# cross_converter.py
# Based on the file conventions documented by crossme 
# at https://github.com/nelhage/crossme/blob/master/docs/file_format.md

import json
import sys
import os
from typing import List

class CrosswordConverter:
    __MAGIC = "ACROSS&DOWN"

    def __init__(self):
        self.__solution = []
        self.__width = 0
        self.__height = 0
        self.__num_clues = 0
        self.__clues = []
        self.__sorted_clues = {}
        self.__binary = None
        self.__puzzle = []

# **********************************************************************
# Public methods
# **********************************************************************

    def read_puz(self, input_file: str):
        self.__binary = open(in_file, "rb")

        # Puzzle information
        self.__binary.seek(0x2C)
        self.__width = int.from_bytes(self.__binary.read(0x1), "big")
        self.__height = int.from_bytes(self.__binary.read(0x1), "big")
        self.__num_clues = int.from_bytes(self.__binary.read(0x2), "little")

        offset = 0x34
        self.__solution = self.__get_solution(offset)

        offset += 2 * self.__width * self.__height
        self.__clues = self.__get_clues(offset)

        self.__organize_clues()

    def write_json(self, output_file: str):
        json_data = self.__encode_puzzle_data_as_json()
        with open(output_file, "w") as json_file:
            json_file.write(json_data)

# **********************************************************************
# Private methods
# **********************************************************************

    def __check_magic(self):
        self.__binary.seek(0x02)
        magic = self.__binary.read(0x0B).decode("iso-8859-1")
        assert (magic == self.__MAGIC), "Invalid magic, cannot read file."

    def __read_next_string(self) -> str:
        byte = self.__binary.read(0x01)
        byte_array = []
        while byte != b'\x00':
            byte_array.append(byte)
            byte = self.__binary.read(0x01)

        next_string = b"".join(byte_array).decode("iso-8859-1")
        return next_string

    def __get_clues(self, offset: int) -> List[str]:
        self.__binary.seek(offset)
        title = self.__read_next_string()
        self.__binary.read(0x02)
        author = self.__read_next_string()
        self.__binary.read(0x02)
        copywrite = self.__read_next_string()

        clues = []
        for i in range(self.__num_clues):
            clue = self.__read_next_string().strip()
            clues.append(clue)
        return clues

    def __get_solution(self, offset: int) -> List[List[str]]:
        self.__binary.seek(offset)
        solution = []
        for row in range(self.__height):
            solution.append([])
            for column in range(self.__width):
                solution_item = self.__binary.read(0x01).decode("iso-8859-1")
                if solution_item == ".":
                    solution_item = "#"
                solution[row].append(solution_item)
        return solution

    def __is_wall(self, x: int, y: int) -> bool:
        if x in range(0, self.__height) and y in range(0, self.__width):
            return (self.__solution[x][y] == "#")
        return False

    def __cell_needs_down_number(self, x: int, y: int) -> bool:
        if ((x == 0) or self.__is_wall(x-1, y)):
            if (((x+1) < self.__width) and (not self.__is_wall(x+1, y))):
                return True
        return False

    def __cell_needs_across_number(self, x: int, y: int) -> bool:
        if ((y == 0) or self.__is_wall(x, y-1)):
            if (((y+1) < self.__height) and (not self.__is_wall(x, y+1))):
                return True
        return False

    def __organize_clues(self):
        across_clues = []
        down_clues = []
        current_cell_number = 1
        clue_index = 0

        for x in range(self.__height):
            self.__puzzle.append([])
            for y in range(self.__width):
                self.__puzzle[x].append(0)

        for x in range(self.__height):
            for y in range(self.__width):
                if self.__is_wall(x, y):
                    self.__puzzle[x][y] = "#"
                else:
                    assigned_number = False
                    if self.__cell_needs_across_number(x, y):
                        across_clues.append([current_cell_number, self.__clues[clue_index]])
                        clue_index += 1
                        assigned_number = True
                    if self.__cell_needs_down_number(x, y):
                        down_clues.append([current_cell_number, self.__clues[clue_index]])
                        clue_index += 1
                        assigned_number = True
                    if assigned_number:
                        self.__puzzle[x][y] = current_cell_number
                        current_cell_number += 1

        self.__sorted_clues["across"] = across_clues
        self.__sorted_clues["down"] = down_clues

    def __encode_puzzle_data_as_json(self):
        json_string = "{\n"
        json_string += "\t\"dimensions\": {\n"
        json_string += "\t\t\"width\": " + str(self.__width) + ",\n"
        json_string += "\t\t\"height\": " + str(self.__height) + "\n"
        json_string += "\t},\n"

        json_string += self.__encode_solution()
        json_string += self.__encode_puzzle()
        json_string += self.__encode_clues()

        json_string += "}"
        return json_string

    def __encode_solution(self):
        json_string = "\t\"solution\": " + "[\n"
        for i in range(len(self.__solution)):
            row = self.__solution[i]
            json_string += "\t\t["
            for j in range(len(row)):
                json_string += json.dumps(row[j])
                if (j+1) != len(row):
                    json_string += ", "
            json_string += "]"
            if (i+1) != len(self.__solution):
                json_string += ","
            json_string += "\n"
        json_string += "\t],\n"
        return json_string

    def __encode_puzzle(self):
        json_string = "\t\"puzzle\": " + "[\n"
        for i in range(len(self.__puzzle)):
            json_string += "\t\t" + json.dumps(self.__puzzle[i])
            if (i+1) != len(self.__puzzle):
                json_string += ","
            json_string += "\n"
        json_string += "\t],\n"
        return json_string

    def __encode_clues(self):
        json_string = "\t\"clues\": {\n"
        for i, key in enumerate(self.__sorted_clues):
            json_string += "\t\t" + json.dumps(key) + ": [\n"
            values = self.__sorted_clues[key]
            for j in range(len(values)):
                json_string += "\t\t\t" + json.dumps(values[j], ensure_ascii=False)
                if (j+1) != len(values):
                    json_string += ","
                json_string += "\n"
            json_string += "\t\t]"
            if (i+1) != len(self.__sorted_clues):
                json_string += ","
            json_string += "\n"

        json_string += "\t}\n"
        return json_string

in_directory = sys.argv[1]
out_directory = sys.argv[2]
for filename in os.listdir(in_directory):
    in_file = os.path.join(in_directory, filename)
    if os.path.isfile(in_file):
        filename_with_no_extension, extension = os.path.splitext(filename)
        if extension == ".puz":
            out_filename = filename_with_no_extension + ".json"
            out_file = os.path.join(out_directory, out_filename)
            converter = CrosswordConverter()
            converter.read_puz(in_file)
            converter.write_json(out_file)
