--- Animation
-- Holds frame index information to build a sprite frame animation.
-- @author Aaron Barrett

-- Required to create an AHAnimate action.  To create an Animation object,
-- drop a sequence of frames into a spritesheet folder and rebuild using
-- `./gradlew build`.  The format of the frames should be
-- `arbitrary_name_000` to `arbitrary_name_###`.  You must use three-digit
-- frame numbers.  The `Rename Finder Items`/`Make Sequential` task in
-- Automator may be helpful.

-- ===========================================================================
class("Animation")
-- ===========================================================================


-- ----------------------------------------------------------------------------
-- Data --
-- ----------------------------------------------------------------------------

local cachedAnimations = {}

-- ----------------------------------------------------------------------------
-- Utility --
-- ----------------------------------------------------------------------------

local function countFrames(animationPrefix)
  local frameCount = 0
  repeat
    frameCount = frameCount + 1
    local frameNumberString = string.format("%03d", frameCount)
    local frameName = animationPrefix .. frameNumberString .. ".png"
    local foundDeck = Sprite.deckAndIndexOfSpritesheetSprite(frameName)
  until (not foundDeck)

  return frameCount
end

-- ----------------------------------------------------------------------------
-- Animation API --
-- ----------------------------------------------------------------------------

--- Call *before* calling Animation.newWithPrefix() to cache the animation
-- @param animationPrefix (string) - The animation name before the numbers and ".png" (something like "fish_swim_")
function Animation.loadAnimationWithPrefix(animationPrefix)
  local animation = Animation.newWithPrefix(animationPrefix)
  cachedAnimations[animationPrefix] = animation
end

--- Uncache the animation with the given prefix
-- @param animationPrefix (string) - The animation name before the numbers and ".png" (something like "fish_swim_")
function Animation.uncacheAnimationWithPrefix(animationPrefix)
  cachedAnimations[animationPrefix] = nil
end

-- ----------------------------------------------------------------------------
-- API --
-- ----------------------------------------------------------------------------

--- Create a sprite with any frame of the animation.
-- @param frameIndex (integer, default 1) - the index of the frame to use to create the sprite.
-- @return A sprite of the frame at frameIndex
function Animation:createSpriteFromFrame(frameIndex)
  frameIndex = frameIndex or 1
  local offsetFrameIndex = self.lowestFrameNumber + frameIndex - 1
  local deck = self.deck

  return Sprite.newWithDeck(deck, offsetFrameIndex)
end

--- Gets the number of frames in the animation
-- @return The number of frames in the animation
function Animation:getFrameCount()
  return self.frameCount
end

-- ----------------------------------------------------------------------------
-- Object Creation --
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Animation object which stores a sequence of frames loaded from a single
-- spritesheet for animating with the animate action helper.  @see AHAnimate
-- Note that creating an Animation does not automatically cache it.  @see loadAnimationWithPrefix
-- @param animationPrefix (string) - The animation name before the numbers and ".png" (something like "fish_swim_")
-- @return (animation) The created animation object
function Animation.newWithPrefix(animationPrefix)
  assert(animationPrefix, "animationPrefix cannot be nil.")

  local self = cachedAnimations[animationPrefix]
  if (self) then
    return self
  end

  local firstFrameName = animationPrefix .. "000.png"
  local deck, firstIndex = Sprite.deckAndIndexOfSpritesheetSprite(firstFrameName)
  assert(deck, "Animation frame not found: " .. tostring(firstFrameName))

  self = {}
  Mixin.mix(self, Animation)
  self.frameCount = countFrames(animationPrefix)
  self.deck = deck
  self.lowestFrameNumber = firstIndex

  return self
end
