--- AudioTimer
-- @author Created by karl on July 06, 2015.

class("AudioTimer")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startCoroutine(self)
  if not self._AudioTimer_coroutine then
    self._AudioTimer_coroutine = runInCoroutine(function()
      local sound = self._AudioTimer_sound
      -- Note: lastTime/thisTime is needed to counter a bug in getTime(), which
      -- will return very small values just before the sound finishes playing.
      local lastTime = 0
      while not sound:isFinished() do
        local thisTime = sound:getTime()
        if thisTime > lastTime then
          self._AudioTimer_updateHandler()
          lastTime = thisTime
        end
        coroutine.yield()
      end
      self._AudioTimer_timeoutHandler()
      self._AudioTimer_coroutine = nil
    end)
  end
end

local function stopCoroutine(self)
  if self._AudioTimer_coroutine then
    self._AudioTimer_coroutine:stop()
    self._AudioTimer_coroutine = nil
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function AudioTimer:getTimeRemaining()
  local currentTime = 0
  if self._AudioTimer_sound then
    currentTime = self._AudioTimer_sound:getTime()
  end
  local remaining = self._AudioTimer_duration - currentTime
  if remaining < 0 then
    remaining = 0
  end
  return remaining
end

function AudioTimer:getNormalizedTimeRemaining()
  return self:getTimeRemaining() / self._AudioTimer_duration
end

function AudioTimer:start()
  self._AudioTimer_sound = Sound.playLongformSoundWithPath(self._AudioTimer_soundPath)
  startCoroutine(self)
end

function AudioTimer:stop()
  if self._AudioTimer_sound then
    self._AudioTimer_sound:stop()
  end
  stopCoroutine(self)
  self._AudioTimer_updateHandler()
end

function AudioTimer:pause()
  self:setPaused(true)
end

function AudioTimer:resume()
  self:setPaused(false)
end

function AudioTimer:isPlaying()
  return self._AudioTimer_sound and not self._AudioTimer_sound:isFinished()
end

function AudioTimer:setPaused(isPaused)
  if isPaused ~= self._AudioTimer_sound:isPaused() then
    self._AudioTimer_sound:setPaused(isPaused)
    if isPaused then
      stopCoroutine(self)
    else
      startCoroutine(self)
    end
  end
end

function AudioTimer:getUpdateHandler()
  return self._AudioTimer_updateHandler
end

function AudioTimer:setUpdateHandler(handler)
  self._AudioTimer_updateHandler = handler or function() end
end

function AudioTimer:getTimeoutHandler()
  return self._AudioTimer_timeoutHandler
end

function AudioTimer:setTimeoutHandler(handler)
  self._AudioTimer_timeoutHandler = handler or function() end
end

function AudioTimer:getDuration()
  return self._AudioTimer_duration
end

function AudioTimer:getSound()
  return self._AudioTimer_sound
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make an audio timer.
-- @param self (table) - Object to turn into a timer.
-- @param soundPath (string) - Absolute path to the sound file that will be played.
-- @param duration (numeric) - Duration of the sound file in seconds.
function AudioTimer.makeTimer(self, soundPath, duration)
  Mixin.mix(self, AudioTimer)

  self._AudioTimer_soundPath = soundPath
  self._AudioTimer_duration = duration
  self:setUpdateHandler(nil)
  self:setTimeoutHandler(nil)

  return self
end
