--- BaseGame
--
-- Every game inherits from BaseGame.  Each instance of BaseGame contains an
-- instance of UIOverlay.
--
-- When the game is started from GameLauncher, it will wait for a short delay to
-- give the fade-in transition from the App time to complete.  After this delay,
-- gameplay starts. Background music starts playing, the UIOverlay Widgets fade in,
-- and gameFlow() is called.
--
-- Before the game calls GameSession.notifyGameWin/Loss, it should start the defualt
-- win/loss outro action and wait for it to complete.  This needs to happen so that
-- UIOverlay Widgets do not clutter the screenshot that appears behind the Post Game.
-- Games should also take it upon themselves to obscure or fade out any bright or
-- distracting elements such as text.
--
-- @author Created by Aaron Barrett on September 27, 2013.

-- ===========================================================================
class("BaseGame")
-- ===========================================================================

require("Analytics")
require("Localizer")
require("UIOverlay")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local APP_FADE_IN_DELAY_DURATION = 0.4

local EVENT_GAME_AUDIO_REQUIRED_SCREEN = "GameAudioRequiredScreen"

local Z_UI_OVERLAY = 10000

local PROPERTIES = require("properties")

local locale = GameSession.getCurrentLocale()
Localizer.setLocale(locale)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

--- Gets this game's UIOverlay
-- @return the UIOverlay instance
function BaseGame:getUIOverlay()
  return self.uiOverlay
end

-- Called from GameLauncher -----------

--- Determines whether a game's UI overlay should fade in automatically at the
-- beginning of the game.
-- It may be useful to override this method to return `false` when an in-game
-- tutorial needs to be shown, and the game's UI overlay needs to start in a
-- hidden state.
function BaseGame:isUIAutoFadeInEnabled()
  return true
end

--- Called automatically by GameLauncher. Waits for 0.4 seconds
-- (to give the fade in from App land time to complete), then Simultaneously
-- fades in the UIOverlay, starts the background music (if any),
-- and calls gameFlow().
function BaseGame:startGame()
  local function showIntroAndStartGameFlow()
    waitFor(APP_FADE_IN_DELAY_DURATION)
    if (self:isUIAutoFadeInEnabled()) then
      self.uiOverlay:fadeIn()
    end

    self.isGameRunning = true
    GameSession.notifyGameTimeStart()
    GameSession.notifyGameStart()
    if (self.uiOverlay.pauseButton) then
      self.uiOverlay.pauseButton:setTouchEnabled(true)
    end
    self:gameFlow()
  end

  self.uiOverlay:prepareForGameIntro()
  if PROPERTIES.requiresAudio then
    require("widgets.AudioRequiredOverlay")
    local overlay = AudioRequiredOverlay.new(function()
      runInCoroutine(showIntroAndStartGameFlow)
    end)
    Screen.rootGroup:insertAtCenter(overlay)
    overlay:show()
    Analytics.reportEvent(EVENT_GAME_AUDIO_REQUIRED_SCREEN)
  else
    runInCoroutine(showIntroAndStartGameFlow)
  end
end

-- Default game endings ---------------

--- Called automaticaly by showDefaultGameWinOutro() and showDefaultGameLossOutro().
-- Tells added UIOverlay Widgets to stop any ongoing actions and prepare to fade out.
function BaseGame:prepareForGameOutro()
  self.uiOverlay:prepareForGameOutro()
end

--- Creates an action that currently just fades out all added UIOverlay widgets.
-- @return the default outro action
function BaseGame:createDefaultGameWinOutroAction()
  -- Disable the default fade action across all games so that each game can control its outro timing.
  -- Added for extra challenges v1; will more than likely be changed again so keeping this method
  -- around until a larger decision is made.
  local defaultGameWinOutroAction = AHDoNothing.new()--self.uiOverlay:createFadeOutAction()

  return defaultGameWinOutroAction
end

--- Calls prepareForGameOutro() and starts the default win outro action.
-- @return the started default outro action
function BaseGame:showDefaultGameWinOutro()
  self:prepareForGameOutro()

  return self:createDefaultGameWinOutroAction():start()
end

--- Creates an action that currently just fades out all added UIOverlay widgets.
function BaseGame:createDefaultGameLossOutroAction()
  local defaultGameLossOutroAction = self.uiOverlay:createFadeOutAction()

  return defaultGameLossOutroAction
end

--- Calls prepareForGameOutro() and starts the default loss outro action.
-- @return the started default outro action
function BaseGame:showDefaultGameLossOutro()
  self:prepareForGameOutro()

  return self:createDefaultGameLossOutroAction():start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new instance of BaseGame.  Called automatically by GameLauncher.
-- @return the new BaseGame instance.  Add stuff to it and make it into a fun game.
function BaseGame.new()
  local self = Group.new(Screen.width, Screen.height)
  Mixin.mix(self, BaseGame)

  self.uiOverlay = UIOverlay.new()
  self:insertAtCenter(self.uiOverlay, Z_UI_OVERLAY)

  return self
end
