--- ContentTracker
-- @author Created by Eduardo Monroy on April 20, 2016

-- =============================================================================
class("ContentTracker")
-- =============================================================================

require("utils.TableUtils")

-- -----------------------------------------------------------------------------
-- Data
-- -----------------------------------------------------------------------------

local CONTENT_TRACKING_RESULTS = {
  CORRECT = "correct",
  INCORRECT = "incorrect",
  TIMEOUT = "timeout",
  SKIPPED = "skipped"
}

-- -----------------------------------------------------------------------------
-- Utility
-- -----------------------------------------------------------------------------

local function onContentTracked(self)
  if self.onContentTrackedCallback ~= nil then
    self.onContentTrackedCallback()
  end
end

local function setCurrentPresentation(self, conceptIdentifier, conceptDifficulty)
  Asserts.assertType(conceptIdentifier, "conceptIdentifier", "string")
  Asserts.assertOptionalType(conceptDifficulty, "conceptDifficulty", "string")
  if (self.seenConcepts[conceptIdentifier] ~= nil and self.shouldSquashData) then
    self.currentPresentation = self.seenConcepts[conceptIdentifier]
  else
    self.currentPresentation = {
      concept_identifier = conceptIdentifier,
      concept_difficulty = conceptDifficulty
    }
    self.seenConcepts[conceptIdentifier] = self.currentPresentation
    table.insert(self.contentTrackingData, self.currentPresentation)
    onContentTracked(self)
  end
end

local function isResultValid(result)
  return TableUtils.contains(CONTENT_TRACKING_RESULTS, result)
end

local function insertAnswer(self, conceptIdentifier, questionIdentifier, userInput, result)
  Asserts.assertType(conceptIdentifier, "conceptIdentifier", "string")
  Asserts.assertOptionalType(questionIdentifier, "questionIdentifier", "string")
  Asserts.assertOptionalType(userInput, "userInput", "string")
  assert(isResultValid(result), "Invalid result: " .. result)

  if self.currentPresentation == nil or self.currentPresentation.concept_identifier ~= conceptIdentifier then
    setCurrentPresentation(self, conceptIdentifier)
  end

  local answerData = {
    question_identifier = questionIdentifier,
    user_input = userInput,
    result = result
  }

  self.currentPresentation.answers_data = self.currentPresentation.answers_data or {}
  table.insert(self.currentPresentation.answers_data, answerData)
  onContentTracked(self)
end

-- -----------------------------------------------------------------------------
-- API
-- -----------------------------------------------------------------------------

-- conceptDifficulty (string) is optional, will be set to nil if excluded
-- nil values get excluded when table is encoded to a json
function ContentTracker:notifyPresentedConcept(conceptIdentifier, conceptDifficulty)
  setCurrentPresentation(self, conceptIdentifier, conceptDifficulty)
end

function ContentTracker:notifyCorrectAnswer(conceptIdentifier, questionIdentifier, userInput)
  insertAnswer(self, conceptIdentifier, questionIdentifier, userInput, CONTENT_TRACKING_RESULTS.CORRECT)
end

function ContentTracker:notifyIncorrectAnswer(conceptIdentifier, questionIdentifier, userInput)
  insertAnswer(self, conceptIdentifier, questionIdentifier, userInput, CONTENT_TRACKING_RESULTS.INCORRECT)
end

function ContentTracker:notifyTimeout(conceptIdentifier, questionIdentifier)
  insertAnswer(self, conceptIdentifier, questionIdentifier, nil, CONTENT_TRACKING_RESULTS.TIMEOUT)
end

function ContentTracker:notifySkippedAnswer(conceptIdentifier, questionIdentifier)
  insertAnswer(self, conceptIdentifier, questionIdentifier, nil, CONTENT_TRACKING_RESULTS.SKIPPED)
end

function ContentTracker:getContentTrackingData()
  return self.contentTrackingData
end

function ContentTracker:setSquashData(shouldSquashData)
  self.shouldSquashData = shouldSquashData
end

function ContentTracker:clearSeenConcepts()
  self.seenConcepts = {}
end

-- -----------------------------------------------------------------------------
-- Object creation
-- -----------------------------------------------------------------------------

function ContentTracker.new(onContentTrackedCallback)
  local self = {}
  Mixin.mix(self, ContentTracker)

  self.seenConcepts = {}
  self.contentTrackingData = {}
  self.currentPresentation = nil;

  self.shouldSquashData = false
  self.onContentTrackedCallback = onContentTrackedCallback

  return self
end
