-- Coroutine functions


--- Convenience function to create a coroutine with the specified function.
-- The coroutine will not run until you call start() on it.
function newCoroutineWithFunction(aFunction)
  local routine = MOAICoroutine.new()
  routine:setFunction(aFunction)
  return routine
end

-- Start a coroutine running the specified function and return it.
function runInCoroutine(aFunction)
  local routine = MOAICoroutine.new()
  routine:run(aFunction)
  return routine
end

--- Start a coroutine running exclusively.
-- Exclusivity means that no two coroutines with the same key will run at the same
-- time. You may optionally specify for the function to either abort running or
-- queue up and run once all other functions with the same key have run.
-- Queueing is done in FIFO order.
-- If you elect not to queue, and there's already a function with the specified key
-- running, this function will return immediately.
--
-- @param self An object to hold the coroutine queues
-- @param key The key to associate this function with
-- @param queueIfBusy If true, queue the call when there's already a function running.
-- @param aFunction The function to run
function runExclusivelyInCoroutine(self, key, queueIfBusy, aFunction)
  if not self.runInCoroutine_queues then
    self.runInCoroutine_queues = {}
  end
  local queue = self.runInCoroutine_queues[key]
  if not queue then
    queue = {}
    self.runInCoroutine_queues[key] = queue
  end

  return runInCoroutine(function()
    local isFirstEntry = #queue == 0
    if queueIfBusy or isFirstEntry then
      table.insert(queue, aFunction)
      if isFirstEntry then -- Only drain the queue if we're the first.
        while #queue > 0 do
          queue[1]()
          table.remove(queue, 1)
        end
      else
        while #queue > 0 do
          coroutine.yield()
        end
      end
    end
  end)
end
