--- DeviceKeyboard
-- @author Created by Aaron Barrett on September 21, 2015.

class("DeviceKeyboard")

require("DesktopKeyboard")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_INPUT_CALLBACK = function(updatedText) end
local DEFAULT_INPUT_REJECTED_CALLBACK = function(updatedText) end
local DEFAULT_RETURN_CALLBACK = function(currentText) end
local DEFAULT_VALIDATE_CALLBACK = function(replacementText) error("Must define callback to accept or reject input text using `DeviceKeyboard.setValidateCallback()`") end
local DEFAULT_WILL_APPEAR_WITH_DIMENSIONS_CALLBACK = function(originXProportion, originYProportion, widthProportion, heightProportion) end
local DEFAULT_RECEIVED_LOCALE_CALLBACK = function(locale) error("Must define callback to handle keyboard locale using `DeviceKeyboard.setReceivedLocaleCallback()`") end

DeviceKeyboard.currentText = ""
DeviceKeyboard._isEnabled = true

DeviceKeyboard.inputCallback = DEFAULT_INPUT_CALLBACK
DeviceKeyboard.inputRejectedCallback = DEFAULT_INPUT_REJECTED_CALLBACK
DeviceKeyboard.returnCallback = DEFAULT_RETURN_CALLBACK
DeviceKeyboard.validateCallback = DEFAULT_VALIDATE_CALLBACK
DeviceKeyboard.willAppearWithDimensionsCallback = DEFAULT_WILL_APPEAR_WITH_DIMENSIONS_CALLBACK
DeviceKeyboard.receivedLocaleCallback = DEFAULT_RECEIVED_LOCALE_CALLBACK

-- NOTE: This is a workaround for the cases on the various versions of iOS and Android
-- where the keyboard text field loses focus.  In a normal app, the user could simply
-- tap on the native text box to make the keyboard display again.  Since the text field
-- is hidden in our games, however, this is not possible.  As a workaround, this action
-- continually asks the app to re-focus the keyboard text field in case the keyboard
-- became hidden.  It's only safe to remove this action if the problem has been solved on
-- both platforms.
DeviceKeyboard.reshowKeyboardAction = AHLoop.new(
  AHDelayed.new(1.0, function()
    Host.showKeyboard()
  end)
)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function DeviceKeyboard.show()
  Host.showKeyboard()
  DesktopKeyboard.setEnabled(true)
  DeviceKeyboard.reshowKeyboardAction:start()
end

function DeviceKeyboard.hide()
  DeviceKeyboard.reshowKeyboardAction:stop()
  Host.hideKeyboard()
  DesktopKeyboard.setEnabled(false)
end

function DeviceKeyboard.getText()
  return DeviceKeyboard.currentText
end

function DeviceKeyboard.setText(text)
  Host.setKeyboardTextFieldText(text)
  DeviceKeyboard.currentText = text
  DeviceKeyboard.inputCallback(DeviceKeyboard.currentText)
end

function DeviceKeyboard.clear()
  DeviceKeyboard.setText("")
end

function DeviceKeyboard.isEnabled()
  return DeviceKeyboard._isEnabled
end

function DeviceKeyboard.setEnabled(isEnabled)
  DeviceKeyboard._isEnabled = isEnabled
end

-- NOTE: getting the keyboard locale is set up through `requestLocale` and `setReceivedLocaleCallback` because
-- it was inconvenient to set it up as a regular `getLocale` method on Android.  Please feel free to refactor.
function DeviceKeyboard.requestLocale()
  Host.requestKeyboardLocale()
end

function DeviceKeyboard.onKeyboardInput(replacementStartIndex, replacementLength, replacementText)
  if DeviceKeyboard.isEnabled() then
    local leftText = string.sub(DeviceKeyboard.currentText, 1, replacementStartIndex)
    local rightText = string.sub(DeviceKeyboard.currentText, replacementStartIndex + replacementLength + 1, #(DeviceKeyboard.currentText))
    local resultingText = leftText .. replacementText .. rightText

    local wasAccepted = DeviceKeyboard.validateCallback(replacementText, resultingText)
    if wasAccepted then
      DeviceKeyboard.currentText = resultingText
      runInCoroutine(function() DeviceKeyboard.inputCallback(DeviceKeyboard.currentText) end)
      return true -- Return values are REQUIRED by the host system.
    else
      runInCoroutine(function() DeviceKeyboard.inputRejectedCallback(replacementText, resultingText) end)
    end
  end
  return false
end

function DeviceKeyboard.onKeyboardReturn()
  if DeviceKeyboard.isEnabled() then
    runInCoroutine(function() DeviceKeyboard.returnCallback(DeviceKeyboard.currentText) end)
  end
end

function DeviceKeyboard.onKeyboardWillAppearWithDimensions(originXProportion, originYProportion, widthProportion, heightProportion)
  runInCoroutine(function()
    DeviceKeyboard.willAppearWithDimensionsCallback(originXProportion, originYProportion, widthProportion, heightProportion)
  end)
end

function DeviceKeyboard.onReceivedKeyboardLocale(locale)
  runInCoroutine(function() DeviceKeyboard.receivedLocaleCallback(locale) end)
end

function DeviceKeyboard.setInputCallback(callback)
  DeviceKeyboard.inputCallback = callback or DEFAULT_INPUT_CALLBACK
end

function DeviceKeyboard.setInputRejectedCallback(callback)
  DeviceKeyboard.inputRejectedCallback = callback or DEFAULT_INPUT_REJECTED_CALLBACK
end

function DeviceKeyboard.setReturnCallback(callback)
  DeviceKeyboard.returnCallback = callback or DEFAULT_RETURN_CALLBACK
end

function DeviceKeyboard.setValidateCallback(callback)
  DeviceKeyboard.validateCallback = callback or DEFAULT_VALIDATE_CALLBACK
end

function DeviceKeyboard.createCharacterBasedValidationCallback(validCharactersInLowercase)
  local validCharacters = TableUtils.createSetFromString(validCharactersInLowercase)
  return function(replacementText)
    local lowerCaseReplacementText = string.lower(replacementText)
    for character in string.gmatch(lowerCaseReplacementText, ".") do
      local characterIsPrintable = validCharacters[character]
      if (not characterIsPrintable) then
        return false
      end
    end
    return true
  end
end

function DeviceKeyboard.setWillAppearWithDimensionsCallback(callback)
  DeviceKeyboard.willAppearWithDimensionsCallback = callback or DEFAULT_WILL_APPEAR_WITH_DIMENSIONS_CALLBACK
end

function DeviceKeyboard.setReceivedLocaleCallback(callback)
  DeviceKeyboard.receivedLocaleCallback = callback or DEFAULT_RECEIVED_LOCALE_CALLBACK
end
