--- Experiment
-- @author Created by aaroncantor on August 10, 2020.

class("Experiment")
require("utils.TableUtils")
require("Analytics")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_VARIANT = "default"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isVariantRecognized(self, variant)
  return ((variant ~= nil) and TableUtils.contains(self.allVariants, variant))
end

local function shouldTreatVariantAsControl(self, variant)
  if ((variant == self.control) or (variant == DEFAULT_VARIANT)) then
    return true
  end
  if (not isVariantRecognized(self, variant)) then
    Log.warn(string.format("Checking unrecognized variant, %s, for experiment %s. Defaulting to control experience.", tostring(variant), tostring(self.name)))
    return true
  end
  return false
end

local function getActiveVariant(self)
  if (self.activeVariant == nil) then
    local variant = Host.getExperimentVariant(self.name)
    Log.verbose(string.format("Getting variant for %s: %s", tostring(self.name), tostring(variant)))
    self.activeVariant = variant
  end
  return self.activeVariant
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Experiment:isControl()
  local activeVariant = getActiveVariant(self)
  return shouldTreatVariantAsControl(self, activeVariant)
end

function Experiment:isActive()
  return (not self:isControl())
end

function Experiment:isVariantActive(variant)
  assert(isVariantRecognized(self, variant), string.format("Attempted to check experiment %s with the invalid variant %s", tostring(self.name), tostring(variant)))
  return (variant == getActiveVariant(self))
end

function Experiment:reportExperimentExposed()
  if (not self.hasSentExposure) then
    Log.verbose(string.format("Experiment Exposed: %s", tostring(self.name)))
    Host.reportExperimentExposureEvent(self.name)

    self.hasSentExposure = true
  else
    Log.verbose(string.format("Ignoring duplicate Experiment Exposure: %s", tostring(self.name)))
  end
end

-- ----------------------------------------------------------------------------
-- For Manual Testing
-- ----------------------------------------------------------------------------

function Experiment:setActiveVariant(variant)
  self.activeVariant = variant
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function validateExperiment(name, control, variants)
  assert(((type(name) == "string") and (#name > 0)), "Experiment must have a valid name, not " .. StringUtils.tostring(name))
  assert(((type(control) == "string") and (#control > 0)), name .. " experiment must have a valid control, not " .. StringUtils.tostring(control))
  assert((#variants > 0), name .. " experiment must have at least one variant.")
  for _, variant in ipairs(variants) do
    assert(((type(variant) == "string") and (#variant > 0)), name .. "experiment has an invalid variant: " .. StringUtils.tostring(variant))
  end
end

-- Constructor ------------------------

function Experiment.new(name, control, variants)
  if (type(variants) ~= "table") then
    variants = {variants}
  end
  validateExperiment(name, control, variants)
  local self = {}
  Mixin.mix(self, Experiment)

  self.name = name
  self.control = control
  self.variants = variants
  self.allVariants = TableUtils.createConcatenatedArray({self.control}, self.variants)

  self.activeVariant = nil
  self.hasSentExposure = false

  return self
end
