--- GameSession

require("data.generators.ConceptGenerator")
require("difficulty.DifficultyCurves")
require("bonuses.BonusReporter")
require("ContentTracker")
require("Timer")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function initializeGameSession()
  -- setup the GameSession.config property
  local configJSON = GameSession.configJSON

  if (configJSON:len() > 0) then
    GameSession.config = MOAIJsonParser.decode(configJSON)
  else
    GameSession.config = {}
  end

  -- Reserved GameSession fields
  GameSession.score = 0
  GameSession.passScore = nil
  GameSession.secondRankScore = nil
  GameSession.thirdRankScore = nil
  GameSession.rank = 0
  GameSession.bonusScores = {}
  GameSession.reportingMap = {}
  GameSession.pretestResult = {}
  GameSession.interestsMap = {}
  GameSession.topInterest = "" -- Currently unused
  GameSession.personalizationMap = {}
  GameSession.bonuses = {}
  GameSession.initialLives = 0
  GameSession.perfectSetsMaximum = 0
  GameSession.gameDuration = 0
  GameSession.totalTime = 0
  GameSession.correctAnswerCount = 0
  GameSession.incorrectAnswerCount = 0
  GameSession.contentTrackingJson = "[]"
  GameSession.bonusTrackingProperties = {}

  GameSession.conceptGenerator = ConceptGenerator.new(GameSession.conceptChooser)
  GameSession.bonusReporter = BonusReporter.new()
  GameSession.contentTracker = ContentTracker.new(GameSession.notifyContentTracking)

  GameSession.gameDurationTimer = MOAITimer.new()
  GameSession.gameDurationTimer:setMode(MOAITimer.CONTINUE)
end

local function setGameDuration()
  GameSession.gameDuration = GameSession.gameDurationTimer:getTime()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Pass conditions and scoring --------------------------------------------------

function GameSession.setPassScore(passScore)
  local difficultyAdjustedPassScore = ScoringUtils.getDifficultyScaledPoints(passScore)

  GameSession.passScore = difficultyAdjustedPassScore
  Log.debug("Setting Pass Score:", GameSession.passScore)
end

function GameSession.getPassScore()
  return GameSession.passScore
end

function GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)
  local difficultyAdjustedRankTwoScore = ScoringUtils.getDifficultyScaledPoints(rankTwoScore)
  local difficultyAdjustedRankThreeScore = ScoringUtils.getDifficultyScaledPoints(rankThreeScore)

  GameSession.secondRankScore = difficultyAdjustedRankTwoScore
  GameSession.thirdRankScore = difficultyAdjustedRankThreeScore
  Log.debug("Setting Rank 2 Score:", GameSession.secondRankScore)
  Log.debug("Setting Rank 3 Score:", GameSession.thirdRankScore)
end

function GameSession.increaseScore(amount)
  local scoreDelta = ScoringUtils.getDifficultyScaledPoints(amount)
  GameSession.score = math.max(0, GameSession.score + scoreDelta)
  Log.debug(string.format("Increasing score by %d [adjusted to %d]", amount, scoreDelta))

  return GameSession.score
end

function GameSession.scoredCorrectAnswer(conceptID)
  GameSession.notifyCorrectAnswer(conceptID)

  local scoreDelta = GameSession.config.correct_answer_points
  GameSession.increaseScore(scoreDelta)

  return GameSession.score, scoreDelta
end

-- This sends a MOAIGameEventCorrectAnswer to the client, but does not
-- actually mark a concept as seen. Use ConceptGenerator:markConceptIdentifierAsSeen(conceptIdentifier)
-- to mark a concept as seen.
function GameSession.reportSeenConcept(conceptID)
  GameSession.notifyCorrectAnswer(conceptID)
end

function GameSession.scoredIncorrectAnswer(conceptID)
  GameSession.notifyIncorrectAnswer(conceptID)

  local penaltyPoints = GameSession.config.incorrect_answer_points
  if (penaltyPoints) then
    assert(penaltyPoints < 0, "Point penalty must be negative.")
    GameSession.increaseScore(penaltyPoints)
  end

  return GameSession.score
end

function GameSession.setScore(score)
  GameSession.score = score
end

function GameSession.getScore()
  return GameSession.score
end

function GameSession.didReachPassScore()
  assert(GameSession.passScore, "Pass score was not set before checking didReachPassScore. Configure GameSession with setPassScore()")
  local didReach = (math.ceil(GameSession.getScore()) >= math.floor(GameSession.getPassScore()))
  return didReach
end

function GameSession.setInitialLives(lives)
  GameSession.initialLives = lives
end

function GameSession.setPerfectSetsMaximum(count)
  GameSession.perfectSetsMaximum = count
end

function GameSession.setTotalTime(time)
  GameSession.totalTime = time
end

function GameSession.incrementCorrectAnswerCount()
  GameSession.correctAnswerCount = GameSession.correctAnswerCount + 1
end

function GameSession.getCorrectAnswerCount()
  return GameSession.correctAnswerCount
end

function GameSession.incrementIncorrectAnswerCount()
  GameSession.incorrectAnswerCount = GameSession.incorrectAnswerCount + 1
end

function GameSession.getIncorrectAnswerCount()
  return GameSession.incorrectAnswerCount
end

function GameSession.notifyGameTimeStart()
  GameSession.gameDurationTimer:start()
end

function GameSession.notifyGameTimePause()
  GameSession.gameDurationTimer:pause()
end

function GameSession.notifyGameTimeEnd()
  GameSession.gameDurationTimer:pause()
end

function GameSession.calculateRank(didWin)
  local score = GameSession.getScore()
  local passScore = GameSession.passScore or 0 -- Setting a pass score is optional
  local secondRankScore = GameSession.secondRankScore
  local thirdRankScore = GameSession.thirdRankScore

  Asserts.assertType(passScore, "passScore", "number")
  Asserts.assertType(secondRankScore, "secondRankScore", "number")
  Asserts.assertType(thirdRankScore, "thirdRankScore", "number")

  -- Prevent floating-point precision errors.
  score = math.ceil(score)
  passScore = math.floor(passScore)
  secondRankScore = math.floor(secondRankScore)
  thirdRankScore = math.floor(thirdRankScore)

  if (not didWin) then
    GameSession.rank = 0
  elseif (score >= thirdRankScore) then
    GameSession.rank = 3
  elseif (score >= secondRankScore) then
    GameSession.rank = 2
  elseif (score >= passScore) then
    GameSession.rank = 1
  else
    error(string.format("GameSession.calculateRank() resulted in rank 0 with didWin: %s, score: %s, passScore: %s, secondRankScore: %s, thirdRankScore: %s",
      tostring(didWin), tostring(score), tostring(passScore), tostring(secondRankScore), tostring(thirdRankScore)))
  end
end

function GameSession.notifyGameWin()
  setGameDuration()

  local didWin = true
  GameSession.calculateRank(didWin)
  GameSession.notifyBonuses()
  GameSession.notifyGameEndScore()
end

function GameSession.notifyGameLoss()
  setGameDuration()

  local didWin = false
  GameSession.calculateRank(didWin)
  GameSession.notifyBonuses()
  GameSession.notifyGameEndScore()
end

function GameSession.getGameTimeElapsed()
  return GameSession.gameDurationTimer:getTime()
end


-- Difficulty configuration------------------------------

function GameSession.getDifficulty()
  assert(GameSession.difficulty ~= nil, "GameSession.difficulty was not set by the host")
  return GameSession.difficulty
end

function GameSession.getDifficultyScoreMultiplier()
  assert(GameSession.difficultyMultiplier ~= nil, "GameSession.difficultyMultiplier was not set by the host")
  return GameSession.difficultyMultiplier
end

function GameSession.hasNeverWon()
  local timesWon = GameSession.timesWon or 0
  return (timesWon == 0)
end

-- Challenge completed status --------------------

function GameSession.canGameBeSwitched()
  assert(GameSession.canSwitchChallenge ~= nil, "GameSession.challengeAlreadyWon was not set by the host")
  return GameSession.canSwitchChallenge
end

-- Starting position -----------------------------

function GameSession.getStartingPositionIdentifier()
  assert((GameSession.startingPositionIdentifier ~= nil and GameSession.startingPositionIdentifier ~= ""),
         "GameSession.startingPositionIdentifier was not set by the host")
  return GameSession.startingPositionIdentifier
end

-- Crosswords -----------------------------

function GameSession.getCrosswordIdentifier()
  if (GameSession.crosswordIdentifier == nil or GameSession.crosswordIdentifier == "") then
    Log.error("GameSession.crosswordIdentifier was not set by the host")
    return ""
  end
  return GameSession.crosswordIdentifier
end

function GameSession.getCrosswordDifficulty()
  if (GameSession.crosswordDifficulty == nil or GameSession.crosswordDifficulty == "") then
    Log.error("GameSession.crosswordDifficulty was not set by the host")
    return ""
  end
  return GameSession.crosswordDifficulty
end

function GameSession.getCrosswordAutoCheckEnabled()
  if (GameSession.crosswordAutoCheckEnabled == nil) then
    Log.error("GameSession.crosswordAutoCheckEnabled was not set by the host")
    return false
  end
  return GameSession.crosswordAutoCheckEnabled
end

function GameSession.getCrosswordNudgeScreenSeen()
  if (GameSession.crosswordNudgeScreenSeen == nil) then
    Log.error("GameSession.crosswordNudgeScreenSeen was not set by the host")
    return false
  end
  return GameSession.crosswordNudgeScreenSeen
end

function GameSession.getCrosswordIsFromArchive()
  if (GameSession.crosswordIsFromArchive == nil) then
    Log.error("GameSession.crosswordIsFromArchive was not set by the host")
    return false
  end
  return GameSession.crosswordIsFromArchive
end

function GameSession.getCrosswordIsEarningXP()
  if (GameSession.crosswordIsEarningXP == nil) then
    -- This is expected, as we're only implementing this on Android first.
    return false
  end
  return GameSession.crosswordIsEarningXP
end

-- Bonuses ---------------------------------------

function GameSession.notifyBonuses()
  local bonuses = GameSession.bonusReporter:getBonusCounts()

  Asserts.assertType(bonuses, "bonuses", "table")
  for bonusType, value in pairs(bonuses) do
    Asserts.assertType(bonusType, "bonusType", "string")
    Asserts.assertType(value, "value for " .. bonusType, "number")
  end

  GameSession.bonuses = bonuses
end

-- Content tracking ---------------------------------------

function GameSession.notifyContentTracking()
  local contentTrackingData = GameSession.contentTracker:getContentTrackingData()

  local contentTrackingJson = MOAIJsonParser.encode(contentTrackingData)
  GameSession.contentTrackingJson = contentTrackingJson
end

-- Dark Mode ---------------------------------------

function GameSession.getIsDarkMode()
  if (GameSession.isDarkMode == nil) then
    Log.error("GameSession.isDarkMode was not set by the host")
    return false
  end
  local isDarkModeFlagEnabled = ExperimentsConfiguration.DARK_MODE_FLAG:isActive()
  return (isDarkModeFlagEnabled and GameSession.isDarkMode)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

initializeGameSession()
