--- Sortable prop composites
-- Supports inserting children into the group,
-- Performs cleanup when removing children from the group,
-- Helps manage bounds and touch logic
-- @author Created by Alexey Chernikov on Feb 19, 2013

-- ===========================================================================
class("Group")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MINIMUM_WIDTH = 2
local MINIMUM_HEIGHT = 2


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function assertIsMOAIProp(object)
  assert(type(object) == "userdata", "Expected a MOAIProp (got " .. type(sceneObject) .. ")")
end

-- Forward function declaration.
local disposeOfProp = nil

--- Remove an object from the screen hierarchy and render loop.
-- @param object - (prop) the child to remove from the hierarchy
local function removeFromScreen(object)
  object:clearAttrLink(MOAIProp2D.INHERIT_TRANSFORM)
  Screen.remove(object)
  if (not Host.environmentIsAppleTv) then
    TouchManager.removeTouchHandler(object)
  end
end

--- Remove a child from the parent's child list.
-- @param parent (Group) - The parent to remove from.
-- @param child (MOAIProp) - The child to remove.
local function removeFromChildren(parent, child)
  TableUtils.removeValueFromArray(parent.children, child)
end

--- Remove all children from a group.
-- @param object (MOAIProp) - The object to remove children from.
-- @param shouldDispose (boolean, default true) - If true, also dispose of the prop.
local function removeAllChildren(object, shouldDispose)
  if shouldDispose == nil then shouldDispose = true end
  for _, child in ipairs(object.children or {}) do
    removeFromScreen(child)
    child.parent = nil
    if shouldDispose then
      disposeOfProp(child)
    end
  end
  object.children = {}
end

--- Remove a prop from the group and screen hierarchy.
-- @param prop (MOAIProp) - The prop to remove.
-- @param shouldDispose (boolean, default true) - If true, also dispose of the prop.
local function removeProp(prop, shouldDispose)
  if shouldDispose == nil then shouldDispose = true end
  if prop.parent then
      removeFromChildren(prop.parent, prop)
      prop.parent = nil
  end
  removeFromScreen(prop)
  if shouldDispose then
    disposeOfProp(prop)
  end
end

--- Dispose of a prop, removing its children and freeing any resources.
-- @param prop (MOAIProp) - The prop to dispose of.
disposeOfProp = function(prop)
  removeAllChildren(prop, true)
  if prop.disposeOfSelf then
    prop:disposeOfSelf()
  end
end

--- Find the ordinal of a parent's children to insert a child at
-- @param self (Group) - Self.
-- @param zIndex (number) - The Z index to insert at.
local function getOrdinalToInsertChildAt(self, zIndex)
  for index, child in ipairs(self.children) do
    if (zIndex < child.zIndex) then
      return index
    end
  end
  return #self.children + 1
end

--- Automatically choose a zIndex for a new child.
-- @param self (Group) - Self.
-- @param sceneObject (MOAIProp) - The object that will be inserted.
local function getZIndexForNewChild(self, sceneObject)
  if (sceneObject.zIndex) then
    return sceneObject.zIndex
  end
  local topChild = self.children[#self.children]
  if (topChild) then
    return topChild.zIndex + 1
  else
    return 1
  end
end

--- Insert a child with an optional zIndex.
-- If zIndex is nil, it will choose the highest current zIndex + 1
-- @param parent (Group) - Parent to insert object into.
-- @param sceneObject (MOAIProp) - The object to insert.
-- @param zIndex (number, optional) - The Z index to insert at, or nil for default.
local function insertToChildrenWithZIndex(parent, sceneObject, zIndex)
  assertIsMOAIProp(sceneObject)
  zIndex = zIndex or getZIndexForNewChild(parent, sceneObject)
  sceneObject.zIndex = zIndex
  local ordinal = getOrdinalToInsertChildAt(parent, zIndex)
  table.insert(parent.children, ordinal, sceneObject)
  sceneObject:setAttrLink(MOAIProp2D.INHERIT_TRANSFORM, parent, MOAIProp2D.TRANSFORM_TRAIT)
  sceneObject.parent = parent
  sceneObject:forceUpdate()
  Screen.notifyStructureChange()
  return zIndex
end

--- Changes the Z index of a child that has already been added to this group.
-- If zIndex is nil, it will choose the highest current zIndex + 1
-- @param parent - (Group) the parent group.
-- @param child - (prop) the child whose Z index should be changed.
-- @param zIndex - (integer, optional) the new zIndex at which the child should be ordered, or nil for default.
local function setZIndexOfChild(parent, child, zIndex)
  removeFromChildren(parent, child)
  insertToChildrenWithZIndex(parent, child, zIndex)
end

local function get2DBounds(sceneObject)
  local xMin, yMin, _, xMax, yMax, _ = sceneObject:getBounds()
  xMin, yMin = xMin or 0, yMin or 0
  xMax, yMax = xMax or xMin, yMax or yMin
  local w, h = xMax - xMin, yMax - yMin
  return xMin, yMin, xMax, yMax, w, h
end

local function getAnchorPointTransform(sceneObject)
  local xMin, yMin, xMax, yMax, w, h = get2DBounds(sceneObject)

  local ax, ay = sceneObject:getAnchorPoint()
  local dx, dy = -w * ax, -h * ay
  xMin, yMin, xMax, yMax = xMin + dx, yMin + dy, xMax + dx, yMax + dy

  return xMin, yMin, xMax, yMax
end

local function getTransformedBounds(sceneObject)
  local xMin, yMin, xMax, yMax, w, h = get2DBounds(sceneObject)
  local x, y = sceneObject:getLoc()
  local scale = sceneObject:getScl()

  xMin = xMin * scale + x
  xMax = xMin + w * scale
  yMin = yMin * scale + y
  yMax = yMin + h * scale

  return xMin, yMin, xMax, yMax
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Insertion --------------------------

--- Adds a child object to the top of the group (i.e. new child will be rendered before all others)
-- If the child was already part of another group, remove from that group and reinsert it here.
-- NOTE: with the current implementation, touch handlers will be cleared as part of the cleanup procedure upon reinsertion, fix when/if necessary, @see remove
-- @param self - (Group) the group itself
-- @param sceneObject - (prop) the child to add to the group
-- @param zIndex - (optional integer) an index that will influence the sort of children within the group. A higher index means the group is on top of elements with a lower index.
function Group:insert(sceneObject, zIndex)
  assertIsMOAIProp(sceneObject)
  assert(sceneObject.parent == nil, "To change an object's parent, use takeChild()")
  insertToChildrenWithZIndex(self, sceneObject, zIndex)
end

-- Adds a child object to the top of the group and at a specified position
--@param sceneObject - (prop) the child to add to the group
--@param x - (number) the x location to set on the scene object
--@param y - (number) the y location to set on the scene object
--@param zIndex - (optional integer) an index that will influence the sort of children within the group. A higher index means the group is on top of elements with a lower index.
function Group:insertAtLocation(sceneObject, x, y, zIndex)
  Asserts.assertType(x, "x", "number")
  Asserts.assertType(y, "y", "number")
  assertIsMOAIProp(sceneObject)

  sceneObject:setLoc(x, y)
  self:insert(sceneObject, zIndex)
end

-- Adds a child object to the top of the group and at a specified fractional position
--@param sceneObject - (prop) the child to add to the group
--@param widthPart - (number) the fractional x location to set on the scene object based on parent group width
--@param heightPart - (number) the fractional y location to set on the scene object based on parent group height
--@param zIndex - (optional integer) an index that will influence the sort of children within the group. A higher index means the group is on top of elements with a lower index.
function Group:insertAtLocationFractional(sceneObject, widthPart, heightPart, zIndex)
  assertIsMOAIProp(sceneObject)
  Asserts.assertType(widthPart, "widthPart", "number")
  Asserts.assertType(heightPart, "heightPart", "number")

  local width, height = self:getDims()
  assert(width and height, "Group bounds unset. Could not insert child object at fractional location.")

  sceneObject:setLoc(width * widthPart, height * heightPart)
  self:insert(sceneObject, zIndex)
end

function Group:insertAtCenter(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.5, 0.5, zIndex)
end

function Group:insertAtRight(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 1.0, 0.5, zIndex)
end

function Group:insertAtBottomRight(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 1.0, 0.0, zIndex)
end

function Group:insertAtBottom(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.5, 0.0, zIndex)
end

function Group:insertAtBottomLeft(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.0, 0.0, zIndex)
end

function Group:insertAtLeft(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.0, 0.5, zIndex)
end

function Group:insertAtTopLeft(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.0, 1.0, zIndex)
end

function Group:insertAtTop(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 0.5, 1.0, zIndex)
end

function Group:insertAtTopRight(sceneObject, zIndex)
  self:insertAtLocationFractional(sceneObject, 1.0, 1.0, zIndex)
end

function Group:reinsert(sceneObject, zIndex)
  assertIsMOAIProp(sceneObject)
  setZIndexOfChild(self, sceneObject, zIndex)
end

--- Take a child from another object, preserving its physical screen location.
-- If the object has no parent, does a simple insert instead.
function Group:takeChild(sceneObject, zIndex)
  assertIsMOAIProp(sceneObject)
  if sceneObject.parent ~= self then
    if sceneObject.parent then
      sceneObject:setLoc(self:transformObjectLoc(sceneObject))
      removeFromChildren(sceneObject.parent, sceneObject)
      insertToChildrenWithZIndex(self, sceneObject, zIndex)
    else
      self:insert(sceneObject, zIndex)
    end
  end
end


-- Removal ----------------------------

--- Removes this group from it's parent, leaving all the group's children orphans.
-- Cleanup is performed for the children so they can be GC'd
-- @param shouldDispose (boolean, default true) - If true, also dispose of self.
function Group:removeSelf(shouldDispose)
  removeProp(self, shouldDispose)
end

--- Remove a prop. This is kept around for cases where speed optimizations preclude adding
-- group functionality (specifically removeSelf) to children.
-- @param prop (MOAIProp) - The prop to remove.
-- @param shouldDispose (boolean, default true) - If true, also dispose of the prop.
function Group:remove(prop, shouldDispose)
  removeProp(prop, shouldDispose)
end

--- Remove all children from this prop.
-- @param shouldDispose (boolean, default true) - If true, also dispose of the children.
function Group:removeAll(shouldDispose)
  removeAllChildren(self, shouldDispose)
end


-- Z Order ----------------------------

--- Change the Z index of this group among its siblings.
-- @param zIndex (integer) - The new zIndex at which the group should be ordered.
function Group:setZIndex(zIndex)
  setZIndexOfChild(self.parent, self, zIndex)
end

--- Get the highest Z index in this group's children.
function Group:getTopZIndex()
  local topChild = self.children[#self.children]
  if (topChild) then
    return topChild.zIndex
  end
  return 0
end

--- Bring this group to the front among its siblings.
function Group:bringToFront()
  if (self.parent == nil) then return end
  local newZIndex = self.parent:getTopZIndex() + 1
  self:setZIndex(newZIndex)
end

--- Get the lowest Z index in this group's children.
function Group:getBottomZIndex()
  local bottomChild = self.children[1]
  if (bottomChild) then
    return bottomChild.zIndex
  end
  return 0
end

--- Bring this group to the back among its siblings.
function Group:bringToBack()
  if (self.parent == nil) then return end
  local newZIndex = self.parent:getBottomZIndex() - 1
  self:setZIndex(newZIndex)
end


-- Location ---------------------------

function Group:getX()
  local x, _ = self:getLoc()
  return x
end

function Group:setX(x)
  self:setAttr(MOAITransform2D.ATTR_X_LOC, x)
end

function Group:getY()
  local _, y = self:getLoc()
  return y
end

function Group:setY(y)
  self:setAttr(MOAITransform2D.ATTR_Y_LOC, y)
end

function Group:offsetX(deltaX)
  self:setX(self:getX() + deltaX)
end

function Group:offsetY(deltaY)
  self:setY(self:getY() + deltaY)
end

--- Convenience helpers for getting the location of the group edges
function Group:getXLeft()
  local anchorX, _ = self:getAnchorPoint()
  local scaleX, _ = self:getScl()
  return self:getX() - self:getWidthPart(anchorX) * scaleX
end

function Group:getXRight()
  local anchorX, _ = self:getAnchorPoint()
  local scaleX, _ = self:getScl()
  return self:getX() + self:getWidthPart(1 - anchorX) * scaleX
end

function Group:getXCenter()
  local anchorX, _ = self:getAnchorPoint()
  local scaleX, _ = self:getScl()
  return self:getX() + self:getWidthPart(0.5 - anchorX) * scaleX
end

function Group:getYBottom()
  local _, anchorY = self:getAnchorPoint()
  local _, scaleY = self:getScl()
  return self:getY() - self:getHeightPart(anchorY) * scaleY
end

function Group:getYTop()
  local _, anchorY = self:getAnchorPoint()
  local _, scaleY = self:getScl()
  return self:getY() + self:getHeightPart(1 - anchorY) * scaleY
end

function Group:getYCenter()
  local _, anchorY = self:getAnchorPoint()
  local _, scaleY = self:getScl()
  return self:getY() + self:getHeightPart(0.5 - anchorY) * scaleY
end

--- Returns the location of self if it had its current physical screen position,
-- but a different anchor point.
-- For example, if self had an anchor point of (0.5, 0.5) and you called this method
-- with (1, 1), the returned location would be the current location + half self's
-- width and height.
-- @param anchorX, anchorY, anchorZ The hypothetical anchor point.
function Group:getLocationWithAnchorPoint(anchorX, anchorY, anchorZ)
  anchorZ = anchorZ or 0
  local selfX, selfY, selfZ = self:getLoc()
  selfZ = selfZ or 0
  local selfAX, selfAY, selfAZ = self:getAnchorPoint()
  selfAZ = selfAZ or 0
  local selfW, selfH, selfD = self:getEffectiveDims()
  selfD = selfD or 0

  -- Get the location of the 0-most corner in each dimension.
  local loc0X = selfX - selfW * selfAX
  local loc0Y = selfY - selfH * selfAY
  local loc0Z = selfZ - selfD * selfAZ

  -- Apply the user-specified anchor point.
  local finalX = loc0X + selfW * anchorX
  local finalY = loc0Y + selfH * anchorY
  local finalZ = loc0Z + selfD * anchorZ

  return finalX, finalY, finalZ
end

--- Set the location, pretending it has the specified anchor point.
function Group:setLocUsingAnchorPoint(locX, locY, locZ, anchorX, anchorY, anchorZ)
  local selfAX, selfAY, selfAZ = self:getAnchorPoint()
  selfAZ = selfAZ or 0
  local selfW, selfH, selfD = self:getDims()
  anchorX = anchorX or selfAX
  anchorY = anchorY or selfAY
  anchorZ = anchorZ or selfAZ
  locZ = locZ or 0
  selfD = selfD or 0
  local selfX, selfY, selfZ = locX - selfW * anchorX, locY - selfH * anchorY, locZ - selfD * anchorZ

  self:setLoc(selfX, selfY, selfZ)
end

--- Get the distance between two objects on the screen in world coordinates, using the specified anchor points.
-- If the anchor points are omitted, it uses each respective object's current anchor point.
-- @param prop The prop to get the distance to.
-- @param selfAnchorX, selfAnchorY, selfAnchorZ Anchor points for self.
-- @param propAnchorX, propAnchorY, propAnchorZ Anchor points for the prop.
function Group:getWorldDistance(prop, selfAnchorX, selfAnchorY, selfAnchorZ, propAnchorX, propAnchorY, propAnchorZ)
  local function worldLocUsingAnchorPoint(object, anchorX, anchorY, anchorZ)
    local actualAnchorX, actualAnchorY, actualAnchorZ = object:getAnchorPoint()
    anchorX = anchorX or actualAnchorX
    anchorY = anchorY or actualAnchorY
    anchorZ = anchorZ or actualAnchorZ or 0

    local loc0X, loc0Y, loc0Z = object:getLocationWithAnchorPoint(anchorX, anchorY, anchorZ)
    local locX, locY, locZ = object.parent:modelToWorld(loc0X, loc0Y, loc0Z)
    locZ = locZ or 0
    return locX, locY, locZ
  end

  local selfX, selfY, selfZ = worldLocUsingAnchorPoint(self, selfAnchorX, selfAnchorY, selfAnchorZ)
  local propX, propY, propZ = worldLocUsingAnchorPoint(prop, propAnchorX, propAnchorY, propAnchorZ)

  return propX - selfX, propY - selfY, propZ - selfZ
end

--- Transforms coordinates in sceneObject to coordinates in self.
-- @param self - (Group) the group itself
-- @param sceneObject the object to transform coordinates from.
-- @param x - (number) X Coordinate to transform.
-- @param y - (number) Y Coordinate to transform.
-- @param z - (number) Z Coordinate to transform.
function Group:transformLoc(sceneObject, x, y, z)
  return self:worldToModel(sceneObject:modelToWorld(x, y, z))
end

-- Transforms sceneObject's location in sceneObject.parent's coordinates to coordinates in self.
-- @param anchorX, anchorY, anchorZ (numbers, optional) The hypothetical anchor point to use. Defaults to sceneObject's current anchor point.
function Group:transformObjectLoc(sceneObject, anchorX, anchorY, anchorZ)
  local locX, locY, locZ
  if anchorX and anchorY then
    locX, locY, locZ = sceneObject:getLocationWithAnchorPoint(anchorX, anchorY, anchorZ)
  else
    locX, locY, locZ = sceneObject:getLoc()
  end
  return self:transformLoc(sceneObject.parent, locX, locY, locZ)
end

--- Move self such that it aligns in real world coordinates with the specified prop along the specified anchor points.
-- If the anchor points are omitted, it uses each respective object's current anchor point.
-- @param prop The prop to get the distance to.
-- @param selfAnchorX, selfAnchorY, selfAnchorZ Anchor points for self.
-- @param propAnchorX, propAnchorY, propAnchorZ Anchor points for the prop.
function Group:alignToProp(prop, selfAnchorX, selfAnchorY, selfAnchorZ, propAnchorX, propAnchorY, propAnchorZ)
  local diffX, diffY, diffZ = self:getWorldDistance(prop, selfAnchorX, selfAnchorY, selfAnchorZ, propAnchorX, propAnchorY, propAnchorZ)
  local selfX, selfY, selfZ = self:getLoc()
  self:setLoc(selfX + diffX, selfY + diffY, (selfZ or 0) + (diffZ or 0))
end

--- Get coordinates of group center in world coordinates. Needed because getWorldLoc apparently just gives you the
--  lower-left corner, not its actual location.
function Group:getCenterWorldLoc()
  local selfX, selfY = self:getWorldLoc()
  selfX = selfX + self:getWidthPart(0.5)
  selfY = selfY + self:getHeightPart(0.5)

  return selfX, selfY
end


-- Anchor Point -----------------------

--- Sets the pivot of the group at the given proportion of its width and height.  @see getAnchorPoint
-- @param anchorPointX - (number) The proportion of the group's width at which to set the pivot.
-- @param anchorPointY - (number) The proportion of the group's height at which to set the pivot.
function Group:setAnchorPoint(anchorPointX, anchorPointY)
  local width, height = self:getDims()

  width, height = width or 0, height or 0
  local pivX = width * anchorPointX
  local pivY = height * anchorPointY

  self:setPiv(pivX, pivY)
end

--- Convenience helpers for setting anchor point to common places
function Group:setAnchorPointCenter()
  self:setAnchorPoint(0.5, 0.5)
end

function Group:setAnchorPointTop()
  self:setAnchorPoint(0.5, 1.0)
end

function Group:setAnchorPointBottom()
  self:setAnchorPoint(0.5, 0.0)
end

function Group:setAnchorPointLeft()
  self:setAnchorPoint(0.0, 0.5)
end

function Group:setAnchorPointRight()
  self:setAnchorPoint(1.0, 0.5)
end

function Group:setAnchorPointTopLeft()
  self:setAnchorPoint(0.0, 1.0)
end

function Group:setAnchorPointTopRight()
  self:setAnchorPoint(1.0, 1.0)
end

function Group:setAnchorPointBottomLeft()
  self:setAnchorPoint(0.0, 0.0)
end

function Group:setAnchorPointBottomRight()
  self:setAnchorPoint(1.0, 0.0)
end

--- Retrieves the location of the group's pivot as a proportion of its width and height
-- @return The x-coordinate of the group's anchor point (pivotX / width).
-- @return The y-coordinate of the group's anchor point (pivotY / height).
function Group:getAnchorPoint()
  local pivX, pivY = self:getPiv()
  local width, height = self:getDims()

  width, height = width or math.huge, height or math.huge
  local anchorPointX = pivX / width
  local anchorPointY = pivY / height

  return anchorPointX, anchorPointY
end

--- Change the anchor point without changing the group's relative position on screen.
-- For example, given a group of height 100 with Y anchor point 0, transforming the anchor
-- point to 1 would increase the group's Y position by 100 to compensate.
function Group:transformAnchorPoint(anchorX, anchorY, anchorZ)
  self:setLoc(self:getLocationWithAnchorPoint(anchorX, anchorY, anchorZ))
  self:setAnchorPoint(anchorX, anchorY, anchorZ)
end


-- Dimensions -------------------------

--- Get this group's effective dimensions, as they would appear after scaling has been applied.
function Group:getEffectiveDims()
  local selfW, selfH = self:getDims()
  local scaleX, scaleY = self:getScl()
  return selfW * scaleX, selfH * scaleY
end

function Group:setDims(width, height, depth)
  depth = depth or 0
  local anchorX, anchorY, anchorZ = self:getAnchorPoint()
  self:setBounds(0.0, 0.0, 0, width, height, depth)
  self:setAnchorPoint(anchorX, anchorY, anchorZ)
end

--- Sets the "effective" dimensions of the group by adjusting the scale.
-- For example, if your group has size 100, 200, and you passed in 800,800
-- it would set X scale to 8.0 and Y scale to 4.0.
-- If either width or height is omitted, it will scale based on whichever dimension is not nil.
function Group:setEffectiveDims(width, height)
  self:setScl(self:scaleForEffectiveDims(width, height))
end

function Group:getWidth()
  local width, _ = self:getDims()
  return width
end

function Group:getEffectiveWidth()
  local scaleX, scaleY = self:getScl()
  return self:getWidth() * scaleX
end

function Group:getScaleX()
  local scaleX, _ = self:getScl()
  return scaleX
end

function Group:setWidth(width)
  local _, height = self:getDims()
  self:setDims(width, height)
end

function Group:setEffectiveWidth(width)
  local scaleX, scaleY = self:getScl()
  scaleX = width / self:getWidth()
  self:setScl(scaleX, scaleY)
end

function Group:setScaleX(scaleX)
  local _, scaleY = self:getScl()
  self:setScl(scaleX, scaleY)
end

function Group:getHeight()
  local _, height = self:getDims()
  return height
end

function Group:getEffectiveHeight()
  local scaleX, scaleY = self:getScl()
  return self:getHeight() * scaleY
end

function Group:getScaleY()
  local _, scaleY = self:getScl()
  return scaleY
end

function Group:setHeight(height)
  local width = self:getDims()
  self:setDims(width, height)
end

function Group:setEffectiveHeight(height)
  local scaleX, scaleY = self:getScl()
  scaleY = height / self:getHeight()
  self:setScl(scaleX, scaleY)
end

function Group:setScaleY(scaleY)
  local scaleX, _ = self:getScl()
  self:setScl(scaleX, scaleY)
end

function Group:getWidthPart(proportion)
  return self:getWidth() * proportion
end

function Group:getEffectiveWidthPart(proportion)
  return self:getEffectiveWidth() * proportion
end

function Group:getHeightPart(proportion)
  return self:getHeight() * proportion
end

function Group:getEffectiveHeightPart(proportion)
  return self:getEffectiveHeight() * proportion
end

function Group:getCenter()
  local selfW, selfH, selfD = self:getDims()
  local halfD
  if selfD then
    halfD = selfD / 2
  end
  return selfW/2, selfH/2, halfD
end

--- Gets the scale that would be required on this object to achieve the specified dimensions.
-- If either width or height is omitted, it will return the scale required to achieve
-- whichever dimension is not nil.
function Group:scaleForEffectiveDims(width, height)
  local selfW, selfH = self:getDims()
  if not width then
    return height/selfH, height/selfH
  elseif not height then
    return width/selfW, width/selfW
  else
    return width/selfW, height/selfH
  end
end

--- Resize this group to fit its children.
-- Does not handle the case where a child overlaps into negative space.
function Group:resizeToFitChildren()
  local maxX, maxY, maxZ = 0, 0, 0
  for _, child in ipairs(self.children) do
    local x, y, z = child:getLocationWithAnchorPoint(1, 1, 1)
    maxX = math.max(x, maxX)
    maxY = math.max(y, maxY)
    maxZ = math.max(z, maxZ)
  end

  self:setDims(maxX, maxY, maxZ)
end


-- Attributes -------------------------

local function inheritAttr(self, parent, attribute, sourceAttribute)
  if parent then
    self:setAttrLink(attribute, parent, sourceAttribute)
  else
    self:clearAttrLink(attribute)
  end
end

--- Inherit color from parent.
function Group:inheritColor(parent)
  inheritAttr(self, parent, MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)
end

--- Allow parent to override this group's color.
function Group:overrideColor(parent)
  inheritAttr(self, parent, MOAIColor.INHERIT_COLOR_RAW, MOAIColor.COLOR_TRAIT)
end

--- Inherit opacity from parent.
function Group:inheritOpacity(parent)
  inheritAttr(self, parent, MOAIColor.ATTR_OPACITY)
end

--- Link an attribute from this group to its children.
-- @param attrID (number) - The attribute ID to link to in the children.
-- @param sourceAttrID (number) - The attribute in this group to link from.
-- @param isRecursive (boolean) - If true, call this method recursively on all children.
function Group:linkAttrToChildren(attrID, sourceAttrID, isRecursive)
  for _, child in ipairs(self.children) do
    child:setAttrLink(attrID, self, sourceAttrID)
    if isRecursive and child.linkAttrToChildren then
      child:linkAttrToChildren(attrID, sourceAttrID, isRecursive)
    end
  end
end

-- Clear an existing attribute link to children.
-- @param attrID (number) - The attribute ID to clear in the children.
-- @param isRecursive (boolean) - If true, call this method recursively on all children.
function Group:clearAttrLinkToChildren(attrID, isRecursive)
  for _, child in ipairs(self.children) do
    child:clearAttrLink(attrID, self)
    if isRecursive and child.clearAttrLinkToChildren then
      child:clearAttrLinkToChildren(attrID, isRecursive)
    end
  end
end

--- Link the color attribute to all children.
function Group:linkColorAttrToChildren(isRecursive)
  self:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT, isRecursive)
end

--- Link the raw color attribute to all children.
function Group:linkRawColorAttrToChildren(isRecursive)
  self:linkAttrToChildren(MOAIColor.INHERIT_COLOR_RAW, MOAIColor.COLOR_TRAIT, isRecursive)
end

--- Link the opacity attribute to all children.
function Group:linkOpacityAttrToChildren(isRecursive)
  self:linkAttrToChildren(MOAIColor.ATTR_OPACITY, nil, isRecursive)
end

--- Clear the color attribute link to all children.
function Group:clearColorAttrLinkToChildren(isRecursive)
  self:clearAttrLinkToChildren(MOAIColor.INHERIT_COLOR, isRecursive)
end

--- Clear the raw color attribute link to all children.
function Group:clearRawColorAttrLinkToChildren(isRecursive)
  self:clearAttrLinkToChildren(MOAIColor.INHERIT_COLOR_RAW, isRecursive)
end

--- Clear the opacity attribute link to all children.
function Group:clearOpacityAttrLinkToChildren(isRecursive)
  self:clearAttrLinkToChildren(MOAIColor.ATTR_OPACITY, isRecursive)
end


-- Tags -------------------------------

function Group:getTag()
  return self._Group_tag
end

function Group:setTag(tag)
  self._Group_tag = tag
end

function Group:findGroupWithTag(tag)
  for _, group in ipairs(self.children) do
    if group._Group_tag == tag then
      return group
    end
    if group.findGroupWithTag then
      local found = group:findGroupWithTag(tag)
      if found then
        return found
      end
    end
  end
  return nil
end


-- Utility ----------------------------

--- Adds group behavior to the specified prop
-- Group behaviour means managing a list of children, scene graph sorting between elements in the hierarchy,
-- automated cleanup for removed children, on demand bounds calculation.
-- @param prop - (MOAIProp) prop that will have group behaviour.
function Group.makeGroup(prop)
  if not prop.children then
    Mixin.mix(prop, Group)
    prop.children = {}
    prop.zIndex = 0
  end
end

--- Mark an object that may not be a group as being 3D
function Group.markAs3DObject(object)
  object._is3DObject = true
end

--- Sets visibility on this prop and any child props, recursively
-- @param self - (Group) the group itself
-- @param isVisible value to set for visibility of props
function Group:setVisibleRecursive(isVisible)
  self:setVisible(isVisible)
  for _, child in ipairs(self.children) do
    if (child.setVisibleRecursive) then
      child:setVisibleRecursive(isVisible)
    else
      child:setVisible(isVisible)
    end
  end
end


-- ---------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create an unsized group. This means that width and height are nil!
-- This constructor should only be called if you will be using the underlying deck to control size.
-- DO NOT CALL THIS CONSTRUCTOR UNLESS YOU HAVE A DAMN GOOD REASON TO DO SO!
function Group.newUnsized()
  local self = MOAIProp2D.new()
  Group.makeGroup(self)
  return self
end

--- DEPRECATED CONSTRUCTOR - Creates an empty group with no size.
function Group.newDeprecated()
  return Group.newUnsized()
end

-- Create a sized group with an anchor point in the center.
function Group.new(width, height)
  assert(type(width) == "number", "width must be a number (got " .. type(width) .. ")")
  assert(type(height) == "number", "height must be a number (got " .. type(height) .. ")")
  local self = Group.newUnsized()
  self:setBounds(0.0, 0.0, 0, width, height, 0)
  self:setAnchorPointCenter()
  return self
end

-- Create group the size of the screen with an anchor point in the center.
function Group.newFullScreen()
  return Group.new(Screen.width, Screen.height)
end

--- Create a group with the minimum allowed size before MOAI starts doing stupid things.
function Group.newMinimumSized()
  return Group.new(MINIMUM_WIDTH, MINIMUM_HEIGHT)
end

--- Create a group supporting custom draw routines.
-- The draw routine can set later by calling:
--     self.deck:setDrawCallback(aFunction)
-- @param width (number) - The group's width
-- @param height (number) - The group's height
-- @param color (table) - The color to apply to the group (optional)
-- @param drawFunction (function) - A function containing draw commands (optional if setting later)
function Group.newCustomDraw(width, height, color, drawFunction)
  local self = Group.new(width, height)

  self.deck = MOAIScriptDeck.new()
  self.deck:setRect(0.0, 0.0, width, height)
  self:setDeck(self.deck)

  if color then
    self:setColor(unpack(color))
  end

  if drawFunction then
    self.deck:setDrawCallback(drawFunction)
  end

  return self
end
