-- Class used to manage localization of strings.
-- The "currentLanguage" is the base localization file used.
-- The "currentLocale" is used for any overrides to the base localization file.

-- Example:
-- "en.json" contains english copy for every string.
-- "en_GB.json" contains a couple strings with British specific copy, that will be
-- used if the user's locale is set to "en_GB". If a string does not have
-- British specific copy, the base version in "en.json" is used.

-- ===========================================================================
class("Localizer")
-- ===========================================================================

require("Log")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local pathToGame = "localization/"
local pathToShared = "../../shared_source/localization/"
local fileExtension = ".json"

local currentLanguage = nil
local currentLocale = nil
local dictionary = {}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function fileExists(filePath)
  local file = io.open(filePath, "rb")
  if (file) then
    file:close()
    return true
  end

  return false
end

local function loadKeyPairs(filePath)
  if fileExists(filePath) then
    local lines = ""
    for line in io.lines(filePath) do
      lines = lines .. line
    end
    return MOAIJsonParser.decode(lines)
  else
    Log.verbose("Localization file path does not exist: " .. tostring(filePath))
    return {}
  end
end

local function createLocalizedDictionaryAtPath(path)
  local languageFilePath = path .. currentLanguage  .. fileExtension
  local localeFilePath = path .. currentLocale  .. fileExtension

  local languageTable = loadKeyPairs(languageFilePath)
  local localeOverrides = loadKeyPairs(localeFilePath)

  for key, override in pairs(localeOverrides) do
    if (languageTable[key] == nil) then
      Log.warn(string.format("Locale file '%s' contains a key '%s' that does not exist in the base '%s' file.", tostring(localeFilePath), tostring(key), tostring(languageFilePath)))
    end
    languageTable[key] = override
  end

  return languageTable
end

local function loadLanguageFiles()
  local sharedDictionary = createLocalizedDictionaryAtPath(pathToShared)
  local gameDictionary = createLocalizedDictionaryAtPath(pathToGame)

  for key, value in pairs(gameDictionary) do
    if (sharedDictionary[key] ~= nil) then
      Log.error(string.format("Conflicting key '%s' exists in both game and shared localization files.", tostring(key)))
    end
    sharedDictionary[key] = value
  end

  dictionary = sharedDictionary
end

local function setBaseLanguage(locale)
  local separatedStrings = StringUtils.split(locale, "_")
  currentLanguage = separatedStrings[1]
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Localizer.setLocale(locale)
  currentLocale = locale
  setBaseLanguage(locale)
  loadLanguageFiles()
end

function Localizer.getCurrentLanguage()
  return currentLanguage
end

function Localizer.getCurrentLocale()
  return currentLocale
end

function Localizer.get(field)
  if (dictionary[field] == nil) then
    Log.warn(string.format("Key '%s' does not exist in the Localizer's dictionary.", tostring(field)))
    return ""
  end
  return dictionary[field]
end

-- ----------------------------------------------------------------------------
-- Exposed for Testing
-- ----------------------------------------------------------------------------

function Localizer._setPathToGame(newPathToGame)
  pathToGame = newPathToGame
end

function Localizer._setPathToShared(newPathToShared)
  pathToShared = newPathToShared
end

function Localizer._setDictionary(newDictionary)
  dictionary = newDictionary
end

function Localizer._getDictionary()
  return dictionary
end

function Localizer._clean()
  currentLanguage = nil
  currente = nil
  dictionary = {}
end
