--- MetalPrebinder
-- @author Created by Aaron Barrett on June 08, 2021.
--
-- The purpose of MetalPrebinder is to eliminate performance issues in MOAI
-- caused by the introduction of MetalANGLE.
--
-- Although MOAI's overall performance with MetalANGLE in place is comparable to
-- its performance under OpenGL, there are some some moments where animations
-- stutter significantly in iOS, particularly in games which display large
-- amounts of text.  It appears that this stuttering occurs when new textures
-- are first shown in a game.  Note that the stuttering does *not* occur on
-- texture creation, but rather at the moment that a texture is rendered for the
-- first time.  It seems that MetalANGLE must perform an expensive operation
-- involving shader compilation whenever a texture is bound for the first time,
-- and this operation was either trivial or unneccessary under OpenGL.
--
-- MetalPrebinder attempts to alleviate the stuttering by binding new textures
-- immediately after they are created.  This serves to get the performance hit
-- out of the way.  Since texture creation tends to occur at times that do not
-- impact running animations, it is generally safe to group the initial binding
-- hit alongside it.  In practice, this technique seems to work well, and the
-- initial binding hits are mostly imperceptible.
--
-- MetalPrebinder works by adding hidden sprites to the scene.  Each time a new
-- texture is created (either loaded from disk or generated by the text system),
-- a new minimum-sized sprite that uses the texture is added to the lower-left
-- corner of the screen at a Z-index behind the game itself.  These hidden
-- sprites trigger the initial binding operation for each texture.
-- MetalPrebinder runs a method each frame that removes hidden sprites as soon
-- as they have been bound and rendered.

class("MetalPrebinder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MINIMUM_DISPLAYABLE_SIZE = 2.0
local SPRITE_AGE_NEEDED_TO_PREBIND = 2

local Z_BEHIND_GAME = -1000000

local spritesToAdd = {}
local addedSpritesSet = {}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createMinimumSizedDeck()
  local deck = MOAIGfxQuad2D.new()
  deck:setRect(0.0, 0.0, MINIMUM_DISPLAYABLE_SIZE, MINIMUM_DISPLAYABLE_SIZE)
  return deck
end

local function createMinimumSizedDeckForShader(shader)
  local deck = createMinimumSizedDeck()
  deck:setShader(shader)
  return deck
end

local function createMinimumSizedDeckForTexture(texture)
  local deck = createMinimumSizedDeck()
  deck:setTexture(texture)
  return deck
end

local function createAndInitializeSpriteForDeck(deck)
  local sprite = Sprite.newWithDeck(deck)
  sprite.age = 0
  table.insert(spritesToAdd, sprite)
end

local function incrementAddedSpriteAges()
  for sprite in pairs(addedSpritesSet) do
    sprite.age = sprite.age + 1
  end
end

local function getSpritesThatHaveBeenRendered()
  local renderedSprites = {}
  for sprite in pairs(addedSpritesSet) do
    if (sprite.age > SPRITE_AGE_NEEDED_TO_PREBIND) then
      table.insert(renderedSprites, sprite)
    end
  end
  return renderedSprites
end

local function clearSprites(spritesToClear)
  for _, sprite in ipairs(spritesToClear) do
    sprite:removeSelf()
    addedSpritesSet[sprite] = nil
  end
end

local function addNewSprites()
  for _, sprite in ipairs(spritesToAdd) do
    sprite:setLoc(0.0, 0.0)
    Screen.rootGroup:insert(sprite, Z_BEHIND_GAME)
    addedSpritesSet[sprite] = true
  end
  spritesToAdd = {}
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MetalPrebinder.prebindShader(shader)
  local deck = createMinimumSizedDeckForShader(shader)
  createAndInitializeSpriteForDeck(deck)
end

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function MetalPrebinder.prebindTexture(texture)
  local deck = createMinimumSizedDeckForTexture(texture)
  createAndInitializeSpriteForDeck(deck)
end

function MetalPrebinder.onRender()
  incrementAddedSpriteAges()

  local spritesToClear = getSpritesThatHaveBeenRendered()
  clearSprites(spritesToClear)

  addNewSprites()
end
