--- Property loader reads in the properties.lua and acts on each contents appropriately.

require("Bootstrap")

-- ===========================================================================
class("PropertyLoader")
-- ===========================================================================

require("Lib")

local isLandscape = false

--- load properties.lua and handle the results
-- properties.lua should live in the root directory of your game and return a table
-- that looks something like
-- return {
--   resources = {
--     sounds = {},
--     animations = {},
--     spritesheets = {"balloons", "clouds"},
--     fonts = {
--       BALLOON_FONT = "din_ot_medium.otf"
--     }
--     scripts = {"PrerenderBalloonText.lua"}
--   }
-- }

------------------------------Override properties file------------------------------------
-- Some games (like Zoomer) require special properties files to be loaded for specific
-- game modes.  To specify a properties override for a game mode, include the following
-- key in the "parameters" section for the game mode in games.json:
-- overridePropertiesName = "your_properties_name_here"
-- ----------------------------------------------------------------------------------------
local gamePropertiesName = GameSession.config.overridePropertiesName or "properties"
print("Loading properties file with name: '" .. tostring(gamePropertiesName) .. "'")
local gameProperties = require(gamePropertiesName)
-- ----------------------------------------------------------------------------------------


-- ---------------------------------glClear Color------------------------------------------
-- Set the glClear color if it is included in the game's properties.lua
-- This is done to prevent black frames from appearing while games are loading.
-- ----------------------------------------------------------------------------------------
if (gameProperties.glClearColor ~= nil) then
  local frameBuffer = MOAIGfxDevice.getFrameBuffer()
  local frameBufferColor = gameProperties.glClearColor
  local isDarkMode = GameSession.getIsDarkMode()
  if (isDarkMode and (gameProperties.glClearColorDark ~= nil)) then
    frameBufferColor = gameProperties.glClearColorDark
  end
  frameBuffer:setClearColor(unpack(frameBufferColor))
end
-- ----------------------------------------------------------------------------------------


function PropertyLoader.numberOfItemsToLoadForProperties(properties)
  local totalItems = 1

  if (properties.resources) then
    if (properties.resources.spritesheets) then
      totalItems = totalItems + #properties.resources.spritesheets
    end
    if (properties.resources.images) then
      totalItems = totalItems + #properties.resources.images
    end
    if (properties.resources.animations) then
      totalItems = totalItems + #properties.resources.animations
    end
    if (properties.resources.sounds) then
      totalItems = totalItems + #properties.resources.sounds
    end
    if (properties.resources.fonts) then
      totalItems = totalItems + #properties.resources.fonts
    end
    if (properties.resources.scripts) then
      totalItems = totalItems + #properties.resources.scripts
    end
  end

  return totalItems
end

function PropertyLoader.loadPropertyTable(properties, progressCallback)
  if (properties) then
    -- Load the reference device
    if (properties.referenceDevice) then
      ReferenceDevice.setDevice(properties.referenceDevice)
    end

    -- Load the spritesheets
    if (properties.resources.spritesheets) then
      for i, sheetName in ipairs(properties.resources.spritesheets) do
        Sprite.loadSpritesheetWithName(sheetName)
        progressCallback()
      end
    end

    -- Load standalone images
    if (properties.resources.images) then
      for i, imageName in ipairs(properties.resources.images) do
        TextureLoader.loadTexture(imageName)
        progressCallback()
      end
    end

    -- Load the animations
    if (properties.resources.animations) then
      for i, animation in ipairs(properties.resources.animations) do
        Animation.loadAnimationWithPrefix(animation)
        progressCallback()
      end
    end

    -- Load sounds
    if (properties.resources.sounds) then
      for i, sound in ipairs(properties.resources.sounds) do
        Sound.loadSoundEffect(sound)
        progressCallback()
      end
    end

    -- Load fonts
    if (properties.resources.fonts) then
      for name, file in pairs(properties.resources.fonts) do
        FontFaces.loadFont(name, file)
        progressCallback()
      end
    end

    -- Run any Lua files to preload anything while still on the loading screen
    if (properties.resources.scripts) then
      for i, script in ipairs(properties.resources.scripts) do
        dofile(script)
        progressCallback()
      end
    end

  else
    print("No properties to load.")
  end

end

function PropertyLoader.loadProperties()
  local sharedProperties = require("shared_properties")
  local progress = 0.0

  MSAssetProgressCallback(progress)

  local totalProperties = PropertyLoader.numberOfItemsToLoadForProperties(gameProperties) + PropertyLoader.numberOfItemsToLoadForProperties(sharedProperties)

  local progressIncrement = 100 / totalProperties
  local progress = 0

  local function bumpProgress()
    progress = progress + progressIncrement
    MSAssetProgressCallback(progress / 100)
  end

  PropertyLoader.loadPropertyTable(gameProperties, bumpProgress)
  PropertyLoader.loadPropertyTable(sharedProperties, bumpProgress)

  bumpProgress() -- this accounts for the initial 1
end

--- Determine if the current game is landscape
-- @return bool - is the current game meant to be landscape
function PropertyLoader.isLandscape()
  local properties = gameProperties
  return properties.isLandscape
end

function PropertyLoader.requiresMultitouch()
  local properties = gameProperties
  return properties.requiresMultitouch
end
