--- A global screen object for display type functionality
-- @module Screen
-- @field width The width of the screen in points
-- @field height The height of the screen in points
-- @author Created by Alexey Chernikov on Feb 19, 2013

-- Builds a series of 2d and 3d layers to adapt to any combination of
-- 2d/3d object trees. 2d layers sort in priority order, whereas 3d layers
-- sort by Z position.

-- ===========================================================================
class("Screen")
-- ===========================================================================

require("MetalPrebinder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

Screen.width = MOAIEnvironment.horizontalResolution / Host.contentScale
Screen.height = MOAIEnvironment.verticalResolution / Host.contentScale
Screen.stepDuration = MOAISim.getStep()

Screen.fullWidth = Screen.width;
Screen.fullHeight = Screen.height;

Screen.safeWidth = Screen.fullWidth - (Host.safeAreaLeftInset + Host.safeAreaRightInset)
Screen.safeHeight = Screen.fullHeight - (Host.safeAreaTopInset + Host.safeAreaBottomInset)
Screen.safeLeft = Host.safeAreaLeftInset
Screen.safeRight = Screen.fullWidth - Host.safeAreaRightInset
Screen.safeRightInset = Host.safeAreaRightInset
Screen.safeBottom = Host.safeAreaBottomInset
Screen.safeTop = Screen.fullHeight - Host.safeAreaTopInset
Screen.safeTopInset = Host.safeAreaTopInset
Screen.safeCenterX = (Screen.safeRight + Screen.safeLeft) / 2
Screen.safeCenterY = (Screen.safeTop + Screen.safeBottom) / 2

Screen.elapsedFrameCount = 0

-- DEPRECATED
Screen.IPHONE_X_BOTTOM_SAFE_ZONE_HEIGHT = 42.0

local TABLET_ASPECT_RATIO_THRESHOLD = 0.6

local viewport = nil
local managedLayers = {}
local initialized = false
local nodeTreeIsDirty = false

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function assignLayer(object, layer)
  local parentLayer = object.parentLayer
  if (parentLayer ~= layer) then
    if (parentLayer ~= nil) then
      parentLayer:removeProp(object)
    end
    object.parentLayer = layer
    layer:insertProp(object)
  end
end

--- Collapse an object's parent-child tree structure into a linear priority set
-- using depth-first, left sided traversal, updating using setPriority().
local function applyTreePriority(object, layer, rootPriority)
  assignLayer(object, layer)

  -- Update priority and recurse into children.
  object:setPriority(rootPriority)
  local priority = rootPriority + 1
  local children = object.children
  if (children ~= nil) then
    for _, child in ipairs(children) do
      priority = applyTreePriority(child, layer, priority)
    end
  end

  return priority
end

local function onRender()
  if (nodeTreeIsDirty) then
    for _, layer in ipairs(managedLayers) do
      local rootGroup = layer.rootGroup
      local rootPriority = 1
      applyTreePriority(rootGroup, layer, rootPriority)
    end
    if (not Host.environmentIsAppleTv) then
      TouchManager.notifyPriorityChange()
    end
    nodeTreeIsDirty = false
  end

  MetalPrebinder.onRender()
  Screen.elapsedFrameCount = Screen.elapsedFrameCount + 1
end

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

-- --- Tell the screen that there has been an internal structure change.
-- -- INTERNAL METHOD. DO NOT CALL FROM CLIENT CODE.
function Screen.notifyStructureChange()
  nodeTreeIsDirty = true
end

--- Remove an object from the screen hierarchy.
-- INTERNAL METHOD. DO NOT CALL FROM CLIENT CODE.
function Screen.remove(object)
  local parentLayer = object.parentLayer
  if (parentLayer ~= nil) then
    parentLayer:removeProp(object)
    object.parentLayer = nil
  end
end

--- Initialize the screen.
function Screen.init()
  if (initialized) then
    return
  end

  viewport = MOAIViewport.new()
  viewport:setSize(MOAIEnvironment.horizontalResolution, MOAIEnvironment.verticalResolution)
  viewport:setScale(Screen.fullWidth, Screen.fullHeight)
  viewport:setOffset(-1.0, -1.0)

  local layer = MOAILayer2D.new()
  layer:setViewport(viewport)
  layer:setSortMode(MOAILayer2D.SORT_PRIORITY_ASCENDING)
  Screen.addManagedLayer(layer)

  local rootGroup = Group.newFullScreen()
  rootGroup:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))
  layer:insertProp(rootGroup)
  rootGroup.parentLayer = layer
  layer.rootGroup = rootGroup
  Screen.rootGroup = rootGroup

  MOAIRenderMgr.addToRenderTable(layer)
  MOAIRenderMgr.setPreRenderCallback(onRender)

  initialized = true
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Screen.getWidthPart(proportion)
  return Screen.width * proportion
end

function Screen.getHeightPart(proportion)
  return Screen.height * proportion
end

function Screen.getCenter()
  return Screen.getWidthPart(0.5), Screen.getHeightPart(0.5)
end

-- getSafeWidthPart() and getSafeHeightPart() return fractions of the safe area
function Screen.getSafeWidthPart(proportion)
  return Screen.safeWidth * proportion
end

function Screen.getSafeHeightPart(proportion)
  return Screen.safeHeight * proportion
end

function Screen.isTabletAspectRatio()
  local aspectRatioFractional = Screen.width / Screen.height
  return (aspectRatioFractional > TABLET_ASPECT_RATIO_THRESHOLD)
end

function Screen.getSafeX(proportion)
  return Screen.safeLeft + Screen.getSafeWidthPart(proportion)
end

function Screen.getSafeXFromRight(proportion)
  return Screen.safeRight - Screen.getSafeWidthPart(proportion)
end

function Screen.getSafeY(proportion)
  return Screen.safeBottom + Screen.getSafeHeightPart(proportion)
end

function Screen.getSafeYFromTop(proportion)
  return Screen.safeTop - Screen.getSafeHeightPart(proportion)
end

function Screen.getViewport()
  return viewport
end

function Screen.addManagedLayer(layer)
  table.insert(managedLayers, layer)
end

function Screen.sort(root, rootPriority, layer)
  assignLayer(root, layer)
  root:setPriority(rootPriority)
  local priority = rootPriority + 1
  if root.children then
    for _, child in ipairs(root.children) do
      priority = Screen.sort(child, priority, layer)
    end
  end
  return priority
end

-- DEPRECATED
function Screen.hasNotch()
  return Host.hasNotch
end
