--- A module for loading and playing sounds

-- ===========================================================================
class("Sound")
-- ===========================================================================

require("SoundSource")


local soundFinishedHandlers = {}
local activeSources = {}

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function getSoundEffectPath(name)
  local extension = Host.soundEffectFileExtension
  local nameInFolder = "sounds/" .. name .. extension
  local path = FileUtils.findFile(nameInFolder)
  assert(path ~= nil, "No sound effect found at '" .. tostring(nameInFolder) .. "'")
  return path
end

local function getLongformPath(name)
  local extension = Host.musicFileExtension
  local nameInFolder = "sounds/" .. name .. extension
  local path = FileUtils.findFile(nameInFolder)
  assert(path ~= nil, "No longform sound found at '" .. tostring(nameInFolder) .. "'")
  return path
end

local function playSoundOnHost(path, volume, looping)
  volume = volume or 1.0
  looping = looping or false
  local soundID = Host.playSound(path, volume, looping)

  local soundSource = SoundSource.new(soundID, path)
  activeSources[soundID] = soundSource
  return soundSource
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

--- Load a sound effect by name
-- Search the asset path for an effect sound with the name.(ext) where
-- ext is the effect extension for that platform
-- @param name (string) - A filename, i.e. bloon_pop
function Sound.loadSoundEffect(name)
  local path = getSoundEffectPath(name)
  Host.loadSound(path)
end

--- Tell the host to play a sound effect
-- Play a sound assumed to be a sound effect and that is in the assets path
-- @param name (string) just the file name of a sound
-- @param volume (optional) (number) The volume to play at (0.0 - 1.0). Default 1.0
function Sound.playSoundEffect(name, volume)
  local path = getSoundEffectPath(name)
  local soundSource = playSoundOnHost(path, volume)
  return soundSource
end

--- Used to play sounds that require a special format (such as the conversations in listening games).
-- Note that you must provide the entire path of the sound, not just the name.
-- @param path (string) absolute path of long form sound
-- @param looping (optional) (boolean) if true, loops the sound forever
function Sound.playLongformSoundWithPath(path, looping)
  assert(MOAIFileSystem.checkFileExists(path), "No long form sound file found at path: '" .. tostring(path) .. "'")
  Host.loadSound(path)
  local volume = 1.0
  local soundSource = playSoundOnHost(path, volume, looping)
  return soundSource
end

--- Used to play sounds that require a special format (such as the conversations in listening games).
-- @param name (string) name of long form sound (for example, "music")
-- @param looping (optional) (boolean) if true, loops the sound forever
function Sound.playLongformSound(name, looping)
  local path = getLongformPath(name)
  return Sound.playLongformSoundWithPath(path, looping)
end

--- Stop a sound, given the sounds's ID.  @see SoundSource
-- @param soundID - the sound ID returned by the host when the sound was started
function Sound.stopSound(soundID)
  Host.stopSound(soundID)
end

--- Get the duration that a sound has been playing on the host, given the sounds's ID.  @see SoundSource
-- @param soundID - the sound ID returned by the host when the sound was started
-- @return the duration that the sound has been playing
function Sound.getTime(soundID)
  return Host.getSoundTime(soundID)
end

function Sound.pauseSound(soundID)
  Host.pauseSound(soundID)
end

function Sound.resumeSound(soundID)
  Host.resumeSound(soundID)
end

function Sound.seek(soundID, time)
  Host.seekSoundAtTime(soundID, time)
end

function Sound.getAveragePowerForAllChannels(soundID)
  return Host.averagePowerForAllChannels(soundID)
end

--- WARNING: Called by Host, don't call from Lua!
function Sound.soundFinished(soundID)
  local source = activeSources[soundID]
  if source then
    activeSources[soundID] = nil
    source:stop() -- MUST be called to avoid multiple-play bug when unpausing.
  end
  local handlers = soundFinishedHandlers[soundID]
  if (handlers == nil) then return end
  for _, handler in pairs(handlers) do
    handler(soundID)
  end
end

function Sound.addSoundFinishedHandler(soundID, handler)
  handlers = soundFinishedHandlers[soundID] or {}
  handlers[#handlers + 1] = handler
  soundFinishedHandlers[soundID] = handlers
end
