--- Sprites and texture sheet loading
-- @author Jay Austria

-- ===========================================================================
class("Sprite")
-- ===========================================================================

require("TextureLoader")

-- ----------------------------------------------------------------------------
-- Configuration --
-- ----------------------------------------------------------------------------

local DEFAULT_SPRITESHEET_EXTENSION = ".pvr.gz"

-- ----------------------------------------------------------------------------
-- Data --
-- ----------------------------------------------------------------------------

-- There are two levels of cache: Sprite sheet and sprite level.
-- The sprite sheet level allows access to a sprite sheet deck based on the sprite sheet's name.
-- The sprite level allows access to a sprite sheet deck based on the name of a sprite contained within the sheet.
local Caches = {
  spritesheetDeckBySpritesheetName = {},
  spritesheetDeckBySpriteName = {}
}


-- ----------------------------------------------------------------------------
-- Caching --
-- ----------------------------------------------------------------------------

local function getCachedSpritesheetSpriteDeck(name)
  return Caches.spritesheetDeckBySpriteName[name]
end

local function cacheSpritesheetSprite(name, spritesheetDeck)
  Caches.spritesheetDeckBySpriteName[name] = spritesheetDeck
end

local function uncacheSpritesheetSprite(name)
  Caches.spritesheetDeckBySpriteName[name] = nil
end

local function getCachedSpritesheetDeck(name)
  return Caches.spritesheetDeckBySpritesheetName[name]
end

local function cacheSpritesheet(name, deck)
  Caches.spritesheetDeckBySpritesheetName[name] = deck
  for name in pairs(deck.frameIndices) do
    cacheSpritesheetSprite(name, deck)
  end
end

local function uncacheSpritesheet(name)
  local deck = getCachedSpritesheetDeck(name)
  for name in pairs(deck.frameIndices) do
    uncacheSpritesheetSprite(name)
  end
  Caches.spritesheetDeckBySpritesheetName[name] = nil
end


-- ----------------------------------------------------------------------------
-- Standalone Sprite Construction --
-- ----------------------------------------------------------------------------

--- Create a MOAIGfxQuad2D (deck) for a MOAITexture of a standalone image.
-- @param texture (MOAITexture) - the texture to use to create the deck.
-- @return (MOAIGfxQuad2D) - a deck set to use the given texture.
local function createDeckForTexture(texture)
  local deck = MOAIGfxQuad2D.new()
  deck:setTexture(texture)

  local width, height = texture:getSize()
  local contentScale  = Host.contentScale
  local rect = {0.0, 0.0, width/contentScale, height/contentScale}
  deck:setRect(unpack(rect))
  deck.originalRect = rect
  deck.visibleRect = TableUtils.shallowCopy(rect)

  return deck
end


-- ----------------------------------------------------------------------------
-- Sprite Sheet Construction --
-- ----------------------------------------------------------------------------

local function getUVQuadMapping(uvRectangle, isTextureRotated)
  local u0, v0 = uvRectangle.u0, uvRectangle.v0
  local u1, v1 = uvRectangle.u1, uvRectangle.v1

  local x0, y0
  local x1, y1
  local x2, y2
  local x3, y3

  if (isTextureRotated) then
    x0, y0 = u1, v0
    x1, y1 = u1, v1
    x2, y2 = u0, v1
    x3, y3 = u0, v0
  else
    x0, y0 = u0, v0
    x1, y1 = u1, v0
    x2, y2 = u1, v1
    x3, y3 = u0, v1
  end

  local uvQuadMapping = {x0, y0, x1, y1, x2, y2, x3, y3}
  return uvQuadMapping
end

local function getSpriteRectangleMapping(spriteColorRectangle, spriteSourceSize)
  local contentScale = Host.contentScale

  local cX, cY = spriteColorRectangle.x, spriteColorRectangle.y
  local cWidth, cHeight = spriteColorRectangle.width, spriteColorRectangle.height
  local sHeight = spriteSourceSize.height

  local x0 = cX / contentScale
  local x1 = (cX + cWidth) / contentScale

  local y0 = (sHeight - cY - cHeight) / contentScale
  local y1 = (sHeight - cY) / contentScale

  local spriteRectangleMapping = {x0, y0, x1, y1}
  return spriteRectangleMapping
end

local function getBoundsMapping(spriteSourceSize)
  local contentScale = Host.contentScale
  local width = spriteSourceSize.width / contentScale
  local height = spriteSourceSize.height / contentScale

  local boundsMapping = {0.0, 0.0, 0, width, height, 0}
  return boundsMapping
end

local function buildSpritesheetDeck(texture, frames)
  local framesCount = #frames

  local spritesheetDeck = MOAIGfxQuadDeck2D.new()
  spritesheetDeck:setTexture(texture)
  spritesheetDeck:reserve(framesCount)

  local boundsDeck = MOAIBoundsDeck.new()
  boundsDeck:reserveIndices(framesCount)
  boundsDeck:reserveBounds(framesCount)

  spritesheetDeck:setBoundsDeck(boundsDeck)

  spritesheetDeck.frameIndices = {}

  for index, frame in ipairs(frames) do
    local uvRectangle = frame.uvRect
    local isTextureRotated = frame.textureRotated
    local uvQuadMapping = getUVQuadMapping(uvRectangle, isTextureRotated)
    spritesheetDeck:setUVQuad(index, unpack(uvQuadMapping))

    local spriteColorRectangle = frame.spriteColorRect
    local spriteSourceSize = frame.spriteSourceSize
    local spriteRectangleMapping = getSpriteRectangleMapping(spriteColorRectangle, spriteSourceSize)
    spritesheetDeck:setRect(index, unpack(spriteRectangleMapping))

    local boundsMapping = getBoundsMapping(spriteSourceSize)
    boundsDeck:setBounds(index, unpack(boundsMapping))

    spritesheetDeck.frameIndices[frame.name] = index
  end

  return spritesheetDeck
end

local function getSpritesheetBasePath(name)
  return "sheets/" .. name .. "/" .. name
end

local function getSpritesheetDefinitionPath(name)
  return getSpritesheetBasePath(name) .. Host.assetSuffix .. ".lua"
end

-- Returns the relative path for a spritesheet without the asset suffix,
-- since the asset suffix will be added later by TextureLoader
local function getSpritesheetTexturePath(name)
  local basePath = getSpritesheetBasePath(name)
  local pathWithExtension = basePath .. DEFAULT_SPRITESHEET_EXTENSION
  local pathWithAssetSuffix = FileUtils.addSuffixBeforeExtension(pathWithExtension, Host.assetSuffix)
  local absolutePath = FileUtils.findFile(pathWithAssetSuffix)
  if (absolutePath ~= nil) then
    return pathWithExtension
  end

  error("No spritesheet found with name '" .. tostring(name) .. "'.")
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- TODO: Visible area only works with standalone sprites right now. Sprite sheets complicate things a lot.

function Sprite:setVisibleRect(minX, minY, maxX, maxY)
  local rect = self.deck.originalRect
  assert(rect, "setVisibleRect currently only works with standalone sprites, not sprite sheets.")

  local width = rect[3] - rect[1]
  local height = rect[4] - rect[2]

  local minXProportion = minX / width
  local minYProportion = minY / height
  local maxXProportion = maxX / width
  local maxYProportion = maxY / height

  self.deck.visibleRect = {minX, maxX, minY, maxY}
  self.deck:setRect(minX, minY, maxX, maxY)
  self.deck:setUVQuad(minXProportion, 1.0 - maxYProportion,
                      maxXProportion, 1.0 - maxYProportion,
                      maxXProportion, 1.0 - minYProportion,
                      minXProportion, 1.0 - minYProportion)
end


-- ----------------------------------------------------------------------------
-- Sprite sheet API --
-- ----------------------------------------------------------------------------

-- Name is something like "swell_sheet_1".  No asset suffix.  No file extension.

--- Load a spritesheet and cache in texture memory so that it can be used to create Sprites.  @see newWithName
-- If a spritesheet has already been cached for the provided name, it will not be loaded again.
-- Spritesheets consist of an image texture file (sheet_name@*x.pvr) and a frames file (sheet_name@*x.pvr).
-- Once loaded, sprites for each of the image files in the spritesheet will be cached by their respective filenames (without asset suffix).
-- @param name (string) - The name to index the spritesheet.  If the filenames for a spritesheet are "image_name@4x.pvr" and "image_name@4x.lua", this will be "image_name"
function Sprite.loadSpritesheetWithName(name)
  Asserts.assertType(name, "name", "string")
  local spritesheetDeck = getCachedSpritesheetDeck(name)
  if not spritesheetDeck then
    local definitionLocalPath = getSpritesheetDefinitionPath(name)
    local definitionPath = FileUtils.findFile(definitionLocalPath)
    assert(definitionPath ~= nil, "No sprite data file found at ''" .. tostring(definitionLocalPath) .. "'")
    local spritesheetData = dofile(definitionPath)

    local textureLocalPath = getSpritesheetTexturePath(name)
    local shouldCache = true
    local texture = TextureLoader.loadTexture(textureLocalPath, shouldCache)

    spritesheetDeck = buildSpritesheetDeck(texture, spritesheetData.frames)
    cacheSpritesheet(name, spritesheetDeck)
  end
end

--- Remove a sprite sheet *and* any sprites contained in that spritesheet from the cache.
-- If no spritesheet is cached by that name, no action will be taken.
-- @param name (string) - The name to index the spritesheet.  If the full filenames for a spritesheet are "image_name@4x.pvr" and "image_name@4x.lua", this will be "image_name"
function Sprite.uncacheSpritesheetWithName(name)
  uncacheSpritesheet(name)
  TextureLoader.uncacheTexture(name)
end

--- Remove a standalone (not part of a sprite sheet) sprite from the cache.
-- @param name The name or filename used to create the sprite.
function Sprite.uncacheSpriteWithName(name)
  TextureLoader.uncacheTexture(name)
end

--- Get the deck and index of a sprite sheet based sprite.
-- For internal use only.
-- @param name The name of the sprite to get info for.
-- @return The info or nil if not found.
function Sprite.deckAndIndexOfSpritesheetSprite(name)
  local spritesheetDeck = getCachedSpritesheetSpriteDeck(name)
  if spritesheetDeck then
    local index = spritesheetDeck.frameIndices[name]
    if index then
      return spritesheetDeck, index
    end
  end

  return nil
end


-- ----------------------------------------------------------------------------
-- Sprite constructors --
-- ----------------------------------------------------------------------------

--- Create a sprite from the given deck with the given index.
-- Note that the sprite will *not* be cached automatically.  @see loadTextureWithFilename @see loadSpritesheetWithName
-- @param deck (MOAIDeck) - the deck to use for this Sprite.
-- @param index (optional integer) - the frame index to use for this Sprite, if the sprite is part of a spritesheet.  Pass 'nil' if it is a standalone image.
-- @return (Sprite) - A new (uncached) Sprite
function Sprite.newWithDeck(deck, index)
  local self = Group.newUnsized() -- The sprite deck controls this group's dimensions.
  Mixin.mix(self, Sprite)
  self:setDeck(deck)
  self.deck = deck
  if index then
    self:setIndex(index)
  end
  self:setAnchorPointCenter()

  return self
end

--- Create a sprite from the given MOAITexture.
-- @param texture (MOAITexture) - the texture to use for this Sprite.
-- @return (Sprite) - A new Sprite
function Sprite.newWithTexture(texture)
  local deck = createDeckForTexture(texture)
  return Sprite.newWithDeck(deck)
end

--- Create a Sprite with the image file at the given filename.
-- @param imageFilenameWithoutAssetSuffix (string) - The filename (without asset suffix) of the image file from which to create a texture. (image_name.png/pvr/jpg)
-- @param shouldCache If true, cache the sprite's resources (default true).
-- @return (Sprite) - A new Sprite
function Sprite.newWithFilename(imageFilenameWithoutAssetSuffix, shouldCache)
  Asserts.assertType(imageFilenameWithoutAssetSuffix, "imageFilenameWithoutAssetSuffix", "string")
  local texture = TextureLoader.loadTexture(imageFilenameWithoutAssetSuffix, shouldCache)
  return Sprite.newWithTexture(texture)
end

--- Create a sprite from a spritesheet or a standalone texture.
-- If no image with the given name has already been cached, this function will attempt to create a
-- sprite from a standalone image on the filesystem.  @see newWithFilename
-- @param name The filename of this sprite. Sprite will be cached under this name.
-- @param shouldCache If true, cache the sprite's resources (default true).
-- @return (Sprite) - A new Sprite
function Sprite.newWithName(name, shouldCache)
  Asserts.assertType(name, "name", "string")

  local deck, index = Sprite.deckAndIndexOfSpritesheetSprite(name)
  if not deck then
    return Sprite.newWithFilename(name, shouldCache)
  end

  return Sprite.newWithDeck(deck, index)
end
