--- Standard UI Builder
-- Builds UI components in a standardized manner

-- ===========================================================================
class("StandardUIBuilder")
-- ===========================================================================

require("UIOverlay")
require("widgets.progress.PieTimer")
require("widgets.progress.DigitTimer")
require("widgets.PauseButton")
require("widgets.LivesDisplay")
require("widgets.ScoreLabel")
require("bonuses.BonusNotificationTray")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

StandardUIBuilder.STANDARD_PIE_TIMER_RADIUS = 12.0
StandardUIBuilder.STANDARD_BUTTON_SMALL_WIDTH = 44.0
StandardUIBuilder.STANDARD_FONT = FontFaces.DIN_OT_BOLD
StandardUIBuilder.STANDARD_FONT_SIZE = 20.0

StandardUIBuilder.STANDARD_INDICATOR_Y = Screen.height * 0.85

StandardUIBuilder.STANDARD_BAR_MARGIN = 40.0
StandardUIBuilder.STANDARD_BAR_RADIUS = 4.0

StandardUIBuilder.STANDARD_SCORE_MARGIN_X = 12.0
StandardUIBuilder.BONUS_SCORE_MARGIN_Y = 4.0

-- Offset to ensure east and west oriented components are properly aligned.
local EASTWEST_OFFSET = 10.0

local STANDARD_PIE_TIMER_OPACITY_ACTIVE = 1.0
local STANDARD_PIE_TIMER_OPACITY_PAUSED = 0.25

local STANDARD_BONUS_TRAY_X_OFFSET = -10.0
local STANDARD_BONUS_TRAY_Y_OFFSET = -12.0


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createHiddenBonusIndicator(self, indicatorClass)
  local indicator = indicatorClass.new()
  indicator:setVisible(false)

  local yPosition = StandardUIBuilder.STANDARD_INDICATOR_Y
  self.uiOverlay:addSideIndicator(indicator, yPosition)

  return indicator
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Applies a standard x offset to a UIOverlay widget and aligns the widget's anchor point (when xOffset is not specified),
-- then adds the widget to the UIOverlay.
-- @param (Group) uiOverlayWidget - the UIOverlay widget to add.
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
function StandardUIBuilder:addWidget(uiOverlayWidget, layoutRegion, xOffset, yOffset)
  if not xOffset then
    xOffset = 0.0
    if layoutRegion == UIOverlay.NORTHWEST or layoutRegion == UIOverlay.SOUTHWEST or layoutRegion == UIOverlay.WEST then
      uiOverlayWidget:setAnchorPointLeft()
      xOffset = -EASTWEST_OFFSET
    elseif layoutRegion == UIOverlay.NORTHEAST or layoutRegion == UIOverlay.SOUTHEAST or layoutRegion == UIOverlay.EAST then
      uiOverlayWidget:setAnchorPointRight()
      xOffset = EASTWEST_OFFSET
    end
  end

  self.uiOverlay:addWidget(uiOverlayWidget, layoutRegion, xOffset, yOffset)
end

--- Creates and adds the standard timer widget to the UIOverlay.
-- @param (number) duration - the duration of the timer
-- @param (function) onCompletion - called with no arguments when the timer completes.
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST) (Default = UIOverlay.NORTHEAST)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new widget.
function StandardUIBuilder:addTimer(duration, onCompletion, layoutRegion, xOffset, yOffset)
  return self:addPieTimer(duration, onCompletion, layoutRegion, xOffset, yOffset)
end

--- Creates and adds a PieTimer to the UIOverlay
-- @param (number) duration - the duration of the timer
-- @param (function) onCompletion - called with no arguments when the timer completes.
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST) (Default = UIOverlay.NORTHEAST)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new widget.
function StandardUIBuilder:addPieTimer(duration, onCompletion, layoutRegion, xOffset, yOffset)
  layoutRegion = layoutRegion or UIOverlay.NORTHEAST

  local timerColor = self.uiColor
  local backOpacity = 0.3
  local timer = PieTimer.new(StandardUIBuilder.STANDARD_PIE_TIMER_RADIUS, timerColor, backOpacity, duration, onCompletion)
  timer:addStandardWarningSequence()

  timer:setActiveOpacity(STANDARD_PIE_TIMER_OPACITY_ACTIVE)
  timer:setPausedOpacity(STANDARD_PIE_TIMER_OPACITY_PAUSED)

  self:addWidget(timer, layoutRegion, xOffset, yOffset)

  return timer
end

--- Creates and adds a DigitTimer to the UIOverlay
-- @param (number) duration - the duration of the timer
-- @param (function) onCompletion - called with no arguments when the timer completes.
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST) (Default = UIOverlay.NORTHEAST)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new widget.
function StandardUIBuilder:addDigitTimer(duration, onCompletion, layoutRegion, xOffset, yOffset, font, fontSize)
  layoutRegion = layoutRegion or UIOverlay.NORTHEAST

  local timerColor = TableUtils.shallowCopy(self.uiColor)
  local timer = DigitTimer.new(duration, onCompletion, font, fontSize, timerColor)
  timer:addStandardWarningSequence()

  self:addWidget(timer, layoutRegion, xOffset, yOffset)

  return timer
end

--- Add a label to the UIOverlay.
-- @param (string) initialText (optional) The text to start with. (Default = "")
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST) (Default = UIOverlay.NORTH)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new widget.
function StandardUIBuilder:addLabel(initialText, layoutRegion, xOffset, yOffset)
  initialText = initialText or ""
  layoutRegion = layoutRegion or UIOverlay.NORTH

  local label = DeprecatedLabel.newSingleLine(initialText, StandardUIBuilder.STANDARD_FONT, StandardUIBuilder.STANDARD_FONT_SIZE)
  label:setColor(unpack(self.uiColor))

  self:addWidget(label, layoutRegion, xOffset, yOffset)

  return label
end

--- Adds the pause button to the northwest corner of the UIOverlay.
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new pause button.
function StandardUIBuilder:addPauseButton(xOffset, yOffset)
  local layoutRegion = UIOverlay.NORTHWEST

  local pauseButton = PauseButton.new(self.uiColor)
  self:addWidget(pauseButton, layoutRegion, xOffset, yOffset)
  self.uiOverlay.pauseButton = pauseButton

  return pauseButton
end

--- Adds the score label to the northwest corner of the UIOverlay.
-- @param (number) xOffset - the x offset (+ standard score margin) to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new score label.
function StandardUIBuilder:addScoreLabel(xOffset, yOffset)
  xOffset = xOffset or 0
  local layoutRegion = UIOverlay.NORTHWEST

  local scoreLabel = ScoreLabel.new(StandardUIBuilder.STANDARD_FONT, StandardUIBuilder.STANDARD_FONT_SIZE, self.uiColor)

  scoreLabel:setAnchorPointLeft()
  xOffset = xOffset + StandardUIBuilder.STANDARD_SCORE_MARGIN_X
  self:addWidget(scoreLabel, layoutRegion, xOffset, yOffset)

  return scoreLabel
end

--- Adds the bonus tray to the northwest corner of the UIOverlay.
-- @param (number) xOffset - the x offset (+ standard bonus tray offset) to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset (+ standard bonus tray offset) to apply to the location in UIOverlay.
-- @return The new bonus tray.
function StandardUIBuilder:addBonusTray(xOffset, yOffset)
  xOffset = xOffset or 0
  yOffset = yOffset or 0
  local layoutRegion = UIOverlay.NORTHWEST

  local bonusTray = BonusNotificationTray.new(self.uiColor)
  bonusTray:setAnchorPointTopLeft()
  xOffset = xOffset + STANDARD_BONUS_TRAY_X_OFFSET
  yOffset = yOffset + STANDARD_BONUS_TRAY_Y_OFFSET
  self:addWidget(bonusTray, layoutRegion, xOffset, yOffset)

  return bonusTray
end

--- Add a lives display to the UIOverlay.
-- @param (optional integer) numberOfLives - Number lives in this lives counter
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
-- @return The new widget.
function StandardUIBuilder:addLivesDisplay(numberOfLives, xOffset, yOffset)
  local display = LivesDisplay.new(numberOfLives, self.uiColor)
  local layoutRegion = UIOverlay.NORTHEAST

  self:addWidget(display, layoutRegion, xOffset, yOffset)

  return display
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new StandardUIBuilder that uses uiColor when creating widgets.
-- @param (Group) uiOverlay - The overlay to which this StandardUIBuilder will add widgets.  (Use BaseGame:getUIOverlay())
-- @param (table) uiColor - The color to apply to widgets created by this StandardUIBuilder
-- @return The new StandardUIBuilder.
function StandardUIBuilder.new(uiOverlay, uiColor)
  local self = {}
  Mixin.mix(self, StandardUIBuilder)

  self.uiOverlay = uiOverlay

  self.uiColor = uiColor
  if not uiColor[4] then
    uiColor[4] = 1
  end

  return self
end
