--- Texture Loader

-- ===========================================================================
class("TextureLoader")
-- ===========================================================================

require("utils.FileUtils")
require("utils.StringUtils")

require("MetalPrebinder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ASSET_SUFFIX = Host.assetSuffix

local TextureCache = {}

local PVR_EXTENSION = ".pvr"
local PVR_ZIPPED_EXTENSION = ".pvr.gz"

local REQUIRED_STANDALONE_IMAGE_WIDTH_DIVISOR = 4

-- http://stackoverflow.com/a/1838702/401024
-- MAX window size + 1111 = 10000 + 1111 = 111111 = (1 << 5) - 1
local ZLIB_WINDOW_FLAGS = (2 ^ 5) - 1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function assertThatTextureLoaded(texture, absolutePath)
  local textureWidth = texture:getSize()
  assert(textureWidth > 0, "Texture from `" .. absolutePath .. "` has width zero (failed to load).")
end

local function cacheTexture(pathWithoutAssetSuffix, texture)
  TextureCache[pathWithoutAssetSuffix] = texture
end

-- Loading PVR and PVR.GZ Images ------

local function isPathForPvr(imagePath)
  return (StringUtils.endsWith(imagePath, PVR_ZIPPED_EXTENSION) or StringUtils.endsWith(imagePath, PVR_EXTENSION))
end

local function createDataSourceForPvr(absolutePath)
  local buffer = MOAIDataBuffer.new()
  buffer:load(absolutePath, MOAIDataBuffer.INFLATE_ON_EXT, ZLIB_WINDOW_FLAGS)

  return buffer
end

-- Loading PNG and JPG Images ---------

local function createDataSourceForPngOrJpg(absolutePath)
  local image = MOAIImage.new()
  image:load(absolutePath, MOAIImage.TRUECOLOR + MOAIImage.PREMULTIPLY_ALPHA)

  -- Moai has a bug that requires the *width* of images without an alpha channel to be divisible
  -- by 4.  Loading an image without an alpha channel whose width is not divisible by 4 will result
  -- in distortion.  If we detect that an image will break, use a copy of the image that includes
  -- an alpha channel as the data source.

  local doesImageHaveAlphaChannel = (image:getFormat() == MOAIImage.COLOR_FMT_RGBA_8888)
  local imageWidth = image:getSize()
  local isImageWidthLoadable = (imageWidth % REQUIRED_STANDALONE_IMAGE_WIDTH_DIVISOR == 0)
  local needsColorFormatChange = (not (doesImageHaveAlphaChannel or isImageWidthLoadable))
  if (needsColorFormatChange) then
    local reformattedImage = image:convertColors(MOAIImage.COLOR_FMT_RGBA_8888)
    print("WARNING: Image at `" .. absolutePath .. "` was reformatted to avoid Moai's distortion bug.")
    return reformattedImage
  end

  return image
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TextureLoader.uncacheTexture(pathWithoutAssetSuffix)
  TextureCache[pathWithoutAssetSuffix] = nil
end

function TextureLoader.uncacheAllTextures()
  TextureCache = {}
end

function TextureLoader.getCachedTexture(pathWithoutAssetSuffix)
  return TextureCache[pathWithoutAssetSuffix]
end

--- Loads a texture from the given path.  Caches the texture if `shouldCache` is `nil` or a non-false value.
-- `pathWithoutAssetSuffix` can be absolute (to load a texture from anywhere in the filesystem) or relative
-- (to load a texture from the game's assets directory, or the shared assets directory).  Throws an error if
-- no texture is found with the given path.
-- @param pathWithoutAssetSuffix (string) - the path of the texture. (For example: "image_name.png")
-- @param shouldCache (boolean) - whether or not to cache the texture once it is loaded.  Defaults to true.
-- @return (MOAITexture) - the loaded texture.
function TextureLoader.loadTexture(pathWithoutAssetSuffix, shouldCache)
  Asserts.assertType(pathWithoutAssetSuffix, "pathWithoutAssetSuffix", "string")

  local texture = TextureCache[pathWithoutAssetSuffix]
  if (texture == nil) then
    local pathPlusAssetSuffix = FileUtils.addSuffixBeforeExtension(pathWithoutAssetSuffix, ASSET_SUFFIX)
    local absolutePath = FileUtils.findFile(pathPlusAssetSuffix)
    assert(absolutePath ~= nil, "Unable to find texture file for `" .. tostring(pathWithoutAssetSuffix) .. "`")

    texture = MOAITexture.new()
    texture:setFilter(MOAITexture.GL_LINEAR, MOAITexture.GL_LINEAR)

    local dataSource = nil
    if (isPathForPvr(absolutePath)) then
      dataSource = createDataSourceForPvr(absolutePath)
    else
      dataSource = createDataSourceForPngOrJpg(absolutePath)
    end

    texture:load(dataSource)
    assertThatTextureLoaded(texture, absolutePath)

    MetalPrebinder.prebindTexture(texture)

    if ((shouldCache) or (shouldCache == nil)) then
      cacheTexture(pathWithoutAssetSuffix, texture)
    end
  end

  return texture
end
