--- Timer mixin.
-- Provides full timer functionality, including start, stop, pause, resume,
-- and time measurement functions in seconds and normalized.
--
-- Timer also supports two callbacks:
-- - Update callback: Called whenever the time is updated.
-- - Timeout callback: Called when the time runs out.

-- ===========================================================================
class("Timer")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- CONFIGURATION
-- ----------------------------------------------------------------------------

TIME_EPSILON = 1/45

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function callWarningHandler(self)
  if self._Timer_nextWarningIndex >= 1 then
    local timeRemaining = self:getTimeRemaining()
    local nextWarningPoint = self._Timer_warningPoints[self._Timer_nextWarningIndex]
    if math.abs(nextWarningPoint - timeRemaining) < TIME_EPSILON then
      self._Timer_nextWarningIndex = self._Timer_nextWarningIndex - 1
      self._Timer_warningHandlers:start()
    end
  end
end

local function updateNextWarningIndex(self)
  if self._Timer_duration then
    local timeRemaining = self:getTimeRemaining()
    for index, warningPoint in ipairs(self._Timer_warningPoints) do
      if warningPoint + TIME_EPSILON < timeRemaining then
        self._Timer_nextWarningIndex = index
      else
        break
      end
    end
  end
end

local function updateWarningIntervals(self)
  if self._Timer_warningInterval and self._Timer_warningCount then
    local interval = self._Timer_warningInterval
    local timeOffset = self._Timer_warningTimeOffset or 0
    local time = interval
    local warningPoints = {}
    for i = 1, self._Timer_warningCount do
      local currentTime = time + timeOffset
      if currentTime >= 0 then
        table.insert(warningPoints, currentTime)
      end
      time = time + interval
    end
    self._Timer_warningPoints = warningPoints
    updateNextWarningIndex(self)
    self._Timer_callWarningHandler = callWarningHandler
  else
    self._Timer_callWarningHandler = function() end
  end
end

local function callUpdateHandler(self)
  self._Timer_callWarningHandler(self)
  self._Timer_updateHandler()
end

local function startCoroutine(self)
  if not self._Timer_coroutine then
    self._Timer_coroutine = runInCoroutine(function()
      while self._Timer_timeCounter:isBusy() do
        callUpdateHandler(self)
        coroutine.yield()
      end

      self._Timer_coroutine = nil
      callUpdateHandler(self)
      if self._Timer_timeoutHandler then
        self._Timer_timeoutHandler()
      end
    end)
  end
end

local function stopCoroutine(self)
  if self._Timer_coroutine then
    self._Timer_coroutine:stop()
    self._Timer_coroutine = nil
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Start/resume the timer.
function Timer:start()
  self._Timer_isPaused = false
  if self:getTimeRemaining() == 0 then
    self:setTime(0)
    updateNextWarningIndex(self)
  end
  self._Timer_timeCounter:start()
  startCoroutine(self)
end

-- Setup the timer, and pause it.
function Timer:initialize()
  self:start()
  self:setPaused(true)
end

--- Stop the timer.
-- Note: This will not trigger the timeout handler.
function Timer:stop()
  self._Timer_isPaused = false
  stopCoroutine(self)
  self._Timer_timeCounter:stop()
end

function Timer:pause()
  self:setPaused(true)
end

function Timer:resume()
  self:setPaused(false)
end

function Timer:setPaused(isPaused)
  if isPaused ~= self._Timer_isPaused then
    self._Timer_isPaused = isPaused
    if isPaused then
      stopCoroutine(self)
      self._Timer_timeCounter:pause()
    else
      self._Timer_timeCounter:pause(false)
      startCoroutine(self)
    end
  end
end

function Timer:getPaused()
  return self._Timer_isPaused
end

function Timer:getUpdateHandler()
  return self._Timer_updateHandler
end

function Timer:setUpdateHandler(handler)
  self._Timer_updateHandler = handler or function() end
end

function Timer:getTimeoutHandler()
  return self._Timer_timeoutHandler
end

function Timer:setTimeoutHandler(handler)
  self._Timer_timeoutHandler = handler
end

function Timer:getDuration()
  return self._Timer_duration
end

function Timer:setDuration(duration)
  self._Timer_duration = duration
  self._Timer_timeCounter:setSpan(self._Timer_duration)
  updateNextWarningIndex(self)
  callUpdateHandler(self)
end

function Timer:getTime()
  return self._Timer_timeCounter:getTime()
end

function Timer:setTime(time)
  time = MathUtils.clamp(time, 0, self:getDuration())
  self._Timer_timeCounter:setTime(time)
  updateNextWarningIndex(self)
  callUpdateHandler(self)
end

function Timer:addTime(amount)
  self:setTime(self:getTime() + amount)
end

function Timer:addTimeRemaining(amount)
  self:addTime(-amount)
end

function Timer:getTimeRemaining()
  local remaining = self:getDuration() - self:getTime()
  if remaining < 0 then
    remaining = 0
  end

  return remaining
end

function Timer:setTimeRemaining(timeRemaining)
  local duration = self:getDuration()
  assert(duration, "Duration not set")
  self:setTime(duration - timeRemaining)
end

--- Time as a value from 0 to 1
-- @param time - (float) Optionally specify the time to calculate for.
function Timer:getNormalizedTime(time)
  time = time or self:getTime()
  return time / self:getDuration()
end

function Timer:setNormalizedTime(normalizedTime)
  self:setTime(normalizedTime * self:getDuration())
end

--- Time remaining as a value from 0 to 1
function Timer:getNormalizedTimeRemaining()
  return self:getTimeRemaining() / self:getDuration()
end

function Timer:setNormalizedTimeRemaining(normalizedTimeRemaining)
  self:setTimeRemaining(normalizedTimeRemaining * self:getDuration())
end

--- Time between warning pings
function Timer:getWarningInterval()
  return self._Timer_warningInterval
end

function Timer:setWarningInterval(duration)
  self._Timer_warningInterval = duration
  updateWarningIntervals(self)
end

--- Number of warning pings
function Timer:getWarningCount()
  return self._Timer_warningCount
end

function Timer:setWarningCount(count)
  self._Timer_warningCount = count
  updateWarningIntervals(self)
end

function Timer:getWarningTimeOffset()
  return self._Timer_warningTimeOffset
end

function Timer:setWarningTimeOffset(offset)
  self._Timer_warningTimeOffset = offset
  updateWarningIntervals(self)
end

function Timer:addWarningHandler(handler)
  self._Timer_warningHandlers:addAction(handler)
end


--- Turn an object into a timer.
function Timer.makeTimer(self, duration)
  Mixin.mix(self, Timer)
  self._Timer_timeCounter = MOAITimer.new()
  self:setUpdateHandler(nil)
  self._Timer_warningPoints = {}
  self._Timer_nextWarningIndex = 0
  self._Timer_warningHandlers = AHSpawn.new()
  updateWarningIntervals(self)
  if duration then
    self:setDuration(duration)
  end

  self.isTimer = true
end


-- ===========================================================================
-- AHTimeTo
-- ===========================================================================

class("AHTimeTo")
function AHTimeTo.new(target, duration, ease, amount)
  return AHNonAttributeTo.new(target, duration, target.getTime, target.setTime, ease, amount)
end

class("AHNormalizedTimeTo")
function AHNormalizedTimeTo.new(target, duration, ease, amount)
  return AHNonAttributeTo.new(target, duration, target.getNormalizedTime, target.setNormalizedTime, ease, amount)
end

class("AHTimeRemainingTo")
function AHTimeRemainingTo.new(target, duration, ease, amount)
  return AHNonAttributeTo.new(target, duration, target.getTimeRemaining, target.setTimeRemaining, ease, amount)
end

class("AHNormalizedTimeRemainingTo")
function AHNormalizedTimeRemainingTo.new(target, duration, ease, amount)
  return AHNonAttributeTo.new(target, duration, target.getNormalizedTimeRemaining, target.setNormalizedTimeRemaining, ease, amount)
end


-- ===========================================================================
-- AHTimeBy
-- ===========================================================================

class("AHTimeBy")
function AHTimeBy.new(target, duration, ease, amount)
  return AHNonAttributeBy.new(target, duration, target.getTime, target.setTime, ease, amount)
end

class("AHNormalizedTimeBy")
function AHNormalizedTimeBy.new(target, duration, ease, amount)
  return AHNonAttributeBy.new(target, duration, target.getNormalizedTime, target.setNormalizedTime, ease, amount)
end

class("AHTimeRemainingBy")
function AHTimeRemainingBy.new(target, duration, ease, amount)
  return AHNonAttributeBy.new(target, duration, target.getTimeRemaining, target.setTimeRemaining, ease, amount)
end

class("AHNormalizedTimeRemainingBy")
function AHNormalizedTimeRemainingBy.new(target, duration, ease, amount)
  return AHNonAttributeBy.new(target, duration, target.getNormalizedTimeRemaining, target.setNormalizedTimeRemaining, ease, amount)
end
