--- TouchHandler
-- @author Created by karl on June 18, 2014.

-- ===========================================================================
class("TouchHandler")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Handler Behaviors ------------------

-- Eclusivity: Choose one of EXCLUSIVE, FOCUS, or ALWAYS.
-- Most likely, you'll want to use FOCUS or ALWAYS. EXCLUSIVE allows you to monopolize the
-- touch event UNTIL the touch moves outside of your target's area.

-- Only handle the touch if no other exclusive handlers have done so already.
TouchHandler.BEHAVIOR_EXCLUSIVE = "TouchHandler.BEHAVIOR_EXCLUSIVE"

-- On top of being exclusive, also take "focus", meaning commandeer all "TouchManager.TOUCH_MOVED"
-- state events for the same touch ID until an "TouchManager.TOUCH_ENDED" state event is received.
TouchHandler.BEHAVIOR_FOCUS     = "TouchHandler.BEHAVIOR_FOCUS"

-- Always try to handle the touch.
TouchHandler.BEHAVIOR_ALWAYS    = "TouchHandler.BEHAVIOR_ALWAYS"

-- Other behaviors --------------------

-- Accept touch hand-offs from other handlers.
TouchHandler.BEHAVIOR_HANDOFF   = "TouchHandler.BEHAVIOR_HANDOFF"


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isInScreenHierarchy(self)
  return self.target.parent ~= nil
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TouchHandler:hasBehavior(behavior)
  return self.handlerBehaviors[behavior]
end

function TouchHandler:handlesAlways()
  return self:hasBehavior(TouchHandler.BEHAVIOR_ALWAYS)
end

function TouchHandler:isExclusive()
  return self:hasBehavior(TouchHandler.BEHAVIOR_EXCLUSIVE)
end

function TouchHandler:isFocused()
  return self:hasBehavior(TouchHandler.BEHAVIOR_FOCUS)
end

function TouchHandler:supportsHandoff()
  return self:hasBehavior(TouchHandler.BEHAVIOR_HANDOFF)
end

function TouchHandler:takesFocus(event)
  return self:isFocused() and event:isTouchStarting()
end

function TouchHandler:canHandle(event)
  if not isInScreenHierarchy(self) then
    return false
  end

  if event:isHandlerBlacklisted(self) then
    return false
  end

  if event:isHandoff() then
    if not self:supportsHandoff() then
      return false
    end
  end

  local isInBounds = self.target:inside(event.x, event.y)

  if self:handlesAlways() then
    return isInBounds
  end

  if self:isExclusive() and event:isExclusive() and not event:isExclusiveTo(self) then
    return false
  end

  if self:isFocused() and event:isExclusiveTo(self) and not event:isTouchStarting() then
    return true
  end

  return isInBounds
end

function TouchHandler:handleEvent(event)
  local eventHandling = self.handlerFunction(event, self.target)
  event:notifyHandledBy(self)

  return eventHandling or TouchEvent.HANDLING.CONTINUE
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function getBehaviors(handlerBehaviors)
  handlerBehaviors = handlerBehaviors or {}
  if type(handlerBehaviors) == "string" then
    handlerBehaviors = {handlerBehaviors}
  end

  handlerBehaviors = TableUtils.createSetFromArray(handlerBehaviors)

  if handlerBehaviors[TouchHandler.BEHAVIOR_ALWAYS] then
    assert(not (handlerBehaviors[TouchHandler.BEHAVIOR_EXCLUSIVE] or handlerBehaviors[TouchHandler.BEHAVIOR_FOCUS]), "Cannot mix behavior ALWAYS with EXCLUSIVE or FOCUS")
  end


  if not handlerBehaviors[TouchHandler.BEHAVIOR_ALWAYS] and
  not handlerBehaviors[TouchHandler.BEHAVIOR_EXCLUSIVE] and
  not handlerBehaviors[TouchHandler.BEHAVIOR_FOCUS] then
    handlerBehaviors[TouchHandler.BEHAVIOR_FOCUS] = true
  end

  if handlerBehaviors[TouchHandler.BEHAVIOR_FOCUS] then
    handlerBehaviors[TouchHandler.BEHAVIOR_EXCLUSIVE] = true
  end

  return handlerBehaviors
end

-- Constructor ------------------------

function TouchHandler.new(target, handlerFunction, handlerBehaviors, priority)
  assert(target, "Target cannot be nil")
  Asserts.assertType(handlerFunction, "handlerFunction", "function")
  Asserts.assertOptionalType(handlerBehaviors, "handlerBehaviors", "table", "string")

  local self = {
    target = target,
    handlerFunction = handlerFunction,
    handlerBehaviors = getBehaviors(handlerBehaviors),
    priority = priority
  }
  Mixin.mix(self, TouchHandler)

  return self
end
