--- TouchManager
--
-- Handles touches from the operating system and forwards them to any interested handlers.
--
-- Touch events are split into three phases: begin, move, and end. In most behavior
-- modes, touch events will flow from begin to move, to end.
--
-- Touch handlers may return a value that the touch manager will interpret. These values
-- are specified in TouchEvent.HANDLING. Returning nil (or not returning a value) implies
-- TouchEvent.HANDLING.CONTINUE.
--
-- Focused touch handlers can also hand off their touch to another handler. This is
-- done by returning TouchEvent.HANDLING.HAND_OFF from the handler function. Returning
-- implies TouchEvent.HANDLING.CONTINUE.
-- Only a "move" touch may be handed off. When a touch is handed off, it reverts to
-- a "begin" touch and is then passed to the next handler capable of receiving handed
-- off touch events (see TouchHandler.lua).
--
-- Handlers operate using a callback function that takes an event object and a
-- touch target.
--
-- By default, priority in the touch system mirrors the grup priority in the screen
-- hierarchy. You can optionally set your own priorities.
--
-- @author Created by karl on June 10, 2014.


-- ===========================================================================
class("TouchManager")
-- ===========================================================================

require("TouchEvent")
require("TouchHandler")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

TouchManager.MAX_ACCIDENTAL_MOVEMENT = 5

-- Touch event phases ------------------

TouchManager.TOUCH_BEGAN      = "TouchManager.TOUCH_BEGAN"
TouchManager.TOUCH_MOVED      = "TouchManager.TOUCH_MOVED"
TouchManager.TOUCH_ENDED      = "TouchManager.TOUCH_ENDED"

local TOUCH_EVENT_PHASES = {
  [0] = TouchManager.TOUCH_BEGAN, -- event 0
  [1] = TouchManager.TOUCH_MOVED, -- event 1
  [2] = TouchManager.TOUCH_ENDED  -- event 2
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local handlersByTarget = {}
local handlersByPriority = {}
local focusedTouches = {}

local activeEvents = {}

local function endEvent(eventID)
  activeEvents[eventID] = nil
end

local function isValidEvent(phase, id)
  return phase == TouchManager.TOUCH_BEGAN or activeEvents[id]
end

local function newEvent(phase, id, x, y, tapCount)
  local event = activeEvents[id]
  if event then
    event = event:clone()
  else
    event = TouchEvent.new(phase, id, x, y, tapCount)
  end
  event.phase = phase
  event.x = x
  event.y = y
  event.tapCount = tapCount
  event:clearHandoff()

  activeEvents[id] = event

  return event
end

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

--- Inform the handler of a touch event.
-- INTERNAL METHOD. DO NOT CALL FROM CLIENT CODE.
-- @param phase (string) - The event phase. See Configuration section.
-- @param id (object) - The touch ID,
-- @param x (number) - The touche's X position.
-- @param y (number) - The touche's Y position.
-- @param tapCount (number) - The number of taps for this event.
function TouchManager.notifyTouchEvent(phase, id, x, y, tapCount)
  if not isValidEvent(phase, id) then
    return
  end

  local event = newEvent(phase, id, x, y, tapCount)

  for _, handler in ipairs(TableUtils.shallowCopy(handlersByPriority)) do
    if handler:canHandle(event) then
      if handler:handleEvent(event) == TouchEvent.HANDLING.HAND_OFF then
        event:blacklistHandler(handler)
        if handler:isFocused() and event:isExclusiveTo(handler) and event:isTouchMoving() then
          event:beginHandoff()
        end
      end
    end
  end

  if event.phase == TouchManager.TOUCH_ENDED then
    endEvent(event.id)
  end
end

--- Notify the touch manager that a screen structure change has changed relative object priorities.
-- INTERNAL METHOD. DO NOT CALL FROM CLIENT CODE.
function TouchManager.notifyPriorityChange()
  local handlers = TableUtils.getValues(handlersByTarget)
  table.sort(handlers, function(a, b)
    local priorityA = a.priority or a.target:getPriority() or 0
    local priorityB = b.priority or b.target:getPriority() or 0
    return priorityA > priorityB
  end)
  handlersByPriority = handlers
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Register an object for touch handling.
-- @param target (prop) - The target that will be tested for touch hits.
-- @param handlerFunction (function taking param event) - Function to call when a touch hits the target.
-- @param handlerBehaviors (string or array, optional) - Special handler behavior(s). Default TouchHandler.BEHAVIOR_FOCUS
-- @param priority (number, optional) - Override to the standard screen-based priority.
function TouchManager.addTouchHandler(target, handlerFunction, handlerBehaviors, priority)
  handlerBehavior = handlerBehavior or TouchHandler.BEHAVIOR_FOCUS
  local handler = TouchHandler.new(target, handlerFunction, handlerBehaviors, priority)
  handlersByTarget[target] = handler
  TouchManager.notifyPriorityChange()
end

--- Remove an object from touch handling.
-- @param target (prop) - The object to remove.
function TouchManager.removeTouchHandler(target)
  local handler = handlersByTarget[target]
  if handler then
    handlersByTarget[target] = nil
    TableUtils.removeValueFromArray(handlersByPriority, handler)
    TableUtils.removeKeyForValue(focusedTouches, handler)
  end
end

--- Add touch enable functions to an object.
-- @param target (prop) - The target that will be tested for touch hits.
-- @param handlerFunction (function taking param event) - Function to call when a touch hits the target.
-- @param handlerBehaviors (string or array, optional) - Special handler behavior(s). Default TouchHandler.BEHAVIOR_FOCUS
-- @param priority (number, optional) - Override to the standard screen-based priority.
-- @param onEnableTouch (function, optional) - Function to call when touches are enabled.
-- @param onDisableTouch (function, optional) - Function to call when touches are disabled.
function TouchManager.addTouchEnabledMethods(target, handlerFunction, handlerBehaviors, priority, onEnableTouch, onDisableTouch)
  onEnableTouch = onEnableTouch or function() end
  onDisableTouch = onDisableTouch or function() end

  target.isTouchEnabled = function(self)
    return self._Touch_isTouchEnabled
  end

  target.setTouchEnabled = function(self, isEnabled)
    if isEnabled ~= self._Touch_isTouchEnabled then
      self._Touch_isTouchEnabled = isEnabled
      if isEnabled then
        TouchManager.addTouchHandler(target, handlerFunction, handlerBehaviors, priority)
        onEnableTouch()
      else
        TouchManager.removeTouchHandler(target)
        onDisableTouch()
      end
    end
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Initializer ------------------------

--- Initialize the touch manager.
function TouchManager.init()
  if not _TouchManager_initialized then
    local touchEventQueue = {}
    MOAIInputMgr.device.touch:setCallback(function(phaseIndex, id, x, y, tapCount)
      runExclusivelyInCoroutine(touchEventQueue, "TOUCH_QUEUE", true, function()
        local phase = TOUCH_EVENT_PHASES[phaseIndex]
        TouchManager.notifyTouchEvent(phase, id, x, y, tapCount)
      end)
    end)
    _TouchManager_initialized = true
  end
end

TouchManager.init() -- Run the initializer
