--- UI Overlay
--
-- An invisible full-screen Group that is automatically added in front of BaseGame.
-- Contains and keeps track of "UIOverlay Widgets," which are just Groups that can
-- be added to UIOverlay.  UIOverlay Widgets can be added at any of the 8 layout
-- regions corresponding to the cardinal directions.  (UIOverlay.NORTH, UIOverlay.WEST,
-- et cetera)
--
-- UIOverlay Widgets fade in and out during the default game intros and outros.
-- To customize the way these fades work, UIOverlay Widgets may implement the
-- following methods:
--
-- createFadeInAction(duration) - called to build game intro action
-- getKeyColor() - If createFadeInAction() is not implemented, UIOverlay will
--                 attempt to create a plain AHOpacityTo action with a destination opacity
--                 of this color.
-- createFadeOutAction(duration) - called to build game outro action
-- prepareForGameIntro() - called before game intro to allow widgets to set their
--                         opacity to 0.0 and prepare to fade in.
-- prepareForGameOutro() - called immediately before game ontro to give widgets
--                         a chance to stop any ongoing actions and prepare to fade out.
--
-- @author Created by Alexey Chernikov on May 6, 2013

-- ===========================================================================
class("UIOverlay")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Layout Regions ---------------------

local POSITIVE_MARGIN = 20.0
local NEGATIVE_MARGIN = -POSITIVE_MARGIN

UIOverlay.NORTH = {x = Screen.getSafeX(0.5), y = Screen.getSafeY(1.0) + NEGATIVE_MARGIN}
UIOverlay.SOUTH = {x = Screen.getSafeX(0.5), y = Screen.getSafeY(0.0) + POSITIVE_MARGIN}
UIOverlay.EAST = {x = Screen.getSafeX(1.0) + NEGATIVE_MARGIN, y = Screen.getSafeY(0.5)}
UIOverlay.WEST = {x = Screen.getSafeX(0.0) + POSITIVE_MARGIN, y = Screen.getSafeY(0.5)}
UIOverlay.NORTHEAST = {x = Screen.getSafeX(1.0) + NEGATIVE_MARGIN, y = Screen.getSafeY(1.0) + NEGATIVE_MARGIN}
UIOverlay.SOUTHEAST = {x = Screen.getSafeX(1.0) + NEGATIVE_MARGIN, y = Screen.getSafeY(0.0) + POSITIVE_MARGIN}
UIOverlay.NORTHWEST = {x = Screen.getSafeX(0.0) + POSITIVE_MARGIN, y = Screen.getSafeY(1.0) + NEGATIVE_MARGIN}
UIOverlay.SOUTHWEST = {x = Screen.getSafeX(0.0) + POSITIVE_MARGIN, y = Screen.getSafeY(0.0) + POSITIVE_MARGIN}

local LAYOUT_REGIONS = {
  [UIOverlay.NORTH] = true,
  [UIOverlay.SOUTH] = true,
  [UIOverlay.EAST] = true,
  [UIOverlay.WEST] = true,
  [UIOverlay.NORTHEAST] = true,
  [UIOverlay.SOUTHEAST] = true,
  [UIOverlay.NORTHWEST] = true,
  [UIOverlay.SOUTHWEST] = true
}

-- Z-Indices --------------------------

local Z_WIDGET = 10

-- Durations --------------------------

local DEFAULT_FADE_IN_DURATION = 0.2
local DEFAULT_FADE_OUT_DURATION = 0.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Default fade actions ---------------

local function createDefaultFadeInAction(widget, duration)
  local keyOpacity = nil
  if (widget.getKeyColor) then
    local keyColor = widget:getKeyColor()
    keyOpacity = keyColor[4]
  else
    keyOpacity = 1.0
  end
  local widgetFadeAction = AHOpacityTo.new(widget, duration, nil, keyOpacity)
  return widgetFadeAction
end

local function createDefaultFadeOutAction(widget, duration)
  local destinationOpacity = 0.0
  local widgetFadeAction = AHOpacityTo.new(widget, duration, nil, destinationOpacity)
  return widgetFadeAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

--- Gets the array of all widgets that have been added to the UIOverlay.
-- @return (table) - the widgets array
function UIOverlay:getWidgets()
  return self.widgets
end

-- Adding widgets ---------------------

--- Adds a UIView Widget to the UIOverlay at the specified layout region.
-- @param (Group) uiOverlayWidget - the UIOverlay widget to add.
-- @param (table) layoutRegion - One of the layout regions defined in UIOverlay.  (For example: UIOverlay.NORTH, UIOverlay.NORTHEAST)
-- @param (number) xOffset - the x offset to apply to the location in UIOverlay.
-- @param (number) yOffset - the y offset to apply to the location in UIOverlay.
function UIOverlay:addWidget(uiOverlayWidget, layoutRegion, xOffset, yOffset)
  assert(uiOverlayWidget, "Must provide a UI overlay widget to add")
  assert(LAYOUT_REGIONS[layoutRegion], "Must provide a layout region (For example, UIOverlay.NORTHWEST)")

  xOffset = xOffset or 0.0
  yOffset = yOffset or 0.0

  local x = layoutRegion.x + xOffset
  local y = layoutRegion.y + yOffset
  uiOverlayWidget:setLoc(x, y)

  self:insert(uiOverlayWidget, Z_WIDGET)
  table.insert(self.widgets, uiOverlayWidget)
end

-- Prepare for intro/outro ------------

--- Calls prepareForGameIntro() on all added UIOverlay Widgets to allow them to prepare to fade in.
-- If a widget does not implement prepareForGameInto(), will call setOpacity(0.0) instead.
function UIOverlay:prepareForGameIntro()
  local widgets = self.widgets
  for i, widget in ipairs(widgets) do
    if (widget.prepareForGameIntro) then
      widget:prepareForGameIntro()
    else
      widget:setOpacity(0.0)
    end
  end
end

--- Calls prepareForGameOutro() on all added UIOverlay Widgets to allow them to prepare to fade out.
-- Gives widgets a chance to stop any ongoing actions.  If a widget does not implement prepareForGameOutro(),
-- nothing will be called.
function UIOverlay:prepareForGameOutro()
  local widgets = self.widgets
  for i, widget in ipairs(widgets) do
    if (widget.prepareForGameOutro) then
      widget:prepareForGameOutro()
    end
  end
end

-- Intro/outro actions ----------------

--- Creates an action that fades in every UIOverlay Widget over the specified duration.  Uses createFadeInAction(duration)
-- from all UIOverlay Widgets that implement it.  If a widget does not implement createFadeInAction, it will use an AHOpacityTo
-- with the opacity from widget:getKeyColor().  If a widget does not implement getKeyColor(), it will fade in to an opacity of 1.0.
-- @param (number) duration - The duration passed to all widgets to create their fade in actions
-- @return (AHSpawn) - An AHSpawn combining all of the widgets' fade in actions.
function UIOverlay:createFadeInAction(duration)
  duration = duration or DEFAULT_FADE_IN_DURATION

  local widgetFadeActions = {}
  for i, widget in ipairs(self.widgets) do
    local widgetFadeAction = nil
    if (widget.createFadeInAction) then
      widgetFadeAction = widget:createFadeInAction(duration)
    else
      widgetFadeAction = createDefaultFadeInAction(widget, duration)
    end
    table.insert(widgetFadeActions, widgetFadeAction)
  end

  local fadeAction = AHSpawn.new(unpack(widgetFadeActions))
  return fadeAction
end

--- Creates and starts a fade in action.  @see UIOverlay:createFadeInAction()
function UIOverlay:fadeIn(duration)
  local fadeAction = self:createFadeInAction(duration)
  return fadeAction:start()
end

--- Creates an action that fades out every UIOverlay Widget over the specified duration.  Uses createFadeOutAction(duration)
-- from all UIOverlay Widgets that implement it.  If a widget does not implement createFadeOutAction, it will use an AHOpacityTo
-- with a destination opacity of 0.0.
-- @param (number) duration - The duration passed to all widgets to create their fade out actions
-- @param (AHSpawn) - An AHSpawn combining all of the widgets' fade out actions.
function UIOverlay:createFadeOutAction(duration)
  duration = duration or DEFAULT_FADE_OUT_DURATION
  local widgetFadeActions = {}
  for i, widget in ipairs(self.widgets) do
    local widgetFadeAction = nil
    if (widget.createFadeOutAction) then
      widgetFadeAction = widget:createFadeOutAction(duration)
    else
      widgetFadeAction = createDefaultFadeOutAction(widget, duration)
    end
    table.insert(widgetFadeActions, widgetFadeAction)
  end

  local fadeAction = AHSpawn.new(unpack(widgetFadeActions))
  return fadeAction
end

--- Creates and starts a fade out action.  @see UIOverlay:createFadeOutAction()
function UIOverlay:fadeOut(duration)
  local fadeAction = self:createFadeOutAction(duration)
  return fadeAction:start()
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new UIOverlay object.  Called automatically by BaseGame.
-- @return the new UIOverlay instance.
function UIOverlay.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, UIOverlay)

  self.widgets = {}

  return self
end
