--- AHAnimate
-- @author Created by Aaron Barrett on January 2, 2014.

-- Create flipbook-style sprite animations using AHAnimate.
-- AHAnimate requires an Animation object (see Animation.lua).
-- AHAnimate should be applied only to Sprite objects.  Note that Moai
-- does not automatically account for frames of varying size.  To compensate,
-- AHAnimate reads the anchor point of the target Sprite when started and
-- calls `setAnchorPoint` with the initial anchor point values each time
-- that the frame index is updated.

class("AHAnimate")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

AHAnimate.DEFAULT_FRAME_RATE = 24.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function onAnimateFrameUpdate(self)
  local target = self.target
  target:setAnchorPoint(self.anchorPointX, self.anchorPointY)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function AHAnimate:start()
  local target = self.target

  local animation = self.animation
  local lowestFrameNumber = animation.lowestFrameNumber
  target:setIndex(lowestFrameNumber)

  local anchorPointX, anchorPointY = target:getAnchorPoint()
  self.anchorPointX = anchorPointX
  self.anchorPointY = anchorPointY
  target:setAnchorPoint(anchorPointX, anchorPointY)

  local curve = self.curve
  self:setLink(1, curve, target, MOAIProp2D.ATTR_INDEX)

  return self.super.start(self)
end

-- ---------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new AHAnimate action.
-- If `duration` is provided, the animation will loop at `frameRate` for `duration` seconds.
-- Otherwise, the animation will run only once, and the duration will be determined by
-- the number of frames times `frameRate`.
-- @param target (MOAIProp2D) - The target to animate.  Should be a Sprite.  @see Sprite
-- @param animation (table) - An Animation object describing the frames to animate.
-- @param frameRate (number) - The number of frames per second at which to run the animation.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param duration (number, optional) - A duration in seconds during which the animation will loop.
-- @return (MOAIAnim) - A new AHAnimate action.
function AHAnimate.new(target, animation, frameRate, ease, duration)
  ease = AHEase.wrap(ease)
  frameRate = frameRate or AHAnimate.DEFAULT_FRAME_RATE
  duration = duration or (animation.frameCount / frameRate)

  assert(type(target) == "userdata", "`target` must be a Sprite.")
  assert(type(animation) == "table", "`animation` must be an Animation object. (see Animation.lua)")
  assert(type(frameRate) == "number", "`frameRate` must be a number.")
  assert(frameRate > 0.0, "`frameRate` must be positive.")
  assert((ease == nil) or (type(ease) == "userdata"), "`ease` must be `nil` or a MOAIEase.")
  assert(duration >= 0.0, "`duration` must not be negative.")

  local lowestFrameNumber = animation.lowestFrameNumber
  local animationFrameCount = animation.frameCount
  local numFramesToShow = math.floor(frameRate * duration)
  local lastFrameIndex = lowestFrameNumber

  local curve = MOAIAnimCurve.new()
  curve:reserveKeys(numFramesToShow)
  curve:setWrapMode(MOAIAnimCurve.APPEND)

  local frameTimeFunction = function(i)
    return ease:getValueAtTime(i / numFramesToShow) * duration
  end

  for i = 1, numFramesToShow do
    local time = frameTimeFunction(i - 1)
    lastFrameIndex = lowestFrameNumber + (i - 1) % animationFrameCount
    curve:setKey(i, time, lastFrameIndex, MOAIEaseType.FLAT)
  end

  local self = MOAIAnim.new()
  AHAction.makeAHAction(self)
  Mixin.mix(self, AHAnimate)

  self:reserveLinks(1)
  self:setSpan(duration)

  self:setCurve(curve)
  self:setListener(MOAITimer.EVENT_TIMER_KEYFRAME, onAnimateFrameUpdate)

  self.target = target
  self.duration = duration
  self.animation = animation
  self.curve = curve

  return self
end
