--- AHAttributeAction
-- @author Created by Aaron Barrett on January 15, 2014.

-- Use AHAttributeAction to move the value of a MOAINode's attribute over
-- time.  AHAttributeAction is usually not called directly.  Instead,
-- use common attribute actions (AHScaleTo, AHLocationBy, etc.).  To create
-- attribute actions for custom attributes, use AHAttributeTo or AHAttributeBy.

-- Here is an example of the constructor for a common attribute action:

-- local target = mySprite
-- local duration = 3.75
-- local ease = AHEaseSimpleInOut.new(4.0)
-- local destinationScaleX = 2.0
-- local destinationScaleY = -0.5
-- local scaleAction = AHScaleTo.new(target, duration, ease, destinationScaleX, destinationScaleY)

-- If you were to call `wait(blahAction:start())`, the following would happen:
-- The current coroutine would block execution for 3.75 seconds.  During that time,
-- the sprite named `mySprite` would stretch from its initial scale to a scale
-- of (2.0, -0.5).  The stretching would following a quartic ease--that is, it would
-- start and end slowly, and most of the stretching would happen towards the middle
-- of the animation.

-- Note that the above example uses absolute destination values.  If we wanted to create
-- an action that used relative destination values, (i.e. one that would increase the
-- sprite's initial x-scale by 2.0 and decrease its initial y-scale by 0.5),
-- we would use AHScaleBy rather than AHScaleTo.

-- In the above example, `ease` is an AHEaseSimpleInOut object.  This is just a
-- MOAIEase object.  `ease` could be any MOAIEase (see the eases in AHEase.lua),
-- or it could be `nil`.  If `ease` is `nil`, it means that a linear ease (no distortion) will be used.

-- Note that an AHAttributeAction can modify only a single attribute.  To modify multiple attributes
-- simultaneously (as is the case in AHLocationTo/By and AHColorTo/By, for example), use AHSpawn to
-- combine the effects of multiple AHAttributeActions.

-- The constructor for AHAttributeAction requires a 'deltaFunction'.  This is a function
-- that is called when the action is started to calculate the delta value that will
-- be passed the the MOAIEaseDriver.  `deltaFunction` should have the signature:
-- deltaFunction((number) initialValue, (number) amount)
-- In most circumstances, AttributeTo actions will use AHAttributeTo.absoluteDeltaFunction,
-- and AttributeBy actions will use AHAttributeBy.relativeDeltaFunction.  The notable
-- exception is AHRotationTo, which needs its own delta function to handle the circular
-- (non-linear) nature of angles.

-- ----------------------------------------------------------------------------
-- AHAttributeAction
-- ----------------------------------------------------------------------------

class("AHAttributeAction")

function AHAttributeAction:getTarget()
  return self.target
end

function AHAttributeAction:getAttributeID()
  return self.attributeID
end

function AHAttributeAction:getAmount()
  return self.amount
end

function AHAttributeAction:_AHAttributeAction_getInitialValue(target)
  local attributeID = self.attributeID
  local initialValue = target:getAttr(attributeID)
  return initialValue
end

function AHAttributeAction:start()
  local target = self.target
  local duration = self.duration
  local attributeID = self.attributeID
  local amount = self.amount

  local deltaFunction = self.deltaFunction
  local initialValue = self:_AHAttributeAction_getInitialValue(target)
  -- Ensure the attribute is properly set (to handle non-attribute values via script nodes)
  target:setAttr(attributeID, initialValue)

  local delta = deltaFunction(initialValue, amount)

  local ease = self.ease

  self:setLink(1, target, attributeID, delta, ease)

  return self.super.start(self)
end


-- Constructor ------------------------

--- Creates a new attribute action to modify a node's attribute over time.
-- @param target (MOAINode) - the node whose attribute will be modified.
-- @param duration (number) - the length of the action in seconds.
-- @param attributeID (number) - the ID number of the node's attribute.
-- @param deltaFunction (function) - the function to call to calculate the action's delta.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - The amount to adjust the attribute (the destination value if `deltaFunction` is absolute, or an offset value if `deltaFunction` is relative).
-- @return (MOAIEaseDriver) - A new AHAttributeAction.
function AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  ease = AHEase.wrap(ease)
  assert(type(target) == "userdata", "`target` must be a MOAINode.")
  assert(type(duration) == "number", "`duration` must be a number.")
  assert(duration >= 0.0, "`duration` must not be negative.")
  assert(type(attributeID) == "number", "`attributeID` must be an integer.")
  assert(type(deltaFunction) == "function", "`deltaFunction` must be a function.")
  assert(type(amount) == "number", "`amount` must be a number.")

  local self = MOAIEaseDriver.new()
  AHAction.makeAHAction(self)
  Mixin.mix(self, AHAttributeAction)

  self.actionType = AHActionTypes.ATTRIBUTE_ACTION

  self:reserveLinks(1)

  self.target = target
  self.duration = duration
  self.attributeID = attributeID
  self.deltaFunction = deltaFunction
  self.ease = ease
  self.amount = amount

  -- Cheat: force at least one update to be called.
  if duration == 0 then
    duration = 0.00000001
  end
  self:setSpan(duration)

  return self
end

-- ----------------------------------------------------------------------------
-- AHAttributeTo
-- ----------------------------------------------------------------------------

class("AHAttributeTo")
function AHAttributeTo.absoluteDeltaFunction(initialValue, amount)
  local delta = amount - initialValue
  return delta
end

-- Constructor ------------------------

--- Creates a new (absolute) AHAttributeTo action.
-- @param target (MOAINode) - the node whose attribute will be modified.
-- @param duration (number) - the length of the action in seconds.
-- @param attributeID (number) - the ID number of the node's attribute.
-- @param ease (MOAIEase or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - The absolute destination value for the action.
-- @return (MOAIEaseDriver) - A new AHAttributeTo action.
function AHAttributeTo.new(target, duration, attributeID, ease, amount)
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHAttributeBy
-- ----------------------------------------------------------------------------

class("AHAttributeBy")
function AHAttributeBy.relativeDeltaFunction(initialValue, amount)
  local delta = amount
  return delta
end

function AHAttributeBy.multiplicationDeltaFunction(initialValue, amount)
  local delta = initialValue * amount
  return delta
end

-- Constructor ------------------------

--- Creates a new (relative) AHAttributeBy action.
-- @param target (MOAINode) - the node whose attribute will be modified.
-- @param duration (number) - the length of the action in seconds.
-- @param attributeID (number) - the ID number of the node's attribute.
-- @param ease (MOAIEase or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - The relative offset value for the action.
-- @return (MOAIEaseDriver) - A new AHAttributeBy action.
function AHAttributeBy.new(target, duration, attributeID, ease, amount)
  local deltaFunction = AHAttributeBy.relativeDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end
