--- AHColor actions
-- @author Created by Aaron Barrett on December 26, 2013.

-- Use AHColor to modify a MOAIColor's value (optionally including opacity).
-- Use AHOpacity to modify the opacity alone.

-- Note that unlike MOAIColor:setColor(red, green, blue, (opacity)), which
-- accepts its argument as a list of numbers, AHColor actions accept their
-- color values as a table with the format {red, green, blue, (opacity)}.  You
-- would have to call unpack() on this color table to use it with setColor().

-- ----------------------------------------------------------------------------
-- AHColorAction
-- ----------------------------------------------------------------------------

class("AHColorAction")
-- Constructor ------------------------

function AHColorAction.new(target, duration, deltaFunction, ease, color)
  assert(type(color) == "table", "`color` must be a table.")
  assert((#color == 3) or (#color == 4), "`color` must have the format {red, green, blue, (opacity)}")

  local redAmount, greenAmount, blueAmount, opacityAmount = unpack(color)

  local redChannelAction = AHAttributeAction.new(target, duration, MOAIColor.ATTR_R_COL, deltaFunction, ease, redAmount)
  local greenChannelAction = AHAttributeAction.new(target, duration, MOAIColor.ATTR_G_COL, deltaFunction, ease, greenAmount)
  local blueChannelAction = AHAttributeAction.new(target, duration, MOAIColor.ATTR_B_COL, deltaFunction, ease, blueAmount)

  local opacityAction = nil
  if (opacityAmount) then
    opacityAction = AHAttributeAction.new(target, duration, MOAIColor.ATTR_OPACITY, deltaFunction, ease, opacityAmount)
  end

  local self = AHSpawn.new(redChannelAction, greenChannelAction, blueChannelAction, opacityAction)
  return self
end

-- ----------------------------------------------------------------------------
-- AHColorTo
-- ----------------------------------------------------------------------------

class("AHColorTo")
-- Constructor ------------------------

--- Creates a new absolute attribute action to modify a MOAIColor's value over time.
-- @param target (MOAIColor) - the MOAIColor (often a MOAIProp2D or Sprite) to modify.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param color (table) - A table containing the destination values for each color channel (and optionally opacity). Format: {r, g, b, (o)}
-- @return (MOAICoroutine) - A new AHColorTo action.
function AHColorTo.new(target, duration, ease, color)
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHColorAction.new(target, duration, deltaFunction, ease, color)
  return self
end

-- ----------------------------------------------------------------------------
-- AHColorBy
-- ----------------------------------------------------------------------------

class("AHColorBy")
-- Constructor ------------------------

--- Creates a new relative attribute action to modify a MOAIColor's value over time.
-- @param target (MOAIColor) - the MOAIColor (often a MOAIProp2D or Sprite) to modify.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param color (table) - A table containing the offset values for each color channel (and optionally opacity). Format: {r, g, b, (o)}
-- @return (MOAICoroutine) - A new AHColorBy action.
function AHColorBy.new(target, duration, ease, color)
  local deltaFunction = AHAttributeBy.relativeDeltaFunction
  local self = AHColorAction.new(target, duration, deltaFunction, ease, color)
  return self
end
