--- AHEase
-- @author Created by Aaron Barrett on January 2, 2014.

-- The constructors in this file are convenince methods to initialize
-- MOAIEase objects.  MOAIEase objects can be passed as the `ease` argument
-- in an attribute action to distort the action's rate (See AHAttributeAction.lua).

-- Although it is more efficient to rely on MOAIEase objects that perform their
-- interpolation in C, it is possible to create custom ease objects that use
-- Lua callbacks with AHEaseCustom.  Custom ease functions must take one argument
-- (time), and must be defined between 0.0 and 1.0 (inclusive).  For an example,
-- see AHEaseBounceInOut.

-- Eases have been named to maintain consistency with Cocos2D.  Note that
-- AHEaseSimple is the same as CCEase.  Do not confuse AHEaseSimple with
-- AHEaseExponential.

-- ----------------------------------------------------------------------------
-- AHEaseLinear
-- ----------------------------------------------------------------------------

class("AHEaseLinear")
-- Constructor ------------------------

--- Creates a new linear (no distortion) ease object.
-- @return (MOAIEaseLinear) - The ease object.
function AHEaseLinear.new()
  return MOAIEaseLinear.new()
end

-- ----------------------------------------------------------------------------
-- AHEaseSine
-- ----------------------------------------------------------------------------

class("AHEaseSineIn")
-- Constructor ------------------------

--- Creates a new sinusoidal ease-in object.
-- @return (MOAIEaseSineIn) - The ease object.
function AHEaseSineIn.new()
  return MOAIEaseSineIn.new()
end

class("AHEaseSineOut")
-- Constructor ------------------------

--- Creates a new sinusoidal ease-out object.
-- @return (MOAIEaseSineOut) - The ease object.
function AHEaseSineOut.new()
  return MOAIEaseSineOut.new()
end

class("AHEaseSineInOut")
-- Constructor ------------------------

--- Creates a new sinusoidal ease-in-out object.
-- @return (MOAIEaseSineInOut) - The ease object.
function AHEaseSineInOut.new()
  return MOAIEaseSineInOut.new()
end

-- ----------------------------------------------------------------------------
-- AHEaseExponential
-- ----------------------------------------------------------------------------

class("AHEaseExponentialIn")
-- Constructor ------------------------

--- Creates a new exponential ease-in object.
-- @return (MOAIEaseExponentialIn) - The ease object.
function AHEaseExponentialIn.new()
  return MOAIEaseExponentialIn.new()
end

class("AHEaseExponentialOut")
-- Constructor ------------------------

--- Creates a new exponential ease-out object.
-- @return (MOAIEaseExponentialOut) - The ease object.
function AHEaseExponentialOut.new()
  return MOAIEaseExponentialOut.new()
end

class("AHEaseExponentialInOut")
-- Constructor ------------------------

--- Creates a new exponential ease-in-out object.
-- @return (MOAIEaseExponentialInOut) - The ease object.
function AHEaseExponentialInOut.new()
  return MOAIEaseExponentialInOut.new()
end

-- ----------------------------------------------------------------------------
-- AHEaseSimple
-- ----------------------------------------------------------------------------

class("AHEaseSimpleIn")
-- Constructor ------------------------

--- Creates a new simple (output = t^rate) ease-in object.
-- @param rate (number) - The sharpness of the ease. (default 2.0)
-- @return (MOAIEaseSimpleIn) - The ease object.
function AHEaseSimpleIn.new(rate)
  rate = rate or 2.0

  local self = MOAIEaseSimpleIn.new()
  self:setRate(rate)
  return self
end

class("AHEaseSimpleOut")
-- Constructor ------------------------

--- Creates a new simple (output = t^rate) ease-out object.
-- @param rate (number) - The sharpness of the ease. (default 2.0)
-- @return (MOAIEaseSimpleOut) - The ease object.
function AHEaseSimpleOut.new(rate)
  rate = rate or 2.0

  local self = MOAIEaseSimpleOut.new()
  self:setRate(rate)
  return self
end

class("AHEaseSimpleInOut")
-- Constructor ------------------------

--- Creates a new simple (output = t^rate) ease-in-out object.
-- @param rate (number) - The sharpness of the ease. (default 2.0)
-- @return (MOAIEaseSimpleInOut) - The ease object.
function AHEaseSimpleInOut.new(rate)
  rate = rate or 2.0

  local self = MOAIEaseSimpleInOut.new()
  self:setRate(rate)
  return self
end

-- ----------------------------------------------------------------------------
-- AHEaseBack
-- ----------------------------------------------------------------------------

class("AHEaseBackIn")
-- Constructor ------------------------

--- Creates a new back ease-in object.
-- @param overshoot (number) - How much to back in. (default 1.70158 (a value copied from Cocos2D))
-- @return (MOAIEaseBackIn) - The ease object.
function AHEaseBackIn.new(overshoot)
  overshoot = overshoot or 1.70158

  local self = MOAIEaseBackIn.new()
  self:setOvershoot(overshoot)
  return self
end

class("AHEaseBackOut")
-- Constructor ------------------------

--- Creates a new back ease-out object.
-- @param overshoot (number) - How much to back out. (default 1.70158 (a value copied from Cocos2D))
-- @return (MOAIEaseBackOut) - The ease object.
function AHEaseBackOut.new(overshoot)
  overshoot = overshoot or 1.70158

  local self = MOAIEaseBackOut.new()
  self:setOvershoot(overshoot)
  return self
end

class("AHEaseBackInOut")
-- Constructor ------------------------

--- Creates a new back ease-in-out object.
-- @param overshoot (number) - How much to back in and out. (default 1.70158 (a value copied from Cocos2D))
-- @return (MOAIEaseBackInOut) - The ease object.
function AHEaseBackInOut.new(overshoot)
  overshoot = overshoot or 1.70158

  local self = MOAIEaseBackInOut.new()
  self:setOvershoot(overshoot)
  return self
end

-- ----------------------------------------------------------------------------
-- AHEaseElastic
-- ----------------------------------------------------------------------------

class("AHEaseElasticIn")
-- Constructor ------------------------

--- Creates a new elastic ease-in object.
-- @param period (number) - How long one wobble takes. (default 0.3)
-- @return (MOAIEaseElasticIn) - The ease object.
function AHEaseElasticIn.new(period)
  period = period or 0.3

  local self = MOAIEaseElasticIn.new()
  self:setPeriod(period)
  return self
end

class("AHEaseElasticOut")
-- Constructor ------------------------

--- Creates a new elastic ease-out object.
-- @param period (number) - How long one wobble takes. (default 0.3)
-- @return (MOAIEaseElasticOut) - The ease object.
function AHEaseElasticOut.new(period)
  period = period or 0.3

  local self = MOAIEaseElasticOut.new()
  self:setPeriod(period)
  return self
end

class("AHEaseElasticInOut")
-- Constructor ------------------------

--- Creates a new elastic ease-in-out object.
-- @param period (number) - How long one wobble takes. (default 0.3)
-- @return (MOAIEaseElasticInOut) - The ease object.
function AHEaseElasticInOut.new(period)
  period = period or 0.3

  local self = MOAIEaseElasticInOut.new()
  self:setPeriod(period)
  return self
end

-- ----------------------------------------------------------------------------
-- AHEaseCustom
-- ----------------------------------------------------------------------------

class("AHEaseCustom")
require("utils.CurveUtils")

-- Constructor ------------------------

--- Creates a new ease object that uses `easeFunction` for interpolation.
-- @param easeFunction (function) - A function defined for one argument, time, from 0.0 to 1.0 (inclusive)
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseCustom.new(easeFunction)
  local self = MOAIEaseCustom.new()
  self:setFunction(easeFunction)
  return self
end

--- Create a custom ease with the specified curve.
-- @param curve (table) - the curve definition, containing:
--                        * points (table) - Array of ordered pairs
--                        * params (table) - Table containing parameters
-- params:
-- * tension (number) - The tension of the resulting spline.
function AHEaseCustom.newWithCurve(curve)
  local points = curve.points
  local params = curve.params or {}
  local tension = params.tension or 0
  local v0, pointInfo = CurveUtils.cardinalSplineForPointsAtTime(points, 0, tension)

  return AHEaseCustom.new(function(t)
    return CurveUtils.cardinalSplineForPointsAtTime(points, t, tension, pointInfo)
  end)
end

-- ----------------------------------------------------------------------------
-- AHEaseBezier
-- ----------------------------------------------------------------------------

class("AHEaseBezier")

--- Create a custom ease with curve specified by Bezier curve control points
-- @param controlPoints (table) - has {x = a, y = b} values for 1st control point
--                                and {x = c, y = d} values for 2nd control point
function AHEaseBezier.new(controlPoints)
  local self = MOAIEaseCustom.new()
  local easeFunction = CurveUtils.getBezierEaseFunction(controlPoints)
  self:setFunction(easeFunction)
  return self
end

--- Same as the above but with the control points in a flat table
function AHEaseBezier.newFlat(controlPoints)
  local formattedControlPoints = {
    {
      x = controlPoints[1],
      y = controlPoints[2]
    },
    {
      x = controlPoints[3],
      y = controlPoints[4]
    }
  }
  return AHEaseBezier.new(formattedControlPoints)
end

function AHEaseBezier.newAppleDefault()
  return AHEaseBezier.newFlat({0.25, 0.1, 0.25, 1.0})
end

-- ----------------------------------------------------------------------------
-- AHEaseBounce
-- ----------------------------------------------------------------------------

class("AHEaseBounceCommon")
function AHEaseBounceCommon.bounce(t)
  if (t < 1 / 2.75) then
    return 7.5625 * t * t
  elseif (t < 2 / 2.75) then
    t = t - 1.5 / 2.75
    return 7.5625 * t * t + 0.75
  elseif (t < 2.5 / 2.75) then
    t = t - 2.25 / 2.75
    return 7.5625 * t * t + 0.9375
  end
  t = t - 2.625 / 2.75
  return 7.5625 * t * t + 0.984375
end

class("AHEaseBounceIn")
-- Constructor ------------------------

--- Creates a new bounce ease-in object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseBounceIn.new()
  local function bounceIn(t)
    return 1.0 - AHEaseBounceCommon.bounce(1.0 - t)
  end

  local self = AHEaseCustom.new(bounceIn)

  return self
end

class("AHEaseBounceOut")
-- Constructor ------------------------

--- Creates a new bounce ease-out object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseBounceOut.new()
  local function bounceOut(t)
    return AHEaseBounceCommon.bounce(t)
  end

  local self = AHEaseCustom.new(bounceOut)

  return self
end

class("AHEaseBounceInOut")
-- Constructor ------------------------

--- Creates a new bounce ease-in-out object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseBounceInOut.new()
  local function bounceInOut(t)
    if (t < 0.5) then
      t = t * 2.0
      return (1.0 - AHEaseBounceCommon.bounce(1.0 - t)) * 0.5
    end
    return AHEaseBounceCommon.bounce(t * 2.0 - 1.0) * 0.5 + 0.5
  end

  local self = AHEaseCustom.new(bounceInOut)

  return self
end

-- ----------------------------------------------------------------------------
-- AHEaseQuadratic
-- ----------------------------------------------------------------------------

class("AHEaseQuadraticCommon")
function AHEaseQuadraticCommon.easeIn(t)
    return t * t
end

function AHEaseQuadraticCommon.easeOut(t)
    return -t * (t - 2)
end

class("AHEaseQuadraticIn")
-- Constructor ------------------------

--- Creates a new quadratic ease-in object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseQuadraticIn.new()
  return AHEaseCustom.new(AHEaseQuadraticCommon.easeIn)
end

class("AHEaseQuadraticOut")
-- Constructor ------------------------

--- Creates a new quadratic ease-out object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseQuadraticOut.new()
  return AHEaseCustom.new(AHEaseQuadraticCommon.easeOut)
end

class("AHEaseQuadraticInOut")
-- Constructor ------------------------

--- Creates a new quadratic ease-in-out object.
-- @return (MOAIEaseCustom) - The ease object.
function AHEaseQuadraticInOut.new()
  return AHEaseCustom.new(function(t)
    if t < 0.5 then
      return AHEaseQuadraticCommon.easeIn(t) * 2
    else
      return AHEaseQuadraticCommon.easeOut(t) * 2 - 1;
    end
  end)
end

-- ----------------------------------------------------------------------------
-- AHEase misc functions
-- ----------------------------------------------------------------------------

class("AHEase")
--- Wrap an unknown object into an ease, if needed.
-- Supports nil (linear), AHEase, MOAIEase, functions, and curves contained in a table.
-- A function must accept a normalized time value (0.0 - 1.0) and return a value.
-- A curve is a table as per AHEaseCustom.newWithCurve().
function AHEase.wrap(object)
  local objectType = type(object)
  if object == nil then
    return AHEaseLinear.new()
  elseif objectType == "function" then
    return AHEaseCustom.new(object)
  elseif objectType == "table" then
    return AHEaseCustom.newWithCurve(object)
  elseif objectType == "userdata" then
    return object
  elseif objectType == "number" then
    return object
  end

  error("Expected a MOAIEase, curve, or ease function. Got " .. objectType)
end
