--- AHLocation actions
-- @author Created by Aaron Barrett on December 26, 2013.

-- Use AHLocation to modify a MOAIProp2D's location values.

-- ----------------------------------------------------------------------------
-- AHLocationXTo
-- ----------------------------------------------------------------------------

class("AHLocationXTo")
--- Creates a new absolute attribute action to modify a prop's X coordinate value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the destination X coordinate.
-- @return (MOAIEaseDriver) - A new AHLocationXTo action.
function AHLocationXTo.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_X_LOC
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHLocationXBy
-- ----------------------------------------------------------------------------

class("AHLocationXBy")
--- Creates a new relative attribute action to modify a prop's X coordinate value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the amount to offset the X coordinate.
-- @return (MOAIEaseDriver) - A new AHLocationXBy action.
function AHLocationXBy.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_X_LOC
  local deltaFunction = AHAttributeBy.relativeDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHLocationYTo
-- ----------------------------------------------------------------------------

class("AHLocationYTo")
--- Creates a new absolute attribute action to modify a prop's Y coordinate value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the destination Y coordinate.
-- @return (MOAIEaseDriver) - A new AHLocationYTo action.
function AHLocationYTo.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Y_LOC
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHLocationYBy
-- ----------------------------------------------------------------------------

class("AHLocationYBy")
--- Creates a new relative attribute action to modify a prop's Y coordinate value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the amount to offset the Y coordinate.
-- @return (MOAIEaseDriver) - A new AHLocationYBy action.
function AHLocationYBy.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Y_LOC
  local deltaFunction = AHAttributeBy.relativeDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHLocationTo
-- ----------------------------------------------------------------------------

class("AHLocationTo")
--- Creates a new absolute attribute action to modify a prop's location over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param xAmount (number) - the destination X coordinate.
-- @param yAmount (number) - the destination Y coordinate.
-- @return (MOAICoroutine) - A new AHLocationTo action.
function AHLocationTo.new(target, duration, ease, xAmount, yAmount)
  assert(type(yAmount) == "number", "`yAmount` must be a number.")

  local xAction = AHLocationXTo.new(target, duration, ease, xAmount)
  local yAction = AHLocationYTo.new(target, duration, ease, yAmount)
  local self = AHSpawn.new(xAction, yAction)
  return self
end

-- ----------------------------------------------------------------------------
-- AHLocationBy
-- ----------------------------------------------------------------------------

class("AHLocationBy")
--- Creates a new relative attribute action to modify a prop's location over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to move.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param xAmount (number) - the amount to offset the X coordinate.
-- @param yAmount (number) - the amount to offset the Y coordinate.
-- @return (MOAICoroutine) - A new AHLocationBy action.
function AHLocationBy.new(target, duration, ease, xAmount, yAmount)
  local xAction = AHLocationXBy.new(target, duration, ease, xAmount)
  local yAction = AHLocationYBy.new(target, duration, ease, yAmount)
  local self = AHSpawn.new(xAction, yAction)
  return self
end
