--- AHLoopForDuration
--
-- Loops an action, coroutine, or function for a specific amount of time.
-- If you pass in an action, it will loop that action until duration is reached.
-- If you pass in a function or coroutine, it will repeatedly call it until duration is reached.
-- Upon reaching the end of the duration, the action is stopped, even if it's in the middle of
-- doing something.
--
--
-- Examples:
--
--  -- Sprite bounces up and down for 2 seconds and then stops.
--  local action = AHSequence.new(
--    AHLocationYBy.new(sprite, 0.3, AHEaseSineInOut.new(), 50),
--    AHLocationYBy.new(sprite, 0.3, AHEaseSineInOut.new(), -50)
--  )
--  AHLoopForDuration.new(2.0, action):start()
--
--  -- Prints "a" repeatedly for 0.5 seconds
--  local action = function()
--    print("a")
--  end
--  AHLoopForDuration.new(0.5, action):start()
--
--  -- Runs the coroutine and stops it after 1.5 seconds.
--  -- Note that it will only print "end" once in this case since it gets stopped.
--  local action = MOAICoroutine.new()
--  action:setFunction(function()
--    print("start")
--    waitFor(1)
--    print("end")
--  end)
--  AHLoopForDuration.new(1.5, action):start()


class("AHLoopForDuration")

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Construct a new AHLoopForDuration
-- @param duration (number) - Duration of this action.
-- @param action (AHAction, MOAICoroutine, or function) - action to run (see description at top of file)
-- @return (AHLoopForDuration) - A new action.
function AHLoopForDuration.new(duration, action)
  action = AHCallLogic.wrapIfIsFunction(action)
  return AHRunForDuration.new(duration, AHLoop.new(action))
end
