--- AHNonAttributeAction
--
-- Works similar to AHAttributeAction except that instead of accessing a prop's
-- attributes directly, it makes use of user-supplied getters and setters.


class("AHNonAttributeAction")

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

-- Attributes -------------------------

local ATTR_ID = 0


------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function makeScriptNode(target, getterFunction, setterFunction)
    local scriptNode = MOAIScriptNode.new()
    scriptNode:reserveAttrs(1)
    scriptNode:setCallback(function(node)
      setterFunction(target, node:getAttr(ATTR_ID))
    end)

    scriptNode._AHNonAttributeAction_getInitialValue = function()
      return getterFunction(target)
    end

    return scriptNode
end

function AHNonAttributeAction:_AHAttributeAction_getInitialValue(target)
  return target._AHNonAttributeAction_getInitialValue()
end

local function makeNonAttributeAction(actionClass, target, duration, getterFunction, setterFunction, ease, amount)
  assert(type(target) == "userdata" or type(target) == "table", "`target` must be a MOAINode or a table.")
  assert(type(getterFunction) == "function", "`getterFunction` must be a function.")
  assert(type(setterFunction) == "function", "`setterFunction` must be a function.")
  local scriptNode = makeScriptNode(target, getterFunction, setterFunction)
  local self = actionClass.new(scriptNode, duration, ATTR_ID, ease, amount)
  Mixin.mix(self, AHNonAttributeAction)

  -- HACK: We have hit bugs where the script node is accessed after it has been garbage collected,
  -- causing a crash in MOAI. To avoid that, we store a reference to the script node. Storing it on
  -- the action itself does not fix it in cases where the action has also been garbage collected,
  -- so we store it on the target.
  target.__AHNonAttributeAction_scriptNodes = target.__AHNonAttributeAction_scriptNodes or {}
  table.insert(target.__AHNonAttributeAction_scriptNodes, scriptNode)

  return self
end

------------------------------------------------------------------------------
-- AHNonAttributeTo
------------------------------------------------------------------------------

class("AHNonAttributeTo")
-- Constructor ------------------------

--- Creates a new (absolute) AHAttributeTo action.
-- @param target (MOAINode) - the node whose attribute will be modified.
-- @param duration (number) - the length of the action in seconds.
-- @param getterFunction (function) - Gets the value in the target. Params: (self), returns number
-- @param setterFunction (function) - Sets the value in the target. Params: (self, value)
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - The absolute destination value for the action.
-- @return (MOAIEaseDriver) - A new AHNonAttributeTo action.
function AHNonAttributeTo.new(target, duration, getterFunction, setterFunction, ease, amount)
  return makeNonAttributeAction(AHAttributeTo, target, duration, getterFunction, setterFunction, ease, amount)
end

------------------------------------------------------------------------------
-- AHNonAttributeBy
------------------------------------------------------------------------------

class("AHNonAttributeBy")
-- Constructor ------------------------

--- Creates a new (relative) AHAttributeBy action.
-- @param target (MOAINode) - the node whose attribute will be modified.
-- @param duration (number) - the length of the action in seconds.
-- @param getterFunction (function) - Gets the value in the target. Params: (self), returns number
-- @param setterFunction (function) - Sets the value in the target. Params: (self, value)
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - The relative offset value for the action.
-- @return (MOAIEaseDriver) - A new AHNonAttributeBy action.
function AHNonAttributeBy.new(target, duration, getterFunction, setterFunction, ease, amount)
  return makeNonAttributeAction(AHAttributeBy, target, duration, getterFunction, setterFunction, ease, amount)
end
