--- AHRotation actions
-- @author Created by Aaron Barrett on December 26, 2013.

-- Use AHRotation to modify a MOAIProp2D's rotation value.

-- Note that AHRotationTo behaves similarly to CCRotateTo from Cocos2D.
-- That is--the destination value of AHRotationTo (but *not* of AHRotationBy)
-- will be "wrapped".  For example, if a prop has an initial rotation of 350
-- degrees, then an AHRotationTo with a destination rotation of 10 degrees will
-- rotate the prop only 20 degrees, rather than 340 degrees.

-- ----------------------------------------------------------------------------
-- AHRotationTo
-- ----------------------------------------------------------------------------

class("AHRotationTo")
local function convertAngleToStandardRange(angle)
  local wrappedAngle = angle % 360.0
  if (wrappedAngle > 180.0) then
    return wrappedAngle - 360.0
  elseif (wrappedAngle <= -180.0) then
    return -wrappedAngle
  end
  return wrappedAngle
end

local function rotationAbsoluteDeltaFunction(initialAngle, amount)
  local delta = nil

  local wrappedInitialAngle = convertAngleToStandardRange(initialAngle)
  local wrappedDestinationAngle = convertAngleToStandardRange(amount)

  local differenceBetweenWrappedAngles = wrappedDestinationAngle - wrappedInitialAngle
  local isDistanceOverHalfRotation = (math.abs(differenceBetweenWrappedAngles) > 180.0)
  if (isDistanceOverHalfRotation) then

    local isShortestPathClockwise = (wrappedDestinationAngle > 0.0)
    if (isShortestPathClockwise) then
      delta = differenceBetweenWrappedAngles - 360.0
    else
      delta = 360.0 - (wrappedInitialAngle - wrappedDestinationAngle)
    end

  else

    delta = differenceBetweenWrappedAngles

  end

  return delta
end

-- Constructor ------------------------

--- Creates a new absolute attribute action to modify a MOAIProp2D's rotation value over time.
-- @param target (MOAIProp2D) - the prop to rotate.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the destination rotation value.
-- @return (MOAIEaseDriver) - A new AHRotationTo action.
function AHRotationTo.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Z_ROT
  local deltaFunction = rotationAbsoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHRotationBy
-- ----------------------------------------------------------------------------

class("AHRotationBy")
-- Constructor ------------------------

--- Creates a new relative attribute action to modify a MOAIProp2D's rotation value over time.
-- @param target (MOAIProp2D) - the prop to rotate.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the amount to offset the rotation value.
-- @return (MOAIEaseDriver) - A new AHRotationBy action.
function AHRotationBy.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Z_ROT
  local deltaFunction = AHAttributeBy.relativeDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end
