--- AHScale actions
-- @author Created by Aaron Barrett on December 26, 2013.

-- Use AHScale to modify a MOAIProp2D's scale values.

-- ----------------------------------------------------------------------------
-- AHScaleXTo
-- ----------------------------------------------------------------------------

class("AHScaleXTo")
-- Constructor ------------------------

--- Creates a new absolute attribute action to modify a prop's X scale value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the destination X scale.
-- @return (MOAIEaseDriver) - A new AHScaleXTo action.
function AHScaleXTo.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_X_SCL
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHScaleXBy
-- ----------------------------------------------------------------------------

class("AHScaleXBy")
-- Constructor ------------------------

--- Creates a new relative attribute action to modify a prop's X scale value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the amount to offset the X scale.
-- @return (MOAIEaseDriver) - A new AHScaleXBy action.
function AHScaleXBy.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_X_SCL
  local deltaFunction = AHAttributeBy.multiplicationDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHScaleYTo
-- ----------------------------------------------------------------------------

class("AHScaleYTo")
-- Constructor ------------------------

--- Creates a new absolute attribute action to modify a prop's Y scale value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the destination Y scale.
-- @return (MOAIEaseDriver) - A new AHScaleYTo action.
function AHScaleYTo.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Y_SCL
  local deltaFunction = AHAttributeTo.absoluteDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHScaleYBy
-- ----------------------------------------------------------------------------

class("AHScaleYBy")
-- Constructor ------------------------

--- Creates a new relative attribute action to modify a prop's Y scale value over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param amount (number) - the amount to offset the Y scale.
-- @return (MOAIEaseDriver) - A new AHScaleYBy action.
function AHScaleYBy.new(target, duration, ease, amount)
  local attributeID = MOAITransform2D.ATTR_Y_SCL
  local deltaFunction = AHAttributeBy.multiplicationDeltaFunction
  local self = AHAttributeAction.new(target, duration, attributeID, deltaFunction, ease, amount)
  return self
end

-- ----------------------------------------------------------------------------
-- AHScaleTo
-- ----------------------------------------------------------------------------

class("AHScaleTo")
-- Constructor ------------------------

--- Creates a new absolute attribute action to modify a prop's scale over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param xAmount (number) - the destination X scale.
-- @param yAmount (number or `nil`) - the destination Y scale (will use `xAmount` if `nil` is passed).
-- @return (MOAICoroutine) - A new AHScaleTo action.
function AHScaleTo.new(target, duration, ease, xAmount, yAmount)
  yAmount = yAmount or xAmount

  local xAction = AHScaleXTo.new(target, duration, ease, xAmount)
  local yAction = AHScaleYTo.new(target, duration, ease, yAmount)
  local self = AHSpawn.new(xAction, yAction)
  return self
end

-- ----------------------------------------------------------------------------
-- AHScaleBy
-- ----------------------------------------------------------------------------

class("AHScaleBy")
-- Constructor ------------------------

--- Creates a new relative attribute action to modify a prop's scale over time.
-- @param target (MOAIProp2D) - the MOAIProp2D to scale.
-- @param duration (number) - the length of the action in seconds.
-- @param ease (MOAIEase, function, curve, or `nil`) - The AHEase object to use (or `nil`, for a linear ease).
-- @param xAmount (number) - the amount to offset the X scale.
-- @param yAmount (number or `nil`) - the amount to offset the Y scale (will use `xAmount` if `nil` is passed).
-- @return (MOAICoroutine) - A new AHScaleBy action.
function AHScaleBy.new(target, duration, ease, xAmount, yAmount)
  yAmount = yAmount or xAmount

  local xAction = AHScaleXBy.new(target, duration, ease, xAmount)
  local yAction = AHScaleYBy.new(target, duration, ease, yAmount)
  local self = AHSpawn.new(xAction, yAction)
  return self
end
