--- AHSpawn
-- @author Created by Aaron Barrett on December 26, 2013.

-- Use AHSpawn to play multiple AHActions that can be started
-- and stopped simultaneously as if they were a single action.

class("AHSequenceSpawn")

-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function internalAddActions(self, actions)
  for _, action in ipairs(actions) do
    if type(action) == "table" then
      internalAddActions(self, action)
    else
      self:addAction(action)
    end
  end
end


-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

--- Gets the array of AHActions that this sequence or spawn will play.
-- @return (table) - the array.
function AHSequenceSpawn:getChildActions()
  return self.childActions
end

--- Adds an additional child action or function to this sequence or spawn.
-- It can only be used before the seuence or spawn has been started.
-- @param action (AHAction or function) - the child action, function, or coroutine.
function AHSequenceSpawn:addAction(action)
  action = AHCallLogic.wrapIfIsFunction(action)

  assert(type(action) == "userdata", "`action` must be a MOAIAction or a function (got " .. type(action) .. ").")
  assert(not action:isActive(), "Cannot add actions that have already started.")
  assert(not self:isActive(), "Cannot add actions afer parent action has started.")

  table.insert(self.childActions, action)
end

--- Add multiple actions.
-- It can only be used before the seuence or spawn has been started.
-- If you pass in a list of arrays, they will be unpacked and their contents added.
function AHSequenceSpawn:addActions(...)
  internalAddActions(self, arg)
end


-- ---------------------------------------------------------------------------
-- Object creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function AHSequenceSpawn.new(...)
  local self = MOAICoroutine.new()
  AHAction.makeAHAction(self)
  Mixin.mix(self, AHSequenceSpawn)

  self.childActions = {}

  internalAddActions(self, arg)

  return self
end



-- ===========================================================================
-- AHSpawn
-- ===========================================================================

class("AHSpawn")
-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function AHSpawn:stop()
  for _, action in ipairs(self.childActions) do
    action:stop()
  end
  return self.super.stop(self)
end

function AHSpawn:pause(isPaused)
  if(isPaused == nil) then isPaused = true end
  for _, action in ipairs(self.childActions) do
    action:pause(isPaused)
  end
  return self.super.pause(self, isPaused)
end

-- ---------------------------------------------------------------------------
-- Object creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new AHSpawn action.
-- @param ... - the list of actions, functions, and coroutines that will be played.
--              You can also provide arrays, which will be unpacked and their contents added.
-- @return (MOAICoroutine) - A new AHSpawn action.
function AHSpawn.new(...)
  local self = AHSequenceSpawn.new(...)
  Mixin.mix(self, AHSpawn)

  self.actionType = AHActionTypes.SPAWN

  local function runAllChildrenAtOnce()
    for _, action in ipairs(self.childActions) do
      action:start()
    end
    wait(unpack(self.childActions))
  end

  self:setFunction(runAllChildrenAtOnce)

  return self
end



-- ===========================================================================
-- AHStaggeredSpawn
-- ===========================================================================

class("AHStaggeredSpawn")
-- ---------------------------------------------------------------------------
-- Object creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new AHStaggeredSpawn action, delays between starting child actions.
-- @param staggerDuration (number) - How long to stagger the spawn actions.
-- @param ... - the list of actions, functions, and coroutines that will be played.
--              You can also provide arrays, which will be unpacked and their contents added.
-- @return (MOAICoroutine) - A new AHSpawn action.
function AHStaggeredSpawn.new(staggerDuration, ...)
  local self = AHSpawn.new(...)

  self:setFunction(function()
    for _, action in ipairs(self.childActions) do
      action:start()
      waitFor(staggerDuration)
    end
    wait(unpack(self.childActions))
  end)

  return self
end



-- ===========================================================================
-- AHSequence
-- ===========================================================================

class("AHSequence")
-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function AHSequence:stop()
  if (self.runningAction) then
    self.runningAction:stop()
  end
  return self.super.stop(self)
end

function AHSequence:pause(isPaused)
  if(isPaused == nil) then isPaused = true end
  if (self.runningAction) then
    self.runningAction:pause(isPaused)
  end
  return self.super.pause(self, isPaused)
end


-- ---------------------------------------------------------------------------
-- Object creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new AHSequence action.
-- @param ... - the list of actions, functions, and coroutines that will be played.
--              You can also provide arrays, which will be unpacked and their contents added.
-- @return (MOAICoroutine) - A new AHSequence action.
function AHSequence.new(...)
  local self = AHSequenceSpawn.new(...)
  Mixin.mix(self, AHSequence)

  self.actionType = AHActionTypes.SEQUENCE

  local function runAllChildrenInSequence()
    for _, action in ipairs(self.childActions) do
      self.runningAction = action
      action:start()
      wait(action)
    end
    self.runningAction = nil
  end

  self:setFunction(runAllChildrenInSequence)

  return self
end
