--- AHUntimedAction
-- @author Created by Aaron Barrett on April 29, 2015.

class("AHUntimedAction")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function calculateDuration(durationsByAction, action)
  local actionType = action.actionType
  if ((actionType == AHActionTypes.ATTRIBUTE_ACTION) or
      (actionType == AHActionTypes.DELAY)) then
    durationsByAction[action] = action.duration
    return action.duration
  elseif (actionType == AHActionTypes.SPAWN) then
    local spawnDuration = 0.0
    for _, childAction in ipairs(action.childActions) do
      spawnDuration = math.max(spawnDuration, calculateDuration(durationsByAction, childAction))
    end
    durationsByAction[action] = spawnDuration
    return spawnDuration
  elseif (actionType == AHActionTypes.SEQUENCE) then
    local sequenceDuration = 0.0
    for _, childAction in ipairs(action.childActions) do
      sequenceDuration = sequenceDuration + calculateDuration(durationsByAction, childAction)
    end
    durationsByAction[action] = sequenceDuration
    return sequenceDuration
  else
    error("No way to calculate duration defined for `actionType`:" .. tostring(actionType))
  end
end

local function initializeActions(self, action)
  local actionType = action.actionType
  if (actionType == AHActionTypes.ATTRIBUTE_ACTION) then
    local initialValuesByAction = self.initialValuesByAction
    local rangesByAction = self.rangesByAction

    local target = action.target
    local attributeID = action.attributeID
    local deltaFunction = action.deltaFunction
    local amount = action.amount

    local initialValue = target:getAttr(attributeID)
    initialValuesByAction[action] = initialValue

    local range = deltaFunction(initialValue, amount)
    rangesByAction[action] = range
  elseif (actionType == AHActionTypes.DELAY) then
    return
  elseif ((actionType == AHActionTypes.SPAWN) or
          (actionType == AHActionTypes.SEQUENCE)) then
    local childActions = action.childActions
    for _, childAction in ipairs(childActions) do
      initializeActions(self, childAction)
    end
  else
    error("No way to initialize action defined for `actionType`:" .. tostring(actionType))
  end
end

local function applyInitialValues(self, action)
  local actionType = action.actionType

  if (actionType == AHActionTypes.ATTRIBUTE_ACTION) then
    local initialValue = self.initialValuesByAction[action]
    local target = action.target
    local attributeID = action.attributeID
    target:setAttr(attributeID, initialValue)
  elseif (actionType == AHActionTypes.DELAY) then
    return
  elseif ((actionType == AHActionTypes.SPAWN) or
          (actionType == AHActionTypes.SEQUENCE)) then
    local childActions = action.childActions
    for _, childAction in ipairs(childActions) do
      applyInitialValues(self, childAction)
    end
  else
    error("No way to apply initial values defined for `actionType`:" .. tostring(actionType))
  end
end

local function applyAtProgress(self, action, progress)
  local actionType = action.actionType

  if (actionType == AHActionTypes.ATTRIBUTE_ACTION) then
    local initialValue = self.initialValuesByAction[action]
    local range = self.rangesByAction[action]

    local ease = action.ease
    local easedProgress = nil
    if (ease ~= nil) then
      easedProgress = ease:getValueAtTime(progress)
    else
      easedProgress = progress
    end

    local target = action.target
    local attributeID = action.attributeID

    local valueToApply = initialValue + easedProgress * range
    target:setAttr(attributeID, valueToApply)
  elseif (actionType == AHActionTypes.DELAY) then
    return
  elseif (actionType == AHActionTypes.SPAWN) then
    local spawnDuration = self.durationsByAction[action]
    local childActions = action.childActions
    for _, childAction in ipairs(childActions) do
      local childActionDuration = self.durationsByAction[childAction]
      local progressOnChild = progress / (childActionDuration / spawnDuration)

      if (progressOnChild < 1.0) then
        applyAtProgress(self, childAction, progressOnChild)
      else
        applyAtProgress(self, childAction, 1.0)
      end
    end
  elseif (actionType == AHActionTypes.SEQUENCE) then
    local sequenceDuration = self.durationsByAction[action]

    local childActions = action.childActions
    local consumedProgress = 0.0
    for _, childAction in ipairs(childActions) do
      local childActionDuration = self.durationsByAction[childAction]
      local progressConsumedByChild = childActionDuration / sequenceDuration
      if (consumedProgress + progressConsumedByChild < progress) then
        applyAtProgress(self, childAction, 1.0)
      else
        local progressOnChild = (progress - consumedProgress) / progressConsumedByChild
        applyAtProgress(self, childAction, progressOnChild)
        break
      end
      consumedProgress = consumedProgress + progressConsumedByChild
    end
  else
    error("No way to apply action defined for `actionType`:" .. tostring(actionType))
  end
  if action.target then
    runInCoroutine(function()
      -- Force an update because MOAI doesn't always send an update at the appropriate
      -- times when doing an untimed action.
      action.target:forceUpdate()
    end)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function AHUntimedAction:start()
  if not self.actionIsRunning then
    local actionTree = self.actionTree
    local durationsByAction = self.durationsByAction
    calculateDuration(durationsByAction, actionTree)
    initializeActions(self, actionTree)
    self.actionIsRunning = true
  end
end

function AHUntimedAction:applyAtTime(time)
  self:start()
  local actionTree = self.actionTree

  applyInitialValues(self, actionTree)

  local duration = self.durationsByAction[actionTree]
  local progress = time / duration
  applyAtProgress(self, actionTree, progress)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function AHUntimedAction.new(actionTree)
  local self = {}
  Mixin.mix(self, AHUntimedAction)

  self.actionTree = actionTree
  self.durationsByAction = {}
  self.initialValuesByAction = {}
  self.rangesByAction = {}

  return self
end
