--- ActionQueue
-- A synchronization tool that queues up actions to be run sequentially one
-- after the other. If the queue is full, new actions will not be accepted until
-- an existing action finishes. Think of it as a kind of action semaphore that
-- drains as the actions finish.
-- Accepts MOAIActions, AHActions, MOAICoroutines, Timers, and raw functions.
-- When an action reaches the front of the queue, it will be started immediately
-- (if it's not already running or done).
--
-- Note: Do not manually pause individual actions in the queue as they will
-- prevent further actions from being started!
--
-- ActionQueue's API is a subset of MOAIAction's API.
--
-- @author Created by karl on July 03, 2014.

class("ActionQueue")
require("actions.AHCallLogic")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function pruneActions(self)
  local actions = {}
  for _, action in ipairs(self.actions) do
    local hasFinished = action._ActionQueue_hasStarted and not action:isActive()
    if not hasFinished then
      table.insert(actions, action)
    end
  end
  self.actions = actions
end

local function startPoller(self)
  if not self.poller then
    self.poller = runInCoroutine(function()
      pruneActions(self)
      while #self.actions > 0 do
        local action = self.actions[1]
        if not action._ActionQueue_hasStarted then
          action._ActionQueue_hasStarted = true
          action:start()
        end
        MOAICoroutine.blockOnAction(action)
        pruneActions(self)
      end
      self.poller = nil
    end)
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Queue up action(s).
-- The action(s) will only be queued if the queue isn't full.
-- If an action is not already running, it will be started once it reaches the
-- head of the queue.
-- @param ... (list of [action, coroutine, timer, function]) - Action(s) to queue.
-- @return (boolean) - True if at least one action was inserted to the queue.
function ActionQueue:insert(...)
  local didInsert = false
  for _, action in ipairs(arg) do
    local canInsert = #self.actions < self.maxQueueSize
    if not canInsert then
      break
    end
    action = AHCallLogic.wrapIfIsFunction(action)
    action._ActionQueue_hasStarted = action:isActive()
    table.insert(self.actions, action)
    didInsert = true
  end

  if didInsert then
    startPoller(self)
  end

  return didInsert
end

--- Checks to see if the current action (if any) is currently in the action tree.
-- @return (boolean) - true if the current action is active, or false if there is no current action.
function ActionQueue:isActive()
  local action = self.actions[1]
  if action then
    return action:isActive()
  end
  return false
end

--- Checks to see if the current action (if any) is currently busy.
-- An action is 'busy' only if it is 'active' and not 'done.'.
-- @return (boolean) - true if the current action is busy, or false if there is no current action.
function ActionQueue:isBusy()
  local action = self.actions[1]
  if action then
    return action:isBusy()
  end
  return false
end

--- Checks to see if the queue is 'done.'
-- Definition of 'done' is up to individual action implementations.
-- For an action queue, it means that the queue is empty.
-- @return (boolean) - true if the queue is done.
function ActionQueue:isDone()
  return #self.actions == 0
end

--- Leaves the current action (if any) and queue in the action tree but
-- prevents them from receiving updates.
-- Call pause(false) or start() to unpause.
function ActionQueue:pause(isPaused)
  local action = self.actions[1]
  if action then
    action:pause(isPaused)
  end
  if not self.poller then
    startPoller(self)
  end
  self.poller:pause(isPaused)
end

--- Adds the current action (if any) and queue to the root of the action tree.
function ActionQueue:start()
  if self.poller then
    self:pause(false)
  else
    startPoller(self)
  end
end

--- Removes the current action (if any) and queue from its parent action.
-- The queue will stop processing actions until you call start() or
-- queue another action.
function ActionQueue:stop()
  if self.poller then
    self.poller:stop()
    self.poller = nil
  end
  local action = self.actions[1]
  if action then
    action:stop()
  end
end

--- Wait on any running actions.
function ActionQueue:wait()
  while self.poller do
    coroutine.yield()
  end
end

--- Removes all child actions.
-- Note: Does not stop the currently running action! Stop the queue BEFORE
-- calling this if you want the current action to end as well.
function ActionQueue:clear()
  self.actions = {}
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new action queue.
-- @param maxQueueSize (number) - The max number of actions that may exist in
--                                the queue.
function ActionQueue.new(maxQueueSize)
  maxQueueSize = maxQueueSize or math.huge
  local self = {}
  Mixin.mix(self, ActionQueue)

  self.maxQueueSize = maxQueueSize
  self.actions = {}

  return self
end



-- ===========================================================================
-- ActionMutex
-- ===========================================================================

class("ActionMutex")
-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make an action mutex (queue size 1).
function ActionMutex.new()
  return ActionQueue.new(1)
end



-- ===========================================================================
-- ActionInterrupt
-- ===========================================================================

--- ActionInterrupt. Runs only one action at a time, interrupting the currently
-- running action when a new action is started.
class("ActionInterrupt")

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Insert a new action.
-- Note: Since ActionInterrupt only allows a queue size of 1, calling this
--       method with multiple parameters will effectively run only the LAST
--       action in the parameter list.
function ActionInterrupt:insert(...)
  self:stop()
  self:clear()
  local action = arg[#arg]
  return self.super.insert(self, action)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make an action interrupt (queue size 1, interrupts the running action).
function ActionInterrupt.new()
  local self = ActionQueue.new(1)
  Mixin.mix(self, ActionInterrupt)

  return self
end
