
class("BonusNotificationTray")
require('bonuses.SimpleTextBonusIndicator')
require('bonuses.StreakIndicator')

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local NOTIFICATION_PADDING_Y = -2
local NOTIFICATION_HEIGHT = 18

local GROUP_WIDTH = Screen.width * 0.3
local MAX_NOTIFICATIONS = 3
local GROUP_HEIGHT = MAX_NOTIFICATIONS * NOTIFICATION_HEIGHT +
                     (MAX_NOTIFICATIONS - 1) * NOTIFICATION_PADDING_Y

local NOTIFICATION_SLIDE_DURATION = 0.3
local NOTIFICATION_FADE_DURATION = 0.3
local DISPLAY_INDICATOR_DURATION = 1.75

local DEFAULT_COLOR = {1.0, 1.0, 1.0, 1.0}
local DEFAULT_OPACITY = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getKeyOpacity(self)
  return self.keyColor[4]
end

local function repositionAllNotifications(self)
  for i, exisitingNotification in ipairs(self.currentNotifications) do
    local notificationY = GROUP_HEIGHT - NOTIFICATION_HEIGHT * 0.5 -
                          (i - 1) * (NOTIFICATION_HEIGHT + NOTIFICATION_PADDING_Y)

    local moveAction = AHLocationYTo.new(exisitingNotification, NOTIFICATION_SLIDE_DURATION, nil, notificationY)
    moveAction:start()
  end
end

local function removeNotification(self, notification)
  local function removeFunction()
    local fadeOutAndRemoveAction = AHSequence.new(
      AHOpacityTo.new(notification, NOTIFICATION_FADE_DURATION, nil, 0),
      AHRemove.new(notification)
    )
    notification.opacityAction = fadeOutAndRemoveAction
    wait(fadeOutAndRemoveAction:start())

    local notificationIndex = nil
    for i, existingNotification in ipairs(self.currentNotifications) do
      if (existingNotification == notification) then
        notificationIndex = i
        break
      end
    end
    table.remove(self.currentNotifications, notificationIndex)

    repositionAllNotifications(self)
  end

  local removeCoroutine = newCoroutineWithFunction(removeFunction)
  return removeCoroutine:start()
end

local function createNewNotification(self, notification)
  local currentNotificationCount = #self.currentNotifications

  local notificationY = GROUP_HEIGHT - NOTIFICATION_HEIGHT * 0.5 -
                        currentNotificationCount * (NOTIFICATION_HEIGHT + NOTIFICATION_PADDING_Y)

  local trayOpacity = getKeyOpacity(self)
  local fadeInAction = AHOpacityTo.new(notification, NOTIFICATION_FADE_DURATION, nil, trayOpacity)
  notification.opacityAction = fadeInAction
  fadeInAction:start()

  notification:setLoc(0, notificationY)
  self:insert(notification)
  notification:setAttrLink(MOAIColor.INHERIT_COLOR, self, MOAIColor.COLOR_TRAIT)
  table.insert(self.currentNotifications, notification)
end

local function getAppendedMessageStringWithBonusPointsEarned(messageString, bonusPointsEarned)
  local appendedMessageString = messageString

  if (bonusPointsEarned and (bonusPointsEarned > 0)) then
    local bonusPointsEarnedDifficultyScaled = ScoringUtils.getDifficultyScaledPoints(bonusPointsEarned)
    local roundedBonusPointsEarnedDifficultyScaled = MathUtils.round(bonusPointsEarnedDifficultyScaled)
    appendedMessageString = string.format("%s +%d", appendedMessageString, roundedBonusPointsEarnedDifficultyScaled)
  end

  return appendedMessageString
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BonusNotificationTray:indicateSimpleTextBonus(notificationString, bonusPointsEarned)
  local notificationString = Localizer.get(notificationString)
  local notificationString = getAppendedMessageStringWithBonusPointsEarned(notificationString, bonusPointsEarned)
  local bonusIndicator = SimpleTextBonusIndicator.newWithMessage(notificationString)

  local isRecursive = true
  self:linkColorAttrToChildren(isRecursive)

  local function showIndicatorFunction()
    createNewNotification(self, bonusIndicator)
    waitFor(DISPLAY_INDICATOR_DURATION)
    wait(removeNotification(self, bonusIndicator))
  end
  runInCoroutine(showIndicatorFunction)
end

function BonusNotificationTray:notifyStreakEnd()
  if self.streakIndicator then
    local function removeNotificationAndReset()
      wait(removeNotification(self, self.streakIndicator))
      self.streakIndicator = nil
    end
    runInCoroutine(removeNotificationAndReset)
  end
end

function BonusNotificationTray:notifyStreakWithCount(count)
  if not self.streakIndicator then
    self.streakIndicator = StreakIndicator.new()
    self.streakIndicator:setColor(self:getColor())
    self.streakIndicator:playAnimation(count)
    createNewNotification(self, self.streakIndicator)
  end

  self.streakIndicator:playAnimation(count)
end


-- UIOverlay Widget -------------------

function BonusNotificationTray:getKeyColor()
  return self.keyColor
end

function BonusNotificationTray:setKeyColor(color)
  self.keyColor = color
end

-- Note: Disabled for showing scoring for end-game bonuses for Extra Challenges
-- function BonusNotificationTray:prepareForGameOutro()
--   for i, notification in ipairs(self.currentNotifications) do
--     local opacityAction = notification.opacityAction
--     if (opacityAction) then
--       opacityAction:stop()
--     end
--   end
-- end

function BonusNotificationTray:createFadeOutAction(duration)
  local notificationFadeOutActions = {}
  for i, notification in ipairs(self.currentNotifications) do
    local notificationFadeOutAction = AHOpacityTo.new(notification, duration, nil, 0.0)
    table.insert(notificationFadeOutActions, notificationFadeOutAction)
  end
  local fadeOutAction = AHSpawn.new(unpack(notificationFadeOutActions))
  return fadeOutAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function BonusNotificationTray.new(color)
  color = color or DEFAULT_COLOR
  color[4] = color[4] or DEFAULT_OPACITY

  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  Mixin.mix(self, BonusNotificationTray)

  self:setColor(unpack(color))
  self.keyColor = color

  self.currentNotifications = {}

  return self
end
