class("ResponseSpeedTracker")
-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------
local function processCorrectResponse(self, attributeValue)
  if #self.earnedBonusProportions >= self.maximumBonusCount then
    return
  end

  local clampedAttributeValue = MathUtils.clamp(attributeValue, self.decayStart, self.bonusWindowDuration)
  local earnedBonusProportion = MathUtils.interpolateLinear(clampedAttributeValue, self.decayStart, self.bonusWindowDuration, 1.0, 0.0)

  if earnedBonusProportion > 0 then
    table.insert(self.earnedBonusProportions, earnedBonusProportion)

    if self.onQuickResponse then
      local earnedBonusProportionsSum = self:getEarnedBonusProportionsSum()
      self.onQuickResponse(earnedBonusProportion, earnedBonusProportionsSum)
    end
  end
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function ResponseSpeedTracker:getEarnedBonusProportionsSum()
  return TableUtils.sum(self.earnedBonusProportions)
end

function ResponseSpeedTracker:notifyCorrectResponseWithValue(value)
  processCorrectResponse(self, value)
end

function ResponseSpeedTracker:notifyCorrectResponse()
  local responseTimer = self.responseTimer

  local hasTimerBeenRestarted = (responseTimer:isActive())
  if (hasTimerBeenRestarted) then
    local timeUntilResponse = responseTimer:getTime()
    processCorrectResponse(self, timeUntilResponse)
    if not self.allowMultipleResponses then
      responseTimer:stop()
    end
  end
end

function ResponseSpeedTracker:notifyIncorrectResponse()
  local responseTimer = self.responseTimer
  if not self.allowMultipleResponses then
    responseTimer:stop()
  end
end

function ResponseSpeedTracker:notifyStart()
  local responseTimer = self.responseTimer
  responseTimer:setTime(0.0)
  responseTimer:start()
end

function ResponseSpeedTracker:setBonusWindowDuration(bonusWindowDuration)
  self.bonusWindowDuration = bonusWindowDuration
end

function ResponseSpeedTracker:setDecayWindowStart(decayWindowStart)
  self.decayStart = decayWindowStart
end

function ResponseSpeedTracker:getCurrentTime()
  local responseTimer = self.responseTimer
  return responseTimer:getTime()
end

function ResponseSpeedTracker:isActive()
  local responseTimer = self.responseTimer
  return responseTimer:isActive()
end

function ResponseSpeedTracker:getEarnedBonusesCount()
  return #self.earnedBonusProportions
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

--- Creates a new speed tracker for quick responses
-- @param bonusWindowDuration (number) Time (in seconds) that a user must respond within to get a bonus
-- @param onQuickResponse - (function) callback called when a bonus is earned
-- @param decayStart - (Optional, number) Time in the bonus window where the reward will start
--   to decay. Ex. If the bonus window is 5 seconds and the decay start is 3, a bonus is worth full points for
--   the first 3 seconds. After that, the bonus decays to zero points within remainder of the bonus window.
-- @param maximumBonusCount - (Optional, number) Maximum number of bonuses taken into account. When calling
--   notifyCorrectResponse or notifyCorrectResponseWithValue, if the maximum bonus count has been reached
--   the call will be ignored.
-- @param allowMultipleResponses - (Optional, boolean) Defaults to False. Setting to True will cause the
--   responseTimer to not stop when calling notifyCorrectResponse, or notifyIncorrectResponse, allowing
--   multiple responses to the same question to earn a bonus.
-- @return (ResponseSpeedTracker) - A new speed tracker object
function ResponseSpeedTracker.new(bonusWindowDuration, onQuickResponse, decayStart, maximumBonusCount, allowMultipleResponses)
  bonusWindowDuration = bonusWindowDuration or 0
  decayStart = decayStart or 0
  maximumBonusCount = maximumBonusCount or math.huge
  allowMultipleResponses = allowMultipleResponses or false
  assert(decayStart <= bonusWindowDuration, "Decay window shouldn't be longer than the bonus window")

  local self = {}
  Mixin.mix(self, ResponseSpeedTracker)

  local responseTimer = MOAITimer.new()
  responseTimer:setMode(MOAITimer.CONTINUE)
  self.responseTimer = responseTimer

  self.bonusWindowDuration = bonusWindowDuration
  self.onQuickResponse = onQuickResponse
  self.decayStart = decayStart
  self.maximumBonusCount = maximumBonusCount
  self.allowMultipleResponses = allowMultipleResponses
  self.earnedBonusProportions = {}

  return self
end
