--- BranchLexer
-- @author Created by andrewc on February 23, 2019
-- This class takes in raw strings of our branching language
-- and separates it into tokens for our parser.
-- Each token has a tokenType and data. Data is almost always just going to be the
-- token type, with the main exception being ARG tokens.

class("BranchLexer")

require("utils.StringUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

BranchLexer.IDENTIFIER = {
  NOT = "NOT",
  AND = "AND",
  OR = "OR",
  OPEN = "OPEN",
  CLOSE = "CLOSE",
  ARG = "ARG",
  COMMA = "COMMA"
}

local PATTERN = {
  [BranchLexer.IDENTIFIER.NOT] = '^NOT',
  [BranchLexer.IDENTIFIER.AND] = '^AND',
  [BranchLexer.IDENTIFIER.OR] = '^OR',
  [BranchLexer.IDENTIFIER.OPEN] = '^%(',
  [BranchLexer.IDENTIFIER.CLOSE] = '^%)',
  [BranchLexer.IDENTIFIER.ARG] = '^[%a%d_-]+',
  [BranchLexer.IDENTIFIER.COMMA] = '^%,'
}

local IDENTIFIER_WEIGHT = {
  [BranchLexer.IDENTIFIER.NOT] = 10,
  [BranchLexer.IDENTIFIER.AND] = 10,
  [BranchLexer.IDENTIFIER.OR] = 10,
  [BranchLexer.IDENTIFIER.OPEN] = 10,
  [BranchLexer.IDENTIFIER.CLOSE] = 10,
  [BranchLexer.IDENTIFIER.COMMA] = 10,
  [BranchLexer.IDENTIFIER.ARG] = 0
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function checkBetterToken(tokenToReturn, endIndexToReturn, tokenToCheck, endIndexToCheck)
  if (tokenToReturn == nil) then
    return true
  elseif (endIndexToCheck > endIndexToReturn) then
    return true
  elseif (IDENTIFIER_WEIGHT[tokenToReturn] < IDENTIFIER_WEIGHT[tokenToCheck]) then
    return true
  end
  return false
end

local function findNextToken(branchString)
  local tokenToReturn = nil
  local endIndexToReturn = nil
  for _, tokenToCheck in pairs(BranchLexer.IDENTIFIER) do
    if (PATTERN[tokenToCheck]) then
      local startIndex, endIndexToCheck = string.find(branchString, PATTERN[tokenToCheck])
      if (startIndex ~= nil and checkBetterToken(tokenToReturn, endIndexToReturn, tokenToCheck, endIndexToCheck)) then
        tokenToReturn = tokenToCheck
        endIndexToReturn = endIndexToCheck
      end
    end
  end
  assert(tokenToReturn, "No token found: " .. branchString)
  assert(endIndexToReturn ~= nil, "Error reading branch: " .. branchString)
  return tokenToReturn, endIndexToReturn
end

local function stringToTokens(branchString)
  local tokens = {}
  branchString = StringUtils.trim(branchString)
  while (#branchString > 0) do
    local token, endIndex = findNextToken(branchString)
    if (token) then
      local newToken = {
        tokenType = token,
        data = string.sub(branchString, 1, endIndex)
      }
      table.insert(tokens, newToken)
      branchString = StringUtils.trim(string.sub(branchString, endIndex + 1, #branchString))
    end
  end
  return tokens
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function BranchLexer.getTokens(branchString)
  return stringToTokens(branchString)
end
