--- BranchParser
-- @author Created by andrewc on February 20, 2019
-- This class parses our branching language.
-- The string is passed into the lexer to be converted into tokens,
-- and then those tokens are parsed into a tree of nodes.

class("BranchParser")

require("branching.BranchLexer")
require("branching.BranchNode")
require("branching.NumericBranchNode")
require("branching.CompositeBranchNode")
require("branching.NotBranchNode")
require("branching.ThresholdBranchNode")
require("branching.ExperimentBranchNode")
require("branching.InterestBranchNode")
require("branching.TopInterestBranchNode")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BRANCH_TYPE = {
  EXPERIMENT = "experiment",
  INTEREST = "hasInterest",
  TOP_INTEREST = "topInterest",
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local tokensToTree, getNextNode -- Forward declaration

local function assertRootIsNil(rootNode)
  assert(rootNode == nil, "Missing a composite token!")
end

local function createExperimentBranchNode(params)
  local experimentId = table.remove(params, 1)
  return ExperimentBranchNode.new(experimentId, params)
end

local function createInterestBranchNode(params)
  local interestId = table.remove(params, 1)
  return InterestBranchNode.new(interestId)
end

local function createTopInterestBranchNode(params)
  return TopInterestBranchNode.new(params)
end

local function createBranchNode(branchType, params)
  local branchFunctions = {
    [BRANCH_TYPE.EXPERIMENT] = createExperimentBranchNode,
    [BRANCH_TYPE.INTEREST] = createInterestBranchNode,
    [BRANCH_TYPE.TOP_INTEREST] = createTopInterestBranchNode,
  }

  assert(TableUtils.contains(BRANCH_TYPE, branchType), "Invalid Branch Type: " .. branchType)
  local branchFunction = branchFunctions[branchType]
  return branchFunction(params)
end

local function parseThresholdToken(self, token, tokens, rootNode)
  local branches = {}
  assert(table.remove(tokens, 1).tokenType == BranchLexer.IDENTIFIER.OPEN, "Threshold requires a OPEN token.")
  local threshold = table.remove(tokens, 1)
  assert(threshold.tokenType == BranchLexer.IDENTIFIER.ARG, "Threshold requires the first arugment to be a number.")
  while (tokens[1].tokenType ~= BranchLexer.IDENTIFIER.CLOSE) do
    assert(table.remove(tokens, 1).tokenType == BranchLexer.IDENTIFIER.COMMA, "Threshold requires a Comma after each argument.")
    local branchNode, tokens = getNextNode(self, tokens, nil)
    table.insert(branches, branchNode)
  end
  table.remove(tokens, 1)

  local node = ThresholdBranchNode.new(threshold.data, branches)
  return node, tokens
end

local function parseArgToken(self, token, tokens, rootNode)
  assertRootIsNil(rootNode)
  local params = {}
  local branchType = token.data

  if (branchType == BRANCH_TYPE.THRESHOLD) then
    return parseThresholdToken(self, token, tokens, rootNode)
  end

  local nextToken = function()
    token = table.remove(tokens, 1)
    assert(token, "Missing close token for branch: " .. branchType)
  end

  local addNextToken = function()
    nextToken()
    assert(token.tokenType == BranchLexer.IDENTIFIER.ARG, "Invalid token in branch. Expected ARG, found: " .. token.tokenType)
    table.insert(params, token.data)
    nextToken()
  end

  nextToken()
  assert(token.tokenType == BranchLexer.IDENTIFIER.OPEN, "Branch missing OPEN token!")
  addNextToken()
  while (token.tokenType ~= BranchLexer.IDENTIFIER.CLOSE) do
    assert(token.tokenType == BranchLexer.IDENTIFIER.COMMA, "Missing comma between branch args.")
    addNextToken()
  end

  local node = createBranchNode(branchType, params)
  return node, tokens
end

local function parseOpenToken(self, token, tokens, rootNode)
  assertRootIsNil(rootNode)
  self.openTokens = self.openTokens + 1
  return tokensToTree(self, tokens)
end

local function parseCloseToken(self, token, tokens, rootNode)
  assert(self.openTokens > 0, "Extra close token found.")
  self.openTokens = self.openTokens - 1
  return rootNode, tokens
end

local function parseSyntaxToken(self, token, tokens)
  assert(false, "Extra Syntax Token found: " .. token.data)
end

local function parseAndToken(self, token, tokens, rootNode)
  local nextNode, remainingTokens = getNextNode(self, tokens)
  local andNode = CompositeBranchNode.newAnd(rootNode, nextNode)
  return andNode, remainingTokens
end

local function parseOrToken(self, token, tokens, rootNode)
  local nextNode, remainingTokens = getNextNode(self, tokens)
  local orNode = CompositeBranchNode.newOr(rootNode, nextNode)
  return orNode, remainingTokens
end

local function parseNotToken(self, token, tokens, rootNode)
  assertRootIsNil(rootNode)
  local negatedNode, remainingTokens = getNextNode(self, tokens)
  local node = NotBranchNode.new(negatedNode)
  return node, remainingTokens
end

function getNextNode(self, tokens, rootNode)
  if (tokens == nil or #tokens < 1) then return nil end

  local token_functions = {
    [BranchLexer.IDENTIFIER.NOT] = parseNotToken,
    [BranchLexer.IDENTIFIER.AND] = parseAndToken,
    [BranchLexer.IDENTIFIER.OR] = parseOrToken,
    [BranchLexer.IDENTIFIER.OPEN] = parseOpenToken,
    [BranchLexer.IDENTIFIER.CLOSE] = parseCloseToken,
    [BranchLexer.IDENTIFIER.ARG] = parseArgToken,
    [BranchLexer.IDENTIFIER.COMMA] = parseSyntaxToken
  }

  local token = table.remove(tokens, 1)
  local tokenFunction = token_functions[token.tokenType]
  return tokenFunction(self, token, tokens, rootNode)
end

function tokensToTree(self, tokens)
  local node = nil
  while (tokens ~= nil and #tokens > 0) do
    node, tokens = getNextNode(self, tokens, node)
  end
  assert(node, "No node returned!")
  assert(self.openTokens == 0, "Extra open token found.")
  return node
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BranchParser:getRootNode(branch)
  self.tokens = BranchLexer.getTokens(branch)
  self.openTokens = 0
  return tokensToTree(self, self.tokens)
end

function BranchParser.buildTree(branch)
  local self = BranchParser.new()
  return self:getRootNode(branch)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function BranchParser.new()
  local self = {}
  Mixin.mix(self, BranchParser)
  return self
end

-- ----------------------------------------------------------------------------
-- External API
-- ----------------------------------------------------------------------------

function evaluateBranchString(branchString)
  local rootNode = BranchParser.buildTree(branchString)
  return rootNode:evaluate()
end
