--- Carousel
-- @author Created by karl on October 26, 2015.

class("Carousel")

require("widgets.progress.CircleStepProgressIndicator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SWIPE_RESOLUTION = 4
local AUTO_SCROLL_SPEED = Screen.width
local SCROLL_DURATION = 0.5

local SCROLL_DIRECTION = {
  NONE = "NONE",
  HORIZONTAL = "HORIZONTAL",
  VERTICAL = "VERTICAL"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getScrollPositionX(self)
  return self.cardContainer:getX()
end

local function getScrollPositionY(self)
  return self.cardContainer:getY()
end

local function setScrollPositionX(self, scrollAmount)
  local cardWidth = self:getWidth()
  local rightEdge = 0
  local leftEdge = -((#self.cardContainer.children - 1) * cardWidth)
  local positionX = MathUtils.clamp(scrollAmount, leftEdge, rightEdge)
  self.cardContainer:setX(positionX)
  local scrollProportion = -positionX / cardWidth
  local progressStep = MathUtils.round(scrollProportion)
  self.progressIndicator:setProgressStep(progressStep)

  local nextScrollIndex = math.floor(scrollProportion) + 1
  local nextScrollAction = self.scrollActionXs[nextScrollIndex]
  local lastScrollIndex = self.lastScrollIndex
  local lastScrollAction = self.scrollActionXs[lastScrollIndex]

  if (lastScrollAction) then
    if (nextScrollIndex > lastScrollIndex) then
      lastScrollAction:applyAtTime(1)
    elseif (nextScrollIndex < lastScrollIndex) then
      lastScrollAction:applyAtTime(0)
    end
  end

  if (nextScrollAction) then
    local inScreenProportion = scrollProportion % 1.0
    nextScrollAction:applyAtTime(inScreenProportion)
  end

  self.lastScrollIndex = nextScrollIndex

  self.onScrollX()
end

local function setScrollPositionY(self, scrollAmount)
  local minimumY = -self:getHeight()
  local maximumY = self:getHeight()
  local positionY = MathUtils.clamp(scrollAmount, minimumY, maximumY)
  self.cardContainer:setY(positionY)

  self.onScrollY()
end

local function getSwipeVelocity(self)
  if (#self.moveEvents <= 1) then
    return 0
  end
  local lastEvent = self.moveEvents[#self.moveEvents]
  local firstEvent = self.moveEvents[1]
  if (#self.moveEvents > SWIPE_RESOLUTION) then
    firstEvent = self.moveEvents[#self.moveEvents - SWIPE_RESOLUTION]
  end
  local duration = lastEvent.time - firstEvent.time
  local distanceX = lastEvent.x - firstEvent.x
  local distanceY = lastEvent.y - firstEvent.y
  local velocityX = distanceX / duration
  local velocityY = distanceY / duration
  return velocityX, velocityY
end

local function getPageModulus(self)
  local cardWidth = self:getWidth()
  local fractionalPage = self.cardContainer:getX() / cardWidth
  return fractionalPage % 1
end

local function getDefaultAutoScrollVelocityX(self)
  local pageModulus = getPageModulus(self)
  if (pageModulus == 0) then
    return 0
  end
  if (pageModulus >= 0.5) then
    return AUTO_SCROLL_SPEED
  end
  return -AUTO_SCROLL_SPEED
end

local function getDefaultAutoScrollVelocityY(self)
  local scrollY = getScrollPositionY(self)
  if (scrollY < 0.0) then
    return AUTO_SCROLL_SPEED
  end
  return -AUTO_SCROLL_SPEED
end

local function getAutoScrollVelocityX(self)
  local scrollVelocityX, _ = getSwipeVelocity(self)
  scrollVelocityX = scrollVelocityX or 0.0
  if math.abs(scrollVelocityX) < AUTO_SCROLL_SPEED then
    scrollVelocityX = getDefaultAutoScrollVelocityX(self)
  end
  return scrollVelocityX
end

local function getAutoScrollVelocityY(self)
  local _, scrollVelocityY = getSwipeVelocity(self)
  scrollVelocityY = scrollVelocityY or 0.0
  if (math.abs(scrollVelocityY) < AUTO_SCROLL_SPEED) then
    scrollVelocityY = getDefaultAutoScrollVelocityY(self)
  end
  return scrollVelocityY
end

local function getAutoScrollAmountX(self, velocity)
  local cardWidth = self:getWidth()
  local pageModulus = getPageModulus(self)
  if (velocity == 0) then
    return 0
  elseif (velocity < 0) then
    return -pageModulus * cardWidth
  end
  return (1 - pageModulus) * cardWidth
end

local function getAutoScrollAmountY(self, velocity)
  if (velocity == 0) then
    return 0
  end
  return -getScrollPositionY(self)
end

local function canScrollX(self, velocity)
  local maxX = 0
  local minX = -(#self.cardContainer.children - 1) * self:getWidth()
  if (velocity > 0) then
    return self.cardContainer:getX() < maxX
  elseif (velocity < 0) then
    return self.cardContainer:getX() > minX
  end
  return false
end

local function canScrollY(self, velocity)
  local scrollY = getScrollPositionY(self)
  local minimumY = -self:getHeight()
  local maximumY = self:getHeight()
  local canScrollUp = (velocity > 0.0 and scrollY < maximumY)
  local canScrollDown = (velocity < 0.0 and scrollY > minimumY)
  local canScroll = (canScrollUp or canScrollDown)
  return canScroll
end

local function startAutoScrollX(self)
  local scrollVelocity = getAutoScrollVelocityX(self)
  local scrollAmount = getAutoScrollAmountX(self, scrollVelocity)
  if (scrollVelocity ~= 0 and scrollAmount ~= 0 and canScrollX(self, scrollVelocity)) then
    self.scrollActionX = AHNonAttributeBy.new(self, self.scrollDuration, getScrollPositionX, setScrollPositionX, self.scrollEase, scrollAmount)
    self.scrollActionX:start()
  end
end

local function startAutoScrollY(self)
  local scrollVelocity = getAutoScrollVelocityY(self)
  local scrollAmount = getAutoScrollAmountY(self, scrollVelocity)
  if (scrollVelocity ~= 0 and scrollAmount ~= 0 and canScrollY(self, scrollVelocity)) then
    self.scrollActionY = AHLocationYBy.new(self.cardContainer, self.scrollDuration, self.scrollEase, scrollAmount)
    self.scrollActionY:start()
  end
end

local function startAutoScroll(self)
  if (self.scrollDirection == SCROLL_DIRECTION.HORIZONTAL) then
    startAutoScrollX(self)
  elseif (self.scrollDirection == SCROLL_DIRECTION.VERTICAL) then
    startAutoScrollY(self)
  end
end

local function stopScrollingX(self)
  if (self.scrollActionX) then
    self.scrollActionX:stop()
    self.scrollActionX = nil
  end
end

local function stopScrollingY(self)
  if (self.scrollActionY) then
    self.scrollActionY:stop()
    self.scrollActionY = nil
  end
end

local function addMoveEvent(self, event)
  local timestamp = MOAISim.getDeviceTime()
  table.insert(self.moveEvents, {x = event.x, y = event.y, time = timestamp})
end

local function updateCarouselX(self, event)
  local lastEvent = self.moveEvents[#self.moveEvents]
  if (#self.moveEvents > 1) then
    lastEvent = self.moveEvents[#self.moveEvents - 1]
  end
  setScrollPositionX(self, getScrollPositionX(self) + event.x - lastEvent.x)
end

local function updateCarouselY(self, event)
  local lastEvent = self.moveEvents[#self.moveEvents]
  if (#self.moveEvents > 1) then
    lastEvent = self.moveEvents[#self.moveEvents - 1]
  end

  local currentScrollPositionY = getScrollPositionY(self)
  local newScrollPositionY = currentScrollPositionY + event.y - lastEvent.y

  local shouldScrollUp = (newScrollPositionY >= currentScrollPositionY and currentScrollPositionY < 0.0 and self.allowVerticalScrolling)
  local shouldScrollDown = (newScrollPositionY <= currentScrollPositionY and self.allowVerticalScrolling)

  if (shouldScrollUp or shouldScrollDown) then
    setScrollPositionY(self, newScrollPositionY)
  end
end

local function updateCarousel(self, event)
  local cardWidth = self:getWidth()
  local fractionalPage = math.abs(self.cardContainer:getX() / cardWidth)
  local cardHorizontallyCentered = (fractionalPage < 0.01 or fractionalPage > 0.99)
  local cardVerticallyCentered = (math.abs(self.cardContainer:getY()) < 0.001)
  local updateHorizontalScroll = ((self.scrollDirection == SCROLL_DIRECTION.HORIZONTAL) and cardVerticallyCentered)
  local updateVerticalScroll = ((self.scrollDirection == SCROLL_DIRECTION.VERTICAL) and cardHorizontallyCentered)

  if (updateHorizontalScroll) then
    updateCarouselX(self, event)
  elseif (updateVerticalScroll) then
    updateCarouselY(self, event)
  end
end

local function determineScrollDirection(self, event)
  local lastEvent = self.moveEvents[#self.moveEvents]
  if (#self.moveEvents > 1) then
    lastEvent = self.moveEvents[#self.moveEvents - 1]
  end

  local deltaX = math.abs(event.x - lastEvent.x)
  local deltaY = math.abs(event.y - lastEvent.y)
  if (deltaX > deltaY) then
    self.scrollDirection = SCROLL_DIRECTION.HORIZONTAL
  elseif (deltaY > deltaX and self.allowVerticalScrolling) then
    self.scrollDirection = SCROLL_DIRECTION.VERTICAL
  else
    self.scrollDirection = SCROLL_DIRECTION.NONE
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Carousel:addCard(card)
  self.cardContainer:insert(card)
  self.cardContainer:alignChildren()
  self.progressIndicator:setStepCount(#self.cardContainer.children)
end

--- Insert a custom scrolling action.
-- @param startIndex (integer) - The index from which the action starts (action runs from index to index + 1)
-- @param action (action) - The action to perform.
function Carousel:setScrollAction(startIndex, action)
  self.scrollActionXs[startIndex] = AHUntimedAction.new(action)
end

function Carousel:setProgressYPosition(yPosition)
  self.progressIndicator:setY(yPosition)
end

function Carousel:setHorizontalScrollCallback(onScroll)
  self.onScrollX = onScroll or function() end
end

function Carousel:setVerticalScrollCallback(onScroll)
  self.onScrollY = onScroll or function() end
end

function Carousel:setScrollEndedCallback(onScrollEnded)
  self.onScrollEnded = onScrollEnded or function() end
end

function Carousel:getScrollProportionX()
  return -self.cardContainer:getX() / self:getWidth()
end

function Carousel:getScrollProportionY()
  local proportion = getScrollPositionY(self) / self:getHeight()
  return proportion
end

function Carousel:setScrollEase(ease)
  self.scrollEase = ease
end

function Carousel:setScrollDuration(scrollDuration)
  self.scrollDuration = scrollDuration
end

function Carousel:setAllowVerticalScrolling(allow)
  self.allowVerticalScrolling = allow
end

function Carousel:getSwipeVelocityY()
  local _, swipeVelocityY = getSwipeVelocity(self)
  return swipeVelocityY
end

function Carousel:scrollToLastCard()
  local scrollAmount = -(#self.cardContainer.children - 1 - self:getScrollProportionX()) * self:getWidth()
  local scrollDuration = 0.2
  self.scrollActionX = AHNonAttributeBy.new(self, scrollDuration, getScrollPositionX, setScrollPositionX, nil, scrollAmount)

  return self.scrollActionX:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createProgressIndicator(highlightColor, neutralColor)
  local radius = Screen.getWidthPart(0.01)
  local spacing = radius * 3
  return CircleStepProgressIndicator.new(radius, spacing, highlightColor, neutralColor)
end

-- Constructor ------------------------

--- Create a new carousel.
-- @param highlightColor (color) - Color for the highlighted progress circle.
-- @param neutralColor (color) - Color for the neutral progress circles.
function Carousel.new(highlightColor, neutralColor, width, height, bottomSafeArea)
  width = width or Screen.fullWidth
  height = height or Screen.fullHeight
  bottomSafeArea = bottomSafeArea or 0
  local self = Group.new(width, height)
  Mixin.mix(self, Carousel)

  self.cardContainer = Row.new(0)
  self.cardContainer:setAnchorPointBottomLeft()
  self:insert(self.cardContainer)

  self.progressIndicator = createProgressIndicator(highlightColor, neutralColor)
  local progressIndicatorY = math.max(self:getHeightPart(0.05), self.progressIndicator:getHeightPart(0.5) + bottomSafeArea)
  self.progressIndicator:setLoc(self:getWidthPart(0.5), progressIndicatorY)
  self:insert(self.progressIndicator)
  self.progressIndicator:setProgressStep(0)
  self.progressIndicator:inheritColor(self)

  TouchManager.addTouchEnabledMethods(self, function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      self.scrollDirection = SCROLL_DIRECTION.NONE
      stopScrollingX(self)
      stopScrollingY(self)
      addMoveEvent(self, event)
    elseif (event.phase == TouchManager.TOUCH_MOVED) then
      addMoveEvent(self, event)
      if (self.scrollDirection == SCROLL_DIRECTION.NONE) then
        determineScrollDirection(self, event)
      end
      updateCarousel(self, event)
    else
      startAutoScroll(self)
      self.moveEvents = {}
      self.onScrollEnded()
    end
  end)

  self.moveEvents = {}
  self.scrollActionXs = {}
  self:setHorizontalScrollCallback(nil)
  self:setVerticalScrollCallback(nil)
  self:setScrollEndedCallback(nil)
  self.lastScrollIndex = nil
  self.scrollEase = AHEaseExponentialOut.new()
  self.scrollDuration = SCROLL_DURATION
  self.scrollDirection = SCROLL_DIRECTION.NONE
  self.allowVerticalScrolling = false

  return self
end
