--- Aligns children in a grid pattern.
-- It spaces everything based on its largest child and the column count specified at construction.
-- All children will have their anchor point set to (0.5, 0.5).
-- @author Created by karl on June 17, 2014.

-- ===========================================================================
class("Grid")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getLargestDimensions(self)
  local maxW, maxH = 0, 0
  for _, child in ipairs(self.children) do
    local childW, childH = child:getDims()
    if childW and childW > maxW then
      maxW = childW
    end
    if childH and childH > maxH then
      maxH = childH
    end
  end
  return math.ceil(maxW), math.ceil(maxH)
end

local function getOptimalDimensions(self, maxW, maxH)
  local rowCount = math.ceil(#self.children / self.columnCount)
  local optimalW = maxW * self.columnCount + self.spacingX * (self.columnCount - 1)
  local optimalH = maxH * rowCount + self.spacingY * (rowCount - 1)
  return optimalW, optimalH
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Grid:alignChildren()
  local maxW, maxH = getLargestDimensions(self)
  local selfW, selfH = getOptimalDimensions(self, maxW, maxH)
  self:setDims(selfW, selfH)

  local colIndex = 1
  local posX, posY = math.floor(maxW/2), math.floor(selfH - maxH/2)
  local deltaX, deltaY = maxW + self.spacingX, -(maxH + self.spacingY)

  for _, child in ipairs(self.children) do
    if colIndex > self.columnCount then
      posX = math.floor(maxW/2)
      posY = posY + deltaY
      colIndex = 1
    end
    child:setLoc(posX, posY)
    child:setAnchorPointCenter()
    posX = posX + deltaX
    colIndex = colIndex + 1
  end
end

function Grid:setColumnCount(count)
  self.columnCount = count
end

function Grid:setSpacingX(spacingX)
  self.spacingX = spacingX
end

function Grid:setSpacingY(spacingY)
  self.spacingY = spacingY
end

function Grid:getColumnCount()
  return self.columnCount
end

function Grid:getSpacingX()
  return self.spacingX
end

function Grid:getSpacingY()
  return self.spacingY
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a grid container.
-- @param columnCount (number) - The number of columns.
-- @param spacingX (number, optional) - spacing between objects on the X axis.
-- @param spacingY (number, optional) - spacing between objects on the Y axis.
function Grid.new(columnCount, spacingX, spacingY)
  spacingX = spacingX or 0
  spacingY = spacingY or 0

  local self = Group.new(2, 2)
  Mixin.mix(self, Grid)

  self.columnCount = columnCount
  self.spacingX = spacingX
  self.spacingY = spacingY

  return self
end
