--- Row & Column


-- ===========================================================================
class("RowColumn")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ALIGNMENT_MIN = "ALIGNMENT_MIN"
local ALIGNMENT_MID = "ALIGNMENT_MID"
local ALIGNMENT_MAX = "ALIGNMENT_MAX"

-- These map external alignments to a mathematically consistent internal
-- representation. Vertical alignment is the only one that comes out differently.
local HORIZONTAL_ALIGNMENT_MAPPING = {
  [Alignment.LEFT] = ALIGNMENT_MIN,
  [Alignment.CENTER] = ALIGNMENT_MID,
  [Alignment.RIGHT] = ALIGNMENT_MAX
}

local VERTICAL_ALIGNMENT_MAPPING = {
  [Alignment.TOP] = ALIGNMENT_MAX,
  [Alignment.CENTER] = ALIGNMENT_MID,
  [Alignment.BOTTOM] = ALIGNMENT_MIN
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Note about the "DS" helper functions.
-- In order to accommodate both row and column operations and avoid lots of code
-- duplication, I've adopted the concept of dynamic and static axes. In a column,
-- X is static and Y is dynamic, because the children get moved along the Y axis.
-- In a row, X is dynamic and Y is static.

local function getDimsDS(self, object)
  return self.convertDynamicStaticDims(object:getDims())
end

local function setDimsDS(self, object, d, s)
  object:setDims(self.convertDynamicStaticDims(d, s))
end

local function setLocDS(self, object, d, s)
  object:setLoc(self.convertDynamicStaticDims(d, s))
end

local function getAnchorPointDS(self, object)
  return self.convertDynamicStaticDims(object:getAnchorPoint())
end

local function setAnchorPointDS(self, object, d, s)
  object:setAnchorPoint(self.convertDynamicStaticDims(d, s))
end

local function getEffectiveDimsDS(self, object)
  return self.convertDynamicStaticDims(object:getEffectiveDims())
end

local function getRequiredDimsDS(self)
  local minEdgeS = math.huge
  local maxEdgeS = -math.huge
  local totalSizeD = 0

  for _, child in pairs(self.children) do
    local childDimD, childDimS = getEffectiveDimsDS(self, child)
    local _, anchorS = getAnchorPointDS(self, child)
    local lowEdgeS = childDimS * anchorS * -1
    local highEdgeS = childDimS * (1 - anchorS)
    if lowEdgeS < minEdgeS then
      minEdgeS = lowEdgeS
    end
    if highEdgeS > maxEdgeS then
      maxEdgeS = highEdgeS
    end
    totalSizeD = totalSizeD + childDimD
  end

  return totalSizeD, maxEdgeS - minEdgeS
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Get the spacing between items in the dynamic axis.
-- @return (number) - The value for spacing.
function RowColumn:getSpacing()
  return self.spacing
end

--- Set the spacing between items in the dynamic axis.
-- Note: This class uses lazy updating, you need to call alignChildren() for this change to take effect.
-- @param spacing (number) - The value for spacing.
function RowColumn:setSpacing(spacing)
  self.spacing = spacing
end

--- Get the maximum spacing allowed between items in the dynamic axis. Used when spacing is not set.
-- @return (number) - The value for maximum spacing.
function RowColumn:getMaxSpacing()
  return self.maxSpacing
end

--- Set the maximum spacing allowed between items in the dynamic axis. Used when spacing is not set.
-- Note: This class uses lazy updating, you need to call alignChildren() for this change to take effect.
-- @param maxSpacing (number) - The value for maximum spacing.
function RowColumn:setMaxSpacing(maxSpacing)
  self.maxSpacing = maxSpacing
end

--- Get the alignment of items within the row or column in the static axis.
-- @return (Alignment) - The alignment.
function RowColumn:getAlignment()
  return self.alignment
end

--- Set the alignment of items within the row or column in the static axis.
-- Note: This class uses lazy updating, you need to call alignChildren() for this change to take effect.
-- @param alignment (Alignment) - The alignment.
function RowColumn:setAlignment(alignment)
  self.alignment = alignment
  self.mappedAlignment = self.alignmentMapping[alignment]
end

--- Insert extra space along the dynamic axis (vertical on a column, horizontal on a row).
-- Note: The spacer is a group like any other, and will have normal spacing added to it
--       according to this row/column's spacing value.
-- @param size (number) - The amount of space to add.
-- @return (Group) - the spacer group.
function RowColumn:insertSpacer(size)
  local spacer = Group.new(self.convertDynamicStaticDims(size, 0))
  self:insert(spacer)
  return spacer
end

function RowColumn:alignChildren()
  local spacing = self.spacing
  local shouldResizeDynamicAxis = spacing ~= nil
  local childCount = #self.children

  if childCount == 0 then
    if shouldResizeDynamicAxis then
      local _, selfDimS = getDimsDS(self, self)
      setDimsDS(self, self, 0, selfDimS)
    end
    return
  end

  local requiredD, requiredS = getRequiredDimsDS(self)
  local selfDimD, selfDimS = getDimsDS(self, self)

  if not shouldResizeDynamicAxis then
    spacing = (selfDimD - requiredD) / (childCount - 1)
    if spacing < 0 then
      spacing = 0
      shouldResizeDynamicAxis = true
    end
  end

  if shouldResizeDynamicAxis then
    local totalSpacing = spacing * (#self.children - 1)
    setDimsDS(self, self, requiredD + totalSpacing, requiredS)
  else
    setDimsDS(self, self, selfDimD, requiredS)
  end

  if spacing > self.maxSpacing then
    spacing = self.maxSpacing
  end

  local selfDimD, selfDimS = getDimsDS(self, self)
  local alignment = self.mappedAlignment
  local growDirectionMultiplier = self.growDirectionMultiplier
  local baseLocD = 0
  if growDirectionMultiplier <= 0 then
    baseLocD = selfDimD
  end
  for _, child in pairs(self.children) do
    local childDimD, childDimS = getEffectiveDimsDS(self, child)
    local childLocD, childLocS = baseLocD + (childDimD / 2) * growDirectionMultiplier, 0
    if alignment == ALIGNMENT_MIN then
      childLocS = childDimS / 2
    elseif alignment == ALIGNMENT_MAX then
      childLocS = selfDimS - childDimS / 2
    elseif alignment == ALIGNMENT_MID then
      childLocS = selfDimS / 2
    end
    setLocDS(self, child, childLocD, childLocS)
    child:setAnchorPointCenter()
    baseLocD = baseLocD + (childDimD + spacing) * growDirectionMultiplier
  end
end

function RowColumn:getChildren()
  return self.children
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Common constructor for abstract rows/columns. Do not instantiate this directly.
-- @param isRow (boolean) - If true, this is a row, not a column.
-- @param spacing (number) - Space to put between children. If nil, self will not resize, but rather dynamically determine spacing based on its current size.
-- @param aligmnent (Alignment) - The alignment.
-- @param alignmentMapping (table) - Mapping to internal alignment values.
-- @param growDirectionMultiplier (number) - The direction to "grow" in. 1 = right (row) or up (col). -1 = left (row) or down (col).
-- @return A new row or column.
function RowColumn.new(isRow, spacing, alignment, alignmentMapping, growDirectionMultiplier)
  alignment = alignment or Alignment.CENTER
  local self = Group.new(2, 2)
  Mixin.mix(self, RowColumn)
  self.alignmentMapping = alignmentMapping
  self.alignment = alignment
  self.mappedAlignment = alignmentMapping[alignment]
  self.spacing = spacing
  self.maxSpacing = math.huge
  self.growDirectionMultiplier = growDirectionMultiplier

  if isRow then
    self.convertDynamicStaticDims = function(x, y) return x, y end
  else
    self.convertDynamicStaticDims = function(x, y) return y, x end
  end

  return self
end



-- ===========================================================================
class("Row")
-- ===========================================================================

Row.GrowLeft = "Row.GrowLeft"
Row.GrowRight = "Row.GrowRight"


--- Row constructor
-- @param spacing Space to put between children. If nil, self will not resize, but rather dynamically determine spacing based on its current size.
-- @param aligmnent Alignment along the Y axis. Default Alignment.CENTER
-- @param growDirection Row.GrowLeft or Row.GrowRight. Default Row.GrowRight
-- @return A new row or column.
function Row.new(spacing, alignment, growDirection)
  growDirection = growDirection or Row.GrowRight
  local growDirectionMultiplier
  if growDirection == Row.GrowLeft then
    growDirectionMultiplier = -1
  elseif growDirection == Row.GrowRight then
    growDirectionMultiplier = 1
  else
    assert(false, "growDirection must be one of: Row.GrowLeft, Row.GrowRight")
  end

  return RowColumn.new(true, spacing, alignment, VERTICAL_ALIGNMENT_MAPPING, growDirectionMultiplier)
end



-- ===========================================================================
class("Column")
-- ===========================================================================

Column.GrowUp = "Column.GrowUp"
Column.GrowDown = "Column.GrowDown"


--- Column constructor
-- @param spacing Space to put between children. If nil, self will not resize, but rather dynamically determine spacing based on its current size.
-- @param aligmnent Alignment along the X axis. Default Alignment.CENTER
-- @param growDirection Column.GrowUp or Column.GrowDown. Default Column.GrowDown
-- @return A new row or column.
function Column.new(spacing, alignment, growDirection)
  growDirection = growDirection or Column.GrowDown
  local growDirectionMultiplier
  if growDirection == Column.GrowUp then
    growDirectionMultiplier = 1
  elseif growDirection == Column.GrowDown then
    growDirectionMultiplier = -1
  else
    assert(false, "growDirection must be one of: Column.GrowUp, Column.GrowDown")
  end

  return RowColumn.new(false, spacing, alignment, HORIZONTAL_ALIGNMENT_MAPPING, growDirectionMultiplier)
end
