--- CardCarouselCard
-- @author Created by Aaron Barrett on April 20, 2015.

class("CardCarouselCard")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CardCarouselCard:getCardType()
  return self.cardType
end

function CardCarouselCard:setEnterCallback(callback)
  self.enterCallback = callback
end

function CardCarouselCard:setNavigatorExitCallback(callback)
  self.navigatorExitCallback = callback
end

function CardCarouselCard:forceScrollNavigation(scrollDirection, duration)
  local forceScrollCallback = self.forceScrollCallback
  assert(forceScrollCallback ~= nil, "`forceScrollCallback` must be set by CardCarouselNavigator before cards can force navigation.")
  forceScrollCallback(scrollDirection, duration)
end

function CardCarouselCard:forceTapNavigation(scrollDirection, duration, willCheckNavigationConstraints)
  duration = duration or 0.5

  local forceTapNavigateCallback = self.forceTapNavigateCallback
  assert(forceTapNavigateCallback ~= nil, "`forceTapNavigateCallback` must be set by CardCarouselNavigator before cards can force navigation.")
  forceTapNavigateCallback(scrollDirection, duration, willCheckNavigationConstraints)
end

function CardCarouselCard:setForceScrollCallback(callback)
  self.forceScrollCallback = callback
end

function CardCarouselCard:setForceTapNavigateCallback(callback)
  self.forceTapNavigateCallback = callback
end

--- Determines whether the navigator is allowed to begin a manual scroll off of this card in the given direction
-- (when the navigator is already displaying this card)
function CardCarouselCard:isScrollAwayAllowed(scrollDirection)
  return true
end

-- Determines whether the navigator is allowed to begin a manual scroll onto this card from the given direction
-- (when the navigator is displaying a neighboring card, and the user tries to swipe onto this card)
function CardCarouselCard:isScrollEnterAllowed(scrollDirection)
  return true
end

--- Determines whether the navigator is allowed to begin a manual scroll onto this card from the given direction
-- *after* this card has already been seen for the first time.  Same as `isScrollEnterAllowed`, but is applied
-- only after the first time this card has been displayed.
function CardCarouselCard:isScrollBackAllowed(scrollDirection)
  return true
end

function CardCarouselCard:canScrollFromThisCard(scrollDirection)
  return self:isScrollAwayAllowed(scrollDirection)
end

function CardCarouselCard:canScrollToThisCard(scrollDirection)
  return self:isScrollEnterAllowed(scrollDirection) and
         ((not self:hasBeenSeen()) or self:isScrollBackAllowed(scrollDirection))
end

--- Determines whether the navigator is allowed to tap navigate off of this card in the given direction
-- (when the navigator is already displaying this card)
function CardCarouselCard:isTapNavigateAwayAllowed(tapDirection)
  return true
end

-- Determines whether the navigator is allowed to tap navigate onto this card from the given direction
-- (when the navigator is displaying a neighboring card, and the user tries to tap navigate onto this card)
function CardCarouselCard:isTapNavigateEnterAllowed(tapDirection)
  return true
end

--- Determines whether the navigator is allowed to tap navigate onto this card from the given direction
-- *after* this card has already been seen for the first time.  Same as `isTapNavigateEnterAllowed`, but is applied
-- only after the first time this card has been displayed.
function CardCarouselCard:isTapNavigateBackAllowed(tapDirection)
  return true
end

function CardCarouselCard:canTapNavigateFromThisCard(tapDirection)
  return self:isTapNavigateAwayAllowed(tapDirection)
end

function CardCarouselCard:canTapNavigateToThisCard(tapDirection)
  return self:isTapNavigateEnterAllowed(tapDirection) and
         ((not self:hasBeenSeen()) or self:isTapNavigateBackAllowed(tapDirection))
end

function CardCarouselCard:handlesOwnTapNavigation()
  return false
end

function CardCarouselCard:hasBeenSeen()
  return self.seen
end

function CardCarouselCard:isDynamic()
  return true
end

function CardCarouselCard:getLayer()
  return self.layer
end

function CardCarouselCard:reset()
  -- Do nothing.  Override in subclass.
end

function CardCarouselCard:onEnter()
  self.isActiveCard = true
  self.seen = true
  self:startAnimation()
  if self.enterCallback then
    self.enterCallback()
  end
end

function CardCarouselCard:onLeaveBegin()
  -- Do nothing.  Override in subclass.
end

function CardCarouselCard:onLeaveComplete()
  self.isActiveCard = false
  self:stopAnimation()
end

function CardCarouselCard:setScrollProgress(progress, isHorizontal)
  -- Do nothing.  Override in subclass.
end

function CardCarouselCard:createAnimationAction()
  -- Create nothing.  Override in subclass.
  return nil
end

function CardCarouselCard:startAnimation()
  local animationAction = self:createAnimationAction()
  if (animationAction ~= nil) then
    self.animationAction = animationAction:start()
  end
end

function CardCarouselCard:stopAnimation()
  if (self.animationAction ~= nil) then
    self.animationAction:stop()
    self.animationAction = nil
  end
end

function CardCarouselCard:overloadOnEnter(aFunction)
  local originalOnEnter = self.onEnter or function() end
  self.onEnter = function(self)
    originalOnEnter(self)
    aFunction(self)
  end
end

function CardCarouselCard:overloadOnLeaveBegin(aFunction)
  local originalOnLeaveBegin = self.onLeaveBegin or function() end
  self.onLeaveBegin = function(self)
    originalOnLeaveBegin(self)
    aFunction(self)
  end
end

function CardCarouselCard:overloadOnLeaveComplete(aFunction)
  local originalOnLeaveComplete = self.onLeaveComplete or function() end
  self.onLeaveComplete = function(self)
    originalOnLeaveComplete(self)
    aFunction(self)
  end
end

function CardCarouselCard:initializeLayer()
  if not self.layer then
    local viewport = Screen.getViewport()

    local layer = MOAILayer2D.new()
    layer:setViewport(viewport)
    layer:setSortMode(MOAILayer2D.SORT_PRIORITY_ASCENDING)
    Screen.addManagedLayer(layer)
    self.layer = layer

    self:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))
    layer:insertProp(self)
    self.parentLayer = layer
    layer.rootGroup = self
  end
end

function CardCarouselCard:getCardGlobals()
  return self.cardGlobals
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CardCarouselCard.new(cardType, cardGlobals)
  local self = Group.newFullScreen()
  Mixin.mix(self, CardCarouselCard)

  self.slideProgress = 0.0
  self.animationAction = nil
  self.forceScrollCallback = nil
  self.seen = false
  self.cardType = cardType
  self.cardGlobals = cardGlobals

  return self
end
