--- CardCarouselFieldMapping
-- @author Created by Aaron Barrett on May 04, 2015.

class("CardCarouselFieldMapping")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CardCarouselFieldMapping.REQUIRED = "required"
CardCarouselFieldMapping.OPTIONAL = "optional"
CardCarouselFieldMapping.XML = "xml"
CardCarouselFieldMapping.AUDIO = "audio"
CardCarouselFieldMapping.IMAGE = "image"
CardCarouselFieldMapping.ARRAY = "array"
CardCarouselFieldMapping.REPEATING = "repeating"

local INDEX_TEXT = 1
local INDEX_IMAGE = 2
local INDEX_AUDIO = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function hasAttribute(self, key, attribute)
  local fieldEntry = self.mapping[key]
  return fieldEntry.attributes[attribute]
end

local function isArrayValue(value)
  return value:find(",") or value:find("-")
end

local function isProperlyFormattedFieldSpecification(fieldSpecifier)
  return fieldSpecifier:find("[^,]+-[^,]+-") == nil and fieldSpecifier:find("[-,]$") == nil
end

local function tokenize(input)
  local tokens = {}
  local fieldMatcher = "%s*([^,-]+)"
  local operationMatcher = "%s*([,-])"
  local offset = 0
  local fieldName
  local operation
  while offset do
    _, offset, fieldName = input:find(fieldMatcher, offset + 1)
    if fieldName then
      table.insert(tokens, StringUtils.trim(fieldName))
      _, offset, operation = input:find(operationMatcher, offset + 1)
      if operation then
        table.insert(tokens, StringUtils.trim(operation))
      end
    end
  end
  return tokens
end

local function getFieldIndex(fieldName)
  local _, _, index = fieldName:find("(%d+)")
  return tonumber(index)
end

local function getFieldBase(fieldName)
  local _, _, base = fieldName:find("([^%d]+)")
  return base
end

local function extractFields(fieldSpecifier)
  assert(isProperlyFormattedFieldSpecification(fieldSpecifier), "field specifier \"" .. fieldSpecifier .. "\" is improperly formed")

  local lastField
  local isRange = false
  local isOperator = false
  local fields = {}
  for _, token in ipairs(tokenize(fieldSpecifier)) do
    if isOperator then
      if token == "," then
        table.insert(fields, lastField)
        isRange = false
        lastField = nil
      else
        isRange = true
      end
    else
      if isRange then
        local baseName = getFieldBase(lastField)
        local startIndex = getFieldIndex(lastField)
        local endIndex = getFieldIndex(token)
        for i = startIndex, endIndex do
          table.insert(fields, baseName .. tostring(i))
        end
        isRange = false
        lastField = nil
      else
        lastField = token
      end
    end
    isOperator = not isOperator
  end
  if lastField then
    table.insert(fields, lastField)
  end

  return fields
end

local function getNumericValueOrDefault(value, default)
  if not value or value == "" then
    return default
  end
  return tonumber(value)
end

local function parseNumericArray(value)
  if value == nil then
    return {}
  end

  local array = TableUtils.csvToArray(value)
  for index, value in ipairs(array) do
    array[index] = tonumber(value)
  end
  return array
end

local function getRowAdvance(self, cardEntry, key)
  local rowAdvanceArray = parseNumericArray(cardEntry.row_advance)
  while #rowAdvanceArray < 3 do
    table.insert(rowAdvanceArray, 1)
  end

  if self:isImage(key) then
    return rowAdvanceArray[INDEX_IMAGE]
  elseif self:isAudio(key) then
    return rowAdvanceArray[INDEX_AUDIO]
  end
  return rowAdvanceArray[INDEX_TEXT]
end

local function getRepeatingValue(self, cardEntry, key)
  local startingKey = self:getForeignKeys(key)[1]
  local baseKey = getFieldBase(startingKey)
  local index = getFieldIndex(startingKey)
  local rowAdvance = getRowAdvance(self, cardEntry, key)

  local values = {}
  while true do
    local nextKey = baseKey .. tostring(index)
    local nextValue = cardEntry[nextKey]
    if not nextValue or nextValue == "" then
      break
    end
    table.insert(values, nextValue)
    index = index + rowAdvance
  end

  return values
end

local function getNormalValue(self, cardEntry, key)
  local values = {}
  for _, foreignKey in ipairs(self:getForeignKeys(key)) do
    value = cardEntry[foreignKey]
    if not value or value == "" then
      break
    end
    table.insert(values, value)
  end
  return values
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CardCarouselFieldMapping:getAllKeys()
  return self.allKeys
end

function CardCarouselFieldMapping:getForeignKeys(key)
  return self.mapping[key].foreignKeys
end

function CardCarouselFieldMapping:getValues(cardEntry, key)
  if self:isRepeating(key) then
    return getRepeatingValue(self, cardEntry, key)
  end
  return getNormalValue(self, cardEntry, key)
end

function CardCarouselFieldMapping:exists(cardEntry, key)
  local value = self:getValues(cardEntry, key)[1]
  return value ~= nil and value ~= ""
end

function CardCarouselFieldMapping:isRequired(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.REQUIRED)
end

function CardCarouselFieldMapping:isOptional(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.OPTIONAL)
end

function CardCarouselFieldMapping:isXml(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.XML)
end

function CardCarouselFieldMapping:isAudio(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.AUDIO)
end

function CardCarouselFieldMapping:isImage(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.IMAGE)
end

function CardCarouselFieldMapping:isArray(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.ARRAY)
end

function CardCarouselFieldMapping:isRepeating(key)
  return hasAttribute(self, key, CardCarouselFieldMapping.REPEATING)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newFieldEntry(localKey, foreignKeys, attributes)
  if isArrayValue(foreignKeys) or attributes[CardCarouselFieldMapping.REPEATING] then
    attributes[CardCarouselFieldMapping.ARRAY] = true
  end
  return {
    localKey = localKey,
    foreignKeys = extractFields(foreignKeys),
    attributes = attributes
  }
end

local function newFieldEntryFromInput(inputField)
  local localKey = inputField[1]
  local foreignKeys = inputField[2]
  local attributes = {}
  local i = 3
  while true do
    local attribute = inputField[i]
    if attribute then
      attributes[attribute] = true
    else
      break
    end
    i = i + 1
  end
  return newFieldEntry(localKey, foreignKeys, attributes)
end

local function addFieldMapping(self, fieldEntry)
  local key = fieldEntry.localKey
  self.mapping[key] = fieldEntry
  table.insert(self.allKeys, key)
end

-- Constructor ------------------------

function CardCarouselFieldMapping.new(...)
  local self = {}
  Mixin.mix(self, CardCarouselFieldMapping)

  self.allKeys = {}
  self.mapping = {}

  for _, inputField in ipairs({...}) do
    addFieldMapping(self, newFieldEntryFromInput(inputField))
  end

  addFieldMapping(self, newFieldEntry("cardType", "card_type", {[CardCarouselFieldMapping.REQUIRED] = true}))

  return self
end
