--- CardCarouselTextView
-- @author Created by karl on April 10, 2015.

class("CardCarouselTextView")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CardCarouselTextView.ITEM_TYPE_DIVIDE = "divide"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createViewForUTF8Text(self, item, text)
  local label = Label.new(text, item:getFont())
  label:inheritColor(self)
  return label
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CardCarouselTextView:setText(text)
  if text ~= self:getText() then
    local textInterpreter = self.textInterpreter
    local textModel = self.textModel

    local items = textInterpreter:interpretXmlText(text)
    textModel:resetItems(items)

    self.text = text
  end
end

function CardCarouselTextView:getText()
  return self.text
end

function CardCarouselTextView:createItemView(item)
  local itemType = item:getItemType()

  local prop = nil
  if itemType == CardCarouselTextView.ITEM_TYPE_DIVIDE then
    prop = createViewForUTF8Text(self, item, "÷")
  else
    prop = self.super.createItemView(self, item)
  end

  local color = item:getAttribute("color")
  if prop and color then
    prop:clearAttrLink(MOAIColor.INHERIT_COLOR)
    prop:setColor(unpack(color))
    prop:inheritOpacity(self)
  end

  return prop
end

function CardCarouselTextView:getDimensionsAndCenterOfViewsWithAttribute(attribute, value)
  local views = self:getViewsWithAttribute(attribute, value)
  local minX, minY, maxX, maxY = GroupUtils.getBoundsOfGroups(unpack(views))
  local width, height = maxX - minX, maxY - minY
  local centerX, centerY = GroupUtils.getCenterOfRectangle(minX, minY, maxX, maxY)
  return width, height, centerX, centerY
end

function CardCarouselTextView:updateLayout()
  self.super.updateLayout(self)
  self.originalXPositions = {}
  for _, line in ipairs(self:getViewsByRow()) do
    local xValues = {}
    table.insert(self.originalXPositions, xValues)
    for _, view in ipairs(line) do
      table.insert(xValues, view:getX())
    end
  end
end

function CardCarouselTextView:prepareForSlideIn(initialStagger, staggerAmount)
  local stagger = initialStagger
  for iLine, line in ipairs(self:getViewsByRow()) do
    for iView, view in ipairs(line) do
      local originalX = self.originalXPositions[iLine][iView]
      view:setX(originalX + stagger)
    end
    stagger = stagger + staggerAmount
  end
end

function CardCarouselTextView:newSlideInAction()
  local slideAction = AHSpawn.new()
  local ease = AHEaseSineIn.new()
  for iLine, line in ipairs(self:getViewsByRow()) do
    for iView, view in ipairs(line) do
      local originalX = self.originalXPositions[iLine][iView]
      slideAction:addAction(AHLocationXTo.new(view, 1, ease, originalX))
    end
  end

  return slideAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CardCarouselTextView.new(text, width, attributes, horizontalAlignment, verticalAlignment)
  Asserts.assertType(text, "text", "string")
  Asserts.assertType(width, "width", "number")
  Asserts.assertType(attributes, "attributes", "table")
  Asserts.assertOptionalType(horizontalAlignment, "horizontalAlignment", "number")
  Asserts.assertOptionalType(verticalAlignment, "verticalAlignment", "number")

  -- NOTE: Couldn't just subclass DynamicTextArea because `createItemView` needs
  -- to be overriden before setText is called
  local self = TextView.new(width, horizontalAlignment, verticalAlignment, attributes.default.lineSpacing)
  Mixin.mix(self, CardCarouselTextView)

  local textInterpreter = TextInterpreter.new(attributes.default, attributes)
  local textModel = TextModel.new(textInterpreter:interpretXmlText(""))
  self:setModel(textModel)

  self.textInterpreter = textInterpreter
  self.textModel = textModel

  self:setText(text)

  return self
end
