--- KeyFrameInterpolator
-- @author Created by Jay Austria on September 30, 2014.

class("KeyFrameInterpolator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local DEFAULT_FILL_IN_VALUE = 0.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function hasValueDefined(keyFrame, valueToCheck)
  for value, _ in pairs(keyFrame.valueProbabilities) do
    if value == valueToCheck then
      return true
    end
  end

  return false
end

local function fillInKeyFrames(keyFrames)
  local allValues = {}
  for _, frame in ipairs(keyFrames) do
    for value, _ in pairs(frame.valueProbabilities) do
      allValues[value] = true
    end
  end

  for _, frame in ipairs(keyFrames) do
    for requiredValue, _ in pairs(allValues) do
      if not hasValueDefined(frame, requiredValue) then
        frame.valueProbabilities[requiredValue] = DEFAULT_FILL_IN_VALUE
      end
    end
  end

  return keyFrames
end

local function areZeroAndOneDefined(keyFrames)
  local hasZero = false
  local hasOne = false
  for _, keyFrame in ipairs(keyFrames) do
    if keyFrame.key == 0.0 then
      hasZero = true
    elseif keyFrame.key == 1.0 then
      hasOne = true
    end
  end

  return hasZero and hasOne
end

local function doesSumOfBucketProportionsEqualOne(keyFrames)
  for _, keyFrame in ipairs(keyFrames) do
    local proportionsSum = 0.0
    local expectedSum = 1.0

    for _, probability in pairs(keyFrame.valueProbabilities) do
      proportionsSum = proportionsSum + probability
    end

    local sumIsCorrect = math.abs(proportionsSum - expectedSum) < 0.0001
    if not sumIsCorrect then
      return false
    end
  end

  return true
end

local function areFramesSorted(keyFrames)
  for i = 1, #keyFrames - 1 do
    if keyFrames[i].key >= keyFrames[i + 1].key then
      return false
    end
  end
  return true
end

local function validateKeyFrames(keyFrames)
  assert(areZeroAndOneDefined(keyFrames), "Keyframes at 0.0 and 1.0 were not defined for the key frames")
  assert(doesSumOfBucketProportionsEqualOne(keyFrames), "A keyframe bucket's proportions did not add up to 1.0")
  assert(areFramesSorted(keyFrames), "Keyframes are not ordered by key")
end

local function getProbability(frame, requestedValue)
  for value, probability in pairs(frame.valueProbabilities) do
    if value == requestedValue then
      return probability
    end
  end

  error("No matching probability found in a frame with requested bucket name")
end

local function getInterpolatedDistribution(currentFrame, nextFrame, interpolationValue)
  local values = {}
  local probabilities = {}

  local keyCurrent = currentFrame.key
  local keyNext = nextFrame.key

  for value, probability in pairs(currentFrame.valueProbabilities) do
    local probabilityCurrent = getProbability(currentFrame, value)
    local probabilityNext = getProbability(nextFrame, value)

    local interpolatedProbability = MathUtils.interpolateLinear(interpolationValue,
                                                                keyCurrent,
                                                                keyNext,
                                                                probabilityCurrent,
                                                                probabilityNext)
    table.insert(values, value)
    table.insert(probabilities, interpolatedProbability)
  end

  return values, probabilities
end

local function getValuesAndWeights(keyFrames, interpolationValue)
  local frameCount = #keyFrames
  for index = 1, frameCount - 1 do
    local currentFrame = keyFrames[index]
    local nextFrame = keyFrames[index + 1]

    local isBetweenFrames = (interpolationValue >= currentFrame.key) and
                            (interpolationValue <= nextFrame.key)
    if isBetweenFrames then
      return getInterpolatedDistribution(currentFrame, nextFrame, interpolationValue)
    end
  end

  error("interpolationValue seems to be outside of all defined keys in the key frames.")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function KeyFrameInterpolator:nextValue()
  local logString = "Selecting a value at random from the following weights:\n"
  for i = 1, #self.values do
    if (self.weights[i] > 0) then
      logString = logString .. string.format("%-10s\t%.4f", self.values[i], self.weights[i]) .. "\n"
    end
  end

  local selectedValue = TableUtils.selectRandomValueWithWeights(self.values, self.weights)
  logString = logString .. "Randomly selected the value: " .. tostring(selectedValue)
  Log.debug(logString)

  return selectedValue
end

function KeyFrameInterpolator:getValues()
  return self.values
end

function KeyFrameInterpolator:getWeights()
  return self.weights
end

function KeyFrameInterpolator:getWeightForValue(value)
  local indexOfValue = TableUtils.keyForValue(self.values, value)
  local weight = self.weights[indexOfValue]

  return weight
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function KeyFrameInterpolator.new(keyFrames, interpolationValue)
  assert(interpolationValue ~= nil, "InterpolationValue not passed to KeyFrameInterpolator")

  local self = {}
  Mixin.mix(self, KeyFrameInterpolator)

  keyFrames = fillInKeyFrames(keyFrames)

  validateKeyFrames(keyFrames)

  local values, weights = getValuesAndWeights(keyFrames, interpolationValue)
  self.values = values
  self.weights = weights

  return self
end
