--- ConceptGenerator
-- Uses the concept chooser to intelligently choose new concepts.
-- Keeps track of previously seen concepts to ensure they're not seen again unless
-- the supply of concepts from the host is exhausted.
-- @author Created by karl on July 16, 2014.

class("ConceptGenerator")
require("data.generators.Generator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local KEY_DIFFICULTY_FILTER = "difficulty"

local EXHAUSTION_HANDLER_RECYCLE_CONCEPTS = function(self)
  self.excludedConceptIdentifiers = TableUtils.shallowCopy(self.perFetchSeenConceptIdentifiers)
  return true
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ConceptGenerator:next(count)
  self.perFetchSeenConceptIdentifiers = {}
  return self.super.next(self, count)
end

function ConceptGenerator:generateNextValue()
  if self.difficultyNameChooser then
    self.filters[KEY_DIFFICULTY_FILTER] = self.difficultyNameChooser:nextValue()
  end

  local allowFewerResults = true
  local concept = self.conceptChooser.nextConcepts(1, self.excludedConceptIdentifiers, self.filters, allowFewerResults)
  if self.markGeneratedConceptsAsSeen then
    if (concept ~= nil) then
      self:markConceptIdentifierAsSeen(concept.identifier)
    end
  end

  if concept then
    table.insert(self.excludedConceptIdentifiers, concept.identifier)
    table.insert(self.perFetchSeenConceptIdentifiers, concept.identifier)
  end

  return concept
end

function ConceptGenerator:reset()
  self.excludedConceptIdentifiers = {}
  self.perFetchSeenConceptIdentifiers = {}
end

--- If true, allow the generator to return less concepts than requested.
-- If false, throw an error if not enough concepts are available.
function ConceptGenerator:setAllowFewerResults(allowFewerResults)
  if allowFewerResults then
    self:setExhaustionHandler(Generator.EXHAUSTION_HANDLER_GIVE_UP)
  else
    self:setExhaustionHandler(EXHAUSTION_HANDLER_RECYCLE_CONCEPTS)
  end
end

--- Filters tell the concept chooser to choose from specific buckets.
-- @param filters (Table) - The filters to apply when generating concepts.
function ConceptGenerator:setFilters(filters)
  self.filters = TableUtils.shallowCopy(filters or {})
end

function ConceptGenerator:setDifficultyNameChooser(difficultyNameChooser)
  self.difficultyNameChooser = difficultyNameChooser
end

function ConceptGenerator:setDifficulty(difficultyName)
  self.filters[KEY_DIFFICULTY_FILTER] = difficultyName
end

function ConceptGenerator:setMarkGeneratedConceptsAsSeen(markGeneratedConceptsAsSeen)
  self.markGeneratedConceptsAsSeen = markGeneratedConceptsAsSeen
end

function ConceptGenerator:markConceptIdentifierAsSeen(conceptIdentifier)
  if (conceptIdentifier ~= nil)then
    self.conceptChooser.markSeenConceptIdentifiers({conceptIdentifier})
  end
end

function ConceptGenerator:removeConceptIdentifierFromExclusionList(conceptIdentifier)
  TableUtils.removeValueFromArray(self.excludedConceptIdentifiers, conceptIdentifier)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make a new concept generator.
-- By default, "allowFewerResults" is false.
-- @param conceptChooser (ConceptChooser) - GameSession's conceptChooser.
function ConceptGenerator.new(conceptChooser)
  local self = Generator.new()
  Mixin.mix(self, ConceptGenerator)

  self.conceptChooser = conceptChooser
  self.difficultyNameChooser = nil
  self.filters = {}
  self.markGeneratedConceptsAsSeen = true

  self:setAllowFewerResults(false)
  self:reset()

  return self
end
