--- Generator superclass.
-- Subclasses must implement reset and generateNextValue.
-- @author Created by karl on June 06, 2014.

require("Mixin")

class("Generator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Strategies are functions taking a self parameter and returning true if the operation should be retried.

-- Common exhaustion strategies

-- On exhaustion, reset the generator and retry.
Generator.EXHAUSTION_HANDLER_RESET = function(self)
  self:reset()
  return true
end

-- On exhaustion, throw an error.
Generator.EXHAUSTION_HANDLER_ERROR = function(self)
  error("Generator has been exhausted")
end

-- On exhaustion, give up and return less values.
Generator.EXHAUSTION_HANDLER_GIVE_UP = function(self)
  return false
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Set the handler to call when the generator has been exhausted.
-- Handler takes a "self" parameter and returns true if the operation should be retried.
-- The initial handler resets the generator and retries.
-- @param onExhaustion (function with param self) - Called whenever the generator is exhausted.
function Generator:setExhaustionHandler(onExhaustion)
  self.onExhaustion = onExhaustion
end

--- Make an iterator suitable for use in a for loop.
--
-- Usage:
--   for element in aGenerator:iterator() do
--     do stuff
--   end
--
-- Warning: Be sure to set a suitable exhaustion handler before calling this!
--          Remember that the default handler resets the generator, giving the
--          potential for an infinite loop!
function Generator:iterator()
  return function() return self:next() end
end

--- Generate more values.
-- When generating multiple values, you may want to wrap the result in {} to make an array.
-- @param count The number of values to generate.
-- @return a list of values (not in a table).
function Generator:next(count)
  count = count or 1

  local hasRetried = false
  local result = {}

  while #result < count do
    local value = self:generateNextValue()
    if value ~= nil then
      table.insert(result, value)
    else
      local shouldRetry = self.onExhaustion(self)
      if not shouldRetry then
        break
      end
      if hasRetried then
        error("Generator failed to generate a value after retrying")
      end
      hasRetried = true
    end
  end

  return unpack(result)
end

--- Generate more values, returning as an array.
-- @param count The number of values to generate.
-- @return an array of values.
function Generator:nextAsArray(count)
  return {self:next(count)}
end

--- Reset the generator.
function Generator:reset()
  error("Implement me")
end

--- Generate the next value.
-- FOR INTERNAL USE ONLY.
function Generator:generateNextValue()
  error("Implement me")
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Construct a generator.
-- @param exhaustionStrategy (function with param self) - Called whenever the generator is exhausted (defaults to reset).
function Generator.new(exhaustionStrategy)
  exhaustionStrategy = exhaustionStrategy or Generator.EXHAUSTION_HANDLER_RESET

  local self = {}
  Mixin.mix(self, Generator)
  self:setExhaustionHandler(exhaustionStrategy)

  return self
end
