--LearningFocusedRoundGenerator
-- Repeats concepts that are marked as incorrect
-- to encourage users to learn.

class("LearningFocusedRoundGenerator")

require("ExperimentsConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_IDENTIFIER_KEY = "identifier"
local DEFAULT_PROGRESS_BEFORE_RESURFACING_MISTAKES = 0.4
local DEFAULT_REMAINING_BEFORE_RESURFACING_MISTAKES = 1.0 - DEFAULT_PROGRESS_BEFORE_RESURFACING_MISTAKES

LearningFocusedRoundGenerator.DEFAULT_PROGRESS_BEFORE_RESURFACING = DEFAULT_PROGRESS_BEFORE_RESURFACING_MISTAKES
LearningFocusedRoundGenerator.DEFAULT_REMAINING_BEFORE_RESURFACING = DEFAULT_REMAINING_BEFORE_RESURFACING_MISTAKES


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getMostRecentlyServedConceptIdentifier(self)
  if (#self.shownConceptIdentifers > 0) then
    return self.shownConceptIdentifers[#self.shownConceptIdentifers]
  end
  return nil
end

local function getMistakeIdentifierToResurface(self)
  if self.areMistakesAllowedToResurface then
    for _, conceptIdentifier in ipairs(self.mistakenConceptIdentifiers) do
      local timesShown = self.conceptIdentifierToTimesShown[conceptIdentifier]
      local canShowAgain = (timesShown <= self.maximumTimesConceptsCanResurface)
      local mostRecentConceptIdentifier = getMostRecentlyServedConceptIdentifier(self)
      local isMostRecentConcept = (conceptIdentifier == mostRecentConceptIdentifier)
      if (canShowAgain and not isMostRecentConcept) then
        return conceptIdentifier
      end
    end
  end
  return nil
end

local function incrementTimesShownCount(self, conceptIdentifier)
  if self.conceptIdentifierToTimesShown[conceptIdentifier] == nil then
    self.conceptIdentifierToTimesShown[conceptIdentifier] = 0
  end
  self.conceptIdentifierToTimesShown[conceptIdentifier] = self.conceptIdentifierToTimesShown[conceptIdentifier] + 1
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LearningFocusedRoundGenerator:next()
  local round = nil
  local conceptIdentifier = getMistakeIdentifierToResurface(self)
  if (conceptIdentifier ~= nil) then
    TableUtils.removeValueFromArray(self.mistakenConceptIdentifiers, conceptIdentifier)
    round = self.conceptIdentiferToRound[conceptIdentifier]
  else
    round = self.conceptGenerator:next()
    conceptIdentifier = round[self.conceptIdentifierKey]
    assert(conceptIdentifier ~= nil, "A round identifier could not be found in the " ..
                                   "generated round object. Has a correct " ..
                                   "identifierKey parameter been provided?")
    self.conceptIdentiferToRound[conceptIdentifier] = round
  end
  table.insert(self.shownConceptIdentifers, conceptIdentifier)
  incrementTimesShownCount(self, conceptIdentifier)
  return round
end

function LearningFocusedRoundGenerator:allowMistakesToResurface()
  self.areMistakesAllowedToResurface = true
end

function LearningFocusedRoundGenerator:setAreMistakesAllowedToResurface(areAllowed)
  self.areMistakesAllowedToResurface = areAllowed
end

function LearningFocusedRoundGenerator:notifyMistake(conceptIdentifier)
  table.insert(self.mistakenConceptIdentifiers, conceptIdentifier)
end

function LearningFocusedRoundGenerator:clearMistakes()
  self.shownConceptIdentifers = {}
  self.mistakenConceptIdentifiers = {}
  self.conceptIdentiferToRound = {}
  self.conceptIdentifierToShownCount = {}
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make a new concept generator.
-- @param conceptGenerator (Generator) - Something like Generator.lua that
--                                       supports next()
-- @param identifierKey (string, optional) - The key to use when accessing the
--                                           conceptIdfrom the generated round
-- @param maximumTimesConceptsCanResurface (int, optional) - The number of times a concept can resurface
--                                                           after being answered incorrectly
function LearningFocusedRoundGenerator.new(conceptGenerator, identifierKey, maximumTimesConceptsCanResurface)
  local self = {}
  Mixin.mix(self, LearningFocusedRoundGenerator)

  Asserts.assertOptionalType(identifierKey, "identifierKey", "string")

  if (identifierKey == nil) then
    identifierKey = DEFAULT_IDENTIFIER_KEY
  end

  self.conceptGenerator = conceptGenerator

  self.areMistakesAllowedToResurface = false
  self.shownConceptIdentifers = {}
  self.mistakenConceptIdentifiers = {}
  self.conceptIdentiferToRound = {}
  self.conceptIdentifierToTimesShown = {}
  self.maximumTimesConceptsCanResurface = maximumTimesConceptsCanResurface or math.huge
  self.conceptIdentifierKey = identifierKey

  return self
end
