--- Various number generators
-- @author Created by karl on June 06, 2014.

require("data.generators.Generator")

-- ----------------------------------------------------------------------------
-- Sequential Integer Generator
-- ----------------------------------------------------------------------------

class("SequentialNumberGenerator")
local function isExhausted(self)
  if self._stepSize < 0 then
    return self._nextValue < self._endValue
  end
  return self._nextValue > self._endValue
end

function SequentialNumberGenerator:generateNextValue()
  if isExhausted(self) then
    return nil
  end

  local value = self._nextValue
  self._nextValue = self._nextValue + self._stepSize
  return value
end

--- Reset the next value to min value.
function SequentialNumberGenerator:reset()
  self._nextValue = self._startValue
end

--- Construct a new generator.
-- @param startValue (number) - The start value. The generator will return this value first.
-- @param endValue (number) - The "end" value. The generator will return values until the end value is passed.
-- @param stepSize (number, default 1) - The amount to add to the current value to generate the next value. Can be negative.
function SequentialNumberGenerator.new(startValue, endValue, stepSize)
  stepSize = stepSize or 1

  local self = Generator.new()
  Mixin.mix(self, SequentialNumberGenerator)

  self._startValue = startValue
  self._endValue = endValue
  self._stepSize = stepSize
  self:reset()

  return self
end


-- ----------------------------------------------------------------------------
-- Random Integer Generator
-- ----------------------------------------------------------------------------

class("RandomIntegerGenerator")
function RandomIntegerGenerator:generateNextValue()
  if self._allowRepeat then
    return math.random(self._minValue, self._maxValue)
  end

  local nextValue = self._lastValue
  while nextValue == self._lastValue do
    nextValue = math.random(self._minValue, self._maxValue)
  end

  self._lastValue = nextValue
  return nextValue
end

function RandomIntegerGenerator:reset()
  -- Nothing to do.
end

--- Construct a new generator.
-- @param minValue (integer) - The minimum value allowed.
-- @param maxValue (integer) - The maximum value allowed.
-- @param allowRepeat (boolean, default true) - If true, allow the generator to return the same value twice in a row.
function RandomIntegerGenerator.new(minValue, maxValue, allowRepeat)
  if allowRepeat == nil then allowRepeat = true end

  local self = Generator.new()
  Mixin.mix(self, RandomIntegerGenerator)

  self._minValue = minValue
  self._maxValue = maxValue
  self._allowRepeat = allowRepeat
  self:reset()

  return self
end


-- ----------------------------------------------------------------------------
-- Random Boolean Generator
-- ----------------------------------------------------------------------------

class("RandomBooleanGenerator")
function RandomBooleanGenerator:generateNextValue()
  return MathUtils.getRandomBoolean()
end

function RandomBooleanGenerator:reset()
  -- Nothing to do.
end

--- Construct a new generator.
function RandomBooleanGenerator.new()
  local self = Generator.new()
  Mixin.mix(self, RandomBooleanGenerator)

  return self
end
