-- @author Created by karl on June 09, 2014.

require("data.generators.Generator")


-- ============================================================================
-- Transformation Generator
-- ============================================================================

--- Transforms the output from another generator.
class("TransformationGenerator")
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransformationGenerator:isExhausted()
  return self.generator:isExhausted()
end

function TransformationGenerator:reset()
  self.generator:reset()
end

function TransformationGenerator:generateNextValue()
  local value = self.generator:next()
  return self.transformFunction(value)
end

function TransformationGenerator:setTransformFunction(transformFunction)
  self.transformFunction = transformFunction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a transformation generator.
-- @param generator (Generator) - The generator to draw values from.
-- @param transformFunction (function with param value) - Transforms the generated value.
function TransformationGenerator.new(generator, transformFunction)
  local self = Generator.new(generator.onExhaustion)
  Mixin.mix(self, TransformationGenerator)

  self.generator = generator
  self.transformFunction = transformFunction

  return self
end


-- ============================================================================
-- Demultiplexing Generator
-- ============================================================================

--- Accepts input from multiple generators, and outputs a single value.
class("DemultiplexingGenerator")
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function DemultiplexingGenerator:isExhausted()
  for _, generator in ipairs(self.generators) do
    if generator:isExhausted() then
      return true
    end
  end
  return false
end

function DemultiplexingGenerator:reset()
  for _, generator in ipairs(self.generators) do
    generator:reset()
  end
end

function DemultiplexingGenerator:generateNextValue()
  local values = {}
  for _, generator in ipairs(self.generators) do
    local value = generator:next()
    if value == nil then
      return nil
    end
    table.insert(values, value)
  end

  return self.demultiplexFunction(unpack(values))
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a demultiplexing generator.
-- The demultiplex function must accept as many parameters as there are generators.
-- @param generators (array of Generator) - The generators to draw values from.
-- @param demultiplexFunction (function with multiple params) - Demultiplexes the generated values.
function DemultiplexingGenerator.new(generators, demultiplexFunction)
  local self = Generator.new()
  Mixin.mix(self, DemultiplexingGenerator)

  self.generators = generators
  self.demultiplexFunction = demultiplexFunction

  return self
end


-- ============================================================================
-- Maximum Run Generator
-- ============================================================================

--- Only allows a maximum number of identical results from its sub-generator.
class("MaximumRunGenerator")
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MaximumRunGenerator:isExhausted()
  return self.generator:isExhausted()
end

function MaximumRunGenerator:reset()
  self.generator:reset()
end

function MaximumRunGenerator:generateNextValue()
  local value = self.generator:next()
  if value == nil then
    return nil
  end

  while value == self.lastValue do
    self.currentRunLength = self.currentRunLength + 1
    if self.currentRunLength <= self.maxRunLength then
      return value
    end

    value = self.generator:next()
    if value == nil then
      return nil
    end
  end

  self.lastValue = value
  self.currentRunLength = 1
  return value
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a maximum run generator.
-- @param generator (Generator) - The generator to draw values from.
-- @param maxRunLength (integer) - The maximum number of identical results in a row to allow.
--                                 For example, 2 means a result of "a", "a" is allowed, but
--                                 "a", "a", "a" is not.
function MaximumRunGenerator.new(generator, maxRunLength)
  local self = Generator.new(generator.onExhaustion)
  Mixin.mix(self, MaximumRunGenerator)

  self.generator = generator
  self.maxRunLength = maxRunLength
  self.currentRunLength = 1

  return self
end


-- ============================================================================
-- Caching Generator
-- ============================================================================

--- Caches the last n items generated by its wrapped generator.
class("CachingGenerator")
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CachingGenerator:reset()
  self.generator:reset()
  self.cache = {}
end

function CachingGenerator:generateNextValue()
  local value = self.generator:next()
  if value then
    table.insert(self.cache, 1, value)
  end
  while #self.cache > self.cacheSize do
    table.remove(self.cache)
  end
  return value
end

--- Get a cached value.
-- @param age (integer) How far back in the cache to look (0 = most recent value).
function CachingGenerator:getCachedValue(age)
  return self.cache[age + 1]
end

function CachingGenerator:setCacheSize(cacheSize)
  self.cacheSize = cacheSize
end

function CachingGenerator:getCacheSize()
  return self.cacheSize
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a CachingGenerator.
-- @param generator (Generator) - The wrapped generator.
-- @param cacheSize (integer) - Number of values to cache (default 1).
function CachingGenerator.new(generator, cacheSize)
  cacheSize = cacheSize or 1
  Asserts.assertType(generator, "generator", "table")
  Asserts.assertType(cacheSize, "cacheSize", "number")
  local self = Generator.new()
  Mixin.mix(self, CachingGenerator)

  self.generator = generator
  self.cacheSize = cacheSize
  self.cache = {}

  return self
end


-- ============================================================================
-- Delayed Generator
-- ============================================================================

--- Returns values from a caching generator delayed by the specified amount.
-- Note: This "generator" is entirely dependent upon the caching generator.
--       You must call next() each time on the caching generator before the
--       dependent generators will be able to generate their next value.
class("DelayedGenerator")
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function DelayedGenerator:reset()
  self.generator:reset()
end

function DelayedGenerator:generateNextValue()
  return self.generator:getCachedValue(self.delay) or self.defaultValue
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a DelayedGenerator.
-- @param generator (CachingGenerator) - The wrapped caching generator.
-- @param delay (integer) - How many values to delay by (example: 0 = get most recently generated value).
-- @param defaultValue (any) - The default value to return while there are insufficient values in the cache.
function DelayedGenerator.new(generator, delay, defaultValue)
  Asserts.assertType(generator, "generator", "table")
  Asserts.assertType(delay, "delay", "number")
  assert(delay >= 0, "delay must be 0 or greater")
  local self = Generator.new()
  Mixin.mix(self, DelayedGenerator)

  if generator:getCacheSize() < delay then
    generator:setCacheSize(delay)
  end

  self.generator = generator
  self.delay = delay
  self.defaultValue = defaultValue

  return self
end
