--- Various value generators.
-- @author Created by karl on June 06, 2014.

require("data.generators.Generator")


-- ----------------------------------------------------------------------------
-- SingleValueGenerator
-- ----------------------------------------------------------------------------

class("SingleValueGenerator")
function SingleValueGenerator:generateNextValue()
  return self.value
end

function SingleValueGenerator:reset()
end

function SingleValueGenerator:getValue()
  return self.value
end

function SingleValueGenerator:setValue(value)
  self.value = value
end

function SingleValueGenerator:newValueToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getValue, self.setValue, ease, amount)
end

function SingleValueGenerator:newValueByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getValue, self.setValue, ease, amount)
end

--- Construct a new generator.
-- This generator will repeatedly return the specified value.
-- @param value (anything) - Value to return.
function SingleValueGenerator.new(value)
  local self = Generator.new()
  Mixin.mix(self, SingleValueGenerator)

  self.value = value

  return self
end


-- ----------------------------------------------------------------------------
-- SequentialValueGenerator
-- ----------------------------------------------------------------------------

class("SequentialValueGenerator")
function SequentialValueGenerator:generateNextValue()
  if self.index >= #self.sourceValues then
    return nil
  end

  self.index = self.index + 1
  return self.sourceValues[self.index]
end

function SequentialValueGenerator:reset()
  self.index = 0
end

--- Construct a new generator.
-- This generator will sequentially produce values from a source array.
-- @param sourceValues (array) - Values to produce from this generator.
function SequentialValueGenerator.new(sourceValues)
  assert(type(sourceValues) == "table", "sourceValues must be an array (got " .. type(sourceValues) .. ").")
  local self = Generator.new()
  Mixin.mix(self, SequentialValueGenerator)

  self.sourceValues = TableUtils.shallowCopy(sourceValues)
  self:reset()

  return self
end


-- ----------------------------------------------------------------------------
-- RandomlySelectedValueGenerator
-- ----------------------------------------------------------------------------

class("RandomlySelectedValueGenerator")
function RandomlySelectedValueGenerator:generateNextValue()
  if #self.workingValues == 0 then
    return nil
  end

  local index = math.random(1, #self.workingValues)
  local value = self.workingValues[index]
  if not self.isWithReplacement then
    table.remove(self.workingValues, index)
  end
  return value
end

function RandomlySelectedValueGenerator:reset()
  self.workingValues = TableUtils.shallowCopy(self.sourceValues)
end

--- Construct a new generator.
-- This generator will randomly select values, with or without replacement.
-- @param sourceValues (array or set) - Values to produce from this generator.
-- @param isWithReplacement (boolean, default false) - If true, replace used values.
function RandomlySelectedValueGenerator.new(sourceValues, isWithReplacement)
  assert(type(sourceValues) == "table", "sourceValues must be an array (got " .. type(sourceValues) .. ").")
  local self = Generator.new()
  Mixin.mix(self, RandomlySelectedValueGenerator)

  self.isWithReplacement = isWithReplacement
  local isArray = #sourceValues > 0 and sourceValues[1] ~= true
  if isArray then
    self.sourceValues = TableUtils.shallowCopy(sourceValues)
  else
    self.sourceValues = TableUtils.getKeys(sourceValues)
  end
  self:reset()

  return self
end

-- ----------------------------------------------------------------------------
-- WeightedRandomlySelectedValueGenerator
-- ----------------------------------------------------------------------------

class("WeightedRandomlySelectedValueGenerator")

function WeightedRandomlySelectedValueGenerator:generateNextValue()
  if #self.workingValues == 0 then
    return nil
  end

  local selectedValue, valueIndex = TableUtils.selectRandomValueWithWeights(self.workingValues, self.workingWeights)

  if not self.isWithReplacement then
    table.remove(self.workingValues, valueIndex)
    table.remove(self.workingWeights, valueIndex)
  end

  return selectedValue
end

function WeightedRandomlySelectedValueGenerator:reset()
  self.workingValues = TableUtils.shallowCopy(self.sourceValues)
  self.workingWeights = TableUtils.shallowCopy(self.valueWeights)
end

--- Construct a new generator.
-- This generator will randomly select values with weighting, with or without replacement.
-- @param sourceValues (array) - Values to produce from this generator.
-- @param valueWeights (array) - Weights associated with each value in the sourceValues
-- @param isWithReplacement (boolean, default false) - If true, replace used values.
function WeightedRandomlySelectedValueGenerator.new(sourceValues, valueWeights, isWithReplacement)
  local self = Generator.new()
  Mixin.mix(self, WeightedRandomlySelectedValueGenerator)

  self.isWithReplacement = isWithReplacement
  self.sourceValues = TableUtils.shallowCopy(sourceValues)
  self.valueWeights = TableUtils.shallowCopy(valueWeights)
  self:reset()

  return self
end
