--- KeyframeSet
-- @author Created by andrewc on February 24, 2021.

class("KeyframeSet")

require("data.keyframes.Keyframe")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local KeyframeSetInterpolated = privateClass()
local KeyframeSetThreshold = privateClass()

-- ----------------------------------------------------------------------------
-- Validation
-- ----------------------------------------------------------------------------

local function validateKeyframesAreInOrder(keyframes)
  assert(#keyframes >= 1, "KeyframeSets must have at least one entry.")
  local expectedFrameNumberSize = TableUtils.getSize(keyframes[1]:getFrameNumbers())
  for i = 2, #keyframes do
    local keyframe = keyframes[i]
    local keyframeNumberSize = TableUtils.getSize(keyframe:getFrameNumbers())
    assert(keyframeNumberSize == expectedFrameNumberSize, "Keyframe frameNumbers do not match in size.")
    local previousKey = keyframes[i - 1]:getKey()
    assert(previousKey < keyframe:getKey(), "keyframeOrderedList keys are not in order.")
  end
end

local function validateInterpolatedSet(keyframes)
  assert(#keyframes >= 2, "Interpolated KeyframeSets must have at least two entries.")
  validateKeyframesAreInOrder(keyframes)
end

local function validateThresholdSet(keyframes)
  validateKeyframesAreInOrder(keyframes)
end

local function validateKeyForInterpolatedSet(keyframes, inputKey)
  assert(type(inputKey) == "number", "Input Key for new frame must be a number.")
  assert(inputKey >= keyframes[1]:getKey(), "Key for interpolation must be equal to or greater than the first keyframe key.")
  assert(inputKey <= keyframes[#keyframes]:getKey(), "Key for interpolation must be equal to or less than the last keyframe key.")
end

local function validateKeyForThresholdSet(keyframes, inputKey)
  assert(type(inputKey) == "number", "Input Key for new frame must be a number.")
  assert(inputKey >= keyframes[1]:getKey(), "Key for threshold must be equal to or greater than the first key.")
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createInterpolatedFrameNumbers(inputKey, previousKeyFrame, nextKeyFrame)
  local newKeyframe = {}
  local previousKey = previousKeyFrame:getKey()
  local nextKey = nextKeyFrame:getKey()
  for frameNumberKey, _ in pairs(previousKeyFrame:getFrameNumbers()) do
    local previousNumber = previousKeyFrame:getFrameNumbers()[frameNumberKey]
    local nextNumber = nextKeyFrame:getFrameNumbers()[frameNumberKey]
    local interpolatedValue = MathUtils.interpolateLinear(inputKey, previousKey, nextKey, previousNumber, nextNumber)
    newKeyframe[frameNumberKey] = interpolatedValue
  end
  return newKeyframe
end

local function getFrameNumbersWithLinearInterpolation(keyframes, inputKey)
  local previousKeyFrame = nil
  local nextKeyFrame = nil
  for _, keyframe in ipairs(keyframes) do
    if (inputKey == keyframe:getKey()) then
      return keyframe:getFrameNumbers()
    end

    if (inputKey > keyframe:getKey()) then
      previousKeyFrame = keyframe
    elseif (nextKeyFrame == nil) then
      nextKeyFrame = keyframe
      break
    end
  end
  return createInterpolatedFrameNumbers(inputKey, previousKeyFrame, nextKeyFrame)
end

local function getFrameNumbersWithLowerThreshold(keyframes, inputKey)
  for keyframeIndex, keyframe in ipairs(keyframes) do
    if (keyframe:getKey() > inputKey) then
      local previousKeyframe = keyframes[keyframeIndex - 1]
      return previousKeyframe:getFrameNumbers()
    end
  end
  local lastKeyframe = keyframes[#keyframes]
  return lastKeyframe:getFrameNumbers()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function KeyframeSetInterpolated:getFrameAt(inputKey)
  validateKeyForInterpolatedSet(self.keyframes, inputKey)
  return getFrameNumbersWithLinearInterpolation(self.keyframes, inputKey)
end

function KeyframeSetThreshold:getFrameAt(inputKey)
  validateKeyForThresholdSet(self.keyframes, inputKey)
  return getFrameNumbersWithLowerThreshold(self.keyframes, inputKey)
end

function KeyframeSet:getFrameAt(inputKey)
  error("Subclasses should implement this method.")
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newKeyframeSet(keyframes)
  local self = {}
  Mixin.mix(self, KeyframeSet)

  self.keyframes = keyframes
  return self
end

-- Constructors ------------------------

function KeyframeSet.newInterpolated(...)
  local self = newKeyframeSet(arg)
  Mixin.mix(self, KeyframeSetInterpolated)
  validateInterpolatedSet(self.keyframes)
  return self
end

function KeyframeSet.newThreshold(...)
  local self = newKeyframeSet(arg)
  Mixin.mix(self, KeyframeSetThreshold)
  validateThresholdSet(self.keyframes)
  return self
end