--- A sine wave data visualization.

class("SineWaveVisualizer")
require("data.visualization.DataVisualizer")

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

SineWaveVisualizer.PENALTY_ANIMATION_DURATION = 0.9

local DEFAULT_COLOR = {1, 1, 1, 1}
local PENALTY_COLOR = {1, 0, 0, 1}
local REWARD_COLOR = {125/255, 249/255, 201/255, 1}
local DEFAULT_NUMBER_OF_PEAKS = 3
local INCREASED_NUMBER_OF_PEAKS = 10
local DECREASED_NUMBER_OF_PEAKS = 0
local DEFAULT_OPACITY_MULTIPLIER = 1.0
local WAVEFORM_ADVANCEMENT_MULTIPLIER = 10.0
local DEFAULT_LINE_THICKNESS = 0.5


------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function setYAtIndex(self, i, y)
  local x = self.points[i][1]
  self.points[i][2] = y
end

local function applyProgress(self, progress, normalizedSamplePower)
  local desiredHeight = math.pow(normalizedSamplePower, 10) * self.maxHeight
  local adjustedDesiredHeight = math.min(desiredHeight + self.heightOverflow, self.maxHeight)
  local newHeight = adjustedDesiredHeight

  local heightChange = adjustedDesiredHeight - self.height
  local maxHeightChange = 1.5
  local heightChangeMultiplier = math.abs(heightChange)/maxHeightChange

  if(heightChangeMultiplier > 1) then
    local cappedHeightChange = heightChange/heightChangeMultiplier
    heightChange = cappedHeightChange
    newHeight = self.height + heightChange
    -- Take the previous frame adjustment in just a little bit, for further amortization
    self.heightOverflow = heightChange - cappedHeightChange
    -- Alternative, single-frame antialising. Safer, but marginally less smooth
    -- self.heightOverflow = desiredHeight - newHeight
  else
    self.heightOverflow = 0
  end

  self.height = newHeight

  local phase = 2 * progress * math.pi

  local startPoint, endPoint = self.leftEnd, self.rightEnd
  local totalPoints = endPoint - startPoint

  local numberOfPeaks = self.numberOfPeaks
  local phaseMultiplier = totalPoints/numberOfPeaks / math.pi * 2
  local easedStep = math.pow(normalizedSamplePower, 30)
  local middle = (self.rightEnd - self.leftEnd)/2
  local reNormalizedSamplePower = middle

  local offsetY = self:getHeightPart(0.5)

  for i = math.floor(startPoint), math.floor(endPoint) do
    local x = self.points[i][1]

    local distanceFromStep = (i - startPoint) - reNormalizedSamplePower
    local normalizedDistance = distanceFromStep/middle
    local heightAdjustment = math.pow(normalizedDistance, 2)

    local penaltyAdjustment = 0
    if (self.isPenalty) then
      local distFromEdges = 1 - math.abs(normalizedDistance)
      penaltyAdjustment = math.sin(x - progress) * distFromEdges * newHeight/5
      penaltyAdjustment = penaltyAdjustment * normalizedSamplePower
    end
    local y = math.sin(x / phaseMultiplier - progress) * newHeight * (1 - heightAdjustment) + penaltyAdjustment * self.penaltyMultiplier
    y = y * self.yScale + offsetY
    self.points[i][2] = y
  end

  for i = 1, math.floor(startPoint) do
    setYAtIndex(self, i, offsetY)
  end
  for i = math.floor(endPoint), #self.points do
    setYAtIndex(self, i, offsetY)
  end
end

local function applyMorph(self, progress, normalizedSamplePower)
  local dt = 0.05 * (math.pow(self.morphAmount, 0.5) + 0.5)
  self.morphAmount = math.min(self.morphAmount + dt, 1)
  local t = self.morphAmount
  local p = 1 - t
  local targetPoints = self.morphTarget
  for i, point in ipairs(self.points) do
    local x, y = point[1], point[2]
    local tx, ty = targetPoints[i][1], targetPoints[i][2]
    local dx, dy = tx - x, ty - y

    local xm = point[1] * p + tx * t
    local ym = point[2] * p + ty * t
  end
end

local function getRubberBandAnimation(self, peaksChangedHandler, minPeaks, expectedPeaks, maxPeaks)
  local setPeaksAndPenalty = function(self, peaks)
    self:setNumberOfPeaks(peaks)
    if (peaksChangedHandler) then
      peaksChangedHandler(peaks, minPeaks, expectedPeaks, maxPeaks)
    end
  end

  local compressEase = MOAIEaseType.EASE_IN
  local decompressEase = MOAIEaseType.EASE_IN
  local normalizeEase = nil

  local compress = AHNonAttributeTo.new(self, 0.3, self.getNumberOfPeaks, setPeaksAndPenalty, compressEase, maxPeaks)
  local decompress = AHNonAttributeTo.new(self, 0.5, self.getNumberOfPeaks, setPeaksAndPenalty, decompressEase, minPeaks)
  local normalize = AHNonAttributeTo.new(self, 0.1, self.getNumberOfPeaks, setPeaksAndPenalty, normalizeEase, expectedPeaks)

  local animation = AHSequence.new(compress, decompress, normalize)
  return animation
end

local function cancelPenaltyAnimation(self)
  if (self.penaltyAnimation) then
    self.penaltyAnimation:stop()
    self.penaltyAnimation = nil
  end
end


------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

-- Getters / Setters ------------------

function SineWaveVisualizer:getNumberOfPeaks()
  return self.numberOfPeaks
end

function SineWaveVisualizer:setNumberOfPeaks(n)
  self.numberOfPeaks = n
end

function SineWaveVisualizer:getTotalPointCount()
  return #self.points
end

function SineWaveVisualizer:setPenaltyColor(color)
  self.penaltyColor = ColorUtils.colorWithOpacityMultiplier(color, self.opacityMultiplier)
  self:setPenaltyMultiplier(self.penaltyMultiplier)
end

function SineWaveVisualizer:getPenaltyMultiplier()
  return self.penaltyMultiplier
end

function SineWaveVisualizer:setPenaltyMultiplier(multiplier)
  multiplier = math.max(multiplier, 0)
  local currentColor = {self:getColor()}
  local color = ColorUtils.makeColorBetweenTwoColors(currentColor, self.penaltyColor, multiplier)
  self:setColor(unpack(color))
  self.penaltyMultiplier = multiplier
end

function SineWaveVisualizer:getRewardMultiplier()
  return self.rewardMultiplier or 0
end

function SineWaveVisualizer:setRewardMultiplier(multiplier)
  multiplier = math.max(multiplier, 0)
  local currentColor = {self:getColor()}
  local color = ColorUtils.makeColorBetweenTwoColors(currentColor, self.rewardColor, multiplier)
  self:setColor(unpack(color))
  self.rewardMultiplier = multiplier

  self.yScale = multiplier * 1.3 + 1
end

-- Other ------------------------------

function SineWaveVisualizer:_DataVisualizer_receiveNextSample(normalizedSamplePower)
  local timeStep = self:getTimeStep()
  if (timeStep > 0.0) then
    local easedStep = normalizedSamplePower * timeStep * WAVEFORM_ADVANCEMENT_MULTIPLIER
    self.progress = self.progress + easedStep
  end

  applyProgress(self, self.progress, normalizedSamplePower)
  if (self.morphing) then
    applyMorph(self, self.progress, normalizedSamplePower)
  end
end

function SineWaveVisualizer:morph(targetPoints)
  self.morphing = true
  self.morphTarget = targetPoints
end

function SineWaveVisualizer:showPenaltyAnimation(minPeaks, maxPeaks)
  cancelPenaltyAnimation(self)
  self.heightOverflow = 0

  minPeaks, maxPeaks = minPeaks or DECREASED_NUMBER_OF_PEAKS, maxPeaks or INCREASED_NUMBER_OF_PEAKS
  local expectedPeaks = DEFAULT_NUMBER_OF_PEAKS

  self.isPenalty = true
  local prevPenalty = self:getPenaltyMultiplier()
  local peaksChangedHandler = function(current, min, expected, max)
    local penalty = (current - expected)/(max - expected)
    if (penalty > prevPenalty) then
      self:setPenaltyMultiplier(penalty)
      prevPenalty = penalty
    end
  end
  local squeezeAnimation = getRubberBandAnimation(self, peaksChangedHandler, minPeaks, expectedPeaks, maxPeaks)
  self.penaltyAnimation = squeezeAnimation

  return squeezeAnimation:start()
end

function SineWaveVisualizer:showRewardAnimation()
  local rewardEase = MOAIEaseType.EASE_IN
  local neutralEase = MOAIEaseType.EASE_IN
  local getReward = AHNonAttributeTo.new(self, 0.3, self.getRewardMultiplier, self.setRewardMultiplier, rewardEase, 1)
  local getNeutral = AHNonAttributeTo.new(self, 0.16, self.getRewardMultiplier, self.setRewardMultiplier, neutralEase, 0)
  local animation = AHSequence.new(getReward, getNeutral)
  return animation:start()
end

function SineWaveVisualizer:showRestoreAnimation()
  local restoreAction = AHNonAttributeTo.new(self, 0.3, self.getPenaltyMultiplier, self.setPenaltyMultiplier, nil, 0)
  return restoreAction:start()
end

--- Let the visualizer settle to its rest state and stop.
function SineWaveVisualizer:settleAndStop()
  self:stop()
  local lastStep = 0 -- Tried saving the last step > 0, but looks like audio waves tend to already have 0's at the end
  self._settleAction = runInCoroutine(function()
    -- The visualizer requires a few more samples to settle down.
    local samplesToSettleDown = 40
    for i=1,samplesToSettleDown do
      local sampleValue = lastStep * (samplesToSettleDown - i)/samplesToSettleDown
      self:_DataVisualizer_receiveNextSample(sampleValue)
      coroutine.yield()
    end
    self._settleAction = nil
  end)

  return self._settleAction
end

function SineWaveVisualizer:stop()
  self.super.stop(self)
  if self._settleAction then
    self._settleAction:stop()
    self._settleAction = nil
  end
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createLayout(self)
  local quality = 128
  local points = {}
  local selfW, selfH = self:getDims()
  local dw = selfW/quality
  for i = 0, quality do
    table.insert(points, {dw * i, selfH/2})
  end

  self.points = points
  self.yScale = 1
  self.maxHeight = selfH
end

-- Constructor ------------------------

function SineWaveVisualizer.new(width, height, color, penaltyColor, rewardColor, opacityMultiplier, lineThickness)
  local self = Group.newCustomDraw(width, height)
  DataVisualizer.makeDataVisualizer(self)
  Mixin.mix(self, SineWaveVisualizer)

  createLayout(self)

  self.progress = 0

  color = color or DEFAULT_COLOR
  penaltyColor = penaltyColor or PENALTY_COLOR
  rewardColor = rewardColor or REWARD_COLOR
  opacityMultiplier = opacityMultiplier or DEFAULT_OPACITY_MULTIPLIER

  color = ColorUtils.colorWithOpacityMultiplier(color, opacityMultiplier)
  penaltyColor = ColorUtils.colorWithOpacityMultiplier(penaltyColor, opacityMultiplier)
  rewardColor = ColorUtils.colorWithOpacityMultiplier(rewardColor, opacityMultiplier)

  self.penaltyColor = penaltyColor
  self.rewardColor = rewardColor
  self.opacityMultiplier = opacityMultiplier

  self:setColor(unpack(color))

  self.leftEnd = 1
  self.rightEnd = #self.points

  self.numberOfPeaks = DEFAULT_NUMBER_OF_PEAKS
  self:setPenaltyMultiplier(0)
  self.morphAmount = 0
  self.height = 0
  self.heightOverflow = 0
  self.lineThickness = lineThickness or DEFAULT_LINE_THICKNESS

  self:_DataVisualizer_receiveNextSample(0)

  self.deck:setDrawCallback(function()
    MOAIDraw.drawBeveledLine(ShapeUtils.flattenCoordPairs(self.points), self.lineThickness)
  end)

  return self
end
