--- ColorTracker
-- @author created by Jay Austria on November 13, 2013 based on code written by Aaron Barrett

-- ===========================================================================
class("ColorTracker")
-- ===========================================================================

local ATTRIBUTE_RED            = 0
local ATTRIBUTE_GREEN          = 1
local ATTRIBUTE_BLUE           = 2
local ATTRIBUTE_ALPHA          = 3
local CHANNEL_ATTRIBUTE_COUNT  = 4
local ATTRIBUTE_COLORING       = 0
local ATTRIBUTE_FOR_INDEX      = {
  MOAIColor.ATTR_R_COL,
  MOAIColor.ATTR_G_COL,
  MOAIColor.ATTR_B_COL,
  MOAIColor.ATTR_OPACITY
}

--- Stop any changes to the tracker's t value
function ColorTracker:stop()
  if (self.coloringAction) then
    self.coloringAction:stop()
  end
end

function ColorTracker:setPaused(isPaused)
  self.coloringAction:pause(isPaused)
end

--- Force set a time value to the color tracker
-- @param t - (number) t value to set the tracker to
function ColorTracker:setT(t)
  assert(t >= 0 and t <= 1.0, "Invalid t value provided")
  self.coloringScriptNode:setAttr(MOAIAnimCurve.ATTR_TIME, t)
end

--- Check the (normalized) t value the tracker is set to
-- @return t - (number) t value the tracker is set to
function ColorTracker:getT()
  return self.coloringScriptNode:getAttr(MOAIAnimCurve.ATTR_TIME)
end

--- Changes the current t value of the tracker over time
-- @param targetColoring - (number) t value to seek to
-- @param duration - (number) Amount of time to seek to the destination t
function ColorTracker:shiftColorWithDuration(targetColoring, duration)
  self:stop()

  local attributeID = ATTRIBUTE_COLORING
  self.coloringAction = AHAttributeTo.new(self.coloringScriptNode, duration, attributeID, nil, targetColoring)
  self.coloringAction:start()
end

--- Link's a MOAIColor to this color tracker. As the tracker's t value changes, the
-- colors are applied to the R,G,B, A channels of the linked MOAINode.
-- @param nodeToLink - (MOAINode) Color userdata containing a color node that will be affected
--                     by value changes in this tracker
function ColorTracker:linkColorNodeToSelf(nodeToLink)
  nodeToLink:setAttrLink(self.attributeLinked, self.currentColor, MOAIColor.COLOR_TRAIT)
  self.linkedNode = nodeToLink
end

function ColorTracker:unlinkColorNode()
  if (self.linkedNode) then
    self.linkedNode:clearAttrLink(self.attributeLinked)
    self.linkedNode = nil
  end
end

--- Turns a table of color shift rules into curves applied to each color
-- channel over time
-- @param coloringTable - (table) Table with the t values and keys (colors) that this
--                        tracker shifts between
function ColorTracker:configureColoringWithTable(coloringTable)
  assert(coloringTable and #coloringTable > 0, "Invalid coloring table provided")
  assert(coloringTable[1].t ~= nil and
         coloringTable[1].keyColor ~= nil, "Coloring table doesnt seem formatted correctly")

  self.coloringScriptNode = MOAIScriptNode:new()
  self.coloringScriptNode:reserveAttrs(1)

  self.currentColor = ColorUtils.channelsToMOAIColor(1.0, 1.0, 1.0)

  local initialColoringLevel = coloringTable[1]
  local initialChannels       = initialColoringLevel.colorChannels

  local keyCount              = #coloringTable

  self.channelCurves = {}

  for channelIndex = 1, CHANNEL_ATTRIBUTE_COUNT do
    local attributeIndex = ATTRIBUTE_FOR_INDEX[channelIndex]

    local channelCurve = MOAIAnimCurve.new()
    channelCurve:reserveKeys(keyCount)

    table.insert(self.channelCurves, channelCurve)

    for levelIndex, coloringLevel in ipairs(coloringTable) do
      channelCurve:setKey(levelIndex, coloringLevel.t, coloringLevel.keyColor[channelIndex], MOAIEaseType.LINEAR)
    end

    self.currentColor:setAttrLink(attributeIndex, channelCurve, MOAIAnimCurve.ATTR_VALUE)

    channelCurve:setAttrLink(MOAIAnimCurve.ATTR_TIME, self.coloringScriptNode, ATTRIBUTE_COLORING)
    channelCurve:setWrapMode(MOAIAnimCurve.CLAMP)
  end
end

--- Creates a color tracker object
-- Color tracker can be linked to a prop's color and configured with an input table
-- which specifies what colors to blend to based on current t value.
-- For example a sample input table
-- EXAMPLE_SHIFT_TABLE = {
--   {t = 0.0, keyColor = {1.0, 1.0, 1.0, 1.0}},
--   {t = 0.5, keyColor = {1.0, 0.0, 0.0, 1.0}},
--   {t = 0.9, keyColor = {0.0, 1.0, 0.0, 1.0}},
--   {t = 1.0, keyColor = {1.0, 0.0, 1.0, 1.0}}
-- }
-- can be faded between black->red->green->blue by shifting the t value with
-- shiftColorWithDuration() and this color will be applied to a prop that has been linked
-- to the tracker with linkColorNodeToSelf()
-- @param coloringTable - (table) Table with the t values and keys (colors) that this
--                        tracker shifts between
-- @param attributeLinked - (number) Attribute id of this object that will be linked to nodes.
--                          Defaults to MOAIColor.INHERIT_COLOR.
-- @return (ColorTracker) the color tracker object

function ColorTracker.newWithTable(coloringTable, attributeLinked)
  attributeLinked = attributeLinked or MOAIColor.INHERIT_COLOR

  local self = {}
  Mixin.mix(self, ColorTracker)
  self:configureColoringWithTable(coloringTable)
  self.coloringTable = coloringTable
  self.attributeLinked = attributeLinked
  return self
end
