--- CompositeLabel
-- @author Created by Aaron Barrett on November 25, 2013.

-- ===========================================================================
class("CompositeLabel")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Calls release() on all of the textures that were rendered to create this label.
-- Use this method to free up texture memory immediately if the garbage collector
-- doesn't get to it quickly enough.
-- @param self - (Label) the CompositeLabel whose textures will be freed.
function CompositeLabel:releaseTexture()
  local lines = self.lines

  for i, line in ipairs(lines) do
    local labels = line.labels
    for j, label in ipairs(labels) do
      label:releaseTexture()
    end
  end
end

--- Applies 'horizontalAlignment' to reposition the lines in this CompositeLabel
-- @param horizontalAlignment (default Alignment.LEFT) - (number) the horizontal alignment to use when formatting this label.  Should be one of Alignment.LEFT, Alignment.CENTER, or Alignment.RIGHT)
function CompositeLabel:setHorizontalAlignment(horizontalAlignment)
  horizontalAlignment = horizontalAlignment or Alignment.LEFT

  local lines = self.lines

  local width = self.width
  local height = self.height
  local lineSpacing = self.lineSpacing
  local maxLineHeight = self.maxLineHeight

  local penY = height - maxLineHeight * 0.5

  for i, line in ipairs(lines) do

    local lineX = nil

    if (horizontalAlignment == Alignment.LEFT) then
      local lineWidth, _ = line:getDims()
      lineX = lineWidth * 0.5
    elseif (horizontalAlignment == Alignment.CENTER) then
      lineX = width * 0.5
    elseif (horizontalAlignment == Alignment.RIGHT) then
      local lineWidth, _ = line:getDims()
      lineX = width - lineWidth * 0.5
    else
      error("Unrecognized horizontal alignment")
    end

    line:setLoc(lineX, penY)

    penY = penY - maxLineHeight - lineSpacing
  end

end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new CompositeLabel with the given width, lines, and line spacing
-- @param width - (number) the width to use for this label.  The Group that is returned will have the dimensions (width, height), where height is the distance from the top of the top line to the bottom of the bottom line.
-- @param lines - (table) a table (generally returned from a CompositeLabelRenderer) containing Groups for each of the lines in this CompositeLabel.
-- @param lineSpacing (default 0.0) - (number) the amount (in 1x coordinates) that each new line should be moved up (to overlap the previous line) or down (for additional spacing)
-- @param horizontalAlignment (default Alignment.LEFT) - (number) the horizontal alignment to use when formatting this label.  Should be one of Alignment.LEFT, Alignment.CENTER, or Alignment.RIGHT)
function CompositeLabel.new(width, lines, lineSpacing, horizontalAlignment)
  lineSpacing = lineSpacing or 0.0

  local self = Group.newDeprecated()
  Mixin.mix(self, CompositeLabel)

  local linesCount = #lines

  local maxLineHeight = 0.0

  for i, line in ipairs(lines) do
    local _, lineHeight = line:getDims()
    maxLineHeight = math.max(lineHeight, maxLineHeight)
    self:insert(line)
  end

  local height = linesCount * maxLineHeight + (linesCount - 1) * lineSpacing

  self:setBounds(0.0, 0.0, 0, width, height, 0)
  self:setAnchorPointCenter()

  self.width = width
  self.height = height
  self.lineSpacing = lineSpacing
  self.maxLineHeight = maxLineHeight
  self.lines = lines

  self:setHorizontalAlignment(horizontalAlignment)

  return self
end
