--- CompositeLabelRenderer
-- @author Created by Aaron Barrett on November 24, 2013.

-- ===========================================================================
class("CompositeLabelRenderer")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Sets the line spacing that this CompositeLabelRenderer should use.  @see CompositeLabel
-- @param lineSpacing (number) - the amount (in 1x coordinates) that each new line should be moved up (to overlap the previous line) or down (for additional spacing)
function CompositeLabelRenderer:setLineSpacing(lineSpacing)
  self.lineSpacing = lineSpacing
end

--- Sets the width that CompositeLabels rendererd by this CompositeLabelRenderer should use
-- @param width - (number) the width to use for this renderer.  This renderer's CompositeLabels will have the dimensions (width, height), where height is the distance from the top of the top line to the bottom of the bottom line.
function CompositeLabelRenderer:setWidth(width)
  self.width = width
end

--- Sets the horizontal alignment to use for this CompositeLabelRenderer
-- @param horizontalAlignment (number, default Alignment.LEFT) - the horizontal alignment to use when formatting this renderer's CompositeLabels.  Should be one of Alignment.LEFT, Alignment.CENTER, or Alignment.RIGHT)
function CompositeLabelRenderer:setHorizontalAlignment(horizontalAlignment)
  self.horizontalAlignment = horizontalAlignment
end

--- Sets the default LabelRenderer that this CompositeLabelRenderer should use.
-- @param renderer (LabelRenderer) - A renderer that will be used by default by this CompositeLabelRenderer to render the single line Labels that make up a CompositeLabel.
function CompositeLabelRenderer:setDefaultRenderer(renderer)
  self.defaultRenderer = renderer
end

--- Adds an additional LabelRenderer to be used to render all single-line labels in the CompositeLabel that fall between startCharacter and endCharacter.
-- Note that nested font faces are not supported.
-- @param renderer (LabelRenderer) - A renderer that will be used by this CompositeLabelRenderer to render the single line Labels that make up a CompositeLabel.
-- @param startCharacter (string) - A character to set off the use of renderer when it occurs in the strings rendered by this CompositeLabelRenderer.
-- @param endCharacter (string) - A character to conclude the use of renderer when it occurs in the strings rendererd by this CompositeLabelRenderer.
function CompositeLabelRenderer:addRenderer(renderer, startCharacter, endCharacter)
  self.rendererForStartCharacter[startCharacter] = renderer

  self.startCharacters = self.startCharacters .. startCharacter
  self.endCharacters = self.endCharacters .. endCharacter
end

-- Render CompositeLabels -------------

--- Renders the string 'text' as a new CompositeLabel using this CompositeLabelRenderer's default settings.
-- If the start or end characters for any of the LabelRenderers that have been added to this CompositeLabelRenderer are encountered in the string,
-- the corresponding LabelRenderer will be used for any words/characters that occur between those characters.  The start/end characters themselves
-- will not be rendererd.
-- @param text (string) - the string to render.
function CompositeLabelRenderer:newCompositeLabel(text)
  local width = self.width
  assert(width, "Must set width before rendering a CompositeLabel.")

  local horizontalAlignment = self.horizontalAlignment
  local lineSpacing = self.lineSpacing

  local currentRenderer = self.defaultRenderer

  local startCharacters = self.startCharacters
  local endCharacters = self.endCharacters

  local textLength = #text

  local penX = 0.0
  local lineHeight = 0.0

  local lines = {}
  local line = Group.newDeprecated()
  line.labels = {}
  local word = ""

  local function writeLine()
    line:setBounds(0.0, 0.0, 0, penX, lineHeight, 0)
    line:setAnchorPointCenter()

    for i, label in ipairs(line.labels) do
      label:setLoc(label.x, lineHeight * 0.5)
      line:insert(label)
    end

    table.insert(lines, line)

    line = Group.newDeprecated()
    line.labels = {}
    penX = 0.0
    lineHeight = 0.0
  end

  local function writeWord()
    if (#word > 0) then
      local label = DeprecatedLabel.new(currentRenderer)
      label:setText(word)

      local labelWidth, labelHeight = label:getDims()
      if (penX + labelWidth >= width) then
        writeLine()
      end

      penX = penX + labelWidth

      label.x = penX - labelWidth * 0.5
      table.insert(line.labels, label)
      lineHeight = math.max(lineHeight, labelHeight)

      word = ""
    end
  end

  -- Iterate through all characters in 'text'
  for i = 1, textLength do
    local c = text:sub(i, i)

    -- If the character is whitespace (or a hyphen, so hyphenated words get broken up onto multiple lines) or the last character
    if (c:find("[%s-]") or i == textLength) then
      word = word .. c .. " " --Must add an extra space, since TrueType doesn't render the last whitespace character
      writeWord()

    -- If the character is the start character for one of the renderers
    elseif (startCharacters:find(c, nil, true)) then
      writeWord()
      currentRenderer = self.rendererForStartCharacter[c]

    -- If the character is the end character for one of the renderers
    elseif (endCharacters:find(c, nil, true)) then
      writeWord()
      currentRenderer = self.defaultRenderer

    -- Otherwise, the character should be added to the word.
    else
      word = word .. c
    end

  end

  writeLine()

  local compositeLabel = CompositeLabel.new(width, lines, lineSpacing, horizontalAlignment)

  return compositeLabel
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new CompositeLabelRenderer that uses the given LabelRenderer as a default renderer.
-- @param defaultRenderer (LabelRenderer) - A renderer that will be used by default by this CompositeLabelRenderer to render the single line Labels that make up a CompositeLabel.
-- @return the CompositeLabelRenderer
function CompositeLabelRenderer.new(defaultRenderer)
  local self = {}
  Mixin.mix(self, CompositeLabelRenderer)

  self:setDefaultRenderer(defaultRenderer)
  self.rendererForStartCharacter = {}
  self.startCharacters = ""
  self.endCharacters = ""

  self.lineSpacing = 0.0
  self.width = nil
  self.horizontalAlignment = Alignment.LEFT

  return self
end
