--- Label
-- @author Rewritten most recently by Aaron Barrett on November 17, 2013.  Previous revisions occurred on April 29, 2013 and June 24, 2013.

-- ===========================================================================
class("DeprecatedLabel")
-- ===========================================================================

require("text.TextCommon")

require("MetalPrebinder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_HORIZONTAL_ALIGNMENT = Alignment.LEFT
local DEFAULT_VERTICAL_ALIGNMENT   = Alignment.TOP

local DEFAULT_WORD_BREAK_MODE      = MOAITextBox.WORD_BREAK_HYPHEN

local DEFAULT_LINE_SPACING = 1.0
local DEFAULT_GRANULARITY  = 1.0
local ROUND_TO_INT_TRUE    = true


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

--- Render a texture of the specified text.
-- Caches the previous call and returns that if the text hasn't changed.
local function renderTexture(self, text)
  if not self.preparedRender or self.preparedRender.text ~= text then
    self.preparedRender = {}
    self.preparedRender.text = text
    local texture, renderedWidth, renderedHeight, glyphTable = self.fontRenderEngine:renderTexture(text)
    texture:setFilter(MOAITexture.GL_LINEAR, MOAITexture.GL_LINEAR)
    self.preparedRender.renderResult = {texture, renderedWidth, renderedHeight, glyphTable}
    texture.renderedWidth = renderedWidth
    texture.renderedHeight = renderedHeight

    MetalPrebinder.prebindTexture(texture)
  end
  return unpack(self.preparedRender.renderResult)
end

--- Replace the label texture.
local function replaceTexture(self, texture, renderedWidth, renderedHeight)
  local textureWidth, textureHeight = texture:getSize()
  local selfW = renderedWidth / Host.contentScale
  local selfH = renderedHeight / Host.contentScale
  local uvWidth = renderedWidth / textureWidth
  local uvHeight = renderedHeight / textureHeight

  self:releaseTexture()

  local deck = MOAIGfxQuad2D.new()
  deck:setTexture(texture)
  deck:setRect(0.0, 0.0, selfW, selfH)
  deck:setUVRect(0.0, uvHeight, uvWidth, 0.0)
  self:setDims(selfW, selfH)
  self:setDeck(deck)

  self.originalUVDims = { uvWidth, uvHeight }

  self.deck = deck
  self.texture = texture
end

local function isPrintable(text)
  if text == "" then
    return false
  end
  return true
end

local function updateMargins(self, glyphTable)
  if not self.isSingleLine then
    return
  end

  local leftmostGlyph = glyphTable[1]
  local leftMargin = leftmostGlyph.xOffset / Host.contentScale
  self.leftMargin = leftMargin

  local rightmostGlyph = glyphTable[#glyphTable]
  local rightMargin = (rightmostGlyph.xAdvance - rightmostGlyph.xOffset - (rightmostGlyph.xMax - rightmostGlyph.xMin)) / Host.contentScale
  self.rightMargin = rightMargin
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Releases this label's internal texture and resources.
-- Use this method to free up texture memory immediately if the garbage collector
-- doesn't get to it quickly enough.
-- @param self - (Label) the Label whose texture will be freed.
function DeprecatedLabel:releaseTexture()
  if self.texture then
    self.texture:release()
    self.texture = nil
    self.deck = nil
  end
  self:setDeck(MOAIGfxQuad2D.new())
  self.glyphTable = nil
  self.preparedRender = nil
end

--- Do as many up-front preparations to set this label's text as possible.
-- Note: If the text did not fit, will set the render engine's font size to its
--       minimum font size. Use textDidFit() to check if this method succeeded
--       in fitting the text.
-- Runs in a coroutine. You can wait() on the returned coroutine.
-- @param text (string or number) the new text to display
function DeprecatedLabel:prepareToSetText(text)
  text = TextCommon.validateAndConvertText(text)

  return runExclusivelyInCoroutine(self, "prepareToSetText", true, function()
    local textDidFit = true
    if isPrintable(text) then
      if self.fontRenderEngine:requiresFontSizeCalculation() then
        wait(self.fontRenderEngine:calculateOptimalFontSizeForText(text))
        if self.fontRenderEngine:getFontSize() < 0 then
          textDidFit = false
          self.fontRenderEngine:setFontSize(self.fontRenderEngine:getMinFontSize())
        end
      end
      renderTexture(self, text)
    end
    self.textDidFitValue = textDidFit
  end)
end

--- Set this label's text.
-- This method will wait on prepareToSetText if it has not run with this text already.
-- @param text (string or number) the new text to display
-- @return The coroutine that is updating the text.
function DeprecatedLabel:setText(text)
  text = TextCommon.validateAndConvertText(text)

  if isPrintable(text) then
    if self.fontRenderEngine:requiresFontSizeCalculation() then
      wait(self:prepareToSetText(text))
      -- An extra yield is required or else the label's new dimensions will not be
      -- available until the next update.
      coroutine.yield()
    end
    local texture, renderedWith, renderedHeight, glyphTable = renderTexture(self, text)
    replaceTexture(self, texture ,renderedWith, renderedHeight)

    if glyphTable then
      updateMargins(self, glyphTable)
    end

    self.glyphTable = glyphTable
  else
    self:releaseTexture()
    self:setDims(2, 2)
  end
  self.text = text
end

function DeprecatedLabel:getText()
  return self.text
end

--- Returns true if the last attempt to make the text fit within this label's
-- parameters succeeded. See prepareToSetText and setText.
function DeprecatedLabel:textDidFit()
  return self.textDidFitValue
end

--- Returns the current glyph table, if any.
function DeprecatedLabel:getGlyphTable()
  return self.glyphTable
end

function DeprecatedLabel:disposeOfSelf()
  self:releaseTexture()
end

--- Clone this label. The cloned label will use the same texture, saving gfx mem.
-- Note: Be VERY CAREFUL when releasing the texture, because many labels could be
--       using the same texture!
function DeprecatedLabel:clone()
  local texture = self.texture
  local renderedWidth = texture.renderedWidth
  local renderedHeight = texture.renderedHeight
  local textureWidth, textureHeight = texture:getSize()
  local cloneW, cloneH = self:getDims()
  local uvWidth = renderedWidth / textureWidth
  local uvHeight = renderedHeight / textureHeight

  local labelClone = Group.new(cloneW, cloneH)
  Mixin.mix(labelClone, DeprecatedLabel)

  local deck = MOAIGfxQuad2D.new()
  deck:setTexture(texture)
  deck:setRect(0.0, 0.0, cloneW, cloneH)
  deck:setUVRect(0.0, uvHeight, uvWidth, 0.0)
  labelClone:setDeck(deck)

  return labelClone
end

function DeprecatedLabel:getLeftMargin()
  return self.leftMargin
end

function DeprecatedLabel:getRightMargin()
  return self.rightMargin
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructors -----------------------

function DeprecatedLabel.new(fontRenderEngine)
  local self
  local selfW, selfH = fontRenderEngine:getDimensions()
  selfW = selfW or 2
  selfH = selfH or 2
  self = Group.new(selfW, selfH)
  Mixin.mix(self, DeprecatedLabel)

  self.fontRenderEngine = fontRenderEngine
  fontRenderEngine:onLabelCreated(self)

  return self
end


local function internalNewFitted(text, width, height, font, maxFontSize, minFontSize, horizontalAlignment, verticalAlignment, forceSingleLine, wordBreakMode, shouldBuildGlyphTable, lineSpacing)
  text = TextCommon.validateAndConvertText(text)
  assert(font, "font cannot be nil.")
  assert(maxFontSize, "maxFontSize cannot be nil.")
  assert(minFontSize, "minFontSize cannot be nil.")

  local engine = FontRenderEngine.new()
  engine:setFont(font)
  engine:setWidth(width)
  engine:setHeight(height)
  engine:setAlignment(horizontalAlignment or DEFAULT_HORIZONTAL_ALIGNMENT, verticalAlignment or DEFAULT_VERTICAL_ALIGNMENT)
  engine:setWordBreak(wordBreakMode or DEFAULT_WORD_BREAK_MODE)
  engine:setMaxFontSize(maxFontSize)
  engine:setMinFontSize(minFontSize)
  engine:setMultiLine(not forceSingleLine)
  engine:setReturnGlyphBounds(shouldBuildGlyphTable)
  engine:setLineSpacing(lineSpacing or DEFAULT_LINE_SPACING)

  local self = DeprecatedLabel.new(engine)
  self.isSingleLine = forceSingleLine
  self:setText(text)

  return self
end

-- Constructors -----------------------


--- Creates a Label that displays 'text' on a single line using the specified font and font size.
-- NOTE: the dimensions of the label will be the size of the rendered text.
-- @param text - (string or number) The string to display.
-- @param font - (MOAIFreeTypeFont) The font in which to draw the text.
-- @param fontSize - (integer) The size in which to draw the font.
-- @param shouldBuildGlyphTable - (optional boolean), defaults to false - Whether to return a table containing glyph coordinate information for all characters in the string.
-- @return - (Group) The Label Group, with text written on a single line.
-- @return glyphTable - (table) A table containing glyph coordinate data for all characters in the string - Only returned if shouldBuildGlyphTable was true.
function DeprecatedLabel.newSingleLine(text, font, fontSize, shouldBuildGlyphTable)
  text = TextCommon.validateAndConvertText(text)
  assert(font, "font cannot be nil.")
  assert(fontSize, "fontSize cannot be nil.")

  local engine = FontRenderEngine.new()
  engine:setFont(font)
  engine:setFontSize(fontSize)
  engine:setReturnGlyphBounds(shouldBuildGlyphTable)

  local self = DeprecatedLabel.new(engine)
  self.isSingleLine = true
  self:setText(text)

  return self
end

--- Creates a Label that displays 'text'.  The text may span multiple lines.
-- NOTE: Labels created using this constructor will cut off words if 'fontSize' is too big to fit in the dimensions.
-- @param text - (string or number) The string to display.
-- @param width - (number) The width of the region in which to draw the text.
-- @param height - (number) The height of the region in which to draw the text.
-- @param font - (MOAIFreeTypeFont) The font in which to draw the text.
-- @param fontSize - (integer) The size in which to draw the font.
-- @param horizontalAlignment - (optional integer, defaults to DEFAULT_HORIZONTAL_ALIGNMENT (Alignment.LEFT) ) The horizontal alignment with which to draw the text.
-- @param verticalAlignment - (optional integer, defaults to DEFAULT_VERTICAL_ALIGNMENT (Alignment.TOP) ) The vertical alignment with which to draw the text.
-- @param wordBreakMode - (optional integer, defaults to DEFAULT_WORD_BREAK_MODE (MOAIFreeTypeTextBox.WORD_BREAK_HYPHEN) ) The word break mode to use for the text.
-- @param shouldBuildGlyphTable - (optional boolean), defaults to false - Whether to return a table containing glyph coordinate information for all characters in the string.
-- @param lineSpacing - (optional number), defaults to 1.0 - Multiplier applied to the line spacing of the font when rendering lines
-- @return - (MOAITextBox) The Label Group.
-- @return glyphTable - (table) A table containing glyph coordinate data for all characters in the string - Only returned if shouldBuildGlyphTable was true.
function DeprecatedLabel.newMultiLine(text, width, height, font, fontSize, horizontalAlignment, verticalAlignment, wordBreakMode, shouldBuildGlyphTable, lineSpacing)
  text = TextCommon.validateAndConvertText(text)
  assert(width, "width cannot be nil.")
  assert(font, "font cannot be nil.")
  assert(fontSize, "fontSize cannot be nil.")

  local engine = FontRenderEngine.new()
  engine:setFont(font)
  engine:setWidth(width)
  if height then
    engine:setHeight(height)
  end
  engine:setAlignment(horizontalAlignment or DEFAULT_HORIZONTAL_ALIGNMENT, verticalAlignment or DEFAULT_VERTICAL_ALIGNMENT)
  engine:setWordBreak(wordBreakMode or DEFAULT_WORD_BREAK_MODE)
  engine:setFontSize(fontSize)
  engine:setMultiLine(true)
  engine:setReturnGlyphBounds(shouldBuildGlyphTable)
  engine:setLineSpacing(lineSpacing or DEFAULT_LINE_SPACING)

  local self = DeprecatedLabel.new(engine)
  self.isSingleLine = false
  self:setText(text)

  return self
end

--- Creates a Label that diplays 'text' at the optimal size to fit within the specified dimensions on one or more lines.
-- @param text - (string or number) The string to display.
-- @param width - (number) The width of the region in which to draw the text.
-- @param height - (number) The height of the region in which to draw the text.
-- @param font - (MOAIFreeTypeFont) The font in which to draw the text.
-- @param maxFontSize - integer) The maximum allowable font size.
-- @param minFontSize - (integer) The minimum allowable font size.
-- @param horizontalAlignment - (optional integer, defaults to DEFAULT_HORIZONTAL_ALIGNMENT (Alignment.LEFT) ) The horizontal alignment with which to draw the text.
-- @param verticalAlignment - (optional integer, defaults to DEFAULT_VERTICAL_ALIGNMENT (Alignment.TOP) ) The vertical alignment with which to draw the text.
-- @param shouldBuildGlyphTable - (optional boolean), defaults to false - Whether to return a table containing glyph coordinate information for all characters in the string.
-- @return self - (Group) The Label Group with a font size fitted to its dimensions.
-- @return fontSize - (integer) The optimal font size that was used to fit the string within the box.  This font size should be the same across content scales.
-- @return renderedFontSize - (integer) The font size, multiplied by Host.contentScale.  This is the actual font size that was used to render the texture.
-- @return glyphTable - (table) A table containing glyph coordinate data for all characters in the string - Only returned if shouldBuildGlyphTable was true.
function DeprecatedLabel.newSingleLineFitted(text, width, height, font, maxFontSize, minFontSize, horizontalAlignment, verticalAlignment, shouldBuildGlyphTable)
  local forceSingleLine = true
  return internalNewFitted(text, width, height, font, maxFontSize, minFontSize, horizontalAlignment, verticalAlignment, forceSingleLine, MOAITextBox.WORD_BREAK_CHAR, shouldBuildGlyphTable, DEFAULT_LINE_SPACING)
end

--- Creates a Label that diplays 'text' at the optimal size to fit within the specified dimensions on one or more lines.
-- @param text - (string or number) The string to display.
-- @param width - (number) The width of the region in which to draw the text.
-- @param height - (number) The height of the region in which to draw the text.
-- @param font - (MOAIFreeTypeFont) The font in which to draw the text.
-- @param maxFontSize - (integer) The maximum allowable font size.
-- @param minFontSize - (integer) The minimum allowable font size.
-- @param horizontalAlignment - (optional integer, defaults to DEFAULT_HORIZONTAL_ALIGNMENT (Alignment.LEFT) ) The horizontal alignment with which to draw the text.
-- @param verticalAlignment - (optional integer, defaults to DEFAULT_VERTICAL_ALIGNMENT (Alignment.TOP) ) The vertical alignment with which to draw the text.
-- @param wordBreakMode - (optional integer, defaults to DEFAULT_WORD_BREAK_MODE (MOAIFreeTypeTextBox.WORD_BREAK_HYPHEN) ) The word break mode to use for the text.
-- @param shouldBuildGlyphTable - (optional boolean), defaults to false - Whether to return a table containing glyph coordinate information for all characters in the text.
-- @param lineSpacing - (optional number), defaults to 1.0 - Multiplier applied to the line spacing of the font when rendering lines
-- @return self - (Group) The Label Group with a font size fitted to its dimensions.
-- @return fontSize - (integer) The optimal font size that was used to fit the text within the box.  This font size should be the same across content scales.
-- @return renderedFontSize - (integer) The font size, multiplied by Host.contentScale.  This is the actual font size that was used to render the texture.
-- @return glyphTable - (table) A table containing glyph coordinate data for all characters in the string - Only returned if shouldBuildGlyphTable was true.
function DeprecatedLabel.newMultiLineFitted(text, width, height, font, maxFontSize, minFontSize, horizontalAlignment, verticalAlignment, wordBreakMode, shouldBuildGlyphTable, lineSpacing)
  local forceSingleLine = false
  return internalNewFitted(text, width, height, font, maxFontSize, minFontSize, horizontalAlignment, verticalAlignment, forceSingleLine, wordBreakMode, shouldBuildGlyphTable, lineSpacing)
end
