--- FontRenderEngine
-- @author Created by Aaron Barrett on November 17, 2013.

-- ===========================================================================
class("FontRenderEngine")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Operating modes. These are set based on the individual settings.

-- Render a single line.
local MODE_UNSIZED = "MODE_UNSIZED"

-- Pick the optimal font size to fit within the specified width/height.
local MODE_OPTIMAL_FONT_SIZE = "MODE_OPTIMAL_FONT_SIZE"

-- Pick the optimal height to fit the width at the specified font size.
local MODE_OPTIMAL_HEIGHT = "MODE_OPTIMAL_HEIGHT"

-- No calculations. Fit whatever can be fit within the area at the specified font size.
local MODE_FIXED_SIZE = "MODE_FIXED_SIZE"


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function hasDimensions(self)
  local selfW, selfH = self:getDimensions()
  return selfW and selfW > 0 and selfH and selfH > 0
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Getters & Setters ------------------

--- If true, will render multiline labels. Otherwise it renders single line labels. Default false
function FontRenderEngine:setMultiLine(isMultiLine)
  self.isMultiLine = isMultiLine
  self:setForceSingleLine(not self.isMultiLine)
end

function FontRenderEngine:getMultiLine()
  return self.isMultiLine
end

--- Tells this FontRenderEngine to use the given horizontal and vertical alignment. Default Alignment.LEFT, Alignment.TOP
-- @param horizontalAlignment - (number) the horizontal alignment to use.  Should be one of Alignment.LEFT, Alignment.CENTER, or Alignment.RIGHT
-- @param verticalAlignment - (number) the vertical alignment to use.  Should be one of Alignment.TOP, Alignment.CENTER, or Alignment.BOTTOM
function FontRenderEngine:setAlignment(horizontalAlignment, verticalAlignment)
  self.horizontalAlignment = horizontalAlignment
  self.verticalAlignment = verticalAlignment
  self.super.setAlignment(self, horizontalAlignment, verticalAlignment)
end

function FontRenderEngine:getAlignment()
  return self.horizontalAlignment, self.verticalAlignment
end

--- Tells this FontRenderEngine to use the given MOAIFreeTypeFont
-- @param font - (MOAIFreeTypeFont) the font to use.
function FontRenderEngine:setFont(font)
  self.font = font
  self.super.setFont(self, font)
end

function FontRenderEngine:getFont()
  return self.font
end

--- Tells this FontRenderEngine to return a glyph bounds table with the Labels it renders. Default false
-- @param shouldReturnGlyphBounds - (boolean) whether or not this FontRenderEngine should return glyph bounds tables.
function FontRenderEngine:setReturnGlyphBounds(shouldReturnGlyphBounds)
  self.shouldReturnGlyphBounds = shouldReturnGlyphBounds
  self.super.setReturnGlyphBounds(self, shouldReturnGlyphBounds)
end

function FontRenderEngine:getReturnGlyphBounds()
  return self.shouldReturnGlyphBounds
end

--- Tells this FontRenderEngine to use a particular word break mode for multi-line labels. Default MOAITextBox.WORD_BREAK_NONE
-- @param wordBreakMode - (number) The word break mode to use.  Should be one of MOAITextBox.WORD_BREAK_HYPHEN, MOAITextBox.WORD_BREAK_NONE, or MOAITextBox.WORD_BREAK_CHAR
function FontRenderEngine:setWordBreak(wordBreakMode)
  self.wordBreakMode = wordBreakMode
  self.super.setWordBreak(self, wordBreakMode)
end

function FontRenderEngine:getWordBreak()
  return self.wordBreakMode
end

--- Tells this FontRenderEngine to use a particular width for non-single-line-labels.
-- @param width - (number) The width to use. If nil, renders will be sized to fit the text & font size.
function FontRenderEngine:setWidth(width)
  self.width = width
  self.super.setWidth(self, width * Host.contentScale)
end

--- Tells this FontRenderEngine to use a particular height for non-single-line-labels.
-- Note that this value is in game coordinates (close to 320x480 full screen, no matter what the contentScale).
-- @param height - (number) The height to use.
function FontRenderEngine:setHeight(height)
  self.height = height
  self.super.setHeight(self, height * Host.contentScale)
end

--- Calls setWidth(width) and setHeight(height)
-- Note that these values are in game coordinates (close to 320x480 full screen, no matter what the contentScale).
-- @param width - (number) The width to use. If nil, renders will be sized to fit the text & font size.
-- @param height - (number) The height to use.
function FontRenderEngine:setDimensions(width, height)
  self:setWidth(width)
  self:setHeight(height)
end

function FontRenderEngine:getDimensions()
  return self.width, self.height
end

--- Tells this FontRenderEngine to use a particular font size for non-fitted Labels.
-- Note that this value is in game coordinates (close to 320x480 full screen, no matter what the contentScale).
-- @param fontSize - (number) The font size to use.
function FontRenderEngine:setFontSize(fontSize)
  self.fontSize = fontSize
  self.super.setFontSize(self, fontSize * Host.contentScale)
end

function FontRenderEngine:getFontSize()
  return self.fontSize
end

--- Tells this FontRenderEngine to use a particular "granularity" value for finding the optimal font size for fitted Labels.
-- "granularity" is the threshold for the difference between maximum and minumum font size parameters for determining when
-- the optimal size processing is complete.  The value will be multiplied by the contentScale. Default 1
-- @param granularity - (number) the granularity value to use.
function FontRenderEngine:setGranularity(granularity)
  self.granularity = granularity
  self.super.setGranularity(self, granularity * Host.contentScale)
end

function FontRenderEngine:getGranularity()
  return self.granularity
end

--- Tells this FontRenderEngine to use the given font size as the upper bound when searching for an optimal font size for fitted Labels.
-- Note that this value is in game coordinates (close to 320x480 full screen, no matter what the contentScale).
-- @param maxFontSize - (number) the maximum font size to use.
function FontRenderEngine:setMaxFontSize(maxFontSize)
  self.maxFontSize = maxFontSize
  self.super.setMaxFontSize(self, maxFontSize * Host.contentScale)
end

function FontRenderEngine:getMaxFontSize()
  return self.maxFontSize
end

--- Tells this FontRenderEngine to use the given font size as the lower bound when searching for an optimal font size for fitted Labels.
-- Note that this value is in game coordinates (close to 320x480 full screen, no matter what the contentScale).
-- @param minFontSize - (number) the minimum font size to use.
function FontRenderEngine:setMinFontSize(minFontSize)
  self.minFontSize = minFontSize
  self.super.setMinFontSize(self, minFontSize * Host.contentScale)
end

function FontRenderEngine:getMinFontSize()
  return self.minFontSize
end

--- How much space to put between lines when rendering a multiline texture.
-- @param lineSpacing - (number) The amount of space between lines. Default 1.0
function FontRenderEngine:setLineSpacing(lineSpacing)
  self.lineSpacing = lineSpacing
  self.super.setLineSpacing(self, lineSpacing)
end

function FontRenderEngine:getLineSpacing()
  return self.lineSpacing
end

--- Set a function to be called for extra processing when a label is created.
-- A label will call this function upon creation.
function FontRenderEngine:setOnLabelCreated(onLabelCreated)
  self.onLabelCreatedFunction = onLabelCreated
end

function FontRenderEngine:getOnLabelCreated()
  return self.onLabelCreatedFunction
end

--- Called by DeprecatedLabel.new()
function FontRenderEngine:onLabelCreated(label)
  if self.onLabelCreatedFunction then
    self.onLabelCreatedFunction(label)
  end
end

function FontRenderEngine:getOperatingMode()
  local hasVariableFontSize = self.minFontSize ~= nil and self.minFontSize > 0 and self.maxFontSize ~= nil and self.maxFontSize > 0
  local hasFixedFontSize = self.fontSize ~= nil and self.fontSize > 0
  local hasWidth = self.width ~= nil and self.width > 0
  local hasHeight = self.height ~= nil and self.height > 0

  assert(hasVariableFontSize or hasFixedFontSize, string.format("Must have fontSize or minFontSize and maxFontSize. fontSize = %s, minFontSize = %s, maxFontSize = %s", tostring(self.fontSize), tostring(self.minFontSize), tostring(self.maxFontSize)))

  if hasVariableFontSize then
    assert(hasWidth and hasHeight, string.format("Variable font size requires width and height. width = %s, height = %s", tostring(self.width), tostring(self.height)))
    return MODE_OPTIMAL_FONT_SIZE
  end

  if not self.isMultiLine and not hasWidth and not hasHeight then
    return MODE_UNSIZED
  end

  -- From here on, only multiline is valid.
  assert(hasWidth, "Must specify width for multiline.")

  if hasHeight then
    return MODE_FIXED_SIZE
  end

  return MODE_OPTIMAL_HEIGHT
end

-- Font Size Calculation --------------

--- Calculate the optimal font size to fit the text within this renderer's dimensions.
-- Sets fontSize on completion. See getFontSize.
-- Note: fontSize will be < 0 if the calculation failed!
-- @param text - (string) The text to get the optimal size for.
-- @return (MOAICoroutine) a running coroutine.
function FontRenderEngine:calculateOptimalFontSizeForText(text)
  return runInCoroutine(function()
    if not self:requiresFontSizeCalculation() then
      return
    end

    assert(not (self.isProcessing), "FontRenderEngine can process only one optimal size at a time.")
    assert(type(text) == "string", "'text' must be a string (got " .. type(text) .. ").")
    assert(self.font, "Must set font before calculating optimal size")
    assert(self.width, "Must set width before calculating optimal size.")
    assert(self.height, "Must set height before calculating optimal size.")
    assert(self.maxFontSize, "Must set max font size before calculating optimal size.")
    assert(self.minFontSize, "Must set min font size before calculating optimal size.")

    if self.fontSizeCalculatedForText ~= text then
      self.isProcessing = true
      self:resetProcess()

      local renderedFontSize = nil
      repeat
        renderedFontSize = self:processOptimalSize(text)
        coroutine.yield()
      until renderedFontSize

      if renderedFontSize <= 0 then
        print(string.format("Error: [%s] does not fit within %dx%d with font size range %d-%d",
          text, self.width, self.height, self.minFontSize, self.maxFontSize))
      end
      self:setFontSize(renderedFontSize / Host.contentScale)
      self.fontSizeCalculatedForText = text
      self.isProcessing = false
    end
  end)
end

function FontRenderEngine:requiresFontSizeCalculation()
  return self:getOperatingMode() == MODE_OPTIMAL_FONT_SIZE
end


-- Rendering --------------------------

--- Renders a new texture using the given string 'text' and the FontRenderEngine's current settings.
-- @param text - (string) the string to render.
-- @return (MOAITexture) the rendered texture
-- @return (number) The "rendered" width (not normalized to the host's content scale)
-- @return (number) The "rendered" height (not normalized to the host's content scale)
-- @return (table) A table containing glyph coordinate data for all characters in the string (Returned only if renderer:setReturnGlyphBounds(true) has been called)
function FontRenderEngine:renderTexture(text)
  assert(type(text) == "string", "'text' must be a string (got " .. type(text) .. ").")
  assert(self.font, "Must set font before creating a label")
  assert(self.fontSize, "Must set font size before creating a label")
  assert(self.fontSize > 0, "Font size is invalid. Did calculateOptimalFontSizeForText fail? [" .. text .. "]")

  local texture, renderedWidth, renderedHeight, glyphTable

  local mode = self:getOperatingMode()

  if mode == MODE_OPTIMAL_HEIGHT then
    local _, height = self.font:dimensionsWithMaxWidth(text, self.fontSize, self.width, self.wordBreakMode, nil, self.lineSpacing)
    -- "unofficially" set the height for the next render.
    self.super.setHeight(self, height * Host.contentScale)
    texture, glyphTable = self:render(text)
    renderedWidth, renderedHeight = self.width * Host.contentScale, height * Host.contentScale
  elseif self.isMultiLine then
    texture, glyphTable = self:render(text)
    renderedWidth, renderedHeight = self.width * Host.contentScale, self.height * Host.contentScale
  else
    texture, renderedWidth, renderedHeight, glyphTable = self:renderSingleLine(text)
  end

  return texture, renderedWidth, renderedHeight, glyphTable
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FontRenderEngine.new()
  local self = MOAITextRenderer.new()
  Mixin.mix(self, FontRenderEngine)

  -- Defaults
  self:setMultiLine(false)
  self:setAlignment(Alignment.LEFT, Alignment.TOP)
  self:setReturnGlyphBounds(false)
  self:setWordBreak(MOAITextBox.WORD_BREAK_NONE)
  self:setGranularity(1)
  self:setLineSpacing(1.0)

  self.isProcessing = false

  return self
end
