--- Difficulty ramping helper class.
--
-- Stores information about how the player is doing and offers a hook for calculating
-- a new difficulty setting based on this.
--
-- Properties of interest:
--
-- self.correctAnswerCount: Total number of correct answers since the game began.
-- self.correctAnswerStreak: Number of correct answers in a row. Resets on incorrect answer.
-- self.incorrectAnswerCount: Total number of incorrect answers since the game began.
-- self.incorrectAnswerStreak: Number of incorrect answers in a row. Resets on incorrect answer.
-- self.answerDurations: Array containing time taken to answer each question.
--
-- Example usage:
--
-- local updateFunction = function(ramp)
--  if ramp.correctAnswerStreak >= 2 then
--    ramp:addDifficulty(1)
--    ramp:resetCorrectAnswerStreak()
--  elseif ramp.incorrectAnswerStreak >= 2 then
--    ramp:addDifficulty(-1)
--    ramp:resetIncorrectAnswerStreak()
--  end
-- end
--
-- self.difficultyRamp = DifficultyRamp.new(minDifficulty, maxDifficulty, updateFunction)


-- ===========================================================================
class("DifficultyRamp")
-- ===========================================================================

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function recordAnswerTimeAndDuration(self)
  local time = self.timer:getTime()
  self.timer:stop()
  table.insert(self.answerDurations, time)
end


------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function DifficultyRamp:notifyQuestionStart()
  self.timer:stop()
  self.timer:start()
end

function DifficultyRamp:notifyCorrectAnswer()
  self.correctAnswerCount = self.correctAnswerCount + 1
  self.correctAnswerStreak = self.correctAnswerStreak + 1
  self:resetIncorrectAnswerStreak()
  recordAnswerTimeAndDuration(self)
  self.updateDifficultyFunction(self)
end

function DifficultyRamp:resetCorrectAnswerStreak()
  self.correctAnswerStreak = 0
end

function DifficultyRamp:notifyIncorrectAnswer()
  self.incorrectAnswerCount = self.incorrectAnswerCount + 1
  self.incorrectAnswerStreak = self.incorrectAnswerStreak + 1
  self:resetCorrectAnswerStreak()
  recordAnswerTimeAndDuration(self)
  self.updateDifficultyFunction(self)
end

function DifficultyRamp:resetIncorrectAnswerStreak()
  self.incorrectAnswerStreak = 0
end

--- Get difficulty as a value from 0 to 1
function DifficultyRamp:getNormalizedDifficulty()
  return self.normalizedDifficulty
end

function DifficultyRamp:setNormalizedDifficulty(normalizedDifficulty)
  self.normalizedDifficulty = MathUtils.clamp(normalizedDifficulty, 0, 1)
end

function DifficultyRamp:addNormalizedDifficulty(delta)
  local difficulty = self:getNormalizedDifficulty()
  self:setNormalizedDifficulty(difficulty + delta)
end

--- Get difficulty as a discrete integer
function DifficultyRamp:getDifficulty()
  local difficulty = self.minDifficulty + self.deltaDifficulty * self.normalizedDifficulty

  if difficulty > self.maxDifficulty - 0.001 then
    difficulty = self.maxDifficulty
  elseif difficulty < self.minDifficulty + 0.001 then
    difficulty = self.minDifficulty
  end

  return math.floor(difficulty)
end

function DifficultyRamp:setDifficulty(difficulty)
  self:setNormalizedDifficulty((difficulty - self.minDifficulty) / (self.maxDifficulty - self.minDifficulty))
end

function DifficultyRamp:addDifficulty(delta)
  local difficulty = self:getDifficulty()
  self:setDifficulty(difficulty + delta)
end


------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

-- Constructor ------------------------

--- Construct a new difficulty ramp.
-- @param minDifficulty - (integer) The minimum difficulty allowed.
-- @param maxDifficulty - (integer) The maximum difficulty allowed.
-- @param updateDifficultyFunction A function to update the current difficulty setting.
--                                 Params: self
--                                 User should modify difficulty from this function.
function DifficultyRamp.new(minDifficulty, maxDifficulty, updateDifficultyFunction)
  assert(updateDifficultyFunction, "Must provide a valid updateDifficultyFunction")

  local self = {}
  Mixin.mix(self, DifficultyRamp)

  self.minDifficulty = minDifficulty
  self.maxDifficulty = maxDifficulty
  self.deltaDifficulty = maxDifficulty - minDifficulty
  self.normalizedDifficulty = 0
  self.updateDifficultyFunction = updateDifficultyFunction

  self.correctAnswerCount = 0
  self.correctAnswerStreak = 0
  self.incorrectAnswerCount = 0
  self.incorrectAnswerStreak = 0
  self.answerDurations = {}

  self.timer = MOAITimer.new()
  self.timer:setSpan(0, 100000) -- Set span to a biggish number (70 days in this case)

  return self
end
