--- wait functions
-- @author Created by Aaron Barrett on January 15, 2014.

-- The wait() function is a utility that works just like MOAICoroutine.blockOnAction(),
-- except that if the parent coroutine is stopped while waiting on another action,
-- that action will be stopped as well.  If you do not wish to use this behavior,
-- simply use MOAICoroutine.blockOnAction() instead.

-- Remember to call start() on an action before waiting on it.  Blocking on a
-- stopped action will do nothing; the coroutine will simply proceed to the
-- next instruction.

-- ----------------------------------------------------------------------------
-- wait functions
-- ----------------------------------------------------------------------------

--- Stop updating the current coroutine until the specified action is no longer
-- busy.  Stopping the current coroutine will also stop any actions being waited on.
-- @param ... (MOAIAction) - Started action(s) to wait for.
function wait(...)
  -- "currentThread" is badly named. It actually returns the currently running MOAIAction.
  local runningMOAIAction = MOAICoroutine.currentThread()
  runningMOAIAction.waitedActions = arg

  for _, action in ipairs(runningMOAIAction.waitedActions) do
    assert(type(action) == "userdata", "Unexpected argument type passed to wait().  Must pass a MOAIAction.")
    if action:isActive() then
      MOAICoroutine.blockOnAction(action)
    end
  end

  runningMOAIAction.waitedActions = nil
end

--- Stop updating the current coroutine for the specified number of seconds.
-- @param duration (number) - the length of time to block execution in seconds.
function waitFor(duration)
  local delayAction = AHDelay.new(duration)
  wait(delayAction:start())
end

--- Stop updating the current coroutine for the specified number of frames.
-- @param framesToWait (number) - the number of frames to block execution
function waitForFrames(framesToWait)
  for frameIndex = 1, framesToWait do
    coroutine.yield()
  end
end

--- Stop updating the current coroutine while the specified function returns true.
-- @param functionReturningBoolean - A function that will eventually return false.
-- @param pollInterval (number) - How long to wait between function calls. nil = no wait.
function waitWhile(functionReturningBoolean, pollInterval)
  assert(type(functionReturningBoolean) == "function", "`functionReturningBoolean` must be a function.")

  while functionReturningBoolean() do
    if pollInterval and pollInterval > 0 then
      waitFor(pollInterval)
    else
      coroutine.yield()
    end
  end
end

--- Stop updating the current coroutine until the specified function returns true.
-- @param functionReturningBoolean - A function that will eventually return true.
-- @param pollInterval (number) - How long to wait between function calls. nil = no wait.
function waitUntil(functionReturningBoolean, pollInterval)
  assert(type(functionReturningBoolean) == "function", "`functionReturningBoolean` must be a function.")

  waitWhile(function() return not functionReturningBoolean() end, pollInterval)
end
