class("LinePathTracer")
require("graphics.LineTrajectoryBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SEGMENT_COLOR = {1, 1, 1, 1}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function createSegment(length, lineWidth)
  local view = FilledRectangle.new(length, lineWidth, SEGMENT_COLOR)
  return view
end
local function preloadViews(self)
  self.preloadedViews = {}
  local totalPoints = self.trajectoryBuilder:getSegmentCount()
  for i = 1, totalPoints do
    local view = createSegment(1, self.lineWidth)
    self:insert(view)
    self.preloadedViews[i] = view
    view:setVisible(false)
    view:inheritColor(self)
  end
end
local function getSegment(self, i, x0, y0, x1, y1)
  local dx, dy = x1 - x0, y1 - y0
  local l = MathUtils.vectorLength(dx, dy)
  local angle = ShapeUtils.getVectorDirectionAngle(dx, dy)

  local view
  if(self.preloadedViews) then
    view = self.preloadedViews[i]
  else
    view = createSegment(1, self.lineWidth)
    self:insert(view)
  end

  view:setVisible(true)
  view:setOpacity(1)
  view:setScl(l, 1)
  view:setRot(math.deg(angle))
  view:setLoc((x0 + x1)/2, (y0 + y1)/2)
  local x, y = view:getLoc()

  return view
end
local function draw(self, trajectory, startIndex, endIndex, viewList)
  local prevPoint = trajectory[startIndex]
  endIndex = math.min(#trajectory, endIndex)
  for i = startIndex + 1, endIndex do
    local currentPoint = trajectory[i]
    local x0, y0 = prevPoint[1], prevPoint[2]
    local x1, y1 = currentPoint[1], currentPoint[2]

    local view = getSegment(self, i - 1, x0, y0, x1, y1)
    prevPoint = currentPoint

    viewList[#viewList + 1] = view
  end
end

local function fade(self, viewList, prevInvisibleIndex, lastInvisibleViewIndex, fullyVisibleViewIndex)
  if(#viewList < 1) then return end
  for i = prevInvisibleIndex, lastInvisibleViewIndex do
    viewList[i]:setOpacity(0)
  end
  fullyVisibleViewIndex = math.min(fullyVisibleViewIndex, #viewList)
  local trailLength = fullyVisibleViewIndex - lastInvisibleViewIndex
  for i = lastInvisibleViewIndex, fullyVisibleViewIndex do
    local opacity = (i - lastInvisibleViewIndex)/trailLength
    viewList[i]:setOpacity(opacity)
  end
end

local function getTracer(self, trajectory, shouldFade, animParams)
  local totalPoints = #trajectory
  local prevDrawIndex = 1
  local viewList = {}
  local extendedTotalPoints = totalPoints + self.maxTailLength + self.maxVisibleViews - 1
  local prevInvisibleIndex = 1
  return function(t)
    local currentDrawIndex = math.floor((extendedTotalPoints - 1) * t) + 1
    draw(self, trajectory, prevDrawIndex, currentDrawIndex, viewList)

    if(shouldFade) then
      local lastInvisibleViewIndex = math.max(1, currentDrawIndex - self.maxVisibleViews - self.maxTailLength)
      local fullyVisibleViewIndex = math.min(currentDrawIndex - 1, lastInvisibleViewIndex + self.maxTailLength)
      fade(self, viewList, prevInvisibleIndex, lastInvisibleViewIndex, fullyVisibleViewIndex)
      prevInvisibleIndex = lastInvisibleViewIndex
    end

    prevDrawIndex = currentDrawIndex
    if(self.tipAdvanceHandler) then
      self.tipAdvanceHandler(trajectory, currentDrawIndex)
    end
  end
end
local function getFadingTracer(self, trajectory, animParams)
  return getTracer(self, trajectory, true, animParams)
end
local function getDrawingTracer(self, trajectory, animParams)
  return getTracer(self, trajectory, false, animParams)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinePathTracer:createTracePathAnimation(x0, y0, x1, y1, duration)
  local trajectory = self.trajectoryBuilder:connect2(x0, y0, x1, y1)
  local duration = duration or 1
  local totalPoints = #trajectory
  local animParams = {t = 0}
  local setterFunction = getFadingTracer(self, trajectory, animParams)

  local beamAnimation = AHCustomAction.new(duration, ease, setterFunction)

  return beamAnimation
end

function LinePathTracer:setTipAdvanceHandler(handler)
  self.tipAdvanceHandler = handler
end

-- ----------------------------------------------------------------------------
-- Constructor
-- ----------------------------------------------------------------------------
-- viewParams = {
-- lineWidth        = w,
-- maxVisibleViews  = linesToShowAtATime, -- number of segments with opacity 1 at a given time
-- maxTailLength    = fadedTailLength     -- number of segments with opacity between 0 and 0.99 at a given time
-- }
function LinePathTracer.new(viewParams, trajectoryBuilder)
  viewParams = viewParams or {}
  local self = Group.new(0, 0)
  Mixin.mix(self, LinePathTracer)

  self.trajectoryBuilder = trajectoryBuilder or LineTrajectoryBuilder.new()
  self.lineWidth = viewParams.lineWidth or 3
  -- Potentially make the fade feature disable-able
  self.maxVisibleViews = viewParams.maxVisibleViews or 8
  self.maxTailLength = viewParams.maxTailLength or 8
  preloadViews(self)

  return self
end
