--- MultiGradient
--
-- A group that contains a series of rectangle gradients positioned end-to-end.
-- Specify the colors to use at each of the key points along the MultiGradient.
-- You must also specify a table of relative thicknesses to use for the spacing between
-- each of the key points.
--
-- @author Created by Aaron Barrett on June 20, 2014.

class("MultiGradient")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MultiGradient:newGradientToAction(duration, ease, colors)
  assert(#colors == #self.bands + 1, "Incorrect number of colors provided for the multi gradient to action")

  local gradientAction = AHSpawn.new()
  for i, band in ipairs(self.bands) do
    gradientAction:addAction(AHSpawn.new(
      AHColorTo.new(band.beginColor, duration, ease, colors[i]),
      AHColorTo.new(band.endColor, duration, ease, colors[i + 1])
    ))
  end

  return gradientAction
end

function MultiGradient:createOpacityTo(duration, ease, opacity)
  local animation = AHSpawn.new()
  for i, band in ipairs(self.bands) do
    local beginOpacity = (self.colors[i][4] or 1) * opacity
    local endOpacity = (self.colors[i + 1][4] or 1) * opacity
    local bandAnimation = band:newOpacityToAction(duration, ease, beginOpacity, endOpacity)
    animation:addAction(bandAnimation)
  end
  return animation
end

function MultiGradient:setOpacity(opacity)
  self:createOpacityTo(0, nil, opacity):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a new MultiGradient that contains a series of RectangleGradients.
-- @param (number) width - the width of the MultiGradient
-- @param (number) height - the height of the MultiGradient
-- @param (boolean) isHorizontal - Whether the MultiGradient will arrange the colors from left-to-right or top-to-bottom.
-- @param (table) colors - A table containing the colors (as tables themselves) to use at the key points of the MultiGradient.
-- @param (table) thicknesses - A table containing numbers to use as the relative thicknesses of each of the RectangleGradients in the MultiGradient.
-- @return (Group) the new MultiGradient
function MultiGradient.new(width, height, isHorizontal, colors, thicknesses)
  assert(type(colors) == "table", "'colors' must be a table of colors (where each color is a table).")
  assert(type(thicknesses) == "table", "'thicknesses' must be a table of numbers (where each number represents the thickness of a gradient band).")
  assert(#colors == #thicknesses + 1, "There must be exactly one less thickness in 'thicknesses' as there are colors in 'colors'.")

  local self = nil

  local spacing = 0.0
  local alignment = Alignment.CENTER
  if (isHorizontal) then
    local growDirection = Row.GrowRight
    self = Row.new(spacing, alignment, growDirection)
  else
    local growDirection = Column.growDown
    self = Column.new(spacing, alignment, growDirection)
  end

  Mixin.mix(self, MultiGradient)

  local totalThickness = 0.0
  for i, thickness in ipairs(thicknesses) do
    totalThickness = totalThickness + thickness
  end

  local bands = {}
  for i, thickness in ipairs(thicknesses) do
    local beginColor = colors[i]
    local endColor = colors[i + 1]

    local thicknessProportion = thickness / totalThickness

    local bandWidth = nil
    local bandHeight = nil
    if (isHorizontal) then
      bandWidth = thicknessProportion * width
      bandHeight = height
    else
      bandWidth = width
      bandHeight = thicknessProportion * height
    end

    local band = RectangleGradient.new(bandWidth, bandHeight, beginColor, endColor, isHorizontal)
    self:insert(band)
    bands[#bands + 1] = band
  end

  self:alignChildren()
  self.bands = bands
  self.colors = colors

  return self
end
