--- MulticolorCircle

class("MulticolorCircle")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getPointsForCircle(centerCoordinate, radius, quality, startAngle)
  local points = {}
  local angle = startAngle
  local deltaAngle = 2 * math.pi/quality

  for i = 0, quality do
    angle = deltaAngle * i + startAngle
    local x = centerCoordinate + radius * math.cos(angle)
    local y = centerCoordinate + radius * math.sin(angle)
    local point = {x, y}
    table.insert(points, point)
  end

  return points
end

local function appendPointsAsFlatList(points, list)
  for _, point in ipairs(points) do
    table.insert(list, point)
  end
end

local function triangulateDonut(innerPoints, outerPoints)
  local triangles = {}
  for i = 1, #innerPoints - 1 do
    local triangle1 = {innerPoints[i], outerPoints[i], outerPoints[i + 1]}
    local triangle2 = {outerPoints[i + 1], innerPoints[i], innerPoints[i + 1]}
    appendPointsAsFlatList(triangle1, triangles)
    appendPointsAsFlatList(triangle2, triangles)
  end
  return triangles
end

local function getColorindex(normalizedPointIndex, colorDistribution)
  for i = 1, #colorDistribution do
    local normalizedColorIndex = colorDistribution[i]
    local nextNormalizedColorIndex = colorDistribution[i + 1] or 1
    if (normalizedPointIndex >= normalizedColorIndex and normalizedPointIndex <= nextNormalizedColorIndex) then
      return i
    end
  end
  return #colorDistribution
end

local function getColorForPoint(colors, pointIndex, totalPoints, colorDistribution)
  local normalizedIndex = (pointIndex - 1)/(totalPoints - 1)
  local colorCount = #colors
  local colorIndex = colorCount * normalizedIndex
  colorIndex = getColorindex(normalizedIndex, colorDistribution)
  if (colorIndex > #colors) then colorIndex = 1 end
  local nextColorIndex = colorIndex + 1


  local normalizedColorIndex = (colorIndex - 1)/#colors
  local normalizedNextColorIndex = (nextColorIndex - 1)/#colors
  local startPointIndex = totalPoints * colorDistribution[colorIndex]
  local nextPointIndex = totalPoints * (colorDistribution[nextColorIndex] or 1)

  if (nextColorIndex > #colors) then nextColorIndex = 1 end

  local pointsPerColor = nextPointIndex - startPointIndex + 1
  local colorProgress = (pointIndex - startPointIndex)/pointsPerColor

  local curColor = colors[colorIndex]
  local nextColor = colors[nextColorIndex]

  local r = curColor[1] * (1 - colorProgress) + nextColor[1] * colorProgress
  local g = curColor[2] * (1 - colorProgress) + nextColor[2] * colorProgress
  local b = curColor[3] * (1 - colorProgress) + nextColor[3] * colorProgress
  local a = curColor[4] * (1 - colorProgress) + nextColor[4] * colorProgress

  return {r, g, b, a}
end

local function writeVertexBuffer(points, colors, vertexBuffer, colorDistribution)
  vertexBuffer:reset()
  for i, point in ipairs(points) do
    local color = getColorForPoint(colors, i, #points, colorDistribution)
    local x, y = point[1], point[2]
    vertexBuffer:writeFloat(x, y)
    vertexBuffer:writeColor32(color[1], color[2], color[3], color[4])
  end
  vertexBuffer:bless()

end

local function makeMesh(points, colors, colorDistribution)
  local vertexFormat = MOAIVertexFormat.new()
  local vertexBuffer = MOAIVertexBuffer.new()
  vertexFormat:declareCoord(1, MOAIVertexFormat.GL_FLOAT, 2)
  vertexFormat:declareColor(3, MOAIVertexFormat.GL_UNSIGNED_BYTE)
  vertexBuffer:setFormat(vertexFormat)
  vertexBuffer:reserveVerts(#points)

  writeVertexBuffer(points, colors, vertexBuffer, colorDistribution)

  local whitePixel = MOAIImageTexture.new()
  whitePixel:init(1, 1)
  whitePixel:setRGBA(0, 0, 1, 1, 1, 1)
  whitePixel:invalidate()

  local mesh = MOAIMesh.new()
  mesh:setTexture(whitePixel)
  mesh:setVertexBuffer(vertexBuffer)
  mesh:setPrimType(MOAIMesh.GL_TRIANGLES)

  return mesh, vertexBuffer
end

local function createLayout(self, innerRadius, outerRadius, colors, quality, startAngle, colorDistribution)
  local centerCoordinate = outerRadius
  local innerPoints = getPointsForCircle(centerCoordinate, innerRadius, quality, startAngle)
  local outerPoints = getPointsForCircle(centerCoordinate, outerRadius, quality, startAngle)
  local triangulatedPoints = triangulateDonut(innerPoints, outerPoints)


  local mesh, vertexBuffer = makeMesh(triangulatedPoints, colors, colorDistribution)
  self:setDeck(mesh)
  self:setShader(MOAIShaderMgr.getShader(MOAIShaderMgr.MESH_SHADER))

  self.triangulatedPoints = triangulatedPoints
  self.vertexBuffer = vertexBuffer
end

local function getDefaultColorAngles(colorCount)
  local angles = {}
  local deltaAngle = 2.0 * math.pi / colorCount
  for i = 0, colorCount - 1 do
    local angle = i * deltaAngle
    table.insert(angles, angle)
  end
  return angles
end

local function getColorDistribution(colorAngles)
  local distribution = {}
  for _, angle in ipairs(colorAngles) do
    local convertedAngle = angle / (2.0 * math.pi)
    table.insert(distribution, convertedAngle)
  end
  return distribution
end

local function getColorsWithDefaultOpacity(originalColors)
  local colors = TableUtils.deepCopy(originalColors)
  for _, color in ipairs(colors) do
    color[4] = color[4] or 1
  end
  return colors
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Set the colors for this multicolor circle and their distribution. Opacity is optional, defaults to 1.
-- @param distribution (table) Array of values in [0-1], must have the same length as colors.
-- @param colors (table) Array of colors, must have the same length as distribution. Optional, defaults to current ones.
function MulticolorCircle:setColorDistribution(distribution, colors)
  self.colorDistribution = distribution
  if colors then
    self.colors = getColorsWithDefaultOpacity(colors)
  end
  writeVertexBuffer(self.triangulatedPoints, self.colors, self.vertexBuffer, distribution)
end

--- Get the colors for this multicolor circle.
-- @return (table) Array containing the colors used in this multicolor circle (copy).
function MulticolorCircle:getColors()
  return TableUtils.deepCopy(self.colors)
end

--- Set the colors for this multicolor circle, without changing the color distribution. Opacity is optional, defaults to 1.
-- @param colors (table) Array of colors, must have the same length as the current ones.
function MulticolorCircle:setColors(colors)
  assert(#colors == #self.colors, "If you change the number of colors you must supply a new distribution, use setColorDistribution instead.")
  self:setColorDistribution(self.colorDistribution, colors)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

-- Normalize all angle parameters to be in [0; 2 * math.pi]
function MulticolorCircle.new(innerRadius, outerRadius, colors, quality, startAngle, colorAngles, colorDistribution)
  quality = quality or 64
  startAngle = startAngle or 0
  colors = getColorsWithDefaultOpacity(colors)
  colorAngles = colorAngles or getDefaultColorAngles(#colors)
  colorDistribution = colorDistribution or getColorDistribution(colorAngles)

  local outerDiameter = outerRadius * 2.0
  local self = Group.new(outerDiameter, outerDiameter)
  Mixin.mix(self, MulticolorCircle)

  self.colors = colors
  self.colorDistribution = colorDistribution

  createLayout(self, innerRadius, outerRadius, colors, quality, startAngle, colorDistribution)

  return self
end
