--- A texture that renders offscreen based on a user-provided group hierarchy.
-- Any changes in the group will be reflected in the next render pass.
-- This special texture can be used anywhere a regular texture can be used.
--
-- Note: Like regular textures, dimensions are in pre-scaled units.


class("OffscreenRenderedTexture")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_CLEAR_COLOR = {0, 0, 0, 0}


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OffscreenRenderedTexture:setSize(selfW, selfH)

  self.viewport:setSize(selfW, selfH)
  -- Note: OpenGL origin is at the bottom, but image origin is at the top.
  -- We must compensate for this.
  self.viewport:setScale(selfW / Host.contentScale, -selfH / Host.contentScale)
  self.viewport:setOffset(-1, 1)

  self:init(selfW, selfH)
  self.width, self.height = selfW, selfH
end

function OffscreenRenderedTexture:getSize()
  return self.width, self.height
end

function OffscreenRenderedTexture:setRenderedGroup(group)
  if group ~= self.group then
    if self.group then
      self.group:removeSelf()
    end

    self.group = group
    self.layer.rootGroup = self.group

    if self.group ~= nil then
      group:setAnchorPointBottomLeft()
      group:setLoc(0, 0)
      Screen.sort(self.group, 1, self.layer)
    end
  end
end

--- Start rendering the group and any changes over time.
function OffscreenRenderedTexture:startRendering()
  self:setSize(self.width, self.height)
  self.layer:setViewport(self.viewport)
  self:setRenderTable({self.layer})
  MOAIRenderMgr.addToBufferTable(self)
end

--- Stop rendering any future changes to the group.
function OffscreenRenderedTexture:stopRendering()
  MOAIRenderMgr.removeFromBufferTable(self)
  self:setRenderTable({})
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a live texture renderer. It will update to match changes in the group.
-- @param group (prop) - The object to render.
-- @param width (optional number) - Width. Defaults to the group width * Host.contentScale.
-- @param height (optional number) - Height. Defaults to the group height * Host.contentScale.
-- @param clearColor (optional color) - Clear color. Defaults to {0, 0, 0, 0}.
-- @return a texture renderer.
function OffscreenRenderedTexture.new(group, width, height, clearColor)
  clearColor = clearColor or DEFAULT_CLEAR_COLOR

  local self = MOAIFrameBufferTexture.new()
  Mixin.mix(self, OffscreenRenderedTexture)

  self.width = width or group:getWidth() * Host.contentScale
  self.height = height or group:getHeight() * Host.contentScale

  self.viewport = MOAIViewport.new()
  self.layer = MOAILayer2D.new()

  Screen.addManagedLayer(self.layer)

  self:setClearColor(unpack(clearColor))
  self:setRenderedGroup(group)

  self:startRendering()

  return self
end
