--- A rectangle prop that has a gradient
-- @author Created by Aaron Barrett on March 20, 2013

-- ===========================================================================
class("RectangleGradient")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function toMOAIColor(color)
  if type(color) == "table" then
    return ColorUtils.tableToMOAIColor(color)
  end
  return color
end

local function setDrawCallback(self)
  local drawFunction = MOAIDraw.fillVerticalRectangularGradient
  if self.isHorizontal then
    drawFunction = MOAIDraw.fillHorizontalRectangularGradient
  end

  self.deck:setDrawCallback(function()
    drawFunction(0.0, self:getHeight(), self:getWidth(), 0.0, self.beginColor, self.endColor)
  end)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Returns the top or left color of the gradient as a MOAIColor.
-- Attribute links/colorTo()/colorBy() may be applied to this color, and the gradient will update its appearance automatically.
-- @return (MOAIColor) the top color of the gradient (if vertical) or the left color (if horizontal)
function RectangleGradient:getBeginColor()
  return self.beginColor
end

--- Set the top or left color of the gradient.
--- @param ... (r, g, b, (a)) - The color components
function RectangleGradient:setBeginColor(...)
  self.beginColor:setColor(...)
end

--- Returns the bottom or right color of the gradient as a MOAIColor.
-- Attribute links/colorTo()/colorBy() may be applied to this color, and the gradient will update its appearance automatically.
-- @return (MOAIColor) the bottom color of the gradient (if vertical) or the right color (if horizontal)
function RectangleGradient:getEndColor()
  return self.endColor
end

--- Set the bottom or right color of the gradient.
--- @param ... (r, g, b, (a)) - The color components
function RectangleGradient:setEndColor(...)
  self.endColor:setColor(...)
end

-- Actions ----------------------------

function RectangleGradient:newGradientToAction(duration, ease, beginColor, endColor)
  return AHSpawn.new(
    AHColorTo.new(self.beginColor, duration, ease, beginColor),
    AHColorTo.new(self.endColor, duration, ease, endColor)
  )
end

function RectangleGradient:newGradientByAction(duration, ease, beginColorDelta, endColorDelta)
  return AHSpawn.new(
    AHColorBy.new(self.beginColor, duration, ease, beginColorDelta),
    AHColorBy.new(self.endColor, duration, ease, endColorDelta)
  )
end

function RectangleGradient:newOpacityToAction(duration, ease, beginOpacity, endOpacity)
  return AHSpawn.new(
    AHOpacityTo.new(self.beginColor, duration, ease, beginOpacity),
    AHOpacityTo.new(self.endColor, duration, ease, endOpacity)
  )
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a rectangle that is filled with a vertical or horizontal gradient running from beginColor to endColor.
-- Colors can be passed either as MOAIColors or as tables in the format {r, g, b, (o)}.
-- @param width - (number) the width of the rectangle
-- @param height - (number) the height of the rectangle
-- @param beginColor - (table, or MOAIColor) The first color of the gradient.  An array of 4 elements - {r, g, b, a}, in the range of [0, 1]. Alpha is optional, defaults to 1.  An instance of MOAIColor may be used instead.
-- @param endColor - (table, or MOAIColor) The second color of the gradient.  An array of 4 elements - {r, g, b, a}, in the range of [0, 1]. Alpha is optional, defaults to 1.  An instance of MOAIColor may be used instead.
-- @param isHorizontal (optional boolean, defaults to false) Draws the gradient from left-to-right if true, top-to-bottom if false.
function RectangleGradient.new(width, height, beginColor, endColor, isHorizontal)
  local self = Group.new(width, height)
  Mixin.mix(self, RectangleGradient)

  self.beginColor = toMOAIColor(beginColor)
  self.endColor = toMOAIColor(endColor)
  self.isHorizontal = isHorizontal

  self.deck = MOAIScriptDeck.new()
  self.deck:setRect(0.0, 0.0, width, height)
  self:setDeck(self.deck)
  setDrawCallback(self)

  return self
end
