class("ShapeUtils")
------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

local DEFAULT_CIRCLE_POINT_COUNT = 60


------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

--- Flatten an array of ordered pairs into a single flat array.
-- Example: {{1, 2}, {3, 4}} becomes {1, 2, 3, 4}
function ShapeUtils.flattenCoordPairs(arrayOfPairs)
  local flatArray = {}
  for _, point in ipairs(arrayOfPairs) do
    flatArray[#flatArray + 1] = point[1]
    flatArray[#flatArray + 1] = point[2]
  end
  return flatArray
end

-- TODO: This is a mid-PR kludge to avoid refactoring tons of legacy code until next PR.
-- Likely, this function will become the new flattenCoordPairs and the old one will go away.
function ShapeUtils.flattenCoordPairsNew(arrayOfPairs)
  local flatArray = {}
  for _, point in ipairs(arrayOfPairs) do
    flatArray[#flatArray + 1] = point.x
    flatArray[#flatArray + 1] = point.y
  end
  return flatArray
end

function ShapeUtils.getMaxEdges(points)
  local maxX, maxY = -math.huge, -math.huge
  for _, point in ipairs(points) do
    maxX = math.max(maxX, point.x)
    maxY = math.max(maxY, point.y)
  end
  return maxX, maxY
end

--- Scale in all points to apply a margin along the edge.
-- Used to compensate for pen width when drawing.
-- Assumes that the center of the object is at maxX/2, maxY/2.
function ShapeUtils.applyMargin(points, margin, maxX, maxY)
  if margin ~= 0 then
    local midX, midY = maxX / 2, maxY / 2
    local proportionX, proportionY = (midX - margin) / midX, (midY - margin) / midY

    for _, point in ipairs(points) do
      point.x = MathUtils.interpolateLinearFromNormalized(proportionX, midX, point.x)
      point.y = MathUtils.interpolateLinearFromNormalized(proportionY, midY, point.y)
    end
  end
end

function ShapeUtils.scalePoints(points, width, height)
  for _, point in ipairs(points) do
    point.x = point.x * width
    point.y = point.y * height
  end
end

function ShapeUtils.normalizePoints(points, width, height)
  for _, point in ipairs(points) do
    point.x = point.x / width
    point.y = point.y / height
  end
end

--- Create a series of points along an arc.
-- If startX and startY are not provided, the points are adjusted such that the lowest point is 0.
-- If pointCount is not specified, it will choose a number of points based off a reference of
-- 60 points for a full circle. For large radii, you may want to choose more.
-- @param startAngle (number, radians) - The starting angle of the arc.
-- @param endAngle (number, radians) - The ending angle of the arc.
-- @param radius (number) - The radius of the arc.
-- @param startX (number) - The "center" X coordinate of the arc (optional).
-- @param startY (number) - The "center" Y coordinate of the arc (optional).
-- @param pointCount (number) - The number of points to generate along the arc (optional)
-- @param points (array) - array to add the points to (optional).
-- @return (table) - An array of flattened coordinates: x1, y1, x2, y2, ...
function ShapeUtils.makeArcPoints(startAngle, endAngle, radius, startX, startY, pointCount, points)
  assert(endAngle >= startAngle, "endAngle must be >= startAngle")
  points = points or {}
  local angleRange = endAngle - startAngle
  pointCount = pointCount or math.ceil(DEFAULT_CIRCLE_POINT_COUNT * angleRange / (math.pi * 2))
  local sideCount = pointCount - 1

  local originX, originY = startX or radius, startY or radius

  local lowX, lowY = math.huge, math.huge
  for i = 0, pointCount - 1 do
    local angle = startAngle + i * angleRange / sideCount
    local x, y = originX + radius * math.cos(angle), originY + radius * math.sin(angle)
    table.insert(points, {x = x, y = y})
    lowX = math.min(lowX, x)
    lowY = math.min(lowY, y)
  end

  if not startX or not startY then
    for _, point in ipairs(points) do
      if not startX then
        point.x = point.x - lowX
      end
      if not startY then
        point.y = point.y - lowY
      end
    end
  end

  return points
end

--- Create a series of points forming a checkmark.
-- @param width (number) - The width of the checkmark.
-- @param shortSegmentProportion (number) - The proportion of the width to use for the short checkmark segment.
-- @param thicknessProportion (number) - The thickness of the checkmark, measured as a proportion of width.
function ShapeUtils.makeCheckmarkPoints(width, shortSegmentProportion, thicknessProportion)
  local shortPos = width * shortSegmentProportion
  local longPos = width * (1 - shortSegmentProportion)
  local thickness = width * thicknessProportion
  local thickPos = math.sqrt(thickness * thickness / 2)

  return {
    {x = shortPos, y = 0},
    {x = width, y = longPos},
    {x = width - thickPos, y = longPos + thickPos},
    {x = shortPos, y = thickPos * 2},
    {x = thickPos, y = shortPos + thickPos},
    {x = 0, y = shortPos}
  }
end

--- Returns the angle of a vector to the X axis in radians
-- @param dx - x length of the vector
-- @param dy - y length of the vector
function ShapeUtils.getVectorDirectionAngle(dx, dy)
  return math.atan2(dy, dx)
end
