-- ===========================================================================
class("SimpleMorphingPolygon")
-- ===========================================================================

-------------------------------------------------------------------------------
-- Configuration
-------------------------------------------------------------------------------

local MIN_VERTICES = 3


-------------------------------------------------------------------------------
-- Utility
-------------------------------------------------------------------------------

local function getBoundsOfVertices(vertexTable)
  local minX, minY = math.huge, math.huge
  local maxX, maxY = -math.huge, -math.huge

  for _, vertex in pairs(vertexTable) do
    local x = vertex.x
    local y = vertex.y

    minX = math.min(x, minX)
    minY = math.min(y, minY)
    maxX = math.max(x, maxX)
    maxY = math.max(y, maxY)
  end

  return minX, minY, maxX, maxY
end

local function getInterpolatedVertices(t, startVertices, endVertices)
  local interpolatedVertices = {}
  local numberOfVertices = #startVertices
  for i = 1, numberOfVertices do
    local vertex1 = startVertices[i]
    local vertex2 = endVertices[i]

    local blendedX = MathUtils.interpolateLinearFromNormalized(t, vertex1.x, vertex2.x)
    local blendedY = MathUtils.interpolateLinearFromNormalized(t, vertex1.y, vertex2.y)

    local blendedVertex =
    {
      x = blendedX,
      y = blendedY
    }
    table.insert(interpolatedVertices, blendedVertex)
  end

  return interpolatedVertices
end

local function getDrawPointsTable(self, vertexTable)
  local drawPoints = {}
  local numberOfVertices = #vertexTable
  local numberOfTriangles = numberOfVertices - 2
  local w, h = self:getDims()
  local halfW = w * 0.5
  local halfH = h * 0.5
  for triangleIndex = 1, numberOfTriangles do
    local vertexIndex = 1
    table.insert(drawPoints, vertexTable[vertexIndex].x + halfW)
    table.insert(drawPoints, vertexTable[vertexIndex].y + halfH)

    vertexIndex = triangleIndex + 1
    table.insert(drawPoints, vertexTable[vertexIndex].x + halfW)
    table.insert(drawPoints, vertexTable[vertexIndex].y + halfH)

    vertexIndex = triangleIndex + 2
    table.insert(drawPoints, vertexTable[vertexIndex].x + halfW)
    table.insert(drawPoints, vertexTable[vertexIndex].y + halfH)
  end

  return drawPoints
end

local function createLayout(width, height, startVertices, endVertices)
  local self = Group.new(width, height)

  self.startVertices = startVertices
  self.endVertices = endVertices

  local deck = MOAIScriptDeck.new()
  deck:setRect(0.0, 0.0, width, height)
  self:setDeck(deck)

  local function drawCallbackWrapper()
    local t = self.t or 0
    t = MathUtils.clamp(t, 0.0, 1.0)
    local interpolatedVertices = getInterpolatedVertices(t, self.startVertices, self.endVertices)
    local drawPoints = getDrawPointsTable(self, interpolatedVertices)

    MOAIDraw.fillFan(unpack(drawPoints))
  end
  deck:setDrawCallback(drawCallbackWrapper)

  return self
end


-------------------------------------------------------------------------------
-- API
-------------------------------------------------------------------------------

function SimpleMorphingPolygon:getMorphAmount()
  return self.t
end

function SimpleMorphingPolygon:setMorphAmount(amount)
  self.t = amount
end

function SimpleMorphingPolygon:createMorphToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getMorphAmount, self.setMorphAmount, ease, amount)
end

function SimpleMorphingPolygon:createMorphByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getMorphAmount, self.setMorphAmount, ease, amount)
end


-------------------------------------------------------------------------------
-- Object Creation
-------------------------------------------------------------------------------

function SimpleMorphingPolygon.new(width, height, startVertices, endVertices)
  assert(startVertices and #startVertices >= MIN_VERTICES, "Invalid table of start vertices")
  assert(endVertices and #endVertices >= MIN_VERTICES, "Invalid table of end vertices")
  assert(#startVertices == #endVertices, "Number of start and end vertices must be equal")

  local self = createLayout(width, height, startVertices, endVertices)
  Mixin.mix(self, SimpleMorphingPolygon)

  self.t = 0

  return self
end
