--- Tiled image.
-- Creates a deck with a repeating pattern based on a tile image.
-- Tiles repeat from the top left towards the bottom right.
-- Note: The tile image MUST have power-of-2 dimensions.

-- ===========================================================================
class("TiledImage")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local RENDERABLE_TILED_DIMENSIONS_SET = {}
local MAX_EXPONENT_BEFORE_OVERFLOW = 62

local function initializeRenderableTiledDimensions()
  for n = 0, MAX_EXPONENT_BEFORE_OVERFLOW do
    local powerOfTwo = 2^n
    RENDERABLE_TILED_DIMENSIONS_SET[powerOfTwo] = true
  end
end
initializeRenderableTiledDimensions()

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isRenderableTiledDimension(dimension)
  return (RENDERABLE_TILED_DIMENSIONS_SET[dimension] ~= nil)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TiledImage:setBounds(lowX, lowY, lowZ, highX, highY, highZ)
  local width = highX - lowX
  local height = highY - lowY

  local numberOfRepeatsX = width / self.textureWidth
  local numberOfRepeatsY = height / self.textureHeight

  self.backgroundDeck:setRect(0.0, 0.0, width, height)
  self.backgroundDeck:setUVRect(0.0, numberOfRepeatsY, numberOfRepeatsX, 0.0)

  return self.super.setBounds(self, lowX, lowY, lowZ, highX, highY, highZ)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new tiled image.
-- @param tilePath Path to the image to tile.
-- @param width The width of the tiled image to create. Defaults to Screen.width.
-- @param height The height of the tiled image to create. Defaults to Screen.height.
-- @param textureScale What scale to display the tile at. Defaults to 1.
-- @param growToFit If true, increase the size beyond width and height to match the edge of a (scaled) tile exactly. Default false.
-- @return A new tiled image.
function TiledImage.newWithFilename(tilePath, width, height, textureScale, growToFit)
  local texture = TextureLoader.loadTexture(tilePath)
  return TiledImage.newWithTexture(texture, width, height, textureScale, growToFit)
end

--- Create a new tiled image.
-- @param texture The texture to tile.
-- @param width The width of the tiled image to create. Defaults to Screen.width.
-- @param height The height of the tiled image to create. Defaults to Screen.height.
-- @param textureScale What scale to display the tile at. Defaults to 1.
-- @param growToFit If true, increase the size beyond width and height to match the edge of a (scaled) tile exactly. Default false.
-- @return A new tiled image.
function TiledImage.newWithTexture(texture, width, height, textureScale, growToFit)
  textureScale = textureScale or 1
  local width = width or Screen.width
  local height = height or Screen.height

  texture:setWrap(true)

  local textureWidth, textureHeight = texture:getSize()
  assert(isRenderableTiledDimension(textureWidth), "OpenGL requires the width of tiled textures to be a power of two on some devices.")
  assert(isRenderableTiledDimension(textureHeight), "OpenGL requires the height of tiled textures to be a power of two on some devices.")
  textureWidth = textureWidth / Host.contentScale * textureScale
  textureHeight = textureHeight / Host.contentScale * textureScale

  local numberOfRepeatsX = width / textureWidth
  local numberOfRepeatsY = height / textureHeight

  if growToFit then
    numberOfRepeatsX = math.floor(numberOfRepeatsX)
    if numberOfRepeatsX * textureWidth < width then
      numberOfRepeatsX = numberOfRepeatsX + 1
      width = numberOfRepeatsX * textureWidth
    end
    numberOfRepeatsY = math.floor(numberOfRepeatsY)
    if numberOfRepeatsY * textureHeight < height then
      numberOfRepeatsY = numberOfRepeatsY + 1
      height = numberOfRepeatsY * textureHeight
    end
  end

  local backgroundDeck = MOAIGfxQuad2D.new()
  backgroundDeck:setTexture(texture)
  backgroundDeck:setRect(0.0, 0.0, width, height)
  backgroundDeck:setUVRect(0.0, numberOfRepeatsY, numberOfRepeatsX, 0.0)

  local self = Group.new(width, height)
  Mixin.mix(self, TiledImage)
  self:setDeck(backgroundDeck)

  self.backgroundDeck = backgroundDeck
  self.textureWidth = textureWidth
  self.textureHeight = textureHeight

  return self
end
