--- Arc
-- Builds a circular arc with configurable start and end angles.

require("graphics.primitives.GraphicsPrimitive")
require("graphics.ShapeUtils")


-- ===========================================================================
local ArcCommon = privateClass()
-- ===========================================================================

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function ArcCommon:getStartAngle()
  return self.startAngle
end

function ArcCommon:setStartAngle(startAngle)
  self.startAngle = startAngle
  self:updateGraphicsPrimitive()
end

function ArcCommon:getEndAngle()
  return self.endAngle
end

function ArcCommon:setEndAngle(endAngle)
  self.endAngle = endAngle
  self:updateGraphicsPrimitive()
end

--- The number of points to use when rendering the arc.
-- nil = let ShapeUtils.makeArcPoints decide.
function ArcCommon:getPointCount()
  return self.pointCount
end

function ArcCommon:setPointCount(pointCount)
  self.pointCount = pointCount
  self:updateGraphicsPrimitive()
end

-- Actions ----------------------------

function ArcCommon:newStartAngleToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getStartAngle, self.setStartAngle, ease, amount)
end

function ArcCommon:newStartAngleByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getStartAngle, self.setStartAngle, ease, amount)
end

function ArcCommon:newEndAngleToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getEndAngle, self.setEndAngle, ease, amount)
end

function ArcCommon:newEndAngleByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getEndAngle, self.setEndAngle, ease, amount)
end


-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function ArcCommon.makeArc(self, startAngle, endAngle)
  CircularPrimitive.makeCircularPrimitive(self)
  Mixin.mix(self, ArcCommon)

  self.startAngle = startAngle
  self.endAngle = endAngle
end



-- ===========================================================================
class("FilledArc")
-- ===========================================================================

function FilledArc:updateGraphicsPrimitive()
  local points = {{x=0.5, y=0.5}}
  ShapeUtils.makeArcPoints(self.startAngle, self.endAngle, 0.5, 0.5, 0.5, self.pointCount, points)
  self._normalizedPointLists = {points}
  self.super.updateGraphicsPrimitive(self)
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new filled arc.
-- @param radius (number) - The arc's radius
-- @param startAngle (number) - The starting angle in radians.
-- @param endAngle (number) - The ending angle in radians.
-- @param color (table) - The color to apply to this object (optional)
function FilledArc.new(radius, startAngle, endAngle, color)
  local self = FilledGraphicsPrimitive.new(radius * 2, radius * 2, color)
  self.hasCenterPoint = true

  ArcCommon.makeArc(self, startAngle, endAngle)
  Mixin.mix(self, FilledArc)
  self:updateGraphicsPrimitive()

  return self
end



-- ===========================================================================
class("HollowArc")
-- ===========================================================================

function HollowArc:updateGraphicsPrimitive()
  local isFullCircle = self.startAngle == self.endAngle
  self:setLooped(isFullCircle or self.hasCenterPoint)
  local points = {}
  if not isFullCircle and self.hasCenterPoint then
    points = {{x=0.5, y=0.5}}
  end
  ShapeUtils.makeArcPoints(self.startAngle, self.endAngle, 0.5, 0.5, 0.5, self.pointCount, points)
  if isFullCircle then
    table.remove(points)
  end
  self._normalizedPointLists = {points}
  self.super.updateGraphicsPrimitive(self)
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new hollow arc.
-- @param radius (number) - The arc's radius
-- @param startAngle (number) - The starting angle in radians.
-- @param endAngle (number) - The ending angle in radians.
-- @param color (table) - The color to apply to this object (optional)
-- @param lineWidth (number) - The width of the line to draw the arc with,
--                             Regardless of line width, the arc will never draw
--                             outside of its bounds (default 1.0)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0.0)
-- @param hasCenterPoint (boolean) - If true, put a center point in the arc (default true)
function HollowArc.new(radius, startAngle, endAngle, color, lineWidth, blurMargin, hasCenterPoint)
  if hasCenterPoint == nil then
    hasCenterPoint = true
  end

  local self = HollowGraphicsPrimitive.new(radius * 2, radius * 2, color, lineWidth, blurMargin)
  self.hasCenterPoint = hasCenterPoint

  ArcCommon.makeArc(self, startAngle, endAngle)
  Mixin.mix(self, HollowArc)
  self:updateGraphicsPrimitive()

  return self
end
