--- GraphicsPrimitive
-- Provides the base classes for filled and hollow graphics primitives.
--
-- Subclasses must store the draw points normalized in an array of arrays of
-- points in self._normalizedPointLists.
-- Your subclass must also call self:updateGraphicsPrimitive() in its constructor.
--
-- Subclasses may override updateGraphicsPrimitive(), but must call the superclass
-- updateGraphicsPrimitive() AFTER creating/updating self._normalizedPointLists.
--
-- @author Created by karl on November 07, 2014.


-- ============================================================================
local GraphicsPrimitive = privateClass()
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function GraphicsPrimitive:updateGraphicsPrimitive()
  assert(self._normalizedPointLists, "Must build self._normalizedPointLists before calling updateGraphicsPrimitive")
  local selfW, selfH = self:getDims()
  local midX, midY, proportionX, proportionY = 0, 0, 1, 1
  if self.shouldApplyAutoMargin then
    local margin = self:getMargin()
    midX = selfW / 2
    midY = selfH / 2
    proportionX = (midX - margin) / midX
    proportionY = (midY - margin) / midY
  end
  local coordinateLists = {}
  for _, points in ipairs(self._normalizedPointLists) do
    local coordinates = {}
    table.insert(coordinateLists, coordinates)
    for _, point in ipairs(points) do
      -- Scale up & apply margin in one go for speed.
      local x = ((point["x"] * selfW) - midX) * proportionX + midX
      local y = ((point["y"] * selfH) - midY) * proportionY + midY
      table.insert(coordinates, x)
      table.insert(coordinates, y)
    end
  end
  self._graphicsPrimitiveCoordinateLists = coordinateLists
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GraphicsPrimitive:setBounds(minX, minY, minZ, maxX, maxY, maxZ)
  self.super.setBounds(self, minX, minY, minZ, maxX, maxY, maxZ)
  self:updateGraphicsPrimitive()
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GraphicsPrimitive.new(width, height, color)
  local self = Group.newCustomDraw(width, height, color)
  Mixin.mix(self, GraphicsPrimitive)

  return self
end



-- ============================================================================
class("FilledGraphicsPrimitive")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new filled graphics primitive.
-- Note: You must call self:updateGraphicsPrimitive() in your constructor!
-- @param width (number) - The group's width
-- @param height (number) - The group's height
-- @param color (table) - The color to apply to this object (optional)
function FilledGraphicsPrimitive.new(width, height, color)
  local self = GraphicsPrimitive.new(width, height, color)
  Mixin.mix(self, FilledGraphicsPrimitive)
  self.deck:setDrawCallback(function()
    for _, coordinates in ipairs(self._graphicsPrimitiveCoordinateLists) do
      MOAIDraw.fillFan(coordinates)
    end
  end)
  self._graphicsPrimitiveCoordinateLists = {}

  return self
end



-- ============================================================================
class("HollowGraphicsPrimitive")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function HollowGraphicsPrimitive:setLooped(isLooped)
  local drawFunction = MOAIDraw.drawBeveledLine
  if isLooped then
    drawFunction = MOAIDraw.drawBeveledLineLoop
  end
  self.deck:setDrawCallback(function()
    for _, coordinates in ipairs(self._graphicsPrimitiveCoordinateLists) do
      drawFunction(coordinates, self.lineWidth, self.blurMargin)
    end
  end)
end

--- If true, primitive will attempt to add a margin to accomodate the line width.
function HollowGraphicsPrimitive:getShouldApplyAutoMargin()
  return self.shouldApplyAutoMargin
end

function HollowGraphicsPrimitive:setShouldApplyAutoMargin(shouldApplyAutoMargin)
  self.shouldApplyAutoMargin = shouldApplyAutoMargin
end

function HollowGraphicsPrimitive:getMargin()
  return self:getLineWidth() * 0.5
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function HollowGraphicsPrimitive:getLineWidth()
  return self.lineWidth
end

function HollowGraphicsPrimitive:setLineWidth(lineWidth)
  self.lineWidth = lineWidth
  self:updateGraphicsPrimitive()
end

function HollowGraphicsPrimitive:getBlurMargin()
  return self.blurMargin
end

function HollowGraphicsPrimitive:setBlurMargin(blurMargin)
  self.blurMargin = blurMargin
end

function HollowGraphicsPrimitive:newLineWidthToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getLineWidth, self.setLineWidth, ease, amount)
end

function HollowGraphicsPrimitive:newLineWidthByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getLineWidth, self.setLineWidth, ease, amount)
end

function HollowGraphicsPrimitive:newBlurMarginToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getBlurMargin, self.setBlurMargin, ease, amount)
end

function HollowGraphicsPrimitive:newBlurMarginByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getBlurMargin, self.setBlurMargin, ease, amount)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new hollow graphics primitive.
-- Note: You must call self:updateGraphicsPrimitive() in your constructor!
-- @param width (number) - The group's width
-- @param height (number) - The group's height
-- @param color (table) - The color to apply to this object (optional)
-- @param lineWidth (number) - The width of the line to draw the primitive with,
--                             Regardless of line width, the primitive will never draw
--                             outside of its bounds (default 1.0)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0.0)
-- @param isLooped (boolean) - If true, the draw routine will loop from the last point to the first.
function HollowGraphicsPrimitive.new(width, height, color, lineWidth, blurMargin, isLooped)
  lineWidth = lineWidth or 1.0
  blurMargin = blurMargin or 0.0
  local margin = lineWidth * 0.5

  local self = GraphicsPrimitive.new(width, height, color)
  Mixin.mix(self, HollowGraphicsPrimitive)
  self.lineWidth = lineWidth
  self.blurMargin = blurMargin
  self:setShouldApplyAutoMargin(true)
  self._graphicsPrimitiveCoordinateLists = {}
  self:setLooped(isLooped)

  return self
end



-- ============================================================================
class("CircularPrimitive")
-- ============================================================================

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function CircularPrimitive:getRadius()
  return self:getWidth() / 2
end

function CircularPrimitive:setRadius(radius)
  self:setDims(radius * 2, radius * 2)
end

function CircularPrimitive:newRadiusToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getRadius, self.setRadius, ease, amount)
end

function CircularPrimitive:newRadiusByAction(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getRadius, self.setRadius, ease, amount)
end


-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

function CircularPrimitive.makeCircularPrimitive(self)
  Mixin.mix(self, CircularPrimitive)
  return self
end
