--- Polygon
-- A generic polygon generation API.
-- @author Created by karl on November 26, 2014.


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function newNormalizedPoints(points, width, height)
  points = TableUtils.deepCopy(points)
  ShapeUtils.normalizePoints(points, width, height)
  return points
end



-- ===========================================================================
class("FilledPolygon")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new filled polygon.
-- @param width (number) - The width of the draw area.
-- @param height (number) - The height of the draw area.
-- @param points (array of flattened coordinates) - The coordinates of the shape to draw.
-- @param color (table) - The color to apply to this object (optional)
function FilledPolygon.new(width, height, points, color)
  local self = FilledGraphicsPrimitive.new(width, height, color)
  Mixin.mix(self, FilledPolygon)
  self._normalizedPointLists = {newNormalizedPoints(points, width, height)}
  self:updateGraphicsPrimitive()

  return self
end

--- Create a new filled polygon, automatically choosing the width & height.
-- Width and height will be determined by the maximums along the X and Y axis.
-- @param points (array of flattened coordinates) - The coordinates of the shape to draw.
-- @param color (table) - The color to apply to this object (optional)
function FilledPolygon.newAutomatic(points, color)
  local width, height = ShapeUtils.getMaxEdges(points)
  local self = FilledGraphicsPrimitive.new(width, height, color)
  Mixin.mix(self, FilledPolygon)
  self._normalizedPointLists = {newNormalizedPoints(points, width, height)}
  self:updateGraphicsPrimitive()

  return self
end



-- ===========================================================================
class("HollowPolygon")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new hollow polygon.
-- @param width (number) - The width of the draw area.
-- @param height (number) - The height of the draw area.
-- @param points (array) - The points to draw.
-- @param color (table) - The color to apply to this object (optional)
-- @param lineWidth (number) - The width of the line to draw the arc with,
--                             Regardless of line width, the arc will never draw
--                             outside of its bounds (default 1.0)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0.0)
-- @param isLooped (boolean) - If true, the draw routine will loop from the last point to the first.
-- @param shouldApplyAutomargin (boolean) - If true, apply automargin (default false)
function HollowPolygon.new(width, height, points, color, lineWidth, blurMargin, isLooped, shouldApplyAutomargin)
  local self = HollowGraphicsPrimitive.new(width, height, color, lineWidth, blurMargin, isLooped)
  Mixin.mix(self, HollowPolygon)
  self._normalizedPointLists = {newNormalizedPoints(points, width, height)}
  self:setShouldApplyAutoMargin(shouldApplyAutomargin)
  self:updateGraphicsPrimitive()

  return self
end

--- Create a new hollow polygon, automatically choosing the width & height.
-- Width and height will be determined by the maximums along the X and Y axis.
-- @param points (array) - The points to draw.
-- @param color (table) - The color to apply to this object (optional)
-- @param lineWidth (number) - The width of the line to draw the arc with,
--                             Regardless of line width, the arc will never draw
--                             outside of its bounds (default 1.0)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0.0)
-- @param isLooped (boolean) - If true, the draw routine will loop from the last point to the first.
function HollowPolygon.newAutomatic(points, color, lineWidth, blurMargin, isLooped)
  local width, height = ShapeUtils.getMaxEdges(points)
  local self = HollowGraphicsPrimitive.new(width, height, color, lineWidth, blurMargin, isLooped)
  Mixin.mix(self, HollowPolygon)
  self._normalizedPointLists = {newNormalizedPoints(points, width, height)}
  self:updateGraphicsPrimitive()

  return self
end
