--- A rectangle with rounded corners
-- @author Created by Alexey Chernikov on May 10, 2013

-- ===========================================================================
local RoundedRectangleCommon = privateClass()
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

RoundedRectangleCommon.CORNERS = {
  TOP_LEFT = "RoundedRectangleCommon.CORNERS.TOP_LEFT",
  TOP_RIGHT = "RoundedRectangleCommon.CORNERS.TOP_RIGHT",
  BOTTOM_LEFT = "RoundedRectangleCommon.CORNERS.BOTTOM_LEFT",
  BOTTOM_RIGHT = "RoundedRectangleCommon.CORNERS.BOTTOM_RIGHT"
}

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RoundedRectangleCommon:updateGraphicsPrimitive(sharpCornersList)
  local selfW, selfH = self:getDims()
  local margin = (self.lineWidth or 0) / 2
  local radius = self.cornerRadius - margin

  local centerL = radius + margin
  local centerR = selfW - radius - margin
  local centerB = radius + margin
  local centerT = selfH - radius - margin

  local points = {}

  local sharpBottomLeft = ((sharpCornersList ~= nil) and (TableUtils.contains(sharpCornersList, RoundedRectangleCommon.CORNERS.BOTTOM_LEFT)))
  if (sharpBottomLeft) then
    table.insert(points, {x = margin, y = margin})
  else
    ShapeUtils.makeArcPoints(-math.pi, -math.pi/2, radius, centerL, centerB, nil, points)
  end

  local sharpBottomRight = ((sharpCornersList ~= nil) and (TableUtils.contains(sharpCornersList, RoundedRectangleCommon.CORNERS.BOTTOM_RIGHT)))
  if (sharpBottomRight) then
    table.insert(points, {x = selfW - margin, y = margin})
  else
    ShapeUtils.makeArcPoints(-math.pi/2, 0, radius, centerR, centerB, nil, points)
  end

  local sharpTopRight = ((sharpCornersList ~= nil) and (TableUtils.contains(sharpCornersList, RoundedRectangleCommon.CORNERS.TOP_RIGHT)))
  if (sharpTopRight) then
    table.insert(points, {x = selfW - margin, y = selfH - margin})
  else
    ShapeUtils.makeArcPoints(0, math.pi/2, radius, centerR, centerT, nil, points)
  end

  local sharpTopLeft = ((sharpCornersList ~= nil) and (TableUtils.contains(sharpCornersList, RoundedRectangleCommon.CORNERS.TOP_LEFT)))
  if (sharpTopLeft) then
    table.insert(points, {x = margin, y = selfH - margin})
  else
    ShapeUtils.makeArcPoints(math.pi/2, math.pi, radius, centerL, centerT, nil, points)
  end

  self._graphicsPrimitiveCoordinateLists = {ShapeUtils.flattenCoordPairsNew(points)}
end

function RoundedRectangleCommon:getCornerRadius()
  return self.cornerRadius
end

--- Set the corner radius.
-- Note: The corner radius cannot be greater than half of the width or height.
function RoundedRectangleCommon:setCornerRadius(radius)
  self.cornerRadius = radius
  self:updateGraphicsPrimitive()
end

function RoundedRectangleCommon:newCornerRadiusTo(duration, ease, radius)
  return AHNonAttributeTo.new(self, duration, self.getCornerRadius, self.setCornerRadius, ease, radius)
end

function RoundedRectangleCommon:newCornerRadiusBy(duration, ease, radius)
  return AHNonAttributeBy.new(self, duration, self.getCornerRadius, self.setCornerRadius, ease, radius)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function RoundedRectangleCommon.makeRoundedRectangle(self, cornerRadius)
  Mixin.mix(self, RoundedRectangleCommon)
  self.cornerRadius = cornerRadius
  return self
end



-- ===========================================================================
class("FilledRoundedRectangle")
-- ===========================================================================

FilledRoundedRectangle.CORNERS = RoundedRectangleCommon.CORNERS

--- Create a new filled rounded rectangle.
-- @param width (number) - The rectangle's width
-- @param height (number) - The rectangle's height
-- @param cornerRadius (number) - The radius of the rounded corner pieces
-- @param color (table) - The color to apply to this object (optional)
-- @param sharpCornersList (table) - List of corners to keep sharp and ignore rounding
function FilledRoundedRectangle.new(width, height, cornerRadius, color, sharpCornersList)
  local self = FilledGraphicsPrimitive.new(width, height, color)
  RoundedRectangleCommon.makeRoundedRectangle(self, cornerRadius)
  Mixin.mix(self, FilledRoundedRectangle)
  self:updateGraphicsPrimitive(sharpCornersList)

  return self
end



-- ===========================================================================
class("HollowRoundedRectangle")
-- ===========================================================================

HollowRoundedRectangle.CORNERS = RoundedRectangleCommon.CORNERS

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

--- Create a new hollow rounded rectangle.
-- @param width (number) - The rectangle's width
-- @param height (number) - The rectangle's height
-- @param cornerRadius (number) - The radius of the rounded corner pieces
-- @param lineWidth (number) - The width of the line to draw the rectangle with,
--                             Regardless of line width, the rectangle will never draw
--                             outside of its bounds (default 1)
-- @param color (table) - The color to apply to this object (optional)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0)
-- @param sharpCornersList (table) - List of corners to keep sharp and ignore rounding
function HollowRoundedRectangle.new(width, height, cornerRadius, color, lineWidth, blurMargin, sharpCornersList)
  local self = HollowGraphicsPrimitive.new(width, height, color, lineWidth, blurMargin, true)
  RoundedRectangleCommon.makeRoundedRectangle(self, cornerRadius)
  Mixin.mix(self, HollowRoundedRectangle)
  self:setShouldApplyAutoMargin(false)
  self:updateGraphicsPrimitive(sharpCornersList)

  return self
end
