--- RoundedRegularPolygon
-- Advanced Hexagon Technology
-- @author Created by Aaron Barrett on September 29, 2014.

require("graphics.ShapeUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local POINTS_PER_CORNER = 8
local CIRCLE_RADIANS = math.pi * 2


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function newRoundedRegularPolygonPoints(sideCount, startAngle, roundingRadius)
  local outerRadius = 0.5
  local angleCoveredByEachArc = CIRCLE_RADIANS / sideCount
  local halfAngleCoveredByEachArc = angleCoveredByEachArc * 0.5
  local radiusToArcCenter = outerRadius - roundingRadius

  local points = {}

  for sideIndex = 0, sideCount-1 do
    local angleToArcCenter = startAngle + sideIndex / sideCount * CIRCLE_RADIANS

    local arcCenterX = outerRadius + math.cos(angleToArcCenter) * radiusToArcCenter
    local arcCenterY = outerRadius + math.sin(angleToArcCenter) * radiusToArcCenter

    local arcStartAngle = angleToArcCenter - halfAngleCoveredByEachArc
    local arcEndAngle = angleToArcCenter + halfAngleCoveredByEachArc

    ShapeUtils.makeArcPoints(arcStartAngle,
                             arcEndAngle,
                             roundingRadius,
                             arcCenterX,
                             arcCenterY,
                             POINTS_PER_CORNER,
                             points)
  end

  return points
end



-- ===========================================================================
class("FilledRoundedRegularPolygon")
-- ===========================================================================

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new filled rounded regular polygon.
-- @param radius (number) - The radius from the center of the polygon to the outer border at each of its vertices.
-- @param sideCount (number) - The number of sides this polygon should have.
-- @param startAngle (number) - The starting angle to draw from in radians.
-- @param roundingRadius (number) - the radius of the arc that will be used to draw the rounding at each of its vertices.
-- @param color (table) - The color to apply to this object (optional)
function FilledRoundedRegularPolygon.new(radius, sideCount, startAngle, roundingRadius, color)
  local self = FilledGraphicsPrimitive.new(radius*2, radius*2, color)
  CircularPrimitive.makeCircularPrimitive(self)
  Mixin.mix(self, FilledRoundedRegularPolygon)

  self._normalizedPointLists = {newRoundedRegularPolygonPoints(sideCount, startAngle, roundingRadius/(radius*2))}
  self:updateGraphicsPrimitive()

  return self
end



-- ===========================================================================
class("HollowRoundedRegularPolygon")
-- ===========================================================================

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new hollow regular polygon.
-- @param radius (number) - The radius from the center of the polygon to the outer border at each of its vertices.
-- @param sideCount (number) - The number of sides this polygon should have.
-- @param startAngle (number) - The starting angle to draw from in radians.
-- @param roundingRadius (number) - the radius of the arc that will be used to draw the rounding at each of its vertices.
-- @param color (table) - The color to apply to this object (optional)
-- @param lineWidth (number) - The width of the line to draw the polygon with. Regardless of line width, the polygon will never draw
--                             outside of its bounds (default 1.0)
-- @param blurMargin (number) - The amount of blur to apply to the edges (default 0.0)
function HollowRoundedRegularPolygon.new(radius, sideCount, startAngle, roundingRadius, color, lineWidth, blurMargin)
  local self = HollowGraphicsPrimitive.new(radius*2, radius*2, color, lineWidth, blurMargin, true)
  CircularPrimitive.makeCircularPrimitive(self)
  Mixin.mix(self, HollowRegularPolygon)

  self._normalizedPointLists = {newRoundedRegularPolygonPoints(sideCount, startAngle, roundingRadius/(radius*2))}
  self:updateGraphicsPrimitive()

  return self
end
