--- Label
--
-- API defininition for single-line labels, and constructor entry point.
-- Use this class for generating all single line labels. The type of font
-- you pass in determines what label implementation will be used.
-- Implementation classes will have their own extended API, but will at the very
-- least have the API specified in this file.
--
-- @author Created by karl on July 23, 2015.

class("Label")

require("text.rendered.RenderedLabel")
require("text.prerendered.PrerenderedLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local IMPLEMENTATIONS_BY_FONT_TYPE = {
  [Font.getFontType()] = RenderedLabel,
  [PrerenderedFont.getFontType()] = PrerenderedLabel
}

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- These denote the API methods expected to be found in all labels.

function Label:getText()
  error("Subclass must define getText")
end

function Label:setText(text)
  error("Subclass must define setText")
end

--- Free any resources taken by this label.
function Label:freeResources()
  -- Do nothing by default
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new label.
-- @param text (string or number) - What to display in the label.
-- @param font (Font) - The font to use.
-- @param width (number, optional) - The width of the label. Will auto-size if unspecified.
-- @param horizontalAlignment (alignment, optional) - What text alignment to use. Default Alignment.LEFT
-- @param shouldBuildGlyphDescriptors (boolean, default false) - If true, build glyph descriptors for this label.
-- @param trackingOffset (number, optional) - Amount of tracking between characters. Defaults to 0.0
function Label.new(text, font, width, horizontalAlignment, shouldBuildGlyphDescriptors, trackingOffset)
  assert(font, "font cannot be nil.")
  if (trackingOffset ~= nil) then
    assert(font:getFontType() == PrerenderedFont.getFontType(), "trackingOffset can only be used with PrerenderedFont")
  end

  text = TextCommon.validateAndConvertText(text)
  horizontalAlignment = horizontalAlignment or Alignment.LEFT

  local implementation = IMPLEMENTATIONS_BY_FONT_TYPE[font:getFontType()]
  return implementation.new(text, font, width, horizontalAlignment, shouldBuildGlyphDescriptors, trackingOffset)
end
