--- SizeConstrainedTextArea
-- @author Created by andrewc on September 23, 2020.

class("SizeConstrainedTextArea")

require("text.dynamic.TextInterpreter")
require("text.dynamic.TextModel")
require("text.dynamic.TextView")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createLabelAttributesWithScaledFontSize(self, fontSize)
  local labelAttributesCopy = TableUtils.deepCopy(self.labelAttributes)
  for _, attributes in pairs(labelAttributesCopy) do
    local scaleFactor = fontSize / self.maxFontSize
    if (attributes["font"] ~= nil) then
      local scaledFontSize = scaleFactor * attributes["font"]:getNominalSize()
      attributes["font"] = attributes["font"]:newFontWithNominalSize(scaledFontSize)
    end
    if (attributes["scaleToSize"] ~= nil) then
      attributes["scaleToSize"] = scaleFactor * attributes["scaleToSize"]
    end
    if (attributes["yOffset"] ~= nil) then
      attributes["yOffset"] = scaleFactor * attributes["yOffset"]
    end
  end
  return labelAttributesCopy
end

local function createTextInterpreter(self, fontSize)
  local baseAttributes = TableUtils.deepCopy(self.baseAttributes)
  baseAttributes["font"] = self.baseAttributes["font"]:newFontWithNominalSize(fontSize)
  local labelAttributesCopy = createLabelAttributesWithScaledFontSize(self, fontSize)
  return TextInterpreter.new(baseAttributes, labelAttributesCopy)
end

local function createTextModel(self, textInterpreter)
  return TextModel.new(textInterpreter:interpretXmlText(self.text))
end

local function setFontSize(self, fontSize)
  self:setLineSpacing(fontSize * self.lineSpacingMultiplier)
  local textInterpreter = createTextInterpreter(self, fontSize)
  local textModel = createTextModel(self, textInterpreter)
  self:setModel(textModel)
end

local function fontSizeIsValid(self, fontSize)
  setFontSize(self, fontSize)
  return ((self:getHeight() <= self.maxHeight) and (fontSize >= self.minimumFontSize))
end

local function binarySearchForFontSize(self, minimumFontSize, maxFontSize)
  if ((maxFontSize - minimumFontSize) < 2) then
    if fontSizeIsValid(self, maxFontSize) then
      return maxFontSize
    else
      return minimumFontSize
    end
  end
  local midPoint = math.floor((minimumFontSize + maxFontSize) / 2)
  if fontSizeIsValid(self, midPoint) then
    return binarySearchForFontSize(self, midPoint, maxFontSize)
  else
    return binarySearchForFontSize(self, minimumFontSize, midPoint)
  end
end

local function resizeFontToFit(self)
  local currentFontSize = self.maxFontSize
  if (self.maxHeight ~= nil) then
    currentFontSize = binarySearchForFontSize(self, self.minimumFontSize, self.maxFontSize)
  end
  setFontSize(self, currentFontSize)
  self.currentBaseFontSize = currentFontSize
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SizeConstrainedTextArea:getCurrentFontSize()
  return self.currentBaseFontSize
end

function SizeConstrainedTextArea:setText(text)
  self.text = text
end

function SizeConstrainedTextArea:getText()
  return self.text
end

function SizeConstrainedTextArea:setWidth(width)
  self.width = width
end

function SizeConstrainedTextArea:setMaxHeight(maxHeight)
  self.maxHeight = maxHeight
end

function SizeConstrainedTextArea:setBaseAttributes(baseAttributes)
  self.baseAttributes = baseAttributes
  self.maxFontSize = baseAttributes["font"]:getNominalSize()
end

function SizeConstrainedTextArea:setLabelAttributes(labelAttributes)
  self.labelAttributes = labelAttributes
end

function SizeConstrainedTextArea:setMinimumFontSize(minimumFontSize)
  self.minimumFontSize = minimumFontSize
end

function SizeConstrainedTextArea:setHorizontalAlignment(alignment)
  self.horizontalAlignment = alignment
end

function SizeConstrainedTextArea:setVerticalAlignment(alignment)
  self.verticalAlignment = alignment
end

function SizeConstrainedTextArea:setLineSpacing(lineSpacing)
  self.lineSpacing = (lineSpacing / self.maxFontSize)
end

function SizeConstrainedTextArea:rerender()
  resizeFontToFit(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

-- @param text (string) - What to display in the text area.
-- @param width (number) - The width of the text area.
-- @param maxHeight (number) - The maximum height of the text area. The font size will scale down to fit.
-- @param baseAttributes (table) - Attributes to apply to every item; must include font.
-- @param attributesForLabels (table, optional) - Attributes to set on an item when a tag is found. Font sizes will scale with the base font size.
-- @param minimumFontSize (number, optional) - The smallest font size before the text will exceed the constraints. Default 1.
-- @param horizontalAlignment (alignment, optional) - What horizontal text alignment to use. Default Alignment.LEFT
-- @param verticalAlignment (alignment, optional) - What vertical text alignment to use. Default Alignment.TOP
-- @param lineSpacing (number, optional) - Spacing between lines. Scales with base font size. Default 0
function SizeConstrainedTextArea.new(text, width, maxHeight, baseAttributes, labelAttributes, minimumFontSize,
  horizontalAlignment, verticalAlignment, lineSpacing)
  lineSpacing = lineSpacing or 0
  minimumFontSize = minimumFontSize or 1
  labelAttributes = labelAttributes or {}

  local self = TextView.new(width, horizontalAlignment, verticalAlignment, lineSpacing, true)
  Mixin.mix(self, SizeConstrainedTextArea)

  self.text = text
  self.width = width
  self.maxHeight = maxHeight
  self.baseAttributes = baseAttributes
  self.labelAttributes = labelAttributes
  self.minimumFontSize = minimumFontSize
  self.horizontalAlignment = horizontalAlignment
  self.verticalAlignment = verticalAlignment

  self.maxFontSize = baseAttributes["font"]:getNominalSize()
  self.lineSpacingMultiplier = (lineSpacing / self.maxFontSize)

  self.currentBaseFontSize = nil
  resizeFontToFit(self)

  return self
end
