--- TextArea
--
-- API defininition for multi-line text areas, and constructor entry point.
-- Use this class for generating all text areas. The type of font you pass in
-- determines what label implementation will be used.
-- Implementation classes will have their own extended API, but will at the very
-- least have the API specified in this file.
--
-- @author Created by karl on July 23, 2015.

class("TextArea")

require("text.rendered.RenderedTextArea")
require("text.prerendered.PrerenderedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local IMPLEMENTATIONS_BY_FONT_TYPE = {
  [Font.getFontType()] = RenderedTextArea,
  [PrerenderedFont.getFontType()] = PrerenderedTextArea
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- These denote the API methods expected to be found in all labels.

function TextArea:getText()
  error("Subclass must define getText")
end

function TextArea:setText(text)
  error("Subclass must define setText")
end

--- Free any resources taken by this text area.
function TextArea:freeResources()
  -- Do nothing by default
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new text area.
-- @param text (string) - What to display in the text area.
-- @param font (Font) - The font to use.
-- @param width (number) - The width of the text area.
-- @param horizontalAlignment (alignment, optional) - What horizontal text alignment to use. Default Alignment.LEFT
-- @param height (number, optional) - The height of the text area. Will auto-size if unspecified.
-- @param verticalAlignment (alignment, optional) - What vertical text alignment to use. Default Alignment.TOP
-- @param lineSpacing (number, optional) - Spacing between lines. Default 1
-- @param wordBreakMode (mode) - Word break mode. Default MOAITextBox.WORD_BREAK_NONE
-- @param trackingOffset (number, optional) - Amount of tracking between characters. Defaults to 0.0
function TextArea.new(text, font, width, horizontalAlignment, height, verticalAlignment, lineSpacing, wordBreakMode, trackingOffset)
  assert(font, "font cannot be nil.")
  assert(width, "width cannot be nil")
  if (trackingOffset ~= nil) then
    assert(font:getFontType() == PrerenderedFont.getFontType(), "trackingOffset can only be used with PrerenderedFont")
  end

  text = TextCommon.validateAndConvertText(text)
  horizontalAlignment = horizontalAlignment or Alignment.LEFT
  verticalAlignment = verticalAlignment or Alignment.TOP
  lineSpacing = lineSpacing or 1.0
  wordBreakMode = wordBreakMode or MOAITextBox.WORD_BREAK_NONE

  local implementation = IMPLEMENTATIONS_BY_FONT_TYPE[font:getFontType()]
  return implementation.new(text, font, width, horizontalAlignment, height, verticalAlignment, lineSpacing, wordBreakMode, trackingOffset)
end
