--- TextAnimationBuilder
-- @author Created by amorant on October 20, 2014.

class("TextAnimationBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_MOVE_DURATION = 0.5
local DEFAULT_SHOW_DURATION = 0.5
local DEFAULT_HIDE_DURATION = 0.5
local DEFAULT_ADJUST_HEIGHT_DURATION = 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------



-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Protected API ----------------------

function TextAnimationBuilder:_createMoveItemAnimation(itemView, itemLayout)
  return AHLocationTo.new(itemView, self.moveDuration, AHEaseSineInOut.new(), itemLayout:getX(), itemLayout:getY())
end

function TextAnimationBuilder:_createAllMoveItemAnimations(oldItemLayouts, newItemLayouts)
  local moveAnimations = AHSpawn.new()

  for itemView, itemLayout in pairs(newItemLayouts) do
    local oldItemLayout = oldItemLayouts[itemView]
    local shouldMoveItem = (oldItemLayout ~= nil) and (not itemLayout:equals(oldItemLayout))

    if shouldMoveItem then
      local itemAnimation = self:_createMoveItemAnimation(itemView, itemLayout)
      moveAnimations:addAction(itemAnimation)
    end
  end

  return moveAnimations
end

function TextAnimationBuilder:_createShowItemAnimation(itemView, itemLayout, textView)
  local originalOpacity = itemView:getOpacity()
  return AHSequence.new(
    function()
      textView:insert(itemView)
      itemView:setOpacity(0)
      itemView:setLoc(itemLayout:getX(), itemLayout:getY())
      itemView:setAnchorPoint(itemLayout:getAnchorX(), itemLayout:getAnchorY())
    end,
    AHOpacityTo.new(itemView, self.showDuration, nil, originalOpacity)
  )
end

function TextAnimationBuilder:_createAllShowItemAnimations(textView, oldItemLayouts, newItemLayouts)
  local showAnimations = AHSpawn.new()

  for itemView, itemLayout in pairs(newItemLayouts) do
    if not oldItemLayouts[itemView] then
      local itemAnimation = self:_createShowItemAnimation(itemView, itemLayout, textView)
      showAnimations:addAction(itemAnimation)
    end
  end

  return showAnimations
end

function TextAnimationBuilder:_createHideItemAnimation(itemView)
  return AHSequence.new(
    AHOpacityTo.new(itemView, self.hideDuration, nil, 0),
    function() itemView:removeSelf() end
  )
end

function TextAnimationBuilder:_createAllHideItemAnimations(oldItemLayouts, newItemLayouts)
  local hideAnimations = AHSpawn.new()

  for itemView, itemLayout in pairs(oldItemLayouts) do
    if not newItemLayouts[itemView] then
      local itemAnimation = self:_createHideItemAnimation(itemView, itemLayout)
      hideAnimations:addAction(itemAnimation)
    end
  end

  return hideAnimations
end

function TextAnimationBuilder:_createAdjustHeightAnimation(textView, newHeight)
  return AHDimYTo.new(textView, self.adjustHeightDuration, AHEaseSineInOut.new(), newHeight)
end

function TextAnimationBuilder:_createAdjustHeightAnimationIfNeeded(textView, newHeight)
  if newHeight ~= textView:getHeight() then
    local adjustHeightAnimation = self:_createAdjustHeightAnimation(textView, newHeight)
    return adjustHeightAnimation
  else
    return AHDoNothing.new()
  end
end

-- Public API -------------------------

function TextAnimationBuilder:createItemsMovedAnimation(textView, oldItemLayouts, newItemLayouts, newHeight)
  local moveAnimations = self:_createAllMoveItemAnimations(oldItemLayouts, newItemLayouts)
  local adjustHeightAnimation = self:_createAdjustHeightAnimationIfNeeded(textView, newHeight)
  local animation = AHSpawn.new(moveAnimations, adjustHeightAnimation)
  return animation
end

function TextAnimationBuilder:createItemsInsertedAnimation(textView, oldItemLayouts, newItemLayouts, newHeight)
  local moveAnimations = self:_createAllMoveItemAnimations(oldItemLayouts, newItemLayouts)
  local showAnimations = self:_createAllShowItemAnimations(textView, oldItemLayouts, newItemLayouts)
  local adjustHeightAnimation = self:_createAdjustHeightAnimationIfNeeded(textView, newHeight)
  local animation = AHSequence.new(AHSpawn.new(moveAnimations, adjustHeightAnimation), showAnimations)
  return animation
end

function TextAnimationBuilder:createItemsRemovedAnimation(textView, oldItemLayouts, newItemLayouts, newHeight)
  local hideAnimations = self:_createAllHideItemAnimations(oldItemLayouts, newItemLayouts)
  local moveAnimations = self:_createAllMoveItemAnimations(oldItemLayouts, newItemLayouts)
  local adjustHeightAnimation = self:_createAdjustHeightAnimationIfNeeded(textView, newHeight)
  local animation = AHSequence.new(hideAnimations, AHSpawn.new(moveAnimations, adjustHeightAnimation))
  return animation
end

function TextAnimationBuilder:createItemsReplacedAnimation(textView, oldItemLayouts, newItemLayouts, newHeight)
  local hideAnimations = self:_createAllHideItemAnimations(oldItemLayouts, newItemLayouts)
  local moveAnimations = self:_createAllMoveItemAnimations(oldItemLayouts, newItemLayouts)
  local showAnimations = self:_createAllShowItemAnimations(textView, oldItemLayouts, newItemLayouts)
  local adjustHeightAnimation = self:_createAdjustHeightAnimationIfNeeded(textView, newHeight)
  local animation = AHSequence.new(hideAnimations, AHSpawn.new(moveAnimations, adjustHeightAnimation, showAnimations))
  return animation
end

function TextAnimationBuilder:setMoveDuration(moveDuration)
  self.moveDuration = moveDuration
end

function TextAnimationBuilder:setShowDuration(showDuration)
  self.showDuration = showDuration
end

function TextAnimationBuilder:setHideDuration(hideDuration)
  self.hideDuration = hideDuration
end

function TextAnimationBuilder:setAdjustHeightDuration(adjustHeightDuration)
  self.adjustHeightDuration = adjustHeightDuration
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TextAnimationBuilder.new()
  local self = {}
  Mixin.mix(self, TextAnimationBuilder)
  self.moveDuration = DEFAULT_MOVE_DURATION
  self.showDuration = DEFAULT_SHOW_DURATION
  self.hideDuration = DEFAULT_HIDE_DURATION
  self.adjustHeightDuration = DEFAULT_ADJUST_HEIGHT_DURATION

  return self
end
