--- TextInterpreter
-- @author Created by amorant on October 13, 2014.

class("TextInterpreter")
require("utils.XmlParser")
require("text.dynamic.TextItem")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PATTERN_EXTRACT_NON_BREAKING_WHITESPACE = "([^ %c]+)"
local PATTERN_IS_BREAKING_WHITESPACE = "^ %c*$"

local TAG_BR = "br"
local TAG_IMG = "img"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateFont(itemAttributes)
  local font = itemAttributes.font
  if font then
    local nominalSize = itemAttributes.size
    local actualSize = itemAttributes.actualSize
    if nominalSize then
      itemAttributes.font = font:newFontWithNominalSize(nominalSize)
    elseif actualSize then
      itemAttributes.font = font:newFontWithActualSize(actualSize)
    end
  end
end

local function getItemAttributes(self, element, inheritedItemAttributes)
  local xmlLabel = element:getLabel()
  local labelAttributes = self.attributesForLabels[xmlLabel] or {}
  local xmlAttributes = element:getAttributes()

  -- NOTE: Order is important here to define attribute priority: xml > label > inherited.
  -- Also, the model does not currently support attribute combining (e.g. bold italics), so the deepest tag wins.
  local itemAttributes = TableUtils.unionSet(xmlAttributes, labelAttributes, inheritedItemAttributes)
  updateFont(itemAttributes)

  return itemAttributes
end

local function parseXmlTree(self, element, fullText, inheritedItemAttributes)
  local inheritedItemAttributes = inheritedItemAttributes or TableUtils.shallowCopy(self.baseAttributes)

  -- Text element: Parse element
  if element:isTextElement() then
    return self:parseXmlTextElement(element, inheritedItemAttributes, fullText)

  -- Root element: Visit children
  elseif element:isRootElement() then
    local items = {}

    for _, child in ipairs(element:getChildren()) do
      local itemsForChild = parseXmlTree(self, child, fullText, inheritedItemAttributes)
      TableUtils.concatenateArrays(items, itemsForChild)
    end

    return items

  -- Tag element: Parse element and visit children
  elseif element:isTagElement() then
    local attributesForElement = getItemAttributes(self, element, inheritedItemAttributes)
    local items = self:parseXmlTagElement(element, attributesForElement)

    for _, child in ipairs(element:getChildren()) do
      local itemsForChild = parseXmlTree(self, child, fullText, attributesForElement)
      TableUtils.concatenateArrays(items, itemsForChild)
    end

    return items
  end
end

local function splitTextIntoWords(text)
  local words = {}
  local startIndex = 1
  local endIndex = nil
  local word = nil

  while startIndex <= text:len() do
    startIndex, endIndex, word = text:find(PATTERN_EXTRACT_NON_BREAKING_WHITESPACE, startIndex)

    if startIndex == nil then
      break
    end

    table.insert(words, { word = word, positionOffset = startIndex - 1})
    startIndex = endIndex + 1
  end

  return words
end

local function findCharacterThatIsNotInsideTag(startingPosition, direction, fullText)
  assert(direction == 1 or direction == -1, "Direction was expected to be +1 or -1, to find next or previous character")

  local characterIsInsideTag = false
  position = startingPosition + direction

  while position > 0 and position <= fullText:len() do
    local character = fullText[position]

    if (character == "<" and direction > 0) or (character == ">" and direction < 0) then
      characterIsInsideTag = true
    end

    if not characterIsInsideTag then
      -- If we have skipped over a br tag, do not consider the words at either side adjacent
      local rangeStartIndex = math.min(startingPosition, position)
      local rangeEndIndex = math.max(startingPosition, position)
      if fullText:sub(rangeStartIndex, rangeEndIndex):find("<%s*" .. TAG_BR .. "%s*/>") then
        return nil
      else
        return character
      end
    end

    if (character == "<" and direction < 0) or (character == ">" and direction > 0) then
      characterIsInsideTag = false
    end

    position = position + direction
  end

  return nil
end

local function findPreviousCharacterThatIsNotInsideTag(position, fullText)
  return findCharacterThatIsNotInsideTag(position, -1, fullText)
end

local function findNextCharacterThatIsNotInsideTag(position, fullText)
  return findCharacterThatIsNotInsideTag(position, 1, fullText)
end

local function getCaseInsensitiveAttributesForLabels(attributesForLabels)
  -- XmlParser returns lower case labels, so we must make sure we match them.
  local caseInsensitiveAttributesForLabels = {}
  for label, attributes in pairs(attributesForLabels) do
    local lowerCaseLabel = label:lower()
    caseInsensitiveAttributesForLabels[lowerCaseLabel] = attributes
  end

  assert(#caseInsensitiveAttributesForLabels == #attributesForLabels, "Some labels clashed when made case-insensitive")

  return caseInsensitiveAttributesForLabels
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Protected API ----------------------
-- To be overriden by subclasses

--- Creates TextItems for an XmlElement containing text.
-- The base implementation splits text into words and generates TextItem objects.
-- @param element (XmlElement) - An XmlElement of text type.
-- @param attributesForElement (table) - A table of TextItem attributes resulting from all the element's parent labels.
-- @return (array) - An array of TextItems to create for this element.
function TextInterpreter:parseXmlTextElement(element, attributesForElement, fullText)
  local items = {}
  local text = element:getText()
  local words = splitTextIntoWords(text)
  local wordsCount = TableUtils.getSize(words)

  for i, wordData in ipairs(words) do
    local word = wordData.word
    local positionOffset = wordData.positionOffset

    local wordIsEmpty = word:find(PATTERN_IS_BREAKING_WHITESPACE) ~= nil
    if not wordIsEmpty then
      local itemAttributes = TableUtils.shallowCopy(attributesForElement)
      local isFirstWord = i == 1
      local isLastWord = i == wordsCount

      -- Handle spacing around tags

      if isFirstWord then
        local previousCharacter = findPreviousCharacterThatIsNotInsideTag(element:getPosition() + positionOffset, fullText)
        local previousCharacterIsWord = previousCharacter ~= nil and previousCharacter:find("%s") == nil

        if previousCharacterIsWord then
          itemAttributes.skipPreviousSpacing = true
          itemAttributes.allowBreakLineBefore = false
        end
      end

      if isLastWord then
        local nextCharacter = findNextCharacterThatIsNotInsideTag(element:getPosition() + positionOffset + word:len() - 1, fullText)
        local nextCharacterIsWord = nextCharacter ~= nil and nextCharacter:find("%s") == nil

        if nextCharacterIsWord then
          itemAttributes.skipNextSpacing = true
          itemAttributes.allowBreakLineAfter = false
        end
      end

      -- Create TextItem for word

      local item = TextItem.newTextItem(word, itemAttributes)
      table.insert(items, item)
    end
  end

  return items
end

--- Creates TextItems for an XmlElement containing a tag.
-- The base implementation recognizes <br/> and <img src=""/> tags, and creates custom item types
-- for XmlElements that have no children if an itemType attribute is available.
-- Note: This method does not recursively process children. That is done in the main algorithm.
-- @param element (XmlElement) - An XmlElement of tag type.
-- @param attributesForElement (table) - A table of TextItem attributes resulting from all the element's parent labels.
-- @return (array) - An array of TextItems to create for this element (often a single one).
function TextInterpreter:parseXmlTagElement(element, attributesForElement)
  local items = {}
  local label = element:getLabel()
  local itemAttributes = TableUtils.shallowCopy(attributesForElement)

  local elementIsEmpty = TableUtils.isEmpty(element:getChildren())
  local itemHasCustomType = itemAttributes.itemType ~= nil
  local isCustomItem = elementIsEmpty and itemHasCustomType

  if isCustomItem then
    itemAttributes = TableUtils.unionSet(itemAttributes, element:getAttributes())
    local item = TextItem.new(itemAttributes)
    return { item }
  elseif label == TAG_BR then
    itemAttributes.mustBreakLine = true
    itemAttributes.skipNextSpacing = true
    itemAttributes.skipPreviousSpacing = true
    local item = TextItem.newEmptyItem(itemAttributes)
    return { item }
  elseif label == TAG_IMG then
    local xmlAttributes = element:getAttributes()
    local imageFilename = xmlAttributes[TextItem.IMAGE_FILENAME_ATTRIBUTE]
    local item = TextItem.newImageItem(imageFilename, itemAttributes)
    return { item }
  end

  return {}
end

-- Public API -------------------------

--- Process an XML Tree producing TextItems.
-- XML data gets processed into items that a TextModel can store and a TextView can display.
-- @param text - (string) Text containing xml tags.
-- @return (array) - TextItems (contents may vary between interpreters as needed for different TextViews)
function TextInterpreter:interpretXmlText(text)
  local xmlTree = XmlParser.parse(text)
  local items = parseXmlTree(self, xmlTree, text)

  return items
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Constructor
-- @param baseAttributes (optional table) - Attributes to set on all items, unless overridden by a tag.
-- @param attributesForLabels (optional table) - Attributes to set on an item when a tag is found.
-- @return (TextInterpreter) - The new instance.
function TextInterpreter.new(baseAttributes, attributesForLabels)
  local self = {}
  Mixin.mix(self, TextInterpreter)

  baseAttributes = baseAttributes or {}
  attributesForLabels = attributesForLabels or {}

  attributesForLabels = getCaseInsensitiveAttributesForLabels(attributesForLabels)

  self.baseAttributes = baseAttributes
  self.attributesForLabels = attributesForLabels

  return self
end
