--- TextItem
-- @author Created by amorant on October 14, 2014.
-- TextItems represent items that can be displayed by a TextView. They contain information
-- extracted from the xml data and are generated by the TextInterpreter.
-- A TextItem can be seen as a simple collection of attributes, like "text" or "font". Al TextItems
-- must have a "itemType", which is what a TextView can use to decide how to handle an item.
-- This class provides some helper methods to access some standard attributes just for the sake of
-- convenience and for documenting such standard attributes. Custom attributes can be obtained
-- through the getAttribute() API.

class("TextItem")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Standard item types ----------------

TextItem.TYPE_TEXT = "TYPE_TEXT"
TextItem.TYPE_IMAGE = "TYPE_IMAGE"
TextItem.TYPE_EMPTY = "TYPE_EMPTY"
TextItem.IMAGE_FILENAME_ATTRIBUTE = "src"

local DEFAULT_ATTRIBUTES = {
  mustBreakLine = false,
  allowBreakLineBefore = true,
  allowBreakLineAfter = true,
  skipPreviousSpacing = false,
  skipNextSpacing = false,
  yOffset = 0
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TextItem:getAttribute(key)
  return self.attributes[key]
end

function TextItem:setAttribute(key, value)
  if key then
    self.attributes[key] = value
  end
end

function TextItem:hasAttribute(key)
  return self.attributes[key] ~= nil
end

function TextItem:equals(textItem)
  if #self.attributes ~= #textItem.attributes then
    return false
  end

  for key, value in pairs(self.attributes) do
    if textItem.attributes[key] ~= value then
      return false
    end
  end

  return true
end

-- Note: This method is exposed only for automated testing purposes, the class is meant to be
-- read-only. If a specific getter is not available, use `getAttribute(key)`.
function TextItem:_test_getAttributes()
  return self.attributes
end

-- Generic attributes -----------------

function TextItem:getItemType()
  return self.attributes.itemType
end

-- Layout attributes ------------------

function TextItem:getMustBreakLine()
  return self.attributes.mustBreakLine
end

function TextItem:setMustBreakLine(allow)
  self.attributes.mustBreakLine = allow
end

function TextItem:getAllowBreakLineBefore()
  return self.attributes.allowBreakLineBefore
end

function TextItem:setAllowBreakLineBefore(allow)
  self.attributes.allowBreakLineBefore = allow
end

function TextItem:getAllowBreakLineAfter()
  return self.attributes.allowBreakLineAfter
end

function TextItem:setAllowBreakLineAfter(allow)
  self.attributes.allowBreakLineAfter = allow
end

function TextItem:getSkipPreviousSpacing()
  return self.attributes.skipPreviousSpacing
end

function TextItem:setSkipPreviousSpacing(skip)
  self.attributes.skipPreviousSpacing = skip
end

function TextItem:getSkipNextSpacing()
  return self.attributes.skipNextSpacing
end

function TextItem:setSkipNextSpacing(skip)
  self.attributes.skipNextSpacing = skip
end

function TextItem:getSpacing()
  return self.attributes.spacing
end

function TextItem:getYOffset()
  return self.attributes.yOffset
end

-- Text attributes --------------------

function TextItem:getText()
  return self.attributes.text
end

function TextItem:getFont()
  return self.attributes.font
end

-- Image attributes -------------------

function TextItem:getImageFilename()
  return self:getAttribute(TextItem.IMAGE_FILENAME_ATTRIBUTE)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TextItem.new(attributes)
  local self = {}
  Mixin.mix(self, TextItem)

  Asserts.assertType(attributes.itemType, "attributes.itemType", "string")

  attributes = TableUtils.unionSet(attributes, DEFAULT_ATTRIBUTES)
  self.attributes = attributes

  return self
end

function TextItem.newTextItem(text, attributes)
  -- TODO This may be temporary, since I'm thinking default font/size might fit in the TextView.
  -- that way a TextItem only has to specify them if it's not using the defaults.
  -- Also  may remove these convenience constructors anyway.
  assert(attributes.font, "No font found in attributes. newTextItem requires a font")

  attributes.itemType = TextItem.TYPE_TEXT
  attributes.text = text

  -- Font and size information used to be separate attributes, but now they are both contained in a
  -- Font object. However, since the TextInterpreter still allow size attributes to update fonts,
  -- we introduce a check here to make all TextItems consistent:
  if attributes.size then
    assert(attributes.size == attributes.font:getNominalSize(), "The size attribute must match the Font's nominal size.")
  else
    attributes.size = attributes.font:getNominalSize()
  end

  local self = TextItem.new(attributes)
  return self
end

function TextItem.newImageItem(imageFilename, attributes)
  attributes = attributes or {}

  attributes.itemType = TextItem.TYPE_IMAGE
  attributes[TextItem.IMAGE_FILENAME_ATTRIBUTE] = imageFilename

  local self = TextItem.new(attributes)
  return self
end

function TextItem.newEmptyItem(attributes)
  attributes = attributes or {}

  attributes.itemType = TextItem.TYPE_EMPTY

  local self = TextItem.new(attributes)
  return self
end
