--- TextModel
-- @author Created by amorant on October 10, 2014.
-- The TextModel stores the text items to display by a TextView. It does not transform or modify
-- the items, as other components in the text system pipeline do, but can support operations on the
-- items such as inserting, removing, moving and replacing. This operations fire events that views
-- can listen to, to update the displayed graphics accordingly.
--
-- NOTE: It is important that the event just contains indices, not actual items or data, so that
-- views must still access the model to retrieve the data. This eliminates the problem of
-- out-of-sync models and views that occurs when the model updates its internal structures in a way
-- and signals something else.

class("TextModel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Events -----------------------------

local function itemsInserted(self, startIndex, itemCount)
  if not self.transactionInProgress then
    for _, listener in ipairs(self.listeners) do
      listener:itemsInserted(startIndex, itemCount)
    end
  end
end

local function itemsRemoved(self, startIndex, itemCount, indicesToRemove)
  if not self.transactionInProgress then
    for _, listener in ipairs(self.listeners) do
      listener:itemsRemoved(startIndex, itemCount, indicesToRemove)
    end
  end
end

local function itemsMoved(self, sourceStartIndex, targetStartIndex, itemCount)
  if not self.transactionInProgress then
    for _, listener in ipairs(self.listeners) do
      listener:itemsMoved(sourceStartIndex, targetStartIndex, itemCount)
    end
  end
end

local function itemsReplaced(self, startIndex, oldItemCount, newItemCount)
  if not self.transactionInProgress then
    for _, listener in ipairs(self.listeners) do
      listener:itemsReplaced(startIndex, oldItemCount, newItemCount)
    end
  end
end

local function itemsReset(self)
  if not self.transactionInProgress then
    for _, listener in ipairs(self.listeners) do
      listener:itemsReset()
    end
  end
end

local function transactionEnded(self)
  assert(not self.transactionInProgress, "Transaction is still in progress")
  for _, listener in ipairs(self.listeners) do
    listener:transactionEnded()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Get items --------------------------

function TextModel:getItems()
  return self.items
end

-- Listener management ----------------

--- Add a listener to this model.
-- The listener must be an object with the following methods:
--  * itemsInserted(startIndex, itemCount)
--  * itemsRemoved(startIndex, itemCount)
--  * itemsMoved(sourceStartIndex, targetStartIndex, itemCount)
--  * itemsReplaced(startIndex, oldItemCount, newItemCount)
--  * itemsReset()
--  * transactionEnded()
-- @param listener (table) - The listener to add.
function TextModel:addListener(listener)
  table.insert(self.listeners, listener)
end

--- Remove a listener from this model.
-- @param listener (table) - The listener to remove.
function TextModel:removeListener(listener)
  TableUtils.removeValueFromArray(self.listeners, listener)
end

-- Operations -------------------------

--- Insert text items into the model. Fires the itemsInserted event.
-- @param items (array) - TextItems to insert
-- @param startIndex (number) - Index where items will be inserted
function TextModel:insertItems(items, startIndex)
  for _, item in rpairs(items) do
    table.insert(self.items, startIndex, item)
  end
  itemsInserted(self, startIndex, #items)
end

--- Remove contiguous text items from the model. Fires the itemsRemoved event.
-- @param startIndex (number) - Index where items will be removed from
-- @param itemCount (number) - How many items to remove
function TextModel:removeItems(startIndex, itemCount)
  for i = 1, itemCount do
    table.remove(self.items, startIndex)
  end
  itemsRemoved(self, startIndex, itemCount)
end

--- Move text items from the model from one position to another. Fires the itemsMoved event.
-- @param sourceStartIndex (number) - Index where items will be removed from
-- @param targetStartIndex (number) - Index where items would be inserted if there was no removal.
--                                    e.g.: Model holds { a, b, c, d }.
--                                          If we want to move `a` after `c`, we will call
--                                          moveItems(1, 1, 4), because the current position after
--                                          `c` is index 4.
-- @param itemCount (number) - How many items to remove
function TextModel:moveItems(sourceStartIndex, targetStartIndex, itemCount)
  local movedItems = {}
  for i = 1, itemCount do
    local item = table.remove(self.items, sourceStartIndex)
    table.insert(movedItems, item)
  end

  local targetStartIndexAfterRemoval = targetStartIndex
  if targetStartIndex > sourceStartIndex then
    targetStartIndexAfterRemoval = targetStartIndex - itemCount
  end

  for _, item in rpairs(movedItems) do
    table.insert(self.items, targetStartIndexAfterRemoval, item)
  end

  itemsMoved(self, sourceStartIndex, targetStartIndex, itemCount)
end

--- Replace text items from the model with new ones. Fires the itemsReplaced event.
-- @param startIndex (number) - Index where items will be replaced from/into
-- @param oldItemCount (number) - How many items to remove
-- @param items (array) - TextItems to insert
function TextModel:replaceItems(startIndex, oldItemCount, newItems)
  for i = 1, oldItemCount do
    table.remove(self.items, startIndex)
  end

  for _, item in rpairs(newItems) do
    table.insert(self.items, startIndex, item)
  end

  itemsReplaced(self, startIndex, oldItemCount, #newItems)
end

--- Reset the model, discarding all current items and setting the new ones.
-- If no items are specified, the current items are kept, but the reset signal is sent anyways.
-- This method also finishes any ongoing transaction.
-- Fires the itemsReset event.
-- @param items (optional array) - New TextItems to insert
function TextModel:resetItems(items)
  if items ~= nil then
    self.items = items
  end

  self.transactionInProgress = false
  itemsReset(self)
end

--- Let the model know that the next changes should be considered as a single one.
-- After this method is called, calls such as moveItems() or replaceItems() will not result in
-- the corresponding events being fired.
-- WARNING: This is just for animation purposes. Changes are notified as a single event, but they
-- are still applied one by one with their respective calls. There is no rollback functionality either.
-- See TextModel:endTransaction().
function TextModel:beginTransaction()
  assert(not self.transactionInProgress, "Transaction already in progress")
  self.transactionInProgress = true
end

--- Let the model know that the transaction started with beginTransaction() has finished.
-- After this method is called, calls such as moveItems() or replaceItems() will result in the
-- corresponding events being fired normally again.
-- Fires the transactionEnded event.
-- See TextModel:beginTransaction().
function TextModel:endTransaction()
  assert(self.transactionInProgress, "Transaction not in progress")
  self.transactionInProgress = false
  transactionEnded(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Constructor
-- @param items (array) - Text items to store.
-- @return (TextModel) - The new instance.
function TextModel.new(items)
  local self = {}
  Mixin.mix(self, TextModel)

  self.items = items
  self.listeners = {}
  self.transactionInProgress = false

  return self
end
