--- TextView
-- @author Created by amorant on October 10, 2014.

class("TextView")
require("text.dynamic.TextAnimationBuilder")
require("text.dynamic.TextItem")
require("text.dynamic.TextLayoutBuilder")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_HORIZONTAL_ALIGNMENT = Alignment.LEFT
local DEFAULT_VERTICAL_ALIGNMENT = Alignment.BOTTOM
local DEFAULT_LINE_SPACING = 0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function resetView(self)
  for _, itemView in ipairs(self.itemViews) do
    itemView:removeSelf()
  end

  if not self.textModel then
    self.itemViews = {}
    self.viewsForItems = {}
    self.itemsForViews = {}
    return
  end

  local items = self.textModel:getItems()
  local itemViews = {}
  for _, item in ipairs(items) do
    local itemView = self:createItemView(item)
    self:insert(itemView)
    table.insert(itemViews, itemView)
  end

  self.itemViews = itemViews
  self.viewsForItems = TableUtils.buildDictionary(items, itemViews)
  self.itemsForViews = TableUtils.buildDictionary(itemViews, items)

  self:updateLayout()
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TextView:getViewsWithAttribute(attribute, value)
  local views = {}
  for index, item in ipairs(self:getModel():getItems()) do
    if item:hasAttribute(attribute) and (item:getAttribute(attribute) == value or value == nil) then
      table.insert(views, self.itemViews[index])
    end
  end

  return views
end

function TextView:getViewsWithoutAttribute(attribute)
  local views = {}
  for index, item in ipairs(self:getModel():getItems()) do
    if not item:hasAttribute(attribute) then
      table.insert(views, self.itemViews[index])
    end
  end

  return views
end

function TextView:getItemViews()
  return self.itemViews
end

-- Protected API ----------------------
-- To be overriden by subclasses

--- Causes a refresh of this view's layout.
-- Exposed so that subclasses can hook into this event.
function TextView:updateLayout()
  local itemViews = self.itemViews
  local itemsForViews = self.itemsForViews
  local layoutBuilder = self.layoutBuilder

  layoutBuilder:updateLayout(itemViews, itemsForViews)
  self:setHeight(layoutBuilder:getTextHeight())

  local itemLayouts = layoutBuilder:getItemLayouts()

  -- Trivially applying the layout, just placing each item view where it belongs.
  -- A more advanced version of the view may use some animation to display the items.
  for itemView, itemLayout in pairs(itemLayouts) do
    itemView:setLoc(itemLayout:getX(), itemLayout:getY())
    itemView:setAnchorPoint(itemLayout:getAnchorX(), itemLayout:getAnchorY())
  end
end

--- Creates an item view for a single item. For instance, a Label or a Sprite.
-- The base implementation expects an item containing text and creates a Label.
-- @param item (table) - A text item (contents depend on TextInterpreter, should to inherit from AbstractTextItem)
-- @return (Group) - An item view.
function TextView:createItemView(item)
  if item:getItemType() == TextItem.TYPE_TEXT then
    return self:createViewForTextItem(item)
  elseif item:getItemType() == TextItem.TYPE_IMAGE then
    return self:createViewForImageItem(item)
  elseif item:getItemType() == TextItem.TYPE_EMPTY then
    return self:createViewForEmptyItem(item)
  else
    error("Unexpected item type: " .. tostring(item:getItemType()))
  end
end

--- Creates an item view for an item of type TextItem.TYPE_TEXT.
-- @param item (table) - A text item of type TextItem.TYPE_TEXT.
-- @return (Group) - An item view.
function TextView:createViewForTextItem(item)
  local text = item:getText()
  local font = item:getFont()
  local color = item:getAttribute("color")

  local label = Label.new(text, font)
  if (color) then
    label:setColor(unpack(color))
    label:inheritOpacity(self)
  else
    label:inheritColor(self)
  end

  local itemWidth = label:getWidth()
  local textWidth = self.layoutBuilder.textWidth
  if item.attributes.fitInLine and itemWidth > textWidth then
    print(string.format("WARNING: Item is too wide for text layout, scaling down. Item width: %s, text width: %s",
                        itemWidth, textWidth))
    local scaling = textWidth / itemWidth
    item.attributes.font = Font.new(item.attributes.font.moaiFont,
                                    item.attributes.font.nominalSize * scaling,
                                    item.attributes.font.nominalMinSize)
    return self:createViewForTextItem(item)
  else
    return label
  end
end

--- Creates an item view for an item of type TextItem.TYPE_IMAGE.
-- @param item (table) - A text item of type TextItem.TYPE_IMAGE.
-- @return (Group) - An item view.
function TextView:createViewForImageItem(item)
  local sprite = Sprite.newWithName(item:getImageFilename())
  sprite:inheritOpacity(self)
  local scaleToSize = item:getAttribute("scaleToSize")
  if (scaleToSize ~= nil) then
    sprite:setEffectiveDims(nil, scaleToSize)
  end
  return sprite
end

function TextView:createViewForEmptyItem(item)
  -- TODO: Ask the font for its line height. Since we don't have that yet, we create a test
  -- character with the same font and size as the surrounding text (same item attributes).
  local testCharacter = self:createViewForTextItem(TextItem.newTextItem("a", {font = item:getFont()}))
  local height = testCharacter:getHeight()
  local group = Group.new(0, height)
  return group
end

-- Public API ----------------------

--- Set the text model for this view.
-- @param horizontalAlignment (table) - Text model to set. Usually an instance of TextModel.
function TextView:setModel(textModel)
  if self.textModel then
    self.textModel:removeListener(self)
  end

  textModel:addListener(self)
  self.textModel = textModel

  resetView(self)
end

function TextView:getModel()
  return self.textModel
end

function TextView:getViewsByRow()
  return self.layoutBuilder:getLines()
end

--- Set the horizontal alignment for this view.
-- @param horizontalAlignment (enum) - Horizontal alignment for the view's layout. Should be one of Alignment.LEFT,
--                                     Alignment.CENTER, Alignment.RIGHT.
function TextView:setHorizontalAlignment(horizontalAlignment)
  local layoutBuilder = self.layoutBuilder
  if horizontalAlignment ~= layoutBuilder:getHorizontalAlignment() then
    layoutBuilder:setHorizontalAlignment(horizontalAlignment)
    self:updateLayout()
  end
end

--- Set the vertical alignment for this view.
-- @param verticalAlignment (enum) - Vertical alignment for the view's layout. Should be one of Alignment.BOTTOM,
--                                   Alignment.CENTER, Alignment.TOP.
function TextView:setVerticalAlignment(verticalAlignment)
  local layoutBuilder = self.layoutBuilder
  if verticalAlignment ~= layoutBuilder:getVerticalAlignment() then
    layoutBuilder:setVerticalAlignment(verticalAlignment)
    self:updateLayout()
  end
end

--- Set extra spacing between lines.
-- @param lineSpacing (number) - Spacing between lines, in pixels. 0 means lines are adjusted to
--                               contents, positive means lines farther apart, negative means
--                               closer together.
function TextView:setLineSpacing(lineSpacing)
  local layoutBuilder = self.layoutBuilder
  if lineSpacing ~= layoutBuilder:getLineSpacing() then
    layoutBuilder:setLineSpacing(lineSpacing)
    self:updateLayout()
  end
end

--- Set a fixed line height
-- @param lineHeight (number) - Total height of each line, to adjust line spacing when you want
--                              each line of text to take up the same height. This value should
--                              larger than the font size.
--                              Note: setting this value will ignore other line spacing.
function TextView:setFixedLineHeight(lineHeight)
  self.layoutBuilder:setFixedLineHeight(lineHeight)
  self:updateLayout()
end

--- Get the animation builder for this view
function TextView:getAnimationBuilder()
  return self.animationBuilder
end

function TextView:getTextWidth()
  return self.layoutBuilder:getTextWidth()
end

function TextView:getTextHeight()
  return self.layoutBuilder:getTextHeight()
end

-- Listeners for TextModel events -----

function TextView:itemsInserted(startIndex, itemCount)
  local items = self.textModel:getItems()
  local endIndex = startIndex + itemCount - 1
  local newItems = TableUtils.subarray(items, startIndex, endIndex)
  local itemViews = self.itemViews
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()

  for _, newItem in rpairs(newItems) do
    local newItemView = self:createItemView(newItem)
    table.insert(itemViews, startIndex, newItemView)
    viewsForItems[newItem] = newItemView
    itemsForViews[newItemView] = newItem
  end

  layoutBuilder:updateLayout(itemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local animation = AHSequence.new(
    animationBuilder:createItemsInsertedAnimation(self, itemLayouts, newItemLayouts, newHeight),
    function() self.animation = nil end
  )

  animation:start()
  self.animation = animation
end

function TextView:itemsRemoved(startIndex, itemCount)
  local itemViews = self.itemViews
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()

  for itemIndex = 1, itemCount do
    local oldItemView = table.remove(itemViews, startIndex)
    local oldItem = itemsForViews[oldItemView]

    viewsForItems[oldItem] = nil
    itemsForViews[oldItemView] = nil
  end

  layoutBuilder:updateLayout(itemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local animation = AHSequence.new(
    animationBuilder:createItemsRemovedAnimation(self, itemLayouts, newItemLayouts, newHeight),
    function() self.animation = nil end
  )

  animation:start()
  self.animation = animation
end

function TextView:itemsMoved(sourceStartIndex, targetStartIndex, itemCount)
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()

  local items = self.textModel:getItems()
  local itemViews = {}

  for _, item in ipairs(items) do
    local itemView = viewsForItems[item]
    table.insert(itemViews, itemView)
  end

  layoutBuilder:updateLayout(itemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local animation = AHSequence.new(
    animationBuilder:createItemsMovedAnimation(self, itemLayouts, newItemLayouts, newHeight),
    function() self.animation = nil end
  )

  animation:start()
  self.animation = animation
  self.itemViews = itemViews
end

function TextView:itemsReplaced(startIndex, oldItemCount, newItemCount)
  local itemViews = self.itemViews
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()

  for i = 1, oldItemCount do
    local oldItemView = table.remove(itemViews, startIndex)
    local oldItem = itemsForViews[oldItemView]

    viewsForItems[oldItem] = nil
    itemsForViews[oldItemView] = nil
  end

  local items = self.textModel:getItems()
  local endIndex = startIndex + newItemCount - 1
  local newItems = TableUtils.subarray(items, startIndex, endIndex)

  for _, newItem in rpairs(newItems) do
    local newItemView = self:createItemView(newItem)
    table.insert(itemViews, startIndex, newItemView)
    viewsForItems[newItem] = newItemView
    itemsForViews[newItemView] = newItem
  end

  layoutBuilder:updateLayout(itemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local animation = AHSequence.new(
    animationBuilder:createItemsReplacedAnimation(self, itemLayouts, newItemLayouts, newHeight),
    function() self.animation = nil end
  )

  animation:start()
  self.animation = animation
end

function TextView:itemsReset()
  resetView(self)
end

function TextView:transactionEnded()
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()
  local newItems = self.textModel:getItems()
  local newItemsSet = TableUtils.createSetFromArray(newItems)

  for oldItem, oldItemView in pairs(viewsForItems) do
    if not newItemsSet[oldItem] then
      viewsForItems[oldItem] = nil
      itemsForViews[oldItemView] = nil
    end
  end

  local updatedItemViews = {}
  for _, newItem in ipairs(newItems) do
    local itemView = viewsForItems[newItem]
    if itemView == nil then
      itemView = self:createItemView(newItem)
      viewsForItems[newItem] = itemView
      itemsForViews[itemView] = newItem
    end
    table.insert(updatedItemViews, itemView)
  end

  self.itemViews = updatedItemViews

  layoutBuilder:updateLayout(updatedItemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local animation = AHSequence.new(
    animationBuilder:createItemsReplacedAnimation(self, itemLayouts, newItemLayouts, newHeight),
    function() self.animation = nil end
  )

  animation:start()
  self.animation = animation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Constructor
-- @param textWdth (number) - Width for the text, which is that of the created group (height is calculated dynamically).
-- @param horizontalAlignment (enum) - Horizontal alignment for the view's layout. Should be one of Alignment.LEFT,
--                                     Alignment.CENTER, Alignment.RIGHT.
-- @param verticalAlignment (enum) - Vertical alignment for the view's layout. Should be one of Alignment.BOTTOM,
--                                   Alignment.CENTER, Alignment.TOP.
-- @param lineSpacing (number) - Spacing between lines, in pixels. 0 means lines are adjusted to contents, positive means lines farther
--                               apart, negative means closer together.
-- @return (TextView) - The new instance.
function TextView.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing, isPointSpaced)
  horizontalAlignment = horizontalAlignment or DEFAULT_HORIZONTAL_ALIGNMENT
  verticalAlignment = verticalAlignment or DEFAULT_VERTICAL_ALIGNMENT
  lineSpacing = lineSpacing or DEFAULT_LINE_SPACING

  local self = Group.new(textWidth, 2)
  Mixin.mix(self, TextView)

  self.textModel = nil
  self.itemViews = {}
  self.viewsForItems = {}
  self.itemsForViews = {}

  local layoutBuilder = TextLayoutBuilder.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing, isPointSpaced)
  self.layoutBuilder = layoutBuilder

  local animationBuilder = TextAnimationBuilder.new()
  self.animationBuilder = animationBuilder

  return self
end
