--- PrerenderedFont
--
-- Pre-renders a set of characters in a specific font and size, allowing
-- lightweight copies to be made of the individual glyphs.
--
-- @author Created by karl on July 21, 2015.

require("ReferenceDevice")
require("text.TextCommon")
require("text.FontScaling")


-- ============================================================================
local PrerenderedGlyphBuilder = privateClass()
-- ============================================================================

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PrerenderedGlyphBuilder:applyGlyphProperties(glyph)
  glyph.texture = self.texture
  glyph.xAdvance = self.xAdvance
  glyph.xOffset = self.xOffset
  glyph:setDeck(self.deck)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PrerenderedGlyphBuilder.new(texture, descriptor)
  local self = {}
  Mixin.mix(self, PrerenderedGlyphBuilder)

  local renderedHeight = texture.renderedHeight
  local xMin, xMax = descriptor.xMin, descriptor.xMax
  local width = (xMax - xMin) / Host.contentScale
  local height = renderedHeight / Host.contentScale
  local textureWidth, textureHeight = texture:getSize()

  self.texture = texture
  self.xAdvance = descriptor.xAdvance / Host.contentScale
  self.xOffset = descriptor.xOffset / Host.contentScale
  self.xMin = xMin / Host.contentScale
  self.xMax = xMax / Host.contentScale

  self.leftMargin = self.xOffset / Host.contentScale
  self.rightMargin = (self.xAdvance - self.xOffset - (self.xMax - self.xMin)) / Host.contentScale

  self.deck = MOAIGfxQuad2D.new()
  self.deck:setTexture(self.texture)
  self.deck:setRect(0, 0, width, height)
  self.deck:setUVRect(xMin / textureWidth,
                      renderedHeight / textureHeight,
                      xMax / textureWidth,
                      0)

  return self
end



-- ============================================================================
class("PrerenderedFont")
-- ============================================================================

require("text.rendered.Font")
require("text.rendered.RenderedLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- TODO: This MAX_TEXTURE_WIDTH at 1024 is producing too large of a texture after
-- MOAI renders it. 1024 => 8192 is twice as large as the max allowable texture size
-- when tested on ipad air. For now, cut in half to 512. Later, investigate where the extra 2x
-- multiplier is coming from (NOTE: a scale factor 4x multiplier is expected and accounted for)
local MAX_TEXTURE_WIDTH = 512

local CACHED_FONTS = {}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeCacheKey(moaiFont, fontSize, characters)
  return tostring(moaiFont) .. ":" .. tostring(fontSize) .. ":" .. characters
end

local function getCachedFont(moaiFont, fontSize, characters)
  return CACHED_FONTS[makeCacheKey(moaiFont, fontSize, characters)]
end

local function cacheFont(font)
  CACHED_FONTS[makeCacheKey(font.moaiFont, font.nominalSize, font.characters)] = font
end

local function applyGlyphPropertiesForCharacter(self, character, glyph)
  local builder = self.glyphBuildersByCharacter[character]
  builder:applyGlyphProperties(glyph)
end

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function PrerenderedFont:getFontType()
  return "Prerendered Font"
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Assign a glyph's texture and properties to prop that has already been instantiated
function PrerenderedFont:makeGlyph(character, existingProp)
  applyGlyphPropertiesForCharacter(self, character, existingProp)
end

--- Generate a glyph from the specified character.
function PrerenderedFont:newGlyph(character)
  local glyph = MOAIProp2D.new()
  applyGlyphPropertiesForCharacter(self, character, glyph)
  return glyph
end

function PrerenderedFont:isVariableSized()
  return false
end

function PrerenderedFont:getMOAIFont()
  return self.moaiFont
end

--- Nominal size specified on creation.
function PrerenderedFont:getNominalSize()
  return self.nominalSize
end

function PrerenderedFont:getNominalMinSize()
  return self.nominalSize
end

--- Actual size after scaling relative to ReferenceDevice
function PrerenderedFont:getActualSize()
  return self.actualSize
end

function PrerenderedFont:getActualMinSize()
  return self.actualSize
end

function PrerenderedFont:getLineHeight()
  return self.lineHeight
end

function PrerenderedFont:newFontWithNominalSize(size)
  if size == self.nominalSize then
    return self
  end
  return PrerenderedFont.new(self.moaiFont, size, self.characters)
end

function PrerenderedFont:newFontWithActualSize(actualSize)
  if actualSize == self.actualSize then
    return self
  end
  return PrerenderedFont.newWithActualSize(self.moaiFont, actualSize, self.characters)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function expandCharacters(characters)
  local newCharacters = ""
  for _, character in ipairs(TableUtils.createArrayFromString(characters)) do
    newCharacters = newCharacters .. character .. " "
  end

  -- Add an extra character to the end of the string to avoid memory corruption bug in `MOAIFreeTypeFont::_dimensionsWithMaxWidth()`
  newCharacters = newCharacters .. "\n"

  return newCharacters
end

local function getLinesToRender(moaiFont, actualSize, expandedCharacters)
  local linesToRender = {}

  local shouldBuildGlyphDescriptors = true
  local _, _, glyphDescriptorLines = moaiFont:dimensionsWithMaxWidth(expandedCharacters, actualSize, MAX_TEXTURE_WIDTH, nil, shouldBuildGlyphDescriptors)

  for _, glyphDescriptorLine in ipairs(glyphDescriptorLines) do
    local lineToRender = glyphDescriptorLine.renderedCharacters

    if ((lineToRender ~= "") and (lineToRender ~= nil)) then
      table.insert(linesToRender, lineToRender)
    end
  end

  return linesToRender
end

local function newPrerenderedFont(moaiFont, nominalSize, actualSize, characters)
  if (characters ~= nil) then
    characters = StringUtils.getUniqueCharacterString(characters)
  end

  characters = characters or TextCommon.getDefaultCharacters()

  local self = getCachedFont(moaiFont, nominalSize, characters)
  if not self then
    self = {}
    Mixin.mix(self, PrerenderedFont)

    local expandedCharacters = expandCharacters(characters)

    local linesToRender = getLinesToRender(moaiFont, actualSize, expandedCharacters)

    self.glyphBuildersByCharacter = {}

    local font = Font.new(moaiFont, nominalSize)

    local firstLineLabel = nil
    for lineIndex, line in ipairs(linesToRender) do
      local shouldBuildGlyphDescriptors = true
      local label = RenderedLabel.new(line, font, nil, nil, shouldBuildGlyphDescriptors)

      local texture = label.texture
      for descriptorIndex, descriptor in ipairs(label:getGlyphDescriptors()) do
        local characterIsSpacer = MathUtils.isEven(descriptorIndex)
        if (not characterIsSpacer) then
          local glyphBuilder = PrerenderedGlyphBuilder.new(texture, descriptor)
          local glyphCharacter = descriptor.character
          self.glyphBuildersByCharacter[glyphCharacter] = glyphBuilder
        end
      end

      if (lineIndex == 1) then
        firstLineLabel = label
      end

    end

    self.moaiFont = moaiFont
    self.nominalSize = nominalSize
    self.actualSize = actualSize
    self.characters = characters

    -- A font's per-line height is shorter than the height of a single line label.
    local approximatePerLineHeightScale = 0.965
    self.lineHeight = firstLineLabel:getHeight() * approximatePerLineHeightScale

    cacheFont(self)
  end

  return self
end

-- Constructor ------------------------

--- Create a new prerendered font. Size will be scaled relative to the reference device to fit the current device.
-- @param moaiFont (MOAIFreeTypeFont) - Font to use.
-- @param nominalSize (number) - Point size to generate (actual size will be scaled using ReferenceDevice).
-- @param characters (string) - String containing all characters to generate glyphs for.
function PrerenderedFont.new(moaiFont, nominalSize, characters)
  local actualSize = FontScaling.getScaledSize(nominalSize)
  return newPrerenderedFont(moaiFont, nominalSize, actualSize, characters)
end

--- Create a new prerendered font with an unscaled size. No scaling to fit the device will be performed.
-- @param moaiFont (MOAIFreeTypeFont) - Font to use.
-- @param nominalSize (number) - Point size to generate (actual size will be scaled using ReferenceDevice).
-- @param characters (string) - String containing all characters to generate glyphs for.
function PrerenderedFont.newWithActualSize(moaiFont, actualSize, characters)
  local nominalSize = FontScaling.getInverseScaledSize(actualSize)
  return newPrerenderedFont(moaiFont, nominalSize, actualSize, characters)
end
