--- PrerenderedGlyphLayout
-- @author Created by karl on July 28, 2015.

class("PrerenderedGlyphLayout")

require("text.prerendered.PrerenderedFont")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LINEFEED = "\n"
local FORCED_LINEFEED = "FORCED_LINEFEED"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addRow(self)
  self.currentRow = {}
  table.insert(self.rows, self.currentRow)
  self.penX = 0
  self.penY = self.penY - self.skipY
end

local function removeRow(self)
  table.remove(self.rows)
  self.currentRow = self.rows[#self.rows]
  if (#self.currentRow > 0) then
    local glyph = self.currentRow[#self.currentRow]
    if (#self.currentRow > 1) then
      local x = glyph:getDims()
      self.penX = x - glyph.xOffset + glyph.xAdvance
    else
      self.penX = glyph.xAdvance
    end
  else
    self.penX = 0
  end
  self.penY = self.penY - self.skipY
end

local function removeForcedLinefeeds(self)
  while self.characters[#self.characters] == FORCED_LINEFEED do
    table.remove(self.characters)
    removeRow(self)
  end
end

local function repositionRow(row, width, horizontalAlignment, yOffset, trackingOffset)
  if (#row > 0) then
    local glyph = row[#row]
    local rowEndX = glyph:getLoc() + glyph:getDims()
    local xOffset = 0
    if (horizontalAlignment == Alignment.CENTER) then
      xOffset = (width - rowEndX + trackingOffset) / 2
    elseif (horizontalAlignment == Alignment.RIGHT) then
      xOffset = width - rowEndX + trackingOffset
    elseif (horizontalAlignment ~= Alignment.LEFT) then
      error(tostring(horizontalAlignment) .. ": Unknown horizontal alignment")
    end

    for _, glyph in ipairs(row) do
      local x, y = glyph:getLoc()
      glyph:setLoc(x + xOffset, y + yOffset)
    end
  end
end

local function activateGlyph(self, character)
  local inactiveGlyphIsAvailable = (#(self.inactiveGlyphs) > 0)
  if (inactiveGlyphIsAvailable) then
    local recycledGlyph = table.remove(self.inactiveGlyphs)
    self.font:makeGlyph(character, recycledGlyph)
    recycledGlyph:setVisible(true)
    return recycledGlyph
  else
    local freshGlyph = self.font:newGlyph(character)
    self.group:insert(freshGlyph)
    return freshGlyph
  end
end

local function recycleGlyph(self, glyph)
  glyph:setVisible(false)
  table.insert(self.inactiveGlyphs, glyph)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PrerenderedGlyphLayout:preAddGlyphs(glyphsCount)
  for glyphIndex = 1, glyphsCount do
    local emptyGlyph = MOAIProp2D.new()
    emptyGlyph:setVisible(false)
    self.group:insert(emptyGlyph)
    table.insert(self.inactiveGlyphs, emptyGlyph)
  end
end

function PrerenderedGlyphLayout:addCharacter(character, ignoreTrackingOffset)
  if ((character == LINEFEED) or (character == FORCED_LINEFEED)) then
    addRow(self)
  else
    local glyph = activateGlyph(self, character)
    local offset = glyph.xOffset
    local trackingOffset = self.trackingOffset
    if (ignoreTrackingOffset) then
      trackingOffset = 0.0
    end
    local advance = glyph.xAdvance + trackingOffset
    if #self.currentRow == 0 then
      advance = advance - offset
      offset = 0
    end

    table.insert(self.currentRow, glyph)
    glyph:setLoc(self.penX + offset, self.penY)
    self.penX = self.penX + advance
  end
  table.insert(self.characters, character)
end

function PrerenderedGlyphLayout:addLinefeed()
  self:addCharacter(FORCED_LINEFEED)
end

function PrerenderedGlyphLayout:removeLastCharacter()
  removeForcedLinefeeds(self)
  if #self.characters > 0 then
    local character = table.remove(self.characters)
    if (character == LINEFEED) then
      removeRow(self)
    else
      local glyph = table.remove(self.currentRow)
      recycleGlyph(self, glyph)
      self.penX = self.penX - glyph.xAdvance
      if (self.penX < 0) then
        self.penX = 0
      end
    end
    removeForcedLinefeeds(self)
  end
end

function PrerenderedGlyphLayout:getRows()
  return self.rows
end

function PrerenderedGlyphLayout:getTextHeight()
  return -self.penY
end

function PrerenderedGlyphLayout:getCurrentRowWidth()
  return self.penX
end

function PrerenderedGlyphLayout:getTrackingOffset()
  return self.trackingOffset
end

function PrerenderedGlyphLayout:setTrackingOffset(offset)
  self.trackingOffset = offset
end

function PrerenderedGlyphLayout:clear()
  if self.rows then
    for _, row in ipairs(self.rows) do
      for _, glyph in ipairs(row) do
        recycleGlyph(self, glyph)
      end
    end
  end

  self.characters = {}
  self.currentRow = {}
  self.rows = {self.currentRow}
  self.penX = 0
  self.penY = -self.font:getLineHeight()
end

function PrerenderedGlyphLayout:reposition(width, height, horizontalAlignment, verticalAlignment)
  local textHeight = self:getTextHeight()
  local trackingOffset = self.trackingOffset
  local yOffset = nil
  if (verticalAlignment == Alignment.TOP) then
    yOffset = height
  elseif (verticalAlignment == Alignment.CENTER) then
    yOffset = textHeight + (height - textHeight) / 2
  elseif (verticalAlignment == Alignment.BOTTOM) then
    yOffset = textHeight
  else
    error(tostring(verticalAlignment) .. ": Unknown vertical alignment")
  end

  for rowIndex, row in ipairs(self.rows) do
    if (rowIndex == #self.rows) then
      trackingOffset = 0.0
    end
    repositionRow(row, width, horizontalAlignment, yOffset, trackingOffset)
  end
end

function PrerenderedGlyphLayout:getTrackingOffset()
  return self.trackingOffset
end

function PrerenderedGlyphLayout:setTrackingOffset(trackingOffset)
  self.trackingOffset = trackingOffset
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PrerenderedGlyphLayout.new(group, font, lineSpacing, trackingOffset)
  local self = {}
  Mixin.mix(self, PrerenderedGlyphLayout)

  self.group = group
  self.font = font
  self.skipY = font:getLineHeight() + lineSpacing
  self.inactiveGlyphs = {}
  self.trackingOffset = trackingOffset or 0.0
  self:clear()

  return self
end
