--- PrerenderedLabel
--
-- A prerendered label uses glyphs from a prerendered font to display text.
-- Each glyph has its own prop, and shares a common texture provided by the font.
--
-- @author Created by karl on July 22, 2015.

class("PrerenderedLabel")

require("text.TextCommon")
require("text.prerendered.PrerenderedFont")
require("text.prerendered.PrerenderedGlyphLayout")

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PrerenderedLabel:preAddGlyphs(glyphsCount)
  self.layout:preAddGlyphs(glyphsCount)
end

function PrerenderedLabel:setText(text)
  text = TextCommon.validateAndConvertText(text)
  self.layout:clear()
  local characters = TableUtils.createArrayFromString(text)
  for characterIndex, character in ipairs(characters) do
    local ignoreTrackingOffset = (characterIndex == #characters)
    self.layout:addCharacter(character, ignoreTrackingOffset)
  end
  if not self.isFixedWidth then
     local width = math.max(self.layout:getCurrentRowWidth(), TextCommon.EMPTY_WIDTH)
    self:setDims(width, self.layout:getTextHeight())
  end
  self.layout:reposition(self:getWidth(), self:getHeight(), self.horizontalAlignment, Alignment.TOP)
  self:linkColorAttrToChildren()
  self.text = text
end

function PrerenderedLabel:getText()
  return self.text
end

function PrerenderedLabel:freeResources()
  -- No resources to free.
end

function PrerenderedLabel:getTrackingOffset()
  return self.layout:getTrackingOffset()
end

function PrerenderedLabel:setTrackingOffset(trackingOffset)
  self.layout:setTrackingOffset(trackingOffset)
  self:setText(self.text)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new label.
-- @param text (string or number) - What to display in the label.
-- @param font (Font) - The font to use.
-- @param width (number, optional) - The width of the label. Will auto-size if unspecified.
-- @param horizontalAlignment (alignment, optional) - What text alignment to use.
-- @param shouldBuildGlyphDescriptors (boolean, IGNORED) - Feel free to implement if you actually need this.
-- @param trackingOffset (number, optional) - Amount of tracking between characters. Defaults to 0.0
function PrerenderedLabel.new(text, font, width, horizontalAlignment, shouldBuildGlyphDescriptors, trackingOffset)
  local self = Group.new(width or TextCommon.EMPTY_WIDTH, font:getLineHeight())
  Mixin.mix(self, PrerenderedLabel)

  self.isFixedWidth = width ~= nil
  self.horizontalAlignment = horizontalAlignment
  local lineSpacing = 0
  self.layout = PrerenderedGlyphLayout.new(self, font, lineSpacing, trackingOffset)
  self:setText(text)

  return self
end
