--- PrerenderedTextArea
--
-- A prerendered text area uses glyphs from a prerendered font to display text.
-- Each glyph has its own prop, and shares a common texture provided by the font.
--
-- Note: The current implementation does not support MOAITextBox.WORD_BREAK_HYPHEN.
--
-- @author Created by karl on July 23, 2015.

require("text.TextCommon")
require("text.prerendered.PrerenderedFont")
require("text.prerendered.PrerenderedGlyphLayout")


class("PrerenderedTextArea")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function characterWrap(self, characters, firstIndex, lastIndex)
  local width = self:getWidth()
  for i = firstIndex, lastIndex do
    local character = characters[i]
    self.layout:addCharacter(character)
    if self.layout:getCurrentRowWidth() > width then
      self.layout:removeLastCharacter()
      self.layout:addLinefeed()
      if character ~= " " then
        self.layout:addCharacter(character)
      end
    end
  end
end

local function wordWrap(self, characters, index)
  local width = self:getWidth()
  local currentIndex = index
  while self.layout:getCurrentRowWidth() > 0 and characters[currentIndex] and characters[currentIndex] ~= " " do
    self.layout:removeLastCharacter()
    currentIndex = currentIndex - 1
  end
  if currentIndex < 1 then
    currentIndex = 1
  end
  if characters[currentIndex] == " " then
    currentIndex = currentIndex + 1
    self.layout:addLinefeed()
  end
  characterWrap(self, characters, currentIndex, index)
end

local function layoutTextWithCharacterWrap(self, text)
  local characters = TableUtils.createArrayFromString(text)
  characterWrap(self, characters, 1, #characters)
end

local function layoutTextWithWordWrap(self, text)
  local characters = TableUtils.createArrayFromString(text)
  local width = self:getWidth()

  for index, character in ipairs(characters) do
    local ignoreTrackingOffset = (index == #characters)
    self.layout:addCharacter(character, ignoreTrackingOffset)
    if self.layout:getCurrentRowWidth() > width then
      wordWrap(self, characters, index)
    end
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PrerenderedTextArea:getText()
  return self.text
end

function PrerenderedTextArea:setText(text)
  text = TextCommon.validateAndConvertText(text)
  self.layout:clear()
  if self.wordBreakMode == MOAITextBox.WORD_BREAK_NONE then
    layoutTextWithWordWrap(self, text)
  elseif self.wordBreakMode == MOAITextBox.WORD_BREAK_CHAR then
    layoutTextWithCharacterWrap(self, text)
  else
    assert(false, "Unsupported wrap mode")
  end
  if not self.isFixedHeight then
    self:setHeight(self.layout:getTextHeight())
  end
  self.layout:reposition(self:getWidth(), self:getHeight(), self.horizontalAlignment, self.verticalAlignment)
  self:linkColorAttrToChildren()
  self.text = text
end

function PrerenderedTextArea:freeResources()
  -- No resources to free.
end

function PrerenderedTextArea:getTrackingOffset()
  return self.layout:getTrackingOffset()
end

function PrerenderedTextArea:setTrackingOffset(trackingOffset)
  self.layout:setTrackingOffset(trackingOffset)
  self:setText(self.text)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new text area.
-- @param text (string) - What to display in the text area.
-- @param font (Font) - The font to use.
-- @param width (number) - The width of the text area.
-- @param horizontalAlignment (alignment, optional) - What horizontal text alignment to use. Default Alignment.LEFT
-- @param height (number, optional) - The height of the text area. Will auto-size if unspecified.
-- @param verticalAlignment (alignment, optional) - What vertical text alignment to use. Default Alignment.TOP
-- @param lineSpacing (number, optional) - Spacing between lines. Default 1
-- @param wordBreakMode (mode) - Word break mode. Default MOAITextBox.WORD_BREAK_NONE
-- @param trackingOffset (number, optional) - Amount of tracking between characters. Defaults to 0.0
function PrerenderedTextArea.new(text, font, width, horizontalAlignment, height, verticalAlignment, lineSpacing, wordBreakMode, trackingOffset)
  local self = Group.new(width, height or font:getLineHeight())
  Mixin.mix(self, PrerenderedTextArea)

  self.isFixedHeight = height ~= nil
  self.horizontalAlignment = horizontalAlignment
  self.verticalAlignment = verticalAlignment
  self.wordBreakMode = wordBreakMode
  self.layout = PrerenderedGlyphLayout.new(self, font, lineSpacing, trackingOffset)
  self:setText(text)

  return self
end
