--- Font
--
-- Font object for the rendered label and text areas.
-- If minSize is specified, the font is "variable", and hints that the text
-- size can be changed to fit inside a bounding box.
--
-- @author Created by karl on July 17, 2015.

class("Font")

require("text.FontScaling")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function Font:getFontType()
  return "Rendered Font"
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Font:isVariableSized()
  return self.nominalSize ~= self.nominalMinSize
end

function Font:getMOAIFont()
  return self.moaiFont
end

--- Nominal size specified on creation.
function Font:getNominalSize()
  return self.nominalSize
end

function Font:getNominalMinSize()
  return self.nominalMinSize
end

--- Actual size after scaling relative to ReferenceDevice
function Font:getActualSize()
  return self.actualSize
end

function Font:getActualMinSize()
  return self.actualMinSize
end

function Font:newFontWithNominalSize(nominalSize)
  if nominalSize == self.nominalSize then
    return self
  end
  return Font.new(self.moaiFont, nominalSize, self.nominalMinSize)
end

function Font:newFontWithActualSize(actualSize)
  if actualSize == self.actualSize then
    return self
  end
  return Font.newWithActualSize(self.moaiFont, actualSize, self.actualMinSize)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newFont(moaiFont, nominalSize, nominalMinSize, actualSize, actualMinSize)
  local self = {}
  Mixin.mix(self, Font)

  self.moaiFont = moaiFont
  self.nominalSize = nominalSize
  self.nominalMinSize = nominalMinSize
  self.actualSize = actualSize
  self.actualMinSize = actualMinSize

  return self
end

-- Constructor ------------------------

--- Create a new font. Sizes will be scaled relative to the reference device to fit the current device.
-- @param moaiFont (MOAIFont) - The font to use.
-- @param nominalSize (number) - The nominal font size (or max size) to display.
-- @param nominalMinSize (number, optional) - If specified, the nominal min size to display.
function Font.new(moaiFont, nominalSize, nominalMinSize)
  assert(nominalSize ~= nil, "Font constructor must be provided a size argument")

  nominalMinSize = nominalMinSize or nominalSize
  local actualSize = FontScaling.getScaledSize(nominalSize)
  local actualMinSize = FontScaling.getScaledSize(nominalMinSize)
  return newFont(moaiFont, nominalSize, nominalMinSize, actualSize, actualMinSize)
end

--- Create a new font with an unscaled size. No scaling to fit the device will be performed.
-- @param moaiFont (MOAIFont) - The font to use.
-- @param actualSize (number) - The actual font size (or max size) to display.
-- @param actualMinSize (number, optional) - If specified, the actual min size to display.
function Font.newWithActualSize(moaiFont, actualSize, actualMinSize)
  assert(actualSize ~= nil, "Font constructor must be provided a size argument")

  actualMinSize = actualMinSize or actualSize
  local nominalSize = FontScaling.getInverseScaledSize(actualSize)
  local nominalMinSize = FontScaling.getInverseScaledSize(actualMinSize)
  return newFont(moaiFont, nominalSize, nominalMinSize, actualSize, actualMinSize)
end
