--- RenderedCommon
--
-- Common code to rendered label and text area.
--
-- @author Created by karl on July 23, 2015.

class("RenderedCommon")

require("text.rendered.Font")
require("text.TextCommon")

require("MetalPrebinder")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function hasTextChanged(self, text)
  local latestText = self.latestRender.text or ""
  return latestText ~= text
end

local function renderTexture(self, text)
  if text == "" then
    self.latestRender = {
      text = text
    }
  else
    local texture, renderedWidth, renderedHeight, glyphDescriptors = self:doRender(text)
    texture.renderedWidth = renderedWidth
    texture.renderedHeight = renderedHeight
    texture:setFilter(MOAITexture.GL_LINEAR, MOAITexture.GL_LINEAR)

    self.latestRender = {
      text = text,
      texture = texture,
      glyphDescriptors = glyphDescriptors
    }

    MetalPrebinder.prebindTexture(texture)
  end
end

local function showNextTexture(self)
  if self.texture ~= self.latestRender.texture then
    local texture = self.latestRender.texture
    local deck = MOAIGfxQuad2D.new()
    local textWidth, textHeight = TextCommon.EMPTY_WIDTH, self.font:getActualSize()

    if self.texture then
      self.texture:release()
    end

    if texture then
      local renderedWidth, renderedHeight = texture.renderedWidth, texture.renderedHeight
      local textureWidth, textureHeight = texture:getSize()
      local uvWidth = renderedWidth / textureWidth
      local uvHeight = renderedHeight / textureHeight
      textWidth = renderedWidth / Host.contentScale
      textHeight = renderedHeight / Host.contentScale

      local offset = 0
      if self.isFixedWidth then
        if self.horizontalAlignment == Alignment.CENTER then
          offset = (self:getWidth() - textWidth) / 2
        elseif self.horizontalAlignment == Alignment.RIGHT then
          offset = self:getWidth() - textWidth
        end
      end

      deck:setTexture(texture)
      deck:setRect(offset, 0, textWidth + offset, textHeight)
      deck:setUVRect(0, uvHeight, uvWidth, 0)
    end

    local selfWidth, selfHeight = textWidth, textHeight
    if self.isFixedWidth then
      selfWidth = self:getWidth()
    end

    self:setDeck(deck)
    self:setDims(selfWidth, selfHeight)
    self.text = self.latestRender.text
    self.texture = texture
    self.glyphDescriptors = self.latestRender.glyphDescriptors
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Protected API --------------------------------------------------------------

function RenderedCommon:_updateMargins()
  error("Must be defined by subclasses")
end


-- Public API -----------------------------------------------------------------

function RenderedCommon:prerenderText(text)
  text = TextCommon.validateAndConvertText(text)

  if hasTextChanged(self, text) then
    self.textDidFit = false
    if self.prerenderCoroutine then
      self.prerenderCoroutine:stop()
      self.prerenderCoroutine = nil
    end

    if text ~= "" and self.requiresFontSizeCalculation then
      self.prerenderCoroutine = runInCoroutine(function()
        self.renderEngine:resetProcess()
        local renderedFontSize = nil
        while true do
          renderedFontSize = self.renderEngine:processOptimalSize(text)
          if renderedFontSize then
            break
          end
          coroutine.yield()
        end
        local contentScaledFontSize = renderedFontSize
        if renderedFontSize <= 0 then
          contentScaledFontSize = self.font:getActualMinSize() * Host.contentScale
        end

        self.renderEngine:setFontSize(contentScaledFontSize)
        renderTexture(self, text)
        self.textDidFit = renderedFontSize > 0
        self.prerenderCoroutine = nil
      end)
    else
      renderTexture(self, text)
      self.textDidFit = true
    end
  end

  return self.prerenderCoroutine
end

function RenderedCommon:getText()
  return self.text
end

function RenderedCommon:setText(text)
  wait(self:prerenderText(text))
  showNextTexture(self)

  self:_updateMargins()
end

function RenderedCommon:freeResources()
  if self.prerenderCoroutine then
    self.prerenderCoroutine:stop()
    self.prerenderCoroutine = nil
  end

  if self.latestRender.texture then
    self.latestRender.texture:release()
  end
  self.latestRender = {}

  if self.texture then
    self.texture:release()
    self.texture = nil
    self:setDeck(MOAIGfxQuad2D.new())
    self:setDimensionsEmptySize()
  end
end

--- Check if the last text fitting operation was successful.
-- Only valid if using a variable sized font and a width.
function RenderedCommon:didTextFit()
  return self.textDidFit
end

function RenderedCommon:getGlyphDescriptors()
  return self.glyphDescriptors
end

function RenderedCommon:getTexture()
  return self.texture
end

--- A label clone shares the original label's texture to save memory.
-- The clone will have none of the methods of the original label. It's just a group.
-- WARNING: Do not release the texture from the original label until
--          all clones have been removed!
function RenderedCommon:clone()
  local texture = self.texture
  local renderedWidth = texture.renderedWidth
  local renderedHeight = texture.renderedHeight
  local textureWidth, textureHeight = texture:getSize()
  local cloneWidth, cloneHeight = self:getDims()
  local uvWidth = renderedWidth / textureWidth
  local uvHeight = renderedHeight / textureHeight

  local labelClone = Group.new(cloneWidth, cloneHeight)
  Mixin.mix(labelClone, Label)

  local deck = MOAIGfxQuad2D.new()
  deck:setTexture(texture)
  deck:setRect(0.0, 0.0, cloneWidth, cloneHeight)
  deck:setUVRect(0.0, uvHeight, uvWidth, 0.0)
  labelClone:setDeck(deck)

  return labelClone
end
