--- RenderedLabel
--
-- Uses the MOAI text rendering system to draw a text label.
-- Each label has its own texture/canvas.
--
-- @author Created by karl on July 20, 2015.

class("RenderedLabel")

require("text.rendered.RenderedCommon")

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Internal -------------------------------------------------------------------

function RenderedLabel:setDimensionsEmptySize()
  self:setDims(TextCommon.EMPTY_WIDTH, self:getHeight())
end

function RenderedLabel:doRender(text)
  return self.renderEngine:renderSingleLine(text)
end

-- Protected API --------------------------------------------------------------

function RenderedLabel:_updateMargins()
  if self.glyphDescriptors then
    local leftmostGlyph = self.glyphDescriptors[1]
    local leftMargin = leftmostGlyph.xOffset / Host.contentScale
    self.leftMargin = leftMargin

    local rightmostGlyph = self.glyphDescriptors[#self.glyphDescriptors]
    local rightMargin = (rightmostGlyph.xAdvance - rightmostGlyph.xOffset - (rightmostGlyph.xMax - rightmostGlyph.xMin)) / Host.contentScale
    self.rightMargin = rightMargin
  end
end

-- Public API -----------------------------------------------------------------

function RenderedLabel:getLeftMargin()
  return self.leftMargin
end

function RenderedLabel:getRightMargin()
  return self.rightMargin
end

function RenderedLabel:disposeOfSelf()
  self:freeResources()
end

function RenderedLabel:getGlyphTable()
  if (self.glyphDescriptors) then
    return self.glyphDescriptors
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new label.
-- @param text (string or number) - What to display in the label.
-- @param font (Font) - The font to use.
-- @param width (number, optional) - The width of the label.
-- @param horizontalAlignment (alignment, optional) - What text alignment to use.
-- @param shouldBuildGlyphDescriptors (boolean, default false) - If true, build glyph descriptors for this label.
function RenderedLabel.new(text, font, width, horizontalAlignment, shouldBuildGlyphDescriptors)
  local isFixedWidth = width ~= nil
  width = width or TextCommon.EMPTY_WIDTH

  local self = Group.new(width, font:getActualSize())
  Mixin.mix(self, RenderedCommon)
  Mixin.mix(self, RenderedLabel)

  self.renderEngine = MOAITextRenderer.new()
  self.renderEngine:setFont(font:getMOAIFont())
  self.renderEngine:setForceSingleLine(true)
  if font:isVariableSized() and isFixedWidth then
    self.renderEngine:setMinFontSize(font:getActualMinSize() * Host.contentScale)
    self.renderEngine:setMaxFontSize(font:getActualSize() * Host.contentScale)
    self.renderEngine:setHeight(font.size * Host.contentScale)
    self.requiresFontSizeCalculation = true
  else
    self.renderEngine:setFontSize(font:getActualSize() * Host.contentScale)
  end
  if isFixedWidth then
    self.renderEngine:setWidth(width * Host.contentScale)
  end
  if horizontalAlignment then
    self.renderEngine:setAlignment(horizontalAlignment)
  end
  self.renderEngine:setReturnGlyphBounds(shouldBuildGlyphDescriptors)

  self.isFixedWidth = isFixedWidth
  self.horizontalAlignment = horizontalAlignment
  self.font = font
  self.latestRender = {}
  self.text = text
  self:setText(text)

  return self
end
