--- RenderedTextArea
--
-- Uses the MOAI text rendering system to draw a text area.
-- Each text area has its own texture/canvas.
--
-- @author Created by karl on July 23, 2015.

class("RenderedTextArea")

require("text.rendered.RenderedCommon")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getHeightForMaxWidth(self, text, maxWidth)
  local moaiFont = self.font:getMOAIFont()
  local actualSize = self.font:getActualSize()

  local scaleNeededByMOAIFontCalculations = Host.contentScale
  local _, height = moaiFont:dimensionsWithMaxWidth(text,
    actualSize * scaleNeededByMOAIFontCalculations,
    maxWidth * scaleNeededByMOAIFontCalculations,
    self.wordBreakMode,
    false,
    self.lineSpacing)
  return height / scaleNeededByMOAIFontCalculations
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Internal -------------------------------------------------------------------

function RenderedTextArea:setDimensionsEmptySize()
  local width = self:getWidth()
  local height = TextCommon.EMPTY_WIDTH
  if self.isFixedHeight then
    height = self:getHeight()
  end
  self:setDims(width, height)
end

function RenderedTextArea:doRender(text)
  local width, height = self:getDims()
  if not self.isFixedHeight then
    height = getHeightForMaxWidth(self, text, width)
    self.renderEngine:setHeight(height * Host.contentScale)
  end

  local texture, glyphDescriptors = self.renderEngine:render(text)
  local renderedWidth, renderedHeight = width * Host.contentScale, height * Host.contentScale

  return texture, renderedWidth, renderedHeight, glyphDescriptors
end

function RenderedTextArea:disposeOfSelf()
  self:freeResources()
end

-- Protected -------------------------------------------------------------------

-- Protected API --------------------------------------------------------------
function RenderedTextArea:_updateMargins()
  -- Do nothing. Margins do not make sense for multilined
  -- text and RenderedTextAreas are setForceSingleLine(false)
  -- by definition.
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

---- Build a new text area.
-- @param text (string) - What to display in the text area.
-- @param font (Font) - The font to use.
-- @param width (number, optional) - The width of the text area.
-- @param horizontalAlignment (alignment, optional) - What horizontal text alignment to use.
-- @param width (number, optional) - The height of the text area.
-- @param verticalAlignment (alignment, optional) - What vertical text alignment to use.
-- @param lineSpacing (numberm optional) - Spacing between lines (default 1).
-- @param wordBreakMode (mode) - Word break mode (default MOAITextBox.WORD_BREAK_HYPHEN)
function RenderedTextArea.new(text, font, width, horizontalAlignment, height, verticalAlignment, lineSpacing, wordBreakMode)
  local isFixedHeight = height ~= nil
  local self = Group.new(width, height or font:getActualSize())
  Mixin.mix(self, RenderedCommon)
  Mixin.mix(self, RenderedTextArea)

  self.renderEngine = MOAITextRenderer.new()
  self.renderEngine:setFont(font:getMOAIFont())
  self.renderEngine:setForceSingleLine(false)
  if font:isVariableSized() and isFixedHeight then
    self.renderEngine:setMinFontSize(font:getActualMinSize() * Host.contentScale)
    self.renderEngine:setMaxFontSize(font:getActualSize() * Host.contentScale)
    self.requiresFontSizeCalculation = true
  else
    self.renderEngine:setFontSize(font:getActualSize() * Host.contentScale)
  end
  self.renderEngine:setWidth(width * Host.contentScale)
  if isFixedHeight then
    self.renderEngine:setHeight(height * Host.contentScale)
  end
  self.renderEngine:setAlignment(horizontalAlignment, verticalAlignment)
  self.renderEngine:setLineSpacing(lineSpacing)
  self.renderEngine:setWordBreak(wordBreakMode)

  self.wordBreakMode = wordBreakMode
  self.lineSpacing = lineSpacing
  self.isFixedHeight = isFixedHeight
  self.font = font
  self.latestRender = {}
  self:setText(text)

  return self
end
