--- BaseTutorialOverlay
-- @author Created by aaroncantor on August 06, 2020.

class("BaseTutorialOverlay")
require("tutorials.TutorialUtils")
require("graphics.OffscreenRenderedTexture")
require("utils.ShaderUtils")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local RIM_COLOR = {0.01, 0.78, 0.98, 1.0}
local FADE_DURATION = 1
local RIM_SCALE_MAX = 10
local INVISIBLE_COLOR = {0.0, 0.0, 0.0, 0.0}
local DEFAULT_OVERLAY_COLOR = {1, 1, 1, 1}
local DEFAULT_ANIM_IN_DURATION = 1
local DEFAULT_ANIM_OUT_DURATION = 1
local DEFAULT_BUTTON_COLOR = {1, 1, 1}
local DEFAULT_BUTTON_OPACITY = 1
local PULSE_GROUPING_DELAY = 0.9
local NUMBER_OF_PULSES = 3
local PULSE_DELAY = 0.2
local BUTTON_FADE_DURATION = 0.167

local MASK_SHADER_UNIFORM_INDEX = {
  object = 1,
  mask = 2,
  minAlpha = 3
}

local PULSE_SHADER_UNIFORM_INDEX = {
  object = 1,
  opacity = 2,
  minAlpha = 3,
  color = 4
}

BaseTutorialOverlay.Z_INDICES = {
  OVERLAY = 10,
  MESSAGE = 15,
  BUTTON = 20
}

-- ----------------------------------------------------------------------------
-- Shaders
-- ----------------------------------------------------------------------------

local MASK_VSH = [[
  attribute vec4 position;
  attribute vec2 uv;
  attribute vec4 color;

  varying vec2 uvVarying1;

  void main ()
  {
    gl_Position = position;
    uvVarying1 = uv;
  }
]]

local MASK_FSH = [[
  varying MEDP vec2 uvVarying1;

  uniform sampler2D object;
  uniform sampler2D mask;
  uniform LOWP float minAlpha;

  void main()
  {
    gl_FragColor = texture2D(object, uvVarying1);
    LOWP float a = 1.0 - ceil(texture2D(mask, uvVarying1).a - minAlpha);
    gl_FragColor.a *= a;
    gl_FragColor.rgb = gl_FragColor.rgb * a;
  }
]]

local PULSE_FSH = [[
  varying MEDP vec2 uvVarying1;

  uniform sampler2D object;
  uniform LOWP float opacity;
  uniform LOWP float minAlpha;
  uniform LOWP vec4 color;

  void main()
  {
    gl_FragColor = texture2D(object, uvVarying1);
    gl_FragColor.a = ceil(gl_FragColor.a - minAlpha) * opacity;
    gl_FragColor.rgb = color.rgb * gl_FragColor.a;
  }
]]

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getMaskShader(self)
  local maskShader = MOAIShader.new()
  local uniformCount = TableUtils.getSize(MASK_SHADER_UNIFORM_INDEX)

  maskShader:reserveUniforms(uniformCount)
  maskShader:declareUniformSampler(MASK_SHADER_UNIFORM_INDEX.object, "object", 1)
  maskShader:declareUniformSampler(MASK_SHADER_UNIFORM_INDEX.mask, "mask", 2)
  maskShader:declareUniformFloat(MASK_SHADER_UNIFORM_INDEX.minAlpha, "minAlpha", self.minAlpha)

  maskShader:setVertexAttribute(1, "position")
  maskShader:setVertexAttribute(2, "uv")
  maskShader:setVertexAttribute(3, "color")

  maskShader:load(MASK_VSH, MASK_FSH)
  return maskShader
end

local function getPulseShader(self)
  local pulseShader = MOAIShader.new()
  local uniformCount = TableUtils.getSize(PULSE_SHADER_UNIFORM_INDEX)

  pulseShader:reserveUniforms(uniformCount)
  pulseShader:declareUniformSampler(PULSE_SHADER_UNIFORM_INDEX.object,"object", 1)
  pulseShader:declareUniformFloat(PULSE_SHADER_UNIFORM_INDEX.opacity, "opacity", 1.0)
  pulseShader:declareUniformFloat(PULSE_SHADER_UNIFORM_INDEX.minAlpha, "minAlpha", self.minAlpha)
  pulseShader:declareUniform(PULSE_SHADER_UNIFORM_INDEX.color, "color", MOAIShader.UNIFORM_COLOR)

  pulseShader:setVertexAttribute(1, "position")
  pulseShader:setVertexAttribute(2, "uv")
  pulseShader:setVertexAttribute(3, "color")

  pulseShader:load(MASK_VSH, PULSE_FSH)

  local moaiColor = MOAIColor.new()
  moaiColor:setColor(unpack(RIM_COLOR))
  pulseShader:setAttrLink(PULSE_SHADER_UNIFORM_INDEX.color, moaiColor, MOAIColor.COLOR_TRAIT)
  return pulseShader
end

local function createPulse(self, shape)
  if shape.trackCoroutine and shape.trackCoroutine:isBusy() then
    shape.trackCoroutine:stop()
  end
  local pulseWidth, pulseHeight = shape:getDims()

  local pulseGroup = Group.new(pulseWidth, pulseHeight)
  pulseGroup:insertAtCenter(shape)

  local pulseSprite = TutorialUtils.createSprite(pulseGroup, shape:getRot())
  local pulseShader = getPulseShader(self)
  pulseSprite:setShader(pulseShader)

  return pulseSprite, pulseShader
end

local function getOverlayDeck(mask, overlay)
  local multitexture = MOAIMultiTexture.new()
  local uniformCount = TableUtils.getSize(MASK_SHADER_UNIFORM_INDEX)
  multitexture:reserve(uniformCount)

  local maskTexture = TutorialUtils.createTexture(mask)
  local overlayTexture = TutorialUtils.createTexture(overlay)
  multitexture:setTexture(MASK_SHADER_UNIFORM_INDEX.mask, maskTexture)
  multitexture:setTexture(MASK_SHADER_UNIFORM_INDEX.object, overlayTexture)

  -- assuming textures are same size
  local gfxQuad = MOAIGfxQuad2D.new()
  gfxQuad:setTexture(multitexture)

  local textureWidth, textureHeight = maskTexture:getSize()

  local rectOriginX = 0
  local rectOriginY = 0
  local rectWidth = textureWidth / Host.contentScale
  local rectHeight = textureHeight / Host.contentScale

  gfxQuad:setRect(
    rectOriginX,
    rectOriginY,
    rectOriginX + rectWidth,
    rectOriginY + rectHeight
  )

  gfxQuad:setUVRect(0, 1, 1, 0)

  return gfxQuad
end

local function createMaskedOverlay(self, overlay, holes)
  local mask = Group.newFullScreen()
  for _, hole in ipairs(holes) do
    mask:takeChild(hole)
  end

  local overlayDeck = getOverlayDeck(mask, overlay)
  local overlaySprite = Sprite.newWithDeck(overlayDeck)

  local overlayShader = getMaskShader(self)
  overlaySprite:setShader(overlayShader)

  return overlaySprite, mask, overlayShader
end

local function runPulseActionForPanel(self, trackingShape, panel, shader, opacityController, scaleXTo, scaleYTo)
  local panelScaleAction = AHScaleTo.new(panel, FADE_DURATION, nil, scaleXTo, scaleYTo)
  local panelFadeAction = AHOpacityTo.new(opacityController, FADE_DURATION, AHEaseSimpleOut.new(), 0)
  local combinedAction = AHSpawn.new(panelScaleAction, panelFadeAction)

  local function updatePulse()
    while true do
      shader:setAttr(PULSE_SHADER_UNIFORM_INDEX.opacity, panel:getOpacity())
      local pulseX, pulseY = trackingShape:getLoc()
      panel:setLoc(pulseX, pulseY)
      coroutine.yield()
    end
  end
  local pulseCoroutine = runInCoroutine(updatePulse)

  local function removePanel()
    if pulseCoroutine and pulseCoroutine:isBusy() then
      pulseCoroutine:stop()
    end

    TutorialUtils.stopMultipyingOpacity(panel)

    local rimPanels = self.rimPanels[trackingShape]
    if rimPanels then
      TableUtils.removeValueFromArray(self.rimPanels[trackingShape], panel)
    end
    panel:removeSelf()
  end
  local fullAction = AHSequence.new(combinedAction, removePanel)
  fullAction:start()
end

local function animatePulse(self, shapeFactory, trackingShape)
  for _ = 1, NUMBER_OF_PULSES do
    waitFor(PULSE_DELAY)
    local pulse, shader = createPulse(self, shapeFactory())

    opacityController = TutorialUtils.multiplyOpacity(pulse, self.highlightFadeController)
    self.overlay:insert(pulse)
    table.insert(self.rimPanels[trackingShape], pulse)

    local scaleTarget = 2 * Host.contentScale * self.pulseSize
    local scaleX = 1 + scaleTarget / trackingShape:getWidth()
    local scaleY = 1 + scaleTarget / trackingShape:getHeight()
    runPulseActionForPanel(self, trackingShape, pulse, shader, opacityController, scaleX, scaleY)
  end
end

local function createRimAnimation(self, shapeFactory, trackingShape)
  self.rimPanels[trackingShape] = {}
  local function animateRimForever()
    while(true) do
      animatePulse(self, shapeFactory, trackingShape)
      waitFor(PULSE_GROUPING_DELAY)
    end
  end
  return runInCoroutine(animateRimForever)
end

local function stopAnimatingRim(self, highlightShape)
  local animateRimCoroutine = self.animateRimCoroutines[highlightShape]
  if animateRimCoroutine and animateRimCoroutine:isBusy() then
    animateRimCoroutine:stop()
  end

  local rimPanels = self.rimPanels[highlightShape]
  if rimPanels then
    for _, panel in ipairs(rimPanels) do
      panel:removeSelf()
    end
  end

  self.rimPanels[highlightShape] = nil
end

local function buttonTapHandler(self, touchEvent, isOverlay)
  if (not self.tapLock or not self.buttonsEnabled) then
    return
  end

  if (isOverlay and not self.overlayTapEnabled) then
    return
  end

  self.screenTapX = touchEvent.x
  self.screenTapY = touchEvent.y
  self.tapLock = false
end

local function setupButton(self, button, isOverlay)
  local function endPulses()
    if not self.tapLock then
      return
    end

    stopAnimatingRim(self, button)
  end

  local function onTargetTap(touchEvent)
    if (not self.tapLock or not self.buttonsEnabled) then
      return
    end

    if (not isOverlay) then
      AHOpacityTo.new(self.buttonOpacityControllers[button], BUTTON_FADE_DURATION, AHEaseLinear.new(), 0):start()
    end

    endPulses()

    buttonTapHandler(self, touchEvent, isOverlay)

    if (not isOverlay) then
      button:removeSelf()
    end
  end

  local function onTouchDown(_, touchEvent)
    if (not self.buttonsEnabled) then
      return
    end

    if not button.onTouchUp then
      onTargetTap(touchEvent)
    end

    if button.onTouchDown then
      button.onTouchDown()
    end
  end

  local function onTouchUp(_, touchEvent)
    if (not self.buttonsEnabled) then
      return
    end

    if button.onTouchUp then
      onTargetTap(touchEvent)
      button.onTouchUp()
    end
  end

  Button.makeButton(button, onTouchDown, onTouchUp, onTouchUp)
end

local function addPassThroughButton(self, button)
  table.insert(self.passThroughButtons, button)
  if self.shouldBlockTaps then
    setupButton(self, button)
  end
end

local function getMasks(self)
  local allMasks = {}
  if self.highlightShapeFactories then
    for _, factory in ipairs(self.highlightShapeFactories) do
      local button = factory()
      button:linkAttrToChildren(MOAIColor.ATTR_OPACITY, MOAIColor_ATTR_OPACITY, true)
      button:setColor(unpack(self.buttonColor))
      button:setOpacity(self.buttonOpacity)
      self.buttonOpacityControllers[button] = TutorialUtils.multiplyOpacity(button, self.highlightFadeController)

      self:insert(button, BaseTutorialOverlay.Z_INDICES.BUTTON)
      addPassThroughButton(self, button)

      local mask = factory()
      table.insert(allMasks, mask)

      if self.shouldAutoPulse then
        self.animateRimCoroutines[button] = createRimAnimation(self, factory, button)
      end
    end
  end

  if self.secondaryHighlightShapes then
    for _, mask in ipairs(self.secondaryHighlightShapes) do
      table.insert(allMasks, mask)
    end
  end

  return allMasks
end

local function createHighlight(self, allMasks)
  local maskedOverlay, mask, shader = createMaskedOverlay(self, self.overlay, allMasks)
  maskedOverlay:setAnchorPointBottomLeft()
  maskedOverlay:setLoc(0, 0)

  if self.shouldBlockTaps then
    setupButton(self, maskedOverlay, true)
  end

  self:insert(maskedOverlay, BaseTutorialOverlay.Z_INDICES.OVERLAY)
  self.maskedOverlay = maskedOverlay
  self.mask = mask
  self.maskShader = shader
end

local function clearHighlights(self)
  self.childrenToDispose = TableUtils.joinArrayToSet(self.mask.children, self.childrenToDispose)
  self.mask:removeAll(false)

  for _, button in ipairs(self.passThroughButtons) do
    stopAnimatingRim(self, button)
    button:removeSelf()
  end

  self.passThroughButtons = {}
  self.animateRimCoroutines = {}
end

local function createFadeInHighlightsAction(self, fadeTime)
  fadeTime = fadeTime or DEFAULT_ANIM_IN_DURATION
  return AHOpacityTo.new(self.highlightFadeController, fadeTime, AHEaseLinear.new(), 1)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BaseTutorialOverlay:linkOpacity(group)
  return TutorialUtils.multiplyOpacity(group, self.fadeController)
end

function BaseTutorialOverlay:setPulseSize(size)
  self.pulseSize = size
end

function BaseTutorialOverlay:runSinglePulse()
  for _, shapeFactory in ipairs(self.highlightShapeFactories) do
    local trackingShape = shapeFactory()
    self.rimPanels[trackingShape] = {}
    animatePulse(self, shapeFactory, trackingShape)
  end
end

function BaseTutorialOverlay:waitForTargetTap()
  self.tapLock = true
  waitUntil(function() return self.tapLock == false end)
  self.overlayTapEnabled = false
  return self.screenTapX, self.screenTapY
end

function BaseTutorialOverlay:tap()
  for _, button in ipairs(self.passThroughButtons) do
    stopAnimatingRim(self, button)
    button:removeSelf()
  end
  self.tapLock = false
end

function BaseTutorialOverlay:setOpacity(opacity)
  self.fadeController:setOpacity(opacity)
end

function BaseTutorialOverlay:createOpacityAnimation(animTime, destOpacity)
  return AHOpacityTo.new(self.fadeController, animTime, AHEaseLinear.new(), destOpacity)
end

function BaseTutorialOverlay:createFadeInAnimation(animTime)
  animTime = animTime or self.animInDuration
  return self:createOpacityAnimation(animTime, 1.0)
end

function BaseTutorialOverlay:createFadeOutAnimation(animTime)
  animTime = animTime or self.animOutDuration
  return self:createOpacityAnimation(animTime, 0.0)
end

function BaseTutorialOverlay:setOverlayColor(r, g, b, a)
  self.overlay:setColor(r, g, b)
  self.overlayOpacityController:setOpacity(a)
end

function BaseTutorialOverlay:enableOverlayTap(enabled, onTouchDown, onTouchUp)
  self.maskedOverlay.onTouchDown = onTouchDown
  self.maskedOverlay.onTouchUp = onTouchUp
  self.overlayTapEnabled = enabled
end

function BaseTutorialOverlay:enableButtons(enabled)
  self.buttonsEnabled = enabled
end

function BaseTutorialOverlay:setButtonColor(r, g, b, a)
  self.buttonColor = {r,g,b}
  self.buttonOpacity = a

  for _, button in ipairs(self.passThroughButtons) do
    button:setColor(unpack(self.buttonColor))
    local opacityController = self.buttonOpacityControllers[button]
    opacityController:setOpacity(self.buttonOpacity)
  end
end

function BaseTutorialOverlay:setAnimInDuration(duration)
  self.animInDuration = duration
end

function BaseTutorialOverlay:setAnimOutDuration(duration)
  self.animOutDuration = duration
end

function BaseTutorialOverlay:createFadeAndClearHighlightsAction(fadeTime)
  fadeTime = fadeTime or DEFAULT_ANIM_OUT_DURATION
  local combinedAction = AHSequence.new(
    AHOpacityTo.new(self.highlightFadeController, fadeTime, AHEaseLinear.new(), 0),
    clearHighlights(self)
  )

  return combinedAction
end

function BaseTutorialOverlay:resetHighlights(highlightShapeFactories, secondaryHighlightShapes, fadeTime)
  wait(self:createFadeAndClearHighlightsAction(fadeTime):start())

  self.highlightShapeFactories = highlightShapeFactories
  self.secondaryHighlightShapes = secondaryHighlightShapes

  local allMasks = getMasks(self)
  for _, mask in ipairs(allMasks) do
    self.mask:insert(mask)
  end

  createFadeInHighlightsAction(self, fadeTime):start()
end

function BaseTutorialOverlay:reinitialize()
  self:resetHighlights(self.highlightShapeFactories, self.secondaryHighlightShapes)
end

function BaseTutorialOverlay:disposeOfSelf()
  if self.childrenToDispose then
    for child, _ in pairs(self.childrenToDispose) do
      child:removeSelf()
    end
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

local function createOverlay(self)
  self:setOpacity(0.0)
  self.overlay = FilledRectangle.new(Screen.fullWidth, Screen.fullHeight, DEFAULT_OVERLAY_COLOR)
  self.overlayOpacityController = TutorialUtils.multiplyOpacity(self.overlay, self.fadeController)
end

local function createMasks(self)
  local allMasks = getMasks(self)
  createHighlight(self, allMasks)
end

function BaseTutorialOverlay.new(highlightShapeFactories, secondaryHighlightShapes, minAlpha, shouldAutoPulse, buttonsEnabled, shouldBlockTaps)
  local self = FilledRectangle.newFullScreen(INVISIBLE_COLOR)
  Mixin.mix(self, BaseTutorialOverlay)

  self.fadeController = Group.newMinimumSized()
  local highlightFadeController = Group.newMinimumSized()
  self.highlightFadeController = TutorialUtils.multiplyOpacity(highlightFadeController, self.fadeController)

  self.pulseSize = RIM_SCALE_MAX
  self.highlightShapeFactories = highlightShapeFactories
  self.secondaryHighlightShapes = secondaryHighlightShapes
  self.minAlpha = minAlpha or 0
  self.buttonColor = DEFAULT_BUTTON_COLOR
  self.buttonOpacity = DEFAULT_BUTTON_OPACITY
  self.animInDuration = DEFAULT_ANIM_IN_DURATION
  self.animOutDuration = DEFAULT_ANIM_OUT_DURATION

  if (shouldAutoPulse == nil) then
    self.shouldAutoPulse = true
  else
    self.shouldAutoPulse = shouldAutoPulse
  end

  if (buttonsEnabled == nil) then
    self.buttonsEnabled = true
  else
    self.buttonsEnabled = buttonsEnabled
  end

  if (shouldBlockTaps == nil) then
    self.shouldBlockTaps = true
  else
    self.shouldBlockTaps = shouldBlockTaps
  end

  self.animateRimCoroutines = {}
  self.rimPanels = {}
  self.passThroughButtons = {}
  self.buttonOpacityControllers = {}
  self.tapLock = false
  self.overlayTapEnabled = false

  createOverlay(self)
  createMasks(self)

  return self
end
