--- TutorialHighlighter
-- @author created by Joe Poka on May 2, 2016

class("TutorialHighlighter")

require("tutorials.BaseTutorialOverlay")
require("tutorials.TutorialAnalytics")
-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FLOAT_DURATION = 2
local FLOAT_DISTANCE = 3
local HOLE_ARROW_PADDING = 20
local LABEL_ARROW_PADDING = 20
local ARROW_THICKNESS = 5
local ARROW_HEAD_ANGLE = -math.pi / 6
local ARROW_HEAD_LENGTH = 15
local DEFAULT_OVERLAY_COLOR = {0.039, 0.055, 0.082, 0.7}
local DEFAULT_ANIM_IN_DURATION = 0.167
local DEFAULT_ANIM_OUT_DURATION = 0.033
local DEFAULT_BUTTON_COLOR = {0, 0, 0, 0}
local INSTRUCTION_LABEL_WIDTH = Screen.getSafeWidthPart(0.6)

local DEFAULT_FONT = FontFaces.DIN_OT_MEDIUM
local DEFAULT_FONT_SIZE = 20
local LABEL_TEXT_COLOR = {1, 1, 1, 1}
local LABEL_X = Screen.getSafeX(0.5)
local LABEL_WIDTH = Screen.getSafeWidthPart(0.8)
local LABEL_BACKGROUND_COLOR = {0, 0, 0, 0.8}
local LABEL_BACKGROUND_CORNER_RADIUS = 5
local LABEL_BACKGROUND_POINTER_SIDE_LENGTH = 18
local LABEL_BACKGROUND_MARGIN_X = 20
local LABEL_BACKGROUND_MARGIN_Y = 20
local LABEL_SLIDE_IN_OFFSET_Y = -20
local LABEL_SLIDE_OUT_OFFSET_Y = 10
local DEFAULT_LABEL_FADE_IN_DURATION = 0.167
local DEFAULT_LABEL_FADE_OUT_DURATION = 0.333
local LABEL_SLIDE_IN_DURATION = 0.667
local LABEL_SLIDE_OUT_DURATION = 0.333
local LABEL_GLOW_FILENAME = "shared_tutorial_glow.png"
local LABEL_GLOW_MARGIN_X = 100
local LABEL_GLOW_MARGIN_Y = 100
local LABEL_POINTER_MARGIN = 2

TutorialHighlighter.DEFAULT_LABEL_Y = Screen.getSafeY(0.24)

-- Static variables
TutorialHighlighter.hasSeenTutorial = false

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createArrowHead(headAngle, startX, startY)
  local x = startX - ARROW_HEAD_LENGTH * math.cos(headAngle)
  local y = startY - ARROW_HEAD_LENGTH * math.sin(headAngle)
  local head = Line.new(startX, startY, x, y)
  head:setLineWidth(ARROW_THICKNESS)
  return head
end

local function createArrow(self, targetX, targetY)
  local arrowStartX = LABEL_X
  local arrowStartY = TutorialHighlighter.DEFAULT_LABEL_Y + self.instructionLabel:getHeightPart(0.5) + LABEL_ARROW_PADDING

  local arrowAngle = math.atan2((targetY-arrowStartY), (targetX-arrowStartX))
  local arrowEndX = targetX - HOLE_ARROW_PADDING * math.cos(arrowAngle)
  local arrowEndY = targetY - HOLE_ARROW_PADDING * math.sin(arrowAngle)

  local arrowBody = Line.new(arrowStartX, arrowStartY, arrowEndX, arrowEndY)
  arrowBody:setAnchorPointBottomLeft()
  arrowBody:setLineWidth(ARROW_THICKNESS)
  self:insert(arrowBody)

  local arrowLeftHead = createArrowHead(ARROW_HEAD_ANGLE + arrowAngle, arrowEndX, arrowEndY)
  arrowLeftHead:setAnchorPointBottomLeft()
  self:insert(arrowLeftHead)

  local arrowRightHead = createArrowHead(-ARROW_HEAD_ANGLE + arrowAngle, arrowEndX, arrowEndY)
  arrowRightHead:setAnchorPointBottomLeft()
  self:insert(arrowRightHead)

  if (self.fadeController) then
    self:linkOpacity(arrowBody)
    self:linkOpacity(arrowLeftHead)
    self:linkOpacity(arrowRightHead)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TutorialHighlighter:addInstructionLabel(instructionMessage, fontSize, instructionLocationY, shouldShowArrows, labelWidth)
  instructionLocationY = instructionLocationY or TutorialHighlighter.DEFAULT_LABEL_Y
  shouldShowArrows = shouldShowArrows or false
  labelWidth = labelWidth or INSTRUCTION_LABEL_WIDTH

  local instructionFont = Font.new(FontFaces.DEFAULT_FONT, fontSize)

  local instructionLabel = DynamicTextArea.newSimple(instructionMessage, labelWidth, instructionFont, Alignment.CENTER)
  instructionLabel:setLoc(LABEL_X, instructionLocationY)
  self:insert(instructionLabel, BaseTutorialOverlay.Z_INDICES.MESSAGE)
  self:linkOpacity(instructionLabel)

  self.instructionLabel = instructionLabel

  local floatUpAction = AHLocationYBy.new(instructionLabel, FLOAT_DURATION, AHEaseSimpleInOut.new(), FLOAT_DISTANCE)
  local floatDownAction = AHLocationYBy.new(instructionLabel, FLOAT_DURATION, AHEaseSimpleInOut.new(), -FLOAT_DISTANCE)
  local floatUpAndDown = AHSequence.new(floatUpAction, floatDownAction)
  local floatForever = AHLoop.new(floatUpAndDown)
  self.floatInstructionLabelAction = floatForever:start()

  if shouldShowArrows then
    for _, tapArea in ipairs(self.passThroughButtons) do
      local maskX, maskY = tapArea:getLoc()
      local maskHeight = tapArea:getHeight()
      createArrow(self, maskX, maskY - (maskHeight / 2))
    end
  end
end

function TutorialHighlighter:addBubbleLabel(text, font, fontSize, labelPosY, shouldShowPointer, isPointerUp, pointerOffsetX, tagAttributes, shouldGlow)
  font = font or DEFAULT_FONT
  fontSize = fontSize or DEFAULT_FONT_SIZE
  labelPosY = labelPosY or TutorialHighlighter.DEFAULT_LABEL_Y
  shouldShowPointer = shouldShowPointer or false
  isPointerUp = isPointerUp or false
  pointerOffsetX = pointerOffsetX or 0

  local helpFont = Font.newWithActualSize(font, fontSize)
  local labelAttributes = {
    font = helpFont,
    color = LABEL_TEXT_COLOR
  }

  local helpLabel = DynamicTextArea.new(text, LABEL_WIDTH, labelAttributes, tagAttributes, Alignment.LEFT, Alignment.CENTER)

  local backgroundWidth = helpLabel:getWidth() + (2 * LABEL_BACKGROUND_MARGIN_X)
  local backgroundHeight = helpLabel:getHeight() + (2 * LABEL_BACKGROUND_MARGIN_Y)
  local helpBackground = FilledRoundedRectangle.new(backgroundWidth, backgroundHeight, LABEL_BACKGROUND_CORNER_RADIUS, LABEL_BACKGROUND_COLOR)
  helpBackground:insertAtCenter(helpLabel)
  helpLabel:inheritOpacity(helpBackground)

  local glow = Sprite.newWithName(LABEL_GLOW_FILENAME)
  local glowOpacity = 0
  if shouldGlow then
    glowOpacity = 1.0
  end
  glow:setOpacity(glowOpacity)

  local glowWidth = backgroundWidth + (2 * LABEL_GLOW_MARGIN_X)
  local glowHeight = backgroundHeight + (2 * LABEL_GLOW_MARGIN_Y)
  glow:setEffectiveDims(glowWidth, glowHeight)

  local glowZ = 0
  local pointerZ = 10
  local backgroundZ = 20
  local helpGroup = Group.new(glowWidth, glowHeight)
  helpGroup:setOpacity(helpBackground:getOpacity())
  helpGroup:insertAtCenter(helpBackground, backgroundZ)
  helpGroup:insertAtCenter(glow, glowZ)
  helpBackground:inheritOpacity(helpGroup)

  if shouldShowPointer then
    local pointer = FilledRoundedRectangle.new(LABEL_BACKGROUND_POINTER_SIDE_LENGTH, LABEL_BACKGROUND_POINTER_SIDE_LENGTH, 3, LABEL_BACKGROUND_COLOR)
    local pointerX = helpGroup:getWidthPart(0.5) + pointerOffsetX
    local pointerY = LABEL_GLOW_MARGIN_Y + LABEL_POINTER_MARGIN
    if isPointerUp then
      pointerY = pointerY + helpBackground:getHeight() - (2 * LABEL_POINTER_MARGIN)
    end
    pointer:setLoc(pointerX, pointerY)
    pointer:setRot(45)

    helpGroup:insert(pointer, pointerZ)
    pointer:inheritOpacity(helpGroup)
  end

  local helpOpacity = helpGroup:getOpacity()
  helpGroup:setOpacity(1.0)
  local helpSprite = TutorialUtils.createSprite(helpGroup)

  helpSprite:setLoc(LABEL_X, labelPosY)
  helpSprite:setOpacity(0.0)
  self:insert(helpSprite, BaseTutorialOverlay.Z_INDICES.MESSAGE)

  self.helpLabelGroup = helpSprite
  self.helpLabelHeight = backgroundHeight
  self.helpLabelGlow = glow
end

function TutorialHighlighter:setBubbleLabelGlowOpacity(opacity)
  self.helpLabelGlow:setOpacity(opacity)
end

function TutorialHighlighter:setBubbleLabelPosition(yPosition, anchorAtBottom)
  local anchorAtBottom = anchorAtBottom or false
  local offset = self.helpLabelHeight / 2
  if not anchorAtBottom then
    offset = -offset
  end
  self.helpLabelGroup:setY(yPosition + offset)
end

function TutorialHighlighter:createBubbleLabelFadeInAction(fadeTime, slideDistance)
  fadeTime = fadeTime or DEFAULT_LABEL_FADE_IN_DURATION
  slideDistance = slideDistance or -LABEL_SLIDE_IN_OFFSET_Y

  local setInitialState = function()
    self.helpLabelGroup:offsetY(-slideDistance)
    self.helpLabelGroup:setOpacity(0.0)
  end

  local appearAction = AHSpawn.new(
    AHOpacityTo.new(self.helpLabelGroup, fadeTime, AHEaseBezier.newFlat({0.2, 0, 0, 1}), LABEL_BACKGROUND_COLOR[4]),
    AHLocationYBy.new(self.helpLabelGroup, LABEL_SLIDE_IN_DURATION, AHEaseBezier.newFlat({0.2, 0, 0, 1}), slideDistance)
  )

  local combinedAction = AHSequence.new(setInitialState, appearAction)
  return combinedAction
end

function TutorialHighlighter:createRemoveBubbleLabelAction(fadeTime, slideDistance)
  if self.helpLabelGroup then
    fadeTime = fadeTime or DEFAULT_LABEL_FADE_OUT_DURATION
    slideDistance = slideDistance or LABEL_SLIDE_OUT_OFFSET_Y

    local disappearAction = AHSpawn.new(
      AHOpacityTo.new(self.helpLabelGroup, fadeTime, AHEaseBezier.newFlat({0.2, 0, 0, 1}), 0.0),
      AHLocationYBy.new(self.helpLabelGroup, LABEL_SLIDE_OUT_DURATION, AHEaseBezier.newFlat({0.2, 0, 0, 1}), slideDistance)
    )

    local combinedAction = AHSequence.new(
      disappearAction,
      AHRemove.new(self.helpLabelGroup)
    )
    return combinedAction
  else
    return AHDoNothing.new()
  end
end

function TutorialHighlighter:createFadeInAnimation(animTime)
  local combinedAction = AHSpawn.new(BaseTutorialOverlay.createFadeInAnimation(self, animTime))

  if (not self.hasSeenTutorial) then
    self.hasSeenTutorial = true
    combinedAction:addAction(TutorialAnalytics.reportDelayedStart)
  end

  return combinedAction
end

function TutorialHighlighter:resetHighlights(highlightShapeFactories, secondaryHighlightShapes, fadeTime)
  fadeTime = fadeTime or 0
  BaseTutorialOverlay.resetHighlights(self, highlightShapeFactories, secondaryHighlightShapes, fadeTime)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function TutorialHighlighter.newWithMasks(highlightShapeFactories, secondaryHighlightShapes, minAlpha, shouldAutoPulse, enableButtons, shouldBlockTaps)
  local self = BaseTutorialOverlay.new(highlightShapeFactories, secondaryHighlightShapes, minAlpha, shouldAutoPulse, enableButtons, shouldBlockTaps)
  Mixin.mix(self, TutorialHighlighter)

  self:setButtonColor(unpack(DEFAULT_BUTTON_COLOR))
  self:setOverlayColor(unpack(DEFAULT_OVERLAY_COLOR))
  self:setAnimInDuration(DEFAULT_ANIM_IN_DURATION)
  self:setAnimOutDuration(DEFAULT_ANIM_OUT_DURATION)

  return self
end
