--- ColorUtils
-- @author Created by Aaron Barrett on October 21, 2013.

-- ===========================================================================
class("ColorUtils")
-- ===========================================================================

local BYTE_RANGE = 256
local BYTE_MAX_VALUE = BYTE_RANGE - 1
local HEX_BASE = 16

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Get a color with the specified opacity.
-- @param color (array) - The color to add default opacity to.
-- @param opacity (number) - The opacity to use.
-- @return (array) A copy of the color array with the specified opacity.
function ColorUtils.colorWithOpacity(color, opacity)
  return {color[1], color[2], color[3], opacity}
end

--- Get a color with a default opacity added if no opacity was present.
-- @param color (array) - The color to add default opacity to.
-- @param defaultOpacity (number) - The opacity to use (default 1).
-- @return (array) A copy of the color array with original opacity, if any, or the default.
function ColorUtils.colorWithDefaultOpacity(color, defaultOpacity)
  defaultOpacity = defaultOpacity or 1
  return ColorUtils.colorWithOpacity(color, color[4] or defaultOpacity)
end

--- Return a copy of a color with a multiplier applied to the opacity.
-- @param color (array) - The color to copy.
-- @param opacityMultiplier (number) - The opacity multiplier.
-- @return (array) A copy of the color array with the new opacity.
function ColorUtils.colorWithOpacityMultiplier(color, opacityMultiplier)
  local opacity = color[4] or 1
  return ColorUtils.colorWithOpacity(color, opacity * opacityMultiplier)
end

function ColorUtils.channelsToMOAIColor(red, green, blue, opacity)
  local color = MOAIColor.new()
  color:setColor(red, green, blue, opacity)
  return color
end

function ColorUtils.tableToMOAIColor(colorTable)
  local red, green, blue, opacity = unpack(colorTable)
  local color = ColorUtils.channelsToMOAIColor(red, green, blue, opacity)
  return color
end

function ColorUtils.colorHex24ToStd(color24bit)
  local twoByteRange = math.pow(BYTE_RANGE, 2)
  local r = math.abs((color24bit / twoByteRange) % BYTE_RANGE) / BYTE_MAX_VALUE
  local g = math.abs((color24bit / BYTE_RANGE) % BYTE_RANGE) / BYTE_MAX_VALUE
  local b = math.abs(color24bit % BYTE_RANGE) / BYTE_MAX_VALUE

  r = MathUtils.clamp(r, 0, 1)
  g = MathUtils.clamp(g, 0, 1)
  b = MathUtils.clamp(b, 0, 1)

  local resultColor = {r, g, b, 1}
  return resultColor
end

function ColorUtils.colorHex32ToStd(color32bit)
  local twoByteRange = math.pow(BYTE_RANGE, 2)
  local threeByteRange = math.pow(BYTE_RANGE, 3)
  local r = math.abs((color32bit / threeByteRange) % BYTE_RANGE) / BYTE_MAX_VALUE
  local g = math.abs((color32bit / twoByteRange) % BYTE_RANGE) / BYTE_MAX_VALUE
  local b = math.abs((color32bit / BYTE_RANGE) % BYTE_RANGE) / BYTE_MAX_VALUE
  local a = math.abs(color32bit % BYTE_RANGE) / BYTE_MAX_VALUE

  local resultColor = {r, g, b, a}
  return resultColor
end

function ColorUtils.colorStdToRGB(colorStdTable)
  local r = colorStdTable[1] * BYTE_MAX_VALUE
  local g = colorStdTable[2] * BYTE_MAX_VALUE
  local b = colorStdTable[3] * BYTE_MAX_VALUE
  local a = BYTE_MAX_VALUE

  if colorStdTable[4] then
    a = colorStdTable[4]
  end

  local resultColor = {r, g, b, a}
  return resultColor
end

function ColorUtils.colorRGBToStd(r, g, b, a)
  r = r / BYTE_MAX_VALUE
  g = g / BYTE_MAX_VALUE
  b = b / BYTE_MAX_VALUE

  if (a ~= nil) then
    a = a / BYTE_MAX_VALUE
  else
    a = 1.0
  end

  local resultColor = {r, g, b, a}
  return resultColor
end

function ColorUtils.colorHex24StringToStd(color24bitString)
  local strippedColorString = string.sub(color24bitString, 2)
  local hex24Number = tonumber(strippedColorString, HEX_BASE)
  return ColorUtils.colorHex24ToStd(hex24Number)
end

function ColorUtils.colorHex32StringToStd(color32bitString)
  local strippedColorString = string.sub(color24bitString, 2)
  local hex32Number = tonumber(strippedColorString, HEX_BASE)
  return ColorUtils.colorHex32ToStd(hex32Number)
end

-- YIQ (NTSC color TV space) weights for determining contrast
-- taken from http://24ways.org/2010/calculating-color-contrast/
function ColorUtils.getTextColorForBackingColorYIQ(color)
  local redWeighted = color[1] * 299
  local greenWeighted = color[2] * 587
  local blueWeighted = color[3] * 114
  local yiq = (redWeighted + greenWeighted + blueWeighted)/1000
  if (yiq >= 0.5) then
    return {0.0, 0.0, 0.0} -- black text would read better
  else
    return {1.0, 1.0, 1.0} -- white text would read better
  end
end

--- Adds the channel values of two color arrays together to produce a combined color
-- Note that channel values can be negative.  If the combined color channel value
-- falls outside of the range [0.0, 1.0] it will be cropped.  If alpha is not included
-- in either color, an alpha of 1.0 will be used.
-- @param color1 - The first color to add
-- @param color2 - The second color to add
-- @return - the color resulting from adding the channels in color1 and color2 together
function ColorUtils.addColors(color1, color2)
  local alphaChannelIndex = 4

  local color1Copy = TableUtils.copyKeysFromTable(color1)
  local color2Copy = TableUtils.copyKeysFromTable(color2)
  color1Copy[alphaChannelIndex] = color1Copy[alphaChannelIndex] or 1.0
  color2Copy[alphaChannelIndex] = color2Copy[alphaChannelIndex] or 1.0

  local combinedColor = {}
  for channelIndex = 1, alphaChannelIndex do
    local color1ChannelValue = color1Copy[channelIndex]
    local color2ChannelValue = color2Copy[channelIndex]
    local combinedChannelValue = color1ChannelValue + color2ChannelValue

    -- Crop each channel value if it is out of the range [0.0, 1.0]
    combinedChannelValue = MathUtils.clamp(combinedChannelValue, 0.0, 1.0)

    table.insert(combinedColor, combinedChannelValue)
  end

  return combinedColor
end

-- specify 2 to get a color more like c2 than c1
-- t == 0 means the color is c1
-- t == 1 means the color is c2
function ColorUtils.makeColorBetweenTwoColors(c1, c2, t)
  c1 = ColorUtils.colorWithDefaultOpacity(c1, c1[4])
  c2 = ColorUtils.colorWithDefaultOpacity(c2, c2[4])

  local c1t = (1 - t)
  local c2t = t

  local r = c1[1] * c1t  + c2[1] * c2t
  local g = c1[2] * c1t  + c2[2] * c2t
  local b = c1[3] * c1t  + c2[3] * c2t
  local a = c1[4] * c1t  + c2[4] * c2t

  local resultColor = {r, g, b, a}
  return resultColor
end

--- Make a color gradient.
-- If colors are missing opacity, 1 is assumed.
-- @param startColor (array) The starting gradient color.
-- @param endColor (array) The ending gradient color.
-- @param colorCount (number) The number of colors to create.
-- @return (array of color arrays) The gradient.
function ColorUtils.makeGradient(startColor, endColor, colorCount)
  startColor = ColorUtils.colorWithDefaultOpacity(startColor)
  endColor = ColorUtils.colorWithDefaultOpacity(endColor)

  local startValues = {}
  local ranges = {}
  for i=1,4 do
    startValues[i] = startColor[i]
    ranges[i] = endColor[i] - startColor[i]
  end

  local gradient = {}
  for iGradient=1,colorCount do
    color = {}
    local proportion = (iGradient-1) / (colorCount-1)
    for iColor=1,4 do
      color[iColor] = startValues[iColor] + ranges[iColor] * proportion
    end
    table.insert(gradient, color)
  end
  return gradient
end

--- Adds the same amount to the red, green, and blue channels of a color to produce
-- a ligher (or, if the amount is negative, darker) color.  If the resulting value of
-- any of the channels falls outside of the range [0.0, 1.0] it will be cropped.  Alpha
-- will not be affected.
-- @param color - The color to lighten
-- @param amountToLighten - The amount to lighten (or if negative, darken) each channel
-- @return - the lightened (or darkened) color
function ColorUtils.lightenColor(color, amountToLighten)
  local colorChannelCount = 3

  local lightenedColor = {}

  for channelIndex = 1, colorChannelCount do
    local channelValue = color[channelIndex]
    local lightenedChannelValue = channelValue + amountToLighten

    -- Crop each channel value if it is out of the range [0.0, 1.0]
    lightenedChannelValue = MathUtils.clamp(lightenedChannelValue, 0.0, 1.0)

    table.insert(lightenedColor, lightenedChannelValue)
  end

  return lightenedColor
end
