class("CurveUtils")

local BEZIER_CURVE_CACHE = {}

function CurveUtils.cardinalSpline1D(a, b, c, d, tension, t)
  -- Formula: s(-ttt + 2tt - t)P1 + s(-ttt + tt)P2 + (2ttt - 3tt + 1)P2 + s(ttt - 2tt + t)P3 + (-2ttt + 3tt)P3 + s(ttt - tt)P4
  local t2 = t * t
  local t3 = t2 * t
  local s = (1 - tension)/2

  local b1 = s * ((-t3 + (2 * t2)) - t)
    local b2 = s * (-t3 + t2) + (2 * t3 - 3 * t2 + 1)
    local b3 = s * (t3 - 2 * t2 + t) + (-2 * t3 + 3 * t2)
    local b4 = s * (t3 - t2)

    return a * b1 + b * b2 + c * b3 + d * b4
end

function CurveUtils.cardinalSpline2D(p0, p1, p2, p3, tension, t)
  -- Formula: s(-ttt + 2tt - t)P1 + s(-ttt + tt)P2 + (2ttt - 3tt + 1)P2 + s(ttt - 2tt + t)P3 + (-2ttt + 3tt)P3 + s(ttt - tt)P4
  local result = {}
  local t2 = t * t
  local t3 = t2 * t
  local s = (1 - tension)/2

  local b1 = s * ((-t3 + (2 * t2)) - t)
  local b2 = s * (-t3 + t2) + (2 * t3 - 3 * t2 + 1)
    local b3 = s * (t3 - 2 * t2 + t) + (-2 * t3 + 3 * t2)
    local b4 = s * (t3 - t2)

    result[1] = p0[1] * b1 + p1[1] * b2 + p2[1] * b3 + p3[1] * b4
    result[2] = p0[2] * b1 + p1[2] * b2 + p2[2] * b3 + p3[2] * b4

    return result
end



function CurveUtils.getPointIndexByTime(points)
  local xMin, xMax = math.huge, -math.huge
  for index, point in ipairs(points) do
    if (point[1] < xMin) then xMin = point[1] end
    if (point[1] > xMax) then xMax = point[2] end
  end
  local xAmp = xMax - xMin
  local tIndices = {xMin = xMin, xMax = xMax, xAmp = xAmp}
  local prevTimeData = nil
  for index, point in ipairs(points) do
    local t = (point[1] - xMin)/xAmp
    local timeData = {t = t, pointIndex = index}
    tIndices[#tIndices + 1] = timeData
    if (prevTimeData) then
      prevTimeData.duration = t - prevTimeData.t
    end
    prevTimeData = timeData
  end
  table.sort(tIndices, function(v1, v2) return v1.t < v2.t end)
  return tIndices
end



local function getPointIndexForTime(points, time, pointIndexByTime)
  for i = 1, #pointIndexByTime - 1 do
    local timeData = pointIndexByTime[i]
    local nextTimeData = pointIndexByTime[i + 1]
    if (timeData.t <= time and nextTimeData.t > time) then
      return i
    end
  end
  return #pointIndexByTime - 1
end
local function getNextCardinalSplineIndex(index, points)
  local nextPointIndex = index + 1
  if (nextPointIndex > #points) then nextPointIndex = #points end
  return nextPointIndex
end

function CurveUtils.cardinalSplineForPointsAtTime(points, time, tension, pointIndexByTime)
  pointIndexByTime = pointIndexByTime or CurveUtils.getPointIndexByTime(points)
  local index = getPointIndexForTime(points, time, pointIndexByTime)
  local timeData = pointIndexByTime[index]

  local p1 = points[index]
  local p0
  if (index >= 2) then
    p0 = points[index - 1]
  else
    p0 = points[index]
  end
  local p2 = points[index + 1]

  local p3
  if (index <= #points - 2) then
    p3 = points[index + 2]
  else
    p3 = p2
  end

  local curveTime = (time - timeData.t)/(timeData.duration or 1)
  local y = CurveUtils.cardinalSpline1D(p0[2], p1[2], p2[2], p3[2], tension, curveTime)

  return y, pointIndexByTime
end


local function getFlatPointIndexByTime(points)
  local xMin, xMax = math.huge, -math.huge
  for i = 1, #points, 2 do
    if (points[i] < xMin) then xMin = points[i] end
    if (points[i + 1] > xMax) then xMax = points[i + 1] end
  end
  local xAmp = xMax - xMin
  local tIndices = {xMin = xMin, xMax = xMax, xAmp = xAmp}
  for i = 1, #points, 2 do
    local t = (points[i] - xMin)/xAmp
    tIndices[#tIndices + 1] = {t = t, pointIndex = i}
  end
  table.sort(tIndices, function(v1, v2) return v1.t < v2.t end)
  return tIndices
end

function CurveUtils.getCardinalSplinePoints(points, subdivide, tension)
  local values = {}
  for index = 1, #points - 1 do
    local i0, i1, i2, i3
    local p1 = points[index]
    local p0
    i1 = index
    if (index >= 2) then
      p0 = points[index - 1]
      i0 = index - 1
    else
      p0 = points[index]
      i0 = index
    end
    local p2 = points[index + 1]
    i2 = index + 1

    local p3
    if (index <= #points - 2) then
      p3 = points[index + 2]
      i3 = index + 2
    else
      p3 = p2
      i3 = i2
    end

    for t = 0, subdivide - 1 do
      local time = t/subdivide
      local x = CurveUtils.cardinalSpline1D(p0[1], p1[1], p2[1], p3[1], tension, time)
      local y = CurveUtils.cardinalSpline1D(p0[2], p1[2], p2[2], p3[2], tension, time)
      values[#values + 1] = x
      values[#values + 1] = y
    end
  end

  return values
end


function CurveUtils.valueAtTime(points, time, pointInfo)
  pointInfo = pointInfo or getFlatPointIndexByTime(points)
  local index = getPointIndexForTime(points, time, pointInfo)
  local value = points[index * 2]

  return value, pointInfo
end

local function getBezierCurveValues(controlPoints)
  local startControlPoint = controlPoints[1]
  local endControlPoint = controlPoints[2]

  local phi = 0.0
  local deltaPhi = 0.01
  local points = {}
  repeat
    local point = {
      x = MathUtils.bezier(0, startControlPoint.x, endControlPoint.x, 1, phi),
      y = MathUtils.bezier(0, startControlPoint.y, endControlPoint.y, 1, phi)
    }
    table.insert(points, point)
    phi = phi + deltaPhi
  until (point.y >= 1.0)

  local index = 1
  local left = points[index]
  local right = points[index + 1]
  local stepCount = 128
  local dt = 1.0 / stepCount
  local t = left.x
  local values = {}
  while (right ~= nil) do
    while (t < right.x) do
      local y = MathUtils.interpolateLinear(t, left.x, right.x, left.y, right.y)
      table.insert(values, y)
      t = t + dt
    end
    index = index + 1
    left = right
    right = points[index]
  end

  return values
end

local function getKey(controlPoints)
  return
    tostring(controlPoints[1].x) ..
    tostring(controlPoints[1].y) ..
    tostring(controlPoints[2].x) ..
    tostring(controlPoints[2].y)
end

local function getCachedBezierCurveValues(controlPoints)
  local key = getKey(controlPoints)
  return BEZIER_CURVE_CACHE[key]
end

local function addCachedBezierCurveValues(controlPoints, points)
  local key = getKey(controlPoints)
  BEZIER_CURVE_CACHE[key] = points
end

function CurveUtils.getBezierEaseFunction(controlPoints)
  local values = getCachedBezierCurveValues(controlPoints)
  if (values == nil) then
    values = getBezierCurveValues(controlPoints)
    addCachedBezierCurveValues(controlPoints, values)
  end

  local easeFunction = function(t)
    local valuesCount = #values
    local index = MathUtils.clamp(math.ceil(t * valuesCount), 1, valuesCount)
    return values[index]
  end
  return easeFunction
end
