--- FileUtils
-- Loading your files with confidence since 1893.

-- ===========================================================================
class("FileUtils")
-- ===========================================================================

-- --------------------------------------------------------------------------------
-- Configuration
-- --------------------------------------------------------------------------------

local ASSET_PATH_LIST_FROM_HOST = asset.path
local HOST_ASSET_PATH_SEPARATOR = ";"
local ASSET_PATHS = StringUtils.split(ASSET_PATH_LIST_FROM_HOST, HOST_ASSET_PATH_SEPARATOR)

local ZIPPED_EXTENSION = ".gz"
local IMAGE_EXTENSIONS = {
  ".png",
  ".webp",
  ".pvr.gz"
}
local IMAGE_EXTENSIONS_SET = TableUtils.createSetFromArray(IMAGE_EXTENSIONS)

-- --------------------------------------------------------------------------------
-- Utility
-- --------------------------------------------------------------------------------


--- For paths that end with a zipped extension (`.gz`), this method removes the extension
-- from the path and checks whether there is a file at that path.  It returns the path
-- with the zipped extension removed in the case there was a file at that path.
--
-- It is necessary to search for zipped files at their unzipped paths because Android
-- automatically uncompresses files before adding them to the app bundle.
local function findFileWithZippedExtensionRemoved(absolutePath)
  local isFileZipped = StringUtils.endsWith(absolutePath, ZIPPED_EXTENSION)
  if (isFileZipped) then
    local pathWithZippedExtensionRemoved = FileUtils.splitPathAtLastExtension(absolutePath)
    if (MOAIFileSystem.checkFileExists(pathWithZippedExtensionRemoved)) then
      return pathWithZippedExtensionRemoved
    end
  end

  return nil
end

--- Returns true if the extension is a member of IMAGE_EXTENSIONS_SET, false otherwise.
local function isImageExtension(extension)
  if (IMAGE_EXTENSIONS_SET[extension]) then
    return true
  else
    return false
  end
end

--- Searches for files that end with an image extension by checking to see if a file with that
-- name exists with all of the other supported image extensions.  For example, if `absolutePath`
-- contains `/path/to/foo.png`, this method will search `/path/to/foo.webp` and
-- `/path/to/foo.pvr.gz` to see whether an image exists at either of those locations.
--
-- Returns the path of the first image file it finds if one exists, otherwise returns `nil`.
local function findFileWithAlternateImageExtension(absolutePath)
  local filePath, fileName, fileExtension = FileUtils.splitPathIntoComponents(absolutePath)
  if (isImageExtension(fileExtension)) then
    local imageExtensionsSetCopy = TableUtils.shallowCopy(IMAGE_EXTENSIONS_SET)
    imageExtensionsSetCopy[fileExtension] = nil
    local pathWithExtensionRemoved = filePath .. fileName
    for imageExtension, _ in pairs(imageExtensionsSetCopy) do
      local pathWithOtherImageExtension = pathWithExtensionRemoved .. imageExtension
      if (MOAIFileSystem.checkFileExists(pathWithOtherImageExtension)) then
        return pathWithOtherImageExtension
      end
    end
  end

  return nil
end

local function findFileWithAbsolutePath(absolutePath)
  if (MOAIFileSystem.checkFileExists(absolutePath)) then
    return absolutePath
  end

  local pathWithZippedExtensionRemoved = findFileWithZippedExtensionRemoved(absolutePath)
  if (pathWithZippedExtensionRemoved ~= nil) then
    return pathWithZippedExtensionRemoved
  end

  return findFileWithAlternateImageExtension(absolutePath)
end

local function findFileWithRelativePath(relativePath)
  for _, assetPath in ipairs(ASSET_PATHS) do
    local absolutePath = assetPath .. "/" .. relativePath
    local foundFile = findFileWithAbsolutePath(absolutePath)
    if (foundFile ~= nil) then
      return foundFile
    end
  end

  return nil
end


-- --------------------------------------------------------------------------------
-- API
-- --------------------------------------------------------------------------------

--- Searches the filesystem for a file with the given absolute or relative path and
-- returns the absolute path where the file was found.  If the given path is relative,
-- searches all known asset paths to find the absolute path.  Android automatically
-- unzips all files before adding them to the app bundle.  Therefore, if a zipped
-- extension is detected on the given path, this method searches for the file both
-- with and without the zipped extension.
-- @param path (string) - the path (relative or absolute) to search for.
-- @return (string) - the absolute path where the file was found.
function FileUtils.findFile(path)
  Asserts.assertType(path, "path", "string")

  local isPathAbsolute = StringUtils.startsWith(path, "/")
  if (isPathAbsolute) then
    return findFileWithAbsolutePath(path)
  end

  return findFileWithRelativePath(path)
end

--- Split a filename into the path part, the name part, and the extension part.
-- For example, if the path is: "/Users/aab29/ms/com.mindsnacks.whatever/thing@7x.pvr.gz"
-- then the result will be:
-- filePath: "/Users/aab29/ms/com.mindsnacks.whatever/"
-- filename: "thing@7x"
-- extension: ".pvr.gz"
-- Note that extension *does* contain the '.'.
-- @param path (string) - the path or filename to split.
-- @return filePath (string) - Everything before (and including) the last '/' character in the path.
-- @return filename (string) - Everything after the last '/' character in the path and before the first '.' in the filename part.
-- @return extension (string) - Everything after (and including) the first '.' character in the filename part.
function FileUtils.splitPathIntoComponents(path)
  Asserts.assertType(path, "path", "string")
  local filePath, filename, extension = string.match(path, "(.-)([^/]-)([%.][^/]+)$")
  return filePath, filename, extension
end

--- Split a path into the part before the last '.' character and the part after
-- (and including) the last '.' character with no slashes following it.
-- For example, if the path is: "/Users/aab29/ms/com.mindsnacks.whatever/thing@7x.pvr.gz"
-- then the result will be:
-- pathWithLastExtensionRemoved: "/Users/aab29/ms/com.mindsnacks.whatever/thing@7x.pvr"
-- lastExtension: ".gz"
-- @param path (string) - the path or filename to split
-- @return pathWithLastExtension (string) - Everything before (and excluding) the last '.' character.
-- @return lastExtension (string) - Everything after (and including) the last '.' character.
function FileUtils.splitPathAtLastExtension(path)
  Asserts.assertType(path, "path", "string")
  local pathWithLastExtensionRemoved, lastExtension = string.match(path, "(.*)([%.][^/]+)$")
  return pathWithLastExtensionRemoved, lastExtension
end

--- Replaces the last file extension (including '.' character) with the provided replacement string.
-- @param path (string) - the path on which to perform the extension substitution
-- @param replacementExtension (string) - the new extension
-- @return pathAfterReplacement (string) - The original path with its extention replaced.
function FileUtils.replaceLastExtension(path, replacementExtension)
  Asserts.assertType(replacementExtension, "replacementExtension", "string")
  local pathWithLastExtensionRemoved, lastExtension = FileUtils.splitPathAtLastExtension(path)
  local pathAfterReplacement = pathWithLastExtensionRemoved .. replacementExtension
  return pathAfterReplacement
end

--- Add a string suffix before the first '.' character in a filename.
-- @param path (string) - the file path or filename to which to add the suffix.
-- @param suffix (string) - the suffix to add.
-- @return (string) - a new string made up of the path with the suffix added.
function FileUtils.addSuffixBeforeExtension(path, suffix)
  Asserts.assertType(suffix, "suffix", "string")
  local filePath, filename, extension = FileUtils.splitPathIntoComponents(path)
  local pathWithExtension = filePath .. filename .. suffix .. extension
  return pathWithExtension
end

-- --------------------------------------------------------------------------------
-- JSON processing
-- --------------------------------------------------------------------------------

function FileUtils.loadString(path)
  Asserts.assertType(path, "path", "string")
  local file, errorMessage = io.open(path, "r")
  assert(not errorMessage, "Loading '" .. tostring(path) .. "' failed with error: " .. tostring(errorMessage))

  local contents = nil
  contents, errorMessage = file:read("*a")
  assert(not errorMessage, "Loading '" .. tostring(path) .. "' failed with error: " .. tostring(errorMessage))
  io.close(file)

  return contents
end

--- Loads the specified file as a json table
-- @param filePath - (string) full path to the file to load, must contain a json string
-- @return (table) json contained in the specified file as a lua table
function FileUtils.loadJson(filePath)
  Asserts.assertType(filePath, "filePath", "string")
  local jsonString = FileUtils.loadString(filePath)
  return MOAIJsonParser.decode(jsonString)
end
