--- FunctionUtils
-- @author Created by karl on January 20, 2015.

class("FunctionUtils")


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

--- Forward a method call from sourceObject to destinationObject.
-- @param sourceObject (table or userdata) - The object to forward calls from.
-- @param destinationObject (table or userdata) - The object to forward calls to.
-- @param methodName (string) - The name of the method to forward.
local function forwardMethod(sourceObject, destinationObject, methodName)
  assert(destinationObject[methodName], "Method '" .. methodName .. "' not found in destinationObject")
  Asserts.assertType(destinationObject[methodName], "method", "function")
  sourceObject[methodName] = function(self, ...)
    return destinationObject[methodName](destinationObject, ...)
  end
end

--- Forward many methods
-- @param sourceObject (table or userdata) - The object to forward calls from.
-- @param destinationObject (table or userdata) - The object to forward calls to.
-- @param methods (table) - Table with keys containing the names of functions to forward.
-- @param excludedSet (set) - names of functions that must NOT be forwarded (optional).
local function forwardMethods(sourceObject, destinationObject, methods, excludedSet)
  excludedSet = excludedSet or {}
  Asserts.assertType(sourceObject, "sourceObject", "table", "userdata")
  Asserts.assertType(destinationObject, "destinationObject", "table", "userdata")
  Asserts.assertType(methods, "methods", "table")
  Asserts.assertType(excludedSet, "excludedSet", "table")
  for methodName, _ in pairs(FunctionUtils.allFunctionsInObject(methods)) do
    if not excludedSet[methodName] then
      forwardMethod(sourceObject, destinationObject, methodName)
    end
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Forward method calls from sourceObject to destinationObject.
-- @param sourceObject (table or userdata) - The object to forward calls from.
-- @param destinationObject (table or userdata) - The object to forward calls to.
-- @param ... (list) - names of functions to forward.
function FunctionUtils.forwardMethodCalls(sourceObject, destinationObject, ...)
  Asserts.assertType(sourceObject, "sourceObject", "table", "userdata")
  Asserts.assertType(destinationObject, "destinationObject", "table", "userdata")
  for _, methodName in ipairs(arg) do
    forwardMethod(sourceObject, destinationObject, methodName)
  end
end

--- Forward method calls from a parent to its children.
-- @param parent (Group or MOAIProp) - The object to forward calls from.
-- @param isRecursive(boolean) - If true, the method call will be applied recursively to the entire child tree.
-- @param ... (list) - names of functions to forward.
function FunctionUtils.forwardMethodCallsToChildren(parent, isRecursive, ...)
  Asserts.assertType(parent, "parent", "table", "userdata")
  local function newForwardingMethod(methodName)
    return function(self, ...)
      for _, child in ipairs(self.children) do
        local childMethod = child[methodName]
        if type(childMethod) == "function" then
          childMethod(child, ...)
        end
      end
    end
  end

  for _, methodName in ipairs({...}) do
    local forwardingMethod = newForwardingMethod(methodName)
    if isRecursive then
      parent[methodName] = function(self, ...) FunctionUtils.callFunctionRecursive(forwardingMethod, parent, ...) end
    else
      parent[methodName] = forwardingMethod
    end
  end
end

--- Forward all functions found in destinationObject via sourceObject.
-- i.e. for every function found in destinationObject, set up a forwarding call from sourceObject.
-- Note: destinationObject must be a table-based object, not userdata.
-- @param sourceObject (table or userdata) - The table whose functions calls to forward.
-- @param destinationObject (table) - The table to forward the calls to.
-- @param excludedSet (set) - names of functions that must NOT be forwarded (optional).
function FunctionUtils.forwardAllMethodCalls(sourceObject, destinationObject, excludedSet)
  forwardMethods(sourceObject, destinationObject, destinationObject, excludedSet)
end

--- Forward all functions specifed in class from sourceObject to destinationObject.
-- Note: class must be a table-based object, not userdata.
-- @param sourceObject (table or userdata) - The object whose functions calls to forward.
-- @param destinationObject (table or userdata) - The object to forward the calls to.
-- @param class (table) - The class containing the functions to forward.
-- @param excludedSet (set) - names of functions that must NOT be forwarded (optional).
function FunctionUtils.forwardMethodCallsInClass(sourceObject, destinationObject, class, excludedSet)
  forwardMethods(sourceObject, destinationObject, class, excludedSet)
end

--- Get all functions from an object.
-- Note: Only works on table-based objects, not userdata.
-- @param object (table) - The object to get functions from.
-- @return a table of named functions.
function FunctionUtils.allFunctionsInObject(object)
  Asserts.assertType(object, "object", "table")
  local functions = {}
  for methodName, method in pairs(object) do
    if type(method) == "function" then
      functions[methodName] = method
    end
  end
  return functions
end

--- Recursively call a method on a target and all of its descendents.
-- This differs from callFunctionRecursive in that it looks up each child's implementation of the method.
-- @param methodName (string) - Name of the method to call.
-- @param target (Group or MOAIProp) - Target to call the method on.
-- @param ... (any) - Any parameters to pass to the method besides the target.
function FunctionUtils.callMethodRecursive(methodName, target, ...)
  local method = target[methodName]
  if method then
    method(target, ...)
  end
  if target.children then
    for _, child in ipairs(target.children) do
      FunctionUtils.callMethodRecursive(methodName, child, ...)
    end
  end
end

--- Recursively call a function on a target and all of its descendents.
-- This differs from callMethodRecursive in that it applies the same function to all descendents.
-- @param aFunction (function) - The function to call.
-- @param target (Group or MOAIProp) - Target to call the method on.
-- @param ... (any) - Any parameters to pass to the function besides the target.
function FunctionUtils.callFunctionRecursive(aFunction, target, ...)
  aFunction(target, ...)
  if target.children then
    for _, child in ipairs(target.children) do
      FunctionUtils.callFunctionRecursive(aFunction, child, ...)
    end
  end
end

--- Call a method (if it exists) on all provided targets
-- @oaram methodname (string) - Name of the method to call.
-- @param targets (array) - An array of targets on which to call the method.
-- @param ... (any) -- any parameters to pass to the method besides the target.
function FunctionUtils.callMethodOnTargets(methodName, targets, ...)
  for _, target in ipairs(targets) do
    local method = target[methodName]
    if (method ~= nil) then
      method(target, ...)
    end
  end
end

--- Override a function call.
-- @param originalFunction (function) - The function to override.
-- @param overridingFunction (function) - The function to override with.
-- @return (function) - An overriding function that can be put in place of the original.
function FunctionUtils.overrideFunction(originalFunction, overridingFunction)
  Asserts.assertType(originalFunction, "originalFunction", "function")
  Asserts.assertType(overridingFunction, "overridingFunction", "function")
  return function(...)
    overridingFunction(originalFunction, ...)
  end
end

--- Override a method in an object.
-- @param object (table) - The object whose method to override.
-- @param methodName (string) - The name of the method to override.
-- @param overridingFunction (function) - The function to override with.
function FunctionUtils.overrideMethod(object, methodName, overridingMethod)
  local originalFunction = object[methodName]
  assert(type(originalFunction) == "function", "Method " .. methodName .. " is not a function. Got " .. type(originalFunction))
  object[methodName] = FunctionUtils.overrideFunction(originalFunction, overridingMethod)
end
