--- GlyphTableHelpers 
-- Helpers for getting info from glyph tables
--  @author Created by Jay Austria on Octomom 30, 2013

-- ===========================================================================
class("GlyphTableHelpers")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Safer/Convenient Table Access ------

--- Gets a glyph in a glyph row
-- @param glyphRow    - (table) The glyph row to access
-- @param glyphIndex  - (number) Index of the glyph to access from the row
-- @return (table) the glyph entree in the row
function GlyphTableHelpers.getGlyphInRowByIndex(glyphRow, glyphIndex)
  assert(glyphRow, "Invalid glyph row")

  local glyph = nil
  if (glyphIndex > 0 and glyphIndex <= #glyphRow) then
    glyph = glyphRow[glyphIndex]
  else
    error("glyphIndex out of bounds")
  end

  return glyph
end

--- Gets a glyph row in a glyph table
-- @param glyphTable - (table) The glyph table to access
-- @param rowIndex   - (number) Index of the glyph row to access from the table
-- @return (table) the glyph row entree in the glyph table
function GlyphTableHelpers.getGlyphRowByIndex(glyphTable, rowIndex)
  assert(glyphTable, "Invlalid glyph table")

  local glyphRow = nil
  if (rowIndex > 0 and rowIndex <= #glyphTable) then
    glyphRow = glyphTable[rowIndex]
  else
    error("rowIndex out of bounds")
  end
  return glyphTable[rowIndex]
end


-- String Index to Table Conversions --

--- Fetch a specific glyph by the source string index
-- @param glyphTable   - (table) The glyph table generated from rendering the
--                       source string to a label
-- @param stringIndex  - (number) Index of the character in the source string
--                       that we are wanting the corresponding glyph info of
-- @return (table) the glyph entree corresponding with the string index
-- @return (table) the row that contained the corresponding glyph entree
function GlyphTableHelpers.getGlyphAndGlyphRowByStringIndex(glyphTable, stringIndex)
  assert(glyphTable, "Invalid glyph table")
  assert(stringIndex > 0, "string index provided was zero")

  local glyph        = nil
  local rowIndex     = 1
  local glyphIndex   = stringIndex
  local numberOfRows = #glyphTable
  local currentRow   = GlyphTableHelpers.getGlyphRowByIndex(glyphTable, rowIndex)
  local rowLength    = #currentRow
  -- Shorten the glyph Index until it accesses the current row
  while (glyphIndex > rowLength and rowIndex <= numberOfRows) do
    glyphIndex = glyphIndex - rowLength
    glyphIndex = glyphIndex  - 1 -- Compensate for missing whitespace during render
    rowIndex   = rowIndex + 1
    currentRow = GlyphTableHelpers.getGlyphRowByIndex(glyphTable, rowIndex)
    rowLength  = #currentRow
  end

  assert(glyphIndex > 0, "Index fell on a whitespace that exceeded the glyphs in a row")
  assert(rowIndex <= numberOfRows, "Index exceeded number of glyphs in the glyph table")

  if (glyphIndex > 0 and rowIndex <= numberOfRows) then
    glyph = GlyphTableHelpers.getGlyphInRowByIndex(currentRow, glyphIndex)
  end

  return glyph, currentRow
end

--- Opposite of above. If you have a glyph, this function will help you determine
-- what the corresopnding character in the source string's index was
-- @param glyphTable   - (table) The glyph table generated from rendering the
--                       source string to a label
-- @param glyph        - (table) Glyph table item whose corresponding string
--                       index we are wanting to find
-- @return (number) the determined string index of the glyph entree
-- @return (table)  the row that contained the corresponding glyph entree
function GlyphTableHelpers.getStringIndexAndRowByGlyphTableAndGlyph(glyphTable, glyph)
  assert(glyphTable, "Invalid glyph table")
  assert(glyph, "Invalid glyph")

  local stringIndex = 1

  -- Iterate through all the rows until the glyph is found
  for _, row in pairs(glyphTable) do
    for _, glyphCandidate in pairs(row) do
      if (type(glyphCandidate) == "table") then
        if (glyphCandidate == glyph) then
          return stringIndex, row
        end
        stringIndex = stringIndex + 1
      end
    end
    stringIndex = stringIndex + 1 -- Compensate for missing whitespace during render
  end

  error("Glyph was not found in row")
  return nil, nil
end

-- Convenience Functions For Fetching Rows --

--- Gets the row that a glyph resides in
-- @param glyphTable - (table) The glyph table generated from rendering the
--                       source string to a label
-- @param glyph      - (table) Glyph entree that resides in the row we are
--                     looking for
-- @return (table) the glyph row that contains the glyph
function GlyphTableHelpers.getRowContainingGlyph(glyphTable, glyph)
  assert(glyphTable, "Invalid glyph table")
  assert(glyph, "Invalid glyph")

  local _, row = GlyphTableHelpers.getStringIndexAndRowByGlyphTableAndGlyph(glyphTable, glyph)
  return row
end

--- Gets the row that a string indexes' corresponding glyph resides in
-- @param glyphTable   - (table) The glyph table generated from rendering the
--                       source string to a label
-- @param stringIndex  - (number) Index of the character in the source string
--                       that we are wanting the corresponding row of
-- @return (table) the glyph row that contains the corresponding glyph from the
--                 requested index
function GlyphTableHelpers.getRowContainingStringIndex(glyphTable, stringIndex)
  assert(glyphTable, "invalid glyph table")
  assert(stringIndex > 0, "string index provided was zero")

  local _, row = GlyphTableHelpers.getGlyphAndGlyphRowByStringIndex(glyphTable, stringIndex)
  return row
end

-- Baseline Info ----------------------

--- Reads the baseline info from a row
-- @param glyphRow - (table) Glyph row we are wanting the baselineY of
-- @return (number) The baseline y for the row
function GlyphTableHelpers.getBaselineFromRow(glyphRow)
  assert(glyphRow, "Invalid glyph row")

  return glyphRow.baselineY
end

--- Reads the baseline info from the row containing the glyph
-- @param glyphTable - (table) Glyph table generated from rendering the source
--                     string to a label
-- @param glyph      - the glyph we are wanting to get baseline info for
-- @return (number) The baseline y for the row containing the glyph
function GlyphTableHelpers.getBaselineFromGlyphTableByGlyph(glyphTable, glyph)
  assert(glyphTable, "Invalid glyph table")
  assert(glyph, "Invalid glyph")

  local baselineY = nil
  local row = GlyphTableHelpers.getRowContainingGlyph(glyphTable, glyph)
  if (row) then
    baselineY = GlyphTableHelpers.getBaselineFromRow(row)
  end
  return baselineY
end

--- Reads the baseline info from the row containing the glyph corresponding
-- with a string index
-- @param glyphTable  - (table) Glyph table generated from rendering the source
--                      string to a label
-- @param stringIndex - the index of the source string corresponding with the
--                      glyph we are trying to find baseline information for
-- @return (number) The baseline y for the row containing the corresponding
--                  glyph
function GlyphTableHelpers.getBaselineFromGlyphTableByStringIndex(glyphTable, stringIndex)
  assert(glyphTable, "invalid glyph table")
  assert(stringIndex > 0, "string index provided was zero")

  local baselineY = nil
  local row = GlyphTableHelpers.getRowContainingStringIndex(glyphTable, stringIndex)
  if (row) then
    baselineY = GlyphTableHelpers.getBaselineFromRow(row)
  end
  return baselineY
end

-- Coordinate Info --------------------

--- Reads the coordinate information from a glyph
-- @param glyph - (table) Glyph we are wanting the coordinate information from
-- @return (number) x location of the glyph
-- @return (number) y location of the glyph
function GlyphTableHelpers.getGlyphCoordinatesByGlyph(glyph)
  assert(glyph, "Invalid glyph")

  local x, y = nil, nil
  if (glyph) then
    x = (glyph.xMin + glyph.xMax) * 0.5
    y = (glyph.yMin + glyph.yMax) * 0.5
  end
  return x, y
end

--- Reads the coordinate information from a glyph corresponding with a string
--  index
-- @param glyphTable  - (table) Glyph table generated from rendering the source
--                      string to a label
-- @param stringIndex - (number) Index in the string corresponding to the glyph
--                      we are wanting to get coordinate information from
-- @return (number) x location of the glyph
-- @return (number) y location of the glyph
function GlyphTableHelpers.getGlyphCoordinatesFromGlyphTableByStringIndex(glyphTable, stringIndex)
  assert(glyphTable, "Invalid glyph table")
  assert(stringIndex > 0, "String index was zero")

  local glyph = GlyphTableHelpers.getGlyphAndGlyphRowByStringIndex(glyphTable, stringIndex)
  local x, y  = GlyphTableHelpers.getGlyphCoordinatesByGlyph(glyph)
  return x, y
end

-- Dimensions Info --------------------

--- Calculates the distance between the minimum and maximum glyph x coordinates
--  in the glyphTable.
-- @param glyphTable  - (table) Glyph table generated from rendering the source
--                      string to a label
-- @return (number) width of the rendered label
function GlyphTableHelpers.getMaxWidth(glyphTable)
  assert(glyphTable, "Invalid glyph table")

  local xMin = math.huge
  local xMax = 0

  for _, row in ipairs(glyphTable) do
    local rowXMin = row[1].xMin
    local rowXMax = row[#row].xMax
    xMin = math.min(rowXMin, xMin)
    xMax = math.max(rowXMax, xMax)
  end

  local width = xMax - xMin
  return width
end

--- Calculates the distance between the minimum and maximum glyph y coordinates
--  in the glyphTable.
-- @param glyphTable  - (table) Glyph table generated from rendering the source
--                      string to a label
-- @return (number) height of the rendered label
function GlyphTableHelpers.getMaxHeight(glyphTable)
  assert(glyphTable, "Invalid glyph table")

  local yMin = math.huge
  local yMax = 0

  local firstRow = glyphTable[1]
  for _, glyph in ipairs(firstRow) do
    local rowYMin = glyph.yMin
    yMin = math.min(rowYMin, yMin)
  end

  local lastRow = glyphTable[#glyphTable]
  for _, glyph in ipairs(lastRow) do
    local rowYMax = glyph.yMax
    yMax = math.max(rowYMax, yMax)
  end

  local height = yMax - yMin
  return height
end
