--- LevenshteinDistance
-- @author Created by Alexey Chernikov on Jul 24, 2014.

class("LevenshteinDistance")

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance
-- Note, this uses the Optimal String Alignment distance algorithm described in the article
-- It produces slightly different results than the true Damerau Levenshtein distance, namely
-- LD("CA", "ABC") would be 2
-- OSA("CA", "ABC") yields 3
function LevenshteinDistance.calculateDamerauLevenshtein(word1, word2)
  Asserts.assertType(word1, "word1", "string")
  Asserts.assertType(word2, "word2", "string")

  local word1Length = #word1
  local word2Length = #word2

  -- OPTIMIZATION: Do comparisons with a character array to avoid having to constantly
  -- access characters through string.sub()
  local word1CharacterArray = TableUtils.createArrayFromString(word1)
  local word2CharacterArray = TableUtils.createArrayFromString(word2)

  -- OPTIMIZATION: 2D array as a 1D array
  -- costs[i][j] -> costs[i * columnCount + j]
  local costs = {}
  local columnCount = word2Length + 1

  -- OPTIMIZATION: Localize for performance
  local min = math.min

  -- initialize costs of deletion
  for i = 0, word1Length do
    costs[i * columnCount] = i
  end

  -- initialize costs of insertion
  for j = 0, word2Length do
    costs[j] = j
  end

  for i = 1, word1Length do
    local iOffset = i * columnCount
    for j = 1, word2Length do
      local addCost = 0
      if (word1CharacterArray[i] ~= word2CharacterArray[j]) then
        addCost = 1
      end

      local deletionCost = costs[iOffset - columnCount + j] + 1
      local insertionCost = costs[iOffset + j - 1] + 1
      local substitutionCost = costs[iOffset - columnCount + j - 1] + addCost

      local lowestCost = min(deletionCost, insertionCost, substitutionCost)
      costs[iOffset + j] = lowestCost

      local isEligibleForTransposition = ((i > 1) and
                                          (j > 1) and
                                          (word1CharacterArray[i] == word2CharacterArray[j - 1]) and
                                          (word1CharacterArray[i - 1] == word2CharacterArray[j]))
      if (isEligibleForTransposition) then
        local transpositionCost = costs[iOffset - 2 * columnCount + j - 2] + addCost
        costs[iOffset + j] = min(lowestCost, transpositionCost)
      end
    end
  end

  local levenshteinDistance = costs[#costs]

  return levenshteinDistance
end
