--- Miscellaneous math related functions


-- ===========================================================================
class("MathUtils")
-- ===========================================================================

-- Round a number to nearest decimal places
-- Ex. MathUtils.round(1.234, 1) => 1.2
--     MathUtils.round(1.234, 2) => 1.23
function MathUtils.round(number, decimalPlaceIndex)
  local mult = 10^(decimalPlaceIndex or 0)
  return math.floor(number * mult + 0.5) / mult
end

-- Round a number to the nearest multiple of another number
-- Ex. MathUtils.roundToNearestMultiple(224, 50) => 200
--     MathUtils.roundToNearestMultiple(226, 50) => 250
function MathUtils.roundToNearestMultiple(number, multiple)
  return MathUtils.round(number / multiple) * multiple
end

--- Generates a random number sign
-- @return -1 or 1
function MathUtils.randomSign()
  return math.random(1, 2) * 2 - 3
end

function MathUtils.sign(x)
  if (x < 0) then
    return -1
  elseif (x > 0) then
    return 1
  else
    return 0
  end
end

function MathUtils.distanceSquared(x0, y0, x1, y1)
  local dx = x1 - x0
  local dy = y1 - y0
  return dx * dx + dy * dy
end

function MathUtils.distance(x0, y0, x1, y1)
  local distanceSquared = MathUtils.distanceSquared(x0, y0, x1, y1)
  return math.sqrt(distanceSquared)
end

function MathUtils.vectorLength(x, y)
  return math.sqrt(x * x + y * y)
end

function MathUtils.dotProduct(x1, y1, x2, y2)
  return x2 * x1 + y2 * y1
end

function MathUtils.angleBetweenVectors(x0, y0, x1, y1, x2, y2)
  local v1x, v1y = x1 - x0, y1 - y0
  local v2x, v2y = x2 - x0, y2 - y0
  local v1Length, v2Length = MathUtils.vectorLength(v1x, v1y), MathUtils.vectorLength(v2x, v2y)
  local dotProduct = MathUtils.dotProduct(v1x, v1y, v2x, v2y)
  local cosAngle = dotProduct/(v1Length * v2Length)

  local angle = math.acos(cosAngle)
  local directionDeterminant = v1x * v2y - v2x * v1y
  if (directionDeterminant < 0) then angle = -angle end

  return angle
end

function MathUtils.clamp(value, min, max)
    return math.min(math.max(value, min), max)
end

-- Get a bool with 50% chance of true and 50% chance of false
function MathUtils.getRandomBoolean()
  return (math.random(2) == 1)
end

--- Get a random float in the [min, max) interval (inclusive min, exclusive max).
-- @min Lower bound for the value (inclusive).
-- @max Upper bound for the value (exclusive).
-- @return Random float value between min and max.
function MathUtils.getRandomFloat(min, max)
    return min + math.random() * (max - min);
end

--- Perform a linear mapping of a value from the [ax, bx] interval to the [ay, by] interval.
-- Note that it is acceptable to pass a value out of the [ax, bx] interval, in which case
-- the mapped value will be outside of the [ay, by] interval.
-- @value Value to map, usually belonging to the [ax, bx] interval.
-- @ax Lower bound of the origin interval. Can be less or greater than bx.
-- @bx Upper bound of the origin interval. Can be greater or less than ax.
-- @ay Lower bound of the destination interval. Can be less or greater than by.
-- @by Upper bound of the destination interval. Can be greater or less than by.
-- @return Mapped value, in the [ay, by] interval.
-- Ex. MathUtils.mapLinear(4.2, 0, 10, 0, 100) => 42
--     MathUtils.mapLinear(4.2, 0, 10, 10, 20) => 14.2
--     MathUtils.mapLinear(13, 10, 25, 100, 50) => 90
function MathUtils.mapLinear(value, ax, bx, ay, by)
  assert(ax ~= bx, "ax and bx should be different")
  return ay + ((value - ax) / (bx - ax)) * (by - ay)
end

--- Normalize a value in the [min, max] interval to the [0, 1] interval.
-- @value Value to normalize, belonging to the [min, max] interval.
-- @min Lower bound of the interval. Must be less than max.
-- @max Upper bound of the interval. Must be greater than min.
-- @return Normalized value, in the [0, 1] interval
-- Ex. MathUtils.normalize(42, 0, 100) => 0.42
--     MathUtils.normalize(14.2, 10, 20) => 0.42
--     MathUtils.normalize(95, 50, 100) => 0.9
function MathUtils.normalize(value, min, max)
  assert(value >= min and value <= max, "value must be in the [min, max] interval.")
  assert(min < max, "min must be less than max")
  return MathUtils.mapLinear(value, min, max, 0, 1)
end


--- Perform linear interpolation using the point set {(xMin, yMin), (xMax, yMax)} and evaluate a value.
-- This is equivalent to MathUtils.mapLinear, but restricts the range of value to the [xMin, xMax] interval.
-- @value Value to evaluate in the interpolation, belonging to the [xMin, xMax] interval.
-- @xMin Lower bound of the origin interval. Must be less than xMax.
-- @xMax Upper bound of the origin interval. Must be greater than xMin.
-- @yMin Lower bound of the destination interval. Can be less or greater than yMax.
-- @yMax Upper bound of the destination interval. Can be greater or less than yMin.
-- @return Interpolated value, in the [yMin, yMax] interval
-- Ex. MathUtils.interpolateLinear(4.2, 0, 10, 0, 100) => 42
--     MathUtils.interpolateLinear(4.2, 0, 10, 10, 20) => 14.2
--     MathUtils.interpolateLinear(13, 10, 25, 100, 50) => 90
function MathUtils.interpolateLinear(value, xMin, xMax, yMin, yMax)
  assert(value >= xMin and value <= xMax, "value must be in the [" .. xMin .. ", " .. xMax .. "] interval.")
  assert(xMin < xMax, "xMin must be less than xMax")
  return MathUtils.mapLinear(value, xMin, xMax, yMin, yMax)
end

--- This is equivalent to MathUtils.interpolateLinear, but clamps value between [xMin, xMax] first
-- @value Value to evaluate in the interpolation.
-- @xMin Lower bound of the origin interval. Must be less than xMax.
-- @xMax Upper bound of the origin interval. Must be greater than xMin.
-- @yMin Lower bound of the destination interval. Can be less or greater than yMax.
-- @yMax Upper bound of the destination interval. Can be greater or less than yMin.
-- @return Interpolated value, in the [yMin, yMax] interval
-- Ex. MathUtils.interpolateLinear(4.2, 0, 10, 0, 100) => 42
--     MathUtils.interpolateLinear(14.2, 0, 10, 0, 100) => 100
--     MathUtils.interpolateLinear(-4.2, 0, 10, 0, 100) => 0
function MathUtils.interpolateLinearWithClamp(value, xMin, xMax, yMin, yMax)
  local clampedValue = MathUtils.clamp(value, xMin, xMax)
  return MathUtils.interpolateLinear(clampedValue, xMin, xMax, yMin, yMax)
end

--- Perform linear interpolation using the point set {(0, yMin), (1, yMax)} and evaluate a normalized value.
-- This is equivalent to MathUtils.mapLinear, but restricts the range of value to the [0, 1] interval.
-- @normalizedValue Value to evaluate in the interpolation, belonging to the [0, 1] interval.
-- @yMin Lower bound of the interval. Can be less or greater than yMax.
-- @yMax Upper bound of the interval. Can be greater or less than yMin.
-- @return Interpolated value, in the [yMin, yMax] interval
-- Ex. MathUtils.interpolateLinearFromNormalized(0.42, 0, 100) => 42
--     MathUtils.interpolateLinearFromNormalized(0.42, 10, 20) => 14.2
--     MathUtils.interpolateLinearFromNormalized(0.1, 100, 50) => 95
function MathUtils.interpolateLinearFromNormalized(normalizedValue, yMin, yMax)
  return MathUtils.interpolateLinear(normalizedValue, 0, 1, yMin, yMax)
end

function MathUtils.getLinearPointsOnSegment(startPoint, endPoint, numberOfInbetweenPoints)
  local result = {}
  local x0, y0 = startPoint[1], startPoint[2]
  local x1, y1 = endPoint[1], endPoint[2]
  local dx = x1 - x0
  local dy = y1 - y0

  --- +==|==|==|==+: 3 points inbetween, each segment is 1/4 of the length
  local xStep = dx/(numberOfInbetweenPoints + 1)
  local yStep = dy/(numberOfInbetweenPoints + 1)
  result[1] = {x0, y0}

  for i = 1, numberOfInbetweenPoints do
    local x = x0 + xStep * i
    local y = y0 + yStep * i
    result[#result + 1] = {x, y}
  end

  result[#result + 1] = {x1, y1}

  return result
end

--- Returns true if number is an integer.
-- Note: Only call this method with number types (not string).
function MathUtils.isInteger(number)
  return number % 1 == 0
end

function MathUtils.isEven(numberValue)
  return (numberValue % 2 == 0)
end

function MathUtils.isOdd(numberValue)
  return (not MathUtils.isEven(numberValue))
end

--- Get the greatest common divisor of two numbers.
-- Example: The GCD of 42 and 56 is 14 (3 * 14 and 4 * 14).
function MathUtils.greatestCommonDivisor(valueA, valueB)
  while valueA > 0 do
    local tmp = valueA
    valueA = valueB % valueA
    valueB = tmp
  end
  return valueB
end

--- Get the lowest common multiple of two numbers.
-- Example: The LCM of 4 and 6 is 12 (12/4 = 3, 12/6 = 2).
function MathUtils.lowestCommonMultiple(valueA, valueB)
  local gcd = MathUtils.greatestCommonDivisor(valueA, valueB)
  if gcd == 0 then
    return 0
  end
  return valueA * valueB / gcd
end

--- Reduce a fraction to its lowest terms.
-- @return (number, number) - The numerator and denominator.
function MathUtils.fractionLowestTerms(numerator, denominator)
  local gcd = MathUtils.greatestCommonDivisor(numerator, denominator)
  return numerator / gcd, denominator / gcd
end

--- Increase a fraction to its highest terms, up to maxDenominator.
-- @return (number, number) - The numerator and denominator.
function MathUtils.fractionHighestTerms(numerator, denominator, maxDenominator)
  local multiple = math.floor(maxDenominator / denominator)
  return (numerator * multiple), (denominator * multiple)
end

--- Get the binary logarithm of a number.
function MathUtils.log2(n)
  return math.log(n)/math.log(2)
end

--- Increment a number until it is a power of 2.
-- If the number is already a power of 2, it remains unchanged.
function MathUtils.incrementUntilPowerOf2(number)
  local exponent = math.ceil(MathUtils.log2(number))
  return 2 ^ exponent
end

--- Determine whether value is NaN.
-- @value (number) - Value to determine wether is NaN or not.
-- @return true if value is NaN, false otherwise.
function MathUtils.isNan(value)
  assert(type(value) == "number", "Value must be a number")
  return tostring(value):find("nan") ~= nil
end

--- Determine whether value is infinite (inf or -inf).
-- @value (number) - Value to determine wether is infinite or not.
-- @return true if value is infinite, false otherwise.
function MathUtils.isInf(value)
  assert(type(value) == "number", "Value must be a number")
  return math.abs(value) == math.huge
end

--- Determine whether value is real (not NaN, inf or -inf).
-- @value (number) - Value to determine wether is real or not.
-- @return true if value is real, false otherwise.
function MathUtils.isRealNumber(value)
  return not MathUtils.isNan(value) and not MathUtils.isInf(value)
end

--- Solve the quadratic equation 'a * x^2 + b * x + c = 0'.
-- @a Coeficient of the second degree term
-- @b Coeficient of the first degree term
-- @c Coeficient of the independent term
-- @return (table) - Array of 0, 1 or 2 real solutions.
function MathUtils.getQuadraticEquationSolutions(a, b, c)
  local solutions = {}
  local squareRootResult = math.sqrt(b * b - 4 * a * c)
  local x1 = (-b + squareRootResult) / (2 * a)
  local x2 = (-b - squareRootResult) / (2 * a)

  if MathUtils.isRealNumber(x1) then
    table.insert(solutions, x1)
  end
  if x2 ~= x1 and MathUtils.isRealNumber(x2) then
    table.insert(solutions, x2)
  end
  return solutions
end

--- Format a number so that commas are inserted every three digits
-- @value Value to format
function MathUtils.getCommaFormattedValue(value)
  local formatted = value
  while true do
    formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
    if (k == 0) then
      break
    end
  end

  return formatted
end

-- Returns one dimension's component of a point along a Bezier curve.
-- @param a (number) - coordinate of the starting point
-- @param b (number) - coordinate of the first control point
-- @param c (number) - coordinate of the second control point
-- @param d (number) - coordinate of the ending point
-- @param t (number) - the proportion (between 0.0 and 1.0) of curve that has already been drawn
function MathUtils.bezier(a, b, c, d, t)
  local t2 = t * t
  local t3 = t2 * t

  local u = 1.0 - t
  local u2 = u * u
  local u3 = u2 * u

  return ((u3 * a) + (3.0 * t * u2 * b) + (3.0 * t2 * u * c) + (t3 * d))
end

--- Return one dimension's component of a point along a cardinal spline.
-- @param a (number) - coordinate of the first control point
-- @param b (number) - coordinate of the starting point
-- @param c (number) - coordinate of the ending point
-- @param d (number) - coordinate of the second control point
-- @param tension (number) - a value between 0.0 and 1.0, inclusive.  Use 0.0 to draw a Catmull-Rom curve.
-- @param t (number) - the proportion (between 0.0 and 1.0) of curve that has already been drawn
function MathUtils.cardinalSpline(a, b, c, d, tension, t)
  local t2 = t * t
  local t3 = t2 * t

  local s = (1.0 - tension) * 0.5

  local b1 = s * ((-t3 + (2.0 * t2)) - t)
  local b2 = s * (-t3 + t2) + (2.0 * t3 - 3.0 * t2 + 1.0)
  local b3 = s * (t3 - 2.0 * t2 + t) + (-2.0 * t3 + 3.0 * t2)
  local b4 = s * (t3 - t2)

  return a * b1 + b * b2 + c * b3 + d * b4
end
