--- StateMachine
-- @author Created by karl on May 06, 2014.

-- ===========================================================================
class("StateMachine")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function queueTransition(self, newState, ...)
  table.insert(self.transitionsQueue, { state = newState, arguments = arg })
end

local function doTransition(self, newState, ...)
  local canTransition = self:canTransition(newState)
  if canTransition then
    self.currentState.onLeave()
    self.currentState = self.states[newState]
    self.currentState.onEnter(...)
  else
  end
end

local function handleTransitions(self)
  self.isHandlingTransitions = true
  while #self.transitionsQueue > 0 do
    local transition = table.remove(self.transitionsQueue, 1)
    doTransition(self, transition.state, unpack(transition.arguments))
  end
  self.isHandlingTransitions = false
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Add a state to this machine.
-- @param name (string) - The name used to reference this state.
-- @param onEnter (function) - Function to call when entering this state. Can be nil.
-- @param onLeave (function) - Function to call when leaving this state. Can be nil.
-- @param ... (strings) - States that may be transitioned to from this state.
function StateMachine:addState(name, onEnter, onLeave, ...)
  local state = {
    name = name,
    onEnter = onEnter or function() end,
    onLeave = onLeave or function() end,
    transitions = TableUtils.createSetFromArray(arg)
  }
  self.states[name] = state
end

--- Check if the current state can transition to the given new state.
-- Note that it only checks the current state, but there may be other states queued
-- for transitioning that prevent a transition to the new state from happening.
-- @param newState (string) - The name used to reference the new state.
-- @return (boolean) - true if the transition would be allowed, false otherwise.
function StateMachine:canTransition(newState)
  if self.currentState.transitions then
    return self.currentState.transitions[newState]
  end
  return true
end

--- Transition to a new state.
-- @param newState (string) - the name of the new state.
-- @param ... (anything) - Passed to the onEnter function
function StateMachine:transition(newState, ...)
  queueTransition(self, newState, ...)
  if not self.isHandlingTransitions then
    handleTransitions(self)
  end
end

function StateMachine:getCurrentState()
  return self.currentState.name
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function StateMachine.new()
  local self = {}
  Mixin.mix(self, StateMachine)
  self.states = {}
  self.currentState = {
    onLeave = function() end
  }

  self.isHandlingTransitions = false
  self.transitionsQueue = {}

  return self
end
