--- StringUtils
-- @author Created by Aaron Barrett on May 04, 2014.

-- ===========================================================================
class("StringUtils")
-- ===========================================================================

require("utils.LevenshteinDistance")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function shouldNotCapitalizeInTitleCase(word)
  local word = word:lower()
  return word == "of" or word == "in"
end

local function titleCaseHelper(first, rest)
  local fullWord = first..rest
  if(shouldNotCapitalizeInTitleCase(fullWord)) then
    return fullWord:lower()
  end
  return first:upper()..rest:lower()
end

local tableToString = nil

local function valueToString(v)
  if ("string" == type(v)) then
    v = string.gsub(v, "\n", "\\n")
    if (string.match(string.gsub(v,"[^'\"]",""), '^"+$'))then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  elseif "table" == type(v) then
    return tableToString(v)
  else
    return _G.tostring(v)
  end
end

local function keyToString(k)
  if "string" == type(k) and string.match(k, "^[_%a][_%a%d]*$") then
    return k
  else
    return "[" .. valueToString( k ) .. "]"
  end
end

tableToString = function(t)
  local result, done = {}, {}
  for k, v in ipairs(t) do
    table.insert(result, valueToString(v))
    done[ k ] = true
  end
  for k, v in pairs( t ) do
    if not done[ k ] then
      table.insert(result, keyToString(k) .. "=" .. valueToString(v))
    end
  end
  return "{" .. table.concat( result, "," ) .. "}"
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Split a string along a seperator.
-- @param stringToSplit - (string) the string to parse.
-- @param separatorString - (string) token to split on (Default = ":")
-- @return A numberic table of strings
function StringUtils.split(stringToSplit, separatorString)
  Asserts.assertType(stringToSplit, "stringToSplit", "string")
  Asserts.assertType(separatorString, "separatorString", "string")

  local separatorString = separatorString or ":"
  local separatedStrings = {}
  local pattern = string.format("([^%s]+)", separatorString)

  local function onStringFound(foundString)
    table.insert(separatedStrings, foundString)
  end

  string.gsub(stringToSplit, pattern, onStringFound)

  return separatedStrings
end

--- Remove leading or trailing whitespace from a string.
-- @param stringToTrim - (string) the string to trim.
-- @return The string with no leading or trailing whitespace
function StringUtils.trim(stringToTrim)
  Asserts.assertType(stringToTrim, "stringToTrim", "string")

  local trimmedString = string.gsub(stringToTrim, "^%s*(.-)%s*$", "%1")
  return trimmedString
end

--- Checks whether string is free of leading or trailing whitespace.
-- @param - stringToCheck - (string) The string to check for whitespace
-- @return true if stringToCheck has no leading or trailing whitespace, otherwise false
function StringUtils.isTrimmed(stringToCheck)
  Asserts.assertType(stringToCheck, "stringToCheck", "string")
  return stringToCheck == StringUtils.trim(stringToCheck)
end

--- Checks whether all characters in a string are uppercase.
-- @param - stringToCheck - (string) The string to check for capitalization
-- @return true if all characters in stringToCheck are uppercase, otherwise false
function StringUtils.isUppercase(stringToCheck)
  Asserts.assertType(stringToCheck, "stringToCheck", "string")
  return stringToCheck == string.upper(stringToCheck)
end

--- Makes the first letter of the specified string uppercase
-- @param stringToCapitalize - The string to capitalize
-- @return the string with the first letter capitalized
function StringUtils.capitalizeFirstLetter(stringToCapitalize)
  Asserts.assertType(stringToCapitalize, "stringToCapitalize", "string")
  local capitalizedString = string.gsub(stringToCapitalize, "^%l", string.upper)
  return capitalizedString
end

--- Convert an object to a string. Supports deep stringifying of tables.
function StringUtils.tostring(object)
  if type(object) == "table" then
    return tableToString(object)
  end
  return _G.tostring(object)
end

--- Check whether a string begins with a prefix.
function StringUtils.startsWith(stringToCheck, prefix)
  local prefixLength = #prefix
  local substring = stringToCheck:sub(1, prefixLength)
  return (substring == prefix)
end

--- Check whether a string ends with a suffix.
function StringUtils.endsWith(stringToCheck, suffix)
  local suffixLength = #suffix
  local substring = stringToCheck:sub(-suffixLength)
  return (substring == suffix)
end

function StringUtils.getLevenshteinDistance(word1, word2)
  Asserts.assertType(word1, "word1", "string")
  Asserts.assertType(word2, "word2", "string")

  local distance = LevenshteinDistance.calculateDamerauLevenshtein(word1, word2)
  return distance
end

--- Returns a string of unique characters from the provided string
function StringUtils.getUniqueCharacterString(unfilteredString)
  local allowedType = "string"
  Asserts.assertType(unfilteredString, "unfilteredString", allowedType)

  if (unfilteredString == "") then
    return unfilteredString
  end

  local characterSet = TableUtils.createSetFromString(unfilteredString)
  local uniqueString = ""

  for character, _ in pairs(characterSet) do
    uniqueString = uniqueString..character
  end

  return uniqueString
end

--- Returns a string with tags stripped. Assumes tags to be anything contained inside of < and > characters
-- ex. "That's a <b>huge</b> donkey" -> "That's a huge donkey"
function StringUtils.getAfterStrippingTags(text)
  text = text:gsub("%b<>", "")

  return text
end

--- Returns all substrings contained within a tag. The "tag" parameter should come without brackets
function StringUtils.findSubstringsInTags(text, tag)
  local taggedStrings = {}
  local openTag = "<" .. tag .. ">"
  local closeTag = "</" .. tag .. ">"
  local parseString = text
  local substrings = {}
  while parseString do
    local startIndex, endIndex = parseString:find(openTag, 1, true)
    if startIndex then
      parseString = parseString:sub(endIndex + 1)
      table.insert(substrings, parseString)
    else
      parseString = nil
    end
  end

  for _, substring in ipairs(substrings) do
    local closeTagIndex = substring:find(closeTag, 1, true)
    if closeTagIndex then
      taggedString = substring:sub(1, closeTagIndex - 1)
      table.insert(taggedStrings, taggedString)
    end
  end

  return taggedStrings
end
