--- Misc. helper functions for table manipulation
-- @author Jay Austria

-- ===========================================================================
class("TableUtils")
-- ===========================================================================

--- Get all keys in a table.
-- @param table - (table) A table.
-- @return a new array (table) containing all of the source table's keys.
function TableUtils.getKeys(t)
  local keys = {}
  for key in pairs(t) do
    table.insert(keys, key)
  end
  return keys
end

--- Get a table's keys, sorted based on their associated values.
-- @param tbl (table) - The table whose keys to get.
-- @param sortFunction (function taking params a and b) - The function to use for sorting.
--        It will be passed the values, not the keys.
--        The return value is a boolean, same as for table.sort().
function TableUtils.getKeysSortedByValue(tbl, sortFunction)
  assert(type(sortFunction) == "function", "sortFunction must be a function")
  local keys = TableUtils.getKeys(tbl)
  table.sort(keys, function(a, b)
    return sortFunction(tbl[a], tbl[b])
  end)
  return keys
end

--- Get all values in a table. If there are duplicate values, there will
-- be duplicate entries in the resulting array.
-- @param table - (table) A table.
-- @return a new array (table) containing all of the source table's values.
function TableUtils.getValues(t)
  local values = {}
  for _, value in pairs(t) do
    table.insert(values, value)
  end
  return values
end

--- Copies all of the keys in table 't' into a new table.
-- Copies only references (not values) from the top-level (no recursion).
-- @param t - (table) The table whose keys will be copied
-- @return a new table containing all of the same keys as 't'.
function TableUtils.copyKeysFromTable(t)
  local t2 = {}
  for k, v in pairs(t) do
    t2[k] = v
  end
  return t2
end

--- Appends all arrays in '...' onto the end of destinationArray
-- @param ... - (array) Arrays to append to the destinationArray
-- @return destinationArray
function TableUtils.concatenateArrays(destinationArray, ...)
  for _, sourceArray in ipairs(arg) do
    for _, value in ipairs(sourceArray) do
      table.insert(destinationArray, value)
    end
  end

  return destinationArray
end

--- Concatenates a variable number of arrays into a newly created array.
-- Original arrays are unmodified. Order is preserved.
-- @param ... - (array) Arrays to concatenate
-- @return New array containing all items from the original arrays.
function TableUtils.createConcatenatedArray(...)
  local result = {}
  TableUtils.concatenateArrays(result, ...)

  return result
end

function TableUtils.indexOf(table, value)
  for index, element in ipairs(table) do
    if (element == value) then
      return index
    end
  end
end

function TableUtils.removeValueFromArray(array, value)
  local indexOfValue = TableUtils.indexOf(array, value)
  if indexOfValue ~= nil then
    table.remove(array, indexOfValue)
  end
end

--- Return a subset of the specified array.
function TableUtils.subarray(array, startIndex, endIndex)
  local result = {}
  for i=startIndex, endIndex do
    table.insert(result, array[i])
  end
  return result
end

--- Find a key that maps to a value.
-- Search order is not guaranteed. If more than one key maps to the same value,
-- there is no guarantee as to which key will be returned.
-- @param table - The table to search in.
-- @param value - The value to search for.
-- @return The corresponding key or nil.
function TableUtils.keyForValue(table, value)
  for k, v in pairs(table) do
    if v == value then
      return k
    end
  end
  return nil
end

--- Remove the key that maps to a value.
-- Search order is not guaranteed. If more than one key maps to the same value,
-- there is no guarantee as to which key will be removed.
-- @param table - The table to search in.
-- @param value - The value to search for.
-- @return true if a key was removed.
function TableUtils.removeKeyForValue(table, value)
  local key = TableUtils.keyForValue(table, value)
  if key then
    table[key] = nil
    return true
  end
  return false
end

function TableUtils.swap(array, i, j)
  if (i == nil) then
    i = 1
  end

  if (j == nil) then
    j = 2
  end

  local hasNilElement = ((array[i] == nil) or (array[j] == nil))
  assert(not hasNilElement, "Nil element found when swapping table values.")

  array[i], array[j] = array[j], array[i]
end

--- Randomly shuffle the contents on an indexed lua table
-- shuffles a table by swapping random indexed items a number of
-- times (based on the size of the table)
-- @param array table of indexed values that will be shuffled
-- @return shuffled array
function TableUtils.shuffleArray(array)
  local arrayCount = #array
  for i = arrayCount, 2, -1 do
    local j = math.random(1, i)
    TableUtils.swap(array, i, j)
  end
  return array
end

--- Creates and shuffles a shallow copy of the given indexed table.
-- @param (indexed table) - the array to shuffle
-- @return a shuffled copy of the array
function TableUtils.createShuffledArray(array)
  local arrayCopy = TableUtils.shallowCopy(array)
  TableUtils.shuffleArray(arrayCopy)
  return arrayCopy
end

--- Remove the first n elements from the beginning of an array
-- @param array - table of indexed values to modify
-- @param elementsToRemoveCount - the number of elements to remove from the front of the array
function TableUtils.removeFromFrontOfArray(array, elementsToRemoveCount)
  for i = 1, elementsToRemoveCount do
    table.remove(array, 1)
  end
end

--- Remove the first n elements from the end of an array
-- @param array - table of indexed values to modify
-- @param elementsToRemoveCount - the number of elements to remove from the end of the array
function TableUtils.removeFromEndOfArray(array, elementsToRemoveCount)
  for i = 1, elementsToRemoveCount do
    table.remove(array)
  end
end

--- Remove the last element from the end of an array
-- @param array - table to remove from
-- @return the element
function TableUtils.arrayPopBack(array)
  assert(type(array) == "table", "Expected array to be a table, but got " .. type(array))

  return table.remove(array)
end

--- Insert an item into an array at a specified index. Causes all indices after
-- the item to increase by one.
-- @param array - table to insert into
-- @param item - the element to inser
-- @param index - the index to insert the item into
function TableUtils.arrayInsertAtIndex(array, item, index)
  assert(type(array) == "table", "Expected array to be a table, but got " .. type(array))
  assert(((index > 0) and (index <= #array + 1)), "Out of bounds index provided to TableUtils.arrayInsertAtIndex")

  table.insert(array, index, item)
end

--- Insert an item into the front of an array
-- @param array - table to insert into
-- @param item - the element to insert
function TableUtils.arrayPushFront(array, item)
  TableUtils.arrayInsertAtIndex(array, item, 1)
end

--- Count the number of occurances of an object in a table
-- Iterates through a table and counts the number of occurences
-- of an item
-- @param tt lua table to search through
-- @param item the item we want to count the occurences of
-- @return count of how  many times the object was found in the table
function TableUtils.tableCount(tt, item)
  local count
  count = 0
  for ii, xx in pairs(tt) do
    if item == xx then count = count + 1 end
  end
  return count
end

--- Return true if a table contains a value
-- @param t lua table to search through
-- @param item the item we want to check for
-- @return true or false
function TableUtils.contains(t, item)
  for k, v in pairs(t) do
    if (v == item) then
      return true
    end
  end
  return false
end

--- Randomly samples 'numberOfElementsToSample' number of elements from an array.
-- copies a number of random elements from an array and returns a
-- new array containing the copied values. (Lottery)
-- @param array (table) - the array from which elements will be randomly sampled
-- @param numberOfElementsToSample (number) - the number of elements to choose from 'array'
-- @return (table) - an array containing randomly chosen values from the input array.
function TableUtils.createRandomSampleOfArray(array, numberOfElementsToSample)
  assert(type(array) == "table", "'array' must be a numeric table.")
  assert(type(numberOfElementsToSample) == "number", "'numberOfElementsToSample' must be a number.")
  assert(#array >= numberOfElementsToSample, "'array' must contain at least as many elements as 'numberOfElementsToSample'.")

  local arrayCopy = TableUtils.shallowCopy(array)
  local result = {}

  for i = 1, numberOfElementsToSample do
    local randomlyChosenIndex = math.random(#arrayCopy)
    local chosenElement = table.remove(arrayCopy, randomlyChosenIndex)
    table.insert(result, chosenElement)
  end

  return result
end

--- Creates an array following a number sequence as a for-loop would generate it.
-- @param firstNumber (number) First number for the sequence.
-- @param lastNumber (number) Limit for the sequence (may or may not be the last element depending on the value of increment).
-- @param increment (number) Difference between one element and the previous one. Optional, defaults to 1 or -1 depending on firstNumber, lastNumber.
-- @return (table) Array of numbers following the sequence.
function TableUtils.getNumberSequence(firstNumber, lastNumber, increment)
  if (firstNumber == lastNumber) then
    return {firstNumber}
  end

  increment = increment or MathUtils.sign(lastNumber - firstNumber)
  assert(MathUtils.sign(increment) == MathUtils.sign(lastNumber - firstNumber), "The increment should have the same sign as (lastNumber - firstNumber)")

  local array = {}
  for n = firstNumber, lastNumber, increment do
    table.insert(array, n)
  end

  return array
end

--- Creates a table of random values
-- Generate a table of n through m values and shuffle it before returning
-- @param n Min number to include in the table (inclusive)
-- @param m Max number to include in the table (inclusive)
-- @return newTable Table of shuffled random values
function TableUtils.getTableOfRandomValues(n, m)
  local newTable = TableUtils.getNumberSequence(n, m)
  return TableUtils.shuffleArray(newTable)
end

--- Get a table containing values showing the ranks of the input's values
-- Takes an array of numbers and returns the value's ranks where lower values
-- are ranked higher. Example, a table containing {4, 23, 2, 100} will get returned
-- a table with ranks {2,3,1,4}
-- @param inTable Table of values to create a rank table of
-- @return a table of values indicating the ranked order of the values in the inTable
function TableUtils.getValueRanks(inTable)
  local tableCopy = {}
  for k, v in pairs(inTable) do
    tableCopy[k] = v
  end
  table.sort(tableCopy)

  local reversedKeysAndValues = {}
  for k, v in pairs(tableCopy) do
    reversedKeysAndValues[v] = k
  end

  local valueRanks = {}
  for _, v in pairs(inTable) do
    table.insert(valueRanks, reversedKeysAndValues[v])
  end

  return valueRanks
end

function TableUtils.shallowCopy(t)
  local copy = {}
  for k, v in pairs(t) do
    copy[k] = v
  end

  return copy
end

function TableUtils.deepCopy(t)
  if type(t) ~= 'table' then return t end

  local copy = {}
  for k, v in pairs(t) do
    if (type(v) == 'table') then
      v = TableUtils.deepCopy(v)
    end
    copy[k] = v
  end

  return copy
end

--- Builds a dictionary table using the passed keys and values, taking them as pairs in the .
-- @param keys (table) Array containing keys.
-- @param values (table) Array containing values. Must have same length as keys.
-- @return (table) Dictionary that maps the ith key to the ith value.
function TableUtils.buildDictionary(keys, values)
  Asserts.assertType(keys, "keys", "table")
  Asserts.assertType(values, "values", "table")
  assert(#keys == #values, string.format("keys and values should have the same length. #keys: %s, #values: %s", #keys, #values))

  local dictionary = {}

  for i, key in ipairs(keys) do
    local value = values[i]
    dictionary[key] = value
  end

  return dictionary
end

function TableUtils.isEmpty(t)
  return next(t) == nil
end

function TableUtils.areEmpty(...)
  local areEmpty = true
  for _, t in ipairs(arg) do
    if (not TableUtils.isEmpty(t)) then
      areEmpty = false
      break
    end
  end

  return areEmpty
end

function TableUtils.createSetFromArray(array)
  local set = {}
  for _, element in ipairs(array) do
    set[element] = true
  end
  return set
end

function TableUtils.joinArrayToSet(array, set)
  set = set or {}
  for _, element in ipairs(array) do
    set[element] = true
  end
  return set
end

--- Creates an array from a set. The ordering of the array
-- is NOT guaranteed.
function TableUtils.createArrayFromSet(set)
  local array = {}
  for value, _ in pairs(set) do
    table.insert(array, value)
  end
  return array
end

function TableUtils.cloneSet(set)
  return TableUtils.copyKeysFromTable(set)
end

-- Useful for any table that is not a simple contiguous array
function TableUtils.getSize(t)
  local count = 0

  for _ in pairs(t) do
    count = count + 1
  end

  return count
end

--- Returns a copy of the array, with only the values that satisfy a condition specified by the passed function.
-- @param array (table) Original array. Will not be modified.
-- @param filterCondition (function) Function that takes an element from the array and returns true/false.
-- @return (table) Array that only contains elements for which filterCondition returned true.
function TableUtils.getFilteredArray(array, filterCondition)
  local result = {}

  for _, value in ipairs(array) do
    if filterCondition(value) then
      table.insert(result, value)
    end
  end

  return result
end

--- Returns a copy of the table, with only the values that satisfy a condition specified by the passed function.
-- @param tbl (table) Original table. Will not be modified.
-- @param filterCondition (function) Function that takes a value from the table and returns true/false.
-- @return (table) Table that only contains elements for which filterCondition returned true.
function TableUtils.getFilteredTable(tbl, filterCondition)
  local result = {}

  for key, value in pairs(tbl) do
    if filterCondition(value) then
      result[key] = value
    end
  end

  return result
end

--- Get a copy of an array without duplicates. Preserves item order. Does not modify original array.
-- @param array (table) Input array to remove duplicates from. It will not be modified.
-- @return (table) New array containing unique elements in the order they first appear in the original array.
function TableUtils.getUniquesArray(array)
  local uniquesArray = {}
  local seenElements = {}
  for _, element in ipairs(array) do
    if not seenElements[element] then
      table.insert(uniquesArray, element)
      seenElements[element] = true
    end
  end
  return uniquesArray
end

--- Get the duplicates in an array. Preserves item order. Does not modify original array.
--  If an item is present more than two times, it will be present in the result more than once.
-- @param array (table) Input array to find duplicates in. It will not be modified.
-- @return (table) New array containing duplicated elements in the order they first appear in the original array.
function TableUtils.getDuplicatesInArray(array)
  local seenItemsSet = {}
  local duplicateItems = {}
  for _, item in ipairs(array) do
    if seenItemsSet[item] then
      table.insert(duplicateItems, item)
    else
      seenItemsSet[item] = true
    end
  end
  return duplicateItems
end

--- Returns a copy of the array, with the same elements in reversed order.
-- @param array (table) Original array. Will not be modified.
-- @return (table) Reversed array.
function TableUtils.getReversedArray(array)
  local result = {}

  for i = #array, 1, -1 do
    table.insert(result, array[i])
  end

  return result
end

--- Sorts an array from highest value to lowest value.
-- @param array (table) Original array to modify in place
function TableUtils.reverseSortArray(array)
  Asserts.assertType(array, "array", "table")
  table.sort(array, function(a, b) return a > b end)
end

--- Returns an array mapping each value using func.
-- @param array (table) Original array. Will not be modified.
-- @param func (function) Function to be applied to each element f(value)
-- @return (table) Mapped array
function TableUtils.map(array, func)
  local result = {}

  for i, v in ipairs(array) do
    result[i] = func(v)
  end

  return result
end

function TableUtils.createArrayFromString(text)
  local array = {}
  for i = 1, #text do
    local char = text:sub(i, i)
    table.insert(array, char)
  end
  return array
end

function TableUtils.createSetFromString(text)
  local array = TableUtils.createArrayFromString(text)
  local set = TableUtils.createSetFromArray(array)
  return set
end

-- Set manipulation functions ---------
--
-- These are meant to be used on sets of the form {a = true, b = true, ...}
-- However, they will work equally well with any key-value table.
-- The only caveat is that intersectionSet and unionSet have special
-- behavior for values contained in both "sets". Namely, values in setA
-- always supercede values in setB.

--- Set of keys contained in both setA and setB.
-- Note: Only values in setA will be stored in the result.
function TableUtils.intersectionSet(setA, setB)
  local result = {}
  for k, v in pairs(setA) do
    if setB[k] then
      result[k] = v
    end
  end
  return result
end

--- Set of keys contained in either of the received sets.
-- Note: Intersecting values will be taken from the set passed as the first argument.
function TableUtils.unionSet(...)
  local result = {}
  for _, set in rpairs(arg) do
    for k, v in pairs(set) do
      result[k] = v
    end
  end
  return result
end

--- Set of keys contained in setU but not in setA.
function TableUtils.differenceSet(setU, setA)
  local result = {}
  for k, v in pairs(setU) do
    if not setA[k] then
      result[k] = v
    end
  end
  return result
end

--- Set of keys contained in either set but not in the other.
function TableUtils.symmetricDifferenceSet(setA, setB)
  local result = {}
  for k, v in pairs(setA) do
    if not setB[k] then
      result[k] = v
    end
  end
  for k, v in pairs(setB) do
    if not setA[k] then
      result[k] = v
    end
  end
  return result
end

--- Takes a string in the format
--  @param csvString - "abc, def, ghijklmno, pqrstu, vwx, yz"
--  and transforms it into an array:
--  @return {"abc", "def", "ghijklmno", "pqrstu", "vwx", "yz"}
function TableUtils.csvToArray(csvString)
  local valueType = type(csvString)
  local isValidType = (valueType == "string" or valueType == "number")
  assert(isValidType, string.format("Must pass a string or number, received %s, of type %s", tostring(csvString), valueType))
  csvString = tostring(csvString)

  local stringValues = StringUtils.split(csvString, ",")
  local result = {}
  for _, value in ipairs(stringValues) do
      result[#result + 1] = StringUtils.trim(value)
  end
  return result
end

-- Return the sum of a table of values
-- @param array (table) Table of values (number)
-- @return Sum of all values in the table
function TableUtils.sum(array)
  assert(type(array) == "table", "Expected array to be a table, but got " .. type(array))

  local sum = 0
  for _, value in pairs(array) do
    sum = sum + value
  end

  return sum
end

-- Randomly select from the table using weights. Doesn't have to be whole numbers
-- @param values (table) possible values to select
-- @param weights (table) weights for each possible value
-- @return Randomly selected value
function TableUtils.selectRandomValueWithWeights(values, weights)
  Asserts.assertType(values, "values", "table")
  Asserts.assertType(weights, "weights", "table")
  assert(#values == #weights, "Number of values and weights must be equal")

  local weightsSum = TableUtils.sum(weights)
  local selectionThreshold = math.random() * weightsSum

  for index, weight in ipairs(weights) do
    if selectionThreshold <= weight then
      return values[index], index
    end

    selectionThreshold = selectionThreshold - weight
  end

  error("Randomly selected threshold somehow exceeded sum of weights in `weights`")
end

function TableUtils.getRandomElement(array, doRemove)
  doRemove = doRemove or false

  local numContents = #array
  local randomElementIndex = math.random(numContents)
  local randomElement = nil

  if doRemove then
    randomElement = table.remove(array, randomElementIndex)
  else
    randomElement = array[randomElementIndex]
  end

  return randomElement
end

function TableUtils.createArrayFromField(t, fieldName)
  local arrayFromField = TableUtils.map(t, function(item) return item[fieldName] end)
  assert((#arrayFromField == #t), "Table passed to createArrayFromField contained items that didn't have the specified fieldName")

  return arrayFromField
end

function TableUtils.createStringToStringMapping(sourceTable)
  local mapping = {}
  for key, value in pairs(sourceTable) do
    mapping[tostring(key)] = tostring(value)
  end
  return mapping
end

function TableUtils.reduce(array, reduceFunc, defaultVal)
  assert(type(array) == "table", "Expected array to be a table, but got " .. type(array))

  local reduction = nil
  for _, value in pairs(array) do
    reduction = reduction and reduceFunc(reduction, value) or value
  end

  return reduction or defaultVal
end

function TableUtils.find(array, testFunc)
  assert(type(array) == "table", "Expected array to be a table, but got " .. type(array))

  for _, value in pairs(array) do
    if (testFunc(value)) then
      return value
    end
  end

  return nil
end
