--- XmlParser
-- @author Created by amorant on October 10, 2014.
-- Parses an XML document and produces a user-friendly, navigable XML tree. Uses XmlElement.
-- Parsing code in `parseargs` and `collect` originally obtained from: http://lua-users.org/wiki/LuaXml
-- and modified to include token position information.

class("XmlParser")
require("utils.XmlElement")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Parsing ----------------------------

local function parseargs(s)
  local arg = {}
  string.gsub(s, "([%-%w]+)=([\"'])(.-)%2", function (w, _, a)
    arg[w] = a
  end)
  return arg
end

local function collect(s)
  local stack = {}
  local top = {}
  table.insert(stack, top)
  local ni,c,label,xarg, empty
  local i, j = 1, 1
  while true do
    ni,j,c,label,xarg, empty = string.find(s, "<(%/?)([%w:]+)(.-)(%/?)>", i)
    if not ni then break end
    label = label:lower()
    local text = string.sub(s, i, ni-1)
    if not string.find(text, "^%s*$") then
      table.insert(top, {text=text, position=i})
    end
    if empty == "/" then  -- empty element tag
      table.insert(top, {label=label, xarg=parseargs(xarg), empty=1, position=ni})
    elseif c == "" then   -- start tag
      top = {label=label, xarg=parseargs(xarg), position=ni}
      table.insert(stack, top)   -- new level
    else  -- end tag
      local toclose = table.remove(stack)  -- remove top
      top = stack[#stack]
      if #stack < 1 then
        error("nothing to close with "..label)
      end
      if toclose.label ~= label then
        error("trying to close "..toclose.label.." with "..label)
      end
      table.insert(top, toclose)
    end
    i = j+1
  end
  local text = string.sub(s, i)
  if not string.find(text, "^%s*$") then
    table.insert(stack[#stack], {text=text, position=i})
  end
  if #stack > 1 then
    error("unclosed "..stack[#stack].label)
  end
  return stack[1]
end

-- Xml tree building ------------------

local function buildXmlTree(parsedElement)
  local xmlElement = nil

  if parsedElement.text ~= nil then
    xmlElement = XmlElement.newTextElement(parsedElement.text, parsedElement.position)
  else
    local children = {}
    for _, parsedChild in ipairs(parsedElement) do
      local childXmlElement = buildXmlTree(parsedChild)
      table.insert(children, childXmlElement)
    end
    xmlElement = XmlElement.newTagElement(parsedElement.label, parsedElement.xarg, children, parsedElement.position)
  end

  return xmlElement
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Parse an XML string into an XML tree.
-- @param text - (string) Text containing xml tags.
-- @return (XmlElement) The root element of the tree.
function XmlParser.parse(text)
  local parsedText = collect(text)
  local xmlTree = buildXmlTree(parsedText)
  return xmlTree
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Constructor
-- @return (XmlParser) - The new instance.
function XmlParser.new()
  local self = {}
  Mixin.mix(self, XmlParser)

  return self
end
