--- ConceptValidator

class("ConceptValidator")
require("validators.Validator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

ConceptValidator.LETTERS_SET = TableUtils.createSetFromString('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZáéíóúàèìòùâêîôûäëïöüÁÉÍÓÚÀÈÌÒÙÂÊÎÔÛÄËÏÖÜñÑøØ¡¿ªºüÜ')
ConceptValidator.NON_LETTERS_SET = TableUtils.createSetFromString('0123456789!"#$€%&\'()*+,-./:;<=>?@[\\]^_`{|}~ °£–—«» ·')
ConceptValidator.PRINTABLE_SET = TableUtils.unionSet(ConceptValidator.LETTERS_SET, ConceptValidator.NON_LETTERS_SET)

ConceptValidator.PUNCTUATION_SET = TableUtils.createSetFromString('.!?')
ConceptValidator.DOUBLE_QUOTATION_SET = TableUtils.createSetFromString('\"')
ConceptValidator.SINGLE_QUOTATION_SET = TableUtils.createSetFromString('\'')
ConceptValidator.WORD_SEPARATORS_SET = TableUtils.unionSet(
  ConceptValidator.PUNCTUATION_SET,
  ConceptValidator.DOUBLE_QUOTATION_SET,
  ConceptValidator.SINGLE_QUOTATION_SET,
  TableUtils.createSetFromString(" ,;:")
)

local SUPPORTED_IMAGE_FORMATS = TableUtils.createSetFromArray({
  "png"
})

-- specfic strings that google sheets may pass to the content.db when there is
-- a forumla error
local CONTENT_ERROR_STRINGS = {
  "#N/A",
  "#ERROR!"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function isPunctuation(character)
  return (ConceptValidator.PUNCTUATION_SET[character] ~= nil)
end

local function isSingleQuote(character)
  return (ConceptValidator.SINGLE_QUOTATION_SET[character] ~= nil)
end

local function isDoubleQuote(character)
  return (ConceptValidator.DOUBLE_QUOTATION_SET[character] ~= nil)
end

local function isQuote(character)
  return isSingleQuote(character) or isDoubleQuote(character)
end

local function getExtension(path)
  if not path then
    return nil
  end
  local extension = string.match(path, "%.([^/]+)$")
  return extension
end

local function isSupportedImageFileExtension(extension)
  return SUPPORTED_IMAGE_FORMATS[extension] ~= nil
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Main API ---------------------------

function ConceptValidator:validate(concept)
  self:runValidation(function()
    self:validateConcept(concept)
    self:validateNoLeadingOrTrailingWhitespace(concept)
  end)
end

-- To be overriden by subclasses

function ConceptValidator:validateConcept(concept)
  error("Not implemented")
end

-- Validation helper methods ----------

function ConceptValidator:validateRequiredKeys(concept, requiredKeys, allowEmpty)
  self:assertHasKeys(concept, requiredKeys, allowEmpty)

  for _, key in ipairs(requiredKeys) do
    local value = concept[key]
    local valueString = StringUtils.tostring(value)
    self:expectStringIsPrintable(valueString, key)
    self:expectNoDoubleSpaces(valueString, key)
  end
end

function ConceptValidator:validateImageFile(filename)
  local extension = getExtension(filename)
  self:expectTrue(isSupportedImageFileExtension(extension), tostring(filename) .. " has an invalid image filename extension '" .. tostring(extension) .. "'")
end

function ConceptValidator:expectStringContains(text, subtext, fieldName)
  fieldName = fieldName or ""

  local plain = true
  local found = text:find(subtext, 1, plain) ~= nil
  local errorMessage = string.format("Field %s should contain '%s', found value: %s", fieldName, subtext, text)
  return self:expectTrue(found, errorMessage)
end

function ConceptValidator:validateCharactersAdjacentToBlanks(text, blankString)
  local acceptableCharacters = TableUtils.createArrayFromSet(ConceptValidator.WORD_SEPARATORS_SET)

  for i = 1, #text do
    local character = text:sub(i, i)
    if character == blankString then
      if i > 1 then
        self:expectIsContained(text:sub(i - 1, i - 1), acceptableCharacters, string.format("character before blank (%s)", text))
      end
      if i < #text then
        self:expectIsContained(text:sub(i + 1, i + 1), acceptableCharacters, string.format("character after blank (%s)", text))
      end
    end
  end
end

function ConceptValidator:validateNoLeadingOrTrailingWhitespace(concept)
  local stack = {concept}

  while (#stack > 0) do
    local top = table.remove(stack, #stack)

    for key, value in pairs(top) do
      local valueType = type(value)
      if (valueType == "table") then
        table.insert(stack, value)
      elseif (valueType == "string") then
        self:expectNoLeadingOrTrailingWhitespace(value, "value for key: " .. key)
      end
    end
  end
end

function ConceptValidator:testForEndingPunctuation(sentence)
  local indexToCheck = #sentence
  local lastCharacter = sentence:sub(indexToCheck, indexToCheck)

  if isQuote(lastCharacter) then
    indexToCheck = indexToCheck - 1
    lastCharacter = sentence:sub(indexToCheck, indexToCheck)
  end

  self:expectTrue(isPunctuation(lastCharacter), "The sentence should end with punctuation: " .. sentence)
end

function ConceptValidator:expectStringCharactersBelongToSet(text, characterSet, fieldName)
  fieldName = fieldName or ""

  local characters = TableUtils.createArrayFromString(text)
  for index, char in ipairs(characters) do
    local belongsToSet = characterSet[char]
    local characterWithContext = text:sub(math.max(index - 3, 1), index + 3)
    self:expectTrue(belongsToSet, string.format("Forbidden character found at position %s of field %s: \"%s\"", index, fieldName, characterWithContext))
    if not belongsToSet then
      return false
    end
  end
  return true
end

function ConceptValidator:expectStringIsLettersOnly(text, fieldName, extraAllowedCharactersString)
  extraAllowedCharactersString = extraAllowedCharactersString or ""
  local extraCharactersSet = TableUtils.createSetFromString(extraAllowedCharactersString)
  local characterSet = TableUtils.unionSet(ConceptValidator.LETTERS_SET, extraCharactersSet)
  return self:expectStringCharactersBelongToSet(text, characterSet, fieldName)
end

function ConceptValidator:expectStringIsPrintable(text, fieldName)
  return self:expectStringCharactersBelongToSet(text, ConceptValidator.PRINTABLE_SET, fieldName)
end

function ConceptValidator:expectStringIsNotContentError(text, fieldName)
  local textIsContentError = TableUtils.contains(CONTENT_ERROR_STRINGS, text)
  return self:expectFalse(textIsContentError, fieldName)
end

function ConceptValidator:expectNoDoubleSpaces(text, fieldName)
  fieldName = fieldName or ""

  local plain = true
  local doubleSpaceString = "  "
  local found = (text:find(doubleSpaceString, 1, plain) ~= nil)
  local errorMessage = string.format("Field %s contained a double space", fieldName)

  return self:expectFalse(found, errorMessage)
end

function ConceptValidator:expectCapitalizationMatches(firstString, secondString, firstFieldName, secondFieldName)
  local EXCEPTIONS = { "I" }
  if TableUtils.contains(EXCEPTIONS, firstString) or TableUtils.contains(EXCEPTIONS, secondString) then
    return true
  end

  firstFieldName = firstFieldName or ""
  secondFieldName = secondFieldName or ""
  local isReplacementUpper = StringUtils.isUppercase(firstString:sub(1, 1))
  local isDistractorUpper = StringUtils.isUppercase(secondString:sub(1, 1))
  local errorMessage = string.format("%s (%s) and %s (%s) should have the same capitalization", firstFieldName, firstString, secondFieldName, secondString)
  self:expectTrue(isReplacementUpper == isDistractorUpper, errorMessage)
end

function ConceptValidator:expectNoLeadingOrTrailingWhitespace(text, fieldName)
  fieldName = fieldName or ""
  local isTrimmed = StringUtils.isTrimmed(text)
  local errorMessage = string.format("%s has a leading or trailing whitespace", fieldName)

  self:expectTrue(isTrimmed, errorMessage)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ConceptValidator.new()
  local self = Validator.new()
  Mixin.mix(self, ConceptValidator)

  return self
end
