--- ParametersValidator

class("ParametersValidator")
require("Mixin")
require("validators.Validator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local DIFFICULTY_PARAMETER = "difficulties"
local VALID_DIFFICULTIES = { "Basic", "Easy", "Medium", "Hard", "Advanced" }
local MINIMUM_DIFFICULTY_COUNT = 3 -- May change, but true at the moment.
local MAXIMUM_DIFFICULTY_COUNT = #VALID_DIFFICULTIES

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function isValidDifficulty(difficulty)
  return TableUtils.contains(VALID_DIFFICULTIES, difficulty)
end

local function isValidLength(difficultyCollection)
  local uniquesTable = TableUtils.getUniquesArray(difficultyCollection)
  local hasNoRedundancies = (#difficultyCollection == #uniquesTable)

  local hasCorrectDifficultyCount = (#uniquesTable >= MINIMUM_DIFFICULTY_COUNT and #uniquesTable <= MAXIMUM_DIFFICULTY_COUNT)

  return hasCorrectDifficultyCount and hasNoRedundancies
end

local function validateDifficultyArray(self, difficultyCollection)
  self:expectTrue(isValidLength(difficultyCollection), "invalid number of difficulties defined in difficulty array")

  for _, difficulty in pairs(difficultyCollection) do
    self:expectTrue(isValidDifficulty(difficulty), "invalid difficulty in difficulty array")
  end
end

local function validateDifficulties(self, parameters)
  for key, parameter in pairs(parameters) do
    if key == DIFFICULTY_PARAMETER then
      validateDifficultyArray(self, parameter)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Main API ---------------------------

function ParametersValidator:validate(parameters)
  self:runValidation(function()
    self:validateParameters(parameters)
    validateDifficulties(self, parameters)
  end)
end

-- To be overriden by subclasses

function ParametersValidator:validateParameters(parameters)
  error("Not implemented")
end

-- Main API ---------------------------

function ParametersValidator:testRequiredParameters(parameters, requiredParameters)
  parameters = parameters or {}
  -- Check that all required parameters are present
  self:assertHasKeys(parameters, requiredParameters)
  -- Check that only the required parameters are present
  for parameter, _ in pairs(parameters) do
    self:expectIsContained(parameter, requiredParameters, "parameters")
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ParametersValidator.new()
  local self = Validator.new()
  Mixin.mix(self, ParametersValidator)

  return self
end
