--- TagValidator
-- @author Created by Anton Morant on April 16, 2014.

class("TagValidator")
require("utils.TagUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

TagValidator.DEFAULT_OPTIONAL_PAIRED_TAGS = {
  "em",
  "i",
  "b"
}

local TAG_REGEX = "%<(.-)%>"
local MARKUP_REGEX = "[/$]"

local STATE_INITIAL = "INITIAL"
local STATE_OPEN = "OPEN"
local STATE_CLOSED = "CLOSED"
local STATE_FOUND = "FOUND"
local STATE_FOUND_TERTIARY = "FOUND_TERTIARY"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getInitialTagStates(tags)
  local tagStates = {}
  for _, tag in ipairs(tags) do
    tagStates[tag] = STATE_INITIAL
  end
  return tagStates
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TagValidator:validateTags(text, singleTags, pairedTags, tertiaryTags, optionalPairedTags, optionalSingleTags)
  singleTags = singleTags or {}
  pairedTags = pairedTags or {}
  tertiaryTags = tertiaryTags or {}
  optionalPairedTags = optionalPairedTags or TagValidator.DEFAULT_OPTIONAL_PAIRED_TAGS
  optionalSingleTags = optionalSingleTags or {}

  local singleTagStates = getInitialTagStates(singleTags)
  local pairedTagStates = getInitialTagStates(pairedTags)
  local tertiaryTagStates = getInitialTagStates(tertiaryTags)
  local optionalPairedTagStates = getInitialTagStates(optionalPairedTags)
  local optionalSingleTagStates = getInitialTagStates(optionalSingleTags)

  for foundTag in text:gmatch(TAG_REGEX) do
    local tag = foundTag:gsub(MARKUP_REGEX,"") -- get rid of /, $

    local isSingleTag = foundTag:sub(-1,-1) == "/"
    local isOpenTag   = foundTag == tag
    local isCloseTag  = foundTag:sub(1,1) == "/"
    local isMiddleTag = foundTag:sub(1,1) == "$" and foundTag:sub(-1,-1) == "$"

    -- Single tags
    if isSingleTag and singleTagStates[tag] then
      self:expectTrue(singleTagStates[tag] == STATE_INITIAL, "More than one occurrence found for tag: \"" .. tag .. "\"")
      singleTagStates[tag] = STATE_FOUND

    -- Paired tags
    elseif isOpenTag and pairedTagStates[tag] then
      self:expectTrue(pairedTagStates[tag] == STATE_INITIAL, "Two consecutive open tags found for: \"" .. tag .. "\"")
      pairedTagStates[tag] = STATE_OPEN

    elseif isCloseTag and pairedTagStates[tag] then
      self:expectTrue(pairedTagStates[tag] == STATE_OPEN, "Close tag before open tag found for: \"" .. tag .. "\"")
      pairedTagStates[tag] = STATE_CLOSED

    -- Optional paired tags
    elseif isOpenTag and optionalPairedTagStates[tag] then
      self:expectTrue(optionalPairedTagStates[tag] == STATE_INITIAL, "Two consecutive open tags found for: \"" .. tag .. "\"")
      optionalPairedTagStates[tag] = STATE_OPEN

    elseif isCloseTag and optionalPairedTagStates[tag] then
      self:expectTrue(optionalPairedTagStates[tag] == STATE_OPEN, "Close tag before open tag found for: \"" .. tag .. "\"")
      optionalPairedTagStates[tag] = STATE_INITIAL -- Allow multiple occurrences for optional tags only

    -- Optional single tags
    elseif isSingleTag and optionalSingleTagStates[tag] then
      -- do nothing but definitely allow single tags

    -- Tertiary tags
    elseif isOpenTag and tertiaryTagStates[tag] then
      self:expectTrue(tertiaryTagStates[tag] == STATE_INITIAL, "Two consecutive open tags found for: \"" .. tag .. "\"")
      tertiaryTagStates[tag] = STATE_OPEN

    elseif isMiddleTag and tertiaryTagStates[tag] then
      self:expectTrue(tertiaryTagStates[tag] == STATE_OPEN, "Middle tag before open tag found for: \"" .. tag .. "\"")
      tertiaryTagStates[tag] = STATE_FOUND_TERTIARY

    elseif isCloseTag and tertiaryTagStates[tag] then
      self:expectTrue(tertiaryTagStates[tag] == STATE_FOUND_TERTIARY, "Close tag before middle tag found for: \"" .. tag .. "\"")
      tertiaryTagStates[tag] = STATE_CLOSED

    -- Other tags
    else
      self:expectTrue(false, "Unexpected tag found: \"" .. foundTag .. "\"")
    end
  end

  for tag, state in pairs(singleTagStates) do
    self:expectTrue(state == STATE_FOUND, "Tag not found for: \"" .. tag .. "\"")
  end

  for tag, state in pairs(pairedTagStates) do
    self:expectTrue(state ~= STATE_INITIAL, "Opening tag not found for: \"" .. tag .. "\"")
    self:expectTrue(state == STATE_CLOSED, "Closing tag not found for: \"" .. tag .. "\"")
  end

  for tag, state in pairs(optionalPairedTagStates) do
    self:expectTrue(state ~= STATE_OPEN, "Closing tag not found for: \"" .. tag .. "\"")
  end

  for tag, state in pairs(tertiaryTagStates) do
    self:expectTrue(state ~= STATE_INITIAL, "Opening tag not found for: \"" .. tag .. "\"")
    self:expectTrue(state ~= STATE_OPEN, "Middle tag not found for: \"" .. tag .. "\"")
    self:expectTrue(state == STATE_CLOSED, "Closing tag not found for: \"" .. tag .. "\"")
  end
end

function TagValidator:expectHasTag(value, tag, message)
  local tagContents = TagUtils.extractTagContents(value, tag)
  return self:expectNotNil(tagContents, nil, message or "Tag [" .. tag .. "] not found in [" .. tostring(value) .. "]")
end

function TagValidator:expectHasContentsInTag(value, tag, message)
  if not self:expectHasTag(value, tag, message) then
    return false
  end
  local tagContents = TagUtils.extractTagContents(value, tag)
  return self:expectNotEmpty(tagContents, nil, message or "Tag [" .. tag .. "] contains no data in [" .. tostring(value) .. "]")
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TagValidator.make(conceptValidator)
  Mixin.mix(conceptValidator, TagValidator)

  return conceptValidator
end
