--- Validator

class("Validator")
require("Mixin")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

Validator.VALIDATION_ERROR_CODE = 1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function reset(self)
  self.passed = true
  self.message = ""
end

local function appendMessage(self, msg)
  if self.message ~= "" then
    self.message = self.message .. "\n"
  end
  self.message = self.message .. "-" .. msg
end

local function throwValidationError()
  error({code = VALIDATION_ERROR_CODE})
end

local function createAssertFunction(expectFunction)
  return function(...)
    if not expectFunction(...) then
      throwValidationError()
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Validator:runValidation(validationFunction)

  local function errorHandler(errorObject)
    local errorIsValidationAssertion = errorObject and errorObject.code == Validator.VALIDATION_ERROR_CODE
    if not errorIsValidationAssertion then
      local traceback = debug.traceback()
      local errorMessage = StringUtils.tostring(errorObject) .. "\n" .. tostring(traceback)
      self.passed = false
      appendMessage(self, errorMessage)
    end
  end

  reset(self)
  xpcall(validationFunction, errorHandler)
end

-- Get state --------------------------

function Validator:getPassed()
  return self.passed
end

function Validator:getMessage()
  return self.message
end

-- Test methods -----------------------

function Validator:expectFailed(message)
  self.passed = false
  appendMessage(self, message)
  return false
end

-- Boolean conditions

function Validator:expectTrue(condition, message)
  if not condition then
    self:expectFailed(message)
  end
  return condition
end

function Validator:expectFalse(condition, message)
  return self:expectTrue(not condition, message)
end

-- Equality

function Validator:expectEqual(actualValue, expectedValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected: %s)", message, tostring(actualValue), tostring(expectedValue))
  return self:expectTrue(actualValue == expectedValue, message)
end

function Validator:expectNotEqual(actualValue, forbiddenValue, message)
  message = message or ""
  message = string.format("%s (forbidden value: %s)", message, tostring(actualValue), tostring(forbiddenValue))
  return self:expectTrue(actualValue ~= forbiddenValue, message)
end

-- Arithmetic comparison

function Validator:expectLessThan(actualValue, maximumValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected maximum: %s)", message, tostring(actualValue), tostring(maximumValue))
  return self:expectTrue(actualValue < maximumValue, message)
end

function Validator:expectGreaterThan(actualValue, minimumValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected minimum: %s)", message, tostring(actualValue), tostring(minimumValue))
  return self:expectTrue(actualValue > minimumValue, message)
end

function Validator:expectLessOrEqual(actualValue, maximumValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected maximum: %s)", message, tostring(actualValue), tostring(maximumValue))
  return self:expectTrue(actualValue <= maximumValue, message)
end

function Validator:expectGreaterOrEqual(actualValue, minimumValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected minimum: %s)", message, tostring(actualValue), tostring(minimumValue))
  return self:expectTrue(actualValue >= minimumValue, message)
end

function Validator:expectNumberInRange(value, min, max, fieldName)
  min = min or -math.huge
  max = max or math.huge
  fieldName = fieldName or "field"

  local isNumber = type(value) == "number"
  local result = self:expectTrue(isNumber, "Value for " .. fieldName .. " should be a number, was: " .. tostring(value))

  if isNumber then
    result = result and self:expectTrue(value >= min, "Value for " .. fieldName .. " should be at least " .. tostring(min) .. ", was: " .. tostring(value))
    result = result and self:expectTrue(value <= max, "Value for " .. fieldName .. " should be at most " .. tostring(max) .. ", was: " .. tostring(value))
  end

  return result
end

-- Float arithmetic comparison

function Validator:expectFloatEqual(actualValue, expectedValue, message)
  message = message or ""
  message = string.format("%s (found: %s, expected: %s)", message, tostring(actualValue), tostring(expectedValue))

  -- Using lua's builtin float-to-string conversion is a neat way of comapring floats, but should
  -- it misbehave we should replace it by a more conventional `math.abs(x - y) < EPSILON`.
  --(http://stackoverflow.com/a/12320854)
  return self:expectTrue(tostring(actualValue) == tostring(expectedValue), message)
end

-- Nil/empty checking

function Validator:expectNil(value, fieldName, message)
  fieldName = fieldName or ""
  return self:expectTrue(value == nil, message or 'Field "' .. fieldName .. '" should be nil, found: ' .. tostring(value))
end

function Validator:expectNotNil(value, fieldName, message)
  fieldName = fieldName or ""
  return self:expectTrue(value ~= nil, message or 'Field "' .. fieldName .. '" should not be nil')
end

function Validator:expectEmpty(value, fieldName, message)
  fieldName = fieldName or ""
  return self:expectTrue(value == "", message or 'Field "' .. fieldName .. '" should be empty, found: ' .. tostring(value))
end

function Validator:expectNotEmpty(value, fieldName, message)
  fieldName = fieldName or ""
  if not self:expectNotNil(value, fieldName, message) then
    return false
  end
  return self:expectTrue(value ~= "", message or 'Field "' .. fieldName .. '" should not be empty')
end

-- Key presence

function Validator:expectHasKey(object, key, allowEmpty, message)
  allowEmpty = allowEmpty or false

  if allowEmpty then
    return self:expectNotNil(object[key], key, message)
  else
    return self:expectNotEmpty(object[key], key, message)
  end
end

function Validator:expectHasKeys(object, keys, allowEmpty, message)
  local allKeysFound = true
  for _, key in ipairs(keys) do
    allKeysFound = allKeysFound and self:expectHasKey(object, key, allowEmpty, message)
  end
  return allKeysFound
end

-- Item belonging

function Validator:expectIsContained(value, array, fieldName, message)
  local isContained = TableUtils.contains(array, value)
  local errorMessage = message or string.format("Incorrect value for %s, found: %s, expected one of: %s", fieldName, value, StringUtils.tostring(array))
  return self:expectTrue(isContained, errorMessage)
end

function Validator:expectValueIsInSet(value, containerSet, valuesName, containerName)
  local errorMessage = string.format("Value '%s' from field '%s' should be contained in field '%s'", value, valuesName, containerName)
  self:expectTrue(containerSet[value], errorMessage)
end

function Validator:expectValueIsNotInSet(value, containerSet, valuesName, containerName)
  local errorMessage = string.format("Value '%s' from field '%s' should not be contained in field '%s'", value, valuesName, containerName)
  self:expectFalse(containerSet[value], errorMessage)
end

function Validator:expectValuesAreInSet(valuesSet, containerSet, valuesName, containerName)
  for value, _ in pairs(valuesSet) do
    self:expectValueIsInSet(value, containerSet, valuesName, containerName)
  end
end

function Validator:expectValuesAreNotInSet(valuesSet, containerSet, valuesName, containerName)
  for value, _ in pairs(valuesSet) do
    self:expectValueIsNotInSet(value, containerSet, valuesName, containerName)
  end
end

-- Misc

function Validator:expectNoDuplicates(array, fieldName, message)
  local duplicates = TableUtils.getDuplicatesInArray(array)
  self:expectEqual(#duplicates, 0, message or string.format("Duplicates found in field %s: %s", fieldName, StringUtils.tostring(duplicates)))
end

function Validator:expectType(value, expectedType, fieldName, message)
  fieldName = fieldName or ""
  return self:expectEqual(type(value), expectedType, message or string.format("Field %s does not have expected type. Found: ", fieldName, type(value)))
end

function Validator:expectNoError(functionThatMayThrow)
  local function errorHandler(errorObject)
    local errorMessage = StringUtils.tostring(errorObject)
    self:expectFailed(errorMessage)
  end

  xpcall(functionThatMayThrow, errorHandler)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function initMetatable(self)
  local mt = {}
  mt.__index = function(validator, key)
    if key:find("assert") == 1 then
      local expectFunctionName = key:gsub("assert", "expect")
      return createAssertFunction(rawget(validator, expectFunctionName))
    else
      return rawget(validator, key)
    end
  end
  setmetatable(self, mt)
end

-- Constructor ------------------------

function Validator.new()
  local self = {}
  Mixin.mix(self, Validator)

  initMetatable(self)
  reset(self)

  return self
end
