--- AudioRequiredOverlay
-- @author Created by karl on November 18, 2014.

-- ===========================================================================
class("AudioRequiredOverlay")
-- ===========================================================================

require("widgets.TapToContinue")
require("widgets.AudioRequiredOverlayButton")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEBUG_OVERRIDE_TOUCH_AREA_SIZE = 8
local DEBUG_OVERRIDE_TOUCH_COUNT_TO_SKIP = 2

local CIRCLE_RADIUS = Screen.getWidthPart(0.25)

local COLOR_OVERLAY = {0, 0, 0, 0.7}
local COLOR_AUDIO_REQUIRED = {1, 1, 1, 0.4}
local COLOR_AUDIO_INSTRUCTIONS = {1, 1, 1}
local COLOR_AUDIO_MUTE = {1, 1, 1, 0.2}
local COLOR_CIRCLE_FILL = {1, 1, 1, 0.15}
local COLOR_CIRCLE_EDGE = {1, 1, 1, 0.7}

local SOUND_EFFECT = "shared_audio_required_tone"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local CirclePulse = privateClass()

function CirclePulse:animate()
  local duration = 1
  return AHStaggeredSpawn.new(duration*0.5,
    AHSpawn.new(
      self.edgeCircle:newRadiusToAction(duration, AHEaseSineOut.new(), CIRCLE_RADIUS),
      self.filledCircle:newRadiusToAction(duration, AHEaseSineOut.new(), CIRCLE_RADIUS)
    ),
    AHOpacityTo.new(self, duration, AHEaseSineOut.new(), 0)
  ):start()
end

function CirclePulse.new()
  local self = Group.new(2, 2)
  Mixin.mix(self, CirclePulse)
  self.filledCircle = FilledCircle.new(1, COLOR_CIRCLE_FILL)
  self.edgeCircle = HollowCircle.new(1, COLOR_CIRCLE_EDGE, 0.4, 1)
  self:insertAtCenter(self.filledCircle)
  self:insertAtCenter(self.edgeCircle)

  self:linkColorAttrToChildren(true)

  return self
end

local function showPulse(self)
  return runInCoroutine(function()
    local pulse = CirclePulse.new()
    pulse:setLoc(self.pulseX, self.pulseY)
    self:insert(pulse)
    wait(pulse:animate())
    pulse:removeSelf()
  end)
end

local function startPulseAnimationLoop(self)
  self.pulseAnimationLoop = runInCoroutine(function()
    wait(AHOpacityTo.new(self, 0.5, AHEaseSineOut.new(), 1):start())
    while true do
      Sound.playLongformSound(SOUND_EFFECT)
      showPulse(self)
      waitFor(0.4)
      showPulse(self)
      waitFor(1.5)
    end
  end)
end

local function stopPulseAnimationLoop(self)
  self.pulseAnimationLoop:stop()
end

local function disableButtons(self)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(false)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function AudioRequiredOverlay:show()
  startPulseAnimationLoop(self)
end

function AudioRequiredOverlay:dismiss()
  disableButtons(self)
  stopPulseAnimationLoop(self)
  self:linkColorAttrToChildren()
  wait(AHOpacityTo.new(self, 0.5, AHEaseSineOut.new(), 0.0):start())
  self:removeSelf()
  if (self.callback ~= nil) then
    self.callback()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createBeginButton(self)
  local function beginCallback()
    self:dismiss()
  end
  return AudioRequiredOverlayButton.new(Localizer.get("AudioRequired.BEGIN"), beginCallback)
end

local function createSwitchGameButton(self)
  local function switchGameCallback()
    disableButtons(self)
    GameSession.notifySwapGamesRequested()
  end
  local useWhiteColor = true
  return AudioRequiredOverlayButton.new(Localizer.get("AudioRequired.SWITCH_GAME"), switchGameCallback, useWhiteColor)
end

-- Tapping 3 times in the corner will bypass this screen
local function createTestingOverrideGesture(self)
  local onTestingOverrideGestureCompleted = function()
    Log.debug("Audio Required Screen bypassed via debug gesture.")
    self:dismiss()
  end

  local overrideTouchUpCount = 0

  local onTouchUp = function()
    overrideTouchUpCount = overrideTouchUpCount + 1
    if (overrideTouchUpCount == DEBUG_OVERRIDE_TOUCH_COUNT_TO_SKIP) then
      onTestingOverrideGestureCompleted()
    end
  end

  local testingOverrideButton = Button.new(DEBUG_OVERRIDE_TOUCH_AREA_SIZE, DEBUG_OVERRIDE_TOUCH_AREA_SIZE, nil, onTouchUp)
  testingOverrideButton:setAnchorPointTopLeft()
  self:insertAtTopLeft(testingOverrideButton)
end

-- Constructor ------------------------

function AudioRequiredOverlay.new(callback)
  local self = Group.newFullScreen()
  Mixin.mix(self, AudioRequiredOverlay)

  local bg = FilledRectangle.new(self:getWidth(), self:getHeight(), COLOR_OVERLAY)
  self:insertAtCenter(bg)

  local font = FontFaces.DIN_OT_BOLD
  local fontSize = Screen.getWidthPart(0.045)
  local spacerH = fontSize * 3

  local column = Column.new()

  local icon = Sprite.newWithName("shared_audio_icon.png")

  local audioRequiredLabel = DeprecatedLabel.newSingleLine(Localizer.get("AudioRequired.REQUIRES_AUDIO"), font, fontSize)
  audioRequiredLabel:setColor(unpack(COLOR_AUDIO_REQUIRED))

  local instructions1 = DeprecatedLabel.newSingleLine(Localizer.get("AudioRequired.HEAR_CLEARLY_1"), font, fontSize)
  instructions1:setColor(unpack(COLOR_AUDIO_INSTRUCTIONS))
  local instructions2 = DeprecatedLabel.newSingleLine(Localizer.get("AudioRequired.HEAR_CLEARLY_2"), font, fontSize)
  instructions2:setColor(unpack(COLOR_AUDIO_INSTRUCTIONS))

  self.buttons = {}

  local beginButton = createBeginButton(self)
  table.insert(self.buttons, beginButton)

  column:insert(audioRequiredLabel)
  column:insertSpacer(self:getWidthPart(0.24))
  column:insert(icon)
  column:insertSpacer(self:getWidthPart(0.22))
  column:insert(instructions1)
  column:insert(instructions2)
  column:insertSpacer(self:getWidthPart(0.06))
  column:insert(beginButton)

  if (GameSession.canGameBeSwitched()) then
    local switchGameButton = createSwitchGameButton(self)
    table.insert(self.buttons, switchGameButton)

    column:insertSpacer(self:getWidthPart(0.02))
    column:insert(switchGameButton)
  end

  column:insertSpacer(self:getWidthPart(0.06))
  column:alignChildren()
  self:insertAtCenter(column)

  column:linkColorAttrToChildren()
  self:linkColorAttrToChildren()

  self.pulseX, self.pulseY = self:transformObjectLoc(icon)

  createTestingOverrideGesture(self)

  self:setOpacity(0)

  self.callback = callback

  return self
end
