--- A basic button widget
-- @author Created by Alexey Chernikov on April 19, 2013

-- ===========================================================================
class("Button")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

--- Forward a button event function to the specified target
local function forwardButtonEvent(target, aFunction)
  if not aFunction then
    return nil
  end

  return function(button, event)
    return aFunction(target, event)
  end
end

local function callIfExists(functionToCall, ...)
  if functionToCall then
    functionToCall(...)
  end
end

local function forwardTouchEnabledMethods(group, virtualButton)
  group.isTouchEnabled = function(self) return virtualButton:isTouchEnabled() end
  group.setTouchEnabled = function(self, isEnabled) virtualButton:setTouchEnabled(isEnabled) end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function Button.createButtonTouchHandler(onTouchBegan, onTouchMoved, onTouchEnded, onTouchCancel, shouldStopProcessingOnCancel)
  local startX, startY
  local isCancelled = false
  shouldStopProcessingOnCancel = shouldStopProcessingOnCancel or false

  local handler = function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      isCancelled = false
      startX = event.x
      startY = event.y
      callIfExists(onTouchBegan, event)
    elseif (event.phase == TouchManager.TOUCH_MOVED) and not isCancelled then
      -- This is needed to plug a hole where MOVED can be sent before BEGAN.
      startX = startX or event.x
      startY = startY or event.y

      if event:isWithinAccidentalMovementRange(startX, startY) then
        callIfExists(onTouchMoved, event)

        return TouchEvent.HANDLING.CONTINUE
      else
        if shouldStopProcessingOnCancel then
          isCancelled = true
        end

        callIfExists(onTouchCancel, event)

        return TouchEvent.HANDLING.HAND_OFF
      end

    elseif (event.phase == TouchManager.TOUCH_ENDED) and not isCancelled then
      callIfExists(onTouchEnded, event)
    end

    return TouchEvent.HANDLING.CONTINUE
  end

  return handler
end

--- Turns any group into a button whose touchable area is the bounds of the group.
-- @param group - the group to buttonify
-- @param onTouchDown - (optional) function that will be called when the button is touched
-- @param onTouchUp - (optional) function that will be called when the touch is released (only tracks initial touch which can be cancelled by movement)
-- @param onTouchCancel - (optional) function that will be called when the touch moves too far from it's initial location
function Button.makeButton(group, onTouchDown, onTouchUp, onTouchCancel)
  Mixin.mix(group, Button)

  local touchDownWrapper = function(event)
    callIfExists(onTouchDown, group, event)
  end

  local touchUpWrapper = function(event)
    callIfExists(onTouchUp, group, event)
  end

  local touchCancelWrapper = function(event)
    callIfExists(onTouchCancel, group, event)
  end

  local shouldStopProcessingOnCancel = true
  local handlerFunction = Button.createButtonTouchHandler(touchDownWrapper,
                                                          nil,
                                                          touchUpWrapper,
                                                          touchCancelWrapper,
                                                          shouldStopProcessingOnCancel)
  TouchManager.addTouchEnabledMethods(group, handlerFunction)
  group:setTouchEnabled(true)
  return group
end

--- Adds an invisible touchable button to the top Z position of a group, of the specified size, relative to the center of the group.
-- @param group - the group to buttonify
-- @param width - (optional) The width of the virtual button. Defaults to the group width.
-- @param height - (optional) The height of the virtual button. Defaults to the group height.
-- @param onTouchDown - (optional) function that will be called when the button is touched
-- @param onTouchUp - (optional) function that will be called when the touch is released (only tracks initial touch which can be cancelled by movement)
-- @param onTouchCancel - (optional) function that will be called when the touch moves too far from its initial location
-- @return The added virtual button.
function Button.addVirtualButton(group, width, height, onTouchDown, onTouchUp, onTouchCancel)
  local groupW, groupH = group:getDims()
  width = width or groupW
  height = height or groupH
  onTouchDown = forwardButtonEvent(group, onTouchDown)
  onTouchUp = forwardButtonEvent(group, onTouchUp)
  onTouchCancel = forwardButtonEvent(group, onTouchCancel)

  local virtualButton = Button.new(width, height, onTouchDown, onTouchUp, onTouchCancel)
  group:insertAtCenter(virtualButton, group:getTopZIndex() + 1)

  forwardTouchEnabledMethods(group, virtualButton)

  return virtualButton
end

-- Constructor ------------------------

--- Make a new button.
-- @param width - The width of the button
-- @param height - The height of the button
-- @param onTouchDown - (optional) function that will be called when the button is touched
-- @param onTouchUp - (optional) function that will be called when the touch is released (only tracks initial touch which can be cancelled by movement)
-- @param onTouchCancel - (optional) function that will be called when the touch moves too far from it's initial location
-- @return The new button.
function Button.new(width, height, onTouchDown, onTouchUp, onTouchCancel)
  local self = Group.new(width, height)
  Button.makeButton(self, onTouchDown, onTouchUp, onTouchCancel)
  return self
end
