--- Drag And Drop Handler
-- Coordinates dragging and dropping of objects into containers.
--
-- Drag and drop containers and objects must adhere to the following protocol:
--
-- Container:
-- canAcceptDNDObject(object): Returns true if the object is acceptable to the container.
-- notifyDNDEnter(object): Notify that an object is hovering over the container.
-- notifyDNDLeave(object): Notify that an object is no longer hovering over the container.
-- notifyDNDMove(object):Notify that an object has moved while hovering over the container.
-- notifyDNDDrop(object): Notify that an object has been dropped on the container.
--
-- Object:
-- notifyDNDPickUp(): Called when the object is picked up.
-- notifyDNDDrop(container): Called when the object is dropped. container will only be present if it was dropped inside.
-- Note: It is up to the object to return to its original pre-drag position, if it so desires.


-- ===========================================================================
class("DragAndDropHandler")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isWithinObjectBounds(object, x, y)
  x, y = object:worldToModel(x, y)
  local xMin, yMin, zMin, xMax, yMax, zMax = object:getBounds()
  if (xMin == nil or xMax == nil or yMin == nil or yMax == nil) then
    return false
  end
  return (xMin <= x and x <= xMax and yMin <= y and y <= yMax)
end

local function getTouchedObject(self, locX, locY)
  for _, object in ipairs(self.objects) do
    if isWithinObjectBounds(object, locX, locY) then
      return object
    end
  end
  return nil
end

local function getCompatibleContainer(self, locX, locY, object)
  for _, container in ipairs(self.containers) do
    if isWithinObjectBounds(container, locX, locY) then
      if container:canAcceptDNDObject(object) then
        return container
      end
    end
  end
  return nil
end

local function setObjectDragDelta(self, object, eventX, eventY)
  local objectX, objectY = object:getLoc()
  local deltaX, deltaY = objectX - eventX, objectY - eventY

  self.dragDeltaX = deltaX
  self.dragDeltaY = deltaY
end

local function moveObject(self, object, eventX, eventY)
  local objectX, objectY = eventX + self.dragDeltaX, eventY + self.dragDeltaY
  object:setLoc(objectX, objectY)
end

local function pickUpObject(self, eventX, eventY)
  local object = getTouchedObject(self, eventX, eventY)
  if not object then
    return
  end

  setObjectDragDelta(self, object, eventX, eventY)

  self.draggingObject = object
  self.draggingObject:notifyDNDPickUp()

  self.currentDNDContainer = getCompatibleContainer(self, eventX, eventY, object)
  if self.currentDNDContainer then
    self.currentDNDContainer:notifyDNDEnter(object)
  end
end

local function dragCurrentObject(self, eventX, eventY)
  if not self.draggingObject then
    return
  end

  local object = self.draggingObject
  moveObject(self, object, eventX, eventY)

  if not self.currentDNDContainer then
    self.currentDNDContainer = getCompatibleContainer(self, eventX, eventY, object)
    if self.currentDNDContainer then
      self.currentDNDContainer:notifyDNDEnter(object)
    end
  else
    if isWithinObjectBounds(self.currentDNDContainer, eventX, eventY) then
      self.currentDNDContainer:notifyDNDMove(object, eventX, eventY)
    else
      self.currentDNDContainer:notifyDNDLeave(object)
      self.currentDNDContainer = nil
    end
  end
end

local function dropCurrentObject(self, eventX, eventY)
  if not self.draggingObject then
    return
  end

  local object = self.draggingObject
  moveObject(self, object, eventX, eventY)

  if self.currentDNDContainer then
    self.currentDNDContainer:notifyDNDDrop(object)
  end
  object:notifyDNDDrop(self.currentDNDContainer)

  self.draggingObject = nil
  self.currentDNDContainer = nil
end

local function getTouchHandler(self)
  return function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      return pickUpObject(self, event.x, event.y)
    elseif (event.phase == TouchManager.TOUCH_MOVED) then
      return dragCurrentObject(self, event.x, event.y)
    elseif (event.phase == TouchManager.TOUCH_ENDED) then
      return dropCurrentObject(self, event.x, event.y)
    end
  end
end

local function insertWithPriority(array, object, priority, priorityFieldName)
  priority = priority or 0
  object[priorityFieldName] = priority
  for index, entry in ipairs(array) do
    local currentPriority = entry[priorityFieldName]
    if not currentPriority or currentPriority > priority then
      table.insert(array, index, object)
      return
    end
  end
  table.insert(array, object)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Start/stop managing objects & containers

function DragAndDropHandler:addObject(object, priority)
  insertWithPriority(self.objects, object, priority, "_DND_objectPriority")
end

function DragAndDropHandler:removeObject(object)
  TableUtils.removeValueFromArray(self.objects, object)

  if self.draggingObject == object then
    self.draggingObject = nil
  end
end

function DragAndDropHandler:addContainer(object, priority)
  insertWithPriority(self.containers, object, priority, "_DND_containerPriority")
end

function DragAndDropHandler:removeContainer(object)
  TableUtils.removeValueFromArray(self.containers, object)

  if self.currentDNDContainer == object then
    self.currentDNDContainer = nil
  end
end

function DragAndDropHandler:setEnabled(enabled)
  if (enabled) then
    local touchHandler = self.touchHandler
    TouchManager.addTouchHandler(self, touchHandler)
  else
    TouchManager.removeTouchHandler(self)
  end
end

function DragAndDropHandler:clearContainers()
  self.currentDNDContainer = nil
  self.containers = {}
end

function DragAndDropHandler:clearObjects()
  self.draggingObject = nil
  self.objects = {}
end

function DragAndDropHandler:clear()
  self:clearContainers()
  self:clearObjects()
end

function DragAndDropHandler:forceDrop()
  local draggingObject = self.draggingObject

  if (draggingObject == nil) then
    return
  end

  local x, y = draggingObject:getLoc()
  dropCurrentObject(self, x, y)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function DragAndDropHandler.new()
  local self = Group.new(Screen.width, Screen.height)
  self:setAnchorPointBottomLeft()
  Screen.rootGroup:insert(self)
  Mixin.mix(self, DragAndDropHandler)

  self.objects = {}
  self.containers = {}

  self.touchHandler = getTouchHandler(self)
  self:setEnabled(true)

  return self
end
