--- DragHelpOverlay
-- @author Created by amorant on August 26, 2014.

-- ===========================================================================
class("DragHelpOverlay")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5

local DEFAULT_MESSAGE_RELATIVE_Y = 0.27
local DEFAULT_ARROW_RELATIVE_X_OFFSET = 0.05
local DEFAULT_ARROW_SCALE = 1

local HOVER_AMOUNT = HEIGHT * 0.008

-- Message Label ----------------------

local MESSAGE_LABEL_FONT = FontFaces.DIN_OT_MEDIUM
local DEFAULT_MESSAGE_LABEL_FONT_SIZE = 20.0

-- Dark Rectangle ---------------------

local DARK_RECTANGLE_COLOR = {0.0, 0.0, 0.0}
local DARK_RECTANGLE_MAX_OPACITY = 0.11

-- Z-Indices --------------------------

local Z_DARK_RECTANGLE = 10
local Z_MESSAGE_LABEL = 20
local Z_ARROW = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createFadeMessageAction(self, destinationOpacity)
  local darkRectangle = self.darkRectangle
  local darkRectangleDestinationOpacity = DARK_RECTANGLE_MAX_OPACITY * destinationOpacity

  local fadeAction = AHSpawn.new(
                                 AHOpacityTo.new(self, 0.2, nil, destinationOpacity),
                                 AHOpacityTo.new(darkRectangle, 0.2, nil, darkRectangleDestinationOpacity)
                                 )
  return fadeAction
end

local function createHoverAction(self)
  local leftArrow = self.leftArrow
  local rightArrow = self.rightArrow

  local hoverEase = AHEaseSimpleInOut.new(2.0)
  local hoverDuration = 0.7

  local leftArrowAction = AHSequence.new(
                                         AHLocationYTo.new(leftArrow, hoverDuration, hoverEase, self.messageY + HOVER_AMOUNT),
                                         AHLocationYTo.new(leftArrow, hoverDuration, hoverEase, self.messageY - HOVER_AMOUNT)
                                         )
  local rightArrowAction = AHSequence.new(
                                         AHLocationYTo.new(rightArrow, hoverDuration, hoverEase, self.messageY + HOVER_AMOUNT),
                                         AHLocationYTo.new(rightArrow, hoverDuration, hoverEase, self.messageY - HOVER_AMOUNT)
                                         )

  local hoverAction = AHLoop.new(
                                 AHSpawn.new(
                                             leftArrowAction,
                                             rightArrowAction
                                             )
                                 )
  return hoverAction
end

local function setElementsVisible(self, isVisible)
  local elements = self.elements
  for i, element in ipairs(elements) do
    element:setVisible(isVisible)
  end
end

local function stopFadeAction(self)
  if (self.fadeAction ~= nil) then
    self.fadeAction:stop()
  end
end

local function stopHoverAction(self)
  if (self.hoverAction ~= nil) then
    self.hoverAction:stop()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function DragHelpOverlay:show()
  setElementsVisible(self, true)

  local destinationOpacity = 1.0
  local fadeMessageAction = createFadeMessageAction(self, destinationOpacity)
  self.fadeAction = fadeMessageAction

  local hoverAction = createHoverAction(self)
  self.hoverAction = hoverAction
  hoverAction:start()

  return fadeMessageAction:start()
end

function DragHelpOverlay:hide()
  stopFadeAction(self)

  local destinationOpacity = 0.0
  local fadeMessageAction = createFadeMessageAction(self, destinationOpacity)

  local fadeAction = AHSequence.new(
                                    fadeMessageAction,
                                    function()
                                      self:removeSelf()
                                      stopHoverAction(self)
                                      self.messageLabel:releaseTexture()
                                    end
                                    )
  self.fadeAction = fadeAction

  return fadeAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createElements(self, message, messageY, arrowRelativeXOffset, messageLabelFontSize, arrowScale)
  local elements = {}

  local darkRectangle = FilledRectangle.new(WIDTH, HEIGHT, DARK_RECTANGLE_COLOR)
  darkRectangle:setVisible(false)
  darkRectangle:setOpacity(0.0)
  self:insertAtCenter(darkRectangle, Z_DARK_RECTANGLE)
  table.insert(elements, darkRectangle)
  self.darkRectangle = darkRectangle

  local messageLabel = DeprecatedLabel.newSingleLine(message, MESSAGE_LABEL_FONT, messageLabelFontSize)
  messageLabel:setVisible(false)
  messageLabel:inheritOpacity(self)
  messageLabel:setLoc(CENTER_X, messageY)
  self:insert(messageLabel, Z_MESSAGE_LABEL)
  table.insert(elements, messageLabel)
  self.messageLabel = messageLabel

  local leftArrow = Sprite.newWithName("shared_arrow.png")
  local leftArrowX = arrowRelativeXOffset * WIDTH
  leftArrow:setScl(arrowScale)
  leftArrow:setVisible(false)
  leftArrow:inheritOpacity(self)
  leftArrow:setLoc(leftArrowX, messageY - HOVER_AMOUNT)
  self:insert(leftArrow, Z_ARROW)
  table.insert(elements, leftArrow)
  self.leftArrow = leftArrow

  local rightArrow = Sprite.newWithName("shared_arrow.png")
  local rightArrowX = (1 - arrowRelativeXOffset) * WIDTH
  rightArrow:setScl(arrowScale)
  rightArrow:setVisible(false)
  rightArrow:inheritOpacity(self)
  rightArrow:setLoc(rightArrowX, messageY - HOVER_AMOUNT)
  self:insert(rightArrow, Z_ARROW)
  table.insert(elements, rightArrow)
  self.rightArrow = rightArrow

  self:setOpacity(0.0)

  self.elements = elements
end

function DragHelpOverlay:getHoverAmount()
  return HOVER_AMOUNT
end

-- Constructor ------------------------

function DragHelpOverlay.new(message, messageRelativeY, arrowRelativeXOffset, messageLabelFontSize, arrowScale)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, DragHelpOverlay)

  messageRelativeY = messageRelativeY or DEFAULT_MESSAGE_RELATIVE_Y
  arrowRelativeXOffset = arrowRelativeXOffset or DEFAULT_ARROW_RELATIVE_X_OFFSET
  messageLabelFontSize = messageLabelFontSize or DEFAULT_MESSAGE_LABEL_FONT_SIZE
  arrowScale = arrowScale or DEFAULT_ARROW_SCALE

  local messageY = messageRelativeY * HEIGHT
  self.messageY = messageY

  createElements(self, message, messageY, arrowRelativeXOffset, messageLabelFontSize, arrowScale)

  return self
end
