-- Image based lives display


-- ===========================================================================
class("LivesDisplay")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ENABLED_OPACITY = 1.0
local DISABLED_OPACITY = 0.2
local LIVES_PADDING = 3
local LIFE_IMAGE_FILENAME = "shared_life_image.png"
local FLICKER_DELAY = 0.05
local FLICKER_COUNT = 2

local DEFAULT_COLOR = {1.0, 1.0, 1.0, 1.0}
local DEFAULT_OPACITY = 1.0

local DEFAULT_LIVES_COUNT = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setLifeImageEnabled(lifeImage, doEnable)
  local opacity = ENABLED_OPACITY

  if (not doEnable) then
    opacity = DISABLED_OPACITY
  end

  lifeImage:setOpacity(opacity)
end

local function disableLifeImage(lifeImage)
  local function flickerOut()
    for i = 1, FLICKER_COUNT do
      lifeImage:setOpacity(DISABLED_OPACITY)
      waitFor(FLICKER_DELAY)
      lifeImage:setOpacity(ENABLED_OPACITY)
      waitFor(FLICKER_DELAY)
    end
    lifeImage:setOpacity(DISABLED_OPACITY)
  end
  runInCoroutine(flickerOut)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LivesDisplay:getLives()
  return self.lives
end

function LivesDisplay:setLives(lives)
  assert((lives >= 0) and (lives <= self.maxLives), "Set lives to number outside of range for possible lives")

  for i, lifeImage in ipairs(self.lifeImageCollection) do
    local isLifeEnabled = (i <= lives)
    setLifeImageEnabled(lifeImage, isLifeEnabled)
  end

  self.lives = lives
end

function LivesDisplay:decrement()
  if (self.lives > 0) then
    local lifeToRemove = self.lifeImageCollection[self.lives]
    disableLifeImage(lifeToRemove)
    self.lives = self.lives - 1
  end
end

-- UIOverlay Widget -------------------

function LivesDisplay:getKeyColor()
  return self.keyColor
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ----------------------------
function LivesDisplay.new(numberOfLives, color)
  numberOfLives = numberOfLives or DEFAULT_LIVES_COUNT

  assert(numberOfLives > 0, "Number of lives for lives display is too damn low")

  color = color or DEFAULT_COLOR
  color[4] = color[4] or DEFAULT_OPACITY

  local lifeImage = Sprite.newWithName(LIFE_IMAGE_FILENAME)

  local imageWidth, imageHeight = lifeImage:getDims()
  local displayWidth = imageWidth * numberOfLives +  LIVES_PADDING * (numberOfLives - 1)
  local displayHeight = imageHeight

  local self = Group.new(displayWidth, displayHeight)
  Mixin.mix(self, LivesDisplay)

  local lifeImageCollection = {}

  local imageY = displayHeight * 0.5
  for i = 1, numberOfLives do
    local lifeImage = Sprite.newWithName(LIFE_IMAGE_FILENAME)
    local xOffset = (i - 1) * (imageWidth + LIVES_PADDING) + imageWidth * 0.5
    local imageX = displayWidth - xOffset
    lifeImage:setLoc(imageX, imageY)
    self:insert(lifeImage)
    table.insert(lifeImageCollection, lifeImage)
  end

  self:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)

  self.keyColor = color
  self:setColor(unpack(color))

  self.lifeImageCollection = lifeImageCollection
  self.maxLives = numberOfLives

  self:setLives(numberOfLives)
  return self
end
