--- Various kinds of vertical scrollers.
-- @module Scroller
-- @author Created by Karl Stenerud on Oct 10, 2013.


-- Abstract scroller class
-- ===========================================================================
class("Scroller")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Attributes -------------------------

Scroller.ATTR_SCROLL = 0
Scroller.ATTR_SCROLL_SPEED = 1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function onScrollUpdated(scriptNode)
  local self = scriptNode.group
  local scroll = scriptNode:getAttr(Scroller.ATTR_SCROLL)
  local scrollSpeed = scriptNode:getAttr(Scroller.ATTR_SCROLL_SPEED)

  self.scrollSpeed = scrollSpeed

  -- Don't update scroll if this scroller has a speed.
  if self.scrollSpeed ~= 0 then
    self.internalSetScroll(self, scroll)
    self.scroll = scroll
  end

end

local function startScrollUpdateCoroutine(self)
  if not self.scrollUpdateCoroutine then
    local updater = MOAICoroutine.new()
    self.scrollUpdateCoroutine = updater

    local timer = MOAITimer.new()
    timer:setSpan(math.huge)
    timer:start()
    local prevTime = timer:getTime()
    updater:run(function()
      while self.scrollUpdateCoroutine == updater do
        if self.scrollSpeed ~= 0 then
          local curTime = timer:getTime()
          local dt = curTime - prevTime
          prevTime = curTime
          local newScroll = self.scroll + self.scrollSpeed * dt
          self:setScroll(newScroll)
        end
        coroutine.yield()
      end
      timer:stop()
    end)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Actions ----------------------------

Scroller.scrollTo = function(target, duration, ease, amount)
  local attributeID = Scroller.ATTR_SCROLL
  return AHAttributeTo.new(target.scriptNode, duration, attributeID, ease, amount)
end

Scroller.scrollBy = function(target, duration, ease, amount)
  local attributeID = Scroller.ATTR_SCROLL
  return AHAttributeBy.new(target.scriptNode, duration, attributeID, ease, amount)
end

Scroller.scrollSpeedTo = function(target, duration, ease, amount)
  local attributeID = Scroller.ATTR_SCROLL_SPEED
  return AHAttributeTo.new(target.scriptNode, duration, attributeID, ease, amount)
end

Scroller.scrollSpeedBy = function(target, duration, ease, amount)
  local attributeID = Scroller.ATTR_SCROLL_SPEED
  return AHAttributeBy.new(target.scriptNode, duration, attributeID, ease, amount)
end


-- Properties -------------------------

function Scroller:getScroll()
  return self.scroll
end

function Scroller:setScroll(scroll)
  self.scriptNode:setAttr(Scroller.ATTR_SCROLL, scroll)
  self.internalSetScroll(self, scroll)
  self.scroll = scroll
end

function Scroller:setScrollSpeed(speedPPS)
  self.scriptNode:setAttr(Scroller.ATTR_SCROLL_SPEED, speedPPS)
  self.scrollSpeed = speedPPS
end

function Scroller:stop()
  self.scrollUpdateCoroutine = nil
end

local function getAxisX(x, y)
  return x
end

local function getAxisY(x, y)
  return y
end

-- Code generator functions.
--
-- These functions allow the scroller class to effectively set an X or Y value
-- without disturbing the other axis.
-- It will first call the getter, save the values, then call the setter with
-- the specified value, putting in the original X or Y value for the axis we're
-- leaving  undisturbed.
--
-- These will be stored in the constructed scroller as "self.setScrollAxis",
-- pointing to either the X or Y version depending on what direction we're
-- configured to scroll in.
--
-- Usage example in a Scroller configured to scroll on the X axis:
--   self.setScrollAxis(myObject, "Loc", 200) -- Put the object at 200 along the scroll axis
--
-- In this case, self.setScrollAxis points to setAxisX.
--
-- The call is equivalent to:
--   local x, y = myObject:getLoc()
--   myObject:setLoc(200, y)

local function setAxisX(object, funcBaseName, value)
  local getter = "get" .. funcBaseName
  local setter = "set" .. funcBaseName
  local x, y = object[getter](object)
  object[setter](object, value, y)
end

local function setAxisY(object, funcBaseName, value)
  local getter = "get" .. funcBaseName
  local setter = "set" .. funcBaseName
  local x, y = object[getter](object)
  object[setter](object, x, value)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function Scroller.new(width, height, isXScroll, setScrollFunc)
  local self = Group.new(width, height)
  Mixin.mix(self, Scroller)

  if isXScroll then
    self.getScrollAxis = getAxisX
    self.setScrollAxis = setAxisX
  else
    self.getScrollAxis = getAxisY
    self.setScrollAxis = setAxisY
  end

  self.internalSetScroll = setScrollFunc

  self.scrollSpeed = 0
  self.scroll = 0

  self.scriptNode = MOAIScriptNode.new()
  self.scriptNode:reserveAttrs(2)
  self.scriptNode.group = self
  self.scriptNode:setCallback(onScrollUpdated)

  startScrollUpdateCoroutine(self)

  return self
end



-- Assumes all assets are of fixed size, and scrolls them in the order they
-- were provided in the constructor.
-- ===========================================================================
class("FixedScroller")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function FixedScroller_setScroll(self, scroll)
  self.scroll = scroll

  local realScroll = scroll % self.assetScrollAxisDim
  local primaryIdx = math.floor(scroll / self.assetScrollAxisDim) % self.assetCount + 1
  local secondaryIdx = (primaryIdx + self.assetCount) % self.assetCount + 1

  self.setScrollAxis(self.assets[primaryIdx], "Loc", realScroll)
  self.setScrollAxis(self.assets[secondaryIdx], "Loc", realScroll - self.assetScrollAxisDim + 1)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function FixedScroller.new(assets, isXScroll)
  local assetW, assetH = assets[1]:getEffectiveDims()
  local self = Scroller.new(assetW, assetH, isXScroll, FixedScroller_setScroll)
  Mixin.mix(self, FixedScroller)

  self.assets = assets
  self.currentIndex = 1
  self.assetCount = #assets
  self.assetScrollAxisDim = self.getScrollAxis(assets[1]:getEffectiveDims())
  self.assetX = 0

  for _, asset in ipairs(assets) do
    asset:setAnchorPointBottomLeft()
    self.setScrollAxis(asset, "Loc", -self.assetScrollAxisDim)
    self:insert(asset)
  end

  self:setScroll(0)

  return self
end



-- Scrolls multiple other scrollers at different rates to give a parallax
-- effect.
-- ===========================================================================
class("ParallaxScroller")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function ParallaxScroller_setScroll(self, scroll)
  self.scroll = scroll

  for i=1,#self.scrollers do
    self.scrollers[i]:setScroll(scroll * self.scrollProportions[i])
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ParallaxScroller.new(scrollers, scrollProportions)
  local scrollerW, scrollerH = scrollers[1]:getEffectiveDims()

  local self = Scroller.new(scrollerW, scrollerH, true, ParallaxScroller_setScroll)
  Mixin.mix(self, ParallaxScroller)

  self.scrollers = scrollers
  self.scrollProportions = scrollProportions

  for _, scroller in ipairs(scrollers) do
    scroller:setLoc(scrollerW/2, scrollerH/2)
    self:insert(scroller)
  end

  self:setScroll(0)

  return self
end



-- Wraps a scroller, allowing precise positioning.
-- ===========================================================================
class("PositionedScroller")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function PositionedScroller_setScroll(self, scroll)
  self.scroll = scroll
  self.scroller:setScroll(scroll)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function PositionedScroller.new(scroller, width, height, posX, posY)
  local self = Scroller.new(width, height, true, PositionedScroller_setScroll)
  Mixin.mix(self, PositionedScroller)
  self:setDims(width, height)

  self.scroller = scroller

  self.scroller:setLoc(posX, posY)
  self:insert(self.scroller)

  self:setScroll(0)

  return self
end



-- Scrolls individual items at a specified X and Y offset.
-- ===========================================================================
class("ItemScroller")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function ItemScroller_setScroll(self, scroll)
  self.scroll = scroll

  for asset, offset in pairs(self.assets) do
    local scrollOffset = self.getScrollAxis(offset.x, offset.y)
    self.setScrollAxis(asset, "Loc", (scroll + scrollOffset) % self.wrapAt)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ItemScroller.new(assets, wrapAt, width, height, isXScroll)
  isXScroll = isXScroll or false
  local self = Scroller.new(width, height, isXScroll, ItemScroller_setScroll)
  Mixin.mix(self, ItemScroller)

  self.assets = assets
  self.wrapAt = wrapAt

  if not width then
    width = Screen.width
  end
  if not height then
    height = Screen.height
  end

  for asset, offset in pairs(assets) do
    asset:setLoc(offset.x, offset.y)
    self:insert(asset)
  end

  self:setScroll(0)

  return self
end
