--- TapToContinue
-- Waits for the user to touch the screen, optionally displaying an overlay.
-- Also supports a timeout operation.
-- @author Created by karl on July 01, 2014.

-- ===========================================================================
class("TapToContinue")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function performCallback(self, isTimeout)
  if self.isArmed then
    local callback = self.callback
    self:disarm()
    if callback then
      if (self.runCallbackInCoroutine) then
        runInCoroutine(function() callback(self, isTimeout) end)
      else
        callback(self, isTimeout)
      end
    end
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Disarm, removing from the touch manager and ignoring timeouts.
function TapToContinue:disarm()
  self.isArmed = false

  TouchManager.removeTouchHandler(self)

  if self.timeoutAction then
    self.timeoutAction:stop()
    self.timeoutAction = nil
  end
end

--- Rearm the tap-to-continue after it has fired.
-- If already armed, it will disarm before rearming.
-- Note: This will also restart the timeout operation if timeout is valid.
function TapToContinue:rearm()
  if self.isArmed then
    self:disarm()
  end

  self.isArmed = true

  TouchManager.addTouchHandler(self, function(event)
    if event.phase == TouchManager.TOUCH_BEGAN then
      local isTimeout = false
      performCallback(self, isTimeout)
    end
  end)

  if self.timeoutInSeconds then
    local isTimeout = true
    self.timeoutAction = AHSequence.new(
      AHDelay.new(self.timeoutInSeconds),
      function() performCallback(self, isTimeout) end
    )
    self.timeoutAction:start()
  end
end

--- Pause the current coroutine until triggered by a tap, timeout, or disarm.
function TapToContinue:waitForTrigger()
  waitWhile(function() return self.isArmed end)
end

function TapToContinue:getTimeoutInSeconds()
  return self.timeoutInSeconds
end

--- Set the timeout duration in seconds.
-- Note: This will not affect the currently running timeout action, if any.
-- To restart the timeout, call rearm.
function TapToContinue:setTimeoutInSeconds(timeoutInSeconds)
  self.timeoutInSeconds = timeoutInSeconds
end

function TapToContinue:getCallback()
  return self.callback
end

function TapToContinue:setCallback(callback)
  self.callback = callback
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Create a new tap-to-continue overlay in the specified parent group.
-- The overlay will accept the next touch and call the callback.
-- Also supports an optional timeout, which also calls the callback.
-- @param parent (prop) - The new TapToContinue will be added to the center of this prop, at max Z order.
-- @param child (prop, optional) - Child to be added to the center of this group.
-- @param timeoutInSeconds (number, optional) - Automatically call the callback after timeout seconds.
-- @param callback (function(self, isTimeout)) - Called on timeout or touch.
--                 self: This TapToContinue object.
--                 isTimeout: If true, callback was called due to a timeout rather than a touch.
-- @param runCallbackInCoroutine (boolean, optional) - Specify if the callback should be processed
--                 on a separate thread. Defaults to true.
function TapToContinue.newInGroup(parent, child, timeoutInSeconds, callback, runCallbackInCoroutine)
  local selfW, selfH = parent:getDims()
  local self = Group.new(selfW, selfH)
  Mixin.mix(self, TapToContinue)

  if child then
    self:insertAtCenter(child)
  end

  if (runCallbackInCoroutine == nil) then
    runCallbackInCoroutine = true
  end

  self.timeoutInSeconds = timeoutInSeconds
  self.callback = callback
  self.runCallbackInCoroutine = runCallbackInCoroutine

  self:rearm()

  parent:insertAtCenter(self, math.huge)

  return self
end


--- Create a new tap-to-continue overlay in the screen's root group.
-- The overlay will accept the next touch and call the callback.
-- Also supports an optional timeout, which also calls the callback.
-- @param child (prop, optional) - Child to be added to the center of this group.
-- @param timeoutInSeconds (number, optional) - Automatically call the callback after timeout seconds.
-- @param callback (function(self, isTimeout)) - Called on timeout or touch.
--                 self: This TapToContinue object.
--                 isTimeout: If true, callback was called due to a timeout rather than a touch.
-- @param runCallbackInCoroutine (boolean, optional) - Specify if the callback should be processed
--                 on a separate thread. Defaults to true.
function TapToContinue.new(child, timeoutInSeconds, callback, runCallbackInCoroutine)
  return TapToContinue.newInGroup(Screen.rootGroup, child, timeoutInSeconds, callback, runCallbackInCoroutine)
end
