--- TouchScroller
-- @author Created by karl on February 11, 2015.

class("TouchScroller")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SCROLL_THRESHOLD = 5
local DEFAULT_SCROLL_DECAY_FACTOR = 0.8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startScrolling(self, touchEvent)
  self.isScrolling = true
  if (self.onScrollStarted) then
    self.onScrollStarted()
  end

  -- Improves scrolling smoothness by avoiding an initial jump of
  -- SCROLL_THRESHOLD units in the scroll direction once we've determined
  -- user intends to scroll.
  if (self.willResetInitialTouchCoordinateUponScrollStart) then
    self.initialTouch.y = touchEvent.y
  end
end

local function stopScrolling(self)
  self.isScrolling = false
  if (self.onScrollEnded) then
    self.onScrollEnded()
  end
end

local function getMaxScroll(self)
  return self.contents:getHeight() - self:getHeight()
end

local function getMinScroll()
  return 0
end

local function getScrollAmount(self, event)
  if (self.initialTouch ~= nil) then
    return event.y - self.initialTouch.y
  end
  return 0
end

local function hasScrolledSignificantlyY(self, event)
  return math.abs(event.y - self.initialTouch.y) > SCROLL_THRESHOLD
end

local function hasScrolledSignificantlyX(self, event)
  return math.abs(event.x - self.initialTouch.x) > SCROLL_THRESHOLD
end

local function canScrollUp(self, event)
  local scrollY = self.initialTouch.scroll + getScrollAmount(self, event)
  return scrollY <= getMaxScroll(self)
end

local function canScrollDown(self, event)
  local scrollY = self.initialTouch.scroll + getScrollAmount(self, event)
  return scrollY >= getMinScroll()
end

local function adjustScroll(self, event)
  if (self.initialTouch ~= nil) then
    self:setScrollY(self.initialTouch.scroll + getScrollAmount(self, event))
  end
end

local function onTouch(self, event)
  if (event.phase == TouchManager.TOUCH_BEGAN) then
    self.initialTouch = {x = event.x, y = event.y, scroll = self:getScrollY()}
    self.isScrolling = false
    self.scrollVelocity = 0
  elseif ((event.phase == TouchManager.TOUCH_MOVED) and (self.initialTouch ~= nil)) then
    if (self.isHandoffOnScrollUpEnabled and not canScrollUp(self, event)) then
      return TouchEvent.HANDLING.HAND_OFF
    elseif (self.isHandoffOnScrollDownEnabled and not canScrollDown(self, event)) then
      return TouchEvent.HANDLING.HAND_OFF
    end
    if (self.isScrolling) then
      adjustScroll(self, event)
    else
      if (hasScrolledSignificantlyY(self, event)) then
        startScrolling(self, event)
        adjustScroll(self, event)
      elseif (self.isHandoffOnHorizontalSwipeEnabled and hasScrolledSignificantlyX(self, event)) then
        return TouchEvent.HANDLING.HAND_OFF
      end
    end
  else
    stopScrolling(self)
  end
  self.lastTouchX, self.lastTouchY = event.x, event.y
end

local function startUpdateInertiaCoroutine(self)
  if (self.updateInertiaCoroutine ~= nil) then
    return
  end

  self.scrollVelocity = 0
  local updateTimer = MOAITimer.new()
  updateTimer:setSpan(math.huge)

  local updateInertia = function()
    updateTimer:start()
    local previousTime = updateTimer:getTime()
    local previousY = self.scrollY
    while (true) do
      waitForFrames(1)

      local currentTime = updateTimer:getTime()
      local currentY = self.scrollY

      if self.isScrolling then
        self.scrollVelocity = (currentY - previousY) / (currentTime - previousTime)
      else
        local deltaTime = currentTime - previousTime

        self.scrollVelocity = self.scrollVelocity * self.scrollDecayFactor

        local epsilon = 0.01
        if (math.abs(self.scrollVelocity) < epsilon) then
          self.scrollVelocity = 0
        end

        local deltaScrollY = self.scrollVelocity * deltaTime
        self:setScrollY(self:getScrollY() + deltaScrollY)
      end

      previousTime = currentTime
      previousY = currentY
    end
  end

  self.updateInertiaCoroutine = runInCoroutine(updateInertia)
end

local function stopUpdateInertiaCoroutine(self)
  if (self.updateInertiaCoroutine ~= nil) then
    self.updateInertiaCoroutine:stop()
  end
  self.updateInertiaCoroutine = nil
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TouchScroller:getContents()
  return self.contents
end

function TouchScroller:setScrollBoundsEnabled(enabled)
  self.isScrollBoundsEnabled = enabled
end

function TouchScroller:setInertialScrollingEnabled(isEnabled)
  if (isEnabled ~= self.isInertialScrollingEnabled) then
    if (isEnabled) then
      startUpdateInertiaCoroutine(self)
    else
      stopUpdateInertiaCoroutine(self)
    end
    self.isInertialScrollingEnabled = isEnabled
  end
end

function TouchScroller:getScrollY()
  return self.scrollY
end

function TouchScroller:setScrollY(scroll)
  if (self.isScrollBoundsEnabled) then
    local diff = math.max(0, self.contents:getHeight() - self:getHeight())
    self.scrollY = MathUtils.clamp(scroll, 0, diff)
  else
    self.scrollY = scroll
  end

  local posY = self:getHeight() + self.scrollY
  self.contents:setY(posY)

  if (self.onScrollUpdatedCallback ~= nil) then
    self.onScrollUpdatedCallback(self.scrollY)
  end
end

function TouchScroller:newScrollYTo(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getScrollY, self.setScrollY, ease, amount)
end

function TouchScroller:newScrollYBy(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getScrollY, self.setScrollY, ease, amount)
end

function TouchScroller:setScrollNotificationFunctions(onScrollStarted, onScrollEnded)
  self.onScrollStarted = onScrollStarted
  self.onScrollEnded = onScrollEnded
end

function TouchScroller:setHandoffOnHorizontalSwipeEnabled(isEnabled)
  self.isHandoffOnHorizontalSwipeEnabled = isEnabled
end

function TouchScroller:setHandoffOnScrollUpEnabled(isEnabled)
  self.isHandoffOnScrollUpEnabled = isEnabled
end

function TouchScroller:setHandoffOnScrollDownEnabled(isEnabled)
  self.isHandoffOnScrollDownEnabled = isEnabled
end

function TouchScroller:setOnScrollUpdatedCallback(onScrollUpdatedCallback)
  self.onScrollUpdatedCallback = onScrollUpdatedCallback
end

function TouchScroller:setScrollDecayFactor(factor)
  assert(factor > 0 and factor < 1, "Scroll decay factor of " .. tostring(factor) .. " does not fall between (0, 1)")

  self.scrollDecayFactor = factor
end

function TouchScroller:setWillResetInitialTouchCoordinateUponScrollStart(isEnabled)
  self.willResetInitialTouchCoordinateUponScrollStart = isEnabled
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TouchScroller.new(width, height, contents)
  local self = Group.new(width, height)
  Mixin.mix(self, TouchScroller)

  self.isScrollBoundsEnabled = true
  self.onScrollUpdatedCallback = nil
  self.scrollDecayFactor = DEFAULT_SCROLL_DECAY_FACTOR

  contents:setAnchorPointTop()
  self:insertAtCenter(contents)
  self.contents = contents
  self:setScrollY(0)

  self.isInertialScrollingEnabled = false
  self.willResetInitialTouchCoordinateUponScrollStart = false
  self.initialTouch = nil

  local handlerFunction = function(event)
    return onTouch(self, event)
  end

  local onEnableTouch = function()
    self.initialTouch = nil
  end

  TouchManager.addTouchEnabledMethods(self, handlerFunction, TouchHandler.BEHAVIOR_HANDOFF, nil, onEnableTouch)
  self:setTouchEnabled(true)

  return self
end
