--- ValidityButtonBar
-- @author Created by Jay Austria on August 14, 2015.


-- ===========================================================================
local ValidityButton = privateClass()
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_HEIGHT = Screen.getSafeHeightPart(0.11)

local BUTTON_CORNER_RADIUS = 6
local INNER_BUTTON_CORNER_RADIUS = BUTTON_CORNER_RADIUS - 1
local NEGATIVE_BUTTON_STRING = Localizer.get("ValidityButton.NEGATIVE")
local POSITIVE_BUTTON_STRING = Localizer.get("ValidityButton.POSITIVE")
local FADE_DURATION = 0.6

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local function initializeButton(self)
  local touchBeganWrapper = function()
    if self.tapCallback then
      self.tapCallback()
    end
  end

  Button.makeButton(self, touchBeganWrapper)
end

local function createButtonColorAction(self, endBackingColor, endTextColor)
  local beginBackingColor = self.styleDescriptor.defaultBackingColor
  local beginTextColor = self.styleDescriptor.defaultTextColor

  local colorAction = AHSpawn.new(
    AHSequence.new(
      AHColorTo.new(self.backing, 0.1, nil, endBackingColor),
      AHDelay.new(0.1),
      AHColorTo.new(self.backing, 0.1, nil, beginBackingColor)
    ),
    AHSequence.new(
      AHColorTo.new(self.labelRow, 0.1, nil, endTextColor),
      AHDelay.new(0.1),
      AHColorTo.new(self.labelRow, 0.1, nil, beginTextColor)
    )
  )

  return colorAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ValidityButton:setTapCallback(callback)
  self.tapCallback = callback
end

function ValidityButton:createFadeAction(opacity)
  local fadeAction = AHSpawn.new(
    AHOpacityTo.new(self, FADE_DURATION, nil, opacity),
    AHOpacityTo.new(self.labelRow, FADE_DURATION, nil, opacity)
  )

  return fadeAction
end

function ValidityButton:createCorrectAnimation()
  return createButtonColorAction(self, self.styleDescriptor.correctBackingColor, self.styleDescriptor.correctTextColor)
end

function ValidityButton:createIncorrectAnimation()
  return createButtonColorAction(self, self.styleDescriptor.incorrectBackingColor, self.styleDescriptor.incorrectTextColor)
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

local function createLabelRow(icon, text, color)
  local elementSpacing = 10
  local labelRow = Row.new(elementSpacing)

  local symbol = Sprite.newWithName(icon)
  local symbolScale = 0.9
  symbol:setScl(symbolScale)
  labelRow:insert(symbol)

  local textLabel = Label.new(text, Font.new(FontFaces.DIN_OT_BOLD, 15))
  labelRow:insert(textLabel)

  labelRow:alignChildren()

  labelRow:linkColorAttrToChildren(true)
  labelRow:setColor(unpack(color))

  return labelRow
end

local function createLayout(self, text, iconFilename, isFlipped)
  isFlipped = isFlipped or false
  local buttonWidth = self:getWidth()
  local buttonHeight = self:getHeight()

  local sharpCorners = nil
  if (isFlipped == true) then
    sharpCorners = {FilledRoundedRectangle.CORNERS.TOP_LEFT, FilledRoundedRectangle.CORNERS.BOTTOM_LEFT}
  else
    sharpCorners = {FilledRoundedRectangle.CORNERS.TOP_RIGHT, FilledRoundedRectangle.CORNERS.BOTTOM_RIGHT}
  end

  local backing = FilledRoundedRectangle.new(buttonWidth, buttonHeight, INNER_BUTTON_CORNER_RADIUS, self.styleDescriptor.defaultBackingColor, sharpCorners)
  self:insertAtCenter(backing)
  self.backing = backing

  if (self.styleDescriptor.outlineColor ~= nil) then
    local outlineColor = self.styleDescriptor.outlineColor
    local lineWidth = self.styleDescriptor.outlineWidth or LINE_THICKNESS
    local blurMargin = self.styleDescriptor.outlineBlur or 0
    local outline = HollowRoundedRectangle.new(buttonWidth, buttonHeight, INNER_BUTTON_CORNER_RADIUS, outlineColor, lineWidth, blurMargin, sharpCorners)
    self:insertAtCenter(outline)
    self.outline = outline
  end

  local labelRow = createLabelRow(iconFilename, text, self.styleDescriptor.defaultTextColor)
  self:insertAtCenter(labelRow)
  self.labelRow = labelRow

  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

function ValidityButton.new(buttonWidth, buttonHeight, text, iconFilename, isFlipped, styleDescriptor)
  local self = Group.new(buttonWidth, buttonHeight)
  Mixin.mix(self, ValidityButton)

  self.styleDescriptor = styleDescriptor

  createLayout(self, text, iconFilename, isFlipped)
  initializeButton(self)

  return self
end

-- ============================================================================
class("ValidityButtonBar")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LINE_THICKNESS = 2


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getButtonToAnimate(self, willGetAffirmativeButton)
  if (willGetAffirmativeButton) then
    return self.affirmativeButton
  else
    return self.negativeButton
  end
end

local function createFadeAction(self, opacity)
  local fadeAction = AHSpawn.new(
    AHOpacityTo.new(self, FADE_DURATION, nil, opacity),
    self.negativeButton:createFadeAction(opacity),
    self.affirmativeButton:createFadeAction(opacity)
  )

  return fadeAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ValidityButtonBar:setAffirmativeButtonCallback(callback)
  self.affirmativeButton:setTapCallback(callback)
end

function ValidityButtonBar:setNegativeButtonCallback(callback)
  self.negativeButton:setTapCallback(callback)
end

function ValidityButtonBar:setTouchEnabled(enabled)
  self.negativeButton:setTouchEnabled(enabled)
  self.affirmativeButton:setTouchEnabled(enabled)
end

function ValidityButtonBar:createFadeOutAction()
  local opacity = 0
  return createFadeAction(self, opacity)
end

function ValidityButtonBar:createFadeInAction()
  local opacity = 1
  return createFadeAction(self, opacity)
end

function ValidityButtonBar:prepareForShowAction()
  self:setOpacity(0)
end

function ValidityButtonBar:playCorrectAnimation(willAnimateAffirmativeButton)
  local buttonToAnimate = getButtonToAnimate(self, willAnimateAffirmativeButton)
  local correctAnimation = buttonToAnimate:createCorrectAnimation()

  return correctAnimation:start()
end

function ValidityButtonBar:playIncorrectAnimation(willAnimateAffirmativeButton)
  local buttonToAnimate = getButtonToAnimate(self, willAnimateAffirmativeButton)
  local incorrectAnimation = buttonToAnimate:createIncorrectAnimation()

  return incorrectAnimation:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local buttonRow = Row.new()
  local buttonWidth = self:getWidthPart(0.5)
  local buttonHeight = self:getHeight()

  local usingCustomOutline = (self.styleDescriptor.outlineWidth ~= nil)

  local isFlipped = false
  local negativeButton = ValidityButton.new(buttonWidth, buttonHeight, NEGATIVE_BUTTON_STRING, "validity_cross.png", isFlipped, self.styleDescriptor)
  buttonRow:insert(negativeButton)

  local lineWidth = LINE_THICKNESS
  if not usingCustomOutline then
    buttonRow:insertSpacer(lineWidth)
  end

  isFlipped = true
  local affirmativeButton = ValidityButton.new(buttonWidth, buttonHeight, POSITIVE_BUTTON_STRING, "validity_tick.png", isFlipped, self.styleDescriptor)
  buttonRow:insert(affirmativeButton)

  buttonRow:alignChildren()
  self:insertAtCenter(buttonRow)

  if not usingCustomOutline then
    local centerDivider = FilledRectangle.new(lineWidth, self:getHeight(), self.styleDescriptor.centerDividerColor)
    self:insertAtCenter(centerDivider)
    centerDivider:inheritOpacity(self)
    self.centerDivider = centerDivider
  end

  self.negativeButton = negativeButton
  self.affirmativeButton = affirmativeButton

  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

function ValidityButtonBar.newWithStyleDescriptor(styleDescriptor)
  Asserts.assertType(styleDescriptor, "styleDescriptor", "table")
  Asserts.assertType(styleDescriptor.defaultBackingColor, "defaultBackingColor", "table")
  Asserts.assertType(styleDescriptor.correctBackingColor, "correctBackingColor", "table")
  Asserts.assertType(styleDescriptor.centerDividerColor, "centerDividerColor", "table")
  Asserts.assertType(styleDescriptor.incorrectBackingColor, "incorrectBackingColor", "table")
  Asserts.assertType(styleDescriptor.defaultTextColor, "defaultTextColor", "table")
  Asserts.assertType(styleDescriptor.correctTextColor, "correctTextColor", "table")
  Asserts.assertType(styleDescriptor.incorrectTextColor, "incorrectTextColor", "table")
  Asserts.assertOptionalType(styleDescriptor.outlineColor, "outlineColor", "table")
  Asserts.assertOptionalType(styleDescriptor.outlineWidth, "outlineWidth", "number")
  Asserts.assertOptionalType(styleDescriptor.outlineBlur, "outlineBlur", "number")
  Asserts.assertOptionalType(styleDescriptor.width, "width", "number")
  Asserts.assertOptionalType(styleDescriptor.height, "height", "number")

  local width = styleDescriptor.width or Screen.getSafeWidthPart(0.85)
  local height = styleDescriptor.height or BUTTON_HEIGHT
  local self = Group.new(width, height)
  Mixin.mix(self, ValidityButtonBar)

  self.styleDescriptor = styleDescriptor

  createLayout(self)

  return self
end
