class("Key")

require("widgets.keyboard.KeyPreview")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

Key.types = {
  normal = "Key.types.normal",
  padding = "Key.types.padding",
  space = "Key.types.space",
  backspace = "Key.types.backspace",
  command = "Key.types.command"
}

local STYLES = {
  light = {
    bgColor = {1, 1, 1},
    shadowColor = {0, 0, 0, 0.15},
    textColor = {0, 0, 0, 0.7},
    font = FontFaces.DIN_OT_LIGHT,
    fontSize = 0.5
  },
  dark = {
    bgColor = ColorUtils.colorRGBToStd(0, 0, 0, 0.8),
    shadowColor = {0, 0, 0, 0.4},
    textColor = {1, 1, 1},
    font = FontFaces.DIN_OT_MEDIUM,
    fontSize = 0.5
  }
}
STYLES.highlighted = TableUtils.shallowCopy(STYLES.light)
STYLES.preview = TableUtils.shallowCopy(STYLES.light)
STYLES[Key.types.normal] = TableUtils.shallowCopy(STYLES.light)
STYLES[Key.types.normal].hasPreview = true
STYLES[Key.types.padding] = TableUtils.shallowCopy(STYLES.light)
STYLES[Key.types.space] = TableUtils.shallowCopy(STYLES.light)
STYLES[Key.types.space].font = FontFaces.DIN_OT_MEDIUM
STYLES[Key.types.space].fontSize = 0.4
STYLES[Key.types.backspace] = TableUtils.shallowCopy(STYLES.light)
STYLES[Key.types.backspace].icon = "shared_keyboard_backspace_icon"
STYLES[Key.types.backspace].isRepeatable = true
STYLES[Key.types.command] = TableUtils.shallowCopy(STYLES.dark)
STYLES[Key.types.command].fontSize = 0.4

Key.SUBMIT_TRANSITION_DELAY = 0.05

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function applyStyle(self, style)
  self.backgroundFront:setColor(unpack(style.bgColor))
  self.backgroundShadow:setColor(unpack(style.shadowColor))
  self.displayNormal:setColor(unpack(style.textColor))
end

local function setHighlighted(self, isHighlighted, transitionDelay)
  transitionDelay = transitionDelay or 0

  if self.preview then
    runInCoroutine(function()
      waitFor(transitionDelay)
      self.preview:setVisibleRecursive(isHighlighted)
    end)
  else
    local style, visibleDisplay, invisibleDisplay

    if isHighlighted then
      style = self.highlightedStyle
      visibleDisplay = self.displayHighlight
      invisibleDisplay = self.displayNormal
    else
      style = self.normalStyle
      visibleDisplay = self.displayNormal
      invisibleDisplay = self.displayHighlight
    end
    self.backgroundFront:setColor(unpack(style.bgColor))
    self.backgroundShadow:setColor(unpack(style.shadowColor))
    visibleDisplay:setVisible(true)
    invisibleDisplay:setVisible(false)
  end
end

local function isRepeatable(self)
  return self.normalStyle.isRepeatable
end

local function startRepeating(self)
  if not self.repeater then
    local repeatPress = AHDelayed.new(0.05, function() self.onInput() end)
    self.repeater = AHDelayed.new(0.2, AHLoop.new(repeatPress))
    self.repeater:start()
  end
end

local function stopRepeating(self)
  if self.repeater then
    self.repeater:stop()
    self.repeater = nil
  end
end

local function isRepeating(self)
  return self.repeater ~= nil
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Key:transitionDown()
  if self._isEnabled then
    setHighlighted(self, true)
    if(isRepeatable(self)) then
      startRepeating(self)
    end
  end
end

function Key:transitionUp(transitionDelay)
  if self._isEnabled then
    stopRepeating(self)
    setHighlighted(self, false, transitionDelay)
  end
end

function Key:activate()
  if self._isEnabled and not isRepeating(self) then
    self.onInput()
  end
end

function Key:reset()
  self:transitionUp()
end

function Key:isEnabled()
  return self._isEnabled
end

function Key:setEnabled(isEnabled)
  if self._isEnabled ~= isEnabled then
    if isEnabled then
      applyStyle(self, self.normalStyle)
    else
      self:transitionUp()
      self.backgroundFront:setOpacity(0.15)
      self.displayNormal:setOpacity(0.15)
      self.backgroundShadow:setOpacity(0)
    end
    self._isEnabled = isEnabled
  end
end

function Key:setInputHandler(onInput)
  self.onInput = onInput
end

function Key:getTouchArea()
  return self.virtualTouchOverlay or self
end

--- Add a virtual overlay that extends beyond the left and/or right of the key.
-- These values extend from the MARGINS, not the visible region.
function Key:setVirtualTouchOverlay(extraWidthLeft, extraWidthRight)
  if self.virtualTouchOverlay then
    self.virtualTouchOverlay:removeSelf()
  end
  local offset = ( (extraWidthRight or 0) - (extraWidthLeft or 0) ) / 2
  self.virtualTouchOverlay = Group.new(self:getWidth() + extraWidthLeft + extraWidthRight, self:getHeight())
  self.virtualTouchOverlay:setLoc(self:getWidthPart(0.5) + offset, self:getHeightPart(0.5))
  self:insert(self.virtualTouchOverlay)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, keyString, width, height)
  local dropShadowOffset = -1
  local cornerRadius = height * 0.1
  local style = self.normalStyle
  self.backgroundFront = FilledRoundedRectangle.new(width, height, cornerRadius)
  self.backgroundShadow = FilledRoundedRectangle.new(width, height, cornerRadius)
  self.backgroundShadow:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.5) + dropShadowOffset)
  self:insert(self.backgroundShadow)
  self:insertAtCenter(self.backgroundFront)

  if style.icon then
    self.displayNormal = Sprite.newWithName(style.icon .. ".png")
    self.displayHighlight = Sprite.newWithName(style.icon .. "_hover.png")
  else
    local fontSize = height * style.fontSize
    self.displayNormal = DeprecatedLabel.newSingleLine(keyString, style.font, fontSize)
    self.displayHighlight = self.displayNormal:clone()
  end
  self.displayHighlight:setColor(unpack(STYLES.light.textColor))
  self.displayHighlight:setVisible(false)
  self:insertAtCenter(self.displayNormal)
  self:insertAtCenter(self.displayHighlight)

  if style.hasPreview then
    self.preview = KeyPreview.new(keyString, width, height, cornerRadius, style.bgColor, style.textColor)
    self:insert(self.preview)
    self.preview:alignToProp(self.backgroundFront, 0.5, 0, 0, 0.5, 0, 0)
    self.preview:setVisibleRecursive(false)
  end
end

-- Constructor ------------------------

function Key.new(keyString, keyType, keyW, keyH, marginX, marginY, onInput)
  local self = Group.new(keyW + marginX * 2, keyH + marginY * 2)

  Mixin.mix(self, Key)

  self.normalStyle = STYLES[keyType]
  self.highlightedStyle = STYLES.highlighted
  self.previewStyle = STYLES.preview

  createLayout(self, keyString, keyW, keyH)

  self.keyString = keyString
  self.onInput = onInput

  self:setEnabled(true)

  return self
end
