-- Created by Alexey Chernikov on 02/25/2013
class("Keyboard")

require("widgets.keyboard.Key")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

Keyboard.keyWidths = {
  normal = Screen.getWidthPart(0.08),
  wide = Screen.getWidthPart(0.11),
  space = Screen.getWidthPart(0.5),
  command = Screen.getWidthPart(0.22),
}

Keyboard.specialKeys = {
  space = "Space",
  backspace = "Backspace",
  enter = "Enter",
  escape = "Escape",
}

-- Layout parameters:
-- name (string) - The name used to reference this key, and also what is displayed on screen.
-- width (number) - The width of this key.
-- type (string) - What kind of key this is (optional, defaults to "normal"). See Key.types.
-- customFunction (function) - Function to call instead of default handling when this key is pressed.
-- isEnabled (boolean) - Set to false to disable this key from the start (optional, defaults to true).
Keyboard.DEFAULT_LAYOUT = {
  {
    {name = "Q", width = Keyboard.keyWidths.normal},
    {name = "W", width = Keyboard.keyWidths.normal},
    {name = "E", width = Keyboard.keyWidths.normal},
    {name = "R", width = Keyboard.keyWidths.normal},
    {name = "T", width = Keyboard.keyWidths.normal},
    {name = "Y", width = Keyboard.keyWidths.normal},
    {name = "U", width = Keyboard.keyWidths.normal},
    {name = "I", width = Keyboard.keyWidths.normal},
    {name = "O", width = Keyboard.keyWidths.normal},
    {name = "P", width = Keyboard.keyWidths.normal},
  },
  {
    {name = "A", width = Keyboard.keyWidths.normal},
    {name = "S", width = Keyboard.keyWidths.normal},
    {name = "D", width = Keyboard.keyWidths.normal},
    {name = "F", width = Keyboard.keyWidths.normal},
    {name = "G", width = Keyboard.keyWidths.normal},
    {name = "H", width = Keyboard.keyWidths.normal},
    {name = "J", width = Keyboard.keyWidths.normal},
    {name = "K", width = Keyboard.keyWidths.normal},
    {name = "L", width = Keyboard.keyWidths.normal},
  },
  {
    {name = "-", width = Keyboard.keyWidths.wide},
    {width = Screen.getWidthPart(0.02), type = Key.types.padding},
    {name = "Z", width = Keyboard.keyWidths.normal},
    {name = "X", width = Keyboard.keyWidths.normal},
    {name = "C", width = Keyboard.keyWidths.normal},
    {name = "V", width = Keyboard.keyWidths.normal},
    {name = "B", width = Keyboard.keyWidths.normal},
    {name = "N", width = Keyboard.keyWidths.normal},
    {name = "M", width = Keyboard.keyWidths.normal},
    {width = Screen.getWidthPart(0.02), type = Key.types.padding},
    {name = Keyboard.specialKeys.backspace, width = Keyboard.keyWidths.wide, type = Key.types.backspace},
  },
  {
    -- These are left in as examples of how to add custom commands beside the space bar.
    -- {name = "Skip", width = Keyboard.keyWidths.command, type = Key.types.command, customFunction = function() print("skip") end},
    {name = Keyboard.specialKeys.space, width = Keyboard.keyWidths.space, type = Key.types.space},
    -- {name = Keyboard.specialKeys.enter, width = Keyboard.keyWidths.command, type = Key.types.command, customFunction = function() print("enter") end}
  },
}

local MARGIN_X = Screen.getWidthPart(0.01)
local MARGIN_Y = Screen.getHeightPart(0.01)

local KNOWN_DEVICE_KEYBOARD_HEIGHT = 224 -- iPhones 4 and 5
local KEYBOARD_HEIGHT = math.max(Screen.getHeightPart(0.36), KNOWN_DEVICE_KEYBOARD_HEIGHT)

local ROWS_COUNT = #Keyboard.DEFAULT_LAYOUT
local KEY_HEIGHT = (KEYBOARD_HEIGHT - (MARGIN_Y * 2 * ROWS_COUNT)) / ROWS_COUNT


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function handleInputStart()
  Sound.playSoundEffect("shared_keytap")
end

local function resetKeys(self)
  -- Need to manually reset because the OS may not update quickly enough to
  -- unselect a previously selected key.
  for _, key in pairs(self.keys) do
    if key.reset then
      key:reset()
    end
  end
end

local function onBackspace(self)
  if self._isTouchEnabled then
    self.onBackspace()
  end
end

local function onLetter(self, letter)
  if self._isTouchEnabled then
    self.onLetter(letter)
  end
end

local function onCustomFunction(self, customFunction)
  if self._isTouchEnabled then
    customFunction()
  end
end

local function newHandlerFunction(self, name, keyType, customFunction)
  if customFunction then
    return function()
      onCustomFunction(self, customFunction)
    end
  elseif keyType == Key.types.backspace then
    return function()
      onBackspace(self)
    end
  elseif keyType == Key.types.space then
    return function()
      onLetter(self, " ")
    end
  elseif keyType == Key.types.normal then
    local letter = name
    return function()
      onLetter(self, letter)
    end
  elseif keytype == nil then
    return function() end
  else
    error(tostring(keyType) .. ": Unsupported key type")
  end
end

local function getTargetKey(self, event)
  for _, key in pairs(self.keys) do
    if key:getTouchArea():inside(event.x, event.y) then
      if key:isEnabled() then
        return key
      end
    end
  end
  return nil
end

local function transitionKeyDown(key)
  if key then
    key:transitionDown()
  end
end

local function transitionKeyUp(key)
  if key then
    key:transitionUp()
  end
end

local function activateKey(self, key, transitionDelay)
  if key then
    key:transitionUp(transitionDelay)
    key:activate()
  end
end

local function handleTouch(self, event)
  local tapIDBeingHandled = self.tapIDBeingHandled
  local keyBeingHandled = self.keyBeingHandled
  local tapID = event.id
  local key = getTargetKey(self, event)

  if (event.phase == TouchManager.TOUCH_BEGAN) then
    tapIDBeingHandled = tapID

    local isAnotherKeyBeingHandled = (keyBeingHandled ~= nil)

    if isAnotherKeyBeingHandled then
      activateKey(self, keyBeingHandled)
    end

    transitionKeyDown(key)
    keyBeingHandled = key
    handleInputStart()

  elseif (event.phase == TouchManager.TOUCH_MOVED) then
    if tapIDBeingHandled == tapID then
      if keyBeingHandled ~= key then
        transitionKeyUp(keyBeingHandled)
        transitionKeyDown(key)
        keyBeingHandled = key
      end
    end
  elseif (event.phase == TouchManager.TOUCH_ENDED) then
    if tapIDBeingHandled == tapID then
      activateKey(self, keyBeingHandled, Key.SUBMIT_TRANSITION_DELAY)
      keyBeingHandled = nil
      tapIDBeingHandled = nil
    end
  end

  self.tapIDBeingHandled = tapIDBeingHandled
  self.keyBeingHandled = keyBeingHandled
end

-- ----------------------------------------------------------------------------
-- Internal API
-- ----------------------------------------------------------------------------

function Keyboard:simulateKeyEvent(keyName, isDown)
  local key = self.keys[keyName]
  if key then
    if isDown then
      key:transitionDown()
    else
      key:transitionUp()
      key:activate()
    end
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Keyboard:reset()
  for _, key in pairs(self.keys) do
    key:reset()
  end
end

function Keyboard:getKey(name)
  local key = self.keys[name]
  assert(key, tostring(name) .. ": Unknown key")
  return key
end

function Keyboard:setKeyEnabled(name, isEnabled, willPersistEnabledState)
  if (willPersistEnabledState == nil) then
    willPersistEnabledState = true
  end

  local key = self:getKey(name)

  if willPersistEnabledState then
    self.persistentEnabledKeys[key] = isEnabled
  end

  key:setEnabled(isEnabled)
end

function Keyboard:setTouchEnabled(isEnabled)
  if isEnabled ~= self._isTouchEnabled then
    self._isTouchEnabled = isEnabled
    if isEnabled then
      TouchManager.addTouchHandler(self, function(event) handleTouch(self, event) end)
    else
      if self.keyBeingHandled then
        self.keyBeingHandled:reset()
      end
      TouchManager.removeTouchHandler(self)
    end
  end
end

--- Same as setTouchEnabled, except it also visually shows the "enabled" state.
function Keyboard:setTouchEnabledVisual(isEnabled)
  for _, key in pairs(self.keys) do
    local persistentIsEnabled = self.persistentEnabledKeys[key]
    if persistentIsEnabled == nil or persistentIsEnabled == isEnabled then
      key:setEnabled(isEnabled)
    end
  end
  self:setTouchEnabled(isEnabled)
end

function Keyboard:setLetterHandler(onLetter)
  self.onLetter = onLetter or function() end
end

function Keyboard:setBackspaceHandler(onBackspace)
  self.onBackspace = onBackspace or function() end
end

function Keyboard:setKeyHandler(keyName, onPressed)
  local key = self.keys[keyName]
  assert(key, tostring(key) .. ": unknown key")
  local handlerFunction = newHandlerFunction(self, keyName, Key.types.command, onPressed)
  key:setInputHandler(onPressed)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newKey(self, entry)
  local isEnabled = entry.isEnabled
  if isEnabled == nil then isEnabled = true end
  local width = entry.width
  local height = KEY_HEIGHT
  local keyType = entry.type or Key.types.normal

  if keyType == Key.types.padding then
    return Group.new(width, height + MARGIN_Y * 2)
  end

  local handlerFunction = newHandlerFunction(self, entry.name, keyType, entry.customFunction)
  local key = Key.new(entry.name, keyType, width, height, MARGIN_X, MARGIN_Y, handlerFunction)
  if not isEnabled then
    key:setEnabled(false)
    self.persistentEnabledKeys[key] = false
  end

  return key
end

local function createLayout(self, keyboardLayout)
  local maxW = 0
  local rows = {}
  self.keys = {}
  for _, layoutRow in ipairs(keyboardLayout) do
    local row = Row.new()
    for _, entry in ipairs(layoutRow) do
      local key = newKey(self, entry)
      row:insert(key)
      if entry.name then
        self.keys[entry.name] = key
      end
    end
    row:alignChildren()
    maxW = math.max(maxW, row:getWidth())
    self:insert(row)
    table.insert(rows, row)
  end

  for _, row in ipairs(rows) do
    local rowW = row:getWidth()
    if rowW < maxW and #row.children > 1 then
      local extraMargin = (maxW - rowW) / 2
      local leftKey, rightKey = row.children[1], row.children[#row.children]
      if leftKey.setVirtualTouchOverlay then
        leftKey:setVirtualTouchOverlay(extraMargin, 0)
      end
      if rightKey.setVirtualTouchOverlay then
        rightKey:setVirtualTouchOverlay(0, extraMargin)
      end
    end
  end
  self:alignChildren()
end

-- Constructor ------------------------

function Keyboard.new(keyboardLayout, onLetter, onBackspace)
  assert(PropertyLoader.requiresMultitouch(), "Field requiresMultitouch in properties.lua must be set to true to instantiate a Keyboard.")

  keyboardLayout = keyboardLayout or Keyboard.DEFAULT_LAYOUT
  local self = Column.new()
  Mixin.mix(self, Keyboard)

  self.persistentEnabledKeys = {}
  self.tapIDBeingHandled = nil
  self.keyBeingHandled = nil

  createLayout(self, keyboardLayout)
  self:setLetterHandler(onLetter)
  self:setBackspaceHandler(onBackspace)
  self:setTouchEnabled(true)

  return self
end
