--- NumberPad
-- @author Created by Jay Austria on May 5, 2015.

class("NumberPad")
NumberPadKey = privateClass()

require("graphics.primitives.RoundedRegularPolygon")

-- ===========================================================================
-- NumberPadKey
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

NumberPad.SpecialKeys =
{
  BACKSPACE = "SpecialKeys.BACKSPACE",
  SUBMIT = "SpecialKeys.SUBMIT"
}

local DEFAULT_KEY_FONT = Font.new(FontFaces.DIN_OT_LIGHT, 28)
local SUBMIT_COLOR = ColorUtils.colorHex24ToStd(0xffffff)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getScaleAction(view, scaleDuration, scaleAmount)
  return AHScaleTo.new(view, scaleDuration, AHEaseSimpleInOut.new(), scaleAmount)
end

local function animateKeyPress(self)
  local scaleUp = function()
    self.keyView:setScl(1.5)
  end
  self.animationInterrupt:insert(scaleUp)
end

local function animateKeyRelease(self)
  local scaleDuration = 0.2
  local scaleAmount = 1.0
  local scaleAction = getScaleAction(self.keyView, scaleDuration, scaleAmount)
  self.animationInterrupt:insert(scaleAction)
end

local function createKeyLayout(self, keyColor, keyFont, underlineThickness, keyViewOffsetY)
  local valueAsString = self.valueAsString
  local keyView = nil
  keyFont = keyFont or DEFAULT_KEY_FONT

  if valueAsString == NumberPad.SpecialKeys.BACKSPACE then
    keyView = Sprite.newWithName("shared_numberpad_backspace.png")
    keyView:setColor(unpack(keyColor))
    self:insertAtCenter(keyView)
  elseif valueAsString == NumberPad.SpecialKeys.SUBMIT then
    keyView = Label.new(Localizer.get("Misc.SUBMIT"), Font.new(FontFaces.DIN_OT_MEDIUM, 13))
    keyView:setColor(unpack(SUBMIT_COLOR))
    self:insertAtCenter(keyView)
  else
    local label = Label.new(valueAsString, keyFont)
    label:setColor(unpack(keyColor))

    local width, height = label:getDims()
    keyView = Group.new(width, height)
    keyView:setAnchorPointBottom()
    keyView:insertAtCenter(label)

    local defaultUnderlineThickness = (1 / Host.contentScale)
    underlineThickness = underlineThickness or defaultUnderlineThickness

    local underline = FilledRectangle.new(self:getWidthPart(0.4), underlineThickness, keyColor)

    local defaultKeyViewOffsetY = keyView:getHeightPart(0.1)
    keyViewOffsetY = keyViewOffsetY or defaultKeyViewOffsetY

    underline:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.5) - keyView:getHeightPart(0.5) - keyViewOffsetY)
    self:insert(underline)

    keyView:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.5) - label:getHeightPart(0.5))
    self:insert(keyView)
  end

  self.keyView = keyView
end

local function initializeButton(self, extendedXLeft, extendedXRight, extendedYTop, extendedYBottom)
  local pressedCallback = self.pressedCallback
  extendedXLeft = extendedXLeft or 0
  extendedXRight = extendedXRight or 0
  extendedYTop = extendedYTop or 0
  extendedYBottom = extendedYBottom or 0

  local onPress = function()
    if pressedCallback then
      runInCoroutine(function()
        pressedCallback(self)
      end)
    end

    animateKeyPress(self)
  end

  local onRelease = function()
    animateKeyRelease(self)
  end

  local onCancel = function()
    animateKeyRelease(self)
  end

  local width = self:getWidth() + (extendedXLeft + extendedXRight)
  local height = self:getHeight() + (extendedYTop + extendedYBottom)
  local tapRegion = Group.new(width, height)
  local regionX = self:getWidthPart(0.5) - extendedXLeft/2 + extendedXRight/2
  local regionY = self:getHeightPart(0.5) + extendedYTop/2 - extendedYBottom/2
  tapRegion:setLoc(regionX, regionY)
  self:insert(tapRegion)

  Button.makeButton(tapRegion, onPress, onRelease, onCancel)

  self.tapRegion = tapRegion
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function NumberPadKey:reset()
  animateKeyRelease(self)
end

function NumberPadKey:setTouchEnabled(enable)
  self.tapRegion:setTouchEnabled(enable)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function NumberPadKey.new(width, height, valueAsString, pressedCallback, extendedXLeft, extendedXRight, extendedYTop, extendedYBottom, keyColor, keyFont, underlineThickness, keyViewOffsetY)
  local self = Group.new(width, height)
  Mixin.mix(self, NumberPadKey)

  self.valueAsString = valueAsString
  self.pressedCallback = pressedCallback
  self.animationInterrupt = ActionInterrupt.new()

  createKeyLayout(self, keyColor, keyFont, underlineThickness, keyViewOffsetY)
  initializeButton(self, extendedXLeft, extendedXRight, extendedYTop, extendedYBottom)

  return self
end

-- ===========================================================================
-- NumberPad
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local NUMBER_PAD_HEIGHT = Screen.getSafeHeightPart(0.47)

local TOP_MARGIN_Y = 10
local BOTTOM_MARGIN_Y = 10
local SIDES_MARGIN_X = 20
local BUTTON_PADDING = 1

local DEFAULT_NUMBER_PAD_LAYOUT =
{
  {"1", "2", "3"},
  {"4", "5", "6"},
  {"7", "8", "9"},
  {NumberPad.SpecialKeys.BACKSPACE, "0", NumberPad.SpecialKeys.SUBMIT}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function registerKeyPressed(self, key)
  if self.onButtonPressCallback then
    self.onButtonPressCallback(key.valueAsString)
  end

  if key.valueAsString == NumberPad.SpecialKeys.BACKSPACE then
    self.inputField:backspace()
  elseif key.valueAsString == NumberPad.SpecialKeys.SUBMIT then
    self:_registerSubmit(key)
  else
    self.inputField:append(key.valueAsString)
  end
end

local function createLayout(self, keyColor, keyFont, numberPadLayout, underlineThickness, keyViewOffsetY)
  local rows = numberPadLayout
  local rowsCount = #rows
  local columnsCount = #rows[1] -- eh, close enough amirite?

  local keyAreaWidth = self:getWidth() - (2 * SIDES_MARGIN_X)
  local keyAreaHeight = self:getHeight() - (BOTTOM_MARGIN_Y + TOP_MARGIN_Y)

  local keyWidth = (keyAreaWidth - ((columnsCount - 1) * BUTTON_PADDING)) / columnsCount
  local keyHeight = (keyAreaHeight - ((rowsCount - 1) * BUTTON_PADDING)) / rowsCount

  local defaultPaddingSize = 50

  local columnGroup = Column.new()
  for rowIndex, row in ipairs(rows) do
    local rowGroup = Row.new()

    local extendedTapYTop = 0
    local extendedTapYBottom = 0

    if rowIndex == rowsCount then
      extendedTapYBottom = defaultPaddingSize
    end

    for columnIndex = 1, columnsCount do
      local extendedTapXLeft = 0
      local extendedTapXRight = 0

      if columnIndex == 1 then
        extendedTapXLeft = defaultPaddingSize
      elseif columnIndex == columnsCount then
        extendedTapXRight = defaultPaddingSize
      end

      local keyValue = row[columnIndex]
      local pressedCallback = function(returnKey) registerKeyPressed(self, returnKey) end

      local key = NumberPadKey.new(keyWidth, keyHeight, keyValue, pressedCallback, extendedTapXLeft, extendedTapXRight, extendedTapYTop, extendedTapYBottom, keyColor, keyFont, underlineThickness, keyViewOffsetY)
      rowGroup:insert(key)
      table.insert(self.interactableElements, key)
    end

    rowGroup:alignChildren()
    columnGroup:insert(rowGroup)
  end

  columnGroup:alignChildren()
  columnGroup:setLoc(self:getWidthPart(0.5), BOTTOM_MARGIN_Y + self:getHeightPart(0.5))
  self:insert(columnGroup)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Protected API --------------------------------------------------------------

function NumberPad:_registerSubmit(key)
  if self.onSubmitCallback then
    local fieldValue = tonumber(self.inputField:getValue())
    self.onSubmitCallback(fieldValue)
  end
end

-- Public API -----------------------------------------------------------------

function NumberPad:setTouchEnabled(enabled)
  for _, interactableElement in ipairs(self.interactableElements) do
    interactableElement:setTouchEnabled(enabled)
  end
end

function NumberPad:reset()
  for _, interactableElement in ipairs(self.interactableElements) do
    interactableElement:reset()
  end
end

function NumberPad:setOnSubmitCallback(callback)
  self.onSubmitCallback = callback
end

function NumberPad:setOnButtonPressCallback(callback)
  self.onButtonPressCallback = callback
end

function NumberPad:attachInputField(inputField)
  self.inputField = inputField
end

function NumberPad:getInputFieldValue()
  return self.inputField:getValue()
end

function NumberPad:forceZeroInput()
  self.inputField:append(0)
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function NumberPad.new(keyColor, keyFont, numberPadLayout, underlineThickness, keyViewOffsetY)
  local self = Group.new(Screen.safeWidth, NUMBER_PAD_HEIGHT)
  Mixin.mix(self, NumberPad)

  self.interactableElements = {}

  local layout = numberPadLayout or DEFAULT_NUMBER_PAD_LAYOUT

  createLayout(self, keyColor, keyFont, layout, underlineThickness, keyViewOffsetY)
  self:setTouchEnabled(false)

  return self
end
