--- NumberPadInputField
-- @author Created by Jay Austria on May 6, 2015.

class("NumberPadInputField")

require("actions.advanced.Shake")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

NumberPadInputField.DEFAULT_MAX_DIGITS = 5
local VISIBLE_OPACITY = 1.0
local DEFAULT_CHARACTERS = "0123456789"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function doShowAnimation(view)
  view:setScl(1.8)
  view:setOpacity(0.0)

  local animationDuration = 0.35

  local fadeIn = AHOpacityTo.new(view, animationDuration, nil, VISIBLE_OPACITY)
  local scaleDown = AHScaleTo.new(view, animationDuration, AHEaseSimpleOut.new(), 1.0)
  local showAnimation = AHSpawn.new(fadeIn, scaleDown)

  showAnimation:start()
end

local function getMargins(font, label)
  local glyphBuildersByCharacter = font.glyphBuildersByCharacter
  local glyphBuilder = glyphBuildersByCharacter[label:getText()]

  return glyphBuilder.leftMargin, glyphBuilder.rightMargin
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Protected API --------------------------------------------------------------

function NumberPadInputField:_createViewForValueString(valueAsString)
  local valueLabel = Label.new(valueAsString, self.font, nil, Alignment.CENTER)
  valueLabel:setColor(unpack(self.defaultColor))

  local leftMargin, rightMargin = getMargins(self.font, valueLabel)
  local combinedWidth = leftMargin + valueLabel:getWidth() + rightMargin
  local viewForValue = Group.new(combinedWidth, valueLabel:getHeight())
  viewForValue:insertAtCenter(valueLabel)
  viewForValue.valueLabel = valueLabel

  return viewForValue
end

function NumberPadInputField:_getMaxDigits()
  return NumberPadInputField.DEFAULT_MAX_DIGITS
end

-- Public API --------------------------------------------------------------

function NumberPadInputField:setInvalidInputCallback(callback)
  self.invalidInputCallback = callback
end

function NumberPadInputField:setValidInputCallback(callback)
  self.validInputCallback = callback
end

function NumberPadInputField:reactToInvalidInput()
  local xShakeAmp = 10
  local yShakeAmp = 0
  local shakesPerSecond = 3
  local shakeSeconds = 0.5
  local shakeAction = AHSmoothShake.new(self, xShakeAmp, yShakeAmp, shakesPerSecond, shakeSeconds)

  local haveCapturedInitialPosition = (self.initialX ~= nil and self.initialY ~= nil)
  if (not haveCapturedInitialPosition) then
    self.initialX = self:getX()
    self.initialY = self:getY()
  end

  local doneShaking = function()
    self:setLoc(self.initialX, self.initialY)
  end

  local sequence = AHSequence.new(shakeAction, doneShaking)
  sequence:start()

  if self.invalidInputCallback then
    self.invalidInputCallback()
  end
end

function NumberPadInputField:append(displayedValueAsString)
  local newView = self:_createViewForValueString(displayedValueAsString)

  local isLeadingZero = ((#self.viewsRow.children == 0) and (displayedValueAsString == '0'))
  local canAppend = (#self.viewsRow.children < self:_getMaxDigits()) and not isLeadingZero

  if canAppend then
    self.viewsRow:insert(newView)
    self.viewsRow:alignChildren()
    doShowAnimation(newView)

    self.displayedValueAsString = self.displayedValueAsString .. displayedValueAsString

    if self.validInputCallback then
      self.validInputCallback()
    end
  else
    self:reactToInvalidInput()
  end
end

function NumberPadInputField:backspace()
  local isEmpty = (#self.viewsRow.children == 0)
  local willNeedAlignment = (#self.viewsRow.children > 1)

  if not isEmpty then
    local lastDigit = self.viewsRow.children[#self.viewsRow.children]
    self.viewsRow:remove(lastDigit)

    if willNeedAlignment then
      self.viewsRow:alignChildren()
    end

    self.displayedValueAsString = string.sub(self.displayedValueAsString, 1, #self.displayedValueAsString - 1)

    if self.validInputCallback then
      self.validInputCallback()
    end
  else
    self:reactToInvalidInput()
  end
end

function NumberPadInputField:clear()
  self.viewsRow:removeAll()
  self.displayedValueAsString = ""
end

function NumberPadInputField:getValue()
  return self.displayedValueAsString
end

-- Warning: We must instantiate a Label in order to get the height of any views
-- that will be created in _createViewForValueString(). To mitigate the performance,
-- impact, the return value will be cached. Still, be mindful of the timing of when
-- this method gets called.
function NumberPadInputField:getViewHeight()
  if (self.cachedViewHeight ~= nil) then
    return self.cachedViewHeight
  end

  local dummyValueString = "1"
  local dummyValueLabel = Label.new(dummyValueString, self.font, nil, Alignment.CENTER)

  self.cachedViewHeight = dummyValueLabel:getHeight()

  return self.cachedViewHeight
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local viewSpacing = 0
  local viewsRow = Row.new(viewSpacing)
  self:insertAtCenter(viewsRow)
  self.viewsRow = viewsRow
end

-- Constructor ------------------------

function NumberPadInputField.new(height, fontFace, fontSize, defaultColor, incorrectColor, characters)
  local self = Group.new(Screen.safeWidth, height)
  Mixin.mix(self, NumberPadInputField)

  characters = characters or DEFAULT_CHARACTERS
  self.font = PrerenderedFont.new(fontFace, fontSize, characters)
  self.displayedValueAsString = ""
  self.defaultColor = defaultColor or {1.0, 1.0, 1.0, 1.0}
  self.incorrectColor = incorrectColor

  createLayout(self)

  return self
end
