--- Bar progress indicator.
-- Shows a 2-color bar progress indicator that can grow in any direction.

-- ===========================================================================
class("BarProgressIndicator")
-- ===========================================================================

require("widgets.progress.ProgressIndicator")
require("widgets.progress.ProgressTimer")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

BarProgressIndicator.GrowDown  = 1
BarProgressIndicator.GrowUp    = 2
BarProgressIndicator.GrowLeft  = 3
BarProgressIndicator.GrowRight = 4


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

function BarProgressIndicator:setProgress(progress)
  self.super.setProgress(self, progress)

  local progress = self:getProgress()
  if self.isScaleX then
    self.progressBar:setScl(progress, 1)
  else
    self.progressBar:setScl(1, progress)
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

--- Make a bar progress indicator.
-- @param width The width
-- @param height The height
-- @param fgColor The color for the foreground rectangle.
-- @param bgColor The color for the background rectangle.
-- @param growDirection The direction the progress indicator will grow in (see Configuration section)
-- @param stepCount The number of steps. Can leave nil if you aren't going to use step functions.
function BarProgressIndicator.new(width, height, fgColor, bgColor, growDirection, stepCount)
  local self = Group.new(width, height)
  ProgressIndicator.makeProgressIndicator(self, stepCount)
  Mixin.mix(self, BarProgressIndicator)

  local selfW, selfH = self:getDims()

  local anchorX, anchorY
  local locX, locY
  local isScaleX
  if growDirection == BarProgressIndicator.GrowDown then
    anchorX, anchorY = 0.5, 1
    locX, locY = selfW/2, selfH
    isScaleX = false
  elseif growDirection == BarProgressIndicator.GrowUp then
    anchorX, anchorY = 0.5, 0
    locX, locY = selfW/2, 0
    isScaleX = false
  elseif growDirection == BarProgressIndicator.GrowLeft then
    anchorX, anchorY = 1, 0.5
    locX, locY = selfW, selfH/2
    isScaleX = true
  elseif growDirection == BarProgressIndicator.GrowRight then
    anchorX, anchorY = 0, 0.5
    locX, locY = 0, selfH/2
    isScaleX = true
  end

  local bgBar = FilledRectangle.new(selfW, selfH, bgColor)
  bgBar:setAnchorPoint(anchorX, anchorY)
  bgBar:setLoc(locX, locY)
  self:insert(bgBar)

  local fgBar = FilledRectangle.new(selfW, selfH, fgColor)
  fgBar:setAnchorPoint(anchorX, anchorY)
  fgBar:setLoc(locX, locY)
  self:insert(fgBar)
  self:linkColorAttrToChildren()

  self.backing = bgBar
  self.progressBar = fgBar
  self.isScaleX = isScaleX

  return self
end


--- Make a bar progress timer.
-- @param (mandatory) width The width
-- @param (mandatory) height The height
-- @param (mandatory) durationSeconds
-- @param (optional)  onCompletion - callback called after the timer is down at 0
-- @param (optional)  fgColor The color for the foreground rectangle.
-- @param (optional)  bgColor The color for the background rectangle.
-- @param (optional)  growDirection The direction the progress indicator will grow in (see Configuration section)
-- @param (optional)  stepCount The number of steps. Can leave nil if you aren't going to use step functions.
function BarProgressIndicator.newTimer(width, height, durationSeconds, onCompletion, fgColor, bgColor, growDirection, stepCount)
  growDirection = growDirection or BarProgressIndicator.GrowUp

  local self = BarProgressIndicator.new(width, height, fgColor, bgColor, growDirection, stepCount)
  ProgressTimer.makeTimer(self, durationSeconds, onCompletion)

  return self
end
