--- DigitTimer
-- @author Created by andrewc on August 28, 2020.

class("DigitTimer")

require("widgets.progress.TimeTimer")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_FONT = FontFaces.DIN_OT_BOLD
local DEFAULT_TIME_FONT_SIZE = 20
local DEFAULT_PAUSED_FONT_SIZE_MULTIPLER = 0.8

local DEFAULT_PAUSED_OPACITY = 0.5
local DEFAULT_ACTIVE_OPACITY = 1.0

local DEFAULT_FADE_DURATION = 0.333
local DEFAULT_FADE_EASE = AHEaseBezier.newFlat({0.7, 0, 0.3, 1})

local TIME_LOST_COLOR = {1, 0, 0}
local TIME_RUN_OUT_COLOR = {0.75, 0, 0}
local FLASH_DURATION = 0.5

local LABEL_MOVE_OFFSET = -5
local ERROR_BLIP_SCALE_REDUCTION_FOR_DIGIT_LOSS = 0.8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isEqualWithEpsilon(a, b)
  local epsilon = 0.00001
  return (math.abs(a - b) < epsilon)
end

local function createPrepareForTimeLostAnimationFunction(self, amount)
  local timeLostString = string.format("-%d ", amount) .. Localizer.get("DigitTimer.SECONDS")
  return function()
    self.timeLostLabel:setText(timeLostString)
    self.timeLostLabel:setOpacity(0)
    self.errorBlip:setScl(0.7 * self.errorBlipTargetScale)
    self.errorBlip:setOpacity(0)
  end
end

local function createTimeLabelColorChangeAnimation(self)
  local timeFadeDelay = 0.667
  local timeFadeDuration = DEFAULT_FADE_DURATION
  local timeFadeEase = DEFAULT_FADE_EASE
  local colorChangeDelay = 0.05
  local labelAnimation = function()
    waitFor(colorChangeDelay)
    self:setColor(unpack(TIME_LOST_COLOR))
    waitFor(timeFadeDelay - colorChangeDelay)
    local targetColor = self.timerColor
    if self:getPaused() then
      targetColor[4] = self:getPausedOpacity()
    else
      targetColor[4] = self:getActiveOpacity()
    end

    wait(AHColorTo.new(self, timeFadeDuration, timeFadeEase, targetColor):start())
  end
  return labelAnimation
end

local function createShowLabelAnimation(label, opacity, duration)
  local duration = duration or DEFAULT_FADE_DURATION
  local ease = DEFAULT_FADE_EASE
  local distance = LABEL_MOVE_OFFSET
  local prepareForAnimation = function() label:offsetX(-LABEL_MOVE_OFFSET) end
  local moveOnScreenAnimation = AHLocationXBy.new(label, duration, ease, distance)
  local fadeInAnimation = AHOpacityTo.new(label, duration, ease, opacity)
  local showAnimation = AHSpawn.new(moveOnScreenAnimation, fadeInAnimation)
  return AHSequence.new(prepareForAnimation, showAnimation)
end

local function createHideLabelAnimation(label)
  local duration = DEFAULT_FADE_DURATION
  local ease = DEFAULT_FADE_EASE
  local distance = LABEL_MOVE_OFFSET
  local moveOffScreenAnimation = AHLocationXBy.new(label, duration, ease, distance)
  local fadeOutAnimation = AHOpacityTo.new(label, duration, ease, 0)
  local hideAnimation = AHSpawn.new(moveOffScreenAnimation, fadeOutAnimation)
  local cleanupAfterAnimation = function() label:offsetX(-LABEL_MOVE_OFFSET) end
  return AHSequence.new(hideAnimation, cleanupAfterAnimation)
end

local function createTimeLostLabelAnimation(self)
  local holdDuration = DEFAULT_FADE_DURATION

  local timeLostOpacity = self:getOpacity()
  if (self.timeLostOpacityOverride ~= nil) then
    timeLostOpacity = self.timeLostOpacityOverride
  end

  local showAnimation = createShowLabelAnimation(self.timeLostLabel, timeLostOpacity)
  local hideAnimation = createHideLabelAnimation(self.timeLostLabel)

  return AHSequence.new(
    showAnimation,
    AHDelayed.new(holdDuration, hideAnimation))
end

local function createErrorBlipAnimation(self)
  local scaleDuration = 1
  local scaleEase = AHEaseBezier.newFlat({0, 0.2, 0.5, 1})
  local scaleAnimation = AHScaleTo.new(self.errorBlip, scaleDuration, scaleEase, self.errorBlipTargetScale)

  local fadeInDuration = 0.05
  local fadeOutDuration = 0.95
  local fadeInAnimation = AHOpacityTo.new(self.errorBlip, fadeInDuration, nil, 0.3)
  local fadeOutAnimation = AHOpacityTo.new(self.errorBlip, fadeOutDuration, nil, 0)
  local fadeAnimation = AHSequence.new(fadeInAnimation, fadeOutAnimation)

  return AHSpawn.new(scaleAnimation, fadeAnimation)
end

local function markLostTime(self, amount)
  local prepareForAnimation = createPrepareForTimeLostAnimationFunction(self, amount)

  local fadeTimeBackToColor = createTimeLabelColorChangeAnimation(self)
  local lostTimeLabelAnimation = createTimeLostLabelAnimation(self)
  local errorBlipAnimation = createErrorBlipAnimation(self)

  local lostTimeAnimation = AHSequence.new(prepareForAnimation,
    AHSpawn.new(
      fadeTimeBackToColor,
      lostTimeLabelAnimation,
      errorBlipAnimation
    )
  )

  self.animationQueue:insert(lostTimeAnimation)
end

local function pauseAndKeepOpacity(self)
  local currentOpacity = self:getOpacity()
  self.TimeTimer.pause(self)
  self:setOpacity(currentOpacity)
end

local function resumeAndKeepOpacity(self)
  local currentOpacity = self:getOpacity()
  self.TimeTimer.resume(self)
  self:setOpacity(currentOpacity)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Add time towards completion of the timer.
-- This subtracts the amount from the time remaining.
function DigitTimer:addTime(amount)
  if (amount > 0) then
    markLostTime(self, amount)
  end
  self.TimeTimer.addTime(self, amount)
end

-- This should really be pauseAndShowLabel, but we're matching PieTimer's API
function DigitTimer:pauseAndShowIcon(fadeDuration)
  pauseAndKeepOpacity(self)
  fadeDuration = fadeDuration or DEFAULT_FADE_DURATION

  local playAnimationsAsNeeded = function()
    local fullAnimation = AHSpawn.new()
    if not isEqualWithEpsilon(self.pausedLabel:getOpacity(), self:getPausedOpacity()) then
      fullAnimation:addAction(createShowLabelAnimation(self.pausedLabel, self:getPausedOpacity(), fadeDuration))
    end
    if not isEqualWithEpsilon(self:getOpacity(), self:getPausedOpacity()) then
      fullAnimation:addAction(AHOpacityTo.new(self, fadeDuration, DEFAULT_FADE_EASE, self:getPausedOpacity()))
    end
    wait(fullAnimation:start())
  end

  self.animationQueue:insert(playAnimationsAsNeeded)
end

-- This should really be resumeAndHideLabel, but we're matching PieTimer's API
function DigitTimer:resumeAndHideIcon(fadeDuration)
  self:resume(fadeDuration)
end

function DigitTimer:pause()
  pauseAndKeepOpacity(self)
  local fadeOutTime = AHOpacityTo.new(self, DEFAULT_FADE_DURATION, DEFAULT_FADE_EASE, self:getPausedOpacity())
  local fadeIfNeeded = function()
    if not isEqualWithEpsilon(self:getOpacity(), self:getPausedOpacity()) then
      wait(fadeOutTime:start())
    end
  end
  self.animationQueue:insert(fadeIfNeeded)
end

function DigitTimer:resume(fadeDuration)
  fadeDuration = fadeDuration or DEFAULT_FADE_DURATION
  resumeAndKeepOpacity(self)

  local playAnimationsAsNeeded = function()
    local fullAnimation = AHSpawn.new()
    if not isEqualWithEpsilon(self:getOpacity(), self:getActiveOpacity()) then
      fullAnimation:addAction(AHOpacityTo.new(self, fadeDuration, DEFAULT_FADE_EASE, self:getActiveOpacity()))
    end
    if not isEqualWithEpsilon(self.pausedLabel:getOpacity(), 0) then
      fullAnimation:addAction(createHideLabelAnimation(self.pausedLabel))
    end
    wait(fullAnimation:start())
  end

  self.animationQueue:insert(playAnimationsAsNeeded)
end

function DigitTimer:getKeyColor()
  return self.timerColor
end

function DigitTimer:setKeyColor(color)
  self.timerColor = color
end

function DigitTimer:setTimeLostOpacityOverride(opacity)
  self.timeLostOpacityOverride = opacity
end

-- Called by UIOverlay
function DigitTimer:createFadeInAction(duration)
  local fadeInAnimation = AHOpacityTo.new(self, duration, nil, self:getActiveOpacity())
  if self.shouldFadeInWithPauseLabel then
    fadeInAnimation = AHSpawn.new(
      AHOpacityTo.new(self, duration, nil, self:getPausedOpacity()),
      AHOpacityTo.new(self.pausedLabel, duration, nil, self:getPausedOpacity())
    )
  end
  local callback = function()
    self.animationQueue:insert(fadeInAnimation)
  end
  return callback
end

function DigitTimer:createFadeOutAction(duration)
  local fadeOutAnimation = AHSpawn.new(function()
    self.animationQueue:stop()
  end)
  for _, element in ipairs({self, self.pausedLabel, self.timeLostLabel, self.errorBlip}) do
    fadeOutAnimation:addAction(AHOpacityTo.new(element, duration, nil, 0))
  end
  return fadeOutAnimation
end

function DigitTimer:addVisualWarningSequence()
  self:setWarningInterval(1)
  self:setWarningCount(5)
  self:addWarningHandler(function()
    local originalColor = {self:getColor()}
    self:setColor(unpack(TIME_RUN_OUT_COLOR))
    AHColorTo.new(self, FLASH_DURATION, nil, originalColor):start()
  end)
end

function DigitTimer:getPausedLabel()
  return self.pausedLabel
end

function DigitTimer:setShouldFadeInWithPauseLabel(shouldFadeInWithPauseLabel)
  self.shouldFadeInWithPauseLabel = shouldFadeInWithPauseLabel
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, font, fontSize, color)
  local labelFont = Font.newWithActualSize(font, fontSize * DEFAULT_PAUSED_FONT_SIZE_MULTIPLER)
  local pausedLabel = Label.new(Localizer.get("DigitTimer.PAUSED"), labelFont)
  pausedLabel:setColor(unpack(color))
  pausedLabel:setOpacity(0)
  pausedLabel:setAnchorPointTopRight()
  self:insertAtBottomRight(pausedLabel)

  local timeLostLabel = Label.new("", labelFont)
  timeLostLabel:setOpacity(0)
  timeLostLabel:setColor(unpack(TIME_LOST_COLOR))
  timeLostLabel:setAnchorPointTopRight()
  self:insertAtBottomRight(timeLostLabel)

  local radius = math.max(self:getWidth(), self:getHeight()) * 0.85
  local errorBlip = FilledCircle.new(radius, TIME_LOST_COLOR)
  errorBlip:setOpacity(0)
  self:insertAtCenter(errorBlip)
  self.errorBlipTargetScale = 1

  self:setMinuteLabelRemovedCallback(function()
    self.errorBlipTargetScale = ERROR_BLIP_SCALE_REDUCTION_FOR_DIGIT_LOSS
    local offsetAmount = errorBlip:getWidth() * (1 - self.errorBlipTargetScale) / 4
    errorBlip:offsetX(offsetAmount)
  end)

  self.pausedLabel = pausedLabel
  self.timeLostLabel = timeLostLabel
  self.errorBlip = errorBlip
end

-- Constructor ------------------------

function DigitTimer.new(duration, onTimeout, font, fontSize, color)
  font = font or DEFAULT_FONT
  fontSize = fontSize or DEFAULT_TIME_FONT_SIZE
  local useMinuteLabel = TimeDisplay.USE_MINUTE_LABEL.ONLY_AT_START
  local self = TimeTimer.new(font, fontSize, duration, onTimeout, useMinuteLabel)
  Mixin.mix(self, DigitTimer, "TimeTimer")

  self:setColor(unpack(color))
  local activeOpacity = color[4] or DEFAULT_ACTIVE_OPACITY
  self:setPausedOpacity(DEFAULT_PAUSED_OPACITY)
  self:setActiveOpacity(activeOpacity)
  self.timeLostOpacityOverride = nil

  createLayout(self, font, fontSize, color)

  self.timerColor = color
  self.shouldFadeInWithPauseLabel = false

  self.animationQueue = ActionQueue.new()

  return self
end
