--- Pie Timer

-- ===========================================================================
class("PieTimer")
-- ===========================================================================

require("widgets.progress.PieProgressIndicator")
require("widgets.progress.ProgressTimer")

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

local COLOR_WARNING            = {0.94, 0.38, 0.32, 0.0}
local COLOR_LOST_TIME_PIE      = {1, 0, 0, 1}
local COLOR_LOST_TIME_RING     = {1, 0, 0, 0.5}

local FLASH_DURATION           = 0.25
local FLASH_IN_OPACITY         = 1.0
local FLASH_OUT_OPACITY        = 0.0

local TIME_LOST_RING_SIZE_MULT = 2.0

local DEFAULT_LABEL_FADE_DURATION = 0.1

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function showTimeLostArc(self, oldNormTimeRemaining, newNormTimeRemaining)
  return runInCoroutine(function()
    local startAngle = self:getAngleForProgress(newNormTimeRemaining)
    local endAngle = self:getAngleForProgress(oldNormTimeRemaining)

    local arc = FilledArc.new(self.radius, startAngle, endAngle, COLOR_LOST_TIME_PIE)
    self:insertAtCenter(arc, self:getTopZIndex() + 1)

    waitFor(0.2)
    wait(AHOpacityTo.new(arc, 0.5, nil, 0):start())
    arc:removeSelf()
  end)
end

local function showTimeLostRing(self)
  return runInCoroutine(function()
    local radius = math.min(self.radius * TIME_LOST_RING_SIZE_MULT)
    local ring = FilledCircle.new(radius, COLOR_LOST_TIME_RING)
    ring:setEffectiveDims(self:getDims())
    self:insertAtCenter(ring, self:getBottomZIndex() - 1)
    local ringAction = AHSpawn.new(
      AHScaleTo.new(ring, 0.6, nil, 1),
      AHOpacityTo.new(ring, 0.6, nil, 0)
      )
    wait(ringAction:start())
    ring:removeSelf()
  end)
end

local function markLostTime(self, amount)
  local currentTime = self:getTime()
  local duration = self:getDuration()
  return runInCoroutine(function()
    local newTime = math.min(currentTime + amount, duration)
    local currentNormTimeRemaining = 1 - currentTime / duration
    local newNormTimeRemaining = 1 - newTime / duration

    local arcAction = showTimeLostArc(self, currentNormTimeRemaining, newNormTimeRemaining)
    local ringsAction = showTimeLostRing(self)

    self.pie:setColor(unpack(COLOR_LOST_TIME_PIE))
    local pieAction = AHColorTo.new(self.pie, 0.5, nil, self.fillColor):start()
    wait(pieAction)
    wait(arcAction)
    wait(ringsAction)
  end)
end

local function getMidZIndex(self)
  local low = self:getBottomZIndex()
  local high = self:getTopZIndex()
  return low + (high - low) / 2
end

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function PieTimer:addTime(amount)
  if amount > 0 then
    markLostTime(self, amount)
  end
  self.super.addTime(self, amount)
end

--- Add only the visual components of the warning sequence.
-- Note: You must still set the interval and count.
function PieTimer:addVisualWarningSequence()
  local selfW, selfH = self:getDims()
  self.warningBacking = FilledCircle.new(self.radius, COLOR_WARNING)
  self.warningBacking:setLoc(selfW * 0.5, selfH * 0.5)
  self:insert(self.warningBacking, getMidZIndex(self))

  self:addWarningHandler(function()
    self.warningBacking:setOpacity(FLASH_OUT_OPACITY)
    local flashOutAction = AHOpacityTo.new(self.warningBacking, FLASH_DURATION, AHEaseSimpleOut.new(), FLASH_IN_OPACITY)
    local flashInAction = AHOpacityTo.new(self.warningBacking, FLASH_DURATION, nil, FLASH_OUT_OPACITY)
    local flashSequence = AHSequence.new(flashOutAction, flashInAction)

    flashSequence:start()
  end)
end

--- Add the standard warning sequence (visual and autible).
-- @param count (number) - The number of warning sequences at the end of the timer.
-- @param interval (number) - The number of seconds between warning sequences.
function PieTimer:addStandardWarningSequence(count, interval)
  self:setWarningInterval(interval or 1)
  self:setWarningCount(count or 5)

  self:addVisualWarningSequence()
  self:addStandardAudibleWarningSequence()
end

function PieTimer:pauseAndShowIcon(duration)
  self:pause()
  duration = duration or DEFAULT_LABEL_FADE_DURATION
  local fadeInIcon = AHOpacityTo.new(self.pausedIcon, duration, nil, 1.0)
  self.labelInterrupt:insert(fadeInIcon)
end

function PieTimer:resumeAndHideIcon(duration)
  self:resume()
  duration = duration or DEFAULT_LABEL_FADE_DURATION
  local fadeOutIcon = AHOpacityTo.new(self.pausedIcon, duration, nil, 0.0)
  self.labelInterrupt:insert(fadeOutIcon)
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

-- Constructor ------------------------

--- Creates a pie-shaped timer.
-- When the value is at maximum, it takes the shape of a circle. As the value decreases, proportionate slices get "cut out"
-- @param radius - (float) the radius of the circle representing the progress pie
-- @param backColor - (table) background color in [0, 1] space. Alpha is optional, defaults to 1
-- @param fillColor - (table) fill color in [0, 1] space. Alpha is optional, defaults to 1
-- @param duration - (float) Timer duration
-- @param onCompletion - (function) Called when the timer completes.
function PieTimer.new(radius, fillColor, backOpacity, duration, onCompletion)
  local self = PieProgressIndicator.new(radius, fillColor, backOpacity)
  ProgressTimer.makeTimer(self, duration, onCompletion)
  Mixin.mix(self, PieTimer)

  local timerPausedIcon = Sprite.newWithName("shared_lock.png")
  timerPausedIcon:setOpacity(0.0)
  self:insertAtCenter(timerPausedIcon)
  self.pausedIcon = timerPausedIcon

  self.labelInterrupt = ActionInterrupt.new()

  return self
end
