--- Progress Indicator
--
-- Your progress indicator implementation must override either setProgress or setProgressStep.
-- You may also override setInitialProgress if you wish your indicator to support starting at
-- an arbitrary progress level.
--
-- Basic implementation:
--
-- function MyIndicator:setProgress(progress)
--  self.super.setProgress(self, progress)
--  progress = self:getProgress() -- progress may have been clamped.
--  -- TODO: visible manifestation of progress level change.
-- end
--
-- function MyIndicator.new()
--  local stepCount = 10
--  local self = something()
--  ProgressIndicator.makeProgressIndicator(self, stepCount) -- Call before other mixins.
--  Mixin.mix(self, MyIndicator)
-- end


-- ===========================================================================
class("ProgressIndicator")
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function stepAtProgress(self, progress)
  if not self._ProgressIndicator_stepCount then
    return 0
  end

  local rawStep = progress * self._ProgressIndicator_stepCount
  local step = math.floor(rawStep)
  local ceilStep = math.ceil(rawStep)
  if ceilStep - rawStep < 0.0001 then -- Handle rounding error.
    step = ceilStep
  end

  return step
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Sets this indicator's progress, without any animations etc that would
-- happen during normal progress level changes.
-- @param progress The progress level to show (0 - 1).
function ProgressIndicator:setInitialProgress(progress)
  progress = MathUtils.clamp(progress, 0, 1)
  self._ProgressIndicator_currentProgress = progress
  self._ProgressIndicator_currentStep = stepAtProgress(self, progress)
end

--- Get the amount of progress represented by the specified number of progress steps.
function ProgressIndicator:progressForSteps(steps)
  return MathUtils.clamp(steps / self._ProgressIndicator_stepCount, 0, 1)
end

--- Sets the indicator's progress normally.
-- Very small and very large values will be rounded to 0 or 1.
-- @param progress The new progress level (will be clamped to stay between 0 and 1).
function ProgressIndicator:setProgress(progress)
  if progress > 0.9999 then
    progress = 1
  elseif progress < 0.0001 then
    progress = 0
  end
  self._ProgressIndicator_currentProgress = progress
  self._ProgressIndicator_currentStep = stepAtProgress(self, progress)
end

function ProgressIndicator:getProgress()
  return self._ProgressIndicator_currentProgress
end

--- Set the current progress step.
-- @param step The step value (will be clamped to stay between 0 and stepCount)
function ProgressIndicator:setProgressStep(step)
  assert(self._ProgressIndicator_stepCount, "Must set step count before calling setProgressStep")
  step = MathUtils.clamp(step, 0, self._ProgressIndicator_stepCount)
  self:setProgress(step / self._ProgressIndicator_stepCount)
end

function ProgressIndicator:getProgressStep()
  return self._ProgressIndicator_currentStep
end

function ProgressIndicator:setStepCount(value)
  self._ProgressIndicator_stepCount = value
end

function ProgressIndicator:getStepCount()
  return self._ProgressIndicator_stepCount
end

function ProgressIndicator:incrementProgress()
  self:setProgressStep(self._ProgressIndicator_currentStep + 1)
end

function ProgressIndicator:decrementProgress()
  self:setProgressStep(self._ProgressIndicator_currentStep - 1)
end

function ProgressIndicator:newProgressTo(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getProgress, self.setProgress, ease, amount)
end

function ProgressIndicator:newProgressBy(duration, ease, amount)
  return AHNonAttributeBy.new(self, duration, self.getProgress, self.setProgress, ease, amount)
end

function ProgressIndicator:newProgressStepTo(duration, ease, amount)
  return self:newProgressTo(duration, ease, self:progressForSteps(amount))
end

function ProgressIndicator:newProgressStepBy(duration, ease, amount)
  return self:newProgressBy(duration, ease, self:progressForSteps(amount))
end

--- Turn an object into a progress indicator.
-- @param self The object to make into a progress indicator.
-- @param stepCount The number of steps. Can leave nil if you aren't going to use step functions.
function ProgressIndicator.makeProgressIndicator(self, stepCount)
  self._ProgressIndicator_currentProgress = 0
  self._ProgressIndicator_currentStep = 0
  self._ProgressIndicator_stepCount = stepCount
  Mixin.mix(self, ProgressIndicator)
end
