--- Displays timer value as minutes and seconds.

require("widgets.progress.ProgressIndicator")
require("widgets.progress.ProgressTimer")


-- ============================================================================
class("TimeDisplay")
-- ============================================================================

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

TimeDisplay.USE_MINUTE_LABEL = {
  ALWAYS = "USE_MINUTE_LABEL.ALWAYS",
  ONLY_AT_START = "USE_MINUTE_LABEL.ONLY_AT_START",
  NEVER = "USE_MINUTE_LABEL.NEVER"
}

local KERNING_PADDING_MULTIPLIER = 0.05
local COLON_PADDING_MULTIPLIER = 0.02

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function setTextIfNotEqualToValue(label, value)
  local isLabelEqualToValue = (label:getText() == tostring(value))
  if (not isLabelEqualToValue) then
    label:setText(tostring(value))
  end
end

local function shouldUseMinutePlace(self, timeRemaining)
  if (self.useMinuteLabel == TimeDisplay.USE_MINUTE_LABEL.ALWAYS) then
    return true
  elseif (self.useMinuteLabel == TimeDisplay.USE_MINUTE_LABEL.NEVER) then
    return false
  elseif (self.useMinuteLabel == TimeDisplay.USE_MINUTE_LABEL.ONLY_AT_START) then
    return (self.minuteLabel ~= nil)
  end
end

local function getDigitValues(self, timeRemaining)
  timeRemaining = math.floor(timeRemaining)
  local minuteValue = 0
  local useMinutePlace = shouldUseMinutePlace(self, timeRemaining)
  if useMinutePlace then
    minuteValue = math.floor(timeRemaining / 60)
    timeRemaining = timeRemaining % 60
  end
  local tensValue = math.floor(timeRemaining / 10)
  local onesValue = timeRemaining % 10
  return {
    minute = minuteValue,
    tens = tensValue,
    ones = onesValue }
end

local function setMinuteValue(self, minuteValue)
  if self.minuteLabel then
    if ((minuteValue == 0) and (self.useMinuteLabel ~= TimeDisplay.USE_MINUTE_LABEL.ALWAYS)) then
      self.minuteLabel:removeSelf()
      if self.minuteLabelRemovedCallback then
        self.minuteLabelRemovedCallback()
      end
      self.minuteLabel = nil
    else
      setTextIfNotEqualToValue(self.minuteLabel, minuteValue)
    end
  end
end

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function TimeDisplay:setTime(time)
  local digitValues = getDigitValues(self, time)

  setMinuteValue(self, digitValues.minute)
  setTextIfNotEqualToValue(self.tensLabel, digitValues.tens)
  setTextIfNotEqualToValue(self.onesLabel, digitValues.ones)
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createLayout(self, font)
  if (self.useMinuteLabel ~= TimeDisplay.USE_MINUTE_LABEL.NEVER) then
    local minuteLabel = Label.new(tostring(0), font)
    self:insert(minuteLabel)
    self.minuteLabel = minuteLabel
  end

  local colonLabel = Label.new(":", font)

  local colonColumn = Column.new()
  colonColumn:insert(colonLabel)
  local visuallyFudgeColonY = 4
  colonColumn:insertSpacer(visuallyFudgeColonY)
  colonColumn:alignChildren()

  self:insert(colonColumn)
  local colonPadding = font:getActualSize() * COLON_PADDING_MULTIPLIER
  self:insertSpacer(colonPadding)
  self.colonLabel = colonLabel

  local tensLabel = Label.new(tostring(0), font)
  self:insert(tensLabel)
  self.tensLabel = tensLabel

  local onesLabel = Label.new(tostring(0), font)
  self:insert(onesLabel)
  self.onesLabel = onesLabel

  self:alignChildren()

  local isRecursive = true
  self:linkColorAttrToChildren(isRecursive)

  return self
end

-- Passing in a PrerenderedFont is the most efficient, however there appears to
-- be an inconsistent error when trying to use the same FontFace to create two
-- different PrerenderedFonts. A regular Font can be used in this case.
function TimeDisplay.new(font, useMinuteLabel)
  local spacingPadding = font:getActualSize() * KERNING_PADDING_MULTIPLIER
  local self = Row.new(spacingPadding)
  Mixin.mix(self, TimeDisplay)

  self.useMinuteLabel = useMinuteLabel or TimeDisplay.USE_MINUTE_LABEL.ONLY_AT_START
  createLayout(self, font)

  return self
end

-- ===========================================================================
class("TimeProgressIndicator")
-- ===========================================================================

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

local WARNING_COLOR = {0.84, 0.38, 0.36, 1.0}
local FLASH_DURATION = 0.5
local TIME_CHARACTERS = "0123456789:"

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function TimeProgressIndicator:setProgress(progress)
  self.super.setProgress(self, progress)

  local timeLeft = self:getTimeRemaining()
  local digitValues = getDigitValues(self, timeLeft)

  setMinuteValue(self, digitValues.minute)
  setTextIfNotEqualToValue(self.tensLabel, digitValues.tens)
  setTextIfNotEqualToValue(self.onesLabel, digitValues.ones)
end

function TimeProgressIndicator:hide()
  self.tensLabel:setOpacity(0.0)
  self.onesLabel:setOpacity(0.0)
  self.colonLabel:setOpacity(0.0)
  if self.minuteLabel then
    self.minuteLabel:setOpacity(0.0)
  end
end

function TimeProgressIndicator:fadeWithStagger(staggerDuration, destinationOpacity)
  local animationSequence = AHSequence.new()
  if self.minuteLabel then
    animationSequence:addAction(AHOpacityTo.new(self.minuteLabel, staggerDuration, nil, destinationOpacity))
  end
  animationSequence:addAction(AHOpacityTo.new(self.colonLabel, staggerDuration, nil, destinationOpacity))
  animationSequence:addAction(AHOpacityTo.new(self.tensLabel, staggerDuration, nil, destinationOpacity))
  animationSequence:addAction(AHOpacityTo.new(self.onesLabel, staggerDuration, nil, destinationOpacity))
  animationSequence:start()
end

function TimeProgressIndicator:setMinuteLabelRemovedCallback(callback)
  self.minuteLabelRemovedCallback = callback
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

function TimeProgressIndicator.new(font, fontSize, duration, useMinuteLabel)
  local prerenderedFont = PrerenderedFont.newWithActualSize(font, fontSize, TIME_CHARACTERS)
  local self = TimeDisplay.new(prerenderedFont, useMinuteLabel)
  ProgressIndicator.makeProgressIndicator(self)
  Mixin.mix(self, TimeProgressIndicator)

  self:setTime(duration)

  return self
end

-- ============================================================================
class("TimeTimer")
-- ============================================================================

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function TimeTimer:addStandardWarningSequence()
  self:addVisualWarningSequence()
  self:addStandardAudibleWarningSequence()
end

function TimeTimer:addVisualWarningSequence()
  self:setWarningInterval(1)
  self:setWarningCount(5)
  self:addWarningHandler(function()
    local originalColor = {self:getColor()}
    self:setColor(unpack(WARNING_COLOR))
    AHColorTo.new(self, FLASH_DURATION, nil, originalColor):start()
  end)
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

function TimeTimer.new(font, fontSize, duration, onTimeout, useMinuteLabel)
  local self = TimeProgressIndicator.new(font, fontSize, duration, useMinuteLabel)
  ProgressTimer.makeTimer(self, duration, onTimeout)
  Mixin.mix(self, TimeTimer)

  return self
end
