--- ChatMessageBubble

class("ChatMessageBubble")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Keep in absolute pixels to avoid a bug on iPhone SE sizing where message bubbles
-- that are too small do not align correctly in the ChatMessageLogScroller
local PADDING_LEFT_RIGHT = 16
local PADDING_TOP = 12
local PADDING_BOTTOM = 16

local FONT_SIZE = 15

local BUBBLE_RADIUS = Screen.getHeightPart(0.0143)
local FLAG_WIDTH = Screen.getHeightPart(0.0143)
local FLAG_HEIGHT = Screen.getHeightPart(0.0237)
local FLAG_STEM_RADIUS = Screen.getHeightPart(0.003)

local SYSTEM_BUBBLE_COLOR = ColorUtils.colorHex24ToStd(0x009cf4)
local USER_BUBBLE_COLOR = ColorUtils.colorHex24ToStd(0xffffff)
local USER_BUBBLE_SHADOW_COLOR = ColorUtils.colorHex24ToStd(0xbbbbbb)
local DROP_SHADOW_OFFSET_Y = -Screen.getHeightPart(0.004)

local SYSTEM_TEXT_COLOR = ColorUtils.colorHex24ToStd(0xffffff)
local USER_TEXT_COLOR = ColorUtils.colorHex24ToStd(0x000000)

local TYPING_DOTS_COUNT = 3
local TYPING_DOTS_MARGIN = Screen.getHeightPart(0.0107)
local TYPING_DOT_RADIUS = Screen.getHeightPart(0.0035)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createBubblePointsLists(width, height, bubble, hasFlag)
  local groupHeight = bubble:getHeight()

  local centerL = FLAG_WIDTH + BUBBLE_RADIUS
  local centerR = width - BUBBLE_RADIUS
  local centerB = (groupHeight - height) + BUBBLE_RADIUS
  local centerT = groupHeight - BUBBLE_RADIUS

  local points = {}

    -- Top right rounded corner
  ShapeUtils.makeArcPoints(0, math.pi/2, BUBBLE_RADIUS, centerR, centerT, nil, points)

  -- Top left flag or rounded corner
  if (hasFlag) then
    -- Rounded tip of the flag stem
    local stemL = FLAG_STEM_RADIUS
    local stemT = groupHeight - FLAG_STEM_RADIUS
    ShapeUtils.makeArcPoints(math.pi/2, 7 * math.pi / 5, FLAG_STEM_RADIUS, stemL, stemT, nil, points)

    -- The curved underside of the flag that connects back to the message bubble
    -- This could have been done with another arc but this ended up being easier
    -- to polish and is more efficient to just hand position some points for roundness
    local flagStartY = groupHeight - FLAG_HEIGHT
    local nudgeAmount = Screen.getHeightPart(0.0009)
    table.insert(points, {x = FLAG_WIDTH - 5 * nudgeAmount, y = flagStartY + 14 * nudgeAmount})
    table.insert(points, {x = FLAG_WIDTH - 4 * nudgeAmount, y = flagStartY + 13 * nudgeAmount})
    table.insert(points, {x = FLAG_WIDTH - 3 * nudgeAmount, y = flagStartY + 11 * nudgeAmount})
    table.insert(points, {x = FLAG_WIDTH - 2 * nudgeAmount, y = flagStartY + 9 * nudgeAmount})
    table.insert(points, {x = FLAG_WIDTH - 1 * nudgeAmount, y = flagStartY + 5 * nudgeAmount})

    table.insert(points, {x = FLAG_WIDTH, y = flagStartY})
  else
    ShapeUtils.makeArcPoints(math.pi/2, math.pi, BUBBLE_RADIUS, centerL, centerT, nil, points)
  end

  -- Bottom left rounded corner
  ShapeUtils.makeArcPoints(-math.pi, -math.pi/2, BUBBLE_RADIUS, centerL, centerB, nil, points)

  -- Bottom right rounded corner
  ShapeUtils.makeArcPoints(-math.pi/2, 0, BUBBLE_RADIUS, centerR, centerB, nil, points)

  bubble._graphicsPrimitiveCoordinateLists = {ShapeUtils.flattenCoordPairsNew(points)}
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ChatMessageBubble:showTypingDots()
  self.typingDots:setVisibleRecursive(true)

  local typingAnimation = AHSpawn.new()

  local fadeDuration = 0.3
  local fadeEase = nil

  for _, graphicsPrimitive in ipairs(self.graphicsPrimitives) do
    typingAnimation:addAction(AHOpacityTo.new(graphicsPrimitive, fadeDuration, fadeEase, 1))
  end

  local jumpHeight = 4
  local jumpDuration = 0.28
  local jumpStagger = 0.12
  local typingAnimationDuration = 0.55

  for dotIndex, dot in ipairs(self.typingDots.children) do
    typingAnimation:addAction(
      AHLoop.new(
        AHSequence.new(
          AHDelay.new(jumpStagger * (dotIndex - 1)),
          AHLocationYBy.new(dot, jumpDuration/2, AHEaseSimpleInOut.new(), jumpHeight),
          AHLocationYBy.new(dot, jumpDuration/2, AHEaseSimpleInOut.new(), -jumpHeight),
          AHDelay.new(typingAnimationDuration - jumpDuration - jumpStagger * (dotIndex - 1))
        )
      )
    )
  end

  self.typingAnimation = typingAnimation:start()
end

function ChatMessageBubble:prepareForFadeIn()
  for _, graphicsPrimitive in ipairs(self.graphicsPrimitives) do
    graphicsPrimitive:setOpacity(0)
  end

  for _, element in ipairs(self.elementsToFade) do
    element:setOpacity(0)
  end
end

function ChatMessageBubble:fadeIn()
  local fadeDuration = 0.3
  local fadeEase = nil
  local fadeAction = AHSpawn.new()

  for _, element in ipairs(self.elementsToFade) do
    fadeAction:addAction(AHOpacityTo.new(element, fadeDuration, fadeEase, 1))
  end

  return fadeAction:start()
end

function ChatMessageBubble:hideTypingDots()
  if (self.typingAnimation ~= nil) then
    self.typingAnimation:stop()
  end

  self.typingDots:setVisibleRecursive(false)
end

function ChatMessageBubble:showMessage()
  local expandDuration = 0.4
  local expandEase = AHEaseExponentialInOut.new()

  local expandUpdateFunction = function(t)
    local width = MathUtils.interpolateLinearFromNormalized(t, self.initialWidth, self.finalWidth)
    local height = MathUtils.interpolateLinearFromNormalized(t, self.initialHeight, self.finalHeight)

    for _, graphicsPrimitive in ipairs(self.graphicsPrimitives) do
      createBubblePointsLists(width, height, graphicsPrimitive, self.hasFlag)
    end
  end

  local showDelayDuration = 0.2
  local showDuration = 0.17
  local scaleTextDuration = 0.1
  local scaleTextStartScale = 0.99
  local showEase = nil

  local showAction = AHSpawn.new(
    AHCustomAction.new(expandDuration, expandEase, expandUpdateFunction),
    AHSequence.new(
      function()
        self.textArea:setVisibleRecursive(true)
        self.textArea:setOpacity(0)
        self.textArea:setScl(scaleTextStartScale)
      end,
      AHDelay.new(showDelayDuration),
      AHSpawn.new(
        AHOpacityTo.new(self.textArea, showDuration, showEase, 1),
        AHScaleTo.new(self.textArea, scaleTextDuration, showEase, 1)
      )
    )
  )

  showAction:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function newCommon(text, width, icon, fontWeight, boldFontWeight, textColor, hasFlag)
  local iconWidth = icon:getEffectiveWidth()
  local textAreaWidth = width - (2 * PADDING_LEFT_RIGHT) - (2 * FLAG_WIDTH) - iconWidth

  local baseAttributes = {
    font = Font.new(fontWeight, FONT_SIZE),
    color = textColor
  }

  local attributesForLabels = {
    b = { font = Font.new(boldFontWeight, FONT_SIZE)},

    --offset needed because the medium italic font file oddly has a lower baseline y
    -- Note: We don't have access to any italic fonts at a weight any bolder than medium!
    i = { font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, FONT_SIZE), yOffset = 0.6}
  }

  local textArea = DynamicTextArea.new(text, textAreaWidth, baseAttributes, attributesForLabels)
  textArea:setVisibleRecursive(false)

  local typingDots = Row.new(TYPING_DOTS_MARGIN)
  for i = 1, TYPING_DOTS_COUNT do
    local dot = FilledCircle.new(TYPING_DOT_RADIUS, textColor)
    typingDots:insert(dot)
  end
  typingDots:alignChildren()
  typingDots:setVisibleRecursive(false)

  local height = PADDING_BOTTOM + textArea:getHeight() + PADDING_TOP

  local self = Group.new(width, height)
  Mixin.mix(self, ChatMessageBubble)

  self.textArea = textArea
  self.typingDots = typingDots

  self.hasFlag = hasFlag

  self.initialWidth = iconWidth * 2
  self.initialHeight = icon:getEffectiveHeight()
  self.finalWidth = FLAG_WIDTH + textArea:getTextWidth() + (2 * PADDING_LEFT_RIGHT)
  self.finalHeight = height

  return self
end

function ChatMessageBubble.newSystemMessage(text, width, hasFlag, showIcon)
  if (showIcon == nil) then
    showIcon = true
  end

  local icon = Sprite.newWithName("shared_brain_icon.png")
  icon:setScl(1.5)
  icon:setVisible(showIcon)

  local self = newCommon(text, width, icon, FontFaces.DIN_OT_MEDIUM, FontFaces.DIN_OT_BLACK, SYSTEM_TEXT_COLOR, hasFlag)
  local height = self:getHeight()

  local bubble = FilledGraphicsPrimitive.new(self.finalWidth, self.finalHeight, SYSTEM_BUBBLE_COLOR)
  createBubblePointsLists(self.initialWidth, self.initialHeight, bubble, hasFlag)
  local textAreaY = PADDING_BOTTOM + self.textArea:getHeightPart(0.5)
  bubble:insertAtLocation(self.textArea, FLAG_WIDTH + PADDING_BOTTOM + self.textArea:getWidthPart(0.5), textAreaY)
  bubble:insertAtLocation(self.typingDots, FLAG_WIDTH * 0.5 + self.initialWidth * 0.5, self.finalHeight - self.initialHeight * 0.5)

  local row = Row.new(0, Alignment.TOP)
  row:insert(icon)
  row:insert(bubble)
  row:alignChildren()

  row:setAnchorPointLeft()

  self:insertAtLocationFractional(row, 0, 0.5)

  self.graphicsPrimitives = {}
  table.insert(self.graphicsPrimitives, bubble)

  self.elementsToFade = {}
  table.insert(self.elementsToFade, icon)

  return self
end

function ChatMessageBubble.newUserMessage(text, width, hasFlag)
  local icon = Sprite.newWithName("shared_brain_icon.png")
  icon:setScl(1.5)

  local self = newCommon(text, width, icon, FontFaces.DIN_OT_REGULAR, FontFaces.DIN_OT_BOLD, USER_TEXT_COLOR, hasFlag)
  local height = self:getHeight()

  local dropShadow = FilledGraphicsPrimitive.new(self.finalWidth, self.finalHeight, USER_BUBBLE_SHADOW_COLOR)
  createBubblePointsLists(self.initialWidth, self.initialHeight, dropShadow, hasFlag)
  dropShadow:setRot(180)
  dropShadow:setAnchorPointLeft()

  local bubble = FilledGraphicsPrimitive.new(self.finalWidth, self.finalHeight, USER_BUBBLE_COLOR)
  createBubblePointsLists(self.initialWidth, self.initialHeight, bubble, hasFlag)
  bubble:setRot(180)
  bubble:setAnchorPointLeft()

  self:insertAtLocationFractional(dropShadow, 1, 0.5)
  dropShadow:offsetY(DROP_SHADOW_OFFSET_Y)
  self:insertAtLocationFractional(bubble, 1, 0.5)

  self.textArea:setAnchorPointLeft()
  local textAreaY = PADDING_BOTTOM + self.textArea:getHeightPart(0.5)
  self:insertAtLocation(self.textArea, width - self.finalWidth + PADDING_LEFT_RIGHT, textAreaY)

  self.graphicsPrimitives = {}
  table.insert(self.graphicsPrimitives, bubble)
  table.insert(self.graphicsPrimitives, dropShadow)

  self.elementsToFade = {}
  table.insert(self.elementsToFade, bubble)
  table.insert(self.elementsToFade, dropShadow)
  table.insert(self.elementsToFade, self.textArea)

  return self
end
