--- ChatMessageInputPanel

class("ChatMessageInputPanel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local INITIAL_HEIGHT = 1
local BACKGROUND_COLOR = ColorUtils.colorHex24ToStd(0xffffff)
local DROP_SHADOW_COLOR = ColorUtils.colorHex24ToStd(0xbbbbbb)
local TEXT_COLOR = ColorUtils.colorHex24ToStd(0x000000)
local BUTTON_RADIUS = 8
local MAX_WIDTH = Screen.getWidthPart(0.83)
local BUTTON_PADDING = 12
local BUTTON_MARGIN = 12
local DROP_SHADOW_OFFSET_Y = -2
local BUTTON_EXTRA_TOUCH_MARGIN = 5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function recalculateDimensions(self)
  local panelHeight = self.buttonColumn:getYTop() + BUTTON_MARGIN
  self.panel:setEffectiveHeight(panelHeight)
end

local function handleButtonPressed(self, id)
  if (self.pressedCallback ~= nil) then
    self.pressedCallback(id)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ChatMessageInputPanel:getPanelHeight()
  return self.panel:getEffectiveHeight()
end

function ChatMessageInputPanel:reset()
  self.buttonColumn:removeAll()
  self.buttonsById = {}
end

function ChatMessageInputPanel:setTouchEnabled(enabled)
  for _, button in pairs(self.buttonsById) do
    button:setTouchEnabled(enabled)
  end
end

function ChatMessageInputPanel:setPressedCallback(callback)
  self.pressedCallback = callback
end

function ChatMessageInputPanel:removeUserMessageById(id, shouldReadjustButtons)
  local buttonToRemove = self.buttonsById[id]

  assert(buttonToRemove ~= nil, string.format("Failed to remove user message. None found with id %s", id))

  if (buttonToRemove ~= nil) then
    buttonToRemove:setTouchEnabled(false)

    local duration = 0.08
    local scaleDestination = 0.75
    local scaleEase = AHEaseSineInOut.new()
    local fadeEase = nil
    local readjustmentDelay = 0.3
    local removeAction = AHSequence.new(
      AHSpawn.new(
        AHScaleTo.new(buttonToRemove, duration, scaleEase, scaleDestination),
        AHOpacityTo.new(buttonToRemove, duration, fadeEase, 0)
      ),
      AHRemove.new(buttonToRemove),
      AHDelay.new(readjustmentDelay),
      function()
        self.buttonsById[id] = nil

        if (shouldReadjustButtons == true) then
          self.buttonColumn:alignChildren()
          recalculateDimensions(self)
        end
      end
    )

    return removeAction:start()
  end
end

function ChatMessageInputPanel:addUserMessage(id, text)
  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_REGULAR, 14),
    color = TEXT_COLOR
  }

  local attributesForLabels = {
    b = {
      font = Font.new(FontFaces.DIN_OT_BOLD, 14)
    }
  }

  local textArea = DynamicTextArea.new(text, MAX_WIDTH, baseAttributes, attributesForLabels, Alignment.RIGHT)
  textArea:setAnchorPointRight()

  local buttonWidth = textArea:getTextWidth() + BUTTON_PADDING * 2
  local buttonHeight = textArea:getHeight() + BUTTON_PADDING * 2

  local buttonBacking = FilledRoundedRectangle.new(buttonWidth, buttonHeight, BUTTON_RADIUS, BACKGROUND_COLOR)
  buttonBacking:setAnchorPointTop()
  buttonBacking:insertAtLocation(textArea, buttonWidth - BUTTON_PADDING, buttonHeight / 2)

  local buttonDropShadow = FilledRoundedRectangle.new(buttonWidth, buttonHeight, BUTTON_RADIUS, DROP_SHADOW_COLOR)
  buttonDropShadow:setAnchorPointBottom()

  local buttonGroup = Group.new(buttonWidth, buttonHeight - DROP_SHADOW_OFFSET_Y)
  buttonGroup:insertAtBottom(buttonDropShadow)
  buttonGroup:insertAtTop(buttonBacking)

  local tapWidth = buttonWidth + BUTTON_EXTRA_TOUCH_MARGIN
  local tapHeight = buttonHeight + BUTTON_EXTRA_TOUCH_MARGIN
  Button.addVirtualButton(buttonGroup, tapWidth, tapHeight, function() handleButtonPressed(self, id) end)

  self.buttonColumn:insert(buttonGroup)
  self.buttonColumn:alignChildren()
  local buttonColumnX = MAX_WIDTH + (Screen.width - MAX_WIDTH) / 2
  self.buttonColumn:setLoc(buttonColumnX, Screen.safeBottom + BUTTON_MARGIN)

  self.buttonsById[id] = buttonGroup

  recalculateDimensions(self)
end

function ChatMessageInputPanel:prepareForFadeIn()
  local buttons = self.buttonColumn:getChildren()

  local isRecursive = true
  self:clearOpacityAttrLinkToChildren(isRecursive)

  for buttonIndex, button in ipairs(buttons) do
    button:linkOpacityAttrToChildren(isRecursive)
    button:setOpacity(0)
    button:setScl(0.95)
  end
end

function ChatMessageInputPanel:fadeIn()
  local buttons = self.buttonColumn:getChildren()

  local fadeInAction = AHSpawn.new()
  local staggerDuration = 0.2
  local duration = 0.1
  local ease = nil
  for buttonIndex, button in ipairs(buttons) do
    fadeInAction:addAction(
      AHDelayed.new((buttonIndex - 1) * staggerDuration,
        AHSpawn.new(AHOpacityTo.new(button, duration, ease, 1),
                    AHScaleTo.new(button, duration, ease, 1)
        )
      )
    )
  end

  return fadeInAction:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function ChatMessageInputPanel.new(color)
  local self = Group.newFullScreen()
  Mixin.mix(self, ChatMessageInputPanel)

  local panel = FilledRectangle.new(self:getWidth(), INITIAL_HEIGHT, color)
  panel:setAnchorPointBottom()
  self:insertAtBottom(panel)
  self.panel = panel

  local buttonColumn = Column.new(BUTTON_MARGIN, Alignment.RIGHT)
  buttonColumn:setAnchorPointBottomRight()
  local buttonColumnX = MAX_WIDTH + (Screen.width - MAX_WIDTH) / 2
  local minY = math.max(Screen.safeBottom, BUTTON_MARGIN)
  self:insertAtLocation(buttonColumn, buttonColumnX, minY)
  self.buttonColumn = buttonColumn

  self.pressedCallback = nil
  self.buttonsById = {}

  return self
end
