--- ChatMessageLogScroller

class("ChatMessageLogScroller")

require("widgets.TouchScroller")
require("widgets.rationales.ChatMessageBubble")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONTENTS_HEIGHT = Screen.height * 8
local BUBBLE_MARGIN = Screen.getHeightPart(0.0057)

local MESSAGE_TYPE = {
  System = "MESSAGE_TYPE.System",
  User = "MESSAGE_TYPE.User"
}

local DELAY_UNTIL_SYSTEM_FINISHES_TYPING = 0.75
local DELAY_UNTIL_SYSTEM_STARTS_TYPING_NEXT_MESSAGE = 0.75
local DELAY_UNTIL_SYSTEM_RESPONDS_AFTER_USER_MESSAGE = 1.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- TouchScroller's setScrollY gets called every frame through the update inertia
-- coroutine. In some cases, mainly when message bubbles are near the input panel
-- just as an input button is being removed, the sudden update in setVisibleHeight
-- causes setScrollY to jump many pixels. This method allows for temporarily
-- disabling the clamping conditions in setScrollY that cause sudden jumps.
function ChatMessageLogScroller:setAllowOverscroll(allow)
  self.allowOverscroll = allow
end

-- Overrides TouchScroller:setScrollY(scroll)
-- Prevents user from scrolling the contents too high and too low. Clamps to only
-- permit scrolling to portions of the content that have message bubbles.
function ChatMessageLogScroller:setScrollY(scroll)
  if (self.allowOverscroll == true) then
    self.scrollY = scroll
  else
    local maxScroll = math.max(0, self.contentsColumn:getHeight() - self.visibleHeight)
    self.scrollY = MathUtils.clamp(scroll, 0, maxScroll)
  end

  local posY = self:getHeight() + self.scrollY
  self.contents:setY(posY)
end

-- Needed for the overscroll protection. As the ChatMessageInputPanel panel's
-- effective height shrinks as buttons disappear, our scrollable dimensions need
-- to be able to change over time.
function ChatMessageLogScroller:setVisibleHeight(height)
  self.visibleHeight = height
end

function ChatMessageLogScroller:setProcessingMessageStartedCallback(callback)
  self.processingMessageStartedCallback = callback
end

function ChatMessageLogScroller:setProcessingMessageFinishedCallback(callback)
  self.processingMessageFinishedCallback = callback
end

function ChatMessageLogScroller:setIntroMessageFinishedCallback(callback)
  self.introMessageFinishedCallback = callback
end

function ChatMessageLogScroller:setMessageAddedCallback(callback)
  self.messageAddedCallback = callback
end

function ChatMessageLogScroller:pushSystemMessages(messages, isIntro)
  if (isIntro == nil) then
    isIntro = false
  end

  for messageIndex, message in ipairs(messages) do
    local hasIcon = (messageIndex == 1)
    local hasFlag = (messageIndex == 1)

    local newMessage = {
      type = MESSAGE_TYPE.System,
      hasIcon = hasIcon,
      hasFlag = hasFlag,
      text = message,
      isIntro = isIntro
    }

    table.insert(self.messageQueue, newMessage)
  end
end

function ChatMessageLogScroller:pushUserMessage(message)
  local newMessage = {
    type = MESSAGE_TYPE.User,
    text = message
  }

  table.insert(self.messageQueue, newMessage)
end

function ChatMessageLogScroller:startProcessingMessages()
  assert(self.messageProcessingCoroutine == nil, "messageProcessingCoroutine has already been started")

  self.contents:clearOpacityAttrLinkToChildren(true)

  self.messageProcessingCoroutine = runInCoroutine(function()
    local isProcessing = false
    local currentMessage = nil
    while true do
      if (TableUtils.getSize(self.messageQueue) > 0) then
        if (isProcessing == false) then
          if (self.processingMessageStartedCallback) then
            self.processingMessageStartedCallback()
          end

          isProcessing = true
        end

        currentMessage = table.remove(self.messageQueue, 1)

        if (currentMessage.type == MESSAGE_TYPE.System) then
          local bubble = ChatMessageBubble.newSystemMessage(currentMessage.text, self.contents:getWidth(), currentMessage.hasFlag, currentMessage.hasIcon)
          self.contentsColumn:insert(bubble)

          self.contentsColumn:alignChildren()

          if (self.messageAddedCallback ~= nil) then
            self.messageAddedCallback()
          end

          bubble:prepareForFadeIn()
          if (currentMessage.hasIcon) then
            bubble:fadeIn()

            if (currentMessage.isIntro) then
              waitFor(0.5)
            end
          end

          bubble:showTypingDots()
          waitFor(DELAY_UNTIL_SYSTEM_FINISHES_TYPING)
          bubble:showMessage()
          waitFor(0.18)
          bubble:hideTypingDots()
          waitFor(DELAY_UNTIL_SYSTEM_STARTS_TYPING_NEXT_MESSAGE)
        else
          local senderChangeMargin = self:getHeightPart(0.0196)

          self.contentsColumn:insertSpacer(senderChangeMargin)
          local hasFlag = true
          local bubble = ChatMessageBubble.newUserMessage(currentMessage.text, self.contents:getWidth(), hasFlag)
          self.contentsColumn:insert(bubble)

          self.contentsColumn:insertSpacer(senderChangeMargin)

          self.contentsColumn:alignChildren()

          if (self.messageAddedCallback ~= nil) then
            self.messageAddedCallback()
          end

          bubble:prepareForFadeIn()
          bubble:fadeIn()
          waitFor(0.3)
          bubble:showMessage()
          waitFor(DELAY_UNTIL_SYSTEM_RESPONDS_AFTER_USER_MESSAGE)
        end
      else
        if (isProcessing == true) then
          if (currentMessage.isIntro == true) then
            if (self.introMessageFinishedCallback) then
              self.introMessageFinishedCallback()
            end
          end

          if (self.processingMessageFinishedCallback) then
            self.processingMessageFinishedCallback()
          end
          isProcessing = false
        end
      end

      coroutine.yield()
    end
  end)
end

function ChatMessageLogScroller:stopProcessingMessages()
  if (self.messageProcessingCoroutine) then
    local isRecursive = true
    self.contents:linkOpacityAttrToChildren(isRecursive)

    self.messageProcessingCoroutine:stop()
  end
  self.messageProcessingCoroutine = nil
end

function ChatMessageLogScroller:reset()
  self.contentsColumn:removeAll()

  self.messageQueue = {}

  self:setScrollY(0)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function ChatMessageLogScroller.new(width, height)
  local contentsColumn = Column.new(BUBBLE_MARGIN)
  contentsColumn:setAnchorPointTop()

  local contents = Group.new(width, CONTENTS_HEIGHT)
  contents:insertAtTop(contentsColumn)

  local self = TouchScroller.new(Screen.width, height, contents)
  Mixin.mix(self, ChatMessageLogScroller)

  self:setInertialScrollingEnabled(true)

  self.contentsColumn = contentsColumn
  self.contents = contents
  self.messageQueue = {}
  self.messageProcessingCoroutine = nil
  self.processingMessageStartedCallback = nil
  self.processingMessageFinishedCallback = nil
  self.introMessageFinishedCallback = nil
  self.messageAddedCallback = nil
  self.visibleHeight = height
  self.allowOverscroll = false

  return self
end
