--- ChatRationale

class("ChatRationale")

require("widgets.rationales.ChatConversationUtils")
require("widgets.rationales.ChatMessageLogScroller")
require("widgets.rationales.ChatMessageInputPanel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local AUTOSCROLL_CUTOFF_Y = Screen.getHeightPart(0.4)
local AUTOSCROLL_DURATION = 0.7

ChatRationale.CONVERSATION_INTRO_KEY = "intro"
ChatRationale.CONVERSATION_OPTIONS_KEY = "options"
ChatRationale.CONVERSATION_OPTIONS_ID_KEY = "id"
ChatRationale.CONVERSATION_OPTIONS_USER_MESSAGE_KEY = "userMessage"
ChatRationale.CONVERSATION_OPTIONS_SYSTEM_MESSAGE_KEY = "systemMessage"

ChatRationale.CONVERSATION_CLOSE_CHAT_ID = "id_dismiss"
ChatRationale.CONVERSATION_CLOSE_CHAT_MESSAGE = "That's it, thanks!"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getUserMessageById(self, id)
  if (self.formattedConversationDescriptor ~= nil) then
    for _, option in ipairs(self.formattedConversationDescriptor[ChatRationale.CONVERSATION_OPTIONS_KEY]) do
      if (option[ChatRationale.CONVERSATION_OPTIONS_ID_KEY] == id) then
        return option[ChatRationale.CONVERSATION_OPTIONS_USER_MESSAGE_KEY]
      end
    end
  end
end

local function getSystemMessagesById(self, id)
  if (self.formattedConversationDescriptor ~= nil) then
    for _, option in ipairs(self.formattedConversationDescriptor[ChatRationale.CONVERSATION_OPTIONS_KEY]) do
      if (option[ChatRationale.CONVERSATION_OPTIONS_ID_KEY] == id) then
        return option[ChatRationale.CONVERSATION_OPTIONS_SYSTEM_MESSAGE_KEY]
      end
    end
  end
end

local function loadInputOptions(self)
  assert((self.formattedConversationDescriptor ~= nil), "setConversationDescriptor() must be called before loading input options")
  local options = self.formattedConversationDescriptor[ChatRationale.CONVERSATION_OPTIONS_KEY]
  for _, option in ipairs(options) do
    local id = option[ChatRationale.CONVERSATION_OPTIONS_ID_KEY]
    local message = option[ChatRationale.CONVERSATION_OPTIONS_USER_MESSAGE_KEY]
    self.inputPanel:addUserMessage(id, message)
  end

  self.inputPanel:addUserMessage(ChatRationale.CONVERSATION_CLOSE_CHAT_ID, ChatRationale.CONVERSATION_CLOSE_CHAT_MESSAGE)
end

local function initiateConversation(self)
  local introMessages = self.formattedConversationDescriptor[ChatRationale.CONVERSATION_INTRO_KEY]
  local isIntro = true
  self.chatMessageLog:pushSystemMessages(introMessages, isIntro)

  self.chatMessageLog:startProcessingMessages()
end

function endConversation(self)
  self.chatMessageLog:stopProcessingMessages()
end

local function setTouchSwallowingEnabled(self, enabled)
  if (enabled == true) then
    local handlerFunction = function(event)
      -- Does nothing, but is necessary to swallow touch events and prevent elements
      -- undearneath this from handling touches
    end
    TouchManager.addTouchHandler(self, handlerFunction, handlerBehaviors)
  else
    TouchManager.removeTouchHandler(self)
  end
end

local function onOpenStarted(self)
  self.chatMessageLog:setTouchEnabled(false)
  self.closeButton:setTouchEnabled(false)
  setTouchSwallowingEnabled(self, true)
end

local function onOpenEnded(self)
  self.closeButton:setTouchEnabled(true)
end

local function onCloseStarted(self)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  self.chatMessageLog:setTouchEnabled(false)
  self.closeButton:setTouchEnabled(false)

  endConversation(self)
end

local function onCloseEnded(self)
  self.chatMessageLog:reset()
  self.inputPanel:reset()
  setTouchSwallowingEnabled(self, false)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ChatRationale:openWithAction(action)
  local duration = 0.7
  local closeButtonDelay = 0.1
  local ease = AHEaseExponentialInOut.new()
  local fadeInAction = AHOpacityTo.new(self, duration / 2, ease, 1.0)
  local fadeInCloseButtonAction = AHDelayed.new(closeButtonDelay, AHOpacityTo.new(self.closeButton, duration - closeButtonDelay, ease, 1.0))
  local slideUpAction = AHLocationYTo.new(self, duration, ease, Screen.getHeightPart(0.5))
  local delayedConversationStart = AHDelayed.new(0.3, function() initiateConversation(self) end)

  local combinedAction = AHSpawn.new(action, fadeInAction, slideUpAction, fadeInCloseButtonAction, delayedConversationStart)

  local openSequence = AHSequence.new(
    function() onOpenStarted(self) end,
    combinedAction,
    function() onOpenEnded(self) end
  )

  openSequence:start()

  -- Necessary to force touch disabling to occur immediately and avoid bug where
  -- consecutive open actions can occur
  coroutine.yield()

  return openSequence
end

function ChatRationale:closeWithAction(action)
  local fadeDuration = 0.2
  local slideDuration = 0.5
  local ease = AHEaseSineInOut.new()
  local fadeOutAction = AHOpacityTo.new(self, fadeDuration, ease, 0)
  local fadeOutCloseButtonAction = AHOpacityTo.new(self.closeButton, fadeDuration, ease, 0)
  local slideDownAction = AHLocationYTo.new(self, slideDuration, ease, 0)

  local combinedAction = AHSpawn.new(action, fadeOutAction, slideDownAction, fadeOutCloseButtonAction)

  local closeSequence = AHSequence.new(
    function() onCloseStarted(self) end,
    combinedAction,
    function() onCloseEnded(self) end
  )

  closeSequence:start()

  -- Necessary to force touch disabling to occur immediately and avoid bug where
  -- consecutive close actions can occur
  coroutine.yield()

  return closeSequence
end

function ChatRationale:setConversationDescriptor(conversationDescriptor)
  ChatConversationUtils.validateConversationDescriptor(conversationDescriptor)
  self.formattedConversationDescriptor = ChatConversationUtils.createFormattedConversationDescriptor(conversationDescriptor)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, color)
  local closeIcon = Sprite.newWithName("close_icon.png")
  local foldMargin = closeIcon:getHeight()

  local foldHeight = math.max(foldMargin * 3, Host.safeAreaTopInset + closeIcon:getHeight() + (2 * foldMargin))
  local fold = FilledRectangle.new(self:getWidth(), foldHeight, color)
  fold:setAnchorPointTop()
  self:insertAtTop(fold)
  self.fold = fold

  local closeIconX = math.max(foldMargin + closeIcon:getWidthPart(0.5), Host.safeAreaLeftInset)
  local closeIconY = Screen.safeTop - foldMargin - closeIcon:getHeightPart(0.5)
  fold:insertAtLocation(closeIcon, closeIconX, closeIcon:getHeightPart(0.5) + foldMargin)

  local virtualButtonSize = closeIcon:getWidth() * 3
  Button.addVirtualButton(closeIcon, virtualButtonSize, virtualButtonSize, function()
    if (self.closedCallback ~= nil) then
      self.closedCallback()
    end

    ChatRationaleAnalytics.reportDismissed()
    Sound.playSoundEffect("shared_rationales_end")
  end)
  self.closeButton = closeIcon

  local chatMessageLogMargin = closeIcon:getWidthPart(0.75)
  local chatMessageLogWidth = math.min(self:getWidth() - (2 * chatMessageLogMargin), Screen.safeWidth)
  local chatMessageLogHeight = fold:getYBottom() - (2 * chatMessageLogMargin)
  local chatMessageLog = ChatMessageLogScroller.new(chatMessageLogWidth, chatMessageLogHeight)
  chatMessageLog:setTouchEnabled(false)
  self:insertAtLocation(chatMessageLog, self:getWidthPart(0.5), chatMessageLogMargin + (chatMessageLogHeight / 2))
  self.chatMessageLog = chatMessageLog

  local inputPanel = ChatMessageInputPanel.new(color)

  chatMessageLog:setProcessingMessageStartedCallback(function()
    inputPanel:setTouchEnabled(false)
    chatMessageLog:setTouchEnabled(false)
  end)

  chatMessageLog:setIntroMessageFinishedCallback(function()
    loadInputOptions(self)
    inputPanel:prepareForFadeIn()
    inputPanel:fadeIn()
  end)

  chatMessageLog:setProcessingMessageFinishedCallback(function()
    inputPanel:setTouchEnabled(true)
    chatMessageLog:setTouchEnabled(true)
    chatMessageLog:setAllowOverscroll(false)
  end)

  chatMessageLog:setMessageAddedCallback(function()
    local bottomYContentsSpace = chatMessageLog.contentsColumn:getYBottom()
    local _, bottomYScrollerSpace, _ = chatMessageLog:transformLoc(chatMessageLog.contents, 0, bottomYContentsSpace, 0)

    local delta = bottomYScrollerSpace - AUTOSCROLL_CUTOFF_Y
    if (delta < 0) then
      local ease = AHEaseSineInOut.new()
      chatMessageLog:newScrollYBy(AUTOSCROLL_DURATION, ease, -delta):start()
    end
  end)

  inputPanel:setPressedCallback(function(id)
    ChatRationaleAnalytics.reportOptionTapped(id)

    if (id == ChatRationale.CONVERSATION_CLOSE_CHAT_ID) then
      Sound.playSoundEffect("shared_rationales_end")

      local shouldReadjustButtons = false
      inputPanel:removeUserMessageById(id, shouldReadjustButtons)

      if (self.closedCallback ~= nil) then
        self.closedCallback()
      end
    else
      Sound.playSoundEffect("shared_rationales_question_button")

      local shouldReadjustButtons = true
      inputPanel:removeUserMessageById(id, shouldReadjustButtons)

      chatMessageLog:setAllowOverscroll(true)

      local userMessage = getUserMessageById(self, id)
      local systemMessages = getSystemMessagesById(self, id)

      chatMessageLog:pushUserMessage(userMessage)
      chatMessageLog:pushSystemMessages(systemMessages)
    end
  end)

  self:insertAtCenter(inputPanel)
  self.inputPanel = inputPanel

  fold:bringToFront()
  closeIcon:bringToFront()
  closeIcon:setTouchEnabled(false)

  -- If message bubbles get below this screen location, autoscroll the chatMessageLog
  -- to keep messages above it
  chatMessageLog:setVisibleHeight(self.chatMessageLog:getHeight() - AUTOSCROLL_CUTOFF_Y)

  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  -- We want to fade this element independently so we may fine tune its fade animation
  closeIcon:clearAttrLink(MOAIColor.ATTR_OPACITY, self)
  self.closeButton:setOpacity(0)

  self:setOpacity(0)
end

function ChatRationale.new(color, closedCallback)
  Asserts.assertType(closedCallback, "closedCallback", "function")

  local self = FilledRectangle.newFullScreen(color)
  Mixin.mix(self, ChatRationale)

  self.closedCallback = closedCallback

  createLayout(self, color)

  return self
end
