--- ArcticGame
-- @author Created by joepoka on October 19, 2016.

class("ArcticGame")

require("StandardUIBuilder")
require("ArcticConfiguration")
require("ArcticRoundGenerator")
require("ArcticScene")
require("bonuses.ResponseSpeedTracker")
require("ExperimentsConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {0.2, 0.2, 0.2}
local UI_COMPONENTS_COLOR = {1, 1, 1}

local function configureGameDifficulty(self)
  local rankTwoScore = ArcticConfiguration.getRankTwoScore()
  local rankThreeScore = ArcticConfiguration.getRankThreeScore()
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = ArcticConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  local roundGenerator = ArcticRoundGenerator.new()
  roundGenerator:setDifficultyNameChooser(difficultyNameChooser)
  roundGenerator:preloadConcepts()
  self.roundGenerator = roundGenerator

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function reset(self)
  self.submittedAnswer = nil
  self.userHasSubmittedAnswer = false
  self.userAnsweredCorrectly = false
  self.userEnteredSum = false
end

local function pauseTimer(self)
  self.timer:pauseAndShowIcon()
end

local function resumeTimer(self)
  self.timer:resumeAndHideIcon()
end

local function isOutOfTime(self)
  return (self.timer:getTime() >= ArcticConfiguration.GAME_DURATION)
end

local function showFirstPlayOverlay(self)
  local isFirstPlay = GameSession.hasNeverWon()
  if (isFirstPlay) then
    local basicInformationAnimation = self.scene:getBasicInformationOverlayAnimation()
    wait(basicInformationAnimation:start())
  end
end

local function showGameIntro(self)
  local gameIntroAnimation = self.scene:getGameIntroAnimation()
  Sound.playSoundEffect("game_start")
  gameIntroAnimation:start()
end

local function showSumHelperIfNeeded(self)
  local hasUserAnsweredCorrectlyBefore = (GameSession.getScore() > 0)
  local willHandleAccidentalSum = (self.userEnteredSum and (not hasUserAnsweredCorrectlyBefore))
  if (willHandleAccidentalSum) then
    pauseTimer(self)
    local accidentalSumAnimation = self.scene:getAccidentalSumOverlayAnimation()

    wait(accidentalSumAnimation:start())
  end
  return willHandleAccidentalSum
end

local function resetScene(self, isFirstPresentation)
  if (not isFirstPresentation) then
    Sound.playSoundEffect("begin_round")
  end

  local resetAnimation = self.scene:getResetAnimation(self.round)
  wait(resetAnimation:start())
end

local function presentConcept(self, startAcceptingInputCallback)
  local presentationAnimation = self.scene:getRoundPresentationAnimation(startAcceptingInputCallback)
  presentationAnimation:start()

  GameSession.contentTracker:notifyPresentedConcept(self.round.conceptId, self.round.difficulty)
end

local function waitForUserInput(self)
  local readyToProcessAnswer = function()
    return self.userHasSubmittedAnswer
  end

  local isValidInput = false
  repeat
    waitUntil(readyToProcessAnswer)

    local didHandleAccidentalSum = showSumHelperIfNeeded(self)
    if (didHandleAccidentalSum) then
      reset(self)
      resumeTimer(self)
    else
      isValidInput = true
    end
  until isValidInput

  self.scene:stopAcceptingUserInput()
end

local function showResultAnimation(self)
  local actions = {}

  local sceneCleanUpAnimation = self.scene:getRoundResultAnimation(self.userAnsweredCorrectly)
  table.insert(actions, sceneCleanUpAnimation)

  if (not self.userAnsweredCorrectly) then
    local tipSequence = self.scene:getTipSequenceAnimation(self.round, self.submittedAnswer)
    table.insert(actions, tipSequence)
  end

  local resultAnimation = AHSpawn.new(actions)
  wait(resultAnimation:start())

  -- Edge case handling of dismissing a rationale with an empty game timer.
  -- Necessary to trigger the timer onCompletion callback
  if ((not self.userAnsweredCorrectly) and isOutOfTime(self)) then
    resumeTimer(self)
  end
end

local function userSubmittedAnswer(self, answer)
  self.userHasSubmittedAnswer = true
  local values = self.round.values
  local sum = TableUtils.sum(values)

  self.submittedAnswer = answer
  self.userEnteredSum = (answer == sum)
  self.userAnsweredCorrectly = (answer == self.round.answer)
end

local function increaseScore(self)
  GameSession.increaseScore(ArcticConfiguration.getPointsForCorrectAnswer(self.round.difficulty))
  self.scoreLabel:setScore(GameSession.getScore())
end

local function handleUserInput(self)
  if (self.userAnsweredCorrectly) then
    GameSession.contentTracker:notifyCorrectAnswer(self.round.conceptId)
    GameSession.incrementCorrectAnswerCount()
    self.responseSpeedTracker:notifyCorrectResponse()
    increaseScore(self)
    Sound.playSoundEffect("correct_answer")
  else
    GameSession.contentTracker:notifyIncorrectAnswer(self.round.conceptId, nil, tostring(self.submittedAnswer))
    GameSession.incrementIncorrectAnswerCount()
    self.timer:addTime(ArcticConfiguration.INCORRECT_RESPONSE_TIME_PENALTY)
    pauseTimer(self)
    self.responseSpeedTracker:notifyIncorrectResponse()
    Sound.playSoundEffect("incorrect_answer")
  end
end

local function gameOver(self)
  local clearSceneAnimation = self.scene:getClearSceneAnimation()

  if (GameSession:getScore() > 0) then
    self.scoreLabel:setScore(GameSession:getScore())

    Sound.playSoundEffect("outro")

    local outroAnimation, duration = self.scene:getOutroAnimation()
    local delayedClearScene = AHDelayed.new(duration * 0.075, clearSceneAnimation)
    local successOutro = AHSpawn.new(outroAnimation, delayedClearScene)
    wait(successOutro:start())

    GameSession.notifyGameWin()
  else
    wait(clearSceneAnimation:start())
    GameSession.notifyGameLoss()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ArcticGame:gameFlow()
  self.timer:initialize()

  showFirstPlayOverlay(self)
  showGameIntro(self)

  local startAcceptingInputCallback = function()
    resumeTimer(self)
    self.responseSpeedTracker:notifyStart()
  end

  local isFirstPresentation = true
  while (not isOutOfTime(self)) do
    self.round = self.roundGenerator:next()
    resetScene(self, isFirstPresentation)
    presentConcept(self, startAcceptingInputCallback)
    waitForUserInput(self)
    handleUserInput(self)
    showResultAnimation(self)
    reset(self)

    isFirstPresentation = false
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)

  local quickResponseWrapper = function(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 ArcticConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(ArcticConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       ArcticConfiguration.SPEED_BONUS_DECAY_START)

  local userSubmittedAnswerWrapper = function(answer)
    userSubmittedAnswer(self, answer)
  end

  self.scene = ArcticScene.new(userSubmittedAnswerWrapper)
  self:insertAtCenter(self.scene)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local onCompletion = function()
    gameOver(self)
  end

  self.timer = builder:addDigitTimer(ArcticConfiguration.GAME_DURATION, onCompletion)
end

-- Constructor ------------------------

function ArcticGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ArcticGame)

  ArcticCommon.initialize()

  FontScaling.setAutoScaling(Screen.width, Screen.height)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
