--- ArcticInputGroup
-- @author Created by Joe Poka on November 16, 2016.

class("ArcticInputGroup")

require("ArcticAnimatedLine")
require("ArcticHollowTriangle")
require("ArcticNumberPadInputField")
require("ArcticScrollingParticles")
require("ArcticShake")
require("ArcticShinyRectanglesInputGroup")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  scrollingParticles = 5,
  shinyRectangles = 10,
  outline = 20,
  additionalTriangles = 30,
  hollowTriangle = 40,
  filledBackground = 50,
  inputField = 60,
  submitLabel = 70,
  rectangleGradient = 80
}

local TRIANGLE_COLORS = {
  default = {1.0, 1.0, 1.0, 1.0},
  frontTriangle = {1.0, 0.9, 0.64, 1.0},
  middleTriangle = {1.0, 0.85, 0.8, 1.0},
  rearTriangle = {0.99, 0.61, 0.62, 1.0},
  muddyYellow = {0.9, 0.9, 0.85, 0.25}
}

ArcticInputGroup.InputState = {
  empty = "empty",
  partialAnswer = "partialAnswer"
}

local BACKGROUND_TRIANGLE_OPACITY = 0.15
local HOLLOW_TRIANGLE_OPACITY = 0.1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function onTouchDown(self, submitAnswerCallback)
  local isEmpty = self:fieldIsEmpty()
  if (isEmpty) then
    return
  end

  local value = self.inputField:getValue()
  local isJustDecimal = (value == ".")
  if (isJustDecimal) then
    return
  end

  self.shouldShowTapToSubmit = false
  local answer = tonumber(value)
  submitAnswerCallback(answer)
end

local function showDelayedTapToSubmit(self)
  local delayedShow = AHSequence.new(
    AHDelay.new(5),
    AHOpacityTo.new(self.submitLabel, 1.0, nil, 0.0),
    AHOpacityTo.new(self.tapToSubmitLabel, 1.0, nil, 1.0),
    function() self.shouldShowTapToSubmit = false end,
    AHLoop.new(
      AHSequence.new(
        AHOpacityTo.new(self.tapToSubmitLabel, 0.5, nil, 0.25),
        AHOpacityTo.new(self.tapToSubmitLabel, 0.5, nil, 1.0)
      )
    )
  )
  delayedShow:start()
  self.delayedTapToSubmitReveal = delayedShow
end

local function getStopAndHideLabelsAction(self, duration)
  self.delayedTapToSubmitReveal:stop()
  return AHSpawn.new(
    AHOpacityTo.new(self.submitLabel, duration, nil, 0.0),
    AHOpacityTo.new(self.tapToSubmitLabel, duration, nil, 0.0)
  )
end

local function showInputStatePartialAnswer(self)
  local duration = 0.2
  local fadeInBackground = AHOpacityTo.new(self.background, duration, nil, BACKGROUND_TRIANGLE_OPACITY)
  local fadeInLabel = AHOpacityTo.new(self.submitLabel, duration, nil, 1.0)

  local fadeInAdditionalBackground = AHSpawn.new()
  for _, triangle in ipairs(self.additionalTriangleBackgrounds) do
    local fadeIn = AHOpacityTo.new(triangle, duration, nil, BACKGROUND_TRIANGLE_OPACITY)
    fadeInAdditionalBackground:addAction(fadeIn)
  end

  local updateBorderColor = AHColorTo.new(self.mainHollowTriangle, duration, nil, TRIANGLE_COLORS.default)
  local updateBorderColors = AHSpawn.new(updateBorderColor)
  for _, triangle in ipairs(self.additionalTriangles) do
    local updateColor = AHColorTo.new(triangle, duration, nil, TRIANGLE_COLORS.default)
    updateBorderColors:addAction(updateColor)
  end

  updateBorderColors:start()
  fadeInAdditionalBackground:start()
  fadeInBackground:start()
  fadeInLabel:start()

  if (self.shouldShowTapToSubmit) then
    showDelayedTapToSubmit(self)
  end
end

local function cleanUpInputStatePartialAnswer(self)
  local duration = 0.15
  local fadeOutBackground = AHOpacityTo.new(self.background, duration, nil, 0.0)
  local fadeOutLabel = getStopAndHideLabelsAction(self, duration)

  local fadeOutAdditionalBackground = AHSpawn.new()
  for _, triangle in ipairs(self.additionalTriangleBackgrounds) do
    local fadeOut = AHOpacityTo.new(triangle, duration, nil, 0.0)
    fadeOutAdditionalBackground:addAction(fadeOut)
  end

  local updateBorderColors = function()
    self.mainHollowTriangle:setColor(unpack(TRIANGLE_COLORS.muddyYellow))
    for i = 1, #self.additionalTriangles do
      local triangle = self.additionalTriangles[i]
      local color = nil
      if (i == 1) then
        color = TRIANGLE_COLORS.middleTriangle
      elseif (i == 2) then
        color = TRIANGLE_COLORS.rearTriangle
      end
      color[4] = HOLLOW_TRIANGLE_OPACITY
      triangle:setColor(unpack(color))
    end
  end

  updateBorderColors()
  fadeOutAdditionalBackground:start()
  fadeOutBackground:start()
  fadeOutLabel:start()
end

local function getOutlineAnimation(self, isFirstRound)
  local delay = 0.0
  local actions = {}

  local deltaXAdjust = 1.35
  local deltaDelay = 0.4
  if (isFirstRound) then
    deltaXAdjust = 1.0
    deltaDelay = 0.55
  end

  for i = 1, #self.outlines do
    local line = self.outlines[i]
    local show = function()
      line:adjustDeltaX(deltaXAdjust)
      line:show()
      local appearAnimation = line:getAppearAnimation()
      appearAnimation:start()

      local showRectangles = nil
      local lineHasRightRectangles = (i == 1)
      local lineHasLeftRectangles = (i == 3)

      if (lineHasRightRectangles) then
        showRectangles = self.shinyRectanglesGroup:getShowRightRectangles(isFirstRound)
      elseif (lineHasLeftRectangles) then
        showRectangles = self.shinyRectanglesGroup:getShowLeftRectangles(isFirstRound)
      end

      if (showRectangles ~= nil) then
        local rectangleShowDelay = 0.1
        local delayedShowRectangles = AHDelayed.new(rectangleShowDelay, showRectangles)
        delayedShowRectangles:start()
      end
    end

    local delayedShowLine = AHDelayed.new(delay, show)
    table.insert(actions, delayedShowLine)
    delay = delay + deltaDelay
  end

  local outlineAnimation = AHSpawn.new(actions)
  return outlineAnimation
end

local function setupOutlineAnimation(self)
  local deltaX = 0.03
  local lineWidth = Screen.getHeightPart(0.002)
  self.outlines = {}

  local lineParameters = {
    {
      length = self.triangleSideLength,
      lineWidth = lineWidth,
      angleDegrees = -60.0,
      deltaX = deltaX,
      x = self.background:getWidthPart(0.5),
      y = self.triangleHeight
    },
    {
      length = self.triangleSideLength * 0.99,
      lineWidth = lineWidth,
      angleDegrees = 0.0,
      deltaX = -deltaX,
      x = self.background:getWidth(),
      y = 0.0
    },
    {
      length = self.triangleSideLength,
      lineWidth = lineWidth,
      angleDegrees = 60.0,
      deltaX = deltaX,
      x = 0.0,
      y = 0.0
    }
  }

  for _, parameters in ipairs(lineParameters) do
    local animatedLine = ArcticAnimatedLine.new(parameters)
    animatedLine:setLoc(parameters.x, parameters.y)
    self:insert(animatedLine, Z_INDEX.outline)
    table.insert(self.outlines, animatedLine)
  end
end

local function resetOutlines(self)
  for _, line in ipairs(self.outlines) do
    line:reset()
  end
end

local function getShowTrianglesAnimation(self, isFirstRound)
  local duration = 0.25
  if (isFirstRound) then
    duration = 0.4
  end

  local fadeActions = {}
  local fadeIn = AHOpacityTo.new(self.mainHollowTriangle, duration, nil, TRIANGLE_COLORS.muddyYellow[4])
  table.insert(fadeActions, fadeIn)

  for _, triangle in ipairs(self.additionalTriangles) do
    triangle:setVisible(true)
    local fade = AHOpacityTo.new(triangle, duration, nil, HOLLOW_TRIANGLE_OPACITY)
    table.insert(fadeActions, fade)
  end

  local actions = {}
  local delay = 0.0
  local deltaDelay = 0.3

  for _, action in ipairs(fadeActions) do
    local delayedAction = AHDelayed.new(delay, action)
    table.insert(actions, delayedAction)
    delay = delay + deltaDelay
  end

  local animation = AHSpawn.new(actions)
  return animation
end

local function toggleInputState(self, inputState)
  if (self.currentInputState == inputState) then
    return
  end

  self.currentInputState = inputState

  local blinkQuickly = (self.currentInputState == ArcticInputGroup.InputState.partialAnswer)
  self.shinyRectanglesGroup:updateTwinkleAnimation(blinkQuickly)

  if (self.currentInputState == ArcticInputGroup.InputState.empty) then
    cleanUpInputStatePartialAnswer(self)
  elseif (self.currentInputState == ArcticInputGroup.InputState.partialAnswer) then
    showInputStatePartialAnswer(self)
  end
end

local function inputFieldWillBeClearAfterBackspace(self)
  local inputValue = self.inputField:getValue()
  local willBeClearAfterBackspace = (#inputValue <= 1)
  return willBeClearAfterBackspace
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ArcticInputGroup:getDefaultLocation()
  local x = Screen.getWidthPart(0.5)
  local y = Screen.getSafeY(0.65)
  return x, y
end

function ArcticInputGroup:toggleInputStateForNumberPadKeyPress(keyString)
  -- don't toggle input state on submit (it's toggled elsewhere)
  local userTappedSubmit = (keyString == NumberPad.SpecialKeys.SUBMIT)
  if (userTappedSubmit) then
    return
  end

  -- don't toggle input state when user taps zero and input field is empty
  local userTappedZero = (keyString == "0")
  if (userTappedZero) then
    local inputIsEmpty = self:fieldIsEmpty()
    if (inputIsEmpty) then
      return
    end
  end

  -- toggle input state between empty and partialAnswer
  local inputState = ArcticInputGroup.InputState.partialAnswer
  local userTappedBackspace = (keyString == NumberPad.SpecialKeys.BACKSPACE)
  if (userTappedBackspace) then
    local willBeClear = inputFieldWillBeClearAfterBackspace(self)
    if (willBeClear) then
      inputState = ArcticInputGroup.InputState.empty
    end
  end

  local showExtraParticles = (inputState == ArcticInputGroup.InputState.partialAnswer)
  self.shinyRectanglesGroup:updateExtraParticles(showExtraParticles)

  toggleInputState(self, inputState)
end

function ArcticInputGroup:partialReset()
  if (self.delayedTapToSubmitReveal) then
    self.delayedTapToSubmitReveal:stop()
    self.tapToSubmitLabel:setOpacity(0.0)
  end
  self.shinyRectanglesGroup:partialReset()
  self.inputField:clear()
  self.background:setOpacity(0.0)
  self.submitLabel:setOpacity(0.0)
  toggleInputState(self, ArcticInputGroup.InputState.empty)
end

function ArcticInputGroup:reset()
  self:partialReset()

  resetOutlines(self)

  for _, triangle in ipairs(self.additionalTriangles) do
    triangle:setOpacity(0.0)
    triangle:setVisible(false)
  end

  for _, triangle in ipairs(self.additionalTriangleBackgrounds) do
    triangle:setOpacity(0.0)
  end

  self.shinyRectanglesGroup:reset()
  self.mainHollowTriangle:setOpacity(0.0)
  self.mainHollowTriangle:setScl(1.0, 1.0)
  self.background:setScl(1.0, 1.0)
end

function ArcticInputGroup:fieldIsEmpty()
  local inputValue = self.inputField:getValue()
  local isEmpty = (inputValue == "")
  return isEmpty
end

function ArcticInputGroup:getInputField()
  return self.inputField
end

function ArcticInputGroup:clearInputField()
  self.inputField:clear()
end

function ArcticInputGroup:getConceptPresentationAnimation(isFirstRound)
  local actions = {}

  local showOutlines = getOutlineAnimation(self, isFirstRound)
  table.insert(actions, showOutlines)

  local showTriangles = getShowTrianglesAnimation(self, isFirstRound)

  local delay = 0.85
  if (isFirstRound) then
    delay = 1.3
  end

  local delayedShowTriangles = AHDelayed.new(delay, showTriangles)
  table.insert(actions, delayedShowTriangles)

  local hideOutlines = function()
    resetOutlines(self)
  end

  delay = delay * 2.0
  local delayedHideOutlines = AHDelayed.new(delay, hideOutlines)
  table.insert(actions, delayedHideOutlines)

  local showScrollingParticles = function()
    self.scrollingParticles:reset()
    self.scrollingParticles:show()
  end

  delay = delay * 0.35
  local delayedShowScrollingParticles = AHDelayed.new(delay, showScrollingParticles)
  table.insert(actions, delayedShowScrollingParticles)

  local presentationAnimation = AHSpawn.new(actions)
  return presentationAnimation
end

function ArcticInputGroup:getCorrectAnswerAnimation()
  local actions = {}
  local totalDuration = 0.5

  local animateOutShinyRectangles = self.shinyRectanglesGroup:getCorrectAnswerAnimation()
  table.insert(actions, animateOutShinyRectangles)

  local scaleFadeDuration = totalDuration
  local scaleUpTriangle = AHScaleTo.new(self.mainHollowTriangle, scaleFadeDuration, AHEaseExponentialOut.new(), 1.5)
  local fadeOutTriangle = AHOpacityTo.new(self.mainHollowTriangle, scaleFadeDuration * 1.5, nil, 0.0)
  table.insert(actions, scaleUpTriangle)
  table.insert(actions, fadeOutTriangle)

  local shorterScaleFadeDuration = scaleFadeDuration * 0.7
  for _, triangle in ipairs(self.additionalTriangles) do
    local fadeOut = AHOpacityTo.new(triangle, shorterScaleFadeDuration, nil, 0.0)
    table.insert(actions, fadeOut)
  end

  for _, triangle in ipairs(self.additionalTriangleBackgrounds) do
    local fadeOut = AHOpacityTo.new(triangle, shorterScaleFadeDuration, nil, 0.0)
    table.insert(actions, fadeOut)
  end

  local scaleUpBackground = AHScaleTo.new(self.background, shorterScaleFadeDuration, AHEaseExponentialOut.new(), 1.05)
  local fadeOutBackground = AHOpacityTo.new(self.background, shorterScaleFadeDuration, nil, 0.0)
  table.insert(actions, scaleUpBackground)
  table.insert(actions, fadeOutBackground)

  local fadeOutLabel = getStopAndHideLabelsAction(self, shorterScaleFadeDuration)
  table.insert(actions, fadeOutLabel)

  local hideScrollingParticles = function()
    self.scrollingParticles:hide()
  end
  table.insert(actions, hideScrollingParticles)

  local animation = AHSpawn.new(actions)
  return animation
end

function ArcticInputGroup:getIncorrectAnswerAnimation(duration)
  local resetInputGroup = function()
    self:reset()
    self.scrollingParticles:hide()
  end
  return resetInputGroup
end

function ArcticInputGroup:getAccidentalSumAnimation(duration)
  local actions = {}

  local fadeSubmitLabel = getStopAndHideLabelsAction(self, duration)
  table.insert(actions, fadeSubmitLabel)

  local trianglesToShake = {
    self.mainHollowTriangle,
    self.background,
    self.shinyRectanglesGroup
  }

  for _, triangle in ipairs(self.additionalTriangles) do
    table.insert(trianglesToShake, triangle)
  end

  for _, triangle in ipairs(self.additionalTriangleBackgrounds) do
    table.insert(trianglesToShake, triangle)
  end

  local parameters = {
    minimumDelta = Screen.getWidthPart(0.0004),
    maximumDelta = Screen.getWidthPart(0.04),
    shakeCount = 8,
    useRandomSigns = true
  }

  for _, triangle in ipairs(trianglesToShake) do
    local shake = ArcticShake.getShakeAnimation(triangle, parameters)
    table.insert(actions, shake)
  end

  local animation = AHSpawn.new(actions)
  return animation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, parameters)
  local inputHeight = parameters.inputHeight
  local fontFace = parameters.fontFace
  local fontSize = parameters.fontSize
  local defaultColor = {0.71, 0.29, 0.3, 1.0}
  self.inputField = ArcticNumberPadInputField.new(inputHeight, fontFace, fontSize, defaultColor)
  local x = self:getWidthPart(0.5)
  local y = self:getHeightPart(0.485)
  self.inputField:setLoc(x, y)
  self:insert(self.inputField, Z_INDEX.inputField)

  local color = TRIANGLE_COLORS.frontTriangle
  self.background = FilledTriangle.new(0.0, 0.0, self.triangleSideLength, 0.0, self.triangleSideLength * 0.5, self.triangleHeight, color)
  self.background:setOpacity(0.0)
  self.background:setBlendMode(MOAIProp2D.BLEND_ADD)
  self:insertAtCenter(self.background, Z_INDEX.filledBackground)

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 18.0)
  self.submitLabel = Label.new(Localizer.get("SUBMIT"), font)
  self.submitLabel:setColor(unpack({0.85, 0.59, 0.59, 1.0}))
  self.submitLabel:setOpacity(0.0)
  y = self:getHeightPart(0.24)
  self.submitLabel:setLoc(x, y)
  self:insert(self.submitLabel, Z_INDEX.submitLabel)

  local tapToSubmitFont = Font.new(FontFaces.DIN_OT_MEDIUM, 16.0)
  self.tapToSubmitLabel = DynamicTextArea.newSimple(Localizer.get("TAP_HERE"), Screen.getWidthPart(0.33), tapToSubmitFont, Alignment.CENTER, Alignment.CENTER)
  self.tapToSubmitLabel:setColor(unpack({0.85, 0.59, 0.59, 1.0}))
  self.tapToSubmitLabel:setOpacity(0.0)
  self.tapToSubmitLabel:setLoc(x, y)
  self:insert(self.tapToSubmitLabel, Z_INDEX.submitLabel)

  self.shinyRectanglesGroup = ArcticShinyRectanglesInputGroup.new(self.triangleHeight)
  self:insertAtCenter(self.shinyRectanglesGroup, Z_INDEX.shinyRectangles)

  self.scrollingParticles = ArcticScrollingParticles.new()
  self:insertAtCenter(self.scrollingParticles, Z_INDEX.scrollingParticles)

  local sideLength = self.triangleSideLength * 1.02
  local triangleHeight = self.triangleHeight * 1.02
  local triangleColor = TRIANGLE_COLORS.muddyYellow
  local lineWidth = self.triangleSideLength * 0.008
  local blurMargin = lineWidth * 0.1
  self.mainHollowTriangle = HollowTriangle.new(0.0, 0.0, sideLength, 0.0, sideLength * 0.5, triangleHeight, triangleColor, lineWidth, blurMargin)
  self.mainHollowTriangle:setOpacity(0.0)
  self:insertAtCenter(self.mainHollowTriangle, Z_INDEX.hollowTriangle)
  self.mainHollowTriangle:setBlendMode(MOAIProp2D.BLEND_ADD)

  self.additionalTriangles = {}
  self.additionalTriangleBackgrounds = {}
  local additionalTriangleCount = 2
  local minimumOffset = self:getWidthPart(0.06)
  local maximumOffset = self:getWidthPart(0.08)
  lineWidth = lineWidth * 0.5
  blurMargin = blurMargin * MathUtils.getRandomFloat(1.2, 1.4)
  x = self:getWidthPart(0.5)
  y = self:getHeightPart(0.5)

  for i = 1, additionalTriangleCount do
    local reverse = (i % 2 == 0)
    local offset = MathUtils.getRandomFloat(minimumOffset, maximumOffset)
    if (reverse) then
      offset = offset * -1.0
    end

    local triangleParameters = {0.0, 0.0, sideLength, 0.0, sideLength * 0.5, triangleHeight, triangleColor, lineWidth, blurMargin}

    local triangle = ArcticHollowTriangle.new(triangleParameters, offset)
    triangle:setOpacity(0.0)
    triangle:setLoc(x + offset, y + offset)
    self:insert(triangle, Z_INDEX.additionalTriangles)
    table.insert(self.additionalTriangles, triangle)
    triangle:setBlendMode(MOAIProp2D.BLEND_ADD)

    local triangleBackgroundColor = nil
    if (i == 1) then
      triangleBackgroundColor = TRIANGLE_COLORS.middleTriangle
    elseif (i == 2) then
      triangleBackgroundColor = TRIANGLE_COLORS.rearTriangle
    end

    local triangleBackground = FilledTriangle.new(0.0, 0.0, sideLength, 0.0, sideLength * 0.5, triangleHeight, triangleBackgroundColor)
    triangleBackground:setOpacity(0.0)
    triangle:insertAtCenter(triangleBackground)
    table.insert(self.additionalTriangleBackgrounds, triangleBackground)
    triangleBackground:setBlendMode(MOAIProp2D.BLEND_ADD)

    triangle:startRotating(reverse)
  end
end

-- Constructor ------------------------

function ArcticInputGroup.new(parameters)
  local triangleHeight = math.min(Screen.getSafeWidthPart(0.455), parameters.triangleMaxHeight)
  local triangleSideLength = triangleHeight / 0.886

  local self = Group.new(triangleSideLength, triangleHeight)
  Mixin.mix(self, ArcticInputGroup)

  self.triangleSideLength = triangleSideLength
  self.triangleHeight = triangleHeight

  self.currentInputState = ArcticInputGroup.InputState.empty

  createLayout(self, parameters)
  setupOutlineAnimation(self)

  local onTouchDownWrapper = function()
    onTouchDown(self, parameters.submitAnswerCallback)
  end

  Button.makeButton(self, onTouchDownWrapper)

  self.shouldShowTapToSubmit = true

  return self
end
