--- ArcticMultiTextureBlend
-- @author Created by Joe Poka on December 14, 2016.

class("ArcticMultiTextureBlend")

require("graphics.OffscreenRenderedTexture")
require("utils.ShaderUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SHADER_UNIFORM_INDEX = {
  backgroundSampler = 1,
  sourceSampler = 2,
  normalizationValueHorizontal = 3,
  normalizationValueVertical = 4
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createTexture(group)
  local groupWidth = group:getWidth()
  local groupHeight = group:getHeight()
  local contentScale = Host.contentScale
  local textureWidth = MathUtils.incrementUntilPowerOf2(groupWidth * contentScale)
  local textureHeight = MathUtils.incrementUntilPowerOf2(groupHeight * contentScale)
  local texture = OffscreenRenderedTexture.new(group, textureWidth, textureHeight)
  return texture
end

local function getShader(self)
  local shader = MOAIShader.new()
  local uniformCount = TableUtils.getSize(SHADER_UNIFORM_INDEX)
  shader:reserveUniforms(uniformCount)
  shader:declareUniformSampler(SHADER_UNIFORM_INDEX.backgroundSampler, "backgroundSampler", 1)
  shader:declareUniformSampler(SHADER_UNIFORM_INDEX.sourceSampler, "sourceSampler", 2)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.normalizationValueHorizontal, "normalizationValueHorizontal", 0.0)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.normalizationValueVertical, "normalizationValueVertical", 0.0)

  shader:setVertexAttribute(1, "position")
  shader:setVertexAttribute(2, "uv")
  shader:setVertexAttribute(3, "color")

  local normalizationValueHorizontal = (Screen.width * Host.contentScale / self.textureWidth)
  local normalizationValueVertical = 1.0 - (Screen.height * Host.contentScale / self.textureHeight)
  shader:setAttr(SHADER_UNIFORM_INDEX.normalizationValueHorizontal, normalizationValueHorizontal)
  shader:setAttr(SHADER_UNIFORM_INDEX.normalizationValueVertical, normalizationValueVertical)

  local vsh = ShaderUtils.loadShader("ScrollingParticles.vsh")
  local fsh = ShaderUtils.loadShader("ScrollingParticles.fsh")
  shader:load(vsh, fsh)

  return shader
end

local function getDeck(self, backgroundGroup, sourceGroup)
  local multitexture = MOAIMultiTexture.new()
  local uniformCount = TableUtils.getSize(SHADER_UNIFORM_INDEX)
  multitexture:reserve(uniformCount)

  local backgroundTexture = createTexture(backgroundGroup)
  local foregroundTexture = createTexture(sourceGroup)
  multitexture:setTexture(SHADER_UNIFORM_INDEX.backgroundSampler, backgroundTexture)
  multitexture:setTexture(SHADER_UNIFORM_INDEX.sourceSampler, foregroundTexture)

  -- assuming textures are same size
  self.textureWidth, self.textureHeight = backgroundTexture:getSize()

  local gfxQuad = MOAIGfxQuad2D.new()
  gfxQuad:setTexture(multitexture)

  local width, height = backgroundGroup:getDims()
  local textureWidth, textureHeight = backgroundTexture:getSize()

  local rectOriginX = -width * 0.5
  local rectOriginY = -height * 0.5
  local rectWidth = textureWidth / Host.contentScale
  local rectHeight = textureHeight / Host.contentScale

  gfxQuad:setRect(
    rectOriginX,
    rectOriginY,
    rectOriginX + rectWidth,
    rectOriginY + rectHeight
  )

  gfxQuad:setUVRect(0, 1, 1, 0)

  return gfxQuad
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ArcticMultiTextureBlend.new(backgroundGroup, sourceGroup)
  local self = Group.newFullScreen()
  Mixin.mix(self, ArcticMultiTextureBlend)

  local deck = getDeck(self, backgroundGroup, sourceGroup)
  self:setDeck(deck)

  local shader = getShader(self)
  self:setShader(shader)

  return self
end
