--- ArcticRoundGenerator
-- @author Created by Steven Mah on November 28, 2016

class("ArcticRoundGenerator")

-- -----------------------------------------------------------------------------
-- Configuration
-- -----------------------------------------------------------------------------

local KEY_DIFFICULTY_FILTER = "difficulty"
local KEY_RULE_FILTER = "rule_num"
local RULE_NUMBER_MAX_BY_DIFFICULTY = {
  ["Easy"] = 3,
  ["Medium"] = 7,
  ["Hard"] = 3,
  ["Advanced"] = 3
}

local FIRST_CONCEPT_RULE_NUMBER_BY_DIFFIUCLTY = {
  ["Easy"] = 1,
  ["Medium"] = 1,
  ["Hard"] = 1,
  ["Advanced"] = 2
}

-- -----------------------------------------------------------------------------
-- Utility
-- -----------------------------------------------------------------------------

local function formatConceptIntoRound(concept)
  local values = {}

  for valueIndex = 1, 4 do
    local key = "value" .. valueIndex
    local value = concept[key]

    if (value == "") then
      break
    else
      table.insert(values, value)
    end
  end

  TableUtils.shuffleArray(values)

  local round = {
    conceptId = concept.identifier,
    difficulty = concept.difficulty,
    answer = concept.answer,
    tipType = concept.tipType,
    values = values
  }

  return round
end

-- @param ruleNumber (int) - optional, will pick randomly if not given
local function setRuleFilter(self, difficulty, ruleNumber)
  ruleNumber = ruleNumber or math.random(1, RULE_NUMBER_MAX_BY_DIFFICULTY[difficulty])
  local rule = string.format("%s %d", difficulty, ruleNumber)
  self.filters[KEY_RULE_FILTER] = rule
end

local function chooseConcept(self)
  local allowFewerResults = true
  local concept = self.conceptChooser.nextConcepts(1, self.excludedConceptIdentifiers, self.filters, allowFewerResults)
  if concept then
    self.conceptChooser.markSeenConceptIdentifiers({concept.identifier})
    table.insert(self.excludedConceptIdentifiers, concept.identifier)
    table.insert(self.perFetchSeenConceptIdentifiers, concept.identifier)
  end

  return concept
end

local function generateNextConcept(self)
  if self.difficultyNameChooser then
    local difficulty = self.difficultyNameChooser:nextValue()

    self.filters[KEY_DIFFICULTY_FILTER] = difficulty
    setRuleFilter(self, difficulty)
  end

  local concept = chooseConcept(self)
  if (concept == nil) then
    self:reset()
    concept = chooseConcept(self)
  end

  return concept
end

local function getOrderedDifficultyNamesForPreloadedConcepts(self)
  local orderedDiffifcultyNames = {}

  if self.difficultyNameChooser:getWeightForValue("Easy") > 0 then
    table.insert(orderedDiffifcultyNames, "Easy")
  end
  if self.difficultyNameChooser:getWeightForValue("Easy") >= 0.2 then
    table.insert(orderedDiffifcultyNames, "Easy")
  end
  if self.difficultyNameChooser:getWeightForValue("Easy") >= 0.4 then
    table.insert(orderedDiffifcultyNames, "Easy")
  end

  if self.difficultyNameChooser:getWeightForValue("Medium") >= 0.1 then
    table.insert(orderedDiffifcultyNames, "Medium")
  end
  if self.difficultyNameChooser:getWeightForValue("Medium") >= 0.25 then
    table.insert(orderedDiffifcultyNames, "Medium")
  end
  if self.difficultyNameChooser:getWeightForValue("Medium") >= 0.45 then
    table.insert(orderedDiffifcultyNames, "Medium")
  end

  if self.difficultyNameChooser:getWeightForValue("Hard") >= 0.2 then
    table.insert(orderedDiffifcultyNames, "Hard")
  end
  if self.difficultyNameChooser:getWeightForValue("Hard") >= 0.4 then
    table.insert(orderedDiffifcultyNames, "Hard")
  end

  return orderedDiffifcultyNames
end

local function logPreloadedConcepts(self, orderedDiffifcultyNames)
  local logString = "Preloading concepts using the following difficulty wieghts:\n"
  local difficultyValues = self.difficultyNameChooser:getValues()
  local difficultyWeights = self.difficultyNameChooser:getWeights()
  for i = 1, #difficultyValues do
    if (difficultyWeights[i] > 0) then
      logString = logString .. string.format("%-10s\t%.4f", difficultyValues[i], difficultyWeights[i]) .. "\n"
    end
  end
  logString = logString .. "Preloading ordered concepts:"

  for _, difficultyName in ipairs(orderedDiffifcultyNames) do
    logString = logString .. "\n\t" .. tostring(difficultyName)
  end

  Log.debug(logString)
end

-- -----------------------------------------------------------------------------
-- API
-- -----------------------------------------------------------------------------

function ArcticRoundGenerator:generateNextValue()
  local concept = table.remove(self.orderedConcepts)
  if (concept == nil) then
    concept = generateNextConcept(self)
  end
  local round = formatConceptIntoRound(concept)

  return round
end

function ArcticRoundGenerator:preloadConcepts()
  if (self.difficultyNameChooser == nil) then
    Log.error("No Difficulty Name Chooser set for the Concept Chooser! Cannot preload concepts.")
    return
  end

  local orderedDiffifcultyNames = getOrderedDifficultyNamesForPreloadedConcepts(self)
  logPreloadedConcepts(self, orderedDiffifcultyNames)

  self.orderedConcepts = {}
  for i, difficultyName in ipairs(orderedDiffifcultyNames) do
    self.filters[KEY_DIFFICULTY_FILTER] = difficultyName
    if (i == 1) then
      -- Show as few numbers as possible for the first concept
      local ruleNumber = FIRST_CONCEPT_RULE_NUMBER_BY_DIFFIUCLTY[difficultyName]
      setRuleFilter(self, difficultyName, ruleNumber)
    else
      setRuleFilter(self, difficultyName)
    end
    local concept = chooseConcept(self)
    table.insert(self.orderedConcepts, concept)
  end
end

-- -----------------------------------------------------------------------------
-- Object Creation
-- -----------------------------------------------------------------------------

function ArcticRoundGenerator.new()
  local self = ConceptGenerator.new(GameSession.conceptChooser)
  Mixin.mix(self, ArcticRoundGenerator)

  self.orderedConcepts = {}

  return self
end
