--- ArcticScene
-- @author Created by Joe Poka on October 19, 2016.

class("ArcticScene")

require("graphics.CircleGradient")
require("graphics.MultiGradient")
require("ArcticBackground")
require("ArcticInputGroup")
require("ArcticMessageOverlayAccidentalSum")
require("ArcticMessageOverlayBasicInformation")
require("ArcticMessageOverlayIncorrectAnswer")
require("ArcticNumberPad")
require("ArcticNumbersGroup")
require("ArcticOutro")
require("ArcticRadialGradientTexture")
require("ArcticShake")
require("ArcticStarsGroup")
require("ArcticVerticalGradientTexture")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  background = 0,
  radialGradient = 5,
  verticalGradient = 7,
  numberPadBacking = 10,
  stars = 20,
  inputGroupShine = 30,
  uiGroup = 40,
  backgroundRectangleGradient = 50,
  correctAnswerRectangleGradient = 60,
  outro = 70
}

local UI_Z_INDEX = {
  numberPad = 10,
  overlay = 20,
  inputGroup = 30,
  numbersGroup = 40,
  instructionsLabel = 50
}

local UI_INTRO_ANIMATION_DURATION = 1.5
local UI_INTRO_DELAY_DURATION = 0.6
local UI_INTRO_ANIMATION_SCALE = 1.12
local INSTRUCTIONS_MAXIMUM_OPACITY = 0.8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function numberPadButtonWasPressed(self, keyString)
  self.inputGroup:toggleInputStateForNumberPadKeyPress(keyString)

  local inputField = self.inputGroup:getInputField()
  local inputString = tostring(inputField:getValue())
  local toneIndex = MathUtils.clamp(#inputString + 1, 1, 5)

  if (keyString ~= NumberPad.SpecialKeys.BACKSPACE) then
    local toneFilename = "input0" .. toneIndex
    Sound.playSoundEffect(toneFilename)
  end
end

local function getStarZoomAnimation(self, destinationX, destinationY, addStarAnimation)
  local color = {1.0, 1.0, 1.0, 1.0}
  local sideLength = self.inputGroup:getWidthPart(0.96)
  local height = sideLength * 0.866
  local triangle = FilledTriangle.new(0.0, 0.0, sideLength, 0.0, sideLength * 0.5, height, color)
  local x, y = self.inputGroup:getLoc()
  triangle:setLoc(x, y)
  self:insert(triangle, Z_INDEX.stars)

  local duration = 0.4
  local ease = AHEaseSimpleOut.new(4)
  local move = AHLocationTo.new(triangle, duration, ease, destinationX, destinationY)
  local scale = AHScaleTo.new(triangle, duration, ease, 0.04)
  local zoom = AHSpawn.new(move, scale)

  local fade = AHOpacityTo.new(triangle, duration * 0.2, nil, 0.0)
  local fadeAndAdd = AHSpawn.new(fade, addStarAnimation)
  local delayedFadeAndAdd = AHDelayed.new(duration * 0.855, fadeAndAdd)
  local zoomAndShow = AHSpawn.new(zoom, delayedFadeAndAdd)
  local remove = AHRemove.new(triangle)

  local zoomAnimation = AHSequence.new(zoomAndShow, remove)
  return zoomAnimation
end

local function getFadeGradientAnimation(self)
  local fadeInDuration = 0.06
  local fadeOutDuration = 20.0 * fadeInDuration
  local destinationAlpha = 0.45
  local fadeInGradient = self.correctAnswerRectangleGradient:newOpacityToAction(fadeInDuration, nil, destinationAlpha, destinationAlpha)
  local fadeOutGradient = self.correctAnswerRectangleGradient:newOpacityToAction(fadeOutDuration, nil, 0.0, 0.0)
  local fadeGradient = AHSequence.new(fadeInGradient, fadeOutGradient)
  return fadeGradient
end

local function getCorrectAnswerAnimation(self)
  local actions = {}

  local clearInput = function()
    self.inputGroup:clearInputField()
  end
  table.insert(actions, clearInput)

  local inputCorrectAnimation = self.inputGroup:getCorrectAnswerAnimation()
  table.insert(actions, inputCorrectAnimation)

  local destinationX, destinationY = self.starsGroup:getNextStarLocation()
  local addStarAnimation = self.starsGroup:getAddStarAnimation(destinationX, destinationY)
  local starZoomAnimation = getStarZoomAnimation(self, destinationX, destinationY, addStarAnimation)
  table.insert(actions, starZoomAnimation)

  local fadeGradient = getFadeGradientAnimation(self)
  table.insert(actions, fadeGradient)

  local parameters = {
    minimumDelta = Screen.getWidthPart(0.0008),
    maximumDelta = Screen.getWidthPart(0.008),
    shakeCount = 3,
    useRandomSigns = true
  }

  local screenShake = ArcticShake.getShakeAnimation(self.uiGroup, parameters)
  table.insert(actions, screenShake)

  local backgroundScaleAnimation = self.background:getCorrectAnswerScaleAnimation()
  table.insert(actions, backgroundScaleAnimation)

  local animation = AHSpawn.new(actions)
  return animation
end

local function getIncorrectAnswerAnimation(self)
  local clearInput = function()
    self.inputGroup:clearInputField()
  end

  return clearInput
end

local function getShowNumberPadAnimation(self, duration)
  local showNumberPad = AHOpacityTo.new(self.numberPad, duration, nil, 1.0)
  return showNumberPad
end

local function getHideNumberPadAnimation(self, duration)
  local hideNumberPad = AHOpacityTo.new(self.numberPad, duration, nil, 0.0)
  return hideNumberPad
end

local function delayedCreateLayout(self, submitAnswerCallback)
  local fontSize = 16.0
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local instructionsText = Localizer.get("INSTRUCTIONS")
  self.instructionsLabel = Label.new(instructionsText, font)
  self.instructionsLabel:setOpacity(0.0)
  self.instructionsLabel:setScl(UI_INTRO_ANIMATION_SCALE)
  self.uiGroup:insert(self.instructionsLabel, UI_Z_INDEX.instructionsLabel)

  self.numberPad = ArcticNumberPad.new()

  local inputHeight = Screen.getHeightPart(0.1)
  local triangleMaxHeight = Screen.getSafeYFromTop(0.25) - (self.numberPad:getHeight() + Screen.safeBottom)
  local fontFace = FontFaces.DIN_OT_MEDIUM
  fontSize = 36.0
  local defaultColor = {1.0, 1.0, 1.0}
  local inputGroupParameters = {
    inputHeight = inputHeight,
    triangleMaxHeight = triangleMaxHeight,
    fontFace = fontFace,
    fontSize = fontSize,
    defaultColor = defaultColor,
    submitAnswerCallback = submitAnswerCallback
  }
  self.inputGroup = ArcticInputGroup.new(inputGroupParameters)
  self.uiGroup:insert(self.inputGroup, UI_Z_INDEX.inputGroup)

  local radius = Screen.getWidthPart(0.5)
  local blurMargin = 0
  local centerColor = {1.0, 0.9, 0.64, 0.0}
  local edgeColor = {0.0, 0.0, 1.0, 0.0}
  self.inputGroupShine = CircleGradient.new(radius, centerColor, edgeColor, blurMargin)
  self.inputGroupShine:setBlendMode(MOAIProp2D.BLEND_ADD)
  self:insert(self.inputGroupShine, Z_INDEX.inputGroupShine)

  local width, height = self:getDims()
  local beginColor = {0.54, 1.0, 0.0, 0.0}
  local endColor = {1.0, 0.38, 0.0, 0.0}
  local isHorizontal = true
  self.correctAnswerRectangleGradient = RectangleGradient.new(width, height, beginColor, endColor, isHorizontal)
  self:insertAtCenter(self.correctAnswerRectangleGradient, Z_INDEX.correctAnswerRectangleGradient)
  self.correctAnswerRectangleGradient:setBlendMode(MOAIProp2D.BLEND_ADD)

  self.backgroundRectangleGradient = RectangleGradient.new(width, height, beginColor, endColor, isHorizontal)
  self:insertAtCenter(self.backgroundRectangleGradient, Z_INDEX.backgroundRectangleGradient)
  self.backgroundRectangleGradient:setBlendMode(MOAIProp2D.BLEND_ADD)

  self.numberPad:attachInputField(self.inputGroup:getInputField())
  self.numberPad:setOpacity(0.0)
  self.uiGroup:insert(self.numberPad, UI_Z_INDEX.numberPad)

  local buttonPressCallback = function(keyString)
    numberPadButtonWasPressed(self, keyString)
  end
  self.numberPad:setOnButtonPressCallback(buttonPressCallback)

  self.starsGroup = ArcticStarsGroup.new()
  self:insertAtCenter(self.starsGroup, Z_INDEX.stars)

  self.outro = ArcticOutro.new()
  self:insertAtCenter(self.outro, Z_INDEX.outro)

  local x = self:getWidthPart(0.5)

  self.instructionsLabelDefaultY = Screen.getSafeYFromTop(0.09)

  local instructionsLabelInitialY = self.instructionsLabelDefaultY + self:getHeightPart(0.075)
  self.instructionsLabel:setLoc(x, instructionsLabelInitialY)

  self.inputGroup:setLoc(self.inputGroup:getDefaultLocation())

  self.inputGroupShine:setLoc(self.inputGroup:getLoc())
  self.inputGroupShine:offsetY(-0.1 * self.inputGroupShine:getY())

  local numberPadY = self.numberPad:getHeightPart(0.5) + Screen.safeBottom
  self.numberPad:setLoc(x, numberPadY)
end

local function presentOverlay(self, overlay, showAnimation, hideAnimation)
  self.uiGroup:insertAtCenter(overlay, UI_Z_INDEX.overlay)

  local showOverlay = AHDelayed.new(0.3, overlay:getShowAnimation())
  local hideOverlay = overlay:getHideAnimation()

  local show = AHSpawn.new(showOverlay)
  if (showAnimation ~= nil) then
    show:addAction(showAnimation)
  end

  local hide = AHSpawn.new(hideOverlay)
  if (hideAnimation ~= nil) then
    hide:addAction(hideAnimation)
  end

  local listenForTaps = function()
    local tapToDismissGroup = overlay:getTapToDismissGroup()

    local overlayWasDismissed = false
    local onTouch = function() overlayWasDismissed = true end
    Button.makeButton(tapToDismissGroup, onTouch)

    local userDismissedOverlay = function() return overlayWasDismissed end
    waitUntil(userDismissedOverlay)

    hide:start()
  end

  local overlaySequence = AHSequence.new(show, listenForTaps)
  return overlaySequence
end

local function getUIIntroAnimation(self)
  local duration = UI_INTRO_ANIMATION_DURATION
  local ease = AHEaseExponentialOut.new()
  local fadeNumberPad = AHOpacityTo.new(self.numberPad, duration, ease, 1.0)
  return fadeNumberPad
end

local function getBackgroundGradientAnimation(self)
  local duration = ArcticConfiguration.GAME_DURATION

  local beginColorTo = {0.0, 0.8, 1.0, 0.4}
  local endColorTo = {1.0, 0.0, 0.74, 0.4}

  local colorChange = self.backgroundRectangleGradient:newGradientToAction(duration, nil, beginColorTo, endColorTo)
  return colorChange
end

local function showInstructions(self, showNumbers)
  local duration = 0.3
  local showLabel = AHOpacityTo.new(self.instructionsLabel, duration, nil, INSTRUCTIONS_MAXIMUM_OPACITY)
  local showAnimation = AHSpawn.new(showLabel)
  if (showNumbers) then
    local showNumbersGroup = AHOpacityTo.new(self.numbersGroup, duration, nil, 1.0)
    showAnimation:addAction(showNumbersGroup)
  end
  showAnimation:start()
end

local function hideInstructions(self)
  local duration = 0.3
  local hideLabel = AHOpacityTo.new(self.instructionsLabel, duration, nil, 0.0)
  local hideNumbers = AHOpacityTo.new(self.numbersGroup, duration, nil, 0.0)
  local hideAnimation = AHSpawn.new(hideLabel, hideNumbers)
  hideAnimation:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ArcticScene:getResetAnimation(round)
  local firstResetAnimation = self.waitingForCreateLayout

  if (self.waitingForCreateLayout) then
    self.waitingForCreateLayout = false
    delayedCreateLayout(self, self.submitAnswerCallback)
  end

  local actions = {}

  local resetInputGroup = function()
    self.inputGroup:reset()
    local isFirstRound = self.waitingForUIIntroAnimation
    local presentationAnimation = self.inputGroup:getConceptPresentationAnimation(isFirstRound)
    presentationAnimation:start()
  end
  table.insert(actions, resetInputGroup)

  if (self.numbersGroup ~= nil) then
    local exitingNumbersGroup = self.numbersGroup
    local exitAnimation = exitingNumbersGroup:getExitAnimation()
    local remove = AHRemove.new(exitingNumbersGroup)
    local cleanUpSequence = AHSequence.new(exitAnimation, remove)
    table.insert(actions, cleanUpSequence)
  end

  local addNewGroup = function()
    self.numbersGroup = ArcticNumbersGroup.new(round)
    local numbersGroupX = self:getWidthPart(0.5)

    local destinationY = Screen.getSafeYFromTop(0.15)

    local numbersGroupY = destinationY
    if (firstResetAnimation) then
      numbersGroupY = numbersGroupY + self:getHeightPart(0.05)
      self.numbersGroup:setScl(UI_INTRO_ANIMATION_SCALE)
    end
    self.numbersGroup:setLoc(numbersGroupX, numbersGroupY)
    self.numbersGroup:setOpacity(0.0)
    self.uiGroup:insert(self.numbersGroup, UI_Z_INDEX.numbersGroup)

    local duration = UI_INTRO_ANIMATION_DURATION
    local ease = AHEaseExponentialOut.new()
    local move = AHLocationYTo.new(self.numbersGroup, duration, ease, destinationY)
    local fade = AHOpacityTo.new(self.numbersGroup, duration, ease, 1.0)
    local scale = AHScaleTo.new(self.numbersGroup, duration, ease, 1.0)

    local introAnimation = AHSpawn.new(fade)
    if (firstResetAnimation) then
      introAnimation:addAction(move)
      introAnimation:addAction(scale)

      local moveLabel = AHLocationYTo.new(self.instructionsLabel, duration, ease, self.instructionsLabelDefaultY)
      local fadeLabel = AHOpacityTo.new(self.instructionsLabel, duration, ease, INSTRUCTIONS_MAXIMUM_OPACITY)
      local scaleLabel = AHScaleTo.new(self.instructionsLabel, duration, ease, 1.0)
      introAnimation:addAction(moveLabel)
      introAnimation:addAction(fadeLabel)
      introAnimation:addAction(scaleLabel)
    end

    local delayedIntroAnimation = AHDelayed.new(UI_INTRO_DELAY_DURATION, introAnimation)
    delayedIntroAnimation:start()
  end
  table.insert(actions, addNewGroup)

  local resetAnimation = AHSequence.new(actions)
  return resetAnimation
end

function ArcticScene:getGameIntroAnimation()
  local shakingCompletedCallback = function()
    local fadeGradient = getFadeGradientAnimation(self)
    fadeGradient:start()
  end

  local backgroundAnimation = self.background:getIntroAnimation(shakingCompletedCallback)
  local startOverlayTinting = function() self.background:startOverlayTinting() end

  local introAnimation = AHSequence.new(backgroundAnimation, startOverlayTinting)
  return introAnimation
end

function ArcticScene:getRoundPresentationAnimation(acceptUserInputCallback)
  local actions = {}
  local acceptInputDelay = UI_INTRO_ANIMATION_DURATION * 0.5

  if (self.waitingForUIIntroAnimation) then
    self.waitingForUIIntroAnimation = false
    acceptInputDelay = acceptInputDelay + UI_INTRO_DELAY_DURATION * 0.5
    local UIIntroAnimation = getUIIntroAnimation(self)
    local delayedUIIntroAnimation = AHDelayed.new(UI_INTRO_DELAY_DURATION, UIIntroAnimation)
    table.insert(actions, delayedUIIntroAnimation)

    local fadeInDuration = 1.5
    local fadeInShine = self.inputGroupShine:newOpacityToAction(fadeInDuration, nil, 0.85, 0.0)
    table.insert(actions, fadeInShine)

    local destinationAlpha = 0.3
    local fadeInRectangleGradient = self.backgroundRectangleGradient:newOpacityToAction(fadeInDuration, nil, destinationAlpha, destinationAlpha)

    local rectangleGradientAnimation = getBackgroundGradientAnimation(self)
    local rectangleGradientSequence = AHSequence.new(fadeInRectangleGradient, rectangleGradientAnimation)
    table.insert(actions, rectangleGradientSequence)
  end

  local startAcceptingInput = function()
    self:startAcceptingUserInput()
    acceptUserInputCallback()
  end

  local delayedStartAcceptingInput = AHDelayed.new(acceptInputDelay, startAcceptingInput)
  table.insert(actions, delayedStartAcceptingInput)

  local presentRoundAnimation = AHSpawn.new(actions)
  return presentRoundAnimation
end

function ArcticScene:startAcceptingUserInput()
  self.numberPad:setTouchEnabled(true)
end

function ArcticScene:stopAcceptingUserInput()
  self.numberPad:setTouchEnabled(false)
end

function ArcticScene:getRoundResultAnimation(userAnsweredCorrectly)
  local resultAnimation = nil
  if (userAnsweredCorrectly) then
    resultAnimation = getCorrectAnswerAnimation(self)
  else
    resultAnimation = getIncorrectAnswerAnimation(self)
  end
  return resultAnimation
end

function ArcticScene:getBasicInformationOverlayAnimation()
  local overlay = ArcticMessageOverlayBasicInformation.new()
  local overlaySequence = presentOverlay(self, overlay)
  return overlaySequence
end

function ArcticScene:getAccidentalSumOverlayAnimation()
  local overlay = ArcticMessageOverlayAccidentalSum.new()
  local overlayDuration = overlay:getAnimationDuration()
  local clearInputField = function() self.inputGroup:partialReset() end
  local showNumberPad = getShowNumberPadAnimation(self, overlayDuration)
  local show = AHSpawn.new(clearInputField, showNumberPad)
  local accidentalSumAnimation = self.inputGroup:getAccidentalSumAnimation(overlayDuration)
  local hideNumberPad = getHideNumberPadAnimation(self, overlayDuration)
  local hide = AHSpawn.new(accidentalSumAnimation, hideNumberPad)
  local overlaySequence = presentOverlay(self, overlay, hide, show)
  return overlaySequence
end

function ArcticScene:getTipSequenceAnimation(round, submittedAnswer)
  local toggleInstructionsCallback = function(shouldShowInstructions, showNumbers)
    if (shouldShowInstructions) then
      showInstructions(self, showNumbers)
    else
      hideInstructions(self)
    end
  end

  local overlay = ArcticMessageOverlayIncorrectAnswer.new(round, submittedAnswer, toggleInstructionsCallback)
  local overlayDuration = overlay:getAnimationDuration()

  local hideNumberPad = getHideNumberPadAnimation(self, overlayDuration)
  local hideInputGroup = self.inputGroup:getIncorrectAnswerAnimation(overlayDuration)
  local delayedHideInputGroup = AHDelayed.new(0.02, hideInputGroup)
  local hideSceneUI = AHSpawn.new(hideNumberPad, delayedHideInputGroup)

  local showNumberPad = getShowNumberPadAnimation(self, overlayDuration)

  local overlaySequence = presentOverlay(self, overlay, hideSceneUI, showNumberPad)
  return overlaySequence
end

function ArcticScene:getClearSceneAnimation()
  local duration = 0.5
  local destinationX = Screen.getWidthPart(1.5)
  local move = AHLocationXTo.new(self.uiGroup, duration, nil, destinationX)
  local resetInputGroup = function() self.inputGroup:reset() end
  local fade = AHOpacityTo.new(self.uiGroup, duration * 0.25, nil, 0.0)
  self.uiGroup:linkOpacityAttrToChildren()
  local moveAndFade = AHSpawn.new(move, fade)
  local clearScene = AHSequence.new(resetInputGroup, moveAndFade)
  return clearScene
end

function ArcticScene:getOutroAnimation()
  local outroAnimation = self.outro:getOutroAnimation()
  local duration = ArcticOutro.getOutroAnimationDuration()
  return outroAnimation, duration
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  self.background = ArcticBackground.new()
  self:insertAtCenter(self.background, Z_INDEX.background)

  local verticalGradientTexture = ArcticVerticalGradientTexture.new()
  local verticalGradient = Sprite.newWithTexture(verticalGradientTexture)
  verticalGradient:setBlendMode(MOAIProp2D.BLEND_MULTIPLY)
  local scaleX = self:getWidth() / verticalGradient:getWidth()
  local scaleY = self:getHeight() / verticalGradient:getHeight()
  verticalGradient:setScl(scaleX, scaleY)
  self:insertAtCenter(verticalGradient, Z_INDEX.verticalGradient)

  local radialGradientTexture = ArcticRadialGradientTexture.new()
  local radialGradient = Sprite.newWithTexture(radialGradientTexture)
  local radialGradientWidth = self:getWidthPart(1.0)
  local radialGradientHeight = self:getHeightPart(0.3)
  scaleX = radialGradientWidth / radialGradient:getWidth()
  scaleY = radialGradientHeight / radialGradient:getHeight()
  radialGradient:setScl(scaleX, scaleY)
  radialGradient:setLoc(self:getWidthPart(0.5), self:getHeight() - radialGradientHeight * 0.5)
  self:insert(radialGradient, Z_INDEX.radialGradient)

  local colors = {
    {0.0, 0.0, 0.0, 0.0},
    {0.0, 0.0, 0.0, 0.4},
    {0.0, 0.0, 0.0, 0.9}
  }

  local thicknesses = {
    0.2,
    0.8
  }

  local numberPadBacking = MultiGradient.new(Screen.width, Screen.getHeightPart(0.5), false, colors, thicknesses)
  numberPadBacking:setAnchorPointBottom()
  self:insertAtBottom(numberPadBacking, Z_INDEX.numberPadBacking)
  numberPadBacking:offsetY(Screen.safeBottom)
  self.numberPadBacking = numberPadBacking

  self.uiGroup = Group.newFullScreen()
  self:insertAtCenter(self.uiGroup, Z_INDEX.uiGroup)
end

-- Constructor ------------------------

function ArcticScene.new(submitAnswerCallback)
  local width = Screen.width
  local height = Screen.height
  local color = {0.0, 0.0, 0.0, 1.0}
  local self = FilledRectangle.new(width, height, color)
  Mixin.mix(self, ArcticScene)

  self.submitAnswerCallback = submitAnswerCallback
  self.waitingForCreateLayout = true
  self.waitingForUIIntroAnimation = true

  createLayout(self)

  return self
end
