--- ArcticShinyRectangles
-- @author Created by Joe Poka on November 29, 2016.

class("ArcticShinyRectangles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getDefaultVerticesAndIndices(halfWidth, halfHeight, horizontalSheer)
  local vertices = {
    {
      x = -halfWidth,
      y = -halfHeight
    },
    {
      x = halfWidth,
      y = -halfHeight
    },
    {
      x = halfWidth + horizontalSheer,
      y = halfHeight
    },
    {
      x = -halfWidth + horizontalSheer,
      y = halfHeight
    }
  }

  local indices = {
    1, 2, 3, 1, 3, 4
  }

  return vertices, indices
end

local function getVerticesAndIndices(self, parameters)
  local thetaRadians = 0.523
  local maximumHeightOffsetMagnitude = parameters.triangleHeight * 0.45
  local defaultColorRGB = parameters.defaultColorRGB

  local rectangleAttributes = {}
  for i = 1, parameters.rectangleCount do
    local heightOffset = MathUtils.getRandomFloat(0.0, maximumHeightOffsetMagnitude)
    if (MathUtils.getRandomBoolean()) then
      heightOffset = heightOffset * -1.0
    end

    local length = maximumHeightOffsetMagnitude - heightOffset
    local widthOffset = length * math.tan(thetaRadians)
    if (parameters.leftSide) then
      widthOffset = widthOffset * -1.0
    end

    local excludeBright = MathUtils.getRandomBoolean()
    if (excludeBright) then
      parameters.maximumAlpha = parameters.maximumAlpha * 0.75
    end

    local alpha = MathUtils.getRandomFloat(parameters.minimumAlpha, parameters.maximumAlpha)
    local color = {defaultColorRGB[1], defaultColorRGB[2], defaultColorRGB[3], alpha}

    local attributes = {
      offset = {
        x = widthOffset,
        y = heightOffset
      },
      color = color
    }
    table.insert(rectangleAttributes, attributes)
  end

  local defaultRectangleWidth = parameters.width or Screen.getWidthPart(0.2)
  local defaultRectangleHeight = parameters.height or Screen.getHeightPart(0.006)
  local defaultRectangleHalfWidth = defaultRectangleWidth * 0.5
  local defaultRectangleHalfHeight = defaultRectangleHeight * 0.5
  local defaultHorizontalSheer = -Screen.getWidthPart(0.02)

  local vertices = {}
  local indices = {}

  for i = 1, parameters.rectangleCount do
    local attributes = rectangleAttributes[i]

    local heightOffsetMagnitude = math.abs(attributes.offset.y)
    local heightFactor = (maximumHeightOffsetMagnitude - heightOffsetMagnitude) / maximumHeightOffsetMagnitude
    local rectangleHalfHeight = defaultRectangleHalfHeight * heightFactor
    local rectangleHalfWidth = defaultRectangleHalfWidth * MathUtils.getRandomFloat(0.35, 1.2)
    local horizontalSheer = defaultHorizontalSheer * heightFactor

    local defaultVertices, defaultIndices = getDefaultVerticesAndIndices(rectangleHalfWidth, rectangleHalfHeight, horizontalSheer)

    for _, vertex in ipairs(defaultVertices) do
      local xOffset = attributes.offset.x
      vertex.x = vertex.x + xOffset
      vertex.y = vertex.y + attributes.offset.y
      vertex.color = attributes.color
      table.insert(vertices, vertex)
    end

    local verticesPerRectangle = 4
    for _, index in ipairs(defaultIndices) do
      table.insert(indices, index + (i - 1) * verticesPerRectangle)
    end
  end

  return vertices, indices
end

local function updateBuffer(self, vertices)
  self.buffer:reset()
  for _, vertex in ipairs(vertices) do
    self.buffer:writeFloat(vertex.x, vertex.y)
    self.buffer:writeColor32(unpack(vertex.color))
  end
  self.buffer:bless()
end

local function decrementAlpha(self, deltaAlpha, vertices)
  for _, vertex in ipairs(vertices) do
    local updatedAlpha = vertex.color[4] - deltaAlpha
    vertex.color[4] = math.max(0.0, updatedAlpha)
  end
  updateBuffer(self, vertices)
end

function ArcticShinyRectangles:setAlpha(alpha)
  local vertices = TableUtils.deepCopy(self.defaultVertices)
  for _, vertex in ipairs(vertices) do
    vertex.color[4] = alpha
  end
  updateBuffer(self, vertices)
 end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ArcticShinyRectangles:setDefaultAlpha()
  updateBuffer(self, self.defaultVertices)
end

function ArcticShinyRectangles:getTwinkleAnimation()
  local verticesPerRectangle = 4
  local rectangleCount = #self.defaultVertices / verticesPerRectangle
  local randomRectangle = math.random(1, rectangleCount)
  local lowerBufferIndex = 1 + verticesPerRectangle * (randomRectangle - 1)
  local offsetsOfInterest = {2, 3, 3, 4}
  local shineColor = {1.0, 1.0, 1.0, 1.0}

  local shine = function()
    local vertices = TableUtils.deepCopy(self.defaultVertices)
    for _, offset in ipairs(offsetsOfInterest) do
      vertices[lowerBufferIndex + offset - 1].color = shineColor
    end
    updateBuffer(self, vertices)
  end

  local dim = function()
    self:setDefaultAlpha()
  end

  local delay = MathUtils.getRandomFloat(0.4, 0.8)
  local delayedDim = AHDelayed.new(delay, dim)

  local animation = AHSequence.new(shine, delayedDim)
  return animation
end

function ArcticShinyRectangles:getFadeOutAnimation(duration)
  local approximateTimePerFrame = Screen.stepDuration
  local iterations = math.ceil(duration / approximateTimePerFrame)
  local approximateDeltaAlpha = 1.0 / iterations
  local vertices = TableUtils.deepCopy(self.defaultVertices)
  local stepFade = function()
    decrementAlpha(self, approximateDeltaAlpha, vertices)
  end
  local fadeOut = AHLoop.new(stepFade, iterations)
  return fadeOut
end

function ArcticShinyRectangles:getPulsingAnimation()
  local verticesPerRectangle = 4
  local rectangleCount = #self.defaultVertices / verticesPerRectangle

  local pulseOnce = function()
    local vertices = TableUtils.deepCopy(self.defaultVertices)
    for i = 1, rectangleCount do
      local update = MathUtils.getRandomBoolean()
      if (update) then
        local deltaAlpha = MathUtils.getRandomFloat(0.01, 0.06) * MathUtils.randomSign()
        for j = 1, verticesPerRectangle do
          local vertex = vertices[(i - 1) * verticesPerRectangle + j]
          vertex.color[4] = vertex.color[4] + deltaAlpha
        end
      end
    end
    updateBuffer(self, vertices)

    local waitDuration = MathUtils.getRandomFloat(0.01, 0.1)
    waitFor(waitDuration)
  end
  local pulseLoop = AHLoop.new(pulseOnce)

  local duration = 3.0
  local deltaX = Screen.getWidthPart(0.005)
  local moveLeft = AHLocationXBy.new(self, duration, nil, -deltaX)
  local moveAcross = AHLocationXBy.new(self, 2.0 * duration, nil, 2.0 * deltaX)
  local moveBack = AHLocationXBy.new(self, duration, nil, -deltaX)
  local moveOnce = AHSequence.new(moveLeft, moveAcross, moveBack)
  local moveLoop = AHLoop.new(moveOnce)

  local pulseAnimation = AHSpawn.new(pulseLoop, moveLoop)
  return pulseAnimation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, parameters)
  local format = MOAIVertexFormat.new()
  format:declareCoord(1, MOAIVertexFormat.GL_FLOAT, 2)
  format:declareColor(2, MOAIVertexFormat.GL_UNSIGNED_BYTE)

  local defaultVertices, indices = getVerticesAndIndices(self, parameters)
  self.defaultVertices = defaultVertices
  local vertices = TableUtils.deepCopy(self.defaultVertices)
  local vertexCount = #vertices
  self.buffer = MOAIVertexBuffer.new()
  self.buffer:setFormat(format)
  self.buffer:reserveVerts(vertexCount)
  updateBuffer(self, vertices)

  local indexBuffer = MOAIIndexBuffer.new()
  indexBuffer:reserve(#indices)

  for i = 1, #indices do
    indexBuffer:setIndex(i, indices[i])
  end

  local mesh = MOAIMesh.new()
  mesh:setVertexBuffer(self.buffer)
  mesh:setIndexBuffer(indexBuffer)
  mesh:setPrimType(MOAIMesh.GL_TRIANGLES)

  local shader = ArcticCommon.SHINY_RECTANGLES_SHADER
  mesh:setShader(shader)

  self:setDeck(mesh)
  self:setBlendMode(MOAIProp2D.BLEND_ADD)
end

function ArcticShinyRectangles.new(triangleHeight, rectangleCount, leftSide, width, height, minimumAlpha, maximumAlpha, defaultColorRGB)
  rectangleCount = rectangleCount or 32
  defaultColorRGB = defaultColorRGB or {0.9, 0.9, 0.85}

  if (leftSide == nil) then
    leftSide = true
  end

  local self = MOAIProp2D.new()
  Mixin.mix(self, ArcticShinyRectangles)

  minimumAlpha = minimumAlpha or 0.2
  maximumAlpha = maximumAlpha or 0.7

  local parameters = {
    triangleHeight = triangleHeight,
    rectangleCount = rectangleCount,
    leftSide = leftSide,
    width = width,
    height = height,
    minimumAlpha = minimumAlpha,
    maximumAlpha = maximumAlpha,
    defaultColorRGB = defaultColorRGB
  }

  createLayout(self, parameters)

  return self
end
