--- ArcticShinyRectanglesInputGroup
-- @author Created by Joe Poka on November 29, 2016.

class("ArcticShinyRectanglesInputGroup")

require("ArcticShinyRectangles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INITIAL_SCALE = 0.85
local DEFAULT_MINIMUM_TWINKLE_DELAY = 0.5
local DEFAULT_MAXIMUM_TWINKLE_DELAY = 2.0
local QUICK_MINIMUM_TWINKLE_DELAY = 0.025
local QUICK_MAXIMUM_TWINKLE_DELAY = 0.1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getShowRectangleGroups(groups, isFirstRound, deltaDelay, skipScale)
  local actions = {}
  local delay = 0.0

  deltaDelay = deltaDelay or 0.15
  if (isFirstRound) then
    deltaDelay = 0.085
  end

  for _, group in ipairs(groups) do
    local show = function()
      group:setVisible(true)

      if (not skipScale) then
        local duration = 1.0
        local scale = AHScaleTo.new(group, duration, AHEaseExponentialOut.new(), 1.0)
        scale:start()
      end
    end

    local delayedShow = AHDelayed.new(delay, show)
    table.insert(actions, delayedShow)

    delay = delay + deltaDelay
  end

  local showGroups = AHSpawn.new(actions)
  return showGroups
end

local function getAnimateOutRectangleGroups(groups, moveLeft)
  local actions = {}
  local delay = 0.0
  local deltaDelay = 0.04
  local duration = 1.3

  local destinationX = Screen.getWidthPart(1.4)
  if (moveLeft) then
    destinationX = -Screen.getWidthPart(0.4)
  end

  for _, group in ipairs(groups) do
    local move = AHLocationXTo.new(group, duration, AHEaseExponentialOut.new(), destinationX)
    local delayedMove = AHDelayed.new(delay, move)
    table.insert(actions, delayedMove)
    delay = delay + deltaDelay
  end

  local animation = AHSpawn.new(actions)
  return animation
end

local function getIdleAnimation(self, rectangleGroup)
  local minimumDeltaX = self:getWidthPart(0.005)
  local maximumDeltaX = self:getWidthPart(0.035)
  local deltaX = MathUtils.getRandomFloat(minimumDeltaX, maximumDeltaX)
  if (MathUtils.getRandomBoolean()) then
    deltaX = deltaX * -1.0
  end

  local moveDuration = 4.0
  local moveLeft = AHLocationXBy.new(rectangleGroup, moveDuration, nil, -deltaX)
  local moveRight = AHLocationXBy.new(rectangleGroup, 2.0 * moveDuration, nil, 2.0 * deltaX)
  local moveBack = AHLocationXBy.new(rectangleGroup, moveDuration, nil, -deltaX)
  local moveOnce = AHSequence.new(moveLeft, moveRight, moveBack)
  local move = AHLoop.new(moveOnce)

  local doDelayedTwinkle = function()
    local delay = MathUtils.getRandomFloat(self.minimumTwinkleDelay, self.maximumTwinkleDelay)
    local twinkleAnimation = rectangleGroup:getTwinkleAnimation()
    local delayedTwinkle = AHDelayed.new(delay, twinkleAnimation)
    wait(delayedTwinkle:start())
  end

  local twinkle = AHLoop.new(doDelayedTwinkle)

  local animation = AHSpawn.new(move, twinkle)
  return animation
end

local function resetGroups(groups, initialX, initialY)
  for _, group in ipairs(groups) do
    group:setVisible(false)
    group:setScl(INITIAL_SCALE)
    group:setLoc(initialX, initialY)
    group:setDefaultAlpha()
  end
end

local function getRectangleGroup(self, triangleHeight, rectangleCount, leftSide, width, height, scale)
  scale = scale or INITIAL_SCALE

  local rectangleGroup = ArcticShinyRectangles.new(triangleHeight, rectangleCount, leftSide, width, height)
  rectangleGroup:setVisible(false)
  rectangleGroup:setScl(scale)

  local idleAnimation = getIdleAnimation(self, rectangleGroup)
  idleAnimation:start()

  return rectangleGroup
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ArcticShinyRectanglesInputGroup:getShowRightRectangles(isFirstRound)
  return getShowRectangleGroups(self.rightGroups, isFirstRound)
end

function ArcticShinyRectanglesInputGroup:getShowLeftRectangles(isFirstRound)
  return getShowRectangleGroups(self.leftGroups, isFirstRound)
end

function ArcticShinyRectanglesInputGroup:getCorrectAnswerAnimation()
  local actions = {}

  local moveLeft = true
  local leftGroupsExit = getAnimateOutRectangleGroups(self.leftGroups, moveLeft)
  table.insert(actions, leftGroupsExit)

  moveLeft = false
  local rightGroupsExit = getAnimateOutRectangleGroups(self.rightGroups, moveLeft)
  table.insert(actions, rightGroupsExit)

  local extraLeftGroups = {}
  local extraRightGroups = {}
  for _, rectangleGroup in ipairs(self.extraGroups) do
    if (moveLeft) then
      table.insert(extraLeftGroups, rectangleGroup)
    else
      table.insert(extraRightGroups, rectangleGroup)
    end
    moveLeft = not moveLeft
  end

  local extraLeftGroupsExit = getAnimateOutRectangleGroups(extraLeftGroups, true)
  local extraRightGroupsExit = getAnimateOutRectangleGroups(extraRightGroups, false)
  table.insert(actions, extraLeftGroupsExit)
  table.insert(actions, extraRightGroupsExit)

  local animation = AHSpawn.new(actions)
  return animation
end

function ArcticShinyRectanglesInputGroup:updateExtraParticles(showExtraParticles)
  if (showExtraParticles) then
    if (not self.showingExtraParticles) then
      self.showingExtraParticles = true
      local isFirstRound = false
      local deltaDelay = 0.02
      local skipScale = true
      self.showExtraParticlesAnimation = getShowRectangleGroups(self.extraGroups, isFirstRound, deltaDelay, skipScale)
      self.showExtraParticlesAnimation:start()
    end
  else
    self:partialReset()
  end
end

function ArcticShinyRectanglesInputGroup:partialReset()
  if (self.showExtraParticlesAnimation) then
    self.showExtraParticlesAnimation:stop()
  end

  local initialX = self:getWidthPart(0.5)
  local initialY = self:getHeightPart(0.5)
  resetGroups(self.extraGroups, initialX, initialY)

  self.showingExtraParticles = false
end

function ArcticShinyRectanglesInputGroup:reset()
  local initialX = self:getWidthPart(0.5)
  local initialY = self:getHeightPart(0.5)
  resetGroups(self.leftGroups, initialX, initialY)
  resetGroups(self.rightGroups, initialX, initialY)
  resetGroups(self.extraGroups, initialX, initialY)
end

function ArcticShinyRectanglesInputGroup:updateTwinkleAnimation(blinkQuickly)
  if (blinkQuickly) then
    self.minimumTwinkleDelay = QUICK_MINIMUM_TWINKLE_DELAY
    self.maximumTwinkleDelay = QUICK_MAXIMUM_TWINKLE_DELAY
  else
    self.minimumTwinkleDelay = DEFAULT_MINIMUM_TWINKLE_DELAY
    self.maximumTwinkleDelay = DEFAULT_MAXIMUM_TWINKLE_DELAY
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, triangleHeight)
  local rectangleGroupCount = 2
  self.leftGroups = {}
  self.rightGroups = {}

  for i = 1, rectangleGroupCount do
    local rectangleCount = math.random(8, 10)
    local leftSide = (i % 2 == 0)
    local rectangleGroup = getRectangleGroup(self, triangleHeight, rectangleCount, leftSide)
    self:insertAtCenter(rectangleGroup)

    if (leftSide) then
      table.insert(self.leftGroups, rectangleGroup)
    else
      table.insert(self.rightGroups, rectangleGroup)
    end
  end

  local extraRectangleGroupCount = 7
  self.extraGroups = {}

  for i = 1, extraRectangleGroupCount do
    local rectangleCount = math.random(8, 10)
    local leftSide = (i % 2 == 0)
    local height = triangleHeight * 0.05
    local width = self:getWidthPart(0.45)
    local scale = 1.0
    local rectangleGroup = getRectangleGroup(self, triangleHeight, rectangleCount, leftSide, width, height, scale)
    self:insertAtCenter(rectangleGroup)
    table.insert(self.extraGroups, rectangleGroup)
  end
end

function ArcticShinyRectanglesInputGroup.new(height)
  local self = Group.newFullScreen()
  Mixin.mix(self, ArcticShinyRectanglesInputGroup)

  self.minimumTwinkleDelay = DEFAULT_MINIMUM_TWINKLE_DELAY
  self.maximumTwinkleDelay = DEFAULT_MAXIMUM_TWINKLE_DELAY

  createLayout(self, height)

  return self
end
