--- BrevityGame
-- @author created by Joe Poka on May 2, 2016

class("BrevityGame")
require("bonuses.ResponseSpeedTracker")
require("BrevityCommon")
require("BrevityConfiguration")
require("BrevityData")
require("BrevityRoundGenerator")
require("BrevityInterstitialLabel")
require("BrevityScene")
require("BrevityTangramCounter")
require("StandardUIBuilder")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INSTRUCTIONS_FADE_TIME = 0.3

local INTERSTITIAL_X = Screen.getSafeX(0.5)
local INTERSTITIAL_Y = Screen.getSafeY(0.88)

local UNLOCK_DELAY_AFTER_CONCEPT_CORRECT = 2.0
local UNLOCK_DELAY_AFTER_CONCEPT_INCORRECT = 2.25
local UNLOCK_DELAY_AFTER_RATIONALE_SHOWN = 0.5

local DO_INTRO_DELAY = 0.5

local HIGHLIGHT_ON_TIME_OUT_DURATION_THRESHOLD = 1
local GAME_CONCLUDED_DELAY = 3

local function configureGameDifficulty()
  local passScore = BrevityConfiguration.getPassScore()
  local rankTwoScore = BrevityConfiguration.getRankTwoScore()
  local rankThreeScore = BrevityConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = BrevityConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function didWin(self)
  return (self.currentTangramGroupIndex > 1)
end

local function getCurrentTangramGroup(self)
  local groupIndex = math.min(self.currentTangramGroupIndex, #self.tangramGroups)
  local tangramGroup = self.tangramGroups[groupIndex]
  return tangramGroup
end

local function unlockAfterDelay(self, delay)
  local function delayedUnlock()
    waitFor(delay)
    if self.responseLock then
      self.responseLock:stop()
      self.responseLock = nil
    end
  end
  runInCoroutine(delayedUnlock)
end

local function increaseDifficultyIfThresholdReached(conceptIndex)
  local willIncreaseDifficulty = (BrevityConfiguration.getCorrectConceptIndexToIncreaseDifficulty() == conceptIndex)
  if (willIncreaseDifficulty) then
    local increasedGameDifficulty = math.min(1.0, GameSession.getDifficulty() + 0.25)

    local difficultyKeyFrames = BrevityConfiguration.getDifficultyKeyFrames()
    local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, increasedGameDifficulty)
    GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)
  end
end

local function onConceptCorrect(self)
  self.responseSpeedTracker:notifyCorrectResponse()

  Sound.playSoundEffect("correct")
  GameSession.increaseScore(BrevityConfiguration.POINTS_PER_CORRECT_RESPONSE)

  local currentConcept = self.currentConcept
  local conceptID = currentConcept.identifier
  GameSession.notifyCorrectAnswer(conceptID)
  GameSession.incrementCorrectAnswerCount()

  increaseDifficultyIfThresholdReached(self.currentTangramGroupIndex)

  unlockAfterDelay(self, UNLOCK_DELAY_AFTER_CONCEPT_CORRECT)

  self.currentTangramGroupIndex = self.currentTangramGroupIndex + 1

  self.scoreCounter:setScore(GameSession.getScore())
  self.tangramCounter:incrementProgress()
end

local function onConceptIncorrect(self)
  self.responseSpeedTracker:notifyIncorrectResponse()

  Sound.playSoundEffect("incorrect")

  local currentConcept = self.currentConcept
  local conceptID = currentConcept.identifier
  GameSession.notifyIncorrectAnswer(conceptID)
  GameSession.incrementIncorrectAnswerCount()

  self.timer:addTime(BrevityConfiguration.INCORRECT_RESPONSE_TIME_PENALTY)
  self.timer:pause()

  if (self.timer:getTimeRemaining() <= 0) then
    self.isGameEnded = true
  end

  local tangramGroup = getCurrentTangramGroup(self)
  local tangramColor = tangramGroup[#tangramGroup].color

  local rationaleDismissedCallback = function() unlockAfterDelay(self, UNLOCK_DELAY_AFTER_RATIONALE_SHOWN) end
  self.gameScene:showRationale(currentConcept.rationale, tangramColor, rationaleDismissedCallback)
end

local function onTapCorrect(self, word)
  self.wordInteractingWith = word

  local tangramGroup = getCurrentTangramGroup(self)

  GameSession.contentTracker:notifyCorrectAnswer(self.currentConcept.identifier, "reduction")
  self.gameScene:onTapCorrect(word, tangramGroup)
  onConceptCorrect(self)
end

local function onTapIncorrect(self, word)
  local tangramGroup = getCurrentTangramGroup(self)
  local tangramColor = tangramGroup[#tangramGroup].color

  self.roundGenerator:notifyMistake(self.currentConcept.identifier)

  GameSession.contentTracker:notifyIncorrectAnswer(self.currentConcept.identifier, "reduction", word:getText())
  self.gameScene:onTapIncorrect(word, tangramColor)
  onConceptIncorrect(self)
end

local function onReducibleResponse(self, word)
  if self.canRespond then
    self.responseActionInProgress = true
    self.canRespond = false

    -- The user tapped something. They know how to play!
    if (self.showHelpIfTimeout == true) then
      self.helperTimer:stop()
      self.showHelpIfTimeout = false
    end

    local textModel = self.gameScene:getModel()
    local wordItems = textModel:getItems()
    local wordItem = wordItems[word.wordIndex]
    if not (wordItem:getAttribute("isReducible") or wordItem:getAttribute("isDroppable")) then
      onTapIncorrect(self, word)
    else
      onTapCorrect(self, word)
    end
  end
end

local function onGameConcluded(self)
  waitWhile(function() return self.responseActionInProgress end)

  if (didWin(self)) then
    GameSession.notifyGameTimeEnd()
    self.outroAction, self.colorChangeActions = self.gameScene:playOutroAnimations()

    wait(self.outroAction)
    self.colorChangeActions:stop()

    -- fix for tangram color flickering as outro transitions to results
    local framesToWaitToAvoidFlickering = 30
    waitForFrames(framesToWaitToAvoidFlickering)

    GameSession.notifyGameWin()
  else
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  end
end

local function onTimeOut(self)
  local couldRespondWhenTimeEnded = self.canRespond

  self.isGameEnded = true
  self.canRespond = false

  local tangramGroup = getCurrentTangramGroup(self)
  local tangramColor = tangramGroup[#tangramGroup].color

  local currentTime = self.timer:getTime()
  local conceptSeenDuration = currentTime - self.timeWhenConceptWasPresented
  local sawLastConceptForLongEnough = (conceptSeenDuration > HIGHLIGHT_ON_TIME_OUT_DURATION_THRESHOLD)
  local shouldHighlightExpectedWords = (couldRespondWhenTimeEnded and sawLastConceptForLongEnough)

  self.gameScene:onTimeOut(tangramColor, shouldHighlightExpectedWords)

  local delayUntilConcludingGame = 0
  if (shouldHighlightExpectedWords) then
    delayUntilConcludingGame = GAME_CONCLUDED_DELAY
  end

  waitFor(delayUntilConcludingGame)

  onGameConcluded(self)
end

local function showHelper(self)
  self.timer:pauseAndShowIcon()
  self.canRespond = false

  local highlightedWord = self.gameScene:showHelpAndWaitForTap()
  -- simulate a correct tap after showing help
  self.canRespond = true
  onReducibleResponse(self, highlightedWord)
end

local function createWordsAndAllowTouch(self, concept)
  local function onResponseWrapper(word)
    onReducibleResponse(self, word)
  end

  self.gameScene:createWordsWithConcept(concept, onResponseWrapper)
end

local function initializeGame(self)
  self.tangramGroups = BrevityData.getTangramGroups()
  self.currentTangramGroupIndex = 1

  self.timer:initialize()
end

local function waitForIntro(self)
  local showInstructionsAction = self.instructionsLabel:getShowAction()
  local showSceneAction = self.gameScene:getShowAction()

  waitFor(DO_INTRO_DELAY)
  showInstructionsAction:start()
  waitFor(INSTRUCTIONS_FADE_TIME * 0.5)
  wait(showSceneAction:start())
end

local function startRoundWithConcept(self, concept)
  createWordsAndAllowTouch(self, concept)

  self.timeWhenConceptWasPresented = self.timer:getTime()

  wait(self.gameScene:onStartRound())

  self.timer:resumeAndHideIcon()

  self.canRespond = true

  self.responseSpeedTracker:notifyStart()
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
end

local function waitForResponse(self)
  self.responseLock = AHDelay.new(math.huge)
  wait(self.responseLock:start())
end

local function waitForRoundClean(self)
  local onCleanRoundCoroutine = self.gameScene:getOnCleanRoundCoroutine()

  -- Don't allow the game to move onto a new concept until the scene
  -- is able to finish its cool transition effect
  local hideInstructionsAction = self.instructionsLabel:getHideAction()
  hideInstructionsAction:start()
  wait(onCleanRoundCoroutine:start())
  self.responseActionInProgress = false
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BrevityGame:gameFlow()
  initializeGame(self)

  waitForIntro(self)
  while (not self.isGameEnded) do
    if (self.timer:getTimeRemaining() < BrevityConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
      self.roundGenerator:allowMistakesToResurface()
    end
    self.currentConcept = self.roundGenerator:next()
    if (self.showHelpIfTimeout) then
      self.helperTimer = AHDelayed.new(
        BrevityConfiguration.TUTORIAL_DELAY,
        function()
          showHelper(self)
        end
      )
      self.helperTimer:start()
    end
    startRoundWithConcept(self, self.currentConcept)
    waitForResponse(self)
    waitForRoundClean(self)
  end

  -- Need to resume if the last answer before time runs out is incorrect
  self.timer:resume()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreCounter = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.timer = builder:addDigitTimer(BrevityConfiguration.GAME_DURATION)
  self.timer:setTimeLostOpacityOverride(1.0)

  if self.shouldShowUndelayedTutorial then
    self.timer:setShouldFadeInWithPauseLabel(true)
  end

  local function timeOutWrapper()
    onTimeOut(self)
  end
  self.timer:setTimeoutHandler(timeOutWrapper)

  local instructionsString = BrevityData.getInstructionsString()
  local maxInstructionWidth = Screen.getSafeWidthPart(0.9)
  local topOfPrompt = self.gameScene:getTopOfPromptBackground()
  local instructionSafeArea = self.scoreCounter:getYBottom() - topOfPrompt
  local maxInstructionHeight = instructionSafeArea * 0.9
  local instructionsLabel = BrevityInterstitialLabel.newWithMessage(instructionsString, maxInstructionWidth, maxInstructionHeight)
  local instructionLocationY = topOfPrompt + (instructionSafeArea / 2)
  instructionsLabel:setLoc(INTERSTITIAL_X, instructionLocationY)
  self:insert(instructionsLabel)
  self.instructionsLabel = instructionsLabel

  self.tangramCounter = BrevityTangramCounter.new()
  builder:addWidget(self.tangramCounter, UIOverlay.NORTH)
end

local function createGameElements(self)
  local gameScene = BrevityScene.newHidden(self)
  self:insertAtCenter(gameScene)
  self.gameScene = gameScene

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 BrevityConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(BrevityConfiguration.SPEED_BONUS_WINDOW,
                                                       quickResponseWrapper,
                                                       BrevityConfiguration.SPEED_BONUS_DECAY_START)
end

function BrevityGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, BrevityGame)

  configureGameDifficulty()
  createGameElements(self)
  createUI(self)

  self.roundGenerator = LearningFocusedRoundGenerator.new(BrevityRoundGenerator.new())
  self.showHelpIfTimeout = true
  self.isGameEnded = false
  self.timeWhenConceptWasPresented = nil

  return self
end
