--- BrevityScene
-- @author created by Joe Poka on May 2, 2016

class("BrevityScene")

require("actions.advanced.Shake")
require("BrevityBackground")
require("BrevityColors")
require("BrevityData")
require("BrevityWords")
require("BrevityParticles")
require("BrevityOutroEffect")
require("BrevityShapeUtils")
require("BrevityRationaleOverlay")
require("graphics.SimpleMorphingPolygon")
require("tutorials.TutorialHighlighter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GAME_AREA_WIDTH = Screen.getSafeWidthPart(0.92)

local NUMBER_OF_PERSISTENT_SHAPES = 6
local NUMBER_OF_SHAPES_NOT_TO_BLEND = 4
local NUMBER_OF_SHAPES_TO_BLEND = NUMBER_OF_PERSISTENT_SHAPES - NUMBER_OF_SHAPES_NOT_TO_BLEND
local PERSISTENT_SHAPE_STAGGER_SCALE = 0.0025
local MULTIPLE_OVERLAYS_STAGGER_TIME = 0.1

local HOLE_PADDING_X = 10
local HOLE_PADDING_Y = 5

local REARRANGE_DELAY = 0.5

local SCENE_FADE_DURATION = 0.3

local OUTRO_START_DELAY = 0.7
local OUTRO_COLOR_CHANGE_DELAY = 0.2

local MORPH_DURATION = 0.5

local SLIVER_HEIGHT = Screen.getSafeHeightPart(0.03)

local OVERLAY_COLOR = {0.0, 0.0, 0.1, 0.6}
local TUTORIAL_FONT_SIZE = 17
local INSTRUCTION_LABEL_WIDTH = Screen.getSafeWidthPart(0.8)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getPaletteIndexFromColor(color)
  local closestIndex = 0
  local lowestScore = math.huge
  local palette = BrevityColors.getPalette()
  for colorIndex, possibleColor in ipairs(palette) do
    local currentScore = 0
    local NUMBER_OF_CHANNELS = 3
    for channelIndex = 1, NUMBER_OF_CHANNELS do
      currentScore = currentScore + math.abs(color[channelIndex] - possibleColor[channelIndex])
    end

    if (currentScore < lowestScore) then
      lowestScore = currentScore
      closestIndex = colorIndex
    end
  end

  return closestIndex
end

local function getLineBoundingRectanglesForWords(self, words)
  -- group words by line first to make this easier to work with
  local lineCollection = {}
  local newLine = {}
  local prevY = nil
  for _, word in pairs(words) do
    local _, y = word:getLoc()

    -- This word isn't on the same line, store everything up to this
    -- point as a 'line' and continue on a new line
    if (prevY and prevY ~= y) then
      table.insert(lineCollection, newLine)
      newLine = {}
    end

    prevY = y
    table.insert(newLine, word)
  end
  table.insert(lineCollection, newLine)

  -- Now for each line, make a bounding rect
  local boundingRects = {}
  for _, line in pairs(lineCollection) do
    local rect = {}

    rect.minX = math.huge
    rect.minY = math.huge
    rect.maxX = -math.huge
    rect.maxY = -math.huge

    for _, word in pairs(line) do
      local x, y = self:transformLoc(word)
      local w, h = word:getDims()

      rect.minX = math.min(x, rect.minX)
      rect.maxX = math.max(x + w, rect.maxX)
      rect.minY = math.min(y, rect.minY)
      rect.maxY = math.max(y + h, rect.maxY)
    end

    table.insert(boundingRects, rect)
  end

  return boundingRects
end

local function playTangramMorphAnimation(self, x, y, rotation, index, overlay, staggerGroup)
  local effectDuration = MORPH_DURATION
  local moveEase = AHEaseSimpleInOut.new()

  local function morphToTangram()
    local opacity = 1.0
    if (index > NUMBER_OF_SHAPES_NOT_TO_BLEND) then
      opacity = 1 - (1.0/(NUMBER_OF_SHAPES_TO_BLEND + 1)) * (index - NUMBER_OF_SHAPES_NOT_TO_BLEND)
    end
    overlay:setOpacity(opacity)

    if ((index == 1) and (staggerGroup == 0)) then
      self.addedTangram = overlay
    end

    local function onArrive()
      overlay:bringToBack()
      if index == 1 then
        if staggerGroup == 0 then
          Sound.playSoundEffect("tangram_piece_insert")
        end
      else
        overlay:removeSelf()
      end
    end

    local staggerAmount = math.pow((index - 1), 2) * PERSISTENT_SHAPE_STAGGER_SCALE
    waitFor(staggerAmount)
    self.overlayLayer:insert(overlay)

    local moveAction = AHLocationTo.new(overlay, effectDuration, moveEase, x, y)
    local morphAction = overlay:createMorphToAction(effectDuration,  AHEaseSimpleInOut.new(), 1.0)
    local rotateAction = AHRotationTo.new(overlay, effectDuration, AHEaseSimpleInOut.new(), rotation)
    local combinedTransformsAction = AHSpawn.new(morphAction, moveAction, rotateAction)

    local staggerDelay = AHDelay.new(MULTIPLE_OVERLAYS_STAGGER_TIME * staggerGroup)
    local onArriveAction = AHCallLogic.new(onArrive)

    local fullSequence = AHSequence.new(staggerDelay, combinedTransformsAction, onArriveAction)
    fullSequence:start()
  end
  runInCoroutine(morphToTangram)
end

local function morphOverlaysToTangramArea(self, overlays, tangramData, staggerGroup)
  local x = TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_X_SCALE * tangramData.x
  local y = TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_Y_SCALE * tangramData.y
  local rotation = tangramData.rotation

  for i, overlay in ipairs(overlays) do
    playTangramMorphAnimation(self, x, y, rotation, i, overlay, staggerGroup)
  end
end

local function getOverlaysFromWord(self, originWord, tangramData)
  local wordsForOverlays = {}

  local wordIndices = self.words.textModel:getIndicesOfWordsToRemove(originWord.wordIndex)
  for _, neighborIndex in pairs(wordIndices) do
    table.insert(wordsForOverlays, self.words:getWordViewByIndex(neighborIndex))
  end

  local lineBoundingRectangles = getLineBoundingRectanglesForWords(self, wordsForOverlays)

  local allOverlays = {}
  for _, rect in pairs(lineBoundingRectangles) do
    local overlayForShape = {}
    local width = rect.maxX - rect.minX
    local height = rect.maxY - rect.minY
    local x = rect.minX + width * 0.5
    local y = rect.minY + height * 0.5

    local scaledWidth = TANGRAM_AREA_X_SCALE * tangramData.width
    local startVertices = BrevityShapeUtils.getVerticesOfBoxWithDims(width, height)
    local endVertices = BrevityShapeUtils.getEndVerticesForShape(tangramData.shape, scaledWidth)

    for i = 1, NUMBER_OF_PERSISTENT_SHAPES do
      local newBoundingRect = SimpleMorphingPolygon.new(scaledWidth, scaledWidth, startVertices, endVertices)
      newBoundingRect:setLoc(x, y)
      newBoundingRect:setColor(unpack(tangramData.color))
      table.insert(overlayForShape, newBoundingRect)
    end
    table.insert(allOverlays, overlayForShape)
  end

  return allOverlays
end

local function morphAllOverlaysToTangramArea(self, word, tangramData)
  local function morphAll()
    self.newTangramOverlays = getOverlaysFromWord(self, word, tangramData)

    for i, overlays in pairs(self.newTangramOverlays) do
      morphOverlaysToTangramArea(self, overlays, tangramData, (i - 1))
    end
  end
  runInCoroutine(morphAll)
end

local function createParticlesAtLocation(self, x, y, color)
  local particles = BrevityParticles.newWithColor(color)
  particles:setLoc(x, y)
  self:insert(particles)

  particles:emit()
end

local function createEmptyLayer(parentGroup)
  local newLayer = FilledRectangle.newFullScreen(WHITE)
  newLayer:setOpacity(INVISIBLE_OPACITY)
  parentGroup:insertAtCenter(newLayer)
  return newLayer
end

local function getRectBounds(self, targets)
  local minimumX = math.huge
  local maximumX = -math.huge
  local minimumY = math.huge
  local maximumY = -math.huge

  for _, target in ipairs(targets) do
    local x, y = self:transformLoc(target)
    local width, height = target:getDims()
    minimumX = math.min(minimumX, x)
    maximumX = math.max(maximumX, x + width)
    minimumY = math.min(minimumY, y)
    maximumY = math.max(maximumY, y + height)
  end

  local targetWidth = maximumX - minimumX
  local targetHeight = maximumY - minimumY
  local targetX = minimumX
  local targetY = minimumY

  local rectWidth = targetWidth + HOLE_PADDING_X
  local rectHeight = targetHeight + HOLE_PADDING_Y
  local rectX = targetX + targetWidth * 0.5
  local rectY = targetY + targetHeight * 0.5

  local rectVerticalOffset = rectHeight * 0.1
  rectY = rectY - rectVerticalOffset

  return rectX, rectY, rectWidth, rectHeight, rectVerticalOffset
end

local function getWordsToHighlightByRow(self)
  local firstReducibleWord = self.words:getFirstReducibleWord()
  local wordsToHighlight = self.words:getWordsToHideForWord(firstReducibleWord)

  if (wordsToHighlight ~= nil) then
    local reducibleWordsSet = TableUtils.createSetFromArray(wordsToHighlight)
    local reducibleRows = {}
    local textArea = self.words:getTextArea()
    local viewsByRow = textArea:getViewsByRow()

    for index, row in ipairs(viewsByRow) do
      for _, word in ipairs(row) do
        local wordIsReducible = (reducibleWordsSet[word] ~= nil)
        if (wordIsReducible) then
          if (reducibleRows[index] == nil) then
            reducibleRows[index] = {}
          end
          table.insert(reducibleRows[index], word)
        end
      end
    end
    return reducibleRows
  end
  return nil
end

local function getWordsToHighlight(self)
  local reducibleWords = self.words:getReducibleWords()
  local firstReducibleWord = reducibleWords[1]

  if (firstReducibleWord ~= nil) then
    local reducibleWordsSet = TableUtils.createSetFromArray(reducibleWords)
    local smallestRowIndex = math.huge
    local reducibleRows = {}
    local textArea = self.words:getTextArea()
    local viewsByRow = textArea:getViewsByRow()

    for index, row in ipairs(viewsByRow) do
      for _, word in ipairs(row) do
        local wordIsReducible = (reducibleWordsSet[word] ~= nil)
        if (wordIsReducible) then
          smallestRowIndex = math.min(smallestRowIndex, index)
          if (reducibleRows[index] == nil) then
            reducibleRows[index] = {}
          end
          table.insert(reducibleRows[index], word)
        end
      end
    end

    local wordsToHighlight = reducibleRows[smallestRowIndex]
    return wordsToHighlight
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BrevityScene:showRationale(rationaleText, tangramColor, rationaleDismissedCallback)
  self.rationaleOverlay:bringToFront()

  local delay = 0.17 + 0.515 -- 0.515 from AHAxisOffsetShake duration
  local showOverlayFunction = function() 
    self.rationaleOverlay:show(rationaleText, tangramColor, rationaleDismissedCallback)
  end
  AHDelayed.new(delay, showOverlayFunction):start()
end

function BrevityScene:getWordsToHighlight()
  return getWordsToHighlight(self)
end

function BrevityScene:getHighlightMaskFactoriesForWords(callback)
  local wordsToHighlightByRow = getWordsToHighlightByRow(self)
  local factories = {}
  if (wordsToHighlightByRow ~= nil) then
    for i, row in pairs(wordsToHighlightByRow) do
      local wordsX, wordsY, wordsWidth, wordsHeight, verticalOffset = getRectBounds(self, row)
      local highlightedText = ""
      for _, wordView in ipairs(row) do
        highlightedText = highlightedText .. " " .. wordView:getText()
      end
      local maskFactory = function()
        wordsRect = FilledRectangle.new(wordsWidth, wordsHeight)
        wordsRect:setLoc(wordsX, wordsY)
        local wordsLabel = BrevityTextArea.new(highlightedText, wordsRect:getWidth(), Alignment.CENTER)
        wordsRect:insertAtCenter(wordsLabel)
        wordsLabel:offsetY(verticalOffset)
        return wordsRect
      end
      table.insert(factories, maskFactory)
    end
  end
  return factories
end

function BrevityScene:showHelpAndWaitForTap()
  local reducibleWords = self.words:getReducibleWords()
  local firstReducibleWord = reducibleWords[1]

  if (firstReducibleWord ~= nil) then
    local wordsToHighlight = self:getWordsToHighlight()
    local maskFactories = self:getHighlightMaskFactoriesForWords()

    local helpLayer = TutorialHighlighter.newWithMasks(maskFactories)
    helpLayer:setOverlayColor(unpack(OVERLAY_COLOR))

    local instructionMessage = BrevityData.getTutorialString()
    local instructionLocationYOverride = nil
    local shouldShowArrows = true
    local labelWidth = INSTRUCTION_LABEL_WIDTH
    helpLayer:addInstructionLabel(instructionMessage,
                                  TUTORIAL_FONT_SIZE,
                                  instructionLocationYOverride,
                                  shouldShowArrows,
                                  INSTRUCTION_LABEL_WIDTH)

    helpLayer:setOpacity(1.0)
    self:insertAtCenter(helpLayer)
    helpLayer:waitForTargetTap()
    helpLayer:removeSelf()
  end

  return firstReducibleWord
end

function BrevityScene:onTimeOut(tangramColor, highlightExpectedWords)
  if (highlightExpectedWords and (self.words ~= nil)) then
    local doShake = true
    self.words:highlightCorrectWords(tangramColor, doShake)
  end
end

function BrevityScene:onTapIncorrect(word, tangramColor)
  self.words:highlightAsIncorrect(word)

  local doShake = false
  self.words:highlightCorrectWords(tangramColor, doShake)
end

local function handleTangramTransform(self, startTangram, endTangram)
  local scaledStartWidth = TANGRAM_AREA_X_SCALE * startTangram.width
  local scaledEndWidth = TANGRAM_AREA_X_SCALE * endTangram.width

  local startVertices = BrevityShapeUtils.getEndVerticesForShape(startTangram.shape, scaledStartWidth)
  local endVertices = BrevityShapeUtils.getEndVerticesForShape(endTangram.shape, scaledEndWidth)

  local startX = TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_X_SCALE * startTangram.x
  local startY = TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_Y_SCALE * startTangram.y
  local endX = TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_X_SCALE * endTangram.x
  local endY = TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_Y_SCALE * endTangram.y

  local startRotation = startTangram.rotation
  local endRotation = endTangram.rotation

  local startColor = startTangram.color
  local endColor = endTangram.color

  local morphingPolygon = SimpleMorphingPolygon.new(scaledStartWidth, scaledStartWidth, startVertices, endVertices)
  morphingPolygon:setColor(unpack(startColor))
  morphingPolygon:setLoc(startX, startY)
  morphingPolygon:setRot(startRotation)
  self:insert(morphingPolygon)

  local duration = MORPH_DURATION
  local morph = morphingPolygon:createMorphToAction(duration, AHEaseSimpleInOut.new(), 1.0)
  local translate = AHLocationTo.new(morphingPolygon, duration, AHEaseSimpleInOut.new(), endX, endY)
  local rotate = AHRotationTo.new(morphingPolygon, duration, AHEaseSimpleInOut.new(), endRotation)
  local color = AHColorTo.new(morphingPolygon, duration, AHEaseSimpleInOut.new(), endColor)

  local transform = AHSpawn.new(morph, translate, rotate, color)
  transform:start()

  return morphingPolygon
end

local function transformTangramGroup(self, newTangramGroup)
  local previousTangramGroup = self.previousTangramGroup
  assert(previousTangramGroup ~= nil)

  local polygons = {}

  for i = 1, #previousTangramGroup do
    local startTangram = previousTangramGroup[i]
    local endTangram = newTangramGroup[i]
    local polygon = handleTangramTransform(self, startTangram, endTangram)
    table.insert(polygons, polygon)
  end

  return polygons
end

function BrevityScene:onTapCorrect(word, newTangramGroup)
  AHRotationShake.newScreenShake():start()

  -- assume new tangram is always last entry in group
  local newTangram = newTangramGroup[#newTangramGroup]

  self.words:hideWordAndNeighbors(word)
  morphAllOverlaysToTangramArea(self, word, newTangram)

  -- adjust previous tangrams if we have them
  if (self.previousTangramGroup ~= nil) then

    local polygonsToRemove = {}

    -- clear new tangram from last time
    for i, overlay in ipairs(self.newTangramOverlays) do
      for j, polygon in ipairs(overlay) do
        table.insert(polygonsToRemove, polygon)
      end
    end

    -- clear other tangrams from last time
    if (self.previousMorphingPolygons ~= nil) then
      for i, polygon in ipairs(self.previousMorphingPolygons) do
        table.insert(polygonsToRemove, polygon)
      end
    end

    self.previousMorphingPolygons = transformTangramGroup(self, newTangramGroup)

    -- remove previous polygons after single-frame delay to avoid empty frame
    runInCoroutine(function()
      local framesToWaitToAvoidEmptyFrame = 1
      waitForFrames(framesToWaitToAvoidEmptyFrame)

      for i, polygon in ipairs(polygonsToRemove) do
        polygon:removeSelf()
      end
    end)
  end

  self.previousTangramGroup = newTangramGroup

  local x, y = self:transformLoc(word)
  local w, h = word:getDims()
  local particlesX = x + w * 0.5
  local particlesY = y + h * 0.5
  createParticlesAtLocation(self, particlesX, particlesY, newTangram.color)

  local function delayedRemove()
    waitFor(REARRANGE_DELAY)
    self.words:removeWordAndNeighbors(word)
  end
  runInCoroutine(delayedRemove)
end

function BrevityScene:getOnCleanRoundCoroutine()
  local hideAction = self.words:getHideAction()

  local function onCleanRound()
    wait(hideAction:start())
    self.words:removeSelf()
    self.words = nil
  end
  return newCoroutineWithFunction(onCleanRound)
end

function BrevityScene:createWordsWithConcept(concept, onResponse)
  local words = BrevityWords.newInvisible(concept, GAME_AREA_WIDTH)
  self:insertAtCenter(words)
  self.words = words

  words:setOnResponseCallback(onResponse)
end

function BrevityScene:onStartRound()
  local showAction = self.words:getShowAction()

  return showAction:start()
end

function BrevityScene:getShowAction()
  local fadeAction = AHOpacityTo.new(self, SCENE_FADE_DURATION, nil, VISIBLE_OPACITY)
  return fadeAction
end

function BrevityScene:playOutroAnimations()
   -- Wait one frame to  make sure the addedTangrams list includes the
   -- most recently added tangram
  coroutine.yield()

  local shapeColorChangeActions = {}

  -- want color changing on all tangrams
  local colorChangeTangrams = {self.addedTangram}

  if (self.previousMorphingPolygons ~= nil) then
    for i, tangram in ipairs(self.previousMorphingPolygons) do
      table.insert(colorChangeTangrams, tangram)
    end
  end

  for i, tangram in ipairs(colorChangeTangrams) do
    local delayBeforeStart = OUTRO_START_DELAY
    if (i % 2 == 0) then
      delayBeforeStart = delayBeforeStart + OUTRO_COLOR_CHANGE_DELAY * 0.5
    end

    local colorIndex = getPaletteIndexFromColor({tangram:getColor()})
    local function changeColorFunction()
      local palette = BrevityColors.getPalette()
      colorIndex = colorIndex + 1
      if (colorIndex > #palette) then
        colorIndex = 1
      end
      tangram:setColor(unpack(BrevityColors.getPaletteColor(colorIndex)))
    end

    local waitBeforeStarting = AHDelay.new(delayBeforeStart)

    local changeColorAction = AHCallLogic.new(changeColorFunction)
    local waitBeforeRepeating = AHDelay.new(OUTRO_COLOR_CHANGE_DELAY)

    local colorChangeSequence = AHSequence.new(changeColorAction, waitBeforeRepeating)
    local changeColorLoop = AHLoop.new(colorChangeSequence)

    local fullSequence = AHSequence.new(waitBeforeStarting, changeColorLoop)
    table.insert(shapeColorChangeActions, fullSequence)
  end
  local allColorChangeAction = AHSpawn.new(unpack(shapeColorChangeActions))

  local outro = BrevityOutroEffect.newWithDiameter(TANGRAM_AREA_HEIGHT)
  outro:setLoc(TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_WIDTH * 0.5, TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_HEIGHT * 0.5)
  self:insert(outro)

  local allDotsAction = outro:getAllDotsAction()
  local hideWordsAction = AHDoNothing.new()
  if (self.words ~= nil) then
    self.words:setOnResponseCallback(nil)
    hideWordsAction = self.words:getHideAction()
  end
  local outroAction = AHSpawn.new(allDotsAction, hideWordsAction)

  self.game:showDefaultGameWinOutro()

  self.overlayLayer:bringToFront()
  outro:bringToFront()

  outroAction:start()
  allColorChangeAction:start()
  AHOpacityTo.new(self.outroOverlay, 0.5, nil, 0.5):start()

  return outroAction, allColorChangeAction
end

function BrevityScene:getModel()
  return self.words:getModel()
end

function BrevityScene:getMaskCenterX()
  local wordsToHighlight = getWordsToHighlight(self)
  local wordsX = getRectBounds(self, wordsToHighlight)
  return wordsX
end

function BrevityScene:getTopOfPromptBackground()
  return self.promptBackground:getYTop()
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createLayout(self)
  local background = BrevityBackground.new()
  self:insertAtCenter(background)
  self.background = background

  local promptAreaColor = BrevityColors.getPromptBackgroundColor()
  local promptBackground = FilledRectangle.new(PROMPT_AREA_WIDTH, PROMPT_AREA_HEIGHT, promptAreaColor)
  promptBackground:setLoc(Screen.getWidthPart(0.5), PROMPT_AREA_Y)
  self:insert(promptBackground)
  self.promptBackground = promptBackground

  local tangramAreaColor = BrevityColors.getTangramAreaColor()
  local tangramAreaHeight = TANGRAM_AREA_HEIGHT + TANGRAM_AREA_BOTTOM_MARGIN + SLIVER_HEIGHT
  local tangramBackingRect = FilledRectangle.new(Screen.fullWidth, tangramAreaHeight, tangramAreaColor)
  tangramBackingRect:setLoc(Screen.getWidthPart(0.5), tangramBackingRect:getHeightPart(0.5))
  self:insert(tangramBackingRect)
  self.tangramBackingRect = tangramBackingRect

  local rationaleOverlay = BrevityRationaleOverlay.new(tangramAreaHeight)
  self:insertAtCenter(rationaleOverlay)
  self.rationaleOverlay = rationaleOverlay

  self.overlayLayer = createEmptyLayer(self)

  local outroOverlay = FilledRectangle.newFullScreen({1.0, 1.0, 1.0, 0.0})
  self:insertAtCenter(outroOverlay)
  self.outroOverlay = outroOverlay
end

function BrevityScene.newHidden(game)
  local self = Group.new(Screen.fullWidth, Screen.fullHeight)
  Mixin.mix(self, BrevityScene)

  createLayout(self)

  self:setOpacity(INVISIBLE_OPACITY)

  self.game = game

  return self
end
