--- BrevityTextModel
-- @author created by Joe Poka on May 2, 2016

class("BrevityTextModel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Returns the item at the specified index
-- @param index (number) - Index of the item
function BrevityTextModel:getWordItemByIndex(index)
  local items = self.items
  assert(index <= #items and index >= 1, "Desired index is out of bounds.")
  return items[index]
end

--- Returns the indicies of words that are to be removed along with the word at wordIndex
-- @param wordIndex (number) - Index of the item to be removed
function BrevityTextModel:getIndicesOfWordsToRemove(wordIndex)
  local indicesToRemove = {}
  local groupingIndex = self:getWordItemByIndex(wordIndex):getAttribute("groupingIndex")

  if self:getWordItemByIndex(wordIndex):getAttribute("isDroppable") then
    groupingIndex = 1
  end

  for index, word in pairs(self:getItems()) do
    local isReducible = word:getAttribute("isReducible")
    local isDroppable = word:getAttribute("isDroppable")
    local sameGrouping = word:getAttribute("groupingIndex") == groupingIndex
    if (isReducible and sameGrouping) or isDroppable then
      table.insert(indicesToRemove, index)
    end
  end

  return indicesToRemove
end

--- Returns the indicies of words that are to be removed along with the word at wordIndex,
--- but ignores droppable commas (as those are non-adjacent)
-- @param wordIndex (number) - Index of the item to be removed
function BrevityTextModel:getIndicesOfWordsToBeRemovedForHint(wordIndex)
  local indicesToRemove = {}
  local groupingIndex = self:getWordItemByIndex(wordIndex):getAttribute("groupingIndex")

  for index, word in pairs(self:getItems()) do
    local isReducible = word:getAttribute("isReducible")
    local isDroppable = word:getAttribute("isDroppable")
    local isComma = (word:getText() == ",")
    local sameGrouping = word:getAttribute("groupingIndex") == groupingIndex
    if (isReducible and sameGrouping) or (isDroppable and not isComma) then
      table.insert(indicesToRemove, index)
    end
  end

  return indicesToRemove
end

--- Returns whether a particular TextItem is reducible
-- @param wordItem - the TextItem in question
function BrevityTextModel:isWordItemReducible(wordItem)
  return wordItem:getAttribute("isReducible")
end

--- Returns whether a particular TextItem is droppable
-- @param wordItem - the TextItem in question
function BrevityTextModel:isWordItemDroppable(wordItem)
  return wordItem:getAttribute("isDroppable")
end

--- Removes the wordItem at the specific index along with all its droppable and
-- reducible partners
-- @param wordIndex (number) - Index of the item to remove
function BrevityTextModel:reduceSentenceByWordIndex(wordIndex)
  local wordItem = self:getWordItemByIndex(wordIndex)

  assert(self:isWordItemReducible(wordItem) or self:isWordItemDroppable(wordItem), "Selected word is neither reducible nor droppable.")
  local indicesToRemove = self:getIndicesOfWordsToRemove(wordIndex)

  -- unset skipNextSpacing for TextItems preceding recently removed commas
  local items = self:getItems()
  for _, index in ipairs(indicesToRemove) do
    local wordItemToRemove = items[index]
    local wordToRemove = wordItemToRemove:getText()
    if (wordToRemove == ",") then
      local previousIndex = index - 1
      if (previousIndex >= 1) then
        local previousWordItem = items[previousIndex]
        previousWordItem:setSkipNextSpacing(false)
      end
    end
  end

  -- Indices need to be reversed before removal
  indicesToRemove = TableUtils.getReversedArray(indicesToRemove)

  self:beginTransaction()
  for _, index in ipairs(indicesToRemove) do
    self:removeItems(index, 1)
  end
  self:endTransaction()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function BrevityTextModel.new(textItems)
  local self = TextModel.new(textItems)
  Mixin.mix(self, BrevityTextModel)
  return self
end
