--- BrevityWords
-- @author created by Joe Poka on May 2, 2016

class("BrevityWords")
require("BrevityColors")
require("BrevityCommon")
require("BrevityWordView")
require("BrevityTextArea")

-- ----------------------------------------------
-- Configuration
-- ----------------------------------------------

local GROUP_WIDTH = Screen.width

local FADE_DURATION = 0.3
local SOFT_SHAKE_WIDTH_X = 280
local SOFT_SHAKE_UP_SCALE_Y = 6
local SOFT_SHAKE_DURATION_MULTIPLIER = 0.8

local BLACKLIST_PUNCTUATION = {
  ["?"] = true,
  ["."] = true,
  ["!"] = true
}

-- ----------------------------------------------
-- Utility
-- ----------------------------------------------
local function getWordViews(self)
  return self.textArea:getItemViews()
end

local function onWordTapped(self, word)
  if self.onResponseCallback then
    self.onResponseCallback(word)
  end
end

local function makeTextArea(self, areaWidth)
  local textColor = BrevityColors.getTextColor()

  local textArea = BrevityTextArea.new(self.rawSentence, areaWidth)
  textArea:setColor(unpack(textColor))
  self.textModel = textArea:getModel()
  self.textArea = textArea

  textArea:setAnchorPoint(0.5, 1.0)
  return textArea
end

local function initializeTouch(self)
  local wordViews = getWordViews(self)
  for i, wordView in ipairs(wordViews) do
    wordView.outOfLineAnim = true
    local isPunctuation = BLACKLIST_PUNCTUATION[wordView:getText()]
    if (not isPunctuation) then
      local function onTapCallback(word)
        onWordTapped(self, word)
      end
      BrevityWordView.makeTappable(wordView, i, onTapCallback)
    end
  end
end

local function getFadeAction(self, opacity)
  if (self.fadeAction) then
    self.fadeAction:stop()
  end

  self.fadeAction = AHOpacityTo.new(self, FADE_DURATION, nil, opacity)
  return self.fadeAction
end

-- ----------------------------------------------
-- API
-- ----------------------------------------------

function BrevityWords:setOnResponseCallback(callback)
  self.onResponseCallback = callback
end

function BrevityWords:getShowAction()
  local fadeAction = getFadeAction(self, VISIBLE_OPACITY)
  return fadeAction
end

function BrevityWords:getHideAction()
  local fadeAction = AHSpawn.new(
    -- HACK: Introduced after text system changes to fix an animation regression where
    -- the word added to the sentence in Clarity fades out at the end of the round at a
    -- different rate than the rest of the sentence. Remove hack by cleaning up how
    -- opacity attribute links are handled throughout the game.
    function() self.textArea:applyItemViewOpacityAttributeLinks() end,
    getFadeAction(self, INVISIBLE_OPACITY)
  )
  return fadeAction
end

function BrevityWords:getWordViewByIndex(index)
  local wordViews = getWordViews(self)
  assert(wordViews and index <= #wordViews, "Tried to access word by invalid index.")
  return wordViews[index]
end

function BrevityWords:highlightAsIncorrect(word)
  local shake = AHAxisOffsetShake.new(word, SOFT_SHAKE_WIDTH_X, AHAxisOffsetShake.SHAKE_AXIS.X, SOFT_SHAKE_DURATION_MULTIPLIER)
  shake:start()
  word:showAsIncorrect()
end

function BrevityWords:getReducibleWords()
  local reducibleWords = {}
  local wordViews = getWordViews(self)
  for index, word in pairs(wordViews) do
    local wordItem = self.textModel:getWordItemByIndex(index)
    if wordItem:getAttribute("isReducible") or wordItem:getAttribute("isDroppable") then
      table.insert(reducibleWords, word)
    end
  end
  return reducibleWords
end

function BrevityWords:highlightCorrectWords(correctColor, doShake)
  local reducibleWords = self:getReducibleWords()
  for _, word in pairs(reducibleWords) do
    local _, wordH = word:getDims()
    if doShake then
      AHAxisOffsetShake.new(word, wordH * SOFT_SHAKE_UP_SCALE_Y, AHAxisOffsetShake.SHAKE_AXIS.Y):start()
    end
    word:showAsCorrect(correctColor)
  end
end

function BrevityWords:hideWordAndNeighbors(word)
  local indicesToRemove = self.textModel:getIndicesOfWordsToRemove(word.wordIndex)

  for _, removableIndex in ipairs(indicesToRemove) do
    local removableWord = self:getWordViewByIndex(removableIndex)
    removableWord:setVisible(false)
  end
end

function BrevityWords:removeWordAndNeighbors(word)
  self.textModel:reduceSentenceByWordIndex(word.wordIndex)
end

function BrevityWords:getWordsToHideForWord(word)
  local indicesOfWords = self.textModel:getIndicesOfWordsToBeRemovedForHint(word.wordIndex)
  local words = {}
  for _, index in ipairs(indicesOfWords) do
    local word = self:getWordViewByIndex(index)
    table.insert(words, word)
  end
  return words
end

function BrevityWords:getFirstReducibleWord()
  local wordViews = getWordViews(self)
  for index, word in pairs(wordViews) do
    local wordItem = self.textModel:getWordItemByIndex(index)
    if wordItem:getAttribute("isReducible") then
      return word
    end
  end
  error("No reducible word found in current word views.")
end

function BrevityWords:getModel()
  return self.textModel
end

function BrevityWords:getTextArea()
  return self.textArea
end

-- ----------------------------------------------
-- Object Creation
-- ----------------------------------------------

local function createLayout(self, collectionWidth)
  self.textArea = makeTextArea(self, collectionWidth)
  self:insertAtCenter(self.textArea)

  self.textModel = self.textArea:getModel()
  local wordsH = self.textArea:getHeight()
  self.textArea:setLoc(GROUP_WIDTH * 0.5, PROMPT_AREA_Y + wordsH * 0.5)
  self.textArea:setAttrLink(MOAIColor.ATTR_OPACITY, self)
end

function BrevityWords.newInvisible(inputRound, collectionWidth)
  local self = Group.new(Screen.width, Screen.height)
  Mixin.mix(self, BrevityWords)

  self.rawSentence = inputRound.rawSentence

  createLayout(self, collectionWidth)
  initializeTouch(self)
  self.textArea:applyItemViewOpacityAttributeLinks()

  self:setOpacity(0.0)

  return self
end
