--- BrevityValidator
-- @author Created by Steven Mah on October 25, 2016.
-- ConceptValidator for Brevity.

class("BrevityValidator")
require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'sentence',
  'conceptType'
}

local NON_EMPTY_KEYS = {
  'sentence',
  'rationale'
}

local SINGLE_TAGS = nil
local PAIRED_TAGS = nil
local TERTIARY_TAGS = nil
local OPTIONAL_PAIRED_TAGS = { "reduce", "drop" } -- use optional tags validator so we can catch multiple instances

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateNonEmptyKeys(self, concept)
  local expectedNonEmptyKeys = TableUtils.shallowCopy(NON_EMPTY_KEYS)

  self:assertHasKeys(concept, expectedNonEmptyKeys)
end

local function validateNoTrailingOrLeadingSpace(self, sentence, tag)
  local openingTagWithTrailingSpace = string.format("<%s> ", tag)
  local closingTagWithLeadingSpace = string.format(" </%s>", tag)

  local hasTrailingSpace = (sentence:find(openingTagWithTrailingSpace) ~= nil)
  local hasLeadingSpace = (sentence:find(closingTagWithLeadingSpace) ~= nil)

  self:expectFalse(hasTrailingSpace, "Found a trailing space after an opening tag")
  self:expectFalse(hasLeadingSpace, "Found a leading space before a closing tag")
end

local function validateNoWhitespaceBetweenTagsAndWords(self, sentence)
  for _, tag in ipairs(OPTIONAL_PAIRED_TAGS) do
    validateNoTrailingOrLeadingSpace(self, sentence, tag)
  end
end

local function validateReductionsForEndingPunctuation(self, sentence)
  local reductions = {}
  local patternToCaptureReductions = "<reduce>(.-)</reduce>"

  -- capture all possible reductions and store in table
  for reduction in sentence:gmatch(patternToCaptureReductions) do
    table.insert(reductions, reduction)
  end

  -- test every version of the reduced sentence (drops included) for punctuation
  for _, reduction in ipairs(reductions) do
    local sentenceReduced = sentence:gsub(string.format("<reduce>%s</reduce>", reduction), "")
    local sentenceReducedNoDrops = sentenceReduced:gsub("<drop>(.*)</drop>", "")
    local sentenceReducedNoDropsStripped = sentenceReducedNoDrops:gsub("%b<>", "")
    local sentenceNoTrailingWhitespace = StringUtils.trim(sentenceReducedNoDropsStripped)

    self:testForEndingPunctuation(sentenceNoTrailingWhitespace)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BrevityValidator:validateConcept(concept)
  local allowEmpty = true
  self:validateRequiredKeys(concept, REQUIRED_KEYS, allowEmpty)
  validateNonEmptyKeys(self, concept)
  validateNoWhitespaceBetweenTagsAndWords(self, concept.sentence)
  validateReductionsForEndingPunctuation(self, concept.sentence)

  self:validateTags(concept.sentence, SINGLE_TAGS, PAIRED_TAGS, TERTIARY_TAGS, OPTIONAL_PAIRED_TAGS)
  self:expectStringContains(concept.sentence, "<reduce>", "sentence") -- at least one set of <reduce> tags must be present

  local rationale = concept.rationale
  self:validateTags(rationale)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function BrevityValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, BrevityValidator)

  return self
end
