--- Castle square shape
-- @author Created by Adam Villalobos on Oct 28, 2013.

class("Castle")
require("CastlesBlockBasic")
require("CastlesCommon")
require("CastlesConfiguration")
require("CastlesOverlayBlend")
require("actions.advanced.Shake")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GRID_WIDTH = 15
local GRID_HEIGHT = 16
local LIMIT_LINE_HEIGHT = Screen.height * 0.75
local BLOCK_HEIGHT = LIMIT_LINE_HEIGHT / GRID_HEIGHT
local BLOCK_WIDTH = Screen.width / GRID_WIDTH
local BOTTOM_PLATFORM_COLOR = ColorUtils.colorRGBToStd(51, 35, 32)
local PLATFORM_START_OFFSET_Y = -30
local PLATFORM_HEIGHT = Screen.getHeightPart(0.11)

local POLE_WIDTH = 4
local POLE_COLOR = ColorUtils.colorRGBToStd(104, 79, 59)
local FLAG_COLOR = ColorUtils.colorRGBToStd(155, 77, 67)

local Z_PLATFORM = 10
local Z_FLAG = 20
local Z_ROW_GROUP = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getFlagTrianglePoints(height, width)
  return {
    {x = 0, y = 0}, 
    {x = width, y = height / 2}, 
    {x = 0, y = height}
  }
end

local function createFlag(block)
  local flagHeight = BLOCK_HEIGHT * 0.75
  local flagWidth = flagHeight * 1.25
  local flagPoints = getFlagTrianglePoints(flagHeight, flagWidth)
  local flagTriangle = FilledTriangle.newWithTable(flagPoints, FLAG_COLOR)

  local poleHeight = flagHeight + (BLOCK_HEIGHT * 2)
  local pole = FilledRectangle.new(POLE_WIDTH, poleHeight, POLE_COLOR)

  local flagGroup = Group.new(flagWidth, poleHeight)
  flagGroup:insertAtLeft(pole)
  flagGroup:insertAtTop(flagTriangle)
  flagTriangle:offsetY(flagTriangle:getHeightPart(-0.5))

  flagGroup.blockRef = block
  local flagGroupX = flagGroup.blockRef.finalX + flagTriangle:getWidthPart(0.5)
  local flagGroupY = Screen.height + poleHeight
  flagGroup:setLoc(flagGroupX, flagGroupY)
  return flagGroup
end

-- Create actions ---------------------

local function createFlagAction(flag)
  local flagMoveDuration = 0.35
  local flagMoveY = flag.blockRef.finalY + flag:getHeightPart(0.5)
  local flagAction = AHLocationYTo.new(flag, flagMoveDuration, AHEaseSimpleIn.new(2.0), flagMoveY)
  return flagAction
end

local function createCollapseAction(self)
  local castleMoveEndDuration = 0.61 * 2
  local castleMoveEndY = -Screen.height * 3.5
  local collapseAction = AHLocationYTo.new(self, castleMoveEndDuration, AHEaseSimpleInOut.new(2.0), castleMoveEndY)
  return collapseAction
end

local function createJumpAction(self)
  local castleMoveDuration = 0.11
  local jumpDestinationY = Screen.height * 0.5 - self.baseHeight + PLATFORM_HEIGHT
  local jumpAction = AHLocationYTo.new(self, castleMoveDuration, AHEaseSimpleOut.new(2.0), jumpDestinationY)
  return jumpAction
end

local function createBlockJumpAction(block,height,duration)
  duration = duration or 0.6
  local blockJumpAction = AHLocationYBy.new(block, duration, AHEaseSimpleInOut.new(2.0), height)
  return blockJumpAction
end

local function createRemoveBlockAnimation(block)
  local removeDuration = 0.5
  local deltaX = block:getWidth() * 2.0 * MathUtils.getRandomFloat(-1.0, 1.0)
  local deltaDegrees = 360.0 * MathUtils.getRandomFloat(-1.0, 1.0)
  local destinationX = block:getX() + deltaX

  local translate = AHLocationXTo.new(block, removeDuration, AHEaseSimpleInOut.new(2.0), destinationX)
  local rotate = AHRotationBy.new(block, removeDuration, nil, deltaDegrees)

  local deltaY = -1.2 * block:getYIndex() * BLOCK_HEIGHT
  local fallDown = AHLocationYBy.new(block, removeDuration, AHEaseSimpleIn.new(2.0), deltaY)

  local fadeDelay = removeDuration * 0.9
  local fadeDuration = removeDuration - fadeDelay
  local fadeOut = AHOpacityTo.new(block, fadeDuration, nil, 0.0)
  local delayedFadeOut = AHDelayed.new(fadeDelay, fadeOut)
  local removeAnimation = AHSpawn.new(translate, rotate, fallDown, delayedFadeOut)

  local removeDelay = MathUtils.getRandomFloat(0.1, 0.3)
  local delayedRemove = AHDelayed.new(removeDelay, removeAnimation)
  return delayedRemove
end

-- Play actions -----------------------

local function startMovingFlag(flag)
  local w,h = flag:getDims()
  flag:setLoc(flag.blockRef.finalX + w * 0.5, Screen.height + h)
  local flagAction = createFlagAction(flag)
  return flagAction:start()
end

local function startCollapse(self)
  local collapseAction = createCollapseAction(self)
  return collapseAction:start()
end

local function startJump(self)
  local jumpAction = createJumpAction(self)
  return jumpAction:start()
end

local function startBlockJump(block,height)
  local blockJumpAction = createBlockJumpAction(block,height)
  return blockJumpAction:start()
end

local function getHorizontalShakeAnimation(object, deltaX)
  local shakeCount = 8
  local shakeDuration = 0.02
  local initialX = object:getX()
  local distanceDampenFactor = 0.35
  local shake = AHSequence.new()

  local shake0 = AHLocationXTo.new(object, shakeDuration, AHEaseSimpleIn.new(2.0), initialX + deltaX)
  shake:addAction(shake0)

  for i = 1, shakeCount do
    shake:addActions(
      AHLocationXTo.new(object, 2.0 * shakeDuration, nil, initialX - deltaX),
      AHLocationXTo.new(object, 2.0 * shakeDuration, nil, initialX + deltaX)
    )
    deltaX = deltaX * distanceDampenFactor
  end

  local shakeN = AHLocationXTo.new(object, shakeDuration, nil, initialX)
  shake:addAction(shakeN)

  return shake
end

local function getShakeBlocksAnimation(self)
  local shakeBlocksAnimation = AHSpawn.new()
  for _, blocks in ipairs(self.blocks) do
    for _, block in ipairs(blocks) do
      local deltaXMagnitude = block:getWidth() * MathUtils.getRandomFloat(0.01, 0.075)
      local deltaX = deltaXMagnitude * MathUtils.randomSign() * block:getYIndex()
      local shake = getHorizontalShakeAnimation(block, deltaX)
      shakeBlocksAnimation:addAction(shake)
    end
  end
  return shakeBlocksAnimation
end

local function getShakeCastleForBlocksRemove(self)
  local deltaX = self:getWidthPart(0.06)
  local shake = getHorizontalShakeAnimation(self.castleHolder, deltaX)
  return shake
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Castle:playIntro(introDuration)
  local introAction = AHLocationYTo.new(self.platform, introDuration, AHEaseBackOut.new(2.0), -PLATFORM_HEIGHT * 0.5)
  introAction:start()
end

function Castle:slideOffScreenLeft()
  local delayAction = AHDelay.new(0.8)

  local moveDuration = 0.6
  local moveEase = AHEaseSimpleIn.new(2.0)
  local moveDestinationX = -self.blendedGroup:getWidthPart(0.5)
  local moveAction = AHLocationXTo.new(self.blendedGroup, moveDuration, moveEase, moveDestinationX)

  local slideAction = AHSequence.new(delayAction, moveAction)

  return slideAction:start()
end

function Castle:playEndingSequence()
  return runInCoroutine(function()
    for i = 1, #self.flags do
      self.flags[i]:setVisibleRecursive(true)
      startMovingFlag(self.flags[i])
    end

    waitFor(0.35)
    local changesCount = 12
    local shuffleDelay = 0.05
    for iteration = 1, changesCount do
      for i = 1, #self.blocks do
        for j = 1, #self.blocks[i] do
          local willRandomize = math.random(1,2) == 1
          if (willRandomize) then
            self.blocks[i][j]:randomizeColor()
          end
        end
      end

      if (iteration < changesCount - 3) then
        waitFor(shuffleDelay)
      elseif (iteration < changesCount - 1) then
        waitFor(shuffleDelay * iteration / 5)
      else
        waitFor(shuffleDelay * iteration / 2)
      end
    end
  end)
end

function Castle:hasGroupsLeft()
  return self.currentGroup < #self.blocks
end

function Castle:hasCollapsed()
  return self.isCollapsed
end

function Castle:collapse()
  self.isCollapsed = true
  local jump = startJump(self)
  waitFor(0.05) --start lifting off blocks right as platform eases out

  local yVar = 10
  local yMin = 200
  local block, runningHeight

  for k, _ in pairs(self.columns) do
    runningHeight = 0 --for each column
    for j = 1, #self.columns[k] do
      block = self.columns[k][j]
      if (j == #self.columns[k]) then
        runningHeight = runningHeight + math.random(yVar * 3.5, yVar * 8)
      elseif (j > #self.columns[k] - (#self.columns[k]/3)) then
        runningHeight = runningHeight + math.random(yVar * 2, yVar * 3.5)
      else
        runningHeight = runningHeight + math.random(yVar, yVar * 2)
      end
      startBlockJump(block,yMin + runningHeight)
    end
  end
  wait(jump)
  startCollapse(self)
end

local function getRemoveBlocksAnimation(self, groupIndex)
  local animation = AHSpawn.new()
  local blocksToRemove = self.blocks[groupIndex]

  for _, block in ipairs(blocksToRemove) do
    local remove = createRemoveBlockAnimation(block)

    local reset = function()
      local column = self.columns[block:getXIndex()]
      TableUtils.removeValueFromArray(column, block)

      block:removeSelf()
      block:resetPosition()
      block:setOpacity(1.0)
      block:setRot(0.0)
    end

    local removeAndReset = AHSequence.new(remove, reset)
    animation:addAction(removeAndReset)
  end

  return animation
end

function Castle:removeBlocksGroup()
  local groupToRemove = self.currentGroup
  self.currentGroup = math.max(0, self.currentGroup - 1)

  local animation = self.precomputedRemoveAnimations[groupToRemove] or AHSpawn.new()

  local showTintFunction = function()
    self.blendedGroup:showTint()
  end
  animation:addAction(showTintFunction)

  return animation
end

function Castle:addBlocksGroup()
  self.currentGroup = math.min(#self.blocks, self.currentGroup + 1)

  local groupToDrop = self.currentGroup
  local fallDuration = self.blocks[groupToDrop][1]:getFallDuration()

  local function shakeFunction()
    waitFor(CastlesCommon.CORRECT_EVENT_DELAY)

    local rowGroups = self.rowGroups
    local columns = self.columns
    for _, block in ipairs(self.blocks[groupToDrop]) do
      local yIndex = block:getYIndex()
      local xIndex = block:getXIndex()

      local rowGroup = rowGroups[yIndex]
      rowGroup:insert(block)

      local column = columns[xIndex]
      table.insert(column, block)

      block:startFalling()
    end

    waitFor(fallDuration)

    AHRotationShake.newScreenShake():start()
  end
  runInCoroutine(shakeFunction)

  return fallDuration
end

function Castle:getMaxHeight()
  return self.maxHeight + self.baseHeight
end

function Castle:getCurrentHeight()
  local height = self.precomputedHeights[self.currentGroup]
  return height + self.baseHeight
end

function Castle:setBaseHeight(height)
  self.baseHeight = height
end

function Castle:hideTint()
  self.blendedGroup:hideTint()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function Castle.new()
  local self = Group.new(Screen.width, Screen.height)
  Mixin.mix(self, Castle)

  --create blocks from data
  self.blocks = {} --blocks in falling groups
  self.columns = {} --for traversing current blocks, laid out column -> row
  self.flags = {}

  self.maxHeight = 0
  self.currentGroup = 0
  self.isCollapsed = false

  --load random castle data
  local castleName = "data_castle" .. math.random(1, 3) .. ".lua"
  local castleData, castleSettings = dofile(castleName)

  local castleHolder = Group.newFullScreen()
  self.castleHolder = castleHolder

  local rowGroups = {}
  for i = 1, GRID_HEIGHT do
    local rowGroup = Group.new(Screen.width, Screen.height)
    castleHolder:insertAtCenter(rowGroup, Z_ROW_GROUP)
    table.insert(rowGroups, rowGroup)
  end
  self.rowGroups = rowGroups

  local blocksPerGroup = #castleData / CastlesConfiguration.getCorrectAnswersToWinCount()

  for i = 1, #castleData do
    local block = CastlesBlockBasic.new(BLOCK_WIDTH, BLOCK_HEIGHT, castleData[i], castleSettings)

    --track maximum height for UI line
    self.maxHeight = math.max(self.maxHeight, block.yIndex * BLOCK_HEIGHT)

    --create flags from marked blocks (flag=true)
    if (castleData[i].flag) then
      local flag = createFlag(block)
      table.insert(self.flags, flag)
      self.castleHolder:insert(flag, Z_FLAG)
      flag:setVisibleRecursive(false)
    end

    --calculate group index and insert into blocks
    local index = math.floor((i - 1)/blocksPerGroup) + 1
    if (self.blocks[index] == nil) then
      self.blocks[index] = {}
    end
    table.insert(self.blocks[index], block)

    --create column tables
    if (not self.columns[block.xIndex]) then
      self.columns[block.xIndex] = {}
    end
  end

  --create platform rect
  local platFormWidth = Screen.width * 1.25 --Wider than the screen so that you won't see white space when it shakes
  self.platform = FilledRectangle.new(platFormWidth, PLATFORM_HEIGHT, BOTTOM_PLATFORM_COLOR)
  self.platform:setLoc(Screen.width * 0.5, -PLATFORM_HEIGHT * 0.5 + PLATFORM_START_OFFSET_Y)
  self:insert(self.platform, Z_PLATFORM)

  --precompute heights
  self.precomputedHeights = {
    [0] = 0.0
  }

  for index, blocks in pairs(self.blocks) do
    local height = 0.0
    for _, block in ipairs(blocks) do
      height = math.max(height, block:getYIndex() * BLOCK_HEIGHT)
    end
    self.precomputedHeights[index] = height
  end

  --precompute remove block animations
  self.precomputedRemoveAnimations = {}
  local shakeCastleForBlocksRemove = getShakeCastleForBlocksRemove(self)
  local shakeAllBlocks = getShakeBlocksAnimation(self)

  for index, _ in pairs(self.blocks) do
    local animation = getRemoveBlocksAnimation(self, index)
    animation:addAction(shakeCastleForBlocksRemove)
    animation:addAction(shakeAllBlocks)

    self.precomputedRemoveAnimations[index] = animation
  end

  self.blendedGroup = CastlesOverlayBlend.new(self.castleHolder)
  local x = self.blendedGroup:getWidthPart(0.5)
  local y = self.blendedGroup:getHeightPart(0.5)
  self.blendedGroup:setLoc(x, y)
  self:insert(self.blendedGroup)

  return self
end
