--- CastlesGame
-- @author created by Adam Villalobos October 28, 2013.

class("CastlesGame")
require("Castle")
require("CastlesBackground")
require("CastlesCommon")
require("CastlesConfiguration")
require("CastlesFlickInteractionLayer")
require("CastlesPrompt")
require("CastlesRoundGenerator")
require("CastlesTip")
require("StandardUIBuilder")
require("widgets.TapToContinue")
require("ExperimentsConfiguration")
require("data.generators.LearningFocusedRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------
local Z_BACKGROUND = 10
local Z_FLASH = 20
local Z_CASTLE = 30
local Z_TIP_BACKGROUND = 35
local Z_UI = 40
local Z_TIP = 50

-- Input -----------------------------
local INPUT_STATE =
{
  WAIT = 0,
  PROCCESS = 1,
  STEP = 2
}

-- Timing --------------------------
local INTRO_DURATION = 1.4
local START_DELAY = 1.0
local ROUND_DELAY = 0.7
local PLAY_END_DELAY = 0.15
local AUTO_EXIT_DELAY = 1.75
local DELAY_TO_LET_SFX_FINISH = 0.6

-- Colors ------------------------------
local RED_TINT_COLOR = {1.0, 0.38, 0.27, 0.4}
local RATIONALE_BACKGROUND_COLOR = {0.1, 0.1, 0.1, 0.9}

local function configureGameDifficulty(self)
  self.gameTime = CastlesConfiguration.GAME_DURATION_IN_SECONDS
  GameSession.setTotalTime(self.gameTime)

  local rankTwoScore = CastlesConfiguration.getRankTwoScore()
  local rankThreeScore = CastlesConfiguration.getRankThreeScore()

  -- No pass score set for castles. Pass condition is finishing a castle.
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = CastlesConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.TIME_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.PERFECT_SET_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function waitForShowWinBar(self)
  local winBar = self.winBar

  self:insert(winBar, Z_UI)

  local TIMER_DURATION = 0.3
  local scaleAction = AHScaleTo.new(winBar, TIMER_DURATION, nil, 1.0, 1.0)
  local moveAction = AHLocationXTo.new(winBar, TIMER_DURATION, nil, Screen.width * 0.5)
  local showWinBarAction = AHSpawn.new(scaleAction, moveAction)

  wait(showWinBarAction:start())
end

local function hideWinBar(self)
  self.background:removeLimitLine()
  local winBar = self.winBar
  local hideWinBarAction = AHOpacityTo.new(winBar, 1.2, AHEaseSimpleIn.new(4.0), 0.0)
  return hideWinBarAction:start()
end

local function waitForPrepScene(self)
  self.background:playIntro(INTRO_DURATION)
  self.castle:playIntro(0.35)
  self.scoreLabel:setVisibleRecursive(true)

  waitFor(INTRO_DURATION)

  self.interactionLayer:playIntroFadeIn()

  waitFor(START_DELAY)
end

local function createPrompt(self)
  local promptWidth = Screen.getWidthPart(0.5)
  local promptHeight = Screen.getHeightPart(0.167)
  local prompt = CastlesPrompt.new(promptWidth, promptHeight, self.promptX, self.promptY)
  self:insert(prompt, Z_UI)
  if (self.currentWord ~= nil) then
    prompt:setText(self.currentWord.word)
  end
  return prompt
end

local function nextQuestion(self)
  if (self.timer:getTimeRemaining() < CastlesConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
    self.roundGenerator:allowMistakesToResurface()
  end

  self.currentWord = self.roundGenerator:next()
  self.currentWordID = self.currentWord.identifier

  GameSession.contentTracker:notifyPresentedConcept(self.currentWordID)
  ChatRationaleAnalytics.setConceptId(self.currentWordID)

  self.correctFlagType = self.currentWord.flagType

  if (not self.timeUp) then
    self.prompt = createPrompt(self)
    local showPromptAnimation = self.prompt:getPlayInAnimation()
    wait(showPromptAnimation:start())
  end

  self.interactionLayer:setTargetObject(self.prompt)
  self.interactionLayer:setEnabled(true)
end

local function isGameActive(self)
  return self.castle:hasGroupsLeft() and not self.timeUp and not self.castle:hasCollapsed()
end

local function isGameWon(self)
  return not self.castle:hasGroupsLeft() and not self.timeUp and not self.castle:hasCollapsed()
end

local function performStepAndCheckIfGameEnded(self)
  if (isGameActive(self)) then
    self.inputState = INPUT_STATE.PROCCESS -- lock input until nextQuestion() below
    self.timer:resumeAndHideIcon()

    local function delayedNextQuestionFunction()
      waitFor(ROUND_DELAY)
      nextQuestion(self)
      self.inputState = INPUT_STATE.WAIT
    end
    runInCoroutine(delayedNextQuestionFunction)

    return false
  else
    return true
  end
end

local function startFinalGameLossFadeOut(self)
  if (not self.finalGameLossFadeOutCoroutine) then
    self.finalGameLossFadeOutCoroutine = runInCoroutine(function()
        self.interactionLayer:hideCoachUI()

        local promptFadeAction = self.prompt:getFadeOutAnimation()
        wait(promptFadeAction:start())

        GameSession.notifyGameLoss()
      end)
  end
end

local function onCorrectResponse(self, flagType)
  GameSession.contentTracker:notifyCorrectAnswer(self.currentWordID)
  GameSession.notifyCorrectAnswer(self.currentWordID)
  GameSession.incrementCorrectAnswerCount()

  self.currentLevel = self.currentLevel + 1
  if (self.currentLevel > self.maximumLevel) then
    self.maximumLevel = self.currentLevel
    GameSession.increaseScore(CastlesConfiguration.getPointsForCorrectAnswer())
  end

  AHOpacityTo.new(self.fullScreenFlash, 0.05, nil, 0.2):start()

  Sound.playSoundEffect("correct")

  local correctAnimation = self.interactionLayer:getCorrectAnimation(flagType)
  correctAnimation:start()

  local animateFall = function()
    local fallDuration = self.castle:addBlocksGroup()
    if (isGameWon(self)) then
      self.timer:pause()
    end
    waitFor(CastlesCommon.CORRECT_EVENT_DELAY)
    AHOpacityTo.new(self.fullScreenFlash, fallDuration, nil, 0.0):start()
    self.background:setNewCastleHeight(self.castle:getCurrentHeight(), fallDuration)

    local sfxFileRuntimeUntilThud = 0.29
    local minimumDelay = 0
    local delayToAlignSfxWithAnimation = math.max(fallDuration - sfxFileRuntimeUntilThud, minimumDelay)
    AHDelayed.new(delayToAlignSfxWithAnimation, function() Sound.playSoundEffect("impact") end):start()

    waitFor(fallDuration)

    local scoreToSet = GameSession.getScore()
    self.scoreLabel:setScore(scoreToSet)
  end
  runInCoroutine(animateFall)
end

local function allowFullScreenExit(self)
  local continueExitingGame = function()
    startFinalGameLossFadeOut(self)
  end

  local child = nil
  local timeoutSeconds = AUTO_EXIT_DELAY
  local tapToContinueOverlay = TapToContinue.new(child, timeoutSeconds, continueExitingGame)
end

local function waitForWinOutro(self)
  waitFor(PLAY_END_DELAY)

  AHDelayed.new(0.3, function() Sound.playSoundEffect("goal_reached") end):start()

  self:showDefaultGameWinOutro()
  waitForShowWinBar(self)

  self.castle:slideOffScreenLeft()
  hideWinBar(self)
  wait(self.interactionLayer:hideCoachUI())
  self.interactionLayer:removeSelf()
  wait(self.castle:playEndingSequence())

  waitFor(DELAY_TO_LET_SFX_FINISH)
end

local function showLossOutro(self)
  self.castle:collapse()

  self:showDefaultGameLossOutro()
end

local function getBounceBackAnimation(self, flagType)
  self.bouncingPrompt:setText(self.currentWord.word)
  self.bouncingPrompt:setScl(1.0)
  self.bouncingPrompt:setLoc(self.prompt:getLoc())
  self.bouncingPrompt:resetOpacity()

  local hideOriginalPrompt = function()
    self.prompt:setScl(0.0)
  end
  local delayedHideOriginalPrompt = AHDelayed.new(0.01, hideOriginalPrompt)

  local moveDuration = 1.3
  local destinationX = Screen.getWidthPart(0.5)
  local destinationY = Screen.getHeightPart(0.635)
  local moveInPrompt = AHLocationTo.new(self.bouncingPrompt, moveDuration, AHEaseExponentialOut.new(), destinationX, destinationY)

  local deltaAngleDegrees = 360.0
  if (flagType == CastlesCommon.FlagType.NEGATIVE) then
    deltaAngleDegrees = deltaAngleDegrees * -1.0
  end

  local backgroundFadeInDuration = 0.2
  local backgroundFadeInDelay = 1.0
  local fadeInBackground = AHOpacityTo.new(self.rationaleBackground, backgroundFadeInDuration, nil, RATIONALE_BACKGROUND_COLOR[4])
  local delayedFadeInBackground = AHDelayed.new(backgroundFadeInDelay, fadeInBackground)

  local spinPrompt = AHRotationBy.new(self.bouncingPrompt, moveDuration, AHEaseExponentialOut.new(), deltaAngleDegrees)

  local bounceBackAnimation = AHSpawn.new(
    delayedHideOriginalPrompt,
    moveInPrompt,
    spinPrompt,
    delayedFadeInBackground
  )

  return bounceBackAnimation
end

local function getBounceBackCleanUpAnimation(self)
  local slideDownDuration = 0.2
  local fadeOutPromptDuration = 0.15
  local fadeOutBackgroundDuration = 0.1

  local fadeOutPrompt = self.bouncingPrompt:getFadeOutAnimation(fadeOutPromptDuration)
  local slideDownPrompt = AHLocationYBy.new(self.bouncingPrompt, slideDownDuration, AHEaseSineIn.new(), -Screen.getHeightPart(0.25))
  local fadeOutBackground = AHOpacityTo.new(self.rationaleBackground, fadeOutBackgroundDuration, nil, 0.0)

  local cleanUpAnimation = AHSpawn.new(fadeOutPrompt, slideDownPrompt, fadeOutBackground)
  return cleanUpAnimation
end

local function playIncorrectSoundBasedOnTowerHeight(self)
  local isTowerEmpty = (self.currentLevel == 0)

  if (isTowerEmpty) then
    Sound.playSoundEffect("incorrect_no_tower")
  else
    Sound.playSoundEffect("incorrect")
  end
end

local function onIncorrectResponse(self, flagType)
  playIncorrectSoundBasedOnTowerHeight(self)

  self.currentLevel = math.max(0, self.currentLevel - 1)

  GameSession.contentTracker:notifyIncorrectAnswer(self.currentWordID)
  self.roundGenerator:notifyMistake(self.currentWordID)

  self.interactionLayer:setEnableEmitParticlesOnExit(false)
  self.timer:pauseAndShowIcon()

  GameSession.scoredIncorrectAnswer(self.currentWordID)
  GameSession.incrementIncorrectAnswerCount()

  local removeAnimation = self.castle:removeBlocksGroup()
  self.background:setNewCastleHeight(self.castle:getCurrentHeight(), 0.5)

  local bounceBack = getBounceBackAnimation(self, flagType)
  local bounceBackCleanUp = getBounceBackCleanUpAnimation(self)

  local tip = CastlesTip.new(self.currentWord)
  self:insertAtCenter(tip, Z_TIP)

  local cleanUpIncorrectAnimation = function()
    local hideFlash = AHOpacityTo.new(self.fullScreenFlash, 0.5, nil, 0.0)
    local resetFlash = function()
      self.fullScreenFlash:setColor(unpack({1.0, 1.0, 1.0, 0.0}))
    end
    local hideAndResetFlash = AHSequence.new(hideFlash, resetFlash)
    hideAndResetFlash:start()

    tip:prepareForOutroAnimation()

    local tipOutroAnimation = tip:getOutroAnimation()
    local showTipOutroAndRemove = AHSequence.new(tipOutroAnimation, AHRemove.new(tip))
    showTipOutroAndRemove:start()

    bounceBackCleanUp:start()

    self.castle:hideTint()
    self.interactionLayer:returnToNeutral()
    self.inputState = INPUT_STATE.STEP
  end

  local showRationale = function()
    local onComplete = function()
      local runCallbackInCoroutine = false
      local tapToContinueOverlay = TapToContinue.new(nil, nil, cleanUpIncorrectAnimation, runCallbackInCoroutine)

      tip:setLearnMorePressedCallback(function()
        tapToContinueOverlay:disarm()
      end)

      tip:setLearnMoreClosedCallback(function()
        tapToContinueOverlay:rearm()
      end)

      tapToContinueOverlay:bringToBack()
    end

    local tipIntroAnimation = tip:getIntroAnimation(onComplete)
    tipIntroAnimation:start()
  end

  local delayedShowRationale = AHDelayed.new(0.8125, showRationale)

  self.fullScreenFlash:setColor(unpack(RED_TINT_COLOR))
  local showFlash = AHOpacityTo.new(self.fullScreenFlash, 0.125, nil, 0.4)
  local interactionLayerIncorrectAnimation = self.interactionLayer:getIncorrectAnimation(flagType)

  local incorrectAnimation = AHSpawn.new(
    removeAnimation,
    showFlash,
    interactionLayerIncorrectAnimation,
    delayedShowRationale,
    bounceBack
  )

  incorrectAnimation:start()
end

local function guess(self, flagType)
  if (isGameActive(self) and self.inputState == INPUT_STATE.WAIT) then
    self.inputState = INPUT_STATE.PROCCESS

    local isResponseCorrect = (flagType == self.correctFlagType)

    if (isResponseCorrect) then
      onCorrectResponse(self, flagType)
      self.inputState = INPUT_STATE.STEP
    else
      onIncorrectResponse(self, flagType)
    end
  end
end

local function playShowCorrectAnswerAnimation(self)
  local returnToNeutral = function()
    self.interactionLayer:returnToNeutral()
  end

  local fadeOutPrompt = self.prompt:getFadeOutAnimation()
  local allowExitAction = function() allowFullScreenExit(self) end

  local combinedSequence = AHSequence.new(
    returnToNeutral,
    fadeOutPrompt,
    allowExitAction
  )

  combinedSequence:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------


function CastlesGame:gameFlow()
  waitForPrepScene(self)
  nextQuestion(self)
  self.timer:start()
  self.inputState = INPUT_STATE.WAIT

  repeat
    while (self.inputState ~= INPUT_STATE.STEP) do
      coroutine.yield()
    end
  until performStepAndCheckIfGameEnded(self)

  self.interactionLayer:setEnabled(false)

  local didWin = isGameWon(self)
  if (didWin) then
    local secondsRemaining = self.timer:getTimeRemaining()
    ScoringUtils.applyTimeBonus(secondsRemaining, CastlesConfiguration.POINTS_PER_EXTRA_SECOND_REMAINING, self.bonusTray)

    local incorrectAnswersCount = GameSession.getIncorrectAnswerCount()
    local consistencyBonus = CastlesConfiguration.getPointsForConsistencyBonus(incorrectAnswersCount)
    if (consistencyBonus > 0) then
      if (incorrectAnswersCount == 0) then
        ScoringUtils.applyPerfectSetBonus(consistencyBonus, self.bonusTray)
      else
        ScoringUtils.applyCustomBonus(consistencyBonus, BonusTypes.TYPES.CONSISTENCY_BONUS, self.bonusTray)
      end
    end

    self.scoreLabel:setScore(GameSession.getScore())

    local roundedSecondsRemaining = MathUtils.round(secondsRemaining)
    GameSession.bonusReporter:setBonusCount(BonusTypes.TYPES.TIME_BONUS, roundedSecondsRemaining)

    waitForWinOutro(self)
    GameSession.notifyGameWin()
  else
    self.background.limitLine:doEraseEffect()

    if (self.timeUp) then
      showLossOutro(self)
      playIncorrectSoundBasedOnTowerHeight(self)
    end

    playShowCorrectAnswerAnimation(self)
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  self.castle = Castle.new()
  local baseHeight = Screen.height/50
  self.castle:setBaseHeight(baseHeight)

  local castleX = Screen.width * 0.5
  local castleY = Screen.height * 0.5 + baseHeight
  self.castle:setLoc(castleX, castleY)
  self:insert(self.castle, Z_CASTLE)

  local castleMaxHeight = self.castle:getMaxHeight()
  self.background = CastlesBackground.new(castleMaxHeight)
  self:insertAtCenter(self.background, Z_BACKGROUND)

  local width = Screen.width
  local height = Screen.height/150
  local winBar = FilledRectangle.new(width, height, CastlesCommon.WIN_COLOR)
  winBar:setY(castleMaxHeight)
  winBar:setScl(0, 1)
  self.winBar = winBar

  local interactionLayer = CastlesFlickInteractionLayer.new()
  interactionLayer:setEnabled(false)
  interactionLayer:setEnableEmitParticlesOnExit(true)
  self:insertAtCenter(interactionLayer, Z_UI)

  local function onFlagAsPositive()
    guess(self, CastlesCommon.FlagType.POSITIVE)
    self.interactionLayer:setEnabled(false)
  end
  interactionLayer:setFlagRightCallback(onFlagAsPositive)

  local function onFlagAsNegative()
    guess(self, CastlesCommon.FlagType.NEGATIVE)
    self.interactionLayer:setEnabled(false)
  end
  interactionLayer:setFlagLeftCallback(onFlagAsNegative)

  self.interactionLayer = interactionLayer

  local fullScreenFlash = FilledRectangle.newFullScreen({1, 1, 1})
  fullScreenFlash:setOpacity(0)
  self:insertAtCenter(fullScreenFlash, Z_FLASH)
  self.fullScreenFlash = fullScreenFlash

  self.rationaleBackground = FilledRectangle.new(Screen.width, Screen.height, RATIONALE_BACKGROUND_COLOR)
  self.rationaleBackground:setOpacity(0.0)
  self:insertAtCenter(self.rationaleBackground, Z_TIP_BACKGROUND)

  self.bouncingPrompt = createPrompt(self)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), CastlesCommon.UI_COLOR)

  local onCompleteHandler = function()
    self.timeUp = true
    self.inputState = INPUT_STATE.STEP
  end

  self.timer = builder:addDigitTimer(CastlesConfiguration.GAME_DURATION_IN_SECONDS, onCompleteHandler)
  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
end

function CastlesGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, CastlesGame)

  self.promptX = Screen.width * 0.5
  self.promptY = CastlesCommon.COACH_MARKS_Y

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  ChatRationaleAnalytics.setSkill("castles")

  self.timeUp = false
  self.inputState = INPUT_STATE.PROCCESS

  self.finalGameLossFadeOutCoroutine = nil

  self.conceptGenerator = CastlesRoundGenerator.new()
  self.roundGenerator = LearningFocusedRoundGenerator.new(self.conceptGenerator)

  self.currentLevel = 0
  self.maximumLevel = 0

  return self
end
