--- CastlesOverlayBlend
-- @author Created by Joe Poka on March 8, 2017.

class("CastlesOverlayBlend")

require("graphics.OffscreenRenderedTexture")
require("utils.ShaderUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SHADER_UNIFORM_INDEX = {
  SAMPLER = 1,
  NORMALIZATION_VALUE_HORIZONTAL = 2,
  NORMALIZATION_VALUE_VERTICAL = 3,
  TINT_COLOR_WEIGHT = 4
}

local MAXIMUM_TINT_WEIGHT = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getShader(self)
  local shader = MOAIShader.new()
  local uniformCount = TableUtils.getSize(SHADER_UNIFORM_INDEX)
  shader:reserveUniforms(uniformCount)
  shader:declareUniformSampler(SHADER_UNIFORM_INDEX.SAMPLER, "sampler", 1)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_HORIZONTAL, "normalizationValueHorizontal", 0.0)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_VERTICAL, "normalizationValueVertical", 0.0)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.TINT_COLOR_WEIGHT, "tintColorWeight", 0.0)
  shader:setVertexAttribute(1, "position")
  shader:setVertexAttribute(2, "uv")

  local normalizationValueHorizontal = (Screen.width * Host.contentScale / self.textureWidth)
  local normalizationValueVertical = 1.0 - (Screen.height * Host.contentScale / self.textureHeight)
  shader:setAttr(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_HORIZONTAL, normalizationValueHorizontal)
  shader:setAttr(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_VERTICAL, normalizationValueVertical)

  local vsh = ShaderUtils.loadShader("overlayShader.vsh")
  local fsh = ShaderUtils.loadShader("overlayShader.fsh")
  shader:load(vsh, fsh)

  return shader
end

local function createTexture(group)
  local groupWidth = group:getWidth()
  local groupHeight = group:getHeight()
  local contentScale = Host.contentScale
  local textureWidth = MathUtils.incrementUntilPowerOf2(groupWidth * contentScale)
  local textureHeight = MathUtils.incrementUntilPowerOf2(groupHeight * contentScale)
  local texture = OffscreenRenderedTexture.new(group, textureWidth, textureHeight)
  return texture
end

local function adjustTintWeight(self, deltaWeight)
  local incrementWeight = function()
    self.tintWeight = math.min(MAXIMUM_TINT_WEIGHT, math.max(0.0, self.tintWeight + deltaWeight))
    self.shader:setAttr(SHADER_UNIFORM_INDEX.TINT_COLOR_WEIGHT, self.tintWeight)
  end

  local loopCount = math.ceil(MAXIMUM_TINT_WEIGHT / math.abs(deltaWeight))
  local changeWeight = AHLoop.new(incrementWeight, loopCount)
  changeWeight:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CastlesOverlayBlend:showTint()
  local deltaWeight = 0.1
  adjustTintWeight(self, deltaWeight)
end

function CastlesOverlayBlend:hideTint()
  local deltaWeight = -0.01
  adjustTintWeight(self, deltaWeight)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CastlesOverlayBlend.new(group)
  local texture = createTexture(group)
  local self = Sprite.newWithTexture(texture)
  Mixin.mix(self, CastlesOverlayBlend)

  self.textureWidth, self.textureHeight = texture:getSize()
  self.shader = getShader(self);
  self:setShader(self.shader)

  self.tintWeight = 0.0

  return self
end
