--- CastlesRoundGenerator
-- @author Created by Joe Poka on March 13, 2017.

class("CastlesRoundGenerator")

require("CastlesCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DYNAMIC_ODDS_MULTIPLIER = 0.8
local INITIAL_STREAK_ODDS = 0.4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Bucket concepts into Positive/Negative so we can pull a
-- concept with a specific connotation on demand
local function bucketConcepts(self, concepts)
  for _, concept in ipairs(concepts) do
    if (concept.connotation == "positive") then
      concept.flagType = CastlesCommon.FlagType.POSITIVE
      table.insert(self.positiveConceptPool, concept)
    elseif (concept.connotation == "negative") then
      concept.flagType = CastlesCommon.FlagType.NEGATIVE
      table.insert(self.negativeConceptPool, concept)
    else
      error(concept.connotation .. ": Unknown connotation")
    end
  end
end

-- This prevents the game from loading too few concepts if there aren't
-- enough Basic or Advanced concepts available that the game is trying to load.
-- Instead it will keep trying to load concepts until it gets enough from
-- other difficulties.
local function loadAndBucketConcepts(self)
  GameSession.conceptGenerator:setAllowFewerResults(true)
  local conceptsToLoadCount = CastlesConfiguration.getConceptsToLoadCount()

  local loadedConcepts = {}

  while (#loadedConcepts < conceptsToLoadCount) do
    loadedConcept = GameSession.conceptGenerator:next()
    if (loadedConcept ~= nil) then
      table.insert(loadedConcepts, loadedConcept)
    else
      Log.debug("Failed to load a new concept, retrying...")
    end
  end

  bucketConcepts(self, loadedConcepts)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- We're using a "Dynamic Odds" system, which causes users to see concepts that have
-- alternating connotations more often. This leads to more enjoyable gameplay and
-- a lower chance of a game having a long streak of words with the same connotation
function CastlesRoundGenerator:next()
  local nextConceptIsPositive = self.lastConceptWasPositive
  local shouldContinueStreak = (math.random() < self.continueStreakOdds)
  if shouldContinueStreak then
    -- Reduce the chance that the streak continues on the next concept
    self.continueStreakOdds = self.continueStreakOdds * DYNAMIC_ODDS_MULTIPLIER
  else
    nextConceptIsPositive = (not nextConceptIsPositive)
    self.lastConceptWasPositive = nextConceptIsPositive
    self.continueStreakOdds = INITIAL_STREAK_ODDS
  end

  -- Ensure we don't try and pull a concept from an empty pool
  local positiveConceptsRemain = (#self.positiveConceptPool > 0)
  local negativeConceptsRemain = (#self.negativeConceptPool > 0)
  local doRemove = true
  if ((nextConceptIsPositive and positiveConceptsRemain) or (not negativeConceptsRemain)) then
    return TableUtils.getRandomElement(self.positiveConceptPool, doRemove)
  else
    return TableUtils.getRandomElement(self.negativeConceptPool, doRemove)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CastlesRoundGenerator.new()
  local self = {}
  Mixin.mix(self, CastlesRoundGenerator)

  self.positiveConceptPool = {}
  self.negativeConceptPool = {}
  loadAndBucketConcepts(self)

  self.lastConceptWasPositive = MathUtils.getRandomBoolean()
  self.continueStreakOdds = INITIAL_STREAK_ODDS

  return self
end
