--- CastlesTip
-- @author Created by joe poka on March 13, 2017.

class("CastlesTip")

require("ExperimentsConfiguration")
require("widgets.rationales.ChatRationale")
require("widgets.rationales.ChatRationaleLearnMoreButton")
require("CastlesChatRationaleHelper")
require("widgets.rationales.ChatRationaleAnalytics")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_TIP_CARD = {0.87, 0.87, 0.86, 1.0}
local COLOR_TIP = ColorUtils.colorHex24ToStd(0x7D7D7D)

local BASE_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 20)

local BASE_ATTRIBUTES = {
  font = BASE_FONT
}

local ATTRIBUTES_FOR_LABELS = {
  b = { font = Font.new(FontFaces.DIN_OT_BOLD, 20) },
   --offset needed because the medium italic font file oddly has a higher baseline y
  i = { font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, 20), yOffset = -1 }
}

local Z_INDEX = {
  BACKGROUND = 0,
  CARD = 10
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getLearnMorePressedCallback(self)
  return function()
    -- Do these function calls immediately. Avoids a bug where there is a delay before
    -- the button is disabled and the callback is called if included as a function
    -- call in an AHSequence.
    self.learnMoreButton:setTouchEnabled(false)
    if (self.learnMorePressedCallback ~= nil) then
      self.learnMorePressedCallback()
    end

    local openAction = AHLocationYTo.new(self.tipCard, 0.7, AHEaseExponentialInOut.new(), self:getYTop())
    self.chatRationale:openWithAction(openAction)
  end
end

local function getChatRationaleClosedCallback(self)
  return function()
    local tipCardVisibleY = self.tipCard:getHeightPart(0.5)

    local closeAction = AHSequence.new(
      AHLocationYTo.new(self.tipCard, 0.5, AHEaseSineInOut.new(), tipCardVisibleY),
      function()
        self.learnMoreButton:setTouchEnabled(true)

        if (self.learnMoreClosedCallback ~= nil) then
          self.learnMoreClosedCallback()
        end
      end
    )

    self.chatRationale:closeWithAction(closeAction)
  end
end

local function reportOnPresentationAnalytics(self)
  ChatRationaleAnalytics.reportExperimentExposed()

  local hasLearnMoreAvailable = (self.learnMoreButton ~= nil)
  ChatRationaleAnalytics.reportMainRationaleSeen(hasLearnMoreAvailable)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CastlesTip:getIntroAnimation(onCompleteCallback)
  local duration = 0.3

  local destinationY = self:getHeightPart(0.5)
  local slideInCard = AHSequence.new(
    AHLocationYTo.new(self.tipCard, duration, AHEaseSineOut.new(), destinationY),
    function()
      reportOnPresentationAnalytics(self)

      if (self.learnMoreButton ~= nil) then
        self.learnMoreButton:setTouchEnabled(true)
      end

      if (onCompleteCallback ~= nil) then
        onCompleteCallback()
      end
    end
  )

  return slideInCard
end

function CastlesTip:prepareForOutroAnimation()
  if (self.learnMoreButton ~= nil) then
    self.learnMoreButton:setTouchEnabled(false)
  end
end

function CastlesTip:getOutroAnimation()
  local duration = 0.2
  local destinationY = 0
  local slideOutCard = AHLocationYTo.new(self.tipCard, duration, AHEaseSineIn.new(), destinationY)

  return slideOutCard
end

function CastlesTip:setLearnMorePressedCallback(callback)
  self.learnMorePressedCallback = callback
end

function CastlesTip:setLearnMoreClosedCallback(callback)
  self.learnMoreClosedCallback = callback
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, wordData)
  local tipCardHeight = self:getHeightPart(0.5)
  self.tipCard = FilledRectangle.new(self:getWidth(), tipCardHeight, COLOR_TIP_CARD)
  self:insert(self.tipCard, Z_INDEX.CARD)
  self.tipCard:setAnchorPointTop()

  local tipCardHiddenY = -self.tipCard:getHeightPart(0.5)
  self.tipCard:setLoc(self.tipCard:getWidthPart(0.5), tipCardHiddenY)

  local marginY = self.tipCard:getHeightPart(0.05)

  local tipLabelY = self.tipCard:getHeight() - marginY
  local tipLabel = Label.new(Localizer.get("TIP"), Font.new(FontFaces.DIN_OT_BOLD, 16))
  tipLabel:setColor(unpack(COLOR_TIP))
  tipLabel:setAnchorPointTopLeft()

  local minimumX = Screen.safeLeft + tipLabel:getWidthPart(0.25)
  local marginX = math.max(self.tipCard:getWidthPart(0.1), minimumX)

  self.tipCard:insertAtLocation(tipLabel, marginX, tipLabelY)

  local continueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 15))
  continueLabel:setColor(unpack(COLOR_TIP))
  continueLabel:setAnchorPointBottom()

  local minimumContinueLabelY = Screen.safeBottom + continueLabel:getHeight()
  local continueLabelY = math.max(continueLabel:getHeightPart(1.5), minimumContinueLabelY)

  self.tipCard:insertAtLocation(continueLabel, self.tipCard:getWidthPart(0.5), continueLabelY)

  local marginYTipLabelToTextArea = marginY * 0.65

  local rationale = wordData.rationale
  local textWidth = self.tipCard:getWidth() - (marginX * 2)
  local textArea = DynamicTextArea.new(rationale, textWidth, BASE_ATTRIBUTES, ATTRIBUTES_FOR_LABELS, Alignment.LEFT, Alignment.TOP)
  textArea:setColor(unpack(COLOR_TIP))
  textArea:setAnchorPointTop()
  textArea:setLoc(self.tipCard:getWidthPart(0.5), tipLabel:getYBottom() - marginYTipLabelToTextArea)
  self.tipCard:insert(textArea)
end

local function createLayoutExpanded(self, wordData)
  local tipCardHeight = self:getHeight()
  self.tipCard = FilledRectangle.new(self:getWidth(), tipCardHeight, COLOR_TIP_CARD)
  self:insert(self.tipCard, Z_INDEX.CARD)
  self.tipCard:setAnchorPointTop()

  local tipCardHiddenY = -self.tipCard:getHeightPart(0.5)
  self.tipCard:setLoc(self.tipCard:getWidthPart(0.5), tipCardHiddenY)

  local continueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 15))
  continueLabel:setColor(unpack(COLOR_TIP))
  continueLabel:setAnchorPointBottom()

  local minimumContinueLabelY = Screen.safeBottom + continueLabel:getHeight()
  local continueLabelY = math.max(continueLabel:getHeightPart(1.5), minimumContinueLabelY)

  self.tipCard:insertAtLocation(continueLabel, self.tipCard:getWidthPart(0.5), self.tipCard:getHeightPart(0.5) + continueLabelY)

  local marginY = 24
  local minimumX = Screen.safeLeft
  local marginX = math.max(self.tipCard:getWidthPart(0.1), minimumX)

  local baseFont = Font.new(FontFaces.DIN_OT_MEDIUM, 24)

  local baseAttributes = {
    font = baseFont
  }

  local attributesForLabels = {
    b = { font = Font.new(FontFaces.DIN_OT_BOLD, 24) },
     --offset needed because the medium italic font file oddly has a higher baseline y
    i = { font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, 24), yOffset = -1 }
  }

  local rationale = wordData.rationale
  local textWidth = self.tipCard:getWidth() - (marginX * 2)
  local textArea = DynamicTextArea.new(rationale, textWidth, baseAttributes, attributesForLabels, Alignment.LEFT, Alignment.TOP)
  textArea:setColor(unpack(COLOR_TIP))
  textArea:setAnchorPointTop()
  textArea:setLoc(self.tipCard:getWidthPart(0.5), self.tipCard:getHeight() - marginY)
  self.tipCard:insert(textArea)

  local learnMorePressedCallback = getLearnMorePressedCallback(self)
  local learnMoreButton = ChatRationaleLearnMoreButton.new(COLOR_TIP_CARD, learnMorePressedCallback)
  learnMoreButton:setAnchorPointTopLeft()
  self.tipCard:insertAtLocation(learnMoreButton, textArea:getXLeft(), textArea:getYBottom() - marginY)
  learnMoreButton:setTouchEnabled(false)
  self.learnMoreButton = learnMoreButton

  local chatRationaleClosedCallback = getChatRationaleClosedCallback(self)
  local chatRationale = ChatRationale.new(COLOR_TIP_CARD, chatRationaleClosedCallback)
  Screen.rootGroup:insertAtLocation(chatRationale, Screen.getWidthPart(0.5), 0)
  chatRationale:bringToFront()
  self.chatRationale = chatRationale

  chatRationale:setConversationDescriptor(CastlesChatRationaleHelper.createConversation(wordData))
end

-- Constructor ------------------------

function CastlesTip.new(wordData)
  local self = Group.newFullScreen()
  Mixin.mix(self, CastlesTip)

  -- This can probably be removed once all the concepts have extended rationale
  -- support and that support is guaranteed through additions to the validator
  local conceptSupportsExtendedRationale = (((wordData.etymology ~= nil) and (string.len(wordData.etymology) > 0)) and
                                            ((wordData.exampleSentence ~= nil) and (string.len(wordData.exampleSentence) > 0)) and
                                            ((wordData.pronunciation ~= nil) and (string.len(wordData.pronunciation) > 0)))

  if (ExperimentsConfiguration.EXPANDED_RATIONALES_EXPERIMENT:isActive() and conceptSupportsExtendedRationale) then
    createLayoutExpanded(self, wordData)
  else
    createLayout(self, wordData)
  end

  self.wordData = wordData

  return self
end
