--- FlickInteractionCoachUI
-- @author created by Jay Austria on January 26, 2015
--
-- A basic CoachUI that implements some standard animations
-- expected by FlickInteractionLayers for use with games that
-- have swipe interactions
--
-- Games can extend the FlickInteractionCoachUI with more game specific
-- behaviors and animations as necessary

class("FlickInteractionCoachUI")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LABEL_FONT_SIZE = 13
local COACH_UI_SCALE = 0.6
local NEUTRAL_SCALE = 1.0
local HIGHLIGHT_SCALE = 1.05
local LABEL_Y_OFFSET = 15
local FADE_DURATION = 0.15
local SCALE_DURATION = 0.2
local STARTING_VISIBLE_OPACITY = 0.35
local HIGHLIGHT_VISIBLE_OPACITY = 0.5
local FADED_VISIBLE_OPACITY = 0.08
local INVISIBLE_OPACITY = 0.0


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function showNeutralInternal(self)
  local fadeAction = self:getFadeOpacityToAction(STARTING_VISIBLE_OPACITY, FADE_DURATION)
  local scaleAction = AHScaleTo.new(self, SCALE_DURATION, nil, NEUTRAL_SCALE)

  local showAction = AHSpawn.new(fadeAction, scaleAction)
  self.currentRunningAction:insert(showAction)
end

local function createLabel(self, labelString, fontName)
  local text = labelString
  local fontSize = LABEL_FONT_SIZE

  local label = DeprecatedLabel.newSingleLine(text, fontName, fontSize)

  local width, height = self:getEffectiveDims()

  label:setLoc(width * 0.5, height + LABEL_Y_OFFSET)

  self:insert(label)
end

local function createLayout(self, spriteName, labelString, fontName, color)
  local iconSprite = Sprite.newWithName(spriteName)
  iconSprite:setScl(COACH_UI_SCALE)

  self:setDims(iconSprite:getEffectiveDims())
  self:setAnchorPointCenter()

  self:insertAtCenter(iconSprite)
  self.iconSprite = iconSprite

  createLabel(self, labelString, fontName)

  local isRecursive = true
  self:linkRawColorAttrToChildren(isRecursive)

  self:setColor(unpack(color))

  return self
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------
function FlickInteractionCoachUI:getFadeOpacityToAction(opacity, duration)
  duration = duration or FADE_DURATION

  local fadeAction = AHOpacityTo.new(self, duration, nil, opacity)

  return fadeAction
end

function FlickInteractionCoachUI:show()
  local currentOpacity = self:getOpacity()
  self:setColor(unpack(self.initialColor))
  self:setOpacity(currentOpacity)
  self:setScl(NEUTRAL_SCALE)

  showNeutralInternal(self)
end

function FlickInteractionCoachUI:returnToNeutral()
  showNeutralInternal(self)
end

function FlickInteractionCoachUI:highlight()
  local fadeAction = self:getFadeOpacityToAction(HIGHLIGHT_VISIBLE_OPACITY, FADE_DURATION)
  local scaleAction = AHScaleTo.new(self, SCALE_DURATION, AHEaseBackOut.new(7.0), HIGHLIGHT_SCALE)

  local highlightAction = AHSpawn.new(fadeAction, scaleAction)
  self.currentRunningAction:insert(highlightAction)
end

function FlickInteractionCoachUI:fadeOut()
  local fadeAction = self:getFadeOpacityToAction(FADED_VISIBLE_OPACITY, FADE_DURATION)

  self.currentRunningAction:insert(fadeAction)
end

function FlickInteractionCoachUI:hide()
  local fadeAction = self:getFadeOpacityToAction(INVISIBLE_OPACITY, FADE_DURATION)
  local resetColor = function()
    local currentOpacity = self:getOpacity()
    self:setColor(unpack(self.initialColor))
    self:setOpacity(currentOpacity)
  end

  local hideAction = AHSequence.new(fadeAction, resetColor)

  self.currentRunningAction:insert(hideAction)
end

function FlickInteractionCoachUI:getStartingVisibleOpacity()
  return self.startingVisibleOpacity
end

function FlickInteractionCoachUI:getHighlightVisibleOpacity()
  return self.highlightVisibleOpacity
end

function FlickInteractionCoachUI:getFadedVisibleOpacity()
  return self.fadedVisibleOpacity
end

function FlickInteractionCoachUI:getNeutralScale()
  return self.neutralScale
end

function FlickInteractionCoachUI:getHighlightScale()
  return self.highlightScale
end

function FlickInteractionCoachUI:getInitialColor()
  return self.initialColor
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function FlickInteractionCoachUI.new(spriteName, labelString, fontName, color)
  local self = Group.newDeprecated()
  Mixin.mix(self, FlickInteractionCoachUI)

  createLayout(self, spriteName, labelString, fontName, color)

  self.initialColor = color
  self.startingVisibleOpacity = STARTING_VISIBLE_OPACITY
  self.highlightVisibleOpacity = HIGHLIGHT_VISIBLE_OPACITY
  self.fadedVisibleOpacity = FADED_VISIBLE_OPACITY
  self.neutralScale = NEUTRAL_SCALE
  self.highlightScale = HIGHLIGHT_SCALE
  self.currentRunningAction = ActionInterrupt.new()

  return self
end
