--- FlickInteractionHandler
-- @author Created by Jay Austria on January 20, 2015
--
-- A group that inserts itself at the center of the
-- rootGroup and captures and interprets touch events
-- for games that utilize the swipe interaction.
--
-- Has a bunch of callbacks that can be hooked into:
--
--  When a swipe interaction ends in a particular region
--    setFlickEndsInLeftRegionCallback()
--    setFlickEndsInRightRegionCallback()
--    setFlickEndsInMiddleRegionCallback()
--
-- When a swipe causes the target object to leave a region
--    setLeaveLeftRegionCallback()
--    setLeaveMiddleRegionCallback()
--    setLeaveRightRegionCallback()
--
-- When a swipe causes the target object to enter a region
--    setEnterLeftRegionCallback()
--    setEnterRightRegionCallback()
--    setEnterMiddleRegionCallback()
--
-- When the swipe starts/ends
--    setFlickInteractionEndCallback()
--    setFlickInteractionBeginCallback()
--
-- When the thrown target object is detected to be off screen
--    setTargetOffScreenCallback()
--
-- In addition, the target object that gets assigned focus
-- must implement the following methods:
--   notifyFlickInteractionBegin()
--   notifyFlickInteractionEnd()

class("FlickInteractionHandler")

require("CastlesStateMachine")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local function vector2D(coordinateX, coordinateY)
  return {
    x = coordinateX,
    y = coordinateY
  }
end

local TOUCHES_QUEUE_LENGTH = 4
local SWIPE_DISTANCE_THRESHOLD = 20
local TARGET_RETURN_TO_ORIGIN_DURATION = 0.5
local TARGET_MOVE_OFFSCREEN_DURATION = 0.5

local LEFT_EDGE_PULL_VECTOR = vector2D(-Screen.width, 0)
local RIGHT_EDGE_PULL_VECTOR = vector2D(Screen.width, 0)

local ORIGINAL_WEIGHT = 0.9
local EDGE_WEIGHT = 0.1
local THROW_MAGNITUDE = Screen.width

local RegionState =
{
  LEFT = "RegionState.LEFT",
  MIDDLE = "RegionState.MIDDLE",
  RIGHT = "RegionState.RIGHT"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function callCallbackIfExists(callback)
  if callback then
    callback()
  end
end

local function moveTargetObject(self, targetObject, eventX, eventY)
  local targetObjectX = eventX + self.dragDelta.x
  local targetObjectY = eventY + self.dragDelta.y

  targetObject:setLoc(targetObjectX, targetObjectY)
end

local function setTargetObjectDragDelta(self, targetObject, eventX, eventY)
  local targetObjectX, targetObjectY = targetObject:getLoc()
  local deltaX = targetObjectX - eventX
  local deltaY = targetObjectY - eventY

  self.dragDelta = vector2D(deltaX, deltaY)
end

local function beginFlickInteraction(self, eventX, eventY)
  local targetObject = self.targetObject
  if not targetObject then
    return
  end

  targetObject:notifyFlickInteractionBegin()
  setTargetObjectDragDelta(self, targetObject, eventX, eventY)

  self.regionStateMachine:setCurrentState(RegionState.MIDDLE)

  callCallbackIfExists(self.flickInteractionBeginCallback)

  self.didStartWithBeginInteraction = true
end

local function insertTouchIntoTouchesQueue(self, eventX, eventY)
  local touchPair = vector2D(eventX, eventY)

  table.insert(self.touchesQueue, touchPair)

  if (#self.touchesQueue > TOUCHES_QUEUE_LENGTH) then
    table.remove(self.touchesQueue, 1)
   end
end

local function updateRegionState(self, eventX)
  local deltaX = eventX - self.flickOrigin.x + self.dragDelta.x

  local isInLeftRegion = (deltaX + SWIPE_DISTANCE_THRESHOLD < 0)
  local isInRightRegion = (SWIPE_DISTANCE_THRESHOLD - deltaX < 0)

  if isInLeftRegion then
    self.regionStateMachine:transition(RegionState.LEFT)
  elseif isInRightRegion then
    self.regionStateMachine:transition(RegionState.RIGHT)
  else
    self.regionStateMachine:transition(RegionState.MIDDLE)
  end
end

local function dragFlickInteraction(self, eventX, eventY)
  local targetObject = self.targetObject
  if not targetObject or not self.didStartWithBeginInteraction then
    return
  end

  moveTargetObject(self, targetObject, eventX, eventY)
  insertTouchIntoTouchesQueue(self, eventX, eventY)
  updateRegionState(self, eventX)
end

local function isAtScreenEdge(object)
  local x, y = object:getWorldLoc()
  local isAtEdge = x < 0 or x > Screen.width or
                   y < 0 or y > Screen.height

  return isAtEdge
end

local function startRunningOffScreenTrackingCoroutine(self, target)
  local trackingFunction = function()
    repeat
      coroutine.yield()
    until(isAtScreenEdge(target))

    if self.targetOffScreenCallback then
      self.targetOffScreenCallback(target)
    end
  end

  local trackingCoroutine = runInCoroutine(trackingFunction)
  self.trackingCoroutine = trackingCoroutine

  return trackingCoroutine
end

local function getAverageVelocityVector(self)
  local numTouches = #self.touchesQueue
  local sumX = 0
  local sumY = 0
  local numberOfVelocities = numTouches - 1
  for i = 1, numberOfVelocities do
    local velocityX = self.touchesQueue[i + 1].x - self.touchesQueue[i].x
    local velocityY = self.touchesQueue[i + 1].y - self.touchesQueue[i].y

    sumX = sumX + velocityX
    sumY = sumY + velocityY
  end

  local averageX = sumX / numberOfVelocities
  local averageY = sumY / numberOfVelocities

  local averageVelocityVector = vector2D(averageX, averageY)
  return averageVelocityVector
end

local function throwTargetObjectOffscreen(self, targetObject, edgePullVector)
  local averageVelocityVector = getAverageVelocityVector(self)

  local velocityVectorX = (ORIGINAL_WEIGHT * averageVelocityVector.x) + (EDGE_WEIGHT * edgePullVector.x)
  local velocityVectorY = (ORIGINAL_WEIGHT * averageVelocityVector.y) + (EDGE_WEIGHT * edgePullVector.y)

  local magnitude = MathUtils.vectorLength(velocityVectorX, velocityVectorY)
  local xNormalized = velocityVectorX / magnitude
  local yNormalized = velocityVectorY / magnitude

  local xEnd = xNormalized * THROW_MAGNITUDE
  local yEnd = yNormalized * THROW_MAGNITUDE

  local moveAction = AHLocationBy.new(targetObject, TARGET_MOVE_OFFSCREEN_DURATION, nil, xEnd, yEnd)
  local removeSelfAction = function() targetObject:removeSelf() end

  startRunningOffScreenTrackingCoroutine(self, targetObject)

  local moveButtonSequence = AHSequence.new(moveAction, removeSelfAction)
  moveButtonSequence:start()
end

local function returnTargetObjectToOrigin(self, targetObject)
  local flickOriginX = self.flickOrigin.x
  local flickOriginY = self.flickOrigin.y

  local moveAction = AHLocationTo.new(targetObject, TARGET_RETURN_TO_ORIGIN_DURATION, AHEaseSimpleOut.new(2.0), flickOriginX, flickOriginY)
  self.returnTargetToOriginInterrupt:insert(moveAction)
end

local function endFlickInteraction(self, eventX, eventY)
  local targetObject = self.targetObject
  if not targetObject or not self.didStartWithBeginInteraction then
    return
  end

  moveTargetObject(self, targetObject, eventX, eventY)

  targetObject:notifyFlickInteractionEnd()

  local finalRegion = self.regionStateMachine:getCurrentState()
  if (finalRegion == RegionState.LEFT) then
    throwTargetObjectOffscreen(self, targetObject, LEFT_EDGE_PULL_VECTOR)
    callCallbackIfExists(self.flickEndsInLeftRegionCallback)
  elseif (finalRegion == RegionState.RIGHT) then
    throwTargetObjectOffscreen(self, targetObject, RIGHT_EDGE_PULL_VECTOR)
    callCallbackIfExists(self.flickEndsInRightRegionCallback)
  else
    returnTargetObjectToOrigin(self, targetObject)
    callCallbackIfExists(self.flickEndsInMiddleRegionCallback)
  end

  callCallbackIfExists(self.flickInteractionEndCallback)

  self.didStartWithBeginInteraction = false
end

local function getTouchHandler(self)
  return function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      return beginFlickInteraction(self, event.x, event.y)
    elseif (event.phase == TouchManager.TOUCH_MOVED) then
      return dragFlickInteraction(self, event.x, event.y)
    elseif (event.phase == TouchManager.TOUCH_ENDED) then
      return endFlickInteraction(self, event.x, event.y)
    end
  end
end

local function createRegionStates(self)
  -- Left region ----------------------
  local function onEnterLeftRegion()
    callCallbackIfExists(self.onEnterLeftRegionCallback)
  end

  local function onLeaveLeftRegion()
    callCallbackIfExists(self.onLeaveLeftRegionCallback)
  end

  self.regionStateMachine:addState(RegionState.LEFT, onEnterLeftRegion, onLeaveLeftRegion)

  -- Right region ---------------------
  local function onEnterRightRegion()
    callCallbackIfExists(self.onEnterRightRegionCallback)
  end

  local function onLeaveRightRegion()
    callCallbackIfExists(self.onLeaveRightRegionCallback)
  end

  self.regionStateMachine:addState(RegionState.RIGHT, onEnterRightRegion, onLeaveRightRegion)

  -- Middle region --------------------
  local function onEnterMiddleRegion()
    callCallbackIfExists(self.onEnterMiddleRegionCallback)
  end

  local function onLeaveMiddleRegion()
    callCallbackIfExists(self.onLeaveMiddleRegionCallback)
  end

  self.regionStateMachine:addState(RegionState.MIDDLE, onEnterMiddleRegion, onLeaveMiddleRegion)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FlickInteractionHandler:setEnabled(enabled)
  if (enabled) then
    local touchHandler = self.touchHandler
    TouchManager.addTouchHandler(self, touchHandler)
  else
    TouchManager.removeTouchHandler(self)
  end
end

function FlickInteractionHandler:setTargetObject(targetObject)
  self:clearTargetObject()

  self.targetObject = targetObject

  local objectX, objectY = targetObject:getLoc()

  local flickOrigin = vector2D(objectX, objectY)
  self.flickOrigin = flickOrigin
end

function FlickInteractionHandler:clearTargetObject()
  if self.trackingCoroutine then
    self.trackingCoroutine:stop()
    self.trackingCoroutine = nil
  end

  self.targetObject = nil
  self.flickOrigin = nil
end

function FlickInteractionHandler:setFlickEndsInLeftRegionCallback(callback)
  self.flickEndsInLeftRegionCallback = callback
end

function FlickInteractionHandler:setFlickEndsInRightRegionCallback(callback)
  self.flickEndsInRightRegionCallback = callback
end

function FlickInteractionHandler:setFlickEndsInMiddleRegionCallback(callback)
  self.flickEndsInMiddleRegionCallback = callback
end

function FlickInteractionHandler:setEnterLeftRegionCallback(callback)
  self.onEnterLeftRegionCallback = callback
end

function FlickInteractionHandler:setLeaveLeftRegionCallback(callback)
  self.onLeaveLeftRegionCallback = callback
end

function FlickInteractionHandler:setEnterRightRegionCallback(callback)
  self.onEnterRightRegionCallback = callback
end

function FlickInteractionHandler:setLeaveRightRegionCallback(callback)
  self.onLeaveRightRegionCallback = callback
end

function FlickInteractionHandler:setEnterMiddleRegionCallback(callback)
  self.onEnterMiddleRegionCallback = callback
end

function FlickInteractionHandler:setLeaveMiddleRegionCallback(callback)
  self.onLeaveMiddleRegionCallback = callback
end

function FlickInteractionHandler:setFlickInteractionEndCallback(callback)
  self.flickInteractionEndCallback = callback
end

function FlickInteractionHandler:setFlickInteractionBeginCallback(callback)
  self.flickInteractionBeginCallback = callback
end

function FlickInteractionHandler:setTargetOffScreenCallback(callback)
  self.targetOffScreenCallback = callback
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function FlickInteractionHandler.new()
  local self = Group.newFullScreen()
  Screen.rootGroup:insertAtCenter(self)

  Mixin.mix(self, FlickInteractionHandler)

  self.dragDelta = nil
  self.didStartWithBeginInteraction = false
  self.targetObject = nil
  self.flickOrigin = nil
  self.touchesQueue = {}
  self.regionStateMachine = CastlesStateMachine.new()
  self.returnTargetToOriginInterrupt = ActionInterrupt.new()

  createRegionStates(self)

  self.touchHandler = getTouchHandler(self)
  self:setEnabled(true)

  return self
end
