--- FlickInteractionLayer
-- @author created by Jay Austria on January 21, 2015
--
-- A basic layer that can be inserted into a scene and will have basic
-- flick interaction behavior. It creates an input.FlickInteractionHandler
-- and hooks a couple edge glow UIs and provided coachUI into the
-- FlickInteractionHandler's callbacks
--
-- Games can extend the FlickInteractionLayer with more game specific
-- behaviors and animations as necessary

class("FlickInteractionLayer")

require("FlickInteractionHandler")
require("FlickInteractionHelper")
require("Particles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local EDGE_SPRITE_NAME = "shared_glow_side.png"
local SCALE_X_REGULAR = 1
local SCALE_X_FLIPPED = -1
local INVISIBLE_OPACITY = 0.0
local VISIBLE_OPACITY = 1.0
local EDGE_FADE_SPEED = 0.15
local PARTICLES_SURGE_COUNT = 128
local PARTICLES_FILE_NAME = "shared_flick_particles_edge.pex"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function callCallbackIfExists(callback)
  if callback then
    callback()
  end
end

local function insertCoachUIs(self, coachUIs)
  for _, coachUI in ipairs(coachUIs) do
    self:insert(coachUI)
  end
end

local function showInteractionHelper(self)
  self.interactionHelper:show()
end

local function hideInteractionHelper(self)
  self.interactionHelper:hide()
end

local function getFlagTypeIndex(self, flagType)
  local flagTypes = self.flagTypes
  local typeIndex = TableUtils.indexOf(flagTypes, flagType)

  return typeIndex
end

local function getFadeEdgeOpacityAnimation(edge, opacity)
  local fadeAction = AHOpacityTo.new(edge, EDGE_FADE_SPEED, nil, opacity)
  edge.fadeInterrupt:insert(fadeAction)

  return fadeAction
end

local function fadeEdgeOut(self, edge)
  local fadeAction = self:getFadeEdgeOutAnimation(edge)
  fadeAction:start()
end

local function fadeEdgeIn(self, edge)
  local fadeAction = self:getFadeEdgeInAnimation(edge)
  fadeAction:start()
end

local function createEdgeGlowHidden(descriptor)
  local edgeGlow = Sprite.newWithName(EDGE_SPRITE_NAME)
  local scaleYToScreenHeight = Screen.height / edgeGlow:getHeight()
  edgeGlow:setScl(descriptor.scaleX, scaleYToScreenHeight)

  local edgeX = descriptor.x
  local edgeY = Screen.getHeightPart(0.5)
  edgeGlow:setLoc(edgeX, edgeY)

  edgeGlow:setOpacity(INVISIBLE_OPACITY)

  edgeGlow.fadeInterrupt = ActionInterrupt.new()

  return edgeGlow
end

local function createEdgeGlows(self, flagTypes)
  local dummyEdgeSprite = Sprite.newWithName(EDGE_SPRITE_NAME)

  local edgeGlowDescriptors =
  {
    {
      x = dummyEdgeSprite:getWidth() * 0.5,
      scaleX = SCALE_X_REGULAR,
      flagType = flagTypes[1]
    },
    {
      x = Screen.width - dummyEdgeSprite:getWidth() * 0.5,
      scaleX = SCALE_X_FLIPPED,
      flagType = flagTypes[2]
    }
  }

  local edgeGlows = {}
  for _, edgeGlowDescriptor in ipairs(edgeGlowDescriptors) do
    local edgeGlow = createEdgeGlowHidden(edgeGlowDescriptor)

    self:insert(edgeGlow)

    table.insert(edgeGlows, edgeGlow)
  end
  self.edgeGlows = edgeGlows
end

local function emitParticles(self, x, y)
  local offscreenParticles = self.offscreenParticles
  offscreenParticles:setSourcePosition(x, y)
  offscreenParticles:surge(PARTICLES_SURGE_COUNT)
end

local function createEventCallbacks(self, flagTypes, interactionHandler)
  local leftFlagType = flagTypes[1]
  local rightFlagType = flagTypes[2]

  local leftCoachUI = self:getCoachUIForFlagType(leftFlagType)
  local rightCoachUI = self:getCoachUIForFlagType(rightFlagType)

  local leftEdgeGlow = self:getEdgeGlowForFlagType(leftFlagType)
  local rightEdgeGlow = self:getEdgeGlowForFlagType(rightFlagType)

  interactionHandler:setFlickEndsInLeftRegionCallback(function()
    if self.isFirstInteraction then
      hideInteractionHelper(self)
    end

    self.isFirstInteraction = false

    callCallbackIfExists(self.flagLeftCallback)
  end)

  interactionHandler:setFlickEndsInRightRegionCallback(function()
    if self.isFirstInteraction then
      hideInteractionHelper(self)
    end

    self.isFirstInteraction = false

    callCallbackIfExists(self.flagRightCallback)
  end)

  interactionHandler:setEnterLeftRegionCallback(function()
    fadeEdgeIn(self, leftEdgeGlow)
    leftCoachUI:highlight()
    rightCoachUI:fadeOut()
  end)

  interactionHandler:setLeaveLeftRegionCallback(function()
    fadeEdgeOut(self, leftEdgeGlow)

    for _, coachUI in pairs(self.coachUIs) do
      coachUI:returnToNeutral()
    end
  end)

  interactionHandler:setEnterRightRegionCallback(function()
    fadeEdgeIn(self, rightEdgeGlow)
    leftCoachUI:fadeOut()
    rightCoachUI:highlight()
  end)

  interactionHandler:setLeaveRightRegionCallback(function()
    fadeEdgeOut(self, rightEdgeGlow)

    for _, coachUI in pairs(self.coachUIs) do
      coachUI:returnToNeutral()
    end
  end)

  interactionHandler:setEnterMiddleRegionCallback(function()
  end)

  interactionHandler:setLeaveMiddleRegionCallback(function()
  end)

  interactionHandler:setFlickInteractionEndCallback(function()
  end)

  interactionHandler:setFlickInteractionBeginCallback(function()
    for _, coachUI in pairs(self.coachUIs) do
      coachUI:show()
    end
  end)

  interactionHandler:setTargetOffScreenCallback(function(target)
    if self.emitParticlesOnExit then
      local x, y = target:getWorldLoc()

      -- getWorldLoc seems to place the particles at the bottom of the target.  Center it.
      y = y + target:getHeightPart(0.5)

      emitParticles(self, x, y)
    end
  end)
end

local function createOffscreenParticles(self)
  local offscreenParticles = Particles.new(PARTICLES_FILE_NAME)
  offscreenParticles:setLoc(0.0, 0.0)
  self:insert(offscreenParticles)
  offscreenParticles:start()
  self.offscreenParticles = offscreenParticles

  self.emitParticlesOnExit = false
end

local function getCombinedResponseAnimation(self, flagType, coachUIAnimation)
  local edgeGlow = self:getEdgeGlowForFlagType(flagType)
  local edgeFadeAction = self:getFadeEdgeOutAnimation(edgeGlow)

  local coachUIs = self.coachUIs
  local resetCoachMarks = function()
    for _, coachUI in ipairs(coachUIs) do
      coachUI:returnToNeutral()
    end
  end

  local combinedFadeOutAction = AHSpawn.new(edgeFadeAction, resetCoachMarks)

  local combinedResponseAnimation = AHSequence.new()
  if (coachUIAnimation ~= nil) then
    combinedResponseAnimation:addAction(coachUIAnimation)
  end
  combinedResponseAnimation:addAction(combinedFadeOutAction)

  return combinedResponseAnimation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------
function FlickInteractionLayer:getCoachUIForFlagType(flagType)
  local coachUIs = self.coachUIs
  local flagTypeIndex = getFlagTypeIndex(self, flagType)
  local requestedCoachUI = coachUIs[flagTypeIndex]

  assert(requestedCoachUI, "Couldn't find a coachUI for flagType")
  return requestedCoachUI
end

function FlickInteractionLayer:setEnabled(isEnabled)
  local interactionHandler = self.interactionHandler
  interactionHandler:setEnabled(isEnabled)
  if self.isFirstInteraction and isEnabled then
    showInteractionHelper(self)
  end
end

function FlickInteractionLayer:setTargetObject(target)
  local interactionHandler = self.interactionHandler
  interactionHandler:setTargetObject(target)
end

function FlickInteractionLayer:setFlagLeftCallback(callback)
  self.flagLeftCallback = callback
end

function FlickInteractionLayer:setFlagRightCallback(callback)
  self.flagRightCallback = callback
end

function FlickInteractionLayer:getEdgeGlowForFlagType(flagType)
  local edgeGlows = self.edgeGlows
  local flagTypeIndex = getFlagTypeIndex(self, flagType)
  local requestedEdgeGlow = edgeGlows[flagTypeIndex]

  assert(requestedEdgeGlow, "Couldn't find an edge glow for flagType")
  return requestedEdgeGlow
end

function FlickInteractionLayer:getFadeEdgeInAnimation(edge)
  return getFadeEdgeOpacityAnimation(edge, VISIBLE_OPACITY)
end

function FlickInteractionLayer:getFadeEdgeOutAnimation(edge)
  return getFadeEdgeOpacityAnimation(edge, INVISIBLE_OPACITY)
end

function FlickInteractionLayer:getIncorrectAnimation(flagType)
  local combinedResponseAnimation = getCombinedResponseAnimation(self, flagType)
  return combinedResponseAnimation
end

function FlickInteractionLayer:getCorrectAnimation(flagType)
  local coachUIToAnimate = self:getCoachUIForFlagType(flagType)
  local coachUICorrectAnimation = coachUIToAnimate:getCorrectAnimation()

  local combinedResponseAnimation = getCombinedResponseAnimation(self, flagType, coachUICorrectAnimation)
  return combinedResponseAnimation
end

function FlickInteractionLayer:setEnableEmitParticlesOnExit(isEnabled)
  self.emitParticlesOnExit = isEnabled
end

function FlickInteractionLayer:hideCoachUI()
  local coachUIs = self.coachUIs
  for _, coachUI in pairs(coachUIs) do
    coachUI:hide()
  end
end

function FlickInteractionLayer:showCoachUI()
  local coachUIs = self.coachUIs
  for _, coachUI in pairs(coachUIs) do
    coachUI:returnToNeutral()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function FlickInteractionLayer.new(flagTypes, coachUIs)
  assert(#flagTypes == 2, "FlickInteractionLayer could not be made with the given number of flagTypes")
  assert(#coachUIs == 2, "FlickInteractionLayer could not be made with the given number of coachUIs")

  local self = Group.newFullScreen()
  Mixin.mix(self, FlickInteractionLayer)

  local interactionHandler = FlickInteractionHandler.new()
  self.interactionHandler = interactionHandler

  self.coachUIs = coachUIs
  self.flagTypes = flagTypes

  insertCoachUIs(self, coachUIs)
  createEdgeGlows(self, flagTypes)
  createEventCallbacks(self, flagTypes, interactionHandler)
  createOffscreenParticles(self)

  self.isFirstInteraction = true
  self.interactionHelper = FlickInteractionHelper.new()
  self:insertAtCenter(self.interactionHelper)

  return self
end
