--- ClarityGame
-- @author Created by Jay Austria on February 17, 2014.

class("ClarityGame")
require("bonuses.BonusNotificationTray")
require("bonuses.ResponseSpeedTracker")
require("widgets.progress.BarProgressIndicator")
require("ClarityCommon")
require("ClarityConfiguration")
require("ClarityData")
require("ClarityRoundGenerator")
require("ClarityHelpLayer")
require("ClarityInterstitialLabel")
require("ClarityReplacementMenu")
require("ClarityScene")
require("StandardUIBuilder")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INSTRUCTIONS_FADE_TIME = 0.3

local INTERSTITIAL_X = Screen.getSafeX(0.5)
local INTERSTITIAL_Y = Screen.getSafeY(0.88)

local UNLOCK_DELAY_AFTER_CONCEPT_CORRECT = 2.0
local UNLOCK_DELAY_AFTER_RATIONALE_SHOWN = 0.5

local DO_INTRO_DELAY = 0.5

local INTERSTITIAL_MESSAGE_LAST = Localizer.get("INTERSTITIAL_MESSAGE_LAST")
local INTERSTITIAL_MESSAGE_HALFWAY_TEMPLATE = Localizer.get("INTERSTITIAL_MESSAGE_HALFWAY_TEMPLATE")
local INTERSTITIAL_BLINK_DELAY = 0.15

local UI_COMPONENTS_COLOR = { 0.3, 0.3, 0.3, 0.75 }

local function configureGameDifficulty(self)
  self.correctNeededToWin = ClarityConfiguration.getCorrectAnswersToWinCount()

  local passScore = ClarityConfiguration.getPassScore()
  local rankTwoScore = ClarityConfiguration.getRankTwoScore()
  local rankThreeScore = ClarityConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = ClarityConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(ClarityConfiguration.NUMBER_OF_LIVES)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function allowMistakesToResurfaceIfNeeded(self)
  local conceptCountForResurfacing = math.floor(self.correctNeededToWin *
                                                LearningFocusedRoundGenerator.DEFAULT_REMAINING_BEFORE_RESURFACING)

  if (self.conceptsRemaining <= conceptCountForResurfacing) then
    self.roundGenerator:allowMistakesToResurface()
  end
end

local function didWin(self)
  local isOutOfConcepts = (self.conceptsRemaining == 0)
  local numberOfLives = self.livesCounter:getLives()
  local hasLivesRemaining = (numberOfLives > 0)

  return (isOutOfConcepts and hasLivesRemaining)
end

local function didLose(self)
  local numberOfLives = self.livesCounter:getLives()

  return numberOfLives == 0
end

local function isFlowEnded(self)
  return didWin(self) or didLose(self)
end

local function getNextTangram(self)
  self.currentTangramIndex = self.currentTangramIndex + 1
  local tangram = self.tangramLayout[self.currentTangramIndex]
  return tangram
end

local function getCurrentTangram(self)
  local index = math.min(self.currentTangramIndex, #self.tangramLayout)
  local tangram = self.tangramLayout[index]
  return tangram
end

local function unlockAfterDelay(self, delay)
  local function delayedUnlock()
    waitFor(delay)
    if self.responseLock then
      self.responseLock:stop()
      self.responseLock = nil
    end
  end
  runInCoroutine(delayedUnlock)
end

local function onConceptCorrect(self)
  self.responseSpeedTracker:notifyCorrectResponse()

  Sound.playSoundEffect("correct")
  GameSession.increaseScore(ClarityConfiguration.getPointsForCorrectAnswer())

  local currentConcept = self.currentConcept
  local conceptID = currentConcept.identifier
  GameSession.notifyCorrectAnswer(conceptID)
  GameSession.incrementCorrectAnswerCount()

  unlockAfterDelay(self, UNLOCK_DELAY_AFTER_CONCEPT_CORRECT)

  self.conceptsRemaining = self.conceptsRemaining - 1
  self.nextTangram = getNextTangram(self)

  if didWin(self) then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), ClarityConfiguration.POINTS_PER_EXTRA_LIFE, self.bonusTray)
    GameSession.notifyGameTimeEnd()
    self.outroAction, self.colorChangeActions = self.gameScene:playOutroAnimations()
  end

  self.scoreCounter:setScore(GameSession.getScore())
end

local function onConceptIncorrect(self)
  self.responseSpeedTracker:notifyIncorrectResponse()

  Sound.playSoundEffect("incorrect")

  local currentConcept = self.currentConcept
  local conceptID = currentConcept.identifier
  self.roundGenerator:notifyMistake(conceptID)
  GameSession.notifyIncorrectAnswer(conceptID)
  GameSession.incrementIncorrectAnswerCount()

  self.livesCounter:decrement()

  local tangram = getCurrentTangram(self)
  local tangramColor = tangram.color

  local rationaleDismissedCallback = function() unlockAfterDelay(self, UNLOCK_DELAY_AFTER_RATIONALE_SHOWN) end
  self.gameScene:showRationale(currentConcept.rationale, tangram.color, rationaleDismissedCallback)
end

local function onSubstitutionResponse(self, response)
  -- The user tapped something. They know how to play!
  if (self.showHelpIfTimeout) then
    self.helperTimer:stop()
    self.showHelpIfTimeout = false
  end

  if (response == self.correctReplacement) then
    local isFirstCorrect = (self.currentTangramIndex == 1)
    local doBounce = not isFirstCorrect
    GameSession.contentTracker:notifyCorrectAnswer(self.currentConcept.identifier, self.questionIdentifier)
    self.gameScene:onSubstitutionCorrect(response, self.wordInteractingWith, self.nextTangram, doBounce)
    onConceptCorrect(self)
  else
    GameSession.contentTracker:notifyIncorrectAnswer(self.currentConcept.identifier, self.questionIdentifier)
    self.gameScene:onSubstitutionIncorrect(response, self.nextTangram)
    onConceptIncorrect(self)
  end
end

local function showSubstitutionHelper(self)
  self.canRespond = false

  local correctResponse = self.correctReplacement
  wait(self.gameScene:showSubstitutionHelpAndWaitForTap(correctResponse))

   -- simulate a correct tap after showing help
  self.canRespond = true
  onSubstitutionResponse(self, correctResponse)
end

local function onReducibleTapCorrect(self, word)
  self.wordInteractingWith = word
  GameSession.contentTracker:notifyCorrectAnswer(self.currentConcept.identifier, self.questionIdentifier)
  self.questionIdentifier = "replacement"
  self.gameScene:onReducibleTapCorrectWithReplacement(self.nextTangram)
  Sound.playSoundEffect("highlight")
end

local function onReducibleTapIncorrect(self, word)
  GameSession.contentTracker:notifyIncorrectAnswer(self.currentConcept.identifier, self.questionIdentifier, word:getText())
  self.gameScene:onReducibleTapIncorrect(word, self.nextTangram)
  onConceptIncorrect(self)
end

local function onReducibleResponse(self, word)
  if self.canRespond then
    self.canRespond = false

    -- The user tapped something. They know how to play!
    if (self.showHelpIfTimeout) then
      self.helperTimer:stop()
      self.showHelpIfTimeout = false
    end
    local textModel = self.gameScene:getModel()
    local wordItems = textModel:getItems()
    local wordItem = wordItems[word.wordIndex]
    if not (wordItem:getAttribute("isReducible")) then
      onReducibleTapIncorrect(self, word)
    else
      onReducibleTapCorrect(self, word)
    end
    local hideInstructionsAction = self.instructionsLabel:getHideAction()
    hideInstructionsAction:start()
  end
end

local function showReducibleHelper(self)
  self.canRespond = false
  local substitutionMessage = ClarityData.getReplacementInstructionsString()
  self.gameScene:addSubstitutionMessage(substitutionMessage)
  local highlightedWord = self.gameScene:showReducibleHelpAndWaitForTap()

   -- simulate a correct tap after showing help
  self.canRespond = true
  onReducibleResponse(self, highlightedWord)
  self.helperTimer = AHDelayed.new(
    ClarityConfiguration.TUTORIAL_DELAY,
    function()
      showSubstitutionHelper(self)
    end
  )
  self.showHelpIfTimeout = true
  self.helperTimer:start()
end

local function createWordsAndAllowTouch(self, concept)
  local function onResponseWrapper(word)
    onReducibleResponse(self, word)
  end

  self.gameScene:createWordsWithConcept(concept, onResponseWrapper)
end

local function initializeReplacements(self, concept)
  local function replacementWrapper(response)
    onSubstitutionResponse(self, response)
  end

  self.gameScene:initializeReplacements(concept, replacementWrapper)
end

local function initializeGame(self)
  self.tangramLayout = ClarityData.getTangramLayout(self.correctNeededToWin)
  self.nextTangram = getNextTangram(self)

  self.conceptsRemaining = self.correctNeededToWin
end

local function waitForIntro(self)
  local showInstructionsAction = self.instructionsLabel:getShowAction()
  local showSceneAction = self.gameScene:getShowAction()

  waitFor(DO_INTRO_DELAY)
  showInstructionsAction:start()
  waitFor(INSTRUCTIONS_FADE_TIME * 0.5)
  wait(showSceneAction:start())
end

local function createInterstitialLabel(self, message)
  local maxLabelWidth = Screen.getSafeWidthPart(0.9)
  local labelSafeArea = self.scoreCounter:getYBottom() - TOP_EDGE_HEIGHT
  local maxLabelHeight = labelSafeArea * 0.9
  local interstitialLabel = ClarityInterstitialLabel.newWithMessage(message, maxLabelWidth, maxLabelHeight)
  local labelLocationY = TOP_EDGE_HEIGHT + (labelSafeArea / 2)
  interstitialLabel:setLoc(INTERSTITIAL_X, labelLocationY)
  self:insert(interstitialLabel)
  return interstitialLabel
end

local function runInterstitialWithMessage(self, message)
  local interstitial = createInterstitialLabel(self, message)
  local wipeAction = interstitial:getFullAnimation()
  wait(wipeAction:start())
  interstitial:removeSelf()
end

local function waitForInterstitial(self)
  local isAtHalf = (self.conceptsRemaining == math.floor(self.correctNeededToWin / 2))
  local isAtLast = (self.conceptsRemaining == 1)

  if (isAtHalf and not self.hasShowedHalfInterstitial) then
    self.hasShowedHalfInterstitial = true
    local combinedString = string.format(INTERSTITIAL_MESSAGE_HALFWAY_TEMPLATE, self.conceptsRemaining)
    runInterstitialWithMessage(self, combinedString)
  elseif (isAtLast and not self.hasShowedLastInterstitial) then
    self.hasShowedLastInterstitial = true
    runInterstitialWithMessage(self, INTERSTITIAL_MESSAGE_LAST)
  end
end

local function startRoundWithConcept(self, concept)
  self.correctReplacement = concept.replacement

  createWordsAndAllowTouch(self, concept)
  initializeReplacements(self, concept)
  wait(self.gameScene:onStartRound())

  self.canRespond = true
  self.questionIdentifier = "reduction"

  self.responseSpeedTracker:notifyStart()
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
end

local function waitForResponse(self)
  self.responseLock = AHDelay.new(math.huge)
  wait(self.responseLock:start())
end

local function waitForRoundClean(self)
  local onCleanRoundCoroutine = self.gameScene:getOnCleanRoundCoroutine()

  -- Don't allow the game to move onto a new concept until the scene
  -- is able to finish its cool transition effect
  wait(onCleanRoundCoroutine:start())
end

local function onGameConcluded(self)
  if didWin(self) then
    wait(self.outroAction)
    self.colorChangeActions:stop()

    GameSession.notifyGameWin()
  else
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ClarityGame:gameFlow()
  initializeGame(self)

  waitForIntro(self)
  repeat
    waitForInterstitial(self)

    allowMistakesToResurfaceIfNeeded(self)
    self.currentConcept = self.roundGenerator:next()

    if self.showHelpIfTimeout then
      self.helperTimer = AHDelayed.new(
        ClarityConfiguration.TUTORIAL_DELAY,
        function()
          showReducibleHelper(self)
        end
      )
      self.helperTimer:start()
    end
    startRoundWithConcept(self, self.currentConcept)
    waitForResponse(self)
    if not didWin(self) then
      waitForRoundClean(self)
    end
  until isFlowEnded(self)

  onGameConcluded(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreCounter = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(ClarityConfiguration.NUMBER_OF_LIVES)

  local instructionsString = ClarityData.getInstructionsString()

  self.instructionsLabel = createInterstitialLabel(self, instructionsString)
end

local function createGameElements(self)
  local gameScene = ClarityScene.newHidden(self)
  self:insertAtCenter(gameScene)
  self.gameScene = gameScene

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 ClarityConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  local speedBonusWindowInSeconds = ClarityConfiguration.getSpeedBonusWindowInSeconds()
  self.responseSpeedTracker = ResponseSpeedTracker.new(speedBonusWindowInSeconds,
                                                       quickResponseWrapper,
                                                       ClarityConfiguration.SPEED_BONUS_DECAY_START)
end

function ClarityGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ClarityGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.roundGenerator = LearningFocusedRoundGenerator.new(ClarityRoundGenerator.new())
  self.hasShowedHalfInterstitial = false
  self.hasShowedLastInterstitial = false
  self.currentTangramIndex = 0
  self.showHelpIfTimeout = true

  return self
end
