--- ClarityReplacementMenu
-- @author Created by Jay Austria, FebruaANY 20, 2014

class("ClarityReplacementMenu")

require("ClarityColors")
require("ClarityCommon")
require("text.SizeConstrainedTextArea")


-- ----------------------------------------------
-- Configuration
-- ----------------------------------------------

local MENU_BUTTON_FONT = Font.new(FontFaces.DIN_OT_LIGHT, 20)

local GROUP_HEIGHT = Screen.getSafeHeightPart(0.38)
local GROUP_WIDTH = Screen.getWidthPart(1.0)

local MENU_FADE_DURATION = 0.4
local MENU_FADE_OUT_DURATION = 0.2
local MENU_FADE_OUT_DELAY = 0.2

local BUTTON_LOCATIONS =
{
  {
    x = GROUP_WIDTH * 0.5,
    y = GROUP_HEIGHT * 0.65
  },
  {
    x = GROUP_WIDTH * 0.5,
    y = GROUP_HEIGHT * 0.32
  }
}

local BUTTON_OFFSET_FOR_MESSAGE = GROUP_HEIGHT * 0.1
local BUTTON_COLOR = ClarityColors.DARK_PAYNES_GRAY
local BUTTON_LABEL_COLOR = ClarityColors.WHITE
local BUTTON_WIDTH = GROUP_WIDTH * 0.92
local BUTTON_HEIGHT = GROUP_HEIGHT * 0.25

local LABEL_PADDING_X = 4
local LABEL_PADDING_Y = 0

local SOFT_SHAKE_WIDTH_X = 300
local SOFT_SHAKE_DURATION_MULTIPLIER = 0.8


-- ----------------------------------------------
-- Utility
-- ----------------------------------------------

local function onButtonTapped(self, response)
  if (self.onButtonTappedCallback and self.canBeTapped) then
    self.onButtonTappedCallback(response)
  end
end

local function getReplacementsMenuOpacityAction(self, opacity)
  if (self.fadeAction) then
    self.fadeAction:stop()
  end

  self.fadeAction = AHOpacityTo.new(self, MENU_FADE_DURATION, nil, opacity)
  return self.fadeAction
end

local function positionButtonsWithOffset(self, offset)
  for i, button in ipairs(self.buttons) do
    button:setY(BUTTON_LOCATIONS[i].y + offset)
  end
end

local function addMessage(self, message)
  local maxHeight = GROUP_HEIGHT * 0.25
  local maxWidth = GROUP_WIDTH
  local baseAttributes = { font = Font.new(FontFaces.DEFAULT_FONT, 18), color = {1,1,1} }
  local label = SizeConstrainedTextArea.new(message, maxWidth, maxHeight, baseAttributes, nil, nil, Alignment.CENTER, Alignment.CENTER)
  local labelLocY = BUTTON_LOCATIONS[1].y + ((BUTTON_HEIGHT + maxHeight - BUTTON_OFFSET_FOR_MESSAGE) / 2)
  label:setLoc(GROUP_WIDTH / 2, labelLocY)
  label:inheritOpacity(self)
  self:insert(label)
  self.messageLabel = label
end

local function createButton(self)
  local newButton = FilledRectangle.new(BUTTON_WIDTH, BUTTON_HEIGHT, ClarityColors.BLACK, true)

  local insideFill = FilledRectangle.new(BUTTON_WIDTH - 1, BUTTON_HEIGHT - 1, BUTTON_COLOR)
  newButton:insertAtCenter(insideFill)
  newButton.insideFill = insideFill

  local buttonLabel = DynamicTextArea.newSimple("-unset-",
                                          BUTTON_WIDTH - LABEL_PADDING_X,
                                          MENU_BUTTON_FONT,
                                          Alignment.CENTER,
                                          Alignment.CENTER)
  buttonLabel:setColor(unpack(BUTTON_LABEL_COLOR))
  newButton:insertAtCenter(buttonLabel)
  newButton.label = buttonLabel

  return newButton
end

-- ----------------------------------------------
-- API
-- ----------------------------------------------

function ClarityReplacementMenu:addMessageLabel(message)
  if (self.messageLabel == nil) then
    positionButtonsWithOffset(self, -BUTTON_OFFSET_FOR_MESSAGE)
  else
    self.messageLabel:removeSelf()
  end
  addMessage(self, message)
end

function ClarityReplacementMenu:removeMessageLabel()
  if (self.messageLabel ~= nil) then
    positionButtonsWithOffset(self, 0)
    self.messageLabel:removeSelf()
    self.messageLabel = nil
  end
end

function ClarityReplacementMenu:isCurrentlyVisible()
  return self.isVisible
end

function ClarityReplacementMenu:lockButtons()
  self.canBeTapped = false
end

function ClarityReplacementMenu:hide()
  self.isVisible = false

  if (self.fadeAction) then
    self.fadeAction:stop()
  end

  self.fadeAction = AHOpacityTo.new(self, MENU_FADE_OUT_DURATION, nil, INVISIBLE_OPACITY)

  local function fadeAndResetColors()
    waitFor(MENU_FADE_OUT_DELAY)
    wait(self.fadeAction:start())
    self:reset()
  end
  runInCoroutine(fadeAndResetColors)
end

function ClarityReplacementMenu:show()
  self.isVisible = true
  local fadeAction = getReplacementsMenuOpacityAction(self, NEAR_VISIBLE_OPACITY)
  fadeAction:start()
  self.canBeTapped = true
end

function ClarityReplacementMenu:reset()
  for _, button in pairs(self.buttons) do
    button.insideFill:setColor(unpack(BUTTON_COLOR))
    button:setVisible(true)
    button.label:setColor(unpack(BUTTON_LABEL_COLOR))
  end
end

function ClarityReplacementMenu:highlightResponseAsCorrect(playerResponse, highlightColor)
  for _, button in pairs(self.buttons) do
    local response = button.label.text
    if (response == playerResponse) then
      button.insideFill:setColor(unpack(highlightColor))
      button:setVisible(false)
    end
  end
end

function ClarityReplacementMenu:showCorrectAnswer(playerResponse, highlightColor)
  for _, button in pairs(self.buttons) do
    local response = button.label.text
    if (response == playerResponse) then
      local shake = AHAxisOffsetShake.new(button, SOFT_SHAKE_WIDTH_X, AHAxisOffsetShake.SHAKE_AXIS.X, SOFT_SHAKE_DURATION_MULTIPLIER)
      shake:start()
      button.insideFill:setColor(unpack(ClarityColors.BLACK))
      button.label:setColor(unpack(ClarityColors.WHITE))
    elseif (response == self.correctResponseString) then
      button.insideFill:setColor(unpack(highlightColor))
      local labelColor = ColorUtils.getTextColorForBackingColorYIQ(highlightColor)
      button.label:setColor(unpack(labelColor))
    end
    button:setVisible(false)
  end
end

function ClarityReplacementMenu:getLocOfResponse(playerResponse)
  local x = 0
  local y = 0
  for _, button in pairs(self.buttons) do
    local response = button.label.text
    if (response == playerResponse) then
      x, y = button:getLoc()
      break
    end
  end

  return x, y
end

function ClarityReplacementMenu:setOptions(correctOption, incorrect)
  self.correctResponseString = correctOption

  local options =
  {
    correctOption,
    incorrect
  }

  TableUtils.shuffleArray(options)

  for i, button in ipairs(self.buttons) do
    button.label:setText(options[i])
  end
end

function ClarityReplacementMenu:setOnButtonTappedCallback(callback)
  self.onButtonTappedCallback = callback
end

function ClarityReplacementMenu:getButtonFactoryForResponse(response)
  for i, button in ipairs(self.buttons) do
    if (button.label.text == response) then
      return function()
        newButton = createButton(self)
        newButton:setLoc(button:getLoc())

        return newButton
      end
    end
  end
end

-- ----------------------------------------------
-- Object Creation
-- ----------------------------------------------

local function createLayout(self)
  self.buttons = {}

  for i = 1, 2 do
    local newButton = createButton(self)

    local function onDownWrapper()
      onButtonTapped(self, newButton.label.text)
      return true
    end
    Button.makeButton(newButton, onDownWrapper)

    newButton:setLoc(BUTTON_LOCATIONS[i].x, BUTTON_LOCATIONS[i].y)
    self:insert(newButton)

    table.insert(self.buttons, newButton)

    newButton.label:setAttrLink(MOAIColor.ATTR_OPACITY, newButton)
    newButton.insideFill:setAttrLink(MOAIColor.ATTR_OPACITY, newButton)
    newButton:setAttrLink(MOAIColor.ATTR_OPACITY, self)
  end

  self:setOpacity(INVISIBLE_OPACITY)
  self.isVisible = false
end

function ClarityReplacementMenu.newTransparent()
  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  Mixin.mix(self, ClarityReplacementMenu)

  createLayout(self)

  self.canBeTapped = false
  return self
end
