--- ClarityScene
-- @author Created by jay austria on March 17, 2014.

class("ClarityScene")

require("actions.advanced.Shake")
require("ClarityBackground")
require("ClarityColors")
require("ClarityWords")
require("ClarityParticles")
require("ClarityOutroEffect")
require("ClarityShapeUtils")
require("ClarityRationaleOverlay")
require("graphics.SimpleMorphingPolygon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GAME_AREA_WIDTH = Screen.getSafeWidthPart(0.92)

local NUMBER_OF_PERSISTENT_SHAPES = 6
local NUMBER_OF_SHAPES_NOT_TO_BLEND = 4
local NUMBER_OF_SHAPES_TO_BLEND = NUMBER_OF_PERSISTENT_SHAPES - NUMBER_OF_SHAPES_NOT_TO_BLEND
local PERSISTENT_SHAPE_STAGGER_SCALE = 0.0025
local MULTIPLE_OVERLAYS_STAGGER_TIME = 0.1

local REARRANGE_DELAY = 0.5

local SCENE_FADE_DURATION = 0.3

local FADE_RECT_FADE_DURATION = 0.2
local FADE_RECT_FADE_OPACITY = 0.90

local OUTRO_START_DELAY = 0.7
local OUTRO_COLOR_CHANGE_DELAY = 0.2

local MENU_HIDE_DELAY = 1

local MORPH_DURATION = 0.7

local TEXT_BACKING_OPACITY = 0.92

local HOLE_PADDING_X = 10
local HOLE_PADDING_Y = 5

local Z_INDEX = {
  BACKGROUND = 10,
  TEXT_BACKING = 20,
  TANGRAM_BACKING = 20,
  RATIONALE = 30,
  OVERLAY = 40,
  WORDS = 43,
  TANGRAM_FADE = 50,
  REPLACEMENT_MENU = 60,
  PARTICLES = 70,
  HELP_LAYER = 80,
  OUTRO = 90
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function bounceOverlayLayerDown(self)
  if (self.overlayLayer.bounceAction) then
    self.overlayLayer.bounceAction:stop()
  end

  local downAction = AHLocationYBy.new(self.overlayLayer, 0.1, AHEaseSineOut.new(), -6)
  local upAction = AHLocationYBy.new(self.overlayLayer, 0.1, AHEaseSineIn.new(), 6)
  local bounceAction = AHSequence.new(downAction, upAction)
  self.overlayLayer.bounceAction = bounceAction
  bounceAction:start()
end

local function getPaletteIndexFromColor(self, color)
  local closestIndex = 0
  local lowestScore = math.huge
  local palette = ClarityColors.getPalette()
  for colorIndex, possibleColor in ipairs(palette) do
    local currentScore = 0
    local NUMBER_OF_CHANNELS = 3
    for channelIndex = 1, NUMBER_OF_CHANNELS do
      currentScore = currentScore + math.abs(color[channelIndex] - possibleColor[channelIndex])
    end
    if (currentScore < lowestScore) then
      lowestScore = currentScore
      closestIndex = colorIndex
    end
  end

  return closestIndex
end

local function hideMenuAfterDelay(self, delayTime)
  local function delayedHide()
    waitFor(delayTime)
    self.replacementMenu:hide()
    AHOpacityTo.new(self.tangramFadeRect, FADE_RECT_FADE_DURATION, nil, INVISIBLE_OPACITY):start()
  end
  runInCoroutine(delayedHide)
end

local function getLineBoundingRectanglesForWords(self, words)
  -- group words by line first to make this easier to work with
  local lineCollection = {}
  local newLine = {}
  local prevY = nil
  for _, word in pairs(words) do
    local _, y = word:getLoc()

    -- This word isn't on the same line, store everything up to this
    -- point as a 'line' and continue on a new line
    if (prevY and prevY ~= y) then
      table.insert(lineCollection, newLine)
      newLine = {}
    end

    prevY = y
    table.insert(newLine, word)
  end
  table.insert(lineCollection, newLine)

  -- Now for each line, make a bounding rect
  local boundingRects = {}
  for _, line in pairs(lineCollection) do
    local rect = {}

    rect.minX = math.huge
    rect.minY = math.huge
    rect.maxX = -math.huge
    rect.maxY = -math.huge

    for _, word in pairs(line) do
      local x, y = self:transformLoc(word)
      local w, h = word:getDims()

      rect.minX = math.min(x, rect.minX)
      rect.minY = math.min(y, rect.minY)
      rect.maxX = math.max(x + w, rect.maxX)
      rect.maxY = math.max(y + h, rect.maxY)
    end

    table.insert(boundingRects, rect)
  end

  return boundingRects
end

local function playTangramMorphAnimation(self, x, y, rotation, index, overlay, staggerGroup, doBounce)
  local moveEase = AHEaseBackIn.new(0.2)
  local effectDuration = MORPH_DURATION

  local function morphToTangram()
    local opacity = 1.0
    if (index > NUMBER_OF_SHAPES_NOT_TO_BLEND) then
      opacity = 1 - (1.0/(NUMBER_OF_SHAPES_TO_BLEND + 1)) * (index - NUMBER_OF_SHAPES_NOT_TO_BLEND)
    end
    overlay:setOpacity(opacity)

    if ((index == 1) and (staggerGroup == 0)) then
      table.insert(self.addedTangrams, overlay)
    end

    local function onArrive()
      overlay:bringToBack()
      if index == 1 then
        if (#self.addedTangrams > 0 and doBounce) then
          bounceOverlayLayerDown(self)
        end
        if staggerGroup == 0 then
          Sound.playSoundEffect("tangram_piece_insert")
        end
      else
        overlay:removeSelf()
      end
    end

    local staggerAmount = math.pow((index - 1), 2) * PERSISTENT_SHAPE_STAGGER_SCALE
    waitFor(staggerAmount)
    self.overlayLayer:insert(overlay)

    local moveAction = AHLocationTo.new(overlay, effectDuration, moveEase, x, y)
    local morphAction = overlay:createMorphToAction(effectDuration,  AHEaseSimpleInOut.new(), 1.0)
    local rotateAction = AHRotationTo.new(overlay, effectDuration, AHEaseSimpleInOut.new(), rotation)
    local combinedTransformsAction = AHSpawn.new(morphAction, moveAction, rotateAction)

    local staggerDelay = AHDelay.new(MULTIPLE_OVERLAYS_STAGGER_TIME * staggerGroup)
    local onArriveAction = AHCallLogic.new(onArrive)

    local fullSequence = AHSequence.new(staggerDelay, combinedTransformsAction, onArriveAction)
    fullSequence:start()
  end
  runInCoroutine(morphToTangram)
end

local function morphOverlaysToTangramArea(self, overlays, tangramData, staggerGroup, doBounce)
  local x = TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_X_SCALE * tangramData.x
  local y = TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_Y_SCALE * tangramData.y
  local rotation = tangramData.rotation

  for i, overlay in ipairs(overlays) do
    playTangramMorphAnimation(self, x, y, rotation, i, overlay, staggerGroup, doBounce)
  end
end

local function getOverlaysFromWord(self, originWord, tangramData)
  local wordsForOverlays = {}

  local wordIndices = self.words.textModel:getIndicesOfWordsToRemove(originWord.wordIndex)
  for _, neighborIndex in pairs(wordIndices) do
    table.insert(wordsForOverlays, self.words:getWordViewByIndex(neighborIndex))
  end

  local lineBoundingRectangles = getLineBoundingRectanglesForWords(self, wordsForOverlays)

  local allOverlays = {}
  for _, rect in pairs(lineBoundingRectangles) do
    local overlayForShape = {}
    local width = rect.maxX - rect.minX
    local height = rect.maxY - rect.minY
    local x = rect.minX + width * 0.5
    local y = rect.minY + height * 0.5

    local scaledWidth = TANGRAM_AREA_X_SCALE * tangramData.width
    local startVertices = ClarityShapeUtils.getVerticesOfBoxWithDims(width, height)
    local endVertices = ClarityShapeUtils.getEndVerticesForShape(tangramData.shape, scaledWidth)

    for i = 1, NUMBER_OF_PERSISTENT_SHAPES do
      local newBoundingRect = SimpleMorphingPolygon.new(scaledWidth, scaledWidth, startVertices, endVertices)
      newBoundingRect:setLoc(x, y)
      newBoundingRect:setColor(unpack(tangramData.color))
      table.insert(overlayForShape, newBoundingRect)
    end
    table.insert(allOverlays, overlayForShape)
  end

  return allOverlays
end

local function morphAllOverlaysToTangramArea(self, word, tangramData, doBounce)
  local function morphAll()
    local allOverlays = getOverlaysFromWord(self, word, tangramData)
    for i, overlays in pairs(allOverlays) do
      local willCauseBounce = (doBounce and (i == #allOverlays))
      morphOverlaysToTangramArea(self, overlays, tangramData, (i - 1), willCauseBounce)
    end
  end
  runInCoroutine(morphAll)
end

local function createParticlesAtLocation(self, x, y, color)
  local particles = ClarityParticles.newWithColor(color)
  particles:setLoc(x, y)
  self:insert(particles, Z_INDEX.PARTICLES)

  particles:emit()
end

local function createEmptyLayer(parentGroup, zIndex)
  local newLayer = FilledRectangle.newFullScreen(ClarityColors.WHITE)
  newLayer:setOpacity(INVISIBLE_OPACITY)
  parentGroup:insertAtCenter(newLayer, zIndex)
  return newLayer
end

local function getWordsToHighlightByRow(self, reducibleWords)
  local reducibleWordsSet = TableUtils.createSetFromArray(reducibleWords)
  local reducibleRows = {}
  local textArea = self.words:getTextArea()
  local viewsByRow = textArea:getViewsByRow()

  for index, row in ipairs(viewsByRow) do
    for _, word in ipairs(row) do
      local wordIsReducible = (reducibleWordsSet[word] ~= nil)
      if (wordIsReducible) then
        if (reducibleRows[index] == nil) then
          reducibleRows[index] = {}
        end
        table.insert(reducibleRows[index], word)
      end
    end
  end

  return reducibleRows
end

local function getRectBounds(self, targets)
  local minimumX = math.huge
  local maximumX = -math.huge
  local minimumY = math.huge
  local maximumY = -math.huge

  for _, target in ipairs(targets) do
    local x, y = self:transformLoc(target)
    local width, height = target:getDims()
    minimumX = math.min(minimumX, x)
    maximumX = math.max(maximumX, x + width)
    minimumY = math.min(minimumY, y)
    maximumY = math.max(maximumY, y + height)
  end

  local targetWidth = maximumX - minimumX
  local targetHeight = maximumY - minimumY
  local targetX = minimumX
  local targetY = minimumY

  local rectWidth = targetWidth + HOLE_PADDING_X
  local rectHeight = targetHeight + HOLE_PADDING_Y
  local rectX = targetX + targetWidth * 0.5
  local rectY = targetY + targetHeight * 0.5

  local rectVerticalOffset = rectHeight * 0.1
  rectY = rectY - rectVerticalOffset

  return rectX, rectY, rectWidth, rectHeight, rectVerticalOffset
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ClarityScene:showRationale(rationaleText, tangramColor, rationaleDismissedCallback)
  self.rationaleOverlay:bringToFront()

  local delay = 0.17 + 0.515 -- 0.515 from AHAxisOffsetShake duration
  local showOverlayFunction = function()
    self.rationaleOverlay:show(rationaleText, tangramColor, rationaleDismissedCallback)
  end
  AHDelayed.new(delay, showOverlayFunction):start()
end

function ClarityScene:showSubstitutionHelpAndWaitForTap(response)
  local buttonFactory = self.replacementMenu:getButtonFactoryForResponse(response)

  local helpLayer = ClarityHelpLayer.newWithMasks({buttonFactory})
  self:insertAtCenter(helpLayer, Z_INDEX.HELP_LAYER)

  helpLayer:waitForTargetTap()

  helpLayer:fadeOutAndRemove()
end

function ClarityScene:addSubstitutionMessage(message)
  self.replacementMenu:addMessageLabel(message)
end

function ClarityScene:showReducibleHelpAndWaitForTap()
  local reducibleWords = self.words:getReducibleWords()
  local firstReducibleWord = reducibleWords[1]

  if firstReducibleWord ~= nil then
    local wordsToHighlightByRow = getWordsToHighlightByRow(self, reducibleWords)
    local factories = {}
    for i, row in pairs(wordsToHighlightByRow) do
      local wordsX, wordsY, wordsWidth, wordsHeight = getRectBounds(self, row)
      local getWordsHole = function()
        local wordsRect = FilledRectangle.new(wordsWidth, wordsHeight, {1.0, 1.0, 1.0, 1.0})
        wordsRect:setLoc(wordsX, wordsY)
        return wordsRect
      end
      table.insert(factories, getWordsHole)
    end

    local helpLayer = ClarityHelpLayer.newWithMasks(factories)
    self:insertAtCenter(helpLayer, Z_INDEX.HELP_LAYER)

    helpLayer:waitForTargetTap()

    helpLayer:fadeOutAndRemove()

    return firstReducibleWord
  end
end

function ClarityScene:onTimeOut(tangramData)
  local response = nil
  self.replacementMenu:lockButtons()
  self.replacementMenu:showCorrectAnswer(response, tangramData.color)

  local doHighlightWords = not self.replacementMenu:isCurrentlyVisible()
  if (doHighlightWords) then
    local doShake = false
    self.words:highlightCorrectWords(tangramData.color, doShake)
  end

  hideMenuAfterDelay(self, MENU_HIDE_DELAY)
end

function ClarityScene:initializeReplacements(concept, onResponse)
  local correctReplacement = concept.replacement
  local incorrectReplacement = concept.distractor

  self.replacementMenu:setOptions(correctReplacement, incorrectReplacement)
  self.correctReplacement = correctReplacement

  self.replacementMenu:setOnButtonTappedCallback(onResponse)
end

function ClarityScene:onSubstitutionIncorrect(response, tangramData)
  self.replacementMenu:lockButtons()
  self.replacementMenu:showCorrectAnswer(response, tangramData.color)

  hideMenuAfterDelay(self, MENU_HIDE_DELAY)
end

function ClarityScene:onSubstitutionCorrect(response, word, tangramData, isFirstCorrect)
  AHRotationShake.newScreenShake():start()

  self.words:hideWordAndNeighbors(word)
  morphAllOverlaysToTangramArea(self, word, tangramData, isFirstCorrect)

  self.replacementMenu:hide()

  local fadeInTangramsAction = AHOpacityTo.new(self.tangramFadeRect, FADE_RECT_FADE_DURATION, nil, INVISIBLE_OPACITY)
  fadeInTangramsAction:start()

  self.replacementMenu:lockButtons()
  self.replacementMenu:highlightResponseAsCorrect(response, tangramData.color)

  local x, y = self.replacementMenu:getLocOfResponse(response)
  createParticlesAtLocation(self, x, y, tangramData.color)

  local function delayedReplace()
    waitFor(REARRANGE_DELAY)
    self.words:replaceWordAndNeighbors(word, response)
  end
  runInCoroutine(delayedReplace)
end

function ClarityScene:onReducibleTapIncorrect(word, tangramData)
  self.words:highlightAsIncorrect(word)

  local doShake = false
  self.words:highlightCorrectWords(tangramData.color, doShake)
end

function ClarityScene:onReducibleTapCorrectNoReplacement(word, tangramData, isFirstCorrect)
  AHRotationShake.newScreenShake():start()

  self.words:hideWordAndNeighbors(word)
  morphAllOverlaysToTangramArea(self, word, tangramData, isFirstCorrect)

  local x, y = self:transformLoc(word)
  local w, h = word:getDims()
  local particlesX = x + w * 0.5
  local particlesY = y + h * 0.5
  createParticlesAtLocation(self, particlesX, particlesY, tangramData.color)

  local function delayedRemove()
    waitFor(REARRANGE_DELAY)
    self.words:removeWordAndNeighbors(word)
  end
  runInCoroutine(delayedRemove)
end

function ClarityScene:onReducibleTapCorrectWithReplacement(tangramData)
  local doShake = true
  self.words:highlightCorrectWords(tangramData.color, doShake)

  self.replacementMenu:show()

  local fadeOutTangramsAction = AHOpacityTo.new(self.tangramFadeRect, FADE_RECT_FADE_DURATION, nil, FADE_RECT_FADE_OPACITY)
  fadeOutTangramsAction:start()
end

function ClarityScene:getOnCleanRoundCoroutine()
  local hideAction = self.words:getHideAction()

  local function onCleanRound()
    wait(hideAction:start())
    self.words:removeSelf()
    self.words = nil
    self.replacementMenu:removeMessageLabel()
  end
  return newCoroutineWithFunction(onCleanRound)
end

function ClarityScene:createWordsWithConcept(concept, onResponse)
  local words = ClarityWords.newInvisible(concept, GAME_AREA_WIDTH)
  self.wordsLayer:insertAtCenter(words)
  self.words = words

  words:setOnResponseCallback(onResponse)
end

function ClarityScene:onStartRound()
  self.replacementMenu:reset()

  local showAction = self.words:getShowAction()
  local completionDelay = AHDelay.new(0.5)
  local showSequence = AHSequence.new(showAction, completionDelay)

  return showSequence:start()
end

function ClarityScene:getShowAction()
  local fadeAction = AHOpacityTo.new(self, SCENE_FADE_DURATION, nil, VISIBLE_OPACITY)
  return fadeAction
end

function ClarityScene:playOutroAnimations()
   -- Wait one frame to  make sure the addedTangrams list includes the
   -- most recently added tangram
  coroutine.yield()

  local shapeColorChangeActions = {}
  for i, tangram in ipairs(self.addedTangrams) do
    local delayBeforeStart = OUTRO_START_DELAY
    if (i % 2 == 0) then
      delayBeforeStart = delayBeforeStart + OUTRO_COLOR_CHANGE_DELAY * 0.5
    end

    local colorIndex = getPaletteIndexFromColor(self, {tangram:getColor()})
    local function changeColorFunction()
      local palette = ClarityColors.getPalette()
      colorIndex = colorIndex + 1
      if (colorIndex > #palette) then
        colorIndex = 1
      end
      tangram:setColor(unpack(ClarityColors.getPaletteColor(colorIndex)))
    end

    local waitBeforeStarting = AHDelay.new(delayBeforeStart)

    local changeColorAction = AHCallLogic.new(changeColorFunction)
    local waitBeforeRepeating = AHDelay.new(OUTRO_COLOR_CHANGE_DELAY)

    local colorChangeSequence = AHSequence.new(changeColorAction, waitBeforeRepeating)
    local changeColorLoop = AHLoop.new(colorChangeSequence)

    local fullSequence = AHSequence.new(waitBeforeStarting, changeColorLoop)
    table.insert(shapeColorChangeActions, fullSequence)
  end
  local allColorChangeAction = AHSpawn.new(unpack(shapeColorChangeActions))

  local outro = ClarityOutroEffect.newWithDiameter(TANGRAM_AREA_HEIGHT)
  outro:setLoc(TANGRAM_AREA_BOUNDS.xMin + TANGRAM_AREA_WIDTH * 0.5, TANGRAM_AREA_BOUNDS.yMin + TANGRAM_AREA_HEIGHT * 0.5)
  self:insert(outro, Z_INDEX.OUTRO)

  local allDotsAction = outro:getAllDotsAction()
  local hideWordsAction = self.words:getHideAction()
  local outroAction = AHSequence.new(
                               allDotsAction,
                               hideWordsAction
                               )

  self.outro:showDefaultGameWinOutro()
  outroAction:start()
  allColorChangeAction:start()
  return outroAction, allColorChangeAction
end

function ClarityScene:getModel()
  return self.words:getModel()
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createLayout(self)
  local background = ClarityBackground.new()
  self:insertAtCenter(background, Z_INDEX.BACKGROUND)
  self.background = background

  local textBacking = FilledRectangle.new(TEXT_BACKING_WIDTH, TEXT_BACKING_HEIGHT, ClarityColors.getTextBackingColor())
  textBacking:setLoc(Screen.getSafeX(0.5), TEXT_BACKING_Y)
  textBacking:setOpacity(TEXT_BACKING_OPACITY)
  self:insert(textBacking, Z_INDEX.TEXT_BACKING)

  local tangramAreaColor = ClarityColors.getTangramAreaColor()
  local tangramBackingRect = FilledRectangle.new(Screen.fullWidth, FREE_AREA_HEIGHT, tangramAreaColor)
  tangramBackingRect:setLoc(Screen.fullWidth * 0.5, FREE_AREA_HEIGHT * 0.5)
  self:insert(tangramBackingRect, Z_INDEX.TANGRAM_BACKING)
  self.tangramBackingRect = tangramBackingRect

  local rationaleOverlay = ClarityRationaleOverlay.new(FREE_AREA_HEIGHT)
  self:insertAtCenter(rationaleOverlay, Z_INDEX.RATIONALE)
  self.rationaleOverlay = rationaleOverlay

  self.overlayLayer = createEmptyLayer(self, Z_INDEX.OVERLAY)

  self.wordsLayer = createEmptyLayer(self, Z_INDEX.WORDS)

  local tangramFadeRect = FilledRectangle.new(Screen.fullWidth, TANGRAM_AREA_HEIGHT, tangramAreaColor)
  tangramFadeRect:setOpacity(INVISIBLE_OPACITY)
  tangramFadeRect:setLoc(Screen.fullWidth * 0.5, TANGRAM_AREA_HEIGHT * 0.5)
  self:insert(tangramFadeRect, Z_INDEX.TANGRAM_FADE)
  self.tangramFadeRect = tangramFadeRect

  local replacementMenu = ClarityReplacementMenu.newTransparent()
  local menuW, menuH = replacementMenu:getDims()
  replacementMenu:setLoc(menuW * 0.5, (menuH * 0.5) + TANGRAM_AREA_BOTTOM_MARGIN)
  self:insert(replacementMenu, Z_INDEX.REPLACEMENT_MENU)
  self.replacementMenu = replacementMenu
end

function ClarityScene.newHidden(outro)
  local self = Group.new(Screen.fullWidth, Screen.fullHeight)
  Mixin.mix(self, ClarityScene)

  createLayout(self)

  self:setOpacity(INVISIBLE_OPACITY)

  self.addedTangrams = {}

  self.outro = outro

  return self
end
