--- ClarityTextModel
-- @author Created by Steven Mah on January 6, 2016.

class("ClarityTextModel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Returns the item at the specified index
-- @param index (number) - Index of the item
function ClarityTextModel:getWordItemByIndex(index)
  local items = self.items
  assert(index <= #items and index >= 1, "Desired index is out of bounds.")
  return items[index]
end

--- Returns the indicies of words that are to be removed along with the word at wordIndex
-- @param wordIndex (number) - Index of the item to be removed
function ClarityTextModel:getIndicesOfWordsToRemove(wordIndex)
  local indicesToRemove = {}

  for index, word in pairs(self:getItems()) do
    local isReducible = word:getAttribute("isReducible")
    if (isReducible) then
      table.insert(indicesToRemove, index)
    end
  end

  return indicesToRemove
end

--- Returns whether a particular TextItem is reducible
-- @param wordItem - the TextItem in question
function ClarityTextModel:isWordItemReducible(wordItem)
  return wordItem:getAttribute("isReducible")
end

--- Removes the wordItem at the specific index along with all its reducible partners
-- @param wordIndex (number) - Index of the item to remove
function ClarityTextModel:reduceSentenceByWordIndex(wordIndex)
  local wordItem = self:getWordItemByIndex(wordIndex)

  assert(self:isWordItemReducible(wordItem), "Selected word is not reducible.")
  local indicesToRemove = self:getIndicesOfWordsToRemove(wordIndex)

  -- unset skipNextSpacing for TextItems preceding recently removed commas
  local items = self:getItems()
  for _, index in ipairs(indicesToRemove) do
    local wordItemToRemove = items[index]
    local wordToRemove = wordItemToRemove:getText()
    if (wordToRemove == ",") then
      local previousIndex = index - 1
      if (previousIndex >= 1) then
        local previousWordItem = items[previousIndex]
        previousWordItem:setSkipNextSpacing(false)
      end
    end
  end

  -- Indices need to be reversed before removal
  indicesToRemove = TableUtils.getReversedArray(indicesToRemove)

  self:beginTransaction()
  for _, index in ipairs(indicesToRemove) do
    self:removeItems(index, 1)
  end
  self:endTransaction()
end

--- Replaces the wordItem(s) starting at a specific index with the table of
-- replacement items given.
-- @param wordIndex (number) - Index in original sentence of word(s) to replace
-- @param replacementItems (table) - Table of replacement TextItems
function ClarityTextModel:replaceWordsInSentence(wordIndex, replacementItems)
  local wordItem = self:getWordItemByIndex(wordIndex)

  assert(self:isWordItemReducible(wordItem), "Selected word is not reducible.")
  local indicesToReplace = self:getIndicesOfWordsToRemove(wordIndex)
  self:replaceItems(indicesToReplace[1], #indicesToReplace, replacementItems)
end
-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ClarityTextModel.new(textItems)
  local self = TextModel.new(textItems)
  Mixin.mix(self, ClarityTextModel)
  return self
end
