--- ClarityWords
-- @author Created by Jay Austria, Febtober 17, 2014

class("ClarityWords")
require("ClarityColors")
require("ClarityCommon")
require("ClarityWordView")
require("ClarityTextArea")

-- ----------------------------------------------
-- Configuration
-- ----------------------------------------------

local FADE_DURATION = 0.3
local SOFT_SHAKE_WIDTH_X = 280
local SOFT_SHAKE_UP_SCALE_Y = 6
local SOFT_SHAKE_DURATION_MULTIPLIER = 0.8

local BLACKLIST_PUNCTUATION = {
  ["?"] = true,
  ["."] = true,
  ["!"] = true
}

-- ----------------------------------------------
-- Utility
-- ----------------------------------------------
local function getWordViews(self)
  return self.textArea:getItemViews()
end

local function onWordTapped(self, word)
  if self.onResponseCallback then
    self.onResponseCallback(word)
  end
end

local function makeTextArea(self, areaWidth)
  local textColor = ClarityColors.getTextColor()

  local textArea = ClarityTextArea.new(self.rawSentence, areaWidth)
  textArea:setColor(unpack(textColor))
  self.textModel = textArea:getModel()
  self.textArea = textArea

  return textArea
end

local function initializeTouch(self)
  local wordViews = getWordViews(self)
  for i, wordView in ipairs(wordViews) do
    wordView.outOfLineAnim = true
    local isPunctuation = BLACKLIST_PUNCTUATION[wordView:getText()]
    if (not isPunctuation) then
      local function onTapCallback(word)
        onWordTapped(self, word)
      end
      ClarityWordView.makeTappable(wordView, i, onTapCallback)
    end
  end
end

local function getFadeAction(self, opacity)
  if (self.fadeAction) then
    self.fadeAction:stop()
  end

  self.fadeAction = AHOpacityTo.new(self, FADE_DURATION, nil, opacity)
  return self.fadeAction
end

-- ----------------------------------------------
-- API
-- ----------------------------------------------

function ClarityWords:setText(text)
  self.textModel:setText(text)
end

function ClarityWords:getTextArea()
  return self.textArea
end

function ClarityWords:setOnResponseCallback(callback)
  self.onResponseCallback = callback
end

function ClarityWords:hasReplacement()
  return (self.replacement ~= nil)
end

function ClarityWords:getReplacement()
  return self.replacement
end

function ClarityWords:getShowAction()
  local fadeAction = getFadeAction(self, VISIBLE_OPACITY)
  return fadeAction
end

function ClarityWords:getHideAction()
  local fadeAction = AHSpawn.new(
    -- HACK: Introduced after text system changes to fix an animation regression where
    -- the word added to the sentence in Clarity fades out at the end of the round at a
    -- different rate than the rest of the sentence. Remove hack by cleaning up how
    -- opacity attribute links are handled throughout the game.
    function() self.textArea:applyItemViewOpacityAttributeLinks() end,
    getFadeAction(self, INVISIBLE_OPACITY)
  )
  return fadeAction
end

function ClarityWords:getWordViewByIndex(index)
  local wordViews = getWordViews(self)
  assert(wordViews and index <= #wordViews, "Tried to access word by invalid index.")
  return wordViews[index]
end

function ClarityWords:highlightAsIncorrect(word)
  local shake = AHAxisOffsetShake.new(word, SOFT_SHAKE_WIDTH_X, AHAxisOffsetShake.SHAKE_AXIS.X, SOFT_SHAKE_DURATION_MULTIPLIER)
  shake:start()
  word:showAsIncorrect()
end

function ClarityWords:getReducibleWords()
  local reducibleWords = {}
  local wordViews = getWordViews(self)
  for index, word in pairs(wordViews) do
    local wordItem = self.textModel:getWordItemByIndex(index)
    if wordItem:getAttribute("isReducible") then
      table.insert(reducibleWords, word)
    end
  end
  return reducibleWords
end

function ClarityWords:highlightCorrectWords(correctColor, doShake)
  local reducibleWords = self:getReducibleWords()
  for _, word in pairs(reducibleWords) do
    local _, wordH = word:getDims()
    if doShake then
      AHAxisOffsetShake.new(word, wordH * SOFT_SHAKE_UP_SCALE_Y, AHAxisOffsetShake.SHAKE_AXIS.Y):start()
    end
    word:showAsCorrect(correctColor)
  end
end

function ClarityWords:hideWordAndNeighbors(word)
  local indicesToRemove = self.textModel:getIndicesOfWordsToRemove(word.wordIndex)

  for _, removableIndex in ipairs(indicesToRemove) do
    local removableWord = self:getWordViewByIndex(removableIndex)
    removableWord:setVisible(false)
  end
end

function ClarityWords:removeWordAndNeighbors(word)
  self.textModel:reduceSentenceByWordIndex(word.wordIndex)
end

function ClarityWords:replaceWordAndNeighbors(selectedWord, replacementTextString)
  local replacementItems = {}
  for word in replacementTextString:gmatch("%S+") do
    local replacementTextItem = TextItem.newTextItem(word, {font = Font.new(FontFaces.DIN_OT_LIGHT, DEFAULT_FONT_SIZE)})
    table.insert(replacementItems, replacementTextItem)
  end

  self.textModel:replaceWordsInSentence(selectedWord.wordIndex, replacementItems)
end

function ClarityWords:getModel()
  return self.textModel
end

-- ----------------------------------------------
-- Object Creation
-- ----------------------------------------------

local function createLayout(self, collectionWidth)
  self.textArea = makeTextArea(self, collectionWidth)
  self:insert(self.textArea)

  self.textModel = self.textArea:getModel()
  self.textArea:setLoc(Screen.getSafeX(0.5), TEXT_BACKING_Y)
  self.textArea:setAttrLink(MOAIColor.ATTR_OPACITY, self)
end

function ClarityWords.newInvisible(inputRound, collectionWidth)
  local self = Group.new(Screen.fullWidth, Screen.fullHeight)
  Mixin.mix(self, ClarityWords)

  self.sentence = inputRound.sentence
  self.replacement = inputRound.replacement
  self.rawSentence = inputRound.rawSentence

  createLayout(self, collectionWidth)
  initializeTouch(self)
  self.textArea:applyItemViewOpacityAttributeLinks()

  self:setOpacity(0.0)

  return self
end
