--- ClarityValidator
-- @author Created by Anton Morant on April 17, 2014.
-- ConceptValidator for Clarity.

class("ClarityValidator")
require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'sentence',
  'replacement',
  'distractor',
  'phrase'
}

local NON_EMPTY_KEYS = {
  'sentence',
  'rationale'
}

local ADDITIONAL_NON_EMPTY = {
  'replacement',
  'distractor'
}

local SINGLE_TAGS = nil
local PAIRED_TAGS = nil
local TERTIARY_TAGS = nil
local OPTIONAL_PAIRED_TAGS = { "reduce" } -- use optional tags validator so we can catch multiple instances

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateNonEmptyKeys(self, concept)
  local expectedNonEmptyKeys = TableUtils.shallowCopy(NON_EMPTY_KEYS)
  TableUtils.concatenateArrays(expectedNonEmptyKeys, ADDITIONAL_NON_EMPTY)

  self:assertHasKeys(concept, expectedNonEmptyKeys)
end

local function validateNoTrailingOrLeadingSpace(self, sentence, tag)
  local openingTagWithTrailingSpace = string.format("<%s> ", tag)
  local closingTagWithLeadingSpace = string.format(" </%s>", tag)

  local hasTrailingSpace = (sentence:find(openingTagWithTrailingSpace) ~= nil)
  local hasLeadingSpace = (sentence:find(closingTagWithLeadingSpace) ~= nil)

  self:expectFalse(hasTrailingSpace, "Found a trailing space after an opening tag")
  self:expectFalse(hasLeadingSpace, "Found a leading space before a closing tag")
end

local function validateNoWhitespaceBetweenTagsAndWords(self, sentence)
  for _, tag in ipairs(OPTIONAL_PAIRED_TAGS) do
    validateNoTrailingOrLeadingSpace(self, sentence, tag)
  end
end

local function validateReductionsForEndingPunctuation(self, sentence, replacement)
  local sentenceWithReplacement = sentence:gsub("<reduce>(.*)</reduce>", replacement)
  self:testForEndingPunctuation(sentenceWithReplacement)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ClarityValidator:validateConcept(concept)
  local allowEmpty = true
  self:validateRequiredKeys(concept, REQUIRED_KEYS, allowEmpty)
  validateNonEmptyKeys(self, concept)
  validateNoWhitespaceBetweenTagsAndWords(self, concept.sentence)
  validateReductionsForEndingPunctuation(self, concept.sentence, concept.replacement)

  self:validateTags(concept.sentence, SINGLE_TAGS, PAIRED_TAGS, TERTIARY_TAGS, OPTIONAL_PAIRED_TAGS)
  self:expectStringContains(concept.sentence, "<reduce>", "sentence") -- at least one set of <reduce> tags must be present

  self:expectCapitalizationMatches(concept.replacement, concept.distractor, "replacement", "distractor")

  local rationale = concept.rationale
  self:validateTags(rationale)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ClarityValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, ClarityValidator)

  return self
end
