--- CollateData
-- @author Created by Jay Austria on May 15, 2014.

class("CollateData")
require("CollateUnitConversionMatrix")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PREFERRED_NUMBER_OF_DECIMALS_TO_DISPLAY = 1
local MAXIMUM_NUMBER_OF_DECIMALS_TO_DISPLAY = 4

CollateData.BASE_UNIT_CONVERSIONS =
{
  {
    category = "length",
    conversions =
    {
      ["millimeters"] = 1,
      ["centimeters"] = 10,
      ["inches"] = 25.4,
      ["feet"] = 304.8,
      ["yards"] = 914.4,
      ["meters"] = 1000,
      ["kilometers"] = 1000000,
      ["miles"] = 1609340
    }
  },
  {
    category = "volume",
    conversions =
    {
      ["milliliters"] = 1,
      ["fluid ounces"] = 29.5735,
      ["cups"] = 236.588,
      ["pints"] = 473.176,
      ["quarts"] = 946.353,
      ["liters"] = 1000,
      ["gallons"] = 3785.41
    }
  },
  {
    category = "weight",
    conversions =
    {
      ["grams"] = 1,
      ["ounces"] = 28.3495,
      ["pounds"] = 453.592,
      ["kilograms"] = 1000,
      ["tons"] = 907185
    }
  }
}

local UNIT_STRINGS =
{
  ["millimeters"] = {singular = "millimeter", plural = "millimeters", abbreviation = "mm"},
  ["centimeters"] =  {singular = "centimeter", plural = "centimeters", abbreviation = "cm"},
  ["inches"] =  {singular = "inch", plural = "inches", abbreviation = "in"},
  ["feet"] =  {singular = "foot", plural = "feet", abbreviation = "ft"},
  ["yards"] =  {singular = "yard", plural = "yards", abbreviation = "yd"},
  ["meters"] =  {singular = "meter", plural = "meters", abbreviation = "m"},
  ["kilometers"] =  {singular = "kilometer", plural = "kilometers", abbreviation = "km"},
  ["miles"] =  {singular = "mile", plural = "miles", abbreviation = "mi"},
  ["grams"] =  {singular = "gram", plural = "grams", abbreviation = "g"},
  ["ounces"] =  {singular = "ounce", plural = "ounces", abbreviation = "oz"},
  ["pounds"] =  {singular = "pound", plural = "pounds", abbreviation = "lb"},
  ["kilograms"] =  {singular = "kilogram", plural = "kilograms", abbreviation = "kg"},
  ["tons"] =  {singular = "ton", plural = "tons", abbreviation = "t"},
  ["milliliters"] =  {singular = "milliliter", plural = "milliliters", abbreviation = "ml"},
  ["fluid ounces"] =  {singular = "fluid ounce", plural = "fluid ounces", abbreviation = "fl oz"},
  ["cups"] =  {singular = "cup", plural = "cups", abbreviation = "c"},
  ["pints"] =  {singular = "pint", plural = "pints", abbreviation = "pt"},
  ["quarts"] =  {singular = "quart", plural = "quarts", abbreviation = "qt"},
  ["liters"] =  {singular = "liter", plural = "liters", abbreviation = "l"},
  ["gallons"] =  {singular = "gallon", plural = "gallons", abbreviation = "gal"}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConversionTableForUnit(unit)
  for _, conversionDefinition in pairs(CollateData.BASE_UNIT_CONVERSIONS) do
    local conversionTable = conversionDefinition.conversions
    if conversionTable[unit] then
      return conversionTable
    end
  end

  error("Conversion table not found")
  return nil
end

local function getConversionToUnitFromTable(conversionTable, targetUnit)
  return conversionTable[targetUnit]
end

local function getConversionFactorForUnit(unit)
  local conversionTable = getConversionTableForUnit(unit)
  local conversion = getConversionToUnitFromTable(conversionTable, unit)
  return conversion
end

local function getUnitStringForValue(value, unit)
  local form = "plural"

  if value == 1 then
    form = "singular"
  end

  return UNIT_STRINGS[unit][form]
end

local function getFormattedString(value, unit)
  local valueStringWithCommas = MathUtils.getCommaFormattedValue(value)
  local formattedString = string.format("%s %s", valueStringWithCommas, unit)
  return formattedString
end

local function getPresentableValue(value)
  local numberOfDecimalPlaces = PREFERRED_NUMBER_OF_DECIMALS_TO_DISPLAY
  local presentableValue = nil
  repeat
    presentableValue = MathUtils.round(value, numberOfDecimalPlaces)
    numberOfDecimalPlaces = numberOfDecimalPlaces + 1
  until (presentableValue > 0) or (numberOfDecimalPlaces > MAXIMUM_NUMBER_OF_DECIMALS_TO_DISPLAY)

  return presentableValue
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Unit conversion --------------------

function CollateData.convertUnitToBaseUnit(value, unit)
  local conversionFactor = getConversionFactorForUnit(unit)
  local valueInBaseUnit = value * conversionFactor

  return valueInBaseUnit
end

function CollateData.convertBaseUnitToUnit(value, unit)
  local conversionFactor = getConversionFactorForUnit(unit)
  local valueInUnit = value / conversionFactor
  local presentableValue = getPresentableValue(valueInUnit)

  return presentableValue
end

function CollateData.convertFromUnitToUnit(value, oldUnit, newUnit)
  local valueInBaseUnit = CollateData.convertUnitToBaseUnit(value, oldUnit)
  local valueInUnit = CollateData.convertBaseUnitToUnit(valueInBaseUnit, newUnit)

  return valueInUnit
end

-- String helpers ---------------------

function CollateData.getAbbreviationForUnit(unit)
  return UNIT_STRINGS[unit].abbreviation
end

function CollateData.getHintStringFromTo(fromValue, fromUnit, toUnit)
  local oneToOneString = HINT_STRINGS[fromUnit][toUnit]

  local originalString = CollateData.getUnitString(fromValue, fromUnit)
  local convertedString = CollateData.getUnitConversionString(fromValue, fromUnit, toUnit)
  local conversionString = string.format("%s = %s", originalString, convertedString)

  return string.format("%s<br/>%s", oneToOneString, conversionString)
end

function CollateData.getUnitString(value, unit)
  local unitString = getUnitStringForValue(value, unit)
  return getFormattedString(value, unitString)
end

function CollateData.getUnitConversionString(valueFrom, unitFrom, unitTo)
  local newValue = CollateData.convertFromUnitToUnit(valueFrom, unitFrom, unitTo)
  return CollateData.getUnitString(newValue, unitTo)
end
