--- CollateProgressIndicator
-- @author Created by Jay Austria on May 15, 2014.

class("CollateProgressIndicator")
require("widgets.progress.ProgressIndicator")
require("CollateOutroPattern")
require("CollateColors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GROUP_WIDTH = Screen.safeWidth
local GROUP_HEIGHT = Screen.height
local MARGIN_X = 10
local PADDING_X = 4
local BOX_COLOR = {0, 0, 0, 0.3}
local SCALE_UP_AMOUNT = 2.2
local SCALE_DOWN_AMOUNT = 1.0
local SCALE_UP_DURATION = 0.15
local SCALE_DOWN_DURATION = 0.3
local WAVE_DELAY = 0.04
local WIN_WAVE_DELAY = 0.04
local UPDATE_DELAY = 0.03
local FALL_DELAY_DURATION = 1.0
local CHECK_Y_OFFSET = Screen.height * 0.7
local START_SCALE = 1.0
local PANIC_SCALE_OFFSET = 0.2
local PANIC_SCALE_ACTION_DURATION = 0.1
local PANIC_COLOR_ACTION_DURATION = 0.4
local COLOR_CHANGE_DURATION = 0.2
local GRAVITY = -0.7
local MAX_VELOCITY_Y = 9.5
local MAX_VELOCITY_X = 6
local MAX_VELOCITY_Z = 0.02
local BOUNCE_DAMPENING = 0.2
local BLOCKS_LEFT_FONT_SIZE = 16
local BOXES_LEFT_LABEL_COUNT = 3
local GHOSTING_EFFECT_STAGGER_DURATION = 0.06
local MAX_DROP_STAGGER_DURATION = 0.08
local MILD_WAVE_UP_AMOUNT = 1.5
local MILD_WAVE_DOWN_AMOUNT = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createWaveEffectAnimation(self, box, waveDelay, waveHeight)
  waveHeight = waveHeight or SCALE_UP_AMOUNT

  local totalScaleDuration = SCALE_UP_DURATION + SCALE_DOWN_DURATION

  local scaleUpAnimation = AHScaleYTo.new(box, SCALE_UP_DURATION, AHEaseSimpleInOut.new(), waveHeight)
  local scaleBackAnimation = AHScaleYTo.new(box, SCALE_DOWN_DURATION, AHEaseSimpleInOut.new(), 1)

  local waveEffectAction = AHSequence.new(scaleUpAnimation, scaleBackAnimation)
  local changeColorAnimation = AHColorTo.new(box, totalScaleDuration, nil, CollateColors.getCorrectColor())
  local waveEffectWithColorChangeAnimation = AHSpawn.new(waveEffectAction, changeColorAnimation)

  local fullWaveEffect = AHLoop.new(waveEffectWithColorChangeAnimation, 3)

  local startDelay = AHDelay.new(waveDelay)
  local delayedFullWaveWithColorEffect = AHSequence.new(startDelay, fullWaveEffect)

  return delayedFullWaveWithColorEffect
end

local function animateBoxes(self)
  local numSteps = #self.boxCollection

  local boxAnimations = AHSpawn.new()
  local waveDelay = 0
  local leftBoxIndex = math.ceil(numSteps * 0.5)
  local rightBoxIndex = math.ceil((numSteps + 1) * 0.5)

  while rightBoxIndex <= numSteps do
    local leftBoxToAnimate = self.boxCollection[leftBoxIndex]
    local rightBoxToAnimate = self.boxCollection[rightBoxIndex]

    local leftWaveAnimation = createWaveEffectAnimation(self, leftBoxToAnimate, waveDelay)
    boxAnimations:addAction(leftWaveAnimation)

    if leftBoxToAnimate ~= rightBoxToAnimate then
      local rightWaveAnimation = createWaveEffectAnimation(self, rightBoxToAnimate, waveDelay)
      boxAnimations:addAction(rightWaveAnimation)
    end

    waveDelay = waveDelay + WIN_WAVE_DELAY
    leftBoxIndex = leftBoxIndex - 1
    rightBoxIndex = rightBoxIndex + 1
  end

  boxAnimations:start()
end

local function createFallingPixelFromIndex(self, boxIndex, color)
  local pixel = FilledRectangle.new(self.boxWidth, self.boxWidth, color)
  self:insert(pixel)

  local originBox = self.boxCollection[boxIndex]
  local startX, startY = originBox:getLoc()

  startY = startY + self.boxWidth * 0.5

  pixel:setLoc(startX, startY)
  return pixel
end

local function isOffscreen(object)
  local _, objectHeight = object:getEffectiveDims()
  local _, y = object:getLoc()
  return y < -objectHeight
end

local function runWinUpdateCoroutine(self, script)
  local fallDistance = self.boxWidth + PADDING_X
  local currentStep = 1

  local function updatePixels()
    while(true) do
      -- update any current pixels
      local i = 1
      while i <= #self.pixelCollection do
        local pixel = self.pixelCollection[i]
        local x, y = pixel:getLoc()
        pixel:setLoc(x, y - fallDistance)

        if isOffscreen(pixel) then
          pixel:removeSelf()
          table.remove(self.pixelCollection, i)
        else
          i = i + 1
        end
      end

      -- Create any new pixels
      if currentStep <= #script then
        local newPixels = script[currentStep]
        for _, pixelData in pairs(newPixels) do
          local index = pixelData.index
          local color = pixelData.color
          local newPixel = createFallingPixelFromIndex(self, index, color)
          table.insert(self.pixelCollection, newPixel)
        end
      end

      waitFor(UPDATE_DELAY)
      currentStep = currentStep + 1
    end
  end

  runInCoroutine(updatePixels)
end

local function collidesWithBounceArea(x, y, bounceY, bounceWidth)
  local noCollisionWidth = (Screen.width - bounceWidth) * 0.5
  local bounceLeftEdge = noCollisionWidth
  local bounceRightEdge = Screen.width - noCollisionWidth
  return (x > bounceLeftEdge) and (x < bounceRightEdge) and y < bounceY
end

local function runLoseUpdateCoroutine(self, collisionY, collisionWidth)
  local function updateFallingBoxes()
    while(#self.droppingBoxes) do
      for _, box in pairs(self.droppingBoxes) do
        if not box.stopUpdating then
          local x, y = box:getLoc()
          local z, _ = box:getScl()
          x = x + box.velocityX
          y = y + box.velocityY
          z = z + box.velocityZ

          if collidesWithBounceArea(x, y, collisionY, collisionWidth) and not box.hasBounced then
            box.velocityY = -box.velocityY * BOUNCE_DAMPENING
            box.hasBounced = true
          end

          box:setLoc(x, y)
          box:setScl(z)

          box.velocityY = box.velocityY + GRAVITY

          if isOffscreen(box) then
            box.stopUpdating = true
          end
        end
      end

      coroutine.yield()
    end
  end
  runInCoroutine(updateFallingBoxes)
end

local function animateFallingPattern(self)
  local numSteps = #self.boxCollection
  local outroScript = CollateOutroPattern.getOutroScript(numSteps)
  runWinUpdateCoroutine(self, outroScript)
end

local function addCheckToScreen(self)
  local spacing = self.boxWidth + PADDING_X
  local checkPattern = CollateOutroPattern.getCheckPattern()
  for i, row in ipairs(checkPattern) do
    local elements = #row
    for j = 1, elements do
      local character = row:sub(j,j)
      if character == "x" then
        local box = FilledRectangle.new(self.boxWidth, self.boxWidth, CollateColors.getCorrectColor())
        box:setLoc(j *  spacing, CHECK_Y_OFFSET + i * -spacing)

        box:setOpacity(0)
        self:insert(box)

        local showDelayDuration = 1.0 + 1.0 * math.random()
        local showDelay = AHDelay.new(showDelayDuration)
        local showAction = AHOpacityTo.new(box, 0.5, nil, 1)
        local showSequence = AHSequence.new(showDelay, showAction)
        showSequence:start()
      end
    end
  end
end

local function getCorrectAnimation(boxToAnimate, endColor)
  local scaleUpAction = AHScaleYTo.new(boxToAnimate, SCALE_UP_DURATION, AHEaseSimpleOut.new(), SCALE_UP_AMOUNT)
  local scaleDownAction = AHScaleYTo.new(boxToAnimate, SCALE_DOWN_DURATION, AHEaseSimpleInOut.new(), SCALE_DOWN_AMOUNT)
  local scaleSequence = AHSequence.new(scaleUpAction, scaleDownAction)

  local colorChangeDuration = SCALE_UP_DURATION + SCALE_DOWN_DURATION
  local colorActionTo = AHColorTo.new(boxToAnimate, colorChangeDuration * 0.5, nil, CollateColors.getSingleCorrectColor())
  local fullVisibilityAction = AHOpacityTo.new(boxToAnimate, colorChangeDuration * 0.1, nil, 1.0)
  local visibleAndColorChangeAction = AHSpawn.new(colorActionTo, fullVisibilityAction)
  local colorActionBack = AHColorTo.new(boxToAnimate, colorChangeDuration * 0.5, nil, endColor)
  local colorSequence = AHSequence.new(visibleAndColorChangeAction, colorActionBack)

  local correctAnimation = AHSpawn.new(scaleSequence, colorSequence)

  return correctAnimation
end

local function getIncorrectAnimation(boxToAnimate, endColor)
  local colorChangeDuration = SCALE_UP_DURATION + SCALE_DOWN_DURATION
  local colorActionTo = AHColorTo.new(boxToAnimate, colorChangeDuration * 0.5, nil, CollateColors.getSingleIncorrectColor())
  local fullVisibilityAction = AHOpacityTo.new(boxToAnimate, colorChangeDuration * 0.1, nil, 1.0)
  local visibleAndColorChangeAction = AHSpawn.new(colorActionTo, fullVisibilityAction)
  local colorActionBack = AHColorTo.new(boxToAnimate, colorChangeDuration * 0.5, nil, endColor)
  local colorSequence = AHSequence.new(visibleAndColorChangeAction, colorActionBack)
  
  return colorSequence
end

local function animateAsCorrect(boxToAnimate, endColor)
  local correctAnimation = getCorrectAnimation(boxToAnimate, endColor)
  correctAnimation:start()
end

local function animateAsIncorrect(boxToAnimate, endColor)
  local incorrectAnimation = getIncorrectAnimation(boxToAnimate, endColor)
  incorrectAnimation:start()
end


local function animateAsPerfectSet(self)
  local step = self:getProgressStep()
  local perfectSetAnimation = AHSpawn.new()
  for i = 1, MAX_CORRECT do
    local boxToAnimate = self.boxCollection[step + i - MAX_CORRECT]

    local returnColor = {boxToAnimate:getColor()}

    local correctAnimation = getCorrectAnimation(boxToAnimate, returnColor)
    local delayDuration = (i - 1) * WAVE_DELAY
    local delayAction = AHDelay.new(delayDuration)

    local correctAnimationSequence = AHSequence.new(delayAction, correctAnimation)
    perfectSetAnimation:addAction(correctAnimationSequence)
  end

  perfectSetAnimation:start()
end

local function turnBoxRed(box)
  local colorAction = AHColorTo.new(box, COLOR_CHANGE_DURATION, nil, CollateColors.getIncorrectColor())
  colorAction:start()
end

local function dropBox(self, box)
  local originalScale = 1
  box:setScl(1, 1)

  box.velocityX = math.random() * MAX_VELOCITY_X - MAX_VELOCITY_X * 0.5
  box.velocityY = math.random() * MAX_VELOCITY_Y
  box.velocityZ = math.random() * MAX_VELOCITY_Z
  box.hasBounced = false

  table.insert(self.droppingBoxes, box)
end

local function doBoxesLeftAnimation(self, numLeft)
  local messageSuffix = "   b l o c k s   t o   g o"
  local sfxName = "two_blocks_left"

  if numLeft == 1 then
    messageSuffix = "   b l o c k   t o   g o"
    sfxName = "one_block_left"
  end

  local numberAsString = tostring(numLeft)
  local formattedString = numberAsString .. messageSuffix
  local boxesLeftLabel =  Label.new(formattedString, Font.new(FontFaces.DEFAULT_FONT, BLOCKS_LEFT_FONT_SIZE))
  local labelY = self.baselineY + self.boxWidth * 0.5
  boxesLeftLabel:setLoc(Screen.width * 0.5, self.baselineY)
  local paletteIndex = 1
  boxesLeftLabel:setColor(unpack(CollateColors.getColorForStrobe(paletteIndex)))
  self:insert(boxesLeftLabel)

  boxesLeftLabel:setOpacity(0)

  local moveAction = AHLocationYBy.new(boxesLeftLabel, 0.4, AHEaseSimpleOut.new(), -20)
  local showAction = AHOpacityTo.new(boxesLeftLabel, 0.4, nil, 1.0)

  local function changeColorWrapper()
    paletteIndex = paletteIndex + 1
    boxesLeftLabel:setColor(unpack(CollateColors.getColorForStrobe(paletteIndex)))
  end

  local function sfxWrapper()
    local volume = 0.5
    Sound.playSoundEffect(sfxName, volume)
  end

  local colorChangeDelay = AHDelay.new(0.4)

  local colorChangeSequence = AHSequence.new(colorChangeDelay, changeColorWrapper)

  local presentAction = AHSpawn.new(moveAction, showAction, colorChangeSequence, sfxWrapper)

  local waitAction = AHDelay.new(0.6)

  local hideAction = AHOpacityTo.new(boxesLeftLabel, 0.4, nil, 0.0)

  local function removeSelfWrapper()
    boxesLeftLabel:removeSelf()
  end

  local fullAnimation = AHSequence.new(presentAction, waitAction, hideAction, removeSelfWrapper)

  fullAnimation:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CollateProgressIndicator:getBoxesLeft()
  return self:getStepCount() - self:getProgressStep()
end

function CollateProgressIndicator:showBoxesLeft(numLeft)
  local function showBoxesLeftWrapper()
    for i = 1, BOXES_LEFT_LABEL_COUNT do
      doBoxesLeftAnimation(self, numLeft)
      waitFor(MAX_DROP_STAGGER_DURATION)
    end
  end
  runInCoroutine(showBoxesLeftWrapper)
end

function CollateProgressIndicator:isFilled()
  return self:getBoxesLeft() == 0
end

function CollateProgressIndicator:incrementProgress(endColor, isCorrect)
  self.super.incrementProgress(self)

  local step = self:getProgressStep()

  local boxToAnimate = self.boxCollection[step]

  if isCorrect then
    animateAsCorrect(boxToAnimate, endColor)
  else
    local incorrectColor = CollateColors.getIncorrectColor()
    animateAsIncorrect(boxToAnimate, incorrectColor)
  end

  local boxesLeft = self:getBoxesLeft()
  if (boxesLeft <= 2 and not self:isFilled()) then
    self:showBoxesLeft(boxesLeft)
  end
end

function CollateProgressIndicator:animatePerfectSet()
  animateAsPerfectSet(self)
end

function CollateProgressIndicator:doLoseAnimation(collisionY, collisionWidth)
  for _, box in pairs(self.boxCollection) do
    local function delayedDropBoxWrapper()
      waitFor(math.random() * MAX_DROP_STAGGER_DURATION)
      dropBox(self, box)
      turnBoxRed(box)
    end
    runInCoroutine(delayedDropBoxWrapper)
  end
  runLoseUpdateCoroutine(self, collisionY, collisionWidth)
end

function CollateProgressIndicator:doWinAnimation()
  local numSteps = self:getStepCount()
  animateBoxes(self)
  animateFallingPattern(self)
  addCheckToScreen(self)
end

function CollateProgressIndicator:startPanicAnimation()
  local numSteps = self:getProgressStep()
  local boxScaleAnimations = AHSpawn.new()
  local boxColorAnimations = AHSpawn.new()
  for i = 1, numSteps do
    local boxToAnimate = self.boxCollection[i]

    -- will alternate between 1.4 and 0.6
    local scaleToAmount = START_SCALE + (2 * (i % 2) - 1) * PANIC_SCALE_OFFSET
    local scaleTo = AHScaleYTo.new(boxToAnimate, PANIC_SCALE_ACTION_DURATION * 0.5, nil, scaleToAmount)
    local scaleBack = AHScaleYTo.new(boxToAnimate, PANIC_SCALE_ACTION_DURATION * 0.5, nil, START_SCALE)
    local scaleSequence = AHSequence.new(scaleTo, scaleBack)
    boxScaleAnimations:addAction(scaleSequence)

    local originalColor = { boxToAnimate:getColor() }
    local colorTo = AHColorTo.new(boxToAnimate, PANIC_COLOR_ACTION_DURATION * 0.5, nil, CollateColors.getIncorrectColor())
    local colorBack = AHColorTo.new(boxToAnimate, PANIC_COLOR_ACTION_DURATION * 0.5, nil, originalColor)
    local colorSequence = AHSequence.new(colorTo, colorBack)
    boxColorAnimations:addAction(colorSequence)
  end

  local panicScaleLoop = AHLoop.new(boxScaleAnimations)
  local panicColorLoop = AHLoop.new(boxColorAnimations)

  local panicAnimation = AHSpawn.new(panicScaleLoop, panicColorLoop)
  panicAnimation:start()

  self.panicAnimation = panicAnimation
end

function CollateProgressIndicator:stopPanicAnimation()
  if self.panicAnimation then
    self.panicAnimation:stop()
    self.panicAnimation = nil
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, stepCount, baselineY)
  local boxWidth = ( (GROUP_WIDTH - (2 * MARGIN_X)) - PADDING_X * (stepCount - 1)) / stepCount

  local boxCollection = {}
  for i = 1, stepCount do
    local newBox = FilledRectangle.new(boxWidth, boxWidth, BOX_COLOR)
    newBox:setAnchorPointBottom()
    self:insert(newBox)

    local offsetX = (i - 1) * (boxWidth + PADDING_X)
    local posX = MARGIN_X + offsetX + boxWidth * 0.5
    local posY = baselineY
    newBox:setLoc(posX, posY)

    table.insert(boxCollection, newBox)
  end

  self.boxWidth = boxWidth
  self.boxCollection = boxCollection
  self.pixelCollection = {}
  self.droppingBoxes = {}
  self.baselineY = baselineY
end

function CollateProgressIndicator.new(stepCount, baselineY)
  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  ProgressIndicator.makeProgressIndicator(self, stepCount)
  Mixin.mix(self, CollateProgressIndicator)

  createLayout(self, stepCount, baselineY)

  return self
end
