--- CollateScene
-- @author Created by andrewcousins on August 24, 2022.

class("CollateScene")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  BACKGROUND = 10,
  START_PLATFORM = 20,
  NOISE = 30,
  SORT_REMINDER = 40,
  SORTABLE_LIST = 90,
  ITEM = 100,
  PROGRESS_BAR = 110,
  TAP_TO_CONTINUE = 120,
  INTERACTION_HELPER = 130,
  FLASH = 140
}

local FLASH_COLOR = {1.0, 1.0, 1.0}
local PLATFORM_COLOR = {0, 0, 0, 0.2}

local ITEMS_LIST_LOCATION_Y = (Screen.safeHeight * 0.16) + (Screen.safeBottom / 2)
local ITEM_HEIGHT = Screen.safeHeight * 0.086
local REMINDER_UI_MARGIN_Y = Screen.safeWidth * 0.11

local CELL_PADDING_Y = 3
local CELL_MARGIN_Y = Screen.safeHeight * 0.08

local FLASH_DURATION = 0.1
local FLASH_IN_OPACITY = 0.03
local FLASH_OUT_OPACITY = 0.0

local WIN_OUTRO_DURATION = 3
local REARRANGE_DELAY_DURATION = 0.5
local INCORRECT_ANSWER_DELAY = 0.5
local END_OF_ROUND_DELAY = 0.4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function doScreenFlash(self)
  local flashIn = AHOpacityTo.new(self.fullScreenFlash, FLASH_DURATION, nil, FLASH_IN_OPACITY)
  local flashOut = AHOpacityTo.new(self.fullScreenFlash, FLASH_DURATION, nil, FLASH_OUT_OPACITY)

  local flashSequence = AHSequence.new(flashIn, flashOut)
  flashSequence:start()
end

local function getHintStringForValues(item, neighbor)
  local itemValue = item.value
  local neighborValue = neighbor.value

  local itemUnitAbbreviation = CollateData.getAbbreviationForUnit(item.unit)
  local neighborUnitAbbreviation = CollateData.getAbbreviationForUnit(neighbor.unit)

  local inequalitySymbol = nil
  if itemValue > neighborValue then
    inequalitySymbol = ">"
  else
    inequalitySymbol = "<"
  end

  local hintString = string.format("%s %s %s %s %s", itemValue, itemUnitAbbreviation, inequalitySymbol, neighborValue, neighborUnitAbbreviation)
  return hintString
end

local function hideInteractionHelper(self)
  self.interactionHelper:setX(OFFSCREEN_X)
end

local function showInteractionHelper(self)
  if (not self.hasPlayerSuccessfullyInteracted) then
    local onScreenX = Screen.getWidthPart(0.5)
    self.interactionHelper:setX(onScreenX)
  end
end

local function removeHelperTrigger(self)
  if self.interactionHelperTrigger then
    TouchManager.removeTouchHandler(self.interactionHelperTrigger)
    self.interactionHelperTrigger:removeSelf()
  end
end

local function showTapToContinue(self)
  self.waitLock = true
  self.tapToContinueLayer:activate()
end

local function onContinueTap(self)
  self.tapToContinueLayer:deactivate()
  self.waitLock = false
end

local function showNextGameItem(self)
  local nextItem = table.remove(self.gameItems, 1)
  self.dragAndDropHandler:addObject(nextItem)
  self:insert(nextItem, Z_INDEX.ITEM)

  nextItem:setHome(Screen.width * 0.5, CELL_MARGIN_Y)

  local showAction = AHLocationTo.new(nextItem, ITEM_SHOW_DURATION, AHEaseSimpleInOut.new(), Screen.width * 0.5, CELL_MARGIN_Y)
  showAction:start()

  self.currentItemToPlace = nextItem
end

local function onAdd(self, item, slotColor)
  self.currentItemToPlace = nil

  local isCorrect = self.itemsList:isItemInPlace(item)
  self.onItemAddedCallback(item, isCorrect)

  if isCorrect then
    local doBandEffect = true

    if self.itemsList:isListFull() then
      slotColor = nil --Don't apply a slot color because a perfect animation will occur
    end

    item:showAsCorrect(doBandEffect, slotColor)
    doScreenFlash(self)
    if not self.itemsList:isListFull() then
      Sound.playSoundEffect("correct")
    end
  else
    -- Format is top - bottom
    local adjacentValuesString = self.itemsList:getAdjacentValuesString(item)
    item:showAsIncorrect()
    Sound.playSoundEffect("incorrect")

    local itemValue = item.value
    local itemUnit = item.unit

    local neighbor = self.itemsList:getNeighborToItem(item)
    local neighborUnit = neighbor.unit

    item:setAlternateTextFromUnit(neighborUnit)

    local hintString = nil
    local isSameUnit = (itemUnit == neighborUnit)

    if isSameUnit then
      hintString = getHintStringForValues(item, neighbor)
    else
      hintString = CollateData.getHintStringFromTo(itemValue, itemUnit, neighborUnit)
    end

    local incorrectAnimation = item:getIncorrectAnimation()
    local showCorrectionAnimation = item:getCorrectionAnimation()
    local showHintAnimation = item:getSlideOutHintAnimation(hintString)
    local rearrangeAnimation = self.itemsList:getRearrangeAnimation(item)

    local hideHintAnimation = item:getSlideInHintAnimation()
    local secondRearrangeAnimation = self.itemsList:getRearrangeAnimation()
    local cleanupAnimation = AHSpawn.new(hideHintAnimation, secondRearrangeAnimation)

    local rearrangeDelay = AHDelay.new(REARRANGE_DELAY_DURATION)

    local rearrangeSpawn = AHSpawn.new(showCorrectionAnimation, showHintAnimation, rearrangeAnimation)

    local function showTapToContinueWrapper()
      self.tapToContinueLayer:activate()
      wait(self.tapToContinueLayer:waitForTrigger())
    end

    local incorrectDelay = AHDelay.new(INCORRECT_ANSWER_DELAY)

    local incorrectSequence = AHSequence.new(incorrectAnimation,
                                             rearrangeDelay,
                                             rearrangeSpawn,
                                             incorrectDelay,
                                             showTapToContinueWrapper,
                                             cleanupAnimation)
    wait(incorrectSequence:start())
  end

  local countSoFar = self.itemsList:getNumberOfItems()
  local slotColor = self.itemsList:getItemColorForSlotIndex(countSoFar)
  self.progressBar:incrementProgress(slotColor, isCorrect)
  self.leftSortReminder:incrementProgress()
  self.rightSortReminder:incrementProgress()

  if self.itemsList:isListFull() then
    Sound.playSoundEffect("set_complete")
    waitFor(END_OF_ROUND_DELAY)
    self.onRoundCompleteCallback()
  else
    showNextGameItem(self)
  end
end

local function shiftBackgroundColors(self, topColor, bottomColor)
  local shiftDuration = ITEM_SHOW_DURATION
  local shiftColorsAction = self.background:newGradientToAction(shiftDuration, nil, topColor, bottomColor)
  shiftColorsAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CollateScene:playWinAnimation()
  self.progressBar:doWinAnimation()
  self.leftSortReminder:hide()
  self.rightSortReminder:hide()

  Sound.playSoundEffect("goal_reached")

  waitFor(WIN_OUTRO_DURATION)
end

function CollateScene:startRoundWithConcept(concept)
  local rawItemDescriptorCollection = concept.items
  local units = concept.units
  local values = concept.values

  self.itemsList:setPalette(concept.palette.itemColors)
  shiftBackgroundColors(self, unpack(concept.palette.backgroundColors))

  local itemDescriptorCollection = {}
  local numDescriptors = #rawItemDescriptorCollection

  for i = 1, numDescriptors do
    local itemDescriptor =
    {
      string = rawItemDescriptorCollection[i],
      value = values[i],
      unit = units[i],
      gameIndex = i
    }
    table.insert(itemDescriptorCollection, itemDescriptor)
  end

  TableUtils.shuffleArray(itemDescriptorCollection)

  local gameItems = {}
  for i, itemDescriptor in ipairs(itemDescriptorCollection) do
    local item = CollateItem.new(ITEM_WIDTH, ITEM_HEIGHT, CollateColors.getNewItemColor(), itemDescriptor)
    item:setLoc(Screen.width * 0.5, Screen.height * 0.0)
    table.insert(gameItems, item)
  end

  local dealLocationY = (START_PLATFORM_HEIGHT * 0.5) + Screen.safeBottom

  local firstItem = table.remove(gameItems, 1)
  local slotColor = self.itemsList:getItemColorForSlotIndex(1)
  firstItem:setBackgroundColor(slotColor)
  self:insert(firstItem, Z_INDEX.ITEM)

  local nextItem = table.remove(gameItems, 1)
  nextItem:setLoc(OFFSCREEN_X, dealLocationY)
  self:insert(nextItem, Z_INDEX.ITEM)
  self.dragAndDropHandler:addObject(nextItem)

  nextItem:setHome(Screen.width * 0.5, CELL_MARGIN_Y)

  if not self.hasPlayerSuccessfullyInteracted then
    local function handleSuccessfulInteraction()
      self.hasPlayerSuccessfullyInteracted = true
      removeHelperTrigger(self)
      hideInteractionHelper(self)
    end
    nextItem:setOnSortCallback(handleSuccessfulInteraction)

    local function reactToFailureToSort()
      removeHelperTrigger(self)
      showInteractionHelper(self)
    end
    nextItem:setReturnHomeCallback(reactToFailureToSort)
  end

  local showAction = AHLocationTo.new(nextItem, ITEM_SHOW_DURATION, AHEaseSimpleInOut.new(), Screen.width * 0.5, dealLocationY)
  showAction:start()

  self.itemsList:insertAtFront(firstItem)

  self.currentItemToPlace = nextItem
  self.gameItems = gameItems

  self.leftSortReminder:setProgressStep(0)
  self.rightSortReminder:setProgressStep(0)

  Sound.playSoundEffect("slide_in")
end

function CollateScene:playPerfectSetAnimation()
  self.progressBar:animatePerfectSet()
  wait(self.itemsList:doPerfectSetAnimation())
end

function CollateScene:clearItems()
  self.itemsList:clearItemsFromScreen()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createAndAddTapToContinueLayer(self)
  local function tapLayerTapWrapper()
    onContinueTap(self)
  end

  local tapToContinueLayer = CollateTapToContinue.new(tapLayerTapWrapper)
  self:insertAtCenter(tapToContinueLayer, Z_INDEX.TAP_TO_CONTINUE)
  self.tapToContinueLayer = tapToContinueLayer
end

local function createDragAndDropHandler(self)
  local itemsList = self.itemsList

  local function onAddWrapper(item, slotColor)
    onAdd(self, item, slotColor)
  end

  local function onDropWrapper(item)
    self.dragAndDropHandler:removeObject(item)
  end

  itemsList:setOnAddCallback(onAddWrapper)
  itemsList:setOnDropCallback(onDropWrapper)

  local dragAndDropHandler = DragAndDropHandler.new()
  dragAndDropHandler:addContainer(itemsList)
  self.dragAndDropHandler = dragAndDropHandler
end

local function createLayout(self)
  local background = RectangleGradient.new(Screen.width, Screen.height, CollateColors.getInitialBackgroundColors())
  self:insertAtCenter(background, Z_INDEX.BACKGROUND)
  self.background = background

  local noise = TiledImage.newWithFilename("tiles/noise.png", WIDTH, HEIGHT)
  self:insertAtCenter(noise, Z_INDEX.NOISE)

  createAndAddTapToContinueLayer(self)

  local itemsList = CollateSortableList.new(ITEM_WIDTH, ITEM_HEIGHT + CELL_PADDING_Y, MAX_LIST_SIZE)
  itemsList:setLoc(Screen.width * 0.5, ITEMS_LIST_LOCATION_Y)
  self:insert(itemsList, Z_INDEX.SORTABLE_LIST)
  self.itemsList = itemsList

  createDragAndDropHandler(self)

  local progressBar = CollateProgressIndicator.new(CollateConfiguration.RESPONSES_NEEDED_TO_WIN, Screen.safeTop - (Screen.safeHeight * 0.14))
  progressBar:setLoc(Screen.width * 0.5, Screen.height * 0.5)
  self.progressBar = progressBar
  self:insert(progressBar, Z_INDEX.PROGRESS_BAR)

  local sortReminderY = ITEMS_LIST_LOCATION_Y + ITEM_HEIGHT

  local leftSortReminder = CollateSortOrderReminder.new(ITEM_HEIGHT + CELL_PADDING_Y, MAX_LIST_SIZE)
  local reminderUIWidth = leftSortReminder:getWidth()
  local leftReminderX = (Screen.width * 0.5 - ITEM_WIDTH * 0.5) * 0.5
  leftSortReminder:setLoc(leftReminderX, sortReminderY)
  leftSortReminder:runUpdateCoroutine()
  self:insert(leftSortReminder, Z_INDEX.SORT_REMINDER)
  self.leftSortReminder = leftSortReminder

  local rightSortReminder = CollateSortOrderReminder.new(ITEM_HEIGHT + CELL_PADDING_Y, MAX_LIST_SIZE)
  local rightReminderX = Screen.width - (Screen.width * 0.5 - ITEM_WIDTH * 0.5) * 0.5
  rightSortReminder:setLoc(rightReminderX, sortReminderY)
  rightSortReminder:runUpdateCoroutine()
  self:insert(rightSortReminder, Z_INDEX.SORT_REMINDER)
  self.rightSortReminder = rightSortReminder

  self.interactionHelper = CollateInteractionHelperLayer.new()
  self:insertAtCenter(self.interactionHelper, Z_INDEX.INTERACTION_HELPER)
  hideInteractionHelper(self)

  local fullScreenFlash = FilledRectangle.newFullScreen(FLASH_COLOR)
  fullScreenFlash:setOpacity(0)
  self:insertAtCenter(fullScreenFlash, Z_INDEX.FLASH)
  self.fullScreenFlash = fullScreenFlash
end

-- Constructor ------------------------

function CollateScene.new(onItemAddedCallback, onRoundCompleteCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, CollateScene)

  self.onItemAddedCallback = onItemAddedCallback
  self.onRoundCompleteCallback = onRoundCompleteCallback

  self.hasPlayerSuccessfullyInteracted = false
  self.waitLock = false
  createLayout(self)

  return self
end
