--- CollateSortableList
-- @author Created by Jay Austria on May 15, 2014.

class("CollateSortableList")
require("CollateCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REARRANGE_MOVE_DURATION = 1.0
local HOME_DISTANCE_THRESHOLD = 450
local STRUM_DELAY = 0.04

local NO_SLOT = math.huge

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function compareY(a, b)
  local _, aY = a:getLoc()
  local _, bY = b:getLoc()
  return aY < bY
end

local function compareGameIndex(a, b)
  return a.gameIndex < b.gameIndex
end

local function getHomePositionFromIndex(self, index)
  assert(index >= 1 and index <= self.maxCellCount, "Invalid home position index.")

  local homeSlot = self.homeSlots[index]
  return homeSlot
end

local function getItemByIndex(self, index)
  local itemAtIndex = nil

  local numItems = #self.itemCollection
  if ( (index > 0) and (index <= numItems) ) then
    itemAtIndex = self.itemCollection[index]
  end

  return itemAtIndex
end

local function reassignHomesByIndexes(self)
  for i, item in ipairs(self.itemCollection) do
    item.homeY = getHomePositionFromIndex(self, i)
    item.listIndex = i
  end
end

local function getMoveToHomeAction(self, item, moveDuration)
  local moveDuration = moveDuration or ITEM_MOVE_HOME_DURATION
  local itemHomeY = item.homeY
  local moveHomeAction = AHLocationTo.new(item, moveDuration, AHEaseSineInOut.new(), self.groupWidth * 0.5, itemHomeY)
  return moveHomeAction
end

local function moveToHome(self, target, moveHomeDuration)
  if target.moveAction then
    target.moveAction:stop()
  end

  target.moveAction = getMoveToHomeAction(self, target, moveHomeDuration)
  target.moveAction:start()
  return target.moveAction
end

local function reassignSlots(self, hoveringIndex)
  table.sort(self.itemCollection, compareY)

  for index, item in ipairs(self.itemCollection) do
    local isIndexAfterHoveringItem = (index >= hoveringIndex)
    if isIndexAfterHoveringItem then
      item.listIndex = index + 1
    else
      item.listIndex = index
    end


    item.homeY = getHomePositionFromIndex(self, item.listIndex)
  end
end

local function getIndexOfNearHomePosition(self, x, y)
  for i, candidateSlot in ipairs(self.homeSlots) do
    local isCloseToSlot = (MathUtils.distanceSquared(0, y, 0, candidateSlot)) < HOME_DISTANCE_THRESHOLD
    if isCloseToSlot then
      local maxSlotIndex = math.min(#self.itemCollection + 1, self.maxCellCount)
      local cappedIndex = math.min(i, maxSlotIndex)
      return cappedIndex
    end
  end

  return NO_SLOT
end

local function isNearANewSlot(self, x, y)
  local nearestSlotIndex = getIndexOfNearHomePosition(self, x, y)
  local isNearSlot = (nearestSlotIndex ~= NO_SLOT)
  local isNearestSlotDifferent = (nearestSlotIndex ~= self.candidateHomeIndex)

  if (isNearSlot and isNearestSlotDifferent) then
    self.candidateHomeIndex = nearestSlotIndex
    return true
  end

  return false
end

local function getMoveAllToHomeFinalAction(self)
  local moveActions = {}
  for _, item in pairs(self.itemCollection) do
    if item.moveAction then
      item.moveAction:stop()
    end

    local moveHomeAction = getMoveToHomeAction(self, item, REARRANGE_MOVE_DURATION)
    table.insert(moveActions, moveHomeAction)
  end

  local moveAllHomeAction = AHSpawn.new(unpack(moveActions))
  return moveAllHomeAction
end

local function moveAllToHomeSlots(self, moveHomeDuration)
  local lastMoveAction = nil
  for _, item in pairs(self.itemCollection) do
    lastMoveAction = moveToHome(self, item, moveHomeDuration)
  end

  return lastMoveAction
end

local function insertItemAtIndex(self, item, index, moveHomeDuration)
  table.insert(self.itemCollection, index, item)
  self:takeChild(item)
  reassignHomesByIndexes(self)
  local lastMoveAction = moveAllToHomeSlots(self, moveHomeDuration)
  return lastMoveAction
end

local function reapplyOtherColors(self, indexToIgnore)
  for i = 1, #self.itemCollection do
    if (i ~= indexToIgnore) then
      local item = self.itemCollection[i]
      local itemColor = self:getItemColorForSlotIndex(i)
      item:animateBackgroundColorTo(itemColor)
    end
  end
end

function addItem(self, item, index)
  local moveHomeDuration = INSERT_ITEM_DURATION
  wait(insertItemAtIndex(self, item, index, moveHomeDuration))

  reapplyOtherColors(self, index)

  local itemColor = self:getItemColorForSlotIndex(index)

  if self.onAddCallback then
    self.onAddCallback(item, itemColor)
  end
end

local function setHomeIndicesToGameIndices(self)
  table.sort(self.itemCollection, compareGameIndex)

  for i, item in ipairs(self.itemCollection) do
    item.homeY = getHomePositionFromIndex(self, i)
    item.listIndex = i
  end
end

local function shiftHomeYPositionsUpAfterItem(self, incorrectItem)
  local itemHeight = incorrectItem:getHeight()
  local shiftOffsetY = itemHeight * CollateItem.SLIDE_OUT_SCALE
  for i, item in ipairs(self.itemCollection) do
    if (i >= incorrectItem.listIndex) then
      item.homeY = item.homeY + shiftOffsetY
    end
  end
end

local function onObjectDNDStateChange(self)
  reassignSlots(self, self.candidateHomeIndex)
  moveAllToHomeSlots(self)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CollateSortableList:getTopY()
  local highestIndex = self.itemCollection[#self.itemCollection].listIndex

  return CELL_MARGIN_Y + highestIndex *(self.cellHeight)
end

function CollateSortableList:getItemColorForSlotIndex(index)
  assert(index > 0 and index <= self.maxCellCount, "Invalid index for item color")
  assert(self.itemColors ~= nil and #self.itemColors > 0, "Item colors has not yet been set")

  return self.itemColors[index]
end

function CollateSortableList:doPerfectSetAnimation()
  local strumAction = AHSpawn.new()
  for i, item in ipairs(self.itemCollection) do
    if item ~= mostRecentCorrectItem then
      local delayDuration = STRUM_DELAY * (i - 1)
      local delayAction = AHDelay.new(delayDuration)

      local function showCorrectWrapper()
        local doBandEffect = false
        item:showAsCorrect(doBandEffect)
      end

      local showCorrectAnimation = AHSequence.new(delayAction, showCorrectWrapper)
      strumAction:addAction(showCorrectAnimation)
    end
  end

  return strumAction:start()
end

function CollateSortableList:getRearrangeAnimation(item)
  setHomeIndicesToGameIndices(self)
  if item ~= nil then
    shiftHomeYPositionsUpAfterItem(self, item)
  end
  local rearrangeAnimation = getMoveAllToHomeFinalAction(self)
  return rearrangeAnimation
end

function CollateSortableList:setOnAddCallback(callback)
  self.onAddCallback = callback
end

function CollateSortableList:setOnDropCallback(callback)
  self.onDropCallback = callback
end

function CollateSortableList:insertAtFront(item)
  local slotIndex = 1
  local moveHomeDuration = ITEM_SHOW_DURATION
  insertItemAtIndex(self, item, slotIndex, moveHomeDuration)
  local homeY = getHomePositionFromIndex(self, slotIndex)
  item:setLoc(OFFSCREEN_X, homeY)

  local itemColor = self:getItemColorForSlotIndex(slotIndex)
  item:setBackgroundColor(itemColor)
end

function CollateSortableList:notifyDNDDrop(object)
  local slotIndex = self.candidateHomeIndex
  if slotIndex then
    if self.onDropCallback then
      self.onDropCallback(object)
    end

    local function addItemWrapper()
      addItem(self, object, slotIndex)
    end
    runInCoroutine(addItemWrapper)
  end
end

function CollateSortableList:notifyDNDMove(object, touchX, touchY)
  local objectX, objectY = object:getLoc()

  local objectLocalX, objectLocalY = self:worldToModel(objectX, objectY)

  if isNearANewSlot(self, objectLocalX, objectLocalY) then
    local hoveringSlotIndex = getIndexOfNearHomePosition(self, objectLocalX, objectLocalY)
    reassignSlots(self, hoveringSlotIndex)
    moveAllToHomeSlots(self)
  end
end

function CollateSortableList:notifyDNDEnter(object)
  self.candidateHomeIndex = 1
  onObjectDNDStateChange(self)
end

function CollateSortableList:notifyDNDLeave(object)
  self.candidateHomeIndex = NO_SLOT
  onObjectDNDStateChange(self)
end

function CollateSortableList:canAcceptDNDObject(object)
  return true
end

function CollateSortableList:isItemInPlace(object)
  local itemBelow = getItemByIndex(self, object.listIndex - 1)
  if itemBelow then
    if (itemBelow.gameIndex > object.gameIndex) then
      return false
    end
  end

  local itemAbove = getItemByIndex(self, object.listIndex + 1)
  if itemAbove then
    if (itemAbove.gameIndex < object.gameIndex) then
      return false
    end
  end

  return true
end

function CollateSortableList:getAdjacentValuesString(object)
  local itemAbove = getItemByIndex(self, object.listIndex + 1)
  local itemBelow = getItemByIndex(self, object.listIndex - 1)

  local topString = "none"
  local bottomString = "none"
  if (itemAbove ~= nil) then
    topString = itemAbove.string
  end
  if (itemBelow ~= nil) then
    bottomString = itemBelow.string
  end
  local adjacentValuesString = topString .. " - " .. bottomString

  return adjacentValuesString
end

function CollateSortableList:isListFull()
  return #self.itemCollection >= self.maxCellCount
end

function CollateSortableList:getNumberOfItems()
  return #self.itemCollection
end

function CollateSortableList:clearItemsFromScreen()
  for _, item in pairs(self.itemCollection) do
    item:onClear()

    local function onArriveWrapper()
      item:removeSelf()
    end

    local leaveScreenAction = AHLocationXTo.new(item, ITEM_SHOW_DURATION, AHEaseSineInOut.new(), END_OFFSCREEN_X)

    local leaveSequence = AHSequence.new(leaveScreenAction, onArriveWrapper)
    leaveSequence:start()
  end

  self.itemCollection = {}
end

function CollateSortableList:setPalette(colors)
  self.itemColors = colors
end

function CollateSortableList:getNeighborToItem(item)
  local indexInsertedAt = item.listIndex

  local indexShouldHaveBeenInsertedAt = nil
  local collectionCopy = TableUtils.shallowCopy(self.itemCollection)
  table.sort(collectionCopy, compareGameIndex)

  for i, itemCopy in ipairs(collectionCopy) do
    if itemCopy == item then
        indexShouldHaveBeenInsertedAt = i
      break
    end
  end

  local didInsertTooHigh = indexInsertedAt > indexShouldHaveBeenInsertedAt

  local neighborIndex = nil
  if didInsertTooHigh then
    neighborIndex = indexShouldHaveBeenInsertedAt + 1
  else
    neighborIndex = indexShouldHaveBeenInsertedAt - 1
  end

  return collectionCopy[neighborIndex]
end

function CollateSortableList:doLoseAnimation()
  local loseAnimation = AHSpawn.new()
  for _, item in pairs(self.itemCollection) do
    local colorChangeAction = item:getAnimateBackgroundColorToAction(CollateColors.getFrozenItemColor())
    loseAnimation:addAction(colorChangeAction)
  end

  loseAnimation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function CollateSortableList.new(cellWidth, cellHeight, maxCellCount)
  local groupWidth = Screen.width
  local groupHeight = Screen.height * 0.8

  local self = Group.new(groupWidth, groupHeight)
  Mixin.mix(self, CollateSortableList)

  self.itemCollection = {}
  self.groupWidth = groupWidth
  self.groupHeight = groupHeight
  self.cellWidth = cellWidth
  self.cellHeight = cellHeight
  self.maxCellCount = maxCellCount
  self.itemColors = {}

  self:setAnchorPointBottom()

  local homeSlots = {}
  for i = 1, maxCellCount do
    local newSlot = CELL_MARGIN_Y + ((i - 1) * self.cellHeight) + (self.cellHeight * 0.5)
    table.insert(homeSlots, newSlot)
  end
  self.homeSlots = homeSlots

  return self
end
