--- CollateValidator
-- @author Created by Jay Austria on May 15, 2014.
-- ConceptValidator for Collate.

class("CollateValidator")

require("validators.ConceptValidator")
require("CollateData")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "value1",
  "value2",
  "value3",
  "value4",
  "value5",
  "value6",
  "unit1",
  "unit2",
  "unit3",
  "unit4",
  "unit5",
  "unit6"
}

local VALUES_COUNT = 6
local UNITS_TO_CATEGORIES_MAP = nil

local RESTRICTED_COMBINATIONS = {
  { "centimeters", "miles" },
  { "millimeters", "miles" },
  { "milliliters", "gallons" }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function initializeUnitsToCategoriesMap()
  if UNITS_TO_CATEGORIES_MAP == nil then
    UNITS_TO_CATEGORIES_MAP = {}

    for _, conversionTable in ipairs(CollateData.BASE_UNIT_CONVERSIONS) do
      for units, _ in pairs(conversionTable.conversions) do
        UNITS_TO_CATEGORIES_MAP[units] = conversionTable.category
      end
    end
  end
end

local function getCategoryForUnits(units)
  initializeUnitsToCategoriesMap()
  return UNITS_TO_CATEGORIES_MAP[units]
end

-- Validate helpers -------------------

local function validateValueTypes(self, concept)
  for i = 1, VALUES_COUNT do
    local fieldName = "value" .. i
    local value = concept[fieldName]
    self:assertType(value, "number", fieldName)
  end
end

local function validateUnitStrings(self, concept)
  for i = 1, VALUES_COUNT do
    local fieldName = "unit" .. i
    local units = concept[fieldName]
    local isValidUnit = getCategoryForUnits(units) ~= nil
    self:assertTrue(isValidUnit, string.format("Units for %s (%s) are not a valid unit string", fieldName, units))
  end
end

local function validateNoMixedCategories(self, concept)
  local conceptCategory = concept.category
  for i = 1, VALUES_COUNT do
    local fieldName = "unit" .. i
    local units = concept[fieldName]
    local unitCategory = getCategoryForUnits(units)
    self:expectEqual(unitCategory, conceptCategory, string.format("Units for %s (%s) are not the right category", fieldName, units))
  end
end

local function validateNoRestrictedCombinations(self, concept)
  local unitsSet = {}
  for i = 1, VALUES_COUNT do
    local fieldName = "unit" .. i
    local units = concept[fieldName]
    unitsSet[units] = true
  end

  for _, combination in ipairs(RESTRICTED_COMBINATIONS) do
    local firstUnit = combination[1]
    local secondUnit = combination[2]
    local containsCombination = unitsSet[firstUnit] and unitsSet[secondUnit]
    self:expectFalse(containsCombination, string.format("Concept contains restricted combination of units: %s, %s", firstUnit, secondUnit))
  end
end

local function validateSortedSet(self, concept)
  local lastConvertedValue = 0
  for i = 1, VALUES_COUNT do
    local units = concept["unit" .. i]
    local value = concept["value" .. i]
    local convertedValue = CollateData.convertUnitToBaseUnit(value, units)
    self:expectGreaterThan(convertedValue, lastConvertedValue, string.format("Value in position %s is not greater than the last", i))
    lastConvertedValue = convertedValue
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CollateValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateValueTypes(self, concept)
  validateUnitStrings(self, concept)
  validateNoMixedCategories(self, concept)
  validateNoRestrictedCombinations(self, concept)
  validateSortedSet(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CollateValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, CollateValidator)

  return self
end
