--- CommasGame
-- @author Created by andrewc on May 29, 2019.

class("CommasGame")

require("StandardUIBuilder")
require("CommasConfiguration")
require("CommasScene")
require("CommasRoundGenerator")
require("CommasColors")
require("bonuses.ResponseSpeedTracker")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = CommasColors.WHITE

local MINIMUM_SECONDS_NEEDED_TO_START_NEW_ROUND = 2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Keeps track of any comma blanks ("words") that were correctly answered before
-- an incorrect response. When this same concept is resurfaced at a later time,
-- we will reference this list of correct words to prepopulate those comma blanks
-- that the user already correctly answered prior to committing their error.
local function trackCorrectWordIfNeeded(self, conceptId, word)
  if (not self.hasAnsweredIncorrectlyDuringRound) then
    if (self.correctWordListsByConceptId[conceptId] == nil) then
      self.correctWordListsByConceptId[conceptId] = {}
    end

    table.insert(self.correctWordListsByConceptId[conceptId], word)
  end
end

local function getCorrectWordsForConceptId(self, conceptId)
  local correctWords = self.correctWordListsByConceptId[conceptId]
  if (correctWords == nil) then
    return {}
  else
    return self.correctWordListsByConceptId[conceptId]
  end
end

local function prepopulateCorrectCommas(self, round)
  local correctWords = getCorrectWordsForConceptId(self, round:getIdentifier())

  local numberOfInteractions = round:getNumberOfInteractions()
  if (#correctWords >= numberOfInteractions) then
    -- Safeguard to abort the prepopulate step if we're going to try
    -- and fill all the correct commas (or more).
    return
  end

  local correctWordIndices = {}
  for _, word in ipairs(correctWords) do
    -- Sometimes, due to how the CommasRound may prepopulate commas upon construction,
    -- a word that we want to prepopulate may have already been prepopulated and
    -- removed from the round's correct word list. Skip past this index if so.
    local correctWordIndex = round:getCorrectWordIndex(word)
    if (correctWordIndex ~= nil) then
      table.insert(correctWordIndices, round:getCorrectWordIndex(word))
    end
  end

  -- Reverse sort the indices to avoid issues with indices changing as array
  -- elements are removed during comma prepopulation
  TableUtils.reverseSortArray(correctWordIndices)

  for _, commaIndex in ipairs(correctWordIndices) do
    round:prepopulateComma(commaIndex)
  end
end

local function allowMistakesToResurfaceIfNeeded(self)
    if (self.timer:getTimeRemaining() < CommasConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
      self.roundGenerator:allowMistakesToResurface()
    end
end

local function reportAnswer(self, inputItem, isCorrect)
  local conceptId = self.round:getIdentifier()
  if isCorrect then
    GameSession.contentTracker:notifyCorrectAnswer(conceptId, inputItem:getText())
    GameSession.incrementCorrectAnswerCount()
    trackCorrectWordIfNeeded(self, conceptId, inputItem:getText())
  else
    self.hasAnsweredIncorrectlyDuringRound = true

    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, inputItem:getText())
    GameSession.incrementIncorrectAnswerCount()

    if (not self.roundGeneratorHasNotifiedMistakeThisRound) then
      self.roundGenerator:notifyMistake(conceptId)
      self.roundGeneratorHasNotifiedMistakeThisRound = true
    end
  end
end

-- Wait lock --------------------------

local function waitForRoundComplete(self)
  self.roundComplete = false
  waitUntil(function() return self.roundComplete end)
end

local function unlockWait(self)
  if (self.roundComplete ~= nil) then
    self.roundComplete = true
  end
end

local function endRoundIfOver(self)
  if (self.commasLeftInRound <= 0) then
    wait(self.scene:waitForAnimationsToComplete())
    unlockWait(self)
  end
end

-- Handlers --------------------------

local function getAnswerHandler(self)
  local endRoundIfOverFunction = function()
    endRoundIfOver(self)
  end
  local rationaleClosedCallback = function()
    self.timer:resume()
  end
  return function(item, itemView)
    if (self.commasLeftInRound > 0) then
      if item:getAttribute(CommasCommon.IS_CORRECT_ATTRIBUTE) then
        self.commasLeftInRound = self.commasLeftInRound - 1
        self.scene:correctAnswer(item, itemView, endRoundIfOverFunction, self.commasLeftInRound)
        reportAnswer(self, item, true)
        self.responseSpeedTracker:notifyCorrectResponse()
        GameSession.increaseScore(CommasConfiguration.getPointsForCorrectComma(GameSession.getCorrectAnswerCount()))
        self.scoreLabel:setScore(GameSession.getScore())
      else
        self.timer:addTime(CommasConfiguration.INCORRECT_RESPONSE_TIME_PENALTY)
        self.timer:pause()
        self.scene:incorrectAnswer(item, itemView, rationaleClosedCallback, self.commasLeftInRound)
        self.responseSpeedTracker:notifyIncorrectResponse()
        reportAnswer(self, item, false)
      end
    end
  end
end

-- Game flow --------------------------

local function startRound(self, isFirstRound)
  GameSession.contentTracker:notifyPresentedConcept(self.round:getIdentifier())
  self.scene:startNewRound(self.round, isFirstRound)
  self.commasLeftInRound = self.round:getNumberOfInteractions()
  self.responseSpeedTracker:notifyStart()
end

local function endFinalRound(self)
  wait(self.scene:revealCorrectCommas())
  wait(self.scene:endFinalRound())
end

local function endGame(self)
  wait(self.scene:playOutro())
  self.scene:clearScene()
  GameSession.notifyGameWin()
end

local function scoreCompletedRound(self, roundNumber)
  GameSession.increaseScore(CommasConfiguration.getPointsForRoundCompleted(roundNumber))
  self.scoreLabel:setScore(GameSession.getScore())
end

local function isGameOver(self)
  local timeRemaining = self.timer:getTimeRemaining()

  return (timeRemaining <= MINIMUM_SECONDS_NEEDED_TO_START_NEW_ROUND)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CommasGame:gameFlow()
  self.scene:startGame()

  local timerStartDelayToSyncWithIntroAnimation = 0.9
  runInCoroutine(function()
    waitFor(timerStartDelayToSyncWithIntroAnimation)
    self.timer:start()
  end)

  local roundNumber = 1
  while not isGameOver(self) do
    allowMistakesToResurfaceIfNeeded(self)
    self.hasAnsweredIncorrectlyDuringRound = false
    self.roundGeneratorHasNotifiedMistakeThisRound = false
    local round = self.roundGenerator:next()

    prepopulateCorrectCommas(self, round)

    self.round = round
    local isFirstRound = (roundNumber == 1)
    startRound(self, isFirstRound)
    waitForRoundComplete(self)

    scoreCompletedRound(self, roundNumber)
    roundNumber = roundNumber + 1
  end

  endFinalRound(self)

  if GameSession.didReachPassScore() then
    endGame(self)
  else
    GameSession.notifyGameLoss()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local passScore = CommasConfiguration.getPassScore()
  local rankTwoScore = CommasConfiguration.getRankTwoScore()
  local rankThreeScore = CommasConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)

  local difficulty = GameSession.getDifficulty()
  local difficultyKeyFrames = CommasConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, difficulty)
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  self.roundGenerator = LearningFocusedRoundGenerator.new(CommasRoundGenerator.new(GameSession.conceptGenerator))
end

local function createGameElements(self)
  self.scene = CommasScene.new(getAnswerHandler(self))
  self:insertAtCenter(self.scene)
  self.numberOfStreakItems = 0
  self.correctWordListsByConceptId = {}
  self.hasAnsweredIncorrectlyDuringRound = false
  self.roundGeneratorHasNotifiedMistakeThisRound = false
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion, earnedBonusProportionSum, CommasConfiguration.POINTS_FOR_FULL_SPEED_BONUS, self.bonusTray)
  end

  local allowMultipleSpeedyRespones = true
  self.responseSpeedTracker = ResponseSpeedTracker.new(CommasConfiguration.SPEED_BONUS_WINDOW, quickResponseWrapper, CommasCommon.SPEED_BONUS_DECAY_START, nil, allowMultipleSpeedyRespones)

  local timeOutHandler = function()
    unlockWait(self)
  end

  self.timer = DigitTimer.new(CommasConfiguration.GAME_DURATION, timeOutHandler, nil, nil, UI_COMPONENTS_COLOR)
  self.timer:addStandardWarningSequence()
  self.timer:setActiveOpacity(1)
  self.timer:setPausedOpacity(0.25)
  builder:addWidget(self.timer, UIOverlay.NORTHEAST)
  self.timer:initialize()

end

-- Constructor ------------------------

function CommasGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, CommasGame)

  CommasCommon.initialize()

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)

  return self
end
