--- CommasStatusbar
-- @author Created by andrewc on May 30, 2019.

class("CommasStatusbar")

require("CommasCommaHolder")
require("CommasAnswerAnimations")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LINE_COLOR = {1, 1, 1, 0.21}

local Z_INDEX = {
  BACKGROUND = 10,
  COMMAS = 20,
  COVER = 30
}

local COMMA_SPACING = 8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getNextComma(self)
  return self.commas[self.nextCommaIndex]
end

local function makeComma(self)
  local commaHolder = CommasCommaHolder.new(self)

  return commaHolder
end

local function getBalanceAnimation(self)
  local animation = AHSpawn.new()
  local duration = CommasCommon.ROUND_TRANSITION_IN
  local ease = AHEaseBezier.newFlat({0,0,0,1})
  local numberOfCommas = #self.commas
  local xOffset = self.commas[1]:getWidth() + COMMA_SPACING
  local initialX = self.commaArea:getXLeft() + ((self.commaArea:getWidth() - ((xOffset) * numberOfCommas)) / 2)
  for i, comma in ipairs(self.commas) do
    local commaX = initialX + ((i - 0.5) * xOffset)
    animation:addAction(AHLocationXTo.new(comma, duration, ease, commaX))
  end
  initialX = self.commaArea:getXLeft() + ((self.commaArea:getWidth() - ((xOffset) * #self.holderLines)) / 2)
  for i, line in ipairs(self.holderLines) do
    local lineX = initialX + ((i - 0.5) * xOffset)
    animation:addAction(AHLocationXTo.new(line, duration, ease, lineX))
  end
  return animation
end

local function drawCommaHolderLines(self)
  for i = 1, #self.commas - 1 do
    local lineWidth = 1
    local line = FilledRectangle.new(COMMA_SPACING, lineWidth, LINE_COLOR)
    self.commaArea:insertAtCenter(line)
    table.insert(self.holderLines, line)
  end
end

local function getHolderLineScaleAnimation(self, commaIndex)
  local duration = 0.05
  local duration2 = 0.5
  local easeUp = AHEaseLinear.new()
  local easeDown = AHEaseBezier.newFlat({0,0,0,1})
  local amount = self.commas[1]:getWidthPart(0.3)

  local fullAnimation = AHSpawn.new()

  local getScaleAnimation = function(obj, moveRight)
    local partOne = AHSpawn.new()
    partOne:addAction(AHScaleXTo.new(obj, duration, easeUp, 1 + (amount / obj:getWidth())))
    local modifier = (moveRight) and 1 or -1
    partOne:addAction(AHLocationXBy.new(obj, duration, easeUp, modifier * amount/2))

    local partTwo = AHSpawn.new()
    partTwo:addAction(AHScaleXTo.new(obj, duration2, easeDown, 1))
    partTwo:addAction(AHLocationXBy.new(obj, duration2, easeDown, -1 * modifier * amount/2))

    return AHSequence.new(partOne, partTwo)
  end

  if (commaIndex ~= #self.commas) then
    local line = self.holderLines[commaIndex]
    local animation = getScaleAnimation(line, false)
    fullAnimation:addAction(animation)
  end
  if (commaIndex ~= 1) then
    local line = self.holderLines[commaIndex - 1]
    local animation = getScaleAnimation(line, true)
    fullAnimation:addAction(animation)
  end

  return fullAnimation
end

local function getMovingCommaAnimation(self, itemView, commaItem, isCorrect, isEndOfGame)
  isEndOfGame = isEndOfGame or false
  local animation = AHSpawn.new()
  animation:addAction(CommasAnswerAnimations.getTapAnimation(itemView, 1, 0.5))
  local lineScale = getHolderLineScaleAnimation(self, self.nextCommaIndex)
  animation:addAction(AHDelayed.new(0.167, lineScale))
  animation:addAction(function() wait(commaItem:animateCommaToItem(itemView, isCorrect, isEndOfGame)) end)
  return animation
end

local function getClearAnimation(self)
  local animation = AHSpawn.new()
  local duration = CommasCommon.ROUND_TRANSITION_OUT
  for _, comma in ipairs(self.commas) do
    local closeAnim = comma:getCloseAnimation()
    local moveAnim = AHLocationXTo.new(comma, duration, AHEaseBezier.newFlat({0,0,0,1}), self.commaArea:getX())
    animation:addAction(closeAnim)
    animation:addAction(moveAnim)
  end
  for _, line in ipairs(self.holderLines) do
    local fadeOut = AHOpacityTo.new(line, duration, AHEaseBezier.newFlat({0,0,0,1}), 0)
    animation:addAction(fadeOut)
  end
  return animation
end

local function getStartAnimation(self)
  local animation = AHSpawn.new()
  local duration = CommasCommon.ROUND_TRANSITION_IN
  animation:addAction(getBalanceAnimation(self))
  for _, comma in ipairs(self.commas) do
    local openAnimation = comma:getOpenAnimation()
    animation:addAction(openAnimation)
  end
  for _, line in ipairs(self.holderLines) do
    local fadeIn = AHOpacityTo.new(line, duration, AHEaseBezier.newFlat({0,0,0,1}), 1)
    animation:addAction(fadeIn)
  end
  return animation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CommasStatusbar:startRound(round)
  local numberOfCommas = round:getNumberOfInteractions()
  while (numberOfCommas > 0) do
    local commaHolder = makeComma(self)
    self.commaArea:insertAtCenter(commaHolder)
    table.insert(self.commas, commaHolder)
    numberOfCommas = numberOfCommas - 1
  end
  self.nextCommaIndex = 1
  drawCommaHolderLines(self)
  getNextComma(self):setSelected(true)
end

function CommasStatusbar:getStartRoundAnimation()
  return getStartAnimation(self)
end

function CommasStatusbar:getEndRoundAnimation()
  local clearAnimation = getClearAnimation(self)
  local clearAndRemove = AHSequence.new(clearAnimation, function() self:clear() end)
  return clearAndRemove
end

function CommasStatusbar:endFinalRound(remainingCommaViews)
  local animation = AHSpawn.new()
  for i, itemView in ipairs(remainingCommaViews) do
    local commaIndex = self.nextCommaIndex + i - 1
    local commaToAnimate = self.commas[commaIndex]
    if itemView and commaToAnimate then
      animation:addAction(getMovingCommaAnimation(self, itemView, commaToAnimate, true))
    else
      print("Warning - Tried to animate the end of round sequence with itemView \'" .. itemView:getText() .. "\' and statusbar comma at index: " ..commaIndex)
    end
  end
  wait(animation:start())
end

function CommasStatusbar:getCommaFoundAnimation(itemView, isEndOfGame)
  local commaToMove = getNextComma(self)

  local commaAnimation = getMovingCommaAnimation(self, itemView, commaToMove, true, isEndOfGame)

  self.nextCommaIndex = self.nextCommaIndex + 1

  commaToMove:setSelected(false)
  local nextComma = getNextComma(self)
  if (nextComma ~= nil) then
    local delay = 0.167
    commaAnimation:addAction(AHDelayed.new(delay, function() nextComma:setSelected(true) end))
  end
  return commaAnimation
end

function CommasStatusbar:animateIncorrectComma(itemView, shouldMoveToNextComma)
  local commaToMove = getNextComma(self)

  local movingAnimation = getMovingCommaAnimation(self, itemView, commaToMove, false)

  local postMoveFunction = function()
    if shouldMoveToNextComma then
      self.nextCommaIndex = self.nextCommaIndex + 1
      commaToMove:setSelected(false)
      commaToMove:markAsFound()
      local nextComma = getNextComma(self)
      if (nextComma ~= nil) then
        nextComma:setSelected(true)
      end
    end
  end

  return AHSequence.new(movingAnimation, postMoveFunction)
end

function CommasStatusbar:clear()
  while #self.commas > 0 do
    table.remove(self.commas):removeSelf()
  end
  while #self.holderLines > 0 do
    table.remove(self.holderLines):removeSelf()
  end
end

function CommasStatusbar:temporarilyHide()
  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0,0.3,0,1})

  self.offset = Screen.height - self:getYBottom() + COMMA_SPACING

  AHLocationYBy.new(self, duration, ease, self.offset):start()
end

function CommasStatusbar:getReshowAnimation()
  local duration = 0.333
  local ease = AHEaseBezier.newFlat({0.25,0,0,1})

  return AHDelayed.new(0.1, AHLocationYBy.new(self, duration, ease, -self.offset))
end

function CommasStatusbar:getScene()
  return self.delegate
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CommasStatusbar.new(delegate, width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, CommasStatusbar)

  self.delegate = delegate
  self.commas = {}
  self.holderLines = {}

  local commaArea = Group.new(width, height)
  self:insertAtCenter(commaArea, Z_INDEX.BACKGROUND)

  self.fullScreenZone = Group.newFullScreen()
  self:insert(self.fullScreenZone, Z_INDEX.COMMAS)

  self.commaArea = commaArea
  self.nextCommaIndex = 1

  return self
end
