--- CommasTextLayoutBuilder
-- @author Created by andrewc on May 30, 2019.

class("CommasTextLayoutBuilder")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_SPACING_MULTIPLIER = 0.3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- State ------------------------------

local function reset(self, itemViews, itemsForViews)
  self.layouts = {}
  self.itemViews = itemViews
  self.itemsForViews = itemsForViews
  self.remainingItemViews = TableUtils.shallowCopy(itemViews)
  self.textHeight = 2 -- Height for an empty view
  self.lines = {}
  self.alreadyJumpedItemViews = {}
end

-- Misc helpers -----------------------

local function getDefaultSpacing(item)
  return item:getFont():getActualSize() * DEFAULT_SPACING_MULTIPLIER
end

local function getSpacing(item)
  if item then
    local spacing = item:getSpacing()
    if spacing then
      return spacing
    end
    return getDefaultSpacing(item)
  end
  return nil
end

local function getSpacingBetweenItems(self, itemView, nextItemView)
  local itemsForViews = self.itemsForViews
  local item = itemsForViews[itemView]
  local nextItem = itemsForViews[nextItemView]

  if item and item:getSkipNextSpacing() then
    local combinedSpacing = 0

    if itemView.getRightMargin then
      combinedSpacing = combinedSpacing + itemView:getRightMargin()
    end

    if nextItemView and nextItemView.getLeftMargin then
      combinedSpacing = combinedSpacing + nextItemView:getLeftMargin()
    end

    return combinedSpacing
  end

  if nextItem and nextItem:getSkipPreviousSpacing() then
    local combinedSpacing = 0

    if nextItemView.getLeftMargin then
      combinedSpacing = combinedSpacing + nextItemView:getLeftMargin()
    end

    if itemView and itemView.getRightMargin then
      combinedSpacing = combinedSpacing + itemView:getRightMargin()
    end

    return combinedSpacing
  end

  local itemSpacing = getSpacing(item)
  local nextItemSpacing = getSpacing(nextItem)
  if itemSpacing and nextItemSpacing then
    return (itemSpacing + nextItemSpacing) / 2
  elseif itemSpacing then
    return itemSpacing
  elseif nextItemSpacing then
    return nextItemSpacing
  else
    error("Either item or nextItem need to be not nil")
  end
end

-- Layout operations ------------------

local function positionItemView(self, itemView, penX, currentLine)
  local layouts = self.layouts

  assert(itemView ~= currentLine[#currentLine], "Item is already positioned")
  table.insert(currentLine, itemView)

  assert(layouts[itemView] == nil, "Item already has a layout")
  local layout = TextItemLayout.new(penX, nil, 0, nil)
  layouts[itemView] = layout
end

local function unpositionItemView(self, itemView, currentLine)
  local remainingItemViews = self.remainingItemViews
  local layouts = self.layouts

  assert(itemView == currentLine[#currentLine], "Item is not positioned")
  table.remove(currentLine)

  assert(itemView ~= remainingItemViews[1])
  table.insert(remainingItemViews, 1, itemView)

  assert(layouts[itemView] ~= nil, "Item was positioned but does not have a layout")
  layouts[itemView] = nil
end

local function jumpToNextLine(self, itemView, currentLine, forceJump)
  if forceJump == nil then
    forceJump = true
  end

  local itemsForViews = self.itemsForViews
  local alreadyJumpedItemViews = self.alreadyJumpedItemViews

  if alreadyJumpedItemViews[itemView] and not forceJump then
    return
  end

  unpositionItemView(self, itemView, currentLine)
  alreadyJumpedItemViews[itemView] = true

  local previousItemView = currentLine[#currentLine]
  if previousItemView then
    local previousItem = itemsForViews[previousItemView]
    local shouldJumpPreviousItem = (not previousItem:getAllowBreakLineAfter())

    if shouldJumpPreviousItem then
      forceJump = false
      jumpToNextLine(self, previousItemView, currentLine, forceJump)
    end
  end
end

local function removeEmptyLines(self)
  local lines = self.lines
  local nonEmptyLines = TableUtils.getFilteredArray(lines, function(line)
    if #line == 0 then
      print("WARNING: An empty line was produced. This is not expected, probably produced by items \
             forcing linebreaks resulting in a too long line where all items jump to the next line.")
    end
    return #line > 0
  end)
  self.lines = nonEmptyLines
end

local function adjustHorizontalAlignment(self)
  local layouts = self.layouts
  local lines = self.lines
  local horizontalAlignment = self.horizontalAlignment
  local textWidth = self.textWidth

  for _, line in ipairs(lines) do
    local lastItemView = line[#line]
    local lastItemLayout = layouts[lastItemView]
    local lineEndX = lastItemLayout:getX() + lastItemView:getWidth()
    local shiftX = nil

    if horizontalAlignment == Alignment.LEFT then
      shiftX = 0
    elseif horizontalAlignment == Alignment.CENTER then
      shiftX = (textWidth - lineEndX)/2
    elseif horizontalAlignment == Alignment.RIGHT then
      shiftX = textWidth - lineEndX
    end

    if shiftX ~= 0 then
      for _, itemView in ipairs(line) do
        local layout = layouts[itemView]
        local newX = layout:getX() + shiftX
        layout:setX(newX)
      end
    end
  end
end

local function adjustVerticalAlignment(self)
  local layouts = self.layouts
  local itemsForViews = self.itemsForViews
  local lines = self.lines
  local verticalAlignment = self.verticalAlignment
  local lineSpacing = self.lineSpacing

  local lineHeights = {}
  for i, line in ipairs(lines) do
    local lineHeight = itemsForViews[line[1]]:getFont():getActualSize()
    lineHeight = lineHeight + lineSpacing
    table.insert(lineHeights, lineHeight)
  end

  local textHeight = TableUtils.sum(lineHeights)
  self.textHeight = textHeight

  local lineBaseY = textHeight

  for i, line in ipairs(lines) do
    local lineHeight = lineHeights[i]
    lineBaseY = lineBaseY - lineHeight

    local shiftY = nil
    local anchorY = nil

    if verticalAlignment == Alignment.BOTTOM then
      shiftY = 0
      anchorY = 0
    elseif verticalAlignment == Alignment.CENTER then
      shiftY = lineHeight/2
      anchorY = 0.5
    elseif verticalAlignment == Alignment.TOP then
      shiftY = lineHeight
      anchorY = 1
    end

    for _, itemView in ipairs(line) do
      local layout = layouts[itemView]
      local item = itemsForViews[itemView]
      local yOffset = item:getYOffset()
      local y = lineBaseY + shiftY + yOffset
      layout:setY(y)
      layout:setAnchorY(anchorY)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CommasTextLayoutBuilder:updateLayout(itemViews, itemsForViews)
  reset(self, itemViews, itemsForViews)

  if #itemViews == 0 then
    return
  end

  local textWidth = self.textWidth
  local remainingItemViews = self.remainingItemViews
  local itemsForViews = self.itemsForViews
  local lines = self.lines

  local currentLine = {}
  table.insert(lines, currentLine)

  local penX = 0

  while #remainingItemViews > 0 do
    local itemView = table.remove(remainingItemViews, 1)
    local item = itemsForViews[itemView]
    local itemWidth = itemView:getWidth()

    if itemWidth > textWidth then
      print(string.format("WARNING: Item is too wide for text layout. Item width: %s, text width: %s", itemWidth, textWidth))
    end

    positionItemView(self, itemView, penX, currentLine)

    local itemDoesNotFit = penX + itemWidth > textWidth
    local mustBreakLine = item:getMustBreakLine()
    local canBreakLine = item:getAllowBreakLineBefore()
    local isAtBeginningOfLine = #currentLine == 1
    local doNextLine = ((canBreakLine and itemDoesNotFit) or mustBreakLine) and not isAtBeginningOfLine

    local nextItemView = remainingItemViews[1]
    local lookAhead = nextItemView ~= nil and canBreakLine and not isAtBeginningOfLine
    if lookAhead then
      local nextItem = itemsForViews[nextItemView]
      if not nextItem:getAllowBreakLineBefore() then
        local nextItemDoesNotFit = penX + itemWidth + nextItemView:getWidth() > textWidth
        if nextItemDoesNotFit then
          doNextLine = true
        end
      end
    end

    if doNextLine then
      penX = 0

      jumpToNextLine(self, itemView, currentLine)

      currentLine = {}
      table.insert(lines, currentLine)
    else
      local horizontalSpacing = getSpacingBetweenItems(self, itemView, nextItemView)
      penX = penX + itemWidth + horizontalSpacing
    end
  end

  removeEmptyLines(self)
  adjustHorizontalAlignment(self)
  adjustVerticalAlignment(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CommasTextLayoutBuilder.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing)
  local self = TextLayoutBuilder.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing)
  Mixin.mix(self, CommasTextLayoutBuilder)

  return self
end
