--- CommasTextView
-- @author Created by andrewc on May 30, 2019.

class("CommasTextView")

require("CommasCommon")
require("CommasTextLayoutBuilder")
require("CommasTextAnimationBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_HORIZONTAL_ALIGNMENT = Alignment.LEFT
local DEFAULT_VERTICAL_ALIGNMENT = Alignment.CENTER
local DEFAULT_LINE_SPACING = 17

local BASE_ATTRIBUTES = {
  font = CommasCommon.FONT.MAIN_TEXT
}

local ATTRIBUTES_FOR_COMMAS = {
  font = CommasCommon.FONT.COMMA,
  skipPreviousSpacing = true,
  allowBreakLineAfter = true,
  allowBreakLineBefore = false
}

local ATTRIBUTES_FOR_LABELS = {
  i = { font = CommasCommon.FONT.MAIN_TEXT_ITALIC },
  [CommasCommon.MISSING_COMMA_TAG_LABEL] = { [CommasCommon.IS_CORRECT_ATTRIBUTE] = true,
                                             [CommasCommon.IS_FOUND_ATTRIBUTE] = false },
  [CommasCommon.DISTRACTOR_COMMA_TAG_LABEL] = { [CommasCommon.IS_INCORRECT_ATTRIBUTE] = true,
                                                [CommasCommon.IS_FOUND_ATTRIBUTE] = false },
  [CommasCommon.PREPOPULATED_COMMA_TAG_LABEL] = ATTRIBUTES_FOR_COMMAS,
  [CommasCommon.OXFORD_COMMA_TAG_LABEL] = { [CommasCommon.IS_OXFORD_ATTRIBUTE] = true },
  [CommasCommon.IGNORE_WORD] = { [CommasCommon.IS_FOUND_ATTRIBUTE] = true }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTouchHandler(self, item, button)
  return function()
    if self.touchEnabled then
      button:setTouchEnabled(false)
      self.answerHandler(item, self.viewsForItems[item])
    end
  end
end

local function showMarkers(self)
  local markers = {}
  local items = {}

  local duration = CommasCommon.ROUND_TRANSITION_OUT
  local ease = AHEaseBezier.newFlat({0.75,0,0.98,0.98})
  local createAndInsertMarker = function(item)
    local radius = item:getFont():getActualSize() * 0.65
    local marker = FilledCircle.new(radius, CommasColors.WHITE)
    marker:setOpacity(0.13)
    local outline = HollowCircle.new(radius, CommasColors.WHITE, 1, 0)
    outline:setOpacity(0.17)
    marker:insertAtCenter(outline)

    marker.getHideAnimation = function()
      marker:setTouchEnabled(false)
      local animation = AHSpawn.new()
      animation:addAction(AHScaleTo.new(marker, duration, ease, 1.1))
      animation:addAction(AHOpacityTo.new(marker, duration, ease, 0))
      animation:addAction(AHScaleTo.new(outline, duration, ease, 1.1))
      animation:addAction(AHOpacityTo.new(outline, duration, ease, 0))
      return animation
    end

    local showEase = AHEaseElasticOut.new(0.7)
    local showDuration = 0.6
    marker.getShowAnimation = function()
      marker:setScl(0)
      outline:setScl(0)
      local animation = AHSpawn.new()
      animation:addAction(AHScaleTo.new(marker, showDuration, showEase, 1))
      animation:addAction(AHScaleTo.new(outline, showDuration, showEase, 1))
      return animation
    end

    local view = self.viewsForItems[item]
    view:insertAtLocationFractional(marker, 1, 0.5)
    CommasCommon.offsetItemToComma(marker)
    table.insert(markers, marker)
    table.insert(items, item)

    local buttonSizeIncrease = 1.75

    Button.addVirtualButton(marker, marker:getWidthPart(buttonSizeIncrease), marker:getHeightPart(buttonSizeIncrease), nil, getTouchHandler(self, item, marker))
  end

  for _, indexAndItemPair in ipairs(self.indexToCommaItems) do
    createAndInsertMarker(indexAndItemPair[2])
  end
  for _, indexAndItemPair in ipairs(self.indexToDistractorItems) do
    createAndInsertMarker(indexAndItemPair[2])
  end
  self.markersForItems = TableUtils.buildDictionary(items, markers)
end

local function updateCommaIndexes(self)
  self.indexToCommaItems = {}
  self.indexToDistractorItems = {}
  self.distractorItemToIndex = {}
  local distractorItemIndex = 1
  for i, item in ipairs(self.textModel:getItems()) do
    if item:getAttribute(CommasCommon.IS_CORRECT_ATTRIBUTE) then
      table.insert(self.indexToCommaItems, {i, item})
    end
    if item:getAttribute(CommasCommon.IS_INCORRECT_ATTRIBUTE) then
      table.insert(self.indexToDistractorItems, {i, item})
      self.distractorItemToIndex[item] = distractorItemIndex
      distractorItemIndex = distractorItemIndex + 1
    end
  end
  table.sort(self.indexToCommaItems, function(a,b) return a[1]<b[1] end)
  table.sort(self.indexToDistractorItems, function(a,b) return a[1]<b[1] end)
end

local function insertComma(self, index)
  local comma = TextItem.newTextItem(",", ATTRIBUTES_FOR_COMMAS)
  self.textModel:insertItems({comma}, index)
end

local function setText(self, text)
  local textInterpreter = self.textInterpreter
  local textModel = self.textModel

  local items = textInterpreter:interpretXmlText(text)
  textModel:resetItems(items)
  updateCommaIndexes(self)
  textModel:beginTransaction()
  textModel:resetItems(items)
  showMarkers(self)
end

local function setAttributesToKeepLinesStatic(self)
  local rows = self:getViewsByRow()
  for _, row in ipairs(rows) do
    for _, itemView in ipairs(row) do
      local item = self.itemsForViews[itemView]
      item:setAllowBreakLineAfter(false)
      item:setAllowBreakLineBefore(false)
    end
    self.itemsForViews[row[1]]:setAllowBreakLineBefore(true)
    self.itemsForViews[row[#row]]:setAllowBreakLineAfter(true)
    self.itemsForViews[row[1]]:setMustBreakLine(true)
  end
end

local function addMissingComma(self, item)
  local itemIndex = 0
  for _, indexAndItemPair in ipairs(self.indexToCommaItems) do
    if (item == indexAndItemPair[2]) then
      indexAndItemPair[2]:setAttribute(CommasCommon.IS_FOUND_ATTRIBUTE, true)
      itemIndex = indexAndItemPair[1]
      break
    end
  end

  insertComma(self, itemIndex + 1)
  updateCommaIndexes(self)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CommasTextView:getRemainingCorrectItemViewPairs()
  local itemViewPairs = {}
  for _, indexAndItemPair in ipairs(self.indexToCommaItems) do
    if (not indexAndItemPair[2]:getAttribute(CommasCommon.IS_FOUND_ATTRIBUTE)) then
      local itemViewPair = {indexAndItemPair[2], self.viewsForItems[indexAndItemPair[2]]}
      table.insert(itemViewPairs, itemViewPair)
    end
  end
  return itemViewPairs
end

function CommasTextView:getHideMarkerAnimation(item)
  if item then
    local marker = self.markersForItems[item]
    if marker then
      return marker.getHideAnimation()
    end
  end
end

function CommasTextView:getShowAllMarkersAnimation()
  local fullAnim = AHSpawn.new()
  for _, marker in pairs(self.markersForItems) do
    fullAnim:addAction(marker.getShowAnimation())
  end
  return fullAnim
end

function CommasTextView:getHideAllMarkersAnimation()
  local fullAnim = AHSpawn.new()
  for _, marker in pairs(self.markersForItems) do
    fullAnim:addAction(marker.getHideAnimation())
  end
  return fullAnim
end

function CommasTextView:setTouchEnabled(enabled)
  self.touchEnabled = enabled
end

function CommasTextView:startRound(round)
  setText(self, round:getText())
  setAttributesToKeepLinesStatic(self)
end

function CommasTextView:clear()
  setText(self, "")
end

function CommasTextView:commaFound(item)
  self:getHideMarkerAnimation(item):start()
  addMissingComma(self, item)
end

function CommasTextView:markErrorItem(item)
  item:setAttribute(CommasCommon.IS_FOUND_ATTRIBUTE, true)
  local commaItem = nil
  -- finds the textItem after the passed item
  for i, itemAtIndex in ipairs(self:getModel():getItems()) do
    if commaItem then
      commaItem = itemAtIndex
      break
    end
    if (item == itemAtIndex) then
      commaItem = true
    end
  end

  self.viewsForItems[commaItem]:bringToFront()
end

function CommasTextView:setAnswerHandler(answerHandler)
  self.answerHandler = answerHandler
end

function CommasTextView:getDistractorIndexForItem(item)
  return self.distractorItemToIndex[item]
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CommasTextView.new(answerHandler, textWidth, horizontalAlignment, verticalAlignment, lineSpacing)
  horizontalAlignment = horizontalAlignment or DEFAULT_HORIZONTAL_ALIGNMENT
  verticalAlignment = verticalAlignment or DEFAULT_VERTICAL_ALIGNMENT
  lineSpacing = lineSpacing or DEFAULT_LINE_SPACING

  local self = TextView.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing)
  Mixin.mix(self, CommasTextView)

  self.layoutBuilder = CommasTextLayoutBuilder.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing)

  self.textInterpreter = TextInterpreter.new(BASE_ATTRIBUTES, ATTRIBUTES_FOR_LABELS)
  self.textModel = TextModel.new(self.textInterpreter:interpretXmlText(""))
  self:setModel(self.textModel)

  self.answerHandler = answerHandler
  self.touchEnabled = false
  self.markersForItems = {}

  self.animationBuilder = CommasTextAnimationBuilder.new()

  return self
end
