--- CommasValidator
-- @author Created by andrewc on May 29, 2019.
-- ConceptValidator for Commas.

class("CommasValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")
require("utils.XmlParser")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'text',
  'numDistractors',
  'numCommas',
  'filter_locale'
}

local REQUIRE_RATIONALE_KEYS = {
  'distractor_1',
  'distractor_2',
  'distractor_3'
}

local MAX_COMMAS = 3
local MAX_DISTRACTORS = 3

local COMMA_TAG = "comma"
local DISTRACTOR_TAG = "distractor"
local PREPOPULATED_TAG = "prepopulated"

local PAIRED_TAGS = {COMMA_TAG, PREPOPULATED_TAG, DISTRACTOR_TAG, "i"}

local TAG_REGEX = "%<(.-)%>"

local VALID_RATIONALE_KEYS = {
  ["dis-that"] = true,
  ["dis-verb_connect"] = true,
  ["dis-necessary"] = true,
  ["dis-sub_conj"] = true,
  ["dis-list"] = true,
  ["dis-name"] = true,
  ["dis-two_actions"] = true,
  ["dis-month_day"] = true,
  ["dis-month_year"] = true,
  ["dis-adverb"] = true,
  ["dis-adj_noun"] = true
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateRequiredRationaleKeys(self, concept)
  local expectedNonEmptyKeys = {}
  for i=1, concept.numDistractors do
    table.insert(expectedNonEmptyKeys, REQUIRE_RATIONALE_KEYS[i])
  end

  self:assertHasKeys(concept, expectedNonEmptyKeys)
end

local function validateCommaIsUsed(self, concept)
  self:expectTrue((concept.numCommas > 0), "All concepts must have at least 1 comma to add.")
  self:expectTrue(concept.numCommas <= MAX_COMMAS, "All concepts must have 3 or less commas to add")
  self:expectStringContains(concept.text, "<"..COMMA_TAG..">", "text")
end

local function validateDistractorIsUsed(self, concept)
  self:expectTrue(concept.numDistractors <= MAX_DISTRACTORS, "All concepts must have 3 or less distractors to remove.")
  self:expectTrue((concept.numDistractors > 0), "All concepts must have at least 1 distractor to remove.")
  self:expectStringContains(concept.text, "<"..DISTRACTOR_TAG..">", "text")
end

local function validateNoTrailingOrLeadingSpace(self, sentence, tag)
  local openingTagWithTrailingSpace = string.format("<%s> ", tag)
  local closingTagWithLeadingSpace = string.format(" </%s>", tag)

  local hasTrailingSpace = (sentence:find(openingTagWithTrailingSpace) ~= nil)
  local hasLeadingSpace = (sentence:find(closingTagWithLeadingSpace) ~= nil)

  self:expectFalse(hasTrailingSpace, "Found a trailing space after an opening tag")
  self:expectFalse(hasLeadingSpace, "Found a leading space before a closing tag")
end

local function validateNoSpaceInsideTags(self, sentence, tag)
  local tag_regex = string.format("<%s>(.-)</%s>", tag, tag)
  for tagContents in sentence:gmatch(tag_regex) do
    self:expectTrue(tagContents:match(" ") == nil, "There should not be any space in the string \'"..tagContents.."\' for tag \'"..tag.."\'.")
  end
end

local function validateNoWhitespaceBetweenTags(self, sentence)
  for _, tag in ipairs(PAIRED_TAGS) do
    validateNoTrailingOrLeadingSpace(self, sentence, tag)
    if tag ~= "i" then
      validateNoSpaceInsideTags(self, sentence, tag)
    end
  end
end

local function validateSpacesBeforeAndAfterTag(self, sentence, tag)
  local openingTag = string.format(".?<%s>.", tag)
  local closingTag = string.format("</%s>.?", tag)

  local syntaxTable = {
    ['\''] = true,
    ['\"'] = true
  }

  local acceptableOpenTagPreceedingChars = {
    ['<'] = true,
    [' '] = true,
    ['-'] = true
  }
  for openTag in sentence:gmatch(openingTag) do
    if not syntaxTable[openTag[-1]] then
      self:expectTrue(acceptableOpenTagPreceedingChars[openTag[1]] == true, "The " .. tag.. " tag should have a space before it starts, not \'"..openTag[1].."\'")
    end
  end

  local acceptableCloseTagFollowingCharacters = {
    ['>'] = true,
    [' '] = true,
    ['"'] = true,
    ['\''] = true,
    ['<'] = true
  }
  for closeTag in sentence:gmatch(closingTag) do
    self:expectTrue(acceptableCloseTagFollowingCharacters[closeTag[-1]] == true, "The " .. tag.. " tag should have a space after it closes, not \'" .. closeTag[-1].."\'")
  end
end

local function validateSpacesAroundCommas(self, sentence)
  validateSpacesBeforeAndAfterTag(self, sentence, COMMA_TAG)
  validateSpacesBeforeAndAfterTag(self, sentence, DISTRACTOR_TAG)
end

local function validateSentenceLength(self, concept)
  local sentence = concept.text
  local cleanedText = sentence:gsub(TAG_REGEX, "")
  local textLength = #cleanedText
  self:expectTrue(textLength >= 115 and textLength <= 185, "Text length out of bounds at " .. textLength)
end

local function validatePrepopulatedTag(self, concept)
  local prepopulatedRegex = "<prepopulated>(.-)</prepopulated>"
  for prepopulatedItem in concept.text:gmatch(prepopulatedRegex) do
    self:expectTrue(prepopulatedItem == ",", "Only commas can be prepopulated, not \"" .. prepopulatedItem .. "\".")
  end
end

local function validateCommaTagContent(self, sentence)
  local tag_regex = string.format("<%s>(.-)</%s>", COMMA_TAG, COMMA_TAG)
  for tagContents in sentence:gmatch(tag_regex) do
    self:expectTrue(tagContents ~= "\'", "Comma tags should not wrap only a single quote, but the entire word instead. Otherwise, users can easily be marked wrong for correct answers.")
    self:expectTrue(tagContents ~= "\"", "Comma tags should not wrap only a double quote, but the entire word instead. Otherwise, users can easily be marked wrong for correct answers.")
  end
end

local function validateRationaleKeys(self, concept)
  local rationaleKeys = {
    concept.distractor_1,
    concept.distractor_2,
    concept.distractor_3
  }

  for i=1, concept.numDistractors do
    local rationaleKey = rationaleKeys[i]
    self:expectTrue(VALID_RATIONALE_KEYS[rationaleKey], "Rationale key was not found: \'" ..rationaleKey.."\'")
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CommasValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateRequiredRationaleKeys(self, concept)

  validateCommaIsUsed(self, concept)

  validateDistractorIsUsed(self, concept)

  self:expectNoError(function() XmlParser.parse(concept.text) end)

  self:validateTags(concept.text, nil, nil, nil, PAIRED_TAGS)

  validateNoWhitespaceBetweenTags(self, concept.text)

  validateSpacesAroundCommas(self, concept.text)

  validateSentenceLength(self, concept)

  validatePrepopulatedTag(self, concept)

  self:testForEndingPunctuation(concept.text)

  self:expectNoDoubleSpaces(concept.text, "text")

  validateCommaTagContent(self, concept.text)

  validateRationaleKeys(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CommasValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, CommasValidator)

  return self
end
