--- ContentReviewCardAnimatedMath
-- @author Created by Steven Mah on January 3, 2017.

class("ContentReviewCardAnimatedMath")

require("ContentReviewCard")
require("ContentReviewDouble")
require("ContentReviewTakeaway")
require("ContentReviewCountup")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TIP_TEXT_MAPPING = {
  DOUBLE = "You can get the answer quickly if you recognize that each digit of the number you're subtracting is half of the corresponding digit of the number you're subtracting from.",
  TAKEAWAY = "Each digit in the top row is larger than the corresponding number in the bottom row. To solve, take away each digit in the bottom row from the one above it to get the answer.",
  COUNTUP = "You can solve this problem without borrowing numbers. Count up from the smaller number in steps until you reach the larger number. Add each step together to get the answer."
}

local TIP_CLASS_MAPPING = {
  DOUBLE = ContentReviewDouble,
  TAKEAWAY = ContentReviewTakeaway,
  COUNTUP = ContentReviewCountup
}

local ANIMATION_STATES = {
  READY = "READY",
  IN_PROGRESS = "IN_PROGRESS",
  COMPLETED = "COMPLETED"
}

local BUTTON_TEXT = {
  PLAY = "PLAY",
  PAUSE = "PAUSE",
  REPLAY = "REPLAY"
}

local BUTTON_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 16)
local BUTTON_COLOR = ContentReviewCommon.Colors.darkGray

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeButtonIcon(label)
  local text = label:getText()
  local height = label:getHeightPart(0.5)
  local icon = Group.new(height, height)

  if (text == BUTTON_TEXT.PLAY) then
    local points = {
      { x = 0, y = 0 },
      { x = 0, y = height },
      { x = height, y = height / 2 }
    }

    local arrow = FilledTriangle.newWithTable(points, BUTTON_COLOR)
    icon:insertAtCenter(arrow)
  elseif (text == BUTTON_TEXT.PAUSE) then
    local leftBar = FilledRectangle.new(height / 3, height, BUTTON_COLOR)
    leftBar:setAnchorPointLeft()
    icon:insertAtLeft(leftBar)

    local rightBar = FilledRectangle.new(height / 3, height, BUTTON_COLOR)
    rightBar:setAnchorPointRight()
    icon:insertAtRight(rightBar)
  elseif (text == BUTTON_TEXT.REPLAY) then
    local startAngle = math.rad(-90)
    local endAngle = math.rad(210)
    local arc = HollowArc.new(height / 2, startAngle, endAngle, BUTTON_COLOR, 1.5, nil, false)
    icon:insertAtCenter(arc)

    local points = {
      { x = 0, y = height / 5 },
      { x = height / 4, y = height / 2.5 },
      { x = height / 4, y = 0 }
    }
    local tinyArrow = FilledTriangle.newWithTable(points, BUTTON_COLOR)
    tinyArrow:setLoc(icon:getWidthPart(0.5), tinyArrow:getHeightPart(0.25))
    icon:insert(tinyArrow)
  else
    error("Invalid button type.")
  end

  return icon
end

local function transitionButtons(self, startingButton, targetButton)
  startingButton:setTouchEnabled(false)
  if (self.buttonTransition ~= nil) then
    self.buttonTransition:stop()
  end
  local transitionAction = AHSequence.new(
    AHStaggeredSpawn.new(
      0.15,
      AHOpacityTo.new(startingButton, 0.1, nil, 0.0),
      AHOpacityTo.new(targetButton, 0.1, nil, 1.0)
    ),
    function() targetButton:setTouchEnabled(true) end
  )
  self.buttonTransition = transitionAction
  transitionAction:start()
end

local function displayReplay(self)
  transitionButtons(self, self.pauseButton, self.replayButton)
  self.currentAnimationState = ANIMATION_STATES.COMPLETED
  self.internalAnimation = nil
end

local function makeButton(self, text, touchDownCallback, content, isVisible)
  local buttonLabel = Label.new(text, BUTTON_FONT)
  local buttonIcon = makeButtonIcon(buttonLabel)
  buttonLabel:setColor(unpack(BUTTON_COLOR))
  buttonIcon:setColor(unpack(BUTTON_COLOR))

  local button = Row.new(buttonLabel:getHeightPart(0.5))
  button:insert(buttonLabel)
  button:insert(buttonIcon)
  button:alignChildren()
  button:linkOpacityAttrToChildren(true)

  Button.makeButton(button, touchDownCallback)
  button:setLoc(Screen.getWidthPart(0.5), self.animationGroup:getYBottom() * 0.75)
  content:insert(button)

  if (not isVisible) then
    button:setOpacity(0.0)
  end

  button:setTouchEnabled(false)

  return button
end

local function makeButtons(self, content)
  local onPlay = function()
    transitionButtons(self, self.playButton, self.pauseButton)
    self.currentAnimationState = ANIMATION_STATES.IN_PROGRESS
    if (self.internalAnimation == nil) then
      self.internalAnimation = AHSequence.new(
        self.animationGroup:createAnimations(),
        function() displayReplay(self) end
      )
      self.internalAnimation:start()
    else
      self.internalAnimation:pause(false)
    end
  end

  local isVisible = true
  local playButton = makeButton(self, BUTTON_TEXT.PLAY, onPlay, content, isVisible)
  self.playButton = playButton

  local onPause = function()
    self.internalAnimation:pause()
    transitionButtons(self, self.pauseButton, self.playButton)
  end

  local pauseButton = makeButton(self, BUTTON_TEXT.PAUSE, onPause, content)
  self.pauseButton = pauseButton

  local onReplay = function()
    transitionButtons(self, self.replayButton, self.pauseButton)
    self.currentAnimationState = ANIMATION_STATES.IN_PROGRESS
    self.animationGroup:reset()
    coroutine.yield() -- NOTE: Needed to avoid errors with transformLoc

    self.internalAnimation = AHSequence.new(
      self.animationGroup:createAnimations(),
      function() displayReplay(self) end
    )
    self.internalAnimation:start()
  end
  local replayButton = makeButton(self, BUTTON_TEXT.REPLAY, onReplay, content)
  self.replayButton = replayButton
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ContentReviewCardAnimatedMath:createAnimationAction()
  return nil
end

function ContentReviewCardAnimatedMath:reset()
  if (self.currentAnimationState == ANIMATION_STATES.IN_PROGRESS) then
    if (self.internalAnimation) then
      self.internalAnimation:stop()
      self.internalAnimation = nil
    end
    self.pauseButton:setTouchEnabled(false)
    self.replayButton:setTouchEnabled(false)
    self.playButton:setTouchEnabled(false)

    self.pauseButton:setOpacity(0.0)
    self.replayButton:setOpacity(0.0)
    self.playButton:setOpacity(1.0)
    self.animationGroup:reset()
    self.currentAnimationState = ANIMATION_STATES.READY
  end
end

function ContentReviewCardAnimatedMath:onLeaveBegin()
  self.pauseButton:setTouchEnabled(false)
  self.replayButton:setTouchEnabled(false)
  self.playButton:setTouchEnabled(false)

  self.exitButton:setTouchEnabled(false)

  self.super.onLeaveBegin(self)
end

function ContentReviewCardAnimatedMath:onEnter()
  self.pauseButton:setTouchEnabled(false)

  if (self.currentAnimationState == ANIMATION_STATES.READY) then
    self.replayButton:setTouchEnabled(false)
    self.playButton:setTouchEnabled(true)
  else
    self.replayButton:setTouchEnabled(true)
    self.playButton:setTouchEnabled(false)
  end


  self.exitButton:setTouchEnabled(true)

  self.super.onEnter(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, concept)
  local tipType = concept.tipType

  local tipText = TIP_TEXT_MAPPING[tipType]
  local textWidth = ContentReviewCommon.defaultContentWidth
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 17)
  local defaultAlignment = nil
  local lineSpacing = Screen.getHeightPart(0.005)
  local tipTextView = DynamicTextArea.newSimple(tipText, textWidth, font, defaultAlignment, defaultAlignment, lineSpacing)
  tipTextView:setColor(unpack(ContentReviewCommon.Colors.lightGray))
  tipTextView:setAnchorPointTop()
  tipTextView:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.9))

  local tipClass = TIP_CLASS_MAPPING[tipType]
  local minuend = concept.minuend
  local subtrahend = concept.subtrahend
  local answer = minuend - subtrahend
  local animationGroup = tipClass.new(textWidth, minuend, subtrahend, answer)
  animationGroup:setAnchorPointTop()
  animationGroup:setLoc(Screen.getWidthPart(0.5), tipTextView:getYBottom() - Screen.getHeightPart(0.05))
  self.animationGroup = animationGroup

  local content = Group.newFullScreen()
  content:insert(tipTextView)
  content:insert(animationGroup)
  makeButtons(self, content)

  self:insertAtCenter(content)

  return content
end


-- Constructor ------------------------

function ContentReviewCardAnimatedMath.new(concept, parameters)
  local self = ContentReviewCard.new()
  Mixin.mix(self, ContentReviewCardAnimatedMath, "ContentReviewCard")

  createLayout(self, concept)

  self.currentAnimationState = ANIMATION_STATES.READY

  return self
end
