--- ContentReviewCardOutro
-- @author Created by Joe Poka on June 29, 2016.

class("ContentReviewCardOutro")

require("ContentReviewCard")
require("graphics.primitives.Checkmark")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DISK_RADIUS = Screen.getWidthPart(0.09)

local DISK_DATA = {
  {
    initialX = -DISK_RADIUS * 0.5,
    initialY = Screen.getHeightPart(0.575),
    initialOpacity = 0.1,
    finalOpacity = 0.4,
    initialScale = 0.3,
    destinationFraction = 0.975
  },
  {
    initialX = -DISK_RADIUS * 0.5,
    initialY = Screen.getHeightPart(0.425),
    initialOpacity = 0.3,
    finalOpacity = 0.5,
    initialScale = 0.5,
    destinationFraction = 0.985
  },
  {
    initialX = Screen.width + DISK_RADIUS * 0.5,
    initialY = Screen.getHeightPart(0.475),
    initialOpacity = 0.6,
    finalOpacity = 1.0,
    initialScale = 0.6
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addAnimatedItems(self)
  local width = DISK_RADIUS
  local shortSegmentProportion = 0.425
  local thicknessProportion = 0.25
  local checkmarkColor = {1.0, 1.0, 1.0, 1.0}
  self.checkmark = FilledCheckmark.new(width, shortSegmentProportion, thicknessProportion, checkmarkColor)
  self.checkmark:setOpacity(0.0)

  local diskColor = ContentReviewCommon.Colors.highlightBlue

  for i = 1, #DISK_DATA do
    local disk = FilledCircle.new(DISK_RADIUS, diskColor)
    disk.hasCheckmark = false
    self:insert(disk)
    table.insert(self.disks, disk)

    if (i == #DISK_DATA) then
      disk.hasCheckmark = true
      disk:insertAtCenter(self.checkmark)
    end
  end
end

local function resetAnimatedItems(self)
  self.checkmark:setOpacity(0.0)

  for i = 1, #DISK_DATA do
    local data = DISK_DATA[i]
    local disk = self.disks[i]
    disk:setLoc(data.initialX, data.initialY)
    disk:setOpacity(data.initialOpacity)
    disk:setScl(data.initialScale)
  end
end

local function showAnimatedItems(self, show)
  for _, disk in ipairs(self.disks) do
    disk:setVisible(show)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ContentReviewCardOutro:createAnimationAction()
  local enterDuration = 1.0
  local settleDuration = 0.7
  local enterActions = {}
  local settleActions = {}

  local fadeInTitle = AHOpacityTo.new(self.titleLabel, enterDuration, nil, 1.0)
  table.insert(enterActions, fadeInTitle)

  local translateTitle = AHLocationYTo.new(self.titleLabel, enterDuration, AHEaseExponentialOut.new(), self.finalTitleLabelY)
  table.insert(enterActions, translateTitle)

  local fadeInNote = AHOpacityTo.new(self.noteLabel, enterDuration, nil, 1.0)
  table.insert(enterActions, fadeInNote)

  local translateNote = AHLocationYTo.new(self.noteLabel, enterDuration, AHEaseExponentialOut.new(), self.finalNoteLabelY)
  table.insert(enterActions, translateNote)

  local fadeInButton = AHOpacityTo.new(self.button, enterDuration, nil, 1.0)
  table.insert(enterActions, fadeInButton)

  local translateButton = AHLocationYTo.new(self.button, enterDuration, AHEaseExponentialOut.new(), self.finalButtonY)
  table.insert(enterActions, translateButton)

  showAnimatedItems(self, true)

  local defaultDestinationX = Screen.getWidthPart(0.5)
  local defaultDestinationY = Screen.getHeightPart(0.65)

  for i = 1, #self.disks do
    local disk = self.disks[i]
    local data = DISK_DATA[i]

    disk.destinationX = defaultDestinationX
    disk.destinationY = defaultDestinationY

    -- lingering/rotating disks don't move all the way to disk destination
    if (not disk.hasCheckmark) then
      disk.destinationX = disk.destinationX * data.destinationFraction
      disk.destinationY = disk.destinationY * data.destinationFraction
      disk.offset = MathUtils.distance(disk.destinationX, disk.destinationY, defaultDestinationX, defaultDestinationY)
      disk.angle = MathUtils.angleBetweenVectors(defaultDestinationX, defaultDestinationY, defaultDestinationX + 1.0, defaultDestinationY, disk.destinationX, disk.destinationY)

      local directionMultiplier = -1.0
      if (i % 2 == 0) then
        directionMultiplier = 1.0
      end

      disk.deltaAngle = i * 0.01 * directionMultiplier
    end

    local translate = AHLocationTo.new(disk, enterDuration, AHEaseExponentialOut.new(), disk.destinationX, disk.destinationY)
    table.insert(enterActions, translate)

    local fadeIn = AHOpacityTo.new(disk, enterDuration, nil, data.finalOpacity)
    table.insert(enterActions, fadeIn)

    local maximumScale = 1.6
    local scaleUp = AHScaleTo.new(disk, enterDuration, AHEaseExponentialOut.new(), maximumScale)
    table.insert(enterActions, scaleUp)

    local fadeInCheckmark = AHOpacityTo.new(self.checkmark, settleDuration, AHEaseSineIn.new(), 1.0)
    table.insert(settleActions, fadeInCheckmark)

    local scaleDown = AHScaleTo.new(disk, settleDuration, AHEaseExponentialInOut.new(), 1.0)
    table.insert(settleActions, scaleDown)
  end

  local lingeringDisks = {}
  for i = 1, #self.disks do
    local disk = self.disks[i]
    if (not disk.hasCheckmark) then
      table.insert(lingeringDisks, disk)
    end
  end

  local rotateDisks = AHLoop.new(function()
    for _, disk in ipairs(lingeringDisks) do
      local offset = disk.offset
      local angle = disk.angle
      local dx = offset * math.cos(angle)
      local dy = offset * math.sin(angle)
      local x = defaultDestinationX + dx
      local y = defaultDestinationY + dy
      disk:setLoc(x, y)
      disk.angle = disk.angle + disk.deltaAngle
    end
  end)

  local enterAnimation = AHSpawn.new(enterActions)
  local settleAnimation = AHSpawn.new(settleActions)

  -- settle animation starts a bit before enter animation is complete
  -- this helps get just the right scale up/down animation
  local settleAnimationDelay = enterDuration * 0.15
  local delayedSettleAnimation = AHDelayed.new(settleAnimationDelay, settleAnimation)

  local diskRotateDelay = enterDuration * 0.85
  local delayedDiskRotate = AHDelayed.new(diskRotateDelay, rotateDisks)

  local animation = AHSpawn.new(enterAnimation, delayedSettleAnimation, delayedDiskRotate)
  return animation
end

function ContentReviewCardOutro:reset()
  self.button:setOpacity(0.0)
  self.button:setY(self.initialButtonY)

  self.titleLabel:setOpacity(0.0)
  self.titleLabel:setY(self.initialTitleLabelY)

  self.noteLabel:setOpacity(0.0)
  self.noteLabel:setY(self.initialNoteLabelY)

  resetAnimatedItems(self)
  showAnimatedItems(self, false)
end

function ContentReviewCardOutro:onEnter()
  self.exitButton:setTouchEnabled(true)
  self.button:setTouchEnabled(true)

  self.super.onEnter(self)
end

function ContentReviewCardOutro:onLeaveBegin()
  self.exitButton:setTouchEnabled(false)
  self.button:setTouchEnabled(false)

  self.super.onLeaveBegin(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, concept)
  local width = ContentReviewCommon.defaultContentWidth
  local height = Screen.height
  local color = {1.0, 1.0, 1.0, 1.0}
  local content = FilledRectangle.new(width, height, color)

  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_BOLD, 36.0),
    color = ContentReviewCommon.Colors.highlightBlue
  }

  local title = "YOU'RE DONE"
  self.titleLabel = ContentReviewTextView.new(title, width, baseAttributes, nil, nil, Alignment.CENTER)
  self.titleLabel:linkOpacityAttrToChildren(true)
  self.titleLabel:setOpacity(0.0)
  content:insertAtCenter(self.titleLabel)

  local note = "You’ll receive more errors to review as you advance in your training"
  local fontSize = nil
  self.noteLabel = ContentReviewTextViewRationale.new(note, width * 0.9, fontSize, Alignment.CENTER)
  self.noteLabel:linkOpacityAttrToChildren(true)
  self.noteLabel:setOpacity(0.0)
  content:insertAtCenter(self.noteLabel)

  local onTouchUp = function()
    local isRecursive = true
    content:inheritOpacity(self)
    content:linkOpacityAttrToChildren(isRecursive)

    local disksFadeAction = AHSpawn.new()
    for _, disk in ipairs(self.disks) do
      disksFadeAction:addAction(
        AHOpacityTo.new(disk, 0.25, nil, 0.0)
      )
    end

    local exitAction = AHSpawn.new(
      self:getExitAction(),
      disksFadeAction
    )
    exitAction:start()
  end

  local onTouchDown = function()
    self.buttonBackground:setColor(unpack(ContentReviewCommon.Colors.buttonDownBlue))
    local fade = AHColorTo.new(self.buttonBackground, 0.3, nil, ContentReviewCommon.Colors.highlightBlue)
    fade:start()
  end

  local buttonWidth = width * 1.1
  local buttonHeight = buttonWidth * 0.165
  local buttonCornerRadius = buttonWidth * 0.0085
  local buttonColor = ContentReviewCommon.Colors.highlightBlue
  self.buttonBackground = FilledRoundedRectangle.new(buttonWidth, buttonHeight, buttonCornerRadius, buttonColor)
  self.button = Button.makeButton(self.buttonBackground, onTouchDown, onTouchUp)
  self.button:setTouchEnabled(false)
  self.button:setOpacity(0.0)
  content:insertAtCenter(self.button)

  local doneFont = Font.new(FontFaces.DIN_OT_MEDIUM, 20.0)
  local doneLabel = Label.new("Finish review", doneFont)
  self.button:insertAtCenter(doneLabel)

  local verticalSpacing = (Screen.width - buttonWidth) * 0.75
  self.finalButtonY = buttonHeight * 0.5 + verticalSpacing

  if (Screen.hasNotch()) then
    self.finalButtonY = Screen.IPHONE_X_BOTTOM_SAFE_ZONE_HEIGHT + buttonHeight * 0.55
  end

  self.initialButtonY = self.finalButtonY - buttonHeight * 0.4
  self.button:setY(self.initialButtonY)

  self.initialTitleLabelY = Screen.getHeightPart(0.35)
  self.finalTitleLabelY = content:getHeightPart(0.5)
  self.titleLabel:setY(self.initialTitleLabelY)

  self.initialNoteLabelY = Screen.getHeightPart(0.285)
  self.finalNoteLabelY = Screen.getHeightPart(0.385)
  self.noteLabel:setY(self.initialNoteLabelY)

  self:insertAtCenter(content)

  return content
end

-- Constructor ------------------------

function ContentReviewCardOutro.new(concept)
  local self = ContentReviewCard.new()
  Mixin.mix(self, ContentReviewCardOutro, "ContentReviewCard")

  self.disks = {}
  createLayout(self, concept)

  addAnimatedItems(self)
  resetAnimatedItems(self)
  showAnimatedItems(self, false)

  return self
end
