--- ContentReviewCardPromptRationale
-- @author Created by Joe Poka on July 13, 2016.

class("ContentReviewCardPromptRationale")

require("ContentReviewCard")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getUserInputTable(concept)
  local userInputJson = GameSession.answersDatasByConceptIdentifier[concept.identifier]
  local userInputTable = {}
  if (#userInputJson > 0) then
    userInputTable = MOAIJsonParser.decode(userInputJson)[1]
  end
  return userInputTable
end

local function getCardDataDiction(concept, parameters)
  local correctReplace = concept.correct
  local incorrectReplace = concept.incorrect
  local prompt = concept.sentence
  local rationale = concept.tip

  -- insert incorrect answer followed by correct answer
  local startIndex, endIndex = string.find(prompt, "<blank/>")
  local startPiece = prompt:sub(1, startIndex - 1)
  local endPiece = prompt:sub(endIndex + 1)
  prompt = startPiece .. "<strike>" .. incorrectReplace .. "</strike> <highlight>" .. correctReplace .. "</highlight>" .. endPiece

  return prompt, rationale
end

local function getCardDataErrorAvoidance(concept, parameters)
  local word1 = concept.wordOne
  local word2 = concept.wordTwo
  local prompt = concept.sentence
  local rationale = concept.rationale

  local replace1StrikeWord = word2
  local replace1HighlightWord = word1

  local replace2StrikeWord = word1
  local replace2HighlightWord = word2

  local replaceWordStartsPrompt = (prompt[1] == "_")
  if (replaceWordStartsPrompt) then
    replace1StrikeWord = StringUtils.capitalizeFirstLetter(replace1StrikeWord)
    replace1HighlightWord = StringUtils.capitalizeFirstLetter(replace1HighlightWord)

    replace2StrikeWord = string.gsub(replace2StrikeWord, "^%u", string.lower)
    replace2HighlightWord = string.gsub(replace2HighlightWord, "^%u", string.lower)
  end

  local replace1 = "<strike>" .. replace1StrikeWord .. "</strike> <highlight>" .. replace1HighlightWord .. "</highlight>"
  local replace2 = "<strike>" .. replace2StrikeWord .. "</strike> <highlight>" .. replace2HighlightWord .. "</highlight>"

  prompt = string.gsub(prompt, "_", replace1, 1)
  prompt = string.gsub(prompt, "_", replace2)

  return prompt, rationale
end

local function getCardDataBrevity(concept, parameters)
  local prompt = concept.sentence
  local rationale = concept.rationale

  -- remove all <reduce></reduce> pairs after the first pair
  local _, endIndex = string.find(prompt, "</reduce>")
  local afterFirstReduce = string.sub(prompt, endIndex + 1)
  afterFirstReduce = string.gsub(afterFirstReduce, "<reduce>", "")
  afterFirstReduce = string.gsub(afterFirstReduce, "</reduce>", "")
  prompt = string.sub(prompt, 1, endIndex) .. afterFirstReduce

  return prompt, rationale
end

local function getCardDataClarity(concept, parameters)
  local word = concept.replacement
  local prompt = concept.sentence
  local rationale = concept.rationale

  -- insert the correct answer after the reduce part
  local _, reduceEndIndex = string.find(prompt, "</reduce>")
  prompt = prompt:sub(1, reduceEndIndex) .. " <highlight>" .. word .. "</highlight>" .. prompt:sub(reduceEndIndex + 1)

  return prompt, rationale
end

local function getCardDataSyntax(concept, parameters)
  local correct = concept.correct
  local incorrect = concept.incorrect
  local prompt = concept.sentence
  local rationale = concept.tip

  local startIndex, endIndex = string.find(prompt, "<blank/>")
  local startPiece = prompt:sub(1, startIndex - 1)
  local endPiece = prompt:sub(endIndex + 1)

  -- insert incorrect and correct answers
  prompt = startPiece .. "<strike>" .. incorrect .. "</strike> <highlight>" .. correct .. "</highlight>" .. endPiece

  return prompt, rationale
end

local function getCardDataSpelling(concept, parameters)
  local prompt = nil
  local rationale = concept.tip
  return prompt, rationale
end

local function getCardDataExpression(concept, parameters)
  local correct = concept.correctReplace
  local incorrect = concept.incorrectReplace
  local prompt = concept.prompt
  local rationale = concept.rationale
  local taggedReplacement = "<strike>" .. incorrect .. "</strike> <highlight>" .. correct .. "</highlight>"
  prompt = string.gsub(prompt, "<blank/>", taggedReplacement)

  return prompt, rationale
end

local function getCardDataContext(concept)
  local prompt = concept.question
  local correctWord = concept.correct

  local userInputTable = getUserInputTable(concept)
  local selectedWord = userInputTable.user_input
  local correctDefinition = concept.correct_definition

  -- NOTE: `incorrectIndex` defaults to 1 in the event that user input is
  -- out of sync with content.db and no match for `selectedWord` can be found;
  -- the extra `incorrectWord` assignment below is to accommodate this
  local incorrectIndex = 1
  for i = 1, 4 do
    if (concept["incorrect_"..i] == selectedWord) then
      incorrectIndex = i
      break
    end
  end

  local incorrectWord = concept["incorrect_"..incorrectIndex]
  local incorrectDefinition = concept["incorrect_definition_"..incorrectIndex]

  local rationale = "<strong>" .. correctWord .. "</strong><br/>" .. correctDefinition .. "<br/><br/>"
  rationale = rationale .. "<strong>" .. incorrectWord .. "</strong><br/>" .. incorrectDefinition

  local taggedReplacement = "<strike>" .. incorrectWord .. "</strike> <highlight>" .. correctWord .. "</highlight>"
  prompt = string.gsub(prompt, "_", taggedReplacement)

  return prompt, rationale
end

local function getCardDataAgility(concept, parameters)
  local prompt = string.upper(concept.question)
  local userInputTable = getUserInputTable(concept)
  local rationale = ""

  -- Determine if we have a definition for the user_input word
  local function getDefinitionIndex(input)
    local distractors = {concept.distractor1, concept.distractor2, concept.distractor3}
    local distractorIndex = TableUtils.keyForValue(distractors, input)
    return distractorIndex
  end

  if (userInputTable.user_input) then
    local definitionIndex = getDefinitionIndex(userInputTable.user_input)
    local incorrectDefinition = ""
    if (definitionIndex) then
      incorrectDefinition = concept["distractorDefinition" .. definitionIndex]
    end
    if (incorrectDefinition ~= "") then
      rationale = "<strong>" .. userInputTable.user_input .. "</strong><br/>" .. incorrectDefinition .. "<br/><br/>"
    end
  end
  rationale = rationale .. "<strong>" .. concept.answer .. "</strong><br/>" .. concept.gameReportDefinition

  return prompt, rationale
end

local function getCardDataPunctuation(concept)
  local basePrompt = concept.sentence

  -- strip italics to prevent overlapping tags.
  -- revisit this when adding italics support to content review.
  basePrompt = string.gsub(basePrompt, "<i>", "")
  basePrompt = string.gsub(basePrompt, "</i>", "")

  local punctuationMapping = {
    apostrophe = "\'",
    colon = ":",
    comma = ",",
    nothing = "",
    hyphen = "-",
    period = ".",
    question_mark = "?",
    semicolon = ";"
  }

  local userInputTable = getUserInputTable(concept)
  local incorrectPunctuation = punctuationMapping[userInputTable.user_input]

  local correctPunctuationKey = concept.correct_1.correct_answer
  local correctPunctuation = punctuationMapping[correctPunctuationKey]

  local blank = "<blank/>"
  local space = " "

  local wordTable = StringUtils.split(basePrompt, space)
  local formattedPrompt = ""
  for _, word in ipairs(wordTable) do
    local formattedString = word
    local containtsBlank = string.match(word, blank)
    if (containtsBlank) then
      local lastCharacter = ""
      local hasEndQuote = (word[#word] == "\'") or (word[#word] == "\"")
      if (hasEndQuote) then
        lastCharacter = word[#word]
        word = string.sub(word, 1, #word - 1)
      end

      -- if the only distractor is a hyphen and the correct
      -- answer is nothing, then "nothing" means a space
      if (incorrectPunctuation == punctuationMapping.hyphen and
          correctPunctuation == punctuationMapping.nothing) then
        correctPunctuation = space
      end

      local strikeString = "<strike>" .. string.gsub(word, blank, incorrectPunctuation) .. "</strike> "
      local highlightString = "<highlight>" .. string.gsub(word, blank, correctPunctuation) .. "</highlight>"
      formattedString = strikeString .. highlightString .. lastCharacter
    end
    formattedPrompt = formattedPrompt .. formattedString .. space
  end
  formattedPrompt = StringUtils.trim(formattedPrompt)
  local rationale = concept.correct_1.body

  return formattedPrompt, rationale
end

local function getRandomSuggestions(suggestionsList, conceptIdsHash)
  local maxSuggestions = 3
  local selectedSuggestions = {}

  -- we want the shuffle to always produce the same results so the suggesitons
  -- shown do not change between visits. Prime the random function after seeding.
  -- via http://stackoverflow.com/questions/18199844/lua-math-random-not-working
  math.randomseed(conceptIdsHash)
  math.random()
  math.random()
  math.random()

  TableUtils.shuffleArray(suggestionsList)

  for suggestionIndex = 1, maxSuggestions do
    local suggestion = suggestionsList[suggestionIndex]
    if (suggestion ~= nil) then
      table.insert(selectedSuggestions, suggestion)
    end
  end

  return selectedSuggestions
end

local function getShuffledSuggestionsString(suggestionsList, conceptIdsHash)
  local selectedSuggestions = getRandomSuggestions(suggestionsList, conceptIdsHash)
  table.sort(selectedSuggestions)

  local suggestionsString = ""
  for suggestionIndex, suggestion in ipairs(selectedSuggestions) do
    suggestionsString = suggestionsString .. suggestion

    local isLastSuggestion = (suggestionIndex == #selectedSuggestions)
    if (not isLastSuggestion) then
      suggestionsString = suggestionsString .. ", "
    end
  end

  return suggestionsString
end

local function getCardDataAdjectiveRecall(concept, parameters)
  local prompt = string.upper(concept.prompt)
  local suggestionsList = StringUtils.split(concept.suggestions, ", ")

  Asserts.assertType(parameters.conceptIdsHash, "parameters.conceptIdsHash", "number")
  local conceptIdsHash = parameters.conceptIdsHash

  local rationale = "<strong>Suggestions:</strong><br/>" .. getShuffledSuggestionsString(suggestionsList, conceptIdsHash)

  return prompt, rationale
end

local function getCardData(concept, parameters)
  local cardDataGetters = {
    [ContentReviewCommon.GAMES.DICTION] = getCardDataDiction,
    [ContentReviewCommon.GAMES.ERROR_AVOIDANCE] = getCardDataErrorAvoidance,
    [ContentReviewCommon.GAMES.BREVITY] = getCardDataBrevity,
    [ContentReviewCommon.GAMES.CLARITY] = getCardDataClarity,
    [ContentReviewCommon.GAMES.SYNTAX] = getCardDataSyntax,
    [ContentReviewCommon.GAMES.SPELLING] = getCardDataSpelling,
    [ContentReviewCommon.GAMES.EXPRESSION] = getCardDataExpression,
    [ContentReviewCommon.GAMES.ADJECTIVE_RECALL] = getCardDataAdjectiveRecall,
    [ContentReviewCommon.GAMES.CONTEXT] = getCardDataContext,
    [ContentReviewCommon.GAMES.PUNCTUATION] = getCardDataPunctuation,
    [ContentReviewCommon.GAMES.AGILITY] = getCardDataAgility
  }

  local dataFunction = cardDataGetters[ContentReviewCommon.SKILL]
  local prompt, rationale = dataFunction(concept, parameters)
  return prompt, rationale
end

local function getPromptFontSize(concept)
  local fontSizes = {
    [ContentReviewCommon.GAMES.SPELLING] = 40.0,
    [ContentReviewCommon.GAMES.ADJECTIVE_RECALL] = 40.0,
    [ContentReviewCommon.GAMES.AGILITY] = 36.0,
    [ContentReviewCommon.GAMES.PUNCTUATION] = 30.0
  }

  local lookupFontSize = fontSizes[ContentReviewCommon.SKILL]
  local defaultFontSize = 32.0
  local fontSize = lookupFontSize or defaultFontSize

  return fontSize
end

local function getPromptLabel(concept, prompt)
  local promptFontSize = getPromptFontSize(concept)
  local promptWidth = ContentReviewCommon.defaultContentWidth

  local promptLabel = nil
  if (ContentReviewCommon.SKILL == ContentReviewCommon.GAMES.AGILITY) then
    local messageFontSize = 28.0

    local message = "A synonym for <br/><strong>" .. concept.question .. "</strong> is"
    local messageLabel = ContentReviewTextViewRationale.new(message, promptWidth, messageFontSize)
    local messageHeight = messageLabel:getHeight()

    local messageSpacer = messageHeight * 0.5

    local correctWord = "<highlight>" .. string.upper(concept.answer) .. "</highlight>"
    local correctLabel = ContentReviewTextViewPrompt.new(correctWord, promptWidth, promptFontSize)
    local correctHeight = correctLabel:getHeight()

    local promptHeight = correctHeight + messageHeight

    promptLabel = Group.new(promptWidth, promptHeight)
    promptLabel:insertAtCenter(messageLabel)
    promptLabel:insertAtCenter(correctLabel)

    local userInputTable = getUserInputTable(concept)
    if (userInputTable.user_input and userInputTable.user_input ~= "") then
      local incorrectWord = "<strike>" .. string.upper(userInputTable.user_input) .. "</strike>"
      local incorrectLabel = ContentReviewTextViewPrompt.new(incorrectWord, promptWidth, promptFontSize)
      local incorrectHeight = incorrectLabel:getHeight()

      promptHeight = promptHeight + incorrectHeight
      promptLabel:insertAtCenter(incorrectLabel)

      local messageLabelY = promptHeight - messageSpacer
      messageLabel:setY(messageLabelY)

      local incorrectLabelY = messageLabelY - messageSpacer - incorrectHeight
      incorrectLabel:setY(incorrectLabelY)

      local correctLabelY = incorrectLabelY - incorrectHeight * 0.5 - correctLabel:getHeightPart(0.5)
      correctLabel:setY(correctLabelY)
    else
      local messageLabelY = promptHeight - messageSpacer
      messageLabel:setY(messageLabelY)

      local correctLabelY = messageLabelY - messageSpacer - correctHeight
      correctLabel:setY(correctLabelY)
    end
  elseif (ContentReviewCommon.SKILL == ContentReviewCommon.GAMES.SPELLING) then
    local correctWord = "<highlight>" .. string.upper(concept.corr_nomark) .. "</highlight>"
    local incorrectWord = "<strike>" .. string.upper(concept.incorr_nomark1) .. "</strike>"

    local correctLabel = ContentReviewTextViewPrompt.new(correctWord, promptWidth, promptFontSize)
    local incorrectLabel = ContentReviewTextViewPrompt.new(incorrectWord, promptWidth, promptFontSize)

    local correctHeight = correctLabel:getHeight()
    local incorrectHeight = incorrectLabel:getHeight()
    local promptHeight = correctHeight + incorrectHeight

    promptLabel = Group.new(promptWidth, promptHeight)
    promptLabel:insertAtCenter(correctLabel)
    promptLabel:insertAtCenter(incorrectLabel)

    local y = promptHeight - incorrectHeight * 0.5
    incorrectLabel:setY(y)

    y = y - incorrectHeight * 0.5 - correctHeight * 0.5
    correctLabel:setY(y)
  else
    promptLabel = ContentReviewTextViewPrompt.new(prompt, promptWidth, promptFontSize)
  end

  return promptLabel
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ContentReviewCardPromptRationale:createAnimationAction()
  local duration = 0.7
  local ease = AHEaseExponentialOut.new()
  local destinationX = self.rationaleLabel:getWidthPart(0.5)
  local animateIn = AHSequence.new(
    function() self.rationaleLabel:setVisibleRecursive(true) end,
    AHLocationXTo.new(self.rationaleLabel, duration, ease, destinationX)
  )
  return animateIn
end

function ContentReviewCardPromptRationale:reset()
  self.rationaleLabel:setVisibleRecursive(false)
  self.rationaleLabel:setLoc(self.rationaleLabelInitialX, self.rationaleLabelInitialY)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, concept)
  local prompt, rationale = getCardData(concept, self.parameters)

  local defaultContentWidth = ContentReviewCommon.defaultContentWidth
  local defaultContentHeight = defaultContentWidth
  local contentHeight = 0.0

  local promptLabel = getPromptLabel(concept, prompt)
  contentHeight = contentHeight + promptLabel:getHeight()

  local largeVerticalSpacing = defaultContentHeight * 0.15
  contentHeight = contentHeight + largeVerticalSpacing

  local marginWidth = (Screen.width - defaultContentWidth) * 0.5
  local horizontalLineWidth = defaultContentWidth + marginWidth
  local horizontalLineHeight = 2 / Host.contentScale
  local horizontalLine = FilledRectangle.new(horizontalLineWidth, horizontalLineHeight, ContentReviewCommon.Colors.lightGray)
  contentHeight = contentHeight + horizontalLineHeight

  local smallVerticalSpacing = defaultContentHeight * 0.075
  contentHeight = contentHeight + smallVerticalSpacing

  local rationaleWidth = defaultContentWidth
  self.rationaleLabel = ContentReviewTextViewRationale.new(rationale, rationaleWidth)
  contentHeight = contentHeight + self.rationaleLabel:getHeight()

  -- layout items
  local playingSpelling = (ContentReviewCommon.SKILL == ContentReviewCommon.GAMES.SPELLING)
  if (playingSpelling) then
    contentHeight = Screen.getHeightPart(0.35)
  end

  local promptWidth = promptLabel:getWidth()
  local promptHeight = promptLabel:getHeight()
  local x = promptWidth * 0.5
  local y = contentHeight - promptHeight * 0.5
  promptLabel:setLoc(x, y)

  x = horizontalLineWidth * 0.5
  y = y - promptHeight * 0.5 - largeVerticalSpacing - horizontalLineHeight
  horizontalLine:setLoc(x, y)

  local rationaleHeight = self.rationaleLabel:getHeight()
  x = rationaleWidth * 1.5 + marginWidth
  y = y - horizontalLineHeight - smallVerticalSpacing - rationaleHeight * 0.5
  self.rationaleLabelInitialX = x
  self.rationaleLabelInitialY = y
  self.rationaleLabel:setLoc(self.rationaleLabelInitialX, self.rationaleLabelInitialY)

  local content = Group.new(defaultContentWidth, contentHeight)
  content:insert(promptLabel)
  content:insert(horizontalLine)
  content:insert(self.rationaleLabel)

  self:insertAtCenter(content)

  return content
end

-- Constructor ------------------------

function ContentReviewCardPromptRationale.new(concept, parameters)
  local self = ContentReviewCard.new()
  Mixin.mix(self, ContentReviewCardPromptRationale, "ContentReviewCard")

  Asserts.assertType(parameters, "parameters", "table")
  self.parameters = parameters

  createLayout(self, concept)

  return self
end
