--- ContentReviewCardWordParts
-- @author Created by Steven Mah on October 19, 2016

class("ContentReviewCardWordParts")

require("ContentReviewCard")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PART_SPACING = 4
local LINE_WIDTH = 2
local LIGHT_GRAY = ContentReviewCommon.Colors.veryLightGray

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setFontSizes(self, concept)
  if (concept.useSmallFont) then
    self.wordFontSize = 26.0
    self.definitionFontSize = 16.0
  else
    self.wordFontSize = 35.0
    self.definitionFontSize = 18.0
  end
end

local function addBracketAndDefinition(self, concept, wordPartLabel)
  local lineWidth = LINE_WIDTH
  local labelWidth, labelHeight = wordPartLabel:getDims()
  local bracketHeight = labelHeight * 0.2 + lineWidth
  local bracketGroup = Group.new(labelWidth, bracketHeight)

  local bracketBase = FilledRectangle.new(labelWidth, lineWidth, LIGHT_GRAY)
  local bracketLeft = FilledRectangle.new(lineWidth, labelHeight * 0.1, LIGHT_GRAY)
  local bracketRight = FilledRectangle.new(lineWidth, labelHeight * 0.1, LIGHT_GRAY)
  local bracketCenter = FilledRectangle.new(lineWidth, labelHeight * 0.1, LIGHT_GRAY)

  bracketLeft:setAnchorPointBottom()
  bracketRight:setAnchorPointBottom()
  bracketCenter:setAnchorPointTop()

  local centerX = labelWidth * 0.5
  bracketLeft:setLoc(centerX, (bracketHeight + lineWidth) * 0.5)
  bracketRight:setLoc(centerX, (bracketHeight + lineWidth) * 0.5)
  bracketCenter:setLoc(centerX, (bracketHeight - lineWidth) * 0.5)
  bracketBase:setEffectiveWidth(lineWidth)
  bracketGroup:setScl(1.0, 0.0)

  bracketGroup:insertAtCenter(bracketBase)
  bracketGroup:insert(bracketLeft)
  bracketGroup:insert(bracketRight)
  bracketGroup:insert(bracketCenter)
  bracketGroup.left = bracketLeft
  bracketGroup.right = bracketRight
  bracketGroup.center = bracketCenter
  bracketGroup.base = bracketBase

  bracketGroup:setAnchorPointTop()
  bracketGroup:setLoc(wordPartLabel:getWidthPart(0.5), -wordPartLabel:getHeightPart(0.2) + bracketHeight * 0.5)
  wordPartLabel:insert(bracketGroup)
  self.bracketGroup = bracketGroup

  local wordPartDefinitionKey = "definition" .. self.index
  local definition = concept[wordPartDefinitionKey]

  local definitionFontSize = self.definitionFontSize
  local definitionWidth = Screen.getWidthPart(0.3)
  local definitionLabel = ContentReviewTextViewRationale.new(definition, definitionWidth, definitionFontSize, Alignment.CENTER)
  definitionLabel:setLoc(wordPartLabel:getWidthPart(0.5), -wordPartLabel:getHeightPart(0.2) + bracketHeight * 0.5)
  definitionLabel:setAnchorPointTop()
  definitionLabel:linkOpacityAttrToChildren(true)
  definitionLabel:setOpacity(0.0)
  wordPartLabel:insert(definitionLabel)
  self.definitionLabel = definitionLabel
end

local function createWordParts(self, concept)
  local wordPartsRow = Row.new(PART_SPACING)
  local font = Font.new(FontFaces.DIN_OT_BOLD, self.wordFontSize)

  for i = 1, concept.partCount do
    local wordPartKey = "part" .. i
    local wordPartText = string.upper(concept[wordPartKey])
    local label = Label.new(wordPartText, font)
    label:setColor(unpack(LIGHT_GRAY))

    if (i == self.index) then
      self.partLabel = label
      addBracketAndDefinition(self, concept, label)
    end
    wordPartsRow:insert(label)
  end

  wordPartsRow:alignChildren()
  return wordPartsRow
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ContentReviewCardWordParts:createAnimationAction()
  local bracketGroup = self.bracketGroup
  local labelWidth = bracketGroup:getWidth()
  local lineWidth = LINE_WIDTH

  local yAction = AHSpawn.new(
    AHOpacityTo.new(self.definitionLabel, 0.25, nil, 1.0),
    AHScaleYTo.new(bracketGroup, 0.25, AHEaseSineOut.new(), 1.0),
    AHLocationYTo.new(self.definitionLabel, 0.25, AHEaseSineOut.new(), -self.partLabel:getHeightPart(0.4))
  )

  local xAction = AHSpawn.new(
    AHColorTo.new(self.partLabel, 0.5, nil, ContentReviewCommon.Colors.darkGray),
    AHScaleXTo.new(bracketGroup.base, 1.0, AHEaseExponentialOut.new(), 1.0),
    AHLocationXTo.new(bracketGroup.left, 1.0, AHEaseExponentialOut.new(), lineWidth * 0.5),
    AHLocationXTo.new(bracketGroup.right, 1.0, AHEaseExponentialOut.new(), labelWidth - lineWidth * 0.5)
  )

  return AHSequence.new(yAction, xAction)
end

function ContentReviewCardWordParts:reset()
  local bracketGroup = self.bracketGroup
  local centerX = bracketGroup:getWidthPart(0.5)

  bracketGroup.left:setX(centerX)
  bracketGroup.right:setX(centerX)
  bracketGroup.center:setX(centerX)
  bracketGroup.base:setEffectiveWidth(LINE_WIDTH)
  bracketGroup:setScl(1.0, 0.0)

  self.definitionLabel:setOpacity(0.0)
  self.definitionLabel:setY(bracketGroup:getY())
  self.partLabel:setColor(unpack(ContentReviewCommon.Colors.veryLightGray))
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, concept)
  local defaultContentWidth = ContentReviewCommon.defaultContentWidth
  local contentHeightFraction = 0.165
  local contentHeight = Screen.getHeightPart(contentHeightFraction)

  local word = createWordParts(self, concept)
  local wordHeight = word:getHeight()
  local x = defaultContentWidth * 0.5
  local y = contentHeight - wordHeight * 0.5
  word:setLoc(x, y)

  local content = Group.new(defaultContentWidth, contentHeight)
  content:insert(word)
  self.wordLabel = word

  self:insertAtCenter(content)

  return content
end

function ContentReviewCardWordParts.new(concept, parameters)
  local self = ContentReviewCard.new()
  Mixin.mix(self, ContentReviewCardWordParts, "ContentReviewCard")

  setFontSizes(self, concept)
  Asserts.assertType(parameters, "parameters", "table")
  Asserts.assertType(parameters.wordPartIndex, "parameters.wordPartIndex", "number")
  self.index = parameters.wordPartIndex

  createLayout(self, concept)

  return self
end
