--- ContentReviewTextView
-- @author Created by Joe Poka on June 30, 2016.

class("ContentReviewTextView")

require("ContentReviewCommon")
require("graphics.primitives.Polygon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local AttributeName = {
  strike = "strike",
  trimStrikeLeft = "trimStrikeLeft",
  trimStrikeRight = "trimStrikeRight"
}

local COLORS = {
  SPEAKING = {0.98, 0.4, 0.26, 1.0},
  WRITING = {0.269, 0.83, 0.77, 1.0},
  READING = {0.965, 0.29, 0.5, 1.0},
  MATH = {0.46, 0.44, 0.98},
  LISTENING = {0.93, 0.69, 0.31}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getStrikethrough(label, approximateSpaceWidth, strikeColor)
  local width = label:getWidth() + approximateSpaceWidth
  local height = label:getHeight() * 0.045
  local strikethrough = FilledRectangle.new(width, height, strikeColor)
  return strikethrough
end

local function trimStartEndStrikes(self)
  local seenStrike = false
  for i = 1, #self.items do
    local item = self.items[i]
    local hasStrike = item:getAttribute(AttributeName.strike)
    if (hasStrike and not seenStrike) then
      seenStrike = true
      item:setAttribute(AttributeName.trimStrikeLeft, true)
    elseif (not hasStrike and seenStrike) then
      seenStrike = false
      local previousItem = self.items[i - 1]
      previousItem:setAttribute(AttributeName.trimStrikeRight, true)
    end
  end

  local lastItem = self.items[#self.items]
  local lastItemHasStrike = lastItem:getAttribute(AttributeName.strike)
  lastItem:setAttribute(AttributeName.trimStrikeRight, lastItemHasStrike)
end

local function trimInnerStrikes(self)
  for i = 1, #self.items - 1 do
    local currentItem = self.items[i]
    local hasStrike = currentItem:getAttribute(AttributeName.strike)
    if (hasStrike) then
      local currentItemView = self.viewsForItems[currentItem]
      local currentY = currentItemView:getY()

      local nextItem = self.items[i + 1]
      local nextItemView = self.viewsForItems[nextItem]
      local nextY = nextItemView:getY()

      if (nextY < currentY) then
        -- trim strike on item that ends line
        currentItem:setAttribute(AttributeName.trimStrikeRight, true)

        -- trim strike on following item that begins line (if it has a strike)
        local nextHasStrike = nextItem:getAttribute(AttributeName.strike)
        if (nextHasStrike) then
          nextItem:setAttribute(AttributeName.trimStrikeLeft, true)
        end
      end
    end
  end
end

local function addStrikes(self)
  local isSingleStrike = (#self.items == 1)
  for i, item in ipairs(self.items) do
    local hasStrikeThrough = item:getAttribute(AttributeName.strike)
    if (hasStrikeThrough) then
      local itemView = self.viewsForItems[item]
      local text = item:getText()

      local playingBrevity = (self.skillName == ContentReviewCommon.GAMES.BREVITY)
      local playingEloquence = (self.skillName == ContentReviewCommon.GAMES.ELOQUENCE)
      local strikeColor = ContentReviewCommon.Colors.lightGray
      if (playingBrevity) then
        strikeColor = ContentReviewCommon.Colors.strikeRed
      elseif (playingEloquence) then
        strikeColor = ContentReviewCommon.Colors.darkGray
      end
      local approximateSpaceWidth = self:getWidth() * 0.045
      local strikethrough = getStrikethrough(itemView, approximateSpaceWidth, strikeColor)
      itemView:insertAtCenter(strikethrough)

      -- determine how much to trim strikethrough width
      local trimLength = approximateSpaceWidth * 0.25
      local trimDelta = 0.0
      local dx = 0.0

      if (item:getAttribute(AttributeName.trimStrikeLeft)) then
        trimDelta = trimDelta + trimLength
        dx = dx + trimLength
      end

      if (item:getAttribute(AttributeName.trimStrikeRight)) then
        trimDelta = trimDelta + trimLength
        dx = dx - trimLength
      end

      local strikeWidth = strikethrough:getWidth()
      strikethrough:setWidth(strikeWidth - trimDelta)

      local x = strikethrough:getX() + dx
      strikethrough:setX(x)

      local allUppercase = StringUtils.isUppercase(text)
      local hasHyphen = string.match(itemView:getText(), "-") ~= nil
      local fraction = 0.4
      if (allUppercase and isSingleStrike) then
        fraction = 0.43
      end
      if (hasHyphen) then
        fraction = 0.47
      end
      local y = itemView:getHeight() * fraction
      strikethrough:setY(y)
    end
  end
end

local function getHexagonColor()
  local colors = {
    [ContentReviewCommon.GAMES.DICTION] = COLORS.SPEAKING,
    [ContentReviewCommon.GAMES.ERROR_AVOIDANCE] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.BREVITY] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.CLARITY] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.SYNTAX] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.SPELLING] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.AGILITY] = COLORS.READING,
    [ContentReviewCommon.GAMES.MEMORY] = COLORS.SPEAKING,
    [ContentReviewCommon.GAMES.ELOQUENCE] = COLORS.SPEAKING,
    [ContentReviewCommon.GAMES.INVERSION] = COLORS.READING,
    [ContentReviewCommon.GAMES.EXPRESSION] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.WORD_PARTS] = COLORS.READING,
    [ContentReviewCommon.GAMES.VISUALIZATION] = COLORS.READING,
    [ContentReviewCommon.GAMES.ADJECTIVE_RECALL] = COLORS.SPEAKING,
    [ContentReviewCommon.GAMES.ASSOCIATION] = COLORS.READING,
    [ContentReviewCommon.GAMES.CONTEXT] = COLORS.READING,
    [ContentReviewCommon.GAMES.PUNCTUATION] = COLORS.WRITING,
    [ContentReviewCommon.GAMES.SUBTRACTION] = COLORS.MATH
  }

  local color = colors[ContentReviewCommon.SKILL]
  return color
end

local function getHexagon(self)
  local length = Screen.getWidthPart(0.018)
  local angleRadians = math.rad(30.0)
  local lengthSin30 = length * math.sin(angleRadians)
  local lengthCos30 = length * math.cos(angleRadians)

  local hexagonPoints = {
    {x = 0.0, y = 0.0},
    {x = -lengthCos30, y = lengthSin30},
    {x = -lengthCos30, y = lengthSin30 + length},
    {x = 0.0, y = length + 2.0 * lengthSin30},
    {x = lengthCos30, y = lengthSin30 + length},
    {x = lengthCos30, y = lengthSin30}
  }

  local horizontalOffset = Screen.getWidthPart(0.015)
  local verticalOffset = Screen.getHeightPart(0.006)
  for _, point in ipairs(hexagonPoints) do
    point.x = point.x + horizontalOffset
    point.y = point.y + verticalOffset
  end

  local hexagonWidth = Screen.getWidthPart(0.03)
  local hexagonHeight = hexagonWidth
  local hexagonColor = getHexagonColor(self.skillName)
  local hexagon = FilledPolygon.new(hexagonWidth, hexagonHeight, hexagonPoints, hexagonColor)
  return hexagon
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ContentReviewTextView:createViewForTextItem(item)
  local text = item:getText()
  local font = item:getFont()
  local color = item:getAttribute("color")

  local label
  if (text == ContentReviewCommon.hexagonPlaceholder) then
    label = getHexagon(self)
  else
    label = Label.new(text, font)
    if (color) then
      label:setColor(unpack(color))
    end
    local maximumTextWidth = self.layoutBuilder.textWidth
    if (label:getWidth() > maximumTextWidth) then
      label:setEffectiveWidth(maximumTextWidth)
    end
  end

  return label
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ContentReviewTextView.new(xmlText, textWidth, baseAttributes, attributesForLabels, lineSpacing, horizontalAlignment, verticalAlignment)
  assert(baseAttributes, "Must provide baseAttributes")
  attributesForLabels = attributesForLabels or {}

  local self = TextView.new(textWidth, horizontalAlignment, verticalAlignment, lineSpacing)
  Mixin.mix(self, ContentReviewTextView)

  self.skillName = ContentReviewCommon.SKILL

  self.textInterpreter = TextInterpreter.new(baseAttributes, attributesForLabels)
  self.items = self.textInterpreter:interpretXmlText(xmlText)

  -- trim strikes on first/last items in strikethrough block
  trimStartEndStrikes(self)

  local textModel = TextModel.new(self.items)
  self:setModel(textModel)

  -- trim strike on "inner" items that start/end lines
  trimInnerStrikes(self)

  -- create and add strikes
  addStrikes(self)

  return self
end
