--- CrossAnalytics
-- @author Created by andrewcousins on April 13, 2022.

class("CrossAnalytics")

require("Analytics")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local EVENT_CROSSWORD_STARTED = "CrosswordStarted"
local EVENT_CROSSWORD_HINT_USED = "CrosswordHintUsed"
local EVENT_CROSSWORD_AUTOCHECK_TOGGLED = "CrosswordAutocheckToggled"
local EVENT_CROSSWORD_COMPLETED = "CrosswordCompleted"
local EVENT_CROSSWORD_RESET = "CrosswordReset"
local EVENT_CROSSWORD_REVIEWED = "CrosswordReviewed"
local EVENT_CROSSWORD_FEEDBACK = "CrosswordFeedbackAction"
local EVENT_CROSSWORD_DIFFICULTY_SWITCHED = "CrosswordDifficultySwitched"
local EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_SEEN = "CrosswordDifficultyNudgeScreenSeen"
local EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_DIFFICULTY_INCREASED = "CrosswordDifficultyNudgeScreenDifficultyIncreased"
local EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_DIFFICULTY_DECREASED = "CrosswordDifficultyNudgeScreenDifficultyDecreased"

local PROPERTY_PUZZLE_ID = "puzzle_id"
local PROPERTY_PUZZLE_DIFFICULTY = "puzzle_difficulty"

local PROPERTY_PERCENT_COMPLETED = "puzzle_percent_completed"
local PROPERTY_PUZZLE_COMPLETED = "is_completed"
local PROPERTY_IS_FROM_ARCHIVE = "is_from_archive"

local PROPERTY_HINT_WAS_ENTIRE_WORD = "hint_entire_word"
local PROPERTY_HINT_CLUE_IDENTIFIER = "hint_clue_identifier"
local PROPERTY_HINT_WORD_STATE = "hint_word_state"
local PROPERTY_HINT_SELECTED_LETTER = "hint_selected_letter"
local PROPERTY_HINT_CORRECT_WORD = "hint_correct_word"

local PROPERTY_AUTOCHECK_TOGGLED_ON = "autocheck_toggled_on"

local PROPERTY_TIME_FOR_COMPLETION = "time_for_completion"
local PROPERTY_EXPERIENCE_EARNED = "xp_earned"

local PROPERTY_FEEDBACK_RATING = "rating"
local PROPERTY_FEEDBACK_IMPROVEMENT_LIST = "improvements"
local PROPERTY_NUMBER_OF_HINTS_USED = "hints_used"

local PROPERTY_ORIGINAL_PUZZLE_DIFFICULTY = "original_difficulty"
local PROPERTY_UPDATED_PUZZLE_DIFFICULTY = "updated_difficulty"
local PROPERTY_IS_INCREASING_DIFFICULTY = "is_increasing_difficulty"

-- Must be instantiate using CrossAnalytics.instantiate(round)
local shared_instance = nil

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function sharedInstanceExists()
  if (shared_instance == nil) then
    Log.error("NOT REPORTING ANALYTICS. Use CrossAnalytics.instantiate(round) before calling CrossAnalytics methods.")
    return false
  end
  return true
end

local function reportEvent(instance, eventKey, propertiesMapping)
  propertiesMapping = propertiesMapping or {}
  propertiesMapping[PROPERTY_PUZZLE_ID] = instance.roundIdentifier
  propertiesMapping[PROPERTY_PUZZLE_DIFFICULTY] = instance.getCrosswordDifficulty()
  Analytics.reportEvent(eventKey, propertiesMapping)
end

local function getClueIdentifier(clue)
  return clue:getDirection() .. ", " .. clue:getClueNumber()
end

local function buildWordState(instance, clue)
  local clueCellCoordinates = clue:getClueCellCoordinates()
  local wordState = ""
  for _, coordinates in ipairs(clueCellCoordinates) do
    local column, row = unpack(coordinates)
    local cellState = instance.round:getCellState(column, row)
    if cellState:hasContent() then
      wordState = wordState .. cellState:getContent()
    else
      wordState = wordState .. "_"
    end
  end
  return wordState
end

local function getCorrectLetterAtCursor(instance)
  local column, row = unpack(instance.round:getCursorCoordinates())
  return instance.round:getSolutionForCell(column, row)
end

local function getCorrectWordForClue(instance, clue)
  local clueCellCoordinates = clue:getClueCellCoordinates()
  local correctAnswer = ""
  for _, coordinates in ipairs(clueCellCoordinates) do
    local column, row = unpack(coordinates)
    correctAnswer = correctAnswer .. instance.round:getSolutionForCell(column, row)
  end
  return correctAnswer
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossAnalytics.puzzleStarted()
  if sharedInstanceExists() then
    local percentCompleted = shared_instance.round:getPercentCompleted()
    local isCompletedCorreclty = shared_instance.round:isCompletedCorrectly()
    local isFromArchive = shared_instance.round:getIsFromArchive()
    local propertiesMapping = {
      [PROPERTY_PERCENT_COMPLETED] = percentCompleted,
      [PROPERTY_PUZZLE_COMPLETED] = isCompletedCorreclty,
      [PROPERTY_IS_FROM_ARCHIVE] = isFromArchive
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_STARTED, propertiesMapping)
  end
end

function CrossAnalytics.puzzleSingleHintUsed()
  if sharedInstanceExists() then
    local clue = shared_instance.round:getCurrentClue()
    local clueIdentifier = getClueIdentifier(clue)
    local wordState = buildWordState(shared_instance, clue)
    local selectedLetter = getCorrectLetterAtCursor(shared_instance)
    local correctWord = getCorrectWordForClue(shared_instance, clue)
    local percentCompleted = shared_instance.round:getPercentCompleted()
    local propertiesMapping = {
      [PROPERTY_HINT_WAS_ENTIRE_WORD] = false,
      [PROPERTY_HINT_CLUE_IDENTIFIER] = clueIdentifier,
      [PROPERTY_HINT_WORD_STATE] = wordState,
      [PROPERTY_HINT_SELECTED_LETTER] = selectedLetter,
      [PROPERTY_HINT_CORRECT_WORD] = correctWord,
      [PROPERTY_PERCENT_COMPLETED] = percentCompleted
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_HINT_USED, propertiesMapping)
  end
end

function CrossAnalytics.puzzleWordHintUsed()
  if sharedInstanceExists() then
    local clue = shared_instance.round:getCurrentClue()
    local clueIdentifier = getClueIdentifier(clue)
    local wordState = buildWordState(shared_instance, clue)
    local selectedLetter = getCorrectLetterAtCursor(shared_instance)
    local correctWord = getCorrectWordForClue(shared_instance, clue)
    local percentCompleted = shared_instance.round:getPercentCompleted()
    local propertiesMapping = {
      [PROPERTY_HINT_WAS_ENTIRE_WORD] = true,
      [PROPERTY_HINT_CLUE_IDENTIFIER] = clueIdentifier,
      [PROPERTY_HINT_WORD_STATE] = wordState,
      [PROPERTY_HINT_SELECTED_LETTER] = selectedLetter,
      [PROPERTY_HINT_CORRECT_WORD] = correctWord,
      [PROPERTY_PERCENT_COMPLETED] = percentCompleted
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_HINT_USED, propertiesMapping)
  end
end

function CrossAnalytics.puzzleAutocheckToggled()
  if sharedInstanceExists() then
    local autocheckState = shared_instance.round:isAutocheckEnabled()
    local percentCompleted = shared_instance.round:getPercentCompleted()
    local propertiesMapping = {
      [PROPERTY_AUTOCHECK_TOGGLED_ON] = autocheckState,
      [PROPERTY_PERCENT_COMPLETED] = percentCompleted
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_AUTOCHECK_TOGGLED, propertiesMapping)
  end
end

function CrossAnalytics.puzzleCompleted(completionTime, xpEarned)
  if sharedInstanceExists() then
    local propertiesMapping = {
      [PROPERTY_TIME_FOR_COMPLETION] = completionTime,
      [PROPERTY_EXPERIENCE_EARNED] = xpEarned
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_COMPLETED, propertiesMapping)
  end
end

function CrossAnalytics.puzzleReset()
  if sharedInstanceExists() then
    local percentCompleted = shared_instance.round:getPercentCompleted()
    local isCompletedCorreclty = shared_instance.round:isCompletedCorrectly()
    local propertiesMapping = {
      [PROPERTY_PERCENT_COMPLETED] = percentCompleted,
      [PROPERTY_PUZZLE_COMPLETED] = isCompletedCorreclty
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_RESET, propertiesMapping)
  end
end

function CrossAnalytics.puzzleReviewed()
  if sharedInstanceExists() then
    reportEvent(shared_instance, EVENT_CROSSWORD_REVIEWED)
  end
end

function CrossAnalytics.puzzleFeedbackSubmitted(rating, feedbackMap, completionTime)
  if sharedInstanceExists() then
    local hintsUsedCount = shared_instance.round:getNumberOfHintsUsed()
    local propertiesMapping = {
      [PROPERTY_FEEDBACK_RATING] = rating,
      [PROPERTY_TIME_FOR_COMPLETION] = completionTime,
      [PROPERTY_NUMBER_OF_HINTS_USED] = hintsUsedCount
    }
    propertiesMapping = TableUtils.unionSet(propertiesMapping, feedbackMap)
    reportEvent(shared_instance, EVENT_CROSSWORD_FEEDBACK, propertiesMapping)
  end
end

function CrossAnalytics.puzzleDifficultySwitched(originalDifficulty, newDifficulty)
  if sharedInstanceExists() then
    local propertiesMapping = {
      [PROPERTY_ORIGINAL_PUZZLE_DIFFICULTY] = originalDifficulty,
      [PROPERTY_UPDATED_PUZZLE_DIFFICULTY] = newDifficulty
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_DIFFICULTY_SWITCHED, propertiesMapping)
  end
end

function CrossAnalytics.difficultyNudgeScreenSeen(originalDifficulty, completionTime, isIncreasingDifficulty)
  if sharedInstanceExists() then
    local propertiesMapping = {
      [PROPERTY_ORIGINAL_PUZZLE_DIFFICULTY] = originalDifficulty,
      [PROPERTY_TIME_FOR_COMPLETION] = completionTime,
      [PROPERTY_IS_INCREASING_DIFFICULTY] = isIncreasingDifficulty
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_SEEN, propertiesMapping)
  end
end

function CrossAnalytics.difficultyNudgeScreenDifficultyIncreased(originalDifficulty, newDifficulty)
  if sharedInstanceExists() then
    local propertiesMapping = {
      [PROPERTY_ORIGINAL_PUZZLE_DIFFICULTY] = originalDifficulty,
      [PROPERTY_UPDATED_PUZZLE_DIFFICULTY] = newDifficulty
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_DIFFICULTY_INCREASED, propertiesMapping)
  end
end

function CrossAnalytics.difficultyNudgeScreenDifficultyDecreased(originalDifficulty, newDifficulty)
  if sharedInstanceExists() then
    local propertiesMapping = {
      [PROPERTY_ORIGINAL_PUZZLE_DIFFICULTY] = originalDifficulty,
      [PROPERTY_UPDATED_PUZZLE_DIFFICULTY] = newDifficulty
    }
    reportEvent(shared_instance, EVENT_CROSSWORD_DIFFICULTY_NUDGE_SCREEN_DIFFICULTY_DECREASED, propertiesMapping)
  end
end

-- ----------------------------------------------------------------------------
-- Shared Instance Instantiation
-- ----------------------------------------------------------------------------

function CrossAnalytics.instantiate(round)
  shared_instance = {}
  shared_instance.round = round
  shared_instance.roundIdentifier = round:getConceptIdentifier()
  shared_instance.getCrosswordDifficulty = function() return round:getCrosswordDifficulty() end
end
