--- CrossCellDisplay
-- @author Created by andrewcousins on March 07, 2022.

class("CrossCellDisplay")

require("CrossCellState")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OUTLINE_OVERLAP = 1

local SPACING_PERCENT = 0.064
local CORNER_RADIUS_PERCENT = 0.1
local OUTLINE_CORNER_RADIUS_PERCENT = 0.14

local Z_INDEX_UNSELECTED = 10
local Z_INDEX_SELECTED = 20

local WALL_SPRITE_NAME = "wall.png"
local OUTLINE_EDGE_SPRITE_NAME = "highlight_edge.png"
local WRONG_LETTER_SPRITE_NAME = "icon_wrong_letter.png"

local REVEAL_DELAY = 0.05

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateContentLabel(self, content)
  if (content == "") then
    -- This prevents a visual artifact that sometimes appears
    -- when replacing an existing label.
    self.contentLabel:setVisible(false)
  else
    self.contentLabel:setVisible(true)
  end
  self.contentLabel:setText(content)
end

local function playOnTouchSound(self)
  if self.isCursor then
    Sound.playSoundEffect("switch_selected_clue")
  else
    Sound.playSoundEffect("select_square")
  end
end

local function updateCellColors(self)
  if self.isDownstateActive then
    if self.isCursor then
      self.surface:setColor(unpack(CrossColors.CELL_CURSOR_SURFACE_DOWNSTATE))
      self.shadow:setColor(unpack(CrossColors.CELL_CURSOR_SURFACE_DOWNSTATE))
    else
      self.surface:setColor(unpack(CrossColors.CELL_SURFACE_DOWNSTATE))
      self.shadow:setColor(unpack(CrossColors.CELL_SURFACE_DOWNSTATE))
    end
  else
    if self.isCursor then
      self.surface:setColor(unpack(CrossColors.CELL_CURSOR_SURFACE))
      self.shadow:setColor(unpack(CrossColors.CELL_CURSOR_SHADOW))
    else
      self.surface:setColor(unpack(CrossColors.CELL_SURFACE))
      self.shadow:setColor(unpack(CrossColors.CELL_SHADOW))
    end
  end
end

local function animateCellBounce(self)
  local duration = 0.167
  local ease = AHEaseBezier.newFlat({0.4, 0.0, 0.6, 1.0})
  local animation = AHSpawn.new(
    AHSequence.new(
      function() self.surface:setScl(0.95) end,
      AHScaleTo.new(self.surface, duration, ease, 1.02),
      AHScaleTo.new(self.surface, duration, ease, 1.0)
    ),
    AHSequence.new(
      function() self.shadow:setScl(0.95) end,
      AHScaleTo.new(self.shadow, duration, ease, 1.02),
      AHScaleTo.new(self.shadow, duration, ease, 1.0)
    )
  )
  self.cellScaleActionInterrupt:insert(animation)
end

local function playIncorrectAnimation(self)
  if (self.cellState:getIsIncorrect() and (not self.isMarkedAsIncorrect)) then
    self.isMarkedAsIncorrect = true
    local duration = 0.167
    local ease = AHEaseBezier.newFlat({0.0, 0.0, 0.0, 1.0})
    animateCellBounce(self)
    local showIncorrectAnimation = AHSpawn.new(
      function()
        self.incorrectIcon:setScl(0)
        self.incorrectIcon:setOpacity(1)
      end,
      AHScaleTo.new(self.incorrectIcon, duration, ease, self.incorrectIcon.baseScale),
      AHColorTo.new(self.contentLabel, duration, ease, CrossColors.CELL_CHECKED_INCORRECT)
    )
    self.incorrectIconActionInterrupt:insert(showIncorrectAnimation)
  elseif (self.isMarkedAsIncorrect and (not self.cellState:getIsIncorrect())) then
    self.isMarkedAsIncorrect = false
    local duration = 0.167
    local ease = AHEaseBezier.newFlat({0.0, 0.0, 0.0, 1.0})
    animateCellBounce(self)
    local hideIncorrectAnimation = AHSpawn.new(
      AHScaleTo.new(self.incorrectIcon, duration, ease, 0),
      AHColorTo.new(self.contentLabel, duration, ease, CrossColors.CELL_LABEL)
    )
    self.incorrectIconActionInterrupt:insert(hideIncorrectAnimation)
  end
end

local function clearIncorrectIcon(self)
  self.isMarkedAsIncorrect = false
  self.incorrectIcon:setOpacity(0)
  self.contentLabel:setColor(unpack(CrossColors.CELL_LABEL))
end

local function updateIncorrectIcon(self)
  if self.incorrectIconActionInterrupt:isDone() then
    if self.cellState:getIsIncorrect() then
      self.isMarkedAsIncorrect = true
      self.incorrectIcon:setOpacity(1)
      self.incorrectIcon:setScl(self.incorrectIcon.baseScale)
      self.contentLabel:setColor(unpack(CrossColors.CELL_CHECKED_INCORRECT))
    else
      clearIncorrectIcon(self)
    end
  end
end

local function getClearIncorrectAnimation(self)
  local delay = REVEAL_DELAY
  return AHDelayed.new(delay, function() clearIncorrectIcon(self) end)
end

local function getCellRevealedScaleAnimation(self)
  local duration = 0.167
  local ease = AHEaseBezier.newFlat({0.42, 0.00, 0.58, 1.00})
  local animation = AHSpawn.new(
    AHSequence.new(
      function() self.surface:setScl(0.95) end,
      AHScaleTo.new(self.surface, duration, ease, 1.02),
      AHScaleTo.new(self.surface, duration, ease, 1.0)
    ),
    AHSequence.new(
      function() self.shadow:setScl(0.95) end,
      AHScaleTo.new(self.shadow, duration, ease, 1.02),
      AHScaleTo.new(self.shadow, duration, ease, 1.0)
    ),
    function()
      self.contentLabel:setScl(1)
      updateContentLabel(self, self.cellState:getContent())
    end
  )
  return animation
end

local function getRevealedIconAnimation(self)
  local duration = 0.167
  local ease = AHEaseBezier.newFlat({0.0, 0.0, 0.0, 1.0})
  local animation = AHSpawn.new(
    function()
      self.revealedIcon:setEffectiveDims(self.surface:getWidth())
      self.revealedIcon:setLoc(self.surface:getWidthPart(0.5), self.surface:getHeightPart(0.5) - self:getWidthPart(SPACING_PERCENT/2))
      self.revealedIcon:setOpacity(1.0)
      self.surface:setColor(unpack(CrossColors.CELL_REVEALED_ICON))
      self.shadow:setColor(unpack(CrossColors.CELL_REVEALED_ICON))
    end,
    AHDelayed.new(REVEAL_DELAY, AHSpawn.new(
      function() updateCellColors(self) end,
      AHScaleTo.new(self.revealedIcon, duration, ease, 1.0),
      AHLocationTo.new(self.revealedIcon, duration, ease, self.revealedIcon.homeX, self.revealedIcon.homeY)
    ))
  )
  return animation
end

local function updateRevealedIcon(self)
  if self.cellState:getRevealed() then
    self.isRevealed = true
    self.revealedIcon:setOpacity(1)
  else
    self.isRevealed = false
    self.revealedIcon:setOpacity(0)
  end
end

local function waitForRevealAnimation(self)
  if self.cellState:getRevealed() and not self.isRevealed then
    self.isRevealed = true
    local delay = self.cellState:getRevealDelay()
    self.revealAnimation = AHDelayed.new(delay,
      AHSpawn.new(
        getClearIncorrectAnimation(self),
        getCellRevealedScaleAnimation(self),
        getRevealedIconAnimation(self),
        function() Sound.playSoundEffect("hint_select") end
    ))
    self.revealAnimation:start()
  end
  wait(self.revealAnimation)
end

function getAdjacentCells(self)
  return self.adjacentCellsCallback()
end

local function shouldShowOutlineEdge(self, selectedDirection, direction)
  if (selectedDirection ~= false) then
    local clueIsAcross = (selectedDirection == CrossClue.DIRECTION.ACROSS)
    local directionIsHorizontal = CrossDirection.isHorizontal(direction)
    if (clueIsAcross == directionIsHorizontal) then
      local adjacentCells = getAdjacentCells(self)
      local cellDisplay = adjacentCells[direction]
      if (cellDisplay == nil) then
        return true
      elseif cellDisplay:getCellState():getIsWall() then
        return true
      end
    end
  end
  return false
end

local function getOutlineEdgeAnimation(self, edge)
  local duration = 0.167
  local ease = AHEaseBezier.newFlat({0.4, 0.0, 0.6, 1.0})
  local baseScale = self.outlineEdgeScale
  local maxScale = self.outlineEdgeScale * 1.1
  -- Note: always scaling X works because the sprites are rotated
  local animation = AHSequence.new(
    function() edge:setScl(baseScale) end,
    AHScaleXTo.new(edge, duration, ease, maxScale),
    AHScaleXTo.new(edge, duration, ease, baseScale)
  )
  return animation
end

local function updateOutlineEdges(self, selectedDirection)
  local animationTable = {}
  for _, direction in pairs(CrossDirection.DIRECTIONS) do
    local edge = self.outlineEdges[direction]
    if (shouldShowOutlineEdge(self, selectedDirection, direction)) then
      if (edge:getOpacity() < 1) then
        edge:setOpacity(1)
        table.insert(animationTable, getOutlineEdgeAnimation(self, edge, direction))
      end
    else
      edge:setOpacity(0)
    end
  end
  if (#animationTable > 0) then
    self.edgeHighlightAnimationInterrupt:insert(AHSpawn.new(unpack(animationTable)))
  end
end

local function updateOutlineForSelected(self, selectedDirection)
  if (selectedDirection ~= false) then
    self.outline:setOpacity(1)
    self:setZIndex(Z_INDEX_SELECTED)
  else
    self.outline:setOpacity(0)
    self:setZIndex(Z_INDEX_UNSELECTED)
  end
end

local function animateCellForSelected(self)
  if (self.selectedDirection ~= false) then
    animateCellBounce(self)
  end
end

local function animateCellUpstate(self)
  animateCellBounce(self)
end

local function updateCellDisplay(self)
  updateCellColors(self)
  if self.isDownstateActive then
    self.surface:setScl(0.95)
    self.shadow:setScl(0.95)
  elseif (not self.animateTouchUpRequired) then
    self.surface:setScl(1)
    self.shadow:setScl(1)
  end
end

local function playUpdateContentAnimation(self)
  local content = self.cellState:getContent()
  if (self.contentLabel:getText() == content) then
    return
  end
  updateContentLabel(self, content)

  local duration1 = 0.167
  local duration2 = 0.083
  local ease = AHEaseBezier.newFlat({0.4, 0.0, 0.6, 1.0})
  local animation = AHSequence.new(
    function() self.contentLabel:setScl(1.15) end,
    AHScaleTo.new(self.contentLabel, duration1, ease, 0.95),
    AHScaleTo.new(self.contentLabel, duration2, ease, 1.0)
  )
  self.contentAnimationInterrupt:insert(animation)
end

local function updateContent(self)
  local content = self.cellState:getContent()
  if (self.contentLabel:getText() ~= content) then
    updateContentLabel(self, content)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossCellDisplay:getWaveAnimation(isReveal)
  if isReveal then
    self:setScl(0)
  end

  local scaleUpDuration = 0.167
  local scaleUpEase = AHEaseBezier.newFlat({0.3, 0.0, 0.0, 1.0})
  local scaleUpAmount = 1.1
  local scaleUpAnimation = AHScaleTo.new(self, scaleUpDuration, scaleUpEase, scaleUpAmount)

  local scaleDownDuration = 0.167
  local scaleDownEase = AHEaseBezier.newFlat({0.3, 0.0, 0.3, 1.0})
  local scaleDownAnimation = AHScaleTo.new(self, scaleDownDuration, scaleDownEase, 1)

  return AHSequence.new(scaleUpAnimation, scaleDownAnimation)
end

function CrossCellDisplay:getPulseAnimation()
  return AHSpawn.new(
    getCellRevealedScaleAnimation(self),
    function() Sound.playSoundEffect("hint_select") end
  )
end

function CrossCellDisplay:setSelected(clueDirection)
  self.updatedSelectedDirection = clueDirection
end

function CrossCellDisplay:setCursor(isCursor)
  self.isCursor = isCursor
end

function CrossCellDisplay:update(isLoadingState)
  if (self.cellState:getIsWall()) then
    return
  end

  local selectedDirectionHasChanged = (self.selectedDirection ~= self.updatedSelectedDirection)
  self.selectedDirection = self.updatedSelectedDirection

  -- Handle animations
  if (not isLoadingState) then
    waitForRevealAnimation(self)
    playIncorrectAnimation(self)
    playUpdateContentAnimation(self)
    if selectedDirectionHasChanged then
      updateOutlineEdges(self, self.selectedDirection)
      animateCellForSelected(self)
    end
    if self.animateTouchUpRequired then
      animateCellUpstate(self)
      self.animateTouchUpRequired = false
    end
  else
    -- When loading, always update outline edges
    updateOutlineEdges(self, self.selectedDirection)
  end

  -- Update State
  updateContent(self)
  updateRevealedIcon(self)
  updateIncorrectIcon(self)

  if selectedDirectionHasChanged then
    updateOutlineForSelected(self, self.selectedDirection)
  end

  updateCellDisplay(self)
end

function CrossCellDisplay:getCellState()
  return self.cellState
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function setupTouchHandler(self, onTouchCallback, isTouchEnabledCallback)
  local onPressDown = function()
    if ((not self.cellState:getIsWall()) and isTouchEnabledCallback()) then
      self.isDownstateActive = true
      self:update()
    end
  end
  local onPressCancel = function()
    self.isDownstateActive = false
    self:update()
  end
  local onPressUp = function()
    self.isDownstateActive = false
    self.animateTouchUpRequired = true
    self:update()
    if ((not self.cellState:getIsWall()) and isTouchEnabledCallback()) then
      playOnTouchSound(self)
      onTouchCallback()
    end
  end
  Button.makeButton(self, onPressDown, onPressUp, onPressCancel)
end

local function buildOutlineEdges(self, outlineHeight)
  self.outlineEdges = {}
  for _, direction in pairs(CrossDirection.DIRECTIONS) do
    local outlineEdgeSprite = Sprite.newWithName(OUTLINE_EDGE_SPRITE_NAME)
    outlineEdgeSprite:setEffectiveDims(nil, outlineHeight)
    self.outlineEdgeScale = outlineEdgeSprite:getScl()
    self:insertAtCenter(outlineEdgeSprite)
    local offsetAmount = outlineHeight * 0.585
    if (direction == CrossDirection.RIGHT) then
      outlineEdgeSprite:setRot(0)
      outlineEdgeSprite:offsetX(offsetAmount)
    elseif (direction == CrossDirection.LEFT) then
      outlineEdgeSprite:setRot(180)
      outlineEdgeSprite:offsetX(-offsetAmount)
    elseif (direction == CrossDirection.UP) then
      outlineEdgeSprite:setRot(90)
      outlineEdgeSprite:offsetY(offsetAmount)
    elseif (direction == CrossDirection.DOWN) then
      outlineEdgeSprite:setRot(270)
      outlineEdgeSprite:offsetY(-offsetAmount)
    end
    outlineEdgeSprite:setOpacity(0)
    self.outlineEdges[direction] = outlineEdgeSprite
  end
end

local function createLayout(self)
  local width, height = self:getDims()
  local spacing = width * SPACING_PERCENT

  if (self.cellState:getIsWall()) then
    local wallSprite = Sprite.newWithName(WALL_SPRITE_NAME)
    wallSprite:setEffectiveDims(width - spacing, height - spacing)
    self:insertAtCenter(wallSprite)
    return
  end

  local cornerRadius = width * CORNER_RADIUS_PERCENT

  local backgroundWidth = width - spacing
  local backgroundHeight = width - spacing
  local background = FilledRoundedRectangle.new(backgroundWidth, backgroundHeight, cornerRadius, CrossColors.CELL_BACKGROUND)
  self:insertAtCenter(background)
  background:offsetY(-spacing / 2)
  self.background = background


  local outlineWidth = width + OUTLINE_OVERLAP
  local outlineHeight = height + OUTLINE_OVERLAP
  local outlineCornerRadius = width * OUTLINE_CORNER_RADIUS_PERCENT
  local outline = FilledRoundedRectangle.new(outlineWidth, outlineHeight, outlineCornerRadius, CrossColors.CELL_OUTLINE)
  outline:setOpacity(0)
  self:insertAtCenter(outline)
  self.outline = outline

  buildOutlineEdges(self, outlineHeight)

  local surfaceWidth = width - (spacing * 2)
  local shadowHeight = height - (spacing * 2)
  local surfaceHeight = shadowHeight - spacing

  local shadow = FilledRoundedRectangle.new(surfaceWidth, shadowHeight, cornerRadius, CrossColors.CELL_SHADOW)
  self:insertAtCenter(shadow)
  self.shadow = shadow

  local surface = FilledRoundedRectangle.new(surfaceWidth, surfaceHeight, cornerRadius, CrossColors.CELL_SURFACE)
  self:insertAtCenter(surface)
  surface:offsetY(spacing / 2)
  self.surface = surface

  self.labels = {}

  local fontSize = (width * 0.51)
  local font = Font.newWithActualSize(FontFaces.DIN_OT_MEDIUM, fontSize)
  local contentLabel = Label.new("", font)
  contentLabel:setColor(unpack(CrossColors.CELL_LABEL))
  self:insertAtCenter(contentLabel)
  contentLabel:offsetY(spacing / 2)
  self.surface:takeChild(contentLabel)
  self.contentLabel = contentLabel
  table.insert(self.labels, contentLabel)

  local incorrectIconSize = width * 0.125
  local incorrectIcon = Sprite.newWithName(WRONG_LETTER_SPRITE_NAME)
  incorrectIcon:setEffectiveDims(incorrectIconSize)
  self:insertAtLocationFractional(incorrectIcon, 0.22, 0.238)
  incorrectIcon:setOpacity(0)
  self.surface:takeChild(incorrectIcon)
  incorrectIcon.baseScale = incorrectIcon:getScl()
  self.incorrectIcon = incorrectIcon

  local revealedIconSize = width * 0.04
  local revealedIcon = FilledCircle.new(revealedIconSize, CrossColors.CELL_REVEALED_ICON)
  self:insertAtLocationFractional(revealedIcon, 0.8, 0.238)
  revealedIcon:setOpacity(0)
  self.surface:takeChild(revealedIcon)
  revealedIcon.homeX = revealedIcon:getX()
  revealedIcon.homeY = revealedIcon:getY()
  self.revealedIcon = revealedIcon

  if (self.cellState:hasClueNumber()) then
    local clueNumber = self.cellState:getClueNumber()
    local clueIndexFontSize = width * 0.22
    local clueIndexFont = Font.newWithActualSize(FontFaces.DIN_OT_MEDIUM, clueIndexFontSize)
    local clueIndexLabel = Label.new(clueNumber, clueIndexFont)
    clueIndexLabel:setColor(unpack(CrossColors.CELL_CLUE_INDEX_LABEL))
    self:insertAtTopLeft(clueIndexLabel)
    clueIndexLabel:offsetY(spacing * -4)
    clueIndexLabel:offsetX(spacing * 3)
    self.surface:takeChild(clueIndexLabel)
    table.insert(self.labels, clueIndexLabel)
  end
end

-- Constructor ------------------------

function CrossCellDisplay.new(size, cellState, onTouchCallback, isTouchEnabledCallback, adjacentCellsCallback)
  local self = Group.new(size, size)
  Mixin.mix(self, CrossCellDisplay)

  self.cellState = cellState
  self.adjacentCellsCallback = adjacentCellsCallback

  self.isCursor = false
  self.selectedDirection = false
  self.updatedSelectedDirection = false
  self.isDownstateActive = false
  self.animateTouchUpRequired = false
  self.isRevealed = false
  self.isMarkedAsIncorrect = false

  self.contentAnimationInterrupt = ActionInterrupt.new()
  self.edgeHighlightAnimationInterrupt = ActionInterrupt.new()
  self.cellScaleActionInterrupt = ActionInterrupt.new()
  self.incorrectIconActionInterrupt = ActionInterrupt.new()
  self.revealAnimation = nil

  createLayout(self)
  setupTouchHandler(self, onTouchCallback, isTouchEnabledCallback)

  return self
end
