--- CrossClueBar
-- @author Created by andrewcousins on March 08, 2022.

class("CrossClueBar")

require("CrossClueBarArrowButton")
require("text.SizeConstrainedTextArea")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ANIMATE_IN_BOUNCE_AMOUNT = 10

-- BaseGame pauses for 0.4 seconds after startGame is called but before it calls gameFlow.
-- The purpose of the clue label builder coroutine created in startBuildingClueLabels is
-- to space out the work of rendering clue labels across this 0.4 seconds to minimize the
-- impact on the framerate.  Assuming a 5x5 puzzle and 60 FPS, two frames of rest after
-- each frame when a label gets rendered is the maximum amount of staggering that can
-- be done without exceeding 0.4 seconds.  If the coroutine has not finished rendering
-- all clue labels by the time `gameFlow` gets called, then the game will wait for it
-- on a blank screen before presenting the puzzle.
local FRAME_COUNT_TO_WAIT_BETWEEN_CLUE_LABEL_RENDERS = 2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getIntroMoveHeight(self)
  local _, moveHeight = self:getWorldLoc()
  return moveHeight
end

local function showLabelForClue(self, clue, crosswordDifficulty)
  local currentClueLabel = self.displayedClueLabel
  currentClueLabel:setVisibleRecursive(false)
  local nextClueLabel = self.clueLabelsByClueByCrosswordDifficulty[clue][crosswordDifficulty]
  nextClueLabel:setVisibleRecursive(true)
  nextClueLabel:setOpacity(1)

  self.displayedClueLabel = nextClueLabel

  local duration = 0.333
  local ease = AHEaseBezier.newFlat({0.0, 0.0, 0.3, 1.0})
  self.displayedClueLabel:setScl(0.9)
  local animation = AHScaleTo.new(self.displayedClueLabel, duration, ease, 1.0)
  self.clueLabelAnimationInterrupt:insert(animation)
end

local function getCurrentClueText(self)
  if (self.displayedClueLabel == nil) then
    return nil
  end
  return self.displayedClueLabel:getText()
end

local function getClueLabelWidth(self)
  local widthBetweenArrows = (self.rightArrow:getXLeft() - self.leftArrow:getXRight())
  return widthBetweenArrows - self:getHeightPart(0.33)
end

local function buildLabelForClue(self, clue, crosswordDifficulty, labelWidth, maxLabelHeight)
  local fontSize = 18
  local baseAttributes = {font = Font.new(FontFaces.DIN_OT_BOLD, fontSize), color = CrossColors.CLUE_TEXT}
  local labelAttributes = {
    i = {font = Font.new(FontFaces.DIN_OT_BOLD_ITALIC, fontSize), color = CrossColors.CLUE_TEXT},
    flagGB = { src = "flag-gb.png", itemType = TextItem.TYPE_IMAGE, scaleToSize = fontSize},
    flagUS = { src = "flag-us.png", itemType = TextItem.TYPE_IMAGE, scaleToSize = fontSize}
  }
  local minimumFontSize = nil
  local horizontalAlignment = Alignment.CENTER
  local verticalAlignment = Alignment.CENTER
  local lineSpacing = nil
  local clueText = clue:getClueText(crosswordDifficulty)

  local clueLabel = SizeConstrainedTextArea.new(
    clueText,
    labelWidth,
    maxLabelHeight,
    baseAttributes,
    labelAttributes,
    minimumFontSize,
    horizontalAlignment,
    verticalAlignment,
    lineSpacing)

  clueLabel:setVisibleRecursive(false)
  if (not self.clueLabelsByClueByCrosswordDifficulty[clue]) then
    self.clueLabelsByClueByCrosswordDifficulty[clue] = {}
  end
  self.clueLabelsByClueByCrosswordDifficulty[clue][crosswordDifficulty] = clueLabel
  return clueLabel
end

-- Pre-renders all clue labels used in the game.  In order to avoid a large performance hit from rendering all
-- labels in a single frame, uses a coroutine and renders one label every few frames (defined in
-- FRAME_COUNT_TO_WAIT_BETWEEN_CLUE_LABEL_RENDERS).  The coroutine is stored in clueLabelBuilderCoroutine, which
-- other parts of the game can wait on to ensure that gameplay does not begin until all labels are available.
-- This way clue labels do not need to be rendered at runtime, but we can still keep any game intro animations
-- reasonably smooth.
local function prerenderClueLabels(self, crosswordDifficulty)
  return runInCoroutine(function()
    self.renderedClueDifficulties[crosswordDifficulty] = true
    local clues = self.round:getClues()

    local labelWidth = getClueLabelWidth(self)
    local maxLabelHeight = self:getHeightPart(0.778)

    for _, clue in ipairs(clues) do
      local clueLabel = buildLabelForClue(self, clue, crosswordDifficulty, labelWidth, maxLabelHeight)
      self.clueButton:insertAtCenter(clueLabel)
      waitForFrames(FRAME_COUNT_TO_WAIT_BETWEEN_CLUE_LABEL_RENDERS)
    end
  end)
end

local function startBuildingClueLabels(self)
  return runInCoroutine(function()
    local crosswordDifficulty = self.round:getCrosswordDifficulty()
    wait(prerenderClueLabels(self, crosswordDifficulty))

    local currentClue = self.round:getCurrentClue()
    local currentLabel = self.clueLabelsByClueByCrosswordDifficulty[currentClue][crosswordDifficulty]
    currentLabel:setVisibleRecursive(true)
    self.displayedClueLabel = currentLabel
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossClueBar:prepareForIntro()
  self.moveHeight = getIntroMoveHeight(self)
  self:offsetY(-self.moveHeight)
  self.displayedClueLabel:setOpacity(0)
  self.displayedClueLabel:setScl(1.1)
end

function CrossClueBar:getIntroAnimation()
  local popUpDuration = 0.333
  local popUpEase = AHEaseBezier.newFlat({0.0, 0.47, 0.0, 1.0})
  local popUpHeight = self.moveHeight + ANIMATE_IN_BOUNCE_AMOUNT
  local popUpAnimation = AHLocationYBy.new(self, popUpDuration, popUpEase, popUpHeight)

  local settleDownDelay = 0.333
  local settleDownDuration = 0.333
  local settleDownEase = AHEaseBezier.newFlat({0.4, 0.0, 0.8, 1.0})
  local settleDownHeight = -ANIMATE_IN_BOUNCE_AMOUNT
  local settleDownAnimation = AHDelayed.new(settleDownDelay, AHLocationYBy.new(self, settleDownDuration, settleDownEase, settleDownHeight))

  local clueFadeInDelay = 0.333
  local clueFadeInDuration = 0.167
  local clueFadeInEase = AHEaseLinear.new()
  local clueFadeInAnimation = AHDelayed.new(clueFadeInDelay, AHOpacityTo.new(self.displayedClueLabel, clueFadeInDuration, clueFadeInEase, 1))

  local clueScaleDownDelay = 0.333
  local clueScaleDownDuration = 0.333
  local clueScaleDownEase = AHEaseLinear.new()
  local clueScaleDownAnimation = AHDelayed.new(clueScaleDownDelay, AHScaleTo.new(self.displayedClueLabel, clueScaleDownDuration, clueScaleDownEase, 1))

  local arrowAnimationDelay = 0.333
  local arrowAnimation = AHDelayed.new(arrowAnimationDelay, AHSpawn.new(self.leftArrow:getIntroAnimation(), self.rightArrow:getIntroAnimation()))

  return AHSpawn.new(popUpAnimation, settleDownAnimation, clueFadeInAnimation, clueScaleDownAnimation, arrowAnimation)
end

function CrossClueBar:getOutroAnimation()
  local _, worldY = self:getWorldLoc()
  local additionalBufferToSyncWithCrossKeyboard = Screen.getHeightPart(0.08)
  local moveDistance = -(worldY + additionalBufferToSyncWithCrossKeyboard)

  local duration = 0.555
  local ease = AHEaseBezier.newFlat({1.0, 0.0, 0.7, 1.0})
  local moveAnimation = AHLocationYBy.new(self, duration, ease, moveDistance)

  local fadeDuration = CrossCommon.UI_FADE_OUT_DURATION
  local fadeEase = AHEaseLinear.new()
  local clueFade = AHOpacityTo.new(self.displayedClueLabel, fadeDuration, fadeEase, 0)
  local fadeAnimation = AHSpawn.new(
    clueFade,
    self.leftArrow:getFadeOutAnimation(fadeDuration, fadeEase),
    self.rightArrow:getFadeOutAnimation(fadeDuration, fadeEase)
  )

  return AHSpawn.new(moveAnimation, fadeAnimation)
end

function CrossClueBar:prepareForReview()
  wait(self.clueLabelBuilderCoroutine)

  self.displayedClueLabel:setOpacity(0)

  self.leftArrow:prepareForReview()
  self.rightArrow:prepareForReview()
end

function CrossClueBar:showForReview()
  self.displayedClueLabel:setVisibleRecursive(true)
  self.displayedClueLabel:setOpacity(1)

  self.leftArrow:showForReview()
  self.rightArrow:showForReview()
end

function CrossClueBar:updateWithClue(clue)
  local crosswordDifficulty = self.round:getCrosswordDifficulty()
  local newClueText = clue:getClueText(crosswordDifficulty)
  local currentClueText = getCurrentClueText(self)
  if (currentClueText == nil) then
    return
  end

  if (newClueText ~= currentClueText) then
    showLabelForClue(self, clue, crosswordDifficulty)
  end
end

function CrossClueBar:setTouchEnabled(enabled)
  self.touchEnabled = enabled
end

function CrossClueBar:getClueLabelBuilderCoroutine()
  return self.clueLabelBuilderCoroutine
end

function CrossClueBar:renderCluesForDifficulty(crosswordDifficulty)
  if not self.renderedClueDifficulties[crosswordDifficulty] then
    self.clueLabelBuilderCoroutine = prerenderClueLabels(self, crosswordDifficulty)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function addArrowButtons(self, isTouchEnabledCallback)
  local onLeftArrowTouchUp = function()
    if isTouchEnabledCallback() then
      self.round:moveToPreviousClue(self)
    end
  end
  local leftArrow = CrossClueBarArrowButton.new(self:getHeight(), onLeftArrowTouchUp)
  leftArrow:setRot(180)
  self:insertAtLeft(leftArrow)
  leftArrow:offsetX(leftArrow:getWidthPart(0.5))
  self.leftArrow = leftArrow

  local onRightArrowTouchUp = function()
    if isTouchEnabledCallback() then
      self.round:moveToNextClue(self)
    end
  end
  local rightArrow = CrossClueBarArrowButton.new(self:getHeight(), onRightArrowTouchUp)
  self:insertAtRight(rightArrow)
  rightArrow:offsetX(-rightArrow:getWidthPart(0.5))
  self.rightArrow = rightArrow
end

local function addClueLabel(self, isTouchEnabledCallback)
  local widthBetweenArrows = (self.rightArrow:getXLeft() - self.leftArrow:getXRight())
  local labelWidth = widthBetweenArrows - self:getHeightPart(0.33)
  local maxHeight = self:getHeightPart(0.778)

  local clueButtonWidth = (labelWidth + widthBetweenArrows) / 2
  local clueButtonHeight = maxHeight
  local clueButtonCornerRadius = clueButtonHeight * 0.11
  local clueButton = FilledRoundedRectangle.new(clueButtonWidth, clueButtonHeight, clueButtonCornerRadius, CrossColors.CLUE_BACKGROUND)

  self.clueButton = clueButton

  local clueButtonActionInterrupt = ActionInterrupt.new()
  local onTouchDown = function()
    clueButtonActionInterrupt:insert(AHColorTo.new(clueButton, 0, nil, CrossColors.CLUE_DOWNSTATE))
  end
  local onTouchUp = function()
    clueButtonActionInterrupt:insert(AHColorTo.new(clueButton, 0.1, nil, CrossColors.CLUE_BACKGROUND))
    if isTouchEnabledCallback() then
      self.round:moveToAlternateClue(self)
      Sound.playSoundEffect("clue_bar_tap")
    end
  end
  local onTouchCancel = function()
    clueButtonActionInterrupt:insert(AHColorTo.new(clueButton, 0.1, nil, CrossColors.CLUE_BACKGROUND))
  end
  Button.makeButton(clueButton, onTouchDown, onTouchUp, onTouchCancel)
  self:insertAtCenter(clueButton)
end

local function createLayout(self, isTouchEnabledCallback)
  local width, height = self:getDims()
  local backing = FilledRectangle.new(width, height, CrossColors.CLUE_BACKGROUND)
  self:insertAtCenter(backing)

  addArrowButtons(self, isTouchEnabledCallback)
  addClueLabel(self, isTouchEnabledCallback)
end

-- Constructor ------------------------

function CrossClueBar.new(width, height, round, isTouchEnabledCallback)
  local self = Group.new(width, height)
  Mixin.mix(self, CrossClueBar)

  self.round = round

  self.clueLabelAnimationInterrupt = ActionInterrupt.new()
  self.touchEnabled = true
  self.clueLabelsByClueByCrosswordDifficulty = {}
  self.renderedClueDifficulties = {}

  local isTouchEnabledFunction = function()
    return (self.touchEnabled and isTouchEnabledCallback())
  end

  createLayout(self, isTouchEnabledFunction)

  self.clueLabelBuilderCoroutine = startBuildingClueLabels(self)

  return self
end
