--- CrossCompletionScreen
-- @author Created by andrewcousins on April 08, 2022.

class("CrossCompletionScreen")

require("CrossTimeDisplay")
require("CrossPostGameButton")
require("CrossPostGameButtonGroup")
require("CrossCompletionScreenHeader")
require("CrossResetConfirmationScreen")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONGRATS_SUBTITLE_TEXT = Localizer.get("COMPLETION_TIME")
local REVIEW_TEXT = Localizer.get("REVIEW")
local XP_EARNED_TEXT = Localizer.get("XP_EARNED")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConfirmResetCallback(self, resetCallback)
  return function()
    local confirmationScreen = CrossResetConfirmationScreen.new(resetCallback)
    self:insertAtCenter(confirmationScreen)
  end
end

local function setButtonsTouchEnabled(self, enabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(enabled)
  end
end

local function getCallbackWithTouchDisabled(self, callback)
  return function(...)
    setButtonsTouchEnabled(self, false)
    callback(...)
  end
end

local function show(self)
  local animation = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self.labelColumn, 0.167, nil, 1.0),
      AHOpacityTo.new(self.reviewButton, 0.5, nil, 1.0),
      self.postGameButtonGroup:getFadeAnimation(0.5, nil, 1.0),
      self.header:getShowAnimation()
    ),
    function() setButtonsTouchEnabled(self, true) end
  )
  AHDelayed.new(0.517, animation):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createReviewButton(reviewCallback)
  local reviewButton = Sprite.newWithName("back_button.png")
  reviewButton:setEffectiveDims(24)
  local buttonSize = reviewButton:getEffectiveWidthPart(2)
  local onTouchDown = function()
    reviewButton:setOpacity(0.7)
  end
  local onTouchUp = function()
    reviewButton:setOpacity(1)
    reviewButton:setTouchEnabled(false)
    reviewCallback()
  end
  local onTouchCancel = function()
    reviewButton:setOpacity(1)
  end
  Button.addVirtualButton(reviewButton, buttonSize, buttonSize, onTouchDown, onTouchUp, onTouchCancel)
  return reviewButton
end

local function createLabel(text, fontSize, color)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local label = Label.new(text, font)
  label:setColor(unpack(color))
  return label
end

local function createXPLabel(xpValue)
  local fontSize = 18
  local fontColor = CrossColors.COMPLETION_SCREEN_TEXT
  local xpLabel = createLabel(XP_EARNED_TEXT, fontSize, fontColor)
  local xpValueLabel = createLabel(tostring(xpValue), fontSize, fontColor)
  local xpIcon = Sprite.newWithName("xp_icon.png")
  xpIcon:setEffectiveDims(nil, xpLabel:getEffectiveHeightPart(0.85))

  local spacingSize = ReferenceDevice.getScaledWidth(4)

  local row = Row.new(spacingSize * 3)
  row:insert(xpIcon)
  row:insert(xpLabel)
  row:insert(xpValueLabel)
  row:alignChildren()

  local backingWidth = row:getWidth() + (spacingSize * 10)
  local backingHeight = row:getHeight() + (spacingSize * 4)
  local cornerRadius = backingHeight * 0.45
  local backingColor = CrossColors.COMPLETION_SCREEN_XP_BACKGROUND
  local backing = FilledRoundedRectangle.new(backingWidth, backingHeight, cornerRadius, backingColor)

  backing:insertAtCenter(row)
  xpLabel:offsetY(spacingSize / 4)
  xpValueLabel:offsetY(spacingSize / 4)
  return backing
end

local function createTimerLabel(endTime)
  local timerFont = Font.new(FontFaces.DIN_OT_BOLD, 80)
  local timerColor = CrossColors.COMPLETION_SCREEN_TEXT
  local timerLabel = CrossTimeDisplay.new(timerFont, timerColor)
  timerLabel:setDisplayedTime(endTime)
  return timerLabel
end

local function createLayout(self, endTime, xpEarned, continueCallback, reviewCallback, resetCallback)
  local header = CrossCompletionScreenHeader.new()
  header:setAnchorPointBottom()
  self:insertAtLocationFractional(header, 0.5, 0.6)
  self.header = header

  local reviewButton = createReviewButton(reviewCallback)
  table.insert(self.buttons, reviewButton)
  self:insertAtLocation(reviewButton, Screen.getSafeX(0.0) + 20, Screen.getSafeY(1.0) - 20)
  reviewButton:setOpacity(0)
  self.reviewButton = reviewButton

  local labelColumn = Column.new()
  labelColumn:setAnchorPointTop()
  local subtitleLabel = createLabel(CONGRATS_SUBTITLE_TEXT, 24, CrossColors.COMPLETION_SCREEN_SUBTITLE)
  labelColumn:insert(subtitleLabel)
  labelColumn:insertSpacer(-4)
  local timerLabel = createTimerLabel(endTime)
  labelColumn:insert(timerLabel)

  if (xpEarned > 0) then
    labelColumn:insertSpacer(28)
    local xpLabel = createXPLabel(xpEarned)
    labelColumn:insert(xpLabel)
  else
    labelColumn:insertSpacer(40)
    local viewPuzzleButton = CrossPostGameButton.newPillButton(REVIEW_TEXT, reviewCallback, "view_puzzle.png")
    table.insert(self.buttons, viewPuzzleButton)
    labelColumn:insert(viewPuzzleButton)
  end

  labelColumn:alignChildren()
  self:insertAtCenter(labelColumn)
  labelColumn:setY(header:getY() + header:getYOffsetForLabel() + subtitleLabel:getHeightPart(0.5))
  local isRecursive = true
  labelColumn:linkOpacityAttrToChildren(isRecursive)
  labelColumn:setOpacity(0)
  self.labelColumn = labelColumn

  local withContinueButton = true
  local resetConfirmationCallback = getConfirmResetCallback(self, resetCallback)
  local shareCallback = function() GameSession.notifyShareCrossword() end
  local postGameButtonGroup = CrossPostGameButtonGroup.new(shareCallback, resetConfirmationCallback, withContinueButton, continueCallback)
  table.insert(self.buttons, postGameButtonGroup)
  self:insertAtCenter(postGameButtonGroup)

  postGameButtonGroup:setButtonGroupYPosition()

  postGameButtonGroup:setOpacity(0)
  self.postGameButtonGroup = postGameButtonGroup
end

-- Constructor ------------------------

function CrossCompletionScreen.new(endTime, xpEarned, continueCallback, reviewCallback, resetCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossCompletionScreen)

  local continueCallback = getCallbackWithTouchDisabled(self, continueCallback)
  local reviewCallback = getCallbackWithTouchDisabled(self, reviewCallback)
  local resetCallback = getCallbackWithTouchDisabled(self, resetCallback)

  self.buttons = {}

  createLayout(self, endTime, xpEarned, continueCallback, reviewCallback, resetCallback)
  setButtonsTouchEnabled(self, false)
  show(self)

  return self
end
