--- CrossCompletionScreenHeader
-- @author Created by andrewcousins on June 28, 2023.

class("CrossCompletionScreenHeader")

require("graphics.primitives.Checkmark")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DISTANCE_TO_OPACITY = {1, 0.4, 0.25, 0.1}

local CELL_SIZE = 92

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getDistanceForCell(column, row, centerColumn, centerRow)
  return (math.abs(column - centerColumn) + math.abs(row - centerRow))
end

local function getOpacityForDistance(distance)
  local adjustedDistance = distance + 1
  if (adjustedDistance <= #DISTANCE_TO_OPACITY) then
    return DISTANCE_TO_OPACITY[adjustedDistance]
  end
  local remainingDistance = adjustedDistance - #DISTANCE_TO_OPACITY
  local opacity = DISTANCE_TO_OPACITY[#DISTANCE_TO_OPACITY]
  opacity = opacity * (0.5^remainingDistance)
  return opacity
end

local function getCellScaleAnimation(cell)
  local scaleUpDuration = 0.167
  local scaleUpEase = AHEaseBezier.newFlat({0.3, 0.0, 0.0, 1.0})
  local scaleUpScale = cell.baseScale * 1.1
  local scaleUpAnimation = AHScaleTo.new(cell, scaleUpDuration, scaleUpEase, scaleUpScale)

  local scaleDownDuration = 0.167
  local scaleDownEase = AHEaseBezier.newFlat({0.3, 0.0, 0.3, 1.0})
  local scaleDownScale = cell.baseScale
  local scaleDownAnimation = AHScaleTo.new(cell, scaleDownDuration, scaleDownEase, scaleDownScale)

  local scaleAnimation = AHSequence.new(scaleUpAnimation, scaleDownAnimation)
  if (cell.distance == 0) then
    return AHSpawn.new(
      function() cell:setScl(cell.baseScale * 0.4) end,
      scaleAnimation
    )
  else
    return scaleAnimation
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossCompletionScreenHeader:getYOffsetForLabel()
  return CELL_SIZE / 2
end

function CrossCompletionScreenHeader:getShowAnimation()
  local showAnimation = AHSpawn.new()
  for _, cell in ipairs(self.cells) do
    cell:setScl(0.0)
    local delay = 0.067 * cell.distance
    showAnimation:addAction(AHDelayed.new(delay, getCellScaleAnimation(cell)))
  end
  return showAnimation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createCellSprite(cellSize, column, row, centerColumnIndex, centerRowIndex)
  local cellSprite = Sprite.newWithName("header_cell.png")
  cellSprite:setEffectiveDims(cellSize)
  cellSprite.baseScale = cellSprite:getScl()
  cellSprite.distance = getDistanceForCell(column, row, centerColumnIndex, centerRowIndex)
  cellSprite:setOpacity(getOpacityForDistance(cellSprite.distance))

  if (column == centerColumnIndex) and (row == centerRowIndex) then
    local checkmark = FilledCheckmark.new(cellSprite:getWidthPart(0.5), nil, nil, CrossColors.BLACK)
    cellSprite:insertAtCenter(checkmark)
    cellSprite:linkOpacityAttrToChildren(true)
  end

  return cellSprite
end


local function createLayout(self)
  local cellSize = CELL_SIZE
  local cellGroupHeight = Screen.getHeightPart(0.4)
  local cellGroupWidth = Screen.getWidthPart(1.0)

  local cellRowCount = math.ceil(cellGroupHeight / cellSize)
  local cellColumnCount = math.ceil(cellGroupWidth / cellSize)
  if (cellColumnCount % 2 == 0) then
    cellColumnCount = cellColumnCount + 1
  end
  local centerColumnIndex = (cellColumnCount / 2) + 0.5
  local centerRowIndex = 2

  self:setDims(cellGroupHeight, cellGroupWidth)
  for row = 1, cellRowCount do
    for column = 1, cellColumnCount do
      if (row == 1) and (column >= centerColumnIndex - 1) and (column <= centerColumnIndex + 1) then
        -- skip creating this cell
      else
        local cellSprite = createCellSprite(cellSize, column, row, centerColumnIndex, centerRowIndex)
        local cellY = ((row - 1) * cellSize) + (cellSize / 2)
        local cellX = (column - centerColumnIndex) * cellSize + self:getWidthPart(0.5)
        self:insertAtLocation(cellSprite, cellX, cellY)
        table.insert(self.cells, cellSprite)
      end
    end
  end
end

-- Constructor ------------------------

function CrossCompletionScreenHeader.new()
  local self = Group.newMinimumSized()
  Mixin.mix(self, CrossCompletionScreenHeader)

  self.cells = {}

  createLayout(self)

  return self
end
