--- CrossDifficultyNudgeScreen
-- @author Created by andrewcousins on March 06, 2024.

class("CrossDifficultyNudgeScreen")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAXIMUM_SOLVE_TIME_IN_SECONDS_TO_SHOW_INCREASE = 30
local MINIMUM_SOLVE_TIME_IN_SECONDS_TO_SHOW_DECREASE = 240 -- 4 minutes

local MAXIMUM_HINTS_USED_TO_SHOW_INCREASE = 0

local INCREASED_DIFFICULTY_TO_SPRITE_NAME = {
  [CrossRound.CROSSWORD_DIFFICULTY.EASY] = "nudge_medium.png",
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = "nudge_hard.png"
}
local DECREASED_DIFFICULTY_TO_SPRITE_NAME = {
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = "nudge_down_easy.png",
  [CrossRound.CROSSWORD_DIFFICULTY.HARD] = "nudge_down_medium.png"
}
local BACKGROUND_SPRITE_NAME = "nudge_background.png"

local HEADER_INCREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_INCREASE_NUDGE_HEADER")
local DESCRIPTION_INCREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_INCREASE_NUDGE_DESCRIPTION")
local YES_BUTTON_INCREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_INCREASE_NUDGE_YES")
local HEADER_DECREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_DECREASE_NUDGE_HEADER")
local DESCRIPTION_DECREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_DECREASE_NUDGE_DESCRIPTION")
local YES_BUTTON_DECREASE_DIFFICULTY_TEXT = Localizer.get("DIFFICULTY_DECREASE_NUDGE_YES")
local NO_BUTTON_TEXT = Localizer.get("DIFFICULTY_NUDGE_NO")

local DIFFICULTY_INCREASE_BY_DIFFICULTY = {
  [CrossRound.CROSSWORD_DIFFICULTY.EASY] = CrossRound.CROSSWORD_DIFFICULTY.MEDIUM,
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = CrossRound.CROSSWORD_DIFFICULTY.HARD
}
local DIFFICULTY_DECREASE_BY_DIFFICULTY = {
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = CrossRound.CROSSWORD_DIFFICULTY.EASY,
  [CrossRound.CROSSWORD_DIFFICULTY.HARD] = CrossRound.CROSSWORD_DIFFICULTY.MEDIUM
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function doesRoundSupportIncreasedDifficulty(supportedDifficulties, currentDifficulty)
  local increasedDifficulty = DIFFICULTY_INCREASE_BY_DIFFICULTY[currentDifficulty]
  if (increasedDifficulty ~= nil) then
    return TableUtils.contains(supportedDifficulties, increasedDifficulty)
  end
  return false
end

local function doesRoundSupportDecreasedDifficulty(supportedDifficulties, currentDifficulty)
  local decreasedDifficulty = DIFFICULTY_DECREASE_BY_DIFFICULTY[currentDifficulty]
  if (decreasedDifficulty ~= nil) then
    return TableUtils.contains(supportedDifficulties, decreasedDifficulty)
  end
  return false
end

local function isTimeFastEnoughForIncreaseDifficultyScreen(time)
  return time < MAXIMUM_SOLVE_TIME_IN_SECONDS_TO_SHOW_INCREASE
end

local function isValidNumberOfHintsForIncreaseDifficultyScreen(hintsUsed)
  return hintsUsed <= MAXIMUM_HINTS_USED_TO_SHOW_INCREASE
end

local function isTimeSlowEnoughForDecreaseDifficultyScreen(time)
  return time > MINIMUM_SOLVE_TIME_IN_SECONDS_TO_SHOW_DECREASE
end

local function show(self)
  local duration = 0.4
  local ease = AHEaseBezier.newFlat({0.15, 0.00, 0.25, 1.00})
  local animation = AHOpacityTo.new(self.fadeInCover, duration, ease, 0.0)
  animation:start()
end

local function getDismissedCallback(self)
  return function()
    for _, button in ipairs(self.buttons) do
      button:setTouchEnabled(false)
    end
    self.screenDismissedCallback()
  end
end

local function getIncreaseDifficultyCallback(self, difficulty)
  local dismissedCallback = getDismissedCallback(self)
  local increasedDifficulty = DIFFICULTY_INCREASE_BY_DIFFICULTY[difficulty]
  return function()
    GameSession.setCrosswordDifficulty(increasedDifficulty)
    CrossAnalytics.difficultyNudgeScreenDifficultyIncreased(difficulty, increasedDifficulty)
    dismissedCallback()
  end
end

local function getDecreaseDifficultyCallback(self, difficulty)
  local dismissedCallback = getDismissedCallback(self)
  local decreasedDifficulty = DIFFICULTY_DECREASE_BY_DIFFICULTY[difficulty]
  return function()
    GameSession.setCrosswordDifficulty(decreasedDifficulty)
    CrossAnalytics.difficultyNudgeScreenDifficultyDecreased(difficulty, decreasedDifficulty)
    dismissedCallback()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossDifficultyNudgeScreen.shouldShowNudgeScreen(time, currentDifficulty, supportedDifficulties, numberOfHintsUsed)
  local shouldShowNudgeScreen = false
  if (isTimeFastEnoughForIncreaseDifficultyScreen(time) and isValidNumberOfHintsForIncreaseDifficultyScreen(numberOfHintsUsed)) then
    shouldShowNudgeScreen = doesRoundSupportIncreasedDifficulty(supportedDifficulties, currentDifficulty)
  elseif isTimeSlowEnoughForDecreaseDifficultyScreen(time) then
    shouldShowNudgeScreen = doesRoundSupportDecreasedDifficulty(supportedDifficulties, currentDifficulty)
  end

  if shouldShowNudgeScreen then
    ExperimentsConfiguration.CROSSWORDS_DIFFICULTY_NUDGE_EXPERIMENT:reportExperimentExposed()
    if ExperimentsConfiguration.CROSSWORDS_DIFFICULTY_NUDGE_EXPERIMENT:isControl() then
      shouldShowNudgeScreen = false
    end
  end
  return shouldShowNudgeScreen
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createBackground(self)
  local background = FilledRectangle.newFullScreen(CrossColors.BACKGROUND)
  self:insertAtCenter(background)
  local backgroundShapes = Sprite.newWithName(BACKGROUND_SPRITE_NAME)
  backgroundShapes:setEffectiveDims(self:getWidth())
  backgroundShapes:setAnchorPointTop()
  background:insertAtTop(backgroundShapes)
end

local function createPrompt(self, difficulty, isIncreasingDifficulty)
  local spriteName = DECREASED_DIFFICULTY_TO_SPRITE_NAME[difficulty]
  local headerText = HEADER_DECREASE_DIFFICULTY_TEXT
  local descriptionText = DESCRIPTION_DECREASE_DIFFICULTY_TEXT
  if isIncreasingDifficulty then
    spriteName = INCREASED_DIFFICULTY_TO_SPRITE_NAME[difficulty]
    headerText = HEADER_INCREASE_DIFFICULTY_TEXT
    descriptionText = DESCRIPTION_INCREASE_DIFFICULTY_TEXT
  end

  local columnSpacing = 8
  local promptGroup = Column.new(columnSpacing)
  local nudgeIcon = Sprite.newWithName(spriteName)
  promptGroup:insert(nudgeIcon)
  promptGroup:insertSpacer(columnSpacing * 1.5)

  local headerFont = Font.new(FontFaces.DIN_OT_MEDIUM, 24)
  local headerLabel = Label.new(headerText, headerFont)
  promptGroup:insert(headerLabel)

  local descriptionFont = Font.new(FontFaces.DIN_OT_REGULAR, 18)
  local descriptionLabel = Label.new(descriptionText, descriptionFont)
  promptGroup:insert(descriptionLabel)

  promptGroup:alignChildren()
  self:insertAtLocationFractional(promptGroup, 0.5, 0.625)
end

local function createButtons(self, difficulty, isIncreasingDifficulty)
  local buttonColumnSpacing = 9
  local buttonColumn = Column.new(buttonColumnSpacing)

  local yesButtonText = YES_BUTTON_DECREASE_DIFFICULTY_TEXT
  local changeDifficultyCallback = getDecreaseDifficultyCallback(self, difficulty)
  if isIncreasingDifficulty then
    yesButtonText = YES_BUTTON_INCREASE_DIFFICULTY_TEXT
    changeDifficultyCallback = getIncreaseDifficultyCallback(self, difficulty)
  end
  local changeDifficultyButton = CrossPostGameButton.new(yesButtonText, changeDifficultyCallback)
  buttonColumn:insert(changeDifficultyButton)
  table.insert(self.buttons, changeDifficultyButton)

  local notNowButton = CrossPostGameButton.newFloatingInverted(NO_BUTTON_TEXT, getDismissedCallback(self))
  buttonColumn:insert(notNowButton)
  table.insert(self.buttons, notNowButton)

  buttonColumn:alignChildren()
  self:insertAtCenter(buttonColumn)
  buttonColumn:setY(Screen.safeBottom + buttonColumn:getHeightPart(0.5) + (CrossPostGameButton.getButtonHeight() * 0.75))
end


local function createLayout(self, difficulty, isIncreasingDifficulty)
  createBackground(self)
  createPrompt(self, difficulty, isIncreasingDifficulty)
  createButtons(self, difficulty, isIncreasingDifficulty)

  local fadeInCover = FilledRectangle.newFullScreen(CrossColors.BACKGROUND)
  self:insertAtCenter(fadeInCover)
  self.fadeInCover = fadeInCover
end

-- Constructor ------------------------

function CrossDifficultyNudgeScreen.new(completionTime, difficulty, screenDismissedCallback, numberOfHintsUsed)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossDifficultyNudgeScreen)

  self.buttons = {}
  self.screenDismissedCallback = screenDismissedCallback

  local isIncreasingDifficulty = nil
  if (isTimeFastEnoughForIncreaseDifficultyScreen(completionTime) and isValidNumberOfHintsForIncreaseDifficultyScreen(numberOfHintsUsed)) then
    isIncreasingDifficulty = true
  elseif isTimeSlowEnoughForDecreaseDifficultyScreen(completionTime) then
    isIncreasingDifficulty = false
  else
    Log.error("Tried to show the difficulty nudge screen, but neither screen was valid to show.")
    screenDismissedCallback()
    return
  end


  GameSession.setCrosswordNudgeScreenSeen(true)
  CrossAnalytics.difficultyNudgeScreenSeen(difficulty, completionTime, isIncreasingDifficulty)

  createLayout(self, difficulty, isIncreasingDifficulty)

  show(self)
  return self
end
