--- CrossDifficultyTray
-- @author Created by andrewcousins on November 28, 2023.

class("CrossDifficultyTray")

require("CrossTrayButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_ORDER = {
  CrossRound.CROSSWORD_DIFFICULTY.EASY,
  CrossRound.CROSSWORD_DIFFICULTY.MEDIUM,
  CrossRound.CROSSWORD_DIFFICULTY.HARD
}

local DIFFICULTY_ICONS = {
  [CrossRound.CROSSWORD_DIFFICULTY.EASY] = "icon_emoji_easy.png",
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = "icon_emoji_medium.png",
  [CrossRound.CROSSWORD_DIFFICULTY.HARD] = "icon_emoji_hard.png"
}

local DIFFICULTY_TEXT_KEYS = {
  [CrossRound.CROSSWORD_DIFFICULTY.EASY] = "EASY",
  [CrossRound.CROSSWORD_DIFFICULTY.MEDIUM] = "MEDIUM",
  [CrossRound.CROSSWORD_DIFFICULTY.HARD] = "HARD"
}

local OVERLAY_OPACITY = CrossColors.TRAY_OVERLAY[4]

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getSupportedDifficulties(supportedDifficulties)
  local supportedDifficultiesInOrder = {}
  for _, crosswordDifficulty in ipairs(BUTTON_ORDER) do
    if TableUtils.contains(supportedDifficulties, crosswordDifficulty) then
      table.insert(supportedDifficultiesInOrder, crosswordDifficulty)
    end
  end
  return supportedDifficultiesInOrder
end

local function setButtonsTouchEnabled(self, touchEnabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(touchEnabled)
  end
end

local function getDifficultyButtonCallback(self, crosswordDifficulty)
  return function()
    local originalCrosswordDifficulty = self.round:getCrosswordDifficulty()
    -- The analytics call should happen before setting the difficulty
    CrossAnalytics.puzzleDifficultySwitched(originalCrosswordDifficulty, crosswordDifficulty)

    self.round:setCrosswordDifficulty(crosswordDifficulty)
    self:hide()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossDifficultyTray:getIsActive()
  return self.isActive
end

function CrossDifficultyTray:show()
  self.isActive = true

  local duration = 0.333
  local ease = AHEaseBezier.newFlat({0.0, 0.5, 0.0, 1.0})
  local animation = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self.overlay, duration, ease, OVERLAY_OPACITY),
      AHLocationYTo.new(self.tray, duration, ease, self.tray:getHeightPart(0.5)),
      function() self.dismissTouchArea:setTouchEnabled(true) end
    ),
    function() setButtonsTouchEnabled(self, true) end
  )
  Sound.playSoundEffect("hint_tray_open")
  self.actionInterrupt:insert(animation)
end

function CrossDifficultyTray:hide()
  self.isActive = false
  setButtonsTouchEnabled(self, false)
  self.dismissTouchArea:setTouchEnabled(false)

  local duration = 0.333
  local ease = AHEaseBezier.newFlat({1.0, 0.0, 1.0, 1.0})
  local animation = AHSpawn.new(
    AHOpacityTo.new(self.overlay, duration, ease, 0),
    AHLocationYTo.new(self.tray, duration, ease, -self.tray:getHeightPart(0.5)),
    AHDelayed.new(duration/4, function() self.trayDismissedCallback() end)
  )
  Sound.playSoundEffect("hint_tray_close")
  self.actionInterrupt:insert(animation)
end

function CrossDifficultyTray:reset()
  -- Do Nothing
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createCloseButton(self)
  local buttonSize = ReferenceDevice.getScaledHeight(60)
  local dismissFunction = function() self:hide() end

  local closeIcon = Sprite.newWithName("icon_close.png")
  closeIcon:setEffectiveDims(buttonSize/4)
  closeIcon:setAnchorPointTopLeft()

  local closeButton = Button.new(buttonSize, buttonSize, nil, dismissFunction)
  closeButton:insertAtTopLeft(closeIcon)
  local xOffset = self:getWidthPart(0.08)
  local yOffset = -xOffset * 0.833
  closeIcon:offsetX(xOffset)
  closeIcon:offsetY(yOffset)
  return closeButton
end

local function createDifficultyLabel()
  local difficultyText = Localizer.get("DIFFICULTY")
  local difficultyLabelFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local difficultyLabel = Label.new(difficultyText, difficultyLabelFont)
  difficultyLabel:setColor(unpack(CrossColors.TRAY_TEXT))
  return difficultyLabel
end

local function createDescriptionTextArea(width)
  local descriptionText = Localizer.get("DIFFICULTY_DESCRIPTION")
  local descriptionTextAreaFont = Font.new(FontFaces.DIN_OT_REGULAR, 14)
  local descriptionTextArea = DynamicTextArea.newSimple(descriptionText, width, descriptionTextAreaFont, Alignment.CENTER, Alignment.CENTER)
  descriptionTextArea:setColor(unpack(CrossColors.TRAY_DESCRIPTION))
  return descriptionTextArea
end

local function createLayout(self, height)
  local overlay = FilledRectangle.newFullScreen(CrossColors.TRAY_OVERLAY)
  overlay:setOpacity(0)
  self:insertAtCenter(overlay)
  self.overlay = overlay

  local cornerRadius = 12
  local trayWidth = self:getWidth()
  local trayHeight = height + cornerRadius/2
  local trayColor = CrossColors.TRAY
  local sharpCornersList = {FilledRoundedRectangle.CORNERS.BOTTOM_LEFT, FilledRoundedRectangle.CORNERS.BOTTOM_RIGHT}
  local tray = FilledRoundedRectangle.new(trayWidth, trayHeight, cornerRadius, trayColor, sharpCornersList)
  self:insertAtBottom(tray)
  tray:offsetY(-trayHeight/2)
  self.tray = tray

  local dismissTouchAreaWidth = self:getWidth()
  local dismissTouchAreaHeight = self:getHeight() - trayHeight
  dismissTouchArea = Group.new(dismissTouchAreaWidth, dismissTouchAreaHeight)
  Button.makeButton(dismissTouchArea, nil, function() self:hide() end)
  self:insertAtLocation(dismissTouchArea, (dismissTouchAreaWidth / 2), ((dismissTouchAreaHeight / 2) + trayHeight))
  self.dismissTouchArea = dismissTouchArea

  local closeButton = createCloseButton(self)
  tray:insertAtTopLeft(closeButton)
  closeButton:offsetX(closeButton:getWidthPart(0.5))
  closeButton:offsetY(closeButton:getHeightPart(-0.5))
  table.insert(self.buttons, closeButton)

  local difficultyLabel = createDifficultyLabel()
  tray:insertAtTop(difficultyLabel)
  difficultyLabel:offsetY(closeButton:getHeightPart(-0.5))

  local descriptionTextArea = createDescriptionTextArea(trayWidth)
  tray:insertAtCenter(descriptionTextArea)
  descriptionTextArea:setY(difficultyLabel:getYBottom() - descriptionTextArea:getHeightPart(0.5) - difficultyLabel:getHeightPart(0.5))

  local buttonPositioningFraction = 0.29
  local bottomSpacer = trayHeight * 0.05
  local maxHeight = ((descriptionTextArea:getYBottom() - Screen.safeBottom) / (1 + buttonPositioningFraction)) - bottomSpacer
  local buttonSize = math.min(trayWidth * 0.261, maxHeight)

  local buttonSpacing = buttonSize * 0.12
  local buttonRow = Row.new(buttonSpacing)

  local difficultiesInOrder = getSupportedDifficulties(self.round:getSupportedDifficulties())
  for _, crosswordDifficulty in ipairs(difficultiesInOrder) do
    local iconName = DIFFICULTY_ICONS[crosswordDifficulty]
    local iconLabelText = Localizer.get(DIFFICULTY_TEXT_KEYS[crosswordDifficulty])
    local button = CrossTrayButton.newSquare(buttonSize, getDifficultyButtonCallback(self, crosswordDifficulty), iconName, iconLabelText)
    table.insert(self.buttons, button)
    buttonRow:insert(button)
  end

  buttonRow:alignChildren()
  tray:insertAtCenter(buttonRow)
  buttonRow:setY(descriptionTextArea:getYBottom() - buttonRow:getHeightPart(0.5 + buttonPositioningFraction))
end

-- Constructor ------------------------

function CrossDifficultyTray.new(height, round, trayDismissedCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossDifficultyTray)

  self.actionInterrupt = ActionInterrupt.new()
  self.round = round
  self.trayDismissedCallback = trayDismissedCallback
  self.isActive = false

  self.buttons = {}

  createLayout(self, height)

  setButtonsTouchEnabled(self, false)
  self.dismissTouchArea:setTouchEnabled(false)

  return self
end
