--- CrossFeedbackScreen
-- @author Created by andrewcousins on June 07, 2023.

class("CrossFeedbackScreen")

require("CrossFeedbackScreenStarRater")
require("CrossFeedbackScreenFeedbackButtons")
require("CrossPostGameButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SUBMIT_TEXT = Localizer.get("SUBMIT")
local SKIP_TEXT = Localizer.get("SKIP")

local PROMPT_TEXT = Localizer.get("RATING_PROMPT")
local SUBPROMPT_TEXT = Localizer.get("RATING_SUB_PROMPT")
local ADDITIONAL_FEEDBACK_TEXT = Localizer.get("ADDITIONAL_FEEDBACK")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function show(self)
  local duration = 0.4
  local ease = AHEaseBezier.newFlat({0.15, 0.00, 0.25, 1.00})
  local animation = AHOpacityTo.new(self.fadeInCover, duration, ease, 0.0)
  animation:start()
end

local function reportFeedbackSubmitted(self)
  local rating = self.starRater:getRating()
  local feedbackMap = self.feedbackButtons:getFeedback()
  local completionTime = self.completionTime
  CrossAnalytics.puzzleFeedbackSubmitted(rating, feedbackMap, completionTime)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossFeedbackScreen:setTouchEnabled(enabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(enabled)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createPromptLabel(text, fontSize)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local label = Label.new(text, font)
  label:setColor(unpack(CrossColors.FEEDBACK_SCREEN_TEXT))
  return label
end

local function createFeedbackLabel(self)
  local baseAttributes = {font = Font.new(FontFaces.DIN_OT_MEDIUM, 14)}
  local width = self:getWidth()
  local feedbackLabel = DynamicTextArea.new(ADDITIONAL_FEEDBACK_TEXT, width, baseAttributes, nil, Alignment.CENTER)
  return feedbackLabel
end

local function createContinueButton(self)
  local callback = function()
    self:setTouchEnabled(false)
    reportFeedbackSubmitted(self)
    self.dismissedCallback()
  end

  return CrossPostGameButton.new(SUBMIT_TEXT, callback)
end

local function createSkipButton(self)
  local callback = function()
    self:setTouchEnabled(false)
    self.dismissedCallback()
  end

  return CrossPostGameButton.newSimpleTextButton(SKIP_TEXT, callback)
end

local function createLayout(self)
  local background = FilledRectangle.newFullScreen(CrossColors.FEEDBACK_SCREEN_BACKGROUND)
  self:insertAtCenter(background)

  self.buttons = {}

  local column  = Column.new()
  local promptLabel = createPromptLabel(PROMPT_TEXT, 28)
  column:insert(promptLabel)

  local starRater = CrossFeedbackScreenStarRater.new()
  column:insert(starRater)
  table.insert(self.buttons, starRater)
  self.starRater = starRater

  local subpromptLabel = createPromptLabel(SUBPROMPT_TEXT, 18)
  column:insert(subpromptLabel)

  local feedbackButtons = CrossFeedbackScreenFeedbackButtons.new()
  column:insert(feedbackButtons)
  table.insert(self.buttons, feedbackButtons)
  self.feedbackButtons = feedbackButtons

  column:insertSpacer(self:getHeightPart(0.05))
  local additionalFeedbackLabel = createFeedbackLabel(self)
  column:insert(additionalFeedbackLabel)

  column:alignChildren()
  self:insertAtCenter(column)

  local buttonColumn = Column.new()
  local continueButton = createContinueButton(self)
  buttonColumn:insert(continueButton)
  table.insert(self.buttons, continueButton)

  local skipButton = createSkipButton(self)
  buttonColumn:insert(skipButton)
  table.insert(self.buttons, skipButton)
  buttonColumn:alignChildren()

  self:insertAtBottom(buttonColumn)
  buttonColumn:offsetY(Screen.safeBottom + buttonColumn:getHeightPart(0.75))

  column:setY((Screen.safeTop + buttonColumn:getYTop()) / 2)

  local fadeInCover = FilledRectangle.newFullScreen(CrossColors.FEEDBACK_SCREEN_BACKGROUND)
  self:insertAtCenter(fadeInCover)
  self.fadeInCover = fadeInCover
end

-- Constructor ------------------------

function CrossFeedbackScreen.new(dismissedCallback, completionTime)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossFeedbackScreen)

  self.dismissedCallback = dismissedCallback
  -- Completion time is required to send with the analytic event
  self.completionTime = completionTime

  createLayout(self)
  show(self)

  return self
end
