--- CrossGame
-- @author Created by andrewcousins on March 03, 2022.

class("CrossGame")

require("StandardUIBuilder")
require("CrossColors")
require("CrossCommon")
require("CrossDirection")
require("CrossGridDisplay")
require("CrossKeyboard")
require("CrossClueBar")
require("CrossRound")
require("CrossTimer")
require("CrossHelpButton")
require("CrossPauseButton")
require("CrossDifficultyButton")
require("CrossDifficultyTray")
require("CrossHelpTray")
require("CrossRoundGenerator")
require("CrossCompletionScreen")
require("CrossNotYetCompleteScreen")
require("CrossReviewScreen")
require("CrossFeedbackScreen")
require("CrossAnalytics")
require("CrossSaveManager")
require("CrossDifficultyNudgeScreen")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  GRID = 10,
  KEYBOARD = 20,
  TRAY = 10001 -- above the UI zIndex
}

local XP_EARNED = 25 -- Currently all crossword completions earn 25 XP

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getXPEarned()
  if GameSession.getCrosswordIsEarningXP() then
    return XP_EARNED
  end
  return 0
end

local function setTouchInputEnabled(self, enabled)
  self.keyboard:setTouchEnabled(enabled)
  self.grid:setTouchEnabled(enabled)
  self.grid:setClueBarTouchEnabled(enabled)
  for _, widget in ipairs(self.uiOverlay:getWidgets()) do
    if (widget.setTouchEnabled ~= nil) then
      widget:setTouchEnabled(enabled)
    end
  end
end

local function getSetTouchEnabledCallback(self)
  return function(enabled)
    setTouchInputEnabled(self, enabled)
  end
end

local function playIntroAnimation(self)
  self.overlayForIntro:setOpacity(1)
  for _, button in ipairs(self.UIButtons) do
    button:setOpacity(1)
  end
  self.keyboard:prepareForIntro()
  self.grid:prepareForIntro()

  local animation = AHSpawn.new(
    function()
      self.overlayForIntro:setOpacity(0)
    end,
    self.keyboard:getIntroAnimation(),
    self.grid:getIntroAnimation()
  )
  for _, button in ipairs(self.UIButtons) do
    animation:addAction(button:getFadeInAnimation())
  end

  wait(animation:start())
end

local function playCompletionAnimation(self)
  local completionAnimation = self.grid:getCompletionAnimation()

  local outroAnimation = AHSpawn.new(
    self.grid:getOutroAnimation(),
    self.keyboard:getOutroAnimation()
  )
  for _, button in ipairs(self.UIButtons) do
    outroAnimation:addAction(button:getFadeOutAnimation())
  end

  wait(completionAnimation:start())
  outroAnimation:start()
end

local function handleSimpleFadeOut(self)
  local overlayForSimpleFadeOut = FilledRectangle.newFullScreen(CrossColors.BACKGROUND)
  overlayForSimpleFadeOut:setOpacity(0)
  self:insertAtCenter(overlayForSimpleFadeOut)
  overlayForSimpleFadeOut:bringToFront()
  local duration = 0.2
  local ease = nil
  local animation = AHSequence.new(
    AHOpacityTo.new(overlayForSimpleFadeOut, duration, ease, 1),
    AHDelay.new(0.1)
  )
  return animation:start()
end

local function shouldShowDifficultyNudgeScreen(self, loadedCompletedPuzzle)
  if loadedCompletedPuzzle then
    return false
  end
  if GameSession.getCrosswordNudgeScreenSeen() then
    return false
  end
  local difficulty = self.round:getCrosswordDifficulty()
  local supportedDifficulties = self.round:getSupportedDifficulties()
  local completionTime = self.timer:getTime()
  local numberOfHintsUsed = self.round:getNumberOfHintsUsed()
  return CrossDifficultyNudgeScreen.shouldShowNudgeScreen(completionTime, difficulty, supportedDifficulties, numberOfHintsUsed)
end

local function showDifficultyNudgeScreen(self, endGameSessionCallback)
  local completionTime = self.timer:getTime()
  local difficulty = self.round:getCrosswordDifficulty()
  local numberOfHintsUsed = self.round:getNumberOfHintsUsed()
  local difficultyNudgeScreen = CrossDifficultyNudgeScreen.new(completionTime, difficulty, endGameSessionCallback, numberOfHintsUsed)

  self:insertAtCenter(difficultyNudgeScreen)
  difficultyNudgeScreen:bringToFront()
end

local function showFeedbackScreen(self, dismissedCallback)
  local completionTime = self.timer:getTime()
  local feedbackScreen = CrossFeedbackScreen.new(dismissedCallback, completionTime)
  self:insertAtCenter(feedbackScreen)
  feedbackScreen:bringToFront()
end

local function fadeOutAndEndGameSession(self)
  wait(handleSimpleFadeOut(self))
  GameSession.notifyGameWin()
end

local function getEndGameSessionCallback(self, loadedCompletedPuzzle)
  return function()
    local shouldSkipFeedbackScreen = (not self.round:getAskForFeedback())
    if (loadedCompletedPuzzle or shouldSkipFeedbackScreen) then
      fadeOutAndEndGameSession(self)
    else
      showFeedbackScreen(self, function() fadeOutAndEndGameSession(self) end)
    end
  end
end

local function getCompletionScreenDismissedCallback(self, loadedCompletedPuzzle)
  local completionScreenDismissed = false
  loadedCompletedPuzzle = loadedCompletedPuzzle or false
  return function()
    if (not completionScreenDismissed) then
      completionScreenDismissed = true
      local endGameSessionCallback = getEndGameSessionCallback(self, loadedCompletedPuzzle)
      if shouldShowDifficultyNudgeScreen(self, loadedCompletedPuzzle) then
        showDifficultyNudgeScreen(self, endGameSessionCallback)
      else
        endGameSessionCallback()
      end
    end
  end
end

local function getResetRoundCallback(self)
  return function()
    CrossAnalytics.puzzleReset()
    if self.completionScreen then
      self.completionScreen:removeSelf()
    end
    if self.reviewScreen then
      self:takeChild(self.grid)
      self.grid:setY(self.grid.homeY)
      self.reviewScreen:removeSelf()
    end
    self.showingOverlay = false
    self.round:reset()
    self.timer:reset()
    self.helpTray:reset()
    self:gameFlow()
  end
end

local function getReviewRoundCallback(self, endTime, loadedCompletedPuzzle)
  return function(loadedCompletedPuzzle)
    self.round:setSaveEnabled(false)
    local reviewScreen = CrossReviewScreen.new(self.grid, endTime, loadedCompletedPuzzle, getCompletionScreenDismissedCallback(self, loadedCompletedPuzzle), getResetRoundCallback(self))
    self:insertAtCenter(reviewScreen)
    reviewScreen:bringToFront()
    self.reviewScreen = reviewScreen

    if (not loadedCompletedPuzzle) then
      CrossAnalytics.puzzleReviewed()
    end
  end
end

local function getRoundCompletedCallback(self)
  return function(loadedCompletedPuzzle)
    if (not self.showingOverlay) then
      self.showingOverlay = true
      self.timer:pause()
      setTouchInputEnabled(self, false)
      local endTime = self.timer:getTime()

      if loadedCompletedPuzzle then
        local showReviewRoundScreen = getReviewRoundCallback(self, endTime, loadedCompletedPuzzle)
        showReviewRoundScreen(loadedCompletedPuzzle)
      else
        GameSession.markCrosswordAsComplete(self.round:getPuzzleIdentifier(), self.round:getIsFromArchive())
        local xpEarned = getXPEarned()
        CrossAnalytics.puzzleCompleted(endTime, xpEarned)
        Sound.playSoundEffect("game_completed")
        playCompletionAnimation(self)
        local completionScreen = CrossCompletionScreen.new(endTime, xpEarned, getCompletionScreenDismissedCallback(self), getReviewRoundCallback(self, endTime), getResetRoundCallback(self))
        self:insertAtCenter(completionScreen)
        completionScreen:bringToFront()
        self.completionScreen = completionScreen
      end
    end
  end
end

local function getRoundCompletedButIncorrectCallback(self)
  return function()
    if (not self.showingOverlay) then
      self.showingOverlay = true
      self.timer:pause()
      setTouchInputEnabled(self, false)
      local gridCompletedButIncorrectScreen = CrossNotYetCompleteScreen.new(function()
        self.timer:start()
        self.showingOverlay = false
        setTouchInputEnabled(self, true)
      end)
      self:insertAtCenter(gridCompletedButIncorrectScreen)
      gridCompletedButIncorrectScreen:bringToFront()
    end
  end
end

local function getHelpButtonUsedCallback(self)
  return function()
    if self.helpTray:getIsActive() then
      self.helpTray:hide()
    else
      if self.difficultyTray:getIsActive() then
        self.difficultyTray:hide()
        waitFor(0.3)
      end
      self.keyboard:setTouchEnabled(false)
      self.grid:setClueBarTouchEnabled(false)
      self.helpTray:show()
    end
  end
end

local function getDifficultyButtonUsedCallback(self)
  return function()
    if self.difficultyTray:getIsActive() then
      self.difficultyTray:hide()
    else
      if self.helpTray:getIsActive() then
        self.helpTray:hide()
        waitFor(0.3)
      end
      self.keyboard:setTouchEnabled(false)
      self.grid:setClueBarTouchEnabled(false)
      self.difficultyTray:show()
    end
  end
end

local function getTrayDismissedCallback(self)
  return function()
    if (not self.showingOverlay) then
      self.keyboard:setTouchEnabled(true)
      self.grid:setClueBarTouchEnabled(true)
    end
  end
end

local function updateDifficultyButtonLabel(self)
  if (self.difficultyButton ~= nil) then
    self.difficultyButton:updateLabelForDifficulty(self.round:getCrosswordDifficulty())
  end
end

local function getDifficultyTrayDismissedCallback(self)
  local trayDismissedCallback = getTrayDismissedCallback(self)
  return function()
    updateDifficultyButtonLabel(self)
    trayDismissedCallback()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossGame:isUIAutoFadeInEnabled()
  return false
end

function CrossGame:gameFlow()
  wait(self.grid:getClueLabelBuilderCoroutine())

  if (not self.showingOverlay) then
    setTouchInputEnabled(self, false)
    playIntroAnimation(self)
    CrossAnalytics.puzzleStarted()
    self.timer:start()
    setTouchInputEnabled(self, true)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local background = FilledRectangle.new(self:getWidth(), self:getHeight(), CrossColors.BACKGROUND)
  self:insertAtCenter(background)

  local UIBarBottomY = self.UIBottomY

  local keyboard = CrossKeyboard.new(self.round)
  local keyboardHeight = keyboard:getHeight()
  self:insertAtLocation(keyboard, self:getWidthPart(0.5), (keyboardHeight / 2), Z_INDEX.KEYBOARD)
  self.keyboard = keyboard

  local gridWidth = self:getWidth()
  local gridHeight = (UIBarBottomY - keyboardHeight)
  local grid = CrossGridDisplay.new(gridWidth, gridHeight, self.round)
  self:insertAtCenter(grid, Z_INDEX.GRID)
  grid:setY((UIBarBottomY - keyboardHeight) / 2 + keyboardHeight)
  grid.homeY = grid:getY()
  self.grid = grid

  local trayHeight = keyboardHeight + grid:getHeightCoveredByHelpTray()

  local helpTray = CrossHelpTray.new(trayHeight, self.round, getTrayDismissedCallback(self))
  self:insertAtCenter(helpTray, Z_INDEX.TRAY)
  self.helpTray = helpTray

  local difficultyTray = CrossDifficultyTray.new(trayHeight, self.round, getDifficultyTrayDismissedCallback(self))
  self:insertAtCenter(difficultyTray, Z_INDEX.TRAY)
  self.difficultyTray = difficultyTray

  self.overlayForIntro = FilledRectangle.newFullScreen(CrossColors.BACKGROUND)
  self:insertAtCenter(self.overlayForIntro)
  self.overlayForIntro:bringToFront()
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, CrossColors.UI_COMPONENTS)
  local yOffset = -5

  self.UIButtons = {}

  self.timer = CrossTimer.new()

  local UISpacer = CrossUIButton.BUTTON_SIZE * 0.3

  local helpButton = CrossHelpButton.new(getHelpButtonUsedCallback(self), (UISpacer / 2))
  local helpButtonXOffset = -6
  helpButton:setAnchorPointRight()
  builder:addWidget(helpButton, UIOverlay.NORTHEAST, helpButtonXOffset, yOffset)
  table.insert(self.UIButtons, helpButton)

  if (#self.round:getSupportedDifficulties() > 1) then
    local difficultyButton = CrossDifficultyButton.new(self.round:getCrosswordDifficulty(), getDifficultyButtonUsedCallback(self), (UISpacer / 2))
    local difficultyButtonXOffset = helpButtonXOffset - helpButton:getWidthPart(1) - (difficultyButton:getWidthPart(0.5) + UISpacer)
    builder:addWidget(difficultyButton, UIOverlay.NORTHEAST, difficultyButtonXOffset, yOffset)
    table.insert(self.UIButtons, difficultyButton)
    self.difficultyButton = difficultyButton
  end

  local pauseButton = CrossPauseButton.new(self.timer)
  local pauseButtonXOffset = 6
  pauseButton:setAnchorPointLeft()
  builder:addWidget(pauseButton, UIOverlay.NORTHWEST, pauseButtonXOffset, yOffset)
  table.insert(self.UIButtons, pauseButton)

  self.UIBottomY = pauseButton:getYBottom()
end

local function setThresholdsSoAppWillNotCrash()
  GameSession.setScoreThresholds(0, 0)
end

local function loadSavedState(self)
  local saveManager = CrossSaveManager.new(self.round, self.timer, self.helpTray)
  saveManager:load()
end

local function loadAutoCheckSetting(self)
  local isLoadingState = true
  local autoCheckEnabled = GameSession.getCrosswordAutoCheckEnabled()
  self.round:enableAutocheck(autoCheckEnabled, isLoadingState)
  self.helpTray:setAutocheckToggleState(autoCheckEnabled)
end

-- Constructor ------------------------

function CrossGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, CrossGame)

  self.showingOverlay = false

  self.round = CrossRoundGenerator.generateRound(getRoundCompletedCallback(self), getRoundCompletedButIncorrectCallback(self), getSetTouchEnabledCallback(self))
  CrossAnalytics.instantiate(self.round)

  setThresholdsSoAppWillNotCrash()
  createUI(self)
  createGameElements(self)

  loadSavedState(self)
  loadAutoCheckSetting(self)

  return self
end
