--- CrossGridDisplay
-- @author Created by andrewcousins on March 07, 2022.

class("CrossGridDisplay")

require("CrossCellDisplay")
require("CrossClueBar")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isValidCell(columnIndex, rowIndex, columnCount, rowCount)
  local columnIsValid = ((columnIndex > 0) and (columnIndex <= columnCount))
  local rowIsValid = ((rowIndex > 0) and (rowIndex <= rowCount))
  return (columnIsValid and rowIsValid)
end

local function getAdjacentCellStatesCallback(grid, columnIndex, rowIndex, columnCount, rowCount)
  return function()
    local adjacentCells = {}
    for direction, modifiers in pairs(CrossDirection.DIRECTION_TO_MODIFIER) do
      local adjacentColumnIndex = columnIndex + modifiers[1]
      local adjacentRowIndex = rowIndex + modifiers[2]
      if isValidCell(adjacentColumnIndex, adjacentRowIndex, columnCount, rowCount) then
        adjacentCells[direction] = grid[adjacentColumnIndex][adjacentRowIndex]
      end
    end
    return adjacentCells
  end
end

local function clearActiveCells(self)
  for columnIndex = 1, #self.grid do
    for rowIndex = 1, #self.grid[columnIndex] do
      local cell = self.grid[columnIndex][rowIndex]
      cell:setCursor(false)
      cell:setSelected(false)
    end
  end
end

local function updateCells(self)
  for columnIndex = 1, #self.grid do
    for rowIndex = 1, #self.grid[columnIndex] do
      runInCoroutine(function() self.grid[columnIndex][rowIndex]:update(self.isLoadingState) end)
    end
  end
end

local function updateGrid(self)
  clearActiveCells(self)

  local clue = self.round:getCurrentClue()
  local clueDirection = clue:getDirection()
  local clueCellCoordinates = clue:getClueCellCoordinates()
  for _, cellCoordinates in ipairs(clueCellCoordinates) do
    local column, row = unpack(cellCoordinates)
    self.grid[column][row]:setSelected(clueDirection)
  end

  local cursorColumn, cursorRow = unpack(self.round:getCursorCoordinates())
  self.grid[cursorColumn][cursorRow]:setCursor(true)

  updateCells(self)

  self.clueBar:updateWithClue(clue)
end

local function getIsTouchEnabledCallback(self)
  return function()
    return self.touchEnabled
  end
end

local function getCellDisplaysByDiagonal(self)
  local diagonals = {}
  for columnIndex = 1, #self.grid do
    for rowIndex = 1, #self.grid[columnIndex] do
      local cell = self.grid[columnIndex][rowIndex]
      local diagonalIndex = rowIndex + columnIndex - 1
      if (diagonals[diagonalIndex] == nil) then
        diagonals[diagonalIndex] = {}
      end
      table.insert(diagonals[diagonalIndex], cell)
    end
  end
  return diagonals
end

local function getGridWaveAnimation(self, isReveal)
  local stackingDelay = 0.067

  local animation = AHSpawn.new()
  for i, diagonal in ipairs(getCellDisplaysByDiagonal(self)) do
    local delay = (i-1) * stackingDelay
    for _, cell in ipairs(diagonal) do
      animation:addAction(AHDelayed.new(delay, cell:getWaveAnimation(isReveal)))
    end
  end
  return animation
end

local function getClueBarYPositionForReview(self)
  local effectiveYBottom = self.grid[1][#self.grid[1]]:getY() + (self.cellSize / 2)
  local clueBarYOffset = self.cellSize / 2
  local idealClueBarYPosition = effectiveYBottom - (clueBarYOffset + self.clueBar:getHeightPart(0.5))
  local minimumClueBarYPosition = self.clueBar:getHeightPart(0.5)

  return math.max(minimumClueBarYPosition, idealClueBarYPosition)
end

local function getClueBarYPositionForIntro(self)
  return self.clueBar:getHeightPart(0.5)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossGridDisplay:prepareForIntro()
  self.gridGroup:setScl(1)
  self.gridGroup:setY(self.gridGroup.homeY)
  self.clueBar:setY(getClueBarYPositionForIntro(self))
  self.clueBar:prepareForIntro()
end

function CrossGridDisplay:getIntroAnimation()
  local isReveal = true
  local animation = AHSpawn.new(
    function() Sound.playSoundEffect("board_fill_in") end,
    self.clueBar:getIntroAnimation(),
    getGridWaveAnimation(self, isReveal)
  )

  return animation
end

function CrossGridDisplay:getOutroAnimation()
  local duration = 0.25
  local scaleUpEase = AHEaseBezier.newFlat({0.4, 0.0, 0.6, 1.0})
  local scaleUpAnimation = AHScaleTo.new(self.gridGroup, duration, scaleUpEase, 1.05)

  local scaleDownEase = AHEaseBezier.newFlat({1.0, 0.0, 0.6, 1.0})
  local scaleDownAnimation = AHScaleTo.new(self.gridGroup, duration, scaleDownEase, 0.1)
  local crossCompletionScreenCheckmarkYCoordinate = Screen.getHeightPart(0.6) + (82 * 1.5) -- based on layout of CrossCompletionScreen
  local _, moveToYLocation = self:worldToModel(nil, crossCompletionScreenCheckmarkYCoordinate)
  local moveToCheckmarkAnimation = AHLocationYTo.new(self.gridGroup, duration, scaleDownEase, moveToYLocation)
  local moveAndScaleDownAnimation = AHSpawn.new(scaleDownAnimation, moveToCheckmarkAnimation)

  local hideAnimation = AHDelayed.new(0.017, function() self.gridGroup:setScl(0) end)

  local gridAnimation = AHSequence.new(
    scaleUpAnimation,
    moveAndScaleDownAnimation,
    hideAnimation
  )

  local clueBarOutroAnimation = self.clueBar:getOutroAnimation()
  return AHSpawn.new(gridAnimation, clueBarOutroAnimation)
end

function CrossGridDisplay:getCompletionAnimation()
  local isReveal = false
  local animation = AHSequence.new(
    getGridWaveAnimation(self, isReveal),
    AHDelay.new(0.5)
  )
  return animation
end

function CrossGridDisplay:getWordPulseAnimation(cellCoordinates)
  local stackingDelay = 0.067

  local animation = AHSpawn.new()
  for i, cellCoordinate in ipairs(cellCoordinates) do
    local delay = (i-1) * stackingDelay
    local columnIndex, rowIndex = unpack(cellCoordinate)
    local cell = self.grid[columnIndex][rowIndex]
    animation:addAction(AHDelayed.new(delay, cell:getPulseAnimation()))
  end
  return animation
end

function CrossGridDisplay:prepareForReview()
  self.gridGroup:setY(self.gridGroup.homeY)
  self.gridGroup:setScl(0)

  self.clueBar:setY(getClueBarYPositionForReview(self))
  self.clueBar:prepareForReview()
end

function CrossGridDisplay:showForReview()
  Sound.playSoundEffect("board_fill_in")
  getGridWaveAnimation(self, true):start()
  self.gridGroup:setScl(1)
  self.clueBar:showForReview()
end

function CrossGridDisplay:update()
  updateGrid(self)
end

function CrossGridDisplay:updateCell(column, row)
  self.grid[column][row]:update(self.isLoadingState)
end

function CrossGridDisplay:getHeightCoveredByHelpTray()
  local topOfClueBar = self.clueBar:getHeight()
  local bottomOfGridCells = self.gridGroup:getY() - (self.cellSize * (#self.grid[1] + 1) / 2)
  return (topOfClueBar + bottomOfGridCells) * (2/3)
end

function CrossGridDisplay:getEffectiveYTop()
  return self:getY() + (self.cellSize * (#self.grid[1] + 1) / 2)
end

function CrossGridDisplay:setTouchEnabled(enabled)
  self.touchEnabled = enabled
end

function CrossGridDisplay:setClueBarTouchEnabled(enabled)
  self.clueBar:setTouchEnabled(enabled)
end

function CrossGridDisplay:setLoadingState(isLoading)
  self.isLoadingState = isLoading
end

function CrossGridDisplay:getClueLabelBuilderCoroutine()
  return self.clueBar:getClueLabelBuilderCoroutine()
end

function CrossGridDisplay:updateClueDifficulties(crosswordDifficulty)
  self.clueBar:renderCluesForDifficulty(crosswordDifficulty)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function addClueBar(self, round)
  local clueBarWidth = self:getWidth()
  local clueBarHeight = math.min(Screen.getHeightPart(0.08), clueBarWidth * 0.2)
  local clueBar = CrossClueBar.new(clueBarWidth, clueBarHeight, round, getIsTouchEnabledCallback(self))
  self:insertAtBottom(clueBar)
  clueBar:offsetY(clueBar:getHeightPart(0.5))
  self.clueBar = clueBar
end

local function createLayout(self, round)
  addClueBar(self, round)

  local columnCount = round:getColumnCount()
  local rowCount = round:getRowCount()
  local gridWidth = self:getWidth()
  local gridHeight = self:getHeight() - self.clueBar:getHeight()

  local grid = {}
  local gridGroup = Group.new(gridWidth, gridHeight)

  local horizontalColumnProportionBuffer = 0.8
  local verticalRowProportionBuffer = 1.0
  local maxHorizontalCellSize = (gridWidth / (columnCount + horizontalColumnProportionBuffer))
  local maxVerticalCellSize = (gridHeight / (rowCount + verticalRowProportionBuffer))
  local cellSize = math.min(maxHorizontalCellSize, maxVerticalCellSize)
  self.cellSize = cellSize

  for columnIndex = 1, columnCount do
    grid[columnIndex] = {}
    for rowIndex = 1, rowCount do
      local onCellTouchCallback = function()
        self.round:cellTouchCallback(columnIndex, rowIndex)
      end
      local adjacentCellsCallback = getAdjacentCellStatesCallback(grid, columnIndex, rowIndex, columnCount, rowCount)
      local cellState = round:getCellState(columnIndex, rowIndex)
      local cell = CrossCellDisplay.new(cellSize, cellState, onCellTouchCallback, getIsTouchEnabledCallback(self), adjacentCellsCallback)
      local cellX = (gridWidth / 2) + ((columnIndex - (columnCount / 2) - 0.5) * cellSize)
      local cellY = (gridHeight / 2) + (((rowCount / 2) - rowIndex + 0.5) * cellSize)
      cell:setLoc(cellX, cellY)
      gridGroup:insert(cell)
      grid[columnIndex][rowIndex] = cell
    end
  end

  self.grid = grid
  self:insertAtCenter(gridGroup, 0.5)
  gridGroup:setY((gridHeight / 2) + self.clueBar:getHeight())
  gridGroup.homeY = gridGroup:getY()
  self.gridGroup = gridGroup
end

-- Constructor ------------------------

function CrossGridDisplay.new(width, height, round)
  local self = Group.new(width, height)
  Mixin.mix(self, CrossGridDisplay)

  self.round = round
  self.touchEnabled = true
  self.isLoadingState = false
  createLayout(self, round)

  round:setGridDisplay(self)

  updateGrid(self)

  return self
end
