--- CrossHelpTray
-- @author Created by andrewcousins on April 05, 2022.

class("CrossHelpTray")

require("CrossTrayButton")
require("CrossTrayToggleButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OVERLAY_OPACITY = CrossColors.TRAY_OVERLAY[4]

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setButtonsTouchEnabled(self, touchEnabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(touchEnabled)
  end
end

local function getRevealLetterCallback(self)
  return function()
    self.round:revealLetter()
    self:hide()
  end
end

local function getRevealWordCallback(self)
  return function()
    self.round:revealWord()
    self:hide()
  end
end

local function getCheckForErrorsCallback(self)
  return function()
    self.round:checkForErrors()
    self:hide()
  end
end

local function getToggleAutocheckCallback(self, checkForErrorsButton)
  return function(enabled)
    self.round:enableAutocheck(enabled)
    checkForErrorsButton.setIsActive(not enabled)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossHelpTray:getIsActive()
  return self.isActive
end

function CrossHelpTray:show()
  self.isActive = true

  local duration = 0.333
  local ease = AHEaseBezier.newFlat({0.0, 0.5, 0.0, 1.0})
  local animation = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self.overlay, duration, ease, OVERLAY_OPACITY),
      AHLocationYTo.new(self.tray, duration, ease, self.tray:getHeightPart(0.5)),
      function() self.dismissTouchArea:setTouchEnabled(true) end
    ),
    function() setButtonsTouchEnabled(self, true) end
  )
  Sound.playSoundEffect("hint_tray_open")
  self.actionInterrupt:insert(animation)
end

function CrossHelpTray:hide()
  self.isActive = false
  setButtonsTouchEnabled(self, false)
  self.dismissTouchArea:setTouchEnabled(false)

  local duration = 0.333
  local ease = AHEaseBezier.newFlat({1.0, 0.0, 1.0, 1.0})
  local animation = AHSpawn.new(
    AHOpacityTo.new(self.overlay, duration, ease, 0),
    AHLocationYTo.new(self.tray, duration, ease, -self.tray:getHeightPart(0.5)),
    AHDelayed.new(duration/4, function() self.trayDismissedCallback() end)
  )
  Sound.playSoundEffect("hint_tray_close")
  self.actionInterrupt:insert(animation)
end

function CrossHelpTray:setAutocheckToggleState(enabled)
  self.toggleAutocheckButton:setState(enabled)
  self.checkForErrorsButton.setIsActive(not enabled)
end

function CrossHelpTray:reset()
  self:setAutocheckToggleState(false)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createCloseButton(self)
  local buttonSize = ReferenceDevice.getScaledHeight(60)
  local dismissFunction = function() self:hide() end

  local closeIcon = Sprite.newWithName("icon_close.png")
  closeIcon:setEffectiveDims(buttonSize/4)
  closeIcon:setAnchorPointTopLeft()

  local closeButton = Button.new(buttonSize, buttonSize, nil, dismissFunction)
  closeButton:insertAtTopLeft(closeIcon)
  local xOffset = self:getWidthPart(0.08)
  local yOffset = -xOffset * 0.833
  closeIcon:offsetX(xOffset)
  closeIcon:offsetY(yOffset)
  return closeButton
end

local function createHintsLabel()
  local hintsText = Localizer.get("HINTS")
  local hintsLabelFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local hintsLabel = Label.new(hintsText, hintsLabelFont)
  hintsLabel:setColor(unpack(CrossColors.TRAY_TEXT))
  return hintsLabel
end

local function createLayout(self, height)
  local overlay = FilledRectangle.newFullScreen(CrossColors.TRAY_OVERLAY)
  overlay:setOpacity(0)
  self:insertAtCenter(overlay)
  self.overlay = overlay

  local cornerRadius = 12
  local trayHeight = height + cornerRadius/2
  local trayColor = CrossColors.TRAY
  local sharpCornersList = {FilledRoundedRectangle.CORNERS.BOTTOM_LEFT, FilledRoundedRectangle.CORNERS.BOTTOM_RIGHT}
  local tray = FilledRoundedRectangle.new(self:getWidth(), trayHeight, cornerRadius, trayColor, sharpCornersList)
  self:insertAtBottom(tray)
  tray:offsetY(-trayHeight/2)
  self.tray = tray

  local dismissTouchAreaWidth = self:getWidth()
  local dismissTouchAreaHeight = self:getHeight() - trayHeight
  dismissTouchArea = Group.new(dismissTouchAreaWidth, dismissTouchAreaHeight)
  Button.makeButton(dismissTouchArea, nil, function() self:hide() end)
  self:insertAtLocation(dismissTouchArea, (dismissTouchAreaWidth / 2), ((dismissTouchAreaHeight / 2) + trayHeight))
  self.dismissTouchArea = dismissTouchArea

  local closeButton = createCloseButton(self)
  tray:insertAtTopLeft(closeButton)
  closeButton:offsetX(closeButton:getWidthPart(0.5))
  closeButton:offsetY(closeButton:getHeightPart(-0.5))
  table.insert(self.buttons, closeButton)

  local hintsLabel = createHintsLabel()
  tray:insertAtTop(hintsLabel)
  hintsLabel:offsetY(closeButton:getHeightPart(-0.5))

  local fullButtonWidth = self:getWidthPart(0.86)
  local fullButtonHeight = math.min(fullButtonWidth * 0.17, trayHeight * 0.175)

  local halfButtonWidth = fullButtonWidth * 0.48
  local halfButtonHeight = fullButtonHeight * 1.42
  local revealLetterButton = CrossTrayButton.newRevealLetter(halfButtonWidth, halfButtonHeight, getRevealLetterCallback(self))
  table.insert(self.buttons, revealLetterButton)
  local revealWordButton = CrossTrayButton.newRevealWord(halfButtonWidth, halfButtonHeight, getRevealWordCallback(self))
  table.insert(self.buttons, revealWordButton)
  local checkForErrorsButton = CrossTrayButton.newCheckForErrors(fullButtonWidth, fullButtonHeight, getCheckForErrorsCallback(self))
  table.insert(self.buttons, checkForErrorsButton)
  local toggleAutocheckButton = CrossTrayToggleButton.newAutocheck(fullButtonWidth, fullButtonHeight * 1.2, getToggleAutocheckCallback(self, checkForErrorsButton))
  table.insert(self.buttons, toggleAutocheckButton)
  self.checkForErrorsButton = checkForErrorsButton
  self.toggleAutocheckButton = toggleAutocheckButton

  local revealButtonRowSpacing = fullButtonWidth - (halfButtonWidth * 2)
  local revealButtonRow = Row.new(revealButtonRowSpacing)
  revealButtonRow:insert(revealLetterButton)
  revealButtonRow:insert(revealWordButton)
  revealButtonRow:alignChildren()

  local hintButtonColumnSpacing = fullButtonHeight * 0.2
  local hintButtonColumn = Column.new(hintButtonColumnSpacing)
  hintButtonColumn:insert(revealButtonRow)
  hintButtonColumn:insertSpacer(0)
  hintButtonColumn:insert(checkForErrorsButton)
  hintButtonColumn:insertSpacer(-hintButtonColumnSpacing * 1.5)
  hintButtonColumn:insert(toggleAutocheckButton)
  hintButtonColumn:alignChildren()
  tray:insertAtCenter(hintButtonColumn)
  hintButtonColumn:setAnchorPointTop()
  hintButtonColumn:setY(closeButton:getYBottom() - (hintButtonColumnSpacing / 3))
end

-- Constructor ------------------------

function CrossHelpTray.new(height, round, trayDismissedCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossHelpTray)

  self.actionInterrupt = ActionInterrupt.new()
  self.round = round
  self.trayDismissedCallback = trayDismissedCallback
  self.isActive = false

  self.buttons = {}

  createLayout(self, height)

  setButtonsTouchEnabled(self, false)
  self.dismissTouchArea:setTouchEnabled(false)

  return self
end
