--- CrossKeyboard
-- @author Created by andrewcousins on March 08, 2022.

class("CrossKeyboard")

require("CrossKeyboardKey")
require("CrossDesktopKeyboard")
require("CrossDirection")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REFERENCE_DEVICE_PHONE_KEYBOARD_WIDTH = 450
local REFERENCE_DEVICE_PHONE_KEYBOARD_HEIGHT = 198

local REFERENCE_DEVICE_PHONE_KEY_WIDTH = 32
local REFERENCE_DEVICE_PHONE_KEY_HEIGHT = 42

local REFERENCE_DEVICE_PHONE_KEY_MARGIN_HORIZONTAL = 5.4
local REFERENCE_DEVICE_PHONE_KEY_MARGIN_VERTICAL = 12

local REFERENCE_DEVICE_TABLET_KEYBOARD_HEIGHT = 143

local REFERENCE_DEVICE_TABLET_KEY_WIDTH = 43.5
local REFERENCE_DEVICE_TABLET_KEY_HEIGHT = 36.7

local REFERENCE_DEVICE_TABLET_KEY_MARGIN_HORIZONTAL = 7.5
local REFERENCE_DEVICE_TABLET_KEY_MARGIN_VERTICAL = 10.46

local KEYS = {
  {"Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P"},
  {"A", "S", "D", "F", "G", "H", "J", "K", "L"},
  {CrossKeyboardKey.EMPTY_KEY, "Z", "X", "C", "V", "B", "N", "M", CrossKeyboardKey.DELETE_KEY}
}

local ANIMATE_IN_BOUNCE_AMOUNT = 10

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getKeyStyle(rowIndex, letterIndex, letterCountInRow)
  if (rowIndex == 1) then
    if (letterIndex == 1) then
      return CrossKeyboardKeyPopUpShape.STYLE.LEFT
    elseif (letterIndex == letterCountInRow) then
      return CrossKeyboardKeyPopUpShape.STYLE.RIGHT
    end
  end
  return CrossKeyboardKeyPopUpShape.STYLE.CENTER
end

local function getKeyHorizontalMargin(defaultHorizontalMargin, rowIndex, letterIndex, letterCountInRow)
  if (
    ((rowIndex <= 2) and ((letterIndex == 1) or -- Update margin for "Q" and "A"
                          (letterIndex == letterCountInRow))) or -- Update margin for "P" and "L"
    ((rowIndex == 3) and (letterIndex == 2)) -- Update margin for "Z"
  ) then
    local increasedMargin = defaultHorizontalMargin * 2
    return increasedMargin
  end
  return defaultHorizontalMargin
end

local function getKeyHorizontalOffset(defaultHorizontalMargin, horizontalMargin, letterIndex, letterCountInRow)
  if (defaultHorizontalMargin == horizontalMargin) then
    return 0
  end
  local offsetAmount = (defaultHorizontalMargin / 2)
  if (letterIndex < (letterCountInRow / 2)) then
    return -offsetAmount
  end
  return offsetAmount
end

local function getIsTouchEnabledCallback(self)
  return function()
    return self.touchEnabled
  end
end

local function getMoveToNextClueCallback(self)
  return function()
    self.round:moveToNextClue()
  end
end

local function getMoveToPreviousClueCallback(self)
  return function()
    self.round:moveToPreviousClue()
  end
end

local function getMoveToAlternateClueCallback(self)
  return function()
    self.round:moveToAlternateClue()
  end
end

local function getMoveCursorCallback(self, direction)
  return function()
    self.round:moveCursorInDirection(direction)
  end
end

local function getClearAndAdvanceCallback(self)
  return function()
    self.round:clearContentAtCursorAndAdvance()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossKeyboard:prepareForIntro()
  self:setY(self:getHeightPart(-0.5))
end

function CrossKeyboard:getIntroAnimation()
  local popUpDuration = 0.333
  local popUpEase = AHEaseBezier.newFlat({0.0, 0.47, 0.0, 1.0})
  local popUpHeight = self:getHeight() + ANIMATE_IN_BOUNCE_AMOUNT
  local popUpAnimation = AHLocationYBy.new(self, popUpDuration, popUpEase, popUpHeight)

  local settleDownDuration = 0.333
  local settleDownEase = AHEaseBezier.newFlat({0.4, 0.0, 0.8, 1.0})
  local settleDownHeight = -ANIMATE_IN_BOUNCE_AMOUNT
  local settleDownAnimation = AHLocationYBy.new(self, settleDownDuration, settleDownEase, settleDownHeight)

  return AHSequence.new(popUpAnimation, settleDownAnimation)
end

function CrossKeyboard:getOutroAnimation()
  local additionalBufferToSyncWithClueBar = Screen.getHeightPart(0.08)
  local moveDistance = -(self:getHeight() + additionalBufferToSyncWithClueBar)

  local duration = 0.555
  local ease = AHEaseBezier.newFlat({1.0, 0.0, 0.7, 1.0})
  local animation = AHLocationYBy.new(self, duration, ease, moveDistance)

  return animation
end

function CrossKeyboard:setTouchEnabled(enabled)
  self.touchEnabled = enabled
  CrossDesktopKeyboard.setEnabled(enabled)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local backing = FilledRectangle.new(self:getWidth(), self:getHeight() + ANIMATE_IN_BOUNCE_AMOUNT, CrossColors.KEYBOARD_BACKGROUND)
  self:insertAtCenter(backing)
  backing:offsetY(-ANIMATE_IN_BOUNCE_AMOUNT / 2)

  local defaultKeyHorizontalMargin = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_PHONE_KEY_MARGIN_HORIZONTAL)
  local keyVerticalMargin = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_PHONE_KEY_MARGIN_VERTICAL)
  local keyWidth = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_PHONE_KEY_WIDTH)
  local keyHeight = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_PHONE_KEY_HEIGHT)

  if Screen.isTabletAspectRatio() then
    defaultKeyHorizontalMargin = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_TABLET_KEY_MARGIN_HORIZONTAL)
    keyVerticalMargin = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_TABLET_KEY_MARGIN_VERTICAL)
    keyWidth = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_TABLET_KEY_WIDTH)
    keyHeight = ReferenceDevice.getScaledWidth(REFERENCE_DEVICE_TABLET_KEY_HEIGHT)
  end

  local keyPressedCallback = function(keyContent)
    if (keyContent == CrossKeyboardKey.DELETE_KEY) then
      self.round:clearContentAtCursor()
      Sound.playSoundEffect("delete_letter")
    else
      self.round:updateContentAtCursor(keyContent)
      Sound.playSoundEffect("type")
    end
  end

  local keyboardColumn = Column.new(keyVerticalMargin)
  for rowIndex, letters in ipairs(KEYS) do
    local letterCountInRow = #letters
    local keyboardRow = Row.new(defaultKeyHorizontalMargin)
    for letterIndex = 1, letterCountInRow do
      local letter = letters[letterIndex]
      local keyStyle = getKeyStyle(rowIndex, letterIndex, letterCountInRow)
      local isKeyboardEnabledCallback = getIsTouchEnabledCallback(self)
      local horizontalMargin = getKeyHorizontalMargin(defaultKeyHorizontalMargin, rowIndex, letterIndex, letterCountInRow)
      local horizontalOffset = getKeyHorizontalOffset(defaultKeyHorizontalMargin, horizontalMargin, letterIndex, letterCountInRow)
      local key = CrossKeyboardKey.new(keyWidth, keyHeight, horizontalMargin, horizontalOffset, keyVerticalMargin, letter, keyStyle, keyPressedCallback, isKeyboardEnabledCallback)
      keyboardRow:insert(key)
    end
    keyboardRow:alignChildren()
    keyboardColumn:insert(keyboardRow)
  end
  keyboardColumn:alignChildren()
  self:insertAtCenter(keyboardColumn)
  keyboardColumn:offsetY(self.emptyLowerHeight/2)
end

-- Constructor ------------------------

function CrossKeyboard.new(round)
  local width = Screen.width
  local height = width * (REFERENCE_DEVICE_PHONE_KEYBOARD_HEIGHT / REFERENCE_DEVICE_PHONE_KEYBOARD_WIDTH)
  local buffer = Screen.getHeightPart(0.025)
  local safeArea = Screen.safeBottom

  if Screen.isTabletAspectRatio() then
    ReferenceDevice.setDevice("ipad pro")
    height = ReferenceDevice.getScaledHeight(REFERENCE_DEVICE_TABLET_KEYBOARD_HEIGHT)
    safeArea = math.max(Screen.safeBottom, Screen.getHeightPart(0.025))
  end

  height = height + safeArea + buffer
  local self = Group.new(width, height)
  Mixin.mix(self, CrossKeyboard)

  self.round = round
  self.emptyLowerHeight = safeArea + buffer
  self.touchEnabled = true
  createLayout(self)

  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveToNextClueCallback(self), CrossDesktopKeyboard.RIGHT_KEY)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveToPreviousClueCallback(self), CrossDesktopKeyboard.LEFT_KEY)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveToAlternateClueCallback(self), CrossDesktopKeyboard.SLASH_KEY)

  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveToNextClueCallback(self), CrossDesktopKeyboard.ENTER)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveToNextClueCallback(self), CrossDesktopKeyboard.TAB)

  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getClearAndAdvanceCallback(self), CrossDesktopKeyboard.SPACE)

  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveCursorCallback(self, CrossDirection.LEFT), CrossDesktopKeyboard.LEFT_ARROW)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveCursorCallback(self, CrossDirection.RIGHT), CrossDesktopKeyboard.RIGHT_ARROW)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveCursorCallback(self, CrossDirection.DOWN), CrossDesktopKeyboard.DOWN_ARROW)
  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(getMoveCursorCallback(self, CrossDirection.UP), CrossDesktopKeyboard.UP_ARROW)

  return self
end
