--- CrossKeyboardKey
-- @author Created by andrewcousins on May 25, 2022.

class("CrossKeyboardKey")

require("CrossKeyboardKeyPopUpShape")
require("CrossHaptics")
require("CrossDesktopKeyboard")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CrossKeyboardKey.EMPTY_KEY = "EMPTY"
CrossKeyboardKey.DELETE_KEY = "DELETE"

local Z_INDEX = {
  DROP_SHADOW = 0,
  BACKGROUND = 10,
  LABEL = 20,
  POP_UP = 30
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isKeyEmpty(keyContent)
  return keyContent == CrossKeyboardKey.EMPTY_KEY
end

local function isKeyDelete(keyContent)
  return keyContent == CrossKeyboardKey.DELETE_KEY
end

local function isSpecialFunctionKey(keyContent)
  return isKeyEmpty(keyContent) or isKeyDelete(keyContent)
end

local function getFontFaceForDeviceType()
  if (Host.systemFontFamily ~= nil) then
    local targetFontFace = tostring(Host.systemFontFamily) .. "_REGULAR"
    if (FontFaces[targetFontFace] ~= nil) then
      return FontFaces[targetFontFace]
    else
      Log.warn("Font face for system not found: " .. tostring(targetFontFace))
    end
  else
    Log.warn("No system font face set for Host.")
  end
  return FontFaces.ROBOTO_REGULAR
end

local function addKeyPopUp(self, keyContent, keyStyle, cornerRadius, fontFace, fontSize)
  local width, height = self:getDims()
  local popUp = Group.new(width, height)

  local popUpSizeMultiplier = 1.8
  local popUpFontSizeMultiplier = 1.7

  local dropShadow = CrossKeyboardKeyPopUpShape.newFilled(width, height, cornerRadius, CrossColors.KEYBOARD_KEY_DROPSHADOW, keyStyle)
  popUp:insertAtCenter(dropShadow, Z_INDEX.DROP_SHADOW)
  local dropShadowOffset = self:getHeightPart(-0.05)
  dropShadow:offsetY(dropShadowOffset)
  local popUpBackground = CrossKeyboardKeyPopUpShape.newFilled(width, height, cornerRadius, CrossColors.KEYBOARD_KEY, keyStyle)
  popUp:insertAtCenter(popUpBackground, Z_INDEX.BACKGROUND)

  local lineWidth = 1
  local popUpOutline = CrossKeyboardKeyPopUpShape.newHollow(width, height, cornerRadius, lineWidth, CrossColors.KEYBOARD_KEY_POP_UP_OUTLINE, keyStyle)
  popUp:insertAtCenter(popUpOutline, Z_INDEX.LABEL)

  local popUpFont = Font.new(fontFace, fontSize * popUpFontSizeMultiplier)
  local popUpLabel = Label.new(keyContent, popUpFont)
  popUpLabel:setColor(unpack(CrossColors.KEYBOARD_TEXT))

  popUpBackground:insertLabel(popUpLabel, Z_INDEX.LABEL)

  popUp:linkOpacityAttrToChildren(true)
  popUp:setOpacity(0)
  self:insertAtCenter(popUp, Z_INDEX.POP_UP)
  self.popUp = popUp
end

local function isMoveOutsideVirtualTouchSurface(self, xLocation, yLocation)
  local width, height = self.virtualTouchSurface:getDims()
  local xCenter, yCenter = self.virtualTouchSurface:getCenterWorldLoc()
  return ((math.abs(xLocation - xCenter) > (width / 2)) or (math.abs(yLocation - yCenter) > (height / 2)))
end

local function createTouchHandler(self, onTouchBegan, onTouchEnded, onTouchCancel)
  local isCancelled = false

  local handler = function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      isCancelled = false
      onTouchBegan()
    elseif ((event.phase == TouchManager.TOUCH_MOVED) and not isCancelled) then
      if isMoveOutsideVirtualTouchSurface(self, event.x, event.y) then
        onTouchCancel()
        isCancelled = true
        return TouchEvent.HANDLING.HAND_OFF
      end
    elseif ((event.phase == TouchManager.TOUCH_ENDED) and not isCancelled) then
      isCancelled = true
      onTouchEnded()
    end

    return TouchEvent.HANDLING.CONTINUE
  end

  return handler
end

local function isKeyboardEnabled(self)
  return self.isKeyboardEnabledCallback()
end

local function keyPressed(self)
  return self.keyPressedCallback(self.keyContent)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureTouchManagement(self, keyHorizontalMargin, keyHorizontalOffset, keyVerticalMargin)
  local width, height = self:getDims()
  local keyBackground = self.keyBackground
  local dropShadowOffset = self.dropShadowOffset
  local keyBackgroundYLocation = keyBackground:getY()

  local onPressDown = function()
    if ((not self.isDown) and isKeyboardEnabled(self)) then
      CrossHaptics:playOnTouchDown()
      local animation = AHSpawn.new()
      if (self.popUp == nil) then
        local duration = 0.01
        local ease = nil
        animation:addAction(AHLocationYTo.new(keyBackground, duration, ease, keyBackgroundYLocation + dropShadowOffset))
      else
        animation:addAction(function() self.popUp:setOpacity(1) end)
      end
      self.actionQueue:insert(animation)
      self.isDown = true
    end
  end
  local onPressCancel = function()
    if (self.isDown) then
      local animation = AHSpawn.new()
      if (self.popUp == nil) then
        local duration = 0.01
        local ease = nil
        animation:addAction(AHLocationYTo.new(keyBackground, duration, ease, keyBackgroundYLocation))
      else
        animation:addAction(function() self.popUp:setOpacity(0) end)
      end
      self.actionQueue:insert(animation)
      self.isDown = false
    end
  end
  local onPressUp = function()
    onPressCancel()
    if isKeyboardEnabled(self) then
      CrossHaptics:playOnTouchUp()
      keyPressed(self)
    end
  end

  local virtualTouchSurfaceWidth = width + keyHorizontalMargin
  local virtualTouchSurfaceHeight = height + keyVerticalMargin
  local virtualTouchSurface = Group.new(virtualTouchSurfaceWidth, virtualTouchSurfaceHeight)
  TouchManager.addTouchEnabledMethods(virtualTouchSurface, createTouchHandler(self, onPressDown, onPressUp, onPressCancel))
  virtualTouchSurface:setTouchEnabled(true)
  self:insertAtCenter(virtualTouchSurface)
  virtualTouchSurface:offsetX(keyHorizontalOffset)
  self.virtualTouchSurface = virtualTouchSurface

  CrossDesktopKeyboard.setPressDownCallbackForKeyContent(onPressDown, self.keyContent)
  CrossDesktopKeyboard.setPressUpCallbackForKeyContent(onPressUp, self.keyContent)
end


local function createLayout(self, keyStyle)
  local width, height = self:getDims()

  local cornerRadius = (height * 0.1)
  local dropShadow = FilledRoundedRectangle.new(width, height, cornerRadius, CrossColors.KEYBOARD_KEY_DROPSHADOW)
  self:insertAtCenter(dropShadow, Z_INDEX.DROP_SHADOW)
  local dropShadowOffset = self:getHeightPart(-0.025)
  dropShadow:offsetY(dropShadowOffset)
  local keyBackground = FilledRoundedRectangle.new(width, height, cornerRadius, CrossColors.KEYBOARD_KEY)
  self:insertAtCenter(keyBackground, Z_INDEX.BACKGROUND)
  self.keyBackground = keyBackground
  self.dropShadowOffset = dropShadowOffset

  local keyLabel = nil
  local keyPopUp = nil
  if isSpecialFunctionKey(self.keyContent) then
    if isKeyDelete(self.keyContent) then
      keyBackground:setColor(unpack(CrossColors.KEYBOARD_DELETE_KEY))
      keyLabel = Sprite.newWithName("shared_keyboard_backspace_icon_hover.png")
      keyLabel:setEffectiveDims(nil, height * 0.4)
    end
  else
    local fontSize = math.min((width * 0.75), (height * 0.5))
    local fontFace = getFontFaceForDeviceType()
    local font = Font.new(fontFace, fontSize)
    keyLabel = Label.new(self.keyContent, font)

    addKeyPopUp(self, self.keyContent, keyStyle, cornerRadius, fontFace, fontSize)
  end
  keyLabel:setColor(unpack(CrossColors.KEYBOARD_TEXT))
  self:insertAtCenter(keyLabel, Z_INDEX.LABEL)
  keyLabel:offsetY(self:getHeightPart(-0.02)) -- small offset makes letters appear visually centered
end

-- Constructor ------------------------

function CrossKeyboardKey.new(keyWidth, keyHeight, keyHorizontalMargin, keyHorizontalOffset, keyVerticalMargin, keyContent, keyStyle, keyPressedCallback, isKeyboardEnabledCallback)
  if isSpecialFunctionKey(keyContent) then
    keyWidth = (keyWidth * 1.5)
  end

  local self = Group.new(keyWidth, keyHeight)
  Mixin.mix(self, CrossKeyboardKey)

  self.keyContent = keyContent
  self.actionQueue = ActionInterrupt.new()
  self.isDown = false
  self.keyPressedCallback = keyPressedCallback
  self.isKeyboardEnabledCallback = isKeyboardEnabledCallback

  if (not isKeyEmpty(keyContent)) then
    createLayout(self, keyStyle)
    configureTouchManagement(self, keyHorizontalMargin, keyHorizontalOffset, keyVerticalMargin)
  end

  return self
end
