--- CrossKeyboardKeyPopUpShape
-- @author Created by andrewcousins on May 25, 2022.

class("CrossKeyboardKeyPopUpShape")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CrossKeyboardKeyPopUpShape.STYLE = {
  LEFT = "left",
  CENTER = "center",
  RIGHT = "right"
}

local WIDTH_MULTIPLIER = 1.81
local HEIGHT_MULTIPLIER = 2.48

local TOP_KEY_PORTION = 0.476
local BOTTOM_KEY_PORTION = 0.403

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getPointsForCenterStyle(width, height, outerRightEdge, margin, cornerRadius, isFilledGraphicsPrimitive)
  local topCornerRadius = cornerRadius * 3
  local points = {}

  table.insert(points, {x = (width / 2), y = 0})

  -- bottom right corner
  local centerBottomRightX = (width * 0.776) - cornerRadius
  local centerBottomRightY = cornerRadius
  ShapeUtils.makeArcPoints((-math.pi / 2), 0, cornerRadius, centerBottomRightX, centerBottomRightY, nil, points)

  -- lower-middle right curve
  local lowerMiddleRightCurveStartX = (width * 0.776)
  local lowerMiddleRightCurveStartY = (height * BOTTOM_KEY_PORTION)
  table.insert(points, {x = lowerMiddleRightCurveStartX, y = lowerMiddleRightCurveStartY})
  table.insert(points, {x = lowerMiddleRightCurveStartX + (cornerRadius * math.cos(math.pi * (1/6))), y = lowerMiddleRightCurveStartY + (cornerRadius * math.sin(math.pi * (1/3)))})

  -- upper-middle right curve
  local upperMiddleRightEndX = outerRightEdge
  local upperMiddleRightEndY = (height * TOP_KEY_PORTION) + cornerRadius
  table.insert(points, {x = upperMiddleRightEndX + (cornerRadius * math.cos(math.pi * (5/4))), y = upperMiddleRightEndY + (cornerRadius * math.sin(math.pi * (5/4)))})
  table.insert(points, {x = upperMiddleRightEndX, y = upperMiddleRightEndY})

  -- top right corner
  local centerTopRightX = outerRightEdge - topCornerRadius
  local centerTopRightY = height - topCornerRadius
  ShapeUtils.makeArcPoints(0, (math.pi / 2), topCornerRadius, centerTopRightX, centerTopRightY, nil, points)

  -- Mirror the right side points to the left side
  for i = #points, 1, -1 do
    local point = points[i]
    local mirroredX = width - point.x
    table.insert(points, {x = mirroredX, y = point.y})
  end

  if isFilledGraphicsPrimitive then
     -- give the first point in the list line of sight to every other point
     -- to ensure proper rendering
    table.insert(points, 1, {x = (width / 2), y = height})
  end

  return points
end

local function getPointsForLeftStyle(width, height, outerRightEdge, margin, cornerRadius, isFilledGraphicsPrimitive)

  local topCornerRadius = cornerRadius * 3
  local points = {}

  table.insert(points, {x = (width / 4), y = height})

  -- top left corner
  local centerTopLeftX = topCornerRadius
  local centerTopLeftY = height - topCornerRadius
  ShapeUtils.makeArcPoints((math.pi / 2), math.pi, topCornerRadius, centerTopLeftX, centerTopLeftY, nil, points)

  -- bottom left corner
  local centerBottomLeftX = cornerRadius
  local centerBottomLeftY = cornerRadius
  ShapeUtils.makeArcPoints(-math.pi, (-math.pi / 2), cornerRadius, centerBottomLeftX, centerBottomLeftY, nil, points)

  -- bottom right corner
  local centerBottomRightX = (width * 0.552) - cornerRadius
  local centerBottomRightY = cornerRadius
  ShapeUtils.makeArcPoints((-math.pi / 2), 0, cornerRadius, centerBottomRightX, centerBottomRightY, nil, points)

  -- lower-middle right curve
  local lowerMiddleRightCurveStartX = (width * 0.552)
  local lowerMiddleRightCurveStartY = (height * BOTTOM_KEY_PORTION) * 0.9
  table.insert(points, {x = lowerMiddleRightCurveStartX, y = lowerMiddleRightCurveStartY})
  table.insert(points, {x = lowerMiddleRightCurveStartX + (cornerRadius * math.cos(math.pi * (1/6))), y = lowerMiddleRightCurveStartY + (cornerRadius * math.sin(math.pi * (1/3)))})

  -- upper-middle right curve
  local upperMiddleRightEndX = outerRightEdge
  local upperMiddleRightEndY = (height * TOP_KEY_PORTION) + cornerRadius
  table.insert(points, {x = upperMiddleRightEndX + (cornerRadius * math.cos(math.pi * (5/4))), y = upperMiddleRightEndY + (cornerRadius * math.sin(math.pi * (5/4)))})
  table.insert(points, {x = upperMiddleRightEndX, y = upperMiddleRightEndY})

  -- top right corner
  local centerTopRightX = outerRightEdge - topCornerRadius
  local centerTopRightY = height - topCornerRadius
  ShapeUtils.makeArcPoints(0, (math.pi / 2), topCornerRadius, centerTopRightX, centerTopRightY, nil, points)

  return points
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossKeyboardKeyPopUpShape:updateGraphicsPrimitive()
  local width = self.width
  local height = self.height
  local margin = (self.lineWidth or 0) / 2
  local cornerRadius = self.cornerRadius - margin
  local isFilledGraphicsPrimitive = self.isFilledGraphicsPrimitive

  local outerRightEdge = width
  if Screen.isTabletAspectRatio() then
    outerRightEdge = width * 0.9
  end

  local points = {}
  if (self.style == CrossKeyboardKeyPopUpShape.STYLE.CENTER) then
    points = getPointsForCenterStyle(width, height, outerRightEdge, margin, cornerRadius, isFilledGraphicsPrimitive)
  elseif (self.style == CrossKeyboardKeyPopUpShape.STYLE.LEFT) then
    points = getPointsForLeftStyle(width, height, outerRightEdge, margin, cornerRadius, isFilledGraphicsPrimitive)
  elseif (self.style == CrossKeyboardKeyPopUpShape.STYLE.RIGHT) then
    points = getPointsForLeftStyle(width, height, outerRightEdge, margin, cornerRadius, isFilledGraphicsPrimitive)
    for _, point in ipairs(points) do
      local mirroredX = width - point.x
      point.x = mirroredX
    end
  end

  self._graphicsPrimitiveCoordinateLists = {ShapeUtils.flattenCoordPairsNew(points)}
end

function CrossKeyboardKeyPopUpShape:insertLabel(label, zIndex)
  local yLocationFractional = ((1 + ((BOTTOM_KEY_PORTION + TOP_KEY_PORTION) / 2)) / 2)
  self:insertAtLocationFractional(label, 0.5, yLocationFractional, zIndex)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makePopUpShape(self, width, height, cornerRadius, style)
  Mixin.mix(self, CrossKeyboardKeyPopUpShape)

  self.width = width
  self.height = height
  self.cornerRadius = cornerRadius
  self.style = style or CrossKeyboardKeyPopUpShape.STYLE.CENTER

  self:setAnchorPoint(0.5, (0.5 / HEIGHT_MULTIPLIER))
  if (style == CrossKeyboardKeyPopUpShape.STYLE.LEFT) then
    self:setAnchorPoint(0.276, (0.5 / HEIGHT_MULTIPLIER))
  elseif (style == CrossKeyboardKeyPopUpShape.STYLE.RIGHT) then
    self:setAnchorPoint(0.724, (0.5 / HEIGHT_MULTIPLIER))
  end
end

-- Constructor ------------------------

function CrossKeyboardKeyPopUpShape.newFilled(keyWidth, keyHeight, cornerRadius, color, style)
  local width = keyWidth * WIDTH_MULTIPLIER
  local height = keyHeight * HEIGHT_MULTIPLIER
  local self = FilledGraphicsPrimitive.new(width, height, color)
  makePopUpShape(self, width, height, cornerRadius, style)

  self.isFilledGraphicsPrimitive = true
  self:updateGraphicsPrimitive()

  return self
end

function CrossKeyboardKeyPopUpShape.newHollow(keyWidth, keyHeight, cornerRadius, lineWidth, color, style)
  local width = keyWidth * WIDTH_MULTIPLIER
  local height = keyHeight * HEIGHT_MULTIPLIER
  local blurMargin = 0
  local isLooped = true
  local self = HollowGraphicsPrimitive.new(width, height, color, lineWidth, blurMargin, isLooped)
  makePopUpShape(self, width, height, cornerRadius, style)

  self.isFilledGraphicsPrimitive = false
  self:setShouldApplyAutoMargin(false)
  self:updateGraphicsPrimitive()

  return self
end
