--- CrossPopUpScreen
-- @author Created by andrewcousins on July 06, 2023.

class("CrossPopUpScreen")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_OPACITY = 0.5

local DOWN_STATE_OPACITY = 0.7

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createMainButton(self, buttonText, buttonWidth, buttonHeight, callback)
  local cornerRadius = buttonWidth * 0.01

  local button = FilledRoundedRectangle.new(buttonWidth, buttonHeight, cornerRadius, CrossColors.POP_UP_SCREEN_CARD_BUTTON)

  local buttonFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local buttonLabel = Label.new(buttonText, buttonFont)
  buttonLabel:setColor(unpack(CrossColors.POP_UP_SCREEN_CARD_BUTTON_TEXT))
  button:insertAtCenter(buttonLabel)

  local onTouchDown = function()
    button:setOpacity(DOWN_STATE_OPACITY)
  end
  local onTouchUp = function()
    button:setOpacity(1)
    button:setTouchEnabled(false)
    self:hide()
    if (callback ~= nil) then
      callback()
    end
  end
  local onTouchCancel = function()
    button:setOpacity(1)
  end
  Button.makeButton(button, onTouchDown, onTouchUp, onTouchCancel)

  return button
end


local function createSecondaryButton(self, buttonText, buttonWidth, buttonHeight, callback)
  local button = Group.new(buttonWidth, buttonHeight)

  local buttonFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local buttonLabel = Label.new(buttonText, buttonFont)
  buttonLabel:setColor(unpack(CrossColors.POP_UP_SCREEN_CARD_SECONDARY_BUTTON_TEXT))
  button:insertAtCenter(buttonLabel)

  local onTouchDown = function()
    buttonLabel:setOpacity(DOWN_STATE_OPACITY)
  end
  local onTouchUp = function()
    buttonLabel:setOpacity(1)
    button:setTouchEnabled(false)
    self:hide()
    if (callback ~= nil) then
      callback()
    end
  end
  local onTouchCancel = function()
    buttonLabel:setOpacity(1)
  end
  Button.makeButton(button, onTouchDown, onTouchUp, onTouchCancel)

  return button
end

local function addCardAndBackground(self, cardWidth, cardHeight, backgroundOpacity, callback)
  local width, height = self:getDims()
  backgroundOpacity = backgroundOpacity or BACKGROUND_OPACITY
  local background = FilledRectangle.new(width, height, CrossColors.POP_UP_SCREEN_BACKGROUND)
  background:setOpacity(backgroundOpacity)
  self:insertAtCenter(background)

  local cornerRadius = cardWidth * 0.02
  local card = FilledRoundedRectangle.new(cardWidth, cardHeight, cornerRadius, CrossColors.POP_UP_SCREEN_CARD)
  self:insertAtCenter(card)
  self.card = card

  -- Tapping outside of the pop up will dismiss it
  Button.makeButton(background, function()
    if (callback ~= nil) then
      callback()
    end
    self:hide()
  end)
  Button.makeButton(card) -- Do nothing, just block input
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossPopUpScreen:hide()
  self:removeSelf()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, titleText, subText, mainButtonText, mainButtonCallback, popUpDismissedCallback, secondaryButtonText, secondaryButtonCallback, backgroundOpacity)
  local cardWidth = Screen.getSafeWidthPart(0.872)
  local cardHeight = cardWidth * 0.868
  addCardAndBackground(self, cardWidth, cardHeight, backgroundOpacity, popUpDismissedCallback)

  local elementColumnSpacing = 8
  local elementColumn = Column.new(elementColumnSpacing)

  local titleFont = Font.new(FontFaces.DIN_OT_MEDIUM, 24)
  local titleLabel = Label.new(titleText, titleFont)
  titleLabel:setColor(unpack(CrossColors.POP_UP_SCREEN_CARD_TEXT))
  elementColumn:insert(titleLabel)
  elementColumn:insertSpacer(0)

  local subtitleFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local textWidth = self.card:getWidthPart(0.85)
  local subtitleLabel = DynamicTextArea.newSimple(subText, textWidth, subtitleFont, Alignment.CENTER)
  subtitleLabel:setColor(unpack(CrossColors.POP_UP_SCREEN_CARD_SUBTITLE))
  elementColumn:insert(subtitleLabel)

  elementColumn:insertSpacer(16)

  local buttonWidth = self.card:getWidthPart(0.85)
  local buttonHeight = buttonWidth * 0.2
  local mainButton = createMainButton(self, mainButtonText, buttonWidth, buttonHeight, mainButtonCallback)
  elementColumn:insert(mainButton)

  if secondaryButtonText then
    local secondaryButton = createSecondaryButton(self, secondaryButtonText, buttonWidth, buttonHeight, secondaryButtonCallback)
    elementColumn:insert(secondaryButton)
  end

  elementColumn:alignChildren()
  self.card:setHeight(elementColumn:getHeight() + 40)
  self.card:insertAtCenter(elementColumn)
end

-- Constructor ------------------------

function CrossPopUpScreen.new(titleText, subText, mainButtonText, mainButtonCallback, popUpDismissedCallback, secondaryButtonText, secondaryButtonCallback, backgroundOpacity)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossPopUpScreen)

  createLayout(self, titleText, subText, mainButtonText, mainButtonCallback, popUpDismissedCallback, secondaryButtonText, secondaryButtonCallback, backgroundOpacity)

  return self
end
