--- CrossPostGameButton
-- @author Created by andrewcousins on June 15, 2023.

class("CrossPostGameButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

  local BUTTON_WIDTH = Screen.getSafeWidthPart(0.872)
  local HALF_BUTTON_WIDTH = BUTTON_WIDTH * 0.486
  local BUTTON_HEIGHT = BUTTON_WIDTH * 0.16
  local CORNER_RADIUS = BUTTON_WIDTH * 0.01

  local PILL_CORNER_RADIUS = BUTTON_WIDTH * 0.5

  local DOWN_STATE_OPACITY = 0.7

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossPostGameButton.getHalfButtonSpacing()
  return BUTTON_WIDTH - (HALF_BUTTON_WIDTH * 2)
end

function CrossPostGameButton.getButtonHeight()
  return BUTTON_HEIGHT
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function addRegularButtonFunctionality(self, callback)
  local onTouchDown = function()
    self:setOpacity(DOWN_STATE_OPACITY)
  end
  local onTouchUp = function()
    self:setOpacity(1)
    callback()
  end
  local onTouchCancel = function()
    self:setOpacity(1)
  end
  Button.makeButton(self, onTouchDown, onTouchUp, onTouchCancel)
end

local function addInvertedButtonFunctionality(self, callback)
  local onTouchDown = function()
    self:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_DOWNSTATE))
  end
  local onTouchUp = function()
    self:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON))
    callback()
  end
  local onTouchCancel = function()
    self:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON))
  end
  Button.makeButton(self, onTouchDown, onTouchUp, onTouchCancel)
end

local function addFloatingInvertedButtonFunctionality(self, callback)
  local onTouchDown = function()
    self.buttonLabel:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT_DOWNSTATE))
  end
  local onTouchUp = function()
    self.buttonLabel:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT))
    callback()
  end
  local onTouchCancel = function()
    self.buttonLabel:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT))
  end
  Button.makeButton(self, onTouchDown, onTouchUp, onTouchCancel)
end

local function createButtonLabel(self, text, spriteName, color)
  local labelRowSpacing = 8
  local labelRow = Row.new(labelRowSpacing)

  local buttonFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local buttonLabel = Label.new(text, buttonFont)
  self.buttonLabel = buttonLabel

  if (spriteName ~= nil) then
    local sprite = Sprite.newWithName(spriteName)
    sprite:setEffectiveDims(nil, buttonLabel:getHeight())
    labelRow:insert(sprite)
  end

  labelRow:insert(buttonLabel)

  labelRow:alignChildren()
  labelRow:linkColorAttrToChildren(true)
  labelRow:setColor(unpack(color))

  self:insertAtCenter(labelRow)
end

local function setupNewButton(self, callback)
  addRegularButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

-- Constructor ------------------------

function CrossPostGameButton.new(text, callback, spriteName)
  local self = FilledRoundedRectangle.new(BUTTON_WIDTH, BUTTON_HEIGHT, CORNER_RADIUS, CrossColors.POST_GAME_SCREEN_BUTTON)
  Mixin.mix(self, CrossPostGameButton)

  createButtonLabel(self, text, spriteName, CrossColors.POST_GAME_SCREEN_BUTTON_TEXT)

  addRegularButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end

function CrossPostGameButton.newSimpleTextButton(text, callback, spriteName)
  local self = Group.new(BUTTON_WIDTH, BUTTON_HEIGHT)
  Mixin.mix(self, CrossPostGameButton)

  createButtonLabel(self, text, spriteName, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT)

  addRegularButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end

function CrossPostGameButton.newInverted(text, callback, spriteName)
  local self = FilledRoundedRectangle.new(HALF_BUTTON_WIDTH, BUTTON_HEIGHT, CORNER_RADIUS, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON)
  Mixin.mix(self, CrossPostGameButton)

  local lineWidth = 1
  local outline = HollowRoundedRectangle.new(HALF_BUTTON_WIDTH, BUTTON_HEIGHT, CORNER_RADIUS, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_OUTLINE, lineWidth)
  self:insertAtCenter(outline)

  createButtonLabel(self, text, spriteName, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT)

  addInvertedButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end

function CrossPostGameButton.newFloatingInverted(text, callback, spriteName)
  local self = FilledRoundedRectangle.new(BUTTON_WIDTH, BUTTON_HEIGHT, CORNER_RADIUS, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON)
  Mixin.mix(self, CrossPostGameButton)

  createButtonLabel(self, text, spriteName, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT)

  addFloatingInvertedButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end

function CrossPostGameButton.newPillButton(text, callback, spriteName)
  -- Build labelRow first, as the sizing of self is dependant on what is in labelRow.
  -- Not built in a helper function as we need references to buttonLabel & sprite to get accurate dimensions
  local labelRowSpacing = 16
  local labelRow = Row.new(labelRowSpacing)

  local buttonFont = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
  local buttonLabel = Label.new(text, buttonFont)

  local sprite = nil
  if (spriteName ~= nil) then
    sprite = Sprite.newWithName(spriteName)
    sprite:setEffectiveDims(nil, buttonLabel:getHeight())
    labelRow:insert(sprite)
    sprite:setOpacity(0)
  end
  labelRow:insert(buttonLabel)

  labelRow:alignChildren()
  labelRow:linkColorAttrToChildren(true)
  labelRow:setColor(unpack(CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_TEXT))

  local width = 48
  local height = 40
  local padding = 16
  if (sprite ~= nil) then
    width = width + sprite:getEffectiveWidth() + padding
    height = math.min(height, sprite:getEffectiveHeight() + padding)
  end
  if (buttonLabel ~= nil) then
    width = width + buttonLabel:getEffectiveWidth()
    height = math.min(height, buttonLabel:getEffectiveHeight() + padding)
  end
  local cornerRadius = (height / 2) - 1
  local self = FilledRoundedRectangle.new(width, height, cornerRadius, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON)
  Mixin.mix(self, CrossPostGameButton)

  local lineWidth = 1
  local outline = HollowRoundedRectangle.new(width, height, cornerRadius, CrossColors.POST_GAME_SCREEN_INVERTED_BUTTON_OUTLINE, lineWidth)
  self:insertAtCenter(outline)

  self:insertAtCenter(labelRow)

  addInvertedButtonFunctionality(self, callback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end
