--- CrossReviewScreen
-- @author Created by andrewcousins on May 31, 2023.

class("CrossReviewScreen")

require("CrossPostGameButtonGroup")
require("CrossResetConfirmationScreen")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SOLVED_TEXT = Localizer.get("SOLVED_TIME")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConfirmResetCallback(self, resetCallback)
  return function()
    local confirmationScreen = CrossResetConfirmationScreen.new(resetCallback)
    self:insertAtCenter(confirmationScreen)
  end
end

local function show(self)
  self.gridDisplay:prepareForReview()
  local duration = 0.4
  local ease = AHEaseBezier.newFlat({0.15, 0.00, 0.25, 1.00})
  local animation = AHSpawn.new(AHOpacityTo.new(self.overlay, duration, ease, 0.0))
  animation:addAction(AHDelayed.new(0.2, function()
    self.gridDisplay:showForReview()
    self.gridDisplay:setTouchEnabled(true)
    self.gridDisplay:setClueBarTouchEnabled(true)
  end))
  animation:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createTimerLabel(endTime)
  local timerFont = Font.new(FontFaces.DIN_OT_BOLD, 20)
  local timerColor = CrossColors.COMPLETION_SCREEN_TEXT
  local timerLabel = CrossTimeDisplay.new(timerFont, timerColor)
  timerLabel:setDisplayedTime(endTime)

  local solvedInLabel = Label.new(SOLVED_TEXT, timerFont)
  solvedInLabel:setColor(unpack(timerColor))

  local timerRow = Row.new(solvedInLabel:getEffectiveHeightPart(0.25))
  timerRow:insert(solvedInLabel)
  timerRow:insert(timerLabel)
  timerRow:alignChildren()
  return timerRow
end

local function createCloseButton(continueCallback)
  local closeIcon = Sprite.newWithName("icon_close_review.png")
  local onTouchDown = function()
    closeIcon:setOpacity(0.5)
  end
  local onTouchUp = function()
    closeIcon:setOpacity(1.0)
    continueCallback()
  end
  local onTouchCancel = function()
    closeIcon:setOpacity(1.0)
  end
  local virtualButtonWidth = closeIcon:getWidthPart(2.0)
  local virtualButtonHeight = closeIcon:getHeightPart(2.0)
  Button.addVirtualButton(closeIcon, virtualButtonWidth, virtualButtonHeight, onTouchDown, onTouchUp, onTouchCancel)
  return closeIcon
end

local function createLayout(self, gridDisplay, endTime, loadedCompletedPuzzle, continueCallback, resetCallback)
  local background = FilledRectangle.newFullScreen(CrossColors.COMPLETION_SCREEN_BACKGROUND)
  self:insertAtCenter(background)
  self.background = background

  self.background:takeChild(gridDisplay)

  if loadedCompletedPuzzle then
    local closeButton = createCloseButton(continueCallback)
    background:insert(closeButton)
    closeButton:setX(Screen.safeLeft + closeButton:getWidthPart(1.5))
    closeButton:setY(Screen.safeTop - closeButton:getHeightPart(1.5))
  end

  local withContinueButton = (not loadedCompletedPuzzle)
  local resetConfirmationCallback = getConfirmResetCallback(self, resetCallback)
  local shareCallback = function() GameSession.notifyShareCrossword() end
  local postGameButtonGroup = CrossPostGameButtonGroup.new(shareCallback, resetConfirmationCallback, withContinueButton, continueCallback)
  self.background:insertAtCenter(postGameButtonGroup)
  postGameButtonGroup:setButtonGroupYPosition()

  local gridDisplayLowerBound = postGameButtonGroup:getYTop()
  local gridDisplayUpperBound = self:getHeight() - Screen.safeTopInset
  if (closeButton ~= nil) then
    gridDisplayUpperBound = closeButton:getYBottom()
  end
  local gridDisplayY = (gridDisplayLowerBound + gridDisplayUpperBound) / 2
  gridDisplay:setLoc(self:getWidthPart(0.5), gridDisplayY)
  self.gridDisplay = gridDisplay

  local completionLabel = createTimerLabel(endTime)
  background:insertAtCenter(completionLabel)
  local completionLabelSpacing = completionLabel:getHeightPart(0.75)

  local maxGridDisplayTopY = gridDisplayUpperBound - completionLabel:getHeight() - (completionLabelSpacing * 2)
  if (gridDisplay:getEffectiveYTop() > maxGridDisplayTopY) then
    gridDisplay:offsetY(maxGridDisplayTopY - gridDisplay:getEffectiveYTop())
  end

  completionLabel:setY(gridDisplay:getEffectiveYTop() + completionLabel:getHeightPart(0.5) + completionLabelSpacing)

  local overlay = FilledRectangle.newFullScreen(CrossColors.COMPLETION_SCREEN_BACKGROUND)
  self:insertAtCenter(overlay)
  overlay:bringToFront()
  self.overlay = overlay
end

-- Constructor ------------------------

function CrossReviewScreen.new(gridDisplay, endTime, loadedCompletedPuzzle, continueCallback, resetCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, CrossReviewScreen)

  createLayout(self, gridDisplay, endTime, loadedCompletedPuzzle, continueCallback, resetCallback)
  show(self)

  return self
end
