--- CrossRound
-- @author Created by andrewcousins on March 10, 2022.

class("CrossRound")

require("CrossClue")
require("CrossCellState")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CrossRound.CROSSWORD_DIFFICULTY = {
  EASY = "Easy",
  MEDIUM = "Medium",
  HARD = "Hard"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getCoordinatesForFirstCellInClue(self)
  local currentClue = self:getCurrentClue()
  local clueCellCoordinates = currentClue:getClueCellCoordinates()
  return clueCellCoordinates[1]
end

-- Target column, row must be in the coordinate list.
local function getCoordinateIndexInList(coordinatesList, targetColumn, targetRow)
  for index, cellCoordinates in ipairs(coordinatesList) do
    local column, row = unpack(cellCoordinates)
    if ((column == targetColumn) and (row == targetRow)) then
      return index
    end
  end

  return nil
end

local function getClueIndexForCursor(self)
  local clueIndexes = self.cellsToClueIndex[self.cursorColumn][self.cursorRow]
  for _, clueIndex in ipairs(clueIndexes) do
    local clue = self.clues[clueIndex]
    if (clue:getDirection() == self:getCurrentClue():getDirection()) then
      return clueIndex
    end
  end
  if (#clueIndexes > 0) then
    return clueIndexes[1]
  else
    Log.error("No clue found for cell.")
  end
end

local function getFirstEmptyOrIncorrectCoordinatesForClue(self, clue, shouldStartAtNextCell)
  -- If shouldStartAtNextCell, start checking from where the cursor would naturally advance to
  shouldStartAtNextCell = shouldStartAtNextCell or false

  local clueCellCoordinates = clue:getClueCellCoordinates()
  local clueCellIndex = 1

  if shouldStartAtNextCell then
    for index, cellCoordinates in ipairs(clueCellCoordinates) do
      local column, row = unpack(cellCoordinates)
      if ((column == self.cursorColumn) and (row == self.cursorRow)) then
        clueCellIndex = math.min(index + 1, #clueCellCoordinates)
        break
      end
    end
  end

  for clueCellIndex = clueCellIndex, #clueCellCoordinates do
    local coordinates = clueCellCoordinates[clueCellIndex]
    local column, row = unpack(coordinates)
    local cell = self.gridState[column][row]
    if (cell:isEmpty() or cell:getIsIncorrect()) then
      return coordinates
    end
  end

  return nil
end

local function updateCursorForClue(self)
  local clue = self:getCurrentClue()
  local firstEmptyCoordinates = getFirstEmptyOrIncorrectCoordinatesForClue(self, clue)
  if (firstEmptyCoordinates == nil) then
    self.cursorColumn, self.cursorRow = unpack(clue:getFirstGridCoordinates())
  else
    self.cursorColumn, self.cursorRow = unpack(firstEmptyCoordinates)
  end
end

local function incrementClueIndex(self)
  self.clueIndex = self.clueIndex + 1
  if (self.clueIndex > #self.clues) then
    self.clueIndex = 1
  end
end

local function decrementClueIndex(self)
  self.clueIndex = self.clueIndex - 1
  if (self.clueIndex < 1) then
    self.clueIndex = #self.clues
  end
end

local function goToNextIncompleteClue(self, goBackwards, alwaysAdvanceClueIndex)
  goBackwards = goBackwards or false
  alwaysAdvanceClueIndex = alwaysAdvanceClueIndex or false
  local startingClueIndex = self.clueIndex
  repeat
    if goBackwards then
      decrementClueIndex(self)
    else
      incrementClueIndex(self)
    end
    local currentClue = self:getCurrentClue()
    local currentClueIsIncomplete = (getFirstEmptyOrIncorrectCoordinatesForClue(self, currentClue) ~= nil)
    local backAtStartingIndex = (startingClueIndex == self.clueIndex)
  until (currentClueIsIncomplete or backAtStartingIndex)

  if (alwaysAdvanceClueIndex and (startingClueIndex == self.clueIndex)) then
    if goBackwards then
      decrementClueIndex(self)
    else
      incrementClueIndex(self)
    end
  end
end

--[[ Cursor movement algorithm:
If cursor is at the last cell of the clue
  If there are empty or incorrect spaces in the clue
    Move to first empty space
  Elseif grid is full
    Do not move the cursor
  Else
    Move to first empty cell in the next incomplete clue
Else
  If just rewrote cell,
    move to the next cell in the clue
  If filled empty cell,
    Move to next empty or incorrect cell.
    If no empty cells,
      Move to first empty cell in the next incomplete clue
--]]
local function advanceCursor(self, didOverwriteCell)
  local clue = self:getCurrentClue()
  local clueCellCoordinates = clue:getClueCellCoordinates()
  local clueCellCoordinatesIndex = getCoordinateIndexInList(clueCellCoordinates, self.cursorColumn, self.cursorRow)
  local isAtEndOfClue = (clueCellCoordinatesIndex == #clueCellCoordinates)
  if isAtEndOfClue then
    local firstEmptyCoordinatesInClue = getFirstEmptyOrIncorrectCoordinatesForClue(self, clue)
    if (firstEmptyCoordinatesInClue ~= nil) then
      self.cursorColumn, self.cursorRow = unpack(firstEmptyCoordinatesInClue)
    else
      local startingClueIndex = self.clueIndex
      goToNextIncompleteClue(self)
      if (startingClueIndex ~= self.clueIndex) then
        updateCursorForClue(self)
      else
        -- Do nothing.
        return
      end
    end
  else
    if didOverwriteCell then
      self.cursorColumn, self.cursorRow = unpack(clueCellCoordinates[clueCellCoordinatesIndex + 1])
    else
      local shouldStartAtNextCell = true
      local nextEmptyCoordinatesInClue = getFirstEmptyOrIncorrectCoordinatesForClue(self, clue, shouldStartAtNextCell)
      if (nextEmptyCoordinatesInClue == nil) then
        shouldStartAtNextCell = false
        nextEmptyCoordinatesInClue = getFirstEmptyOrIncorrectCoordinatesForClue(self, clue, shouldStartAtNextCell)
      end
      if (nextEmptyCoordinatesInClue == nil) then
        local startingClueIndex = self.clueIndex
        goToNextIncompleteClue(self)
        if (startingClueIndex ~= self.clueIndex) then
          updateCursorForClue(self)
        end
      else
        self.cursorColumn, self.cursorRow = unpack(nextEmptyCoordinatesInClue)
      end
    end
  end
end

-- Go to the next cell in the clue
-- If at the end of the clue, go to the first cell of the current clue
local function advanceCursorWithinClue(self)
  local clue = self:getCurrentClue()
  local clueCellCoordinates = clue:getClueCellCoordinates()
  local clueCellCoordinatesIndex = getCoordinateIndexInList(clueCellCoordinates, self.cursorColumn, self.cursorRow)
  local isAtEndOfClue = (clueCellCoordinatesIndex == #clueCellCoordinates)
  if isAtEndOfClue then
    clueCellCoordinatesIndex = 1
  else
    clueCellCoordinatesIndex = clueCellCoordinatesIndex + 1
  end
  self.cursorColumn, self.cursorRow = unpack(clueCellCoordinates[clueCellCoordinatesIndex])
end

-- Go to the next cell in the clue
-- If at the end of the clue, go to the first cell of the next clue
local function advanceCursorToNextCell(self)
  local clue = self:getCurrentClue()
  local clueCellCoordinates = clue:getClueCellCoordinates()
  local clueCellCoordinatesIndex = getCoordinateIndexInList(clueCellCoordinates, self.cursorColumn, self.cursorRow)
  local isAtEndOfClue = (clueCellCoordinatesIndex == #clueCellCoordinates)
  if isAtEndOfClue then
    incrementClueIndex(self)
    clue = self:getCurrentClue()
    clueCellCoordinates = clue:getClueCellCoordinates()
    clueCellCoordinatesIndex = 1
  else
    clueCellCoordinatesIndex = clueCellCoordinatesIndex + 1
  end
  self.cursorColumn, self.cursorRow = unpack(clueCellCoordinates[clueCellCoordinatesIndex])
end

local function getCoordinatesOfPreviousSpaceInClue(self)
  local clue = self:getCurrentClue()
  local clueCoordinates = clue:getClueCellCoordinates()
  for index = 2, #clueCoordinates do
    local column, row = unpack(clueCoordinates[index])
    if ((column == self.cursorColumn) and (row == self.cursorRow)) then
      return clueCoordinates[index - 1]
    end
  end
end

local function retreatCursor(self)
  local newCoordinates = getCoordinatesOfPreviousSpaceInClue(self)
  if (newCoordinates == nil) then
    self.clueIndex = self.clueIndex - 1
    if (self.clueIndex < 1) then
      self.clueIndex = #self.clues
    end
    local clue = self:getCurrentClue()
    local clueCoordinates = clue:getClueCellCoordinates()
    newCoordinates = clueCoordinates[#clueCoordinates]
  end
  self.cursorColumn, self.cursorRow = unpack(newCoordinates)
end

local function moveCursorInDirection(self, direction)
  local modifier = CrossDirection.DIRECTION_TO_MODIFIER[direction]
  local newColumn = self.cursorColumn + modifier[1]
  local newRow = self.cursorRow + modifier[2]
  self:setCursorCoordinates(newColumn, newRow)
end

local function saveState(self)
  if self.savesEnabled then
    if (self.saveManager ~= nil) then
      self.saveManager:save()
    else
      Log.warn("No save manager has been set - not saving progress.")
    end
  end
end

local function handleCompletedGridIfNeeded(self, didRewriteLastCell, loadedSavedPuzzle)
  didRewriteLastCell = didRewriteLastCell or false
  loadedSavedPuzzle = loadedSavedPuzzle or false
  local isError = false
  for column = 1, #self.gridState do
    for row = 1, #self.gridState[column] do
      local cell = self.gridState[column][row]
      if (not cell:getIsWall()) then
        if (not cell:hasContent()) then
          return
        elseif (cell:getContent() ~= self:getSolutionForCell(column, row)) then
          isError = true
        end
      end
    end
  end
  if isError then
    if ((not didRewriteLastCell) and (not loadedSavedPuzzle)) then
      self.gridCompletedButIncorrectCallback()
    end
  else
    self.completedCorrectly = true
    saveState(self)
    self.roundCompletedCallback(loadedSavedPuzzle)
  end
end

local function getPercentCompleted(self)
  local totalCells = 0
  local filledCells = 0
  for column = 1, #self.gridState do
    for row = 1, #self.gridState[column] do
      local cell = self.gridState[column][row]
      if (not cell:getIsWall()) then
        totalCells = totalCells + 1
        if (cell:hasContent()) then
          filledCells = filledCells + 1
        end
      end
    end
  end
  return filledCells / totalCells
end

local function checkCell(self, column, row)
  local correctContent = self:getSolutionForCell(column, row)
  local cellState = self.gridState[column][row]
  local hasContent = cellState:hasContent()
  local isContentIncorrect = (cellState:getContent() ~= correctContent)
  if (hasContent and isContentIncorrect) then
    cellState:setIsIncorrect(true)
  else
    cellState:setIsIncorrect(false)
  end
end

local function checkGrid(self)
   for column = 1, #self.gridState do
    for row = 1, #self.gridState[column] do
      checkCell(self, column, row)
    end
  end
end

local function updateContentAtLocation(self, column, row, content, isReveal, delay)
  if self.gridState[column][row]:getRevealed() then
    -- Cannot overwrite already revealed cells
    return false
  end
  isReveal = isReveal or false
  local didOverwriteCell = self.gridState[column][row]:hasContent()
  local didChangeContent = (self.gridState[column][row]:getContent() ~= content)
  self.gridState[column][row]:setContent(content, isReveal)
  if self.autocheckEnabled then
    checkCell(self, self.cursorColumn, self.cursorRow)
  elseif didChangeContent then
    -- Remove the incorrect marker if autocheck is disabled, regardless of if this is the correct
    -- answer. If the updated content is the same as what is displayed, leave the incorrect marker.
     self.gridState[column][row]:setIsIncorrect(false)
  end
  if isReveal then
    delay = delay or 0
    self.gridState[column][row]:setRevealDelay(delay)
  end
  self.gridDisplay:update()

  return didOverwriteCell
end

local function isCellCorrect(self, column, row)
  local cell = self.gridState[column][row]
  local solution = self:getSolutionForCell(column, row)
  if cell:hasContent() then
    return (cell:getContent() == solution)
  end
  return false
end

local function revealLetterAtLocation(self, column, row, delay)
  local correctContent = self:getSolutionForCell(column, row)
  local isReveal = true
  self.gridState[column][row]:setIsIncorrect(false)
  return updateContentAtLocation(self, column, row, correctContent, isReveal, delay)
end

local function updateGridAndSaveState(self)
  self.gridDisplay:update()
  saveState(self)
end

local function refreshGrid(self)
  for column = 1, #self.gridState do
    for row = 1, #self.gridState[column] do
      self.gridDisplay:updateCell(column, row)
    end
  end

  self.gridDisplay:update()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossRound:getCellState(column, row)
  return self.gridState[column][row]
end

function CrossRound:getSolutionForCell(column, row)
  return self.gridSolution[column][row]
end

function CrossRound:getCluesForCell(column, row)
  return self.cellsToClueIndex[column][row]
end

function CrossRound:getClues()
  return self.clues
end

function CrossRound:getColumnCount()
  return self.columnCount
end

function CrossRound:getRowCount()
  return self.rowCount
end

function CrossRound:setGridDisplay(gridDisplay)
  self.gridDisplay = gridDisplay
end

function CrossRound:updateContentAtCursor(content)
  -- If cell is revealed, content cannot be updated
  if (not self.gridState[self.cursorColumn][self.cursorRow]:getRevealed()) then
    local didOverwriteCell = updateContentAtLocation(self, self.cursorColumn, self.cursorRow, content)
    handleCompletedGridIfNeeded(self, didOverwriteCell)

    advanceCursor(self, didOverwriteCell)
  else
    advanceCursorToNextCell(self)
  end
  updateGridAndSaveState(self)
end

function CrossRound:clearContentAtCursor()
  local currentlySelectedCell = self:getCellState(self.cursorColumn, self.cursorRow)
  -- If cell is revealed, content cannot be updated
  if not currentlySelectedCell:getRevealed() then
    if currentlySelectedCell:isEmpty() then
      retreatCursor(self)
    end
    updateContentAtLocation(self, self.cursorColumn, self.cursorRow, "")
  end
  updateGridAndSaveState(self)
end

function CrossRound:clearContentAtCursorAndAdvance()
  -- If cell is revealed, content cannot be updated
  if (not self.gridState[self.cursorColumn][self.cursorRow]:getRevealed()) then
    updateContentAtLocation(self, self.cursorColumn, self.cursorRow, "")

    advanceCursorWithinClue(self)
  end
  updateGridAndSaveState(self)
end

function CrossRound:getCurrentClue()
  return self.clues[self.clueIndex]
end

function CrossRound:moveToNextClue()
  local goBackwards = false
  local alwaysAdvanceClueIndex = true
  goToNextIncompleteClue(self, goBackwards, alwaysAdvanceClueIndex)
  updateCursorForClue(self)
  updateGridAndSaveState(self)
end

function CrossRound:moveToPreviousClue()
  local goBackwards = true
  local alwaysAdvanceClueIndex = true
  goToNextIncompleteClue(self, goBackwards, alwaysAdvanceClueIndex)
  updateCursorForClue(self)
  updateGridAndSaveState(self)
end

function CrossRound:moveToAlternateClue()
  for _, index in ipairs(self.cellsToClueIndex[self.cursorColumn][self.cursorRow]) do
    if (index ~= self.clueIndex) then
      self.clueIndex = index
      break
    end
  end
  updateGridAndSaveState(self)
end

function CrossRound:moveCursorInDirection(direction)
  if (CrossDirection.isHorizontal(direction) ~= self:getCurrentClue():isAcross()) then
    self:moveToAlternateClue()
  else
    moveCursorInDirection(self, direction)
  end
end

function CrossRound:moveToClueForCursor()
  self.clueIndex = getClueIndexForCursor(self)
  updateGridAndSaveState(self)
end

function CrossRound:getCursorCoordinates()
  return {self.cursorColumn, self.cursorRow}
end

function CrossRound:setCursorCoordinates(column, row)
  if ((0 < column) and (column <= self.columnCount)) then
    if ((0 < row) and (row <= self.rowCount)) then
      if not self:getCellState(column, row):getIsWall() then
        self.cursorColumn = column
        self.cursorRow = row
        self:moveToClueForCursor()
      end
    end
  end
end

function CrossRound:cellTouchCallback(column, row)
  if ((self.cursorColumn == column) and (self.cursorRow == row)) then
    self:moveToAlternateClue()
  else
    self.cursorColumn = column
    self.cursorRow = row
    self:moveToClueForCursor()
  end
end

function CrossRound:revealLetter()
  CrossAnalytics.puzzleSingleHintUsed()
  self.hintsUsedCount = self.hintsUsedCount + 1
  if (not isCellCorrect(self, self.cursorColumn, self.cursorRow)) then
    local didOverwriteCell = revealLetterAtLocation(self, self.cursorColumn, self.cursorRow)
    handleCompletedGridIfNeeded(self, didOverwriteCell)
  else
    local cellCoordinates = {self.cursorColumn, self.cursorRow}
    local cellCoordinatesList = {cellCoordinates}
    self.gridDisplay:getWordPulseAnimation(cellCoordinatesList):start()
  end

  updateGridAndSaveState(self)
end

function CrossRound:revealWord()
  CrossAnalytics.puzzleWordHintUsed()
  self.hintsUsedCount = self.hintsUsedCount + 1
  local currentClue = self:getCurrentClue()
  local clueCellCoordinates = currentClue:getClueCellCoordinates()
  self.setTouchEnabledCallback(false)
  local delay = 0
  local wordWasCorrect = true
  local wasEntireRewrite = true
  for _, coordinates in ipairs(clueCellCoordinates) do
    local column, row = unpack(coordinates)
    if (not isCellCorrect(self, column, row)) then
      local didOverwriteCell = revealLetterAtLocation(self, column, row, delay)
      delay = delay + 0.083
      wordWasCorrect = false
      wasEntireRewrite = wasEntireRewrite and didOverwriteCell
    end
  end
  if wordWasCorrect then
    self.gridDisplay:getWordPulseAnimation(clueCellCoordinates):start()
  else
    handleCompletedGridIfNeeded(self, wasEntireRewrite)
  end
  updateGridAndSaveState(self)
  -- Lock out input during hint animation. The added delay should be slightly shorter than
  -- the animation duration of CrossCellDisplay's updateRevealedIcon() animations
  AHDelayed.new(delay + 0.25, function() self.setTouchEnabledCallback(true) end):start()
end

function CrossRound:getPuzzleIdentifier()
  return self.puzzleIdentifier
end

function CrossRound:getConceptIdentifier()
  return self.conceptIdentifier
end

function CrossRound:setCrosswordDifficulty(crosswordDifficulty)
  self.crosswordDifficulty = crosswordDifficulty
  GameSession.setCrosswordDifficulty(crosswordDifficulty)
  self.gridDisplay:updateClueDifficulties(crosswordDifficulty)
  wait(self.gridDisplay:getClueLabelBuilderCoroutine())
  refreshGrid(self)
end

function CrossRound:getCrosswordDifficulty()
  return self.crosswordDifficulty
end

function CrossRound:getSupportedDifficulties()
  return self.supportedDifficulties
end

function CrossRound:getAskForFeedback()
  return self.askForFeedback
end

function CrossRound:getPercentCompleted()
  return getPercentCompleted(self)
end

function CrossRound:getNumberOfHintsUsed()
  return self.hintsUsedCount
end

-- param: autocheckEnabled is deprecated, and is only used by old versions of saved data
function CrossRound:loadSavedGridData(gridData, autocheckEnabled)
  self.gridDisplay:setLoadingState(true)
  for columnIndex, column in ipairs(gridData) do
    for rowIndex, savedCellState in ipairs(column) do
      local cellState = self:getCellState(columnIndex, rowIndex)
      local content = savedCellState[CrossSaveManager.CELL_CONTENT_KEY]
      if (content == nil) then
        content = savedCellState[CrossSaveManager.CELL_CONTENT_KEY_OLD]
      end
      local isRevealed = savedCellState[CrossSaveManager.CELL_REVEALED_KEY]
      if (isRevealed == nil) then
        isRevealed = savedCellState[CrossSaveManager.CELL_REVEALED_KEY_OLD]
      end
      cellState:setContent(content, isRevealed)
      local isIncorrect = savedCellState[CrossSaveManager.CELL_INCORRECT_KEY]
      if isIncorrect then
        cellState:setIsIncorrect(isIncorrect)
      end
    end
  end

  if (autocheckEnabled ~= nil) then
    -- We no longer use saved data to store autocheck, but if an older save file
    -- had autocheck enabled, reveal any errors to keep the state the same
    if autocheckEnabled then
      checkGrid(self)
    end
  end

  refreshGrid(self)
  local loadedSavedPuzzle = true
  handleCompletedGridIfNeeded(self, nil, loadedSavedPuzzle)
  self.gridDisplay:setLoadingState(false)
end

function CrossRound:setSaveManager(saveManager)
  self.saveManager = saveManager
end

function CrossRound:isCompletedCorrectly()
  return self.completedCorrectly
end

function CrossRound:reset()
  for _, column in ipairs(self.gridState) do
    for _, cellState in ipairs(column) do
      cellState:reset()
    end
  end

  self.clueIndex = 1
  self.cursorColumn, self.cursorRow = unpack(getCoordinatesForFirstCellInClue(self))

  self.completedCorrectly = false
  self.autocheckEnabled = false
  self.savesEnabled = true

  refreshGrid(self)
  updateGridAndSaveState(self)
end

function CrossRound:setSaveEnabled(enabled)
  self.savesEnabled = enabled
end

function CrossRound:checkForErrors()
  checkGrid(self)
  refreshGrid(self)
  updateGridAndSaveState(self)
end

function CrossRound:enableAutocheck(enabled, isLoadingState)
  self.autocheckEnabled = enabled
  if (not isLoadingState) then
    GameSession.setCrosswordAutoCheckEnabled(enabled)
    CrossAnalytics.puzzleAutocheckToggled()
  end
  if enabled then
    checkGrid(self)
  end
  refreshGrid(self)
  updateGridAndSaveState(self)
end

function CrossRound:isAutocheckEnabled()
  return self.autocheckEnabled
end

function CrossRound:getIsFromArchive()
  return self.isFromArchive
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CrossRound.new(puzzleIdentifier, conceptIdentifier, askForFeedback, isFromArchive, crosswordDifficulty, supportedDifficulties, columnCount, rowCount, gridState, gridSolution, clues, cellsToClueIndex, roundCompletedCallback, gridCompletedButIncorrectCallback, setTouchEnabledCallback)
  local self = {}
  Mixin.mix(self, CrossRound)

  self.puzzleIdentifier = puzzleIdentifier
  self.conceptIdentifier = conceptIdentifier
  self.crosswordDifficulty = crosswordDifficulty
  self.supportedDifficulties = supportedDifficulties
  self.askForFeedback = askForFeedback
  self.isFromArchive = isFromArchive

  self.columnCount = columnCount
  self.rowCount = rowCount
  self.gridState = gridState
  self.gridSolution = gridSolution
  self.clues = clues
  self.cellsToClueIndex = cellsToClueIndex
  self.roundCompletedCallback = roundCompletedCallback
  self.gridCompletedButIncorrectCallback = gridCompletedButIncorrectCallback
  self.setTouchEnabledCallback = setTouchEnabledCallback

  self.autocheckEnabled = false
  self.savesEnabled = true

  self.clueIndex = 1
  self.cursorColumn, self.cursorRow = unpack(getCoordinatesForFirstCellInClue(self))

  self.completedCorrectly = false
  self.hintsUsedCount = 0

  return self
end
