--- CrossRoundGenerator
-- @author Created by andrewcousins on April 12, 2022.

class("CrossRoundGenerator")

require("CrossRound")
require("CrossCellState")
require("CrossClue")
require("ExperimentsConfiguration")

-- Note: some 2D arrays are read [row][column] and others are read [column][row]. This is because:
--      Puzzle grid arrays from concepts are organized by:
--          [row (top to bottom)][column (left to right)]
--      This keeps them human readable and easy to edit.
--
--      This game organizes grid arrays with
--          [column (left to right)][row (top to buttom)]
--      This follows the traditional graphics display convention of having coordinates in
--      [x][y], starting from the top left of the screen.



-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local PLACEHOLDER_IDENTIFIER = "CrosswordIdentifierPlaceholder"
local PLACEHOLDER_PUZZLE_SETUP_STRING = '{"puzzle": {"id": "cw_cld_batch9_p128", "version": 2, "ask_for_feedback": true, "dimensions": {"width": 5, "height": 5}, "solution": [["#", "#", "H", "E", "N"], ["#", "P", "I", "P", "E"], ["A", "U", "D", "I", "O"], ["S", "P", "E", "C", "#"], ["H", "A", "S", "#", "#"]], "puzzle": [["#", "#", 1, 2, 3], ["#", 4, 0, 0, 0], [5, 0, 0, 0, 0], [6, 0, 0, 0, "#"], [7, 0, 0, "#", "#"]], "clues": {"Easy": {"across": [[1, "Barnyard animal that lays eggs"], [4, "Piece of plumbing"], [5, "Related to sound"], [6, "Shortening of specification"], [7, "I is to have as she is to ___"]], "down": [[1, "Rearrangement of shied (hint: conceals)"], [2, "Very big or significant"], [3, "Word beginning meaning new before -natal"], [4, "Singular of pupae"], [5, "What might be left in a fireplace"]]}, "Medium": {"across": [[1, "Barnyard animal that lays eggs"], [4, "Piece of plumbing"], [5, "Related to sound"], [6, "Shortening of specification"], [7, "I is to have as she is to ___"]], "down": [[1, "Rearrangement of shied (hint: conceals)"], [2, "Very big or significant"], [3, "Word beginning meaning new before -natal"], [4, "Singular of pupae"], [5, "What might be left in a fireplace"]]}, "Hard": {"across": [[1, "Barnyard animal that lays eggs"], [4, "Piece of plumbing"], [5, "Related to sound"], [6, "Shortening of specification"], [7, "I is to have as she is to ___"]], "down": [[1, "Rearrangement of shied (hint: conceals)"], [2, "Very big or significant"], [3, "Word beginning meaning new before -natal"], [4, "Singular of pupae"], [5, "What might be left in a fireplace"]]}}}}'

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createGridState(columnCount, rowCount, puzzleLayout)
  local gridState = {}
  for columnIndex = 1, columnCount do
    gridState[columnIndex] = {}
    for rowIndex = 1, rowCount do
      local cellContents = tostring(puzzleLayout[rowIndex][columnIndex]) -- intentionally reversed
      local isWall = (cellContents == "#")
      local clueLabel = cellContents
      if ((clueLabel == "0") or isWall) then
        clueLabel = ""
      end
      gridState[columnIndex][rowIndex] = CrossCellState.new(isWall, clueLabel)
    end
  end
  return gridState
end

local function createGridSolution(columnCount, rowCount, solution)
  local gridSolution = {}
  for columnIndex = 1, columnCount do
    gridSolution[columnIndex] = {}
    for rowIndex = 1, rowCount do
      gridSolution[columnIndex][rowIndex] = tostring(solution[rowIndex][columnIndex]) -- intentionally reversed
    end
  end
  return gridSolution
end

local function getSortedClueLabels(cluesToAdd)
  local sortFunction = function(a, b)
    return (a < b)
  end
  local sortedClueLabels = TableUtils.getKeys(cluesToAdd)
  table.sort(sortedClueLabels, sortFunction)
  return sortedClueLabels
end

local function addCluesToList(cluesToAdd, clueDirection, clueList, supportedDifficulties)
  local sortedClueLabels = getSortedClueLabels(cluesToAdd)
  for _, clueLabel in ipairs(sortedClueLabels) do
    local clueTextByDifficulty = cluesToAdd[clueLabel]
    local direction = clueDirection
    local newClue = CrossClue.new(clueLabel, clueTextByDifficulty, direction, supportedDifficulties)
    table.insert(clueList, newClue)
  end
  return clueList
end

local function createCombinedClueList(clueData, supportedDifficulties)
  local clueList = {}

  local acrossClues = clueData.across
  local downClues = clueData.down
  clueList = addCluesToList(acrossClues, CrossClue.DIRECTION.ACROSS, clueList, supportedDifficulties)
  clueList = addCluesToList(downClues, CrossClue.DIRECTION.DOWN, clueList, supportedDifficulties)

  return clueList
end

local function createClueList(clues, supportedDifficulties)
  local clueData = {across = {}, down = {}}
  for crosswordDifficulty, cluesForDifficulty in pairs(clues) do
    for direction, cluesForDirection in pairs(cluesForDifficulty) do
      for _, clue in ipairs(cluesForDirection) do
        local clueLabel = clue[1]
        local clueText = clue[2]
        if (not clueData[direction][clueLabel]) then
          clueData[direction][clueLabel] = {}
        end
        clueData[direction][clueLabel][crosswordDifficulty] = clueText
      end
    end
  end
  return createCombinedClueList(clueData, supportedDifficulties)
end

local function getSupportedDifficulties(clues)
  ExperimentsConfiguration.CROSSWORDS_DIFFICULTY_TOGGLE_EXPERIMENT:reportExperimentExposed()
  if ExperimentsConfiguration.CROSSWORDS_DIFFICULTY_TOGGLE_EXPERIMENT:isControl() then
    return {CrossRound.CROSSWORD_DIFFICULTY.MEDIUM}
  else
    return TableUtils.getKeys(clues)
  end
end

local function createCellsToClueIndex(gridState, clueList)
  local cellsToClueIndex = {}
  for columnIndex = 1, #gridState do
    cellsToClueIndex[columnIndex] = {}
    for rowIndex = 1, #gridState[columnIndex] do
      cellsToClueIndex[columnIndex][rowIndex] = {}
    end
  end

  for columnIndex = 1, #gridState do
    for rowIndex = 1, #gridState[columnIndex] do
      local cell = gridState[columnIndex][rowIndex]
      if cell:hasClueNumber() then
        local cellClueNumber = cell:getClueNumber()
        for clueListIndex, clue in ipairs(clueList) do
          local clueNumber = clue:getClueNumber()
          if (tostring(clueNumber) == cellClueNumber) then
            local clueColumn = columnIndex
            local clueRow = rowIndex
            while ((clueColumn <= #gridState) and (clueRow <= #gridState[clueColumn])) do
              local cell = gridState[clueColumn][clueRow]
              if cell:getIsWall() then
                break
              end

              clue:addCellCoordinatesToClue(clueColumn, clueRow)
              table.insert(cellsToClueIndex[clueColumn][clueRow], clueListIndex)

              if clue:isAcross() then
                clueColumn = clueColumn + 1
              else
                clueRow = clueRow + 1
              end
            end
          end
        end
      end
    end
  end

  return cellsToClueIndex
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CrossRoundGenerator.generateRound(roundCompletedCallback, gridCompletedButIncorrectCallback, setTouchEnabledCallback)
  local puzzleIdentifier = GameSession.getCrosswordIdentifier()
  local puzzleSetupString = ""

  if (puzzleIdentifier ~= "") then
    puzzleSetupString = GameSession.getCrosswordSetupData(puzzleIdentifier)
  else
    -- Let Achilles have a puzzle to show
    puzzleSetupString = PLACEHOLDER_PUZZLE_SETUP_STRING
  end

  local puzzleSetupData = MOAIJsonParser.decode(puzzleSetupString)
  local puzzle = puzzleSetupData.puzzle

  local version = puzzle.version
  if ((version == nil) or (version == "")) then
    version = 1
  end

  local conceptIdentifier = puzzle.id
  local crosswordDifficulty = GameSession.getCrosswordDifficulty()
  if ((crosswordDifficulty == nil) or (crosswordDifficulty == "")) then
    if (puzzle.difficulty ~= nil) then
      -- Support v1 puzzles
      crosswordDifficulty = puzzle.difficulty
    else
      -- Only used as a fallback for Achilles
      crosswordDifficulty = CrossRound.CROSSWORD_DIFFICULTY.MEDIUM
    end
  end
  local askForFeedback = puzzle.ask_for_feedback
  local isFromArchive = GameSession.getCrosswordIsFromArchive()

  local puzzleDimensions = puzzle.dimensions
  local columnCount = puzzleDimensions.width
  local rowCount = puzzleDimensions.height

  local puzzleLayout = puzzle.puzzle

  local gridState = createGridState(columnCount, rowCount, puzzleLayout)

  local solution = puzzle.solution
  local gridSolution = createGridSolution(columnCount, rowCount, solution)

  local clues = puzzle.clues
  if (version == 1) then
    clues = {[crosswordDifficulty] = clues}
  end
  local supportedDifficulties = getSupportedDifficulties(clues)

  if (not TableUtils.contains(supportedDifficulties, crosswordDifficulty)) then
    crosswordDifficulty = supportedDifficulties[1]
  end

  local clueList = createClueList(clues, supportedDifficulties)

  local cellsToClueIndex = createCellsToClueIndex(gridState, clueList)

  return CrossRound.new(puzzleIdentifier, conceptIdentifier, askForFeedback, isFromArchive, crosswordDifficulty, supportedDifficulties, columnCount, rowCount, gridState, gridSolution, clueList, cellsToClueIndex, roundCompletedCallback, gridCompletedButIncorrectCallback, setTouchEnabledCallback)
end

-- Used for testing in TestCrossRound.lua
function CrossRoundGenerator.generateTestRound()
  local roundCompletedCallback = function() end
  local gridCompletedButIncorrectCallback = function() end
  local setTouchEnabledCallback = function() end
  local puzzleIdentifier = "test_cross_puzzle_identifier"
  local conceptIdentifier = "test_cross_concept_identifier"
  local crosswordDifficulty = CrossRound.CROSSWORD_DIFFICULTY.MEDIUM
  local askForFeedback = false
  local isFromArchive = false

  local columnCount = 5
  local rowCount = 5

  local puzzleLayout = {{1, 2, 3, 4, 5}, {6, 0, 0, 0, 0}, {7, 0, 0, 0, 0}, {8, 0, 0, 0, 0}, {9, 0, 0, 0, 0}}
  local gridState = createGridState(columnCount, rowCount, puzzleLayout)

  local solution = {{"A", "A", "A", "A", "A"}, {"A", "A", "A", "A", "A"}, {"A", "A", "A", "A", "A"}, {"A", "A", "A", "A", "A"}, {"A", "A", "A", "A", "A"}}
  local gridSolution = createGridSolution(columnCount, rowCount, solution)

  local clues = {Medium = {across = {{1, "Test clue"}, {6, "Test clue"}, {7, "Test clue"}, {8, "Test clue"}, {9, "Test clue"}}, down = {{1, "Test clue"}, {2, "Test clue"}, {3, "Test clue"}, {4, "Test clue"}, {5, "Test clue"}}}}
  local clueList = createClueList(clues)

  local cellsToClueIndex = createCellsToClueIndex(gridState, clueList)

  local supportedDifficulties = getSupportedDifficulties(clues)

  return CrossRound.new(puzzleIdentifier, conceptIdentifier, askForFeedback, isFromArchive, crosswordDifficulty, supportedDifficulties, columnCount, rowCount, gridState, gridSolution, clueList, cellsToClueIndex, roundCompletedCallback, gridCompletedButIncorrectCallback, setTouchEnabledCallback)
end
