--- CrossSaveManager
-- @author Created by andrewcousins on April 21, 2022.

class("CrossSaveManager")

require("CrossSaveValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- "time_elapsed" is also used in CoreMS, and must be kept in sync and
-- compatabile with existing crossword saves
CrossSaveManager.TIME_ELAPSED_KEY = "time_elapsed"
CrossSaveManager.GRID_DATA_KEY = "grid_data"
CrossSaveManager.AUTO_CHECK_ENABLED = "autocheck"

-- We shortened these old keys to reduce file size. We continue to support them
-- to let users load crossword saves from the early rollout
CrossSaveManager.CELL_CONTENT_KEY_OLD = "cell_content"
CrossSaveManager.CELL_REVEALED_KEY_OLD = "cell_revealed"

CrossSaveManager.CELL_CONTENT_KEY = "c"
CrossSaveManager.CELL_REVEALED_KEY = "r"
CrossSaveManager.CELL_INCORRECT_KEY = "x"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addGridData(state, round)
  local columnCount = round:getColumnCount()
  local rowCount = round:getRowCount()
  local gridData = {}
  for columnIndex = 1, columnCount do
    gridData[columnIndex] = {}
    for rowIndex = 1, rowCount do
      local cell = round:getCellState(columnIndex, rowIndex)
      local cellState = {}
      if not cell:getIsWall() then
        -- If a value in a table is set to nil, it won't be included
        -- in the JSON string.
        local cellContent = cell:getContent()
        if (cellContent == nil) then
          cellContent = ""
        end
        local isRevealed = cell:getRevealed()
        if (isRevealed == nil) then
          isRevealed = false
        end
        local isIncorrect = cell:getIsIncorrect()
        if (isIncorrect == nil) then
          isIncorrect = false
        end
        -- Saving isRevealed and isIncorrect for all cells significantly increases the size of the save json,
        -- but this is not currently an issue. A future update could ignore these fields if they are false,
        -- or find some other form of compression to save space.
        cellState = {
          [CrossSaveManager.CELL_CONTENT_KEY] = cellContent,
          [CrossSaveManager.CELL_REVEALED_KEY] = isRevealed,
          [CrossSaveManager.CELL_INCORRECT_KEY] = isIncorrect
        }
      end
      gridData[columnIndex][rowIndex] = cellState
    end
  end
  state[CrossSaveManager.GRID_DATA_KEY] = gridData
end

local function addTimerData(state, timer)
  local timeElapsed = timer:getTime()
  local timeToPreviousSecond = math.floor(timeElapsed)
  state[CrossSaveManager.TIME_ELAPSED_KEY] = timeToPreviousSecond
end

local function addAutocheckData(state, round)
  state[CrossSaveManager.AUTO_CHECK_ENABLED] = round:isAutocheckEnabled()
end

local function buildState(self)
  local round = self.round
  local timer = self.timer

  local state = {}
  addGridData(state, round)
  addTimerData(state, timer)
  addAutocheckData(state, round)

  return state
end

local function loadValidatedState(self, savedState)
  self.timer:setInitialTime(savedState[CrossSaveManager.TIME_ELAPSED_KEY])
  local autocheckEnabled = savedState[CrossSaveManager.AUTO_CHECK_ENABLED]
  self.round:loadSavedGridData(savedState[CrossSaveManager.GRID_DATA_KEY], autocheckEnabled)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossSaveManager:save()
  local state = buildState(self)
  self.saveValidator:validateSave(state)
  if (not self.saveValidator:getPassed()) then
    Log.warn("Invalid save state built, cancelling save.")
    Log.warn(self.saveValidator:getMessage())
    return
  end
  local stateAsJSON = MOAIJsonParser.encode(state)
  GameSession.saveCrosswordData(self.round:getPuzzleIdentifier(), stateAsJSON)
end

function CrossSaveManager:load()
  local savedData = GameSession.getCrosswordSavedData(self.round:getPuzzleIdentifier())
  if ((savedData == nil) or (savedData == "")) then
    Log.verbose("No saved data found.")
    return
  end

  Log.verbose("Loading saved data...")
  local savedState = MOAIJsonParser.decode(savedData)
  if (savedState == nil) then
    Log.warn("Load failed, unable to decode saved data.")
    return
  end

  self.saveValidator:validateSave(savedState)
  if self.saveValidator:getPassed() then
    Log.verbose("Valid save found, restoring state.")
  else
    Log.warn("Invalid save found, failed to restore state.")
    Log.warn(self.saveValidator:getMessage())
    return
  end

  loadValidatedState(self, savedState)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function CrossSaveManager.new(round, timer, helpTray)
  local self = {}
  Mixin.mix(self, CrossSaveManager)

  self.round = round
  self.timer = timer
  self.helpTray = helpTray

  self.saveValidator = CrossSaveValidator.new()

  self.round:setSaveManager(self)
  self.timer:setSaveManager(self)

  return self
end
