--- CrossSaveValidator
-- @author Created by andrewcousins on May 04, 2022.

class("CrossSaveValidator")

require("validators.Validator")
require("utils.TableUtils")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- These match the strings in CrossSaveManager
local TIME_ELAPSED = "time_elapsed"
local GRID_DATA = "grid_data"
local CELL_CONTENT = "cell_content"

local CELL_CONTENT_KEYS = {"cell_content", "c"}
local CELL_REVEALED_KEYS = {"cell_revealed", "r"}
local CELL_INCORRECT_KEYS = {"x"}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateKey(self, savedState, key, allowEmpty)
  local allowEmpty = allowEmpty or false
  local message = "Key does not exist in saved state: " .. key
  self:expectHasKey(savedState, key, allowEmpty, message)
end

local function validateExactlyOneKey(self, savedState, keys)
  local noKeysFoundMessage = "No key exists in saved state: "
  local moreThanOneKeyMessage = "Only one key should exist in state: "
  local validKeyFound = false
  for _, key in ipairs(keys) do
    noKeysFoundMessage = noKeysFoundMessage .. key .. ", "
    if (savedState[key] ~= nil) then
      moreThanOneKeyMessage = moreThanOneKeyMessage .. key .. ", "
      self:expectFalse(validKeyFound, moreThanOneKeyMessage)
      validKeyFound = true
    end
  end
  self:expectTrue(validKeyFound, noKeysFoundMessage)
end

local function validateUpToOneKey(self, savedState, keys)
  local moreThanOneKeyMessage = "Only one key should exist in state: "
  local validKeyFound = false
  for _, key in ipairs(keys) do
    if (savedState[key] ~= nil) then
      moreThanOneKeyMessage = moreThanOneKeyMessage .. key .. ", "
      self:expectFalse(validKeyFound, moreThanOneKeyMessage)
      validKeyFound = true
    end
  end
end

local function validateTimeElapsed(self, savedState)
  local key = TIME_ELAPSED
  validateKey(self, savedState, key)
  self:expectGreaterOrEqual(savedState[key], 0, key)
end

local function validateCellState(self, cellState)
  if #TableUtils.getKeys(cellState) > 0 then
    validateExactlyOneKey(self, cellState, CELL_CONTENT_KEYS)
    validateExactlyOneKey(self, cellState, CELL_REVEALED_KEYS)
    validateUpToOneKey(self, cellState, CELL_INCORRECT_KEYS)
  end
end

local function validateGridData(self, savedState)
  local key = GRID_DATA
  validateKey(self, savedState, key)
  for _, column in ipairs(savedState[key]) do
    for _, cellState in ipairs(column) do
      validateCellState(self, cellState)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossSaveValidator:validateSave(savedState)
  self:runValidation(function()
    validateTimeElapsed(self, savedState)
    validateGridData(self, savedState)
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CrossSaveValidator.new()
  local self = Validator.new()
  Mixin.mix(self, CrossSaveValidator)

  return self
end
