--- CrossTimeDisplay
-- @author Created by andrewcousins on May 02, 2022.

class("CrossTimeDisplay")


local KERNING_PADDING_MULTIPLIER = 0.05
local COLON_PADDING_MULTIPLIER = 0.02

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function getDigitValues(time)
  local timeRemaining = math.floor(time)

  local hoursValue = math.floor(timeRemaining / 3600)
  timeRemaining = timeRemaining % 3600

  local minutesValue = math.floor(timeRemaining / 60)
  timeRemaining = timeRemaining % 60

  return {
    hours = hoursValue,
    minutes = minutesValue,
    seconds = timeRemaining }
end

local function setTextIfNotEqualToValue(label, value, showZeroes)
  local isLabelEqualToValue = (label:getText() == tostring(value))
  if (not isLabelEqualToValue) then
    label:setText(tostring(value))
  end
end

local function addColonLabel(self)
  local colonLabel = Label.new(":", self.font)
  colonLabel:setColor(unpack(self.color))

  local colonColumn = Column.new()
  colonColumn:insert(colonLabel)
  local visuallyFudgeColonY = 4
  colonColumn:insertSpacer(visuallyFudgeColonY)
  colonColumn:alignChildren()

  local colonPadding = self.font:getActualSize() * COLON_PADDING_MULTIPLIER
  self:insertSpacer(colonPadding)
  self:insert(colonColumn)
  self:insertSpacer(colonPadding)
  table.insert(self.colonLabels, colonLabel)
end

local function addLabel(self, labelTable)
  local label = Label.new("0", self.font)
  label:setColor(unpack(self.color))

  self:insert(label)
  table.insert(labelTable, label)

  for _, callback in ipairs(self.labelAddedCallbacks) do
    callback()
  end
end

local function updateLabels(self, labelTable, time, showAtLeastTwoDigits)
  local labelIndex = 1
  while (time > 0) do
    local digit = time % 10
    if (#labelTable < labelIndex) then
      if (labelIndex == 1) then
        addColonLabel(self)
      end
      addLabel(self, labelTable)
    end
    setTextIfNotEqualToValue(labelTable[labelIndex], digit)
    time = math.floor(time / 10)
    labelIndex = labelIndex + 1
  end

  local minimumDigitsToUpdate = #labelTable
  if (showAtLeastTwoDigits and (minimumDigitsToUpdate < 2)) then
    minimumDigitsToUpdate = 2
  end
  for labelIndex = labelIndex, minimumDigitsToUpdate do
    local digit = 0
    if (#labelTable < labelIndex) then
      addLabel(self, labelTable)
    end
    setTextIfNotEqualToValue(labelTable[labelIndex], digit)
  end
end

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

function CrossTimeDisplay:setDisplayedTime(time)
  local digitValues = getDigitValues(time)
  local showAtLeastTwoDigits = true
  updateLabels(self, self.secondsLabels, digitValues.seconds, showAtLeastTwoDigits)
  showAtLeastTwoDigits = (digitValues.hours > 0)
  updateLabels(self, self.minutesLabels, digitValues.minutes, showAtLeastTwoDigits)
  showAtLeastTwoDigits = false
  updateLabels(self, self.hoursLabels, digitValues.hours, showAtLeastTwoDigits)

  self:alignChildren()
end

function CrossTimeDisplay:setOpacity(opacity)
  for _, labelGroup in ipairs(self.labelGroups) do
    for _, label in ipairs(labelGroup) do
      label:setOpacity(opacity)
    end
  end
end

function CrossTimeDisplay:getFadeAnimation(duration, ease, opacity)
  local fadeAnimation = AHSpawn.new()
  for _, labelGroup in ipairs(self.labelGroups) do
    for _, label in ipairs(labelGroup) do
      fadeAnimation:addAction(AHOpacityTo.new(label, duration, ease, opacity))
    end
  end
  return fadeAnimation
end

function CrossTimeDisplay:addLabelAddedCallback(callback)
  table.insert(self.labelAddedCallbacks, callback)
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createLayout(self, font)
  addLabel(self, self.secondsLabels)
  addLabel(self, self.secondsLabels)
  addColonLabel(self)

  addLabel(self, self.minutesLabels)

  self:alignChildren()

  return self
end

-- Passing in a PrerenderedFont is the most efficient, however there appears to
-- be an inconsistent error when trying to use the same FontFace to create two
-- different PrerenderedFonts. A regular Font can be used in this case.
function CrossTimeDisplay.new(font, color)
  local spacingPadding = font:getActualSize() * KERNING_PADDING_MULTIPLIER
  local self = Row.new(spacingPadding, Alignment.CENTER, Row.GrowLeft)
  Mixin.mix(self, CrossTimeDisplay)

  self.secondsLabels = {}
  self.minutesLabels = {}
  self.hoursLabels = {}
  self.colonLabels = {}
  self.labelGroups = {self.secondsLabels, self.minutesLabels, self.hoursLabels, self.colonLabels}

  self.labelAddedCallbacks = {}

  self.font = font
  self.color = color or CrossColors.UI_COMPONENTS
  createLayout(self, font)

  return self
end