--- CrossTrayButton
-- @author Created by andrewcousins on May 26, 2023.

class("CrossTrayButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REVEAL_LETTER_TEXT = Localizer.get("REVEAL_LETTER")
local REVEAL_WORD_TEXT = Localizer.get("REVEAL_WORD")
local CHECK_FOR_ERRORS_TEXT = Localizer.get("CHECK_FOR_ERRORS")

local REVEAL_LETTER_SPRITE_NAME = "icon_hints_letter.png"
local REVEAL_WORD_SPRITE_NAME = "icon_hints_word.png"
local CHECK_FOR_ERRORS_SPRITE_NAME = "icon_hints_check.png"

local CORNER_RADIUS_PERCENT = 0.18
local BUTTON_SHADOW_OPACITY = 0.05

local ACTIVE_OPACITY = 1.0
local INACTIVE_OPACITY = 0.3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getOnPressDown(self)
  return function()
    if self.isActive then
      self.pressActionInterrupt:insert(function()
        self.buttonSurface:setColor(unpack(CrossColors.TRAY_BUTTON_OUTLINE))
        self.buttonSurface:setScaleY(0.98)
      end)
    end
  end
end

local function getOnPressUp(self)
  local duration = 0.167
  local ease = AHEaseBezier.newFlat({0.42, 0.00, 0.58, 1.00})
  local animation = AHSequence.new(
    AHScaleYTo.new(self.buttonSurface, duration, ease, 1.01),
    AHScaleYTo.new(self.buttonSurface, duration, ease, 1.00)
  )
  return function()
    if self.isActive then
      self.pressActionInterrupt:insert(AHSpawn.new(
        function()
          self.buttonSurface:setColor(unpack(CrossColors.TRAY_BUTTON_SURFACE))
        end,
        animation
      ))
      self.callback()
    end
  end
end

local function getOnPressCancel(self)
  return function()
    if self.isActive then
      self.pressActionInterrupt:insert(function()
        self.buttonSurface:setColor(unpack(CrossColors.TRAY_BUTTON_SURFACE))
        self.buttonSurface:setScaleY(1.0)
      end)
    end
  end
end

local function getSetIsActiveCallback(self, fullOpacityObjects, buttonShadow)
  return function(isActive)
    local duration = 0.1
    local ease = AHEaseLinear.new()
    local animation = AHSpawn.new()
    local opacity = ACTIVE_OPACITY
    if not isActive then
      opacity = INACTIVE_OPACITY
    end
    animation:addAction(AHOpacityTo.new(self, duration, ease, opacity))
    for _, object in ipairs(fullOpacityObjects) do
      animation:addAction(AHOpacityTo.new(object, duration, ease, opacity))
    end
    animation:addAction(AHOpacityTo.new(buttonShadow, duration, ease, opacity * BUTTON_SHADOW_OPACITY))

    if isActive then
      animation = AHSequence.new(animation, function() self.isActive = true end)
    else
      self.isActive = false
    end

    self.isActiveActionInterrupt:insert(animation)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, cornerRadius, spriteName, text, isSpriteAbove, spriteHeightPercentage)
  local width, height = self:getDims()
  local spacing = height * 0.06
  if isSpriteAbove then
    spacing = spacing / 2
  end
  local surfaceWidth = width - spacing
  local outlineHeight = height - spacing
  local shadowHeight = outlineHeight - spacing
  local surfaceHeight = shadowHeight - (spacing * 0.5)


  local buttonOutline = FilledRoundedRectangle.new(surfaceWidth, outlineHeight, cornerRadius, CrossColors.TRAY_BUTTON_OUTLINE)
  self:insertAtCenter(buttonOutline)
  buttonOutline:offsetY(spacing * 0.25)

  local buttonShadow = FilledRoundedRectangle.new(surfaceWidth, shadowHeight, cornerRadius, CrossColors.TRAY_BUTTON_SHADOW)
  -- Use AnchorPointBottom so scaling animation on downstate works as intended
  buttonShadow:setAnchorPointBottom()
  buttonShadow:setOpacity(BUTTON_SHADOW_OPACITY)
  self:insertAtBottom(buttonShadow)
  buttonShadow:offsetY(spacing * 1.5)

  local buttonSurface = FilledRoundedRectangle.new(surfaceWidth, surfaceHeight, cornerRadius, CrossColors.TRAY_BUTTON_SURFACE)
  -- Use AnchorPointBottom so scaling animation on downstate works as intended
  buttonSurface:setAnchorPointBottom()
  self:insertAtBottom(buttonSurface)
  buttonSurface:offsetY(spacing * 2)


  local sprite = nil
  if (spriteName ~= nil) then
    sprite = Sprite.newWithName(spriteName)
    sprite:setEffectiveDims(nil, height * spriteHeightPercentage)
    if isSpriteAbove then
      local spriteLocationFractionalY = 0.63
      buttonSurface:insertAtLocationFractional(sprite, 0.5, spriteLocationFractionalY)
    else
      local spriteLocationFractionalX = 0.89
      buttonSurface:insertAtLocationFractional(sprite, spriteLocationFractionalX, 0.5)
    end
  end

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 16)
  local label = Label.new(text, font)
  label:setColor(unpack(CrossColors.TRAY_BUTTON_TEXT))

  if isSpriteAbove then
    local labelLocationFractionalY = 0.24
    buttonSurface:insertAtLocationFractional(label, 0.5, labelLocationFractionalY)
  else
    local horizontalPadding = width * 0.05
    buttonSurface:insertAtLeft(label)
    label:offsetX(label:getWidthPart(0.5) + horizontalPadding)
  end

  self.buttonSurface = buttonSurface

  Button.makeButton(self, getOnPressDown(self), getOnPressUp(self), getOnPressCancel(self))

  self.setIsActive = getSetIsActiveCallback(self, {buttonOutline, label, sprite}, buttonShadow)
end

-- Constructor ------------------------

function CrossTrayButton.new(width, height, callback, spriteName, text, isSpriteAbove, spriteHeightPercentage)
  local isSpriteAbove = isSpriteAbove or false
  local spriteHeightPercentage = spriteHeightPercentage or 0.45
  local cornerRadius = height * CORNER_RADIUS_PERCENT
  if isSpriteAbove then
    cornerRadius = cornerRadius / 2
  end
  self = FilledRoundedRectangle.new(width, height, cornerRadius, CrossColors.TRAY_BUTTON_BACKGROUND)
  Mixin.mix(self, CrossTrayButton)

  self.callback = callback
  self.pressActionInterrupt = ActionInterrupt.new()
  self.isActiveActionInterrupt = ActionInterrupt.new()
  self.isActive = true

  createLayout(self, cornerRadius, spriteName, text, isSpriteAbove, spriteHeightPercentage)

  return self
end

function CrossTrayButton.newRevealLetter(width, height, callback)
  local spriteName = REVEAL_LETTER_SPRITE_NAME
  local text = REVEAL_LETTER_TEXT
  local isSpriteAbove = true
  local spriteHeightPercentage = 0.25

  return CrossTrayButton.new(width, height, callback, spriteName, text, isSpriteAbove, spriteHeightPercentage)
end

function CrossTrayButton.newRevealWord(width, height, callback)
  local spriteName = REVEAL_WORD_SPRITE_NAME
  local text = REVEAL_WORD_TEXT
  local isSpriteAbove = true
  local spriteHeightPercentage = 0.25

  return CrossTrayButton.new(width, height, callback, spriteName, text, isSpriteAbove, spriteHeightPercentage)
end

function CrossTrayButton.newCheckForErrors(width, height, callback)
  local spriteName = CHECK_FOR_ERRORS_SPRITE_NAME
  local text = CHECK_FOR_ERRORS_TEXT

  return CrossTrayButton.new(width, height, callback, spriteName, text)
end

function CrossTrayButton.newSquare(size, callback, spriteName, labelText)
  local isSpriteAbove = true
  return CrossTrayButton.new(size, size, callback, spriteName, labelText, isSpriteAbove)
end
