--- CrossUIButton
-- @author Created by andrewcousins on November 27, 2023.

class("CrossUIButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CrossUIButton.BUTTON_SIZE = 16 + FontScaling.getScaledSize(20)
if Screen.isTabletAspectRatio() then
  CrossUIButton.BUTTON_SIZE = CrossUIButton.BUTTON_SIZE - 5
end

local VIRTUAL_BUTTON_EXTENSION = CrossUIButton.BUTTON_SIZE * 0.75

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addButtonFunctionality(self, buttonPressedCallback)
  local duration = 0.05
  local ease = nil
  local onPressDown = function()
    if (not self.isDown) then
      local animation = AHColorTo.new(self.buttonBackground, duration, ease, CrossColors.UI_BUTTON_DOWNSTATE)
      self.actionQueue:insert(animation)
      self.isDown = true
    end
  end
  local onPressCancel = function()
    if self.isDown then
      local animation = AHColorTo.new(self.buttonBackground, duration, ease, CrossColors.UI_BUTTON)
      self.actionQueue:insert(animation)
      self.isDown = false
    end
  end
  local onPressUp = function()
    onPressCancel()
    buttonPressedCallback()
  end

  local virtualButtonWidth = self:getEffectiveWidth() + VIRTUAL_BUTTON_EXTENSION
  local virtualButtonHeight = self:getEffectiveHeight() + VIRTUAL_BUTTON_EXTENSION
  self.virtualButton = Button.addVirtualButton(self, virtualButtonWidth, virtualButtonHeight, onPressDown, onPressUp, onPressCancel)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossUIButton:getFadeInAnimation()
  local fadeDuration = CrossCommon.UI_FADE_IN_DURATION
  local fadeEase = AHEaseLinear.new()
  return AHSequence.new(
    function() self:setOpacity(0) end,
    AHOpacityTo.new(self, fadeDuration, fadeEase, 1)
  )
end

function CrossUIButton:getFadeOutAnimation()
  local fadeDuration = CrossCommon.UI_FADE_OUT_DURATION
  local fadeEase = AHEaseLinear.new()
  return AHOpacityTo.new(self, fadeDuration, fadeEase, 0)
end

function CrossUIButton:setButtonWidth(width, virtualButtonDirectionCapIsLeft, virtualButtonExtensionCap)
  self:setWidth(width)

  self.buttonOutline:setWidth(width)
  self.buttonOutline:setX(width/2)

  self.buttonBackground:setWidth(width)
  self.buttonBackground:setX(width/2)

  if ((virtualButtonDirectionCapIsLeft ~= nil) and (virtualButtonExtensionCap ~= nil)) then
    local standardButtonExtension = (VIRTUAL_BUTTON_EXTENSION / 2)
    local virtualButtonExtension = math.min(virtualButtonExtensionCap, standardButtonExtension)
    local virtualButtonWidth = width + standardButtonExtension + virtualButtonExtension
    self.virtualButton:setWidth(virtualButtonWidth)
    self.virtualButton:setX(width/2)
    if virtualButtonDirectionCapIsLeft then
      self.virtualButton:offsetX((standardButtonExtension - virtualButtonExtension) / 2)
    else
      self.virtualButton:offsetX((virtualButtonExtension - standardButtonExtension) / 2)
    end
  else
    self.virtualButton:setWidth(width + VIRTUAL_BUTTON_EXTENSION)
    self.virtualButton:setX(width/2)
  end
end

-- This should be called after any children added.
function CrossUIButton:linkOpacityToChildren()
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local cornerRadius = CrossUIButton.BUTTON_SIZE * 0.15
  local buttonBackground = FilledRoundedRectangle.new(CrossUIButton.BUTTON_SIZE, CrossUIButton.BUTTON_SIZE, cornerRadius, CrossColors.UI_BUTTON)

  local lineWidth = 1
  local buttonOutline = HollowRoundedRectangle.new(CrossUIButton.BUTTON_SIZE, CrossUIButton.BUTTON_SIZE, cornerRadius, CrossColors.UI_BUTTON_BORDER, lineWidth)

  self:insertAtCenter(buttonBackground)
  self:insertAtCenter(buttonOutline)

  self.buttonOutline = buttonOutline
  self.buttonBackground = buttonBackground
end

-- Constructor ------------------------

function CrossUIButton.new(buttonPressedCallback)
  local self = Group.new(CrossUIButton.BUTTON_SIZE, CrossUIButton.BUTTON_SIZE)
  Mixin.mix(self, CrossUIButton)

  createLayout(self)

  self.actionQueue = ActionQueue.new()
  self.isDown = false

  addButtonFunctionality(self, buttonPressedCallback)

  return self
end
