--- CrossValidator
-- @author Created by andrewcousins on March 03, 2022.
-- ConceptValidator for Cross.

class("CrossValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'dimensions',
  'difficulty',
  'puzzle',
  'clues',
  'solution'
}

local REQUIRED_DIMENSION_KEYS = {
  'width',
  'height'
}

local REQUIRED_CLUES_KEYS = {
  'across',
  'down'
}

local VALID_SOLUTION_SYMBOLS = TableUtils.createSetFromString('ABCDEFGHIJKLMNOPQRSTUVWXYZ#')

local VALID_DIFFICULTY = {
  "Basic",
  "Easy",
  "Medium",
  "Hard",
  "Advanced"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateDifficulty(self, concept)
  self:expectIsContained(concept.difficulty, VALID_DIFFICULTY, "difficulty")
end

local function validateRequiredDimensionKeys(self, concept)
  self:validateRequiredKeys(concept.dimensions, REQUIRED_DIMENSION_KEYS)
end

local function validateRequiredCluesKeys(self, concept)
  self:validateRequiredKeys(concept.clues, REQUIRED_CLUES_KEYS)
end

local function validatePuzzleDataSize(self, dataName, puzzleData, width, height)
  self:expectEqual(#puzzleData, height, string.format("The %s's height does not match the set value.", dataName))
  for rowIndex, row in ipairs(puzzleData) do
    self:expectEqual(#row, width, string.format("The %s's width does not match the set value (row %d).", dataName, rowIndex))
  end
end

local function validatePuzzle(self, concept)
  local width = concept.dimensions.width
  local height = concept.dimensions.height
  local puzzle = concept.puzzle

  validatePuzzleDataSize(self, "puzzle", puzzle, width, height)

  for _, row in ipairs(puzzle) do
    for _, cell in ipairs(row) do
      if (cell ~= "#") then
        self:expectType(cell, 'number', "puzzle", "Characters in Puzzle must be numbers or #.")
      end
    end
  end
end

local function validateSolution(self, concept)
  local width = concept.dimensions.width
  local height = concept.dimensions.height
  local solution = concept.solution

  validatePuzzleDataSize(self, "solution", solution, width, height)

  for _, row in ipairs(solution) do
    for _, cell in ipairs(row) do
      self:expectStringCharactersBelongToSet(cell, VALID_SOLUTION_SYMBOLS, "Solution")
    end
  end
end

local function validateClueIndexInPuzzle(self, concept, clueIndex, fieldName)
  for _, row in ipairs(concept.puzzle) do
    for _, cell in ipairs(row) do
      if (cell == clueIndex) then
        return
      end
    end
  end
  self:expectFailed(string.format("%s has a clue number not found in the puzzle: %d", fieldName, clueIndex))
end

local function validateClues(self, concept)
  local clues = concept.clues
  local maxClueNumber = concept.dimensions.width * concept.dimensions.height
  for clueCategory, clueList in pairs(clues) do
    for clueListIndex, clue in ipairs(clueList) do
      clueIndex, clueText = unpack(clue)
      self:expectType(clueIndex, "number", "clue index", "Clue index is not a number. Is this clue missing an index?")
      self:expectType(clueText, "string", "clue text", "Clue text is not a string. Clues must have a single index followed by a single string.")
      local fieldName = string.format("%s clue number %d", clueCategory, clueListIndex)
      self:expectNumberInRange(clueIndex, 1, maxClueNumber, fieldName)
      validateClueIndexInPuzzle(self, concept, clueIndex, fieldName)
      self:expectNoLeadingOrTrailingWhitespace(clueText, fieldName)
      self:expectStringIsPrintable(clueText, fieldName)
    end
  end
end

local function validateWallLocations(self, concept)
  for rowIndex=1, #concept.puzzle do
    for columnIndex=1, #concept.puzzle[rowIndex] do
      local puzzleCell = concept.puzzle[rowIndex][columnIndex]
      local solutionCell = concept.solution[rowIndex][columnIndex]
      if ((puzzleCell == "#") or (solutionCell == "#")) then
        self:expectEqual(puzzleCell, solutionCell,
          string.format("A wall was found in either the solution or puzzle \
            that does not exist in the other. Row: %d, Column: %d",
            rowIndex, columnIndex))
      end
    end
  end
end

local function validatePuzzleClueLocations(self, concept)
  local clueIndexesInPuzzle = {}
  local maxClueIndex = 0
  for _, row in ipairs(concept.puzzle) do
    for _, clueIndex in ipairs(row) do
      if ((clueIndex ~= "#") and (clueIndex ~= 0)) then
        table.insert(clueIndexesInPuzzle, clueIndex)
        if (clueIndex > maxClueIndex) then
          maxClueIndex = clueIndex
        end
      end
    end
  end

  self:expectNoDuplicates(clueIndexesInPuzzle, "Puzzle", "Duplicate clue index(es) found in puzzle.")
  table.sort(clueIndexesInPuzzle)

  local clueIndexesInClues = {}
  for _, clueList in pairs(concept.clues) do
    for _, clue in ipairs(clueList) do
      table.insert(clueIndexesInClues, clue[1])
    end
  end
  clueIndexesInClues = TableUtils.getUniquesArray(clueIndexesInClues)
  table.sort(clueIndexesInClues)

  self:expectEqual(#clueIndexesInPuzzle, maxClueIndex, "There appears to be an incorrect number of clues in the puzzle.")
  self:expectEqual(#clueIndexesInClues, maxClueIndex, "There appears to be an incorrect number of clues in the puzzle.")
  for i=1, maxClueIndex do
    self:expectEqual(clueIndexesInPuzzle[i], i, "Clue indexes in puzzle is missing " .. tostring(i))
    self:expectEqual(clueIndexesInClues[i], i, "Clue indexes in clues is missing " .. tostring(i))
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CrossValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  validateRequiredDimensionKeys(self, concept)
  validateRequiredCluesKeys(self, concept)

  validatePuzzle(self, concept)
  validateSolution(self, concept)
  validateClues(self, concept)
  validateWallLocations(self, concept)
  validatePuzzleClueLocations(self, concept)

  validateDifficulty(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CrossValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, CrossValidator)

  return self
end
