--- DepthBubble
-- @author Created by Aaron Barrett on January 25, 2014.

class("DepthBubble")
require("DepthBubbleInsidesView")
require("DepthInsidesIncorrectView")
-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local Z_SQUIGGLEBURST = 10
local Z_CIRCLEBURST = 20
local Z_BACK = 30
local Z_FACE = 40
local Z_STARBURST = 50
local Z_EVIL_CIRCLE = 60

local CORRECT_COLOR = {0.76, 1.00, 0.99}
local INCORRECT_COLOR = {0.99, 0.18, 0.22}

local EVIL_CIRCLE_OPACITY = 0.2
local EVIL_CIRCLE_STEPS = 32



-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function destroySelf(self)
  if (self.squishAction) then
    self.squishAction:stop()
    self.squishAction = nil
  end

  self.velocityRegulator = nil
  self.face = nil
  self.back = nil

  self:removeSelf()
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function DepthBubble:getIdentifier()
  return self.identifier
end

function DepthBubble:getIsGood()
  return self.isGood
end

function DepthBubble:getRadius()
  return self.radius
end

function DepthBubble:getBody()
  return self.body
end

function DepthBubble:getPosition()
  local body = self.body
  return body:getPosition()
end

function DepthBubble:getFractionAsString()
  local fraction = self.face
  local numerator = fraction:getNumerator()
  local denominator = fraction:getDenominator()

  local fractionAsString = numerator .. "/" .. denominator
  return fractionAsString
end

function DepthBubble:setBody(body)
  self.body = body

  if (body ~= nil) then
    local velocityRegulator = self.velocityRegulator
    velocityRegulator:setNodeLink(body)
  end
end

function DepthBubble:startAnimating()
  local back = self.back

  local backScaleFactor = self.backScaleFactor

  local squishAmount = 0.025
  local squishedScaleA = backScaleFactor + squishAmount
  local squishedScaleB = backScaleFactor - squishAmount

  local squishEase = AHEaseSimpleInOut.new(2.3)

  local squishAction = AHLoop.new(
                                  AHSequence.new(
                                                 AHScaleTo.new(back, 1.2, squishEase, squishedScaleA, squishedScaleB),
                                                 AHScaleTo.new(back, 1.2, squishEase, squishedScaleB, squishedScaleA)
                                                 )
                                  )
  self.squishAction = squishAction
  return squishAction:start()
end

function DepthBubble:preparePopCorrect()
  local radius = self.radius

  local starburst = DepthStarburst.new()
  starburst:setLoc(radius, radius)
  self:insert(starburst, Z_STARBURST)
  self.starburst = starburst

  local circleburst = DepthCircleBurst.new()
  circleburst:setLoc(radius, radius)
  self:insert(circleburst, Z_CIRCLEBURST)
  self.circleburst = circleburst

  local squiggleburst = DepthSquiggleBurst.new()
  squiggleburst:setLoc(radius, radius)
  self:insert(squiggleburst, Z_SQUIGGLEBURST)
  self.squiggleburst = squiggleburst
end

function DepthBubble:createPopCorrectAction()
  local back = self.back
  local face = self.face

  local backScaleAction = AHScaleTo.new(back, 0.2, AHEaseBackIn.new(1.4), 1.1)
  local backFadeAction = AHOpacityTo.new(back, 0.2, nil, 0.0)
  local backColorAction = AHColorTo.new(back, 0.05, nil, CORRECT_COLOR)

  local faceColorAction = AHColorTo.new(face, 0.05, nil, CORRECT_COLOR)
  local faceFadeAction = AHSequence.new(
                                        AHOpacityTo.new(face, 1.7, nil, 0.0),
                                        AHCallLogic.new(function()
                                                            destroySelf(self)
                                                          end)
                                        )

  local showParticles = function()
    -- WARNING: if you want to modify this code, be sure to do a lot of device testing.
    -- This block of code is carefully structured to sidestep lag issues on devices.
    coroutine.yield()
    local squiggleburst = self.squiggleburst
    local squiggleburstAction = squiggleburst:createBurstAction()
    squiggleburstAction:start()

    coroutine.yield()
    local circleburst = self.circleburst
    local circleburstAction = circleburst:createBurstAction()
    circleburstAction:start()

    coroutine.yield()
    local starburst = self.starburst
    local starburstAction = starburst:createBurstAction()
    coroutine.yield()
    starburstAction:start()
  end


  local popAction = AHSpawn.new(
                                backScaleAction,
                                backFadeAction,
                                backColorAction,
                                faceColorAction,
                                faceFadeAction,
                                showParticles
                                )

  return popAction
end

function DepthBubble:preparePopIncorrect()
  local radius = self.radius

  local evilCircle = FilledCircle.new(radius, INCORRECT_COLOR, EVIL_CIRCLE_STEPS)
  evilCircle:setOpacity(EVIL_CIRCLE_OPACITY)
  evilCircle:setLoc(radius, radius)
  self:insert(evilCircle, Z_EVIL_CIRCLE)
  self.evilCircle = evilCircle
end

function DepthBubble:createPopIncorrectAction()
  local w, h = self:getDims()
  local evilCircle = self.evilCircle
  local back = self.back
  local face = self.face
  local trail = DepthInsidesIncorrectView.new(self.concept)
  self:insert(trail)
  trail:die()
  trail:setLoc(w/2, h)
  Screen.rootGroup:takeChild(trail)

  local radius = self.radius
  local finalEvilCircleScale = DepthDot.RADIUS / radius

  local evilCircleScaleAction = AHSequence.new(
                                               AHScaleTo.new(evilCircle, 0.0375, nil, 1.4),
                                               AHScaleTo.new(evilCircle, 0.0375, nil, 1.3),
                                               AHScaleTo.new(evilCircle, 0.0375, nil, 1.4),
                                               AHScaleTo.new(evilCircle, 0.0375, nil, 1.3),
                                               AHScaleTo.new(evilCircle, 0.15, AHEaseSimpleIn.new(2.0), finalEvilCircleScale),
                                               AHCallLogic.new(function()
                                                                   destroySelf(self)
                                                                 end)
                                               )

  local backScaleFactor = self.backScaleFactor
  local backScaleLow = backScaleFactor * 0.95
  local backScaleHigh = backScaleFactor * 1.05
  local finalBackScale = DepthDot.RADIUS / self.backWidth

  local backScaleAction = AHSequence.new(
                                         AHScaleTo.new(back, 0.025, nil, backScaleFactor),
                                         AHScaleTo.new(back, 0.05, nil, backScaleHigh),
                                         AHScaleTo.new(back, 0.05, nil, backScaleLow),
                                         AHScaleTo.new(back, 0.05, nil, backScaleHigh),
                                         AHScaleTo.new(back, 0.05, nil, backScaleLow),
                                         AHScaleTo.new(back, 0.075, AHEaseSimpleIn.new(2.0), finalBackScale)
                                         )

  local backColorAction = AHColorTo.new(back, 0.05, nil, INCORRECT_COLOR)

  local faceScaleAction = AHSequence.new(
                                         AHScaleTo.new(face, 0.025, nil, 1.0),
                                         AHScaleTo.new(face, 0.05, nil, 1.05),
                                         AHScaleTo.new(face, 0.05, nil, 0.95),
                                         AHScaleTo.new(face, 0.05, nil, 1.05),
                                         AHScaleTo.new(face, 0.05, nil, 0.95),
                                         AHScaleTo.new(face, 0.075, AHEaseSimpleIn.new(2.0), finalBackScale)
                                         )

  local faceColorAction = AHColorTo.new(face, 0.05, nil, INCORRECT_COLOR)

  local popAction = AHSpawn.new(
                                evilCircleScaleAction,
                                backScaleAction,
                                backColorAction,
                                faceScaleAction,
                                faceColorAction
                                )

  return popAction
end

function DepthBubble:pop(isCorrect)
  local body = self.body

  local velocityRegulator = self.velocityRegulator
  velocityRegulator:clearNodeLink(body)

  local popAction = nil
  if (isCorrect) then
    self:preparePopCorrect()
    popAction = self:createPopCorrectAction()
  else
    self:preparePopIncorrect()
    popAction = self:createPopIncorrectAction()
  end

  return popAction:start()
end


function DepthBubble:hide(time)
  time = time or 0.4

  local hideFace = AHOpacityTo.new(self.face, time, nil, 0)
  local hideBack = AHOpacityTo.new(self.back, time, nil, 0)
  local hide = AHSpawn.new(hideFace, hideBack)
  return hide:start()
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function DepthBubble.new(concept, getBubbleSpeedFunction)
  local radius = DepthConstants.BUBBLE_BASE_RADIUS + math.random() * DepthConstants.BUBBLE_MAX_RANDOM_RADIUS
  local diameter = radius * 2.0

  local self = Group.new(diameter, diameter)
  Mixin.mix(self, DepthBubble)

  self.radius = radius
  self.identifier = concept.identifier
  self.isGood = concept.isCorrect
  self.concept = concept

  local back = Sprite.newWithName("bubble.png")
  local backWidth, backHeight = back:getDims()
  local backScaleFactor = diameter / backWidth
  back:setScl(backScaleFactor)
  self.backScaleFactor = backScaleFactor
  self.backWidth = backWidth
  back:setLoc(radius, radius)
  self:insert(back, Z_BACK)
  self.back = back

  local velocityRegulator = MOAIScriptNode.new()
  local function onPhysicsUpdate()
    local body = self.body
    local bodyVelocityX, bodyVelocityY = body:getLinearVelocity()
    local bubbleSpeed = getBubbleSpeedFunction()
    body:setLinearVelocity(bodyVelocityX, bubbleSpeed)
  end
  velocityRegulator:setCallback(onPhysicsUpdate)
  self.velocityRegulator = velocityRegulator

  local face = DepthBubbleInsidesView.new(concept, radius)
  face:setLoc(radius, radius)
  self:insert(face, Z_FACE)
  self.face = face

  return self
end
