require("FractionGenerator")
require("DepthContentDifficultyChooser")

class("DepthConceptsGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUBBLE_TYPES =
{
  CORRECT_BUBBLE = 0,
  FRACTION_REPEAT = 1,
  TOTAL_REPEAT = 2,
  RANDOM_ERROR = 3,
}

local NAMES_BY_TYPE = {}
NAMES_BY_TYPE[BUBBLE_TYPES.CORRECT_BUBBLE] = "Correct"
NAMES_BY_TYPE[BUBBLE_TYPES.FRACTION_REPEAT] = "Fract Rep"
NAMES_BY_TYPE[BUBBLE_TYPES.TOTAL_REPEAT] = "Total Rep"
NAMES_BY_TYPE[BUBBLE_TYPES.RANDOM_ERROR] = "Random"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBubbleInfoFromConcept(concept, bubbleType, excludedDistractors)
  local result = {}
  result.total = concept.total
  result.fraction = concept.fraction
  result.name = NAMES_BY_TYPE[bubbleType]
  if(bubbleType == BUBBLE_TYPES.CORRECT_BUBBLE) then
    result.isCorrect = true
    result.difficulty = concept.difficulty
    -- Fully automatic game:
    -- local fractionInfo = FractionGenerator.generatePercentageFraction(concept.percent)
    -- result.fraction = fractionInfo.numerator
    -- result.total = fractionInfo.denominator
    -- result.difficulty = fractionInfo.difficulty
    -- result.modifiers = fractionInfo.modifiers
    -- result.isCorrect = true
    -- TODO: swap for the above behavior once we confirm that fully automatic content is a-okay
  elseif(bubbleType == BUBBLE_TYPES.TOTAL_REPEAT) then
    local errorIndex = math.random(#concept.fraction_errors)
    result.fraction = concept.fraction_errors[errorIndex]
    table.remove(concept.fraction_errors, errorIndex)
    result.isCorrect = false
    result.difficulty = concept.difficulty
  elseif(bubbleType == BUBBLE_TYPES.FRACTION_REPEAT) then
    local errorIndex = math.random(#concept.total_errors)
    result.total = concept.total_errors[errorIndex]
    table.remove(concept.total_errors, errorIndex)
    result.isCorrect = false
    result.difficulty = concept.difficulty
  elseif(bubbleType == BUBBLE_TYPES.RANDOM_ERROR) then
    local fractionInfo = FractionGenerator.generateRandomPercentageFraction(concept.percent, excludedDistractors)
    result.fraction = fractionInfo.numerator
    result.total = fractionInfo.denominator
    result.difficulty = fractionInfo.difficulty
    result.modifiers = fractionInfo.modifiers
    result.isCorrect = false

    excludedDistractors[#excludedDistractors + 1] = fractionInfo
  end
  return result
end

local function chooseBubbleTypeWithTypeChances(concept, fractionRepeatChance, totalRepeatChance, randomErrorChance)
  -- Yes, it's supposed to be backwards, cause when we're repeating the fraction we need an error of the total
  local canRepeatFraction = concept.total_errors and #concept.total_errors > 0
  local canRepeatTotal = concept.fraction_errors and #concept.fraction_errors > 0
  local canDoRandomError = true
  local canDoCorrect = (fractionRepeatChance + totalRepeatChance + randomErrorChance) < 1

  if(not canRepeatFraction) then
    fractionRepeatChance = 0
  end
  if(not canRepeatTotal) then
    totalRepeatChance = 0
  end

  local typeRoll = math.random()

  local fractionRoll = fractionRepeatChance
  local totalRoll = fractionRepeatChance + totalRepeatChance
  local randomRoll = fractionRepeatChance + totalRepeatChance + randomErrorChance


  if(typeRoll < fractionRoll and canRepeatFraction) then
    return BUBBLE_TYPES.FRACTION_REPEAT
  elseif(typeRoll < totalRoll and canRepeatTotal) then
    return BUBBLE_TYPES.TOTAL_REPEAT
  elseif((typeRoll < randomRoll and canDoRandomError) or not canDoCorrect) then
    return BUBBLE_TYPES.RANDOM_ERROR
  else
    return BUBBLE_TYPES.CORRECT_BUBBLE
  end
end

local function getErrorChances(correctInARow, incorrectInARow)
  local difficulty = GameSession.getDifficulty()
  local fractionRepeatChance =  0.2 - 0.1 * difficulty -- go from 0.2 to 0.1 linearly
  local totalRepeatChance =     0.2 - 0.1 * difficulty -- go from 0.2 to 0.1 linearly
  local randomErrorChance =     0.1 + 0.2 * difficulty -- go from 0.1 to 0.3 linearly
  if(correctInARow >= DepthConstants.MAX_CORRECT_BUBBLES_IN_A_ROW) then
    fractionRepeatChance = 0.33
    totalRepeatChance = 0.33
    randomErrorChance = 0.34
  end
  return fractionRepeatChance, totalRepeatChance, randomErrorChance
end

local function chooseBubbleType(concept, correctInARow, incorrectInARow)
  if(incorrectInARow >= DepthConstants.MAX_WRONG_BUBBLES_IN_A_ROW) then
    return BUBBLE_TYPES.CORRECT_BUBBLE
  end
  -- IDEA FOR BETTER FEEL: the repeats aren't actually handled together with a correct answer
  -- Maybe force there to be a correct answer for a streak of repeats?
  local fractionRepeatChance, totalRepeatChance, randomErrorChance = getErrorChances(correctInARow, incorrectInARow)
  return chooseBubbleTypeWithTypeChances(concept, fractionRepeatChance, totalRepeatChance, randomErrorChance)
end

local function adjustBubbleType(bubbleType, bubbleIndex)
  if(bubbleIndex < 2 and bubbleType == BUBBLE_TYPES.RANDOM_ERROR) then
    return BUBBLE_TYPES.CORRECT_BUBBLE
  end
  return bubbleType
end

local function parseSourceConcept(sourceConcept)
  local result = TableUtils.deepCopy(sourceConcept)
  result.fraction_errors = TableUtils.csvToArray(result.fraction_errors)
  result.total_errors = TableUtils.csvToArray(result.total_errors)
  return result
end

local function chooseBaseConcept(self, conceptIndex, percentage)
  local difficultyNameChooser = DepthContentDifficultyChooser.new()
  self.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)
  self.conceptGenerator:setFilters({percent_concept = percentage})
  return self.conceptGenerator:next()
end

local function selectBubbleInfoSequence(self, percentage)
  local bubblesToChoose = 30
  local sourceConceptsToChoose = 30
  local result = {}

  local baseConcepts = {}
  for i = 1, sourceConceptsToChoose do
    baseConcepts[#baseConcepts + 1] = chooseBaseConcept(self, i, percentage)
  end

  local baseConceptIndex = 1
  local currentSourceConcept = baseConcepts[baseConceptIndex]
  -- Copy the concept so we can know which error values have already been used
  local currentConcept = parseSourceConcept(currentSourceConcept)

  local correctInARow = 0
  local incorrectInARow = 0
  local excludedDistractors = {}

  while(bubblesToChoose > 0) do
    local bubbleType = chooseBubbleType(currentConcept, correctInARow, incorrectInARow)
    bubbleType = adjustBubbleType(bubbleType, #result + 1)

    local bubbleInfo = getBubbleInfoFromConcept(currentConcept, bubbleType, excludedDistractors)
    result[#result + 1] = bubbleInfo
    bubbleInfo.sourceConcept = currentSourceConcept
    bubbleInfo.identifier = currentSourceConcept.identifier

    if(bubbleType == BUBBLE_TYPES.CORRECT_BUBBLE) then
      correctInARow = correctInARow + 1
      incorrectInARow = 0
    else
      correctInARow = 0
      incorrectInARow = incorrectInARow + 1
    end

    bubblesToChoose = bubblesToChoose - 1
    if(bubbleType == BUBBLE_TYPES.CORRECT_BUBBLE or bubbleType == BUBBLE_TYPES.RANDOM_ERROR) then
      baseConceptIndex = baseConceptIndex % #baseConcepts + 1
      currentSourceConcept = baseConcepts[baseConceptIndex]
      currentConcept = parseSourceConcept(currentSourceConcept)
    end
  end
  return result
end

local function getRound(self, percentage)
  local result = {}
  local percentage = percentage or DepthConstants.CORRECT_ANSWER_PERCENTAGE
  local problems = selectBubbleInfoSequence(self, percentage)
  result.problems = problems
  result.task = percentage.."%"
  return result
end

local function increaseConceptNumber(self)
  self.conceptNumber = self.conceptNumber % #self.round.problems + 1
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function DepthConceptsGenerator:chooseNextConcept()
  local concept = self.round.problems[self.conceptNumber]
  increaseConceptNumber(self)
  return concept
end

function DepthConceptsGenerator:getTask()
  return self.round.task
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function DepthConceptsGenerator.new(percentage)
  local self = {}
  Mixin.mix(self, DepthConceptsGenerator)
  self.conceptGenerator = GameSession.conceptGenerator
  self.round = getRound(self, percentage)
  self.conceptNumber = 1
  return self
end
