--- DepthGame
-- @author Created by Aaron Barrett on January 25, 2014.

class("DepthGame")
require("DepthConfiguration")
require("DepthConstants")

require("DepthPrize")
require("DepthDot")
require("DepthBubble")
require("DepthPhysics")
require("DepthBubbleGenerator")

require("DepthLevel")
require("DepthLevelLedge")
require("DepthLevelWhale")
require("DepthLevelJellyfish")
require("DepthSchool")
require("DepthSchoolFish")
require("DepthScroller")
require("DepthBackground")

require("DepthMeter")

require("DepthStarburst")
require("DepthCircleBurst")
require("DepthSquiggleBurst")
require("DepthRedAlert")

require("DepthFishOutroDance")

require("DepthStreak")
require("DepthStreakScroller")

require("StandardUIBuilder")

require("DepthConceptsGenerator")

require("DepthIntro")
require("bonuses.ResponseSpeedTracker")

-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_STREAK_SCROLLER = 20
local Z_SCROLLER = 30
local Z_DOT = 40
local Z_BUBBLE = 50
local Z_PRIZE = 60
local Z_RED_ALERT = 70

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

-- Colors -----------------------------

local UI_COMPONENTS_COLOR = {1.0, 1.0, 1.0, 0.8}


local function configureGameDifficulty(self)
  local passScore = DepthConfiguration.getPassScore()
  local rankTwoScore = DepthConfiguration.getRankTwoScore()
  local rankThreeScore = DepthConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(DepthConfiguration.getNumberOfLives())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)

  GameSession.contentTracker:setSquashData(true)
end


-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function updateScoreLabel(self)
  local scoreLabel = self.scoreLabel
  scoreLabel:setScoreInstant(GameSession.getScore())
end

local function processCorrectResponse(self, bubble, y)
  local identifier = bubble:getIdentifier()
  GameSession.contentTracker:notifyCorrectAnswer(identifier)
  GameSession.notifyCorrectAnswer(identifier)
  GameSession.incrementCorrectAnswerCount()

  local bubbleRadius = bubble:getHeightPart(0.5)
  -- Wait until the bubble is on screen before reducing the speed bonus
  local responseValue = math.max(0, y - bubbleRadius)

  self.responseSpeedTracker:notifyCorrectResponseWithValue(responseValue)
  updateScoreLabel(self)
end

local function startScoreTrackingCoroutine(self)
  local function trackScore()
    while not self.gameOver do
      GameSession.increaseScore(DepthConfiguration.POINTS_PER_PASSING_SECOND)
      updateScoreLabel(self)
      waitFor(1.0)
    end
  end
  local scoreTrackingCoroutine = runInCoroutine(trackScore)
  return scoreTrackingCoroutine
end

local function handleMissedBubble(self, bubble, x, y)
  GameSession.notifyIncorrectAnswer(bubble:getIdentifier())
  GameSession.incrementIncorrectAnswerCount()

  if(self.gameOver) then return end
  self.isTouchEnabled = false
  self:setAllTheThingsPaused(true)
  self.redAlert:flash()
  Sound.playSoundEffect("depth_incorrect")

  self.livesDisplay:decrement()
  local areLivesRemaining = (self.livesDisplay:getLives() > 0)
  if (areLivesRemaining) then
    local resumerAction = AHSequence.new(
                                         AHDelay.new(0.45),
                                         function()
                                           self:setAllTheThingsPaused(false)
                                           self.isTouchEnabled = true
                                           self:createDotAt(x, y)
                                         end
                                         )
    resumerAction:start()
  else
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  end
end

local function onBubbleTapped(self, bubble)
  if(self.gameOver) then return end

  if (self.isTouchEnabled) then
    local physics = self.physics

    TouchManager.removeTouchHandler(bubble)
    self.bubbles[bubble] = nil

    local x, y = bubble:getPosition()

    local shouldShowPrize = (bubble:getIsGood())

    if (shouldShowPrize) then
      runInCoroutine(function()
        coroutine.yield()
        self:createPrizeAt(x, y)
        Sound.playSoundEffect("depth_correct")
        processCorrectResponse(self, bubble, y)
      end)
    else
      local identifier = bubble:getIdentifier()
      local fractionAsString = bubble:getFractionAsString()
      GameSession.contentTracker:notifyIncorrectAnswer(identifier, nil, fractionAsString)
      GameSession.incrementIncorrectAnswerCount()

      handleMissedBubble(self, bubble, x, y)
    end

    bubble:pop(shouldShowPrize)

    runInCoroutine(function()
      coroutine.yield()
      coroutine.yield()
      physics:removeBubble(bubble)
    end)
  end
end

local function showIntroInstructions(self)
  self.scroller:pause(true)
  self.intro:setContinueHandler(function()
    self.scroller:pause(false)
  end)
  return self.intro:getActionLock()
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function DepthGame:incrementBubbleSpeed()
  self.bubbleSpeed = self.bubbleSpeed + DepthConstants.BUBBLE_SPEED_INCREMENT
  if (self.bubbleSpeed > DepthConstants.MAX_BUBBLE_SPEED) then
    self.bubbleSpeed = DepthConstants.MAX_BUBBLE_SPEED
  end
end

function DepthGame:setAllTheThingsPaused(isAllTheThingsPaused)
  self.scroller:pause(isAllTheThingsPaused)
  self.depthMeter:pause(isAllTheThingsPaused)
  self.physics:pause(isAllTheThingsPaused)
  self.bubbleGenerator:pause(isAllTheThingsPaused)
  self.background:pause(isAllTheThingsPaused)
end

function DepthGame:createDotAt(x, y)
  local physics = self.physics

  local dot = DepthDot.new()
  dot:setLoc(-Screen.width, -Screen.height)
  self:insert(dot, Z_DOT)
  physics:addDot(dot, x, y)
  dot:setLoc(0.0, 0.0)
end

function DepthGame:createPrizeAt(x, y)
  local depthMeter = self.depthMeter
  local progressProportion = depthMeter:getProgressProportion()

  local prize = DepthPrize.new(progressProportion)
  prize:setLoc(x, y)
  prize:setScl(0.0)
  self:insert(prize, Z_PRIZE)

  local shouldSwimLeft = (x > CENTER_X)
  prize:swimOut(shouldSwimLeft)

end

function DepthGame:sendBubble()
  if(self.gameOver) then return end
  local concept = self.conceptGenerator:chooseNextConcept()
  local function getBubbleSpeed()
    return self.bubbleSpeed
  end

  local bubble = DepthBubble.new(concept, getBubbleSpeed)
  self.bubbles[bubble] = true
  -- Move the bubble off-screen before adding it to the scene to prevent
  -- it from appearing at (0, 0) for one frame because of MOAI's node update bug
  bubble:setLoc(-Screen.width, -Screen.height)

  self:insert(bubble, Z_BUBBLE)
  self.physics:addBubble(bubble)

  -- Return the bubble to (0, 0) so it can mirror the physics body's "transform"
  bubble:setLoc(0.0, 0.0)

  local function bubbleTouchHandler(event)
    if(self.gameOver) then return end

    if event.phase == TouchManager.TOUCH_BEGAN then
      onBubbleTapped(self, bubble)
    end
  end
  TouchManager.addTouchHandler(bubble, bubbleTouchHandler)

  bubble:startAnimating()
end

function DepthGame:gameFlow()
  wait(showIntroInstructions(self))
  self.uiOverlay:fadeIn()

  self.background:startTinting()
  self.physics:start()
  self.bubbleGenerator:startGenerating()
  self.depthMeter:start()
  self.isTouchEnabled = true

  startScoreTrackingCoroutine(self)
end

function DepthGame:isUIAutoFadeInEnabled()
  return false
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------
local function createGameElements(self)
  local conceptGenerator = DepthConceptsGenerator.new()
  local taskString = conceptGenerator:getTask()
  local background = DepthBackground.new()
  local intro = DepthIntro.new(taskString)

  background:setLoc(CENTER_X, CENTER_Y)
  self:insert(background, Z_BACKGROUND)
  self:insert(intro, math.huge)

  self.background = background
  self.intro = intro
  self.conceptGenerator = conceptGenerator
  self.bubbles = {}
end

local function destroyAllBubbles(self)
  for bubble, _ in pairs(self.bubbles) do
    bubble:hide()
  end
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local depthMeterDuration = DepthConstants.GAME_DURATION

  local function finishGame()
    wait(self:showDefaultGameWinOutro())
    GameSession.notifyGameWin()
  end
  local function onFinishCallback()
    self.gameOver = true
    destroyAllBubbles(self)
    self.scroller:pause(true)

    ScoringUtils.applyLivesBonus(self.livesDisplay:getLives(), DepthConfiguration.getExtraLivesBonus(), self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    local outro = DepthFishOutroDance.new()
    Screen.rootGroup:insert(outro)
    outro:start(finishGame)
  end

  local depthMeter = DepthMeter.new(depthMeterDuration, onFinishCallback)
  local depthMeterWidth, depthMeterHeight = depthMeter:getDims()
  local xOffset = -2
  local yOffset = Screen.safeHeight * 0.5 - depthMeterHeight * 0.5 - StandardUIBuilder.STANDARD_BAR_MARGIN

  builder:addWidget(depthMeter, UIOverlay.EAST, xOffset, yOffset)
  self.depthMeter = depthMeter

  local livesDisplay = builder:addLivesDisplay(DepthConfiguration.getNumberOfLives())
  self.livesDisplay = livesDisplay

  local bonusWindowSize = (DepthConfiguration.SPEED_BONUS_SCREEN_PROPORTION_THRESHOLD * Screen.height)
  local function onQuickResponse(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 DepthConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end
  local bonusDecay = 0
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowSize, onQuickResponse, bonusDecay)
end

local function createPhysics(self)
  local function onMissedBubbleCallback(bubble)
    runInCoroutine(function()
      if self.bubbles[bubble] then
        TouchManager.removeTouchHandler(bubble)

        local identifier = bubble:getIdentifier()
        GameSession.contentTracker:notifyTimeout(identifier)

        local x, y = bubble:getPosition()
        handleMissedBubble(self, bubble, x, y)

        bubble:pop()
        self.physics:removeBubble(bubble)
        self.bubbles[bubble] = nil
      end
    end)
  end

  local physics = DepthPhysics.new(onMissedBubbleCallback)
  self.physics = physics
end

local function createBubbleGenerator(self)
  local function sendBubbleCallback()
    self:sendBubble()
  end

  local function incrementBubbleSpeedCallback()
    self:incrementBubbleSpeed()
  end

  local bubbleGenerator = DepthBubbleGenerator.new(sendBubbleCallback, incrementBubbleSpeedCallback)
  self.bubbleGenerator = bubbleGenerator
end

local function createStreakScroller(self)
  local streakScroller = DepthStreakScroller.new()
  self:insertAtCenter(streakScroller, Z_SCROLLER)
  self.streakScroller = streakScroller
end

local function createScroller(self)
  local scroller = DepthScroller.new()
  scroller:setLoc(CENTER_X, CENTER_Y)
  self:insert(scroller, Z_SCROLLER)
  self.scroller = scroller
end

local function createRedAlert(self)
  local redAlert = DepthRedAlert.new()
  redAlert:setLoc(CENTER_X, CENTER_Y)
  self:insert(redAlert, Z_RED_ALERT)
  self.redAlert = redAlert
end

-- Constructor ------------------------

function DepthGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, DepthGame)

  createGameElements(self)
  createUI(self)
  configureGameDifficulty(self)

  createPhysics(self)

  createBubbleGenerator(self)

  createStreakScroller(self)

  createScroller(self)

  createRedAlert(self)

  self.bubbleSpeed = DepthConstants.INITIAL_GAME_SPEED * DepthConfiguration.getInitialBubbleSpeedModifier()
  self.goodBubblesInARow = 0
  self.badBubblesInARow = 0

  self.isTouchEnabled = false

  self.scroller:startScrolling()
  self.streakScroller:startScrolling()

  return self
end
