require("actions.advanced.Shake")
require("DepthStarburst")
require("DepthCircleBurst")
require("DepthSquiggleBurst")

class("DepthIntro")

local OVERLAY_COLOR = ColorUtils.colorRGBToStd(24, 26, 54, 122)
local BUBBLE_COLOR = ColorUtils.colorRGBToStd(169, 188, 246)
local TAP_TO_START_COLOR = ColorUtils.colorRGBToStd(180, 180, 180)

local BUBBLE_EASE = AHEaseSimpleIn.new(0.9)

local function spawnAnimationBubble(self)
  local bubbleType = math.random(6)
  local bubble = Sprite.newWithName("intro_bubble_"..bubbleType..".png")
  bubble:setColor(unpack(BUBBLE_COLOR))

  local d = bubble:getDims()
  local r = d/2
  local bubbleGroup = Group.newDeprecated()
  local w, h = bubble:getDims()

  local margin = r * 2
  local x = math.random() * (Screen.width - margin * 2) + margin
  local y = -r
  local time = (math.random() * 0.75 + 0.25) * 10
  local scale = math.min(d/w, d/h)

  bubble:setScl(scale)
  bubbleGroup:setLoc(x, y)
  bubbleGroup:insert(bubble)
  bubble:setAttrLink(MOAIColor.ATTR_OPACITY, self.bubbles)
  self.bubbles:insert(bubbleGroup)

  local moveUp = AHLocationTo.new(bubbleGroup, time, BUBBLE_EASE, x, Screen.height + r)
  local die = function() bubbleGroup:removeSelf() end
  local animation = AHSequence.new(moveUp, die)
  animation:start()

  local xShakeAmp = 60
  local yShakeAmp = 0
  local shakesPerSecond = 20 * (math.random()/2 + 0.5)
  local durationSeconds = time * 2
  local shake = AHSmoothShake.new(bubble, xShakeAmp, yShakeAmp, shakesPerSecond, durationSeconds, completeHandler)
  shake:start()
end
local function animate(self)
  self.animating = true
  local bubbler = function()
    while(self.animating) do
      local delay = math.random() * 1
      spawnAnimationBubble(self)
      if(math.random() < 0.5) then
        spawnAnimationBubble(self)
      end
      waitFor(delay)
    end
  end
  runInCoroutine(bubbler)
end
local function createTip(self, sessionDescriptionString)
  local tip = Group.newDeprecated()
  local tipText = nil
  local difficulty = GameSession.getDifficulty()
  if(sessionDescriptionString == "15%"      and difficulty < 0.6) then
    tipText = "Tip: To calculate 15%, divide the denominator by 10; then take half of that amount and add together."
  elseif(sessionDescriptionString == "20%"  and difficulty < 0.55) then
    tipText = "Tip: To calculate 20%, divide the denominator by 10 to get 10%; then multiply by 2."
  elseif(sessionDescriptionString == "25%"  and difficulty < 0.55) then
    tipText = "Tip: To calculate 25%, multiply the numerator by 4; then verify it equals the denominator."
  end

  if(tipText) then
    local width = Screen.width * 0.9
    local font = Font.new(FontFaces.DIN_OT_LIGHT, 18)
    local tipView = DynamicTextArea.newSimple(tipText, width, font, horizontalAlignment, verticalAlignment, lineSpacing)
    tip:insert(tipView)
    tipView:setAttrLink(MOAIColor.ATTR_OPACITY, tip)
  end
  return tip
end

local function createLayout(self, sessionDescriptionString)
  local tapToStart = Label.new("pop to start", Font.new(FontFaces.DIN_OT_LIGHT, 18))
  local mainBubble = Sprite.newWithName("intro_large_bubble.png")
  local sessionDescriptionLabel = Label.new(sessionDescriptionString, Font.new(FontFaces.DIN_OT_LIGHT, 64))
  local overlay = FilledRectangle.newFullScreen(OVERLAY_COLOR)
  local tip = createTip(self, sessionDescriptionString)
  local bubbles = Group.newDeprecated()

  local bubbleW, bubbleH = mainBubble:getDims()
  sessionDescriptionLabel:setLoc(bubbleW/2, bubbleH/2)

  local centerX = Screen.width * 0.5
  local bubbleY = Screen.height * 0.535
  local tipY = Screen.safeBottom + 50

  mainBubble:insert(sessionDescriptionLabel)
  mainBubble:setLoc(centerX, bubbleY)
  overlay:setLoc(centerX, Screen.height/2)
  tip:setLoc(centerX, tipY)

  tapToStart:setLoc(centerX, bubbleY - bubbleH/4)

  self:insert(overlay)
  self:insert(bubbles)
  self:insert(tapToStart)
  self:insert(mainBubble)
  self:insert(tip)

  self.bubbles = bubbles
  self.radius = bubbleW/2
  self.mainBubble = mainBubble
  self.label = sessionDescriptionLabel
  self.background = overlay
  self.label.text = sessionDescriptionString
  self.tip = tip
  self.tapToStart = tapToStart

  self.bubbles:setOpacity(0.75)
  tapToStart:setColor(unpack(TAP_TO_START_COLOR))

  animate(self)
end


local function pop(self)
  local radius = self.radius

  local backScaleAction = AHScaleTo.new(self.mainBubble, 0.2, AHEaseBackIn.new(1.4), 1.1)
  local backFadeAction = AHOpacityTo.new(self.mainBubble, 0.2, nil, 0.0)
  local overlayFadeAction = AHOpacityTo.new(self.background, 0.2, nil, 0.0)
  local bubblesFadeAction = AHOpacityTo.new(self.bubbles, 1.7, nil, 0.0)
  local tipFadeAction = AHOpacityTo.new(self.tip, 0.2, nil, 0)

  self:takeChild(self.label)
  local labelAnimTime = 0.6
  local labelX = Screen.getWidthPart(0.5)
  local labelY = Screen.getHeightPart(0.945) - (Screen.fullHeight - Screen.safeTop)
  local scaleOutLabel = AHScaleTo.new(self.label, labelAnimTime, AHEaseBackIn.new(5), 0.25)
  local fadeOutLabel = AHOpacityTo.new(self.label, labelAnimTime, nil, 0.8)
  local fadeOutTapToStart = AHOpacityTo.new(self.tapToStart, labelAnimTime / 2, nil, 0.0)

  local replaceLabel = function()
    local x, y = labelX, labelY

    local newLabel = Label.new(self.label.text, Font.new(FontFaces.DIN_OT_BOLD, 28))
    self:insert(newLabel)
    newLabel:setLoc(x, y)
    newLabel:setOpacity(0)

    runInCoroutine(function()
      coroutine.yield()
      local hide = AHOpacityTo.new(self.label, 0.1, nil, 0)
      AHOpacityTo.new(newLabel, 0.1, nil, 0.8):start()
      local oldLabel = self.label
      self.label = newLabel
      hide:start()
      wait(hide)
      oldLabel:removeSelf()
    end)
  end
  AHSequence.new(AHDelay.new(labelAnimTime - 0.09), replaceLabel):start()
  local faceFadeAction = AHSequence.new(AHSpawn.new(scaleOutLabel, fadeOutLabel, fadeOutTapToStart), nil)


  local popAction = AHSpawn.new(
                                backScaleAction,
                                backFadeAction,
                                faceFadeAction,
                                overlayFadeAction,
                                bubblesFadeAction,
                                tipFadeAction
                                )

  runInCoroutine(function()
    coroutine.yield()
    local starburst = DepthStarburst.new(nil, 160, 180)
    local circleburst = DepthCircleBurst.new(30, nil, nil, 3, 6)
    starburst:setOpacity(0.45)

    starburst:setLoc(radius, radius)
    circleburst:setLoc(radius, radius)

    self.mainBubble:insert(starburst)
    self.mainBubble:insert(circleburst)

    self.starburst = starburst
    self.circleburst = circleburst

    coroutine.yield()
    local circleburstAction = circleburst:createBurstAction(0.8)
    circleburstAction:start()

    coroutine.yield()
    local starburstAction = starburst:createBurstAction()
    starburstAction:start()
  end)

  popAction:start()
  Sound.playSoundEffect("depth_correct")
  self.animating = false
end
local function initTouchHandling(self)
  TouchManager.addTouchHandler(self, function(event)
    if(event.phase == TouchManager.TOUCH_BEGAN) then
      pop(self)
      if(self.continueHandler) then self.continueHandler() end
      TouchManager.removeTouchHandler(self)
      self.actionLock:stop()
    end
  end)
end

function DepthIntro.new(sessionDescriptionString)
  local self = Group.newFullScreen()
  Mixin.mix(self, DepthIntro)
  createLayout(self, sessionDescriptionString)
  self:setAnchorPointBottomLeft()
  self.actionLock = AHDelay.new(math.huge)
  self.actionLock:start()
  initTouchHandling(self)
  return self
end

function DepthIntro.newWithPercentage(percentage)
  return DepthIntro.new(percentage.."%")
end

function DepthIntro:setContinueHandler(handler)
  self.continueHandler = handler
end
--- Returns a waitable that releases when the intro is over
function DepthIntro:getActionLock()
  return self.actionLock
end
