--- DepthPhysics
-- @author Created by Aaron Barrett on January 25, 2014.

class("DepthPhysics")
-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local GRAVITY = -14.4
local UNITS_TO_METERS_RATIO = 1.0 / 64.0

local VELOCITY_ITERATIONS = 1
local POSITION_ITERATIONS = 1

local WALL_THICKNESS = Screen.height * 0.075
local WALL_TOP = Screen.height * 2.0
local WALL_RIGHT = Screen.width
local WALL_LEFT = 0.0
local WALL_BOTTOM = -Screen.height

local DOT_REMOVER_TOP = -Screen.height * 0.2
local POPPER_BOTTOM = Screen.height
local BUBBLE_REMOVER_BOTTOM = Screen.height * 1.3

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function handleDotRemoverSensorCollision(self, fixtureA, fixtureB)
  local dot = nil
  if (fixtureA.isDotFixture) then
    dot = fixtureA.dot
  elseif (fixtureB.isDotFixture) then
    dot = fixtureB.dot
  end

  if (dot) then
    self:removeDot(dot)
    dot:removeSelf()
  end
end

local function handlePopperSensorCollision(self, fixtureA, fixtureB)
  local bubble = nil
  if (fixtureA.isBubbleFixture) then
    bubble = fixtureA.bubble
  elseif (fixtureB.isBubbleFixture) then
    bubble = fixtureB.bubble
  end

  if (bubble) then
    local shouldShowMissedAnswer = bubble:getIsGood()
    if (shouldShowMissedAnswer) then
      self.onMissedBubbleCallback(bubble)
    end
  end
end

local function handleBubbleRemoverSensorCollision(self, fixtureA, fixtureB)
  local bubble = nil
  if (fixtureA.isBubbleFixture) then
    bubble = fixtureA.bubble
  elseif (fixtureB.isBubbleFixture) then
    bubble = fixtureB.bubble
  end

  if (bubble) then
    self:removeBubble(bubble)
    bubble:removeSelf()
  end
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function DepthPhysics:destroy()
  local world = self.world
  world:stop()
end

function DepthPhysics:pause(isPaused)
  local world = self.world
  world:pause(isPaused)
end

function DepthPhysics:addBubble(bubble)
  local world = self.world

  local radius = bubble:getRadius()

  local body = world:addBody(MOAIBox2DBody.DYNAMIC)
  body:setFixedRotation(true)

  bubble:setBody(body)

  local fixture = body:addCircle(0.0, 0.0, radius)
  fixture.isBubbleFixture = true
  fixture.bubble = bubble
  fixture:setDensity(0.1)
  fixture:setFriction(0.0)
  fixture:setRestitution(0.3)
  body:resetMassData()

  local x = radius * 1.5 + math.random() * (Screen.width - radius * 3.0)
  local y = -radius
  body:setTransform(x, y)
  body:setLinearVelocity(0.0, 0.0)

  bubble:setAttrLink(MOAITransform.INHERIT_TRANSFORM, body, MOAITransform.TRANSFORM_TRAIT)
end

function DepthPhysics:addDot(dot, x, y)
  local world = self.world

  local radius = DepthDot.RADIUS

  local body = world:addBody(MOAIBox2DBody.DYNAMIC)
  body:setFixedRotation(true)

  dot:setBody(body)

  local fixture = body:addCircle(0.0, 0.0, radius)
  fixture.isDotFixture = true
  fixture.dot = dot
  fixture:setDensity(5.0)
  fixture:setFriction(0.0)
  fixture:setRestitution(0.2)
  body:resetMassData()

  body:setTransform(x, y)
  body:setLinearVelocity(0.0, 0.0)

  dot:setAttrLink(MOAITransform.INHERIT_TRANSFORM, body, MOAITransform.TRANSFORM_TRAIT)
end

function DepthPhysics:removeBubble(bubble)
  local body = bubble:getBody()
  body:destroy()

  bubble:setBody(nil)
end

function DepthPhysics:removeDot(dot)
  local body = dot:getBody()
  body:destroy()

  dot:setBody(nil)
end

function DepthPhysics:start()
  local world = self.world
  world:start()
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

local function addDotRemoverSensor(self, wallsBody)
  local dotRemoverSensor = wallsBody:addRect(WALL_LEFT, DOT_REMOVER_TOP - WALL_THICKNESS, WALL_RIGHT, DOT_REMOVER_TOP)
  dotRemoverSensor:setSensor(true)

  local function onDotRemoverSensorCollision(phase, fixtureA, fixtureB)
    handleDotRemoverSensorCollision(self, fixtureA, fixtureB)
  end

  local phasesToHandle = MOAIBox2DArbiter.BEGIN
  dotRemoverSensor:setCollisionHandler(onDotRemoverSensorCollision, phasesToHandle)
end

local function addPopperSensor(self, wallsBody)
  local popperSensor = wallsBody:addRect(WALL_LEFT, POPPER_BOTTOM, WALL_RIGHT, POPPER_BOTTOM + WALL_THICKNESS)
  popperSensor:setSensor(true)

  local function onPopperSensorCollision(phase, fixtureA, fixtureB)
    handlePopperSensorCollision(self, fixtureA, fixtureB)
  end

  local phasesToHandle = MOAIBox2DArbiter.BEGIN
  popperSensor:setCollisionHandler(onPopperSensorCollision, phasesToHandle)
end

local function addBubbleRemoverSensor(self, wallsBody)
  local bubbleRemoverSensor = wallsBody:addRect(WALL_LEFT, BUBBLE_REMOVER_BOTTOM, WALL_RIGHT, BUBBLE_REMOVER_BOTTOM + WALL_THICKNESS)
  bubbleRemoverSensor:setSensor(true)

  local function onBubbleRemoverSensorCollision(phase, fixtureA, fixtureB)
    handleBubbleRemoverSensorCollision(self, fixtureA, fixtureB)
  end

  local phasesToHandle = MOAIBox2DArbiter.BEGIN
  bubbleRemoverSensor:setCollisionHandler(onBubbleRemoverSensorCollision, phasesToHandle)
end

local function createWorld(self)
  local world = MOAIBox2DWorld.new()
  world:setIterations(VELOCITY_ITERATIONS, POSITION_ITERATIONS)
  world:setGravity(0, GRAVITY)
  world:setUnitsToMeters(UNITS_TO_METERS_RATIO)

  local wallsBody = world:addBody(MOAIBox2DBody.STATIC)
  wallsBody:addRect(WALL_LEFT - WALL_THICKNESS, WALL_BOTTOM, WALL_LEFT, WALL_TOP)
  wallsBody:addRect(WALL_RIGHT, WALL_BOTTOM, WALL_RIGHT + WALL_THICKNESS, WALL_TOP)

  addDotRemoverSensor(self, wallsBody)
  addPopperSensor(self, wallsBody)
  addBubbleRemoverSensor(self, wallsBody)

  self.world = world
end

-- Constructor ------------------------

function DepthPhysics.new(onMissedBubbleCallback)
  local self = {}
  Mixin.mix(self, DepthPhysics)

  self.onMissedBubbleCallback = onMissedBubbleCallback

  createWorld(self)

  return self
end
