--- DepthPrize
-- @author Created by Aaron Barrett on January 27, 2014.

class("DepthPrize")
-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local PRIZE_SIZE = 100.0
local WIDTH = PRIZE_SIZE
local HEIGHT = PRIZE_SIZE

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

local Z_BODY = 10
local Z_FIN = 20

local SWIM_AMOUNT = Screen.width * 1.2
local SWIM_DURATION = 0.9

local FIN_MIN_ROTATION = -15.0
local FIN_MAX_ROTATION = 17.5

local GROUPING_COUNT = 6
local PRIZES_PER_GROUPING = 4

local PRIZE_DESCRIPTORS = {
  {x = 8.0, y = 46.0},  --  1: Black cyan zombiefish
  {x = 44.0, y = 31.0}, --  2: Anglerfish
  {x = 22.0, y = 48.0}, --  3: Black squarefin squarepants
  {x = 2.0, y = 47.0},  --  4: Longfish is long
  {x = 30.0, y = 45.0}, --  5: Boxfish
  {x = 36.0, y = 52.0}, --  6: Ovalfish
  {x = 31.0, y = 57.0}, --  7: Black diamondfish -- ***Wrong fin?
  {x = 28.0, y = 49.0}, --  8: Rounhead flatbottom
  {x = 2.0, y = 38.0},  --  9: Blue diamondfish
  {x = 23.0, y = 43.0}, -- 10: Brown pineapplefish
  {x = 40.0, y = 49.0}, -- 11: Blue pixelated ninjastarfish
  {x = 29.0, y = 49.0}, -- 12: Blue striped birdonose
  {x = 17.0, y = 38.0}, -- 13: Testfish       -- ***Fin got trimmed?
  {x = 41.0, y = 50.0}, -- 14: Purple snootynose
  {x = 24.0, y = 40.0}, -- 15: Purple symmetricfish
  {x = 20.0, y = 49.0}, -- 16: Purple skinnyfish
  {x = 32.0, y = 70.0}, -- 17: Yellow stubblefish -- ***Fin got trimmed?
  {x = -8.0, y = 43.0}, -- 18: Purple wafflefish
  {x = 30.0, y = 55.0}, -- 19: Purple circledots
  {x = 48.0, y = 53.0}, -- 20: Purple birdonose -- Eye is higher in mockup
  {x = 40.0, y = 55.0}, -- 21: Yellow-finned tallfish --Has yellow stripe in mockup
  {x = 22.0, y = 55.0}, -- 22: Green stripedpantsfish
  {x = 34.0, y = 42.0}, -- 23: Yellow anvilfish
  {x = 28.0, y = 54.0}  -- 24: Turquoise oddfish
}

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------
-- Remember the sequence of prizes
local spawnedFish = {}
local FIN_EASE = AHEaseSimpleInOut.new(2.0)

local function registerFishSpawned(index)
  spawnedFish[#spawnedFish + 1] = index
end
function DepthPrize.getSpawnedFishSequence()
  return spawnedFish
end


local function startFinMotion(fish)
  local fin = fish.fin
  local finAction = AHLoop.new(
                               AHSequence.new(
                                              AHRotationTo.new(fin, 0.4, FIN_EASE, FIN_MAX_ROTATION),
                                              AHRotationTo.new(fin, 0.4, FIN_EASE, FIN_MIN_ROTATION)
                                              )
                               )
  finAction:start()
  return finAction
end


-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------


function DepthPrize:swimOut(shouldSwimLeft)
  local fin = self.fin

  local swimX = SWIM_AMOUNT
  local destinationXScale = 1.0

  if (shouldSwimLeft) then
    swimX = -swimX
  else
    destinationXScale = -destinationXScale
  end

  local swimAction = AHSequence.new(
                                    AHSpawn.new(
                                                AHScaleTo.new(self, 0.2, AHEaseSimpleOut.new(2.0), destinationXScale, 1.0),
                                                AHLocationXBy.new(self, SWIM_DURATION, AHEaseBackIn.new(1.1), swimX)
                                                ),
                                    AHCallLogic.new(function()
                                                        self:removeSelf()
                                                      end)
                                    )

  swimAction:start()
end

function DepthPrize:removeSelf()
  if(self.finAction) then
    self.finAction:stop()
  end
  self.super.removeSelf(self)
end


-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function DepthPrize.new(progressProportion)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, DepthPrize)

  local groupingNumber = math.floor(progressProportion * GROUPING_COUNT)
  if (groupingNumber >= GROUPING_COUNT) then
    groupingNumber = GROUPING_COUNT - 1
  end
  local groupingBottomIndex = groupingNumber * PRIZES_PER_GROUPING
  local prizeIndex = groupingBottomIndex + math.random(PRIZES_PER_GROUPING)

  local fish = DepthPrize.createFish(prizeIndex)
  self:insert(fish)
  self.finAction = fish:startFinMotion()

  registerFishSpawned(prizeIndex)

  return self
end

function DepthPrize.createFish(prizeIndex)
  local descriptor = PRIZE_DESCRIPTORS[prizeIndex]

  local bodyName = "fish_" .. prizeIndex .. ".png"
  local body = Sprite.newWithName(bodyName)
  body:setLoc(CENTER_X, CENTER_Y)

  local finX = descriptor.x
  local finY = descriptor.y
  local finName = "fin_" .. prizeIndex .. ".png"
  local fin = Sprite.newWithName(finName)
  fin:setAnchorPointTopLeft()
  fin:setLoc(finX, finY)
  fin:setRot(FIN_MIN_ROTATION)

  local w, h = body:getDims()
  local fish = Group.new(w, h)
  fish:insert(body, Z_BODY)
  fish:insert(fin, Z_FIN)
  fish.fin = fin
  fish.startFinMotion = startFinMotion
  fish:setAnchorPointBottomLeft()

  return fish
end
