--- DepthScroller
-- @author Created by Aaron Barrett on January 27, 2014.

class("DepthScroller")
-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

local LEVEL_SCROLL_SPEED = HEIGHT * 0.057

local LEVEL_SCROLL_DISTANCE = HEIGHT + DepthLevel.HEIGHT
local LEVEL_SCROLL_DURATION = LEVEL_SCROLL_DISTANCE / LEVEL_SCROLL_SPEED

local LEVEL_APPEAR_DURATION = DepthLevel.HEIGHT / LEVEL_SCROLL_SPEED

local LEVELS_PER_GAME = DepthConstants.GAME_DURATION / LEVEL_APPEAR_DURATION

local INITIAL_LEVEL_COUNT = 1

local INITIAL_LEVEL_OPACITY = 0.2
local FINAL_LEVEL_OPACITY = 0.1
local LEVEL_OPACITY_DECREMENT = (INITIAL_LEVEL_OPACITY - FINAL_LEVEL_OPACITY) / LEVELS_PER_GAME

local INTERVAL_BETWEEN_LEVELS = LEVEL_APPEAR_DURATION

local MIN_INTERVAL_BETWEEN_SCHOOLS = 5.2
local MAX_INTERVAL_BETWEEN_SCHOOLS = 16.1

local MIN_INTERVAL_BETWEEN_SPECKS = 2.5
local MAX_INTERVAL_BETWEEN_SPECKS = 5.8

local LEVEL_CLASSES = {
  DepthLevelWhale,
  DepthLevelJellyfish,
  DepthLevelLedge
}

local SPECKS_FILENAMES = {
  "specks.png",
  "haze.png"
}

local Z_SPECKS = 10
local Z_LEVEL = 20
local Z_SCHOOL = 30

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function getRandomLevelClass()
  local randomLevelClass = LEVEL_CLASSES[math.random(#LEVEL_CLASSES)]
  return randomLevelClass
end

local function createLevelAtY(self, y)
  local levelClass = getRandomLevelClass()
  local levelOpacity = self.levelOpacity
  local level = levelClass.new(levelOpacity)
  level:setLoc(CENTER_X, y)
  self:insert(level, Z_LEVEL)

  return level
end

local function startMovingLevel(self, level)
  local scrollActions = self.scrollActions

  local levelWidth, levelHeight = level:getDims()
  local levelScrollDistance = HEIGHT + levelHeight
  local levelScrollDuration = levelScrollDistance / LEVEL_SCROLL_SPEED

  local scrollAction = AHSequence.new(
                                      AHLocationYBy.new(level, levelScrollDuration, nil, -levelScrollDistance),
                                      AHCallLogic.new(function()
                                                          level:removeSelf()
                                                          local scrollAction = level.scrollAction
                                                          scrollActions[scrollAction] = nil
                                                        end)
                                      )
  level.scrollAction = scrollAction

  scrollActions[scrollAction] = true

  scrollAction:start()
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function DepthScroller:pause(isPaused)
  self.isPaused = isPaused

  self.intervalAction:pause(isPaused)

  local scrollActions = self.scrollActions
  for scrollAction in pairs(scrollActions) do
    scrollAction:pause(isPaused)
  end
end

function DepthScroller:startScrolling()
  local initialLevels = self.levels
  for i, initialLevel in ipairs(initialLevels) do
    startMovingLevel(self, initialLevel)
  end

  local intervalAction = AHLoop.new(
                                    AHSequence.new(
                                                   AHCallLogic.new(function()
                                                                       if (not (self.isPaused)) then
                                                                         local level = createLevelAtY(self, HEIGHT + DepthLevel.HEIGHT * 0.5)
                                                                         startMovingLevel(self, level)
                                                                         self.levelOpacity = self.levelOpacity - LEVEL_OPACITY_DECREMENT
                                                                         if (self.levelOpacity < FINAL_LEVEL_OPACITY) then
                                                                           self.levelOpacity = FINAL_LEVEL_OPACITY
                                                                         end
                                                                       end
                                                                     end),
                                                   AHDelay.new(INTERVAL_BETWEEN_LEVELS)
                                                   )
                                    )
  self.intervalAction = intervalAction
  intervalAction:start()

  local specksCoroutine = newCoroutineWithFunction(function()
                                                      while (true) do
                                                        local specksIntervalDuration = MIN_INTERVAL_BETWEEN_SPECKS + math.random() * (MAX_INTERVAL_BETWEEN_SPECKS - MIN_INTERVAL_BETWEEN_SPECKS)
                                                        local specksIntervalAction = AHDelay.new(specksIntervalDuration)
                                                        wait(specksIntervalAction:start())

                                                        if (not (self.isPaused)) then
                                                          local filename = SPECKS_FILENAMES[math.random(#SPECKS_FILENAMES)]

                                                          local specks = Sprite.newWithName(filename)
                                                          local specksWidth, specksHeight = specks:getDims()

                                                          local shouldFlipHorizontally = MathUtils.getRandomBoolean()
                                                          if (shouldFlipHorizontally) then
                                                            specks:setScl(-1.0, 1.0)
                                                          end

                                                          local specksX = math.random() * WIDTH

                                                          specks:setOpacity(0.3)
                                                          specks:setLoc(specksX, HEIGHT + specksHeight * 0.5)
                                                          self:insert(specks, Z_SPECKS)

                                                          startMovingLevel(self, specks)
                                                        end
                                                      end
                                                    end)
  specksCoroutine:start()

  local schoolCoroutine = newCoroutineWithFunction(function()
                                                       while (true) do
                                                         local schoolIntervalDuration = MIN_INTERVAL_BETWEEN_SCHOOLS + math.random() * (MAX_INTERVAL_BETWEEN_SCHOOLS - MIN_INTERVAL_BETWEEN_SCHOOLS)
                                                         local schoolIntervalAction = AHDelay.new(schoolIntervalDuration)
                                                         wait(schoolIntervalAction:start())

                                                         local schoolOpacity = self.levelOpacity
                                                         local school = DepthSchool.new(schoolOpacity)
                                                         local schoolInitialY = math.random() * HEIGHT * 1.1
                                                         school:setLoc(CENTER_X, schoolInitialY)
                                                         self:insert(school, Z_SCHOOL)
                                                         school:startSchool()
                                                         if (not (self.isPaused)) then
                                                           startMovingLevel(self, school)
                                                         end
                                                       end
                                                     end)
  schoolCoroutine:start()

  self.isPaused = false
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function DepthScroller.new()
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, DepthScroller)

  self.scrollActions = {}
  self.levelOpacity = INITIAL_LEVEL_OPACITY

  local levels = {}
  local y = DepthLevel.HEIGHT * 0.5
  for i = 1, INITIAL_LEVEL_COUNT do
    local initialLevel = createLevelAtY(self, y)
    y = y + DepthLevel.HEIGHT
    table.insert(levels, initialLevel)
  end
  self.levels = levels

  self.isPaused = true

  return self
end
