--- DepthStarburst
-- @author Created by Aaron Barrett on January 30, 2014.

class("DepthStarburst")

-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local Z_DASH = 10

local DASH_COUNT = 12
local DEGREES_PER_DASH = 360.0 / DASH_COUNT

local BIG_DASH_INDEX = 1
local SMALL_DASH_INDEX = 2
local DASH_INDEX_COUNT = 2

local BIG_DASH_RADIUS = 130.0
local SMALL_DASH_RADIUS = 120.0


local DIAMETER = BIG_DASH_RADIUS * 2.0
local WIDTH = DIAMETER
local HEIGHT = DIAMETER

local BIG_DASH_WIDTH = 4.0
local SMALL_DASH_WIDTH = 2.0
local DASH_WIDTHS = {
  BIG_DASH_WIDTH,
  SMALL_DASH_WIDTH
}

local BIG_DASH_HEIGHT = 45.0
local SMALL_DASH_HEIGHT = 22.5
local DASH_HEIGHTS = {
  BIG_DASH_HEIGHT,
  SMALL_DASH_HEIGHT
}

local DASH_COLOR = {0.76, 1.00, 0.99}

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function DepthStarburst:createBurstAction()
  local dashes = self.dashes
  for _, dash in ipairs(dashes) do
    dash:setScl(1, 1)
  end

  local fullBurstAction = AHSpawn.new()

  for _, dash in ipairs(dashes) do
    local dashAction = AHSpawn.new(
      AHLocationBy.new(dash, 0.4, AHEaseExponentialOut.new(), dash.dx, dash.dy),
      AHScaleTo.new(dash, 0.4, AHEaseSineOut.new(), 0.0),
      AHOpacityTo.new(dash, 0.4, nil, 0.0)
    )
    fullBurstAction:addAction(dashAction)
  end

  return fullBurstAction
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

-- Constructor ------------------------

function DepthStarburst.new(opacity, smallRadius, bigRadius)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, DepthStarburst)

  local bigRadius = bigRadius or BIG_DASH_RADIUS
  local smallRadius = smallRadius or SMALL_DASH_RADIUS
  local DASH_RADII =
  {
    bigRadius,
    smallRadius
  }

  local dashes = {}
  local startOpacity = 0.5

  local dashIndex = 1
  for i = 1, DASH_COUNT do
    local dashWidth = DASH_WIDTHS[dashIndex]
    local dashHeight = DASH_HEIGHTS[dashIndex]

    local dashDestinationRadius = DASH_RADII[dashIndex]

    local dash = FilledRectangle.new(dashWidth, dashHeight, DASH_COLOR)

    local angle = i * DEGREES_PER_DASH
    local dashAngle = angle + 90.0
    local angleInRadians = math.rad(angle)

    dash:setRot(dashAngle)

    local dashDestinationX = BIG_DASH_RADIUS + math.cos(angleInRadians) * dashDestinationRadius
    local dashDestinationY = BIG_DASH_RADIUS + math.sin(angleInRadians) * dashDestinationRadius
    local startRadius = 10

    dash.destinationX = dashDestinationX
    dash.destinationY = dashDestinationY
    dash.startX, dash.startY = BIG_DASH_RADIUS + startRadius * math.cos(angleInRadians), BIG_DASH_RADIUS + startRadius * math.sin(angleInRadians)
    dash.startOpacity = startOpacity
    dash.dx, dash.dy = dash.destinationX - dash.startX, dash.destinationY - dash.startY

    dash:setLoc(dash.startX, dash.startY)
    dash:setScl(1.0, 0.0)
    dash:setOpacity(startOpacity)

    self:insert(dash, Z_DASH)
    table.insert(dashes, dash)

    dashIndex = dashIndex + 1
    if (dashIndex > DASH_INDEX_COUNT) then
      dashIndex = 1
    end
  end

  self.dashes = dashes

  return self
end
