--- EggcornsButtonBar
-- @author Created by Jay Austria on August 14, 2015.

require("EggcornsCommon")

-- ===========================================================================
local EggcornsButton = privateClass()
-- ===========================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_HEIGHT = Screen.getHeightPart(0.12)

local BUTTON_CORNER_RADIUS = 6
local INNER_BUTTON_CORNER_RADIUS = BUTTON_CORNER_RADIUS - 1
local NEGATIVE_BUTTON_STRING = Localizer.get("ERROR")
local POSITIVE_BUTTON_STRING = Localizer.get("VALID")
local INVERTED_COLOR_CORRECT = EggcornsCommon.COLORS.BLACK


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local function animateTap(self)
  local activeOverlay = self.activeOverlay
  local tapAnimation = AHSequence.new(
    AHOpacityTo.new(activeOverlay, 0.1, nil, EggcornsCommon.VISIBLE_OPACITY),
    AHDelay.new(0.1),
    AHOpacityTo.new(activeOverlay, 0.1, nil, EggcornsCommon.INVISIBLE_OPACITY)
  )

  tapAnimation:start()
end

local function initializeButton(self)
  local touchBeganWrapper = function()
    if self.tapCallback then
      animateTap(self)
      self.tapCallback()
    end
  end

  Button.makeButton(self, touchBeganWrapper)
end

local function createFadeAction(self, endOpacity)
  local fadeDuration = EggcornsCommon.UI_FADE_DURATION
  local fadeOut = AHSpawn.new(
    AHOpacityTo.new(self, fadeDuration, nil, endOpacity),
    AHOpacityTo.new(self.labelRow, fadeDuration, nil, endOpacity)
  )

  return fadeOut
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsButton:setTapCallback(callback)
  self.tapCallback = callback
end

function EggcornsButton:createFadeOutAction(opacity)
  return createFadeAction(self, opacity)
end

function EggcornsButton:createFadeInAction()
  return createFadeAction(self, EggcornsCommon.VISIBLE_OPACITY)
end

function EggcornsButton:prepareForShowAction()
  self:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  self.labelRow:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

local function createLabelRow(icon, text, color)
  local elementSpacing = 10
  local labelRow = Row.new(elementSpacing)

  local symbol = Sprite.newWithName(icon)
  symbol:setColor(unpack(color))
  labelRow:insert(symbol)

  local textLabel = Label.new(text, Font.new(FontFaces.DIN_OT_BOLD, 18))
  textLabel:setColor(unpack(color))
  labelRow:insert(textLabel)

  labelRow:alignChildren()

  local isRecursive = true
  labelRow:linkOpacityAttrToChildren(isRecursive)

  return labelRow
end

local function createLayout(self, text, iconFilename)
  local labelRow = createLabelRow(iconFilename, text, EggcornsCommon.COLORS.WHITE)
  self:insertAtCenter(labelRow)
  self.labelRow = labelRow

  local activeOverlay = FilledRoundedRectangle.new(self:getWidth(), self:getHeight(), INNER_BUTTON_CORNER_RADIUS, EggcornsCommon.COLORS.WHITE)
  activeOverlay:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  self:insertAtCenter(activeOverlay)
  self.activeOverlay = activeOverlay

  local invertedLabelRow = createLabelRow(iconFilename, text, INVERTED_COLOR_CORRECT)
  invertedLabelRow:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  activeOverlay:insertAtCenter(invertedLabelRow)

  local isRecursive = true
  activeOverlay:linkOpacityAttrToChildren(isRecursive)
end

function EggcornsButton.new(buttonWidth, text, iconFilename)
  local self = HollowRoundedRectangle.new(buttonWidth, BUTTON_HEIGHT, INNER_BUTTON_CORNER_RADIUS, EggcornsCommon.COLORS.WHITE, 0.5)
  Mixin.mix(self, EggcornsButton)

  createLayout(self, text, iconFilename)
  initializeButton(self)

  return self
end

-- ============================================================================
class("EggcornsButtonBar")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BAR_CORNER_RADIUS = 6
local LINE_THICKNESS = 2
local GRADIENT_OVERLAY_OPACITY_START = 0.0
local GRADIENT_OVERLAY_OPACITY_END =  0.15

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsButtonBar:setAffirmativeButtonCallback(callback)
  self.affirmativeButton:setTapCallback(callback)
end

function EggcornsButtonBar:setNegativeButtonCallback(callback)
  self.negativeButton:setTapCallback(callback)
end

function EggcornsButtonBar:setTouchEnabled(enabled)
  self.negativeButton:setTouchEnabled(enabled)
  self.affirmativeButton:setTouchEnabled(enabled)
end

function EggcornsButtonBar:createFadeOutAction(opacity)
  opacity = opacity or EggcornsCommon.UI_FADE_OPACITY

  local fadeOut = AHSpawn.new(
    AHOpacityTo.new(self, EggcornsCommon.UI_FADE_DURATION, nil, opacity),
    self.gradientOverlay:newOpacityToAction(EggcornsCommon.UI_FADE_DURATION, nil, EggcornsCommon.INVISIBLE_OPACITY, EggcornsCommon.INVISIBLE_OPACITY),
    self.negativeButton:createFadeOutAction(opacity),
    self.affirmativeButton:createFadeOutAction(opacity)
  )

  return fadeOut
end

function EggcornsButtonBar:createFadeInAction()
  local fadeIn = AHSpawn.new(
    AHOpacityTo.new(self, EggcornsCommon.UI_FADE_DURATION, nil, EggcornsCommon.VISIBLE_OPACITY),
    self.gradientOverlay:newOpacityToAction(EggcornsCommon.UI_FADE_DURATION, nil, GRADIENT_OVERLAY_OPACITY_START, GRADIENT_OVERLAY_OPACITY_END),
    self.negativeButton:createFadeInAction(),
    self.affirmativeButton:createFadeInAction()
  )

  return fadeIn
end

function EggcornsButtonBar:prepareForShowAction()
  self:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  self.gradientOverlay.beginColor:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  self.gradientOverlay.endColor:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  self.negativeButton:prepareForShowAction()
  self.affirmativeButton:prepareForShowAction()
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, outsideRim)
  local buttonRow = Row.new()
  local buttonWidth = self:getWidthPart(0.5)

  local negativeButton = EggcornsButton.new(buttonWidth, NEGATIVE_BUTTON_STRING, "cross.png")
  buttonRow:insert(negativeButton)

  local affirmativeButton = EggcornsButton.new(buttonWidth, POSITIVE_BUTTON_STRING, "tick.png")
  buttonRow:insert(affirmativeButton)

  buttonRow:alignChildren()
  self:insertAtCenter(buttonRow)

  local centerDivider = FilledRectangle.new(LINE_THICKNESS, self:getHeight(), EggcornsCommon.COLORS.WHITE)
  self:insertAtCenter(centerDivider)
  centerDivider:inheritOpacity(self)
  self.centerDivider = centerDivider

  local startColor = {1, 1, 1, GRADIENT_OVERLAY_OPACITY_START}
  local endColor = {1, 1, 1, GRADIENT_OVERLAY_OPACITY_END}
  local gradientOverlay = RectangleGradient.new(self:getWidth(), self:getHeight(), startColor, endColor)
  self:insertAtCenter(gradientOverlay)
  self.gradientOverlay = gradientOverlay

  self:insertAtCenter(outsideRim)
  outsideRim:inheritOpacity(self)

  self.negativeButton = negativeButton
  self.affirmativeButton = affirmativeButton
end

function EggcornsButtonBar.new()
  local outsideRim = HollowRoundedRectangle.new(EggcornsCommon.GAME_UI_WIDTH, BUTTON_HEIGHT, BAR_CORNER_RADIUS, EggcornsCommon.COLORS.WHITE, LINE_THICKNESS)

  local self = Group.new(outsideRim:getDims())
  Mixin.mix(self, EggcornsButtonBar)

  createLayout(self, outsideRim)

  return self
end
