--- EggcornsConfiguration
-- @author Created by Jay Austria on August 14, 2015.

class("EggcornsConfiguration")

require("data.keyframes.KeyframeSet")
require("difficulty.DifficultyUtils")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

EggcornsConfiguration.CONCEPTS_TO_COMPLETE = 8
EggcornsConfiguration.CORRECT_ANSWER_POINTS = 800
EggcornsConfiguration.SPEED_BONUS_SECONDS = 2.5
EggcornsConfiguration.SPEED_BONUS_DECAY_START = 1.0
EggcornsConfiguration.MAX_SPEED_BONUS_POINTS = 200
EggcornsConfiguration.TUTORIAL_DELAY = 12
EggcornsConfiguration.DIFFICULTY_ESCALATION = 0.25
EggcornsConfiguration.ESCALATE_DIFFICULTY_AFTER_CORRECT_ANSWER_COUNT = 4

EggcornsConfiguration.NUMBER_OF_CONCEPTS_BEFORE_RESURFACING_MISTAKES = math.floor(
  EggcornsConfiguration.CONCEPTS_TO_COMPLETE *
  LearningFocusedRoundGenerator.DEFAULT_PROGRESS_BEFORE_RESURFACING
)

local INITIAL_LIVES_THRESHOLDS = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {lives = 6}),
  Keyframe.new(0.125, {lives = 5}),
  Keyframe.new(0.3, {lives = 4})
)

local TOTAL_POINTS_FOR_BONUS_LIVES = 3600
local MAX_LIVES_LOST_RANK_TWO = 1
local MAX_LIVES_LOST_RANK_THREE = 0


-- Value used in CoreMS's HighDifficultyReachedHighlightProducer, update if modified
local HIGHLIGHT_DIFFICULTY_THRESHOLD = 0.6

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsConfiguration.getInitialLivesCount()
  local difficulty = GameSession.getDifficulty()
  local frameForDifficulty = INITIAL_LIVES_THRESHOLDS:getFrameAt(difficulty)
  return frameForDifficulty.lives
end

function EggcornsConfiguration.getInstructionText()
  local text = nil
  if EggcornsConfiguration.willHighlightPhrases() then
    text = Localizer.get("INSTRUCTIONS")
  else
    text = Localizer.get("HIGH_DIFFICULTY_INSTRUCTIONS")
  end

  return text
end

function EggcornsConfiguration.getMaxSpeedBonusPointsPerQuestion()
  local pointsPerQuestion = (EggcornsConfiguration.MAX_SPEED_BONUS_POINTS / EggcornsConfiguration.CONCEPTS_TO_COMPLETE)

  return pointsPerQuestion
end

function EggcornsConfiguration.willHighlightPhrases()
  local isBelowThreshold = (GameSession.getDifficulty() < HIGHLIGHT_DIFFICULTY_THRESHOLD)

  return isBelowThreshold
end

function EggcornsConfiguration.shouldShowHighDifficultyInterstitial()
  return DifficultyUtils.shouldShowHighDifficultyInterstitial(HIGHLIGHT_DIFFICULTY_THRESHOLD, GameSession.getDifficulty())
end

function EggcornsConfiguration.getPointsPerLifeBonus()
  local maximumBonusLives = (EggcornsConfiguration.getInitialLivesCount() - 1)
  return TOTAL_POINTS_FOR_BONUS_LIVES / maximumBonusLives
end

function EggcornsConfiguration.getLivesBonusPointsFromTotalLives(livesCount)
  local extraLives = livesCount - 1
  local pointsPerLivesBonus = EggcornsConfiguration.getPointsPerLifeBonus()
  local livesBonusPoints = extraLives * pointsPerLivesBonus

  return livesBonusPoints
end

function EggcornsConfiguration.getPassScore()
  local passScore = EggcornsConfiguration.CORRECT_ANSWER_POINTS * EggcornsConfiguration.CONCEPTS_TO_COMPLETE

  return passScore
end

function EggcornsConfiguration.getRankTwoScore()
  local passScore = EggcornsConfiguration.getPassScore()

  local livesNeededRankTwo = EggcornsConfiguration.getInitialLivesCount() - MAX_LIVES_LOST_RANK_TWO
  local extraPointsRankTwo = EggcornsConfiguration.getLivesBonusPointsFromTotalLives(livesNeededRankTwo)
  local rankTwoScore = passScore + extraPointsRankTwo

  return rankTwoScore
end

function EggcornsConfiguration.getRankThreeScore()
  local passScore = EggcornsConfiguration.getPassScore()

  local livesNeededRankThree = EggcornsConfiguration.getInitialLivesCount() - MAX_LIVES_LOST_RANK_THREE
  local extraPointsRankThree = EggcornsConfiguration.getLivesBonusPointsFromTotalLives(livesNeededRankThree)
  local rankThreeScore = passScore + extraPointsRankThree

  return rankThreeScore
end

function EggcornsConfiguration.getDifficultyKeyFrames()
  local keyFrames = {
    {key = 0.0, valueProbabilities = {Easy = 1.0}},

    {key = 0.075, valueProbabilities = {Easy = 0.95,
                                        Medium = 0.05}},

    {key = 0.125, valueProbabilities = {Easy = 0.85,
                                        Medium = 0.15}},

    {key = 0.15, valueProbabilities = {Easy = 0.75,
                                       Medium = 0.25}},

    {key = 0.225, valueProbabilities = {Easy = 0.60,
                                        Medium = 0.40}},

    {key = 0.3, valueProbabilities = {Easy = 0.50,
                                      Medium = 0.45,
                                      Hard = 0.05}},

    {key = 0.375, valueProbabilities = {Easy = 0.40,
                                        Medium = 0.45,
                                        Hard = 0.15}},

    {key = 0.45, valueProbabilities = {Easy = 0.30,
                                       Medium = 0.45,
                                       Hard = 0.25}},

    {key = 0.525, valueProbabilities = {Easy = 0.20,
                                      Medium = 0.40,
                                      Hard = 0.40}},

    {key = 0.6, valueProbabilities = {Easy = 0.05,
                                      Medium = 0.45,
                                      Hard = 0.50}},

    {key = 0.675, valueProbabilities = {Medium = 0.30,
                                        Hard = 0.70}},

    {key = 0.75, valueProbabilities = {Medium = 0.05,
                                       Hard = 0.95}},

    {key = 0.8, valueProbabilities = {Hard = 0.9,
                                      Advanced = 0.1}},

    {key = 0.9, valueProbabilities = {Hard = 0.8,
                                      Advanced = 0.2}},

    {key = 1.0, valueProbabilities = {Hard = 0.7,
                                      Advanced = 0.3}},
  }

  return keyFrames
end
