--- EggcornsGame
-- @author Created by Jay Austria on August 14, 2015.

class("EggcornsGame")

require("StandardUIBuilder")
require("EggcornsConfiguration")
require("EggcornsScene")
require("EggcornsButtonBar")
require("EggcornsProgressBar")
require("EggcornsRoundGenerator")
require("EggcornsHighLevelInterstitial")
require("widgets.TapToContinue")
require("bonuses.ResponseSpeedTracker")
require("tutorials.TutorialHighlighter")
require("data.generators.LearningFocusedRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {1, 1, 1}

local TUTORIAL_OVERLAY_COLOR = {0, 0, 0, 0}
local TUTORIAL_MASK_COLOR = {1.0, 1.0, 1.0, 1.0}

local function configureGameDifficulty(self)
  local passScore = EggcornsConfiguration.getPassScore()
  local rankTwoScore = EggcornsConfiguration.getRankTwoScore()
  local rankThreeScore = EggcornsConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = EggcornsConfiguration.getDifficultyKeyFrames()
  local difficulty = GameSession.getDifficulty()
  self.baseDifficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, difficulty)
  local escalatedDifficulty = MathUtils.clamp(difficulty + EggcornsConfiguration.DIFFICULTY_ESCALATION, 0.0, 1.0)
  self.escalatedDifficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, escalatedDifficulty)
  GameSession.conceptGenerator:setDifficultyNameChooser(self.baseDifficultyNameChooser)

  GameSession.setInitialLives(EggcornsConfiguration.getInitialLivesCount())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function adjustDifficultyIfNeeded(self)
  local isAtOrAboveEscalationPoint = (GameSession.getCorrectAnswerCount() >= EggcornsConfiguration.ESCALATE_DIFFICULTY_AFTER_CORRECT_ANSWER_COUNT)
  local noMistakesMade = (GameSession.getIncorrectAnswerCount() == 0)
  if (noMistakesMade and isAtOrAboveEscalationPoint) then
    GameSession.conceptGenerator:setDifficultyNameChooser(self.escalatedDifficultyNameChooser)
  else
    GameSession.conceptGenerator:setDifficultyNameChooser(self.baseDifficultyNameChooser)
  end
end

local function allowMistakesToResurfaceIfNeeded(self)
  if (GameSession.getCorrectAnswerCount() > EggcornsConfiguration.NUMBER_OF_CONCEPTS_BEFORE_RESURFACING_MISTAKES) then
    self.roundGenerator:allowMistakesToResurface()
  end
end

local function getQuestionIdentifier(self)
  if (self.isConceptCorrect) then
    return "correct_form"
  else
    return "incorrect_form"
  end
end

local function didLose(self)
  return (self.livesCounter:getLives() <= 0)
end

local function didWin(self)
  return (self.progressBar:getProgressStep() >= EggcornsConfiguration.CONCEPTS_TO_COMPLETE)
end

local function showAndWaitForAnswerRationale(self)
  local scene = self.scene

  local showRationaleAction = scene:createShowRationaleAction()
  wait(showRationaleAction:start())

  local tapToContinueLayer = TapToContinue.new()
  tapToContinueLayer:waitForTrigger()

  local dismissRationaleAction = scene:createDismissRationaleAction(didLose(self))
  wait(dismissRationaleAction:start())
end

local function processIncorrectResponse(self)
  Sound.playSoundEffect("incorrect")
  GameSession.contentTracker:notifyIncorrectAnswer(self.currentConceptId, getQuestionIdentifier(self))
  self.roundGenerator:notifyMistake(self.currentConceptId)
  GameSession.notifyIncorrectAnswer(self.currentConceptId)
  GameSession.incrementIncorrectAnswerCount()
  self.livesCounter:decrement()
  self.responseSpeedTracker:notifyIncorrectResponse()

  local incorrectAnimation = self.scene:createIncorrectAnimation()
  wait(incorrectAnimation:start())

  runInCoroutine(function()
    showAndWaitForAnswerRationale(self)
    self.isWaitingForResponse = false
  end)
end

local function processCorrectResponse(self)
  Sound.playSoundEffect("correct")
  GameSession.contentTracker:notifyCorrectAnswer(self.currentConceptId, getQuestionIdentifier(self))
  GameSession.notifyCorrectAnswer(self.currentConceptId)
  GameSession.incrementCorrectAnswerCount()
  self.responseSpeedTracker:notifyCorrectResponse()

  local delayedUpdate = function()
    local scoreDelta = EggcornsConfiguration.CORRECT_ANSWER_POINTS
    GameSession.increaseScore(scoreDelta)

    self.progressBar:incrementProgress()

    if didWin(self) then
      ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), EggcornsConfiguration.getPointsPerLifeBonus(), self.bonusTray)
    end

    self.scoreLabel:setScore(GameSession.getScore())
  end

  local scene = self.scene
  scene:setDelayedUpdateCallback(delayedUpdate)

  local normalizedProgress = self.progressBar:getProgress()
  local correctAnimation = scene:createCorrectAnimation(normalizedProgress, self.isConceptCorrect)
  wait(correctAnimation:start())

  self.isWaitingForResponse = false
end

local function processNegativeButtonPressed(self)
  self.buttonBar:setTouchEnabled(false)
  self.shouldShowTutorial = false

  if not self.userHasResponded then
    self.userHasResponded = true
    local wasCorrect = not self.isConceptCorrect
    self.scene:toggleInstructionsAndProgressBar(wasCorrect)
  end

  if self.isConceptCorrect then
    processIncorrectResponse(self)
  else
    processCorrectResponse(self)
  end
end

local function processAffirmativeButtonPressed(self)
  self.buttonBar:setTouchEnabled(false)
  self.shouldShowTutorial = false

  if not self.userHasResponded then
    self.userHasResponded = true
    local wasCorrect = self.isConceptCorrect
    self.scene:toggleInstructionsAndProgressBar(wasCorrect)
    self.scene:toggleInstructionsAndProgressBar()
  end

  if self.isConceptCorrect then
    processCorrectResponse(self)
  else
    processIncorrectResponse(self)
  end
end

local function isGameEnded(self)
  return (didWin(self) or didLose(self))
end

local function showTutorial(self)
  self.buttonBar:setTouchEnabled(false)

  local highlightButton = (self.isConceptCorrect and self.buttonBar.affirmativeButton) or self.buttonBar.negativeButton
  local generateHole = function()
    local sharpCorners
    if self.isConceptCorrect then
      sharpCorners = {FilledRoundedRectangle.CORNERS.TOP_LEFT, FilledRoundedRectangle.CORNERS.BOTTOM_LEFT}
    else
      sharpCorners = {FilledRoundedRectangle.CORNERS.TOP_RIGHT, FilledRoundedRectangle.CORNERS.BOTTOM_RIGHT}
    end

    local hole = FilledRoundedRectangle.new(highlightButton:getWidth(), highlightButton:getHeight(), 5, TUTORIAL_MASK_COLOR, sharpCorners)
    local x, y = self:transformLoc(highlightButton)
    x = x + hole:getWidthPart(0.5)
    y = y + hole:getHeightPart(0.5)
    hole:setLoc(x, y)
    return hole
  end

  local tutorialOverlay = TutorialHighlighter.newWithMasks({generateHole})
  tutorialOverlay:setOverlayColor(unpack(TUTORIAL_OVERLAY_COLOR))
  self:insertAtCenter(tutorialOverlay)
  tutorialOverlay:createFadeInAnimation():start()
  tutorialOverlay:waitForTargetTap()

  if self.isConceptCorrect then
    processAffirmativeButtonPressed(self)
  else
    processNegativeButtonPressed(self)
  end

  wait(tutorialOverlay:createFadeOutAnimation():start())
  tutorialOverlay:removeSelf()
end

local function startRoundWithConcept(self, concept, isCorrectForm)
  self.currentConceptId = concept.identifier
  self.isConceptCorrect = isCorrectForm
  self.isWaitingForResponse = true

  GameSession.contentTracker:notifyPresentedConcept(self.currentConceptId)

  if self.shouldShowTutorial then
    local tutorialDelay = EggcornsConfiguration.TUTORIAL_DELAY
    Timer.makeTimer(self, tutorialDelay)
    self:setTimeoutHandler(function()
      if self.shouldShowTutorial then
        showTutorial(self)
      end
    end)
    self:start()
  end

  local isWaitingForResponseWrapper = function()
    return self.isWaitingForResponse
  end

  local formattedPrompt = EggcornsRoundGenerator:getFormattedPromptForConcept(concept, isCorrectForm)
  self.scene:presentPrompt(formattedPrompt, concept.rationale, concept.correctedReplacementText)

  self.buttonBar:setTouchEnabled(true)
  self.responseSpeedTracker:notifyStart()

  waitWhile(isWaitingForResponseWrapper)
end

local function showAndWaitForHighLevelInterstitial(self)
  local interstitial = EggcornsHighLevelInterstitial.new()
  interstitial:prepareForShowAction()

  local tapToContinue = TapToContinue.new(interstitial)
  wait(interstitial:startShowAction())
  tapToContinue:waitForTrigger()
  wait(interstitial:startHideAction())
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsGame:gameFlow()
  if EggcornsConfiguration.shouldShowHighDifficultyInterstitial() then
    showAndWaitForHighLevelInterstitial(self)
  end

  wait(self.scene:startFadeInAction())

  self.shouldShowTutorial = true
  repeat
    adjustDifficultyIfNeeded(self)
    allowMistakesToResurfaceIfNeeded(self)

    local concept = self.roundGenerator:next()
    local isCorrectForm = MathUtils.getRandomBoolean()
    startRoundWithConcept(self, concept, isCorrectForm)
  until isGameEnded(self)

  if didWin(self) then
    self.scoreLabel:setScore(GameSession.getScore())

    local delayedUIFadeAction = AHDelayed.new(1.5, self.uiOverlay:createFadeOutAction())
    delayedUIFadeAction:start()

    self:showDefaultGameWinOutro()

    wait(self.scene:playOutroAnimation())

    GameSession.notifyGameWin()
  else
    GameSession.notifyGameLoss()
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local negativeButtonWrapper = function()
    processNegativeButtonPressed(self)
  end

  local affirmativeButtonWrapper = function()
    processAffirmativeButtonPressed(self)
  end

  local buttonBar = EggcornsButtonBar.new()
  buttonBar:setNegativeButtonCallback(negativeButtonWrapper)
  buttonBar:setAffirmativeButtonCallback(affirmativeButtonWrapper)
  buttonBar:setTouchEnabled(false)
  self.buttonBar = buttonBar

  local progressBar = EggcornsProgressBar.newHidden(EggcornsConfiguration.CONCEPTS_TO_COMPLETE)
  self.progressBar = progressBar

  local scene = EggcornsScene.new(buttonBar, progressBar)
  scene:prepareForFadeInAction()
  self:insertAtCenter(scene)
  self.scene = scene

  local quickResponseWrapper = function(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 EggcornsConfiguration.getMaxSpeedBonusPointsPerQuestion(),
                                 self.bonusTray)
  end

  local responseSpeedTracker = ResponseSpeedTracker.new(EggcornsConfiguration.SPEED_BONUS_SECONDS,
                                                        quickResponseWrapper,
                                                        EggcornsConfiguration.SPEED_BONUS_DECAY_START)
  self.responseSpeedTracker = responseSpeedTracker
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(EggcornsConfiguration.getInitialLivesCount())
end

-- Constructor ------------------------

function EggcornsGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, EggcornsGame)

  EggcornsCommon.initialize()

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.roundGenerator = LearningFocusedRoundGenerator.new(EggcornsRoundGenerator.new())
  self.userHasResponded = false

  return self
end
