--- EggcornsOutroLayer
-- @author Created by Jay Austria on September 3, 2015.

class("EggcornsOutroLayer")

require("EggcornsScreenShake")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GLOBE_SPRITE_SCALE = 2

-- ASSUMES 8 DOTS
local DOT_END_POINTS = {
  {x = 0.26, y = 0.16},
  {x = 0.33, y = 0.14},
  {x = 0.40, y = 0.13},
  {x = 0.47, y = 0.13},
  {x = 0.52, y = 0.16},
  {x = 0.57, y = 0.19},
  {x = 0.6, y = 0.22},
  {x = 0.62, y = 0.24},
}

local CHECKMARK_ANIMATION_DESCRIPTORS = {
  {
    filename = "check_1.png",
    opacity = EggcornsCommon.VISIBLE_OPACITY,
    centerXOffsetMultiplier = 0,
    centerYOffsetMultiplier = 0,
    driftXMultiplier = 0,
    driftYMultiplier = 0
  },
  {
    filename = "check_2.png",
    opacity = 0.7 ,
    centerXOffsetMultiplier = 0.06,
    centerYOffsetMultiplier = 0.04,
    driftXMultiplier = -0.02,
    driftYMultiplier = -0.04
  },
  {
    filename = "check_2.png",
    opacity = EggcornsCommon.VISIBLE_OPACITY,
    centerXOffsetMultiplier = 0.02,
    centerYOffsetMultiplier = 0.04,
    driftXMultiplier = -0.04,
    driftYMultiplier = -0.02
  }
}

local DISTANCE_PADDING = 5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createCheckmarkAction(self, checkmarkIndex, checkmark)
  local descriptor = CHECKMARK_ANIMATION_DESCRIPTORS[checkmarkIndex]

  local driftX = self:getWidthPart(descriptor.driftXMultiplier)
  local driftY = self:getHeightPart(descriptor.driftYMultiplier)

  local checkmarkAction = AHSpawn.new(
    AHScaleTo.new(checkmark, 1.0, AHEaseBackOut.new(), 1.0),
    AHLocationBy.new(checkmark, 2.5, AHEaseSimpleInOut.new(), driftX, driftY)
  )

  return checkmarkAction
end

local function createCachedAnimations(self)
  local globeSprite = self.globeSprite
  local globeAnimation = Animation.newWithPrefix("outro_")
  local animateGlobeAction = AHAnimate.new(globeSprite, globeAnimation, 30)

  self.animateGlobeAction = animateGlobeAction
end

local function createSwallowDotsAction(self)
  local swallowDotsAction = AHSpawn.new()
  local takenDots = self.takenDots
  local globeSprite = self.globeSprite

  for dotIndex, dot in ipairs(takenDots) do
    local endPoint = DOT_END_POINTS[dotIndex]

    local targetX = globeSprite:getXLeft() + globeSprite:getEffectiveWidthPart(endPoint.x)
    local targetY = globeSprite:getYBottom() + globeSprite:getEffectiveHeightPart(endPoint.y)

    local actionDuration = 0.8
    swallowDotsAction:addAction(AHSpawn.new(
      AHLocationXTo.new(dot, actionDuration, AHEaseExponentialIn.new(), targetX),
      AHLocationYTo.new(dot, actionDuration, AHEaseExponentialIn.new(), targetY)
    ))
  end

  return swallowDotsAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsOutroLayer:createOutroAction()
  local swallowDots = createSwallowDotsAction(self)
  local animateGlobe = self.animateGlobeAction
  local globeSprite = self.globeSprite

  local beginTrackingTrails = function()
    for _, dot in ipairs(self.takenDots) do
      local trail = self.trailsForDots[dot]
      trail:follow()
    end
  end

  local removeDots = function()
    for _, dot in ipairs(self.takenDots) do
      local trail = self.trailsForDots[dot]
      local destroyTrail = trail:createDestroyAction()

      destroyTrail()
      dot:removeSelf()
    end
  end

  local checkmarkActions = AHSpawn.new()
  local animationDescriptor = nil
  for checkmarkIndex, checkmark in ipairs(self.checkmarks) do
    checkmarkActions:addAction(createCheckmarkAction(self, checkmarkIndex, checkmark))
  end

  local explosionDelay = 2.1
  local outroAction = AHSpawn.new(
    function() Sound.playSoundEffect("outro") end,
    beginTrackingTrails,
    AHSpawn.new(
      swallowDots,
      AHDelayed.new(0.8, AHSpawn.new(
        AHOpacityTo.new(globeSprite, 0.05, nil, EggcornsCommon.VISIBLE_OPACITY)),
        AHSpawn.new(animateGlobe, removeDots)
      )
    ),
    AHDelayed.new(explosionDelay, AHSpawn.new(
      EggcornsScreenShake.createOutroScreenShake(),
      AHScaleTo.new(self.explosionCircle, 0.2, AHEaseSimpleIn.new(), 1.0),
      checkmarkActions
    ))
  )

  return outroAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, dots)
  local globeSprite = Sprite.newWithName("outro_000.png")
  globeSprite:setScl(GLOBE_SPRITE_SCALE)
  globeSprite:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
  globeSprite:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  self:insert(globeSprite)
  self.globeSprite = globeSprite

  local trailsForDots = {}

  for _, dot in ipairs(dots) do
    self:takeChild(dot)

    local trailRadius = dot:getWidthPart(0.5)
    local maxScale = 1
    local trail = EggcornsBlipTrail.newDisabled(dot, trailRadius, maxScale)
    self:insertAtCenter(trail)

    trailsForDots[dot] = trail
  end

  self.takenDots = dots
  self.trailsForDots = trailsForDots

  local distanceFromMidToCorner = MathUtils.distance(0, 0, Screen.getWidthPart(0.5), Screen.getHeightPart(0.5)) + DISTANCE_PADDING
  local explosionCircle = FilledCircle.new(distanceFromMidToCorner, EggcornsCommon.COLORS.OVEREXPOSED_AQUA)
  explosionCircle:setScl(0)
  self:insertAtCenter(explosionCircle)
  self.explosionCircle = explosionCircle

  local checkmarks = {}
  local x, y = nil
  for _, checkmarkDescriptor in ipairs(CHECKMARK_ANIMATION_DESCRIPTORS) do
    local checkmark = Sprite.newWithName(checkmarkDescriptor.filename)
    checkmark:setScl(0)
    checkmark:setOpacity(checkmarkDescriptor.opacity)
    x = Screen.getSafeX(0.5 + checkmarkDescriptor.centerXOffsetMultiplier)
    y = Screen.getSafeY(0.5 + checkmarkDescriptor.centerYOffsetMultiplier)

    checkmark:setLoc(x, y)
    self:insert(checkmark)
    table.insert(checkmarks, checkmark)
  end
  self.checkmarks = checkmarks
end

function EggcornsOutroLayer.newWithDots(dots)
  local numberOfDots = #dots
  local numberOfDefinedPosition = #DOT_END_POINTS
  assert(numberOfDots <= numberOfDefinedPosition, string.format("Insufficient positions defined for where to place %d dots.", numberOfDots))

  local self = Group.newFullScreen()
  Mixin.mix(self, EggcornsOutroLayer)

  createLayout(self, dots)
  createCachedAnimations(self)

  return self
end
