--- EggcornsScene
-- @author Created by Jay Austria on August 14, 2015.

class("EggcornsScene")

require("EggcornsBackground")
require("EggcornsPromptMorpher")
require("EggcornsBlipTrail")
require("EggcornsTextView")
require("EggcornsRationaleOverlay")
require("EggcornsOutroLayer")
require("EggcornsScreenShake")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COLUMN_MARGIN_Y = Screen.getHeightPart(0.0675)

local EMPTY_HEIGHT = 0
local PROMPT_Y = Screen.getSafeY(0.637)
local INSTRUCTION_LABEL_X = Screen.getSafeX(0.5)
local INSTRUCTIONS_LABEL_OFFSET_Y = -Screen.getHeightPart(0.035)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function waitOnLowerPrompt(self, textArea)
  local lowerAction = AHSpawn.new()

  lowerAction:addAction(textArea:createAnimateInAction())

  if self.backings then
    local backingsFallDuration = EggcornsCommon.FALL_DURATION - 0.2
    for _, backing in ipairs(self.backings) do
      lowerAction:addAction(AHSpawn.new(
        AHLocationYBy.new(backing, backingsFallDuration, AHEaseExponentialOut.new(), -Screen.fullHeight),
        AHRotationTo.new(backing, backingsFallDuration, AHEaseExponentialOut.new(), 0)
      ))
    end
  end

  wait(lowerAction:start())
end

local function createDropEverythingAction(self)
  local textArea = self.currentTextArea
  local dropAction = textArea:createDropAction()
  local dropBackings = AHSpawn.new()
  local minAngle = -90
  local maxAngle = 90
  local dropHeight = -Screen.fullHeight
  for _, backing in ipairs(self.backings) do
    local randomAngle = MathUtils.getRandomFloat(minAngle, maxAngle)
    local backingAction = AHSequence.new(
      AHSpawn.new(
        AHLocationYBy.new(backing, 0.8, AHEaseExponentialIn.new(), dropHeight),
        AHRotationTo.new(backing, 0.6, nil, randomAngle)
      ),
      AHRemove.new(backing)
    )

    dropBackings:addAction(backingAction)
  end

  return AHSpawn.new(dropAction, dropBackings)
end

local function getLeftEdge(view)
  local leftEdgeX = view:getX() - EggcornsCommon.BACKING_PADDING_X

  return leftEdgeX
end

local function getRightEdge(view)
  local rightEdgeX = view:getX() + view:getWidth() + EggcornsCommon.BACKING_PADDING_X

  return rightEdgeX
end

local function createBacking(self, viewsInBackingInLine, textArea, color)
  local leftmostView = viewsInBackingInLine[1]
  local leftX = getLeftEdge(leftmostView)

  local rightmostView = viewsInBackingInLine[#viewsInBackingInLine]
  local rightX = getRightEdge(rightmostView)

  local width = rightX - leftX
  local height = leftmostView:getHeight()

  local cornerRadius = 4
  local backing = FilledRoundedRectangle.new(width, height, cornerRadius, color)
  local x = leftX + (width * 0.5)
  local y = leftmostView:getY() + leftmostView:getHeightPart(0.5)
  backing:setLoc(x, y)
  textArea:insert(backing)
  backing:bringToBack()

  return backing
end

local function addBackings(self, textArea, color)
  local startX = nil
  local endX = nil
  local didFindHighlightStart = false
  local didFindHighlightEnd = false

  local backings = {}
  local viewsInBackingInLine = {}
  local lines = textArea.layoutBuilder:getLines()

  for _, line in ipairs(lines) do
    for _, view in ipairs(line) do
      local item = textArea.itemsForViews[view]
      if item:hasAttribute("hasBacking") then
        table.insert(viewsInBackingInLine, view)
      end
    end

    local lineNeedsBackings = (#viewsInBackingInLine > 0)

    if lineNeedsBackings then
      local backing = createBacking(self, viewsInBackingInLine, textArea, color)
      table.insert(backings, backing)
    end

    -- Start fresh for next line
    viewsInBackingInLine = {}
  end

  self.backings = backings
end

local function playPromptMorphAnimation(self, isConceptCorrect)
  if isConceptCorrect and EggcornsConfiguration.willHighlightPhrases() then
    return AHDoNothing.new()
  end

  -- If we aren't changing the text and the correct text is not already highlighted,
  -- we don't want to introduce margins that could change line breaks
  local includeMargins = not isConceptCorrect

  -- PROBLEM: This fires animations immediately! As a result, the area of code that calls this
  -- playPromptMorphAnimation can't be wait()ed on, normally.
  self.currentTextArea:replaceWithCorrected(self.correctedReplacementText, includeMargins)

  local backingsAction = AHSpawn.new()
  local backings = self.backings
  local animationDuration = 0.01
  for _, backing in ipairs(backings) do
    local fadeOut = AHOpacityTo.new(backing, animationDuration, nil, EggcornsCommon.INVISIBLE_OPACITY)
    local shrink = AHScaleTo.new(backing, animationDuration, nil, 0)

    backingsAction:addAction(AHSequence.new(
      function() backing:setColor(unpack(EggcornsCommon.COLORS.AQUA)) end,
      AHSpawn.new(fadeOut, shrink),
      AHRemove.new(backing),
      AHDelay.new(0.5) -- HACK, this addresses the problem above where we can't wait on the text replacement actions
    ))
  end

  return backingsAction:start()
end

local function removeAllBackings(self)
  for _, backing in ipairs(self.backings) do
    backing:removeSelf()
  end
end

local function createColorizeBackingsAction(self, color)
  local colorizeAction = AHSpawn.new()

  local backings = self.backings
  for _, backing in ipairs(backings) do
    colorizeAction:addAction(function() backing:setColor(unpack(color)) end)
  end

  return colorizeAction
end

local function prepareBackingsForFallAction(self)
  local backings = self.backings
  for _, backing in ipairs(backings) do
    local x, y = backing:getLoc()
    local offscreenY = y + Screen.fullHeight

    backing:setLoc(x, offscreenY)
  end
end

local function prepareBackingsForAppearAction(self)
  local backings = self.backings
  for _, backing in ipairs(backings) do
    backing:setScl(0)
  end
end

local function createBackingsAppearAction(self)
  local appearAction = AHSpawn.new()
  local backings = self.backings
  for _, backing in ipairs(backings) do
    appearAction:addAction(AHScaleTo.new(backing, 0.01, nil, 1))
  end

  return appearAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------


function EggcornsScene:presentPrompt(prompt, rationaleText, correctedReplacementText)
  self.correctedReplacementText = correctedReplacementText

  local textArea = EggcornsTextView.new(Screen.getSafeWidthPart(0.8))
  textArea:setText(prompt)
  self:insert(textArea)
  self.currentTextArea = textArea

  textArea:setLoc(Screen.getSafeX(0.5), PROMPT_Y)

  textArea:prepareToAnimateIn()

  if EggcornsConfiguration.willHighlightPhrases() then
    addBackings(self, textArea, EggcornsCommon.COLORS.WHITE)
    prepareBackingsForFallAction(self)
  end

  local rationaleOverlay = self.rationaleOverlay
  local bottomY = textArea:getYBottom()
  rationaleOverlay:prepareToDisplay(rationaleText, bottomY)

  waitOnLowerPrompt(self, textArea)

  local promptMorpher = EggcornsPromptMorpher.new()
  promptMorpher:setLoc(textArea:getLoc())
  self:insert(promptMorpher)
  promptMorpher:attachTextArea(textArea)
  promptMorpher:setEndLocation(self.progressBar:getNextDotLocation())
  self.promptMorpher = promptMorpher

  local blipRadius = EggcornsProgressBar.CIRCLE_DIAMETER / 2
  local blipTrail = EggcornsBlipTrail.newDisabled(promptMorpher, blipRadius)
  self:insertAtCenter(blipTrail)
  self.blipTrail = blipTrail
end

function EggcornsScene:createCorrectAnimation(progress, isConceptCorrect)
  local delayedCallbackWrapper = function()
    if self.delayedUpdateCallback then
      self.delayedUpdateCallback()
    end

    EggcornsScreenShake.playScreenShake(progress)
    Sound.playSoundEffect("impact")

    wait(self.background:fadeInScreenGlow())
    self.background:fadeOutScreenGlow()
  end

  local beginTrackingTrails = function()
    local blipTrail = self.blipTrail
    blipTrail:follow()
  end
  self.promptMorpher:setOnShowBlipCallback(beginTrackingTrails)

  local willAddBackings = (not isConceptCorrect) or
                          (not EggcornsConfiguration.willHighlightPhrases())

  local morphDelay = nil
  if not willAddBackings then
    morphDelay = 0.5
  else
    morphDelay = 1.1
  end

  local impactDelay = morphDelay + 0.7

  local correctAnimation =
    AHSpawn.new(
      AHDelayed.new(impactDelay, delayedCallbackWrapper),
      self.currentTextArea:createBounceAction(),
      function() playPromptMorphAnimation(self, isConceptCorrect) end,
      createColorizeBackingsAction(self, EggcornsCommon.COLORS.AQUA),
      AHSequence.new(
        self.background:createFadeInEdgeGlowAction(),
        AHDelayed.new(0.1, self.background:createFadeOutEdgeGlowAction())
      ),
      AHDelayed.new(0.1, function()
        if willAddBackings then
          addBackings(self, self.currentTextArea, EggcornsCommon.COLORS.AQUA)
          prepareBackingsForAppearAction(self)
          createBackingsAppearAction(self):start()
        end
      end),
      AHDelayed.new(morphDelay, AHSequence.new(
        function()
          removeAllBackings(self)

          Sound.playSoundEffect("woosh")
          wait(self.promptMorpher:createAndPlayMorphAction())
        end,
        AHSpawn.new(
          AHRemove.new(self.promptMorpher, self.currentTextArea),
          self.blipTrail:createDestroyAction()
        )
      ))
    )

  return correctAnimation
end

function EggcornsScene:createIncorrectAnimation()
  local bouncePrompt = self.currentTextArea:createBounceAction()

  local ignoreDots = false
  local fadeDuration = 0.1
  local incorrectAnimation = AHSpawn.new(
    bouncePrompt,
    self.background:createTintRedAction(),
    self.progressBar:createFadeOutAction(ignoreDots, fadeDuration),
    createColorizeBackingsAction(self, EggcornsCommon.COLORS.LIGHT_RED)
  )

  return incorrectAnimation
end

function EggcornsScene:setDelayedUpdateCallback(callback)
  self.delayedUpdateCallback = callback
end

function EggcornsScene:createShowRationaleAction()
  local buttonBar = self.buttonBar
  local rationaleOverlay = self.rationaleOverlay

  local showAction = AHSpawn.new(
    buttonBar:createFadeOutAction(),
    rationaleOverlay:createDisplayAction()
  )

  return showAction
end

function EggcornsScene:createDismissRationaleAction(didUserLose)
  local buttonBar = self.buttonBar
  local progressBar = self.progressBar
  local rationaleOverlay = self.rationaleOverlay

  local dismissAction = AHSpawn.new(
    rationaleOverlay:createDismissAction(),
    AHSequence.new(
      createDropEverythingAction(self),
      AHRemove.new(self.currentTextArea)
    )
  )

  if not didUserLose then
    dismissAction:addAction(AHDelayed.new(0.75, AHSpawn.new(
      self.background:createTintDefaultAction(),
      buttonBar:createFadeInAction(),
      progressBar:createFadeInAction())
    ))
  end

  return dismissAction
end

function EggcornsScene:playOutroAnimation()
  local dots = self.progressBar:getDots()
  local outro = EggcornsOutroLayer.newWithDots(dots)
  self:insertAtCenter(outro)

  local globeExplodeDelayDuration = 2.1
  local ignoreDots = true
  local outroAnimation = AHSpawn.new(
    outro:createOutroAction(),
    AHDelayed.new(globeExplodeDelayDuration, self.background:createOutroAction()),
    self.buttonBar:createFadeOutAction(EggcornsCommon.INVISIBLE_OPACITY),
    self.progressBar:createFadeOutAction(ignoreDots)
  )

  return outroAnimation:start()
end

function EggcornsScene:toggleInstructionsAndProgressBar(doFadeInProgressBar)
  local instructionsLabel = self.instructionsLabel
  local fadeOutInstructions = AHSequence.new(
    AHOpacityTo.new(instructionsLabel, 0.1, nil, EggcornsCommon.INVISIBLE_OPACITY),
    AHRemove.new(instructionsLabel)
  )

  local fadeInProgressBar = self.progressBar:createFadeInAction()

  local toggleAction = AHSpawn.new(fadeOutInstructions)
  if doFadeInProgressBar then
    toggleAction:addAction(fadeInProgressBar)
  end

  toggleAction:start()
end

function EggcornsScene:prepareForFadeInAction()
  self.buttonBar:prepareForShowAction()
  self.instructionsLabel:setOpacity(EggcornsCommon.INVISIBLE_OPACITY)
end

function EggcornsScene:startFadeInAction()
  local fadeInAction = AHSpawn.new(
    self.buttonBar:createFadeInAction(),
    AHOpacityTo.new(self.instructionsLabel, EggcornsCommon.UI_FADE_DURATION, nil, EggcornsCommon.VISIBLE_OPACITY)
  )

  fadeInAction:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, buttonBar, progressBar)
  local background = EggcornsBackground.new()
  self:insertAtCenter(background)
  self.background = background

  local elementSpacing = UI_COLUMN_MARGIN_Y
  local alignment = Alignment.CENTER
  local growDirection = Column.GrowUp
  local gameUIColumn = Column.new(elementSpacing, alignment, growDirection)

  gameUIColumn:insertSpacer(EMPTY_HEIGHT)
  gameUIColumn:insert(buttonBar)
  gameUIColumn:insert(progressBar)

  gameUIColumn:alignChildren()

  gameUIColumn:setAnchorPointBottom()
  gameUIColumn:setLoc(Screen.getSafeX(0.5), Screen.safeBottom)
  self:insert(gameUIColumn)

  local rationaleOverlay = EggcornsRationaleOverlay.new()
  self:insertAtCenter(rationaleOverlay)
  self.rationaleOverlay = rationaleOverlay

  local instructionText = EggcornsConfiguration.getInstructionText()
  local instructionsLabel = Label.new(instructionText, Font.new(FontFaces.DIN_OT_BOLD, 13))
  instructionsLabel:setColor(unpack(EggcornsCommon.COLORS.AQUA))
  instructionsLabel:setLoc(INSTRUCTION_LABEL_X, Screen.safeBottom + progressBar:getY() + INSTRUCTIONS_LABEL_OFFSET_Y)
  self:insert(instructionsLabel)
  self.instructionsLabel = instructionsLabel
end

function EggcornsScene.new(buttonBar, progressBar)
  local self = Group.newFullScreen()
  Mixin.mix(self, EggcornsScene)

  createLayout(self, buttonBar, progressBar)

  self.buttonBar = buttonBar
  self.progressBar = progressBar
  self.backings = {}

  return self
end
