--- EggcornsTextView
-- @author Created by Jay Austria on August 30, 2015.

class("EggcornsTextView")

require("EggcornsCommon")
require("EggcornsTextAnimationBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ADDITIONAL_MARGIN_X = 4

local PUNCTUATION_SET = {
  ['.'] = true,
  ['?'] = true,
  [','] = true,
  ['!'] = true,
  [';'] = true,
  [':'] = true,
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBaseAttributes()
  return {
    font = EggcornsCommon.EGGCORNS_TEXT_VIEW_FONT,
    size = EggcornsCommon.FONT_SIZE,
    spacing = MathUtils.round(0.4 * EggcornsCommon.FONT_SIZE),
    color = EggcornsCommon.COLORS.WHITE
  }
end

local function getAttributesForLabels()
  return {
    default = {color = EggcornsCommon.COLORS.WHITE },
    i = {font = EggcornsCommon.EGGCORNS_TEXT_VIEW_FONT_ITALIC},
    leftMargin = {itemType = "margin", spacing = 0, skipNextSpacing = true, allowBreakLineAfter = false},
    rightMargin = {itemType = "margin", spacing = 0, skipPreviousSpacing = true, allowBreakLineBefore = false},
    replacement = {hasBacking = true, isReplacement = true},
    replacementBlack = {hasBacking = true, isReplacement = true, color = EggcornsCommon.COLORS.BLACK}
  }
end

local function getItemStartIndexAndCountForReplacement(self)
  local model = self:getModel()
  local startIndex = nil
  local count = 0

  for index, item in ipairs(model:getItems()) do
    local isMargin = (item:getAttribute("itemType") == "margin")
    if (item:getAttribute("isReplacement") or isMargin) then
      startIndex = startIndex or index
      count = count + 1
    elseif (startIndex ~= nil) then
      break
    end
  end

  return startIndex, count
end

local function getTrailingPunctuationFromReplacementText(self)
  local model = self:getModel()
  local startIndex = nil

  for index, item in ipairs(model:getItems()) do
    if item:getAttribute("isReplacement") then
      startIndex = startIndex or index
    elseif (startIndex ~= nil) then
      local isPunctuation = PUNCTUATION_SET[item:getText()]
      if isPunctuation then
        return item:getText()
      end
    end
  end

  return ""
end

local function getFormattedTextForMorph(text, includeMargins, trailingPunctuation)
  local formattedText = "<replacementBlack>" .. text .. "</replacementBlack>"
  if includeMargins then
    formattedText = "<leftMargin/>" .. formattedText .. "<rightMargin/>"
  elseif (trailingPunctuation ~= "") then
    -- Trailing punctuation is added a seperate item to avoid changes to linebreaks.
    formattedText = formattedText .. "<replacementBlack>" .. trailingPunctuation .. "</replacementBlack>"
  end

  return formattedText
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function EggcornsTextView:prepareToAnimateIn()
  local startingYsForGlyphs = {}

  local characterSizeOffset = 10 -- keep rotated characters out of view
  local offscreenOffsetY = Screen.fullHeight - self:getYBottom() + characterSizeOffset

  for _, view in ipairs(self.itemViews) do
    if view.layout then
      local rows = view.layout:getRows()
      for _, row in ipairs(rows) do
        for _, glyph in ipairs(row) do
          local startX, startY = glyph:getLoc()
          startingYsForGlyphs[glyph] = startY

          local offscreenY = startY + offscreenOffsetY
          glyph:setLoc(startX, offscreenY)
        end
      end
    end
  end

  self.startingYsForGlyphs = startingYsForGlyphs
end

function EggcornsTextView:createAnimateInAction()
  local maxDelay = 0.3
  local duration = EggcornsCommon.FALL_DURATION
  local ease = AHEaseExponentialOut.new()

  local action = AHSpawn.new()
  for glyph, startingY in pairs(self.startingYsForGlyphs) do
    local delayDuration = math.random() * maxDelay
    action:addAction(AHDelayed.new(delayDuration, AHSpawn.new(
      AHLocationYTo.new(glyph, duration, ease, startingY)
    )))
  end

  return action
end

function EggcornsTextView:createDropAction()
  local maxDelay = 0.3
  local duration = 0.7
  local ease = AHEaseExponentialIn.new()

  local action = AHSpawn.new()
  for glyph, startingY in pairs(self.startingYsForGlyphs) do
    local randomRotation = MathUtils.getRandomFloat(-90, 90)
    local delayDuration = math.random() * maxDelay
    action:addAction(AHDelayed.new(delayDuration, AHSpawn.new(
      AHLocationYBy.new(glyph, duration, ease, -Screen.fullHeight),
      AHRotationTo.new(glyph, 0.4, nil, randomRotation),
      function()
        glyph:setColor(unpack(EggcornsCommon.COLORS.WHITE))
        glyph.parent:clearColorAttrLinkToChildren()
      end
    )))
  end

  return action
end

function EggcornsTextView:createItemView(item)
  local itemType = item:getItemType()

  local prop = nil
  if (itemType == "margin") then
    local marginPadding = EggcornsCommon.BACKING_PADDING_X + ADDITIONAL_MARGIN_X
    prop = Group.new(marginPadding, 1)
  else
    prop = self.super.createItemView(self, item)
  end

  return prop
end

function EggcornsTextView:setText(text)
  if text ~= self:getText() then
    local textInterpreter = self.textInterpreter
    local textModel = self.textModel

    local items = textInterpreter:interpretXmlText(text)
    textModel:resetItems(items)

    self.text = text
  end
end

function EggcornsTextView:getText()
  return self.text
end

function EggcornsTextView:replaceWithCorrected(correctedText, includeMargins)
  local model = self:getModel()

  local startIndex, endIndex = getItemStartIndexAndCountForReplacement(self)

  -- If we don't add margins, we also have to replace the trailing punctation,
  -- as it will not get pushed outside of the highlight.
  local includeTrailingPunctuation = not includeMargins
  local trailingPunctuation = ""
  if includeTrailingPunctuation then
    trailingPunctuation = getTrailingPunctuationFromReplacementText(self)
    if trailingPunctuation ~= "" then
      endIndex = endIndex + 1
    end
  end

  local formattedText = getFormattedTextForMorph(correctedText, includeMargins, trailingPunctuation)
  local newItems = self.textInterpreter:interpretXmlText(formattedText)

  -- Note: This triggers animations to start
  model:replaceItems(startIndex, endIndex, newItems)
end

function EggcornsTextView:createBounceAction()
  local bounceAction = AHSequence.new(
    AHScaleTo.new(self, 0.08, AHEaseSineInOut.new(), 0.9),
    AHScaleTo.new(self, 0.13, AHEaseSineInOut.new(), 1.05),
    AHScaleTo.new(self, 0.23, AHEaseSineInOut.new(), 0.95),
    AHScaleTo.new(self, 0.38, AHEaseSineInOut.new(), 1.0)
  )

  return bounceAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createElements(self)
  local baseAttributes = getBaseAttributes()
  local attributesForLabels = getAttributesForLabels()
  local textInterpreter = TextInterpreter.new(baseAttributes, attributesForLabels)
  self.textInterpreter = textInterpreter

  local textModel = TextModel.new(textInterpreter:interpretXmlText(""))
  self:setModel(textModel)

  self.animationBuilder = EggcornsTextAnimationBuilder.new()
end

function EggcornsTextView.new(textWidth)
  local lineSpacing = 5
  local self = TextView.new(textWidth, nil, nil, lineSpacing)
  Mixin.mix(self, EggcornsTextView)

  createElements(self)

  self.startingYsForGlyphs = {}

  return self
end
