--- ExerciseWPMSlider
-- @author Created by Aaron Barrett on May 13, 2015.

class("ExerciseWPMSlider")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.getHeightPart(0.15)

local MARGIN_WIDTH = Screen.getWidthPart(0.14)
local FILLABLE_BAR_WIDTH = Screen.width - MARGIN_WIDTH * 2.0
local SLIDING_GROUP_HEIGHT = 50.0

local BAR_HEIGHT = 2.0
local INNER_CIRCLE_RADIUS = 10
local OUTER_CIRCLE_RADIUS = 18
local OUTER_CIRCLE_OPACITY = 0.3

local FILLING_COLOR = {0.23, 0.77, 0.98}
local EMPTY_COLOR = {0.95, 0.95, 0.95}

local Z_BAR = 10
local Z_FILLING = 20
local Z_SLIDING_GROUP = 30
local Z_TOUCH_AREA = 40

local TIP_TEXT_ATTRIBUTES = {
  default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_SMALL, CardCarouselCardCommon.COLORS.WHITE)
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateValue(self, event)
  local x = event.x
  local normalizedValue = (x - MARGIN_WIDTH) / FILLABLE_BAR_WIDTH
  self:setNormalizedValue(normalizedValue)
end

local function getCirclesScaleAction(self, scale, duration)
  local innerCircle = self.innerCircle
  local outerCircle = self.outerCircle

  local scaleAction = AHSpawn.new(
                        AHScaleTo.new(innerCircle, duration, AHEaseSimpleInOut.new(), scale),
                        AHScaleTo.new(outerCircle, duration, AHEaseSimpleInOut.new(), scale)
                      )

  return scaleAction
end

local function getTipBoxFadeAction(self, opacity, duration)
  local tipBox = self.tipBox

  local fadeAction = AHOpacityTo.new(tipBox, duration, nil, opacity)

  return fadeAction
end

local function getCounterSlideAction(self, offset, duration)
  local counterRow = self.counterRow

  local baseY = self.rowBaseY
  local slideAction = AHLocationYTo.new(counterRow, duration, AHEaseSimpleInOut.new(), baseY + offset)

  return slideAction
end

local function animateTouchBegan(self)
  local circlesAction = getCirclesScaleAction(self, 1.5, 0.05)
  local rowAction = getCounterSlideAction(self, 14, 0.05)

  local combinedAction = AHSpawn.new(circlesAction, rowAction)

  if self.tipBox then
    local tipAction = getTipBoxFadeAction(self, 1.0, 0.1)
    combinedAction:addAction(tipAction)
  end

  self.interactionInterrupt:insert(combinedAction)
end

local function animateTouchEnded(self)
  local circlesAction = getCirclesScaleAction(self, 1.0, 0.3)
  local rowAction = getCounterSlideAction(self, 0, 0.3)

  local combinedAction = AHSpawn.new(circlesAction, rowAction)

  if self.tipBox then
    local tipAction = getTipBoxFadeAction(self, 0.0, 0.3)
    combinedAction:addAction(tipAction)
  end

  self.interactionInterrupt:insert(combinedAction)
end

local function createTipBox(self, tipText)
  local boxWidth = self:getWidthPart(0.8)
  local boxPadding = 15

  local tipLabel = CardCarouselTextView.new(tipText, boxWidth - 2 * boxPadding, TIP_TEXT_ATTRIBUTES, Alignment.CENTER)

  local cornerRadius = 4
  local boxOffsetY = 36
  local box = FilledRoundedRectangle.new(boxWidth, tipLabel:getHeight() + 2 * boxPadding, cornerRadius, FILLING_COLOR)
  box:setLoc(self:getWidthPart(0.5), self:getHeight() + box:getHeightPart(0.5) + boxOffsetY)
  self:insert(box)

  box:insertAtCenter(tipLabel)

  local isRecursive = true
  box:linkOpacityAttrToChildren(isRecursive)
  box:setOpacity(0)

  return box
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseWPMSlider:reset()
  -- This element normally removes itself during the animation that creates it
  -- but exiting the card may interrupt that animation. Remove it if it is stuck.
  if self.explodingCircle then
    self.explodingCircle:removeSelf()
  end
end

function ExerciseWPMSlider:prepareForAnimateIn()
  local filling = self.filling
  filling:setColor(unpack(EMPTY_COLOR))

  local innerCircle = self.innerCircle
  innerCircle:setColor(unpack(EMPTY_COLOR))

  local outerCircle = self.outerCircle
  outerCircle:setScl(0.0)

  self.counterRow:setOpacity(0.0)
end

function ExerciseWPMSlider:getAnimateInAction()
  local filling = self.filling
  local innerCircle = self.innerCircle
  local outerCircle = self.outerCircle

  local fadeDuration = 0.5

  local animateIn = AHSpawn.new(
                      AHColorTo.new(filling, fadeDuration, nil, FILLING_COLOR),
                      AHColorTo.new(innerCircle, fadeDuration, nil, FILLING_COLOR),
                      AHDelayed.new(0.2, AHScaleTo.new(outerCircle, fadeDuration, nil, 1.0))
                    )

  local worldX, worldY = innerCircle:getCenterWorldLoc()
  local radius = math.max(MathUtils.distance(worldX, worldY, 0, Screen.height),
                          MathUtils.distance(worldX, worldY, Screen.width, Screen.height))
  local circle = FilledCircle.new(radius, FILLING_COLOR)
  self.explodingCircle = circle

  local presentGiantCircle = function()
    innerCircle:insertAtCenter(circle)
    circle:setScl(0.0)
  end

  local circleAction = AHSpawn.new(
                         AHScaleTo.new(circle, 2.0, AHEaseExponentialInOut.new(), 1.0),
                         AHDelayed.new(1.0, AHOpacityTo.new(circle, 0.5, nil, 0.0))
                       )

  local counterAction = AHDelayed.new(1.1, AHOpacityTo.new(self.counterRow, 0.3, nil, 1.0))

  local animateAction = AHSpawn.new(animateIn, counterAction, AHDelayed.new(0.1, circleAction))

  return AHSequence.new(presentGiantCircle, animateAction, AHRemove.new(circle))
end

function ExerciseWPMSlider:setValue(value)
  local minValue = self.minValue
  local maxValue = self.maxValue

  local normalizedValue = (value - minValue) / (maxValue - minValue)
  self:setNormalizedValue(normalizedValue)
end

function ExerciseWPMSlider:setNormalizedValue(normalizedValue)
  local minValue = self.minValue
  local maxValue = self.maxValue
  local filling = self.filling
  local slidingGroup = self.slidingGroup
  local counter = self.counter

  normalizedValue = MathUtils.clamp(normalizedValue, 0.0, 1.0)
  local value = minValue + normalizedValue * (maxValue - minValue)
  local roundedValue = MathUtils.round(value)
  counter:setText(roundedValue)

  local slidingGroupX = MARGIN_WIDTH + normalizedValue * FILLABLE_BAR_WIDTH

  local fillingScale = slidingGroupX / WIDTH
  filling:setScl(fillingScale, 1.0)

  slidingGroup:setX(slidingGroupX)

  self.value = value
end

function ExerciseWPMSlider:getValue()
  return self.value
end

function ExerciseWPMSlider:createTip()
  local tipText = "Choosing a speed higher than your comfort level will help increase reading speed."
  self.tipBox = createTipBox(self, tipText)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ExerciseWPMSlider.new(minValue, maxValue)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, ExerciseWPMSlider)

  self.maxValue = maxValue
  self.minValue = minValue

  local bar = FilledRectangle.new(WIDTH, BAR_HEIGHT, EMPTY_COLOR)
  self:insertAtCenter(bar, Z_BAR)

  local filling = FilledRectangle.new(WIDTH, BAR_HEIGHT, FILLING_COLOR)
  filling:setAnchorPointLeft()
  filling:setLoc(0.0, self:getHeightPart(0.5))
  self:insert(filling, Z_FILLING)
  self.filling = filling

  local outerCircle = FilledCircle.new(OUTER_CIRCLE_RADIUS, FILLING_COLOR)
  outerCircle:setOpacity(OUTER_CIRCLE_OPACITY)
  self.outerCircle = outerCircle

  local innerCircle = FilledCircle.new(INNER_CIRCLE_RADIUS, FILLING_COLOR)
  self.innerCircle = innerCircle

  local spacing = 4.0
  local row = Row.new(spacing)

  local counterFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 18, "WPM0123456789")
  local counter = Label.new("000", counterFont)
  counter:setColor(unpack(FILLING_COLOR))
  row:insert(counter)
  self.counter = counter

  local wpmLabel = Label.new("WPM", counterFont)
  wpmLabel:setColor(unpack(FILLING_COLOR))
  row:insert(wpmLabel)

  row:alignChildren()
  row:linkOpacityAttrToChildren()

  self.counterRow = row

  local slidingGroup = Group.new(row:getWidth(), SLIDING_GROUP_HEIGHT)
  local rowOffsetY = 20
  self.rowBaseY = SLIDING_GROUP_HEIGHT - row:getHeightPart(0.5) + rowOffsetY
  row:setLoc(row:getWidthPart(0.5), self.rowBaseY)
  slidingGroup:insert(row)

  slidingGroup:insertAtCenter(outerCircle)
  slidingGroup:insertAtCenter(innerCircle)

  self:insertAtCenter(slidingGroup, Z_SLIDING_GROUP)
  self.slidingGroup = slidingGroup

  local defaltValue = (minValue + maxValue) * 0.15
  self:setValue(defaltValue)

  local touchArea = Group.new(self:getDims())
  self:insertAtCenter(touchArea, Z_TOUCH_AREA)

  TouchManager.addTouchEnabledMethods(touchArea, function(event)
    local phase = event.phase

    if (phase == TouchManager.TOUCH_BEGAN) then
      updateValue(self, event)
      animateTouchBegan(self)
    elseif (phase == TouchManager.TOUCH_MOVED) then
      updateValue(self, event)
    elseif (phase == TouchManager.TOUCH_ENDED) then
      animateTouchEnded(self)
    end
  end)

  FunctionUtils.forwardMethodCalls(self, touchArea, "setTouchEnabled")

  self.interactionInterrupt = ActionInterrupt.new()

  return self
end
