--- ExerciseCardFlashCard
-- @author Created by karl on May 14, 2015.

class("ExerciseCardFlashCard")

require("cards.ExerciseCard")
require("containers.card_carousel.CardCarouselCardCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local totalCards = 0
local BACKING_COLOR = ColorUtils.colorHex24ToStd(0xeaeaea)
local DROP_SHADOW_COLOR = ColorUtils.colorHex24ToStd(0xcccccc)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseCardFlashCard:reset()
  self.cardIndexLabel:setText(tostring(self.cardIndex) .. " of " .. tostring(totalCards))
  self.frontCard:setOpacity(1)
  self.backCard:setOpacity(0)
end

function ExerciseCardFlashCard:onEnter()
  self.frontCard:setTouchEnabled(true)
  self.super.onEnter(self)
end

function ExerciseCardFlashCard:onLeaveComplete()
  self.frontCard:setTouchEnabled(false)
  self.backCard:setTouchEnabled(false)
  self.super.onLeaveComplete(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newFrontCard(width, height, data, showUsabilityMessage)
  local card = Group.new(width, height)
  local word = data.word:upper()
  local fontSize = 63
  if #word > 8 and string.find(word, " ") == nil then
    fontSize = 45
  end

  local label = CardCarouselCardCommon.newMultilineLabel(word,
                                                     FontFaces.DIN_OT_BOLD,
                                                     fontSize,
                                                     Screen.width,
                                                     Alignment.CENTER,
                                                     CardCarouselCardCommon.COLORS.GREY)
  card:insertAtCenter(label)

  if showUsabilityMessage then
    local messageAttrbutes = {
      default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_SMALL, CardCarouselCardCommon.COLORS.LIGHT_GREY)
    }

    local usabilityLabel = CardCarouselTextView.new("TAP TO FLIP", card:getWidth(), messageAttrbutes, Alignment.CENTER)
    usabilityLabel:setLoc(card:getWidthPart(0.5), card:getHeightPart(0.05))
    card:insert(usabilityLabel)
  end

  card:linkOpacityAttrToChildren(true)

  return card
end

local function newBackCard(width, height, data)
  local alignment = Alignment.CENTER

  local concept = data.concept
  local card = Group.new(width, height)
  local col = Column.new(0)
  local labelPadding = CardCarouselCardCommon.MARGINS.HORIZONTAL * 2

  local wordDefinitionAttrbutes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_LARGE, CardCarouselCardCommon.COLORS.DARK_GREY)
  }

  local exampleAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM, CardCarouselCardCommon.COLORS.LIGHT_GREY)
  }

  local exampleDefinitionColor = ColorUtils.colorHex24ToStd(0xbababa)
  local exampleDefinitionAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM, exampleDefinitionColor)
  }

  local sectionSpaceY = 14
  local labelWidth = width - labelPadding

  local image = nil
  local imageScale = 1
  if data.image then
    image = AssetHelper.createPhotoSpriteForConcept(concept, data.image)
    local imageW = image:getWidth()
    if imageW > width then
      imageScale = width / imageW
      image:setScl(imageScale)
    end
    col:insert(image)
  end

  if data.wordDefinition then
    col:insert(CardCarouselTextView.new(data.wordDefinition, labelWidth, wordDefinitionAttrbutes, alignment))
    col:insertSpacer(sectionSpaceY)
  end
  if data.example1 then
    col:insert(CardCarouselTextView.new(data.example1, labelWidth, exampleAttributes, alignment))
    if data.exampleDefinition1 then
      col:insert(CardCarouselTextView.new(data.exampleDefinition1, labelWidth, exampleDefinitionAttributes, alignment))
    end
    if data.example2 then
      col:insertSpacer(sectionSpaceY)
      col:insert(CardCarouselTextView.new(data.example2, labelWidth, exampleAttributes, alignment))
      if data.exampleDefinition2 then
        col:insert(CardCarouselTextView.new(data.exampleDefinition2, labelWidth, exampleDefinitionAttributes, alignment))
      end
    end
  end

  col:alignChildren()
  card:insertAtCenter(col)
  card:linkOpacityAttrToChildren(true)

  local targetHeight = height * 0.95
  if col:getHeight() > targetHeight and image then
    local overH = col:getHeight() - targetHeight
    local imageH = image:getHeight()
    local scale = imageH / (imageH + overH)
    if scale < imageScale then
      image:setScl(scale)
      col:alignChildren()
    end
  end

  return card
end

local function getFlipSequence(group, targetOpacity)
  local flipDuration = 0.5

  local scaleDown = AHScaleYTo.new(group, flipDuration/2, AHEaseQuadraticInOut.new(), 0.0)
  local setTargetOpacity = function() group:setOpacity(targetOpacity) end
  local scaleUp = AHScaleYTo.new(group, flipDuration/2, AHEaseQuadraticInOut.new(), 1.0)

  return AHSequence.new(scaleDown, setTargetOpacity, scaleUp)
end

local function addFlipTouchHandler(self, srcGroup, dstGroup)
  local onTouchDown = nil

  local onTouchUp = function()
    AHSequence.new(
      function() srcGroup:setTouchEnabled(false) end,
      AHSpawn.new(
        getFlipSequence(srcGroup, 0.0),
        getFlipSequence(dstGroup, 1.0),
        getFlipSequence(self.workArea, 1.0)
      ),
      function() dstGroup:setTouchEnabled(true) end
    ):start()
  end

  local onTouchCancel = nil

  local buttonWidthReduced = Screen.safeWidth - (2 * CardCarouselNavigator.TAP_NAVIGATE_LEFT_BOUNDARY)
  local buttonHeight = srcGroup:getHeight()
  Button.addVirtualButton(srcGroup, buttonWidthReduced, buttonHeight, onTouchDown, onTouchUp, onTouchCancel)
end

local function createLayout(self, data)
  totalCards = totalCards + 1
  self.cardIndex = totalCards

  local backing = CardCarouselCardCommon.newStandardBackground(BACKING_COLOR)
  self:insertAtCenter(backing)

  local dropShadowPaddingX = 2
  local dropShadowPaddingY = 4
  local cardEdgePaddingX = 16
  local cardEdgePaddingY = cardEdgePaddingX * 0.5

  local workWidth = self:getWidth() - dropShadowPaddingX * 2 - cardEdgePaddingX
  local workHeight = self:getHeightPart(0.85)
  local dropShadowWidth = self:getWidth() - cardEdgePaddingX
  local dropShadowHeight = workHeight + dropShadowPaddingY
  local cornerRadius = 6

  local dropShadow = FilledRoundedRectangle.new(dropShadowWidth, dropShadowHeight, cornerRadius, DROP_SHADOW_COLOR)
  local cardFace = FilledRoundedRectangle.new(workWidth, workHeight, cornerRadius, CardCarouselCardCommon.COLORS.WHITE)

  local workArea = Group.new(dropShadow:getDims())
  dropShadow:setLoc(workArea:getWidthPart(0.5), workArea:getHeightPart(0.5) - cardEdgePaddingY)
  workArea:insert(dropShadow)
  cardFace:setLoc(workArea:getWidthPart(0.5), workArea:getHeight() - cardFace:getHeightPart(0.5) - cardEdgePaddingY)
  workArea:insert(cardFace)

  workArea:setLoc(self:getWidthPart(0.5), self:getHeight() - workArea:getHeightPart(0.55))
  self:insert(workArea)

  local showUsabilityMessage = (self.cardIndex == 1)

  self.workArea = workArea
  self.frontCard = newFrontCard(workWidth, workHeight, data, showUsabilityMessage)
  self.backCard = newBackCard(workWidth, workHeight, data)

  addFlipTouchHandler(self, self.frontCard, self.backCard)
  self.frontCard:setTouchEnabled(false)
  addFlipTouchHandler(self, self.backCard, self.frontCard)
  self.backCard:setTouchEnabled(false)

  workArea:insertAtCenter(self.backCard)
  workArea:insertAtCenter(self.frontCard)

  self.cardIndexLabel = CardCarouselCardCommon.newCachedFontLabel(" ", CardCarouselCardCommon.CACHED_FONTS.MEDIUM_SMALL, CardCarouselCardCommon.COLORS.LIGHT_GREY)
  local labelY = (self:getHeight() - workArea:getHeight() - dropShadowPaddingX) / 2
  self.cardIndexLabel:setLoc(self:getWidthPart(0.5), labelY)
  self:insert(self.cardIndexLabel)
end

-- Constructor ------------------------

function ExerciseCardFlashCard.new(cardType, cardGlobals, data)
  local self = ExerciseCard.new(cardType, cardGlobals)
  Mixin.mix(self, ExerciseCardFlashCard)

  createLayout(self, data)

  return self
end
