--- ExerciseCardImageExplanation
-- @author Created by karl on May 11, 2015.

require("cards.ExerciseCard")
require("cards.ExerciseCardExplanationCommon")

-- ============================================================================
class("ExerciseCardImageExplanation")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DOT_COUNT = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function extractDotLocation(data, index, image)
  local height = image:getEffectiveHeight()
  local scaleX, scaleY = image:getScl()
  local field = data["location" .. tostring(index)]

  local scaleFactor = AssetHelper.getAssetScaleFactor()

  if field and field ~= "" then
    _, _, x, y = string.find(field, "%s*(%d+)%s*,%s*(%d+)")
    x = tonumber(x) * scaleX / scaleFactor
    y = tonumber(y) * scaleY / scaleFactor
    return x, height - y
  end
end

local function newExplanation(description, examples)
  local descriptionFont = CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM
  local exampleFont = CardCarouselCardCommon.CACHED_FONTS.MEDIUM_SMALL
  local width = CardCarouselCardCommon.CONTENT_WIDTH

  local descriptionAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(descriptionFont, CardCarouselCardCommon.COLORS.BLUE)
  }

  local examplesAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(exampleFont, CardCarouselCardCommon.COLORS.LIGHT_GREY)
  }

  local col = Column.new(0, Alignment.LEFT)
  col:insert(CardCarouselTextView.new(description, width, descriptionAttributes))
  col:insertSpacer(10)
  col:insert(CardCarouselTextView.new(examples, width, examplesAttributes))
  col:alignChildren()
  col:linkOpacityAttrToChildren(true)
  SlidingGroup.makeSlidingGroup(col)

  return col
end

local function updateSlideDistance(self)
  local _, containerTop = self.workArea:transformLoc(self.explanationContainer, 0, self.explanationContainer:getHeight())
  self.slideDistance = containerTop
end

local function onDotTouched(self, dot, explanationGroup)
  if dot ~= self.lastDotTouched then
    for _, group in ipairs(self.dots) do
      group:setActive(false)
    end
    dot:setActive(true)

    updateSlideDistance(self)

    explanationGroup:setLoc(0, self.explanationContainer:getHeight() + explanationGroup:getHeight())

    local lastAction
    for _, group in ipairs(self.explanationContainer.children) do
      local willBeActive = group == explanationGroup
      lastAction = group:slideDown(group:getHeight(), willBeActive)

      if willBeActive and group:getHeight() > self.slideDistance then
        local slideUpDistance = group:getHeight() - self.slideDistance;
        lastAction = self.col:slideDown(-slideUpDistance, true)
      end
    end
    wait(lastAction)

    updateSlideDistance(self)

    for _, group in ipairs(self.explanationContainer.children) do
      if group ~= explanationGroup then
        group:setLoc(0, self.explanationContainer:getHeight() + self.slideDistance)
        group:reset()
      end
    end
    self.lastDotTouched = dot
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseCardImageExplanation:createAnimationAction()
  local action = AHSpawn.new()
  for _, dot in ipairs(self.dots) do
    action:addAction(dot:newLoopAction())
  end

  return action
end

function ExerciseCardImageExplanation:setTouchEnabled(isEnabled)
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(isEnabled)
  end
end

function ExerciseCardImageExplanation:reset()
  for _, dot in ipairs(self.dots) do
    dot:reset()
  end

  for _, group in ipairs(self.explanationContainer.children) do
    group:setLoc(0, self.explanationContainer:getHeight())
    local isActive = group == self.tapLabel
    group:reset(isActive)
  end

  self.lastDotTouched = nil
end

function ExerciseCardImageExplanation:onEnter()
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(true)
  end
  self.super.onEnter(self)
end

function ExerciseCardImageExplanation:onLeaveComplete()
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(false)
  end
  self.super.onLeaveComplete(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newImage(concept, filename)
  return AssetHelper.createPhotoSpriteForConcept(concept, filename)
end

-- Constructor ------------------------

function ExerciseCardImageExplanation.new(cardType, cardGlobals, data)
  local self = ExerciseCard.new(cardType, cardGlobals)
  Mixin.mix(self, ExerciseCardImageExplanation)

  self:insertAtCenter(CardCarouselCardCommon.newStandardBackground())
  local workArea = CardCarouselCardCommon.newStandardWorkArea()
  self:insertAtCenter(workArea)
  self.workArea = workArea

  local col = Column.new(0, Alignment.LEFT)
  SlidingGroup.makeSlidingGroup(col)
  self.col = col

  local image = newImage(data.concept, data.image)
  image:setEffectiveDims(workArea:getWidth())
  local imageContainer = Group.new(image:getEffectiveDims())
  imageContainer:insertAtCenter(image)
  col:insert(imageContainer)
  col:insertSpacer(self:getHeightPart(0.05))

  local descriptionAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM, CardCarouselCardCommon.COLORS.GREY)
  }

  col:insert(CardCarouselTextView.new(data.overallDescription, workArea:getWidth(), descriptionAttributes))
  col:insertSpacer(self:getHeightPart(0.05))

  col:insert(CardCarouselCardCommon.newSeparatorLine())
  col:insertSpacer(self:getHeightPart(0.05))

  self.tapLabel = CardCarouselCardCommon.newLabel("TAP ONE OF THE DOTS TO LEARN MORE", FontFaces.DIN_OT_MEDIUM, 14, CardCarouselCardCommon.COLORS.BLUE)
  SlidingGroup.makeSlidingGroup(self.tapLabel)
  self.tapLabel:reset(true)
  col:insert(self.tapLabel)

  col:alignChildren()

  local containerX, containerY = self.tapLabel:getLocationWithAnchorPoint(0, 1)
  self.explanationContainer = Group.new(workArea:getDims())
  self.explanationContainer:setAnchorPointTopLeft()
  self.explanationContainer:setLoc(containerX, containerY)
  col:insert(self.explanationContainer)
  self.tapLabel:removeSelf(false)
  self.tapLabel:setAnchorPointTopLeft()
  self.tapLabel:setLoc(0, self.explanationContainer:getHeight())
  self.explanationContainer:insert(self.tapLabel)

  self.dots = {}

  for i = 1, DOT_COUNT do
    local x, y = extractDotLocation(data, i, image)
    if x and y then
      local description = data["description" .. tostring(i)]
      local examples = data["examples" .. tostring(i)]
      local explanationGroup = newExplanation(description, examples)
      explanationGroup:setAnchorPointTopLeft()
      self.explanationContainer:insert(explanationGroup)
      explanationGroup:reset()
      local dot
      dot = AnimatedDot.new(AnimatedDot.DIRECTIONS.RIGHT, function() onDotTouched(self, dot, explanationGroup) end)
      dot:setLoc(x, y)
      imageContainer:insert(dot)
      table.insert(self.dots, dot)
    end
  end

  col:setAnchorPointLeft()
  col:setLoc(0, workArea:getHeightPart(0.5))
  workArea:insert(col)

  return self
end
