--- ExerciseCardMultipleChoice
-- @author Created by Jay Austria on April 30, 2015.

class("ExerciseCardMultipleChoice")

require("cards.ExerciseCard")
require("containers.card_carousel.CardCarouselCardCommon")
require("ExerciseMultipleChoiceButton")
require("containers.card_carousel.CardCarouselNavigator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LINE_OFFSET_X = Screen.getWidthPart(0.25)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setButtonsTouchEnabled(self, enabled)
  self.scrollEnabled = enabled
  local buttons = self.optionButtons
  for _, button in ipairs(buttons) do
    button:setTouchEnabled(enabled)
  end
end


local function getViewLines(textView)
  local lines = {}
  local currentLine = {}
  local firstView = textView.itemViews[1]
  local currentY = firstView:getY()
  for _, view in ipairs(textView.itemViews) do
    local viewY = view:getY()

    if viewY ~= currentY then
      table.insert(lines, currentLine)
      currentLine = {}
      currentY = viewY
    end

    table.insert(currentLine, view)
  end

  table.insert(lines, currentLine)

  return lines
end

local function createLayout(self)
  local background = CardCarouselCardCommon.newStandardBackground()
  self:insertAtCenter(background)

  local prompt = self.prompt
  local promptAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.LIGHT_MEDIUM, CardCarouselCardCommon.COLORS.GREY)
  }
  local alignment = Alignment.CENTER
  local promptLabel = CardCarouselTextView.new(prompt, CardCarouselCardCommon.CONTENT_WIDTH, promptAttributes, alignment)
  promptLabel:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.75))
  self:insert(promptLabel)

  local lines = getViewLines(promptLabel)

  for lineIndex, line in ipairs(lines) do
    local offsetX = LINE_OFFSET_X * (lineIndex - 1)
    for _, view in ipairs(line) do
      view.originX = view:getX()
    end
  end

  local options = self.options
  local optionButtons = {}
  local currentY = ExerciseMultipleChoiceButton.BUTTON_AREA_HEIGHT - (ExerciseMultipleChoiceButton.BUTTON_HEIGHT * 0.5)
  local offsetY = -(ExerciseMultipleChoiceButton.BUTTON_HEIGHT + ExerciseMultipleChoiceButton.BUTTON_MARGIN_Y)

  for _, option in ipairs(options) do
    local optionButton = ExerciseMultipleChoiceButton.new(option)

    local function onButtonPressedCallback(buttonText)
      runInCoroutine(function()
        setButtonsTouchEnabled(self, false)

        local cardGlobals = self:getCardGlobals()
        local questionsCount = cardGlobals.multipleChoiceQuestionsCount or 0
        local correctCount = cardGlobals.multipleChoiceCorrectCount or 0

        questionsCount = questionsCount + 1

        local wasAnswerCorrect = (buttonText == self.correctAnswer)
        if (wasAnswerCorrect) then
          correctCount = correctCount + 1
        end

        cardGlobals.multipleChoiceQuestionsCount = questionsCount
        cardGlobals.multipleChoiceCorrectCount = correctCount

        wait(optionButton:playPressed())
        self:forceScrollNavigation(CardCarouselNavigator.DIRECTION_LEFT, CardCarouselNavigator.SCROLL_ANIMATION_DURATION_HORIZONTAL)
      end)
    end

    optionButton:setButtonCallback(onButtonPressedCallback)
    optionButton:setOriginX(self:getWidthPart(0.5))
    optionButton:setLoc(optionButton.originX, currentY)
    table.insert(optionButtons, optionButton)
    self:insert(optionButton)

    currentY = currentY + offsetY
  end

  self.lines = lines
  self.promptLabel = promptLabel
  self.optionButtons = optionButtons

  setButtonsTouchEnabled(self, false)
end

local function insertOptionIntoTableIfValid(optionTable, option)
  if option and option ~= "" then
    table.insert(optionTable, option)
  end
end

local function initializeCard(self, data)
  local options = {}
  insertOptionIntoTableIfValid(options, data.correctAnswer)
  insertOptionIntoTableIfValid(options, data.distractor1)
  insertOptionIntoTableIfValid(options, data.distractor2)
  insertOptionIntoTableIfValid(options, data.distractor3)

  TableUtils.shuffleArray(options)

  self.prompt = data.prompt
  self.correctAnswer = data.correctAnswer
  self.options = options
  self.scrollEnabled = true
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseCardMultipleChoice:createAnimationAction()
  local arriveAction = AHSpawn.new()
  for buttonIndex, button in ipairs(self.optionButtons) do
    local delayDuration = (buttonIndex - 1) * 0.1
    local centerAction = button:getEnterAnimation()
    arriveAction:addAction(AHDelayed.new(delayDuration, centerAction))
  end

  local animationAction = AHSequence.new(
                                         arriveAction,
                                         function()
                                           setButtonsTouchEnabled(self, true)
                                         end
                                         )

  return animationAction
end

function ExerciseCardMultipleChoice:isScrollAwayAllowed(scrollDirection)
  if ((scrollDirection == CardCarouselNavigator.DIRECTION_LEFT) or
      (scrollDirection == CardCarouselNavigator.DIRECTION_RIGHT)) then
    return false
  else
    return self.scrollEnabled
  end
end

function ExerciseCardMultipleChoice:isTapNavigateAwayAllowed(tapDirection)
  if ((tapDirection == CardCarouselNavigator.TAP_DIRECTION_FORWARD) or
      (tapDirection == CardCarouselNavigator.TAP_DIRECTION_BACK)) then
    return false
  else
    return self.scrollEnabled
  end
end

function ExerciseCardMultipleChoice:isScrollEnterAllowed(scrollDirection)
  return false
end

function ExerciseCardMultipleChoice:isTapNavigateEnterAllowed(tapDirection)
  return false
end

function ExerciseCardMultipleChoice:onLeaveComplete()
  self.super.onLeaveComplete(self)

  setButtonsTouchEnabled(self, false)
end

function ExerciseCardMultipleChoice:reset()
  local moveToEndLocationSpawn = AHSpawn.new()
  local lines = self.lines

  for lineIndex, line in ipairs(lines) do
    local offsetX = LINE_OFFSET_X * (lineIndex - 1)
    for _, view in ipairs(line) do
      local startX = view.originX + offsetX
      local endX = view.originX

      view:setX(startX)

      local moveToEndLocationAction = AHLocationXTo.new(view, 1.0, nil, endX)
      moveToEndLocationSpawn:addAction(moveToEndLocationAction)
    end
  end

  local untimedAction = AHUntimedAction.new(moveToEndLocationSpawn)
  untimedAction:start()

  self.untimedAction = untimedAction

  local buttons = self.optionButtons
  for buttonIndex, button in ipairs(buttons) do
    local directionParity = 1
    if (buttonIndex % 2 == 0) then
      directionParity = -1
    end

    button:prepareForEnterAnimation(directionParity)
  end

  setButtonsTouchEnabled(self, false)
end

function ExerciseCardMultipleChoice:setScrollProgress(progress, isHorizontal)
  self.untimedAction:applyAtTime(progress)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ExerciseCardMultipleChoice.new(cardType, cardGlobals, data)
  local self = ExerciseCard.new(cardType, cardGlobals)
  Mixin.mix(self, ExerciseCardMultipleChoice)

  initializeCard(self, data)
  createLayout(self)

  return self
end
