--- ExerciseCardTipExplanation
-- @author Created by karl on May 11, 2015.

require("cards.ExerciseCard")
require("cards.ExerciseCardExplanationCommon")

-- ============================================================================
class("ExerciseCardTipExplanation")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DOT_COUNT = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function extractDotLocation(data, image)
  local height = image:getEffectiveHeight()
  local scaleX, scaleY = image:getScl()
  local field = data.location

  local scaleFactor = AssetHelper.getAssetScaleFactor()

  _, _, x, y = string.find(field, "%s*(%d+)%s*,%s*(%d+)")
  x = tonumber(x) * scaleX / scaleFactor
  y = tonumber(y) * scaleY / scaleFactor
  return x, height - y
end

local function newTipElement(self, tip)
  local tipButtonColor = ColorUtils.colorRGBToStd(205, 205, 205)
  local buttonLabel = CardCarouselCardCommon.newLabel("VIEW TIP", FontFaces.DIN_OT_MEDIUM, 15, tipButtonColor)

  local buttonWidth = buttonLabel:getWidthPart(1.4)
  local buttonHeight = buttonLabel:getHeightPart(1.8)
  local cornerRadius = buttonHeight * 0.1
  local lineWidth = cornerRadius * 0.4
  local tipButton = HollowRoundedRectangle.new(buttonWidth, buttonHeight, cornerRadius, tipButtonColor, lineWidth)
  tipButton:insertAtCenter(buttonLabel)
  tipButton:linkColorAttrToChildren(true)

  local tipLabel = CardCarouselCardCommon.newMultilineLabel(tip,
                                                        FontFaces.DIN_OT_MEDIUM,
                                                        16,
                                                        CardCarouselCardCommon.CONTENT_WIDTH,
                                                        Alignment.LEFT,
                                                        CardCarouselCardCommon.COLORS.GREY)
  local container = Group.new(tipLabel:getDims())
  container:insertAtCenter(tipLabel)
  tipButton:setAnchorPointTopLeft()
  tipButton:setLoc(0, container:getHeight())
  container:insert(tipButton)

  tipLabel:setOpacity(0)
  Button.makeButton(tipButton, function()
    local duration = 0.2
    AHSequence.new(
      AHOpacityTo.new(tipButton, duration, nil, 0),
      AHOpacityTo.new(tipLabel, duration, nil, 1)
    ):start()

    local _, tipBottom = self.workArea:transformLoc(tipLabel, 0, 0)
    if tipBottom < 0 then
      AHLocationYBy.new(self.col, duration, AHEaseSineInOut.new(), -tipBottom):start()
    end
  end)

  container.setTouchEnabled = function(self, isEnabled)
    tipButton:setTouchEnabled(isEnabled)
  end

  container:linkColorAttrToChildren()

  return container
end

local function newExplanation(self, description, tip)
  local descriptionFont = CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM
  local width = CardCarouselCardCommon.CONTENT_WIDTH

  local descriptionAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(descriptionFont, CardCarouselCardCommon.COLORS.BLUE)
  }

  local col = Column.new(0, Alignment.LEFT)
  col:insert(CardCarouselTextView.new(description, width, descriptionAttributes))
  col:insertSpacer(10)
  local tipElement = newTipElement(self, tip)
  col:insert(tipElement)
  col:alignChildren()
  col:linkOpacityAttrToChildren()
  SlidingGroup.makeSlidingGroup(col)

  col.setTouchEnabled = function(self, isEnabled)
    tipElement:setTouchEnabled(isEnabled)
  end

  return col
end

local function onDotTouched(self, dot, explanationGroup)
  if dot ~= self.lastDotTouched then
    for _, group in ipairs(self.dots) do
      group:setActive(false)
    end
    dot:setActive(true)

    explanationGroup:setLoc(0, self.explanationContainer:getHeight() + self.slideDistance)

    local lastAction
    for _, group in ipairs(self.explanationContainer.children) do
      local willBeActive = group == explanationGroup
      lastAction = group:slideDown(self.slideDistance, willBeActive)
    end
    wait(lastAction)

    for _, group in ipairs(self.explanationContainer.children) do
      if group ~= explanationGroup then
        group:setLoc(0, self.explanationContainer:getHeight() + self.slideDistance)
        group:reset()
      end
    end
    self.lastDotTouched = dot
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseCardTipExplanation:createAnimationAction()
  local action = AHSpawn.new()
  for _, dot in ipairs(self.dots) do
    action:addAction(dot:newLoopAction())
  end

  return action
end

function ExerciseCardTipExplanation:setTouchEnabled(isEnabled)
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(isEnabled)
  end
end

function ExerciseCardTipExplanation:reset()
  for _, dot in ipairs(self.dots) do
    dot:reset()
  end

  for _, group in ipairs(self.explanationContainer.children) do
    group:setLoc(0, self.explanationContainer:getHeight())
    local isActive = group == self.tapLabel
    group:reset(isActive)
  end

  self.lastDotTouched = nil
end

function ExerciseCardTipExplanation:onEnter()
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(true)
  end
  for _, explanation in ipairs(self.explanations) do
    explanation:setTouchEnabled(true)
  end
  self.super.onEnter(self)
end

function ExerciseCardTipExplanation:onLeaveComplete()
  for _, dot in ipairs(self.dots) do
    dot:setTouchEnabled(false)
  end
  for _, explanation in ipairs(self.explanations) do
    explanation:setTouchEnabled(false)
  end
  self.super.onLeaveComplete(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newImage(concept, filename)
  return AssetHelper.createPhotoSpriteForConcept(concept, filename)
end

-- Constructor ------------------------

function ExerciseCardTipExplanation.new(cardType, cardGlobals, data)
  local self = ExerciseCard.new(cardType, cardGlobals)
  Mixin.mix(self, ExerciseCardTipExplanation)

  local image = data.image

  self:insertAtCenter(CardCarouselCardCommon.newStandardBackground())
  local workArea = CardCarouselCardCommon.newStandardWorkArea()
  self:insertAtCenter(workArea)
  self.workArea = workArea

  local col = Column.new(0, Alignment.LEFT)
  self.col = col

  local image = newImage(data.concept, image)
  image:setEffectiveDims(workArea:getWidth())
  local imageContainer = Group.new(image:getEffectiveDims())
  imageContainer:insertAtCenter(image)
  col:insert(imageContainer)

  local descriptionAttributes = {
    default = CardCarouselCardCommon.newTextAttribute(CardCarouselCardCommon.CACHED_FONTS.MEDIUM_MEDIUM, CardCarouselCardCommon.COLORS.GREY)
  }

  col:insert(CardCarouselCardCommon.newSeparatorLine())
  col:insertSpacer(self:getHeightPart(0.05))

  self.tapLabel = CardCarouselCardCommon.newLabel("TAP THE DOT TO LEARN MORE", FontFaces.DIN_OT_MEDIUM, 14, CardCarouselCardCommon.COLORS.BLUE)
  SlidingGroup.makeSlidingGroup(self.tapLabel)
  self.tapLabel:reset(true)
  col:insert(self.tapLabel)

  col:alignChildren()

  local containerX, containerY = self.tapLabel:getLocationWithAnchorPoint(0, 1)
  self.explanationContainer = Group.new(workArea:getDims())
  self.explanationContainer:setAnchorPointTopLeft()
  self.explanationContainer:setLoc(containerX, containerY)
  col:insert(self.explanationContainer)
  self.tapLabel:removeSelf(false)
  self.tapLabel:setAnchorPointTopLeft()
  self.tapLabel:setLoc(0, self.explanationContainer:getHeight())
  self.explanationContainer:insert(self.tapLabel)

  self.dots = {}
  local maxHeight = self.tapLabel:getHeight()
  self.explanations = {}

  local x, y = extractDotLocation(data, image)
  local description = data.description
  local tip = data.tip
  local explanationGroup = newExplanation(self, description, tip)
  explanationGroup:setTouchEnabled(false)
  table.insert(self.explanations, explanationGroup)
  explanationGroup:setAnchorPointTopLeft()
  self.explanationContainer:insert(explanationGroup)
  explanationGroup:reset()
  maxHeight = math.max(maxHeight, explanationGroup:getHeight())
  local dot
  dot = AnimatedDot.new(AnimatedDot.DIRECTIONS.RIGHT, function() onDotTouched(self, dot, explanationGroup) end)
  dot:setLoc(x, y)
  imageContainer:insert(dot)
  table.insert(self.dots, dot)

  self.slideDistance = maxHeight

  col:setAnchorPointLeft()
  col:setLoc(0, workArea:getHeightPart(0.53))
  workArea:insert(col)

  return self
end
