--- ExerciseValidator
-- @author Created by karl on April 02, 2015.
-- ConceptValidator for Exercise.

class("ExerciseValidator")

require("validators.ConceptValidator")
require("utils.XmlParser")
require("validators.TagValidator")
require("ExerciseCardDefinitions")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "title",
  "description",
  "cards",
  "filterExerciseIdentifier",
  "categoryId",
  "skillGroupId",
  "requiredSkillGroupProgressLevel",
  "isPro",
  "exerciseOrder",
  "initialSRSStep"
}

local OPTIONAL_XML_TAGS = {
  "i",
  "title",
  "h1",
  "h2",
  "h1highlight",
  "highlight",
}

local OPTIONAL_XML_SINGLE_TAGS = {
  "br",
  "divide"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateAttributes(self, mapping, key)
  self:expectTrue(mapping:isRequired(key) or mapping:isOptional(key), "Field mapping " .. key .. " must be either optional or required.")
  self:expectFalse(mapping:isRequired(key) and mapping:isOptional(key), "Field mapping " .. key .. " cannot be both optional and required.")
  self:expectFalse(mapping:isXml(key) and mapping:isImage(key), "Field mapping " .. key .. " cannot be both xml and image.")
end

local function validateRequiredField(self, card, mapping, key)
  self:expectTrue(mapping:exists(card, key),
                  "The field `" .. tostring(mapping:getForeignKeys(key)[1]) .. "` must not be empty on card with rank `" ..
                  tostring(card.rank) .. "` and file `" .. tostring(card.file) .. "` of card type `" ..
                  tostring(mapping:getValues(card, "cardType")[1]) .. "`.  This key maps to local key `" .. tostring(key) .. "`." )
end

local function validateImageField(self, card, mapping, key)
  for _, value in ipairs(mapping:getValues(card, key)) do
    self:validateImageFile(value)
  end
end


local function validateXmlField(self, card, mapping, key)
  for _, value in ipairs(mapping:getValues(card, key)) do
    XmlParser.parse(value) -- throws errors if it fails
    self:validateTags(value, nil, nil, nil, OPTIONAL_XML_TAGS, OPTIONAL_XML_SINGLE_TAGS)
  end
end

local function validateCards(self, concept)
  local cards = concept.cards
  self:expectType(cards, "table", "cards")

  for _, card in ipairs(cards) do
    local cardType = card["card_type"]
    self:assertIsContained(cardType, ExerciseCardDefinitions.CARD_TYPES, "card_type")

    local mapping = ExerciseCardDefinitions.FIELD_MAPPING_BY_CARD_TYPE[cardType]
    self:assertNotNil(mapping, "mapping", "No field mapping info defined for card type: `" .. tostring(cardType) .. "`")

    for _, key in ipairs(mapping:getAllKeys()) do
      validateAttributes(self, mapping, key)
      self:expectNotNil(mapping:getForeignKeys(key)[1], "Key " .. key .. " has no foreign mapping")

      if mapping:isRequired(key) then
        validateRequiredField(self, card, mapping, key)
      end

      if mapping:exists(card, key) then
        if mapping:isImage(key) then
          validateImageField(self, card, mapping, key)
        end

        if mapping:isXml(key) then
          validateXmlField(self, card, mapping, key)
        end
      end
    end
  end
end

local function validateCategoryIdentifier(self, concept)
  local categoryIdentifier = concept["categoryId"]
  local CATEGORY_IDENTIFIERS = { "public_speaking", "vocab", "speed_reading", "math" }
  self:expectIsContained(categoryIdentifier, CATEGORY_IDENTIFIERS, "categoryId")
end

local function validateSkillGroupIdentifier(self, concept)
  local skillGroupIdentifier = concept["skillGroupId"]
  local SKILL_GROUP_IDENTIFIERS = { "reading", "writing", "listening", "speaking", "math" }
  self:expectIsContained(skillGroupIdentifier, SKILL_GROUP_IDENTIFIERS, "skillGroupId")
end

local function validateSkillGroupProgressLevel(self, concept)
  local skillGroupProgressLevel = concept["requiredSkillGroupProgressLevel"]
  local SKILL_GROUP_PROGRESS_LEVELS = { "novice", "intermediate", "advanced", "expert", "elite", "master" }
  self:expectIsContained(skillGroupProgressLevel, SKILL_GROUP_PROGRESS_LEVELS, "requiredSkillGroupProgressLevel")
end

local function validateIsPro(self, concept)
  local isPro = concept["isPro"]
  local BOOLEAN_STRING_VALUES = { "true", "false" }
  self:expectIsContained(isPro, BOOLEAN_STRING_VALUES, "isPro")
end

local function validateProProgressLevelMatch(self, concept)
  local isPro = concept["isPro"]
  local requiredSkillGroupProgressLevel = concept["requiredSkillGroupProgressLevel"]

  local shouldBeNovice = (isPro == "true")
  if shouldBeNovice then
    local isLevelNovice = (requiredSkillGroupProgressLevel == "novice")
    local errorMessage = string.format("Exercises marked is_pro = true must be novice. Exercise was %s", requiredSkillGroupProgressLevel)
    self:expectTrue(isLevelNovice, errorMessage)
  end
end

local function validateExerciseOrder(self, concept)
  local exerciseOrder = concept["exerciseOrder"]
  self:assertType(exerciseOrder, "number", "exerciseOrder")
  self:expectGreaterThan(exerciseOrder, 0, "exerciseOrder")
end

local function validateInitialSRSStep(self, concept)
  local initialSRSStep = concept["initialSRSStep"]
  self:assertType(initialSRSStep, "number", "initialSRSStep")
  self:expectGreaterThan(initialSRSStep, 0, "initialSRSStep")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExerciseValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateCards(self, concept)
  validateCategoryIdentifier(self, concept)
  validateSkillGroupIdentifier(self, concept)
  validateSkillGroupProgressLevel(self, concept)
  validateIsPro(self, concept)
  validateProProgressLevelMatch(self, concept)
  validateExerciseOrder(self, concept)
  validateInitialSRSStep(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ExerciseValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, ExerciseValidator)

  return self
end
