--- ExpelliarmusGame
-- @author Created by karl on August 21, 2014.

class("ExpelliarmusGame")
FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)
require("StandardUIBuilder")
require("ExpelliarmusConfiguration")
require("ExpelliarmusCommon")
require("ExpelliarmusQuestionCard")
require("widgets.progress.BarProgressIndicator")
require("widgets.progress.SpriteRevealProgressIndicator")
require("widgets.progress.ProgressTimer")
require("ExpelliarmusAnswerSelector")
require("ExpelliarmusConceptGenerator")
require("ExpelliarmusTopicList")
require("widgets.TapToContinue")
require("data.generators.ValueGenerators")
require("ExpelliarmusEndScreen")
require("ExpelliarmusTopicInterstitial")
require("bonuses.ResponseSpeedTracker")
require("data.generators.LearningFocusedRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_BG = 0
local Z_NORMAL = 10
local Z_PROGRESS = 20
local Z_OVERLAY = 30

local function configureGameDifficulty(self)
  local rankTwoScore = ExpelliarmusConfiguration.getRankTwoScore()
  local rankThreeScore = ExpelliarmusConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(ExpelliarmusConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local onQuestionAnswered

local function startNextRound(self)
  GameSession.notifyGameTimeStart()
  local selfWidth, selfHeight = self:getDims()

  local isNewTopic = false
  local progressStep = self.progressIndicator:getProgressStep()
  local progressStepInTopic = progressStep % ExpelliarmusConfiguration.QUESTIONS_PER_TOPIC_COUNT
  self.timer:setTime(0)

  if (progressStep >= self.progressStepForNextTopic) then
    self.progressStepForNextTopic = self.progressStepForNextTopic + ExpelliarmusConfiguration.QUESTIONS_PER_TOPIC_COUNT
    self.topicNumber = self.topicNumber + 1
    isNewTopic = self.topicNumber > 1
  end

  local progressInTopicWhenMistakesResurface = LearningFocusedRoundGenerator.DEFAULT_PROGRESS_BEFORE_RESURFACING *
                                               ExpelliarmusConfiguration.QUESTIONS_PER_TOPIC_COUNT
  if (progressStepInTopic > progressInTopicWhenMistakesResurface) then
    self.conceptGenerator:allowMistakesToResurface()
  end

  if self.currentCard then
    self.currentCard:removeSelf()
  end

  local topic = self.topics[self.topicNumber]

  if isNewTopic then
    self.conceptGenerator:clearMistakes()
    self.conceptGenerator:setAreMistakesAllowedToResurface(false)
    GameSession.notifyGameTimePause()
    Sound.playSoundEffect("complete_set")
    local customHeader = nil
    if (self.topicNumber == ExpelliarmusConfiguration.TOPIC_COUNT) then
      customHeader = Localizer.get("LAST_ROUND")
    end
    local interstitial = ExpelliarmusTopicInterstitial.new(self.topicNumber, topic, customHeader)
    self:insertAtCenter(interstitial, Z_NORMAL)
    wait(interstitial:animateIn())
    waitFor(0.5)
    wait(interstitial:animateOut())
    interstitial:removeSelf()
    GameSession.notifyGameTimeStart()
  end

  self.setBaseConceptGeneratorTopic(topic)

  local concept = self.conceptGenerator:next()
  local card = ExpelliarmusQuestionCard.new(concept, self.topicNumber, self.uiOverlay, function(answer) onQuestionAnswered(self, answer) end)
  card:setLoc(selfWidth * 1.5, selfHeight * 0.5)
  self:insert(card, Z_NORMAL)

  local slideAction = AHSequence.new(
    AHLocationXTo.new(card, 0.3, AHEaseSineOut.new(), selfWidth/2-10),
    AHLocationXTo.new(card, 0.3, AHEaseSineOut.new(), selfWidth/2)
  )

  self.currentCard = card
  self.currentConcept = concept
  GameSession.contentTracker:notifyPresentedConcept(self.currentConcept.identifier)

  slideAction:start()
  Sound.playSoundEffect("new_question", 0.5)
  waitFor(0.5)
  self.responseSpeedTracker:notifyStart()
  self.timer:setDuration(ExpelliarmusConfiguration.QUESTION_TIMER_DURATIONS[self.topicNumber])
  self.timer:start()
  card:setTouchEnabled(true)
end

local function shakeScreen(self)
  local screenGroup = Screen.rootGroup
  local shakeCount = 2
  local shakeDuration = 0.1
  local rotation = math.pi
  local ease = AHEaseSineInOut.new()
  local action = AHSequence.new(
    AHRotationTo.new(screenGroup, shakeDuration/2, ease, rotation),
    AHRotationTo.new(screenGroup, shakeDuration/2, ease, -rotation)
  )

  return AHSequence.new(
    AHLoop.new(action, shakeCount),
    AHRotationTo.new(screenGroup, shakeDuration/4, ease, 0)
  ):start()
end

local function loseLife(self)
  self.livesCounter:decrement()
end

local function isLoss(self)
  return self.livesCounter:getLives() == 0
end

local function isWin(self)
  return self.progressIndicator:getProgress() == 1
end

local function onWin(self)
  local endScreen = ExpelliarmusEndScreen.new()
  self:insertAtCenter(endScreen, Z_OVERLAY)
  wait(endScreen:start())

  GameSession.notifyGameWin()
end

local function onLoss(self)
  GameSession.notifyGameLoss()
end

local function playCorrectSoundEffect(topicNumber)
  assert((topicNumber == 1) or (topicNumber == 2) or (topicNumber == 3),
    "No sound defined for topic number: " .. tostring(topicNumber))
  local soundName = "correct_" .. topicNumber
  Sound.playSoundEffect(soundName)
end

local function increaseProgress(self)
  return self.progressIndicator:newProgressStepBy(0.3, AHEaseSineInOut.new(), 1):start()
end

local function onCorrectAnswer(self, answer)
  GameSession.contentTracker:notifyCorrectAnswer(self.currentConcept.identifier)
  GameSession.notifyCorrectAnswer(self.currentConcept.identifier)
  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(ExpelliarmusConfiguration.POINTS_PER_CORRECT_RESPONSE)

  self.responseSpeedTracker:notifyCorrectResponse()
  self.scoreLabel:setScore(GameSession.getScore())
  playCorrectSoundEffect(self.topicNumber)
  increaseProgress(self)
  local correctAnswerAction = self.currentCard:markCorrect(answer)

  local willWin = (self.progressIndicator:getProgressStep() + 1 == self.progressIndicator:getStepCount())
  if willWin then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), ExpelliarmusConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())
  end

  waitFor(1)
  self.timer:setTime(0)
  wait(correctAnswerAction)
  if isWin(self) then
    GameSession.notifyGameTimePause()
    local delayedUIFadeAction = AHDelayed.new(1.7, self.uiOverlay:createFadeOutAction())
    delayedUIFadeAction:start()

    onWin(self)
  else
    startNextRound(self)
  end
end

local function onIncorrectAnswer(self, answer)
  GameSession.notifyGameTimePause()
  if (answer ~= nil) then
    GameSession.contentTracker:notifyIncorrectAnswer(self.currentConcept.identifier)
  else
    GameSession.contentTracker:notifyTimeout(self.currentConcept.identifier)
  end
  GameSession.scoredIncorrectAnswer(self.currentConcept.identifier)
  GameSession.incrementIncorrectAnswerCount()
  self.conceptGenerator:notifyMistake(self.currentConcept.identifier)
  self.responseSpeedTracker:notifyIncorrectResponse()
  Sound.playSoundEffect("incorrect")
  loseLife(self)
  self.currentCard:setZIndex(Z_OVERLAY)
  shakeScreen(self)
  wait(self.currentCard:markIncorrect(answer))
  if isLoss(self) then
    onLoss(self)
  else
    startNextRound(self)
  end
end

onQuestionAnswered = function(self, answer)
  self.timer:stop()
  self.currentCard:setTouchEnabled(false)
  return runInCoroutine(function()
    if answer == self.currentConcept.correct then
      onCorrectAnswer(self, answer)
    else
      onIncorrectAnswer(self, answer)
    end
  end)
end

local function onTimeout(self)
  self.currentCard:setTouchEnabled(false)
  onIncorrectAnswer(self, nil)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpelliarmusGame:gameFlow()
  self.topicNumber = 1
  self.progressStepForNextTopic = ExpelliarmusConfiguration.QUESTIONS_PER_TOPIC_COUNT
  self.topicList:animateIn()
  GameSession.notifyGameTimePause()
  waitFor(0.3)
  Sound.playSoundEffect("game_start")
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function getTopics(self)
  local topicCount = ExpelliarmusConfiguration.TOPIC_COUNT

  local selectedTopicTypes = {}
  local roundKeyFrames = ExpelliarmusConfiguration.getRoundKeyFrames()

  repeat
    local topicNameChooser = KeyFrameInterpolator.new(roundKeyFrames, GameSession.getDifficulty())

    local nextTopic = nil
    local isUniqueTopic = false
    repeat
      nextTopic = topicNameChooser:nextValue()

      isUniqueTopic = (not TableUtils.contains(selectedTopicTypes, nextTopic))
    until isUniqueTopic

    table.insert(selectedTopicTypes, nextTopic)
  until (#selectedTopicTypes == topicCount)

  return selectedTopicTypes
end

local function getTotalQuestionCount()
  return ExpelliarmusConfiguration.TOPIC_COUNT * ExpelliarmusConfiguration.QUESTIONS_PER_TOPIC_COUNT
end

local function newProgressIndicator(self)
  local stepCount = getTotalQuestionCount()
  local fgSprite = Sprite.newWithFilename("rainbowBar_filled.png")
  local bgSprite = Sprite.newWithFilename("rainbowBar_unfilled.png")
  local growDirection = SpriteRevealProgressIndicator.GROW_RIGHT
  local progressIndicator = SpriteRevealProgressIndicator.new(fgSprite, bgSprite, growDirection, stepCount)
  progressIndicator:setEffectiveDims(Screen.getSafeWidthPart(0.9), progressIndicator:getHeight())
  return progressIndicator
end

local function createGameElements(self)
  local selfWidth, selfHeight = self:getDims()

  self.timer = BarProgressIndicator.new(selfWidth, selfHeight, ExpelliarmusCommon.COLOR_TIMER, ExpelliarmusCommon.COLOR_BACKGROUND, BarProgressIndicator.GrowUp)
  ProgressTimer.makeTimer(self.timer, ExpelliarmusConfiguration.QUESTION_TIMER_DURATIONS[1], function() onTimeout(self) end)
  self:insertAtCenter(self.timer, Z_BG)

  local stepCount = getTotalQuestionCount()
  local fgColor = ExpelliarmusCommon.TOPIC_COLORS[2]
  local bgColor = ColorUtils.colorWithOpacityMultiplier(fgColor, 0.5)
  self.progressIndicator = newProgressIndicator(self)
  local bottomMargin = 42.0
  local minimumY = bottomMargin + self.progressIndicator:getHeightPart(0.5)
  local safeY = Screen.safeBottom + self.progressIndicator:getHeightPart(2.0)
  local progressIndicatorY = math.max(minimumY, safeY)
  self.progressIndicator:setLoc(selfWidth/2, progressIndicatorY)
  self:insert(self.progressIndicator, Z_PROGRESS)
  self.progressIndicator:setProgress(0)

  self.topicList = ExpelliarmusTopicList.new(self.topics)
  local tapToContinue = TapToContinue.new(self.topicList, nil, function()
    wait(self.topicList:animateOut())
    self.topicList:removeSelf()
    startNextRound(self)
  end)

  local function onQuickResponse(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 ExpelliarmusConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(ExpelliarmusConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       onQuickResponse,
                                                       ExpelliarmusConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, ExpelliarmusCommon.COLOR_UI)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(ExpelliarmusConfiguration.INITIAL_LIVES_COUNT)
end

-- Constructor ------------------------

function ExpelliarmusGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ExpelliarmusGame)

  self.topics = getTopics(self)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  local baseConceptGenerator = ExpelliarmusConceptGenerator.new()

  self.setBaseConceptGeneratorTopic = function(topic)
    baseConceptGenerator:setTopic(topic)
  end

  self.conceptGenerator = LearningFocusedRoundGenerator.new(baseConceptGenerator)

  return self
end
