--- ExpendituresConceptBuilder
-- @author Created by Jay Austria on October 22, 2014.

class("ExpendituresConceptBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- To make things easier to read.
local YEARLY = ExpendituresCommon.FrequencyType.YEARLY
local MONTHLY = ExpendituresCommon.FrequencyType.MONTHLY
local WEEKLY = ExpendituresCommon.FrequencyType.WEEKLY
local EASY = ExpendituresCommon.ContentDifficulty.EASY
local MEDIUM = ExpendituresCommon.ContentDifficulty.MEDIUM
local HARD = ExpendituresCommon.ContentDifficulty.HARD

local PRICE_MIN = "min_price"
local PRICE_MAX = "max_price"
local FREQUENCY_MIN = "freq_min"
local FREQUENCY_MAX = "freq_max"
local FREQUENCY_TYPE = "frequency_type"
local ROUNDING = "rounding"

local ROUND_TO_ZERO = "round_to_zero"
local ROUND_TO_FIVE = "round_to_five"

local ConceptConstraints =
{
  [YEARLY] =
  {
    [EASY] = {FREQUENCY_MIN = 1, FREQUENCY_MAX = 2, PRICE_MIN = 0, PRICE_MAX = 100, ROUNDING = ROUND_TO_ZERO},
    [MEDIUM] = {FREQUENCY_MIN = 2, FREQUENCY_MAX = 3, PRICE_MIN = 0, PRICE_MAX = math.huge, ROUNDING = ROUND_TO_FIVE},
    [HARD] = {FREQUENCY_MIN = 4, FREQUENCY_MAX = math.huge, PRICE_MIN = 0, PRICE_MAX = math.huge, ROUDING = ROUND_TO_FIVE}
  },
  [MONTHLY] =
  {
    [EASY] = {FREQUENCY_MIN = 1, FREQUENCY_MAX = 1, PRICE_MIN = 1, PRICE_MAX = 12},
    [MEDIUM] = {FREQUENCY_MIN = 2, FREQUENCY_MAX = 3, PRICE_MIN = 1, PRICE_MAX = 100, ROUNDING = ROUND_TO_FIVE},
    [HARD] = {FREQUENCY_MIN = 3, FREQUENCY_MAX = math.huge, PRICE_MIN = 1, PRICE_MAX = 200}
  },
  [WEEKLY] =
  {
    [EASY] = {FREQUENCY_MIN = 1, FREQUENCY_MAX = 1, PRICE_MIN = 1, PRICE_MAX = 12},
    [MEDIUM] = {FREQUENCY_MIN = 2, FREQUENCY_MAX = 2, PRICE_MIN = 1, PRICE_MAX = 200, ROUNDING = ROUND_TO_FIVE},
    [HARD] = {FREQUENCY_MIN = 3, FREQUENCY_MAX = math.huge, PRICE_MIN = 1, PRICE_MAX = 200, ROUNDING = ROUND_TO_FIVE}
  }
}

local FrequencyMultiplier =
{
  [YEARLY] = 1,
  [MONTHLY] = 12,
  [WEEKLY] = 52
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConstraintsForContentDifficultyAndFrequencyType(contentDifficulty, frequencyType)
  return ConceptConstraints[frequencyType][contentDifficulty]
end

local function doRangesOverlap(aMin, aMax, bMin, bMax)
  local doOverlap = (aMin <= bMax) and (bMin <= aMax)

  return doOverlap
end

local function generateFrequency(descriptor, contentDifficulty)
  local frequencyType = descriptor[FREQUENCY_TYPE]
  local frequencyMin = descriptor[FREQUENCY_MIN]
  local frequencyMax = descriptor[FREQUENCY_MAX]

  local constraints = getConstraintsForContentDifficultyAndFrequencyType(contentDifficulty, frequencyType)

  local min = math.max(frequencyMin, constraints.FREQUENCY_MIN)
  local max = math.min(frequencyMax, constraints.FREQUENCY_MAX)

  assert(min <= max, "Invalid min and max constraints. Frequency min exceeded frequency max.")

  local randomFrequency = math.random(min, max)
  return randomFrequency
end

local function generatePrice(descriptor, contentDifficulty)
  local frequencyType = descriptor[FREQUENCY_TYPE]
  local priceMin = descriptor[PRICE_MIN]
  local priceMax = descriptor[PRICE_MAX]

  local constraints = getConstraintsForContentDifficultyAndFrequencyType(contentDifficulty, frequencyType)
  local roundingType = constraints.ROUNDING

  local min = math.max(priceMin, constraints.PRICE_MIN)
  local max = math.min(priceMax, constraints.PRICE_MAX)

  assert(min <= max, "Invalid min and max constraints. Price min exceeded price max.")

  local randomPrice = math.random(min, max)

  if randomPrice > 10 then
    if roundingType then
      local roundingAmount = 0
      if roundingType == ROUND_TO_ZERO then
        roundingAmount = 10
      elseif roundingType == ROUND_TO_FIVE then
        roundingAmount = 5
      end

      randomPrice = MathUtils.roundToNearestMultiple(randomPrice, roundingAmount)
    end
  end

  return randomPrice
end

local function isTooCloseToAnswer(answer, value)
  return math.abs(answer - value) / answer < 0.15
end

local function generateDistractors(price, frequency, frequencyTypeMultiplier)
  local distractors = {}

  local answer =  price * frequency * frequencyTypeMultiplier

  -- Add some distractors which are variants of the price
  local variantInterval = 1
  if answer > 10  and answer < 30 then
    variantInterval = 2
  elseif answer > 30 and answer < 100 then
    variantInterval = 5
  elseif answer > 100 then
    variantInterval = 100
  end

  -- Add some distractors which are variants of the final price
  table.insert(distractors, answer - 3 * variantInterval)
  table.insert(distractors, answer - 2 * variantInterval)
  table.insert(distractors, answer - 1 * variantInterval)
  table.insert(distractors, answer + 1 * variantInterval)
  table.insert(distractors, answer + 2 * variantInterval)
  table.insert(distractors, answer + 3 * variantInterval)

  table.insert(distractors, MathUtils.round(answer * 0.67))
  table.insert(distractors, MathUtils.round(answer * 0.71))
  table.insert(distractors, MathUtils.round(answer * 0.75))
  table.insert(distractors, MathUtils.round(answer * 0.79))
  table.insert(distractors, MathUtils.round(answer * 1.21))
  table.insert(distractors, MathUtils.round(answer * 1.25))

  -- Add some distractors which are variants of the frequency
  table.insert(distractors, price * (frequency - 2) * frequencyTypeMultiplier)
  table.insert(distractors, price * (frequency - 1) * frequencyTypeMultiplier)
  table.insert(distractors, price * (frequency + 1) * frequencyTypeMultiplier)
  table.insert(distractors, price * (frequency + 2) * frequencyTypeMultiplier)

  local function isValidDistractor(distractor)
    local isInRange = distractor > 0
    local isDifferentEnough = not isTooCloseToAnswer(answer, distractor)

    return isInRange and isDifferentEnough
  end

  local validDistractors = TableUtils.getFilteredArray(distractors, isValidDistractor)
  local uniqueDistractors = TableUtils.getUniquesArray(validDistractors)

  return uniqueDistractors
end

local function getRandomFromArrayWithBackup(array, backupArray)
  local doRemove = true
  local selectedValue = nil

  if #array > 0 then
    selectedValue = TableUtils.getRandomElement(array, doRemove)
  else
    selectedValue = TableUtils.getRandomElement(backupArray, doRemove)
  end

  return selectedValue
end

local function selectDistractors(calculatedAnswer, distractors, optionsCount)
  local function isBelowAnswer(distractor)
    return distractor < calculatedAnswer
  end

  local function isAboveAnswer(distractor)
    return distractor > calculatedAnswer
  end

  local distractorsLessThanAnswer = TableUtils.getFilteredArray(distractors, isBelowAnswer)
  local distractorsMoreThanAnswer = TableUtils.getFilteredArray(distractors, isAboveAnswer)

  local chosenAnswerIndex = math.random(1, optionsCount)
  local selectedDistractors = {}
  local selectedDistractor = nil

  for responseIndex = 1, optionsCount do
    if responseIndex < chosenAnswerIndex then
      selectedDistractor = getRandomFromArrayWithBackup(distractorsLessThanAnswer, distractorsMoreThanAnswer)
    elseif responseIndex > chosenAnswerIndex then
      selectedDistractor = getRandomFromArrayWithBackup(distractorsMoreThanAnswer, distractorsLessThanAnswer)
    end

    if responseIndex ~= chosenAnswerIndex then
      table.insert(selectedDistractors, selectedDistractor)
    end
  end

  return selectedDistractors
end

local function isUniqueConcept(pastConcepts, frequency, price)
  for _, concept in pairs(pastConcepts) do
    if ((frequency == concept.frequency) and (price == concept.price)) then
       return false
    end
  end

  return true
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresConceptBuilder:isMatchingFrequencyType(descriptor, frequencyType)
  return descriptor[FREQUENCY_TYPE] == frequencyType
end

function ExpendituresConceptBuilder:generateUniqueConcept(descriptor, contentDifficulty, pastConcepts)
  local gameConcept = {}

  local frequency = nil
  local price = nil
  local frequencyTypeMultiplier = FrequencyMultiplier[descriptor[FREQUENCY_TYPE]]

  local iterationCap = 100
  local i = 0
  repeat
    frequency = generateFrequency(descriptor, contentDifficulty)
    price = generatePrice(descriptor, contentDifficulty)
    i = i + 1
  until isUniqueConcept(pastConcepts, frequency, price) or i == iterationCap

  -- As a last resort, if a unique concept couldnt be generated--let's at least make it *some* different value
  -- Leads to failure < 2.0% of the time after many tests
  if i == iterationCap then
    price = price + 5
  end

  local calculatedAnswer = price * frequency * frequencyTypeMultiplier

  local allDistractors = generateDistractors(price, frequency, frequencyTypeMultiplier)
  local selectDistractors = selectDistractors(calculatedAnswer, allDistractors, self.optionsCount)
  table.sort(selectDistractors)

  gameConcept.distractors = selectDistractors
  gameConcept.frequency = frequency
  gameConcept.frequencyType = descriptor[FREQUENCY_TYPE]
  gameConcept.price = price
  gameConcept.itemName = descriptor.art_link
  gameConcept.answer = calculatedAnswer

  gameConcept.attemptedResponses = {}

  return gameConcept
end

function ExpendituresConceptBuilder:isEligible(descriptor, contentDifficulty)
  local frequencyMin = descriptor[FREQUENCY_MIN]
  local frequencyMax = descriptor[FREQUENCY_MAX]
  local priceMin = descriptor[PRICE_MIN]
  local priceMax = descriptor[PRICE_MAX]
  local frequencyType = descriptor[FREQUENCY_TYPE]

  local constraints = getConstraintsForContentDifficultyAndFrequencyType(contentDifficulty, frequencyType)

  local frequencyInRange = doRangesOverlap(constraints.FREQUENCY_MIN, constraints.FREQUENCY_MAX, frequencyMin, frequencyMax)
  local priceInRange = doRangesOverlap(constraints.PRICE_MIN, constraints.PRICE_MAX, priceMin, priceMax)

  return (frequencyInRange and priceInRange)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresConceptBuilder.new(optionsCount)
  local self = {}
  Mixin.mix(self, ExpendituresConceptBuilder)

  self.optionsCount = optionsCount

  return self
end
