--- ExpendituresFrequencyDisplay
-- @author Created by Jay Austria on October 27, 2014.

class("ExpendituresFrequencyDisplay")
require("ExpendituresCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DISPLAY_FONT_SIZE = 28
local SLIDE_IN_DURATION = 0.55
local SLIDE_OUT_DURATION = 0.45
local SLIDE_IN_EASE = AHEaseBackOut.new()
local SLIDE_OUT_EASE = AHEaseBackIn.new()
local UPDATE_TYPE_STAGGER_DURATION = 0.05

local TYPE_STRINGS =
{
  [ExpendituresCommon.FrequencyType.WEEKLY] = Localizer.get("A_WEEK"),
  [ExpendituresCommon.FrequencyType.MONTHLY] = Localizer.get("A_MONTH"),
  [ExpendituresCommon.FrequencyType.YEARLY] = Localizer.get("A_YEAR")
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getSlideOutLabelAction(label, distance)
  local slideOut = AHLocationYBy.new(label, SLIDE_OUT_DURATION, SLIDE_OUT_EASE, distance)
  local removeLabel = AHRemove.new(label)

  local slideOutSequence = AHSequence.new(slideOut, removeLabel)
  return slideOutSequence
end

local function getSlideInLabelAction(label, distance)
  local slideInAction = AHLocationYBy.new(label, SLIDE_IN_DURATION, SLIDE_IN_EASE, distance)
  return slideInAction
end

local function createNewDisplayLabel(text)
  local newLabel = Label.new(text, Font.new(FontFaces.DIN_OT_BOLD, DISPLAY_FONT_SIZE))
  return newLabel
end

local function createNewAmountLabel(self, amount)
  local frequencyString = string.format("%d×", amount)
  local newAmountLabel = createNewDisplayLabel(frequencyString)

  local amountBoxWidth = ExpendituresCommon.DIVIDER_PLACEMENT_X_MULTIPLIER * self:getWidth()
  local amountLocationX = amountBoxWidth * 0.5
  newAmountLabel:setLoc(amountLocationX, self:getHeightPart(1.5))
  self:insert(newAmountLabel)

  return newAmountLabel
end

local function createNewTypeLabel(self, type)
  local typeString = TYPE_STRINGS[type]
  local newTypeLabel = createNewDisplayLabel(typeString)

  local typeBoxWidth = (1.0 - ExpendituresCommon.DIVIDER_PLACEMENT_X_MULTIPLIER) * self:getWidth()
  local typeLocationX = self:getWidth() - typeBoxWidth * 0.5
  newTypeLabel:setLoc(typeLocationX, self:getHeightPart(1.5))
  self:insert(newTypeLabel)

  return newTypeLabel
end

local function getUpdateLabelAction(self, newValue, cachedValueProperty, labelProperty, newLabel)
  local updateLabelAction = AHSpawn.new()

  local slideDistance = -self:getHeight()

  if self[cachedValueProperty] then
    local slideOut = getSlideOutLabelAction(self[labelProperty], slideDistance)
    updateLabelAction:addAction(slideOut)
  end

  self[labelProperty] = newLabel
  local slideIn = getSlideInLabelAction(self[labelProperty], slideDistance)
  local slideInDelay = AHDelay.new(0.3)
  local slideInSequence = AHSequence.new(slideInDelay, slideIn)
  updateLabelAction:addAction(slideInSequence)
  self[cachedValueProperty] = newValue

  return updateLabelAction
end

local function getUpdateFrequencyTypeAction(self, type)
  local newLabel = createNewTypeLabel(self, type)
  local updateAction = getUpdateLabelAction(self, type, "frequencyType", "frequencyTypeLabel", newLabel)
  return updateAction
end

local function getUpdateFrequencyAmountAction(self, amount)
  local newLabel = createNewAmountLabel(self, amount)
  local updateAction = getUpdateLabelAction(self, amount, "frequencyAmount", "frequencyAmountLabel", newLabel)
  return updateAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresFrequencyDisplay:setFrequency(frequencyAmount, frequencyType)
  local function updateFrequencyDisplay()
    if frequencyAmount ~= self.frequencyAmount then
      local frequencyUpdateAction = getUpdateFrequencyAmountAction(self, frequencyAmount)
      frequencyUpdateAction:start()
    end

    waitFor(UPDATE_TYPE_STAGGER_DURATION)

    if frequencyType ~= self.frequencyType then
      local typeUpdateAction = getUpdateFrequencyTypeAction(self, frequencyType)
      typeUpdateAction:start()
    end
  end
  runInCoroutine(updateFrequencyDisplay)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresFrequencyDisplay.new(width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, ExpendituresFrequencyDisplay)

  return self
end
