--- ExpendituresGame
-- @author Created by Jay Austria on October 20, 2014.

class("ExpendituresGame")
require("StandardUIBuilder")
require("ExpendituresCommon")
require("ExpendituresConfiguration")
require("ExpendituresProgressIndicator")
require("ExpendituresRoundGenerator")
require("ExpendituresScene")

require("bonuses.StreakTracker")
require("widgets.TapToContinue")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TIMER_COLOR = ExpendituresCommon.Color.CYAN

local FIRST_CONCEPT_DELAY = 1.4

local function configureGameDifficulty(self)
  self.roundsRemainingCount = ExpendituresConfiguration.getConceptsToComplete()

  local rankTwoScore = ExpendituresConfiguration.getRankTwoScore()
  local rankThreeScore = ExpendituresConfiguration.getRankThreeScore()

  GameSession.setPassScore(ExpendituresConfiguration.getPassScore())
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setTotalTime(ExpendituresConfiguration.getGameDuration())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.MAX_STREAK_BONUS)
  GameSession.contentTracker:setSquashData(true)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateScore(self)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function didWin(self)
  return (not self.progressIndicator:doesContainIcons())
end

local function isGameEnded(self)
  local playerWon = didWin(self)
  local timeEnded = (self.timer:getTime() >= ExpendituresConfiguration.getGameDuration())

  return playerWon or timeEnded
end

local function doGameIntro(self)
  self.scene:playIntroAnimation(self.concepts)
  self.progressIndicator:playIntroAnimation(self.concepts)

  local firstConceptDelay = AHDelay.new(FIRST_CONCEPT_DELAY)
  wait(firstConceptDelay:start())
end

local function waitForResponse(self)
  self.responseLock = AHDelay.new(math.huge)
  wait(self.responseLock:start())
end

local function unlockResponseLock(self)
  if self.responseLock then
    self.responseLock:stop()
    self.responseLock = nil
  end
end

local function showNextQueuedConcept(self)
  local nextConcept = table.remove(self.concepts, 1)

  self.scene:presentConcept(nextConcept)
  -- NOTE: This game doesn't have a content sheet or concept IDs; dynamically
  -- create an "ID" based on the content itself in the form of
  -- basePrice-frequency-frequencyType, eg 20-1-monthly
  local customID = string.format("%d-%d-%s", nextConcept.price, nextConcept.frequency, nextConcept.frequencyType)
  self.currentID = customID
  GameSession.contentTracker:notifyPresentedConcept(customID)
  self.progressIndicator:arrangeIcons()

  self.currentConcept =  nextConcept
end

local function reinsertCurrentConceptAtEnd(self, response)
  table.insert(self.currentConcept.attemptedResponses, response)
  table.insert(self.concepts, self.currentConcept)
end

local function isAtHalfwayPoint(self)
  local numberToComplete = ExpendituresConfiguration.getConceptsToComplete()
  return (self.roundsRemainingCount == math.ceil(numberToComplete / 2))
end

local function isAtTwoRemainingPoint(self)
  return (self.roundsRemainingCount == 2)
end

local function shouldShowHalfwayProgress(self)
  return isAtHalfwayPoint(self) and not self.hasShownHalfwayProgress
end

local function shouldShowTwoRemainingProgress(self)
  return isAtTwoRemainingPoint(self) and not self.hasShownTwoRemainingProgress
end

local function getWillShowProgress(self)
  local willShowProgress = shouldShowHalfwayProgress(self) or shouldShowTwoRemainingProgress(self)
  return willShowProgress
end

local function handleCorrectResponse(self, response)
  local pointReward = ExpendituresConfiguration.CORRECT_ANSWER_POINTS
  GameSession.increaseScore(pointReward)
  updateScore(self)

  self.roundsRemainingCount = self.roundsRemainingCount - 1

  local willShowProgress = getWillShowProgress(self)

  if shouldShowHalfwayProgress(self) then
    self.hasShownHalfwayProgress = true
  end
  if shouldShowTwoRemainingProgress(self) then
    self.hasShownTwoRemainingProgress = true
  end

  self.streakTracker:notifyCorrectAnswer()

  self.progressIndicator:popFrontAsCorrect()
  Sound.playSoundEffect("expenses_correct_ans")

  self.scene:waitForCorrectAnimation(response, willShowProgress, self.roundsRemainingCount)

  unlockResponseLock(self)
end

local function waitForHideHelper(self)
  local hideHelperAnimation = self.scene:waitForHideHelper()
  Sound.playSoundEffect("expenses_tip")
  wait(hideHelperAnimation)

  local fadeDuration = 0.2
  local resumeAction = AHSpawn.new(
    AHOpacityTo.new(self.pauseButton, fadeDuration, nil, 1.0),
    AHOpacityTo.new(self.scoreLabel, fadeDuration, nil, 1.0),
    function()
      self.timer:resumeAndHideIcon(fadeDuration)
    end
  )
  resumeAction:start()

  waitFor(fadeDuration)
  self.progressIndicator:rebootOpacityCoroutine()
end

local function waitForShowHelper(self, price, frequency, frequencyType)
  Sound.playSoundEffect("expenses_tip")
  self.scene:waitForShowHelper(price, frequency, frequencyType)

  local fadeDuration = 0.3
  local pauseAction = AHSpawn.new(
    self.progressIndicator:getFadeOutIconsAction(fadeDuration, 0.0),
    self.uiOverlay:createFadeOutAction(fadeDuration)
  )
  pauseAction:start()
end

local function waitForHelpFlow(self, price, frequency, frequencyType)
  self.scene:fadeOutAndResetButtons()
  waitForShowHelper(self, price, frequency, frequencyType)

  waitForHideHelper(self)
end

local function waitForGameWinOutro(self)
  self.timer:pause()
  Sound.playSoundEffect("expenses_win_game")
  self.scene:waitForOutroAnimation(self)
end

local function shouldShowHint(self)
  local currentFrequencyIncorrectCount = self.incorrectCounts[self.currentConcept.frequencyType]
  return (currentFrequencyIncorrectCount == ExpendituresConfiguration.NUMBER_INCORRECT_IN_FREQUENCY_TO_SHOW_HINT)
end

local function handleIncorrectResponse(self, response)
  self.incorrectCounts[self.currentConcept.frequencyType] = self.incorrectCounts[self.currentConcept.frequencyType] + 1
  local shouldShowHint = shouldShowHint(self)
  if shouldShowHint then
    local fadeDuration = 0.3
    self.timer:pauseAndShowIcon(fadeDuration)
  end
  self.streakTracker:notifyIncorrectAnswer()

  self.progressIndicator:animateFrontAsIncorrect()
  Sound.playSoundEffect("expenses_incorrect")

  local willShowProgress = getWillShowProgress(self)

  self.scene:waitForIncorrectAnimation(response, willShowProgress)

  local doButtonFadeDuringReject = true

  if shouldShowHint then
    doButtonFadeDuringReject = false
    waitForHelpFlow(self, self.currentConcept.price, self.currentConcept.frequency, self.currentConcept.frequencyType)
  end

  self.progressIndicator:reinsertFront()
  self.scene:waitForRejectAnimation(doButtonFadeDuringReject)

  reinsertCurrentConceptAtEnd(self, response)

  unlockResponseLock(self)

end

local function handleUserResponse(self, response)
  local isValidResponse = (response and (response > 0) and not isGameEnded(self))
  if isValidResponse then
    self.scene:setEnableInteraction(false)

    local function handleResponse()
      if (response == self.currentConcept.answer) then
        if (not self.incorrectResponseSet[self.currentID]) then
          GameSession.contentTracker:notifyCorrectAnswer(self.currentID)
          GameSession.incrementCorrectAnswerCount()
        end
        handleCorrectResponse(self, response)
      else
        self.incorrectResponseSet[self.currentID] = true
        GameSession.contentTracker:notifyIncorrectAnswer(self.currentID, nil, tostring(response))
        GameSession.incrementIncorrectAnswerCount()
        handleIncorrectResponse(self, response)
      end
    end
    runInCoroutine(handleResponse)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresGame:gameFlow()
  doGameIntro(self)

  -- Start and pause timer here to make it look visible full but frozen.
  -- Resume the timer once the round begins.
  self.timer:initialize()

  repeat
    self.timer:resume()
    showNextQueuedConcept(self)
    waitForResponse(self)

  until isGameEnded(self)

  if didWin(self) then
    ScoringUtils.applyMaxStreakBonus(self.streakTracker:getLongestStreak(), ExpendituresConfiguration.MAX_STREAK_BONUS_POINTS, self.bonusTray)
    ScoringUtils.applyTimeBonus(self.timer:getTimeRemaining(), ExpendituresConfiguration.getTimeBonusPerSecond(), self.bonusTray)
    local consistencyBonus = ExpendituresConfiguration.getConsistencyBonusForMistakes(GameSession.getIncorrectAnswerCount())
    if (consistencyBonus > 0) then
      ScoringUtils.applyCustomBonus(consistencyBonus, BonusTypes.TYPES.CONSISTENCY_BONUS, self.bonusTray)
    end
    updateScore(self)

    local delayedUIFadeAction = AHDelayed.new(1.8, self.uiOverlay:createFadeOutAction())
    delayedUIFadeAction:start()

    waitForGameWinOutro(self)

    GameSession.notifyGameWin()
  else
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local scene = ExpendituresScene.new()

  local function responseHandler(buttonValue)
    handleUserResponse(self, buttonValue)
  end

  scene:setResponseHandler(responseHandler)
  self:insertAtCenter(scene)
  self.scene = scene

  local function onStreakWrapper()
    local streakCount = self.streakTracker:getCurrentStreak()
    self.bonusTray:notifyStreakWithCount(streakCount)
  end

  local function onStreakEndWrapper()
    self.bonusTray:notifyStreakEnd()
  end

  local streakMinimimum = 2
  local streakTracker = StreakTracker.new(streakMinimimum, onStreakWrapper, onStreakEndWrapper)
  self.streakTracker = streakTracker
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, TIMER_COLOR)

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local function timeOutWrapper()
    GameSession.contentTracker:notifyTimeout(self.currentID)
    unlockResponseLock(self)
  end

  self.timer = builder:addDigitTimer(ExpendituresConfiguration.getGameDuration())
  self.timer:setTimeoutHandler(timeOutWrapper)

  local progressIndicator = ExpendituresProgressIndicator.new()
  local xOffset = 0.0
  local yOffset = -Screen.getSafeHeightPart(0.025)
  builder:addWidget(progressIndicator, UIOverlay.NORTH, xOffset, yOffset)
  self.progressIndicator = progressIndicator
end

local function createConcepts(self)
  self.concepts = {}

  local roundGenerator = ExpendituresRoundGenerator.new(ExpendituresCommon.RESPONSE_OPTIONS_COUNT)

  roundGenerator:setFrequencyTypeMode(ExpendituresCommon.FrequencyType.YEARLY)
  local yearlyCount = ExpendituresConfiguration.CONCEPT_COUNT_YEARLY
  if yearlyCount then
    TableUtils.concatenateArrays(self.concepts, roundGenerator:nextAsArray(yearlyCount))
  end

  roundGenerator:setFrequencyTypeMode(ExpendituresCommon.FrequencyType.MONTHLY)
  local monthlyCount = ExpendituresConfiguration.CONCEPT_COUNT_MONTHLY
  if monthlyCount then
    TableUtils.concatenateArrays(self.concepts, roundGenerator:nextAsArray(monthlyCount))
  end

  roundGenerator:setFrequencyTypeMode(ExpendituresCommon.FrequencyType.WEEKLY)
  local weeklyCount = ExpendituresConfiguration.CONCEPT_COUNT_WEEKLY
  if weeklyCount then
    TableUtils.concatenateArrays(self.concepts, roundGenerator:nextAsArray(weeklyCount))
  end
end

-- Constructor ------------------------

function ExpendituresGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ExpendituresGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)
  createConcepts(self)

  self.incorrectCounts =
  {
    [ExpendituresCommon.FrequencyType.YEARLY] = 0,
    [ExpendituresCommon.FrequencyType.MONTHLY] = 0,
    [ExpendituresCommon.FrequencyType.WEEKLY] = 0
  }
  self.incorrectResponseSet = {}
  self.hasShownHalfwayProgress = false
  self.hasShownTwoRemainingProgress = false

  return self
end
