--- ExpendituresIcon
-- @author Created by Jay Austria on October 30, 2014.

class("ExpendituresIcon")
require("ExpendituresCommon")
require("ExpendituresRockShake")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PRICE_FONT_SIZE = 30
local PRICE_LOCATION_Y_OFFSET = 10
local PRICE_COLOR = ExpendituresCommon.Color.CYAN
local FADE_IN_DURATION = 0.1

local TILT_ROTATION_AMOUNT = 10
local TILT_DELAY_DURATION = 0.21
local TILT_ROCK_DURATION = 0.1
local TILT_EXIT_ROCK_DURATION = 0.05

local SHRINK_PRICE_DURATION = 0.1
local JUMP_UP_DURATION = 0.12
local FALL_DELAY_DURATION = 0.05
local FALL_DOWN_DURATION = 0.20
local JUMP_UP_OFFSET_Y = 20

local GROUP_WIDTH = 100
local GROUP_HEIGHT = 100

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresIcon:getExitRockAnimation()
  local rockAction = AHRotationTo.new(self.itemIcon, TILT_EXIT_ROCK_DURATION, nil, -TILT_ROTATION_AMOUNT * 0.5)
  return rockAction
end

function ExpendituresIcon:getCorrectAnimation()
  local shrinkPrice = AHScaleTo.new(self.priceLabel, SHRINK_PRICE_DURATION, nil, 0.0)
  local expandCheck = AHScaleTo.new(self.checkMark, SHRINK_PRICE_DURATION, nil, 1.0)
  local priceAndCheckSpawn = AHSpawn.new(shrinkPrice, expandCheck)

  local jumpUp = AHLocationYBy.new(self.itemIcon, JUMP_UP_DURATION, AHEaseSimpleOut.new(), JUMP_UP_OFFSET_Y)
  local fallDelay = AHDelay.new(FALL_DELAY_DURATION)
  local fallDown = AHLocationYBy.new(self.itemIcon, FALL_DOWN_DURATION, AHEaseSimpleIn.new(), -JUMP_UP_OFFSET_Y)

  local iconJumpSequence = AHSequence.new(jumpUp, fallDelay, fallDown)

  local function showParticles()
    for _, particles in ipairs(self.particles) do
      particles:setOpacity(1.0)
    end
  end

  local playParticles = AHSequence.new(showParticles, self.combinedParticlesAnimation)

  local fadeInGlow = AHOpacityTo.new(self.blurCircle, 0.2, nil, 0.4)
  local expandGlow = AHScaleTo.new(self.blurCircle, 0.2, nil, 5)

  local showGlow = AHSpawn.new(fadeInGlow, expandGlow)

  local correctAnimation = AHSpawn.new(priceAndCheckSpawn, iconJumpSequence, playParticles, showGlow)

  return correctAnimation
end

function ExpendituresIcon:getIncorrectAnimation()
  local ROCK_AMPLITUDE = 2.5

  local shakeAction = ExpendituresRockShake.new(self.itemIcon,
                                                ROCK_AMPLITUDE,
                                                ExpendituresCommon.INCORRECT_ROCK_DURATION,
                                                ExpendituresCommon.INCORRECT_ROCK_CYCLE_COUNT)

  local changeLabelColor = AHColorTo.new(self.priceLabel, ExpendituresCommon.INCORRECT_MORPH_DURATION, nil, ExpendituresCommon.Color.WHITE)

  local incorrectAnimation = AHSpawn.new(shakeAction, changeLabelColor)

  return incorrectAnimation
end

function ExpendituresIcon:getPriceFadeInAction()
  return AHOpacityTo.new(self.priceLabel, FADE_IN_DURATION, nil, 1.0)
end

function ExpendituresIcon:getRockAnimation()
  local initializeRotation = function() self.itemIcon:setRot(-TILT_ROTATION_AMOUNT) end
  local delayRotation = AHDelay.new(TILT_DELAY_DURATION)
  local rotationSequence = AHSequence.new(AHRotationTo.new(self.itemIcon, TILT_ROCK_DURATION, nil, TILT_ROTATION_AMOUNT * 0.5),
                                          AHRotationTo.new(self.itemIcon, TILT_ROCK_DURATION, nil, TILT_ROTATION_AMOUNT * -0.25),
                                          AHRotationTo.new(self.itemIcon, TILT_ROCK_DURATION, nil, 0))

  return AHSequence.new(initializeRotation, delayRotation, rotationSequence)
end

function ExpendituresIcon:prepForPresent()
  self.priceLabel:setColor(unpack(PRICE_COLOR))
  self.priceLabel:setOpacity(0)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresIcon.new(iconName, price)
  local iconFilename = string.format("%s.png", iconName)

  local itemIcon = Sprite.newWithName(iconFilename)

  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  Mixin.mix(self, ExpendituresIcon)

  itemIcon:setAnchorPointBottom()
  self:insertAtCenter(itemIcon)
  self.itemIcon = itemIcon

  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_LIGHT, PRICE_FONT_SIZE),
    skipNextSpacing = true,
    skipPreviousSpacing = true
  }

  local attributesForLabels = {
    b = {font = Font.new(FontFaces.DIN_OT_BOLD, PRICE_FONT_SIZE)}
  }

  local priceString = ExpendituresCommon.getPriceString(price)
  local priceLabel = DynamicTextArea.new(priceString, GROUP_WIDTH, baseAttributes, attributesForLabels, Alignment.CENTER)
  priceLabel:setColor(unpack(PRICE_COLOR))
  priceLabel:setOpacity(0)
  priceLabel:setLoc(self:getWidthPart(0.5), PRICE_LOCATION_Y_OFFSET)
  self:insert(priceLabel)
  self.priceLabel = priceLabel

  local checkMark = Sprite.newWithName("checkmark.png")
  checkMark:setLoc(self:getWidthPart(0.5), PRICE_LOCATION_Y_OFFSET)
  checkMark:setScl(0)
  self:insert(checkMark)
  self.checkMark = checkMark

  local particlesDescriptors =
  {
    {x = GROUP_WIDTH * 0.5, y = GROUP_HEIGHT * 0.6, scale = 1.8},
    {x = GROUP_WIDTH * 0.5, y = GROUP_HEIGHT * 1.0, scale = 1.3},
    {x = GROUP_WIDTH * 0.5, y = GROUP_HEIGHT * 1.4, scale = 1.8}
  }

  local particles = {}
  local combinedParticlesAnimation = AHSpawn.new()

  local startDelayDuration = 0
  for _, particlesDescriptor in ipairs(particlesDescriptors) do
    local newParticles = Sprite.newWithName("particle_000.png")
    newParticles:setScl(0.2)
    newParticles:setOpacity(0)
    newParticles:setLoc(particlesDescriptor.x, particlesDescriptor.y)
    self:insert(newParticles)
    table.insert(particles, newParticles)

    local particlesAnimation = Animation.newWithPrefix("particle_")
    local particlesAnimateAction = AHAnimate.new(newParticles, particlesAnimation, 48)

    local expandParticles = AHScaleTo.new(newParticles, 0.1, nil, particlesDescriptor.scale)
    local expandAndAnimate = AHSpawn.new(expandParticles, particlesAnimateAction)
    local delayedParticlesAnimation = AHDelayed.new(startDelayDuration, expandAndAnimate)

    combinedParticlesAnimation:addAction(delayedParticlesAnimation)

    startDelayDuration = startDelayDuration + 0.04
  end

  self.particles = particles
  self.combinedParticlesAnimation = combinedParticlesAnimation

  local blurCircle = Sprite.newWithName("blur_circle.png")
  blurCircle:setColor(unpack(ExpendituresCommon.Color.CYAN))
  blurCircle:setLoc(itemIcon:getWidthPart(0.5), itemIcon:getHeightPart(0.5))
  blurCircle:setOpacity(0.0)
  blurCircle:setScl(0)
  itemIcon:insert(blurCircle)
  self.blurCircle = blurCircle

  return self
end
