--- ExpendituresProgressIndicator
-- @author Created by Jay Austria on November 5, 2014.

class("ExpendituresProgressIndicator")
-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INDICATOR_HEIGHT = 30

local ICON_END_LOCATION_X = Screen.width * 1.2

local SLIDE_ACROSS_DURATION = 0.40
local SLIDE_ACROSS_STAGGER_DURATION = 0.1
local VISIBLE_BAND = 0.7
local ICON_SPACING = 32

local COLOR_MORPH_DURATION = 0.3
local EXIT_DELAY_DURATION = 0.3
local EXIT_DURATION = 0.2

local TOTAL_MOVE_DURATION = 0.63
local MOVE_DURATION = 0.33
local MOVE_DELAY = TOTAL_MOVE_DURATION - MOVE_DURATION

local SURFACE_WIDTH = Screen.width * 0.8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function runUpdateChildOpacityCoroutine(self)
  local updateOpacities = function()
    local visibleWidth = Screen.width * VISIBLE_BAND
    local visibleStartX = (Screen.width - visibleWidth) / 2

    while(true) do
      for _, child in ipairs(self.children) do
        local x = child:getX()
        local shiftedX = x - visibleStartX
        local distanceFromCenterOfBand = math.abs(shiftedX - visibleWidth * 0.5)
        local normalizedDistanceFromCenter = distanceFromCenterOfBand / visibleWidth
        local opacity = MathUtils.clamp(1 - (2 * normalizedDistanceFromCenter), 0, 1)

        child:setOpacity(opacity)
      end
      coroutine.yield()
    end
  end

  return runInCoroutine(updateOpacities)
end

local function getMoveIconToAssignedLocation(icon, locationX)
  local moveAction = AHLocationXTo.new(icon, MOVE_DURATION, AHEaseSimpleInOut.new(), locationX)
  local delayedMoveAction = AHDelayed.new(MOVE_DELAY, moveAction)

  return delayedMoveAction
end

local function getArrangeAction(self)
  local arrangeSpawn = AHSpawn.new()
  local centerX = self:getWidthPart(0.5)
  for i, icon in ipairs(self.icons) do
    local assignedX = centerX + (i - 1) * ICON_SPACING
    local moveAction = getMoveIconToAssignedLocation(icon, assignedX)
    arrangeSpawn:addAction(moveAction)
  end

  return arrangeSpawn
end

local function getAnimateAsCorrect(icon)
  local morphColor = AHColorTo.new(icon, COLOR_MORPH_DURATION, nil, ExpendituresCommon.Color.CYAN)
  local exitDelay = AHDelay.new(EXIT_DELAY_DURATION)
  local leaveAction = AHLocationXTo.new(icon, EXIT_DURATION, AHEaseSimpleIn.new(), ExpendituresCommon.ICON_OFFSCREEN_LOCATION_X)

  local correctAnimation = AHSequence.new(morphColor, exitDelay, leaveAction)
  return correctAnimation
end

local function getAnimateAsIncorrect(icon)
  local morphColor = AHColorTo.new(icon, COLOR_MORPH_DURATION, nil, ExpendituresCommon.Color.ORANGE)
  return morphColor
end

local function getReinsert(self, icon)
  local destinationX = (#self.icons + 1) * ICON_SPACING + self:getWidthPart(0.5)

  local leaveAction = AHLocationXTo.new(icon, EXIT_DURATION, AHEaseSimpleIn.new(), destinationX)
  return leaveAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresProgressIndicator:doesContainIcons()
  return (#self.icons > 0.0)
end

function ExpendituresProgressIndicator:arrangeIcons()
  local arrangeAction = getArrangeAction(self)
  arrangeAction:start()
end

function ExpendituresProgressIndicator:popFrontAsCorrect()
  local frontIcon = table.remove(self.icons, 1)
  local correctAnimation = getAnimateAsCorrect(frontIcon)

  correctAnimation:start()
end

function ExpendituresProgressIndicator:animateFrontAsIncorrect()
  local frontIcon = self.icons[1]
  local incorrectAnimation = getAnimateAsIncorrect(frontIcon)

  incorrectAnimation:start()
end

function ExpendituresProgressIndicator:reinsertFront()
  local frontIcon = table.remove(self.icons, 1)

  -- Don't do any animation if it was the last icon
  if #self.icons > 0 then
    local reinsertAnimation = getReinsert(self, frontIcon)
    reinsertAnimation:start()
  end

  table.insert(self.icons, frontIcon)
end

function ExpendituresProgressIndicator:playIntroAnimation(concepts)
  local iconY = self:getHeightPart(0.2)

  local introAnimation = AHSpawn.new()

  local icons = {}
  for i, concept in ipairs(concepts) do
    local iconFilename = string.format("small_%s.png", concept.itemName)
    local icon = Sprite.newWithName(iconFilename)
    icon:setLoc(ExpendituresCommon.ICON_OFFSCREEN_LOCATION_X, iconY)

    local appearanceDelay = SLIDE_ACROSS_STAGGER_DURATION * (#concepts - (i - 1))
    local slideAction = AHLocationXTo.new(icon, SLIDE_ACROSS_DURATION, nil, ICON_END_LOCATION_X)

    local iconIntroAction = AHDelayed.new(appearanceDelay, slideAction)
    introAnimation:addAction(iconIntroAction)

    self:insert(icon)
    table.insert(icons, icon)
  end

  self.icons = icons
  return introAnimation:start()
end

function ExpendituresProgressIndicator:getFadeOutIconsAction(duration)
  self.opacityUpdate:stop()
  self.opacityUpdate = nil

  local targetOpacity = 0.0
  local fadeAction = AHSpawn.new()
  for _, icon in ipairs(self.icons) do
      fadeAction:addAction(
      AHOpacityTo.new(icon, duration, nil, targetOpacity)
    )
  end

  return fadeAction
end

function ExpendituresProgressIndicator:rebootOpacityCoroutine()
  self.opacityUpdate = runUpdateChildOpacityCoroutine(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresProgressIndicator.new()
  local self = Group.new(Screen.width, INDICATOR_HEIGHT)
  Mixin.mix(self, ExpendituresProgressIndicator)

  local surface = Sprite.newWithName("progress_indicator_surface.png")
  local surfaceScaleX = SURFACE_WIDTH / surface:getWidth()
  surface:setScl(surfaceScaleX, 1.0)
  surface:setLoc(Screen.width * 0.5, 0)
  self:insert(surface)

  self.opacityUpdate = runUpdateChildOpacityCoroutine(self)

  return self
end
