--- ExpendituresResponseButton
-- @author Created by Jay Austria on October 31, 2014.

class("ExpendituresResponseButton")
require("ExpendituresCommon")
require("ExpendituresRockShake")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CORNER_RADIUS = 4
local BACKING_COLOR = ColorUtils.colorHex24ToStd(0x2E14A9)
local FACE_COLOR = ColorUtils.colorHex24ToStd(0x4122C3)
local LABEL_COLOR = ExpendituresCommon.Color.WHITE
local VALUE_FONT_SIZE = 18
local INITIAL_VALUE = 0
local FADE_OUT_DURATION = 0.2
local SCALE_DOWN_AMOUNT = 0.9
local SCALE_DOWN_DURATION = 0.05
local SCALE_UP_DURATION = 0.1
local COLOR_CHANGE_DURATION = SCALE_DOWN_DURATION
local DESTINATION_BUTTON_COLOR_CORRECT = ExpendituresCommon.Color.CYAN
local DESTINATION_BUTTON_COLOR_INCORRECT = ExpendituresCommon.Color.ORANGE
local DESINATION_LABEL_COLOR = ExpendituresCommon.Color.BLACK
local FADE_IN_DELAY_DURATION = 0.05
local FADE_IN_DURATION = 0.35

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createButtonRectangle(width, height, color)
  local buttonRectangle = FilledRoundedRectangle.new(width, height, CORNER_RADIUS, color)
  return buttonRectangle
end

local function updateValue(self, value)
  self.value = value

  local valueString = ExpendituresCommon.getPriceString(value)

  if self.valueLabel then
    self.valueLabel:removeSelf()
  end

  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_LIGHT, VALUE_FONT_SIZE),
    skipNextSpacing = true,
    skipPreviousSpacing = true
  }

  local attributesForLabels = {
    b = {font = Font.new(FontFaces.DIN_OT_BOLD, VALUE_FONT_SIZE)}
  }

  local valueLabel = DynamicTextArea.new(valueString, self:getWidth(), baseAttributes, attributesForLabels, Alignment.CENTER)
  self:insertAtCenter(valueLabel)
  self.valueLabel = valueLabel

  self:linkOpacityAttrToChildren()
end

local function initialize(self, responseHandler)
  self.value = INITIAL_VALUE
  self.responseHandler = responseHandler
  self.didAttempt = false
end

local function makeInteractable(self)
  local function onDown()
    if self.responseHandler then
      self.responseHandler(self.value)
    end
  end

  Button.makeButton(self, onDown, nil, nil)
end

local function reset(self)
  local currentOpacity = self:getOpacity()

  self:setColor(unpack(FACE_COLOR))
  self.valueLabel:setColor(unpack(LABEL_COLOR))

  self:setOpacity(currentOpacity)

  self.didAttempt = false
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresResponseButton:setAttemptedFlag(didAttempt)
  self.didAttempt = didAttempt
end

function ExpendituresResponseButton:hasBeenAttempted()
  return self.didAttempt
end

function ExpendituresResponseButton:getFadeInAction()
  local fadeInAction = AHSequence.new()

  if not self.didAttempt then
    local buttonFadeDelay = AHDelay.new(FADE_IN_DELAY_DURATION)
    local buttonFadeIn = AHOpacityTo.new(self, FADE_IN_DURATION, nil, 1.0)

    fadeInAction:addAction(buttonFadeDelay)
    fadeInAction:addAction(buttonFadeIn)
  end

  return fadeInAction
end

function ExpendituresResponseButton:getFadeAndResetAction()
  local fadeAction = AHOpacityTo.new(self, FADE_OUT_DURATION, nil, 0.0)
  local resetAction = function() reset(self) end

  local fadeAndResetColorAction = AHSequence.new(fadeAction, resetAction)

  return fadeAndResetColorAction
end

function ExpendituresResponseButton:setValue(value)
  updateValue(self, value)
end

function ExpendituresResponseButton:getCorrectAnimation()
  local scaleDown = AHScaleTo.new(self, SCALE_DOWN_DURATION, nil, SCALE_DOWN_AMOUNT)
  local scaleUp = AHScaleTo.new(self, SCALE_UP_DURATION, nil, 1.0)

  local scaleSequence = AHSequence.new(scaleDown, scaleUp)

  local buttonColorChangeAction = AHColorTo.new(self, COLOR_CHANGE_DURATION, nil, DESTINATION_BUTTON_COLOR_CORRECT)
  local labelColorChangeAction = AHColorTo.new(self.valueLabel, COLOR_CHANGE_DURATION, nil, DESINATION_LABEL_COLOR)

  local correctAnimation = AHSpawn.new(scaleSequence, buttonColorChangeAction, labelColorChangeAction)
  return correctAnimation
end

function ExpendituresResponseButton:getIncorrectAnimation(doColorChange)
  local incorrectAnimation = AHSpawn.new()

  if doColorChange then
    local buttonColorChangeAction = AHColorTo.new(self, COLOR_CHANGE_DURATION, nil, DESTINATION_BUTTON_COLOR_INCORRECT)
    incorrectAnimation:addAction(buttonColorChangeAction)
  end

  local shakeAction = ExpendituresRockShake.new(self,
                                                ExpendituresCommon.INCORRECT_ROCK_AMPLITUDE,
                                                ExpendituresCommon.INCORRECT_ROCK_DURATION,
                                                ExpendituresCommon.INCORRECT_ROCK_CYCLE_COUNT)

  incorrectAnimation:addAction(shakeAction)

  return incorrectAnimation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresResponseButton.newButton(width, height, responseHandler)
  local self = createButtonRectangle(width, height, FACE_COLOR)
  Mixin.mix(self, ExpendituresResponseButton)

  initialize(self, responseHandler)
  makeInteractable(self)

  self:linkOpacityAttrToChildren()

  return self
end

function ExpendituresResponseButton.newBacking(width, height)
  local self = createButtonRectangle(width, height, BACKING_COLOR)

  return self
end
