--- ExpendituresRoundGenerator
-- @author Created by Jay Austria on October 20, 2014.

class("ExpendituresRoundGenerator")
require("data.generators.Generator")
require("ExpendituresCommon")
require("ExpendituresConceptBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONCEPT_DESCRIPTORS_FILENAME = "ExpendituresConceptDescriptors.json"

local function addToBucket(self, conceptDescriptor, bucketName)
  local bucket = self.buckets[bucketName]

  if bucket then
    table.insert(bucket, conceptDescriptor)
  else
    error("Tried to add concept to nonexistent bucket")
  end
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Note: concept descriptors can be eligibile for and reside in multiple buckets
local function fillBucketsForContentDifficulties(self, allConceptDescriptors)
  for _, conceptDescriptor in ipairs(allConceptDescriptors) do
    for _, contentDifficulty in pairs(ExpendituresCommon.ContentDifficulty) do
       if self.conceptBuilder:isEligible(conceptDescriptor, contentDifficulty) then
         addToBucket(self, conceptDescriptor, contentDifficulty)
       end
    end
  end
end

local function getAllConceptDescriptors()
  local descriptors = FileUtils.loadJson(CONCEPT_DESCRIPTORS_FILENAME)
  return descriptors
end

local function initialize(self, optionsCount)
  self.buckets =
  {
    [ExpendituresCommon.ContentDifficulty.EASY] = {},
    [ExpendituresCommon.ContentDifficulty.MEDIUM] = {},
    [ExpendituresCommon.ContentDifficulty.HARD] = {}
  }

  local allConcepts = getAllConceptDescriptors()

  local conceptBuilder = ExpendituresConceptBuilder.new(optionsCount)
  self.conceptBuilder = conceptBuilder

  fillBucketsForContentDifficulties(self, allConcepts)

  local difficultiesArray = { ExpendituresCommon.ContentDifficulty.EASY,
                              ExpendituresCommon.ContentDifficulty.MEDIUM,
                              ExpendituresCommon.ContentDifficulty.HARD }

  local difficultyKeyFrames = ExpendituresConfiguration.getDifficultyKeyFrames(difficultiesArray)
  self.difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())

  self.frequencyTypeMode = ExpendituresCommon.FrequencyType.WEEKLY
  self.generatedConcepts = {}
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresRoundGenerator:reset()
end

function ExpendituresRoundGenerator:generateNextValue()
  local contentDifficulty = self.difficultyNameChooser:nextValue()

  local bucketToSelectFrom = self.buckets[contentDifficulty]

  local function filterByFrequencyType(descriptor)
    return self.conceptBuilder:isMatchingFrequencyType(descriptor, self.frequencyTypeMode)
  end

  local bucketFilteredWithFrequencyType = TableUtils.getFilteredArray(bucketToSelectFrom, filterByFrequencyType)

  if #bucketFilteredWithFrequencyType == 0 then
    error("Ran out of concepts when trying to select from "..self.frequencyTypeMode)
  end

  local randomConceptDescriptor = TableUtils.getRandomElement(bucketFilteredWithFrequencyType)

  local function filterIsDifferentConcept(descriptor)
    local isSameDescriptor = (descriptor == randomConceptDescriptor)
    local isSameItemType = (descriptor.item == randomConceptDescriptor.item)

    return (not isSameDescriptor) and (not isSameItemType)
  end

  for k, bucket in pairs(self.buckets) do
    -- Only keep descriptors if they are different than the descriptor just used
    local unusedDescriptors = TableUtils.getFilteredArray(bucket, filterIsDifferentConcept)
    self.buckets[k] = unusedDescriptors
  end

  local gameConcept = self.conceptBuilder:generateUniqueConcept(randomConceptDescriptor, contentDifficulty, self.generatedConcepts)
  table.insert(self.generatedConcepts, gameConcept)

  return gameConcept
end

function ExpendituresRoundGenerator:setFrequencyTypeMode(mode)
  self.frequencyTypeMode = mode
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function ExpendituresRoundGenerator.new(optionsCount)
  local self = Generator.new()
  Mixin.mix(self, ExpendituresRoundGenerator)

  initialize(self, optionsCount)

  return self
end
