--- ExpendituresScene
-- @author Created by Jay Austria on October 27, 2014.

class("ExpendituresScene")
require("actions.advanced.Shake")

require("ExpendituresFrequencyDisplay")
require("ExpendituresGrid")
require("ExpendituresHelper")
require("ExpendituresIconStage")
require("ExpendituresOutroLayer")
require("ExpendituresResponseButton")
require("ExpendituresWindowBorder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BG_COLOR = ColorUtils.colorHex24ToStd(0x2006A0)
local BG_GRADIENT_TOP = ColorUtils.colorHex24ToStd(0x000714)

local COMBINED_TOP_MIDDLE_HEIGHT = Screen.safeHeight * 0.53

local WINDOW_TOP_WIDTH = Screen.width
local WINDOW_TOP_HEIGHT = Screen.safeHeight * 0.1

local WINDOW_MIDDLE_WIDTH = Screen.width
local WINDOW_MIDDLE_HEIGHT = COMBINED_TOP_MIDDLE_HEIGHT - WINDOW_TOP_HEIGHT + Screen.safeTopInset

local WINDOW_BOTTOM_WIDTH = Screen.width
local WINDOW_BOTTOM_HEIGHT = Screen.getSafeY(0.35)

local WINDOW_HOLE_WIDTH = Screen.safeWidth * 0.76
local WINDOW_HOLE_HEIGHT = Screen.height - WINDOW_TOP_HEIGHT - WINDOW_MIDDLE_HEIGHT - WINDOW_BOTTOM_HEIGHT

local WINDOW_BORDER_COLOR = ExpendituresCommon.Color.CYAN
local WINDOW_BORDER_LOCATION_X = Screen.width * 0.5
local WINDOW_BORDER_LOCATION_Y = WINDOW_BOTTOM_HEIGHT

local FREQUENCY_DISPLAY_LOCATION_X = Screen.width * 0.5
local FREQUENCY_DISPLAY_LOCATION_Y = WINDOW_BORDER_LOCATION_Y + (WINDOW_HOLE_HEIGHT * 0.5)

local STAGE_LOCATION_X = Screen.width * 0.50
local STAGE_LOCATION_Y = Screen.height * 0.65

local WINDOW_EXPAND_DELAY_DURATION = 0.5

local BOTTOM_UI_GROUP_WIDTH = Screen.safeWidth
local BOTTOM_UI_GROUP_HEIGHT = Screen.height * 0.24
local BOTTOM_UI_GROUP_LOCATION_X = Screen.width * 0.5
local BOTTOM_UI_GROUP_LOCATION_Y = (BOTTOM_UI_GROUP_HEIGHT * 0.5) + Screen.safeBottom

local IS_ANNUALLY_LABEL_TEXT = Localizer.get("IS_ANNUALLY")
local IS_ANNUALLY_LABEL_COLOR = ExpendituresCommon.Color.CYAN
local IS_ANNUALLY_LABEL_LOCATION_X = Screen.width * 0.5
local IS_ANNUALLY_LABEL_LOCATION_Y = 0.5 * ((BOTTOM_UI_GROUP_LOCATION_Y + (BOTTOM_UI_GROUP_HEIGHT / 2)) + (FREQUENCY_DISPLAY_LOCATION_Y - (WINDOW_HOLE_HEIGHT / 2)))
local IS_ANNUALLY_LABEL_FONT = FontFaces.DIN_OT_BOLD
local IS_ANNUALLY_LABEL_FONT_SIZE = 15
local IS_ANNUALLY_LABEL_FADE_DELAY_DURATION = 1.45
local IS_ANNUALLY_LABEL_FADE_IN_DURATION = 0.35

local BUTTON_MARGIN = 9

local ICON_OFFSET_Y = 12

local BETWEEN_FREQUENCY_AND_PRICE_DELAY_DURATION = 0.25

local SCREEN_SHAKE_COUNT = 6
local SCREEN_SHAKE_AMPLITUDE = 2

local FADE_BUTTONS_DELAY_DURATION = 0.5

local Z_INDEX = {
  FREQUENCY = 10,
  WINDOW = 20,
  GRID = 30,
  STAGE = 40,
  WINDOW_BORDER = 50,
  INSTRUCTIONS = 60,
  UI = 70,
  HELPER = 80
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function presentIcon(self, concept)
  local icon = concept.icon

  self.stage:getPresentIconAnimation(icon)
end

local function displayConceptPrices(self, answer, distractors, attemptedResponses)
  assert(answer and (#distractors == 3), "Must provide one answer and three distractors")

  local prices = {answer, unpack(distractors)}
  table.sort(prices)

  for buttonIndex, button in ipairs(self.responseButtons) do
    local price = prices[buttonIndex]
    button:setValue(price)

    if TableUtils.contains(attemptedResponses, price) then
      button:setAttemptedFlag(true)
      button:setTouchEnabled(false)
    end
  end
end

local function getButtonWithValue(self, response)
  for _, button in ipairs(self.responseButtons) do
    if button.value == response then
      return button
    end
  end

  error("No response button with value found")
  return nil
end

local function createSpawnOfButtonActions(self, getterName)
  local spawnAction = AHSpawn.new()
  for _, button in ipairs(self.responseButtons) do
    local buttonAction = button[getterName](button)
    spawnAction:addAction(buttonAction)
  end

  return spawnAction
end

local function getFadeInButtonsAnimation(self)
  local fadeInButtons = createSpawnOfButtonActions(self, "getFadeInAction")
  return fadeInButtons
end

local function fadeInButtons(self)
  local fadeInAction = getFadeInButtonsAnimation(self)

  return fadeInAction:start()
end

local function getFadeOutAndResetButtonsAnimation(self)
  local fadeOutAndResetButtons = createSpawnOfButtonActions(self, "getFadeAndResetAction")
  return fadeOutAndResetButtons
end

local function getGameLayerMoveDistanceForHelper(self)
  local topOfHelper = ExpendituresCommon.HELPER_HEIGHT
  local bottomOfFrequencyDisplay = self.frequencyDisplay:getYBottom()
  local moveDistance = topOfHelper - bottomOfFrequencyDisplay

  local bufferAmount = self:getHeightPart(0.036)
  local clearenceOfIcon = self:getHeight() - self.currentConcept.icon:getYTop() + moveDistance
  if (clearenceOfIcon < (bufferAmount * 2)) then
    bufferAmount = clearenceOfIcon / 2
  end

  return (moveDistance + bufferAmount)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ExpendituresScene:waitForOutroAnimation()
  local outroLayerAnimation = self.outroLayer:getAnimation()

  local outroAnimation = AHSpawn.new(outroLayerAnimation)

  wait(outroAnimation:start())
end

function ExpendituresScene:fadeOutAndResetButtons()
  local fadeAndReset = getFadeOutAndResetButtonsAnimation(self)
  fadeAndReset:start()
end

function ExpendituresScene:waitForCorrectAnimation(response, willShowProgress, progressValue)
  if willShowProgress then
    self.stage:showProgress(progressValue)
  else
    self.stage:hideProgress()
  end

  local correctButton = getButtonWithValue(self, response)

  local buttonAnimation = correctButton:getCorrectAnimation()
  local fadeButtonsDelay = AHDelay.new(FADE_BUTTONS_DELAY_DURATION)
  local fadeButtonsAnimation = getFadeOutAndResetButtonsAnimation(self)
  local fadeButtonsSequence = AHSequence.new(fadeButtonsDelay, fadeButtonsAnimation)

  local correctIcon = self.currentConcept.icon
  local iconAnimation = self.stage:getCorrectAnimation(correctIcon)

  local correctAnimation = AHSpawn.new(fadeButtonsSequence, buttonAnimation, iconAnimation)

  wait(correctAnimation:start())
end

function ExpendituresScene:waitForIncorrectAnimation(response, willShowProgress)
  if not willShowProgress then
    self.stage:hideProgress()
  end

  local incorrectButton = getButtonWithValue(self, response)

  local buttonAnimations = AHSpawn.new()

  for _, button in ipairs(self.responseButtons) do
    local doColorChange = (button == incorrectButton)
    local buttonAction = button:getIncorrectAnimation(doColorChange)
    buttonAnimations:addAction(buttonAction)
  end

  local incorrectIcon = self.currentConcept.icon
  local iconAnimation = self.stage:getIncorrectAnimation(incorrectIcon)

  local windowBorderAnimation = ExpendituresRockShake.new(self.windowBorder,
                                                          ExpendituresCommon.INCORRECT_ROCK_AMPLITUDE,
                                                          ExpendituresCommon.INCORRECT_ROCK_DURATION,
                                                          ExpendituresCommon.INCORRECT_ROCK_CYCLE_COUNT)

  local frequencyDisplayShake = ExpendituresRockShake.new(self.frequencyDisplay,
                                                          ExpendituresCommon.INCORRECT_ROCK_AMPLITUDE,
                                                          ExpendituresCommon.INCORRECT_ROCK_DURATION,
                                                          ExpendituresCommon.INCORRECT_ROCK_CYCLE_COUNT)

  local screenShake = AHRotationShake.newScreenShake(SCREEN_SHAKE_COUNT, SCREEN_SHAKE_AMPLITUDE)

  local gridIncorrectAnimation = self.gridLayer:getIncorrectAnimation()

   -- Protect from some abuse by making incorrect sequences annoying
  local incorrectAnimationTotalDelay = AHDelay.new(1.2)

  local incorrectAnimation = AHSpawn.new(buttonAnimations, windowBorderAnimation, frequencyDisplayShake, screenShake, iconAnimation, gridIncorrectAnimation, incorrectAnimationTotalDelay)

  wait(incorrectAnimation:start())
end

function ExpendituresScene:waitForRejectAnimation(doButtonFade)
  local rejectAnimation = AHSpawn.new()

  local incorrectIcon = self.currentConcept.icon
  local slideAwayIcon = self.stage:getRejectAnimation(incorrectIcon)
  rejectAnimation:addAction(slideAwayIcon)

  if doButtonFade then
    local fadeButtonsAnimation = getFadeOutAndResetButtonsAnimation(self)
    rejectAnimation:addAction(fadeButtonsAnimation)
  end

  wait(rejectAnimation:start())
end

function ExpendituresScene:setEnableInteraction(isEnabled)
  for _, button in ipairs(self.responseButtons) do
    if isEnabled and not button:hasBeenAttempted() then
      button:setTouchEnabled(true)
    else
      button:setTouchEnabled(false)
    end
  end
end

function ExpendituresScene:playIntroAnimation(concepts)
  local introAnimation = AHSpawn.new()

  local icons = {}
  for _, concept in ipairs(concepts) do
    local icon = ExpendituresIcon.new(concept.itemName, concept.price)
    icon:setLoc(ExpendituresCommon.ICON_OFFSCREEN_LOCATION_X, STAGE_LOCATION_Y - self.stage:getHeightPart(0.5) + ICON_OFFSET_Y)
    self.gameLayer:insert(icon, Z_INDEX.UI)
    table.insert(icons, icon)

    concept.icon = icon
  end

  local introSlideAnimation = self.stage:getIntroAnimation(icons)
  introAnimation:addAction(introSlideAnimation)

  local borderExpandDelay = AHDelay.new(WINDOW_EXPAND_DELAY_DURATION)
  local expandWindowAnimation = self.windowBorder:getExpandAnimation()

  local expandWindowSequence = AHSequence.new(borderExpandDelay, expandWindowAnimation)
  introAnimation:addAction(expandWindowSequence)

  local isAnnuallyLabelFadeDelay = AHDelay.new(IS_ANNUALLY_LABEL_FADE_DELAY_DURATION)
  local isAnnuallyLabelFadeIn = AHOpacityTo.new(self.isAnnuallyLabel, IS_ANNUALLY_LABEL_FADE_IN_DURATION, nil, 1.0)
  local isAnnuallyLabelIntroSequence = AHSequence.new(isAnnuallyLabelFadeDelay, isAnnuallyLabelFadeIn)
  introAnimation:addAction(isAnnuallyLabelIntroSequence)

  return introAnimation:start()
end

function ExpendituresScene:presentConcept(concept)
  self.frequencyDisplay:setFrequency(concept.frequency, concept.frequencyType)

  local function delayedPresent()
    waitFor(BETWEEN_FREQUENCY_AND_PRICE_DELAY_DURATION)
    presentIcon(self, concept)
    displayConceptPrices(self, concept.answer, concept.distractors, concept.attemptedResponses)
    self.currentConcept = concept

    wait(fadeInButtons(self))
    self.moveDistanceForHelper = getGameLayerMoveDistanceForHelper(self)
    self:setEnableInteraction(true)
  end
  runInCoroutine(delayedPresent)
end

function ExpendituresScene:setResponseHandler(handler)
  self.responseHandler = handler
end

function ExpendituresScene:waitForShowHelper(price, frequency, frequencyType)
  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0.6, 0, 0.4, 1})
  local moveDistance = self.moveDistanceForHelper
  local moveLayerUp = AHLocationYBy.new(self.gameLayer, duration, ease, moveDistance)
  local isAnnuallyLabelFadeOut = AHOpacityTo.new(self.isAnnuallyLabel, duration / 2, nil, 0.0)
  local animation = AHSpawn.new(moveLayerUp, isAnnuallyLabelFadeOut)

  local helper = ExpendituresHelper.new(price, frequency, frequencyType)
  self.gameLayer:insertAtCenter(helper, Z_INDEX.HELPER)
  helper:setY(ExpendituresCommon.HELPER_HEIGHT - helper:getHeightPart(0.5) - moveDistance)
  self.currentHelper = helper

  animation:addAction(helper:getIntroAction())
  animation:start()
end

function ExpendituresScene:waitForHideHelper()
  waitUntil(function() return self.currentHelper:returnToGame() end)
  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0.6, 0, 0.4, 1})
  local moveDistance = self.moveDistanceForHelper
  local moveLayerDown = AHLocationYBy.new(self.gameLayer, duration, ease, -moveDistance)
  local isAnnuallyLabelFadeIn = AHDelayed.new(0.25, AHOpacityTo.new(self.isAnnuallyLabel, duration, nil, 1.0))
  local currentHelperHideAction = self.currentHelper:getHideAction()
  local animation = AHSequence.new(
    AHSpawn.new(moveLayerDown, isAnnuallyLabelFadeIn, currentHelperHideAction),
    AHRemove.new(self.currentHelper))

  return animation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local gameLayer = Group.newFullScreen()
  self:insertAtCenter(gameLayer)
  self.gameLayer = gameLayer

  local frequencyDisplay = ExpendituresFrequencyDisplay.new(WINDOW_HOLE_WIDTH, WINDOW_HOLE_HEIGHT)
  frequencyDisplay:setLoc(FREQUENCY_DISPLAY_LOCATION_X, FREQUENCY_DISPLAY_LOCATION_Y)
  self.gameLayer:insert(frequencyDisplay, Z_INDEX.FREQUENCY)
  self.frequencyDisplay = frequencyDisplay

  local windowTop = FilledRectangle.new(WINDOW_TOP_WIDTH, WINDOW_TOP_HEIGHT, BG_GRADIENT_TOP)
  windowTop:setLoc(WINDOW_TOP_WIDTH * 0.5, Screen.height - WINDOW_TOP_HEIGHT * 0.5)
  self.gameLayer:insert(windowTop, Z_INDEX.WINDOW)

  local windowMiddle = RectangleGradient.new(WINDOW_MIDDLE_WIDTH, WINDOW_MIDDLE_HEIGHT, BG_GRADIENT_TOP, BG_COLOR)
  windowMiddle:setLoc(WINDOW_MIDDLE_WIDTH * 0.5, Screen.height - WINDOW_TOP_HEIGHT - WINDOW_MIDDLE_HEIGHT * 0.5)
  self.gameLayer:insert(windowMiddle, Z_INDEX.WINDOW)

  local windowBottom = FilledRectangle.new(WINDOW_BOTTOM_WIDTH, WINDOW_BOTTOM_HEIGHT, BG_COLOR)
  windowBottom:setLoc(WINDOW_BOTTOM_WIDTH * 0.5, WINDOW_BOTTOM_HEIGHT * 0.5)
  self.gameLayer:insert(windowBottom, Z_INDEX.WINDOW)

  local gridTextureCutOffY = windowMiddle:getYBottom()
  local gridLayer = ExpendituresGrid.new(gridTextureCutOffY)
  self.gameLayer:insertAtCenter(gridLayer, Z_INDEX.GRID)
  self.gridLayer = gridLayer

  local stage = ExpendituresIconStage.new()
  stage:setLoc(STAGE_LOCATION_X, STAGE_LOCATION_Y)
  self.gameLayer:insert(stage, Z_INDEX.STAGE)
  self.stage = stage

  local windowBorder = ExpendituresWindowBorder.newCollapsed(WINDOW_HOLE_WIDTH, WINDOW_HOLE_HEIGHT, WINDOW_BORDER_COLOR)
  windowBorder:setLoc(WINDOW_BORDER_LOCATION_X, WINDOW_BORDER_LOCATION_Y)
  self.gameLayer:insert(windowBorder, Z_INDEX.WINDOW_BORDER)
  self.windowBorder = windowBorder

  local isAnnuallyLabel = Label.new(IS_ANNUALLY_LABEL_TEXT, Font.new(IS_ANNUALLY_LABEL_FONT, IS_ANNUALLY_LABEL_FONT_SIZE))
  isAnnuallyLabel:setLoc(IS_ANNUALLY_LABEL_LOCATION_X, IS_ANNUALLY_LABEL_LOCATION_Y)
  isAnnuallyLabel:setColor(unpack(IS_ANNUALLY_LABEL_COLOR))
  isAnnuallyLabel:setOpacity(0)
  self.gameLayer:insert(isAnnuallyLabel, Z_INDEX.INSTRUCTIONS)
  self.isAnnuallyLabel = isAnnuallyLabel

  local bottomUIGroup = Group.new(BOTTOM_UI_GROUP_WIDTH, BOTTOM_UI_GROUP_HEIGHT)

  local bottomUIGroupY = BOTTOM_UI_GROUP_LOCATION_Y

  bottomUIGroup:setLoc(BOTTOM_UI_GROUP_LOCATION_X, bottomUIGroupY)
  self.gameLayer:insert(bottomUIGroup, Z_INDEX.UI)

  local numberOfHorizontalSpaces = 3
  local numberOfVerticalSpaces = 3
  local buttonColumnCount = 2
  local buttonRowCount = ExpendituresCommon.RESPONSE_OPTIONS_COUNT / buttonColumnCount
  local buttonWidth = (BOTTOM_UI_GROUP_WIDTH - (BUTTON_MARGIN * numberOfHorizontalSpaces)) / buttonColumnCount
  local buttonHeight = (BOTTOM_UI_GROUP_HEIGHT - (BUTTON_MARGIN * numberOfVerticalSpaces)) / buttonRowCount

  local buttonBackingGrid = Grid.new(buttonColumnCount, BUTTON_MARGIN, BUTTON_MARGIN)
  local responseButtonGrid = Grid.new(buttonColumnCount, BUTTON_MARGIN, BUTTON_MARGIN)

  bottomUIGroup:insertAtCenter(buttonBackingGrid)
  bottomUIGroup:insertAtCenter(responseButtonGrid)

  local totalButtonCount = buttonRowCount * buttonColumnCount

  local responseButtons = {}
  for backingIndex = 1, totalButtonCount do
    local backing = ExpendituresResponseButton.newBacking(buttonWidth, buttonHeight)
    buttonBackingGrid:insert(backing)

    local function responseWrapper(buttonValue)
      if self.responseHandler then
        self.responseHandler(buttonValue)
      end
    end

    local button = ExpendituresResponseButton.newButton(buttonWidth, buttonHeight, responseWrapper)
    button:setOpacity(0)
    button:setTouchEnabled(false)
    responseButtonGrid:insert(button)
    table.insert(responseButtons, button)
  end

  buttonBackingGrid:alignChildren()
  responseButtonGrid:alignChildren()

  self.responseButtonGrid = responseButtonGrid
  self.responseButtons = responseButtons

  local outroLayer = ExpendituresOutroLayer.new()
  self:insertAtCenter(outroLayer)
  self.outroLayer = outroLayer

  self.moveDistanceForHelper = nil
end

function ExpendituresScene.new()
  local self = FilledRectangle.newFullScreen(BG_COLOR)
  Mixin.mix(self, ExpendituresScene)

  createLayout(self)

  return self
end
