--- FoodAnswerOptionButton
-- @author Created by Aaron Barrett on November 26, 2014.

class("FoodAnswerOptionButton")

require("FoodRoundColors")
require("FoodItem")
require("FoodFractionLabel")
require("FoodPieFraction")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.getSafeWidthPart(0.20)
local HEIGHT = Screen.getSafeHeightPart(0.14)

FoodAnswerOptionButton.ICON_X_OFFSET = WIDTH * 0.5
FoodAnswerOptionButton.ICON_Y_OFFSET = HEIGHT * 0.65

local Z_ICON = 10
local Z_LABEL = 20
local Z_HIGHLIGHT = 30

local ICON_NAMES_BY_MIXED_FORM = {
  ["1"] = "answer_option_whole.png",
  ["1/2"] = "answer_option_half.png",
  ["1/3"] = "answer_option_third.png",
  ["1/4"] = "answer_option_fourth.png",
  ["1/6"] = "answer_option_sixth.png",
  ["1/8"] = "answer_option_eighth.png"
}

local TUTORIAL_COLOR = {0.6, 0.6, 0.6}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function morphColor(self, color)
  local fadeIconAction = AHColorTo.new(self.icon, FoodConstants.TUTORIAL_MORPH_DURATION_SHORT, nil, color)
  local fadeLabelAction = AHColorTo.new(self.label, FoodConstants.TUTORIAL_MORPH_DURATION_SHORT, nil, color)

  local fadeSpawn = AHSpawn.new(fadeIconAction, fadeLabelAction)

  local colorInterrupt = self.colorInterrupt
  colorInterrupt:insert(fadeSpawn)
end

local function createSprite(mixedForm)
  local iconName = ICON_NAMES_BY_MIXED_FORM[mixedForm]
  assert(iconName ~= nil, "No answer option button icon image for amount: `" .. tostring(mixedForm) .. "`")
  local icon = Sprite.newWithName(iconName)
  icon:setLoc(FoodAnswerOptionButton.ICON_X_OFFSET, FoodAnswerOptionButton.ICON_Y_OFFSET)

  return icon
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodAnswerOptionButton:setEnabledForTutorial(isEnabled)
  self:setTouchEnabled(isEnabled)

  local destinationColor = nil

  if isEnabled then
    destinationColor = self.roundColor
  else
    destinationColor = TUTORIAL_COLOR
  end

  morphColor(self, destinationColor)
end

function FoodAnswerOptionButton:getAmount()
  return self.amount
end

function FoodAnswerOptionButton:stayGrayOnTap()
  self.doesReturnToColor = false
end

-- Wrong Animation --------------------

function FoodAnswerOptionButton:playWrong()
  local icon = self.icon
  local label = self.label
  local roundColor = self.roundColor

  local redRemainDuration = 1.55

  local ease = AHEaseSimpleOut.new(1.4)

  local wrongAction = AHSequence.new(
                                     AHSpawn.new(
                                                 AHColorTo.new(icon, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                                 AHColorTo.new(label, 0.1, nil, FoodConstants.WRONG_RED_COLOR)
                                                 ),
                                     AHSpawn.new(
                                                 AHColorTo.new(icon, redRemainDuration, ease, roundColor),
                                                 AHColorTo.new(label, redRemainDuration, ease, roundColor)
                                                 )
                                     )
  return wrongAction:start()
end

function FoodAnswerOptionButton:highlight()
  local highlight = FoodPieFraction.new(self.amount, {0.01, 0.78, 0.98}, 0.5, 1)
  highlight:setEffectiveDims(self.icon:getEffectiveDims())
  highlight:setLoc(self.icon:getLoc())
  self:insert(highlight, Z_HIGHLIGHT)
  self.highlightSprite = highlight
end

function FoodAnswerOptionButton:removeHighlight()
  if self.highlightSprite then
    AHSequence.new(
      self.highlightSprite:createFadeAction(0.1, nil, 0),
      AHRemove.new(self.highlightSprite)
    ):start()
  end
end

function FoodAnswerOptionButton:getSprite()
  local sprite = FilledCircle.new(self.icon:getWidthPart(0.45), {1,1,1})
  sprite:setLoc(FoodAnswerOptionButton.ICON_X_OFFSET, FoodAnswerOptionButton.ICON_Y_OFFSET)
  return sprite
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildIcon(self, answerOptionAmount, roundColor)
  local mixedForm = answerOptionAmount:getMixedForm()
  local icon = createSprite(mixedForm)
  icon:setColor(unpack(roundColor))
  self:insert(icon, Z_ICON)
  self.icon = icon
end

local function buildLabel(self, answerOptionAmount, roundColor)
  local labelFontSize = 20.0
  local label = FoodFractionLabel.new(labelFontSize, answerOptionAmount)
  label:setColor(unpack(roundColor))
  label:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.24))
  self:insert(label, Z_LABEL)
  self.label = label
end

local function createOnTouchDownCallback(self, onPressedCallback)
  local function onTouchDown()
    local icon = self.icon
    local label = self.label
    local colorInterrupt = self.colorInterrupt

    local colorAction = AHSpawn.new(
                                    AHColorTo.new(icon, 0.1, nil, {1.0, 1.0, 1.0}),
                                    AHColorTo.new(label, 0.1, nil, {1.0, 1.0, 1.0})
                                    )
    colorInterrupt:insert(colorAction)

    onPressedCallback(self)
  end
  return onTouchDown
end

local function createOnTouchUpCallback(self)
  local function onTouchUp()
    local icon = self.icon
    local label = self.label
    local roundColor = self.roundColor
    local colorInterrupt = self.colorInterrupt

    if self.doesReturnToColor then
      local colorAction = AHSpawn.new(
                                      AHColorTo.new(icon, 0.5, nil, roundColor),
                                      AHColorTo.new(label, 0.5, nil, roundColor)
                                      )
      colorInterrupt:insert(colorAction)
    end
  end
  return onTouchUp
end

-- Constructor ------------------------

function FoodAnswerOptionButton.new(answerOptionAmount, roundColor, onPressedCallback)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, FoodAnswerOptionButton)

  buildIcon(self, answerOptionAmount, roundColor)
  buildLabel(self, answerOptionAmount, roundColor)

  self.colorAction = nil
  self.amount = answerOptionAmount
  self.roundColor = roundColor
  self.doesReturnToColor = true
  self.colorInterrupt = ActionInterrupt.new()
  self.onPressedCallback = onPressedCallback

  self.onTouchDown = createOnTouchDownCallback(self, onPressedCallback)
  self.onTouchUp = createOnTouchUpCallback(self)

  Button.makeButton(self, self.onTouchDown, self.onTouchUp, self.onTouchUp)

  return self
end
