--- FoodDot
-- @author Created by Aaron Barrett on January 05, 2015.

class("FoodDot")

require("FoodConstants")
require("FoodRoundColors")
require("FoodFractionLabel")
require("graphics.CircleGradient")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_TRAIL_GLOW = 10
local Z_TRAIL = 20

local Z_HEAD = 30

local Z_GLOW = 10
local Z_LABEL = 20
local Z_FILLING = 30
local Z_OUTLINE = 40
local Z_PERFECT_ANSWER_GLOW = 50

local LABEL_COLOR = {1.0, 1.0, 1.0}
local LABEL_FONT_SIZE = 30.0

local DOT_SMALL_SCALE = 0.16
local DOT_RADIUS = 42
local DOT_LINE_WIDTH_FULL = 2
local DOT_LINE_WIDTH_SMALL = 6

local CENTER_Y = Screen.getSafeY(0.5)
local TUTORIAL_DOT_COLOR = {0.7, 0.7, 0.7}

local SHADOW_DOT_COUNT = 6
local BASE_SHADOW_DOT_OPACITY = 0.35
local SHADOW_DOT_OPACITY_DIFFERENCE = BASE_SHADOW_DOT_OPACITY / SHADOW_DOT_COUNT

local INTRO_SCALE = 0.45
local CIRCLE_POINT_COUNT = 30


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Compensate for Moai ----------------

function FoodDot:getWorldLocationFromParent(parent)
  local x, y = self:getLoc()
  local worldX, worldY = parent:modelToWorld(x, y)
  return worldX, worldY
end

-- Accessors --------------------------

function FoodDot:getRoundColor()
  return self.roundColor
end

-- FoodProgressIndicator Animations ---

function FoodDot:playGrow(destinationX)
  local head = self.head
  local outline = self.outline
  local label = self.label
  local roundColor = self.roundColor

  local ease = AHEaseSimpleInOut.new(4.0)

  local growAction = AHDelayed.new(
                                   0.6,
                                   AHSpawn.new(
                                               AHLocationXTo.new(self, 0.8, AHEaseBackInOut.new(0.4), destinationX),
                                               AHScaleTo.new(head, 0.8, ease, 1.0),
                                               self.outline:newLineWidthToAction(1.0, AHEaseExponentialInOut.new(4.0), DOT_LINE_WIDTH_FULL),
                                               AHColorTo.new(outline, 0.8, nil, roundColor),
                                               AHOpacityTo.new(label, 0.8, ease, 1.0)
                                               )
                                   )
  return growAction:start()
end

function FoodDot:playShrink(destinationX)
  local head = self.head
  local roundColor = self.roundColor
  local filling = self.filling

  filling:setColor(unpack(roundColor))
  filling:setOpacity(0.0)

  local shrinkAction = AHDelayed.new(
                                     0.48,
                                     AHSpawn.new(
                                                 AHScaleTo.new(head, 0.5, AHEaseBackIn.new(1.8), DOT_SMALL_SCALE),
                                                 self.outline:newLineWidthToAction(0.5, nil, DOT_LINE_WIDTH_SMALL),
                                                 AHDelayed.new(0.05, AHOpacityTo.new(filling, 0.1, AHEaseSimpleOut.new(2.0), 1.0)),
                                                 AHLocationXTo.new(self, 1.2, AHEaseBackInOut.new(1.6), destinationX)
                                                 )
                                     )
  return shrinkAction:start()
end

function FoodDot:playAdoptRoundColor(currentRoundColor)
  local outline = self.outline
  local changeToNewRoundColorAction = AHDelayed.new(
                                                    0.8,
                                                    AHColorTo.new(outline, 0.8, nil, currentRoundColor)
                                                    )
  return changeToNewRoundColorAction:start()
end

function FoodDot:playMorphToColor(color)
  local fadeAction = AHColorTo.new(self.outline, FoodConstants.TUTORIAL_MORPH_DURATION, nil, color)
  return fadeAction:start()
end

function FoodDot:playMorphToOwnColor()
  local roundColor = self.roundColor
  return self:playMorphToColor(roundColor)
end

function FoodDot:playMorphToTutorialColor()
  return self:playMorphToColor(TUTORIAL_DOT_COLOR)
end

-- Intro Animation --------------------

function FoodDot:playFlyToCenter()
  local head = self.head
  local filling = self.filling
  local glow = self.glow
  local trailGlow = self.trailGlow
  local trail = self.trail

  local flyInAction = AHSequence.new(
                                     AHSpawn.new(
                                                 AHLocationYTo.new(self, 0.4, AHEaseExponentialOut.new(), CENTER_Y),
                                                 AHScaleYTo.new(head, 0.4, AHEaseExponentialOut.new(), INTRO_SCALE),
                                                 AHOpacityTo.new(glow, 0.4, AHEaseSimpleOut.new(1.5), 0.0),
                                                 AHOpacityTo.new(trailGlow, 0.4, AHEaseSimpleOut.new(1.5), 0.0),
                                                 AHOpacityTo.new(trail, 0.4, AHEaseSimpleOut.new(1.5), 0.0),
                                                 AHScaleXTo.new(trail, 0.4, AHEaseSimpleInOut.new(2.0), INTRO_SCALE),
                                                 AHScaleYTo.new(trail, 0.4, AHEaseSimpleOut.new(1.5), 0.0)
                                                 ),
                                     AHOpacityTo.new(filling, 0.5, AHEaseSimpleInOut.new(2.0), 0.0)
                                     )
  return flyInAction:start()
end

function FoodDot:playFlyToDestination(destinationX, destinationY, destinationRoundColor, shouldShrink)
  local head = self.head
  local filling = self.filling
  local glow = self.glow
  local outline = self.outline
  local label = self.label

  local chargeUpDuration = 0.4
  local intoPlaceDuration = 1.3

  local easeInOut = AHEaseExponentialInOut.new()
  local easeIn = AHEaseSimpleIn.new(2.0)
  local easeOut = AHEaseSimpleOut.new(2.0)

  local chargeUpAction = AHSpawn.new(
                                     AHOpacityTo.new(filling, chargeUpDuration, easeIn, 1.0),
                                     AHOpacityTo.new(glow, chargeUpDuration, easeIn, 1.0)
                                     )

  local intoPlaceAction = AHSpawn.new(
                                      AHLocationTo.new(self, intoPlaceDuration, easeInOut, destinationX, destinationY),
                                      AHOpacityTo.new(glow, intoPlaceDuration, easeInOut, 0.0)
                                      )

  if (shouldShrink) then
    intoPlaceAction:addAction(AHScaleTo.new(head, intoPlaceDuration, easeInOut, DOT_SMALL_SCALE))
    intoPlaceAction:addAction(self.outline:newLineWidthToAction(intoPlaceDuration, nil, DOT_LINE_WIDTH_SMALL))
    intoPlaceAction:addAction(AHColorTo.new(filling, intoPlaceDuration, easeIn, destinationRoundColor))
    intoPlaceAction:addAction(AHColorTo.new(outline, intoPlaceDuration, easeIn, destinationRoundColor))
    intoPlaceAction:addAction(AHColorTo.new(glow, intoPlaceDuration, easeIn, destinationRoundColor))
    intoPlaceAction:addAction(AHOpacityTo.new(filling, intoPlaceDuration, easeIn, 0.0))
    intoPlaceAction:addAction(AHOpacityTo.new(label, 0.1, easeOut, 0.0))
  else
    intoPlaceAction:addAction(AHScaleTo.new(head, intoPlaceDuration, easeInOut, 1.0))
    intoPlaceAction:addAction(AHOpacityTo.new(filling, intoPlaceDuration, easeInOut, 0.0))
  end

  local flyAction = AHSequence.new(
                                   chargeUpAction,
                                   intoPlaceAction
                                   )

  return flyAction:start()
end

-- Outro Animation --------------------

function FoodDot:addShadowDots(parentToHoldDots, centrifuge, zIndex)
  local filling = self.filling

  local shadowDots = {}

  local initialX, initialY = self:getWorldLocationFromParent(centrifuge.spinner)

  for shadowDotIndex = 1, SHADOW_DOT_COUNT do
    local shadowDot = FilledCircle.new(DOT_RADIUS)
    shadowDot:setPointCount(CIRCLE_POINT_COUNT)
    shadowDot:setLoc(initialX, initialY)
    shadowDot:setScl(DOT_SMALL_SCALE)
    shadowDot:inheritColor(filling)
    local opacity = BASE_SHADOW_DOT_OPACITY - (shadowDotIndex - 1) * SHADOW_DOT_OPACITY_DIFFERENCE
    shadowDot:setOpacity(opacity)
    parentToHoldDots:insert(shadowDot, zIndex)
    table.insert(shadowDots, shadowDot)
  end

  runInCoroutine(function()
    while (true) do
      local firstShadowDot = shadowDots[1]
      local dotWorldX, dotWorldY = self:getWorldLocationFromParent(centrifuge.spinner)
      firstShadowDot:setLoc(dotWorldX, dotWorldY)
      coroutine.yield()
    end
  end)

  for shadowDotIndex = 2, SHADOW_DOT_COUNT do
    local shadowDot = shadowDots[shadowDotIndex]
    local previousDotIndex = shadowDotIndex - 1
    local previousDot = shadowDots[previousDotIndex]

    runInCoroutine(function()
      while (true) do
        local previousDotX, previousDotY = previousDot:getLoc()
        coroutine.yield()
        shadowDot:setLoc(previousDotX, previousDotY)
      end
    end)
  end

end

function FoodDot:playOutro()
  local label = self.label
  local head = self.head
  local filling = self.filling
  local glow = self.glow
  local roundColor = self.roundColor

  glow:setColor(unpack(roundColor))
  filling:setColor(unpack(roundColor))

  local labelAction = AHOpacityTo.new(label, 0.3, nil, 0.0)
  local scaleDownAction = AHScaleTo.new(head, INTRO_SCALE, nil, DOT_SMALL_SCALE)
  local fillingAction = AHOpacityTo.new(filling, 0.6, nil, 1.0)
  local glowAction = AHOpacityTo.new(glow, 0.6, nil, 0.20)
  local glowScaleAction = AHDelayed.new(0.5, AHScaleTo.new(glow, 3.1, AHEaseElasticOut.new(0.3), 1.6))

  local outroAction = AHSpawn.new(
                                  labelAction,
                                  scaleDownAction,
                                  fillingAction,
                                  glowAction,
                                  glowScaleAction
                                  )

  return outroAction:start()
end

-- Wrong Animation --------------------

function FoodDot:playWrong(colorToRevertTo)
  local outline = self.outline
  local filling = self.filling

  local redRemainDuration = 1.55

  local ease = AHEaseSimpleOut.new(1.4)

  local wrongAction = AHSequence.new(
                                     AHSpawn.new(
                                                 AHColorTo.new(outline, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                                 AHColorTo.new(filling, 0.1, nil, FoodConstants.WRONG_RED_COLOR)
                                                 ),
                                     AHSpawn.new(
                                                 AHColorTo.new(outline, redRemainDuration, ease, colorToRevertTo),
                                                 AHColorTo.new(filling, redRemainDuration, ease, colorToRevertTo)
                                                 )
                                     )
  wrongAction:start()
end

function FoodDot:playEfficientAnswer()
  local glowRadius = self:getWidthPart(0.165)
  local centerColor = self.glowColor
  local edgeColor = {0.0, 0.0, 0.0, 0.0}
  local blurMargin = 0.0
  local glow = CircleGradient.new(glowRadius, centerColor, edgeColor, blurMargin)
  glow:setBlendMode(MOAIProp2D.BLEND_ADD)

  local maximumWhiteOpacity = 0.5
  centerColor = {1.0, 1.0, 1.0, maximumWhiteOpacity}
  local whiteGlow = CircleGradient.new(glowRadius, centerColor, edgeColor, blurMargin)
  whiteGlow:setBlendMode(MOAIProp2D.BLEND_ADD)

  local delay = 1.9
  local showGlow = AHDelayed.new(delay, function()
    self:insertAtCenter(glow, Z_PERFECT_ANSWER_GLOW)
    self:insertAtCenter(whiteGlow, glow.zIndex + 1)
  end)

  local duration = 1.5
  local ease = nil
  local centerOpacity = 0.4
  local edgeOpacity = 0.0
  local fadeOut = glow:newOpacityToAction(duration, ease, centerOpacity, edgeOpacity)
  local fadeOutWhite = whiteGlow:newOpacityToAction(duration, ease, centerOpacity, edgeOpacity)

  centerOpacity = 1.0
  local fadeIn = glow:newOpacityToAction(duration, ease, centerOpacity, edgeOpacity)
  local fadeInWhite = whiteGlow:newOpacityToAction(duration, ease, maximumWhiteOpacity, edgeOpacity)

  local fadeInOutGlow = AHSequence.new(fadeOut, fadeIn)
  local fadeInOutWhite = AHSequence.new(fadeOutWhite, fadeInWhite)
  local fadeInOut = AHSpawn.new(fadeInOutGlow, fadeInOutWhite)
  local fadeLoop = AHLoop.new(fadeInOut)

  local showThenPulse = AHSequence.new(showGlow, fadeLoop)
  showThenPulse:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createTrailBeginColor(roundColor)
  local opacity = 0.5
  return ColorUtils.colorWithOpacityMultiplier(roundColor, opacity)
end

local function createTrailEndColor(roundColor)
  local opacity = 0.0
  return ColorUtils.colorWithOpacityMultiplier(roundColor, opacity)
end

-- Constructor ------------------------

function FoodDot.new(round, roundIndex)
  local roundColor = FoodRoundColors.getColorForRoundIndex(roundIndex)
  local glowColor = FoodRoundColors.getGlowColorForRoundIndex(roundIndex)

  local glow = Sprite.newWithName("dot_glow.png")
  glow:setColor(unpack(glowColor))

  local filling = FilledCircle.new(DOT_RADIUS - DOT_LINE_WIDTH_FULL, roundColor)
  filling:setPointCount(CIRCLE_POINT_COUNT)
  filling:setColor(unpack(roundColor))

  local outline = HollowCircle.new(DOT_RADIUS, roundColor, DOT_LINE_WIDTH_FULL)
  outline:setPointCount(CIRCLE_POINT_COUNT)
  outline:setColor(unpack(roundColor))

  local trailGlow = Sprite.newWithName("dot_trail_glow.png")
  trailGlow:setAnchorPointTop()
  trailGlow:setScl(1.0, 1.0)
  trailGlow:setColor(unpack(glowColor))
  trailGlow:setOpacity(0.65)

  local width, height = glow:getDims()

  local self = Group.new(width, height)
  Mixin.mix(self, FoodDot)

  self:insertAtCenter(trailGlow, Z_TRAIL_GLOW)

  local head = Group.new(width, height)
  self:insertAtCenter(head, Z_HEAD)

  head:insertAtCenter(glow, Z_GLOW)
  head:insertAtCenter(filling, Z_FILLING)
  head:insertAtCenter(outline, Z_OUTLINE)

  local trailWidth = outline:getWidthPart(INTRO_SCALE)
  local trailHeight = Screen.getHeightPart(3.75)
  local trailBeginColor = createTrailBeginColor(roundColor)
  local trailEndColor = createTrailEndColor(roundColor)
  local trail = RectangleGradient.new(trailWidth, trailHeight, trailBeginColor, trailEndColor)
  trail:setAnchorPointTop()
  self:insertAtCenter(trail, Z_TRAIL)

  local amount = round:getGivenAmount()
  local label = FoodFractionLabel.new(LABEL_FONT_SIZE, amount)
  label:setColor(unpack(LABEL_COLOR))
  head:insertAtCenter(label, Z_LABEL)

  head:setScl(INTRO_SCALE, 4.8)

  self.roundColor = roundColor
  self.glowColor = glowColor
  self.glow = glow
  self.filling = filling
  self.outline = outline
  self.trailGlow = trailGlow
  self.trail = trail
  self.head = head
  self.label = label

  return self
end
