--- FoodFraction
-- @author Created by Aaron Barrett on September 17, 2014.

class("FoodFraction")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TITLES_BY_MIXED_FORM = {
  ["1/2"] = "FIND ONE-HALF",
  ["1/3"] = "FIND ONE-THIRD",
  ["2/3"] = "FIND TWO-THIRDS",
  ["1/4"] = "FIND ONE-FOURTH",
  ["3/4"] = "FIND THREE-FOURTHS",
  ["1/6"] = "FIND ONE-SIXTH",
  ["1/8"] = "FIND ONE-EIGHTH",
  ["2"] = "DOUBLE IT",
  ["3"] = "TRIPLE IT"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodFraction:getNumerator()
  return self.numerator
end

function FoodFraction:getDenominator()
  return self.denominator
end

function FoodFraction:getReciprocal()
  local numerator = self.numerator
  local denominator = self.denominator

  local reciprocal = FoodFraction.new(denominator, numerator)
  return reciprocal
end

function FoodFraction:add(otherFraction)
  local numerator = self.numerator
  local denominator = self.denominator

  local otherNumerator = otherFraction:getNumerator()
  local otherDenominator = otherFraction:getDenominator()

  local summedNumerators = numerator * otherDenominator + otherNumerator * denominator
  local commonBase = denominator * otherDenominator

  local sum = FoodFraction.new(summedNumerators, commonBase)
  return sum
end

function FoodFraction:subtract(otherFraction)
  local numerator = self.numerator
  local denominator = self.denominator

  local otherNumerator = otherFraction:getNumerator()
  local otherDenominator = otherFraction:getDenominator()

  local differenceOfNumerators = numerator * otherDenominator - otherNumerator * denominator
  local commonBase = denominator * otherDenominator

  local difference = FoodFraction.new(differenceOfNumerators, commonBase)
  return difference
end

function FoodFraction:multiply(otherFraction)
  local numerator = self.numerator
  local denominator = self.denominator

  local otherNumerator = otherFraction:getNumerator()
  local otherDenominator = otherFraction:getDenominator()

  local productNumerator = numerator * otherNumerator
  local productDenominator = denominator * otherDenominator

  local product = FoodFraction.new(productNumerator, productDenominator)
  return product
end

function FoodFraction:divide(otherFraction)
  local reciprocal = otherFraction:getReciprocal()

  local product = self:multiply(reciprocal)
  return product
end

function FoodFraction:getMixedForm()
  local numerator = self.numerator
  local denominator = self.denominator

  local wholePart = math.floor(numerator / denominator)
  local remainder = (numerator - wholePart * denominator)

  local mixedForm = nil

  if (remainder > 0) then
    local fractionString = remainder .. "/" .. denominator

    if (wholePart > 0) then
      mixedForm = wholePart .. " " .. fractionString
    else
      mixedForm = fractionString
    end
  else
    mixedForm = tostring(wholePart)
  end

  return mixedForm
end

function FoodFraction:equals(otherFraction)
  return ((self.numerator == otherFraction:getNumerator())
          and (self.denominator == otherFraction:getDenominator()))
end

function FoodFraction:toNumber()
  return self.numerator / self.denominator
end

function FoodFraction:getTitle()
  local mixedForm = self:getMixedForm()
  local title = TITLES_BY_MIXED_FORM[mixedForm]
  if (title == nil) then
    if (self:toNumber() < 1.0) then
      title = "FIND " .. mixedForm
    else
      title = "MULTIPLY BY " .. mixedForm
    end
  end
  return title
end

-- Static API -------------------------

function FoodFraction.getComponentsFromString(fractionString)
  local wholePart = 0
  local numeratorPart = 0
  local denominatorPart = 1

  local parts = {}
  for part in fractionString:gmatch("%d+") do
    table.insert(parts, part)
  end

  local firstPart = parts[1]
  local secondPart = parts[2]
  local thirdPart = parts[3]

  if (firstPart ~= nil) then
    if (secondPart ~= nil) then
      if (thirdPart ~= nil) then
        -- Interpret as mixed fraction ("1 1/2")
        wholePart = tonumber(firstPart)
        numeratorPart = tonumber(secondPart)
        denominatorPart = tonumber(thirdPart)
      else
        -- Interpret as fraction ("1/2")
        numeratorPart = tonumber(firstPart)
        denominatorPart = tonumber(secondPart)
      end
    else
      -- Interpret as whole number ("1")
      wholePart = tonumber(firstPart)
    end
  else
    error("Invalid fraction format.  No numbers found in string `" .. tostring(fractionString) .. "`.")
  end

  return wholePart, numeratorPart, denominatorPart
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FoodFraction.new(numerator, denominator)
  local self = {}
  Mixin.mix(self, FoodFraction)

  local reducedNumerator, reducedDenominator = MathUtils.fractionLowestTerms(numerator, denominator)

  self.numerator = reducedNumerator
  self.denominator = reducedDenominator

  return self
end

function FoodFraction.newWithMixedForm(wholeNumber, numerator, denominator)
  local combinedNumerator = numerator + wholeNumber * denominator
  return FoodFraction.new(combinedNumerator, denominator)
end

function FoodFraction.newWithString(fractionString)
  local wholePart, numeratorPart, denominatorPart = FoodFraction.getComponentsFromString(fractionString)
  return FoodFraction.newWithMixedForm(wholePart, numeratorPart, denominatorPart)
end
