--- FoodGame
-- @author Created by Aaron Barrett on September 17, 2014.

class("FoodGame")

require("StandardUIBuilder")
require("FoodConfiguration")
require("widgets.TapToContinue")
require("FoodRound")
require("FoodProgressIndicator")
require("FoodPanelRound")
require("FoodPanelRecipe")
require("FoodConceptGenerator")
require("FoodOutro")
require("FoodMathTipOverlay")
require("FoodTapToRemoveTipOverlay")
require("FoodBackground")
require("bonuses.ResponseSpeedTracker")
require("tutorials.TutorialHighlighter")
require("tutorials.TutorialButton")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TITLE_LABEL_Y = Screen.getSafeY(0.89)
local PROGRESS_INDICATOR_Y = Screen.getSafeY(0.78)
local TAP_TO_REMOVE_Y_PADDING = 20

local UI_COMPONENTS_COLOR = {1.0, 1.0, 1.0}

local Z_BACKGROUND = 10
local Z_TITLE_LABEL = 20
local Z_PANEL = 30
local Z_PROGRESS_INDICATOR = 40
local Z_OUTRO = 50
local Z_TIP_LAYER = 60

local TIME_LIMIT_TO_SHOW_TIP = 7.0
local GUESS_LIMIT_TO_SHOW_TIP = 2

local function configureGameDifficulty(self)
  self.setsToComplete = FoodConfiguration.SETS_TO_COMPLETE_COUNT

  local rankTwoScore = FoodConfiguration.getRankTwoScore()
  local rankThreeScore = FoodConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local roundKeyFrames = FoodConfiguration.getRoundKeyFrames()
  local difficultyKeyFrames = FoodConfiguration.getDifficultyKeyFrames()
  local conceptGenerator = FoodConceptGenerator.new(roundKeyFrames, difficultyKeyFrames)

  conceptGenerator:setRandomRound()

  self.conceptGenerator = conceptGenerator

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.EFFICIENCY_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function endTutorial(self)
  self.isInTutorial = false
  self.roundPanel:endTutorial()
  self.background:displayForGame()
  self.progressIndicator:displayForGame()

  GameSession.notifyGameTimeStart()
end

local function enableTutorial(self)
  GameSession.notifyGameTimePause()
end

local function addTutorialOverlayInstructionalLabel(self, tutorialOverlay)
  local submitButton = self.roundPanel:getHolder():getSubmitButton()
  local _, instructionTopY = self:transformObjectLoc(submitButton)
  instructionTopY = instructionTopY - submitButton:getHeightPart(0.5)

  local answerBar = self.roundPanel:getAnswerBar()
  local _, instructionBottomY = self:transformObjectLoc(answerBar)
  instructionBottomY = instructionBottomY + answerBar:getHeightPart(0.5)

  local baseFontSize = 20
  local maxInstructionHeight = (instructionTopY - instructionBottomY)
  local largestFontSizeForComfortableFit = maxInstructionHeight / 4
  local instructionFontSize = math.min(baseFontSize, largestFontSizeForComfortableFit)

  local instructionLocationY = (instructionTopY + instructionBottomY) / 2

  tutorialOverlay:addInstructionLabel("Tap the shapes to build your answer", instructionFontSize, instructionLocationY)
end

local function showTutorialIfEligible(self, roundPanelAppearAction)
  local isFirstRound = (self.roundIndex == 1)
  local isEligible = isFirstRound and not self.roundPanel.hasPressedButton

  if (isEligible) then
    wait(roundPanelAppearAction)

    self.roundPanel:setTouchEnabled(false)

    if GameSession.hasNeverWon() then
      local tutorialLabelY = self.roundPanel:getHolder():getTopY() + Screen.getSafeHeightPart(0.04)
      self.background:displayForTutorial(tutorialLabelY)
    end

    enableTutorial(self)
    local tutorialStartDelayDuration = 0.5
    waitFor(tutorialStartDelayDuration)
    local optimalSteps = self.round:getOptimalSteps()

    local getButton = function(step)
      return self.roundPanel.answerOptionBar:getButton(step)
    end

    local item = getButton(optimalSteps[1])
    local createSpriteCallback = function()
      local icon = item:getSprite()
      local itemX, itemY = self:transformLoc(item)
      local iconX, iconY = icon:getLoc()

      local onTouchDown = item.onTouchDown
      local onTouchUp = item.onTouchUp
      local tutorialButton = TutorialButton.new(icon, onTouchDown, onTouchUp)
      tutorialButton:setLoc(itemX + iconX, itemY + iconY)
      return tutorialButton
    end

    local tutorialOverlay = TutorialHighlighter.newWithMasks( {createSpriteCallback},
                                                              nil,
                                                              0.3 )

    tutorialOverlay:setOverlayColor(0, 0, 0, 0)
    tutorialOverlay:setPulseSize(5)
    addTutorialOverlayInstructionalLabel(self, tutorialOverlay)

    self:insertAtCenter(tutorialOverlay, Z_TIP_LAYER)

    tutorialOverlay:createFadeInAnimation():start()

    for index, step in ipairs(optimalSteps) do
      item = getButton(step)
      item:highlight()
      tutorialOverlay:reinitialize()

      tutorialOverlay:waitForTargetTap()
      item:removeHighlight()
    end

    wait(AHDelay.new(0.5):start())
    item = self.roundPanel:getHolder():getSubmitButton()
    tutorialOverlay:resetHighlights({createSpriteCallback})
    item:highlight()
    tutorialOverlay:waitForTargetTap()
    item:removeHighlight()

    wait(tutorialOverlay:createFadeOutAnimation():start())
    tutorialOverlay:removeSelf()

    if GameSession.hasNeverWon() then
      self.background:displayForGame()
    end

    self.roundPanel:setTouchEnabled(true)
  end
end

local function isEligibleForMathTip(self)
  local hasSeenMathTipThisRound = self.hasSeenMathTipThisRound

  local responseSpeedTracker = self.responseSpeedTracker
  local didAnswerOverTimeLimit = (responseSpeedTracker:getCurrentTime() >= TIME_LIMIT_TO_SHOW_TIP)

  local didHitGuessLimit = (self.guessesCount == GUESS_LIMIT_TO_SHOW_TIP)

  local isEligible = ((not hasSeenMathTipThisRound) and
                      (didAnswerOverTimeLimit or didHitGuessLimit))
  return isEligible
end

local function isEligibleForTapToRemoveTip(self, wasAnswerTooHigh)
  local isCurrentlyShowingTip = (self.displayedTapToRemoveTip ~= nil)
  local isEligible = ((wasAnswerTooHigh) and
                      (not self.hasSeenTapToRemoveTip) and
                      (not isCurrentlyShowingTip) and
                      (not self.hasPerformedRemoval))
  return isEligible
end

local function showTapToRemoveTip(self)
  local tipOverlay = FoodTapToRemoveTipOverlay.new()
  self:insertAtCenter(tipOverlay, Z_TIP_LAYER)

  self.hasSeenTapToRemoveTip = true
  self.displayedTapToRemoveTip = tipOverlay

  local function keepPinnedAboveHolder()
    while (true) do
      local tipPaddedY = self.roundPanel:getHolder():getTopY() + TAP_TO_REMOVE_Y_PADDING
      local x = tipOverlay:getX()
      tipOverlay:setLoc(x, tipPaddedY)
      coroutine.yield()
    end
  end
  self.tapToRemovePositionUpdater = runInCoroutine(keepPinnedAboveHolder)
end

local function removeTapToRemoveTip(self)
  self.tapToRemovePositionUpdater:stop()
  self.tapToRemovePositionUpdater = nil
  self.displayedTapToRemoveTip:removeSelf()
  self.displayedTapToRemoveTip = nil
end

local function showMathTipAndWaitForDismissal(self)
  self.hasSeenMathTipThisRound = true

  self.roundPanel:setTouchEnabled(false)

  local tipType = self.round:getTipType()
  local tipOverlay = FoodMathTipOverlay.newHidden(tipType)
  self:insertAtCenter(tipOverlay, Z_TIP_LAYER)

  wait(tipOverlay:showAndWaitForDismissal())
end

local function getNotifyRemoveCallback(self)
  return function()
    self.hasPerformedRemoval = true

    if (self.displayedTapToRemoveTip) then
      removeTapToRemoveTip(self)
    end
  end
end

local function isAnswerTooHigh(submittedAmount, round)
  return (submittedAmount:toNumber() > round:getCorrectAnswer():toNumber())
end

local function isAnswerEfficient(submittedSteps, optimalSteps)
  return (#submittedSteps == #optimalSteps)
end

local function waitForCorrectReaction(self, isFirstAnswer, isLastRound)
  if (isFirstAnswer) then
    self.responseSpeedTracker:notifyCorrectResponse()
  end

  if (isLastRound) then
    local consistencyBonus = FoodConfiguration.getConsistencyBonus(self.roundsWithMistakesCount)
    if (consistencyBonus > 0) then
      ScoringUtils.applyCustomBonus(consistencyBonus, BonusTypes.TYPES.CONSISTENCY_BONUS, self.bonusTray)
    end
    local titleLabelAction = AHOpacityTo.new(self.titleLabel, 0.5, nil, 0.0)
    titleLabelAction:start()
  end

  self.roundPanel:playHideAnswerOptionButtons()

  local submittedSteps = TableUtils.shallowCopy(self.roundPanel:getSubmittedSteps())
  local optimalSteps = self.round:getOptimalSteps()

  local wasAnswerEfficient = isAnswerEfficient(submittedSteps, optimalSteps)

  if (wasAnswerEfficient) then
    local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.EFFICIENCY_BONUS)
    local bonusAmount = FoodConfiguration.POINTS_PER_EFFICIENCY_BONUS
    if (isFirstAnswer) then
      bonusAmount = FoodConfiguration.POINTS_PER_PERFECT_BONUS
      self.progressIndicator:markEfficientAnswer()
    end
    self.bonusTray:indicateSimpleTextBonus(notificationString, bonusAmount)
    GameSession.increaseScore(bonusAmount)
    GameSession.bonusReporter:incrementBonusCount(BonusTypes.TYPES.EFFICIENCY_BONUS)
  end

  local shouldShowStrobe = (isFirstAnswer and wasAnswerEfficient)
  if (shouldShowStrobe) then
    Sound.playSoundEffect("perfect")
    self.roundPanel:playStrobe()
  else
    Sound.playSoundEffect("correct")
  end

  GameSession.increaseScore(FoodConfiguration.POINTS_PER_CORRECT_RESPONSE)

  self.scoreLabel:setScore(GameSession.getScore())

  wait(self.roundPanel:playCorrect(submittedSteps, optimalSteps))
end

local waitForRoundInput = nil -- Forward declaration

local function waitForWrongReaction(self, submittedAmount, isLastRound)
  self.responseSpeedTracker:notifyIncorrectResponse()

  Sound.playSoundEffect("wrong")

  if (self.guessesCount == 0) then
    self.roundsWithMistakesCount = self.roundsWithMistakesCount + 1
  end

  local wasAnswerTooHigh = isAnswerTooHigh(submittedAmount, self.round)
  self.background:playWrong()
  self.progressIndicator:playWrong()
  wait(self.roundPanel:playWrong(wasAnswerTooHigh))

  self.guessesCount = self.guessesCount + 1
  if (isEligibleForMathTip(self)) then
    showMathTipAndWaitForDismissal(self)
  end

  if (isEligibleForTapToRemoveTip(self, wasAnswerTooHigh)) then
    showTapToRemoveTip(self)
  end

  self.roundPanel:allowRetry()

  local isFirstAnswer = false
  waitForRoundInput(self, isFirstAnswer, isLastRound)
end

waitForRoundInput = function(self, isFirstAnswer, isLastRound)  -- Implements forward declaration above
  self.roundPanel:setTouchEnabled(true)

  self.responseSpeedTracker:notifyStart()
  waitUntil(function() return self.roundPanel:getSubmittedAmount() ~= nil end)

  GameSession.reportSeenConcept(self.round:getIdentifier())

  if (self.isInTutorial) then
    endTutorial(self)
  end

  local conceptId = self.round.identifier
  local submittedAmount = self.roundPanel:getSubmittedAmount()
  local correctAnswer = self.round:getCorrectAnswer()
  local wasAnswerCorrect = correctAnswer:equals(submittedAmount)
  if (wasAnswerCorrect) then
    if (self.guessesCount == 0) then
      GameSession.contentTracker:notifyCorrectAnswer(conceptId)
      GameSession.incrementCorrectAnswerCount()
    end
    waitForCorrectReaction(self, isFirstAnswer, isLastRound)
  else
    local answerAsString = submittedAmount.numerator .. "/" .. submittedAmount.denominator
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, nil, answerAsString)
    GameSession.incrementIncorrectAnswerCount()
    waitForWrongReaction(self, submittedAmount, isLastRound)
  end
end

local function playTakeTitleLabel(self)
  local recipePanel = self.recipePanel
  local titleLabel = recipePanel:getTitleLabel()
  self.titleLabel = titleLabel

  self:takeChild(titleLabel, Z_TITLE_LABEL)

  local ease = AHEaseSineInOut.new();

  local titleLabelAction = AHDelayed.new(
                                         0.8,
                                         AHSpawn.new(
                                                     AHLocationYTo.new(titleLabel, 0.5, ease, TITLE_LABEL_Y),
                                                     AHOpacityTo.new(titleLabel, 0.5, ease, 0.78),
                                                     AHScaleTo.new(titleLabel, 0.5, ease, 0.5)
                                                     )
                                         )

  return titleLabelAction:start()
end

local function waitForBuildAndPlayOutro(self)
  GameSession.notifyGameTimePause()

  local progressIndicator = self.progressIndicator
  local titleLabel = self.titleLabel

  local outro = FoodOutro.new(progressIndicator)
  self:insertAtCenter(outro, Z_OUTRO)

  Sound.playSoundEffect("goal_reached")
  self:showDefaultGameWinOutro()

  local titleLabelAction = AHOpacityTo.new(titleLabel, 0.5, nil, 0.0)
  titleLabelAction:start()

  wait(outro:playOutro())
end

local function notifyBonusesAndWin(self)
  GameSession.notifyGameWin()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodGame:gameFlow()
  GameSession.notifyGameTimePause()
  wait(self.recipePanel:playFlyToCenter())
  playTakeTitleLabel(self)
  self.recipePanel:playFadeIngredientsLabel()
  wait(self.progressIndicator:playTakeDots())

  self.recipePanel:removeSelf()
  self.recipePanel = nil

  local maxHeightOfRoundPanel = PROGRESS_INDICATOR_Y - self.progressIndicator:getFocusedDot():getHeightPart(0.5)

  local rounds = self.rounds
  for roundIndex, round in ipairs(rounds) do
    GameSession.notifyGameTimeStart()
    self.roundIndex = roundIndex
    self.round = round

    self.roundPanel = FoodPanelRound.new(self.round, roundIndex, maxHeightOfRoundPanel)
    self.roundPanel:getHolder():setOnRemoveCallback(getNotifyRemoveCallback(self))
    self:insertAtCenter(self.roundPanel, Z_PANEL)

    local roundPanelAppearAction = self.roundPanel:playAppear()
    GameSession.contentTracker:notifyPresentedConcept(round:getIdentifier())

    if not GameSession.hasNeverWon() then
      self.roundPanel:setTouchEnabled(true)
    end

    local tutorialDelay = FoodConfiguration.TUTORIAL_DELAY
    if GameSession.hasNeverWon() then
      tutorialDelay = 0
    end

    local delayTimer = Group.newUnsized()
    Timer.makeTimer(delayTimer, tutorialDelay)
    delayTimer:start()
    waitUntil(function() return self.roundPanel.hasPressedButton or delayTimer:getTimeRemaining() <= 0 end)

    showTutorialIfEligible(self, roundPanelAppearAction)

    local isFirstAnswer = true
    local isLastRound = (roundIndex == #rounds)
    self.hasSeenMathTipThisRound = false
    self.guessesCount = 0
    waitForRoundInput(self, isFirstAnswer, isLastRound)

    wait(self.roundPanel:playDismissWithProgressIndicator(self.progressIndicator))
  end

  waitForBuildAndPlayOutro(self)

  notifyBonusesAndWin(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildRounds(self)
  local rounds = {}

  local numberOfRounds = FoodConfiguration.SETS_TO_COMPLETE_COUNT
  local concepts = self.conceptGenerator:nextAsArray(numberOfRounds)

  for _, concept in ipairs(concepts) do
    local round = FoodRound.newWithConcept(concept)
    table.insert(rounds, round)
  end

  self.rounds = rounds
end


local function createGameElements(self)
  local background = FoodBackground.new()
  self:insertAtCenter(background, Z_BACKGROUND)
  self.background = background

  local rounds = self.rounds
  local recipePanel = FoodPanelRecipe.new(rounds)
  self:insertAtCenter(recipePanel, Z_PANEL)
  self.recipePanel = recipePanel

  local dots = recipePanel:getDots()

  local progressIndicator = FoodProgressIndicator.new(dots)
  progressIndicator:setLoc(Screen.getSafeX(0.5), PROGRESS_INDICATOR_Y)
  self:insert(progressIndicator, Z_PROGRESS_INDICATOR)
  self.progressIndicator = progressIndicator

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 FoodConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(FoodConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       FoodConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
end

-- Constructor ------------------------

function FoodGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, FoodGame)

  configureGameDifficulty(self)
  buildRounds(self)
  createGameElements(self)
  createUI(self)

  self.round = nil
  self.roundIndex = 0
  self.roundPanel = nil
  self.titleLabel = nil
  self.guessesCount = 0
  self.hasSeenMathTipThisRound = false
  self.hasSeenTapToRemoveTip = false
  self.isShowingTapToRemoveTip = false
  self.hasPerformedRemoval = false
  self.roundsWithMistakesCount = 0

  return self
end
