--- FoodHolder
-- @author Created by Aaron Barrett on January 06, 2015.

class("FoodHolder")

require("FoodHolderOutline")
require("FoodAnswerOptionButton")
require("FoodSubmitButton")
require("FoodFractionConstants")
require("FoodStrobeAction")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Babylonian Systems of Measurement --

local DEGREES_IN_A_FULL_CIRCLE = 360.0

-- Selection --------------------------

local INITIAL_ITEMS_PER_ROW = 3

local MAX_ITEMS_PER_ROW = 4
local MAX_ROWS = 3
local MAX_SELECTED_ITEMS = MAX_ITEMS_PER_ROW * MAX_ROWS

-- Dimensions -------------------------

local WIDTH = Screen.getSafeWidthPart(0.85)
local HEIGHT = Screen.getSafeHeightPart(0.33)

local TOP_ITEM_MARGIN = Screen.getSafeHeightPart(0.01)
local BOTTOM_ITEM_MARGIN = Screen.getSafeHeightPart(0.038)

local MAX_HORIZONTAL_ITEM_SPACING = WIDTH / MAX_ITEMS_PER_ROW
local VERTICAL_ITEM_SPACING = (HEIGHT - (TOP_ITEM_MARGIN + BOTTOM_ITEM_MARGIN)) / MAX_ROWS

local MIN_RECTANGLE_HEIGHT = Screen.getSafeHeightPart(0.17)
local RECTANGLE_HEIGHT = math.max(TOP_ITEM_MARGIN + BOTTOM_ITEM_MARGIN + MAX_ROWS * VERTICAL_ITEM_SPACING, MIN_RECTANGLE_HEIGHT)
local INITIAL_RECTANGLE_WIDTH = Screen.getSafeWidthPart(0.7)

local SUBMIT_BUTTON_WIDTH = Screen.getSafeWidthPart(0.45)
local OUTLINE_LINE_WIDTH = 0.6
local OUTLINE_BLUR_MARGIN = 0.5
local OUTLINE_LINE_BEVEL = Screen.getSafeHeightPart(0.003)
local SUBMIT_BUTTON_Y = HEIGHT * 0.5 - RECTANGLE_HEIGHT * 0.5

local MAX_GLOBAL_Y = Screen.getSafeY(0.5)
local MIN_GLOBAL_Y = Screen.getSafeY(0.41)

-- Z-Indices --------------------------

local Z_OUTLINE = 10
local Z_SUBMIT_BUTTON = 20
local Z_ITEM = 30

-- Reduction Animation ----------------

local REDUCTION_PATTERN_DESCRIPTORS = {
  {replacementAmounts = {FoodFractionConstants.ONE_WHOLE}, pattern = {FoodFractionConstants.ONE_HALF, FoodFractionConstants.ONE_HALF}},
  {replacementAmounts = {FoodFractionConstants.ONE_WHOLE}, pattern = {FoodFractionConstants.ONE_THIRD, FoodFractionConstants.ONE_THIRD, FoodFractionConstants.ONE_THIRD}},
  {replacementAmounts = {FoodFractionConstants.ONE_HALF}, pattern = {FoodFractionConstants.ONE_FOURTH, FoodFractionConstants.ONE_FOURTH}},
  {replacementAmounts = {FoodFractionConstants.ONE_HALF}, pattern = {FoodFractionConstants.ONE_SIXTH, FoodFractionConstants.ONE_SIXTH, FoodFractionConstants.ONE_SIXTH}},
  {replacementAmounts = {FoodFractionConstants.ONE_THIRD}, pattern = {FoodFractionConstants.ONE_SIXTH, FoodFractionConstants.ONE_SIXTH}},
  {replacementAmounts = {FoodFractionConstants.ONE_FOURTH}, pattern = {FoodFractionConstants.ONE_EIGHTH, FoodFractionConstants.ONE_EIGHTH}},
  {replacementAmounts = {FoodFractionConstants.ONE_HALF}, pattern = {FoodFractionConstants.ONE_THIRD, FoodFractionConstants.ONE_SIXTH}},
  {replacementAmounts = {FoodFractionConstants.ONE_SIXTH, FoodFractionConstants.ONE_SIXTH}, pattern = {FoodFractionConstants.ONE_THIRD}}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Selection --------------------------

local function setSelectedAmount(self, selectedAmount)
  local submitButton = self.submitButton

  self.selectedAmount = selectedAmount

  submitButton:setDisplayedAmount(selectedAmount)

  local shouldAppear = (selectedAmount:getNumerator() > 0)
  if (shouldAppear) then
    self:playAppear()
  else
    self:playShrinkOut()
  end
end

local function addToSelection(self, amount)
  local selectedAmount = self.selectedAmount
  local updatedAmount = selectedAmount:add(amount)
  setSelectedAmount(self, updatedAmount)
end

local function subtractFromSelection(self, amount)
  local selectedAmount = self.selectedAmount
  local updatedAmount = selectedAmount:subtract(amount)
  setSelectedAmount(self, updatedAmount)
end

-- Appearance Actions -----------------

local function replaceAppearanceAction(self, appearanceAction)
  local appearanceInterrupt = self.appearanceInterrupt
  appearanceInterrupt:insert(appearanceAction)

  return runInCoroutine(function() appearanceInterrupt:wait() end)
end

-- Arrangement Action -----------------

local function replaceArrangementAction(self, arrangementAction)
  local arrangementInterrupt = self.arrangementInterrupt
  arrangementInterrupt:insert(arrangementAction)

  return runInCoroutine(function() arrangementInterrupt:wait() end)
end

local function playArrangeItems(self)
  local items = self.items
  local outline = self.outline
  local submitButton = self.submitButton

  local itemsCount = #items

  local itemsPerRow = nil
  if (itemsCount <= INITIAL_ITEMS_PER_ROW * MAX_ROWS) then
    itemsPerRow = INITIAL_ITEMS_PER_ROW
  else
    itemsPerRow = MAX_ITEMS_PER_ROW
  end

  local horizontalItemSpacing = MAX_HORIZONTAL_ITEM_SPACING * (MAX_ITEMS_PER_ROW / itemsPerRow)

  local rowsCount = math.ceil(itemsCount / itemsPerRow)

  local arrangementAction = AHSpawn.new()

  local ease = AHEaseSimpleInOut.new(4.0)

  local topY = (HEIGHT + RECTANGLE_HEIGHT - VERTICAL_ITEM_SPACING) * 0.5 - TOP_ITEM_MARGIN
  local centerX = WIDTH * 0.5

  local itemsOnBottomRow = ((itemsCount - 1) % itemsPerRow) + 1

  for itemIndex, item in ipairs(items) do
    local itemXIndex = (itemIndex - 1) % itemsPerRow
    local itemYIndex = math.floor((itemIndex - 1) / itemsPerRow)

    local itemsThisRow = nil
    if (itemYIndex == rowsCount - 1) then
      itemsThisRow = itemsOnBottomRow
    else
      itemsThisRow = itemsPerRow
    end

    local destinationX = centerX - itemsThisRow * 0.5 * horizontalItemSpacing + (itemXIndex + 0.5) * horizontalItemSpacing
    local destinationY = topY - itemYIndex * VERTICAL_ITEM_SPACING - (MAX_ROWS - rowsCount) * VERTICAL_ITEM_SPACING / 2

    local flyItemAction = AHLocationTo.new(item, 0.5, ease, destinationX, destinationY)
    arrangementAction:addAction(flyItemAction)
  end

  return replaceArrangementAction(self, arrangementAction)
end

-- Adding Items -----------------------

local function addItemAtAscendingIndex(self, item)
  local items = self.items
  local itemValue = item:getAmount():toNumber()
  local indexToAddItem = #items + 1
  for existingItemIndex, existingItem in ipairs(items) do
    local existingItemValue = existingItem:getAmount():toNumber()
    if (existingItemValue < itemValue) then
      indexToAddItem = existingItemIndex
      break
    end
  end

  table.insert(items, indexToAddItem, item)
end

-- Removing Items ---------------------

local function playRemoveItem(self, item)
  local items = self.items
  TableUtils.removeValueFromArray(items, item)

  Sound.playSoundEffect("deselect")
  item:playDisappear()

  local itemAmount = item:getAmount()
  subtractFromSelection(self, itemAmount)

  return playArrangeItems(self)
end

-- Reduction Animation ----------------

local function doesPatternExistAtIndex(self, pattern, startingIndex)
  local items = self.items

  for patternIndex, patternAmount in ipairs(pattern) do
    local itemIndex = startingIndex + patternIndex - 1
    local item = items[itemIndex]

    local didRunOutOfItems = (item == nil)
    if (didRunOutOfItems) then
      return false
    end

    local itemAmount = item:getAmount()
    local doesItemBreakPattern = (not itemAmount:equals(patternAmount))
    if (doesItemBreakPattern) then
      return false
    end
  end

  return true
end

local function findPatternFromIndex(self, startingIndex)
  local items = self.items
  local itemsCount = #items

  for patternSearchIndex = startingIndex, itemsCount do

    for _, descriptor in ipairs(REDUCTION_PATTERN_DESCRIPTORS) do
      local pattern = descriptor.pattern
      if (doesPatternExistAtIndex(self, pattern, patternSearchIndex)) then
        return descriptor, patternSearchIndex
      end
    end

  end

  error("Reduction animation is incomplete, but unable to find any patterns.")
end

local function playReduction(self, patternStartIndex, patternEndIndex, replacementAmount)
  self.isReducing = true

  return runInCoroutine(function()
    local items = self.items

    local firstItemToReduce = items[patternStartIndex]
    local reductionDestinationX, reductionDestinationY = firstItemToReduce:getLoc()

    local previousItemAngle = firstItemToReduce:getAmount():toNumber() * DEGREES_IN_A_FULL_CIRCLE
    local destinationRotation = -previousItemAngle

    for reducibleIndex = patternStartIndex, patternEndIndex do
      local itemToReduce = items[reducibleIndex]

      local itemAmount = itemToReduce:getAmount()
      local itemProportion = itemAmount:toNumber()
      local itemAngle = itemProportion * DEGREES_IN_A_FULL_CIRCLE

      destinationRotation = destinationRotation + previousItemAngle
      itemToReduce:playReduceInto(reductionDestinationX, reductionDestinationY, destinationRotation)

      previousItemAngle = itemAngle
    end

    local numberOfItemsToReduce = patternEndIndex - patternStartIndex + 1
    for i = 1, numberOfItemsToReduce do
      table.remove(items, patternStartIndex)
    end

    local roundColor = self.roundColor
    local replacementItem = FoodItem.new(replacementAmount, roundColor)
    replacementItem:hideInitially()
    replacementItem:setLoc(reductionDestinationX, reductionDestinationY)
    self:insert(replacementItem, Z_ITEM)
    addItemAtAscendingIndex(self, replacementItem)

    replacementItem:playReplace()
    waitFor(0.8)
    Sound.playSoundEffect("combine")

    wait(playArrangeItems(self))
  end)
end

local function playExpansion(self, patternStartIndex, replacementAmounts)
  return runInCoroutine(function()
    local items = self.items

    local itemToExpand = items[patternStartIndex]
    local replacementStartX, replacementStartY = itemToExpand:getLoc()

    itemToExpand:removeSelf()
    table.remove(items, patternStartIndex)

    local firstReplacementAmount = replacementAmounts[1]

    local previousItemAngle = firstReplacementAmount:toNumber() * DEGREES_IN_A_FULL_CIRCLE
    local initialRotation = -previousItemAngle

    for replacementIndex, replacementAmount in ipairs(replacementAmounts) do
      local itemProportion = replacementAmount:toNumber()
      local itemAngle = itemProportion * DEGREES_IN_A_FULL_CIRCLE

      initialRotation = initialRotation + previousItemAngle

      local roundColor = self.roundColor
      local replacementItem = FoodItem.new(replacementAmount, roundColor)
      replacementItem:setRot(initialRotation)
      replacementItem:setLoc(replacementStartX, replacementStartY)
      self:insert(replacementItem, Z_ITEM)
      addItemAtAscendingIndex(self, replacementItem)

      replacementItem:playRotateBackToZero()

      previousItemAngle = itemAngle
    end

    Sound.playSoundEffect("combine")
    wait(playArrangeItems(self))
  end)
end

local function playReplacement(self, patternStartIndex, patternEndIndex, replacementAmounts)
  local replacementsAmountsCount = #replacementAmounts
  if (replacementsAmountsCount == 1) then
    local replacementAmount = replacementAmounts[1]
    return playReduction(self, patternStartIndex, patternEndIndex, replacementAmount)
  else
    return playExpansion(self, patternStartIndex, replacementAmounts)
  end
end

local function hasReachedOptimalNumberOfSteps(self, optimalSteps)
  local items = self.items
  local itemsCount = #items

  local optimalStepsCount = #optimalSteps

  return (itemsCount == optimalStepsCount)
end

function isReducing(self, optimalSteps, optimalStepIndex)
  local items = self.items

  local actualStep = items[optimalStepIndex]:getAmount()
  local optimalStep = optimalSteps[optimalStepIndex]

  if ((actualStep:equals(optimalStep)) or (hasReachedOptimalNumberOfSteps(self, optimalSteps))) then
    return false
  end

  return true
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodHolder:getSubmitButton()
  return self.submitButton
end

-- Add and Remove Items ---------------

function FoodHolder:playAddItemFromButton(answerOptionButton)
  local roundColor = self.roundColor
  local itemAmount = answerOptionButton:getAmount()
  local item = FoodItem.new(itemAmount, roundColor)

  if self.isInTutorial then
    item:freeze()
  end

  local function onTappedCallback()
     playRemoveItem(self, item)

     if self.onRemoveCallback then
       self.onRemoveCallback()
     end
  end
  item:setOnTappedCallback(onTappedCallback)

  addItemAtAscendingIndex(self, item)

  local answerOptionButtonX, answerOptionButtonY = answerOptionButton:getWorldLoc()
  local convertedX, convertedY = self:worldToModel(answerOptionButtonX, answerOptionButtonY)
  local itemStartingX = convertedX + FoodAnswerOptionButton.ICON_X_OFFSET
  local itemStartingY = convertedY + FoodAnswerOptionButton.ICON_Y_OFFSET

  item:setLoc(itemStartingX, itemStartingY)

  self:insert(item, Z_ITEM)

  item:playAppear()
  addToSelection(self, itemAmount)

  return playArrangeItems(self)
end

function FoodHolder:setOnRemoveCallback(callback)
  self.onRemoveCallback = callback
end

-- Reduction Animation ----------------

function FoodHolder:playReduce(optimalSteps)
  return runInCoroutine(function()
    for optimalStepIndex in ipairs(optimalSteps) do
      while (isReducing(self, optimalSteps, optimalStepIndex)) do
        Sound.playSoundEffect("swipe")

        local patternDescriptor, patternStartIndex = findPatternFromIndex(self, optimalStepIndex)

        local pattern = patternDescriptor.pattern
        local patternLength = #pattern
        local patternEndIndex = patternStartIndex + patternLength - 1

        local replacementAmounts = patternDescriptor.replacementAmounts

        wait(playReplacement(self, patternStartIndex, patternEndIndex, replacementAmounts))
      end
    end

    waitFor(0.2)

  end)
end

-- Strobe Animation -------------------

function FoodHolder:playStrobe()
  local outline = self.outline
  local submitButton = self.submitButton
  local items = self.items

  outline:playStrobe()

  for _, item in ipairs(items) do
    item:playStrobe()
  end

  return submitButton:playStrobe()
end

-- Correct Animation ------------------

function FoodHolder:playCorrect()
  local outline = self.outline
  return outline:playCorrect()
end

-- Wrong Animation ---------------------

function FoodHolder:playWrong()
  local items = self.items
  local outline = self.outline
  local submitButton = self.submitButton

  for _, item in ipairs(items) do
    item:playWrong()
  end

  outline:playWrong()
  submitButton:playWrong()

  return AHDelay.new(1.6):start()
end

-- Appear Animation -------------------

function FoodHolder:hideInitially()
  local outline = self.outline
  local submitButton = self.submitButton

  outline:hideInitially()
  submitButton:hideInitially()
end

function FoodHolder:playAppear()
  local outline = self.outline
  local submitButton = self.submitButton

  submitButton:setY(SUBMIT_BUTTON_Y)
  outline:setRectangleWidth(WIDTH)
  outline:setRectangleHeight(RECTANGLE_HEIGHT)

  local appearAction = AHSpawn.new(
                                   outline:createAppearAction(),
                                   submitButton:createAppearAction()
                                   )
  return replaceAppearanceAction(self, appearAction)
end

-- Disappear Animations ---------------

function FoodHolder:playFadeOut()
  local outline = self.outline
  local submitButton = self.submitButton

  local fadeOutAction = AHSpawn.new(
                                    outline:createFadeOutAction(),
                                    submitButton:createFadeOutAction()
                                    )

  return replaceAppearanceAction(self, fadeOutAction)
end

function FoodHolder:playShrinkOut()
  local outline = self.outline
  local submitButton = self.submitButton

  local destinationRectangleWidth = INITIAL_RECTANGLE_WIDTH
  local destinationRectangleHeight = MIN_RECTANGLE_HEIGHT

  local submitButtonDestinationY = HEIGHT * 0.5 - destinationRectangleHeight * 0.5
  local submitButtonLocationAction = AHLocationYTo.new(submitButton, 0.3, AHEaseSimpleIn.new(2.0), submitButtonDestinationY)

  local shrinkOutAction = AHSpawn.new(
                                      outline:createShrinkOutAction(destinationRectangleWidth, destinationRectangleHeight),
                                      submitButtonLocationAction,
                                      submitButton:createShrinkOutAction()
                                      )
  return replaceAppearanceAction(self, shrinkOutAction)
end

-- Flush Animation --------------------

function FoodHolder:playFlushWithProgressIndicator(progressIndicator)
  local destinationX, destinationY = progressIndicator:getFlushDestinationLocation()

  local items = self.items
  for _, item in ipairs(items) do
    runInCoroutine(function()
      local staggerDuration = math.random() * 0.1
      waitFor(staggerDuration)
      item:playFlushWithDestination(destinationX, destinationY)
    end)
  end
end

-- Loss Animation ---------------------

function FoodHolder:playLoss()
  local outline = self.outline
  local submitButton = self.submitButton
  local items = self.items

  for _, item in ipairs(items) do
    runInCoroutine(function()
      local delayDuration = math.random() * 0.14
      waitFor(delayDuration)
      item:playDisappear()
    end)
  end

  submitButton:playLoss()

  return outline:playLoss()
end

-- Tutorial Animations -------------------------

function FoodHolder:enableTutorial()
  local outline = self.outline
  outline:displayForTutorial()

  local submitButton = self.submitButton
  submitButton:displayForTutorial()

  self.isInTutorial = true
end

function FoodHolder:tutorialEnableSubmitButton()
  local submitButton = self.submitButton
  submitButton:tutorialEnable()

  submitButton:setTouchEnabled(true)

  local centerX = self:getWidthPart(0.5)
  submitButton:playTutorialShake(centerX)
end

function FoodHolder:endTutorial()
  local submitButton = self.submitButton
  local centerX = self:getWidthPart(0.5)
  submitButton:playStopTutorialShake(centerX)
end

function FoodHolder:getItemCount()
  return #self.items
end

-- Selection --------------------------

function FoodHolder:canAcceptMoreItems()
  local items = self.items
  local itemsCount = #items
  return (itemsCount < MAX_SELECTED_ITEMS)
end

-- Offsets ----------------------------

function FoodHolder:getTopY()
  local y = self:getY()
  local topY = y + RECTANGLE_HEIGHT * 0.5
  return topY
end

-- Touch Management -------------------

function FoodHolder:setTouchEnabled(touchEnabled)
  local submitButton = self.submitButton
  local items = self.items

  submitButton:setTouchEnabled(touchEnabled)
  for _, item in ipairs(items) do
    item:setTouchEnabled(touchEnabled)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createSubmitCallback(self, roundPanelSubmitCallback)
  local function submitCallback()
    local selectedAmount = self.selectedAmount

    local isSelectedAmountValid = (selectedAmount:getNumerator() > 0)
    if (isSelectedAmountValid) then
      local submittedSteps = self.items
      roundPanelSubmitCallback(selectedAmount, submittedSteps)
    end
  end
  return submitCallback
end

-- Constructor ------------------------

function FoodHolder.new(roundColor, roundPanelSubmitCallback)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, FoodHolder)

  local outline = FoodHolderOutline.new(roundColor, WIDTH, RECTANGLE_HEIGHT, SUBMIT_BUTTON_WIDTH, OUTLINE_LINE_WIDTH, OUTLINE_BLUR_MARGIN, OUTLINE_LINE_BEVEL)
  outline:setRectangleWidth(INITIAL_RECTANGLE_WIDTH)
  self:insertAtCenter(outline, Z_OUTLINE)
  self.outline = outline

  local submitCallback = createSubmitCallback(self, roundPanelSubmitCallback)
  local submitButton = FoodSubmitButton.new(roundColor, submitCallback)
  submitButton:setLoc(self:getWidthPart(0.5), SUBMIT_BUTTON_Y)
  self:insert(submitButton, Z_SUBMIT_BUTTON)
  self.submitButton = submitButton

  self.items = {}
  self.roundColor = roundColor
  self.selectedAmount = FoodFractionConstants.ZERO
  self.appearanceInterrupt = ActionInterrupt.new()
  self.arrangementInterrupt = ActionInterrupt.new()
  self.isInTutorial = false
  self.isReducing = false

  return self
end
