--- FoodHolderOutline
-- @author Created by Aaron Barrett on January 06, 2015.

class("FoodHolderOutline")

require("FoodConstants")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FULL_OPACITY = 0.3
local TUTORIAL_COLOR = {0.6, 0.6, 0.6}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateCoordinates(self)
  local rectangleHeight = self.rectangleHeight
  local rectangleWidth = self.rectangleWidth
  local bevelAmount = self.bevelAmount

  local width, height = self:getDims()

  local top = (height + rectangleHeight) * 0.5
  local bottom = (height - rectangleHeight) * 0.5
  local left = (width - rectangleWidth) * 0.5
  local right = (width + rectangleWidth) * 0.5

  local gapWidth = self.gapWidth
  local leftOfSubmitButton = (width - gapWidth) * 0.5
  local rightOfSubmitButton = (width + gapWidth) * 0.5

  local coordinates = {
    leftOfSubmitButton, bottom,

    left + self.bevelAmount, bottom,
    left, bottom + bevelAmount,

    left, top - bevelAmount,
    left + bevelAmount, top,

    right - bevelAmount, top,
    right, top - bevelAmount,

    right, bottom + bevelAmount,
    right - bevelAmount, bottom,

    rightOfSubmitButton, bottom
  }

  self.coordinates = coordinates
end

local function linkColorButNotOpacity(self, child)
  child:setAttrLink(MOAIColor.ATTR_R_COL, self)
  child:setAttrLink(MOAIColor.ATTR_G_COL, self)
  child:setAttrLink(MOAIColor.ATTR_B_COL, self)
end

local function playCorrectHalf(self, submitButtonX, maxX, sideRotationOffset)
  return runInCoroutine(function()
    local width, height = self:getDims()
    local rectangleHeight = self.rectangleHeight

    local top = (height + rectangleHeight) * 0.5
    local bottom = (height - rectangleHeight) * 0.5

    local centerX = width * 0.5

    local sideSprites = self.sideSprites

    local firstSide = Sprite.newWithName("correct_side.png")
    local sideWidth, sideHeight = firstSide:getDims()
    local sideThickness = Screen.getSafeWidthPart(0.005)
    local sideThicknessScale = sideThickness / sideHeight

    firstSide:setRot(sideRotationOffset)
    firstSide:setAnchorPointLeft()
    firstSide:setScl(0.0, sideThicknessScale)
    firstSide:setLoc(submitButtonX, bottom)
    linkColorButNotOpacity(self, firstSide)
    self:insert(firstSide)
    table.insert(sideSprites, firstSide)

    local secondSide = Sprite.newWithName("correct_side.png")
    secondSide:setAnchorPointLeft()
    secondSide:setRot(90.0)
    secondSide:setScl(0.0, sideThicknessScale)
    secondSide:setLoc(maxX, bottom)
    linkColorButNotOpacity(self, secondSide)
    self:insert(secondSide)
    table.insert(sideSprites, secondSide)

    local thirdSide = Sprite.newWithName("correct_side.png")
    thirdSide:setAnchorPointLeft()
    thirdSide:setRot(180.0 + sideRotationOffset)
    thirdSide:setScl(0.0, sideThicknessScale)
    thirdSide:setLoc(maxX, top)
    linkColorButNotOpacity(self, thirdSide)
    self:insert(thirdSide)
    table.insert(sideSprites, thirdSide)

    local firstDistance = math.abs(maxX - submitButtonX)
    local secondDistance = rectangleHeight
    local thirdDistance = centerX

    local totalDistance = firstDistance + secondDistance + thirdDistance
    local totalDuration = (firstDistance + secondDistance + thirdDistance) * 0.0012

    local firstDuration = firstDistance / totalDistance * totalDuration * 0.5
    local secondDuration = secondDistance / totalDistance * totalDuration * 0.9
    local thirdDuration = thirdDistance / totalDistance * totalDuration * 1.6

    local firstScale = firstDistance / sideWidth
    local secondScale = secondDistance / sideWidth
    local thirdScale = thirdDistance / sideWidth

    local fudgeDuration = 0.055
    local firstFudgedDuration = math.max(0.0, firstDuration - fudgeDuration)
    local secondFudgedDuration = math.max(0.0, firstDuration + secondDuration - fudgeDuration * 2.0)

    local correctAction = AHSpawn.new(
                                      AHScaleXTo.new(firstSide, firstDuration, AHEaseSineInOut.new(), firstScale),
                                      AHDelayed.new(firstFudgedDuration, AHScaleXTo.new(secondSide, secondDuration, AHEaseSineInOut.new(), secondScale)),
                                      AHDelayed.new(secondFudgedDuration, AHScaleXTo.new(thirdSide, thirdDuration, AHEaseSineInOut.new(), thirdScale))
                                      )

    wait(correctAction:start())
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodHolderOutline:setRectangleHeight(rectangleHeight)
  self.rectangleHeight = rectangleHeight
  updateCoordinates(self)
end

function FoodHolderOutline:setRectangleWidth(rectangleWidth)
  self.rectangleWidth = rectangleWidth
  updateCoordinates(self)
end

function FoodHolderOutline:getRectangleHeight()
  return self.rectangleHeight
end

function FoodHolderOutline:getRectangleWidth()
  return self.rectangleWidth
end

function FoodHolderOutline:getBlurMargin()
  return self.blurMargin
end

function FoodHolderOutline:setBlurMargin(blurMargin)
  self.blurMargin = blurMargin
end

function FoodHolderOutline:createRectangleHeightToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getRectangleHeight, self.setRectangleHeight, ease, amount)
end

function FoodHolderOutline:createRectangleWidthToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getRectangleWidth, self.setRectangleWidth, ease, amount)
end

function FoodHolderOutline:createBlurMarginToAction(duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, self.getBlurMargin, self.setBlurMargin, ease, amount)
end

-- Appear Animation -------------------

function FoodHolderOutline:hideInitially()
  self:setOpacity(0.0)
end

function FoodHolderOutline:createAppearAction()
    local ease = AHEaseExponentialOut.new()

    local appearAction = AHSpawn.new(
                                     AHOpacityTo.new(self, 0.8, ease, FULL_OPACITY),
                                     self:createRectangleWidthToAction(0.8, ease, self:getWidth())
                                     )
    return appearAction
end

-- Disappear Animations ---------------

function FoodHolderOutline:createFadeOutAction()
  local sideSprites = self.sideSprites

  local fadeOutAction = AHSpawn.new()

  local outlineFadeAction = AHOpacityTo.new(self, 0.7, AHEaseSimpleIn.new(2.0), 0.0)
  fadeOutAction:addAction(outlineFadeAction)

  for _, sideSprite in ipairs(sideSprites) do
    local sideSpriteFadeAction = AHOpacityTo.new(sideSprite, 0.7, AHEaseSimpleIn.new(2.0), 0.0)
    fadeOutAction:addAction(sideSpriteFadeAction)
  end

  return fadeOutAction
end

function FoodHolderOutline:createShrinkOutAction(destinationRectangleWidth, destinationRectangleHeight)
  local ease = AHEaseSimpleIn.new(2.0)

  local shrinkOutAction = AHSpawn.new(
                                      AHOpacityTo.new(self, 0.3, ease, 0.0),
                                      self:createRectangleWidthToAction(0.3, ease, destinationRectangleWidth),
                                      self:createRectangleHeightToAction(0.3, ease, destinationRectangleHeight)
                                      )
  return shrinkOutAction
end

-- Strobe Animation -------------------

function FoodHolderOutline:playStrobe()
  local strobeAction = FoodStrobeAction.new(self)
  return strobeAction:start()
end

-- Correct Animation ------------------

function FoodHolderOutline:playCorrect()
  return runInCoroutine(function()
    local width = self:getWidth()
    local rectangleWidth = self.rectangleWidth
    local gapWidth = self.gapWidth

    local leftOfSubmitButton = (width - gapWidth) * 0.5
    local rightOfSubmitButton = (width + gapWidth) * 0.5

    local left = (width - rectangleWidth) * 0.5
    local right = (width + rectangleWidth) * 0.5

    playCorrectHalf(self, leftOfSubmitButton, left, 180.0)
    wait(playCorrectHalf(self, rightOfSubmitButton, right, 0.0))
  end)
end

-- Wrong Animation --------------------

function FoodHolderOutline:playWrong()
  local roundColor = self.roundColor

  local redRemainDuration = 1.55

  local wrongAction = AHSequence.new(
                                     AHColorTo.new(self, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                     AHColorTo.new(self, redRemainDuration, AHEaseSimpleOut.new(1.4), roundColor)
                                     )
  return wrongAction:start()
end

-- Loss Animation ---------------------

function FoodHolderOutline:playLoss()
  local lossAction = AHSpawn.new(
                                 AHScaleTo.new(self, 0.6, AHEaseExponentialIn.new(), 0.0),
                                 AHOpacityTo.new(self, 0.6, nil, 0.0)
                                 )
  return lossAction:start()
end

-- Tutorial Animations -----------------------

function FoodHolderOutline:displayForTutorial()
  local fadeAction = AHColorTo.new(self, FoodConstants.TUTORIAL_MORPH_DURATION, nil, TUTORIAL_COLOR)
  fadeAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FoodHolderOutline.new(roundColor, width, height, gapWidth, lineWidth, blurMargin, bevelAmount)
  local self = Group.newCustomDraw(width, height, roundColor)
  Mixin.mix(self, FoodHolderOutline)

  self.sideSprites = {}
  self.rectangleHeight = height
  self.rectangleWidth = width
  self.gapWidth = gapWidth
  self.bevelAmount = bevelAmount
  self.roundColor = roundColor
  updateCoordinates(self)

  self.lineWidth = lineWidth
  self.blurMargin = blurMargin

  self.deck:setDrawCallback(function() MOAIDraw.drawBeveledLine(self.coordinates, self.lineWidth, self.blurMargin) end)

  return self
end
