--- FoodItem
-- @author Created by Aaron Barrett on December 03, 2014.

class("FoodItem")

require("graphics.ShapeUtils")
require("FoodConstants")
require("FoodFractionConstants")
require("FoodStrobeAction")
require("FoodPieFraction")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local RADIUS = Screen.getSafeHeightPart(0.04)
local DIAMETER = RADIUS * 2.0

local BACK_OUTLINE_OPACITY = 0.15
local FRONT_FILLING_OPACITY = 0.25

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodItem:freeze()
  self.isFrozen = true
end

function FoodItem:getAmount()
  return self.amount
end

function FoodItem:playAppear()
  local backOutline = self.backOutline
  local frontOutline = self.frontOutline
  local frontFilling = self.frontFilling

  backOutline:setOpacity(0.0)
  frontOutline:setOpacity(0.0)
  frontFilling:setOpacity(0.0)
  self:setScl(0.5)

  local ease = AHEaseSimpleInOut.new(4.0)

  local appearAction = AHSpawn.new(
                                   AHDelayed.new(0.3, function()
                                                        -- Don't allow items to be tapped and removed from
                                                        -- the holder until *slightly* after they're added.
                                                        self.allowRemoval = not self.isFrozen
                                                      end),
                                   AHOpacityTo.new(backOutline, 0.5, ease, BACK_OUTLINE_OPACITY),
                                   AHOpacityTo.new(frontOutline, 0.5, ease, 1.0),
                                   AHOpacityTo.new(frontFilling, 0.5, ease, FRONT_FILLING_OPACITY),
                                   AHScaleTo.new(self, 0.5, ease, 1.0)
                                   )
  return appearAction:start()
end

function FoodItem:playDisappear()
  local disappearAction = AHSequence.new(
                                         AHScaleTo.new(self, 0.1, AHEaseSimpleIn.new(2.0), 0.0),
                                         AHRemove.new(self)
                                         )
  return disappearAction:start()
end

function FoodItem:playStrobe()
  local frontOutline = self.frontOutline
  local backOutline = self.backOutline
  local frontFilling = self.frontFilling

  local strobeAction = AHSpawn.new(
                                   FoodStrobeAction.new(frontOutline),
                                   FoodStrobeAction.new(backOutline),
                                   FoodStrobeAction.new(frontFilling)
                                   )

  return strobeAction:start()
end

function FoodItem:playWrong()
  local shakeGroup = self.shakeGroup
  local roundColor = self.roundColor
  local backOutline = self.backOutline
  local frontOutline = self.frontOutline
  local frontFilling = self.frontFilling

  local redRemainDuration = 1.55

  local colorEase = AHEaseSimpleOut.new(1.4)

  local colorAction = AHSequence.new(
                                     AHSpawn.new(
                                                 AHColorTo.new(backOutline, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                                 AHColorTo.new(frontOutline, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                                 AHColorTo.new(frontFilling, 0.1, nil, FoodConstants.WRONG_RED_COLOR)
                                                 ),
                                     AHSpawn.new(
                                                 AHColorTo.new(backOutline, redRemainDuration, colorEase, roundColor),
                                                 AHColorTo.new(frontOutline, redRemainDuration, colorEase, roundColor),
                                                 AHColorTo.new(frontFilling, redRemainDuration, colorEase, roundColor)
                                                 )
                                     )


  local scaleUpDuration = 0.2 + math.random() * 0.1
  local scaleUpAmount = 1.1 + math.random() * 0.1
  local scaleDownDuration = 0.3 + math.random() * 0.1


  local scaleAction = AHSequence.new(
                                     AHScaleTo.new(self, scaleUpDuration, AHEaseSimpleIn.new(2.2), scaleUpAmount),
                                     AHScaleTo.new(self, scaleDownDuration, AHEaseSimpleOut.new(2.2), 1.0)
                                     )

  local shakeEase = AHEaseSimpleInOut.new(2.0)
  local signX = MathUtils.randomSign()
  local signY = MathUtils.randomSign()

  local shakeAction = AHSpawn.new(
                                  AHSequence.new(
                                                 AHLocationXTo.new(shakeGroup, 0.05 + math.random() * 0.1, shakeEase, RADIUS + signX * (2.0 + math.random() * 0.1)),
                                                 AHLocationXTo.new(shakeGroup, 0.05 + math.random() * 0.1, shakeEase, RADIUS - signX * (3.0 + math.random() * 0.1)),
                                                 AHLocationXTo.new(shakeGroup, 0.06 + math.random() * 0.1, shakeEase, RADIUS)
                                                 ),
                                  AHSequence.new(
                                                 AHLocationYTo.new(shakeGroup, 0.05 + math.random() * 0.1, shakeEase, RADIUS - signY * (2.0 + math.random() * 0.1)),
                                                 AHLocationYTo.new(shakeGroup, 0.05 + math.random() * 0.1, shakeEase, RADIUS + signY * (3.0 + math.random() * 0.1)),
                                                 AHLocationYTo.new(shakeGroup, 0.06 + math.random() * 0.1, shakeEase, RADIUS)
                                                 )
                                  )

  local wrongAction = AHSpawn.new(
                                  colorAction,
                                  scaleAction,
                                  shakeAction
                                  )
  return wrongAction:start()
end

function FoodItem:playReduceInto(destinationX, destinationY, destinationRotation)
  local backOutline = self.backOutline
  local frontFilling = self.frontFilling

  local reduceAction = AHSequence.new(
                                      AHSpawn.new(
                                                  AHOpacityTo.new(backOutline, 0.1, nil, 0.0),
                                                  AHOpacityTo.new(frontFilling, 0.1, nil, 1.0),
                                                  AHRotationBy.new(self, 0.4, AHEaseSimpleInOut.new(2.0), destinationRotation),
                                                  AHLocationTo.new(self, 0.4, AHEaseSimpleInOut.new(2.0), destinationX, destinationY)
                                                  ),
                                      AHDelay.new(0.4),
                                      AHRemove.new(self)
                                      )

  return reduceAction:start()
end

function FoodItem:playReplace()
  local backOutline = self.backOutline
  local frontFilling = self.frontFilling

  local replaceAction = AHSequence.new(
                                       AHDelay.new(0.78),
                                       function()
                                         self.frontOutline:setOpacity(1.0)
                                         self.frontFilling:setOpacity(1.0)
                                       end,
                                       AHSpawn.new(
                                                   AHOpacityTo.new(backOutline, 0.4, nil, BACK_OUTLINE_OPACITY),
                                                   AHOpacityTo.new(frontFilling, 0.4, nil, FRONT_FILLING_OPACITY)
                                                   )
                                       )
  return replaceAction:start()
end

function FoodItem:playRotateBackToZero()
  local rotateAction = AHRotationTo.new(self, 0.4, AHEaseSimpleInOut.new(2.0), 0.0)
  return rotateAction:start()
end

function FoodItem:hideInitially()
  self.frontFilling:setOpacity(0.0)
  self.frontOutline:setOpacity(0.0)
  self.backOutline:setOpacity(0.0)
end

function FoodItem:playFlushWithDestination(destinationX, destinationY)
  local x, y = self:getWorldLoc()

  local backOutline = self.backOutline
  local frontOutline = self.frontOutline
  local frontFilling = self.frontFilling

  -- Fudge coordinates because getWorldLoc apparently just gives you the lower-left corner, not its actual location.
  x = x + self:getWidthPart(0.5)
  y = y + self:getHeightPart(0.5)

  local deltaX = destinationX - x
  local deltaY = destinationY - y

  local destinationRotation = -90.0 + math.deg(math.atan2(deltaY, deltaX))

  local fadeEase = AHEaseExponentialOut.new()
  local fadeAction = AHSpawn.new(
                                 AHOpacityTo.new(backOutline, 0.05, fadeEase, 0.0),
                                 AHOpacityTo.new(frontOutline, 0.05, fadeEase, 0.0),
                                 AHOpacityTo.new(frontFilling, 0.05, fadeEase, 0.0)
                                 )

  local scaleEase = AHEaseSimpleIn.new(2.0)
  local flushAction = AHSpawn.new(
                                  AHLocationBy.new(self, 0.6, AHEaseExponentialIn.new(), deltaX, deltaY),
                                  AHRotationTo.new(self, 0.6, AHEaseExponentialOut.new(), destinationRotation),
                                  AHScaleXTo.new(self, 0.6, scaleEase, 0.2),
                                  AHScaleYTo.new(self, 0.6, scaleEase, 0.6),
                                  AHDelayed.new(0.55, fadeAction)
                                  )
  return flushAction:start()
end

function FoodItem:setOnTappedCallback(onTappedCallback)
  self.onTappedCallback = onTappedCallback
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FoodItem.new(amount, roundColor)
  local self = Group.new(DIAMETER, DIAMETER)
  Mixin.mix(self, FoodItem)

  local shakeGroup = FoodPieFraction.new(amount, roundColor, FRONT_FILLING_OPACITY, BACK_OUTLINE_OPACITY)
  shakeGroup:setEffectiveDims(DIAMETER, DIAMETER)
  self:insertAtCenter(shakeGroup)
  self.shakeGroup = shakeGroup
  self.frontOutline = shakeGroup.frontOutline
  self.frontFilling = shakeGroup.frontFilling
  self.backOutline = shakeGroup.backOutline

  self.roundColor = roundColor

  self.amount = amount
  self.onTappedCallback = nil
  self.allowRemoval = false
  self.isFrozen = false

  local function onTouchDown()
    if ((self.onTappedCallback ~= nil) and (self.allowRemoval)) then
      self.allowRemoval = false
      self.onTappedCallback()
    end
  end
  Button.makeButton(self, onTouchDown)

  return self
end
