--- FoodMathTipOverlay
-- @author Created by Jay Austria on January 12, 2015

class("FoodMathTipOverlay")

require("widgets.TapToContinue")
require("FoodTipsData")
require("FoodTipItem")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TIP_BACKING_COLOR_TOP = {0.1, 0.04, 0.06}
local TIP_BACKING_COLOR_BOTTOM = {0.3, 0.3, 0.3}
local INVISIBLE_OPACITY = 0.0
local VISIBLE_OPACITY = 1.0

local SUBJECT_LABEL_X = Screen.getSafeX(0.5)
local SUBJECT_LABEL_Y = Screen.getSafeY(0.9)
local SUBJECT_LABEL_FONT_SIZE = 14
local TIP_SUBJECT_LABEL_COLOR = {1.0, 1.0, 1.0}

local SUBTITLE_LABEL_X = Screen.getSafeX(0.5)
local SUBTITLE_LABEL_Y = SUBJECT_LABEL_Y - 20
local SUBTITLE_LABEL_FONT_SIZE = 14
local TIP_SUBTITLE_LABEL_COLOR = {0.5, 0.5, 0.5}

local TAP_INSTRUCTIONS_TEXT = Localizer.get("Misc.TAP_TO_CONTINUE")
local TAP_INSTRUCTIONS_LABEL_X = Screen.getSafeX(0.5)
local TAP_INSTRUCTIONS_LABEL_Y = Screen.getSafeY(0.05)
local TAP_INSTRUCTIONS_LABEL_FONT_SIZE = 13
local TAP_INSTRUCTIONS_LABEL_COLOR = ColorUtils.colorHex24ToStd(0xcccccc)

local STEPS_AREA_X = Screen.getSafeX(0.5)
local STEPS_AREA_Y_TOP = Screen.getSafeY(0.84)
local STEPS_COLUMN_SPACING_Y = Screen.getSafeHeightPart(0.01)
local STEPS_PADDING_Y = Screen.getSafeHeightPart(0.015)
local STEP_HEADER_FONT_SIZE = 15

local ILLUSTRATION_ROW_SPACING_X = 20
local EQUAL_SEPARATOR_FONT_SIZE = 32
local ILLUSTRATION_OUTLINE_COLOR = {0.4, 0.4, 0.4}
local ILLUSTRATION_OUTLINE_LINE_WIDTH = 0.5
local ILLUSTRATION_OUTLINE_BLUR_MARGIN = 0.5
local ILLUSTRATION_OUTLINE_BEVEL_AMOUNT = 2
local ILLUSTRATION_WIDTH = Screen.getSafeWidthPart(0.375)

local FORMULA_ROW_SPACING_X = 6
local FRACTION_LABEL_FONT_SIZE = 14
local GAP_PADDING_X = 10
local IMAGES_SPACING = 4
local IMAGES_SPACER_SIZE = 12

local EXAMPLE_FONT_SIZE = 13
local EXAMPLE_SPACING_X = 8

local ROW_COUNT_UNTIL_BREAK = 3
local FORCED_PADDING_FOR_CENTERING_STEP = Screen.getSafeHeightPart(0.12)
local MATH_OPERATOR_STRINGS =
{
  "×", "=", "÷"
}

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodMathTipOverlay:showAndWaitForDismissal()
  local fadeInDuration = 0.3
  local fadeIn = AHOpacityTo.new(self, fadeInDuration, nil, VISIBLE_OPACITY)

  wait(fadeIn:start())

  local tapLayer = TapToContinue.new()
  tapLayer:waitForTrigger()

  local fadeOutDuration = 0.1
  local fadeOut = AHOpacityTo.new(self, fadeOutDuration, nil, INVISIBLE_OPACITY)

  wait(fadeOut:start())
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getStrippedString(inputString)
  local startIndex = 1
  local endIndex = #inputString

  local hasLeadingWhitespace = (inputString[startIndex] == " ")
  local hasTrailingWhitespace = (inputString[endIndex] == " ")

  if hasLeadingWhitespace then
    startIndex = 2
  end

  if hasTrailingWhitespace then
    endIndex = #inputString - 1
  end

  return inputString:sub(startIndex, endIndex)
end

local function createSubjectLabel(self, tipSubject)
  local subjectText = string.format("TIP - %s", tipSubject)
  local subjectLabel = Label.new(subjectText, Font.new(FontFaces.DIN_OT_MEDIUM, SUBJECT_LABEL_FONT_SIZE))
  subjectLabel:setColor(unpack(TIP_SUBJECT_LABEL_COLOR))
  subjectLabel:setLoc(SUBJECT_LABEL_X, SUBJECT_LABEL_Y)

  self:insert(subjectLabel)
end

local function createSubtitleLabel(self, tipSubtitle)
  local subtitleLabel = Label.new(tipSubtitle, Font.new(FontFaces.DIN_OT_LIGHT, SUBTITLE_LABEL_FONT_SIZE))
  subtitleLabel:setColor(unpack(TIP_SUBTITLE_LABEL_COLOR))
  subtitleLabel:setLoc(SUBTITLE_LABEL_X, SUBTITLE_LABEL_Y)

  self:insert(subtitleLabel)
end

local function createTapInstructionsLabel(self)
  local instructionsLabel = Label.new(TAP_INSTRUCTIONS_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, TAP_INSTRUCTIONS_LABEL_FONT_SIZE))
  instructionsLabel:setColor(unpack(TAP_INSTRUCTIONS_LABEL_COLOR))

  instructionsLabel:setLoc(TAP_INSTRUCTIONS_LABEL_X, TAP_INSTRUCTIONS_LABEL_Y)

  self:insert(instructionsLabel)
end

local function createHeader(headerText)
  local headerLabel = Label.new(headerText, Font.new(FontFaces.DIN_OT_LIGHT, STEP_HEADER_FONT_SIZE))
  return headerLabel
end

local function createExample(exampleValueString)
  local exampleRow = Row.new(EXAMPLE_SPACING_X)

  local exampleString = "EXAMPLE:"
  local exampleLabel = Label.new(exampleString, Font.new(FontFaces.DIN_OT_MEDIUM, EXAMPLE_FONT_SIZE))

  local amount = FoodFraction.newWithString(exampleValueString)
  local fractionLabel = FoodFractionLabel.new(EXAMPLE_FONT_SIZE, amount)

  exampleRow:insert(exampleLabel)
  exampleRow:insert(fractionLabel)
  exampleRow:alignChildren()

  return exampleRow
end

local function createFractions(fractionStrings)
  local fractions = {}

  for _, fractionString in ipairs(fractionStrings) do
    local wholePart, numerator, denominator = FoodFraction.getComponentsFromString(fractionString)
    local newFraction = FoodTipItem.new(wholePart, numerator, denominator)

    table.insert(fractions, newFraction)
  end

  return fractions
end

local function splitFormulaByMathOperator(formula)
  for _, operator in ipairs(MATH_OPERATOR_STRINGS) do
    if string.find(formula, operator) then
      local formulaParts = StringUtils.split(formula, operator)
      return formulaParts, operator
    end
  end

  local formulaInTable = {formula}

  return formulaInTable
end

local function createIllustration(illustrationDescriptor)
  local displayedFractions = illustrationDescriptor.displayedFractions
  local fractionStrings = StringUtils.split(displayedFractions, " ")
  local fractionImages = createFractions(fractionStrings)

  local fractionImagesColumn = Column.new(IMAGES_SPACING, Alignment.LEFT)
  local fractionImagesRow = Row.new(IMAGES_SPACING)

  fractionImagesColumn:insertSpacer(IMAGES_SPACER_SIZE)
  fractionImagesRow:insertSpacer(IMAGES_SPACER_SIZE)

  local imageCount = #fractionImages

  for i = 1, imageCount do
    local fractionImage = fractionImages[i]
    fractionImagesRow:insert(fractionImage)

    local isLastImage = (i == imageCount)
    local isNewRowNeeded = (i % ROW_COUNT_UNTIL_BREAK == 0) and not isLastImage

    if isNewRowNeeded then
      fractionImagesRow:insertSpacer(IMAGES_SPACER_SIZE)
      fractionImagesRow:alignChildren()
      fractionImagesColumn:insert(fractionImagesRow)

      fractionImagesRow = Row.new(IMAGES_SPACING)
      fractionImagesRow:insertSpacer(IMAGES_SPACER_SIZE)
    end
  end

  fractionImagesRow:insertSpacer(IMAGES_SPACER_SIZE)
  fractionImagesRow:alignChildren()

  fractionImagesColumn:insert(fractionImagesRow)
  fractionImagesColumn:insertSpacer(IMAGES_SPACER_SIZE)
  fractionImagesColumn:alignChildren()

  local illustrationWidth = ILLUSTRATION_WIDTH
  local illustrationHeight = fractionImagesColumn:getHeight()

  local illustration = Group.new(illustrationWidth, illustrationHeight)
  illustration:insertAtCenter(fractionImagesColumn)

  local formula = illustrationDescriptor.formula
  local formulaParts, splittingOperator = splitFormulaByMathOperator(formula)

  local formulaRow = Row.new(FORMULA_ROW_SPACING_X)

  for i, formulaPart in ipairs(formulaParts) do
    formulaPart = getStrippedString(formulaPart)
    local willNeedSymbol = (#formulaParts == 2) and (i == 1)

    local fractionLabel = FoodFractionLabel.new(FRACTION_LABEL_FONT_SIZE)
    local wholePart, numerator, denominator = FoodFraction.getComponentsFromString(formulaPart)
    fractionLabel:setAmountFromComponents(wholePart, numerator, denominator)
    formulaRow:insert(fractionLabel)

    if willNeedSymbol then
      local symbolLabel = Label.new(splittingOperator, Font.new(FontFaces.DIN_OT_MEDIUM, FRACTION_LABEL_FONT_SIZE))
      formulaRow:insert(symbolLabel)
    end
  end

  formulaRow:alignChildren()
  formulaRow:setLoc(illustration:getWidthPart(0.5, 0))
  illustration:insert(formulaRow)
  illustration.formulaRow = formulaRow

  local width, height = illustration:getDims()
  local gapWidth = formulaRow:getWidth() + GAP_PADDING_X
  local outline = FoodHolderOutline.new(ILLUSTRATION_OUTLINE_COLOR, width, height, gapWidth, ILLUSTRATION_OUTLINE_LINE_WIDTH, ILLUSTRATION_OUTLINE_BLUR_MARGIN, ILLUSTRATION_OUTLINE_BEVEL_AMOUNT)
  outline:setRectangleHeight(height)
  illustration:insertAtCenter(outline)
  illustration.outline = outline

  return illustration
end

local function createIllustrationRow(illustrationDescriptors)
  local illustrationRow = Row.new(ILLUSTRATION_ROW_SPACING_X)
  local illustrations = {}


  local maxOutlineHeight = -math.huge
  for i, illustrationDescriptor in ipairs(illustrationDescriptors) do
    local willNeedSeparator = (#illustrationDescriptors == 2) and (i == 1)

    local illustration = createIllustration(illustrationDescriptor)
    illustrationRow:insert(illustration)
    table.insert(illustrations, illustration)
    maxOutlineHeight = math.max(maxOutlineHeight, illustration:getHeight())

    if willNeedSeparator then
      local equalSeparator = Label.new("=", Font.new(FontFaces.DIN_OT_MEDIUM, EQUAL_SEPARATOR_FONT_SIZE))
      illustrationRow:insert(equalSeparator)
    end
  end

  -- Kind of messy. Force the labels and outlines of this illustration row to match the height dimensions
  -- of the tallest illustration
  for _, illustration in ipairs(illustrations) do
    illustration.outline:setRectangleHeight(maxOutlineHeight)
    local labelOffsetY = (illustration:getHeight() - maxOutlineHeight) * 0.5
    local labelX, labelY = illustration.formulaRow:getLoc()
    illustration.formulaRow:setLoc(labelX, labelY + labelOffsetY)
  end

  illustrationRow:alignChildren()

  return illustrationRow
end


local function createLayout(self, tipDescriptor)
  local width, height = self:getDims()

  local backing = RectangleGradient.new(width, height, TIP_BACKING_COLOR_TOP, TIP_BACKING_COLOR_BOTTOM)
  self:insertAtCenter(backing)

  backing.beginColor:setAttrLink(MOAIColor.ATTR_OPACITY, self)
  backing.endColor:setAttrLink(MOAIColor.ATTR_OPACITY, self)

  self:setOpacity(INVISIBLE_OPACITY)

  self.backing = backing

  local subjectText = tipDescriptor.subjectText
  createSubjectLabel(self, subjectText)

  local subtitleText = tipDescriptor.subtitleText
  if subtitleText then
    createSubtitleLabel(self, subtitleText)
  end

  createTapInstructionsLabel(self)

  local stepsColumn = Column.new(STEPS_COLUMN_SPACING_Y, Alignment.LEFT)

  if #tipDescriptor.steps == 1 then
    stepsColumn:insertSpacer(FORCED_PADDING_FOR_CENTERING_STEP)
  end

  if tipDescriptor.example then
    local exampleLabel = createExample(tipDescriptor.example)
    stepsColumn:insert(exampleLabel)
  end

  for _, stepDescriptor in ipairs(tipDescriptor.steps) do
    local headerText = stepDescriptor.headerText
    local headerLabel = createHeader(headerText)
    stepsColumn:insert(headerLabel)

    local illustrationDescriptors = stepDescriptor.illustrations
    if illustrationDescriptors then
      local illustrationRow = createIllustrationRow(illustrationDescriptors)
      stepsColumn:insert(illustrationRow)
    end

    stepsColumn:insertSpacer(STEPS_PADDING_Y)
  end

  stepsColumn:setAnchorPointTop()
  stepsColumn:alignChildren()
  stepsColumn:setLoc(STEPS_AREA_X, STEPS_AREA_Y_TOP)
  self:insert(stepsColumn)

  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function FoodMathTipOverlay.newHidden(tipType)
  local self = Group.newFullScreen()
  Mixin.mix(self, FoodMathTipOverlay)

  local tipDescriptor = FoodTipsData.getTipDescriptor(tipType)
  assert(tipDescriptor, "No tip layout defined for tip type" .. tostring(tipType))

  createLayout(self, tipDescriptor)

  return self
end
