--- FoodPanelRound
-- @author Created by Aaron Barrett on December 09, 2014.

class("FoodPanelRound")

require("FoodHolder")
require("FoodAnswerOptionBar")
require("FoodConstants")
require("FoodRoundColors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GOOD_LABEL_RISE_HEIGHT = Screen.getSafeHeightPart(0.07)

local Z_ANSWER_OPTION_BAR = 10
local Z_HOLDER = 20
local Z_TOO_HIGH_LABEL = 30
local Z_TOO_LOW_LABEL = 40
local Z_UNNECESSARY_PIECE_LABEL = 50
local Z_GOOD_LABEL = 60
local Z_PERFECT_LABEL = 70
local Z_HIGHLIGHT_ARROW = 80

local NUMBERS_EXPRESSED_AS_STRINGS = {
  [1] = "one",
  [2] = "two",
  [3] = "three",
  [4] = "four",
  [5] = "five",
  [6] = "six",
  [7] = "seven",
  [8] = "eight",
  [9] = "nine",
  [10] = "ten",
  [11] = "eleven",
  [12] = "twelve"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Message Animations -----------------

local function getPluralizedFormOfTheWordPieces(pieceCount)
  if (pieceCount == 1) then
    return "piece"
  else
    return "pieces"
  end
end

local function expressNumberAsString(numberToExpress)
  local numberExpressedAsString = NUMBERS_EXPRESSED_AS_STRINGS[numberToExpress]
  if (numberExpressedAsString == nil) then
    numberExpressedAsString = tostring(numberToExpress)
  end
  return numberExpressedAsString
end

local function getUnnecessaryPieceString(unnecessaryPieceCount)
  local unnecessaryPieceString = expressNumberAsString(unnecessaryPieceCount)
                                  .. " unnecessary "
                                  .. getPluralizedFormOfTheWordPieces(unnecessaryPieceCount)
                                  .. " used"
  return unnecessaryPieceString
end

local function playShowGood(self, unnecessaryPieceCount)
  local goodLabel = self.goodLabel
  local unnecessaryPieceLabel = self.unnecessaryPieceLabel
  local holder = self.holder

  local unnecessaryPieceString = getUnnecessaryPieceString(unnecessaryPieceCount)
  unnecessaryPieceLabel:setText(unnecessaryPieceString)

  local holderTopY = holder:getTopY()
  local labelSeparationDistance = Screen.getSafeHeightPart(0.04)

  goodLabel:setLoc(Screen.getSafeX(0.5), holderTopY)
  goodLabel:setOpacity(0.0)
  goodLabel:setVisible(true)
  unnecessaryPieceLabel:setLoc(self:getWidthPart(0.5), holderTopY - labelSeparationDistance)
  unnecessaryPieceLabel:setOpacity(0.0)
  unnecessaryPieceLabel:setVisible(true)

  local distanceToRise = GOOD_LABEL_RISE_HEIGHT

  local goodRiseAction = AHLocationYBy.new(goodLabel, 0.8, AHEaseSimpleInOut.new(4.0), distanceToRise)
  local goodFadeAction = AHOpacityTo.new(goodLabel, 0.8, AHEaseSimpleIn.new(4.0), 1.0)

  local unnecessaryRiseAction = AHLocationYBy.new(unnecessaryPieceLabel, 0.8, AHEaseSimpleInOut.new(4.0), distanceToRise)
  local unnecessaryFadeAction = AHOpacityTo.new(unnecessaryPieceLabel, 0.8, AHEaseSimpleIn.new(4.0), 1.0)

  local messageAction = AHSpawn.new(
                                    goodRiseAction,
                                    goodFadeAction,
                                    unnecessaryRiseAction,
                                    AHDelayed.new(0.2, unnecessaryFadeAction)
                                    )

  return messageAction:start()
end

local function playShowSingleLabelMessage(self, label)
  local holder = self.holder
  local holderTopY = holder:getTopY()

  local distanceToRise = Screen.getSafeHeightPart(0.04)

  label:setLoc(Screen.getSafeX(0.5), holderTopY)
  label:setOpacity(0.0)
  label:setVisible(true)

  local riseAction = AHSpawn.new(
                                 AHLocationYBy.new(label, 0.8, AHEaseSimpleInOut.new(4.0), distanceToRise),
                                 AHOpacityTo.new(label, 0.8, AHEaseSimpleIn.new(4.0), 1.0)
                                 )
  return riseAction:start()
end

local function playShowPerfect(self)
  local perfectLabel = self.perfectLabel
  return playShowSingleLabelMessage(self, perfectLabel)
end

local function playClearMessageLabels(self, ...)
  local labels = {...}

  local clearAction = AHSpawn.new()

  for _, label in ipairs(labels) do
    local labelAction = AHOpacityTo.new(label, 0.7, AHEaseSimpleIn.new(2.0), 0.0)
    clearAction:addAction(labelAction)
  end

  return clearAction:start()
end

local function playClearPerfect(self)
  local perfectLabel = self.perfectLabel
  return playClearMessageLabels(self, perfectLabel)
end

local function playClearGood(self)
  local goodLabel = self.goodLabel
  local unnecessaryPieceLabel = self.unnecessaryPieceLabel
  return playClearMessageLabels(self, goodLabel, unnecessaryPieceLabel)
end

-- Tutorial ---------------------------

local function removeHighlight(self)
  if self.currentHighlightArrow then
    self.currentHighlightArrow:removeSelf()
  end
end

local function highlightButton(self, button)
  removeHighlight(self)

  local highlightArrow = Sprite.newWithName("shared_arrow.png")
  highlightArrow:setRot(180)
  highlightArrow:setColor(unpack(self.roundColor))

  self:insert(highlightArrow, Z_HIGHLIGHT_ARROW)

  local buttonX, buttonY = self:transformObjectLoc(button)

  local arrowX = buttonX
  local arrowY = buttonY + 65
  highlightArrow:setLoc(arrowX, arrowY)

  local hoverEase = AHEaseSimpleInOut.new(2.0)
  local hoverAmount = Screen.getSafeHeightPart(0.008)
  local hoverDuration = 0.7
  local hoverAction = AHSequence.new(AHLocationYBy.new(highlightArrow, hoverDuration, hoverEase, hoverAmount),
                                     AHLocationYBy.new(highlightArrow, hoverDuration, hoverEase, -hoverAmount))
  local hoverLoop = AHLoop.new(hoverAction)

  hoverLoop:start()

  self.currentHighlightArrow = highlightArrow
end

local function updateWalkThroughStep(self)
  local answerOptionBar = self.answerOptionBar
  local stepIndex = self.holder:getItemCount() + 1

  if stepIndex <= #self.walkThroughSteps then
    local step = self.walkThroughSteps[stepIndex]
    local answerOptionButton = answerOptionBar:walkThroughAnswerOptionInteraction(step)
    highlightButton(self, answerOptionButton)
  else
    answerOptionBar:endWalkThrough()

    removeHighlight(self)

    local holder = self.holder
    holder:tutorialEnableSubmitButton()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodPanelRound:getHolder()
  return self.holder
end

function FoodPanelRound:getAnswerBar()
  return self.answerOptionBar
end

-- Correct Animation ------------------

function FoodPanelRound:playCorrect(submittedSteps, optimalSteps)
  return runInCoroutine(function()
    local holder = self.holder

    local submittedStepsCount = #submittedSteps
    local optimalStepsCount = #optimalSteps
    local unnecessaryPieceCount = submittedStepsCount - optimalStepsCount

    local shouldShowPerfect = (unnecessaryPieceCount == 0)

    if (shouldShowPerfect) then
      playShowPerfect(self)
    else
      playShowGood(self, unnecessaryPieceCount)
    end

    wait(holder:playCorrect())

    local didShowPerfect = shouldShowPerfect
    if (didShowPerfect) then
      waitFor(0.4)
      playClearPerfect(self)
    else
      wait(holder:playReduce(optimalSteps))
      playClearGood(self)
    end
  end)
end

-- Strobe Animation -------------------

function FoodPanelRound:playStrobe()
  local perfectLabel = self.perfectLabel
  local holder = self.holder

  local perfectLabelStrobeAction = FoodStrobeAction.new(perfectLabel)
  perfectLabelStrobeAction:start()

  return holder:playStrobe()
end

-- Wrong Animation --------------------

function FoodPanelRound:playWrong(wasAnswerTooHigh)
  local holder = self.holder
  local answerOptionBar = self.answerOptionBar
  local roundColor = self.roundColor

  answerOptionBar:playWrong()

  holder:playWrong()

  local comparisonLabel = nil
  if (wasAnswerTooHigh) then
    comparisonLabel = self.tooHighLabel
  else
    comparisonLabel = self.tooLowLabel
  end

  local redRemainDuration = 1.55

  comparisonLabel:setColor(unpack(FoodConstants.WRONG_RED_COLOR))
  local colorAction = AHSequence.new(
                                     AHColorTo.new(comparisonLabel, 0.1, nil, FoodConstants.WRONG_RED_COLOR),
                                     AHColorTo.new(comparisonLabel, redRemainDuration, AHEaseSimpleOut.new(1.4), roundColor)
                                     )
  colorAction:start()


  return runInCoroutine(function()
      wait(playShowSingleLabelMessage(self, comparisonLabel))
      waitFor(1)
      wait(playClearMessageLabels(self, comparisonLabel))
    end)
end

function FoodPanelRound:allowRetry()
  Sound.playSoundEffect("deselect")

  self.submittedAmount = nil
  self.submittedSteps = nil
end

-- Appear Animation -------------------

function FoodPanelRound:playAppear()
  local answerOptionBar = self.answerOptionBar
  return answerOptionBar:playAppear()
end

-- Dismiss Animation ------------------

function FoodPanelRound:playHideAnswerOptionButtons()
  local answerOptionBar = self.answerOptionBar
  return answerOptionBar:playDisappear()
end

function FoodPanelRound:playDismissWithProgressIndicator(progressIndicator)
  local holder = self.holder

  local isGameFinished = (progressIndicator:isOnLastDot())
  if (isGameFinished) then
    return runInCoroutine(function()
        holder:playFlushWithProgressIndicator(progressIndicator)
        holder:playFadeOut()
        waitFor(0.7)
      end)
  else
    return runInCoroutine(function()
        holder:playFlushWithProgressIndicator(progressIndicator)
        progressIndicator:incrementProgress()
        wait(holder:playFadeOut())
        waitFor(0.3)
        self:removeSelf()
      end)
  end
end

-- Loss Animation ---------------------

function FoodPanelRound:playLoss()
  self:playHideAnswerOptionButtons()

  local holder = self.holder
  return holder:playLoss()
end

-- Tutorial Animations -----------------
function FoodPanelRound:enableTutorial()
  local holder = self.holder
  holder:enableTutorial()

  local answerOptionBar = self.answerOptionBar
  answerOptionBar:displayForTutorial()
end

-- Submitted Answer -------------------

function FoodPanelRound:getSubmittedAmount()
  return self.submittedAmount
end

function FoodPanelRound:getSubmittedSteps()
  return self.submittedSteps
end

-- Touch Management -------------------

function FoodPanelRound:setTouchEnabled(touchEnabled)
  local answerOptionBar = self.answerOptionBar
  local holder = self.holder

  answerOptionBar:setTouchEnabled(touchEnabled)
  holder:setTouchEnabled(touchEnabled)
end

-- Tutoral -----------------------------
function FoodPanelRound:walkThroughOptimalSteps(optimalSteps)
  self.isInStepsWalkthrough = true
  self.walkThroughSteps = optimalSteps

  updateWalkThroughStep(self)
end

function FoodPanelRound:endTutorial()
  local holder = self.holder
  holder:endTutorial()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createOnAnswerOptionButtonPressedCallback(self)
  local function onAnswerOptionButtonPressedCallback(answerOptionButton)
    self.hasPressedButton = true
    local holder = self.holder
    if (holder:canAcceptMoreItems()) then
      Sound.playSoundEffect("select")
      holder:playAddItemFromButton(answerOptionButton)
      if self.isInStepsWalkthrough then
        updateWalkThroughStep(self)
      end
    else
      Sound.playSoundEffect("invalid")
    end
  end
  return onAnswerOptionButtonPressedCallback
end

local function createSubmitCallback(self)
  local function submitCallback(submittedAmount, submittedSteps)
    self:setTouchEnabled(false)
    self.submittedAmount = submittedAmount
    self.submittedSteps = submittedSteps
  end
  return submitCallback
end

-- Constructor ------------------------

function FoodPanelRound.new(round, roundIndex, maxHeight)
  local self = Group.newFullScreen()
  Mixin.mix(self, FoodPanelRound)

  local roundColor = FoodRoundColors.getColorForRoundIndex(roundIndex)

  local onAnswerOptionButtonPressedCallback = createOnAnswerOptionButtonPressedCallback(self)

  local answerOptionBar = FoodAnswerOptionBar.new(round, roundColor, onAnswerOptionButtonPressedCallback)
  local answerOptionBarY = -answerOptionBar:getHeightPart(0.5)
  answerOptionBar:setLoc(self:getWidthPart(0.5), answerOptionBarY)
  self:insert(answerOptionBar, Z_ANSWER_OPTION_BAR)
  self.answerOptionBar = answerOptionBar

  local submitCallback = createSubmitCallback(self)

  local holder = FoodHolder.new(roundColor, submitCallback)
  holder:hideInitially()
  holder:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  maxHeight = maxHeight - GOOD_LABEL_RISE_HEIGHT
  if (holder:getYTop() > maxHeight) then
    holder:offsetY(maxHeight - holder:getYTop())
  end
  self:insert(holder, Z_HOLDER)
  self.holder = holder

  local unnecessaryPieceLabel = Label.new("", Font.new(FontFaces.DIN_OT_BOLD, 15.0))
  unnecessaryPieceLabel:setColor(unpack(roundColor))
  unnecessaryPieceLabel:setOpacity(0.0)
  unnecessaryPieceLabel:setVisible(false)
  self:insert(unnecessaryPieceLabel, Z_UNNECESSARY_PIECE_LABEL)
  self.unnecessaryPieceLabel = unnecessaryPieceLabel

  local goodLabel = Label.new("GOOD", Font.new(FontFaces.DIN_OT_BOLD, 25.0))
  goodLabel:setColor(unpack(roundColor))
  goodLabel:setOpacity(0.0)
  goodLabel:setVisible(false)
  self:insert(goodLabel, Z_GOOD_LABEL)
  self.goodLabel = goodLabel

  local perfectLabel = Label.new("PERFECT", Font.new(FontFaces.DIN_OT_BOLD, 25.0))
  perfectLabel:setColor(unpack(roundColor))
  perfectLabel:setOpacity(0.0)
  perfectLabel:setVisible(false)
  self:insert(perfectLabel, Z_PERFECT_LABEL)
  self.perfectLabel = perfectLabel

  local tooHighLabel = Label.new("TOO HIGH", Font.new(FontFaces.DIN_OT_BOLD, 25.0))
  tooHighLabel:setOpacity(0.0)
  tooHighLabel:setVisible(false)
  self:insert(tooHighLabel, Z_TOO_HIGH_LABEL)
  self.tooHighLabel = tooHighLabel

  local tooLowLabel = Label.new("TOO LOW", Font.new(FontFaces.DIN_OT_BOLD, 25.0))
  tooLowLabel:setOpacity(0.0)
  tooLowLabel:setVisible(false)
  self:insert(tooLowLabel, Z_TOO_LOW_LABEL)
  self.tooLowLabel = tooLowLabel

  self:setTouchEnabled(false)

  self.round = round
  self.roundColor = roundColor
  self.submittedAmount = nil
  self.submittedSteps = nil
  self.roundColor = roundColor
  self.hasPressedButton = false

  return self
end
