--- FoodProgressIndicator
-- @author Created by Aaron Barrett on November 13, 2014.

class("FoodProgressIndicator")

require("widgets.progress.ProgressIndicator")
require("FoodRoundColors")
require("FoodConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUNDS_PER_GAME = FoodConfiguration.SETS_TO_COMPLETE_COUNT

local FOCUSED_DOT_WIDTH = 88.0
local SMALL_DOT_WIDTH = 16.0
local DOT_SPACING = Screen.getSafeWidthPart(0.03)
local EXPANDED_DOT_SPACING = Screen.getSafeWidthPart(0.034)

local WIDTH = (ROUNDS_PER_GAME - 1) * (DOT_SPACING + SMALL_DOT_WIDTH) + FOCUSED_DOT_WIDTH
local HEIGHT = FOCUSED_DOT_WIDTH

local DOT_FLY_START_DELAY_DURATIONS = {
  [1] = 0.13,
  [2] = 0.00,
  [3] = 0.20,
  [4] = 0.22,
  [5] = 0.05,
  [6] = 0.14
}

local ALL_DOTS_FLY_IN_DURATION = 1.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getDotXCoordinatesWithSpacing(self, spacing)
  local dots = self.dots

  local dotsCount = #dots

  local focusedDotIndex = self:getProgressStep() + 1
  local focusedDotCenterX = ((focusedDotIndex / (dotsCount + 1)) * 9.0 + 0.5) / 10.0 * WIDTH

  local xCoordinates = {}

  for dotIndex, dot in ipairs(dots) do

    local indexDifference = dotIndex - focusedDotIndex

    local x = focusedDotCenterX + indexDifference * (spacing + SMALL_DOT_WIDTH)

    local isLeftOfFocus = (indexDifference < 0)
    local isRightOfFocus = (indexDifference > 0)

    if (isLeftOfFocus) then
      x = x - 0.4 * FOCUSED_DOT_WIDTH
    elseif (isRightOfFocus) then
      x = x + 0.4 * FOCUSED_DOT_WIDTH
    end

    table.insert(xCoordinates, x)
  end

  return xCoordinates
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FoodProgressIndicator:isOnLastDot()
  local dots = self.dots
  local dotsCount = #dots
  local progressStep = self:getProgressStep()
  return (progressStep == dotsCount - 1)
end

function FoodProgressIndicator:getDots()
  return self.dots
end

function FoodProgressIndicator:getFlushDestinationLocation()
  local x, y = self:getLoc()
  local focusedDot = self:getFocusedDot()
  local focusedDotX, focusedDotY = focusedDot:getLoc()

  local destinationX = x - self:getWidthPart(0.5) + focusedDotX
  local destinationY = y - self:getHeightPart(0.5) + focusedDotY

  return destinationX, destinationY
end

function FoodProgressIndicator:setProgressStep(step)
  self.super.setProgressStep(self, step)

  local dots = self.dots
  local originalDotIndex = self:getProgressStep()
  local focusedDotIndex = self:getProgressStep() + 1

  local xCoordinates = getDotXCoordinatesWithSpacing(self, DOT_SPACING)
  local expandedXCoordinates = getDotXCoordinatesWithSpacing(self, EXPANDED_DOT_SPACING)

  local dotToGrow = dots[focusedDotIndex]
  local colorOfDotToGrow = dotToGrow:getRoundColor()

  for dotIndex, dot in ipairs(dots) do
    local isDotToGrow = (dotIndex == focusedDotIndex)
    local isDotToShrink = (dotIndex == originalDotIndex)
    local isRightOfFocus = (dotIndex > focusedDotIndex)

    local destinationX = xCoordinates[dotIndex]
    local expandedX = expandedXCoordinates[dotIndex]

    if (isDotToGrow) then
      dot:playGrow(destinationX)
    elseif (isDotToShrink) then
      dot:playShrink(destinationX)
    elseif (isRightOfFocus) then
      dot:playAdoptRoundColor(colorOfDotToGrow)
    end

    local needsLocationAction = (not (isDotToGrow or isDotToShrink))
    if (needsLocationAction) then
      local distanceFactor = 0.035 * math.abs(originalDotIndex - dotIndex)
      local locationDelayDuration = 0.98 + distanceFactor

      local xLocationAction = AHSequence.new(
                                             AHDelay.new(locationDelayDuration),
                                             AHLocationXTo.new(dot, 0.35, AHEaseSimpleInOut.new(2.0), expandedX),
                                             AHLocationXTo.new(dot, 0.45, AHEaseSimpleInOut.new(2.0), destinationX)
                                             )
      xLocationAction:start()
    end
  end
end

function FoodProgressIndicator:playTakeDots()
  return runInCoroutine(function()
    local dots = self.dots

    local dotXCoordinates = getDotXCoordinatesWithSpacing(self, DOT_SPACING)

    for dotIndex, dot in ipairs(dots) do
      self:takeChild(dot)

      runInCoroutine(function()
        local destinationX = dotXCoordinates[dotIndex]
        local destinationY = self:getHeightPart(0.5)

        local firstRoundIndex = 1
        local destinationRoundColor = FoodRoundColors.getColorForRoundIndex(firstRoundIndex)

        local shouldShrink = (dotIndex ~= firstRoundIndex)

        local dotFlyStartDelayDuration = DOT_FLY_START_DELAY_DURATIONS[dotIndex]
        assert(dotFlyStartDelayDuration ~= nil, "No dot fly start delay duration defined")
        waitFor(dotFlyStartDelayDuration)

        dot:playFlyToDestination(destinationX, destinationY, destinationRoundColor, shouldShrink)
      end)
    end

    waitFor(ALL_DOTS_FLY_IN_DURATION)
  end)
end

function FoodProgressIndicator:getFocusedDot()
  local dots = self.dots
  local focusedDotIndex = self:getProgressStep() + 1
  local focusedDot = dots[focusedDotIndex]
  return focusedDot
end

-- Wrong Animation --------------------

function FoodProgressIndicator:playWrong()
  local dots = self.dots

  local focusedDotIndex = self:getProgressStep() + 1
  local focusedDot = self:getFocusedDot()
  local focusedDotColor = focusedDot:getRoundColor()

  for dotIndex, dot in ipairs(dots) do
    local hasAlreadyPassedDot = (dotIndex < focusedDotIndex)

    local colorToRevertTo = nil
    if (hasAlreadyPassedDot) then
      colorToRevertTo = dot:getRoundColor()
    else
      colorToRevertTo = focusedDotColor
    end

    dot:playWrong(colorToRevertTo)
  end
end

-- Loss Animation ---------------------

function FoodProgressIndicator:playLoss()
  local dots = self.dots

  local ease = AHEaseBackIn.new(0.9)

  local staggerDuration = 0.01
  local lossAction = AHStaggeredSpawn.new(staggerDuration)

  for _, dot in ipairs(dots) do
    local dotAction = AHScaleTo.new(dot, 0.4, ease, 0.0)
    lossAction:addAction(dotAction)
  end

  return lossAction:start()
end

function FoodProgressIndicator:displayForTutorial()
  for _, dot in ipairs(self.dots) do
    dot:playMorphToTutorialColor()
  end
end

function FoodProgressIndicator:displayForGame()
  local dots = self.dots
  local focusedDotIndex = self:getProgressStep() + 1
  local focusedDot = dots[focusedDotIndex]
  local currentRoundColor = focusedDot:getRoundColor()


  for dotIndex, dot in ipairs(dots) do
    local usesOwnColor = (dotIndex <= focusedDotIndex)
    local usesCurrentRoundColor = (dotIndex > focusedDotIndex)

    if (usesOwnColor) then
      dot:playMorphToOwnColor()
    elseif (usesCurrentRoundColor) then
      dot:playMorphToColor(currentRoundColor)
    end
  end
end

function FoodProgressIndicator:markEfficientAnswer()
  local focusedDot = self:getFocusedDot()
  focusedDot:playEfficientAnswer()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FoodProgressIndicator.new(dots)
  local self = Group.new(WIDTH, HEIGHT)

  local dotsCount = #dots
  ProgressIndicator.makeProgressIndicator(self, dotsCount)

  Mixin.mix(self, FoodProgressIndicator)

  self.dots = dots

  return self
end
