--- FunkyBackground
-- @author Created by Aaron Barrett on February 25, 2016.

class("FunkyBackground")

require("FunkyDiamond")
require("FunkyOutroRibbonOverlay")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

FunkyBackground.INDEX_TOP_DIAMOND = 1
FunkyBackground.INDEX_TOP_LEFT_DIAMOND = 2
FunkyBackground.INDEX_TOP_RIGHT_DIAMOND = 3
FunkyBackground.INDEX_BOTTOM_DIAMOND = 4
FunkyBackground.INDEX_BOTTOM_LEFT_DIAMOND = 5
FunkyBackground.INDEX_BOTTOM_RIGHT_DIAMOND = 6
FunkyBackground.INDEX_EXTRA_CHECKMARK_DIAMOND = 7

FunkyBackground.SCALE_AFTER_INTRO = 0.8

local COLOR_BLACK = {0.0, 0.0, 0.0, 1.0}
local COLOR_WHITE = {1.0, 1.0, 1.0, 1.0}

local CLOUD_COLORS = {
  {0.11, 0.11, 0.95},
  {0.85, 0.0, 1.0}
}

local CLOUD_COLORS_GREEN = {
  {0.0, 0.88, 1.0},
  {0.17, 1.0, 0.81}
}

local CLOUD_COLORS_PINK = {
  {0.98, 0.0, 0.52},
  {0.98, 0.0, 0.29}
}

local Z_DIAMOND = 10
local Z_CLOUD = 20
local Z_MASK_OVERLAY = 30
local Z_OUTRO_RIBBON_OVERLAY = 40

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createOutroPanAction(self)
  local minimumScale = 0.6
  local settledScale = 0.667

  local ease = AHEaseSineInOut.new()

  local panAction = AHSequence.new(
    AHLocationXBy.new(self, 0.2, ease, ReferenceDevice.getScaledWidth(12.0)),
    AHSpawn.new(
      AHLocationXBy.new(self, 0.2, ease, -ReferenceDevice.getScaledWidth(48.0)),
      AHScaleTo.new(self, 0.2, ease, minimumScale)
    ),
    AHScaleTo.new(self, 0.2, ease, settledScale)
  )

  return panAction
end

local function buildOutroRibbonOverlay(self)
  local checkmarkDiamonds = self.checkmarkDiamonds
  local outroRibbonOverlay = FunkyOutroRibbonOverlay.new(self:getWidth(), self:getHeight(), checkmarkDiamonds)
  self:insertAtCenter(outroRibbonOverlay, Z_OUTRO_RIBBON_OVERLAY)
  self.outroRibbonOverlay = outroRibbonOverlay
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyBackground:getCentralDiamonds()
  return self.centralDiamonds
end

function FunkyBackground:playIntro()
  for _, diamond in ipairs(self.diamonds) do
    diamond:playIntro()
  end

  local introAction = AHSpawn.new(
    AHColorTo.new(self, 0.9, nil, COLOR_BLACK),
    AHScaleTo.new(self, 0.9, AHEaseExponentialOut.new(), FunkyBackground.SCALE_AFTER_INTRO)
  )

  return introAction:start()
end

function FunkyBackground:playCorrect()
  local correctAction = AHSpawn.new()

  for _, diamond in ipairs(self.diamonds) do
    correctAction:addAction(diamond:createCorrectAction())
  end

  for cloudIndex, cloud in ipairs(self.clouds) do
    local cloudColorGreen = CLOUD_COLORS_GREEN[cloudIndex]
    local cloudColor = CLOUD_COLORS[cloudIndex]
    local cloudAction = AHSequence.new(
      AHColorTo.new(cloud, 0.1, nil, cloudColorGreen),
      AHDelay.new(0.25),
      AHColorTo.new(cloud, 0.8, nil, cloudColor)
    )
    correctAction:addAction(cloudAction)
  end

  self.actionInterrupt:insert(AHDelayed.new(0.05, correctAction))
end

function FunkyBackground:playWrong()
  local wrongAction = AHSpawn.new()

  for _, diamond in ipairs(self.diamonds) do
    wrongAction:addAction(diamond:createWrongAction())
  end

  for cloudIndex, cloud in ipairs(self.clouds) do
    local cloudColorPink = CLOUD_COLORS_PINK[cloudIndex]
    local cloudColor = CLOUD_COLORS[cloudIndex]
    local cloudAction = AHSequence.new(
      AHColorTo.new(cloud, 0.4, nil, cloudColorPink),
      AHColorTo.new(cloud, 0.4, nil, cloudColor)
    )
    wrongAction:addAction(cloudAction)
  end

  self.actionInterrupt:insert(wrongAction)
end

function FunkyBackground:playOutro()
  buildOutroRibbonOverlay(self)
  self.outroRibbonOverlay:playOutro()

  local outroAction = AHSpawn.new()

  for _, diamond in ipairs(self.nonCheckmarkDiamonds) do
    outroAction:addAction(diamond:createNonCheckmarkOutroAction())
  end

  for _, diamond in ipairs(self.checkmarkDiamonds) do
    outroAction:addAction(diamond:createCheckmarkOutroAction())
  end

  for cloudIndex, cloud in ipairs(self.clouds) do
    local cloudColorGreen = CLOUD_COLORS_GREEN[cloudIndex]
    local cloudAction = AHColorTo.new(cloud, 0.1, nil, cloudColorGreen)
    outroAction:addAction(cloudAction)
  end

  local panAction = createOutroPanAction(self)
  outroAction:addAction(panAction)

  self.actionInterrupt:insert(outroAction)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildDiamondDimensions(self)
  local sampleDiamond = FunkyDiamond.new()
  self.diamondSize = sampleDiamond:getWidth()
end

local function buildMaskOverlay(self)
  local maskOverlay = Group.new(self:getWidth(), self:getHeight())
  self:insertAtCenter(maskOverlay, Z_MASK_OVERLAY)
  self.maskOverlay = maskOverlay
end

local function buildDiamond(self, centerX, centerY)
  local diamond = FunkyDiamond.new()

  local maskDiamond = Sprite.newWithName("diamond_mask.png")
  maskDiamond:inheritColor(self)
  maskDiamond:setRot(45.0)
  local scaleToPreventSeams = 1.01
  maskDiamond:setScl(scaleToPreventSeams)

  diamond:setLoc(centerX, centerY)
  maskDiamond:setLoc(centerX, centerY)

  self:insert(diamond, Z_DIAMOND)
  self.maskOverlay:insert(maskDiamond)
  
  return diamond
end

local function buildDiamondsGrid(self)
  local diamondSize = self.diamondSize
  -- Note: FunkyBackground is larger than the full screen size
  local width = self:getWidth()
  local height = self:getHeight()

  -- The "centerDiamond" is vertically offset to center all diamonds that are buttons
  local centerDiamondX = width * 0.5
  local centerDiamondY = (height * 0.5) - (diamondSize / 4)

  local totalColumnCount = (math.ceil(width / diamondSize) * 2) + 1
  local totalRowCount = math.ceil(height / diamondSize) + 1

  local rowCountToCenter = math.ceil(totalRowCount / 2)
  local startingY = centerDiamondY - (rowCountToCenter * diamondSize)

  local columnCountLeftOfCenter = (totalColumnCount - 1) / 2  -- totalColumnCount is always odd
  local startingX = centerDiamondX - ((columnCountLeftOfCenter+1) * diamondSize/2)

  self.centerColumnIndex = columnCountLeftOfCenter + 1
  self.centerRowIndex = rowCountToCenter

  local diamondsGrid = {}
  for columnIndex = 1, totalColumnCount do
    diamondsGrid[columnIndex] = {}
    for rowIndex = 1, totalRowCount do
      -- Each row is a full diamond apart, but each column is only a half-diamond apart
      -- Every other row is then offset vertically by half a diamond so everything fits together
      local x = startingX + (columnIndex / 2 * diamondSize)
      local y = startingY + (rowIndex * diamondSize)

      -- Vertically offset every other row, but the center column should not be one of those offset
      if MathUtils.isEven(columnIndex + columnCountLeftOfCenter) then
        y = y + (diamondSize * 0.5)
      end

      local diamond = buildDiamond(self, x, y)
       diamondsGrid[columnIndex][rowIndex] = diamond
    end
  end

  return diamondsGrid
end


local function getCentralDiamonds(self)
  local centerColumnIndex = self.centerColumnIndex
  local centerRowIndex = self.centerRowIndex

  local topDiamond = self.diamondsGrid[centerColumnIndex][centerRowIndex + 1]
  local topLeftDiamond = self.diamondsGrid[centerColumnIndex - 1][centerRowIndex]
  local topRightDiamond = self.diamondsGrid[centerColumnIndex + 1][centerRowIndex]

  local bottomDiamond = self.diamondsGrid[centerColumnIndex][centerRowIndex]
  local bottomLeftDiamond = self.diamondsGrid[centerColumnIndex - 1][centerRowIndex - 1]
  local bottomRightDiamond = self.diamondsGrid[centerColumnIndex + 1][centerRowIndex - 1]

  local extraCheckmarkDiamond = self.diamondsGrid[centerColumnIndex + 2][centerRowIndex + 1]

  return {topDiamond, topLeftDiamond, topRightDiamond, bottomDiamond, bottomLeftDiamond, bottomRightDiamond, extraCheckmarkDiamond}
end

local function getCheckmarkDiamonds(self)
  local centralDiamonds = self.centralDiamonds
  return {
    centralDiamonds[FunkyBackground.INDEX_TOP_LEFT_DIAMOND],
    centralDiamonds[FunkyBackground.INDEX_BOTTOM_DIAMOND],
    centralDiamonds[FunkyBackground.INDEX_TOP_RIGHT_DIAMOND],
    centralDiamonds[FunkyBackground.INDEX_EXTRA_CHECKMARK_DIAMOND]
  }
end

local function getNonCheckmarkDiamonds(self)
  local nonCheckmarkDiamonds = {}
  local checkmarkDiamonds = self.checkmarkDiamonds
  for _, diamond in ipairs(self.diamonds) do
    if (not TableUtils.contains(checkmarkDiamonds, diamond)) then
      table.insert(nonCheckmarkDiamonds, diamond)
    end
  end
  return nonCheckmarkDiamonds
end

local function getAllDiamonds(grid)
  local diamonds = {}
  for _, column in ipairs(grid) do
    for _, diamond in ipairs(column) do
      table.insert(diamonds, diamond)
    end
  end
  return diamonds
end

local function buildDiamonds(self)
  self.diamondsGrid = buildDiamondsGrid(self)

  self.diamonds = getAllDiamonds(self.diamondsGrid)
  self.centralDiamonds = getCentralDiamonds(self)
  self.checkmarkDiamonds = getCheckmarkDiamonds(self)
  self.nonCheckmarkDiamonds = getNonCheckmarkDiamonds(self)
end

local function buildClouds(self)
  local clouds = {}
  for cloudIndex, color in ipairs(CLOUD_COLORS) do
    local cloudFilename = "cloud_" .. tostring(cloudIndex) .. ".png"
    local cloud = Sprite.newWithName(cloudFilename)
    cloud:setBlendMode(MOAIProp2D.BLEND_ADD)
    cloud:setColor(unpack(color))
    cloud:setOpacity(0.5)
    cloud:setEffectiveDims(self:getWidthPart(1.1), self:getHeightPart(1.1))
    self:insertAtCenter(cloud, Z_CLOUD)
    table.insert(clouds, cloud)
  end

  self.clouds = clouds
end

-- Constructor ------------------------

function FunkyBackground.new(width, height)
  local self = FilledRectangle.new(width, height, COLOR_WHITE)
  Mixin.mix(self, FunkyBackground)

  buildMaskOverlay(self)
  buildDiamondDimensions(self)
  buildDiamonds(self)
  buildClouds(self)

  self.actionInterrupt = ActionInterrupt.new()

  return self
end
