--- FunkyBunch
-- @author Created by Aaron Barrett on February 26, 2016.

class("FunkyBunch")

require("FunkyItem")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

FunkyBunch.Z_ITEM_UNSELECTED = 10
FunkyBunch.Z_ITEM_SELECTED = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function playCorrectMatch(self, firstItem, secondItem)
  Sound.playSoundEffect("correct")

  firstItem:playCorrect()
  secondItem:playCorrect()

  TableUtils.removeValueFromArray(self.items, firstItem)
  TableUtils.removeValueFromArray(self.items, secondItem)

  for _, item in ipairs(self.items) do
    item:playGray()
  end

  self.selectedItem = nil
end

local function ingoreInputUntilGameDismissesRationalePanel(self, firstItem, secondItem)
  firstItem:playWrong()
  secondItem:playWrong()

  local function resumeCallback()
    self:setTouchEnabled(true)
  end
  self.onWrongCallback(firstItem, secondItem, resumeCallback)
end

local function waitForWrongMatch(self, firstItem, secondItem)
  Sound.playSoundEffect("incorrect")

  firstItem:setSelected(false)
  secondItem:setSelected(false)

  self.selectedItem = nil

  ingoreInputUntilGameDismissesRationalePanel(self, firstItem, secondItem)
end

local function waitForCheckForMatch(self, firstItem, secondItem)
  self:setTouchEnabled(false)

  local firstItemId = firstItem:getConceptId()
  local secondItemId = secondItem:getConceptId()

  if (firstItem:matches(secondItem)) then
    self.onCorrectCallback()

    GameSession.contentTracker:notifyCorrectAnswer(firstItemId)
    GameSession.contentTracker:notifyCorrectAnswer(secondItemId)

    GameSession.incrementCorrectAnswerCount()

    playCorrectMatch(self, firstItem, secondItem)

    if (self:haveAllMatchesBeenFound()) then
      self:playDisappear()
    else
      if (not self.outroHasStarted) then
        self:setTouchEnabled(true)
      end
    end
  else
    GameSession.contentTracker:notifyIncorrectAnswer(firstItemId)
    GameSession.contentTracker:notifyIncorrectAnswer(secondItemId)

    GameSession.incrementIncorrectAnswerCount()

    waitForWrongMatch(self, firstItem, secondItem)
  end
end

local function onItemTapped(self, tappedItem)
  if (self.selectedItem == nil) then
    Sound.playSoundEffect("select")
    self.selectedItem = tappedItem
    tappedItem:setSelected(true)
  else

    if (self.selectedItem == tappedItem) then
      Sound.playSoundEffect("deselect")
      tappedItem:setSelected(false)
    else
      tappedItem:setSelected(true)
      waitForCheckForMatch(self, self.selectedItem, tappedItem)
    end

    self.selectedItem = nil

  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyBunch:setTouchEnabled(enabled)
  for _, item in ipairs(self.items) do
    item:setTouchEnabled(enabled)
  end
end

function FunkyBunch:playAppear()
  runInCoroutine(function()
    for _, item in ipairs(self.items) do
      GameSession.contentTracker:notifyPresentedConcept(item:getConceptId(), item:getDifficulty())

      item:playAppear()
      waitFor(0.15)
    end
  end)

  return runInCoroutine(function()
    Sound.playSoundEffect("bunch_appear")
    waitFor(1.5)
    self:setTouchEnabled(true)
  end)
end

function FunkyBunch:playDisappear()
  runInCoroutine(function()
    local duration = 1.5

    for _, item in ipairs(self.items) do
      item:playDisappear(duration)
    end

    waitFor(duration)
    self:removeSelf()
    self.dismissed = true
  end)
end

function FunkyBunch:isDismissed()
  return self.dismissed
end

function FunkyBunch:haveAllMatchesBeenFound()
  local remainingItemsCount = #(self.items)
  return (remainingItemsCount < 2)
end

function FunkyBunch:playOutro()
  self.outroHasStarted = true
  self:setTouchEnabled(false)
  local duration = 0.25
  for _, item in ipairs(self.items) do
    item:playDisappear(duration)
  end
end

function FunkyBunch:getMatchingPair()
  for _, firstItem in ipairs(self.items) do
    for _, secondItem in ipairs(self.items) do
      if firstItem ~= secondItem and firstItem:matches(secondItem) then
        return firstItem, secondItem
      end
    end
  end
end

function FunkyBunch:itemTapped(item)
  self.wasTapped = true
  onItemTapped(self, item)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FunkyBunch.new(width, height, itemLocations, permutations, onCorrectCallback, onWrongCallback)
  local self = Group.new(width, height)
  Mixin.mix(self, FunkyBunch)

  self.wasTapped = false

  local function onItemTappedCallback(item)
    self:itemTapped(item)
  end

  local items = {}
  for itemIndex, location in ipairs(itemLocations) do
    local orientationAngle = math.random(0, 3) * 90.0 + 45.0
    local permutation = permutations[itemIndex]

    local item = FunkyItem.new(permutation, orientationAngle, onItemTappedCallback)
    item:hide()
    item:setX(location.x)
    item:setY(location.y)
    self:insert(item)
    table.insert(items, item)
  end

  self.items = items
  self.selectedItem = nil
  self.dismissed = false
  self.onCorrectCallback = onCorrectCallback
  self.onWrongCallback = onWrongCallback
  self.outroHasStarted = false
  self.onItemTappedCallback = onItemTappedCallback

  return self
end
