--- FunkyExpressionFraction
-- @author Created by Aaron Barrett on February 23, 2016.

class("FunkyExpressionFraction")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BAR_HEIGHT = ReferenceDevice.getScaledWidth(2.0)
FunkyExpressionFraction.VERTICAL_BAR_OFFSET = ReferenceDevice.getScaledWidth(2.0)

local COLOR_WHITE = {1.0, 1.0, 1.0}
local COLOR_GRAYED_OUT = {0.27, 0.23, 0.25}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyExpressionFraction:jumpToUnstruckState()
  self.strikethrough:setScl(1.0, 0.0)
  self.topMultiplierLabel:setColor(unpack(COLOR_WHITE))
  self.topMultiplicationSignLabel:setColor(unpack(COLOR_WHITE))
  self.bottomMultiplierLabel:setColor(unpack(COLOR_WHITE))
  self.bottomMultiplicationSignLabel:setColor(unpack(COLOR_WHITE))
end

function FunkyExpressionFraction:createStrikeAction()
  local strikeAction = AHSpawn.new(
    AHScaleYTo.new(self.strikethrough, 0.5, AHEaseSineOut.new(), 1.0),
    AHColorTo.new(self.topMultiplierLabel, 0.5, nil, COLOR_GRAYED_OUT),
    AHColorTo.new(self.topMultiplicationSignLabel, 0.5, nil, COLOR_GRAYED_OUT),
    AHColorTo.new(self.bottomMultiplierLabel, 0.5, nil, COLOR_GRAYED_OUT),
    AHColorTo.new(self.bottomMultiplicationSignLabel, 0.5, nil, COLOR_GRAYED_OUT)
  )
  return strikeAction
end

function FunkyExpressionFraction:jumpToStruckState()
  self.topMultiplierLabel:setColor(unpack(COLOR_GRAYED_OUT))
  self.topMultiplicationSignLabel:setColor(unpack(COLOR_GRAYED_OUT))
  self.bottomMultiplierLabel:setColor(unpack(COLOR_GRAYED_OUT))
  self.bottomMultiplicationSignLabel:setColor(unpack(COLOR_GRAYED_OUT))
end

function FunkyExpressionFraction:getNumeratorY()
  return self.numeratorGroup:getY()
end

function FunkyExpressionFraction:getDenominatorY()
  return self.denominatorGroup:getY()
end

function FunkyExpressionFraction:getBarY()
  return self.bar:getY()
end

function FunkyExpressionFraction:createDuplicateBar()
  local bar = self.bar
  local duplicateBar = FilledRectangle.new(bar:getWidth(), bar:getHeight(), COLOR_WHITE)
  return duplicateBar
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildStrikethrough(self)
  local topMultiplierLabel = self.topMultiplierLabel
  local bottomMultiplierLabel = self.bottomMultiplierLabel

  assert((topMultiplierLabel ~= nil), "Fraction expression must store top and bottom multiplier groups in order to add strikethrough.")
  assert((bottomMultiplierLabel ~= nil), "Fraction expression must store top and bottom multiplier groups in order to add strikethrough.")

  local strikeXPadding = ReferenceDevice.getScaledWidth(4.0)

  local topStrikeX = self:getWidthPart(0.5) - self.numeratorGroup:getWidthPart(0.5) + topMultiplierLabel:getX() + topMultiplierLabel:getWidthPart(0.5) - strikeXPadding
  local bottomStrikeX = self:getWidthPart(0.5) - self.denominatorGroup:getWidthPart(0.5) + bottomMultiplierLabel:getX() - bottomMultiplierLabel:getWidthPart(0.5) + strikeXPadding

  local topY = self:getHeight()
  local bottomY = 0.0

  local strikeWidth = ReferenceDevice.getScaledWidth(4.0)
  local strikeLength = MathUtils.distance(topStrikeX, topY, bottomStrikeX, bottomY)
  local strikethrough = FilledRectangle.new(strikeWidth, strikeLength, COLOR_GRAYED_OUT)
  strikethrough:inheritOpacity(self)
  strikethrough:setLoc(0.5 * (topStrikeX + bottomStrikeX), self:getHeightPart(0.5))

  local deltaY = self:getHeight()
  local deltaX = topStrikeX - bottomStrikeX
  local rotation = math.deg(math.atan2(deltaY, deltaX) - math.pi * 0.5)
  strikethrough:setRot(rotation)

  self:insert(strikethrough)
  self.strikethrough = strikethrough
end

local function createMultiplicationSignLabel()
  return Sprite.newWithName("times.png")
end

local function createMultipliedRow(factor, multiplier, font)
  local spacing = ReferenceDevice.getScaledWidth(5.0)
  local row = Row.new(spacing)
  local factorLabel = Label.new(factor, font)
  factorLabel:inheritColor(row)
  row:insert(factorLabel)
  local multiplicationSignLabel = createMultiplicationSignLabel()
  multiplicationSignLabel:inheritColor(row)
  row:insert(multiplicationSignLabel)
  local multiplierLabel = Label.new(multiplier, font)
  multiplierLabel:inheritColor(row)
  row:insert(multiplierLabel)
  row:alignChildren()

  row.multiplierLabel = multiplierLabel
  row.multiplicationSignLabel = multiplicationSignLabel
  row.factorLabel = factorLabel

  return row
end

local function newInternal(numeratorGroup, denominatorGroup)
  local self = Column.new()
  Mixin.mix(self, FunkyExpressionFraction)

  numeratorGroup:inheritColor(self)
  self:insert(numeratorGroup)
  self.numeratorGroup = numeratorGroup
  denominatorGroup:inheritColor(self)
  self:insert(denominatorGroup)
  self.denominatorGroup = denominatorGroup
  self:alignChildren()

  local bar = FilledRectangle.new(self:getWidth(), BAR_HEIGHT, COLOR_WHITE)
  bar:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.5) - FunkyExpressionFraction.VERTICAL_BAR_OFFSET)
  bar:inheritColor(self)
  self:insert(bar)
  self.bar = bar

  return self
end

-- Constructor ------------------------

function FunkyExpressionFraction.new(numerator, denominator, fontSize)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local numeratorLabel = Label.new(numerator, font)
  local denominatorLabel = Label.new(denominator, font)
  return newInternal(numeratorLabel, denominatorLabel)
end

function FunkyExpressionFraction.newWithReducedFraction(fraction, fontSize)
  local numerator = fraction:getNumerator()
  local denominator = fraction:getDenominator()
  return FunkyExpressionFraction.new(numerator, denominator, fontSize)
end

function FunkyExpressionFraction.newMultiplied(numerator, denominator, multiplier, fontSize)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local numeratorGroup = createMultipliedRow(numerator, multiplier, font)
  local denominatorGroup = createMultipliedRow(denominator, multiplier, font)
  local self = newInternal(numeratorGroup, denominatorGroup)

  self.topMultiplierLabel = numeratorGroup.multiplierLabel
  self.topMultiplicationSignLabel = numeratorGroup.multiplicationSignLabel
  self.topFactorLabel = numeratorGroup.factorLabel
  self.topFactorLabelStartX = numeratorGroup.factorLabelStartX
  self.bottomMultiplierLabel = denominatorGroup.multiplierLabel
  self.bottomMultiplicationSignLabel = denominatorGroup.multiplicationSignLabel
  self.bottomFactorLabel = denominatorGroup.factorLabel
  self.bottomFactorLabelStartX = denominatorGroup.factorLabelStartX

  buildStrikethrough(self)

  return self
end
