--- FunkyExpressionMoney
-- @author Created by Aaron Barrett on February 24, 2016.

class("FunkyExpressionMoney")

require("FunkyMoneyFormatter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CURRENCY = Localizer.get("CURRENCY")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getWidthForLabels(labels)
  local width = 0.0

  for _, label in ipairs(labels) do
    width = width + label:getLeftMargin() + label:getWidth() + label:getRightMargin()
  end

  return width
end

local function getAlignedXLocationsForLabels(labels, startX)
  local xLocations = {}

  local penX = startX or 0.0

  for _, label in ipairs(labels) do
    local labelWidth = label:getWidth()
    local labelLeftMargin = label:getLeftMargin()

    local labelX = penX + labelLeftMargin + labelWidth * 0.5
    table.insert(xLocations, labelX)
    penX = penX + labelLeftMargin + labelWidth + label:getRightMargin()
  end

  return xLocations
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyExpressionMoney:createHideDecimalAction()
  local visibleLabels = self.decimalHiddenLabels
  local widthOfVisibleLabels = getWidthForLabels(visibleLabels)
  local startX = 0.5 * (self:getWidth() - widthOfVisibleLabels)
  local destinationXLocations = getAlignedXLocationsForLabels(visibleLabels, startX)

  local hideAction = AHSpawn.new()
  local ease = AHEaseSineInOut.new()

  for labelIndex, label in ipairs(visibleLabels) do
    local destinationX = destinationXLocations[labelIndex]
    local labelAction = AHLocationXTo.new(label, 0.5, ease, destinationX)
    hideAction:addAction(labelAction)
  end

  for _, label in ipairs(self.vanishingLabels) do
    local labelAction = AHOpacityTo.new(label, 0.5, nil, 0.0)
    hideAction:addAction(labelAction)
  end

  return hideAction
end

function FunkyExpressionMoney:jumpToDecimalShownState()
  local decimalShownLabels = self.decimalShownLabels
  local xLocations = getAlignedXLocationsForLabels(decimalShownLabels)
  for labelIndex, label in ipairs(decimalShownLabels) do
    local xLocation = xLocations[labelIndex]
    label:setOpacity(1.0)
    label:setX(xLocation)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FunkyExpressionMoney.new(baseAmount, multiplier, fontSize)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)

  -- TODO: Handle other currency formatting/syntax
  local amountText = FunkyMoneyFormatter.getTextFormattedAsDollarsAndCents(baseAmount, multiplier)
  local text = CURRENCY .. amountText

  local self = Label.new(text, font)
  Mixin.mix(self, FunkyExpressionMoney)

  return self
end

function FunkyExpressionMoney.newWithVanishingDecimal(baseAmount, multiplier, fontSize)
  local decimalHiddenLabels = {}
  local decimalShownLabels = {}
  local vanishingLabels = {}

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)

  local dollarSignLabel = Label.new(CURRENCY, font)
  table.insert(decimalShownLabels, dollarSignLabel)
  table.insert(vanishingLabels, dollarSignLabel)

  local dollarsText = FunkyMoneyFormatter.getTextOfDollarsPart(baseAmount, multiplier)
  local dollarsLabel = Label.new(dollarsText, font)
  table.insert(decimalShownLabels, dollarsLabel)
  if (dollarsText == "0") then
    table.insert(vanishingLabels, dollarsLabel)
  else
    table.insert(decimalHiddenLabels, dollarsLabel)
  end

  local decimalPointLabel = Label.new(".", font)
  table.insert(decimalShownLabels, decimalPointLabel)
  table.insert(vanishingLabels, decimalPointLabel)

  local centsText = FunkyMoneyFormatter.getTextOfCentsPart(baseAmount, multiplier)
  local centsLeadingZeroesText = FunkyMoneyFormatter.getLeadingZeroesText(centsText)
  if ((centsLeadingZeroesText ~= "") and (centsLeadingZeroesText ~= "00")) then
    local centsLeadingZeroesLabel = Label.new(centsLeadingZeroesText, font)
    table.insert(decimalShownLabels, centsLeadingZeroesLabel)
    if (dollarsText == "0") then
      table.insert(vanishingLabels, centsLeadingZeroesLabel)
    else
      table.insert(decimalHiddenLabels, centsLeadingZeroesLabel)
    end
  end

  local centsAfterZeroesText = FunkyMoneyFormatter.getTextWithLeadingZeroesRemoved(centsText)
  if (centsAfterZeroesText == "") then
    centsAfterZeroesText = "00"
  end
  local centsAfterZeroesLabel = Label.new(centsAfterZeroesText, font)
  table.insert(decimalShownLabels, centsAfterZeroesLabel)
  table.insert(decimalHiddenLabels, centsAfterZeroesLabel)

  local width = getWidthForLabels(decimalShownLabels)
  local height = centsAfterZeroesLabel:getHeight()

  local self = Group.new(width, height)
  Mixin.mix(self, FunkyExpressionMoney)

  for _, label in ipairs(decimalShownLabels) do
    label:setY(height - label:getHeightPart(0.5))
    label:inheritColor(self)
    self:insert(label)
  end

  self.decimalShownLabels = decimalShownLabels
  self.decimalHiddenLabels = decimalHiddenLabels
  self.vanishingLabels = vanishingLabels

  self:jumpToDecimalShownState()

  return self
end
