--- FunkyGame
-- @author Created by Aaron Barrett on February 01, 2016.

class("FunkyGame")

require("StandardUIBuilder")
require("FunkyConfiguration")

require("FunkyBackground")
require("FunkyBunch")
require("FunkyRoundPicker")
require("FunkyItemLayoutManager")
require("FunkyTipPanel")
require("FunkyIntroOverlay")
require("FunkyOutroFlashOverlay")
require("FunkyShaker")
require("FunkySkewedGradient")
require("tutorials.TutorialHighlighter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {1.0, 1.0, 1.0}
local COLOR_BLACK = {0.0, 0.0, 0.0}
local TAP_INSTRUCTION_LABEL_COLOR = {1, 1, 1}
local COLOR_OUTRO_GRADIENT = {0.17, 1.0, 0.81}
local COLOR_TUTORIAL_OVERLAY = {0.039, 0.055, 0.082, 0.6}
local COLOR_TUTORIAL_MASK = {1.0, 1.0, 1.0, 1.0}

local Z_SHAKER = 10
local Z_UPPER_GRADIENT = 20
local Z_INTRO_OVERLAY = 30
local Z_PROGRESS_METER = 40
local Z_NEEDED_MATCHES_BOX = 50
local Z_TUTORIAL = 55
local Z_TAP_INSTRUCTION_LABEL = 60
local Z_PANEL = 70
local Z_OUTRO_GRADIENT = 80
local Z_OUTRO_FLASH_OVERLAY = 90

local function configureGameDifficulty()
  local rankTwoScore = FunkyConfiguration.getRankTwoScore()
  local rankThreeScore = FunkyConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = FunkyConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.EXTRA_MATCH_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function playShowTapInstruction(self)
  local showAction = AHDelayed.new(1.6, AHOpacityTo.new(self.tapInstructionLabel, 0.5, nil, 1.0))
  showAction:start()
end

local function updateTapInstructionLabel(self)
  local tapInstructionLabel = self.tapInstructionLabel
  if (tapInstructionLabel ~= nil) then
    local hideAndRemoveAction = AHSequence.new(
      AHOpacityTo.new(tapInstructionLabel, 0.1, nil, 0.0),
      AHRemove.new(tapInstructionLabel)
    )

    self.tapInstructionLabel = nil
    hideAndRemoveAction:start()
  end
end

local function updateNeededMatchesBox(self)
  local neededMatchesBox = self.neededMatchesBox
  if (neededMatchesBox ~= nil) then
    self.neededMatchesBox = nil
    local duration = 1.0
    neededMatchesBox:playDisappear(duration)
  end
end

local function onCorrect(self)
  self.progressMeter:incrementMatchesCount()
  GameSession.bonusReporter:incrementBonusCount(BonusTypes.TYPES.EXTRA_MATCH_BONUS)

  updateNeededMatchesBox(self)
  updateTapInstructionLabel(self)

  local numberOverMinimum = self.progressMeter:getMatchesCount() - FunkyConfiguration.getMinimumMatchesNeeded()

  if (numberOverMinimum >= 1) then
    local decayedBonusPoints = FunkyConfiguration.getDecayingBonusPoints(numberOverMinimum)
    GameSession.increaseScore(decayedBonusPoints)

    local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.EXTRA_MATCH_BONUS)
    self.bonusTray:indicateSimpleTextBonus(notificationString, decayedBonusPoints)
  else
    GameSession.increaseScore(FunkyConfiguration.POINTS_PER_REGULAR_MATCH)
  end

  self.scoreLabel:setScore(GameSession.getScore())
  self.background:playCorrect()
end

local function subtractFromRemainingTime(self)
  self.timer:addTime(FunkyConfiguration.WRONG_ANSWER_PENALTY_IN_SECONDS)
end

local function waitForBackgroundWrong(self)
  self.background:playWrong()
  self.shaker:playWrong()
  waitFor(0.7)
end

local function waitForShowRationalePanel(self, firstItem, secondItem, resumeCallback)
  local panel = FunkyTipPanel.new(firstItem, secondItem)
  self:insertAtCenter(panel, Z_PANEL)

  panel:playAppear()

  waitUntil(function() return panel:isDismissed() end)

  firstItem:playRevert()
  secondItem:playRevert()

  resumeCallback()
  self.timer:resumeAndHideIcon()
end

local function onWrong(self, firstItem, secondItem, resumeCallback)
  runInCoroutine(function()
    subtractFromRemainingTime(self)
    self.timer:pauseAndShowIcon()
    waitForBackgroundWrong(self)
    waitForShowRationalePanel(self, firstItem, secondItem, resumeCallback)
  end)
end

local function buildNextBunch(self)

  local function onCorrectCallback()
    onCorrect(self)
  end

  local function onWrongCallback(firstItem, secondItem, resumeCallback)
    onWrong(self, firstItem, secondItem, resumeCallback)
  end

  local itemLocations = self.itemLayoutManager:getNextLocationsArray()
  local itemsCount = #(itemLocations)

  local permutations = self.roundPicker:getNextRoundPermutations(itemsCount)

  local bunchWidth = self.background:getWidth()
  local bunchHeight = self.background:getHeight()
  local bunchScale = self.background:getScl()

  local bunch = FunkyBunch.new(bunchWidth, bunchHeight, itemLocations, permutations, onCorrectCallback, onWrongCallback)
  bunch:setScl(bunchScale)
  bunch:setLoc(self.shaker:getWidthPart(0.5), self.shaker:getHeightPart(0.5))
  self.shaker:insert(bunch, FunkyShaker.Z_BUNCH)

  return bunch
end

local function waitForShowIntro(self)
  local introOverlay = self.introOverlay
  introOverlay:playAppear()

  waitUntil(function() return introOverlay:hasBeenTapped() end)
  self.background:playIntro()
  self.upperGradient:playAppear()
  playShowTapInstruction(self)
  waitFor(0.95)
end

local function buildOutroGradient(self)
  local outroGradient = FunkySkewedGradient.new()
  outroGradient:setColor(unpack(COLOR_OUTRO_GRADIENT))
  outroGradient:setOpacity(0.0)
  outroGradient:setScl(1.0, -2.0)
  outroGradient:setLoc(self:getWidthPart(0.5), outroGradient:getHeight())
  self:insert(outroGradient, Z_OUTRO_GRADIENT)
  self.outroGradient = outroGradient
end

local function buildOutroFlashOverlay(self)
  local outroFlashOverlay = FunkyOutroFlashOverlay.new()
  self:insertAtCenter(outroFlashOverlay, Z_OUTRO_FLASH_OVERLAY)
  self.outroFlashOverlay = outroFlashOverlay
end

local function playOutroGradient(self)
  local outroGradientAction = AHSequence.new(
    AHOpacityTo.new(self.outroGradient, 0.25, nil, 1.0),
    AHOpacityTo.new(self.outroGradient, 0.75, nil, 0.0)
  )
  return outroGradientAction:start()
end

local function waitForShowOutro(self)
  buildOutroGradient(self)
  buildOutroFlashOverlay(self)

  Sound.playSoundEffect("outro")

  local duration = 0.05
  self.progressMeter:playDisappear(duration)

  self.bunch:playOutro()
  self.background:playOutro()
  self.outroFlashOverlay:playOutro()
  playOutroGradient(self)
  waitFor(3.0)
end

local function playFadeInUi(self)
  local duration = 1.0
  self.uiOverlay:fadeIn(duration)
end

local function setMaskPos(self, mask, element, scale)
  scale = scale or 1
  local x, y = self:transformLoc(element)
  local anchorX, anchorY = mask:getAnchorPoint()
  x = x + element:getEffectiveWidthPart(anchorX * scale)
  y = y + element:getEffectiveHeightPart(anchorY * scale)
  mask:setLoc(x, y)
end

local function showHelpAndWaitForTap(self)
  self.bunch:setTouchEnabled(false)

  local firstItem, secondItem = self.bunch:getMatchingPair()
  local maskScale = self.background:getScl()

  local holeItem = firstItem

  local generateHoleCallback = function()
    local hole = Sprite.newWithName("diamond_filling.png")
    hole:setScl(maskScale)
    local holeWidth, holeHeight = hole:getDims()
    local sqrt2 = 1.414
    local holeGroup = Group.new(sqrt2 * holeWidth, sqrt2 * holeHeight)
    hole:setLoc(sqrt2 * holeWidth / 2, sqrt2 * holeHeight / 2)
    hole:setRot(45)
    holeGroup:insert(hole)
    setMaskPos(self, holeGroup, holeItem, maskScale)
    return holeGroup
  end

  local instructionLabel = self.progressMeter
  local instructionMask = FilledRectangle.new(instructionLabel:getWidth(), instructionLabel:getHeight(), COLOR_TUTORIAL_MASK)
  setMaskPos(self, instructionMask, instructionLabel)

  local secondaryMasks = {instructionMask}
  local helpLayer = TutorialHighlighter.newWithMasks({generateHoleCallback},
                                                     secondaryMasks)

  helpLayer:setOverlayColor(COLOR_TUTORIAL_OVERLAY)
  helpLayer:setPulseSize(20)
  self:insertAtCenter(helpLayer, Z_TUTORIAL)
  helpLayer:createFadeInAnimation():start()
  helpLayer:waitForTargetTap()
  self.bunch.onItemTappedCallback(firstItem)

  holeItem = secondItem
  helpLayer:reinitialize()
  helpLayer:waitForTargetTap()
  self.bunch.onItemTappedCallback(secondItem)

  wait(helpLayer:createFadeOutAnimation():start())
  helpLayer:removeSelf()

  self.bunch:setTouchEnabled(true)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyGame:isUIAutoFadeInEnabled()
  return false
end

function FunkyGame:gameFlow()
  waitForShowIntro(self)
  playFadeInUi(self)

  self.timer:start()
  self.timer:pause()

  local isFirstBunch = true
  while (true) do
    self.bunch = buildNextBunch(self)
    wait(self.bunch:playAppear())
    self.timer:resume()
    local showTutorial = true
    while showTutorial do
      waitUntil(function()
        local isTimerFinished = (self.timer:getTime() > FunkyConfiguration.TUTORIAL_DELAY)
        showTutorial = isTimerFinished and isFirstBunch and not self.bunch.wasTapped
        return showTutorial or self.bunch:haveAllMatchesBeenFound()
      end)

      if showTutorial then
        self.timer:pauseAndShowIcon()
        showHelpAndWaitForTap(self)
        self.timer:resumeAndHideIcon()
      end
    end
    self.timer:pause()
    waitUntil(function() return self.bunch:isDismissed() end)
    isFirstBunch = false
  end

end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildShaker(self)
  local shaker = FunkyShaker.new()
  self:insertAtCenter(shaker, Z_SHAKER)
  self.shaker = shaker
end

local function buildBackground(self)
  local backgroundWidth = self:getWidthPart(1.8)
  local backgroundHeight = self:getHeightPart(1.6)
  local background = FunkyBackground.new(backgroundWidth, backgroundHeight)
  background:setLoc(self.shaker:getWidthPart(0.5), self.shaker:getHeightPart(0.5))
  self.shaker:insert(background, FunkyShaker.Z_BACKGROUND)
  self.background = background
end

local function buildUpperGradient(self)
  local upperGradient = FunkySkewedGradient.new()
  upperGradient:setColor(unpack(COLOR_BLACK))
  upperGradient:setOpacity(0.0)
  upperGradient:setLoc(self:getWidthPart(0.5), self:getHeight() - upperGradient:getHeightPart(0.5))
  self:insert(upperGradient, Z_UPPER_GRADIENT)
  self.upperGradient = upperGradient
end

local function buildIntroOverlay(self)
  local function handoffProgressMeterCallback(progressMeter, neededMatchesBox)
    self:takeChild(progressMeter, Z_PROGRESS_METER)
    self:takeChild(neededMatchesBox, Z_NEEDED_MATCHES_BOX)
    self.progressMeter = progressMeter
    self.neededMatchesBox = neededMatchesBox
  end

  local neededMatchesCount = FunkyConfiguration.getMinimumMatchesNeeded()
  local introOverlay = FunkyIntroOverlay.new(neededMatchesCount, handoffProgressMeterCallback)
  self:insertAtCenter(introOverlay, Z_INTRO_OVERLAY)
  self.introOverlay = introOverlay

  self.progressMeter = nil
  self.neededMatchesBox = nil
end

local function buildUi(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  self.timer = builder:addDigitTimer(FunkyConfiguration.GAME_DURATION)

  self.timer:setTimeoutHandler(function()
    if (self.progressMeter:getMatchesCount() >= FunkyConfiguration.getMinimumMatchesNeeded()) then
      waitForShowOutro(self)
      GameSession.notifyGameWin()
    else
      GameSession.notifyGameLoss()
    end
  end)

  local tapInstructionLabel = Label.new(Localizer.get("TAP_INSTRUCTIONS"), Font.new(FontFaces.DIN_OT_MEDIUM, 14.0))
  tapInstructionLabel:setColor(unpack(TAP_INSTRUCTION_LABEL_COLOR))
  self:insert(tapInstructionLabel, Z_TAP_INSTRUCTION_LABEL)
  tapInstructionLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.25))
  tapInstructionLabel:setOpacity(0)
  self.tapInstructionLabel = tapInstructionLabel
end

-- Constructor ------------------------

function FunkyGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, FunkyGame)

  configureGameDifficulty()
  buildShaker(self)
  buildBackground(self)
  buildUpperGradient(self)
  buildIntroOverlay(self)
  buildUi(self)

  self.roundPicker = FunkyRoundPicker.new()
  self.itemLayoutManager = FunkyItemLayoutManager.new(self.background)
  self.bunch = nil

  GameSession.contentTracker:setSquashData(true)

  return self
end
