--- FunkyPresentationPercentage
-- @author Created by Aaron Barrett on February 24, 2016.

class("FunkyPresentationPercentage")

require("FunkyPresentationBase")
require("FunkyFraction")
require("FunkyPercentageFormatter")
require("FunkyExpressionPercentage")
require("FunkyExpressionFraction")
require("FunkyEquation")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TIP_TYPE_MEMORIZATION = "TIP_TYPE_MEMORIZATION"
local TIP_TYPE_MULTIPLY = "TIP_TYPE_MULTIPLY"
local TIP_TYPE_HALVE = "TIP_TYPE_HALVE"


local FRACTIONS_THAT_REQUIRE_MEMORIZATION = {
  FunkyFraction.new(1, 4),
  FunkyFraction.new(1, 2),
  FunkyFraction.new(3, 4),
  FunkyFraction.new(1, 3),
  FunkyFraction.new(1, 5),

  FunkyFraction.new(1, 10),
  FunkyFraction.new(3, 10),
  FunkyFraction.new(7, 10),
  FunkyFraction.new(9, 10)
}

local FRACTIONS_THAT_REQUIRE_MULTIPLICATION = {
  FunkyFraction.new(2, 3),

  FunkyFraction.new(2, 5),
  FunkyFraction.new(3, 5),
  FunkyFraction.new(4, 5),

  FunkyFraction.new(5, 6),

  FunkyFraction.new(3, 8),
  FunkyFraction.new(5, 8),
  FunkyFraction.new(7, 8)
}

local FRACTIONS_THAT_REQUIRE_HALVING = {
  FunkyFraction.new(1, 6),
  FunkyFraction.new(1, 8)
}

local FONT_SIZE = 26.0
local FONT = Font.new(FontFaces.DIN_OT_MEDIUM, FONT_SIZE)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getYLocationForEquation(self, equationIndex, pageIndex)
  local divisionHeightProportion = 1.0 / (pageIndex + 1.0)
  return self:getHeightPart(0.125) + self:getHeightPart(0.75) * (1.0 - divisionHeightProportion * equationIndex)
end

local function getOpacityForEquation(equationIndex, pageIndex)
  if (pageIndex == equationIndex) then
    return 1.0
  else
    return 0.25
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- FunkyPresentation Protocol ---------

function FunkyPresentationPercentage:getTitleText()
  return Localizer.get("CONVERTING_PERCENTAGES")
end

function FunkyPresentationPercentage:prepareToShowPage(pageIndex)
  local previousPageIndex = pageIndex - 1

  for equationIndex, equation in ipairs(self.equations) do
    local equationShouldStartVisible = (equationIndex < pageIndex)
    if (equationShouldStartVisible) then
      local opacity = getOpacityForEquation(equationIndex, previousPageIndex)
      equation:setOpacity(opacity)
    else
      equation:setOpacity(0.0)
    end

    local startY = getYLocationForEquation(self, equationIndex, previousPageIndex)
    equation:setY(startY)
  end
end

function FunkyPresentationPercentage:playShowPage(pageIndex)
  local showAction = AHSpawn.new()
  local ease = AHEaseSineInOut.new()

  for equationIndex, equation in ipairs(self.equations) do
    local destinationY = getYLocationForEquation(self, equationIndex, pageIndex)
    local locationAction = AHLocationYTo.new(equation, 0.5, ease, destinationY)
    showAction:addAction(locationAction)

    local equationShouldBecomeVisible = (equationIndex <= pageIndex)
    if (equationShouldBecomeVisible) then
      local destinationOpacity = getOpacityForEquation(equationIndex, pageIndex)
      local opacityAction = AHOpacityTo.new(equation, 0.5, nil, destinationOpacity)
      showAction:addAction(opacityAction)
    end
  end

  return showAction:start()
end

function FunkyPresentationPercentage:getBlurbTextForPageIndex(pageIndex)
  return self.blurbText
end

function FunkyPresentationPercentage:playDisappear()
  local disappearAction = AHSpawn.new()

  for _, equation in ipairs(self.equations) do
    local equationAction = AHOpacityTo.new(equation, 0.5, nil, 0.0)
    disappearAction:addAction(equationAction)
  end

  return disappearAction:start()
end

function FunkyPresentationPercentage:hide()
  for _, equation in ipairs(self.equations) do
    equation:setOpacity(0.0)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function doesArrayContainFraction(fractionsArray, fraction)
  for _, fractionFromArray in ipairs(fractionsArray) do
    if (fraction:equals(fractionFromArray)) then
      return true
    end
  end
  return false
end

local function getTipType(permutation)
  local fraction = permutation:getFraction()

  if (doesArrayContainFraction(FRACTIONS_THAT_REQUIRE_MEMORIZATION, fraction)) then
    return TIP_TYPE_MEMORIZATION
  elseif (doesArrayContainFraction(FRACTIONS_THAT_REQUIRE_MULTIPLICATION, fraction)) then
    return TIP_TYPE_MULTIPLY
  elseif (doesArrayContainFraction(FRACTIONS_THAT_REQUIRE_HALVING, fraction)) then
    return TIP_TYPE_HALVE
  else
    error("Unable to find percentage tip type for fraction: `" .. fraction:getReducedForm(fraction) .. "`")
  end
end

local function getPagesCountForTipType(tipType)
  if (tipType == TIP_TYPE_MEMORIZATION) then
    return 1
  elseif ((tipType == TIP_TYPE_MULTIPLY) or (tipType == TIP_TYPE_HALVE)) then
    return 3
  else
    error("No pages count specified for percentage tip type `" .. tostring(tipType) .. "`")
  end
end

local function createEqualsSignLabel()
  return Label.new("=", FONT)
end

local function createMultiplicationSignLabel()
  return Sprite.newWithName("times.png")
end

local function createPercentageToFractionEquation(self, numerator, denominator)
  local percentageExpression = FunkyExpressionPercentage.new(numerator, denominator, FONT_SIZE)
  local equalsSignLabel = createEqualsSignLabel()
  local fractionExpression = FunkyExpressionFraction.new(numerator, denominator, FONT_SIZE)
  local equation = FunkyEquation.new({percentageExpression, equalsSignLabel, fractionExpression})
  return equation
end

local function createPercentageToMultipliedFractionEquation(self, numerator, denominator)
  local percentageExpression = FunkyExpressionPercentage.new(numerator, denominator, FONT_SIZE)
  local equalsSignLabel = createEqualsSignLabel()
  local factorLabel = Label.new(numerator, FONT)
  local multiplicationSignLabel = createMultiplicationSignLabel()
  local factoredFractionExpression = FunkyExpressionFraction.new(1, denominator, FONT_SIZE)
  local equation = FunkyEquation.new({percentageExpression, equalsSignLabel, factorLabel, multiplicationSignLabel, factoredFractionExpression})
  return equation
end

local function createPercentageToHalvedFractionEquation(self, numerator, denominator)
  local percentageExpression = FunkyExpressionPercentage.new(numerator, denominator, FONT_SIZE)
  local equalsSignLabel = createEqualsSignLabel()
  local oneHalfFractionExpression = FunkyExpressionFraction.new(1, 2, FONT_SIZE)
  local multiplicationSignLabel = createMultiplicationSignLabel()
  local doubleDenominator = denominator / 2
  local halvedFractionExpression = FunkyExpressionFraction.new(1, doubleDenominator, FONT_SIZE)
  local equation = FunkyEquation.new({percentageExpression, equalsSignLabel, oneHalfFractionExpression, multiplicationSignLabel, halvedFractionExpression})
  return equation
end

local function buildEquations(self, permutation, tipType)
  local fraction = permutation:getFraction()
  local numerator = fraction:getNumerator()
  local denominator = fraction:getDenominator()

  local equations = {}

  if (tipType == TIP_TYPE_MEMORIZATION) then

    local equation = createPercentageToFractionEquation(self, numerator, denominator)
    table.insert(equations, equation)

  elseif (tipType == TIP_TYPE_MULTIPLY) then

    local firstEquation = createPercentageToFractionEquation(self, 1, denominator)
    table.insert(equations, firstEquation)

    local secondEquation = createPercentageToMultipliedFractionEquation(self, numerator, denominator)
    table.insert(equations, secondEquation)

    local thirdEquation = createPercentageToFractionEquation(self, numerator, denominator)
    table.insert(equations, thirdEquation)

  elseif (tipType == TIP_TYPE_HALVE) then

    local doubleDenominator = denominator / 2
    local firstEquation = createPercentageToFractionEquation(self, 1, doubleDenominator)
    table.insert(equations, firstEquation)

    local secondEquation = createPercentageToHalvedFractionEquation(self, numerator, denominator)
    table.insert(equations, secondEquation)

    local thirdEquation = createPercentageToFractionEquation(self, numerator, denominator)
    table.insert(equations, thirdEquation)

  else

    error("No way to build equations specified for percentage tip type `" .. tostring(tipType) .. "`")

  end

  for _, equation in rpairs(equations) do
    self:insertAtCenter(equation)
  end

  self.equations = equations
end

local function buildBlurbText(self, permutation, tipType)
  local fraction = permutation:getFraction()
  local numerator = fraction:getNumerator()
  local denominator = fraction:getDenominator()

  local percentageText = FunkyPercentageFormatter.getTextFormattedAsPercentage(numerator, denominator)

  local numeratorText = tostring(numerator)
  local denominatorText = tostring(denominator)

  local blurbText = nil

  if (tipType == TIP_TYPE_MEMORIZATION) then
    local blurbTextTemplate = Localizer.get("REMEMBER_THAT_TEMPLATE_1")
    blurbText = string.format(blurbTextTemplate, percentageText, numeratorText, denominatorText)
  elseif (tipType == TIP_TYPE_MULTIPLY) then
    local factorPercentageText = FunkyPercentageFormatter.getTextFormattedAsPercentage(1, denominator)
    local blurbTextTemplate = Localizer.get("REMEMBER_THAT_TEMPLATE_2")
    blurbText = string.format(blurbTextTemplate, factorPercentageText, denominatorText,
                              percentageText, numeratorText, factorPercentageText,
                              percentageText, numeratorText, denominatorText)
  elseif (tipType == TIP_TYPE_HALVE) then
    local doubleDenominator = denominator / 2
    local doublePercentageText = FunkyPercentageFormatter.getTextFormattedAsPercentage(1, doubleDenominator)
    local blurbTextTemplate = Localizer.get("REMEMBER_THAT_TEMPLATE_3")
    blurbText = string.format(blurbTextTemplate, doublePercentageText, doubleDenominator,
                              percentageText, doublePercentageText, percentageText,
                              numeratorText, denominatorText)
  else
    error("No way to build blurb text specified for percentage tip type `" .. tostring(tipType) .. "`")
  end

  self.blurbText = blurbText
end

-- Constructor ------------------------

function FunkyPresentationPercentage.new(permutation)
  local tipType = getTipType(permutation)

  local pagesCount = getPagesCountForTipType(tipType)

  local self = FunkyPresentationBase.new(permutation, pagesCount)
  Mixin.mix(self, FunkyPresentationPercentage)

  buildEquations(self, permutation, tipType)
  buildBlurbText(self, permutation, tipType)

  return self
end
