--- FunkyRoundPicker
-- @author Created by Aaron Barrett on February 05, 2016.

class("FunkyRoundPicker")

require("FunkyFraction")
require("FunkyFractionPermutation")
require("FunkyFractionFormat")
require("FunkyConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FORMATS_BY_CONCEPT_FORMAT_NAME = {
  ["fraction"] = FunkyFractionFormat.FRACTION,
  ["image"] = FunkyFractionFormat.IMAGE,
  ["money"] = FunkyFractionFormat.MONEY,
  ["percentage"] = FunkyFractionFormat.PERCENTAGE
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getNextPermutationIndexToMatch(self, firstPermutation)
  for permutationIndex, permutation in ipairs(self.permutationsQueue) do
    if (firstPermutation:canMatch(permutation)) then
      return permutationIndex
    end
  end
  return nil
end

local function recycleSeenPermutations(self)
  self.permutationsQueue = TableUtils.createShuffledArray(self.seenPermutations)
  self.seenPermutations = {}
end

local function dequeuePermutationAtIndex(self, index)
  local permutation = table.remove(self.permutationsQueue, index)
  table.insert(self.seenPermutations, permutation)

  if (#(self.permutationsQueue) == 0) then
    recycleSeenPermutations(self)
  end

  return permutation
end

local function getNextPermutationIndexWithoutFractions(self, excludedFractions)
  for permutationIndex, permutation in ipairs(self.permutationsQueue) do
    if (not permutation:hasFractionFrom(excludedFractions)) then
      return permutationIndex
    end
  end
  return nil
end

local function getComplementaryFractions(self, fractionsList)
  local complementaryFractions = {}

  for _, fraction in ipairs(fractionsList) do
    local numerator = fraction:getNumerator()
    local denominator = fraction:getDenominator()
    local complementaryNumerator = denominator - numerator

    if (numerator ~= complementaryNumerator) then
      local complementaryFraction = FunkyFraction.new(complementaryNumerator, denominator)
      table.insert(complementaryFractions, complementaryFraction)
    end
  end

  return complementaryFractions
end

local function getNextPermutationsPair(self, usedFractions)
  local complementaryFractions = getComplementaryFractions(self, usedFractions)
  local excludedFractions = TableUtils.createConcatenatedArray(usedFractions, complementaryFractions)

  local firstIndex = getNextPermutationIndexWithoutFractions(self, excludedFractions)
  local firstPermutation = nil
  if (firstIndex ~= nil) then
    firstPermutation = dequeuePermutationAtIndex(self, firstIndex)
  else
    local permutationToChange = dequeuePermutationAtIndex(self, 1)
    firstPermutation = permutationToChange:createPermutationWithChangedFraction(usedFractions)
  end

  local secondIndex = getNextPermutationIndexToMatch(self, firstPermutation)
  local secondPermutation = nil
  if (secondIndex ~= nil) then
    secondPermutation = dequeuePermutationAtIndex(self, secondIndex)
  else
    local permutationToChange = dequeuePermutationAtIndex(self, 1)
    secondPermutation = permutationToChange:createForcedMatch(firstPermutation)
  end

  return {firstPermutation, secondPermutation}
end

local function createPermutationFromConcept(concept)
  local numerator = concept["numerator"]
  local denominator = concept["denominator"]

  local fraction = FunkyFraction.new(numerator, denominator)

  local formatName = concept["format"]
  local format = FORMATS_BY_CONCEPT_FORMAT_NAME[formatName]

  assert(format ~= nil, "No format provided for formatName `" .. tostring(formatName) .. "` in `" .. tostring(concept["identifier"])  .."`")

  local multiplier = concept["multiplier"]
  if (multiplier ~= nil) then
    multiplier = tonumber(multiplier)
    multiplier = MathUtils.round(multiplier)
  end

  local difficulty = concept["difficulty"]
  local conceptId = concept["identifier"]

  local permutation = FunkyFractionPermutation.newFromConcept(fraction, format, multiplier, difficulty, conceptId)
  return permutation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyRoundPicker:getNextRoundPermutations(itemsCountInRound)
  local pairsCountInRound = math.ceil(itemsCountInRound / 2)
  local usedFractions = {}
  local permutations = {}
  for pairIndex = 1, pairsCountInRound do
    local permutationsPair = getNextPermutationsPair(self, usedFractions)
    TableUtils.concatenateArrays(permutations, permutationsPair)

    local firstPermutation = permutationsPair[1]
    local usedFraction = firstPermutation:getFraction()
    table.insert(usedFractions, usedFraction)
  end

  TableUtils.shuffleArray(permutations)

  return permutations
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function FunkyRoundPicker.new()
  local self = {}
  Mixin.mix(self, FunkyRoundPicker)

  local conceptGenerator = GameSession.conceptGenerator
  local concepts = conceptGenerator:nextAsArray(FunkyConfiguration.CONCEPTS_COUNT_PER_GAME)

  local permutationsQueue = {}
  for _, concept in ipairs(concepts) do
    local permutation = createPermutationFromConcept(concept)
    table.insert(permutationsQueue, permutation)
  end

  self.permutationsQueue = permutationsQueue
  self.seenPermutations = {}

  return self
end
