--- FunkyTipAdvancedOverlay
-- @author Created by Aaron Barrett on March 02, 2016.

class("FunkyTipAdvancedOverlay")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_TITLE_LABEL = 10
local Z_ARROW_BUTTON = 20
local Z_PRESENTATION = 30
local Z_BLURB_LABEL = 40

local PRESENTATION_Y = Screen.getSafeY(0.55)

local PRESENTATION_CLASS_BY_FORMAT = {
  [FunkyFractionFormat.FRACTION] = FunkyPresentationFraction,
  [FunkyFractionFormat.IMAGE] = FunkyPresentationImage,
  [FunkyFractionFormat.MONEY] = FunkyPresentationFraction,
  [FunkyFractionFormat.PERCENTAGE] = FunkyPresentationPercentage
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setTouchEnabled(self, enabled)
  self.leftArrowButton:setTouchEnabled(enabled)
  self.rightArrowButton:setTouchEnabled(enabled)
  self.returnToGameButton:setTouchEnabled(enabled)
end

local function playButtonPress(button)
  local pressAction = AHSequence.new(
    AHScaleTo.new(button, 0.15, AHEaseSineIn.new(), 1.2),
    AHScaleTo.new(button, 0.15, AHEaseSineOut.new(), 1.0)
  )
  return pressAction:start()
end

local function getCurrentPresentation(self)
  local overallPageIndex = self.overallPageIndex
  return self.presentationsByOverallPageIndex[overallPageIndex]
end

local function getCurrentPresentationPageIndex(self)
  local overallPageIndex = self.overallPageIndex
  return self.presentationPageIndicesByOverallPageIndex[overallPageIndex]
end

local function doesPageExistAtOverallIndex(self, overallPageIndex)
  return (self.presentationsByOverallPageIndex[overallPageIndex] ~= nil)
end

local function playFadeBlurbLabelToText(self, blurbText, fadeDuration)
  local fadeAction = AHSequence.new(
    AHOpacityTo.new(self.blurbLabel, fadeDuration * 0.5, nil, 0.0),
    function()
      self.blurbLabel:setText(blurbText)
    end,
    AHOpacityTo.new(self.blurbLabel, fadeDuration * 0.5, nil, 1.0)
  )
  return fadeAction:start()
end

local function playHighlightItemForPresentation(self, presentation)
  local highlightItemCallback = presentation:getHighlightItemCallback()
  local titleText = presentation:getTitleText()

  highlightItemCallback()

  local titleLabel = self.titleLabel
  local fadeTitleAction = AHSequence.new(
    AHOpacityTo.new(titleLabel, 0.5, nil, 0.0),
    function()
      titleLabel:setText(titleText)
    end,
    AHOpacityTo.new(titleLabel, 0.5, nil, 1.0)
  )
  return fadeTitleAction:start()
end

local function getOpacityForLeftArrowAtOverallPageIndex(overallPageIndex)
  if (overallPageIndex > 1) then
    return 1.0
  else
    return 0.25
  end
end

local function getOpacityForRightArrowAtOverallPageIndex(self, overallPageIndex)
  local overallPagesCount = #(self.presentationsByOverallPageIndex)
  if (overallPageIndex < overallPagesCount) then
    return 1.0
  else
    return 0.25
  end
end

local function playFadeArrowButtons(self, overallPageIndex)
  local destinationLeftOpacity = getOpacityForLeftArrowAtOverallPageIndex(overallPageIndex)
  local destinationRightOpacity = getOpacityForRightArrowAtOverallPageIndex(self, overallPageIndex)

  local fadeAction = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self.leftArrowButtonIcon, 0.5, nil, 0.0),
      AHOpacityTo.new(self.rightArrowButtonIcon, 0.5, nil, 0.0)
    ),
    AHSpawn.new(
      AHOpacityTo.new(self.leftArrowButtonIcon, 0.5, nil, destinationLeftOpacity),
      AHOpacityTo.new(self.rightArrowButtonIcon, 0.5, nil, destinationRightOpacity)
    )
  )
  return fadeAction:start()
end

local function playUpdateArrowButtons(self, overallPageIndex)
  local destinationLeftOpacity = getOpacityForLeftArrowAtOverallPageIndex(overallPageIndex)
  local destinationRightOpacity = getOpacityForRightArrowAtOverallPageIndex(self, overallPageIndex)

  local fadeAction = AHSpawn.new(
    AHOpacityTo.new(self.leftArrowButtonIcon, 0.25, nil, destinationLeftOpacity),
    AHOpacityTo.new(self.rightArrowButtonIcon, 0.25, nil, destinationRightOpacity)
  )
  return fadeAction:start()
end

local function showPageAtOverallIndex(self, overallPageIndex)
  setTouchEnabled(self, false)

  local originalPresentation = getCurrentPresentation(self)

  self.overallPageIndex = overallPageIndex

  runInCoroutine(function()
    local presentation = getCurrentPresentation(self)
    local presentationPageIndex = getCurrentPresentationPageIndex(self)

    local blurbText = presentation:getBlurbTextForPageIndex(presentationPageIndex)

    local shouldTransitionBetweenPresentations = (presentation ~= originalPresentation)
    if (shouldTransitionBetweenPresentations) then
      local fadeDuration = 1.0
      playFadeBlurbLabelToText(self, blurbText, fadeDuration)
      playFadeArrowButtons(self, overallPageIndex)
      playHighlightItemForPresentation(self, presentation)

      wait(originalPresentation:playDisappear())
    else

      playUpdateArrowButtons(self, overallPageIndex)

      if (blurbText ~= self.blurbLabel:getText()) then
        local fadeDuration = 0.5
        playFadeBlurbLabelToText(self, blurbText, fadeDuration)
      end
    end

    presentation:prepareToShowPage(presentationPageIndex)
    local framesCountForPerformance = 1
    waitForFrames(framesCountForPerformance)

    wait(presentation:playShowPage(presentationPageIndex))
    setTouchEnabled(self, true)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyTipAdvancedOverlay:playAppear()
  showPageAtOverallIndex(self, 1)

  local titleLabel = self.titleLabel

  local appearAction = AHSpawn.new(
    AHOpacityTo.new(self.leftArrowButtonIcon, 0.5, nil, 0.25),
    AHOpacityTo.new(self.rightArrowButtonIcon, 0.5, nil, 1.0),
    AHOpacityTo.new(self.returnToGameLabel, 0.5, nil, 1.0),
    AHOpacityTo.new(self.blurbLabel, 0.5, nil, 1.0),
    AHOpacityTo.new(titleLabel, 0.5, nil, 1.0),
    AHLocationYTo.new(titleLabel, 0.5, AHEaseSineInOut.new(), Screen.getSafeY(0.9))
  )
  self.actionInterrupt:insert(appearAction)
end

function FunkyTipAdvancedOverlay:playDisappear()
  local presentation = getCurrentPresentation(self)
  presentation:playDisappear()

  local disappearAction = AHSpawn.new(
    AHOpacityTo.new(self.leftArrowButtonIcon, 0.5, nil, 0.0),
    AHOpacityTo.new(self.rightArrowButtonIcon, 0.5, nil, 0.0),
    AHOpacityTo.new(self.returnToGameLabel, 0.5, nil, 0.0),
    AHOpacityTo.new(self.titleLabel, 0.5, nil, 0.0),
    AHOpacityTo.new(self.blurbLabel, 0.5, nil, 0.0)
  )
  self.actionInterrupt:insert(disappearAction)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildTitleLabel(self)
  local titleFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 14.0)
  local firstPresentation = self.presentations[1]
  local titleText = firstPresentation:getTitleText()
  local titleLabel = Label.new(titleText, titleFont)
  titleLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.74))
  titleLabel:setOpacity(0.0)
  self:insert(titleLabel, Z_TITLE_LABEL)
  self.titleLabel = titleLabel
end

local function createPresentationForPermutation(self, permutation)
  local format = permutation:getFormat()
  local presentationClass = PRESENTATION_CLASS_BY_FORMAT[format]
  assert(presentationClass, "No presentation class provided for format `" .. tostring(format) .. "`")
  local presentation = presentationClass.new(permutation)
  presentation:hide()
  presentation:setLoc(Screen.getSafeX(0.5), PRESENTATION_Y)
  self:insert(presentation, Z_PRESENTATION)

  local pagesCount = presentation:getPagesCount()
  for presentationPageIndex = 1, pagesCount do
    table.insert(self.presentationsByOverallPageIndex, presentation)
    table.insert(self.presentationPageIndicesByOverallPageIndex, presentationPageIndex)
  end

  return presentation
end

local function buildPresentations(self, highlightLeftItemCallback, highlightRightItemCallback)
  self.presentationsByOverallPageIndex = {}
  self.presentationPageIndicesByOverallPageIndex = {}
  self.overallPageIndex = 1

  local presentations = {}
  for _, permutation in ipairs(self.permutations) do
    local presentation = createPresentationForPermutation(self, permutation)
    table.insert(presentations, presentation)
  end

  local leftPresentation = presentations[1]
  local rightPresentation = presentations[2]
  leftPresentation:setHighlightItemCallback(highlightLeftItemCallback)
  rightPresentation:setHighlightItemCallback(highlightRightItemCallback)

  self.presentations = presentations
end

local function buildBlurbLabel(self)
  local initialPresentation = getCurrentPresentation(self)
  local initialBlurbText = initialPresentation:getBlurbTextForPageIndex(1)

  local width = Screen.getSafeWidthPart(0.8)
  local font = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 19.0)
  local horizontalAlignment = Alignment.CENTER
  local verticalAlignment = nil
  local lineSpacing = ReferenceDevice.getScaledHeight(2.0)
  local blurbLabel = DynamicTextArea.newSimple(initialBlurbText, width, font, horizontalAlignment, verticalAlignment, lineSpacing)
  blurbLabel:setOpacity(0.0)
  blurbLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.25))
  self:insert(blurbLabel, Z_BLURB_LABEL)
  self.blurbLabel = blurbLabel
end

local function createArrowButton(self, xScale, xLocation, pageIndexIncrement)
  local arrowButtonIcon = Sprite.newWithName("tip_arrow.png")
  arrowButtonIcon:setOpacity(0.0)
  arrowButtonIcon:setScl(xScale, 1.0)

  local arrowButton = Group.new(Screen.getSafeWidthPart(0.35), Screen.getSafeHeightPart(0.25))
  arrowButton:insertAtCenter(arrowButtonIcon)
  arrowButton:setLoc(xLocation, PRESENTATION_Y)
  self:insert(arrowButton, Z_ARROW_BUTTON)

  arrowButton.icon = arrowButtonIcon

  local function onArrowButtonPressed()
    local updatedOverallPageIndex = self.overallPageIndex + pageIndexIncrement
    if (doesPageExistAtOverallIndex(self, updatedOverallPageIndex)) then
      Sound.playSoundEffect("click")
      playButtonPress(arrowButton)
      showPageAtOverallIndex(self, updatedOverallPageIndex)
    end
  end
  Button.makeButton(arrowButton, onArrowButtonPressed)
  arrowButton:setTouchEnabled(false)

  return arrowButton
end

local function buildArrowButtons(self)
  local leftArrowButton = createArrowButton(self, -1.0, Screen.getSafeX(0.08), -1)
  self.leftArrowButton = leftArrowButton
  self.leftArrowButtonIcon = leftArrowButton.icon

  local rightArrowButton = createArrowButton(self, 1.0, Screen.getSafeX(0.92), 1)
  self.rightArrowButton = rightArrowButton
  self.rightArrowButtonIcon = rightArrowButton.icon
end

local function buildReturnToGameButton(self, onPanelDismissedCallback)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 15.0)
  local label = Label.new(Localizer.get("Misc.RETURN_TO_GAME"), font)
  label:setOpacity(0.0)
  self.returnToGameLabel = label

  local returnToGameButton = Group.new(Screen.getSafeWidthPart(0.75), Screen.getSafeHeightPart(0.2))
  returnToGameButton:insertAtCenter(label)

  local minimumY = Screen.safeBottom + label:getHeightPart(0.5)
  local returnToGameButtonY = math.max(minimumY, label:getHeightPart(1.5))
  returnToGameButton:setLoc(Screen.getSafeX(0.5), returnToGameButtonY)
  self:insert(returnToGameButton)

  local function onReturnToGameButtonPressedCallback()
    Sound.playSoundEffect("click")
    setTouchEnabled(self, false)
    onPanelDismissedCallback()
    self:playDisappear()
    playButtonPress(returnToGameButton)
  end

  Button.makeButton(returnToGameButton, onReturnToGameButtonPressedCallback)

  self.returnToGameButton = returnToGameButton
end

-- Constructor ------------------------

function FunkyTipAdvancedOverlay.new(leftPermutation, rightPermutation, onPanelDismissedCallback, highlightLeftItemCallback, highlightRightItemCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, FunkyTipAdvancedOverlay)

  self.permutations = {leftPermutation, rightPermutation}

  buildPresentations(self, highlightLeftItemCallback, highlightRightItemCallback)
  buildTitleLabel(self)
  buildBlurbLabel(self)
  buildArrowButtons(self)
  buildReturnToGameButton(self, onPanelDismissedCallback)

  self.highlightLeftItemCallback = highlightLeftItemCallback
  self.highlightRightItemCallback = highlightRightItemCallback
  self.actionInterrupt = ActionInterrupt.new()

  setTouchEnabled(self, false)

  return self
end
