--- FunkyTipBasicOverlay
-- @author Created by Aaron Barrett on March 02, 2016.

class("FunkyTipBasicOverlay")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_BIG_X = {0.98, 0.0, 0.52}
local COLOR_CAPTION_LABEL = {1.0, 0.42, 0.55}

local Z_BIG_X = 10
local Z_NOT_EQUIVALENT_LABEL = 20
local Z_ITEM_CAPTION_LABEL = 30
local Z_LEARN_MORE_BUTTON = 40

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setTouchEnabled(self, enabled)
  self.returnToGameButton:setTouchEnabled(enabled)
  self.learnMoreButton:setTouchEnabled(enabled)
end

local function playButtonPress(button)
  local pressAction = AHSequence.new(
    AHScaleTo.new(button, 0.15, AHEaseSineIn.new(), 1.2),
    AHScaleTo.new(button, 0.15, AHEaseSineOut.new(), 1.0)
  )

  return pressAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function FunkyTipBasicOverlay:playAppear()
  local appearAction = AHSequence.new(
    AHOpacityTo.new(self, 0.5, nil, 1.0),
    function()
      setTouchEnabled(self, true)
    end
  )
  self.actionInterrupt:insert(appearAction)
end

function FunkyTipBasicOverlay:playDisappear()
  local disappearAction = AHSequence.new(
    AHOpacityTo.new(self, 0.5, nil, 0.0),
    AHRemove.new(self)
  )
  self.actionInterrupt:insert(disappearAction)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildBigX(self)
  local bigX = Sprite.newWithName("big_x.png")
  bigX:inheritOpacity(self)
  bigX:setColor(unpack(COLOR_BIG_X))
  bigX:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.85))
  self:insert(bigX, Z_BIG_X)
end

local function buildNotEquivalentLabel(self)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 14.0)
  local label = Label.new(Localizer.get("NOT_EQUIVALENT"), font)
  label:inheritOpacity(self)
  label:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.74))
  self:insert(label, Z_NOT_EQUIVALENT_LABEL)
end

local function createItemCaptionLabel(self, permutation)
  local fraction = permutation:getFraction()
  local text = fraction:getReducedForm()
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 20.0)
  local label = Label.new(text, font)
  label:setColor(unpack(COLOR_CAPTION_LABEL))
  label:inheritOpacity(self)
  return label
end

local function buildItemCaptionLabels(self, leftPermutation, rightPermutation, captionOffset)
  local leftCaptionLabel = createItemCaptionLabel(self, leftPermutation)

  local centerX = Screen.getSafeX(0.5)
  local captionY = Screen.getSafeY(0.55) - captionOffset - ReferenceDevice.getScaledHeight(15.0)

  leftCaptionLabel:setLoc(centerX - captionOffset, captionY)
  self:insert(leftCaptionLabel, Z_ITEM_CAPTION_LABEL)

  local rightCaptionLabel = createItemCaptionLabel(self, rightPermutation)
  rightCaptionLabel:setLoc(centerX + captionOffset, captionY)
  self:insert(rightCaptionLabel, Z_ITEM_CAPTION_LABEL)
end

local function buildLearnMoreButton(self, onMovedToAdvancedOverlayCallback)
  local spacing = ReferenceDevice.getScaledWidth(8.0)
  local row = Row.new(spacing)

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 13.0)
  local learnMoreLabel = Label.new(Localizer.get("LEARN_MORE"), font)
  learnMoreLabel:inheritOpacity(self)
  row:insert(learnMoreLabel)

  local arrow = Sprite.newWithName("learn_more_arrow.png")
  arrow:inheritOpacity(self)
  row:insert(arrow)

  row:alignChildren()

  local learnMoreButton = Group.new(Screen.getSafeX(0.6), Screen.getSafeY(0.2))
  learnMoreButton:insertAtCenter(row)
  learnMoreButton:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.3))
  self:insert(learnMoreButton, Z_LEARN_MORE_BUTTON)
  self.learnMoreButton = learnMoreButton

  local function onLearnMoreButtonPressedCallback()
    Sound.playSoundEffect("click")
    setTouchEnabled(self, false)
    playButtonPress(learnMoreButton)
    self:playDisappear()
    runInCoroutine(function()
      waitFor(0.5)
      onMovedToAdvancedOverlayCallback()
    end)
  end

  Button.makeButton(learnMoreButton, onLearnMoreButtonPressedCallback)
end

local function buildReturnToGameButton(self, onPanelDismissedCallback)
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 15.0)
  local label = Label.new(Localizer.get("Misc.RETURN_TO_GAME"), font)
  label:inheritOpacity(self)

  local returnToGameButton = Group.new(Screen.getSafeWidthPart(0.75), Screen.getSafeHeightPart(0.2))
  returnToGameButton:insertAtCenter(label)

  local minimumY = Screen.safeBottom + label:getHeightPart(0.5)
  local returnToGameButtonY = math.max(minimumY, label:getHeightPart(1.5))
  returnToGameButton:setLoc(Screen.getSafeX(0.5), returnToGameButtonY)
  self:insert(returnToGameButton)

  local function onReturnToGameButtonPressedCallback()
    Sound.playSoundEffect("click")
    setTouchEnabled(self, false)
    onPanelDismissedCallback()
    self:playDisappear()
    playButtonPress(returnToGameButton)
  end

  Button.makeButton(returnToGameButton, onReturnToGameButtonPressedCallback)

  self.returnToGameButton = returnToGameButton
end

-- Constructor ------------------------

function FunkyTipBasicOverlay.new(leftPermutation, rightPermutation, captionOffset, onPanelDismissedCallback, onMovedToAdvancedOverlayCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, FunkyTipBasicOverlay)

  buildBigX(self)
  buildNotEquivalentLabel(self)
  buildItemCaptionLabels(self, leftPermutation, rightPermutation, captionOffset)
  buildLearnMoreButton(self, onMovedToAdvancedOverlayCallback)
  buildReturnToGameButton(self, onPanelDismissedCallback)

  self:setOpacity(0.0)
  setTouchEnabled(self, false)

  self.actionInterrupt = ActionInterrupt.new()

  return self
end
