--- GameTipsLongValidator
-- @author Created by Jay Austria on August 28, 2023.
-- ConceptValidator for Weekly Report long game tips.

class("GameTipsLongValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "intro_text",
  "header",
  "body_text"
}

local INTRO_TEXT_SINGLE_TAGS = nil
local INTRO_TEXT_PAIRED_TAGS = nil
local INTRO_TEXT_TERTIARY_TAGS = nil
local INTRO_TEXT_OPTIONAL_PAIRED_TAGS = { "b" } -- use optional tags validator so we can catch multiple instances

local BODY_TEXT_SINGLE_TAGS = nil
local BODY_TEXT_PAIRED_TAGS = nil
local BODY_TEXT_TERTIARY_TAGS = nil
local BODY_TEXT_OPTIONAL_PAIRED_TAGS = { "b" } -- use optional tags validator so we can catch multiple instances

-- Keep these synced with the identifiers available in games.json/subject.json, the
-- the available tip_game_[identifier]_icon assets integrated into griffin/pegasus,
-- and the prefixes assumed in GameTipWeeklyReportItemProducer.cpp
local CONCEPT_PREFIX_GAME_TIP_LONG = "game_tip_long_"
local VALID_SKILL_IDENTIFIERS = {
  "association",
  "brevity",
  "clarity",
  "commas",
  "connotation",
  "contextualization",
  "diction",
  "eloquence",
  "expression",
  "memory",
  "precision",
  "punctuation",
  "syntax",
  "wordparts"
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateIntroText(self, introText)
  self:validateTags(introText,
                    INTRO_TEXT_SINGLE_TAGS,
                    INTRO_TEXT_PAIRED_TAGS,
                    INTRO_TEXT_TERTIARY_TAGS,
                    INTRO_TEXT_OPTIONAL_PAIRED_TAGS)
end

local function validateBodyText(self, bodyText)
  self:validateTags(bodyText,
                    BODY_TEXT_SINGLE_TAGS,
                    BODY_TEXT_PAIRED_TAGS,
                    BODY_TEXT_TERTIARY_TAGS,
                    BODY_TEXT_OPTIONAL_PAIRED_TAGS)
end

local function validateIdentifier(self, identifier)
  local skillIdentifier = string.gsub(identifier, CONCEPT_PREFIX_GAME_TIP_LONG, "")

  local errorString = string.format("Incorrect value for identifier, found: '%s' after removing '%s' prefix, expected one of: %s",
                                    skillIdentifier,
                                    CONCEPT_PREFIX_GAME_TIP_LONG,
                                    StringUtils.tostring(VALID_SKILL_IDENTIFIERS))
  self:expectIsContained(skillIdentifier, VALID_SKILL_IDENTIFIERS, "identifier", errorString)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GameTipsLongValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local introText = concept.intro_text
  if (introText ~= nil) then
    validateIntroText(self, introText)
  end

  local bodyText = concept.body_text
  if (bodyText ~= nil) then
    validateBodyText(self, bodyText)
  end

  local identifier = concept.identifier
  if (identifier ~= nil) then
    validateIdentifier(self, identifier)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GameTipsLongValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, GameTipsLongValidator)

  return self
end
