--- GameTipsShortValidator
-- @author Created by Jay Austria on November 19, 2023.
-- ConceptValidator for Weekly Report short game tips.

class("GameTipsShortValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "text"
}

-- Keep these synced with the identifiers available in games.json/subject.json, the
-- the available tip_game_[identifier] and tip_game_[identifier]_icon assets integrated
-- into griffin/pegasus, and the prefixes assumed in GameTipWeeklyReportItemProducer.cpp
local CONCEPT_PREFIX_GAME_TIP_SHORT = "game_tip_short_"
local VALID_SKILL_IDENTIFIERS = {
  "detail",
  "erroravoidance",
  -- "spelling",         -- Unsupported until tip can be reworked
  -- "adjectiverecall",  -- Unsupported until tip can be reworked
  "agility",
  "comprehension",
  "extraction",
  -- "estimation",       -- Unsupported until tip can be reworked
  "percentages",
  -- "focus",            -- Unsupported until tip can be reworked
  "names",
  "retention",
  "visualization",
  "equivalence"
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateText(self, text)
  -- As of 2023-11-19, text formatting is unsupported for game short/image tips.
  -- Ensure no formatting tags are present. Remove/replace this validation after
  -- formatting support gets added.
  local textWithoutTags = StringUtils.getAfterStrippingTags(text)

  local errorString = "Encountered formatting tags in 'text'. Text formatting not currently supported."
  self:expectEqual(text, textWithoutTags, errorString)
end

local function validateIdentifier(self, identifier)
  local skillIdentifier = string.gsub(identifier, CONCEPT_PREFIX_GAME_TIP_SHORT, "")

  local errorString = string.format("Incorrect value for identifier, found: '%s' after removing '%s' prefix, expected one of: %s",
                                    skillIdentifier,
                                    CONCEPT_PREFIX_GAME_TIP_SHORT,
                                    StringUtils.tostring(VALID_SKILL_IDENTIFIERS))
  self:expectIsContained(skillIdentifier, VALID_SKILL_IDENTIFIERS, "identifier", errorString)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GameTipsShortValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local text = concept.text
  if (text ~= nil) then
    validateText(self, text)
  end

  local identifier = concept.identifier
  if (identifier ~= nil) then
    validateIdentifier(self, identifier)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GameTipsShortValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, GameTipsShortValidator)

  return self
end
