--- GoodsEffect
-- @author Created by Aaron Barrett on May 29, 2014.

class("GoodsEffect")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUND_TO_TWO_DECIMAL_PLACES = 2
local CURRENCY_FORMAT = Localizer.get("CURRENCY_TEMPLATE")

-- Operation Types --------------------

GoodsEffect.OPERATION_NONE = "none"
GoodsEffect.OPERATION_PERCENT_DISCOUNT = "%_disc"
GoodsEffect.OPERATION_PERCENT_MARKUP = "%_up"
GoodsEffect.OPERATION_DOLLAR_DISCOUNT = "$_disc"
GoodsEffect.OPERATION_DOLLAR_MARKUP = "$_up"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function crashWithUnknownOperationError()
  error("Unknown effect operation")
end

local function getFormattedEffectString(amountString, effectText)
  return "<amount>" .. amountString .. "</amount>" .. "<br/>" ..
         "<effect>" .. effectText .. "</effect>"
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function GoodsEffect:getOperation()
  return self.operation
end

function GoodsEffect:getAmount()
  return self.amount
end

-- Building GoodsItems ----------------

function GoodsEffect:hasOperation()
  local hasOperation = (self.operation ~= GoodsEffect.OPERATION_NONE)
  return hasOperation
end

function GoodsEffect:hasPercentageOperation()
  return ((self.operation == GoodsEffect.OPERATION_PERCENT_MARKUP) or (self.operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT))
end

function GoodsEffect:getAdjustedPriceForBasePrice(basePrice)
  local operation = self.operation

  if (operation == GoodsEffect.OPERATION_NONE) then
    return MathUtils.round(basePrice, ROUND_TO_TWO_DECIMAL_PLACES)
  elseif (operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT) then
    return MathUtils.round(basePrice - self.amount * basePrice, ROUND_TO_TWO_DECIMAL_PLACES)
  elseif (operation == GoodsEffect.OPERATION_PERCENT_MARKUP) then
    return MathUtils.round(basePrice + self.amount * basePrice, ROUND_TO_TWO_DECIMAL_PLACES)
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_DISCOUNT) then
    return MathUtils.round(basePrice - self.amount, ROUND_TO_TWO_DECIMAL_PLACES)
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_MARKUP) then
    return MathUtils.round(basePrice + self.amount, ROUND_TO_TWO_DECIMAL_PLACES)
  end

  crashWithUnknownOperationError()
end

function GoodsEffect:getEffectString()
  local operation = self.operation
  local amountString = nil
  local effectText = nil

  if (operation == GoodsEffect.OPERATION_NONE) then
    return nil
  elseif (operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT) then
    amountString = tostring(self.amount * 100) .. "%"
    effectText = Localizer.get("DISCOUNT")
  elseif (operation == GoodsEffect.OPERATION_PERCENT_MARKUP) then
    local markupPercentage = self.amount * 100
    amountString = tostring(markupPercentage) .. "%"

    if (markupPercentage <= 10) then
      effectText = Localizer.get("SALES_TAX")
    else
      effectText = Localizer.get("MARKUP")
    end
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_DISCOUNT) then
    amountString = string.format(CURRENCY_FORMAT, self.amount)
    effectText = Localizer.get("OFF")
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_MARKUP) then
    amountString = string.format(CURRENCY_FORMAT, self.amount)
    effectText = Localizer.get("MARKUP")
  else
    crashWithUnknownOperationError()
  end

  return getFormattedEffectString(amountString, effectText)
end

function GoodsEffect:getPlainEffectString()
  local operation = self.operation

  if (operation == GoodsEffect.OPERATION_NONE) then
    return nil
  elseif (operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT) then
    return "a " .. tostring(self.amount * 100) .. "% " .. Localizer.get("DISCOUNT"):lower()
  elseif (operation == GoodsEffect.OPERATION_PERCENT_MARKUP) then
    return "a " .. tostring(self.amount * 100) .. "% " .. Localizer.get("MARKUP"):lower()
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_DISCOUNT) then
    return GoodsCurrencyFormatter.convertDecimalToCurrency(self.amount) .. " " .. Localizer.get("OFF"):lower()
  elseif (operation == GoodsEffect.OPERATION_DOLLAR_MARKUP) then
    return "a " .. tostring(self.amount) .. " " .. Localizer.get("MARKUP"):lower()
  else
    crashWithUnknownOperationError()
  end
end

function GoodsEffect:getEffectBannerFilename()
  local operation = self.operation

  if (operation == GoodsEffect.OPERATION_NONE) then
    return nil
  elseif (operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT or operation == GoodsEffect.OPERATION_DOLLAR_DISCOUNT) then
    return "effect_banner_orange.png"
  elseif (operation == GoodsEffect.OPERATION_PERCENT_MARKUP or operation == GoodsEffect.OPERATION_DOLLAR_MARKUP) then
    return "effect_banner_pink.png"
  end

  crashWithUnknownOperationError()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GoodsEffect.new(operation, amount)
  local self = {}
  Mixin.mix(self, GoodsEffect)

  self.operation = operation
  self.amount = amount

  return self
end
