--- GoodsGame
-- @author Created by Aaron Barrett on May 21, 2014.

class("GoodsGame")

FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("graphics.TiledImage")
require("bonuses.ResponseSpeedTracker")

require("GoodsRound")
require("GoodsItem")
require("GoodsGreenParticles")
require("GoodsInstructions")
require("GoodsTiledBackground")
require("GoodsCountdown")
require("GoodsOutro")
require("GoodsRoundGenerator")
require("GoodsConfiguration")
require("GoodsArrowButton")
require("GoodsAdvanceButton")
require("GoodsRationaleOverlay")
require("GoodsTipTextBuilder")
require("GoodsRationaleCard")
require("GoodsRationaleExample")
require("GoodsExampleLinesBuilder")
require("GoodsItemArranger")
require("GoodsCurrencyFormatter")
require("GoodsRationaleTextArea")
require("GoodsRationaleTextView")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5

local TOP_RATIONALE_Y = HEIGHT * 0.85

-- Z-Indices --------------------------

local Z_TILED_BACKGROUND = 10
local Z_COUNTDOWN = 20
local Z_GOODS_OUTRO = 30
local Z_GREEN_PARTICLES = 40
local Z_ITEM = 50
local Z_RATIONALE_OVERLAY = 60
local Z_PROMOTED_ITEM = 70
local Z_GAME_LOSS_OUTRO_TILED_BACKGROUND = 80
local Z_INSTRUCTIONS = 90
local Z_MESSAGE_LABEL = 100

-- Colors -----------------------------

local UI_COMPONENTS_COLOR = {1.0, 1.0, 1.0}

-- Game Rules -------------------------

local ROUNDS_COUNT = 5

local INITIAL_LIVES_COUNT = GoodsConfiguration.INITIAL_LIVES_COUNT

local INITIAL_ITEM_COUNT = 2
local ADDITIONAL_ITEMS_PER_ROUND = 1

local SPEED_BONUS_DURATION_PROPORTION = 0.40

-- Performance Optimizations ----------

GoodsGame.PRE_RATIONALE_OPTIMIZATION_DURATION = 0.15

-- Game Constants ---------------------

local COUNTDOWN_EXPIRED = "countdown_expired"

-- Scoring ----------------------------

local function configureGameDifficulty()
  local rankTwoScore = GoodsConfiguration.getRankTwoScore()
  local rankThreeScore = GoodsConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = GoodsConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(GoodsConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Play Sound Effects -----------------

local function playShowEffectSound()
  Sound.playSoundEffect("show_effect")
end

-- Input System -----------------------

local function waitForResponse(self)
  self.chosenItem = nil

  local inputLockAction = AHDelay.new(math.huge)
  self.inputLockAction = inputLockAction

  wait(inputLockAction:start())
  local chosenItem = self.chosenItem

  return chosenItem
end

local function submitResponse(self, chosenItem)
  self.chosenItem = chosenItem

  if (self.inputLockAction) then
    self.inputLockAction:stop()
    self.inputLockAction = nil
  end
end

-- Callback Wrappers ------------------

local function getChooseItemCallback(self, item)
  local function onChooseItem()
    submitResponse(self, item)
  end

  return onChooseItem
end

local function assignChooseItemCallbacks(self, round)
  local items = round:getItems()

  for _, item in ipairs(items) do
    Button.makeButton(item, getChooseItemCallback(self, item))
  end
end

local function getCountdownExpirationCallback(self)
  local function onCountdownExpirationCallback()
    submitResponse(self, COUNTDOWN_EXPIRED)
  end

  return onCountdownExpirationCallback
end

-- Pre-Round Presentation -------------

local function prepareItems(self, round)
  local items = round:getItems()

  self.itemArranger:arrangeItems(items)

  local greenParticles = self.greenParticles
  greenParticles:clearItems()

  for _, item in ipairs(items) do
    item:prepareToUnveil()
    self:insert(item, Z_ITEM)

    greenParticles:addItem(item)
  end
end

local function waitForPresentRound(self, round)
  local countdown = self.countdown
  local messageLabel = self.messageLabel

  -- Shuffle the items so that they are not presented in the same order
  -- they are listed in the content.
  local shuffledItems = round:createShuffledItems()

  local unveilAction = AHSpawn.new()
  local unveilEffectsAction = AHSpawn.new()

  local unveilDelayDurationIncrement = 0.15
  for i, item in ipairs(shuffledItems) do
    local presentationDelayDuration = i * unveilDelayDurationIncrement

    local itemUnveilAction = item:createUnveilAction()
    local delayedItemUnveilAction = AHSequence.new(
                                                  AHDelay.new(presentationDelayDuration),
                                                  itemUnveilAction
                                                  )
    unveilAction:addAction(delayedItemUnveilAction)

    local itemUnveilEffectAction = item:createUnveilEffectAction()
    unveilEffectsAction:addAction(itemUnveilEffectAction)
  end

  local roundNumber = round:getRoundNumber()
  local roundTextTemplate = Localizer.get("ROUND_OF")
  messageLabel:setText(string.format(roundTextTemplate, roundNumber, ROUNDS_COUNT))
  AHOpacityTo.new(messageLabel, 0.2, nil, 1.0):start()

  unveilAction:start()

  wait(countdown:playRefill())

  playShowEffectSound()
  wait(unveilEffectsAction:start())
  wait(AHOpacityTo.new(messageLabel, 0.2, nil, 0.0):start())

  local countdownDuration = round:getCountdownDuration()
  countdown:restartWithDuration(countdownDuration)
  local speedBonusWindowDuration = SPEED_BONUS_DURATION_PROPORTION * countdownDuration
  local decayWindowStart = 0
  self.responseSpeedTracker:setBonusWindowDuration(speedBonusWindowDuration)
  self.responseSpeedTracker:setDecayWindowStart(decayWindowStart)
  self.responseSpeedTracker:notifyStart()

  assignChooseItemCallbacks(self, round)
end

-- Dismiss Displays -------------------

local function dismissInstructions(self)
  if (self.instructions ~= nil) then
    self.instructions:playDisappear()
    self.messageLabel:setLoc(CENTER_X, HEIGHT * 0.83)
    self.instructions = nil
  end
end

local function resumeAfterIncorrect(self, round)
  if (not self.countdown:isTimedOut()) then
    self.countdown:resumeAfterIncorrect()
  end

  local unfinishedItems = round:getUnfinishedItems()
  for item in pairs(unfinishedItems) do
    item:setTouchEnabled(true)
  end
end

local function restoreAllItemsToOriginalZIndex(self, round)
  local items = round:getItems()
  for _, item in ipairs(items) do
    item:setZIndex(Z_ITEM)
  end
end

local function dismissRationaleOverlay(self, round)
  if (self.rationaleOverlay ~= nil) then
    local fadeOutAction = self.rationaleOverlay:playFadeOut()
    runInCoroutine(function()
      wait(fadeOutAction)
      restoreAllItemsToOriginalZIndex(self, round)
    end)
    self.rationaleOverlay = nil
  end
end

local function playMoveItemsToHomeY(self, round)
  local items = round:getItems()
  for _, item in ipairs(items) do
    item:playMoveToHomeY()
  end
end

local function dismissIncorrect(self, round)
  local incorrectChosenItem = self.incorrectChosenItem
  if (incorrectChosenItem ~= nil) then

    self.incorrectChosenItem = nil

    incorrectChosenItem:revertRed()
  end

  dismissRationaleOverlay(self, round)

  waitFor(0.4)
  playMoveItemsToHomeY(self, round)
  waitFor(1.05)

  resumeAfterIncorrect(self, round)
end

-- Notify Correct/Incorrect Answers ---

local function updateScoreLabel(self)
  self.scoreLabel:setScore(GameSession:getScore())
end

local function loseLife(self, round)
  local responseSpeedTracker = self.responseSpeedTracker
  responseSpeedTracker:notifyIncorrectResponse()

  round:setDidMissAny(true)

  local livesDisplay = self.livesDisplay
  livesDisplay:decrement()

  local didLoseGame = (livesDisplay:getLives() <= 0)
  return didLoseGame
end

local function scoreCorrect(self, round, chosenItem)
  if (chosenItem:isScoringAllowed()) then
    GameSession.increaseScore(GoodsConfiguration.POINTS_PER_CORRECT_RESPONSE)
    updateScoreLabel(self)
  end

  local cheapestItem = round:getCheapestItem()
  cheapestItem:stopPickMe()

  Sound.playSoundEffect("correct")

  chosenItem:setTouchEnabled(false)
  chosenItem:playCorrect()

  round:finishItem(chosenItem)
end

-- Auto-tap ---------------------------

local function autoTapFinalItem(self, round)
  if not self.countdown:isTimedOut() then
    waitForResponse(self)
  else
    -- If already timed out, it means we have shown the round 1 help: Just wait for a small delay
    -- before auto-tapping the last item.
    local autoTapAfterHelpDelay = 0.5
    waitFor(autoTapAfterHelpDelay)
  end

  self.countdown:stop()

  local mostExpensiveItem = round:getMostExpensiveItem()
  scoreCorrect(self, round, mostExpensiveItem)
end

-- Rationale Overlay ------------------

local function waitForFadeInRationaleOverlay(self, chosenItem, cheapestItem)
  local autoTapAndDismissCallback = getChooseItemCallback(self, cheapestItem)
  local rationaleOverlay = GoodsRationaleOverlay.new(chosenItem, cheapestItem, autoTapAndDismissCallback)
  self:insertAtCenter(rationaleOverlay, Z_RATIONALE_OVERLAY)
  self.rationaleOverlay = rationaleOverlay
  wait(rationaleOverlay:playFadeIn())
end

-- Incorrect Actions ------------------

local function createAllItemsIncorrectAction(self, round)
  local allItemsIncorrectAction = AHSpawn.new()

  local unfinishedItems = round:getUnfinishedItems()
  local shouldRevertRed = true
  for item in pairs(unfinishedItems) do
    local itemIncorrectAction = item:createIncorrectAction(shouldRevertRed)
    allItemsIncorrectAction:addAction(itemIncorrectAction)
  end

  return allItemsIncorrectAction
end

-- Incorrect Animation ----------------

local function playMoveItemsToRationaleY(self, round)
  runInCoroutine(function()
    waitFor(GoodsGame.PRE_RATIONALE_OPTIMIZATION_DURATION)
    local items = round:getItems()
    for _, item in ipairs(items) do
      item:playMoveToRationaleY()
    end
  end)
end

local function waitForCountdownExpired(self, round)
  dismissInstructions(self)

  local cheapestItem = round:getCheapestItem()
  cheapestItem:disallowScoring()

  local unfinishedItems = round:getUnfinishedItems()
  for item in pairs(unfinishedItems) do
    item:setTouchEnabled(false)
  end

  local incorrectAction = createAllItemsIncorrectAction(self, round)
  incorrectAction:start()

  cheapestItem:setZIndex(Z_PROMOTED_ITEM)

  Sound.playSoundEffect("incorrect")

  waitFor(0.25)

  self.incorrectChosenItem = nil

  playMoveItemsToRationaleY(self, round)
  self.countdown:playRefill()

  local chosenItem = nil
  waitForFadeInRationaleOverlay(self, chosenItem, cheapestItem)
end

local function waitForIncorrect(self, round, chosenItem)
  dismissInstructions(self)

  local cheapestItem = round:getCheapestItem()
  cheapestItem:disallowScoring()

  local unfinishedItems = round:getUnfinishedItems()
  for item in pairs(unfinishedItems) do
    item:setTouchEnabled(false)
  end

  chosenItem:setZIndex(Z_PROMOTED_ITEM)
  cheapestItem:setZIndex(Z_PROMOTED_ITEM)

  self.countdown:fluster()
  Sound.playSoundEffect("incorrect")
  chosenItem:playIncorrect()
  waitFor(0.25)

  self.incorrectChosenItem = chosenItem

  playMoveItemsToRationaleY(self, round)
  waitForFadeInRationaleOverlay(self, chosenItem, cheapestItem)
end

-- Post-round Transitions -------------

local function waitForShrinkTransition(round)
  local shrinkAction = AHSpawn.new()

  local orderedItems = round:getOrderedItems()
  local itemShrinkStaggerDelayDuration = 0.07
  for i, item in ipairs(orderedItems) do
    local itemShrinkAction = item:createShrinkAction()
    local delayedItemShrinkAction = AHSequence.new(
                                                   AHDelay.new(itemShrinkStaggerDelayDuration * (i - 1)),
                                                   itemShrinkAction
                                                   )

    shrinkAction:addAction(delayedItemShrinkAction)
  end

  local preShrinkDelayDuration = 0.8
  waitFor(preShrinkDelayDuration)
  shrinkAction:start()

  -- Start switching to the next round immediately as the final item starts shrinking.
  local postShrinkDelayDuration = (round:getItemCount() - 1.0) * itemShrinkStaggerDelayDuration
  waitFor(postShrinkDelayDuration)
end

-- Instructions -----------------------

local function showInstructions(self)
  self.instructions:playAppear()

  local postInstructionsDelayDuration = 0.5
  waitFor(postInstructionsDelayDuration)
end

-- Win Outro --------------------------

local function waitForGameWinOutro(self)
  local goodsOutro = GoodsOutro.new()
  self:insertAtCenter(goodsOutro, Z_GOODS_OUTRO)

  local preOutroDelayDuration = 0.7
  waitFor(preOutroDelayDuration)

  Sound.playSoundEffect("win")
  self:showDefaultGameWinOutro()
  wait(goodsOutro:playOutro())
end

-- Loss Outro -------------------------

local function playGameLossOutroFade(self)
  local outroBackground = GoodsTiledBackground.new()
  self:insertAtCenter(outroBackground, Z_GAME_LOSS_OUTRO_TILED_BACKGROUND)

  return outroBackground:playLossOutroFadeIn()
end

local function waitForGameLossOutro(self, round, chosenItem)
  scoreCorrect(self, round, chosenItem)

  dismissRationaleOverlay(self, round)
  waitFor(0.4)
  Sound.playSoundEffect("fall")
  playMoveItemsToHomeY(self, round)
  playGameLossOutroFade(self)
  waitFor(1.05)

  self:showDefaultGameLossOutro()
  GameSession.notifyGameLoss()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function GoodsGame:gameFlow()
  showInstructions(self)

  local countdownDuration = GoodsConfiguration.INITIAL_COUNTDOWN_DURATION
  local itemCount = INITIAL_ITEM_COUNT

  for roundNumber = 1, ROUNDS_COUNT do
    local round = self.roundGenerator:getNextRound(roundNumber, itemCount, countdownDuration)

    local conceptId = round:getIdentifier()
    local difficulty = round:getDifficulty()
    GameSession.reportSeenConcept(conceptId)
    GameSession.contentTracker:notifyPresentedConcept(conceptId, difficulty)

    prepareItems(self, round)
    waitForPresentRound(self, round)

    local didCountdownExpire = false
    local didLoseGame = false
    local questionIdentifier = 1
    local missedCurrentQuestion = false
    local didLoseLife = false

    repeat
      local chosenItem = waitForResponse(self, round)
      didCountdownExpire = (chosenItem == COUNTDOWN_EXPIRED)

      local questionIdentifierAsString = tostring(questionIdentifier)

      -- If the countdown expired...
      if (didCountdownExpire) then
        GameSession.contentTracker:notifyTimeout(conceptId, questionIdentifierAsString)
        missedCurrentQuestion = true

        if (roundNumber > 1) then
          didLoseGame = loseLife(self, round)
          didLoseLife = true
        end
        waitForCountdownExpired(self, round)

      -- If the player chose the wrong item...
      elseif (chosenItem:getAdjustedPrice() > round:getCheapestItem():getAdjustedPrice()) then
        GameSession.contentTracker:notifyIncorrectAnswer(conceptId, questionIdentifierAsString)
        GameSession.incrementIncorrectAnswerCount()
        missedCurrentQuestion = true

        if (roundNumber > 1) then
          didLoseGame = loseLife(self, round)
          didLoseLife = true
        end
        waitForIncorrect(self, round, chosenItem)

      -- If the player chose the correct item...
      else

        -- NOTE: Player HAS to pick the correct answer after an incorrect one
        -- to continue the game
        if (didLoseLife) then
          if (didLoseGame) then
            waitForGameLossOutro(self, round, chosenItem)
            return
          end
        else
          if (not missedCurrentQuestion) then
            GameSession.contentTracker:notifyCorrectAnswer(conceptId, questionIdentifierAsString)
          end

          GameSession.incrementCorrectAnswerCount()
        end

        scoreCorrect(self, round, chosenItem)

        if (self.rationaleOverlay ~= nil) then
          dismissIncorrect(self, round)
        end

        dismissInstructions(self)

        questionIdentifier = questionIdentifier + 1
        missedCurrentQuestion = false
        didLoseLife = false
        didLoseGame = false
      end

    -- Don't handle the final item as part of the loop
    until (round:isOnFinalItem())

    if (not didCountdownExpire) then
      autoTapFinalItem(self, round)
      self.responseSpeedTracker:notifyCorrectResponse()
      updateScoreLabel(self)
      waitForShrinkTransition(round)
    end

    if (not round:getDidMissAny()) then
      countdownDuration = countdownDuration + GoodsConfiguration.ADDITIONAL_COUNTDOWN_DURATION_PER_ROUND
      itemCount = itemCount + ADDITIONAL_ITEMS_PER_ROUND
    end

  end
  ScoringUtils.applyLivesBonus(self.livesDisplay:getLives(), GoodsConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
  self.scoreLabel:setScore(GameSession.getScore())

  waitForGameWinOutro(self)

  GameSession.notifyGameWin()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local tiledBackground = GoodsTiledBackground.new()
  self:insertAtCenter(tiledBackground, Z_TILED_BACKGROUND)

  local instructions = GoodsInstructions.new()
  instructions:setLoc(CENTER_X, HEIGHT - instructions:getHeightPart(0.5))
  self:insert(instructions, Z_INSTRUCTIONS)
  self.instructions = instructions

  local messageLabel = Label.new("", Font.new(FontFaces.DIN_OT_BOLD, 28.0))
  messageLabel:setColor(unpack(UI_COMPONENTS_COLOR))
  messageLabel:setOpacity(0.0)
  messageLabel:setLoc(CENTER_X, Screen.getSafeY(0.76))
  self:insert(messageLabel, Z_MESSAGE_LABEL)
  self.messageLabel = messageLabel

  local countdown = GoodsCountdown.new(getCountdownExpirationCallback(self))
  self:insertAtCenter(countdown, Z_COUNTDOWN)
  self.countdown = countdown

  local greenParticles = GoodsGreenParticles.new()
  self:insertAtCenter(greenParticles, Z_GREEN_PARTICLES)
  self.greenParticles = greenParticles
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesDisplay = builder:addLivesDisplay(INITIAL_LIVES_COUNT)

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 GoodsConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end
  local bonusWindowDuration = nil --Goods sets the bonus window duration at the beginning of each round.
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowDuration, quickResponseWrapper)
end

-- Constructor ------------------------

function GoodsGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, GoodsGame)

  createGameElements(self)
  createUI(self)
  configureGameDifficulty()

  self.itemArranger = GoodsItemArranger.new()
  self.countdownAction = nil
  self.inputLockAction = nil
  self.chosenItem = nil

  self.incorrectChosenItem = nil

  self.rationaleOverlay = nil

  self.roundGenerator = GoodsRoundGenerator.new()

  return self
end
