--- GoodsItem
-- @author Created by Aaron Barrett on May 22, 2014.

class("GoodsItem")
require("GoodsEffect")
require("Particles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CURRENCY_FORMAT = Localizer.get("CURRENCY_TEMPLATE")

-- Dimensions -------------------------

GoodsItem.WIDTH = 100.0
GoodsItem.HEIGHT = 137.5

local TAG_HOLDER_INITIAL_Y = 110.0

local CENTER_X = GoodsItem.WIDTH * 0.5

local BANNER_CENTER_X = 40.0

local ICON_Y = 76.0

local BASE_PRICE_LABEL_WIDTH = 82.0

local EFFECT_BANNER_INITIAL_Y = 48.0

local EFFECT_LABEL_WIDTH = 72.0

local ADJUSTED_PRICE_BANNER_INITIAL_Y = 68.0
local ADJUSTED_PRICE_BANNER_VISIBLE_Y = 32.0

local ADJUSTED_PRICE_LABEL_WIDTH = 72.0

local ADJUSTED_PRICE_LABEL_Y = 21.0

local HIGHLIGHT_SCALE_MAX = 0.93
local HIGHLIGHT_SCALE_MIN_X = 0.87
local HIGHLIGHT_SCALE_MIN_Y = 0.89
local HIGHLIGHT_SCALE_MIN_FADE_OUT = HIGHLIGHT_SCALE_MIN_X

-- Particles --------------------------

local PARTICLES_COUNT_PER_ITEM = 48

-- Fonts ------------------------------

local PRICE_LABEL_FONT = FontFaces.DIN_OT_MEDIUM

local EFFECT_TEXT_ATTRIBUTES = {
  amount = {font = Font.new(PRICE_LABEL_FONT, FontScaling.getInverseScaledSize(17))},
  effect = {font = Font.new(PRICE_LABEL_FONT, FontScaling.getInverseScaledSize(12)),
            fitInLine = true}
}

-- Z-Indices --------------------------

-- Housed in self
local Z_ADJUSTED_PRICE_BANNER = 10
local Z_TAG_HOLDER = 20

-- Housed in tagHolder
local Z_HIGHLIGHT = 10
local Z_EFFECT_BANNER = 20
local Z_TAG_TOP = 30
local Z_TAG_BOTTOM = 40
local Z_BASE_PRICE_LABEL = 50

-- Housed in tagTop/tagBottom
local Z_RED = 10
local Z_ICON = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Icons ------------------------------

local function setIcon(self, icon)
  if (self.icon) then
    self.icon:removeSelf()
  end

  icon:setLoc(CENTER_X, ICON_Y)

  self.tagTop:insert(icon, Z_ICON)
  self.icon = icon
end

-- Play Sound Effects -----------------

local function playShowItemSound()
  Sound.playSoundEffect("show_item")
end

-- Turning red ------------------------

local function setRedVisible(self, isRedVisible)
  self.tagTopRed:setVisible(isRedVisible)
  self.tagBottomRed:setVisible(isRedVisible)

  local isBlueVisible = (not isRedVisible)
  self.tagTop:setVisible(isBlueVisible)
  self.tagBottom:setVisible(isBlueVisible)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function GoodsItem:getEffect()
  return self.effect
end

function GoodsItem:hasPercentageOperation()
  return self.effect:hasPercentageOperation()
end

function GoodsItem:getAdjustedPrice()
  return self.adjustedPrice
end

function GoodsItem:getBasePrice()
  return self.basePrice
end

function GoodsItem:setIconName(name)
  self.iconName = name
  local icon = Sprite.newWithName(name)
  setIcon(self, icon)
end

function GoodsItem:getIconName()
  return self.iconName
end

function GoodsItem:setParticleSystem(particleSystem)
  self.particleSystem = particleSystem
end

function GoodsItem:isScoringAllowed()
  return self.allowScoring
end

function GoodsItem:setHomeLocation(x, y)
  self.homeY = y
  self:setLoc(x, y)
end

function GoodsItem:setRationaleY(y)
  self.rationaleY = y
end

-- Correct Animation ------------------

function GoodsItem:playCorrect()
  self:hideHighlightImmediately()
  local correctAction = self:createCorrectAction()
  return correctAction:start()
end

function GoodsItem:createCorrectAction()
  local initialDipDuration = 0.11
  local riseDuration = 0.15
  local finalDipDuration = 0.125

  local tagHolder = self.tagHolder
  local verticalEase = AHEaseSimpleInOut.new(1.5)
  local initialDipAction = AHLocationYTo.new(tagHolder, initialDipDuration, verticalEase, 75.0)
  local riseAction = AHLocationYTo.new(tagHolder, riseDuration, verticalEase, 134.0)
  local finalDipAction = AHLocationYTo.new(tagHolder, finalDipDuration, verticalEase, 96.0)

  local retractEffectBannerAction = AHDoNothing.new()
  local shouldRetractEffectBanner = (self.effect:getOperation() ~= GoodsEffect.OPERATION_NONE)
  if (shouldRetractEffectBanner) then
    local effectBanner = self.effectBanner
    retractEffectBannerAction = AHSequence.new(
                                               AHDelay.new(initialDipDuration * 0.35),
                                               AHLocationYTo.new(effectBanner, initialDipDuration * 0.65, nil, 32.0),
                                               function()
                                                 effectBanner:setVisibleRecursive(false)
                                               end
                                               )
  end

  local flipAnimation = self.flipAnimation
  local flipDuration = (riseDuration + finalDipDuration) * 1.2
  local flipFramerate = flipAnimation:getFrameCount() / flipDuration
  local flipAction = AHAnimate.new(self.tagBottom, flipAnimation, flipFramerate)

  local basePriceShrinkDuration = flipDuration * 0.45
  local basePriceAction = AHScaleYTo.new(self.basePriceLabel, basePriceShrinkDuration, AHEaseSimpleIn.new(2.5), 0.0)

  local adjustedPriceBanner = self.adjustedPriceBanner
  local adjustedPriceDelayDuration = flipDuration * 0.45
  local adjustedPriceAppearDuration = flipDuration - adjustedPriceDelayDuration
  local adjustedPriceAction = AHSequence.new(
    AHDelay.new(adjustedPriceDelayDuration),
    function()
      adjustedPriceBanner:setVisibleRecursive(true)
    end,
    AHLocationYTo.new(adjustedPriceBanner,
                      adjustedPriceAppearDuration * 0.6,
                      verticalEase,
                      ADJUSTED_PRICE_BANNER_VISIBLE_Y - 6.5),
    AHLocationYTo.new(adjustedPriceBanner,
                      adjustedPriceAppearDuration * 0.4,
                      verticalEase,
                      ADJUSTED_PRICE_BANNER_VISIBLE_Y)
  )

  local particlesDelayDuration = initialDipDuration * 0.25
  local resetParticlesAction = AHDelayed.new(particlesDelayDuration, function()
                                                                    local particleSystem = self.particleSystem
                                                                    particleSystem:surge(PARTICLES_COUNT_PER_ITEM)
                                                                  end)


  local correctAction = AHSequence.new(
                                       function()
                                         setRedVisible(self, false)
                                       end,
                                       AHSpawn.new(
                                                   retractEffectBannerAction,
                                                   initialDipAction,
                                                   resetParticlesAction
                                                   ),
                                       AHSpawn.new(
                                                   flipAction,
                                                   basePriceAction,
                                                   adjustedPriceAction,
                                                   AHSequence.new(
                                                                  riseAction,
                                                                  finalDipAction
                                                                  )
                                                   )
                                       )

  return correctAction
end

-- Incorrect Animation ----------------

function GoodsItem:createIncorrectAction(shouldRevertRed)
  local tagHolder = self.tagHolder
  local halfCycleDuration = 0.05
  local ease = AHEaseSimpleInOut.new(2.0)
  local incorrectAction = AHSequence.new(
                                         function()
                                           setRedVisible(self, true)
                                         end,
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, 5.0),
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, -5.0),
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, 2.5),
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, -2.5),
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, 2.0),
                                         AHLocationXBy.new(tagHolder, halfCycleDuration, ease, -2.0),
                                         function()
                                           if (shouldRevertRed) then
                                             setRedVisible(self, false)
                                           end
                                         end
                                         )

  return incorrectAction
end

function GoodsItem:revertRed()
  setRedVisible(self, false)
end

function GoodsItem:playIncorrect()
  local incorrectAction = self:createIncorrectAction()
  return incorrectAction:start()
end

function GoodsItem:disallowScoring()
  self.allowScoring = false
end

function GoodsItem:playMoveToRationaleY()
  return AHLocationYTo.new(self, 1.0, AHEaseExponentialOut.new(), self.rationaleY):start()
end

function GoodsItem:playMoveToHomeY()
  return AHLocationYTo.new(self, 1.0, AHEaseExponentialOut.new(), self.homeY):start()
end

function GoodsItem:hideHighlightImmediately()
  self.highlightInterrupt:stop()
  self.highlight:setVisible(false)
  self.isHighlightPulsing = false
end

function GoodsItem:setHighlightPulsing(isPulsing)
  if (isPulsing ~= self.isHighlightPulsing) then
    local pulseAction = nil
    local ease = AHEaseBezier.newFlat({0.52, 0.0, 0.5, 1.0})

    if (isPulsing) then
        self.highlight:setVisible(true)
        self.highlight:setOpacity(0.0)
        pulseAction = AHLoop.new(
          AHSpawn.new(
            function()
              self.highlight:setOpacity(0.0)
              self.highlight:setScl(HIGHLIGHT_SCALE_MIN_X, HIGHLIGHT_SCALE_MIN_Y)
            end,
            AHScaleTo.new(self.highlight, 2.5, ease, HIGHLIGHT_SCALE_MAX),
            AHSequence.new(
              AHOpacityTo.new(self.highlight, 1.5, ease, 1.0),
              AHOpacityTo.new(self.highlight, 1.0, ease, 0.0)
            )
          )
        )
    else
      pulseAction = AHSequence.new(
        AHSpawn.new(
          AHScaleTo.new(self.highlight, 0.5, ease, HIGHLIGHT_SCALE_MIN_FADE_OUT),
          AHOpacityTo.new(self.highlight, 0.5, ease, 0.0)
        ),
        function()
          self.highlight:setVisible(false)
        end
      )
    end
    self.highlightInterrupt:insert(pulseAction)
    self.isHighlightPulsing = isPulsing
  end
end

-- Fall Animation ---------------------

function GoodsItem:createFallAction()
  local destinationXOffset = math.random() * 10.0 - 5.0
  local destinationRotation = math.random() * 12.0 - 6.0

  local fallAction = AHSpawn.new(
    AHSequence.new(
      AHLocationYBy.new(self, 0.2, AHEaseSimpleInOut.new(4.0), 8.0),
      AHLocationYBy.new(self, 0.6, AHEaseSimpleInOut.new(4.0), -Screen.height),
      function()
        self:removeSelf()
      end
      ),
    AHLocationXBy.new(self, 0.8, AHEaseSimpleInOut.new(2.0), destinationXOffset),
    AHRotationTo.new(self, 0.8, AHEaseSimpleInOut.new(2.0), destinationRotation)
  )

  return fallAction
end

-- Unveil Animations ------------------

function GoodsItem:prepareToUnveil()
  self:setScl(1.5)

  self.tagTop:setOpacity(0.0)
  self.tagBottom:setOpacity(0.0)
  self.icon:setOpacity(0.0)
  self.basePriceLabel:setOpacity(0.0)
end

function GoodsItem:createUnveilAction()
  local tagHolder = self.tagHolder

  local swayEase = AHEaseSimpleInOut.new(2.0)
  local animDuration = 0.15

  local unveilAction = AHSequence.new(
    AHSpawn.new(
      playShowItemSound,
      AHScaleTo.new(self, animDuration, AHEaseSimpleOut.new(2.0), 0.9),
      AHOpacityTo.new(self.tagTop, animDuration, nil, 1.0),
      AHOpacityTo.new(self.tagBottom, animDuration, nil, 1.0),
      AHOpacityTo.new(self.icon, animDuration, nil, 1.0),
      AHOpacityTo.new(self.basePriceLabel, animDuration, nil, 1.0)
    ),
    AHSequence.new(
      AHSpawn.new(
        AHScaleTo.new(self, 0.05, AHEaseSimpleOut.new(2.0), 1.0),
        AHRotationTo.new(tagHolder, 0.075, swayEase, 4.0)
      ),
      AHRotationTo.new(tagHolder, animDuration, swayEase, -3.0),
      AHRotationTo.new(tagHolder, animDuration, swayEase, 2.0),
      AHRotationTo.new(tagHolder, animDuration, swayEase, -1.0),
      AHRotationTo.new(tagHolder, animDuration, swayEase, 0.5),
      AHRotationTo.new(tagHolder, animDuration, swayEase, 0.0)
    )
  )

  return unveilAction
end

function GoodsItem:createUnveilEffectAction()
  local effectBanner = self.effectBanner
  local effectLabel = self.effectLabel
  if (effectLabel ~= nil) then
    local deltaY = -effectBanner:getHeightPart(0.5)
    local unveilEffectAction = AHSequence.new(
                                              function()
                                                effectBanner:setVisibleRecursive(true)
                                              end,
                                              AHLocationYBy.new(effectBanner, 0.5, AHEaseElasticOut.new(0.36), deltaY)
                                              )

    return unveilEffectAction
  end

  return AHDoNothing.new()
end

-- Shrink Transition Animation --------

function GoodsItem:createShrinkAction()
  local swayEase = AHEaseSimpleInOut.new(2.0)

  local shrinkAction = AHSpawn.new(
                                   AHSequence.new(
                                                  AHRotationTo.new(self, 0.075, swayEase, 3.0),
                                                  AHRotationTo.new(self, 0.15, swayEase, -2.0),
                                                  AHRotationTo.new(self, 0.15, swayEase, 1.0),
                                                  function()
                                                    self:removeSelf()
                                                  end
                                                  ),
                                   AHSequence.new(
                                                  AHScaleTo.new(self, 0.075, AHEaseSimpleOut.new(2.0), 1.08),
                                                  AHScaleTo.new(self, 0.2, AHEaseSimpleIn.new(2.0), 0.0)
                                                  )
                                   )

  return shrinkAction
end

-- "Pick Me!" Animation ---------------

function GoodsItem:createPickMeAction()
  local tagHolder = self.tagHolder

  local easeInOut = AHEaseSimpleInOut.new(2.0)
  local pickMeAction = AHLoop.new(
    AHSequence.new(
      AHLocationYTo.new(tagHolder, 0.037, easeInOut, TAG_HOLDER_INITIAL_Y - 10.0),
      AHLocationYTo.new(tagHolder, 0.15, easeInOut, TAG_HOLDER_INITIAL_Y + 7.5),
      AHLocationYTo.new(tagHolder, 0.1, easeInOut, TAG_HOLDER_INITIAL_Y - 3.5),
      AHLocationYTo.new(tagHolder, 0.075, easeInOut, TAG_HOLDER_INITIAL_Y),
      AHDelay.new(2.0)
    )
  )

  return pickMeAction
end

function GoodsItem:playPickMe()
  local pickMeAction = self:createPickMeAction()

  self.pickMeAction = pickMeAction

  return pickMeAction:start()
end

function GoodsItem:stopPickMe()
  if (self.pickMeAction) then
    self.pickMeAction:stop()
  end

  local settleAction = AHLocationYTo.new(self.tagHolder, 0.075, AHEaseSimpleInOut.new(2.0), TAG_HOLDER_INITIAL_Y)
  self.pickMeAction = settleAction
  settleAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createPriceLabel(text, labelWidth, baseAttributes, taggedAttributes)
  local label = DynamicTextArea.new(text, labelWidth, baseAttributes, taggedAttributes, Alignment.CENTER, nil, 0, true)

  return label
end

-- Constructor ------------------------

function GoodsItem.new(basePrice, effect)
  local self = Group.new(GoodsItem.WIDTH, GoodsItem.HEIGHT)
  Mixin.mix(self, GoodsItem)

  self.basePrice = basePrice
  self.effect = effect

  local tagHolder = Group.new(GoodsItem.WIDTH, GoodsItem.HEIGHT)
  tagHolder:setAnchorPoint(0.5, 0.8)
  tagHolder:setLoc(CENTER_X, TAG_HOLDER_INITIAL_Y)
  self:insert(tagHolder, Z_TAG_HOLDER)
  self.tagHolder = tagHolder

  local tagTop = Sprite.newWithName("tag_top.png")
  tagHolder:insertAtCenter(tagTop, Z_TAG_TOP)
  self.tagTop = tagTop

  local tagTopRed = Sprite.newWithName("tag_top_red.png")
  tagTop:insertAtCenter(tagTopRed, Z_RED)
  self.tagTopRed = tagTopRed

  local tagBottom = Sprite.newWithName("tag_bottom.png")
  tagHolder:insertAtCenter(tagBottom, Z_TAG_BOTTOM)
  self.tagBottom = tagBottom

  local tagBottomRed = Sprite.newWithName("tag_bottom_red.png")
  tagBottom:insertAtCenter(tagBottomRed, Z_RED)
  self.tagBottomRed = tagBottomRed

  setRedVisible(self, false)

  local highlight = Sprite.newWithName("highlight.png")
  highlight:setScl(HIGHLIGHT_SCALE_MIN_X, HIGHLIGHT_SCALE_MIN_Y)
  highlight:setX(GoodsItem.WIDTH * 0.5)
  local verticalFudgeAmount = 1.5
  highlight:setY(GoodsItem.HEIGHT * 0.5 - ADJUSTED_PRICE_LABEL_Y - verticalFudgeAmount)
  highlight:setVisible(false)
  self.tagHolder:insert(highlight, Z_HIGHLIGHT)
  self.highlight = highlight

  local basePriceString = string.format(CURRENCY_FORMAT, basePrice)
  local basePriceAttributes = { font = Font.new(PRICE_LABEL_FONT, 18) }
  local basePriceLabel = createPriceLabel(basePriceString, BASE_PRICE_LABEL_WIDTH, basePriceAttributes)
  local priceLabelY = tagHolder:getHeightPart(0.27)
  basePriceLabel:setLoc(CENTER_X, priceLabelY)
  tagHolder:insert(basePriceLabel, Z_BASE_PRICE_LABEL)
  self.basePriceLabel = basePriceLabel

  local adjustedPriceBanner = Sprite.newWithName("adjusted_price_banner.png")
  adjustedPriceBanner:setLoc(CENTER_X, ADJUSTED_PRICE_BANNER_INITIAL_Y)
  self:insert(adjustedPriceBanner, Z_ADJUSTED_PRICE_BANNER)
  self.adjustedPriceBanner = adjustedPriceBanner

  local adjustedPrice = effect:getAdjustedPriceForBasePrice(basePrice)
  self.adjustedPrice = adjustedPrice

  local effectBanner = nil
  local effectLabel = nil
  if (effect:hasOperation()) then
    local effectBannerFilename = effect:getEffectBannerFilename()
    effectBanner = Sprite.newWithName(effectBannerFilename)
    effectBanner:setLoc(CENTER_X, EFFECT_BANNER_INITIAL_Y)
    tagHolder:insert(effectBanner, Z_EFFECT_BANNER)
    local effectString = effect:getEffectString()
    local effectAttributes = { font = Font.new(PRICE_LABEL_FONT, 11) }
    effectLabel = createPriceLabel(effectString, EFFECT_LABEL_WIDTH, effectAttributes, EFFECT_TEXT_ATTRIBUTES)
    local effectLabelY = effectBanner:getHeightPart(0.4)
    effectLabel:setLoc(BANNER_CENTER_X, effectLabelY)
    effectBanner:insert(effectLabel)

    effectBanner:setVisibleRecursive(false)
  end

  local adjustedPriceString = string.format(CURRENCY_FORMAT, adjustedPrice)
  local adjustedPriceLabel = createPriceLabel(adjustedPriceString, ADJUSTED_PRICE_LABEL_WIDTH, basePriceAttributes)
  adjustedPriceLabel:setLoc(BANNER_CENTER_X, ADJUSTED_PRICE_LABEL_Y)
  adjustedPriceBanner:insert(adjustedPriceLabel)
  self.adjustedPriceLabel = adjustedPriceLabel

  self.effectBanner = effectBanner
  self.effectLabel = effectLabel
  self.highlightInterrupt = ActionInterrupt.new()
  self.isHighlightPulsing = false
  self.iconName = nil
  self.icon = nil
  self.pickMeAction = nil
  self.particleSystem = nil
  self.homeY = nil
  self.rationaleY = nil
  self.allowScoring = true

  adjustedPriceBanner:setVisibleRecursive(false)

  self.flipAnimation = Animation.newWithPrefix("tag_bottom_correct_")

  return self
end
