--- GoodsRationaleCard
-- @author Created by Aaron Barrett on July 12, 2021.

class("GoodsRationaleCard")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.safeWidth

local X_OFFSCREEN_RIGHT = Screen.getWidthPart(1.5)
local X_ONSCREEN_CENTER = Screen.getWidthPart(0.5)
local X_OFFSCREEN_LEFT = -Screen.getWidthPart(0.5)

local HORIZONTAL_MARGIN = 30.0

local BOTTOM_MARGIN = math.max(28.0, Screen.safeBottom)

local EXAMPLE_CENTER_Y = BOTTOM_MARGIN + Screen.getHeightPart(0.11)

local ARROW_BUTTON_SIDE_MARGIN = 10.0

local TIP_HEADER_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 16.0)

local LEARN_MORE_TEXT = "LEARN MORE →"
local RETURN_TO_GAME_TEXT = Localizer.get("Misc.RETURN_TO_GAME")

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function doesEitherItemHavePercentageOperation(self)
  return (self.cheapestItem:hasPercentageOperation() or ((self.chosenItem ~= nil) and self.chosenItem:hasPercentageOperation()))
end

local function doesNeitherItemHavePercentageOperation(self)
  return (not doesEitherItemHavePercentageOperation(self))
end

local function doBothItemsHavePercentageOperation(self)
  return ((self.chosenItem ~= nil) and self.chosenItem:hasPercentageOperation() and self.cheapestItem:hasPercentageOperation())
end

local function getHighlightedItem(self)
  if (self.cardIndex == 2) then
    if ((self.chosenItem ~= nil) and self.chosenItem:hasPercentageOperation()) then
      return self.chosenItem
    else
      return self.cheapestItem
    end
  elseif (self.cardIndex == 3) then
    return self.cheapestItem
  else
    error("Unhandled value for 'cardIndex'")
  end
end

local function needsToHighlightItem(self)
  return (self.cardIndex > 1)
end

local function needsTipHeaderLabel(self)
  return (self.cardIndex > 1)
end

local function needsAdvanceButton(self)
  return ((self.cardIndex > 1) or (doesEitherItemHavePercentageOperation(self)))
end

local function getAdvanceButtonText(self)
  if (self.cardIndex == 1) then
    return LEARN_MORE_TEXT
  else
    return RETURN_TO_GAME_TEXT
  end
end

local function getAdvanceButtonY(self, advanceLabelHeight)
  local halfHeight = advanceLabelHeight * 0.5
  if (self.cardIndex == 1) then
    local bodyLabel = self.bodyLabel
    return bodyLabel:getYBottom() - halfHeight
  else
    return BOTTOM_MARGIN + halfHeight
  end
end

local function getAdvanceButtonCallback(self)
  if (self.cardIndex == 1) then
    return function()
      self.presentCardAtIndexCallback(2)
    end
  else
    return self.autoTapAndDismissCallback
  end
end

local function needsExample(self)
  return (self.cardIndex > 1)
end

local function needsArrowButtons(self)
  return needsExample(self)
end

local function getLeftArrowButtonDestinationIndex(self)
  if (self.cardIndex == 2) then
    return 1
  elseif (self.cardIndex == 3) then
    return 2
  else
    error("Unhandled value for 'cardIndex'")
  end
end

local function getRightArrowButtonDestinationIndex(self)
  if (self.cardIndex == 2) then
    if (doBothItemsHavePercentageOperation(self)) then
      return 3
    else
      return nil
    end
  elseif (self.cardIndex == 3) then
    return nil
  else
    error("Unhandled value for 'cardIndex'")
  end
end

local function getArrowButtonCallback(self, destinationIndex)
  if (destinationIndex ~= nil) then
    return function()
      self.presentCardAtIndexCallback(destinationIndex)
    end
  else
    return nil
  end
end

local function needsLargestBodyFontSize(self)
  return (self.cardIndex == 1)
end

local function getSummaryBodyTextForItem(self, item)
  local effect = item:getEffect()
  local operation = effect:getOperation()

  if (operation == GoodsEffect.OPERATION_NONE) then
    return ""
  end

  local iconName = item:getIconName()
  local basePriceText = GoodsCurrencyFormatter.convertDecimalToCurrency(item:getBasePrice())
  local effectText = effect:getPlainEffectString()
  local adjustedPriceText = GoodsCurrencyFormatter.convertDecimalToCurrency(item:getAdjustedPrice())
  return string.format("A <img src='%s' /> that costs %s with %s is %s.<br/><br/>", iconName, basePriceText, effectText, adjustedPriceText)
end

local function getCheapWordText(self)
  if (self.chosenItem ~= nil) then
    return "cheaper"
  else
    return "cheapest"
  end
end

local function getSummaryBodyText(self)
  local cheapWordText = getCheapWordText(self)
  local tapToContinueText = string.format("<b>Tap the %s item to continue.</b><br/><br/>", cheapWordText)

  local text = nil
  if (doBothItemsHavePercentageOperation(self)) then
    text = getSummaryBodyTextForItem(self, self.chosenItem) .. getSummaryBodyTextForItem(self, self.cheapestItem)
  elseif ((self.chosenItem ~= nil) and self.chosenItem:hasPercentageOperation()) then
    text = getSummaryBodyTextForItem(self, self.chosenItem)
  elseif (self.cheapestItem:hasPercentageOperation()) then
    text = getSummaryBodyTextForItem(self, self.cheapestItem)
  elseif (doesNeitherItemHavePercentageOperation(self)) then
    text = ""
  else
    error("Unhandled state")
  end

  text = text .. tapToContinueText
  return text
end

local function getTipTextForItem(self, item)
  local effect = item:getEffect()
  local percentageAsDecimal = effect:getAmount()
  local percentageAsInteger = MathUtils.round(percentageAsDecimal * 100)
  return GoodsTipTextBuilder.getTipTextForPercentage(percentageAsInteger)
end

local function getBodyText(self)
  if (self.cardIndex == 1) then
    return getSummaryBodyText(self)
  else
    local highlightedItem = getHighlightedItem(self)
    return getTipTextForItem(self, highlightedItem)
  end
end

local function playHorizontalSlide(self, startX, endX, isAppearing)
  self:setOpacity(1.0)
  self:setVisibleRecursive(true)
  self:setTouchEnabled(false)
  self:setY(self:getHeightPart(0.5))
  self:setX(startX)

  if (needsToHighlightItem(self)) then
    local highlightedItem = getHighlightedItem(self)
    highlightedItem:setHighlightPulsing(isAppearing)
  else
    if (not isAppearing) then
      self.cheapestItem:stopPickMe()
    end
  end

  return AHSequence.new(
    AHLocationXTo.new(self, 1.0, AHEaseExponentialOut.new(), endX),
    function()
      self:setTouchEnabled(isAppearing)

      if (isAppearing) then
        if (not needsToHighlightItem(self)) then
          self.cheapestItem:playPickMe()
        end
      else
        self:setVisibleRecursive(false)
      end
    end
  ):start()
end

local function buildTipHeaderLabel(self)
  local tipHeaderLabel = Label.new(Localizer.get("Misc.TIP"), TIP_HEADER_FONT)
  tipHeaderLabel:setX(HORIZONTAL_MARGIN + tipHeaderLabel:getWidthPart(0.5))
  tipHeaderLabel:setY(self:getHeight() - tipHeaderLabel:getHeightPart(0.5))
  self:insert(tipHeaderLabel)

  tipHeaderLabel:inheritOpacity(self)

  self.tipHeaderLabel = tipHeaderLabel
end

local function buildBodyLabel(self)
  local text = getBodyText(self)

  local width = self:getWidth() - (HORIZONTAL_MARGIN * 2.0)
  local useLargestFont = needsLargestBodyFontSize(self)
  local bodyLabel = GoodsRationaleTextArea.new(width, text, useLargestFont)

  local topY = nil
  if (self.tipHeaderLabel ~= nil) then
    local spacingBelowTipLabel = FontScaling.getScaledSize(10.0)
    topY = self.tipHeaderLabel:getYBottom() - spacingBelowTipLabel
  else
    local marginAboveBodyLabel = FontScaling.getScaledSize(28.0)
    topY = self:getHeight() - marginAboveBodyLabel
  end

  bodyLabel:setX(self:getWidthPart(0.5))
  bodyLabel:setY(topY - bodyLabel:getHeightPart(0.5))

  self:insert(bodyLabel)
  self.bodyLabel = bodyLabel

  bodyLabel:inheritOpacity(self)
end

local function getExampleFontSize(self)
  local bodyFontSize = self.bodyLabel:getBodyFontSize()
  return bodyFontSize + 2.0
end

local function buildExample(self)
  local highlightedItem = getHighlightedItem(self)
  local exampleFontSize = getExampleFontSize(self)
  local example = GoodsRationaleExample.new(highlightedItem, exampleFontSize)
  example:setX(self:getWidthPart(0.5))
  example:setY(EXAMPLE_CENTER_Y)
  self:insert(example)
  example:inheritOpacity(self)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GoodsRationaleCard:getCardIndex()
  return self.cardIndex
end

function GoodsRationaleCard:playSlideUp()
  self:setOpacity(0.0)
  self:setVisibleRecursive(true)
  self:setTouchEnabled(false)

  local slideUpAmount = Screen.getHeightPart(0.25)
  self:setY(self:getHeightPart(0.5) - slideUpAmount)

  local ease = AHEaseExponentialOut.new()

  return AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self, 1.0, ease, 1.0),
      AHLocationYTo.new(self, 1.0, ease, self:getHeightPart(0.5))
    ),
    function()
      self.cheapestItem:playPickMe()
      self:setTouchEnabled(true)
    end
  ):start()
end

function GoodsRationaleCard:prepareToDismiss()
  self:setTouchEnabled(false)

  if (needsToHighlightItem(self)) then
    local highlightedItem = getHighlightedItem(self)
    highlightedItem:hideHighlightImmediately()
  else
    self.cheapestItem:stopPickMe()
  end
end

function GoodsRationaleCard:createSlideDownAction()
  local slideDownAmount = Screen.getHeightPart(0.25)
  local ease = AHEaseExponentialOut.new()

  return AHSpawn.new(
    AHOpacityTo.new(self, 1.0, ease, 0.0),
    AHLocationYTo.new(self, 1.0, ease, self:getHeightPart(0.5) - slideDownAmount)
  )
end

function GoodsRationaleCard:playSlideLeft(isAppearing)
  local startX = nil
  local endX = nil
  if (isAppearing) then
    startX = X_OFFSCREEN_RIGHT
    endX = X_ONSCREEN_CENTER
  else
    startX = X_ONSCREEN_CENTER
    endX = X_OFFSCREEN_LEFT
  end

  return playHorizontalSlide(self, startX, endX, isAppearing)
end

function GoodsRationaleCard:playSlideRight(isAppearing)
  local startX = nil
  local endX = nil
  if (isAppearing) then
    startX = X_OFFSCREEN_LEFT
    endX = X_ONSCREEN_CENTER
  else
    startX = X_ONSCREEN_CENTER
    endX = X_OFFSCREEN_RIGHT
  end

  return playHorizontalSlide(self, startX, endX, isAppearing)
end

function GoodsRationaleCard:setTouchEnabled(enabled)
  for _, element in ipairs(self.touchableElements) do
    element:setTouchEnabled(enabled)
  end
end

function GoodsRationaleCard:hideImmediately()
  self:setOpacity(0.0)
  self:setVisibleRecursive(false)
  self:setTouchEnabled(false)
end

function GoodsRationaleCard:buildDeferrableTextElements()
  if (not self.hasBuiltDeferrableTextElements) then
    buildBodyLabel(self)

    if (needsExample(self)) then
      buildExample(self)
    end
  end

  self.hasBuiltDeferrableTextElements = true
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function getHeight(self)
  if (Screen.isTabletAspectRatio()) then
    return Screen.getSafeHeightPart(0.42)
  else
    return Screen.getSafeHeightPart(0.44)
  end
end

local function buildAdvanceButton(self)
  local text = getAdvanceButtonText(self)
  local callback = getAdvanceButtonCallback(self)
  local advanceButton = GoodsAdvanceButton.new(text, callback)

  local advanceLabelHeight = advanceButton:getLabelHeight()
  advanceButton:setX(self:getWidthPart(0.5))
  advanceButton:setY(getAdvanceButtonY(self, advanceLabelHeight))

  self:insert(advanceButton)
  advanceButton:inheritOpacity(self)

  table.insert(self.touchableElements, advanceButton)
  self.advanceButton = advanceButton
end

local function buildArrowButtons(self)
  local leftButtonDestinationIndex = getLeftArrowButtonDestinationIndex(self)
  local leftButtonCallback = getArrowButtonCallback(self, leftButtonDestinationIndex)
  local leftButton = GoodsArrowButton.new(leftButtonCallback)
  leftButton:setX(ARROW_BUTTON_SIDE_MARGIN + leftButton:getWidthPart(0.5))
  leftButton:setY(EXAMPLE_CENTER_Y)
  leftButton:setRot(180.0)
  self:insert(leftButton)
  if (leftButtonCallback ~= nil) then
    table.insert(self.touchableElements, leftButton)
  end

  local rightButtonDestinationIndex = getRightArrowButtonDestinationIndex(self)
  local rightButtonCallback = getArrowButtonCallback(self, rightButtonDestinationIndex)
  local rightButton = GoodsArrowButton.new(rightButtonCallback)
  rightButton:setX(self:getWidth() - ARROW_BUTTON_SIDE_MARGIN - rightButton:getWidthPart(0.5))
  rightButton:setY(EXAMPLE_CENTER_Y)
  self:insert(rightButton)
  if (rightButtonCallback ~= nil) then
    table.insert(self.touchableElements, rightButton)
  end
end

-- Constructor ------------------------

function GoodsRationaleCard.new(chosenItem, cheapestItem, cardIndex, presentCardAtIndexCallback, autoTapAndDismissCallback, useDeferredTextOptimization)
  local height = getHeight(self)
  local self = Group.new(WIDTH, height)
  Mixin.mix(self, GoodsRationaleCard)

  self.chosenItem = chosenItem
  self.cheapestItem = cheapestItem
  self.cardIndex = cardIndex
  self.presentCardAtIndexCallback = presentCardAtIndexCallback
  self.autoTapAndDismissCallback = autoTapAndDismissCallback
  self.touchableElements = {self.cheapestItem}
  self.hasBuiltDeferrableTextElements = false

  if (needsTipHeaderLabel(self)) then
    buildTipHeaderLabel(self)
  end

  if (not useDeferredTextOptimization) then
    self:buildDeferrableTextElements()
  end

  if (needsArrowButtons(self)) then
    buildArrowButtons(self)
  end

  if (needsAdvanceButton(self)) then
    buildAdvanceButton(self)
  end

  return self
end
