--- GoodsRound
-- @author Created by Aaron Barrett on May 29, 2014.

class("GoodsRound")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Icon Buckets -----------------------

local ICON_NAME_BUCKETS = {
  {
    maxCentralPrice = 6.00,
    iconNames = {
      "icon_coffee.png",
      "icon_bottle.png",
      "icon_donut.png",
      "icon_frozen_treat.png",
      "icon_muffin.png",
      "icon_shake.png",
      "icon_turkey_leg.png",
    }
  },
  {
    maxCentralPrice = 45.00,
    iconNames = {
      "icon_fancy_drink.png",
      "icon_headphones.png",
      "icon_lamp.png",
      "icon_martini.png",
      "icon_ramen.png",
      "icon_soccer_ball.png",
      "icon_video_game.png",
    }
  },
  {
    maxCentralPrice = 150.00,
    iconNames = {
      "icon_bike.png",
      "icon_duck.png",
      "icon_robot.png",
      "icon_rocket.png",
      "icon_shoe.png",
      "icon_snow_globe.png",
      "icon_water_tower.png",
    }
  },
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Icon Buckets -----------------------

local function getIconNameBucket(centralPrice)
  for _, iconNameBucket in ipairs(ICON_NAME_BUCKETS) do
    if (iconNameBucket.maxCentralPrice >= centralPrice) then
      return iconNameBucket
    end
  end

  return ICON_NAME_BUCKETS[#ICON_NAME_BUCKETS]
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function GoodsRound:getIdentifier()
  return self.identifier
end

function GoodsRound:getDifficulty()
  return self.difficulty
end

function GoodsRound:getRoundNumber()
  return self.roundNumber
end

function GoodsRound:getItems()
  return self.items
end

function GoodsRound:getItemCount()
  local items = self.items
  local itemCount = #items
  return itemCount
end

function GoodsRound:getCentralPrice()
  return self.centralPrice
end

function GoodsRound:getCountdownDuration()
  return self.countdownDuration
end

function GoodsRound:getDidMissAny()
  return self.didMissAny
end

function GoodsRound:setDidMissAny(didMissAny)
  self.didMissAny = didMissAny
end

function GoodsRound:getOrderedItems()
  return self.orderedItems
end

function GoodsRound:getUnfinishedItems()
  return self.unfinishedItems
end

function GoodsRound:getCheapestItem()
  local orderedItems = self.orderedItems
  local cheapestItemIndex = self.cheapestItemIndex
  local cheapestItem = orderedItems[cheapestItemIndex]
  return cheapestItem
end

function GoodsRound:getMostExpensiveItem()
  local orderedItems = self.orderedItems
  local mostExpensiveItem = orderedItems[#orderedItems]
  return mostExpensiveItem
end

function GoodsRound:isOnFinalItem()
  local cheapestItemIndex = self.cheapestItemIndex
  local itemCount = self:getItemCount()
  return (cheapestItemIndex > itemCount - 1)
end

-- Copy Items -------------------------

function GoodsRound:createShuffledItems()
  local items = self.items

  local itemsCopy = TableUtils.shallowCopy(items)
  local shuffledItems = TableUtils.shuffleArray(itemsCopy)

  return shuffledItems
end

-- Mark Items as Finished -------------

function GoodsRound:finishItem(chosenItem)
  self.unfinishedItems[chosenItem] = nil
  self.cheapestItemIndex = self.cheapestItemIndex + 1
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Build Tables -----------------------

local function buildOrderedItems(self)
  local items = self.items

  local orderedItems = TableUtils.shallowCopy(items)
  local function orderByAdjustedPrice(firstItem, secondItem)
    return (firstItem:getAdjustedPrice() < secondItem:getAdjustedPrice())
  end
  table.sort(orderedItems, orderByAdjustedPrice)

  self.orderedItems = orderedItems
end

local function buildUnfinishedItems(self)
  local items = self.items

  local unfinishedItems = TableUtils.createSetFromArray(items)
  self.unfinishedItems = unfinishedItems
end

-- Assign Icons -----------------------

local function assignIconsToItems(self)
  local centralPrice = self.centralPrice
  assert(type(centralPrice) == "number", "centralPrice is wrong.  Got: " .. tostring(centralPrice))

  local iconNameBucket = getIconNameBucket(centralPrice)
  local iconNames = iconNameBucket.iconNames

  local items = self.items
  local numberOfElementsToSample = #(items)
  local randomSample = TableUtils.createRandomSampleOfArray(iconNames, numberOfElementsToSample)

  for i, iconName in ipairs(randomSample) do
    local item = items[i]

    item:setIconName(iconName)
  end
end

-- Constructor ------------------------

function GoodsRound.new(identifier, difficulty, roundNumber, items, centralPrice, countdownDuration)
  local self = {}
  Mixin.mix(self, GoodsRound)

  self.identifier = identifier
  self.difficulty = difficulty
  self.roundNumber = roundNumber
  self.items = items
  self.centralPrice = centralPrice
  self.countdownDuration = countdownDuration

  self.didMissAny = false
  self.cheapestItemIndex = 1

  buildOrderedItems(self)
  buildUnfinishedItems(self)

  assignIconsToItems(self)

  return self
end
