--- GoodsValidator
-- @author Created by Aaron Barrett on May 21, 2014.
-- ConceptValidator for Goods.

class("GoodsValidator")

-- NOTE: Because GoodsEffect is being used for testing here and relies on Localizer,
-- which is currently not available for use in the validators, the following is needed.
-- We are only testing for American currency formats.
Localizer = {}
Localizer.get = function(someString)
  return "$%.2f"
end

require("validators.ConceptValidator")
require("GoodsEffect")
require("GoodsTipTextBuilder")
require("GoodsCurrencyFormatter")
require("GoodsExampleLinesBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "item_count",
  "central_price",

  "base_price_1",
  "operation_1",
  "op_amount_1",

  "base_price_2",
  "operation_2",
  "op_amount_2",

  "base_price_3",
  "operation_3",
  "op_amount_3",

  "base_price_4",
  "operation_4",
  "op_amount_4",

  "base_price_5",
  "operation_5",
  "op_amount_5",

  "base_price_6",
  "operation_6",
  "op_amount_6"
}

local MAX_ITEM_COUNT = 6

local OPERATIONS = {
  GoodsEffect.OPERATION_NONE,
  GoodsEffect.OPERATION_PERCENT_DISCOUNT,
  GoodsEffect.OPERATION_PERCENT_MARKUP,
  GoodsEffect.OPERATION_DOLLAR_DISCOUNT,
  GoodsEffect.OPERATION_DOLLAR_MARKUP
}

local MIN_DIFFERENCE_BETWEEN_PRICES = 0.01
local MAX_CENTRAL_PRICE = 500

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Helpers ----------------------------

local function getBasePriceFieldName(itemIndex)
  return "base_price_" .. itemIndex
end

local function getOperationFieldName(itemIndex)
  return "operation_" .. itemIndex
end

local function getAmountFieldName(itemIndex)
  return "op_amount_" .. itemIndex
end

local function isOperationPercentageBased(operation)
  return ((operation == GoodsEffect.OPERATION_PERCENT_DISCOUNT) or (operation == GoodsEffect.OPERATION_PERCENT_MARKUP))
end

local function validateAmount(self, operation, amount, itemIndex)
  local amountField = getAmountFieldName(itemIndex)

  if operation == GoodsEffect.OPERATION_NONE then
    self:expectEqual(amount, 0, amountField .. " should be zero because operation is none.")
  else
    self:expectGreaterThan(amount, 0, amountField .. " should be more than zero because operation is not none.")

    if (isOperationPercentageBased(operation)) then
      self:expectLessOrEqual(amount, 0.9, amountField .. " should be at most 0.9 for percentage operations.")
    end
  end
end

local function validatePercentageOperationIsSupported(self, operation, amount)
  if (isOperationPercentageBased(operation)) then
    local percentageAsInteger = MathUtils.round(amount * 100)

    local function ensureTipTextIsDefinedForPercentage()
      GoodsTipTextBuilder.getTipTextForPercentage(percentageAsInteger)
    end
    self:expectNoError(ensureTipTextIsDefinedForPercentage)

    local function ensureExampleLinesAreDefinedForPercentage()
      GoodsExampleLinesBuilder.getLines(amount, percentageAsInteger)
    end
    self:expectNoError(ensureExampleLinesAreDefinedForPercentage)
  end
end

local function validateItem(self, concept, itemIndex)
  local basePriceField = getBasePriceFieldName(itemIndex)
  local operationField = getOperationFieldName(itemIndex)
  local amountField = getAmountFieldName(itemIndex)

  local basePrice = concept[basePriceField]
  local operation = concept[operationField]
  local amount = concept[amountField]

  self:expectType(basePrice, "number", basePriceField)
  self:expectType(operation, "string", operationField)
  self:expectType(amount, "number", amountField)

  local isValidOperation = TableUtils.contains(OPERATIONS, operation)
  self:expectTrue(isValidOperation, string.format("Invalid value for %s: %s", operationField, operation))
  validateAmount(self, operation, amount, itemIndex)

  validatePercentageOperationIsSupported(self, operation, amount)
end

-- Validate methods -------------------

local function validateItems(self, concept)
  local itemCount = concept.item_count

  for itemIndex = 1, itemCount do
    validateItem(self, concept, itemIndex)
  end
end

local function validateItemCount(self, concept)
  local itemCount = concept.item_count
  self:expectType(itemCount, "number", "item_count")

  for itemIndex = 1, MAX_ITEM_COUNT do
    local basePriceField = getBasePriceFieldName(itemIndex)
    local operationField = getOperationFieldName(itemIndex)
    local amountField = getAmountFieldName(itemIndex)

    local basePrice = concept[basePriceField]
    local operation = concept[operationField]
    local amount = concept[amountField]

    if itemIndex <= itemCount then
      self:assertNotEmpty(basePrice, basePriceField)
      self:assertNotEmpty(operationField, operationField)
      self:assertNotEmpty(amount, amountField)
    else
      self:expectEmpty(basePrice, string.format("%s (item_count is %s)", basePriceField, itemCount))
      self:expectEmpty(operation, string.format("%s (item_count is %s)", operationField, itemCount))
      self:expectEmpty(amount, string.format("%s (item_count is %s)", amountField, itemCount))
    end
  end
end

local function validateCentralPrice(self, concept)
  local centralPrice = concept.central_price
  self:expectType(centralPrice, "number", "central_price")
  self:expectGreaterThan(centralPrice, 0, "central_price")
  self:expectLessOrEqual(centralPrice, MAX_CENTRAL_PRICE, "central_price")
end

local function validateMinimumDifferenceBetweenPrices(self, concept)
  local itemCount = concept.item_count
  local adjustedPrices = {}
  for itemIndex = 1, itemCount do
    local basePrice = concept[getBasePriceFieldName(itemIndex)]
    local operation = concept[getOperationFieldName(itemIndex)]
    local amount = concept[getAmountFieldName(itemIndex)]

    local effect = GoodsEffect.new(operation, amount)
    local adjustedPrice = effect:getAdjustedPriceForBasePrice(basePrice)
    table.insert(adjustedPrices, adjustedPrice)
  end

  for i = 1, itemCount do
    for j = i + 1, itemCount do
      local adjustedPriceI = adjustedPrices[i]
      local adjustedPriceJ = adjustedPrices[j]
      local difference = math.abs(adjustedPriceI - adjustedPriceJ)
      difference = MathUtils.round(difference, 6) -- disregard floating point precision errors

      local errorMessage = string.format("The difference between adjusted prices is too small for items %s ($%s) and %s ($%s)", i, adjustedPriceI, j, adjustedPriceJ)
      self:expectGreaterOrEqual(difference, MIN_DIFFERENCE_BETWEEN_PRICES, errorMessage)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function GoodsValidator:validateConcept(concept)
  local allowEmpty = true
  self:validateRequiredKeys(concept, REQUIRED_KEYS, allowEmpty)

  validateItemCount(self, concept)
  validateItems(self, concept)
  validateCentralPrice(self, concept)
  validateMinimumDifferenceBetweenPrices(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function GoodsValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, GoodsValidator)

  return self
end
