--- LinksCircleIntersector
-- @author Created by amorant on May 27, 2014.

class("LinksCircleIntersector")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

--- Calculate intersection points (0, 1 or 2) between the circumferences with radious ra and rb,
--  centered at A = (ax, ay) and B = (bx, by), respectively.
-- @param ax (number) - x coordinate of A, center of the first circle.
-- @param ay (number) - y coordinate of A, center of the first circle.
-- @param bx (number) - x coordinate of B, center of the second circle.
-- @param by (number) - y coordinate of B, center of the second circle.
-- @param ra (number) - Radius of the first circle.
-- @param rb (number) - Radius of the second circle.
-- @return (table) - Array of 0, 1 or 2 intersection points in the form {mx, my}.
function LinksCircleIntersector.getCircleIntersections(ax, ay, bx, by, ra, rb)
  -- The math to solve the problem is as follows:
  -- 1) The intersection points (mx, my) must verify the circumference equations for both circles:
  --    (mx - ax)^2 + (mx - ay)^2 = ra^2
  --    (mx - bx)^2 + (mx - by)^2 = rb^2
  -- 2) To simplify calculations, we translate all points by (-ax, -ay), so that our system is
  --    centered at A. (ax, ay) becomes (0, 0), and (bx, by) becomes (bx - ax, by - ay).
  --    By making these substitutions and expanding and combining the equations above, we obtain:
  --    2 * rb * bx * mx + 2 * rb * by * my = ra^2 - rb^2 + bx^2 + by^2
  -- 3) In that equation we can put mx in function of my, or vice versa. We will choose one or the
  --    other in order to avoid divisions by zero. Substituting the resulting expression in one of
  --    the original equations, we obtain a quadratic equation on mx or my.
  -- 4) We solve the quadratic equation for mx (or my), and for each obtained solution we compute
  --    the other coordinate using the equation in 2).


  if bx == ax and by == ay then
    -- If the end point is the same as the origin, we cannot solve the problem, because our
    -- formulae would encounter divisions by zero. In this case the problem has a solution if and
    -- only if ra == rb, and in that case it has infinitely many solutions: every point in either
    -- circle is a solution because the two circles are equal in position and size. But we cannot
    -- represent that, so we return zero solutions instead which just fits our use case.
    return {}
  end

  -- Translate points A, B to set reference system at A.
  -- This simplifies the math substantially.
  local originalA = {ax, ay}
  bx, by = bx - ax, by - ay
  ax, ay = 0, 0

  -- Some intermediate variables that will help writing formulae:
  local ra2 = ra * ra
  local rb2 = rb * rb
  local bx2 = bx * bx
  local by2 = by * by
  local rc2 = bx2 + by2
  local k = ra2 - rb2 + rc2

  -- Get quadratic equation to solve. To avoid divisions by zero, we choose to solve for mx or for my,
  -- then the other coordinate can be obtained without the zero division.
  local ea, eb, ec = nil, nil, nil
  local solvingForX = true
  if by ~= 0 then
    -- equation for mx:
    -- ea * mx^2 + eb * mx + ec = 0
    ea = 1 + bx2/by2
    eb = -k * bx / by2
    ec = k * k / (4 * by2) - ra2
  else
    -- equation for my:
    -- ea * my^2 + eb * my + ec = 0
    solvingForX = false
    ea = 1 + by2/bx2
    eb = -k * by / bx2
    ec = k * k / (4 * bx2) - ra2
  end

  local solutions = MathUtils.getQuadraticEquationSolutions(ea, eb, ec)
  local intersections = {}
  for _, solution in ipairs(solutions) do
    local mx, my = nil, nil
    if solvingForX then
      mx = solution
      my = (k - 2 * bx * mx) / (2 * by)
    else
      my = solution
      mx = (k - 2 * by * my) / (2 * bx)
    end

    -- Translate back to the original system.
    mx = mx + originalA[1]
    my = my + originalA[2]

    table.insert(intersections, {mx, my})
  end

  return intersections
end
