--- LinksGame
-- @author Created by amorant on May 07, 2014.

class("LinksGame")
require("LinksAudioTimer")
require("LinksBackground")
require("LinksDustParticles")
require("LinksGraphPattern")
require("LinksConfiguration")
require("LinksConceptGenerator")
require("LinksDragHelpOverlay")
require("LinksPanel")
require("LinksPanelScroller")
require("StandardUIBuilder")
require("utils.StateMachine")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_GRAPH_PATTERN = 20
local Z_DUST_PARTICLES = 30
local Z_FOG_SCROLLER = 40
local Z_LINKS_PANEL = 50
local Z_AUDIO_WARNING = 60
local Z_DRAG_HELP_OVERLAY = 70

-- Colors -----------------------------

local UI_COMPONENTS_COLOR = { 1.0, 1.0, 1.0, 0.5 }

-- States -----------------------------

local STATE_INITIAL = "initial"
local STATE_LISTENING = "listening"
local STATE_ANSWER_WAIT = "answer_wait"
local STATE_ANSWER_HANDLE = "answer_handle"
local STATE_TIMEOUT = "timeout"
local STATE_END = "end"

-- Difficulty -------------------------

local function configureGameDifficulty(self)
  local rankTwoScore = LinksConfiguration.getRankTwoScore()
  local rankThreeScore = LinksConfiguration.getRankThreeScore()
  local difficulty = GameSession.getDifficulty()
  local questionTimerDuration = LinksConfiguration.getQuestionTimerDuration(difficulty)

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)
  self.questionTimerDuration = questionTimerDuration

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Misc -------------------------------

local function getConnectionsDataForOutro(self)
  local connectionsData = {}
  for _, round in ipairs(self.rounds) do
    for _, set in ipairs(round.sets) do
      for _, link in ipairs(set.links) do
        for _, node in ipairs(set.nodes) do
          if link.node == node.text then
            table.insert(connectionsData, {link = link, node = node})
            break
          end
        end
      end
    end
  end
  return connectionsData
end

local function getTotalNumberOfAnswers(self)
  local numAnswers = #getConnectionsDataForOutro(self)
  return numAnswers
end


-- Game end condition -----------------

local function isGameWin(self)
  local hasLivesRemaining = (self.livesCounter:getLives() > 0)
  local isOutOfRounds = (self.remainingRounds <= 0)

  return (hasLivesRemaining and isOutOfRounds)
end

local function isGameLoss(self)
  return self.livesCounter:getLives() <= 0
end

local function isGameEnd(self)
  return isGameWin(self) or isGameLoss(self)
end

-- Round flow -------------------------

local function startSet(self)
  self.setIndex = self.setIndex + 1
  self.currentSet = self.currentRound.sets[self.setIndex]
  self.remainingAnswersInSet = self.currentSet.numConnections

  self.linksPanel:setSet(self.currentSet)
end

local function endSet(self)
  self.remainingSetsInRound = self.remainingSetsInRound - 1
  self.currentSet = nil

  if self.lastAnswerAnimation then
    wait(self.lastAnswerAnimation)
  end
end

local function startRound(self)
  self.roundIndex = self.roundIndex + 1
  self.setIndex = 0
  self.currentRound = self.rounds[self.roundIndex]
  self.remainingSetsInRound = #self.currentRound.sets
end

local function endRound(self)
  self.remainingRounds = self.remainingRounds - 1
  self.currentRound = nil

  self.linksPanel:clearRound()
end

local function isSetComplete(self)
  return self.remainingAnswersInSet == 0
end

local function isRoundComplete(self)
  return self.remainingSetsInRound == 0
end

-- Game flow --------------------------

local function afterAnswerTransition(self, answerWasCorrect)
  if isSetComplete(self) then
    self.totalRemainingSeconds = self.totalRemainingSeconds + self.questionTimer:getTimeRemaining()
    self.questionTimer:stop()
    endSet(self)
  end

  if isRoundComplete(self) then
    if answerWasCorrect then
      Sound.playSoundEffect("set_complete")
    end
    endRound(self)
  end

  if isGameEnd(self) then
    self.machine:transition(STATE_END)
  elseif isRoundComplete(self) then
    self.machine:transition(STATE_LISTENING)
  else
    self.machine:transition(STATE_ANSWER_WAIT)
  end
end

-- States -----------------------------

local function onListeningEnter(self)
  startRound(self)

  local concept = self.concept
  local clip = self.currentRound.audioClip
  local clipDuration = self.currentRound.audioClipDuration
  self.audioTimer:playAudioAndWait(concept, clip, clipDuration)

  self.machine:transition(STATE_ANSWER_WAIT)
end

local function onAnswerWaitEnter(self)
  if not self.currentSet then
    self.questionTimer:setTime(0)
    startSet(self)
    self.dragHelpOverlay:showAfterInactivityDelay()
    self.questionTimer:start()
  end
end

local function onAnswerHandleEnter(self, node, link)
  local isCorrect = link:isCorrectNode(node)
  local conceptId = self.concept.identifier
  local questionIdentifier = link:getQuestionIdentifier()
  local answer = node:getText()
  local animation = nil

  if isCorrect then
    GameSession.contentTracker:notifyCorrectAnswer(conceptId, questionIdentifier, answer)

    link:setLinkedNode(node)
    self.graphPattern:createIncrementProgressAction():start()
    animation = AHSpawn.new(
      link:createCorrectAnimation(),
      node:createCorrectAnimation(),
      function() Sound.playSoundEffect("correct") end
    )
  else
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, questionIdentifier, answer)
    self.livesCounter:decrement()
    local correctNode = self.linksPanel:getCorrectNode(link)

    local incorrectAnimation = AHSpawn.new(
      node:createIncorrectAnimation(),
      link:createIncorrectAnimation(),
      function() Sound.playSoundEffect("incorrect") end
    )
    wait(incorrectAnimation:start())
    animation = AHSpawn.new(
      correctNode:createAutoCorrectAnimation(link),
      link:createTimeoutAnimation()
    )
  end

  GameSession.increaseScore(LinksConfiguration.POINTS_PER_CORRECT_RESPONSE)
  local isLastResponse = (self.roundIndex == #self.rounds) and (self.remainingAnswersInSet == 1)

  if isLastResponse then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), LinksConfiguration.getPointsPerLifeBonus(), self.bonusTray)

    local secondsForTimeBonus = LinksConfiguration.getSecondsForTimeBonus(self.totalRemainingSeconds)
    ScoringUtils.applyTimeBonus(secondsForTimeBonus, LinksConfiguration.POINTS_PER_EXTRA_SECOND_REMAINING, self.bonusTray)
  end

  self.scoreLabel:setScore(GameSession.getScore())
  self.remainingAnswersInSet = self.remainingAnswersInSet - 1

  self.lastAnswerAnimation = animation
  animation:start()
  afterAnswerTransition(self, isCorrect)
end

local function onTimeoutEnter(self)
  local linkTimeoutCallback = function(link)
    GameSession.contentTracker:notifyTimeout(self.concept.identifier, link:getQuestionIdentifier())
    self.livesCounter:decrement()
    self.remainingAnswersInSet = self.remainingAnswersInSet - 1
  end

  self.linksPanel:waitForAnimations()

  local animation = self.linksPanel:createRevealAnswersAnimation(linkTimeoutCallback)
  Sound.playSoundEffect("incorrect")
  wait(animation:start())

  assert(self.remainingAnswersInSet == 0, "There should be zero remaining answers after timeout")
  GameSession.scoredIncorrectAnswer("some concept ID")

  local answerWasCorrect = false
  afterAnswerTransition(self, answerWasCorrect)
end

local function onEndEnter(self)
  if isGameWin(self) then
    GameSession.notifyCorrectAnswer(self.concept.identifier)

    local connectionsData = getConnectionsDataForOutro(self)
    local animation = self.linksPanel:createOutroAnimation(connectionsData)
    wait(animation:start())

    waitFor(2)

    GameSession.notifyGameWin()
  else
    GameSession.notifyIncorrectAnswer(self.concept.identifier)
    GameSession.notifyGameLoss()
  end
end

-- Listeners --------------------------

local function answerListener(self)
  return function(node, link)
    if link then
      self.dragHelpOverlay:hide()
      self.dragHelpOverlay:disable()
      self.machine:transition(STATE_ANSWER_HANDLE, node, link)
    else
      self.dragHelpOverlay:show()
    end
  end
end

local function timeoutListener(self)
  return function()
    self.dragHelpOverlay:hide()
    self.dragHelpOverlay:disable()
    self.machine:transition(STATE_TIMEOUT)
  end
end

-- Misc -------------------------------

local function initializeGame(self)
  local concept = self.conceptGenerator:next()
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
  self.concept = concept
  self.rounds = concept.rounds
  self.remainingRounds = #self.rounds
  self.roundIndex = 0
  self.setIndex = 0
  self.currentRound = nil
  self.currentSet = nil
  self.remainingAnswersInSet = 0
  self.remainingSetsInRound = 0

  self.totalRemainingSeconds = 0
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinksGame:gameFlow()
  self.fogScroller:startAnimation()
  self.dustParticles:start()

  wait(self.rootNode:createShowAnimation():start())
  self.graphPattern:createGrowAction():start()

  self.machine:transition(STATE_LISTENING)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createStateMachine(self)
  local machine = StateMachine.new()

  machine:addState(
    STATE_INITIAL,
    nil,
    nil,
    STATE_LISTENING
  )
  machine:addState(
    STATE_LISTENING,
    function() onListeningEnter(self) end,
    nil,
    STATE_ANSWER_WAIT
  )
  machine:addState(
    STATE_ANSWER_WAIT,
    function() onAnswerWaitEnter(self) end,
    nil,
    STATE_ANSWER_HANDLE, STATE_TIMEOUT
  )
  machine:addState(
    STATE_ANSWER_HANDLE,
    function(node, link) onAnswerHandleEnter(self, node, link) end,
    nil,
    STATE_ANSWER_WAIT, STATE_TIMEOUT, STATE_LISTENING, STATE_END
  )
  machine:addState(
    STATE_TIMEOUT,
    function() onTimeoutEnter(self) end,
    nil,
    STATE_LISTENING, STATE_END
  )
  machine:addState(
    STATE_END,
    function() onEndEnter(self) end,
    nil
  )

  machine:transition(STATE_INITIAL)

  self.machine = machine
end

local function createGameElements(self)
  local linksPanel = LinksPanel.new(self.concept.rootNode, self.remainingRounds)
  linksPanel:setNodeDropListener(answerListener(self))
  self:insertAtCenter(linksPanel, Z_LINKS_PANEL)
  self.linksPanel = linksPanel
  self.rootNode = linksPanel:getRootNode()

  local background = LinksBackground.new()
  self:insertAtCenter(background, Z_BACKGROUND)
  self.background = background

  local fogScroller = LinksPanelScroller.new()
  self:insertAtCenter(fogScroller, Z_FOG_SCROLLER)
  self.fogScroller = fogScroller

  local patternPoints = LinkLayouts.getPatternPoints()
  local patternEdges = LinkLayouts.getPatternEdges()
  local stepCount = getTotalNumberOfAnswers(self)
  local graphPattern = LinksGraphPattern.new(patternPoints, patternEdges, stepCount)
  self:insertAtCenter(graphPattern, Z_GRAPH_PATTERN)
  self.graphPattern = graphPattern

  local questionTimer = linksPanel:getQuestionTimer()
  questionTimer:setTimeoutHandler(timeoutListener(self))
  questionTimer:setDuration(self.questionTimerDuration)
  self.questionTimer = questionTimer

  local dustParticles = LinksDustParticles.new()
  self:insertAtCenter(dustParticles, Z_DUST_PARTICLES)
  self.dustParticles = dustParticles

  local dragHelpOverlay = LinksDragHelpOverlay.new()
  self:insertAtCenter(dragHelpOverlay, Z_DRAG_HELP_OVERLAY)
  self.dragHelpOverlay = dragHelpOverlay
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.livesCounter = builder:addLivesDisplay(LinksConfiguration.getInitialLivesCount())
  self.bonusTray = builder:addBonusTray()

  local audioTimer = LinksAudioTimer.new(UI_COMPONENTS_COLOR)
  audioTimer:setAnchorPointRight()
  local xOffset = 22
  local yOffset = - (self.livesCounter:getHeight() + 10)
  builder:addWidget(audioTimer, UIOverlay.NORTHEAST, xOffset, yOffset)
  self.audioTimer = audioTimer
end

-- Constructor ------------------------

function LinksGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, LinksGame)

  self.conceptGenerator = LinksConceptGenerator.new()
  createStateMachine(self)
  initializeGame(self)
  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
