--- LinksGraphPattern
-- @author Created by amorant on May 27, 2014.

class("LinksGraphPattern")

require("LinksCollapsibleLine")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local DOT_RADIUS = 5
local EDGE_LINE_WIDTH = 4

-- Z-Indices --------------------------

local Z_LINES = 10
local Z_DOTS = 20

-- Colors -----------------------------

local DOT_COLOR = ColorUtils.colorRGBToStd(66, 133, 141)
local EDGE_COLOR = ColorUtils.colorRGBToStd(51, 111, 121)
local LIT_UP_DOT_COLOR = ColorUtils.colorRGBToStd(49, 221, 233)
local LIT_UP_EDGE_COLOR = ColorUtils.colorRGBToStd(35, 201, 213)
local LIGHTUP_BRIGHTEST_COLOR = ColorUtils.colorRGBToStd(74, 233, 244)

-- Animation --------------------------

local ADD_DOT_DURATION = 0.05
local GROW_LINE_DURATION = 0.05
local REMOVE_DOT_DURATION = 0.05
local SHRINK_LINE_DURATION = 0.05

local LIGHTUP_RIPPLE_SPEED = 1.5 -- self heights per second
local LIGHTUP_UP_DURATION = 0.25
local LIGHTUP_DOWN_DURATION = 0.25

-- Misc -------------------------------

local EDGE_ID_MULTIPLIER = 1000

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Misc -------------------------------

local function getPointScaledToSelfDimensions(self, point)
  local x, y = unpack(point)
  x = x * self:getWidth()
  y = y * self:getHeight()
  return x, y
end

local function getPointsAdjustedToRadius(x1, y1, x2, y2, radius)
  local edgeLength = MathUtils.distance(x1, y1, x2, y2)
  local adjustedX1 = MathUtils.mapLinear(radius/edgeLength, 0, 1, x1, x2)
  local adjustedY1 = MathUtils.mapLinear(radius/edgeLength, 0, 1, y1, y2)
  local adjustedX2 = MathUtils.mapLinear(1 - radius/edgeLength, 0, 1, x1, x2)
  local adjustedY2 = MathUtils.mapLinear(1 - radius/edgeLength, 0, 1, y1, y2)

  adjustedX1 = math.floor(adjustedX1)
  adjustedY1 = math.floor(adjustedY1)
  adjustedX2 = math.floor(adjustedX2)
  adjustedY2 = math.floor(adjustedY2)

  return adjustedX1, adjustedY1, adjustedX2, adjustedY2
end

-- Graph traversal --------------------

local function getEdgeId(originIndex, destinationIndex)
  return originIndex * EDGE_ID_MULTIPLIER + destinationIndex
end

local function getEdgeOrigin(edgeId)
  return math.floor(edgeId / EDGE_ID_MULTIPLIER)
end

local function getEdgeDestination(edgeId)
  return edgeId % EDGE_ID_MULTIPLIER
end

local function getReversePatternEdges(patternEdges)
  local reversePatternEdges = {}

  for originIndex, destinationIndices in pairs(patternEdges) do
    for destinationIndex in pairs(destinationIndices) do
      -- Note that this assumes a tree-like structure, with one parent per node.
      reversePatternEdges[destinationIndex] = originIndex
    end
  end

  return reversePatternEdges
end

local function getRootIndices(self)
  local rootIndices = {}

  for pointIndex, _ in ipairs(self.patternPoints) do
    if not self.reversePatternEdges[pointIndex] then
      table.insert(rootIndices, pointIndex)
    end
  end

  return rootIndices
end

local function getContiguousLitOffPoints(self, startingPointIndex, numPoints)
  local selectedPointIndices = { startingPointIndex }

  local function isLitOffAndUnselected(pointIndex)
    return not self.litUpDots[pointIndex] and not TableUtils.contains(selectedPointIndices, pointIndex)
  end

  for k, currentPointIndex in ipairs(selectedPointIndices) do
    local parentIndex = self.reversePatternEdges[currentPointIndex]
    local childIndicesSet = self.patternEdges[currentPointIndex] or {}

    local neighbourIndices = TableUtils.createArrayFromSet(childIndicesSet)
    table.insert(neighbourIndices, parentIndex)
    neighbourIndices = TableUtils.getFilteredArray(neighbourIndices, isLitOffAndUnselected)

    local extraNeighbours = #selectedPointIndices + #neighbourIndices - numPoints
    if extraNeighbours > 0 then
      TableUtils.removeFromEndOfArray(neighbourIndices, extraNeighbours)
    end

    TableUtils.concatenateArrays(selectedPointIndices, neighbourIndices)

    if #selectedPointIndices >= numPoints then
      break
    end
  end

  assert(#selectedPointIndices <= numPoints, "It is okay to return less than the required number points, but not more!")

  return selectedPointIndices
end

-- Animations - Grow/Shrink -----------

local function createAddPointAction(self, pointIndex, duration)
  if self.points[pointIndex] then
    return AHDoNothing.new()
  end

  local point = self.patternPoints[pointIndex]
  local x, y = getPointScaledToSelfDimensions(self, point)

  x = math.floor(x)
  y = math.floor(y)

  local dot = Sprite.newWithName("pattern_circle.png")
  dot:setEffectiveDims(2 * DOT_RADIUS)
  dot:setColor(unpack(DOT_COLOR))
  dot:setLoc(x, y)
  dot:setLoc(x, y)
  local initialScale = dot:getScl()
  dot:setScl(0)

  self.points[pointIndex] = dot

  return AHSequence.new(
    function() self:insert(dot, Z_DOTS) end,
    AHScaleTo.new(dot, duration/2, AHEaseSineOut.new(), 1.2 * initialScale),
    AHScaleTo.new(dot, duration/2, AHEaseSineOut.new(), 1 * initialScale)
  )
end

local function createRemovePointAction(self, pointIndex, duration)
  local dot = self.points[pointIndex]

  if not dot then
    return AHDoNothing.new()
  end

  self.points[pointIndex] = nil

  return AHSequence.new(
    AHScaleTo.new(dot, duration, AHEaseSineOut.new(), 0),
    function() dot:removeSelf() end
  )
end

local function createAddEdgesAction(self)
  local addEdgesAction = AHSequence.new()
  local rootIndices = getRootIndices(self)

  for _, rootIndex in ipairs(rootIndices) do
    local addOriginAction = createAddPointAction(self, rootIndex, ADD_DOT_DURATION)
    addEdgesAction:addAction(addOriginAction)
  end

  local pointIndicesQueue = rootIndices
  local pointIndicesByLevel = {}

  while #pointIndicesQueue > 0 do
    local nextLevelPointIndices = {}
    local currentLevelAction = AHSpawn.new()

    for _, originIndex in ipairs(pointIndicesQueue) do
      local destinationIndices = self.patternEdges[originIndex] or {}

      for destinationIndex in pairs(destinationIndices) do
        local originPoint = self.patternPoints[originIndex]
        local destinationPoint = self.patternPoints[destinationIndex]

        local x1, y1 = getPointScaledToSelfDimensions(self, originPoint)
        local x2, y2 = getPointScaledToSelfDimensions(self, destinationPoint)
        x1, y1, x2, y2 = getPointsAdjustedToRadius(x1, y1, x2, y2, DOT_RADIUS)

        local line = LinksCollapsibleLine.new(x1, y1, x2, y2)
        line:setAnchorPointBottomLeft()
        line:setLineWidth(EDGE_LINE_WIDTH)
        line:setColor(unpack(EDGE_COLOR))
        line:setLengthScale(0)
        self:insert(line, Z_LINES)

        local edgeId = getEdgeId(originIndex, destinationIndex)
        self.edges[edgeId] = line

        local growLineAction = line:createGrowAnimation(GROW_LINE_DURATION)
        local addDestinationAction = createAddPointAction(self, destinationIndex, ADD_DOT_DURATION)
        local edgeSequence = AHSequence.new(growLineAction, addDestinationAction)
        currentLevelAction:addAction(edgeSequence)

        table.insert(nextLevelPointIndices, destinationIndex)
      end
    end

    addEdgesAction:addAction(currentLevelAction)
    table.insert(pointIndicesByLevel, pointIndicesQueue)
    pointIndicesQueue = nextLevelPointIndices
  end

  self.pointIndicesByLevel = pointIndicesByLevel

  return addEdgesAction
end

local function createRemoveEdgesAction(self)
  local removeEdgesAction = AHSequence.new()

  local reversedPointIndicesByLevel = TableUtils.getReversedArray(self.pointIndicesByLevel)

  for _, pointIndicesQueue in ipairs(reversedPointIndicesByLevel) do
    local currentLevelAction = AHSpawn.new()

    for _, destinationIndex in ipairs(pointIndicesQueue) do
      local originIndex = self.reversePatternEdges[destinationIndex]

      if originIndex then
        local edgeId = getEdgeId(originIndex, destinationIndex)
        local line = self.edges[edgeId]

        self.edges[edgeId] = nil

        local removeDestinationAction = createRemovePointAction(self, destinationIndex, REMOVE_DOT_DURATION)
        local shrinkLineAction = line:createShrinkAnimation(SHRINK_LINE_DURATION)
        currentLevelAction:addAction(removeDestinationAction)
        currentLevelAction:addAction(shrinkLineAction)
      end
    end

    removeEdgesAction:addAction(currentLevelAction)
  end

  local removeOriginAction = createRemovePointAction(self, 1, REMOVE_DOT_DURATION)
  removeEdgesAction:addAction(removeOriginAction)

  return removeEdgesAction
end

-- Animations - Light up --------------

local function getLightUpRippleDelayForPosition(self, x, y)
  local originX = self:getWidth()/2
  local originY = self:getHeight()/2
  local distanceToOrigin = MathUtils.distance(x, y, originX, originY)
  local rippleSpeed = self:getHeight() * LIGHTUP_RIPPLE_SPEED
  local delayTime = distanceToOrigin / rippleSpeed
  return delayTime
end

local function createDotColorToAction(dot, duration, ease, targetColor)
  return AHColorTo.new(dot, duration, ease, targetColor)
end

local function getRandomLitOffPointIndex(self)
  local allPointIndices = TableUtils.getKeys(self.patternPoints)
  local allPointIndicesSet = TableUtils.createSetFromArray(allPointIndices)
  local litOffPointIndicesSet = TableUtils.differenceSet(allPointIndicesSet, self.litUpDots)
  local litOffPointIndices = TableUtils.createArrayFromSet(litOffPointIndicesSet)

  return litOffPointIndices[math.random(#litOffPointIndices)]
end

local function shouldLightUpEdge(self, edgeId, pointsToLightUpSet)
  local originIndex = getEdgeOrigin(edgeId)
  local destinationIndex = getEdgeDestination(edgeId)
  local isOriginLitUp = self.litUpDots[originIndex] or pointsToLightUpSet[originIndex]
  local isDestinationLitUp = self.litUpDots[destinationIndex] or pointsToLightUpSet[destinationIndex]
  local oneOfThePointsIsNew = pointsToLightUpSet[originIndex] or pointsToLightUpSet[destinationIndex]

  return isOriginLitUp and isDestinationLitUp and oneOfThePointsIsNew
end

local function getEdgesToLightUpFromPoints(self, pointsToLightUpSet)
  local allEdgeIds = TableUtils.getKeys(self.edges)
  local edgesToLightUp = {}
  for _, edgeId in ipairs(allEdgeIds) do
    edgesToLightUp[edgeId] = shouldLightUpEdge(self, edgeId, pointsToLightUpSet)
  end

  return edgesToLightUp
end

local function createLightUpElementAction(target, delayTime, endColor, timeAtBrightestColor)
  return AHSequence.new(
    AHDelay.new(delayTime),
    AHColorTo.new(target, LIGHTUP_UP_DURATION, AHEaseSineInOut.new(), LIGHTUP_BRIGHTEST_COLOR),
    AHDelay.new(timeAtBrightestColor),
    AHColorTo.new(target, LIGHTUP_DOWN_DURATION, AHEaseSineInOut.new(), endColor)
  )
end

local function createLightUpPointsAction(self, targetNumLitPoints)
  assert(targetNumLitPoints <= #self.points, "Error: targetNumLitPoints beyond the number of points.")

  -- Select points to light up
  local pointIndicesToLightUp = {}
  local numPointsToLightUp = targetNumLitPoints - TableUtils.getSize(self.litUpDots)
  while #pointIndicesToLightUp < numPointsToLightUp do
    -- Start with a random point
    local pointIndex = getRandomLitOffPointIndex(self)

    -- Get a cluster as big as needed around that point, repeat until enough points obtained
    local remainingPointsToFind = numPointsToLightUp - #pointIndicesToLightUp
    local pointsCluster = getContiguousLitOffPoints(self, pointIndex, remainingPointsToFind)
    TableUtils.concatenateArrays(pointIndicesToLightUp, pointsCluster)
    pointIndicesToLightUp = TableUtils.getUniquesArray(pointIndicesToLightUp)
  end

  local pointIndicesToLightUpSet = TableUtils.createSetFromArray(pointIndicesToLightUp)
  local edgeIdsToLightUpSet = getEdgesToLightUpFromPoints(self, pointIndicesToLightUpSet)

  -- Create animation
  local lightUpAnimation = AHSpawn.new()

  for edgeId, line in pairs(self.edges) do
    local originDot = self.points[getEdgeOrigin(edgeId)]
    local delayTime = getLightUpRippleDelayForPosition(self, originDot:getLoc())
    local endColor = EDGE_COLOR
    local timeAtBrightestColor = 0
    if edgeIdsToLightUpSet[edgeId] or self.litUpEdges[edgeId] then
      endColor = LIT_UP_EDGE_COLOR
      self.litUpEdges[edgeId] = true
      timeAtBrightestColor = 0.3
    end
    local thisLineAnimation = createLightUpElementAction(line, delayTime, endColor, timeAtBrightestColor)
    lightUpAnimation:addAction(thisLineAnimation)
    coroutine.yield()
  end

  for pointIndex, dot in ipairs(self.points) do
    local delayTime = getLightUpRippleDelayForPosition(self, dot:getLoc())
    local endColor = DOT_COLOR
    local timeAtBrightestColor = 0
    if pointIndicesToLightUpSet[pointIndex] or self.litUpDots[pointIndex] then
      endColor = LIT_UP_DOT_COLOR
      self.litUpDots[pointIndex] = true
      timeAtBrightestColor = 0.3
    end
    local thisDotAnimation = createLightUpElementAction(dot, delayTime, endColor, timeAtBrightestColor)
    lightUpAnimation:addAction(thisDotAnimation)
    coroutine.yield()
  end

  return lightUpAnimation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinksGraphPattern:createGrowAction()
  return createAddEdgesAction(self)
end

function LinksGraphPattern:createShrinkAction()
  return createRemoveEdgesAction(self)
end

function LinksGraphPattern:createIncrementProgressAction()
  waitUntil(function() return #self.cachedIncrementActions >= self.currentAction end)
  local action = self.cachedIncrementActions[self.currentAction]
  self.currentAction = self.currentAction + 1
  return action
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function prepareIncrementProgressAction(self)
  self:incrementProgress()

  local progress = self:getProgress()
  local targetNumLitPoints = math.ceil(progress * #self.points)

  return createLightUpPointsAction(self, targetNumLitPoints)
end

local function prepareIncrementProgressActions(self)
  self.cachedIncrementActions = {}
  self.currentAction = 1

  runInCoroutine(function()
    waitUntil(function() return #self.points == #self.patternPoints end)
    for i = 1, self:getStepCount() do
      local action = prepareIncrementProgressAction(self)
      table.insert(self.cachedIncrementActions, action)
      coroutine.yield()
    end
  end)
end

-- Constructor ------------------------

function LinksGraphPattern.new(patternPoints, patternEdges, stepCount)
  local self = Group.newFullScreen()
  ProgressIndicator.makeProgressIndicator(self, stepCount)
  Mixin.mix(self, LinksGraphPattern)

  self.patternPoints = patternPoints
  self.patternEdges = patternEdges
  self.reversePatternEdges = getReversePatternEdges(patternEdges)
  self.points = {} -- pointIndex to dot (Sprite)
  self.edges = {} -- edgeId to line (LinksCollapsibleLine)
  self.litUpDots = {} -- pointIndex to bool/nil
  self.litUpEdges = {} -- edgeId to bool/nil

  prepareIncrementProgressActions(self)

  return self
end
