--- LinksLink
-- @author Created by Anton Morant on May 12, 2014.

class("LinksLink")

require("LinksCircleIntersector")
require("LinksRectangleIntersector")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local DROP_AREA_WIDTH = 150
local DROP_AREA_HEIGHT = 150

local LABEL_WIDTH = 0.33 * Screen.width
local LABEL_WIDTH_SMALL = 0.7 * LABEL_WIDTH
local LABEL_HEIGHT = 0.2 * Screen.height
local LABEL_MARGIN_X = 8
local LABEL_MARGIN_Y = 8

local LINE_WIDTH = 2

local BALL_RADIUS = 10
local BALL_OUTLINE_WIDTH = 3

-- Text -------------------------------

local FONT = FontFaces.DIN_OT_LIGHT
local FONT_SIZE = 14
local SMALL_FONT_SIZE = 10

-- Z-Indices --------------------------

local Z_LINE = 10
local Z_LABEL_BACKGROUND = 20
local Z_LABEL = 30
local Z_CORRECT_RING = 40
local Z_BALL = 50

-- Colors -----------------------------

local COLOR_LINE = { 1, 1, 1 }

local COLOR_INITIAL = ColorUtils.colorRGBToStd(110, 236, 216)
local COLOR_HOVER_1 = ColorUtils.colorRGBToStd(71, 81, 237)
local COLOR_HOVER_2 = ColorUtils.colorRGBToStd(237, 101, 71)
local COLOR_HOVER_3 = ColorUtils.colorRGBToStd(237, 71, 189)
local COLOR_CORRECT_1 = ColorUtils.colorRGBToStd(237, 71, 233)
local COLOR_CORRECT_2 = ColorUtils.colorRGBToStd(214, 237, 71)
local COLOR_INCORRECT = ColorUtils.colorRGBToStd(236, 71, 109)

-- Animations -------------------------

local MIN_ENLARGED_BALL_RADIUS = 0.03 * MathUtils.vectorLength(Screen.width, Screen.height)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Element location update ------------

local function getLineEndAdjustedToLabel(self, lineStartX, lineStartY, lineEndX, lineEndY)
  local labelX = lineEndX - self.textWidth/2
  local labelY = lineEndY - self.textHeight/2
  local labelWidth, labelHeight = self.textWidth, self.textHeight

  if self.labelIsVisible then
    local lineIsTooSmall = LinksRectangleIntersector.doesRectangleContainPoint(labelX, labelY, labelWidth, labelHeight, lineStartX, lineStartY)
    if not lineIsTooSmall then
      local intersections = LinksRectangleIntersector.getRectangleAndSegmentIntersections(labelX, labelY, labelWidth, labelHeight, lineStartX, lineStartY, lineEndX, lineEndY)
      if #intersections > 0 then
        local chosenIntersection = intersections[1]
        local minDistance = math.huge
        for _, intersection in ipairs(intersections) do
          local distance = MathUtils.distance(intersection[1], intersection[2], lineStartX, lineStartY)
          if distance < minDistance then
            chosenIntersection = intersection
            minDistance = distance
          end
        end
        lineEndX, lineEndY = unpack(chosenIntersection)
      end
    end
  end

  return lineEndX, lineEndY
end

local function drawLines(self, endX, endY, startX, startY, jointX, jointY)
  local labelX = jointX - self.textWidth/2
  local labelY = jointY - self.textHeight/2
  local labelWidth, labelHeight = self.textWidth, self.textHeight
  local lineAStartX, lineAStartY, lineAEndX, lineAEndY = endX, endY, jointX, jointY
  local lineBStartX, lineBStartY, lineBEndX, lineBEndY = startX, startY, jointX, jointY

  lineAEndX, lineAEndY = getLineEndAdjustedToLabel(self, lineAStartX, lineAStartY, lineAEndX, lineAEndY)
  lineBEndX, lineBEndY = getLineEndAdjustedToLabel(self, lineBStartX, lineBStartY, lineBEndX, lineBEndY)

  self.lineA:setVertexPosition(1, lineAStartX, lineAStartY)
  self.lineA:setVertexPosition(2, lineAEndX, lineAEndY)
  self.lineB:setVertexPosition(1, lineBStartX, lineBStartY)
  self.lineB:setVertexPosition(2, lineBEndX, lineBEndY)
end

local function updateLineLocation(self)
  local currentX, currentY = self:getLoc()
  local distanceToStartX = self.layout.startX - currentX
  local distanceToStartY = self.layout.startY - currentY

  -- The following points are expressed taking the link as the origin of coordinates:
  -- endX, endY: link end position (is 0,0 to match link object position)
  -- startX, startY: link start position, where attached to root node
  -- jointX, jointY: joint position
  -- smallSegmentLength: segment closer to the link (small segment)
  -- bigSegmentLength: segment closer to the root node (big segment)
  local endX, endY = 0, 0
  local startX, startY = distanceToStartX, distanceToStartY
  local jointX, jointY = nil, nil
  local smallSegmentLength = self.layout.smallSegmentLength
  local bigSegmentLength = self.layout.bigSegmentLength

  local linkCurrentLength = MathUtils.distance(startX, startY, endX, endY)

  -- If the link is over-stretched, show a straight line.
  if linkCurrentLength > smallSegmentLength + bigSegmentLength then
    local q = smallSegmentLength/(smallSegmentLength + bigSegmentLength)
    jointX, jointY = q * (startX-endX), q * (startY-endY)

  -- If the link is at zero length, position everything at zero
  elseif linkCurrentLength == 0 then
    jointX, jointY = 0, 0

  -- If the link is under-stretched, do magic
  elseif linkCurrentLength < math.abs(bigSegmentLength - smallSegmentLength) then
    assert(smallSegmentLength < bigSegmentLength, "smallSegmentLength is expected to be less than bigSegmentLength, at least while in the danger zone.")

    -- This is the danger zone. The link's current length is not enough to build it with
    -- smallSegmentLength and bigSegmentLength full lengths, so we are going to scale them. But we
    -- need to do so in a continuous way, so that there is no jump after leaving the danger zone
    -- and using proper intersections.
    local underStretchRatio = linkCurrentLength / (bigSegmentLength - smallSegmentLength)
    bigSegmentLength = math.floor(bigSegmentLength * underStretchRatio)
    smallSegmentLength = bigSegmentLength - linkCurrentLength
    jointX = math.floor(startX * (1 - bigSegmentLength / linkCurrentLength))
    jointY = math.floor(startY * (1 - bigSegmentLength / linkCurrentLength))

  -- If the link is stretched within bounds, calculate intersection points.
  else
    local intersections = LinksCircleIntersector.getCircleIntersections(endX, endY, startX, startY, smallSegmentLength, bigSegmentLength)

    if #intersections == 1 then
      jointX, jointY = unpack(intersections[1])
    elseif #intersections == 2 then
      local m1x, m1y = unpack(intersections[1])
      local m2x, m2y = unpack(intersections[2])
      if (m1x < m2x and self.layout.side == "left") or (m1x > m2x and self.layout.side == "right") then
        jointX, jointY = m1x, m1y
      else
        jointX, jointY = m2x, m2y
      end
    else
      error("Something went really wrong with link geometry calculations " .. StringUtils.tostring(self.layout))
    end
  end

  local labelX = self:getWidth()/2 + jointX
  local labelY = self:getHeight()/2 + jointY
  self.label:setLoc(labelX, labelY)

  drawLines(self, endX, endY, startX, startY, jointX, jointY)
end

local function updateLinkedNodeLocation(self)
  if self.linkedNode then
    local nodeX, nodeY = self.linkedNode:getLoc()
    local distanceToNodeX, distanceToNodeY = self:getWorldDistance(self.linkedNode)
    self.linkedNode:setLoc(nodeX - distanceToNodeX, nodeY - distanceToNodeY)
  end
end

local function updateElementsLocation(self)
  updateLineLocation(self)
  updateLinkedNodeLocation(self)
end

-- Animations -------------------------

local function makeSynchronizedColorAnimation(self, colorAnimation)
  return AHCallLogic.new(function()
    if self.colorAnimation and self.colorAnimation:isActive() then
      self.colorAnimation:stop()
    end
    self.colorAnimation = colorAnimation
    wait(colorAnimation:start())
  end)
end

local function createColorChangeAction(self, color, duration)
  local colorAnimation = AHColorTo.new(self.ball.filling, duration, AHEaseSineInOut.new(), color)
  return makeSynchronizedColorAnimation(self, colorAnimation)
end

local function createHoverColorAction(self)
  local duration = 0.5
  local colorAnimation = AHLoop.new(
    AHSequence.new(
      AHColorTo.new(self.ball.filling, duration, nil, COLOR_HOVER_1),
      AHColorTo.new(self.ball.filling, duration, nil, COLOR_HOVER_2),
      AHColorTo.new(self.ball.filling, duration, nil, COLOR_HOVER_3),
      AHColorTo.new(self.ball.filling, duration, nil, COLOR_INITIAL)
    )
  )
  return makeSynchronizedColorAnimation(self, colorAnimation)
end

local function createLocationToAction(self, duration, ease, targetX, targetY)
  return AHSpawn.new(
    AHLocationTo.new(self, duration, ease, targetX, targetY),
    AHCustomAction.new(duration, ease, function(time) updateElementsLocation(self) end)
  )
end

local function createSmallSegmentLengthToAction(self, duration, ease, targetLength)
  -- Warning:
  -- It is a requirement in the layout that smallSegmentLength < bigSegmentLength, otherwise our
  -- current animation will break when in the "danger zone" section.
  -- If this action makes smallSegmentLength > bigSegmentLength, then this function must be called
  -- again to revert the size of smallSegmentLength before allowing a suck in animation to happen
  -- or the game will crash.

  local function setSmallSegmentLength(link, amount)
    link.layout.smallSegmentLength = amount
    updateLineLocation(self)
  end

  local function getSmallSegmentLength(link)
    return link.layout.smallSegmentLength
  end

  return AHNonAttributeTo.new(self, duration, getSmallSegmentLength, setSmallSegmentLength, AHEaseSineInOut.new(), targetLength)
end

local function getTargetBallRadius(self)
  assert(self.linkedNode, "No linked node")
  local outline = self.ball.outline
  local filling = self.ball.filling
  local linkedNodeTextRadius = self.linkedNode:getTextRadius()
  local linkedNodeBoxRadius = self.linkedNode:getBoxRadius()
  local targetRadius = MathUtils.interpolateLinearFromNormalized(linkedNodeTextRadius/linkedNodeBoxRadius, MIN_ENLARGED_BALL_RADIUS, linkedNodeBoxRadius + 2)
  return targetRadius
end

local function createBallEnlargeAnimation(self, duration)
  assert(self.linkedNode, "No linked node")
  local outline = self.ball.outline
  local filling = self.ball.filling

  local outlineRadius = outline:getWidth() / 2
  local fillingRadius = filling:getWidth() / 2
  local targetRadius = getTargetBallRadius(self)
  local targetScaleOutline = targetRadius / outlineRadius
  local targetScaleFilling = (targetRadius - BALL_OUTLINE_WIDTH) / fillingRadius

  local targetSmallSegmentLength = self.layout.smallSegmentLength + targetRadius - outlineRadius

  local bumpFactor = 1.1
  return AHSequence.new(
    AHSpawn.new(
      createSmallSegmentLengthToAction(self, duration/2, AHEaseSineInOut.new(), targetSmallSegmentLength * bumpFactor),
      AHScaleTo.new(outline, duration/2, AHEaseSineInOut.new(), targetScaleOutline * bumpFactor),
      AHScaleTo.new(filling, duration/2, AHEaseSineInOut.new(), targetScaleFilling * bumpFactor)
    ),
    AHSpawn.new(
      createSmallSegmentLengthToAction(self, duration/2, AHEaseSineInOut.new(), targetSmallSegmentLength),
      AHScaleTo.new(outline, duration/2, AHEaseSineInOut.new(), targetScaleOutline),
      AHScaleTo.new(filling, duration/2, AHEaseSineInOut.new(), targetScaleFilling)
    )
  )
end

local function createBallShrinkAnimation(self, duration)
  assert(self.linkedNode, "No linked node")
  local outline = self.ball.outline
  local filling = self.ball.filling
  local targetScaleOutline = 1
  local targetScaleFilling = 1
  local targetSmallSegmentLength = self.layout.originalRa
  local scaleDown = true

  return AHSpawn.new(
    createSmallSegmentLengthToAction(self, duration, AHEaseSineIn.new(), targetSmallSegmentLength),
    self.linkedNode:createFadeOutAction(duration, scaleDown),
    AHScaleTo.new(outline, duration, AHEaseSineIn.new(), targetScaleOutline),
    AHScaleTo.new(filling, duration, AHEaseSineIn.new(), targetScaleFilling)
  )
end

local function createBallFadeAnimation(self, duration, initialOpacity, targetOpacity)
  self.ball.outline:setOpacity(initialOpacity)
  self.ball.filling:setOpacity(initialOpacity)
  return AHSpawn.new(
    AHOpacityTo.new(self.ball.outline, duration, nil, targetOpacity),
    AHOpacityTo.new(self.ball.filling, duration, nil, targetOpacity)
  )
end

local function createLabelFadeAnimation(self, duration, initialOpacity, targetOpacity)
  self.label:setOpacity(initialOpacity)
  return AHOpacityTo.new(self.label, duration, nil, targetOpacity)
end

local function createLineFadeAnimation(self, duration, initialOpacity, targetOpacity)
  self.lineA:setOpacity(initialOpacity)
  self.lineB:setOpacity(initialOpacity)
  return AHSpawn.new(
    AHOpacityTo.new(self.lineA, duration, nil, targetOpacity),
    AHOpacityTo.new(self.lineB, duration, nil, targetOpacity)
  )
end

local function createShootRingAnimation(self, delayTime, duration, lingerDuration, finalScale)
  local ring = Sprite.newWithName("ring.png")
  local ringRadius = ring:getWidth()/2
  finalScale = finalScale * getTargetBallRadius(self) / ringRadius
  ring:setScl(0)

  local firstColor = TableUtils.shallowCopy(COLOR_CORRECT_1)
  local secondColor = TableUtils.shallowCopy(COLOR_CORRECT_2)
  local thirdColor = TableUtils.shallowCopy(COLOR_INITIAL)
  local fourthColor = TableUtils.shallowCopy(COLOR_INITIAL)

  firstColor[4] = 0.75
  secondColor[4] = 0.5
  thirdColor[4] = 0.25
  fourthColor[4] = 0

  return AHSpawn.new(
    AHSequence.new(
      function() self:insertAtCenter(ring, Z_CORRECT_RING) end,
      AHDelay.new(delayTime),
      AHScaleTo.new(ring, duration - delayTime, AHEaseSineOut.new(), finalScale),
      function() waitFor(lingerDuration) ring:removeSelf() end
    ),
    AHSequence.new(
      AHColorTo.new(ring, duration/3, AHEaseSineInOut.new(), firstColor),
      AHColorTo.new(ring, duration/3, AHEaseSineInOut.new(), secondColor),
      AHColorTo.new(ring, duration/3, AHEaseSineInOut.new(), thirdColor),
      AHColorTo.new(ring, lingerDuration, AHEaseSineInOut.new(), fourthColor)
    )
  )
end

local function createThreeRingsAnimation(self, duration, lingerDuration)
  local firstRingDuration = duration
  local ringDelay = duration/4

  return AHSpawn.new(
    createShootRingAnimation(self, 0, duration, lingerDuration, 1.75),
    createShootRingAnimation(self, ringDelay, duration, 0.75 * lingerDuration, 1.5),
    createShootRingAnimation(self, 2 * ringDelay, duration, 0.5 * lingerDuration, 1.25)
  )
end

local function createRefuseNodeAnimation(self, duration)
  assert(self.hoveringNode, "No hovering node")
  local distanceToNodeX, distanceToNodeY = self:getWorldDistance(self.hoveringNode)
  local distanceToNode = MathUtils.vectorLength(distanceToNodeX, distanceToNodeY)
  local moveLength = 0.10 * Screen.width
  local currentX, currentY = self:getLoc()
  local targetX = currentX - distanceToNodeX  * moveLength / distanceToNode
  local targetY = currentY - distanceToNodeY * moveLength / distanceToNode

  return AHSequence.new(
    createLocationToAction(self, 1/8 * duration, AHEaseSineInOut.new(), targetX, targetY),
    createLocationToAction(self, 7/8 * duration, AHEaseSineInOut.new(), currentX, currentY)
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- DragAndDropHandler protocol --------

function LinksLink:canAcceptDNDObject(object)
  return self.canAcceptDrops
end

function LinksLink:notifyDNDEnter(object)
  createHoverColorAction(self):start()
  self.hoveringNode = object
end

function LinksLink:notifyDNDLeave(object)
  createColorChangeAction(self, COLOR_INITIAL, 0.1):start()
  self.hoveringNode = nil
end

function LinksLink:notifyDNDDrop(object)
  -- Empty implementation
  -- (LinksPanel already notified by LinksNode)
end

function LinksLink:notifyDNDMove(object)
  -- Empty implementation
end

-- Animations -------------------------

function LinksLink:createCorrectAnimation()
  if self.dndHandler then
    self.dndHandler:removeContainer(self)
  end

  local ringsDuration = 0.4
  local lingerDuration = 0.4
  local enlargeDuration = 0.4

  local ringsAction = createThreeRingsAnimation(self, ringsDuration, lingerDuration)

  self.currentAnimation = AHSpawn.new(
    AHSequence.new(
      createColorChangeAction(self, COLOR_CORRECT_1, ringsDuration/3),
      createColorChangeAction(self, COLOR_CORRECT_2, ringsDuration/3),
      createColorChangeAction(self, COLOR_INITIAL, ringsDuration/3)
    ),
    createBallEnlargeAnimation(self, enlargeDuration),
    ringsAction
  )
  return self.currentAnimation
end

function LinksLink:createIncorrectAnimation()
  local refuseDuration = 0.5
  local resetDuration = 0.1

  self.currentAnimation = AHSequence.new(
    function() self.canAcceptDrops = false end,
    AHSpawn.new(
      AHSequence.new(
        createColorChangeAction(self, COLOR_INCORRECT, 0.01),
        AHDelay.new(refuseDuration),
        createColorChangeAction(self, COLOR_INITIAL, resetDuration)
      ),
      createRefuseNodeAnimation(self, refuseDuration)
    ),
    function() self.canAcceptDrops = true end
  )
  return self.currentAnimation
end

function LinksLink:createTimeoutAnimation()
  self.dndHandler:removeContainer(self)
  local enlargeDuration = 0.4
  local colorDuration = 0.05
  local resetColorDuration = 0.1

  self.currentAnimation = AHSpawn.new(
    AHSequence.new(
      AHDelay.new(0.2),
      createColorChangeAction(self, COLOR_INCORRECT, colorDuration),
      createColorChangeAction(self, COLOR_INITIAL, resetColorDuration),
      createColorChangeAction(self, COLOR_INCORRECT, colorDuration),
      createColorChangeAction(self, COLOR_INITIAL, resetColorDuration),
      createColorChangeAction(self, COLOR_INCORRECT, colorDuration),
      createColorChangeAction(self, COLOR_INITIAL, resetColorDuration)
    ),
    createBallEnlargeAnimation(self, enlargeDuration),
    function() self.canAcceptDrops = false end
  )
  return self.currentAnimation
end

function LinksLink:createSpringAction()
  local startX, startY = self.layout.startX, self.layout.startY
  local endX, endY = self.layout.endX, self.layout.endY
  local extraBumpX = (endX - startX) * 0.035
  local extraBumpY = (endY - startY) * 0.035
  local duration = 0.75

  self:setLoc(startX, startY)

  self.currentAnimation = AHSequence.new(
    AHSpawn.new(
      AHSequence.new(
        createLocationToAction(self, 0.6 * duration, AHEaseSimpleIn.new(), endX, endY),
        createLocationToAction(self, 0.1 * duration, AHEaseSineOut.new(), endX + extraBumpX, endY + extraBumpY),
        createLocationToAction(self, 0.15 * duration, AHEaseSineInOut.new(), endX - 0.5 * extraBumpX, endY - 0.5 * extraBumpY),
        createLocationToAction(self, 0.15 * duration, AHEaseSineInOut.new(), endX, endY)
      ),
      createBallFadeAnimation(self, 0.2 * duration, 0, 1),
      createLabelFadeAnimation(self, 0.2 * duration, 0, 1)
    ),
    function() self.canAcceptDrops = true end
  )
  return self.currentAnimation
end

function LinksLink:createSuckInAction()
  local startX, startY = self.layout.startX, self.layout.startY
  local endX, endY = self.layout.endX, self.layout.endY
  local extraBumpX = (endX - startX) * 0.2
  local extraBumpY = (endY - startY) * 0.2
  local duration = 0.6
  self.currentAnimation = AHSequence.new(
    function() self.canAcceptDrops = false end,
    createLocationToAction(self, 0.3 * duration, AHEaseSineInOut.new(), endX + extraBumpX, endY + extraBumpY),
    AHSpawn.new(
      createLocationToAction(self, duration, AHEaseSineInOut.new(), startX, startY),
      createLabelFadeAnimation(self, 0.7 * duration, 1, 0),
      AHSequence.new(
        createBallShrinkAnimation(self, 0.25 * duration),
        createBallFadeAnimation(self, 0.4 * duration, 1, 0)
      ),
      AHSequence.new(
        AHDelay.new(0.7 * duration),
        createLineFadeAnimation(self, 0.1 * duration, 1, 0)
      )
    )
  )
  return self.currentAnimation
end

-- State ------------------------------

function LinksLink:setLoc(x, y)
  self.super.setLoc(self, x, y)
  updateElementsLocation(self)
end

function LinksLink:isLinked()
  return self.linkedNode ~= nil
end

function LinksLink:setLinkedNode(node)
  assert(self.linkedNode == nil, "Link is already linked")
  assert(node.link == nil, "LinksNode is already linked")
  self.linkedNode = node
  node.link = self

  self.parent:takeChild(node)
end

function LinksLink:setLabelVisible(visible)
  self.labelIsVisible = visible
  self.label:setVisible(visible)
end

function LinksLink:isAnimating()
  return self.currentAnimation and self.currentAnimation:isActive()
end

function LinksLink:getQuestionIdentifier()
  return self.data.questionIdentifier
end

function LinksLink:isCorrectNode(node)
  return (node:getCorrectLink() == self.data.text)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function initLayoutData(self, originalLayout, panelWidth, panelHeight, rootNode)
  local layout = TableUtils.shallowCopy(originalLayout)
  -- Note: It is a requirement that rootNode and this Link share the same parent for positioning
  -- to work.
  local rootX, rootY = rootNode:getLoc()
  local rootHeight = rootNode:getHeight()

  layout.endX = math.floor(layout.relativeEndX * panelWidth)
  layout.endY = math.floor(layout.relativeEndY * panelHeight)

  layout.startX = math.floor(layout.relativeStartX * panelWidth )
  if layout.relativeEndY < 0.5 then
    layout.startY = math.floor(rootY - rootHeight/2)
  else
    layout.startY = math.floor(rootY + rootHeight/2)
  end

  local standardLength = MathUtils.distance(layout.startX, layout.startY, layout.endX, layout.endY)
  layout.smallSegmentLength = math.floor(layout.relativeSmallSegmentLength * standardLength)
  layout.bigSegmentLength = math.floor(layout.relativeBigSegmentLength * standardLength)
  layout.originalRa = layout.smallSegmentLength

  assert(layout.relativeSmallSegmentLength < layout.relativeBigSegmentLength,
    "relativeSmallSegmentLength is expected to be less than relativeBigSegmentLength, please check the layouts.")
  assert(layout.smallSegmentLength < layout.bigSegmentLength,
    "smallSegmentLength is expected to be less than bigSegmentLength, please check the layouts or the math.")

  self.layout = layout
end

local function createElements(self, text, labelWidth, fontSize)
  -- Note: All elements are be positioned in updateElementsLocation, called by setLoc.
  --       The link is initialized in a "compressed" state, and createSpringAction moves it to its
  --       normal position using the layout data initialized in initLayoutData.

  local lineA = Line.new(0, 0, 1, 1)
  lineA:setColor(unpack(COLOR_LINE))
  lineA:setLineWidth(LINE_WIDTH)
  self:insertAtCenter(lineA, Z_LINE)
  local lineB = Line.new(0, 0, 1, 1)
  lineB:setColor(unpack(COLOR_LINE))
  lineB:setLineWidth(LINE_WIDTH)
  self:insertAtCenter(lineB, Z_LINE)

  local label = TextArea.new(text, Font.new(FONT, fontSize), labelWidth, Alignment.CENTER, nil, Alignment.CENTER)
  label:setColor(unpack(COLOR_LINE))
  self:insert(label, Z_LABEL)

  self.textWidth = label:getWidth() + LABEL_MARGIN_X
  self.textHeight = label:getHeight() + LABEL_MARGIN_Y

  local ball = Group.new(2 * BALL_RADIUS, 2 * BALL_RADIUS)
  local ballOutline = FilledCircle.new(BALL_RADIUS, COLOR_LINE)
  local ballFilling = FilledCircle.new(BALL_RADIUS - BALL_OUTLINE_WIDTH, COLOR_INITIAL)
  ball.outline = ballOutline
  ball.filling = ballFilling
  ball:insertAtCenter(ballOutline)
  ball:insertAtCenter(ballFilling)
  self:insertAtCenter(ball, Z_BALL)
  ball:setVisible(false)

  self.lineA = lineA
  self.lineB = lineB
  self.label = label
  self.labelIsVisible = true
  self.ball = ball
end

local function newLink(dndHandler, rootNode, linkData, layout, panelWidth, panelHeight, labelWidth, fontSize)
  local self = Group.new(DROP_AREA_WIDTH, DROP_AREA_HEIGHT)
  Mixin.mix(self, LinksLink)

  self.dndHandler = dndHandler
  self.data = linkData
  self.linkedNode = nil
  self.canAcceptDrops = false
  self.incorrectMatches = {}

  initLayoutData(self, layout, panelWidth, panelHeight, rootNode)
  createElements(self, linkData.text, labelWidth, fontSize)

  return self
end

-- Constructor ------------------------

function LinksLink.new(dndHandler, rootNode, linkData, layout, panelWidth, panelHeight)
  return newLink(dndHandler, rootNode, linkData, layout, panelWidth, panelHeight, LABEL_WIDTH, FONT_SIZE)
end

function LinksLink.newForOutro(rootNode, linkData, layout, panelWidth, panelHeight)
  return newLink(nil, rootNode, linkData, layout, panelWidth, panelHeight, LABEL_WIDTH_SMALL, SMALL_FONT_SIZE)
end
