--- LinksNode
-- @author Created by Anton Morant on May 12, 2014.

class("LinksNode")

require("utils.GlyphTableHelpers")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_LABEL = 20

-- Colors -----------------------------

local COLOR_BACKGROUND = { 0, 0, 0, 0.3 }
local OPACITY_MIN = 0
local OPACITY_MAX = COLOR_BACKGROUND[4]
local COLOR_TEXT_LINKED = { 0, 0, 0 }

-- Text -------------------------------

local FONT = Font.new(FontFaces.DIN_OT_LIGHT, 14)
local SMALL_FONT = Font.new(FontFaces.DIN_OT_LIGHT, 12)
local LABEL_MARGIN = 6

-- Animation parameters ---------------

local DURATION_MOVE_CORRECT = 0.2
local DURATION_MOVE_TIMEOUT = 0.6
local DURATION_BOUNCE = 0.2
local SPEED_MOVE_INCORRECT = Screen.height / 0.5

local DISTANCE_FOR_MIN_OPACITY = 0.2 * Screen.height
local OPACITY_LINK_ATTR_ID = 0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Animations -------------------------

local function createMoveToOriginAction(self)
  local currentX, currentY = self:getLoc()
  local originX, originY = self.originX, self.originY
  local distanceToOrigin = MathUtils.distance(currentX, currentY, originX, originY)
  local duration = distanceToOrigin / SPEED_MOVE_INCORRECT

  return AHSequence.new(
    AHLocationTo.new(self, duration, AHEaseSineInOut.new(), originX, originY),
    function() self.picked = false end
  )
end

local function createMoveToLinkAction(self, link, duration)
  local x, y = self:getWorldDistance(link)
  return AHLocationBy.new(self, duration, AHEaseSineInOut.new(), x, y)
end

local function createScaleBounceAction(self)
  local duration = DURATION_BOUNCE/2
  return AHSequence.new(
    AHScaleTo.new(self, duration, AHEaseSineInOut.new(), 0.3),
    AHScaleTo.new(self, duration, AHEaseSineInOut.new(), 1)
  )
end

-- Opacity link -----------------------

local function updateBackgroundOpacity(self)
  local opacity = nil

  if self:isLinked() then
    opacity = OPACITY_MIN
  elseif not self:isPicked() then
    opacity = OPACITY_MAX
  else
    local _, currentY = self:getLoc()
    local normalizedDistance = (currentY - self.originY) / DISTANCE_FOR_MIN_OPACITY
    normalizedDistance = MathUtils.clamp(normalizedDistance, 0, 1)
    opacity = MathUtils.interpolateLinearFromNormalized(normalizedDistance, OPACITY_MAX, OPACITY_MIN)
  end

  self.background:setOpacity(opacity)
end

local function initializeOpacityLink(self)
  local scriptNode = MOAIScriptNode.new()
  scriptNode:reserveAttrs(1)
  scriptNode:setCallback(function(node)
    updateBackgroundOpacity(self)
  end)
  scriptNode:setAttrLink(OPACITY_LINK_ATTR_ID, self, MOAITransform2D.ATTR_Y_LOC)
  self.scriptNode = scriptNode
end

local function stopOpacityLink(self)
  self.scriptNode:clearAttrLink(OPACITY_LINK_ATTR_ID)
  -- Must wait a frame for the link to stop having effect
  coroutine.yield()
end

-- Misc -------------------------------

local function getTextRadius(label)
  local minX = Screen.width
  local maxX = 0

  for _, view in ipairs(label:getItemViews()) do
    local left = view:getXLeft()
    local right = view:getXRight()
    minX = math.min(left, minX)
    maxX = math.max(right, maxX)
  end

  local textWidth = maxX - minX
  local textHeight = label:getHeight()
  local textRadius = MathUtils.vectorLength(textWidth, textHeight) / 2

  return textRadius
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- DragAndDropHandler protocol --------

function LinksNode:notifyDNDPickUp()
  assert(self.link == nil, "Breaking links is not supported, this should not have happened")
  self.picked = true
end

function LinksNode:notifyDNDDrop(container)
  runInCoroutine(function()
    if self.dropListener then
      self.dropListener(self, container)
    end
    if not container then
      self.currentAnimation = createMoveToOriginAction(self)
      self.currentAnimation:start()
    end
  end)
end

-- Animations -------------------------

function LinksNode:createDropAnimation()
  self.currentAnimation = AHSpawn.new(
    createScaleBounceAction(self),
    AHColorTo.new(self.label, DURATION_BOUNCE, AHEaseSineInOut.new(), COLOR_TEXT_LINKED)
  )
  return self.currentAnimation
end

function LinksNode:createCorrectAnimation()
  assert(self.link, "Need to connect to link with Link:setLinkedNode first")
  self.dndHandler:removeObject(self)
  self.currentAnimation = AHSequence.new(
    AHSpawn.new(
      createMoveToLinkAction(self, self.link, DURATION_MOVE_CORRECT),
      AHOpacityTo.new(self.background, DURATION_MOVE_CORRECT, nil, 0)
    ),
    self:createDropAnimation()
  )
  return self.currentAnimation
end

function LinksNode:createIncorrectAnimation()
  self.currentAnimation = createMoveToOriginAction(self)
  return self.currentAnimation
end

function LinksNode:createAutoCorrectAnimation(link)
  self.dndHandler:removeObject(self)
  link:setLinkedNode(self)
  self.currentAnimation = AHSequence.new(
    AHSpawn.new(
      createMoveToLinkAction(self, link, DURATION_MOVE_TIMEOUT),
      AHOpacityTo.new(self.background, DURATION_MOVE_TIMEOUT, nil, OPACITY_MIN)
    ),
    AHColorTo.new(self.label, DURATION_BOUNCE, AHEaseSineInOut.new(), COLOR_TEXT_LINKED)
  )
  return self.currentAnimation
end

function LinksNode:createTimeoutRevealAnimation(link)
  self.dndHandler:removeObject(self)
  link:setLinkedNode(self)
  self.currentAnimation = AHSpawn.new(
    createMoveToLinkAction(self, link, DURATION_MOVE_TIMEOUT),
    AHOpacityTo.new(self.background, DURATION_MOVE_TIMEOUT, nil, OPACITY_MIN)
  )
  return self.currentAnimation
end

function LinksNode:createTimeoutDiscardAnimation()
  self.dndHandler:removeObject(self)
  if self:isPicked() then
    self.currentAnimation = createMoveToOriginAction(self)
  else
    self.currentAnimation = AHDoNothing.new()
  end

  return self.currentAnimation
end

function LinksNode:createFadeOutAction(duration, scaleDown)
  stopOpacityLink(self)

  local animation = AHSpawn.new(
    AHOpacityTo.new(self.label, duration, nil, 0),
    AHOpacityTo.new(self.background, duration, nil, 0)
  )

  if scaleDown then
    animation:addAction(AHScaleTo.new(self, duration, nil, 0.5))
  end

  self.currentAnimation = animation
  return animation
end

-- Events -----------------------------

function LinksNode:setDropListener(listener)
  self.dropListener = listener
end

-- State ------------------------------

function LinksNode:isLinked()
  return self.link ~= nil
end

function LinksNode:isPicked()
  return self.picked
end

function LinksNode:getTextRadius()
  return self.textRadius
end

function LinksNode:getBoxRadius()
  return self.boxRadius
end

function LinksNode:isAnimating()
  return self.currentAnimation and self.currentAnimation:isActive()
end

function LinksNode:useCurrentLocationAsOrigin()
  self.originX, self.originY = self:getLoc()
end

function LinksNode:getText()
  return self.data.text
end

function LinksNode:getCorrectLink()
  return self.data.link
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(text, maxWidth, height, font)
  local labelMaxWidth = maxWidth - LABEL_MARGIN
  local label = DynamicTextArea.newSimple(text, labelMaxWidth, font, Alignment.CENTER, Alignment.CENTER)
  local backgroundWidth = maxWidth
  local backgroundHeight = height
  local background = FilledRectangle.new(backgroundWidth, backgroundHeight, COLOR_BACKGROUND)

  local self = Group.new(backgroundWidth, backgroundHeight)
  self:insertAtCenter(background, Z_BACKGROUND)
  self:insertAtCenter(label, Z_LABEL)

  self.background = background
  self.label = label
  self.boxRadius = MathUtils.vectorLength(self:getDims())/2
  self.textRadius = getTextRadius(label)

  return self
end

local function newNode(dndHandler, nodeData, maxWidth, height, maxFontSize)
  local self = createLayout(nodeData.text, maxWidth, height, maxFontSize)
  Mixin.mix(self, LinksNode)

  self.dndHandler = dndHandler
  self.data = nodeData
  self.picked = false

  initializeOpacityLink(self)

  return self
end

-- Constructor ------------------------

function LinksNode.new(dndHandler, nodeData, maxWidth, height)
  return newNode(dndHandler, nodeData, maxWidth, height, FONT)
end

function LinksNode.newForOutro(nodeData, maxWidth, height)
  return newNode(nil, nodeData, maxWidth, height, SMALL_FONT)
end
