--- LinksPanel
-- @author Created by Anton Morant on May 09, 2014.

class("LinksPanel")
require("LinksLink")
require("LinkLayouts")
require("LinksNode")
require("LinksQuestionTimer")
require("LinksRootNode")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local GRAPH_PANEL_WIDTH = Screen.width
local GRAPH_PANEL_HEIGHT = 0.9 * Screen.height
local NODES_PANEL_WIDTH = Screen.width
local NODES_PANEL_HEIGHT = 0.15 * Screen.height
local NODE_HEIGHT = 0.6 * NODES_PANEL_HEIGHT
local ROOT_NODE_WIDTH = GRAPH_PANEL_WIDTH
local ROOT_NODE_HEIGHT = 0.13 * Screen.height

local NODES_ROW_SPACING = 5
local ROOT_FONT_SIZE = 55

-- Z-Indices --------------------------

local Z_GRAPH_PANEL = 10
local Z_NODES_PANEL = 20

local Z_QUESTION_TIMER = 10
local Z_NODES_ROW = 20

local Z_GRAPH_LINKS = 10
local Z_GRAPH_NODES = 20
local Z_GRAPH_ROOT = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Clear helpers ----------------------

function clearNodes(self, ignoreLinked)
  local ignoredNodes = {}

  for key, node in pairs(self.nodes) do
    if node:isLinked() and ignoreLinked then
      ignoredNodes[key] = node
    else
      node:removeSelf()
      self.dndHandler:removeObject(node)
    end
  end

  self.nodes = ignoredNodes
end

function clearLinks(self, ignoreLinked)
  local ignoredLinks = {}

  for key, link in pairs(self.links) do
    if link:isLinked() and ignoreLinked then
      ignoredLinks[key] = link
    else
      link:removeSelf()
      self.dndHandler:removeContainer(link)
    end
  end

  self.links = ignoredLinks
end

local function resetLinkLayouts(self)
  local linkLayouts = LinkLayouts.getLinksLayouts()
  TableUtils.shuffleArray(linkLayouts)
  self.linkLayouts = linkLayouts
end

-- LinksNode dimensions helpers ------------

local function getNodeDimsForSet(nodesInSet)
  local width = math.floor(NODES_PANEL_WIDTH / nodesInSet - 1.5 * NODES_ROW_SPACING)
  local height = NODE_HEIGHT
  return width, height
end

local function getNodeDimsForOutro()
  local scale = 0.8
  local width, height = getNodeDimsForSet(3)
  width = scale * width
  height = height
  return width, height
end

-- Animations -------------------------

local function createLinksSpringAction(self)
  local animation = AHSpawn.new()
  local delayTime = 0
  local delayIncrement = 0.35
  local firstLink = nil

  for _, link in pairs(self.links) do
    if not link:isLinked() then
      local spring = link:createSpringAction()
      link:setVisibleRecursive(true)
      local delay = AHDelay.new(delayTime)
      local currentLinkSequence = AHSequence.new(
        delay,
        function() Sound.playSoundEffect("spring_link") end,
        spring
      )
      animation:addAction(currentLinkSequence)
      delayTime = delayTime + delayIncrement
      firstLink = firstLink or link
    end
  end

  if firstLink then
    local shakeAnimation = self.rootNode:createShakeAnimation(firstLink)
    animation:addAction(shakeAnimation)
  end

  return animation
end

local function createLinksSuckInAction(self, ignoreLinked)
  local animation = AHSpawn.new()
  local delayTime = 0
  local delayIncrement = 0.2
  local firstLinkDone = false

  for _, link in pairs(self.links) do
    if not (ignoreLinked and link:isLinked()) then
      local suckIn = link:createSuckInAction()
      local delay = AHDelay.new(delayTime)
      local currentLinkSequence = AHSequence.new(AHDelay.new(delayTime), suckIn)

      if not firstLinkDone then
        local shakeAnimation = self.rootNode:createShakeAnimation(link)
        currentLinkSequence:addAction(shakeAnimation)
        firstLinkDone = true
      end

      animation:addAction(currentLinkSequence)
      delayTime = delayTime + delayIncrement
    end
  end

  return animation
end

local function createFadeUnlinkedNodesAction(self)
  local animation = AHSpawn.new()
  local duration = 0.5
  local scaleDown = false

  for _, node in pairs(self.nodes) do
    if not node:isLinked() then
      animation:addAction(node:createFadeOutAction(duration, scaleDown))
    end
  end

  return animation
end

local function createSpringAllAction(self, connectionsData)
  local nodeWidth, nodeHeight = getNodeDimsForOutro()
  local graphPanelW, graphPanelH = self.graphPanel:getDims()

  local layouts = TableUtils.shallowCopy(LinkLayouts.getOutroLayouts())
  TableUtils.shuffleArray(layouts)

  local animation = AHSpawn.new()
  local firstLink = nil
  for index, connection in ipairs(connectionsData) do
    local linkData = connection.link
    local nodeData = connection.node

    local layout = layouts[index]

    local node = LinksNode.newForOutro(nodeData, nodeWidth, nodeHeight)
    local link = LinksLink.newForOutro(self.rootNode, linkData, layout, graphPanelW, graphPanelH)
    self.links[link.data.text] = link

    local insertLinkAction = AHCallLogic.new(function()
      self.graphPanel:insert(link, Z_GRAPH_NODES)
    end)

    local insertNodeAction = AHCallLogic.new(function()
      node:setLoc(link:getLoc())
      self.graphPanel:insert(node, Z_GRAPH_NODES)
      link:setLinkedNode(node)
    end)

    local staggerDelay = AHDelay.new(index * 0.1 + math.random() * 0.1)
    local spring = link:createSpringAction()
    local drop = AHSpawn.new(
      node:createDropAnimation(),
      function()
        -- Must create this action at runtime because it requires setLinkedNode
        link:createCorrectAnimation():start()
      end
    )

    local currentLinkAnimation = AHSequence.new(
      staggerDelay,
      insertLinkAction,
      AHSpawn.new(
        spring,
        AHSequence.new(
          AHDelay.new(0.2),
          insertNodeAction,
          drop
        )
      )
    )

    animation:addAction(currentLinkAnimation)
    firstLink = firstLink or link
  end

  if firstLink then
    local shakeAnimation = self.rootNode:createShakeAnimation(firstLink)
    animation:addAction(shakeAnimation)
  end

  return animation
end

local function createWaitForTapAction(self)
  local maxSecondsToWait = 2
  local dismissTimer = AHDelay.new(maxSecondsToWait)

  local tapToContinue = Button.new(self:getWidth(), self:getHeight(), function()
    dismissTimer:stop()
  end)

  return AHCallLogic.new(function()
    self:insertAtCenter(tapToContinue)
    tapToContinue:bringToFront()
    wait(dismissTimer:start())
    tapToContinue:removeSelf()
  end)
end

local function createSlideNodesAction(self)
  local animation = AHSpawn.new()

  for _, node in pairs(self.nodes) do
    local nodeX, nodeY = node:getLoc()
    node:setLoc(nodeX, nodeY - 0.3 * Screen.height)
    node:setVisibleRecursive(true)
    local slide = AHLocationYTo.new(node, 0.5, AHEaseSineOut.new(), nodeY)
    local delay = AHDelay.new(nodeX / self:getWidth() * 0.4)
    animation:addAction(AHSequence.new(delay, slide))
  end

  return animation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinksPanel:clearSet()
  local animation = AHSpawn.new(
    createLinksSuckInAction(self, true),
    createFadeUnlinkedNodesAction(self)
  )
  wait(animation:start())

  local ignoreLinked = true
  clearNodes(self, ignoreLinked)
  clearLinks(self, ignoreLinked)
end

function LinksPanel:clearRound()
  local animation = AHSpawn.new(
    createLinksSuckInAction(self, false),
    createFadeUnlinkedNodesAction(self)
  )
  self.questionTimer:hide()
  wait(animation:start())

  local ignoreLinked = false
  clearNodes(self, ignoreLinked)
  clearLinks(self, ignoreLinked)

  resetLinkLayouts(self)
end

function LinksPanel:setSet(set)
  self:clearSet()

  local nodeWidth, nodeHeight = getNodeDimsForSet(#set.nodes)
  for _, nodeData in ipairs(set.nodes) do
    local node = LinksNode.new(self.dndHandler, nodeData, nodeWidth, nodeHeight)
    node:setDropListener(self.nodeDropListener)
    node:setVisibleRecursive(false)
    -- We use Z_GRAPH_NODES because the z index is maintained when moving the node to the graphPanel.
    self.nodesPanel.row:insert(node, Z_GRAPH_NODES)
    self.dndHandler:addObject(node)
    self.nodes[node.data.text] = node
  end
  self.nodesPanel.row:alignChildren()

  for _, node in pairs(self.nodes) do
    node:useCurrentLocationAsOrigin()
  end

  local graphPanelW, graphPanelH = self.graphPanel:getDims()
  for i, linkData in ipairs(set.links) do
    local linkLayout = table.remove(self.linkLayouts, 1)
    local link = LinksLink.new(self.dndHandler, self.rootNode, linkData, linkLayout, graphPanelW, graphPanelH)
    link:setVisibleRecursive(false)
    self.graphPanel:insert(link, Z_GRAPH_LINKS)
    self.dndHandler:addContainer(link)
    self.links[link.data.text] = link
  end

  self.questionTimer:show()

  local animation = AHSpawn.new(
    createLinksSpringAction(self),
    createSlideNodesAction(self)
  )
  wait(animation:start())
end

function LinksPanel:createRevealAnswersAnimation(linkTimeoutCallback)
  local moveNodes = AHSpawn.new()
  local delayTime = 0
  local staggerTime = 0.75
  for _, node in pairs(self.nodes) do
    if not node:isLinked() then
      if node.data.link ~= nil then
        local link = self.links[node.data.link]
        assert(link, "No link \"" .. node.data.link .. "\" found for node \"" .. node.data.text .. "\"")
        moveNodes:addAction(
          AHSequence.new(
            AHDelay.new(delayTime),
            node:createTimeoutRevealAnimation(link),
            AHSpawn.new(
              node:createDropAnimation(),
              link:createTimeoutAnimation(),
              function() linkTimeoutCallback(link) end
            )
          )
        )
      elseif node:isPicked() then
        moveNodes:addAction(node:createTimeoutDiscardAnimation())
      end
      delayTime = delayTime + staggerTime
    end
  end

  return AHSequence.new(moveNodes, AHDelay.new(1))
end

function LinksPanel:createOutroAnimation(connectionsData)
  return AHSequence.new(
    function() Sound.playSoundEffect("win") end,
    createSpringAllAction(self, connectionsData),
    createWaitForTapAction(self),
    AHSpawn.new(
      -- Must create this action at runtime because it requires setLinkedNode
      function() wait(createLinksSuckInAction(self):start()) end,
      AHSequence.new(
        AHDelay.new(1.5),
        self.rootNode:createHideAnimation()
      )
    )
  )
end

function LinksPanel:setNodeDropListener(listener)
  self.nodeDropListener = listener
end

function LinksPanel:getQuestionTimer()
  return self.questionTimer
end

function LinksPanel:getRootNode()
  return self.rootNode
end

function LinksPanel:waitForAnimations()
  for _, node in pairs(self.nodes) do
    waitWhile(function() return node:isAnimating() end)
  end
  for _, link in pairs(self.links) do
    waitWhile(function() return link:isAnimating() end)
  end
end

function LinksPanel:getCorrectNode(link)
  for _, node in pairs(self.nodes) do
    if link:isCorrectNode(node) then
      return node
    end
  end
  error("No correct node found.")
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, rootNodeText, stepCount)
  local nodesPanel = Group.new(NODES_PANEL_WIDTH, NODES_PANEL_HEIGHT)
  local nodesRow = Row.new(NODES_ROW_SPACING)
  local questionTimer = LinksQuestionTimer.new(NODES_PANEL_WIDTH)
  questionTimer:setAnchorPointTop()
  questionTimer:setLoc(NODES_PANEL_WIDTH/2, NODES_PANEL_HEIGHT)

  nodesPanel:insertAtCenter(nodesRow, Z_NODES_ROW)
  nodesPanel:insert(questionTimer, Z_QUESTION_TIMER)
  nodesPanel.row = nodesRow

  nodesPanel:setLoc(self:getWidth()/2, nodesPanel:getHeight()/2)
  self:insert(nodesPanel, Z_NODES_PANEL)
  self.nodesPanel = nodesPanel
  self.questionTimer = questionTimer

  local graphPanel = Group.new(GRAPH_PANEL_WIDTH, GRAPH_PANEL_HEIGHT)
  graphPanel:setLoc(self:getWidth()/2, self:getHeight() - graphPanel:getHeight()/2)
  self:insert(graphPanel, Z_GRAPH_PANEL)
  self.graphPanel = graphPanel

  local rootNode = LinksRootNode.new(ROOT_NODE_WIDTH, ROOT_NODE_HEIGHT, rootNodeText:upper(), stepCount)
  graphPanel:insertAtCenter(rootNode, Z_GRAPH_ROOT)
  self.rootNode = rootNode
end

-- Constructor ------------------------

function LinksPanel.new(rootNodeText, stepCount)
  local self = Group.newFullScreen()
  Mixin.mix(self, LinksPanel)

  createLayout(self, rootNodeText, stepCount)
  resetLinkLayouts(self)

  self.dndHandler = DragAndDropHandler.new()
  self.nodes = {}
  self.links = {}

  return self
end
