--- LinusBackground
-- @author Created by Jay Austria on October 20, 2015.

class("LinusBackground")

require("graphics.TiledImage")
require("graphics.MultiGradient")
require("LinusCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

LinusBackground.PURPLE_COLOR = ColorUtils.colorHex24ToStd(0x400061)
LinusBackground.NOODLE_UPSCALE = 2.56
LinusBackground.ZOOM_EFFECT_SCALE = 1.2

local INCORRECT_TINT_COLOR = ColorUtils.colorHex24ToStd(0xFF0000)
local CORRECT_TINT_COLOR = ColorUtils.colorHex24ToStd(0x00FFFF)
local BOTTOM_GRADIENT_SCALE = 4
local TOP_GRADIENT_SCALE = 4
local GRADIENT_LOCATION_Y_NORMALIZED = 0.6
local SCALE_PRESS_EFFECT = 0.9
local SCALE_DEFAULT = 1.0
local TOP_GRADIENT_PURPLE_COLOR = ColorUtils.colorHex24ToStd(0xEB13BB)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createMorphGradientOpactiesAction(self, opacity)
  local morphDuration = 0.2
  local morphAction = AHSpawn.new(
    AHOpacityTo.new(self.topGlowGradient, morphDuration, nil, opacity),
    AHOpacityTo.new(self.bottomGlowGradient, morphDuration, nil, opacity)
  )

  return morphAction
end

local function createTintAction(self, tintColor)
  local tintAction = AHColorTo.new(self.topGlowGradient, 0.2, nil, tintColor)

  return tintAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusBackground:createMorphToFlatDisplayAction()
  local morphAction = createMorphGradientOpactiesAction(self, LinusCommon.INVISIBLE_OPACITY)

  return morphAction
end

function LinusBackground:startMorphToGradiatedDisplayAction()
  local morphAction = createMorphGradientOpactiesAction(self, LinusCommon.VISIBLE_OPACITY)

  return morphAction:start()
end

function LinusBackground:createIncorrectShakeAction()
  local shakeMagnitude = self:getWidthPart(0.06)
  local shakeAction = AHSequence.new(
    AHLocationXBy.new(self.noodleLayer, 0.15, AHEaseSineOut.new(), -shakeMagnitude),
    AHLocationXBy.new(self.noodleLayer, 0.4, AHEaseBackOut.new(1.2), shakeMagnitude)
  )

  return shakeAction
end

function LinusBackground:createTintIncorrectAction()
  local tintAction = createTintAction(self, INCORRECT_TINT_COLOR)

  return tintAction
end

function LinusBackground:createTintCorrectAction()
  local tintAction = createTintAction(self, CORRECT_TINT_COLOR)

  return tintAction
end

function LinusBackground:createPressAction()
  local pressAction = AHSequence.new(
    AHScaleTo.new(self, 0.3, AHEaseSineInOut.new(), SCALE_PRESS_EFFECT),
    AHScaleTo.new(self, 1.0, AHEaseExponentialOut.new(), SCALE_DEFAULT)
  )

  return pressAction
end

function LinusBackground:reset()
  local resetAction = AHSpawn.new(
    createMorphGradientOpactiesAction(self, LinusCommon.VISIBLE_OPACITY),
    createTintAction(self, TOP_GRADIENT_PURPLE_COLOR)
  )

  resetAction:start()
end

function LinusBackground:startZoomAction()
  local zoomAction = self:createZoomAction()

  return zoomAction:start()
end

function LinusBackground:createZoomAction()
  local zoomEffectLayer = self.zoomEffectLayer
  local noodleLayer = self.noodleLayer
  local zoomAction = AHSpawn.new(
    AHDelayed.new(0.40, AHOpacityTo.new(zoomEffectLayer, 0.25, nil, LinusCommon.INVISIBLE_OPACITY)),
    AHScaleTo.new(zoomEffectLayer, 0.8, AHEaseSimpleInOut.new(5.0), 1.0),
    AHSequence.new(
      AHScaleTo.new(noodleLayer, 0.6, AHEaseSimpleInOut.new(5.0), 0.95),
      AHScaleTo.new(noodleLayer, 1.0, AHEaseExponentialOut.new(), 1.0)
    ),
    AHOpacityTo.new(self.topGlowGradient, 0.5, nil, LinusCommon.VISIBLE_OPACITY)
  )

  return zoomAction
end

function LinusBackground:createRewindZoomAction()
  local zoomEffectLayer = self.zoomEffectLayer
  local duration = 0.5
  local zoomAction = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(zoomEffectLayer, duration, nil, LinusCommon.VISIBLE_OPACITY),
      AHScaleTo.new(zoomEffectLayer, duration, AHEaseSimpleInOut.new(), LinusBackground.ZOOM_EFFECT_SCALE)
    )
  )

  return zoomAction
end

function LinusBackground:setGradientY(y)
  self.bottomGlowGradient:setLoc(self:getWidthPart(0.5), y)
  self.topGlowGradient:setLoc(self:getWidthPart(0.5), y)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local bottomLayer = FilledRectangle.newFullScreen(LinusBackground.PURPLE_COLOR)
  self:insertAtCenter(bottomLayer)
  self.bottomLayer = bottomLayer

  local bottomGlowGradient = Sprite.newWithName("bottom_gradient.png")
  bottomGlowGradient:setLoc(self:getWidthPart(0.5), self:getHeightPart(GRADIENT_LOCATION_Y_NORMALIZED))
  bottomGlowGradient:setScl(BOTTOM_GRADIENT_SCALE)
  self:insert(bottomGlowGradient)
  self.bottomGlowGradient = bottomGlowGradient

  local widthWithBuffer = self:getWidthPart(1.2)
  local noodleLayer = TiledImage.newWithFilename("tiles/noodles.png", widthWithBuffer, nil, LinusBackground.NOODLE_UPSCALE)
  self:insertAtCenter(noodleLayer)
  self.noodleLayer = noodleLayer

  local zoomEffectLayer = FilledRectangle.newFullScreen(LinusBackground.PURPLE_COLOR)
  local zoomEffectNoodleLayer = TiledImage.newWithFilename("tiles/noodles.png", widthWithBuffer, nil, LinusBackground.NOODLE_UPSCALE)
  zoomEffectLayer:insertAtCenter(zoomEffectNoodleLayer)
  zoomEffectLayer:setScl(LinusBackground.ZOOM_EFFECT_SCALE)
  zoomEffectLayer:linkOpacityAttrToChildren()
  self:insertAtCenter(zoomEffectLayer)
  self.zoomEffectLayer = zoomEffectLayer

  local topGlowGradient = Sprite.newWithName("top_gradient.png")
  topGlowGradient:setLoc(self:getWidthPart(0.5), self:getHeightPart(GRADIENT_LOCATION_Y_NORMALIZED))
  topGlowGradient:setScl(TOP_GRADIENT_SCALE)
  topGlowGradient:setColor(unpack(TOP_GRADIENT_PURPLE_COLOR))
  topGlowGradient:setOpacity(LinusCommon.INVISIBLE_OPACITY)
  self:insert(topGlowGradient)
  self.topGlowGradient = topGlowGradient

  local isHorizontal = false
  local colors = {{0,0,0},
                  {1,1,1,0},
                  {1,1,1,0},
                  {0,0,0},
                  {0,0,0}}
  local thicknesses = {0.45, 0.05, 0.40, 0.10}
  local topGradient = MultiGradient.new(self:getWidth(), self:getHeightPart(1.5), isHorizontal, colors, thicknesses)
  self:insertAtCenter(topGradient)
end

function LinusBackground.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, LinusBackground)

  createLayout(self)

  return self
end
