--- LinusDiamondClusterLayer
-- @author Created by Jay Austria on October 27, 2015.

class("LinusDiamondClusterLayer")

require("LinusDiamond")
require("LinusDiamondRow")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local CLUSTER_WIDTH = Screen.getSafeHeightPart(0.09)
local CLUSTER_HEIGHT = CLUSTER_WIDTH * LinusDiamond.HEIGHT_TO_WIDTH_RATIO
local OVERSIZED_SCALE = 1.35
local FACET_DESCRIPTORS = {
  {
    scale = 0.7,
    locationX = 0.4,
    locationY = 0.3,
    endTransformIndex = 1
  },
  {
    scale = 0.7,
    locationX = 0.2,
    locationY = 0.7,
    endTransformIndex = 2
  },
  {
    scale = 0.7,
    locationX = 0.1,
    locationY = 0.4,
    endTransformIndex = 5
  },
  {
    scale = 0.7,
    locationX = 0.8,
    locationY = 0.3,
    endTransformIndex = 3
  },
  {
    scale = 1.8,
    locationX = 0.35,
    locationY = 0.5,
    endTransformIndex = 5
  },
  {
    scale = 1.0,
    locationX = 0.65,
    locationY = 0.7,
    endTransformIndex = 1
  },
  {
    scale = 1.0,
    locationX = 0.55,
    locationY = 0.2,
    endTransformIndex = 2
  },
  {
    scale = 1.0,
    locationX = 0.3,
    locationY = 0.25,
    endTransformIndex = 3
  },
  {
    scale = 1.0,
    locationX = 0.4,
    locationY = 0.8,
    endTransformIndex = 4
  },
  {
    scale = 1.0,
    locationX = 1.0,
    locationY = 0.5,
    endTransformIndex = 5
  },
  {
    scale = 0.7,
    locationX = 0.0,
    locationY = 0.5,
    endTransformIndex = 1
  },
  {
    scale = 0.5,
    locationX = 0.15,
    locationY = 0.3,
    endTransformIndex = 2
  },
  {
    scale = 0.5,
    locationX = 0.5,
    locationY = 0.0,
    endTransformIndex = 3
  },
  {
    scale = 0.5,
    locationX = 0.5,
    locationY = 1.0,
    endTransformIndex = 4
  }
}

local EXPAND_END_TRANSFORMS = {
  {
    x = Screen.getSafeX(0.2),
    y = Screen.getSafeY(0.58),
    scale = 0.4,
    contractX = Screen.getSafeX(0.2),
    contractY = Screen.getSafeY(0.60),
    contractScale = 0.4
  },
  {
    x = Screen.getSafeX(0.32),
    y = Screen.getSafeY(0.5),
    scale = 0.3,
    contractX = Screen.getSafeX(0.2),
    contractY = Screen.getSafeY(0.35),
    contractScale = 0.7
  },
  {
    x = Screen.getSafeX(0.65),
    y = Screen.getSafeY(0.70),
    scale = 0.4,
    contractX = Screen.getSafeX(0.7),
    contractY = Screen.getSafeY(0.75),
    contractScale = 0.7
  },
  {
    x = Screen.getSafeX(0.70),
    y = Screen.getSafeY(0.53),
    scale = 0.5,
    contractX = Screen.getSafeX(0.85),
    contractY = Screen.getSafeY(0.50),
    contractScale = 0.8
  },
  {
    x = Screen.getSafeX(0.8),
    y = Screen.getSafeY(0.69),
    scale = 0.2,
    contractX = Screen.getSafeX(0.85),
    contractY = Screen.getSafeY(0.77),
    contractScale = 0.4
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusDiamondClusterLayer:createDropAction()
  local dropAction = AHSpawn.new()

  for _, facet in ipairs(self.facets) do
    local facetAction = AHSequence.new(
      facet:createShakeAction(),
      facet:createDropAction()
    )

    dropAction:addAction(facetAction)
  end

  return dropAction
end

function LinusDiamondClusterLayer:createContractAction()
  local contractAction = AHSpawn.new()

  for _, facet in ipairs(self.facets) do
    contractAction:addAction(facet:createContractAction())
  end

  return contractAction
end

function LinusDiamondClusterLayer:createShakeAction()
  local initialX, initialY = self:getLoc()
  local shakeAction = AHSequence.new(
    AHLocationBy.new(self, 0.05, nil, 6, -5),
    AHLocationBy.new(self, 0.05, nil, -8, 10),
    AHLocationBy.new(self, 0.11, nil, 2, -10),
    AHLocationTo.new(self, 0.16, nil, initialX, initialY)
  )

  return shakeAction
end

function LinusDiamondClusterLayer:startExpandAction()
  local expandAction = AHSpawn.new()

  for _, facet in ipairs(self.facets) do
    local endTransformIndex = facet:getEndTransformIndex()
    local endTransform = EXPAND_END_TRANSFORMS[endTransformIndex]

    local clusterExpandX, clusterExpandY = self.cluster:transformLoc(self, endTransform.x, endTransform.y)
    local clusterContractX, clusterContractY = self.cluster:transformLoc(self, endTransform.contractX, endTransform.contractY)

    facet:setContractParameters(clusterContractX, clusterContractY, endTransform.contractScale)
    expandAction:addAction(facet:createExpandAction(clusterExpandX, clusterExpandY, endTransform.scale))
  end

  expandAction:start()
end

function LinusDiamondClusterLayer:reset()
  for _, facet in ipairs(self.facets) do
    facet:reset()
  end
end

function LinusDiamondClusterLayer:prepareForShowAction()
  self:setScl(OVERSIZED_SCALE)

  for _, facet in ipairs(self.facets) do
    facet:hide()
  end
end

function LinusDiamondClusterLayer:startShowAction()
  local facetActions = AHSpawn.new()
  for _, facet in ipairs(self.facets) do
    facetActions:addAction(AHDelayed.new(0.15, facet:createShowAction()))
  end

  local showAction = AHSpawn.new(
    facetActions,
    AHDelayed.new(0.20, AHScaleTo.new(self, 0.4, AHEaseSimpleOut.new(3.0), 1.0))
  )

  return showAction:start()
end

local function createSwallowExtraFacetsAction(self)
  local swallowAction = AHSpawn.new()
  for _, extraFacet in ipairs(self.extraFacets) do
    local randomDelay = MathUtils.getRandomFloat(0.0, 0.3)
    local randomOffset = MathUtils.getRandomFloat(self:getWidthPart(-0.05), self:getWidthPart(0.05))
    swallowAction:addAction(
      AHDelayed.new(randomDelay,
        AHSequence.new(
          function()
            for _, facet in ipairs(self.facets) do
              facet:enableFullAdditiveBlending()
            end
          end,
          AHLocationTo.new(extraFacet, 0.6, AHEaseExponentialIn.new(), self.cluster:getX() + randomOffset, self.cluster:getY() + randomOffset),
          function() extraFacet:enableFullAdditiveBlending() end,
          AHScaleTo.new(extraFacet, 0.3, nil, 0.0)
        )
      )
    )
  end

  return swallowAction
end

function LinusDiamondClusterLayer:createMorphToCheckAction()
  local facetCount = 10
  local extraFacets = {}
  for facetIndex = 1, facetCount do
    local randomX = MathUtils.getRandomFloat(0, Screen.width)
    local randomY = Screen.getSafeHeightPart(-0.3)
    local randomScale = MathUtils.getRandomFloat(0.6, 0.8)

    local facet = LinusDiamondRow.new()
    facet:initializeLocation(randomX, randomY)
    facet:initializeScale(randomScale)

    table.insert(extraFacets, facet)
    self:insert(facet)
  end

  self.extraFacets = extraFacets

  local morphAction = AHSpawn.new(
    createSwallowExtraFacetsAction(self),
    AHSequence.new(
      AHDelay.new(0.8),
      AHScaleTo.new(self.cluster, 0.4, AHEaseSimpleOut.new(), 0.3),
      AHScaleTo.new(self.cluster, 0.2, AHEaseSimpleInOut.new(), 0.0)
    )
  )

  return morphAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local canvas = Group.newFullScreen()
  self.canvas = canvas

  self.scaleForVerticalInsets = Screen.safeHeight / Screen.fullHeight

  local cluster = Group.new(CLUSTER_WIDTH * self.scaleForVerticalInsets,
                            CLUSTER_HEIGHT * self.scaleForVerticalInsets)
  cluster:setLoc(self:getWidthPart(0.5), self.baselineY)
  canvas:insert(cluster)
  self.cluster = cluster

  for _, facetDescriptor in ipairs(FACET_DESCRIPTORS) do
    local additionalScaling = self.scaleForVerticalInsets
    local facet = LinusDiamondRow.new(additionalScaling)
    cluster:insert(facet)

    local x = cluster:getWidthPart(facetDescriptor.locationX)
    local y = cluster:getHeightPart(facetDescriptor.locationY)
    facet:initializeLocation(x, y)
    facet:initializeScale(facetDescriptor.scale * self.scaleForVerticalInsets)
    facet:setEndTransformIndex(facetDescriptor.endTransformIndex)

    table.insert(self.facets, facet)

    facet:startHoveringAnimation()
  end

  -- Renders to an offscreen texture to allow interesting color effects using additive
  -- blending on the diamonds w/o the colors of the background and scene interfering
  local offscreenTexture = OffscreenRenderedTexture.new(canvas, nil, nil)
  local renderedCanvas = Sprite.newWithTexture(offscreenTexture)
  self:insertAtCenter(renderedCanvas)
end

function LinusDiamondClusterLayer.new(baselineY)
  local self = Group.newFullScreen()
  Mixin.mix(self, LinusDiamondClusterLayer)

  self.baselineY = baselineY
  self.facets = {}

  createLayout(self)

  return self
end
