--- LinusDiamondRow
-- @author Created by Jay Austria on October 26, 2015.

class("LinusDiamondRow")

require("LinusDiamond")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DIAMOND_DESCRIPTORS = {
  {
    color = ColorUtils.colorHex32ToStd(0xa7f1fe10),
    sortIndexFromLeft = 0,
    useAdditiveBlend = false
  },
  {
    color = ColorUtils.colorHex32ToStd(0xf25eea10),
    sortIndexFromLeft = 4,
    useAdditiveBlend = false
  },
  {
    color = ColorUtils.colorHex32ToStd(0x66e6fe33),
    sortIndexFromLeft = 1,
    useAdditiveBlend = false
  },
  {
    color = ColorUtils.colorHex32ToStd(0xf246f433),
    sortIndexFromLeft = 3,
    useAdditiveBlend = false
  },
  {
    color = ColorUtils.colorHex32ToStd(0x50aadeff),
    sortIndexFromLeft = 2,
    useAdditiveBlend = true
  }
}

local TILING_FACTOR = 7/15
local SHAKE_MAGNITUDE_MINIMUM = Screen.getSafeWidthPart(0.05)
local SHAKE_MAGNITUDE_MAXIMUM = Screen.getSafeWidthPart(0.10)
local DROP_ROTATION_MINIMUM = -45
local DROP_ROTATION_MAXIMUM = 45

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusDiamondRow:enableFullAdditiveBlending()
  for _, diamond in pairs(self.diamonds) do
    diamond:setBlendMode(MOAIProp2D.BLEND_ADD)
  end
end

function LinusDiamondRow:initializeScale(scale)
  self:setScl(scale)
  self.initialScale = scale
end

function LinusDiamondRow:initializeLocation(x, y)
  self:setLoc(x, y)
  self.initialLocation = {x = x, y = y}

  if self.parent then
    local parentX, parentY = self.parent:getDims()

    -- The rationale behind this math makes little sense, but it ended up looking good
    local driftParameter = MathUtils.distance(x, y, parentX, parentY)
    driftParameter = driftParameter / math.max(self.parent:getWidthPart(0.75), self.parent:getHeightPart(0.75))
    driftParameter = MathUtils.clamp(driftParameter, 0.0, 0.7)
    self.driftParameter = driftParameter
  end
end

function LinusDiamondRow:startHoveringAnimation()
  local hoveringAnimation = AHSpawn.new()
  for _, diamond in pairs(self.diamonds) do
    hoveringAnimation:addAction(diamond:getRandomHoverAnimation(self.initialScale, self.driftParameter, self.additionalScaling))
  end

  self.hoveringAnimation = hoveringAnimation:start()

  return hoveringAnimation
end

function LinusDiamondRow:createExpandAction(x, y, scale)
  local randomDuration = MathUtils.getRandomFloat(0.6, 2.0)
  local expandAction = AHSpawn.new(
    AHLocationTo.new(self, randomDuration, AHEaseExponentialOut.new(), x, y),
    AHScaleTo.new(self, 0.1, AHEaseSimpleOut.new(), scale)
  )

  return expandAction
end

function LinusDiamondRow:createContractAction()
  local contractAction = AHSequence.new(
    AHDelay.new(math.random() * 0.06),
    AHSpawn.new(
      AHLocationTo.new(self, 0.10, AHEaseSineInOut.new(), self.contractLocation.x, self.contractLocation.y),
      AHScaleTo.new(self, 0.10, AHEaseSineInOut.new(), self.contractScale)
    ),
    AHSpawn.new(
      AHLocationTo.new(self, 0.3, AHEaseSimpleIn.new(5.0), self.initialLocation.x, self.initialLocation.y),
      AHDelayed.new(0.2, AHScaleTo.new(self, 0.1, AHEaseSimpleInOut.new(), self.initialScale))
    )
  )

  return contractAction
end

function LinusDiamondRow:setContractParameters(x, y, scale)
  self.contractLocation = {x = x, y = y}
  self.contractScale = scale
end

function LinusDiamondRow:createShakeAction()
  local shakeMagnitude = MathUtils.getRandomFloat(SHAKE_MAGNITUDE_MINIMUM, SHAKE_MAGNITUDE_MAXIMUM)

  local shakeAction = AHSequence.new(
  AHLocationXBy.new(self, 0.1, AHEaseSineOut.new(), shakeMagnitude),
  AHLocationXBy.new(self, 0.3, AHEaseBackOut.new(1.2), -shakeMagnitude)
  )

  return shakeAction
end

function LinusDiamondRow:createDropAction()
  local randomRotation = MathUtils.getRandomFloat(DROP_ROTATION_MINIMUM, DROP_ROTATION_MAXIMUM)

  local dropAction = AHSequence.new(
    AHLocationYBy.new(self, 0.6, AHEaseExponentialIn.new(), -Screen.height),
    AHRotationBy.new(self, 0.4, AHEaseSimpleInOut.new(), randomRotation)
  )

  return dropAction
end

function LinusDiamondRow:hide()
  for _, diamond in ipairs(self.diamonds) do
    diamond:hide()
  end
end

function LinusDiamondRow:createShowAction()
  local showAction = AHSpawn.new()
  for _, diamond in ipairs(self.diamonds) do
    showAction:addAction(diamond:createShowAction())
  end

  return showAction
end

function LinusDiamondRow:reset()
  self:setRot(0)
  self:setLoc(self.initialLocation.x, self.initialLocation.y)
  self:setScl(self.initialScale)
end

function LinusDiamondRow:setEndTransformIndex(index)
  self.endTransformIndex = index
end

function LinusDiamondRow:getEndTransformIndex()
  return self.endTransformIndex
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  for _, diamondDescriptor in ipairs(DIAMOND_DESCRIPTORS) do
    local diamond = LinusDiamond.newWithDescriptor(diamondDescriptor)
    local spanX = self:getWidth() - diamond:getWidth()
    local offsetX = diamond:getWidth() * 0.5
    local diamondCount = #DIAMOND_DESCRIPTORS
    local x = diamondDescriptor.sortIndexFromLeft * (spanX / (diamondCount - 1)) + offsetX
    diamond:setLoc(x, self:getHeightPart(0.5))

    if diamondDescriptor.useAdditiveBlend then
      diamond:setBlendMode(MOAIProp2D.BLEND_ADD)
    end

    self:insert(diamond)
    table.insert(self.diamonds, diamond)
  end
end

function LinusDiamondRow.new(additionalScaling)
  local groupWidth = #DIAMOND_DESCRIPTORS * TILING_FACTOR * LinusDiamond.DEFAULT_WIDTH
  local groupHeight = LinusDiamond.DEFAULT_HEIGHT

  local self = Group.new(groupWidth, groupHeight)
  Mixin.mix(self, LinusDiamondRow)

  self.diamonds = {}
  self.initialScale = 1.0
  self.initialLocation = {x = 0, y = 0}
  self.contractScale = 1.0
  self.contractLocation = {x = 0, y = 0}
  self.driftParameter = 0
  self.endTransformIndex = nil
  self.additionalScaling = additionalScaling

  createLayout(self)

  return self
end
