--- LinusGame
-- @author Created by Jay Austria on October 20, 2015.

class("LinusGame")

require("StandardUIBuilder")
require("LinusConfiguration")
require("LinusScene")
require("LinusConceptGenerator")
require("LinusRoundIndicator")
require("LinusNumberPad")
require("LinusCentralSubmitButton")
require("LinusCommon")
require("bonuses.ResponseSpeedTracker")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local passScore = LinusConfiguration.getLowestPassScore()
  local rankTwoScore = LinusConfiguration.getRankTwoScore()
  local rankThreeScore = LinusConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = LinusConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  local escalatedDifficulty = math.min(1.0, GameSession.getDifficulty() + LinusConfiguration.DIFFICULTY_ESCALATION_AMOUNT)
  self.escalatedDifficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, escalatedDifficulty)
  self.normalDifficultyNameChooser = difficultyNameChooser

  GameSession.setInitialLives(LinusConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setDifficultyEscalation(self)
  if (GameSession.getIncorrectAnswerCount() > 0) then
    GameSession.conceptGenerator:setDifficultyNameChooser(self.normalDifficultyNameChooser)
  elseif (GameSession.getCorrectAnswerCount() >= LinusConfiguration.DIFFICULTY_ESCALATION_AFTER_CORRECT_ANSWER_COUNT) then
    GameSession.conceptGenerator:setDifficultyNameChooser(self.escalatedDifficultyNameChooser)
  end
end

local function didLose(self)
  return (self.livesCounter:getLives() <= 0)
end

local function showAndWaitForHighLevelInterstitial(self)
  wait(self.scene:showHighLevelInterstitial())
end

local function registerCorrectResponse(self)
  GameSession.contentTracker:notifyCorrectAnswer(self.conceptIdentifier)
  GameSession.notifyCorrectAnswer(self.conceptIdentifier)
  GameSession.incrementCorrectAnswerCount()
  self.responseSpeedTracker:notifyCorrectResponse()
  self.hiddenResponseSpeedTracker:notifyCorrectResponse()

  GameSession.increaseScore(LinusConfiguration.CORRECT_ANSWER_POINTS)

  if (self.isFinalRound) then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), LinusConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    local instructionsLabel = self.scene:getInstructionLabel()
    local numberPad = self.numberPad

    instructionsLabel:setVisible(false)
    numberPad:setOpacity(0.0)
  end

  self.scoreLabel:setScore(GameSession.getScore())

  local outroAction = nil
  runInCoroutine(function()
    if (self.isFinalRound) then
      outroAction = self.scene:startOutroAction()
    else
      Sound.playSoundEffect("correct_answer")
      self.scene:startCorrectResponseAction()
    end
  end)

  local unlockDelay = 1.0
  waitFor(unlockDelay)

  if (self.isFinalRound and outroAction) then
    wait(outroAction)
  end
end

local function registerIncorrectResponse(self, value)
  Sound.playSoundEffect("incorrect_answer")
  self.livesCounter:decrement()
  local valueAsString = tostring(value)
  GameSession.contentTracker:notifyIncorrectAnswer(self.conceptIdentifier, nil, valueAsString)
  GameSession.incrementIncorrectAnswerCount()
  self.responseSpeedTracker:notifyIncorrectResponse()
  self.hiddenResponseSpeedTracker:notifyIncorrectResponse()

  wait(self.scene:startIncorrectResponseAction(self.expectedAnswer))

  local isWaitingForRationaleDismissal = true
  local onRationaleDismissed = function()
    isWaitingForRationaleDismissal = false
  end
  self.scene:showRationale(self.concept, onRationaleDismissed)

  local isWaitingForRationaleDismissalWrapper = function()
    return isWaitingForRationaleDismissal
  end

  waitWhile(isWaitingForRationaleDismissalWrapper)

  if (self.isFinalRound and not didLose(self)) then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), LinusConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    self.scene:startRationaleDismissIntoOutroAction()
    self.scene:prepareForOutroAction()
    wait(self.scene:startOutroFromRationaleAction())
  else
    self.scene:startRationaleDismissAction()
  end
end

local function registerResponse(self, value)
  local isEmptyResponse = ((value == nil) or (value == 0))
  if (isEmptyResponse) then
    self.scene:reactToInvalidInput()
    return
  end
  self.numberPad:setTouchEnabled(false)
  self.centralSubmitButton:setTouchEnabled(false)

  local isCorrectResponse = (value == self.expectedAnswer)
  if (isCorrectResponse)then
    registerCorrectResponse(self)
  else
    registerIncorrectResponse(self, value)
  end

  self.wasPreviousAnswerCorrect = isCorrectResponse
  self.isWaitingForResponse = false
end

local function beginRoundWithConcept(self, concept, doIntroAction)
  if (doIntroAction) then
    Sound.playSoundEffect("intro")
    self.scene:startIntroWithConcept(concept)
  else
    self.scene:beginRoundWithConcept(concept, self.wasPreviousAnswerCorrect)
  end

  self.concept = concept
  self.expectedAnswer = concept.answer
  self.conceptIdentifier = concept.identifier
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)

  local delayedEnable = function()
    local enableDelayDuration = 0.65
    waitFor(enableDelayDuration)
    self.numberPad:setTouchEnabled(true)
    self.centralSubmitButton:setTouchEnabled(true)
    self.responseSpeedTracker:notifyStart()
    self.hiddenResponseSpeedTracker:notifyStart()
  end
  runInCoroutine(delayedEnable)

  self.isWaitingForResponse = true

  local isWaitingForResponseWrapper = function()
    return self.isWaitingForResponse
  end

  waitWhile(isWaitingForResponseWrapper)

  self.scene:reset()
end

local function registerGameWon(self)
  GameSession.notifyGameWin()
end

local function registerGameLoss(self)
  GameSession.notifyGameLoss()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusGame:gameFlow()
  local roundCount = LinusConfiguration.ROUNDS_TO_COMPLETE
  for roundIndex = 1, roundCount do
    local doIntroAction = (roundIndex == 1)
    self.isFinalRound = (roundIndex == roundCount)

    setDifficultyEscalation(self)
    local concept = self.conceptGenerator:next()
    beginRoundWithConcept(self, concept, doIntroAction)

    if (didLose(self)) then
      break
    end

    local areSetsRemaining = (self.roundIndicator:getRound() < roundCount)
    if (areSetsRemaining) then
      self.roundIndicator:incrementRound()
    end
  end

  if (not didLose(self)) then
    registerGameWon(self)
  else
    registerGameLoss(self)
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local conceptGenerator = LinusConceptGenerator.new()
  self.conceptGenerator = conceptGenerator

  local numberPad = LinusNumberPad.new()
  local onSubmit = function(value)
    registerResponse(self, value)
  end
  numberPad:setOnSubmitCallback(onSubmit)

  local centralSubmitButton = LinusCentralSubmitButton.new(onSubmit)

  self.numberPad = numberPad
  self.centralSubmitButton = centralSubmitButton

  local scene = LinusScene.new(numberPad, centralSubmitButton, self.colorizedUiElements)
  self:insertAtCenter(scene)
  self.scene = scene

  scene:prepareForIntro()

  local quickResponseWrapper = function(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 LinusConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  local responseSpeedTracker = ResponseSpeedTracker.new(LinusConfiguration.SPEED_BONUS_SECONDS,
                                                        quickResponseWrapper,
                                                        LinusConfiguration.SPEED_BONUS_DECAY_START)
  self.responseSpeedTracker = responseSpeedTracker

  local hiddenQuickResponseWrapper = function(earnedBonusProportion, bonusProportionsSum)
    local speedBonusPoints = LinusConfiguration.HIDDEN_POINTS_PER_FULL_SPEED_BONUS * earnedBonusProportion
    GameSession.increaseScore(speedBonusPoints)
  end
  local hiddenResponseSpeedTracker = ResponseSpeedTracker.new(LinusConfiguration.HIDDEN_SPEED_BONUS_SECONDS,
                                                              hiddenQuickResponseWrapper,
                                                              LinusConfiguration.HIDDEN_SPEED_BONUS_DECAY_START)
  self.hiddenResponseSpeedTracker = hiddenResponseSpeedTracker
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, LinusCommon.UI_COMPONENTS_COLOR_DEFAULT)

  local roundIndicator = LinusRoundIndicator.new(LinusConfiguration.ROUNDS_TO_COMPLETE)
  local roundIndicatorOffsetX = nil
  local roundIndicatorOffsetY = -25
  builder:addWidget(roundIndicator, UIOverlay.NORTHEAST, roundIndicatorOffsetX, roundIndicatorOffsetY)
  self.roundIndicator = roundIndicator

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(LinusConfiguration.INITIAL_LIVES_COUNT)

  local colorizedUiElements = {
    self.roundIndicator,
    self.pauseButton,
    self.bonusTray,
    self.scoreLabel,
    self.livesCounter
  }

  self.colorizedUiElements = colorizedUiElements
end

-- Constructor ------------------------

function LinusGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, LinusGame)

  self.expectedAnswer = nil
  self.wasPreviousAnswerCorrect = false
  self.isFinalRound = false

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)

  if LinusConfiguration.shouldShowHighDifficultyInterstitial() then
    showAndWaitForHighLevelInterstitial(self)
  end

  return self
end
