--- LinusHighLevelInterstitial
-- @author Created by Jay Austria on November 3, 2015.

class("LinusHighLevelInterstitial")

require("LinusCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INTERSTITIAL_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 18)
local MORE_INFO_BUTTON_PADDING_Y = Screen.getHeightPart(0.05)
local PAGE_DESCRIPTORS = {
  {
    text = Localizer.get("HIGH_DIFFICULTY"),
    locationY = Screen.getHeightPart(0.55),
    moreInfoAnchorFunction = "getYBottom",
    moreInfoAnchorOffsetY = -MORE_INFO_BUTTON_PADDING_Y
  },
  {
    text = Localizer.get("HIGH_DIFFICULTY_EXPLANATION"),
    locationY = Screen.getHeightPart(0.48),
    moreInfoAnchorFunction = "getYTop",
    moreInfoAnchorOffsetY = MORE_INFO_BUTTON_PADDING_Y
  }
}
local TEXT_AREA_WIDTH_MULTIPLIER = 0.8

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createHiddenPage(self, pageIndex)
  local descriptor = PAGE_DESCRIPTORS[pageIndex]

  local pageText = DynamicTextArea.newSimple(descriptor.text, Screen.getWidthPart(TEXT_AREA_WIDTH_MULTIPLIER), INTERSTITIAL_FONT, Alignment.LEFT)
  pageText:setLoc(self:getWidthPart(0.5), descriptor.locationY)

  pageText:setOpacity(LinusCommon.INVISIBLE_OPACITY)

  return pageText
end

local function createFadeAction(element, opacity)
  local fadeAction = AHOpacityTo.new(element, 0.2, nil, opacity)

  return fadeAction
end

local function createMoveMoreInfoAction(self, pageIndex)
  local descriptor = PAGE_DESCRIPTORS[pageIndex]
  local locationY = self.currentPage[descriptor.moreInfoAnchorFunction](self.currentPage) + descriptor.moreInfoAnchorOffsetY
  local moveAction = AHLocationYTo.new(self.moreInfoButton, 0.3, AHEaseSimpleInOut.new(), locationY)

  return moveAction
end

local function startShowNextPageAction(self)
  local oldPage = self.currentPage

  self.currentPageIndex = (self.currentPageIndex % #PAGE_DESCRIPTORS) + 1

  local newPage = createHiddenPage(self, self.currentPageIndex)
  self:insert(newPage)
  self.currentPage = newPage

  local moreInfoAction = AHSequence.new(
    function() self.virtualButton:setTouchEnabled(false) end,
    createMoveMoreInfoAction(self, self.currentPageIndex),
    function() self.virtualButton:setTouchEnabled(true) end
  )

  local fadeOldPageAction = createFadeAction(oldPage, LinusCommon.INVISIBLE_OPACITY)
  local fadeNewPageAction = createFadeAction(newPage, LinusCommon.VISIBLE_OPACITY)

  local nextPageAction = AHSpawn.new(
    moreInfoAction,
    AHSequence.new(
      fadeOldPageAction,
      fadeNewPageAction
    )
  )

  nextPageAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusHighLevelInterstitial:startShowAction()
  local pageIndex = self.currentPageIndex
  local page = createHiddenPage(self, pageIndex)
  local moreInfoButton = self.moreInfoButton

  local moreInfoButtonX = (self:getWidth() - self:getWidthPart(TEXT_AREA_WIDTH_MULTIPLIER)) * 0.5
  moreInfoButton:setLoc(moreInfoButtonX, page:getYBottom() - MORE_INFO_BUTTON_PADDING_Y)

  self:insert(page)
  self:insert(moreInfoButton)

  moreInfoButton:setOpacity(LinusCommon.INVISIBLE_OPACITY)

  local showAction = AHSpawn.new(
    createFadeAction(page, LinusCommon.VISIBLE_OPACITY),
    createFadeAction(moreInfoButton, LinusCommon.VISIBLE_OPACITY)
  )

  showAction:start()

  self.currentPage = page
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local widthWithBuffer = self:getWidthPart(1.2)
  local noodleLayer = TiledImage.newWithFilename("tiles/noodles.png", widthWithBuffer, nil, LinusBackground.NOODLE_UPSCALE)
  noodleLayer:setScl(LinusBackground.ZOOM_EFFECT_SCALE)
  self:insertAtCenter(noodleLayer)
  self.noodleLayer = noodleLayer

  local isHorizontal = false
  local colors = {{0,0,0,0.2},
                  {1,1,1,0},
                  {0,0,0}}
  local thicknesses = {0.4, 0.6}
  local topGradient = MultiGradient.new(self:getWidth(), self:getHeight(), isHorizontal, colors, thicknesses)
  self:insertAtCenter(topGradient)

  local moreInfoButton = Label.new(Localizer.get("MORE_INFO"), INTERSTITIAL_FONT)
  moreInfoButton:setAnchorPointLeft()
  moreInfoButton:setColor(unpack(LinusCommon.COLORS.MAGENTA))

  local underlineWidth = moreInfoButton:getWidthPart(1.05)
  local underlineThickness = LinusCommon.SINGLE_PIXEL_DIMENSION
  local underline = FilledRectangle.new(underlineWidth, underlineThickness, LinusCommon.COLORS.MAGENTA)
  underline:setLoc(moreInfoButton:getWidthPart(0.5), moreInfoButton:getHeightPart(0.0))
  moreInfoButton:insert(underline)

  local registerPress = function()
    startShowNextPageAction(self)
  end

  local oversizedButtonWidth = moreInfoButton:getWidthPart(1.5)
  local oversizedButtonHeight = moreInfoButton:getHeightPart(2)
  local virtualButton = Button.addVirtualButton(moreInfoButton, oversizedButtonWidth, oversizedButtonHeight, registerPress)

  local continueLabel = Label.new(Localizer.get("Misc.CONTINUE"), Font.new(FontFaces.DIN_OT_LIGHT, 16))
  continueLabel:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.05))
  self:insert(continueLabel)

  self.moreInfoButton = moreInfoButton
  self.virtualButton = virtualButton
end

function LinusHighLevelInterstitial.new()
  local self = FilledRectangle.newFullScreen(LinusBackground.PURPLE_COLOR)
  Mixin.mix(self, LinusHighLevelInterstitial)

  self.currentPageIndex = 1
  createLayout(self)

  return self
end
