--- LinusInputField
-- @author Created by Jay Austria on October 20, 2015.

class("LinusInputField")

require("widgets.number_pad.NumberPadInputField")
require("LinusLightSource")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

LinusInputField.FONT_FACE = FontFaces.DIN_OT_LIGHT
LinusInputField.FONT_SIZE = 129

local GROUP_HEIGHT = Screen.getHeightPart(0.1)
local DEFAULT_COLOR = {1.0, 1.0, 1.0}
local INCORRECT_COLOR = {1, 0, 0}
local SCALE_EXPANDED = 1.75
local SCALE_DEFAULT = 1.0
local INPUT_FIELD_Z = 10
local LIGHT_SOURCE_DESCRIPTORS = {
  {
    x = Screen.getWidthPart(0.4),
    y = Screen.getHeightPart(0.45),
    z = 100,
    intensity = 0.15
  },
  {
    x = Screen.getWidthPart(0.6),
    y = Screen.getHeightPart(0.5),
    z = 70,
    intensity = 0.2
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createShadowFromValueString(self, valueAsString)
  local label = Label.new(valueAsString, self.font, nil, Alignment.CENTER)
  label:setColor(unpack(LinusCommon.COLORS.BLACK))

  return label
end

local function updateShadows(self, view)
  for lightIndex, lightSource in ipairs(self.lightSources) do
    local worldX, worldY = view:getWorldLoc()
    local distanceX = worldX - lightSource:getX()
    local distanceY = worldY - lightSource:getY()

    local offsetX = distanceX * 0.1
    local offsetY = distanceY * 0.1

    local shadow = view.shadows[lightIndex]
    shadow:setLoc(view:getWidthPart(0.5) + offsetX, view:getHeightPart(0.5) + offsetY)

    local lightZ = lightSource:getZ()
    local shadowScale = lightZ / (lightZ - INPUT_FIELD_Z)
    shadow:setScl(shadowScale)

    shadow:setOpacity(lightSource:getIntensity())
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusInputField:_createViewForValueString(valueAsString)
  local viewForValue = self.super._createViewForValueString(self, valueAsString)

  local shadows = {}
  for shadowIndex = 1, #self.lightSources do
    local shadow = createShadowFromValueString(self, valueAsString)

    viewForValue:insertAtCenter(shadow)
    shadow:bringToBack()
    table.insert(shadows, shadow)
  end

  viewForValue.shadows = shadows

  return viewForValue
end

function LinusInputField:reactToInvalidInput()
  local startX = self:getWidthPart(0.5)

  local shakeAction = AHSequence.new(
    AHLocationXTo.new(self.viewsRow, 0.06, AHEaseSineOut.new(), startX + 20),
    AHLocationXTo.new(self.viewsRow, 0.06, AHEaseSineInOut.new(), startX - 10),
    AHLocationXTo.new(self.viewsRow, 0.06, AHEaseSineInOut.new(), startX + 5),
    AHLocationXTo.new(self.viewsRow, 0.06, AHEaseSineInOut.new(), startX)
  )

  self.shakeActionInterrupt:insert(shakeAction)

  if self.invalidInputCallback then
    self.invalidInputCallback()
  end
end

function LinusInputField:createDropAction()
  local dropAction = AHSpawn.new()

  for _, view in ipairs(self.viewsRow.children) do
    local randomRotation = MathUtils.getRandomFloat(-45, 45)

    dropAction:addAction(
      AHSpawn.new(
        AHLocationYBy.new(view.valueLabel, 0.8, AHEaseExponentialIn.new(), -Screen.height / self.scaleAtStart),
        AHRotationBy.new(view.valueLabel, 0.5, AHEaseSimpleInOut.new(), randomRotation)
      )
    )
  end

  return dropAction
end

function LinusInputField:createSwallowAction()
  local centerX, _ = self.viewsRow:getWidthPart(0.5)
  local viewsAction = AHSpawn.new()

  for _, view in ipairs(self.viewsRow.children) do
    viewsAction:addAction(AHLocationXTo.new(view, 0.2, AHEaseSineInOut.new(), centerX))
  end

  local swallowAction = AHSequence.new(
    AHScaleTo.new(self, 0.1, AHEaseSineInOut.new(), SCALE_EXPANDED * self.scaleAtStart),
    AHDelay.new(0.1),
    AHSpawn.new(
      viewsAction,
      AHScaleTo.new(self, 0.2, AHEaseSineInOut.new(), 0.6 * self.scaleAtStart)
    )
  )

  return swallowAction
end

function LinusInputField:createHideAction()
  local hideAction = AHSpawn.new()

  for _, view in ipairs(self.viewsRow.children) do
    hideAction:addAction(AHScaleTo.new(view, 0.1, nil, 0.0))
  end

  return hideAction
end

function LinusInputField:reset()
  self:setScl(self.scaleAtStart)
  self:clear()
end

function LinusInputField:startShadowTrackingCoroutine()
  local trackShadows = function()
    while true do
      for _, view in ipairs(self.viewsRow.children) do
        updateShadows(self, view)
      end

      coroutine.yield()
    end
  end

  self.shadowTrackingCoroutine = runInCoroutine(trackShadows)
end

function LinusInputField:startFadeLightSourcesOnAction()
  local fadeAction = AHSpawn.new()
  for _, lightSource in ipairs(self.lightSources) do
    fadeAction:addAction(lightSource:createFadeOnAction())
  end

  return fadeAction:start()
end

function LinusInputField:createFadeLightSourcesOffAction()
  local fadeAction = AHSpawn.new()
  for _, lightSource in ipairs(self.lightSources) do
    fadeAction:addAction(lightSource:createFadeOffAction())
  end

  return fadeAction
end

function LinusInputField:setScaleByAreaHeight(areaHeight)
  local viewHeight = self:getViewHeight()
  assert(viewHeight > 0, "Computed view height for input field must be greater than zero.")

  local scale = math.min(1.0, areaHeight / self:getViewHeight())
  self:setScl(scale)

  self.scaleAtStart = scale
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLightSources(self)
  local lightSources = {}
  for _, lightDescriptor in ipairs(LIGHT_SOURCE_DESCRIPTORS) do
    local lightSource = LinusLightSource.new()

    lightSource:setLoc(lightDescriptor.x, lightDescriptor.y)
    lightSource:setZ(lightDescriptor.z)
    lightSource:initializeIntensity(lightDescriptor.intensity)
    lightSource:startRandomWobbleAction()

    table.insert(lightSources, lightSource)
  end

  self.lightSources = lightSources
end

function LinusInputField.new()
  local self = NumberPadInputField.new(GROUP_HEIGHT, LinusInputField.FONT_FACE, LinusInputField.FONT_SIZE, DEFAULT_COLOR, INCORRECT_COLOR)
  Mixin.mix(self, LinusInputField)

  self.shakeActionInterrupt = ActionInterrupt.new()
  createLightSources(self)

  self.scaleAtStart = SCALE_DEFAULT

  return self
end
