--- LinusPrompt
-- @author Created by Jay Austria on October 20, 2015.

class("LinusPrompt")

require("LinusCommon")
require("LinusDigitsLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local EQUATION_FONT_SIZE = LinusCommon.getAspectRatioAdjustedFontSize(36)
local EQUATION_FONT = Font.new(FontFaces.DIN_OT_BOLD, EQUATION_FONT_SIZE)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function alignGroupsNeedingAlignment(self)
  for _, group in ipairs(self.groupsNeedingAlignment) do
    group:alignChildren()
  end
end

local function setEquationOffscreen(self, scene)
  local _, worldBottomY = scene:transformLoc(self, self.equationGroup:getYBottom())
  local offscreenYOffset = Screen.height - worldBottomY + self.equationGroup:getHeightPart(0.5)

  self.equationGroup:setLoc(self:getWidthPart(0.5), offscreenYOffset)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusPrompt:getInstructionLabel()
  return self.instructionsLabel
end

function LinusPrompt:setEquation(minuend, subtrahend)
  local minuendString = tostring(minuend)
  local subtrahendString = tostring(subtrahend)

  if (self.isHorizontal) then
    self.minuendLabel:setText(minuendString)
    self.subtrahendLabel:setText(subtrahendString)
  else
    local newMinuendWidth = self.minuendLabel:setText(minuendString)
    local newSubtrahendWidth = self.subtrahendLabel:setText(subtrahendString)
    local commonWidth = math.max(newMinuendWidth, newSubtrahendWidth)
    self.minuendLabel:updateDigitsLayout(commonWidth)
    self.subtrahendLabel:updateDigitsLayout(commonWidth)
  end

  alignGroupsNeedingAlignment(self)
end

function LinusPrompt:prepareForPresentAction(scene)
  setEquationOffscreen(self, scene)
  self.equationGroup:setScl(1.0)
  self:setOpacity(LinusCommon.INVISIBLE_OPACITY)
end

function LinusPrompt:createPresentAction()
  local presentAction = AHSpawn.new(
    AHOpacityTo.new(self, 0.5, nil, LinusCommon.VISIBLE_OPACITY),
    AHDelayed.new(0.2, AHLocationTo.new(self.equationGroup, 0.5, AHEaseBackOut.new(0.7), self.equationOriginLocation.x, self.equationOriginLocation.y))
  )

  return presentAction
end

function LinusPrompt:createSwallowAtLocationAction(x, y)
  local swallowAction = AHSpawn.new(
    AHDelayed.new(0.1, AHLocationTo.new(self.equationGroup, 0.3, AHEaseExponentialIn.new(), x, y)),
    AHDelayed.new(0.2, AHScaleTo.new(self.equationGroup, 0.5, AHEaseExponentialIn.new(), 0.0))
  )

  return swallowAction
end

function LinusPrompt:createShowAsCorrectAction(willFlipBack)
  if willFlipBack == nil then
    willFlipBack = true
  end

  local showAction = AHSequence.new(
    AHColorTo.new(self.instructionsLabel, 0.1, nil, LinusCommon.UI_COMPONENTS_COLOR_POSITIVE)
  )

  if willFlipBack then
    showAction:addAction(AHDelay.new(0.5))
    showAction:addAction(AHColorTo.new(self.instructionsLabel, 0.1, nil, LinusCommon.UI_COMPONENTS_COLOR_DEFAULT))
  end

  return showAction
end

function LinusPrompt:createHideAction()
  local hideAction = AHOpacityTo.new(self, 0.3, nil, LinusCommon.INVISIBLE_OPACITY)

  return hideAction
end

function LinusPrompt:createShowAction()
  local hideAction = AHOpacityTo.new(self, 0.3, nil, LinusCommon.VISIBLE_OPACITY)

  return hideAction
end

function LinusPrompt:prepareForShowAction(scene)
  setEquationOffscreen(self, scene)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local isHorizontal = self.isHorizontal
  local instructionsLabel = Label.new(Localizer.get("FIND_DIFFERENCE"), Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  instructionsLabel:setAnchorPointTop()
  instructionsLabel:setLoc(self:getWidthPart(0.5), self:getHeight())
  instructionsLabel:setColor(unpack(LinusCommon.UI_COMPONENTS_COLOR_DEFAULT))
  self:insert(instructionsLabel)
  self.instructionsLabel = instructionsLabel

  if isHorizontal then
    local equationRow = Row.new(10)
    table.insert(self.groupsNeedingAlignment, equationRow)

    local minuendLabel = Label.new("", EQUATION_FONT)
    equationRow:insert(minuendLabel)
    self.minuendLabel = minuendLabel

    local subtractionSymbol = Label.new("-", EQUATION_FONT)
    equationRow:insert(subtractionSymbol)

    local subtrahendLabel = Label.new("", EQUATION_FONT)
    equationRow:insert(subtrahendLabel)
    self.subtrahendLabel = subtrahendLabel

    self:insert(equationRow)
    self.equationGroup = equationRow
  else
    local equationRow = Row.new(17)

    local subtractionSymbolColumn = Column.new()
    local valuesColumn = Column.new(nil, Alignment.RIGHT)

    table.insert(self.groupsNeedingAlignment, subtractionSymbolColumn)
    table.insert(self.groupsNeedingAlignment, valuesColumn)
    table.insert(self.groupsNeedingAlignment, equationRow)

    equationRow:insert(subtractionSymbolColumn)
    equationRow:insert(valuesColumn)

    local minuendLabel = LinusDigitsLabel.new(EQUATION_FONT)
    valuesColumn:insert(minuendLabel)
    self.minuendLabel = minuendLabel

    subtractionSymbolColumn:insertSpacer(minuendLabel:getHeight())

    local subtractionSymbol = Label.new("-", EQUATION_FONT)
    subtractionSymbolColumn:insert(subtractionSymbol)

    equationRow:insertSpacer(subtractionSymbolColumn:getWidth())

    local subtrahendLabel = LinusDigitsLabel.new(EQUATION_FONT)
    valuesColumn:insert(subtrahendLabel)
    self.subtrahendLabel = subtrahendLabel

    self:insert(equationRow)
    self.equationGroup = equationRow
  end

  alignGroupsNeedingAlignment(self)
  self:alignChildren()

  local originX, originY = self.equationGroup:getLoc()
  self.equationOriginLocation.x = originX
  self.equationOriginLocation.y = originY

  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)
end

function LinusPrompt.new(isHorizontal)
  local paddingY = 5
  local self = Column.new(paddingY, nil, Column.GrowDown)
  Mixin.mix(self, LinusPrompt)

  self.groupsNeedingAlignment = {}
  self.equationOriginLocation = {x = 0, y = 0}
  self.equationGroup = nil
  self.isHorizontal = isHorizontal
  createLayout(self)

  return self
end
