--- LinusScene
-- @author Created by Jay Austria on October 20, 2015.

class("LinusScene")

require("widgets.TapToContinue")
require("LinusInputField")
require("LinusBackground")
require("graphics.OffscreenRenderedTexture")
require("LinusDiamondClusterLayer")
require("LinusCommon")
require("LinusPrompt")
require("LinusOverExposure")
require("LinusHighLevelInterstitial")
require("LinusRationale")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CHECK_DESCRIPTORS = {
  {
    scale = 1.4,
    opacity = 0.7,
    hoverDuration = 1,
    hoverAmount = 10
  },
  {
    scale = 2.0,
    opacity = 0.4,
    hoverDuration = 2.0,
    hoverAmount = 40
  },
  {
    scale = 2.5,
    opacity = 0.1,
    hoverDuration = 3.0,
    hoverAmount = 60
  }
}
local OUTRO_DURATION = 4.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createOnButtonPressCallback(self)
  return function(keyValue)
    local randomIndex = math.random(1, 3)
    local randomFilename = "input0" .. randomIndex
    Sound.playSoundEffect(randomFilename)

    local keyAsNumber = tonumber(keyValue)
    local isValidNumber = (keyAsNumber ~= nil) and (keyAsNumber ~= 0)
    if isValidNumber and not self.hasUserInputtedNumber then
      if not self.hasUserInputtedNumber then

        local expandDiamondCoroutine = function()
          self.overExposure:startQuickGlowAction()
          waitFor(0.1)
          self.diamondClusterLayer:startExpandAction()
        end

        runInCoroutine(expandDiamondCoroutine)

        self.hasUserInputtedNumber = true
      end
    end
  end
end

local function createShowUiAsCorrectAction(self)
  local showAction = AHSpawn.new()

  for _, uiElement in ipairs(self.uiElements) do
    showAction:addAction(AHSequence.new(
      AHColorTo.new(uiElement, 0.1, nil, LinusCommon.UI_COMPONENTS_COLOR_POSITIVE),
      AHDelay.new(0.5),
      AHColorTo.new(uiElement, 0.1, nil, LinusCommon.UI_COMPONENTS_COLOR_DEFAULT)
    ))
  end

  showAction:addAction(self.prompt:createShowAsCorrectAction())

  return showAction
end

local function createShowAnswerAction(self, expectedAnswer)
  local expectedAnswerLabel = self.expectedAnswerLabel
  local showAction = AHSequence.new(
    function() expectedAnswerLabel:setText(expectedAnswer) end,
    AHOpacityTo.new(expectedAnswerLabel, 0.3, nil, LinusCommon.VISIBLE_OPACITY)
  )

  return showAction
end

local function createHideAnswerAction(self)
  local hideAction = AHOpacityTo.new(self.expectedAnswerLabel, 0.3, nil, LinusCommon.INVISIBLE_OPACITY)

  return hideAction
end

local function createShowExpectedAnswerAction(self, expectedAnswer)
  local backgroundAction = AHSequence.new(
    AHSpawn.new(
      self.background:createIncorrectShakeAction(),
      self.background:createTintIncorrectAction()
    ),
    AHDelay.new(0.5),
    self.background:createMorphToFlatDisplayAction()
  )

  local inputFieldAction = AHSpawn.new(
    self.inputField:createDropAction(),
    AHDelayed.new(1.3, AHSpawn.new(
      self.inputField:createFadeLightSourcesOffAction(),
      createShowAnswerAction(self, expectedAnswer)
    ))
  )

  local diamondClusterLayerAction = self.diamondClusterLayer:createDropAction()

  local numberPadAction = function() self.numberPad:reset() end

  local showExpectedAnswerAction = AHSpawn.new(
    backgroundAction,
    inputFieldAction,
    diamondClusterLayerAction,
    numberPadAction
  )

  return showExpectedAnswerAction
end

local function createUiOpacityAction(self, opacity)
  local combinedAction = AHSpawn.new()

  for _, uiElement in ipairs(self.uiElements) do
    combinedAction:addAction(AHOpacityTo.new(uiElement, 0.3, nil, opacity))
  end

  return combinedAction
end

local function createHideUiAction(self, delay)
  local hideAction = AHDelayed.new(delay, createUiOpacityAction(self, LinusCommon.INVISIBLE_OPACITY))

  return hideAction
end

local function createShowUiAction(self)
  local showAction = createUiOpacityAction(self, LinusCommon.VISIBLE_OPACITY)

  return showAction
end

local function createFadeOutAllAction(self, uiHideDelay)
  local backgroundAction = self.background:createRewindZoomAction()

  local numberPadAction = self.numberPad:createHideAction()

  local promptAction = self.prompt:createHideAction()

  local fadeAction = AHSpawn.new(
    backgroundAction,
    createHideAnswerAction(self),
    numberPadAction,
    promptAction,
    createHideUiAction(self, uiHideDelay)
  )

  return fadeAction
end

local function createShowCorrectAndHideUiAction(self)
  local uiHideDelay = 2.4
  local action = AHSpawn.new(
    AHDelayed.new(0.5, createFadeOutAllAction(self, uiHideDelay))
  )

  for _, uiElement in ipairs(self.uiElements) do
    action:addAction(AHColorTo.new(uiElement, 0.1, nil, LinusCommon.UI_COMPONENTS_COLOR_POSITIVE))
  end

  local willFlipBack = false
  action:addAction(self.prompt:createShowAsCorrectAction(willFlipBack))

  return action
end

local function createOutroAction(self)
  local inputFieldX, inputFieldY = self.inputField:getLoc()
  local inputFieldAction = AHSpawn.new(
    self.inputField:createSwallowAction(),
    AHDelayed.new(0.6, self.inputField:createHideAction())
  )
  local promptSpaceX, promptSpaceY = self.prompt:transformLoc(self, inputFieldX, inputFieldY)

  local equationGroupAction = self.prompt:createSwallowAtLocationAction(promptSpaceX, promptSpaceY)
  local uiAction = createShowCorrectAndHideUiAction(self)

  local diamondClusterLayerAction = AHSpawn.new(
    self.diamondClusterLayer:createContractAction(),
    AHDelayed.new(0.7, self.diamondClusterLayer:createShakeAction()),
    self.diamondClusterLayer:createMorphToCheckAction()
  )

  local overExposureAction = self.overExposure:createOutroOverExposeAction()

  local backgroundAction = self.background:createTintCorrectAction()

  local createHoveringChecks = function()
    local inputFieldBaselineY = self.prompt:getYBottom() + (self.numberPad:getYTop() - self.prompt:getYBottom()) * 0.5

    for _, checkDescriptor in ipairs(CHECK_DESCRIPTORS) do
      local check = Sprite.newWithName("check.png")
      check:setScl(0)
      check:setOpacity(checkDescriptor.opacity)
      check:setBlendMode(MOAIProp2D.BLEND_ADD)

      check:setLoc(self:getWidthPart(0.5), inputFieldBaselineY)
      self:insert(check)

      local checkAction = AHSpawn.new(
        AHDelayed.new(0.7, AHScaleTo.new(check, 1.3, AHEaseExponentialOut.new(), checkDescriptor.scale)),
        AHSequence.new(
          AHLocationXBy.new(check, checkDescriptor.hoverDuration, AHEaseSimpleInOut.new(), checkDescriptor.hoverAmount * 0.5),
          AHLoop.new(
            AHSequence.new(
              AHLocationXBy.new(check, checkDescriptor.hoverDuration, AHEaseSimpleInOut.new(), -checkDescriptor.hoverAmount),
              AHLocationXBy.new(check, checkDescriptor.hoverDuration, AHEaseSimpleInOut.new(), checkDescriptor.hoverAmount)
            )
          )
        ),
        AHSequence.new(
          AHLocationYBy.new(check, checkDescriptor.hoverDuration * 2, AHEaseSimpleInOut.new(), checkDescriptor.hoverAmount * 0.5),
          AHLoop.new(
            AHSequence.new(
              AHLocationYBy.new(check, checkDescriptor.hoverDuration * 2, AHEaseSimpleInOut.new(), -checkDescriptor.hoverAmount),
              AHLocationYBy.new(check, checkDescriptor.hoverDuration * 2, AHEaseSimpleInOut.new(), checkDescriptor.hoverAmount)
            )
          )
        )
      )

      checkAction:start()
    end
  end

  local outroAction = AHSpawn.new(
    function() Sound.playSoundEffect("outro") end,
    function() createHoveringChecks() end,
    inputFieldAction,
    equationGroupAction,
    diamondClusterLayerAction,
    overExposureAction,
    backgroundAction,
    uiAction,
    AHDelay.new(OUTRO_DURATION)
  )

  return outroAction
end

local function createScreenShakeAction()
  local rootGroup = Screen.rootGroup
  local x, y = rootGroup:getLoc()

  local shakeOffsetX = -3
  local shakeOffsetY = -3

  local screenShakeAction = AHSequence.new(
    function() rootGroup:setLoc(x + shakeOffsetX, y + shakeOffsetY) end,
    AHLocationTo.new(rootGroup, 1.0, AHEaseElasticOut.new(), x, y)
  )

  return screenShakeAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusScene:getInstructionLabel()
  local prompt = self.prompt
  local instructionsLabel = prompt:getInstructionLabel()

  return instructionsLabel
end

function LinusScene:showHighLevelInterstitial()
  local interstitial = LinusHighLevelInterstitial.new()
  interstitial:startShowAction()

  local tapToContinue = TapToContinue.new(interstitial)
  tapToContinue:waitForTrigger()

  tapToContinue:removeSelf()
end

function LinusScene:prepareForIntro()
  self.prompt:prepareForPresentAction(self)
  self.numberPad:hide()
  self.diamondClusterLayer:prepareForShowAction()
end


function LinusScene:startIntroWithConcept(concept)
  self.background:startZoomAction()

  local showNumberPadAction = AHDelayed.new(0.5, self.numberPad:createShowAction())
  showNumberPadAction:start()

  local wasPreviousAnswerCorrect = false
  self:beginRoundWithConcept(concept, wasPreviousAnswerCorrect)
end

function LinusScene:beginRoundWithConcept(concept, wasPreviousAnswerCorrect)
  self.hasUserInputtedNumber = false

  self.prompt:setEquation(concept.minuend, concept.subtrahend)
  self.prompt:prepareForPresentAction(self)

  local promptAction = AHDelayed.new(0.5, self.prompt:createPresentAction())
  promptAction:start()

  self.inputField:startFadeLightSourcesOnAction()

  if not wasPreviousAnswerCorrect then
    self.diamondClusterLayer:prepareForShowAction()
    self.diamondClusterLayer:startShowAction()
  end
end

function LinusScene:prepareForOutroAction()
  self.diamondClusterLayer:reset()
end

function LinusScene:startOutroAction()
  local outroAction = AHSpawn.new(
    createOutroAction(self),
    self.background:createPressAction()
  )

  return outroAction:start()
end

function LinusScene:startOutroFromRationaleAction()
  local outroAction = createOutroAction(self)

  return outroAction:start()
end

function LinusScene:startCorrectResponseAction()
  local inputFieldX, inputFieldY = self.inputField:getLoc()
  local inputFieldAction = self.inputField:createSwallowAction()
  local promptSpaceX, promptSpaceY = self.prompt:transformLoc(self, inputFieldX, inputFieldY)
  local equationGroupAction = self.prompt:createSwallowAtLocationAction(promptSpaceX, promptSpaceY)
  local uiAction = createShowUiAsCorrectAction(self)
  local screenAction = AHDelayed.new(0.4, createScreenShakeAction())

  local numberPadAction = AHSpawn.new(
    function() self.numberPad:reset() end,
    self.numberPad:createCorrectResponseAction()
  )

  local diamondClusterLayerAction = AHSpawn.new(
    self.diamondClusterLayer:createContractAction(),
    AHDelayed.new(0.30, self.diamondClusterLayer:createShakeAction())
  )

  local overExposureAction = self.overExposure:createCorrectResponseOverExposeAction()

  local backgroundAction = AHSpawn.new(
    self.background:createPressAction(),
    self.background:createTintCorrectAction()
  )

  local correctResponseAction = AHSpawn.new(
    inputFieldAction,
    equationGroupAction,
    diamondClusterLayerAction,
    overExposureAction,
    backgroundAction,
    uiAction,
    numberPadAction,
    screenAction
  )

  correctResponseAction:start()
end

function LinusScene:startIncorrectResponseAction(expectedAnswer)
  local showExpectedAnswerAction = createShowExpectedAnswerAction(self, expectedAnswer)

  local uiHideDelay = 0.0
  local fadeOutAllAction = createFadeOutAllAction(self, uiHideDelay)

  local incorrectResponseAction = AHSequence.new(
    showExpectedAnswerAction,
    AHDelay.new(0.5),
    fadeOutAllAction
  )

  return incorrectResponseAction:start()
end

function LinusScene:startRationaleDismissAction()
  local backgroundAction = self.background:createZoomAction()

  local prompt = self.prompt
  prompt:prepareForShowAction(self)
  local promptAction = prompt:createShowAction()

  local numberPadAction = self.numberPad:createShowAction()

  local dismissAction = AHSpawn.new(
    function() Sound.playSoundEffect("tip_dismiss") end,
    self.currentRationale:createDismissAction(),
    backgroundAction,
    createShowUiAction(self),
    promptAction,
    numberPadAction
  )

  return dismissAction:start()
end

function LinusScene:startRationaleDismissIntoOutroAction()
  local dismissAction = AHSpawn.new(
    self.background:createZoomAction(),
    self.currentRationale:createDismissAction()
  )

  return dismissAction:start()
end

function LinusScene:reset()
  self.inputField:reset()
  self.diamondClusterLayer:reset()
  self.background:reset()
end

function LinusScene:reactToInvalidInput()
  self.inputField:reactToInvalidInput()
end

function LinusScene:showRationale(concept, onDismissPressed)
  local tipType = concept.tipType
  local minuend = concept.minuend
  local subtrahend = concept.subtrahend
  local answer = concept.answer

  local rationale = LinusRationale.new(tipType, minuend, subtrahend, answer, onDismissPressed)
  self:insertAtCenter(rationale)
  rationale:startShowAction()
  self.currentRationale = rationale

  Sound.playSoundEffect("tip_appear")
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, numberPad, centralSubmitButton)
  local background = LinusBackground.new()
  self:insertAtCenter(background)
  self.background = background

  local inputField = LinusInputField.new()
  inputField:startShadowTrackingCoroutine()
  self:insert(inputField)
  self.inputField = inputField

  centralSubmitButton:attachInputField(inputField)
  centralSubmitButton:setX(self:getWidthPart(0.5))
  self:insert(centralSubmitButton)
  self.centralSubmitButton = centralSubmitButton

  local numberPadLocationY = numberPad:getHeightPart(0.5) + Screen.safeBottom
  numberPad:setLoc(self:getWidthPart(0.5), numberPadLocationY)
  self:insert(numberPad)
  numberPad:setOnButtonPressCallback(createOnButtonPressCallback(self))
  numberPad:attachInputField(inputField)
  self.numberPad = numberPad

  local isHorizontal = LinusConfiguration.getIsHighLevel()
  local prompt = LinusPrompt.new(isHorizontal)
  local promptOffsetY = self:getHeightPart(0.09)
  local promptLocationY = Screen.safeTop - promptOffsetY
  prompt:setAnchorPointTop()
  prompt:setLoc(self:getWidthPart(0.5), promptLocationY)
  self:insert(prompt)
  self.prompt = prompt

  local inputFieldAreaHeight = prompt:getYBottom() - numberPad:getYTop()
  local inputFieldBaselineY = numberPad:getYTop() + (inputFieldAreaHeight * 0.5)

  inputField:setLoc(self:getWidthPart(0.5), inputFieldBaselineY)
  inputField:setScaleByAreaHeight(inputFieldAreaHeight)
  background:setGradientY(inputFieldBaselineY)
  centralSubmitButton:setY(inputFieldBaselineY)

  local diamondClusterLayer = LinusDiamondClusterLayer.new(inputFieldBaselineY)
  self:insertAtCenter(diamondClusterLayer)
  self.diamondClusterLayer = diamondClusterLayer

  local overExposure = LinusOverExposure.new()
  overExposure:setLoc(self:getWidthPart(0.5), inputFieldBaselineY)
  self:insert(overExposure)
  self.overExposure = overExposure

  local expectedAnswerLabel = Label.new("", Font.new(LinusInputField.FONT_FACE, LinusInputField.FONT_SIZE))
  expectedAnswerLabel:setLoc(self:getWidthPart(0.5), inputFieldBaselineY)
  expectedAnswerLabel:setScl(inputField:getScl())
  self:insert(expectedAnswerLabel)
  expectedAnswerLabel:setOpacity(LinusCommon.INVISIBLE_OPACITY)
  self.expectedAnswerLabel = expectedAnswerLabel
end

function LinusScene.new(numberPad, centralSubmitButton, uiElements)
  local self = Group.newFullScreen()
  Mixin.mix(self, LinusScene)

  self.hasUserInputtedNumber = false
  self.uiElements = uiElements

  createLayout(self, numberPad, centralSubmitButton)

  return self
end
