--- LinusTipPagesBase
-- @author Created by Jay Austria on November 4, 2015.

-- ===========================================================================
local PageButton = privateClass()
-- ===========================================================================


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_MARGINS_X = Screen.getWidthPart(0.1)
local BUTTON_STATES = {
  INACTIVE = "BUTTON_STATES.INACTIVE",
  ACTIVE = "BUTTON_STATES.ACTIVE",
  RECOMMENDED = "BUTTON_STATES.RECOMMENDED"
}
local BUTTON_COLORS = {
  [BUTTON_STATES.INACTIVE] = LinusCommon.COLORS.DIMMED_MAGENTA,
  [BUTTON_STATES.ACTIVE] = LinusCommon.COLORS.MAGENTA,
  [BUTTON_STATES.RECOMMENDED] = LinusCommon.COLORS.CYAN
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startPressAction(self)
  local pressAction = AHSequence.new(
    function() self.virtualButton:setTouchEnabled(false) end,
    AHScaleTo.new(self, 0.1, AHEaseSimpleInOut.new(), 2.0)
  )

  pressAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PageButton:setState(state)
  local color = BUTTON_COLORS[state]
  self:setColor(unpack(color))

  local shouldBeActive = (state == BUTTON_STATES.ACTIVE) or (state == BUTTON_STATES.RECOMMENDED)
  self.virtualButton:setTouchEnabled(shouldBeActive)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------


function PageButton.new(isFlipped, pressedCallback)
  local arrowSprite = Sprite.newWithName("tiparrow.png")
  local width, height = arrowSprite:getDims()

  local self = Group.new(width, height)
  Mixin.mix(self, PageButton)

  self:insertAtCenter(arrowSprite)

  if isFlipped then
    arrowSprite:setScl(-1.0, 1.0)
  end

  local oversizedDiameter = Screen.getWidthPart(0.2)

  local onPressWrapper = function()
    startPressAction(self)

    if pressedCallback then
      pressedCallback()
    end
  end

  local virtualButton = Button.addVirtualButton(self, oversizedDiameter, oversizedDiameter, onPressWrapper)
  self.virtualButton = virtualButton

  self:linkColorAttrToChildren()

  return self
end



-- ===========================================================================
class("LinusTipPagesBase")
-- ===========================================================================


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.width


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function delayedHidePageButtons(self)
  local hideAction = AHDelayed.new(0.2,
    AHSpawn.new(
      AHScaleTo.new(self.backButton, 0.1, nil, 0.0),
      AHScaleTo.new(self.nextButton, 0.1, nil, 0.0)
  ))

  hideAction:start()
end

local function showPageButtons(self)
  local hasPreviousPage = (self.currentPageIndex > 1)
  local backButtonState = nil

  if hasPreviousPage then
    backButtonState = BUTTON_STATES.ACTIVE
  else
    backButtonState = BUTTON_STATES.INACTIVE
  end
  self.backButton:setState(backButtonState)

  local hasNextPage = (self.currentPageIndex < #self.pageFunctions)
  local nextButtonState = nil
  if hasNextPage then
    nextButtonState = BUTTON_STATES.RECOMMENDED
  else
    nextButtonState = BUTTON_STATES.INACTIVE
  end
  self.nextButton:setState(nextButtonState)

  local showAction = AHSpawn.new(
    AHScaleTo.new(self.backButton, 0.2, AHEaseBackOut.new(2.5), 1.0),
    AHScaleTo.new(self.nextButton, 0.2, AHEaseBackOut.new(2.5), 1.0)
  )

  showAction:start()
end

local function startPage(self, pageIndex)
  local pageFunctions = self:_getPageFunctions()
  self.pageFunctions = pageFunctions

  local pageFunction = pageFunctions[pageIndex]

  local pageFunctionWrapper = function()
    local runningPageAction = pageFunction()

    self.currentPageIndex = pageIndex
    wait(runningPageAction)

    showPageButtons(self)
  end
  runInCoroutine(pageFunctionWrapper)
end

local function registerNextPressed(self)
  Sound.playSoundEffect("left_right_button")

  delayedHidePageButtons(self)

  local nextPageIndex = self.currentPageIndex + 1
  startPage(self, nextPageIndex)
end

local function registerBackPressed(self)
  Sound.playSoundEffect("left_right_button")

  delayedHidePageButtons(self)

  local previousPageIndex = self.currentPageIndex - 1
  startPage(self, previousPageIndex)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusTipPagesBase:_getPageFunctions()
  error("_getPageFunctions must be overriden by subclass")
end

function LinusTipPagesBase:removeCreatedElements()
  for _, element in ipairs(self.createdElements) do
    element:removeSelf()
  end

  self.createdElements = {}
end

function LinusTipPagesBase:createDismissAction()
  local dismissAction = AHSpawn.new()

  for _, element in ipairs(self.createdElements) do
    dismissAction:addAction(AHSequence.new(
      AHOpacityTo.new(element, 0.1, nil, LinusCommon.INVISIBLE_OPACITY)),
      AHRemove.new(element)
    )
  end

  return dismissAction
end

function LinusTipPagesBase:initialize()
  self.currentPageIndex = 1
  self.pages = self:_getPageFunctions()
end

function LinusTipPagesBase:startFirstPage()
  startPage(self, 1)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local backPressedCallback = function()
    registerBackPressed(self)
  end
  local isFlipped = true
  local backButton = PageButton.new(isFlipped, backPressedCallback)
  backButton:setLoc(Screen.safeLeft + BUTTON_MARGINS_X, self:getHeightPart(0.5))
  backButton:setScl(0)
  self:insert(backButton)
  self.backButton = backButton

  local nextPressedCallback = function()
    registerNextPressed(self)
  end
  isFlipped = false
  local nextButton = PageButton.new(isFlipped, nextPressedCallback)
  nextButton:setLoc(Screen.safeRight - BUTTON_MARGINS_X, self:getHeightPart(0.5))
  nextButton:setScl(0)
  self:insert(nextButton)
  self.nextButton = nextButton
end

function LinusTipPagesBase.new(tipPagesHeight, minuend, subtrahend, answer)
  local self = Group.new(WIDTH, tipPagesHeight)
  Mixin.mix(self, LinusTipPagesBase)

  self.createdElements = {}
  self.minuend = minuend
  self.subtrahend = subtrahend
  self.answer = answer

  createLayout(self)

  return self
end
