--- LinusValidator
-- @author Created by Jay Austria on October 20, 2015.
-- ConceptValidator for Linus.

class("LinusValidator")

require("LinusConfiguration")
require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAXIMUM_DIGIT_COUNT = LinusConfiguration.MAXIMUM_DIGIT_COUNT

local REQUIRED_KEYS = {
  'minuend',
  'subtrahend',
  'tipType'
}

local EXPECTED_TIP_TYPES = {
  "COUNTUP",
  "TAKEAWAY",
  "DOUBLE"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function hasSubtrahendDigitGreaterThanMinuend(minuend, subtrahend)
  while (subtrahend > 0) do
    local minuendDigit = minuend % 10
    local subtrahendDigit = subtrahend % 10

    if subtrahendDigit > minuendDigit then
      return true
    end

    minuend = math.floor(minuend / 10)
    subtrahend = math.floor(subtrahend / 10)
  end

  return false
end

local function validateValuesForCountup(self, minuend, subtrahend)
  local willRequireCountup = hasSubtrahendDigitGreaterThanMinuend(minuend, subtrahend)
  self:expectTrue(willRequireCountup, string.format("Concept does not require count up method with minuend %d and subtrahend %d", minuend, subtrahend))
end

local function validateValuesForTakeaway(self, minuend, subtrahend)
  local canSafeleyTakeaway = not hasSubtrahendDigitGreaterThanMinuend(minuend, subtrahend)
  self:expectTrue(canSafeleyTakeaway, string.format("Concept cannot be solved with takeaway method with minuend %d and subtrahend %d", minuend, subtrahend))
end

local function validateValuesForDouble(self, minuend, subtrahend)
  local subtrahendDoubled = subtrahend * 2
  self:expectTrue((minuend == subtrahendDoubled), string.format("Concept cannot be solved by doubling with minuend %d and subtrahend %d", minuend, subtrahend))
end

local function validateMinuendAndSubtrahendValues(self, minuend, subtrahend)
  self:expectTrue(minuend > subtrahend, "Minuend must be greater than subtrahend")
  self:expectTrue(#tostring(minuend) <= MAXIMUM_DIGIT_COUNT, string.format("Minuend must have a number of digits not exceeding %d. If this is intended, the MAXIMUM_DIGIT_COUNT must be increased.", MAXIMUM_DIGIT_COUNT))
  self:expectTrue(#tostring(subtrahend) <= MAXIMUM_DIGIT_COUNT, string.format("Subtrahend must have a number of digits not exceeding %d.  If this is intended, the MAXIMUM_DIGIT_COUNT must be increased.", MAXIMUM_DIGIT_COUNT))
end

local function validateValuesForTipType(self, concept)
  local tipType = concept.tipType

  local hasValidTipType = TableUtils.contains(EXPECTED_TIP_TYPES, tipType)
  self:expectTrue(hasValidTipType, "Unexpected or missing tipType")

  validateMinuendAndSubtrahendValues(self, concept.minuend, concept.subtrahend)

  if hasValidTipType then
    if (tipType == "COUNTUP") then
      validateValuesForCountup(self, concept.minuend, concept.subtrahend)
    elseif (tipType == "TAKEAWAY") then
      validateValuesForTakeaway(self, concept.minuend, concept.subtrahend)
    elseif (tipType == "DOUBLE") then
      validateValuesForDouble(self, concept.minuend, concept.subtrahend)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LinusValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateValuesForTipType(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function LinusValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, LinusValidator)

  return self
end
