--- ManslapGame
-- @author Created by alexeychernikov on March 31, 2014.

class("ManslapDistort")
require("utils.ShaderUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIND_UP_TIME = 0.5
local WIND_DOWN_TIME = WIND_UP_TIME

local DEFAULT_PARAMS =
{
  width = Screen.width,
  height = Screen.height,
  bigDistortion = 0.0,
  distortionMagnitude = 0.2,
  smallDistortion = 10,
  speed = 60.0,
  rollSpeed = 0.0
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function fillMissingParams(params)
  for paramName, defaulValue in pairs(DEFAULT_PARAMS) do
    params[paramName] = params[paramName] or defaulValue
  end
end

local function getVertexShaderSource()
  local source = [[
    attribute vec4 position;
    attribute vec2 uv;

    varying vec2 vUv;

    void main ()
    {
      gl_Position = position;
      vUv = uv;
    }
  ]]
  return source
end

local function getFragmentShaderSource()
  local source = [[
    varying MEDP vec2 vUv;

    uniform sampler2D object;

    uniform MEDP float time;
    uniform MEDP float distortionMagnitude;
    uniform MEDP float distortion;
    uniform MEDP float distortion2;
    uniform MEDP float speed;
    uniform MEDP float rollSpeed;


    // Start Ashima 2D Simplex Noise
    MEDP vec3 mod289(MEDP vec3 x) {
      return fract(x * (1.0 / 289.0)) * 289.0;
    }
    MEDP vec2 mod289(MEDP vec2 x) {
      return fract(x * (1.0 / 289.0)) * 289.0;
    }
    MEDP vec3 permute(MEDP vec3 x) {
      return fract(mod289(((x*34.0)+1.0)*x));
    }
    MEDP float snoise(MEDP vec2 v)
      {
      const MEDP vec4 C = vec4(0.211324865405187,  // (3.0-sqrt(3.0))/6.0",
                          0.366025403784439,  // 0.5*(sqrt(3.0)-1.0)",
                         -0.577350269189626,  // -1.0 + 2.0 * C.x",
                          0.024390243902439); // 1.0 / 41.0",
      MEDP vec2 i  = floor(v + dot(v, C.yy) );
      MEDP vec2 x0 = v -   i + dot(i, C.xx);

      MEDP vec2 i1;
      i1 = (x0.x > x0.y) ? vec2(1.0, 0.0) : vec2(0.0, 1.0);
      MEDP vec4 x12 = x0.xyxy + C.xxzz;
      x12.xy -= i1;

      i = mod289(i); // Avoid truncation effects in permutation",
      MEDP vec3 p = permute( permute( i.y + vec3(0.0, i1.y, 1.0 ))
        + i.x + vec3(0.0, i1.x, 1.0 ));

      MEDP vec3 m = max(0.5 - vec3(dot(x0,x0), dot(x12.xy,x12.xy), dot(x12.zw,x12.zw)), 0.0);
      m = m*m ;
      m = m*m ;

      MEDP vec3 x = 2.0 * fract(p * C.www) - 1.0;
      MEDP vec3 h = abs(x) - 0.5;
      MEDP vec3 ox = floor(x + 0.5);
      MEDP vec3 a0 = x - ox;

      m *= 1.79284291400159 - 0.85373472095314 * ( a0*a0 + h*h );

      MEDP vec3 g;
      g.x  = a0.x  * x0.x  + h.x  * x0.y;
      g.yz = a0.yz * x12.xz + h.yz * x12.yw;
      return 130.0 * dot(m, g);
    }

    // // End Ashima 2D Simplex Noise

    void main() {
      MEDP vec2 p = vUv;
      MEDP float ty = time * speed;
      // float ty = distortion * distortion;
      MEDP float yt = p.y - ty;

      // //smooth distortion
      MEDP float offset = snoise(vec2(yt * 3.0, 0.0)) * distortionMagnitude;
      // // boost distortion
      // offset = pow(offset * distortion, 3.0)/(1.0 + distortion);
      // //add fine grain distortion
      offset += snoise(vec2(yt * 50.0,0.0)) * distortion2 * 0.1;
      // //combine distortion on X with roll on Y
      gl_FragColor = texture2D(object, vec2(fract(p.x + offset), fract(p.y - time*rollSpeed)));
    }
  ]]
  return source
end

local function initEffect(self, params)
   local shader = MOAIShader.new()
  local vsh = getVertexShaderSource()
  local fsh = getFragmentShaderSource()
  shader:reserveUniforms(7)
  shader:declareUniformSampler(1, 'object', 1)
  shader:declareUniformFloat(2, 'time', 0.0)
  shader:declareUniformFloat(3, 'distortionMagnitude', params.distortionMagnitude)
  shader:declareUniformFloat(4, 'distortion', params.bigDistortion)
  shader:declareUniformFloat(5, 'distortion2', params.smallDistortion)
  shader:declareUniformFloat(6, 'speed', params.speed)
  shader:declareUniformFloat(7, 'rollSpeed', params.rollSpeed)

  shader:setVertexAttribute(1, 'position')
  shader:setVertexAttribute(2, 'uv')
  shader:load(vsh, fsh)

  local imageW, imageH = self.frameBuffer:getSize()
  local w, h = imageW/Host.contentScale, imageH/Host.contentScale

  local deck = MOAIGfxQuad2D.new()
  deck:setTexture(self.frameBuffer)
  deck:setRect(0.0, 0.0, w, h)
  deck:setUVRect(0.0, 1.0, 1.0, 0.0)

  self:setDeck(deck)
  self:setShader(shader)

  self.deck = deck

  return shader
end

local function setViewportSize(self, width, height)
  local scaledWidth = width * Host.contentScale
  local scaledHeight = height * Host.contentScale
  self.viewport:setSize(scaledWidth, scaledHeight)
  self.viewport:setScale(width, -height) -- HACK: Render-to-texture seems to produce a flipped texture

  self.width = scaledWidth
  self.height = scaledHeight
end

local function initRendering(self, params)
  local viewport = MOAIViewport.new()
  local layer = MOAILayer2D.new()
  layer:setViewport(viewport)

   local frameBuffer = MOAIFrameBufferTexture.new()
  frameBuffer:setClearColor(0, 0, 0, 0)
  frameBuffer:setRenderTable({layer})
  MOAIRenderMgr.setBufferTable({frameBuffer})

  self.viewport = viewport
  self.layer = layer
  self.frameBuffer = frameBuffer

  setViewportSize(self, params.width, params.height)
  frameBuffer:init(self.width, self.height)
end

local function setAttributesAmortization(self)
  local coefficient = self.distortionCoefficient

  self.shader:setAttr(3, coefficient * self.params.distortionMagnitude)
  self.shader:setAttr(4, coefficient * self.params.bigDistortion)
  self.shader:setAttr(5, coefficient * self.params.smallDistortion)
  self.shader:setAttr(6, coefficient * self.params.speed)
  self.shader:setAttr(7, coefficient * self.params.rollSpeed)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ManslapDistort:setRenderedGroup(group)
  if (self.group and self.group ~= group) then
    self.group:removeSelf()
  end

  self.group = group

  if (group ~= nil) then
    Screen.sort(group, 1, self.layer)
  end
end

function ManslapDistort:setSize(width, height)
  setViewportSize(self, width, height)
  self.frameBuffer:init(self.width, self.height)
  self.deck:setRect(0, 0, width, height)
end

function ManslapDistort:getDistortionCoefficient()
  return self.distortionCoefficient
end

function ManslapDistort:setDistortionCoefficient(value)
  self.distortionCoefficient = value
  setAttributesAmortization(self)
end

function ManslapDistort:setTime(t)
  self.shader:setAttr(2, t)
end

function ManslapDistort:setUndistorted()
  self:setDistortionCoefficient(0.0)
end

function ManslapDistort:windDown()
  AHNonAttributeTo.new(self, WIND_DOWN_TIME, self.getDistortionCoefficient, self.setDistortionCoefficient, nil, 0.0):start()
end

function ManslapDistort:windUp()
  AHNonAttributeTo.new(self, WIND_DOWN_TIME, self.getDistortionCoefficient, self.setDistortionCoefficient, nil, 1.0):start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function ManslapDistort.new(params)
  params = params or {}

  local self = Group.newFullScreen()
  fillMissingParams(params)
  Mixin.mix(self, ManslapDistort)

  initRendering(self, params)

  local shader = initEffect(self, params)
  self.shader = shader

  self.params = params
  self.distortionCoefficient = 0
  self.group = nil

  return self
end
