--- ManslapGame
-- @author Created by alexeychernikov on March 31, 2014.


class("ManslapGameManager")

require("ManslapRoundGenerator")
require("bonuses.ResponseSpeedTracker")
require("ManslapConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MIN_TIME_TO_READ_CORRECTION = 0.75

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function disableInteraction(self)
  self.scene:setAnswerSubmittedCallback(nil)
end

local function loseGameImmediately(self)
  self.baseGame:showDefaultGameLossOutro()
  AHDelayed.new(
    0.7,
    function()
      GameSession.notifyIncorrectAnswer(self.round.identifier)
      GameSession.notifyGameLoss()
    end
  ):start()
end

local function loseGameAfterRationale(self)
  self.baseGame:showDefaultGameLossOutro()
  GameSession.notifyIncorrectAnswer(self.round.identifier)
  GameSession.notifyGameLoss()
end

local function setLives(self, lives)
  self.lives = lives
  if (self.lives <= 0) then
    disableInteraction(self)
    if (not self.showingRationale) then
      loseGameImmediately(self)
    end
  end

  self.livesView:setLives(self.lives)
end

local function loseLife(self)
  setLives(self, self.lives - 1)
end

local function getSoundTime(self)
  return self.audioTrack:getTime()
end

local function win(self)
  runInCoroutine(function()
    disableInteraction(self)
    waitWhile(function() return self.playing end)

    ScoringUtils.applyLivesBonus(self.lives, ManslapConfiguration.POINTS_PER_LIVES_BONUS, self.baseGame.bonusTray)
    self.scoreView:setScore(GameSession.getScore())

    self.baseGame:showDefaultGameWinOutro()
    local outro = self.scene:doVictoriousOutro()
    wait(outro)

    AHDelayed.new(
      0.6,
      function()
        GameSession.notifyCorrectAnswer(self.round.identifier)
        GameSession.notifyGameWin()
      end
    ):start()
  end)
end

local function getSlapTimeForIndex(self, index)
  return tonumber(self.round.correctionTimes[index].beginTime)
end

local function getCorrectionTimeoutTime(self, index)
  return tonumber(self.round.correctionTimes[index].endTime)
end

local function getRewindTime(self, index)
  return tonumber(self.round.correctionTimes[index].rewindToTime)
end

local function increaseCorrectionIndex(self)
  self.currentCorrectionIndex = math.min(self.currentCorrectionIndex + 1, #self.round.correctionTimes)
  self.currentCorrectionTimeoutTime = getCorrectionTimeoutTime(self, self.currentCorrectionIndex)
end

local function submitCorrectAnswer(self, index)
  if (self.acceptedAnswerIndices[index]) then
    return
  end
  if (index >= #self.round.correctionTimes) then
    win(self)
  end

  local isAfterATimeout = (self.timedOutIndices[index] ~= nil)
  if (not isAfterATimeout) then
    GameSession.incrementCorrectAnswerCount()
    GameSession.contentTracker:notifyCorrectAnswer(self.round.identifier, tostring(index))
  end

  increaseCorrectionIndex(self)

  self.scene:exitBossMode()
  self.scene:doCorrectAnswerAnimation(index)
  self.acceptedAnswerIndices[index] = true

  GameSession.increaseScore(self.scorePerCorrectAnswer)
  self.scoreView:setScore(GameSession.getScore())

  self.scene:displayCorrectionInformation(self.round.correctionTimes[index], index, true)

  AHDelayed.new(
    1.0,
    function()
      self.scene:hideCorrectionInformation()
    end
  ):start()
end

local function getSoundTimeFormattedString(self)
  local time = getSoundTime(self)
  local timeString = string.format("%.1f", time)

  return timeString
end

local function submitIncorrectAnswer(self, index)
  if (not self.acceptedAnswerIndices[index]) then
    local isAfterATimeout = (self.timedOutIndices[index] ~= nil)
    if (not isAfterATimeout) then
      GameSession.incrementIncorrectAnswerCount()
      GameSession.contentTracker:notifyIncorrectAnswer(self.round.identifier, tostring(index), getSoundTimeFormattedString(self))
    end

    loseLife(self)
  end
end

local function initEventHandlers(self)
  self.scene:setAnswerSubmittedCallback(function()
    local willProcessSubmit = ((self.round ~= nil) and (not self.audioTrack:isPaused()))
    if (willProcessSubmit) then
      local correctionIndex = self.currentCorrectionIndex
      local submitTime = getSoundTime(self)
      local slapStartTime = getSlapTimeForIndex(self, correctionIndex)
      local slapEndTime = getCorrectionTimeoutTime(self, correctionIndex)
      local alreadyPassed = self.timedOutIndices[correctionIndex]
      local answerSubmittedOnTime = (submitTime > slapStartTime) and (submitTime < slapEndTime)

      if (alreadyPassed or answerSubmittedOnTime) then
        -- Don't allow speed bonuses if we've already shown the user the answer
        if (not alreadyPassed) then
          local responseDuration = submitTime - slapStartTime
          self.responseSpeedTracker:notifyCorrectResponseWithValue(responseDuration)
        end

        submitCorrectAnswer(self, correctionIndex)
      else
        if (not self.scene:isFocused()) then
          self.scene:doIncorrectAnswerAnimation()
          submitIncorrectAnswer(self, correctionIndex)
        end
      end
    end
  end)

  self.showingRationale = false
  self.scene:setResumeAfterRationaleCallback(function()
    self.showingRationale = false
  end)
end

local function configureGameDifficulty(self, round)
  local totalQuestions = #round.correctionTimes
  self.scorePerCorrectAnswer = ManslapConfiguration.TOTAL_POINTS_FOR_GAME / totalQuestions
  self.pointsPerFullSpeedBonus = ManslapConfiguration.MAX_POINTS_FROM_ALL_SPEED_BONUSES / totalQuestions

  local rankTwoPoints = ManslapConfiguration.getRankTwoScore()
  local rankThreePoints = ManslapConfiguration.getRankThreeScore()
  GameSession.setScoreThresholds(rankTwoPoints, rankThreePoints)

  GameSession.setInitialLives(ManslapConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

local function initGame(self)
  self.acceptedAnswerIndices = {}
  self.timedOutIndices = {}

  setLives(self, self.livesView:getLives())

  self.currentCorrectionIndex = 0
  increaseCorrectionIndex(self)
  configureGameDifficulty(self, self.round)
end

local function doTimeout(self)
  local correctionIndex = self.currentCorrectionIndex

  if (self.timedOutIndices[correctionIndex]) then
    submitCorrectAnswer(self, correctionIndex)
    return
  end

  GameSession.contentTracker:notifyTimeout(self.round.identifier, tostring(correctionIndex))

  self.showingRationale = true
  self.timedOutIndices[correctionIndex] = true
  submitIncorrectAnswer(self, correctionIndex)

  self.scene:doTimeoutAnswerAnimation(correctionIndex)
  self.scene:enableBossMode()
  local wasAnsweredCorrectly = false
  local gameHasEnded = (self.lives <= 0)
  self.scene:displayCorrectionInformation(self.round.correctionTimes[correctionIndex], correctionIndex, wasAnsweredCorrectly, gameHasEnded)

  self.audioTrack:pause()
  waitFor(MIN_TIME_TO_READ_CORRECTION)

  self.scene:setErrorButtonEnabled(false)
  waitWhile(function() return self.showingRationale end)
  if (gameHasEnded) then
    loseGameAfterRationale(self)
  end
  self.scene:setErrorButtonEnabled(true)

  self.scene:setNeutralStateForCorrectionInformation()

  Sound.playSoundEffect("precision_rewind", 0.2)

  self.scene:relaxTimeoutReaction()
  self.scene:doRewindAnimation()
  -- TODO: wait for rewind sound to play through
  waitFor(1)

  local rewindTarget = getRewindTime(self, correctionIndex)
  self.lastRewindTarget = rewindTarget
  self.audioTrack:seek(rewindTarget)
  self.audioTrack:resume()

  self.scene:stopRewindAnimation()
end

local function tryToDoTimeout(self, time)
  if ((time > self.currentCorrectionTimeoutTime) or self.audioTrack:isFinished()) then
    doTimeout(self)
  end
end

local function areAllAnswersPassed(self)
  local atLastCorrection = self.currentCorrectionIndex >= #self.round.correctionTimes
  local lastCorrectionApplied = self.acceptedAnswerIndices[#self.round.correctionTimes]
  return atLastCorrection and lastCorrectionApplied
end

local function trackCurrentCorrection(self)
  runInCoroutine(function()
    while(true) do
      if (areAllAnswersPassed(self)) then
        return
      end

      local time = self.audioTrack:getTime()
      tryToDoTimeout(self, time)
      coroutine.yield()
    end
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ManslapGameManager:startGame()
  self.round = self.roundGenerator:next()
  
  GameSession.contentTracker:notifyPresentedConcept(self.round.identifier)
  initGame(self)
  self.audioTrack = AssetHelper.playSoundForConcept(self.round, self.round.sound)

  self.playing = true
  self.audioTrack:addFinishedHandler(function() self.playing = false end)

  trackCurrentCorrection(self)

  self.scene:setRound(self.round, self.audioTrack)
  self.scene:doIntro()
end

-- ----------------------------------------------------------------------------
-- Object construction
-- ----------------------------------------------------------------------------

function ManslapGameManager.new(scene, scoreView, livesView, baseGame)
  local self = {}
  Mixin.mix(self, ManslapGameManager)

  self.scene = scene
  self.scoreView = scoreView
  self.livesView = livesView
  self.baseGame = baseGame
  self.roundGenerator = ManslapRoundGenerator.new()

  local bonusWindowSize = ManslapConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS
  local function onQuickResponse(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 self.pointsPerFullSpeedBonus,
                                 self.baseGame.bonusTray)
  end
  local bonusDecay = ManslapConfiguration.SPEED_BONUS_DECAY_START
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowSize, onQuickResponse, bonusDecay)

  initEventHandlers(self)

  return self
end
