--- ManslapGame
-- @author Created by alexeychernikov on March 31, 2014.

class("ManslapParticleBall")
require("ManslapParticleSpeck")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_SETS =
{
  {125/255, 249/255, 201/255},
  {18/255, 173/255, 138/255},
  {0/255, 122/255, 105/255},
  {0/255, 74/255, 76/255},
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createParticle(self)
  local color = TableUtils.getRandomElement(COLOR_SETS)

  local radius = self.minRadius + math.random() * self.radiusAmp
  local particle = ManslapParticleSpeck.new(color, radius)

  return particle
end

local function addParticle(self, particle, time)
  particle.birthday = time
  self.particles[particle] = true
  self.particleCount = self.particleCount + 1
  if (particle.parent ~= self) then
    self:insertAtCenter(particle)
  end
end

local function randomXYForCircle(radiusAmp)
  local angle = math.random() * 360
  local R = math.random() * radiusAmp
  local x = math.cos(angle) * R
  local y = math.sin(angle) * R
  local xSign, ySign = MathUtils.randomSign(), MathUtils.randomSign()
  x = x * xSign
  y = y * ySign
  return x, y
end

local function randomInRange(min, amp)
  local value = min + math.random() * amp
  return value
end

local function initializeParticle(self, particle)
  particle.xSpeed, particle.ySpeed = randomXYForCircle(self.speedAmp)
  particle.xAcceleration = randomInRange(self.minAcceleration, self.accelerationAmp)
  particle.yAcceleration = randomInRange(self.minAcceleration, self.accelerationAmp)

  local scaleSpeed = math.random()
  local scaleStartTime = self.particlesLifeTime * (math.random()/10 + 0.10)
  local decelerateStartTime = 0.1
  particle.xScaleSpeed, particle.yScaleSpeed = scaleSpeed, scaleSpeed

  particle.scaleStartTime = scaleStartTime
  particle.decelerateStartTime = decelerateStartTime
end

local function spawnParticlesImmediately(self, particlesToSpawn, time)
  if (self.particlePool) then
    for particle, _ in pairs(self.particlePool) do
      addParticle(self, particle, time)
      initializeParticle(self, particle)
      particlesToSpawn = particlesToSpawn - 1
      self.particlePool[particle] = nil
      particle.isVisible = true
    end
  end
  for i = 1, particlesToSpawn do
    local particle = createParticle(self, time)
    addParticle(self, particle, time)
    initializeParticle(self, particle)
  end
end

local function spawnParticlesByBurst(self, time)
  if (time - self.lastBurstTime < self.timePerBurst) then
    return
  end

  if (self.burstCount >= self.maxBursts) then
    return
  end

  self.burstCount = self.burstCount + 1

  self.lastBurstTime = time
  spawnParticlesImmediately(self, self.particlesPerBurst, time)
end

local function stashParticleForLater(self, particle)
  if (self.particlePool) then
    self.particlePool[particle] = true
    particle.isVisible = false
  else
    particle:removeSelf()
  end
end

local function killOneParticle(self, particle)
  self.particles[particle] = nil
  self.particleCount = self.particleCount - 1
  stashParticleForLater(self, particle)
end

local function killParticles(self, time)
  local timeToDie = self.particlesLifeTime
  for particle, _ in pairs(self.particles) do
    local particleLife = time - particle.birthday
    if (particleLife >= timeToDie) then
      killOneParticle(self, particle)
    end
  end
end

local function applyParticleMotion(self, particle, dt, age)
  local x, y = particle:getLoc()
  local dx, dy = particle.xSpeed * dt, particle.ySpeed * dt
  local dxSpeed, dySpeed = 1 + particle.xAcceleration * dt, 1 + particle.yAcceleration * dt

  particle.xSpeed, particle.ySpeed = particle.xSpeed * dxSpeed, particle.ySpeed * dySpeed

  if (age > particle.scaleStartTime) then
    local scaleTime = (age - particle.scaleStartTime)/(self.particlesLifeTime - particle.scaleStartTime)
    particle:setScl(1 - scaleTime)
  end
  if (age > particle.decelerateStartTime) then
    particle.xAcceleration, particle.yAcceleration = self.decelerationRate, self.decelerationRate
  end

  local xTurbulence, yTurbulence = self.turbulence * math.random() * 2 - self.turbulence, self.turbulence * math.random() * 2 - self.turbulence
  particle:setLoc(x + dx + xTurbulence, y + dy + yTurbulence)
end

local function moveParticles(self, time, dt)
  for particle, _ in pairs(self.particles) do
    local age = time - particle.birthday
    applyParticleMotion(self, particle, dt, age)
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ManslapParticleBall:start(runTime, onComplete)
  runTime = runTime or self.particlesLifeTime
  local prevTime = 0
  self.lastBurstTime = -self.timePerBurst
  self.burstCount = 0

  local particlesTimestep = function(self, t)
    local dt = t - prevTime
    prevTime = t
    killParticles(self, t)

    spawnParticlesByBurst(self, t)
    moveParticles(self, t, dt)
  end

  local getPreviousTime = function()
    return prevTime
  end

  local finishEffect = nil
  if (onComplete) then
    finishEffect = function() onComplete(self) end
  end

  local particleAction = AHNonAttributeTo.new(self, runTime, getPreviousTime, particlesTimestep, nil, runTime)

  return AHSequence.new(
    particleAction,
    finishEffect
  ):start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function ManslapParticleBall.new(speed, acceleration, innerRadius)
  local self = Group.newFullScreen()
  Mixin.mix(self, ManslapParticleBall)

  self.particlesLifeTime = 1

  -- It's either particles per burst, or per second
  self.particlesPerBurst = 5
  self.timePerBurst = 0.01
  self.lastBurstTime = -self.timePerBurst
  self.particlesPerSecond = 30
  self.maxBursts = 8

  self.particleCount = 0
  self.particleSpawnRemainder = 0
  self.minRadius, self.maxRadius = 4, 6
  self.minSpeed, self.maxSpeed = 30 * speed, 60 * speed
  self.innerRadius = innerRadius
  self.minAcceleration, self.maxAcceleration = acceleration, acceleration

  self.turbulence = 0.8
  self.turbulenceStart = self.particlesLifeTime/2
  self.decelerationRate = -2 * self.minAcceleration * 0.8

  -- Derived parameters
  self.radiusAmp = self.maxRadius - self.minRadius
  self.speedAmp = self.maxSpeed - self.minSpeed
  self.accelerationAmp = self.maxAcceleration - self.minAcceleration

  self.particles = {}

  return self
end
