--- ManslapGame
-- @author Created by alexeychernikov on March 31, 2014.


class("ManslapScene")
require("ManslapColors")
require("ManslapStateProgressBar")
require("ManslapBossModeBackground")
require("ManslapCorrectionInformationDisplay")
require("ManslapSoundVisualizer")
require("ManslapWaveUtils")
require("ManslapRewindEffectGroup")
require("ManslapParticleBall")
require("ManslapParticleRow")
require("ManslapErrorButton")
require("ManslapRationaleTextArea")
require("ManslapTapToRewindLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INSTRUCTIONS_TEXT = "Tap the error button when you hear a mistake."

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createProgressVisualizer(self, round, sound)
  local visualizer = ManslapSoundVisualizer.new(ManslapColors.RAGE_LINE_COLOR)
  visualizer:setX(Screen.getSafeX(0.5))
  visualizer:setY(Screen.getSafeY(0.45))
  self.mainVisuals:insert(visualizer)
  self.visualizer = visualizer

  local progressMeter = ManslapStateProgressBar.new(#round.correctionTimes, ManslapColors.PROGRESS_BAR_COLORS_PASSIVE)
  progressMeter:hide()
  progressMeter:setX(Screen.getSafeX(0.5))
  progressMeter:setY(Screen.getSafeY(0.9))
  self.mainVisuals:insert(progressMeter)
  self.progressMeter = progressMeter

  visualizer:visualizeSound(sound)

  self.correctionY = (UIOverlay.NORTH.y + visualizer:getY()) / 2
end

local function flashBackground(self, color)
  if (self.flashAnimation) then
    self.flashAnimation:stop()
    self.flashAnimation = nil
  end

  local endColor = {unpack(ManslapColors.BACKGROUND_COLOR)}
  endColor[4] = 0
  self.flash:setColor(unpack(color))
  local animTime = 0.3
  local delay = AHDelay.new(0.3)
  local fadeOut = AHColorTo.new(self.flash, animTime, nil, endColor)
  local animation = AHSequence.new(delay, fadeOut)
  animation:start()

  self.flashAnimation = animation
end

local function setUIOverlayWidgetsColor(self, color)
  local uiOverlayWidgets = self.uiOverlay:getWidgets()
  for i, widget in ipairs(uiOverlayWidgets) do
    widget:setColor(unpack(color))
  end
end

local function waitWhileExplanationIsOnScreen(self)
  while (self.isInFocusMode) do
    coroutine.yield()
  end
end

local function showOutroCheckMarkAndParticles(self)
  local numberOfPoints = self.visualizer:getTotalPointCount()
  local checkmarkWidth, checkmarkHeight = 100, 80
  local checkmarkShape = ManslapWaveUtils.getCheckmarkShape(numberOfPoints, checkmarkWidth, checkmarkHeight)
  self.visualizer:morph(checkmarkShape)
  self.visualizer:setColor(unpack(ManslapColors.CORRECT_ANSWER_LINE_COLOR))

  local shrinkAmount = 0.8
  local scaleEase = MOAIEaseType.EASE_IN
  local scaleDown = AHScaleTo.new(self.visualizer, 0.3, scaleEase, shrinkAmount, shrinkAmount)
  local scaleUp = AHScaleTo.new(self.visualizer, 0.6, scaleEase, 1.25, 1.25)
  local scaleBounce = AHSequence.new(scaleDown, scaleUp)
  local moveUp = AHLocationYTo.new(self.visualizer, 0.5, AHEaseSineInOut.new(), Screen.getHeightPart(0.5))
  local checkTransform = AHSpawn.new(scaleBounce, moveUp)
  checkTransform:start()

  local visualizerX = self.visualizer:getX()
  local progressBarX = self.progressMeter:getX()
  local progressBarDX = visualizerX - progressBarX
  local progressBarDY = Screen.getHeightPart(0.5)
  local stagger = 0.03
  local progressBarTranslationTime = 0.3
  local shrinkProgressBar = AHScaleTo.new(self.progressMeter, 0.5, scaleEase, 0.8, 0.8)
  local moveProgressBar = self.progressMeter:getStaggeredSuckInAction(progressBarDX, progressBarDY, stagger, progressBarTranslationTime)
  local progressBarAnimation = AHSpawn.new(shrinkProgressBar, moveProgressBar)
  progressBarAnimation:start()

  local speed = 12
  local acceleration = 8
  local radius = Screen.width
  local fireworks = ManslapParticleBall.new(speed, acceleration, radius)

  fireworks:setLoc(Screen.getWidthPart(0.5) - checkmarkWidth * 0.1, checkmarkHeight * 0.43)
  self.visualizer:insert(fireworks)

  local particleBurstSeconds = 1
  return runInCoroutine(function()
      waitFor(0.3)
      wait(fireworks:start(particleBurstSeconds))
      fireworks:removeSelf()
  end)
end

local function stopInstructionAnimations(self)
  if (self.instructions.animation) then
    self.instructions.animation:stop()
    self.instructions.animation = nil
  end
end

local function animateInstructionsColor(self, color, time)
  time = time or 0.4
  stopInstructionAnimations(self)
  local showInstructions = AHColorTo.new(self.instructions, time, nil, color)
  showInstructions:start()
  self.instructions.animation = showInstructions
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ManslapScene:doTimeoutAnswerAnimation(index)
  self:hideInstructions()

  self.progressMeter:fail(index)
  self.failedIndex = index
  flashBackground(self, ManslapColors.INCORRECT_ANSWER_BACKGROUND_FLASH_COLOR)
  AHSmoothShake.new(self.flash, 40, 40):start()

  Sound.playSoundEffect("precision_timeout", 1)
  runInCoroutine(function()
    self.visualizer:hide()
    self.errorButton:hide()
    self.progressMeter:hide()

    self.visualizer:setPenaltyColor(ManslapColors.FLATLINE_COLOR)
  end)
end

function ManslapScene:doIncorrectAnswerAnimation()
  self:hideInstructions()

  flashBackground(self, ManslapColors.INCORRECT_ANSWER_BACKGROUND_FLASH_COLOR)
  Sound.playSoundEffect("precision_incorrect", 0.1)
  runInCoroutine(function()
    self.visualizer:setPenaltyColor(ManslapColors.RAGE_LINE_COLOR)
    self.visualizer:showPenaltyAnimation()
    -- Let the penalty animation play for a bit
    waitFor(0.2)
    -- Then fade it out
    self.visualizer:showRestoreAnimation()
  end)
end

function ManslapScene:relaxTimeoutReaction()
  self.visualizer:showRestoreAnimation()
end

function ManslapScene:doRewindAnimation()
  self.bossModeBackground:enterRewindMode()
  self.rewindEffectGroup:start()

  self:hideRationale()
end

function ManslapScene:stopRewindAnimation()
  self.rewindEffectGroup:stop()
  self.bossModeBackground:enterNormalMode()
  self.progressMeter:neutralize(self.failedIndex)
  self.failedIndex = nil

  self.visualizer:reappear()
  self.progressMeter:reappear()
  self.errorButton:showBlackOutline()
end

function ManslapScene:doCorrectAnswerAnimation(index)
  self:hideInstructions()

  Sound.playSoundEffect("precision_correct", 0.3)

  self.progressMeter:complete(index)
  self.visualizer:showRewardAnimation()

  local _, visualiserY = self.visualizer:getLoc()
  local speed = 4
  local acceleration = 2
  local particlesX = Screen.getWidthPart(0.5)
  local particlesY = visualiserY - self.visualizer:getHeightPart(0.5)
  local particles = ManslapParticleRow.new(speed, acceleration)
  self:insertAtLocation(particles, particlesX, particlesY)

  local time = 1.5
  particles:start(time)

  self.errorButton:showGreen()
  runInCoroutine(function()
    waitFor(1.1)
    self.errorButton:showWhiteOutline()
  end)
end

function ManslapScene:setRound(round, sound)
  createProgressVisualizer(self, round, sound)
end

function ManslapScene:displayCorrectionInformation(information, index, wasAnsweredCorrectly, gameHasEnded)
  self:hideInstructions()

  if (self.correctionInformation) then
    self.correctionInformation:setCorrectionState(ManslapColors.CORRECT_ANSWER_TEXT_COLOR)
  else
    self.correctionInformation = ManslapCorrectionInformationDisplay.new(information, wasAnsweredCorrectly, self.correctionY)
  end
  self.rewindEffectGroup:setGroupToDistort(self.correctionInformation)

  if (wasAnsweredCorrectly) then
    self.progressMeter:focus(index)
    self.focusIndex = index
    self.visualizer:setColor(unpack(ManslapColors.CORRECT_ANSWER_LINE_COLOR))
    self.isInFocusMode = true
  else
    AHSmoothShake.new(self.correctionInformation, 30, 30, 14):start()

    local rationaleText = information["tip"]
    self:displayRationale(rationaleText, gameHasEnded)
  end
end

function ManslapScene:setNeutralStateForCorrectionInformation()
  if (self.correctionInformation) then
    self.correctionInformation:setNeutralState()
  end
end

function ManslapScene:hideCorrectionInformation()
  if (self.correctionInformation) then
    self.correctionInformation:removeSelf()
    self.correctionInformation = nil
  end
  if (self.focusIndex) then
    self.progressMeter:removeFocus(self.focusIndex)
    self.focusIndex = nil
    -- Wait for the focus animation to play
    AHDelayed.new(
      0.3,
      function()
        self.visualizer:setColor(unpack(ManslapColors.NEUTRAL_LINE_COLOR))
        self.isInFocusMode = false
      end
    ):start()
  end
end

function ManslapScene:displayRationale(rationaleText, gameHasEnded)
  local width = Screen.getWidthPart(0.82)
  local rationaleTextArea = ManslapRationaleTextArea.new(rationaleText, width)
  rationaleTextArea:setX(Screen.getSafeX(0.5))
  rationaleTextArea:setY(Screen.getSafeY(0.34))
  self:insert(rationaleTextArea)
  self.rationaleTextArea = rationaleTextArea

  rationaleTextArea:showAppearAnimation()
  self.tapToRewindLabel:showAppearAnimation(gameHasEnded)
end

function ManslapScene:hideRationale()
  if (self.rationaleTextArea ~= nil) then
    self.rationaleTextArea:removeSelf()
  end

  self.tapToRewindLabel:setOpacity(0.0)
end

function ManslapScene:isFocused()
  return self.isInFocusMode
end

function ManslapScene:enableBossMode()
  self:hideInstructions()

  self.bossModeBackground:enable()
  setUIOverlayWidgetsColor(self, ManslapColors.BOSS_MODE_UI_COLOR)
  self.progressMeter:setIndicatorColors(ManslapColors.PROGRESS_BAR_COLORS_BOSS)
  self.visualizer:setColor(unpack(ManslapColors.VISUALIZER_COLORS_BOSS))
end

function ManslapScene:exitBossMode()
  self.bossModeBackground:disable()
  setUIOverlayWidgetsColor(self, ManslapColors.UI_COLOR)
  self.progressMeter:setIndicatorColors(ManslapColors.PROGRESS_BAR_COLORS_PASSIVE)
  self.visualizer:setColor(unpack(ManslapColors.VISUALIZER_COLORS_PASSIVE))
end

function ManslapScene:doVictoriousOutro()
  self.errorButton:hide()

  return runInCoroutine(function()
    local outroSoundEffectVolume = 0.7
    Sound.playSoundEffect("precision_goal_reached", outroSoundEffectVolume)

    waitWhileExplanationIsOnScreen(self)
    local animation = showOutroCheckMarkAndParticles(self)
    wait(animation)
  end)
end

function ManslapScene:doIntro()
  local lineCover = FilledRectangle.newFullScreen(ManslapColors.BACKGROUND_COLOR)
  self.visualizer:insertAtCenter(lineCover)

  local x0 = Screen.width - self.visualizer:getX()
  self.flash:setVisible(false)

  local slideIn = AHLocationXTo.new(lineCover, 0.8, nil, x0 + Screen.width)
  local animation = AHSequence.new(
    slideIn,
    function()
      lineCover:removeSelf()
      self.visualizer:bringToFront()
      self.flash:setVisible(true)

      self.errorButton:showAppearAnimation()
      self:setErrorButtonEnabled(true)
    end
  )
  animation:start()

  self.progressMeter:showAppearAnimation()
  self:showInstructions()
end

function ManslapScene:showInstructions()
  if (self.areInstructionsHidden) then
    animateInstructionsColor(self, ManslapColors.VISIBLE_TAP_WHEN_ERROR_INSTRUCTIONS_COLOR)
    self.areInstructionsHidden = false
  end
end

function ManslapScene:hideInstructions()
  if (not self.areInstructionsHidden) then
    animateInstructionsColor(self, ManslapColors.INVISIBLE_TAP_WHEN_ERROR_INSTRUCTIONS_COLOR, 0.08)
    self.areInstructionsHidden = true
  end
end

function ManslapScene:setAnswerSubmittedCallback(callback)
  self.answerSubmittedCallback = callback
end

function ManslapScene:setResumeAfterRationaleCallback(callback)
  self.resumeAfterRationaleCallback = callback
end

function ManslapScene:setErrorButtonEnabled(enabled)
  self.errorButton:setTouchEnabled(enabled)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local mainVisuals = Group.newFullScreen()
  local bossModeBackground = ManslapBossModeBackground.new()
  local rewindEffectGroup = ManslapRewindEffectGroup.new()
  local flash = FilledRectangle.newFullScreen({0, 0, 0, 0})

  local instructionsFont = Font.new(FontFaces.LIGHT_FONT, 22)
  local instructionsWidth = Screen.getWidthPart(0.75)
  local instructions = TextArea.new(INSTRUCTIONS_TEXT, instructionsFont, instructionsWidth, Alignment.CENTER)
  instructions:setColor(unpack(ManslapColors.INVISIBLE_TAP_WHEN_ERROR_INSTRUCTIONS_COLOR))
  self.areInstructionsHidden = true

  local tapToRewindLabel = ManslapTapToRewindLabel.new()

  local errorButton = ManslapErrorButton.new()

  self:insertAtCenter(bossModeBackground)
  self:insertAtCenter(flash)
  self:insertAtCenter(mainVisuals)
  self:insertAtCenter(rewindEffectGroup)
  self:insertAtLocationFractional(instructions, 0.5, 0.8)
  self:insertAtLocationFractional(tapToRewindLabel, 0.5, 0.08)
  self:insertAtLocationFractional(errorButton, 0.5, 0.12)

  bossModeBackground:disable()
  self.bossModeBackground = bossModeBackground
  self.mainVisuals = mainVisuals
  self.flash = flash
  self.rewindEffectGroup = rewindEffectGroup
  self.instructions = instructions
  self.tapToRewindLabel = tapToRewindLabel
  self.errorButton = errorButton

  Button.makeButton(errorButton, function()
    if (self.answerSubmittedCallback ~= nil) then
      self.answerSubmittedCallback()
    end
  end)
  self:setErrorButtonEnabled(false)

  Button.makeButton(self, function()
    if (self.resumeAfterRationaleCallback ~= nil) then
      self.resumeAfterRationaleCallback()
    end
  end)
end

function ManslapScene.new(uiOverlay)
  local self = FilledRectangle.newFullScreen(ManslapColors.BACKGROUND_COLOR)

  Mixin.mix(self, ManslapScene)
  createLayout(self)

  self.uiOverlay = uiOverlay

  return self
end
