--- ManslapValidator
-- @author Created by Anton Morant on April 25, 2014.
-- ConceptValidator for Manslap.

class("ManslapValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

require("utils.XmlParser")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "sound",
  "correctionTimes"
}

local CORRECTION_TIMES_FIELDS = {
  "beginTime",
  "endTime",
  "rewindToTime",
  "incorrect",
  "correct",
  "tip"
}

local RATIONALE_TIP_TEXT_REQUIRED_PAIRED_TAGS = {
  "correct",
  "incorrect"
}

local RATIONALE_TIP_TEXT_OPTIONAL_PAIRED_TAGS = {
  "i"
}

local CORRECTION_TIMES_EMPTY_ALLOWED_FIELDS = {
  "preContext",
  "postContext"
}

local MIN_CORRECTION_TIMES = 5
local MAX_CORRECTION_TIMES = 6

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateCorrectionFields(self, correction)
  self:expectHasKeys(correction, CORRECTION_TIMES_FIELDS)
  self:expectHasKeys(correction, CORRECTION_TIMES_EMPTY_ALLOWED_FIELDS, true)
end

local function validateRationaleTipText(self, correction)
  local tipText = correction["tip"]

  self:expectNoError(function()
    XmlParser.parse(tipText)
  end)
  self:validateTags(tipText, nil, RATIONALE_TIP_TEXT_REQUIRED_PAIRED_TAGS, nil, RATIONALE_TIP_TEXT_OPTIONAL_PAIRED_TAGS)
end

local function validateContext(self, correction)
  local preContextIsEmtpy = correction.preContext == ""
  local postContextIsEmtpy = correction.postContext == ""
  local errorMessage = string.format("At least one of preContext, postContext must not be empty (correct: %s)", correction.correct)
  self:expectFalse(preContextIsEmtpy and postContextIsEmtpy, errorMessage)
end

local function validateTimes(self, correction)
  local beginTime = tonumber(correction.beginTime)
  local endTime = tonumber(correction.endTime)
  local rewindToTime = tonumber(correction.rewindToTime)
  self:expectLessThan(beginTime, endTime, "beginTime should be < endTime")
  self:expectLessThan(rewindToTime, beginTime, "rewindToTime should be < beginTime")
end

local function validateText(self, correction)
  self:expectCapitalizationMatches(correction.correct, correction.incorrect, "correct", "incorrect")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ManslapValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local correctionTimes = concept.correctionTimes
  self:expectNumberInRange(#correctionTimes, MIN_CORRECTION_TIMES, MAX_CORRECTION_TIMES, "number of correctionTimes")

  for _, correction in ipairs(correctionTimes) do
    validateCorrectionFields(self, correction)
    validateContext(self, correction)
    validateTimes(self, correction)
    validateText(self, correction)
    validateRationaleTipText(self, correction)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ManslapValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, ManslapValidator)

  return self
end
