--- MatryoshkaGame
-- @author Template created by Aaron Barrett on September 27, 2013.

class("MatryoshkaGame")
require("bonuses.ResponseSpeedTracker")
require("MatryoshkaBackground")
require("MatryoshkaInGameUi")
require("MatryoshkaCommon")
require("MatryoshkaDoll")
require("MatryoshkaImageGrid")
require("StandardUIBuilder")
require("MatryoshkaConfiguration")
require("MatryoshkaConceptGenerator")
require("MatryoshkaRoundLabel")
require("MatryoshkaRationaleLayer")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------

local Z_BACKGROUND = 0
local Z_BASE = 10
local Z_IMAGES = 10
local Z_SHADOW = 20
local Z_DOLL = 30
local Z_RATIONALE = 40
local Z_UI = 50
local Z_PARTICLES = 60
local Z_BANNER = 70

-- Colors -----------------------------

local COLOR_UI = {0.2, 0.2, 0.2, 0.58}

-- Other ------------------------------
local END_SEQUENCE_DOLL_SPACING_MULTIPLIER = 0.1

local function configureGameDifficulty(self)
  self.setsToComplete = MatryoshkaConfiguration.SET_COUNT

  local rankTwoScore = MatryoshkaConfiguration.getRankTwoScore()
  local rankThreeScore = MatryoshkaConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = MatryoshkaConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  local contentDifficulty = difficultyNameChooser:nextValue()
  GameSession.conceptGenerator:setDifficulty(contentDifficulty)

  GameSession.setInitialLives(MatryoshkaConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setTouchEnabled(self, isEnabled)
  self.imageGrid:setTouchEnabled(isEnabled)
end

local function selectNextConcept(self)
  if (#self.selectionOrder < 1) then
    return false
  end
  self.currentImage = table.remove(self.selectionOrder, #self.selectionOrder)
  GameSession.contentTracker:notifyPresentedConcept(self.currentImage.concept.identifier)
  self.inGameUi:setQuestionAndAnswer(self.currentImage.concept.prompt, self.currentImage.concept.answer)
  self.responseSpeedTracker:notifyStart()
  return true
end

local function insertPreviousSetMistakes(self)
  local targetSetSize = #self.nextConcepts
  for _, previousMistake in ipairs(self.previouslyIncorrectConcepts) do
    for index, concept in ipairs(self.nextConcepts) do
      if (concept.identifier == previousMistake.identifier) then
        table.remove(self.nextConcepts, index)
      end
    end
  end
  local removedConceptCount = targetSetSize - #self.nextConcepts
  local additionalConceptsToRemove = #self.previouslyIncorrectConcepts - removedConceptCount
  TableUtils.removeFromEndOfArray(self.nextConcepts, additionalConceptsToRemove)
  self.nextConcepts = TableUtils.concatenateArrays(self.nextConcepts, self.previouslyIncorrectConcepts)
  self.previouslyIncorrectConcepts = {}
end

local function loadNextConcepts(self)
  local shuffledConcepts = TableUtils.shuffleArray(self.nextConcepts)
  self.imageGrid:setNextConcepts(shuffledConcepts)
end

local function startSet(self)
  insertPreviousSetMistakes(self)
  loadNextConcepts(self)
  return runInCoroutine(function()
    self.imageGrid:showNextConcepts()
    self.selectionOrder = TableUtils.shallowCopy(self.imageGrid:getCurrentImages())
    TableUtils.shuffleArray(self.selectionOrder)
    selectNextConcept(self)
    self.hasIncorrectAnswers = false
  end)
end

local function prepareSet(self, index)
  local imageCounts = MatryoshkaConfiguration.IMAGE_COUNTS
  local conceptCount = imageCounts[index]
  self.nextConcepts = self.conceptGenerator:nextAsArray(conceptCount)
end

local function advanceSet(self)
  local imageCounts = MatryoshkaConfiguration.IMAGE_COUNTS
  self.setIndex = math.min(self.setIndex + 1, #imageCounts)
end

local function hasNextSet(self)
  return self.setIndex < self.setsToComplete
end

local function getDollStartPosition(self, doll)
  local dollWidth = self.doll:getEffectiveWidth()
  local dollLocationX = math.floor(dollWidth * 0.85)

  return dollLocationX, self.dollLocationY
end

local function insertNewDoll(self)
  self.currentDollType = self.currentDollType + 1
  self.doll = MatryoshkaDoll.new(self.currentDollType)
  self.doll:setScl(math.min(self.doll:scaleForEffectiveDims(nil, self.maxDollHeight)))
  self.doll:setLoc(getDollStartPosition(self, self.doll))
  self:insert(self.doll, Z_DOLL)
end

local function makeParticles()
  local spriteLeft = Sprite.newWithName("particles/particles_000.png")
  local animation = Animation.newWithPrefix("particles/particles_")

  local particleWidth, particleHeight = spriteLeft:getDims()
  local self = Group.new(particleWidth*2, particleHeight)

  local selfWidth, selfHeight = self:getDims()

  spriteLeft:setLoc(selfWidth*0.25, selfHeight*0.5)
  self:insert(spriteLeft)

  local spriteRight = Sprite.newWithName("particles/particles_000.png")
  spriteRight:setScl(-1, 1)
  spriteRight:setLoc(selfWidth*0.75, selfHeight*0.5)
  self:insert(spriteRight)

  runInCoroutine(function()
    AHAnimate.new(spriteLeft, animation, 25):start()
    wait(AHAnimate.new(spriteRight, animation, 25):start())
    spriteLeft:removeSelf()
    spriteRight:removeSelf()
  end)

  return self
end

local function showBanner(self, banner, locationX, locationY, duration)
  duration = duration or 1
  return runInCoroutine(function()
    banner:setScl(0, 0.9)
    banner:setLoc(locationX, locationY)
    wait(AHSequence.new(
      AHScaleTo.new(banner, 0.1, AHEaseSineOut.new(), 1.1, 0.9),
      AHScaleTo.new(banner, 0.1, AHEaseSineInOut.new(), 1.0, 1.0),
      AHDelay.new(duration)
      ):start())
    AHOpacityTo.new(banner, 0.2, AHEaseSineOut.new(), 0):start()

  end)
end

local function showFinalSet(self)
  return runInCoroutine(function()
    local banner = MatryoshkaRoundLabel.newFinalSetLabel()
    local bannerLocationX = Screen.getSafeX(0.5)
    local bannerLocationY = Screen.getSafeY(0.5)
    self:insert(banner, Z_BANNER)
    local duration = 0.5
    wait(showBanner(self, banner, bannerLocationX, bannerLocationY, duration))
  end)
end

local function showCompleteSet(self)
  return runInCoroutine(function()
    waitFor(0.5)
    local oldDoll = self.doll
    insertNewDoll(self)
    self.doll:prepareToEatDoll(oldDoll, false)
    wait(self.doll:eatDoll(oldDoll))
  end)
end

local function showPerfectSet(self)
  return runInCoroutine(function()
    GameSession.notifyGameTimePause()
    waitFor(0.2)
    local originalLocationX, originalLocationY = self.doll:getLoc()
    local originalScale = self.doll:getScl()
    local dollWidth, dollHeight = self.doll:getEffectiveDims()
    local dollLocationX = Screen.getSafeX(0.5)
    local dollLocationY =  Screen.getSafeY(0.5) - (dollHeight / 2)
    AHOpacityTo.new(self.shadow, 0.1, nil, 0):start()
    wait(AHLocationTo.new(self.doll, 0.2, AHEaseSineInOut.new(), dollLocationX, dollLocationY):start())
    local particles = makeParticles()
    particles:setEffectiveDims(dollWidth * 3)
    particles:setAnchorPointBottom()
    particles:setLoc(dollLocationX, (dollLocationY + (dollHeight * 0.5)))
    self:insert(particles, Z_PARTICLES)
    waitFor(0.2)
    Sound.playSoundEffect("matryoshka_perfect")
    local banner = MatryoshkaRoundLabel.newRoundLabel(Localizer.get("PERFECT_SET"), {211,135,48})
    local bannerHeight = banner:getHeight()
    local bannerLocationX = dollLocationX
    local bannerLocationY =  dollLocationY - (dollHeight * 0.1) - (bannerHeight / 2)
    self:insert(banner, Z_PARTICLES)
    showBanner(self, banner, bannerLocationX, bannerLocationY)
    wait(AHScaleTo.new(self.doll, 0.1, AHEaseSineInOut.new(), originalScale * 1.15):start())
    wait(AHScaleTo.new(self.doll, 0.1, AHEaseSineInOut.new(), originalScale * 1.1):start())

    if hasNextSet(self) then
      local roundsLeftBanner = nil
      local roundsLeft =  self.setsToComplete - self.setIndex
      if (roundsLeft > 1) then
        local roundsLeftString = string.format(Localizer.get("ROUNDS_LEFT"), roundsLeft)
        roundsLeftBanner = MatryoshkaRoundLabel.newRoundLabel(roundsLeftString, {108,154,159})
        roundsLeftBanner:setLoc(banner:getLoc())
        roundsLeftBanner:setScl(1, 0)
        self:insert(roundsLeftBanner, Z_PARTICLES)
      end
      local oldDoll = self.doll
      insertNewDoll(self)
      originalScale = self.doll:getScl()
      local oldDollLocationX, oldDollLocationY = oldDoll:getLoc()
      oldDollLocationY = oldDollLocationY - 4
      self.doll:setLoc(oldDollLocationX, oldDollLocationY)
      self.doll:prepareToEatDoll(oldDoll, true)
      if roundsLeftBanner then
        waitFor(0.5)
        AHSequence.new(
          AHScaleTo.new(banner, 0.1, AHEaseSineInOut.new(), 1, 0),
          AHScaleTo.new(roundsLeftBanner, 0.1, AHEaseSineInOut.new(), 1, 1)
          ):start()
        waitFor(0.1)
      end
      waitFor(0.8)
      runInCoroutine(function()
        waitFor(0.2)
        banner:removeSelf()
        if roundsLeftBanner then
          roundsLeftBanner:removeSelf()
        end
      end)
      wait(self.doll:eatDoll(oldDoll))
      waitFor(0.2)
      AHSequence.new(
        AHDelay.new(0.1),
        AHOpacityTo.new(self.shadow, 0.1, nil, 1)
        ):start()
      wait(AHSpawn.new(
        AHLocationTo.new(self.doll, 0.2, AHEaseSineInOut.new(), originalLocationX, originalLocationY),
        AHScaleTo.new(self.doll, 0.2, AHEaseSineInOut.new(), originalScale)
        ):start())
    end
    GameSession.notifyGameTimeStart()
  end)
end

local function getMultiplierForSuccessiveDolls(dollCount)
  local proportion = 1
  local accumulator = 0
  for i = 1, dollCount do
    accumulator = accumulator + proportion
    proportion = proportion * MatryoshkaCommon.END_SEQUENCE_DOLL_SIZE_PROPORTION
  end

  return accumulator
end

local function getEndDollDistanceFromEdge(self)
  return Screen.getSafeX(0.05)
end

local function getEndFirstDollSequenceWidth(self, dollCount)
  local selfWidth = self:getWidth()
  local margin = getEndDollDistanceFromEdge(self)
  local effectiveWidth = selfWidth - margin * 2
  local dollMultiplier = getMultiplierForSuccessiveDolls(dollCount) + END_SEQUENCE_DOLL_SPACING_MULTIPLIER * dollCount

  return math.floor(effectiveWidth / dollMultiplier)
end

local function showEndSequence(self)
  return runInCoroutine(function()
    GameSession.notifyGameTimePause()
    Sound.playSoundEffect("matryoshka_perfect")
    self:showDefaultGameWinOutro()
    local dollCount = self.setsToComplete
    local firstDollEffectiveW = getEndFirstDollSequenceWidth(self, dollCount)
    local initialScale = self.doll:scaleForEffectiveDims(firstDollEffectiveW)
    local margin = getEndDollDistanceFromEdge(self)
    local dollLocationX = math.floor(margin + firstDollEffectiveW * 0.5)
    local dollSpacing = firstDollEffectiveW * END_SEQUENCE_DOLL_SPACING_MULTIPLIER

    AHOpacityTo.new(self.shadow, 0.1, nil, 0):start()

    local resizeFirstDollDuration = 0.3
    wait(AHSpawn.new(
      AHScaleTo.new(self.doll, resizeFirstDollDuration, AHEaseSineInOut.new(), initialScale),
      AHLocationTo.new(self.doll, resizeFirstDollDuration, AHEaseSineInOut.new(), dollLocationX, self.dollStandardY)
    ):start())

    local currentDoll = self.doll
    local dollIndex = 1
    for i = dollCount-1, 1, -1 do
      local nextDoll = MatryoshkaDoll.new(i)
      wait(currentDoll:spitOutDoll(nextDoll, dollSpacing, dollIndex))
      currentDoll = nextDoll
      dollIndex = dollIndex + 1
    end
  end)
end

local function loseLife(self)
  self.livesDisplay:decrement()
end

local function hasLivesRemaining(self)
  return self.livesDisplay:getLives() > 0
end

local function onGameLoss(self)
  return runInCoroutine(function()
    setTouchEnabled(self, false)
    self.inGameUi:clearQuestion()
    local outroAction = self:showDefaultGameLossOutro()
    wait(self.imageGrid:removeImages())
    wait(outroAction)
    GameSession.notifyGameLoss()
  end)
end

local function onGameWin(self)
  setTouchEnabled(self, false)
  wait(showEndSequence(self))
  waitFor(0.9)

  GameSession.notifyGameWin()
end

local function onSetComplete(self)
  return runInCoroutine(function()
    self.imageGrid:removeImages()
    self.inGameUi:clearQuestion()
    if not hasNextSet(self) then
      ScoringUtils.applyLivesBonus(self.livesDisplay:getLives(), MatryoshkaConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
      self.scoreLabel:setScore(GameSession.getScore())
      onGameWin(self)
    else
      local isPerfectSet = not self.hasIncorrectAnswers
      if isPerfectSet then
        wait(showPerfectSet(self))
      else
        wait(showCompleteSet(self))
      end
      advanceSet(self)
      if hasNextSet(self) then
        startSet(self)
        prepareSet(self, self.setIndex+1)
      else
        wait(showFinalSet(self))
        startSet(self)
      end
    end
  end)
end

local function onCorrectAnswer(self, image)
  return runInCoroutine(function()
    local concept = image.concept

    GameSession.contentTracker:notifyCorrectAnswer(self.currentImage.concept.identifier, "set_" .. self.setIndex)
    GameSession.notifyCorrectAnswer(self.currentImage.concept.identifier)
    GameSession.incrementCorrectAnswerCount()
    GameSession.increaseScore(MatryoshkaConfiguration.POINTS_PER_CORRECT_RESPONSE)
    Sound.playSoundEffect("matryoshka_correct")
    self.responseSpeedTracker:notifyCorrectResponse()
    self.scoreLabel:setScore(GameSession.getScore())

    wait(self.inGameUi:runCorrectSequence(image))
    wait(self.doll:eatImage(image))
    if not selectNextConcept(self) then
      wait(onSetComplete(self))
    end
  end)
end

local function onIncorrectAnswer(self, imageTapped)
  return runInCoroutine(function()
    self.hasIncorrectAnswers = true
    table.insert(self.previouslyIncorrectConcepts, self.currentImage.concept)
    GameSession.contentTracker:notifyIncorrectAnswer(self.currentImage.concept.identifier, "set_" .. self.setIndex, imageTapped.concept.identifier)
    GameSession.scoredIncorrectAnswer(self.currentImage.concept.identifier)
    GameSession.incrementIncorrectAnswerCount()
    Sound.playSoundEffect("matryoshka_incorrect")
    self.responseSpeedTracker:notifyIncorrectResponse()
    self.doll:rock()

    local imageForRationale = self.currentImage
    local otherImages = {}
    local images = self.imageGrid:getCurrentImages()
    for _, image in ipairs(images) do
      if (image ~= imageForRationale) then
        table.insert(otherImages, image)
      end
      image:stopWiggle()
    end

    loseLife(self)
    self.inGameUi:runIncorrectSequence(imageTapped, imageForRationale, otherImages)
    wait(self.rationaleLayer:runIncorrectSequence(imageForRationale, self.currentImage.concept.prompt, self.currentImage.concept.definition))

    local tapToContinue = TapToContinue.new()
    tapToContinue:waitForTrigger()
    tapToContinue:removeSelf()

    self.inGameUi:runReturnFromIncorrectSequence(imageForRationale, otherImages)
    wait(self.rationaleLayer:runReturnFromIncorrectSequence())

    for _, image in ipairs(otherImages) do
      image:wiggle()
    end

    if not hasLivesRemaining(self) then
      wait(onGameLoss(self))
    elseif not selectNextConcept(self) then
      wait(onSetComplete(self))
    end
  end)
end

local function getImagePressedHandler(self)
  return function(image)
    runInCoroutine(function()
      setTouchEnabled(self, false)
      if image == self.currentImage then
        wait(onCorrectAnswer(self, image))
      else
        wait(onIncorrectAnswer(self, image))
      end
      setTouchEnabled(self, true)
    end)
  end
end

local function applyCustomSpeedBonus(earnedBonusProportion, bonusProportionsSum, pointsPerFullSpeedBonus, bonusTray)
  local speedBonusPoints = pointsPerFullSpeedBonus * earnedBonusProportion
  GameSession.increaseScore(speedBonusPoints)

  if (speedBonusPoints >= 1.0) then
    local notificationStringKey = "CUSTOM_SPEED_BONUS"
    bonusTray:indicateSimpleTextBonus(notificationStringKey, speedBonusPoints)
  end

  GameSession.bonusReporter:incrementBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function MatryoshkaGame:gameFlow()
  return runInCoroutine(function()
    self.isGameRunning = true

    waitFor(0.1)
    wait(self.moveDollInAction:start())
    wait(startSet(self))
    prepareSet(self, self.setIndex + 1)
  end)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local selfWidth, selfHeight = self:getDims()

  self.background = MatryoshkaBackground.new()
  self.background:setLoc(selfWidth / 2, selfHeight / 2)
  self:insert(self.background, Z_BACKGROUND)

  self.base = Sprite.newWithName("base.png")
  local baseHeight = self.base:getHeight()
  self.base:setEffectiveDims(selfWidth, baseHeight)
  self.base:setLoc(selfWidth / 2, math.floor(baseHeight / 4))
  self:insert(self.base, Z_BASE)

  assert(self.scoreLabel ~= nil, "scoreLabel must be created before MatryoshkaInGameUi")

  self.inGameUi = MatryoshkaInGameUi.new()
  local uiLocationY = self.scoreLabel:getY()
  self:insertAtLocation(self.inGameUi, self:getWidthPart(0.5), uiLocationY, Z_UI)

  local imageGridWidth = math.floor(Screen.safeWidth)
  local imageGridLowerBound = self.base:getYTop() + self:getHeightPart(0.2)
  local imageGridHeight = math.floor(0.95 * (self.inGameUi:getYBottom() - imageGridLowerBound))
  local imageGridLocationX = selfWidth / 2
  local imageGridLocationY = (imageGridHeight / 2) + imageGridLowerBound

  self.imageGrid = MatryoshkaImageGrid.new(imageGridWidth, imageGridHeight, getImagePressedHandler(self))
  self.imageGrid:setLoc(imageGridLocationX, imageGridLocationY)
  self:insert(self.imageGrid, Z_IMAGES)

  self.rationaleLayer = MatryoshkaRationaleLayer.new()
  self:insertAtCenter(self.rationaleLayer, Z_RATIONALE)

  self.dollLocationY = baseHeight / 2
  self.maxDollHeight = self:getHeightPart(0.2)

  self.currentDollType = 0
  insertNewDoll(self)
  self.shadow = Sprite.newWithName("doll_shadow.png")
  self.shadow:setEffectiveDims(self.doll:getEffectiveWidth())
  self.shadow:setLoc(self.base:transformObjectLoc(self.doll))
  self.base:insert(self.shadow, Z_SHADOW)

  local shadowScale = self.shadow:getScl()
  local baseLocationX, baseLocationY = self.base:getLoc()
  local dollLocationX, dollLocationY = self.doll:getLoc()

  self.dollStandardY = dollLocationY

  self.base:setLoc(baseLocationX, baseLocationY - self.maxDollHeight - baseHeight/2)
  self.doll:setLoc(dollLocationX, dollLocationY - self.maxDollHeight - baseHeight/2)

  self.moveDollInAction = AHSpawn.new(
    AHLocationTo.new(self.base, 0.2, AHEaseSineOut.new(), baseLocationX, baseLocationY),
    AHLocationTo.new(self.doll, 0.2, AHEaseSineOut.new(), dollLocationX, dollLocationY + selfHeight * 0.06),
    function() Sound.playSoundEffect("matryoshka_game_start") end,
    AHSequence.new(
      AHDelay.new(0.1),
      AHSpawn.new(
        AHOpacityTo.new(self.shadow, 0.1, AHEaseSineInOut.new(), 0.2),
        AHScaleTo.new(self.shadow, 0.1, AHEaseSineInOut.new(), shadowScale * 0.5)
        ),
      AHSpawn.new(
        AHOpacityTo.new(self.shadow, 0.05, AHEaseSineInOut.new(), 1),
        AHScaleTo.new(self.shadow, 0.05, AHEaseSineInOut.new(), shadowScale),
        AHLocationTo.new(self.doll, 0.05, AHEaseSineIn.new(), dollLocationX, dollLocationY)
        ),
        AHLocationTo.new(self.doll, 0.03, AHEaseExponentialOut.new(), dollLocationX, dollLocationY + selfHeight * 0.01),
        AHLocationTo.new(self.doll, 0.03, AHEaseSineIn.new(), dollLocationX, dollLocationY)
      )
    )


    local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
      applyCustomSpeedBonus(earnedBonusProportion,
                            bonusProportionsSum,
                            MatryoshkaConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                            self.bonusTray)
    end

    self.responseSpeedTracker = ResponseSpeedTracker.new(MatryoshkaConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                         quickResponseWrapper,
                                                         MatryoshkaConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), COLOR_UI)

  local baseYOffset = 10
  local safeAreaTop = (Screen.fullHeight - Screen.safeTop)
  local yOffsetWithSafeArea = baseYOffset - safeAreaTop
  local yOffsetForUI = -1 * math.max(yOffsetWithSafeArea, 0)
  local xOffsetForUI = nil

  builder:addPauseButton(xOffsetForUI, yOffsetForUI)
  self.scoreLabel = builder:addScoreLabel(xOffsetForUI, yOffsetForUI)
  self.bonusTray = builder:addBonusTray(xOffsetForUI, yOffsetForUI)
  self.livesDisplay = builder:addLivesDisplay(MatryoshkaConfiguration.INITIAL_LIVES_COUNT, xOffsetForUI, yOffsetForUI)
end

-- Constructor ------------------------

function MatryoshkaGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, MatryoshkaGame)

  configureGameDifficulty(self)

  createUI(self)
  createGameElements(self)

  self.conceptGenerator = MatryoshkaConceptGenerator.new()
  self.previouslyIncorrectConcepts = {}
  self.setIndex = 1
  prepareSet(self, self.setIndex)

  return self
end
