--- MatryoshkaRationaleLayer
-- @author Created by Jay A. around November 11, 2023.

class("MatryoshkaRationaleLayer")

require("utils.ColorUtils")
require("text.SizeConstrainedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OVERLAY_COLOR = {0, 0, 0}
local OVERLAY_OPACITY_VISIBLE = 0.2
local OVERLAY_OPACITY_INVISIBLE = 0.0

local IMAGE_OVERSIZE_SCALE_AMOUNT = 1.5
local IMAGE_LOCATION_Y_PROPORTION = 0.585

local MARGIN_X_LEFT_MINIMUM_PROPORTION = 0.082
local MARGIN_Y_TOP_PROPORTION = 0.05
local MARGIN_Y_IMAGE_PROPORTION_IMAGE_TO_PANEL = 0.47

local PANEL_COLOR = ColorUtils.colorHex24ToStd(0x432D0E)
local WORD_COLOR = ColorUtils.colorHex24ToStd(0xD04545)
local DEFINITION_COLOR = ColorUtils.colorHex24ToStd(0xD9D9D9)
local TAP_TO_CONTINUE_COLOR = ColorUtils.colorHex24ToStd(0xCCCCCC)

local INDENT_LINE_COLOR = ColorUtils.colorHex24ToStd(0x857762)
local INDENT_LINE_THICKNESS = 1
local INDENT_LINE_HEIGHT_REDUCTION_PROPORTION = 0.86
local INDENT_LINE_LOCATION_X_WORD_TO_DEFINITION_PROPORTION = 0.25

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createOrReplaceIndentLine(self)
  if (self.indentLine ~= nil) then
    self.indentLine:removeSelf()
  end

  local wordLabel = self.wordLabel
  local definitionTextArea = self.definitionTextArea

  local lineWidth = INDENT_LINE_THICKNESS
  local lineHeight = definitionTextArea:getHeightPart(INDENT_LINE_HEIGHT_REDUCTION_PROPORTION)
  local lineX = MathUtils.interpolateLinearFromNormalized(INDENT_LINE_LOCATION_X_WORD_TO_DEFINITION_PROPORTION,
                                                          wordLabel:getXLeft(),
                                                          definitionTextArea:getXLeft())
  local lineY = definitionTextArea:getYCenter()

  local line = FilledRectangle.new(lineWidth, lineHeight, INDENT_LINE_COLOR)
  self.panel:insertAtLocation(line, lineX, lineY)
  self.indentLine = line
end

local function createPanelIfNeededAndUpdateLabels(self, height, word, definition)
  if (self.panel == nil) then
    local panel = FilledRectangle.new(Screen.width, height, PANEL_COLOR)
    panel:setAnchorPointBottom()
    self:insertAtLocation(panel, self:getWidthPart(0.5), -height)

    self.panel = panel

    local marginLeft = math.max(self:getWidthPart(MARGIN_X_LEFT_MINIMUM_PROPORTION), Screen.safeLeft)
    local marginTop = self:getHeightPart(MARGIN_Y_TOP_PROPORTION)
    local definitionIndentX = (self:getWidth() - (2 * marginLeft)) * 0.05
    local textWidth = (self:getWidth() - (2 * marginLeft)) - definitionIndentX

    local wordLabel = Label.new("", Font.new(FontFaces.DIN_OT_BOLD, 20))
    wordLabel:setColor(unpack(WORD_COLOR))
    wordLabel:setAnchorPointLeft()
    self.panel:insertAtLocation(wordLabel, marginLeft, self.panel:getHeight() - marginTop)
    self.wordLabel = wordLabel

    local tapToContinueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 13))
    tapToContinueLabel:setColor(unpack(TAP_TO_CONTINUE_COLOR))
    tapToContinueLabel:setAnchorPointBottom()
    local tapToContinueLocationY = math.max(marginTop, Screen.safeBottom)
    self.panel:insertAtLocation(tapToContinueLabel, self.panel:getWidthPart(0.5), tapToContinueLocationY)

    local marginWordToDefinition = marginTop * 0.25
    local definitionLocationY = wordLabel:getYBottom() - marginWordToDefinition

    local baseAttributes = {
      font = Font.new(FontFaces.DIN_OT_REGULAR, 20),
      color = DEFINITION_COLOR
    }
    local maxHeight = definitionLocationY - (tapToContinueLabel:getYTop() + tapToContinueLabel:getHeight())
    local mininumFontSize = 14
    local definitionTextArea = SizeConstrainedTextArea.new("", textWidth, maxHeight, baseAttributes, nil, mininumFontSize)
    definitionTextArea:setAnchorPointTopLeft()
    self.panel:insertAtLocation(definitionTextArea, marginLeft + definitionIndentX, definitionLocationY)
    self.definitionTextArea = definitionTextArea
  end

  self.wordLabel:setText(word)

  self.definitionTextArea:setText(definition)
  self.definitionTextArea:rerender()
  createOrReplaceIndentLine(self)
end

local function getPanelHeight(self, image)
  local expectedImageYBottom = self:getHeightPart(IMAGE_LOCATION_Y_PROPORTION) - (image:getHeightPart(0.5) * IMAGE_OVERSIZE_SCALE_AMOUNT)
  local marginY = image:getHeightPart(MARGIN_Y_IMAGE_PROPORTION_IMAGE_TO_PANEL) * IMAGE_OVERSIZE_SCALE_AMOUNT
  local height = expectedImageYBottom - marginY

  return height
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MatryoshkaRationaleLayer:runIncorrectSequence(image, word, definition)
  self:takeChild(image)
  self.image = image
  self.imageOriginalLocation = {image:getLoc()}

  local panelHeight = getPanelHeight(self, image)
  createPanelIfNeededAndUpdateLabels(self, panelHeight, word, definition)

  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0.7, 0.0, 0.3, 1.0})

  local incorrectSequence = AHSequence.new(
    AHDelay.new(0.65),
    AHSpawn.new(
      AHOpacityTo.new(self.overlay, duration, ease, OVERLAY_OPACITY_VISIBLE),
      AHLocationTo.new(image, duration, ease, self:getWidthPart(0.5), self:getHeightPart(IMAGE_LOCATION_Y_PROPORTION)),
      AHScaleTo.new(image, duration, ease, IMAGE_OVERSIZE_SCALE_AMOUNT),
      AHLocationYTo.new(self.panel, duration, ease, 0)
    )
  )

  return incorrectSequence:start()
end

function MatryoshkaRationaleLayer:runReturnFromIncorrectSequence()
  assert(self.image ~= nil, "Rationale image must be set before returning from the incorrect sequence.")
  assert(self.imageOriginalLocation ~= nil, "Rationale image location must be cached before returning from the incorrect sequence.")
  assert(self.imageOriginalLocation[1] ~= nil, "Rationale image x must be cached before returning from the incorrect sequence.")
  assert(self.imageOriginalLocation[2] ~= nil,  "Rationale image y must be cached before returning from the incorrect sequence.")

  local duration = 0.5
  local ease = AHEaseBezier.newFlat({0.7, 0.0, 0.3, 1.0})

  local returnSequence = AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(self.overlay, duration, ease, OVERLAY_OPACITY_INVISIBLE),
      AHScaleTo.new(self.image, duration, ease, 0),
      AHLocationTo.new(self.image, duration, ease, self.imageOriginalLocation[1], self.imageOriginalLocation[2]),
      AHLocationYTo.new(self.panel, duration, ease, -self.panel:getHeight())
    ),
    function()
      self.image:removeSelf()
      self.image = nil
      self.imageOriginalLocation = nil
    end
  )

  return returnSequence:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local overlay = FilledRectangle.new(Screen.width, Screen.height, OVERLAY_COLOR)
  overlay:setOpacity(0)
  self:insertAtCenter(overlay)
  self.overlay = overlay

  -- The rest of the layout (e.g. panel and all text) must be created later during
  -- the incorrect sequence since reference to an image is necessary to determine
  -- exact dimensions. See MatryoshkaRationaleLayer:runIncorrectSequence().
end

-- Constructor ------------------------

function MatryoshkaRationaleLayer.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, MatryoshkaRationaleLayer)

  createLayout(self)

  return self
end
