--- MoonsetGame
-- @author created by Jay Austria on Octomom 16, 2013

class("MoonsetGame")

require("bonuses.BonusNotificationTray")
require("bonuses.StreakTracker")
require("StandardUIBuilder")
require("MoonsetOverlayBlend")
require("MoonsetCommon")
require("MoonsetConfiguration")
require("MoonsetRoundGenerator")
require("MoonsetPromptTextView")
require("MoonsetScene")
require("MoonsetTip")
require("widgets.ValidityButtonBar")
require("tutorials.TutorialHighlighter")
require("ExperimentsConfiguration")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------
local Z_INDICES = {
  GAME_CONTAINER = 0,
  GAME_SCENE = 10,
  TUTORIAL_UI = 15,
  DIMMER = 20,
  TIP = 30,
  PROMPT = 40
}

-- Colors -----------------------------
local UI_COLOR = {1, 1, 1}
local TUTORIAL_OVERLAY_COLOR = {0.039, 0.055, 0.082, 0.7}
local TUTORIAL_MASK_COLOR = {1.0, 1.0, 1.0, 1.0}
local OPACITY_DIM = 0.7

-- Placement --------------------------
local PROMPT_OFFSCREEN_X_LEFT = Screen.getWidthPart(-1.5)
local PROMPT_OFFSCREEN_X_RIGHT = Screen.getWidthPart(1.5)
local PROMPT_LOCATION_X = Screen.getSafeX(0.5)
local PROMPT_LOCATION_Y = Screen.getSafeY(0.62)

-- Sizing -----------------------------
local PROMPT_PADDING_X = Screen.getSafeWidthPart(0.2)
local PROMPT_WIDTH = Screen.safeWidth - PROMPT_PADDING_X
local PROMPT_HEIGHT = Screen.getSafeHeightPart(0.6)

-- Gameplay --------------------------
local CORRECT_AFTER_RESPONSE_DELAY = 1.0
local INCORRECT_AFTER_RESPONSE_DELAY = 1.7
local PROMPT_PLAY_IN_SPEED = 0.5
local PROMPT_PLAY_OUT_SPEED = 0.5

local function configureGameDifficulty(self)
  local passScore = MoonsetConfiguration.getPassScore()
  local rankTwoScore = MoonsetConfiguration.getRankTwoScore()
  local rankThreeScore = MoonsetConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = MoonsetConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  self.maxScore = ScoringUtils.getDifficultyScaledPoints(passScore)
  self.correctNeeded = MoonsetConfiguration.getNumberOfConcepts()
  self.roundTime = MoonsetConfiguration.getRoundTime()
  GameSession.setTotalTime(self.roundTime)

  assert(MoonsetConfiguration.getIncorrectAnswerPoints() < 0, "Incorrect answer point penalty must be negative.")

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.MAX_STREAK_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function waitUntilPressed(group)
  local pressed = false
  TouchManager.addTouchHandler(group, function() pressed = true end, TouchHandler.BEHAVIOR_FOCUS)
  waitUntil(function() return pressed end)
  TouchManager.removeTouchHandler(group)
end

local function setPaused(self, isPaused)
  self.scene:setPaused(isPaused)
  if isPaused then
    self.gameTimer:pauseAndShowIcon()
    GameSession.notifyGameTimePause()
  else
    self.gameTimer:resumeAndHideIcon()
    GameSession.notifyGameTimeStart()
  end
end

local function isOutOfTime(self)
  return (self.gameTimer:getTime() >= self.roundTime)
end

local function waitAndShowTip(self)
  if isOutOfTime(self) then
    return
  end

  local tip = MoonsetTip.new(self.roundData.tip)
  tip:setAnchorPointBottom()
  tip:setLoc(Screen.getWidthPart(0.5), -tip:getHeight())
  self:insert(tip, Z_INDICES.TIP)

  local originalY = self.promptGroup:getY()
  local duration = 0.4
  local promptY = Screen.safeTop - (Screen.safeTop - tip:getHeight()) / 2 + self.promptLabel:getHeightPart(0.5)
  wait(AHSpawn.new(
    AHLocationYTo.new(tip, duration, AHEaseSineOut.new(), 0),
    AHLocationYTo.new(self.promptGroup, duration, AHEaseSineOut.new(), promptY),
    AHOpacityTo.new(self.dimmer, duration, nil, OPACITY_DIM)
  ):start())
  waitUntilPressed(self.dimmer)

  duration = 0.2
  wait(AHSpawn.new(
    AHLocationYTo.new(tip, duration, AHEaseSineIn.new(), -tip:getHeight()),
    AHLocationYTo.new(self.promptGroup, duration, AHEaseSineOut.new(), originalY),
    AHOpacityTo.new(self.dimmer, duration, nil, 0)

  ):start())
  tip:removeSelf()
end

local function startSceneAnimation(self, duration)
  self.scene:startAnimation(duration)
end

local function isGameOver(self)
  return (isOutOfTime(self) or GameSession.didReachPassScore())
end

local function setScore(self, scoreToSet)
  local oldScore = self.score
  self.score = scoreToSet
  self.scoreLabel:setScore(scoreToSet)

  if(self.score > oldScore) then
    Sound.playSoundEffect("whoosh")
  end

  if (oldScore ~= self.score) then
    self.scene:moveScoreIcon(scoreToSet, self.maxScore)
  end
end

local function stopRound(self)
  if (self.responseWaitAction) then
    self.responseWaitAction:stop()
    self.responseWaitAction = nil
  end
end

local function getQuestionIdentifier(self)
  if (self.isConceptCorrect) then
    return "correct_form"
  else
    return "incorrect_form"
  end
end

local function registerUserResponse(self, didFlagValid)
  self.shouldShowTutorial = false
  if self.hintLabel then
    runInCoroutine(function()
      wait(AHOpacityTo.new(self.hintLabel, 0.033, AHEaseLinear.new(), 0):start())
      self.hintLabel:removeSelf()
    end)
  end

  runInCoroutine(function()
    local isRoundValid = self.isConceptCorrect
     self.isCorrectResponse = (isRoundValid == didFlagValid)
     self.buttonBar:setTouchEnabled(false)

     local conceptId = self.roundData[MoonsetRoundGenerator.CONCEPT_IDENTIFIER_KEY]

    if (self.isCorrectResponse) then
      self.scene:playCorrectAnimation(didFlagValid)
      GameSession.contentTracker:notifyCorrectAnswer(conceptId, getQuestionIdentifier(self))
      GameSession.notifyCorrectAnswer(conceptId)
      GameSession.incrementCorrectAnswerCount()
      GameSession.increaseScore(MoonsetConfiguration.getCorrectAnswerPoints())
      self.streakTracker:notifyCorrectAnswer()
    else
      self.scene:playIncorrectAnimation(didFlagValid)
      GameSession.contentTracker:notifyIncorrectAnswer(conceptId, getQuestionIdentifier(self))
      self.roundGenerator:notifyMistake(conceptId)
      GameSession.notifyIncorrectAnswer(conceptId)
      GameSession.incrementIncorrectAnswerCount()
      GameSession.increaseScore(MoonsetConfiguration.getIncorrectAnswerPoints())
      self.streakTracker:notifyIncorrectAnswer()
    end

    local scoreToSet = GameSession.getScore()
    setScore(self, scoreToSet)

    local roundEndDelayDuration = nil
    if (self.isCorrectResponse) then
      roundEndDelayDuration = CORRECT_AFTER_RESPONSE_DELAY
    else
      roundEndDelayDuration = INCORRECT_AFTER_RESPONSE_DELAY
    end

    waitFor(roundEndDelayDuration)

    if (not self.isCorrectResponse) then
      setPaused(self, true)
      waitAndShowTip(self)
      setPaused(self, false)
    end

    stopRound(self)
  end)
end

local function initGame(self)
  startSceneAnimation(self, MoonsetConfiguration.getRoundTime())
  AHOpacityTo.new(self.hintLabel, 0.167, AHEaseLinear.new(), 1.0):start()
  self.scene:showStartAnimation()
  self.gameTimer:initialize()
end

local function waitAndPrepareConceptForNextRound(self)
  local promptText = MoonsetRoundGenerator:getFormattedPromptForConcept(self.roundData, self.isConceptCorrect)
  local highlightText = MoonsetRoundGenerator:getHighlightTextForConcept(self.roundData, self.isConceptCorrect)
  local correctionText = MoonsetRoundGenerator:getCorrectionTextForConcept(self.roundData)

  local prompt = MoonsetPromptTextView.new(promptText, highlightText, correctionText, PROMPT_WIDTH)
  prompt:setAnchorPointTop()
  prompt:setLoc(self.promptGroup:getWidthPart(0.5), self.promptGroup:getHeightPart(0.5))
  self.promptLabel = prompt

  self.promptGroup:setLoc(PROMPT_OFFSCREEN_X_RIGHT, PROMPT_LOCATION_Y)
  self.scene:setPrompt(prompt)
end

local function presentPrompt(self)
  self.promptGroup:insert(self.promptLabel)

  local slideAction = AHLocationXTo.new(self.promptGroup, PROMPT_PLAY_IN_SPEED, AHEaseBackOut.new(0.65), PROMPT_LOCATION_X)
  slideAction:start()
end

local function startRound(self)
  self.isCorrectResponse = nil

  presentPrompt(self)

  self.gameTimer:resumeAndHideIcon()
  self.buttonBar:setTouchEnabled(true)
end

local function showTutorial(self)
  setPaused(self, true)
  self.buttonBar:setTouchEnabled(false)

  local uiOverlay = self:getUIOverlay()
  local oldUIZIndex = uiOverlay.zIndex
  self.gameElementsContainer:takeChild(uiOverlay, Z_INDICES.TUTORIAL_UI)

  local correctButton = (self.isConceptCorrect and self.buttonBar.affirmativeButton) or self.buttonBar.negativeButton
  local generateHole = function()
    local hole = FilledRoundedRectangle.new(correctButton:getWidth(), correctButton:getHeight(), 5, TUTORIAL_MASK_COLOR)
    local x, y = self:transformLoc(correctButton)
    x = x + hole:getWidthPart(0.5)
    y = y + hole:getHeightPart(0.5)
    hole:setLoc(x, y)
    return hole
  end

  local tutorialOverlay = TutorialHighlighter.newWithMasks({generateHole})
  self.gameElementsContainer:insertAtCenter(tutorialOverlay, Z_INDICES.DIMMER)
  tutorialOverlay:createFadeInAnimation():start()
  tutorialOverlay:waitForTargetTap()

  registerUserResponse(self, self.isConceptCorrect)

  wait(tutorialOverlay:createFadeOutAnimation():start())
  tutorialOverlay:removeSelf()

  self:takeChild(uiOverlay, oldUIZIndex)
  setPaused(self, false)
end

local function waitForResponse(self)
  local tutorialTimer = nil
  if self.shouldShowTutorial then
    tutorialTimer = {}
    Timer.makeTimer(tutorialTimer, MoonsetConfiguration.getTutorialDelay())
    tutorialTimer:setTimeoutHandler(
      function()
        if self.shouldShowTutorial then
          showTutorial(self)
        end
      end)
    tutorialTimer:start()
  end

  self.responseWaitAction = AHDelay.new(math.huge)
  wait(self.responseWaitAction:start())
end

local function waitForOutro(self)
  if (GameSession.didReachPassScore()) then
    ScoringUtils.applyMaxStreakBonus(self.streakTracker:getLongestStreak(), MoonsetConfiguration.getMaxStreakBonusPoints(), self.bonusTray)
    ScoringUtils.applyTimeBonus(self.gameTimer:getTimeRemaining(), MoonsetConfiguration.getTimeBonusPerSecond(), self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    Sound.playSoundEffect("goal_reached")
    self:showDefaultGameWinOutro()
    self.scene:showEndAnimation()
    waitFor(2.5)
    GameSession.notifyGameWin()
  else
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  end
end

local function waitForConcludeRound(self)
  local playOutAction = AHLocationXTo.new(self.promptGroup, PROMPT_PLAY_OUT_SPEED, AHEaseBackIn.new(0.65), PROMPT_OFFSCREEN_X_LEFT)

  local function removeLabelFunction()
    self.promptLabel:removeSelf()
    self.promptLabel = nil
  end

  local cleanupSequence = AHSequence.new(playOutAction, removeLabelFunction)

  wait(cleanupSequence:start())
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MoonsetGame:gameFlow()
  initGame(self)
  self.shouldShowTutorial = true

  repeat
    if (self.gameTimer:getTimeRemaining() < MoonsetConfiguration.getTimeRemainingToStartResurfacingMistakes()) then
      self.roundGenerator:allowMistakesToResurface()
    end

    self.roundData = self.roundGenerator:next()
    self.isConceptCorrect = MathUtils.getRandomBoolean()
    GameSession.contentTracker:notifyPresentedConcept(self.roundData[MoonsetRoundGenerator.CONCEPT_IDENTIFIER_KEY])
    waitAndPrepareConceptForNextRound(self)
    startRound(self)
    waitForResponse(self)
    waitForConcludeRound(self)
  until (isGameOver(self))

  self.gameTimer:stop()
  waitForOutro(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createStreakTracker(self)
  local function onStreakWrapper()
    local streakCount = self.streakTracker:getCurrentStreak()
    self.bonusTray:notifyStreakWithCount(streakCount)
  end

  local function onStreakEndWrapper()
    self.bonusTray:notifyStreakEnd()
  end

  local streakMinimum = 2
  local streakTracker = StreakTracker.new(streakMinimum, onStreakWrapper, onStreakEndWrapper)
  self.streakTracker = streakTracker
end

local function createLayout(self)
  local gameElementsContainer = Group.new(Screen.fullWidth, Screen.fullHeight)
  self.gameElementsContainer = gameElementsContainer

  local function onMarkAsValid()
    local didFlagValid = true
    registerUserResponse(self, didFlagValid)
  end

  local function onMarkAsError()
    local didFlagValid = false
    registerUserResponse(self, didFlagValid)
  end

  local styleDescriptor = {
    correctBackingColor = {0.04, 0.93, 0.78},
    incorrectBackingColor = {1.00, 0.31, 0.47},
    correctTextColor = {0.0, 0.25, 0.29},
    incorrectTextColor = {0.11, 0.22, 0.27},
    defaultTextColor = {0.04, 0.23, 0.27},
    defaultBackingColor = {1,1,1},
    centerDividerColor = {0.84, 0.91, 0.91}
  }

  local buttonBar = ValidityButtonBar.newWithStyleDescriptor(styleDescriptor)
  buttonBar:setNegativeButtonCallback(onMarkAsError)
  buttonBar:setAffirmativeButtonCallback(onMarkAsValid)
  buttonBar:setTouchEnabled(false)
  self.buttonBar = buttonBar

  local gameScene = MoonsetScene.new(buttonBar)
  self.gameElementsContainer:insertAtCenter(gameScene, Z_INDICES.GAME_SCENE)
  self.scene = gameScene

  self:insertAtCenter(gameElementsContainer, Z_INDICES.GAME_CONTAINER)

  self.dimmer = FilledRectangle.new(self:getWidth(), self:getHeight(), {0, 0, 0, 0})
  self.gameElementsContainer:insertAtCenter(self.dimmer, Z_INDICES.DIMMER)

  local promptGroup = Group.new(PROMPT_WIDTH, PROMPT_HEIGHT)
  promptGroup:setLoc(PROMPT_OFFSCREEN_X_RIGHT, Screen.getSafeY(0.5))
  self.gameElementsContainer:insert(promptGroup, Z_INDICES.PROMPT)
  self.promptGroup = promptGroup

  local hintFont = Font.new(FontFaces.DIN_OT_BOLD, 12)
  local hintLabel = DynamicTextArea.newSimple(Localizer.get("HINT"), Screen.safeWidth, hintFont, Alignment.CENTER, Alignment.CENTER)
  hintLabel:setLoc(Screen.getSafeWidthPart(0.5), self.buttonBar:getY() + self.buttonBar:getHeight())
  hintLabel:setOpacity(0)
  self.gameElementsContainer:insert(hintLabel, Z_INDICES.PROMPT)
  self.hintLabel = hintLabel

  self.score = 0
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), UI_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local function timeOutWrapper()
    stopRound(self)
  end

  self.gameTimer = builder:addDigitTimer(self.roundTime)
  self.gameTimer:setTimeoutHandler(timeOutWrapper)

  -- overlay completely visible when this percentage of game is complete
  local doneAtPercentComplete = 0.75
  local totalTime = MoonsetConfiguration.getRoundTime()
  local speedUpFactor = 1.0 / doneAtPercentComplete

  local defaultUpdateHandler = self.gameTimer:getUpdateHandler()
  self.gameTimer:setUpdateHandler(function()
    defaultUpdateHandler()

    local fractionElapsed = math.min(1.0, speedUpFactor * (totalTime - self.gameTimer:getTimeRemaining()) / totalTime)
    self.scene:updateTime(fractionElapsed)
  end)

  createStreakTracker(self)
end

function MoonsetGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, MoonsetGame)

  configureGameDifficulty(self)
  createUI(self)
  createLayout(self)

  local conceptIdKey = MoonsetRoundGenerator.CONCEPT_IDENTIFIER_KEY
  self.roundGenerator = LearningFocusedRoundGenerator.new(MoonsetRoundGenerator.new(), conceptIdKey)

  return self
end
