--- MoonsetHighlight
-- @author created by Jay Austria on Octomom 17, 2013

class("MoonsetHighlight")

require("MoonsetCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_RADIUS = 3
local PANEL_PAD_X = 15
local PANEL_PAD_Y = 5
local BACKGROUND_COLOR = {1.0, 1.0, 1.0}
local DEFAULT_TEXT_COLOR = {0.28, 0.28, 0.28}
local CORRECT_GREEN_COLOR = {0.03, 0.90, 0.76}
local CORRECT_BLACK_TEXT_COLOR = {0.15, 0.21, 0.20}
local INCORRECT_RED_COLOR = {1.0, 0.31, 0.47}
local INCORRECT_BLACK_TEXT_COLOR = {0.29, 0.21, 0.22}

local MINIMUM_BUBBLE_WIDTH = 20
local BUBBLE_PADDING_X = 14
local BUBBLE_PADDING_Y = 2
local BUBBLE_RADIUS = BUTTON_RADIUS
local BUBBLE_NEUTRAL_COLOR = {1, 1, 1}
local PARTICLES_SURGE_COUNT = 128


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function prepareParticleStyling(self)
  local width, height = self:getDims()
  if (self.bubble ~= nil) then
    self.particles:setScl(0.7)
    self.particles:setLoc(width * 0.5, height * 0.8)
  else
    self.particles:setScl(1.0)
    self.particles:setLoc(width * 0.5, height * 0.5)
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MoonsetHighlight:showAsCorrect()
  self.background:setColor(unpack(CORRECT_GREEN_COLOR))
  self.panelLabel:setColor(unpack(CORRECT_BLACK_TEXT_COLOR))
  self.particles:surge(PARTICLES_SURGE_COUNT)
end

function MoonsetHighlight:showWithCorrectionBubble()
  prepareParticleStyling(self)

  local showAction = AHSequence.new(
    function()
      self.bubble:setColor(unpack(CORRECT_GREEN_COLOR))
      self.correctionLabel:setColor(unpack(CORRECT_BLACK_TEXT_COLOR))
    end,
    AHOpacityTo.new(self.strikethrough, 0.1, nil, 1.0),
    AHSpawn.new(
      AHScaleTo.new(self.bubble, 0.1, AHEaseSimpleInOut.new(), 1.0),
      function() self.particles:surge(PARTICLES_SURGE_COUNT) end
    )
  )

  showAction:start()
end

function MoonsetHighlight:showAsIncorrectWithCorrectionBubble()
  local showAction = AHSequence.new(
    AHSpawn.new(
      function()
        self.background:setColor(unpack(INCORRECT_RED_COLOR))
        self.panelLabel:setColor(unpack(INCORRECT_BLACK_TEXT_COLOR))
        self.strikethrough:setColor(unpack(INCORRECT_BLACK_TEXT_COLOR))
      end,
      AHOpacityTo.new(self.strikethrough, 0.1, nil, 1.0),
      AHScaleTo.new(self.bubble, 0.1, AHEaseSimpleInOut.new(), 1.0)
    )
  )

  showAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createBubble(self, correctionText)
  local correctionLabel = Label.new(correctionText, Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  correctionLabel:setColor(unpack({0, 0, 0}))
  self.correctionLabel = correctionLabel

  local bubbleWidth = math.max(MINIMUM_BUBBLE_WIDTH, correctionLabel:getWidth()) + BUBBLE_PADDING_X
  local bubbleHeight = correctionLabel:getHeight() + BUBBLE_PADDING_Y

  local bubble = FilledRoundedRectangle.new(bubbleWidth, bubbleHeight, BUBBLE_RADIUS, BUBBLE_NEUTRAL_COLOR)
  local flag = FilledRectangle.new(5, 5, BUBBLE_NEUTRAL_COLOR)
  flag:setRot(45)
  bubble:insertAtLocationFractional(flag, 0.5, 0.0)

  flag:setAttrLink(MOAIColor.INHERIT_COLOR, bubble, MOAIColor.COLOR_TRAIT)

  bubble:setAnchorPointBottom()
  bubble:insertAtCenter(correctionLabel)

  self:insertAtLocationFractional(bubble, 0.5, 0.8)

  bubble:setScl(0)
  self.bubble = bubble
end

local function createParticles(self)
  local particles = Particles.new("shared_flick_particles_edge.pex")
  self:insertAtCenter(particles)
  particles:start()

  self.particles = particles
end

local function createLayout(text, correctionText)
  local panelLabel = Label.new(text, Font.new(MoonsetCommon.NORMAL_FONT, MoonsetCommon.PROMPT_FONT_SIZE))
  panelLabel:setColor(unpack(DEFAULT_TEXT_COLOR))
  local panelWidth = panelLabel:getWidth() + panelLabel:getLeftMargin() + panelLabel:getRightMargin() + PANEL_PAD_X
  local panelHeight = panelLabel:getHeight() + PANEL_PAD_Y

  local self = Group.new(panelWidth, panelHeight)

  createParticles(self)

  local background = FilledRoundedRectangle.new(panelWidth, panelHeight, BUTTON_RADIUS, BACKGROUND_COLOR)
  self:insertAtCenter(background)
  self.background = background

  -- todo fix the centering of this label
  self:insertAtLocationFractional(panelLabel, 0.5, 0.52)
  self.panelLabel = panelLabel

  local strikethroughWidth = panelLabel:getWidth() + 4
  local strikethrough = FilledRectangle.new(strikethroughWidth, 1, DEFAULT_TEXT_COLOR)
  strikethrough:setOpacity(0)
  self:insertAtLocationFractional(strikethrough, 0.5, 0.4)
  self.strikethrough = strikethrough

  if (correctionText) then
    createBubble(self, correctionText)
  end

  return self
end

function MoonsetHighlight.new(text, correctionText)
  local self = createLayout(text, correctionText)
  Mixin.mix(self, MoonsetHighlight)

  return self
end
