--- MoonsetSailboat
-- @author created by Jay Austria on November 14, 2013

class("MoonsetSailboat")
require("MoonsetCommon")

local GROUP_WIDTH = 40
local GROUP_HEIGHT = 70
local SAIL_GAP_Y = 3
local PARTICLES_SCALE = 0.5

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_SAIL = ColorUtils.colorHex24ToStd(0x1497AD)
local COLOR_HULL = ColorUtils.colorHex24ToStd(0x0D5666)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getReflectedColor(color)
  return ColorUtils.colorWithOpacity(color, MoonsetCommon.REFLECTION_OPACITY)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MoonsetSailboat:startMovingAnimation()
  if (self.ocillateAction) then
    self.ocillateAction:stop()
    self.ocillateAction = nil
  end

  if (self.boatAnimation) then
    self.boatAnimation:stop()
    self.boatAnimation = nil
  end

  local function floatAnimationFunction()
    local TILT_DURATION = 0.1
    local TILT_AMOUNT = 10

    local tiltForwardBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, nil, -TILT_AMOUNT)
    local tiltForwardReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, nil, TILT_AMOUNT)
    local tiltForwardSpawn = AHSpawn.new(tiltForwardBoatAction, tiltForwardReflectionAction)
    local freezeDelay = AHDelay.new(MoonsetCommon.BOAT_MOVE_DURATION)
    local floatAction = AHSequence.new(tiltForwardSpawn, freezeDelay)
    floatAction:start()
  end

  local boatAnimation = newCoroutineWithFunction(floatAnimationFunction)
  self.boatAnimation = boatAnimation:start()
end

function MoonsetSailboat:startJoltAnimation()
  local TILT_DURATION = 0.6
  local TILT_AMOUNT = -3
  local MOVE_UP_DURATION = 0.2
  local MOVE_DOWN_DURATION = 0.4
  local MOVE_AMOUNT = 1

  local tiltBackBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, nil, -TILT_AMOUNT)
  local tiltBackReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, nil, TILT_AMOUNT)
  local moveUpBoatAction = AHLocationYBy.new(self.topGroup, MOVE_UP_DURATION, nil, MOVE_AMOUNT)
  local moveDownBoatAction = AHLocationYBy.new(self.topGroup, MOVE_DOWN_DURATION, nil, -MOVE_AMOUNT)
  local moveUpDownBoatSequence = AHSequence.new(moveUpBoatAction, moveDownBoatAction)

  local moveUpReflectionAction = AHLocationYBy.new(self.bottomGroup, MOVE_UP_DURATION, nil, -MOVE_AMOUNT)
  local moveDownReflectionAction = AHLocationYBy.new(self.bottomGroup, MOVE_DOWN_DURATION, nil, MOVE_AMOUNT)
  local moveUpDowReflectionSequence = AHSequence.new(moveUpReflectionAction, moveDownReflectionAction)
  local tiltBackSpawn = AHSpawn.new(tiltBackBoatAction, tiltBackReflectionAction, moveUpDownBoatSequence, moveUpDowReflectionSequence)

  return tiltBackSpawn:start()
end

function MoonsetSailboat:startStationaryAnimation()
  if (self.boatAnimation) then
    self.boatAnimation:stop()
    self.boatAnimation = nil
  end

  local function stationaryAnimationFunction()
    local TILT_DURATION = 1.4
    local TILT_AMOUNT = 5

    local tiltEase = AHEaseSimpleInOut.new(2.0)

    local tiltBackBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltBackReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltBackSpawn = AHSpawn.new(tiltBackBoatAction, tiltBackReflectionAction)

    local tiltForwardBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, tiltEase, -TILT_AMOUNT)
    local tiltForwardReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltForwardSpawn = AHSpawn.new(tiltForwardBoatAction, tiltForwardReflectionAction)

    local tiltSequence = AHSequence.new(tiltForwardSpawn, tiltBackSpawn)
    local ocillateAction = AHLoop.new(tiltSequence)

    self.ocillateAction = ocillateAction:start()
  end

  local boatAnimation = newCoroutineWithFunction(stationaryAnimationFunction)
  self.boatAnimation = boatAnimation:start()

  self.rippleParticles:start()
  self.rippleParticles:startSystem()
end

function MoonsetSailboat.new(boatType)
  boatType = boatType or 1
  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  Mixin.mix(self, MoonsetSailboat)

  local topGroup = Group.new(GROUP_WIDTH, GROUP_HEIGHT * 0.5)
  topGroup:setAnchorPointBottom()
  self:insertAtCenter(topGroup)
  self.topGroup = topGroup

  local hull = Sprite.newWithName("sailboat_bottom.png")
  hull:setColor(unpack(COLOR_HULL))
  local bottomH = hull:getHeight()
  hull:setLoc(GROUP_WIDTH * 0.5, bottomH * 0.5)
  topGroup:insert(hull)

  local sail = Sprite.newWithName("sailboat_top.png")
  sail:setColor(unpack(COLOR_SAIL))
  local topH = sail:getHeight()
  sail:setLoc(GROUP_WIDTH * 0.5, bottomH + topH * 0.5 + SAIL_GAP_Y)
  topGroup:insert(sail)

  local bottomGroup = Group.new(GROUP_WIDTH, GROUP_HEIGHT * 0.5)
  bottomGroup:setAnchorPointTop()
  self:insertAtCenter(bottomGroup)
  self.bottomGroup = bottomGroup

  local reflectedHull = Sprite.newWithName("sailboat_bottom.png")
  reflectedHull:setColor(unpack(getReflectedColor(COLOR_HULL)))
  reflectedHull:setScl(1.0, -1.0)
  reflectedHull:setLoc(GROUP_WIDTH * 0.5, GROUP_HEIGHT * 0.5 - bottomH * 0.5)
  bottomGroup:insert(reflectedHull)

  local reflectedSail = Sprite.newWithName("sailboat_top.png")
  reflectedSail:setColor(unpack(getReflectedColor(COLOR_SAIL)))
  reflectedSail:setScl(1.0, -1.0)
  reflectedSail:setLoc(GROUP_WIDTH * 0.5, GROUP_HEIGHT * 0.5 - bottomH - topH * 0.5 - SAIL_GAP_Y)
  bottomGroup:insert(reflectedSail)

  local rippleParticles = Particles.new("ripples.pex")
  rippleParticles:setScl(PARTICLES_SCALE)
  self:insertAtCenter(rippleParticles)
  self.rippleParticles = rippleParticles

  self:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)

  return self
end
