--- MoonsetScene
-- @author created by Jay Austria on November 18, 2013

class("MoonsetScene")
require("Particles")
require("MoonsetBackground")
require("MoonsetSailboat")
require("MoonsetMoon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MOON_RADIUS_MULTIPLIER = 0.2
local BOAT_SCALE = 0.65
local BOAT_X_START = Screen.getSafeX(0.1)
local BOAT_X_END = Screen.getSafeX(0.95)
local SAILBOAT_OFFSET_Y = -2
local FLASH_COLOR = {1.0, 1.0, 1.0}
local BOAT_PARTICLES_BASE_X = Screen.getWidthPart(0.05)
local COLOR_WATER = ColorUtils.colorHex24ToStd(0x5AD7C4)


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function animateFullScreenFlash(self)
  local flashLayer = self.fullScreenFlash

  local flashSequence = AHSequence.new(
    AHOpacityTo.new(flashLayer, 0.35, AHEaseSimpleOut.new(), 0.2),
    AHDelay.new(0.2),
    AHOpacityTo.new(flashLayer, 0.35, nil, 0.0)
  )

  self.flashActionInterrupt:insert(flashSequence)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function MoonsetScene:moveScoreIcon(scoreToSet, maxScore)
  runInCoroutine(function()
      local boat = self.scoreIcon
      local xDifference = BOAT_X_END - BOAT_X_START
      local scoreProportion = math.min(1.0, (scoreToSet / maxScore))
      local boatDestinationX = scoreProportion * xDifference + BOAT_X_START
      local boatMoveAction = AHLocationXTo.new(boat, MoonsetCommon.BOAT_MOVE_DURATION, nil, boatDestinationX)

      local boatParticles = self.boatParticles
      local boatWidth = boat:getWidth()
      local particlesDestinationX = BOAT_PARTICLES_BASE_X + boatDestinationX - boatWidth
      local particlesMoveAction = boatParticles:createSourcePositionXTo(MoonsetCommon.BOAT_MOVE_DURATION, nil, particlesDestinationX)

      local particlesEmitDuration = MoonsetCommon.BOAT_MOVE_DURATION * 0.75
      local particlesEmitAction = boatParticles:createEmitAction(particlesEmitDuration)

      boat:startMovingAnimation()
      particlesMoveAction:start()
      particlesEmitAction:start()

      wait(boatMoveAction:start())
      wait(boat:startJoltAnimation())

      boat:startStationaryAnimation()
    end)
end

function MoonsetScene:startAnimation(duration)
  self.moon:markReferencePosition()
  self.moon:setDuration(duration)
  self.moon:start()
  self.background:startColorAnimation(duration)
end

function MoonsetScene:setPaused(isPaused)
  self.moon:setPaused(isPaused)
  self.background:setPaused(isPaused)
end

function MoonsetScene:showEndAnimation()
  self.background:showEndAnimation()
  local selfW = self:getDims()
  local bgW = self.background:getDims()
  local distance = bgW - selfW
  local endAction = AHSpawn.new(
    self.buttonBar:createFadeOutAction(),
    AHLocationXBy.new(self, 1, nil, -distance),
    AHLocationXBy.new(self.buttonBar, 1, nil, -distance)
  )

  endAction:start()
end

function MoonsetScene:showStartAnimation()
  local fadeAction = self.buttonBar:createFadeInAction()
  fadeAction:start()
end

function MoonsetScene:playCorrectAnimation(didFlagValid)
  Sound.playSoundEffect("correct")
  self.buttonBar:playCorrectAnimation(didFlagValid)
  self.prompt:playCorrectAnimation(didFlagValid)
  animateFullScreenFlash(self)
end

function MoonsetScene:playIncorrectAnimation(didFlagValid)
  Sound.playSoundEffect("incorrect")
  self.buttonBar:playIncorrectAnimation(didFlagValid)
  self.prompt:playIncorrectAnimation(didFlagValid)
end

function MoonsetScene:setPrompt(prompt)
  self.prompt = prompt
end

function MoonsetScene:updateTime(fractionElapsed)
  self.blendedGroup:updateTime(fractionElapsed)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, buttonBar)
  local background = MoonsetBackground.new(MoonsetCommon.HORIZON_Y)
  background:setAnchorPointBottomLeft()
  self.background = background

  local backgroundWidth, backgroundHeight = background:getDims()
  local blendCanvas = Group.new(backgroundWidth, backgroundHeight)
  blendCanvas:insert(background)

  local waterTexture = Sprite.newWithName("watertexture.png")
  waterTexture:setColor(unpack(COLOR_WATER))
  local waterTextureWidth, waterTextureHeight = waterTexture:getDims()
  waterTexture:setScl(Screen.fullWidth / waterTextureWidth, 1.0)
  waterTexture:setLoc(Screen.getWidthPart(0.5), MoonsetCommon.HORIZON_Y - waterTextureHeight * 0.5)
  blendCanvas:insert(waterTexture)

  local moon = MoonsetMoon.new(self:getWidthPart(MOON_RADIUS_MULTIPLIER))
  moon:setLoc(self:getWidthPart(0.5), MoonsetCommon.HORIZON_Y)
  blendCanvas:insert(moon)
  self.moon = moon

  local boatParticles = Particles.new("flowers.pex")
  boatParticles:setLoc(0.0, 0.0)
  boatParticles:setSourcePosition(BOAT_PARTICLES_BASE_X, MoonsetCommon.HORIZON_Y)
  boatParticles:start()
  blendCanvas:insert(boatParticles)
  self.boatParticles = boatParticles

  local scoreIcon = MoonsetSailboat.new()
  local scoreIconW, scoreIconH = scoreIcon:getDims()
  scoreIcon:setScl(BOAT_SCALE)
  scoreIcon:setLoc(BOAT_X_START, MoonsetCommon.HORIZON_Y + SAILBOAT_OFFSET_Y)
  blendCanvas:insert(scoreIcon)
  self.scoreIcon = scoreIcon
  self.scoreIcon:startStationaryAnimation()

  local blendedGroup = MoonsetOverlayBlend.new(blendCanvas)
  local x = blendedGroup:getWidthPart(0.5)
  local y = blendedGroup:getHeightPart(0.5)
  blendedGroup:setLoc(x, y)
  self:insert(blendedGroup)
  self.blendedGroup = blendedGroup

  buttonBar:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.12))
  self:insert(buttonBar)
  buttonBar:prepareForShowAction()
  self.buttonBar = buttonBar

  local fullScreenFlash = FilledRectangle.newFullScreen(FLASH_COLOR)
  fullScreenFlash:setOpacity(0.0)
  self:insertAtCenter(fullScreenFlash)
  self.fullScreenFlash = fullScreenFlash
end

function MoonsetScene.new(buttonBar)
  local self = Group.newFullScreen()
  Mixin.mix(self, MoonsetScene)

  createLayout(self, buttonBar)

  self.flashActionInterrupt = ActionInterrupt.new()

  return self
end
