--- NominaAudioPlayer
-- @author Created by Aaron Barrett on August 04, 2014.

class("NominaAudioPlayer")
require("data.visualization.SineWaveVisualizer")
require("data.generators.SoundAveragePowerGenerator")
require("NominaHumanCircle")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Kinds ------------------------------

NominaAudioPlayer.KIND_IMAGE_ONLY = "NominaAudioPlayer.KIND_IMAGE_ONLY"
NominaAudioPlayer.KIND_IMAGE_AND_NAMES = "NominaAudioPlayer.KIND_IMAGE_AND_NAMES"
NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS = "NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS"

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5

local VISUALIZER_HEIGHT = HEIGHT * 0.12

local VISUALIZER_Y_PROPORTION_INDEXED_BY_AUDIO_PLAYER_KIND = {
  [NominaAudioPlayer.KIND_IMAGE_ONLY] = 0.23,
  [NominaAudioPlayer.KIND_IMAGE_AND_NAMES] = 0.19,
  [NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS] = 0.10
}

local LABEL_MARGIN_Y_HUMAN_TO_NAME = NominaHumanCircle.CIRCLE_DIAMETER * 0.072
local LABEL_MARGIN_Y_NAME_TO_FACT = NominaHumanCircle.CIRCLE_DIAMETER * 0.074
local LABEL_MARGIN_Y_FACT_TO_FACT = NominaHumanCircle.CIRCLE_DIAMETER * 0.06
local FACT_LABEL_COUNT = 2

local VISUALIZER_LINE_THICKNESS = 0.5

-- Labels and Fonts -------------------

local NAME_LABEL_FONT = FontFaces.DIN_OT_MEDIUM
local NAME_LABEL_FONT_SIZE = 24.0

local FACT_LABEL_FONT = FontFaces.DIN_OT_LIGHT
local FACT_LABEL_FONT_SIZE = 20.0

-- Colors -----------------------------

local NAME_LABEL_COLOR = ColorUtils.colorHex24ToStd(0x2e9699)
local FACT_LABEL_COLOR = {0.28, 0.28, 0.28}
local VISUALIZER_COLOR = {0.6, 0.6, 0.6}

-- Durations --------------------------

local FADE_DURATION = 0.15

-- Z-Indices --------------------------

local Z_VISUALIZER = 10
local Z_FACT_LABEL = 20
local Z_NAME_LABEL = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getVisualizerYForAudioPlayerKind(audioPlayerKind)
  local visualizerYProportion = VISUALIZER_Y_PROPORTION_INDEXED_BY_AUDIO_PLAYER_KIND[audioPlayerKind]
  assert(visualizerYProportion ~= nil, "No visualizerYProportion provided for that audioPlayerKind")
  local visualizerY = visualizerYProportion * HEIGHT

  local minimumSafeY = Screen.safeBottom + (VISUALIZER_HEIGHT * 0.5)

  return math.max(visualizerY, minimumSafeY)
end

local function stopFadeOnLabel(label)
  if (label.fadeAction ~= nil) then
    label.fadeAction:stop()
  end
end

local function showLabelWithText(label, text)
  stopFadeOnLabel(label)

  label:setOpacity(0.0)
  label:setText(text)

  local fadeInAction = AHOpacityTo.new(label, FADE_DURATION, nil, 1.0)
  label.fadeAction = fadeInAction
  return fadeInAction:start()
end

local function clearLabel(label)
  stopFadeOnLabel(label)

  local fadeOutAction = AHOpacityTo.new(label, FADE_DURATION, nil, 0.0)
  label.fadeAction = fadeOutAction
  return fadeOutAction:start()
end

local function getFactLabel(self, factIndex)
  local factLabels = self.factLabels
  local factLabel = factLabels[factIndex]
  assert(factLabel ~= nil, "Not enough fact labels for fact index")
  return factLabel
end

local function waitUntilAudioTime(audio, time)
  waitUntil(function() return audio:getTime() >= time end)
end

local function waitUntilAudioIsFinished(audio)
  waitUntil(function() return audio:isFinished() end)
end

local function showNameForHuman(self, human)
  if (self.nameLabel == nil) then
    return
  end

  local name = human:getName()
  local nameLabel = self.nameLabel
  return showLabelWithText(nameLabel, name)
end

local function showFact(self, fact, factIndex)
  if (self.factLabels == nil) then
    return
  end

  local factLabel = getFactLabel(self, factIndex)

  local longVersion = fact:getLongVersion()
  local longVersionSentenceCased = StringUtils.capitalizeFirstLetter(longVersion)

  return showLabelWithText(factLabel, longVersionSentenceCased)
end

local function clearAllLabels(self)
  local nameLabel = self.nameLabel
  if (nameLabel ~= nil) then
    clearLabel(nameLabel)
  end

  local factLabels = self.factLabels
  if (factLabels ~= nil) then
    for _, factLabel in ipairs(factLabels) do
      clearLabel(factLabel)
    end
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function NominaAudioPlayer:playAudio()
  return runInCoroutine(function()
    local stage = self.stage
    local visualizer = self.visualizer

    local fadeInAction = AHOpacityTo.new(visualizer, 0.2, nil, 1.0)
    fadeInAction:start()

    local concept = self.concept
    local soundFilename = concept.sound_filename
    local audio = AssetHelper.playSoundForConcept(concept, soundFilename)

    local generator = SoundAveragePowerGenerator.new(audio)
    visualizer:setGenerator(generator)
    visualizer:start()
    self:insert(visualizer, Z_VISUALIZER) -- Have to insert the visualizer *after*
    -- calling start() to prevent it from jumping.  I would just make it invisible,
    -- but setVisible() doesn't seem to work on it.

    local humans = self.humans
    local previousHuman = nil
    for _, human in ipairs(humans) do
      local humanAppearTime = human:getAppearTime()
      waitUntilAudioTime(audio, humanAppearTime)

      if (previousHuman ~= nil) then
        clearAllLabels(self)
        stage:finishIntroducingHuman(previousHuman)
        waitFor(FADE_DURATION)
      end
      previousHuman = human

      stage:introduceHuman(human)
      showNameForHuman(self, human)

      local facts = human:getFacts()
      for factIndex, fact in ipairs(facts) do
        local factAppearTime = fact:getAppearTime()

        waitUntilAudioTime(audio, factAppearTime)

        showFact(self, fact, factIndex)
      end
    end

    waitUntilAudioIsFinished(audio)
    visualizer:stop()

    clearAllLabels(self)
    wait(stage:finishIntroducingHuman(previousHuman))

    local fadeOutAction = AHOpacityTo.new(visualizer, 0.2, nil, 0.0)
    wait(fadeOutAction:start())
    visualizer:removeSelf()
    self:removeSelf()
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaAudioPlayer.new(concept, humans, stage, audioPlayerKind)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaAudioPlayer)

  self.concept = concept
  self.humans = humans
  self.stage = stage

  local visualizer = SineWaveVisualizer.new(WIDTH,
                                            VISUALIZER_HEIGHT,
                                            VISUALIZER_COLOR,
                                            nil,
                                            nil,
                                            nil,
                                            VISUALIZER_LINE_THICKNESS)
  visualizer:setOpacity(0.0)
  self.visualizer = visualizer

  local visualizerY = getVisualizerYForAudioPlayerKind(audioPlayerKind)
  visualizer:setLoc(CENTER_X, visualizerY)

  local shouldShowNames = ((audioPlayerKind == NominaAudioPlayer.KIND_IMAGE_AND_NAMES) or (audioPlayerKind == NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS))
  if (shouldShowNames) then
    local nameLabel = Label.new("", Font.new(NAME_LABEL_FONT, NAME_LABEL_FONT_SIZE))
    local nameLabelY = stage:getIntroductionHumanCircleBottomY() - nameLabel:getHeightPart(0.5) - LABEL_MARGIN_Y_HUMAN_TO_NAME
    nameLabel:setLoc(CENTER_X, nameLabelY)
    nameLabel:setColor(unpack(NAME_LABEL_COLOR))
    nameLabel:setOpacity(0.0)
    self:insert(nameLabel, Z_NAME_LABEL)
    self.nameLabel = nameLabel

    local shouldShowFacts = (audioPlayerKind == NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS)
    if (shouldShowFacts) then
      local factLabels = {}
      local firstFactLabelYTop = nameLabel:getYBottom() - LABEL_MARGIN_Y_NAME_TO_FACT
      for factIndex = 1, FACT_LABEL_COUNT do
        local factLabel = Label.new("", Font.new(FACT_LABEL_FONT, FACT_LABEL_FONT_SIZE))
        local labelHeight = factLabel:getHeight()
        local labelOffsetY = ((LABEL_MARGIN_Y_FACT_TO_FACT + labelHeight) * (factIndex - 1))

        factLabel:setColor(unpack(FACT_LABEL_COLOR))
        factLabel:setOpacity(0.0)
        factLabel:setAnchorPointTop()
        factLabel:setLoc(CENTER_X, firstFactLabelYTop - labelOffsetY)
        self:insert(factLabel, Z_FACT_LABEL)

        table.insert(factLabels, factLabel)
      end

      self.factLabels = factLabels
    end
  end

  return self
end
