--- NominaCentrifuge
-- @author Created by Aaron Barrett on August 12, 2014.

class("NominaCentrifuge")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

local OUTER_SHUFFLE_RADIUS = HEIGHT * 0.1
local INNER_SHUFFLE_RADIUS = HEIGHT * 0.05

local SHUFFLE_DEGREES_TO_SPIN = 2160.0

local HUMAN_CIRCLE_COUNT = 4
local FULL_CIRLCE_IN_RADIANS = math.pi * 2.0
local HUMAN_CIRCLE_SPACING_IN_RADIANS = FULL_CIRLCE_IN_RADIANS / HUMAN_CIRCLE_COUNT

local HUMAN_CIRCLE_SPACING_INDEX_OFFSET = -7.0
local SHUFFLE_DISTANCE_TO_SHARPNESS_RATIO = 0.009

local OUTRO_RADIUS = HEIGHT * 0.04
local OUTRO_DEGREES_TO_SPIN = 2160.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function compareXLocation(humanCircleA, humanCircleB)
  local aX, _ = humanCircleA:getLoc()
  local bX, _ = humanCircleB:getLoc()
  local shouldHumanCircleAComeBeforeHumanCircleB = (aX < bX)
  return shouldHumanCircleAComeBeforeHumanCircleB
end

local function createSortedHumanCircles(humanCircles)
  local humanCirclesCopy = TableUtils.shallowCopy(humanCircles)
  table.sort(humanCirclesCopy, compareXLocation)
  return humanCirclesCopy
end

local function getCentrifugalLocation(humanCircleIndex, radius)
  local angle = HUMAN_CIRCLE_SPACING_IN_RADIANS * (humanCircleIndex + HUMAN_CIRCLE_SPACING_INDEX_OFFSET)
  local centrifugalX = CENTER_X + math.cos(angle) * radius
  local centrifugalY = CENTER_Y + math.sin(angle) * radius
  return centrifugalX, centrifugalY
end

local function createShuffleConvergeActionForHumanCircle(humanCircle, humanCircleIndex)
  local currentX, currentY = humanCircle:getLoc()

  local outerX, outerY = getCentrifugalLocation(humanCircleIndex, OUTER_SHUFFLE_RADIUS)
  local innerX, innerY = getCentrifugalLocation(humanCircleIndex, INNER_SHUFFLE_RADIUS)

  local distanceToOuterLocation = MathUtils.distance(currentX, currentY, outerX, outerY)
  local easeSharpness = distanceToOuterLocation * SHUFFLE_DISTANCE_TO_SHARPNESS_RATIO
  local outerEase = AHEaseSimpleInOut.new(easeSharpness)

  local moveToOuterLocationAction = AHLocationTo.new(humanCircle, 0.5, outerEase, outerX, outerY)
  local moveToInnerLocationAction = AHLocationTo.new(humanCircle, 0.5, nil, innerX, innerY)
  local convergeAction = AHSequence.new(
                                        moveToOuterLocationAction,
                                        moveToInnerLocationAction
                                        )
  return convergeAction
end

local function createShuffleShrinkActionForHumanCircle(humanCircle)
  local shrinkAction = AHScaleTo.new(humanCircle, 0.4, AHEaseSimpleInOut.new(4.0), 0.15)
  return shrinkAction
end

local function createOutroConvergeActionForHumanCircle(humanCircle, humanCircleIndex)
  local centrifugalX, centrifugalY = getCentrifugalLocation(humanCircleIndex, OUTRO_RADIUS)

  local moveToCentrifugalLocationAction = AHLocationTo.new(humanCircle, 1.1, AHEaseSimpleIn.new(2.0), centrifugalX, centrifugalY)
  local moveToCenterAction = AHLocationTo.new(humanCircle, 0.1, AHEaseSimpleOut.new(1.5), CENTER_X, CENTER_Y)
  local convergeAction = AHSequence.new(
                                        moveToCentrifugalLocationAction,
                                        moveToCenterAction
                                        )
  return convergeAction
end

local function createOutroShrinkActionForHumanCircle(humanCircle)
  local shrinkAction = AHScaleTo.new(humanCircle, 1.2, AHEaseSimpleIn.new(3.0), 0.15)
  return shrinkAction
end

local function takeHumanCircles(self, humanCircles)
  for _, humanCircle in ipairs(humanCircles) do
    self:takeChild(humanCircle)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function NominaCentrifuge:createShuffleActionWithHumanCircles(humanCircles)
  local sortedHumanCircles = createSortedHumanCircles(humanCircles)

  local spinAction = AHSpawn.new()

  for humanCircleIndex, humanCircle in ipairs(sortedHumanCircles) do
    local convergeAction = createShuffleConvergeActionForHumanCircle(humanCircle, humanCircleIndex)
    spinAction:addAction(convergeAction)
    local shrinkAction = createShuffleShrinkActionForHumanCircle(humanCircle)
    spinAction:addAction(shrinkAction)
    local revertAction = humanCircle:createRevertAction()
    spinAction:addAction(revertAction)
    local turnGrayAction = humanCircle:createTurnGrayAction(humanCircleIndex)
    spinAction:addAction(turnGrayAction)
  end

  local rotationAction = AHRotationBy.new(self, 0.85, AHEaseSimpleIn.new(4.0), SHUFFLE_DEGREES_TO_SPIN)
  local delayedRotationAction = AHSequence.new(
                                               AHDelay.new(0.15),
                                               rotationAction
                                               )
  spinAction:addAction(delayedRotationAction)

  local shuffleAction = AHSequence.new(
                                       function()
                                         takeHumanCircles(self, humanCircles)
                                       end,
                                       spinAction
                                       )

  return shuffleAction
end

function NominaCentrifuge:createOutroAction(humanCircles)
  local sortedHumanCircles = createSortedHumanCircles(humanCircles)

  local spinAction = AHSpawn.new()

  local rotationAction = AHRotationBy.new(self, 1.2, AHEaseSimpleIn.new(3.0), OUTRO_DEGREES_TO_SPIN)
  spinAction:addAction(rotationAction)

  for humanCircleIndex, humanCircle in ipairs(sortedHumanCircles) do
    local convergeAction = createOutroConvergeActionForHumanCircle(humanCircle, humanCircleIndex)
    spinAction:addAction(convergeAction)
    local shrinkAction = createOutroShrinkActionForHumanCircle(humanCircle)
    spinAction:addAction(shrinkAction)
    local turnGrayAction = humanCircle:createTurnGrayForOutroAction()
    spinAction:addAction(turnGrayAction)
  end

  local outroAction = AHSequence.new(
                                     function()
                                       takeHumanCircles(self, humanCircles)
                                     end,
                                     spinAction
                                     )

  return outroAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaCentrifuge.new()
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaCentrifuge)

  return self
end
