--- NominaGame
-- @author Created by Aaron Barrett on July 16, 2014.

class("NominaGame")
require("StandardUIBuilder")
require("NominaConfiguration")
require("NominaHuman")
require("NominaStage")
require("NominaAudioPlayer")
require("widgets.TapToContinue")
require("bonuses.ResponseSpeedTracker")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

-- Colors -----------------------------

local BACKGROUND_COLOR = {0.94, 0.94, 0.94}
local UI_COMPONENTS_COLOR = ColorUtils.colorHex24ToStd(0x8a8a8a)

-- Z Indices --------------------------

local Z_BACKGROUND = 10
local Z_STAGE = 20
local Z_AUDIO_PLAYER = 30

local function configureGameDifficulty()
  local passScore = NominaConfiguration.getPassScore()
  local rankTwoScore = NominaConfiguration.getRankTwoScore()
  local rankThreeScore = NominaConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(NominaConfiguration.getInitialLivesCount())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Select Items for Rounds ------------

local function getNamesFromHumans(humans)
  local names = {}
  for _, human in ipairs(humans) do
    local name = human:getName()
    table.insert(names, name)
  end
  return names
end

local function chooseNamesForHumanPair(humanPair)
  local allDistractorNamesSet = {}
  for _, human in ipairs(humanPair) do
    local distractorNames = human:getDistractorNames()
    TableUtils.joinArrayToSet(distractorNames, allDistractorNamesSet)
  end

  local chosenNames = {}
  for _, human in ipairs(humanPair) do
    local name = human:getName()
    table.insert(chosenNames, name)

    allDistractorNamesSet[name] = nil
  end

  local allDistractorNamesWithNoDuplicatesOrRealNames = TableUtils.createArrayFromSet(allDistractorNamesSet)

  local neededDistractorNamesCount = 2
  local chosenDistractorNames = TableUtils.createRandomSampleOfArray(allDistractorNamesWithNoDuplicatesOrRealNames,
                                                                     neededDistractorNamesCount)

  TableUtils.concatenateArrays(chosenNames, chosenDistractorNames)

  return chosenNames
end

local function getFirstHumanMatchingGender(remainingHumans, gender)
  for humanIndex, human in ipairs(remainingHumans) do
    if (human:getGender() == gender) then
      return human, humanIndex
    end
  end
  error("Ran out of humans matching gender `" .. tostring(gender) .. "`")
end

local function getRoundTwoHumanPairs(allHumans)
  local remainingHumans = TableUtils.shallowCopy(allHumans)
  remainingHumans = TableUtils.shuffleArray(remainingHumans)

  local firstHuman = table.remove(remainingHumans, 1)

  local firstGender = firstHuman:getGender()

  local secondHuman, secondHumanIndex = getFirstHumanMatchingGender(remainingHumans, firstGender)
  table.remove(remainingHumans, secondHumanIndex)

  local firstPair = {firstHuman, secondHuman}
  local secondPair = remainingHumans

  local roundTwoHumanPairs = {firstPair, secondPair}
  return roundTwoHumanPairs
end

local function getNamesToMatchForRoundOne(human)
  local namesToMatch = {}

  local actualName = human:getName()
  table.insert(namesToMatch, actualName)
  local distractorNames = human:getDistractorNames()
  TableUtils.concatenateArrays(namesToMatch, distractorNames)

  return namesToMatch
end

-- Game Loss --------------------------

local function loseGame(self)
  local stage = self.stage

  stage:setDragAndDropEnabled(false)

  self:showDefaultGameLossOutro()
  wait(stage:showLoss())

  GameSession.notifyGameLoss()

  -- This infinite delay is intentional and necessary.
  -- notifyGameLoss doesn't switch back to the app immediately, so
  -- if you don't wait for infinity after reporting the game loss,
  -- the first few frames of the next animation that would have played
  -- if you hadn't lost will start playing.
  waitFor(math.huge)
end


-- Correct and Incorrect Answers ------

local function updateScoreLabel(self)
  local scoreLabel = self.scoreLabel
  scoreLabel:setScore(GameSession.getScore())
end

local function scoreCorrectResponse(self)
  Sound.playSoundEffect("correct")

  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(NominaConfiguration.POINTS_PER_CORRECT_RESPONSE)

  updateScoreLabel(self)
end

local function scoreIncorrectResponse(self)
  local quieterVolume = 0.50
  Sound.playSoundEffect("incorrect", quieterVolume)

  GameSession.incrementIncorrectAnswerCount()

  local responseSpeedTracker = self.responseSpeedTracker
  responseSpeedTracker:notifyIncorrectResponse()

  local livesCounter = self.livesCounter
  livesCounter:decrement()

  local didLose = (livesCounter:getLives() <= 0)
  if (didLose) then
    loseGame(self)
  end
end

-- Wait For Audio ---------------------

local function waitForIntroAudio(self)
  local audioPlayer = self.audioPlayer
  wait(audioPlayer:playAudio())
  audioPlayer:removeSelf()
  self.audioPlayer = nil
end

-- Wait For Rounds --------------------

local function startQuickResponseBonus(self, countdownDuration)
  local speedBonusWindowDuration = countdownDuration * NominaConfiguration.SPEED_BONUS_DURATION_PROPORTION
  local decayWindowStart = speedBonusWindowDuration * NominaConfiguration.SPEED_BONUS_DECAY_WINDOW_START_PROPORTION
  local responseSpeedTracker = self.responseSpeedTracker
  responseSpeedTracker:setBonusWindowDuration(speedBonusWindowDuration)
  responseSpeedTracker:setDecayWindowStart(decayWindowStart)
  responseSpeedTracker:notifyStart()
end

local function waitForMatchingSessionWithoutMnemonics(self, humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages)
  local stage = self.stage

  wait(stage:showMatchingSession(humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages))
  startQuickResponseBonus(self, countdownDuration)

  stage:setDragAndDropEnabled(true)

  local remainingMatches = #humansToMatch

  repeat
    waitWhile(function() return stage:isMatchingInProgress() end)
    local wasResponseCorrect = stage:wasLastResponseCorrect()
    local didTimeExpire = stage:didTimeExpire()

    if (wasResponseCorrect) then
      scoreCorrectResponse(self)
      remainingMatches = remainingMatches - 1
      if (remainingMatches <= 0) then
        self.responseSpeedTracker:notifyCorrectResponse()
        updateScoreLabel(self)
        stage:setDragAndDropEnabled(false)
        stage:pauseCountdown()
        wait(stage:showCorrect())
      else
        stage:showCorrect()
      end
    else
      stage:setDragAndDropEnabled(false)
      scoreIncorrectResponse(self)
      stage:moveChosenNameTagHome()
      if (didTimeExpire) then
        stage:showTimeExpired()
      else
        stage:pauseCountdown()
        wait(stage:showIncorrect())
        wait(stage:revertIncorrect())
        stage:resumeCountdown()
      end
      stage:setDragAndDropEnabled(true)
    end

    stage:resumeMatching()

  until (remainingMatches <= 0)
  stage:setDragAndDropEnabled(false)
end

local function waitForRoundOne(self)
  local stage = self.stage
  local humans = self.humans
  local roundNumber = 1
  local countdownDuration = NominaConfiguration.getQuestionTime(roundNumber)

  for i, human in ipairs(humans) do
    local humansToMatch = {human}
    local namesToMatch = getNamesToMatchForRoundOne(human)
    local shouldIncludeShuffle = (i == 1)
    local shouldShowMysteryImages = false
    wait(stage:showMatchingSession(humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages))
    startQuickResponseBonus(self, countdownDuration)

    stage:setDragAndDropEnabled(true)
    waitWhile(function() return stage:isMatchingInProgress() end)
    local wasResponseCorrect = stage:wasLastResponseCorrect()
    local didTimeExpire = stage:didTimeExpire()
    stage:setDragAndDropEnabled(false)
    stage:pauseCountdown()

    if (wasResponseCorrect) then
      self.responseSpeedTracker:notifyCorrectResponse()
      scoreCorrectResponse(self)

      wait(stage:showCorrect())
    else
      scoreIncorrectResponse(self)

      stage:clearAllNameTags()

      if (didTimeExpire) then
        stage:showTimeExpired(human)
        local delayDurationBeforePresentingMnemonic = 0.15
        waitFor(delayDurationBeforePresentingMnemonic)
      else
        stage:showIncorrect()
        local delayDurationBeforePresentingMnemonic = 0.35
        waitFor(delayDurationBeforePresentingMnemonic)
      end

      wait(stage:presentMnemonic(human))
      waitUntil(function() return stage:isMnemonicFinished() end)
      wait(stage:dismissMnemonic())
    end

    stage:dismissMatchingSession()
  end
end

local function waitForRoundTwo(self)
  local stage = self.stage
  local humans = self.humans
  local humanPairs = getRoundTwoHumanPairs(humans)
  local roundNumber = 2
  local countdownDuration = NominaConfiguration.getQuestionTime(roundNumber)

  for i, humanPair in ipairs(humanPairs) do
    local humansToMatch = TableUtils.shallowCopy(humanPair)
    local namesToMatch = chooseNamesForHumanPair(humanPair)
    local shouldIncludeShuffle = (i == 1)
    local shouldShowMysteryImages = false
    waitForMatchingSessionWithoutMnemonics(self, humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages)

    stage:dismissMatchingSession()
  end
end

local function waitForRoundThree(self)
  local humans = self.humans
  local roundNumber = 3
  local countdownDuration = NominaConfiguration.getQuestionTime(roundNumber)

  local humansToMatch = TableUtils.shallowCopy(humans)
  local namesToMatch = getNamesFromHumans(humansToMatch)
  local shouldIncludeShuffle = true
  local shouldShowMysteryImages = true
  waitForMatchingSessionWithoutMnemonics(self, humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages)
end

-- Victory Outro ----------------------

local function waitForOutro(self)
  local stage = self.stage

  stage:pauseCountdown()

  self:showDefaultGameWinOutro()
  local delayedUIFadeAction = AHDelayed.new(1.8, self.uiOverlay:createFadeOutAction())
  delayedUIFadeAction:start()

  wait(stage:showOutro())
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function NominaGame:gameFlow()
  waitForIntroAudio(self)

  waitForRoundOne(self)
  waitForRoundTwo(self)
  waitForRoundThree(self)

  ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), NominaConfiguration.getExtraLifeBonusPoints(), self.bonusTray)
  updateScoreLabel(self)

  waitForOutro(self)

  GameSession.notifyGameWin()
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Build Humans -----------------------

local function buildHumans(self)
  local concept = self.concept

  local humans = {}

  for humanIndex = 1, NominaConfiguration.HUMANS_PER_GAME do
    local human = NominaHuman.new(concept, humanIndex)
    table.insert(humans, human)
  end

  self.humans = humans
end

-- Build Game Elements ----------------------

local function buildStage(self)
  local humans = self.humans
  local conceptId = self.concept.identifier
  local stage = NominaStage.new(humans, conceptId)
  self:insertAtCenter(stage, Z_STAGE)
  self.stage = stage
end

local function getAudioPlayerKindForDifficulty(difficulty)
  if (difficulty > NominaConfiguration.MAX_DIFFICULTY_TO_SHOW_NAMES) then
    return NominaAudioPlayer.KIND_IMAGE_ONLY
  elseif (difficulty > NominaConfiguration.MAX_DIFFICULTY_TO_SHOW_FACTS) then
    return NominaAudioPlayer.KIND_IMAGE_AND_NAMES
  end
  return NominaAudioPlayer.KIND_IMAGE_NAMES_AND_FACTS
end

local function buildAudioPlayer(self)
  local concept = self.concept
  local humans = self.humans
  local stage = self.stage

  local difficulty = GameSession.getDifficulty()
  local audioPlayerKind = getAudioPlayerKindForDifficulty(difficulty)
  local audioPlayer = NominaAudioPlayer.new(concept, humans, stage, audioPlayerKind)
  audioPlayer:setLoc(CENTER_X, CENTER_Y)
  self:insert(audioPlayer, Z_AUDIO_PLAYER)
  self.audioPlayer = audioPlayer
end

local function createGameElements(self)
  local background = FilledRectangle.new(WIDTH, HEIGHT, BACKGROUND_COLOR)
  self:insertAtCenter(background, Z_BACKGROUND)

  buildStage(self)
  buildAudioPlayer(self)
end

-- Build UI ---------------------------

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(NominaConfiguration.getInitialLivesCount())

  local function onQuickResponse(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 NominaConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end
  local bonusWindowDuration = nil -- Nomina sets the bonus window duration at the beginning of each matching session.
  local decayStart = nil -- Nomina sets the decay start at the beginning of each matching session.
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowDuration, onQuickResponse, decayStart)
end

-- Constructor ------------------------

function NominaGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, NominaGame)

  configureGameDifficulty()

  self.concept = GameSession.conceptGenerator:next()
  GameSession.reportSeenConcept(self.concept.identifier)
  GameSession.contentTracker:notifyPresentedConcept(self.concept.identifier)

  buildHumans(self)
  createGameElements(self)
  createUI(self)

  FontScaling.setAutoScaling(Screen.width, Screen.height)

  return self
end
