--- NominaHumanCircle
-- @author Created by Aaron Barrett on July 28, 2014.

class("NominaHumanCircle")

require("graphics.primitives.Checkmark")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

-- Circle dimensions are tuned to look acceptable across different resolutions.
-- Many other game elements will base their sizes off of the circle dimensions
-- so it's critical the circle proportions look correct (see NominaAudioPlayer.lua)
NominaHumanCircle.CIRCLE_DIAMETER = math.min(Screen.getWidthPart(0.58),
                                             Screen.getHeightPart(0.30))
NominaHumanCircle.CIRCLE_RADIUS = NominaHumanCircle.CIRCLE_DIAMETER * 0.5
NominaHumanCircle.BACK_CIRCLE_OVERSCALE_AMOUNT = 0.20

local WIDTH = NominaHumanCircle.CIRCLE_DIAMETER
local HEIGHT = NominaHumanCircle.CIRCLE_DIAMETER

local RING_THICKNESS = NominaHumanCircle.CIRCLE_RADIUS * 0.04
local RING_BLUR_MARGIN = NominaHumanCircle.CIRCLE_RADIUS * 0.02

local UNDERSCALE_FACTOR_FOR_ARTIFACT = 0.99

-- Colors -----------------------------

local CORRECT_COLOR = {0.46, 0.90, 0.72}
local INCORRECT_COLOR = {0.86, 0.29, 0.27}

local WHITE = {1.0, 1.0, 1.0}
local OUTRO_GRAY_COLOR = {0.50, 0.50, 0.50}

-- Round Scales -----------------------

local HOME_SCALE = 0.15

-- Z-Indices --------------------------

local Z_RING = 10
local Z_SCALE_PARENT = 20
local Z_BACK_CIRCLE = 30
local Z_AVATAR_IMAGE = 40
local Z_MYSTERY_IMAGE = 50
local Z_FRONT_CIRCLE = 60
local Z_CHECK_MARK = 70
local Z_CROSS = 80
local Z_GRAY_CIRCLE = 90

-- Gray Colors ------------------------

local GRAY_COLOR_INDEXED_BY_HUMAN_CIRCLE_INDEX = {
  [1] = {0.48, 0.48, 0.48},
  [2] = {0.52, 0.52, 0.52},
  [3] = {0.62, 0.62, 0.62},
  [4] = {0.67, 0.67, 0.67}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getGrayColorForHumanCircleIndex(humanCircleIndex)
  local grayColor = GRAY_COLOR_INDEXED_BY_HUMAN_CIRCLE_INDEX[humanCircleIndex]
  assert(grayColor ~= nil, "No gray color provided for that humanCircleIndex")
  return grayColor
end

local function createMoveHomeAction(self, duration, ease)
  local homeX = self.homeX
  local homeY = self.homeY

  local locationAction = AHLocationTo.new(self, duration, ease, homeX, homeY)
  local scaleDownAction = AHScaleTo.new(self, duration, ease, HOME_SCALE)
  local moveHomeAction = AHSpawn.new(
                                     locationAction,
                                     scaleDownAction
                                     )

  return moveHomeAction
end

local function createTurnGrayActionWithColor(self, grayColor)
  local grayCircle = self.grayCircle

  local turnGrayAction = AHSequence.new(
                                        function()
                                          grayCircle:setColor(unpack(grayColor))
                                          grayCircle:setOpacity(0.0)
                                          grayCircle:setVisible(true)
                                        end,
                                        AHOpacityTo.new(grayCircle, 0.2, nil, 1.0)
                                        )

  return turnGrayAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function NominaHumanCircle:setHomeLocation(homeX, homeY)
  self.homeX = homeX
  self.homeY = homeY
end

function NominaHumanCircle:getHuman()
  return self.human
end

-- Mystery Face Image -----------------

function NominaHumanCircle:showMysteryImage()
  local avatarImage = self.avatarImage
  avatarImage:setVisible(false)

  local mysteryImage = self.mysteryImage
  mysteryImage:setVisible(true)
end

-- Create Actions ---------------------

function NominaHumanCircle:createMoveHomeAfterIntroductionAction()
  local duration = 0.4
  local ease = AHEaseSimpleInOut.new(2.0)
  local moveHomeAction = createMoveHomeAction(self, duration, ease)
  return moveHomeAction
end

function NominaHumanCircle:createMoveHomeAfterShuffleAction()
  local duration = 0.2
  local ease = AHEaseSimpleOut.new(2.0)
  local moveHomeAction = createMoveHomeAction(self, duration, ease)
  return moveHomeAction
end

function NominaHumanCircle:createMoveHomeAfterMatchingSessionAction()
  return self:createMoveHomeAfterIntroductionAction()
end

function NominaHumanCircle:createCorrectAction()
  local frontCircle = self.frontCircle
  local checkMark = self.checkMark
  local ring = self.ring
  local scaleParent = self.scaleParent

  local correctAction = AHSequence.new(
                                       function()
                                         frontCircle:setColor(unpack(CORRECT_COLOR))
                                         frontCircle:setOpacity(0.0)
                                         frontCircle:setVisible(true)

                                         checkMark:setOpacity(0.0)
                                         checkMark:setVisible(true)

                                         ring:setVisible(true)
                                       end,
                                       AHSpawn.new(
                                                   AHOpacityTo.new(frontCircle, 0.1, nil, 1.0),
                                                   AHOpacityTo.new(checkMark, 0.1, nil, 1.0),
                                                   AHOpacityTo.new(ring, 0.75, AHEaseSimpleOut.new(2.0), 0.0),
                                                   AHSequence.new(
                                                                  AHDelay.new(0.05),
                                                                  AHScaleTo.new(ring, 0.75, AHEaseSimpleOut.new(2.0), 1.5)
                                                                  ),
                                                   AHSequence.new(
                                                                  AHScaleTo.new(scaleParent, 0.07, AHEaseSimpleOut.new(2.0), 1.05),
                                                                  AHScaleTo.new(scaleParent, 0.25, AHEaseSimpleInOut.new(2.0), 1.0)
                                                                  )
                                                   ),
                                       function()
                                         ring:setVisible(false)
                                         ring:setScl(1.0)
                                         ring:setOpacity(1.0)
                                       end
                                       )
  return correctAction
end

function NominaHumanCircle:createIncorrectAction()
  local frontCircle = self.frontCircle
  local backCircle = self.backCircle
  local cross = self.cross

  local incorrectAction = AHSequence.new(
                                         function()
                                           frontCircle:setColor(unpack(INCORRECT_COLOR))
                                           frontCircle:setOpacity(0.0)
                                           frontCircle:setVisible(true)

                                           cross:setOpacity(0.0)
                                           cross:setScl(1.0)
                                           cross:setVisible(true)

                                           backCircle:setScl(1.05)
                                           backCircle:setVisible(true)
                                         end,
                                         AHSpawn.new(
                                                     AHOpacityTo.new(frontCircle, 0.1, nil, 0.5),
                                                     AHOpacityTo.new(cross, 0.1, nil, 1.0),
                                                     AHScaleTo.new(backCircle, 0.5, AHEaseElasticOut.new(0.2), 1 + NominaHumanCircle.BACK_CIRCLE_OVERSCALE_AMOUNT),
                                                     AHScaleTo.new(cross, 0.5, AHEaseElasticOut.new(0.2), 0.9)
                                                     )
                                         )
  return incorrectAction
end

function NominaHumanCircle:createRevertAction()
  local frontCircle = self.frontCircle
  local checkMark = self.checkMark
  local cross = self.cross
  local backCircle = self.backCircle

  local revertAction = AHSequence.new(
                                      AHSpawn.new(
                                                  AHOpacityTo.new(frontCircle, 0.4, nil, 0.0),
                                                  AHOpacityTo.new(checkMark, 0.2, nil, 0.0),
                                                  AHOpacityTo.new(cross, 0.2, nil, 0.0),
                                                  AHScaleTo.new(backCircle, 0.2, AHEaseSimpleIn.new(2.0), UNDERSCALE_FACTOR_FOR_ARTIFACT)
                                                  ),
                                      function()
                                        frontCircle:setVisible(false)
                                        checkMark:setVisible(false)
                                        cross:setVisible(false)
                                        backCircle:setVisible(false)
                                      end
                                      )

  return revertAction
end

function NominaHumanCircle:createAppearAction()
  local avatarImage = self.avatarImage
  local appearAction = AHSpawn.new(
                                   AHOpacityTo.new(avatarImage, 0.4, nil, 1.0),
                                   AHScaleTo.new(self, 0.4, AHEaseSimpleInOut.new(2.0), 1.0)
                                   )
  return appearAction
end

function NominaHumanCircle:createMatchingSessionPresentationAction(container, destinationScale, shouldIncludeShuffle)
  local containerX, containerY = container:getLoc()
  local destinationX = containerX
  local destinationY = containerY + container:getHumanCircleYOffset()

  local duration = nil
  if (shouldIncludeShuffle) then
    duration = 0.2
  else
    duration = 0.4
  end

  local ease = AHEaseSimpleInOut.new(2.0)

  local moveAction = AHLocationTo.new(self, duration, ease, destinationX, destinationY)
  local scaleAction = AHScaleTo.new(self, duration, ease, destinationScale)

  local presentationAction =  AHSpawn.new(
                                          moveAction,
                                          scaleAction
                                          )
  return presentationAction
end

function NominaHumanCircle:createTurnGrayAction(humanCircleIndex)
  local grayColor = getGrayColorForHumanCircleIndex(humanCircleIndex)
  local turnGrayAction = createTurnGrayActionWithColor(self, grayColor)
  return turnGrayAction
end

function NominaHumanCircle:createTurnGrayForOutroAction()
  local grayColor = OUTRO_GRAY_COLOR
  local turnGrayAction = createTurnGrayActionWithColor(self, grayColor)
  return turnGrayAction
end

function NominaHumanCircle:createUngrayAction()
  local grayCircle = self.grayCircle
  local ungrayAction = AHSequence.new(
                                      AHOpacityTo.new(grayCircle, 0.2, nil, 0.0),
                                      function()
                                        grayCircle:setVisible(false)
                                      end
                                      )
  return ungrayAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaHumanCircle.new(human)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaHumanCircle)

  local ring = HollowCircle.new(NominaHumanCircle.CIRCLE_RADIUS, CORRECT_COLOR, RING_THICKNESS, RING_BLUR_MARGIN)
  ring:setVisible(false)
  self:insertAtCenter(ring, Z_RING)
  self.ring = ring

  local scaleParent = Group.new(WIDTH, HEIGHT)
  self:insertAtCenter(scaleParent, Z_SCALE_PARENT)
  self.scaleParent = scaleParent

  local backCircle = FilledCircle.new(NominaHumanCircle.CIRCLE_RADIUS, INCORRECT_COLOR)
  backCircle:setVisible(false)
  scaleParent:insertAtCenter(backCircle, Z_BACK_CIRCLE)
  self.backCircle = backCircle

  local avatarImageTexture = human:getAvatarTexture()
  local avatarImage = Sprite.newWithTexture(avatarImageTexture)
  avatarImage:setEffectiveDims(WIDTH * UNDERSCALE_FACTOR_FOR_ARTIFACT, WIDTH * UNDERSCALE_FACTOR_FOR_ARTIFACT)
  avatarImage:setOpacity(0.0)
  scaleParent:insertAtCenter(avatarImage, Z_AVATAR_IMAGE)
  self.avatarImage = avatarImage

  local mysteryImage = Sprite.newWithName("mystery.png")
  mysteryImage:setEffectiveDims(WIDTH * UNDERSCALE_FACTOR_FOR_ARTIFACT, WIDTH * UNDERSCALE_FACTOR_FOR_ARTIFACT)
  mysteryImage:setVisible(false)
  scaleParent:insertAtCenter(mysteryImage, Z_MYSTERY_IMAGE)
  self.mysteryImage = mysteryImage

  local frontCircle = FilledCircle.new(NominaHumanCircle.CIRCLE_RADIUS, WHITE)
  frontCircle:setVisible(false)
  scaleParent:insertAtCenter(frontCircle, Z_FRONT_CIRCLE)
  self.frontCircle = frontCircle

  local checkWidth = WIDTH * 0.6
  local shortSegmentProportion = 0.4
  local thicknessProportion = 0.25
  local checkMark = FilledCheckmark.new(checkWidth, shortSegmentProportion, thicknessProportion, WHITE)
  checkMark:setVisible(false)
  scaleParent:insertAtCenter(checkMark, Z_CHECK_MARK)
  self.checkMark = checkMark

  local cross = Sprite.newWithName("cross.png")
  cross:setVisible(false)
  cross:setColor(unpack(INCORRECT_COLOR))
  scaleParent:insertAtCenter(cross, Z_CROSS)
  self.cross = cross

  local grayCircle = FilledCircle.new(NominaHumanCircle.CIRCLE_RADIUS, WHITE)
  grayCircle:setVisible(false)
  scaleParent:insertAtCenter(grayCircle, Z_GRAY_CIRCLE)
  self.grayCircle = grayCircle

  self.human = human
  self.homeX = nil
  self.homeY = nil

  self:setScl(0.0)

  return self
end
