--- NominaNameTag
-- @author Created by Aaron Barrett on July 29, 2014.

class("NominaNameTag")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width * 0.46
local HEIGHT = Screen.height * 0.10

local GLOW_MARGIN = HEIGHT * 0.048
local GLOW_WIDTH = WIDTH + GLOW_MARGIN
local GLOW_HEIGHT = HEIGHT + GLOW_MARGIN

local FADE_Y_TRANSLATION_DISTANCE = HEIGHT * 0.3

local BACKGROUND_CORNER_RADIUS = WIDTH * 0.04

-- Colors -----------------------------

local GLOW_COLOR = {0.70, 0.82, 0.96}
local BACKGROUND_COLOR = {1.0, 1.0, 1.0}
local LABEL_COLOR = ColorUtils.colorHex24ToStd(0x6997c9)

-- Labels and Fonts -------------------

local LABEL_FONT = FontFaces.DIN_OT_MEDIUM
local LABEL_FONT_SIZE = 24.0

-- Z-Indices --------------------------

local Z_GLOW = 10
local Z_BACKGROUND = 20
local Z_LABEL = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createFadeAction(self, destinationOpacity, shouldIncludeTranslation)
  local label = self.label
  local background = self.background
  local fadeAction = AHSpawn.new(
                                 AHOpacityTo.new(label, 0.3, nil, destinationOpacity),
                                 AHOpacityTo.new(background, 0.3, nil, destinationOpacity)
                                 )
  if (shouldIncludeTranslation) then
    local translationAction = AHLocationYBy.new(self, 0.3, AHEaseSimpleInOut.new(2.0), FADE_Y_TRANSLATION_DISTANCE)
    fadeAction:addAction(translationAction)
  end

  return fadeAction
end

local function stopMovingHome(self)
  if (self.moveHomeAction ~= nil) then
    self.moveHomeAction:stop()
  end
end

local function stopGlow(self)
  if (self.glowAction ~= nil) then
    self.glowAction:stop()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function NominaNameTag:getName()
  return self.name
end

-- Play Actions -----------------------

function NominaNameTag:showGlow()
  stopGlow(self)

  local glow = self.glow
  local glowAction = AHSpawn.new(
                                 AHOpacityTo.new(glow, 0.1, nil, 1.0),
                                 AHRotationTo.new(self, 0.1, AHEaseSimpleInOut.new(2.0), 3.5)
                                 )
  self.glowAction = glowAction

  return glowAction:start()
end

function NominaNameTag:hideGlow()
  stopGlow(self)

  local glow = self.glow
  local glowAction = AHSpawn.new(
                                 AHOpacityTo.new(glow, 0.1, nil, 0.0),
                                 AHRotationTo.new(self, 0.1, AHEaseSimpleInOut.new(2.0), 0.0)
                                 )
  self.glowAction = glowAction

  return glowAction:start()
end

function NominaNameTag:moveHome()
  stopMovingHome(self)

  self:setZIndex(NominaNameTagHolder.Z_NAME_TAG_UNSELECTED)

  local destinationX = self.homeX
  local destinationY = self.homeY
  local moveHomeAction = AHLocationTo.new(self, 0.5, AHEaseSimpleInOut.new(3.0), destinationX, destinationY)
  self.moveHomeAction = moveHomeAction
  return moveHomeAction:start()
end

-- Create Actions ---------------------

function NominaNameTag:createAppearAction()
  local destinationOpacity = 1.0
  local shouldIncludeTranslation = true
  local appearAction = createFadeAction(self, destinationOpacity, shouldIncludeTranslation)
  return appearAction
end

function NominaNameTag:createDisappearAction()
  local destinationOpacity = 0.0
  local shouldIncludeTranslation = true
  local fadeOutAction = createFadeAction(self, destinationOpacity, shouldIncludeTranslation)

  local disappearAction = AHSequence.new(
                                         function()
                                           self:hideGlow()
                                         end,
                                         fadeOutAction,
                                         function()
                                           self:removeSelf()
                                         end
                                         )

  return disappearAction
end

function NominaNameTag:createCorrectAction()
  local destinationOpacity = 0.0
  local shouldIncludeTranslation = false
  local correctAction = createFadeAction(self, destinationOpacity, shouldIncludeTranslation)

  return correctAction
end

function NominaNameTag:createIncorrectAction()
  return self:createCorrectAction()
end

-- Prepare For Actions ----------------

function NominaNameTag:prepareToAppear()
  local background = self.background
  local label = self.label

  background:setOpacity(0.0)
  label:setOpacity(0.0)

  local homeX = self.homeX
  local homeY = self.homeY
  local initialY = homeY - FADE_Y_TRANSLATION_DISTANCE
  self:setLoc(homeX, initialY)
end

-- DragAndDropHandler Object Protocol -

-- notifyDNDPickUp(): Called when the object is picked up.
function NominaNameTag:notifyDNDPickUp()
  stopMovingHome(self)

  Sound.playSoundEffect("select")
  self:setZIndex(NominaNameTagHolder.Z_NAME_TAG_SELECTED)
  self:showGlow()
end

-- notifyDNDDrop(container): Called when the object is dropped. container will only be present if it was dropped inside.
function NominaNameTag:notifyDNDDrop(container)
  self:hideGlow()

  local wasDroppedInContainer = (container ~= nil)
  if (wasDroppedInContainer) then
    self.nameTagChosenCallback(self)
  else
    Sound.playSoundEffect("deselect")
    self:moveHome()

    self.nameTagDroppedOutsideAnyContainerCallback()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaNameTag.new(name, homeX, homeY, nameTagChosenCallback, nameTagDroppedOutsideAnyContainerCallback)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaNameTag)

  local glow = FilledRoundedRectangle.new(GLOW_WIDTH,
                                          GLOW_HEIGHT,
                                          BACKGROUND_CORNER_RADIUS,
                                          GLOW_COLOR)
  glow:setOpacity(0.0)
  self:insertAtCenter(glow, Z_GLOW)
  self.glow = glow

  local background = FilledRoundedRectangle.new(WIDTH,
                                                HEIGHT,
                                                BACKGROUND_CORNER_RADIUS,
                                                BACKGROUND_COLOR)
  self:insertAtCenter(background, Z_BACKGROUND)
  self.background = background

  local label = Label.new(name, Font.new(LABEL_FONT, LABEL_FONT_SIZE))
  label:setColor(unpack(LABEL_COLOR))
  self:insertAtCenter(label, Z_LABEL)
  self.label = label

  self.name = name
  self.homeX = homeX
  self.homeY = homeY
  self.nameTagChosenCallback = nameTagChosenCallback
  self.nameTagDroppedOutsideAnyContainerCallback = nameTagDroppedOutsideAnyContainerCallback

  return self
end
