--- NominaNameTagHolder
-- @author Created by Aaron Barrett on July 29, 2014.

class("NominaNameTagHolder")
require("NominaNameTag")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height * 0.22

local STATIONS = {
  {x = 0.26, y = 0.78}, {x = 0.74, y = 0.78},
  {x = 0.26, y = 0.28}, {x = 0.74, y = 0.28}
}
local STATIONS_COUNT = #STATIONS

-- Z-Indices --------------------------

NominaNameTagHolder.Z_NAME_TAG_UNSELECTED = 10
NominaNameTagHolder.Z_NAME_TAG_SELECTED = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getHomeLocationForNameIndex(nameIndex)
  local station = STATIONS[nameIndex]
  assert(station ~= nil, "No station provided for nameIndex: `" .. tostring(nameIndex) .. "`")
  local homeX = station.x * WIDTH
  local homeY = station.y * HEIGHT
  return homeX, homeY
end

local function createNameTagChosenCallback(self)
  local function nameTagChosenCallback(nameTag)
    self.chosenNameTag = nameTag

    local dragHelpOverlay = self.dragHelpOverlay
    dragHelpOverlay:hide()
  end

  return nameTagChosenCallback
end

local function createNameTagDroppedOutsideAnyContainerCallback(self)
  local function nameTagDroppedOutsideAnyContainerCallback()
    local dragHelpOverlay = self.dragHelpOverlay
    dragHelpOverlay:show()
  end

  return nameTagDroppedOutsideAnyContainerCallback
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Accessors --------------------------

function NominaNameTagHolder:getNameTagsSet()
  return self.nameTagsSet
end

function NominaNameTagHolder:getChosenNameTag()
  return self.chosenNameTag
end

-- Add/Remove Names -------------------

function NominaNameTagHolder:addNames(names)
  assert(#names == STATIONS_COUNT, "Number of names passed was not exact")

  local namesCopy = TableUtils.shallowCopy(names)
  local shuffledNames = TableUtils.shuffleArray(namesCopy)

  local nameTagsSet = self.nameTagsSet

  local nameTagChosenCallback = self.nameTagChosenCallback
  local nameTagDroppedOutsideAnyContainerCallback = self.nameTagDroppedOutsideAnyContainerCallback

  for nameIndex, name in ipairs(shuffledNames) do
    local homeX, homeY = getHomeLocationForNameIndex(nameIndex)

    local nameTag = NominaNameTag.new(name, homeX, homeY, nameTagChosenCallback, nameTagDroppedOutsideAnyContainerCallback)
    nameTag:prepareToAppear()
    self:insert(nameTag, NominaNameTagHolder.Z_NAME_TAG_UNSELECTED)
    nameTagsSet[nameTag] = true
  end
end

function NominaNameTagHolder:clearNameTagsSet()
  self.nameTagsSet = {}
  self.chosenNameTag = nil
end

-- Create Actions ---------------------

function NominaNameTagHolder:createAppearAction()
  local nameTagsSet = self.nameTagsSet

  local appearAction = AHSpawn.new()

  for nameTag in pairs(nameTagsSet) do
    local nameTagAppearAction = nameTag:createAppearAction()
    appearAction:addAction(nameTagAppearAction)
  end

  local initialDelay = AHDelay.new(0.2)
  local delayedAppearAction = AHSequence.new(
                                             initialDelay,
                                             appearAction
                                             )

  return delayedAppearAction
end

function NominaNameTagHolder:createDisappearAction()
  local disappearAction = AHSpawn.new()

  local nameTagsSet = self.nameTagsSet
  for nameTag in pairs(nameTagsSet) do
    local nameTagDisappearAction = nameTag:createDisappearAction()
    disappearAction:addAction(nameTagDisappearAction)
  end

  return disappearAction
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaNameTagHolder.new(dragHelpOverlay)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaNameTagHolder)

  self.dragHelpOverlay = dragHelpOverlay

  self.nameTagChosenCallback = createNameTagChosenCallback(self)
  self.nameTagDroppedOutsideAnyContainerCallback = createNameTagDroppedOutsideAnyContainerCallback(self)

  self.nameTagsSet = {}
  self.chosenNameTag = nil

  return self
end
