--- NominaNameWithFactsPanel
-- @author Created by Aaron Barrett on August 07, 2014.

class("NominaNameWithFactsPanel")
require("NominaHumanCircle")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------
local HEIGHT_PROPORTION = 0.47
local EXTRA_SLIDE_PROPORTION_Y = 0.02

-- Descriptors ------------------------

local DESCRIPTORS_INDEXED_BY_HUMAN_COUNT = {
  [1] = {
    nameLabelFontSize = 24.0,
    factLabelFontSize = 20.0,
    factsCount = 2,
    marginYHumanToNameLabelProportion = 0.07,
    marginYHumanToFactLabelProportion = 0.11,
    marginYFactLabelToFactLabelProportion = 0.02
  },
  [2] = {
    nameLabelFontSize = 24.0,
    factLabelFontSize = nil,
    factsCount = 0,
    marginYHumanToNameLabelProportion = 0.085,
    marginYHumanToFactLabelProportion = nil,
    marginYFactLabelToFactLabelProportion = nil
  },
  [4] = {
    nameLabelFontSize = 16.0,
    factLabelFontSize = 16.0,
    factsCount = 1,
    marginYHumanToNameLabelProportion = 0.06,
    marginYHumanToFactLabelProportion = 0.06,
    marginYFactLabelToFactLabelProportion = 0.02
  }
}

-- Fonts ------------------------------

local NAME_LABEL_FONT = FontFaces.DIN_OT_MEDIUM
local FACT_LABEL_FONT = FontFaces.DIN_OT_LIGHT

-- Colors -----------------------------

local CORRECT_COLOR = ColorUtils.colorHex24ToStd(0x298589)
local INCORRECT_COLOR = {0.86, 0.29, 0.27}
local FACT_LABEL_NORMAL_COLOR = {0.22, 0.22, 0.22}

-- Z-Indices --------------------------

local Z_FACT_LABEL = 10
local Z_NAME_LABEL = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createFadeAction(self, nameLabelDestinationOpacity, factLabelsDestinationOpacity)
  local fadeAction = AHSpawn.new()

  local nameLabel = self.nameLabel
  local nameLabelFadeAction = AHOpacityTo.new(nameLabel, 0.3, nil, nameLabelDestinationOpacity)
  fadeAction:addAction(nameLabelFadeAction)

  local factLabels = self.factLabels
  for _, factLabel in ipairs(factLabels) do
    local factLabelFadeAction = AHOpacityTo.new(factLabel, 0.3, nil, factLabelsDestinationOpacity)
    fadeAction:addAction(factLabelFadeAction)
  end

  return fadeAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Prepare for Actions ----------------

function NominaNameWithFactsPanel:prepareForCorrect()
  local nameLabel = self.nameLabel
  nameLabel:setColor(unpack(CORRECT_COLOR))
  nameLabel:setOpacity(0.0)
end

function NominaNameWithFactsPanel:prepareForMnemonic()
  local nameLabel = self.nameLabel
  nameLabel:setColor(unpack(INCORRECT_COLOR))
  nameLabel:setOpacity(0.0)
end

-- Create Actions ---------------------

function NominaNameWithFactsPanel:createAppearAction()
  local nameLabelDestinationOpacity = 0.0
  local factLabelsDestinationOpacity = 1.0
  local appearAction = createFadeAction(self, nameLabelDestinationOpacity, factLabelsDestinationOpacity)
  return appearAction
end

function NominaNameWithFactsPanel:createDisappearAction()
  local nameLabelDestinationOpacity = 0.0
  local factLabelsDestinationOpacity = 0.0
  local disappearAction = createFadeAction(self, nameLabelDestinationOpacity, factLabelsDestinationOpacity)
  return disappearAction
end

function NominaNameWithFactsPanel:createCorrectAction()
  local nameLabelDestinationOpacity = 1.0
  local factLabelsDestinationOpacity = 0.0
  local correctAction = createFadeAction(self, nameLabelDestinationOpacity, factLabelsDestinationOpacity)
  return correctAction
end

function NominaNameWithFactsPanel:createRevealNameAction()
  local nameLabelDestinationOpacity = 1.0
  local factLabelsDestinationOpacity = 1.0
  local revealNameAction = AHSpawn.new(
    function() self.nameLabel:offsetY(-self.backCircleExpansionThickness) end,
    createFadeAction(self, nameLabelDestinationOpacity, factLabelsDestinationOpacity)
  )

  local slideAmount = self.slideAmount
  local shouldSlide = (slideAmount ~= 0.0)
  if (shouldSlide) then
    local factLabels = self.factLabels
    for _, factLabel in ipairs(factLabels) do
      local slideAction = AHLocationYBy.new(factLabel, 0.3, AHEaseSimpleInOut.new(2.0), slideAmount)
      revealNameAction:addAction(slideAction)
    end
  end

  return revealNameAction
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Trimmed facts ----------------------

local function getTrimmedFactsForHumanAndFactsCount(human, factsCount)
  local facts = human:getFacts()
  local trimmedFacts = TableUtils.createRandomSampleOfArray(facts, factsCount)
  return trimmedFacts
end

-- Constructor ------------------------

function NominaNameWithFactsPanel.new(width, human, humanCount)
  local descriptor = DESCRIPTORS_INDEXED_BY_HUMAN_COUNT[humanCount]
  local height = width * HEIGHT_PROPORTION
  local centerX = width * 0.5

  local self = Group.new(width, height)
  Mixin.mix(self, NominaNameWithFactsPanel)

  local marginYHumanToNameLabel = descriptor.marginYHumanToNameLabelProportion * width

  local name = human:getName()
  local nameLabelFontSize = descriptor.nameLabelFontSize
  local nameLabel = Label.new(name, Font.new(NAME_LABEL_FONT, nameLabelFontSize))
  nameLabel:setOpacity(0.0)
  local nameLabelYLocation = height - marginYHumanToNameLabel
  nameLabel:setAnchorPointTop()
  nameLabel:setLoc(centerX, nameLabelYLocation)
  self:insert(nameLabel, Z_NAME_LABEL)
  self.nameLabel = nameLabel

  local factLabels = {}
  local factsCount = descriptor.factsCount
  local doesRoundShowFactLabels = (factsCount > 0)
  if (doesRoundShowFactLabels) then
    local trimmedFacts = getTrimmedFactsForHumanAndFactsCount(human, factsCount)
    local factLabelFontSize = descriptor.factLabelFontSize
    for i = 1, factsCount do
      local fact = trimmedFacts[i]
      local shortVersion = fact:getShortVersion()
      local shortVersionSentenceCased = StringUtils.capitalizeFirstLetter(shortVersion)
      local factLabel = Label.new(shortVersionSentenceCased, Font.new(FACT_LABEL_FONT, factLabelFontSize))

      local factLabelOffsetY = (i - 1) * (factLabel:getHeight() + descriptor.marginYFactLabelToFactLabelProportion * width)
      local marginYHumanToFactLabel = descriptor.marginYHumanToFactLabelProportion * width
      local factLabelY =  height - marginYHumanToFactLabel - factLabelOffsetY
      factLabel:setAnchorPointTop()
      factLabel:setLoc(centerX, factLabelY)

      factLabel:setColor(unpack(FACT_LABEL_NORMAL_COLOR))
      factLabel:setOpacity(0.0)

      self:insert(factLabel, Z_FACT_LABEL)
      table.insert(factLabels, factLabel)
    end
  end

  self.factLabels = factLabels

  self.backCircleExpansionThickness = NominaHumanCircle.BACK_CIRCLE_OVERSCALE_AMOUNT * width * 0.5
  self.slideAmount = -(self.backCircleExpansionThickness + nameLabel:getHeight() + EXTRA_SLIDE_PROPORTION_Y * width)

  return self
end
