--- NominaOutroBurst
-- @author Created by Aaron Barrett on August 04, 2014.

class("NominaOutroBurst")

require("Particles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_CIRCLE_RADIUS = WIDTH * 0.265
local CENTER_CIRCLE_STARTING_SCALE = 0.15

local RING_OUTER_RADIUS = HEIGHT * 1.05
local RING_THICKNESS_PROPORTION_INDEXED_BY_RING_INDEX = {
  0.32,
  0.20,
  0.09
}
local RING_BLUR_MARGIN = 3.0

-- Colors -----------------------------

local GRAY_COLOR = {0.50, 0.50, 0.50}
local GREEN_COLOR = {0.46, 0.90, 0.72}

-- Particles --------------------------

local PARTICLE_FILENAMES = {
  "filled_victory_particles.pex",
  "hollow_victory_particles.pex"
}

-- Z-Indices --------------------------

local Z_PARTICLES = 10
local Z_RING = 20
local Z_CENTER_CIRCLE = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Create Actions ---------------------

function NominaOutroBurst:createBurstAction()
  local burstAction = AHSpawn.new()

  local function emitParticles()
    Sound.playSoundEffect("win")

    local particleSystems = self.particleSystems
    local particleCountPerSystem = 64
    for _, particleSystem in ipairs(particleSystems) do
      particleSystem:surge(particleCountPerSystem)
    end
  end
  burstAction:addAction(emitParticles)

  local centerCircle = self.centerCircle
  local checkMark = self.checkMark
  local centerCircleAction = AHSpawn.new(
                                         AHScaleTo.new(centerCircle, 0.6, AHEaseElasticOut.new(0.3), 1.0),
                                         AHColorTo.new(centerCircle, 0.6, AHEaseSimpleOut.new(4.0), GREEN_COLOR),
                                         AHOpacityTo.new(checkMark, 0.3, AHEaseSimpleOut.new(2.0), 1.0)
                                         )

  burstAction:addAction(centerCircleAction)

  local rings = self.rings

  for ringIndex, ring in ipairs(rings) do
    local baseRingAppearDelayDuration = 0.04
    local ringAppearDelayDurationIncrement = 0.13
    local ringAppearDelayDuration = baseRingAppearDelayDuration + (ringIndex - 1.0) * ringAppearDelayDurationIncrement

    local ringAction = AHSequence.new(
                                      AHDelay.new(ringAppearDelayDuration),
                                      function()
                                        ring:setVisible(true)
                                      end,
                                      AHScaleTo.new(ring, 0.6, AHEaseSimpleOut.new(2.0), 1.0),
                                      function()
                                        ring:setVisible(false)
                                      end
                                      )

    burstAction:addAction(ringAction)
  end

  return burstAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaOutroBurst.new()
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaOutroBurst)

  local particleSystems = {}
  for _, particleFilename in ipairs(PARTICLE_FILENAMES) do
    local particleSystem = Particles.new(particleFilename)
    particleSystem:start()
    self:insertAtCenter(particleSystem, Z_PARTICLES)
    table.insert(particleSystems, particleSystem)
  end
  self.particleSystems = particleSystems

  local centerCircle = FilledCircle.new(CENTER_CIRCLE_RADIUS, GRAY_COLOR)
  centerCircle:setScl(CENTER_CIRCLE_STARTING_SCALE)
  self:insertAtCenter(centerCircle, Z_CENTER_CIRCLE)
  self.centerCircle = centerCircle

  local checkWidth = WIDTH * 0.35
  local shortSegmentProportion = 0.4
  local thicknessProportion = 0.25
  local color = {1.0, 1.0, 1.0}
  local checkMark = FilledCheckmark.new(checkWidth, shortSegmentProportion, thicknessProportion, color)

  checkMark:setOpacity(0.0)
  centerCircle:insertAtCenter(checkMark)
  self.checkMark = checkMark

  local rings = {}

  for _, ringThicknessProportion in ipairs(RING_THICKNESS_PROPORTION_INDEXED_BY_RING_INDEX) do
    local ringThickness = ringThicknessProportion * WIDTH
    local ring = HollowCircle.new(RING_OUTER_RADIUS, GREEN_COLOR, ringThickness, RING_BLUR_MARGIN)
    ring:setVisible(false)
    ring:setScl(0.0)
    self:insertAtCenter(ring, Z_RING)
    table.insert(rings, ring)
  end

  self.rings = rings

  return self
end
