--- NominaStage
-- @author Created by Aaron Barrett on July 29, 2014.

class("NominaStage")

require("NominaHumanCircle")
require("NominaContainer")
require("NominaNameTagHolder")
require("NominaMnemonicPanel")
require("NominaOutroBurst")
require("NominaCountdown")
require("NominaRedFlash")
require("NominaCentrifuge")
require("NominaDragHelpOverlay")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5

local HOME_Y = math.min(Screen.height * 0.89,
                        Screen.safeTop - Screen.height * 0.095)


local NAME_TAG_HOLDER_Y = HEIGHT * 0.13

local INTRODUCTION_Y = HEIGHT * 0.59

-- Layouts ----------------------------

local HOME_X_PROPORTIONS = {
  0.2, 0.4, 0.6, 0.8
}

local LAYOUTS_INDEXED_BY_HUMAN_COUNT = {
	[1] = {{xProportion = 0.5, yProportion = 0.58}},

	[2] = {{xProportion = 0.26, yProportion = 0.55}, {xProportion = 0.74, yProportion = 0.55}},

	[4] = {{xProportion = 0.26, yProportion = 0.735}, {xProportion = 0.74, yProportion = 0.735},
				 {xProportion = 0.26, yProportion = 0.486}, {xProportion = 0.74, yProportion = 0.486}}
}

local HUMAN_CIRCLE_SCALE_INDEXED_BY_HUMAN_COUNT = {
  [1] = 0.85,
  [2] = 0.67,
  [4] = 0.5
}

-- Z-Indices --------------------------

local Z_COUNTDOWN = 10
local Z_CONTAINER = 20
local Z_HUMAN_CIRCLE_BACK = 30
local Z_HUMAN_CIRCLE_PRESENTED = 40
local Z_CENTRIFUGE = 50
local Z_DRAG_HELP_OVERLAY = 60
local Z_NAME_TAG_HOLDER = 70
local Z_OUTRO_BURST = 80
local Z_MNEMONIC_PANEL = 90
local Z_RED_FLASH = 100

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Play Sound Effects -----------------

local function playShuffleSound()
  Sound.playSoundEffect("shuffle")
end

local function playSwishSound()
  local quieterVolume = 0.50
  Sound.playSoundEffect("swish", quieterVolume)
end

-- Access Values from Constant Tables -

local function getHomeXForHumanIndex(humanIndex)
  local homeXProportion = HOME_X_PROPORTIONS[humanIndex]
  assert(homeXProportion ~= nil, "No homeX for humanIndex: `" .. humanIndex .. "`")
  local homeX = homeXProportion * WIDTH
  return homeX
end

local function getHumanCircleScaleForHumanCount(humanCount)
  local scale = HUMAN_CIRCLE_SCALE_INDEXED_BY_HUMAN_COUNT[humanCount]
  assert(scale ~= nil, "No scale provided for human count: `" .. tostring(humanCount) .. "`")
  return scale
end

local function getContainerWidthForHumanCount(humanCount)
  local humanCircleWidth = NominaHumanCircle.CIRCLE_DIAMETER * getHumanCircleScaleForHumanCount(humanCount)
  return humanCircleWidth
end

local function getLayoutForHumanCount(humanCount)
  local layout = LAYOUTS_INDEXED_BY_HUMAN_COUNT[humanCount]
  assert(layout ~= nil, "No layout provided for human count: `" .. tostring(humanCount) .. "`")
  return layout
end

local function getLocationForHumanCountAndHumanIndex(humanCount, humanIndex)
  local layout = getLayoutForHumanCount(humanCount)
  local station = layout[humanIndex]
  assert(station ~= nil, "No station in layout for human count: `" .. tostring(humanCount) .. "` and humanIndex: `" .. tostring(humanIndex) .. "`")
  local x = station.xProportion * WIDTH
  local y = station.yProportion * HEIGHT
  return x, y
end

-- Access HumanCircles By Human -------

local function getHumanCircleForHuman(self, human)
  local humanCirclesIndexedByHuman = self.humanCirclesIndexedByHuman
  local humanCircle = humanCirclesIndexedByHuman[human]
  assert(humanCircle ~= nil, "No human circle found")
  return humanCircle
end

local function getHumanCirclesExcludedFromCurrentMatchingSession(self)
  local containers = self.containers
  local allHumanCircles = self.humanCircles
  local excludedHumanCircles = TableUtils.shallowCopy(allHumanCircles)

  for _, container in ipairs(containers) do
    local includedHumanCircle = container:getHumanCircle()
    TableUtils.removeValueFromArray(excludedHumanCircles, includedHumanCircle)
  end

  return excludedHumanCircles
end

-- NameTag Dropped Callback -----------

local function createNameTagDroppedCallback(self, container, humanCircle, human)
  local function nameTagDroppedCallback(nameTag)
    local containerName = human:getName()
    local previouslyMatchedIncorrectly = container.previouslyMatchedIncorrectly
    local tagName = nameTag:getName()
    local doesMatch = (containerName == tagName)
    local conceptId = self.conceptIdentifier

    if (doesMatch) then
      if (not previouslyMatchedIncorrectly) then
        GameSession.contentTracker:notifyCorrectAnswer(conceptId, containerName)
      end
    else
      container.previouslyMatchedIncorrectly = true
      GameSession.contentTracker:notifyIncorrectAnswer(conceptId, containerName, tagName)
    end

    self.wasResponseCorrect = doesMatch
    self.chosenHumanCircle = humanCircle
    self.chosenContainer = container
    self.timeExpired = false
    self.isMatching = false
  end

  return nameTagDroppedCallback
end

-- Add/Remove Containers and NameTags -

local function removeChosenContainerAndNameTag(self)
  local dragAndDropHandler = self.dragAndDropHandler

  local nameTagHolder = self.nameTagHolder
  local chosenNameTag = nameTagHolder:getChosenNameTag()
  dragAndDropHandler:removeObject(chosenNameTag)

  local chosenContainer = self.chosenContainer
  dragAndDropHandler:removeContainer(chosenContainer)
end

local function removeAllContainersAndNameTags(self)
  local dragAndDropHandler = self.dragAndDropHandler

  local nameTagHolder = self.nameTagHolder
  local nameTagsSet = nameTagHolder:getNameTagsSet()
  for nameTag in pairs(nameTagsSet) do
    dragAndDropHandler:removeObject(nameTag)
  end

  local containers = self.containers
  for _, container in ipairs(containers) do
    dragAndDropHandler:removeContainer(container)
  end
end

local function addAllContainersAndNameTags(self)
  local dragAndDropHandler = self.dragAndDropHandler
  local nameTagHolder = self.nameTagHolder

  local containers = self.containers
  local nameTagsSet = nameTagHolder:getNameTagsSet()

  for _, container in ipairs(containers) do
    dragAndDropHandler:addContainer(container)
  end

  for nameTag in pairs(nameTagsSet) do
    dragAndDropHandler:addObject(nameTag)
  end
end

local function addContainerAtLayoutLocation(self, human, humanCount, humanIndex)
  local humanCircle = getHumanCircleForHuman(self, human)

  local containers = self.containers

  local containerX, containerY = getLocationForHumanCountAndHumanIndex(humanCount, humanIndex)

  local containerWidth = getContainerWidthForHumanCount(humanCount)
  local container = NominaContainer.new(containerWidth, humanCircle, humanCount)
  container:setNameTagDroppedCallback(createNameTagDroppedCallback(self, container, humanCircle, human))
  container:setLoc(containerX, containerY)
  self:insert(container, Z_CONTAINER)
  table.insert(containers, container)

  return container
end

-- Shuffle Animation ------------------

local function shuffleHomeLocations(self)
  local humanCircles = self.humanCircles
  local shuffledHumanCircles = TableUtils.createShuffledArray(humanCircles)

  for shuffledHumanIndex, humanCircle in ipairs(shuffledHumanCircles) do
    local homeX = getHomeXForHumanIndex(shuffledHumanIndex)
    humanCircle:setHomeLocation(homeX, HOME_Y)
  end
end

local function createShuffleAction(self)
  local humanCircles = self.humanCircles
  local centrifuge = self.centrifuge

  local centrifugeAction = centrifuge:createShuffleActionWithHumanCircles(humanCircles)

  local function retakeHumanCircles()
    for _, humanCircle in ipairs(humanCircles) do
      self:takeChild(humanCircle, Z_HUMAN_CIRCLE_BACK)
    end
  end

  local shuffleAction = AHSequence.new(
                                       playShuffleSound,
                                       centrifugeAction,
                                       retakeHumanCircles
                                       )

  return shuffleAction
end

-- Matching Session Presentation ------

local function bringPresentedHumanCirclesToFront(self)
  local containers = self.containers
  for _, container in ipairs(containers) do
    local humanCircle = container:getHumanCircle()
    humanCircle:setZIndex(Z_HUMAN_CIRCLE_PRESENTED)
  end
end

local function createPresentHumanCirclesAction(self, shouldIncludeShuffle)
  local containers = self.containers
  local humanCount = #containers

  local humanCircleDestinationScale = getHumanCircleScaleForHumanCount(humanCount)

  local appearAction = AHSpawn.new()

  for _, container in ipairs(containers) do
    local humanCircle = container:getHumanCircle()

    local containerAction = container:createMatchingSessionPresentationAction()
    appearAction:addAction(containerAction)

    local humanCircleAction = humanCircle:createMatchingSessionPresentationAction(container, humanCircleDestinationScale, shouldIncludeShuffle)
    appearAction:addAction(humanCircleAction)
  end

  return appearAction
end

local function createPresentExcludedHumanCirclesAction(self, shouldIncludeShuffle)
  local remainingHumanCircles = getHumanCirclesExcludedFromCurrentMatchingSession(self)

  local presentAction = AHSpawn.new()

  for _, humanCircle in ipairs(remainingHumanCircles) do
    local moveHomeAction = nil
    if (shouldIncludeShuffle) then
      moveHomeAction = humanCircle:createMoveHomeAfterShuffleAction()
    else
      moveHomeAction = humanCircle:createMoveHomeAfterMatchingSessionAction()
    end

    presentAction:addAction(moveHomeAction)
  end

  return presentAction
end

local function createUngrayAllHumanCirclesAction(self)
  local humanCircles = self.humanCircles

  local ungrayAction = AHSpawn.new()

  for _, humanCircle in ipairs(humanCircles) do
    local humanCircleUngrayAction = humanCircle:createUngrayAction()
    ungrayAction:addAction(humanCircleUngrayAction)
  end

  return ungrayAction
end


local function startCountdownWithDuration(self, countdownDuration)
  Log.debug("Starting countdown lasting " .. tostring(countdownDuration) .. " seconds.")
  local countdown = self.countdown
  countdown:setDuration(countdownDuration)
  countdown:start()
end

local function showMysteryImages(self)
  local humanCircles = self.humanCircles
  for _, humanCircle in ipairs(humanCircles) do
    humanCircle:showMysteryImage()
  end
end

local function prepareForMatchingSession(self, humansToMatch, namesToMatch)
  return runInCoroutine(function()
    local shuffledHumansToMatch = TableUtils.createShuffledArray(humansToMatch)
    local humanCount = #shuffledHumansToMatch

    for humanIndex, human in ipairs(shuffledHumansToMatch) do
      addContainerAtLayoutLocation(self, human, humanCount, humanIndex)

      -- Yield for one frame after adding each container so that the framerate
      -- doesn't drop at the beginning of the matching session presentation
      coroutine.yield()
    end

    local nameTagHolder = self.nameTagHolder
    nameTagHolder:addNames(namesToMatch)

    self.timeExpired = false
    self.isMatching = true
    self.wasResponseCorrect = false
    self.canResumeCountdown = true
    self.chosenHumanCircle = nil
    self.chosenContainer = nil
  end)
end

local function createShowMatchingSessionAction(self, countdownDuration, shouldIncludeShuffle)
  local presentHumanCirclesAction = createPresentHumanCirclesAction(self, shouldIncludeShuffle)

  local presentExcludedHumanCirclesAction = createPresentExcludedHumanCirclesAction(self, shouldIncludeShuffle)

  local ungrayAction = createUngrayAllHumanCirclesAction(self)

  local nameTagHolder = self.nameTagHolder
  local nameTagHolderAction = nameTagHolder:createAppearAction()


  local appearAction = AHSpawn.new(
                                   presentHumanCirclesAction,
                                   presentExcludedHumanCirclesAction,
                                   ungrayAction,
                                   nameTagHolderAction
                                   )

  local showMatchingSessionAction = AHSequence.new(
                                                   function()
                                                     playSwishSound()
                                                     bringPresentedHumanCirclesToFront(self)
                                                   end,
                                                   appearAction,
                                                   function()
                                                     addAllContainersAndNameTags(self)
                                                     startCountdownWithDuration(self, countdownDuration)
                                                   end
                                                   )

  return showMatchingSessionAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Audio Introduction -----------------

function NominaStage:introduceHuman(human)
  local humanCircle = getHumanCircleForHuman(self, human)
  local appearAction = humanCircle:createAppearAction()
  return appearAction:start()
end

function NominaStage:finishIntroducingHuman(human)
  local humanCircle = getHumanCircleForHuman(self, human)
  local moveHomeAction = humanCircle:createMoveHomeAfterIntroductionAction()

  return moveHomeAction:start()
end

function NominaStage:getIntroductionHumanCircleBottomY()
  return INTRODUCTION_Y - NominaHumanCircle.CIRCLE_RADIUS
end

-- Enable/Disable Drag And Drop -------

function NominaStage:setDragAndDropEnabled(enabled)
  local dragAndDropHandler = self.dragAndDropHandler
  dragAndDropHandler:setEnabled(enabled)
end

-- Matching Session Presentation ------

function NominaStage:showMatchingSession(humansToMatch, namesToMatch, countdownDuration, shouldIncludeShuffle, shouldShowMysteryImages)
  return runInCoroutine(function()

    wait(prepareForMatchingSession(self, humansToMatch, namesToMatch))

    local matchingSessionAction = AHSequence.new()

    if (shouldIncludeShuffle) then
      shuffleHomeLocations(self)

      local shuffleAction = createShuffleAction(self)
      matchingSessionAction:addAction(shuffleAction)
    end

    if (shouldShowMysteryImages) then
      matchingSessionAction:addAction(function()
                                        showMysteryImages(self)
                                      end)
    end

    local showMatchingSessionAction = createShowMatchingSessionAction(self, countdownDuration, shouldIncludeShuffle)
    matchingSessionAction:addAction(showMatchingSessionAction)

    wait(matchingSessionAction:start())
  end)
end

-- Matching Session Dismissal ---------

function NominaStage:dismissMatchingSession()
  local dismissAction = AHSpawn.new()

  local containers = self.containers
  for _, container in ipairs(containers) do
    local containerAction = container:createDisappearAction()
    dismissAction:addAction(containerAction)
  end

  local nameTagHolder = self.nameTagHolder
  local nameTagsAction = nameTagHolder:createDisappearAction()
  dismissAction:addAction(nameTagsAction)

  local countdown = self.countdown
  local refillAction = countdown:createRefillAction()
  dismissAction:addAction(refillAction)

  removeAllContainersAndNameTags(self)

  self.containers = {}
  nameTagHolder:clearNameTagsSet()

  return dismissAction:start()
end

-- Correct/Incorrect Answers ----------

function NominaStage:showCorrect()
  removeChosenContainerAndNameTag(self)

  local nameTagHolder = self.nameTagHolder
  local chosenNameTag = nameTagHolder:getChosenNameTag()
  local nameTagAction = chosenNameTag:createCorrectAction()

  local chosenHumanCircle = self.chosenHumanCircle
  local humanCircleAction = chosenHumanCircle:createCorrectAction()

  local chosenContainer = self.chosenContainer
  chosenContainer:prepareForCorrect()
  local containerAction = chosenContainer:createCorrectAction()

  local correctAction = AHSpawn.new(
                                    nameTagAction,
                                    humanCircleAction,
                                    containerAction
                                    )

  return correctAction:start()
end

function NominaStage:showIncorrect()
  local chosenHumanCircle = self.chosenHumanCircle
  local incorrectAction = chosenHumanCircle:createIncorrectAction()

  return incorrectAction:start()
end

function NominaStage:showTimeExpired(human)
  local timeExpiredAction = AHSpawn.new()

  local redFlash = self.redFlash
  local flashAction = redFlash:createFlashAction()
  timeExpiredAction:addAction(flashAction)

  if (human ~= nil) then
    local humanCircle = getHumanCircleForHuman(self, human)
    local humanCircleAction = humanCircle:createIncorrectAction()
    timeExpiredAction:addAction(humanCircleAction)
  end

  return timeExpiredAction:start()
end

-- Name Tag Management ----------------

function NominaStage:moveChosenNameTagHome()
  local nameTagHolder = self.nameTagHolder
  local chosenNameTag = nameTagHolder:getChosenNameTag()

  if (chosenNameTag ~= nil) then
    chosenNameTag:moveHome()
  end
end

function NominaStage:clearAllNameTags()
  removeAllContainersAndNameTags(self)

  local nameTagHolder = self.nameTagHolder

  local clearAllNameTagsAction = nameTagHolder:createDisappearAction()

  return clearAllNameTagsAction:start()
end

-- Flags for gameFlow Control ---------

function NominaStage:isMatchingInProgress()
  return self.isMatching
end

function NominaStage:wasLastResponseCorrect()
  return self.wasResponseCorrect
end

function NominaStage:didTimeExpire()
  return self.timeExpired
end

-- Matching Controls ------------------

function NominaStage:resumeMatching()
  self.chosenHumanCircle = nil
  self.chosenContainer = nil

  self.isMatching = true
end

-- Mnemonic Panels --------------------

function NominaStage:presentMnemonic(human)
  local mnemonicText = human:getMnemonic()
  local mnemonicPanel = NominaMnemonicPanel.new(mnemonicText)
  self:insertAtCenter(mnemonicPanel, Z_MNEMONIC_PANEL)
  self.mnemonicPanel = mnemonicPanel

  local mnemonicPresentAction = mnemonicPanel:createPresentAction()

  local containers = self.containers
  local firstContainer = containers[1]
  firstContainer:prepareForMnemonic()
  local revealNameAction = firstContainer:createRevealNameAction()

  local presentAction = AHSpawn.new(
                                    mnemonicPresentAction,
                                    revealNameAction
                                    )

  return presentAction:start()
end

function NominaStage:isMnemonicFinished()
  local mnemonicPanel = self.mnemonicPanel
  local finished = mnemonicPanel:isFinished()
  return finished
end

function NominaStage:dismissMnemonic()
  local mnemonicPanel = self.mnemonicPanel
  local dismissAction = mnemonicPanel:createDismissAction()
  return dismissAction:start()
end

-- Revert HumanCircles ----------------

function NominaStage:revertAllHumanCircles()
  local humanCircles = self.humanCircles

  local revertAction = AHSpawn.new()
  for _, humanCircle in ipairs(humanCircles) do
    local humanCircleRevertAction = humanCircle:createRevertAction()
    revertAction:addAction(humanCircleRevertAction)
  end

  return revertAction:start()
end

function NominaStage:revertIncorrect()
  local chosenHumanCircle = self.chosenHumanCircle
  local revertAction = chosenHumanCircle:createRevertAction()

  return revertAction:start()
end

-- Pause/Resume Countdown -------------

function NominaStage:pauseCountdown()
  local countdown = self.countdown
  countdown:pause()
end

function NominaStage:resumeCountdown()
  if (self.canResumeCountdown) then
    local countdown = self.countdown
    countdown:resume()
  end
end

-- Loss Outro -------------------------

function NominaStage:showLoss()
  local lossAction = AHSpawn.new()

  local containers = self.containers
  for _, container in pairs(containers) do
    local containerDisappearAction = container:createDisappearAction()
    lossAction:addAction(containerDisappearAction)
  end

  local nameTagHolder = self.nameTagHolder
  local clearAllNameTagsAction = nameTagHolder:createDisappearAction()
  lossAction:addAction(clearAllNameTagsAction)

  local chosenHumanCircle = self.chosenHumanCircle
  if (chosenHumanCircle ~= nil) then
    local incorrectAction = chosenHumanCircle:createIncorrectAction()
    lossAction:addAction(incorrectAction)
  end

  return lossAction:start()
end

-- Victory Outro ----------------------

function NominaStage:showOutro()
  playShuffleSound()

  local centrifuge = self.centrifuge
  local countdown = self.countdown
  local humanCircles = self.humanCircles
  local containers = self.containers

  local outroBurst = NominaOutroBurst.new()
  outroBurst:setLoc(CENTER_X, INTRODUCTION_Y)

  local countdownAction = countdown:createFadeOutAction()

  local spinAction = AHSpawn.new()

  local centrifugeAction = centrifuge:createOutroAction(humanCircles)
  spinAction:addAction(centrifugeAction)

  for _, container in ipairs(containers) do
    local containerAction = container:createDisappearAction()
    spinAction:addAction(containerAction)
  end

  local burstAction = outroBurst:createBurstAction()
  local delayBeforeAddingBurst = AHDelay.new(1.15)
  local addAndShowBurstAction = AHSequence.new(
                                               delayBeforeAddingBurst,
                                               function()
                                                 self:insert(outroBurst, Z_OUTRO_BURST)
                                               end,
                                               burstAction
                                               )
  spinAction:addAction(addAndShowBurstAction)

  local outroAction = AHSequence.new(
                                     countdownAction,
                                     spinAction
                                     )

  return outroAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Build Elements ---------------------

local function buildCountdown(self)
  local function onTimeoutCallback()
    local dragHelpOverlay = self.dragHelpOverlay
    dragHelpOverlay:hide()

    self.canResumeCountdown = false
    self.timeExpired = true
    self.wasResponseCorrect = false
    self.isMatching = false
  end

  local countdown = NominaCountdown.new(onTimeoutCallback)
  self:insertAtCenter(countdown, Z_COUNTDOWN)
  self.countdown = countdown
end

local function buildHumanCircles(self, humans)
  local humanCircles = {}
  local humanCirclesIndexedByHuman = {}

  for humanIndex, human in ipairs(humans) do
    local humanCircle = NominaHumanCircle.new(human)
    humanCircle:setLoc(CENTER_X, INTRODUCTION_Y)
    self:insert(humanCircle, Z_HUMAN_CIRCLE_BACK)
    table.insert(humanCircles, humanCircle)
    humanCirclesIndexedByHuman[human] = humanCircle

    local homeX = getHomeXForHumanIndex(humanIndex)
    humanCircle:setHomeLocation(homeX, HOME_Y)
  end

  self.humanCircles = humanCircles
  self.humanCirclesIndexedByHuman = humanCirclesIndexedByHuman
end

local function buildDragHelpOverlay(self)
  local dragHelpOverlay = NominaDragHelpOverlay.new()
  self:insertAtCenter(dragHelpOverlay, Z_DRAG_HELP_OVERLAY)
  self.dragHelpOverlay = dragHelpOverlay
end


local function buildNameTagHolder(self)
  local dragHelpOverlay = self.dragHelpOverlay

  local nameTagHolder = NominaNameTagHolder.new(dragHelpOverlay)

  local nameTagHolderY = NAME_TAG_HOLDER_Y + Screen.safeBottom
  nameTagHolder:setLoc(CENTER_X, nameTagHolderY)
  self:insert(nameTagHolder, Z_NAME_TAG_HOLDER)
  self.nameTagHolder = nameTagHolder
end

local function buildCentrifuge(self)
  local centrifuge = NominaCentrifuge.new()
  centrifuge:setLoc(CENTER_X, INTRODUCTION_Y)
  self:insert(centrifuge, Z_CENTRIFUGE)
  self.centrifuge = centrifuge
end

local function buildRedFlash(self)
  local redFlash = NominaRedFlash.new()
  self:insertAtCenter(redFlash, Z_RED_FLASH)
  self.redFlash = redFlash
end

-- Constructor ------------------------

function NominaStage.new(humans, conceptId)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, NominaStage)

  buildCountdown(self)
  buildHumanCircles(self, humans)
  buildCentrifuge(self)
  buildDragHelpOverlay(self)
  buildNameTagHolder(self)
  buildRedFlash(self)

  self.dragAndDropHandler = DragAndDropHandler.new()
  self.containers = {}
  self.conceptIdentifier = conceptId
  self.isMatching = false
  self.timeExpired = false
  self.wasResponseCorrect = false
  self.canResumeCountdown = false
  self.chosenHumanCircle = nil
  self.chosenContainer = nil
  self.mnemonicPanel = nil

  return self
end
