--- NominaValidator
-- @author Created by Aaron Barrett on July 16, 2014.
-- ConceptValidator for Nomina.

class("NominaValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "sound_filename",
  "humans"
}

local MNEMONIC_FIELD = "mnemonic"

local HUMAN_FIELDS = {
  "art_link",
  "nm",
  "nm_app",
  "fct_lg_1",
  "fct_sh_1",
  "fct_app_1",
  "fct_lg_2",
  "fct_sh_2",
  "fct_app_2",
  "dist_1",
  "dist_2",
  "dist_3",
  MNEMONIC_FIELD
}

local NUMBER_OF_HUMANS = 4

local PREFIX_MALE = "nomina_boy"
local PREFIX_FEMALE = "nomina_girl"

local MNEMONIC_SINGLE_TAGS = nil
local MNEMONIC_PAIRED_TAGS = nil
local MNEMONIC_TERTIARY_TAGS = nil
local MNEMONIC_OPTIONAL_PAIRED_TAGS = { "i" } -- use optional tags validator so we can catch multiple instances

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateMnemonicFieldEndsInPunctuation(self, mnemonic)
  self:testForEndingPunctuation(mnemonic)
end

local function validateHumanFields(self, human)
  self:assertHasKeys(self, human, HUMAN_FIELDS)
  validateMnemonicFieldEndsInPunctuation(self, human[MNEMONIC_FIELD])
  self:validateTags(human[MNEMONIC_FIELD], MNEMONIC_SINGLE_TAGS, MNEMONIC_PAIRED_TAGS, MNEMONIC_TERTIARY_TAGS, MNEMONIC_OPTIONAL_PAIRED_TAGS)
end

local function validateHumanNameNotAlsoListedInDistractors(self, human)
  local name = human.nm

  local distractorCount = 3
  local distractorKeyBase = "dist_"
  for distractorIndex = 1, distractorCount do
    local distractorKey = distractorKeyBase .. distractorIndex
    local distractor = human[distractorKey]

    self:expectNotEqual(distractor, name, "Distractor names must be different from the correct name.")
  end

end

local function validateImageGenders(self, humans)
  local boysFound = 0
  local girlsFound = 0

  for _, human in ipairs(humans) do
    local imageName = human.art_link

    if (StringUtils.startsWith(imageName, PREFIX_MALE)) then
      boysFound = boysFound + 1
    elseif (StringUtils.startsWith(imageName, PREFIX_FEMALE)) then
      girlsFound = girlsFound + 1
    end
  end

  local hasFourBoys = boysFound == 4
  local hasFourGirls = girlsFound == 4
  local hasTwoOfEach = (boysFound == 2) and (girlsFound == 2)
  local hasCorrectDistribution = hasFourBoys or hasFourGirls or hasTwoOfEach

  local errorMessage = string.format("There should be 4 boys, 4 girls, or 2 of each. Found %s boys, %s girls.", boysFound, girlsFound)
  self:expectTrue(hasCorrectDistribution, errorMessage)
end

local function validateHumanTimes(self, human)
  local nameTime = tonumber(human.nm_app)
  local firstFactTime = tonumber(human.fct_app_1)
  local secondFactTime = tonumber(human.fct_app_2)

  self:expectType(nameTime, "number", "nm_app")
  self:expectType(firstFactTime, "number", "fct_app_1")
  self:expectType(secondFactTime, "number", "fct_app_2")

  self:expectLessThan(nameTime, firstFactTime, "Name should appear before first fact")
  self:expectLessThan(firstFactTime, secondFactTime, "First fact should appear before second fact")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function NominaValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local humans = concept.humans

  self:expectEqual(#humans, NUMBER_OF_HUMANS, "Invalid number of humans")

  for _, human in ipairs(humans) do
    validateHumanFields(self, human)
    validateHumanNameNotAlsoListedInDistractors(self, human)
    validateHumanTimes(self, human)
  end

  validateImageGenders(self, humans)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function NominaValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, NominaValidator)

  return self
end
