--- ObliviateButton
-- @author Created by amorant on November 11, 2014.

class("ObliviateButton")

require("ObliviateColors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local WIDTH = Screen.getWidthPart(0.5) - 15
local HEIGHT = Screen.getHeightPart(0.22)
local ICON_X = 0.5 * WIDTH
local ICON_Y = 0.6 * HEIGHT
local TEXT_WIDTH = 0.9 * WIDTH
local TEXT_X = 0.5 * WIDTH
local TEXT_Y = 0.2 * HEIGHT

-- Colors -----------------------------

local COLOR_TEXT_INITIAL = ObliviateColors.PURE_WHITE
local COLOR_TEXT_CORRECTED = ObliviateColors.PURE_BLACK
local COLOR_ICON_INITIAL = ObliviateColors.PINK
local COLOR_ICON_CORRECTED = ObliviateColors.TRANSPARENT_BLACK

-- Text -------------------------------

local FONT = Font.new(FontFaces.DIN_OT_BOLD, 14)

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_FOREGROUND = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createCard(filename)
  local card = Sprite.newWithName(filename)
  card:setEffectiveDims(WIDTH, HEIGHT)
  return card
end

local function createCardAnimation(self, cardFilename)
  local label = self.label
  local changeLabelColor = AHColorTo.new(label, 0.1, nil, COLOR_TEXT_CORRECTED)

  local icon = self.icon
  local changeIconColor = AHColorTo.new(icon, 0.1, nil, COLOR_ICON_CORRECTED)

  local newCard = createCard(cardFilename)
  self:insertAtCenter(newCard, Z_BACKGROUND)
  newCard:setOpacity(0)
  local showCard = AHOpacityTo.new(newCard, 0.1, nil, 1)

  return AHSpawn.new(showCard, changeLabelColor, changeIconColor)
end

local function createFlyAnimation(self)
  local DOWN_DISPLACEMENT = 25
  local moveDown = AHLocationYBy.new(self, 0.2, AHEaseSineOut.new(), -DOWN_DISPLACEMENT)
  local moveUp = AHLocationYBy.new(self, 0.2, AHEaseSineIn.new(), Screen.height + DOWN_DISPLACEMENT)
  return AHSequence.new(moveDown, moveUp)
end

local function createShrinkAnimation(self)
  local shrink = AHScaleTo.new(self, 0.2, AHEaseSineIn.new(), 0)
  return shrink
end

local function createShakeAnimation(self)
  local duration = 0.4
  local stepDuration = duration / 7
  local amplitude = self:getWidthPart(0.1)
  local shake = AHSequence.new(
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), -amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), 2 * amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), - 1.5 * amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), 1 * amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), - 0.75 * amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), 0.5 * amplitude),
    AHLocationXBy.new(self, stepDuration, AHEaseSimpleInOut.new(2.0), - 0.25 * amplitude)
  )
  return shake
end

local function createShakeAndShrinkAnimation(self)
  local shake = createShakeAnimation(self)
  local shrink = createShrinkAnimation(self)
  return AHSequence.new(shake, shrink)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ObliviateButton:getAnswerText()
  return self.text
end

function ObliviateButton:wasTapped()
  return self.tapped
end

function ObliviateButton:animateCorrect()
  local cardAnimation = createCardAnimation(self, "card_correct.png")
  local flyAnimation = createFlyAnimation(self)
  local playSound = AHSequence.new(
    function() Sound.playSoundEffect("correct_answer_tap", 0.2) end,
    AHDelayed.new(0.2, function() Sound.playSoundEffect("correct_answer") end)
  )
  local animation = AHSpawn.new(cardAnimation, flyAnimation, playSound)

  self:setTouchEnabled(false)

  return animation:start()
end

function ObliviateButton:animateWrong()
  local cardAnimation = createCardAnimation(self, "card_wrong.png")
  local shrinkAnimation = createShakeAndShrinkAnimation(self)
  local playSound = function() Sound.playSoundEffect("wrong_answer") end
  local animation = AHSpawn.new(cardAnimation, shrinkAnimation, playSound)

  self:setTouchEnabled(false)

  return animation:start()
end

function ObliviateButton:createRevealCorrectAnimation()
  local animation = createFlyAnimation(self)
  return animation
end

function ObliviateButton:createRevealWrongAnimation()
  local animation = createShrinkAnimation(self)
  return animation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createElements(self, concept, answer)
  local text = answer.text:upper()
  local iconFile = answer.icon

  local icon = AssetHelper.createPhotoSpriteForConcept(concept, iconFile)
  icon:setLoc(ICON_X, ICON_Y)
  icon:setColor(unpack(COLOR_ICON_INITIAL))
  self:insert(icon, Z_FOREGROUND)
  self.icon = icon

  local label = TextArea.new(text, FONT, TEXT_WIDTH, Alignment.CENTER)
  label:setLoc(TEXT_X, TEXT_Y)
  label:setColor(unpack(COLOR_TEXT_INITIAL))
  self:insert(label, Z_FOREGROUND)
  self.label = label

  local card = createCard("card_default.png")
  self:insertAtCenter(card, Z_BACKGROUND)
  self.card = card
end

-- Constructor ------------------------

function ObliviateButton.new(concept, answer, answerListener)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, ObliviateButton)

  createElements(self, concept, answer)
  Button.makeButton(self, function(button)
    self.tapped = true
    answerListener(button)
  end)

  self.text = answer.text

  return self
end
