--- ObliviateGame
-- @author Created by amorant on November 11, 2014.

class("ObliviateGame")
require("bonuses.ResponseSpeedTracker")
require("ObliviateAudioPanel")
require("ObliviateColors")
require("ObliviateConfiguration")
require("ObliviateRoundGenerator")
require("ObliviateTimer")
require("ObliviateOutroPanel")
require("ObliviateQuestionPanel")
require("StandardUIBuilder")
require("ObliviateTipPanel")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Colors -----------------------------

local UI_COMPONENTS_COLOR = ObliviateColors.PINK

-- Z-Indices --------------------------

local Z_TIMER = 10
local Z_QUESTION_PANEL = 20
local Z_AUDIO_PANEL = 30
local Z_OUTRO_PANEL = 40

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Wait lock --------------------------

local function waitForSetCompleteOrTimeout(self)
  local waitLock = AHDelay.new(math.huge)
  self.waitLock = waitLock
  wait(waitLock:start())
end

local function unlockWait(self)
  if self.waitLock then
    self.waitLock:stop()
  end
end

-- Game end condition -----------------

local function isGameLoss(self)
  return self.livesCounter:getLives() <= 0
end

local function isOneLifeRemaining(self)
  return self.livesCounter:getLives() == 1
end

-- Answer handling --------------------

local function handleCorrectAnswer(self, button)
  local concept = self.currentConcept
  local remainingCorrectAnswers = self.remainingCorrectAnswers
  local questionPanel = self.questionPanel
  local speedBonusTracker = self.speedBonusTracker
  local scoreLabel = self.scoreLabel

  local scoreDelta = ObliviateConfiguration.POINTS_PER_CORRECT_RESPONSE
  GameSession.increaseScore(scoreDelta)

  GameSession.incrementCorrectAnswerCount()

  local answer = button:getAnswerText()
  TableUtils.removeValueFromArray(remainingCorrectAnswers, answer)

  GameSession.contentTracker:notifyCorrectAnswer(concept.identifier, nil, answer)

  local isSetComplete = TableUtils.isEmpty(remainingCorrectAnswers)

  if isSetComplete then
    questionPanel:disable()
  end

  wait(questionPanel:animateCorrect(button))

  if isSetComplete then
    speedBonusTracker:notifyCorrectResponse()
    unlockWait(self)
  end

  scoreLabel:setScore(GameSession.getScore())
end

local function handleWrongAnswer(self, button)
  local questionPanel = self.questionPanel
  local livesCounter = self.livesCounter
  local concept = self.currentConcept
  local answer = button:getAnswerText()

  GameSession.incrementIncorrectAnswerCount()

  GameSession.contentTracker:notifyIncorrectAnswer(concept.identifier, nil, answer)

  if isOneLifeRemaining(self) then
    questionPanel:disable()
  end

  wait(button:animateWrong())
  livesCounter:decrement()

  if isGameLoss(self) then
    unlockWait(self)
  end
end

local function handleAnswer(self, button)
  local answer = button:getAnswerText()
  local remainingCorrectAnswers = self.remainingCorrectAnswers
  local isCorrect = TableUtils.contains(remainingCorrectAnswers, answer)

  if isCorrect then
    handleCorrectAnswer(self, button)
  else
    handleWrongAnswer(self, button)
  end
end

local function handleTimeout(self)
  local questionPanel = self.questionPanel
  local remainingCorrectAnswers = self.remainingCorrectAnswers

  local volume = 1
  local function revealCorrectCallback()
    self.livesCounter:decrement()

    Sound.playSoundEffect("wrong_answer", volume)
    volume = volume * 0.5
  end

  wait(questionPanel:revealAnswers(remainingCorrectAnswers, revealCorrectCallback))

  local roundGenerator = self.roundGenerator
  roundGenerator:notifyFailedRound()

  GameSession.contentTracker:notifyTimeout(self.currentConcept.identifier)

  unlockWait(self)
end

-- Listeners --------------------------

local function getAnswerListener(self)
  return function(button)
    runInCoroutine(function()
      handleAnswer(self, button)
    end)
  end
end

local function getTimeoutListener(self)
  return function()
    handleTimeout(self)
    unlockWait(self)
  end
end

-- One-time panels --------------------

local function showTipPanel(self)
  local tipPanel = self.tipPanel

  local tipTapToContinue = TapToContinue.new(tipPanel)
  wait(tipPanel:show())
  tipTapToContinue:waitForTrigger()
  wait(tipPanel:hide())
end

local function showOutro(self)
  local outroPanel = self.outroPanel

  self:insertAtCenter(outroPanel, Z_OUTRO_PANEL)

  local uiOverlay = self.uiOverlay
  local colorAction = AHSpawn.new()
  for i, widget in ipairs(uiOverlay.widgets) do
    local widgetColorAction = AHColorTo.new(widget, 0.2, nil, {1.0, 1.0, 1.0})
    colorAction:addAction(widgetColorAction)
  end
  colorAction:start()

  wait(outroPanel:show())
end

-- Game end ---------------------------

local function endGame(self)
  if isGameLoss(self) then
    wait(self:showDefaultGameLossOutro())
    GameSession.notifyGameLoss()
  else
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), ObliviateConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    self:showDefaultGameWinOutro()
    showOutro(self)

    GameSession.notifyGameWin()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ObliviateGame:gameFlow()
  local audioPanel = self.audioPanel
  local questionPanel = self.questionPanel
  local roundGenerator = self.roundGenerator
  local timer = self.timer
  local speedBonusTracker = self.speedBonusTracker

  showTipPanel(self)

  for i = 1, ObliviateConfiguration.ROUNDS_COUNT do
    local concept = roundGenerator:next()
    GameSession.contentTracker:notifyPresentedConcept(concept.identifier)

    self.currentConcept = concept
    self.remainingCorrectAnswers = TableUtils.map(concept.correctAnswers, function(answer) return answer.text end)
    local questionTimerDuration = concept.questionTimerDuration

    audioPanel:playAudioAndWait(concept)

    questionPanel:showQuestion(concept)
    timer:setDuration(questionTimerDuration)
    timer:start()
    speedBonusTracker:notifyStart()

    waitForSetCompleteOrTimeout(self)
    timer:stop()

    questionPanel:hideQuestion()

    GameSession.reportSeenConcept(concept.identifier)
    if isGameLoss(self) then
      break
    end


    wait(timer:refill())
  end

  endGame(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local rankTwoScore = ObliviateConfiguration.getRankTwoScore()
  local rankThreeScore = ObliviateConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(ObliviateConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

local function createGameElements(self)
  local timer = ObliviateTimer.new(getTimeoutListener(self))
  self:insertAtCenter(timer, Z_TIMER)
  self.timer = timer

  local questionPanel = ObliviateQuestionPanel.new(getAnswerListener(self), getTimeoutListener(self), QUESTION_TIMER_DURATION)
  self:insertAtCenter(questionPanel, Z_QUESTION_PANEL)
  self.questionPanel = questionPanel

  local audioPanel = ObliviateAudioPanel.new()
  self:insertAtCenter(audioPanel, Z_AUDIO_PANEL)
  self.audioPanel = audioPanel

  local tipPanel = ObliviateTipPanel.new()
  self.tipPanel = tipPanel

  local outroPanel = ObliviateOutroPanel.new()
  self.outroPanel = outroPanel

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 ObliviateConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end
  local bonusWindowDuration = ObliviateConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS
  local decayStart = ObliviateConfiguration.SPEED_BONUS_DECAY_START
  local speedBonusTracker = ResponseSpeedTracker.new(bonusWindowDuration, quickResponseWrapper, decayStart)
  self.speedBonusTracker = speedBonusTracker
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  local livesCount = ObliviateConfiguration.INITIAL_LIVES_COUNT
  self.livesCounter = builder:addLivesDisplay(livesCount)
end

-- Constructor ------------------------

function ObliviateGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ObliviateGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.currentConcept = nil
  self.remainingCorrectAnswers = nil
  self.roundGenerator = ObliviateRoundGenerator.new()

  return self
end
