--- ObliviateQuestionPanel
-- @author Created by amorant on November 11, 2014.

class("ObliviateQuestionPanel")
require("ObliviateGlow")
require("ObliviateButton")
require("ObliviateColors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local BUTTONS_SPACING = 10
local ANSWERS_PER_ROW = 2

local LAYOUT_FOR_FOUR = {
  QUESTION_Y = Screen.getHeightPart(0.77),
  BUTTONS_Y = Screen.getHeightPart(0.38)
}
local LAYOUT_FOR_SIX = {
  QUESTION_Y = Screen.getHeightPart(0.85),
  BUTTONS_Y = Screen.getHeightPart(0.37)
}
local LAYOUTS_FOR_BUTTON_COUNTS = {
  [4] = LAYOUT_FOR_FOUR,
  [5] = LAYOUT_FOR_SIX,
  [6] = LAYOUT_FOR_SIX,
}

-- Colors -----------------------------

local QUESTION_COLOR = ObliviateColors.LIGHT_PURPLE

-- Z-Indices --------------------------

local Z_QUESTION = 10
local Z_GLOW = 20
local Z_ANSWERS = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Button positioning and animations --

local function applyLayout(self)
  local buttonCount = #self.buttons
  local layout = LAYOUTS_FOR_BUTTON_COUNTS[buttonCount]
  assert(layout, "No layout found for button count " .. buttonCount)
  local questionLabel = self.questionLabel
  local buttonsColumn = self.buttonsColumn
  questionLabel:setY(layout.QUESTION_Y)
  buttonsColumn:setY(layout.BUTTONS_Y)
end

local function createAndInsertButtons(self, concept)
  local correctAnswers = concept.correctAnswers
  local wrongAnswers = concept.wrongAnswers
  local allAnswers = TableUtils.createConcatenatedArray(correctAnswers, wrongAnswers)
  TableUtils.shuffleArray(allAnswers)

  local buttonsColumn = self.buttonsColumn
  local answerListener = self.answerListener
  local currentRow = Row.new(BUTTONS_SPACING)
  local buttons = {}

  for i, answer in ipairs(allAnswers) do
    local button = ObliviateButton.new(concept, answer, answerListener)
    currentRow:insert(button)
    table.insert(buttons, button)

    if i % ANSWERS_PER_ROW == 0 then
      currentRow:alignChildren()
      buttonsColumn:insert(currentRow)
      currentRow = Row.new(BUTTONS_SPACING)
    end
  end

  if #allAnswers % ANSWERS_PER_ROW ~= 0 then
    currentRow:alignChildren()
    buttonsColumn:insert(currentRow)
  end

  buttonsColumn:alignChildren()

  return buttons
end

local function getOffscreenX(buttonIndex, button)
  local buttonWidth = button:getWidth()
  local moveDistance = Screen.width/2 + buttonWidth/2

  if buttonIndex % 2 == 1 then
    return button:getX() - moveDistance
  else
    return button:getX() + moveDistance
  end
end

local function animateButtonsIn(self)
  local buttons = self.buttons
  local buttonsAnimation = AHStaggeredSpawn.new(0.03)

  for i, button in ipairs(buttons) do
    local finalX = button:getX()
    local offscreenX = getOffscreenX(i, button)
    button:setX(offscreenX)
    local flyIn = AHLocationXTo.new(button, 0.4, AHEaseSineOut.new(), finalX)
    buttonsAnimation:addAction(flyIn)
  end

  local playSound = AHDelayed.new(0.1, function() Sound.playSoundEffect("answers_fly_in") end)
  local animation = AHSpawn.new(buttonsAnimation, playSound)

  return animation:start()
end

local function animateButtonsOut(self)
  local buttons = self.buttons
  local buttonsAnimation = AHStaggeredSpawn.new(0.03)
  local buttonsOnScreen = TableUtils.getFilteredTable(buttons, function(button)
    return not button:wasTapped()
  end)

  local noButtonsLeft = TableUtils.getSize(buttonsOnScreen) == 0
  if noButtonsLeft then
    return AHDoNothing.new()
  end

  for i, button in pairs(buttonsOnScreen) do
    local finalX = getOffscreenX(i, button)
    local flyOut = AHLocationXTo.new(button, 0.4, AHEaseSineOut.new(), finalX)
    buttonsAnimation:addAction(flyOut)
  end

  local playSound = function() Sound.playSoundEffect("answers_fly_in") end
  local animation = AHSpawn.new(buttonsAnimation, playSound)

  return animation:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ObliviateQuestionPanel:showQuestion(concept)
  local questionLabel = self.questionLabel
  questionLabel:setText(concept.question:upper())
  local showQuestion = AHOpacityTo.new(questionLabel, 0.25, nil, 1)

  -- We do this at the beginning of the new round instead of at the end because some button
  -- animations linger and overlap with the hideQuestion animation. Instead of waiting for them
  -- before removing the button rows, we remove them before presenting the new round.
  local buttonsColumn = self.buttonsColumn
  buttonsColumn:removeAll()

  local buttons = createAndInsertButtons(self, concept)
  self.buttons = buttons
  applyLayout(self)

  showQuestion:start()
  wait(animateButtonsIn(self))
end

function ObliviateQuestionPanel:hideQuestion()
  local questionLabel = self.questionLabel
  local hideQuestion = AHOpacityTo.new(questionLabel, 0.25, nil, 0)

  hideQuestion:start()
  wait(animateButtonsOut(self))
end

function ObliviateQuestionPanel:animateCorrect(button)
  local blueGlow = self.blueGlow
  return runInCoroutine(function()
    blueGlow:trackButton(button)
    wait(button:animateCorrect())
  end)
end

function ObliviateQuestionPanel:revealAnswers(remainingCorrectAnswers, revealCorrectCallback)
  local buttons = self.buttons
  local revealCorrectAnswers = AHStaggeredSpawn.new(0.15)
  local revealWrongAnswers = AHStaggeredSpawn.new(0.1)
  local redGlow = self.redGlow

  for _, button in ipairs(buttons) do
    local buttonAnswer = button:getAnswerText()
    local buttonIsCorrect = TableUtils.contains(remainingCorrectAnswers, buttonAnswer)
    if buttonIsCorrect then
      local revealAction = AHSequence.new(button:createRevealCorrectAnimation(), revealCorrectCallback)
      revealCorrectAnswers:addAction(revealAction)
      redGlow:trackButton(button)
    else
      local revealAction = button:createRevealWrongAnimation()
      revealWrongAnswers:addAction(revealAction)
    end
  end

  local questionLabel = self.questionLabel
  local dimOutQuestion = AHOpacityTo.new(questionLabel, 0.3, nil, 0.2)

  local revealAnimation = AHSequence.new(
    dimOutQuestion,
    AHDelay.new(0.2),
    revealWrongAnswers,
    AHDelay.new(0.7),
    revealCorrectAnswers
  )

  self:disable()

  return revealAnimation:start()
end

function ObliviateQuestionPanel:disable()
  local buttons = self.buttons
  for _, button in ipairs(buttons) do
    button:setTouchEnabled(false)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGlow(self, filename, particlesEnabled)
  local glow = ObliviateGlow.new(filename, particlesEnabled)
  self:insertAtCenter(glow, Z_GLOW)
  return glow
end

local function createElements(self)
  local questionLabel = DynamicTextArea.newSimple("", self:getWidthPart(0.8), Font.new(FontFaces.DIN_OT_MEDIUM, 20), Alignment.CENTER)
  questionLabel:setColor(unpack(QUESTION_COLOR))
  questionLabel:setX(self:getWidthPart(0.5))
  self:insert(questionLabel, Z_QUESTION)
  self.questionLabel = questionLabel

  local buttonsColumn = Column.new(BUTTONS_SPACING)
  buttonsColumn:setX(self:getWidthPart(0.5))
  self:insert(buttonsColumn, Z_ANSWERS)
  self.buttonsColumn = buttonsColumn

  local blueGlow = createGlow(self, "blue_glow.png", true)
  self.blueGlow = blueGlow

  local redGlow = createGlow(self, "red_glow.png", false)
  self.redGlow = redGlow
end

-- Constructor ------------------------

function ObliviateQuestionPanel.new(answerListener)
  local self = Group.newFullScreen()
  Mixin.mix(self, ObliviateQuestionPanel)

  createElements(self)
  self.answerListener = answerListener

  return self
end
