--- ObliviateRoundGenerator
-- @author Created by antonmorant on November 12, 2014.

class("ObliviateRoundGenerator")
require("ObliviateConfiguration")
require("data.generators.UtilityGenerators")
require("data.generators.ValueGenerators")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONCEPTS_TO_PULL_COUNT = 8 -- must match games.json bucket minimums

local CORRECT_ANSWERS_TO_SHOW_PER_GAME = ObliviateConfiguration.CORRECT_ANSWERS_TO_SHOW_PER_GAME
local TOTAL_ANSWERS_TO_SHOW_PER_ROUND = ObliviateConfiguration.TOTAL_ANSWERS_TO_SHOW_PER_ROUND
local MINIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND = ObliviateConfiguration.MINIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND
local MAXIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND = ObliviateConfiguration.MAXIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND
local ROUNDS_COUNT = ObliviateConfiguration.ROUNDS_COUNT

local MINIMUM_QUESTION_TIMER_DURATION = 15
local MAXIMUM_QUESTION_TIMER_DURATION = 17.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConceptPool()
  local conceptGenerator = GameSession.conceptGenerator
  local difficultyKeyFrames = ObliviateConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  local allConcepts = conceptGenerator:nextAsArray(CONCEPTS_TO_PULL_COUNT)
  local selectedConcepts = {}
  local discardedConcepts = {}
  local seenCategories = {}

  for _, concept in ipairs(allConcepts) do
    local category = concept.category
    if seenCategories[category] == nil then
      table.insert(selectedConcepts, concept)
      seenCategories[category] = true
    else
      table.insert(discardedConcepts, concept)
    end

    if #selectedConcepts >= ROUNDS_COUNT then
      break
    end
  end

  while #selectedConcepts < ROUNDS_COUNT and #discardedConcepts > 0 do
    table.insert(selectedConcepts, table.remove(discardedConcepts))
  end

  assert(#selectedConcepts == ROUNDS_COUNT, string.format("Unexpected final concept pool size. Pulled %s, selected %s, discarded %s",
                                                          #allConcepts, #selectedConcepts, #discardedConcepts))

  return selectedConcepts
end

local function getQuestionTimerDuration(answersCount)
  local minAnswersCount = TOTAL_ANSWERS_TO_SHOW_PER_ROUND[1]
  local maxAnswersCount = TOTAL_ANSWERS_TO_SHOW_PER_ROUND[ROUNDS_COUNT]
  local questionTimerDuration = MathUtils.interpolateLinear(answersCount,
                                                            minAnswersCount,
                                                            maxAnswersCount,
                                                            MINIMUM_QUESTION_TIMER_DURATION,
                                                            MAXIMUM_QUESTION_TIMER_DURATION)
  return questionTimerDuration
end

local function getTransformationFunction(self)
  return function(rawConcept)
    local currentRoundIndex = self.currentRoundIndex
    assert(currentRoundIndex <= ROUNDS_COUNT, "The rounds generator did not expect so many rounds: " .. currentRoundIndex)

    local correctAnswersPerRound = self.correctAnswersPerRound
    local numCorrect = correctAnswersPerRound[currentRoundIndex]

    local failedRoundsCount = self.failedRoundsCount
    local totalAnswersPerRound = TOTAL_ANSWERS_TO_SHOW_PER_ROUND[currentRoundIndex - failedRoundsCount]
    local numWrong = totalAnswersPerRound - numCorrect

    local answers = rawConcept.answers
    local correctAnswers = TableUtils.getFilteredArray(answers, function(answer) return answer.type == "correct" end)
    local wrongAnswers = TableUtils.getFilteredArray(answers, function(answer) return answer.type == "wrong" end)
    local selectedCorrectAnswers = TableUtils.createRandomSampleOfArray(correctAnswers, numCorrect)
    local selectedWrongAnswers = TableUtils.createRandomSampleOfArray(wrongAnswers, numWrong)

    local function makeAnswer(rawAnswer)
      return {
        text = rawAnswer.text,
        icon = rawAnswer.icon_link
      }
    end

    selectedCorrectAnswers = TableUtils.map(selectedCorrectAnswers, makeAnswer)
    selectedWrongAnswers = TableUtils.map(selectedWrongAnswers, makeAnswer)

    local concept = {
      identifier = rawConcept.identifier,
      clip = rawConcept.clip,
      question = rawConcept.question,
      context = rawConcept.context,
      correctAnswers = selectedCorrectAnswers,
      wrongAnswers = selectedWrongAnswers,
      questionTimerDuration = getQuestionTimerDuration(totalAnswersPerRound),
      assets_path = rawConcept.assets_path
    }

    self.currentRoundIndex = currentRoundIndex + 1

    return concept
  end
end

local function calculateCorrectAnswersPerRound(self)
  -- Start with random values for correct answers for each round (within established ranges)
  local correctAnswersPerRound = {}
  for i = 1, ROUNDS_COUNT do
    local minimumCorrectAnswers = MINIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND[i]
    local maximumCorrectAnswers = MAXIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND[i]
    local correctAnswers = math.random(minimumCorrectAnswers, maximumCorrectAnswers)
    correctAnswersPerRound[i] = correctAnswers
  end

  -- Now, adjust so that, in total, we show the right number of correct answers per game
  local totalCorrectAnswersPerGame = TableUtils.sum(correctAnswersPerRound)
  local roundIndex = math.random(ROUNDS_COUNT)

  while totalCorrectAnswersPerGame > CORRECT_ANSWERS_TO_SHOW_PER_GAME do
    local correctAnswers = correctAnswersPerRound[roundIndex]
    if correctAnswers > MINIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND[roundIndex] then
      correctAnswersPerRound[roundIndex] = correctAnswers - 1
      totalCorrectAnswersPerGame = TableUtils.sum(correctAnswersPerRound)
    end
    roundIndex = roundIndex % ROUNDS_COUNT + 1
  end

  while totalCorrectAnswersPerGame < CORRECT_ANSWERS_TO_SHOW_PER_GAME do
    local roundIndex = math.random(ROUNDS_COUNT)
    local correctAnswers = correctAnswersPerRound[roundIndex]
    if correctAnswers < MAXIMUM_CORRECT_ANSWERS_TO_SHOW_PER_ROUND[roundIndex] then
      correctAnswersPerRound[roundIndex] = correctAnswers + 1
      totalCorrectAnswersPerGame = TableUtils.sum(correctAnswersPerRound)
    end
    roundIndex = roundIndex % ROUNDS_COUNT + 1
  end

  self.correctAnswersPerRound = correctAnswersPerRound
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ObliviateRoundGenerator:notifyFailedRound()
  local failedRoundsCount = self.failedRoundsCount
  local roundsGeneratedSoFar = self.currentRoundIndex - 1
  failedRoundsCount = failedRoundsCount + 1
  assert(failedRoundsCount <= roundsGeneratedSoFar, "Cannot fail more rounds than have been generated")
  self.failedRoundsCount = failedRoundsCount
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ObliviateRoundGenerator.new()
  local conceptPool = getConceptPool()
  local selectedConceptsGenerator = SequentialValueGenerator.new(conceptPool)

  local self = TransformationGenerator.new(selectedConceptsGenerator)
  Mixin.mix(self, ObliviateRoundGenerator)

  self:setTransformFunction(getTransformationFunction(self))

  calculateCorrectAnswersPerRound(self)
  self.currentRoundIndex = 1
  self.failedRoundsCount = 0

  return self
end
