--- OnboardioButton
-- @author Created by Aaron Barrett on October 01, 2014.

class("OnboardioButton")
require("OnboardioCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local HORIZONTAL_MARGIN = 16
local NORMAL_WIDTH = Screen.fullWidth - 2 * HORIZONTAL_MARGIN
local NORMAL_HEIGHT = Screen.getHeightPart(0.125)

local SKINNY_HEIGHT = Screen.getHeightPart(0.10)

local SMALL_HEIGHT = Screen.getHeightPart(0.08)
local SMALL_WIDTH = math.min(NORMAL_WIDTH, SMALL_HEIGHT * 6.125)

local INVERTED_PRESSABLE_MARGIN = Screen.getHeightPart(0.035)
local INVERTED_FITTED_SIZE_MARGIN = Screen.getHeightPart(0.05)

local CORNER_RADIUS = NORMAL_HEIGHT * 0.05

local BACKGROUND_COLOR_NORMAL = OnboardioCommon.COLOR_BLUE
local BACKGROUND_COLOR_NORMAL_DOWN_STATE = OnboardioCommon.COLOR_BLUE_DOWNSTATE
local BACKGROUND_COLOR_PRETEST = OnboardioCommon.PRETEST_BUTTON_BACKGROUND
local LABEL_COLOR_NORMAL = OnboardioCommon.COLOR_WHITE_TEXT
local LABEL_COLOR_INVERTED_DOWN_STATE = BACKGROUND_COLOR_NORMAL_DOWN_STATE
local LABEL_COLOR_INVERTED_OUTRO = OnboardioCommon.COLOR_LABEL_INVERTED
local DEFAULT_INVERTED_LABEL_TEXT_COLOR = OnboardioCommon.COLOR_BLUE

local FONT_NORMAL = OnboardioCommon.FONT_NORMAL
local FONT_INVERTED = FONT_NORMAL


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function playSelected(self)
  local backgroundColor = self.unselectedBackgroundColor
  local textColor = self.unselectedTextColor
  if (self.wasPressed) then
    backgroundColor = self.selectedBackgroundColor
    textColor = self.selectedTextColor
  end

  AHSpawn.new(
    AHColorTo.new(self.background, 0.35, nil, backgroundColor),
    AHColorTo.new(self.conceptView, 0.35, nil, textColor)
  ):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioButton:setTouchEnabled(isEnabled)
  self.pressableGroup:setTouchEnabled(isEnabled)
end

function OnboardioButton:getWasPressed()
  return self.wasPressed
end

function OnboardioButton:playPressedColor()
  local background = self.background
  if (background ~= nil) then
    background:setColor(unpack(BACKGROUND_COLOR_NORMAL_DOWN_STATE))

    AHColorTo.new(background, 0.35, nil, BACKGROUND_COLOR_NORMAL):start()
  else
    local label = self.conceptView
    label:setColor(unpack(LABEL_COLOR_INVERTED_DOWN_STATE))

    AHColorTo.new(label, 0.35, nil, LABEL_COLOR_NORMAL):start()
  end
end

function OnboardioButton:getConceptView()
  return self.conceptView
end

function OnboardioButton:getBackground()
  return self.background
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLabel(self, text, labelColor, font)
  local label = Label.new(text, font)
  label:setColor(unpack(labelColor))
  label:inheritOpacity(self)
  return label
end

local function createOnDown(self, onPressedCallback)
  local function onDown()
    self.wasPressed = true
    self:playPressedColor()
    onPressedCallback()
  end
  return onDown
end

local function createOnSelected(self, onPressedCallback)
  local function onSelected()
    self.wasPressed = (not self.wasPressed)
    playSelected(self)
    onPressedCallback(self.wasPressed)
  end
  return onSelected
end

local function newNormalButton(text, onPressedCallback, width, height, font, labelColor, isHollow, backgroundColor)
  local self = Group.new(width, height)
  Mixin.mix(self, OnboardioButton)

  local backgroundColor = backgroundColor or BACKGROUND_COLOR_NORMAL

  local rectangle = nil

  if (isHollow) then
    rectangle = HollowRoundedRectangle.new(width, height, CORNER_RADIUS, backgroundColor)
  else
    rectangle = FilledRoundedRectangle.new(width, height, CORNER_RADIUS, backgroundColor)
  end

  rectangle:inheritOpacity(self)
  self:insertAtCenter(rectangle)
  self.pressableGroup = rectangle
  self.background = rectangle

  local label = createLabel(self, text, labelColor, font)
  rectangle:insertAtCenter(label)
  self.conceptView = label

  local onDown = createOnDown(self, onPressedCallback)
  Button.makeButton(rectangle, onDown)

  return self
end

local function newSelectableButton(text, onPressedCallback, width, height, font, unselectedTextColor, selectedTextColor, unselectedBackgroundColor, selectedBackgroundColor)
  local self = Group.new(width, height)
  Mixin.mix(self, OnboardioButton)

  local rectangle = HollowRoundedRectangle.new(width, height, CORNER_RADIUS, unselectedBackgroundColor)
  rectangle:inheritOpacity(self)
  self:insertAtCenter(rectangle)
  self.pressableGroup = rectangle
  self.background = rectangle

  local label = createLabel(self, text, unselectedTextColor, font)
  self.background:insertAtCenter(label)
  self.conceptView = label

  self.unselectedBackgroundColor = unselectedBackgroundColor
  self.selectedBackgroundColor = selectedBackgroundColor
  self.unselectedTextColor = unselectedTextColor
  self.selectedTextColor = selectedTextColor

  local onDown = createOnSelected(self, onPressedCallback)
  Button.makeButton(rectangle, onDown)

  return self
end

-- Constructor ------------------------


function OnboardioButton.new(text, onPressedCallback)
  local width, height = NORMAL_WIDTH, NORMAL_HEIGHT
  local font = FONT_NORMAL
  local labelColor = LABEL_COLOR_NORMAL

  local self = newNormalButton(text, onPressedCallback, width, height, font, labelColor)
  return self
end

function OnboardioButton.newSmall(text, onPressedCallback)
  local width, height = SMALL_WIDTH, SMALL_HEIGHT
  local font = FONT_NORMAL
  local labelColor = LABEL_COLOR_NORMAL

  local self = newNormalButton(text, onPressedCallback, width, height, font, labelColor)
  return self
end

function OnboardioButton.newSkinny(text, onPressedCallback, usePretestStyle)
  local width, height = NORMAL_WIDTH, SKINNY_HEIGHT
  local font = FONT_NORMAL
  local labelColor = LABEL_COLOR_NORMAL
  local backgroundColor = nil
  if usePretestStyle then
    labelColor = LABEL_COLOR_INVERTED_OUTRO
    backgroundColor = BACKGROUND_COLOR_PRETEST
  end

  local self = newNormalButton(text, onPressedCallback, width, height, font, labelColor, nil, backgroundColor)
  return self
end

function OnboardioButton.newHollow(text, onPressedCallback)
  local width, height = NORMAL_WIDTH, SKINNY_HEIGHT
  local font = FONT_NORMAL
  local labelColor = BACKGROUND_COLOR_NORMAL
  local isHollow = true

  local self = newNormalButton(text, onPressedCallback, width, height, font, labelColor, isHollow)
  return self
end

function OnboardioButton.newInverted(text, onPressedCallback, shouldFitToLabel, labelColor)
  local self = Group.new(NORMAL_WIDTH, NORMAL_HEIGHT)

  Mixin.mix(self, OnboardioButton)

  labelColor = labelColor or DEFAULT_INVERTED_LABEL_TEXT_COLOR
  local label = createLabel(self, text, labelColor, FONT_INVERTED)
  local labelWidth, labelHeight = label:getDims()
  self.conceptView = label

  if shouldFitToLabel then
    self:setWidth(labelWidth + INVERTED_FITTED_SIZE_MARGIN)
    self:setHeight(labelHeight + INVERTED_FITTED_SIZE_MARGIN)
  end

  local pressableGroup = Group.new(labelWidth + INVERTED_PRESSABLE_MARGIN, labelHeight + INVERTED_PRESSABLE_MARGIN)
  self:insertAtCenter(pressableGroup)
  self.pressableGroup = pressableGroup

  pressableGroup:insertAtCenter(label)

  local onDown = createOnDown(self, onPressedCallback)
  Button.makeButton(pressableGroup, onDown)

  self.background = nil
  self.wasPressed = false

  return self
end

function OnboardioButton.newHollowSelectable(text, onPressedCallback, unselectedTextColor, selectedTextColor, unselectedBackgroundColor, selectedBackgroundColor)
  local width, height = NORMAL_WIDTH, SMALL_HEIGHT
  local font = FONT_NORMAL

  local self = newSelectableButton(text, onPressedCallback, width, height, font, unselectedTextColor, selectedTextColor, unselectedBackgroundColor, selectedBackgroundColor)
  return self
end
