--- OnboardioGame
-- @author Created by Aaron Barrett on September 29, 2014.

class("OnboardioGame")

require("StandardUIBuilder")
require("PretestSession")
require("OnboardioCommon")
require("OnboardioPanelIntro")
require("OnboardioContent")
require("OnboardioTransitioner")
require("OnboardioPersonalizationManager")
require("OnboardioBackButtonHandler")
require("OnboardioAnalytics")
require("OnboardioPanelTitle")
require("OnboardioPanelPretest")
require("OnboardioPanelEpqIntroduction")
require("OnboardioNavigationOverlay")
require("OnboardioPanelHowDidYouHearAboutUs")
require("OnboardioPanelSocialProof")
require("OnboardioPanelMotivationQuestion")
require("OnboardioPanelMotivationReinforcement")
require("OnboardioPanelADHDQuestion")
require("OnboardioPanelADHDResponse")
require("OnboardioPanelAddAnotherAccount")
require("OnboardioPanelPersonalizationQuestions")
require("OnboardioPanelReviewProof")
require("OnboardioPanelTimeCommitment")
require("OnboardioPanelLearningPace")
require("OnboardioPanelEnglishSecondLanguage")
require("OnboardioPanelOtherBrainActivities")
require("OnboardioPanelFirstGoalQuestion")
require("OnboardioPanelFirstGoalReinforcement")
require("OnboardioPanelMoreQuestionsBreak")
require("ExperimentsConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = OnboardioCommon.COLOR_TABLE.SURFACE100

local Z_BACKGROUND = 10
local Z_PANEL = 20
local Z_OVERLAY = 30

OnboardioGame.ANSWER_BACK = "back"
OnboardioGame.ANSWER_SKIP = "skip"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setDefaultOverlayCallbacks(self)
  self.overlay:setSkipButtonCallback(function()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end)
  self.overlay:setBackButtonCallback(function()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_BACK)
    self.currentPanel:setDismissed()
  end)
end

local function calculatePretestResults(self)
  local pretestResults = {}

  local scoresBySkillGroupIdentifier = self.scoresBySkillGroupIdentifier
  for skillGroupIdentifier, scores in pairs(scoresBySkillGroupIdentifier) do
    local correctProportion = scores.correct / scores.possible
    pretestResults[skillGroupIdentifier] = correctProportion
  end

  return pretestResults
end

local function getResultsToDisplayFromTheApp(self, pretestResults)
  local resultsToDisplay = {}

  local filledProportions = PretestSession.getPerformanceIndices(pretestResults)
  local epqDisplayValues = PretestSession.getPerformanceDisplayValues(filledProportions)

  local displayNames = OnboardioContent.DISPLAY_NAMES_BY_SKILL_GROUP_IDENTIFIER
  local identifiersInOrder = OnboardioContent.SKILL_GROUP_IDENTIFIERS_IN_DISPLAY_ORDER

  for _, identifier in ipairs(identifiersInOrder) do
    local filledProportion = filledProportions[identifier]
    assert(filledProportion ~= nil, "No 'performance index' was returned by the app for skill group with identifier `" .. tostring(identifier) .. "`.")
    local epqDisplayValue = MathUtils.round(epqDisplayValues[identifier])
    assert(epqDisplayValue ~= nil, "No 'EPQ display value' was returned by the app for the skill group with identifier `" .. tostring(identifier) .. "`.")
    local displayName = displayNames[identifier]
    assert(displayName ~= nil, "No display name listed for skill group with identifier `" .. tostring(identifier) .. "`.")

    local resultsEntry = {
      filledProportion = filledProportion,
      epqDisplayValue = epqDisplayValue,
      displayName = displayName,
      skillIdentifier = identifier
    }
    table.insert(resultsToDisplay, resultsEntry)
  end
  return resultsToDisplay
end

local function submitPretestResultsToHost(pretestResults)
  PretestSession.setPretestResults(pretestResults)
end

local function closePreviousSection(self)
  local previousPanel = self.currentPanel
  self.currentPanel = nil
  OnboardioTransitioner.playTransitionOut(previousPanel)
end

local function waitToTransitionBackToNative(self)
  closePreviousSection(self)

  waitFor(0.2)

  local framesToPreventFlickerOnApp = 2
  waitForFrames(framesToPreventFlickerOnApp)
end

local function calculateAndSubmitPretestResultsIfNeeded(self)
  local needToCalculateAndSubmitPretestResults = (self.pretestResults == nil)
  if needToCalculateAndSubmitPretestResults then
    self.pretestResults = calculatePretestResults(self)
    submitPretestResultsToHost(self.pretestResults)
  end
end

local function getPreviousPanelAnswer(self)
  if (self.previousPanel ~= nil) then
    return self.previousPanel:getSelectedAnswer()
  end
  return nil
end

-- Create and return a new panel
-- @param panelClass The class of the panel to create
-- @param backButtonVisible Whether the back button should be visible (defaults to false)
-- @param skipButtonVisible Whether the skip button should be visible (defaults to false)
-- @param skipButtonCallback The callback to call when the skip button is pressed (optional)
-- @param ... Additional arguments to pass to the panel constructor
local function createPanel(self, panelClass, backButtonVisible, skipButtonVisible, skipButtonCallback, ...)
  local panel = panelClass.new(...)
  backButtonVisible = backButtonVisible or false
  skipButtonVisible = skipButtonVisible or false
  if (skipButtonCallback ~= nil) then
    self.overlay:setSkipButtonCallback(skipButtonCallback)
  end
  self.overlay:setBackButtonVisible(backButtonVisible)
  self.overlay:setSkipButtonVisible(skipButtonVisible)
  return panel
end

local function createTitlePanel(self)
  if (self.previousPanel == nil) then
    self.skipTransition = true
  end

  return createPanel(self, OnboardioPanelTitle)
end

local function createAddAnotherAccountPanel(self)
  local backButtonVisible = true
  return createPanel(self, OnboardioPanelAddAnotherAccount, backButtonVisible, nil, nil, self.overlay)
end

local function createIntroPanel(self)
  local isFirstPanel = OnboardioCommon.shouldStartAtIntroPanel()
  return createPanel(self, OnboardioPanelIntro, nil, nil, nil, isFirstPanel)
end

local function createMoreQuestionsBreakPanel(self)
  ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:reportExperimentExposed()
  return createPanel(self, OnboardioPanelMoreQuestionsBreak)
end

local function createPersonalizationQuestionsPanel(self)
  return createPanel(self, OnboardioPanelPersonalizationQuestions)
end

local function createReviewProofPanel(self)
  local skipButtonVisible = true
  return createPanel(self, OnboardioPanelReviewProof, nil, skipButtonVisible)
end

local function createTimeCommitmentPanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportTimeCommitmentScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelTimeCommitment, nil, skipButtonVisible, skipButtonCallback)
end

local function createLearningPacePanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportLearningPaceScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelLearningPace, nil, skipButtonVisible, skipButtonCallback)
end

local function createHowDidYouHearAboutUsPanel(self)
  local skipButtonCallback = function()
    OnboardioAnalytics.reportHowDidYouHearAboutUsSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end

  local skipButtonVisible = true
  return createPanel(self, OnboardioPanelHowDidYouHearAboutUs, nil, skipButtonVisible, skipButtonCallback)
end

local function createSocialProofPanel(self)
  local skipButtonCallback = function()
    OnboardioAnalytics.reportSocialProofSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end

  local skipButtonVisible = true
  return createPanel(self, OnboardioPanelSocialProof, nil, skipButtonVisible, skipButtonCallback)
end

local function createMotivationQuestionPanel(self)
  self.motivationQuestionResponse = nil
  local setMotivationResponseCallback = function(answerID)
    self.motivationQuestionResponse = answerID
  end

  local skipButtonCallback = function()
    OnboardioAnalytics.reportMotivationScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end

  if (ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:isControl()) then
    ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:reportExperimentExposed()
  end

  local skipButtonVisible = true
  return createPanel(self, OnboardioPanelMotivationQuestion, nil, skipButtonVisible, skipButtonCallback, setMotivationResponseCallback)
end

local function createMotivationReinforcementPanel(self)
  local answer = getPreviousPanelAnswer(self)
  return createPanel(self, OnboardioPanelMotivationReinforcement, nil, nil, nil, answer)
end

local function createADHDQuestionPanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportADHDScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelADHDQuestion, nil, skipButtonVisible, skipButtonCallback, self.motivationQuestionResponse)
end

local function createADHDResponsePanel(self)
  local answer = getPreviousPanelAnswer(self)
  return createPanel(self, OnboardioPanelADHDResponse, nil, nil, nil, answer)
end

local function createESLPanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportESLScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelEnglishSecondLanguage, nil, skipButtonVisible, skipButtonCallback)
end

local function createOtherBrainActivitiesPanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportOtherBrainActivitiesScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelOtherBrainActivities, nil, skipButtonVisible, skipButtonCallback)
end

local function createFirstGoalQuestionPanel(self)
  local skipButtonVisible = true
  local skipButtonCallback = function()
    OnboardioAnalytics.reportFirstGoalScreenSkipped()
    self.currentPanel:setSelectedAnswer(OnboardioGame.ANSWER_SKIP)
    self.currentPanel:setDismissed()
  end
  return createPanel(self, OnboardioPanelFirstGoalQuestion, nil, skipButtonVisible, skipButtonCallback)
end

local function createFirstGoalReinforcementPanel(self)
  local answer = getPreviousPanelAnswer(self)
  return createPanel(self, OnboardioPanelFirstGoalReinforcement, nil, nil, nil, answer)
end

local function createPretestPanel(self)
  local onboardingStartedAtPretest = OnboardioCommon.shouldStartAtPretest()
  local skipButtonVisible = true
  return createPanel(self, OnboardioPanelPretest, nil, skipButtonVisible, nil, self.overlay, self.scoresBySkillGroupIdentifier, onboardingStartedAtPretest)
end

local function createEpqIntroductionPanel(self)
  calculateAndSubmitPretestResultsIfNeeded(self)
  local useAlreadyCreatedAccountCopy = OnboardioCommon.shouldStartAtAddSkillSelection()
  return createPanel(self, OnboardioPanelEpqIntroduction, nil, nil, nil, getResultsToDisplayFromTheApp(self, self.pretestResults), useAlreadyCreatedAccountCopy)
end

local function validateCurrentPanelIndex(currentPanelIndex)
  if (currentPanelIndex < 1) then
    Log.error("Attempted to show an invalid Panel. Reverting to start.")
    return 1
  end
  return currentPanelIndex
end

local function runCurrentPanel(self, currentPanelIndex, isNavigatingBackward)
  if (not self.skipTransition) then
    local frameToImprovePerformanceOnOldDevices = 1
    waitForFrames(frameToImprovePerformanceOnOldDevices)

    if (self.currentPanel:usesHorizontalTransitionIn()) then
      self.currentPanel:setX(self:getWidthPart(1.5))
    end

    local transitionDelay = self.currentPanel:getTransitionDelayDuration()
    if (transitionDelay > 0) then
      waitFor(transitionDelay)
    end

    wait(OnboardioTransitioner.playSimpleTransitionBetweenPanels(self.previousPanel, self.currentPanel, isNavigatingBackward))
  end
  self.currentPanel:start()

  waitUntil(function() return self.currentPanel:isDismissed() end)
end

local function updateCurrentPanelIndexFromAnswer(currentPanelIndex, answer)
  if (answer == OnboardioGame.ANSWER_BACK) then
    return (currentPanelIndex - 1)
  else
    return (currentPanelIndex + 1)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioGame:gameFlow()
  local numberOfPanels = #self.orderedPanelCreationMethods

  local doGame = function()
    local currentPanelIndex = 1
    self.previousPanel = nil
    self.currentPanel = nil
    self.skipTransition = false
    local isNavigatingBackward = false
    while (currentPanelIndex <= numberOfPanels) do
      currentPanelIndex = validateCurrentPanelIndex(currentPanelIndex)

      if (self.currentPanel ~= nil) then
        self.previousPanel = self.currentPanel
      end
      local createPanelFunction = self.orderedPanelCreationMethods[currentPanelIndex]

      setDefaultOverlayCallbacks(self)
      self.currentPanel = createPanelFunction(self)

      if (self.currentPanel == nil) then
        local answer = nil
        if (self.previousPanel ~= nil) then
          answer = self.previousPanel:getSelectedAnswer()
        end
        currentPanelIndex = updateCurrentPanelIndexFromAnswer(currentPanelIndex, answer)
      else
        self:insertAtCenter(self.currentPanel, Z_PANEL)
        runCurrentPanel(self, currentPanelIndex, isNavigatingBackward)
        isNavigatingBackward = false

        local answer = self.currentPanel:getSelectedAnswer()
        currentPanelIndex = updateCurrentPanelIndexFromAnswer(currentPanelIndex, answer)
        if (answer == OnboardioGame.ANSWER_BACK) then
          isNavigatingBackward = true
        end
        self.skipTransition = false
      end
    end

    waitToTransitionBackToNative(self)
    GameSession.notifyGameWin()
  end

  self.gameAction = runInCoroutine(doGame)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createSetOfPanelInitializersToShow(self)
  local createPanelFunctions = {}

  if OnboardioCommon.shouldStartAtDefaultPosition() then
    table.insert(createPanelFunctions, createTitlePanel)
  elseif OnboardioCommon.shouldStartAtAddAnotherAccountPosition() then
    table.insert(createPanelFunctions, createAddAnotherAccountPanel)
  end

  if (not OnboardioCommon.shouldStartAtPretest()) then
    table.insert(createPanelFunctions, createIntroPanel)
    table.insert(createPanelFunctions, createSocialProofPanel)

    if (not OnboardioCommon.isStreamlinedOnboarding()) then
      if (ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:isActive()) then
        table.insert(createPanelFunctions, createMoreQuestionsBreakPanel)
      end
      table.insert(createPanelFunctions, createMotivationQuestionPanel)
      table.insert(createPanelFunctions, createMotivationReinforcementPanel)
      if (ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:isActive()) then
        table.insert(createPanelFunctions, createPersonalizationQuestionsPanel)
        table.insert(createPanelFunctions, createReviewProofPanel)
        table.insert(createPanelFunctions, createTimeCommitmentPanel)
        table.insert(createPanelFunctions, createLearningPacePanel)
      end
      table.insert(createPanelFunctions, createADHDQuestionPanel)
      table.insert(createPanelFunctions, createADHDResponsePanel)
      if (ExperimentsConfiguration.ENHANCED_ONBOARDING_EXPERIMENT:isActive()) then
        table.insert(createPanelFunctions, createESLPanel)
        table.insert(createPanelFunctions, createOtherBrainActivitiesPanel)
        table.insert(createPanelFunctions, createFirstGoalQuestionPanel)
        table.insert(createPanelFunctions, createFirstGoalReinforcementPanel)
      end
      table.insert(createPanelFunctions, createHowDidYouHearAboutUsPanel)
    end
  end

  table.insert(createPanelFunctions, createPretestPanel)
  table.insert(createPanelFunctions, createEpqIntroductionPanel)

  return createPanelFunctions
end

local function createGameElements(self)
  local width = self:getWidth()
  local height = self:getHeight()

  local background = FilledRectangle.new(width, height, BACKGROUND_COLOR)
  self:insertAtCenter(background, Z_BACKGROUND)

  self.overlay = OnboardioNavigationOverlay.new()
  self:insertAtCenter(self.overlay, Z_OVERLAY)
end

local function setThresholdsSoAppWontCrash()
  GameSession.setScoreThresholds(0, 0)
end

-- Constructor ------------------------

function OnboardioGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, OnboardioGame)

  setThresholdsSoAppWontCrash()
  createGameElements(self)

  OnboardioContent.load()

  self.scoresBySkillGroupIdentifier = {
    [OnboardioContent.SKILL_GROUP_IDENTIFIER_READING] = {correct = 0.0, possible = 0.0},
    [OnboardioContent.SKILL_GROUP_IDENTIFIER_WRITING] = {correct = 0.0, possible = 0.0},
    [OnboardioContent.SKILL_GROUP_IDENTIFIER_SPEAKING] = {correct = 0.0, possible = 0.0},
    [OnboardioContent.SKILL_GROUP_IDENTIFIER_MATH] = {correct = 0.0, possible = 0.0},
    [OnboardioContent.SKILL_GROUP_IDENTIFIER_MEMORY] = {correct = 0.0, possible = 0.0}
  }

  self.orderedPanelCreationMethods = createSetOfPanelInitializersToShow(self)

  return self
end
