--- OnboardioPanelChooseYourLanguage
-- @author Created by andrewcousins on July 13, 2022.

class("OnboardioPanelChooseYourLanguage")

require("OnboardioPanel")
require("OnboardioButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DO_NOT_SHOW_PANEL_FOR_LOCALES = {
  "en_US",
  "en_CA",
  "en_AU"
}

local BUTTON_DESCRIPTORS = {
  {
    text = Localizer.get("CHOOSE_YOUR_LANGUAGE_AMERICAN_ENGLISH"),
    locale = "en_US"
  },
  {
    text = Localizer.get("CHOOSE_YOUR_LANGUAGE_BRITISH_ENGLISH"),
    locale = "en_GB"
  }
}

local PROMPT_COLOR = OnboardioCommon.COLOR_PROMPT_TEXT
local FOOTER_COLOR = OnboardioCommon.COLOR_PROMPT_ADDITIONAL_TEXT

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateLocaleForSelection(locale)
  Localizer.setLocale(locale)
  GameSession.setCurrentLocale(locale)
  OnboardioAnalytics.reportEnglishTypeSelected(locale)

  OnboardioContent.load() -- reload content for the selected locale
end

local function createButton(self, text, locale)
  local callback = function()
    updateLocaleForSelection(locale)
    self:setDismissed()
  end
  local button = nil
  if Screen.isTabletAspectRatio() then
    button = OnboardioButton.newSmall(text, callback)
  else
    button = OnboardioButton.newSkinny(text, callback)
  end
  button:setOpacity(0)
  return button
end

local function createButtons(self)
  local buttonDescriptors = BUTTON_DESCRIPTORS
  local buttons = {}
  for _, buttonDescription in ipairs(buttonDescriptors) do
    local button = createButton(self, buttonDescription.text, buttonDescription.locale)
    table.insert(buttons, button)
  end
  return buttons
end

local function createTextArea(text, font, color)
  local label = TextArea.new(text, font, OnboardioCommon.TEXT_WIDTH, Alignment.CENTER)
  label:setColor(unpack(color))
  label:setOpacity(0)
  return label
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelChooseYourLanguage:getAppearanceDelayDuration()
  return 0.25
end

function OnboardioPanelChooseYourLanguage:playAppear()
  local animation = AHSpawn.new()
  local duration = 0.4
  local ease = AHEaseSimpleInOut.new()
  for _, label in ipairs(self.fadableLabels) do
    animation:addAction(AHOpacityTo.new(label, duration, ease, 1))
  end
  for _, button in ipairs(self.buttons) do
    animation:addAction(AHOpacityTo.new(button, duration, ease, 1))
    button:setTouchEnabled(true)
  end

  local delay = self:getAppearanceDelayDuration()
  return AHDelayed.new(delay, animation):start()
end

function OnboardioPanelChooseYourLanguage:playDisappear()
  local animation = AHSpawn.new()
  local duration = 0.3
  local ease = AHEaseSimpleInOut.new()
  for _, label in ipairs(self.fadableLabels) do
    animation:addAction(AHOpacityTo.new(label, duration, ease, 0))
  end
  for _, button in ipairs(self.buttons) do
    animation:addAction(AHOpacityTo.new(button, duration, ease, 0))
    button:setTouchEnabled(false)
  end
  return animation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local columnSpaceSize = Screen.getHeightPart(0.022)
  if Screen.isTabletAspectRatio() then
    columnSpaceSize = columnSpaceSize / 2
  end

  local column = Column.new(columnSpaceSize)
  local promptText = Localizer.get("CHOOSE_YOUR_LANGUAGE_PROMPT")
  local promptTextArea = createTextArea(promptText, Font.new(FontFaces.DIN_OT_MEDIUM, 22), PROMPT_COLOR)
  table.insert(self.fadableLabels, promptTextArea)
  column:insert(promptTextArea)
  column:insertSpacer(0)

  local buttons = createButtons(self)
  for _, button in ipairs(buttons) do
    table.insert(self.buttons, button)
    column:insert(button)
  end

  column:alignChildren()
  self:insertAtCenter(column)

  local youCanChangeThisLaterText = Localizer.get("CHOOSE_YOUR_LANGUAGE_YOU_CAN_CHANGE_THIS_LATER")
  local youCanChangeThisLaterTextArea = createTextArea(youCanChangeThisLaterText, Font.new(FontFaces.DIN_OT_MEDIUM, 15), FOOTER_COLOR)
  table.insert(self.fadableLabels, youCanChangeThisLaterTextArea)

  local youCanChangeThisLaterTextAreaY = math.max(self:getHeightPart(0.05), (Screen.safeBottom + youCanChangeThisLaterTextArea:getHeightPart(0.5)))

  youCanChangeThisLaterTextArea:setLoc(self:getWidthPart(0.5), youCanChangeThisLaterTextAreaY)
  self:insert(youCanChangeThisLaterTextArea)

  column:offsetY(youCanChangeThisLaterTextAreaY/2)
end

-- Constructor ------------------------

function OnboardioPanelChooseYourLanguage.new()
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioPanelChooseYourLanguage)

  OnboardioAnalytics.reportEnglishTypeScreenSeen()

  createLayout(self)

  return self
end

function OnboardioPanelChooseYourLanguage.shouldShowChooseYourLanguagePanel()
  assert(GameSession.shouldClarifyLocaleInOnboarding ~= nil, "shouldClarifyLocaleInOnboarding() was not accessible through GameSession")
  local shouldShow = GameSession.shouldClarifyLocaleInOnboarding()

  return shouldShow
end
