--- OnboardioPanelMotivationReinforcement
-- @author Created by andrewcousins on November 16, 2023.

class("OnboardioPanelMotivationReinforcement")

require("OnboardioPanel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GAME_SCREENS_FILE_NAME = "game_screens.png"

local MOTIVATION_ID_TO_LOCALIZATION_KEY = {
  [OnboardioPanelMotivationQuestion.BUTTON_ID.WORK_SCHOOL] = "MOTIVATION_REINFORCEMENT_WORK_SCHOOL",
  [OnboardioPanelMotivationQuestion.BUTTON_ID.COMMUNICATION] = "MOTIVATION_REINFORCEMENT_COMMUNICATION",
  [OnboardioPanelMotivationQuestion.BUTTON_ID.COGNITIVE_HEALTH] = "MOTIVATION_REINFORCEMENT_COGNITIVE_HEALTH",
  [OnboardioPanelMotivationQuestion.BUTTON_ID.FUN_GAMES] = "MOTIVATION_REINFORCEMENT_FUN_GAMES",
  [OnboardioPanelMotivationQuestion.BUTTON_ID.VOCABULARY] = "MOTIVATION_REINFORCEMENT_VOCABULARY",
  [OnboardioPanelMotivationQuestion.BUTTON_ID.OTHER] = "MOTIVATION_REINFORCEMENT_OTHER"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function motivationIdIsValid(motivationId)
  for id, _ in pairs(MOTIVATION_ID_TO_LOCALIZATION_KEY) do
    if (motivationId == id) then
      return true
    end
  end
  return false
end

local function hideFadeableElements(self)
  for _, element in ipairs(self.fadeableElements) do
    element:setOpacity(0)
  end
end

local function createAndInsertGameScreens(self)
  local sprite = Sprite.newWithName(GAME_SCREENS_FILE_NAME)
  sprite:setEffectiveDims(self:getWidth())
  sprite:setAnchorPointTop()
  self:insertAtTop(sprite)
  local maxHeightDisplayedOnScreen = self:getHeightPart(0.6)
  if (sprite:getEffectiveHeight() > maxHeightDisplayedOnScreen) then
    sprite:offsetY(sprite:getEffectiveHeight() - maxHeightDisplayedOnScreen)
  end
  table.insert(self.fadeableElements, sprite)

  self.gameScreensSprite = sprite
end

local function createAndInsertReinforcementText(self, motivationId)
  local fontSize = 20
  local fontColor = OnboardioCommon.COLOR_PROMPT_TEXT
  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_REGULAR, fontSize),
    color = fontColor
  }
  local labelAttributes = {
    b = {
      font = Font.new(FontFaces.DIN_OT_BOLD, fontSize)
    }
  }

  local reinforcementTextKey = MOTIVATION_ID_TO_LOCALIZATION_KEY[motivationId]
  local reinforcementText = Localizer.get(reinforcementTextKey)
  local textWidth = OnboardioCommon.TEXT_WIDTH

  local textArea = DynamicTextArea.new(reinforcementText, textWidth, baseAttributes, labelAttributes, Alignment.CENTER)

  self:insertAtCenter(textArea)
  local textAreaMaxY = Screen.getHeightPart(0.44)
  local minimumBuffer = Screen.getHeightPart(0.05)
  local textAreaShiftedY = self.gameScreensSprite:getYBottom() - textArea:getHeightPart(0.5) - minimumBuffer
  local textAreaY = math.min(textAreaMaxY, textAreaShiftedY)
  textArea:setY(textAreaY)

  table.insert(self.fadeableElements, textArea)
end

local function createAndInsertContinueButton(self)
  local usePretestStyle = false
  local buttonGroup, continueButton = OnboardioOutroCommon.newButtonGroup(
    Localizer.get("CONTINUE"),
    function()
      self:setDismissed()
    end,
    usePretestStyle
  )

  local buttonGroupY = math.max(self:getHeightPart(0.01), Screen.safeBottom)
  buttonGroup:setAnchorPointBottom()
  self:insertAtLocation(buttonGroup, self:getWidthPart(0.5), buttonGroupY)
  table.insert(self.fadeableElements, continueButton)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelMotivationReinforcement:getAppearanceDelayDuration()
  return 0.25
end

function OnboardioPanelMotivationReinforcement:playAppear()
  OnboardioAnalytics.reportMotivationReinforcementScreenSeen()

  local animation = AHSpawn.new()
  local duration = 0.4
  local ease = AHEaseSimpleInOut.new()
  for _, element in ipairs(self.fadeableElements) do
    animation:addAction(AHOpacityTo.new(element, duration, ease, 1))
  end

  local delay = self:getAppearanceDelayDuration()
  return AHDelayed.new(delay, animation):start()
end

function OnboardioPanelMotivationReinforcement:playDisappear()
  local animation = AHSpawn.new()
  local duration = 0.3
  local ease = AHEaseSimpleInOut.new()
  for _, element in ipairs(self.fadeableElements) do
    animation:addAction(AHOpacityTo.new(element, duration, ease, 0))
  end

  return animation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, motivationId)
  createAndInsertGameScreens(self)
  createAndInsertReinforcementText(self, motivationId)
  createAndInsertContinueButton(self)
end

-- Constructor ------------------------

function OnboardioPanelMotivationReinforcement.new(motivationId)
  if (not motivationIdIsValid(motivationId)) then
    return nil
  end
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioPanelMotivationReinforcement)

  self.fadeableElements = {}

  createLayout(self, motivationId)

  hideFadeableElements(self)

  return self
end