--- OnboardioPanelPersonalizationQuestions
-- Much of the layout logic is based on OnboardioSkillSelector
-- @author Created by andrewcousins on August 27, 2024.

class("OnboardioPanelPersonalizationQuestions")

require("OnboardioModernPanel")
require("OnboardioSkillButtonBar")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local QUESTION_DESCRIPTORS = {
  {
    iconName = "personalization_lightbulb.png",
    questionText = Localizer.get("PERSONALIZATION_QUESTION_ARTICULATION"),
    analyticsEventMethod = OnboardioAnalytics.reportPersonalizationArticulateThoughtsAnswered,
    validInterests = {"articulation"}
  },
  {
    iconName = "personalization_notebook.png",
    questionText = Localizer.get("PERSONALIZATION_QUESTION_LISTENING"),
    analyticsEventMethod = OnboardioAnalytics.reportPersonalizationEasilyDistractedAnswered,
    validInterests = {"listening"}
  },
  {
    iconName = "personalization_pencil.png",
    questionText = Localizer.get("PERSONALIZATION_QUESTION_VOCABULARY"),
    analyticsEventMethod = OnboardioAnalytics.reportPersonalizationThoughtsOnPaperAnswered,
    validInterests = {"improve_your_mental_vocabulary"}
  },
  {
    iconName = "personalization_calendar.png",
    questionText = Localizer.get("PERSONALIZATION_QUESTION_RETAIN_MORE"),
    analyticsEventMethod = OnboardioAnalytics.reportPersonalizationForgetNamesAnswered,
    validInterests = {"retain_more"}
  },
  {
    iconName = "personalization_calculator.png",
    questionText = Localizer.get("PERSONALIZATION_QUESTION_PROBLEM_SOLVING"),
    analyticsEventMethod = OnboardioAnalytics.reportPersonalizationUseCalculatorAnswered,
    validInterests = {"problem_solving"}
  }
}

local COLOR_SEPARATOR = OnboardioCommon.COLOR_TABLE.GRAY200
local COLOR_DESCRIPTION = OnboardioCommon.COLOR_TABLE.ONBOARDING_TEXT_DARK

local BOUNCE_DURATION = 0.65
local IMAGE_SIZE = Screen.getWidthPart(0.3)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function prepareNextImage(self)
  local description = self.questionDescriptions[self.index]
  self.image = Sprite.newWithName(description.iconName)
  table.insert(self.fadeableElements, self.image)
  self.image:setScl(2)
  self.image:setOpacity(0)
  self.imageHolder:insertAtCenter(self.image)
end

local function switchImage(self)
  local duration = 0.15
  if (self.image ~= nil) then
    AHOpacityTo.new(self.image, duration, nil, 0):start()
    AHScaleTo.new(self.image, duration, nil, 0.2):start()
  end

  prepareNextImage(self)
  self.image:setScl(0.2)
  AHDelayed.new(duration, AHSpawn.new(
    AHOpacityTo.new(self.image, duration/2, nil, 1),
    AHScaleTo.new(self.image, duration/2, nil, 2))
  ):start()
end

local function createBounceAnimation(view, wasPressed)
  local scaleDuration = BOUNCE_DURATION
  local scaleDownAmount = 0.95

  if wasPressed then
    scaleDownAmount = 0.85
  end

  return AHSequence.new(
    AHScaleTo.new(view, 0.2 * scaleDuration, AHEaseSimpleOut.new(2.0), scaleDownAmount),
    AHScaleTo.new(view, 0.8 * scaleDuration, AHEaseElasticOut.new(0.45), 1.0)
  )
end

local function createSwitchTextAndImageAnimation(self)
  local wasPressed = false
  local bounceAction = createBounceAnimation(self.descriptionArea, wasPressed)

  local switchDuration = BOUNCE_DURATION

  local fadeAndTextSwitch = AHSequence.new(
    AHOpacityTo.new(self.descriptionArea, 0.25 * switchDuration, nil, 0.0),
    function()
      local descriptionText = self.questionDescriptions[self.index].questionText
      self.descriptionLabel:setText(descriptionText)
      switchImage(self)
    end,
    AHOpacityTo.new(self.descriptionArea, 0.75 * switchDuration, nil, 1.0)
  )

  return AHSpawn.new(bounceAction, fadeAndTextSwitch)
end

local function animateAndIncrementIndex(self, interestWasSelected)
  self.index = self.index + 1

  self.skillButtons:setTouchEnabled(false)

  local selectedButton = self.skillButtons:getYesButton()
  local unselectedButton = self.skillButtons:getNoButton()

  if not interestWasSelected then
    selectedButton, unselectedButton = unselectedButton, selectedButton
  end

  selectedButton:showAsSelected()

  AHSpawn.new(
    createSwitchTextAndImageAnimation(self),
    createBounceAnimation(unselectedButton, false),
    createBounceAnimation(selectedButton, true),
    AHDelayed.new(0.5 * BOUNCE_DURATION, function()
      wait(selectedButton:fadeInLabel())
      self.skillButtons:setTouchEnabled(true)
    end)
  ):start()
end

local function animateLastIteration(self, interestWasSelected)
  self.skillButtons:setTouchEnabled(false)
  local selectedButton = self.skillButtons:getYesButton()
  if not interestWasSelected then
    selectedButton = self.skillButtons:getNoButton()
  end

  selectedButton:showAsSelected()

  return AHSpawn.new(
    createBounceAnimation(selectedButton, true),
    AHDelayed.new(0.5 * BOUNCE_DURATION, function()
      selectedButton:fadeInLabel()
    end)
  ):start()
end

local function getOnAnswerHandler(self)
  return function(interestWasSelected)
    local currentInterest = self.questionDescriptions[self.index]
    currentInterest.analyticsEventMethod(interestWasSelected)
    if (self.index >= #self.questionDescriptions) then
      wait(animateLastIteration(self, interestWasSelected))
      self:setDismissed()
    else
      animateAndIncrementIndex(self, interestWasSelected)
    end
  end
end

local function createQuestionTextArea(text)
  local font = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 22)
  local label = DynamicTextArea.newSimple(text, Screen.getWidthPart(0.8), font, Alignment.CENTER)
  label:setColor(unpack(COLOR_DESCRIPTION))
  return label
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelPersonalizationQuestions:playAppear()
  AHDelayed.new(self:getAppearanceDelayDuration(), function()
      self.skillButtons:fadeIn()
    end):start()
  self.super.playAppear(self)
end

function OnboardioPanelPersonalizationQuestions:playDisappear()
  self.skillButtons:fadeOut()
  self.super.playDisappear(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function shouldShowDescription(interests, interestsMap)
  for _, interest in ipairs(interests) do
    if interestsMap[interest] then
      return true
    end
  end
  return false
end

local function getQuestionDescriptions()
  local interestsMap = GameSession.interestsMap
  if (interestsMap == nil) then
    Log.error("Pretest Interests have not been set - skipping Social Proof screen.")
    return {}
  end

  local descriptions = {}
  for _, description in ipairs(QUESTION_DESCRIPTORS) do
    local interests = description.validInterests
    if shouldShowDescription(interests, interestsMap) then
      table.insert(descriptions, description)
    end
  end

  return descriptions
end



local function createLayout(self)
  self.imageHolder = Group.new(IMAGE_SIZE, IMAGE_SIZE)
  local descriptionText = self.questionDescriptions[self.index].questionText
  self.descriptionLabel = createQuestionTextArea(descriptionText)
  table.insert(self.fadeableElements, self.descriptionLabel)
  self.descriptionLabel:setOpacity(0)
  self.skillButtons = OnboardioSkillButtonBar.new(getOnAnswerHandler(self))
  self.skillButtons:setVisible(false)

  self.descriptionArea = Column.new(24)
  self.descriptionArea:insert(self.imageHolder)
  self.descriptionArea:insert(self.descriptionLabel)
  self.descriptionArea:alignChildren()

  self.seperatorLine = FilledRectangle.new(Screen.getWidthPart(0.9), 1, COLOR_SEPARATOR)
  table.insert(self.fadeableElements, self.seperatorLine)
  self.seperatorLine:setOpacity(0)

  local backgroundTexture = Sprite.newWithName("personalization_background.png")
  backgroundTexture:setScl(2)
  table.insert(self.fadeableElements, backgroundTexture)
  backgroundTexture:setOpacity(0)

  local column = Column.new()
  column:insert(backgroundTexture)
  column:insertSpacer(backgroundTexture:getHeightPart(-0.95))
  column:insert(self.descriptionArea)
  column:insertSpacer(Screen.getHeightPart(0.065))
  column:insert(self.seperatorLine)
  column:insertSpacer(Screen.getHeightPart(0.065))
  column:insert(self.skillButtons)
  column:insertSpacer(backgroundTexture:getHeightPart(0.95))
  column:alignChildren()
  column:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.485))
  self:insert(column)

end

-- Constructor ------------------------

function OnboardioPanelPersonalizationQuestions.new()
  local self = OnboardioModernPanel.new()
  Mixin.mix(self, OnboardioPanelPersonalizationQuestions)

  local questionDescriptions = getQuestionDescriptions()
  if (#questionDescriptions < 1) then
    -- If no heck yes questions to show, skip the whole panel
    return nil
  end

  self.index = 1
  self.questionDescriptions = questionDescriptions

  createLayout(self)
  prepareNextImage(self)

  return self
end
