--- OnboardioPanelSocialProof
-- @author Created by andrewcousins on November 16, 2023.

class("OnboardioPanelSocialProof")

require("OnboardioPanel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TITLE_TEXT_COLOR = OnboardioCommon.COLOR_PROMPT_TEXT

local RESULTS_MARGIN_Y = Screen.getHeightPart(0.03)
local RESULTS_MARGIN_X_ICON_TO_PERCENTAGE = Screen.getWidthPart(0.02)
local RESULTS_MARGIN_X_INFOGRAPHIC_TO_TEXT = Screen.getWidthPart(0.03)
local RESULTS_TEXT_COLOR = OnboardioCommon.COLOR_TABLE.GRAY600
local RESULTS_ICON_OVERSCALE = 1.25
local RESULTS_BACKING_OVERSCALE = 1.20
local RESULTS_RATIO_BACKING_TO_PERCENTAGE = 2.40

local RESULT_SEPARATOR_COLOR = OnboardioCommon.COLOR_TABLE.GRAY200
local RESULT_SEPARATOR_THICKNESS = 1

local RESULTS_DESCRIPTORS = {
  {
    iconName = "icon_social_proof_sharper.png",
    percentage = 93,
    bodyText = Localizer.get("SOCIAL_PROOF_RESULT_BODY_SHARPER"),
    backingColor = OnboardioCommon.COLOR_TABLE.LISTENING100,
    validInterests = {"articulation", "listening", "retain_more"}
  },
  {
    iconName = "icon_social_proof_vocabulary.png",
    percentage = 90,
    bodyText = Localizer.get("SOCIAL_PROOF_RESULT_BODY_VOCABULARY"),
    backingColor = OnboardioCommon.COLOR_TABLE.SPEAKING100,
    validInterests = {"process_information_faster"}
  },
  {
    iconName = "icon_social_proof_math.png",
    percentage = 89,
    bodyText = Localizer.get("SOCIAL_PROOF_RESULT_BODY_MATH"),
    backingColor = OnboardioCommon.COLOR_TABLE.MATH100,
    validInterests = {"problem_solving"}

  },
  {
    iconName = "icon_social_proof_writing.png",
    percentage = 88,
    bodyText = Localizer.get("SOCIAL_PROOF_RESULT_BODY_WRITING"),
    backingColor = OnboardioCommon.COLOR_TABLE.WRITING100,
    validInterests = {"improve_your_mental_vocabulary"}
  }
}

local LAYOUT_MARGIN_Y = Screen.getHeightPart(0.06)
local DISCLAIMER_TEXT_COLOR = OnboardioCommon.COLOR_TABLE.GRAY400

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function shouldShowDescription(interests, interestsMap)
  for _, interest in ipairs(interests) do
    if interestsMap[interest] then
      return true
    end
  end
  return false
end

local function getDescriptionsToShow(self)
  local interestsMap = GameSession.interestsMap
  if (interestsMap == nil) then
    Log.error("Pretest Interests have not been set - skipping Social Proof screen.")
    return {}
  end

  local descriptions = {}
  for _, description in ipairs(RESULTS_DESCRIPTORS) do
    local interests = description.validInterests
    if shouldShowDescription(interests, interestsMap) then
      table.insert(descriptions, description)
    end
  end

  return descriptions
end

local function hideFadeableElements(self)
  for _, element in ipairs(self.fadeableElements) do
    element:setOpacity(0)
  end
end

local function createTextArea(text, font, color, lineSpacing)
  local textWidth = Screen.getSafeWidthPart(0.95)
  local textArea = DynamicTextArea.newSimple(text, textWidth, font, Alignment.CENTER, Alignment.CENTER, lineSpacing)
  textArea:setColor(unpack(color))

  return textArea
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelSocialProof:playAppear()
  OnboardioAnalytics.reportSocialProofScreenSeen()

  local animation = AHSpawn.new()
  local duration = 0.4
  local ease = AHEaseSimpleInOut.new()
  for _, element in ipairs(self.fadeableElements) do
    animation:addAction(AHOpacityTo.new(element, duration, ease, 1))
  end

  return animation:start()
end

function OnboardioPanelSocialProof:playDisappear()
  local animation = AHSpawn.new()
  local duration = 0.3
  local ease = AHEaseSimpleInOut.new()
  for _, element in ipairs(self.fadeableElements) do
    animation:addAction(AHOpacityTo.new(element, duration, ease, 0))
  end

  return animation:start()
end

function OnboardioPanelSocialProof:getAppearanceDelayDuration()
  return 0.25
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Consider the screen locations of two rows and make a separator halfway between.
-- getCenterWorldLoc only transforms to the coordinate space of the rows' parent
-- (the Column). One additional transform from the Column's space to its parent's
-- (the OnboardioPanelSocialProof group) is required to get into screen coords.
local function createPositionedSeparator(parentGroup, groupAbove, groupBelow)
  local width = math.min(parentGroup:getWidth(), OnboardioCommon.TEXT_WIDTH)
  local height = RESULT_SEPARATOR_THICKNESS
  local separator = FilledRectangle.new(width, height, RESULT_SEPARATOR_COLOR)

  local _, groupAboveYParentSpace = groupAbove:getCenterWorldLoc()
  local _, groupBelowYParentSpace = groupBelow:getCenterWorldLoc()

  local groupAboveYScreenSpace = groupAboveYParentSpace + parentGroup:getY()
  local groupBelowYScreenSpace = groupBelowYParentSpace + parentGroup:getY()

  local groupAboveBottomY = groupAboveYScreenSpace - groupAbove:getHeightPart(0.5)
  local groupBelowTopY = groupBelowYScreenSpace + groupBelow:getHeightPart(0.5)

  local x = Screen.getWidthPart(0.5)
  local y = groupBelowTopY + ((groupAboveBottomY - groupBelowTopY) / 2)

  separator:setLoc(x, y)

  return separator
end

-- A row containing [ {percentage infographic} + {result text} ]
-- Ex: [ { [icon] 93% } + "felt mentally stronger" ]
local function createResultRow(percentageInfographic, resultText)
  local resultLabel = Label.new(resultText, Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  resultLabel:setColor(unpack(RESULTS_TEXT_COLOR))

  local row = Row.new(RESULTS_MARGIN_X_INFOGRAPHIC_TO_TEXT)
  row:insert(percentageInfographic)
  row:insert(resultLabel)

  row:alignChildren()

  local isRecursive = true
  row:linkOpacityAttrToChildren(isRecursive)

  return row
end

-- A group containing [ {icon} + {percent} ] inside a colored rounded backing
-- Ex: [ {icon} + 93% ]
local function createPercentageInfographic(iconName, percentage, backingColor)
  local percentageLabel = Label.new(string.format("%d%%", percentage), Font.new(FontFaces.DIN_OT_MEDIUM, 24))
  percentageLabel:setColor(unpack(RESULTS_TEXT_COLOR))

  local icon = Sprite.newWithName(iconName)
  local desiredIconHeight = percentageLabel:getHeight() * RESULTS_ICON_OVERSCALE
  icon:setEffectiveDims(desiredIconHeight, desiredIconHeight)

  local desiredBackingWidth = percentageLabel:getWidth() * RESULTS_RATIO_BACKING_TO_PERCENTAGE
  local desiredBackingHeight = icon:getEffectiveHeight() * RESULTS_BACKING_OVERSCALE
  local radius = desiredBackingHeight/2
  local group = FilledRoundedRectangle.new(desiredBackingWidth, desiredBackingHeight, radius, backingColor)

  local iconOffsetXToMaintainMargin = ((desiredBackingHeight - icon:getEffectiveHeight()) * 0.5)
  local iconLocationX = icon:getEffectiveHeightPart(0.5) + iconOffsetXToMaintainMargin
  group:insertAtLocation(icon, iconLocationX, group:getHeightPart(0.5))

  percentageLabel:setAnchorPointLeft()
  local percentageLabelX = iconLocationX + icon:getEffectiveWidthPart(0.5) + RESULTS_MARGIN_X_ICON_TO_PERCENTAGE
  local percentageLabelYVisuallyCentered = group:getHeightPart(0.53)
  group:insertAtLocation(percentageLabel, percentageLabelX, percentageLabelYVisuallyCentered)

  local isRecursive = true
  group:linkOpacityAttrToChildren(isRecursive)

  return group
end

local function createAndInsertPositionedTitleText(self, resultsColumn)
  assert(self.fadeableElements ~= nil, "fadeableElements table must be created prior to creating title")

  local titleText = Localizer.get("SOCIAL_PROOF_TITLE")
  local titleTextArea = createTextArea(titleText, Font.new(FontFaces.DIN_OT_MEDIUM, 24), TITLE_TEXT_COLOR)
  local titleTextY = resultsColumn:getYTop() + LAYOUT_MARGIN_Y + titleTextArea:getHeightPart(0.5)
  self:insertAtLocation(titleTextArea, self:getWidthPart(0.5), titleTextY)
  table.insert(self.fadeableElements, titleTextArea)
end

-- A column containing a bunch of result rows
--
--  [ result row ]
--  [ result row ]
--       ...
--
--  ex:
--      [ { [icon] 93% } + "felt mentally stronger" ]
--      [ { [icon] 90% } + "expanded their vocabulary" ]
--                       ...
local function createAndInsertPositionedResultsColumn(self, resultDescriptors)
  assert(self.fadeableElements ~= nil, "fadeableElements table must be created prior to creating results column")

  local resultsColumn = Column.new(RESULTS_MARGIN_Y, Alignment.LEFT)
  for _, descriptor in ipairs(resultDescriptors) do
    local percentageInfographic = createPercentageInfographic(descriptor.iconName, descriptor.percentage, descriptor.backingColor)
    local resultRow = createResultRow(percentageInfographic, descriptor.bodyText)
    resultsColumn:insert(resultRow)
  end
  resultsColumn:alignChildren()

  local isRecursive = true
  resultsColumn:linkOpacityAttrToChildren(isRecursive)

  self:insertAtLocationFractional(resultsColumn, 0.5, 0.5)
  table.insert(self.fadeableElements, resultsColumn)

  return resultsColumn
end

-- Takes in a results column. Calculates where separators should be based on the
-- world positions of all the rows. Creates separators for each pair of rows and
-- inserts them at the appropriate world locations in the OnboardioPanelSocialProof
-- group.
local function createAndInsertPositionedResultsSeparators(self, resultsColumn)
  assert(self.fadeableElements ~= nil, "fadeableElements table must be created prior to creating results separators")

  -- Necessary to ensure that position changes are applied to all nodes prior to
  -- creating positioned separators between row objects. Without this yield, group
  -- world location getter returns incorrect values.
  coroutine.yield()

  for rowIndex = 1, #resultsColumn.children - 1 do
    local rowAbove = resultsColumn.children[rowIndex]
    local rowBelow = resultsColumn.children[rowIndex + 1]

    local separator = createPositionedSeparator(resultsColumn, rowAbove, rowBelow)
    self:insert(separator)
    table.insert(self.fadeableElements, separator)
  end
end

local function createAndInsertPositionedContinueButton(self)
  assert(self.fadeableElements ~= nil, "fadeableElements table must be created prior to creating continue button")

  local usePretestStyle = false
  local buttonGroup, continueButton = OnboardioOutroCommon.newButtonGroup(
    Localizer.get("CONTINUE"),
    function()
      OnboardioAnalytics.reportSocialProofContinueTapped()
      self:setDismissed()
    end,
    usePretestStyle
  )

  local buttonGroupY = math.max(self:getHeightPart(0.01), Screen.safeBottom)
  buttonGroup:setAnchorPointBottom()
  self:insertAtLocation(buttonGroup, self:getWidthPart(0.5), buttonGroupY)
  table.insert(self.fadeableElements, continueButton)

  return buttonGroup
end

local function createAndInsertPositionedDisclaimerText(self, resultsColumn, buttonGroup)
  assert(self.fadeableElements ~= nil, "fadeableElements table must be created prior to creating disclaimer text")

  local disclaimerText = Localizer.get("SOCIAL_PROOF_DISCLAIMER")
  local lineSpacing = 3
  local disclaimerTextArea = createTextArea(disclaimerText, Font.new(FontFaces.DIN_OT_MEDIUM, 14), DISCLAIMER_TEXT_COLOR, lineSpacing)
  local disclaimerY = resultsColumn:getYBottom() - LAYOUT_MARGIN_Y
  local disclaimerMinimumY = buttonGroup:getYTop() + disclaimerTextArea:getHeightPart(1.5)
  disclaimerTextArea:setAnchorPointTop()
  self:insertAtLocation(disclaimerTextArea, self:getWidthPart(0.5), math.max(disclaimerY, disclaimerMinimumY))
  table.insert(self.fadeableElements, disclaimerTextArea)
end

local function createLayout(self, resultDescriptors)
  self.fadeableElements = {}

  local resultsColumn = createAndInsertPositionedResultsColumn(self, resultDescriptors)
  createAndInsertPositionedResultsSeparators(self, resultsColumn)
  createAndInsertPositionedTitleText(self, resultsColumn)
  local buttonGroup = createAndInsertPositionedContinueButton(self)
  createAndInsertPositionedDisclaimerText(self, resultsColumn, buttonGroup)

  hideFadeableElements(self)
end

-- Constructor ------------------------

function OnboardioPanelSocialProof.new()
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioPanelSocialProof)

  local resultDescriptors = getDescriptionsToShow(self)
  if (#resultDescriptors < 1) then
    return nil
  end

  createLayout(self, resultDescriptors)

  return self
end
