--- OnboardioQuestionPanel
-- @author Created by andrewcousins on November 17, 2023.

class("OnboardioQuestionPanel")

require("OnboardioPanel")
require("OnboardioButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createButton(self, text, answerID, isInverted, continueCallback)
  isInverted = isInverted or false
  local callback = function()
    continueCallback(self, answerID)
  end
  local button = nil
  if isInverted then
    local shouldFitToLabel = true
    button = OnboardioButton.newInverted(text, callback, shouldFitToLabel)
  else
    button = OnboardioButton.newSmall(text, callback)
  end
  button:setOpacity(0)
  return button
end

local function createButtons(self, buttonDescriptors, continueCallback)
  local buttons = {}
  for _, buttonDescription in ipairs(buttonDescriptors) do
    local button = createButton(self, buttonDescription.text, buttonDescription.id, buttonDescription.isInverted, continueCallback)
    table.insert(buttons, button)
  end
  return buttons
end

local function createTextArea(text, font, color)
  local textWidth = Screen.getSafeWidthPart(0.95)
  local label = DynamicTextArea.newSimple(text, textWidth, font, Alignment.CENTER)
  label:setColor(unpack(color))
  label:setOpacity(0)
  return label
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioQuestionPanel:getAppearanceDelayDuration()
  return 0.25
end

function OnboardioQuestionPanel:playAppear()
  local animation = AHSpawn.new()
  local duration = 0.4
  local ease = AHEaseSimpleInOut.new()
  for _, label in ipairs(self.fadableLabels) do
    animation:addAction(AHOpacityTo.new(label, duration, ease, 1))
  end
  for _, button in ipairs(self.buttons) do
    animation:addAction(AHOpacityTo.new(button, duration, ease, 1))
    animation:addAction(function() button:setTouchEnabled(true) end)
  end

  local delay = self:getAppearanceDelayDuration()
  return AHDelayed.new(delay, animation):start()
end

function OnboardioQuestionPanel:playDisappear()
  local animation = AHSpawn.new()
  local duration = 0.3
  local ease = AHEaseSimpleInOut.new()
  for _, label in ipairs(self.fadableLabels) do
    animation:addAction(AHOpacityTo.new(label, duration, ease, 0))
  end
  for _, button in ipairs(self.buttons) do
    animation:addAction(AHOpacityTo.new(button, duration, ease, 0))
    button:setTouchEnabled(false)
  end
  return animation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

-- @param prompText (String) - The main question text
-- @param buttonDescriptors (Table) - descriptions of each button, in order. Each button should have
-- -- text (String) - The button text, already pulled from the localizer
-- -- id (String) - The button id, to be passed into continueCallback when pressed
-- -- isInverted (Boolean, Optional) - If true, display as an inverted button
-- @param continueCallback (Function) - Called when any button is pressed, takes self and the buttonID as arguments
-- @param subtitleText (String, Optional) - The subtitle text, displayed just below the question
function OnboardioQuestionPanel.new(promptText, buttonDescriptors, continueCallback, subtitleText)
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioQuestionPanel)

  local columnSpacer = Screen.getHeightPart(0.01)
  local column = Column.new(columnSpacer)
  local textArea = createTextArea(promptText, Font.new(FontFaces.DIN_OT_MEDIUM, 24), OnboardioCommon.COLOR_PROMPT_TEXT)
  table.insert(self.fadableLabels, textArea)
  column:insert(textArea)

  if (subtitleText ~= nil) then
    local subtitleTextArea = createTextArea(subtitleText, Font.new(FontFaces.DIN_OT_MEDIUM, 14), OnboardioCommon.COLOR_PROMPT_ADDITIONAL_TEXT)
    table.insert(self.fadableLabels, subtitleTextArea)
    column:insert(subtitleTextArea)
  end

  column:insertSpacer(columnSpacer * 4)

  local buttons = createButtons(self, buttonDescriptors, continueCallback)
  for _, button in ipairs(buttons) do
    table.insert(self.buttons, button)
    column:insert(button)
  end

  column:alignChildren()
  self:insertAtCenter(column)

  return self
end
