--- OnboardioDemoProgress
-- @author Created by karl on October 16, 2014.

class("OnboardioDemoProgress")
require("OnboardioCommon")
require("widgets.progress.SpriteRevealProgressIndicator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SPRITE_NAME = "progress_bar.webp"

local FADE_IN_DURATION = 0.8
local HOLD_DURATION = 0.5
local FADE_OUT_DURATION = 0.5

local BUBBLE_WIDTH = Screen.getWidthPart(0.31)
local BUBBLE_HEIGHT = Screen.getHeightPart(0.1)

local LABEL_FONT = Font.new(OnboardioCommon.FONT_FACE_MEDIUM, 14)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function newFadeInAction(self, index)
  local bubble = self.bubbles[index]
  local progress = self.progressMeters[index]

  local duration = FADE_IN_DURATION
  local ease = AHEaseSineInOut.new()

  local action =  AHSpawn.new(
    AHOpacityTo.new(bubble, duration, ease, 1),
    AHOpacityTo.new(progress, duration, ease, 1)
  )

  -- Hide the previous progress meter if necessary
  if index > 1 then
    action = AHSequence.new(
      action,
      AHOpacityTo.new(self.progressMeters[index-1], 0, nil, 0)
    )
  end

  return action
end

local function newFadeOutAction(self, index)
  local bubble = self.bubbles[index]

  local duration = FADE_OUT_DURATION
  local ease = AHEaseSineInOut.new()

  local action = AHSpawn.new(
    AHOpacityTo.new(bubble, duration, ease, 0)
  )

  -- Last fade out includes the progress meter.
  if index == self.stepCount then
    action:addAction(AHOpacityTo.new(self.progressMeters[index], duration, ease, 0))
  end

  return action
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioDemoProgress:animate()
  local action = AHSequence.new()
  for i=1,self.stepCount do
    action:addAction(newFadeInAction(self, i))
    action:addAction(AHDelay.new(HOLD_DURATION))
    action:addAction(newFadeOutAction(self, i))
  end
  self.animateAction = AHLoop.new(action):start()
  return self.animateAction
end

function OnboardioDemoProgress:stop()
  if self.animateAction then
    self.animateAction:stop()
    self.animateAction = nil
  end
end

function OnboardioDemoProgress:fadeOut(duration)
  local action = AHSpawn.new()
  for _, child in ipairs(self.children) do
    action:addAction(AHOpacityTo.new(child, duration, nil, 0))
  end
  return action:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newProgressIndicator(width, progress)
  local height = width * 0.04
  local color = OnboardioCommon.COLOR_BLUE
  local fgSprite = Sprite.newWithFilename(SPRITE_NAME)
  fgSprite:setEffectiveDims(width, height)
  fgSprite:setColor(unpack(color))
  local bgSprite = Sprite.newWithFilename(SPRITE_NAME)
  bgSprite:setEffectiveDims(width, height)
  bgSprite:setColor(unpack(OnboardioCommon.COLOR_PROGRESS))

  local indicator = SpriteRevealProgressIndicator.new(fgSprite, bgSprite, SpriteRevealProgressIndicator.GROW_RIGHT)
  indicator:setProgress(progress)

  return indicator
end

local function newLabel(text)
  local label = Label.new(text, LABEL_FONT)
  label:setColor(unpack(OnboardioCommon.COLOR_WHITE))

  return label
end

local function newBubble(formattedProgressLevel)
  local width = BUBBLE_WIDTH
  local height = BUBBLE_HEIGHT
  local cornerRadius = 4
  local bubble = FilledRoundedRectangle.new(width, height, cornerRadius, OnboardioCommon.COLOR_BLUE)

  local tailWidth = 7
  local tailHeight = 6
  local tail = FilledTriangle.new(0, tailHeight, tailWidth, tailHeight, tailWidth/2, 0, OnboardioCommon.COLOR_BLUE)
  tail:setLoc(bubble:getWidthPart(0.5), -tail:getHeightPart(0.5))
  bubble:insert(tail)

  local displayLevel = formattedProgressLevel.displayLevel
  local levelLabel = newLabel(string.upper(displayLevel))
  levelLabel:setLoc(bubble:getWidthPart(0.5), bubble:getHeightPart(0.65))
  bubble:insert(levelLabel)

  local scoreString = string.format("%d - %d", formattedProgressLevel.displayValueMinimum, formattedProgressLevel.displayValueMaximum)
  local scoreLabel = newLabel(scoreString)
  scoreLabel:setLoc(bubble:getWidthPart(0.5), bubble:getHeightPart(0.35))
  bubble:insert(scoreLabel)

  local isRecursive = true
  bubble:linkOpacityAttrToChildren(isRecursive)

  return bubble
end

-- PretestSession returns are a bit confusing so this creates a table that is easier to understand
-- and better matches the expected information for creating the demo progress layout
local function getFormattedProgressLevels()
  local formattedProgressLevels = {}

  local rawProgressLevels = {0}
  TableUtils.concatenateArrays(rawProgressLevels, PretestSession.getProgressLevels())

  for progressLevelIndex, progressLevel in ipairs(rawProgressLevels) do
    local displayValue = PretestSession.getPerformanceIndexDisplayValue(progressLevel)
    local displayLevel = PretestSession.getDisplayLevelFromIndex(progressLevel)
    local nextProgressLevel = rawProgressLevels[progressLevelIndex + 1] or 1.0

    local formattedProgressLevel = {
      displayLevel = displayLevel,
      displayValueMinimum = displayValue,
      displayValueMaximum = PretestSession.getPerformanceIndexDisplayValue(nextProgressLevel),
      progressLevelMinimum = progressLevel,
      progressLevelMaximum = nextProgressLevel
    }

    table.insert(formattedProgressLevels, formattedProgressLevel)
  end

  return formattedProgressLevels
end

local function initializeProgressLevels(self, width)
  self.progressMeters = {}
  self.bubbles = {}

  local formattedProgressLevels = getFormattedProgressLevels()

  for _, formattedProgressLevel in ipairs(formattedProgressLevels) do
    table.insert(self.progressMeters, newProgressIndicator(width, formattedProgressLevel.progressLevelMaximum))
    table.insert(self.bubbles, newBubble(formattedProgressLevel))
  end

  return formattedProgressLevels
end


-- Constructor ------------------------

function OnboardioDemoProgress.new(width)
  local emptyProgress = newProgressIndicator(width, 0)

  -- Without this, adding bubbles as children makes the group asymettrical and hard
  -- to center if bubbles extend farther on one side than another
  local bubblePaddingX = BUBBLE_WIDTH * 2
  local bubblePaddingY = emptyProgress:getHeight() + BUBBLE_HEIGHT
  local groupWidth = width + bubblePaddingX
  local groupHeight = bubblePaddingY
  local self = Group.new(groupWidth, groupHeight)
  Mixin.mix(self, OnboardioDemoProgress)

  local formattedProgressLevels = initializeProgressLevels(self, width)

  local progressX = self:getWidthPart(0.5)
  local progressWidth = self.progressMeters[1]:getWidth()
  local progressLeftEdge = (self:getWidth() - progressWidth) / 2

  local bubbleX = nil
  local bubbleY = BUBBLE_HEIGHT * 0.8

  emptyProgress:setLoc(progressX, 0)
  self:insert(emptyProgress)

  self.stepCount = #self.progressMeters

  for i = 1, self.stepCount do
    local progress = self.progressMeters[i]
    local bubble = self.bubbles[i]

    local progressLevelInfo = formattedProgressLevels[i]
    local normalizedMin = progressLevelInfo.progressLevelMinimum
    local normalizedMax = progressLevelInfo.progressLevelMaximum
    bubbleX = progressLeftEdge + progressWidth * (normalizedMin + (normalizedMax - normalizedMin) * 0.5)

    progress:setLoc(progressX, 0)
    bubble:setLoc(bubbleX, bubbleY)

    self:insert(progress)
    self:insert(bubble)

    progress:setOpacity(0)
    bubble:setOpacity(0)
  end

  return self
end
