--- OnboardioCarousel
-- @author Created by karl on November 06, 2015.

class("OnboardioCarousel")
require("OnboardioPanel")
require("OnboardioButton")
require("containers.Carousel")
require("OnboardioIntroPanel1")
require("OnboardioIntroPanel2")
require("OnboardioIntroPanel3")
require("OnboardioIntroPanel4")
require("OnboardioIntroPanel5")
require("OnboardioAnalytics")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function forceUpdateWhileActionRunning(self, action)
  -- HACK: Need to force updates because if certain other actions are
  -- running in a certain way, cardContainer's children will not update
  -- their absolute positions, and the scroll will not show on screen.
  runInCoroutine(function()
    while action:isActive() do
      self.carousel.cardContainer:forceUpdate()
      coroutine.yield()
    end
  end)
end

local function interceptProgressIndicatorToReportEvents(self)
  local originalSetProgressStepFunction = self.carousel.progressIndicator.setProgressStep
  -- HACK: We would normally include this step as part of the action of
  -- presenting a card. However, because we are using AHUntimedAction for the
  -- carousel, we cannot include simple AHCallLogic calls. Instead, we are
  -- intercepting the carousel's progress indicator (dots at the screen bottom)
  -- to get a discrete index signalling the scroll progress.
  self.carousel.progressIndicator.setProgressStep = function(progressIndicator, progressStep)
    local startingProgressStep = self:getCurrentProgressStep()
    originalSetProgressStepFunction(progressIndicator, progressStep)
    local endingProgressStep = self:getCurrentProgressStep()

    local lastCardIndex = #self.cards

    if (startingProgressStep ~= endingProgressStep) then
      local index = startingProgressStep
      if self.isScrollingToLastCard then
        if (endingProgressStep == lastCardIndex) then
          self.isScrollingToLastCard = false
        end
      else
        local swipedForward = startingProgressStep < endingProgressStep
        OnboardioAnalytics.reportPitchScreenSwiped(index, swipedForward)
      end
    end
  end
end

local function startAnimation(card)
  if card.startAnimation then
    card:startAnimation()
  end
end

local function pauseAnimation(card)
  if card.pauseAnimation then
    card:pauseAnimation()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioCarousel:animate()
  local action = self.cards[1]:animateOnce()
  forceUpdateWhileActionRunning(self, action)
end

function OnboardioCarousel:scrollToLastCard()
  self.isScrollingToLastCard = true
  return self.carousel:scrollToLastCard()
end

function OnboardioCarousel:getIcons()
  return self.icons
end

function OnboardioCarousel:setTouchEnabled(enabled)
  if self.carousel then
    self.carousel:setTouchEnabled(enabled)
  end
end

function OnboardioCarousel:getCurrentProgressStep()
  return self.carousel.progressIndicator:getProgressStep() + 1
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioCarousel.new(callbacks, width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, OnboardioCarousel)

  local onComplete = callbacks.onComplete
  local onEnterLastCard = callbacks.onEnterLastCard
  local onLeaveLastCard = callbacks.onLeaveLastCard
  local onEnterFirstCard = callbacks.onEnterFirstCard
  local onLeaveFirstCard = callbacks.onLeaveFirstCard

  local highlightColor = OnboardioCommon.COLOR_TABLE.PRIMARY400
  local neutralColor = OnboardioCommon.COLOR_TABLE.GRAY300

  self:linkColorAttrToChildren(true)

  local carouselBottomSafeArea = Screen.safeBottom
  self.carousel = Carousel.new(highlightColor, neutralColor, width, height, carouselBottomSafeArea)
  self:insertAtCenter(self.carousel)
  self.carousel:inheritColor(self)

  local canvasSize = math.min(width * 0.93, height * 0.7)

  local panel2 = OnboardioIntroPanel2.new(width, height, canvasSize)
  local icons = panel2:getIcons()
  self.icons = icons
  local panel3 = OnboardioIntroPanel3.new(icons, width, height, canvasSize)
  local panel4 = OnboardioIntroPanel4.new(icons, width, height, canvasSize)

  self.cards = {
    panel2,
    panel3,
    panel4
  }

  local panel5 = OnboardioIntroPanel5.new(icons, onComplete, width, height)
  table.insert(self.cards, panel5)

  self.overlays = Group.new(width, height)
  self:insertAtCenter(self.overlays)

  for index, card in ipairs(self.cards) do
    self.carousel:addCard(card)
    if index < #self.cards then
      card:inheritColor(self.carousel)
    end
    if card.getOverlayedGroup then
      self.overlays:insertAtCenter(card:getOverlayedGroup())
    end
    local scrollAction = AHSpawn.new()
    if card.getSlideInAction then
      scrollAction:addAction(card:getSlideInAction())
    end
    local previousCard = self.cards[index - 1]
    if previousCard and previousCard.getSlideOutAction then
      scrollAction:addAction(previousCard:getSlideOutAction())
    end
    self.carousel:setScrollAction(index-1, scrollAction)
  end

  self.carousel:setHorizontalScrollCallback(function()
    local scrollAmount = self.carousel:getScrollProportionX()
    local difference = scrollAmount % 1

    local visibleCardIndex = math.floor(scrollAmount) + 1
    if difference < 0.01 then
      local card = self.cards[visibleCardIndex]
      startAnimation(card)
    elseif difference >= 0.01 and difference < 0.5 then
      local card = self.cards[visibleCardIndex]
      pauseAnimation(card)
    elseif difference >= 0.5 and difference <= 0.99 then
      visibleCardIndex = visibleCardIndex + 1
      local card = self.cards[visibleCardIndex]
      pauseAnimation(card)
    else
      visibleCardIndex = visibleCardIndex + 1
      local card = self.cards[visibleCardIndex]
      startAnimation(card)
    end

    if (visibleCardIndex ~= self.lastVisibleCardIndex) then
      if (visibleCardIndex == #self.cards) then
        onEnterLastCard()
      elseif (self.lastVisibleCardIndex == #self.cards) then
        onLeaveLastCard()
      elseif (visibleCardIndex == 1) then
        onEnterFirstCard()
      elseif (self.lastVisibleCardIndex == 1) then
        onLeaveFirstCard()
      end
    end
    self.lastVisibleCardIndex = visibleCardIndex
  end)

  self.isScrollingToLastCard = false
  interceptProgressIndicatorToReportEvents(self)

  return self
end
