--- OnboardioSkillSelector
-- @author Created by karl on November 09, 2015.

class("OnboardioSkillSelector")

require("OnboardioSkillButtonBar")
require("OnboardioAnalytics")
require("OnboardioSkillAnimationWriting")
require("OnboardioSkillAnimationListening")
require("OnboardioSkillAnimationProcessing")
require("OnboardioSkillAnimationCalculation")
require("OnboardioSkillAnimationArticulation")
require("OnboardioSkillAnimationRetention")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- The identifiers are defined in onboarding.json. They also are referenced in first-session.json,
-- LevelGenerator.cpp, and OnboardioPanelSocialProof.lua. Keep these synced.
local DESCRIPTIONS = {
  {
    identifiers = {"process_information_faster"},
    displayKey = "GOAL_DESCRIPTION_VOCABULARY",
    analyticsIdentifier = "improve_vocabulary",
    animationClass = OnboardioSkillAnimationProcessing,
    width = Screen.getWidthPart(0.4)
  },
  {
    identifiers = {"articulation"},
    displayKey = "GOAL_DESCRIPTION_ARTICULATE",
    analyticsIdentifier = "articulate_thoughts_more_clearly",
    animationClass = OnboardioSkillAnimationArticulation,
    width = Screen.getWidthPart(0.45)
  },
  {
    identifiers = {"listening"},
    displayKey = "GOAL_DESCRIPTION_FOCUS",
    analyticsIdentifier = "strengthen_focus_while_reading",
    animationClass = OnboardioSkillAnimationListening,
    width = Screen.getWidthPart(0.33)
  },
  {
    identifiers = {"improve_your_mental_vocabulary"},
    displayKey = "GOAL_DESCRIPTION_WRITING",
    analyticsIdentifier = "improve_writing",
    animationClass = OnboardioSkillAnimationWriting,
    width = Screen.getWidthPart(0.51)
  },
  {
    identifiers = {"retain_more"},
    displayKey = "GOAL_DESCRIPTION_RETAIN",
    analyticsIdentifier = "enchance_memory",
    animationClass = OnboardioSkillAnimationRetention,
    width = Screen.getWidthPart(0.32)
  },
  {
    identifiers = {"problem_solving"},
    displayKey = "GOAL_DESCRIPTION_CALCULATE",
    analyticsIdentifier = "caluclate_more_quickly_and_precisely",
    animationClass = OnboardioSkillAnimationCalculation,
    width = Screen.getWidthPart(0.45)
  }
}

local COLOR_SEPARATOR = OnboardioCommon.COLOR_TABLE.GRAY200
local COLOR_DESCRIPTION = OnboardioCommon.COLOR_TABLE.ONBOARDING_TEXT_DARK

local BOUNCE_DURATION = 0.65
local ANIMATION_WIDTH = Screen.getWidthPart(0.3)

local FADE_OUT_DURATION = 0.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function switchImage(self)
  if self.animation then
    self.animation:removeSelf()
  end

  local description = self.skillDescriptions[self.index]
  local animationClass = description.animationClass
  self.animation = animationClass.new(description.width)
  self.animation:getAnimationAction():start()
  self.animation:setAnchorPointBottom()
  self.animation:setLoc(self.animationHolder:getWidthPart(0.5), -self.animationHolder:getHeightPart(0.125))
  self.animationHolder:insert(self.animation)
end

local function createBounceAnimation(view, wasPressed)
  local scaleDuration = BOUNCE_DURATION
  local scaleDownAmount = 0.95

  if wasPressed then
    scaleDownAmount = 0.85
  end

  return AHSequence.new(
    AHScaleTo.new(view, 0.2 * scaleDuration, AHEaseSimpleOut.new(2.0), scaleDownAmount),
    AHScaleTo.new(view, 0.8 * scaleDuration, AHEaseElasticOut.new(0.45), 1.0)
  )
end

local function createSwitchTextAndImageAnimation(self)
  local wasPressed = false
  local bounceAction = createBounceAnimation(self.descriptionArea, wasPressed)

  local switchDuration = BOUNCE_DURATION

  local fadeAndTextSwitch = AHSequence.new(
    AHOpacityTo.new(self.descriptionArea, 0.25 * switchDuration, nil, 0.0),
    function()
      local descriptionDisplay = Localizer.get(self.skillDescriptions[self.index].displayKey)
      self.descriptionLabel:setText(descriptionDisplay)
      switchImage(self)
    end,
    AHOpacityTo.new(self.descriptionArea, 0.75 * switchDuration, nil, 1.0)
  )

  return AHSpawn.new(bounceAction, fadeAndTextSwitch)
end

local function animateAndIncrementIndex(self, interestWasSelected)
  self.index = self.index + 1

  self.skillButtons:setTouchEnabled(false)

  local selectedButton = self.skillButtons:getYesButton()
  local unselectedButton = self.skillButtons:getNoButton()

  if not interestWasSelected then
    selectedButton, unselectedButton = unselectedButton, selectedButton
  end

  selectedButton:showAsSelected()

  AHSpawn.new(
    createSwitchTextAndImageAnimation(self),
    createBounceAnimation(unselectedButton, false),
    createBounceAnimation(selectedButton, true),
    AHDelayed.new(0.5 * BOUNCE_DURATION, function()
      wait(selectedButton:fadeInLabel())
      self.skillButtons:setTouchEnabled(true)
    end)
  ):start()
end

local function animateLastIteration(self, interestWasSelected)
  self.skillButtons:setTouchEnabled(false)
  local selectedButton = self.skillButtons:getYesButton()
  if not interestWasSelected then
    selectedButton = self.skillButtons:getNoButton()
  end

  selectedButton:showAsSelected()

  return AHSpawn.new(
    createBounceAnimation(selectedButton, true),
    AHDelayed.new(0.5 * BOUNCE_DURATION, function()
      selectedButton:fadeInLabel()
    end)
  ):start()
end

local function getOnAnswerHandler(self, onFinished)
  return function(interestWasSelected)
    local currentInterest = self.skillDescriptions[self.index]
    OnboardioAnalytics.reportGoalSelectionCardAnswered(currentInterest.analyticsIdentifier, interestWasSelected)
    for _, identifier in ipairs(currentInterest.identifiers) do
      self.answers[identifier] = interestWasSelected
    end
    if (self.index >= #self.skillDescriptions) then
      wait(animateLastIteration(self, interestWasSelected))
      onFinished(self.answers)
    else
      animateAndIncrementIndex(self, interestWasSelected)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioSkillSelector:playFadeAction()
  if (self.animation ~= nil) then
    if (self.animation.playFadeAction) then
      self.animation:playFadeAction(FADE_OUT_DURATION)
    end
  end
end

function OnboardioSkillSelector:reset()
  self.index = 1
  self:setOpacity(0)

  if self.animation then
    self.animation:removeSelf()
    self.animation = nil
  end

  self.descriptionLabel:setText("")
end

function OnboardioSkillSelector:getSkillButtonBar()
  return self.skillButtons
end

function OnboardioSkillSelector:playShowAnimation(previousScreen, lastCard, pretestButton)
  local buttonBar = self.skillButtons
  local yesButton = buttonBar:getYesButton()
  local noButton = buttonBar:getNoButton()

  self:bringToFront()

  local coverWidth = pretestButton:getWidthPart(0.57)
  local coverHeight = pretestButton:getHeight()
  local coverCornerRadius = yesButton:getCornerRadius()
  local coverColor = OnboardioCommon.COLOR_BLUE
  local yesCover = FilledRoundedRectangle.new(coverWidth, coverHeight, coverCornerRadius, coverColor)
  local noCover = FilledRoundedRectangle.new(coverWidth, coverHeight, coverCornerRadius, coverColor)
  self:insert(yesCover)
  self:insert(noCover)

  yesButton:setVisibleRecursive(false)
  noButton:setVisibleRecursive(false)

  local pretestButtonXLeft, pretestButtonY = pretestButton:getWorldLoc()
  local pretestButtonXRight = pretestButtonXLeft + pretestButton:getWidth()
  local pretestButtonHeight = pretestButton:getHeight()
  yesCover:setX(pretestButtonXRight - coverWidth/2)
  yesCover:setY(pretestButtonY + pretestButtonHeight/2)
  noCover:setX(pretestButtonXLeft + coverWidth/2)
  noCover:setY(pretestButtonY + pretestButtonHeight/2)

  local bounceInDuration = BOUNCE_DURATION * 0.2
  local bounceOutDuration = BOUNCE_DURATION * 0.8

  local bounceLastCard = createBounceAnimation(lastCard)
  local bouncePretestButton = createBounceAnimation(pretestButton)
  local bounceCovers = AHSpawn.new(
    createBounceAnimation(yesCover),
    createBounceAnimation(noCover)
  )

  local switchTextAndImage = createSwitchTextAndImageAnimation(self)
  local fadePreviousScreenOut = AHOpacityTo.new(previousScreen, bounceInDuration, nil, 0)
  local fadeLastCardOut = AHOpacityTo.new(lastCard, bounceInDuration, nil, 0)
  local fadeSelfIn = AHOpacityTo.new(self, bounceOutDuration, nil, 1)

  local buttonWidth, buttonHeight = yesButton:getDims()
  local yesCoverX, yesCoverY = self:transformObjectLoc(yesButton)
  local noCoverX, noCoverY = self:transformObjectLoc(noButton)
  local moveDuration = 0.6
  local moveAndScaleCovers = AHSpawn.new(
    AHDimsTo.new(yesCover, moveDuration, AHEaseExponentialOut.new(), buttonWidth, buttonHeight),
    AHDimsTo.new(noCover, moveDuration, AHEaseExponentialOut.new(), buttonWidth, buttonHeight),
    AHLocationTo.new(yesCover, moveDuration, AHEaseExponentialOut.new(), yesCoverX, yesCoverY),
    AHLocationTo.new(noCover, moveDuration, AHEaseExponentialOut.new(), noCoverX, noCoverY)
  )

  local fadeCoversDuration = 0.3
  local fadeCovers = AHSpawn.new(
    AHOpacityTo.new(yesCover, fadeCoversDuration, nil, 0),
    AHOpacityTo.new(noCover, fadeCoversDuration, nil, 0)
  )

  return runInCoroutine(function()
    bounceLastCard:start()
    bouncePretestButton:start()
    bounceCovers:start()
    fadePreviousScreenOut:start()
    fadeLastCardOut:start()
    switchTextAndImage:start()
    waitFor(bounceInDuration)

    fadeSelfIn:start()
    moveAndScaleCovers:start()
    waitFor(moveDuration - 0.3)

    yesButton:setVisibleRecursive(true)
    noButton:setVisibleRecursive(true)
    fadeCovers:start()
  end)
end

function OnboardioSkillSelector:getCurrentInterestAnalyticsIdentifier()
  local currentInterest = self.skillDescriptions[self.index]
  return currentInterest.analyticsIdentifier
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createTextArea(text, font, color)
  local label = DynamicTextArea.newSimple(text, Screen.getWidthPart(0.8), font, Alignment.CENTER)
  label:setColor(unpack(color))
  return label
end

local function createAnimationSpriteHolder()
  return Group.new(ANIMATION_WIDTH, ANIMATION_WIDTH)
end

local function createSeparatorLine()
  return FilledRectangle.new(Screen.getWidthPart(0.9), 1, COLOR_SEPARATOR)
end


-- Constructor ------------------------

function OnboardioSkillSelector.new(onFinished)
  local self = Group.newFullScreen()
  Mixin.mix(self, OnboardioSkillSelector)

  self.index = 0
  self.answers = {}
  self.skillDescriptions = DESCRIPTIONS

  self.animationHolder = createAnimationSpriteHolder()
  local descriptionDisplay = Localizer.get(self.skillDescriptions[1].displayKey)
  self.descriptionLabel = createTextArea(descriptionDisplay, PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 24), COLOR_DESCRIPTION)
  self.skillButtons = OnboardioSkillButtonBar.new(getOnAnswerHandler(self, onFinished))

  self.descriptionArea = Column.new(Screen.getHeightPart(0.09))
  self.descriptionArea:insert(self.animationHolder)
  self.descriptionArea:insert(self.descriptionLabel)
  self.descriptionArea:alignChildren()
  self.descriptionArea:inheritOpacity(self)

  local column = Column.new()
  column:insert(self.descriptionArea)
  column:insertSpacer(Screen.getHeightPart(0.065))
  column:insert(createSeparatorLine())
  column:insertSpacer(Screen.getHeightPart(0.065))
  column:insert(self.skillButtons)
  column:alignChildren()
  column:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.485))
  self:insert(column)

  self.skillButtonsX, self.skillButtonsY = self.skillButtons:getWorldLoc()

  self:linkColorAttrToChildren(true)

  self:reset()

  return self
end
