--- OnboardioGraphAnimationWriting
-- @author Created by Joe Poka on July 18, 2016.

class("OnboardioGraphAnimationWriting")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- dimensions based on OnboardioGraphAnimationReading
local DIMENSIONS = {
  width = 95.0,
  graphHeight = 30.0,
  verticalSpacing = 6.0,
  rectangleHeight = 2.0
}

local MOVING_DIRECTION = {
  up = "up",
  down = "down",
  none = "none"
}

local HUE = {
  bar = {0.5, 0.5, 0.5},
  circle = {0.6, 0.6, 0.6},
  line = {0.92, 0.92, 0.975}
}

local EASE_FACTORS = {
  distance = 0.13,
  decay = 0.9,
  accelerate = 1.12
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateGraph(self)
  for i, pointData in ipairs(self.points) do
    local direction = pointData.direction
    local movingUp = (direction == MOVING_DIRECTION.up)
    local movingDown = (direction == MOVING_DIRECTION.down)

    -- recall end points are stationary
    if (movingUp or movingDown) then
      local deltaY = pointData.deltaY
      local updatedY = pointData.y + deltaY
      pointData.y = updatedY
      self.line:setVertexPosition(i, pointData.x, updatedY)
      self.circles[i]:setY(updatedY)

      -- ease in/out near top/bottom
      local distance = math.abs(pointData.maxY - pointData.minY)
      local easeDistance = distance * EASE_FACTORS.distance
      local distanceFromTop = math.abs(pointData.maxY - pointData.y)
      local distanceFromBottom = math.abs(pointData.y - pointData.minY)
      local inTopZone = (distanceFromTop < easeDistance)
      local inBottomZone = (distanceFromBottom < easeDistance)

      local slowDown = (inTopZone and movingUp) or (inBottomZone and movingDown)
      local speedUp = (inTopZone and movingDown) or (inBottomZone and movingUp)

      local minimumDeltaYMagnitude = pointData.minimumDeltaYMagnitude
      local maximumDeltaYMagnitude = pointData.deltaYMagnitude
      local directionFactor = (movingDown and -1.0 or 1.0)

      if (slowDown) then
        local newMagnitude = math.max(math.abs(deltaY * EASE_FACTORS.decay), minimumDeltaYMagnitude)
        pointData.deltaY = newMagnitude * directionFactor
      elseif (speedUp) then
        local newMagnitude = math.min(math.abs(deltaY * EASE_FACTORS.accelerate), maximumDeltaYMagnitude)
        pointData.deltaY = newMagnitude * directionFactor
      end

      -- start moving in opposite direction?
      local startMovingDown = (updatedY > pointData.maxY and movingUp)
      local startMovingUp = (updatedY < pointData.minY and movingDown)

      if (startMovingDown) then
        pointData.direction = MOVING_DIRECTION.down
        pointData.deltaY = -1.0 * minimumDeltaYMagnitude
      elseif (startMovingUp) then
        pointData.direction = MOVING_DIRECTION.up
        pointData.deltaY = minimumDeltaYMagnitude
      end
    end
  end
end

local function startLoopingAction(self)
  local duration = 0.615
  local frameRate = 60.0
  for _, pointData in ipairs(self.points) do
    if (pointData.direction ~= MOVING_DIRECTION.none) then
      local distance = math.abs(pointData.maxY - pointData.minY)
      local movingDown = (pointData.direction == MOVING_DIRECTION.down)
      local deltaY = distance / duration / frameRate
      pointData.deltaYMagnitude = deltaY
      pointData.minimumDeltaYMagnitude = deltaY * 0.05
      pointData.deltaY = pointData.minimumDeltaYMagnitude * (movingDown and -1.0 or 1.0)
    end
  end

  local loopingAction = AHLoop.new(function()
    updateGraph(self)
  end)

  loopingAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  -- bottom rectangle
  local rectangle = FilledRectangle.new(DIMENSIONS.width, DIMENSIONS.rectangleHeight, HUE.bar)
  self:insertAtCenter(rectangle)

  local rectangleY = DIMENSIONS.rectangleHeight * 0.5
  rectangle:setY(rectangleY)

  -- points and line
  local graphHeight = DIMENSIONS.graphHeight
  self.points = {
    {
      y = 0.15 * graphHeight,
      direction = MOVING_DIRECTION.none
    },
    {
      y = 0.0 * graphHeight,
      minY = 0.0 * graphHeight,
      maxY = 0.5 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.525 * graphHeight,
      minY = 0.4 * graphHeight,
      maxY = 0.525 * graphHeight,
      direction = MOVING_DIRECTION.down
    },
    {
      y = 0.2 * graphHeight,
      minY = 0.2 * graphHeight,
      maxY = 0.5 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.1 * graphHeight,
      minY = 0.1 * graphHeight,
      maxY = 0.45 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.456 * graphHeight,
      minY = 0.465 * graphHeight,
      maxY = 0.7 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.6 * graphHeight,
      minY = 0.3 * graphHeight,
      maxY = 0.6 * graphHeight,
      direction = MOVING_DIRECTION.down
    },
    {
      y = 0.25 * graphHeight,
      minY = 0.25 * graphHeight,
      maxY = 0.45 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.2 * graphHeight,
      minY = 0.2 * graphHeight,
      maxY = 0.6 * graphHeight,
      direction = MOVING_DIRECTION.up
    },
    {
      y = 0.7 * graphHeight,
      direction = MOVING_DIRECTION.none
    }
  }

  local radius = Screen.getWidthPart(0.006)
  local pointCount = #self.points
  local deltaX = (DIMENSIONS.width * 0.92) / (pointCount - 1)
  local x = 0.0
  local coordinates = {}
  self.circles = {}

  for i, pointData in ipairs(self.points) do
    pointData.x = x
    x = x + deltaX

    local coordinate = {pointData.x, pointData.y}
    table.insert(coordinates, coordinate)

    local circle = FilledCircle.new(radius, HUE.circle)
    circle:setLoc(pointData.x, pointData.y)
    table.insert(self.circles, circle)

    -- hide first and last circles
    if (i == 1 or i == #self.points) then
      circle:setVisible(false)
    end
  end

  self.line = Line.new(coordinates)
  self.line:setColor(unpack(HUE.line))
  self:insertAtCenter(self.line)

  local lineY = DIMENSIONS.rectangleHeight + DIMENSIONS.verticalSpacing + DIMENSIONS.graphHeight * 0.35
  self.line:setY(lineY)

  local lineWidth = Screen.getWidthPart(0.0075)
  self.line:setLineWidth(lineWidth)

  for i, circle in ipairs(self.circles) do
    self.line:insert(circle)
  end

  local isRecursive = true
  self:linkColorAttrToChildren(isRecursive)
end

-- Constructor ------------------------

function OnboardioGraphAnimationWriting.new()
  local width = DIMENSIONS.width
  local height = DIMENSIONS.graphHeight + DIMENSIONS.verticalSpacing + DIMENSIONS.rectangleHeight

  local self = Group.new(width, height)
  Mixin.mix(self, OnboardioGraphAnimationWriting)

  createLayout(self)
  startLoopingAction(self)

  return self
end
