--- OnboardioPanelRound
-- @author Created by andrewcousins on March 11, 2024.

class("OnboardioPanelRound")
require("OnboardioCommon")
require("OnboardioPanel")
require("OnboardioButton")
require("OnboardioTextFactory")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local I_DONT_KNOW_BUTTON_TEXT_KEY = "I_DONT_KNOW_TEXT"

local Z_BUTTON = 20

local BOX_CENTER_OFFSET = 37.5

-- Note: Both these colors have an opacity of 0
local CORRECT_OVERLAY_TOP_COLOR = ColorUtils.colorHex32ToStd(0x07063800)
local CORRECT_OVERLAY_BOTTOM_COLOR = ColorUtils.colorHex32ToStd(0x5EFF7800)
local CORRECT_OVERLAY_BOTTOM_OPACITY = 0.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setButtonsTouchEnabled(self, enabled)
  self.iDontKnowButton:setTouchEnabled(enabled)
  self.buttonManager:setButtonsEnabled(enabled)
end


local function getPromptOpacityAnimation(self, duration, ease, opacity)
  local promptDelay = 0
  local animation = AHSpawn.new()
  if self.headerPromptLabel then
    animation:addAction(AHOpacityTo.new(self.headerPromptLabel, duration, ease, opacity))
    promptDelay = 0.05
  end
  animation:addAction(AHDelayed.new(promptDelay, AHSpawn.new(AHOpacityTo.new(self.promptLabel, duration, ease, opacity))))

  return animation
end

local function setPromptOpacity(self, opacity)
  self.promptLabel:setOpacity(opacity)
  if self.headerPromptLabel then
    self.headerPromptLabel:setOpacity(opacity)
  end
end

local function getPromptScaleAnimation(self, duration, ease, scale)
  local promptDelay = 0
  local animation = AHSpawn.new()
  if self.headerPromptLabel then
    animation:addAction(AHScaleTo.new(self.headerPromptLabel, duration, ease, scale))
    promptDelay = 0.05
  end
  animation:addAction(AHDelayed.new(promptDelay, AHSpawn.new(AHScaleTo.new(self.promptLabel, duration, ease, scale))))

  return animation
end

local function setPromptScale(self, scale)
  self.promptLabel:setScl(scale)
  if self.headerPromptLabel then
    self.headerPromptLabel:setScl(scale)
  end
end

local function animateCorrectOverlay(self)
  local duration = 0.25
  local fadeIn = self.correctOverlay:newOpacityToAction(duration, nil, 0, CORRECT_OVERLAY_BOTTOM_OPACITY)
  local fadeOut = self.correctOverlay:newOpacityToAction(duration, nil, 0, 0)
  return AHSequence.new(fadeIn, fadeOut):start()
end

local function createOnPressedCallbackWithAnswerId(self, answerId)
  local function onPressedCallback()
    setButtonsTouchEnabled(self, false)
    self.selectedAnswer = answerId
    self:setDismissed()
  end
  return onPressedCallback
end

local function createOnCorrectAnimationCallback(self)
  return function()
    animateCorrectOverlay(self)
  end
end

local function setupAnswerButtons(self)
  local round = self.round
  local correctAnswerId = round:getCorrectAnswerId()
  local distractorAnswerIds = round:getDistractorAnswerIds()

  local answerIds = TableUtils.shallowCopy(distractorAnswerIds)
  table.insert(answerIds, correctAnswerId)
  TableUtils.shuffleArray(answerIds)

  local onCorrectAnimationCallback = createOnCorrectAnimationCallback(self)

  local buttonContext = {}
  for _, answerId in ipairs(answerIds) do
    local isCorrect = (answerId == correctAnswerId)
    local onPressedCallback = createOnPressedCallbackWithAnswerId(self, answerId)
    local buttonText = Localizer.get(answerId)
    local context = {
      text = buttonText,
      onPressedCallback = onPressedCallback,
      correctAnimationCallback = onCorrectAnimationCallback,
      isCorrect = isCorrect
    }
    table.insert(buttonContext, context)
  end
  self.buttonManager:setContext(buttonContext)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelRound:playAppear()
  self.iDontKnowButton:setOpacity(0)
  setPromptOpacity(self, 0)
  if self.isFirstQuestion then
    setPromptScale(self, 0.9)
    self.buttonManager:prepareForAppearAnimation()
  else
    setPromptScale(self, 0.8)
  end

  local iDontKnowButtonAnimation = AHOpacityTo.new(self.iDontKnowButton, 0.2, nil, 1)

  local promptLabelAnimation = AHSpawn.new(
    function() setupAnswerButtons(self) end,
    getPromptOpacityAnimation(self, 0.2, nil, 1),
    AHSequence.new(
      getPromptScaleAnimation(self, 0.2, AHEaseBezier.newFlat({0.6, 0.0, 0.4, 1.0}), 1.05),
      getPromptScaleAnimation(self, 0.2, AHEaseBezier.newFlat({0.2, 0.0, 0.5, 1.0}), 1.0)
    ),
    AHDelayed.new(0.25, function() setButtonsTouchEnabled(self, true) end)
  )

  local buttonDelay = 0
  if self.headerPromptLabel then
    buttonDelay = 0.05
  end
  local buttonAppearanceAnimation = AHDelayed.new(buttonDelay, self.buttonManager:getAppearAnimation())

  local animation = AHSpawn.new(
    buttonAppearanceAnimation,
    iDontKnowButtonAnimation,
    promptLabelAnimation
  )

  local transitionDelay = 0.25
  if self.isFirstQuestion then
    transitionDelay = 0
  end
  return AHDelayed.new(transitionDelay, animation):start()
end

function OnboardioPanelRound:playDisappear()
  local neutralDelay = 0.0 -- Delay when not showing correct/incorrect
  local scaleAnimation = getPromptScaleAnimation(self, 0.2, AHEaseBezier.newFlat({0.5, 0.0, 0.8, 1.0}), 0.8)
  local fadeAnimation = getPromptOpacityAnimation(self, 0.2, nil, 0)
  local buttonAnimations = self.buttonManager:getRoundOverAnimation()
  return AHDelayed.new(neutralDelay, AHSpawn.new(scaleAnimation, fadeAnimation, buttonAnimations)):start()
end

function OnboardioPanelRound:handleCountdownExpired()
  setButtonsTouchEnabled(self, false)
  self.buttonManager:revealCorrectAnswer()
end

function OnboardioPanelRound:setButtonsEnabled()
  -- Do nothing
  -- This method is called by onboardio transitioner, but we want finer control
  -- of when buttons are touch active
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function positionAnswerButtons(self)
  if (self.buttonType == OnboardioContent.BUTTON_SQUARE) then
    self.buttonManager:positionButtonsForMathLayout(self.promptColumn:getYBottom(), self.boxCenterX)
  else
    local buttonsCenterY = ((self.promptColumn:getYBottom() + self.iDontKnowButton:getYTop()) / 2)
    if self.isFirstQuestion then
      self.buttonManager:positionButtonsWithNoAnimation(buttonsCenterY)
    else
      self.buttonManager:positionButtons(buttonsCenterY)
    end
  end
  if (self.isFirstQuestionOfSection and (not self.isFirstQuestion)) then
    Sound.playSoundEffect("bubble_move_down_whoosh")
  end
end

local function getPromptSpacing(self)
  if self.buttonType == OnboardioContent.BUTTON_SQUARE then
    return self:getHeightPart(0.025)
  else
    return self:getHeightPart(0.085)
  end
end

local function buildPromptColumnStandard(self, progressIndicator)
  local promptSpacing = getPromptSpacing(self)
  local promptColumn = Column.new(promptSpacing)
  local headerPromptLabel = self.headerPromptLabel
  if headerPromptLabel then
    promptColumn:insert(headerPromptLabel)
    promptColumn:insertSpacer(promptSpacing * -1.5)
  end

  local promptLabel = self.promptLabel
  promptColumn:insert(promptLabel)
  promptColumn:alignChildren()

  promptColumn:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.7))

  local yOffset = (promptColumn:getYTop() - progressIndicator:getYBottom()) + progressIndicator:getHeight()
  if (yOffset > 0) then
    promptColumn:offsetY(-yOffset)
  end

  return promptColumn
end

local function positionAndInsertIDontKnowButton(self)
  local iDontKnowButton = self.iDontKnowButton

  local iDontKnowButtonY = math.max(self:getHeightPart(0.05), Screen.safeBottom + iDontKnowButton:getHeightPart(0.5))

  iDontKnowButton:setLoc(self:getWidthPart(0.5), iDontKnowButtonY)
  self:insert(iDontKnowButton, Z_BUTTON)
end

local function buildLayout(self, progressIndicator)
  local promptColumn = buildPromptColumnStandard(self, progressIndicator)
  self:insert(promptColumn)
  self.promptColumn = promptColumn

  positionAndInsertIDontKnowButton(self)
  positionAnswerButtons(self)
end

local function buildPromptLabel(self, round)
  local prompt = round:getPrompt()
  local promptLabel = OnboardioTextFactory.newFancyTextView(prompt)
  self.promptLabel = promptLabel
end

local function buildHeaderLabelIfNeeded(self, round)
  local headerPrompt = round:getHeaderPrompt()
  if headerPrompt == nil then
    return
  end

  local headerPromptLabel = OnboardioTextFactory.newFancyTextView(headerPrompt)
  self.headerPromptLabel = headerPromptLabel
end

local function checkPromptForBoxView(self)
  local promptLabel = self.promptLabel
  local box = promptLabel:getBoxView()
  if self.buttonType == OnboardioContent.BUTTON_SQUARE then
    local boxX = self:getWidthPart(0.5) - promptLabel:getWidthPart(0.5) + box:getX()
    local boxCenterX = boxX + BOX_CENTER_OFFSET
    self.boxCenterX = boxCenterX
  end
end

local function buildIDontKnowButton(self)
  local onIDontKnowCallback = createOnPressedCallbackWithAnswerId(self, "I_DONT_KNOW_ANSWER")
  local callbackWithSound = function()
    self.countdown:stop()
    Sound.playSoundEffect("button_press")
    self.buttonManager:revealCorrectAnswer()
    onIDontKnowCallback()
  end
  local buttonText = Localizer.get(I_DONT_KNOW_BUTTON_TEXT_KEY)
  local buttonColor = OnboardioCommon.COLOR_WHITE
  local iDontKnowButton = OnboardioButton.newInverted(buttonText, callbackWithSound, nil, buttonColor)
  iDontKnowButton:setTouchEnabled(false)
  self.iDontKnowButton = iDontKnowButton

  table.insert(self.buttons, iDontKnowButton)
end

local function buildCorrectOverlay(self)
  local correctOverlay = RectangleGradient.new(self:getWidth(), self:getHeightPart(0.25), CORRECT_OVERLAY_TOP_COLOR, CORRECT_OVERLAY_BOTTOM_COLOR)
  correctOverlay:setAnchorPointBottom()
  self:insertAtBottom(correctOverlay)
  self.correctOverlay = correctOverlay
end

-- Constructor ------------------------

function OnboardioPanelRound.new(round, buttonManager, countdown, progressIndicator, isFirstQuestion, isFirstQuestionOfSection)
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioPanelRound)

  self.round = round
  self.buttonType = round.buttonType
  self.buttonManager = buttonManager
  self.countdown = countdown
  self.isFirstQuestion = isFirstQuestion
  self.isFirstQuestionOfSection = isFirstQuestionOfSection

  -- Note: we don't add anything to OnboardioPanel's fadableLabels or stationaryButtons tables
  -- to avoid OnboardioTransitioner playing undesired animations

  buildPromptLabel(self, round)
  buildHeaderLabelIfNeeded(self, round)
  checkPromptForBoxView(self)
  buildIDontKnowButton(self)
  buildCorrectOverlay(self)
  buildLayout(self, progressIndicator)

  setButtonsTouchEnabled(self, false)

  return self
end
