--- OnboardioPanelSelfRate
-- @author Created by andrewcousins on March 21, 2024.

class("OnboardioPanelSelfRate")
require("OnboardioPanel")
require("OnboardioTextFactory")
require("OnboardioPretestSelfRateButton")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_LAYOUT_COLUMN = 10

local RATING_LEVELS = {
  {amount = 1.0, name = Localizer.get("EXCELLENT")},
  {amount = 0.67, name = Localizer.get("ABOVE_AVERAGE")},
  {amount = 0.33, name = Localizer.get("AVERAGE")},
  {amount = 0.00, name = Localizer.get("BELOW_AVERAGE")}
}

local ANIMATION_TIMING = 0.2
local STAGGERED_APPEARANCE_DELAY = 0.05
local SCALE_OUT_AMOUNT = 0.8
local OVERSCALE_AMOUNT = 1.05

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getPromptAnimateInAnimation(self)
  self.promptLabel:setOpacity(0)
  return AHDelayed.new(self:getAppearanceDelayDuration(), AHSpawn.new(
    AHOpacityTo.new(self.promptLabel, ANIMATION_TIMING, nil, 1),
    AHSequence.new(
      AHScaleTo.new(self.promptLabel, ANIMATION_TIMING, AHEaseBezier.newFlat({0.6, 0.0, 0.4, 1.0}), OVERSCALE_AMOUNT),
      AHScaleTo.new(self.promptLabel, ANIMATION_TIMING, AHEaseBezier.newFlat({0.2, 0.0, 0.5, 1.0}), 1.0)
    )
  ))
end

local function getPromptAnimateOutAnimation(self)
  return AHSpawn.new(
    AHOpacityTo.new(self.promptLabel, ANIMATION_TIMING, nil, 0),
    AHScaleTo.new(self.promptLabel, ANIMATION_TIMING, AHEaseBezier.newFlat({0.5, 0.0, 0.8, 1.0}), SCALE_OUT_AMOUNT)
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPanelSelfRate:playAppear()
  local animation = AHSpawn.new(getPromptAnimateInAnimation(self))
  local soundDelay = ANIMATION_TIMING
  animation:addAction(AHDelayed.new(soundDelay, function()
    Sound.playSoundEffect("buttons_in")
  end))
  for i, button in ipairs(self.selfRateButtons) do
    local delay = self:getAppearanceDelayDuration() + (i * STAGGERED_APPEARANCE_DELAY)
    animation:addAction(AHDelayed.new(delay, button:getAppearAnimation()))
  end
  return animation:start()
end

function OnboardioPanelSelfRate:playDisappear()
  local animation = AHSpawn.new(getPromptAnimateOutAnimation(self))
  for i, button in ipairs(self.selfRateButtons) do
    local delay = i * STAGGERED_APPEARANCE_DELAY
    animation:addAction(AHDelayed.new(delay, button:getDisappearAnimation(self.isLastSelfRateRound)))
  end
  return animation:start()
end

function OnboardioPanelSelfRate:getAppearanceDelayDuration()
  return 0.15
end

function OnboardioPanelSelfRate:setButtonsEnabled(enabled)
  for _, button in ipairs(self.selfRateButtons) do
    button:setTouchEnabled(enabled)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function positionAnswerButtons(self)
  local absoluteX = self:getWidthPart(0.5)

  local layoutColumn = self.layoutColumn
  local layoutColumnHeightPart = layoutColumn:getHeightPart(0.5)
  local layoutColumnY = layoutColumn:getY()

  local answerButtonColumn = self.answerButtonColumn
  local answerButtonColumnHeightPart = answerButtonColumn:getHeightPart(0.5)
  local answerButtonColumnY = answerButtonColumn:getY()

  local answerButtonColumnAbsoluteY = layoutColumnY - layoutColumnHeightPart + answerButtonColumnY - answerButtonColumnHeightPart

  local answerButtons = self.slidingButtons
  for _, answerButton in ipairs(answerButtons) do
    local answerButtonY = answerButton:getY()

    local absoluteY = answerButtonColumnAbsoluteY + answerButtonY

    self:takeChild(answerButton)

    answerButton:setLoc(absoluteX, absoluteY)
  end

end

local function buildLayout(self)
  local promptLabel = self.promptLabel
  local answerButtonColumn = self.answerButtonColumn

  local spacing = self:getHeightPart(0.04)
  local layoutColumn = Column.new(spacing)

  layoutColumn:insert(promptLabel)
  layoutColumn:insertSpacer(spacing)
  layoutColumn:insert(answerButtonColumn)
  layoutColumn:alignChildren()
  self.layoutColumn = layoutColumn

  self:insertAtCenter(layoutColumn, Z_LAYOUT_COLUMN)

  positionAnswerButtons(self)
end

local function buildPromptLabel(self, round)
  local text = round:getPrompt()
  local width = OnboardioCommon.TEXT_WIDTH
  local font = Font.new(FontFaces.DIN_OT_REGULAR, 24)
  local lineSpacing = FontScaling.getScaledSize(7.0)
  textView = DynamicTextArea.newSimple(text, width, font, Alignment.CENTER, Alignment.CENTER, lineSpacing)

  self.promptLabel = textView
end

local function buildAnswerButtons(self, isFirstSelfRateRound)
  local spacing = self:getHeightPart(0.025)
  local alignment = Alignment.CENTER
  local growDirection = Column.GrowDown
  local column = Column.new(spacing, alignment, growDirection)

  for _, ratingLevel in ipairs(RATING_LEVELS) do
    local name = ratingLevel.name
    local function onPressed()
      self:setButtonsEnabled(false)
      self.selectedAnswer = ratingLevel
      self:setDismissed()
    end

    local button = OnboardioPretestSelfRateButton.new(name, onPressed, isFirstSelfRateRound)
    table.insert(self.selfRateButtons, button)
    column:insert(button)
  end

  column:alignChildren()

  self.answerButtonColumn = column
end

-- Constructor ------------------------

function OnboardioPanelSelfRate.new(round, isFirstSelfRateRound, isLastSelfRateRound)
  local self = OnboardioPanel.new()
  Mixin.mix(self, OnboardioPanelSelfRate)

  self.selfRateButtons = {}
  self.isLastSelfRateRound = isLastSelfRateRound

  buildPromptLabel(self, round)
  buildAnswerButtons(self, isFirstSelfRateRound)
  buildLayout(self)

  return self
end
