--- OnboardioPretestCircleButtonManager
-- @author Created by andrewcousins on March 13, 2024.

class("OnboardioPretestCircleButtonManager")

require("OnboardioPretestCircleButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local POSITION_ANIMATION_DURATION = 0.6
local POSITION_ANIMATION_EASE = AHEaseBezier.newFlat({0.15, 0.00, 0.30, 1.00})

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function animateButtonPositions(self, centerYPosition, duration)
  self.leftButton:setBaseScale(1.0)
  self.rightButton:setBaseScale(1.0)
  local scaleAnimations = AHSpawn.new(
    self.leftButton:getScaleToBaseAnimation(duration, POSITION_ANIMATION_EASE),
    self.rightButton:getScaleToBaseAnimation(duration, POSITION_ANIMATION_EASE)
  )

  local horizontalOffset = self.leftButton:getEffectiveWidthPart(0.5)
  local verticalOffset = self.leftButton:getEffectiveHeightPart(0.2)
  local leftButtonX = self:getWidthPart(0.5) - horizontalOffset
  local leftButtonY = centerYPosition + verticalOffset
  local rightButtonX = self:getWidthPart(0.5) + horizontalOffset
  local rightButtonY = centerYPosition - verticalOffset

  local threeButtonLayoutVerticalOffset = self.leftButton:getEffectiveHeightPart(0.6)
  local thirdButtonX = leftButtonX + self.leftButton:getEffectiveWidthPart(0.1)
  local thirdButtonY = centerYPosition - threeButtonLayoutVerticalOffset
  if (self.nextButtonCount == 3) then
    leftButtonY = centerYPosition + threeButtonLayoutVerticalOffset
    rightButtonY = centerYPosition + self.leftButton:getEffectiveHeightPart(0.05)
  end

  local moveAnimations = AHSpawn.new(
    self.leftButton:getMoveToNewHome(leftButtonX, leftButtonY, duration, POSITION_ANIMATION_EASE),
    AHDelayed.new(0.05, self.rightButton:getMoveToNewHome(rightButtonX, rightButtonY, duration, POSITION_ANIMATION_EASE))
  )

  if (self.nextButtonCount == 3) then
    scaleAnimations:addAction(self.thirdButton:getScaleToBaseAnimation(duration, POSITION_ANIMATION_EASE))
    moveAnimations:addAction(AHDelayed.new(0.1, self.thirdButton:getMoveToNewHome(thirdButtonX, thirdButtonY, duration, POSITION_ANIMATION_EASE)))
  end

  return AHSpawn.new(scaleAnimations, moveAnimations):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioPretestCircleButtonManager:positionButtons(centerYPosition)
  return animateButtonPositions(self, centerYPosition, POSITION_ANIMATION_DURATION)
end

function OnboardioPretestCircleButtonManager:positionButtonsWithNoAnimation(centerYPosition)
  return animateButtonPositions(self, centerYPosition, 0)
end

function OnboardioPretestCircleButtonManager:positionButtonsForMathLayout(promptBottomY, promptPositionX)
  self.leftButton:setBaseScale(0.7)
  self.rightButton:setBaseScale(0.7)
  self.thirdButton:setBaseScale(0.7)
  local scaleAnimations = AHSpawn.new(
    self.leftButton:getScaleToBaseAnimation(POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE),
    self.rightButton:getScaleToBaseAnimation(POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE),
    self.thirdButton:getScaleToBaseAnimation(POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE)
  )

  local horizontalOffset = self.leftButton:getEffectiveWidthPart(0.02)
  local verticalOffsetToCenter = self.leftButton:getEffectiveHeightPart(1.1)
  local verticalOffset = self.leftButton:getEffectiveHeightPart(0.55)
  local leftButtonX = promptPositionX - horizontalOffset
  local leftButtonY = promptBottomY - verticalOffsetToCenter + verticalOffset
  local rightButtonX = promptPositionX + horizontalOffset
  local rightButtonY = promptBottomY - verticalOffsetToCenter - verticalOffset

  -- Third button moves with the other two, even though it is animating off screen
  local thirdButtonX = promptPositionX + horizontalOffset
  local thirdButtonY = promptBottomY - verticalOffsetToCenter - (verticalOffset * 2.5)

  local moveAnimations = AHSpawn.new(
    self.leftButton:getMoveToNewHome(leftButtonX, leftButtonY, POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE),
    AHDelayed.new(0.05, self.rightButton:getMoveToNewHome(rightButtonX, rightButtonY, POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE)),
    AHDelayed.new(0.1, self.thirdButton:getMoveToNewHome(thirdButtonX, thirdButtonY, POSITION_ANIMATION_DURATION, POSITION_ANIMATION_EASE))
  )

  return AHSpawn.new(scaleAnimations, moveAnimations):start()
end

function OnboardioPretestCircleButtonManager:setContext(buttonContext)
  for i, button in ipairs(self.buttons) do
    if (i <= #buttonContext) then
      local context = buttonContext[i]
      button:setAnswer(context.text, context.isCorrect)
      button:setOnPressedCallback(context.onPressedCallback)
      button:setCorrectAnimationCallback(context.correctAnimationCallback)
    end
  end
end

function OnboardioPretestCircleButtonManager:prepareForAppearAnimation()
  for i = 1, self.nextButtonCount do
    local button = self.buttons[i]
    button:prepareForAppearAnimation()
  end
end

function OnboardioPretestCircleButtonManager:getAppearAnimation()
  local appearAnimation = AHSpawn.new()
  for i = 1, self.nextButtonCount do
    local button = self.buttons[i]
    local delay = 0.05 * i
    appearAnimation:addAction(AHDelayed.new(delay, button:getAppearAnimation()))
  end
  return appearAnimation
end

function OnboardioPretestCircleButtonManager:getRoundOverAnimation()
  local roundOverAnimation = AHSpawn.new()
  for i = 1, self.currentButtonCount do
    local button = self.buttons[i]
    local delay = 0.05 * (i - 1)
    if (i > self.nextButtonCount) then -- Animate the third button out
      roundOverAnimation:addAction(AHDelayed.new(delay, button:getRoundOverAndHideAnimation()))
    else
      roundOverAnimation:addAction(AHDelayed.new(delay, button:getRoundOverAnimation(self.willFadeOut)))
    end
  end
  if (self.currentButtonCount < self.nextButtonCount) then -- Prep third button to be shown
    self.thirdButton:prepareForAppearAnimation(0.8)
  end
  self.currentButtonCount = self.nextButtonCount
  return roundOverAnimation
end

function OnboardioPretestCircleButtonManager:hide()
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(false)
    button:hide()
  end
end

function OnboardioPretestCircleButtonManager:fadeOutInNextTransition()
  self.willFadeOut = true
end

function OnboardioPretestCircleButtonManager:setButtonsEnabled(enabled)
  for i = 1, self.currentButtonCount do
    local button = self.buttons[i]
    button:setTouchEnabled(enabled)
  end
end

function OnboardioPretestCircleButtonManager:revealCorrectAnswer()
  for i = 1, self.currentButtonCount do
    local button = self.buttons[i]
    if button:getIsCorrect() then
      wait(button:getRevealAnswerAnimation():start())
    end
  end
end

function OnboardioPretestCircleButtonManager:setNextSectionButtonCount(answerOptions)
  if (not (answerOptions == 2 or answerOptions == 3)) then
    Log.error("Invalid number of answer options passed to Onboarding Button Manager, ignoring.")
    return
  end
  self.nextButtonCount = answerOptions
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioPretestCircleButtonManager.new(countdown)
  local self = Group.newFullScreen()
  Mixin.mix(self, OnboardioPretestCircleButtonManager)

  local leftButton = OnboardioPretestCircleButton.new(countdown)
  self:insertAtLocationFractional(leftButton, 0.35, 0.6)
  self.leftButton = leftButton

  local rightButton = OnboardioPretestCircleButton.new(countdown)
  self:insertAtLocationFractional(rightButton, 0.65, 0.4)
  self.rightButton = rightButton

  local thirdButton = OnboardioPretestCircleButton.new(countdown)
  self:insertAtLocationFractional(thirdButton, 0.35, 0.2)
  thirdButton:hide()
  thirdButton:setTouchEnabled(false)
  self.thirdButton = thirdButton

  self.buttons = {self.leftButton, self.rightButton, self.thirdButton}

  self.currentButtonCount = 2
  self.nextButtonCount = 2

  self.willFadeOut = false

  return self
end
