--- OnboardioProgressIndicator
-- @author Created by Aaron Barrett on October 14, 2014.

class("OnboardioProgressIndicator")

require("OnboardioCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.getWidthPart(0.9)
local HEIGHT = Screen.getHeightPart(0.01125)


local FOREGROUND_COLOR = OnboardioCommon.COLOR_WHITE
local BACKGROUND_COLOR = ColorUtils.colorHex24ToStd(0x3A3958)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function stopFillAction(self)
  local fillAction = self.fillAction
  if (fillAction ~= nil) then
    fillAction:stop()
  end
end

local function getBarAnimation(self, targetScale)
  local duration = 0.5
  local ease = AHEaseSimpleInOut.new(2.0)
  local scaleAnimation = AHScaleXTo.new(self.progressBar, duration, ease, targetScale)
  local leftCap = self.progressBar.leftCap
  local rightCap = self.progressBar.rightCap
  local targetX = self.progressBar:getWidthPart(targetScale) + self:getHeight()
  local moveCapAnimation = AHLocationXTo.new(rightCap, duration, ease, targetX)

  local barAnimation = AHSpawn.new(scaleAnimation, moveCapAnimation)

  if (targetScale > 0) and (not leftCap:getVisible()) then
    leftCap:setVisible(true)
  elseif (targetScale <= 0) and (leftCap:getVisible()) then
    barAnimation:addAction(AHDelayed.new(duration, function()
      leftCap:setVisible(true)
    end))
  end

  if (targetScale > 0) and (not rightCap:getVisible()) then
    rightCap:setVisible(true)
  elseif (targetScale <= 0) and (rightCap:getVisible()) then
    rightCap:setVisible(false)
  end

  return barAnimation
end

local function createFadeAnimation(self, targetOpacity, duration)
  local fadeFill = AHOpacityTo.new(self.progressBar, duration, nil, targetOpacity)
  local fadeFillBackground = AHOpacityTo.new(self.backing, duration, nil, targetOpacity)
  local fadeLeftCap = AHOpacityTo.new(self.progressBar.leftCap, duration, nil, targetOpacity)
  local fadeRightCap = AHOpacityTo.new(self.progressBar.rightCap, duration, nil, targetOpacity)

  return AHSpawn.new(fadeFill, fadeFillBackground, fadeLeftCap, fadeRightCap)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioProgressIndicator:setRemainingDivisionCount(remainingDivisionCount)
  self.destinationProgress = self.currentProgress + remainingDivisionCount
end

function OnboardioProgressIndicator:fillDivisions(divisionCount)
  stopFillAction(self)

  self.currentProgress = self.currentProgress + divisionCount
  local fillAmount = (self.currentProgress / self.destinationProgress)

  local fillAction = getBarAnimation(self, fillAmount)
  self.fillAction = fillAction
  return fillAction:start()
end

function OnboardioProgressIndicator:fadeIn()
  local duration = 0.2
  local targetOpacity = 1.0
  return createFadeAnimation(self, targetOpacity, duration):start()
end

function OnboardioProgressIndicator:fadeOut()
  local duration = 0.3
  local targetOpacity = 0.0
  return createFadeAnimation(self, targetOpacity, duration):start()
end

function OnboardioProgressIndicator:hide()
  self.backing:setOpacity(0)
  self.progressBar:setOpacity(0)
  self.progressBar.leftCap:setOpacity(0)
  self.progressBar.rightCap:setOpacity(0)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioProgressIndicator.new()
  local width = WIDTH
  local height = HEIGHT
  local backgroundColor = BACKGROUND_COLOR
  local foregroundColor = FOREGROUND_COLOR

  local cornerRadius = height / 2

  local self = Group.new(width, height)
  Mixin.mix(self, OnboardioProgressIndicator)
  local backgroundBar = FilledRoundedRectangle.new(width, height, cornerRadius, backgroundColor)
  self:insertAtCenter(backgroundBar)

  local foregroundBar = FilledRectangle.new(width - (cornerRadius * 2), height, foregroundColor)
  foregroundBar:setAnchorPointLeft()
  foregroundBar:setScl(0.0, 1.0)
  self:insertAtLeft(foregroundBar)
  foregroundBar:offsetX(cornerRadius)

  local leftCap = FilledArc.new(cornerRadius, math.pi / 2, math.pi * 1.5, foregroundColor)
  leftCap:setAnchorPointLeft()
  leftCap:setVisible(false)
  self:insertAtLeft(leftCap)

  local rightCap = FilledArc.new(cornerRadius, -math.pi / 2, math.pi / 2, foregroundColor)
  rightCap:setAnchorPointRight()
  rightCap:setVisible(false)
  self:insertAtLeft(rightCap)
  rightCap:offsetX(cornerRadius*2)

  self.backing = backgroundBar
  self.progressBar = foregroundBar
  self.progressBar.leftCap = leftCap
  self.progressBar.rightCap = rightCap

  self.currentProgress = 0.0
  self.destinationProgress = 0.0
  self.fillAction = nil

  return self
end
