--- ONboardioCircleGradient
-- @author Created by Steven on April 12, 2017

class("OnboardioCircleGradient")

require("utils.ShaderUtils")

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

local COLORS = {
  "topLeftColor",
  "topRightColor",
  "bottomRightColor",
  "bottomLeftColor"
}

local DEFAULT_CIRCLE_POINT_COUNT = 60

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

--- Same functionality as ShapeUtils.makeArcPoints but incorporates uv inverting
-- for OpenGl coordinates into the loop for optimization
local function makeArcPoints(radius)
  local points = {}
  local diameter = radius * 2
  local angleRange =  2 * math.pi
  local sideCount = DEFAULT_CIRCLE_POINT_COUNT - 1

  for i = 0, sideCount do
    local vertex = {}
    local angle = i * angleRange / sideCount
    local xComponent, yComponent = math.cos(angle), math.sin(angle)
    vertex.x, vertex.y = radius + radius * xComponent, radius + radius * yComponent
    vertex.u = vertex.x / diameter
    vertex.v = 1 - (vertex.y / diameter)
    table.insert(points, vertex)
  end

  return points
end

local function getShader(colors)
  local uniformIndices = {
    transform = 1,
    topLeftColor = 2,
    topRightColor = 3,
    bottomLeftColor = 4,
    bottomRightColor = 5
  }

  local uniformCount = TableUtils.getSize(uniformIndices)

  local shader = MOAIShader.new()
  shader:reserveUniforms(uniformCount)
  shader:declareUniform(uniformIndices.transform, 'transform', MOAIShader.UNIFORM_WORLD_VIEW_PROJ)
  shader:declareUniform(uniformIndices.topLeftColor, "topLeftColor", MOAIShader.UNIFORM_COLOR)
  shader:declareUniform(uniformIndices.topRightColor, "topRightColor", MOAIShader.UNIFORM_COLOR)
  shader:declareUniform(uniformIndices.bottomLeftColor, "bottomLeftColor", MOAIShader.UNIFORM_COLOR)
  shader:declareUniform(uniformIndices.bottomRightColor, "bottomRightColor", MOAIShader.UNIFORM_COLOR)

  shader:setVertexAttribute(1, "position")
  shader:setVertexAttribute(2, "uv")

  local vsh = ShaderUtils.loadShader("GradientShader.vsh")
  local fsh = ShaderUtils.loadShader("GradientShader.fsh")
  shader:load(vsh, fsh)

  for colorIndex, color in ipairs(colors) do
    local uniformName = COLORS[colorIndex]
    local moaiColor = MOAIColor.new()
    moaiColor:setColor(unpack(color))
    local index = uniformIndices[uniformName]
    shader:setAttrLink(index, moaiColor, MOAIColor.COLOR_TRAIT)
  end

  return shader
end

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Object creation
------------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioCircleGradient.new(radius, colors)
  local diameter = radius * 2
  local self = Group.new(diameter, diameter)
  Mixin.mix(self, OnboardioCircleGradient)

  local format = MOAIVertexFormat.new()
  format:declareCoord(1, MOAIVertexFormat.GL_FLOAT, 2)
  format:declareUV(2, MOAIVertexFormat.GL_FLOAT, 2)

  local points = makeArcPoints(radius)
  local vertexCount = #points
  local vertextBuffer = MOAIVertexBuffer.new()
  vertextBuffer:setFormat(format)
  vertextBuffer:reserveVerts(vertexCount)

  for _, point in ipairs(points) do
    vertextBuffer:writeFloat(point.x, point.y)
    vertextBuffer:writeFloat(point.u, point.v)
  end
  vertextBuffer:bless()

  local gradientShader = getShader(colors)

  local mesh = MOAIMesh.new()
  mesh:setVertexBuffer(vertextBuffer)
  mesh:setPrimType(MOAIMesh.GL_TRIANGLE_FAN)
  mesh:setShader(gradientShader)

  self:setDeck(mesh)

  return self
end
