--- OnboardioSkillAnimationCalculation
-- @author Created by Steven on April 5, 2017

class("OnboardioSkillAnimationCalculation")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = OnboardioCommon.COLOR_BACKGROUND

local DELAY_SYMBOL_1 = 0
local DELAY_SYMBOL_2 = 0.05
local DELAY_SYMBOL_3 = 0.1

local IMAGES = {
  {
    x = 0.15,
    y = 0.51,
    character = "2",
    imageStartTime = DELAY_SYMBOL_3,
    angle = 15,
    targetWidth = 0.12
  },
  {
    x = 0.36,
    y = 0.27,
    character = "%",
    imageStartTime = DELAY_SYMBOL_3,
    angle = -15,
    targetWidth = 0.19
  },
  {
    x = 0.39,
    y = 0.7,
    image = "calculate_gear.png",
    imageStartTime = DELAY_SYMBOL_1,
    targetWidth = 0.25
  },
  {
    x = 0.75,
    y = 0.35,
    character = "3",
    imageStartTime = DELAY_SYMBOL_1,
    angle = -15,
    targetWidth = 0.12
  },
  {
    x = 0.55,
    y = 0.28,
    image = "calculate_gear.png",
    imageStartTime = DELAY_SYMBOL_2,
    targetWidth = 0.15
  },
  {
    x = 0.62,
    y = 0.65,
    character = "$",
    imageStartTime = DELAY_SYMBOL_2,
    angle = -15,
    targetWidth = 0.13
  },
  {
    x = 0.8,
    y = 0.61,
    image = "calculate_gear.png",
    imageStartTime = DELAY_SYMBOL_2,
    targetWidth = 0.1
  }
}

local LITTLE_BUBBLES = {
  {
    x = 0.15,
    y = 0.09,
    image = "calculate_bubble_medium.png",
    targetWidth = 0.1
  },
  {
    x = 0.085,
    y = 0.02,
    image = "calculate_bubble_small.png",
    targetWidth = 0.04
  }
}

local FONT = Font.new(FontFaces.DIN_OT_BOLD, 48)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTargetScale(self, object, targetWidthProportion)
  local targetWidth = self:getWidthPart(targetWidthProportion)
  local targetScale = object:scaleForEffectiveDims(targetWidth)
  return targetScale
end

local function addSmallBubbles(self)
  local bubbles = {}
  for _, descriptor in ipairs(LITTLE_BUBBLES) do
    local bubble = Sprite.newWithName(descriptor.image)
    bubble.scaleAmount = getTargetScale(self, bubble, descriptor.targetWidth)
    self:insertAtLocationFractional(bubble, descriptor.x, descriptor.y)
    table.insert(bubbles, bubble)
  end

  self.littleBubbles = bubbles
end

local function addImages(self)
  local images = {}
  for _, descriptor in ipairs(IMAGES) do
    local image
    if (descriptor.character) then
      image = Label.new(descriptor.character, FONT)
      image:setRot(descriptor.angle)
    else
      image = Sprite.newWithName(descriptor.image)
    end
    image.scaleAmount = getTargetScale(self, image, descriptor.targetWidth)
    image.imageStartTime = descriptor.imageStartTime
    self:insertAtLocationFractional(image, descriptor.x, descriptor.y)
    table.insert(images, image)
  end

  self.images = images
end

local function createScaleBounce(object, fullDuration, maxFactor)
  local finalScale = object.scaleAmount
  local amplitude = maxFactor - 1.0
  local ease = AHEaseSineInOut.new()

  local fullAnimation = AHSequence.new(
    AHScaleTo.new(object, fullDuration * 0.25, AHEaseSineOut.new(), finalScale * (1.0 + amplitude)),
    AHScaleTo.new(object, fullDuration * 0.35, ease, finalScale)
  )

  return fullAnimation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioSkillAnimationCalculation:playFadeAction(duration)
  return AHSequence.new(
    AHOpacityTo.new(self.overlay, duration, nil, 1.0),
    AHRemove.new(self)
  ):start()
end

function OnboardioSkillAnimationCalculation:reset()
  self.littleBubbles[1]:setScl(0.0)
  self.littleBubbles[2]:setScl(0.0)
  self.largeBubble:setScl(0.0)

  for _, image in ipairs(self.images) do
    image:setScl(0)
  end
end

function OnboardioSkillAnimationCalculation:getAnimationAction()
  local scaleDuration = 0.8
  local largeBubbleScale = createScaleBounce(self.largeBubble, scaleDuration, 1.1)

  local imagesScaleAction = AHSpawn.new()
  for _, image in ipairs(self.images) do
    local imageStartTime = image.imageStartTime
    imagesScaleAction:addAction(
      AHDelayed.new(
        imageStartTime,
        createScaleBounce(image, scaleDuration - imageStartTime, 1.15)
      )
    )
  end

  local ease = AHEaseExponentialOut.new()
  local smallBubblesAction = AHStaggeredSpawn.new(
    0.15,
    AHScaleTo.new(self.littleBubbles[1], scaleDuration, ease, self.littleBubbles[1].scaleAmount),
    AHScaleTo.new(self.littleBubbles[2], scaleDuration, ease, self.littleBubbles[2].scaleAmount)
  )

  local bigBubbleDelay = 0.1
  local smallBubbleDelay = bigBubbleDelay + 0.2
  return AHSpawn.new(
    AHDelayed.new(bigBubbleDelay, largeBubbleScale, imagesScaleAction),
    AHDelayed.new(smallBubbleDelay, smallBubblesAction)
  )
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioSkillAnimationCalculation.new(width)
  local largeBubble = Sprite.newWithName("calculate_bubble_large.png")
  local scaleAmount = largeBubble:scaleForEffectiveDims(width)
  local selfWidth = width
  local selfHeight = largeBubble:getHeightPart(scaleAmount)

  local self = FilledRectangle.new(selfWidth, selfHeight, BACKGROUND_COLOR)
  Mixin.mix(self, OnboardioSkillAnimationCalculation)

  addSmallBubbles(self)

  largeBubble.scaleAmount = scaleAmount
  self.largeBubble = largeBubble
  self:insertAtCenter(largeBubble)

  addImages(self)

  self.overlay = FilledRectangle.new(Screen.width, selfWidth, BACKGROUND_COLOR)
  self.overlay:setOpacity(0.0)
  self:insertAtCenter(self.overlay)

  self:reset()

  return self
end
