--- OnboardioSkillAnimationProcessing
-- @author Created by Joe Poka on April 5, 2017.

class("OnboardioSkillAnimationProcessing")

require("utils.ShaderUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  EXTRA_PAPER_PIECE = 10,
  PAPER_BOTTOM_SHADOW = 20,
  PAPER = 30,
  PAPER_COVER = 40,
  GEARS = 50,
  OVERLAY = 60
}

local COLORS = {
  WHITE = OnboardioCommon.COLOR_BACKGROUND,
  PAPER_COLOR = {0.89, 0.9, 0.94, 1.0},
  BACKGROUND_PAPER_COLOR = OnboardioCommon.COLOR_TABLE.GRAY100,
  PAPER_SHADOW_COLOR = {0.74, 0.75, 0.81, 1.0},
  PAPER_SHADOW_SIDE_COLOR = {0.37, 0.39, 0.49, 1.0},
  WORD_COLOR = {0.35, 0.37, 0.47, 1.0}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getAnimationPartOne(self, duration)
  local translatePaper = AHLocationYBy.new(self.paper, duration, nil, self.paperDeltaYPartOne)
  local animation = AHSpawn.new(translatePaper, self.bigGearAnimationAction, self.littleGearAnimationAction)
  return animation
end

local function getAnimationPartTwo(self, duration)
  local actions = {}

  local paperTranslateduration = duration * 0.85
  local paperEase = AHEaseElasticOut.new(0.8)
  local wordTranslationDuration = duration * 0.75
  local wordTranslationEase = AHEaseElasticOut.new(1.3)

  local translatePaper = AHLocationYBy.new(self.paper, paperTranslateduration, paperEase, self.paperDeltaYPartTwo)
  table.insert(actions, translatePaper)

  local wordTranslationDelay = 0.0
  local wordTranslationDelayDelta = 0.028

  for _, word in ipairs(self.words) do
    local fadeWord = AHOpacityTo.new(word, duration * 0.35, nil, 1.0)
    table.insert(actions, fadeWord)

    local wordDeltaY = word:getHeightPart(6.5)
    local translateWord = AHLocationYBy.new(word, wordTranslationDuration, wordTranslationEase, wordDeltaY)
    local delayedTranslateWord = AHDelayed.new(wordTranslationDelay, translateWord)
    table.insert(actions, delayedTranslateWord)

    wordTranslationDelay = wordTranslationDelay + wordTranslationDelayDelta
  end

  local fadeInPaperBottomShadow = AHDelayed.new(duration * 0.125, function()
    local fadeInDuration = duration * 0.1
    local fadeInShadowUpper = self.shadowUpper:newGradientToAction(fadeInDuration, nil, COLORS.PAPER_COLOR, COLORS.PAPER_SHADOW_COLOR)
    local fadeInShadowUpperSide = AHOpacityTo.new(self.shadowUpperSide, fadeInDuration, nil, 1.0)
    local fadeInShadowLower = self.shadowLower:newGradientToAction(fadeInDuration, nil, COLORS.PAPER_SHADOW_COLOR, COLORS.WHITE)
    local fadeInShadowLowerSide = self.shadowLowerSide:newGradientToAction(fadeInDuration, nil, COLORS.PAPER_SHADOW_SIDE_COLOR, COLORS.WHITE)
    fadeInShadowUpper:start()
    fadeInShadowUpperSide:start()
    fadeInShadowLower:start()
    fadeInShadowLowerSide:start()
  end)
  table.insert(actions, fadeInPaperBottomShadow)

  local fadeInLeftBackgroundPaper = self.backgroundPaperLeft:newGradientToAction(duration * 0.01, nil, COLORS.BACKGROUND_PAPER_COLOR, COLORS.WHITE)
  local delayedFadeInLeftBackgroundPaper = AHDelayed.new(duration * 0.1, fadeInLeftBackgroundPaper)
  local scaleUpLeftBackgroundPaper = AHScaleTo.new(self.backgroundPaperLeft, duration, paperEase, 1.0, 1.0)
  table.insert(actions, delayedFadeInLeftBackgroundPaper)
  table.insert(actions, scaleUpLeftBackgroundPaper)

  local fadeInRightBackgroundPaper = self.backgroundPaperRight:newGradientToAction(duration * 0.01, nil, COLORS.BACKGROUND_PAPER_COLOR, COLORS.WHITE)
  local delayedFadeInRightBackgroundPaper = AHDelayed.new(duration * 0.1, fadeInRightBackgroundPaper)
  local scaleUpRightBackgroundPaper = AHScaleTo.new(self.backgroundPaperRight, duration, paperEase, 1.0, 1.0)
  table.insert(actions, delayedFadeInRightBackgroundPaper)
  table.insert(actions, scaleUpRightBackgroundPaper)

  local showExtraPaperPiece = AHDelayed.new(duration * 0.15, function()
    self.extraPaper:setVisible(true)
    self.extraShadowSide:setVisible(true)
  end)
  table.insert(actions, showExtraPaperPiece)

  local animation = AHSpawn.new(actions)
  return animation
end

local function getProcessingWordsAndBackground(rowWidth, wordHeight, horizontalSpacing)
  local wordsLayout = {
    {
      rowWidth
    },
    {
      (rowWidth - horizontalSpacing) * 0.3,
      (rowWidth - horizontalSpacing) * 0.7
    },
    {
      rowWidth
    },
    {
      (rowWidth - horizontalSpacing) * 0.7,
      (rowWidth - horizontalSpacing) * 0.3
    }
  }

  local wordsRowCount = #wordsLayout
  local verticalSpacing = wordHeight
  local totalWordsHeight = wordsRowCount * wordHeight + (wordsRowCount - 1) * verticalSpacing

  local cornerRadius = wordHeight * 0.24
  local wordX = 0.0
  local wordY = totalWordsHeight - wordHeight * 0.5

  local words = {}
  local wordsGroup = Group.new(rowWidth, totalWordsHeight)

  for _, row in ipairs(wordsLayout) do
    for _, wordWidth in ipairs(row) do
      local word = FilledRoundedRectangle.new(wordWidth, wordHeight, cornerRadius, COLORS.WORD_COLOR)
      word:setLoc(wordX + wordWidth * 0.5, wordY)
      wordsGroup:insert(word)
      table.insert(words, word)
      wordX = wordX + wordWidth + horizontalSpacing
    end

    wordX = 0.0
    wordY = wordY - wordHeight - verticalSpacing
  end

  return words, wordsGroup
end

local function setUpPaperAndShadow(self)
  self.paper = FilledRectangle.new(self:getWidthPart(0.865), self:getHeightPart(0.6), COLORS.PAPER_COLOR)
  local x = self:getWidthPart(0.525)
  local y = -self:getHeightPart(0.125)
  self.paper:setLoc(x, y)
  self:insert(self.paper, Z_INDEX.PAPER)

  local paperShadowSide = FilledRectangle.new(self.paper:getWidthPart(0.0125), self.paper:getHeight(), COLORS.PAPER_SHADOW_SIDE_COLOR)
  x = self.paper:getWidth() + paperShadowSide:getWidthPart(0.5)
  y = self.paper:getHeightPart(0.5)
  paperShadowSide:setLoc(x, y)
  self.paper:insert(paperShadowSide)

  self.paperDeltaYPartOne = self:getHeightPart(0.1)
  self.paperDeltaYPartTwo = self:getHeightPart(1.1)

  self.shadowUpper = RectangleGradient.new(self.paper:getWidth(), self.paper:getHeightPart(0.25), COLORS.WHITE, COLORS.WHITE)
  x = self.paper:getX()
  y = self.paper:getY() + self.paperDeltaYPartTwo + self.paperDeltaYPartOne - self.paper:getHeightPart(0.5) - self.shadowUpper:getHeightPart(0.5)
  self.shadowUpper:setLoc(x, y)
  self:insert(self.shadowUpper, Z_INDEX.PAPER_BOTTOM_SHADOW)

  self.shadowUpperSide = FilledRectangle.new(paperShadowSide:getWidth(), self.shadowUpper:getHeight(), COLORS.PAPER_SHADOW_SIDE_COLOR)
  self.shadowUpperSide:setOpacity(0.0)
  x = self.shadowUpper:getX() + self.shadowUpper:getWidthPart(0.5) + self.shadowUpperSide:getWidthPart(0.5)
  self.shadowUpperSide:setLoc(x, y)
  self:insert(self.shadowUpperSide, Z_INDEX.PAPER_BOTTOM_SHADOW)

  self.shadowLower = RectangleGradient.new(self.paper:getWidth(), self.paper:getHeightPart(0.25), COLORS.WHITE, COLORS.WHITE)
  x = self.paper:getX()
  y = self.shadowUpper:getY() - self.shadowUpper:getHeightPart(0.5) - self.shadowLower:getHeightPart(0.5)
  self.shadowLower:setLoc(x, y)
  self:insert(self.shadowLower, Z_INDEX.PAPER_BOTTOM_SHADOW)

  self.shadowLowerSide = RectangleGradient.new(paperShadowSide:getWidth(), self.shadowLower:getHeight(), COLORS.WHITE, COLORS.WHITE)
  x = self.shadowLower:getX() + self.shadowLower:getWidthPart(0.5) + self.shadowLowerSide:getWidthPart(0.5)
  y = self.shadowLower:getY()
  self.shadowLowerSide:setLoc(x, y)
  self:insert(self.shadowLowerSide, Z_INDEX.PAPER_BOTTOM_SHADOW)

  self.extraPaper = FilledRectangle.new(self.paper:getWidth(), self.paper:getHeightPart(0.4), COLORS.PAPER_COLOR)
  self.extraPaper:setVisible(false)
  x = self.paper:getX()
  y = y + self.extraPaper:getHeight()
  self.extraPaper:setLoc(x, y)
  self:insert(self.extraPaper, Z_INDEX.EXTRA_PAPER_PIECE)

  self.extraShadowSide = FilledRectangle.new(paperShadowSide:getWidth(), self.paper:getHeightPart(0.4), COLORS.PAPER_SHADOW_SIDE_COLOR)
  self.extraShadowSide:setVisible(false)
  x = self.extraPaper:getX() + self.extraPaper:getWidthPart(0.5) + self.extraShadowSide:getWidthPart(0.5)
  self.extraShadowSide:setLoc(x, y)
  self:insert(self.extraShadowSide, Z_INDEX.EXTRA_PAPER_PIECE)

  local backgroundPaperWidth = self.paper:getWidthPart(0.185)
  self.backgroundPaperLeft = RectangleGradient.new(backgroundPaperWidth, self.paper:getHeightPart(0.7), COLORS.WHITE, COLORS.WHITE)
  self.backgroundPaperLeft:setAnchorPointBottomRight()
  self.backgroundPaperLeft:setScl(0.125, 0.25)
  x = self.paper:getX() - self.paper:getWidthPart(0.5)
  y = self.paper:getY() + self.paperDeltaYPartOne + self.paperDeltaYPartTwo - self.backgroundPaperLeft:getHeightPart(1.2)
  self.backgroundPaperLeft:setLoc(x, y)
  self:insert(self.backgroundPaperLeft, Z_INDEX.PAPER)

  self.backgroundPaperRight = RectangleGradient.new(backgroundPaperWidth, self.paper:getHeightPart(0.95), COLORS.WHITE, COLORS.WHITE)
  self.backgroundPaperRight:setAnchorPointBottomLeft()
  self.backgroundPaperRight:setScl(0.125, 0.25)
  x = self.paper:getX() + self.paper:getWidthPart(0.5) + paperShadowSide:getWidth()
  y = self.paper:getY() + self.paperDeltaYPartOne + self.paperDeltaYPartTwo - self.backgroundPaperRight:getHeightPart(0.85)
  self.backgroundPaperRight:setLoc(x, y)
  self:insert(self.backgroundPaperRight, Z_INDEX.PAPER)

  local p1x = 0.0
  local p1y = 0.0
  local p2x = backgroundPaperWidth
  local p2y = 0.0
  local p3x = 0.0
  local p3y = self.backgroundPaperLeft:getHeight()
  local backgroundPaperLeftTriangle = FilledTriangle.new(p1x, p1y, p2x, p2y, p3x, p3y, COLORS.WHITE)
  self.backgroundPaperLeft:insertAtCenter(backgroundPaperLeftTriangle)

  p3x = backgroundPaperWidth
  p3y = self.backgroundPaperRight:getHeight()
  local backgroundPaperRightTriangle = FilledTriangle.new(p1x, p1y, p2x, p2y, p3x, p3y, COLORS.WHITE)
  self.backgroundPaperRight:insertAtCenter(backgroundPaperRightTriangle)

  self.paperCover = FilledRectangle.new(self.paper:getWidthPart(1.1), self.paper:getHeightPart(2.5), COLORS.WHITE)
  x = self.paper:getX()
  y = -self:getHeightPart(0.3)
  self.paperCover:setLoc(x, y)
  self:insert(self.paperCover, Z_INDEX.PAPER_COVER)

  local rowWidth = self.paper:getWidthPart(0.725)
  local wordHeight = self.paper:getWidthPart(0.1)
  local horizontalSpacing = self.paper:getWidthPart(0.065)
  local words, wordsGroup = getProcessingWordsAndBackground(rowWidth, wordHeight, horizontalSpacing)
  x = self.paper:getWidthPart(0.5)
  y = self.paper:getHeightPart(0.6)
  wordsGroup:setLoc(x, y)
  self.paper:insert(wordsGroup)

  self.words = words

  local wordDeltaY = -wordHeight * 8.25
  for _, word in ipairs(self.words) do
    word:setOpacity(0.0)
    word:setY(word:getY() + wordDeltaY)
  end
end

local function swapInFinalFrame(sprite, finalFrameName)
  local deck, index = Sprite.deckAndIndexOfSpritesheetSprite(finalFrameName)
  sprite:setDeck(deck)
  sprite:setIndex(index)
end

local function setUpGears(self)
  local bigGearPrefix = "BigGear_"
  local littleGearPrefix = "LittleGear_"

  local firstFrameBigGear = Sprite.newWithName(bigGearPrefix .. "000.png")
  local firstFrameLittleGear = Sprite.newWithName(littleGearPrefix .. "000.png")

  local scale = 1.25
  firstFrameBigGear:setScl(scale)
  firstFrameLittleGear:setScl(scale)

  local bigGearAnimation = Animation.newWithPrefix(bigGearPrefix)
  local littleGearAnimation = Animation.newWithPrefix(littleGearPrefix)
  local frameRate = 30
  self.bigGearAnimationAction = AHSequence.new(
    AHAnimate.new(firstFrameBigGear, bigGearAnimation, frameRate),
    function()
      swapInFinalFrame(firstFrameBigGear, "big_gear_final_frame.png")
    end
  )
  self.littleGearAnimationAction = AHSequence.new(
    AHAnimate.new(firstFrameLittleGear, littleGearAnimation, frameRate),
    function()
      swapInFinalFrame(firstFrameLittleGear, "little_gear_final_frame.png")
    end
  )

  local x = self:getWidthPart(0.55)
  local y = self:getHeightPart(0.225)
  firstFrameBigGear:setLoc(x, y)
  firstFrameLittleGear:setLoc(x, y)
  self:insert(firstFrameBigGear, Z_INDEX.GEARS)
  self:insert(firstFrameLittleGear, Z_INDEX.GEARS)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioSkillAnimationProcessing:playFadeAction()
  local removeAnimation = AHSequence.new(
    AHOpacityTo.new(self.overlay, 0.2, nil, 1.0),
    AHRemove.new(self)
  )
  return removeAnimation:start()
end

function OnboardioSkillAnimationProcessing:getAnimationAction()
  local durationPartOne = 0.5
  local durationPartTwo = 1.0

  local animationPartOne = getAnimationPartOne(self, durationPartOne)
  local animationPartTwo = getAnimationPartTwo(self, durationPartTwo)

  local partTwoDelay = durationPartOne * 0.225
  local delayedPartTwo = AHDelayed.new(partTwoDelay, animationPartTwo)

  local animation = AHSpawn.new(animationPartOne, delayedPartTwo)
  return animation
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  setUpGears(self)
  setUpPaperAndShadow(self)

  self.overlay = FilledRectangle.new(self:getWidthPart(1.75), self:getHeightPart(1.95), COLORS.WHITE)
  local x = self:getWidthPart(0.5)
  local y = self:getHeightPart(0.65)
  self.overlay:setLoc(x, y)
  self.overlay:setOpacity(0.0)
  self:insert(self.overlay, Z_INDEX.OVERLAY)
end

-- Constructor ------------------------

function OnboardioSkillAnimationProcessing.new(width)
  local self = FilledRectangle.new(width, width, COLORS.WHITE)
  Mixin.mix(self, OnboardioSkillAnimationProcessing)

  createLayout(self)

  local referenceWidth = 215.0
  local scale = width / referenceWidth
  self:setScl(scale)

  return self
end
