--- OnboardioSkillAnimationRetention
-- @author Created by Jay Austria in April 2017

class("OnboardioSkillAnimationRetention")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WHITE = OnboardioCommon.COLOR_WHITE
local BACKGROUND_COLOR = OnboardioCommon.COLOR_BACKGROUND

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioSkillAnimationRetention:playFadeAction(duration)
  local fadeAction = AHSequence.new(
    AHOpacityTo.new(self.overlay, duration, nil, 1.0),
    AHRemove.new(self)
  )
  return fadeAction:start()
end

function OnboardioSkillAnimationRetention:reset()
  self.backing:setScl(0)
  self.head:setScl(0)
  self.overlay:setOpacity(1)
end

function OnboardioSkillAnimationRetention:getAnimationAction()
  local shiftAmountX = self.headWidth * self.widthShiftPercentage * 0.5
  local duration = 1.0
  local fadeInDuration = duration * 0.08
  local scaleDuration = duration * 0.7
  local translateDuration = duration * 1.5
  local arrowRotateDuration = duration * 1.45
  local minuteHandRotateDuration = duration * 1.04
  local hourHandRotateDuration = duration * 0.91

  local scaleEase = AHEaseElasticOut.new(1.2)
  local translateEase = AHEaseElasticOut.new(0.75)
  local arrowRotateEase = AHEaseElasticOut.new(1.2)
  local clockHandRotateEase = AHEaseExponentialOut.new()

  return AHSpawn.new(
    AHOpacityTo.new(self.overlay, fadeInDuration, nil, 0.0),
    AHScaleTo.new(self.head, scaleDuration, scaleEase, self.scaleToFit),
    AHScaleTo.new(self.backing, scaleDuration, scaleEase, self.scaleToFit),
    AHLocationXBy.new(self.backing, translateDuration, translateEase, shiftAmountX),
    AHLocationXBy.new(self.head, translateDuration, translateEase, -shiftAmountX),
    AHRotationBy.new(self.arrow, arrowRotateDuration, arrowRotateEase, 540.0),
    AHRotationTo.new(self.minuteHand, minuteHandRotateDuration, clockHandRotateEase, -135.0),
    AHRotationTo.new(self.hourHand, hourHandRotateDuration, clockHandRotateEase, 0.0)
  )
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makeArrow(width)
  local arrowRadius = width * 0.34
  local arcWidth = arrowRadius * 2
  local startAngle = math.rad(-105)
  local endAngle = math.rad(181)
  local lineWidth = arrowRadius * 0.175
  local blurMargin = nil
  local hasCenterPoint = false
  local mainArc = HollowArc.new(arrowRadius, startAngle, endAngle, WHITE, lineWidth, blurMargin, hasCenterPoint)

  local arcPoints = mainArc._normalizedPointLists[1]
  local startPoint = arcPoints[1]
  local roundCapRadius = lineWidth / 2

  local startCap = FilledCircle.new(roundCapRadius)
  local startPointAngle = math.pi - math.abs(startAngle)
  local startX = startPoint.x * arcWidth + roundCapRadius * math.cos(startPointAngle)
  local startY = startPoint.y * arcWidth + roundCapRadius * math.sin(startPointAngle)
  startCap:setLoc(startX, startY)
  mainArc:insert(startCap)

  local endExtender = FilledRectangle.new(lineWidth, arrowRadius * 0.3)
  endExtender:setAnchorPointTop()
  endExtender:setLoc(roundCapRadius, arrowRadius)
  mainArc:insert(endExtender)

  local endCap = FilledCircle.new(roundCapRadius)
  endExtender:insertAtBottom(endCap)

  local leftWing = FilledRectangle.new(lineWidth, arrowRadius * 0.4)
  local leftCap = FilledCircle.new(roundCapRadius)
  leftWing:setAnchorPointBottom()
  leftWing:setRot(45)
  leftWing:insertAtTop(leftCap)
  endExtender:insertAtBottom(leftWing)

  local rightWing = FilledRectangle.new(lineWidth, arrowRadius * 0.4)
  local rightCap = FilledCircle.new(roundCapRadius)
  rightWing:setAnchorPointBottom()
  rightWing:setRot(-45)
  rightWing:insertAtTop(rightCap)
  endExtender:insertAtBottom(rightWing)

  return mainArc
end

local function createClockHand(lengthFraction, arrowWidth)
  local handWidth = arrowWidth * 0.05
  local length = arrowWidth * lengthFraction
  return FilledRoundedRectangle.new(handWidth, length, 2, WHITE)
end

-- Constructor ------------------------

function OnboardioSkillAnimationRetention.new(width)
  local widthShiftPercentage = 0.1

  local head = Sprite.newWithName("retention_head_front.png")

  local fullWidth = head:getWidth() * (1.0 + widthShiftPercentage)
  local scaleToFit = width / fullWidth

  head:setScl(scaleToFit)

  local headWidth = head:getEffectiveWidth()
  local headHeight = head:getEffectiveHeight()

  local self = Group.new(width, headHeight)
  Mixin.mix(self, OnboardioSkillAnimationRetention)

  self.head = head
  self.headWidth = headWidth
  self.scaleToFit = scaleToFit
  self.widthShiftPercentage = widthShiftPercentage

  local backing = Sprite.newWithName("retention_head_back.png")
  backing:setAnchorPointLeft()
  self:insertAtLocationFractional(backing, 0.48, 0.5)
  backing:setScl(scaleToFit)
  self.backing = backing

  self:insertAtCenter(head)

  local arrow = makeArrow(width)
  arrow:setRot(180.0)
  local arrowWidth, arrowHeight = arrow:getEffectiveDims()

  local clockGroup = Group.new(arrowWidth, arrowHeight)
  head:insertAtLocationFractional(clockGroup, 0.55, 0.6)

  self.clockGroup = clockGroup
  self.arrow = arrow

  self.clockGroup:insertAtCenter(arrow)

  local minuteHand = createClockHand(0.34, arrowWidth)
  minuteHand:setAnchorPointBottom()
  minuteHand:setRot(-50.0)
  self.clockGroup:insertAtCenter(minuteHand)
  self.minuteHand = minuteHand

  local hourHand = createClockHand(0.18, arrowWidth)
  hourHand:setAnchorPointBottom()
  hourHand:setRot(15.0)
  self.clockGroup:insertAtCenter(hourHand)
  self.hourHand = hourHand

  local handConnector = FilledCircle.new(arrowWidth * 0.025)
  self.clockGroup:insertAtCenter(handConnector)

  local paddedWidth = width * 1.2
  local overlay = FilledRectangle.new(paddedWidth, self:getHeight(), BACKGROUND_COLOR)
  self:insertAtCenter(overlay)
  self.overlay = overlay

  self:reset()

  return self
end
