--- OnboardioSlideTransition
-- @author Created by Aaron Barrett on November 10, 2014.

class("OnboardioSlideTransition")
require("OnboardioViewTransitioner")
require("OnboardioPerspective")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SLIDE_KIND_APPEAR = "SLIDE_KIND_APPEAR"
local SLIDE_KIND_DISAPPEAR = "SLIDE_KIND_DISAPPEAR"
local SLIDE_KIND_MOVE = "SLIDE_KIND_MOVE"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Button Transparency ----------------

local function turnFrontButtonTransparent(oldButton, newButton)
  local oldLabel = oldButton:getConceptView()
  oldLabel:clearAttrLink(MOAIColor.ATTR_OPACITY)

  local newLabel = newButton:getConceptView()
  newLabel:clearAttrLink(MOAIColor.ATTR_OPACITY)
  newLabel:setOpacity(0.0)
  newButton:setOpacity(0.0)
end

local function restoreFrontButtonOpacity(oldButton, newButton)
  local newLabel = newButton:getConceptView()

  newButton:setOpacity(1.0)
  newLabel:inheritOpacity(newButton)
  oldButton:setOpacity(0.0)
end

-- Play Transition --------------------

local function doButtonsHaveSameText(button1, button2)
  local label1 = button1:getConceptView()
  local label2 = button2:getConceptView()

  return label1.getText ~= nil and label2.getText ~= nil and label1:getText() == label2:getText()
end

local function playMoveWithDescriptor(descriptor)
  local oldView = descriptor.oldView
  local newView = descriptor.newView

  local scaleDownAmount = 0.9

  local wasPressed = descriptor.wasPressed
  if (wasPressed) then
    newView:playPressedColor()

    scaleDownAmount = 0.8
  end

  local slideDuration = 0.5
  local slideEase = AHEaseSimpleInOut.new(2.5)

  local destinationX = descriptor.destinationX
  local destinationY = descriptor.destinationY
  AHLocationTo.new(newView, slideDuration, slideEase, destinationX, destinationY):start()
  AHLocationTo.new(oldView, slideDuration, slideEase, destinationX, destinationY):start()


  local scaleDuration = 0.65
  local scaleEase = AHEaseElasticOut.new(0.45)

  AHSequence.new(
                 AHScaleTo.new(newView, scaleDuration * 0.2, AHEaseSimpleOut.new(2.0), scaleDownAmount),
                 AHScaleTo.new(newView, scaleDuration * 0.8, scaleEase, 1.0)
                 ):start()

  local newViewWidth, newViewHeight = newView:getDims()
  local oldViewWidth, oldViewHeight = oldView:getDims()

  local destinationScaleX = newViewWidth / oldViewWidth
  local destinationScaleY = newViewHeight / oldViewHeight
  local intermediateScaleX = destinationScaleX * scaleDownAmount
  local intermediateScaleY = destinationScaleY * scaleDownAmount
  AHSequence.new(
                 AHScaleTo.new(oldView, scaleDuration * 0.2, AHEaseSimpleOut.new(1.5), intermediateScaleX, intermediateScaleY),
                 AHScaleTo.new(oldView, scaleDuration * 0.8, scaleEase, destinationScaleX, destinationScaleY)
                 ):start()

  if (doButtonsHaveSameText(oldView, newView)) then
    restoreFrontButtonOpacity(oldView, newView)
  else
    AHSequence.new(
                   AHSpawn.new(
                               AHOpacityTo.new(oldView:getConceptView(), 0.05, nil, 0.0),
                               AHOpacityTo.new(newView:getConceptView(), 0.15, nil, 1.0)
                               ),
                   function()
                     restoreFrontButtonOpacity(oldView, newView)
                   end
                   ):start()
  end

end

local function playAppearWithDescriptor(descriptor)
  local oldView = descriptor.oldView
  local newView = descriptor.newView

  if (oldView ~= nil) then
    local scaleDownAmount = 0.9

    local slideDuration = 0.5
    local slideEase = AHEaseSimpleInOut.new(2.5)

    local destinationX = descriptor.destinationX
    local destinationY = descriptor.destinationY
    AHLocationTo.new(newView, slideDuration, slideEase, destinationX, destinationY):start()

    local scaleDuration = 0.65
    local scaleEase = AHEaseElasticOut.new(0.45)

    AHSequence.new(
                   AHScaleTo.new(newView, scaleDuration * 0.2, AHEaseSimpleOut.new(2.0), scaleDownAmount),
                   AHScaleTo.new(newView, scaleDuration * 0.8, scaleEase, 1.0)
                   ):start()

    AHOpacityTo.new(newView, 0.3, AHEaseExponentialIn.new(), 1.0):start()
  else
    newView:setScl(0.9)

    local perspectiveDistance = descriptor.perspectiveDistance
    runInCoroutine(function()
                     waitFor(0.15)
                     OnboardioViewTransitioner.playStationaryTransitionInOnView(newView, perspectiveDistance)
                   end)
  end
end

local function playDisappearWithDescriptor(descriptor)
  local oldView = descriptor.oldView
  local perspectiveDistance = descriptor.perspectiveDistance
  local wasPressed = descriptor.wasPressed

  OnboardioViewTransitioner.playStationaryTransitionOutOnView(oldView, perspectiveDistance, wasPressed)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OnboardioSlideTransition:buildDescriptors()
  local oldPanel = self.oldPanel
  local newPanel = self.newPanel

  local descriptors = {}

  local oldSlidingButtons = nil
  local newSlidingButtons = nil

  if (oldPanel ~= nil) then
    oldSlidingButtons = oldPanel:getSlidingButtons()
  else
    oldSlidingButtons = {}
  end

  if (newPanel ~= nil) then
    newSlidingButtons = newPanel:getSlidingButtons()
  else
    newSlidingButtons = {}
  end

  local descriptorCount = math.max(#oldSlidingButtons, #newSlidingButtons)
  local lastOldView = nil
  for i = 1, descriptorCount do
    local oldView = oldSlidingButtons[i]
    local newView = newSlidingButtons[i]

    local slideKind = nil

    local perspectiveDistance = nil
    local wasPressed = false

    if ((oldView ~= nil) and (newView ~= nil)) then
      slideKind = SLIDE_KIND_MOVE

      wasPressed = oldView:getWasPressed()
    elseif (oldView ~= nil) then
      slideKind = SLIDE_KIND_DISAPPEAR
      perspectiveDistance = OnboardioPerspective.getPerspectiveDistanceForY(oldView:getY())

      wasPressed = oldView:getWasPressed()
    else
      slideKind = SLIDE_KIND_APPEAR
      oldView = lastOldView
      perspectiveDistance = OnboardioPerspective.getPerspectiveDistanceForY(newView:getY())
    end

    local destinationX = nil
    local destinationY = nil
    if (newView ~= nil) then
      destinationX, destinationY = newView:getLoc()
    end

    local descriptor = {
      oldView = oldView,
      newView = newView,
      slideKind = slideKind,
      destinationX = destinationX,
      destinationY = destinationY,
      perspectiveDistance = perspectiveDistance,
      wasPressed = wasPressed
    }
    table.insert(descriptors, descriptor)

    if (oldView ~= nil) then
      lastOldView = oldView
    end
  end

  self.descriptors = descriptors
end

function OnboardioSlideTransition:prepareToTransition()
  local descriptors = self.descriptors

  for _, descriptor in ipairs(descriptors) do
    local slideKind = descriptor.slideKind

    local oldView = descriptor.oldView
    local newView = descriptor.newView

    if (slideKind == SLIDE_KIND_MOVE) then
      local x, y = oldView:getLoc()
      newView:setLoc(x, y)

      local width, height = oldView:getDims()
      newView:setEffectiveDims(width, height)

      turnFrontButtonTransparent(oldView, newView)

    elseif (slideKind == SLIDE_KIND_APPEAR) then
      if (oldView ~= nil) then
        local x, y = oldView:getLoc()
        newView:setLoc(x, y)

        local width, height = oldView:getDims()
        newView:setEffectiveDims(width, height)
      else
        local perspectiveDistance = descriptor.perspectiveDistance
        local y = newView:getY()
        newView:setY(y - perspectiveDistance)
      end

      newView:setOpacity(0.0)
    end
  end
end

function OnboardioSlideTransition:playTransition()
  local descriptors = self.descriptors

  for _, descriptor in ipairs(descriptors) do
    local slideKind = descriptor.slideKind
    if (slideKind == SLIDE_KIND_MOVE) then
      playMoveWithDescriptor(descriptor)
    elseif (slideKind == SLIDE_KIND_APPEAR) then
      playAppearWithDescriptor(descriptor)
    elseif (slideKind == SLIDE_KIND_DISAPPEAR) then
      playDisappearWithDescriptor(descriptor)
    end
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OnboardioSlideTransition.new(oldPanel, newPanel)
  local self = {}
  Mixin.mix(self, OnboardioSlideTransition)

  self.oldPanel = oldPanel
  self.newPanel = newPanel

  return self
end
