--- CircularParticleExplosion
-- @author Created by Alexey Chernikov

class("CircularParticleExplosion")

require("animations.ParticleWithGlow")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_SETS = {
  {0.49, 0.98, 0.79},
  {0.07, 0.68, 0.54},
  {0.0, 0.48, 0.41},
  {0.0, 0.29, 0.3},
}

local preloadedInstance = nil

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function chooseColor(self)
  local randomColorIndex = math.random(#self.colors)
  return self.colors[randomColorIndex]
end

local function createParticle(self)
  local color = chooseColor(self)

  local radius = self.minRadius + math.random() * self.radiusAmp
  local particle = ParticleWithGlow.new(color, radius)

  return particle
end

local function createNParticles(self, N, container)
  container = container or {}
  for i = 1, N do
    local particle = createParticle(self)
    container[particle] = true
    self:insert(particle)
  end
  return container
end

local function addParticle(self, particle, time)
  particle.birthday = time
  self.particles[particle] = true
  self.particleCount = self.particleCount + 1
  if (particle.parent ~= self) then
    self:insert(particle)
  end
end

local function randomXYForCircle(radiusAmp)
  local angle = math.random() * 360
  local R = math.random() * radiusAmp
  local x = math.cos(angle) * R
  local y = math.sin(angle) * R
  local xSign, ySign = MathUtils.randomSign(), MathUtils.randomSign()
  x = x * xSign
  y = y * ySign
  return x, y
end

local function initializeParticle(self, particle)
  particle.xSpeed, particle.ySpeed = randomXYForCircle(self.speedAmp)
  particle.xAcceleration = self.acceleration
  particle.yAcceleration = self.acceleration

  local scaleSpeed = math.random()
  local scaleStartTime = self.particlesLifeTime * (math.random() / 10 + 0.10)
  local decelerateStartTime = 0.1
  particle.xScaleSpeed, particle.yScaleSpeed = scaleSpeed, scaleSpeed

  particle.scaleStartTime = scaleStartTime
  particle.decelerateStartTime = decelerateStartTime

  local centerCoordinate = self.innerRadius * 2.0
  particle:setLoc(centerCoordinate, centerCoordinate)

end

local function spawnParticlesImmediately(self, particlesToSpawn, time)
  if (self.particlePool) then
    for particle, _ in pairs(self.particlePool) do
      addParticle(self, particle, time)
      initializeParticle(self, particle)
      particlesToSpawn = particlesToSpawn - 1
      self.particlePool[particle] = nil
      particle.isVisible = true
    end
  end
  for i = 1, particlesToSpawn do
    local particle = createParticle(self, time)
    addParticle(self, particle, time)
    initializeParticle(self, particle)
  end
end

local function spawnParticlesByBurst(self, time)
  local hasEnoughTimeElapsedSinceLastBurst = (time - self.lastBurstTime >= self.timePerBurst)
  if (not hasEnoughTimeElapsedSinceLastBurst) then
    return
  end

  local hasBurstCountReachedMaximum = (self.burstCount >= self.maxBursts)
  if (hasBurstCountReachedMaximum) then
    return
  end

  self.burstCount = self.burstCount + 1

  self.lastBurstTime = time
  spawnParticlesImmediately(self, self.particlesPerBurst, time)
end

local function stashParticleForLater(self, particle)
  if (self.particlePool) then
    self.particlePool[particle] = true
    particle.isVisible = false
  else
    particle:removeSelf()
  end
end

local function killOneParticle(self, particle)
  self.particles[particle] = nil
  self.particleCount = self.particleCount - 1
  stashParticleForLater(self, particle)
end

local function killParticles(self, time)
  local timeToDie = self.particlesLifeTime
  for particle, _ in pairs(self.particles) do
    local particleLife = time - particle.birthday
    if (particleLife >= timeToDie) then
      killOneParticle(self, particle)
    end
  end
end

local function applyParticleMotion(self, particle, age)
  local dt = Screen.stepDuration
  local x, y = particle:getLoc()
  local dx = particle.xSpeed * dt
  local dy = particle.ySpeed * dt
  local dxSpeed = 1 + particle.xAcceleration * dt
  local dySpeed = 1 + particle.yAcceleration * dt

  particle.xSpeed = particle.xSpeed * dxSpeed
  particle.ySpeed = particle.ySpeed * dySpeed

  if (age > particle.scaleStartTime) then
    local scaleTime = (age - particle.scaleStartTime)/(self.particlesLifeTime - particle.scaleStartTime)
    particle:setScl(1 - scaleTime)
  end
  if (age > particle.decelerateStartTime) then
    particle.xAcceleration, particle.yAcceleration = self.decelerationRate, self.decelerationRate
  end

  local xTurbulence = self.turbulence * math.random() * 2 - self.turbulence
  local yTurbulence = self.turbulence * math.random() * 2 - self.turbulence
  local particleX = x + dx + xTurbulence
  local particleY = y + dy + yTurbulence
  particle:setLoc(particleX, particleY)
end

local function moveParticles(self, time)
  for particle, _ in pairs(self.particles) do
    local age = time - particle.birthday
    applyParticleMotion(self, particle, age)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CircularParticleExplosion.preload()
  preloadedInstance = RightAnswerParticleEffect.new()
  preloadedInstance.particlePool = createNParticles(preloadedInstance, preloadedInstance.particlesPerBurst)
  return preloadedInstance
end

function CircularParticleExplosion.getPreloadedInstance()
  return preloadedInstance or RightAnswerParticleEffect.preload()
end

function CircularParticleExplosion:setParticleLifeTime(particlesLifeTime)
  self.particlesLifeTime = particlesLifeTime
end

function CircularParticleExplosion:start(runTime, onComplete)
  runTime = runTime or self.particlesLifeTime
  self.lastBurstTime = -self.timePerBurst
  self.burstCount = 0

  local time = 0.0
  local deltaTime = Screen.stepDuration
  local stepCount = math.floor(runTime / deltaTime)

  local particlesTimestep = function()
    killParticles(self, time)
    spawnParticlesByBurst(self, time)
    moveParticles(self, time)
    time = time + deltaTime
  end

  local stepParticles = AHLoop.new(particlesTimestep, stepCount)
  local actions = {stepParticles}

  local finishEffect = nil
  if (onComplete) then
    finishEffect = function() onComplete(self) end
    table.insert(actions, finishEffect)
  end

  local stepParticlesAndComplete = AHSequence.new(actions)
  stepParticlesAndComplete:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function CircularParticleExplosion.new(speed, acceleration, innerRadius, colors)
  colors = colors or COLOR_SETS
  speed = speed or 1
  acceleration = acceleration or 1
  local innerDiameter = innerRadius * 2.0
  local self = Group.new(innerDiameter, innerDiameter)
  Mixin.mix(self, CircularParticleExplosion)

  self.particlesLifeTime = 1

  -- It's either particles per burst, or per second
  self.particlesPerBurst = 5
  self.timePerBurst = 0.01
  self.lastBurstTime = -self.timePerBurst
  self.maxBursts = 8

  self.particleCount = 0
  self.minRadius, self.maxRadius = 4, 6
  self.minSpeed, self.maxSpeed = 30 * speed, 60 * speed
  self.innerRadius = innerRadius or 0
  self.acceleration = acceleration

  self.turbulence = 0.8
  self.turbulenceStart = self.particlesLifeTime/2
  self.decelerationRate = -2 * self.acceleration * 0.8

  self.colors = colors

  -- Derived parameters
  self.radiusAmp = self.maxRadius - self.minRadius
  self.speedAmp = self.maxSpeed - self.minSpeed

  self.particles = {}

  return self
end
