--- TextStrobeIn
-- @author Created by Alexey Chernikov

class("TextStrobeIn")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_COLORS = {
  {0.8, 0.8, 0.0, 1},
  {0.8, 0.8, 0.8, 1},
  {0.0, 0.8, 0.8, 1},
  {0.0, 0.0, 0.8, 1},
  {0.8, 0.0, 0.8, 1},
  {0.8, 0.0, 0.0, 1}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function positionEase(t)
  return math.pow(t, 0.3) * 0.6 + math.pow(t, 0.8) * 0.4
end

local function setFlasherColor(flasher, color)
  flasher.color = color
  flasher.invisibleColor = {unpack(flasher.color)}
  flasher.invisibleColor[4] = 0
  flasher:setColor(unpack(flasher.color))
end

local function createFlashers(self, message, font, colors, endColor)
  local maxHeight = Screen.getHeightPart(0.5)
  local flashers = {}
  local flasherGroup = Group.new(Screen.width, maxHeight)

  local desiredFlashers = 25
  local currentFlasherCount = 0

  local function getFlasherY()
    local normalizedFlasherIndex = currentFlasherCount / (desiredFlashers - 1)
    normalizedFlasherIndex = positionEase(normalizedFlasherIndex)
    return maxHeight * normalizedFlasherIndex
  end

  local function addFlasher()
    local flasher = Label.new(message, font)
    flasher:setVisible(false)
    local w, h = flasher:getDims()

    local y = getFlasherY()
    flasher:setLoc(Screen.getWidthPart(0.5), y + h / 2)

    flasherGroup:insert(flasher)

    local index = #flashers + 1
    local colorIndex = (index - 1) % #colors + 1
    setFlasherColor(flasher, colors[colorIndex])
    flashers[index] = flasher

    currentFlasherCount = currentFlasherCount + 1
    flasher:setAttrLink(MOAIColor.INHERIT_COLOR, self, MOAIColor.COLOR_TRAIT)

    return flasher, w, h
  end

  self.initialized = false
  local function finishInitialization()
    if (endColor) then
      setFlasherColor(flashers[#flashers], endColor)
    end
    self.initialized = true
  end
  local function initializeAsync()
    runInCoroutine(function()
      while(currentFlasherCount < desiredFlashers) do
        addFlasher()
        coroutine.yield()
      end

      finishInitialization()
    end)
  end

  initializeAsync()

  return flasherGroup, flashers
end

local function createLayout(self, message, font, colors, endColor)
  local flasherGroup, flashers = createFlashers(self, message, font, colors, endColor)
  self:insertAtCenter(flasherGroup)
  self.flashers = flashers
  self.flasherGroup = flasherGroup
end

local function strobeOneFlasher(self, flasher, time)
  if (flasher.animation) then
    flasher.animation:stop()
  end
  flasher:setVisible(true)
  flasher:setColor(unpack(flasher.color))
  flasher.animation = AHColorTo.new(flasher, time, nil, flasher.invisibleColor):start()
end

local function startAnimation(self, duration, completeHandler)
  local totalFlashers = #self.flashers
  local timePerFlash = 0.6
  local lastVisibleIndex = 0

  local deltaTime = Screen.stepDuration
  local time = 0.0

  local updateFlashersOnce = function()
    local curVisibleIndex = math.floor(time * totalFlashers)
    while(lastVisibleIndex < curVisibleIndex) do
      lastVisibleIndex = lastVisibleIndex + 1
      strobeOneFlasher(self, self.flashers[lastVisibleIndex], timePerFlash)
    end

    time = time + deltaTime

    if (time > 1.0) then
      self.flashers[curVisibleIndex].animation:stop()
      self.updateFlashers:stop()
    end
  end

  self.updateFlashers = AHLoop.new(updateFlashersOnce)
  local updateFlashersAndComplete = AHSequence.new(self.updateFlashers, completeHandler)
  updateFlashersAndComplete:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TextStrobeIn:start(time, completeHandler)
  runInCoroutine(function()
    while (not self.initialized) do
      coroutine.yield()
    end
    startAnimation(self, time, completeHandler)
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TextStrobeIn.new(message, font, colors, endColor)
  local self = Group.new(Screen.width, Screen.getHeightPart(0.5))
  Mixin.mix(self, TextStrobeIn)
  colors = colors or DEFAULT_COLORS
  createLayout(self, message, font, colors, endColor)

  return self
end
