  --- TriangleRingExplosion
-- @author Created by Joe Poka on March 31, 2017.

class("TriangleRingExplosion")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createTriangles(self, triangleParameters)
  local trianglesCount = triangleParameters.trianglesCount
  local triangleWidth = triangleParameters.particleWidth
  local triangleHeight = triangleParameters.particleHeight
  local startRadius = triangleParameters.startRadius
  local startRadiusRandomness = triangleParameters.startRadiusRandomness
  local endRadius = triangleParameters.endRadius

  local p1x, p1y = -triangleWidth / 2, 0
  local p2x, p2y = 0, triangleHeight
  local p3x, p3y = triangleWidth / 2, 0
  local color = {1, 1, 1, 1}

  local angleDegrees = 0.0
  local deltaAngleDegrees = 360.0 / trianglesCount
  self.triangleDestinationPositions = {}

  local isRecursive = true

  for i = 1, trianglesCount do
    local triangle = FilledTriangle.new(p1x, p1y, p2x, p2y, p3x, p3y, color)
    triangle:setRot(angleDegrees + 90.0)
    self:insertAtCenter(triangle)

    local startRadiusOffset = (i % 2) * startRadiusRandomness
    local randomStartRadius = startRadius + startRadiusOffset

    local angleRadians = math.rad(angleDegrees)
    local startOffsetX = randomStartRadius * math.cos(angleRadians)
    local startOffsetY = randomStartRadius * math.sin(angleRadians)
    triangle:setLoc(triangle:getX() + startOffsetX, triangle:getY() + startOffsetY)

    local deltaRadius = endRadius - startRadius
    local destinationX = triangle:getX() + deltaRadius * math.cos(angleRadians)
    local destinationY = triangle:getY() + deltaRadius * math.sin(angleRadians)
    local destinationPosition = {
      x = destinationX,
      y = destinationY
    }
    self.triangleDestinationPositions[triangle] = destinationPosition

    angleDegrees = angleDegrees + deltaAngleDegrees
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TriangleRingExplosion:start(duration, startColor, endColor)
  local moveTriangles = AHSpawn.new()

  for triangle, destinationPosition in pairs(self.triangleDestinationPositions) do
    local moveRadiallyOutward = AHLocationTo.new(triangle, duration, AHEaseExponentialOut.new(), destinationPosition.x, destinationPosition.y)
    moveTriangles:addAction(moveRadiallyOutward)

    triangle:setScl(1.0)

    local scaleDownDuration = 0.2
    local scaleDownDelay = duration - scaleDownDuration
    local scaleDown = AHScaleTo.new(triangle, scaleDownDuration, nil, 0.7)
    local delayedScaleDown = AHDelayed.new(scaleDownDelay, scaleDown)
    moveTriangles:addAction(delayedScaleDown)
  end

  local cleanUp = AHRemove.new(self)

  local moveAndCleanUp = AHSequence.new(moveTriangles, cleanUp)
  moveAndCleanUp:start()

  if (startColor and endColor) then
    self:setColor(unpack(startColor))
    local colorEase = AHEaseSimpleOut.new()
    AHColorTo.new(self, duration, colorEase, endColor):start()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TriangleRingExplosion.new(startRadius, endRadius, trianglesCount, startRadiusRandomness, particleWidth, particleHeight)
  local endDiameter = endRadius * 2.0
  local self = Group.new(endDiameter, endDiameter)
  Mixin.mix(self, TriangleRingExplosion)

  local triangleParameters = {
    trianglesCount = trianglesCount,
    particleWidth = particleWidth,
    particleHeight = particleHeight,
    startRadius = startRadius,
    startRadiusRandomness = startRadiusRandomness,
    endRadius = endRadius
  }

  createTriangles(self, triangleParameters)

  self:linkColorAttrToChildren(isRecursive)

  return self
end
