--- OrbitGameManager
-- @author Created by Alexey Chernikov

class("OrbitGameManager")

require("game.OrbitRoundGenerator")
require("bonuses.ResponseSpeedTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self, round)
  local rankTwoScore = OrbitConfiguration.getRankTwoScore()
  local rankThreeScore = OrbitConfiguration.getRankThreeScore()
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  self.scorePerAnswer = OrbitConfiguration.getScorePerAnswer(round)
  self.pointsPerFullSpeedBonus = OrbitConfiguration.getPointsPerFullSpeedBonus(round)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function showGameTip(self)
  self.audioTrack:pause()
  self.scene:showTip(function()
    self.audioTrack:resume()
  end)
end

local function setLives(self, lives)
  self.lives = lives
  if (self.lives <= 0) then
    self.disabled = true
    runInCoroutine(function()
      wait(self.baseGame:showDefaultGameLossOutro())
      GameSession.notifyIncorrectAnswer(self.round.id)
      GameSession.notifyGameLoss()
    end)
  else
    self.livesView:setLives(self.lives)
  end
end

local function loseLife(self)
  setLives(self, self.lives - 1)
end

local function win(self, delay)
  delay = delay or 2
  MOAICoroutine.new():run(function()
    ScoringUtils.applyLivesBonus(self.lives, OrbitConfiguration.getPointsPerLifeBonus(), self.baseGame.bonusTray)
    self.scoreView:setScore(GameSession.getScore())

    waitFor(delay)
    GameSession.notifyCorrectAnswer(self.round.id)

    GameSession.notifyGameWin()
  end)
end

local function isTopicComplete(self, topic)
  return TableUtils.isEmpty(self.phrasesByTopic[topic])
end

local function isVictoryAchieved(self)
  for topic, _ in pairs(self.phrasesByTopic) do
    if (not isTopicComplete(self, topic)) then
      return false
    end
  end
  return true
end

local function acceptBonusAnswer(self, phraseInfo, assignedTopic)
    Sound.playSoundEffect("focus_goal_reached", 0.7)
    GameSession.increaseScore(OrbitConfiguration.POINTS_PER_BONUS_ANSWER)
    self.scoreView:setScore(GameSession.getScore())

    local phraseText = phraseInfo.phrase
    GameSession.contentTracker:notifyCorrectAnswer(self.round.id, phraseText)

    self.scene:assignBonusPhraseToTopic(phraseInfo)
    self.scene:victoriouslyCollapseTopic(assignedTopic)
    self.correctAnswersRemaining = self.correctAnswersRemaining - 1

    if (self.correctAnswersRemaining < 1) then
      win(self, 2) -- Let the player marvel at the beauty of a complete scene
    end
end

local function rejectBonusAnswer(self, phraseInfo, assignedTopic)
  Sound.playSoundEffect("focus_incorrect", 0.7)
  self.scene:rejectPhraseFromLabelBucket(phraseInfo)
  self.scene:playRejectionAnimationForTopic(assignedTopic)
  loseLife(self)

  local phraseText = phraseInfo.phrase
  GameSession.contentTracker:notifyIncorrectAnswer(self.round.id, phraseText, assignedTopic)
end

local function startBonusRound(self)
  if (self.round.bonus ~= nil) then
    self.round.bonus.topics = self.round.topics
    self.scene:startBonusRound(self.round.bonus)
    self.correctAnswerHandler = acceptBonusAnswer
    self.wrongAnswerHandler = rejectBonusAnswer
    self.correctAnswersRemaining = 3
  else
    Log.error("No bonus round found, ending game.")
    win(self)
  end
end

local function acceptAnswer(self, phraseInfo, assignedTopic)
  local currentTime = self.audioTrack:getTime()
  local spawnTime = phraseInfo.spawnTime
  local responseDuration = currentTime - spawnTime
  if(responseDuration < 0) then
    -- When track ends, current time is 0. Don't award speed bonuses.
    responseDuration = math.huge
  end
  self.responseSpeedTracker:notifyCorrectResponseWithValue(responseDuration)
  local phraseText = phraseInfo.phrase
  if (not self.incorrectlyMatchedPhrases[phraseText]) then
    GameSession.contentTracker:notifyCorrectAnswer(self.round.id, phraseText)
  end
  GameSession.increaseScore(self.scorePerAnswer)
  self.scoreView:setScore(GameSession.getScore())

  local topic = phraseInfo.topic
  self.phrasesByTopic[topic][phraseInfo.phrase] = nil
  if (isTopicComplete(self, topic)) then
    Sound.playSoundEffect("focus_planet_complete", 0.7)
    self.scene:assignLastPhraseToTopic(phraseInfo, assignedTopic)
  else
    Sound.playSoundEffect("focus_correct", 0.7)
    self.scene:assignPhraseToTopic(phraseInfo, assignedTopic)
  end
end

local function rejectAnswer(self, phraseInfo, assignedTopic)
  Sound.playSoundEffect("focus_incorrect", 0.7)
  loseLife(self)
  local phraseText = phraseInfo.phrase
  self.incorrectlyMatchedPhrases[phraseText] = true
  GameSession.contentTracker:notifyIncorrectAnswer(self.round.id, phraseText, assignedTopic)
  self.scene:rejectPhraseFromLabelBucket(phraseInfo)
  self.scene:playRejectionAnimationForTopic(assignedTopic)
end

local function isCorrectTopicAssignment(self, phraseInfo, assignedTopic)
  local phraseTopic = phraseInfo.topic
  return phraseTopic == assignedTopic
end

local function initGame(self)
  setLives(self, self.livesView:getLives())
  self.correctAnswerHandler = acceptAnswer
  self.wrongAnswerHandler = rejectAnswer

  local function onQuickResponse(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 self.pointsPerFullSpeedBonus,
                                 self.baseGame.bonusTray)
  end
  local bonusWindowSize = OrbitConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS
  local bonusDecay = OrbitConfiguration.SPEED_BONUS_DECAY_START
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowSize, onQuickResponse, bonusDecay)
end

local function debugCompleteGameAutomatically(self, phrase, phraseIndex)
  -- Automatically assign all phrases to their planets - useful for debugging bonus round transitions
  -- Code goes into spawnPhrase() below
  local delay = 0.25 * phraseIndex
  local delayedFunction = function() self.correctAnswerHandler(self, phrase, phrase.topic) end
  local delayedFunctionCall = AHDelayed.new(delay, delayedFunction)
  delayedFunctionCall:start()
end

local function spawnPhrase(self, phrase)
  self.scene:spawnPhrase(phrase)
end

local function getPhrasesSortedBySpawnTime(self, round)
  local phrasesSortedBySpawnTime = TableUtils.shallowCopy(round.phrases)
  table.sort(phrasesSortedBySpawnTime, function(phrase1, phrase2) return phrase1.spawnTime < phrase2.spawnTime end)
  return phrasesSortedBySpawnTime
end

local function tryToNotifyUserOfGameplayRules(self, spawnIndex, trackTime)
  -- Spawn index == 3 means the next phrase to be spawned is 3, aka 2 concepts on screen
  local conceptsOnScreen = spawnIndex - 1
  local shouldNotify = (trackTime > 14 and conceptsOnScreen > 1) or (conceptsOnScreen > 2)
  if(shouldNotify) then
    self.callbacks.userIsWaiting()
  end
end

local function initPhraseSpawning(self, round)
  local phrasesSortedBySpawnTime = getPhrasesSortedBySpawnTime(self, round)
  local spawnIndex = 1
  local currentPhrase = phrasesSortedBySpawnTime[spawnIndex]
  local currentSpawnTime = currentPhrase.spawnTime

  local shouldSpawnPhrases = true
  runInCoroutine(function()
    while(shouldSpawnPhrases) do
      local time = self.audioTrack:getTime()
      while ((time > currentSpawnTime) or self.audioTrack:isFinished()) do
        spawnPhrase(self, currentPhrase)
        spawnIndex = spawnIndex + 1
        currentPhrase = phrasesSortedBySpawnTime[spawnIndex]
        if (currentPhrase == nil) then
          shouldSpawnPhrases = false
          return
        end
        currentSpawnTime = currentPhrase.spawnTime
        tryToNotifyUserOfGameplayRules(self, spawnIndex, time)
      end

      coroutine.yield()
    end
  end)
end

local function getCallbacksTable(self)
  local userUnderstandsGame = false
  local invalidInputAttempts = 0

  local answerSubmitted = function(answerData)
    if (self.disabled) then return end -- e.g. when the game's already over

    local phraseInfo = answerData.phraseInfo
    local assignedTopic = answerData.assignedTopic
    local isCorrect = isCorrectTopicAssignment(self, phraseInfo, assignedTopic)
    if (isCorrect) then
      userUnderstandsGame = true -- if the user can submit a correct answer, he don't need tips
      self.correctAnswerHandler(self, phraseInfo, assignedTopic)
    else
      self.wrongAnswerHandler(self, phraseInfo, assignedTopic)
    end
  end

  local invalidInput = function()
    if (self.audioTrack == nil) then
      return
    end
    if (userUnderstandsGame) then return end
    invalidInputAttempts = invalidInputAttempts + 1
    if (invalidInputAttempts >= 3 and not self.hasSeenGameTip) then
      showGameTip(self)
      self.hasSeenGameTip = true
    end
  end

  local userWaitingHandled = false
  local userIsWaiting = function()
    if (not userUnderstandsGame and not userWaitingHandled and not self.hasSeenGameTip) then
      userWaitingHandled = true
      showGameTip(self)
      self.hasSeenGameTip = true
    end
  end

  local callbacks = {
    answerSubmitted = answerSubmitted,
    invalidInput = invalidInput,
    userIsWaiting = userIsWaiting
  }

  return callbacks
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OrbitGameManager:start()
  local round = self.roundGenerator:next()
  GameSession.contentTracker:notifyPresentedConcept(round.id)
  self.scene:setRound(round)
  self.round = round
  self.phrasesByTopic = round.phrasesByTopic

  configureGameDifficulty(self, round)

  -- Best way to debug the memory round avoiding inconsistencies in content management with achilles and the app
  -- local audioTrack = {getTime = function() return 10 end}
  local delay = 1.5
  local delayedFunction = function()
    local audioTrack = AssetHelper.playSoundForConcept(round.concept, round.concept.sound)
    self.audioTrack = audioTrack
    initPhraseSpawning(self, round)
  end
  local delayedFunctionCall = AHDelayed.new(delay, delayedFunction)
  delayedFunctionCall:start()

  runInCoroutine(function()
    waitUntil(function()
      return (isVictoryAchieved(self) and (self.audioTrack and self.audioTrack:isFinished()))
    end)
    startBonusRound(self)
  end)
end

function OrbitGameManager:setScene(scene)
  self.scene = scene
end

function OrbitGameManager:getCallbacks()
  return self.callbacks
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function OrbitGameManager.new(baseGame)
  local self = {}
  self.scoreView = baseGame.scoreCounter
  self.livesView = baseGame.livesCounter
  self.baseGame = baseGame
  Mixin.mix(self, OrbitGameManager)

  self.roundGenerator = OrbitRoundGenerator.new()
  self.hasSeenGameTip = false
  self.incorrectlyMatchedPhrases = {}

  initGame(self)

  self.callbacks = getCallbacksTable(self)

  return self
end
