--- OrbitScene
-- @author Created by Alexey Chernikov

class("OrbitScene")

require("scene.floaters.FloatingObject")
require("scene.planets.LayeredPlanet")
require("scene.planets.PlanetsCluster")
require("scene.LayoutInfo")
require("scene.OrbitTip")
require("animations.TextStrobeIn")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)
  self.background = background
end

local function stopFloaterMovement(self, floater)
  if (floater.inertia) then floater.inertia:stop() end
end

local function removeFloater(self, floater)
  stopFloaterMovement(self, floater)
  floater:removeSelf()
  TableUtils.removeValueFromArray(self.allFloaters, floater)
end

local function getDragHandler(self)
  local dragPointDx, dragPointDy = nil, nil
  local started = false

  local lastTime = nil
  local dx, dy, dt = nil, nil, nil
  local lastX, lastY = nil, nil
  local startX, startY = nil, nil

  local startDrag = function(event, target)
      self.inputReceived = true
      target:select()
      stopFloaterMovement(self, target)

      startX, startY = target:getLoc()
      lastX, lastY = startX, startY
      dragPointDx, dragPointDy = startX - event.x, startY - event.y

      lastTime = self.movementTimer:getTime()
      dx, dy = 0, 0
      dt = math.huge
      started = true
  end

  local doInertia = function(target, dx, dy, dt)
    if not MathUtils.isRealNumber(dt) then
      return
    end
    
    local isDifferentFrame = (dt ~= 0)
    if not isDifferentFrame then
      return
    end

    local inertiaTime = 0.4
    local initialSpeedScale = 0.5
    local xSpeed = dx/dt * initialSpeedScale
    local ySpeed = dy/dt * initialSpeedScale
    local xDeceleration = 0.9
    local yDeceleration = 0.9
    local floaterAssigned = false

    local minimumStepCount = 1
    local stepCount = math.max(minimumStepCount, math.floor(inertiaTime / dt))

    local inertiaFunction = function()
      if (floaterAssigned) then
        target.inertia:stop()
        return
      end

      local x, y = target:getLoc()
      local dx, dy = xSpeed * dt, ySpeed * dt
      xSpeed = xSpeed * xDeceleration
      ySpeed = ySpeed * yDeceleration
      local newX, newY = x + dx, y + dy
      target:setLoc(newX, newY)

      floaterAssigned = self.planetsCluster:tryToAssignFloater(target)
    end

    target.inertia = AHLoop.new(inertiaFunction, stepCount)
    target.inertia:start()
  end

  local updateDeltaTime = function()
    local curTime = self.movementTimer:getTime()
    if (curTime ~= lastTime) then
      dt = curTime - lastTime
      lastTime = curTime
    end
  end

  local updateDeltaPosition = function(target)
      local curX, curY = target:getLoc()
      dx, dy = curX - lastX, curY - lastY
      lastX, lastY = curX, curY
  end

  local tryToNotifyInvalidInput = function()
    if(startX) then
      local distanceDragged = MathUtils.distance(startX, startY, lastX, lastY)
      if(distanceDragged < 3) then
        runInCoroutine(function() self.callbacks.invalidInput() end)
      end
    end
  end

  return function(event, target)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      startDrag(event, target)
    elseif (not started) then
      return
    elseif (event.phase == TouchManager.TOUCH_MOVED) then
      if (not started) then startDrag(event, target) end

      target:setLoc(event.x + dragPointDx, event.y + dragPointDy)
      updateDeltaTime()
      updateDeltaPosition(target)
    elseif (event.phase == TouchManager.TOUCH_ENDED) then
      -- Don't assign the floater to a planet on drop.
      -- Intertia always applies, and if there is no distance to be travelled,
      -- And the floater is close enough to the planet, it'll be assigned to the planet as part of the inertia
      target:deselect()
      updateDeltaTime()
      doInertia(target, dx, dy, dt)
      tryToNotifyInvalidInput()

      dragPointDx, dragPointDy = nil, nil
      lastTime = nil
      lastX, lastY = nil, nil
      startX, startY = nil, nil
      dx, dy, dt = nil, nil, nil
      started = false
    end
  end
end

local function getConsecutiveArray(n, maxValue)
  local result = {}
  maxValue = maxValue or n
  for i = 0, n - 1 do
    result[i + 1] = i % maxValue + 1
  end
  return result
end

local function chooseSpawnLocationOrder(self, numberOfPlanets, spawnPoints)
  -- The spawn order is defined by the cluster number, and then the spawn location within that cluster
  -- The goal is to avoid spawning 2 objects in the same cluster consecutively, and avoid reusing the same locations
  self.spawnPoints = spawnPoints or LayoutInfo.getSpawnPointsForPlanetCount(numberOfPlanets)
  self.lastClusterIndex = 1
  local totalClusters = #self.spawnPoints
  local clusterOrderings = {}
  local totalSpawnLocations = 0
  for i = 1, totalClusters do
    local pointsInCluster = #self.spawnPoints[i]
    local clusterOrder = getConsecutiveArray(pointsInCluster)
    TableUtils.shuffleArray(clusterOrder)
    clusterOrderings[i] = clusterOrder
    clusterOrder.spawnIndex = 1 -- Which location to choose inside this ordering
    totalSpawnLocations = totalSpawnLocations + pointsInCluster
  end

  -- Don't shuffle this array - shift it maybe, but no shuffling
  -- We want each new spawn to be in a different cluster, rather than in a random cluster
  local clusterSequence = getConsecutiveArray(totalSpawnLocations, totalClusters)

  self.spawnClusterOrder, self.spawnLocationOrder = clusterSequence, clusterOrderings
end

local function chooseSpawnLocation(self)
  local clusterIndex = self.spawnClusterOrder[self.lastClusterIndex]
  self.lastClusterIndex = self.lastClusterIndex % #self.spawnClusterOrder + 1

  local spawnOrderInCluster = self.spawnLocationOrder[clusterIndex]
  local spawnPointIndex = spawnOrderInCluster[spawnOrderInCluster.spawnIndex]
  local cluster = self.spawnPoints[clusterIndex]

  spawnOrderInCluster.spawnIndex = spawnOrderInCluster.spawnIndex % #spawnOrderInCluster + 1

  local spawnPoint = cluster[spawnPointIndex]
  local center = spawnPoint.center

  local radius = math.random() * spawnPoint.radius
  local angle = math.random() * math.pi * 2
  local dx, dy = math.cos(angle) * radius, math.sin(angle) * radius

  return center[1] + dx, center[2] + dy
end

local function findFloaterByPhrase(self, phraseInfo)
  for _, floater in pairs(self.allFloaters) do
    if (floater.phraseInfo == phraseInfo) then
      return floater
    end
  end
end

local function removeFloaterByPhrase(self, phraseInfo)
  local floater = findFloaterByPhrase(self, phraseInfo)
  removeFloater(self, floater)
end

local function chooseBonusFloaterSpawnPoints(self, bonusRoundInfo)
  local numberOfPlanets = #bonusRoundInfo.topics
  local spawnPoints = LayoutInfo.getBonusSpawnPointsForPlanetCount(numberOfPlanets)
  chooseSpawnLocationOrder(self, numberOfPlanets, spawnPoints)
  for _, spawnInfo in pairs(self.spawnLocationOrder) do
    spawnInfo.spawnIndex = 1
    spawnInfo[1] = 1
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OrbitScene:setRound(roundInfo)
    if (self.planetsCluster) then
    self.planetsCluster:removeSelf()
  end
  self.planetsCluster = PlanetsCluster.new(roundInfo.topics, self.callbacks)
  self:insertAtCenter(self.planetsCluster)

  chooseSpawnLocationOrder(self, #roundInfo.topics)
end

function OrbitScene:spawnPhrase(phraseInfo)
  local text = phraseInfo["phrase"]

  local planetLocations = self.planetsCluster:getColorAnchors()
  local radius = LayoutInfo.getFloaterRadius()
  local phraseView = FloatingObject.new(text, radius, planetLocations)
  local x, y = chooseSpawnLocation(self)
  phraseView:setLoc(x, y)
  phraseView.spawnX, phraseView.spawnY = x, y
  phraseView:setMovementBounds(radius, radius, Screen.width - radius, Screen.height - radius)

  self:insert(phraseView)
  TouchManager.addTouchHandler(phraseView, self.dragHandler, TouchHandler.BEHAVIOR_FOCUS)

  phraseView.phraseInfo = phraseInfo
  table.insert(self.allFloaters, phraseView)

  if(#self.allFloaters >= 2 and not self.inputReceived) then
    -- Expect input by the time the 4th floater comes in
    AHDelayed.new(2, function()
      if(self.inputReceived) then return end
      self.callbacks.invalidInput()
    end):start()
  end

  phraseView:appear()
  phraseView:wobble()
end

function OrbitScene:assignPhraseToTopic(phraseInfo, topic)
  removeFloaterByPhrase(self, phraseInfo)
  self.planetsCluster:addLayerToPlanet(topic, phraseInfo.phrase)
end

function OrbitScene:assignBonusPhraseToTopic(phraseInfo)
  removeFloaterByPhrase(self, phraseInfo)
end

function OrbitScene:assignLastPhraseToTopic(phraseInfo, topic)
  removeFloaterByPhrase(self, phraseInfo)
  self.planetsCluster:addLastLayerToPlanet(topic, phraseInfo.phrase)
end

function OrbitScene:victoriouslyCollapseTopic(topic)
  self.planetsCluster:collapsePlanet(topic)
end

function OrbitScene:rejectPhraseFromLabelBucket(phraseInfo)
  local floater = findFloaterByPhrase(self, phraseInfo)
  floater:shake()
  stopFloaterMovement(self, floater)

  local waitInPlace = AHDelay.new(0.1)
  local moveFloaterOut = AHLocationTo.new(floater, 0.8, MOAIEaseType.EASE_IN, floater.spawnX, floater.spawnY)
  AHSequence.new(waitInPlace, moveFloaterOut):start()
end

function OrbitScene:playRejectionAnimationForTopic(topic)
  self.planetsCluster:playRejectionAnimation(topic)
end

function OrbitScene:startBonusRound(bonusRoundInfo)
  local overlay = FilledRectangle.newFullScreen({0, 0, 0, 0})
  self:insertAtCenter(overlay)

  local function spawnOneBonusFloater(floaterInfo)
      self:spawnPhrase(floaterInfo)
  end

  local function spawnOneBonusFloaterWithDelay(floaterInfo, delay)
    local delayedFunction = function() spawnOneBonusFloater(floaterInfo) end
    local delayedFunctionCall = AHDelayed.new(delay, delayedFunction)
    delayedFunctionCall:start()
  end

  local function spawnBonusFloaters()
    local delayPerIndex = 0.08
    chooseBonusFloaterSpawnPoints(self, bonusRoundInfo)
    for index, floaterInfo in pairs(bonusRoundInfo.phrases) do
      spawnOneBonusFloaterWithDelay(floaterInfo, delayPerIndex * (index - 1))
    end
  end

  local font = Font.newWithActualSize(FontFaces.BONUS_TRANSITION_FONT, 16.0)
  local bonusLabel = TextStrobeIn.new("MEMORY CHALLENGE", font, BONUS_ROUND_STROBE_EFFECT_COLORS, LAST_BONUS_TEXT_STROBE_COLOR)
  bonusLabel:setLoc(Screen.getWidthPart(0.5), bonusLabel:getHeightPart(0.5))
  local animTime = 0.8
  local delayTime = 0.12

  local function spawnBonusFloatersWithDelay()
    local delayedFunction = function()
      -- Hide the "MEMORY CHALLENGE" sign before spawning floaters
      AHColorTo.new(overlay, 0.8, nil, {0, 0, 0, 0}):start()
      AHColorTo.new(bonusLabel, 0.8, nil, {0, 0, 0, 0}):start()
      local delayedFloatersSpawn = AHDelayed.new(0.38, spawnBonusFloaters)
      delayedFloatersSpawn:start()
    end
    local delayedFunctionCall = AHDelayed.new(0.2, delayedFunction)
    delayedFunctionCall:start()
  end
  local bonusRoundStarter = function()
    AHColorTo.new(overlay, animTime, nil, {0, 0, 0, 0.6}):start()
    bonusLabel:start(animTime, spawnBonusFloatersWithDelay)
  end
  local bonusRoundSoundEffect = function()
    Sound.playSoundEffect("focus_memory_challenge")
  end

  local delayedBonusRoundStarter = AHDelayed.new(delayTime, bonusRoundStarter)
  delayedBonusRoundStarter:start()

  local delayedBonusRoundSoundEffect = AHDelayed.new(delayTime + 0.5, bonusRoundSoundEffect)
  delayedBonusRoundSoundEffect:start()

  self:insert(bonusLabel)
  self.planetsCluster:enableAllPlanets()
end

function OrbitScene:destroyFloater(phraseInfo)
  local floater = findFloaterByPhrase(self, phraseInfo)
  removeFloater(self, floater)
end

function OrbitScene:showTip(continueHandler)
  if (self.showingTip) then
    return
  end

  self.showingTip = true

  local tip = OrbitTip.new()
  self:insertAtCenter(tip)

  local secondsBeforeTipCanBeDismissed = 1
  waitFor(secondsBeforeTipCanBeDismissed)

  local tapToContinueOverlay = TapToContinue.new()
  tapToContinueOverlay:waitForTrigger()

  local remove = tip:createRemoveAction()
  wait(remove:start())

  if(continueHandler) then continueHandler() end

  self.showingTip = false
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function OrbitScene.new(uiOverlay, callbacks)
  local self = Group.newFullScreen()
  createLayout(self)
  Mixin.mix(self, OrbitScene)
  self.uiOverlay = uiOverlay
  self.dragHandler = getDragHandler(self)
  self.allFloaters = {}
  self.callbacks = callbacks

  self.movementTimer = MOAITimer.new()
  self.movementTimer:setSpan(math.huge)
  self.movementTimer:start()

  return self
end
