--- OrbitTip
-- @author Created by Alexey Chernikov

class("OrbitTip")

require("scene.planets.LayeredPlanet")
require("scene.floaters.FloatingObject")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PLANET_COLOR = ColorUtils.colorHex24ToStd(0xbfbfbf)
local PLANET_TEXT_COLOR = {0, 0, 0, 1}
local FLOATER_TEXT_COLOR = {1, 1, 1, 1}
local COVER_COLOR = {0, 0, 0, 0.92}
local INSTRUCTIONS_COLOR = {1, 1, 1, 1}
local TIP_COLOR = {1, 1, 1, 0.5}
local TAP_TO_CONTINUE_COLOR = {1, 1, 1, 0.36}
local PLANET_RADIUS = 42
local FLOATER_RADIUS = 36
local DISTANCE_TO_PLANET = 20
local PLANET_NAME = "SUBJECT"
local TERM_NAME = "fact"
local INSTRUCTIONS_TEXT = "Drag facts to their subjects as they appear"
local TIP_TEXT = "Don't wait until the end of the audio clip"
local TAP_TO_CONTINUE_TEXT = Localizer.get("Misc.TAP_TO_CONTINUE")
local KNOWN_TARGET_WIDTH = 640
local FLOATER_REFERENCE_POINTS = {
  {
    location = {Screen.getWidthPart(0.236), Screen.getHeightPart(0.78)},
    color = {1, 1, 1, 1},
  },
}
local FLOATER_THICKNESS = 4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getScaledFontSize(size)
  local screenScale = Screen.width / KNOWN_TARGET_WIDTH
  local adjustedFontSize = size * screenScale

  return MathUtils.round(adjustedFontSize)
end

local function createLayout(self)
  local layerColors = {PLANET_COLOR, PLANET_COLOR, PLANET_TEXT_COLOR}
  local colorReferencePoints = FLOATER_REFERENCE_POINTS

  local planet = LayeredPlanet.new(layerColors, PLANET_RADIUS, PLANET_NAME)
  local floater = FloatingObject.new(TERM_NAME, FLOATER_RADIUS, colorReferencePoints, FLOATER_THICKNESS)
  local cover = FilledRectangle.new(Screen.width, Screen.height, COVER_COLOR)

  local fontSize = getScaledFontSize(30)
  local font = Font.newWithActualSize(FontFaces.DIN_OT_MEDIUM, fontSize)
  local instructionsLabel = Label.new(INSTRUCTIONS_TEXT, font)
  local tipLabel = Label.new(TIP_TEXT, font)
  local tapToContinueLabel = Label.new(TAP_TO_CONTINUE_TEXT, font)

  floater.view.label:setColor(unpack(FLOATER_TEXT_COLOR))
  instructionsLabel:setColor(unpack(INSTRUCTIONS_COLOR))
  tipLabel:setColor(unpack(TIP_COLOR))
  tapToContinueLabel:setColor(unpack(TAP_TO_CONTINUE_COLOR))

  local centerX = Screen.getWidthPart(0.5)
  local x0, y0 = centerX, Screen.getHeightPart(0.625)
  local floaterX, floaterY = x0 + FLOATER_RADIUS + DISTANCE_TO_PLANET, y0 + FLOATER_RADIUS * 1.5 + DISTANCE_TO_PLANET
  local planetX, planetY = x0 - FLOATER_RADIUS, y0 - FLOATER_RADIUS
  local instructionsY = planetY - Screen.getHeightPart(0.25)
  local tipY = instructionsY - Screen.getHeightPart(0.05)
  local tapToContinueY = tapToContinueLabel:getHeight() + Screen.getHeightPart(0.01)

  floater:setLoc(floaterX, floaterY)
  planet:setLoc(planetX, planetY)
  instructionsLabel:setLoc(centerX, instructionsY)
  tipLabel:setLoc(centerX, tipY)
  tapToContinueLabel:setLoc(centerX, tapToContinueY)

  self:insertAtCenter(cover)
  self:insert(planet)
  self:insert(floater)
  self:insert(instructionsLabel)
  self:insert(tipLabel)
  self:insert(tapToContinueLabel)

  floater:setVisibleRecursive(false)

  self:linkColorAttrToChildren(true)

  self.planet = planet
  self.floater = floater
end

local function startAnimation(self)
  local planetX, planetY = self.planet:getLoc()
  local floaterX, floaterY = self.floater:getLoc()
  local animTime = 0.5
  local ease = AHEaseSimpleIn.new(2)
  local getConsumed = function()
    self.floater:setVisibleRecursive(false)
    self.planet:scaleCoreUpAndDown()
  end

  local appear = AHOpacityTo.new(self, 0.3, nil, 1)
  local showFloater = function()
    self.floater:setLoc(floaterX, floaterY)
    self.floater:setVisibleRecursive(true)
    self.floater:appear()
  end
  local init = AHSequence.new(showFloater, AHDelay.new(1.0))
  local move = AHLocationTo.new(self.floater, animTime, ease, planetX + PLANET_RADIUS/3, planetY + PLANET_RADIUS/3)
  local finish = AHDelay.new(0.6)

  local hintSequence = AHSequence.new(init, move, getConsumed, finish)
  local hintLoop = AHLoop.new(hintSequence, math.huge)

  self.animation = AHSequence.new(appear, hintLoop):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OrbitTip:createRemoveAction()
  local remove = AHRemove.new(self)
  local hide = AHOpacityTo.new(self, 0.1, nil, 0)
  local stopAnimation = function()
    if (self.animation ~= nil) then
      self.animation:stop()
    end
  end

  return AHSequence.new(hide, remove, stopAnimation)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OrbitTip.new()
  local self = Group.new(Screen.width, Screen.height)
  Mixin.mix(self, OrbitTip)
  createLayout(self)
  self:setOpacity(0)
  startAnimation(self)

  return self
end
