--- TextInColoredCircle
-- @author Created by Alexey Chernikov

class("TextInColoredCircle")

require("graphics.MulticolorCircle")
require("scene.text.CircleBoundedText")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getColorsFromReferencePoints(colorReferencePoints)
  local colors = {}
  for _, referencePoint in ipairs(colorReferencePoints) do
    table.insert(colors, referencePoint.color)
  end
  return colors
end

local function getAngleForReferencePoint(x0, y0, referencePoint)
  local x, y = referencePoint[1] - x0, referencePoint[2] - y0
  local angle = math.atan2(y, x) - math.atan2(1, 0)

  while(angle < 0) do
    angle = angle + math.pi * 2
  end
  angle = angle + math.pi / 2

  return angle
end

local function getDistributionPointForAngle(angle)
  return angle / (math.pi * 2)
end

local function findIndexOfMinElement(array, startIndex)
  startIndex = startIndex or 1
  local minElement, indexOfMin = math.huge, -1
  for index = startIndex, #array do
    local value = array[index]
    if (value < minElement) then
      minElement = value
      indexOfMin = index
    end
  end
  return indexOfMin
end

local function swap(array, index1, index2)
  local temp = array[index2]
  array[index2] = array[index1]
  array[index1] = temp
end

local function sortDistributionAndRearrangeColors(self, distribution, newColors)
  for index = 1, #distribution - 1 do
    local curMinIndex = findIndexOfMinElement(distribution, index)
    swap(distribution, index, curMinIndex)
    swap(newColors, index, curMinIndex)
  end

  local startAngle = distribution[1]
  for i, distributionPoint in ipairs(distribution) do
    distribution[i] = distributionPoint - startAngle
  end
  return distribution, newColors, startAngle
end

local function tintColors(x0, y0, colorReferencePoints, colors)
  local distances = {}
  for _, referencePoint in ipairs(colorReferencePoints) do
    local point = referencePoint.location
    local dx, dy = point[1] - x0, point[2] - y0
    local distance = math.sqrt(dx * dx + dy * dy)
    table.insert(distances, distance)
  end
  local minDistanceIndex = findIndexOfMinElement(distances)
  local minDistance = distances[minDistanceIndex]
  local closestColor = colorReferencePoints[minDistanceIndex].color
  local maxDistance = 200
  for index, color in ipairs(colors) do
    local distance = distances[index]
    local normalizedDistance = math.min((distance - minDistance)/maxDistance, 1)
    local r = color[1] * (1 - normalizedDistance) + closestColor[1] * normalizedDistance
    local g = color[2] * (1 - normalizedDistance) + closestColor[2] * normalizedDistance
    local b = color[3] * (1 - normalizedDistance) + closestColor[3] * normalizedDistance
    local a = color[4] * (1 - normalizedDistance) + closestColor[4] * normalizedDistance
    color[1], color[2], color[3], color[4] = r, g, b, a
  end
end

local function figureOutColorDistribution(self, colorReferencePoints)
  local x0, y0 = self:getWorldLoc()
  x0, y0 = self.x or x0, self.y or y0
  local distribution = self.colorDistribution

  local newColors = TableUtils.deepCopy(self.originalColors)
  tintColors(x0, y0, colorReferencePoints, newColors)

  for i, referencePoint in ipairs(colorReferencePoints) do
    local angle = getAngleForReferencePoint(x0, y0, referencePoint.location)
    local distributionPoint = getDistributionPointForAngle(angle)
    distribution[i] = distributionPoint
  end
  local distribution, newColors, startAngle = sortDistributionAndRearrangeColors(self, distribution, newColors)
  return distribution, newColors, startAngle
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, text, radius, colorReferencePoints, thickness, quality)
  local thickness = thickness or 2
  local innerRadius = radius - thickness
  local outerRadius = radius

  local colors = getColorsFromReferencePoints(colorReferencePoints)
  local colorOrdering = TableUtils.deepCopy(colors)
  local textRadius = radius - thickness - 2
  local font = Font.newWithActualSize(FontFaces.FLOATER_FONT, 15.0)
  local label = CircleBoundedText.new(text, textRadius, font)
  label:setColor(unpack(FLOATER_TEXT_COLOR))

  local quality = quality or 36
  local distribution = {0, 0.5}
  local colorCircle = MulticolorCircle.new(innerRadius, outerRadius, colorOrdering, quality, nil, nil, distribution)
  self:insertAtCenter(colorCircle)
  self:insertAtCenter(label)
  self.circle = colorCircle
  self.radius = radius
  self.label = label
  self.originalColors = colors
end

-- Constructor ------------------------

--- Makes a circle that calculates its gradient based on external reference points
-- colorReferencePoints = {{location = {x, y}, color = {r, g, b, a}}, ..., {location = {x, y}, color = {r, g, b, a}}}
function TextInColoredCircle.new(text, radius, colorReferencePoints, thickness, quality)
  local diameter = radius * 2.0
  local self = Group.new(diameter, diameter)
  Mixin.mix(self, TextInColoredCircle)
  self.colorDistribution = {}
  self.colorReferencePoints = colorReferencePoints
  createLayout(self, text, radius, colorReferencePoints, thickness, quality)

  return self
end

function TextInColoredCircle:setLoc(x, y)
  self.super.setLoc(self, x, y)
  local newX, newY = self:getWorldLoc()
  self:updateGradient(newX, newY)
end

function TextInColoredCircle:updateGradient(x, y)
  self.x = x
  self.y = y
  local distribution, newColors, startAngle = figureOutColorDistribution(self, self.colorReferencePoints)
  self.circle:setColorDistribution(distribution, newColors)
  self.circle:setRot(startAngle * 360.0)
end
