--- LayeredPlanet
-- @author Created by Alexey Chernikov

class("LayeredPlanet")

require("scene.text.CircularText")
require("scene.planets.LayerView")
require("scene.text.CircleBoundedText")
require("animations.RadioWaveEffect")
require("animations.TriangleRingExplosion")
require("animations.CircularParticleExplosion")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_HAPPY_PARTICLE_EFFECT = 10
local Z_RADIO_WAVE_EFFECT = 20
local Z_VIEW_GROUP = 30
local Z_PARTICLE_EFFECT = 40

local TOP_LAYER_Z_INDEX = 1000

local MAX_LAYERS = 20

local COMPLETE_PARTICLE_COLOR = COMPLETE_LAYER_COLORS[1]
local COMPLETE_PARTICLE_FADEOUT_COLOR = {unpack(COMPLETE_PARTICLE_COLOR)}
COMPLETE_PARTICLE_FADEOUT_COLOR[4] = 0

local COMPLETE_TEXT_COLOR = COMPLETE_LAYER_COLORS[3]

local WHITE = {1, 1, 1, 1}

local VISIBLE_ORBIT_COLOR = COMPLETE_LAYER_COLORS[1]
VISIBLE_ORBIT_COLOR[4] = 1

local INVISIBLE_ORBIT_COLOR = {unpack(VISIBLE_ORBIT_COLOR)}
INVISIBLE_ORBIT_COLOR[4] = 0

local CIRCULAR_TEXT_FONT_SIZE = 10
local THICKNESS_PER_LAYER = CIRCULAR_TEXT_FONT_SIZE + 2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function dimColorNormalized(color, index, maxIndex)
  local r, g, b, a = unpack(color)
  a = a or 1
  local normalizedIndex = index/maxIndex
  r = r * 1 / (1 - normalizedIndex)
  g = g * 1 / (1 - normalizedIndex)
  b = b * 1 / (1 - normalizedIndex)
  return {r, g, b, a}
end

local function getColorForIndex(self, index, colors)
  colors = colors or self.layerColors
  local color = colors[index]
  return color or dimColorNormalized(colors[#colors], index, MAX_LAYERS)
end

local function createLayout(self, layerColors, radius, name)
  local coreColor = getColorForIndex(self, 1, layerColors)

  local core = FilledCircle.new(radius, coreColor)
  local orbit = HollowCircle.new(radius, coreColor)
  local glow = FilledCircle.new(radius, coreColor)

  local font = Font.newWithActualSize(StandardUIBuilder.STANDARD_FONT, 14.0)
  local label = CircleBoundedText.new(string.upper(name), radius, font)
  self.currentZIndex = TOP_LAYER_Z_INDEX
  self.currentRadius = radius

  local diameter = radius * 2.0
  local viewGroup = Group.new(diameter, diameter)

  core:insertAtCenter(label)

  viewGroup:insertAtCenter(glow)
  viewGroup:insertAtCenter(orbit)
  viewGroup:insertAtCenter(core, self.currentZIndex)

  self:insertAtCenter(viewGroup, Z_VIEW_GROUP)
  self.viewGroup = viewGroup

  core.index = 0
  self.layers = {core}
  self.core = core
  self.orbit = orbit
  self.glow = glow
  self.label = label

  self.labelColor = layerColors[3]
  label:setColor(unpack(self.labelColor))
end

local function rotateIndefinitely(layer, index)
  local normalizedIndex = index / MAX_LAYERS
  local degreesPerSecond = 100.0 * math.sqrt(normalizedIndex, 0.5)
  local intervalToAvoidJumpyAnimation = 10000.0
  local degreesPerInterval = degreesPerSecond * intervalToAvoidJumpyAnimation

  local animation = AHLoop.new(AHRotationBy.new(layer, intervalToAvoidJumpyAnimation, nil, degreesPerInterval))
  animation:start()
  layer.animation = animation
end

local function getCorrectColorForIndex(self, index)
  local color = COMPLETE_LAYER_COLORS[index] or COMPLETE_LAYER_COLORS[#COMPLETE_LAYER_COLORS]
  return dimColorNormalized(color, index, MAX_LAYERS)
end

local function createLayer(self, label)
  self.currentZIndex = self.currentZIndex - 1
  self.currentRadius = self.currentRadius + THICKNESS_PER_LAYER
  local fontSize = CIRCULAR_TEXT_FONT_SIZE

  local layer = LayerView.new(self.currentRadius, WHITE)
  local font = Font.newWithActualSize(FontFaces.RING_FONT, fontSize)
  layer:setText(label, font)

  return layer
end

local function cancelAllAnimations(self)
  self.orbit:setVisible(false)
  self.glow:setVisible(false)
  for _, animation in ipairs(self.animations) do
    animation:stop()
  end
  self.animations = {}
end

local function rotateToTop(self, layer, time)
  time = time or 0.6
  layer.animation:stop()
  local layerSize = layer:getSizeInDegrees()
  AHRotationTo.new(layer, time, MOAIEaseType.EASE_IN, layerSize / 2 + 86):start()
end

local function startHappyOrbitAnimation(self)
  local duration = 0.3

  self.orbit:setVisible(true)
  self.orbit:setColor(unpack(VISIBLE_ORBIT_COLOR))

  local startStrokeThickness = 3
  local orbitStartRadius = self.currentRadius - THICKNESS_PER_LAYER
  local orbitTargetRadius = self.currentRadius + THICKNESS_PER_LAYER
  local startScale = orbitStartRadius/self.coreRadius
  local targetScale = orbitTargetRadius/self.coreRadius

  self.orbit:setScl(startScale, startScale)

  local orbitScale = AHScaleTo.new(self.orbit, duration, nil, targetScale, targetScale)
  local fadeDelay = AHDelay.new(duration / 2)
  local orbitFade = AHColorTo.new(self.orbit, duration / 2, nil, INVISIBLE_ORBIT_COLOR)
  local waitThenFade = AHSequence.new(fadeDelay, orbitFade)
  local orbitAnim = AHSpawn.new(waitThenFade, orbitScale)

  local updatedLineWidth = startStrokeThickness
  local decrementLineWidthCount = math.floor(duration / Screen.stepDuration)
  local deltaLineWidth = startStrokeThickness / decrementLineWidthCount
  local decrementLineWidthOnce = function()
    updatedLineWidth = math.max(0.0, updatedLineWidth - deltaLineWidth)
    self.orbit:setLineWidth(updatedLineWidth)
  end

  local decrementLineWidth = AHLoop.new(decrementLineWidthOnce, decrementLineWidthCount)
  local cleanUp = function() self.orbit:setVisible(false) end

  local decrementAndCleanUp = AHSequence.new(decrementLineWidth, cleanUp)
  decrementAndCleanUp:start()
  table.insert(self.animations, decrementAndCleanUp)

  orbitAnim:start()
  table.insert(self.animations, orbitAnim)
end

local function showTriangularParticleEffect(self, animTime)
  local startRadius = self.coreRadius * 0.8
  local endRadius = self.currentRadius * 1.2
  local particlesCount = 12
  local startRadiusRandomness = THICKNESS_PER_LAYER
  local startColor = COMPLETE_PARTICLE_COLOR
  local endColor = COMPLETE_PARTICLE_FADEOUT_COLOR
  local particleWidth = 6
  local particleHeight = 9.6
  local particleEffect = TriangleRingExplosion.new(startRadius, endRadius, particlesCount, startRadiusRandomness, particleWidth, particleHeight)
  particleEffect:start(animTime, startColor, endColor)
  self:insertAtCenter(particleEffect, Z_PARTICLE_EFFECT)
end

local function strobeLayers(self, colorSets)
  local delayPerLayer = 0.03
  local secondsPerStrobe = 0.05
  local animTimeInStrobe = secondsPerStrobe
  local strobeCount = #colorSets

  for index, layer in ipairs(self.layers) do
    local delay = delayPerLayer

    local strobeOneLayer = function(tickNumber)
      local colorToSet = colorSets[(tickNumber - 1) % #colorSets + 1]
      local color = getColorForIndex(self, index, colorToSet)
      AHColorTo.new(layer, animTimeInStrobe, nil, color):start()
    end

    local tickNumber = 1
    strobeOneLayer(tickNumber)

    local delayedFunctionCall = AHDelayed.new(delay, function()
      local delayBetweenStrobes = secondsPerStrobe / strobeCount
      local strobeOneLayerWrapper = function()
        tickNumber = tickNumber + 1
        strobeOneLayer(tickNumber)
      end
      local waitAndStrobe = AHSpawn.new(AHDelay.new(delayBetweenStrobes), strobeOneLayerWrapper)
      local waitAndStrobeRepeat = AHLoop.new(waitAndStrobe, strobeCount)
      waitAndStrobeRepeat:start()
    end)
    delayedFunctionCall:start()
  end
end

local function playHappyColorAnimationForLayer(self, layer)
  local startColor = getCorrectColorForIndex(self, layer.index + 1)
  local targetColor = getColorForIndex(self, layer.index + 1)
  local toTargetColor = AHColorTo.new(layer, 0.22, nil, targetColor)
  local toStartColor = AHColorTo.new(layer, 0.22, nil, startColor)
  local shine = AHSequence.new(toStartColor, toTargetColor)

  shine:start()
  table.insert(self.animations, shine)
end

local function playHappyColorAnimationForAllLayers(self)
  for _, layer in ipairs(self.layers) do
    playHappyColorAnimationForLayer(self, layer)
  end
end

local function playCorrectAnimationForLayer(self, layer)
  layer:setScl((self.currentRadius - THICKNESS_PER_LAYER)/self.currentRadius)

  local scaleUp = AHScaleTo.new(layer, 0.3, nil, 1, 1)
  scaleUp:start()
  table.insert(self.animations, scaleUp)
end

local function prepareForOutermostLayerAnimation(self)
  local layer = self.layers[#self.layers]
  layer:setScl((self.currentRadius - THICKNESS_PER_LAYER) / self.currentRadius * 0.9)
end

local function playCorrectAnimationForOutermostLayer(self)
  local layer = self.layers[#self.layers]
  playCorrectAnimationForLayer(self, layer)

  local delayedFunctionCall = AHDelayed.new(0, function()
    startHappyOrbitAnimation(self)
  end)
  delayedFunctionCall:start()
end

local function scaleUpAndDown(object, scaleMagnitudeMultiplier)
  local dsDown = 0.1 * scaleMagnitudeMultiplier
  local dsUp = 0.13 * scaleMagnitudeMultiplier
  local downScale = 1 - dsDown
  local upScale = 1 + dsUp

  local anticipate = AHScaleTo.new(object, 0.08, MOAIEaseType.SOFT_EASE_IN, downScale, downScale)
  local up = AHScaleTo.new(object, 0.16, nil, upScale, upScale)
  local down = AHScaleTo.new(object, 0.16, MOAIEaseType.EASE_IN, 1, 1)

  local animation = AHSequence.new(anticipate, up, down)
  animation:start()
  return animation

end

local function scaleAllLayersThenShowOutermostLayer(self)
  local delayPerLayer = 0.03
  -- Decrease wobble:
  local maxLayersInAnimation = math.max(5, #self.layers)
  local scaleMagnitudeAmplitude = 1
  for i = 1, #self.layers - 1 do
    local normalizedIndex = (i - 1) / maxLayersInAnimation
    local layer = self.layers[i]

    local delayedFunctionCall = AHDelayed.new((i - 1) * delayPerLayer, function()
      local animation = scaleUpAndDown(layer, 1 - normalizedIndex * scaleMagnitudeAmplitude)
      table.insert(self.animations, animation)
    end)
    delayedFunctionCall:start()
  end

  prepareForOutermostLayerAnimation(self)
  local outerRingSpawnDelay = 0.22 + delayPerLayer * (#self.layers - 1)
  local delayedFunctionCall = AHDelayed.new(outerRingSpawnDelay, function()
    playCorrectAnimationForOutermostLayer(self)
  end)
  delayedFunctionCall:start()
end

local function colorThenRestore(self, object, targetColor, restoreColor, changeTime, keepTime)
  local applyColor = AHColorTo.new(object, changeTime, nil, targetColor)
  local keep = AHDelay.new(keepTime)
  local restoreColorAction = AHColorTo.new(object, changeTime, nil, restoreColor)
  local animation = AHSequence.new(applyColor, keep, restoreColorAction)
  animation:start()

  table.insert(self.animations, animation)
end

local function addHappyParticles(self)
  local speed = 12
  local acceleration = 5
  local innerRadius = 0
  local colors = nil
  local duration = 0.8
  local particleLifeTime = duration

  local effect = CircularParticleExplosion.new(speed, acceleration, innerRadius, colors)
  effect:setParticleLifeTime(particleLifeTime)

  self:insertAtCenter(effect, Z_HAPPY_PARTICLE_EFFECT)

  effect:start(duration, function(effect) effect:removeSelf() end)
end

local function scaleAndColorTwiceThenRestore(self, object, downScale, upScale, targetColor, originalColor)
  targetColor[4] = targetColor[4] or 1
  originalColor[4] = originalColor[4] or 1

  local colorAnim = AHSequence.new(
    AHColorTo.new(object, 0.3, nil, targetColor),
    AHColorTo.new(object, 0.2, nil, originalColor)
  )
  colorAnim:start()

  local scaleAnimation = AHSequence.new(
    AHScaleTo.new(object, 0.1, nil, downScale),
    AHScaleTo.new(object, 0.1, nil, upScale),
    AHScaleTo.new(object, 0.1, nil, downScale),
    AHScaleTo.new(object, 0.1, nil, 1)
  )
  scaleAnimation:start()

  table.insert(self.animations, colorAnim)
  table.insert(self.animations, scaleAnimation)
end

local function getArrayWithStartValueAndOffsets(startValue, offset, count)
  local result = {startValue}
  for i = 1, count - 1 do
    table.insert(result, startValue + offset * i)
  end
  return result
end

local function collapseLayer(layer, delay, duration, scaleUpTime)
  local waitBeforeScaling = AHDelay.new(delay)
  local scaleUp = AHScaleTo.new(layer, scaleUpTime, MOAIEaseType.SOFT_EASE_OUT, 1.1, 1.1)
  local scaleDown = AHScaleTo.new(layer, duration, MOAIEaseType.SOFT_EASE_OUT, 0, 0)
  local animation = AHSequence.new(scaleUp, waitBeforeScaling, scaleDown)
  animation:start()
end

local function collapseCore(self)
  local core = self.layers[1]

  local initialScaleUpTime = 0.15
  local waitAfterInitialScaleUpTime = 0.15
  local slightScaleTime = 0.15
  local waitAfterSlightScaleTime = 0.01
  local explodeTime = 0.25
  local waitAfterExplodeTime = 0.3
  local collapseTime = 0.3

  local scaleUp = AHScaleTo.new(core, initialScaleUpTime, MOAIEaseType.SOFT_EASE_OUT, 1.1, 1.1)
  local waitAfterInitialScaleUp = AHDelay.new(waitAfterInitialScaleUpTime)
  local slightScaleDown = AHScaleTo.new(core, slightScaleTime, nil, 0.8, 0.8)
  local waitAfterScaleDown = AHDelay.new(waitAfterSlightScaleTime)

  local explode = AHScaleTo.new(core, explodeTime, MOAIEaseType.EASE_IN, 1.3, 1.3)

  local timeBeforeExplosion = initialScaleUpTime + waitAfterInitialScaleUpTime + slightScaleTime + waitAfterSlightScaleTime
  local particlesDelay = AHDelay.new(timeBeforeExplosion - 0.05)
  local particleExplode = AHCallLogic.new(function() addHappyParticles(self) end)
  local particleEffect = AHSequence.new(particlesDelay, particleExplode)

  local waitAfterExplode = AHDelay.new(waitAfterExplodeTime)

  local collapseCore = AHScaleTo.new(core, collapseTime, MOAIEaseType.EASE_OUT, 0, 0)
  local collapseOrbit = AHScaleTo.new(self.orbit, collapseTime + 0.08, MOAIEaseType.EASE_OUT, 0, 0)
  local showOrbit = AHCallLogic.new(function()
    self.orbit:setScl(1.3, 1.3)
    self.orbit:setVisible(true)
    self.orbit:setColor(unpack(COMPLETE_LAYER_COLORS[1]))
  end)
  local collapse = AHSpawn.new(collapseCore, showOrbit, collapseOrbit)

  local coreAnim = AHSequence.new(scaleUp, waitAfterInitialScaleUp, slightScaleDown, waitAfterScaleDown, explode, waitAfterExplode, collapse)
  coreAnim = AHSpawn.new(coreAnim, particleEffect)
  coreAnim:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function LayeredPlanet:appear(delay)
  self:setScl(0, 0)
  local wait = AHDelay.new(delay)
  local scale = AHScaleTo.new(self, 1, MOAIEaseType.EASE_IN, 1, 1)
  local appear = AHSequence.new(wait, scale)
  appear:start()
end

function LayeredPlanet:stopAcceptingAnswers()
  self.disabled = true
end

function LayeredPlanet:startAcceptingAnswers()
  self.disabled = false
end

function LayeredPlanet:getName()
  return self.name
end

function LayeredPlanet:addLayer(label)
  if (self.radioWaves) then
    self.radioWaves:stop()
  end
  local layer = createLayer(self, label)

  layer.index = self.layerCount
  table.insert(self.layers, layer)
  self.layerCount = self.layerCount + 1

  self.viewGroup:insertAtCenter(layer, self.currentZIndex)
  rotateIndefinitely(layer, self.layerCount)

  local layerColor = getColorForIndex(self, layer.index + 1)
  layer:setColor(unpack(layerColor))
end

function LayeredPlanet:playHappyAnimation()
  cancelAllAnimations(self)
  scaleAllLayersThenShowOutermostLayer(self)
  playHappyColorAnimationForAllLayers(self)
  colorThenRestore(self, self.label, COMPLETE_TEXT_COLOR, self.labelColor, 0.1, 0.2)
end

function LayeredPlanet:scaleCoreUpAndDown()
  cancelAllAnimations(self)
  scaleUpAndDown(self.layers[1], 0.6)
end

function LayeredPlanet:setHappyColors()
  cancelAllAnimations(self)
  local animTime = 0.5
  local particleEffectTime = 0.6

  scaleAllLayersThenShowOutermostLayer(self)
  for index, layer in ipairs(self.layers) do
    if (index > 1) then
      -- Keep the core where it is that is
      rotateToTop(self, layer, animTime)
    end
  end
  self.labelColor = COMPLETE_TEXT_COLOR
  table.insert(self.animations, AHColorTo.new(self.label, 0.3, nil, COMPLETE_TEXT_COLOR):start())

  local delayedShowTriangleParticles = AHDelayed.new(0.3, function()
    showTriangularParticleEffect(self, particleEffectTime)
  end)
  delayedShowTriangleParticles:start()

  local strobeColorSets = COMPLETE_LAYER_STROBE_COLOR_SETS

  local delayedShowStrobeLayers = AHDelayed.new(0.2, function()
    strobeLayers(self, strobeColorSets)
  end)
  delayedShowStrobeLayers:start()

  self.layerColors = COMPLETE_LAYER_COLORS
end

function LayeredPlanet:playSadAnimation()
  cancelAllAnimations(self)

  local glowColor = getColorForIndex(self, #self.layers + 1, WRONG_ANSWER_LAYER_COLORS)
  local glowHideColor = TableUtils.deepCopy(glowColor)
  glowHideColor[4] = 0

  self.glow:setColor(unpack(glowColor))
  self.orbit:setColor(unpack(glowColor))
  self.orbit:setVisible(true)
  self.glow:setVisible(true)

  colorThenRestore(self, self.label, WRONG_ANSWER_LAYER_COLORS[3], self.labelColor, 0.1, 0.3)

  local layerScaleDownTimes = getArrayWithStartValueAndOffsets(0.4, 0.01, #self.layers)
  local layerScaleUpTimes = getArrayWithStartValueAndOffsets(0.5, 0.01, #self.layers)
  local maxScale = 0.3
  local delayPerLayer = 0.01
  for index, layer in ipairs(self.layers) do
    local targetColor = getColorForIndex(self, index, WRONG_ANSWER_LAYER_COLORS)
    local originalColor = getColorForIndex(self, index)

    local downScale = 1 - maxScale * (1 - (index - 1) / #self.layers)
    local upScale = 1

    local delayedScaleAndColor = AHDelayed.new((index - 1) * delayPerLayer, function()
      scaleAndColorTwiceThenRestore(self, layer, downScale, upScale, targetColor, originalColor)
    end)
    delayedScaleAndColor:start()
  end

  local scaleToOuterLayer = self.currentRadius / self.coreRadius
  local glowStartScale = scaleToOuterLayer * 1.1
  local glowExpandScale = glowStartScale * 1.1
  local downTime, upTime = layerScaleDownTimes[#layerScaleDownTimes], layerScaleUpTimes[#layerScaleUpTimes]

  scaleAndColorTwiceThenRestore(self, self.glow, glowExpandScale, glowStartScale, glowColor, glowHideColor)

  local orbitStartScale = glowStartScale * 1.03
  local orbitExpandScale = orbitStartScale * 1.22
  local orbitStrokeWidth = 3
  local orbitDownTime = downTime * 1.15
  local orbitUpTime = upTime * 1.15

  scaleAndColorTwiceThenRestore(self, self.orbit, orbitExpandScale, orbitStartScale, glowColor, glowHideColor)

  local decrementLineWidthDuration = (orbitDownTime + orbitUpTime) * 2
  local decrementLineWidthCount = math.floor(decrementLineWidthDuration / Screen.stepDuration)
  local deltaLineWidth = orbitStrokeWidth / decrementLineWidthCount
  local updatedLineWidth = orbitStrokeWidth
  local decrementLineWidthOnce = function()
    updatedLineWidth = math.max(0.0, updatedLineWidth - deltaLineWidth)
    self.orbit:setLineWidth(updatedLineWidth)
  end
  local decrementLineWidth = AHLoop.new(decrementLineWidthOnce, decrementLineWidthCount)
  decrementLineWidth:start()
end

function LayeredPlanet:collapse()
  local scaleUpTime = 0.15
  local scaleDownTime = 0.4
  local delayPerLayer = scaleDownTime / 8
  for index = 2, #self.layers do
    local layer = self.layers[index]
    local delay = delayPerLayer * (index - 2)
    collapseLayer(layer, delay, scaleDownTime, scaleUpTime)
  end

  collapseCore(self)
  self.collapsed = true
end

function LayeredPlanet:canAcceptAnswers()
  return not self.collapsed and not self.disabled
end

function LayeredPlanet:wobble(delay)
  local effect = RadioWaveEffect.new(3)
  self.viewGroup:insertAtCenter(effect, Z_RADIO_WAVE_EFFECT)
  effect:setColor(unpack(getColorForIndex(self, 1)))
  local timePerWobble = 6
  local timePerRadioSignalBurst = timePerWobble / 2
  local timeForRadioSignalAnimation = timePerRadioSignalBurst / 2

  local startThickness = 0.25
  local endThickness = 1.2

  local yAmplitude = 4
  local phaseOffset = delay or math.random() * math.pi * 2
  local centerY = self:getHeightPart(0.5)

  local thetaRadians = 0.0
  local deltaThetaRadians = 0.02

  local updateViewGroupY = function()
    local yOffset = yAmplitude * math.sin(thetaRadians + phaseOffset)
    self.viewGroup:setY(centerY + yOffset)

    thetaRadians = thetaRadians + deltaThetaRadians
    if (thetaRadians > 2.0 * math.pi) then
      thetaRadians = 0.0
    end
  end

  local wobble = AHLoop.new(updateViewGroupY)
  wobble:start()

  local doRadioWaves = function()
    local startRadius = self.currentRadius + 2 * THICKNESS_PER_LAYER
    local endRadius =  self.currentRadius
    effect:start(timeForRadioSignalAnimation, startRadius, endRadius, startThickness, endThickness)
  end

  local delayedRadioWaves = AHDelayed.new(phaseOffset / math.pi, function()
    if (#self.layers > 1) then return end
    doRadioWaves()

    local waitThenDoRadioWaves = AHDelayed.new(timePerWobble, doRadioWaves)
    self.radioWaves = AHLoop.new(waitThenDoRadioWaves)
    self.radioWaves:start()
  end)
  delayedRadioWaves:start()
end

function LayeredPlanet:getDropRadius()
  -- Alternatively, to use a constant radius, get it from LayoutInfo.getDropRadius()
  return math.min(self.currentRadius, self.coreRadius + THICKNESS_PER_LAYER * 2)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function LayeredPlanet.new(layerColors, radius, name)
  name = name or ""
  local diameter = radius * 2.0
  local self = Group.new(diameter, diameter)
  Mixin.mix(self, LayeredPlanet)

  self.layerCount = 1
  self.coreRadius = radius
  self.name = name
  self.layerColors = layerColors

  createLayout(self, layerColors, radius, name)
  self.animations = {}

  return self
end
