--- PlanetsCluster
-- @author Created by Alexey Chernikov

class("PlanetsCluster")

require("scene.LayoutInfo")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createPlanets(self, topics)
  -- Add orbiter points
  local planets = {}
  local locations = LayoutInfo.getLayoutForPlanetCount(#topics)
  self.planetLocations = locations
  local wobbleDelayPerIndex = 1
  local appearDelayPerIndex = 0.1

  local planetTouchHandler = function(event)
    if(event.phase == TouchManager.TOUCH_BEGAN) then
      runInCoroutine(function() self.callbacks.invalidInput() end)
    end
  end

  local planetRadius = LayoutInfo.getPlanetRadius()

  for i, topic in ipairs(topics) do
    local pointInfo = locations[i]
    local planet = LayeredPlanet.new(pointInfo.layerColors, planetRadius, topic)
    planet:setLoc(pointInfo.location[1], pointInfo.location[2])
    self:insert(planet)
    planets[topic] = planet

    local wobbleDelay = wobbleDelayPerIndex * (i - 1)
    local appearDelay = appearDelayPerIndex * (i - 1)
    planet:wobble(wobbleDelay)
    planet:appear(appearDelay)

    TouchManager.addTouchHandler(planet, planetTouchHandler)
  end
  self.planets = planets
end

local function forceDropFloater(self, floater, planet)
  local answerData = {
    phraseInfo = floater.phraseInfo,
    assignedTopic = planet:getName()
  }

  self.callbacks.answerSubmitted(answerData)

  if (self.dropHandler) then
    self.dropHandler(floater, planet)
  end
end

local function tryToDropFloaterToPlanet(self, floater, planet)
  if (not planet:canAcceptAnswers()) then return false end

  local x0, y0 = floater:getLoc()
  local xp, yp = planet:getLoc()
  local distance = MathUtils.distance(x0, y0, xp, yp)
  local floaterRadius = floater:getRadius()
  distance = math.abs(distance - floaterRadius * 0.3)

  if (distance < planet:getDropRadius()) then
    forceDropFloater(self, floater, planet)
    return true
  end
  return false
end

local function findPlanetByName(self, name)
  for _, planet in pairs(self.planets) do
    if (planet:getName() == name) then
      return planet
    end
  end
end

local function getPlanetAndAddLayer(self, planetName, layerName)
  local planet = findPlanetByName(self, planetName)
  planet:addLayer(layerName)
  return planet
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PlanetsCluster:setDropHandler(handler)
  self.dropHandler = handler
end

--- Used to let floating objects know how to recolor themselves based on planet positions
function PlanetsCluster:getColorAnchors()
  return self.planetLocations
end

function PlanetsCluster:tryToAssignFloater(floater)
  for topic, planet in pairs(self.planets) do
    local dropSuccessful = tryToDropFloaterToPlanet(self, floater, planet)
    if (dropSuccessful) then
      return true
    end
  end
end

function PlanetsCluster:playRejectionAnimation(planetName)
  local planet = findPlanetByName(self, planetName)
  planet:playSadAnimation()
end

function PlanetsCluster:addLayerToPlanet(planetName, layerName)
  local planet = getPlanetAndAddLayer(self, planetName, layerName)
  planet:playHappyAnimation()
end

function PlanetsCluster:addLastLayerToPlanet(planetName, layerName)
  local planet = getPlanetAndAddLayer(self, planetName, layerName)
  planet:setHappyColors()
  planet:stopAcceptingAnswers()
end

function PlanetsCluster:enableAllPlanets()
  for _, planet in pairs(self.planets) do
    planet:startAcceptingAnswers()
  end
end

function PlanetsCluster:collapsePlanet(planetName)
  local planet = findPlanetByName(self, planetName)
  planet:collapse()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PlanetsCluster.new(topics, callbacks)
  local self = Group.newFullScreen()
  Mixin.mix(self, PlanetsCluster)
  self.callbacks = callbacks
  createPlanets(self, topics)
  return self
end
