--- CircleBoundedText
-- @author Created by Alexey Chernikov

class("CircleBoundedText")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

CircleBoundedText.WORD_MAX_CHARACTERS = 13

-- "hello worldy blabidy bla blahsona besome mucho gracias si"
-- 5 6 7 3 8 6 5 7 2
-- 5 11 18 21 29 35 40 47 49
-- 5
-- 13
-- 11
-- 11
-- 9
-- If there's 1, 3 or 5 lines, position the middle word in the exact center
local ODD_LINES_CHARACTER_COUNTS =
{
  CircleBoundedText.WORD_MAX_CHARACTERS - 5,        ---
  CircleBoundedText.WORD_MAX_CHARACTERS - 1,      -------
  CircleBoundedText.WORD_MAX_CHARACTERS,         ---------
  CircleBoundedText.WORD_MAX_CHARACTERS - 1,      -------
  CircleBoundedText.WORD_MAX_CHARACTERS - 5         ---
}
-- If there's 2 or 4 lines, position them offset from the center. This changes the character count
local EVEN_LINES_CHARACTER_COUNTS =
{
  CircleBoundedText.WORD_MAX_CHARACTERS - 1,    ------
  CircleBoundedText.WORD_MAX_CHARACTERS,      ----------
  --                                            CENTER   <=== Unused space smaller than may appear in comments
  CircleBoundedText.WORD_MAX_CHARACTERS,      ----------
  CircleBoundedText.WORD_MAX_CHARACTERS - 1     ------
}

local ODD_LINE_INDICES_TO_TRY =
{
  {3},
  {2, 3, 4},
  {1, 2, 3, 4, 5}
}
local EVEN_LINE_INDICES_TO_TRY =
{
  {2, 3},
  {1, 2, 3, 4}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function tryToGetLineCountForWordsPlacedOnLines(words, characterCountsPerLine, indicesSet)
  local currentWordIndex = 1
  local lineCount = 1
  local wordIndicesPerLine = {}
  for _, characterCountIndex in ipairs(indicesSet) do
    local maxCharacters = characterCountsPerLine[characterCountIndex]
    local charactersOnCurrentLine = 0
    local curLineWords = {}
    table.insert(wordIndicesPerLine, curLineWords)

    -- First word comes w/o space
    local currentWordLength = string.len(words[currentWordIndex])
    charactersOnCurrentLine = charactersOnCurrentLine + currentWordLength
    table.insert(curLineWords, currentWordIndex)
    while(charactersOnCurrentLine <= maxCharacters and currentWordIndex < #words) do
      charactersOnCurrentLine = charactersOnCurrentLine + 2 -- add space
      currentWordIndex = currentWordIndex + 1
      local currentWordLength = string.len(words[currentWordIndex])
      charactersOnCurrentLine = charactersOnCurrentLine + currentWordLength
      table.insert(curLineWords, currentWordIndex)
    end

    if (charactersOnCurrentLine <= maxCharacters and currentWordIndex >= #words) then
      -- It fits! For whatever reason, we didn't overfill the line margin, so no need to transfer line
      -- This should imply that currentWordIndex >= #words
      return lineCount, wordIndicesPerLine
    end

    lineCount = lineCount + 1
    curLineWords[#curLineWords] = nil


  end
  if (currentWordIndex < #words) then
    -- it don't fit - we haven't reached the end of the word stream
    return nil
  end
end

local function getLineCountForLengthsAndLimits(words, characterCountsPerLine, indicesToTry)
  for _, indicesSet in ipairs(indicesToTry) do
    local lineCount, currentWordIndicesPerLine = tryToGetLineCountForWordsPlacedOnLines(words, characterCountsPerLine, indicesSet)
    if (lineCount and #indicesSet == lineCount) then
      -- Found the perfect match!
      return lineCount, currentWordIndicesPerLine, indicesSet
    end
  end
end

local function createAndPositionWords(self, words, font)
  local wordsOnEachLine = CircleBoundedText.getOptimalWordByLineDistribution(words)

  local fontSize = font:getActualSize()

  local lineSpacing = fontSize * 0.25
  local h = fontSize + lineSpacing
  local xSpacing = fontSize / 2.0
  local totalHeight = 0
  local width = -math.huge
  local lines = {}
  for lineIndex, wordIndicesOnLine in ipairs(wordsOnEachLine) do
    local x = 0
    local labels = {}
    for _, wordIndex in ipairs(wordIndicesOnLine) do
      local label = Label.new(words[wordIndex], font)

      local labelX = x + xSpacing * 0.5
      label:setX(labelX)

      label:setAnchorPointBottomLeft()
      label:inheritColor(self)

      table.insert(labels, label)

      x = x + label:getWidth() + xSpacing
    end

    local currentLine = Group.new(x, h)
    for _, label in ipairs(labels) do
      currentLine:insert(label)
    end

    x = x - xSpacing
    currentLine:setLoc(0, -(lineIndex - 1) * h - fontSize / 2)
    currentLine:setAnchorPointCenter()
    totalHeight = totalHeight + h
    width = math.max(width, x)
    table.insert(lines, currentLine)
  end
  totalHeight = totalHeight - lineSpacing

  local result = Group.new(width, totalHeight)
  for _, line in ipairs(lines) do
    line:setX(line:getX() + width * 0.5)
    line:setY(line:getY() + totalHeight)
    result:insert(line)
  end

  local outerGroup = Group.new(self:getWidth(), self:getHeight())
  outerGroup:insertAtCenter(result)

  return outerGroup
end

local function createLayout(self, text, font)
  local words = StringUtils.split(text, " ")
  local view = createAndPositionWords(self, words, font)

  self:insertAtCenter(view)
  self.view = view
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function CircleBoundedText.getOptimalWordByLineDistribution(words)
  if (#words == 2) then
    -- When there's 2 words, place them on two lines
    return {{1}, {2}}, {2, 3}
  end
  if (#words == 3) then
    -- A little riskier, helps deal with ORBIT-026, for "GREAT BARRIER REEF"
    -- If it breaks other cases, try hardcoding that one instance
    return {{1}, {2}, {3}}, {2, 3, 4}
  end

  local lineCount, wordsOnEachLine, indicesSet = getLineCountForLengthsAndLimits(words, ODD_LINES_CHARACTER_COUNTS, ODD_LINE_INDICES_TO_TRY)
  if (lineCount) then
  else
    lineCount, wordsOnEachLine, indicesSet = getLineCountForLengthsAndLimits(words, EVEN_LINES_CHARACTER_COUNTS, EVEN_LINE_INDICES_TO_TRY)
  end

  assert((wordsOnEachLine ~= nil), "Could not find an optimal word by line distribution for the words.")

  return wordsOnEachLine, indicesSet
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function CircleBoundedText.new(text, radius, font)
  local diameter = radius * 2.0
  local self = Group.new(diameter, diameter)
  Mixin.mix(self, CircleBoundedText)

  createLayout(self, text, font)

  return self
end
