--- OrbitValidator
-- @author Created by Anton Morant on April 17, 2014.
-- ConceptValidator for Orbit.

class("OrbitValidator")

require("validators.ConceptValidator")
require("scene.text.CircleBoundedText")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'phrases',
  'topic_1',
  'topic_2',
  'topic_3',
  'sound'
}

local REQUIRED_PHRASE_FIELDS = {
  'text',
  'category',
  'round',
  'time_said' -- Not used in the game, but content requested validation
}

local REQUIRED_EMPTY_ALLOWED_PHRASE_FIELDS = {
  'spawn_time'
}

local ALLOWED_ROUND_VALUES = {
  "phrases",
  "bonus"
}

local MINIMUM_PHRASES = 12
local MAXIMUM_PHRASES = 18
local MINIMUM_PHRASES_PER_TOPIC = 4
local MAXIMUM_PHRASES_PER_TOPIC = 6

-- based on the word "CHALLENGER" in ORBIT-034
local MAXIMUM_TOPIC_LENGTH = 10

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validatePhrases(self, phrases, topics)
  for _, phrase in ipairs(phrases) do
    self:expectHasKeys(phrase, REQUIRED_PHRASE_FIELDS)
    self:expectHasKeys(phrase, REQUIRED_EMPTY_ALLOWED_PHRASE_FIELDS, true)

    self:expectIsContained(phrase.category, topics, "category")
    self:expectIsContained(phrase.round, ALLOWED_ROUND_VALUES, "round")

    local spawnTime = phrase.spawn_time
    local saidTime = phrase.time_said
    if phrase.round == "phrases" then
      self:expectNotEmpty(spawnTime, string.format("spawn_time (phrase: %s)", phrase.text))
      self:expectGreaterOrEqual(tonumber(spawnTime), tonumber(saidTime), "spawn_time should be after time_said")
    else
      self:expectTrue(spawnTime == "", string.format("spawn_time should be empty for bonus round (phrase: %s)", phrase.text))
    end

    self:expectStringIsPrintable(phrase.text, "phrase.text")

    local words = StringUtils.split(phrase.text, " ")
    for _, word in ipairs(words) do
      self:expectLessOrEqual(#word, CircleBoundedText.WORD_MAX_CHARACTERS, string.format("Phrase contained a word that exceeded the max characters: %s", word))
    end

    -- Will throw an assertion, internally, if an optimal line distribution cannot be found
    CircleBoundedText.getOptimalWordByLineDistribution(words)
  end
end

local function validateNumberOfPhrases(self, phrases, topics)
  self:expectNumberInRange(#phrases, MINIMUM_PHRASES, MAXIMUM_PHRASES, "number of phrases")

  local phrasesPerTopic = {}
  for _, topic in ipairs(topics) do
    phrasesPerTopic[topic] = 0
  end

  for _, phrase in ipairs(phrases) do
    phrasesPerTopic[phrase.category] = phrasesPerTopic[phrase.category] + 1
  end

  for _, topic in ipairs(topics) do
    self:expectNumberInRange(phrasesPerTopic[topic], MINIMUM_PHRASES_PER_TOPIC, MAXIMUM_PHRASES_PER_TOPIC, string.format("phrases for topic %s", topic))
  end
end

local function validateBonusRound(self, phrases, topics)
  local topicsFound = {}

  for _, phrase in ipairs(phrases) do
    if phrase.round == "bonus" then
      local topic = phrase.category
      self:assertFalse(topicsFound[topic], string.format("More than one bonus phrase found for topic %s", topic))
      topicsFound[topic] = true
    end
  end

  for _, topic in ipairs(topics) do
    self:assertTrue(topicsFound[topic], string.format("No bonus phrase found for topic %s", topic))
  end
end

local function validateUppercaseTopics(self, topics)
  for _, topic in ipairs(topics) do
    local isUppercase = StringUtils.isUppercase(topic:sub(1, 1))
    self:expectTrue(isUppercase, string.format("Topic is not upper-case: %s", topic))
  end
end

local function validateTopicMaximumLengths(self, topics)
  for _, topic in ipairs(topics) do
    topic = string.gsub(topic, "-", " ")
    local pieces = StringUtils.split(topic, " ")
    for _, piece in ipairs(pieces) do
      local isSufficientlyShort = (#piece <= MAXIMUM_TOPIC_LENGTH)
      self:expectTrue(isSufficientlyShort, string.format("Topic is too long: %s", topic))
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function OrbitValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local topics = { concept.topic_1, concept.topic_2, concept.topic_3 }
  local phrases = concept.phrases

  validatePhrases(self, phrases, topics)
  validateNumberOfPhrases(self, phrases, topics)
  validateBonusRound(self, phrases, topics)
  validateUppercaseTopics(self, topics)
  validateTopicMaximumLengths(self, topics)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function OrbitValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, OrbitValidator)

  return self
end
