--- ParagraphsGame
-- @author Template created by Aaron Barrett on September 27, 2013.

class("ParagraphsGame")

-- Needed for proper scaling on iPad
FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("ParagraphsCommon")
require("ParagraphsConfiguration")
require("ParagraphsConceptGenerator")
require("ParagraphsScene")
require("bonuses.ResponseSpeedTracker")

------------------------------------------------------------------------------
-- Configuration
------------------------------------------------------------------------------

local UI_COLOR = {0.4, 0.4, 0.4, 0.75}

------------------------------------------------------------------------------
-- Utility
------------------------------------------------------------------------------

local function configureGameDifficulty()
  local rankTwoScore = ParagraphsConfiguration.getRankTwoScore()
  local rankThreeScore = ParagraphsConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

local function updateLivesOrScore(self, isCorrect)
  if (isCorrect) then
    self.responseSpeedTracker:notifyCorrectResponse()
    GameSession.incrementCorrectAnswerCount()
    GameSession.increaseScore(ParagraphsConfiguration.getCorrectAnswerScore(GameSession.getCorrectAnswerCount()))
    self.scoreLabel:setScore(GameSession:getScore())
  else
    self.responseSpeedTracker:notifyIncorrectResponse()
    GameSession.incrementIncorrectAnswerCount()
    self.livesCounter:decrement()
  end
end

local function onResponse(self, answerButton)
  local correctAnswerSelected = answerButton.isCorrect
  updateLivesOrScore(self, correctAnswerSelected)

  self.scene:setScrollLayerTouchEnabled(false)

  local responseAction
  if (correctAnswerSelected) then
    Sound.playSoundEffect("paragraphs_correct")
    GameSession.contentTracker:notifyCorrectAnswer(self.concept.id, self.currentQuestionType)
    responseAction = self.scene:getCorrectAction(answerButton)
  else
    Sound.playSoundEffect("paragraphs_incorrect")
    GameSession.contentTracker:notifyIncorrectAnswer(self.concept.id, self.currentQuestionType)
    responseAction = self.scene:getIncorrectAction(answerButton)
  end

  wait(responseAction:start())
  self.hasSubmittedResponse = true
end

local function onBonusResponse(self, answerButton, correctAnswerButton)
  local responseAction
  if (answerButton.isCorrect) then
    Sound.playSoundEffect("paragraphs_correct")
    GameSession.contentTracker:notifyCorrectAnswer(self.concept.id, self.currentQuestionType)
    GameSession.incrementCorrectAnswerCount()
    responseAction = answerButton:createCorrectResponseAction(true)

    local bonusPoints = ParagraphsConfiguration.getCorrectAnswerScore(GameSession.getCorrectAnswerCount())
    local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.EXTRA_QUESTION_BONUS)
    GameSession.increaseScore(bonusPoints)
    self.bonusTray:indicateSimpleTextBonus(notificationString, bonusPoints)
    self.scoreLabel:setScore(GameSession:getScore())
  else
    Sound.playSoundEffect("paragraphs_incorrect")
    GameSession.contentTracker:notifyIncorrectAnswer(self.concept.id, self.currentQuestionType, answerButton:getText())
    GameSession.incrementIncorrectAnswerCount()
    responseAction = AHSpawn.new(
      answerButton:createIncorrectResponseAction(),
      correctAnswerButton:createCorrectResponseAction()
    )
  end

  wait(responseAction:start())
  self.hasSubmittedResponse = true
end

local function waitForResponse(self)
  waitUntil(function()
    return self.hasSubmittedResponse
  end)
  self.hasSubmittedResponse = false
end

local function startNextQuestion(self)
  local questionTab = self.scene:getNextQuestion()
  self.currentQuestionType = questionTab.questionType
  questionTab.hasBeenActivated = true
  questionTab:bringToBack()
  Sound.playSoundEffect("paragraphs_icon_tap")
  self.scene:startQuestion(questionTab)
  self.scene:setScrollLayerTouchEnabled(true)

  local bonusWindowDuration = ParagraphsConfiguration.getSecondsForSpeedBonus(self.currentQuestionType)
  local decayWindowStart = ParagraphsConfiguration.getSpeedBonusDecayStart(self.currentQuestionType)
  self.responseSpeedTracker:setBonusWindowDuration(bonusWindowDuration)
  self.responseSpeedTracker:setDecayWindowStart(decayWindowStart)
  self.responseSpeedTracker:notifyStart()
end

local function checkIfOutOfLives(self)
  if (self.livesCounter:getLives() == 0)then
    GameSession.notifyGameLoss()
  end
end

local function showAndWaitForBonusRound(self)
  Sound.playSoundEffect("paragraphs_bonus_round_transition")
  wait(self.scene:getShowBonusAction():start())

  waitForResponse(self)
end

------------------------------------------------------------------------------
-- API
------------------------------------------------------------------------------

-- gameFlow ---------------------------

function ParagraphsGame:gameFlow()
  self.scene:waitForFirstQuestion()
  local numberOfQuestions = ParagraphsConfiguration.QUESTIONS_COUNT

  for questionIndex = 1, numberOfQuestions do
    startNextQuestion(self)
    waitForResponse(self)
    self.scene:waitWhileTextAnimationInProgess()

    checkIfOutOfLives(self)
  end

  showAndWaitForBonusRound(self)

  -- NOTE: Wait for sound and animations to settle before game ends
  waitFor(1.0)
  GameSession.notifyGameWin()
end

------------------------------------------------------------------------------
-- Object Creation
------------------------------------------------------------------------------

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(ParagraphsConfiguration.NUMBER_OF_LIVES)
end

local function createGameElements(self)
  local conceptGenerator = ParagraphsConceptGenerator.new()
  local concept = conceptGenerator:next()
  self.concept = concept

  GameSession.contentTracker:notifyPresentedConcept(concept.id)

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 ParagraphsConfiguration.POINTS_FOR_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(ParagraphsConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       ParagraphsConfiguration.SPEED_BONUS_DECAY_START)

  local onResponseCallback = function(answerButton)
    onResponse(self, answerButton)
  end

  local bonusResponseCallback = function(answerButton, correctAnswerButton)
    onBonusResponse(self, answerButton, correctAnswerButton)
  end

  local scene = ParagraphsScene.new(concept, onResponseCallback, bonusResponseCallback)
  self:insertAtCenter(scene)
  self.scene = scene
end

-- Constructor ------------------------

function ParagraphsGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, ParagraphsGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.hasSubmittedResponse = false
  self.currentQuestionType = nil

  return self
end
