--- ParagraphsQuestionTab
-- @author created by Steven Mah on April 5, 2017

class("ParagraphsQuestionTab")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INACTIVE_WIDTH = Screen.getWidthPart(0.33)
local INACTIVE_HEIGHT = ParagraphsCommon.QUESTION_MENU_HEIGHT

local ACTIVE_WIDTH = Screen.width
local ACTIVE_HEIGHT = ParagraphsCommon.ACTIVE_QUESTION_HEIGHT

local MENU_BUTTON_STYLES_BY_QUESTION_TYPE = ParagraphsCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE

local INCORRECT_BACKGROUND_COLOR = {0.96, 0.22, 0.46}
local INCORRECT_ICON_COLOR = {0.52, 0.05, 0.14, 1.0}
local CORRECT_BACKGROUND_COLOR = {0.17, 0.88, 0.55}
local CORRECT_ICON_COLOR = {0.05, 0.4, 0.24, 1.0}

local TAB_ACTION_COMPONENT_DURATION = 0.15

local Z_INDICES = {
  OVERLAY = 10,
  ICON = 20
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createMoveAndScaleDownAction(self)
  local xAnchorPoint, _ = self:getAnchorPoint()
  local ease = AHEaseSimpleInOut.new(4.0)

  local fullAction = AHSequence.new(
    AHSpawn.new(
      AHLocationYBy.new(self, TAB_ACTION_COMPONENT_DURATION * 2, ease, self:getHeightPart(0.5)),
      AHScaleTo.new(self, TAB_ACTION_COMPONENT_DURATION * 2, nil, 1.0, 1.0)
    ),
    function() self:transformAnchorPoint(xAnchorPoint, 0.0) end,
    AHLocationYTo.new(self, TAB_ACTION_COMPONENT_DURATION, ease, self:getHeightPart(-0.25)),
    AHLocationYTo.new(self, TAB_ACTION_COMPONENT_DURATION, ease, 0.0)
  )

  return fullAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ParagraphsQuestionTab:createMoveAndScaleUpAction()
  local xAnchorPoint, _ = self:getAnchorPoint()
  self:transformAnchorPoint(xAnchorPoint, 1.0)

  local destinationHeight = ACTIVE_HEIGHT

  if (self.questionType == ParagraphsCommon.QUESTION_TYPES.MOVE) then
    destinationHeight = ParagraphsCommon.ACTIVE_MOVE_QUESTION_HEIGHT
  end

  local width, height = self:getDims()
  local xScale = ACTIVE_WIDTH / width
  local yScale = destinationHeight / height
  local ease = AHEaseSimpleInOut.new(3.0)
  local fullAction = AHSequence.new(
    AHLocationYBy.new(self, TAB_ACTION_COMPONENT_DURATION, ease, self:getHeightPart(-0.25)),
    AHSpawn.new(
      AHOpacityTo.new(self.icon, TAB_ACTION_COMPONENT_DURATION * 2, nil, 0.0),
      AHLocationYTo.new(self, TAB_ACTION_COMPONENT_DURATION * 2, ease, destinationHeight * 1.1)
    ),
    AHSpawn.new(
      AHLocationYTo.new(self, TAB_ACTION_COMPONENT_DURATION * 2, ease, destinationHeight),
      AHScaleTo.new(self, TAB_ACTION_COMPONENT_DURATION * 2, nil, xScale, yScale)
    )
  )

  return fullAction
end

function ParagraphsQuestionTab:createCorrectResponseAction()
  local correctResponseAction = AHSequence.new(
    createMoveAndScaleDownAction(self),
    function() self.icon = self.checkIcon end,
    AHSpawn.new(
      AHColorTo.new(self, TAB_ACTION_COMPONENT_DURATION, nil, CORRECT_BACKGROUND_COLOR),
      AHOpacityTo.new(self.checkIcon, TAB_ACTION_COMPONENT_DURATION, nil, 1.0)
    )
  )

  return correctResponseAction
end

function ParagraphsQuestionTab:createIncorrectResponseAction()
  local incorrectResponseAction = AHSequence.new(
    createMoveAndScaleDownAction(self),
    function() self.icon = self.crossIcon end,
    AHSpawn.new(
      AHColorTo.new(self, TAB_ACTION_COMPONENT_DURATION, nil, INCORRECT_BACKGROUND_COLOR),
      AHOpacityTo.new(self.crossIcon, TAB_ACTION_COMPONENT_DURATION, nil, 1.0)
    )
  )

  return incorrectResponseAction
end

function ParagraphsQuestionTab:createOutroAction()
  local styleData = MENU_BUTTON_STYLES_BY_QUESTION_TYPE[self.questionType]
  local iconColor = {0.0, 0.0, 0.0, 0.2}

  local fullOutroAction = AHSequence.new(
    AHSpawn.new(
      AHColorTo.new(self, TAB_ACTION_COMPONENT_DURATION, nil, styleData.color),
      AHColorTo.new(self.icon, TAB_ACTION_COMPONENT_DURATION, nil, iconColor),
      AHSequence.new(
        AHLocationYBy.new(self, 0.2, AHEaseQuadraticOut.new(), self:getHeightPart(0.2)),
        AHLocationYBy.new(self, 0.2, AHEaseQuadraticIn.new(), -self:getHeightPart(0.2))
      )
    )
  )

  return fullOutroAction
end

function ParagraphsQuestionTab:triggerDownState()
  local width, height = self:getDims()
  local gray = {0.1, 0.1, 0.1}
  local overlay = FilledRectangle.new(width, height, gray)
  overlay:setOpacity(0.0)

  self:insertAtCenter(overlay, Z_INDICES.OVERLAY)

  AHSequence.new(
    AHOpacityTo.new(overlay, 0.1, nil, 0.75),
    AHOpacityTo.new(overlay, 0.5, nil, 0.0),
    AHRemove.new(overlay)
  ):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createIcon(self, spriteName, color)
  local icon = Sprite.newWithName(spriteName)
  if color then
    icon:setColor(unpack(color))
  end
  self:insertAtCenter(icon, Z_INDICES.ICON)
  return icon
end

local function createLayout(self, styleData)
  self.icon = createIcon(self, styleData.iconName)

  self.crossIcon = createIcon(self, "cross.png", INCORRECT_ICON_COLOR)
  self.crossIcon:setOpacity(0.0)

  self.checkIcon = createIcon(self, "checkmark.png", CORRECT_ICON_COLOR)
  self.checkIcon:setOpacity(0.0)
end

-- Constructor ------------------------

function ParagraphsQuestionTab.new(question)
  local styleData = MENU_BUTTON_STYLES_BY_QUESTION_TYPE[question.type]
  assert(styleData, "Style Data is undefined for question type")
  local width = INACTIVE_WIDTH

  local self = FilledRectangle.new(width, INACTIVE_HEIGHT, styleData.color)
  Mixin.mix(self, ParagraphsQuestionTab)

  self.hasBeenActivated = false
  self.questionType = question.type

  createLayout(self, styleData)

  return self
end
