--- ParagraphsQuestionsMenu
-- @author created by Steven Mah on April 5, 2017

class("ParagraphsQuestionsMenu")
require("ParagraphsQuestionTab")
require("ParagraphsQuestionLayout")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.width
local HEIGHT = ParagraphsCommon.QUESTION_MENU_HEIGHT

local SHINE_FILENAME = "button_shine.png"

local SHINE_DURATION = 0.5
local SHINE_WAIT_DURATION = 5
local DELAY_BEFORE_NEXT_QUESTION = 0.5


local Z_INDICES = {
  OUTRO_BARS = 0,
  QUESTION_TABS = 5,
  SHINE = 10,
  QUESTION_LAYOUTS = 15
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createDisableAndHideInactiveButtonsAction(self, selectedButton)
  self.activeQuestionTab = selectedButton

  local fullShiftAction = AHSpawn.new(
    AHLocationYBy.new(self.background, 0.1, nil, -(HEIGHT*0.9)),
    function() self.background:setOpacity(0.0) end
  )

  for _, questionTab in ipairs(self.questionTabs) do
    if (questionTab ~= selectedButton) then
      fullShiftAction:addAction(
        AHLocationYBy.new(questionTab, 0.1, nil, -(HEIGHT*0.9))
      )
    end
  end

  return fullShiftAction
end

local function setQuestionButtonTouch(self, button)
  local firstQuestionSelected = function()
    button:triggerDownState()

    self.shineAnimation:stop()
    self.shineImage:removeSelf()

    self.firstQuestion = button
    self:setTouchEnabled(false)
  end

  Button.makeButton(button, firstQuestionSelected)
  button:setTouchEnabled(false)
end

local function createShineAnimation(self)
  local shineImage = Sprite.newWithName(SHINE_FILENAME)
  shineImage:setEffectiveDims(nil, HEIGHT)
  shineImage:setAnchorPointBottomLeft()

  local imageW = shineImage:getEffectiveWidth()

  shineImage:setX(-imageW)
  self:insert(shineImage, Z_INDICES.SHINE)

  local reset = function()
    shineImage:setLoc(-imageW)
  end

  local delay = AHDelay.new(SHINE_WAIT_DURATION)
  local move = AHLocationXTo.new(shineImage, SHINE_DURATION, nil, WIDTH)
  local resetMove = AHCallLogic.new(function() reset() end)
  local loop = AHLoop.new(AHSequence.new(delay, move, resetMove))
  loop:start()

  self.shineImage = shineImage
  self.shineAnimation = loop
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ParagraphsQuestionsMenu:setTouchEnabled(isEnabled)
  for _, questionTab in ipairs(self.questionTabs) do
    questionTab:setTouchEnabled(isEnabled)
  end
end

function ParagraphsQuestionsMenu:hasFirstQuestionBeenSelected()
  return (self.firstQuestion ~= nil)
end

function ParagraphsQuestionsMenu:getNextQuestion()
  if (self.firstQuestion ~= nil) then
    local firstQuestion = self.firstQuestion
    -- self.firstQuestion should always be nil after this
    self.firstQuestion = nil
    return firstQuestion
  else
    for questionIndex, question in ipairs(self.questionTabs) do
      if (question.hasBeenActivated == false) then
        self.activeQuestionTab = question
        return question
      end
    end

    assert(false, "Attempted to get another question when there were none.")
  end
end

function ParagraphsQuestionsMenu:startQuestion(questionTab)
  local questionLayout = self.questionLayouts[questionTab.questionType]
  questionLayout:setX(Screen.getWidthPart(0.5))
  self:insert(questionLayout, Z_INDICES.QUESTION_LAYOUTS)

  local activationAction = AHSequence.new(
    AHSpawn.new(
      questionTab:createMoveAndScaleUpAction(),
      AHDelayed.new(0.45, createDisableAndHideInactiveButtonsAction(self, questionTab))
    ),
    questionLayout:createFadeInAction(),
    function()
      questionLayout:setTouchEnabled(true)
      self.activeQuestionLayout = questionLayout
    end
  )

  activationAction:start()
end

function ParagraphsQuestionsMenu:getEnableAndShowInactiveButtonsAction()
  local fullShiftAction = AHSpawn.new(
    AHLocationYTo.new(self.background, 0.25, nil, 0.0),
    AHOpacityTo.new(self.background, 0.1, nil, 1.0)
  )

  for _, questionTab in ipairs(self.questionTabs) do
    if (questionTab ~= self.activeQuestionTab) then
      fullShiftAction:addAction(
        AHLocationYTo.new(questionTab, 0.25, nil, 0.0)
      )
    end
  end

  fullShiftAction:addAction(function() self.activeQuestionTab = nil end)

  return AHSequence.new(fullShiftAction, AHDelay.new(DELAY_BEFORE_NEXT_QUESTION))
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, concept)
  local xAnchorPoint = 0.0
  local yAnchorPoint = 0.0
  local xPosition = 0

  local backgroundGroup = Group.new(WIDTH, HEIGHT)
  local leftBacking = FilledRectangle.new(Screen.getWidthPart(0.005), HEIGHT, {1.0, 1.0, 1.0})
  leftBacking:setX(Screen.getWidthPart(0.33))
  leftBacking:setAnchorPointBottomLeft()

  local rightBacking = FilledRectangle.new(Screen.getWidthPart(0.005), HEIGHT, {1.0, 1.0, 1.0})
  rightBacking:setX(Screen.getWidthPart(0.665))
  rightBacking:setAnchorPointBottomLeft()

  backgroundGroup:insert(leftBacking)
  backgroundGroup:insert(rightBacking)
  backgroundGroup:setAnchorPointBottom()
  backgroundGroup:setX(WIDTH / 2)
  self:insert(backgroundGroup)
  backgroundGroup:linkOpacityAttrToChildren()
  self.background = backgroundGroup

  -- Assumes that there will always be three questions
  local questionButtons = {}
  for _, type in ipairs(ParagraphsCommon.QUESTION_ORDER) do
    local round = concept.rounds[type]
    local questionButton = ParagraphsQuestionTab.new(round)

    setQuestionButtonTouch(self, questionButton)

    questionButton:setAnchorPoint(xAnchorPoint, yAnchorPoint)
    questionButton:setX(xPosition)

    self:insert(questionButton, Z_INDICES.QUESTION_TABS)

    xPosition = xPosition + Screen.getWidthPart(0.5)
    xAnchorPoint = xAnchorPoint + 0.5
    table.insert(questionButtons, questionButton)
  end

  self.questionTabs = questionButtons
end

local function createElements(self, concept)
  local questionLayouts = {}
  for _, type in ipairs(ParagraphsCommon.QUESTION_ORDER) do
    local round = concept.rounds[type]
    local questionLayout = ParagraphsQuestionLayout.new(round, self.onResponseCallback)
    questionLayout:setAnchorPointBottom()
    questionLayouts[type] = questionLayout
  end

  self.questionLayouts = questionLayouts

  createShineAnimation(self)
end

-- Constructor ------------------------

function ParagraphsQuestionsMenu.new(concept, onResponseCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, ParagraphsQuestionsMenu)

  self.onResponseCallback = onResponseCallback

  createLayout(self, concept)
  createElements(self, concept)

  self.activeQuestionLayout = nil
  self.firstQuestion = nil
  self:setTouchEnabled(true)

  return self
end
