--- ParagraphsScene
-- @author Created by Steven Mah on March 30, 2017

class("ParagraphsScene")

require("ParagraphsScrollLayer")
require("ParagraphsQuestionsMenu")
require("ParagraphsBonusLayout")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {1.0, 1.0, 1.0}

local Z_INDICES = {
  SCROLL_LAYER = 0,
  QUESTIONS_MENU = 10,
  COVER = 20
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ParagraphsScene:activateQuestion(questionType)
  self.scrollLayer:activateQuestion(questionType)
end

function ParagraphsScene:activateResponseSequence(isCorrect, answerText)
  self.scrollLayer:activateResponseSequence(isCorrect, answerText)
end

function ParagraphsScene:waitForFirstQuestion()
  waitUntil(function()
    return self.questionsMenu:hasFirstQuestionBeenSelected()
  end)
end

function ParagraphsScene:getNextQuestion()
  return self.questionsMenu:getNextQuestion()
end

function ParagraphsScene:startQuestion(questionTab)
  self:activateQuestion(questionTab.questionType)
  waitFor(ParagraphsCommon.TEXT_MOVE_DURATION)
  self.questionsMenu:startQuestion(questionTab)
end

function ParagraphsScene:getCorrectAction(answerButton)
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout
  local questionTab = questionsMenu.activeQuestionTab

  questionLayout:setTouchEnabled(false)

  local fullResponseAction = AHStaggeredSpawn.new(
    ParagraphsCommon.TEXT_MOVE_DURATION,
    answerButton:createCorrectResponseAction(answerButton.isCorrect),
    function()
      self.scrollLayer:activateResponseSequence(true, answerButton:getText())
    end,
    AHSpawn.new(
      questionLayout:createFadeOutAction(),
      questionTab:createCorrectResponseAction(),
      questionsMenu:getEnableAndShowInactiveButtonsAction()
    )
  )

  return fullResponseAction
end

function ParagraphsScene:getIncorrectAction(answerButton)
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout
  local questionTab = questionsMenu.activeQuestionTab

  questionLayout:setTouchEnabled(false)

  local buttonAction
  if (answerButton ~= nil) then
    buttonAction = AHSequence.new(
      AHSpawn.new(
        answerButton:createIncorrectResponseAction(),
        questionLayout.correctAnswer:createCorrectResponseAction(answerButton.isCorrect)
      ),
      AHDelay.new(1.0)
    )
  else
    buttonAction = AHDoNothing.new()
  end

  local fullResponseAction = AHStaggeredSpawn.new(
    ParagraphsCommon.TEXT_MOVE_DURATION,
    buttonAction,
    function()
      self.scrollLayer:activateResponseSequence(false, questionLayout.correctAnswer:getText())
    end,
    AHSpawn.new(
      questionLayout:createFadeOutAction(),
      questionTab:createIncorrectResponseAction(),
      questionsMenu:getEnableAndShowInactiveButtonsAction()
    )
  )

  return fullResponseAction
end

function ParagraphsScene:waitWhileTextAnimationInProgess()
  self.scrollLayer.passageView:waitWhileAnimationInProgess()
end

function ParagraphsScene:getShowBonusAction()
  local yDelta = Screen.height - ParagraphsCommon.QUESTION_MENU_HEIGHT
  local duration = 0.65
  local ease = AHEaseSineInOut.new()
  local shiftUpAction = AHSpawn.new(
    AHLocationYBy.new(self.scrollLayer, duration, ease, yDelta),
    AHLocationYBy.new(self.questionsMenu, duration, ease, yDelta),
    AHLocationYBy.new(self.topCover, duration, ease, yDelta),
    AHLocationYBy.new(self.bonusLayout, duration, ease, yDelta)
  )

  return shiftUpAction
end

function ParagraphsScene:setScrollLayerTouchEnabled(isEnabled)
  if (self.scrollLayer ~= nil) then
    self.scrollLayer:setTouchEnabled(isEnabled)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)
  self.background = background

  local passageText = self.concept.paragraph
  local scrollLayer = ParagraphsScrollLayer.new(passageText)
  scrollLayer:setAnchorPointTop()
  scrollLayer:setLoc(self:getWidthPart(0.5), self:getHeight())
  self:insert(scrollLayer)
  scrollLayer:setZIndex(Z_INDICES.SCROLL_LAYER)
  self.scrollLayer = scrollLayer

  local questionsMenu = ParagraphsQuestionsMenu.new(self.concept, self.onResponseCallback)
  questionsMenu:setAnchorPointBottom()
  self:insertAtBottom(questionsMenu, Z_INDICES.QUESTIONS_MENU)
  self.questionsMenu = questionsMenu

  local coverHeight = ParagraphsCommon.PASSAGE_MARGIN_HEIGHT

  local topCover = FilledRectangle.new(Screen.width, coverHeight, BACKGROUND_COLOR)
  local lineSeparator = FilledRectangle.new(Screen.width, 0.5, {0.5, 0.5, 0.5})
  lineSeparator:setOpacity(0.25)
  topCover:setAnchorPointTop()
  lineSeparator:setAnchorPointBottom()
  topCover:insertAtBottom(lineSeparator)
  self:insertAtTop(topCover, Z_INDICES.COVER)
  self.topCover = topCover

  local bonusLayout = ParagraphsBonusLayout.new(self.concept.bonus, self.onBonusResponseCallback)
  bonusLayout:setAnchorPointTop()
  self:insertAtBottom(bonusLayout)
  self.bonusLayout = bonusLayout
end

-- Constructor ------------------------

function ParagraphsScene.new(concept, onResponseCallback, onBonusResponseCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, ParagraphsScene)

  self.concept = concept
  self.onResponseCallback = onResponseCallback
  self.onBonusResponseCallback = onBonusResponseCallback

  createLayout(self)

  return self
end
