--- ParagraphsScrollLayer
-- @author created by Steven Mah on April 5, 2017

class("ParagraphsScrollLayer")

require("widgets.TouchScroller")
require("ParagraphsCommon")
require("ParagraphsTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local AUTO_SHIFT_PADDING = Screen.getHeightPart(0.04)
local AUTO_SCROLL_DURATION = 0.75

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function autoScroll(self, shiftDistance)
  self.autoScrollAction = self.scroller:newScrollYBy(AUTO_SCROLL_DURATION, AHEaseExponentialOut.new(), shiftDistance)
  self.autoScrollAction:start()
end

local function ensureHighlightedTextVisible(self, questionType)
  local views = self.passageView:getViewsWithAttribute(questionType)
  local layouts = self.passageView:getLayouts()
  local firstView = views[1]
  local lastView = views[#views]
  local firstLayout = layouts[firstView]
  local lastLayout = layouts[lastView]

  --- NOTE: If highlighted text to both be above and below the window, prioritize
  -- the top highlighted item
  local _, topMostY = self:transformLoc(self.passageView, nil, firstLayout:getY() + firstView:getHeightPart(0.5))
  local _, bottomMostY = self:transformLoc(self.passageView, nil, lastLayout:getY() - lastView:getHeightPart(0.5))

  local upperBound = Screen.height - ParagraphsCommon.PASSAGE_MARGIN_HEIGHT - AUTO_SHIFT_PADDING
  local lowerBound = ParagraphsCommon.ACTIVE_QUESTION_HEIGHT + AUTO_SHIFT_PADDING
  if (questionType == ParagraphsCommon.QUESTION_TYPES.MOVE) then
    lowerBound = ParagraphsCommon.ACTIVE_MOVE_QUESTION_HEIGHT
  end

  local isUnderLowerBound = (bottomMostY < lowerBound)
  local isAboveUpperBound = (topMostY > upperBound)
  if (isAboveUpperBound) then
    local shiftDistance = upperBound - topMostY - AUTO_SHIFT_PADDING
    autoScroll(self, shiftDistance)
  elseif (isUnderLowerBound) then
    local passageHeightDelta = self.passageView.layoutBuilder:getTextHeight() - self.passageView:getHeight()
    local shiftDistance = lowerBound - bottomMostY + AUTO_SHIFT_PADDING + passageHeightDelta
    autoScroll(self, shiftDistance)
  end
end

local function passageAutoScrollIfNeeded(self)
  local _, passageY = self.passageView:getCenterWorldLoc()
  local topMostY = passageY + self.passageView:getHeightPart(0.5)
  local bottomMostY = passageY - self.passageView:getHeightPart(0.5)

  local upperMarginHeights = ParagraphsCommon.PASSAGE_MARGIN_HEIGHT + AUTO_SHIFT_PADDING
  local upperBound = self.scroller:getHeight() - upperMarginHeights
  local lowerBound = ParagraphsCommon.ACTIVE_QUESTION_HEIGHT + AUTO_SHIFT_PADDING

  local isTopLineTooLow = (topMostY < upperBound)
  local isBottomLineTooHigh = (bottomMostY > lowerBound)

  if (isTopLineTooLow) then
    local shiftDistance = upperBound - topMostY
    autoScroll(self, shiftDistance)
  elseif (isBottomLineTooHigh) then
    local shiftDistance = lowerBound - bottomMostY
    autoScroll(self, shiftDistance)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function ParagraphsScrollLayer:getOutroTextAction()
  local passageView = self.passageView

  return passageView:getOutroTextAction()
end

function ParagraphsScrollLayer:setTouchEnabled(isEnabled)
  self.scroller:setTouchEnabled(isEnabled)
end

function ParagraphsScrollLayer:activateQuestion(questionType)
  self.activeQuestionType = questionType
  self.passageView:activateQuestion(questionType)
  ensureHighlightedTextVisible(self, questionType)
end

function ParagraphsScrollLayer:activateResponseSequence(isCorrect, answerText)
  self.passageView:activateResponseSequence(isCorrect, answerText)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, passageText)
  local width, height = self:getDims()

  local passageView = ParagraphsTextArea.new(passageText)
  self.passageView = passageView

  local marginHeights = ParagraphsCommon.QUESTION_MENU_HEIGHT +
                        ParagraphsCommon.PASSAGE_MARGIN_HEIGHT +
                        AUTO_SHIFT_PADDING
  local contentsHeight = passageView:getHeight() + marginHeights
  local contents = Group.new(width, contentsHeight)
  passageView:setAnchorPointTop()
  passageView:setLoc(Screen.getSafeX(0.5), passageView:getHeight() + ParagraphsCommon.QUESTION_MENU_HEIGHT)
  contents:insert(passageView)
  self.contents = contents

  local scroller = TouchScroller.new(self:getWidth(), self:getHeight(), self.contents)
  scroller:setInertialScrollingEnabled(false)
  scroller:setScrollBoundsEnabled(false)
  scroller:setWillResetInitialTouchCoordinateUponScrollStart(true)

  local stopScrollIfInProgress = function()
    if (self.autoScrollAction ~= nil) then
      self.autoScrollAction:stop()
      self.autoScrollAction = nil
    end
  end

  local autoScrollWrapper = function()
    passageAutoScrollIfNeeded(self)
  end

  scroller:setScrollNotificationFunctions(stopScrollIfInProgress, autoScrollWrapper)
  self:insertAtCenter(scroller)
  self.scroller = scroller
end

function ParagraphsScrollLayer.new(passageText)
  local self = Group.newFullScreen()
  Mixin.mix(self, ParagraphsScrollLayer)

  self.autoScrollAction = nil
  self.activeQuestionType = nil

  createLayout(self, passageText)

  return self
end
