--- ParagraphsTextArea
-- @author Created by Steven Mah on March 30, 2017

class("ParagraphsTextArea")

require("ParagraphsTextModel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_WIDTH = Screen.getWidthPart(0.9)
local DEFAULT_TEXT_COLOR = {0.3, 0.3, 0.3}

local QUESTION_TYPES = ParagraphsCommon.QUESTION_TYPES

local HIGHLIGHT_COLORS = {
  [QUESTION_TYPES.REPLACE] = { 0.58, 0.38, 0.9 },
  [QUESTION_TYPES.MOVE] = { 0.76, 0.39, 0.76 },
  [QUESTION_TYPES.COMBINE] = { 0.68, 0.36, 0.83 }
}

local HIGHLIGHT_FADE_DURATION = ParagraphsCommon.FADE_DURATION
local HIGHLIGHT_PADDING = 5
local HIGHLIGHT_HEIGHT = Screen.getHeightPart(0.04)

local Y_COMPARISON_PRECISION = 2

local BASE_ATTRIBUTES = {
  font = Font.new(FontFaces.TISA_REGULAR, 20)
}

local LABEL_ATTRIBUTES = {
  r = { replace = true },
  m = { move = true },
  g = { combine = true },
  m1 = { moveIcon = "inline_A.png", itemType = TextItem.TYPE_EMPTY, skipNextSpacing = true },
  m2 = { moveIcon = "inline_B.png", itemType = TextItem.TYPE_EMPTY, skipNextSpacing = true },
  gi = { combineIcon = "inline_combine.png", itemType = TextItem.TYPE_EMPTY, skipNextSpacing = true },
  correct = { src = "inline_check.png", color = { 0.05, 0.4, 0.24	}, backingColor = { 0.17, 0.88, 0.55 }, itemType = TextItem.TYPE_IMAGE, allowBreakLineAfter = false },
  incorrect = { src = "inline_cross.png", color = { 0.52, 0.05, 0.14 }, backingColor = { 0.96, 0.22, 0.46 }, itemType = TextItem.TYPE_IMAGE, allowBreakLineAfter = false }
}

local MOVE_DURATION = ParagraphsCommon.TEXT_MOVE_DURATION
local SHOW_DURATION = ParagraphsCommon.FADE_DURATION
local HIDE_DURATION = ParagraphsCommon.FADE_DURATION
local ADJUST_HEIGHT_DURATION = ParagraphsCommon.TEXT_MOVE_DURATION

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function insertIcons(self, questionType)
  local model = self.textModel

  model:beginTransaction()
  if (questionType == QUESTION_TYPES.MOVE) then
    local items, indices = model:getItemsAndIndicesWithAttributes(questionType)
    local headerIcon = TextItem.newImageItem("inline_move.png", BASE_ATTRIBUTES)
    headerIcon:setAttribute("move", true)
    headerIcon:setAttribute("allowBreakLineAfter", false)
    model:insertItems({headerIcon}, indices[1])

    local destinationItems, destinationIndices = model:getItemsAndIndicesWithAttributes("moveIcon")

    for index, item in ipairs(destinationItems) do
      local icon = TextItem.newImageItem(item:getAttribute("moveIcon"), BASE_ATTRIBUTES)
      icon:setAttribute("move", true)
      icon:setAttribute("locationTag", "m"..index)
      icon:setAttribute("allowBreakLineAfter", false)
      model:replaceItems(destinationIndices[index], 1, {icon})
    end

    self.firstActiveIndex = indices[1]
    self.lastActiveIndex = indices[#indices] + 1 -- NOTE: +1 to account for additional inserted icon
  elseif (questionType == QUESTION_TYPES.COMBINE) then
    local items, indices = model:getItemsAndIndicesWithAttributes("combineIcon")
    local emptyCombineTag, indexOfCombineTag = items[1], indices[1]
    local icon = TextItem.newImageItem(emptyCombineTag:getAttribute("combineIcon"), BASE_ATTRIBUTES)
    icon:setAttribute("combine", true)
    icon:setAttribute("allowBreakLineAfter", false)
    model:replaceItems(indexOfCombineTag, 1, {icon})

    local _, indices = model:getItemsAndIndicesWithAttributes(questionType)
    self.firstActiveIndex = indices[1]
    self.lastActiveIndex = indices[#indices]
  elseif (questionType == QUESTION_TYPES.REPLACE) then
    local items, indices = model:getItemsAndIndicesWithAttributes(questionType)

    local icon = TextItem.newImageItem("inline_replace.png", BASE_ATTRIBUTES)
    icon:setAttribute("replace", true)
    icon:setAttribute("allowBreakLineAfter", false)
    model:insertItems({icon}, indices[1])

    self.firstActiveIndex = indices[1]
    self.lastActiveIndex = indices[#indices] + 1 -- NOTE: +1 to account for additional inserted icon
  else
    error("Invalid question type.")
  end
  local shouldAddHighlights = true
  model:endTransaction(shouldAddHighlights)
end

local function addTextFades(textView, fadeInAction, fadeOutAction)
  -- NOTE: Needed because views in a TextView automatically inherit color
  textView:clearAttrLink(MOAIColor.INHERIT_COLOR)
  textView:setColor(unpack(DEFAULT_TEXT_COLOR))

  fadeInAction:addAction(
    AHColorTo.new(textView, HIGHLIGHT_FADE_DURATION, nil, {1.0, 1.0, 1.0})
  )

  fadeOutAction:addAction(
    AHColorTo.new(textView, HIGHLIGHT_FADE_DURATION, nil, DEFAULT_TEXT_COLOR)
  )
end

local function createFittedHighlight(self, views, viewIndex, color, fadeInAction, fadeOutAction)
  local layouts = self.layoutBuilder:getItemLayouts()
  local firstView = views[viewIndex]
  local firstLayout = layouts[firstView]

  local x, y = firstLayout:getX(), firstLayout:getY()
  local roundedY = MathUtils.round(y, Y_COMPARISON_PRECISION)
  local width = firstView:getWidth()

  addTextFades(firstView, fadeInAction, fadeOutAction)

  local leftBound = x - HIGHLIGHT_PADDING
  local rightBound = x + width

  local indexToJumpTo = viewIndex

  for iterationIndex = viewIndex, #views do
    indexToJumpTo = indexToJumpTo + 1
    local nextView = views[indexToJumpTo]
    local nextLayout = layouts[nextView]

    local nextY = -1
    if (nextLayout ~= nil) then
      nextY = MathUtils.round(nextLayout:getY(), Y_COMPARISON_PRECISION)
    end

    if (nextY == roundedY) then
      rightBound = nextLayout:getX() + nextView:getWidth()
      addTextFades(nextView, fadeInAction, fadeOutAction)
    else
      --- need to bring index back one since the next view
      -- isn't on the same line
      indexToJumpTo = indexToJumpTo - 1
      break
    end
  end

  width = rightBound - leftBound + HIGHLIGHT_PADDING
  local highlight = FilledRectangle.new(width, HIGHLIGHT_HEIGHT, color)
  highlight:setAnchorPointLeft()
  highlight:setLoc(leftBound, y)
  highlight:setOpacity(0.0)
  self:insert(highlight)
  highlight:bringToBack()
  return highlight, indexToJumpTo
end

local function addInvisibleHighlight(self, type)
  local views = self:getViewsWithAttribute(type)

  local fadeInHighlightsAction = AHSpawn.new()
  local fadeOutHighlightsAction = AHSpawn.new()

  local color = HIGHLIGHT_COLORS[type]

  local viewIndex = 1
  while (viewIndex <= #views) do
    local highlight, jumpIndex = createFittedHighlight(self, views, viewIndex, color, fadeInHighlightsAction, fadeOutHighlightsAction)
    viewIndex = jumpIndex

    fadeInHighlightsAction:addAction(
      AHOpacityTo.new(highlight, HIGHLIGHT_FADE_DURATION, nil, 1.0)
    )

    fadeOutHighlightsAction:addAction(
      AHSequence.new(
        AHOpacityTo.new(highlight, HIGHLIGHT_FADE_DURATION, nil, 0.0),
        AHRemove.new(highlight)
      )
    )

    viewIndex = viewIndex + 1
  end

  self.fadeInHighlightsAction = fadeInHighlightsAction
  self.fadeOutHighlightsAction  = fadeOutHighlightsAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Overriden API ------------------------

function ParagraphsTextArea:createViewForTextItem(item)
  local label = Label.new(item:getText(), item:getFont())
  local color = item:getAttribute("color")
  if (color) then
    label:setColor(unpack(DEFAULT_TEXT_COLOR))
    label:inheritOpacity(self)
  else
    label:inheritColor(self)
  end

  return label
end

function ParagraphsTextArea:createViewForImageItem(item)
  local fileName = item:getImageFilename()
  local image = Sprite.newWithName(fileName)
  image:setColor(unpack(DEFAULT_TEXT_COLOR))
  if (item:hasAttribute("color")) then
    image:setColor(unpack(item:getAttribute("color")))
  else
    image:setColor(unpack(DEFAULT_TEXT_COLOR))
  end

  if (item:hasAttribute("backingColor")) then
    local background = FilledRectangle.new(HIGHLIGHT_HEIGHT, HIGHLIGHT_HEIGHT, item:getAttribute("backingColor"))
    background:insertAtCenter(image)
    return background
  end
  return image
end

-- Outside API ------------------------

function ParagraphsTextArea:waitWhileAnimationInProgess()
  waitWhile(function() return self.animation ~= nil end)
end

function ParagraphsTextArea:getViewsAndIndicesWithAttribute(attribute, value)
  local views = {}
  local indices = {}
  for index, item in ipairs(self:getModel():getItems()) do
    if item:hasAttribute(attribute) and (item:getAttribute(attribute) == value or value == nil) then
      table.insert(indices, index)
      table.insert(views, self.itemViews[index])
    end
  end

  return views, indices
end

function ParagraphsTextArea:activateQuestion(questionType)
  self.activeQuestionType = questionType
  insertIcons(self, questionType)
end

function ParagraphsTextArea:activateResponseSequence(isCorrect, answerText)
  self.fadeOutHighlightsAction:start()
  local model = self.textModel

  model:beginTransaction()
  if (self.activeQuestionType == QUESTION_TYPES.MOVE) then
    if (answerText == "Location A") then
      answerText = "m1"
    else
      answerText = "m2"
    end
    local destinationIcons, destinationIndices = model:getItemsAndIndicesWithAttributes("locationTag")
    local destinationIndex
    for index, icon in ipairs(destinationIcons) do
      if (icon:getAttribute("locationTag") == answerText) then
        destinationIndex = destinationIndices[index]
      end
      local emptyItem = TextItem.newEmptyItem(BASE_ATTRIBUTES)
      emptyItem:setAttribute("skipNextSpacing", true)
      model:replaceItems(destinationIndices[index], 1, {emptyItem})
    end
    local tagText
    if (isCorrect) then
      tagText = "<correct/>"
    else
      tagText = "<incorrect/>"
    end
    model:replaceItems(self.firstActiveIndex, 1, self.textInterpreter:interpretXmlText(tagText))
    local numberToMove = self.lastActiveIndex - self.firstActiveIndex + 1
    model:moveItems(self.firstActiveIndex, destinationIndex, numberToMove)
  else
    local numberToReplace = self.lastActiveIndex - self.firstActiveIndex + 1

    if (answerText == "KEEP AS IS") then
      numberToReplace = 1
      answerText = ""
    end

    if (isCorrect) then
      answerText = "<correct/>" .. answerText
    else
      answerText = "<incorrect/>" .. answerText
    end

    local correctSentence = self.textInterpreter:interpretXmlText(answerText)
    model:replaceItems(self.firstActiveIndex, numberToReplace, correctSentence)
  end
  local shouldAddHighlights = false
  model:endTransaction(shouldAddHighlights)
end

function ParagraphsTextArea:transactionEnded(shouldAddHighlights)
  local itemsForViews = self.itemsForViews
  local viewsForItems = self.viewsForItems
  local layoutBuilder = self.layoutBuilder
  local animationBuilder = self.animationBuilder
  local itemLayouts = layoutBuilder:getItemLayouts()
  local newItems = self.textModel:getItems()
  local newItemsSet = TableUtils.createSetFromArray(newItems)

  for oldItem, oldItemView in pairs(viewsForItems) do
    if not newItemsSet[oldItem] then
      viewsForItems[oldItem] = nil
      itemsForViews[oldItemView] = nil
    end
  end

  local updatedItemViews = {}
  for _, newItem in ipairs(newItems) do
    local itemView = viewsForItems[newItem]
    if itemView == nil then
      itemView = self:createItemView(newItem)
      viewsForItems[newItem] = itemView
      itemsForViews[itemView] = newItem
    end
    table.insert(updatedItemViews, itemView)
  end

  self.itemViews = updatedItemViews

  layoutBuilder:updateLayout(updatedItemViews, itemsForViews)
  local newItemLayouts = layoutBuilder:getItemLayouts()
  local newHeight = layoutBuilder:getTextHeight()

  local highlightsAction
  if (shouldAddHighlights) then
    -- NOTE: Process highlights here so that we can fade them in simultaneously
    -- with text repositions
    addInvisibleHighlight(self, self.activeQuestionType)
    highlightsAction = AHDelayed.new(
      0.5,
      self.fadeInHighlightsAction
    )
  else
    highlightsAction = AHDoNothing.new()
  end

  local animation = AHSequence.new(
    AHSpawn.new(
      animationBuilder:createItemsReplacedAnimation(self, itemLayouts, newItemLayouts, newHeight),
      highlightsAction
    ),
    function()
      self.animation = nil
    end
  )

  animation:start()
  self.animation = animation
end

function ParagraphsTextArea:getLayouts()
  return self.layoutBuilder:getItemLayouts()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function ParagraphsTextArea.new(text)
  local self = TextView.new(DEFAULT_WIDTH, Alignment.LEFT, Alignment.CENTER)
  Mixin.mix(self, ParagraphsTextArea)

  self.animationBuilder = TextAnimationBuilder.new()
  self.animationBuilder:setMoveDuration(MOVE_DURATION)
  self.animationBuilder:setShowDuration(SHOW_DURATION)
  self.animationBuilder:setHideDuration(HIDE_DURATION)
  self.animationBuilder:setAdjustHeightDuration(ADJUST_HEIGHT_DURATION)

  local textInterpreter = TextInterpreter.new(BASE_ATTRIBUTES, LABEL_ATTRIBUTES)
  local textModel = ParagraphsTextModel.new(textInterpreter:interpretXmlText(text))

  self:setModel(textModel)

  self.textInterpreter = textInterpreter
  self.textModel = textModel
  self.activeQuestionType = nil
  self.fadeOutHighlightsAction = nil
  self.firstActiveIndex = nil
  self.lastActiveIndex = nil

  self:setColor(unpack(DEFAULT_TEXT_COLOR))

  return self
end
