--- PassagesGame
-- @author Created by Jay Austria on December 23, 2015.

class("PassagesGame")
-- Needed for proper scaling on iPad
FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("PassagesConfiguration")
require("PassagesScene")
require("PassagesRoundGenerator")
require("bonuses.ResponseSpeedTracker")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {1.0, 1.0, 1.0}
local UI_COMPONENTS_COLOR = {0.4, 0.4, 0.4, 0.75}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function initializeWithConcept(self)
  local roundGenerator = PassagesRoundGenerator.new(PassagesConfiguration.NUMBER_OF_QUESTIONS)
  local concept = roundGenerator:next()
  self.concept = concept
  GameSession.contentTracker:notifyPresentedConcept(concept.id)
  self.scene:initializeWithConcept(concept)
end

local function showIntroInterstitial(self)
  local introInterstitial = self.scene:getIntroInterstitial()
  local tapToContinue = TapToContinue.new(introInterstitial)
  self.tapToContinue = tapToContinue
end

local function waitForIntroDismissal(self)
  self.tapToContinue:waitForTrigger()
  local hideAction = self.scene:getHideIntroAction()
  wait(hideAction:start())
  self.tapToContinue:removeSelf()
end

local function checkIfOutOfLives(self)
  if (self.livesCounter:getLives() == 0)then
    GameSession.notifyGameLoss()
  end
end

local function updateLivesOrScore(self, isCorrect)
  if (isCorrect == true) then
    self.responseSpeedTracker:notifyCorrectResponse()
    GameSession.increaseScore(PassagesConfiguration.BASE_SCORE_FOR_CORRECT_ANSWER)
    self.scoreLabel:setScore(GameSession:getScore())
  else
    self.responseSpeedTracker:notifyIncorrectResponse()
    self.livesCounter:decrement()
  end
end

local function onResponse(self, answerButton)
  local correctAnswerSelected = answerButton.isCorrect
  updateLivesOrScore(self, correctAnswerSelected)

  if (correctAnswerSelected) then
    Sound.playSoundEffect("passages_correct")
    GameSession.contentTracker:notifyCorrectAnswer(self.concept.id, self.currentQuestionType)
    GameSession.incrementCorrectAnswerCount()
    wait(self.scene:getCorrectAction(answerButton):start())
    self.hasSubmittedResponse = true
  else
    Sound.playSoundEffect("passages_incorrect")
    GameSession.contentTracker:notifyIncorrectAnswer(self.concept.id, self.currentQuestionType, answerButton:getText())
    GameSession.incrementIncorrectAnswerCount()
    wait(self.scene:getIncorrectAction(answerButton):start())
    ExperimentsConfiguration.COMPREHENSION_RATIONALES_EXPERIMENT:reportExperimentExposed()
    if ExperimentsConfiguration.COMPREHENSION_RATIONALES_EXPERIMENT:isActive() then
      self.scene:showRationale()
    else
      wait(self.scene:getQuestionDismissedAction():start())
      self.hasSubmittedResponse = true
    end
  end
end

local function onRationaleDismissed(self)
  wait(self.scene:getQuestionDismissedAction():start())
  self.hasSubmittedResponse = true
end

local function waitForResponse(self)
  waitUntil(function()
    return self.hasSubmittedResponse
  end)

  self.hasSubmittedResponse = false
end

local function startNextQuestion(self)
  local questionTab = self.scene:getNextQuestion()
  self.currentQuestionType = questionTab.questionType
  questionTab.hasBeenActivated = true
  questionTab:bringToBack()
  Sound.playSoundEffect("passages_icon_tap")
  self.scene:startQuestion(questionTab)

  self.responseSpeedTracker:notifyStart()
end

local function doOutro(self)
  ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), PassagesConfiguration.getPointsPerLifeBonus(), self.bonusTray)
  self.scoreLabel:setScore(GameSession:getScore())
  GameSession.notifyGameTimeEnd()

  local outroAction = self.scene:createOutroAction()
  Sound.playSoundEffect("passages_win")
  wait(outroAction:start())
  GameSession.notifyGameWin()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesGame:gameFlow()
  local numberOfQuestions = PassagesConfiguration.NUMBER_OF_QUESTIONS

  waitForIntroDismissal(self)
  self.scene:waitForFirstQuestion()

  for questionIndex = 1, numberOfQuestions do
    startNextQuestion(self)
    waitForResponse(self)
    checkIfOutOfLives(self)
  end

  doOutro(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------
local function configureGameDifficulty()
  local rankTwoScore = PassagesConfiguration.getRankTwoScore()
  local rankThreeScore = PassagesConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(PassagesConfiguration.getNumberOfLives())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

local function createGameElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 PassagesConfiguration.POINTS_FOR_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(PassagesConfiguration.getSpeedBonusWindow(),
                                                       quickResponseWrapper,
                                                       PassagesConfiguration.SPEED_BONUS_DECAY_START)

  local scene = PassagesScene.new()

  scene:setOnResponseCallback(function(answerButton)
    onResponse(self, answerButton)
  end)

  scene:setOnRationaleDismissedCallback(function()
    onRationaleDismissed(self)
  end)

  self:insertAtCenter(scene)
  self.scene = scene
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(PassagesConfiguration.getNumberOfLives())
end

-- Constructor ------------------------

function PassagesGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, PassagesGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  initializeWithConcept(self)
  showIntroInterstitial(self)

  self.hasSubmittedResponse = false
  self.currentQuestionType = nil

  return self
end
