--- PassagesQuestionLayout
-- @author Created by Steven Mah on May 27, 2016.

class("PassagesQuestionLayout")
require("PassagesAnswerButton")
require("text.SizeConstrainedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local HEADER_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 14)
local TAP_TO_CONTINUE_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 12)

local QUESTION_TEXT_COLOR = {0.0, 0.0, 0.0, 0.7}
local QUESTION_MINIMUM_FONT_SIZE = 13
local QUESTION_FONT_SIZE = 17

local QUESTION_BASE_ATTRIBUTES = {
  font = Font.new(FontFaces.DIN_OT_MEDIUM, QUESTION_FONT_SIZE),
  color = QUESTION_TEXT_COLOR
}

local QUESTION_LABEL_ATTRIBUTES = {
  em = {
    font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, QUESTION_FONT_SIZE),
    color = QUESTION_TEXT_COLOR
  }
}

local RATIONALE_TITLE = "TIP"
local TAP_TO_CONTINUE = "TAP TO CONTINUE"

local WIDTH = Screen.fullWidth

local HEADER_PADDING_X = 8
local HEADER_PADDING_Y = 12
local ICON_PADDING_X = 5
local ANSWER_SPACING = 6
local HEADER_QUESTION_MARGIN_Y = 16
local QUESTION_PADDING_Y = 5

local RATIONALE_CURVE_DESCRIPTOR = {
  points = {
    {0, 0},
    {0.5, 0.17222222222222},
    {1, 1}
  },
  params = {
    tension = 0,
    pointsPerSegment = 36
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesQuestionLayout:setTouchEnabled(isEnabled)
  for _, answerButton in ipairs(self.answers) do
    answerButton:setTouchEnabled(isEnabled)
  end
end

function PassagesQuestionLayout:createFadeOutAction()
  local duration = 0.1

  local fadeOutAction = AHSpawn.new(
    AHOpacityTo.new(self.header, duration, nil, 0.0),
    AHOpacityTo.new(self.questionTextArea, duration, nil, 0.0),
    AHOpacityTo.new(self.rationaleGroup, duration, nil, 0.0)
  )

  for _, answerButton in ipairs(self.answers) do
    fadeOutAction:addAction(
      answerButton:createFadeOutAction(duration)
    )
  end

  return fadeOutAction
end

function PassagesQuestionLayout:createFadeInAction()
  local duration = 0.1

  local fadeOutAction = AHSpawn.new(
    AHOpacityTo.new(self.header, duration, nil, 0.7),
    AHOpacityTo.new(self.questionTextArea, duration, nil, 0.7)
  )

  for _, answerButton in ipairs(self.answers) do
    fadeOutAction:addAction(
      answerButton:createFadeInAction(duration)
    )
  end

  return fadeOutAction
end

function PassagesQuestionLayout:showRationale()
  self.rationaleGroup:setOpacity(1)
  local duration = 0.4
  local ease = AHEaseCustom.newWithCurve(RATIONALE_CURVE_DESCRIPTOR)
  local moveAction = AHLocationXBy.new(self, duration, ease, -self:getWidth())
  AHSequence.new(
    moveAction,
    function()
      self.rationaleGroup:setTouchEnabled(true)
    end
  ):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createHeader(self, question)
  local headerView = Label.new(string.upper(question.type), HEADER_FONT)
  local headerIcon = Sprite.newWithName(question.type.."_icon.png")
  local headerIconScale = headerView:getHeight() / headerIcon:getHeight()
  headerIcon:setScl(headerIconScale)

  local headerWidth = headerView:getWidthPart(2.0) + ICON_PADDING_X
  local headerHeight = headerView:getHeight()

  local header = Group.new(headerWidth, headerHeight)

  local color = QUESTION_TEXT_COLOR
  header:insert(headerView)
  header:insert(headerIcon)
  headerView:setColor(unpack(color))
  headerIcon:setColor(unpack(color))

  headerView:setAnchorPointLeft()
  headerIcon:setAnchorPointLeft()

  headerView:setLoc(headerIcon:getXRight() + ICON_PADDING_X, header:getHeightPart(0.5))
  headerIcon:setLoc(0, header:getHeightPart(0.5))

  header:setAnchorPointTopLeft()
  header:setLoc(HEADER_PADDING_X + Screen.safeLeft, self:getHeight() - HEADER_PADDING_Y)
  local isRecursive = true
  header:linkOpacityAttrToChildren(isRecursive)
  header:setOpacity(0.0)
  self:insert(header)
  self.header = header
end

local function createQuestion(self, question)
  assert(self.header ~= nil, "Header for Passages question must be created before the header question.")
  assert(self.answerGroup ~= nil, "Answer group for Passages question must be created before the header question.")

  local textAreaWidth = Screen.getSafeWidthPart(0.95)
  local textAreaHeight = self.header:getYBottom() - HEADER_QUESTION_MARGIN_Y - self.answerGroup:getYTop() - QUESTION_PADDING_Y
  local questionTextArea = SizeConstrainedTextArea.new(question.text,
                                                       textAreaWidth,
                                                       textAreaHeight,
                                                       QUESTION_BASE_ATTRIBUTES,
                                                       QUESTION_LABEL_ATTRIBUTES,
                                                       QUESTION_MINIMUM_FONT_SIZE)
  local x = Screen.getSafeX(0.5)
  local y = self.header:getYBottom() - HEADER_QUESTION_MARGIN_Y
  questionTextArea:setAnchorPointTop()
  questionTextArea:setLoc(x, y)
  questionTextArea:setOpacity(0.0)
  self:insert(questionTextArea)
  self.questionTextArea = questionTextArea
end

local function createAnswers(self, question, isTwoAnswers)
  local answerGroup = Column.new(ANSWER_SPACING)
  local answerWidth = Screen.getSafeWidthPart(0.96)
  local answerHeight = Screen.getHeightPart(0.09)

  if (isTwoAnswers) then
    answerWidth = Screen.getSafeWidthPart(0.47)
    answerHeight = Screen.getHeightPart(0.12)
    answerGroup = Row.new(ANSWER_SPACING)
  end

  local answers = {}
  for _, answer in ipairs(question.answers) do
    local answerButton = PassagesAnswerButton.new(answer, answerWidth, answerHeight)
    answerGroup:insert(answerButton)

    if (answerButton.isCorrect) then
      self.correctAnswer = answerButton
    end

    local onResponseWrapper = function()
      for _, button in ipairs(self.answers) do
        button:setTouchEnabled(false)
      end
      self.onResponseCallback(answerButton)
    end

    Button.makeButton(answerButton, onResponseWrapper)
    answerButton:setTouchEnabled(false)

    table.insert(answers, answerButton)
  end

  answerGroup:alignChildren()

  answerGroup:setAnchorPointBottom()
  answerGroup:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.05))
  self:insert(answerGroup)
  self.answerGroup = answerGroup
  self.answers = answers
end

local function createRationale(self, question, rationaleDismissedCallback)
  local rationaleGroup = Group.new(self:getDims())
  local titleLabel = Label.new(RATIONALE_TITLE, HEADER_FONT)
  titleLabel:setColor(unpack(QUESTION_TEXT_COLOR))
  titleLabel:setAnchorPointTopLeft()
  titleLabel:setLoc(Screen.safeLeft + Screen.getSafeWidthPart(0.05), self:getHeight() - HEADER_PADDING_Y)
  rationaleGroup:insert(titleLabel)

  local tapToContinueLabel = Label.new(TAP_TO_CONTINUE, TAP_TO_CONTINUE_FONT)
  rationaleGroup:insertAtBottom(tapToContinueLabel)
  tapToContinueLabel:offsetY(Screen.getSafeY(0.07))
  tapToContinueLabel:setColor(unpack(QUESTION_TEXT_COLOR))

  local rationaleLabelAttributes = {
    i = {
      font = Font.new(FontFaces.DIN_OT_BOLD, QUESTION_FONT_SIZE),
      color = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE[question.type].rationaleHighlightColor
    }
  }

  local textAreaWidth = Screen.getSafeWidthPart(0.9)
  local textAreaHeight = self.header:getYBottom() - HEADER_QUESTION_MARGIN_Y - tapToContinueLabel:getYTop() - QUESTION_PADDING_Y
  local rationaleTextArea = SizeConstrainedTextArea.new(question.rationale,
                                                        textAreaWidth,
                                                        textAreaHeight,
                                                        QUESTION_BASE_ATTRIBUTES,
                                                        rationaleLabelAttributes,
                                                        QUESTION_MINIMUM_FONT_SIZE)
  local x = Screen.getSafeX(0.5)
  local y = self.header:getYBottom() - HEADER_QUESTION_MARGIN_Y
  rationaleTextArea:setAnchorPointTop()
  rationaleTextArea:setLoc(x, y)
  rationaleTextArea:setOpacity(0.0)
  rationaleGroup:insert(rationaleTextArea)

  local isRecursive = true
  rationaleGroup:linkOpacityAttrToChildren(isRecursive)
  rationaleGroup:setOpacity(0)

  local buttonTappedCallback = function()
    rationaleGroup:setTouchEnabled(false)
    rationaleDismissedCallback()
  end
  Button.makeButton(rationaleGroup, buttonTappedCallback)
  rationaleGroup:setTouchEnabled(false)

  self:insertAtCenter(rationaleGroup)
  rationaleGroup:offsetX(self:getWidth())
  self.rationaleGroup = rationaleGroup
end

-- Constructor ------------------------

function PassagesQuestionLayout.new(question, onResponseCallback, rationaleDismissedCallback)
  local isTwoAnswers = (#question.answers == 2)
  local height = PassagesCommon.ACTIVE_QUESTION_HEIGHT
  if (isTwoAnswers) then
    height = PassagesCommon.ACTIVE_TRUE_FALSE_HEIGHT
  end

  local self = Group.new(WIDTH, height)
  Mixin.mix(self, PassagesQuestionLayout)

  self.onResponseCallback = onResponseCallback

  createHeader(self, question)
  createAnswers(self, question, isTwoAnswers)
  createQuestion(self, question)
  createRationale(self, question, rationaleDismissedCallback)

  return self
end
