--- PassagesQuestionsMenu
-- @author Created by Steven Mah on May 20, 2015.

class("PassagesQuestionsMenu")
require("PassagesQuestionTab")
require("PassagesQuestionLayout")
require("Particles")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.fullWidth
local HEIGHT = PassagesCommon.QUESTION_MENU_HEIGHT

local SHINE_FILENAME = "button_shine.png"

local SHINE_DURATION = 0.5
local SHINE_WAIT_DURATION = 5
local DELAY_BEFORE_NEXT_QUESTION = 0.5


local Z_INDICES = {
  OUTRO_BARS = 0,
  OUTRO_PARTICLES = 5,
  QUESTION_TABS = 10,
  SHINE = 15,
  QUESTION_LAYOUTS = 20
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createDisableAndHideInactiveButtonsAction(self, selectedButton)
  self.activeQuestionTab = selectedButton

  local fullShiftAction = AHSpawn.new(
    AHLocationYBy.new(self.background, 0.1, nil, -(HEIGHT*0.9)),
    function() self.background:setOpacity(0.0) end
  )

  for _, questionTab in ipairs(self.questionTabs) do
    if (questionTab ~= selectedButton) then
      fullShiftAction:addAction(
        AHLocationYBy.new(questionTab, 0.1, nil, -(HEIGHT*0.9))
      )
    end
  end

  return fullShiftAction
end

local function setQuestionButtonTouch(self, button)
  local firstQuestionSelected = function()
    self.shineAnimation:stop()
    self.shineImage:removeSelf()

    self.firstQuestion = button
    self:setTouchEnabled(false)
  end

  Button.makeButton(button, firstQuestionSelected)
  button:setTouchEnabled(false)
end

local function createBarAndAnimation(self, tab)
  local duration = 0.4
  local tabWidth = tab:getWidth()
  local tabX = tab:getX()
  local height = Screen.fullHeight
  local xAnchorPoint, _ = tab:getAnchorPoint()

  local questionTypeColors = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE[tab.questionType]
  local topColor = questionTypeColors.winBarTopColor
  local bottomColor = questionTypeColors.winBarBottomColor

  local bar = RectangleGradient.new(tabWidth, height, topColor, bottomColor)
  bar:setAnchorPoint(xAnchorPoint, 1.0)
  bar:setX(tabX)
  self:insert(bar, Z_INDICES.OUTRO_BARS)

  local slideUpAction = AHLocationYTo.new(bar, duration, AHEaseSineIn.new(), Screen.fullHeight)

  return slideUpAction
end

local function createShineAnimation(self)
  local shineImage = Sprite.newWithName(SHINE_FILENAME)
  shineImage:setEffectiveDims(nil, HEIGHT)
  shineImage:setAnchorPointBottomLeft()

  local imageW = shineImage:getEffectiveWidth()

  shineImage:setX(-imageW)
  self:insert(shineImage, Z_INDICES.SHINE)

  local reset = function()
    shineImage:setLoc(-imageW)
  end

  local delay = AHDelay.new(SHINE_WAIT_DURATION)
  local move = AHLocationXTo.new(shineImage, SHINE_DURATION, nil, WIDTH)
  local resetMove = AHCallLogic.new(function() reset() end)
  local loop = AHLoop.new(AHSequence.new(delay, move, resetMove))
  loop:start()

  self.shineImage = shineImage
  self.shineAnimation = loop
end

local function createOutroParticlesAction(self, tab)
  local PARTICLES_FILENAME = "win_particles.pex"
  local DURATION_PARTICLES = 2

  local tabX, tabY = tab:getLoc()
  local tabWidth, tabHeight = tab:getDims()

  local particlesActions = AHSpawn.new()
  local locationOffsets = { {  0,     0    },
                            { -0.25,  0.20 },
                            {  0.25,  0.20 },
                            { -0.10, -0.15 },
                            {  0.10, -0.15 } }

  for _, offset in ipairs(locationOffsets) do
    local particles = Particles.new(PARTICLES_FILENAME)
    local x = tabX + math.ceil(offset[1] * tabWidth)
    local y = tabY - tabHeight + math.ceil(offset[2] * tabHeight)
    particles:setSourcePosition(x, y)
    self:insert(particles, Z_INDICES.OUTRO_PARTICLES)

    local particlesAction = function()
      particles:start()
      particles:surge(10)
      waitFor(DURATION_PARTICLES)
      particles:stop()
      particles:removeSelf()
    end

    particlesActions:addAction(particlesAction)
  end

  return AHSpawn.new(AHDelay.new(DURATION_PARTICLES), particlesActions)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesQuestionsMenu:setTouchEnabled(isEnabled)
  for _, questionTab in ipairs(self.questionTabs) do
    questionTab:setTouchEnabled(isEnabled)
  end
end

function PassagesQuestionsMenu:hasFirstQuestionBeenSelected()
  return (self.firstQuestion ~= nil)
end

function PassagesQuestionsMenu:getNextQuestion()
  if (self.firstQuestion ~= nil) then
    local firstQuestion = self.firstQuestion
    -- self.firstQuestion should always be nil after this
    self.firstQuestion = nil
    return firstQuestion
  else
    for questionIndex, question in ipairs(self.questionTabs) do
      if (question.hasBeenActivated == false) then
        self.activeQuestionTab = question
        return question
      end
    end

    assert(false, "Attempted to get another question when there were none.")
  end
end

function PassagesQuestionsMenu:startQuestion(questionTab)
  local questionLayout = self.questionLayouts[questionTab.questionType]
  questionLayout:setX(Screen.getSafeX(0.5))
  self:insert(questionLayout, Z_INDICES.QUESTION_LAYOUTS)

  local activationAction = AHSequence.new(
    AHSpawn.new(
      questionTab:createMoveAndScaleUpAction(),
      AHDelayed.new(0.45, createDisableAndHideInactiveButtonsAction(self, questionTab))
    ),
    questionLayout:createFadeInAction(),
    function()
      questionLayout:setTouchEnabled(true)
      self.activeQuestionLayout = questionLayout
    end
  )

  activationAction:start()
end

function PassagesQuestionsMenu:getEnableAndShowInactiveButtonsAction()
  local fullShiftAction = AHSpawn.new(
    AHLocationYTo.new(self.background, 0.25, nil, 0.0),
    AHOpacityTo.new(self.background, 0.1, nil, 1.0)
  )

  for _, questionTab in ipairs(self.questionTabs) do
    if (questionTab ~= self.activeQuestionTab) then
      fullShiftAction:addAction(
        AHLocationYTo.new(questionTab, 0.25, nil, 0.0)
      )
    end
  end

  fullShiftAction:addAction(function() self.activeQuestionTab = nil end)

  return AHSequence.new(fullShiftAction, AHDelay.new(DELAY_BEFORE_NEXT_QUESTION))
end

function PassagesQuestionsMenu:createOutroAction()
  local fullOutroAction = AHStaggeredSpawn.new(0.05)

  for _, questionTab in ipairs(self.questionTabs) do
    local barAnimation = createBarAndAnimation(self, questionTab)
    fullOutroAction:addAction(barAnimation)

    local tabAnimation = questionTab:createOutroAction()
    fullOutroAction:addAction(tabAnimation)

    local particleAnimation = createOutroParticlesAction(self, questionTab)
    fullOutroAction:addAction(particleAnimation)
  end

  return fullOutroAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, round)
  local questions = round.questions

  local xAnchorPoint = 0.0
  local yAnchorPoint = 0.0

  local xPositions = {0, Screen.getSafeX(0.5), Screen.fullWidth}
  local widths = {Screen.getSafeX(1.0/3), Screen.getSafeWidthPart(1.0/3), Screen.fullWidth - Screen.getSafeX(2.0/3)}
  -- Icon positions should be halfway across the safe zone on each of the three buttons
  local iconXRelative = widths[2]/2
  local iconX = {Screen.safeLeft + iconXRelative, iconXRelative, iconXRelative}

  local backgroundGroup = Group.new(WIDTH, HEIGHT)
  local leftBacking = FilledRectangle.new(Screen.getWidthPart(0.005), HEIGHT, {1.0, 1.0, 1.0})
  leftBacking:setX(Screen.getSafeX(0.33))
  leftBacking:setAnchorPointBottomLeft()

  local rightBacking = FilledRectangle.new(Screen.getWidthPart(0.005), HEIGHT, {1.0, 1.0, 1.0})
  rightBacking:setX(Screen.getSafeX(0.665))
  rightBacking:setAnchorPointBottomLeft()

  backgroundGroup:insert(leftBacking)
  backgroundGroup:insert(rightBacking)
  backgroundGroup:setAnchorPointBottom()
  backgroundGroup:setX(WIDTH / 2)
  self:insert(backgroundGroup)
  backgroundGroup:linkOpacityAttrToChildren()
  self.background = backgroundGroup

  -- Assumes that there will always be three questions
  local questionButtons = {}
  for buttonIndex, question in ipairs(questions) do
    local questionButton = PassagesQuestionTab.new(question, widths[buttonIndex], iconX[buttonIndex], PassagesCommon.QUESTION_ICON_MENU_HEIGHT)

    setQuestionButtonTouch(self, questionButton)

    questionButton:setAnchorPoint(xAnchorPoint, yAnchorPoint)
    questionButton:setX(xPositions[buttonIndex])

    self:insert(questionButton, Z_INDICES.QUESTION_TABS)

    xAnchorPoint = xAnchorPoint + 0.5
    table.insert(questionButtons, questionButton)
  end

  self.questionTabs = questionButtons
end

local function createElements(self, round)
  local questionLayouts = {}
  for _, question in ipairs(round.questions) do
    local questionLayout = PassagesQuestionLayout.new(question, self.onResponseCallback, self.onRationaleDismissedCallback)
    questionLayout:setAnchorPointBottom()
    questionLayouts[question.type] = questionLayout
  end

  self.questionLayouts = questionLayouts

  createShineAnimation(self)
end

-- Constructor ------------------------

function PassagesQuestionsMenu.new(round, onResponseCallback, onRationaleDismissedCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, PassagesQuestionsMenu)

  self.onResponseCallback = onResponseCallback
  self.onRationaleDismissedCallback = onRationaleDismissedCallback

  createLayout(self, round)
  createElements(self, round)

  self.activeQuestionLayout = nil
  self.firstQuestion = nil

  return self
end
