--- PassagesScene
-- @author Created by Jay Austria on December 23, 2015.

class("PassagesScene")

require("PassagesConfiguration")
require("PassagesIntroInterstitial")
require("PassagesScrollLayer")
require("widgets.TapToContinue")
require("PassagesQuestionsMenu")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {1.0, 1.0, 1.0}

local Z_INDICES = {
  SCROLL_LAYER = 0,
  COVER_LAYER = 5,
  QUESTIONS_MENU = 10
}

local EXIT_CURVE_DESCRIPTOR = {
  points = {
    {0, 0},
    {0.5, 0.17222222222222},
    {1, 1}
  },
  params = {
    tension = 0,
    pointsPerSegment = 36
  }
}

local TOTAL_MARGIN_HEIGHT = PassagesCommon.PASSAGE_MARGIN_HEIGHT

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesScene:initializeWithConcept(concept)
  local passageText = concept.passage
  self.concept = concept
  local scrollLayer = PassagesScrollLayer.new(passageText)
  scrollLayer:setAnchorPointTop()
  scrollLayer:setLoc(self:getWidthPart(0.5), self:getHeight())
  self:insert(scrollLayer)
  scrollLayer:setZIndex(Z_INDICES.SCROLL_LAYER)
  self.scrollLayer = scrollLayer

  local questionsMenu = PassagesQuestionsMenu.new(concept, self.onResponseCallback, self.onRationaleDismissedCallback)
  questionsMenu:setAnchorPointBottomLeft()
  self:insert(questionsMenu)
  questionsMenu:setZIndex(Z_INDICES.QUESTIONS_MENU)
  self.questionsMenu = questionsMenu
end

function PassagesScene:getIntroInterstitial()
  local introDescriptor = self.concept.intro
  local introInterstitial = PassagesIntroInterstitial.new(introDescriptor)
  self.introInterstitial = introInterstitial

  return introInterstitial
end

function PassagesScene:getHideIntroAction()
  local hideAction = AHSequence.new(
    AHLocationYBy.new(self.introInterstitial, 0.4, AHEaseCustom.newWithCurve(EXIT_CURVE_DESCRIPTOR), Screen.fullHeight),
    AHRemove.new(self.introInterstitial),
    function() self.questionsMenu:setTouchEnabled(true) end
  )

  return hideAction
end

function PassagesScene:waitForFirstQuestion()
  waitUntil(function()
    return self.questionsMenu:hasFirstQuestionBeenSelected()
  end)
end

function PassagesScene:getNextQuestion()
  return self.questionsMenu:getNextQuestion()
end

function PassagesScene:startQuestion(questionTab)
  self.scrollLayer:setQuestionActive(true)
  local fadeInHighlights = self.scrollLayer:getHighlightsFadeInActionByType(questionTab.questionType)
  if (fadeInHighlights ~= nil) then
    fadeInHighlights:start()
  end

  self.questionsMenu:startQuestion(questionTab)
end

function PassagesScene:getCorrectAction(answerButton)
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout
  local questionTab = questionsMenu.activeQuestionTab

  questionLayout:setTouchEnabled(false)

  local fullResponseAction = AHSequence.new(
    answerButton:createCorrectResponseAction(answerButton.isCorrect),
    AHDelay.new(1.0),
    questionLayout:createFadeOutAction(),
    AHSpawn.new(
      questionTab:createCorrectResponseAction(),
      questionsMenu:getEnableAndShowInactiveButtonsAction(),
      self:getFadeOutHighlightsAction(questionTab.questionType)
    )
  )

  return fullResponseAction
end

function PassagesScene:getQuestionDismissedAction()
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout
  local questionTab = questionsMenu.activeQuestionTab


  local fullResponseAction = AHSequence.new(
    questionLayout:createFadeOutAction(),
    AHSpawn.new(
      questionTab:createIncorrectResponseAction(),
      questionsMenu:getEnableAndShowInactiveButtonsAction(),
      self:getFadeOutHighlightsAction(questionTab.questionType)
    )
  )

  return fullResponseAction
end

function PassagesScene:getIncorrectAction(answerButton)
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout

  questionLayout:setTouchEnabled(false)

  local buttonAction
  if (answerButton ~= nil) then
    buttonAction = AHSequence.new(
      AHSpawn.new(
        answerButton:createIncorrectResponseAction(),
        questionLayout.correctAnswer:createCorrectResponseAction(answerButton.isCorrect)
      ),
      AHDelay.new(1.0)
    )
  else
    buttonAction = AHDoNothing.new()
  end

  return buttonAction
end

function PassagesScene:showRationale()
  local questionsMenu = self.questionsMenu
  local questionLayout = questionsMenu.activeQuestionLayout
  questionLayout:showRationale()
end

function PassagesScene:createOutroAction()
  self.scrollLayer:setTouchEnabled(false)
  local fullOutroAction = AHSpawn.new(
    AHOpacityTo.new(self.separator, 0.25, nil, 0.0),
    self.questionsMenu:createOutroAction(),
    self.scrollLayer:getOutroTextAction()
  )
  return fullOutroAction
end

function PassagesScene:setOnResponseCallback(callback)
  self.onResponseCallback = callback
end

function PassagesScene:setOnRationaleDismissedCallback(callback)
  self.onRationaleDismissedCallback = callback
end

function PassagesScene:getQuestionsMenu()
  return self.questionsMenu
end

function PassagesScene:getFadeOutHighlightsAction(questionType)
  return self.scrollLayer:getHighlightsFadeOutActionByType(questionType)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local coverHeight = TOTAL_MARGIN_HEIGHT

  local topCover = FilledRectangle.new(Screen.fullWidth, coverHeight, BACKGROUND_COLOR)
  topCover:setLoc(Screen.getWidthPart(0.5), Screen.fullHeight - coverHeight/2)

  local lineSeparator = FilledRectangle.new(Screen.fullWidth, 0.5, {0.5, 0.5, 0.5})
  lineSeparator:setAnchorPointBottom()
  lineSeparator:setX(Screen.getWidthPart(0.5))
  topCover:insert(lineSeparator)
  self.separator = lineSeparator

  self:insert(topCover, Z_INDICES.COVER_LAYER)
end

function PassagesScene.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, PassagesScene)

  createLayout(self)

  self.onResponseCallback = nil
  self.onRationaleDismissedCallback = nil

  return self
end
