--- PassagesScrollLayer
-- @author Created by Jay Austria on December 23, 2015.

class("PassagesScrollLayer")

require("widgets.TouchScroller")
require("PassagesCommon")
require("PassagesTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PASSAGE_MARGIN_X = Screen.getWidthPart(0.08)
local PASSAGE_MARGIN_Y_TOP = Screen.fullHeight - Screen.getSafeY(0.9)
local PASSAGE_MARGIN_Y_BOTTOM = Screen.getSafeY(0.5)

local AUTO_SHIFT_PADDING = 5

local AUTO_SCROLL_DURATION = 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function autoScroll(self, shiftDistance)
  self.autoScrollAction = self.scroller:newScrollYBy(AUTO_SCROLL_DURATION, AHEaseExponentialOut.new(), shiftDistance)
  self.autoScrollAction:start()
end

local function ensureHighlightedTextVisible(self, views)
  local firstView = views[1]
  local lastView = views[#views]

  --- assume that it's impossible for highlighted text to both be above and
  -- below the window
  local _, topMostY = firstView:getWorldLoc()
  topMostY = topMostY + firstView:getHeight()
  local _, bottomMostY = lastView:getWorldLoc()
  local upperBound = Screen.safeHeight - PassagesCommon.PASSAGE_MARGIN_HEIGHT
  local lowerBound = PassagesCommon.ACTIVE_QUESTION_HEIGHT

  local isUnderLowerBound = (bottomMostY < lowerBound)
  local isAboveUpperBound = (topMostY > upperBound)

  if (isUnderLowerBound) then
    local shiftDistance = lowerBound - bottomMostY + AUTO_SHIFT_PADDING
    autoScroll(self, shiftDistance)
  elseif (isAboveUpperBound) then
    local shiftDistance = upperBound - topMostY - AUTO_SHIFT_PADDING
    autoScroll(self, shiftDistance)
  end
end

local function passageAutoScrollIfNeeded(self)
  local _, passageY = self.passageView:getCenterWorldLoc()
  local topMostY = passageY + self.passageView:getHeightPart(0.5)
  local bottomMostY = passageY - self.passageView:getHeightPart(0.5)

  local upperBound = Screen.safeHeight - PASSAGE_MARGIN_Y_TOP
  local lowerBound = PassagesCommon.QUESTION_MENU_HEIGHT * 2
  if self.isQuestionActive then
    lowerBound = PassagesCommon.ACTIVE_QUESTION_HEIGHT
  end

  local isTopLineTooLow = (topMostY < upperBound)
  local isBottomLineTooHigh = (bottomMostY > lowerBound)

  if (isTopLineTooLow) then
    local shiftDistance = upperBound - topMostY - AUTO_SHIFT_PADDING
    autoScroll(self, shiftDistance)
  elseif (isBottomLineTooHigh) then
    local shiftDistance = lowerBound - bottomMostY + AUTO_SHIFT_PADDING
    autoScroll(self, shiftDistance)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesScrollLayer:getHighlightsFadeInActionByType(type)
  local views = self.passageView:getViewsWithAttribute(type)

  if (#views > 0) then
    ensureHighlightedTextVisible(self, views)
  end

  return self.passageView:getHighlightsFadeInActionByType(type)
end

function PassagesScrollLayer:getHighlightsFadeOutActionByType(type)
  return self.passageView:getHighlightsFadeOutActionByType(type)
end

function PassagesScrollLayer:getOutroTextAction()
  local passageView = self.passageView

  return passageView:getOutroTextAction()
end

function PassagesScrollLayer:setTouchEnabled(isEnabled)
  self.scroller:setTouchEnabled(isEnabled)
end

function PassagesScrollLayer:setQuestionActive(isActive)
  self.isQuestionActive = isActive
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createPassageView(self, passageText)
  local width = Screen.getSafeX(1.0) - (2 * PASSAGE_MARGIN_X)
  local passageView = PassagesTextArea.new(passageText, width)

  return passageView
end

local function createLayout(self, passageText)
  local width, height = self:getDims()

  local passageView = createPassageView(self, passageText)
  self.passageView = passageView

  local contentsHeight = passageView:getHeight() + PASSAGE_MARGIN_Y_TOP + PASSAGE_MARGIN_Y_BOTTOM
  local contents = Group.new(width, contentsHeight)
  passageView:setAnchorPointTop()
  passageView:setLoc(Screen.getSafeX(0.5), passageView:getHeight() + PASSAGE_MARGIN_Y_BOTTOM)
  contents:insert(passageView)
  self.contents = contents

  local scroller = TouchScroller.new(width, height, contents)
  scroller:setInertialScrollingEnabled(true)
  scroller:setScrollBoundsEnabled(false)
  scroller:setWillResetInitialTouchCoordinateUponScrollStart(true)

  local stopScrollIfInProgress = function()
    if (self.autoScrollAction ~= nil) then
      self.autoScrollAction:stop()
      self.autoScrollAction = nil
    end
  end

  local autoScrollWrapper = function()
    passageAutoScrollIfNeeded(self)
  end

  scroller:setScrollNotificationFunctions(stopScrollIfInProgress, autoScrollWrapper)
  self:insertAtCenter(scroller)
  self.scroller = scroller
end

function PassagesScrollLayer.new(passageText)
  local self = Group.newFullScreen()
  Mixin.mix(self, PassagesScrollLayer)

  self.autoScrollAction = nil
  self.isQuestionActive = false

  createLayout(self, passageText)

  return self
end
