--- PassagesTextArea
-- @author Created by Steven Mah on June 2, 2016.

class("PassagesTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BASE_ATTRIBUTES = {
  font = Font.new(FontFaces.TISA_REGULAR, 20)
}

local LABEL_ATTRIBUTES = {
  details = {details = true},
  inference = {inference = true},
  technique = {technique = true},
  vocab = {vocab = true}
}

local HIGHLIGHT_COLORS = {
  details = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE.details.fillColor,
  inference = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE.inference.fillColor,
  technique = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE.technique.fillColor,
  vocab = PassagesCommon.MENU_BUTTON_STYLES_BY_QUESTION_TYPE.vocab.fillColor
}

local HIGHLIGHT_FADE_INS_BY_TYPE = {
  details = nil,
  inference = nil,
  technique = nil,
  vocab = nil
}

local HIGHLIGHT_FADE_OUTS_BY_TYPE = {
  details = nil,
  inference = nil,
  technique = nil,
  vocab = nil
}

local HIGHLIGHT_FADE_DURATION = 0.5
local HIGHLIGHT_PADDING = 5

local TEXT_COLOR = PassagesCommon.PASSAGE_TEXT_COLOR

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addTextFades(textView, fadeInAction, fadeOutAction)
  -- Needed because views in a DynamicTextArea automatically inherit color
  textView:clearAttrLink(MOAIColor.INHERIT_COLOR)
  textView:setColor(unpack(TEXT_COLOR))

  fadeInAction:addAction(
    AHColorTo.new(textView, HIGHLIGHT_FADE_DURATION, nil, {1.0, 1.0, 1.0})
  )

  fadeOutAction:addAction(
    AHColorTo.new(textView, HIGHLIGHT_FADE_DURATION, nil, TEXT_COLOR)
  )
end

local function createFittedHighlight(views, viewIndex, color, fadeInAction, fadeOutAction)
  if (#views == 1) then
    local view = views[1]
    local width, height = view:getDims()
    local x, y = view:getLoc()
    local highlight = FilledRectangle.new(width + 2 * HIGHLIGHT_PADDING, height, color)
    highlight:setAnchorPointBottomLeft()
    highlight:setLoc(x - HIGHLIGHT_PADDING, y)
    highlight:setOpacity(0.0)
    view.parent:insert(highlight)
    highlight:bringToBack()

    addTextFades(view, fadeInAction, fadeOutAction)

    return highlight, viewIndex
  else
    local firstView = views[viewIndex]
    local x, y = firstView:getLoc()
    local width, height = firstView:getDims()

    addTextFades(firstView, fadeInAction, fadeOutAction)

    local leftBound = x - HIGHLIGHT_PADDING
    local rightBound = x + width

    local indexToJumpTo = viewIndex

    for iterationIndex = viewIndex, #views do
      indexToJumpTo = indexToJumpTo + 1
      local nextView = views[indexToJumpTo]

      local nextY = -1
      if (nextView ~= nil) then
        nextY = nextView:getY()
      end

      if (nextY == y) then
        rightBound = nextView:getX() + nextView:getWidth()
        addTextFades(nextView, fadeInAction, fadeOutAction)
      else
        --- need to bring index back one since the next view
        -- isn't on the same line
        indexToJumpTo = indexToJumpTo - 1
        break
      end
    end

    width = rightBound - leftBound + HIGHLIGHT_PADDING

    local highlight = FilledRectangle.new(width, height * 0.9, color)
    highlight:setAnchorPointLeft()
    highlight:setLoc(leftBound, y + firstView:getHeightPart(0.5))
    highlight:setOpacity(0.0)
    firstView.parent:insert(highlight)
    highlight:bringToBack()
    return highlight, indexToJumpTo
  end
end

local function addInvisibleHighlight(type, views)
  local fadeInHighlightsAction = AHSpawn.new()
  local fadeOutHighlightsAction = AHSpawn.new()

  local color = HIGHLIGHT_COLORS[type]

  local viewIndex = 1
  while (viewIndex <= #views) do
    local highlight, jumpIndex = createFittedHighlight(views, viewIndex, color, fadeInHighlightsAction, fadeOutHighlightsAction)
    viewIndex = jumpIndex

    fadeInHighlightsAction:addAction(
      AHOpacityTo.new(highlight, HIGHLIGHT_FADE_DURATION, nil, 1.0)
    )

    fadeOutHighlightsAction:addAction(
      AHOpacityTo.new(highlight, HIGHLIGHT_FADE_DURATION, nil, 0.0)
    )

    viewIndex = viewIndex + 1
  end

  HIGHLIGHT_FADE_INS_BY_TYPE[type] = fadeInHighlightsAction
  HIGHLIGHT_FADE_OUTS_BY_TYPE[type] = fadeOutHighlightsAction
end

local function addHighlightsForTags(self)
  for key, _ in pairs(LABEL_ATTRIBUTES) do
    local views = self:getViewsWithAttribute(key)

    if (#views > 0) then
      addInvisibleHighlight(key, views)
    end
  end
end

local function createOutroTextAction(self)
  local wordViews = self:getItemViews()
  local textWidth = self:getWidth()
  local maxLineY = self.layoutBuilder:getTextHeight()

  local textAnimation = AHSpawn.new()
  for _, wordView in ipairs(wordViews) do
    local wordWidth = wordView:getWidth()

    local delayTime = nil
    local moveDistance = nil

    local wordIsOnLeftHalf = (wordView:getX() + wordWidth/2 < textWidth/2)

    if wordIsOnLeftHalf then
      moveDistance = -Screen.fullWidth
      delayTime = 0.01 * (maxLineY - wordView:getY()) + 0.01 * wordView:getX()
    else
      moveDistance = Screen.fullWidth
      -- TODO: What is the point of everything after the + and should it be able to bring delayTime negative?
      delayTime = 0.01 * wordView:getY() + 0.01 * (textWidth - wordView:getX() - wordView:getWidth())
      delayTime = math.max(delayTime, 0)
    end
    delayTime = delayTime / 5

    local slideAction = AHLocationXBy.new(wordView, 0.4, AHEaseSineIn.new(), moveDistance)

    textAnimation:addAction(AHDelayed.new(delayTime, slideAction))
  end

  self.outroTextAction = textAnimation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesTextArea:getHighlightsFadeInActionByType(type)
  local fadeInAction = HIGHLIGHT_FADE_INS_BY_TYPE[type]
  return fadeInAction
end

function PassagesTextArea:getHighlightsFadeOutActionByType(type)
  local fadeOutAction = HIGHLIGHT_FADE_OUTS_BY_TYPE[type]

  return fadeOutAction
end

function PassagesTextArea:getOutroTextAction()
  return self.outroTextAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PassagesTextArea.new(passageText, width)
  local self = DynamicTextArea.new(passageText, width, BASE_ATTRIBUTES, LABEL_ATTRIBUTES)
  Mixin.mix(self, PassagesTextArea)

  addHighlightsForTags(self)
  createOutroTextAction(self)
  self:setColor(unpack(TEXT_COLOR))

  return self
end
