--- PassagesValidator
-- @author Created by Jay Austria on December 23, 2015.
-- ConceptValidator for Passages.

class("PassagesValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'passage',
  'excerpt',
  'illustration'
}

local REQUIRED_ALLOWED_EMPTY_KEYS = {
  'overview_question',
  'overview_correct',
  'overview_incorrect_1',
  'overview_incorrect_2',
  'overview_rationale',
  'technique_question',
  'technique_correct',
  'technique_incorrect_1',
  'technique_incorrect_2',
  'technique_rationale',
  'vocab_question',
  'vocab_correct',
  'vocab_incorrect_1',
  'vocab_incorrect_2',
  'vocab_rationale',
  'inference_question',
  'inference_correct',
  'inference_incorrect_1',
  'inference_incorrect_2',
  'inference_rationale',
  'details_question',
  'details_correct',
  'details_incorrect_1',
  'details_incorrect_2',
  'details_rationale'
}

local QUESTION_TYPES = {
  "overview",
  "technique",
  "vocab",
  "inference",
  "details"
}

local SINGLE_TAGS = nil
local PAIRED_TAGS = nil
local TERTIARY_TAGS = nil
local OPTIONAL_PAIRED_TAGS = { "technique", "vocab", "inference", "details" }
TableUtils.concatenateArrays(OPTIONAL_PAIRED_TAGS, TagValidator.DEFAULT_OPTIONAL_PAIRED_TAGS)

local MINIMUM_QUESTIONS_PRESENT = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

function validatePassagePunctuation(self, passage)
  local cleanPassage = passage
  for _, questionType in ipairs(QUESTION_TYPES) do
    cleanPassage = cleanPassage:gsub("<" .. questionType .. ">", "")
    cleanPassage = cleanPassage:gsub("</" .. questionType .. ">", "")
  end

  self:testForEndingPunctuation(cleanPassage)
end

function validateRationale(self, rationale, rationaleField)
  self:expectNotEmpty(rationale, rationaleField)
  self:validateTags(rationale)
  self:testForEndingPunctuation(rationale)
end

function validateQuestions(self, concept)
  local questionsFound = 0

  for _, questionType in ipairs(QUESTION_TYPES) do
    local questionField = questionType .. "_question"
    local correctField = questionType .. "_correct"
    local incorrectOneField = questionType .. "_incorrect_1"
    local incorrectTwoField = questionType .. "_incorrect_1"
    local rationaleField = questionType .. "_rationale"

    local question = concept[questionField]
    local correct = concept[correctField]
    local incorrectOne = concept[incorrectOneField]
    local incorrectTwo = concept[incorrectTwoField]
    local rationale = concept[rationaleField]
    if question ~= "" then
      self:validateTags(question) -- TagValidator checks "em", "i", & "b" tags
      questionsFound = questionsFound + 1
      self:expectNotEmpty(correct, correctField)
      self:expectNotEmpty(incorrectOne, incorrectOneField)
      -- Note: We don't test incorrect_2 because it is optional for all question types
      validateRationale(self, rationale, rationaleField)
    else
      self:expectTrue(correct == "")
      self:expectTrue(incorrectOne == "")
      self:expectTrue(incorrectTwo == "")
      self:expectTrue(rationale == "")
    end
  end

  self:expectGreaterOrEqual(questionsFound, MINIMUM_QUESTIONS_PRESENT, "number of questions")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PassagesValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  self:validateRequiredKeys(concept, REQUIRED_ALLOWED_EMPTY_KEYS, true)

  self:validateTags(concept.passage, SINGLE_TAGS, PAIRED_TAGS, TERTIARY_TAGS, OPTIONAL_PAIRED_TAGS)
  validatePassagePunctuation(self, concept.passage)
  validateQuestions(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PassagesValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, PassagesValidator)

  return self
end
