--- PercentagesGame
-- @author Created by Jay Austria on April 15, 2015.

class("PercentagesGame")

require("StandardUIBuilder")
require("PercentagesConfiguration")
require("PercentagesCommon")
require("PercentagesScene")
require("PercentagesConceptGenerator")
require("PercentagesRoundIndicator")
require("PercentagesCentralSubmitButton")
require("PercentagesChatRationaleHelper")
require("widgets.number_pad.NumberPad")
require("widgets.TapToContinue")
require("bonuses.BonusNotificationTray")
require("bonuses.ResponseSpeedTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = ColorUtils.colorHex24ToStd(0xffffff)
local NUMBER_PAD_KEY_COLOR = ColorUtils.colorHex24ToStd(0xffabbd)

local function configureGameDifficulty(self)
  self.setsToComplete = PercentagesConfiguration.SETS_TO_COMPLETE_COUNT

  local rankTwoScore = PercentagesConfiguration.getRankTwoScore()
  local rankThreeScore = PercentagesConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local conceptGenerator = PercentagesConceptGenerator.new()
  self.conceptGenerator = conceptGenerator

  local difficultyKeyFrames = PercentagesConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(PercentagesConfiguration.getNumberOfLives())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function registerGameWon(self)
  Sound.playSoundEffect("goal_reach")
  GameSession.notifyGameTimePause()

  local endGameAction = AHStaggeredSpawn.new(
    1.75,
    function() self.scene:playOutroAnimation() end,
    function() self.uiOverlay:fadeOut() end
  )

  wait(endGameAction:start())

  GameSession.notifyGameWin()
end

local function registerGameLoss(self)
  GameSession.notifyGameLoss()
end

local function didLose(self)
  return (self.livesCounter:getLives() == 0)
end

local function checkForGameLoss(self)
  if didLose(self) then
    registerGameLoss(self)
  end
end

local function isCorrectAnswer(self, value)
  return (self.expectedAnswer == value)
end

local function registerCorrectResponse(self)
  local scene = self.scene

  self.responseSpeedTracker:notifyCorrectResponse()
  GameSession.notifyCorrectAnswer(self.conceptIdentifier)
  GameSession.incrementCorrectAnswerCount()

  GameSession.increaseScore(PercentagesConfiguration.POINTS_PER_CORRECT_RESPONSE)

  self.scoreLabel:setScore(GameSession.getScore())

  Sound.playSoundEffect("correct_answer")

  scene:animateCorrectResponse(self.percentage, self.expectedAnswer)

  local unlockDelay = 0.1
  waitFor(unlockDelay)

  self.wasPreviousCorrect = true
end

local function registerIncorrectResponse(self)
  self.livesCounter:decrement()
  self.responseSpeedTracker:notifyIncorrectResponse()
  GameSession.incrementIncorrectAnswerCount()
  Sound.playSoundEffect("incorrect_answer")
  GameSession.notifyGameTimePause()

  local scene = self.scene
  wait(scene:animateIncorrectResponse())

  scene:showTip(self.percentage, self.expectedAnswer)

  local isWaitingForTap = true

  local delayedTapToContinue = function()
    waitFor(0.5)
    local tapToContinueCallback = function()
      isWaitingForTap = false
      scene:prepareForHide()
    end
    local runCallbackInCoroutine = false
    local tapToContinue = TapToContinue.new(nil, nil, tapToContinueCallback, runCallbackInCoroutine)

    scene:setLearnMorePressedCallback(function()
      tapToContinue:disarm()
    end)

    scene:setLearnMoreClosedCallback(function()
      tapToContinue:rearm()
    end)

    tapToContinue:bringToBack()
  end

  runInCoroutine(delayedTapToContinue)

  waitWhile(function() return isWaitingForTap end)

  wait(scene:hideTip())

  self.wasPreviousCorrect = false
  GameSession.notifyGameTimeStart()
end

local function registerResponse(self, value)
  local isEmptyResponse = ((value == nil) or (value == 0))
  if isEmptyResponse then
    self.scene:reactToInvalidInput()
    return
  end

  self.numberPad:setTouchEnabled(false)
  self.centralSubmitButton:setTouchEnabled(false)

  local conceptId = self.conceptIdentifier
  local percentageAsString = tostring(self.percentage)

  if isCorrectAnswer(self, value) then
    GameSession.contentTracker:notifyCorrectAnswer(conceptId, percentageAsString)
    registerCorrectResponse(self)
  else
    local responseAsString = tostring(value)
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, percentageAsString, responseAsString)
    registerIncorrectResponse(self)
  end

  self.isWaitingForResponse = false
end

local function beginRoundWithConcept(self, concept, setIndex)
  local value = concept.value
  self.conceptIdentifier = concept.identifier

  local setsCount = PercentagesConfiguration.SETS_TO_COMPLETE_COUNT
  local willDoIntroAnimation = (setIndex == 1)
  local percentage = concept.percentages[1]

  if willDoIntroAnimation then
    self.scene:playIntroAnimation(percentage, value)
  else
    self.scene:playSetSwitchAnimation(percentage, value)
  end

  for questionIndex = 1, PercentagesConfiguration.QUESTIONS_COUNT_PER_SET do
    self.questionIndex = questionIndex
    self.isWaitingForResponse = true

    local isWaitingForResponseWrapper = function() return self.isWaitingForResponse end

    self.expectedAnswer = concept.expectedAnswers[questionIndex]

    local percentage = concept.percentages[questionIndex]
    self.percentage = percentage

    local tipText = concept.tips[questionIndex]
    local conversationDescriptor = concept.conversations[questionIndex]
    wait(self.scene:presentQuestion(percentage, tipText, conversationDescriptor, questionIndex, self.wasPreviousCorrect, concept.identifier))

    self.responseSpeedTracker:notifyStart()
    self.numberPad:setTouchEnabled(true)
    self.centralSubmitButton:setTouchEnabled(true)

    waitWhile(isWaitingForResponseWrapper)

    checkForGameLoss(self)
  end

  self.scene:presentQuestion(nil, nil, nil, nil, self.wasPreviousCorrect)

  self.scene:endSet()
end

local function prepareForNextRound(self)
  self.scene:doSetSwitchAnimation()
end

local function showAndWaitForHighLevelInterstitial(self)
  wait(self.scene:showHighLevelInterstitial())
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PercentagesGame:gameFlow()
  if PercentagesConfiguration.shouldShowHighDifficultyInterstitial() then
    showAndWaitForHighLevelInterstitial(self)
  end

  local setsCount = PercentagesConfiguration.SETS_TO_COMPLETE_COUNT
  Sound.playSoundEffect("game_start")

  for setIndex = 1, setsCount do
    local concept = self.conceptGenerator:next()
    GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
    ChatRationaleAnalytics.setConceptId(concept.identifier)
    beginRoundWithConcept(self, concept, setIndex)

    local setTransitionDelay = 2.0
    waitFor(setTransitionDelay)

    local areSetsRemaining = (self.roundIndicator:getRound() < setsCount)
    if areSetsRemaining then
      self.roundIndicator:incrementRound()
    end
  end

  if not didLose(self) then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), PercentagesConfiguration.getExtraLivesBonus(), self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())
    registerGameWon(self)
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local onSubmit = function(value)
    registerResponse(self, value)
  end

  local numberPad = NumberPad.new(NUMBER_PAD_KEY_COLOR)
  numberPad:setOnSubmitCallback(onSubmit)
  self.numberPad = numberPad

  local willShowReminders = (not PercentagesConfiguration.getIsHighLevel())

  local centralSubmitButton = PercentagesCentralSubmitButton.new(onSubmit)
  self.centralSubmitButton = centralSubmitButton

  local scene = PercentagesScene.new(numberPad, centralSubmitButton, willShowReminders)
  self:insertAtCenter(scene)
  self.scene = scene

  self.scene:initializeForIntro()

  local quickResponseWrapper = function(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 PercentagesConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(PercentagesConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       PercentagesConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(PercentagesConfiguration.getNumberOfLives())

  local roundIndicator = PercentagesRoundIndicator.new(PercentagesConfiguration.SETS_TO_COMPLETE_COUNT)
  local roundIndicatorOffsetX = nil
  local roundIndicatorOffsetY = -25
  builder:addWidget(roundIndicator, UIOverlay.NORTHEAST, roundIndicatorOffsetX, roundIndicatorOffsetY)
  self.roundIndicator = roundIndicator
end

-- Constructor ------------------------

function PercentagesGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, PercentagesGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  PercentagesChatRationaleHelper.initialize()

  return self
end
