--- PercentagesScene
-- @author Created by Jay Austria on May 5, 2015.

class("PercentagesScene")

require("PercentagesInputField")
require("PercentagesHelpPanel")
require("PercentagesCorrectAnswerBurst")
require("PercentagesPrompt")
require("PercentagesGradientRectangle")
require("PercentagesHighLevelInterstitial")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INPUT_FIELD_HEIGHT = 100
local INPUT_FIELD_Y = Screen.getSafeY(0.575)
local REMINDER_VISIBLE_OPACITY = 0.5

local OUTRO_ARM_MOTION_DESCRIPTORS =
{
  {
    rotationAmount = 90
  },
  {
    rotationAmount = -60
  },
  {
    rotationAmount = 180
  },
  {
    rotationAmount = 60
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function clearReminders(self)
  if self.willShowReminders then
    local fadeDuration = 1.0
    for _, child in ipairs(self.reminders.children) do
      local fadeOut = AHOpacityTo.new(child, fadeDuration, nil, PercentagesCommon.INVISIBLE_OPACITY)
      child.fadeInterrupt:insert(fadeOut)
    end

    self.currentReminderIndex = 1
  end
end

local function createLayout(self, centralSubmitButton)
  local background = PercentagesGradientRectangle.new(Screen.fullWidth,
                                                      Screen.fullHeight,
                                                      PercentagesCommon.BACKGROUND_COLOR_TOP,
                                                      PercentagesCommon.BACKGROUND_COLOR_BOTTOM)
  self:insertAtCenter(background)
  self.background = background

  local numberPad = self.numberPad
  numberPad:setLoc(Screen.getSafeX(0.5), Screen.safeBottom + numberPad:getHeightPart(0.5))
  self:insert(numberPad)

  local minimumInputFieldHeight = numberPad:getYTop() + (INPUT_FIELD_HEIGHT / 2)
  local inputFieldY = math.max(INPUT_FIELD_Y, minimumInputFieldHeight)

  local minimumCentralSubmitButtonY = numberPad:getYTop() + (centralSubmitButton:getHeightPart(0.5))
  local centralSubmitButtonY = math.max(inputFieldY, minimumCentralSubmitButtonY)
  centralSubmitButton:setLoc(Screen.getSafeX(0.5), centralSubmitButtonY)
  self:insert(centralSubmitButton)
  self.centralSubmitButton = centralSubmitButton

  local prompt = PercentagesPrompt.new()
  prompt:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  self:insert(prompt)
  self.prompt = prompt

  local inputField = PercentagesInputField.new(INPUT_FIELD_HEIGHT,
                                               FontFaces.DIN_OT_LIGHT,
                                               PercentagesCommon.INPUT_FIELD_FONT_SIZE,
                                               PercentagesCommon.COLOR.WHITE,
                                               PercentagesCommon.COLOR.INCORRECT_RED)
  inputField:setLoc(Screen.getSafeX(0.5), inputFieldY)
  inputField:setValidInputCallback(function() PercentagesCommon.playValidInputSound() end)
  inputField:setInvalidInputCallback(function() PercentagesCommon.playInvalidInputSound() end)
  self:insert(inputField)
  self.inputField = inputField

  numberPad:attachInputField(inputField)
  centralSubmitButton:attachInputField(inputField)

  if self.willShowReminders then
    local reminders = Column.new(6)

    local remindersCount = PercentagesConfiguration.QUESTIONS_COUNT_PER_SET - 1
    local baseAttributes = {
      font = Font.new(FontFaces.DIN_OT_MEDIUM, 16)
    }
    local boldAttributes = {
      b = { font = Font.new(FontFaces.DIN_OT_BOLD, 16)}
    }
    for reminderIndex = 1, remindersCount do
      -- Placeholder text is required for `reminders:alignChildren()` to properly space the DynamicTextAreas.
      local reminder = DynamicTextArea.new("-unset-", Screen.safeWidth, baseAttributes, boldAttributes, Alignment.CENTER, Alignment.CENTER)
      reminder.fadeInterrupt = ActionInterrupt.new()
      reminders:insert(reminder)
    end
    reminders:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.69))
    reminders:alignChildren()

    for _, reminder in ipairs(reminders.children) do
      reminder:setOpacity(PercentagesCommon.INVISIBLE_OPACITY)
    end

    self:insert(reminders)
    self.reminders = reminders
  end

  local helpPanel = PercentagesHelpPanel.new()
  self:insertAtCenter(helpPanel)
  self.helpPanel = helpPanel
end

local function addReminder(self, percent, value)
  if self.currentReminderIndex >= (PercentagesConfiguration.QUESTIONS_COUNT_PER_SET) then
    return
  end

  local reminder = self.reminders.children[self.currentReminderIndex]
  local reminderText = string.format(Localizer.get("PERCENT_TEMPLATE"), percent, value)
  reminder:setText(reminderText)

  local fadeIn = AHOpacityTo.new(reminder, 2.0, nil, REMINDER_VISIBLE_OPACITY)
  reminder.fadeInterrupt:insert(fadeIn)

  self.currentReminderIndex = self.currentReminderIndex + 1
end

local function createRotatingArmGroup(self, shape)
  local shapeX, shapeY = shape:getCenterWorldLoc()

  local centerX = Screen.getSafeX(0.5)
  local centerY = Screen.getSafeY(0.5)

  local radius = MathUtils.distance(shapeX, shapeY, centerX, centerY)
  local arm = Group.new(radius, 0)
  arm:setAnchorPointLeft()
  self:insertAtCenter(arm)

  local angle = MathUtils.angleBetweenVectors(centerX, centerY, centerX + 1, centerY, shapeX, shapeY)
  local angleInDegrees = math.deg(angle)
  arm:setRot(angleInDegrees)

  shape:setAnchorPointCenter()
  arm:takeChild(shape)
  shape:setLoc(radius, arm:getHeightPart(0.5))
  shape:setRot(-angleInDegrees)

  return arm
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PercentagesScene:reactToInvalidInput()
  self.inputField:reactToInvalidInput()
end

function PercentagesScene:presentQuestion(percentage, tipText, conversationDescriptor, index, wasPreviousCorrect, conceptId)
  self.helpPanel:setHelpText(tipText, conversationDescriptor, conceptId)

  local presentingAction = nil

  -- First set is already animating by this point due to playIntroAnimation() or playSetSwitchAnimation()
  if index == 1 then
    presentingAction = self.lockingAnimation
  else
    presentingAction = self.prompt:presentNextShape(percentage, wasPreviousCorrect)
  end

  return presentingAction
end

function PercentagesScene:animateCorrectResponse(percent, value)
  local correctAnswerBurst = PercentagesCorrectAnswerBurst.new(value)
  correctAnswerBurst:setLoc(self.inputField:getLoc())
  self:insert(correctAnswerBurst)

  local burstAnimation = correctAnswerBurst:getBurstAnimation()
  local burstSequence = AHSequence.new(burstAnimation, AHRemove.new(correctAnswerBurst))

  local clearInputField = function()
    self.inputField:clear()
  end

  local delayedClearInputField = AHDelayed.new(1.0, clearInputField)

  local correctAnimation = AHSpawn.new(burstSequence, delayedClearInputField)

  local pinReminder = function()
    if self.willShowReminders then
      addReminder(self, percent, value)
    end
  end

  correctAnimation:addAction(pinReminder)
  correctAnimation:addAction(function() self.numberPad:reset() end)

  return correctAnimation:start()
end

function PercentagesScene:animateIncorrectResponse()
  return AHSequence.new(AHDelay.new(0.1),
                        function() self.numberPad:reset() end,
                        AHDelay.new(0.2)):start()
end

function PercentagesScene:showTip(percent, value)
  runInCoroutine(function()
    self.helpPanel:animateShow()
    self.inputField:showCorrection(value)

    if self.willShowReminders then
      addReminder(self, percent, value)
    end
  end)
end

function PercentagesScene:prepareForHide()
  self.helpPanel:prepareForHide()
end

function PercentagesScene:hideTip()
  runInCoroutine(function()
    self.helpPanel:animateHide()
    self.inputField:clearCorrection()
  end)
end

function PercentagesScene:initializeForIntro()
  local numberPad = self.numberPad
  local isRecursive = true
  numberPad:linkOpacityAttrToChildren(isRecursive)
  numberPad:setOpacity(PercentagesCommon.INVISIBLE_OPACITY)

  numberPad:setLoc(self:getWidthPart(0.5), numberPad:getHeightPart(0.4))
end

function PercentagesScene:playIntroAnimation(percentage, value)
  self.prompt:presentFirstQuestion(percentage, value)

  local numberPad = self.numberPad
  local fadeIn = AHOpacityTo.new(numberPad, 1.5, nil, PercentagesCommon.VISIBLE_OPACITY)
  local numberPadY = Screen.safeBottom + numberPad:getHeightPart(0.5)
  local slideOut = AHLocationYTo.new(numberPad, 0.5, AHEaseSimpleInOut.new(), numberPadY)

  local doFadeBackdrop = true
  local numberOfDigits = math.log10(value)
  local promptAnimation = self.prompt:getIntroAnimation(doFadeBackdrop, numberOfDigits)

  local introAnimation = AHSpawn.new(fadeIn, slideOut, promptAnimation)

  self.lockingAnimation = introAnimation

  return introAnimation:start()
end

function PercentagesScene:playSetSwitchAnimation(percentage, value)
  local setSwitchAnimation = self.prompt:playSetSwitchAnimation(percentage, value)
  self.background:playColorChange()
  Sound.playSoundEffect("complete_set")

  self.lockingAnimation = setSwitchAnimation
end

function PercentagesScene:endSet()
  runInCoroutine(function()
    local clearDelay = 1.0
    waitFor(clearDelay)

    if self.willShowReminders then
      clearReminders(self)
    end
  end)
end

function PercentagesScene:playOutroAnimation()
  local promptOutroAnimation = self.prompt:getOutroAnimation()

  local outroDuration = 0.8

  local outroAnimation = AHSpawn.new(promptOutro)
  local shapes = TableUtils.shallowCopy(self.prompt:getShapes())
  local shapeSpawn = AHSpawn.new()
  for shapeIndex, shape in ipairs(shapes) do
    local rotatingArm = createRotatingArmGroup(self, shape)
    local motionDescriptor = OUTRO_ARM_MOTION_DESCRIPTORS[shapeIndex]
    shapeSpawn:addAction(AHLocationXTo.new(shape, outroDuration, AHEaseSimpleIn.new(), 0))
    shapeSpawn:addAction(AHRotationBy.new(rotatingArm, outroDuration, AHEaseSimpleIn.new(), motionDescriptor.rotationAmount))
    shapeSpawn:addAction(AHDelayed.new(outroDuration, AHRemove.new(rotatingArm)))
  end

  local valueLabel = self.prompt:getValueLabel()
  local valueArm = createRotatingArmGroup(self, valueLabel)

  local motionDescriptor = OUTRO_ARM_MOTION_DESCRIPTORS[PercentagesConfiguration.QUESTIONS_COUNT_PER_SET + 1]
  shapeSpawn:addAction(AHLocationXTo.new(valueLabel, outroDuration, AHEaseSimpleIn.new(), 0))
  shapeSpawn:addAction(AHRotationBy.new(valueArm, outroDuration, AHEaseSimpleIn.new(), motionDescriptor.rotationAmount))
  shapeSpawn:addAction(AHOpacityTo.new(valueLabel, outroDuration * 0.5, nil, PercentagesCommon.INVISIBLE_OPACITY))
  shapeSpawn:addAction(AHScaleTo.new(valueLabel, outroDuration, nil, 0.5))
  shapeSpawn:addAction(AHDelayed.new(outroDuration, AHRemove.new(valueArm)))

  local stayGreen = true
  local backgroundColorChange = self.background:getColorChangeAction(stayGreen)

  local burst = PercentagesCorrectAnswerBurst.newWithImage("checkmark.png")
  self:insertAtCenter(burst)
  local burstAnimation = AHDelayed.new(outroDuration * 0.95, burst:getBurstAnimation())

  local numberPad = self.numberPad
  local fadeNumberPad = AHOpacityTo.new(numberPad, outroDuration, nil, PercentagesCommon.INVISIBLE_OPACITY)

  local outroDelay = AHDelay.new(3.5) --needed to keep the game from exiting before sfx finishes

  local combinedOutroAnimation = AHSpawn.new(promptOutroAnimation, shapeSpawn, backgroundColorChange, burstAnimation, fadeNumberPad, outroDelay)

  return combinedOutroAnimation:start()
end

function PercentagesScene:showHighLevelInterstitial()
  local interstitial = PercentagesHighLevelInterstitial.new()
  interstitial:prepareForAnimateIn()

  local tapToContinue = TapToContinue.new(interstitial)
  wait(interstitial:doAnimateIn())
  tapToContinue:waitForTrigger()
  wait(interstitial:doAnimateOut())
end

function PercentagesScene:setLearnMorePressedCallback(callback)
  self.helpPanel:setLearnMorePressedCallback(callback)
end

function PercentagesScene:setLearnMoreClosedCallback(callback)
  self.helpPanel:setLearnMoreClosedCallback(callback)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PercentagesScene.new(numberPad, centralSubmitButton, willShowReminders)
  local self = Group.newFullScreen()
  Mixin.mix(self, PercentagesScene)

  self.numberPad = numberPad
  self.currentReminderIndex = 1
  self.willShowReminders = willShowReminders

  createLayout(self, centralSubmitButton)

  return self
end
