--- PercentagesShapeDeck
-- @author Created by Jay Austria on May 14, 2015.

class("PercentagesShapeDeck")

require("PercentagesShape")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local X_OFFSET = 9
local Y_OFFSET = 9
local LIFT_OFFSET_Y = 145

local SHAPE_OPACITIES =
{
  1.0,
  0.8,
  0.4
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getShapeOpacityForIndex(self, index)
  local shapeOpacity = SHAPE_OPACITIES[index]

  return shapeOpacity
end

local function getShapeLocationForIndex(self, index)
  local x = self:getWidthPart(0.5)
  local y = self:getHeightPart(0.5)

  local xOffset = (index - 1) * X_OFFSET
  local yOffset = (index - 1) * Y_OFFSET

  local shapeX = x + xOffset
  local shapeY = y + yOffset

  return shapeX, shapeY
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PercentagesShapeDeck:prepareForIntro()
  local x = self:getWidthPart(0.5)
  local y = self:getHeightPart(0.5)

  for shapeIndex, shape in rpairs(self.children) do
    local invertedIndex = #self.children - shapeIndex  + 1 --we want to iterate from front to back
    local rotationOffset = (invertedIndex - 1) * 15
    shape:setRot(-90 + rotationOffset)
    shape:setLoc(x, y)
  end
end

function PercentagesShapeDeck:getIntroAnimation()
  local introAnimation = AHSpawn.new()
  local animationDuration = 1.9

  local x = self:getWidthPart(0.5)
  local y = self:getHeightPart(0.5)
  for shapeIndex, shape in rpairs(self.children) do
    local invertedIndex = #self.children - shapeIndex + 1 --we want to iterate from front to back
    local shapeX, shapeY = getShapeLocationForIndex(self, invertedIndex)
    local rotationSequence = AHSequence.new(
                              AHRotationBy.new(shape, animationDuration * 0.3, AHEaseExponentialIn.new(), -160),
                              AHRotationTo.new(shape, animationDuration * 0.65, AHEaseExponentialOut.new(), 0)
                             )
    local shapeColor = TableUtils.shallowCopy(PercentagesShape.START_COLOR)
    shapeColor[4] = getShapeOpacityForIndex(self, invertedIndex)
    local morphColor = AHDelayed.new(animationDuration * 0.7, AHColorTo.new(shape, animationDuration * 0.3, nil, shapeColor))
    local growShape = AHScaleTo.new(shape, animationDuration * 0.8, AHEaseSimpleInOut.new(), 1.0)
    introAnimation:addAction(rotationSequence)
    introAnimation:addAction(AHLocationTo.new(shape, animationDuration * 0.6, AHEaseExponentialInOut.new(), shapeX, shapeY))
    introAnimation:addAction(growShape)
    introAnimation:addAction(morphColor)
  end

  return introAnimation
end

function PercentagesShapeDeck:getPopFrontAnimation(value, wasCorrect)
  local frontShape = self.children[#self.children]
  local nextShape = self.children[#self.children - 1]

  local animationDuration = 1.6

  local liftUp = AHDelayed.new(0.3, AHLocationTo.new(frontShape, animationDuration * 0.5, AHEaseSimpleInOut.new(), frontShape:getX() + X_OFFSET * 0.5, frontShape:getY() + LIFT_OFFSET_Y))
  local liftUpSpawn = AHSpawn.new(liftUp)

  if wasCorrect then
    liftUpSpawn:addAction(frontShape:getShowAsCorrectAction())
  else
    liftUpSpawn:addAction(frontShape:getShowAsIncorrectAction())
  end

  local resortShapes = function()
    frontShape:bringToBack()
  end

  local rearrangeShapes = function()
    local rearrangeSpawn = AHSpawn.new()
    for shapeIndex, shape in rpairs(self.children) do
      local invertedIndex = #self.children - shapeIndex + 1 --we want to iterate from front to back
      local x, y = getShapeLocationForIndex(self, invertedIndex)
      local updatedOpacity = getShapeOpacityForIndex(self, invertedIndex)
      rearrangeSpawn:addAction(AHLocationTo.new(shape, animationDuration * 0.5, AHEaseSimpleInOut.new(), x, y))
      rearrangeSpawn:addAction(AHOpacityTo.new(shape, animationDuration * 0.5, nil, updatedOpacity))
    end

    wait(rearrangeSpawn:start())
  end

  local shuffleAction = AHSequence.new(liftUpSpawn, resortShapes, rearrangeShapes)

  return shuffleAction
end

function PercentagesShapeDeck:animateCollapse()
  local collapseSpawn = AHSpawn.new()
  local animationDuration = 0.6
  for shapeIndex, shape in ipairs(self.children) do
    local collapseDelay = (shapeIndex - 1) * 0.05
    local rotateAction = AHRotationBy.new(shape, animationDuration, AHEaseSimpleInOut.new(), -360)
    local scaleAction = AHScaleTo.new(shape, animationDuration, AHEaseSimpleIn.new(), 0.0)
    local combinedCollapse = AHDelayed.new(collapseDelay, AHSpawn.new(rotateAction, scaleAction))

    collapseSpawn:addAction(combinedCollapse)
  end

  local collapseSequence = AHSequence.new(collapseSpawn, AHRemove.new(self))

  collapseSequence:start()
end

function PercentagesShapeDeck:setShapesScale(scale)
  for _, shape in ipairs(self.children) do
    shape:setScl(scale)
  end
end

function PercentagesShapeDeck:setShapesColor(color)
  for _, shape in ipairs(self.children) do
    shape:setColor(unpack(color))
  end
end

function PercentagesShapeDeck:getChildren()
  return self.children
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PercentagesShapeDeck.new(shapeType)
  local testShape = PercentagesShape.new()

  local self = Group.new(PercentagesShape.GROUP_WIDTH, PercentagesShape.GROUP_HEIGHT)
  Mixin.mix(self, PercentagesShapeDeck)

  local shapeCount = PercentagesConfiguration.QUESTIONS_COUNT_PER_SET
  for shapeIndex = 1, shapeCount do
    local shape = PercentagesShape.new(shapeType)
    local shapeOpacity = getShapeOpacityForIndex(self, (shapeCount - shapeIndex + 1))
    shape:setOpacity(shapeOpacity)
    self:insert(shape)
  end

  self:prepareForIntro()

  return self
end
