--- PercentagesValidator
-- @author Created by Jay Austria on April 15, 2015.
-- ConceptValidator for Percentages.

class("PercentagesValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'value',
  'percent1',
  'percent2',
  'percent3',
  'tip1',
  'tip2',
  'tip3',
  "anotherValue"
}

local OPTIONAL_KEYS = {
  "alternativeTip1",
  "alternativeTip2",
  "alternativeTip3"
}

local VALID_TIP_KEYS = {
  ['TIP_ADD'] = true,
  ['TIP_SUBTRACT'] = true,
  ['TIP_MULTIPLY_4'] = true,
  ['TIP_MULTIPLY_3'] = true,
  ['TIP_FIND_1_PERCENT'] = true,
  ['TIP_FIND_2_PERCENT'] = true,
  ['TIP_FIND_2_PERCENT_ALT'] = true,
  ['TIP_FIND_3_PERCENT'] = true,
  ['TIP_FIND_3_PERCENT_ALT'] = true,
  ['TIP_FIND_4_PERCENT'] = true,
  ['TIP_FIND_4_PERCENT_ALT'] = true,
  ['TIP_FIND_5_PERCENT'] = true,
  ['TIP_FIND_10_PERCENT'] = true,
  ['TIP_FIND_20_PERCENT'] = true,
  ['TIP_FIND_20_PERCENT_ALT'] = true,
  ['TIP_FIND_25_PERCENT'] = true,
  ['TIP_FIND_25_PERCENT_ALT'] = true,
  ['TIP_FIND_40_PERCENT'] = true,
  ['TIP_FIND_50_PERCENT'] = true,
  ['TIP_FIND_60_PERCENT'] = true,
  ['TIP_FIND_90_PERCENT'] = true
}

local VALID_ALTERNATIVE_TIP_KEYS = {
  ['EXPANDED_TIP_FIND_1_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_5_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_10_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_30_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_75_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_95_PERCENT'] = true,
  ['EXPANDED_TIP_FIND_PERCENT_OF_100'] = true,
  ['EXPANDED_TIP_Y_OF_X_10_PERCENT'] = true,
  ['EXPANDED_TIP_Y_OF_X_50_PERCENT'] = true,
  ['EXPANDED_TIP_Y_OF_X_200_PERCENT'] = true
}

local EXPECTED_QUESTION_COUNT = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function containsDecimal(value)
  -- Rely on Lua's builtin float to string conversion to prevent floating point precision errors.
  -- MathUtils.isInteger() will not work here because the floating point precision error introduced for particular percentages will produce false positivies.
  return (tostring(value):find(".", 0, true) ~= nil)
end


local function validateNoDecimals(self, concept)
  local value = concept.value

  for questionIndex = 1, EXPECTED_QUESTION_COUNT do
    local percentage = concept["percent" .. questionIndex]
    local result = value * (percentage / 100.0)

    local errorMessage = string.format("Concept contained percentage %d%% and value %d, which produced a decimal.  result == %.16f", percentage, value, result)
    self:expectFalse(containsDecimal(result), errorMessage)
  end
end

local function validateTipsAreValid(self, concept)
  for questionIndex = 1, EXPECTED_QUESTION_COUNT do
    local tip = concept["tip"..questionIndex]

    self:expectTrue(VALID_TIP_KEYS[tip], "Tip is not found in list of valid tips: " .. tip)
  end
end

local function validateNumbers(self, concept)
  local tip3 = concept["tip3"]
  local standardTips = {
    'TIP_ADD',
    'TIP_SUBTRACT'
  }
  if standardTips[tip3] then
    local percent1 = concept["percent1"]
    local percent2 = concept["percent2"]
    local percent3 = concept["percent3"]
    local differenceIsEqual = (math.abs(percent1 - percent2) == percent3)
    local sumIsEqual = ((percent1 + percent2) == percent3)
    self:expectTrue(differenceIsEqual or sumIsEqual, "Numbers do not add up for valid add or subtract.")
  end
end

local function validateOptionalAlternativeTipsAreValid(self, concept)
  for questionIndex = 1, EXPECTED_QUESTION_COUNT do
    local alternativeTip = concept["alternativeTip"..questionIndex]

    if ((alternativeTip ~= nil) and (string.len(alternativeTip) > 0)) then
      local errorMessage = string.format("Optional alternative tip is not found in list of valid alternative tips: %s", alternativeTip)
      self:expectTrue(VALID_ALTERNATIVE_TIP_KEYS[alternativeTip], errorMessage)
    end
  end
end

local function createErrorMessageForInappropriateTipAndPercent(tip, percent)
  return string.format("Tip type %s is not appropriate for a round with a percent %d%%", tip, percent)
end

local function createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent)
  return string.format("Alternative tip type %s is not appropriate for a round with a percent %d%%", alternativeTip, percent)
end

local function createErrorMessageForInappropriateAlternativeTipAndValue(alternativeTip, value)
  return string.format("Alternative tip type %s is not appropriate for value %d", alternativeTip, value)
end

local function validateTipsAreAppropriateForNumbers(self, concept)
  for questionIndex = 1, EXPECTED_QUESTION_COUNT do
    local tip = concept["tip"..questionIndex]
    local percent = concept["percent"..questionIndex]

    if ((tip == 'TIP_ADD') or (tip == 'TIP_SUBTRACT') or (tip == 'TIP_MULTIPLY_4') or (tip == 'TIP_MULTIPLY_3')) then
      local errorMessage = string.format("Tip type %s may only be assigned to the third question in a round", tip)
      self:expectTrue(questionIndex == 3, errorMessage)
    elseif (tip == 'TIP_FIND_1_PERCENT') then
      self:expectTrue(percent == 1, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif ((tip == 'TIP_FIND_2_PERCENT') or (tip == 'TIP_FIND_2_PERCENT_ALT')) then
      self:expectTrue(percent == 2, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif ((tip == 'TIP_FIND_3_PERCENT') or (tip == 'TIP_FIND_3_PERCENT_ALT')) then
      self:expectTrue(percent == 3, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif ((tip == 'TIP_FIND_4_PERCENT') or (tip == 'TIP_FIND_4_PERCENT_ALT')) then
      self:expectTrue(percent == 4, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_5_PERCENT') then
      self:expectTrue(percent == 5, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_10_PERCENT') then
      self:expectTrue(percent == 10, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif ((tip == 'TIP_FIND_20_PERCENT') or (tip == 'TIP_FIND_20_PERCENT_ALT')) then
      self:expectTrue(percent == 20, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif ((tip == 'TIP_FIND_25_PERCENT') or (tip == 'TIP_FIND_25_PERCENT_ALT')) then
      self:expectTrue(percent == 25, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_40_PERCENT') then
      self:expectTrue(percent == 40, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_50_PERCENT') then
      self:expectTrue(percent == 50, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_60_PERCENT') then
      self:expectTrue(percent == 60, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    elseif (tip == 'TIP_FIND_90_PERCENT') then
      self:expectTrue(percent == 90, createErrorMessageForInappropriateTipAndPercent(tip, percent))
    end
  end
end

local function validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, percentage)
  local result = value * (percentage / 100.0)

  local errorMessage = string.format("Optional alternative tip " .. alternativeTip .. " contained percentage %d%% and value %d, which produced a decimal", percentage, value)
  self:expectFalse(containsDecimal(result), errorMessage)
end

local function validateOptionalAlternativeTipsAreAppropriateForNumbers(self, concept)
  local value = concept.value

  for questionIndex = 1, EXPECTED_QUESTION_COUNT do
    local alternativeTip = concept["alternativeTip" .. questionIndex]
    if ((alternativeTip ~= nil) and (string.len(alternativeTip) > 0)) then
      local percent = concept["percent" .. questionIndex]

      if (alternativeTip == 'EXPANDED_TIP_FIND_1_PERCENT') then
        self:expectTrue(percent == 1, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 1)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_5_PERCENT') then
        self:expectTrue(percent == 5, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 1)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_10_PERCENT') then
        self:expectTrue(percent == 10, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 10)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_30_PERCENT') then
        self:expectTrue(percent == 30, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 10)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_75_PERCENT') then
        self:expectTrue(percent == 75, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 25)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_95_PERCENT') then
        self:expectTrue(percent == 95, createErrorMessageForInappropriateAlternativeTipAndPercent(alternativeTip, percent))
        validateNoDecimalsInOptionalAlternativeTip(self, value, alternativeTip, 1)
      elseif (alternativeTip == 'EXPANDED_TIP_FIND_PERCENT_OF_100') then
        self:expectTrue(value == 100, createErrorMessageForInappropriateAlternativeTipAndValue(alternativeTip, value))
      elseif (alternativeTip == 'EXPANDED_TIP_Y_OF_X_10_PERCENT') then
        self:expectTrue(value == 10, createErrorMessageForInappropriateAlternativeTipAndValue(alternativeTip, value))
      elseif (alternativeTip == 'EXPANDED_TIP_Y_OF_X_50_PERCENT') then
        self:expectTrue(value == 50, createErrorMessageForInappropriateAlternativeTipAndValue(alternativeTip, value))
      elseif (alternativeTip == 'EXPANDED_TIP_Y_OF_X_200_PERCENT') then
        self:expectTrue(value == 200, createErrorMessageForInappropriateAlternativeTipAndValue(alternativeTip, value))
      end
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PercentagesValidator:validateConcept(concept)
  local allowEmpty = false
  self:validateRequiredKeys(concept, REQUIRED_KEYS, allowEmpty)

  allowEmpty = true
  self:validateRequiredKeys(concept, OPTIONAL_KEYS, allowEmpty)

  validateNoDecimals(self, concept)
  validateTipsAreValid(self, concept)
  validateNumbers(self, concept)
  validateOptionalAlternativeTipsAreValid(self, concept)
  validateTipsAreAppropriateForNumbers(self, concept)
  validateOptionalAlternativeTipsAreAppropriateForNumbers(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PercentagesValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, PercentagesValidator)

  return self
end
