class("PounceGameManager")
require("bonuses.BonusNotificationTray")
require("bonuses.ResponseSpeedTracker")
require("game.PounceRoundGenerator")
require("game.PounceConfiguration")
require("data.generators.LearningFocusedRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BEAN_STALKS_ON_SCREEN_SECONDS = 2.75

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function allowMistakesToResurfaceIfNeeded(self)
  if (self.ui:getProgress() >= LearningFocusedRoundGenerator.DEFAULT_PROGRESS_BEFORE_RESURFACING) then
    self.roundGenerator:allowMistakesToResurface()
  end
end

local function resetGameStats(self)
  self.points = 0
end

local function isGameWin(self)
  return self.ui:isProgressComplete()
end

local function isGameLoss(self)
  return self.ui:getLivesCount() <= 0
end

local function isGameEnd(self)
  return isGameWin(self) or isGameLoss(self)
end

local function registerCorrectAnswer(self)
  self.responseSpeedTracker:notifyCorrectResponse()

  GameSession.notifyCorrectAnswer(self.roundInfo.identifier)
  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(PounceConfiguration.POINTS_PER_CORRECT_RESPONSE)

  local points = GameSession.getScore()
  self.ui:setPoints(points)
end

local function registerIncorrectAnswer(self)
  self.responseSpeedTracker:notifyIncorrectResponse()
  self.roundGenerator:notifyMistake(self.roundInfo.identifier)
  GameSession.scoredIncorrectAnswer(self.roundInfo.identifier)
  GameSession.incrementIncorrectAnswerCount()
  self.ui:decrementLives()
end

local function doRound(self)
  local roundInfo = self.roundGenerator:next()
  self.roundInfo = roundInfo

  local conceptID = self.roundInfo.identifier
  GameSession.contentTracker:notifyPresentedConcept(conceptID)

  self.scene:setRound(roundInfo)
  self.roundEnded = false
  while(not self.scene:readyForAction()) do
    if (isGameEnd(self)) then
      return
    end
    coroutine.yield()
  end

  local shuffleTimer = MOAITimer.new()
  shuffleTimer:setSpan(BEAN_STALKS_ON_SCREEN_SECONDS)
  self.shuffleTimer = shuffleTimer

  while(not self.roundEnded and not isGameLoss(self)) do
    if (not shuffleTimer:isActive()) then
      shuffleTimer:start()
      self.scene:changeShownAnswerSet()
    end
    coroutine.yield()
  end
end

local function doGameEnd(self)
  self.scene:waitForProgressAnimation()

  if isGameWin(self) then
    GameSession.notifyGameWin()
  else
    GameSession.notifyGameLoss()
  end
end

local function doRoundTransition(self)
  local revealAnim = self.scene:revealAnswer(self.lastAnswerWasCorrect, self.lastSubmittedAnswer)
  if (self.lastAnswerWasCorrect) then
    self.scene:playCorrectAnimation(self.lastAnswerID)
  else
    self.scene:playIncorrectAnimation(self.lastAnswerID)
  end

  wait(revealAnim)
  if self.lastAnswerWasCorrect then
    self.ui:notifyCorrectAnswer()
  end
  wait(self.scene:slideOutQuestion(self.lastAnswerWasCorrect))

  self.lastAnswerWasCorrect = nil
end

local function getGameLoop(self)
  return function()
    while(not isGameEnd(self)) do
      allowMistakesToResurfaceIfNeeded(self)
      doRound(self)
      doRoundTransition(self)
    end

    if isGameWin(self) then
      ScoringUtils.applyLivesBonus(self.ui:getLivesCount(), PounceConfiguration.getExtraLivesBonus(), self.ui.bonusTray)
      self.ui:setPoints(GameSession.getScore())
      wait(self.scene:playWinAnimation())
    else
      wait(self.scene:playLoseAnimation())
    end

    doGameEnd(self)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PounceGameManager:getCallBacks()
  local answerSubmitted = function(answerData)
    -- Used to prevent an edge-case that may happen within a 1-frame window where
    -- the shuffle timer elapses just as this "answerSubmitted" event is fired
    local didAnswerBeforeShuffle = (self.shuffleTimer:getTime() > 0.0)
    if not didAnswerBeforeShuffle then
      return
    end

    local submittedAnswer = answerData.answer
    local correctAnswer = self.roundInfo.correct
    local conceptID = self.roundInfo.identifier

    if (submittedAnswer == correctAnswer) then
      GameSession.contentTracker:notifyCorrectAnswer(conceptID)
      registerCorrectAnswer(self)
      self.lastAnswerWasCorrect = true
    else
      GameSession.contentTracker:notifyIncorrectAnswer(conceptID, nil, submittedAnswer)
      registerIncorrectAnswer(self)
      self.lastAnswerWasCorrect = false
    end
    self.roundEnded = true
    self.lastAnswerID = answerData.answerID
    self.lastSubmittedAnswer = submittedAnswer
  end

  local screenTappedDuringGame = function()
    if (self.shuffleTimer == nil) then return end
    self.shuffleTimer:stop()
  end

  local answerShown = function(answer)
    local correctAnswer = self.roundInfo.correct
    local isCorrectAnswer = answer == correctAnswer

    local responseSpeedTracker = self.responseSpeedTracker
    local isFirstTimeShown = not responseSpeedTracker:isActive()

    local shouldNotifyStart = isCorrectAnswer and isFirstTimeShown
    if (shouldNotifyStart) then
      responseSpeedTracker:notifyStart()
    end
  end

  local callbacks = {
    answerSubmitted = answerSubmitted,
    screenTappedDuringGame = screenTappedDuringGame,
    answerShown = answerShown
  }

  return callbacks
end

function PounceGameManager:configureGameDifficulty()
  local correctNeeded = PounceConfiguration.CORRECT_NEEDED_TO_WIN
  self.ui:setMaxCorrectAnswers(correctNeeded)

  local passScore = PounceConfiguration.getPassScore()
  local rankTwoScore = PounceConfiguration.getRankTwoScore()
  local rankThreeScore = PounceConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyNameChooser = KeyFrameInterpolator.new(PounceConfiguration.getDifficultyKeyFrames(), GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(PounceConfiguration.getNumberOfLives())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

function PounceGameManager:start()
  local gameLoop = getGameLoop(self)
  self.gameProcess = MOAICoroutine.new():run(gameLoop)
  return self.gameProcess
end

function PounceGameManager:setScene(scene)
  self.scene = scene
end

function PounceGameManager:setUI(ui)
  self.ui = ui
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PounceGameManager.new()
  local self = {}
  Mixin.mix(self, PounceGameManager)

  self.roundGenerator = LearningFocusedRoundGenerator.new(PounceRoundGenerator.new())

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 PounceConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.ui.bonusTray)
  end

  local decayWindowStart = 0
  self.responseSpeedTracker = ResponseSpeedTracker.new(PounceConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       decayWindowStart)

  resetGameStats(self)

  return self
end
