class("AnswersView")

require("scene.answers.AnswerButton")
require("scene.answers.BeanStalkAnswer")
require("data.StalkDescriptors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ANSWER_COUNT_VALUES = { 1, 2, 3 }
local ANSWER_COUNT_WEIGHTS = { 0.5, 0.3, 0.2 }
local ANSWER_COUNT_VALUES_WITHOUT_ONE = { 2, 3 }
local ANSWER_COUNT_WEIGHTS_WITHOUT_ONE = { 0.7, 0.3 }
local MAX_ONE_ANSWER_SET_PER_ROUND = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getAnswersToShow(self)
  local oneAnswerSetCount = self.oneAnswerSetCount

  local values = ANSWER_COUNT_VALUES
  local weights = ANSWER_COUNT_WEIGHTS

  if (oneAnswerSetCount >= MAX_ONE_ANSWER_SET_PER_ROUND) then
    values = ANSWER_COUNT_VALUES_WITHOUT_ONE
    weights = ANSWER_COUNT_WEIGHTS_WITHOUT_ONE
  end

  local answersToShow = TableUtils.selectRandomValueWithWeights(values, weights)

  if (answersToShow == 1) then
    self.oneAnswerSetCount = oneAnswerSetCount + 1
  end

  return answersToShow
end

local function createAnswersWithIndex(self, index)
  local answersToShow = getAnswersToShow(self)

  -- Use this to show the entire layout set (to debug positions)
  --answersToShow = #self.layoutSet

  local newAnswersList = {}
  local usedPositions = {count = 0}

  --------------- Make sure we don't reuse positions that we have just used, if possible --------------
  local i = 1
  while (i <= #self.answersList and #self.answers - usedPositions.count > answersToShow) do
    usedPositions[self.answersList[i].layoutIndex] = true
    usedPositions.count = usedPositions.count + 1
    i = i + 1
  end

  --------------------------- Remove current stalks ----------------------------------------------------
  for _, answer in ipairs(self.answersList) do
    MOAICoroutine.new():run(function()
      local anim = answer:springOutImmediately()
      answer:deactivate()
      -- Timer doesn't start right away, if it's started in a different coroutine
      coroutine.yield()
      while(anim and anim:isActive()) do coroutine.yield() end
      answer:removeSelf()
    end)
  end

  ---------------------------- Create new stalks ----------------------------------------------------
  for j = 0, answersToShow - 1 do
    while(index > #self.answers) do index = index - #self.answers end
    local layoutIndex = math.random(#self.layoutSet)
    while(usedPositions[layoutIndex]) do layoutIndex = math.random(#self.layoutSet) end
    usedPositions[layoutIndex] = true

    local layout = self.layoutSet[layoutIndex]
    layout.colorSet = self.colorIndex
    local answerText = self.answers[index]
    local answer = BeanStalkAnswer.new(answerText, self.answerSelectHandler, layout)
    answer:setLoc(layout.x, layout.y)
    if (layout.overlay and self.overlayGroup) then
      self.overlayGroup:insert(answer, layout.z)
    else
      self.answersGroup:insert(answer, layout.z)
    end

    MOAICoroutine.new():run(function()
      coroutine.yield()
      answer:show()
      local activateAnswer = function() answer:activate() end
      self.lastSpringInAnimation = answer:springIn(nil, activateAnswer)
      self.answerShown(answerText)
    end)

    newAnswersList[#newAnswersList + 1] = answer
    answer:setID(#newAnswersList)
    answer.layoutIndex = layoutIndex
    index = index + 1
  end

  self.answersList = newAnswersList

  return answersToShow
end

local function setShowAnswerIndex(self, index)
  while (index > #self.answers) do
    index = index - #self.answers
  end
  self.showAnswersIndex = index
  self.currentAnswersShown = createAnswersWithIndex(self, index)
end

local function cancelAnimation(self)
  if (self.animation) then
    self.animation:stop()
    self.animation = nil
  end
end

local function clearAnswers(self)
  if (self.answersGroup) then
    self.answersGroup:removeSelf()
    self.answersGroup = nil
    self.answersList = nil
  end
  if (self.answersList) then
    for _, answer in ipairs(self.answersList) do
      answer:removeSelf()
    end
  end
  self.oneAnswerSetCount = 0
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function AnswersView:setAnswers(answers, colorIndex)
  clearAnswers(self)
  answers = TableUtils.shuffleArray(answers)

  local answersGroup = Group.newFullScreen()
  local answersList = {}

  local layoutSet = StalkDescriptors.getRandomLayoutSet()

  self.answers = answers
  self.layoutSet = layoutSet

  self:insertAtCenter(answersGroup)

  self.answersList = answersList
  self.answersGroup = answersGroup
  self.showAnswersIndex = 1
  self.currentAnswersShown = 0
  self.colorIndex = colorIndex
end

function AnswersView:changeShownAnswerSet()
  if (self.lastSpringInAnimation and self.lastSpringInAnimation:isActive()) then
    return
  end

  setShowAnswerIndex(self, self.showAnswersIndex + self.currentAnswersShown or 0)
end

function AnswersView:playCorrectAnimation(answerID)
  local animation = nil
  for index, answer in ipairs(self.answersList) do
    answer:deactivate()

    local completeHandler = function() answer:removeSelf() end
    if (index == answerID) then
      answer:bringToFront()
      answer:stretchOut(completeHandler)
    else
      local springOutAnim = answer:springOut(nil, completeHandler)
      animation = springOutAnim
    end
  end
  return animation
end

function AnswersView:playIncorrectAnimation(answerID)
  local animation = nil
  local stagger = 0.03

  for index, answer in ipairs(self.answersList) do
    answer:deactivate()
    local completeHandler =  function() answer:removeSelf() end
    if (index == answerID) then
      answer:bringToFront()
      answer:droopOut(completeHandler)
    else
      local springOutAnim = answer:springOut((index - 1) * stagger + 0.3, completeHandler)
      animation = animation or springOutAnim
    end
  end
  return animation
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function AnswersView.new(overlayGroup, callbacks)
  local self = Group.newFullScreen()
  Mixin.mix(self, AnswersView)

  self.overlayGroup = overlayGroup

  self.answerSelectHandler = function(answerButton)
    local answerData = {
      answer = answerButton:getAnswer(),
      answerID = answerButton:getID()
    }
    callbacks.answerSubmitted(answerData)
  end

  self.oneAnswerSetCount = 0
  self.answerShown = callbacks.answerShown

  return self
end
