class("BeanStalk")
require("scene.shapes.DynamicSpline")
require("data.StalkDescriptors")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setVertexPosition(self, index, x, y)
  self.points[index][1], self.points[index][2] = x, y
  self.line:setPoints(self.points)
  if (index == #self.points) then
    self.attachment:setLoc(x + 20, y + (self.dy or 0))
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function BeanStalk:setMiddlePointLocation(x, y)
  setVertexPosition(self, 2, x, y)
end

function BeanStalk:setTopPointLocation(x, y)
  setVertexPosition(self, 3, x, y)
end

function BeanStalk:animate(animationClass, params, completeHandler)
  if (self.animation) then
    self.animation:stop()
  end
  self.animation = animationClass.new(self, params)
  return self.animation:start(nil, completeHandler)
end

function BeanStalk:attachObject(object)
  self.attachmentView = object
  self.attachment:insertAtCenter(object)
  object:setRot(-self.rotation)
end

function BeanStalk:getPoints()
  return TableUtils.deepCopy(self.points)
end

function BeanStalk:getShape()
  return TableUtils.deepCopy(self.shape)
end

function BeanStalk:setPointLocations(locations)
  for index, point in ipairs(locations) do
    setVertexPosition(self, index, point[1], point[2])
  end
end

function BeanStalk:translateStem(dx, dy)
  self.line:setLoc(Screen.width/2 + dx, Screen.height/2 + dy)
  self.dx, self.dy = dx, dy
end

function BeanStalk:setAttachmentScale(scale)
  self.attachment:setScl(scale)
end

function BeanStalk:setAttachmentRotation(rotation)
  self.attachment:setRot(rotation)
end

function BeanStalk:getLastSegmentAngle()
  local lastPoint = self.points[#self.points]
  local prevPoint = self.points[#self.points - 1]

  local angle = MathUtils.angleBetweenVectors(prevPoint[1], prevPoint[2], 0, Screen.height, lastPoint[1], lastPoint[2])
  return angle
end

function BeanStalk:hide()
  self.line:setColor(0, 0, 0, 0)
end

function BeanStalk:show()
  self.line:setColor(1, 1, 1, 1)
end

-- Returns the height percentage of the maximum height for the beanstalk.
-- This parameter can be used to query properties of the bean stalk at different heights
function BeanStalk:getHeightPercent(height)
  return height/Screen.height * 0.5
end

function BeanStalk:getCurvatureAtHeightPercent(heightPercent)
  return heightPercent * self.curvatureAmplitude + self.shape[2][1]
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function BeanStalk.new(stalkParams)
  local self = Group.newDeprecated()
  Mixin.mix(self, BeanStalk)

  local length = stalkParams.length
  local curvature = stalkParams.curvature or MathUtils.randomSign() * (math.random() * 12 + 14)
  local topXOffset = stalkParams.topXOffset or 0
  local middleYOffset = stalkParams.middleYOffset or 0
  local points = {{0, 0}, {curvature, length/2 + middleYOffset}, {topXOffset, length}}
  local rotation = stalkParams.rotation or 0

  local pointsPerSegment = 24
  local tension = -0.2
  local view = DynamicSpline.new(pointsPerSegment, tension, StalkDescriptors.STALK_THICKNESS, points)
  local attachment = Group.newMinimumSized()
  view:setLoc(Screen.width/2, length)

  self:insert(view)
  self:insert(attachment, 100)
  self:setRot(rotation)

  local attachmentXNoise = math.random() * 24 - 12
  attachment:setLoc(points[#points][1], points[#points][2] + attachmentXNoise)

  self.line = view
  view:setAttrLink(MOAIColor.INHERIT_COLOR, self, MOAIColor.COLOR_TRAIT)
  self.points = points
  self.attachment = attachment
  self.startAngle = self:getLastSegmentAngle()

  -- self.points can change, but always keep a reference to the original points, to be able to calculate the current offsets and such
  self.shape = TableUtils.deepCopy(points)
  self.curvatureAmplitude = points[2][1]
  self.rotation = rotation

  return self
end
