--- StalkForeshadowBackground
-- Provides a spotlight overlay to place over the scene that can be activated/deactivated.

class("StalkForeshadowBackground")
require("widgets.AlertPulse")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OPACITY_VISIBLE = 0.7
local OPACITY_INVISIBLE = 0

local SPRING_IN_DURATION = 0.6
local SPRING_OUT_DURATION = 0.3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function stopTapToContinue(self)
  if self.tapAction then
    self.tapAction:stop()
    self.tapAction = nil
  end
  if self.pulseColumn then
    self.pulseColumn:removeSelf()
    self.pulseColumn = nil
  end
end

local function getTapText(isFirstRound)
  if isFirstRound then
    return Localizer.get("TIP")
  end

  return Localizer.get("Misc.TAP_TO_CONTINUE")
end

local function runTapToContinue(self, isFirstRound)
  local tipText = getTapText(isFirstRound)
  local label = DynamicTextArea.newSimple(tipText,
                                          Screen.width,
                                          Font.new(FontFaces.DIN_OT_MEDIUM, 14),
                                          Alignment.CENTER)

  local pulse = AlertPulse.new(nil, nil, 0.8, 0.2)

  local columnSpacing = 20
  local column = Column.new(columnSpacing)
  column:insert(label)
  column:insert(pulse)
  column:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)
  column:alignChildren()
  column:setOpacity(0)
  column:setLoc(Screen.width/2, math.floor(Screen.height * 0.25))
  self:insert(column)

  self.pulseColumn = column
  self.tapAction = runInCoroutine(function()
    waitFor(0.2)
    wait(AHOpacityTo.new(self.pulseColumn, 0.5, nil, 0.9):start())

    for i = 1, 10 do
      wait(pulse:pulse())
      waitFor(0.1)
    end
  end)
  return self.tapAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function StalkForeshadowBackground:springIn(shouldShowTapMessage, isFirstRound)
  return runInCoroutine(function()
    if (self.onScreen) then return end
    self.onScreen = true

    if shouldShowTapMessage then
      runTapToContinue(self, isFirstRound)
    end

    local lastAction = nil
    for _, overlay in ipairs(self.overlays) do
      lastAction = AHOpacityTo.new(overlay, SPRING_IN_DURATION, nil, OPACITY_VISIBLE):start()
    end
    wait(lastAction)
  end)
end

function StalkForeshadowBackground:springOut()
  if (not self.onScreen) then return end
  self.onScreen = false
  stopTapToContinue(self)

  for _, overlay in ipairs(self.overlays) do
    AHOpacityTo.new(overlay, SPRING_OUT_DURATION, AHEaseSimpleIn.new(2), OPACITY_INVISIBLE):start()
  end
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local scaleFactors = { { x=1, y=1 }, { x=-1, y=1 }, { x=-1, y=-1 }, { x=1, y=-1 } }
  local overlays = {}

  for _, scaleFactor in ipairs(scaleFactors) do
    local overlay = Sprite.newWithName("spotlight_corner.png")
    local imageW, imageH = overlay:getDims()
    local widthRatio = Screen.width/(2 * imageW)
    local heightRatio = Screen.height/(2 * imageH)

    overlay:setAnchorPointBottomLeft()
    overlay:setScl(scaleFactor.x * widthRatio, scaleFactor.y * heightRatio)
    overlay:setOpacity(0)

    self:insertAtCenter(overlay)
    table.insert(overlays, overlay)
  end

  self.overlays = overlays
end

function StalkForeshadowBackground.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, StalkForeshadowBackground)
  createLayout(self)

  return self
end
