--- QuestionView
-- @author Created by alexeychernikov on November 03, 2013.

class("QuestionView")

require("actions.advanced.AHAxisOffsetShake")
require("scene.answers.StalkForeshadowBackground")
require("scene.question.CollectionOfWordsWithBlanks")
require("scene.question.PounceRationaleView")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {0/255, 177/255, 193/255}
local CORRECT_ANSWER_BACKGROUND_COLOR = {0/255, 177/255, 193/255}
local WRONG_ANSWER_BACKGROUND_COLOR = {1.0, 1.0, 1.0}
local REVEALED_ANSWER_TEXT_COLOR = {0.65, 0.65, 0.65}

local DEFAULT_TEXT_COLOR = {1, 1, 1}

local Z_INDEX =
{
  rationale = 1,
  background = 2,
  textView = 3
}

local BLANK_STRING = "___________"

local MIN_FONT_SIZE = 15
local MAX_FONT_SIZE = 20
local LABEL_X_MARGIN = math.floor(0.09 * Screen.width)
local LABEL_Y_MARGIN = 36
local BACKGROUND_WIDTH = math.floor(0.85 * Screen.width)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function clearQuestion(self)
  if (self.textView) then
    self.textView:removeSelf()
    self.textView = nil
  end
  if (self.answerView) then
    self.answerView:removeSelf()
    self.answerView = nil
  end
  if (self.underscore) then
    self.underscore:removeSelf()
    self.underscore = nil
  end
end

local function createRevealAnimation(self, submittedAnswer, isCorrect, scene)
  local answerShowAnim = self.answers:getRevealAnswerAnimation(submittedAnswer)

  local inertiaTime = 0.3
  local inertiaMoveAmount = 60.0

  local moveView = AHLocationYBy.new(self, inertiaTime, AHEaseSimpleIn.new(), inertiaMoveAmount)
  local moveAnswerDownASmidge = self.answers:getSettleAnswersDownAnimation(inertiaTime)
  local inertia = AHSpawn.new(
    moveAnswerDownASmidge,
    moveView
  )

  local delayedShowRationale = function()
    if (not isCorrect) then
      local rationaleDelay = 0.1
      waitFor(rationaleDelay)
      self.rationale:setSubmittedAnswer(submittedAnswer)
      inertia:start()
      local shouldShowTapMessage = false
      scene:showSpotlightBackground(shouldShowTapMessage)

      wait(self.rationale:show())

      local tapToContinue = TapToContinue.new()
      tapToContinue:waitForTrigger()
      tapToContinue:removeSelf()
      self.rationale:hide()
      scene:hideSpotlightBackground()
    end
  end

  local spawnActions = {}

  if (submittedAnswer ~= nil) then
    table.insert(spawnActions, answerShowAnim)
  end

  if (not isCorrect) then
    local shakeWidth = self:getWidth()
    local shakeAxis = AHAxisOffsetShake.SHAKE_AXIS.X
    local shake = AHAxisOffsetShake.new(self, shakeWidth, shakeAxis)
    table.insert(spawnActions, shake)
  end

  local showAnswerWithInertia = AHSpawn.new(spawnActions)

  local actions = {
    showAnswerWithInertia,
    delayedShowRationale
  }

  local waitTime = 0.7
  local hangAround = AHDelay.new(waitTime)
  if (isCorrect) then
    table.insert(actions, hangAround)
  end

  local animation = AHSequence.new(actions)
  return animation
end

local function cancelAnimation(self)
  if (self.animation) then
    self.animation:stop()
    self.animation = nil
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function QuestionView:setQuestion(roundInfo)
  local question = roundInfo.question
  local answer = roundInfo.correct

  clearQuestion(self)

  local questionWithBlank = string.gsub(question, "_", BLANK_STRING)
  local width, height = self:getDims()
  local xMargin, yMargin = LABEL_X_MARGIN, LABEL_Y_MARGIN
  width, height = width - xMargin, height - yMargin
  local horizontalAlignment = Alignment.CENTER
  local verticalAlignment = Alignment.CENTER
  local lineBreakMode = MOAITextBox.WORD_BREAK_NONE
  local willBuildGlyphTable = true

  local textView = DeprecatedLabel.newMultiLineFitted(questionWithBlank,
                                                      width,
                                                      height,
                                                      FontFaces.DEFAULT_FONT,
                                                      MAX_FONT_SIZE,
                                                      MIN_FONT_SIZE,
                                                      horizontalAlignment,
                                                      verticalAlignment,
                                                      lineBreakMode,
                                                      willBuildGlyphTable)

  local fontSize = textView.fontRenderEngine:getFontSize()
  local glyphs = textView:getGlyphTable()
  local h = textView:getHeight()
  local answers = CollectionOfWordsWithBlanks.new(answer, glyphs, fontSize, h)

  textView:setColor(unpack(DEFAULT_TEXT_COLOR))
  textView:setLoc(0, 5)

  textView:insert(answers)
  self:insertAtCenter(textView, Z_INDEX.textView)

  self.background:setColor(unpack(BACKGROUND_COLOR))

  self.question = question
  self.answers = answers
  self.textView = textView

  local definitions = {
    [roundInfo.correct] = roundInfo.correct_definition,
    [roundInfo.incorrect_1] = roundInfo.incorrect_definition_1,
    [roundInfo.incorrect_2] = roundInfo.incorrect_definition_2,
    [roundInfo.incorrect_3] = roundInfo.incorrect_definition_3,
    [roundInfo.incorrect_4] = roundInfo.incorrect_definition_4
  }

  self.rationale:setAnswer(answer)
  self.rationale:setDefinitions(definitions)
end

function QuestionView:revealAnswer(scene, isCorrect, submittedAnswer)
  local textColor = REVEALED_ANSWER_TEXT_COLOR
  local backgroundColor = WRONG_ANSWER_BACKGROUND_COLOR

  if (isCorrect) then
    textColor = DEFAULT_TEXT_COLOR
    backgroundColor = CORRECT_ANSWER_BACKGROUND_COLOR
  end

  local fadeDuration = 0.3
  local fadeBackground = AHColorTo.new(self.background, fadeDuration, nil, backgroundColor)
  local fadeText = AHColorTo.new(self.textView, fadeDuration, nil, textColor)
  fadeBackground:start()
  fadeText:start()

  local animation = createRevealAnimation(self, submittedAnswer, isCorrect, scene)
  self.animation = animation
  return animation:start()
end

function QuestionView:maximize()
  if (self.maximized) then return end
  cancelAnimation(self)
  self.maximized = true

  local h = self:getHeight()

  self:setLoc(Screen.width/2, -h/2)
  self:setRot(6)
  self:setScl(1)
  local animTime = 0.5

  local rotation = AHRotationTo.new(self, animTime, AHEaseSimpleInOut.new(), 0)
  local movement = AHLocationTo.new(self, 0.5, AHEaseSimpleInOut.new(), Screen.width / 2, Screen.height / 2)
  local animation = AHSpawn.new(
                                rotation,
                                movement
                                )

  self.animation = animation
  self.animationProcess = self.animation:start()

  self.background:setColor(unpack(BACKGROUND_COLOR))
end

function QuestionView:minimize()
  if (not self.maximized) then return end
  cancelAnimation(self)
  self.maximized = false

  local animTime = 0.28
  local rotTime = animTime * 0.3

  local scaleDown = AHScaleTo.new(self, animTime / 2, nil, 0.93)
  local scaleBack = AHScaleTo.new(self, animTime / 2, nil, 1)
  local scale = AHSequence.new(
                               scaleDown,
                               scaleBack
                               )

  local movement = AHLocationTo.new(self, animTime, AHEaseBackIn.new(0.9), Screen.width / 2, Screen.height / 2 + 10)
  local rotationAnticipation = AHRotationTo.new(self, rotTime, AHEaseSimpleIn.new(), 3)
  local rotationRestore = AHRotationTo.new(self, animTime - rotTime, nil, 0)
  local rotation = AHSequence.new(
                                  rotationAnticipation,
                                  rotationRestore
                                  )

  Sound.playSoundEffect("tap_start")
  rotation:start()
  self.animation = AHSpawn.new(
                               scale,
                               movement
                               )
  self.animationProcess = self.animation:start()
end

function QuestionView:slideOut()
  cancelAnimation(self)
  self.maximized = false

  local h = self:getHeight()
  local animTime = 0.4
  local rotTime = animTime * 0.3

  local scale = AHScaleTo.new(self, animTime, AHEaseBackIn.new(), 0.95)
  local movement = AHLocationTo.new(self, animTime, AHEaseBackIn.new(0.9), Screen.width / 2, Screen.height + h / 2 + 30)
  local rotationAnticipation = AHRotationTo.new(self, rotTime, AHEaseSimpleIn.new(), -3)
  local rotationRestore = AHRotationTo.new(self, animTime - rotTime, nil, 0)
  local rotation = AHSequence.new(
                                  rotationAnticipation,
                                  rotationRestore
                                  )

  rotation:start()
  self.animation = AHSpawn.new(
                               scale,
                               movement
                               )
  self.animationProcess = self.animation:start()
  return self.animationProcess
end

function QuestionView:isReadyForAction()
  return (self.animationProcess == nil) or (not self.animationProcess:isActive())
end

function QuestionView:getRationaleHeight()
  if (self.rationale ~= nil) then
    return self.rationale:getHeight()
  end

  return 0
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createBackground()
  local background = Sprite.newWithName("pounce_question_box.png")
  background:setEffectiveDims(BACKGROUND_WIDTH)

  return background
end

local function createLayout(background)
  local self = Group.new(background:getEffectiveWidth(), background:getEffectiveHeight())

  background:setColor(unpack(BACKGROUND_COLOR))
  self:insertAtCenter(background, Z_INDEX.background)
  background:setAttrLink(MOAIColor.INHERIT_COLOR, self, MOAIColor.COLOR_TRAIT)
  self.background = background

  local rationaleWidth = self:getWidthPart(0.975)
  local rationaleHeight = self:getHeightPart(0.9)
  self.rationale = PounceRationaleView.new(rationaleWidth, rationaleHeight)
  self:insertAtCenter(self.rationale, Z_INDEX.rationale)

  return self
end

function QuestionView.new()
  local background = createBackground()

  local self = createLayout(background)
  Mixin.mix(self, QuestionView)

  return self
end
