--- HexagonBackground
-- @author Created by Anton Morant on March 11, 2014.

class("HexagonBackground")
require("widgets.progress.ProgressIndicator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BORDER_WIDTH = 1
local HEXAGON_RADIUS = 24

local Z_BACKGROUND = 0
local Z_HEXAGONS = 10
local Z_OVERLAY = 20

local BORDER_COLOR = {  22/255,  126/255,  135/255 }

local BACKGROUND_COLOR_TOP = {  13/255,   36/255,   37/255 }
local BACKGROUND_COLOR_BOTTOM = {   4/255,   11/255,   21/255 }
local OVERLAY_COLOR = { 255/255,  255/255,  255/255, 0.5 }

local HIGHLIGHT_DURATION = 0.2
local HIGHLIGHT_MAX_DELAY = 0.7
local HIGHLIGHT_END_DELAY = 0.1
local DRAIN_TOTAL_DURATION = 1
local DRAIN_DURATION = 0.2
local DRAIN_MAX_DELAY = 0.7
local DRAIN_END_DELAY = 0.1

local COLOR_PALETTE_BOTTOM = {
  { 12/255 , 96/255,  99/255 },
  {  4/255,  61/255,  72/255 },
  {  9/255,  73/255,  80/255 },
  {  5/255,  50/255,  59/255 },
  {  5/255,  59/255,  70/255 },

  {  2/255,  69/255,  70/255 },
  {  4/255,  62/255,  73/255 },
  { 12/255,  96/255, 102/255 },

  {  9/255,  74/255,  80/255 },
  {  7/255,  53/255,  63/255 },
  {  5/255,  48/255,  60/255 },
  { 10/255,  85/255,  93/255 },

  { 14/255,  59/255,  72/255 },
  {  6/255,  48/255,  59/255 },
  {  9/255,  66/255,  79/255 },
}

local COLOR_PALETTE_CENTER = {
  {  8/255,  46/255,  59/255 },
  {  5/255,  48/255,  63/255 },
  {  6/255,  49/255,  66/255 },

  {  4/255,  47/255,  63/255 },
  {  7/255,  56/255,  77/255 },
  {  3/255,  45/255,  63/255 },
  {  5/255,  63/255,  77/255 },

  {  4/255,  47/255,  64/255 },
  {  3/255,  60/255,  85/255 },
  {  4/255,  48/255,  67/255 },
  {  6/255,  49/255,  63/255 },

  {  4/255,  47/255,  64/255 },
  {  3/255,  40/255,  56/255 },
  {  1/255,  54/255,  77/255 },
  {  3/255,  45/255,  62/255 },
}

local COLOR_PALETTE_TOP = {
  {  1/255,  54/255,  77/255 },
  {  4/255,  47/255,  65/255 },
  {  3/255,  40/255,  56/255 },

  {  3/255,  46/255,  65/255 },
  {  9/255,  43/255,  59/255 },
  {  2/255,  55/255,  78/255 },
  {  3/255,  47/255,  67/255 },

  {  8/255,  46/255,  60/255 },
  {  2/255,  60/255,  85/255 },
  { 13/255,  58/255,  71/255 },
  {  3/255,  41/255,  58/255 },

  {  4/255,  44/255,  61/255 },
  {  8/255,  45/255,  58/255 },
  {  5/255,  41/255,  57/255 },
  {  3/255,  41/255,  57/255 },
}

local SHINE_COLORS = {
  {  1/255, 128/255, 139/255 },
  {  0/255, 201/255, 196/255 },
  {  1/255, 190/255, 186/255 },
  {  3/255, 147/255, 150/255 },
  {  1/255, 176/255, 174/255 },
  {  3/255, 252/255, 246/255 },
  {  0/255, 234/255, 229/255 },
  {  4/255,  78/255,  85/255 },
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Sounds -----------------------------

local function playProgressSound()
  Sound.playSoundEffect("progress")
end

-- Hexagon creation -------------------

local shuffledPalettes = {}
local lastColorIndices = {}

local function getNextColorIndex(palette)
  if not shuffledPalettes[palette] then
    shuffledPalettes[palette] = TableUtils.shuffleArray({unpack(palette)})
    lastColorIndices[palette] = 0
  end

  local colorIndex = lastColorIndices[palette] + 1
  local color = shuffledPalettes[palette][colorIndex]

  if colorIndex == #palette then
    shuffledPalettes[palette] = nil
    lastColorIndices[palette] = nil
  else
    lastColorIndices[palette] = colorIndex
  end

  return color
end

local function getColor(y, height)
  local x = MathUtils.clamp(y/height, 0, 1)
  local palette = nil
  local flipCoin = math.random(2) == 1

  -- This code divides the background tiling into 5 bands.
  -- The bottom, middle and top band use colors from their respective COLOR_PALETTE.
  -- The two remaining bands, i.e. between bottom and middle and between middle and top, use
  -- a mix of colors from their neighbour palettes, using flipCoin to randomize.
  -- Figures are chosen arbitrarily for the desired look.
  if x < 1/6 or (x < 2/6 and flipCoin) then
    palette = COLOR_PALETTE_BOTTOM
  elseif x < 3/6 or (x < 4/6 and flipCoin)  then
    palette = COLOR_PALETTE_CENTER
  else
    palette = COLOR_PALETTE_TOP
  end

  return getNextColorIndex(palette)
end

local function newHexagon(radius, fillColor)
  local sideCount = 6
  local startAngle = 0

  local hexagon = FilledRegularPolygon.new(radius, sideCount, startAngle, fillColor)
  local hexagonBorder = HollowRegularPolygon.new(radius + BORDER_WIDTH/2, sideCount, startAngle, BORDER_COLOR, BORDER_WIDTH)
  hexagon:insertAtCenter(hexagonBorder)

  return hexagon
end

local function createHexagons(self)
  local hexagons = {}
  local numHexagons = 0

  local dx = 2 * self.radius * math.cos(math.pi/6)
  local dy = dx * math.sin(math.pi/3)

  local neededWidth = self.width - dx/2
  local neededHeight = self.height - dy/2

  local row = 1
  local y = -dy

  while y < neededHeight do
    hexagons[row] = {}
    y = y + dy

    local col = 1
    local x = -dx

    if row % 2 == 0 then
      x = x + dx/2
    end

    while x < neededWidth do
      x = x + dx

      local color = getColor(y, self.height)
      local hexagon = newHexagon(self.radius, color)
      hexagon:setOpacity(0)
      hexagon:setLoc(x, y)
      hexagon:setRot(90)
      hexagons[row][col] = hexagon
      self:insert(hexagon, Z_HEXAGONS)
      numHexagons = numHexagons + 1

      col = col + 1
    end

    row = row + 1
  end

  self.hexagons = hexagons
  self.numHexagons = numHexagons
end

-- Animation: Highlight progress ------

local function getHighlightAction(self, hexagon)
  assert(not hexagon.isHighlighted, "This hexagon is already highlighted")

  hexagon.isHighlighted = true
  self.numHighlighted = self.numHighlighted + 1

  local opacityChange = AHOpacityTo.new(hexagon, HIGHLIGHT_DURATION, nil, 1)
  local delay = AHDelay.new(math.random() * HIGHLIGHT_MAX_DELAY)

  return AHSequence.new(delay, opacityChange)
end

local function highlightHexagons(self, targetNumHighlighted)
  assert(targetNumHighlighted <= self.numHexagons, "Cannot highlight more hexagons than we have.")

  local highlights = AHSpawn.new()
  highlights:addAction(playProgressSound)

  local endDelay = AHDelay.new(HIGHLIGHT_END_DELAY)

  for _, row in ipairs(self.hexagons) do
    local shuffledRow = TableUtils.shuffleArray({unpack(row)})

    for _, hexagon in ipairs(shuffledRow) do
      if not hexagon.isHighlighted then
        local highlightAction = getHighlightAction(self, hexagon)
        highlights:addAction(highlightAction)

        if self.numHighlighted >= targetNumHighlighted then
          return AHSequence.new(highlights, endDelay):start()
        end
      end
    end
  end

  assert(false, "Logic fail. We should never get here!")
end

-- Animation: Progress complete -------

local function getShineAction(self, hexagon, extraDelayTime)
  local originalColor = table.pack(hexagon:getColor())
  local newColor = SHINE_COLORS[math.random(#SHINE_COLORS)]
  local shineIn = AHColorTo.new(hexagon, 0.2 + math.random() * 0.1, nil, newColor)
  local shineOut = AHColorTo.new(hexagon, math.random() * 0.8, nil, originalColor)
  local delay = AHDelay.new(extraDelayTime + math.random() * 0.2)

  return AHSequence.new(delay, shineIn, shineOut)
end

local function getShineAnimation(self)
  local shineAnimation = AHSpawn.new()

  local SHINE_DURATION = 1
  local delayFactor = SHINE_DURATION / #self.hexagons
  local delayScale = #self.hexagons / #self.hexagons[1]

  for i, row in ipairs(self.hexagons) do

    for j, hexagon in ipairs(row) do
      local delay = (i + j/delayScale - 1) * delayFactor
      local highlightAction = getShineAction(self, hexagon, delay)
      shineAnimation:addAction(highlightAction)
    end
  end

  return shineAnimation
end

-- Animation: Drain all highlighted ---

local function getDrainAction(self, hexagon)
  assert(hexagon.isHighlighted, "This hexagon is not highlighted yet")

  hexagon.isHighlighted = false
  self.numHighlighted = self.numHighlighted - 1

  local opacityChange = AHOpacityTo.new(hexagon, DRAIN_DURATION, nil, 0)
  local delay = AHDelay.new(math.random() * DRAIN_MAX_DELAY)

  return AHSequence.new(delay, opacityChange)
end

local function getDrainAnimation(self)
  local drains = AHSpawn.new()
  local endDelay = AHDelay.new(DRAIN_END_DELAY)
  local delayFactor = DRAIN_TOTAL_DURATION / #self.hexagons

  for i = #self.hexagons, 1, -1 do
    local row = self.hexagons[i]
    local rowDelay = (#self.hexagons - i) * delayFactor

    for _, hexagon in ipairs(row) do
      if hexagon.isHighlighted then
        local drainAction = getDrainAction(self, hexagon)
        local delay = AHDelay.new(rowDelay)
        drains:addAction(AHSequence.new(delay, drainAction))
      end
    end
  end

  return AHSequence.new(drains, endDelay)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function HexagonBackground:setProgress(progress)
  self.super.setProgress(self, progress)
  progress = self:getProgress()

  local hexagonsToHighlight = math.floor(progress * self.numHexagons)
  self.animation = highlightHexagons(self, hexagonsToHighlight)
end

function HexagonBackground:waitForHighlightAnimation()
  if self.animation then
    wait(self.animation)
  end
end

function HexagonBackground:doShineAnimation()
  self:waitForHighlightAnimation()
  return getShineAnimation(self):start()
end

function HexagonBackground:doDrainAnimation()
  return getDrainAnimation(self):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createBackground(self)
  local background = RectangleGradient.new(Screen.width, Screen.height, BACKGROUND_COLOR_TOP, BACKGROUND_COLOR_BOTTOM)
  self:insertAtCenter(background, Z_BACKGROUND)
  self.background = background
end

local function createOverlay(self)
  local overlay = Sprite.newWithName("foreshadow.png")
  local imageW, imageH = overlay:getDims()
  local scaleW = Screen.width / imageW
  local scaleH = Screen.height / imageH

  overlay:setScl(scaleW, scaleH)
  overlay:setColor(unpack(OVERLAY_COLOR))
  -- This gives us the desired multiply blend mode.
  -- MOAIProp2D.BLEND_MULTIPLY, for some reason, does not.
  overlay:setBlendMode(MOAIProp.GL_DST_COLOR, MOAIProp.GL_ONE_MINUS_SRC_ALPHA)
  self:insertAtCenter(overlay, Z_OVERLAY)
  self.overlay = overlay
end

-- Constructor ------------------------

function HexagonBackground.new(width, height, stepCount)
  local self = Group.new(width, height)
  ProgressIndicator.makeProgressIndicator(self, stepCount)
  Mixin.mix(self, HexagonBackground)

  self.width = width
  self.height = height
  self.radius = HEXAGON_RADIUS

  createHexagons(self)
  createBackground(self)
  createOverlay(self)

  self.numHighlighted = 0

  return self
end
