--- PronunciationCard
-- @author Created by karl on July 15, 2014.

class("PronunciationCard")

require("PronunciationCommon")
require("data.visualization.SineWaveVisualizer")
require("data.generators.SoundAveragePowerGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_DEFINITION = ColorUtils.colorHex24ToStd(0xD4ECFF)
local COLOR_TAP_TO_REPEAT = ColorUtils.colorHex24ToStd(0x0ae8f0)
local COLOR_BACKING = ColorUtils.colorWithOpacity(PronunciationCommon.COLOR_OUTLINE, 0.5)

PronunciationCard.WIDTH = math.floor(Screen.width * 0.8)
PronunciationCard.HEIGHT = math.floor(Screen.height * 0.6)


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PronunciationCard:setOutlineOpacity(opacity)
  self.outline:setOpacity(opacity)
end

function PronunciationCard:setConcept(concept)
  self.concept = concept
  self.definitionLabel:setText(concept.definition)
end

function PronunciationCard:stopAudio()
  if self.currentSound then
    self.visualization:stop()
    self.currentSound:stop()
    self.currentSound = nil
  end
end

function PronunciationCard:playAudio(shouldPlayCorrectAudio)
  if not self.currentSound then
    local filename = self.concept.audio_correct_link
    if not shouldPlayCorrectAudio then
      filename = self.concept.audio_incorrect_link
    end
    self.currentSound = AssetHelper.playSoundForConcept(self.concept, filename)
    self.currentSound:addFinishedHandler(function()
      runInCoroutine(function()
        wait(self.visualization:settleAndStop())
        self.currentSound = nil
      end)
    end)
    local generator = SoundAveragePowerGenerator.new(self.currentSound)
    self.visualization:setGenerator(generator)
    self.visualization:start()
  end
end

function PronunciationCard:fadeInContents(duration)
  local action = AHSpawn.new(
    AHOpacityTo.new(self.contents, duration, AHEaseSineInOut.new(), 1),
    AHOpacityTo.new(self.gradient:getEndColor(), duration, AHEaseSineInOut.new(), self.gradient.maxOpacity)
  )
  return action:start()
end

function PronunciationCard:markCorrect()
  self.colorables:setColor(unpack(PronunciationCommon.COLOR_CORRECT))
  self.bottomRect:setColor(unpack(PronunciationCommon.COLOR_CORRECT))
  self.gradient:getEndColor():setColor(unpack(PronunciationCommon.COLOR_CORRECT))
  self.tapLabel:setVisible(false)
end

function PronunciationCard:markIncorrect()
  self.colorables:setColor(unpack(PronunciationCommon.COLOR_INCORRECT))
  self.bottomRect:setColor(unpack(PronunciationCommon.COLOR_INCORRECT))
  self.gradient:getEndColor():setColor(unpack(PronunciationCommon.COLOR_INCORRECT))
  self.tapLabel:setVisible(false)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PronunciationCard.new()
  local self = Group.new(PronunciationCard.WIDTH, PronunciationCard.HEIGHT)
  Mixin.mix(self, PronunciationCard)

  local selfW, selfH = self:getDims()

  local cornerRadius = 4
  local lineWidth = PronunciationCommon.LINE_WIDTH
  local blurMargin = PronunciationCommon.BLUR_MARGIN

  local bg = FilledRoundedRectangle.new(selfW, selfH, cornerRadius, PronunciationCommon.COLOR_BACKGROUND)
  self:insertAtCenter(bg)

  self.colorables = Group.new(selfW, selfH)
  self:insertAtCenter(self.colorables)

  self.outline = HollowRoundedRectangle.new(selfW, selfH, cornerRadius, {1, 1, 1}, lineWidth, blurMargin)
  self.colorables:insertAtCenter(self.outline)

  self.visualization = SineWaveVisualizer.new(selfW-2, 50, {1, 1, 1})
  self.visualization:setAnchorPointBottom()
  self.visualization:setLoc(selfW/2, 50)
  self.visualization:setNumberOfPeaks(6)
  self.colorables:insert(self.visualization)

  self.contents = Group.new(selfW, selfH)
  self:insertAtCenter(self.contents)

  local gradientColor = ColorUtils.colorWithOpacity(COLOR_BACKING, 0)
  self.gradient = RectangleGradient.new(selfW, selfH-2, gradientColor, gradientColor)
  self.gradient:setAnchorPointBottom()
  self.gradient:setLoc(selfW/2, 0)
  self.gradient.maxOpacity = COLOR_BACKING[4] or 1
  self.contents:insert(self.gradient)

  self.bottomRect = FilledRectangle.new(selfW-4, 2, COLOR_BACKING)
  self.bottomRect:setAnchorPointBottom()
  self.bottomRect:setLoc(selfW/2, 0)
  self.contents:insert(self.bottomRect)

  self.definitionLabel = TextArea.new("", Font.new(FontFaces.DIN_OT_LIGHT, 22), math.floor(selfW*0.8), Alignment.CENTER)
  self.definitionLabel:setColor(unpack(COLOR_DEFINITION))
  self.definitionLabel:setAnchorPointTop()
  self.definitionLabel:setLoc(selfW/2, selfH - 24)
  self.contents:insert(self.definitionLabel)

  self.tapLabel = Label.new("tap to repeat", Font.new(FontFaces.DIN_OT_MEDIUM, 12))
  self.tapLabel:setColor(unpack(COLOR_TAP_TO_REPEAT))
  self.tapLabel:setAnchorPointBottom()
  self.tapLabel:setLoc(selfW/2, 10)
  self.contents:insert(self.tapLabel)

  self:linkColorAttrToChildren(true)

  self.colorables:setColor(unpack(PronunciationCommon.COLOR_OUTLINE))

  self.contents:setOpacity(0)

  return self
end
