--- PronunciationCardSurface
-- @author Created by karl on July 16, 2014.

class("PronunciationCardSurface")
require("PronunciationCard")
require("PronunciationSquaresEmitter")
require("PronunciationCardSurfaceButtons")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAX_ROTATION = 45

local Y_START = Screen.height * 0.55 + PronunciationCard.HEIGHT * 0.5
local Y_END = Y_START + Screen.height * 0.08

local SCALE_START = 1.0
local SCALE_END = 0.7

local OPACITY_START = 1.0
local OPACITY_END = 0.0

local Z_EMITTER = 100

local TOLERANCE_TAP = 8
local TOLERANCE_SWIPE = Screen.width * 0.3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isCorrectChoice(self, index)
  return (index == self.correctChoiceIndex)
end

local function playAudio(self)
  self.currentCard:playAudio(self.isCorrect)
end

local function getYForZDistance(distance)
  distance = distance ^ (1/1.2)
  return MathUtils.interpolateLinearFromNormalized(distance, Y_START, Y_END)
end

local function getScaleForZDistance(distance)
  distance = distance ^ (1/1.2)
  return MathUtils.interpolateLinearFromNormalized(distance, SCALE_START, SCALE_END)
end

local function getOpacityForZDistance(distance)
  distance = distance ^ (1/3)
  return MathUtils.interpolateLinearFromNormalized(distance, OPACITY_START, OPACITY_END)
end

local function getCardZDistance(card)
  return card.zDistance
end

local function setCardZDistance(card, zDistance)
  -- Work around a bizarre bug.
  -- Making an AHNonAttributeAction that goes from 1.666... to 0 with a sine-in-out ease
  -- will generate a final value of -2.2351741790771e-08, which breaks the linear interpolation.
  if zDistance < 0 then zDistance = 0 end

  card.zDistance = zDistance
  local cardX = Screen.width / 2
  local cardY = getYForZDistance(zDistance)
  local scale = getScaleForZDistance(zDistance)
  local opacity = getOpacityForZDistance(zDistance)
  card:setLoc(cardX, cardY)
  card:setScl(scale)
  card:setOutlineOpacity(opacity)
end

local function markChoiceActive(self, index)
  local sideColor = self.choiceSides[index].sideColor

  local action = AHSpawn.new(
    AHColorTo.new(sideColor, 0.2, AHEaseSineInOut.new(), PronunciationCommon.COLOR_OUTLINE)
  )
  self.iconActionQueue:insert(action)
end

local function markChoiceInactive(self, sideDuration)
  sideDuration = sideDuration or 0.2
  local action = AHSpawn.new()
  for _, side in ipairs(self.choiceSides) do
    action:addAction(AHColorTo.new(side.sideColor, sideDuration, AHEaseSineInOut.new(), PronunciationCommon.COLOR_CLEAR))
  end
  self.iconActionQueue:insert(action)
end

local function getXDistance(self)
  return self.xDistance
end

local function setXDistance(self, distX, shouldMarkActive)
  self.lastXDistance = self.xDistance or 0
  self.xDistance = distX
  local card = self.currentCard
  local _, cardY = card:getLoc()

  local selfW, selfH = self:getDims()
  local distN = distX / (selfW/2)
  local cardX, cardY = selfW/2 + distX, cardY

  card:setLoc(cardX, cardY)
  card:setRot(-MAX_ROTATION * distN * 0.3)

  if shouldMarkActive then
    if self.xDistance >= TOLERANCE_SWIPE and self.lastXDistance < TOLERANCE_SWIPE then
      markChoiceActive(self, PronunciationCommon.SIDE_INDEX.RIGHT)
    elseif self.xDistance < TOLERANCE_SWIPE and self.lastXDistance >= TOLERANCE_SWIPE then
      markChoiceInactive(self)
    elseif self.xDistance <= -TOLERANCE_SWIPE and self.lastXDistance > -TOLERANCE_SWIPE then
      markChoiceActive(self, PronunciationCommon.SIDE_INDEX.LEFT)
    elseif self.xDistance > -TOLERANCE_SWIPE and self.lastXDistance <= -TOLERANCE_SWIPE then
      markChoiceInactive(self)
    end
  end
end

local function pickUpCard(self)
  if self.dropAction then
    self.dropAction:stop()
  end
  self.currentCard:setScl(1.0)
  self.currentCard:setRot(0)
  local locX, locY = self.currentCard:getLoc()
  self.currentCard:setLoc(self:getWidth()/2, locY)
  AHScaleTo.new(self.currentCard, 0.05, AHEaseSineInOut.new(), 1.04):start()
end

local function dropCard(self)
  self.dropAction = runInCoroutine(function()
    local selfW, selfH = self:getDims()
    local duration = 0.1
    wait(AHSpawn.new(
      AHScaleTo.new(self.currentCard, duration, AHEaseSineInOut.new(), 1.0),
      AHRotationTo.new(self.currentCard, duration, nil, 0),
      AHLocationXTo.new(self.currentCard, duration, nil, selfW/2)
    ):start())
  end)
  return dropAction
end

local function newXDistanceTo(self, duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, getXDistance, setXDistance, ease, amount)
end

local function newCardZDistanceTo(self, duration, ease, amount)
  return AHNonAttributeTo.new(self, duration, getCardZDistance, setCardZDistance, ease, amount)
end

local function slideCurrentCardOut(self)
  return runInCoroutine(function()
    local ease = AHEaseSineIn.new()
    local duration = 0.5
    local distance = self:getWidth() * 2
    if self.chosenIndex == PronunciationCommon.SIDE_INDEX.LEFT then
      wait(newXDistanceTo(self, duration, ease, -distance):start())
    else
      wait(newXDistanceTo(self, duration, ease, distance):start())
    end
    self.currentCard:removeSelf()
    table.remove(self.cards, 1)
  end)
end

local function slideCurrentCardDown(self)
  return runInCoroutine(function()
    local locY = -self:getHeight()
    local duration = 0.8
    self.currentCard:markIncorrect()
    wait(AHLocationYTo.new(self.currentCard, duration, AHEaseSimpleIn.new(), locY):start())
    self.currentCard:removeSelf()
    table.remove(self.cards, 1)
  end)
end

local function setTouchEnabled(self, isTouchEnabled)
  self.isTouchEnabled = isTouchEnabled
  if not isTouchEnabled then
    self.handlingTouch = false
  end
end

local function slideCardsForward(self)
  return runInCoroutine(function()
    local duration = 0.2
    local delay = 0.02
    local ease = AHEaseSineInOut.new()
    local action = AHSpawn.new()
    local zDistance = 0

    local lastAction = nil
    for _, card in ipairs(self.cards) do
      lastAction = newCardZDistanceTo(card, duration, ease, zDistance):start()
      waitFor(delay)
      zDistance = zDistance + self.dZDistance
    end
    wait(lastAction)
  end)
end

local function slideCardsDown(self)
  return runInCoroutine(function()
    local durationDown = 0.5
    local durationUp = 0.3
    local delay = 0.1
    local ease = AHEaseSineInOut.new()
    for _, card in rpairs(self.cards) do
      waitFor(delay)
      local locY = getYForZDistance(card.zDistance)
      local lowLocY = locY - 20
      AHSequence.new(
        AHLocationYTo.new(card, durationDown, ease, lowLocY),
        AHLocationYTo.new(card, durationUp, ease, locY)
      ):start()
    end

    local iconX, iconY = self.iconsGroup:getLoc()
    self.iconsGroup:setLoc(iconX, self:getHeight())
    waitFor(0.1)
    self.iconsGroup:setOpacity(1)
    wait(AHSequence.new(
      AHLocationYTo.new(self.iconsGroup, durationDown, ease, iconY-10),
      AHLocationYTo.new(self.iconsGroup, durationUp, ease, iconY)
    ):start())
  end)
end

local function runCorrectSequence(self)
  return runInCoroutine(function()
    local sideColor = self.choiceSides[self.chosenIndex].sideColor
    local emitter = self.choiceSides[self.chosenIndex].emitter

    sideColor:setColor(unpack(PronunciationCommon.COLOR_CORRECT))
    self.currentCard:markCorrect()
    waitFor(0.2)
    emitter:emit()
  end)
end

local function runIncorrectSequence(self)
  return runInCoroutine(function()
    local sideColor = self.choiceSides[self.chosenIndex].sideColor

    sideColor:setColor(unpack(PronunciationCommon.COLOR_INCORRECT))
    self.currentCard:markIncorrect()
  end)
end

local function runChoiceSequence(self, index)
  if isCorrectChoice(self, index) then
    return runCorrectSequence(self)
  else
    return runIncorrectSequence(self)
  end
end

local function reportChoice(self, index)
  runInCoroutine(function()
    if isCorrectChoice(self, index) then
      self.onCorrectAnswer(self.currentCard.concept)
    else
      self.onIncorrectAnswer(self.currentCard.concept)
    end
  end)
end

local function endRound(self, chosenIndex)

  setTouchEnabled(self, false)
  self.chosenIndex = chosenIndex
end

-- Derived Events ---------------------

local function onChooseDirection(self, index)
  endRound(self, index)
  self.roundEndSequence = runInCoroutine(function()
    local slideAction = slideCurrentCardOut(self)
    runChoiceSequence(self, index)
    markChoiceInactive(self, 0.5)
    reportChoice(self, index)
    wait(slideAction)
  end)
end

local function onChoiceButtonTapped(self, index)
  endRound(self, index)
  self.roundEndSequence = runInCoroutine(function()
    self.iconsGroup:runTapAnimation(index, isCorrectChoice(self, index))
    local slideAction = slideCurrentCardOut(self)
    waitFor(0.1)
    runChoiceSequence(self, index)
    markChoiceInactive(self, 0.5)
    reportChoice(self, index)
    wait(slideAction)
  end)
end

local function onTap(self)
  playAudio(self)
end

-- Primary Events ---------------------

local function onTouchDown(self, x)
  self.handlingTouch = true
  self.touchStartX = x
  pickUpCard(self)
end

local function onTouchMove(self, x)
  if self.handlingTouch then
    local distance = x - self.touchStartX
    local shouldMarkActive = true
    setXDistance(self, distance, shouldMarkActive)

  end
end

local function onTouchUp(self, x)
  if self.handlingTouch then
    local distance = x - self.touchStartX
    local absDistance = math.abs(distance)
    if absDistance < TOLERANCE_TAP then
      onTap(self)
    end
    if absDistance >= TOLERANCE_SWIPE then
      local index = PronunciationCommon.SIDE_INDEX.LEFT
      if distance > 0 then
        index = PronunciationCommon.SIDE_INDEX.RIGHT
      end
      onChooseDirection(self, index)
    else
      wait(dropCard(self))
    end
    self.handlingTouch = false
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PronunciationCardSurface:waitRoundEnd()
  wait(self.roundEndSequence)
end

function PronunciationCardSurface:startRound(concept, isCorrect)
  return runInCoroutine(function()
    self.currentCard = self.cards[1]
    self.currentCard:setConcept(concept)

    local duration = 0.2

    if not self.hasDisplayedFirstRound then
      wait(slideCardsDown(self))
      self.currentCard:fadeInContents(duration)
      self.hasDisplayedFirstRound = true
    else
      wait(slideCardsForward(self))
      self.currentCard:fadeInContents(duration)
    end

    self.xDistance = 0
    self.isCorrect = isCorrect
    if isCorrect then
      self.correctChoiceIndex = PronunciationCommon.SIDE_INDEX.RIGHT
    else
      self.correctChoiceIndex = PronunciationCommon.SIDE_INDEX.LEFT
    end

    playAudio(self)

    setTouchEnabled(self, true)
  end)
end

function PronunciationCardSurface:cardCount()
  return #self.cards
end

function PronunciationCardSurface:fadeIcons()
  return self.iconsGroup:runFadeButtonsAnimation()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makeEmitter(isLeftSide)
  local color = PronunciationCommon.COLOR_CORRECT
  local distance = 70
  local angles = {math.pi * 0.75, math.pi, math.pi * 1.25}
  if isLeftSide then
    angles = {math.pi * 0.25, 0, math.pi * -0.25}
  end

  return PronunciationSquaresEmitter.new(color, distance, angles)
end

local function buildCards(self, cardCount)
  local selfW, selfH = self:getDims()

  self.dZDistance = 1 / (cardCount - 1)
  self.cards = {}
  local zDistance = 0
  for i=1,cardCount do
    local card = PronunciationCard.new()
    card:setAnchorPointTop()
    table.insert(self.cards, card)
    setCardZDistance(card, zDistance)
    zDistance = zDistance + self.dZDistance
    local cardX = card:getLoc()
    card:setLoc(cardX, selfH + card:getHeight() + 10)
  end

  for _, card in rpairs(self.cards) do
    self:insert(card)
  end
end

local function addTouchHandler(self)
  TouchManager.addTouchHandler(self, function(event)
    if self.isTouchEnabled then
      runInCoroutine(function()
        if event.phase == TouchManager.TOUCH_BEGAN then
          onTouchDown(self, event.x)
        elseif event.phase == TouchManager.TOUCH_MOVED then
          onTouchMove(self, event.x)
        elseif event.phase == TouchManager.TOUCH_ENDED then
          onTouchUp(self, event.x)
        end
      end)
    end
  end, TouchHandler.BEHAVIOR_FOCUS)
end

local function makeChoiceSide(self, index)
  local isLeft = index == PronunciationCommon.SIDE_INDEX.LEFT
  local selfW, selfH = self:getDims()
  local sideW = 16
  local sideX, sideY = sideW/2, selfH/2
  local emitterX, emitterY = -25, selfH/2

  local side = RectangleGradient.new(sideW, selfH, PronunciationCommon.COLOR_CLEAR, PronunciationCommon.COLOR_CLEAR, true)
  local emitter = makeEmitter(isLeft)
  local sideColor = side:getBeginColor()

  if not isLeft then
    sideX = selfW - sideX
    emitterX = selfW - emitterX
    sideColor = side:getEndColor()
  end

  side:setLoc(sideX, sideY)
  emitter:setLoc(emitterX, emitterY)

  self:insert(side, Z_EMITTER)
  self:insert(emitter, Z_EMITTER)

  self.choiceSides[index] = {
    sideColor = sideColor,
    emitter = emitter
  }
end

-- Constructor ------------------------

function PronunciationCardSurface.new(cardCount, onCorrectAnswer, onIncorrectAnswer)
  local self = Group.newFullScreen()
  Mixin.mix(self, PronunciationCardSurface)

  self.choiceSides = {}
  makeChoiceSide(self, PronunciationCommon.SIDE_INDEX.LEFT)
  makeChoiceSide(self, PronunciationCommon.SIDE_INDEX.RIGHT)

  local selfWidth = self:getWidth()

  local buttonCallback = function(index)
    if self.isTouchEnabled then
      onChoiceButtonTapped(self, index)
    end
  end
  self.iconsGroup = PronunciationCardSurfaceButtons.new(PronunciationCard.WIDTH, buttonCallback)

  local bottomOfCard = Y_START - PronunciationCard.HEIGHT
  local iconY = (bottomOfCard + Screen.safeBottom) / 2
  self.iconsGroup:setLoc(selfWidth/2, iconY)
  self:insert(self.iconsGroup)

  self.onCorrectAnswer = onCorrectAnswer
  self.onIncorrectAnswer = onIncorrectAnswer

  buildCards(self, cardCount)
  addTouchHandler(self)

  self.iconActionQueue = ActionInterrupt.new()

  return self
end
