--- PronunciationGame
-- @author Created by karl on July 15, 2014.

class("PronunciationGame")
require("bonuses.ResponseSpeedTracker")
require("StandardUIBuilder")
require("data.generators.NumberGenerators")
require("data.generators.UtilityGenerators")
require("PronunciationCommon")
require("PronunciationConfiguration")
require("PronunciationCardSurface")
require("PronunciationIncorrectInterstitial")
require("graphics.TiledImage")
require("PronunciationEndSequence")
require("PronunciationConceptGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {1, 1, 1}

local MAX_RUN_LENGTH = 3

local function configureGameDifficulty(self)
  local rankTwoScore = PronunciationConfiguration.getRankTwoScore()
  local rankThreeScore = PronunciationConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(PronunciationConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startNextRound(self)
  if (self.cardSurface:cardCount() <= PronunciationConfiguration.CONCEPTS_REMAINING_BEFORE_RESURFACING_MISTAKES) then
    self.conceptGenerator:allowMistakesToResurface()
  end

  local isCorrect = self.correctGenerator:next()
  local concept = self.conceptGenerator:next()
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
  wait(self.cardSurface:startRound(concept, isCorrect))
  self.currentConcept = concept
  self.isCurrentConceptCorrect = isCorrect
  self.responseSpeedTracker:notifyStart()
end

local function onGameWin(self)
  self.cardSurface:fadeIcons()

  waitFor(0.2)
  local endSequence = PronunciationEndSequence.new()
  self:insertAtCenter(endSequence)
  Sound.playSoundEffect("pronounciation_win")
  wait(endSequence:play())

  GameSession.notifyGameWin()
end

local function onGameLoss(self)
  GameSession.notifyGameLoss()
end

local function continueLoseOrWin(self)
  if self.livesCounter:getLives() == 0 then
    onGameLoss(self)
  elseif self.cardSurface:cardCount() == 0 then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), PronunciationConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    onGameWin(self)
  else
    startNextRound(self)
  end
end

local function getQuestionIdentifier(self)
  local questionIdentifier = nil
  if (self.isCurrentConceptCorrect) then
    questionIdentifier = "correct_form"
  else
    questionIdentifier = "incorrect_form"
  end

  return questionIdentifier
end

local function onCorrectAnswer(self, concept)
  GameSession.notifyCorrectAnswer(concept.identifier)
  GameSession.incrementCorrectAnswerCount()
  GameSession.contentTracker:notifyCorrectAnswer(concept.identifier, getQuestionIdentifier(self))
  GameSession.increaseScore(PronunciationConfiguration.POINTS_PER_CORRECT_RESPONSE)

  self.responseSpeedTracker:notifyCorrectResponse()
  self.scoreLabel:setScore(GameSession.getScore())
  Sound.playSoundEffect("pronounciation_correct")

  self.cardSurface:waitRoundEnd()
  continueLoseOrWin(self)
end

local function incorrectCommon(self, concept)
  GameSession.scoredIncorrectAnswer(concept.identifier)
  GameSession.incrementIncorrectAnswerCount()
  GameSession.contentTracker:notifyIncorrectAnswer(concept.identifier, getQuestionIdentifier(self))
  self.responseSpeedTracker:notifyIncorrectResponse()
  self.conceptGenerator:notifyMistake(concept.identifier)
  self.livesCounter:decrement()
  Sound.playSoundEffect("pronounciation_incorrect")

  waitFor(0.5)
  local incorrectInterstitial = PronunciationIncorrectInterstitial.new(concept)
  incorrectInterstitial:setOpacity(0)
  self:insertAtCenter(incorrectInterstitial)
  AHOpacityTo.new(self.dimOverlay, 0.3, nil, 0.95):start()
  wait(AHOpacityTo.new(incorrectInterstitial, 0.3, nil, 1):start())
  wait(incorrectInterstitial:run())
  AHOpacityTo.new(self.dimOverlay, 0.3, nil, 0.0):start()
  wait(AHOpacityTo.new(incorrectInterstitial, 0.3, nil, 0):start())
  incorrectInterstitial:removeSelf()
end

local function onIncorrectAnswer(self, concept)
  return runInCoroutine(function()
    incorrectCommon(self, concept)
    continueLoseOrWin(self)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PronunciationGame:gameFlow()
  startNextRound(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local selfW, selfH = self:getDims()

  local background = TiledImage.newWithTexture(TextureLoader.loadTexture("tiles/tile.png"), selfW, selfH)
  self:insertAtCenter(background)

  local cardCount = PronunciationConfiguration.CORRECT_RESPONSES_NEEDED_TO_WIN
  local onCorrect = function(concept) onCorrectAnswer(self, concept) end
  local onIncorrect = function(concept) onIncorrectAnswer(self, concept) end
  self.cardSurface = PronunciationCardSurface.new(cardCount, onCorrect, onIncorrect)
  self:insertAtCenter(self.cardSurface)

  self.dimOverlay = FilledRectangle.new(selfW, selfH, {0, 0, 0, 0})
  self:insertAtCenter(self.dimOverlay)

  local function onQuickResponse(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 PronunciationConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(PronunciationConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       onQuickResponse,
                                                       PronunciationConfiguration.SPEED_BONUS_DECAY_START)

  self.correctGenerator = MaximumRunGenerator.new(RandomBooleanGenerator.new(), MAX_RUN_LENGTH)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(PronunciationConfiguration.INITIAL_LIVES_COUNT)
end

-- Constructor ------------------------

function PronunciationGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, PronunciationGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

  self.conceptGenerator = LearningFocusedRoundGenerator.new(PronunciationConceptGenerator.new())

  return self
end
