--- PronunciationIncorrectInterstitial
-- @author Created by karl on July 22, 2014.

class("PronunciationIncorrectInterstitial")
require("PronunciationCommon")
require("PronunciationPlayIcon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PronunciationIncorrectInterstitial:run()
  return runInCoroutine(function()
    self.correctPronunciation:play()
    self.isContinuePressed = false
    waitUntil(function() return self.isContinuePressed end)
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newPlayFunction(visualizer)
  local currentSound = nil
  return function(concept, filename, color)
    if currentSound then
      visualizer:stop()
      currentSound:stop()
    end
    visualizer:setColor(unpack(color))
    currentSound = AssetHelper.playSoundForConcept(concept, filename)

    local generator = SoundAveragePowerGenerator.new(currentSound)
    visualizer:setGenerator(generator)
    visualizer:start()

    currentSound:addFinishedHandler(function()
      runInCoroutine(function()
        currentSound = nil
        wait(visualizer:settleAndStop())
      end)
    end)

    return runInCoroutine(function()
      local soundSource = currentSound
      waitUntil(function() return soundSource:isFinished() end)
    end)
  end
end

local function newPronunciation(text, pronunciation, color, concept, filename, playFunction)
  local width = math.floor(Screen.width * 0.8)
  local textFontSize = 12
  local pronunciationFontSize = 23
  local letterTracking = 0.7
  local textLabel = Label.new(text, PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, textFontSize), nil, nil, nil, letterTracking)

  local pronunciationLabel = Label.new(pronunciation, Font.new(FontFaces.DIN_OT_LIGHT, pronunciationFontSize))
  local pronunciationLabelWidth = math.min(pronunciationLabel:getWidth(), width)
  pronunciationLabel:setEffectiveDims(pronunciationLabelWidth)

  local playIcon = PronunciationPlayIcon.new(26, color, PronunciationCommon.COLOR_BACKGROUND)

  local columnSpacing = -4
  local col = Column.new(columnSpacing, Alignment.CENTER)
  local extraSpacerSize = pronunciationLabel:getHeight() - pronunciationLabel:getEffectiveHeight() - columnSpacing
  local shouldInsertExtraSpacer = (pronunciationLabel:getScl() < 1)

  col:insert(textLabel)

  col:insertSpacer(Screen.getHeightPart(0.01))

  if (shouldInsertExtraSpacer) then
    col:insertSpacer(extraSpacerSize)
  end

  col:insert(pronunciationLabel)

  col:insertSpacer(Screen.getHeightPart(0.02))

  if (shouldInsertExtraSpacer) then
    col:insertSpacer(extraSpacerSize)
  end

  col:insert(playIcon)

  col:alignChildren()

  textLabel:setColor(unpack(color))
  pronunciationLabel:setColor(unpack(color))

  local group = Group.new(width, col:getHeight())
  group:insertAtCenter(col)

  group.play = function()
    runInCoroutine(function()
      local playAction = playFunction(concept, filename, color)
      -- Delay 2 frames so that the showPressed functions don't fall over each other due to coroutine timing.
      coroutine.yield()
      coroutine.yield()
      playIcon:showPressed(true)
      wait(playAction)
      playIcon:showPressed(false)
    end)
  end

  Button.makeButton(group, group.play)

  group.textLabel = textLabel
  group.pronunciationLabel = pronunciationLabel

  return group
end

local function createPromptColumn(self, word, definition)
  local promptColumn = Column.new()

  local wordLabel = Label.new(word, Font.new(FontFaces.DIN_OT_BOLD, 20))
  self.wordLabel = wordLabel

  local definitionLabelWidth = self:getWidthPart(0.8)
  local definitionLabel = DynamicTextArea.newSimple(definition, definitionLabelWidth, Font.new(FontFaces.DIN_OT_REGULAR, 17), Alignment.CENTER)
  self.definitionLabel = definitionLabel

  promptColumn:insert(wordLabel)
  promptColumn:insertSpacer(self:getHeightPart(0.006))
  promptColumn:insert(definitionLabel)

  promptColumn:alignChildren()

  return promptColumn
end

local function createPronunciationsColumn(self, concept)
  local pronunciationsColumn = Column.new()

  local headerLabel = Label.new("Pronunciation", Font.new(FontFaces.DIN_OT_MEDIUM, 17))
  self.headerLabel = headerLabel

  local groupWidth = self:getWidth()
  local visualizer = SineWaveVisualizer.new(groupWidth, 50, PronunciationCommon.COLOR_CORRECT)
  visualizer:setNumberOfPeaks(4)

  local playFunction = newPlayFunction(visualizer)

  local incorrectPronunciation = newPronunciation("Incorrect",
                                                  concept.incorrect_phonetic,
                                                  PronunciationCommon.COLOR_INCORRECT,
                                                  concept,
                                                  concept.audio_incorrect_link,
                                                  playFunction)
  self.incorrectPronunciation = incorrectPronunciation

  local correctPronunciation = newPronunciation("Correct",
                                                concept.phonetic,
                                                PronunciationCommon.COLOR_CORRECT,
                                                concept,
                                                concept.audio_correct_link,
                                                playFunction)
  self.correctPronunciation = correctPronunciation

  local letterTracking = 2.0
  local continueLabel = Label.new(Localizer.get("Misc.RETURN_TO_GAME"), PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 13), nil, nil, nil, letterTracking)
  continueLabel:setColor(unpack(PronunciationCommon.COLOR_CORRECT))
  self.continueLabel = continueLabel

  pronunciationsColumn:insert(headerLabel)
  pronunciationsColumn:insertSpacer(self:getHeightPart(0.018))
  pronunciationsColumn:insert(incorrectPronunciation)
  pronunciationsColumn:insertSpacer(self:getHeightPart(0.037))
  pronunciationsColumn:insert(correctPronunciation)
  pronunciationsColumn:insertSpacer(self:getHeightPart(0.025))
  pronunciationsColumn:insert(visualizer)
  pronunciationsColumn:insertSpacer(self:getHeightPart(0.03))
  pronunciationsColumn:insert(continueLabel)

  if (Screen.hasNotch()) then
    pronunciationsColumn:insertSpacer(self:getHeightPart(0.04))
  end

  pronunciationsColumn:alignChildren()

  return pronunciationsColumn
end

local function createLayout(self, concept)
  local sprite = Sprite.newWithName("x.png")
  local spriteSideLength = self:getWidthPart(0.165)
  sprite:setEffectiveDims(spriteSideLength, spriteSideLength)
  self:insertAtLocationFractional(sprite, 0.5, 0.85)
  sprite:setColor(unpack(PronunciationCommon.COLOR_INCORRECT))

  local promptColumn = createPromptColumn(self, concept.word, concept.definition)
  promptColumn:setAnchorPointTop()
  self:insertAtLocationFractional(promptColumn, 0.5, 0.77)

  local pronunciationsColumn = createPronunciationsColumn(self, concept)
  pronunciationsColumn:setAnchorPointBottom()
  self:insertAtLocationFractional(pronunciationsColumn, 0.5, 0.04)

  self:linkColorAttrToChildren(true)
end

local function createContinueLabelVirtualButton(self)
  local continueLabel = self.continueLabel
  local buttonWidth = continueLabel:getWidth() * 4
  local buttonHeight = continueLabel:getHeight() * 4
  local onContinueLabelPressed = function()
    self.isContinuePressed = true
  end

  Button.addVirtualButton(continueLabel, buttonWidth, buttonHeight, onContinueLabelPressed)
end

-- Constructor ------------------------

function PronunciationIncorrectInterstitial.new(concept)
  local self = Group.newFullScreen()
  Mixin.mix(self, PronunciationIncorrectInterstitial)

  createLayout(self, concept)

  createContinueLabelVirtualButton(self)

  return self
end
