class("PunctuationAnswerMenu")
require("PunctuationAnswerButton")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MENU_ORDER =
{
  ["comma"] = 1,
  ["colon"] = 2,
  ["semicolon"] = 3,
  ["hyphen"] = 4,
  ["question_mark"] = 5,
  ["period"] = 6,
  ["apostrophe"] = 7,
  ["nothing"] = 8
}

local MAX_OPTIONS_COUNT = 4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getOptionWidthForNumberOfItems(safeWidth, itemCount, layoutHorizontalSpacing)
  return ((safeWidth - ((itemCount + 1) * layoutHorizontalSpacing)) / itemCount)
end

local function getOptionHeightForNumberOfItems(totalHeight, safeAreaOffsetY, layoutVeritcalSpacing)
  local functionalMenuHeight = (totalHeight - safeAreaOffsetY)
  return (functionalMenuHeight - (3  * layoutVeritcalSpacing))
end

local function getOptionSize(safeWidth, totalHeight, safeAreaOffsetY, itemCount, layoutHorizontalSpacing, layoutVeritcalSpacing)
  local optionWidth = getOptionWidthForNumberOfItems(safeWidth, itemCount, layoutHorizontalSpacing)
  local optionHeight = getOptionHeightForNumberOfItems(totalHeight, safeAreaOffsetY, layoutVeritcalSpacing)
  return math.min(optionWidth, optionHeight)
end

local function getAnswerTouchHandler(self)
  return function(answerButton)
    if (not self.enabled) then
      return
    end

    local answer = answerButton:getAnswer()
    if (self.answerSelectHandler) then
      self.answerSelectHandler(answer, answerButton)
    end
  end
end

local function clearOptions(self)
  self.layout:removeAll()
  self.options = nil
end

local function sortOptions(options)
  table.sort(options, function(option1, option2)
    local ordinal1 = MENU_ORDER[option1]
    local ordinal2 = MENU_ORDER[option2]
    return (ordinal1 < ordinal2)
  end)
  return options
end

local function initOptions(self, options, character)
  options = sortOptions(options)
  local columnCount = #options

  local answerHandler = self.answerTouchHandler
  local layout = self.layout

  local widthPerItem = getOptionWidthForNumberOfItems(Screen.safeWidth, columnCount, layout:getSpacingX())
  local heightPerItem = getOptionHeightForNumberOfItems(self:getHeight(), self.safeAreaOffsetY, layout:getSpacingY())

  for _, option in ipairs(options) do
    local answerView = PunctuationAnswerButton.new(widthPerItem, heightPerItem, self.optionSize, option, character, answerHandler)
    layout:insert(answerView)
    table.insert(self.items, answerView)
    answerView:setTouchEnabled(self.enabled)
  end

  layout:setColumnCount(columnCount)
  layout:alignChildren()

  local isRecursive = true
  layout:linkOpacityAttrToChildren(isRecursive)

  self.options = options
end

local function createLayout(self, width, height, safeAreaOffsetY)
  local backingHeight = Screen.getHeightPart(0.5)

  local backgroundColor = PunctuationColors.getMenuBackgroundColor()
  local background = FilledRectangle.new(width, height, backgroundColor)
  local backingToBottom = FilledRectangle.new(width, backingHeight, backgroundColor)
  local layout = Grid.new()

  local dx = Screen.getWidthPart(0.0083)
  local dy = Screen.getHeightPart(0.0045)
  layout:setSpacingX(dx)
  layout:setSpacingY(dy)

  local x0 = (width / 2)
  local y0 = (height / 2)
  background:setLoc(x0, y0)

  local layoutY = y0 + ((dy + safeAreaOffsetY) / 2)
  layout:setLoc(x0, layoutY)

  local backingToBottomY = y0 - (backingHeight / 2) - (height / 2)
  backingToBottom:setLoc(x0, backingToBottomY)

  self:insert(backingToBottom)
  self:insert(background)
  self:insert(layout)

  self:linkColorAttrToChildren(true)

  self.answerTouchHandler = getAnswerTouchHandler(self)
  self.layout = layout

  self.optionSize = getOptionSize(Screen.safeWidth, height, safeAreaOffsetY, MAX_OPTIONS_COUNT, self.layout:getSpacingX(), self.layout:getSpacingY())
end
-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------
function PunctuationAnswerMenu:setAnswerSelectHandler(handler)
  self.answerSelectHandler = handler
end

function PunctuationAnswerMenu:disable()
  self.enabled = false
  for _, item in pairs(self.items) do
    item:setTouchEnabled(false)
  end
end

function PunctuationAnswerMenu:enable()
  self.enabled = true
  for _, item in pairs(self.items) do
    item:setTouchEnabled(true)
  end
end

function PunctuationAnswerMenu:createHideOptionsAnimation(seconds)
  self.options = {}

  local fade = AHOpacityTo.new(self.layout, seconds, nil, 0)

  local remove = function()
    clearOptions(self)
  end

  local animation = AHSequence.new(fade, remove)
  return animation
end

function PunctuationAnswerMenu:createOptionsSwapAnimation(newOptions, seconds, character)
  seconds = seconds or 0.8

  local createNew = function()
    initOptions(self, newOptions, character)
  end

  local show = AHOpacityTo.new(self.layout, seconds / 2, nil, 1)
  local removeOld = self:createHideOptionsAnimation(seconds / 2)
  local showNew = AHSequence.new(createNew, show)
  local animation = AHSequence.new(removeOld, showNew)

  return animation
end

function PunctuationAnswerMenu:getSafeAreaOffsetY()
  return self.safeAreaOffsetY
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PunctuationAnswerMenu.new(width, fullHeight, safeAreaOffsetY)
  local self = Group.new(width, fullHeight)
  Mixin.mix(self, PunctuationAnswerMenu)
  createLayout(self, width, fullHeight, safeAreaOffsetY)
  self.items = {}
  self.safeAreaOffsetY = safeAreaOffsetY
  return self
end
