--- PunctuationGame
-- @author Created by alexeychernikov on September 05, 2014.

class("PunctuationGame")
require("StandardUIBuilder")
require("PunctuationConfiguration")
require("PunctuationScene")
require("PunctuationRoundGenerator")
require("bonuses.ResponseSpeedTracker")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {1, 1, 1}
local MINIMUM_TIME_NEEDED_TO_ANSWER = 1.4

local function configureGameDifficulty()
  local rankTwoScore = PunctuationConfiguration.getRankTwoScore()
  local rankThreeScore = PunctuationConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isOutOfTime(self)
  return (self.timer:getTime() >= PunctuationConfiguration.GAME_DURATION)
end

local function endGame(self)
  if (GameSession.getScore() < PunctuationConfiguration.getPassScore()) then
    GameSession.notifyGameLoss()
  else
    self.scene:displayForOutro()
    self.scene:disableInput()

    local outro = self.scene:createOutroAction()
    wait(outro:start())

    GameSession.notifyGameWin()
  end
end

local function startNextRound(self)
  self.canGameEnd = true
  if (not isOutOfTime(self)) then
    if (self.timer:getTimeRemaining() < PunctuationConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
      self.roundGenerator:allowMistakesToResurface()
    end

    local hasTimeForAnotherRound = ((PunctuationConfiguration.GAME_DURATION - self.timer:getTime()) >= MINIMUM_TIME_NEEDED_TO_ANSWER)
    local round = self.roundGenerator:next()
    GameSession.reportSeenConcept(round[PunctuationRoundGenerator.CONCEPT_IDENTIFIER_KEY])
    GameSession.contentTracker:notifyPresentedConcept(round[PunctuationRoundGenerator.CONCEPT_IDENTIFIER_KEY])
    self.currentRound = round
    self.didSeeRationale = false

    local nextQuestionAnimation = self.scene:setRound(round, self.didSeeRationale, hasTimeForAnotherRound)
    if (hasTimeForAnotherRound) then
      self.scene:startNextQuestion()
    end

    wait(nextQuestionAnimation)
    self.responseSpeedTracker:notifyStart()
  end
end

local function getInvalidInputHandler(self)
  return function()
    self.scene:setInvalidInputHandler(nil)
    self.scene:showHint()
  end
end

local function getCorrectAnswerAppliedHandler(self)
  return function()
    if (isOutOfTime(self)) then
      endGame(self)
    else
      startNextRound(self)
    end
  end
end

local function getCorrectAnswerSubmittedHandler(self)
  return function()
    GameSession.increaseScore(PunctuationConfiguration.getPointsForCorrectAnswer(self.currentRound.difficulty))
    self.responseSpeedTracker:notifyCorrectResponse()
    GameSession.contentTracker:notifyCorrectAnswer(self.currentRound[PunctuationRoundGenerator.CONCEPT_IDENTIFIER_KEY])
    GameSession.incrementCorrectAnswerCount()
    self.scoreLabel:setScore(GameSession.getScore())
  end
end

local function getWrongAnswerHandler(self)
  return function(userResponse)
    local conceptId = self.currentRound[PunctuationRoundGenerator.CONCEPT_IDENTIFIER_KEY]

    self.responseSpeedTracker:notifyIncorrectResponse()
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, nil, userResponse)
    self.roundGenerator:notifyMistake(conceptId)
    GameSession.incrementIncorrectAnswerCount()
    self.timer:addTime(PunctuationConfiguration.INCORRECT_TIME_PENALTY)
    self.timer:pauseAndShowIcon()

    local hintWaitable = self.scene:createShowHintAction()
    wait(hintWaitable:start())
    self.didSeeRationale = true
    self.timer:resumeAndHideIcon()

    if (isOutOfTime(self)) then
      endGame(self)
    else
      startNextRound(self)
    end
  end
end

local function getMenuTouchedHandler(self)
  return function()
    self.canGameEnd = false
    self.scene:setInvalidInputHandler(nil)
    self.scene:hideHint()
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationGame:gameFlow()
  self.timer:start()
  startNextRound(self)
  self.scene:setInvalidInputHandler(getInvalidInputHandler(self))
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function initBonusTracking(self, builder)
  self.bonusTray = builder:addBonusTray()
  local bonusWindowSize = PunctuationConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS

  local onQuickResponse = function(earnedBonusProportion, bonusProportionsSum)
      ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                   bonusProportionsSum,
                                   PunctuationConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                   self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowSize, onQuickResponse)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)
  builder:addPauseButton()

  local handleTimeOut = function()
    if (self.canGameEnd) then
      endGame(self)
    end
  end
  self.timer = builder:addDigitTimer(PunctuationConfiguration.GAME_DURATION, handleTimeOut)

  self.scoreLabel = builder:addScoreLabel()

  initBonusTracking(self, builder)

  local conceptIdKey = PunctuationRoundGenerator.CONCEPT_IDENTIFIER_KEY
  self.roundGenerator = LearningFocusedRoundGenerator.new(PunctuationRoundGenerator.new(),
                                                          conceptIdKey)

  local scene = PunctuationScene.new()
  self:insertAtCenter(scene)

  scene:setCorrectAnswerAppliedHandler(getCorrectAnswerAppliedHandler(self))
  scene:setWrongAnswerHandler(getWrongAnswerHandler(self))
  scene:setCorrectAnswerSubmittedHandler(getCorrectAnswerSubmittedHandler(self))
  scene:setMenuTouchedHandler(getMenuTouchedHandler(self))

  self.scene = scene
end

-- Constructor ------------------------

function PunctuationGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, PunctuationGame)

  configureGameDifficulty()
  createUI(self)

  self.didSeeRationale = false
  self.canGameEnd = true

  return self
end
