class("PunctuationHintSnippetView")

require("text.SizeConstrainedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TEXT_COLOR = {1, 1, 1, 0.8}
local TAP_TO_CONTINUE_COLOR = {1, 1, 1, 0.34}
local MORE_INFO_COLOR = {1, 1, 1, 0.34}
local TIP_LABEL_COLOR = {1, 1, 1, 0.34}
local TEXT_WIDTH = Screen.safeWidth - 44 -- polished margin

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getMainTextXYSpacing()
  local xSpacing = Screen.getHeightPart(0.01) -- from height because of font size
  local ySpacing = Screen.getHeightPart(0.007)
  return xSpacing, ySpacing
end

local function createBody(self, hintInfo, width, height)
  local body = Group.new(width, height)

  local tipLabelView = Label.new("TIP", Font.new(FontFaces.DIN_OT_MEDIUM, 16))
  tipLabelView:setColor(unpack(TIP_LABEL_COLOR))

  local baseAttributes = {
     font = Font.new(FontFaces.DIN_OT_LIGHT, 18)
  }

  local attributesForLabels = {
    em = { font = Font.new(FontFaces.DIN_OT_BOLD, 18) },
    i = { font = Font.new(FontFaces.DIN_OT_LIGHT_ITALIC, 18)}
  }

  local marginY = 14

  local maxHeight = height - tipLabelView:getHeight() - marginY
  self.textViewMaxHeight = maxHeight
  local bodyView = SizeConstrainedTextArea.new(hintInfo.body, TEXT_WIDTH, maxHeight, baseAttributes, attributesForLabels)
  bodyView:setColor(unpack(TEXT_COLOR))

  local bodyWidth, bodyHeight = bodyView:getDims()
  local bodyX = width / 2
  local tipLabelX = bodyX - (bodyWidth / 2) + tipLabelView:getWidthPart(0.5)
  local tipLabelY = height - marginY - tipLabelView:getHeightPart(0.5)
  local bodyY = tipLabelY - bodyView:getHeightPart(0.5) - tipLabelView:getHeightPart(0.5)

  bodyView:setLoc(bodyX, bodyY)
  tipLabelView:setLoc(tipLabelX, tipLabelY)

  body:insert(tipLabelView)
  body:insert(bodyView)

  self.tipView = tipLabelView

  return body, bodyView
end

local function createContinue()
  local view = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 15))
  view:setAnchorPointBottom()
  view:setColor(unpack(TAP_TO_CONTINUE_COLOR))
  return view
end

local function getMoreInfoOffscreenPosition(self)
  local _, mainTextBottomY = self:transformLoc(self.body, 0, self.mainText:getYBottom())
  return mainTextBottomY - self.moreInfoIcon:getHeightPart(0.5)
end

local function getMoreInfoOnscreenPosition(self)
  local _, ySpacing = getMainTextXYSpacing()
  return self.height - self.moreInfoIcon:getHeight()
end

local function createMoreInfoView(self, hintInfo)
  local _, ySpacing = getMainTextXYSpacing(20)

  local baseAttributes = {
     font = Font.new(FontFaces.DIN_OT_LIGHT, 18)
  }

  local attributesForLabels = {
    em = { font = Font.new(FontFaces.DIN_OT_BOLD, 18) },
    i = { font = Font.new(FontFaces.DIN_OT_LIGHT_ITALIC, 18)}
  }

  local maxHeight = self.textViewMaxHeight
  local moreInfoView = SizeConstrainedTextArea.new(hintInfo.grammar_note, TEXT_WIDTH, maxHeight, baseAttributes, attributesForLabels)
  moreInfoView:setY(ySpacing)

  local width, height = moreInfoView:getDims()

  local result = Group.new(width, height)
  result:insertAtCenter(moreInfoView)
  result:linkColorAttrToChildren(true)
  result:setOpacity(0)
  result:setAnchorPointTop()
  self:insertAtTop(result, math.huge)

  result:offsetY(self.moreInfoIcon:getHeightPart(-2.0))

  self.moreInfoView = result
end

local function createShowMoreInfoAction(self)
  local moreInfoView = self.moreInfoView
  local targetIconY = getMoreInfoOnscreenPosition(self)
  local moreInfoIconEase = AHEaseElasticOut.new(1.5)

  local hideBody = AHOpacityTo.new(self.mainText, 0.1, nil, 0)
  local showMoreInfo = AHOpacityTo.new(moreInfoView, 0.3, nil, TEXT_COLOR[4])
  local hideTipHeader = AHOpacityTo.new(self.tipView, 0.1, nil, 0)
  local animateMoreInfo = AHDelayed.new(0.1, showMoreInfo)
  local moveMoreInfoIcon = AHLocationYTo.new(self.moreInfoIcon, 1, moreInfoIconEase, targetIconY)
  local animation = AHSpawn.new(hideBody, animateMoreInfo, moveMoreInfoIcon, hideTipHeader)

  return animation
end

local function createShowLessInfoAction(self)
  local moreInfoView = self.moreInfoView
  local targetIconY = getMoreInfoOffscreenPosition(self)

  local moreInfoIconEase = AHEaseElasticOut.new(1.5)

  local showBody = AHOpacityTo.new(self.mainText, 0.3, nil, TEXT_COLOR[4])
  local hideMoreInfo = AHOpacityTo.new(moreInfoView, 0.1, nil, 0)
  local animateBody = AHDelayed.new(0.1, showBody)
  local moveMoreInfoIcon = AHLocationYTo.new(self.moreInfoIcon, 1, moreInfoIconEase, targetIconY)
  local showTipHeader = AHOpacityTo.new(self.tipView, 0.1, nil, TIP_LABEL_COLOR[4])
  local animation = AHSpawn.new(hideMoreInfo, animateBody, moveMoreInfoIcon, showTipHeader)

  return animation
end

local function getMoreInfoTouchHandler(self)
  local showingMoreInfo = false
  local animation = nil

  return function(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      if (not self.additionalInfoInitialized) then
        return
      end

      if (animation) then
        animation:stop()
      end

      if (showingMoreInfo) then
        animation = createShowLessInfoAction(self)
        animation:start()
        showingMoreInfo = false
      else
        animation = createShowMoreInfoAction(self)
        animation:start()
        showingMoreInfo = true
      end
    end
  end
end

local function hasGrammarNote(hintInfo)
  return (hintInfo.grammar_note and #hintInfo.grammar_note > 0)
end

local function createMoreInfoIcon(self, hintInfo)
  if (not hasGrammarNote(hintInfo)) then
    return Group.newMinimumSized()
  end

  local moreInfo = Label.new("more information", Font.new(FontFaces.DIN_OT_MEDIUM, 12))
  local width, height = moreInfo:getDims()

  local view = Group.new(width, height)
  view.textLabel = moreInfo
  local underline = Line.new(0, 0, width, 0)

  moreInfo:setColor(unpack(MORE_INFO_COLOR))
  underline:setColor(unpack(MORE_INFO_COLOR))

  view:insertAtCenter(moreInfo)

  view:insert(underline)
  underline:setAnchorPointBottomLeft()

  view:setAnchorPointTopLeft()

  TouchManager.addTouchHandler(moreInfo, getMoreInfoTouchHandler(self))

  local touchWidth, touchHeight = Screen.getWidthPart(0.6), Screen.getHeightPart(0.12)
  moreInfo:setBounds(-touchWidth / 2, -touchHeight / 2, 0, touchWidth / 2, touchHeight / 2, 0)
  self.hitArea = moreInfo

  return view
end

local function createLayout(self, hintInfo, width, height)
  local background = FilledRectangle.new(width, height, PunctuationColors.getMenuBackgroundColor())
  self:insertAtCenter(background)

  local bodyWidth = width
  local bodyHeight = (height - Screen.safeBottom) * 0.8

  coroutine.yield()
  local body, mainText = createBody(self, hintInfo, bodyWidth, bodyHeight)
  coroutine.yield()
  local continue = createContinue()
  coroutine.yield()
  local moreInfoIcon = createMoreInfoIcon(self, hintInfo)
  coroutine.yield()

  self.body = body
  self.mainText = mainText
  self.continue = continue
  self.moreInfoIcon = moreInfoIcon

  body:setAnchorPointTop()
  self:insertAtTop(body)

  local centerX = self:getWidthPart(0.5)
  local centerY = self:getWidthPart(0.5) + Screen.safeBottom

  local moreInfoIconX = self:getWidthPart(0.5) - mainText:getWidthPart(0.5)
  local moreInfoIconY = getMoreInfoOffscreenPosition(self)
  moreInfoIcon:setLoc(moreInfoIconX, moreInfoIconY)
  self:insert(moreInfoIcon)

  local continueY = math.max(Screen.safeBottom, continue:getHeightPart(1.5))
  continue:setLoc(centerX, continueY)
  self:insert(continue)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationHintSnippetView:createAdditionalInfo()
  self.additionalInfoInitialized = true
  if (hasGrammarNote(self.hintInfo)) then
    createMoreInfoView(self, self.hintInfo)
  end
end

function PunctuationHintSnippetView:setInfoButtonHandler(handler)
  self.infoButtonHandler = handler
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PunctuationHintSnippetView.new(hintInfo, width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, PunctuationHintSnippetView)

  self.hintInfo = hintInfo
  self.width = width
  self.height = height
  runInCoroutine(function()
    createLayout(self, hintInfo, width, height)
  end)

  return self
end
