require("PunctuationProgressMeterStalk")

class("PunctuationProgressMeter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- the first five descriptors should be shuffled between themselves because
-- they are selected to look nicely spaced:
-- A B C D E | F G H I J K L M N O ... suffled in segments at index 5 ->
-- E C B A D | N O G H J F L I K M ...
local STALK_SHUFFLE_STRATEGY_INDEX = 5
local STALK_DESCRIPTORS = {
  {
    multiplierX = 0.57,
    heightMultiplier = 0.4
  },
  {
    multiplierX = 0.89,
    heightMultiplier = 0.25
  },
  {
    multiplierX = 0.33,
    heightMultiplier = 0.13
  },
  {
    multiplierX = 0.0,
    heightMultiplier = 0.15
  },
  {
    multiplierX = 0.25,
    heightMultiplier = 0.35
  },
  {
    multiplierX = 0.44,
    heightMultiplier = 0.32
  },
  {
    multiplierX = 0.52,
    heightMultiplier = 0.21
  },
  {
    multiplierX = 0.61,
    heightMultiplier = 0.12
  },
  {
    multiplierX = 0.70,
    heightMultiplier = 0.33
  },
  {
    multiplierX = 0.81,
    heightMultiplier = 0.12
  },
  {
    multiplierX = 1.0,
    heightMultiplier = 0.33
  },
  {
    multiplierX = 0.04,
    heightMultiplier = 0.55
  },
  {
    multiplierX = 0.17,
    heightMultiplier = 0.07
  },
  {
    multiplierX = 0.11,
    heightMultiplier = 0.22
  },
  {
    multiplierX = 0.29,
    heightMultiplier = 0.7
  },
  {
    multiplierX = 0.38,
    heightMultiplier = 0.42
  },
  {
    multiplierX = 0.48,
    heightMultiplier = 0.82
  },
  {
    multiplierX = 0.66,
    heightMultiplier = 0.62
  },
  {
    multiplierX = 0.75,
    heightMultiplier = 0.1
  },
  {
    multiplierX = 0.85,
    heightMultiplier = 0.72
  },
  {
    multiplierX = 0.93,
    heightMultiplier = 0.57
  },
  {
    multiplierX = 0.96,
    heightMultiplier = 0.93
  }
}
local ORB_RADIUS = 7

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createShuffledArrayInSegments(array)
  local arrayLength = #array

  local firstSegment = TableUtils.subarray(array, 1, STALK_SHUFFLE_STRATEGY_INDEX)
  local secondSegment = TableUtils.subarray(array, STALK_SHUFFLE_STRATEGY_INDEX + 1, arrayLength)

  TableUtils.shuffleArray(firstSegment)
  TableUtils.shuffleArray(secondSegment)

  return TableUtils.createConcatenatedArray(firstSegment, secondSegment)
end

local function incrementAndGetStep(self)
  local index = self.currentIndex
  self.currentIndex = math.min(self.currentIndex + 1, #STALK_DESCRIPTORS)

  return index
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationProgressMeter:increment()
  local step = incrementAndGetStep(self)

  self.stalks[self.stalkFillIndices[step]]:setCorrectState()
end

function PunctuationProgressMeter:getClosestStalkIndex(x0)
  local minDistance = math.huge
  local closestIndex = nil
  for index, stalk in ipairs(self.stalks) do
    local dx = math.abs(x0 - stalk:getX())
    if (dx < minDistance) then
      closestIndex = index
      minDistance = dx
    end
  end
  return closestIndex
end

function PunctuationProgressMeter:createBumpAnimation(seconds, centerIndex, minYOffset, maxYOffset, staggerCoef, upEase, downEase)
  seconds = seconds or 2
  centerIndex = centerIndex or 1
  staggerCoef = staggerCoef or 0.55
  maxYOffset = maxYOffset or 20
  minYOffset = minYOffset or 10

  local totalStalks = #self.stalks
  local yOffsetAmp = maxYOffset - minYOffset
  local maxIndexDist = math.max(totalStalks - centerIndex, centerIndex - 1)

  local totalDelayTime = seconds * staggerCoef
  local maxDelay = totalDelayTime
  local secondsPerStalk = seconds - totalDelayTime
  local animation = AHSpawn.new()

  for index, stalk in ipairs(self.stalks) do
    local distance = math.abs(index - centerIndex)
    local normalizedDistance = distance/maxIndexDist
    normalizedDistance = math.pow(normalizedDistance, 0.75)
    if (staggerCoef < 0) then
      normalizedDistance = 1 - normalizedDistance
    end

    local delay = maxDelay * normalizedDistance
    local yOffset = minYOffset + yOffsetAmp * (1 - normalizedDistance)

    local moveStalk = stalk:createUpDownAbsoluteAnimation(secondsPerStalk, yOffset, upEase, downEase)
    local currentAnimation = AHDelayed.new(delay, moveStalk)
    animation:addAction(currentAnimation)
  end

  return animation
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local width, height = self:getDims()

  local stalks = {}
  local stalkFillIndices = {}
  local stalkGroup = Group.new(width, height)
  local minHeight = height * 0.3
  local maxHeight = height
  local orbDiameter = ORB_RADIUS * 2
  for descriptorIndex, stalkDescriptor in ipairs(STALK_DESCRIPTORS) do
    local stalkHeight = (maxHeight - minHeight) * stalkDescriptor.heightMultiplier + minHeight
    local stalk = PunctuationProgressMeterStalk.newHidden(orbDiameter, stalkHeight)
    stalk:setAnchorPointBottom()

    local stalkWidth = stalk:getWidth()
    local x = (width - stalkWidth) * stalkDescriptor.multiplierX + (stalkWidth * 0.5)
    stalk:setX(x)
    stalkGroup:insert(stalk)
    table.insert(stalks, stalk)

    stalkFillIndices[descriptorIndex] = descriptorIndex
  end

  local particles = PunctuationAnswerReaction.getPreloadedInstance("stalk", PunctuationAnswerReactionPositive, orbDiameter, orbDiameter)
  self:insert(particles)
  local offscreenX = -Screen.width
  local offscreenY = -Screen.height
  particles:setLoc(offscreenX, offscreenY)

  self:insertAtCenter(stalkGroup)

  stalkFillIndices = createShuffledArrayInSegments(stalkFillIndices)

  self.stalks = stalks
  self.stalkFillIndices = stalkFillIndices
end

function PunctuationProgressMeter.new(width)
  local height = Screen.getHeightPart(0.12)

  local self = Group.new(width, height)
  Mixin.mix(self, PunctuationProgressMeter)
  createLayout(self)

  self.currentIndex = 1

  return self
end
