class("PunctuationScene")
require("graphics.LinePathTracer")

require("PunctuationAnswerMenu")
require("PunctuationHintSnippetView")
require("PunctuationProgressMeter")
require("PunctuationSceneAnimations")
require("PunctuationBackground")
require("PunctuationOutroEffect")
require("PunctuationInputHelper")
require("PunctuationFairyDust")
require("widgets.TapToContinue")
require("graphics.MultiGradient")
require("PunctuationPromptTextView")
require("PunctuationAnswerReactionPositive")
require("PunctuationAnswerReactionNegative")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BEAM_FLY_DURATION_SECONDS = 0.6

local OUTRO_COLORS = {
  {31/255, 4/255, 28/255},
  {109/255, 34/255, 98/255},
  {255/255, 0/255, 113/255},
}

local OUTRO_BACKGROUND_BAND_SIZES = {
  0.6,
  0.4,
}

local TRACED_PATH_COLOR = {44/255, 255/255, 255/255}
local POSITIVE_ANIMATION_DURATION = 0.65
local NEGATIVE_ANIMATION_DURATION = 0.9
local PROMPT_WIDTH = Screen.getSafeWidthPart(0.89)


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function preloadEffects()
  local sizer = Sprite.newWithName("punctuation_no_mark.png")
  local width, height = sizer:getDims()
  local radius = math.max(width, height) * 0.9

  local positiveParticles = PunctuationAnswerReaction.getPreloadedInstance("sentence_correct", PunctuationAnswerReactionPositive, radius, radius)
  local negativeParticles = PunctuationAnswerReaction.getPreloadedInstance("sentence_wrong", PunctuationAnswerReactionNegative, radius, radius)
  PunctuationAnswerReaction.getAnimationForName("sentence_correct", POSITIVE_ANIMATION_DURATION)
  PunctuationAnswerReaction.getAnimationForName("sentence_wrong", NEGATIVE_ANIMATION_DURATION)

  local offscreenX = -Screen.width
  local offscreenY = -Screen.height
  positiveParticles:setLoc(offscreenX, offscreenY)
  negativeParticles:setLoc(offscreenX, offscreenY)

  Screen.rootGroup:insert(positiveParticles, 9000)
  Screen.rootGroup:insert(negativeParticles, 9001)

  sizer:removeSelf()
end

local function getFairyDustSprinkler(self)
  local duration = 1.2

  local totalSprinklers = #self.fairySprinklers
  local prevSprinklerIndex = 1
  return function(trajectory, index)
    if (index > #trajectory) then
      return
    end

    local normalizedIndex = index/#trajectory
    local sprinklerIndex = math.floor(normalizedIndex * totalSprinklers)
    if (prevSprinklerIndex > sprinklerIndex) then
      prevSprinklerIndex = 1
    end

    local position = trajectory[index]
    for i = prevSprinklerIndex, sprinklerIndex do
      local sprinkler = self.fairySprinklers[i]
      local sprinkle = sprinkler:createSprinkleAction(duration)

      local x = position[1]
      local y = position[2]

      sprinkler:setLoc(x, y)
      sprinkle:start()
    end
    prevSprinklerIndex = math.max(sprinklerIndex, 1)
  end
end

local function createBeamToCurrentPunctuationAnim(self, x0, y0)

  local x1, y1 = self.sentenceView:getBlankWorldLocation()
  self.lineTracer:bringToFront()
  self.fairySprinklerGroup:setColor(unpack(PunctuationColors.getCorrectColor()))
  return self.lineTracer:createTracePathAnimation(x0, y0, x1, y1, BEAM_FLY_DURATION_SECONDS), BEAM_FLY_DURATION_SECONDS
end

local function doCorrectAnswer(self)
  if (self.correctAnswerSubmittedHandler) then
    self.correctAnswerSubmittedHandler()
  end

  self.background:transitionToCorrectColors()

  Sound.playSoundEffect("punctuation_correct")
  self.roundProgress:increment()

  local sentenceAnimation = self.sentenceView:createSetCorrectAction(self.roundInfo.rules[1].correct)
  sentenceAnimation:start()

  AHDelayed.new(POSITIVE_ANIMATION_DURATION - 0.1, self.correctAnswerAppliedHandler):start()
end

local function transitionIntoWrongColors(self)
  local wrongBeamColor = PunctuationColors.getWrongColor()
  self.lineTracer:setColor(unpack(wrongBeamColor))
  self.fairySprinklerGroup:setColor(unpack(wrongBeamColor))

  self.background:transitionToIncorrectAndRationaleColors()
end

local function doWrongAnswer(self, userResponse)
  Sound.playSoundEffect("punctuation_wrong")

  local wrongAnimation = self.sentenceView:createSetIncorrectAction(self.roundInfo.rules[1].correct)
  wrongAnimation:start()

  transitionIntoWrongColors(self)

  if (self.wrongAnswerHandler) then
    self.wrongAnswerHandler(userResponse)
  end
end

local function clearInvalidInputHandler(self)
  if (self.invalidInputCover) then
    self.invalidInputCover:removeSelf()
    self.invalidInputCover = nil
  end
end

local function getAnswerHandler(self)
  return function(answer, source)
    clearInvalidInputHandler(self)
    if (self.menuTouchedHandler) then
      self.menuTouchedHandler()
    end

    self:disableInput()
    self.lineTracer:setColor(unpack(TRACED_PATH_COLOR))

    local x0, y0 = source:getLoc()
    y0 = y0 + Screen.safeBottom
    local beamAnimation, beamSeconds = createBeamToCurrentPunctuationAnim(self, x0, y0)
    local dispatchAnswer = function()
      if (answer == self.correctAnswer) then
        doCorrectAnswer(self)
      else
        doWrongAnswer(self, answer)
      end
    end

    local progressBarAnimation = PunctuationSceneAnimations.createProgressBarReactionToInput(source, self.roundProgress, self.answerMenu, 1.2)
    local secondsBeforeDispatchingAnswer = beamSeconds * 0.55
    local delayBeforeDispatchingAnswer = AHDelay.new(secondsBeforeDispatchingAnswer)
    local dispatchAnswerAnimation = AHSequence.new(delayBeforeDispatchingAnswer, dispatchAnswer)
    local answerAnimation = AHSpawn.new(beamAnimation, dispatchAnswerAnimation, progressBarAnimation)
    answerAnimation:start()
  end
end

local function showAnswerMenu(self)
  local width = Screen.width
  local height = Screen.getHeightPart(0.14)
  local safeAreaHeightBuffer = Screen.safeBottom
  local fullHeight = height + safeAreaHeightBuffer

  local answerMenu = PunctuationAnswerMenu.new(width, fullHeight, safeAreaHeightBuffer)
  self:insert(answerMenu)
  answerMenu:setAnswerSelectHandler(getAnswerHandler(self))

  return answerMenu
end

local function createFairyDustParticles()
  local spawners = {}
  local spawnerCount = 8
  local particlesPerPoint = 2
  local w, h = Screen.getWidthPart(0.05), Screen.getHeightPart(0.05)
  local group = Group.newFullScreen()
  for i = 1, spawnerCount do
    local spawner = PunctuationFairyDust.new(particlesPerPoint, w, h)
    spawners[i] = spawner
    group:insert(spawner)
    spawner:inheritColor(group)
  end
  return spawners, group
end

local function initInputHintIfNecessary(self)
  if (self.inputHintInitialized) then
    return
  end
  self.inputHintInitialized = true
  local secondsBeforeShowingHint = PunctuationConfiguration.SECONDS_BEFORE_HELPER
  local doInvalidInput = function()
    if (self.invalidInputHandler) then
      self.invalidInputHandler()
    end
  end
  AHDelayed.new(secondsBeforeShowingHint, doInvalidInput):start()
end

local function initInvalidInputRegion(self, menu)
  local height = menu:getHeight()
  local inputRegion = Group.new(Screen.width, Screen.height - height)
  inputRegion:setLoc(Screen.width / 2, Screen.height / 2 + height)
  self:insert(inputRegion)
  self.invalidInputCover = inputRegion
  TouchManager.addTouchHandler(inputRegion, function()
    if (self.invalidInputHandler) then
      self.invalidInputHandler()
    end
  end)
end

local function createLayout(self)
  local answerMenu = showAnswerMenu(self)
  local menuHeight = answerMenu:getHeight()
  self.answerMenu = answerMenu

  local midGround = Group.newFullScreen()
  self:insertAtCenter(midGround)
  self.midGround = midGround

  local background = PunctuationBackground.new()
  self:insertAtCenter(background)
  self.background = background

  local roundProgress = PunctuationProgressMeter.new(PROMPT_WIDTH)
  self:insert(roundProgress)
  roundProgress:setAnchorPointBottom()
  self.roundProgress = roundProgress

  self.fairySprinklers, self.fairySprinklerGroup = createFairyDustParticles()
  self:insertAtCenter(self.fairySprinklerGroup)

  local lineTracer = LinePathTracer.new({maxVisibleViews = 25, maxTailLength = 15, lineWidth = 3})
  lineTracer:setTipAdvanceHandler(getFairyDustSprinkler(self))
  self:insert(lineTracer)
  self.lineTracer = lineTracer

  local progressHeight = roundProgress:getHeight()
  local menuY = -menuHeight/2 - progressHeight/2 - Screen.getHeightPart(0.08)
  local progressY = -progressHeight - Screen.getHeightPart(0.08)
  roundProgress:setLoc(Screen.width/2, progressY)
  answerMenu:setLoc(Screen.width/2, menuY)
  answerMenu:bringToFront()

  initInvalidInputRegion(self, answerMenu)
end

local function initIfNecessary(self)
  if (not self.initialized) then
    self.initialized = true
    createLayout(self)
    preloadEffects(self)
  end
end

local function createRoundView(self, roundInfo)
  self:disableInput()

  local sentenceView = PunctuationPromptTextView.new(PROMPT_WIDTH)
  sentenceView:setText(roundInfo.sentence)
  sentenceView:setLoc(self:getWidthPart(0.5), Screen.getHeightPart(0.65))
  self:insert(sentenceView)

  self.previousSentenceView = self.sentenceView
  self.sentenceView = sentenceView
end

local function swapOptions(self, newOptions, character)
  local timeToSwapOptions = 0.7
  local swapAnimation, button = self.answerMenu:createOptionsSwapAnimation(newOptions, timeToSwapOptions, character)
  swapAnimation:start()
end

local function preloadHintSnippet(self)
  if (self.currentHintSnippet) then
    if (self.currentHintSnippet.parent) then
      self.currentHintSnippet:removeSelf()
    else
      self.currentHintSnippet:removeSelf()
    end
  end

  local hintInfo = self.roundInfo.rules[1].hint
  local width = Screen.width
  local height = Screen.getSafeHeightPart(0.4) + Screen.safeBottom
  self.currentHintSnippet = PunctuationHintSnippetView.new(hintInfo, width, height)
  self.currentHintSnippet:setLoc(-Screen.width, -Screen.height)
end

local function createShowHintAnimation(self)
  local hintView = self.currentHintSnippet
  self:insert(hintView)
  self.hintView = hintView
  hintView:bringToFront()
  local width, height = hintView:getDims()
  hintView:setLoc(width / 2, height / 2)

  local hintAnimation = PunctuationSceneAnimations.createShowHintAnimation(hintView, self.answerMenu, self.roundProgress, 1)
  return hintAnimation
end

local function setOutroColors(self)
  local isHorizontal = false
  local overlay = MultiGradient.new(Screen.width, Screen.height, isHorizontal, OUTRO_COLORS, OUTRO_BACKGROUND_BAND_SIZES)
  self.midGround:insert(overlay)
  overlay:setAnchorPointBottomLeft()

  local hide = overlay:createOpacityTo(0, nil, 0)
  local show = overlay:createOpacityTo(0.3, nil, 1)
  AHSequence.new(hide, show):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationScene:setRound(roundInfo, didSeeRationale, hasTimeForAnotherRound)
  initIfNecessary(self)
  initInputHintIfNecessary(self)

  local animationSeconds = 0.7
  local cleanSeconds = 0.7

  self.background:restoreColor()
  local cleanRoundAnimation = PunctuationSceneAnimations.createRoundCleanAnimation(self.sentenceView, self.roundProgress, cleanSeconds, didSeeRationale)
  createRoundView(self, roundInfo)

  self.roundInfo = roundInfo

  self:disableInput()
  local roundPresentAnimation, roundInitAnimation = PunctuationSceneAnimations.createRoundPresentAnimation(self.sentenceView, self.answerMenu, self.roundProgress, animationSeconds)

  if (not hasTimeForAnotherRound) then
    self.sentenceView:removeSelf()
  end

  local swapOptionsAction = function()
    local answerOptions = self.roundInfo.rules[1].options
    local character = self.roundInfo.characterBeforeBlank
    swapOptions(self, answerOptions, character)
  end

  local presentAnswerOptions = AHDelayed.new(animationSeconds * 0.7, swapOptionsAction)

  local roundStartAnimation = nil
  if (hasTimeForAnotherRound) then
    roundStartAnimation = AHSequence.new(roundInitAnimation, cleanRoundAnimation, roundPresentAnimation)
  else
    roundStartAnimation = AHSequence.new(roundInitAnimation, cleanRoundAnimation)
  end

  local roundTransitionAnimation = AHSpawn.new(roundStartAnimation, presentAnswerOptions)
  roundTransitionAnimation:start()

  return roundTransitionAnimation
end

function PunctuationScene:startNextQuestion()
  self.correctAnswer = self.roundInfo.rules[1].correct

  local enableInput = function()
    self:enableInput()
  end

  local secondsBeforePreloadingHint = 1.3
  local waitToPreloadHint = AHDelay.new(secondsBeforePreloadingHint)
  local doPreloadHint = function()
    preloadHintSnippet(self)
  end

  local preloadHintAction = AHSequence.new(waitToPreloadHint, doPreloadHint, enableInput)
  preloadHintAction:start()

  return preloadHintAction
end

function PunctuationScene:setInvalidInputHandler(handler)
  self.invalidInputHandler = handler
end

function PunctuationScene:setCorrectAnswerAppliedHandler(handler)
  self.correctAnswerAppliedHandler = handler
end

function PunctuationScene:setCorrectAnswerSubmittedHandler(handler)
  self.correctAnswerSubmittedHandler = handler
end

function PunctuationScene:setMenuTouchedHandler(handler)
  self.menuTouchedHandler = handler
end

function PunctuationScene:setWrongAnswerHandler(handler)
  self.wrongAnswerHandler = handler
end

function PunctuationScene:disableInput()
  self.answerMenu:disable()
end

function PunctuationScene:enableInput()
  self.answerMenu:enable()
end

function PunctuationScene:createShowHintAction()
  local showHintAnimation = createShowHintAnimation(self)
  local hideHintAnimation = PunctuationSceneAnimations.createHideHintAnimation(self.hintView, self.answerMenu, self.roundProgress, 0.4)
  local restoreBackground = function()
    self.background:restoreColor()
  end
  local restoreGlyphColor = function()
    self.sentenceView:restoreVisualState()
  end
  local restoreGameplay = AHSpawn.new(hideHintAnimation, restoreBackground, restoreGlyphColor)
  local waitForHint = AHDelay.new(math.huge)

  local createTapToContinue = function()
    local tapToContinue = nil
    tapToContinue = TapToContinue.new(nil, nil, function()
      waitForHint:stop()
      tapToContinue:removeSelf()
      restoreGameplay:start()
    end)
    tapToContinue:bringToBack()
  end

  local animation = AHSequence.new(showHintAnimation, createTapToContinue, waitForHint)

  return animation
end

function PunctuationScene:createOutroAction()
  local celebrationView = PunctuationOutroEffect.new()
  self:insert(celebrationView)

  local moveOutSeconds = 0.4
  local celebrateSeconds = 3
  local secondsBeforeCelebration = moveOutSeconds - 0.1

  local moveUIOut = PunctuationSceneAnimations.createUIOutroAnimation(self.sentenceView, self.previousSentenceView, self.answerMenu, self.roundProgress, moveOutSeconds)
  local celebrate = celebrationView:createOutroAnimation(celebrateSeconds)
  local waitToStartCeleberation = AHDelay.new(secondsBeforeCelebration)
  local celebrationAnimation = AHSequence.new(waitToStartCeleberation, celebrate)
  local outro = AHSpawn.new(moveUIOut, celebrationAnimation)
  return outro
end

function PunctuationScene:showHint()
  if (self.showingHint) then
    return
  end

  self.showingHint = true
  local helper = PunctuationInputHelper.new()
  helper:show()
  helper:setLoc(0, self.answerMenu:getHeight() + 10)
  self:insert(helper)
  self.helper = helper
end

function PunctuationScene:hideHint()
  if (not self.showingHint) then
    return
  end

  self.showingHint = false
  self.helper:hide()
  self.helper = nil
end

function PunctuationScene:displayForOutro()
  setOutroColors(self)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function PunctuationScene.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, PunctuationScene)

  self.initialized = false
  self.sentenceView = nil
  self.previousSentenceView = nil

  FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

  initIfNecessary(self)

  return self
end
