class("PunctuationSceneAnimations")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getMenuPosition(menu)
  return Screen.width / 2, menu:getHeightPart(0.5)
end

local function getProgressBarPosition(menu)
  local x = Screen.getWidthPart(0.5)
  local safeAreaOffsetY = menu:getSafeAreaOffsetY()
  local menuHeight = menu:getHeight() - safeAreaOffsetY
  local y = safeAreaOffsetY + (menuHeight * 0.8)
  return x, y
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationSceneAnimations.createRoundPresentAnimation(sentence, menu, progressBar, seconds)
  local sentenceX, sentenceY = sentence:getLoc()
  local _, menuY = getMenuPosition(menu)
  local _, progressY = getProgressBarPosition(menu)
  local progressW = progressBar:getWidth()
  local menuDy = math.abs(menu:getY() - menuY)
  local normalizedMenuDy = menuDy / menu:getHeight()

  local sentenceStartY = -sentence:getHeight()
  local sentenceEase = AHEaseElasticOut.new(1.5)
  local menuEase = AHEaseElasticOut.new(0.8)
  local progressEase = sentenceEase
  local progressBumpTime = seconds * 0.7
  local progressCenterIndex = progressBar:getClosestStalkIndex(progressW/2)
  local minYOffset = Screen.height * (0.005 + normalizedMenuDy * 0.0015)
  local maxYOffset = Screen.height * (0.015 + normalizedMenuDy * 0.0025)
  local staggerCoef = 0.2

  local delayBeforeProgressStart = seconds * normalizedMenuDy  * 0.1
  local delayBeforeMenuStart = seconds * 0.3

  local init = function()
    sentence:setLoc(sentenceX, sentenceStartY)
  end

  local stalksUpEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.1, 0.1}, {0.85, 0.97}, {1, 1}}, params = {tension = 0.9}})
  local stalksDownEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.15, 0.03}, {0.75, 0.95}, {1, 1}}, params = {tension = 0}})

  local moveSentence = AHLocationYTo.new(sentence, seconds, sentenceEase, sentenceY)
  local moveMenu = AHSequence.new(AHDelay.new(delayBeforeMenuStart), AHLocationYTo.new(menu, seconds, menuEase, menuY))

  local moveProgress = AHLocationYTo.new(progressBar, seconds, progressEase, progressY)
  local bumpProgress = progressBar:createBumpAnimation(progressBumpTime, progressCenterIndex, minYOffset, maxYOffset, staggerCoef, stalksUpEase, stalksDownEase)
  local moveAndBumpProgress = AHSequence.new(AHDelay.new(delayBeforeProgressStart), AHSpawn.new(moveProgress, bumpProgress))

  local playSound = function() Sound.playSoundEffect("punctuation_switch_terms") end

  local animation = AHSpawn.new(moveSentence, moveMenu, moveAndBumpProgress, playSound)

  return animation, init
end

function PunctuationSceneAnimations.createRoundCleanAnimation(sentence, progressBar, seconds, didSeeRationale)
  if (sentence == nil) then -- first round
    return AHDelay.new(0)
  end

  local progressW = progressBar:getWidth()

  local sentenceEase = AHEaseElasticIn.new(1.5)
  local progressBumpTime = 0.5
  local progressCenterIndex = progressBar:getClosestStalkIndex(progressW/2)
  local minYOffset = Screen.height * 0.03
  local maxYOffset = Screen.height * 0.06
  local staggerCoef = 0.1

  local moveSentence = AHLocationYTo.new(sentence, seconds, sentenceEase, Screen.height)
  local clearSentence = function() sentence:removeSelf() end
  local bumpProgress = progressBar:createBumpAnimation(progressBumpTime, progressCenterIndex, minYOffset, maxYOffset, staggerCoef)

  local sentenceAnimation = AHSequence.new(moveSentence, clearSentence)
  local animation = AHSpawn.new(sentenceAnimation)

  if (didSeeRationale == true) then
    animation:addAction(bumpProgress)
  end

  return animation
end

function PunctuationSceneAnimations.createShowHintAnimation(hint, menu, progressBar, seconds)
  local hintX, hintY = hint:getLoc()
  local menuH = menu:getHeight()

  local hintStartY = -hint:getHeightPart(0.5)
  local dy = hintY - hintStartY
  local menuDy = dy - menuH
  local menuStartTime = 0 * seconds * menuH / dy / 2
  local menuSeconds = seconds - menuStartTime

  local hintEase = AHEaseElasticOut.new(1.5)
  local stalksEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.35, 1.02}, {1, 1}}, params = {tension = 0}})
  local stalksSeconds = menuSeconds * 0.8

  local initHint = function()
    hint:setLoc(hintX, hintStartY)
  end

  local moveHintUp = AHLocationYTo.new(hint, seconds, hintEase, hintY)

  local hideMenu = menu:createHideOptionsAnimation(seconds)
  local moveMenu = AHLocationYBy.new(menu, menuSeconds, nil, menuDy)
  local moveProgress = AHLocationYBy.new(progressBar, stalksSeconds, stalksEase, menuDy)
  local waitToMoveUI = AHDelay.new(menuStartTime)
  local moveUI = AHSequence.new(waitToMoveUI, AHSpawn.new(moveMenu, moveProgress))
  local menuAnimation = AHSpawn.new(hideMenu, moveUI)
  local createAdditionalHintInfo = function()
    hint:createAdditionalInfo()
  end

  local animation = AHSequence.new(initHint, AHSpawn.new(menuAnimation, moveHintUp), createAdditionalHintInfo)
  return animation
end

function PunctuationSceneAnimations.createHideHintAnimation(hint, menu, progressBar, seconds)
  local _, menuY = getMenuPosition(menu)
  local _, progressY = getProgressBarPosition(menu)
  local hintY = -hint:getHeightPart(0.5)

  local menuEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.3, -0.1}, {1, 1}}, params = {tension = 0.9}})
  local hintEase = menuEase
  local progressEase = hintEase

  local moveMenu = AHLocationYTo.new(menu, seconds, menuEase, menuY)
  local moveProgress = AHLocationYTo.new(progressBar, seconds, progressEase, progressY)
  local moveHint = AHLocationYTo.new(hint, seconds, hintEase, hintY)
  local removeHint = function() hint:removeSelf() end
  local hintAnimation = AHSequence.new(moveHint, removeHint)

  local animation = AHSpawn.new(moveMenu, moveProgress, hintAnimation)
  return animation
end

function PunctuationSceneAnimations.createProgressBarReactionToInput(answerButton, progressBar, menu, seconds)
  local _, menuY = getMenuPosition(menu)

  local progressBumpTime = seconds * 0.6
  local menuBumpTime = seconds * 0.2
  local minYOffset = Screen.height * 0.0125
  local maxYOffset = Screen.height * 0.025
  local staggerCoef = 0.2
  local progressCenterIndex = progressBar:getClosestStalkIndex(answerButton:getX() - answerButton:getWidthPart(0.5))
  local menuDelayTime = seconds * 0.0
  local menuUpEase =  AHEaseSineOut.new()
  local menuDownEase = AHEaseSineInOut.new()
  local menuBumpHeight = (menu:getHeight() - menu:getSafeAreaOffsetY()) * 0.25

  local stalksUpEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.1, 0.1}, {0.85, 0.97}, {1, 1}}, params = {tension = 0.9}})
  local stalksDownEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.15, 0.03}, {0.6, 1.1}, {1, 1}}, params = {tension = 0}})

  local bump = progressBar:createBumpAnimation(progressBumpTime, progressCenterIndex, minYOffset, maxYOffset, staggerCoef, stalksUpEase, stalksDownEase)
  local menuUp = AHLocationYTo.new(menu, menuBumpTime/2, menuUpEase, menuY + menuBumpHeight)
  local menuDown = AHLocationYTo.new(menu, menuBumpTime, menuDownEase, menuY)
  local waitToBumpMenu = AHDelay.new(menuDelayTime)
  local menuBump = AHSequence.new(waitToBumpMenu, menuUp, menuDown)

  local animation = AHSpawn.new(bump, menuBump)

  return animation
end

function PunctuationSceneAnimations.createUIOutroAnimation(sentence, previousSentence, menu, progressBar, seconds)
  local sentenceY = Screen.height + sentence:getHeightPart(0.5)
  local menuY = -menu:getHeightPart(0.2)
  local progressY = -progressBar:getHeight() - Screen.height * 0.05

  local moveOutSeconds = seconds
  local menuEase = AHEaseCustom.newWithCurve({points = {{0, 0}, {0.3, 1.1}, {1, 1}}, params = {tension = 0}})
  local progressEase = nil
  local sentenceEase = AHEaseElasticIn.new(1.5)
  local progressOutSeconds = seconds * 0.7
  local sentenceOutSeconds = seconds
  local delayBeforeMenuOut = seconds * 0.3


  local moveMenu = AHLocationYTo.new(menu, moveOutSeconds, menuEase, menuY)
  local hideMenuOptions = menu:createHideOptionsAnimation(moveOutSeconds/2)
  local moveProgress = AHLocationYTo.new(progressBar, progressOutSeconds, progressEase, progressY)
  local moveMenuAndProgress = AHSpawn.new(moveMenu, moveProgress, hideMenuOptions)

  local animateMenuAndProgress = AHSequence.new(AHDelay.new(delayBeforeMenuOut), moveMenuAndProgress)
  local moveSentence = AHLocationYTo.new(sentence, sentenceOutSeconds, sentenceEase, sentenceY)

  local animation = AHSpawn.new(animateMenuAndProgress, moveSentence)

  if (previousSentence) then
    local movePreviousSentence = AHLocationYTo.new(previousSentence, sentenceOutSeconds, sentenceEase, sentenceY)
    animation:addAction(movePreviousSentence)
  end

  return animation
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------
