--- PunctuationValidator
-- @author Created by alexeychernikov on September 05, 2014.
-- ConceptValidator for Punctuation.

class("PunctuationValidator")

require("PunctuationRoundGenerator")
require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'correct_1',
  'sentence'
}

local REQUIRED_ANSWER_KEYS = {
  'body',
  'correct_answer',
  'incorrect_answers',
  'rule_name'
}

local ALLOWED_PUNCTUATION_TYPES = TableUtils.createSetFromArray({
  "comma", "semicolon", "colon", "period", "hyphen", "apostrophe", "question_mark", "nothing"
})

local SENTENCE_SINGLE_TAGS = {"blank"}
local SENTENCE_OPTIONAL_PAIRED_TAGS = {"i"}

local RATIONALE_OPTIONAL_PAIRED_TAGS = {"i", "em"}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function areAnswerOptionsUnique(rule)
  local answerOptions = rule.options
  local takenOptions = {}
  for _, option in ipairs(answerOptions) do
    if (takenOptions[option]) then
      return false
    end
    takenOptions[option] = true
  end
  return true
end

local function validateAnswerOptionsAreUnique(self, round)
  for _, rule in ipairs(round.rules) do
    local isValid = areAnswerOptionsUnique(rule)
    self:expectTrue(isValid, "list of distractors contains correct answer")
  end
end

local function validateSupportedAnswerTypes(self, round)
  for i, rule in ipairs(round.rules) do
    for _, option in ipairs(rule.options) do
      local isAllowed = ALLOWED_PUNCTUATION_TYPES[option]
      self:expectTrue(isAllowed, "Unsupported punctuation type: "..option)
    end
  end
end

local function validateCharacterBeforeBlankTag(self, round)
  local character = round.characterBeforeBlank
  local validCharacterSet = TableUtils.differenceSet(ConceptValidator.PRINTABLE_SET, {" "})
  local fieldName = "CharacterBeforeBlankTag"

  self:expectNotNil(character, fieldName, "The blank tag should directly follow a word, with no space in between.")
end

local function validateTags(self, round)
  self:validateTags(round.rules[1].hint.body, nil, nil, nil, RATIONALE_OPTIONAL_PAIRED_TAGS, nil)
  self:validateTags(round.rules[1].hint.grammar_note, nil, nil, nil, RATIONALE_OPTIONAL_PAIRED_TAGS, nil)
  self:validateTags(round.sentence, SENTENCE_SINGLE_TAGS, nil, nil, SENTENCE_OPTIONAL_PAIRED_TAGS, nil)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function PunctuationValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  self:validateRequiredKeys(concept.correct_1, REQUIRED_ANSWER_KEYS)

  local round = PunctuationRoundGenerator.conceptToRound(concept)

  validateSupportedAnswerTypes(self, round)
  validateTags(self, round)
  validateAnswerOptionsAreUnique(self, round)
  validateCharacterBeforeBlankTag(self, round)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function PunctuationValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, PunctuationValidator)

  return self
end
