--- RainGame
-- @author Created by Aaron Barrett on July 06, 2015.

class("RainGame")

require("StandardUIBuilder")
require("RainConfiguration")
require("RainRound")
require("RainPanelRound")
require("RainPanelTip")
require("RainRoundGenerator")
require("RainOutro")
require("bonuses.BonusNotificationTray")
require("bonuses.ResponseSpeedTracker")
require("widgets.progress.NumericProgressIndicator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {1.0, 1.0, 1.0}
local UI_COMPONENTS_COLOR = ColorUtils.colorHex24ToStd(0x949494)
local NOTIFICATION_TRAY_COLOR = {1.0, 1.0, 1.0}

local Z_BACKGROUND = 10
local Z_ROUND_PANEL = 20
local Z_TIP_PANEL = 30
local Z_OUTRO = 40

local ROUND_COLORS = {
  ColorUtils.colorHex24ToStd(0xD7B004),
  ColorUtils.colorHex24ToStd(0xf4a82f),
  ColorUtils.colorHex24ToStd(0xf28244),
  ColorUtils.colorHex24ToStd(0x03c6d8)
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getRoundColor(roundIndex)
  local colorIndex = ((roundIndex - 1) % (#ROUND_COLORS)) + 1
  local roundColor = ROUND_COLORS[colorIndex]
  return roundColor
end

local function registerGameWin(self)
  GameSession.notifyGameWin()
end

local function uiColorTo(self, color)
  local uiOverlay = self.uiOverlay
  local colorAction = AHSpawn.new()
  for i, widget in ipairs(uiOverlay.widgets) do
    local widgetColorAction = AHColorTo.new(widget, 0.2, nil, color)
    colorAction:addAction(widgetColorAction)
  end
  colorAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainGame:gameFlow()
  local gameWon = false
  local playerHasReducedAnAnswer = false

  for roundIndex = 1, RainConfiguration.ROUNDS_COUNT do
    local round = self.roundGenerator:next()
    local conceptId = round:getConceptIdentifier()
    local roundColor = getRoundColor(roundIndex)
    local isLastRound = (roundIndex == RainConfiguration.ROUNDS_COUNT)

    local includeTapHelper = (GameSession.hasNeverWon() and (roundIndex == 1))
    local allowReduceHelper = (not playerHasReducedAnAnswer)
    local roundPanel = RainPanelRound.new(round, roundColor, includeTapHelper, allowReduceHelper)
    self:insertAtCenter(roundPanel, Z_ROUND_PANEL)

    wait(roundPanel:playAppear())
    GameSession.contentTracker:notifyPresentedConcept(conceptId)
    self.responseSpeedTracker:notifyStart()

    waitUntil(function() return roundPanel:getSubmittedAnswer() ~= nil end)
    if (roundPanel:didPlayerReduceAnswer()) then
      playerHasReducedAnAnswer = true
    end

    local submittedAnswer = roundPanel:getSubmittedAnswer()
    self.roundCounter:incrementProgress()

    local wasAnswerCorrect = (submittedAnswer == round:getCorrectAnswer())
    if (wasAnswerCorrect) then
      GameSession.contentTracker:notifyCorrectAnswer(conceptId)

      self.responseSpeedTracker:notifyCorrectResponse()
      GameSession.notifyCorrectAnswer(round:getConceptIdentifier())
      GameSession.incrementCorrectAnswerCount()

      GameSession.increaseScore(RainConfiguration.POINTS_PER_CORRECT_RESPONSE)

      self.scoreLabel:setScore(GameSession.getScore())

      uiColorTo(self, NOTIFICATION_TRAY_COLOR)
      roundPanel:playCorrect()

      if not isLastRound then
        Sound.playSoundEffect("correct")
        waitFor(0.5)
        roundPanel:playPostCorrectFade()
        waitFor(0.6)
      else
        Sound.playSoundEffect("win")
        waitFor(0.5)
        self:showDefaultGameWinOutro()
        roundPanel:playFadeSelectionLabel()
        waitFor(0.6)
      end
    else
      GameSession.contentTracker:notifyIncorrectAnswer(conceptId, nil, tostring(submittedAnswer))
      Sound.playSoundEffect("incorrect")
      self.responseSpeedTracker:notifyIncorrectResponse()
      GameSession.notifyIncorrectAnswer(round:getConceptIdentifier())
      GameSession.incrementIncorrectAnswerCount()
      self.livesCounter:decrement()

      wait(roundPanel:playWrong(round:getCorrectAnswer()))

      local tipPanel = RainPanelTip.new(round, roundColor)
      self:insertAtCenter(tipPanel, Z_TIP_PANEL)

      tipPanel:playAppear()
      waitWhile(function() return tipPanel:isShowing() end)
      tipPanel:playDisappear()
      local removeDelay = 0.4
      AHDelayed.new(removeDelay, AHRemove.new(tipPanel)):start()
      if not isLastRound then
        waitFor(0.4)
      else
        uiColorTo(self, NOTIFICATION_TRAY_COLOR)
      end

      if (self.livesCounter:getLives() <= 0) then
        wait(self:showDefaultGameLossOutro())
        GameSession.notifyGameLoss()
      end

      if isLastRound then
        Sound.playSoundEffect("win")
        self:showDefaultGameWinOutro()
        waitFor(0.5)
      end
    end

    if not isLastRound then
      roundPanel:removeSelf()
      uiColorTo(self, UI_COMPONENTS_COLOR)
    end
  end

  ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), RainConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
  self.scoreLabel:setScore(GameSession.getScore())

  local outroColors = {}
  for colorIndex = 1, RainOutro.OUTRO_COLORS_COUNT do
    local outroColor = getRoundColor(RainConfiguration.ROUNDS_COUNT + colorIndex)
    table.insert(outroColors, outroColor)
  end

  local outro = RainOutro.new(outroColors)
  self:insertAtCenter(outro, Z_OUTRO)

  wait(outro:playOutro())

  registerGameWin(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local rankTwoScore = RainConfiguration.getRankTwoScore()
  local rankThreeScore = RainConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = RainConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  self.roundGenerator = RainRoundGenerator.new()

  GameSession.setInitialLives(RainConfiguration.INITIAL_LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

local function createGameElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background, Z_BACKGROUND)

  local speedBonusSeconds = RainConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 RainConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(speedBonusSeconds,
                                                       quickResponseWrapper,
                                                       RainConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()

  local bonusTray = builder:addBonusTray()
  bonusTray:setColor(unpack(NOTIFICATION_TRAY_COLOR))
  self.bonusTray = bonusTray

  self.livesCounter = builder:addLivesDisplay(RainConfiguration.INITIAL_LIVES_COUNT)

  local minValue = 0
  local maxValue = RainConfiguration.ROUNDS_COUNT
  local shouldShowMaxValue = true
  local roundCounter = NumericProgressIndicator.new(StandardUIBuilder.STANDARD_FONT,
                                                    StandardUIBuilder.STANDARD_FONT_SIZE,
                                                    minValue,
                                                    maxValue,
                                                    shouldShowMaxValue)
  local roundCounterOffsetX = nil
  local roundCounterOffsetY = -24
  builder:addWidget(roundCounter, UIOverlay.NORTHEAST, roundCounterOffsetX, roundCounterOffsetY)
  roundCounter:setColor(unpack(UI_COMPONENTS_COLOR))
  self.roundCounter = roundCounter
end

-- Constructor ------------------------

function RainGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, RainGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
