--- RainMoneyBarColumn
-- @author Created by Aaron Barrett on July 06, 2015.

class("RainMoneyBarColumn")

require("RainBill")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = Screen.getSafeWidthPart(0.23)
local HEIGHT = Screen.getHeightPart(0.24)

local STACK_HEIGHT_PER_BILL = Screen.getHeightPart(0.0085)

local LABEL_COLOR = ColorUtils.colorHex24ToStd(0x949494)

local MAX_BILLS_COUNT = 10

local Z_SHAKER = 10
local Z_BILL_STACK = 20
local Z_ADD_BUTTON = 30
local Z_SUBTRACT_BUTTON = 40

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function canAddBill(self)
  local bills = self.bills
  local billsCount = #bills
  return (billsCount < MAX_BILLS_COUNT)
end

local function playAddBill(self)
  Sound.playSoundEffect("add_bill")

  local shaker = self.shaker
  local bills = self.bills
  local roundColor = self.roundColor

  local bill = RainBill.new(roundColor)
  bill:setOpacity(0.0)
  bill:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.25))
  shaker:insert(bill, Z_BILL_STACK)

  table.insert(bills, bill)
  local billsCount = #bills

  local destinationY = self:getHeightPart(0.5) + bill:getHeightPart(0.5) + (billsCount - 1) * STACK_HEIGHT_PER_BILL
  bill:playAppear(destinationY)

  self.onBillAddedCallback()
end

local function canSubtractBill(self)
  local bills = self.bills
  local billsCount = #bills
  return (billsCount >= 1)
end

local function playSubtractBill(self)
  Sound.playSoundEffect("subtract_bill")

  local bills = self.bills
  local bill = table.remove(bills)

  bill:playDisappear()

  self.onBillRemovedCallback()
end

local function playShake(self)
  local shaker = self.shaker
  local shakerInterrupt = self.shakerInterrupt

  local shakeAction = AHSequence.new(
                                     AHLocationXTo.new(shaker, 0.05, nil, self:getWidthPart(0.5 + 0.08)),
                                     AHLocationXTo.new(shaker, 0.08, nil, self:getWidthPart(0.5 - 0.08)),
                                     AHLocationXTo.new(shaker, 0.06, nil, self:getWidthPart(0.5 + 0.03)),
                                     AHLocationXTo.new(shaker, 0.06, nil, self:getWidthPart(0.5 - 0.03)),
                                     AHLocationXTo.new(shaker, 0.04, nil, self:getWidthPart(0.5 + 0.01)),
                                     AHLocationXTo.new(shaker, 0.04, nil, self:getWidthPart(0.5))
                                     )

  shakerInterrupt:insert(shakeAction)
end

local function playUnableToAdd(self)
  Sound.playSoundEffect("unable_to_add")
  playShake(self)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainMoneyBarColumn:getSelectedAmount()
  local bills = self.bills
  local billsCount = #bills
  local denomination = self.denomination

  return billsCount * denomination
end

function RainMoneyBarColumn:setTouchEnabled(enabled)
  self.addButton:setTouchEnabled(enabled)
  self.subtractButton:setTouchEnabled(enabled)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RainMoneyBarColumn.new(denomination, onBillAddedCallback, onBillRemovedCallback, roundColor)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, RainMoneyBarColumn)

  local shaker = Group.new(WIDTH, HEIGHT)
  self:insertAtCenter(shaker, Z_SHAKER)
  self.shaker = shaker
  self.shakerInterrupt = ActionInterrupt.new()

  local denominationText = tostring(denomination)
  local font = Font.newWithActualSize(FontFaces.DIN_OT_MEDIUM, 20)
  local label = Label.new(denominationText, font)
  label:setColor(unpack(LABEL_COLOR))
  label:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.25))
  self:insert(label, Z_LABEL)
  self.label = label

  self.denomination = denomination
  self.bills = {}

  local addButton = Group.new(WIDTH, self:getHeightPart(0.5))
  addButton:setLoc(self:getWidthPart(0.5), addButton:getHeightPart(0.5))
  self:insert(addButton, Z_ADD_BUTTON)
  self.addButton = addButton

  Button.makeButton(addButton, function()
    if (canAddBill(self)) then
      playAddBill(self)
    else
      playUnableToAdd(self)
    end
  end)

  local subtractButton = Group.new(WIDTH, self:getHeightPart(0.5))
  subtractButton:setLoc(self:getWidthPart(0.5), HEIGHT - subtractButton:getHeightPart(0.5))
  self:insert(subtractButton, Z_SUBTRACT_BUTTON)
  self.subtractButton = subtractButton

  Button.makeButton(subtractButton, function()
    if (canSubtractBill(self)) then
      playSubtractBill(self)
    end
  end)

  self.onBillAddedCallback = onBillAddedCallback
  self.onBillRemovedCallback = onBillRemovedCallback
  self.roundColor = roundColor

  self:setTouchEnabled(false)

  return self
end
