--- RainPanelRound
-- @author Created by Aaron Barrett on July 06, 2015.

class("RainPanelRound")

require("RainMoneyBar")
require("RainDotCluster")
require("RainDot")
require("RainHelper")
require("RainWhiteOverlay")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_CLUSTER = 10
local Z_PROMPT_COLUMN = 20
local Z_SELECTION_LABEL = 30
local Z_WHITE_OVERLAY = 40
local Z_MONEY_BAR = 50
local Z_SUBMIT_BUTTON = 60
local Z_HELPER = 70

local QUESTION_LABEL_COLOR = ColorUtils.colorHex24ToStd(0x949494)
local INCORRECT_DANCING_DOTS_COLOR = {0.78, 0.78, 0.78}
local COLOR_WHITE = {1.0, 1.0, 1.0}

local SUBMIT_BUTTON_WIDTH = RainDot.RADIUS * 1.75
local SUBMIT_BUTTON_HEIGHT = SUBMIT_BUTTON_WIDTH

local DIVIDEND_FONT_SIZE = 61
local DIVISOR_FONT_SIZE = 18

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateSelection(self)
  local moneyBar = self.moneyBar
  local submitButton = self.submitButton
  local cluster = self.cluster
  local reduceHelper = self.reduceHelper

  local selectedAmount = moneyBar:getSelectedAmount()

  local showDots = (selectedAmount > 0)
  local submitEnabled = ((showDots) and
                         ((reduceHelper == nil) or (not reduceHelper:isShowing())))

  submitButton:setTouchEnabled(submitEnabled)

  if (showDots) then
    cluster:setSelectedAmount(selectedAmount)
    cluster:playAppear()
  else
    cluster:playDisappear()
  end
end

local function submitAnswer(self)
  local moneyBar = self.moneyBar
  local submitButton = self.submitButton

  submitButton:setTouchEnabled(false)

  moneyBar:playDisappear()

  local selectedAmount = moneyBar:getSelectedAmount()
  self.submittedAnswer = selectedAmount
end

local function dismissHelper(self, helper)
  if ((helper ~= nil) and (helper:isShowing())) then
    helper:playDisappear()
    self.whiteOverlay:playDisappear()
  end
end

local function updateBillsExceedingTotalCountIfNeeded(self)
  if (self.allowReduceHelper) then
    local selectedAmount = self.moneyBar:getSelectedAmount()
    local correctAnswer = self.round:getCorrectAnswer()
    if (selectedAmount > correctAnswer) then
      self.billsExceedingTotalCount = self.billsExceedingTotalCount + 1
    else
      self.billsExceedingTotalCount = 0
    end
  end
end

local function shouldShowReduceHelper(self)
  if (self.allowReduceHelper) then
    local existingReduceHelper = self.reduceHelper
    if (existingReduceHelper == nil) then
      return RainConfiguration.shouldShowReduceHelper(self.billsExceedingTotalCount)
    end
  end
  return false
end

local function showReduceHelper(self)
  local roundColor = self.roundColor
  local reduceHelper = RainHelper.new(RainHelper.KIND_REDUCE, roundColor)
  self:insertAtCenter(reduceHelper, Z_HELPER)
  self.reduceHelper = reduceHelper

  reduceHelper:playAppear()
  self.whiteOverlay:playAppear()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainPanelRound:getSubmittedAnswer()
  return self.submittedAnswer
end

function RainPanelRound:playAppear()
  local moneyBar = self.moneyBar
  local promptColumn = self.promptColumn

  local tapHelper = self.tapHelper
  if (tapHelper ~= nil) then
    tapHelper:playAppear()
    self.whiteOverlay:playAppear()
  end

  local slideAction = AHLocationYTo.new(promptColumn, 0.65, AHEaseExponentialOut.new(), Screen.getSafeY(0.87))
  local enableOnlyDemoColumn = nil

  local barAction = moneyBar:createAppearAction()

  local appearAction = AHStaggeredSpawn.new(0.05,
                                            slideAction,
                                            barAction
                                            )

  return appearAction:start()
end

function RainPanelRound:playCorrect()
  local cluster = self.cluster
  local promptColumn = self.promptColumn

  cluster:playCorrect()

  local fadeAction = AHOpacityTo.new(promptColumn, 0.2, nil, 0.0)
  return fadeAction:start()
end

function RainPanelRound:playPostCorrectFade()
  local cluster = self.cluster
  cluster:playPostCorrectFade()
end

function RainPanelRound:playFadeSelectionLabel()
  local cluster = self.cluster
  return cluster:playFadeSelectionLabel()
end

function RainPanelRound:playWrong(correctAnswer)
  local cluster = self.cluster
  local promptColumn = self.promptColumn

  cluster:playWrong(correctAnswer)

  local wrongAction = AHSequence.new(
                                     AHDelay.new(2.4),
                                     AHOpacityTo.new(promptColumn, 0.5, nil, 0.0)
                                     )

  return wrongAction:start()
end

function RainPanelRound:didPlayerReduceAnswer()
  return self.playerReducedAnswer
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RainPanelRound.new(round, roundColor, includeTapHelper, allowReduceHelper)
  local self = Group.newFullScreen()
  Mixin.mix(self, RainPanelRound)

  local promptColumn = Column.new()

  local dividendText = tostring(round:getDividend())
  local dividendFont = Font.new(FontFaces.DIN_OT_MEDIUM, DIVIDEND_FONT_SIZE)
  local dividendLabel = Label.new(dividendText, dividendFont)
  dividendLabel:inheritColor(promptColumn)
  promptColumn:insert(dividendLabel)

  local divisorText = Localizer.get("DIVIDED_BY") .. " " .. tostring(round:getDivisor())
  local divisorFont = Font.new(FontFaces.DIN_OT_MEDIUM, DIVISOR_FONT_SIZE)
  local divisorLabel = Label.new(divisorText, divisorFont)
  divisorLabel:inheritColor(promptColumn)
  promptColumn:insert(divisorLabel)

  promptColumn:alignChildren()
  promptColumn:setColor(unpack(QUESTION_LABEL_COLOR))
  promptColumn:setLoc(Screen.getSafeX(0.5), self:getHeight() + promptColumn:getHeightPart(0.5))
  self:insert(promptColumn, Z_PROMPT_COLUMN)
  self.promptColumn = promptColumn

  local cluster = RainDotCluster.new(roundColor)
  cluster:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))
  self:insert(cluster, Z_CLUSTER)
  self.cluster = cluster

  local submitButton = Group.new(SUBMIT_BUTTON_WIDTH, SUBMIT_BUTTON_HEIGHT)
  submitButton:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  self:insert(submitButton, Z_SUBMIT_BUTTON)
  self.submitButton = submitButton

  local whiteOverlay = RainWhiteOverlay.new()
  self:insertAtCenter(whiteOverlay, Z_WHITE_OVERLAY)
  self.whiteOverlay = whiteOverlay

  if (includeTapHelper) then
    local tapHelper = RainHelper.new(RainHelper.KIND_TAP, roundColor)
    self:insertAtCenter(tapHelper, Z_HELPER)
    self.tapHelper = tapHelper
  end

  local function onSubmitPressedCallback()
    submitAnswer(self)
  end
  Button.makeButton(submitButton, onSubmitPressedCallback)
  submitButton:setTouchEnabled(false)

  local function onBillAddedCallback()
    dismissHelper(self, self.tapHelper)

    updateBillsExceedingTotalCountIfNeeded(self)
    if (shouldShowReduceHelper(self)) then
      showReduceHelper(self)
    end
    updateSelection(self)
  end

  local function onBillRemovedCallback()
    self.playerReducedAnswer = true
    self.allowReduceHelper = false

    dismissHelper(self, self.reduceHelper)
    updateSelection(self)
  end

  local moneyBar = RainMoneyBar.new(round:getDenominations(), onBillAddedCallback, onBillRemovedCallback, roundColor)
  moneyBar:setLoc(Screen.getSafeX(0.5), Screen.safeBottom + moneyBar:getHeightPart(0.5))
  self:insert(moneyBar, Z_MONEY_BAR)
  self.moneyBar = moneyBar

  self.round = round
  self.roundColor = roundColor
  self.submittedAnswer = nil
  self.allowReduceHelper = allowReduceHelper
  self.playerReducedAnswer = false
  self.billsExceedingTotalCount = 0

  return self
end
