--- RainPanelTip
-- @author Created by Aaron Barrett on July 06, 2015.

class("RainPanelTip")

require("RainEquation")

require("RainTipClassNames")

require("RainTipReverseMultiplication")
require("RainTipDropZeroes")
require("RainTipSplitDivisor")
require("RainTipSplitDividend")
require("RainTipGroupDigits")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local HEADING_LABEL_COLOR = ColorUtils.colorHex24ToStd(0x949494)
local TIP_LABEL_COLOR = ColorUtils.colorHex24ToStd(0x636363)

local Z_MAIN_COLUMN = 10
local Z_CONTINUE_LABEL = 20

local RISE_DISTANCE = Screen.getHeightPart(0.02)

local THEREFORE_DESCENT_DISTANCE = Screen.getHeightPart(0.04)

local HEADING_FONT_SIZE = 24
local THEREFORE_FONT_SIZE = 24
local CONTINUE_FONT_SIZE = 18
local TIP_FONT_SIZE = 27

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createShowThereforeAction(self)
  local thereforeColumn = self.thereforeColumn
  local thereforeLabel = self.thereforeLabel
  local solutionLine = self.solutionLine

  local appearAction = AHSpawn.new(
                                   AHLocationYBy.new(thereforeColumn, 0.5, AHEaseSimpleInOut.new(2.0), -THEREFORE_DESCENT_DISTANCE),
                                   AHOpacityTo.new(thereforeLabel, 0.5, nil, 1.0),
                                   solutionLine:createAppearAction()
                                   )
  return appearAction
end

local function playAnimationLoop(self)
  local animatedTip = self.animatedTip

  if (animatedTip ~= nil) then
    animatedTip:playTip()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainPanelTip:playAppear()
  local actionInterrupt = self.actionInterrupt
  local continueLabel = self.continueLabel
  local wordsColumn = self.wordsColumn
  local animatedTip = self.animatedTip
  local riser = self.riser

  local delayForPerformance = AHDelay.new(0.15)

  animatedTip:prepareToAppear()

  local initialSpawn = AHSpawn.new(
                                   AHOpacityTo.new(wordsColumn, 0.4, nil, 1.0),
                                   AHOpacityTo.new(continueLabel, 0.5, nil, 1.0),
                                   AHLocationYTo.new(riser, 0.4, AHEaseSimpleInOut.new(2.0), Screen.getSafeY(0.5)),
                                   animatedTip:createAppearAction()
                                   )

  local appearAction = AHSequence.new(
                                      delayForPerformance,
                                      initialSpawn,
                                      function()
                                        self:setTouchEnabled(true)
                                      end
                                      )

  local showThereforeAction = createShowThereforeAction(self)

  if (animatedTip.createTipAction ~= nil) then
    local thereforeShown = false

    local animatedLoop = AHLoop.new(
                                    AHSequence.new(
                                                   AHDelay.new(1.4),
                                                   animatedTip:createTipAction(),
                                                   AHDelay.new(1.4),
                                                   function()
                                                     if (not thereforeShown) then
                                                       thereforeShown = true
                                                       wait(showThereforeAction:start())
                                                     end
                                                   end,
                                                   AHDelay.new(2.8),
                                                   animatedTip:createDisappearAction(),
                                                   function()
                                                     animatedTip:prepareToAppear()
                                                   end,
                                                   animatedTip:createAppearAction()
                                                   )
                                    )
    appearAction:addAction(animatedLoop)
  else
    appearAction:addAction(showThereforeAction)
  end

  actionInterrupt:insert(appearAction)
end

function RainPanelTip:playDisappear()
  local actionInterrupt = self.actionInterrupt
  local continueLabel = self.continueLabel
  local wordsColumn = self.wordsColumn
  local thereforeLabel = self.thereforeLabel
  local solutionLine = self.solutionLine
  local animatedTip = self.animatedTip

  local disappearAction = AHSpawn.new(
                                      AHOpacityTo.new(wordsColumn, 0.2, nil, 0.0),
                                      AHOpacityTo.new(continueLabel, 0.2, nil, 0.0),
                                      AHOpacityTo.new(thereforeLabel, 0.2, nil, 0.0),
                                      solutionLine:createDisappearAction(),
                                      animatedTip:createDisappearAction()
                                      )

  actionInterrupt:insert(disappearAction)
end

function RainPanelTip:isShowing()
  return self.showing
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RainPanelTip.new(round, roundColor)
  local self = Group.newFullScreen()
  Mixin.mix(self, RainPanelTip)

  local riser = Group.newFullScreen()
  riser:setLoc(self:getWidthPart(0.5), Screen.getSafeY(0.5) - RISE_DISTANCE)
  self:insert(riser)
  self.riser = riser

  local mainColumnSpacing = Screen.getHeightPart(0.04)
  local mainColumn = Column.new(mainColumnSpacing)

  local wordsColumnSpacing = Screen.getHeightPart(0.02)
  local wordsColumnAlignment = Alignment.LEFT
  local wordsColumn = Column.new(wordsColumnSpacing, wordsColumnAlignment)

  local headingFont = Font.new(FontFaces.DIN_OT_MEDIUM, HEADING_FONT_SIZE)
  local headingLabel = Label.new(Localizer.get("Misc.TIP"), headingFont)
  headingLabel:setColor(unpack(HEADING_LABEL_COLOR))
  headingLabel:inheritOpacity(wordsColumn)
  wordsColumn:insert(headingLabel)

  local tipText = round:getTipText()
  local tipWidth = Screen.getSafeWidthPart(0.85)
  local tipFont = Font.new(FontFaces.DIN_OT_MEDIUM, TIP_FONT_SIZE)
  local attributesForLabels = nil
  local tipLabel = DynamicTextArea.newSimple(tipText, tipWidth, tipFont, Alignment.LEFT)
  tipLabel:setColor(unpack(TIP_LABEL_COLOR))
  tipLabel:inheritOpacity(wordsColumn)
  wordsColumn:insert(tipLabel)
  self.wordsColumn = wordsColumn

  wordsColumn:alignChildren()
  wordsColumn:setOpacity(0.0)
  mainColumn:insert(wordsColumn)

  local tipIdentifier = round:getTipIdentifier()
  local animatedClassName = RainTipClassNames.CLASS_NAMES_BY_TIP_IDENTIFIER[tipIdentifier]
  local animatedClass = _G[animatedClassName]

  local animatedTip = animatedClass.new(round, roundColor)
  mainColumn:insert(animatedTip)
  self.animatedTip = animatedTip

  local thereforeColumnSpacing = Screen.getHeightPart(0.009)
  local thereforeColumn = Column.new(thereforeColumnSpacing)

  local thereforeFont = Font.new(FontFaces.DIN_OT_MEDIUM, THEREFORE_FONT_SIZE)
  local thereforeLabel = Label.new(Localizer.get("THEREFORE"), thereforeFont)
  thereforeLabel:setColor(unpack(TIP_LABEL_COLOR))
  thereforeLabel:setOpacity(0.0)
  thereforeColumn:insert(thereforeLabel)
  self.thereforeLabel = thereforeLabel

  local solutionText = round:getDividend() .. " / " .. round:getDivisor() .. " = " .. round:getCorrectAnswer()
  local solutionLine = RainEquation.new(solutionText, roundColor)
  thereforeColumn:insert(solutionLine)
  self.solutionLine = solutionLine

  thereforeColumn:alignChildren()
  thereforeColumn:setOpacity(0.0)

  mainColumn:insert(thereforeColumn)

  mainColumn:alignChildren()
  mainColumn:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.55))
  riser:insert(mainColumn, Z_MAIN_COLUMN)

  thereforeColumn:setY(thereforeColumn:getY() + THEREFORE_DESCENT_DISTANCE)
  self.thereforeColumn = thereforeColumn

  local continueFont = Font.new(FontFaces.DIN_OT_MEDIUM, CONTINUE_FONT_SIZE)
  local continueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), continueFont)
  continueLabel:setColor(unpack(HEADING_LABEL_COLOR))
  continueLabel:setOpacity(0.0)
  continueLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.05))
  riser:insert(continueLabel, Z_CONTINUE_LABEL)
  self.continueLabel = continueLabel

  self.actionInterrupt = ActionInterrupt.new()

  self.showing = true

  Button.makeButton(self, function()
    self:setTouchEnabled(false)
    self.showing = false
  end)
  self:setTouchEnabled(false)

  return self
end
