--- RainTipGroupDigits
-- @author Created by Aaron Barrett on July 13, 2015.

class("RainTipGroupDigits")

require("RainEquation")
require("RainDividendGrouper")
require("RainMathLayoutBuilder")
require("RainMathActionBuilder")
require("actions.advanced.PathBezier")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GROUPING_SPACING_WIDTH = 8.0
local FRACTION_BAR_HEIGHT = 1.0

local FRACTION_BAR_COLOR = {0.73, 0.73, 0.73}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createNumeratorAction(self)
  local numeratorGroupings = self.numeratorGroupings
  local destinationXByGlyph = self.destinationXByGlyph

  local ease = AHEaseSimpleInOut.new(2.0)

  local numeratorAction = AHSpawn.new()
  for _, numeratorGrouping in ipairs(numeratorGroupings) do
    for _, glyph in ipairs(numeratorGrouping) do
      local destinationX = destinationXByGlyph[glyph]
      local glyphAction = AHLocationXTo.new(glyph, 0.8, ease, destinationX)
      numeratorAction:addAction(glyphAction)
    end
  end
  return numeratorAction
end

local function createDenominatorAction(self)
  local denominatorPaths = self.denominatorPaths

  local ease = AHEaseSimpleInOut.new(2.0)

  local denominatorAction = AHSpawn.new()
  for _, path in ipairs(denominatorPaths) do
    local pathAction = path:createMoveAction(0.8, ease)
    denominatorAction:addAction(pathAction)
  end

  return denominatorAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainTipGroupDigits:createAppearAction()
  local equationLine = self.equationLine
  return equationLine:createAppearAction()
end

function RainTipGroupDigits:createDisappearAction()
  local equationLine = self.equationLine
  local fractionBars = self.fractionBars

  local lineDisappearAction = equationLine:createDisappearAction()

  local fractionBarsAction = AHSpawn.new()
  for _, bar in ipairs(fractionBars) do
    local barAction = AHOpacityTo.new(bar, 0.2, nil, 0.0)
    fractionBarsAction:addAction(barAction)
  end

  local disappearAction = AHSpawn.new(
                                      lineDisappearAction,
                                      fractionBarsAction
                                      )
  return disappearAction
end

function RainTipGroupDigits:prepareToAppear()
  local equationLine = self.equationLine
  local denominatorGroupings = self.denominatorGroupings
  local divisorGlyphs = self.divisorGlyphs
  local denominatorPaths = self.denominatorPaths

  equationLine:centerGlyphs()

  for _, denominatorGrouping in ipairs(denominatorGroupings) do
    FunctionUtils.callMethodOnTargets("setVisible", denominatorGrouping, false)
    FunctionUtils.callMethodOnTargets("setOpacity", denominatorGrouping, 1.0)
  end

  for _, path in ipairs(denominatorPaths) do
    path:setProgress(0.0)
  end

  FunctionUtils.callMethodOnTargets("setVisible", divisorGlyphs, true)
end

function RainTipGroupDigits:createTipAction()
  local equationLine = self.equationLine

  local dividendGlyphs = self.dividendGlyphs
  local denominatorGroupings = self.denominatorGroupings
  local divisorGlyphs = self.divisorGlyphs
  local quotientGlyphs = self.quotientGlyphs
  local questionMarkGlyph = self.questionMarkGlyph
  local divisionSignGlyph = self.divisionSignGlyph
  local equalsSignGlyph = self.equalsSignGlyph
  local equalsSignDestinationX = self.equalsSignDestinationX
  local groupingsCount = self.groupingsCount
  local fractionBars = self.fractionBars

  local initialGlyphs = equationLine:getInitialGlyphs()

  local tipAction = AHSequence.new()

  tipAction:addAction(function()
    FunctionUtils.callMethodOnTargets("setVisible", divisorGlyphs, false)
    for _, denominatorGrouping in ipairs(denominatorGroupings) do
      FunctionUtils.callMethodOnTargets("setVisible", denominatorGrouping, true)
    end
  end)

  local shiftPhaseAction = AHSpawn.new()
  local numeratorAction = createNumeratorAction(self)
  shiftPhaseAction:addAction(numeratorAction)
  local denominatorAction = createDenominatorAction(self)
  shiftPhaseAction:addAction(denominatorAction)
  local fractionBarsAction = AHSpawn.new()
  for _, bar in ipairs(fractionBars) do
    local barAction = AHOpacityTo.new(bar, 0.4, nil, 1.0)
    fractionBarsAction:addAction(barAction)
  end
  shiftPhaseAction:addAction(AHDelayed.new(0.6, fractionBarsAction))
  local questionMarkAction = AHOpacityTo.new(questionMarkGlyph, 0.2, nil, 0.0)
  shiftPhaseAction:addAction(questionMarkAction)
  local divisionSignAction = AHOpacityTo.new(divisionSignGlyph, 0.2, nil, 0.0)
  shiftPhaseAction:addAction(divisionSignAction)
  local equalsSignAction = AHLocationTo.new(equalsSignGlyph, 0.8, AHEaseSimpleInOut.new(2.0), equalsSignDestinationX, 0.0)
  shiftPhaseAction:addAction(equalsSignAction)
  tipAction:addAction(shiftPhaseAction)

  RainMathActionBuilder.addGlyphSpawnToAction(tipAction, quotientGlyphs, function(glyph)
    return AHOpacityTo.new(glyph, 0.2, nil, 1.0)
  end)

  return tipAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildNumeratorGroupings(self, dividend, divisor, initialGlyphs)
  local groupingStrings = RainDividendGrouper.getDividendGroupingStrings(dividend, divisor)

  local glyphIndex = 1
  local numeratorGroupings = {}
  for _, groupingString in ipairs(groupingStrings) do
    local glyphGrouping = {}
    local groupingLength = #groupingString
    for characterIndex = 1, groupingLength do
      local glyph = initialGlyphs[glyphIndex]
      table.insert(glyphGrouping, glyph)
      glyphIndex = glyphIndex + 1
    end
    table.insert(numeratorGroupings, glyphGrouping)
  end

  self.numeratorGroupings = numeratorGroupings
  return numeratorGroupings
end

local function buildDenominatorGroupings(self, divisorText, groupingsCount, equationLine)
  local denominatorGroupings = {}
  for groupingIndex = 1, groupingsCount do
    local denominatorGrouping = equationLine:addGlyphsForText(divisorText)
    table.insert(denominatorGroupings, denominatorGrouping)
  end
  self.denominatorGroupings = denominatorGroupings
  return denominatorGroupings
end

local function getGroupingStartXCoordinates(self, numeratorGroupings, dividendGlyphs)
  local groupingsCount = #numeratorGroupings
  local dividendWidth = RainMathLayoutBuilder.getWidthOfGlyphs(dividendGlyphs)
  local dividendCenterX = dividendWidth * 0.5
  local leftmostGroupingStartX = dividendCenterX - 0.5 * (dividendWidth + groupingsCount * GROUPING_SPACING_WIDTH)

  local groupingStartXCoordinates = {}
  local groupingStartX = leftmostGroupingStartX
  for _, numeratorGrouping in ipairs(numeratorGroupings) do
    table.insert(groupingStartXCoordinates, groupingStartX)
    local groupingWidth = RainMathLayoutBuilder.getWidthOfGlyphs(numeratorGrouping)
    groupingStartX = groupingStartX + groupingWidth + GROUPING_SPACING_WIDTH
  end

  return groupingStartXCoordinates
end

local function buildDestinationXByGlyph(self, numeratorGroupings, denominatorGroupings, groupingStartXCoordinates)
  local destinationXByGlyph = {}
  for numeratorGroupingIndex, glyphGrouping in ipairs(numeratorGroupings) do
    local groupingWidth = RainMathLayoutBuilder.getWidthOfGlyphs(glyphGrouping)

    local groupingStartX = groupingStartXCoordinates[numeratorGroupingIndex]
    local numeratorXCoordinates = RainMathLayoutBuilder.getXCoordinatesForGlyphs(glyphGrouping, groupingStartX)
    for glyphIndex, glyph in ipairs(glyphGrouping) do
      local xCoordinate = numeratorXCoordinates[glyphIndex]
      destinationXByGlyph[glyph] = xCoordinate
    end

    local denominatorGrouping = denominatorGroupings[numeratorGroupingIndex]
    local denominatorWidth = RainMathLayoutBuilder.getWidthOfGlyphs(denominatorGrouping)
    local denominatorStartX = groupingStartX + (groupingWidth - denominatorWidth) * 0.5
    local denominatorXCoordinates = RainMathLayoutBuilder.getXCoordinatesForGlyphs(denominatorGrouping, denominatorStartX)
    for glyphIndex, glyph in ipairs(denominatorGrouping) do
      local xCoordinate = denominatorXCoordinates[glyphIndex]
      destinationXByGlyph[glyph] = xCoordinate
    end
  end
  self.destinationXByGlyph = destinationXByGlyph
  return destinationXByGlyph
end

local function buildFractionBars(self, numeratorGroupings, groupingStartXCoordinates)
  local fractionBars = {}
  for groupingIndex, numeratorGrouping in ipairs(numeratorGroupings) do
    local groupingStartX = groupingStartXCoordinates[groupingIndex]
    local groupingWidth = RainMathLayoutBuilder.getWidthOfGlyphs(numeratorGrouping)

    local fractionBar = FilledRectangle.new(groupingWidth, FRACTION_BAR_HEIGHT, FRACTION_BAR_COLOR)
    fractionBar:setOpacity(0.0)
    fractionBar:setLoc(groupingStartX + groupingWidth * 0.5, self:getHeightPart(0.5))
    self:insert(fractionBar)
    table.insert(fractionBars, fractionBar)
  end

  self.fractionBars = fractionBars
  return fractionBars
end

local function buildDenominatorPaths(self, denominatorGroupings, destinationXByGlyph, initialGlyphs, firstDivisorDigitIndex)
  local glyphsBeforeFirstDivisorDigit = TableUtils.subarray(initialGlyphs, 1, firstDivisorDigitIndex - 1)
  local widthOfGlyphsBeforeFirstDivisorDigit = RainMathLayoutBuilder.getWidthOfGlyphs(glyphsBeforeFirstDivisorDigit)

  local firstDenominatorGrouping = denominatorGroupings[1]
  local startXCoordinates = RainMathLayoutBuilder.getXCoordinatesForGlyphs(firstDenominatorGrouping, widthOfGlyphsBeforeFirstDivisorDigit)

  local denominatorPaths = {}

  for _, denominatorGrouping in ipairs(denominatorGroupings) do
    for glyphIndex, glyph in ipairs(denominatorGrouping) do
      local startX = startXCoordinates[glyphIndex]
      local endX = destinationXByGlyph[glyph]

      local startY = glyph:getHeightPart(0.5)
      local endY = -glyph:getHeightPart(0.5)

      local startControlPoint = {x = startX * 0.9, y = -glyph:getHeight()}
      local startPoint = {x = startX, y = startY}
      local endPoint = {x = endX, y = endY}
      local endControlPoint = {x = endX * 1.2, y = -glyph:getHeight()}

      local path = PathBezier.new(glyph, startControlPoint, startPoint, endPoint, endControlPoint)
      table.insert(denominatorPaths, path)
    end
  end

  self.denominatorPaths = denominatorPaths
end

-- Constructor ------------------------

function RainTipGroupDigits.new(round, roundColor)
  local dividend = round:getDividend()
  local divisor = round:getDivisor()

  local quotient = round:getCorrectAnswer()

  local dividendText = tostring(dividend)
  local divisorText = tostring(divisor)
  local quotientText = tostring(quotient)
  local spaceText = " "
  local divisionSignText = spaceText .. "/" .. spaceText

  local equationText = dividendText .. divisionSignText .. divisorText .. spaceText .. "=" .. spaceText .. "?"
  local equationLine = RainEquation.new(equationText, roundColor)

  local width = equationLine:getWidth()
  local height = equationLine:getHeightPart(2.0)

  local self = Group.new(width, height)
  Mixin.mix(self, RainTipGroupDigits)

  equationLine:setLoc(self:getWidthPart(0.5), equationLine:getHeightPart(1.5))
  self:insert(equationLine)
  self.equationLine = equationLine

  local quotientGlyphs = equationLine:addGlyphsForText(quotientText)
  self.quotientGlyphs = quotientGlyphs

  local equationTextLength = #equationText
  local dividendTextLength = #dividendText
  local divisorTextLength = #divisorText
  local spaceTextLength = #spaceText
  local divisionSignTextLength = #divisionSignText

  local initialGlyphs = equationLine:getInitialGlyphs()
  local questionMarkGlyph = initialGlyphs[equationTextLength]
  self.questionMarkGlyph = questionMarkGlyph
  local divisionSignGlyph = initialGlyphs[dividendTextLength + spaceTextLength + 1]
  self.divisionSignGlyph = divisionSignGlyph
  local equalsSignIndex = dividendTextLength + divisionSignTextLength + divisorTextLength + spaceTextLength + 1
  local equalsSignGlyph = initialGlyphs[equalsSignIndex]
  self.equalsSignGlyph = equalsSignGlyph

  local numeratorGroupings = buildNumeratorGroupings(self, dividend, divisor, initialGlyphs)
  local groupingsCount = #numeratorGroupings
  self.groupingsCount = groupingsCount

  local denominatorGroupings = buildDenominatorGroupings(self, divisorText, groupingsCount, equationLine)

  local dividendGlyphs = TableUtils.subarray(initialGlyphs, 1, dividendTextLength)
  self.dividendGlyphs = dividendGlyphs

  local groupingStartXCoordinates = getGroupingStartXCoordinates(self, numeratorGroupings, dividendGlyphs)

  local destinationXByGlyph = buildDestinationXByGlyph(self, numeratorGroupings, denominatorGroupings, groupingStartXCoordinates)

  buildFractionBars(self, numeratorGroupings, groupingStartXCoordinates)

  local divisionSignTextLength = #divisionSignText
  local firstDivisorDigitIndex = dividendTextLength + divisionSignTextLength + 1
  self.firstDivisorDigitIndex = firstDivisorDigitIndex

  buildDenominatorPaths(self, denominatorGroupings, destinationXByGlyph, initialGlyphs, firstDivisorDigitIndex)

  self.divisorGlyphs = TableUtils.subarray(initialGlyphs, firstDivisorDigitIndex, firstDivisorDigitIndex + divisorTextLength - 1)

  local finalDividendGrouping = numeratorGroupings[groupingsCount]
  local finalDividendGroupingGlyph = finalDividendGrouping[#finalDividendGrouping]
  local finalDividendGroupingGlyphDestinationX = destinationXByGlyph[finalDividendGroupingGlyph]
  local equalsSignDestinationX = (finalDividendGroupingGlyphDestinationX + equalsSignGlyph:getX()) * 0.5
  self.equalsSignDestinationX = equalsSignDestinationX

  local equalsSignPlusSpace = TableUtils.subarray(initialGlyphs, equalsSignIndex, equalsSignIndex + 1)
  local equalsSignPlusSpaceWidth = RainMathLayoutBuilder.getWidthOfGlyphs(equalsSignPlusSpace)
  local quotientStartX = equalsSignDestinationX + equalsSignPlusSpaceWidth - equalsSignGlyph:getWidthPart(0.5) - equalsSignGlyph:getLeftMargin()
  RainMathLayoutBuilder.alignGlyphs(quotientGlyphs, quotientStartX, 0.0)

  return self
end
