--- RainTipSplitDivisor
-- @author Created by Aaron Barrett on July 09, 2015.

class("RainTipSplitDivisor")

require("RainEquation")
require("RainMathLayoutBuilder")
require("RainDivisorFactors")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainTipSplitDivisor:createAppearAction()
  local equationLine = self.equationLine
  return equationLine:createAppearAction()
end

function RainTipSplitDivisor:createDisappearAction()
  local equationLine = self.equationLine
  return equationLine:createDisappearAction()
end

function RainTipSplitDivisor:prepareToAppear()
  local equationLine = self.equationLine
  local factorGlyphGroupings = self.factorGlyphGroupings
  local widthOfGlyphsBeforeDivisionSign = self.widthOfGlyphsBeforeDivisionSign

  equationLine:centerGlyphs()

  for _, factorGlyphGrouping in ipairs(factorGlyphGroupings) do
    RainMathLayoutBuilder.alignGlyphs(factorGlyphGrouping, widthOfGlyphsBeforeDivisionSign)
  end
end

function RainTipSplitDivisor:createTipAction()
  local equationLine = self.equationLine

  local glyphs = equationLine:getGlyphs()
  local glyphsCount = #glyphs
  local initialGlyphs = equationLine:getInitialGlyphs()
  local initialGlyphsCount = #initialGlyphs
  local questionMarkIndex = initialGlyphsCount
  local questionMarkGlyph = initialGlyphs[questionMarkIndex]
  local finalSpaceIndex = questionMarkIndex - 1
  local finalSpaceGlyph = initialGlyphs[finalSpaceIndex]

  local glyphsToRemove = self.glyphsToRemove
  local factorGlyphGroupings = self.factorGlyphGroupings
  local postFactorizationGlyphs = self.postFactorizationGlyphs
  local solutionGlyphs = self.solutionGlyphs
  local postRevealGlyphs = self.postRevealGlyphs

  local tipAction = AHSequence.new()

  RainMathActionBuilder.addGlyphSpawnToAction(tipAction, glyphsToRemove, function(glyph)
    return AHOpacityTo.new(glyph, 0.2, nil, 0.0)
  end)

  local factorizationAction = AHSpawn.new()
  local factorizationAlignAction = equationLine:createCenterAction(postFactorizationGlyphs)
  factorizationAction:addAction(factorizationAlignAction)
  for _, factorGlyphGrouping in ipairs(factorGlyphGroupings) do
    RainMathActionBuilder.addGlyphActionsToAction(factorizationAction, factorGlyphGrouping, function(glyph)
      return AHOpacityTo.new(glyph, 0.8, nil, 1.0)
    end)
  end
  tipAction:addAction(factorizationAction)

  tipAction:addAction(AHDelay.new(1.4))

  local questionMarkFadePhaseAction = AHSpawn.new()
  local revealAlignAction = equationLine:createCenterAction(postRevealGlyphs)
  questionMarkFadePhaseAction:addAction(revealAlignAction)
  local questionMarkFadeAction = AHOpacityTo.new(questionMarkGlyph, 0.2, nil, 0.0)
  questionMarkFadePhaseAction:addAction(questionMarkFadeAction)
  tipAction:addAction(questionMarkFadePhaseAction)

  RainMathActionBuilder.addGlyphSpawnToAction(tipAction, solutionGlyphs, function(glyph)
    return AHOpacityTo.new(glyph, 0.5, nil, 1.0)
  end)

  return tipAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RainTipSplitDivisor.new(round, roundColor)
  local dividend = round:getDividend()
  local divisor = round:getDivisor()

  local tipIdentifier = round:getTipIdentifier()
  local divisorFactors = RainDivisorFactors.DIVISOR_FACTORS_BY_TIP_IDENTIFIER[tipIdentifier]

  local dividendText = tostring(dividend)
  local divisorText = tostring(divisor)
  local firstSpaceText = " "
  local divisionSignText = "/ "
  local equationText = dividendText .. firstSpaceText .. divisionSignText .. divisorText .. " = ?"
  local equationLine = RainEquation.new(equationText, roundColor)

  local width, height = equationLine:getDims()

  local self = Group.new(width, height)
  Mixin.mix(self, RainTipSplitDivisor)

  self:insertAtCenter(equationLine)
  self.equationLine = equationLine

  local initialGlyphs = equationLine:getInitialGlyphs()
  local initialGlyphsCount = #initialGlyphs

  local dividendTextLength = #dividendText
  local divisorTextLength = #divisorText
  local firstSpaceTextLength = #firstSpaceText
  local divisionSignTextLength = #divisionSignText

  local removalStartIndex = dividendTextLength + firstSpaceTextLength + 1
  local removalEndIndex = removalStartIndex + divisionSignTextLength + divisorTextLength

  local glyphsToRemove = TableUtils.subarray(initialGlyphs, removalStartIndex, removalEndIndex)
  self.glyphsToRemove = glyphsToRemove

  local postFactorizationGlyphs = TableUtils.subarray(initialGlyphs, 1, removalStartIndex - 1)
  self.postFactorizationGlyphs = postFactorizationGlyphs

  local widthOfGlyphsBeforeDivisionSign = RainMathLayoutBuilder.getWidthOfGlyphs(postFactorizationGlyphs)
  self.widthOfGlyphsBeforeDivisionSign = widthOfGlyphsBeforeDivisionSign

  local factorGlyphGroupings = {}
  for _, factor in ipairs(divisorFactors) do
    local factorText = nil
    if (factor > 0) then
      factorText = "/ " .. tostring(factor) .. " "
    else
      factorText = "* " .. tostring(-factor) .. " "
    end

    local factorGlyphGrouping = equationLine:addGlyphsForText(factorText)
    table.insert(factorGlyphGroupings, factorGlyphGrouping)

    TableUtils.concatenateArrays(postFactorizationGlyphs, factorGlyphGrouping)
  end

  local remainingPostFactorizationGlyphs = TableUtils.subarray(initialGlyphs, removalEndIndex + 1, initialGlyphsCount)
  TableUtils.concatenateArrays(postFactorizationGlyphs, remainingPostFactorizationGlyphs)

  self.factorGlyphGroupings = factorGlyphGroupings
  self.postFactorizationGlyphs = postFactorizationGlyphs

  local solution = round:getCorrectAnswer()
  local solutionText = tostring(solution)
  local solutionGlyphs = equationLine:addGlyphsForText(solutionText)
  RainMathLayoutBuilder.alignGlyphs(solutionGlyphs)
  self.solutionGlyphs = solutionGlyphs

  local postFactorizationGlyphsCount = #postFactorizationGlyphs
  local postRevealGlyphs = TableUtils.subarray(postFactorizationGlyphs, 1, postFactorizationGlyphsCount - 1)
  TableUtils.concatenateArrays(postRevealGlyphs, solutionGlyphs)
  self.postRevealGlyphs = postRevealGlyphs

  return self
end
