--- RainValidator
-- @author Created by Aaron Barrett on July 06, 2015.
-- ConceptValidator for Rain.

class("RainValidator")

require("validators.ConceptValidator")

require("RainTipClassNames")
require("RainDividendGrouper")
require("RainDivisorFactors")
require("RainZeroHelpers")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "dividend",
  "divisor",
  "tip_identifier"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateNoDecimals(self, concept)
  local dividend = concept["dividend"]
  local divisor = concept["divisor"]

  local quotient = dividend / divisor

  local errorMessage = string.format("Concept contained dividend `%d` and divisor `%d`, which produced a decimal", dividend, divisor)
  self:expectTrue(MathUtils.isInteger(quotient), errorMessage)
end

local function validateTipHasAnimation(self, concept)
  local tipIdentifier = concept["tip_identifier"]
  self:expectNotNil(tipIdentifier, "tip_identifier")

  local tipClassName = RainTipClassNames.CLASS_NAMES_BY_TIP_IDENTIFIER[tipIdentifier]

  local message = "No animated tip class listed in RainTipClassNames.lua for `tip_identifier` of `" .. tostring(tipIdentifier) .. "`"
  self:expectNotNil(tipClassName, "tip_identifier", message)

  if (tipClassName == "RainTipSplitDividend") then
    local tipPartition = concept["part_number"]

    local message = "Tip identifier `" .. tipIdentifier .. "` (associated with `RainTipSplitDividend`) requires a positive integer `part_number`."
    self:expectGreaterThan(tipPartition, 0, message)
  elseif (tipClassName == "RainTipSplitDivisor") then
    local divisorFactors = RainDivisorFactors.DIVISOR_FACTORS_BY_TIP_IDENTIFIER[tipIdentifier]
    local message = "No divisor factors defined in RainDivisorFactors.lua for `tip_identifier` of `" .. tostring(tipIdentifier) .. "`."
    self:expectNotNil(divisorFactors, "divisorFactors", message)
  elseif (tipClassName == "RainTipGroupDigits") then
    local dividend = concept["dividend"]
    local divisor = concept["divisor"]

    local groupingStrings = RainDividendGrouper.getDividendGroupingStrings(dividend, divisor)
    local groupingStringsCount = #groupingStrings
    local message = "Tip identifier `" .. tipIdentifier .. "` is associated with `RainTipGroupDigits`, but dividend cannot be grouped into enough divisible substrings."
    self:expectTrue(groupingStringsCount > 1, message)
  elseif (tipClassName == "RainTipDropZeroes") then
    local dividend = concept["dividend"]
    local divisor = concept["divisor"]

    local dividendText = tostring(dividend)
    local divisorText = tostring(divisor)
    local dividendZeroesToDrop, divisorZeroesToDrop = RainZeroHelpers.getZeroesToDrop(dividendText, divisorText)
    local message = "Tip identifier `" .. tipIdentifier .. "` is associated with `RainTipDropZeroes`, but no zeroes could be dropped."
    self:expectGreaterThan(math.max(dividendZeroesToDrop, divisorZeroesToDrop), 0, message)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RainValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateNoDecimals(self, concept)
  validateTipHasAnimation(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RainValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, RainValidator)

  return self
end
