--- RiffleArticle
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleArticle")

require("RiffleArticleTextLabel")
require("RiffleCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local TEXT_WIDTH = Screen.getWidthPart(0.88)
local HEADER_SPACING_Y = 60
local HEADER_SUBHEADER_SPACING_Y = 0
local SUBHEADER_BODY_SPACING_Y = 30
local SUBHEADER_IMAGE_SPACING_Y = 15
local BODY_BOTTOM_SPACING_Y = Screen.getHeightPart(0.25)
local ARTICLE_END_PADDING_Y = Screen.getHeightPart(0.03)
local NO_SUBHEADER_HEADER_PADDING_Y = Screen.getHeightPart(0.01)

local LayoutDescriptor =
{
  [RiffleCommon.LayoutStyle.IMAGE_AND_SUBHEADER] =
  {
    headerFont = FontFaces.DIN_OT_BOLD,
    headerFontSize = 30,
    headerColor = RiffleCommon.COLORS.MUTED_BLACK,
    subheaderFont = FontFaces.DIN_OT_MEDIUM,
    subheaderFontSize = 15,
    subheaderColor = RiffleCommon.COLORS.GREY,
    bodyFont = FontFaces.TISA_REGULAR,
    bodyFontSize = 18,
    bodyColor = RiffleCommon.COLORS.BLACK,
  },
  [RiffleCommon.LayoutStyle.IMAGE_ONLY] =
  {
    headerFont = FontFaces.DIN_OT_BOLD,
    headerFontSize = 26,
    headerColor = RiffleCommon.COLORS.WHITE,
    bodyFont = FontFaces.TISA_REGULAR,
    bodyFontSize = 18,
    bodyColor = RiffleCommon.COLORS.BLACK
  },
  [RiffleCommon.LayoutStyle.SUBHEADER_ONLY] =
  {
    headerFont = FontFaces.DIN_OT_BOLD,
    headerFontSize = 30,
    headerColor = RiffleCommon.COLORS.MUTED_BLACK,
    subheaderFont = FontFaces.DIN_OT_MEDIUM,
    subheaderFontSize = 16,
    subheaderColor = RiffleCommon.COLORS.GREY,
    bodyFont = FontFaces.TISA_REGULAR,
    bodyFontSize = 18,
    bodyColor = RiffleCommon.COLORS.BLACK
  },
  [RiffleCommon.LayoutStyle.NO_ORNAMENTS] =
  {
    headerFont = FontFaces.DIN_OT_BOLD,
    headerFontSize = 30,
    headerColor = RiffleCommon.COLORS.MUTED_BLACK,
    subheaderFont = FontFaces.DIN_OT_MEDIUM,
    subheaderFontSize = 16,
    subheaderColor = RiffleCommon.COLORS.GREY,
    bodyFont = FontFaces.TISA_REGULAR,
    bodyFontSize = 18,
    bodyColor = RiffleCommon.COLORS.BLACK
  }
}
local IMAGE_AND_SUBHEADER_HEADER_FONT = FontFaces.DIN_OT_MEDIUM
local SUBHEADER_FONT = FontFaces.DIN_OT_MEDIUM
local CREDIT_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 12)
local CREDIT_OFFSET_Y = 6
local MINIMUM_IMAGE_HEIGHT_PROPORTION = Screen.getHeightPart(0.37)
local MINIMUM_IMAGE_SCALE = 1.0
local MAXIMUM_IMAGE_SCALE = 1.4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function fitToScreen(image)
  local horizontalCorrectionScale = Screen.width / image:getWidth()
  local verticalCorrectionScale = MINIMUM_IMAGE_HEIGHT_PROPORTION / image:getHeight()

  -- Should NEVER downscale! Requirements are that the image always stretches to the edges
  -- and should occupy a good chunk of screen height
  local largestScale = math.max(horizontalCorrectionScale, verticalCorrectionScale)
  local imageScale = MathUtils.clamp(largestScale, MINIMUM_IMAGE_SCALE, MAXIMUM_IMAGE_SCALE)
  image:setScl(imageScale)
end

local function createImageCredit(round)
  local creditText = "Image credit: " .. round.imageCredit
  local creditLabel = DynamicTextArea.newSimple(creditText, Screen.getWidthPart(0.87), CREDIT_FONT, Alignment.LEFT)
  creditLabel:setColor(unpack(RiffleCommon.COLORS.LIGHT_GREY))

  return creditLabel
end

local function createYSpacer(spacingY)
  local spacer = Group.new(Screen.width, spacingY)

  return spacer
end

local function createLayoutFromColumn(elementsColumn)
  local width = Screen.width
  local height = elementsColumn:getHeight()
  local layout = FilledRectangle.new(width, height, RiffleCommon.COLORS.WHITE)
  layout:insertAtCenter(elementsColumn)

  return layout
end

local function createHeader(round, layoutDescriptor)
  local headerText = round.title
  local header = DynamicTextArea.newSimple(headerText, TEXT_WIDTH, Font.new(layoutDescriptor.headerFont, layoutDescriptor.headerFontSize))
  header:setColor(unpack(layoutDescriptor.headerColor))

  return header
end

local function createSubheader(round, layoutDescriptor)
  local subheaderText = round.subheader
  local subheader = DynamicTextArea.newSimple(subheaderText, TEXT_WIDTH, Font.new(layoutDescriptor.subheaderFont, layoutDescriptor.subheaderFontSize))
  subheader:setColor(unpack(layoutDescriptor.subheaderColor))

  return subheader
end

local function createArticleImage(round, layoutDescriptor)
  local imageFilename = round.image
  local photo = AssetHelper.createPhotoSpriteForConcept(round, imageFilename)
  fitToScreen(photo)

  local articleImage = Group.new(photo:getEffectiveWidth(), photo:getEffectiveHeight())

  articleImage:insertAtCenter(photo)

  return articleImage
end

local function createBody(round, layoutDescriptor)
  local bodyText = round.passage
  local body = RiffleArticleTextLabel.new(bodyText, TEXT_WIDTH, layoutDescriptor.bodyFont, layoutDescriptor.bodyFontSize)
  body:setColor(unpack(layoutDescriptor.bodyColor))

  return body
end

local function createNothingTouchedRegion(self, articleBody)
  local nothingTouchedRegion = Group.new(articleBody:getWidth(), articleBody:getHeight())

  nothingTouchedRegion:setLoc(self:getWidthPart(0.5), articleBody:getY())
  self:insert(nothingTouchedRegion)

  -- Make sure the nothing touched region will be behind the article objects
  -- or else they will intercept taps and all scanning responses will be wrong
  nothingTouchedRegion:bringToBack()

  return nothingTouchedRegion
end

local function createLayoutImageAndSubheader(round)
  local layoutDescriptor = LayoutDescriptor[RiffleCommon.LayoutStyle.IMAGE_AND_SUBHEADER]

  local header = createHeader(round, layoutDescriptor)
  local subheader = createSubheader(round, layoutDescriptor)
  local articleImage = createArticleImage(round, layoutDescriptor)
  local body = createBody(round, layoutDescriptor)
  local imageCredit = createImageCredit(round)

  local elementsColumn = Column.new()
  elementsColumn:insertSpacer(RiffleCommon.UI_BAR_HEIGHT)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(header)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.01))
  elementsColumn:insert(subheader)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(articleImage)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(body)
  elementsColumn:insertSpacer(ARTICLE_END_PADDING_Y)
  elementsColumn:insert(imageCredit)
  elementsColumn:insertSpacer(RiffleCommon.COLLAPSED_QUESTION_PANEL_HEIGHT - imageCredit:getHeight())

  elementsColumn:alignChildren()

  local self = createLayoutFromColumn(elementsColumn)

  self.articleLabel = body
  self.nothingTouchedRegion = createNothingTouchedRegion(self, body)

  return self
end

local function createLayoutNoImage(round)
  local layoutDescriptor = LayoutDescriptor[RiffleCommon.LayoutStyle.SUBHEADER_ONLY]

  local header = createHeader(round, layoutDescriptor)
  local subheader = createSubheader(round, layoutDescriptor)
  local body = createBody(round, layoutDescriptor)

  local elementsColumn = Column.new()
  elementsColumn:insertSpacer(RiffleCommon.UI_BAR_HEIGHT)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(header)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.01))
  elementsColumn:insert(subheader)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.05))
  elementsColumn:insert(body)
  elementsColumn:insertSpacer(ARTICLE_END_PADDING_Y)
  elementsColumn:insertSpacer(RiffleCommon.COLLAPSED_QUESTION_PANEL_HEIGHT)

  elementsColumn:alignChildren()

  local self = createLayoutFromColumn(elementsColumn)

  self.articleLabel = body
  self.nothingTouchedRegion = createNothingTouchedRegion(self, body)

  return self
end

local function createLayoutNoSubheader(round)
  local layoutDescriptor = LayoutDescriptor[RiffleCommon.LayoutStyle.IMAGE_ONLY]

  local articleImage = createArticleImage(round, layoutDescriptor)
  local body = createBody(round, layoutDescriptor)
  local imageCredit = createImageCredit(round)

  local elementsColumn = Column.new()
  elementsColumn:insert(articleImage)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(body)
  elementsColumn:insertSpacer(ARTICLE_END_PADDING_Y)
  elementsColumn:insert(imageCredit)
  elementsColumn:insertSpacer(RiffleCommon.COLLAPSED_QUESTION_PANEL_HEIGHT - imageCredit:getHeight())

  elementsColumn:alignChildren()

  local self = createLayoutFromColumn(elementsColumn)

  local header = createHeader(round, layoutDescriptor)
  header:setAnchorPointBottom()
  header:setColor(unpack(layoutDescriptor.headerColor))
  header:setLoc(articleImage:getWidthPart(0.5), NO_SUBHEADER_HEADER_PADDING_Y)

  local backdropHeight = header:getHeightPart(2.0)
  local backdropWidth = Screen.width
  local backdropColorStart = RiffleCommon.COLORS.INVISIBLE_BLACK
  local backdropColorEnd = RiffleCommon.COLORS.BLACK
  local backdrop = RectangleGradient.new(backdropWidth, backdropHeight, backdropColorStart, backdropColorEnd)
  backdrop:setLoc(articleImage:getWidthPart(0.5), 0)
  backdrop:setAnchorPointBottom()

  articleImage:insert(backdrop)
  articleImage:insert(header)

  self.articleLabel = body
  self.nothingTouchedRegion = createNothingTouchedRegion(self, body)

  return self
end

local function createLayoutNoOrnaments(round)
  local layoutDescriptor = LayoutDescriptor[RiffleCommon.LayoutStyle.SUBHEADER_ONLY]

  local header = createHeader(round, layoutDescriptor)
  local body = createBody(round, layoutDescriptor)

  local elementsColumn = Column.new()
  elementsColumn:insertSpacer(RiffleCommon.UI_BAR_HEIGHT)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.03))
  elementsColumn:insert(header)
  elementsColumn:insertSpacer(Screen.getHeightPart(0.05))
  elementsColumn:insert(body)
  elementsColumn:insertSpacer(ARTICLE_END_PADDING_Y)
  elementsColumn:insertSpacer(RiffleCommon.COLLAPSED_QUESTION_PANEL_HEIGHT)

  elementsColumn:alignChildren()

  local self = createLayoutFromColumn(elementsColumn)

  self.articleLabel = body
  self.nothingTouchedRegion = createNothingTouchedRegion(self, body)

  return self
end

local function chooseLayoutConstructor(round)
  if (round.layoutStyle == RiffleCommon.LayoutStyle.IMAGE_AND_SUBHEADER) then
    return createLayoutImageAndSubheader
  elseif (round.layoutStyle == RiffleCommon.LayoutStyle.SUBHEADER_ONLY) then
    return createLayoutNoImage
  elseif (round.layoutStyle == RiffleCommon.LayoutStyle.IMAGE_ONLY) then
    return createLayoutNoSubheader
  else
    return createLayoutNoOrnaments
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleArticle:setBoldingEnabled(enabled)
  self.articleLabel:setBoldingEnabled(enabled)
end

function RiffleArticle:addTouchToViewsByString(text, viewTappedCallback)
  return self.articleLabel:addTouchToViewsByString(text, viewTappedCallback)
end

function RiffleArticle:clearTouchActiveViews()
  self.articleLabel:clearTouchActiveViews()
end

function RiffleArticle:getNothingTouchedRegion()
  return self.nothingTouchedRegion
end

function RiffleArticle:prepareForCorrectAnimation(responseString)
  self.articleLabel:createHiddenHighlightsForText(responseString)
end

function RiffleArticle:createCorrectAnimation(responseString)
  return self.articleLabel:createHighlightTextAnimation()
end

function RiffleArticle:prepareForIncorrectAnimation(expectedString)
  self.articleLabel:createHiddenHighlightsForText(expectedString)
end

function RiffleArticle:createIncorrectAnimation(expectedString)
  self.articleLabel:createHiddenHighlightsForText(expectedString)
  return self.articleLabel:createHighlightTextAnimation()
end

function RiffleArticle:getTouchViews()
  return self.touchActiveViews
end

function RiffleArticle:clearHighlights()
  local clearAction = AHDelayed.new(1.5, function() self.articleLabel:removeHighlights() end)

  return clearAction:start()
end

function RiffleArticle:getScrollLocationY(text)
  local textY = self.articleLabel:getYBottom() + self.articleLabel:getFirstViewLocationY(text)
  local height = self:getHeight()
  local halfScreenOffset = Screen.getHeightPart(0.5)
  local scrollY = height - textY - halfScreenOffset

  return scrollY
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleArticle.new(round)
  local selectedLayoutConstructor = chooseLayoutConstructor(round)
  local self = selectedLayoutConstructor(round)
  Mixin.mix(self, RiffleArticle)

  return self
end
