--- RiffleArticleTextLabel
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleArticleTextLabel")

require("RiffleCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ATTRIBUTES_FOR_LABELS = {
  i = {font = nil},
  strong = {strong = true}
}

local LIGHT_OPACITY = 0.5
local BOLDING_ATTRIBUTE_TAG = "strong"
local TAP_AREA_PADDING_X = 40
local TAP_AREA_PADDING_Y = 40
local CORRECT_COLOR = RiffleCommon.COLORS.AQUA
local BACKING_WIDTH_PADDING = 10
local BACKING_HEIGHT_PROPORTION = 0.8
local HIGHLIGHT_EASE = AHEaseBezier.newFlat({0.4, 0, 0.6, 1})

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function removeNonAlphaNumeric(text)
  text = text:gsub("%W", "")

  return text
end

local function setItalicFontSize(fontSize)
  ATTRIBUTES_FOR_LABELS.i.font = Font.new(FontFaces.TISA_ITALIC, fontSize)
end

local function itemForView(self, view)
  return self.itemsForViews[view]
end

local function getLongestViewChain(self, startIndex, text, itemText, view)
  local itemTextChain = itemText
  local viewsChain = {view}
  local viewsCount = #self.itemViews

  for viewIndex = startIndex + 1, viewsCount do
    local view = self.itemViews[viewIndex]
    local item = self.itemsForViews[view]
    local itemText = item.attributes.text

    if itemText then
      local strippedItemText = removeNonAlphaNumeric(itemText)
      local candidateChain = itemTextChain .. strippedItemText
      local shouldSearchPlainText = true
      if text:find(candidateChain, 1, shouldSearchPlainText) then
        itemTextChain = candidateChain
        table.insert(viewsChain, view)
      else
        break
      end
    end
  end

  return viewsChain, itemTextChain
end

local function isMatch(searchText, resultText)
  return (#searchText == #resultText)
end

local function getViewsForString(self, text)
  local longestFoundChain = {}

  local strippedText = removeNonAlphaNumeric(text)
  for itemViewIndex, itemView in ipairs(self.itemViews) do
    local item = self.itemsForViews[itemView]
    local itemText = item.attributes.text

    if itemText then
      local strippedItemText = removeNonAlphaNumeric(itemText)
      local isEmptyString = (#strippedItemText == 0)

      if not isEmptyString then
        local shouldSearchPlainText = true
        local foundIndex = strippedText:find(strippedItemText, 1, shouldSearchPlainText)
        if foundIndex then
          local viewChain, itemTextChain = getLongestViewChain(self, itemViewIndex, strippedText, strippedItemText, itemView)

          if (#viewChain > #longestFoundChain) then
            longestFoundChain = viewChain
          end

          if isMatch(strippedText, itemTextChain) then
            return viewChain
          end
        end
      end
    end
  end

 -- fallback if isMatch never finds a match. Just to be safe.
 -- May happen in exceptional cases with weird punctuation or accented characters
  return longestFoundChain
end

local function createHiddenHighlightForView(self, view)
  local width = view:getWidth() + BACKING_WIDTH_PADDING
  local height = view:getHeight() * BACKING_HEIGHT_PROPORTION
  local backingRectangle = FilledRectangle.new(width, height, CORRECT_COLOR)

  local viewX, viewY = view:getLoc()
  local viewWidth, viewHeight = view:getDims()
  backingRectangle:setLoc(viewX + viewWidth * 0.5, viewY + viewHeight * 0.5)
  backingRectangle:setOpacity(0)

  local zBehindParent = -1

  table.insert(self.highlights, backingRectangle)
  view.parent:insert(backingRectangle, zBehindParent)
end

local function createHighlightAction(view)
  local duration = 0.2
  local targetOpacity = 1

  local highlightAction = AHOpacityTo.new(view, duration, HIGHLIGHT_EASE, targetOpacity)

  return highlightAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleArticleTextLabel:setBoldingEnabled(enabled)
  for _, view in ipairs(self.itemViews) do
    local item = self.itemsForViews[view]
    local opacity = RiffleCommon.VISIBLE_OPACITY

    local isLight = (enabled and (not item.attributes[BOLDING_ATTRIBUTE_TAG]))

    if isLight then
      opacity = LIGHT_OPACITY
    end
    view:setOpacity(opacity)
  end
end

function RiffleArticleTextLabel:createHiddenHighlightsForText(text)
  local strippedText = StringUtils.getAfterStrippingTags(text)
  local views = getViewsForString(self, strippedText)

  for _, view in ipairs(views) do
    createHiddenHighlightForView(self, view)
  end
end

function RiffleArticleTextLabel:createHighlightTextAnimation()
  local combinedColorAction = AHSpawn.new()
  for _, view in ipairs(self.highlights) do
    local viewAction = createHighlightAction(view)

    combinedColorAction:addAction(viewAction)
  end

  return combinedColorAction
end

function RiffleArticleTextLabel:addTouchToViewsByString(text, viewTappedCallback)
  local strippedText = StringUtils.getAfterStrippingTags(text)
  local views = getViewsForString(self, strippedText)
  local touchAreas = {}

  for _, view in ipairs(views) do
    local width = view:getWidth() + TAP_AREA_PADDING_X
    local height = view:getHeight() + TAP_AREA_PADDING_Y
    local touchRect = Group.new(width, height)
    view:insertAtCenter(touchRect)

    local touchEndedWrapper = function(event)
      if viewTappedCallback then
        viewTappedCallback(event.x, event.y)
      end
    end

    local touchHandler = Button.createButtonTouchHandler(nil, nil, touchEndedWrapper)
    TouchManager.addTouchEnabledMethods(touchRect, touchHandler)

    table.insert(touchAreas, touchRect)
    touchRect:setTouchEnabled(true)
  end

  self.touchActiveViews = views
  self.touchAreas = touchAreas

  return touchAreas
end

function RiffleArticleTextLabel:getTouchViews()
  return self.touchActiveViews
end

function RiffleArticleTextLabel:disableTouchActiveViews()
  for _, view in ipairs(self.touchActiveViews) do
    if view.setTouchEnabled then
      view:setTouchEnabled(false)
    end
  end
end

function RiffleArticleTextLabel:clearTouchActiveViews()
  self:disableTouchActiveViews()

  for _, touchArea in ipairs(self.touchAreas) do
    touchArea:removeSelf()
  end
end

function RiffleArticleTextLabel:removeHighlights()
  for _, highlight in ipairs(self.highlights) do
    highlight:removeSelf()
  end

  self.highlights = {}
end

function RiffleArticleTextLabel:getFirstViewLocationY(text)
  local strippedText = StringUtils.getAfterStrippingTags(text)
  local views = getViewsForString(self, strippedText)
  local firstView = views[1]

  assert(firstView ~= nil, "No views were returned for the string: " .. text)

  return firstView:getY()
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleArticleTextLabel.new(text, width, fontFace, fontSize)
  local basicAttributes = { font = Font.new(fontFace, fontSize)}
  setItalicFontSize(fontSize)

  local self = DynamicTextArea.new(text, width, basicAttributes, ATTRIBUTES_FOR_LABELS)
  Mixin.mix(self, RiffleArticleTextLabel)

  self.highlights = {}

  return self
end
