--- RiffleConfiguration
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleConfiguration")

require("data.keyframes.KeyframeSet")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INITIAL_LIVES_COUNT_THRESHOLDS = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {lives = 4}),
  Keyframe.new(0.3, {lives = 3})
)
local SKIMMING_DURATION_KEYFRAMES = KeyframeSet.newInterpolated(
  Keyframe.new(0.0, {skimDuration = 60}),
  Keyframe.new(0.25, {skimDuration = 60}),
  Keyframe.new(1.0, {skimDuration = 15})
)
local SCANNING_DURATION_KEYFRAMES = KeyframeSet.newInterpolated(
  Keyframe.new(0.0, {scanDuration = 17}),
  Keyframe.new(0.5, {scanDuration = 13}),
  Keyframe.new(1.0, {scanDuration = 11})
)

local SKIMMING_QUESTIONS_COUNT = 2
local SKIM_QUESTION_DIFFICULTY_KEYFRAMES = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {easyCount = 2, hardCount = 0}),
  Keyframe.new(0.25, {easyCount = 2, hardCount = 0}),
  Keyframe.new(0.375, {easyCount = 1, hardCount = 1}),
  Keyframe.new(0.5, {easyCount = 0, hardCount = 2})
)

local SCANNING_QUESTIONS_COUNT = 3
local SCAN_QUESTION_DIFFICULTY_KEYFRAMES = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {easyCount = 3, hardCount = 0}),
  Keyframe.new(0.25, {easyCount = 2, hardCount = 1}),
  Keyframe.new(0.5, {easyCount = 1, hardCount = 2}),
  Keyframe.new(0.75, {easyCount = 0, hardCount = 3})
)

local DURATION_INTERVAL = 5
local BEGINNER_THRESHOLD = 0.3
local EASY_FILTER_STRING = "Easy"
local HARD_FILTER_STRING = "Hard"
local SKIM_REFIL_TIME = 15
local SKIMMING_FREE_POINTS = 3000
local SKIMMING_EARLY_COMPLETION_BONUS = 1000
local CORRECT_ANSWER_POINTS = 500
local MISTAKES_ALLOWED_RANK_TWO = 1
local MISTAKES_ALLOWED_RANK_THREE = 0
local MAX_EXTRA_LIVES_POINTS = 3600
local POINTS_PER_FULL_SPEED_BONUS = 300
local SPEED_BONUS_WINDOW_IN_SECONDS = 6
local SPEED_BONUS_DECAY_START = 0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function insertNCopies(t, value, copiesCount)
  for i = 1, copiesCount do
    table.insert(t, value)
  end
end

local function getDifficultyNamesArray(keyframeSet)
  local difficulty = GameSession.getDifficulty()
  local frameForDifficulty = keyframeSet:getFrameAt(difficulty)
  local easyConceptCount = frameForDifficulty.easyCount
  local hardConceptCount = frameForDifficulty.hardCount

  local difficultyNames = {}
  insertNCopies(difficultyNames, EASY_FILTER_STRING, easyConceptCount)
  insertNCopies(difficultyNames, HARD_FILTER_STRING, hardConceptCount)

  return difficultyNames
end

local function getMaxNumberOfExtraLives()
  return RiffleConfiguration.getInitialLivesCount() - 1
end

local function getPassScore()
  return SKIMMING_FREE_POINTS + CORRECT_ANSWER_POINTS * (SKIMMING_QUESTIONS_COUNT + SCANNING_QUESTIONS_COUNT)
end

local function getMinimumPassScore()
  local maxPassScore = getPassScore()
  local possibleLostScore = getMaxNumberOfExtraLives() * CORRECT_ANSWER_POINTS

  local minimumPassScore = maxPassScore - possibleLostScore
  return minimumPassScore
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleConfiguration.getScanningDuration()
  local difficulty = GameSession.getDifficulty()
  local frameForDifficulty = SCANNING_DURATION_KEYFRAMES:getFrameAt(difficulty)
  local interpolatedDuration = frameForDifficulty.scanDuration
  local roundedDuration = MathUtils.round(interpolatedDuration)

  return roundedDuration
end

function RiffleConfiguration.getCorrectAnswerPoints()
  return CORRECT_ANSWER_POINTS
end

function RiffleConfiguration.getSkimmingFreePoints()
  return SKIMMING_FREE_POINTS
end

function RiffleConfiguration.getSkimmingBonusPoints(timeRemaining, totalDuration)
  return SKIMMING_EARLY_COMPLETION_BONUS * (timeRemaining / totalDuration)
end

function RiffleConfiguration.getSkimmingQuestionCount()
  return SKIMMING_QUESTIONS_COUNT
end

function RiffleConfiguration.getScanningQuestionCount()
  return SCANNING_QUESTIONS_COUNT
end

function RiffleConfiguration.isBeginner()
  return (GameSession.getDifficulty() <= BEGINNER_THRESHOLD)
end

function RiffleConfiguration.getInitialLivesCount()
  local difficulty = GameSession.getDifficulty()
  local frameForDifficulty = INITIAL_LIVES_COUNT_THRESHOLDS:getFrameAt(difficulty)
  return frameForDifficulty.lives
end

function RiffleConfiguration.getSkimRefilTime()
  return SKIM_REFIL_TIME
end

function RiffleConfiguration.getExtraLivesPoints()
  return (MAX_EXTRA_LIVES_POINTS / getMaxNumberOfExtraLives())
end

function RiffleConfiguration.getSkimmingRoundDuration()
  local difficulty = GameSession.getDifficulty()
  local frameForDifficulty = SKIMMING_DURATION_KEYFRAMES:getFrameAt(difficulty)
  local interpolatedDuration = frameForDifficulty.skimDuration
  local roundedDuration = MathUtils.roundToNearestMultiple(interpolatedDuration, DURATION_INTERVAL)

  return roundedDuration
end

function RiffleConfiguration.getRankTwoScore()
  local passScore = getMinimumPassScore()
  local extraLivesNeeded = getMaxNumberOfExtraLives() - MISTAKES_ALLOWED_RANK_TWO
  local extraPointsRankTwo = RiffleConfiguration.getExtraLivesPoints() * extraLivesNeeded
  local rankTwoScore = passScore + extraPointsRankTwo

  return rankTwoScore
end

function RiffleConfiguration.getRankThreeScore()
  local passScore = getMinimumPassScore()
  local extraLivesNeeded = getMaxNumberOfExtraLives() - MISTAKES_ALLOWED_RANK_THREE
  local extraPointsRankThree = RiffleConfiguration.getExtraLivesPoints() * extraLivesNeeded
  local rankThreeScore = passScore + extraPointsRankThree

  return rankThreeScore
end

function RiffleConfiguration.getSkimmingDifficulties()
  local difficultyNames = getDifficultyNamesArray(SKIM_QUESTION_DIFFICULTY_KEYFRAMES)

  return difficultyNames
end

function RiffleConfiguration.getScanningDifficulties()
  local difficultyNames = getDifficultyNamesArray(SCAN_QUESTION_DIFFICULTY_KEYFRAMES)

  return difficultyNames
end

function RiffleConfiguration.getPointsPerFullSpeedBonus()
  return POINTS_PER_FULL_SPEED_BONUS
end

function RiffleConfiguration.getSpeedBonusWindowInSeconds()
  return SPEED_BONUS_WINDOW_IN_SECONDS
end

function RiffleConfiguration.getSpeedBonusDecayStart()
  return SPEED_BONUS_DECAY_START
end
