--- RiffleGame
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleGame")

require("StandardUIBuilder")
require("RiffleConfiguration")
require("RiffleCommon")
require("RiffleScene")
require("RifflePreRoll")
require("RiffleConceptGenerator")
require("RiffleTimer")
require("RiffleOutro")
require("RiffleOvertimePenaltyPopup")
require("bonuses.ResponseSpeedTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local SLIDE_SCENE_DURATION = 0.7
local UI_BAR_ANIMATE_DELAY_DURATION = 0.7
local SKIMMING_CORRECT_DELAY = 0.6
local SKIMMING_INCORRECT_DELAY = 3.6
local TIMER_TEXT_SKIMMING = "SKIMMING"
local TIMER_TEXT_SCANNING = "SCANNING"
local TIMER_TEXT_WAITING = "WAITING"

local function configureGameDifficulty(self)
  local rankTwoScore = RiffleConfiguration.getRankTwoScore()
  local rankThreeScore = RiffleConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  self.initialLivesCount = RiffleConfiguration.getInitialLivesCount()
  self.skimmingRoundDuration = RiffleConfiguration.getSkimmingRoundDuration()
  GameSession.setInitialLives(self.initialLivesCount)

  local conceptChooser = RiffleConceptGenerator.new()
  self.round = conceptChooser:next()

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function showAndWaitForPreRoll(self)
  local waitLock = AHDelay.new(math.huge)

  local onPreRollDismiss = function()
    waitLock:stop()
  end

  self.preRoll:setDismissCallback(onPreRollDismiss)
  self:insertAtCenter(self.preRoll)

  wait(waitLock:start())
end

local function showScene(self)
  local sceneHeight = self.scene:getHeight()
  self:insertAtCenter(self.scene)
  self.scene:bringToBack()

  local ease = AHEaseExponentialIn.new()
  local slidePreRoll = AHLocationYBy.new(self.preRoll, SLIDE_SCENE_DURATION, ease, -sceneHeight)
  slidePreRoll:start()

  local delayedBarAnimateIn = function()
    waitFor(UI_BAR_ANIMATE_DELAY_DURATION)
    self.uiBar:animateIn()
    self.pauseButton:setTouchEnabled(true)
  end
  runInCoroutine(delayedBarAnimateIn)

  if RiffleConfiguration.isBeginner() then
    self.scene:setBoldingEnabled(true)
  end

  GameSession.contentTracker:notifyPresentedConcept(self.round.identifier)

  wait(self.scene:animateIn())
  self.scene:setScrollEnabled(true)
end

local function checkForLossState(self)
  local didLose = (self.livesCounter:getLives() == 0)
  if didLose then
    GameSession.notifyGameLoss()
  end

  return didLose
end

local function registerCorrectResponse(self)
  self.responseSpeedTracker:notifyCorrectResponse()

  local scoreDelta = RiffleConfiguration.getCorrectAnswerPoints()
  GameSession.increaseScore(scoreDelta)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function registerIncorrectResponse(self)
  self.responseSpeedTracker:notifyIncorrectResponse()

  self.livesCounter:decrement()
end

local function registerSkimmingTimeOut(self, waitLock)
  self.scene:setScrollEnabled(false)

  local uiBarHeight = self.uiBar:getHeight()
  local canKeepSkimming = (self.livesCounter:getLives() > 1)
  local overtimePopup = RiffleOvertimePenaltyPopup.new(uiBarHeight, canKeepSkimming)
  self:insertAtCenter(overtimePopup)

  local registerContinueTapped = function()
    registerIncorrectResponse(self)
    Sound.playSoundEffect("skimming_incorrect")
    overtimePopup:animateOutAndRemove()
    self.timer:setDuration(RiffleConfiguration.getSkimRefilTime())
    self.timer:start()
    self.scene:setScrollEnabled(true)
    self.uiBar:setDisplayForTimeOut(false)
  end
  overtimePopup:setContinueTappedCallback(registerContinueTapped)

  local registerStopTapped = function()
    waitLock:stop()
    overtimePopup:animateOutAndRemove()
    self.uiBar:setDisplayForTimeOut(false)
  end
  overtimePopup:setStopTappedCallback(registerStopTapped)

  overtimePopup:prepareForAnimateIn()
  overtimePopup:animateIn()

  self.uiBar:setDisplayForTimeOut(true)
end

local function waitForFinishReading(self)
  local waitLock = AHDelay.new(math.huge)

  self.timer:start()
  self.timer:setTimeoutHandler(function() registerSkimmingTimeOut(self, waitLock) end)

  local onFinishedReading = function()
    waitLock:stop()
    self.timer:stop()
  end
  self.scene:setFinishedReadingCallback(onFinishedReading)

  wait(waitLock:start())

  local freePoints = RiffleConfiguration.getSkimmingFreePoints()
  local bonusPoints = RiffleConfiguration.getSkimmingBonusPoints(self.timer:getTimeRemaining(), self.timer:getDuration())
  GameSession.increaseScore(freePoints + bonusPoints)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function beginSkimmingQuestions(self)
  self.scene:setScrollEnabled(false)
  self.scene:prepForSkimmingQuestions()
  self.timer:setLabelText(TIMER_TEXT_WAITING)
  self.timer:setDuration(0)
end

local function insertIntoTableIfValid(t, option)
  if option and (#option > 0) then
    table.insert(t, option)
  end
end

local function presentSkimmingQuestion(self, question, questionIndex)
  self.responseSpeedTracker:notifyStart()

  local prompt = question.prompt
  local options = {}

  local isTrueFalse = (question.answer:lower() == "true") or (question.distractor_1:lower() == "true")

  if isTrueFalse then
    -- Keep true on top. Always.
    if question.answer:lower() == "true" then
      options = {question.answer, question.distractor_1}
    else
      options = {question.distractor_1, question.answer}
    end
  else
    insertIntoTableIfValid(options, question.answer)
    insertIntoTableIfValid(options, question.distractor_1)
    insertIntoTableIfValid(options, question.distractor_2)

    TableUtils.shuffleArray(options)
  end

  self.skimQuestionIdentifier = question.question_identifier
  self.correctAnswer = question.answer
  self.scene:presentMultipleChoiceQuestion(prompt, options, questionIndex)
end

local function waitForSkimmingResponse(self, skimmingQuestion)
  local waitLock = AHDelay.new(math.huge)
  local userResponse = nil

  local responseCallback = function(response)
    userResponse = response
    waitLock:stop()
  end

  self.scene:setMultipleChoiceResponseCallback(responseCallback)

  wait(waitLock:start())

  return userResponse
end

local function registerCorrectSkimmingResponse(self, response)
  Sound.playSoundEffect("skimming_correct")
  self.scene:animateSkimmingResponseCorrect(response)
  registerCorrectResponse(self)
  waitFor(SKIMMING_CORRECT_DELAY)
end

local function registerIncorrectSkimmingResponse(self, response, correctAnswer)
  Sound.playSoundEffect("skimming_incorrect")
  registerIncorrectResponse(self)
  wait(self.scene:animateSkimmingResponseIncorrect(response, correctAnswer))
  waitFor(SKIMMING_INCORRECT_DELAY)
end

local function registerSkimmingResponse(self, response)
  local isCorrect = (response == self.correctAnswer)
  local conceptId = self.round.identifier
  local questionId = self.skimQuestionIdentifier

  if isCorrect then
    GameSession.contentTracker:notifyCorrectAnswer(conceptId, questionId)
    GameSession.incrementCorrectAnswerCount()
    registerCorrectSkimmingResponse(self, response)
  else
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, questionId, response)
    GameSession.incrementIncorrectAnswerCount()
    registerIncorrectSkimmingResponse(self, response, self.correctAnswer)
  end
end

local function endSkimmingQuestion(self)
  self.scene:endSkimmingQuestion()
end

local function presentScanningQuestion(self, question, beginScanningCallback, questionIndex, numberIncorrect)
  self.scene:presentScanningQuestion(question, beginScanningCallback, questionIndex, numberIncorrect)
end

local function registerCorrectScanningResponse(self)
  Sound.playSoundEffect("skimming_correct")
  registerCorrectResponse(self)
end

local function registerIncorrectScanningResponse(self)
  Sound.playSoundEffect("skimming_incorrect")
  registerIncorrectResponse(self)
end

local function waitForScanningQuestionResponse(self, question)
  local waitLock = AHDelay.new(math.huge)
  local tapX, tapY = nil, nil
  local isCorrectAnswer = nil
  local conceptId = self.round.identifier
  local questionId = question.question_identifier

  local expectedAnswer = question.answer
  local onCorrectCallback = function(x, y)
    GameSession.contentTracker:notifyCorrectAnswer(conceptId, questionId)
    GameSession.incrementCorrectAnswerCount()
    isCorrectAnswer = true
    tapX, tapY = x, y
    waitLock:stop()
  end

  local onIncorrectCallback = function(x, y)
    GameSession.contentTracker:notifyIncorrectAnswer(conceptId, questionId)
    GameSession.incrementIncorrectAnswerCount()
    isCorrectAnswer = false
    tapX, tapY = x, y
    waitLock:stop()
  end

  local onScanningTimeout = function()
    GameSession.contentTracker:notifyTimeout(conceptId, questionId)
    isCorrectAnswer = false
    tapX = Screen.getWidthPart(0.5)
    tapY = Screen.getHeightPart(0.5)
    waitLock:stop()
  end

  wait(self.scene:displayForScanningQuestion(expectedAnswer, onCorrectCallback, onIncorrectCallback))

  self.scene:setScrollEnabled(true)

  self.timer:setLabelText(TIMER_TEXT_SCANNING)
  self.timer:setDuration(RiffleConfiguration.getScanningDuration())
  self.timer:start()
  self.timer:setTimeoutHandler(onScanningTimeout)

  wait(waitLock:start())

  self.scene:setScrollEnabled(false)
  self.timer:stop()
  self.timer:setDuration(0)
  self.timer:setLabelText(TIMER_TEXT_WAITING)

  if isCorrectAnswer then
    registerCorrectScanningResponse(self)
    wait(self.scene:doCorrectAnimation(tapX, tapY, expectedAnswer))
  else
    registerIncorrectScanningResponse(self)
    wait(self.scene:doIncorrectAnimation(tapX, tapY, expectedAnswer))
  end
end

local function endScanningQuestion(self, questionIndex)
  local isLastQuestion = (questionIndex == RiffleConfiguration.getScanningQuestionCount())

  self.scene:clearScanningRound(isLastQuestion)
end

local function doSkimmingPhase(self)
  local skimmingQuestions = self.round.skimmingQuestions

  waitForFinishReading(self)
  beginSkimmingQuestions(self)

  for i, skimmingQuestion in ipairs(skimmingQuestions) do
    presentSkimmingQuestion(self, skimmingQuestion, i)
    local response = waitForSkimmingResponse(self, skimmingQuestion)
    registerSkimmingResponse(self, response)
    endSkimmingQuestion(self)
    checkForLossState(self)
  end
end

local function doScanningPhase(self)
  self.scene:setBoldingEnabled(false)
  self.scene:removeDoneButton()

  local scanningQuestions = self.round.scanningQuestions

  local numberIncorrect = RiffleConfiguration.getInitialLivesCount() - self.livesCounter:getLives()

  for i, scanningQuestion in ipairs(scanningQuestions) do
    local waitLock = AHDelay.new(math.huge)
    local onBeginScanning = function()
      self.responseSpeedTracker:notifyStart()
      Sound.playSoundEffect("skimming_button")
      waitLock:stop()
    end

    presentScanningQuestion(self, scanningQuestion, onBeginScanning, i, numberIncorrect)

    wait(waitLock:start())

    waitForScanningQuestionResponse(self, scanningQuestion)
    endScanningQuestion(self, i)
    checkForLossState(self)
  end
end

local function registerGameWin(self)
  ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), RiffleConfiguration.getExtraLivesPoints(), self.bonusTray)
  self.scoreLabel:setScore(GameSession.getScore())

  local scoreUpdateDelay = 1.25
  waitFor(scoreUpdateDelay)

  local outro = RiffleOutro.new()
  self:insertAtCenter(outro)

  local outroAction = outro:getAnimation()
  wait(outroAction:start())

  GameSession.notifyGameWin()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleGame:gameFlow()
  showAndWaitForPreRoll(self)
  showScene(self)
  doSkimmingPhase(self)
  doScanningPhase(self)
  registerGameWin(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local scene = RiffleScene.new(self.round)
  self.scene = scene

  local isFirstPlay = GameSession.hasNeverWon()
  local durationToDisplay = self.skimmingRoundDuration
  local preRoll = RifflePreRoll.new(isFirstPlay, durationToDisplay)
  self.preRoll = preRoll

  local quickResponseWrapper = function(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 RiffleConfiguration.getPointsPerFullSpeedBonus(),
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(RiffleConfiguration.getSpeedBonusWindowInSeconds(),
                                                       quickResponseWrapper,
                                                       RiffleConfiguration.getSpeedBonusDecayStart())

end

local function createUI(self)
  local uiColor = RiffleCommon.COLORS.WHITE

  local scoreLabel = ScoreLabel.new(StandardUIBuilder.STANDARD_FONT, StandardUIBuilder.STANDARD_FONT_SIZE, uiColor)

  local pauseButton = PauseButton.new(uiColor)
  local bonusTray = BonusNotificationTray.new(uiColor) -- May need special treatment
  local livesCounter = LivesDisplay.new(self.initialLivesCount, uiColor)
  local timer = RiffleTimer.new(self.skimmingRoundDuration, function() end)
  timer:setLabelText(TIMER_TEXT_SKIMMING)

  local layoutStyle = self.round.layoutStyle
  local uiBar = RiffleUIBar.new(pauseButton, scoreLabel, bonusTray, livesCounter, timer, layoutStyle)
  uiBar:setLoc(self:getWidthPart(0.5), self:getHeight() - uiBar:getHeightPart(0.5))
  self:insert(uiBar)

  uiBar:hide()

  self.pauseButton = pauseButton
  self.scoreLabel = scoreLabel
  self.bonusTray = bonusTray
  self.livesCounter = livesCounter
  self.timer = timer
  self.uiBar = uiBar
end

-- Constructor ------------------------

function RiffleGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, RiffleGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
