--- RifflePreRoll
-- @author Created by Jay Austria on February 23, 2015.
--
-- RifflePreRoll is a pre roll screen that appears before the user starts the skimming round
-- The screen also functions as a tips screen through a more info button
-- In a first-time play, the pre roll screen defaults to starting at the tips screen
--

class("RifflePreRoll")

require("utils.StateMachine")
require("RiffleButton")
require("RiffleCommon")
require("RifflePerCharacterAnimatingLabel")
require("RiffleTipLabel")
require("text.SizeConstrainedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SCREEN_CENTER_X = Screen.getWidthPart(0.5)
local BACKGROUND_COLOR = RiffleCommon.COLORS.BLACK
local TRANSITION_DELAY = 0.6
local BUTTON_LOCATION_Y_PREP = Screen.getHeightPart(0.2)
local TIP_PAGE_BUTTON_TEXT = "Okay"
local PREP_PAGE_BUTTON_TEXT = "Begin"

local TIP_PAGE_HEADER_TEXT = "BEFORE WE BEGIN"
local TIP_PAGE_HEADER_COLOR = RiffleCommon.COLORS.ORANGE

local TIP_PAGE_TEXT_AREA_WIDTH = Screen.getWidthPart(0.85)

local LABEL_ANIMATE_DY = Screen.getHeightPart(1.0)

local TIP_PAGE_COPY = [[This game is split into two portions: skimming and scanning.<br/>
<br/>
Improving these skills allows you to evaluate and consume information more effectively in real-world situations.<br/>
<br/>
<br/>
<h1>SKIMMING</h1><br/>
Quickly look across the text and identify important words or ideas. It's fine to skip lines if they don't seem important.<br/>
<br/>
<br/>
<h2>SCANNING</h2><br/>
Keep the information you're looking for in the front of your mind and move rapidly through the text to identify where it's likely to occur.]]

local PREP_HEADER_TEXT = "SKIM THE ARTICLE"
local PREP_HEADER_COLOR = RiffleCommon.COLORS.ORANGE

local PREP_HEADER_MIN_MARGIN = 50
local PREP_HEADER_MARGIN_BASED_ON_SAFE_AREA = Host.safeAreaTopInset + 15
local PREP_HEADER_MARGIN = math.max(PREP_HEADER_MIN_MARGIN, PREP_HEADER_MARGIN_BASED_ON_SAFE_AREA)
local PREP_HEADER_LOCATION_Y = Screen.height - PREP_HEADER_MARGIN
local PAGE_COLUMN_LOCATION_Y = PREP_HEADER_LOCATION_Y - Screen.getHeightPart(0.05)
local PAGE_COLUMN_SPACING_Y = Screen.getHeightPart(0.04)

local PREP_ROUND_TIME_LOCATION_Y = Screen.getHeightPart(0.6)
local PREP_ROUND_TIME_FONT = FontFaces.DIN_OT_BOLD
local PREP_ROUND_TIME_FONT_SIZE = 200
local PREP_ROUND_TIME_COLOR = RiffleCommon.COLORS.WHITE

local PREP_SECONDS_OFFSET_Y_PROPORTION = 0.03
local PREP_SECONDS_FONT = FontFaces.DIN_OT_BOLD
local PREP_SECONDS_FONT_SIZE = 48
local PREP_SECONDS_TEXT = "SECONDS"
local PREP_SECONDS_COLOR = RiffleCommon.COLORS.WHITE

local MORE_INFO_BUTTON_TEXT = "more info"
local MORE_INFO_Y_OFFSET = -5
local MORE_INFO_BUTTON_COLOR = RiffleCommon.COLORS.GREY
local MORE_INFO_BUTTON_UNDERLINE_COLOR = RiffleCommon.COLORS.ORANGE

local PageState =
{
  TIP_PAGE = "PageState.TIP_PAGE",
  PREP_PAGE = "PageState.PREP_PAGE",
  DISMISSED = "PageState.DISMISSED"
}

local FADE_IN_DURATION = 0.2
local FADE_OUT_DURATION = 0.5
local UNDERLINE_THICKNESS = 2
local UNDERLINE_OFFSET_Y = -2
local BUTTON_PADDING = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createMoreInfoButton(self)
  local label = Label.new(MORE_INFO_BUTTON_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, 15))
  label:setColor(unpack(MORE_INFO_BUTTON_COLOR))

  local buttonWidth = label:getWidth() + BUTTON_PADDING
  local buttonHeight = label:getHeight() + BUTTON_PADDING
  local button = Group.new(buttonWidth, buttonHeight)

  button:insertAtCenter(label)

  local underline = FilledRectangle.new(label:getWidth(), UNDERLINE_THICKNESS, RiffleCommon.COLORS.ORANGE)
  underline:setLoc(button:getWidthPart(0.5), button:getHeightPart(0.5) - label:getHeightPart(0.5) + UNDERLINE_OFFSET_Y)
  button:insert(underline)

  local onDown = function()
    Sound.playSoundEffect("skimming_button")
    self.pageStateMachine:transition(PageState.TIP_PAGE)
  end

  Button.makeButton(button, onDown)

  return button
end

local function getFadeInGroupAction(group)
  local fadeIn = AHOpacityTo.new(group, FADE_IN_DURATION, nil, RiffleCommon.VISIBLE_OPACITY)
  return fadeIn
end

local function getFadeOutGroupAction(group)
  local fadeOut = AHOpacityTo.new(group, FADE_OUT_DURATION, nil, RiffleCommon.INVISIBLE_OPACITY)
  return fadeOut
end

local function processButtonTap(self)
  local isShowingTip = (self.pageStateMachine:getCurrentState() == PageState.TIP_PAGE)

  if isShowingTip then
    waitFor(TRANSITION_DELAY)
    self.pageStateMachine:transition(PageState.PREP_PAGE)
  else
    if self.dismissCallback then
      self.dismissCallback()
    end
    waitFor(TRANSITION_DELAY)
    self.isDismissed = true
    self.pageStateMachine:transition(PageState.DISMISSED)
  end
end

local function createButton(self, text)
  local onButtonTap = function(button)
    Sound.playSoundEffect("skimming_button")
    button:setTouchEnabled(false)
    button:doPressedAnimation()
    processButtonTap(self)
  end

  local button = RiffleButton.new(RiffleButton.BUTTON_TYPE.ORANGE, onButtonTap)
  button:setText(text)
  button:setTouchEnabled(false)
  return button
end

local function clearPageLayout(self)
  if self.pageLayout then
    local fadeLayout = getFadeOutGroupAction(self.pageLayout)

    local removeAction = AHRemove.new(self.pageLayout)

    local clearSequence = AHSequence.new(fadeLayout, removeAction)
    wait(clearSequence:start())
  end
end

local function showTipPageLayout(self)
  local pageLayout = Group.newFullScreen()
  self:insertAtCenter(pageLayout)

  local button = createButton(self, TIP_PAGE_BUTTON_TEXT)
  button:setAnchorPointBottom()
  local minimumY = Screen.safeBottom + button:getHeightPart(0.25)
  local buttonY = math.max(minimumY, button:getHeightPart(0.5))
  pageLayout:insertAtLocation(button, SCREEN_CENTER_X, buttonY)

  local header = Label.new(TIP_PAGE_HEADER_TEXT, Font.new(FontFaces.DIN_OT_BOLD, 18))
  header:setAnchorPointTop()
  header:setColor(unpack(TIP_PAGE_HEADER_COLOR))
  local maximumY = Screen.safeTop - header:getHeightPart(0.5)
  local headerY = math.min(maximumY, Screen.height - header:getHeightPart(1.5))
  pageLayout:insertAtLocation(header, SCREEN_CENTER_X, headerY)

  local marginY = header:getHeightPart(0.8)

  local basicAttributes = {
    font = Font.new(FontFaces.DIN_OT_MEDIUM, 18),
    color = RiffleCommon.COLORS.WHITE
  }

  local customAttributes = {
    h1 = {
      color = RiffleCommon.COLORS.ORANGE,
      yOffset = 8
    },
    h2 = {
      color = RiffleCommon.COLORS.AQUA,
      yOffset = 8
    },
  }

  local maxHeight = header:getYBottom() - button:getYTop() - (2 * marginY)
  local minimumFontSize = 14
  local contents = SizeConstrainedTextArea.new(TIP_PAGE_COPY,
                                               TIP_PAGE_TEXT_AREA_WIDTH,
                                               maxHeight,
                                               basicAttributes,
                                               customAttributes,
                                               minimumFontSize)
  contents:setAnchorPointTop()
  pageLayout:insertAtLocation(contents, SCREEN_CENTER_X, header:getYBottom() - marginY)

  local isRecursive = true
  pageLayout:linkOpacityAttrToChildren(isRecursive)

  pageLayout:setOpacity(RiffleCommon.INVISIBLE_OPACITY)
  local fadeLayout = getFadeInGroupAction(pageLayout)
  local fadeButton = getFadeInGroupAction(button)

  local fadeInSpawn = AHSpawn.new(fadeLayout, fadeButton)

  local enableButton = function() button:setTouchEnabled(true) end

  local fadeInSequence = AHSequence.new(fadeInSpawn, enableButton)
  fadeInSequence:start()

  self.pageLayout = pageLayout
end

local function removeTipPageLayout(self)
  clearPageLayout(self)
end

local function showPrepPageLayout(self)
  local pageLayout = Group.newFullScreen()
  self:insertAtCenter(pageLayout)

  local button = createButton(self, PREP_PAGE_BUTTON_TEXT)
  button:setLoc(SCREEN_CENTER_X, BUTTON_LOCATION_Y_PREP)
  pageLayout:insert(button)

  local headerLabel = Label.new(PREP_HEADER_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, 22))
  headerLabel:setLoc(SCREEN_CENTER_X, PREP_HEADER_LOCATION_Y)
  headerLabel:setColor(unpack(PREP_HEADER_COLOR))
  pageLayout:insert(headerLabel)

  local moreInfoButton = createMoreInfoButton(self)
  local moreInfoButtonY = headerLabel:getYBottom() + MORE_INFO_Y_OFFSET
  moreInfoButton:setLoc(SCREEN_CENTER_X, moreInfoButtonY)
  pageLayout:insert(moreInfoButton)
  moreInfoButton:setTouchEnabled(false)
  self.moreInfoButton = moreInfoButton

  local durationAsText = tostring(self.roundDuration)
  local roundTimeLabel = RifflePerCharacterAnimatingLabel.new(durationAsText, PREP_ROUND_TIME_FONT, PREP_ROUND_TIME_FONT_SIZE, PREP_ROUND_TIME_COLOR)
  roundTimeLabel:setLoc(SCREEN_CENTER_X, PREP_ROUND_TIME_LOCATION_Y)
  pageLayout:insert(roundTimeLabel)

  local secondsLabel = RifflePerCharacterAnimatingLabel.new(PREP_SECONDS_TEXT, PREP_SECONDS_FONT, PREP_SECONDS_FONT_SIZE, PREP_SECONDS_COLOR)
  local secondsLabelY = roundTimeLabel:getYBottom() + roundTimeLabel:getHeightPart(PREP_SECONDS_OFFSET_Y_PROPORTION)
  secondsLabel:setLoc(SCREEN_CENTER_X, secondsLabelY)
  pageLayout:insert(secondsLabel)

  local isBeginner = RiffleConfiguration.isBeginner()
  local tipLabel = RiffleTipLabel.new(isBeginner, RiffleTipLabel.TYPE.SKIM)
  local tipLabelY = (button:getY() - button:getHeightPart(0.5)) * 0.5
  tipLabel:setLoc(SCREEN_CENTER_X, tipLabelY)
  pageLayout:insert(tipLabel)

  local isRecursive = true
  pageLayout:linkOpacityAttrToChildren(isRecursive)
  roundTimeLabel:clearAttrLink(MOAIColor.ATTR_OPACITY)
  secondsLabel:clearAttrLink(MOAIColor.ATTR_OPACITY)

  pageLayout:setOpacity(RiffleCommon.INVISIBLE_OPACITY)

  local animateLabels = function()
    roundTimeLabel:animateIn(LABEL_ANIMATE_DY)
    secondsLabel:animateIn(LABEL_ANIMATE_DY)
  end

  local fadeLayout = getFadeInGroupAction(pageLayout)
  local fadeButton = getFadeInGroupAction(button)

  local fadeInSpawn = AHSpawn.new(fadeLayout, fadeButton)
  local delayedFadeIn = AHDelayed.new(0.6, fadeInSpawn)

  local onFadeFinished = function()
    button:setTouchEnabled(true)
    self.moreInfoButton:setTouchEnabled(true)

    local isRecursive = true
    pageLayout:linkOpacityAttrToChildren(isRecursive)
  end

  local fadeInSequence = AHSequence.new(animateLabels, delayedFadeIn, onFadeFinished)
  fadeInSequence:start()

  self.pageLayout = pageLayout
end

local function removePrepPageLayout(self)
  clearPageLayout(self)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------
function RifflePreRoll:setDismissCallback(callback)
  self.dismissCallback = callback
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createPageStateMachine(self)
  local pageStateMachine = StateMachine.new()

  -- Tip page -----------------------------------------------------------------
  local onEnterTipPage = function()
    showTipPageLayout(self)
  end

  local onLeaveTipPage = function()
    removeTipPageLayout(self)
  end

  local validStatesFromTipPage = { PageState.PREP_PAGE }
  pageStateMachine:addState(PageState.TIP_PAGE, onEnterTipPage, onLeaveTipPage, unpack(validStatesFromTipPage))

  -- Prep page ----------------------------------------------------------------
  local onEnterPrepPage = function()
    showPrepPageLayout(self)
  end

  local onLeavePrepPage = function()
    if self.isDismissed then
      self.moreInfoButton:setTouchEnabled(false)
    else
      removePrepPageLayout(self)
    end
  end

  local validStatesFromPrepPage = { PageState.TIP_PAGE, PageState.DISMISSED }
  pageStateMachine:addState(PageState.PREP_PAGE, onEnterPrepPage, onLeavePrepPage, unpack(validStatesFromPrepPage))

  -- Dismissed ----------------------------------------------------------------
  local onEnterDismissed = function()
  end

  local onLeaveDismissed = function()
  end

  local validStatesFromDismissed = {}
  pageStateMachine:addState(PageState.DISMISSED, onEnterDismissed, onLeaveDismissed, unpack(validStatesFromDismissed))

  self.pageStateMachine = pageStateMachine
end

function RifflePreRoll.new(isFirstPlay, roundDuration)
  local self = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  Mixin.mix(self, RifflePreRoll)

  self.roundDuration = roundDuration
  self.isDismissed = false

  createPageStateMachine(self)

  if isFirstPlay then
    self.pageStateMachine:transition(PageState.TIP_PAGE)
  else
    self.pageStateMachine:transition(PageState.PREP_PAGE)
  end

  return self
end
