--- RiffleQuestionLayer
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleQuestionLayer")

require("graphics.SimpleMorphingPolygon")
require("RiffleButton")
require("RiffleQuestionLayerScanningQuestionGroup")
require("RiffleQuestionLayerSkimmingQuestionGroup")
require("RiffleQuestionLayerInterstitialGroup")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local PANEL_HEIGHT = RiffleCommon.QUESTION_PANEL_HEIGHT
local PANEL_EXTENSION_HEIGHT = Screen.getHeightPart(0.15)
local PANEL_HEIGHT_MAX = PANEL_HEIGHT + PANEL_EXTENSION_HEIGHT
local LAYOUT_WIDTH = Screen.getWidthPart(0.85)
local LEFT_X = -Screen.getWidthPart(0.5)
local RIGHT_X = Screen.getWidthPart(0.5)
local BOTTOM_Y = PANEL_HEIGHT * -0.5
local TOP_Y_MIN = PANEL_HEIGHT * 0.5
local TOP_Y_MAX = PANEL_HEIGHT_MAX * 0.5

local START_VERTICIES =
{
  {x = LEFT_X, y = BOTTOM_Y},
  {x = RIGHT_X, y = BOTTOM_Y},
  {x = RIGHT_X, y = TOP_Y_MIN},
  {x = LEFT_X, y = TOP_Y_MIN}
}

local END_VERTICIES =
{
  {x = LEFT_X, y = BOTTOM_Y},
  {x = RIGHT_X, y = BOTTOM_Y},
  {x = RIGHT_X, y = TOP_Y_MAX},
  {x = LEFT_X, y = TOP_Y_MIN}
}
local PANEL_COLOR = RiffleCommon.COLORS.BLACK
local HIDE_DURATION = 1.5
local SHOW_DURATION = 1.5
local MORPH_AMOUNT_OUT = 0.0
local MORPH_AMOUNT_IN = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------
local function registerSkimmingResponse(self, responseText)
  if self.multipleChoiceResponseCallback then
    self.multipleChoiceResponseCallback(responseText)
  end
end

local function registerScanningBegin(self)
  if self.beginScanningCallback then
    self.layout:registerBeginScanningPressed()
    self.beginScanningCallback()
  end
end

local function createLayout(self)
  local panel = SimpleMorphingPolygon.new(Screen.width, PANEL_HEIGHT, START_VERTICIES, END_VERTICIES)
  panel:setAnchorPointBottom()
  panel:setColor(unpack(PANEL_COLOR))
  panel:setLoc(Screen.getWidthPart(0.5), 0)
  self:insert(panel)

  self.panel = panel
end

local function createAndPresentMultipleChoiceLayout(self, question, options, questionIndex)
  local width = LAYOUT_WIDTH
  local height = PANEL_HEIGHT

  local optionPressedCallback = function(option)
    registerSkimmingResponse(self, option)
  end

  local layout, buttons = RiffleQuestionLayerSkimmingQuestionGroup.new(width, height, question, options, optionPressedCallback, questionIndex)
  layout:hide()
  layout:setAnchorPointBottom()
  layout:setLoc(self:getWidthPart(0.5), 0)

  local fadeIn = layout:getFadeInAction()

  self:insert(layout)
  self.layout = layout
  self.buttons = buttons

  wait(fadeIn:start())
end

local function createAndPresentScanningInterstitialAndQuestion(self, question, questionIndex, numberIncorrect)
  local width = LAYOUT_WIDTH
  local height = PANEL_HEIGHT
  local layout = RiffleQuestionLayerInterstitialGroup.new(width, height, function() registerScanningBegin(self) end, question, questionIndex, numberIncorrect)
  layout:setLoc(Screen.getWidthPart(0.5), PANEL_HEIGHT - 40)

  self:insert(layout)
  self.layout = layout

  return layout:animateIn()
end

local function createAndPresentScanningQuestion(self, question, questionIndex)
  local width = LAYOUT_WIDTH
  local height = PANEL_HEIGHT
  local layout = RiffleQuestionLayerScanningQuestionGroup.new(width, height, function() registerScanningBegin(self) end, question, questionIndex)
  layout:setLoc(Screen.getWidthPart(0.5), PANEL_HEIGHT - 40)

  self:insert(layout)
  self.layout = layout

  return layout:animateIn()
end

local function getButtonWithText(self, response)
  for _, button in ipairs(self.buttons) do
    local isMatchingText = (button.label.text == response)
    if isMatchingText then
      return button
    end
  end

  error("Couldnt find a button with the following text: "..response)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleQuestionLayer:collapseForScanningRound()
  local collapseAction = AHSpawn.new(
    function() self.layout:collapseForScanningRound() end,
    self.panel:createMorphToAction(1.0, AHEaseExponentialInOut.new(), MORPH_AMOUNT_OUT),
    AHLocationYTo.new(self, 1.0, AHEaseExponentialInOut.new(), 0)
  )

  return collapseAction:start()
end

function RiffleQuestionLayer:setBeginScanningCallback(callback)
  self.beginScanningCallback = callback
end

function RiffleQuestionLayer:animateIn()
  local slideIn = AHLocationYTo.new(self, SHOW_DURATION, AHEaseExponentialInOut.new(), self.originPosition.y)
  local morphPanel = self.panel:createMorphToAction(SHOW_DURATION, AHEaseExponentialInOut.new(), MORPH_AMOUNT_IN)

  local showAction = AHSpawn.new(slideIn, morphPanel)
  showAction:start()
end

function RiffleQuestionLayer:animateOut()
  local hideAction = self.panel:createMorphToAction(HIDE_DURATION, AHEaseExponentialInOut.new(), MORPH_AMOUNT_OUT)

  hideAction:start()
end

function RiffleQuestionLayer:prepareForAnimateIn()
  self.panel:setMorphAmount(MORPH_AMOUNT_OUT)

  local currentX, currentY = self:getLoc()
  self.originPosition = {x = currentX, y = currentY}
  self:setLoc(currentX, currentY - self:getHeight())
end

function RiffleQuestionLayer:presentMultipleChoiceQuestion(question, options, questionIndex)
  createAndPresentMultipleChoiceLayout(self, question, options, questionIndex)
end

function RiffleQuestionLayer:setMultipleChoiceResponseCallback(callback)
  self.multipleChoiceResponseCallback = callback
end

function RiffleQuestionLayer:animateSkimmingResponseCorrect(response)
  local correctButton = getButtonWithText(self, response)
  correctButton:doPressedAnimation()
end

function RiffleQuestionLayer:animateSkimmingResponseIncorrect(response, correctAnswer)
  local pressedButton = getButtonWithText(self, response)
  pressedButton:doShakeAnimation()

  waitFor(1)

  local correctEffect = nil
  local correctButton = getButtonWithText(self, correctAnswer)

  for _, button in ipairs(self.buttons) do
    if (button ~= correctButton) then
      button:fadeOut()
    else
      correctEffect = button:showAsCorrect()
    end
  end

  wait(correctEffect)
end

function RiffleQuestionLayer:presentIntroIntoScanningQuestion(question, questionIndex, numberIncorrect)
  return createAndPresentScanningInterstitialAndQuestion(self, question, questionIndex, numberIncorrect)
end

function RiffleQuestionLayer:presentScanningQuestion(question, questionIndex)
  return createAndPresentScanningQuestion(self, question, questionIndex)
end

function RiffleQuestionLayer:clearLayout()
  if self.layout then
    local fadeAction = self.layout:getFadeOutAction()
    local removeAction = AHRemove.new(self.layout)

    local clearSequence = AHSequence.new(fadeAction, removeAction)
    wait(clearSequence:start())
  end
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleQuestionLayer.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, RiffleQuestionLayer)

  createLayout(self)

  return self
end
