--- RiffleScene
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleScene")

require("RiffleScrollLayer")
require("RiffleArticle")
require("RiffleUIBar")
require("RiffleQuestionLayer")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------


local function getPressEffect(self)
  local scrollLayer = self.scrollLayer
  local pressAnimation = AHSequence.new(
    AHDelay.new(0.2),
    AHScaleTo.new(scrollLayer, 0.2, AHEaseSimpleOut.new(), 0.95),
    AHScaleTo.new(scrollLayer, 0.5, AHEaseBackOut.new(), 1.0)
  )

  return pressAnimation
end

local function doIconStamp(self, iconFile, x, y)
  local icon = Sprite.newWithName(iconFile)
  icon:setLoc(x, y)
  self:insert(icon)

  icon:setScl(1.3)

  local stampAction = AHSequence.new(
    AHSpawn.new(
      AHDelayed.new(0.5, AHOpacityTo.new(icon, 0.5, nil, 0.0)),
      AHScaleTo.new(icon, 0.2, nil, 1.0)
    ),
    AHRemove.new(icon)
  )

  stampAction:start()
end

local function createLayout(self, round)
  local backing = FilledRectangle.newFullScreen(RiffleCommon.COLORS.LIGHT_GREY)
  self:insertAtCenter(backing)

  local article = RiffleArticle.new(round)
  self.article = article

  local scrollLayer = RiffleScrollLayer.new(article)
  self:insertAtCenter(scrollLayer)
  scrollLayer:prepareForAnimateIn()
  self.scrollLayer = scrollLayer

  local questionLayer = RiffleQuestionLayer.new()
  self:insertAtCenter(questionLayer)
  questionLayer:prepareForAnimateIn()
  self.questionLayer = questionLayer
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleScene:setScrollEnabled(enabled)
  self.scrollLayer:setTouchEnabled(enabled)
end

function RiffleScene:animateIn()
  return self.scrollLayer:animateIn()
end

function RiffleScene:setFinishedReadingCallback(callback)
  self.scrollLayer:setFinishedReadingCallback(callback)
end

function RiffleScene:prepForSkimmingQuestions()
  self.scrollLayer:scrollToTop()
  Sound.playSoundEffect("skimming_menuWoosh")
  self.questionLayer:animateIn()
end

function RiffleScene:setMultipleChoiceResponseCallback(callback)
  self.questionLayer:setMultipleChoiceResponseCallback(callback)
end

function RiffleScene:presentMultipleChoiceQuestion(question, options, questionIndex)
  self.questionLayer:presentMultipleChoiceQuestion(question, options, questionIndex)
end

function RiffleScene:animateSkimmingResponseCorrect(response)
  local correctAnimation = function()
    self.questionLayer:animateSkimmingResponseCorrect(response)
  end
  runInCoroutine(correctAnimation)
end

function RiffleScene:animateSkimmingResponseIncorrect(response, correctAnswer)
  local incorrectAnimation = function()
    self.questionLayer:animateSkimmingResponseIncorrect(response, correctAnswer)
  end
  runInCoroutine(incorrectAnimation)
end

function RiffleScene:endSkimmingQuestion()
  self.questionLayer:clearLayout()
end

function RiffleScene:presentScanningQuestion(question, beginScanningCallback, questionIndex, numberIncorrect)
  local showIntro = (questionIndex == 1)
  if showIntro then
    Sound.playSoundEffect("skimming_nextUp")
    wait(self.questionLayer:presentIntroIntoScanningQuestion(question, questionIndex, numberIncorrect))
  else
    wait(self.questionLayer:presentScanningQuestion(question, questionIndex))
  end

  self.questionLayer:setBeginScanningCallback(beginScanningCallback)
end

function RiffleScene:displayForScanningQuestion(answerString, correctCallback, incorrectCallback)
  self.scrollLayer:addTouchToViewsByString(answerString, correctCallback)
  self.scrollLayer:setNothingTappedCallback(incorrectCallback)

  Sound.playSoundEffect("skimming_menuWoosh")

  return self.questionLayer:collapseForScanningRound()
end

function RiffleScene:setBoldingEnabled(enabled)
  self.article:setBoldingEnabled(enabled)
end

function RiffleScene:removeDoneButton()
  self.scrollLayer:removeDoneButton()
end

function RiffleScene:doCorrectAnimation(x, y, responseText)
  doIconStamp(self, "tick.png", x, y)
  local pressEffect = getPressEffect(self)
  pressEffect:start()

  local delayBeforeContinuing = 1.0

  self.article:prepareForCorrectAnimation(responseText)

  local correctAnimation = AHSequence.new(
    self.article:createCorrectAnimation(responseText),
    AHDelay.new(delayBeforeContinuing)
  )

  return correctAnimation:start()
end

function RiffleScene:doIncorrectAnimation(x, y, expectedText)
  doIconStamp(self, "cross.png", x, y)
  local pressAction = getPressEffect(self)

  local scrollLocationY = self.article:getScrollLocationY(expectedText)
  local scrollDuration = 1.1
  local delayBeforeScrolling = 0.2
  local delayBeforeHighlighting = 0.8
  local delayBeforeContinuing = 1.8

  self.article:prepareForIncorrectAnimation(expectedText)

  local incorrectAnimation = AHSequence.new(
    pressAction,
    AHDelay.new(delayBeforeScrolling),
    function() self.scrollLayer:scrollToLocationY(scrollLocationY, scrollDuration) end,
    AHDelay.new(delayBeforeHighlighting),
    self.article:createIncorrectAnimation(expectedText),
    AHDelay.new(delayBeforeContinuing)
  )

  return incorrectAnimation:start()
end

function RiffleScene:clearScanningRound(isLastRound)
  self.scrollLayer:clearTouchActiveViews()

  if not isLastRound then
    self.scrollLayer:scrollToTop()
    self.article:clearHighlights()
    Sound.playSoundEffect("skimming_menuWoosh")
    self.questionLayer:animateIn()
    self.questionLayer:setBeginScanningCallback(nil)
  end

  self.questionLayer:clearLayout()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleScene.new(round)
  local self = FilledRectangle.newFullScreen({1,1,1})
  Mixin.mix(self, RiffleScene)

  createLayout(self, round)

  return self
end
