--- RiffleScrollLayer
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleScrollLayer")

require("widgets.TouchScroller")
require("RiffleCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BUTTON_COLOR = RiffleCommon.COLORS.BLACK
local DONE_BUTTON_LOCATION_X = Screen.getWidthPart(0.5)
local DONE_BUTTON_LOCATION_Y = 60 + Host.safeAreaBottomInset * 0.3
local DONE_BUTTON_TEXT = "I'm done"
local SCROLL_TO_TOP_DURATION = 1.5
local COLLAPSED_QUESTION_DEAD_AREA_Y = RiffleCommon.COLLAPSED_QUESTION_PANEL_HEIGHT

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setContentsTouchEnabled(self, enabled)
  for _, item in ipairs(self.interactableItems) do
    item:setTouchEnabled(enabled)
  end
end

local function registerFinishedReading(self)
  if self.finishedReadingCallback then
    self.finishedReadingCallback()
  end
end

local function createContents(self, article)
  local contents = Group.new(article:getDims())

  contents:insertAtCenter(article)

  local doneButton = RiffleButton.newNoTouchHandling(RiffleButton.BUTTON_TYPE.BLACK)

  TouchManager.addTouchEnabledMethods(doneButton, function(event)
    if (event.phase == TouchManager.TOUCH_ENDED) then
      Sound.playSoundEffect("skimming_button")
      doneButton:doPressedAnimation()
      registerFinishedReading(self)
    end
  end)

  doneButton:setText(DONE_BUTTON_TEXT)
  local buttonX = DONE_BUTTON_LOCATION_X
  local buttonY = DONE_BUTTON_LOCATION_Y
  doneButton:setLoc(buttonX, buttonY)
  doneButton:setOpacity(RiffleCommon.INVISIBLE_OPACITY)
  contents:insert(doneButton)

  return contents, doneButton
end

local function registerNothingTapped(self, event)
  if self.nothingTappedCallback then
    self.nothingTappedCallback(event.x, event.y)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------
function RiffleScrollLayer:prepareForAnimateIn()
  self.scroller:setScrollBoundsEnabled(false)
  self.scroller:setScrollY(self.contents:getHeightPart(0.5))
end

function RiffleScrollLayer:animateIn()
  local scrollDuration = 1.25
  local ease = AHEaseExponentialInOut.new()
  local destinationScrollY = 0
  local scrollAction = self.scroller:newScrollYTo(scrollDuration, ease, destinationScrollY)

  local onFinish = function()
    self.doneButton:setOpacity(RiffleCommon.VISIBLE_OPACITY)
    self.scroller:setScrollBoundsEnabled(true)
  end

  local animateInSequence = AHSequence.new(scrollAction, onFinish)

  return animateInSequence:start()
end

function RiffleScrollLayer:setFinishedReadingCallback(callback)
  self.finishedReadingCallback = callback
end

function RiffleScrollLayer:setTouchEnabled(enabled)
  self.scroller:setTouchEnabled(enabled)

  local articleTouchViews = self.article:getTouchViews()
  if articleTouchViews then
    for _, view in ipairs(articleTouchViews) do
      view:setTouchEnabled(enabled)
    end
  end

  for _, interactableItem in ipairs(self.interactableItems) do
    interactableItem:setTouchEnabled(enabled)
  end
end

function RiffleScrollLayer:scrollToLocationY(locationY, duration, ease)
  if (ease == nil) then
    ease = AHEaseBezier.newFlat({0.6, 0.0, 0.3, 1.0})
  end

  local maxScrollY = math.max(0, self.scroller:getContents():getHeight() - self.scroller:getHeight())
  locationY = math.min(locationY, maxScrollY)

  local scrollAction = self.scroller:newScrollYTo(duration, ease, locationY)
  scrollAction:start()

  setContentsTouchEnabled(self, false)
end

function RiffleScrollLayer:scrollToTop()
  local topY = 0
  local ease = AHEaseExponentialInOut.new()
  self:scrollToLocationY(topY, SCROLL_TO_TOP_DURATION, ease)
end

function RiffleScrollLayer:clearTouchActiveViews()
  self.article:clearTouchActiveViews()
end

function RiffleScrollLayer:addTouchToViewsByString(text, viewTappedCallback)
  local touchViews = self.article:addTouchToViewsByString(text, viewTappedCallback)

  for _, view in ipairs(touchViews) do
    table.insert(self.interactableItems, view)
  end
end

function RiffleScrollLayer:setNothingTappedCallback(callback)
  self.nothingTappedCallback = callback
end

function RiffleScrollLayer:removeDoneButton(enabled)
  if self.doneButton then
    TableUtils.removeValueFromArray(self.interactableItems, self.doneButton)
    self.doneButton:removeSelf()
    self.doneButton = nil
  end
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleScrollLayer.new(article)
  local width = Screen.width
  local height = Screen.height

  local self = Group.new(width, height)
  Mixin.mix(self, RiffleScrollLayer)

  local contents, doneButton = createContents(self, article)

  self.scroller = TouchScroller.new(width, height, contents)
  self.scroller:setInertialScrollingEnabled(true)
  self.scroller:setWillResetInitialTouchCoordinateUponScrollStart(true)
  self:insertAtCenter(self.scroller)

  local function onScrollStarted() setContentsTouchEnabled(self, false) end
  local function onScrollEnded() setContentsTouchEnabled(self, true) end
  self.scroller:setScrollNotificationFunctions(onScrollStarted, onScrollEnded)

  self.interactableItems = {}
  self.contents = contents
  self.doneButton = doneButton
  self.article = article

  table.insert(self.interactableItems, doneButton)

  local nothingTappedRegion = article:getNothingTouchedRegion()

  local touchEndedWrapper = function(event)
    local wasAboveQuestion = (event.y >= COLLAPSED_QUESTION_DEAD_AREA_Y)
    if wasAboveQuestion then
      registerNothingTapped(self, event)
    end
  end

  local touchHandler = Button.createButtonTouchHandler(nil, nil, touchEndedWrapper)
  TouchManager.addTouchEnabledMethods(nothingTappedRegion, touchHandler)

  table.insert(self.interactableItems, nothingTappedRegion)

  self:setTouchEnabled(false)
  setContentsTouchEnabled(self, false)

  self.nothingTappedRegion = nothingTappedRegion

  return self
end
