--- RiffleTipLabel
-- @author Created by Jay Austria on February 23, 2015.

class("RiffleTipLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
RiffleTipLabel.TYPE =
{
  SKIM = "RiffleTipLabel.TYPE.SKIM",
  SCAN = "RiffleTipLabel.TYPE.SCAN"
}

local WIDTH = Screen.getWidthPart(0.875)
local BOX_WIDTH = Screen.getWidthPart(0.11)
local BOX_LABEL_TEXT = "TIP"
local TIP_SPACING_X = 8
local TIP_FONT = FontFaces.DIN_OT_MEDIUM
local TIP_FONT_SIZE = 15
local LABEL_OFFSET_NUDGE_Y = 3

local BEGINNER_TIP_TEXT =
{
  [RiffleTipLabel.TYPE.SKIM] = "The type of content generally important in skimming is bolded.",
  [RiffleTipLabel.TYPE.SCAN] = "Keep the information you're looking for in mind as you scan."
}

local TIP_TEXT =
{
  [RiffleTipLabel.TYPE.SKIM] =
  {
    "Read the first and last sentences of each paragraph.",
    "Pay attention to headers and images.",
    "Pay attention to names, dates, and numbers.",
    "Look for qualifying adjectives like <i>most</i>, <i>all</i>, or <i>never</i>.",
    "Take note of unusual or unfamiliar words.",
    "Let your eyes follow a spiral pattern down the page.",
    "Consider the language: is it technical, formal, or casual?"
  },
  [RiffleTipLabel.TYPE.SCAN] =
  {
    "Consider whether you're looking for a name, number, or idea.",
    "Consider the likely context of the idea you're looking for.",
    "Scan in a spiral pattern, not top to bottom or left to right."
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local fontSize = 18
  if Screen.isTabletAspectRatio() then
    fontSize = 15
  end

  local centerY = self:getHeightPart(0.5)

  local box = HollowRectangle.new(BOX_WIDTH, BOX_WIDTH, 1, nil, RiffleCommon.COLORS.GREY)
  local boxX = box:getWidthPart(0.5)

  box:setLoc(boxX, centerY)
  self:insert(box)

  local boxLabel = Label.new(BOX_LABEL_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, fontSize))
  boxLabel:setLoc(boxX, centerY)
  boxLabel:setColor(unpack(RiffleCommon.COLORS.GREY))
  self:insert(boxLabel)

  local initialText = "-unset-"
  local tipWidth = self:getWidth() - BOX_WIDTH - TIP_SPACING_X

  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  }

  local attributesForLabels = {
    b = { font = Font.new(FontFaces.DIN_OT_BOLD, fontSize) },
    i = { font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, fontSize) }
  }

  local tipLabel = DynamicTextArea.new(initialText, tipWidth, baseAttributes, attributesForLabels)
  tipLabel:setAnchorPointTop()
  tipLabel:setColor(unpack(RiffleCommon.COLORS.GREY))
  local tipLabelX = BOX_WIDTH + TIP_SPACING_X + tipLabel:getWidthPart(0.5)
  tipLabel:setLoc(tipLabelX, self:getHeight() + LABEL_OFFSET_NUDGE_Y)
  self:insert(tipLabel)

  self.tipLabel = tipLabel
end

local function incrementTipIndex(self)
  self.tipIndex = self.tipIndex + 1

  if (self.tipIndex > #self.tipTexts) then
    self.tipIndex = 1
  end
end

local function animateNextTip(self)
  local disableTouch = function() self:setTouchEnabled(false) end
  local enableTouch = function() self:setTouchEnabled(true) end
  local changeTip = function() self:showNextTip() end

  local animateTip = AHSequence.new(
    disableTouch,
    AHColorTo.new(self.tipLabel, 0.1, nil, RiffleCommon.COLORS.BLACK),
    changeTip,
    AHColorTo.new(self.tipLabel, 0.1, nil, RiffleCommon.COLORS.GREY),
    enableTouch
  )

  animateTip:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleTipLabel:showBeginnerTip()
  local tipText = BEGINNER_TIP_TEXT[self.type]
  self.tipLabel:setText(tipText)
end

function RiffleTipLabel:showNextTip()
  incrementTipIndex(self)
  local nextText = self.tipTexts[self.tipIndex]
  self.tipLabel:setText(nextText)
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

function RiffleTipLabel.new(isBeginner, type)
  local width = WIDTH
  local height = BOX_WIDTH -- set label height to match width of square box
  local self = Group.new(width, height)
  Mixin.mix(self, RiffleTipLabel)

  createLayout(self)

  local tipTexts = TIP_TEXT[type]
  self.tipTexts = tipTexts
  self.type = type
  self.tipIndex = math.random(#tipTexts)

  if (isBeginner) then
    self:showBeginnerTip()
  else
    self:showNextTip()
  end

  local showNextTipWrapper = function()
    animateNextTip(self)
  end

  Button.makeButton(self, showNextTipWrapper)

  self:linkOpacityAttrToChildren()

  return self
end
