--- RiffleValidator
-- @author Created by Jay Austria on February 23, 2015.
-- ConceptValidator for Riffle.

class("RiffleValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "identifier",
  "passage",
  "title",
  "content"
}

local OPTIONAL_KEYS = {
  "subheader",
  "image",
  "imageCredit"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Validate
--
-- Must have at least 2 easy skimming and 2 hard skimming
-- Must have at least 4 easy scanning and 4 hard scanning
-- Must have properly formatted tags
-- All questions must be matched somewhere in the text
--
--
--
--
--

local MIN_SKIMMING_ROUND_COUNT = 2
local MIN_SCANNING_ROUND_COUNT = 3
local ROUND_KEY_SKIM = "skim"
local ROUND_KEY_SCAN = "scan"
local OPTIONAL_PAIRED_TAGS = {"strong", "u", "i"} -- use optional tags validator so we can catch multiple instances
local OPTIONAL_SINGLE_TAGS = {"br"}

local function validateRoundCount(self, concept, roundType, minimum)
  local questions = concept.content
  local allQuestionsOfType = TableUtils.getFilteredArray(questions, function(value) return (value.round == roundType) end)

  local easyQuestions = TableUtils.getFilteredArray(allQuestionsOfType, function(value) return (value.filter_difficulty == "Easy") end)
  local hardQuestions = TableUtils.getFilteredArray(allQuestionsOfType, function(value) return (value.filter_difficulty == "Hard") end)

  self:expectGreaterOrEqual(#easyQuestions, minimum, "Not enough easy questions for round")
  self:expectGreaterOrEqual(#hardQuestions, minimum, "Not enough hard questions for round")
end

local function validateSkimmingRoundCount(self, concept)
  validateRoundCount(self, concept, ROUND_KEY_SKIM, MIN_SKIMMING_ROUND_COUNT)
end

local function validateScanningRoundCount(self, concept)
  validateRoundCount(self, concept, ROUND_KEY_SCAN, MIN_SCANNING_ROUND_COUNT)
end

local function validateConceptTags(self, concept)
  local singleTags = nil
  local pairedTags = nil
  local tertiaryTags = nil

  self:validateTags(concept.passage, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS, OPTIONAL_SINGLE_TAGS)
  self:validateTags(concept.subheader, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS)

  for _, round in ipairs(concept.content) do
    self:validateTags(round.prompt, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS)

    if round.answer then
      if round.round == ROUND_KEY_SCAN then
        -- scanning rounds are allowed to break tag rules because answers are not displayed
      else
        self:validateTags(round.answer, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS)
      end
    end

    if round.distractor_1 then
      self:validateTags(round.distractor_1, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS)
    end

    if round.distractor_2 then
      self:validateTags(round.distractor_2, singleTags, pairedTags, tertiaryTags, OPTIONAL_PAIRED_TAGS)
    end
  end
end

local function validateSkimmingQuestions(self, concept)
  local skimmingQuestions = TableUtils.getFilteredArray(concept.content, function(value) return (value.round == ROUND_KEY_SKIM) end)

  for _, question in ipairs(skimmingQuestions) do
    self:expectNotEmpty(question.prompt, "prompt")
    self:expectNotEmpty(question.answer, "answer")
    self:expectNotEmpty(question.distractor_1, "distractor_1")
    self:expectNotEmpty(question.question_identifier, "question_identifier")

    -- special logic for true and false formatted questions
    if question.distractor_1 and ((question.distractor_1 ~= "true") and (question.distractor_1 ~= "false")) then
      self:expectNotEmpty(question.distractor_2, "distractor_2")
    end
  end
end

local function validateScanningQuestions(self, concept)
  local scanningQuestions = TableUtils.getFilteredArray(concept.content, function(value) return (value.round == ROUND_KEY_SCAN) end)

  for _, question in ipairs(scanningQuestions) do
    local answer = question.answer
    local isAnswerInPrompt = (concept.passage:find(answer, 1, true) ~= nil)

    self:expectNotEmpty(question.question_identifier, "question_identifier")
    self:expectTrue(isAnswerInPrompt, "Answer was not found anywhere in the prompt:" .. answer:sub(1, 20) .. " (cont.)")
  end
end

local function validateNoPunctuationAfterTag(self)
end

local function validateImageFile(self, concept)
  if concept.image and (#concept.image > 0) then
    self:validateImageFile(concept.image)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RiffleValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateSkimmingRoundCount(self, concept)
  validateScanningRoundCount(self, concept)
  validateSkimmingQuestions(self, concept)
  validateScanningQuestions(self, concept)
  validateNoPunctuationAfterTag(self, concept)
  validateImageFile(self, concept)
  validateConceptTags(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RiffleValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, RiffleValidator)

  return self
end
