--- Rocket In-game UI
-- @author Created by Karl Stenerud on Oct 14, 2013.

class("InGameUI")
require("SidebarButton")
require("RocketButton")
require("RocketCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_QUESTION_TEXT = {1.0, 1.0, 1.0}
local COLOR_ANSWER_TEXT = {0.2, 1.0, 0.2}

local LABEL_HEIGHT = 50
local TOP_GUTTER_HEIGHT = 30

local FONT = Font.new(RocketCommon.FONT_BOLD, 24)

local BUTTON_BEGIN_Y = 100
local BUTTON_SPACING_Y = 10

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getButtonTouchedHandler(self)
  return function(button)
    if not self.isUIEnabled then
      return
    end

    local answer = button.label:getText()

    if button.isCorrect then
      self.onCorrectAnswer(answer)
    else
      self.onIncorrectAnswer(answer)
    end
  end
end

local function setAnswerLabelVisible(self, isVisible)
  self.isAnswerLabelVisible = isVisible
  self.answerLabel:setVisible(isVisible and self.isVisible)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function InGameUI:prepareNextQuestion(questionText, correctAnswerText, incorrectAnswerText, isFirstButtonCorrect)
  return runInCoroutine(function()
    self.nextQuestion = {
      questionText = questionText,
      correctAnswerText = correctAnswerText,
      incorrectAnswerText = incorrectAnswerText,
      correctIdx = 1,
      incorrectIdx = 2
    }

    if not isFirstButtonCorrect then
      self.nextQuestion.correctIdx = 2
      self.nextQuestion.incorrectIdx = 1
    end
  end)
end

function InGameUI:showNextQuestion()
  setAnswerLabelVisible(self, false)

  self.correctIdx = self.nextQuestion.correctIdx
  self.incorrectIdx = self.nextQuestion.incorrectIdx

  self.buttons[self.correctIdx].isCorrect = true
  self.buttons[self.correctIdx]:unmark()
  self.buttons[self.incorrectIdx].isCorrect = false
  self.buttons[self.incorrectIdx]:unmark()

  self.buttons[self.correctIdx]:setText(self.nextQuestion.correctAnswerText)
  self.buttons[self.incorrectIdx]:setText(self.nextQuestion.incorrectAnswerText)
  self.questionLabel:setText(self.nextQuestion.questionText)
  self.answerLabel:setText(self.nextQuestion.correctAnswerText)
end

function InGameUI:runCorrectAnswerSequence()
  self.buttons[self.correctIdx]:markCorrect()
end

function InGameUI:runIncorrectAnswerSequence()
  self.buttons[self.incorrectIdx]:markIncorrect()
end

function InGameUI:showCorrectAnswer()
  setAnswerLabelVisible(self, true)
  for _, button in ipairs(self.buttons) do
    button:setVisible(false)
  end
end

function InGameUI:showContinue()
  self.continueButton:show()
end

function InGameUI:hideContinue()
  Sound.playSoundEffect("rocket_continue_click")
  self.continueButton:hide()
end

function InGameUI:getCorrectButton()
  for _, button in ipairs(self.buttons) do
    if (button.isCorrect) then
      return button
    end
  end
end

function InGameUI:reset()
  setAnswerLabelVisible(self, false)
  for _, button in ipairs(self.buttons) do
    button:setVisible(true)
  end
end

function InGameUI:setUIEnabled(isEnabled)
  self.isUIEnabled = isEnabled
end

function InGameUI:setVisible(isVisible)
  self.isVisible = isVisible
  for _, component in ipairs(self.uiComponents) do
    if component ~= self.answerLabel or self.isAnswerLabelVisible then
      component:setVisible(isVisible)
    end
  end
end

function InGameUI:getPlayAreaHeight()
  return Screen.safeTop - TOP_GUTTER_HEIGHT
end

function InGameUI:fadeOut()
  local duration = 0.25
  local ease = nil
  local destinationOpacity = 0.0

  local firstButton = self.buttons[1]
  local firstButtonAction = firstButton:createFadeAction(duration, ease, destinationOpacity)

  local secondButton = self.buttons[2]
  local secondButtonAction = secondButton:createFadeAction(duration, ease, destinationOpacity)

  local questionLabel = self.questionLabel
  local questionLabelAction = AHOpacityTo.new(questionLabel, duration, ease, destinationOpacity)

  local answerLabel = self.answerLabel
  local answerLabelAction = AHOpacityTo.new(answerLabel, duration, ease, destinationOpacity)

  local fadeAction = AHSpawn.new(
                                 firstButtonAction,
                                 secondButtonAction,
                                 answerLabelAction,
                                 questionLabelAction
                                 )
  fadeAction:start()
end

function InGameUI:getQuestionLabelCopy()
  local questionLabel = Label.new("", FONT)
  questionLabel:setColor(unpack(COLOR_QUESTION_TEXT))
  questionLabel:setAnchorPointBottom()
  questionLabel:setLoc(Screen.getSafeX(0.5), 0)

  return questionLabel
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function InGameUI.new(onCorrectAnswer, onIncorrectAnswer)
  local self = Group.new(Screen.fullWidth, Screen.fullHeight)
  Mixin.mix(self, InGameUI)

  self.isUIEnabled = false
  self.isVisible = true

  self.onCorrectAnswer = onCorrectAnswer
  self.onIncorrectAnswer = onIncorrectAnswer

  self.buttons = {
    RocketButton.new(getButtonTouchedHandler(self)),
    RocketButton.new(getButtonTouchedHandler(self))
  }

  local _, buttonH = self.buttons[1]:getDims()
  local currentY = BUTTON_BEGIN_Y

  self.buttons[1]:setLoc(Screen.getSafeX(0.5), currentY)
  self:insert(self.buttons[1], Z_BUTTON)

  currentY = currentY + buttonH + BUTTON_SPACING_Y
  self.buttons[2]:setLoc(Screen.getSafeX(0.5), currentY)
  self:insert(self.buttons[2], Z_BUTTON)

  currentY = math.floor(currentY + buttonH - Screen.safeHeight * 0.05)
  self.questionLabel = self:getQuestionLabelCopy()
  self.questionLabel:setLoc(self.questionLabel:getX(), currentY)
  self:insert(self.questionLabel)

  currentY = currentY - LABEL_HEIGHT
  self.answerLabel = Label.new("", FONT)
  self.answerLabel:setColor(unpack(COLOR_ANSWER_TEXT))
  self.answerLabel:setLoc(self.buttons[2]:getLoc())
  self:insert(self.answerLabel)
  setAnswerLabelVisible(self, false)

  self.continueButton = SidebarButton.newWithImageAndText("rocket_continue.png", string.lower(Localizer.get("Misc.CONTINUE")), {1, 1, 1})
  self.continueButton:positionAtStart(self.buttons[1]:getY())
  self:insert(self.continueButton)

  self.uiComponents = {}
  table.insert(self.uiComponents, self.buttons[1])
  table.insert(self.uiComponents, self.buttons[2])
  table.insert(self.uiComponents, self.questionLabel)
  table.insert(self.uiComponents, self.answerLabel)

  self:setBounds(0.0, 0.0, 0, Screen.fullWidth, Screen.fullHeight, 0)
  self:setAnchorPointCenter()
  self:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))

  return self
end
