--- Rocket Game
-- @author Created by Karl Stenerud on Oct 8, 2013.

class("RocketGame")
require("bonuses.BonusNotificationTray")
require("bonuses.ResponseSpeedTracker")
require("Background")
require("StandardUIBuilder")
require("Rocket")
require("LaunchPlatform")
require("InGameUI")
require("RocketProgressIndicator")
require("RocketConfiguration")
require("RocketConceptGenerator")
require("widgets.TapToContinue")
require("tutorials.TutorialHighlighter")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Colors -----------------------------

local COLOR_UI_COMPONENTS = {1, 1, 1, 0.8}
local COLOR_TUTORIAL_OVERLAY = {0.039, 0.055, 0.082, 0.7}
local COLOR_TUTORIAL_MASK = {1.0, 1.0, 1.0, 1.0}

local TUTORIAL_FONT_SIZE = 16

-- Z Ordering -------------------------

local Z_BACKGROUND = 0
local Z_LAUNCH_PLATFORM = 10
local Z_ROCKET = 20
local Z_BUTTON = 30
local Z_GAME_OVER = 40
local Z_GAME_OVER_FADE = 50
local Z_IN_GAME_UI = 100
local Z_TUTORIAL = 110
local Z_FULLSCREENTAP = 200

-- Tags -------------------------------

local INTERRUPT_PRIORITY_1 = 1
local INTERRUPT_PRIORITY_2 = 2
local INTERRUPT_PRIORITY_3 = 3

local INTERRUPT_PRIORITY_COUNT = 2

-- Lives
local STARTING_SHIPS = 3

-- Rocket launch timing
local SCROLL_RAMPUP_DURATION = 0.3
local THRUST_START_DURATION = 0.2
local THRUST_DURATION = 0.5
local THRUST_PREPARE_DURATION = 1.0
local LAUNCH_PLATFORM_SLIDE_DURATION = 0.8
local SHIP_FALL_DURATION = 0.8

-- How long it takes for the ship to move from one side of the screen to the other
local ROCKET_SIDE_TO_SIDE_DURATION = 3

-- Background scrolling speed in pixels per second.
local SCROLL_SPEED_MIN = 200
local SCROLL_SPEED_MAX = 1200
local SCROLL_SPEED_IDLE = 50
local SCROLL_SPEED_TUTORIAL = 30
local SCROLL_SPEED_END_GAME = 0
local SCROLL_SPEED_BOOST = 1800

-- Slow down duration after incorrect answer
local SCROLL_SLOWDOWN_DURATION = 1.2

-- Time to wait between showing correct answer and ending game
local GAME_OVER_END_SEQUENCE_DELAY = 1.4
local GAME_OVER_PLATFORM_DELAY = 0.5
local GAME_OVER_PLATFORM_SLIDE_DURATION = 1.7

-- Used for the fullscreen tap to exit at end
local FULLY_TRANSPARENT_COLOR = {0.0, 0.0, 0.0, 0.0}

local AUTO_EXIT_DELAY = 2.0

local function configureGameDifficulty(self)
  self.correctNeeded = RocketConfiguration.CONCEPTS_TO_WIN_COUNT

  local passScore = RocketConfiguration.getPassScore()
  local rankTwoScore = RocketConfiguration.getRankTwoScore()
  local rankThreeScore = RocketConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = RocketConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(STARTING_SHIPS)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)

  self.initialAnswerDuration = RocketConfiguration.getAnswerDurationStart()
  self.answerDurationMultiplier = RocketConfiguration.getAnswerDurationMultiplier()
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local onTimeout

local function interruptInterruptibleActions(self, priority)
  local interruptibleActions = self.interruptibleActions

  for i = 1, priority do
    local actionsAtPriority = interruptibleActions[i]
    if actionsAtPriority then
      for action in pairs(actionsAtPriority) do
        action:stop()
      end
    end
    interruptibleActions[i] = {}
  end
end

local function tagInterruptibleAndRunAction(self, action, priority)
  local interruptibleActions = self.interruptibleActions

  local actionsAtPriority = interruptibleActions[priority]
  actionsAtPriority[action] = true

  return action:start()
end

local function getRefreshSequenceForY(self, endY)
  return AHSequence.new(
           AHCallLogic.new(function()
                           self.rocket:runBoost(THRUST_DURATION*0.7)
                           self.background:boostScrollSpeed(THRUST_DURATION*0.5, SCROLL_SPEED_BOOST)
                         end),
                         AHLocationYTo.new(self.rocket, THRUST_DURATION, AHEaseSineInOut.new(), endY),
                         AHCallLogic.new(function()
                                           self.rocket:startNormalThrust()
                                         end)
       )
end

local function getBoostOffscreenAction(self)
  return getRefreshSequenceForY(self, self.rocketOffscreenY)
end

local function getRefreshFuelAction(self)
  return getRefreshSequenceForY(self, self.rocketCruiseMaxY)
end

local function getDrainFuelAction(self)
  return AHSequence.new(
                        AHLocationYTo.new(self.rocket, self.answerDuration, nil, self.rocketCruiseMinY),
                        AHCallLogic.new(function()
                                            onTimeout(self)
                                          end)
                        )
end

local function getRocketSideToSideAction(self)

  local selfW, selfH = self:getDims()
  local rocketX, rocketY = self.rocket:getDims()
  local minX = rocketX
  local maxX = selfW - rocketX
  local curX = self.rocket:getX()
  local proportion = curX / (maxX - minX)
  return AHSequence.new(
    AHLocationXTo.new(self.rocket, ROCKET_SIDE_TO_SIDE_DURATION * proportion, AHEaseSineOut.new(), minX),
    AHLoop.new(
      AHSequence.new(
        AHLocationXTo.new(self.rocket, ROCKET_SIDE_TO_SIDE_DURATION, AHEaseSineInOut.new(), maxX),
        AHLocationXTo.new(self.rocket, ROCKET_SIDE_TO_SIDE_DURATION, AHEaseSineInOut.new(), minX)
      )
    )
  )
end

local function startRocketSideToSide(self)
  return tagInterruptibleAndRunAction(self, getRocketSideToSideAction(self), INTERRUPT_PRIORITY_2)
end

local function refreshFuel(self)
  return tagInterruptibleAndRunAction(self, getRefreshFuelAction(self), INTERRUPT_PRIORITY_1)
end

local function drainFuel(self)
  return tagInterruptibleAndRunAction(self, getDrainFuelAction(self), INTERRUPT_PRIORITY_1)
end

local function refreshAndDrainFuel(self)
  local action = AHSequence.new(
                                getRefreshFuelAction(self),
                                AHDelay.new(0.5),
                                getDrainFuelAction(self)
                                )
  return tagInterruptibleAndRunAction(self, action, INTERRUPT_PRIORITY_1)
end

local function increaseSpeed(self)
  self.answerDuration = self.answerDuration * self.answerDurationMultiplier
  if self.answerDuration < RocketConfiguration.ANSWER_DURATION_MINIMUM then
    self.answerDuration = RocketConfiguration.ANSWER_DURATION_MINIMUM
  end
end

local function getBGScrollSpeed(self)
  return MathUtils.mapLinear(self.answerDuration, self.initialAnswerDuration, RocketConfiguration.ANSWER_DURATION_MINIMUM, SCROLL_SPEED_MIN, SCROLL_SPEED_MAX)
end

local function updateScore(self)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function setNumShips(self, ships)
  self.ships = ships
end

local function reduceShips(self)
  local ships = self.ships - 1
  if ships < 0 then
    ships = 0
  end
  setNumShips(self, ships)
  self.shipsCounter:decrement()
end

local function getRandomDistractor(concept)
  local distractors = {
    concept.distractor1,
    concept.distractor2,
    concept.distractor3
  }

  local randomDistractor = TableUtils.getRandomElement(distractors)

  return randomDistractor
end

local function prepareNextQuestion(self)
  return runInCoroutine(function()
    if not self.gameRunning then
      return
    end

    local isFirstButtonCorrect = MathUtils.getRandomBoolean()

    if (GameSession.getCorrectAnswerCount() > RocketConfiguration.NUMBER_OF_CONCEPTS_BEFORE_RESURFACING_MISTAKES) then
      self.conceptGenerator:allowMistakesToResurface()
    end

    local concept = self.conceptGenerator:next()

    local questionField = concept.question
    local answerField = concept.answer
    local incorrectAnswerField = getRandomDistractor(concept)

    self.nextConcept = concept
    wait(self.inGameUI:prepareNextQuestion(questionField, answerField, incorrectAnswerField, isFirstButtonCorrect))

  end)
end

local function startNextQuestion(self)
  if not self.gameRunning then
    return
  end

  self.responseSpeedTracker:notifyStart()

  self.concept = self.nextConcept
  GameSession.contentTracker:notifyPresentedConcept(self.concept.identifier)
  self.inGameUI:showNextQuestion()
  self.inGameUI:reset()
  self.inGameUI:setVisible(true)
  self.inGameUI:setUIEnabled(true)
end

local function raiseLaunchPlatform(self)
  return runInCoroutine(function()
    local diffY = self.launchPlatformYHigh - self.launchPlatformYLow
    AHLocationYBy.new(self.rocket, THRUST_PREPARE_DURATION, AHEaseSineOut.new(), diffY):start()
    wait(AHLocationYTo.new(self.launchPlatform, THRUST_PREPARE_DURATION, AHEaseSineOut.new(), self.launchPlatformYHigh):start())
  end)
end

local function launchRocket(self, launchFromPlatform)
  return runInCoroutine(function()
    self.inGameUI:setVisible(false)

    self.rocket:setLoc(self.launchPlatform:getRocketLandingPosition(self.rocket, self))
    if launchFromPlatform then
      wait(raiseLaunchPlatform(self))
      AHLocationYTo.new(self.launchPlatform, LAUNCH_PLATFORM_SLIDE_DURATION, AHEaseExponentialIn.new(), self.launchPlatformYLow):start()
    end
    self.rocket:startIgnition()
    self.launchPlatform:startSmoke()
    waitFor(0.3)
    wait(self.background:rampScrollSpeed(SCROLL_RAMPUP_DURATION, getBGScrollSpeed(self)))
    if launchFromPlatform then
      self.progressIndicator:playIntroAnimation()
    end

    startRocketSideToSide(self)
  end)
end

local function relaunchRocket(self)
  GameSession.notifyGameTimePause()

  self.background:rampScrollSpeed(SCROLL_SLOWDOWN_DURATION, SCROLL_SPEED_IDLE)
  self.inGameUI:showContinue()

  local tapToContinueOverlay = TapToContinue.new()
  tapToContinueOverlay:waitForTrigger()

  GameSession.notifyGameTimeStart()

  self.inGameUI:hideContinue()
  wait(launchRocket(self, false))
  startNextQuestion(self)
  refreshAndDrainFuel(self)
end

local function notifyGameWin(self)
  GameSession.notifyGameWin()
end

local function notifyGameLoss(self)
  GameSession.notifyGameLoss()
end

local function showLossOutroAndEndGame(self)
  self.hasLossOutroStarted = true
  self.inGameUI:fadeOut()
  wait(self:showDefaultGameLossOutro())
  notifyGameLoss(self)
end

local function allowFullScreenExit(self)
  local child = nil
  local timeoutSeconds = nil
  local onFullScreenTap = function()
    if (not self.hasLossOutroStarted) then
      showLossOutroAndEndGame(self)
    end
  end

  local tapToContinueOverlay = TapToContinue.new(child, timeoutSeconds, onFullScreenTap)
end

local function playPlatformLandingSequence(self)
  ScoringUtils.applyLivesBonus(self.ships, RocketConfiguration.POINTS_PER_LIVES_BONUS, self.bonusTray)
  updateScore(self)

  Sound.playSoundEffect("rocket_goal_reached")
  self:showDefaultGameWinOutro()

  local boostAction = getBoostOffscreenAction(self)
  tagInterruptibleAndRunAction(self, boostAction, INTERRUPT_PRIORITY_1)

  self.inGameUI:setUIEnabled(false)
  AHDelayed.new(0.25, function()
    self.inGameUI:fadeOut()
  end):start()
  self.gameRunning = false

  wait(self.background:rampScrollSpeed(SCROLL_SLOWDOWN_DURATION, SCROLL_SPEED_END_GAME))
  local platformSprite = Sprite.newWithName("rocket_launch_platform2.png")
  platformSprite:setEffectiveDims(Screen.width)
  local platformHeight = platformSprite:getEffectiveHeight()

  local bufferPlatformHeight = platformHeight * 0.28
  local safeAreaAdjust = math.min(bufferPlatformHeight, Screen.safeBottom)
  local platformDestination = bufferPlatformHeight - safeAreaAdjust

  platformSprite:setAnchorPointBottom()
  platformSprite:setLoc(Screen.fullWidth/2, -platformHeight)
  self:insert(platformSprite, Z_GAME_OVER)

  waitFor(GAME_OVER_PLATFORM_DELAY)
  wait(AHLocationYTo.new(platformSprite, GAME_OVER_PLATFORM_SLIDE_DURATION, AHEaseSineOut.new(), -platformDestination):start())

  waitFor(GAME_OVER_END_SEQUENCE_DELAY)

  notifyGameWin(self)
end

local function onGameOver(self)
  self.background:rampScrollSpeed(SCROLL_SLOWDOWN_DURATION, SCROLL_SPEED_END_GAME)
  local function delayedAllowExitAndAutoExitFunction()
    waitFor(SCROLL_SLOWDOWN_DURATION)
    allowFullScreenExit(self)
    waitFor(AUTO_EXIT_DELAY)
    if (not self.hasLossOutroStarted) then
      showLossOutroAndEndGame(self)
    end
  end

  runInCoroutine(delayedAllowExitAndAutoExitFunction)
end

local function didWin(self)
  return (self.progressIndicator:getProgress() >= 1.0)
end

local function getOnCorrectAnswerHandler(self)
  return function(answer)
    self.shouldShowTutorial = false
    self.responseSpeedTracker:notifyCorrectResponse()
    self.inGameUI:setUIEnabled(false)
    return runInCoroutine(function()
      Sound.playSoundEffect("rocket_correct")
      interruptInterruptibleActions(self, INTERRUPT_PRIORITY_1)
      GameSession.contentTracker:notifyCorrectAnswer(self.concept.identifier)
      GameSession.notifyCorrectAnswer(self.concept.identifier)
      GameSession.incrementCorrectAnswerCount()
      GameSession.increaseScore(RocketConfiguration.POINTS_PER_CORRECT_RESPONSE)

      self.progressIndicator:incrementProgress()
      updateScore(self)
      self.background:rampScrollSpeed(0.1, getBGScrollSpeed(self))
      self.inGameUI:runCorrectAnswerSequence()

      if (didWin(self)) then
        GameSession.notifyGameTimeEnd()
        playPlatformLandingSequence(self)
      else
        prepareNextQuestion(self)


        increaseSpeed(self)
        wait(refreshFuel(self))
        startNextQuestion(self)
        drainFuel(self)
      end
    end)
  end
end

local function commonIncorrectProcess(self, isTimeout, answer)
  return runInCoroutine(function()
    self.responseSpeedTracker:notifyIncorrectResponse()
    Sound.playSoundEffect("rocket_incorrect")
    GameSession.scoredIncorrectAnswer(self.concept.identifier)
    GameSession.incrementIncorrectAnswerCount()
    self.conceptGenerator:notifyMistake(self.concept.identifier)
    prepareNextQuestion(self)

    interruptInterruptibleActions(self, INTERRUPT_PRIORITY_3)

    reduceShips(self)
    self.rocket:reset()

    if not isTimeout then
      GameSession.contentTracker:notifyIncorrectAnswer(self.concept.identifier, nil, answer)
      self.inGameUI:runIncorrectAnswerSequence()
      wait(AHLocationYTo.new(self.rocket, SHIP_FALL_DURATION, AHEaseSineIn.new(), self.rocketCruiseMinY):start())
    else
      GameSession.contentTracker:notifyTimeout(self.concept.identifier)
    end

    self.inGameUI:showCorrectAnswer()

    local isGameOver = self.ships < 1
    if isGameOver then
      onGameOver(self)
    else
      relaunchRocket(self)
    end
  end)
end

local function getOnIncorrectAnswerHandler(self)
  return function(answer)
    self.shouldShowTutorial = false
    self.inGameUI:setUIEnabled(false)
    return commonIncorrectProcess(self, false, answer)
  end
end

local function showTutorial(self)
  self.inGameUI:setUIEnabled(false)
  local answerButton = self.inGameUI:getCorrectButton()
  local generateHoleCallback = function()
    local holeRect = FilledRectangle.new(answerButton:getWidth(), answerButton:getHeight(), COLOR_TUTORIAL_MASK)
    local x, y = self:transformLoc(answerButton)
    x = x + holeRect:getWidthPart(0.5)
    y = y + holeRect:getHeightPart(0.5)
    holeRect:setLoc(x, y)
    return holeRect
  end

  local labelMask = self.inGameUI:getQuestionLabelCopy()
  labelMask:setText(self.inGameUI.questionLabel:getText())
  local labelX, labelY = self:transformLoc(self.inGameUI.questionLabel)
  labelMask:setLoc(labelX + labelMask:getWidthPart(0.5), labelY)

  local tutorialOverlay = TutorialHighlighter.newWithMasks(
    {generateHoleCallback},
    {labelMask}
  )
  local instructionText = "Tap the matching word to<br/>keep the rocket flying"
  local instructionOffsetY =  self.inGameUI.questionLabel:getHeightPart(0.5) + answerButton:getHeight()
  local instructionLocationY = self.inGameUI.questionLabel:getY() + instructionOffsetY
  tutorialOverlay:addInstructionLabel(instructionText, TUTORIAL_FONT_SIZE, instructionLocationY)
  self:insertAtCenter(tutorialOverlay, Z_TUTORIAL)

  local fadeDuration = 0.167
  self.background:rampScrollSpeed(fadeDuration, SCROLL_SPEED_TUTORIAL)
  tutorialOverlay:createFadeInAnimation(fadeDuration):start()
  tutorialOverlay:waitForTargetTap()

  wait(tutorialOverlay:createFadeOutAnimation():start())
  tutorialOverlay:removeSelf()

  startRocketSideToSide(self)
  local answer = answerButton.label:getText()
  local answerHandler = getOnCorrectAnswerHandler(self)
  return answerHandler(answer)
end

onTimeout = function(self)
  self.inGameUI:setUIEnabled(false)
  return commonIncorrectProcess(self, true)
end

local function checkForTutorial(self)
  runInCoroutine(function()
    wait(AHDelay.new(RocketConfiguration.TUTORIAL_DELAY):start())
    if self.shouldShowTutorial then
      interruptInterruptibleActions(self, INTERRUPT_PRIORITY_2)
      showTutorial(self)
    end
  end)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function RocketGame:gameFlow()
  self.gameRunning = true
  updateScore(self)
  setNumShips(self, STARTING_SHIPS)
  self.answerDuration = self.initialAnswerDuration

  prepareNextQuestion(self)
  wait(launchRocket(self, true))
  startNextQuestion(self)
  self.shouldShowTutorial = true
  refreshAndDrainFuel(self)
  checkForTutorial(self)
  self.inGameUI:setUIEnabled(true)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)

  local selfW, selfH = Screen.fullWidth, Screen.fullHeight

  self.background = Background.new()
  self.background:setAnchorPointBottomLeft()
  self:insert(self.background, Z_BACKGROUND)

  self.inGameUI = InGameUI.new(getOnCorrectAnswerHandler(self), getOnIncorrectAnswerHandler(self))
  self:insert(self.inGameUI, Z_IN_GAME_UI)
  self.inGameUI:setVisible(false)

  self.rocket = Rocket.new()
  local rocketW, rocketH = self.rocket:getDims()
  self.rocketCruiseMinY = -rocketH/2
  self.rocketCruiseMaxY = self.inGameUI:getPlayAreaHeight() - rocketH/2
  self.rocketOffscreenY = Screen.fullHeight + rocketH
  self.rocket:setLoc(-selfW, -selfH)
  self:insert(self.rocket, Z_ROCKET)

  self.launchPlatform = LaunchPlatform.new()
  local launchPlatformW, launchPlatformH = self.launchPlatform:getDims()
  self.launchPlatformYHigh = 0
  self.launchPlatformYLow = -(launchPlatformH + rocketH/2)
  self.launchPlatform:setAnchorPointBottom()
  self.launchPlatform:setLoc(selfW/2, self.launchPlatformYLow)
  self:insert(self.launchPlatform, Z_LAUNCH_PLATFORM)

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 RocketConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(RocketConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       RocketConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), COLOR_UI_COMPONENTS)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.shipsCounter = builder:addLivesDisplay()
  self.shipsCounter:setLives(STARTING_SHIPS)

  local stepsNeeded = self.correctNeeded
  self.progressIndicator = RocketProgressIndicator.newWithStepCount(stepsNeeded)
  self.progressIndicator:setColor(unpack(COLOR_UI_COMPONENTS))
  self.progressIndicator:setOpacity(0.2)
  local progressIndicatorWidth, progressIndicatorHeight = self.progressIndicator:getDims()
  local xOffset = -2
  local yOffset = Screen.safeHeight * 0.5 - progressIndicatorHeight * 0.5 - StandardUIBuilder.STANDARD_BAR_MARGIN
  builder:addWidget(self.progressIndicator, UIOverlay.EAST, xOffset, yOffset)
end

local function createInterruptibleActions(self)
  local interruptibleActions = {}

  for i = 1, INTERRUPT_PRIORITY_COUNT do
    interruptibleActions[i] = {}
  end

  self.interruptibleActions = interruptibleActions
end

-- Constructor ------------------------

local function playIntroSoundEffect()
  runInCoroutine(function()
    waitFor(0.4)
    -- Synched with the rocket launching off
    Sound.playSoundEffect("rocket_launch", 0.75)
  end)
end

function RocketGame.new()

  local self = BaseGame.new()
  Mixin.mix(self, RocketGame)

  self.hasLossOutroStarted = false

  configureGameDifficulty(self)
  playIntroSoundEffect()
  createGameElements(self)
  createUI(self)

  createInterruptibleActions(self)

  self.conceptGenerator = LearningFocusedRoundGenerator.new(RocketConceptGenerator.new())

  return self
end
