class("RocketProgressIndicator")
require("widgets.progress.ProgressIndicator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INDICATOR_WIDTH = 24
local INDICATOR_HEIGHT = Screen.safeHeight * 0.3
local BASE_WIDTH = INDICATOR_WIDTH * 0.5
local BASE_HEIGHT = INDICATOR_WIDTH * 0.1
local BASE_CORNER_RADIUS = 1
local BASE_START_Y = INDICATOR_HEIGHT * 0.05
local VISIBLE_UI_OPACITY = 0.8
local VISIBLE_UI_COLOR = {1.0, 1.0, 1.0, VISIBLE_UI_OPACITY}
local FADED_UI_OPACITY = 0.5
local FADED_UI_COLOR = {1.0, 1.0, 1.0, FADED_UI_OPACITY}
local BACKING_UI_OPACITY = 0.3
local BACKING_UI_COLOR = {1.0, 1.0, 1.0, BACKING_UI_OPACITY}
local PLANET_RADIUS = INDICATOR_WIDTH * 0.25
local PLANET_Y = INDICATOR_HEIGHT * 0.9
local BLIP_HEIGHT = INDICATOR_WIDTH * 0.4
local BLIP_WIDTH = INDICATOR_WIDTH * 0.4
local BLIP_FADE_SPEED = 0.9
local BLIP_START_OPACITY = VISIBLE_UI_OPACITY
local BLIP_END_OPACITY = 0.15
local BLIP_DELAY = 0.2
local BLIP_START_Y = INDICATOR_HEIGHT * 0.1
local BLIP_END_Y = INDICATOR_HEIGHT * 0.85
local BLIP_START_OFFSET = -3
local BLIP_MOVE_SPEED = 0.5
local MOON_RADIUS = INDICATOR_WIDTH * 0.08
local MOON_GROUP_Y = INDICATOR_HEIGHT - MOON_RADIUS
local MOON_ORBIT_DISTANCE = INDICATOR_WIDTH * 0.48
local ORBIT_SPEED = 10.0
local ORBIT_AMOUNT = -360
local DASH_WIDTH = 1.5
local DASH_HEIGHT_REDUCED = 0.5
local BACKING_CORNER_RADIUS = 4

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function animateComponents(self)
  local blipResetOpacityAction = AHCallLogic.new(function() self.shipBlip:setOpacity(BLIP_START_OPACITY) end)
  local blipPulseDownAction = AHOpacityTo.new(self.shipBlip, BLIP_FADE_SPEED, nil, BLIP_END_OPACITY)
  local blipPulseSequence = AHSequence.new(blipResetOpacityAction, blipPulseDownAction, AHDelay.new(BLIP_DELAY))
  local blipLoopAction = AHLoop.new(blipPulseSequence)
  self.blipLoopAction = blipLoopAction

  local moonOrbitAction = AHRotationBy.new(self.moonGroup, ORBIT_SPEED, nil, ORBIT_AMOUNT)
  local moonOrbitLoopAction = AHLoop.new(moonOrbitAction)

  blipLoopAction:start()
  moonOrbitLoopAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RocketProgressIndicator:setProgress(progress)
  self.super.setProgress(self, progress)
  progress = self:getProgress()

  local totalProgressHeight = BLIP_END_Y - BLIP_START_Y
  local newBlipY = totalProgressHeight * progress + BLIP_START_Y
  if (progress >= 1.0) then
    newBlipY = PLANET_Y
  end
  local moveBlipAction = AHLocationYTo.new(self.shipBlip, BLIP_MOVE_SPEED, AHEaseSimpleInOut.new(), newBlipY)

  local function moveAndUpdateDashFunction()
    wait(moveBlipAction:start())
    self.dashes[self:getProgressStep()]:setVisible(true)
    if (self:getProgress() >= 1.0) then
      AHScaleTo.new(self.shipBlip, 0.1, nil, 0.0):start()
    end
  end
  runInCoroutine(moveAndUpdateDashFunction)
end

function RocketProgressIndicator:playIntroAnimation()
  local moveBlipAction = AHLocationYTo.new(self.shipBlip, BLIP_MOVE_SPEED, AHEaseSimpleInOut.new(), BLIP_START_Y)
  moveBlipAction:start()
end

-- UIOverlay Widget -------------------

function RocketProgressIndicator:prepareForGameIntro()
  self:setOpacity(0.0)
  self.planet:setOpacity(0.0)
  self.moon:setOpacity(0.0)
  self.shipBlip:setOpacity(0.0)
  self.base:setOpacity(0.0)
end

function RocketProgressIndicator:createFadeInAction(duration)
  local fadeInAction = AHSequence.new(
                                      AHSpawn.new(
                                                  AHOpacityTo.new(self, duration, nil, BACKING_UI_OPACITY),
                                                  AHOpacityTo.new(self.planet, duration, nil, VISIBLE_UI_OPACITY),
                                                  AHOpacityTo.new(self.moon, duration, nil, VISIBLE_UI_OPACITY),
                                                  AHOpacityTo.new(self.shipBlip, duration, nil, VISIBLE_UI_OPACITY),
                                                  AHOpacityTo.new(self.base, duration, nil, VISIBLE_UI_OPACITY)
                                                  ),
                                      function()
                                        animateComponents(self)
                                      end
                                      )
  return fadeInAction
end

function RocketProgressIndicator:prepareForGameOutro()
  if (self.blipLoopAction) then
    self.blipLoopAction:stop()
  end
end

function RocketProgressIndicator:createFadeOutAction(duration)
  local fadeOutActions = {
    AHOpacityTo.new(self, duration, nil, 0.0),
    AHOpacityTo.new(self.planet, duration, nil, 0.0),
    AHOpacityTo.new(self.moon, duration, nil, 0.0),
    AHOpacityTo.new(self.shipBlip, duration, nil, 0.0),
    AHOpacityTo.new(self.base, duration, nil, 0.0)
  }

  for i, dash in ipairs(self.dashes) do
    local dashFadeOutAction = AHOpacityTo.new(dash, duration, nil, 0.0)
    table.insert(fadeOutActions, dashFadeOutAction)
  end

  local fadeOutAction = AHSpawn.new(unpack(fadeOutActions))
  return fadeOutAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createDashTrail(self)
  local stepCount = self:getStepCount()
  local dashHeight = (BLIP_END_Y - BLIP_START_Y) / stepCount
  local adjustedDashHeight = dashHeight * DASH_HEIGHT_REDUCED
  for i = 1, stepCount do
    local dash = FilledRectangle.new(DASH_WIDTH, adjustedDashHeight, FADED_UI_COLOR)
    local dashX = INDICATOR_WIDTH * 0.5
    local dashY = BLIP_START_Y + (i - 1) * dashHeight + adjustedDashHeight * 0.5
    dash:setLoc(dashX, dashY)

    dash:setVisible(false)

    self:insert(dash)
    table.insert(self.dashes, dash)
  end
end

local function createPlanetAndMoon(self)
  local planet = FilledCircle.new(PLANET_RADIUS, VISIBLE_UI_COLOR)
  planet:setLoc(INDICATOR_WIDTH * 0.5, PLANET_Y)
  self:insert(planet)
  self.planet = planet

  local moonGroupWidth = MOON_ORBIT_DISTANCE
  local moonGroupHeight = 2.0 * MOON_RADIUS
  local moonGroup = Group.new(moonGroupWidth, moonGroupHeight)
  moonGroup:setLoc(INDICATOR_WIDTH * 0.5, PLANET_Y)
  moonGroup:setAnchorPointLeft()
  self:insert(moonGroup)
  self.moonGroup = moonGroup

  local moon = FilledCircle.new(MOON_RADIUS, VISIBLE_UI_COLOR)
  moon:setLoc(moonGroupWidth - MOON_RADIUS, moonGroupHeight * 0.5)
  moonGroup:insert(moon)
  self.moon = moon
end

local function createLayout(self)
  local base = FilledRoundedRectangle.new(BASE_WIDTH, BASE_HEIGHT, BASE_CORNER_RADIUS, VISIBLE_UI_COLOR)
  base:setLoc(INDICATOR_WIDTH * 0.5, BASE_START_Y)
  self:insert(base)
  self.base = base

  local shipBlip = FilledTriangle.new(0, 0, BLIP_WIDTH * 0.5, BLIP_HEIGHT, BLIP_WIDTH, 0, VISIBLE_UI_COLOR)
  shipBlip:setLoc(INDICATOR_WIDTH * 0.5, BASE_START_Y + BLIP_HEIGHT * 0.5)
  self:insert(shipBlip)
  self.shipBlip = shipBlip

  createDashTrail(self)
  createPlanetAndMoon(self)
end

-- Constructor ------------------------

function RocketProgressIndicator.newWithStepCount(stepCount)
  local self = FilledRoundedRectangle.new(INDICATOR_WIDTH, INDICATOR_HEIGHT, BACKING_CORNER_RADIUS, BACKING_UI_COLOR)
  ProgressIndicator.makeProgressIndicator(self, stepCount)
  Mixin.mix(self, RocketProgressIndicator)

  self.dashes = {}

  createLayout(self)

  return self
end
