--- RutabagaConfiguration
-- @author Created by Jay Austria on May 16, 2016.

class("RutabagaConfiguration")

require("data.keyframes.KeyframeSet")
require("difficulty.DifficultyUtils")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

RutabagaConfiguration.GAME_DURATION_IN_SECONDS = 45
RutabagaConfiguration.INCORRECT_RESPONSE_TIME_PENALTY = 6
RutabagaConfiguration.POINTS_PER_FULL_SPEED_BONUS = 100
RutabagaConfiguration.SPEED_BONUS_WINDOW_FOR_PART_COUNT = {
  [1] = 2.0,
  [2] = 2.6,
  [3] = 3.2,
  [4] = 3.8,
}
RutabagaConfiguration.SPEED_BONUS_DECAY_START = 1.0
RutabagaConfiguration.BONUS_POINTS_FOR_FOUR_PART_WORD = 10

RutabagaConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES =
  RutabagaConfiguration.GAME_DURATION_IN_SECONDS *
  LearningFocusedRoundGenerator.DEFAULT_REMAINING_BEFORE_RESURFACING

-- Value used in CoreMS's HighDifficultyReachedHighlightProducer, update if modified
local HIGH_DIFFICULTY_THRESHOLD = 0.6

local CORRECT_NEEDED_RANK_2 = 8
local CORRECT_NEEDED_RANK_3 = 11
local CORRECT_NEEDED_THRESHOLDS = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {rank2 = 8, rank3 = 11}),
  Keyframe.new(0.4, {rank2 = 9, rank3 = 12})
)

local POINTS_PER_CORRECT_ANSWER_BY_CONTENT_DIFFICULTY = {
  Easy = 625,
  Medium = 625,
  Hard = 635
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getPointsForCorrectAnswerForRankThreshold()
  return RutabagaConfiguration.getPointsForCorrectAnswer("Easy")
end

local function getRankThresholdsFrame()
  local difficulty = GameSession.getDifficulty()
  return CORRECT_NEEDED_THRESHOLDS:getFrameAt(difficulty)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaConfiguration.getPointsForCorrectAnswer(contentDifficulty)
  local points = POINTS_PER_CORRECT_ANSWER_BY_CONTENT_DIFFICULTY[contentDifficulty]
  assert(points ~= nil, "ContentDifficulty is not associated with a point value: " .. tostring(contentDifficulty))
  return points
end

function RutabagaConfiguration.validateDifficultyCurves()
  local keyFrames = RutabagaConfiguration.getDifficultyKeyFrames()
  local interpolator = KeyFrameInterpolator.new(keyFrames, HIGH_DIFFICULTY_THRESHOLD)
  local hasEasyOutcomes = (interpolator:getWeightForValue("Easy") > 0.0)
  assert(not hasEasyOutcomes, "Current configuration is invalid: Game may try to pull Easy four-part words, which is undefined by Content.")
end

function RutabagaConfiguration.getIsHighLevel()
  return (GameSession.getDifficulty() >= HIGH_DIFFICULTY_THRESHOLD)
end

function RutabagaConfiguration.shouldShowHighDifficultyInterstitial()
  return DifficultyUtils.shouldShowHighDifficultyInterstitial(HIGH_DIFFICULTY_THRESHOLD, GameSession.getDifficulty())
end

function RutabagaConfiguration.getRankTwoScore()
  local correctNeeded = getRankThresholdsFrame().rank2
  local rankTwoScore = getPointsForCorrectAnswerForRankThreshold() * correctNeeded

  return rankTwoScore
end

function RutabagaConfiguration.getRankThreeScore()
  local correctNeeded = getRankThresholdsFrame().rank3
  local rankThreeScore = getPointsForCorrectAnswerForRankThreshold() * correctNeeded

  return rankThreeScore
end

function RutabagaConfiguration.getDifficultyKeyFrames()
  return {
    {key = 0.0, valueProbabilities = {Easy = 0.95,
                                      Medium = 0.05}},

    {key = 0.075, valueProbabilities = {Easy = 0.78,
                                        Medium = 0.22}},

    {key = 0.15, valueProbabilities = {Easy = 0.65,
                                       Medium = 0.30,
                                       Hard = 0.05}},

    {key = 0.225, valueProbabilities = {Easy = 0.50,
                                        Medium = 0.40,
                                        Hard = 0.10}},

    {key = 0.3, valueProbabilities = {Easy = 0.35,
                                      Medium = 0.50,
                                      Hard = 0.15}},

    {key = 0.375, valueProbabilities = {Easy = 0.30,
                                        Medium = 0.40,
                                        Hard = 0.30}},

    {key = 0.45, valueProbabilities = {Easy = 0.15,
                                       Medium = 0.50,
                                       Hard = 0.35}},

    {key = 0.525, valueProbabilities = {Easy = 0.10,
                                        Medium = 0.40,
                                        Hard = 0.50}},

    {key = 0.6, valueProbabilities = {Medium = 0.35,
                                      Hard = 0.65}},

    {key = 0.675, valueProbabilities = {Medium = 0.22,
                                        Hard = 0.78}},

    {key = 0.75, valueProbabilities = {Medium = 0.05,
                                       Hard = 0.95}},

    {key = 1.0, valueProbabilities = {Hard = 1.0}}
  }
end
