--- RutabagaGame
-- @author Created by Jay Austria on May 16, 2016.

class("RutabagaGame")

require("StandardUIBuilder")
require("RutabagaConfiguration")
require("RutabagaRoundGenerator")
require("RutabagaScene")
require("RutabagaNoisyCascadeLayer")
require("RutabagaRoundProgressionTracker")
require("widgets.TapToContinue")
require("bonuses.ResponseSpeedTracker")
require("RutabagaOutroCheckmarkLayer")
require("ExperimentsConfiguration")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {0.36, 0.01, 0.0}
local UI_COMPONENTS_COLOR = {0.98, 0.91, 0.7}
local EXTRA_DELAY_PER_WORD_PART = 0.8 / 3 -- derived from some very specific polish tuning

local function configureGameDifficulty(self)
  local rankTwoScore = RutabagaConfiguration.getRankTwoScore()
  local rankThreeScore = RutabagaConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  RutabagaConfiguration.validateDifficultyCurves()

  local difficultyKeyFrames = RutabagaConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())

  self.baseRoundGenerator = RutabagaPreloadConceptPool.preloadedRoundGenerator
  self.roundGenerator = LearningFocusedRoundGenerator.new(self.baseRoundGenerator)
  self.baseRoundGenerator:setDifficultyNameChooser(difficultyNameChooser)
  self.roundProgressionTracker = RutabagaRoundProgressionTracker.new()

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function allowMistakesToResurfaceIfNeeded(self)
  if (self.timer:getTimeRemaining() < RutabagaConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
    self.roundGenerator:allowMistakesToResurface()
  end
end

local function generateAndPresentRound(self, isFirstRound)
  local partCount = self.roundProgressionTracker:getPartCount()
  self.baseRoundGenerator:setPartCount(partCount)

  local didRoundProgressionPromote = self.roundProgressionTracker:getDidPromote()
  local isFirstPromotion = (didRoundProgressionPromote and partCount < 3)
  local doesRequireShrink = (partCount >= RutabagaCommon.COLUMNS_COUNT_SHRUNKEN)

  if (didRoundProgressionPromote) then
    self.roundGenerator:clearMistakes()
  end

  self.currentRound = self.roundGenerator:next()

  self.responseSpeedTracker:setBonusWindowDuration(RutabagaConfiguration.SPEED_BONUS_WINDOW_FOR_PART_COUNT[partCount])
  local onStartWrapper = function()
    self.responseSpeedTracker:notifyStart()
    self.timer:resume()
  end

  GameSession.contentTracker:notifyPresentedConcept(self.currentRound.identifier)

  wait(self.scene:presentRound(self.currentRound, isFirstRound, didRoundProgressionPromote, isFirstPromotion, onStartWrapper, doesRequireShrink))
end

local function startCorrectResponseAction(self, additionalDelayForReadingResult)
  local correctResponseAction = self.scene:createCorrectResponseAction(additionalDelayForReadingResult)
  correctResponseAction:start()
end

local function waitForIncorrectResponseAction(self, pressedPillar)
  local incorrectResponseAction = self.scene:createIncorrectResponseAction(pressedPillar)
  wait(incorrectResponseAction:start())
end

local function registerCorrectResponse(self, additionalDelayForReadingResult)
  self.timer:pause()

  self.roundProgressionTracker:notifyCorrectResponse()
  self.responseSpeedTracker:notifyCorrectResponse()

  if (not self.hasAlreadyTrackedRound) then
    GameSession.contentTracker:notifyCorrectAnswer(self.currentRound.identifier)
    local hasPostGameRationale = (#self.currentRound.parts > 1)
    if (hasPostGameRationale) then
      GameSession.notifyCorrectAnswer(self.currentRound.identifier)
      GameSession.incrementCorrectAnswerCount()
    end
  end

  GameSession.increaseScore(RutabagaConfiguration.getPointsForCorrectAnswer(self.currentRound.difficulty))
  if (#self.currentRound.parts >= 4) then
    GameSession.increaseScore(RutabagaConfiguration.BONUS_POINTS_FOR_FOUR_PART_WORD)
  end
  self.scoreLabel:setScore(GameSession.getScore())

  startCorrectResponseAction(self, additionalDelayForReadingResult)
end

local function startRationaleDismissAction(self)
  self.timer:resume()

  local rationaleDismissAction = self.scene:createRationaleDismissAction()
  rationaleDismissAction:start()
end

local function registerIncorrectResponse(self)
  self.timer:addTime(RutabagaConfiguration.INCORRECT_RESPONSE_TIME_PENALTY)
  self.timer:pause()

  self.responseSpeedTracker:notifyIncorrectResponse()
  GameSession.contentTracker:notifyIncorrectAnswer(self.currentRound.identifier, nil, self.scene:getAssembledResponse())
  GameSession.incrementIncorrectAnswerCount()

  if (not self.hasAlreadyTrackedRound) then
    self.roundGenerator:notifyMistake(self.currentRound.identifier)

    local hasPostGameRationale = (#self.currentRound.parts > 1)
    if (hasPostGameRationale) then
      GameSession.notifyIncorrectAnswer(self.currentRound.identifier)
    end

    -- Prevent duplicate entries in the post-game report since players get a second
    -- chance to answer correctly after answering incorrectly
    self.hasAlreadyTrackedRound = true
  end

  waitForIncorrectResponseAction(self)

  local fadeDuration = 0.2 -- From UIOverlay's default fade durations
  local fadedOpacity = 0.0
  local fadeOutUiAction = AHDelayed.new(
    0.1,
    AHOpacityTo.new(self.pauseButton, fadeDuration, nil, fadedOpacity),
    AHOpacityTo.new(self.scoreLabel, fadeDuration, nil, fadedOpacity),
    self.timer:pauseAndShowIcon(fadeDuration)
  )
  self.uiFadeInterrupt:insert(fadeOutUiAction)

  local tapToContinue = TapToContinue.new()
  tapToContinue:waitForTrigger()
  tapToContinue:removeSelf()

  local fadeInUiAction = AHSpawn.new(
    AHOpacityTo.new(self.pauseButton, fadeDuration, nil, 1.0),
    AHOpacityTo.new(self.scoreLabel, fadeDuration, nil, 1.0),
    self.timer:resumeAndHideIcon(fadeDuration)
  )
  self.uiFadeInterrupt:insert(fadeInUiAction)

  startRationaleDismissAction(self)
end

local function registerResponse(self, isComplete, isCorrect)
  if (isComplete) then
    if (isCorrect) then
      self.isWaitingForRoundComplete = false
    else
      registerIncorrectResponse(self)
    end
  end
end

local function registerTimeOut(self)
  if (GameSession.getScore() <= 0) then
    GameSession.notifyGameLoss()
  else
    local outroAnimation = AHSpawn.new()
    outroAnimation:addAction(self.scene:prepareForAndCreateOutro())
    outroAnimation:addAction(function()
      self.uiFadeInterrupt:insert(self.uiOverlay:createFadeOutAction())
    end)

    local outro = RutabagaOutroCheckmarkLayer.new()
    self:insertAtCenter(outro)
    outroAnimation:addAction(function()
      outro:startAnimation()
    end)

    local durationOfOutro = 4.5
    outroAnimation:addAction(AHDelay.new(durationOfOutro))


    local showFinalAnswerAction = self.scene:createShowAnswerForTimeOutAction()
    local fullAction = AHSpawn.new(
      showFinalAnswerAction,
      AHDelayed.new(2.0, outroAnimation)
    )

    wait(fullAction:start())

    GameSession.notifyGameWin()
  end
end

local function showAndWaitForHighLevelInterstitial(self)
  wait(self.scene:showHighLevelInterstitial())
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaGame:isUIAutoFadeInEnabled()
  return false
end

function RutabagaGame:gameFlow()
  local isFirstRound = true

  self.baseRoundGenerator:prepareOnePartWords()

  self.timer:initialize()

  if RutabagaConfiguration.shouldShowHighDifficultyInterstitial() then
    showAndWaitForHighLevelInterstitial(self)
  end

  self.uiOverlay:fadeIn()

  while (true) do
    runInCoroutine(function()
      allowMistakesToResurfaceIfNeeded(self)
      generateAndPresentRound(self, isFirstRound)
      isFirstRound = false
    end)

    self.isWaitingForRoundComplete = true
    self.hasAlreadyTrackedRound = false

    waitWhile(function()
      return (self.isWaitingForRoundComplete == true)
    end)

    local additionalDelayForReadingResult = (self.roundProgressionTracker:getPartCount() - 1) * EXTRA_DELAY_PER_WORD_PART

    registerCorrectResponse(self, additionalDelayForReadingResult)
    waitFor(0.8 + additionalDelayForReadingResult)
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)

  local scene = RutabagaScene.new()
  self:insertAtCenter(scene)
  self.scene = scene

  local responseWrapper = function(isComplete, isCorrect)
    runInCoroutine(function()
      registerResponse(self, isComplete, isCorrect)
    end)
  end
  scene:setResponseCallback(responseWrapper)

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 RutabagaConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(RutabagaConfiguration.SPEED_BONUS_WINDOW_FOR_PART_COUNT[1],
                                                       quickResponseWrapper,
                                                       RutabagaConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.timer = builder:addDigitTimer(RutabagaConfiguration.GAME_DURATION_IN_SECONDS)

  local registerTimeOutWrapper = function()
    registerTimeOut(self)
  end
  self.timer:setTimeoutHandler(registerTimeOutWrapper)

  local polishedBonusTrayColor = {
    UI_COMPONENTS_COLOR[1],
    UI_COMPONENTS_COLOR[2],
    UI_COMPONENTS_COLOR[3],
    0.6
  }
  self.bonusTray:setKeyColor(polishedBonusTrayColor)
end

-- Constructor ------------------------

function RutabagaGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, RutabagaGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.isWaitingForRoundComplete = false
  self.hasAlreadyTrackedRound = false
  self.currentRound = nil
  self.uiFadeInterrupt = ActionInterrupt.new()

  return self
end
