--- RutabagaNoisyCascadeLayer
-- @author Created by Jay Austria on May 24, 2016.

class("RutabagaNoisyCascadeLayer")

require("graphics.FlintstonesImage")
require("graphics.OffscreenRenderedTexture")
require("graphics.MultiGradient")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SCALES_TINT_COLOR = {1.0, 0.93, 0.23}
local SHAPES_COLOR = {0.98, 0.91, 0.7}

local GRADIENT_COLORS = {
  {1.0, 0.93, 0.23},
  {0.99, 0.49, 0.08},
  {0.97, 0.23, 0.0},
  {0.45, 0.09, 0.09}
}

local GRADIENT_THICKNESSES = {
  0.1,
  0.45,
  0.54
}
local RECTANGLE_DENSITY_FACTOR = 0.5
local LINE_DENSITY_FACTOR = 0.05
local RECTANGLE_WIDTH = 15

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaNoisyCascadeLayer:release()
  self.updateCoroutine:stop()
  self.updateCoroutine = nil
  self.offscreenTexture:release()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createRandomRectangle(self)
  local possibleHeights = {20, 30, 60}
  local height = TableUtils.getRandomElement(possibleHeights)
  local possibleOpacities = {0.05, 0.1, 0.2, 0.3, 0.4}
  local baseOpacity = TableUtils.getRandomElement(possibleOpacities)
  local rectangle = FilledRectangle.new(RECTANGLE_WIDTH, height, SHAPES_COLOR)

  local minVelocity = 0.1
  local maxVelocity = 1.5
  rectangle.velocity = MathUtils.getRandomFloat(minVelocity, maxVelocity)

  rectangle:setBlendMode(MOAIProp2D.BLEND_ADD)
  rectangle.baseOpacity = baseOpacity
  rectangle:setOpacity(baseOpacity)

  local randomX = self:getWidthPart(math.random())
  local randomY = self:getHeightPart(math.random())
  rectangle:setLoc(randomX, randomY)

  rectangle.framesToHide = nil

  return rectangle
end

local function createRandomLine(self)
  local width = self:getWidthPart(0.5)
  local height = 1
  local baseOpacity = math.random()
  local line = FilledRectangle.new(width, height, SHAPES_COLOR)

  local minVelocity = 0.1
  local maxVelocity = 2.0
  line.velocity = MathUtils.getRandomFloat(minVelocity, maxVelocity)

  line:setBlendMode(MOAIProp2D.BLEND_ADD)
  line.baseOpacity = baseOpacity
  line:setOpacity(baseOpacity)

  local randomX = self:getWidthPart(math.random())
  local randomY = self:getHeightPart(math.random())
  line:setLoc(randomX, randomY)

  line.framesToHide = nil

  return line
end

local function createLayout(self)
  local scaleMultiplier = 1
  local scales = FlintstonesImage.new("tiles/scales.png", self:getHeight(), self:getWidth(), scaleMultiplier)
  scales:setBlendMode(MOAIProp2D.BLEND_ADD)
  scales:setColor(unpack(SCALES_TINT_COLOR))
  scales:setRot(90)
  scales:playCycle(0.5)
  self:insertAtCenter(scales)

  local canvas = Group.new(self:getWidth(), self:getHeight())
  self.canvas = canvas

  local rectangleCount = math.ceil(self:getWidth() * RECTANGLE_DENSITY_FACTOR)
  local lineCount = math.ceil(self:getWidth() * LINE_DENSITY_FACTOR)

  local slidingShapes = {}
  for rectangleIndex = 1, rectangleCount do
    local newShape = createRandomRectangle(self)
    canvas:insert(newShape)

    table.insert(slidingShapes, newShape)
  end

  for lineIndex = 1, lineCount do
    local newShape = createRandomLine(self)
    canvas:insert(newShape)

    table.insert(slidingShapes, newShape)
  end

  self.updateCoroutine = runInCoroutine(function()
    while (true) do
      local resetWidth = canvas:getWidth() + RECTANGLE_WIDTH
      for _, shape in ipairs(slidingShapes) do
        local x = shape:getX() - shape.velocity
        local isOutside = ((x + shape:getWidthPart(0.5)) <= 0)

        if (isOutside) then
          x = x + resetWidth
        end
        shape:setX(x)

        if (shape.framesToHide == nil) then
          if (math.random() < 0.025) then
            shape:setOpacity(0.0)
            shape.framesToHide = 3
          end
        elseif (shape.framesToHide > 0) then
          shape.framesToHide = shape.framesToHide - 1
        else
          shape:setOpacity(shape.baseOpacity)
        end
      end
      coroutine.yield()
    end
  end)

  local offscreenTexture = OffscreenRenderedTexture.new(canvas, nil, nil)
  self.offscreenTexture = offscreenTexture
  local renderedCanvas = Sprite.newWithTexture(offscreenTexture)
  self:insertAtCenter(renderedCanvas)
end

-- Constructor ------------------------

function RutabagaNoisyCascadeLayer.new(width, height)
  local isHorizontal = false
  local self = MultiGradient.new(width, height, isHorizontal, GRADIENT_COLORS, GRADIENT_THICKNESSES)

  Mixin.mix(self, RutabagaNoisyCascadeLayer)

  createLayout(self)

  return self
end
