--- RutabagaOutroCheckmarkLayer
-- @author Created by Joe Poka on May 25, 2015.

class("RutabagaOutroCheckmarkLayer")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PIECE_ROTATION_DEGREES = 45.0
local CENTER_PIECE_COLOR = {0.93, 0.44, 0.16}
local LEFT_PIECE_COLOR = {0.93, 0.51, 0.24}
local RIGHT_PIECE_COLOR = {0.89, 0.32, 0.15}
local WHITE_COLOR = {1.0, 1.0, 1.0, 1.0}
local PIECE_GAP = Screen.getWidthPart(0.0025)
local ANIMATION_DURATION = 1.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createPieces(self, dimension)
  -- just an eye-balled position
  local centerPieceX = self:getWidthPart(0.4)
  local centerPieceY = self:getHeightPart(0.43)

  local pieceCenter = FilledRectangle.new(dimension, dimension, CENTER_PIECE_COLOR)
  local pieceLeft = FilledRectangle.new(dimension * 1.5, dimension, LEFT_PIECE_COLOR)
  local pieceRight = FilledRectangle.new(dimension, dimension * 2.5, RIGHT_PIECE_COLOR)

  self.pieces = {
    center = pieceCenter,
    left = pieceLeft,
    right = pieceRight
  }

  self.pieceDestinationPositions = {
    center = {
      x = centerPieceX,
      y = centerPieceY
    }
  }

  local thetaRadians = math.rad(PIECE_ROTATION_DEGREES)
  local length = pieceCenter:getWidthPart(0.5) + pieceLeft:getWidthPart(0.5)
  local dx = -length * math.cos(thetaRadians) - PIECE_GAP
  local dy = length * math.sin(thetaRadians) + PIECE_GAP

  self.pieceDestinationPositions.left = {
    x = centerPieceX + dx,
    y = centerPieceY + dy
  }

  length = pieceCenter:getWidthPart(0.5) + pieceRight:getHeightPart(0.5)
  dx = length * math.cos(thetaRadians) + PIECE_GAP
  dy = length * math.sin(thetaRadians) + PIECE_GAP

  self.pieceDestinationPositions.right = {
    x = centerPieceX + dx,
    y = centerPieceY + dy
  }
end

local function createPolygons(self, dimension)
  local centerPos = self.pieceDestinationPositions.center

  local smallDimension = dimension / math.sqrt(2.0)
  local crossDimension = dimension * math.sqrt(2.0)

  local polygonHeight = Screen.fullHeight - centerPos.y
  local polygonY = polygonHeight * 0.5

  -- for left piece polygons
  local leftPieceWidth = self.pieces.left:getWidth()
  local leftSideDimension = leftPieceWidth / math.sqrt(2.0)

  -- for right piece polygons
  local rightPieceHeight = self.pieces.right:getHeight()
  local rightSideDimension = rightPieceHeight / math.sqrt(2.0)

  local polygonData = {
    centerRightPolygon = {
      color = {0.76, 0.34, 0.1},
      width = smallDimension,
      offset = {
        x = smallDimension * 0.5,
        y = polygonY
      }
    },
    centerLeftPolygon = {
      color = {0.76, 0.34, 0.1},
      width = smallDimension,
      offset = {
        x = -smallDimension * 0.5,
        y = polygonY
      }
    },
    leftRightPolygon = {
      color = {0.87, 0.43, 0.19},
      width = leftSideDimension,
      offset = {
        x = -leftSideDimension * 0.5 - PIECE_GAP,
        y = crossDimension * 0.5 + polygonY + PIECE_GAP
      }
    },
    leftLeftPolygon = {
      color = {0.88, 0.47, 0.22},
      width = smallDimension,
      offset = {
        x = -leftSideDimension - smallDimension * 0.5 - PIECE_GAP,
        y = leftSideDimension + polygonHeight * 0.5 + PIECE_GAP
      }
    },
    rightLeftPolygon = {
      color = {0.73, 0.23, 0.07},
      width = rightSideDimension,
      offset = {
        x = rightSideDimension * 0.5 + PIECE_GAP,
        y = polygonY + crossDimension * 0.5 + PIECE_GAP
      }
    },
    rightRightPolygon = {
      color = {0.63, 0.19, 0.07},
      width = smallDimension,
      offset = {
        x = rightSideDimension + smallDimension * 0.5 + PIECE_GAP,
        y = rightSideDimension + polygonHeight * 0.5 + PIECE_GAP
      }
    }
  }

  -- create polygons and calculate destination positions
  self.polygons = {}
  self.polygonDestinationPositions = {}

  for key, data in pairs(polygonData) do
    local polygonWidth = data.width
    local offset = data.offset

    local lowerRightDeltaY = 0.0
    if (key == "rightLeftPolygon") then
      lowerRightDeltaY = polygonWidth
    end

    local points = {
      {x = polygonWidth, y = 0.0 + lowerRightDeltaY}, -- lower right
      {x = 0.0, y = 0.0},                             -- lower left
      {x = 0.0, y = polygonHeight},                   -- upper left
      {x = polygonWidth, y = polygonHeight}           -- upper right
    }
    local color = data.color
    local polygon = FilledPolygon.newAutomatic(points, color)

    self.polygons[key] = polygon
    self.polygonDestinationPositions[key] = {
      x = centerPos.x + offset.x,
      y = centerPos.y + offset.y
    }
  end
end

local function fadeTo(objects, color, duration)
  for _, object in pairs(objects) do
    local colorTo = AHColorTo.new(object, duration, nil, color)
    colorTo:start()
  end
end

local function exitAnimation(self)
  local duration = ANIMATION_DURATION

  local exitDelays = {
    right = 0.0,
    center = duration * 0.12,
    left = duration * 0.24
  }

  local animateOut = function(object, destination, delay)
    local easeOut = AHEaseExponentialIn.new()
    local move = AHLocationTo.new(object, duration, easeOut, destination.x, destination.y)
    local delayedMove = AHDelayed.new(delay, move)
    delayedMove:start()
  end

  -- animate out checkmark pieces
  for key, piece in pairs(self.pieces) do
    local destination = self.pieceInitialPositions[key]
    local delay = exitDelays[key]
    animateOut(piece, destination, delay)
  end

  -- animate out noisy layers
  for key, layer in pairs(self.noisyLayers) do
    local destination = self.polygonInitialPositions[key]
    local delay = exitDelays[self.polygonDelayKeys[key]]
    animateOut(layer, destination, delay)
  end
end

local function createNoisyLayers(self)
  -- adjust layer widths for right checkmark piece
  -- helps avoid layer sticking out below checkmark
  local widePolygon = self.polygons.rightLeftPolygon
  local widePolygonWidth = widePolygon:getWidth()
  local widthAdjustment = widePolygonWidth * 0.3

  self.noisyLayers = {}
  for key, polygon in pairs(self.polygons) do
    local width, height = polygon:getDims()

    -- wide right polygon shrinks + thin right polygon expands
    local adjustment = 0.0
    if (key == "rightLeftPolygon") then
      adjustment = -widthAdjustment
    elseif (key == "rightRightPolygon") then
      adjustment = widthAdjustment
    end

    width = width + adjustment
    local noisyLayer = RutabagaNoisyCascadeLayer.new(width, height)
    noisyLayer:setScl(1, -1)

    self:insert(noisyLayer, self.zOrder.noisyLayer)
    self.noisyLayers[key] = noisyLayer

    local position = self.polygonDestinationPositions[key]
    noisyLayer:setLoc(position.x - math.abs(adjustment * 0.5), position.y)
    noisyLayer:setVisibleRecursive(false)
  end
end

local function createBorders(self)
  local borderThickness = 2

  local leftBorderHeight = self.noisyLayers.leftLeftPolygon:getHeight()
  local leftBorder = FilledRectangle.new(borderThickness, leftBorderHeight, WHITE_COLOR)
  leftBorder:setAnchorPointBottomLeft()
  self.noisyLayers.leftLeftPolygon:insert(leftBorder)

  local rightBorderHeight = self.noisyLayers.rightRightPolygon:getHeight()
  local rightBorder = FilledRectangle.new(borderThickness, rightBorderHeight, WHITE_COLOR)
  rightBorder:setAnchorPointBottomRight()
  rightBorder:setX(self.noisyLayers.rightRightPolygon:getWidth())
  self.noisyLayers.rightRightPolygon:insert(rightBorder)
end

local function removePolygons(self)
  for _, polygon in pairs(self.polygons) do
    polygon:removeSelf()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaOutroCheckmarkLayer:startAnimation()
  Sound.playSoundEffect("outro")

  local duration = ANIMATION_DURATION

  local entranceDelays = {
    left = 0.0,
    center = duration * 0.12,
    right = duration * 0.24
  }

  -- helps match polygon entrance/exit delays to checkmark piece delays
  self.polygonDelayKeys = {
    centerLeftPolygon = "center",
    centerRightPolygon = "center",
    leftRightPolygon = "left",
    leftLeftPolygon = "left",
    rightLeftPolygon = "right",
    rightRightPolygon = "right"
  }

  local animateIn = function(object, destination, delay)
    local easeOut = AHEaseExponentialOut.new()
    local move = AHLocationTo.new(object, duration, easeOut, destination.x, destination.y)
    local delayedMove = AHDelayed.new(delay, move)
    delayedMove:start()
  end

  -- animate in checkmark pieces
  for key, piece in pairs(self.pieces) do
    local destination = self.pieceDestinationPositions[key]
    local delay = entranceDelays[key]
    animateIn(piece, destination, delay)
  end

  -- animate in polygon pieces
  for key, polygon in pairs(self.polygons) do
    local destination = self.polygonDestinationPositions[key]
    local delay = entranceDelays[self.polygonDelayKeys[key]]
    animateIn(polygon, destination, delay)
  end

  -- fade, etc once everything is in
  local maxDelay = 0.0
  for _, delay in pairs(entranceDelays) do
    maxDelay = math.max(maxDelay, delay)
  end

  local fadeInOut = AHCallLogic.new(function()
    -- fade checkmark and polygons to white
    local fadeInDuration = 0.16
    fadeTo(self.pieces, WHITE_COLOR, fadeInDuration)
    fadeTo(self.polygons, WHITE_COLOR, fadeInDuration)

    -- wait, show noisy layers, fade out polygons, wait, then remove checkmark
    local fadeOut = AHDelayed.new(fadeInDuration, function()
      createBorders(self)

      for _, layer in pairs(self.noisyLayers) do
        layer:setVisibleRecursive(true)
      end

      local fadeOutDuration = fadeInDuration * 5.6
      local transparentColor = {1.0, 1.0, 1.0, 0.0}
      fadeTo(self.polygons, transparentColor, fadeOutDuration)

      local delayBeforeExit = fadeOutDuration
      local cleanupCheckmark = AHDelayed.new(delayBeforeExit, function()
        removePolygons(self)
        exitAnimation(self)
      end)
      cleanupCheckmark:start()
    end)
    fadeOut:start()
  end)

  local totalDelayTime = (duration + maxDelay)
  local sequence = AHDelayed.new(totalDelayTime, fadeInOut)
  sequence:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local width = Screen.getWidthPart(0.333)
  local height = Screen.fullHeight
  self:setDims(width, height)

  self.zOrder = {
    noisyLayer = -1,
    centerRightPolygon = 0,
    centerLeftPolygon = 1,
    center = 2,
    leftRightPolygon = 3,
    leftLeftPolygon = 4,
    left = 5,
    rightLeftPolygon = 6,
    rightRightPolygon = 7,
    right = 8,
  }

  -- dimension of square center checkmark piece
  local dimension = Screen.getWidthPart(0.1)
  createPieces(self, dimension)
  createPolygons(self, dimension)
  createNoisyLayers(self)

  -- insert and z-order checkmark pieces
  for key, piece in pairs(self.pieces) do
    self:insert(piece, self.zOrder[key])
    piece:setRot(-PIECE_ROTATION_DEGREES)
  end

  -- insert and z-order polygons
  for key, polygon in pairs(self.polygons) do
    self:insert(polygon, self.zOrder[key])
  end

  self.pieceInitialPositions = {}
  self.polygonInitialPositions = {}

  -- set initial checkmark piece positions
  for key, piece in pairs(self.pieces) do
    local position = self.pieceDestinationPositions[key]
    local initialX = position.x
    local initialY = position.y + Screen.fullHeight
    piece:setLoc(initialX, initialY)
    self.pieceInitialPositions[key] = {
      x = initialX,
      y = initialY
    }
  end

  -- set initial polygon postions
  for key, polygon in pairs(self.polygons) do
    local position = self.polygonDestinationPositions[key]
    local initialX = position.x
    local initialY = position.y + Screen.fullHeight
    polygon:setLoc(initialX, initialY)

    self.polygonInitialPositions[key] = {
      x = initialX,
      y = initialY
    }
  end
end

function RutabagaOutroCheckmarkLayer.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, RutabagaOutroCheckmarkLayer)

  createLayout(self)

  return self
end
