--- RutabagaPartsStackColumn
-- @author Created by Jay Austria on May 18, 2015.

class("RutabagaPartsStackColumn")

require("RutabagaPartPillar")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PILLAR_PADDING_Y = Screen.getHeightPart(0.01)
local PILLAR_COUNT = 2
local PILLAR_PRESENT_TRANSLATION_Y = RutabagaCommon.COLUMN_HEIGHT * 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getOppositePillar(self, pillar)
  if (pillar == self.pillars[1]) then
    return self.pillars[2]
  else
    return self.pillars[1]
  end
end

local function activateAndDeactivePillars(pressedPillar, oppositePillar)
  if (not pressedPillar:getActivated()) then
    pressedPillar:flagForActivatation()
  end

  oppositePillar:flagForDeactivation()
end

local function registerPillarPress(self, pressedPillar)
  if (self.responseCallback) then
    local oppositePillar = getOppositePillar(self, pressedPillar)

    activateAndDeactivePillars(pressedPillar, oppositePillar)

    self.responseCallback(pressedPillar)
  end
end

local function initializeTouch(self, pillar)
  local onPressWrapper = function(pressedPillar)
    registerPillarPress(self, pressedPillar)
  end
  pillar:setOnPressCallback(onPressWrapper)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaPartsStackColumn:setTouchEnabled(isEnabled)
  for _, pillar in ipairs(self.pillars) do
    if (pillar:getIsTouchAllowed()) then
      pillar:setTouchEnabled(isEnabled)
    else
      pillar:setTouchEnabled(false)
    end
  end
end

function RutabagaPartsStackColumn:updateVisualsSimple()
  for _, pillar in ipairs(self.pillars) do
    if (pillar:getWillActivate()) then
      local topPillarY = nil
      local isLastTap = false
      local onActivateWrapper = function()
        pillar:setIsActivated(true)
        pillar:setVisualState(RutabagaCommon.STATES.ACTIVATED)
      end
      pillar:playActivateAction(topPillarY, isLastTap, onActivateWrapper)
    elseif (pillar:getWillDeactivate()) then
      pillar:setVisualState(RutabagaCommon.STATES.UNACTIVATED)
      pillar:setIsActivated(false)
    end

    pillar:resetActivateFlags()
  end
end

function RutabagaPartsStackColumn:updateVisualsAfterRationale()
  local correctPillar = self.correctPillar
  local incorrectPillar = getOppositePillar(self, correctPillar)

  for _, pillar in ipairs(self.pillars) do
    if (pillar:getWasPreviouslyActivated()) then
      if (pillar ~= incorrectPillar) then
        pillar:setVisualState(RutabagaCommon.STATES.ACTIVATED)
      end

      local shouldReturnActivated = (pillar == correctPillar)
      pillar:setIsActivated(shouldReturnActivated)
    else
      pillar:setVisualState(RutabagaCommon.STATES.UNACTIVATED)
      pillar:setIsActivated(false)
    end

    pillar:resetActivateFlags()
  end
end

function RutabagaPartsStackColumn:updateVisualsIncorrect()
  local correctPillar = self.correctPillar
  local incorrectPillar = getOppositePillar(self, correctPillar)

  for _, pillar in ipairs(self.pillars) do
    local isActivatedOrWillActivate = (pillar:getActivated() or pillar:getWillActivate())
    pillar:setWasPreviouslyActivated(isActivatedOrWillActivate)

    if(pillar:getWillActivate()) then
      local topPillarY = nil
      local isLastTap = false
      local onActivateWrapper = nil
      pillar:playActivateAction(topPillarY, isLastTap, onActivateWrapper)
    end

    if (pillar == incorrectPillar and isActivatedOrWillActivate) then
      pillar:setVisualState(RutabagaCommon.STATES.PAUSED_INCORRECT)
    elseif (not pillar:getActivated() and not pillar:getWillActivate()) then
      pillar:setVisualState(RutabagaCommon.STATES.PAUSED_UNACTIVATED)
    else
      pillar:setVisualState(RutabagaCommon.STATES.ACTIVATED)
    end
  end
end

function RutabagaPartsStackColumn:updateVisualsCorrect()
  local topPillarY = self.topPillarY

  local isLastTap = true
  for _, pillar in ipairs(self.pillars) do
    if (pillar:getWillActivate()) then
      local onActivateWrapper = function()
        pillar:setIsActivated(true)
        pillar:setVisualState(RutabagaCommon.STATES.ACTIVATED)
      end
      pillar:playActivateAction(topPillarY, isLastTap, onActivateWrapper)
    elseif(pillar:getActivated()) then
      pillar:playRepositionAction(topPillarY, isLastTap)
    end

    pillar:resetActivateFlags()
  end
end

function RutabagaPartsStackColumn:hasActivatedCorrectPillar()
  return (self.correctPillar:getActivated() or self.correctPillar:getWillActivate())
end

function RutabagaPartsStackColumn:hasActivatedPillar()
  for _, pillar in ipairs(self.pillars) do
    if (pillar:getActivated() or pillar:getWillActivate()) then
      return true
    end
  end

  return false
end

function RutabagaPartsStackColumn:setResponseCallback(callback)
  self.responseCallback = callback
end

function RutabagaPartsStackColumn:establishPillarsHomeY()
  for _, pillar in ipairs(self.pillars) do
    pillar:establishHomeY()
  end
end

function RutabagaPartsStackColumn:createPresentAction()
  local lowerPillar = self.pillars[2]

  local initializeTranslation = function()
    local loweredY = lowerPillar:getY() - PILLAR_PRESENT_TRANSLATION_Y
    lowerPillar:setY(loweredY)
  end

  local presentAction = AHSequence.new(
    initializeTranslation,
    AHLocationYBy.new(lowerPillar, 1.2, AHEaseExponentialOut.new(5.0), PILLAR_PRESENT_TRANSLATION_Y)
  )

  return presentAction
end

function RutabagaPartsStackColumn:getCorrectTextLabel()
  return self.correctPillar:getPartLabel()
end

function RutabagaPartsStackColumn:getIsCorrectPillarHighlighted()
  return (self.correctPillar:getActivated() or self.correctPillar:getWillActivate())
end

function RutabagaPartsStackColumn:getIncorrectPartAndDefinition()
  local incorrectPillar = getOppositePillar(self, self.correctPillar)

  local partAndDefinition = {
    part = incorrectPillar:getPartText(),
    definition = incorrectPillar:getPartDefinition()
  }

  return partAndDefinition
end

function RutabagaPartsStackColumn:disallowInputOnIncorrectPillar()
  local incorrectPillar = getOppositePillar(self, self.correctPillar)
  incorrectPillar:setIsTouchAllowed(false)
end

function RutabagaPartsStackColumn:disallowAllInput()
  for _, pillar in ipairs(self.pillars) do
    pillar:setIsTouchAllowed(false)
  end
end

function RutabagaPartsStackColumn:createExitAction()
  local incorrectPillar = getOppositePillar(self, self.correctPillar)

  return AHLocationYBy.new(incorrectPillar, 0.6, AHEaseExponentialIn.new(0.5), -incorrectPillar:getHeightPart(1.3))
end

function RutabagaPartsStackColumn:displayForPromotedPresent()
  for _, pillar in ipairs(self.pillars) do
    pillar:displayForPromotedPresent()
  end
end

function RutabagaPartsStackColumn:createPromotedPresentAction()
  local promotedPresentAction = AHSpawn.new()
  for _, pillar in ipairs(self.pillars) do
    promotedPresentAction:addAction(pillar:createPromotedPresentAction())
  end

  return promotedPresentAction
end

function RutabagaPartsStackColumn:getHighlightedPartText()
  for _, pillar in ipairs(self.pillars) do
    if (pillar:getActivated() or pillar:getWillActivate()) then
      return pillar:getPartText()
    end
  end

  return nil
end

function RutabagaPartsStackColumn:highlightCorrectPillar()
  local oppositePillar = getOppositePillar(self, self.correctPillar)
  activateAndDeactivePillars(self.correctPillar, oppositePillar)
  self:updateVisualsCorrect()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, partsDescriptor, columnColorDescriptor)
  local pillarWidth = self:getWidth()
  local buttonHeight = RutabagaCommon.COLUMN_BUTTON_HEIGHT
  local pillarOffsetY = buttonHeight + PILLAR_PADDING_Y

  local isTopPillarCorrect = MathUtils.getRandomBoolean()

  local correctPillar = nil
  local pillars = {}
  for pillarIndex = 1, PILLAR_COUNT do
    local isDistractor = ((isTopPillarCorrect and (pillarIndex == 2)) or
                         (not isTopPillarCorrect) and (pillarIndex == 1))

    local pillarText = nil
    local partDefinition = nil
    if (isDistractor) then
      pillarText = partsDescriptor.distractor
      partDefinition = partsDescriptor.distractorDefinition
    else
      pillarText = partsDescriptor.prompt
      partDefinition = partsDescriptor.definition
    end

    local buttonColor = columnColorDescriptor.BUTTON_COLOR

    local newPillar = RutabagaPartPillar.new(pillarWidth, pillarText, partDefinition, columnColorDescriptor, buttonColor)
    newPillar:setAnchorPointTop()
    self:insert(newPillar)

    local pillarY = self:getHeight() - (pillarOffsetY * (pillarIndex - 1))
    newPillar:setLoc(self:getWidthPart(0.5), pillarY)

    initializeTouch(self, newPillar)

    table.insert(pillars, newPillar)

    local isCorrectPillar = ((isTopPillarCorrect and (pillarIndex == 1)) or
                            (not isTopPillarCorrect) and (pillarIndex == 2))
    if (isCorrectPillar) then
      correctPillar = newPillar
    end
  end

  self.pillars = pillars
  self.topPillarY = pillars[1]:getY()
  self.correctPillar = correctPillar
end

function RutabagaPartsStackColumn.newWithColumnDescriptor(partsDescriptor, colorDescriptor)
  local width = RutabagaCommon.COLUMN_WIDTH

  local self = Group.new(width, RutabagaCommon.COLUMN_HEIGHT)
  Mixin.mix(self, RutabagaPartsStackColumn)

  createLayout(self, partsDescriptor, colorDescriptor)

  return self
end
