--- RutabagaPartsStackLayer
-- @author Created by Jay Austria on May 18, 2015.

class("RutabagaPartsStackLayer")

require("RutabagaPartsStackColumn")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLUMN_PRESENT_STAGGER_DURATION = 0.1
local COLUMN_PADDING_X = Screen.getHeightPart(0.01)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getColumnsLocationXs(columnsCount)
  local locationXs = {}

  local columnWidth = RutabagaCommon.COLUMN_WIDTH
  local offsetX = (columnWidth + COLUMN_PADDING_X)
  local x = nil
  for columnIndex = 1, columnsCount do
    x = Screen.getSafeX(0.5) - ((columnsCount - 1) * offsetX * 0.5) + ((columnIndex - 1) * offsetX)
    table.insert(locationXs, x)
  end

  return locationXs
end

local function getColumnsLocationXsForPromotedPresent(columnsCount)
  local columnWidth = RutabagaCommon.COLUMN_WIDTH
  local oneFewerColumnsCount = columnsCount - 1
  local locationXs = getColumnsLocationXs(oneFewerColumnsCount)

  local leftX = locationXs[1]
  local rightX = locationXs[#locationXs]
  local spanX = rightX - leftX

  local promotedLocationXs = {}
  local x = leftX
  for columnIndex = 1, columnsCount do
    table.insert(promotedLocationXs, x)

    x = x + (spanX / (columnsCount - 1))
  end

  return promotedLocationXs
end

local function didUserRespond(self)
  for _, column in ipairs(self.columns) do
    if (not column:hasActivatedPillar()) then
      return false
    end
  end

  return true
end

local function areAllColumnsCorrect(self)
  for _, column in ipairs(self.columns) do
    if (not column:hasActivatedCorrectPillar()) then
      return false
    end
  end

  return true
end

local function updateColumnsVisualsSimple(self)
  Sound.playSoundEffect("active_tap")
  for _, column in ipairs(self.columns) do
    column:updateVisualsSimple()
  end
end

local function updateColumnsVisualsIncorrect(self)
  Sound.playSoundEffect("incorrect")
  for _, column in ipairs(self.columns) do
    column:updateVisualsIncorrect()
  end
end

local function updateColumnsVisualsCorrect(self)
  Sound.playSoundEffect("correct")
  for _, column in ipairs(self.columns) do
    column:updateVisualsCorrect()
  end
end

local function registerResponse(self)
  if (self.responseCallback ~= nil) then
    local isPartsStackComplete = didUserRespond(self)
    local isCorrect = areAllColumnsCorrect(self)

    if (not isPartsStackComplete) then
      updateColumnsVisualsSimple(self)
    else
      self:setColumnsTouchEnabled(false)
      if (isCorrect) then
        updateColumnsVisualsCorrect(self)
      else
        updateColumnsVisualsIncorrect(self)
      end
    end

    self.responseCallback(isPartsStackComplete, isCorrect)
  end
end

local function cacheColumnsInitialLocations(self)
  for _, column in ipairs(self.columns) do
    local x, y = column:getLoc()

    self.initialLocationsForColumns[column] = {
      x = x,
      y = y
    }

    column:establishPillarsHomeY()
  end
end

local function repositionColumnsBy(self, offset)
  for _, column in ipairs(self.columns) do
    local newY = self.initialLocationsForColumns[column].y - offset
    column:setY(newY)
  end
end

local function realignColumnsForPromotedPresent(self)
  local columnsCount = #self.columns

  local realignedXs = getColumnsLocationXsForPromotedPresent(columnsCount)
  for columnIndex, column in ipairs(self.columns) do
    local realignedX = realignedXs[columnIndex]
    column:setLoc(realignedX, column:getY())
  end
end

local function displayColumnsForPromotedPresent(self)
  for _, column in ipairs(self.columns) do
    column:displayForPromotedPresent()
  end
end

local function createStandardPresentAction(self)
  local presentAction = AHSpawn.new()

  for columnIndex, column in ipairs(self.columns) do
    local combinedAction = AHSpawn.new(
      column:createPresentAction(),
      AHLocationYTo.new(column, 1.2, AHEaseExponentialOut.new(), self.initialLocationsForColumns[column].y)
    )

    local presentDelayDuration = (#self.columns - columnIndex) * COLUMN_PRESENT_STAGGER_DURATION
    presentAction:addAction(AHDelayed.new(presentDelayDuration, combinedAction))
  end

  return presentAction
end

local function createPromotedPresentAction(self)
  local presentAction = AHSpawn.new()

  local columns = self.columns

  local expandLocationDuration = 0.8
  local expandElasticEase = AHEaseElasticOut.new(1.0)
  local expandDelayDuration = 1.5

  for columnIndex, column in ipairs(columns) do
    local expandLocationAction = nil
    local isSpecialCasedMiddleColumn = ((#columns == 3) and (columnIndex == 2))
    if (isSpecialCasedMiddleColumn) then
      expandLocationAction = AHSequence.new(
        AHLocationXTo.new(column, expandLocationDuration * 0.25, AHEaseSimpleOut.new(), self.initialLocationsForColumns[column].x - column:getWidthPart(0.02)),
        AHLocationXTo.new(column, expandLocationDuration * 0.75, expandElasticEase, self.initialLocationsForColumns[column].x)
      )
    else
      expandLocationAction = AHLocationXTo.new(column, expandLocationDuration, expandElasticEase, self.initialLocationsForColumns[column].x)
    end

    presentAction:addAction(AHSpawn.new(
      AHLocationYTo.new(column, 1.2, AHEaseExponentialOut.new(), self.initialLocationsForColumns[column].y),
      AHDelayed.new(expandDelayDuration, AHSpawn.new(
        expandLocationAction,
        column:createPromotedPresentAction()
      )
    )))
  end

  presentAction:addAction(
    AHDelayed.new(expandDelayDuration, function() Sound.playSoundEffect("split") end)
  )

  return presentAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaPartsStackLayer:getAssembledResponse()
  local assembledResponse = ""
  for _, column in ipairs(self.columns) do
    local partText = column:getHighlightedPartText()
    assembledResponse = assembledResponse .. partText
  end

  return assembledResponse
end

function RutabagaPartsStackLayer:setColumnsTouchEnabled(isEnabled)
  for _, column in ipairs(self.columns) do
    column:setTouchEnabled(isEnabled)
  end
end

function RutabagaPartsStackLayer:disallowAllInput()
  for _, column in ipairs(self.columns) do
    column:disallowAllInput()
  end
end

function RutabagaPartsStackLayer:getIncorrectPartsAndDefinitions()
  local incorrectPartsAndDefinitions = {}
  for _, column in ipairs(self.columns) do
    if (not column:getIsCorrectPillarHighlighted()) then
      table.insert(incorrectPartsAndDefinitions, column:getIncorrectPartAndDefinition())
    end
  end

  return incorrectPartsAndDefinitions
end

function RutabagaPartsStackLayer:createRationaleDismissAction()
  local rationaleDismissAction = AHSpawn.new(
    function()
      Sound.playSoundEffect("switch")
      self:setColumnsTouchEnabled(true)
    end
  )

  for _, column in ipairs(self.columns) do
    rationaleDismissAction:addAction(function()
      column:updateVisualsAfterRationale()
    end)

    if (not column:getIsCorrectPillarHighlighted()) then
      column:disallowInputOnIncorrectPillar()
      rationaleDismissAction:addAction(column:createExitAction())
    end
  end

  return rationaleDismissAction
end

function RutabagaPartsStackLayer:getColumnsCount()
  return #self.columns
end

function RutabagaPartsStackLayer:getColumnsDimensions()
  local leftX = self.columns[1]:getXLeft()
  local rightX = self.columns[#self.columns]:getXRight()
  local width = rightX - leftX
  local height = self:getHeight()

  return width, height
end

function RutabagaPartsStackLayer:setResponseCallback(callback)
  self.responseCallback = callback
end

function RutabagaPartsStackLayer:initializeColumnsHidden(didRoundProgressionPromote)
  cacheColumnsInitialLocations(self)

  if (didRoundProgressionPromote) then
    realignColumnsForPromotedPresent(self)
    displayColumnsForPromotedPresent(self)
  end

  local guaranteedOffscreenOffsetY = self.columns[1]:getHeight() / RutabagaCommon.SHRUNKEN_SCALE
  repositionColumnsBy(self, guaranteedOffscreenOffsetY)
end

function RutabagaPartsStackLayer:createPresentAction(didRoundProgressionPromote, onStartCallback)
  local presentAction = nil
  local touchDelayDuration = nil
  if (didRoundProgressionPromote) then
    presentAction = createPromotedPresentAction(self)
    touchDelayDuration = 1.8
  else
    presentAction = createStandardPresentAction(self)
    touchDelayDuration = 0.7
  end

  local delayedTouchEnableWrapper = function()
    if (onStartCallback ~= nil) then
      onStartCallback()
    end
    self:setColumnsTouchEnabled(true)
  end

  return AHSpawn.new(
    function()
      Sound.playSoundEffect("switch")
    end,
    presentAction,
    AHDelayed.new(touchDelayDuration, delayedTouchEnableWrapper)
  )
end

function RutabagaPartsStackLayer:getCorrectTextLabels()
  local correctTextLabels = {}
  for _, column in ipairs(self.columns) do
    local correctTextLabel = column:getCorrectTextLabel()

    table.insert(correctTextLabels, correctTextLabel)
  end

  return correctTextLabels
end

function RutabagaPartsStackLayer:highlightCorrectText()
  for _, column in ipairs(self.columns) do
    column:highlightCorrectPillar()
  end
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, partsDescriptors)
  local responseWrapper = function()
    registerResponse(self)
  end

  local colorPalette = RutabagaCommon.PILLAR_COLOR_PALETTES[#partsDescriptors]

  local locationXs = getColumnsLocationXs(#partsDescriptors)

  for columnIndex, columnDescriptor in ipairs(partsDescriptors) do
    local colorIndex = colorPalette[columnIndex]
    local colorDescriptor = RutabagaCommon.COLORS.COLUMN_COLORS[colorIndex]
    local newColumn = RutabagaPartsStackColumn.newWithColumnDescriptor(columnDescriptor, colorDescriptor)
    newColumn:setAnchorPointBottom()

    self:insert(newColumn)
    newColumn:setX(locationXs[columnIndex])
    table.insert(self.columns, newColumn)

    newColumn:setResponseCallback(responseWrapper)
  end

  self:setColumnsTouchEnabled(false)
end

function RutabagaPartsStackLayer.newWithPartsDescriptors(partsDescriptors)
  local self = Group.new(Screen.fullWidth, RutabagaCommon.COLUMN_HEIGHT)
  Mixin.mix(self, RutabagaPartsStackLayer)

  self.columns = {}
  self.initialLocationsForColumns = {}

  createLayout(self, partsDescriptors)

  return self
end
