--- RutabagaRoundGenerator
-- @author Created by Jay Austria on May 16, 2015.

class("RutabagaRoundGenerator")

require("data.generators.UtilityGenerators")
require("RutabagaCommon")
require("RutabagaConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ATTENTION:
-- These pool sizes must match the bucket sizes in the game configuration and
-- may need to change if we want to adjust the odds of seeing related concepts in
-- future rounds of gameplay.
local HACKED_EXPECTED_POOL_SIZES = {
  ["1"] = {
    Easy = 5,
    Medium = 5,
    Hard = 5
  },
  ["2"] = {
    Easy = 40,
    Medium = 40,
    Hard = 40
  },
  ["3"] = {
    Easy = 35,
    Medium = 35,
    Hard = 35
  },
  ["4"] = {
    Medium = 20,
    Hard = 30
  }
}

local DISTRACTOR_DELIMITER = " | "
local DISTRACTOR_DELIMITER_FORMATTED = "|"
local INSTRUCTIONS_TEXT_MAP = {
  ["noun"] = "noun",
  ["adjective"] = "adjective",
  ["verb"] = "verb",
  ["word part"] = "",
  ["adverb"] = "adverb"
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------


local function getSeparatedValues(valuesString)
  -- Our shared split function cannot handle splitting by substring. Format them so that
  -- we can split by a single character.
  local formattedValuesString = valuesString:gsub(DISTRACTOR_DELIMITER, DISTRACTOR_DELIMITER_FORMATTED)

  return StringUtils.split(formattedValuesString, DISTRACTOR_DELIMITER_FORMATTED)
end

local function isValidValue(value)
  return (value ~= nil) and (value ~= "")
end

local function getInstructionTextForPartOfSpeech(partOfSpeech)
  local instructionsText = INSTRUCTIONS_TEXT_MAP[partOfSpeech]
  assert(instructionsText, "Invalid part of speech encountered")

  return instructionsText
end

local function packageConceptIntoRound(concept)
  local round = {}
  round.assembledParts = concept.assembledParts
  round.instructionsText = getInstructionTextForPartOfSpeech(concept.partOfSpeech)
  round.isFullWord = (concept.partOfSpeech ~= "word part")
  round.identifier = concept.identifier
  round.difficulty = concept.difficulty

  if (round.isFullWord) then
    round.wordDefinition = concept.wordDefinition
  else
    round.wordDefinition = concept.definition1
  end


  round.parts = {}

  local partPrompts = {}
  local partDefinitions = {}
  local partDistractors = {}
  local partDistractorDefinitions = {}

  local maxParts = 4
  for partIndex = 1, maxParts do
    -- Package word parts
    local part = concept["part"..partIndex]
    if (isValidValue(part)) then
      table.insert(partPrompts, part)
    end

    -- Package part definitions
    local partDefinition = concept["definition"..partIndex]
    if (isValidValue(partDefinition)) then
      table.insert(partDefinitions, partDefinition)
    end

    -- Package part distractors
    local delimiterSeparatedDistractors = concept["distractors"..partIndex]
    if (isValidValue(delimiterSeparatedDistractors)) then
      table.insert(partDistractors, getSeparatedValues(delimiterSeparatedDistractors))
    end

    -- Package part distractors definitions
    local delimiterSeparatedDistractorDefinitions = concept["distractorDefinitions"..partIndex]
    if (isValidValue(delimiterSeparatedDistractorDefinitions)) then
      table.insert(partDistractorDefinitions, getSeparatedValues(delimiterSeparatedDistractorDefinitions))
    end
  end

  local partCount = #partPrompts
  for partIndex = 1, partCount do
    local part = {}
    part.prompt = partPrompts[partIndex]
    part.definition = partDefinitions[partIndex]

    local distractors = partDistractors[partIndex]
    local randomDistractorIndex = math.random(#distractors)
    part.distractor = distractors[randomDistractorIndex]

    local distractorDefinitions = partDistractorDefinitions[partIndex]
    part.distractorDefinition = distractorDefinitions[randomDistractorIndex]

    table.insert(round.parts, part)
  end

  return round
end

local function getConceptPoolByPartCountAndDifficulty()
  local conceptGenerator = GameSession.conceptGenerator
  conceptGenerator:setMarkGeneratedConceptsAsSeen(false)
  local conceptsByPartCount = {}

  for partCount, difficultyCounts in pairs(HACKED_EXPECTED_POOL_SIZES) do
    local conceptsByDifficulty = {}
    for difficultyName, difficultyCount in pairs(difficultyCounts) do

      local filters = {
        difficulty = difficultyName,
        partcount = partCount
      }

      conceptGenerator:setFilters(filters)
      conceptsByDifficulty[difficultyName] = conceptGenerator:nextAsArray(difficultyCount)

      -- Randomize our pool
      TableUtils.shuffleArray(conceptsByDifficulty[difficultyName])
    end

    conceptsByPartCount[partCount] = conceptsByDifficulty
  end

  return conceptsByPartCount
end

local function getConceptTags(self, concept)
  if (self.memoizedTags[concept] == nil) then
    self.memoizedTags[concept] = StringUtils.split(concept.tags, ", ")
  end

  return self.memoizedTags[concept]
end

local function doesConceptContainAnySeenTags(self, concept, tags)
  local conceptTags = getConceptTags(self, concept)

  for _, conceptTag in ipairs(conceptTags) do
    if (tags[conceptTag]) then
      return true
    end
  end

  return false
end

local function getSeenTagsForTwoOrMoreParts(seenTagsByPartCount)
  local seenTags = {}
  for i=2, #seenTagsByPartCount do
    table.insert(seenTags, seenTagsByPartCount[i])
  end
  return seenTags
end

local function sortConceptsByTags(self, chosenDifficulty)
  local seenTagsByPartCount = self.seenTagsByPartCount
  local partCountKey = tostring(self.partCount)
  local conceptsForPartCount = self.conceptPoolByPartCountAndDifficulty[partCountKey]
  local conceptsForPartCountAndDifficulty = conceptsForPartCount[chosenDifficulty]

  local hasTagMatchingSinglePart = {}
  local hasTagMatchingAny = {}

  for conceptIndex = #conceptsForPartCountAndDifficulty, 1, -1 do
    local conceptAtIndex = conceptsForPartCountAndDifficulty[conceptIndex]

    local seenTagsForSinglePart = seenTagsByPartCount[1]
    local seenTagsForTwoOrMoreParts = getSeenTagsForTwoOrMoreParts(seenTagsByPartCount)

    if (doesConceptContainAnySeenTags(self, conceptAtIndex, seenTagsForSinglePart)) then
      table.insert(hasTagMatchingSinglePart, conceptAtIndex)
      table.remove(conceptsForPartCountAndDifficulty, conceptIndex)
    elseif (doesConceptContainAnySeenTags(self, conceptAtIndex, seenTagsForTwoOrMoreParts)) then
      table.insert(hasTagMatchingAny, conceptAtIndex)
      table.remove(conceptsForPartCountAndDifficulty, conceptIndex)
    end
  end

  TableUtils.concatenateArrays(conceptsForPartCountAndDifficulty, hasTagMatchingAny)
  TableUtils.concatenateArrays(conceptsForPartCountAndDifficulty, hasTagMatchingSinglePart)
end

local function cacheTagsAsSeen(self, concept)
  local conceptTags = getConceptTags(self, concept)
  for _, conceptTag in ipairs(conceptTags) do
    while (#self.seenTagsByPartCount < self.partCount) do
      table.insert(self.seenTagsByPartCount, {})
    end
    self.seenTagsByPartCount[self.partCount][conceptTag] = true
  end
end

local function initializeConceptPool(self)
  self.conceptPoolByPartCountAndDifficulty = getConceptPoolByPartCountAndDifficulty()
end

local function removeLastConceptFromCurrentPool(self, chosenDifficulty, isConceptPoolReinitialized)
  local conceptsToRemoveFrom = nil
  if (self.partCount == 1) then
    conceptsToRemoveFrom = self.singlePartConcepts
  else
    local partCountKey = tostring(self.partCount)
    local conceptsForPartCount = self.conceptPoolByPartCountAndDifficulty[partCountKey]
    conceptsToRemoveFrom = conceptsForPartCount[chosenDifficulty]
  end

  local concept = table.remove(conceptsToRemoveFrom)

  if (concept == nil) then
    if isConceptPoolReinitialized then
      error("Failed to reinitialize concept pool, no concepts to choose from.")
    end
    initializeConceptPool(self)
    isConceptPoolReinitialized = true
    return removeLastConceptFromCurrentPool(self, chosenDifficulty, isConceptPoolReinitialized)
  end

  return concept
end

local function getPartOneWordsByMostCommonTags(self)
  local appearanceCountByTag = {}
  local tagsByConcepts = {}

  local singlePartKey = tostring(1)
  local singlePartConceptsByDifficulty = self.conceptPoolByPartCountAndDifficulty[singlePartKey]
  for _, concepts in pairs(singlePartConceptsByDifficulty) do
    for _, concept in ipairs(concepts) do
      local tags = getConceptTags(self, concept)
      for _, tag in ipairs(tags) do
        appearanceCountByTag[tag] = 0
        tagsByConcepts[concept] = tag
      end
    end
  end

  local partCountsToQuery = {2, 3}
  if RutabagaConfiguration.getIsHighLevel() then
    table.insert(partCountsToQuery, 4)
  end

  for _, partCount in ipairs(partCountsToQuery) do
    local partCountKey = tostring(partCount)
    local conceptsByDifficulty = self.conceptPoolByPartCountAndDifficulty[partCountKey]
    for difficulty, concepts in pairs(conceptsByDifficulty) do
      local difficultyWeight = self.difficultyNameChooser:getWeightForValue(difficulty)
      for _, concept in ipairs(concepts) do
        local tags = getConceptTags(self, concept)
        for _, tag in ipairs(tags) do
          if appearanceCountByTag[tag] then
            appearanceCountByTag[tag] = appearanceCountByTag[tag] + difficultyWeight
          end
        end
      end
    end
  end

  local sortedConcepts = TableUtils.getKeysSortedByValue(tagsByConcepts, function(a, b)
    return appearanceCountByTag[a] < appearanceCountByTag[b]
  end)

  return sortedConcepts
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaRoundGenerator:setPartCount(count)
  self.partCount = count
end

function RutabagaRoundGenerator:setDifficultyNameChooser(chooser)
  self.difficultyNameChooser = chooser
end

function RutabagaRoundGenerator:prepareOnePartWords()
  assert(self.difficultyNameChooser, "Cannot sort concepts until a difficultyNameChooser has been set")

  self.singlePartConcepts = getPartOneWordsByMostCommonTags(self)
end

function RutabagaRoundGenerator:reset()
  self.generator:reset()
end

function RutabagaRoundGenerator:generateNextValue()
  local chosenDifficulty = self.difficultyNameChooser:nextValue()

  local willPreferTags = (self.partCount > 1)
  if (willPreferTags) then
    sortConceptsByTags(self, chosenDifficulty)
  end

  local concept = removeLastConceptFromCurrentPool(self, chosenDifficulty)
  if (concept ~= nil) then
    GameSession.conceptGenerator:markConceptIdentifierAsSeen(concept.identifier)
  end

  local willCacheTags = (self.partCount <= 2)
  if (willCacheTags) then
    cacheTagsAsSeen(self, concept)
  end

  local packagedConcept = packageConceptIntoRound(concept)

  return packagedConcept
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function RutabagaRoundGenerator.new()
  local self = Generator.new()
  Mixin.mix(self, RutabagaRoundGenerator)

  initializeConceptPool(self)
  self.partCount = 1
  self.difficultyNameChooser = nil
  self.seenTagsByPartCount = {{}}
  self.memoizedTags = {}
  self.singlePartConcepts = {}

  return self
end
