--- RutabagaValidator
-- @author Created by Jay Austria on May 16, 2016.
-- ConceptValidator for Rutabaga.

class("RutabagaValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "assembledParts",
  "part1",
  "definition1",
  "distractors1",
  "distractorDefinitions1",
  "partOfSpeech",
  "partCount",
  "difficulty"
}

local LIST_SEPARATOR_CHARACTER = " | "
local FORMATTED_LIST_SEPARATOR_CHARACTER = "|"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validatePrintable(self, concept)
  for partIndex = 1, 4 do
    local distractorList = concept["distractors"..partIndex]
    if ((distractorList ~= nil) and (distractorList ~= "")) then
      local formattedDistractorList = distractorList:gsub(LIST_SEPARATOR_CHARACTER, FORMATTED_LIST_SEPARATOR_CHARACTER)
      local distractors = StringUtils.split(formattedDistractorList, FORMATTED_LIST_SEPARATOR_CHARACTER)
      for _, distractor in ipairs(distractors) do
        self:expectStringIsPrintable(distractor, string.format("%s%d", "distractors", partIndex))
        self:expectStringIsNotContentError(distractor, string.format("%s%d", "distractors", partIndex))
      end
    end

    local part = concept["part"..partIndex]
    if ((part ~= nil) and (part ~= "")) then
      self:expectStringIsPrintable(part, string.format("%s%d", "part", partIndex))
      self:expectStringIsNotContentError(part, string.format("%s%d", "part", partIndex))
    end
  end
end

local function validateTags(self, concept)
  local tags = concept.tags

  if ((tags ~= nil) and (tags ~= "")) then
    local expectedNonLetters = TableUtils.createSetFromString("_, ")
    local allowableSet = TableUtils.unionSet(ConceptValidator.LETTERS_SET, expectedNonLetters)
    self:expectStringCharactersBelongToSet(tags, allowableSet, "tags")

    local splitCharacter = ", "
    local tagsSplit = StringUtils.split(tags, splitCharacter)
    for _, tag in ipairs(tagsSplit) do
      local hasWhitespace = (string.find(tag, " ") ~= nil)
      self:expectFalse(hasWhitespace, "Tags had white space when split by the separating commas")
    end
  end
end

local function validatePartCountDifficulty(self, concept)
  local isFourPartWord = (concept.partCount == 4)
  if (isFourPartWord) then
    local isValidDifficulty = ((concept.difficulty == "Medium") or (concept.difficulty == "Hard"))
    self:expectTrue(isValidDifficulty, "Four part words must be Medium or Hard difficulty")
  end
end

local function getListCount(list)
  local count = 0
  if ((list ~= nil) and (list ~= "")) then
    local listValues = StringUtils.split(list, FORMATTED_LIST_SEPARATOR_CHARACTER)
    count = #listValues
  end

  return count
end

local function validateDistractorListCounts(self, concept)
  for partIndex = 1, 4 do
    local areListsMatching = (getListCount(concept["distractors"..partIndex]) == getListCount(concept["distractorDefinitions"..partIndex]))
    self:expectTrue(areListsMatching, "Number of distractors listed did not match number of distractor definitions listed")
  end
end

local function validateEnoughDistractors(self, concept)
  for partIndex = 1, concept.partCount do
    local distractorList = concept["distractors"..partIndex]
    local isListValid = ((distractorList ~= nil) and (distractorList ~= ""))
    self:expectTrue(isListValid, "Invalid number of distractors provided for word. Each word part must have at least one distractor.")
  end
end

local function validateCombinedWord(self, concept)
  local combinedString = ""
  for partIndex = 1, 4 do
    local part = concept["part"..partIndex]
    if ((part ~= nil) and (part ~= "")) then
      combinedString = combinedString .. part
    end
  end

  self:expectTrue((combinedString == concept.assembledParts), "Assembled parts do not seem to match the expected word when combined.")
end

local function validateUniqueDistractors(self, concept)
  local distractorsSeen = {}
  local uniqueDistractors = {}
  for partIndex = 1, 4 do
    local distractorList = concept["distractors"..partIndex]
    if ((distractorList ~= nil) and (distractorList ~= "")) then
      local formattedDistractorList = distractorList:gsub(LIST_SEPARATOR_CHARACTER, FORMATTED_LIST_SEPARATOR_CHARACTER)

      local distractors = StringUtils.split(formattedDistractorList, FORMATTED_LIST_SEPARATOR_CHARACTER)

      local distractorsForThisPart = {}

      for _, distractor in ipairs(distractors) do
        -- It's okay if a word part has redundant distractors. Duplicates are exempt in this case.
        local hasSeenThisDistractorForThisPart = distractorsForThisPart[distractor]
        if (not hasSeenThisDistractorForThisPart) then
          table.insert(distractorsSeen, distractor)
          uniqueDistractors[distractor] = true
          distractorsForThisPart[distractor] = true
        end
      end
    end

    -- Also ensure that distractors do not match any correct parts
    local correctPart = concept["part"..partIndex]
    if ((correctPart ~= nil) and (correctPart ~= "")) then
      table.insert(distractorsSeen, correctPart)
      uniqueDistractors[correctPart] = true
    end
  end

  local areSameSize = (TableUtils.getSize(distractorsSeen) == TableUtils.getSize(uniqueDistractors))
  self:expectTrue(areSameSize, "A distractor was found to exist in multiple word parts or matched the correct part of another word part")
end

local function validateDistractorNotCorrectPart(self, concept)
  for partIndex = 1, 4 do
    local distractorList = concept["distractors"..partIndex]
    if ((distractorList ~= nil) and (distractorList ~= "")) then
      local correctPart = concept["part"..partIndex]

      local formattedDistractorList = distractorList:gsub(LIST_SEPARATOR_CHARACTER, FORMATTED_LIST_SEPARATOR_CHARACTER)
      local distractors = StringUtils.split(formattedDistractorList, FORMATTED_LIST_SEPARATOR_CHARACTER)

      for _, distractor in ipairs(distractors) do
        if (distractor == correctPart) then
          self:expectTrue(false, "A distractor was found that matches the correct part")
        end
      end
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function RutabagaValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validatePartCountDifficulty(self, concept)
  validateDistractorListCounts(self, concept)
  validateUniqueDistractors(self, concept)
  validateDistractorNotCorrectPart(self, concept)
  validateCombinedWord(self, concept)
  validateEnoughDistractors(self, concept)
  validateTags(self, concept)
  validatePrintable(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function RutabagaValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, RutabagaValidator)

  return self
end
