--- SimonGame
-- @author Created by Aaron Barrett on February 23, 2015.

class("SimonGame")

require("StandardUIBuilder")
require("SimonConfiguration")
require("SimonContentManager")
require("SimonRound")
require("SimonPanelAudio")
require("SimonPanelOrdering")
require("SimonPanelIntermission")
require("SimonPanelOverview")
require("SimonPanelBegin")
require("SimonOutro")
require("SimonSubmitButton")
require("SimonTheme")
require("SimonStarField")
require("bonuses.ResponseSpeedTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Panel choice options ---------------

SimonGame.PANEL_CHOICE_MORE_INFO = "NEXT_PANEL_CHOICE_MORE_INFO"
SimonGame.PANEL_CHOICE_START_GAME = "PANEL_CHOICE_START_GAME"

-- Colors -----------------------------

local UI_COMPONENTS_COLOR = {1.0, 1.0, 1.0}

-- Intermission panel text ------------

local TEXT_EXPLAIN_REVIEW = "NEXT, PRACTICE EACH PART AND COMMIT IT TO MEMORY"
local TEXT_EXPLAIN_FINAL_ORDERING = "NOW THAT YOU'VE PRACTICED EACH PART, ORGANIZE BOTH HALVES AS ONE"

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_COUNTDOWN = 20
local Z_STAR_FIELD = 30
local Z_SUBMIT_BUTTON = 40
local Z_PANEL = 50

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Lives ------------------------------

local function takeALife(self)
  local livesCounter = self.livesCounter
  Sound.playSoundEffect("incorrect")
  livesCounter:decrement()
end

local function loseIfOutOfLives(self)
  local livesCounter = self.livesCounter
  if (livesCounter:getLives() <= 0) then
    GameSession.notifyGameLoss()
    waitFor(math.huge)
  end
end

-- Ordering panels --------------------

local function startQuickResponseBonus(self, speedBonusWindowDuration)
  local decayWindowStart = speedBonusWindowDuration * SimonConfiguration.SPEED_BONUS_DECAY_START_PROPORTION
  local responseSpeedTracker = self.responseSpeedTracker
  responseSpeedTracker:setBonusWindowDuration(speedBonusWindowDuration)
  responseSpeedTracker:setDecayWindowStart(decayWindowStart)
  responseSpeedTracker:notifyStart()
end

local function waitForShowOrderingPanelWithSteps(self, steps, speedBonusWindowDuration, shouldShowTip, currentRoundIndex)
  local submitButton = self.submitButton
  local theme = self.theme

  local function stopSpeedBonusCallback()
    self.responseSpeedTracker:notifyCorrectResponse()
  end

  local orderingPanel = SimonPanelOrdering.new(theme, steps, self.dragAndDropHandler, submitButton, shouldShowTip, stopSpeedBonusCallback, self.concept, currentRoundIndex)
  self:insertAtCenter(orderingPanel, Z_PANEL)

  wait(orderingPanel:playFlyIntoPlace())

  self.dragAndDropHandler:setEnabled(true)
  startQuickResponseBonus(self, speedBonusWindowDuration)
  waitUntil(function() return orderingPanel:hasFinishedCheckingAnswers() end)
  GameSession.increaseScore(SimonConfiguration.POINTS_PER_CORRECT_RESPONSE)

  if (orderingPanel:wasAnswerPerfect()) then
    ScoringUtils.applyPerfectSetBonus(SimonConfiguration.POINTS_PER_PERFECT_SET_BONUS, self.bonusTray)
  else
    takeALife(self)
  end

  self.scoreLabel:setScore(GameSession.getScore())

  wait(orderingPanel:playCorrections())
  orderingPanel:removeSelf()
  self.dragAndDropHandler:clear()

  loseIfOutOfLives(self)
end

-- Intermission Panels ----------------

local function waitForIntermissionPanel(self, panelText)
  local submitButton = self.submitButton

  local intermissionPanel = SimonPanelIntermission.new(panelText, submitButton)
  self:insertAtCenter(intermissionPanel, Z_PANEL)
  submitButton:playEmergeWithText("OKAY")
  wait(intermissionPanel:playAppear())
  waitUntil(function() return intermissionPanel:isDismissed() end)
  submitButton:playSubmerge()
  wait(intermissionPanel:playDisappear())
  intermissionPanel:removeSelf()
end

-- Intro Panels -----------------------

local function waitForIntroPanels(self)
  local submitButton = self.submitButton
  local currentPanel = self.beginPanel
  local previousPanel = self.overviewPanel
  local currentText = "BEGIN"
  local previousText = "OKAY"

  local function swapPanels()
    currentPanel, previousPanel = previousPanel, currentPanel
    currentText, previousText = previousText, currentText
  end

  if (GameSession.hasNeverWon()) then
    swapPanels()
  end

  submitButton:playFadeIn()

  repeat
    submitButton:playEmergeWithText(currentText)
    wait(currentPanel:playAppear())
    waitUntil(function() return currentPanel:isDismissed() end)
    submitButton:playSubmerge()
    wait(currentPanel:playDisappear())

    swapPanels()
  until (previousPanel:getPanelChoice() == SimonGame.PANEL_CHOICE_START_GAME)

  previousPanel:removeSelf()
  currentPanel:removeSelf()
  self.beginPanel = nil
  self.overviewPanel = nil
end


-- Outro ------------------------------

local function waitForOutro(self)
  ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), SimonConfiguration.getPointsPerLivesBonus(), self.bonusTray)
  self.scoreLabel:setScore(GameSession.getScore())

  Sound.playSoundEffect("win")

  local submitButton = self.submitButton

  local outro = SimonOutro.new()
  self:insertAtCenter(outro, Z_PANEL)

  submitButton:playFadeOut()
  wait(outro:playOutro())
end

local function notifyBonusesAndWin(self)
  GameSession.notifyGameWin()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SimonGame:gameFlow()
  local theme = self.theme
  local contentManager = self.contentManager
  local description = contentManager:getDescription()
  local audioRounds = contentManager:getAudioRounds()
  local audioRoundsCount = #audioRounds
  local currentRoundIndex = 1

  waitForIntroPanels(self)

  for roundIndex, round in ipairs(audioRounds) do
    local audioPanel = SimonPanelAudio.new(theme, round, description, roundIndex, audioRoundsCount)
    self:insertAtCenter(audioPanel, Z_PANEL)

    wait(audioPanel:playAudio())
    audioPanel:removeSelf()

    local roundSteps = round:getSteps()

    local speedBonusWindowDuration = SimonConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS_TYPE_HALF
    local shouldShowTip = (roundIndex == 1)
    waitForShowOrderingPanelWithSteps(self, roundSteps, speedBonusWindowDuration, shouldShowTip, currentRoundIndex)
    currentRoundIndex = currentRoundIndex + 1
  end

  waitForIntermissionPanel(self, TEXT_EXPLAIN_REVIEW)

  for roundIndex, round in ipairs(audioRounds) do
    local roundSteps = round:getSteps()
    local speedBonusWindowDuration = SimonConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS_TYPE_HALF
    local shouldShowTip = false
    waitForShowOrderingPanelWithSteps(self, roundSteps, speedBonusWindowDuration, shouldShowTip, currentRoundIndex)
    currentRoundIndex = currentRoundIndex + 1
  end

  waitForIntermissionPanel(self, TEXT_EXPLAIN_FINAL_ORDERING)

  local finalRoundSteps = contentManager:getAllSteps()
  local speedBonusWindowDuration = SimonConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS_TYPE_COMBINED
  local shouldShowTip = false
  waitForShowOrderingPanelWithSteps(self, finalRoundSteps, speedBonusWindowDuration, shouldShowTip, currentRoundIndex)

  waitForOutro(self)

  notifyBonusesAndWin(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local rankTwoScore = SimonConfiguration.getRankTwoScore()
  local rankThreeScore = SimonConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.PERFECT_SET_BONUS)
end

local function configureContent(self)
  local contentManager = SimonContentManager.new()
  self.contentManager = contentManager
  self.concept = contentManager:getConcept()
  GameSession.contentTracker:notifyPresentedConcept(self.concept.identifier)
end

local function configureTheme(self)
  local contentManager = self.contentManager
  local category = contentManager:getCategory()
  self.theme = SimonTheme.new(category)
end

local function createGameElements(self)
  local theme = self.theme

  local backgroundColorTop = theme:getBackgroundTopColor()
  local backgroundColorBottom = theme:getBackgroundBottomColor()

  local background = RectangleGradient.new(Screen.width, Screen.height, backgroundColorTop, backgroundColorBottom)
  self:insertAtCenter(background, Z_BACKGROUND)

  local submitButton = SimonSubmitButton.new(theme)
  submitButton:setLoc(self:getWidthPart(0.5), submitButton:getHeightPart(0.5))
  self:insert(submitButton, Z_SUBMIT_BUTTON)
  submitButton:setTouchEnabled(false)
  submitButton:prepareToFadeIn()
  self.submitButton = submitButton

  local starField = SimonStarField.new()
  self:insertAtCenter(starField, Z_STAR_FIELD)
  starField:playTwinkle()
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(SimonConfiguration.INITIAL_LIVES_COUNT)

  local function onQuickResponseCallback(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 SimonConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end
  local bonusWindowDuration = nil -- Simon sets the bonus window duration at the beginning of each matching session.
  local decayStart = nil -- Simon sets the decay start at the beginning of each matching session.
  self.responseSpeedTracker = ResponseSpeedTracker.new(bonusWindowDuration, onQuickResponseCallback, decayStart)
end

local function buildIntroPanels(self)
  local theme = self.theme
  local submitButton = self.submitButton

  local overviewPanel = SimonPanelOverview.new(theme, submitButton)
  self:insertAtCenter(overviewPanel, Z_PANEL)
  self.overviewPanel = overviewPanel

  local beginPanel = SimonPanelBegin.new(theme, submitButton)
  self:insertAtCenter(beginPanel, Z_PANEL)
  self.beginPanel = beginPanel
end

-- Constructor ------------------------

function SimonGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SimonGame)

  local dragAndDropHandler = DragAndDropHandler.new()
  dragAndDropHandler:setEnabled(false)
  self.dragAndDropHandler = dragAndDropHandler

  configureGameDifficulty(self)
  configureContent(self)
  configureTheme(self)
  createGameElements(self)
  createUI(self)
  buildIntroPanels(self)

  return self
end
