--- SimonMarker
-- @author Created by Aaron Barrett on March 23, 2015.

class("SimonMarker")

require("SimonSparks")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local RADIUS = 2.0

local WIDTH = Screen.width
local HEIGHT = Screen.height

local SPARKS_X = Screen.getWidthPart(0.07)
local DOT_X = Screen.getWidthPart(0.02)
local PEN_WIDTH = Screen.getWidthPart(0.01)

local Z_LINE = 10
local Z_DOT = 20
local Z_SPARKS = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function playSparks(self)
  local selectedItem = self.selectedItem
  local sparksY = selectedItem:getY()

  local sparks = SimonSparks.new()
  sparks:setScl(0.5)
  sparks:setRot(90)
  sparks:setLoc(SPARKS_X, sparksY)
  self:insert(sparks, Z_SPARKS)

  return sparks:playExplode()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SimonMarker:setSnapTime(snapTime)
  local dot = self.dot
  local selectedItem = self.selectedItem
  local selectedItemX = selectedItem:getLeftEdgeX()
  local selectedItemY = selectedItem:getY()
  local snapStartX = DOT_X
  local snapStartY = self.snapStartY

  local dotX = MathUtils.bezier(snapStartX, snapStartX, snapStartX, selectedItemX, snapTime)
  local dotY = MathUtils.bezier(snapStartY, selectedItemY, selectedItemY, selectedItemY, snapTime)

  dot:setLoc(dotX, dotY)

  self.snapTime = snapTime
end

function SimonMarker:getSnapTime()
  return self.snapTime
end

function SimonMarker:playAppear(dotY, selectedItem)
  local dot = self.dot

  self:setVisible(true)

  self.snapping = false
  self.snapTime = 0.0
  self.snapStartY = dotY

  dot:setLoc(DOT_X, dotY)
  self.selectedItem = selectedItem

  local appearAction = AHScaleTo.new(dot, 0.2, AHEaseSimpleOut.new(2.0), 1.0)
  self.appearanceInterrupt:insert(appearAction)
end

function SimonMarker:playDisappear()
  local dot = self.dot

  dot:setX(DOT_X)

  self.snapping = true
  self.snapTime = 0.0
  self.snapStartY = dot:getY()

  local notQuiteFullDuration = 0.22
  local fullDuration = 0.24

  local turnOffSnapBeforeItFinishesAction = AHDelayed.new(notQuiteFullDuration,
                                                          function()
                                                            self:setVisible(false)
                                                            playSparks(self)
                                                            self.snapping = false
                                                          end
                                                          )

  local disappearAction = AHSpawn.new(
                                      AHScaleTo.new(dot, notQuiteFullDuration, AHEaseSimpleIn.new(2.0), 0.0),
                                      AHNonAttributeTo.new(self, fullDuration, self.getSnapTime, self.setSnapTime, AHEaseSineIn.new(), 1.0),
                                      turnOffSnapBeforeItFinishesAction
                                      )
  self.appearanceInterrupt:insert(disappearAction)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SimonMarker.new()
  local self = Group.newCustomDraw(WIDTH, HEIGHT)
  Mixin.mix(self, SimonMarker)

  local function drawCallback()
    local startX, startY = self.dot:getLoc()
    local endX = 0.0
    local endY = 0.0

    local selectedItem = self.selectedItem
    if (selectedItem ~= nil) then
      local fudgeAmount = 2.5
      local scale = selectedItem:getScl()
      endX = selectedItem:getLeftEdgeX()
      endY = selectedItem:getY()
    end

    MOAIGfxDevice.setPenWidth(PEN_WIDTH)

    if (self.snapping) then
      local snapTime = self.snapTime

      local x0 = startX
      local y0 = startY
      local x1 = endX
      local y1 = endY
      local cx0 = startX
      local cy0 = startY * (1.0 - snapTime) + endY * snapTime
      local cx1 = startX * snapTime + endX * (1.0 - snapTime)
      local cy1 = endY
      local stepsCount = 8
      MOAIDraw.drawBezierCurve(x0, y0, x1, y1, cx0, cy0, cx1, cy1, stepsCount)
    else
      MOAIDraw.drawLine(startX, startY, endX, endY)
    end
  end
  self.deck:setDrawCallback(drawCallback)

  local dot = FilledCircle.new(RADIUS)
  dot:setX(DOT_X)
  dot:setScl(0.0)
  dot:inheritColor(self)
  self:insert(dot, Z_DOT)
  self.dot = dot

  self:setVisible(false)

  self.appearanceInterrupt = ActionInterrupt.new()

  self.selectedItem = nil
  self.snapStartY = 0.0
  self.snapTime = 0.0
  self.snapping = false

  return self
end
