--- SimonPanelAudio
-- @author Created by Aaron Barrett on February 23, 2015.

class("SimonPanelAudio")

require("data.visualization.SineWaveVisualizer")
require("data.generators.SoundAveragePowerGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local VISUALIZER_COLOR = {1.0, 1.0, 1.0}

local Z_VISUALIZER = 10
local Z_COLUMN = 20

local WORD_FORMS_BY_NUMBER = {
  [1] = "ONE",
  [2] = "TWO"
}

local FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 16.0)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function waitUntilAudioIsFinished(audio)
  waitUntil(function() return audio:isFinished() end)
end

local function getWordFormOfNumber(number)
  local wordForm = WORD_FORMS_BY_NUMBER[number]
  if (wordForm ~= nil) then
    return wordForm
  else
    error("No word form provided for number: `" .. tostring(number) .. "`")
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SimonPanelAudio:playAudio()
  return runInCoroutine(function()
    local visualizer = self.visualizer

    local fadeInAction = AHOpacityTo.new(self, 0.2, nil, 1.0)
    fadeInAction:start()

    local concept = self.concept
    local audioFilename = self.audioFilename
    local audio = AssetHelper.playSoundForConcept(concept, audioFilename)

    local generator = SoundAveragePowerGenerator.new(audio)
    visualizer:setGenerator(generator)
    visualizer:start()

    waitUntilAudioIsFinished(audio)

    local fadeOutAction = AHOpacityTo.new(self, 0.2, nil, 1.0)
    wait(fadeOutAction:start())
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildVisualizer(self)
  local visualizerWidth = self:getWidth()
  local visualizerHeight = self:getHeightPart(0.12)
  local visualizer = SineWaveVisualizer.new(visualizerWidth,
                                            visualizerHeight,
                                            VISUALIZER_COLOR)
  visualizer:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.45))
  visualizer:inheritOpacity(self)
  self:insert(visualizer, Z_VISUALIZER)
  self.visualizer = visualizer
end

local function buildDescriptionLabels(self, theme, description, part, partsCount)
  local spacing = self:getHeightPart(0.005)
  local column = Column.new(spacing)

  local color = theme:getGrayedTextColor()

  local upperCaseDescription = description:upper()
  local descriptionLabel = Label.new(upperCaseDescription, FONT)
  descriptionLabel:inheritOpacity(self)
  column:insert(descriptionLabel)

  local partWordForm = getWordFormOfNumber(part)
  local partsCountWordForm = getWordFormOfNumber(partsCount)
  local partString = "PART " .. partWordForm .. " OF " .. partsCountWordForm
  local partLabel = Label.new(partString, FONT)
  partLabel:setColor(unpack(color))
  partLabel:inheritOpacity(self)
  column:insert(partLabel)

  column:alignChildren()
  column:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.6))
  self:insert(column, Z_COLUMN)
end

-- Constructor ------------------------

function SimonPanelAudio.new(theme, round, description, part, partsCount)
  local self = Group.newFullScreen()
  Mixin.mix(self, SimonPanelAudio)

  buildVisualizer(self)
  buildDescriptionLabels(self, theme, description, part, partsCount)

  self:setOpacity(0.0)

  self.concept = round:getConcept()
  self.audioFilename = round:getAudioFilename()

  return self
end
