--- SimonPanelOrdering
-- @author Created by Aaron Barrett on February 23, 2015.

class("SimonPanelOrdering")

require("SimonItem")
require("SimonContainer")
require("SimonSubmitButton")
require("SimonCorrection")
require("SimonCorrectionBuilder")
require("SimonMarker")
require("SimonTriangle")
require("SimonZoomer")
require("SimonSparks")
require("SimonRectangleSpewer")
require("SimonItemPalette")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Dimensions -------------------------

local LIST_HEIGHT = Screen.getSafeHeightPart(0.78)
local MAX_ITEM_HEIGHT = Screen.getSafeHeightPart(0.1)

-- Z-Indices --------------------------

local Z_TIP_LABEL = 10
local Z_ZOOMER = 20
local Z_MARKER = 30
local Z_TAP_TO_CONTINUE_LABEL = 40
local Z_RECTANGLE_SPEWER = 50
local Z_CONTAINER = 60
local Z_TRIANGLE = 70
SimonPanelOrdering.Z_ITEM = 80
SimonPanelOrdering.Z_ITEM_SELECTED = 90
local Z_SPARKS = 100

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Guaranteed randomization -----------

local function isItemsAlreadyInOrder(items)
  for itemIndex, item in ipairs(items) do
    if (itemIndex ~= item:getCorrectIndex()) then
      return false
    end
  end
  return true
end

local function swapTwoElements(items)
  local itemsCount = #items

  local indexA = math.random(itemsCount)
  local indexB = math.random(itemsCount - 1)
  if (indexB >= indexA) then
    indexB = indexB + 1
  end

  items[indexA], items[indexB] = items[indexB], items[indexA]
end

-- Automatically arrange items --------

local function playArrangeItems(self, excludedItem)
  local arrangeAction = AHSpawn.new()

  local items = self.items
  local containers = self.containers
  for itemIndex, item in ipairs(items) do
    local shouldMoveitem = (item ~= excludedItem)
    if (shouldMoveitem) then
      local correspondingContainer = containers[itemIndex]
      local destinationX = correspondingContainer:getX()
      local destinationY = correspondingContainer:getY()
      local itemAction = item:createArrangeAction(destinationX, destinationY, itemIndex)
      arrangeAction:addAction(itemAction)
    end
  end

  self.arrangeInterrupt:insert(arrangeAction)
end

-- Drag and Drop ----------------------

local function notifyItemPickedUp(self, pickedUpItem)
  local items = self.items
  local containers = self.containers
  local marker = self.marker
  local zoomer = self.zoomer

  local itemIndex = TableUtils.indexOf(items, pickedUpItem)
  local correspondingContainer = containers[itemIndex]
  local dotY = correspondingContainer:getY()

  marker:playAppear(dotY, pickedUpItem)
  zoomer:playZoomBack()
end

local function notifyItemEntered(self, enteringItem, container)
  local items = self.items
  local containerIndex = container:getContainerIndex()
  TableUtils.removeValueFromArray(items, enteringItem)
  table.insert(items, containerIndex, enteringItem)

  local excludedItem = enteringItem

  playArrangeItems(self, excludedItem)
end

local function notifyItemDropped(self, droppedItem, droppedContainer)
  local marker = self.marker
  local zoomer = self.zoomer

  local wasDroppedOnAContainer = (droppedContainer ~= nil)
  if (wasDroppedOnAContainer) then
    local items = self.items
    local containerIndex = droppedContainer:getContainerIndex()
    TableUtils.removeValueFromArray(items, droppedItem)
    table.insert(items, containerIndex, droppedItem)
  end

  playArrangeItems(self)
  marker:playDisappear()
  zoomer:playZoomForward()
end

-- Corrections ------------------------

local function buildCorrections(self)
  local items = self.items
  self.corrections = SimonCorrectionBuilder.createCorrections(items)
end

local function waitForMarkAnswers(self)
  local containers = self.containers
  local items = self.items
  local triangle = self.triangle
  local dragAndDropHandler = self.dragAndDropHandler
  local sparks = self.sparks
  local rectangleSpewer = self.rectangleSpewer
  local tipLabel = self.tipLabel
  local tapToContinueLabel = self.tapToContinueLabel
  local tapToContinueInterrupt = self.tapToContinueInterrupt

  local staggerDuration = 0.01
  local markAction = AHStaggeredSpawn.new(staggerDuration)

  local orderString = ""
  for _, item in ipairs(items) do
    local itemAction = item:createMarkAction()
    orderString = orderString .. " " .. item:getCorrectIndex()
    markAction:addAction(itemAction)
  end
  orderString = StringUtils.trim(orderString)

  if (tipLabel ~= nil) then
    local tipLabelFadeOutAction = AHOpacityTo.new(tipLabel, 0.9, nil, 0.0)
    tipLabelFadeOutAction:start()
  end

  local topContainer = containers[1]
  local triangleDestinationY = topContainer:getY() + topContainer:getHeightPart(0.5) + triangle:getHeightPart(0.5)

  dragAndDropHandler:forceDrop()
  triangle:playDrop(triangleDestinationY)
  waitFor(0.35)
  sparks:playExplode()
  wait(markAction:start())

  local fadeTapToContinueAction = AHDelayed.new(0.3, AHOpacityTo.new(tapToContinueLabel, 1.2, nil, 1.0))
  tapToContinueInterrupt:insert(fadeTapToContinueAction)

  local roundIndexAsString = tostring(self.currentRoundIndex)
  if (self:wasAnswerPerfect()) then
    GameSession.contentTracker:notifyCorrectAnswer(self.conceptIdentifier, roundIndexAsString)
    Sound.playSoundEffect("perfect")
    rectangleSpewer:playSpew()
  else
    GameSession.contentTracker:notifyIncorrectAnswer(self.conceptIdentifier, roundIndexAsString, orderString)
  end
end

local function playFadeOutItems(self)
  local items = self.items
  local triangle = self.triangle
  local tapToContinueLabel = self.tapToContinueLabel
  local tapToContinueInterrupt = self.tapToContinueInterrupt

  self:setTouchEnabled(false)

  local quickFadeOutAction = AHSpawn.new()

  local triangleAction = triangle:createQuickFadeOutAction()
  quickFadeOutAction:addAction(triangleAction)

  for _, item in ipairs(items) do
    local itemAction = item:createQuickFadeOutAction()
    quickFadeOutAction:addAction(itemAction)
  end

  local fadeOutAndDismissAction = AHSequence.new(
                                                 quickFadeOutAction,
                                                 function()
                                                   self.finishedFadingOutItems = true
                                                 end
                                                 )

  local fadeTapToContinueAction = AHOpacityTo.new(tapToContinueLabel, 0.2, nil, 0.0)
  tapToContinueInterrupt:insert(fadeTapToContinueAction)

  return fadeOutAndDismissAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Corrections ------------------------

function SimonPanelOrdering:playCorrections()
  return runInCoroutine(function()
    local items = self.items
    local containers = self.containers
    local triangle = self.triangle

    local staggerDuration = 0.01
    local correctionAction = AHStaggeredSpawn.new(staggerDuration)

    for _, item in ipairs(items) do
      local correctIndex = item:getCorrectIndex()
      local correspondingContainer = containers[correctIndex]
      local destinationY = correspondingContainer:getY()

      if (item:getShouldMarkWrong()) then
        item:setZIndex(SimonPanelOrdering.Z_ITEM_SELECTED)
      end

      local itemAction = item:createCorrectionAction(destinationY)
      correctionAction:addAction(itemAction)
    end

    wait(correctionAction:start())

    Button.makeButton(self, function() playFadeOutItems(self) end)
    waitUntil(function() return self.finishedFadingOutItems end)
  end)
end

function SimonPanelOrdering:playCheckAnswers()
  runInCoroutine(function()
    self.submitButton:playSubmerge()
    self.dragAndDropHandler:setEnabled(false)

    buildCorrections(self)
    waitForMarkAnswers(self)

    self.finishedCheckingAnswers = true
  end)
end

-- Intro animation --------------------

local function waitBrieflyForPerformanceReasons()
  waitForFrames(2)
end

function SimonPanelOrdering:playFlyIntoPlace()
  local triangle = self.triangle
  local zoomer = self.zoomer
  local items = self.items
  local containers = self.containers
  local tipLabel = self.tipLabel
  local submitButton = self.submitButton

  for _, item in ipairs(items) do
    zoomer:insert(item, SimonPanelOrdering.Z_ITEM)
    item:setScl(0.9 + math.random() * 0.15)
  end

  local flyIntoPlaceAction = AHSpawn.new()

  local itemsCount = #items
  local minimumFirstFloppyIndex = 3
  local firstFloppyItemIndex = math.random(minimumFirstFloppyIndex, itemsCount)
  local firstFloppyItem = items[firstFloppyItemIndex]

  local middleScale = 1.05 + math.random() * 0.2
  local overshootY = containers[1]:getY()
  local destinationY = containers[firstFloppyItemIndex]:getY()
  local destinationIndex = firstFloppyItemIndex
  firstFloppyAction = firstFloppyItem:createFloppyFlyInAction(middleScale, overshootY, destinationY, destinationIndex)
  flyIntoPlaceAction:addAction(firstFloppyAction)

  for itemIndex, item in ipairs(items) do
    local isSteady = (not (itemIndex == firstFloppyItemIndex))
    if (isSteady) then
      local delayDuration = math.random() * 0.5
      local duration = 1.1 - delayDuration
      local destinationY = containers[itemIndex]:getY()
      local destinationIndex = itemIndex
      local steadyAction = item:createSteadyFlyInAction(duration, destinationY, destinationIndex)
      local delayedAction = AHDelayed.new(delayDuration, steadyAction)
      flyIntoPlaceAction:addAction(delayedAction)
    end
  end

  waitBrieflyForPerformanceReasons()

  if (tipLabel ~= nil) then
    local tipLabelFadeInAction = AHOpacityTo.new(tipLabel, 0.9, nil, 1.0)
    tipLabelFadeInAction:start()
  end

  local topContainer = containers[1]
  local topContainerY = topContainer:getY()
  local triangleDestinationY = topContainerY + topContainer:getHeightPart(0.9)
  triangle:playFlyIntoPlace(triangleDestinationY)

  runInCoroutine(function()
      waitFor(0.7)
      submitButton:playEmergeWithText("SUBMIT")
    end)

  Sound.playSoundEffect("fly_in")

  return flyIntoPlaceAction:start()
end

-- Check status from SimonGame --------

function SimonPanelOrdering:hasFinishedCheckingAnswers()
  return self.finishedCheckingAnswers
end

function SimonPanelOrdering:wasAnswerPerfect()
  local corrections = self.corrections
  return (#corrections == 0)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function buildZoomer(self)
  local zoomer = SimonZoomer.new()
  self:insertAtCenter(zoomer, Z_ZOOMER)
  self.zoomer = zoomer
end

local function buildContainers(self, correctSteps)
  local dragAndDropHandler = self.dragAndDropHandler

  local column = Column.new(0.0)
  local containers = {}

  local stepsCount = #correctSteps

  local itemHeight = math.min(MAX_ITEM_HEIGHT, LIST_HEIGHT / stepsCount)
  self.itemHeight = itemHeight

  for containerIndex = 1, stepsCount do
    local function onEnteredCallback(enteringItem, container)
      notifyItemEntered(self, enteringItem, container)
    end

    local container = SimonContainer.new(itemHeight, containerIndex, onEnteredCallback)
    column:insert(container)
    table.insert(containers, container)
  end
  column:alignChildren()

  local columnHeight = column:getHeight()
  local columnX = Screen.getSafeX(0.5)
  local columnY = Screen.getSafeY(0.53)

  column:setLoc(columnX, columnY)

  self.zoomer:insert(column)
  for _, container in ipairs(containers) do
    self:takeChild(container, Z_CONTAINER)
    dragAndDropHandler:addContainer(container)
  end
  column:removeSelf()

  local rectangleSpewer = SimonRectangleSpewer.new(SimonItem.WIDTH, columnHeight, SimonRectangleSpewer.KIND_PERFECT_ANSWER)
  rectangleSpewer:setLoc(columnX, columnY)
  self.zoomer:insert(rectangleSpewer, Z_RECTANGLE_SPEWER)
  self.rectangleSpewer = rectangleSpewer

  self.containers = containers
end

local function buildItems(self, correctSteps, theme)
  local dragAndDropHandler = self.dragAndDropHandler
  local itemHeight = self.itemHeight
  local zoomer = self.zoomer

  local items = {}

  local stepsCount = #correctSteps

  local palette = SimonItemPalette.new(theme, stepsCount)

  for stepIndex, step in ipairs(correctSteps) do
    local function onPickedUpCallback(pickedUpItem)
      notifyItemPickedUp(self, pickedUpItem)
    end

    local function onDroppedCallback(droppedItem, droppedContainer)
      notifyItemDropped(self, droppedItem, droppedContainer)
    end

    local item = SimonItem.new(itemHeight, step, stepIndex, onPickedUpCallback, onDroppedCallback, palette, zoomer)
    item:setLoc(self:getWidthPart(0.5), -self:getHeightPart(0.15))
    dragAndDropHandler:addObject(item)

    table.insert(items, item)
  end

  TableUtils.shuffleArray(items)

  if (isItemsAlreadyInOrder(items)) then
    swapTwoElements(items)
  end

  self.items = items
end

local function buildMarker(self)
  local marker = SimonMarker.new()
  self:insertAtCenter(marker, Z_MARKER)
  self.marker = marker
end

local function buildTriangle(self)
  local triangle = SimonTriangle.new()

  local containers = self.containers
  local topContainer = containers[1]
  local topContainerY = topContainer:getY()

  local triangleY = math.max(topContainerY + self:getHeightPart(0.5), self:getHeight() + triangle:getHeightPart(0.5))

  triangle:setLoc(self:getWidthPart(0.5), triangleY)

  self.zoomer:insert(triangle, Z_TRIANGLE)
  self.triangle = triangle

  local sparks = SimonSparks.new()

  local sparksY = topContainerY + topContainer:getHeightPart(0.57)

  sparks:setLoc(self:getWidthPart(0.5), sparksY)
  self.zoomer:insert(sparks, Z_SPARKS)
  self.sparks = sparks
end

local function configureSubmitButton(self, submitButton)
  local function onSubmitCallback()
    Sound.playSoundEffect("submit")
    self.stopSpeedBonusCallback()
    submitButton:playPressed()
    self:playCheckAnswers()
  end
  submitButton:setCallback(onSubmitCallback)
  self.submitButton = submitButton
end

local function buildTapToContinueLabel(self, theme)
  local color = theme:getGrayedTextColor()

  local tapToContinueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 16.0))
  tapToContinueLabel:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.07) + Screen.safeBottom)
  tapToContinueLabel:setColor(unpack(color))
  tapToContinueLabel:setOpacity(0.0)
  self:insert(tapToContinueLabel, Z_TAP_TO_CONTINUE_LABEL)
  self.tapToContinueLabel = tapToContinueLabel
end

local function buildTipLabel(self, theme, shouldShowTip)
  if (shouldShowTip) then
    local text = "ARRANGE ITEMS SEQUENTIALLY<br />PLACE THE FIRST ITEM AT THE TOP"
    local width = self:getWidth()
    local fontFace = FontFaces.DIN_OT_MEDIUM
    local fontSize = 18
    local attributesForLabels = nil
    local horizontalAlignment = Alignment.CENTER
    local tipLabel = SimonTextLabel.new(text, width, fontFace, fontSize, attributesForLabels, horizontalAlignment)
    local color = theme:getGrayedTextColor()
    tipLabel:setColor(unpack(color))
    tipLabel:setOpacity(0.0)
    tipLabel:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.89))
    self:insert(tipLabel, Z_TIP_LABEL)
    self.tipLabel = tipLabel
  end
end

-- Constructor ------------------------

function SimonPanelOrdering.new(theme, correctSteps, dragAndDropHandler, submitButton, shouldShowTip, stopSpeedBonusCallback, concept, currentRoundIndex)
  local self = Group.newFullScreen()
  Mixin.mix(self, SimonPanelOrdering)

  self.dragAndDropHandler = dragAndDropHandler
  self.conceptIdentifier = concept.identifier
  self.currentRoundIndex = currentRoundIndex

  local stepsCount = #correctSteps

  buildZoomer(self)
  buildContainers(self, correctSteps)
  buildItems(self, correctSteps, theme)
  buildMarker(self)
  buildTriangle(self)
  configureSubmitButton(self, submitButton)
  buildTapToContinueLabel(self, theme)
  buildTipLabel(self, theme, shouldShowTip)

  self.stopSpeedBonusCallback = stopSpeedBonusCallback

  self.correctSteps = correctSteps
  self.arrangeInterrupt = ActionInterrupt.new()
  self.tapToContinueInterrupt = ActionInterrupt.new()

  self.corrections = nil
  self.currentCorrectionIndex = 0

  self.finishedCheckingAnswers = false
  self.finishedFadingOutItems = false

  return self
end
