--- SimonPanelOverview
-- @author Created by Aaron Barrett on March 26, 2015.

class("SimonPanelOverview")

require("SimonTextLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SLIDE_IN_DISTANCE = Screen.getWidthPart(0.25)
local SLIDE_OUT_DISTANCE = -Screen.getHeightPart(0.05)

local Z_MAIN_COLUMN = 10

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function enableDismissal(self)
  local submitButton = self.submitButton
  local onDismissCallback = self.onDismissCallback
  submitButton:setCallback(onDismissCallback)
  self:setTouchEnabled(true)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SimonPanelOverview:playAppear()
  local mainColumn = self.mainColumn
  local headerLabel = self.headerLabel
  local chunkingParagraphLabel = self.chunkingParagraphLabel
  local gameParagraphLabel = self.gameParagraphLabel

  mainColumn:alignChildren()
  headerLabel:setX(headerLabel:getX() - SLIDE_IN_DISTANCE)
  chunkingParagraphLabel:setX(chunkingParagraphLabel:getX() - SLIDE_IN_DISTANCE)
  gameParagraphLabel:setX(gameParagraphLabel:getX() - SLIDE_IN_DISTANCE)

  self.panelChoice = nil
  self.dismissed = false

  local chunkingColumn = self.chunkingColumn
  local gameOverviewColumn = self.gameOverviewColumn

  local staggerDuration = 0.05
  local appearAnimation = AHStaggeredSpawn.new(staggerDuration,
                                               AHSpawn.new(
                                                           AHLocationXBy.new(headerLabel, 0.5, AHEaseExponentialOut.new(), SLIDE_IN_DISTANCE),
                                                           AHOpacityTo.new(headerLabel, 0.5, nil, 1.0)
                                                           ),
                                               AHSpawn.new(
                                                           AHLocationXBy.new(chunkingParagraphLabel, 0.5, AHEaseExponentialOut.new(), SLIDE_IN_DISTANCE),
                                                           AHOpacityTo.new(chunkingParagraphLabel, 0.5, nil, 1.0)
                                                           ),
                                               AHSequence.new(
                                                              AHSpawn.new(
                                                                          AHLocationXBy.new(gameParagraphLabel, 0.5, AHEaseExponentialOut.new(), SLIDE_IN_DISTANCE),
                                                                          AHOpacityTo.new(gameParagraphLabel, 0.5, nil, 1.0)
                                                                          ),
                                                              function()
                                                                enableDismissal(self)
                                                              end
                                                              )
                                               )
  return appearAnimation:start()
end

function SimonPanelOverview:playDisappear()
  local headerLabel = self.headerLabel
  local chunkingParagraphLabel = self.chunkingParagraphLabel
  local gameParagraphLabel = self.gameParagraphLabel

  local staggerDuration = 0.05
  local appearAnimation = AHStaggeredSpawn.new(staggerDuration,
                                               AHSpawn.new(
                                                           AHLocationYBy.new(gameParagraphLabel, 0.5, AHEaseSimpleIn.new(4.0), SLIDE_OUT_DISTANCE),
                                                           AHOpacityTo.new(gameParagraphLabel, 0.5, nil, 0.0)
                                                           ),
                                               AHSpawn.new(
                                                           AHLocationYBy.new(chunkingParagraphLabel, 0.5, AHEaseSimpleIn.new(4.0), SLIDE_OUT_DISTANCE),
                                                           AHOpacityTo.new(chunkingParagraphLabel, 0.5, nil, 0.0)
                                                           ),
                                               AHSpawn.new(
                                                           AHLocationYBy.new(headerLabel, 0.5, AHEaseSimpleIn.new(4.0), SLIDE_OUT_DISTANCE),
                                                           AHOpacityTo.new(headerLabel, 0.5, nil, 0.0)
                                                           )
                                               )
  appearAnimation:start()
  return AHDelay.new(0.5):start()
end

function SimonPanelOverview:isDismissed()
  return self.dismissed
end

function SimonPanelOverview:getPanelChoice()
  return nil
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureDismissCallback(self)
  local submitButton = self.submitButton
  local function onDismissCallback()
    submitButton:playPressed()
    self:setTouchEnabled(false)
    self.dismissed = true
  end
  self.onDismissCallback = onDismissCallback
  Button.makeButton(self, onDismissCallback)
  self:setTouchEnabled(false)
end

local function createHeaderLabel(text)
  local label = Label.new(text, Font.new(FontFaces.DIN_OT_MEDIUM, 16.0))
  return label
end

local function createParagraphLabel(text, textColor)
  local width = Screen.getWidthPart(0.8)
  local fontFace = FontFaces.DIN_OT_MEDIUM
  local fontSize = 18
  local attributesForLabels = nil
  local horizontalAlignment = Alignment.LEFT
  local label = SimonTextLabel.new(text, width, fontFace, fontSize, attributesForLabels, horizontalAlignment)
  label:setColor(unpack(textColor))
  return label
end

-- Constructor ------------------------

function SimonPanelOverview.new(theme, submitButton)
  local self = Group.newFullScreen()
  Mixin.mix(self, SimonPanelOverview)

  self.submitButton = submitButton

  configureDismissCallback(self)

  local textColor = theme:getGrayedTextColor()

  local mainColumn = Column.new(18.0, Alignment.LEFT)

  local headerLabel = createHeaderLabel("SEQUENCING OVERVIEW")
  headerLabel:setOpacity(0.0)
  mainColumn:insert(headerLabel)
  self.headerLabel = headerLabel

  local chunkingParagraphLabel = createParagraphLabel("This game utilizes the chunking strategy, which helps memorization by breaking down a long string of information into smaller parts.", textColor)
  chunkingParagraphLabel:setOpacity(0.0)
  mainColumn:insert(chunkingParagraphLabel)
  self.chunkingParagraphLabel = chunkingParagraphLabel

  local gameParagraphLabel = createParagraphLabel("The material is split into two audio clips. Listen to each and order the items. Practice each set again without listening. Then sequence all the items together.", textColor)
  gameParagraphLabel:setOpacity(0.0)
  mainColumn:insert(gameParagraphLabel)
  self.gameParagraphLabel = gameParagraphLabel

  mainColumn:alignChildren()
  mainColumn:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.55))

  self:insert(mainColumn, Z_MAIN_COLUMN)
  self.mainColumn = mainColumn

  self.dismissed = false

  return self
end
