--- SimonValidator
-- @author Created by Aaron Barrett on February 23, 2015.
-- ConceptValidator for Simon.

class("SimonValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "steps",
  "audio_filename_1",
  "audio_filename_2",
  "steps_count_1",
  "steps_count_2",
  "description",
  "category"
}

local CATEGORIES = {
  "Directions",
  "Dates",
  "Media"
}

local MIN_STEPS_COUNT_PER_CLIP = 3
local MAX_STEPS_COUNT_PER_CLIP = 5
local MAX_TOTAL_STEPS_COUNT = 10

local MAX_DESCRIPTION_STRING_LENGTH = 32
local MAX_STEP_STRING_LENGTH = 31

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateCategory(self, concept)
  local category = concept["category"]
  self:expectIsContained(category, CATEGORIES, "category")
end

local function validateStepsCounts(self, concept)
  local stepsCount1 = concept["steps_count_1"]
  self:expectNumberInRange(stepsCount1, MIN_STEPS_COUNT_PER_CLIP, MAX_STEPS_COUNT_PER_CLIP, "steps_count_1")

  local stepsCount2 = concept["steps_count_2"]
  self:expectNumberInRange(stepsCount2, MIN_STEPS_COUNT_PER_CLIP, MAX_STEPS_COUNT_PER_CLIP, "steps_count_2")

  local totalStepsCount = stepsCount1 + stepsCount2
  self:expectLessOrEqual(totalStepsCount, MAX_TOTAL_STEPS_COUNT, "The sum of `steps_count_1` and `steps_count_2` must not exceed " .. MAX_TOTAL_STEPS_COUNT .. ".")

  local steps = concept["steps"]
  local actualStepsCount = #steps
  self:expectEqual(actualStepsCount, totalStepsCount, "`steps_count_1` is " .. stepsCount1 ..
                   ", and `steps_count_2` is " .. stepsCount2 .. ", but there were " .. actualStepsCount .. " steps provided.")
end

local function validateStringLengths(self, concept)
  local description = concept["description"]
  local descriptionLength = #description
  self:expectLessOrEqual(descriptionLength, MAX_DESCRIPTION_STRING_LENGTH, "Description `" .. description .. "` is longer than " .. MAX_DESCRIPTION_STRING_LENGTH .. " characters.")

  local steps = concept["steps"]
  for stepIndex, step in ipairs(steps) do
    local stepString = step.item
    local stepStringLength = #stepString
    self:expectLessOrEqual(stepStringLength, MAX_STEP_STRING_LENGTH, "Step `" .. stepString .. "` is longer than " .. MAX_STEP_STRING_LENGTH .. " characters.")
  end
end

local function validatePrintableFields(self, concept)
  local description = concept["description"]
  self:expectStringIsPrintable(description, "description")

  local steps = concept["steps"]
  for stepIndex, step in ipairs(steps) do
    local stepString = step.item
    local stepName = "step " .. stepIndex
    self:expectStringIsPrintable(stepString, stepName)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SimonValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateCategory(self, concept)
  validateStepsCounts(self, concept)
  validateStringLengths(self, concept)
  validatePrintableFields(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SimonValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, SimonValidator)

  return self
end
