--- SquaresBalloonGenerator
-- @author Created by Aaron Barrett on January 9, 2014.

class("SquaresBalloonGenerator")
-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BASE_INTERVAL = 0.6
local MAX_ADDITIONAL_INTERVAL = 8.8

local INITIAL_SKILL_FACTOR = 0.60
local INITIAL_TIME_FACTOR = 1.0

local TIME_FACTOR_RATE = 0.985

local QUICK_REFILL_INTERVAL = 0.5

local MAX_MATCHES_FOR_QUICK_REFILL = 4

local EASY_GENERATION_CUTOFF = 0.25
local MEDIUM_GENERATION_CUTOFF = 0.45

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function needsQuickRefill(self)
  local balloonTracker = self.balloonTracker
  local matchesCount = balloonTracker:getMatchesCount()

  local doesNeedQuickRefill = (matchesCount <= MAX_MATCHES_FOR_QUICK_REFILL)
  return doesNeedQuickRefill
end

local function startInterval(self, intervalDuration)
  local intervalAction = AHDelay.new(intervalDuration)
  self.intervalAction = intervalAction
  wait(intervalAction:start())
end

local function createBalloon(self, skipPowerup)
  self.timeFactor = self.timeFactor * TIME_FACTOR_RATE

  local game = self.game
  local balloonTracker = self.balloonTracker

  local concept, kind = balloonTracker:nextConceptAndKind()
  game:sendBalloon(concept, kind, skipPowerup)
end

local function createMatchingInitialBalloons(self)
  local conceptAndKinds = self.balloonTracker:getInitialConceptAndKinds()
  local concept = conceptAndKinds.concept
  local skipPowerup = true
  self.game:sendBalloon(concept, conceptAndKinds.kindA, skipPowerup)
  startInterval(self, BASE_INTERVAL)
  self.game:sendBalloon(concept, conceptAndKinds.kindB, skipPowerup)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SquaresBalloonGenerator:stop()
  self.generationCoroutine:stop()
end

function SquaresBalloonGenerator:setPaused(isPaused)
  self.isPaused = isPaused
  local intervalAction = self.intervalAction
  intervalAction:pause(isPaused)
end

function SquaresBalloonGenerator:start()
  self.isPaused = false
  local generationCoroutine = self.generationCoroutine
  generationCoroutine:start()
end

function SquaresBalloonGenerator:balloonsPopped()
  self.timesBalloonsWerePopped = self.timesBalloonsWerePopped + 1
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SquaresBalloonGenerator.new(game, balloonTracker)
  local self = {}
  Mixin.mix(self, SquaresBalloonGenerator)

  self.isPaused = true

  self.game = game
  self.balloonTracker = balloonTracker

  self.skillFactor = INITIAL_SKILL_FACTOR
  self.timeFactor = INITIAL_TIME_FACTOR
  self.difficultyFactor = SquaresConfiguration.getBalloonGenerationDifficultyFactor()

  self.timesBalloonsWerePopped = 0

  local defaultGenerationFlow = function()
    while (true) do
      while (needsQuickRefill(self)) do
        createBalloon(self)
        startInterval(self, QUICK_REFILL_INTERVAL)
      end

      createBalloon(self)

      local skillFactor = self.skillFactor
      local timeFactor = self.timeFactor
      local difficultyFactor = self.difficultyFactor
      local intervalDuration = BASE_INTERVAL + (MAX_ADDITIONAL_INTERVAL * skillFactor * timeFactor * difficultyFactor)

      startInterval(self, intervalDuration)
    end
  end

  local switchToDefaultGenerationFlow = function()
    self.generationCoroutine:stop()
    self.generationCoroutine = newCoroutineWithFunction(defaultGenerationFlow)
    self.generationCoroutine:start()
  end

  local createBalloonsWithMatchingPair = function(skipPowerup)
    createMatchingInitialBalloons(self)
    startInterval(self, BASE_INTERVAL)
    createBalloon(self, skipPowerup)
  end

  local easyGenerationFlow = function()
    local skipPowerup = true
    createBalloonsWithMatchingPair(skipPowerup)
    waitUntil(function() return (self.timesBalloonsWerePopped > 0) end)
    createBalloonsWithMatchingPair(skipPowerup)
    waitUntil(function() return (self.timesBalloonsWerePopped > 1) end)
    switchToDefaultGenerationFlow()
  end

  local mediumGenerationFlow = function()
    local skipPowerup = true
    createBalloonsWithMatchingPair(skipPowerup)
    waitUntil(function() return (self.timesBalloonsWerePopped > 0) end)
    switchToDefaultGenerationFlow()
  end

  local generationFlow = defaultGenerationFlow
  local difficulty = GameSession.getDifficulty()
  if (difficulty <= EASY_GENERATION_CUTOFF) then
    generationFlow = easyGenerationFlow
  elseif (difficulty <= MEDIUM_GENERATION_CUTOFF) then
    generationFlow = mediumGenerationFlow
  end

  self.generationCoroutine = newCoroutineWithFunction(generationFlow)

  return self
end
