--- SquaresBalloonTracker
-- @author Created by Aaron Barrett on January 9, 2014.

class("SquaresBalloonTracker")
-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UNMATCHED_BALLOON_LIMIT = 6
local NEW_CONCEPT_PROCLIVITY = 1

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateKindDistribution(self, increment)
  self.kindDistribution = self.kindDistribution + increment
end

local function updateUnmatchedBalloonCount(self, originalDistribution, updatedDistribution)
  local didNumberOfMatchesIncrease = (math.abs(updatedDistribution) < math.abs(originalDistribution))
  if (didNumberOfMatchesIncrease) then
    self.unmatchedBalloonCount = self.unmatchedBalloonCount - 1
  else
    self.unmatchedBalloonCount = self.unmatchedBalloonCount + 1
  end
end

local function addConceptAndKind(self, concept, kind)
  local identifier = concept.identifier

  local entry = self.entries[identifier]
  if (entry == nil) then
    entry = {
            distribution = 0,
            concept = concept
            }
    self.entries[identifier] = entry
  end

  local originalDistribution = entry.distribution
  local increment = kind.distributionIncrement
  local updatedDistribution = originalDistribution + increment
  entry.distribution = updatedDistribution

  updateUnmatchedBalloonCount(self, originalDistribution, updatedDistribution)
  updateKindDistribution(self, increment)

  self.balloonCount = self.balloonCount + 1
end

local function chooseRandomConcept(self)
  local availableConcepts = self.availableConcepts
  local randomConcept = availableConcepts[math.random(#availableConcepts)]
  return randomConcept
end

local function chooseRandomKind(self)
  local availableKinds = self.availableKinds
  local randomKind = availableKinds[math.random(#availableKinds)]
  return randomKind
end

local function tryToChooseConceptByWeight(self)
  local totalWeight = self.unmatchedBalloonCount
  local randomlyChosenWeight = math.random(totalWeight)
  local remainingWeight = randomlyChosenWeight

  for identifier, entry in pairs(self.entries) do
    local entryDistribution = entry.distribution
    local entryWeight = math.abs(entryDistribution)
    remainingWeight = remainingWeight - entryWeight

    local isOutOfWeight = (remainingWeight <= 0)
    if (isOutOfWeight) then

      local kind = nil
      if (entryDistribution > 0) then
        kind = self.kindA
      elseif (entryDistribution < 0) then
        kind = self.kindB
      else
        error("Ran out of weight, but last entry's distribution is 0.  Distributions must be messed up.")
      end

      local existingConcept = entry.concept

      return existingConcept, kind
    end
  end

  error("Still " .. remainingWeight .. " weight remaining.  Distributions must be messed up.")
end

local function tryToChooseUnseenConcept(self)
  local randomlyChosenConcept = chooseRandomConcept(self)
  local randomlyChosenIdentifier = randomlyChosenConcept.identifier

  local existingEntry = self.entries[randomlyChosenIdentifier]
  if (existingEntry) then
    -- print("FAILED TO FIND UNSEEN CONCEPT")

    -- return tryToChooseConceptByWeight(self)

    local existingDistribution = existingEntry.distribution

    local kind = nil
    if (existingDistribution > 0) then
      kind = self.kindA
    elseif (existingDistribution < 0) then
      kind = self.kindB
    elseif (self.kindDistribution > 0) then
      kind = self.kindA
    elseif (self.kindDistribution < 0) then
      kind = self.kindB
    else
      kind = chooseRandomKind(self)
    end

    local existingConcept = existingEntry.concept

    return existingConcept, kind
  end

  local kind = nil
  if (self.kindDistribution > 0) then
    kind = self.kindA
  elseif (self.kindDistribution < 0) then
    kind = self.kindB
  else
    kind = chooseRandomKind(self)
  end

  return randomlyChosenConcept, kind
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SquaresBalloonTracker:getMatchesCount()
  local balloonCount = self.balloonCount
  local unmatchedBalloonCount = self.unmatchedBalloonCount
  local matchesCount = balloonCount - unmatchedBalloonCount
  return matchesCount
end

function SquaresBalloonTracker:nextConceptAndKind()
  local unmatchedBalloonCount = self.unmatchedBalloonCount

  local maxRandomIndex = unmatchedBalloonCount
  if (unmatchedBalloonCount < UNMATCHED_BALLOON_LIMIT) then
    maxRandomIndex = maxRandomIndex + NEW_CONCEPT_PROCLIVITY
  end

  local concept = nil
  local kind = nil

  local randomIndex = math.random(maxRandomIndex)
  local shouldTryToChooseUnseenConcept = (randomIndex > unmatchedBalloonCount)
  if (shouldTryToChooseUnseenConcept) then
    concept, kind = tryToChooseUnseenConcept(self)
  else
    concept, kind = tryToChooseConceptByWeight(self)
  end

  addConceptAndKind(self, concept, kind)
  return concept, kind
end

function SquaresBalloonTracker:getInitialConceptAndKinds()
  local concept = tryToChooseUnseenConcept(self)
  addConceptAndKind(self, concept, self.kindA)
  addConceptAndKind(self, concept, self.kindB)

  local conceptAndKinds = {
    concept = concept,
    kindA = self.kindA,
    kindB = self.kindB
  }

  return conceptAndKinds
end

function SquaresBalloonTracker:removeBalloon(balloon)
  self.balloonCount = self.balloonCount - 1

  local identifier = balloon:getIdentifier()
  local kind = balloon:getKind()
  local increment = kind.distributionIncrement

  local entry = self.entries[identifier]
  local originalDistribution = entry.distribution
  local updatedDistribution = originalDistribution - increment
  entry.distribution = updatedDistribution

  updateUnmatchedBalloonCount(self, originalDistribution, updatedDistribution)
  updateKindDistribution(self, -increment)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SquaresBalloonTracker.new(kindA, kindB, availableConcepts)
  local self = {}
  Mixin.mix(self, SquaresBalloonTracker)

  self.kindA = kindA
  self.kindB = kindB
  self.availableKinds = {kindA, kindB}
  self.availableConcepts = availableConcepts

  self.entries = {}

  self.balloonCount = 0
  self.unmatchedBalloonCount = 0
  self.kindDistribution = 0

  return self
end
