--- SquaresFloatingHeaderOverlay
-- @author Created by Aaron Barrett on November 12, 2021.

class("SquaresFloatingHeaderOverlay")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_TOP_MARGIN_COVER = 10
local Z_HEADER = 20

local SCROLL_DIRECTION_TIPPING_POINT = 0.01

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function activateHeader(self, headerToActivate)
  local hasSeenHeaderToActivate = false
  for _, header in ipairs(self.headers) do
    if (header == headerToActivate) then
      header:setState(SquaresHeaderStates.ACTIVE)
      hasSeenHeaderToActivate = true
    elseif (hasSeenHeaderToActivate) then
      header:setState(SquaresHeaderStates.VISIBLE)
    else
      header:setState(SquaresHeaderStates.HIDDEN)
    end
  end

  self.activeHeader = headerToActivate
end

local function takeHeadersWithTopHeaderAtBottomZIndex(self)
  for _, header in rpairs(self.headers) do
    self:takeChild(header, Z_HEADER)
  end
end

local function arrangeHeaders(self, offsetY)
  self.topY = self.topHeader:getY() + offsetY
  self.activationY = self.topY - self.topHeader:getLabelHeight()
  local offScreenY = self:getHeight() + self.topHeader:getLabelHeight() * 0.5

  for _, header in ipairs(self.headers) do
    header:setHomeY(header:getY() + offsetY)
    header:setTopY(self.topY)
    header:setOffScreenY(offScreenY)
  end
end

local function initializeHeaderStates(self)
  for _, header in ipairs(self.headers) do
    local headerIsTopHeader = (header == self.topHeader)
    header:initializeState(headerIsTopHeader)
  end
end

local function getThresholdY(self)
  local moderatedScrollSpeed = self.scrollMotionMeasurer:getCurrentModeratedScrollSpeed()

  if (moderatedScrollSpeed <= SCROLL_DIRECTION_TIPPING_POINT) then
    return self.topY
  else
    return self.activationY
  end
end

local function updateHeaderLocations(self, scrollY)
  for _, header in ipairs(self.headers) do
    if (header:getState() == SquaresHeaderStates.VISIBLE) then
      header:stopHideOrShowAction()
      header:setY(header:getHomeY() + scrollY)
    end
  end
end

local function getLowestHeaderAboveThresholdY(self, scrollY, thresholdY)
  local lowestHeader = self.topHeader

  for _, header in ipairs(self.headers) do
    local effectiveY = header:getHomeY() + scrollY
    if (effectiveY > thresholdY) then
      lowestHeader = header
    end
  end

  return lowestHeader
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SquaresFloatingHeaderOverlay:configureHeaders(headers, offsetY)
  self.headers = headers

  self.topHeader = headers[1]
  self.activeHeader = self.topHeader

  takeHeadersWithTopHeaderAtBottomZIndex(self)
  arrangeHeaders(self, offsetY)
  initializeHeaderStates(self)
end

function SquaresFloatingHeaderOverlay:updateScrollY(scrollY)
  local deltaScrollY = scrollY - self.previousScrollY
  self.scrollMotionMeasurer:takeSnapshot(deltaScrollY)

  local thresholdY = getThresholdY(self)
  local lowestHeaderAboveThresholdY = getLowestHeaderAboveThresholdY(self, scrollY, thresholdY)

  updateHeaderLocations(self, scrollY)

  if (lowestHeaderAboveThresholdY ~= self.activeHeader) then
    activateHeader(self, lowestHeaderAboveThresholdY)
  end

  self.previousScrollY = scrollY
end

function SquaresFloatingHeaderOverlay:addTopMarginCover()
  local coverHeight = self:getHeight() - self.topY + self.topHeader:getLabelHeight() * 0.5
  local topMarginCover = FilledRectangle.new(self:getWidth(), coverHeight)
  topMarginCover:setColor(unpack(SquaresColors.SCORE_CARD_BACKGROUND_COLOR))

  topMarginCover:setX(self:getWidthPart(0.5))
  topMarginCover:setY(self:getHeight() - coverHeight * 0.5)
  self:insert(topMarginCover, Z_TOP_MARGIN_COVER)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SquaresFloatingHeaderOverlay.new(width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, SquaresFloatingHeaderOverlay)

  self.headers = nil
  self.topHeader = nil
  self.activeHeader = nil
  self.topY = nil
  self.previousScrollY = 0.0

  self.scrollMotionMeasurer = SquaresScrollMotionMeasurer.new()

  return self
end
