--- SquaresFocusTracker
-- @author Created by Aaron Barrett on December 16, 2013.

-- Note on touch management:
-- `squares` has some more specific touch handling requirements that currently
-- aren't possible to implement using the standard functionality of the
-- `TouchManager`.  Specifically, balloons need to consume `TOUCH_BEGAN` events
-- only when those events fall within their circle's radius, and when they fall
-- outside of the radius, they still should be handed off to the touch handlers
-- for other balloons and for the background.  While the `TouchManager` supports
-- handing off touches for `TOUCH_MOVED` events, it doesn't currently allow
-- handoffs for `TOUCH_BEGAN`.
--
-- To work around this, `squares` uses `TouchHandler.BEHAVIOR_ALWAYS` for all
-- balloons and for the background, so that the `TouchManager` will *always*
-- call the touch handler callbacks for any balloons whose bounding boxes
-- intersect the touch event.  Whenever a touch intersects with a balloon's
-- circle, the balloon sets the `hasTouchBeenIntercepted` flag to `true` on
-- `SquaresFocusTracker` to "consume" the touch event.  Other balloons and the
-- background check this flag and only call their touch handling logic when it
-- is `false`.  When the touch event ultimately hits the background's touch
-- handler, the background resets the `hasTouchBeenIntercepted` flag to `false`
-- to prepare for the next `TOUCH_BEGAN` event.

class("SquaresFocusTracker")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function doesMatchExistOnScreen(balloon, trackedBalloons)
  for otherBalloon in pairs(trackedBalloons) do
    if ((balloon ~= otherBalloon) and (balloon:doesMatch(otherBalloon)) and (balloon:getIsOnScreen()) and (otherBalloon:getIsOnScreen())) then
      return true
    end
  end
  return false
end

local function areMatchesOnScreen(self)
  local balloonTracker = self.balloonTracker

  if (balloonTracker:getMatchesCount() == 0) then
    return false
  end

  local trackedBalloons = self.trackedBalloons

  for balloon in pairs(trackedBalloons) do
    if (doesMatchExistOnScreen(balloon, trackedBalloons)) then
      return true
    end
  end

  return false
end

local function checkToStopSnowEarly(self)
  local snowMachine = self.snowMachine
  local isSnowing = snowMachine:getIsSnowing()
  if (isSnowing) then
    local shouldStopSnowEarly = (not(areMatchesOnScreen(self)))
    if (shouldStopSnowEarly) then
      snowMachine:unfreeze()
    end
  end
end

local function addTouchHandlerAfterOneFrame(self, balloon)
  local function onBalloonTouched(event, target)
    if ((self.isTouchEnabled) and (not self.hasTouchBeenIntercepted)) then
      local phase = event.phase
      if (phase == TouchManager.TOUCH_BEGAN) then
        local circle = target
        local balloon = circle.balloon

        local isBalloonTracked = self.trackedBalloons[balloon]
        if (isBalloonTracked) then
          local x = event.x
          local y = event.y
          local balloonX, balloonY = balloon:getPosition()

          local distance = MathUtils.distance(x, y, balloonX, balloonY)
          local radius = balloon:getRadius()

          local isTouchWithinRadius = (distance < radius)
          if (isTouchWithinRadius) then
            self.hasTouchBeenIntercepted = true
            self:onBalloonChosen(balloon)
          end
        end
      end
    end
  end

  local addTouchHandlerCoroutine = newCoroutineWithFunction(function()
    coroutine.yield()
    balloon:addTouchHandlerToCircle(onBalloonTouched)
  end)
  addTouchHandlerCoroutine:start()
end

local function growPair(self, balloonA, balloonB)
  local growAmount = SquaresConstants.PENALTY_RESIZE_AMOUNT
  local delayDuration = 0.0
  local growAction = AHSpawn.new(
                                 balloonA:resize(growAmount, delayDuration),
                                 balloonB:resize(growAmount, delayDuration)
                                 )
  growAction:start()

  Sound.playSoundEffect("grow")
end

local function filterBalloonsForPopping(...)
  local balloonsToFilter = arg

  local balloonsToPopNow = {}
  local balloonsWithSnow = {}
  local balloonsWithOtherPowerups = {}

  for i, balloon in ipairs(balloonsToFilter) do
    local flavor = balloon:getPowerupFlavor()
    if (flavor == SquaresPowerup.FLAVOR_SHRINKER or flavor == SquaresPowerup.FLAVOR_FIREWORK or flavor == SquaresPowerup.FLAVOR_BOW) then
      table.insert(balloonsWithOtherPowerups, balloon)
    else
      table.insert(balloonsToPopNow, balloon)
      if (flavor == SquaresPowerup.FLAVOR_SNOW) then
        table.insert(balloonsWithSnow, balloon)
      end
    end
  end

  return balloonsToPopNow, balloonsWithSnow, balloonsWithOtherPowerups
end

local function popPair(self, balloonA, balloonB)
  self.isTouchEnabled = false

  local flasher = self.flasher
  local popParticles = self.popParticles
  local physics = self.physics
  local snowMachine = self.snowMachine
  local balloonTracker = self.balloonTracker
  local balloonGenerator = self.balloonGenerator

  local trackedBalloons = self.trackedBalloons
  trackedBalloons[balloonA] = nil
  trackedBalloons[balloonB] = nil
  balloonTracker:removeBalloon(balloonA)
  balloonTracker:removeBalloon(balloonB)

  local balloonsToPopNow, balloonsWithSnow, balloonsWithOtherPowerups = filterBalloonsForPopping(balloonA, balloonB)
  local poppedBalloonCount = #balloonsToPopNow + #balloonsWithOtherPowerups

  runInCoroutine(function()
    local areThereBalloonsToPopNow = (#balloonsToPopNow > 0)
    if (areThereBalloonsToPopNow) then
      for i, balloon in ipairs(balloonsToPopNow) do
        popParticles:show(balloon:getPosition())

        local allowPowerups = true
        balloon:pop(allowPowerups)
      end
      Sound.playSoundEffect("pop")
    end

    local areThereBalloonsWithSnow = (#balloonsWithSnow > 0)
    if (areThereBalloonsWithSnow) then
      flasher:flash()
      for i, balloon in ipairs(balloonsWithSnow) do
        snowMachine:freeze()
      end
      Sound.playSoundEffect("freeze")
    end

    local areThereBalloonsWithOtherPowerups = (#balloonsWithOtherPowerups > 0)
    if (areThereBalloonsWithOtherPowerups) then

      physics:setPaused(true)
      balloonGenerator:setPaused(true)
      for i, balloon in ipairs(balloonsWithOtherPowerups) do
        balloon:stopBowIcon()
      end

      local postPowerupDuration = nil

      for i, balloon in ipairs(balloonsWithOtherPowerups) do
        local effect = balloon:createEffect(trackedBalloons, popParticles)

        local hitBalloons = effect:getHitBalloons()
        poppedBalloonCount = poppedBalloonCount + #hitBalloons
        for i, balloon in ipairs(hitBalloons) do
          trackedBalloons[balloon] = nil
          physics:removeBalloon(balloon)
          balloonTracker:removeBalloon(balloon)
        end

        local chargeUpAction = balloon:createChargeUpAction()
        snowMachine:remainFrozenForExtraDuration(SquaresBalloon.CHARGE_UP_DURATION)
        wait(chargeUpAction:start())

        popParticles:show(balloon:getPosition())

        local allowPowerups = true
        balloon:pop(allowPowerups)

        flasher:flash()

        local effectGroup = effect:getGroup()
        if (effectGroup) then
          local effectPlane = self.effectPlane
          effectPlane:insert(effectGroup)
        end

        local effectAction = effect:getAction()
        effectAction:start()

        postPowerupDuration = effect:getDelayDuration()
      end

      snowMachine:remainFrozenForExtraDuration(postPowerupDuration)
      waitFor(postPowerupDuration)

      checkToStopSnowEarly(self)

      local isSnowing = snowMachine:getIsSnowing()
      if (not(isSnowing)) then
        physics:setPaused(false)
        balloonGenerator:setPaused(false)
      end

    else
      checkToStopSnowEarly(self)
    end

    physics:removeBalloon(balloonA)
    physics:removeBalloon(balloonB)

    self.onBalloonsPopped(poppedBalloonCount)

    self.isTouchEnabled = true
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SquaresFocusTracker:disableTouchOnAllBalloons()
  for balloon in pairs(self.trackedBalloons) do
    balloon:removeTouchHandler()
  end
end

function SquaresFocusTracker:getMatchedBalloonPair()
  for balloon in pairs(self.trackedBalloons) do
    for matchBalloon in pairs(self.trackedBalloons) do
      if ((balloon:getKind() ~= matchBalloon:getKind()) and (balloon:doesMatch(matchBalloon))) then
        return {balloon, matchBalloon}
      end
    end
  end
  return nil
end

function SquaresFocusTracker:popAllBalloons()
  return runInCoroutine(function()
    local delay = 0.02
    for balloon in pairs(self.trackedBalloons) do
      Sound.playSoundEffect("pop", 0.5)
      balloon:pop(false)
      waitFor(delay)
    end

    local extraDurationSoBalloonsCanFinishPopping = 0.6
    waitFor(extraDurationSoBalloonsCanFinishPopping)
  end)
end

function SquaresFocusTracker:fadeAllBalloonLabels()
  return runInCoroutine(function()
      for balloon in pairs(self.trackedBalloons) do
        balloon:fadeLabel()
      end

      local fadeDelayDuration = 0.4
      waitFor(fadeDelayDuration)
    end)
end

function SquaresFocusTracker:destroy()
  self.isTouchEnabled = false
end

function SquaresFocusTracker:addBalloon(balloon)
  local trackedBalloons = self.trackedBalloons
  trackedBalloons[balloon] = true

  addTouchHandlerAfterOneFrame(self, balloon)
end

function SquaresFocusTracker:removeBalloon(balloon)
  local trackedBalloons = self.trackedBalloons
  trackedBalloons[balloon] = nil
end

function SquaresFocusTracker:focusBalloon(balloonToFocus)
  local trackedBalloons = self.trackedBalloons

  local kindToGray = balloonToFocus:getKind()

  for balloon in pairs(trackedBalloons) do
    local balloonKind = balloon:getKind()
    if (balloonKind == kindToGray) then
      if (balloon == balloonToFocus) then
        balloon:setFocusState(SquaresBalloon.FOCUS_STATE_FOCUSED)
      else
        balloon:setFocusState(SquaresBalloon.FOCUS_STATE_GRAYED)
      end
    else
      balloon:setFocusState(SquaresBalloon.FOCUS_STATE_NORMAL)
    end
  end

  self.focusedBalloon = balloonToFocus

  Sound.playSoundEffect("select")
end

function SquaresFocusTracker:clearFocus()
  local trackedBalloons = self.trackedBalloons

  for balloon in pairs(trackedBalloons) do
    balloon:setFocusState(SquaresBalloon.FOCUS_STATE_NORMAL)
  end

  self.focusedBalloon = nil
end

function SquaresFocusTracker:onBalloonChosen(balloon)
  local focusedBalloon = self.focusedBalloon
  self.hasBeenTouched = true

  if (self.focusedBalloon == nil) then
    self:focusBalloon(balloon)
    return
  end

  local isBalloonAlreadyFocused = (balloon == focusedBalloon)
  if (isBalloonAlreadyFocused) then
    self:clearFocus()
    Sound.playSoundEffect("deselect")
    return
  end

  local isBalloonSameKindAsFocusedBalloon = (balloon:getKind() == focusedBalloon:getKind())
  if (isBalloonSameKindAsFocusedBalloon) then
    self:focusBalloon(balloon)
    return
  else
    self:clearFocus()

    local didChooseMatchingPair = (balloon:doesMatch(focusedBalloon))
    if (didChooseMatchingPair) then
      self.onCorrectAnswer(balloon:getConcept())
      popPair(self, balloon, focusedBalloon)
    else
      self.onIncorrectAnswer(balloon:getIdentifier(), focusedBalloon:getIdentifier())
      growPair(self, balloon, focusedBalloon)
    end
    return
  end

end

function SquaresFocusTracker:onBackgroundChosen()
  self:clearFocus()
  Sound.playSoundEffect("deselect")
end

function SquaresFocusTracker:onBackgroundTouchBegan()
  if (not self.hasTouchBeenIntercepted) then
    self:onBackgroundChosen()
  end
  self.hasTouchBeenIntercepted = false
end

function SquaresFocusTracker:onTouchGuardTouchBegan()
  self.hasTouchBeenIntercepted = true
end

function SquaresFocusTracker:getFocusedKind()
  local focusedBalloon = self.focusedBalloon
  if (focusedBalloon) then
    local focusedKind = focusedBalloon:getKind()
    return focusedKind
  end
  return nil
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SquaresFocusTracker.new(kindA, kindB, balloonTracker, physics, balloonGenerator, snowMachine, flasher, popParticles, effectPlane, onBalloonsPopped, onCorrectAnswer, onIncorrectAnswer)
  local self = {}
  Mixin.mix(self, SquaresFocusTracker)

  self.isTouchEnabled = true

  self.kindA = kindA
  self.kindB = kindB
  self.balloonTracker = balloonTracker
  self.physics = physics
  self.balloonGenerator = balloonGenerator
  self.snowMachine = snowMachine
  self.flasher = flasher
  self.popParticles = popParticles
  self.effectPlane = effectPlane
  self.onBalloonsPopped = onBalloonsPopped
  self.onCorrectAnswer = onCorrectAnswer
  self.onIncorrectAnswer = onIncorrectAnswer
  self.hasBeenTouched = false

  self.trackedBalloons = {}

  self.focusedBalloon = nil
  self.hasTouchBeenIntercepted = false

  return self
end
