--- SquaresGame
-- @author Created by Aaron Barrett on December 16, 2013.

class("SquaresGame")
require("Particles")
require("SquaresKinds")
require("SquaresConstants")
require("SquaresColors")
require("SquaresCounter")
require("SquaresBackground")
require("SquaresFocusTracker")
require("SquaresBalloon")
require("SquaresPhysics")
require("SquaresHalo")
require("SquaresPowerup")
require("SquaresIcon")
require("SquaresFlasher")
require("SquaresBalloonGenerator")
require("SquaresBalloonTracker")
require("SquaresSnowMachine")
require("SquaresScanner")
require("SquaresEffectShrinker")
require("SquaresEffectFirework")
require("SquaresEffectBow")
require("SquaresPopParticles")
require("SquaresConfiguration")
require("StandardUIBuilder")
require("SquaresContentTracker")
require("SquaresTipLocalizer")
require("SquaresPauseButtonTouchGuard")

require("SquaresMatchRecord")
require("SquaresMatchTracker")

require("SquaresScoreCard")
require("SquaresScoreScroller")
require("widgets.TouchScroller")
require("SquaresScrollMotionMeasurer")
require("SquaresScoreColumn")
require("SquaresScoreSlidingElement")
require("SquaresScoreCell")
require("SquaresScorePairIcon")
require("SquaresScoreContinueButton")
require("SquaresScoreHeader")
require("SquaresHeaderStates")
require("SquaresFloatingHeaderOverlay")
require("SquaresScoreFooter")
require("SquaresScoreDefinition")

require("tutorials.TutorialHighlighter")
require("tutorials.TutorialButton")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Z-Indices --------------------------

local Z_BACKGROUND = 10
local Z_POP_PARTICLES = 20
local Z_BALLOON = 30
local Z_POWERUP_EFFECTS = 40
local Z_SNOW_MACHINE = 50
local Z_FLASH = 60
local Z_TUTORIAL = 70
local Z_TIP = 80
local Z_SCORE_CARD = 90

-- Dimensions -------------------------

local WIDTH = Screen.width
local HEIGHT = Screen.height

local CENTER_X = WIDTH * 0.5
local CENTER_Y = HEIGHT * 0.5

local SNOW_MACHINE_Y = Screen.getHeightPart(1.15)

-- Tutorial ---------------------------

local TUTORIAL_TIP_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 20.0)

-- Difficulty -------------------------

local function configureGameDifficulty(self)
  local passScore = SquaresConfiguration.getPassScore()
  local rankTwoScore = SquaresConfiguration.getRankTwoScore()
  local rankThreeScore = SquaresConfiguration.getRankThreeScore()
  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  self.balloonPopPoints = SquaresConfiguration.POINTS_PER_CORRECT_RESPONSE
  self.passBalloonCount = SquaresConfiguration.CORRECT_NEEDED_TO_PASS
  self.incorrectAnswersCount = 0

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.CONSISTENCY_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function destroyGameObjects(self)
  self.scanner:destroy()
  self.physics:destroy()
  self.snowMachine:destroy()
  self.focusTracker:destroy()
  self.popParticles:destroy()
end

local function removeGameElements(self)
  self.background:removeSelf()
  self.popParticles:removeSelf()
  self.flasher:removeSelf()
  self.snowMachine:removeSelf()
end

local function getConcepts()
  local conceptGenerator = GameSession.conceptGenerator
  local difficultyKeyFrames = SquaresConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)
  conceptGenerator:setAllowFewerResults(true)

  local numberToPrepare = 35
  local concepts = conceptGenerator:nextAsArray(numberToPrepare)
  return concepts
end

local function trackPos(self, tracker, tracked, offsetX, offsetY)
  local x, y = self:transformLoc(tracked)
  x = x + offsetX
  y = y + offsetY
  tracker:setLoc(x, y)
end

local function showTutorial(self)
  local matchedPair = self.focusTracker:getMatchedBalloonPair()
  assert(matchedPair ~= nil, "No matched balloon pair found")
  local firstBalloon, secondBalloon = unpack(matchedPair)

  self.background:setTouchEnabled(false)
  self.focusTracker.isTouchEnabled = false
  self.balloonGenerator:setPaused(true)

  local trackingCoroutines = {}
  local lastBalloonTapped = nil

  local function getBalloonHole(balloon)
    local balloonCircle = balloon:getCircle()
    local hole = FilledCircle.new(balloonCircle:getScl(), {1, 1, 1, 1})
    local startX, startY = balloonCircle:getLoc()

    local tappedCallback = function()
      self.focusTracker:onBalloonChosen(balloon)
      lastBalloonTapped = balloon
    end
    local holeButton = TutorialButton.new(hole, tappedCallback)

    local function trackBalloon()
      while true do
        trackPos(self, holeButton, balloon, startX, startY)
        coroutine.yield()
      end
    end

    holeButton.trackCoroutine = runInCoroutine(trackBalloon)
    table.insert(trackingCoroutines, hole.trackCoroutine)

    return holeButton
  end

  local firstHoleFactory = function() return getBalloonHole(firstBalloon) end
  local secondHoleFactory = function() return getBalloonHole(secondBalloon) end

  local generateHoleCallbacks = {
    firstHoleFactory,
    secondHoleFactory
  }

  local tutorialLayer = TutorialHighlighter.newWithMasks(generateHoleCallbacks)
  tutorialLayer:setOverlayColor(unpack(SquaresColors.TUTORIAL_OVERLAY_COLOR))
  self:insertAtCenter(tutorialLayer, Z_TUTORIAL)
  tutorialLayer:createFadeInAnimation():start()

  tutorialLayer:waitForTargetTap()

  local halo = lastBalloonTapped.halo
  self:takeChild(halo, Z_TIP)

  local function trackHalo()
    while true do
      trackPos(self, halo, lastBalloonTapped, lastBalloonTapped.radius, lastBalloonTapped.radius)
      coroutine.yield()
    end
  end
  local haloCoroutine = runInCoroutine(trackHalo)
  table.insert(trackingCoroutines, haloCoroutine)

  tutorialLayer:waitForTargetTap()

  wait(tutorialLayer:createFadeOutAnimation():start())
  tutorialLayer:removeSelf()

  for _, coroutine in ipairs(trackingCoroutines) do
    if coroutine and coroutine:isBusy() then
      coroutine:stop()
    end
  end

  self.balloonGenerator:setPaused(false)
  self.focusTracker.isTouchEnabled = true
  self.background:setTouchEnabled(true)
end

-- Wait for a short delay to avoid a framerate drop when the initial balloons spawn in.
local function startScanningAfterDelay(self)
  local scanningStartDelay = 3.0
  local delayedScanningStart = AHDelayed.new(3.0, function() self.scanner:startScanning() end)
  return delayedScanningStart:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SquaresGame:sendBalloon(concept, kind, skipPowerup)
  self.totalBalloonCount = self.totalBalloonCount + 1
  if self.totalBalloonCount == 2 then
    self.tipAction = AHSequence.new(
                        AHDelay.new(2.0),
                        AHOpacityTo.new(self.tipLabel, 0.15, nil, 1.0)
                      )
    self.tipAction:start()
  end

  local radius = SquaresConstants.BASE_RADIUS + math.random() * SquaresConstants.MAX_RANDOM_RADIUS
  local powerupFlavor = SquaresPowerup.getNextPowerupFlavor()
  if (skipPowerup) then
    powerupFlavor = nil
  end
  local isGrayed = (kind == self.focusTracker:getFocusedKind())
  local balloon = SquaresBalloon.new(kind, concept, radius, powerupFlavor, isGrayed)

  self.focusTracker:addBalloon(balloon)
  self.matchTracker:notifyConceptPresented(concept)

  -- Move the balloon off-screen before adding it to the scene to prevent
  -- it from appearing at (0, 0) for one frame because of MOAI's node update bug
  balloon:setLoc(-Screen.width, -Screen.height)

  self:insert(balloon, Z_BALLOON)
  self.physics:addBalloon(balloon)

  -- Return the balloon to (0, 0) so it can mirror the physics body's "transform"
  balloon:setLoc(0.0, 0.0)

  balloon:startAnimating()
end

-- gameFlow ---------------------------

function SquaresGame:gameFlow()
  self.currentBalloonCount = 0
  self.counter:setMaxValue(self.passBalloonCount)

  self.physics:start()
  self.snowMachine:start()

  self.balloonGenerator:start()

  Timer.makeTimer(self, SquaresConfiguration.TUTORIAL_DELAY)
  self:setTimeoutHandler(function()
    if not self.focusTracker.hasBeenTouched then
      showTutorial(self)
    end
  end)
  self:start()
  
  startScanningAfterDelay(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local background = SquaresBackground.new()
  background:setLoc(CENTER_X, CENTER_Y)
  self:insert(background, Z_BACKGROUND)
  self.background = background

  local function onBackgroundTouched(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      self.focusTracker:onBackgroundTouchBegan()
    end
  end

  background:addTouchHandler(onBackgroundTouched)
end

local function createPauseButtonTouchGuard(self)
  local pauseButton = self.pauseButton
  local virtualButton = pauseButton:getVirtualButton()
  local width = virtualButton:getWidth()
  local height = virtualButton:getHeight()

  local function onTouchGuardTouched(event)
    if (event.phase == TouchManager.TOUCH_BEGAN) then
      self.focusTracker:onTouchGuardTouchBegan()
    end
  end

  self.pauseButtonTouchGuard = SquaresPauseButtonTouchGuard.new(width, height, onTouchGuardTouched)
  pauseButton:insertAtCenter(self.pauseButtonTouchGuard)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), SquaresColors.UI_COMPONENTS_COLOR)

  self.pauseButton = builder:addPauseButton()
  createPauseButtonTouchGuard(self)

  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  self.counter = SquaresCounter.new(nil, SquaresColors.UI_COMPONENTS_COLOR)
  self.counter:setAnchorPointRight()

  local counterXOffset = 12.0
  local counterYOffset = 0.0
  builder:addWidget(self.counter, UIOverlay.NORTHEAST, counterXOffset, counterYOffset)

  local tipProperties = {
    font = TUTORIAL_TIP_FONT,
    color = SquaresColors.TUTORIAL_TIP_TEXT_COLOR
  }
  local tipText = SquaresTipLocalizer.getLocalizedTip()
  self.tipLabel = DynamicTextArea.new(tipText, self:getWidthPart(0.9), tipProperties, nil, Alignment.CENTER, nil, 0, true)
  self:insert(self.tipLabel, Z_TIP)
  self.tipLabel:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.1))
  self.tipLabel:setOpacity(0.0)
  self.totalBalloonCount = 0
end

local function createKinds(self)
  self.kindA = SquaresKinds.KIND_A
  self.kindB = SquaresKinds.KIND_B
end

local function createBalloonTracker(self)
  local kindA = self.kindA
  local kindB = self.kindB
  local availableConcepts = self.concepts
  self.balloonTracker = SquaresBalloonTracker.new(kindA, kindB, availableConcepts)
end

local function createBalloonGenerator(self)
  local balloonTracker = self.balloonTracker

  self.balloonGenerator = SquaresBalloonGenerator.new(self, balloonTracker)
end

local function createSnowMachine(self)
  local physics = self.physics
  local balloonGenerator = self.balloonGenerator

  local function onFreezeCallback()
    physics:setPaused(true)
    balloonGenerator:setPaused(true)
  end

  local function onResumeCallback()
    physics:setPaused(false)
    balloonGenerator:setPaused(false)
  end

  local snowMachine = SquaresSnowMachine.new(onFreezeCallback, onResumeCallback)
  -- Moai seems to cull out props with no dimensions that are positioned off-screen,
  -- so just keep the snow system at 0.0, 0.0 and set its source position off-screen
  snowMachine:setLoc(0.0, 0.0)
  snowMachine:setSourcePosition(CENTER_X, SNOW_MACHINE_Y)
  self:insert(snowMachine, Z_SNOW_MACHINE)

  self.snowMachine = snowMachine
end

local function updateScoreLabel(self)
  local score = GameSession.getScore()
  self.scoreLabel:setScore(score)
end

local function playFadeOutUIElements(self)
  self.pauseButton:setTouchEnabled(false)

  local overlay = self:getUIOverlay()
  local duration = 0.2
  return overlay:fadeOut(duration)
end

local function endGame(self)
  self.hasGameEnded = true
  self.focusTracker:disableTouchOnAllBalloons()
  TouchManager.removeTouchHandler(self.background)
  self.balloonGenerator:stop()
end

local function onGameLoss(self)
  endGame(self)

  destroyGameObjects(self)
  self:showDefaultGameLossOutro()
  wait(self.focusTracker:fadeAllBalloonLabels())
  GameSession.notifyGameLoss()
end

local function onGameWin(self)
  endGame(self)

  Sound.playSoundEffect("game_win")

  runInCoroutine(function()
    wait(self:showDefaultGameWinOutro())
    wait(self.focusTracker:popAllBalloons())

    destroyGameObjects(self)

    playFadeOutUIElements(self)
    waitFor(0.2)

    local matchedConcepts = self.matchTracker:getMatchedConcepts()
    local unmatchedConcepts = self.matchTracker:getUnmatchedConcepts()
    local scoreCard = SquaresScoreCard.new(matchedConcepts, unmatchedConcepts)
    self:insertAtCenter(scoreCard, Z_SCORE_CARD)
    waitFor(0.1)
    wait(scoreCard:playShow())

    removeGameElements(self)
    waitUntil(function() return scoreCard:isDismissed() end)

    GameSession.notifyGameWin()
  end)
end

local function onBalloonsPopped(self, count)
  if (self.tipAction ~= nil) then
    self.tipAction:stop()
    self.tipAction = nil
  end

  if (self.tipLabel ~= nil) then
    self.tipLabel:removeSelf()
    self.tipLabel = nil
  end

  GameSession.increaseScore(self.balloonPopPoints * count)
  updateScoreLabel(self)
  self.currentBalloonCount = math.min(self.currentBalloonCount + count, self.passBalloonCount)
  self.counter:setValue(self.currentBalloonCount)
  self.balloonGenerator:balloonsPopped()
  if ((not self.hasGameEnded) and (self.currentBalloonCount == self.passBalloonCount)) then
    local accuracyBonus = SquaresConfiguration.getAccuracyBonusScoreForIncorrectCount(self.incorrectAnswersCount)
    GameSession.increaseScore(accuracyBonus)
    local consistencyBonusString = "BonusTray.CONSISTENCY_BONUS"
    self.bonusTray:indicateSimpleTextBonus(consistencyBonusString, accuracyBonus)
    self.scoreLabel:setScore(GameSession.getScore())
    GameSession.bonusReporter:setBonusCount(BonusTypes.TYPES.CONSISTENCY_BONUS, self.incorrectAnswersCount)

    onGameWin(self)
  end
end

local function onCorrectAnswer(self, concept)
  local conceptID = concept.identifier

  self.contentTracker:notifyCorrectAnswer(conceptID)
  GameSession.notifyCorrectAnswer(conceptID)
  GameSession.incrementCorrectAnswerCount()

  self.matchTracker:notifyCorrectAnswer(concept)
end

local function onIncorrectAnswer(self, conceptID, focusedConceptID)
  self.contentTracker:notifyIncorrectAnswer(focusedConceptID, conceptID)
  GameSession.notifyIncorrectAnswer(conceptID)
  GameSession.incrementIncorrectAnswerCount()
  self.incorrectAnswersCount = self.incorrectAnswersCount + 1
end

local function createFlasher(self)
  local flasher = SquaresFlasher.new()
  flasher:setLoc(CENTER_X, CENTER_Y)
  self:insert(flasher, Z_FLASH)

  self.flasher = flasher
end

local function createFocusTracker(self)
  local effectPlane = Group.new(WIDTH, HEIGHT)
  effectPlane:setLoc(CENTER_X, CENTER_Y)
  self:insert(effectPlane, Z_POWERUP_EFFECTS)

  local function handleBalloonsPopped(count)
    onBalloonsPopped(self, count)
  end

  local function handleCorrectAnswer(concept)
    onCorrectAnswer(self, concept)
  end

  local function handleIncorrectAnswer(conceptID, focusedConceptID)
    onIncorrectAnswer(self, conceptID, focusedConceptID)
  end

  local focusTracker = SquaresFocusTracker.new(self.kindA,
                                               self.kindB,
                                               self.balloonTracker,
                                               self.physics,
                                               self.balloonGenerator,
                                               self.snowMachine,
                                               self.flasher,
                                               self.popParticles,
                                               effectPlane,
                                               handleBalloonsPopped,
                                               handleCorrectAnswer,
                                               handleIncorrectAnswer)
  self.focusTracker = focusTracker
end

local function createScanner(self)
  local function onGameEndCallback()
    if (not self.hasGameEnded) then
      onGameLoss(self)
    end
  end

  local scanner = SquaresScanner.new(self.physics, self.background, onGameEndCallback)
  self.scanner = scanner
end

local function createPopParticles(self)
  local popParticles = SquaresPopParticles.new()
  popParticles:setLoc(0.0, 0.0)
  self:insert(popParticles, Z_POP_PARTICLES)
  self.popParticles = popParticles
end

local function createContentTracker(self)
  self.contentTracker = SquaresContentTracker.new(GameSession.contentTracker)
end

local function createMatchTracker(self)
  self.matchTracker = SquaresMatchTracker.new()
end

-- Constructor ------------------------

function SquaresGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SquaresGame)

  self.concepts = getConcepts()

  configureGameDifficulty(self)
  createKinds(self)

  self.physics = SquaresPhysics.new()

  createPopParticles(self)

  createBalloonTracker(self)
  createBalloonGenerator(self)

  createSnowMachine(self)
  createFlasher(self)
  createFocusTracker(self)

  createGameElements(self)
  createUI(self)

  createScanner(self)

  createContentTracker(self)
  createMatchTracker(self)

  self.hasGameEnded = false

  return self
end
