--- SummationFireworks
-- @author Created by Jay Austria on August 4, 2014.

class("SummationFireworks")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SHAPE_COUNT = 3
local DEFAULT_OUTER_RING_SUBDIVISIONS = 6
local DEFAULT_INNER_RING_SUBDIVISIONS = 3
local DEFAULT_OUTER_RING_SHAPE_SIZE = 8
local DEFAULT_INNER_RING_SHAPE_SIZE = 5
local VARIATION_MAX = 30
local BURST_DURATION = 2.0
local FADE_DELAY = 0.4
local FADE_DURATION = BURST_DURATION - FADE_DELAY
local OUTER_VELOCITY = 20
local INNER_VELOCITY = -10


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationFireworks:runAnimation()
  self:setVisibleRecursive(true)

  if self.cachedAnimation then
    self.cachedAnimation:start()
  end
end

function SummationFireworks:reset()
  for _, shape in pairs(self.shapeCollection) do
    shape:setLoc(shape.originX, originY)
    shape:setOpacity(shape.originOpacity)
    shape:setRot(shape.originRot)
  end

  for _, ring in pairs(self.ringCollection) do
    ring:setRot(ring.originRot)
  end

  self:setVisibleRecursive(false)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createShapeInternal(shape, radius, degrees, shapeRotation, opacity)
  local rad = math.rad(degrees)
  local randomVariation = VARIATION_MAX * (math.random() - 0.5)
  local x = radius * math.cos(rad) + radius
  local y = radius * math.sin(rad) + radius

  shape.destinationX = x
  shape.destinationY = y
  shape.originX = radius
  shape.originY = radius
  shape.destinationRot = shapeRotation
  shape.originRot = degrees + randomVariation
  shape.originOpacity =  opacity
  shape.destinationOpacity = 0.0

  shape:setLoc(shape.originX, shape.originY)
  shape:setRot(shape.originRot)

  shape:setOpacity(opacity)
end

local function createSquareAtDegrees(width, radius, degrees, shapeRotation, opacity, color)
  local square = FilledRectangle.new(width, width, color)
  createShapeInternal(square, radius, degrees, shapeRotation, opacity)

  return square
end

local function createTriangleAtDegrees(width, radius, degrees, shapeRotation, opacity, color)
  local triangle = FilledTriangle.new(-width * 0.5, 0, 0, width, width * 0.5, 0)
  triangle:setColor(unpack(color))
  createShapeInternal(triangle, radius, degrees, shapeRotation, opacity)

  return triangle
end

local function createLineAtDegrees(width, radius, degrees, shapeRotation, opacity, color)
  local line = Line.new(0, 0, width, 0)
  line:setLineWidth(2)
  line:setColor(unpack(color))
  createShapeInternal(line, radius, degrees, shapeRotation, opacity)

  return line
end

local function createFireworkRing(self, radius, shapeWidth, opacity, rotationVelocity, ringSubdivisions, color)
  local width = radius * 2
  local height = width
  local ringGroup = Group.new(width, height)

  ringGroup.originRot = 0
  ringGroup.destinationRot = rotationVelocity

  local degrees = 0
  local degreeIncrement = 360 / (SHAPE_COUNT * ringSubdivisions)
  local shapeRotation = -rotationVelocity / 4
  for i = 1, ringSubdivisions do
    local square = createSquareAtDegrees(shapeWidth, radius, degrees, shapeRotation, opacity, color)
    ringGroup:insert(square)
    table.insert(self.shapeCollection, square)
    degrees = degrees + degreeIncrement

    local triangle = createTriangleAtDegrees(shapeWidth, radius, degrees, shapeRotation, opacity, color)
    ringGroup:insert(triangle)
    table.insert(self.shapeCollection, triangle)
    degrees = degrees + degreeIncrement

    local line = createLineAtDegrees(shapeWidth, radius, degrees, shapeRotation, opacity, color)
    ringGroup:insert(line)
    table.insert(self.shapeCollection, line)
    degrees = degrees + degreeIncrement
  end

  table.insert(self.ringCollection, ringGroup)

  return ringGroup
end

local function initializeAnimation(self)
  local burstAction = AHSpawn.new()
  for _, shape in pairs(self.shapeCollection) do
    local moveAction = AHLocationTo.new(shape, BURST_DURATION, AHEaseExponentialOut.new(), shape.destinationX, shape.destinationY)
    local fadeDelay = AHDelay.new(FADE_DELAY)
    local fadeAction = AHOpacityTo.new(shape, FADE_DURATION, nil, shape.destinationOpacity)
    local combinedFadeAction = AHSequence.new(fadeDelay, fadeAction)

    burstAction:addAction(moveAction)
    burstAction:addAction(combinedFadeAction)
  end

  for _, ring in pairs(self.ringCollection) do
    local rotationAction = AHRotationTo.new(ring, BURST_DURATION, nil, ring.destinationRot)
    burstAction:addAction(rotationAction)
  end

  self.cachedAnimation = burstAction
end

local function createLayout(self, color, outerSubdivisions, innerSubdivisions, outerSize, innerSize)
  local width, height = self:getDims()

  local outerRing = createFireworkRing(self, width/2, outerSize, 1.0, OUTER_VELOCITY, outerSubdivisions, color)
  self:insertAtCenter(outerRing)

  local innerRing = createFireworkRing(self, width/4, innerSize, 0.6, INNER_VELOCITY, innerSubdivisions, color)
  self:insertAtCenter(innerRing)
end

function SummationFireworks.new(radius, color, outerSubdivisions, innerSubdivisions, outerSize, innerSize)
  local width = radius * 2
  local height = width
  local self = Group.new(width, height)
  Mixin.mix(self, SummationFireworks)

  color = color or SummationCommon.WHITE_COLOR
  outerSubdivisions = outerSubdivisions or DEFAULT_OUTER_RING_SUBDIVISIONS
  innerSubdivisions = innerSubdivisions or DEFAULT_INNER_RING_SUBDIVISIONS
  outerSize = outerSize or DEFAULT_OUTER_RING_SHAPE_SIZE
  innerSize = innerSize or DEFAULT_INNER_RING_SHAPE_SIZE

  self.radius = radius
  self.shapeCollection = {}
  self.ringCollection = {}
  createLayout(self, color, outerSubdivisions, innerSubdivisions, outerSize, innerSize)
  initializeAnimation(self)

  self:setVisibleRecursive(false)

  return self
end
