--- SummationGame
-- @author Created by Jay Austria on July 17, 2014.

class("SummationGame")
require("StandardUIBuilder")
require("SummationColorUtils")
require("SummationConfiguration")
require("SummationRound")
require("SummationScene")
require("SummationWinAnimation")
require("SummationScreenShake")
require("ExperimentsConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OVERLAY_X_OFFSET = -2
local OVERLAY_Y_OFFSET = Screen.getSafeHeightPart(0.3)
local GAME_OUTRO_DURATION = 2.0

local function configureGameDifficulty(self)
  self.gameDuration = SummationConfiguration.GAME_DURATION
  GameSession.setTotalTime(self.gameDuration)

  local passScore = SummationConfiguration.getPassScore()
  local rankTwoScore = SummationConfiguration.getRankTwoScore()
  local rankThreeScore = SummationConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.ACCURACY_BONUS)

  self.partiallyCorrectDelta = SummationConfiguration.getDeltaForPartiallyCorrectAnswer()
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateScore(self)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function isResponseFullyCorrect(responseDelta)
  return responseDelta <= SummationConfiguration.FULLY_CORRECT_DELTA_THRESHOLD
end

local function isResponseCorrect(self, responseDelta)
  return responseDelta <= self.partiallyCorrectDelta
end

local function didPlayerWin()
  return (GameSession.getCorrectAnswerCount() >= SummationConfiguration.CORRECT_ANSWERS_TO_PASS)
end

local function isGameEnded(self)
  return self.hasTimeExpired
end

local function getClampedY(self, y)
  local bottomY = SummationCommon.NUMBER_LINE_LOCATION_Y - SummationCommon.NUMBER_LINE_HEIGHT * 0.5
  local topY = SummationCommon.NUMBER_LINE_LOCATION_Y + SummationCommon.NUMBER_LINE_HEIGHT * 0.5

  if self.isInTutorial then
    topY = self.scene:getLocationYFromValue(self.correctSum)
  end
  return MathUtils.clamp(y, bottomY, topY)
end

local function waitForResponseOrTutorial(self)
  local responseLock = true
  self.responseLock = responseLock

  self.isInputEnabled = true

  waitUntil(function()
    local unlocked = not self.responseLock
    local shouldShowTutorial = (not self.hasTouched and self.gameTimer:getTime() > SummationConfiguration.TUTORIAL_DELAY)
    return unlocked or shouldShowTutorial
  end)
end

local function waitForResponse(self)
  local responseLock = true
  self.responseLock = responseLock
  self.isInputEnabled = true

  waitUntil(function() return self.responseLock == false end)
end

local function unlockWait(self)
  self.responseLock = false
end

local function unlockAfterSceneAnimation(self)
  local isSceneAnimatingWrapper = function()
    return self.scene:isAnimating()
  end

  waitWhile(isSceneAnimatingWrapper)
  unlockWait(self)
end

local function getTimeoutHandler(self)
  return function()
    self.hasTimeExpired = true
    unlockAfterSceneAnimation(self)
  end
end

local function startRoundWithRoundDescriptor(self, currentRoundDescriptor, isFirstRound)
  local priceCollection = currentRoundDescriptor.prices
  local sum = currentRoundDescriptor.sum

  self.scene:startRound(priceCollection, sum, isFirstRound)

  self.priceCollection = priceCollection
  self.correctSum = sum
end

local function rewardForCorrect(self, isPerfect)
  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(SummationConfiguration.getCorrectAnswerPoints(GameSession.getCorrectAnswerCount()))

  if isPerfect then
    ScoringUtils.applyAccuracyBonus(SummationConfiguration.POINTS_PER_ACCURACY_BONUS, self.bonusTray)
    self.scene:showPerfectVisual()
  end

  updateScore(self)
end

local function handleCorrectResponse(self, response, isPerfect)
  local onPuckArriveCallback = function()
    rewardForCorrect(self, isPerfect)
  end

  self.sceneAnimation = self.scene:animateCorrectResponse(response, self.correctSum, onPuckArriveCallback, isPerfect)
  unlockAfterSceneAnimation(self)
end

local function handleIncorrectResponse(self, response)
  GameSession.incrementIncorrectAnswerCount()

  self.sceneAnimation = self.scene:animateIncorrectResponse(response, self.correctSum)
  unlockAfterSceneAnimation(self)
end

local function cleanRound(self)
  self.sceneAnimation = nil

  if isGameEnded(self) and not didPlayerWin() then
    self.scene:prepForLoseAnimation()
  else
    self.scene:clear()
  end
end

local function submitUserResponse(self, response)
  self.isInputEnabled = false

  local distanceFromActualSum = math.abs(response - self.correctSum)

  local isCorrect = isResponseCorrect(self, distanceFromActualSum)
  local isPerfect = isResponseFullyCorrect(distanceFromActualSum)

  self.restartTutorial = false

  if self.isInTutorial then
    if isPerfect then
      self.gameTimer:resumeAndHideIcon()
      handleCorrectResponse(self, response, isPerfect)
    else
      unlockAfterSceneAnimation(self)
      self.scene:clearNumberLine()
      self.scene:setInteractionHelperVisible(true)
      self.restartTutorial = true
    end
  else
    if isCorrect then
      handleCorrectResponse(self, response, isPerfect)
    else
      handleIncorrectResponse(self, response)
    end
  end
end

local function waitForGameWinOutro(self)
  self.gameTimer:stop()

  local uiWidgets = self.uiOverlay:getWidgets()
  for _, widget in pairs(uiWidgets) do
    SummationColorUtils.morphElementColors(widget, SummationCommon.UI_COMPONENTS_COLOR, SummationCommon.COLOR_MORPH_RESPONDING_DURATION)
  end

  local gameWinOutro = SummationWinAnimation.new()
  self:insertAtCenter(gameWinOutro)
  gameWinOutro:runAndWaitForOutro()
end

local function onTouchBegan(self, x, y)
  local clampedY = getClampedY(self, y)

  self.scene:onTouchBegan(x, clampedY)
  self.value = self.scene:getValueFromLocationY(clampedY)

  self.didTouchStartWithBegan = true
  self.hasTouched = true
end

local function onTouchMoved(self, x, y)
  if self.didTouchStartWithBegan then
    local clampedY = getClampedY(self, y)
    self.scene:onTouchMoved(x, clampedY)
    self.value = self.scene:getValueFromLocationY(clampedY)
  end
end

local function onTouchEnded(self, x, y)
  if self.didTouchStartWithBegan then
    self.scene:onTouchEnded(x, y)
    submitUserResponse(self, self.value)
  end

  self.didTouchStartWithBegan = false
end

local function showTutorial(self)
  runInCoroutine(function()
    self.isInputEnabled = true

    local uiOverlay = self:getUIOverlay()
    local oldUIZIndex = uiOverlay.zIndex
    self.scene:takeUI(uiOverlay)

    self.scene:waitForTutorial()
    self:takeChild(uiOverlay, oldUIZIndex)
  end)
end

local function createInputHandler(self)
  local handlerWrapper = function(touchEvent)
    if self.isInputEnabled then
      local screenX = touchEvent.x
      local screenY = touchEvent.y

      self.scene:setInteractionHelperVisible(false)

      if touchEvent.phase == TouchManager.TOUCH_BEGAN then
        onTouchBegan(self, screenX, screenY)
      elseif touchEvent.phase == TouchManager.TOUCH_MOVED then
        onTouchMoved(self, screenX, screenY)
      elseif touchEvent.phase == TouchManager.TOUCH_ENDED then
        onTouchEnded(self, screenX, screenY)
      end
    end
  end
  TouchManager.addTouchHandler(self, handlerWrapper)
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationGame:gameFlow()
  self.gameTimer:start()
  self.scene:doIntro()

  local isFirstRound = true
  repeat
    local gameProgress = GameSession.getCorrectAnswerCount()
    local currentRoundDescriptor = SummationRound.newRandomRound(GameSession.getDifficulty(), gameProgress, isFirstRound)

    startRoundWithRoundDescriptor(self, currentRoundDescriptor, isFirstRound)
    waitForResponseOrTutorial(self)

    local shouldShowTutorial = (not self.hasTouched and isFirstRound)
    if shouldShowTutorial then
      self.gameTimer:pauseAndShowIcon()

      repeat
        self.isInTutorial = true
        self.scene:setTutorialMode(true)
        showTutorial(self)
        waitForResponse(self)
      until not self.restartTutorial
    end

    self.isInTutorial = false
    self.scene:setTutorialMode(false)

    cleanRound(self)
    isFirstRound = false
  until isGameEnded(self)
  self.isInputEnabled = false
  GameSession.notifyGameTimeEnd()

  if didPlayerWin() then
    waitForGameWinOutro(self)

    GameSession.notifyGameWin()
  else
    self.scene:animateGameLoss()
    waitFor(GAME_OUTRO_DURATION)
    GameSession.notifyGameLoss()
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local widgets = self.uiOverlay:getWidgets()

  local scene = SummationScene.new(widgets, self.partiallyCorrectDelta)
  self:insertAtCenter(scene)
  self.scene = scene
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, SummationCommon.UI_COMPONENTS_COLOR)

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()

  self.gameTimer = builder:addDigitTimer(self.gameDuration)
  self.gameTimer:setTimeoutHandler( getTimeoutHandler(self) )

  self.bonusTray = builder:addBonusTray()
  self.bonusTray:setColor(unpack(SummationCommon.WHITE_COLOR)) -- because we only see it when the bg is inverted
end

-- Constructor ------------------------

function SummationGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SummationGame)

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)
  createInputHandler(self)

  self.hasInteracted = false
  self.hasTimeExpired = false
  self.hasTouched = false
  self.isInTutorial = false

  return self
end
