--- SummationInteractionHelper
-- @author Created by Jay Austria on August 13, 2014.

class("SummationInteractionHelper")
require("graphics.primitives.Arrow")
require("SummationCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FAKE_CURSOR_COLOR = ColorUtils.colorRGBToStd(175, 153, 189)
local FAKE_BAR_COLOR = ColorUtils.colorRGBToStd(175, 153, 189)
local HELP_MESSAGE = Localizer.get("HELPER")
local HELP_MESSAGE_FONT_SIZE = ReferenceDevice.getScaledWidth(16)
local HELP_LABEL_COLOR = ColorUtils.colorHex24ToStd(0x90679E)
local ARROW_COLOR = ColorUtils.colorRGBToStd(175, 153, 189)
local ARROW_LENGTH = 15
local ARROW_HEAD_LENGTH = 5
local ARROW_HEAD_ANGLE = math.pi / 4
local ARROW_THICKNESS = 2
local UP_ARROW_ROTATION = 90
local UP_ARROW_OFFSET_Y = 15
local DOWN_ARROW_ROTATION = -90
local FADE_IN_DURATION = 0.333
local FADE_OUT_DELAY = 0.333
local FADE_OUT_DURATION = 0.333
local FADE_IN_DELAY = 0.65
local INITIAL_FADE_DURATION = 0.6

local LABEL_OPACITY_OUT = 0.8
local LABEL_OPACITY_IN = 1
local BACKING_OPACITY_OUT = 0.1
local BACKING_OPACITY_IN = 0.4
local CURSOR_OPACITY_OUT = 0.3
local CURSOR_OPACITY_IN = 0.6


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createFadeAction(self, duration, destinationOpacityLabel, destinationOpacityBacking, destinationOpacityCursor)
  local labelFadeAction = AHOpacityTo.new(self.helpLabel, duration, nil, destinationOpacityLabel)
  local backingFadeAction = AHOpacityTo.new(self.fakeBacking, duration, nil, destinationOpacityBacking)
  local cursorFadeAction = AHOpacityTo.new(self.fakeCursor, duration, nil, destinationOpacityCursor)

  local combinedFade = AHSpawn.new(labelFadeAction, backingFadeAction, cursorFadeAction)
  return combinedFade
end

local function createLoopedFadeSequence(self)
  local fadeIn = createFadeAction(self, FADE_IN_DURATION, LABEL_OPACITY_IN, BACKING_OPACITY_IN, CURSOR_OPACITY_IN)
  local fadeOutDelay = AHDelay.new(FADE_OUT_DELAY)
  local fadeOut = createFadeAction(self, FADE_OUT_DURATION, LABEL_OPACITY_OUT, BACKING_OPACITY_OUT, CURSOR_OPACITY_OUT)
  local fadeInDelay = AHDelay.new(FADE_IN_DELAY)

  local fadeSequence = AHSequence.new(fadeOutDelay, fadeOut, fadeInDelay, fadeIn)
  return AHLoop.new(fadeSequence)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationInteractionHelper:stopAnimating()
  if self.loopedSequence then
    self.loopedSequence:stop()
  end
end

function SummationInteractionHelper:createDisplayAnimation(isFirstRound)
  if isFirstRound then
    local initialFade = AHSpawn.new(
      AHOpacityTo.new(self, INITIAL_FADE_DURATION, nil, SummationCommon.VISIBLE_OPACITY),
      createFadeAction(self, INITIAL_FADE_DURATION, LABEL_OPACITY_IN, BACKING_OPACITY_IN, CURSOR_OPACITY_IN)
    )
    local loopedFadeSequence = createLoopedFadeSequence(self)
    return function()
      self:stopAnimating()
      self.loopedSequence = AHSequence.new(initialFade, loopedFadeSequence):start()
    end
  else
    return createFadeAction(self, INITIAL_FADE_DURATION, 0, BACKING_OPACITY_OUT, CURSOR_OPACITY_OUT)
  end
end

function SummationInteractionHelper:show()
  self:stopAnimating()
  self:setOpacity(SummationCommon.VISIBLE_OPACITY)
  self.helpLabel:setOpacity(LABEL_OPACITY_OUT)
  self.fakeBacking:setOpacity(BACKING_OPACITY_OUT)
  self.fakeCursor:setOpacity(CURSOR_OPACITY_OUT)
end

function SummationInteractionHelper:hide()
  self:stopAnimating()
  self:setOpacity(0)
  self.helpLabel:setOpacity(0)
  self.fakeBacking:setOpacity(0)
  self.fakeCursor:setOpacity(0)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local width, height = self:getDims()

  local fakeBacking = FilledRectangle.new(width, height, FAKE_BAR_COLOR)
  fakeBacking:setOpacity(SummationCommon.INVISIBLE_OPACITY)
  self:insertAtCenter(fakeBacking)
  self.fakeBacking = fakeBacking

  local fakeCursor = FilledRectangle.new(SummationCommon.CURSOR_WIDTH, height, FAKE_CURSOR_COLOR)
  fakeCursor:setLoc(self:getWidthPart(0.95) + SummationCommon.CURSOR_WIDTH * -0.5, height * 0.5)
  self:insert(fakeCursor)
  fakeCursor:setOpacity(SummationCommon.INVISIBLE_OPACITY)
  self.fakeCursor = fakeCursor

  local helpLabel = Label.new(HELP_MESSAGE, Font.new(FontFaces.DEFAULT_FONT, HELP_MESSAGE_FONT_SIZE))
  helpLabel:setColor(unpack(HELP_LABEL_COLOR))
  local helpLabelX = (fakeCursor:getX() - fakeCursor:getWidthPart(0.5)) * 0.5
  helpLabel:setLoc(helpLabelX, height * 0.5)
  self:insert(helpLabel)
  helpLabel:setOpacity(SummationCommon.INVISIBLE_OPACITY)
  self.helpLabel = helpLabel

  local upArrow = StickArrow.new(ARROW_LENGTH, ARROW_HEAD_LENGTH, ARROW_HEAD_ANGLE, ARROW_THICKNESS)
  upArrow:setRot(UP_ARROW_ROTATION)
  upArrow:setLoc(SummationCommon.CURSOR_WIDTH * 0.5, height + UP_ARROW_OFFSET_Y)
  upArrow:setColor(unpack(ARROW_COLOR))
  fakeCursor:insert(upArrow)
  upArrow:inheritOpacity(self)

  local downArrow = StickArrow.new(ARROW_LENGTH, ARROW_HEAD_LENGTH, ARROW_HEAD_ANGLE, ARROW_THICKNESS)
  downArrow:setRot(DOWN_ARROW_ROTATION)
  downArrow:setLoc(SummationCommon.CURSOR_WIDTH * 0.5, -UP_ARROW_OFFSET_Y)
  downArrow:setColor(unpack(ARROW_COLOR))
  fakeCursor:insert(downArrow)
  downArrow:inheritOpacity(self)
end

function SummationInteractionHelper.newHidden(width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, SummationInteractionHelper)

  createLayout(self)

  self:setAnchorPointRight()
  self:setOpacity(SummationCommon.INVISIBLE_OPACITY)

  return self
end
