--- SummationNumberLine
-- @author Created by Jay Austria on July 30, 2014.

class("SummationNumberLine")
require("SummationColorUtils")
require("SummationCommon")
require("SummationCursor")
require("SummationHighlightBar")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MARK_SIZE_DEFAULT = 2
local MARK_SIZE_FIVES = 5
local MARK_SIZE_TENS = 10
local MARK_COLOR = ColorUtils.colorHex24ToStd(0xA68AB2)
local MARK_COLOR_DIMMED = ColorUtils.colorRGBToStd(141, 122, 153)
local VALUE_LABEL_FONT_SIZE = 20
local CURRENCY_LABEL_FONT_SIZE = 12
local VALUE_LABEL_OFFSET_X = ReferenceDevice.getScaledWidth(-25)
local CURRENCY_LABEL_OFFSET_X = VALUE_LABEL_OFFSET_X + (ReferenceDevice.getScaledWidth(15) * SummationCommon.PRICE_LABEL_OFFSET_MULTIPLIER)
local HIGHLIGHT_MARKS_THRESHOLD = 1
local HIGHLIGHT_T_MINIMUM = 0.5
local HIGHLIGHT_T_MAXIMUM = 1.0
local COLOR_HIGHLIGHT_DURATION = 0.1


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isValueDivisibleByFive(value)
  return (value % 5 == 0)
end

local function isValueDivisibleByTen(value)
  return (value % 10 == 0)
end

local function getMarkSize(value)
  local markSize = MARK_SIZE_DEFAULT
  if isValueDivisibleByTen(value) then
    markSize = MARK_SIZE_TENS
  elseif isValueDivisibleByFive(value) then
    markSize = MARK_SIZE_FIVES
  end

  return markSize
end

local function createValueLabel(value)
  local valueString = tostring(value)
  local valueFont = Font.new(FontFaces.DIN_OT_MEDIUM, VALUE_LABEL_FONT_SIZE)
  local valueLabel = Label.new(valueString, valueFont)
  valueLabel:setColor(unpack(MARK_COLOR))
  return valueLabel
end

local function createCurrencyLabel()
  local currencyString = Localizer.get("CURRENCY")
  local currencyFont = Font.new(FontFaces.DIN_OT_MEDIUM, CURRENCY_LABEL_FONT_SIZE)
  local currencyLabel = Label.new(currencyString, currencyFont)
  currencyLabel:setColor(unpack(MARK_COLOR))
  return currencyLabel
end

local function createMark(self, value)
  local markSize = getMarkSize(value)
  local newMark = FilledRectangle.new(markSize, markSize, MARK_COLOR)

  if isValueDivisibleByTen(value) then
    local valueLabel = createValueLabel(value)
    local valueLocationX = VALUE_LABEL_OFFSET_X
    local valueLocationY = markSize * 0.5
    valueLabel:setLoc(valueLocationX, valueLocationY)
    newMark:insert(valueLabel)

    local currencyLabel = createCurrencyLabel()
    local currencyLocationX = CURRENCY_LABEL_OFFSET_X
    local currencyLocationY = markSize * 0.8
    currencyLabel:setLoc(currencyLocationX, currencyLocationY)
    newMark:insert(currencyLabel)
  end

  return newMark
end

local function initPerfectStateAction(self)
  self.cachedPerfectStateAction = SummationColorUtils.getFlipElementColorToAction(self, SummationCommon.WHITE_COLOR, MARK_COLOR, 0.0, 1.4, 0.4)
end

local function restoreAllMarks(self)
  self.runningAction:insert(SummationColorUtils.getColorToAction(self, MARK_COLOR, COLOR_HIGHLIGHT_DURATION))
end

local function dimAllMarks(self)
  self.runningAction:insert(SummationColorUtils.getColorToAction(self, MARK_COLOR_DIMMED, COLOR_HIGHLIGHT_DURATION))
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationNumberLine:getMarkYOffset()
  return self.markYOffset
end

function SummationNumberLine:resetMarkColors()
  restoreAllMarks(self)
end

function SummationNumberLine:transitionColor(transitionAction)
  self.runningAction:insert(transitionAction)
end

function SummationNumberLine:morphToPerfectState()
  if self.cachedPerfectStateAction then
    self.cachedPerfectStateAction:start()
  end
end

function SummationNumberLine:doLossAnimation()
  local colorAction = SummationColorUtils.getColorToAction(self, SummationCommon.INCORRECT_COLOR, SummationCommon.LOSE_COLOR_DURATION)
  local fadeDelay = AHDelay.new(SummationCommon.LOSE_FADE_DELAY_DURATION)
  local fadeAction = AHOpacityTo.new(self, SummationCommon.LOSE_FADE_DURATION, nil, SummationCommon.INVISIBLE_OPACITY)
  local combinedAction = AHSequence.new(colorAction, fadeDelay, fadeAction)
  combinedAction:start()
end

function SummationNumberLine:enterRespondingState()
  dimAllMarks(self)
end

function SummationNumberLine:leaveRespondingState()
  restoreAllMarks(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local firstMarkValue = self.minimum
  local lastMarkValue = self.maximum

  local numberOfMarks = lastMarkValue - firstMarkValue

  local width, height = self:getDims()
  local markXLocation = width * 0.5
  local markYOffset = height / numberOfMarks

  for markIndex = 1, numberOfMarks do
    local value =  firstMarkValue + markIndex
    local mark = createMark(self, value)
    mark:setLoc(width * 0.5, markIndex * markYOffset)
    self:insert(mark)
    table.insert(self.markCollection, mark)
  end

  self.runningAction = ActionInterrupt.new()

  local isRecursive = true
  self:setColor(unpack(MARK_COLOR))
  self:linkOpacityAttrToChildren(isRecursive)

  self.markYOffset = markYOffset
end

function SummationNumberLine.new(width, height, minimum, maximum)
  local self = Group.new(width, height)
  Mixin.mix(self, SummationNumberLine)

  self.minimum = minimum
  self.maximum = maximum
  self.markCollection = {}

  createLayout(self)
  initPerfectStateAction(self)

  return self
end
