--- SummationPriceDisplay
-- @author Created by Jay Austria on July 31, 2014.

class("SummationPriceDisplay")
require("SummationColorUtils")
require("SummationCommon")
require("SummationPrice")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LABEL_CELL_HEIGHT = ReferenceDevice.getScaledHeight(80)
local LABEL_CELL_WIDTH = ReferenceDevice.getScaledWidth(160)
local LABEL_STAGGER_X = ReferenceDevice.getScaledWidth(30)
local FADE_IN_DURATION = 0.6
local FADE_OUT_DURATION = 0.1


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getFirstPriceLocationY(self, priceCount)
  local groupHeight = self:getHeight()
  local firstPriceLocationY = (groupHeight * 0.5) - (priceCount * 0.5) * LABEL_CELL_HEIGHT + LABEL_CELL_HEIGHT * 0.5

  return firstPriceLocationY
end

local function createFadeLabelsToOpacityAction(self, targetOpacity, duration)
  local fadeActionCollection = AHSpawn.new()
  local labels = self.priceLabelCollection
  for _, label in pairs(labels) do
    local fadeAction = AHOpacityTo.new(label, duration, nil, targetOpacity)
    fadeActionCollection:addAction(fadeAction)
  end

  return fadeActionCollection
end

local function createLabelsToColorAction(self, targetColor, duration)
  local colorActionCollection = AHSpawn.new()
  local labels = self.priceLabelCollection
  for _, label in pairs(labels) do
    local colorAction = AHColorTo.new(label, duration, nil, targetColor)
    colorActionCollection:addAction(colorAction)
  end

  return colorActionCollection
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationPriceDisplay:createDisplayPricesAnimation(priceCollection)
  local priceCount = #priceCollection
  local width = self:getWidth()

  local doFlipStagger = MathUtils.getRandomBoolean()
  local staggerParity = 1.0

  if doFlipStagger then
    staggerParity = -1.0
  end

  local firstPriceLocationX = width * 0.5 - staggerParity * ((width - LABEL_CELL_WIDTH) * 0.5)
  local firstPriceLocationY = getFirstPriceLocationY(self, priceCount)

  for priceIndex = 1, priceCount do
    local priceValue = priceCollection[priceIndex]
    local priceLabel = SummationPrice.new(priceValue)

    local labelLocationXOffset = staggerParity * ((priceIndex % 2) * LABEL_STAGGER_X)

    local labelLocationX = firstPriceLocationX + labelLocationXOffset
    local labelLocationY = firstPriceLocationY + LABEL_CELL_HEIGHT * (priceIndex - 1)

    priceLabel:setLoc(labelLocationX, labelLocationY)
    self:insert(priceLabel)
    table.insert(self.priceLabelCollection, priceLabel)
  end

  return createFadeLabelsToOpacityAction(self, SummationCommon.VISIBLE_OPACITY, FADE_IN_DURATION)
end

function SummationPriceDisplay:clear()
  local fadeAndRemoveLabels = function()
    local fadeOutAction = createFadeLabelsToOpacityAction(self, SummationCommon.INVISIBLE_OPACITY, FADE_OUT_DURATION)
    wait(fadeOutAction:start())

    for _, priceLabel in pairs(self.priceLabelCollection) do
      priceLabel:removeSelf()
    end

    self.priceLabelCollection = {}
  end
  runInCoroutine(fadeAndRemoveLabels)
end

function SummationPriceDisplay:morphToRespondingState()
  for _, priceLabel in pairs(self.priceLabelCollection) do
    priceLabel:morphToRespondingState()
  end
end

function SummationPriceDisplay:doLossAnimation()
  for _, priceLabel in pairs(self.priceLabelCollection) do
    local colorAction = SummationColorUtils.getColorToAction(priceLabel, SummationCommon.INCORRECT_COLOR, SummationCommon.LOSE_COLOR_DURATION)
    local fadeDelay = AHDelay.new(SummationCommon.LOSE_FADE_DELAY_DURATION)
    local fadeAction = AHOpacityTo.new(priceLabel, SummationCommon.LOSE_FADE_DURATION, nil, SummationCommon.INVISIBLE_OPACITY)
    local combinedAction = AHSequence.new(colorAction, fadeDelay, fadeAction)
    combinedAction:start()
  end
end

function SummationPriceDisplay:animateToColor(color, duration)
  createLabelsToColorAction(self, color, duration):start()
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function SummationPriceDisplay.new(width, height)
  local self = Group.new(width, height)
  Mixin.mix(self, SummationPriceDisplay)

  self.priceLabelCollection = {}

  return self
end
