--- SummationRound
-- @author Created by Jay Austria on July 17, 2014.

class("SummationRound")
require("data.generators.UtilityGenerators")
require("data.generators.ValueGenerators")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUND_COUNT_TO_MAX_DIFFICULTY = 12
local MIN_DOLLARS = 1
local MAX_DOLLARS = 17
local VALID_MINIMUM = 2
local VALID_MAXIMUM = 38
local VALID_MAXIMUM_FIRST_ROUND = 15
local DOLLARS_VALUE_WEIGHTS_EASY = {3, 3, 2, 2, 2, 2, 2, 3, 3, 1, 1, 1, 0, 0, 0, 0, 0}
local DOLLARS_VALUE_WEIGHTS_MEDIUM = {4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 1, 1, 1, 0, 0, 0}
local DOLLARS_VALUE_WEIGHTS_HARD = {1, 4, 4, 4, 1, 4, 4, 4, 4, 1, 2, 2, 1, 1, 1, 1, 1}
local POSSIBLE_VALUES_CENTS = {1, 2, 3, 4, 5, 6, 7, 8, 9, 0}
local ONES_VALUE_WEIGHTS_EASY = {1, 1, 1, 0, 0, 0, 1, 1, 10, 5}
local ONES_VALUE_WEIGHTS_MEDIUM = {1, 1, 1, 1, 5, 1, 1, 1, 15, 5}
local ONES_VALUE_WEIGHTS_HARD = {5, 5, 5, 5, 1, 5, 5, 5, 1, 1}
local TENS_VALUE_WEIGHTS_EASY = {1, 0, 0, 0, 0, 0, 0, 0, 1, 1}
local TENS_VALUE_WEIGHTS_MEDIUM = {0, 1, 1, 0, 0, 0, 1, 1, 1, 0}
local TENS_VALUE_WEIGHTS_HARD = {0, 1, 1, 2, 2, 2, 1, 1, 0, 0}
local MIN_PRICES_COUNT_BASIC = 2
local MAX_PRICES_COUNT_BASIC = 3
local MIN_PRICES_COUNT_EASY = 2
local MAX_PRICES_COUNT_EASY = 4
local MIN_PRICES_COUNT_MEDIUM = 3
local MAX_PRICES_COUNT_MEDIUM = 5
local MIN_PRICES_COUNT_HARD = 3
local MAX_PRICES_COUNT_HARD = 6
local MAX_PRICES_COUNT_FIRST_ROUND = 3
local MAX_REDUCE = 4
local DIFFICULTY_NAMES = {"EASY", "MEDIUM", "HARD"}
local POSSIBLE_VALUES_DOLLARS = TableUtils.getNumberSequence(MIN_DOLLARS, MAX_DOLLARS)

local DIFFICULTY_THRESHOLD_BASIC = 0.1125
local DIFFICULTY_THRESHOLD_EASY = 0.25
local DIFFICULTY_TRESHOLD_MEDIUM = 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Set size -------------------------------------------------------------------
local function isDifficultyBasic(difficulty)
  return (difficulty < DIFFICULTY_THRESHOLD_BASIC)
end

local function isDifficultyEasy(difficulty)
  return (difficulty >= DIFFICULTY_THRESHOLD_BASIC) and (difficulty < DIFFICULTY_THRESHOLD_EASY)
end

local function isDifficultyMedium(difficulty)
  return (difficulty >= DIFFICULTY_THRESHOLD_EASY) and (difficulty < DIFFICULTY_TRESHOLD_MEDIUM)
end

local function isDifficultyHard(difficulty)
  return (difficulty >= DIFFICULTY_TRESHOLD_MEDIUM)
end

local function getNumberOfPrices(difficulty, gameProgress, isFirstRound)
  local minCount = nil
  local maxCount = nil

  if isDifficultyBasic(difficulty) then
    minCount = MIN_PRICES_COUNT_BASIC
    maxCount = MAX_PRICES_COUNT_BASIC
  elseif isDifficultyEasy(difficulty) then
    minCount = MIN_PRICES_COUNT_EASY
    maxCount = MAX_PRICES_COUNT_EASY
  elseif isDifficultyMedium(difficulty) then
    minCount = MIN_PRICES_COUNT_MEDIUM
    maxCount = MAX_PRICES_COUNT_MEDIUM
  elseif isDifficultyHard(difficulty) then
    minCount = MIN_PRICES_COUNT_HARD
    maxCount = MAX_PRICES_COUNT_HARD
  end

  if isFirstRound then
    maxCount = MAX_PRICES_COUNT_FIRST_ROUND
    minCount = math.min(minCount, maxCount)
  end

  local pricesCount = MathUtils.interpolateLinearFromNormalized(gameProgress, minCount, maxCount)
  local roundedCount = MathUtils.round(pricesCount)
  return roundedCount
end

-- Generators --------------------------------------------------------------------

local function getGeneratorForWeights(weightsDollars, weightsTens, weightsOnes)
  local doReplace = true
  local dollarGenerator = WeightedRandomlySelectedValueGenerator.new(POSSIBLE_VALUES_DOLLARS, weightsDollars, doReplace)
  local tensGenerator = WeightedRandomlySelectedValueGenerator.new(POSSIBLE_VALUES_CENTS, weightsTens, doReplace)
  local onesGenerator = WeightedRandomlySelectedValueGenerator.new(POSSIBLE_VALUES_CENTS, weightsOnes, doReplace)

  local priceReturnWrapper = function(dollars, tens, ones)
    local price = dollars + (tens / 10) + (ones / 100)
    return price
  end

  local randomGenerator = DemultiplexingGenerator.new({dollarGenerator, tensGenerator, onesGenerator}, priceReturnWrapper)
  return randomGenerator
end

local function getEasyGenerator()
  return getGeneratorForWeights(DOLLARS_VALUE_WEIGHTS_EASY, TENS_VALUE_WEIGHTS_EASY, ONES_VALUE_WEIGHTS_EASY)
end

local function getMediumGenerator()
  return getGeneratorForWeights(DOLLARS_VALUE_WEIGHTS_MEDIUM, TENS_VALUE_WEIGHTS_MEDIUM, ONES_VALUE_WEIGHTS_MEDIUM)
end

local function getHardGenerator()
  return getGeneratorForWeights(DOLLARS_VALUE_WEIGHTS_HARD, TENS_VALUE_WEIGHTS_HARD, ONES_VALUE_WEIGHTS_HARD)
end

-- Round generation -------------------------------------------------------------

local priceGenerators =
{
  ["EASY"] = getEasyGenerator(),
  ["MEDIUM"] = getMediumGenerator(),
  ["HARD"] = getHardGenerator()
}

local function generateRandomPrice(difficultyName)
  local price = priceGenerators[difficultyName]:next()
  return price
end

local function isSumValid(priceSum, isFirstRound)
  local max = (isFirstRound and VALID_MAXIMUM_FIRST_ROUND) or VALID_MAXIMUM
  return ((priceSum >= VALID_MINIMUM) and (priceSum <= max))
end

local function generateRandomRound(normalizedDifficulty, normalizedGameProgress, isFirstRound)
  local self = {}
  local priceCollection = {}

  local numberOfPrices = getNumberOfPrices(normalizedDifficulty, normalizedGameProgress, isFirstRound)

  local difficultyKeyFrames = DifficultyCurves.getKeyFrames(DifficultyCurves.LESS_MIXING, DIFFICULTY_NAMES)
  local difficultyChooser = KeyFrameInterpolator.new(difficultyKeyFrames, normalizedDifficulty)

  for priceIndex = 1, numberOfPrices do
    local difficultyName = difficultyChooser:nextValue()

    local newPrice = generateRandomPrice(difficultyName)
    table.insert(priceCollection, newPrice)
  end

  -- If the prices are not within valid range, randomly reduce prices
  -- evenly by dollar amounts (if possible) until the sum is valid
  local priceSum = TableUtils.sum(priceCollection)
  while not isSumValid(priceSum, isFirstRound) do
    local newCollection = {}
    local randomReduceAmount = math.random(MAX_REDUCE)
    for _, price in pairs(priceCollection) do
      if price > randomReduceAmount + 1 then
        price = price - randomReduceAmount
      end
      table.insert(newCollection, price)
    end
    priceCollection = newCollection
    priceSum = TableUtils.sum(priceCollection)
  end

  self.prices = priceCollection
  self.sum = priceSum

  return self
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function SummationRound.newRandomRound(normalizedDifficulty, gameProgress, isFirstRound)
  local normalizedGameProgress = math.min(1, gameProgress / ROUND_COUNT_TO_MAX_DIFFICULTY)
  local self = generateRandomRound(normalizedDifficulty, normalizedGameProgress, isFirstRound)

  return self
end
