--- SummationScene
-- @author Created by Jay Austria on August 2, 2014.

class("SummationScene")
require("utils.StateMachine")
require("SummationColorUtils")
require("SummationCommon")
require("SummationFireworks")
require("SummationGameNumberLine")
require("SummationPriceDisplay")
require("SummationInteractionHelper")
require("tutorials.TutorialHighlighter")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKING_COLOR = ColorUtils.colorRGBToStd(47, 41, 51)
local TUTORIAL_OVERLAY_COLOR = {0.039, 0.055, 0.082, 0.6}
local TUTORIAL_MASK_COLOR = {1, 1, 1, 1}
local PRICE_DISPLAY_WIDTH = Screen.getSafeWidthPart(0.6)
local PRICE_DISPLAY_MARGIN_X = Screen.getSafeWidthPart(0.15)
local PRICE_DISPLAY_HEIGHT = Screen.fullHeight
local PRICE_DISPLAY_LOCATION_X = Screen.safeRight - PRICE_DISPLAY_WIDTH * 0.5 - PRICE_DISPLAY_MARGIN_X
local PRICE_DISPLAY_LOCATION_Y = PRICE_DISPLAY_HEIGHT * 0.5
local FIREWORKS_BASE_X = Screen.getWidthPart(0.6)
local FIREWORKS_OFFSET_X_MAX = Screen.getWidthPart(0.3)
local FIREWORKS_RADIUS = 50
local SCREEN_DIVISIONS = 3
local STATE_UNCHANGED = "UNCHANGED"
local STATE_RESPONDING = "RESPONDING"
local STATE_PERFECT = "PERFECT"
local INTERACTION_HELPER_LOCATION_Y = Screen.getSafeY(0.12)
local BAR_HEIGHT_PADDING_Y = 2

local Z_INDEX = {
  BACKING = 0,
  UI = 5,
  TUTORIAL = 10,
  GAME_ELEMENTS = 20,
  FX = 30
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function runBackingColorToAction(self, targetColor)
  SummationColorUtils.morphElementColors(self, targetColor, SummationCommon.COLOR_MORPH_RESPONDING_DURATION)
end

local function showBacking(self)
  runBackingColorToAction(self.backingLayer, BACKING_COLOR)
end

local function hideBacking(self)
  runBackingColorToAction(self.backingLayer, SummationCommon.WHITE_COLOR)
end

local function returnToUnchanged(self)
  hideBacking(self)
end

local function onResponseInternal(self, launchPuckActions)
  local responseInternalWrapper = function()
    -- This is to allow a user to tap to respond
    -- We let the bar finish animating before we fall through to
    -- disabling the isAnimatingResponse flag
    waitUntil(function() return self.gameNumberLine:isDoneAnimating() end)
    Sound.playSoundEffect("estimation_puck_woosh")
    launchPuckActions:start()
  end

  self.isAnimatingResponse = true

  return runInCoroutine(responseInternalWrapper)
end

local function isPuckInRegion(puckY, bottom, top)
  return (bottom < puckY) and (puckY < top)
end

local function getRandomXForFireworks()
  local randomOffset = math.random() * FIREWORKS_OFFSET_X_MAX - FIREWORKS_OFFSET_X_MAX * 0.5
  local x = FIREWORKS_BASE_X + randomOffset
  return x
end

local function getFireworksYsFromPuckY(puckY)
  local regionHeight = Screen.fullHeight / SCREEN_DIVISIONS

  local bottomThreshold = 0
  local topThreshold = regionHeight

  local fireworksYs = {}
  for i = 1, SCREEN_DIVISIONS do
    if not isPuckInRegion(puckY, bottomThreshold, topThreshold) then
      table.insert(fireworksYs, bottomThreshold + regionHeight * 0.5)
    end

    bottomThreshold = bottomThreshold + regionHeight
    topThreshold = topThreshold + regionHeight
  end

  return fireworksYs
end

local function doFireworks(self)
  local puckYModelSpace = self.gameNumberLine:getPuckY()
  local _, puckYWorldSpace, _ = self:transformLoc(self.gameNumberLine, 0, puckYModelSpace, 0)
  local fireworksYs = getFireworksYsFromPuckY(puckYWorldSpace)

  for _, fireworks in pairs(self.fireworkCollection) do
    local x = getRandomXForFireworks()

    local y = fireworksYs[1]
    table.remove(fireworksYs, 1)

    fireworks:setLoc(x, y)
    fireworks:runAnimation()
  end
end

local function addVisualFunctionsToWidgets(self)
  -- Note: some widgets may already have these functions defined so only inject them if they dont exist
  for _, widget in pairs(self.uiWidgets) do
    if not widget.enterRespondingState then
      local enterRespondingStateFunction = function()
        SummationColorUtils.morphElementColors(self, SummationCommon.UI_COMPONENTS_COLOR_RESPONDING, SummationCommon.COLOR_MORPH_RESPONDING_DURATION)
      end
      widget.enterRespondingState = enterRespondingStateFunction
    end

    if not widget.leaveRespondingState then
      local leaveRespondingStateFunction = function()
        SummationColorUtils.morphElementColors(self, SummationCommon.UI_COMPONENTS_COLOR, SummationCommon.COLOR_MORPH_RESPONDING_DURATION)
      end
      widget.leaveRespondingState = leaveRespondingStateFunction
    end

    if not widget.enterPerfectState then
      local enterPerfectStateFunction = function()
        SummationColorUtils.flipElementColor(self,
                                             SummationCommon.WHITE_COLOR,
                                             SummationCommon.UI_COMPONENTS_COLOR,
                                             SummationCommon.FLIP_TO_DURATION,
                                             SummationCommon.FLIP_FROM_DURATION,
                                             SummationCommon.FLIP_DELAY)
      end
      widget.enterPerfectState = enterPerfectStateFunction
    end
  end
end

local function onEnterUnchangedState(self)
  self.gameNumberLine:resetMarkColors()
end

local function onEnterRespondingState(self)
  showBacking(self)
  self.priceDisplay:morphToRespondingState()
  self.gameNumberLine:enterRespondingState()

  for _, widget in pairs(self.uiWidgets) do
    if widget.enterRespondingState then
      widget:enterRespondingState()
    end
  end
end

local function onLeaveRespondingState(self)
  if not self.isInTutorial then
    self.priceDisplay:clear()
  end

  returnToUnchanged(self)

  self.gameNumberLine:leaveRespondingState()

  for _, widget in pairs(self.uiWidgets) do
    if widget.leaveRespondingState then
      widget:leaveRespondingState()
    end
  end
end

local function onEnterPerfectState(self)
  self.priceDisplay:clear()
  self.perfectAnimation = SummationColorUtils.flipElementColor(self.backingLayer,
                                                               SummationCommon.FULLY_CORRECT_COLOR,
                                                               SummationCommon.WHITE_COLOR,
                                                               SummationCommon.FLIP_TO_DURATION,
                                                               SummationCommon.FLIP_FROM_DURATION,
                                                               SummationCommon.FLIP_DELAY)
  doFireworks(self)
  self.gameNumberLine:morphToPerfectState()

  for _, widget in pairs(self.uiWidgets) do
    if widget.enterPerfectState then
      widget:enterPerfectState()
    end
  end
end

local function initializeVisualStateMachine(self, stateMachine)
  local enterUnchangedWrapper = function()
    onEnterUnchangedState(self)
  end

  local unchangedTransitions = { STATE_RESPONDING, STATE_PERFECT }
  stateMachine:addState(STATE_UNCHANGED, enterUnchangedWrapper, nil, unpack(unchangedTransitions))


  local enterRespondingWrapper = function()
    onEnterRespondingState(self)
  end

  local leaveRespondingWrapper = function()
    onLeaveRespondingState(self)
  end

  local respondingTransitions = { STATE_UNCHANGED, STATE_PERFECT }
  stateMachine:addState(STATE_RESPONDING, enterRespondingWrapper, leaveRespondingWrapper, unpack(respondingTransitions))

  local enterPerfectWrapper = function()
    onEnterPerfectState(self)
  end

  local perfectTransitions = { STATE_UNCHANGED }
  stateMachine:addState(STATE_PERFECT, enterPerfectWrapper, nil, unpack(perfectTransitions))

  stateMachine:transition(STATE_UNCHANGED)
end

local function getNumberlineY(self, y)
  local height = self.gameNumberLine:getHeight()
  local nx, ny = self.gameNumberLine:getLoc()

  return y - (ny - (height / 2))
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SummationScene:animateCorrectResponse(response, correctSum, puckArriveCallback, isPerfect)
  local puckArriveCorrectWrapper = function()
    if puckArriveCallback then
      puckArriveCallback()
    end
    if isPerfect then
      Sound.playSoundEffect("estimation_perfect")
    else
      Sound.playSoundEffect("estimation_nice")
    end
  end

  local launchAction = self.gameNumberLine:getLaunchPuckToCorrectValueAction(correctSum, response, puckArriveCorrectWrapper, isPerfect)

  return onResponseInternal(self, launchAction)
end

function SummationScene:animateIncorrectResponse(response, correctSum)
  local puckArriveIncorrectWrapper = function()
    Sound.playSoundEffect("estimation_wrong")
  end

  local launchPuckAction = self.gameNumberLine:getLaunchPuckToValueIncorrectAction(correctSum, response, puckArriveIncorrectWrapper)

  return onResponseInternal(self, launchPuckAction)
end

function SummationScene:startRound(priceCollection, correctSum, isFirstRound)
  if (not self.shouldAdjustPriceY) then
    self.priceDisplay:setLoc(PRICE_DISPLAY_LOCATION_X, PRICE_DISPLAY_LOCATION_Y)
  end

  local displayPriceAnimation = self.priceDisplay:createDisplayPricesAnimation(priceCollection)
  self.correctSum = correctSum
  self.shouldAdjustPriceY = false

  self.interactionHelper:setLoc(self.barLocationX, INTERACTION_HELPER_LOCATION_Y)
  local displayInteractionHelper = self.interactionHelper:createDisplayAnimation(isFirstRound)

  local roundStartAnimation = AHSpawn.new(displayPriceAnimation, displayInteractionHelper)
  roundStartAnimation:start()
end

function SummationScene:onTouchBegan(x, y)
  self.tutorialLock = false
  self.gameNumberLine:notifyTouchBegan(getNumberlineY(self, y))

  self.visualStateMachine:transition(STATE_RESPONDING)
  local reducedVolume = 0.5
  Sound.playSoundEffect("estimation_tap_hold", reducedVolume)
end

function SummationScene:onTouchMoved(x, y)
  self.gameNumberLine:notifyUpdateHighlightLocationY(getNumberlineY(self, y))
end

function SummationScene:onTouchEnded(x, y)
  self.visualStateMachine:transition(STATE_UNCHANGED)
end

function SummationScene:getValueFromLocationY(y)
  local height = self.gameNumberLine:getHeight()
  local t = getNumberlineY(self, y) / height
  local tClamped = MathUtils.clamp(t, 0.0, 1.0)

  local value = MathUtils.interpolateLinearFromNormalized(tClamped, SummationCommon.NUMBER_LINE_MINIMUM, SummationCommon.NUMBER_LINE_MAXIMUM)

  return value
end

function SummationScene:getLocationYFromValue(value)
  local fraction = (value - SummationCommon.NUMBER_LINE_MINIMUM) / (SummationCommon.NUMBER_LINE_MAXIMUM - SummationCommon. NUMBER_LINE_MINIMUM)
  local numberLineBottom = self.gameNumberLine:getY() - self.gameNumberLine:getHeightPart(0.5)
  return numberLineBottom + self.gameNumberLine:getHeightPart(fraction)
end

function SummationScene:clear()
  self.visualStateMachine:transition(STATE_UNCHANGED)

  for _, fireworks in pairs(self.fireworkCollection) do
    fireworks:reset()
  end
end

function SummationScene:prepForLoseAnimation()
  if self.perfectAnimation then
    self.perfectAnimation:stop()
  end

  returnToUnchanged(self)

  self.gameNumberLine:clear()

  for _, fireworks in pairs(self.fireworkCollection) do
    fireworks:reset()
  end
end

function SummationScene:clearNumberLine()
  self.gameNumberLine:clear()
end

function SummationScene:isAnimating()
  return self.isAnimatingResponse
end

function SummationScene:showPerfectVisual()
  self.visualStateMachine:transition(STATE_PERFECT)
end

function SummationScene:animateGameLoss()
  SummationScreenShake.playScreenShake()
  self.priceDisplay:doLossAnimation()
  self.gameNumberLine:doLossAnimation()
end

function SummationScene:doIntro()
  self.gameNumberLine:doIntro()
end

function SummationScene:setInteractionHelperVisible(shouldShow)
  if shouldShow then
    self.interactionHelper:setLoc(self.barLocationX, INTERACTION_HELPER_LOCATION_Y)
    self.interactionHelper:show()
  else
    self.interactionHelper:hide()
  end
end

function SummationScene:setTutorialMode(isInTutorial)
  self.isInTutorial = isInTutorial
end

function SummationScene:waitForTutorial()
  self.tutorialLock = true
  local interactionHelper = self.interactionHelper
  interactionHelper:stopAnimating()

  local trackingCoroutines = {}

  local generateHoleCallback = function()
    local target = interactionHelper
    local targetWidth, targetHeight = target:getDims()
    local paddingToHideFrameLagWhenTracking = 4
    targetWidth = targetWidth - paddingToHideFrameLagWhenTracking * 0.5
    targetHeight = targetHeight - paddingToHideFrameLagWhenTracking

    local hole = FilledRectangle.new(targetWidth, targetHeight, TUTORIAL_MASK_COLOR)

    local trackTarget = function()
      local targetVelocityY = 0
      local targetYPrevious = nil
      while true do
        local targetX, targetY = self:transformLoc(target)

        if (targetYPrevious ~= nil) then
          targetVelocityY = targetY - targetYPrevious
        end

        local xOffsetToHideFrameLagWhenTracking = paddingToHideFrameLagWhenTracking * 0.5
        local yOffsetToHideFrameLagWhenTracking = targetVelocityY + paddingToHideFrameLagWhenTracking * 0.5
        local holeX = targetX + hole:getWidthPart(0.5) + xOffsetToHideFrameLagWhenTracking
        local holeY = targetY + hole:getHeightPart(0.5) + yOffsetToHideFrameLagWhenTracking

        hole:setLoc(holeX, holeY)

        targetYPrevious = targetY

        coroutine.yield()
      end
    end

    hole.trackCoroutine = runInCoroutine(trackTarget)
    table.insert(trackingCoroutines, hole.trackCoroutine)

    return hole
  end

  local width, height = interactionHelper:getDims()
  local zSortBehindInteractionHelper = -10
  local backingRect = FilledRectangle.new(width, height, TUTORIAL_MASK_COLOR)
  backingRect:setLoc(interactionHelper:getWidthPart(0.5), interactionHelper:getHeightPart(0.5))
  interactionHelper:insert(backingRect, zSortBehindInteractionHelper)

  local tutorialOverlay = TutorialHighlighter.newWithMasks(
    nil,
    nil,
    0,
    false,
    false,
    false
  )

  tutorialOverlay:setOverlayColor(unpack(TUTORIAL_OVERLAY_COLOR))
  self:insertAtCenter(tutorialOverlay, Z_INDEX.TUTORIAL)

  local numberLine = self.gameNumberLine.numberLine
  local elementColor = {1, 1, 1, 0.7}
  numberLine:transitionColor(AHColorTo.new(numberLine, 0.167, AHEaseLinear.new(), elementColor))
  self.priceDisplay:animateToColor(elementColor, 0.167)

  local originY = interactionHelper:getY()
  local destinationY = self:getLocationYFromValue(self.correctSum)

  local slideLoop = AHSequence.new(
    tutorialOverlay:createFadeInAnimation(),
    AHLoop.new(
      AHSequence.new(
        AHSpawn.new(
          AHDelayed.new(0.55, function()
              tutorialOverlay:resetHighlights({generateHoleCallback})
              tutorialOverlay:runSinglePulse()
            end
          ),
          AHLocationYTo.new(interactionHelper, 1.0, AHEaseBezier.newFlat({0.7, 0, 0.4, 1}), destinationY)
        ),
        AHDelay.new(0.833),
        AHLocationYTo.new(interactionHelper, 1.0, AHEaseBezier.newFlat({0.5, 0, 0.3, 1}), originY),
        AHDelay.new(0.833),
        function()
          tutorialOverlay:resetHighlights()
        end
      )
    )
  )
  slideLoop:start()

  waitUntil(function() return self.tutorialLock == false end)
  slideLoop:stop()

  numberLine:resetMarkColors()
  wait(tutorialOverlay:createFadeOutAnimation():start())

  for _, coroutine in ipairs(trackingCoroutines) do
    if coroutine and coroutine:isBusy() then
      coroutine:stop()
    end
  end

  backingRect:removeSelf()
  tutorialOverlay:removeSelf()
end

function SummationScene:takeUI(uiOverlay)
  self:takeChild(uiOverlay, Z_INDEX.UI)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, highlightBarValueSpan)
  local widthWithShakeBuffer = Screen.getWidthPart(1.1)
  local heightWithShakeBuffer = Screen.getHeightPart(1.1)
  local backingLayer = FilledRectangle.new(widthWithShakeBuffer, heightWithShakeBuffer, SummationCommon.WHITE_COLOR)
  self:insertAtCenter(backingLayer, Z_INDEX.BACKING)
  self.backingLayer = backingLayer

  local priceDisplay = SummationPriceDisplay.new(PRICE_DISPLAY_WIDTH, PRICE_DISPLAY_HEIGHT)
  local firstRoundYAdjust = 50
  self.shouldAdjustPriceY = true
  priceDisplay:setLoc(PRICE_DISPLAY_LOCATION_X, PRICE_DISPLAY_LOCATION_Y + firstRoundYAdjust)
  self:insert(priceDisplay, Z_INDEX.GAME_ELEMENTS)
  self.priceDisplay = priceDisplay

  local gameNumberLine = SummationGameNumberLine.new(SummationCommon.NUMBER_LINE_WIDTH,
                                                     SummationCommon.NUMBER_LINE_HEIGHT,
                                                     SummationCommon.NUMBER_LINE_MINIMUM,
                                                     SummationCommon.NUMBER_LINE_MAXIMUM,
                                                     highlightBarValueSpan)
  gameNumberLine:setLoc(SummationCommon.NUMBER_LINE_LOCATION_X, SummationCommon.NUMBER_LINE_LOCATION_Y)
  self:insert(gameNumberLine, Z_INDEX.GAME_ELEMENTS)
  local resumeGameWrapper = function() self.isAnimatingResponse = false end
  gameNumberLine:setResumeGameCallback(resumeGameWrapper)
  self.gameNumberLine = gameNumberLine

  local barHeight = self.gameNumberLine:getMarkYOffset() * highlightBarValueSpan - SummationAnswerPuck.PUCK_HEIGHT - BAR_HEIGHT_PADDING_Y
  local barWidth = Screen.fullWidth - SummationCommon.NUMBER_LINE_LOCATION_X
  self.barLocationX = self.gameNumberLine:getX() - self.gameNumberLine:getWidthPart(0.5) + self.gameNumberLine:getBarLocationX()
  local interactionHelper = SummationInteractionHelper.newHidden(barWidth, barHeight)
  interactionHelper:setLoc(self.barLocationX, INTERACTION_HELPER_LOCATION_Y)
  self:insert(interactionHelper, Z_INDEX.GAME_ELEMENTS)
  self.interactionHelper = interactionHelper

  local fireworkCollection = {}
  for i = 1, 2 do
    local fireworks = SummationFireworks.new(FIREWORKS_RADIUS)
    local x = Screen.getWidthPart(0.5)
    local y = Screen.getWidthPart(0.25 * i)
    fireworks:setLoc(x, y)
    self:insert(fireworks, Z_INDEX.FX)
    table.insert(fireworkCollection, fireworks)
  end
  self.fireworkCollection = fireworkCollection
end

function SummationScene.new(widgets, partiallyCorrectDelta)
  local self = Group.newFullScreen()
  Mixin.mix(self, SummationScene)

  self.isAnimatingResponse = false
  self.uiWidgets = widgets
  self.isInTutorial = false

  local highlightBarValueSpan = partiallyCorrectDelta * 2
  createLayout(self, highlightBarValueSpan)
  addVisualFunctionsToWidgets(self)

  local visualStateMachine = StateMachine.new()
  initializeVisualStateMachine(self, visualStateMachine)
  self.visualStateMachine = visualStateMachine

  return self
end
