--- SunsetBackground
-- @author created by Jay Austria on November 7, 2013

class("SunsetBackground")
require("deprecated.ColorTracker")
require("SunsetCommon")
require("SunsetStarOverlay")
require("SunsetSailboat")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DEFAULT_COLOR = SunsetCommon.WHITE
local BOTTOM_GRADIENT_BOTTOM = ColorUtils.colorRGBToStd(200, 123, 148)
local BOTTOM_GRADIENT_TOP = SunsetCommon.WHITE
local TOP_GRADIENT_BOTTOM = ColorUtils.colorRGBToStd(255, 230, 210)
local TOP_GRADIENT_TOP = SunsetCommon.WHITE
local BOTTOM_GRADIENT_HEIGHT_PROP = 0.3
local TOP_GRADIENT_HEIGHT_PROP = 0.2

local COLORING_DOCK =
{
  {t = 0.0, keyColor = {1, 1, 1, 1}},
  {t = 1.0, keyColor = {1, 1, 1, 0.75}}
}

local COLORING_STARS =
{
  {t = 0.0, keyColor = {1, 1, 1, 1}},
  {t = 1.0, keyColor = {1, 1, 1, 0.7}}
}

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SunsetBackground:shiftColorWithDuration(value, duration)
  self.colorTrackerTop:shiftColorWithDuration(value, duration)
  self.colorTrackerBottom:shiftColorWithDuration(value, duration)
  self.colorTrackerDock:shiftColorWithDuration(value, duration)
  self.colorTrackerStars:shiftColorWithDuration(value, duration)
  for _, boat in ipairs(self.boats) do
    boat:shiftColorWithDuration(value, duration)
  end
end

function SunsetBackground:setPaused(isPaused)
  self.colorTrackerTop:setPaused(isPaused)
  self.colorTrackerBottom:setPaused(isPaused)
  self.colorTrackerDock:setPaused(isPaused)
  self.colorTrackerStars:setPaused(isPaused)
  for _, boat in ipairs(self.boats) do
    boat:setPaused(isPaused)
  end
end

function SunsetBackground:showEndAnimation()
  self.starsOverlay:animate()
  AHOpacityTo.new(self.starsOverlay, 1, nil, 1):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function SunsetBackground.new(horizonY)
  local dock = Sprite.newWithName("dock.png")
  local dockWidth, dockHeight = dock:getDims()

  local selfWidth = Screen.fullWidth + dockWidth
  local selfHeight = Screen.fullHeight
  local self = RectangleGradient.new(selfWidth, selfHeight, DEFAULT_COLOR, DEFAULT_COLOR)
  Mixin.mix(self, SunsetBackground)

  local colorTrackerTop = ColorTracker.newWithTable(SunsetCommon.COLORING_LEVELS_SKY_TOP)
  colorTrackerTop:linkColorNodeToSelf(self:getBeginColor())

  local colorTrackerBottom = ColorTracker.newWithTable(SunsetCommon.COLORING_LEVELS_SKY_BOTTOM)
  colorTrackerBottom:linkColorNodeToSelf(self:getEndColor())

  self.colorTrackerTop = colorTrackerTop
  self.colorTrackerBottom = colorTrackerBottom

  local bottomGradientWidth = selfWidth
  local bottomGradientHeight = selfHeight * BOTTOM_GRADIENT_HEIGHT_PROP
  local bottomGradient = RectangleGradient.new(bottomGradientWidth, bottomGradientHeight, BOTTOM_GRADIENT_TOP, BOTTOM_GRADIENT_BOTTOM)
  bottomGradient:setLoc(bottomGradientWidth * 0.5, bottomGradientHeight * 0.5)
  bottomGradient:setBlendMode(MOAIProp2D.BLEND_MULTIPLY)
  self:insert(bottomGradient)

  local topGradientWidth = selfWidth
  local topGradientHeight = selfHeight * TOP_GRADIENT_HEIGHT_PROP
  local topGradient = RectangleGradient.new(topGradientWidth, topGradientHeight, TOP_GRADIENT_TOP, TOP_GRADIENT_BOTTOM)
  topGradient:setLoc(topGradientWidth * 0.5, horizonY + topGradientHeight * 0.5)
  topGradient:setBlendMode(MOAIProp2D.BLEND_MULTIPLY)
  self:insert(topGradient)

  local horizonYFloored = math.floor(horizonY)
  local starsOverlay = SunsetStarOverlay.new(selfHeight, horizonYFloored)
  starsOverlay:setAnchorPointBottomLeft()
  starsOverlay:setOpacity(0)
  self:insert(starsOverlay)
  self.starsOverlay = starsOverlay

  local colorTrackerStars = ColorTracker.newWithTable(COLORING_STARS)
  colorTrackerStars:linkColorNodeToSelf(starsOverlay)
  self.colorTrackerStars = colorTrackerStars

  local dockX = selfWidth
  local dockY = horizonYFloored - math.floor(dockWidth * 0.45)
  dock:setAnchorPointBottomRight()
  dock:setLoc(dockX, dockY)
  self:insert(dock)

  local colorTrackerDock = ColorTracker.newWithTable(COLORING_DOCK)
  colorTrackerDock:linkColorNodeToSelf(dock)
  self.colorTrackerDock = colorTrackerDock

  local boatDescriptors =
  {
    {
      boatStyleIndex = 2,
      x = dockX - math.floor(dockWidth * 0.6),
      y = dockY - math.floor(dockHeight * 0.2),
      scale = 0.8,
      animationDelay = 0.0
    },
    {
      boatStyleIndex = 3,
      x = dockX - math.floor(dockWidth * 0.3),
      y = dockY - math.floor(dockHeight * 0.45),
      scale = 1.0,
      animationDelay = 0.7
    }
  }

  self.boats = {}

  for _, boatDescriptor in ipairs(boatDescriptors) do
    local sailboat = SunsetSailboat.new(boatDescriptor.boatStyleIndex)
    sailboat:setLoc(boatDescriptor.x, boatDescriptor.y)
    sailboat:setScl(boatDescriptor.scale)

    runInCoroutine(function()
      waitFor(boatDescriptor.animationDelay)
      sailboat:startStationaryAnimation()
    end)

    self:insert(sailboat)
    table.insert(self.boats, sailboat)
  end

  return self
end
