--- SunsetPromptTextView
-- @author Created by Jay Austria on August 30, 2015.

class("SunsetPromptTextView")
require("SunsetHighlight")
require("SunsetPromptTextModel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local LINE_SPACING = 10


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBaseAttributes()
  return {
    font = Font.new(FontFaces.DIN_OT_MEDIUM, SunsetCommon.PROMPT_FONT_SIZE),
    color = {1, 1, 1}
  }
end

local function getAttributesForLabels()
  return {
    default = {color = {1, 1, 1} },
    leftMargin = {itemType = "margin", spacing = 0, skipNextSpacing = true, allowBreakLineAfter = false},
    rightMargin = {itemType = "margin", spacing = 0, skipPreviousSpacing = true, allowBreakLineBefore = false},
    highlighted = {hasHighlight = true, isHighlighted = true}
  }
end

local function createHighlight(self, highlightText, correctionText)
  local viewsInHighlight = {}

  local minimumX = math.huge
  local maximumX = -math.huge

  local highlightedItems = self.textModel:getHighlightedItems()

  for _, item in ipairs(highlightedItems) do
    local view = self.viewsForItems[item]
    table.insert(viewsInHighlight, view)

    minimumX = math.min(minimumX, view:getXLeft())
    maximumX = math.max(maximumX, view:getXRight())
  end

  local highlight = SunsetHighlight.new(highlightText, correctionText)
  highlight:setAnchorPointBottom()
  local highlightX = minimumX + (maximumX - minimumX) * 0.5
  local offsetY = -3
  local highlightY = viewsInHighlight[1]:getYBottom() + offsetY

  self:insertAtLocation(highlight, highlightX, highlightY)

  self.highlight = highlight
end

local function playPromptErrorShake(self)
  local x = self:getX()
  local y = self:getY()
  local shakeOffsetX = 20
  local promptShakeAction = AHSequence.new(
    AHLocationXBy.new(self, 0.1, AHEaseSimpleInOut.new(), shakeOffsetX),
    function() self:setX(x + shakeOffsetX) end,
    AHLocationTo.new(self, 1.5, AHEaseElasticOut.new(), x, y)
  )

  promptShakeAction:start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SunsetPromptTextView:createItemView(item)
  local itemType = item:getItemType()

  local prop = nil
  if (itemType == "margin") then
    local marginPadding = SunsetCommon.HIGHLIGHT_PADDING_X
    prop = Group.new(marginPadding, 1)
  else
    prop = self.super.createItemView(self, item)

    local isVisible = (not item:getAttribute("isHighlighted"))
    prop:setVisible(isVisible)
  end

  return prop
end

function SunsetPromptTextView:setText(text)
  if (text ~= self:getText()) then
    local textInterpreter = self.textInterpreter
    local textModel = self.textModel

    local items = textInterpreter:interpretXmlText(text)
    textModel:resetItems(items)

    self.text = text
  end
end

function SunsetPromptTextView:getText()
  return self.text
end

function SunsetPromptTextView:playCorrectAnimation(didFlagValid)
  if (didFlagValid) then
    self.highlight:showAsCorrect()
  else
    self.highlight:showWithCorrectionBubble()
  end
end

function SunsetPromptTextView:playIncorrectAnimation(didFlagValid)
  if (didFlagValid) then
    self.highlight:showAsIncorrectWithCorrectionBubble()
  end

  playPromptErrorShake(self)
end


-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createElements(self)
  local baseAttributes = getBaseAttributes()
  local attributesForLabels = getAttributesForLabels()
  local textInterpreter = TextInterpreter.new(baseAttributes, attributesForLabels)
  self.textInterpreter = textInterpreter

  local textModel = SunsetPromptTextModel.new(textInterpreter:interpretXmlText(""))
  self:setModel(textModel)
end

function SunsetPromptTextView.new(prompt, highlightText, correctionText, textWidth)
  local self = TextView.new(textWidth, Alignment.CENTER, nil, LINE_SPACING)
  Mixin.mix(self, SunsetPromptTextView)

  createElements(self)

  self:setText(prompt)

  createHighlight(self, highlightText, correctionText)

  return self
end
