--- SunsetSailboat
-- @author created by Jay Austria on November 14, 2013

class("SunsetSailboat")
require("SunsetCommon")

local GROUP_WIDTH = 40
local GROUP_HEIGHT = 70
local SAIL_GAP_Y = 3
local PARTICLES_SCALE = 0.5

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BOAT_COLORS = {
  {
    top = {
      startColor = ColorUtils.colorRGBToStd(252, 102, 85),
      endColor = ColorUtils.colorRGBToStd(173, 43, 33)
    },
    bottom = {
      startColor = ColorUtils.colorRGBToStd(133, 41, 66),
      endColor = ColorUtils.colorRGBToStd(92, 18, 41)
    }
  },
  {
    top = {
      startColor = ColorUtils.colorRGBToStd(250, 156, 62),
      endColor = ColorUtils.colorRGBToStd(219, 111, 52)
    },
    bottom = {
      startColor = ColorUtils.colorRGBToStd(120, 15, 32),
      endColor = ColorUtils.colorRGBToStd(105, 12, 41)
    }
  },
  {
    top = {
      startColor = ColorUtils.colorRGBToStd(250, 62, 87),
      endColor = ColorUtils.colorRGBToStd(223, 55, 79)
    },
    bottom = {
      startColor = ColorUtils.colorRGBToStd(120, 15, 23),
      endColor = ColorUtils.colorRGBToStd(107, 13, 21)
    }
  }
}

local function getColoring(colorStruct, isReflection)
  local opacity = 1.0
  if isReflection then opacity = SunsetCommon.REFLECTION_OPACITY end

  return {
    {t = 0.0, keyColor = {colorStruct.startColor[1], colorStruct.startColor[2], colorStruct.startColor[3], opacity}},
    {t = 1.0, keyColor = {colorStruct.endColor[1], colorStruct.endColor[2], colorStruct.endColor[3], opacity}}
  }
end

local function getTopColor(boatType, isReflection)
  return getColoring(BOAT_COLORS[boatType].top, isReflection)
end

local function getBottomColor(boatType, isReflection)
  return getColoring(BOAT_COLORS[boatType].bottom, isReflection)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SunsetSailboat:startMovingAnimation()
  if (self.ocillateAction) then
    self.ocillateAction:stop()
    self.ocillateAction = nil
  end

  if (self.boatAnimation) then
    self.boatAnimation:stop()
    self.boatAnimation = nil
  end

  local function floatAnimationFunction()
    local TILT_DURATION = 0.1
    local TILT_AMOUNT = 10

    local tiltForwardBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, nil, -TILT_AMOUNT)
    local tiltForwardReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, nil, TILT_AMOUNT)
    local tiltForwardSpawn = AHSpawn.new(tiltForwardBoatAction, tiltForwardReflectionAction)
    local freezeDelay = AHDelay.new(SunsetCommon.BOAT_MOVE_DURATION)
    local floatAction = AHSequence.new(tiltForwardSpawn, freezeDelay)
    floatAction:start()
  end

  local boatAnimation = newCoroutineWithFunction(floatAnimationFunction)
  self.boatAnimation = boatAnimation:start()
end

function SunsetSailboat:startJoltAnimation()
  local TILT_DURATION = 0.6
  local TILT_AMOUNT = -3
  local MOVE_UP_DURATION = 0.2
  local MOVE_DOWN_DURATION = 0.4
  local MOVE_AMOUNT = 1

  local tiltBackBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, nil, -TILT_AMOUNT)
  local tiltBackReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, nil, TILT_AMOUNT)
  local moveUpBoatAction = AHLocationYBy.new(self.topGroup, MOVE_UP_DURATION, nil, MOVE_AMOUNT)
  local moveDownBoatAction = AHLocationYBy.new(self.topGroup, MOVE_DOWN_DURATION, nil, -MOVE_AMOUNT)
  local moveUpDownBoatSequence = AHSequence.new(moveUpBoatAction, moveDownBoatAction)

  local moveUpReflectionAction = AHLocationYBy.new(self.bottomGroup, MOVE_UP_DURATION, nil, -MOVE_AMOUNT)
  local moveDownReflectionAction = AHLocationYBy.new(self.bottomGroup, MOVE_DOWN_DURATION, nil, MOVE_AMOUNT)
  local moveUpDowReflectionSequence = AHSequence.new(moveUpReflectionAction, moveDownReflectionAction)
  local tiltBackSpawn = AHSpawn.new(tiltBackBoatAction, tiltBackReflectionAction, moveUpDownBoatSequence, moveUpDowReflectionSequence)

  return tiltBackSpawn:start()
end

function SunsetSailboat:startStationaryAnimation()
  if (self.boatAnimation) then
    self.boatAnimation:stop()
    self.boatAnimation = nil
  end

  local function stationaryAnimationFunction()
    local TILT_DURATION = 1.4
    local TILT_AMOUNT = 5

    local tiltEase = AHEaseSimpleInOut.new(2.0)

    local tiltBackBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltBackReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltBackSpawn = AHSpawn.new(tiltBackBoatAction, tiltBackReflectionAction)

    local tiltForwardBoatAction = AHRotationTo.new(self.topGroup, TILT_DURATION, tiltEase, -TILT_AMOUNT)
    local tiltForwardReflectionAction = AHRotationTo.new(self.bottomGroup, TILT_DURATION, tiltEase, TILT_AMOUNT)
    local tiltForwardSpawn = AHSpawn.new(tiltForwardBoatAction, tiltForwardReflectionAction)

    local tiltSequence = AHSequence.new(tiltForwardSpawn, tiltBackSpawn)
    local ocillateAction = AHLoop.new(tiltSequence)

    self.ocillateAction = ocillateAction:start()
  end

  local boatAnimation = newCoroutineWithFunction(stationaryAnimationFunction)
  self.boatAnimation = boatAnimation:start()

  self.rippleParticles:start()
  self.rippleParticles:startSystem()
end

function SunsetSailboat:shiftColorWithDuration(value, duration)
  self.colorTrackerBoatTop:shiftColorWithDuration(value, duration)
  self.colorTrackerBoatBottom:shiftColorWithDuration(value, duration)
  self.colorTrackerReflectionTop:shiftColorWithDuration(value, duration)
  self.colorTrackerReflectionBottom:shiftColorWithDuration(value, duration)
end

function SunsetSailboat:setPaused(isPaused)
  self.colorTrackerBoatTop:setPaused(isPaused)
  self.colorTrackerBoatBottom:setPaused(isPaused)
  self.colorTrackerReflectionTop:setPaused(isPaused)
  self.colorTrackerReflectionBottom:setPaused(isPaused)
end

function SunsetSailboat.new(boatType)
  boatType = boatType or 1
  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  Mixin.mix(self, SunsetSailboat)

  local topGroup = Group.new(GROUP_WIDTH, GROUP_HEIGHT * 0.5)
  topGroup:setAnchorPointBottom()
  self:insertAtCenter(topGroup)
  self.topGroup = topGroup

  local boatBottomHalf = Sprite.newWithName("sailboat_bottom.png")
  local _, bottomH = boatBottomHalf:getDims()
  boatBottomHalf:setLoc(GROUP_WIDTH * 0.5, bottomH * 0.5)
  topGroup:insert(boatBottomHalf)
  self.boatBottomHalf = boatBottomHalf

  local boatTopHalf = Sprite.newWithName("sailboat_top.png")
  local _, topH = boatTopHalf:getDims()
  boatTopHalf:setLoc(GROUP_WIDTH * 0.5, bottomH + topH * 0.5 + SAIL_GAP_Y)
  topGroup:insert(boatTopHalf)

  local bottomGroup = Group.new(GROUP_WIDTH, GROUP_HEIGHT * 0.5)
  bottomGroup:setAnchorPointTop()
  self:insertAtCenter(bottomGroup)
  self.bottomGroup = bottomGroup

  local reflectionBottomHalf = Sprite.newWithName("sailboat_bottom.png")
  reflectionBottomHalf:setScl(1.0, -1.0)
  reflectionBottomHalf:setLoc(GROUP_WIDTH * 0.5, GROUP_HEIGHT * 0.5 - bottomH * 0.5)
  bottomGroup:insert(reflectionBottomHalf)

  local reflectionTopHalf = Sprite.newWithName("sailboat_top.png")
  reflectionTopHalf:setScl(1.0, -1.0)
  reflectionTopHalf:setLoc(GROUP_WIDTH * 0.5, GROUP_HEIGHT * 0.5 - bottomH - topH * 0.5 - SAIL_GAP_Y)
  bottomGroup:insert(reflectionTopHalf)

  local rippleParticles = Particles.new("ripples.pex")
  rippleParticles:setScl(PARTICLES_SCALE)
  self:insertAtCenter(rippleParticles)
  self.rippleParticles = rippleParticles

  local colorTrackerBoatBottom = ColorTracker.newWithTable(getBottomColor(boatType))
  colorTrackerBoatBottom:linkColorNodeToSelf(boatBottomHalf)
  self.colorTrackerBoatBottom = colorTrackerBoatBottom

  local colorTrackerBoatTop = ColorTracker.newWithTable(getTopColor(boatType))
  colorTrackerBoatTop:linkColorNodeToSelf(boatTopHalf)
  self.colorTrackerBoatTop = colorTrackerBoatTop

  local colorTrackerReflectionBottom = ColorTracker.newWithTable(getBottomColor(boatType, true))
  colorTrackerReflectionBottom:linkColorNodeToSelf(reflectionBottomHalf)
  self.colorTrackerReflectionBottom = colorTrackerReflectionBottom

  local colorTrackerReflectionTop = ColorTracker.newWithTable(getTopColor(boatType, true))
  colorTrackerReflectionTop:linkColorNodeToSelf(reflectionTopHalf)
  self.colorTrackerReflectionTop = colorTrackerReflectionTop

  self:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)

  return self
end
