--- SunsetScene
-- @author created by Jay Austria on November 18, 2013

class("SunsetScene")
require("Particles")
require("SunsetBackground")
require("SunsetCloudScroller")
require("SunsetSailboat")
require("SunsetSun")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SUN_RADIUS_MULTIPLIER = 0.28
local BOAT_SCALE = 0.65
local BOAT_X_START = Screen.fullWidth * 0.1
local BOAT_X_END = Screen.fullWidth * 0.95
local SAILBOAT_OFFSET_Y = -2
local FLASH_COLOR = {1.0, 1.0, 1.0}
local BOAT_PARTICLES_BASE_X = Screen.getWidthPart(0.05)


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function animateFullScreenFlash(self)
  local flashLayer = self.fullScreenFlash

  local flashSequence = AHSequence.new(
    AHOpacityTo.new(flashLayer, 0.35, AHEaseSimpleOut.new(), 0.2),
    AHDelay.new(0.2),
    AHOpacityTo.new(flashLayer, 0.35, nil, 0.0)
  )

  self.flashActionInterrupt:insert(flashSequence)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SunsetScene:startScroller(roundTime)
  self.clouds:start(roundTime)
end

function SunsetScene:moveScoreIcon(scoreToSet, maxScore)
  runInCoroutine(function()
      local boat = self.scoreIcon
      local xDifference = BOAT_X_END - BOAT_X_START
      local scoreProportion = math.min(1.0, (scoreToSet / maxScore))
      local boatDestinationX = scoreProportion * xDifference + BOAT_X_START
      local boatMoveAction = AHLocationXTo.new(boat, SunsetCommon.BOAT_MOVE_DURATION, nil, boatDestinationX)

      local boatParticles = self.boatParticles
      local boatWidth = boat:getWidth()
      local particlesDestinationX = BOAT_PARTICLES_BASE_X + boatDestinationX - boatWidth
      local particlesMoveAction = boatParticles:createSourcePositionXTo(SunsetCommon.BOAT_MOVE_DURATION, nil, particlesDestinationX)

      local particlesEmitDuration = SunsetCommon.BOAT_MOVE_DURATION * 0.75
      local particlesEmitAction = boatParticles:createEmitAction(particlesEmitDuration)

      boat:startMovingAnimation()
      particlesMoveAction:start()
      particlesEmitAction:start()

      wait(boatMoveAction:start())
      wait(boat:startJoltAnimation())

      boat:startStationaryAnimation()
    end)
end

function SunsetScene:updateScene(startT, endT, duration)
  self.sun:fadeSetAmount(startT, endT, duration)
  self.clouds:shiftColorWithDuration(endT, duration)
  self.background:shiftColorWithDuration(endT, duration)
  self.waterColorTracker:shiftColorWithDuration(endT, duration)
  self.scoreIcon:shiftColorWithDuration(endT, duration)
end

function SunsetScene:setPaused(isPaused)
  self.sun:setPaused(isPaused)
  self.clouds:setPaused(isPaused)
  self.background:setPaused(isPaused)
  self.waterColorTracker:setPaused(isPaused)
  self.scoreIcon:setPaused(isPaused)
end

function SunsetScene:showEndAnimation()
  self.background:showEndAnimation()
  local selfW = self:getDims()
  local bgW = self.background:getDims()
  local distance = bgW - selfW

  local endAction = AHSpawn.new(
    self.buttonBar:createFadeOutAction(),
    AHLocationXBy.new(self, 1, nil, -distance),
    AHLocationXBy.new(self.buttonBar, 1, nil, -distance)
  )

  endAction:start()
end

function SunsetScene:showStartAnimation()
  local fadeAction = self.buttonBar:createFadeInAction()
  fadeAction:start()
end

function SunsetScene:playCorrectAnimation(didFlagValid)
    Sound.playSoundEffect("correct")
    self.buttonBar:playCorrectAnimation(didFlagValid)
    self.prompt:playCorrectAnimation(didFlagValid)
    animateFullScreenFlash(self)
end

function SunsetScene:playIncorrectAnimation(didFlagValid)
    Sound.playSoundEffect("incorrect")
    self.buttonBar:playIncorrectAnimation(didFlagValid)
    self.prompt:playIncorrectAnimation(didFlagValid)
end

function SunsetScene:setPrompt(prompt)
  self.prompt = prompt
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, buttonBar)
  local background = SunsetBackground.new(SunsetCommon.HORIZON_Y)
  background:setAnchorPointBottomLeft()
  self:insert(background)
  self.background = background

  local waterTexture = Sprite.newWithName("watertexture.png")
  local waterTextureWidth, waterTextureHeight = waterTexture:getDims()
  waterTexture:setScl(Screen.fullWidth / waterTextureWidth, 1.0)
  waterTexture:setLoc(Screen.fullWidth * 0.5, SunsetCommon.HORIZON_Y - waterTextureHeight * 0.5)
  self:insert(waterTexture)

  local waterColorTracker = ColorTracker.newWithTable(SunsetCommon.COLORING_LEVELS_SKY_TOP)
  waterColorTracker:linkColorNodeToSelf(waterTexture)
  self.waterColorTracker = waterColorTracker

  local clouds = SunsetCloudScroller.new()
  clouds:setLoc(Screen.fullWidth * 0.5, SunsetCommon.HORIZON_Y)
  self:insert(clouds)
  self.clouds = clouds

  local sun = SunsetSun.new(SUN_RADIUS_MULTIPLIER * Screen.fullWidth)
  sun:setLoc(Screen.fullWidth * 0.50, SunsetCommon.HORIZON_Y)
  self:insert(sun)
  self.sun = sun

  local boatParticles = Particles.new("sun_flowers.pex")
  boatParticles:setLoc(0.0, 0.0)
  boatParticles:setSourcePosition(BOAT_PARTICLES_BASE_X, SunsetCommon.HORIZON_Y)
  boatParticles:start()
  self:insert(boatParticles)
  self.boatParticles = boatParticles

  local scoreIcon = SunsetSailboat.new()
  scoreIcon:setScl(BOAT_SCALE)
  scoreIcon:setLoc(BOAT_X_START, SunsetCommon.HORIZON_Y + SAILBOAT_OFFSET_Y)
  self:insert(scoreIcon)
  self.scoreIcon = scoreIcon
  self.scoreIcon:startStationaryAnimation()

  self:insertAtLocationFractional(buttonBar, 0.5, 0.12)
  buttonBar:prepareForShowAction()
  self.buttonBar = buttonBar

  local fullScreenFlash = FilledRectangle.newFullScreen(FLASH_COLOR)
  fullScreenFlash:setOpacity(0.0)
  self:insertAtCenter(fullScreenFlash)
  self.fullScreenFlash = fullScreenFlash
end

function SunsetScene.new(buttonBar)
  local self = Group.newFullScreen()
  Mixin.mix(self, SunsetScene)

  createLayout(self, buttonBar)

  self.flashActionInterrupt = ActionInterrupt.new()

  return self
end
