--- SwapGame
-- @author Template created by Aaron Barrett on September 27, 2013.

class("SwapGame")
require("bonuses.ResponseSpeedTracker")
require("StandardUIBuilder")
require("SwapButton")
require("SwapCommon")
require("SwapConfiguration")
require("SwapProgressIndicator")
require("SwapRoundGenerator")
require("SwapScene")
require("data.generators.LearningFocusedRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = ColorUtils.colorHex24ToStd(0x5f5745)
local CORRECT_RESPONSE_DELAY = 1.0

local function configureGameDifficulty(self)
  self.correctNeeded = SwapConfiguration.CORRECT_NEEDED_TO_WIN
  self.slideSeconds = SwapConfiguration.SLIDE_DURATION_IN_SECONDS
  self.incorrectAllowedRankTwo = SwapConfiguration.INCORRECT_RESPONSES_ALLOWED_RANK_TWO
  self.incorrectAllowedRankThree = SwapConfiguration.INCORRECT_RESPONSES_ALLOWED_RANK_THREE

  local rankTwoScore = SwapConfiguration.getRankTwoScore()
  local rankThreeScore = SwapConfiguration.getRankThreeScore()

  -- No pass score. Game ends with progress indicator is filled
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = SwapConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.CONSISTENCY_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- End of game conditions -------------

local function didWin(self)
  return self.progressIndicator:getProgress() >= 1.0
end

local function didLose(self)
  return self.progressIndicator:getProgress() <= 0
end

local function isGameEnd(self)
  return didWin(self) or didLose(self)
end

-- Misc -------------------------------

local function updateScore(self)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function applyScorePenalty()
  local penaltyPoints = -SwapConfiguration.POINTS_PER_CORRECT_RESPONSE
  GameSession.increaseScore(penaltyPoints)
end

local function setButtonsEnabled(self, enabled)
  self.swapButton:setTouchEnabled(enabled)
  self.acceptButton:setTouchEnabled(enabled)
end

local function getQuestionIdentifier(self)
  if (self.isPromptCorrect) then
    return "correct_form"
  else
    return "incorrect_form"
  end
end

local function endRound(self)
  self.responseWaitLock:stop()
end

-- Handlers ---------------------------

local function onCorrectResponse(self, pressedButton)
  pressedButton:showAsCorrect()

  self.responseSpeedTracker:notifyCorrectResponse()

  GameSession.contentTracker:notifyCorrectAnswer(self.currentConceptId, getQuestionIdentifier(self))
  GameSession.notifyCorrectAnswer(self.currentConceptId)
  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(SwapConfiguration.POINTS_PER_CORRECT_RESPONSE)

  Sound.playSoundEffect("correct")

  local shouldSwap = not self.isPromptCorrect
  self.scene:playCorrectAnimation(shouldSwap)

  updateScore(self)
  local progressAction = self.progressIndicator:addBar()

  self.progressIndicator:pauseSliding(true)
  waitFor(CORRECT_RESPONSE_DELAY)
  self.progressIndicator:pauseSliding(false)

  coroutine.yield()
  if didWin(self) then
    local accuracyBonusCountReceived = 0
    if (self.incorrectResponses <= self.incorrectAllowedRankThree) then
      accuracyBonusCountReceived = 2
    elseif (self.incorrectResponses <= self.incorrectAllowedRankTwo) then
      accuracyBonusCountReceived = 1
    end

    local accuracyBonusPoints = accuracyBonusCountReceived * SwapConfiguration.POINTS_PER_ACCURACY_BONUS
    local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.CONSISTENCY_BONUS)
    if (accuracyBonusPoints > 0) then
      self.bonusTray:indicateSimpleTextBonus(notificationString, accuracyBonusPoints)
    end
    GameSession.bonusReporter:setBonusCount(BonusTypes.TYPES.CONSISTENCY_BONUS, self.incorrectResponses)
    GameSession.increaseScore(accuracyBonusPoints)
    updateScore(self)

    GameSession.notifyGameTimeEnd()
    Sound.playSoundEffect("win")
    self:showDefaultGameWinOutro()

    self.scene:playGameWinAnimation()
  end
  wait(progressAction)

  pressedButton:revert()
  endRound(self)
end

local function onIncorrectResponse(self, pressedButton)
  runInCoroutine(function()
    self.progressIndicator:pauseSliding(true)
    self.incorrectResponses = self.incorrectResponses + 1
    self.responseSpeedTracker:notifyIncorrectResponse()
    self.roundGenerator:notifyMistake(self.currentConceptId)
    GameSession.contentTracker:notifyIncorrectAnswer(self.currentConceptId, getQuestionIdentifier(self))
    GameSession.scoredIncorrectAnswer(self.currentConceptId)
    GameSession.incrementIncorrectAnswerCount()
    Sound.playSoundEffect("incorrect")

    pressedButton:showAsIncorrect()

    local shouldSwap = not self.isPromptCorrect
    self.scene:playHighlightIncorrectAnimationAndWaitForRationaleTap(shouldSwap)

    if not SwapConfiguration.scoreProtectionIsActiveForProgress(self.progressIndicator:getProgressStep()) then
      self.progressIndicator:removeBarFraction(SwapConfiguration.INCORRECT_PROGRESS_PENALTY_FRACTION)
    end

    self.progressIndicator:pauseSliding(false)
    pressedButton:revert()

    endRound(self)
  end)
end

local function onSwapPressed(self)
  if self.tutorialTimer then
    self.tutorialTimer:stop()
  end

  setButtonsEnabled(self, false)
  if (not self.isPromptCorrect) then
    onCorrectResponse(self, self.swapButton)
  else
    onIncorrectResponse(self, self.swapButton)
  end
end

local function onAcceptPressed(self)
  if self.tutorialTimer then
    self.tutorialTimer:stop()
  end

  setButtonsEnabled(self, false)
  if (self.isPromptCorrect) then
    onCorrectResponse(self, self.acceptButton)
  else
    onIncorrectResponse(self, self.acceptButton)
  end
end

local function onProgressTimeOut(self)
  return function()
    runInCoroutine(function()
      setButtonsEnabled(self, false)

      local shouldSwap = not self.isPromptCorrect
      self.scene:playHighlightIncorrectAnimationAndWaitForRationaleTap(shouldSwap)

      endRound(self)
    end)
  end
end

local function onProgressBarLost(self)
  return function()
    applyScorePenalty()
    updateScore(self)
  end
end

-- Round flow -------------------------

local function prepareNextRound(self, isFirstConcept)
  return runInCoroutine(function()
    local conceptData = self.roundGenerator:next()

    self.nextIsPromptCorrect = MathUtils.getRandomBoolean()
    local areWordsOrderedCorrectFirst = self.nextIsPromptCorrect
    self.scene:createNextRoundElements(conceptData.sentence, conceptData.words, areWordsOrderedCorrectFirst, isFirstConcept)

    self.nextConceptId = conceptData[SwapRoundGenerator.CONCEPT_IDENTIFIER_KEY]
    self.nextRationaleText = conceptData.rationale

    self.prepareRoundLock:stop()
  end)
end

local function showTutorial(self)
  setButtonsEnabled(self, false)
  self.progressIndicator:pauseSliding(true)

  local uiOverlay = self:getUIOverlay()

  local correctButton = nil
  local correctCallback = nil
  if (self.isPromptCorrect) then
    correctButton = self.acceptButton
    correctCallback = function()
      onAcceptPressed(self)
    end
  else
    correctButton = self.swapButton
    correctCallback = function()
      onSwapPressed(self)
    end
  end

  self.scene:showTutorialOverlayAndWaitForDismissal(uiOverlay, correctButton, correctCallback)

  self.progressIndicator:pauseSliding(false)
end

local function runRound(self)
  self.scene:setRationaleText(self.nextRationaleText)
  wait(self.scene:playRoundTransitionAnimation(self.isFirstRound))

  self.isPromptCorrect = self.nextIsPromptCorrect
  self.currentConceptId = self.nextConceptId

  setButtonsEnabled(self, true)

  if self.isFirstRound then
    local tutorialTimer = {}
    Timer.makeTimer(tutorialTimer, SwapConfiguration.TUTORIAL_DELAY)
    tutorialTimer:setTimeoutHandler(function() showTutorial(self) end)
    tutorialTimer:start()
    self.tutorialTimer = tutorialTimer
  end
end

-- Game flow --------------------------

local function initializeGame(self)
  self.scene:fadeInHintLabel()
  self.isFirstRound = true

  self.responseWaitLock = AHDelay.new(math.huge)
  self.prepareRoundLock = AHDelay.new(math.huge)
  self.prepareRoundLock:start()
  prepareNextRound(self, true)

  local function quickResponseWrapper(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 SwapConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  local maximumBonusesAllowed = self.correctNeeded
  self.responseSpeedTracker = ResponseSpeedTracker.new(SwapConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       SwapConfiguration.SPEED_BONUS_DECAY_START,
                                                       maximumBonusesAllowed)
  self.incorrectResponses = 0
end

local function endGame(self)
  self.progressIndicator:stopSliding()

  if didWin(self) then
    GameSession.notifyGameWin()
  else
    GameSession.notifyGameLoss()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function SwapGame:gameFlow()
  initializeGame(self)

  self.progressIndicator:addBar()

  repeat
    wait(self.prepareRoundLock)

    runRound(self)
    GameSession.contentTracker:notifyPresentedConcept(self.currentConceptId)

    -- Note that this is a bit imperfect because Swap pulls concepts for the next
    -- round in advance. Thus, the first resurfaced concept will appear 1 round later.
    if (self.progressIndicator:getProgress() >= LearningFocusedRoundGenerator.DEFAULT_PROGRESS_BEFORE_RESURFACING) then
      self.roundGenerator:allowMistakesToResurface()
    end

    self.prepareRoundLock:start()
    prepareNextRound(self, false)

    self.responseSpeedTracker:notifyStart()
    wait(self.responseWaitLock:start())
    self.isFirstRound = false
  until (isGameEnd(self))

  if didLose(self) then
    self.scene:playGameLoseAnimation()
    wait(self:showDefaultGameLossOutro())
  end

  endGame(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local onSwapPressedCallback = function()
    onSwapPressed(self)
  end

  local onAcceptPressedCallback = function()
    onAcceptPressed(self)
  end

  local swapButton = SwapButton.new("revert_icon.png", Localizer.get("SWAP"), onSwapPressedCallback)
  local acceptButton = SwapButton.new("tick_icon.png", Localizer.get("KEEP"), onAcceptPressedCallback)
  self.swapButton = swapButton
  self.acceptButton = acceptButton

  local numberOfSteps = self.correctNeeded + 1
  local progressIndicator = SwapProgressIndicator.new(numberOfSteps, self.slideSeconds, onProgressTimeOut(self), onProgressBarLost(self))
  self.progressIndicator = progressIndicator

  local scene = SwapScene.new(self.swapButton, self.acceptButton, self.progressIndicator)
  self:insertAtCenter(scene)
  self.scene = scene

  setButtonsEnabled(self, false)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  if self.scene then
    self.scene:setGetUiYBottomCallback(function()
      return self.scoreLabel:getYBottom()
    end)
  end
end

-- Constructor ------------------------

function SwapGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SwapGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.roundGenerator = LearningFocusedRoundGenerator.new(SwapRoundGenerator.new(), SwapRoundGenerator.CONCEPT_IDENTIFIER_KEY)

  return self
end
