--- SwapProgressIndicator
-- @author created by jay austria on febtober 11, 2014
class("SwapProgressIndicator")
require("widgets.progress.ProgressIndicator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BAR_WIDTH = Screen.fullWidth * 0.04
local GROUP_WIDTH = Screen.fullWidth
local GROUP_HEIGHT = Screen.fullHeight

local COLORS = { { 230/255, 184/255,  97/255, 1 },
                 { 230/255, 148/255,  98/255, 1 },
                 { 202/255, 147/255,  96/255, 1 },
                 { 233/255, 165/255, 100/255, 1 } }

local ADD_BAR_SCALE_DURATION = 0.1
local ADD_BAR_MOVE_DURATION = 0.15
local ADD_BAR_DURATION = ADD_BAR_SCALE_DURATION + ADD_BAR_MOVE_DURATION
local REMOVE_BAR_FALL_DURATION = 0.2

local FULL_BAR_EXTRA_CYCLES = 3
local FULL_BAR_STEP_DELAY = 0.05

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function colorIterator(colors)
  local i = 0
  return function()
    i = i % #colors + 1
    return i, colors[i]
  end
end

local function createLayout(self, numberOfSteps)
  self.leftBars = {}
  self.rightBars = {}
  self.barHeight = Screen.fullHeight / numberOfSteps

  self.chooseColor = colorIterator(COLORS)
end

-- Animations: Add bars ---------------

local function simpleBounceEase()
  local TIME_FIRST_FALL = 0.33
  local TIME_BOUNCE = 0.33
  local TIME_SECOND_FALL = 1 - TIME_FIRST_FALL - TIME_BOUNCE
  local HEIGHT_BOUNCE = 0.2

  local function easeFunction(t)
    if t < TIME_FIRST_FALL then
      t = t / TIME_FIRST_FALL
      return t * t
    elseif t < TIME_FIRST_FALL + TIME_BOUNCE then
      t = (t - TIME_FIRST_FALL) / TIME_BOUNCE
      t = 1 - t
      return t * t * HEIGHT_BOUNCE + (1 - HEIGHT_BOUNCE)
    else
      t = (t - TIME_FIRST_FALL - TIME_BOUNCE) / TIME_SECOND_FALL
      return t * t * HEIGHT_BOUNCE + (1 - HEIGHT_BOUNCE)
    end
  end

  local ease = MOAIEaseCustom.new()
  ease:setFunction(easeFunction)
  return ease
end

local function addBarAction(self, newBar, barX, barY)
  local SHOW_EFFECT_OFFSET = self.barHeight * 0.3

  self:insert(newBar)
  newBar:setLoc(barX, barY + SHOW_EFFECT_OFFSET)
  newBar:setScl(0)

  local moveAction = AHLocationYBy.new(newBar, ADD_BAR_MOVE_DURATION, simpleBounceEase(), -SHOW_EFFECT_OFFSET)
  local scaleAction = AHScaleTo.new(newBar, ADD_BAR_SCALE_DURATION, nil, 1.0)

  return AHSequence.new(scaleAction, moveAction)
end

local function addBarsAnimation(self)
  local step = self:getProgressStep()
  local colorIndex, color = self.chooseColor()

  local barX = BAR_WIDTH * 0.5
  local barY = self.barHeight * step - self.barHeight * 0.5

  local newBarLeft = FilledRectangle.new(BAR_WIDTH, self.barHeight, color)
  local addLeft = addBarAction(self, newBarLeft, barX, barY)
  newBarLeft.colorIndex = colorIndex
  table.insert(self.leftBars, newBarLeft)

  barX = Screen.fullWidth - BAR_WIDTH * 0.5
  local newBarRight = FilledRectangle.new(BAR_WIDTH, self.barHeight, color)
  local addRight = addBarAction(self, newBarRight, barX, barY)
  newBarRight.colorIndex = colorIndex
  table.insert(self.rightBars, newBarRight)

  return AHSpawn.new(addLeft, addRight)
end

-- Animations: Slide down -------------

local function removeBarAndResetLocation(self)
  self:offsetY(self.barHeight)
  for _, bars in ipairs({self.leftBars, self.rightBars}) do
    bars[1]:removeSelf()
    table.remove(bars, 1)
    for _, bar in ipairs(bars) do
      bar:offsetY(-self.barHeight)
    end
  end
end

local function barOffScreenHandler(self)
  self:stopSliding()
  self:decrementProgress()
  removeBarAndResetLocation(self)
  self.onBarLost()
  if (self:getProgressStep() == 0) then
    self.onTimeOut()
  end
end

local function getBeepsCount(slideDuration)
  local maxBeepsCount = 3
  if (slideDuration < maxBeepsCount) then
    return math.floor(slideDuration)
  end
  return maxBeepsCount
end

local function createDelayedBeepsLoopAction(self, beepsCount, totalDuration)
  local delay = AHDelay.new(totalDuration - beepsCount)
  local playAction = AHSpawn.new(
    function()
      if (self:getProgressStep() == 1) then
        Sound.playSoundEffect("shared_countdown_warning")
      end
    end,
    AHDelay.new(1.0)
  )

  return AHSequence.new(delay, AHLoop.new(playAction, beepsCount))
end

local function createTimeoutSoundAction(self, slideDuration)
  local beepsCount = getBeepsCount(slideDuration)

  if (beepsCount < 1) then
    return AHDoNothing.new()
  else
    return createDelayedBeepsLoopAction(self, beepsCount, slideDuration)
  end
end

local function startSlideDownAction(self, fractionOfTimeRemaining)
  if (fractionOfTimeRemaining == nil) then
    fractionOfTimeRemaining = 1.0
  end

  local slideDuration = self.slideSeconds * fractionOfTimeRemaining
  local slideDistance = (-self.barHeight) * fractionOfTimeRemaining
  local slide = AHLocationYBy.new(self, slideDuration, nil, slideDistance)
  local slideCompletedFunction = function()
    barOffScreenHandler(self)
    startSlideDownAction(self)
  end
  local slideAndDecrement = AHSequence.new(slide, slideCompletedFunction)
  local timeoutSound = createTimeoutSoundAction(self, slideDuration)
  self.slideLoop = AHSpawn.new(slideAndDecrement, timeoutSound)

  self.slideAction = slide
  self.slideActionDuration = slideDuration

  return self.slideLoop:start()
end

-- Animations: Remove bars ------------

local function getFractionOfCurrentBarElapsed(self)
  local timeElapsedInPreviousAnimations = self.slideSeconds - self.slideActionDuration
  local timeElapsedInCurrentAnimation = self.slideAction:getTime()

  local totalTimeElapsedForCurrentBar = timeElapsedInPreviousAnimations + timeElapsedInCurrentAnimation
  return totalTimeElapsedForCurrentBar / self.slideSeconds
end

local function removeFractionOfBarAnimation(self, barFraction)
  local fallDistance = (-self.barHeight) * barFraction
  local moveBarsAnimation = AHLocationYBy.new(self, REMOVE_BAR_FALL_DURATION, AHEaseSineIn.new(), fallDistance)

  local fallAnimation = AHSpawn.new(
    function()
      self:stopSliding()
    end,
    moveBarsAnimation
  )

  local cleanupFunctions = AHSpawn.new()

  local fractionOfCurrentBarElapsed = getFractionOfCurrentBarElapsed(self)
  local newFractionOfBarElapsed = fractionOfCurrentBarElapsed + barFraction
  local barIsOffscreen = (newFractionOfBarElapsed > 1.0)
  if barIsOffscreen then
    newFractionOfBarElapsed = newFractionOfBarElapsed - 1
    cleanupFunctions:addAction(function() barOffScreenHandler(self) end)
  end
  cleanupFunctions:addAction(function() startSlideDownAction(self, 1 - newFractionOfBarElapsed) end)

  return AHSequence.new(fallAnimation, cleanupFunctions)
end

-- Animations: Full bar ---------------

local function setColor(target, colorIndex)
  return AHCallLogic.new(function()
    target:setColor(unpack(COLORS[colorIndex]))
    target.colorIndex = colorIndex
  end)
end

local function shiftColor(target)
  return AHCallLogic.new(function()
    local nextColorIndex = (target.colorIndex - 2) % #COLORS + 1
    setColor(target, nextColorIndex):start()
  end)
end

local function fullBarAnimation(self)
  local allActions = {}

  for _, bars in ipairs({self.leftBars, self.rightBars}) do
    local thisSideActions = {}
    local firstColorIndex = bars[1].colorIndex

    for n = 2, #bars do
      local colorActions = {}
      table.insert(colorActions, setColor(bars[n], firstColorIndex))

      for i = 1, n-1 do
        table.insert(colorActions, shiftColor(bars[i]))
      end

      table.insert(thisSideActions, AHSpawn.new(unpack(colorActions)))
      table.insert(thisSideActions, AHDelay.new(FULL_BAR_STEP_DELAY))
    end

    for i = 1, #bars * FULL_BAR_EXTRA_CYCLES do
      local thisCycleActions = {}
      for _, bar in ipairs(bars) do
        table.insert(thisCycleActions, shiftColor(bar))
      end
      table.insert(thisSideActions, AHSpawn.new(unpack(thisCycleActions)))
      table.insert(thisSideActions, AHDelay.new(FULL_BAR_STEP_DELAY))
    end

    table.insert(allActions, AHSequence.new(unpack(thisSideActions)))
  end

  return AHSequence.new(AHDelay.new(ADD_BAR_DURATION), AHSpawn.new(unpack(allActions)))
end

-- Animations: Victory explosion ------

local function addVictoryExplosionSprites(self)
  local explosionSprites = {}

  local rowsCount = 7
  local sidesCount = 2
  local rowSpacing = GROUP_HEIGHT / rowsCount

  for rowIndex = 1, rowsCount do
    for sideIndex = 1, sidesCount do
      local explosionSprite = Sprite.newWithName("end_explosion_000.png")
      local halfExplosionSpriteWidth = explosionSprite:getWidth() * 0.5

      local y = rowSpacing * (0.5 + rowIndex)

      local x = nil
      local xScale = nil
      local isExplosionOnRight = (sideIndex == sidesCount)
      if (isExplosionOnRight) then
        x = GROUP_WIDTH - halfExplosionSpriteWidth
        xScale = -1.0
      else
        x = halfExplosionSpriteWidth
        xScale = 1.0
      end

      explosionSprite:setLoc(x, y)
      explosionSprite:setScl(xScale, 1.0)

      explosionSprite:setVisible(false)

      self:insert(explosionSprite)
      table.insert(explosionSprites, explosionSprite)
    end
  end

  return explosionSprites
end

local function createVictoryExplosionAction(explosionSprites)
  local explosionActions = {}

  local explosionAnimation = Animation.newWithPrefix("end_explosion_")
  local explosionFramerate = 38.0
  local explosionEase = nil

  for i, explosionSprite in ipairs(explosionSprites) do
    local explosionDelayDuration = (i - 1) * 0.04
    local explosionAction = AHSequence.new(
                                           AHDelay.new(explosionDelayDuration),
                                           function()
                                             explosionSprite:setVisible(true)
                                           end,
                                           AHAnimate.new(explosionSprite, explosionAnimation, explosionFramerate, explosionEase)
                                           )
    table.insert(explosionActions, explosionAction)
  end

  local victoryExplosionAction = AHSpawn.new(unpack(explosionActions))
  return victoryExplosionAction
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SwapProgressIndicator:addBar()
  local actions = {}

  self:incrementProgress()
  local addBars = addBarsAnimation(self)
  table.insert(actions, addBars)

  if not (self.slideLoop and self.slideLoop:isActive()) then
    startSlideDownAction(self)
  end

  local shouldShowVictoryAnimation = (self:getProgressStep() == self:getStepCount())
  if (shouldShowVictoryAnimation) then
    self:stopSliding()

    local explosionSprites = addVictoryExplosionSprites(self)
    local explosionAction = createVictoryExplosionAction(explosionSprites)

    local fullBarAction = fullBarAnimation(self)

    local victoryAction = AHSpawn.new(fullBarAction, explosionAction)

    table.insert(actions, victoryAction)
    table.insert(actions, AHCallLogic.new(self.onTimeOut))
  end

  return AHSequence.new(unpack(actions)):start()
end

function SwapProgressIndicator:removeBarFraction(barFraction)
  barFraction = barFraction or 1.0

  assert(barFraction >= 0.0, "Cannot remove a fraction of a bar that is less than 0.")
  assert(barFraction <= 1.0, "Cannot remove a fraction of a bar that is greater than 1.")

  return removeFractionOfBarAnimation(self, barFraction):start()
end

function SwapProgressIndicator:stopSliding()
  if self.slideLoop then
    self.slideLoop:stop()
  end
end

function SwapProgressIndicator:pauseSliding(isPaused)
  if self.slideLoop then
    self.slideLoop:pause(isPaused)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SwapProgressIndicator.new(numberOfSteps, slideSeconds, onTimeOut, onBarLost)
  local self = Group.new(GROUP_WIDTH, GROUP_HEIGHT)
  ProgressIndicator.makeProgressIndicator(self, numberOfSteps)
  Mixin.mix(self, SwapProgressIndicator)

  createLayout(self, numberOfSteps - 1)
  self.slideSeconds = slideSeconds
  self.onTimeOut = onTimeOut
  self.onBarLost = onBarLost

  return self
end
