--- SwapScene
-- @author created by Jay Austria on June 5, 2021

class("SwapScene")
require("SwapCommon")
require("SwapRationalePanel")
require("SwapTextManager")
require("SwapVictoryShine")
require("tutorials.TutorialHighlighter")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Colors -----------------------------

local BACKGROUND_COLOR_TOP = {0.86, 0.82, 0.72}
local BACKGROUND_COLOR_BOTTOM = {0.95, 0.89, 0.79}
local HINT_TEXT_COLOR = {0.24, 0.26, 0.24}
local BACKING_COLOR = {0.96, 0.91, 0.81}
local RATIONALE_OVERLAY_COLOR = {0.11, 0.01, 0.01}
local RATIONALE_OVERLAY_OPACITY_HIDDEN = 0.0
local RATIONALE_OVERLAY_OPACITY_VISIBLE = 0.2

-- Sizes and locations ----------------

local BUTTON_Y = Screen.getSafeY(0.15)
local BACKING_WIDTH = Screen.safeWidth * 0.75
local PROMPT_PADDING_X = 15
local PROMPT_PADDING_Y = 13
local SENTENCE_BOX_PADDING_Y = 25

local SENTENCE_BOX_X = Screen.getSafeX(0.5)
local SENTENCE_BOX_Y = Screen.getSafeY(0.6)

-- Animation params -------------------

local DURATION_SENTENCE_SWIPE = 0.7
local DURATION_INTO_RATIONALE = 0.5
local DURATION_OUT_OF_RATIONALE = 0.33
local EASE_INTO_RATIONALE = AHEaseBezier.newFlat({0.5, 0.4, 0.0, 1.0})
local EASE_OUT_OF_RATIONALE = AHEaseBezier.newFlat({0.6, 0.0, 0.4, 1.0})

-- Z-Indices --------------------------

local Z_INDICES = {
  BACKGROUND = 0,
  BACKING = 10,
  BUTTON = 20,
  TUTORIAL_UI = 30,
  TUTORIAL = 40,
  RATIONALE_OVERLAY = 50,
  SENTENCES = 60,
  PROGRESS_INDICATOR = 70,
  VICTORY_SHINE = 80,
  RATIONALE_PANEL = 90
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setProgressIndicatorZIndexUnderRationaleOverlay(self)
  local desiredZIndex = Z_INDICES.RATIONALE_OVERLAY - 1
  self.progressIndicator:setZIndex(desiredZIndex)
end

local function restoreProgressIndicatorZIndex(self)
  self.progressIndicator:setZIndex(Z_INDICES.PROGRESS_INDICATOR)
end

local function fadeHintLabelOpacity(self, duration, targetOpacity)
  local fadeAction = AHOpacityTo.new(self.hintLabel, duration, AHEaseLinear.new(), targetOpacity)
  fadeAction:start()
end

local function fadeOutHintLabel(self)
  local duration = 0.033
  local targetOpacity = 0.0
  fadeHintLabelOpacity(self, duration, targetOpacity)
end

local function invertHintLabelColor(self)
  local duration = 0.167
  local targetColor = {1, 1, 1, 1}

  local invertColorSpawn = AHSpawn.new()
  for _, itemView in ipairs(self.hintLabel.itemViews) do
    invertColorSpawn:addAction(AHColorTo.new(itemView, duration, AHEaseLinear.new(), targetColor))
  end

  invertColorSpawn:start()
end

local function fadeRationaleOverlay(self, targetOpacity, duration, ease)
  local fadeAction = AHOpacityTo.new(self.rationaleOverlay, duration, ease, targetOpacity)

  return fadeAction:start()
end

local function slideSentenceBox(self, destinationY, duration, ease)
  local slideAction = AHLocationYTo.new(self.sentenceBox, duration, ease, destinationY)

  return slideAction:start()
end

local function getBackingHeight(labelHeight)
  return (labelHeight + (2 * PROMPT_PADDING_Y))
end

local function createBacking(sentenceLabel, textManager)
  local backing = FilledRoundedRectangle.new(BACKING_WIDTH,
                                             getBackingHeight(sentenceLabel:getHeight()),
                                             SwapCommon.STANDARD_CORNER_RADIUS,
                                             BACKING_COLOR)

  textManager:setWordsSwappedCallback(function(labelHeight)
    local duration = 0.5
    local ease = AHEaseSineInOut.new()
    AHDimYTo.new(backing, duration, ease, getBackingHeight(labelHeight)):start()
  end)

  return backing
end

local function getMaxHeightForSentenceLabel(self, isFirstConcept)
  local uiYBottom = 0
  if self.getUiYBottomCallback then
    uiYBottom = self.getUiYBottomCallback()
  end

  local rationaleHeight = self.rationalePanel:getHeight()

  local maxTextHeight = uiYBottom - rationaleHeight - (2 * SENTENCE_BOX_PADDING_Y)

  return maxTextHeight
end

local function getYLocationForSentenceBoxShifted(self, sentenceBox)
  local rationaleYTop = self.rationalePanel:getHeight()

  if self.getUiYBottomCallback then
    uiYBottom = self.getUiYBottomCallback()
  end

  local rationaleToUiHeight = uiYBottom - rationaleYTop

  local shiftedY = rationaleYTop + (rationaleToUiHeight * 0.5)
  return shiftedY
end

local function getYLocationForSentenceBox(self, sentenceBox, backing, isFirstConcept)
  local yLocation = SENTENCE_BOX_Y
  if isFirstConcept then
    local boxBottom = SENTENCE_BOX_Y - backing:getHeightPart(0.5)
    local hintTop = self.hintLabel:getYTop()

    local padding = 20
    if (boxBottom < (hintTop + padding)) then
      yLocation = yLocation + (hintTop - boxBottom) + padding
    end
  end
  return yLocation
end

local function createSentenceBox(self, textManager, isFirstConcept)
  local maxTextHeight = getMaxHeightForSentenceLabel(self)
  local textWidth = BACKING_WIDTH - (2 * PROMPT_PADDING_X)
  local sentenceLabel = textManager:createSentenceLabel(textWidth, maxTextHeight)

  local sentenceBox = Group.newMinimumSized()
  sentenceBox:insert(sentenceLabel, Z_INDICES.SENTENCES)
  local backing = createBacking(sentenceLabel, textManager)
  sentenceBox:insert(backing, Z_INDICES.BACKING)

  local adjustedY = getYLocationForSentenceBox(self, sentenceBox, backing, isFirstConcept)
  sentenceBox:setLoc(SENTENCE_BOX_X, adjustedY)

  sentenceBox.sentenceLabel = sentenceLabel
  sentenceBox.backing = backing

  return sentenceBox
end

local function playSentenceBoxTransitionAnimation(self)
  local actions = {}

  if self.sentenceBox then
    local oldSentenceBox = self.sentenceBox
    local swipeOutAction = AHLocationXBy.new(oldSentenceBox, DURATION_SENTENCE_SWIPE, AHEaseSimpleInOut.new(3.5), -Screen.fullWidth)
    local function removeOldSentenceBox()
      oldSentenceBox:removeSelf()
    end
    table.insert(actions, AHSequence.new(swipeOutAction, removeOldSentenceBox))
  end

  if self.nextSentenceBox then
    local newSentenceBox = self.nextSentenceBox
    local swipeInAction = AHLocationXBy.new(newSentenceBox, DURATION_SENTENCE_SWIPE, AHEaseSimpleInOut.new(3.5), -Screen.fullWidth)
    table.insert(actions, swipeInAction)

    local x, y = newSentenceBox:getLoc()
    newSentenceBox:setLoc(x + Screen.fullWidth, y)
    self:insert(newSentenceBox, Z_INDICES.SENTENCES)
  end

  self.sentenceBox = self.nextSentenceBox
  self.textManager = self.nextTextManager

  return AHSpawn.new(unpack(actions)):start()
end

local function createAndAnimateVictoryShine(self)
  local victoryShine = SwapVictoryShine.new()
  self:insertAtCenter(victoryShine, Z_INDICES.VICTORY_SHINE)
  victoryShine:shine()
end

local function fadeOutButtonsAndText(self)
  local swapButton = self.swapButton
  local acceptButton = self.acceptButton

  local swapAction = swapButton:createFadeOutAction()
  local acceptAction = acceptButton:createFadeOutAction()

  local fadeOutAction = AHSpawn.new(
                                    swapAction,
                                    acceptAction
                                    )

  local sentenceBox = self.sentenceBox
  if (sentenceBox ~= nil) then
    local sentenceLabel = sentenceBox.sentenceLabel
    if (sentenceLabel ~= nil) then
      local isRecursive = true
      sentenceLabel:linkOpacityAttrToChildren(isRecursive)

      local sentenceLabelAction = AHOpacityTo.new(sentenceLabel, SwapCommon.FADE_OUT_DURATION, SwapCommon.FADE_OUT_EASE, 0.0)
      fadeOutAction:addAction(sentenceLabelAction)
    end
  end

  local fadeAndHideAction = AHSequence.new(
                                           fadeOutAction,
                                           function()
                                             -- To prevent a bug where one or both buttons intermittently reappears immediately
                                             -- following the fade sequence, set their visibility to false.
                                             swapButton:setVisibleRecursive(false)
                                             acceptButton:setVisibleRecursive(false)
                                           end
                                           )

  fadeAndHideAction:start()
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SwapScene:setGetUiYBottomCallback(callback)
  self.getUiYBottomCallback = callback
end

function SwapScene:fadeInHintLabel()
  local duration = 0.167
  local targetOpacity = 1.0
  fadeHintLabelOpacity(self, duration, targetOpacity)
end

function SwapScene:playRoundTransitionAnimation(isFirstRound)
  if (not isFirstRound) then
    fadeOutHintLabel(self)
  end

  return playSentenceBoxTransitionAnimation(self)
end

function SwapScene:setRationaleText(text)
  self.rationalePanel:setText(text)
end

function SwapScene:playHighlightIncorrectAnimationAndWaitForRationaleTap(shouldSwap)
  local sentenceBoxYOriginal = self.sentenceBox:getY()
  local sentenceBoxYShifted = getYLocationForSentenceBoxShifted(self)

  wait(self.textManager:playHighlightIncorrectAnimation())
  waitFor(0.170)
  setProgressIndicatorZIndexUnderRationaleOverlay(self)
  fadeRationaleOverlay(self, RATIONALE_OVERLAY_OPACITY_VISIBLE, DURATION_INTO_RATIONALE, EASE_INTO_RATIONALE)
  slideSentenceBox(self, sentenceBoxYShifted, DURATION_INTO_RATIONALE, EASE_INTO_RATIONALE)
  wait(self.rationalePanel:playSlideOnscreenAnimation(DURATION_INTO_RATIONALE, EASE_INTO_RATIONALE))

  local tapToContinue = TapToContinue.newInGroup(self.rationalePanel)
  tapToContinue:waitForTrigger()
  tapToContinue:removeSelf()

  fadeRationaleOverlay(self, RATIONALE_OVERLAY_OPACITY_HIDDEN, DURATION_OUT_OF_RATIONALE, EASE_OUT_OF_RATIONALE)
  slideSentenceBox(self, sentenceBoxYOriginal, DURATION_OUT_OF_RATIONALE, EASE_OUT_OF_RATIONALE)
  wait(self.rationalePanel:playSlideOffscreenAnimation(DURATION_OUT_OF_RATIONALE, EASE_OUT_OF_RATIONALE))


  if shouldSwap then
    waitFor(0.170)
    wait(self.textManager:playIncorrectAnimationSwapWords())
  end

  waitFor(0.5)

  restoreProgressIndicatorZIndex(self)
end

function SwapScene:playCorrectAnimation(shouldSwap)
  return self.textManager:playCorrectAnimation(shouldSwap)
end

function SwapScene:createNextRoundElements(sentence, words, areWordsOrderedCorrectFirst, isFirstConcept)
  local wordsInPresentationOrder = TableUtils.deepCopy(words)

  if (not areWordsOrderedCorrectFirst) then
    TableUtils.swap(wordsInPresentationOrder)
  end

  local textManager = SwapTextManager.new(sentence, wordsInPresentationOrder)
  local sentenceBox = createSentenceBox(self, textManager, isFirstConcept)

  self.nextSentenceBox = sentenceBox
  self.nextTextManager = textManager
end

function SwapScene:showTutorialOverlayAndWaitForDismissal(uiOverlay, correctButton, correctCallback)
  local oldUIZIndex = uiOverlay.zIndex
  uiOverlay:setZIndex(Z_INDICES.TUTORIAL_UI)

  local generateHole = function()
    local hole = FilledRoundedRectangle.new(correctButton:getWidth(), correctButton:getHeight(), SwapCommon.STANDARD_CORNER_RADIUS)
    local x, y = self:transformLoc(correctButton)
    x = x + hole:getWidthPart(0.5)
    y = y + hole:getHeightPart(0.5)
    hole:setLoc(x, y)
    return hole
  end

  local tutorialOverlay = TutorialHighlighter.newWithMasks({generateHole})

  self:insertAtCenter(tutorialOverlay, Z_INDICES.TUTORIAL)
  tutorialOverlay:createFadeInAnimation():start()
  invertHintLabelColor(self)
  tutorialOverlay:waitForTargetTap()

  if correctCallback then
    correctCallback()
  end

  wait(tutorialOverlay:createFadeOutAnimation():start())
  tutorialOverlay:removeSelf()
  uiOverlay:setZIndex(oldUIZIndex)
end

function SwapScene:playGameWinAnimation()
  createAndAnimateVictoryShine(self)
  return fadeOutButtonsAndText(self)
end

function SwapScene:playGameLoseAnimation()
  return fadeOutButtonsAndText(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, swapButton, acceptButton, progressIndicator)
  local background = RectangleGradient.new(Screen.fullWidth, Screen.fullHeight, BACKGROUND_COLOR_TOP, BACKGROUND_COLOR_BOTTOM)
  background:setLoc(Screen.fullWidth * 0.5, Screen.fullHeight * 0.5)
  self:insert(background, Z_INDICES.BACKGROUND)
  self.background = background

  local buttonsRow = Row.new(Screen.getSafeWidthPart(0.1))
  buttonsRow:insert(swapButton)
  buttonsRow:insert(acceptButton)
  buttonsRow:alignChildren()
  buttonsRow:setLoc(Screen.getSafeX(0.5), BUTTON_Y)
  self:insert(buttonsRow, Z_INDICES.BUTTON)

  local hintFont = Font.new(FontFaces.DIN_OT_BOLD, 16)
  local hintAttributes = {
    font = hintFont,
    color = HINT_TEXT_COLOR
  }
  local hintLabel = DynamicTextArea.new(Localizer.get("HINT"), Screen.safeWidth, hintAttributes, nil, Alignment.CENTER, Alignment.CENTER)
  hintLabel:setLoc(Screen.getSafeWidthPart(0.5), BUTTON_Y + swapButton:getHeight())
  hintLabel:setOpacity(0)
  self:insert(hintLabel, Z_INDICES.SENTENCES)
  self.hintLabel = hintLabel

  self:insertAtCenter(progressIndicator, Z_INDICES.PROGRESS_INDICATOR)

  local rationalePanel = SwapRationalePanel.new()
  local onscreenY = 0
  local offscreenY = -rationalePanel:getHeight()
  rationalePanel:setScreenYLocations(onscreenY, offscreenY)
  self:insertAtLocation(rationalePanel, self:getWidthPart(0.5), offscreenY, Z_INDICES.RATIONALE_PANEL)
  self.rationalePanel = rationalePanel

  local rationaleOverlay = FilledRectangle.newFullScreen(RATIONALE_OVERLAY_COLOR)
  rationaleOverlay:setOpacity(RATIONALE_OVERLAY_OPACITY_HIDDEN)
  self:insertAtCenter(rationaleOverlay, Z_INDICES.RATIONALE_OVERLAY)
  self.rationaleOverlay = rationaleOverlay
end

function SwapScene.new(swapButton, acceptButton, progressIndicator)
  local self = Group.newFullScreen()
  Mixin.mix(self, SwapScene)

  createLayout(self, swapButton, acceptButton, progressIndicator)

  self.swapButton = swapButton
  self.acceptButton = acceptButton
  self.progressIndicator = progressIndicator

  return self
end
