--- SwapTextManager
-- @author Created by Anton Morant on February 13, 2014.

class("SwapTextManager")

require("deprecated.CompositeLabel")
require("deprecated.CompositeLabelRenderer")
require("GapFontRenderEngine")
require("Particles")
require("SwapCommon")
require("text.SizeConstrainedTextArea")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TEXT_COLOR = {  61/255,  66/255,  62/255 }
local WORD_BG_COLOR = { 230/255, 184/255,  97/255 }
local WORD_BG_CORRECT_COLOR = SwapCommon.CORRECT_COLOR
local WORD_BG_INCORRECT_COLOR = SwapCommon.INCORRECT_COLOR
local WORD_BG_NEUTRAL_COLOR = { 212/255, 198/255, 171/255 }

local DEFAULT_FONT = FontFaces.DIN_OT_LIGHT
local WORD_FONT = FontFaces.DIN_OT_MEDIUM
local ITALIC_FONT = FontFaces.DIN_OT_LIGHT_ITALIC
local INITIAL_FONT_SIZE = 25
local MINIMUM_FONT_SIZE = 15
local LINE_SPACING = 5

local BACKGROUND_PADDING_X = 3
local BACKGROUND_PADDING_Y = 1

local Z_PARTICLES = 10

local DURATION_TOUCH = 0.2
local DURATION_SWAP = 0.2
local DURATION_PARTICLES_DELAY = 0.3 * DURATION_SWAP
local DURATION_PARTICLES = DURATION_PARTICLES_DELAY
local DURATION_MISSED_DELAY = 0.5

local PARTICLES_FILENAME = "green_particles.pex"

local GAP_PLACEHOLDER_CHAR = "_"
local GAP_PLACEHOLDER_PUNCTUATION_PATTERN = "(%p*)_(%p*)"

local CAPITALIZED_WORDS_WHITELIST = {
  "I"
}

local WORD_ATTRIBUTE = "word"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Misc helpers -----------------------

local function getSpaceCount(text)
  local _, spaceCount = text:gsub(" ", "")
  return spaceCount
end

local function getBackgroundWidth(label)
  return label:getWidth() + 2 * BACKGROUND_PADDING_X
end

local function getBackgroundHeight(label)
  return label:getHeight() + 2 * BACKGROUND_PADDING_Y
end

local function createBackground(label)
  local background = FilledRoundedRectangle.new(getBackgroundWidth(label),
                                                getBackgroundHeight(label),
                                                SwapCommon.STANDARD_CORNER_RADIUS,
                                                WORD_BG_COLOR)
  return background
end

local function alignToCenterOfParent(group)
  local parent = group.parent
  local width, height = parent:getDims()
  group:setLoc(width * 0.5, height * 0.5)
end

local function updateSwapWordLabels(self)
  for i, wordLabel in ipairs(self.swapWordLabels) do
    local background = self.swapWordBackgrounds[i]
    background:setWidth(getBackgroundWidth(wordLabel))
    alignToCenterOfParent(background)

    background.label:setText(wordLabel:getText())
    alignToCenterOfParent(background.label)
  end
end

local function capitalizeText(text)
  return text:sub(1,1):upper() .. text:sub(2)
end

local function addPunctuation(self, wordIndex, word)
  local punctuationBeforeWord = self.punctuation[wordIndex][1]
  local punctuationAfterWord = self.punctuation[wordIndex][2]
  return punctuationBeforeWord .. word .. punctuationAfterWord
end

local function capitalizeWordIfNeeded(self, wordIndex, word)
  if ((wordIndex == 1) and self.shouldCapitalizeFirstWord) then
    return capitalizeText(word)
  end
  return word
end

local function getWordForIndex(self, word, wordIndex)
  word = capitalizeWordIfNeeded(self, wordIndex, word)
  return addPunctuation(self, wordIndex, word)
end

local function getWord(self, wordIndex)
  local word = self.words[wordIndex]
  return getWordForIndex(self, word, wordIndex)
end

local function swapTextInLabels(self, swapWordLabels)
  for i, word in ipairs(self.words) do
    local newIndexForWord = (i % 2) + 1
    local wordForIndex = getWordForIndex(self, word, newIndexForWord)
    swapWordLabels[i]:setText(wordForIndex)
  end
end

local function getSwappedWordIndex(words, wordIndex)
  return #words - wordIndex + 1
end

local function splitSentence(sentence)
  local fragments = StringUtils.split(sentence, GAP_PLACEHOLDER_CHAR)

  if StringUtils.startsWith(sentence, GAP_PLACEHOLDER_CHAR) then
    table.insert(fragments, 1, "")
  end

  return fragments
end

local function extractPunctuation(sentence)
  local punctuation = {}

  for leadingPunctuation, trailingPunctuation in sentence:gmatch(GAP_PLACEHOLDER_PUNCTUATION_PATTERN) do
    table.insert(punctuation, {leadingPunctuation, trailingPunctuation})
  end

  for i = 1, 2 do
    for j = 1, 2 do
      assert(punctuation[i][j] ~= nil, "extractPunctuation failed for sentence:\n" .. tostring(sentence))
    end
  end

  return punctuation
end

local function removePunctuation(sentence)
  return sentence:gsub(GAP_PLACEHOLDER_PUNCTUATION_PATTERN, GAP_PLACEHOLDER_CHAR)
end

local function swapWords(self, textModel, swapWordLabels)
  swapTextInLabels(self, swapWordLabels)
  textModel:beginTransaction()
  for i, word in ipairs(self.words) do
    local swappedWordIndex = getSwappedWordIndex(self.words, i)
    local wordCount = getSpaceCount(word) + 1
    textModel:moveItems(self.wordIndices[i], self.wordIndices[swappedWordIndex], wordCount)
  end
  textModel:endTransaction()
end

-- Animation helpers ------------------

local function stretchAction(target, duration)
  return AHSequence.new(
    AHScaleTo.new(target, 0.5 * duration, nil, 0.9, 0.95),
    AHScaleTo.new(target, 0.3 * duration, AHEaseSineOut.new(), 1.05, 0.93),
    AHScaleTo.new(target, 0.2 * duration, AHEaseSineOut.new(), 1, 1)
  )
end

local function colorBackgroundsAction(self, color)
  return AHCallLogic.new(function()
    for _, background in ipairs(self.swapWordBackgrounds) do
      background:setColor(unpack(color))
    end
  end)
end

local function swapAnimationAction(self, color)
  local actions = {}
  local duration = DURATION_SWAP

  local swapAction = AHCallLogic.new(function()
    swapWords(self, self.sentenceModel, self.swapWordLabels)
    updateSwapWordLabels(self)
    self.wordsSwappedCallback(self.swappedLabelForReference:getHeight())
  end)
  table.insert(actions, swapAction)

  for _, label in ipairs(self.swapWordLabels) do
    table.insert(actions, stretchAction(label, duration))
  end

  table.insert(actions, colorBackgroundsAction(self, color))

  return AHSpawn.new(unpack(actions))
end

local function smallStretchAction(target, duration)
  return AHSequence.new(
    AHScaleTo.new(target, 0.3 * duration, AHEaseSineOut.new(), 1.05, 1.05),
    AHScaleTo.new(target, 0.7 * duration, AHEaseSineInOut.new(), 1, 1)
  )
end

local function touchAnimationAction(self)
  local duration = DURATION_TOUCH
  local actions = {}
  for _, label in ipairs(self.swapWordLabels) do
    table.insert(actions, smallStretchAction(label, duration))
  end
  return AHSpawn.new(unpack(actions))
end

local function swapAnimation(self, color)
  local swap = swapAnimationAction(self, color)
  return swap
end

local function colorAnimation(self, color)
  local colorBackgrounds = colorBackgroundsAction(self, color)
  local touch = touchAnimationAction(self)
  return AHSequence.new(colorBackgrounds, touch)
end

-- Particles --------------------------

local function particlesAction(self)
  local actions = {}
  for _, label in ipairs(self.swapWordLabels) do
    local labelX, labelY = label:getLoc()
    local labelWidth, labelHeight = label:getDims()
    labelX = labelX + labelWidth / 2
    labelY = labelY + labelHeight / 2
    local compositeLabel = label.parent
    local xPositions = {labelX - labelWidth / 4, labelX + labelWidth / 4}

    for _, x in ipairs(xPositions) do
      local particles = Particles.new(PARTICLES_FILENAME)
      particles:setLoc(x, labelY)
      compositeLabel:insert(particles, Z_PARTICLES)

      table.insert(actions, AHCallLogic.new(function()
        waitFor(DURATION_PARTICLES_DELAY)
        particles:start()
        particles:surge(25)
        waitFor(3.0)
        particles:stop()
        particles:removeSelf()
      end))
    end
  end
  return AHSpawn.new(AHDelay.new(DURATION_PARTICLES), unpack(actions))
end

local function buildSentenceText(self)
  for i,_ in ipairs(self.words) do
    local sentenceIndex = 2*i --Insert words at even indices
    local word = getWord(self, i)
    table.insert(self.sentenceFragments, sentenceIndex, "<w>" .. word .. "</w>")
  end
  return table.concat(self.sentenceFragments)
end

local function createSizeConstrainedTextAreaForLabel(text, width, height, maxFontSize)
  local baseAttributes = {
    font = Font.new(DEFAULT_FONT, maxFontSize),
    color = TEXT_COLOR,
    fitInLine = true
  }

  local italicAttributes = {
    font = Font.new(ITALIC_FONT, maxFontSize)
  }

  local wordAttributes = {
    font = Font.new(WORD_FONT, maxFontSize),
    [WORD_ATTRIBUTE] = true
  }

  local labelAttributes = {
    i = italicAttributes,
    w = wordAttributes
  }

  return SizeConstrainedTextArea.new(text, width, height, baseAttributes, labelAttributes, MINIMUM_FONT_SIZE, nil, nil, LINE_SPACING)
end

local function createSwappedLabelForReference(self, sentenceText, fontSize, width)
  local labelCopy = createSizeConstrainedTextAreaForLabel(sentenceText, width, nil, fontSize)
  local swapWordLabelsCopy = labelCopy:getViewsWithAttribute(WORD_ATTRIBUTE)
  local textModel = labelCopy:getModel()
  swapWords(self, textModel, swapWordLabelsCopy)
  return labelCopy
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SwapTextManager:createSentenceLabel(width, maxHeight)
  local sentenceText = buildSentenceText(self)

  local sentenceLabel = createSizeConstrainedTextAreaForLabel(sentenceText, width, maxHeight, INITIAL_FONT_SIZE)
  local fontSize = sentenceLabel:getCurrentFontSize()

  self.swapWordLabels = sentenceLabel:getViewsWithAttribute(WORD_ATTRIBUTE)
  for _, label in ipairs(self.swapWordLabels) do
    local background = createBackground(label)
    table.insert(self.swapWordBackgrounds, background)
    label:insertAtCenter(background)

    local labelCopy = Label.new(label:getText(), Font.new(WORD_FONT, fontSize), label:getWidth(), Alignment.CENTER)
    labelCopy:setColor(unpack(TEXT_COLOR))
    label:insertAtCenter(labelCopy)

    background.label = labelCopy
  end

  self.swappedLabelForReference = createSwappedLabelForReference(self, sentenceText, fontSize, width)
  self.sentenceModel = sentenceLabel:getModel()
  return sentenceLabel
end

function SwapTextManager:playCorrectAnimation(shouldSwap)
  local particles = particlesAction(self)

  if (shouldSwap) then
    return AHSpawn.new(swapAnimation(self, WORD_BG_CORRECT_COLOR), particles):start()
  else
    return AHSpawn.new(colorAnimation(self, WORD_BG_CORRECT_COLOR), particles):start()
  end
end

function SwapTextManager:playHighlightIncorrectAnimation()
  return colorAnimation(self, WORD_BG_INCORRECT_COLOR):start()
end

function SwapTextManager:playIncorrectAnimationSwapWords()
  return swapAnimation(self, WORD_BG_INCORRECT_COLOR, WORD_BG_NEUTRAL_COLOR):start()
end

function SwapTextManager:setWordsSwappedCallback(callback)
  self.wordsSwappedCallback = callback
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SwapTextManager.new(sentence, words)
  local self = {}
  Mixin.mix(self, SwapTextManager)

  self.punctuation = extractPunctuation(sentence)
  sentence = removePunctuation(sentence)

  self.sentenceFragments = splitSentence(sentence)
  self.words = words
  self.shouldCapitalizeFirstWord = (#(self.sentenceFragments[1]) == 0)

  self.swapWordLabels = {}
  self.swapWordBackgrounds = {}

  self.wordIndices = {}
  local wordIndex = 1
  for i, word in ipairs(self.words) do
    local space = " "
    local nonBreakingSpace = " "
    local processedWord = word:gsub(space, nonBreakingSpace)
    if (not TableUtils.contains(CAPITALIZED_WORDS_WHITELIST, word)) then
      processedWord = processedWord:lower()
    end

    self.words[i] = processedWord

    wordIndex = wordIndex + getSpaceCount(self.sentenceFragments[i])
    table.insert(self.wordIndices, wordIndex)
    wordIndex = wordIndex + getSpaceCount(processedWord)
  end

  self.wordsSwappedCallback = function(labelHeight) end

  return self
end
