--- SwapValidator
-- @author Created by Anton Morant on April 17, 2014.
-- ConceptValidator for Swap.

class("SwapValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'sentence',
  'wordOne',
  'wordTwo',
  'rationale'
}

local BLANK_CHAR = "_"
local MAX_CHARACTER_COUNT = 135

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateSentenceLength(self, concept)
  local sentence = concept.sentence
  local wordOne = concept.wordOne
  local wordTwo = concept.wordTwo

  local replacementCountLimit = 1
  local modifiedSentence = StringUtils.getAfterStrippingTags(sentence)

  -- Assumes wordOne goes into the first blank and wordTwo goes into the second. This is fine for our purposes of string length validation
  modifiedSentence = string.gsub(modifiedSentence, BLANK_CHAR, wordOne, replacementCountLimit)
  modifiedSentence = string.gsub(modifiedSentence, BLANK_CHAR, wordTwo, replacementCountLimit)

  local sentenceLength = #modifiedSentence
  local formattedErrorString = string.format("Sentence contained %d characters but should contain %d or fewer", sentenceLength, MAX_CHARACTER_COUNT)
  self:expectLessOrEqual(sentenceLength, MAX_CHARACTER_COUNT, formattedErrorString)
end

local function validateSentence(self, sentence)
  local numBlanks = 0
  local numSeparators = 0

  local blankMatchString = "[" .. BLANK_CHAR .. "]"

  for char in sentence:gmatch(blankMatchString) do
    if (char == BLANK_CHAR) then
      numBlanks = numBlanks + 1
    end
  end

  self:expectTrue(numBlanks == 2, "The sentence should have two blanks, it has " .. numBlanks)

  self:testForEndingPunctuation(sentence)
  self:expectStringIsPrintable(sentence, "sentence")
  self:expectNoDoubleSpaces(sentence, "sentence")
end

local function validateWords(self, wordOne, wordTwo)
  self:expectStringIsPrintable(wordOne, "wordOne")
  self:expectNoDoubleSpaces(wordOne, "wordOne")

  self:expectStringIsPrintable(wordTwo, "wordTwo")
  self:expectNoDoubleSpaces(wordTwo, "wordTwo")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SwapValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  validateSentence(self, concept.sentence)
  validateWords(self, concept.wordOne, concept.wordTwo)
  validateSentenceLength(self, concept)

  self:validateCharactersAdjacentToBlanks(concept.sentence, BLANK_CHAR)

  local rationale = concept.rationale
  self:expectNotEmpty(rationale, "rationale", "rationale is empty")
  self:validateTags(rationale)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SwapValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, SwapValidator)

  TagValidator.make(self)

  return self
end
