--- SynonymsAnswerMode
-- @author Created by andrewc on May 06, 2021.

class("SynonymsAnswerMode")

require("SynonymsInputField")
require("SynonymsPromptLabel")
require("SynonymsSuggestionLabel")
require("widgets.TapToContinue")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local AUTOCORRECT_STRING = Localizer.get("Misc.AUTOCORRECT")
local AUTOCORRECT_LABEL_FONT = Font.new(FontFaces.DIN_OT_BOLD, 15)
local AUTOCORRECT_LABEL_COLOR = SynonymsCommon.COLORS.BLACK
local AUTOCORRECT_LABEL_HEIGHT = Screen.getHeightPart(0.95)

local AUTOCORRECT_LABEL_OPACITY_VISIBLE = 0.6
local AUTOCORRECT_LABEL_OPACITY_INVISIBLE = 0
local AUTOCORRECT_LABEL_FADE_IN_DURATION = 0.1
local AUTOCORRECT_LABEL_FADE_DELAY_DURATION = 1.0
local AUTOCORRECT_LABEL_FADE_OUT_DURATION = 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function animateAutocorrectLabel(self)
  local autocorrectLabel = Label.new(AUTOCORRECT_STRING, AUTOCORRECT_LABEL_FONT)
  autocorrectLabel:setColor(unpack(AUTOCORRECT_LABEL_COLOR))
  autocorrectLabel:setOpacity(AUTOCORRECT_LABEL_OPACITY_INVISIBLE)

  self:insertAtCenter(autocorrectLabel)
  autocorrectLabel:setY(AUTOCORRECT_LABEL_HEIGHT)

  local fadeIn = AHOpacityTo.new(autocorrectLabel, AUTOCORRECT_LABEL_FADE_IN_DURATION, nil, AUTOCORRECT_LABEL_OPACITY_VISIBLE)
  local fadeDelay = AHDelay.new(AUTOCORRECT_LABEL_FADE_DELAY_DURATION)
  local fadeOut = AHOpacityTo.new(autocorrectLabel, AUTOCORRECT_LABEL_FADE_OUT_DURATION, nil, AUTOCORRECT_LABEL_OPACITY_INVISIBLE)
  local removeSelf = AHRemove.new(autocorrectLabel)

  local labelAnimation = AHSequence.new(fadeIn, fadeDelay, fadeOut, removeSelf)
  labelAnimation:start()
end

local function createElementColumnSlowScale(self)
  local duration = 3.033
  local ease = AHEaseLinear.new()
  local scale = 0.95
  local scaleAnimation = AHScaleTo.new(self.elementColumn, duration, ease, scale)
  return scaleAnimation
end

local function allowSkipPastAnimation(skipPastAnimationFunction)
  local tapToContinue = TapToContinue.new(nil, nil, skipPastAnimationFunction)
  DeviceKeyboard.setReturnCallback(skipPastAnimationFunction)
end

local function animateIncorrectAnswer(self)
  local scaleAnimation = createElementColumnSlowScale(self)
  local animationComplete = false
  allowSkipPastAnimation(function() animationComplete = true end)

  local incorrectAnimation = runInCoroutine(function()
    wait(
      scaleAnimation:start(),
      self.inputField:animateIncorrect(),
      self.promptLabel:animateIncorrect(),
      self.suggestionLabel:animateIncorrect()
    )
    animationComplete = true
  end)
  waitUntil(function() return animationComplete end)
  incorrectAnimation:stop()

  wait(
    self.inputField:animateExit(),
    self.promptLabel:animateExit(),
    self.suggestionLabel:animateExit()
  )
end

local function animateSkipAnswer(self)
  local fadeDuration = 0.167
  local fadeEase = AHEaseLinear.new()
  local fadeBackground = AHSpawn.new(
    AHOpacityTo.new(self.background, fadeDuration, fadeEase, 1),
    self.inputField:newColorTo(fadeDuration, fadeEase, SynonymsCommon.COLORS.WHITE),
    self.promptLabel:newColorTo(fadeDuration, fadeEase, SynonymsCommon.COLORS.WHITE)
  )

  local scaleAnimation = AHDelayed.new(fadeDuration, createElementColumnSlowScale(self))

  local animationComplete = false
  allowSkipPastAnimation(function() animationComplete = true end)

  local incorrectAnimation = runInCoroutine(function()
    wait(
      fadeBackground:start(),
      scaleAnimation:start(),
      self.suggestionLabel:animateSkip(SynonymsCommon.COLORS.WHITE)
    )
    animationComplete = true
  end)
  waitUntil(function() return animationComplete end)
  incorrectAnimation:stop()

  wait(
    self.inputField:animateExit(),
    self.promptLabel:animateExit(),
    self.suggestionLabel:animateExit()
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsAnswerMode:setKeyboardEnabled(enabled)
  self.isKeyboardEnabled = enabled
end

function SynonymsAnswerMode:animateIn()
  -- Do nothing
end

function SynonymsAnswerMode:animateAnswer(roundResponse)
  self.isKeyboardEnabled = false
  local isCorrect = roundResponse:getIsCorrect()
  return runInCoroutine(function()
    self.suggestionLabel:updateLabelTextForAnswer(self.mode, roundResponse:getAnswer(), isCorrect)
    self.inputField:trimWhitespace()
    if isCorrect then
      wait(self.inputField:animateCorrect(), self.promptLabel:animateCorrect())
    else
      animateIncorrectAnswer(self)
    end
    self.inputField:removeSelf(true)
  end)
end

function SynonymsAnswerMode:animateSkip()
  self.isKeyboardEnabled = false
  return runInCoroutine(function()
    self.suggestionLabel:updateLabelTextForAnswer(self.mode)
    animateSkipAnswer(self)
  end)
end

function SynonymsAnswerMode:animateAutocorrect(answer, answerLabel)
  answerLabel = answerLabel or self.inputField
  answerLabel:setText(answer)

  animateAutocorrectLabel(self)
end

function SynonymsAnswerMode:setTouchEnabled(enabled)
  self.isKeyboardEnabled = enabled
end

function SynonymsAnswerMode:prepareForRemoval()
  self.isKeyboardEnabled = false
  self.inputField:stopBlink()
end

function SynonymsAnswerMode:getSubmittedText(answerText)
  return string.lower(answerText)
end

function SynonymsAnswerMode:getMode()
  return self.mode
end

function SynonymsAnswerMode:getPromptLabel()
  return self.promptLabel
end

function SynonymsAnswerMode:getInputField()
  return self.inputField
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsAnswerMode.new(concept, onAnswer)
  local self = Group.newFullScreen()
  Mixin.mix(self, SynonymsAnswerMode)
  self.mode = "NO_MODE"

  self.background = FilledRectangle.new(self:getWidth(), self:getHeight(), {0, 0, 0, 0})
  self:insertAtCenter(self.background)

  self.promptLabel = SynonymsPromptLabel.new(concept.prompt, SynonymsCommon.COLORS.WHITE)
  self.inputField = SynonymsInputField.new(self:getWidthPart(0.7))
  self.suggestionLabel = SynonymsSuggestionLabel.new(concept)

  local contents = Group.new(self:getWidth(), self:getHeightPart(0.8))
  contents:setAnchorPointTop()
  contents:setLoc(self:getWidthPart(0.5), self:getHeight())
  self:insert(contents)

  local elementColumn = Column.new(0)
  elementColumn:insert(self.promptLabel)
  elementColumn:insert(self.inputField)
  elementColumn:insert(self.suggestionLabel)
  elementColumn:alignChildren()
  contents:insertAtCenter(elementColumn)
  self.elementColumn = elementColumn

  DeviceKeyboard.setInputCallback(function(updatedText)
    if self.isKeyboardEnabled then
      self.inputField:setText(string.lower(updatedText))
    end
  end)

  DeviceKeyboard.setReturnCallback(function(answerText)
    answerText = StringUtils.trim(answerText)

    if (self.isKeyboardEnabled and (#answerText > 0)) then
      onAnswer(self:getSubmittedText(answerText))
    end
  end)

  return self
end
