--- SynonymsConceptGenerator
-- @author Created by karl on January 08, 2015.

class("SynonymsConceptGenerator")

require("data.generators.UtilityGenerators")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SCORE_BASIC = 1
local SCORE_MEDIUM = 2
local SCORE_ADVANCED = 3

local MAX_RETRY_ATTEMPTS = 3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTransformationFunction()
  return function(concept)
    concept.basic = TableUtils.csvToArray(concept.basic)
    concept.medium = TableUtils.csvToArray(concept.medium)
    concept.advanced = TableUtils.csvToArray(concept.advanced)
    concept.hints = TableUtils.csvToArray(concept.hints)
    concept.correct = TableUtils.csvToArray(concept.correct)
    concept.distractors = TableUtils.csvToArray(concept.distractors)
    concept.suggestions = TableUtils.csvToArray(concept.suggestions)
    concept.nouns = TableUtils.createSetFromArray(TableUtils.csvToArray(concept.nouns))
    concept.verbs = TableUtils.createSetFromArray(TableUtils.csvToArray(concept.verbs))

    concept.synonyms = {}
    for _, synonym in pairs(concept.basic) do
      concept.synonyms[synonym] = SCORE_BASIC
    end
    for _, synonym in pairs(concept.medium) do
      concept.synonyms[synonym] = SCORE_MEDIUM
    end
    for _, synonym in pairs(concept.advanced) do
      concept.synonyms[synonym] = SCORE_ADVANCED
    end

    return concept
  end
end

local function getExhaustionHandler(self)
  return function()
    if (#self.deferredConceptIdentifiers <= 0) then
      self:reset()
    else
      Log.debug("Concepts Exhausted: Using deferred concepts.")
      for _, conceptIdentifier in ipairs(self.deferredConceptIdentifiers) do
        self.generator:removeConceptIdentifierFromExclusionList(conceptIdentifier)
      end
      self.deferredConceptIdentifiers = {}
      self.promptsToIgnore = {}
    end

    return true
  end
end

local function ignoreSimilarWords(self, concept)
  for _, word in ipairs(concept.hints) do
    self:ignorePrompt(word)
  end
  for _, word in ipairs(concept.correct) do
    self:ignorePrompt(word)
  end
  for _, word in ipairs(concept.suggestions) do
    self:ignorePrompt(word)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsConceptGenerator:generateNextValue()
  local value = self.generator:next()
  for i = 1, MAX_RETRY_ATTEMPTS do
    local prompt = value.prompt
    if (not self.promptsToIgnore[prompt]) then
      break
    else
      table.insert(self.deferredConceptIdentifiers, value.identifier)
      value = self.generator:next()
    end
  end

  self.generator:markConceptIdentifierAsSeen(value.identifier)
  local transformedValue = self.transformFunction(value)
  ignoreSimilarWords(self, transformedValue)
  return transformedValue
end

function SynonymsConceptGenerator:ignorePrompt(word)
  self.promptsToIgnore[word] = true
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsConceptGenerator.new()
  local conceptGenerator = GameSession.conceptGenerator
  local transformFunction = getTransformationFunction()

  local self = TransformationGenerator.new(conceptGenerator, transformFunction)
  Mixin.mix(self, SynonymsConceptGenerator)

  self.promptsToIgnore = {}
  self.deferredConceptIdentifiers = {}

  conceptGenerator:setMarkGeneratedConceptsAsSeen(false)
  conceptGenerator:setExhaustionHandler(getExhaustionHandler(self))

  return self
end
