--- SynonymsGame
-- @author Created by karl on January 05, 2015.

class("SynonymsGame")

require("SynonymsUI")
require("StandardUIBuilder")
require("SynonymsConfiguration")
require("SynonymsConceptGenerator")
require("SynonymsIntro")
require("SynonymsTimer")
require("SynonymsSuggestionGenerator")
require("utils.WordMatcher")
require("data.generators.UtilityGenerators")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Global
SUGGESTION_GENERATOR = SynonymsSuggestionGenerator.new()

local BACKGROUND_COLOR = SynonymsCommon.COLORS.WHITE
local UI_COMPONENTS_COLOR = SynonymsCommon.COLORS.BLACK_30

local WARNING_TIME = 0.5

local Z_TIMER = 10
local Z_INTRO = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startNextQuestion(self)
  if (self.timer:getTimeRemaining() < SynonymsConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
    self.conceptGenerator:allowMistakesToResurface()
  end
  self.concept = self.conceptGenerator:next()
  GameSession.contentTracker:notifyPresentedConcept(self.concept.identifier)
  self.wordMatcher = WordMatcher.new(self.concept.synonyms)
  self.ui:setConcept(self.concept)
  self.timer:resume()
end

local function scoreAnswer(concept, answer)
  return concept.synonyms[answer] or 0
end

local function closestMatch(self, word)
  local match = self.wordMatcher:getClosestMatch(word)
  return match or word
end

local function onRoundComplete(self)
  startNextQuestion(self)
end

local function checkPartOfSpeech(self, answer)
  local isNoun, isVerb = SynonymsCommon.checkPartOfSpeech(self.concept, answer)
  return isNoun, isVerb
end

local function onRationaleTriggered(self, answer, isNoun, isVerb)
  self.timer:pause()
  self.ui:showAndWaitForRationale(answer, isNoun, isVerb)
  self.timer:resume()
end

local function onAnswerSubmitted(self, answer, ignoreRationale)
  Sound.playSoundEffect("adjectives_wordsubmit")

  local isNoun, isVerb = checkPartOfSpeech(self, answer)
  local isIncorrectPartOfSpeech = (isNoun or isVerb)
  local isPickMode = (self.ui:getActiveMode() == SynonymsCommon.MODES.PICK)
  local isNormalMode = (self.ui:getActiveMode() == SynonymsCommon.MODES.NORMAL)

  if ((not isIncorrectPartOfSpeech) or isPickMode or ignoreRationale) then
    local originalAnswer = answer
    if (not isPickMode) then
      answer = closestMatch(self, answer)
    end

    local didAutocorrect = (originalAnswer ~= answer)
    if didAutocorrect then
      self.ui:animateAutocorrect(answer)
    end

    local score = scoreAnswer(self.concept, answer)
    table.insert(self.answers, {concept = self.concept, answer = answer, score = score})
    local isCorrect = (score > 0)
    if isCorrect then
      GameSession.notifyCorrectAnswer(self.concept.identifier)
      GameSession.contentTracker:notifyCorrectAnswer(self.concept.identifier, nil, answer)
      self.accumulatedPoints = self.accumulatedPoints + score
      self.ignorePromptFunction(answer)
    else
      GameSession.notifyIncorrectAnswer(self.concept.identifier)
      GameSession.contentTracker:notifyIncorrectAnswer(self.concept.identifier, nil, answer)
      if isNormalMode then
        self.conceptGenerator:notifyMistake(self.concept.identifier)
      end
    end
    isIncorrectPartOfSpeech = false
    return SynonymsRoundResponse.new(isCorrect, isIncorrectPartOfSpeech, answer)
  else
    onRationaleTriggered(self, answer, isNoun, isVerb)
    local isCorrect = false
    return SynonymsRoundResponse.new(isCorrect, isIncorrectPartOfSpeech, answer)
  end
end

local function onSkip(self)
  table.insert(self.answers, {concept = self.concept, answer = nil, score = 0})
  GameSession.contentTracker:notifySkippedAnswer(self.concept.identifier)
  self.conceptGenerator:notifyMistake(self.concept.identifier)
  Sound.playSoundEffect("adjectives_skip")
  self.timer:addTime(SynonymsConfiguration.LOST_SECONDS_ON_SKIP)
  self.timer:pauseForSkip()
  wait(self.ui:animateSkip())
  waitFor(0.167)
  self.timer:resumeFromSkip()
  startNextQuestion(self)
end

local function onGameOver(self)
  local points = self.accumulatedPoints
  local passPointThreshold = SynonymsConfiguration.getQuestionsCountToPass()
  local didPassGame = (points >= passPointThreshold)
  if didPassGame then
    local baseScore = SynonymsConfiguration.PASS_SCORE
    GameSession.increaseScore(baseScore)

    local bonusCount = points - passPointThreshold
    local bonusPoints = SynonymsConfiguration.getDecayingBonusPoints(bonusCount)
    GameSession.increaseScore(bonusPoints)
    GameSession.bonusReporter:setBonusCount(BonusTypes.TYPES.EXTRA_COLLECTED_BONUS, points)

    GameSession.notifyGameWin()
  else
    GameSession.notifyGameLoss()
  end
end

local function hideUIOverlay(self)
  local uiOverlay = self:getUIOverlay()
  uiOverlay:linkColorAttrToChildren(true)
  self.timer:linkColorAttrToChildren(true)
  AHOpacityTo.new(uiOverlay, 0.2, nil, 0):start()
  AHOpacityTo.new(self.timer, 0.2, nil, 0):start()
end

local function onTimeout(self)
  self.ui:stop()
  self.pauseButton:setTouchEnabled(false)
  hideUIOverlay(self)
  local shouldShowTutorial = GameSession.hasNeverWon()
  self.ui:runEndSequence(self.answers, shouldShowTutorial, function() onGameOver(self) end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsGame:gameFlow()
  self.ui:setTouchEnabled(false)
  wait(self.intro:run())
  self.ui:setTouchEnabled(true)
  self.intro:removeSelf()
  self.intro = nil
  self.pauseButton:setVisibleRecursive(true)
  self.ui:setTouchEnabled(false)
  startNextQuestion(self)
  self.timer:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------
local function isRepeatAnswer(self, text)
  for _, submission in ipairs(self.answers) do
    local isSubmissionForCurrentConcept = (submission.concept == self.concept)
    local skippedPreviousSubmission = (submission.answer == nil)

    if (isSubmissionForCurrentConcept and (not skippedPreviousSubmission)) then
      local isRepeat = (submission.answer == closestMatch(self, text))

      if isRepeat then
        return true
      end
    end
  end

  return false
end

local function configureGameDifficulty(self)
  local difficultyKeyFrames = SynonymsConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  local passScore = SynonymsConfiguration.PASS_SCORE
  local rankTwoScore = SynonymsConfiguration.getRankTwoScore()
  local rankThreeScore = SynonymsConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)
end

local function createGameElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  self:insertAtCenter(background)

  local onAnswerSubmittedCallback = function(answer, ignoreRationale)
    return onAnswerSubmitted(self, answer, ignoreRationale)
  end
  local onRoundCompleteCallback = function()
    onRoundComplete(self)
  end
  local onSkipCallback = function()
    onSkip(self)
  end
  local isRepeatAnswerCallback = function(text)
    return isRepeatAnswer(self, text)
  end
  local isCorrectAnswerCallback = function(answer)
    return (scoreAnswer(self.concept, answer) > 0)
  end
  self.ui = SynonymsUI.new(self.timer, onAnswerSubmittedCallback, onRoundCompleteCallback, onSkipCallback, isRepeatAnswerCallback, isCorrectAnswerCallback)
  self:insertAtCenter(self.ui)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.EXTRA_COLLECTED_BONUS)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  self.pauseButton = builder:addPauseButton()

  self.timer = SynonymsTimer.new(SynonymsConfiguration.GAME_DURATION_IN_SECONDS, WARNING_TIME, UI_COMPONENTS_COLOR, function() onTimeout(self) end)
  self:insertAtCenter(self.timer, Z_TIMER)
end

-- Constructor ------------------------

function SynonymsGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, SynonymsGame)

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)

  self.intro = SynonymsIntro.new()
  self:insertAtCenter(self.intro, Z_INTRO)

  local synonymsConceptGenerator = SynonymsConceptGenerator.new()
  self.ignorePromptFunction = function(word)
    synonymsConceptGenerator:ignorePrompt(word)
  end
  self.conceptGenerator = LearningFocusedRoundGenerator.new(synonymsConceptGenerator, nil, SynonymsConfiguration.MAXIMUM_CONTENT_REPETITION_TIMES)
  self.answers = {}
  self.accumulatedPoints = 0
  self.pauseButton:setVisibleRecursive(false)

  return self
end
