--- SynonymsInputField
-- @author Created by karl on January 22, 2015.

class("SynonymsInputField")

require("SynonymsCommon")
require("SynonymsCursor")
require("ShapeEmitter")
require("text.ComposedGlyphLabel")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAX_LETTERS = 30

local COLORFUL_COLORS = SynonymsCommon.COLOR_PROGRESSION
local WHITE_COLORS = {SynonymsCommon.COLORS.WHITE, SynonymsCommon.COLORS.WHITE, SynonymsCommon.COLORS.WHITE}

local FONT_FACE = FontFaces.DIN_OT_LIGHT
local FONT_SIZE = 54
local LABEL_COLOR = SynonymsCommon.COLORS.BLACK_60

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getColors(isColorful)
  if isColorful then
    return COLORFUL_COLORS
  end
  return WHITE_COLORS
end

local function storeColors(letters, startIndex, endIndex, color)
  for i = startIndex, endIndex do
    letters[i].finalColor = color
  end
end

local function assignColors(letters, forcedColor)
  local letterCount = #letters
  if letterCount == 1 and not forcedColor then
    forcedColor = SynonymsCommon.COLORS.BLACK
  end

  if forcedColor then
    for _, letter in ipairs(letters) do
      letter.finalColor = forcedColor
    end
  elseif (letterCount > 2) then
    local third = MathUtils.round(letterCount / 3)
    storeColors(letters, 1, third, SynonymsCommon.COLOR_PROGRESSION[1])
    storeColors(letters, third + 1, letterCount - third, SynonymsCommon.COLOR_PROGRESSION[2])
    storeColors(letters, letterCount - third + 1, letterCount, SynonymsCommon.COLOR_PROGRESSION[3])
  elseif (letterCount == 2) then
    letters[1].finalColor = SynonymsCommon.COLOR_PROGRESSION[1]
    letters[2].finalColor = SynonymsCommon.COLOR_PROGRESSION[3]
  end
end

local function newColorLettersAction(label, letters)
  local duration = 0.2
  local action = AHSpawn.new()
  for _, letter in ipairs(letters) do
    action:addAction(AHColorTo.new(letter, duration, nil, letter.finalColor))
  end

  return AHSequence.new(
    function()
      local labelColor = {label:getColor()}
      label:setColor(1, 1, 1, 1)
      for _, letter in ipairs(letters) do
        letter:setColor(unpack(labelColor))
      end
    end,
    action
  )
end

local function newDropLettersAction(self, letters)
  local scale = self:getScl()
  local deltaY = -Screen.getHeightPart(0.7 * (1 / scale))
  local minRotation = -90
  local maxRotation = 90
  local maxDelay = 0.35

  local duration = 0.7
  local ease = AHEaseSimpleIn.new()

  letters = TableUtils.shallowCopy(letters)
  TableUtils.shuffleArray(letters)

  local action = AHSpawn.new()
  for index, letter in ipairs(letters) do
    local delay = 0
    if #letters > 1 then
      delay = MathUtils.mapLinear(index, 1, #letters, 0, maxDelay)
    end
    local rotation = math.random(minRotation, maxRotation)
    action:addAction(AHDelayed.new(delay, AHSpawn.new(
      AHLocationYBy.new(letter, duration, ease, deltaY),
      AHRotationTo.new(letter, duration, ease, rotation)
    )))
  end
  return action
end

local function adjustScaleIfNecessary(self)
  if (self:getWidth() > self.maxWidth) then
    self:setEffectiveDims(self.maxWidth)
  else
    self:setScl(1)
  end
end

local function newBoldLabel(text)
  local buildGlyphs = true
  local boldLabel = Label.new(text, Font.new(SynonymsCommon.FONTS.BOLD, 54), nil, nil, buildGlyphs)
  return ComposedGlyphLabel.new(boldLabel)
end

local function createIncorrectRedFlashAnimation(self)
  local startingColor = {self.answerLabel:getColor()}
  local delay = 0.02
  local fadeDuration = 0.08
  local ease = AHEaseLinear.new()
  local hold = 0.1
  local changeToRed = AHSpawn.new()
  local changeBack = AHSpawn.new()
  local errorColor = SynonymsCommon.COLORS.PENALTY_RED
  if self.hintLabel then
    local startingHintColor = {self.hintLabel:getColor()}
    changeToRed:addAction(AHColorTo.new(self.hintLabel, fadeDuration, ease, errorColor))
    changeBack:addAction(AHColorTo.new(self.hintLabel, fadeDuration, ease, startingHintColor))
  end
  changeToRed:addAction(AHColorTo.new(self.answerLabel, fadeDuration, ease, errorColor))
  changeBack:addAction(AHColorTo.new(self.answerLabel, fadeDuration, ease, startingColor))

  return AHSequence.new(
    AHDelay.new(delay),
    changeToRed,
    AHDelay.new(hold),
    changeBack
  )
end

local function createIncorrectShrinkAndStrikethroughAnimation(self)
  local delay = 0.1 + SynonymsCommon.TIMINGS.INCORRECT_FADE_DELAY
  local scaleAmount = 0.5
  local scaleEase = SynonymsCommon.COMMON_EASE
  local scaleDuration = SynonymsCommon.TIMINGS.INCORRECT_SCALE_DURATION
  local scaleAnimation = AHScaleTo.new(self, scaleDuration, scaleEase, scaleAmount)

  local boldDuration = SynonymsCommon.TIMINGS.INCORRECT_FADE_DURATION
  local boldLabel = newBoldLabel(self:getText())
  local boldColor = self.boldColor
  local boldOpacity = boldColor[4]
  boldLabel:setColor(unpack(boldColor))
  boldLabel:setOpacity(0)
  self:insertAtCenter(boldLabel)
  boldLabel:setX(boldLabel:getX() - 2.5)
  local boldLabelAnimation = AHSpawn.new(
    AHOpacityTo.new(self.answerLabel, boldDuration, nil, 0),
    AHOpacityTo.new(boldLabel, boldDuration, nil, boldOpacity))

  local strikethroughAnimation = SynonymsCommon.createStrikethroughAnimation(self)

  local moveDistance = SynonymsCommon.INCORRECT_MOVE_DISTANCE
  local moveAnimation = AHLocationYBy.new(self, scaleDuration, scaleEase, moveDistance)

  return AHDelayed.new(delay, AHSpawn.new(scaleAnimation, strikethroughAnimation, boldLabelAnimation, moveAnimation))
end

local function createIncorrectScaleAndExitAnimation(self)
  local scaleUpFunction = function()
    wait(AHScaleTo.new(self, 0.1, AHEaseSineInOut.new(), self:getScl() * 1.2):start())
  end
  return AHSequence.new(
    scaleUpFunction,
    AHScaleTo.new(self, 0.3, AHEaseSineInOut.new(), 0)
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsInputField:trimWhitespace()
  self.answerLabel.text = StringUtils.trim(self.answerLabel.text)
end

function SynonymsInputField:getText()
  if self.hintLabel then
    return self.hintLabel.text .. self.answerLabel.text
  end
  return self.answerLabel.text
end

function SynonymsInputField:setText(text)
  self.answerLabel:setText(text)
  self:alignChildren()
  adjustScaleIfNecessary(self)
end

function SynonymsInputField:stopBlink()
  self.cursor:stopBlink()
end

function SynonymsInputField:toggleBlink(isPaused)
  self.cursor:toggleBlink(isPaused)
end

function SynonymsInputField:setCorrectForcedColor(color)
  self.forcedColor = color
end

function SynonymsInputField:setIncorrectBoldColor(color)
  self.boldColor = color
end

function SynonymsInputField:animateCorrect()
  local boldLabel = newBoldLabel(self:getText())
  boldLabel:setColor(unpack(SynonymsCommon.COLORS.BLACK))
  boldLabel:setOpacity(0)
  self:insertAtCenter(boldLabel)
  boldLabel:setX(boldLabel:getX() - 2.5)
  local letters = boldLabel.children
  assignColors(letters, self.forcedColor)
  local scale = self:getScl()
  local scaleLow = scale * 0.9
  local scalePeak = scale * 1.4
  local scaleHigh = scale * 1.2
  local durationLow = 0.05
  local durationPeak = 0.3
  local durationHigh = 0.15
  local easeLow = AHEaseSineIn.new()
  local easePeak = AHEaseExponentialInOut.new()
  local easeHigh = AHEaseSineOut.new()

  local scaleUpAndFade = AHSpawn.new(
    AHScaleTo.new(self, durationPeak, easePeak, scalePeak),
    AHOpacityTo.new(self.answerLabel, durationPeak * 0.3, nil, 0),
    AHOpacityTo.new(boldLabel, durationPeak, nil, 0.6)
  )
  if self.hintLabel then
    scaleUpAndFade:addAction(AHOpacityTo.new(self.hintLabel, durationPeak, nil, 0))
  end

  return AHSequence.new(
    function() self.cursor:setVisibleRecursive(false) end,
    AHScaleTo.new(self, durationLow, easeLow, scaleLow),
    AHSpawn.new(
      scaleUpAndFade,
      AHDelayed.new(durationPeak * 0.6, function()
        ShapeEmitter.emitFrom(self, getColors(self.isColorful), true)
      end)
    ),
    AHScaleTo.new(self, durationHigh, easeHigh, scaleHigh),
    AHSpawn.new(
      newColorLettersAction(boldLabel, letters),
      newDropLettersAction(self, letters)
    )
  ):start()
end

function SynonymsInputField:animateIncorrect()
  self.cursor:setVisibleRecursive(false)
  local wiggleAnimation = SynonymsCommon.createWiggleAnimation(self)
  local redFlashAnimation = createIncorrectRedFlashAnimation(self)
  local shrinkAndStrikethroughAnimation = createIncorrectShrinkAndStrikethroughAnimation(self)
  return AHSpawn.new(
    wiggleAnimation,
    redFlashAnimation,
    shrinkAndStrikethroughAnimation
  ):start()
end

function SynonymsInputField:animateExit()
  return createIncorrectScaleAndExitAnimation(self):start()
end

function SynonymsInputField:newColorTo(duration, ease, color)
  local action = AHSpawn.new(
    AHColorTo.new(self.answerLabel, duration, ease, color)
  )
  if self.hintLabel then
    action:addAction(AHColorTo.new(self.hintLabel, duration, ease, color))
  end
  return action
end

function SynonymsInputField:setIsColorful(isColorful)
  self.isColorful = isColorful
end

function SynonymsInputField:addHint(hint)
  local hintColor = SynonymsCommon.COLORS.BLACK_30
  self.hintLabel = Label.new(hint, self.font)
  self.hintLabel:setColor(unpack(hintColor))
  local underline = FilledRectangle.new(self.hintLabel:getWidth(), 2)
  underline:setLoc(self.hintLabel:getWidthPart(0.5), 4)
  self.hintLabel:insert(underline)
  underline:inheritColor(self.hintLabel)
  self:insert(self.hintLabel)
  self:insertSpacer(4)

  -- Move hintLabel and spacer to the front of the Row
  local children = self:getChildren()
  table.insert(children, 1, table.remove(children))
  table.insert(children, 1, table.remove(children))
  self:alignChildren()
end

function SynonymsInputField:getFont()
  return self.font
end

function SynonymsInputField:hide()
  self.answerLabel:setOpacity(0)
  self.cursor:setVisibleRecursive(false)
end

function SynonymsInputField:show()
  local alpha = LABEL_COLOR[4]
  self.answerLabel:setOpacity(alpha)
  self.cursor:setVisibleRecursive(true)
end

function SynonymsInputField:createClone(text)
  local clone = Row.new(0)

  local label = Label.new(text, Font.new(FONT_FACE, FONT_SIZE))
  label:setColor(unpack(LABEL_COLOR))
  clone:insert(label)

  local cursorHeight = self.answerLabel:getHeight()
  local cursor = self.cursor:createClone(cursorHeight)
  clone:insert(cursor)

  clone:alignChildren()

  return clone
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function preventLabelFlickerBugByPreAddingGlyphs(answerLabel)
  answerLabel:preAddGlyphs(MAX_LETTERS)
end

-- Constructor ------------------------

function SynonymsInputField.new(maxWidth)
  local self = Row.new(0)
  Mixin.mix(self, SynonymsInputField)

  self.font = PrerenderedFont.new(FONT_FACE, FONT_SIZE)

  self.answerLabel = Label.new("", self.font)
  preventLabelFlickerBugByPreAddingGlyphs(self.answerLabel)
  self.answerLabel:setColor(unpack(LABEL_COLOR))
  self:insert(self.answerLabel)

  local cursorHeight = self.answerLabel:getHeight()
  self.cursor = SynonymsCursor.new(cursorHeight, SynonymsCommon.COLORS.GREY, SynonymsCommon.COLORS.WHITE)
  self:insert(self.cursor)

  self:alignChildren()

  self.isColorful = false
  self.maxWidth = maxWidth

  self.forcedColor = nil
  self.boldColor = SynonymsCommon.COLORS.DARK_GREY_2

  return self
end
