--- SynonymsMultiAnswerMode
-- @author Created by karl on February 02, 2015.

class("SynonymsMultiAnswerMode")

require("SynonymsAnswerMode")
require("SynonymsInputField")
require("actions.advanced.Shake")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAX_WIDTH = Screen.getWidthPart(0.7)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function replaceInputField(self)
  if (self.inputField ~= nil) then
    self.inputField:stopBlink()
    self.inputField:removeSelf()
  end

  local elementColumn = self.elementColumn
  local inputField = SynonymsInputField.new(MAX_WIDTH)
  elementColumn:insert(inputField)
  elementColumn:alignChildren()

  elementColumn:forceUpdate()

  self.inputField = inputField
end

local function scaleLabelIfNeeded(label)
  if label:getWidth() > MAX_WIDTH then
    label:setEffectiveDims(MAX_WIDTH)
  end
end

local function onRepeatAnswer(self)
  local doRejectAnimationAndReplace = function()
    local shakeAction = AHAxisOffsetShake.new(self.inputField, self.inputField:getWidth())
    local fadeAction = AHOpacityTo.new(self.inputField.answerLabel, 0.1, nil, 0)
    local delayedFadeAction = AHDelayed.new(0.5, fadeAction)
    local rejectionAnimation = AHSpawn.new(shakeAction, delayedFadeAction)

    wait(rejectionAnimation:start())
    replaceInputField(self)
  end
  return runInCoroutine(doRejectAnimationAndReplace)
end

local function onSingleAnswer(self, answer)
  local answerLabel = Label.new(answer, self.inputField:getFont())
  scaleLabelIfNeeded(answerLabel)
  answerLabel:setColor(unpack(SynonymsCommon.COLORS.BLACK_60))

  local answerColumn = self.answerColumn
  answerColumn:insert(answerLabel)
  answerColumn:alignChildren()

  replaceInputField(self)

  local maxHeight = self.instructionsLabel:getYBottom() - self.endButton:getYTop()

  if self.elementColumn:getHeight() > maxHeight then
    self.elementColumn:setEffectiveDims(nil, maxHeight)
  end

  answerLabel:setOpacity(0.0)
  local fadeInAction = AHOpacityTo.new(answerLabel, 0.1, nil, 1.0)
  fadeInAction:start()

  self.previousAnswerLabel = answerLabel
end

local function animateIncorrectAnswers(self)
  for _, label in ipairs(self.answerColumn:getChildren()) do
    local isCorrect = self.isCorrectAnswerCallback(label:getText())
    if (not isCorrect) then
      SynonymsCommon.createStrikethroughAnimation(label):start()
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsMultiAnswerMode:animateIn()
  AHDelayed.new(
    0.21,
    function()
      self.promptLabel:show()
      self.inputField:show()
      self.isKeyboardEnabled = true
      DeviceKeyboard.setEnabled(true)
    end
  ):start()
end

function SynonymsMultiAnswerMode:animateSkip()
  -- Do nothing
end

function SynonymsMultiAnswerMode:animateAnswer()
  -- Do nothing
end

function SynonymsMultiAnswerMode:animateAutocorrect(answer)
  self.super.animateAutocorrect(self, answer, self.previousAnswerLabel)
end

function SynonymsMultiAnswerMode:setTouchEnabled(enabled)
  self.endButton:setTouchEnabled(enabled)
  self.super.setTouchEnabled(self, enabled)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newEndButton(self, onAnswer, onComplete)
  local buttonText = Localizer.get("END_MULTI")
  local height = Screen.getWidthPart(0.12)
  local font = SynonymsCommon.FONTS.MEDIUM
  local normalColor = SynonymsCommon.COLORS.WHITE
  local hightlightColor = SynonymsCommon.COLORS.BLUE
  local endButton = nil
  endButton = SynonymsChoiceButton.new(buttonText, height, font, normalColor, hightlightColor, function()
    self.isKeyboardEnabled = false
    local ignoreRationale = true
    local text = StringUtils.trim(string.lower(self.inputField:getText()))
    if (#text > 0) then
      onSingleAnswer(self, text)
      onAnswer(text, ignoreRationale)
    end
    wait(endButton:animateTouch(true))
    animateIncorrectAnswers(self)
    waitFor(0.3)
    wait(AHOpacityTo.new(self.endButton, 0.4, nil, 0):start())
    wait(AHLocationYBy.new(self.elementColumn, 0.5, AHEaseExponentialIn.new(), -Screen.height):start())
    onComplete()
  end)
  return endButton
end

-- Constructor ------------------------

function SynonymsMultiAnswerMode.new(concept, onAnswer, onComplete, isRepeatAnswerCallback, isCorrectAnswerCallback)
  local self = SynonymsAnswerMode.new(concept, onAnswer)
  Mixin.mix(self, SynonymsMultiAnswerMode)
  self.mode = SynonymsCommon.MODES.MULTI

  self.isCorrectAnswerCallback = isCorrectAnswerCallback
  self.suggestionLabel:removeSelf()

  self.keyboardHeight = self:getHeightPart(0.4)
  local endButton = newEndButton(self, onAnswer, onComplete)
  local endButtonMargin = 20
  local endButtonY = self.keyboardHeight + endButton:getHeightPart(0.5) + endButtonMargin
  endButton:setLoc(self:getWidthPart(0.5), endButtonY)
  self:insert(endButton)
  self.endButton = endButton

  local spacingThatAllowsOverlap = -14.0
  local answerColumn = Column.new(spacingThatAllowsOverlap)
  self.answerColumn = answerColumn

  self.elementColumn:insert(answerColumn)
  local elementColumnY = self.keyboardHeight + 0.5 * (self:getHeight() - self.keyboardHeight)
  self:takeChild(self.elementColumn)
  self.elementColumn:setLoc(self:getWidthPart(0.5), elementColumnY)

  replaceInputField(self)

  local label = Label.new(Localizer.get("INSTRUCTIONS_MULTI"), Font.new(SynonymsCommon.FONTS.BOLD, 17))
  label:setColor(unpack(SynonymsCommon.COLORS.BLACK_50))
  label:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.9))
  self:insert(label)
  self.instructionsLabel = label

  DeviceKeyboard.setReturnCallback(function(answerText)
    if (self.isKeyboardEnabled and (#answerText > 0)) then
      self.isKeyboardEnabled = false
      answerText = StringUtils.trim(string.lower(answerText))
      local isRepeatAnswer = isRepeatAnswerCallback(answerText)
      if isRepeatAnswer then
        wait(onRepeatAnswer(self))
      else
        local isNoun, isVerb = SynonymsCommon.checkPartOfSpeech(concept, answerText)
        local isIncorrectPartOfSpeech = isNoun or isVerb
        if (not isIncorrectPartOfSpeech) then
          onSingleAnswer(self, answerText)
        end
        onAnswer(answerText)
      end
      DeviceKeyboard.clear()
      self.isKeyboardEnabled = true
    end
  end)

  self.previousAnswerLabel = nil

  self.promptLabel:hide()
  self.inputField:hide()

  return self
end
