--- SynonymsPromptLabel
-- @author Created by karl on January 28, 2015.

class("SynonymsPromptLabel")

require("text.ComposedGlyphLabel")
require("SynonymsCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FONT = Font.new(SynonymsCommon.FONTS.MEDIUM, 56)
local DEFAULT_LABEL_COLOR = SynonymsCommon.COLORS.BLACK

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setChildColors(self, startIndex, endIndex, color)
  for i = startIndex, endIndex do
    self.children[i]:setColor(unpack(color))
  end
end

local function assignChildColors(self)
  if (#self.text > 2) then
    local childCount = #self.children
    local third = MathUtils.round(childCount / 3)
    setChildColors(self, 1, third, SynonymsCommon.COLOR_PROGRESSION[1])
    setChildColors(self, third + 1, childCount - third, SynonymsCommon.COLOR_PROGRESSION[2])
    setChildColors(self, childCount - third + 1, childCount, SynonymsCommon.COLOR_PROGRESSION[3])
  end
end

local function createIncorrectScaleAndExitAnimation(self)
  local scaleUpFunction = function()
    wait(AHScaleTo.new(self, 0.1, AHEaseSineInOut.new(), self:getScl() * 1.2):start())
  end
  return AHSequence.new(
    scaleUpFunction,
    AHScaleTo.new(self, 0.3, AHEaseSineInOut.new(), 0)
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsPromptLabel:setText(text)
  self.text = text
  local label = Label.new(text, FONT)
  self:setLabel(label)
end

function SynonymsPromptLabel:prepareToAnimateIn()
  self.baseY = self.children[1]:getY()
  local startY = self.baseY + Screen.getHeightPart(0.5)

  for _, child in ipairs(self.children) do
    child:setY(startY)
  end
end

function SynonymsPromptLabel:animateIn()
  local minRotation = -90
  local maxRotation = 90
  local maxDelay = 0.5

  local duration = 1
  local ease = AHEaseExponentialOut.new()

  local action = AHSpawn.new()
  for _, child in ipairs(self.children) do
    local delay = math.random() * maxDelay
    local rotation = math.random(minRotation, maxRotation)
    local colorChangeAction = AHDelayed.new(duration * 0.6, AHColorTo.new(child, duration * 0.4, nil, {0, 0, 0}))
    child._colorAction = colorChangeAction

    child:setRot(rotation)
    action:addAction(AHDelayed.new(delay, AHSpawn.new(
      AHLocationYTo.new(child, duration, ease, self.baseY),
      AHRotationTo.new(child, duration, ease, 0)
    )))
    action:addAction(colorChangeAction)
  end

  assignChildColors(self)
  return action:start()
end

function SynonymsPromptLabel:animateCorrect()
  local lightLabel = Label.new(self.text, Font.new(SynonymsCommon.FONTS.LIGHT, 56))
  lightLabel:setColor(unpack(SynonymsCommon.COLORS.BLACK))
  lightLabel:setOpacity(0)
  self:insertAtCenter(lightLabel)
  return AHSequence.new(
    AHSpawn.new(
      AHSequence.new(
        AHScaleTo.new(self, 0.2, nil, 1.2),
        AHScaleTo.new(self, 0.3, AHEaseSineIn.new(), 0.5)
      ),
      AHOpacityTo.new(self, 0.5, nil, 0),
      AHOpacityTo.new(lightLabel, 0.5, nil, 0.3)
    ),
    AHDelay.new(0.4),
    AHOpacityTo.new(lightLabel, 0.3, nil, 0)
  ):start()
end

function SynonymsPromptLabel:animateIncorrect()
  local fadeDelay = SynonymsCommon.TIMINGS.INCORRECT_FADE_DELAY

  local scale = 0.5
  local scaleDuration = SynonymsCommon.TIMINGS.INCORRECT_SCALE_DURATION
  local scaleEase = SynonymsCommon.COMMON_EASE
  local scaleAnimation = AHScaleTo.new(self, scaleDuration, scaleDuration, scale)

  local colorDuration = SynonymsCommon.TIMINGS.INCORRECT_FADE_DURATION
  local colorEase = AHEaseLinear.new()
  local color = self.incorrectColor
  local colorAnimation = self:newColorTo(colorDuration, colorEase, color)

  local moveDistance = SynonymsCommon.INCORRECT_MOVE_DISTANCE
  local moveAnimation = AHLocationYBy.new(self, scaleDuration, scaleEase, moveDistance)

  local fadeAnimation = AHSpawn.new(
    scaleAnimation,
    colorAnimation,
    moveAnimation
  )

  return AHDelayed.new(fadeDelay, fadeAnimation):start()
end

function SynonymsPromptLabel:setIncorrectColor(color)
  self.incorrectColor = color
end

function SynonymsPromptLabel:animateExit()
  return createIncorrectScaleAndExitAnimation(self):start()
end

function SynonymsPromptLabel:newColorTo(duration, ease, color)
  local action = AHSpawn.new()
  for _, child in ipairs(self.children) do
    local createAndRunColorChange = function()
      local colorChangeAction = AHColorTo.new(child, duration, ease, color)

      if child._colorAction then
        child._colorAction:stop()
      end

      colorChangeAction:start()
    end
    action:addAction(createAndRunColorChange)
  end
  return action
end

function SynonymsPromptLabel:getText()
  return self.text
end

function SynonymsPromptLabel:hide()
  self:setOpacity(0)
end

function SynonymsPromptLabel:show()
  self:setOpacity(1)
end

function SynonymsPromptLabel:createClone(text)
  local clone = Label.new(text, FONT)
  clone:setColor(unpack(DEFAULT_LABEL_COLOR))

  return clone
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsPromptLabel.new(text, initialColor)
  text = text or " "
  initialColor = initialColor or DEFAULT_LABEL_COLOR
  local default = nil
  local buildGlyphs = true
  local label = Label.new(text, FONT, default, default, buildGlyphs)
  local self = ComposedGlyphLabel.new(label)
  Mixin.mix(self, SynonymsPromptLabel)

  self.text = text

  for _, child in ipairs(self.children) do
    child:setColor(unpack(initialColor))
  end

  self.incorrectColor = SynonymsCommon.COLORS.LIGHT_GREY_2

  return self
end
