--- SynonymsScoreCard
-- @author Created by karl on January 08, 2015.

class("SynonymsScoreCard")

require("SynonymsCommon")
require("SynonymsTutorial")
require("SynonymsScoreScroller")
require("SynonymsScoreRectangle")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_LOW = 0
local Z_HIGH = 10
local Z_TUTORIAL = 20

local FOOTER_HEIGHT = Screen.safeBottom + 160

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTotalScore(answers)
  local score = 0
  for _, answer in ipairs(answers) do
    score = score + answer.score
  end
  return score
end

local function newLabel(text, font, fontSize, color)
  local label = Label.new(text, Font.new(font, fontSize))
  if color then
    label:setColor(unpack(color))
  end
  return label
end

local function newMultilineLabel(text, font, fontSize, color)
  local label = DynamicTextArea.newSimple(text, Screen.getWidthPart(0.8), Font.new(font, fontSize), Alignment.CENTER)
  if color then
    label:setColor(unpack(color))
  end
  return label
end

local function newScoreLabel(leftText, rightText, fontSize, color)
  local spacer = fontSize * 0.4
  local row = Row.new(0)
  if leftText and #leftText > 0 then
    row:insert(newLabel(leftText, SynonymsCommon.FONTS.MEDIUM, fontSize, color))
    row:insertSpacer(spacer * 1.4)
  end
  row:insert(SynonymsScoreRectangle.new(fontSize * 0.9, SynonymsCommon.COLORS.YELLOW))
  if rightText and #rightText > 0 then
    row:insertSpacer(spacer)
    row:insert(newLabel(rightText, SynonymsCommon.FONTS.MEDIUM, fontSize, color))
  end
  row:alignChildren()
  return row
end

local function newContinueButton(color, onPressed)
  local buttonW = Screen.getWidthPart(0.85)
  local buttonH = buttonW * 0.15
  local cornerRadius = buttonH * 0.1
  local fontSize = buttonH * 0.45
  local button = FilledRoundedRectangle.new(buttonW, buttonH, cornerRadius, color)
  button:insertAtCenter(newLabel(Localizer.get("END_GAME"), SynonymsCommon.FONTS.MEDIUM, fontSize, SynonymsCommon.COLORS.WHITE))
  Button.makeButton(button, function()
    button:setColor(unpack(SynonymsCommon.COLORS.WHITE))
    onPressed()
    button:setTouchEnabled(false)
  end)
  return button
end

local function newMarkedUpLabel(text, font, fontSize, color)
  local baseAttributes = {
    font = Font.new(font, fontSize)
  }
  local tagAttributes = {
    b = {font = Font.new(SynonymsCommon.FONTS.BOLD, fontSize)}
  }
  local label = DynamicTextArea.new(text, Screen.width, baseAttributes, tagAttributes, Alignment.CENTER)
  if color then
    label:setColor(unpack(color))
  end
  return label
end

local function showFirstEntry(self)
  return self.scroller:animateFirstEntry()
end

local function newExample(rectangleCount, rating, examples)
  local bigFontSize = 19
  local smallFontSize = 15
  local smallColor = {1, 1, 1, 0.6}
  local col = Column.new(bigFontSize * 0.07)
  local examplesString = string.format("%s %s", Localizer.get("EXAMPLES_PREFIX"), examples)

  col:insert(SynonymsScoreRectangleSet.new(Screen.getWidthPart(0.14), rectangleCount))
  col:insert(newLabel(rating, SynonymsCommon.FONTS.BOLD, bigFontSize))
  col:insert(newLabel(examplesString, SynonymsCommon.FONTS.LIGHT, smallFontSize, smallColor))
  col:alignChildren()
  return col
end

local function newScoreExplanationScreen()
  local col = Column.new(Screen.getHeightPart(0.045))
  local text = Localizer.get("END_SCORE_EXPLAINER")
  col:insert(newMultilineLabel(text, SynonymsCommon.FONTS.LIGHT, Screen.getWidthPart(0.04)))
  col:insert(newExample(1, Localizer.get("COMMON"), Localizer.get("EXAMPLES_COMMON")))
  col:insert(newExample(2, Localizer.get("UNCOMMON"), Localizer.get("EXAMPLES_UNCOMMON")))
  col:insert(newExample(3, Localizer.get("EXPERT"), Localizer.get("EXAMPLES_EXPERT")))

  col:alignChildren()
  col:linkColorAttrToChildren(true)
  return col
end

local function showResult(self)
  self:insert(self.resultsFooter, Z_HIGH)
  return AHSequence.new(
    AHLocationYBy.new(self.resultsFooter, 0.3, AHEaseSineIn.new(), self.resultsFooter:getHeight())
  ):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsScoreCard:showCard()
  return runInCoroutine(function()
    wait(self.scroller:animateIn())
    self.scroller:setTouchEnabled(true)
    wait(self.scroller:fadeInDisclaimer())
    wait(showResult(self))
  end)
end

function SynonymsScoreCard:showTutorial()
  return runInCoroutine(function()
    local screens = {
      newScoreExplanationScreen()
    }
    local pointToGroups = {
      self:findGroupWithTag(SynonymsCommon.TAGS.SCORE)
    }

    local function completeTutorial()
      wait(self.tutorial:endTutorial())
      self.tutorial:removeSelf()
      wait(self:showCard())
    end

    local function showScoreExplanationPart()
      local screen = screens[1]
      local pointToGroup = pointToGroups[1]
      assert(pointToGroup ~= nil, "Group not found to display Score Explanation. Make sure a group is tagged with " .. SynonymsCommon.TAGS.SCORE)
      wait(showFirstEntry(self))
      wait(self.tutorial:pointTo(pointToGroup))
      wait(self.tutorial:showArrow(pointToGroup), self.tutorial:showScreen(screen))
      self.tutorial:waitForTap(function() completeTutorial() end)
    end

    self.tutorial = SynonymsTutorial.new()
    self:insertAtCenter(self.tutorial, Z_TUTORIAL)
    coroutine.yield()
    coroutine.yield()
    showScoreExplanationPart()
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newHeader(width)
  local marginW = (Screen.width - width) / 2
  local fontSize = 19
  local textColor = {0, 0, 0, 0.8}
  local underlineColor = {0, 0, 0, 0.2}
  local row = Row.new()
  row:setWidth(width)
  row:insert(newLabel(Localizer.get("SYNONYMS"), SynonymsCommon.FONTS.MEDIUM, fontSize, textColor))
  row:insert(newLabel(Localizer.get("SCORE"), SynonymsCommon.FONTS.MEDIUM, fontSize, textColor))
  row:alignChildren()
  local col = Column.new(0, Alignment.LEFT)
  local topSpacerSize = math.max(marginW, Screen.safeTopInset)
  col:insertSpacer(topSpacerSize)
  col:insert(row)
  col:insert(FilledRectangle.new(Screen.width - marginW, 1, underlineColor))
  col:alignChildren()
  local header = FilledRectangle.new(Screen.width, col:getHeight(), SynonymsCommon.COLORS.WHITE)
  col:setAnchorPointBottomLeft()
  col:setLoc(marginW, 0)
  header:insert(col)
  return header
end

local function newSuccessFooter(score, onEndGame)
  local largeFontSize = 34
  local smallFontSize = largeFontSize * 0.55
  local rect = FilledRectangle.new(Screen.width, FOOTER_HEIGHT, SynonymsCommon.COLORS.BLACK)
  local col = Column.new(0)
  col:insertSpacer(smallFontSize * 0.8)
  col:insert(newLabel(Localizer.get("GAME_WON"), SynonymsCommon.FONTS.BOLD, largeFontSize, SynonymsCommon.COLORS.YELLOW))
  col:insert(newScoreLabel(tostring(score), Localizer.get("EARNED"), smallFontSize, SynonymsCommon.COLORS.WHITE))
  col:insertSpacer(smallFontSize * 0.8)
  local button = newContinueButton(SynonymsCommon.COLORS.YELLOW, onEndGame)
  col:insert(button)
  col:alignChildren()
  col:setAnchorPointTop()
  if (col:getHeight() > (FOOTER_HEIGHT - Screen.safeBottom)) then
    col:setEffectiveDims(nil, (FOOTER_HEIGHT - Screen.safeBottom))
  end
  col:setLoc(rect:getWidth() / 2, FOOTER_HEIGHT)
  rect:insert(col)
  return rect
end

local function newFailedFooter(score, onEndGame)
  local largeFontSize = 34
  local smallFontSize = largeFontSize * 0.55
  local rect = FilledRectangle.new(Screen.width, FOOTER_HEIGHT, SynonymsCommon.COLORS.BLACK)
  local col = Column.new(0)
  col:insertSpacer(smallFontSize * 0.8)
  col:insert(newLabel(Localizer.get("GAME_LOST"), SynonymsCommon.FONTS.BOLD, largeFontSize, SynonymsCommon.COLORS.RED))
  col:insert(newScoreLabel(tostring(score), Localizer.get("EARNED"), smallFontSize, SynonymsCommon.COLORS.WHITE))
  col:insertSpacer(smallFontSize * 0.8)
  local button = newContinueButton(SynonymsCommon.COLORS.RED, onEndGame)
  col:insert(button)
  col:alignChildren()
  col:setAnchorPointTop()
  if (col:getHeight() > (FOOTER_HEIGHT - Screen.safeBottom)) then
    col:setEffectiveDims(nil, (FOOTER_HEIGHT - Screen.safeBottom))
  end
  col:setLoc(rect:getWidth() / 2, FOOTER_HEIGHT)
  rect:insert(col)
  return rect
end

-- Constructor ------------------------

function SynonymsScoreCard.new(answers, onComplete)
  local self = Group.newFullScreen()
  Mixin.mix(self, SynonymsScoreCard)

  local score = getTotalScore(answers)
  local didPass = (score >= SynonymsConfiguration.getQuestionsCountToPass())

  local bg = FilledRectangle.new(self:getWidth(), self:getHeight(), {1, 1, 1})
  self:insertAtCenter(bg, Z_LOW)

  local width = Screen.getWidthPart(0.8)
  local xOffset = (self:getWidth() - width) / 2

  local header = newHeader(width)
  header:setAnchorPointTop()
  header:setLoc(self:getWidthPart(0.5), self:getHeight())
  self:insert(header, Z_HIGH)

  if didPass then
    self.resultsFooter = newSuccessFooter(score, onComplete)
  else
    self.resultsFooter = newFailedFooter(score, onComplete)
  end
  self.resultsFooter:setAnchorPointBottom()
  self.resultsFooter:setLoc(self:getWidthPart(0.5), -self.resultsFooter:getHeight())

  local scrollerVerticalOffset = self.resultsFooter:getHeight()
  local remainingY = self:getHeight() - header:getHeight() - scrollerVerticalOffset
  local scroller = SynonymsScoreScroller.new(width, remainingY, answers)
  scroller:setAnchorPointBottomLeft()
  scroller:setLoc(xOffset, scrollerVerticalOffset)
  self:insert(scroller, Z_LOW)
  scroller:setTouchEnabled(false)

  self.scroller = scroller
  self.xOffset = xOffset
  self.contentsIndex = 0
  self.header = header

  return self
end
