--- SynonymsScoreCell
-- @author Created by karl on February 11, 2015.

class("SynonymsScoreCell")

require("SynonymsCommon")
require("data.generators.ValueGenerators")
require("SynonymsScoreRectangle")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FONT_SIZE = 20.5
local SUGGESTION_FONT_SIZE = 15.5
local COMPLEXITY_FONT_SIZE = 15.5

local SUGGESTION_TAG = "suggestionPrefix"
local SUGGESTION_TAG_PATTERN = "<"..SUGGESTION_TAG..">%s</"..SUGGESTION_TAG..">"
local SHOULD_REDUCE_OPACITY_ATTRIBUTE = "shouldReduceOpacity"

local Z_INDEX = {
  BACKING = 10,
  CONTENT = 20
}

local SUGGESTION_BASE_ATTRIBUTES = {
  font = Font.new(SynonymsCommon.FONTS.BOLD, SUGGESTION_FONT_SIZE),
  color = SynonymsCommon.COLORS.DARK_GREY_3
}
local SUGGESTION_LABEL_ATTRIBUTES = {
  [SUGGESTION_TAG] = {
    font = Font.new(SynonymsCommon.FONTS.LIGHT, SUGGESTION_FONT_SIZE),
    color = {0, 0, 0},
    [SHOULD_REDUCE_OPACITY_ATTRIBUTE] = true
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getWordComplexity(score)
  local scoreTypes = {
    [0] = Localizer.get("INCORRECT"),
    [1] = Localizer.get("COMMON"),
    [2] = Localizer.get("UNCOMMON"),
    [3] = Localizer.get("EXPERT")
  }
  return scoreTypes[score]
end

local function newLabel(text, font, fontSize, color)
  local label = Label.new(text, Font.new(font, fontSize))
  label:setColor(unpack(color))
  return label
end

local function overrideDynamicTextAreaOpacityInheritanceForSuggestionPrefix(suggestionLabel)
  suggestionLabel:clearOpacityAttrLinkToChildren()
  local itemViews = suggestionLabel:getViewsWithAttribute(SHOULD_REDUCE_OPACITY_ATTRIBUTE, true)
  for _, item in ipairs(itemViews) do
    item:setOpacity(0.3)
  end
end

local function getSuggestionTextWithTags(concept, answer)
  local suggestions = SUGGESTION_GENERATOR:getMultipleSuggestions(concept, answer, 2)
  local suggestionPrefix = ""
  if (#suggestions == 1) then
    suggestionPrefix = Localizer.get("SUGGESTION_PREFIX")
  elseif (#suggestions > 1) then
    suggestionPrefix = Localizer.get("SUGGESTIONS_PREFIX")
  end

  local suggestionText = string.format(SUGGESTION_TAG_PATTERN, suggestionPrefix)
  for i, suggestion in ipairs(suggestions) do
    if (i > 1) then
      suggestionText = suggestionText .. ", "
    end
    suggestionText = suggestionText .. suggestion
  end
  return suggestionText
end

local function newSuggestionLabel(concept, answer, width)
  local suggestionText = getSuggestionTextWithTags(concept, answer)

  local suggestionLabel = DynamicTextArea.new(suggestionText, width, SUGGESTION_BASE_ATTRIBUTES, SUGGESTION_LABEL_ATTRIBUTES)
  overrideDynamicTextAreaOpacityInheritanceForSuggestionPrefix(suggestionLabel)

  return suggestionLabel
end

local function newScore(score, width)
  if (score == 0) then
    local cross = Sprite.newWithName("cross.png")
    cross:setEffectiveDims(width/2)
    cross:setColor(unpack(SynonymsCommon.COLORS.RED))
    return cross
  end

  return SynonymsScoreRectangleSet.new(width, score)
end

local function createRightColumn(score, scoreWidth, wasSkipped, isTagged)
  local rightColumn = Column.new(0, Alignment.RIGHT)
  rightColumn:setAnchorPointTopRight()
  local scoreGroup = newScore(score, scoreWidth)
  if (isTagged) then
    scoreGroup:setTag(SynonymsCommon.TAGS.SCORE)
  end
  rightColumn:insert(scoreGroup)
  local scoreText = getWordComplexity(score)
  if wasSkipped then
    scoreText = Localizer.get("SKIPPED")
  end
  local scoreTextLowercase = scoreText:lower()
  local scoreLabel = newLabel(scoreTextLowercase, SynonymsCommon.FONTS.BOLD, COMPLEXITY_FONT_SIZE, {0, 0, 0, 0.3})
  rightColumn:insert(scoreLabel)

  rightColumn.scoreLabel = scoreLabel
  rightColumn:alignChildren()

  return rightColumn
end

local function createLeftColumn(concept, answer, wasSkipped, isCorrect, suggestionLabelWidth)
  local leftColumn = Column.new(-FONT_SIZE * 0.1, Alignment.LEFT)
  leftColumn:setAnchorPointTopLeft()
  leftColumn:insert(newLabel(concept.prompt, SynonymsCommon.FONTS.LIGHT, FONT_SIZE, {0, 0, 0, 0.3}))
  if (not wasSkipped) then
    local answerLabel = newLabel(answer, SynonymsCommon.FONTS.BOLD, FONT_SIZE, SynonymsCommon.COLORS.DARK_GREY_3)
    if (not isCorrect) then
      local strikethroughWidth = answerLabel:getWidth() + 4
      local strikethrough = FilledRectangle.new(strikethroughWidth, 2, SynonymsCommon.COLORS.BLACK)
      strikethrough:setLoc(answerLabel:getWidthPart(0.5), answerLabel:getHeightPart(0.4))
      answerLabel:insert(strikethrough)
    end
    leftColumn:insert(answerLabel)
  else
    leftColumn:insertSpacer(FONT_SIZE * 0.4)
  end

  local suggestionLabel = newSuggestionLabel(concept, answer, suggestionLabelWidth)
  leftColumn:insert(suggestionLabel)
  leftColumn:alignChildren()

  return leftColumn
end

local function getTopElementOfColumn(column)
  local children = column:getChildren()
  if ((children ~= nil) and (#children > 0)) then
    return children[1]
  end
end

local function getCenterYOfColumnElement(self, column, element)
  local x, y = element:getLoc()
  local _, centerYOfElement = self:transformLoc(column, x, y)
  return centerYOfElement
end

local function alignTopElementsOfColumns(self, leftColumn, rightColumn)
  local topLeftElement = getTopElementOfColumn(leftColumn)
  local topRightElement = getTopElementOfColumn(rightColumn)
  if ((topLeftElement ~= nil) and (topRightElement ~= nil)) then
    local centerYOfTopLeftElement = getCenterYOfColumnElement(self, leftColumn, topLeftElement)
    local centerYOfTopRightElement = getCenterYOfColumnElement(self, rightColumn, topRightElement)
    local differenceInHeight = math.abs(centerYOfTopLeftElement - centerYOfTopRightElement)
    if (centerYOfTopLeftElement < centerYOfTopRightElement) then
      leftColumn:offsetY(differenceInHeight)
    else
      rightColumn:offsetY(differenceInHeight)
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsScoreCell.new(width, concept, answer, score, isTagged)
  local isCorrect = (score > 0)
  local wasSkipped = (answer == nil)
  local scoreWidth = width * 0.16

  local rightColumn = createRightColumn(score, scoreWidth, wasSkipped, isTagged)
  local suggestionLabelWidth = width
  if wasSkipped then
    suggestionLabelWidth = suggestionLabelWidth - rightColumn.scoreLabel:getWidth() - 10
  end
  local leftColumn = createLeftColumn(concept, answer, wasSkipped, isCorrect, suggestionLabelWidth)

  local height = math.max(leftColumn:getHeight(), rightColumn:getHeight())
  local self = Group.new(width, height)
  Mixin.mix(self, SynonymsScoreCell)

  local colBacking = FilledRectangle.new(Screen.getWidthPart(1.5), self:getHeight(), SynonymsCommon.COLORS.WHITE)
  self:insertAtCenter(colBacking, Z_INDEX.BACKING)
  self:insertAtTopLeft(leftColumn, Z_INDEX.CONTENT)
  self:insertAtTopRight(rightColumn, Z_INDEX.CONTENT)

  alignTopElementsOfColumns(self, leftColumn, rightColumn)

  return self
end
