--- SynonymsSuggestionLabel
-- @author Created by andrewc on May 03, 2021.

class("SynonymsSuggestionLabel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local FONT_SIZE = 33
local HORIZONTAL_PADDING = 10

local SUGGESTION_PREFIX = "SUGGESTION_PREFIX"
local HINT_PREFIX = "CORRECT_ANSWER_PREFIX"

local SUGGESTION_TAGS = "<word>%s</word>"

local SUGGESTION_COLOR = SynonymsCommon.COLORS.LIGHT_GREY_2
local WORD_COLOR = SynonymsCommon.COLORS.DARK_GREY_5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBaseAttributesWithColor(color)
  return {
    font = Font.new(SynonymsCommon.FONTS.LIGHT, FONT_SIZE),
    color = color
  }
end

local function getLabelAttributesWithColor(color)
  return {
    word = {
      font = Font.new(SynonymsCommon.FONTS.BOLD, FONT_SIZE),
      color = color
    }
  }
end

local function createLabel(text, suggestionColor, wordColor)
  local baseAttributes = getBaseAttributesWithColor(suggestionColor)
  local labelAttributes = getLabelAttributesWithColor(wordColor)
  local width = Screen.width * 3
  return DynamicTextArea.new(text, width, baseAttributes, labelAttributes, Alignment.CENTER)
end

local function setColor(self, color)
  if color then
    local text = self.label:getText()
    local scale = self.label:getScl()
    self.label:removeSelf()
    self.label = createLabel(text, color, color)
    self.label:setOpacity(0)
    self.label:setScl(scale)
    self:insertAtCenter(self.label)
  end
end

local function createAppearAnimation(self, delay)
  local scaleDuration = SynonymsCommon.TIMINGS.INCORRECT_SCALE_DURATION
  local scaleEase = SynonymsCommon.COMMON_EASE
  local scaleAmount = 0.85 * self.label:getScl()
  local scaleAnimation = AHScaleTo.new(self.label, scaleDuration, scaleEase, scaleAmount)

  local fadeDuration = SynonymsCommon.TIMINGS.INCORRECT_FADE_DURATION
  local fadeEase = AHEaseLinear.new()
  local fadeAmount = 1.0
  local fadeAnimation = AHOpacityTo.new(self.label, fadeDuration, fadeEase, fadeAmount)

  local moveDistance = SynonymsCommon.INCORRECT_MOVE_DISTANCE
  local moveAnimation = AHLocationYBy.new(self.label, scaleDuration, scaleEase, moveDistance)

  return AHDelayed.new(delay,
    AHSpawn.new(
      scaleAnimation,
      fadeAnimation,
      moveAnimation
  ))
end

local function createExitAnimation(self)
  local scaleUpFunction = function()
    wait(AHScaleTo.new(self.label, 0.1, AHEaseSineInOut.new(), self.label:getScl() * 1.2):start())
  end
  return AHSequence.new(
    scaleUpFunction,
    AHScaleTo.new(self.label, 0.3, AHEaseSineInOut.new(), 0)
  )
end

local function getTextWidth(self)
  local itemViews = self.label:getItemViews()
  local leftMostPoint = itemViews[1]:getXLeft()
  local rightMostPoint = itemViews[#itemViews]:getXRight()
  return rightMostPoint - leftMostPoint
end

local function setTextWithPrefixAndSuggestion(self, prefixText, suggestionText)
  local suggestionWithTags = string.format(SUGGESTION_TAGS, suggestionText)
  local text = prefixText .. " " .. suggestionWithTags
  self.label:setText(text)
  local textWidth = getTextWidth(self)
  if (textWidth > self.maximumTextWidth) then
    self.label:setScl(self.maximumTextWidth / textWidth)
  end
end

local function updateLabelTextForAnswerWithHint(self, submittedAnswer)
  prefixText = Localizer.get(HINT_PREFIX)
  SUGGESTION_GENERATOR:cacheSuggestionForAnswer(self.concept, self.correctAnswer, submittedAnswer)
  setTextWithPrefixAndSuggestion(self, prefixText, self.correctAnswer)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsSuggestionLabel:animateIncorrect()
  setColor(self, self.incorrectColor)
  local appearDelay = 0.2 + SynonymsCommon.TIMINGS.INCORRECT_FADE_DELAY
  return createAppearAnimation(self, appearDelay):start()
end

function SynonymsSuggestionLabel:setIncorrectColor(color)
  self.incorrectColor = color
end

function SynonymsSuggestionLabel:animateSkip(color)
  setColor(self, color)
  self.label:offsetY(-Screen.height * 0.05)

  local appearDelay = 0.326
  return createAppearAnimation(self, appearDelay):start()
end

function SynonymsSuggestionLabel:animateExit()
  return createExitAnimation(self):start()
end

function SynonymsSuggestionLabel:updateLabelTextForAnswer(mode, submittedAnswer, isCorrect)
  if isCorrect then
    SUGGESTION_GENERATOR:markAnswersAsUsed(self.concept, {submittedAnswer})
  else
    if (mode == SynonymsCommon.MODES.HINT) then
      updateLabelTextForAnswerWithHint(self, submittedAnswer)
    else
      local suggestionText = SUGGESTION_GENERATOR:getSuggestion(self.concept, submittedAnswer)
      if (suggestionText == nil) then
        suggestionText = Localizer.get("NO_SUGGESTION")
      end
      local prefixText = Localizer.get(SUGGESTION_PREFIX)
      setTextWithPrefixAndSuggestion(self, prefixText, suggestionText)
    end
  end
end

function SynonymsSuggestionLabel:setCorrectAnswer(correctAnswer)
  self.correctAnswer = correctAnswer
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsSuggestionLabel.new(concept)
  local self = Group.new(Screen.width, FONT_SIZE)
  Mixin.mix(self, SynonymsSuggestionLabel)

  self.label = createLabel("suggestion", SUGGESTION_COLOR, WORD_COLOR)
  self:insertAtCenter(self.label)

  self:setHeight(self.label:getHeight())

  self.label:setOpacity(0)

  self.concept = concept
  self.correctAnswer = nil
  self.incorrectColor = nil
  self.maximumTextWidth = Screen.safeWidth - HORIZONTAL_PADDING

  return self
end
