--- SynonymsTutorial
-- @author Created by karl on February 12, 2015.


require("SynonymsCommon")
require("widgets.TapToContinue")

local ORIENTATION_UP = "up"
local ORIENTATION_DOWN = "down"
local ARROW_SPACING = Screen.getHeightPart(0.06)

-- ============================================================================
class("SynonymsTutorialOverlay")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_ARROW = SynonymsCommon.COLORS.BLACK
local COLOR_BG = SynonymsCommon.COLORS.BLACK
local FADE_DURATION = 0.5


local function getAnchorY(prop)
  local _, anchorY = prop:getAnchorPoint()
  return anchorY
end

local function getCenterTransformed(self, prop, anchorY)
  anchorY = anchorY or 0.5
  local x, y = prop:getLocationWithAnchorPoint(0.5, anchorY)
  return self:transformLoc(prop.parent, x, y)
end

local function getOrientation(self, prop)
  local x, y = getCenterTransformed(self, prop)
  if y >= self:getHeightPart(0.5) then
    return ORIENTATION_UP
  end
  return ORIENTATION_DOWN
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsTutorialOverlay:setOrientation(orientation)
  if orientation == ORIENTATION_UP then
    self:setAnchorPointTop()
  else
    self:setAnchorPointBottom()
  end
end

function SynonymsTutorialOverlay:pointTo(pointToGroup)
  local x = getCenterTransformed(self, pointToGroup)
  self.upArrow:setX(x)
  self.downArrow:setX(x)
end

function SynonymsTutorialOverlay:showArrow()
  return AHSpawn.new(
    AHOpacityTo.new(self.upArrow, FADE_DURATION, nil, 1),
    AHOpacityTo.new(self.downArrow, FADE_DURATION, nil, 1)
  ):start()
end

function SynonymsTutorialOverlay:hideArrow()
  return AHSpawn.new(
    AHOpacityTo.new(self.upArrow, FADE_DURATION, nil, 0),
    AHOpacityTo.new(self.downArrow, FADE_DURATION, nil, 0)
  ):start()
end

function SynonymsTutorialOverlay:showScreen(screen)
  local _, topY = self:modelToWorld(self.screenArea:getLoc())
  local maxScreenHeight = topY + self.screenArea:getHeightPart(0.5) - self.minimumContentHeight
  if (screen:getHeight() > maxScreenHeight) then
    screen:setEffectiveDims(nil, maxScreenHeight)
  end

  self.screenArea:removeAll()
  local anchorY = getAnchorY(self)
  screen:setAnchorPoint(0.5, anchorY)
  screen:setLoc(self.screenArea:getWidthPart(0.5), self.screenArea:getHeightPart(anchorY))
  self.screenArea:insert(screen)
  self.screen = screen
  return AHOpacityTo.new(self.cover, FADE_DURATION, nil, 0):start()
end

function SynonymsTutorialOverlay:hideScreen()
  local action = AHDoNothing.new()
  if self.screen then
    action = AHOpacityTo.new(self.cover, FADE_DURATION, nil, 1)
  end
  return action:start()
end

function SynonymsTutorialOverlay:setMinimumContentHeight(minimumContentHeight)
  self.minimumContentHeight = minimumContentHeight
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newArrow()
  local width, height = Screen.getWidthPart(0.1), Screen.getWidthPart(0.12)
  local headHeight = height * 0.4
  local tailHeight = height - headHeight
  local tailWidth = width * 0.5

  local arrow = Group.new(width, height)
  local triangle = FilledTriangle.new(0, 0, width, 0, width / 2, headHeight)
  local rectangle = FilledRectangle.new(tailWidth, tailHeight)
  triangle:setAnchorPointTop()
  triangle:setLoc(width / 2, height)
  arrow:insert(triangle)
  rectangle:setAnchorPointBottom()
  rectangle:setLoc(width / 2, 0)
  arrow:insert(rectangle)
  arrow:linkColorAttrToChildren()

  arrow:setColor(unpack(COLOR_ARROW))
  return arrow
end

-- Constructor ------------------------

function SynonymsTutorialOverlay.new()
  local extraSpacing = 8
  local upArrow = newArrow()
  local downArrow = newArrow()
  downArrow:setScl(1, -1)

  local self = Group.new(Screen.width, Screen.height + (upArrow:getHeight() + ARROW_SPACING + extraSpacing) * 2)
  Mixin.mix(self, SynonymsTutorialOverlay)

  local bg = FilledRectangle.new(Screen.width, Screen.height, COLOR_BG)
  self:insertAtCenter(bg)

  upArrow:setAnchorPointTop()
  upArrow:setLoc(self:getWidthPart(0.5), self:getHeight() - ARROW_SPACING)
  upArrow:setOpacity(0)
  self:insert(upArrow)
  self.upArrow = upArrow

  downArrow:setAnchorPointTop()
  downArrow:setLoc(self:getWidthPart(0.5), ARROW_SPACING)
  downArrow:setOpacity(0)
  self:insert(downArrow)
  self.downArrow = downArrow

  self.screenArea = Group.new(Screen.width, Screen.getHeightPart(0.95))
  self:insertAtCenter(self.screenArea)

  self.cover = FilledRectangle.new(bg:getWidth(), bg:getHeight(), COLOR_BG)
  self:insertAtCenter(self.cover)

  self.minimumContentHeight = 0

  return self
end


-- ============================================================================
class("SynonymsTutorial")
-- ============================================================================

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setOrientation(self, orientation)
  self.overlay:setOrientation(orientation)
  local anchorY = 0
  local tapLabelY = math.max(Screen.getHeightPart(0.05), Screen.getSafeY(0.02))
  if orientation == ORIENTATION_DOWN then
    anchorY = 1
    tapLabelY = self:getHeight() - tapLabelY
  end
  self.tapLabel:setAnchorPoint(0.5, anchorY)
  self.tapLabel:setLoc(self:getWidthPart(0.5), tapLabelY)
  self.currentOrientation = orientation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsTutorial:slideOut()
  return runInCoroutine(function()
    local duration = 0.5
    local anchorY = getAnchorY(self.overlay)
    local y = self:getHeightPart(1 - anchorY)
    if self.overlay:getY() ~= y then
      wait(AHLocationYTo.new(self.overlay, duration, AHEaseExponentialIn.new(), y):start())
    end
  end)
end

function SynonymsTutorial:waitForTap(onTap)
  local tapToContinue
  tapToContinue = TapToContinue.new(nil, nil, function()
    tapToContinue:removeSelf()
    onTap()
  end)
end

function SynonymsTutorial:endTutorial()
  return runInCoroutine(function()
    wait(self:hideScreen(), self:hideArrow())
    wait(self:slideOut())
  end)
end

function SynonymsTutorial:moveBy(deltaY)
  return AHLocationYBy.new(self.overlay, 0.8, AHEaseExponentialInOut.new(), deltaY):start()
end

function SynonymsTutorial:pointArrowTo(pointToGroup)
  self.overlay:pointTo(pointToGroup)
end

function SynonymsTutorial:pointTo(pointToGroup)
  return runInCoroutine(function()
    local newOrientation = getOrientation(self, pointToGroup)
    if not self.currentOrientation then
      self.overlay:setOrientation(newOrientation)
      local anchorY = getAnchorY(self.overlay)
      self.overlay:setY(self:getHeightPart(1 - anchorY))
    elseif self.currentOrientation ~= newOrientation then
      self.currentOrientation = newOrientation
      if self.overlay:getY() == 0 or self.overlay:getY() == self:getHeight() then
        wait(self:slideOut())
      end
      local anchorY = getAnchorY(self.overlay)
      self.overlay:setY(self:getHeightPart(anchorY))
    end
    setOrientation(self, newOrientation)
    self:pointArrowTo(pointToGroup)
    local anchorY = getAnchorY(self.overlay)
    local _, y = getCenterTransformed(self, pointToGroup, 1 - anchorY)
    wait(AHLocationYTo.new(self.overlay, 0.8, AHEaseExponentialInOut.new(), y):start())
  end)
end

function SynonymsTutorial:showScreen(screen)
  self.overlay:showScreen(screen)
  return AHOpacityTo.new(self.tapLabel, FADE_DURATION, nil, 1):start()
end

function SynonymsTutorial:hideScreen()
  self.overlay:hideScreen()
  return AHOpacityTo.new(self.tapLabel, FADE_DURATION, nil, 0):start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsTutorial.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, SynonymsTutorial)

  self.overlay = SynonymsTutorialOverlay.new()
  self.overlay:setAnchorPointTop()
  self.overlay:setLoc(self:getWidthPart(0.5), 0)
  self:insert(self.overlay)

  self.tapLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(SynonymsCommon.FONTS.LIGHT, 15))
  self.tapLabel:setOpacity(0)
  self.tapLabel:setAnchorPointBottom()
  local tapLabelY = math.max(Screen.getHeightPart(0.05), Screen.getSafeY(0.02))
  self.tapLabel:setLoc(self:getWidthPart(0.5), tapLabelY)
  self:insert(self.tapLabel)

  self.overlay:setMinimumContentHeight(self.tapLabel:getYTop())

  FunctionUtils.forwardMethodCalls(self, self.overlay, "showArrow", "hideArrow")

  return self
end
