--- SynonymsUI
-- @author Created by karl on January 07, 2015.

class("SynonymsUI")

require("SynonymsHelperBank")
require("SynonymsCommon")
require("SynonymsNormalAnswerMode")
require("SynonymsHintAnswerMode")
require("SynonymsMultiAnswerMode")
require("SynonymsPickAnswerMode")
require("SynonymsTimeUpSequence")
require("SynonymsScoreCard")
require("SynonymsRationale")
require("SynonymsRoundResponse")
require("DeviceKeyboard")
require("widgets.KeyboardWarning")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDICES = {
  HELPER_BUTTONS = 0,
  CIRCLE_OVERLAY = 10,
  SKIP = 20,
  ANSWER_MODE = 30,
  SCORE_CARD = 40,
  TIMEOUT = 50,
  RATIONALE = 60,
  TRANSITION_EFFECTS = 70
}


local VALID_KEYS = Localizer.get("VALID_KEYS")
local MAX_INPUT_LENGTH = 30

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function containsConsecutiveWhitespace(text)
  return (text:find("%s%s") ~= nil)
end

local function containsPrintableCharacterAfterWhitespace(text)
  return (text:find("%s%g") ~= nil)
end

local function onAnswerSubmitted(self, answer, ignoreRationale)
  return self.onAnswerSubmittedCallback(answer, ignoreRationale)
end

local function onRoundComplete(self, roundResponse)
  self:setTouchEnabled(false)
  runInCoroutine(function()
    wait(self.answerMode:animateAnswer(roundResponse))
    self.onRoundCompleteCallback()
  end)
end

local function onSkip(self)
  self:setTouchEnabled(false)
  runInCoroutine(function()
    self.onSkipCallback()
  end)
end

local function getKeyboardTopY(self)
  return self.keyboardTopY or self:getHeightPart(0.4)
end

local function animateHelperButtonsIn(self)
  self.keyboardActionQueue:insert(function()
    waitFor(0.2) -- Give keyboard position time to settle.
    local keyboardTopY = getKeyboardTopY(self)
    if keyboardTopY > 0 then
      wait(AHLocationYTo.new(self.helperButtons, 0.8, AHEaseExponentialOut.new(), keyboardTopY):start())
    end
  end)
end

local function repositionHelperButtons(self)
  self.keyboardActionQueue:insert(AHLocationYTo.new(self.helperButtons, 0.7, AHEaseSineInOut.new(), self.keyboardTopY))
end

local function slideKeyboardIn(self)
  DeviceKeyboard.show()
end

local function slideKeyboardOut(self)
  DeviceKeyboard.hide()
end

local function showCircleOverlay(self, button, color)
  local duration = 0.35
  local startRadius = Screen.getWidthPart(0.1)
  local endRadius = math.sqrt(Screen.width * Screen.width + (Screen.height * Screen.height) / 2)
  local endScale = endRadius / startRadius
  self.circleOverlay = FilledCircle.new(startRadius, color)
  self.circleOverlay:setLoc(self:transformObjectLoc(button))

  self:insert(self.circleOverlay, Z_INDICES.CIRCLE_OVERLAY)

  return AHScaleTo.new(self.circleOverlay, duration, AHEaseSineIn.new(), endScale):start()
end

local function replaceAnswerMode(self, newAnswerMode)
  if self.answerMode then
    self.answerMode:prepareForRemoval()
    self.answerMode:removeSelf()
  end
  self.answerMode = newAnswerMode
  self:insertAtCenter(self.answerMode, Z_INDICES.ANSWER_MODE)
  self.answerMode:animateIn()
end

local function setModeInternal(self, setModeFunction)
  local newModeCoroutine = runInCoroutine(setModeFunction)
  self.setModeInterrupt:insert(newModeCoroutine)

  wait(newModeCoroutine)
end

local function setModeNormal(self)
  local setModeFunction = function()
    local oldAnswerMode = self.answerMode
    local newAnswerMode = SynonymsNormalAnswerMode.new(self.concept, function(answer)
      local roundResponse = onAnswerSubmitted(self, answer)
      local isCorrect = roundResponse:getIsCorrect()
      local isIncorrectPartOfSpeech = roundResponse:getIsIncorrectPartOfSpeech()
      if (not isCorrect and not isIncorrectPartOfSpeech) then
        self.timer:pauseAndShowIcon()
      end
      if (not isIncorrectPartOfSpeech) then
        onRoundComplete(self, roundResponse)
      end
    end)

    self:insertAtCenter(newAnswerMode, Z_INDICES.CIRCLE_OVERLAY - 1)
    if oldAnswerMode then
      local duration = 0.2
      local action = AHSpawn.new(
        AHRecursive.new(AHOpacityTo.new, oldAnswerMode, duration, nil, 0)
      )
      if self.circleOverlay then
        action:addAction(AHRecursive.new(AHOpacityTo.new, self.circleOverlay, duration, nil, 0))
      end
      wait(action:start())
    end

    if oldAnswerMode then
      oldAnswerMode:removeSelf()
    end
    if self.circleOverlay then
      self.circleOverlay:removeSelf()
    end
    newAnswerMode:setZIndex(Z_INDICES.CIRCLE_OVERLAY)
    slideKeyboardIn(self)
    animateHelperButtonsIn(self)
    newAnswerMode:animateIn()
    self.helperButtons:retakeSkipButton()

    self.answerMode = newAnswerMode
    self.helperButtons:reset()
  end

  setModeInternal(self, setModeFunction)
end

local function setModeHint(self)
  local setModeFunction = function()
    replaceAnswerMode(self, SynonymsHintAnswerMode.new(self.concept, function(answer)
      local roundResponse = onAnswerSubmitted(self, answer)
      local isCorrect = roundResponse:getIsCorrect()
      if (not isCorrect) then
        self.timer:pauseAndShowIcon()
      end
      local isIncorrectPartOfSpeech = roundResponse:getIsIncorrectPartOfSpeech()
      if (not isIncorrectPartOfSpeech) then
        onRoundComplete(self, roundResponse)
      end
    end))
  end

  setModeInternal(self, setModeFunction)
  self.helperButtons:relinquishSkipButton(self, Z_INDICES.SKIP)
end

local function setModePick(self)
  local setModeFunction = function()
    slideKeyboardOut(self)
    replaceAnswerMode(self, SynonymsPickAnswerMode.new(self.concept, function(answer)
      local roundResponse = onAnswerSubmitted(self, answer)
      onRoundComplete(self, roundResponse)
    end))
  end

  setModeInternal(self, setModeFunction)
end

local function preloadModeMulti(self)
  local allAnswers = {}
  local submitHandler = function(answer, ignoreRationale)
    local roundResponse = onAnswerSubmitted(self, answer, ignoreRationale)
    table.insert(allAnswers, roundResponse:getAnswer())
    return roundResponse:getIsIncorrectPartOfSpeech()
  end
  local completeHandler = function()
    SUGGESTION_GENERATOR:markAnswersAsUsed(self.concept, allAnswers)
    onRoundComplete(self)
  end

  self.preloadedAnswerMode = SynonymsMultiAnswerMode.new(self.concept, submitHandler, completeHandler, self.isRepeatAnswerCallback, self.isCorrectAnswerCallback)
end

local function getTransitionEffectToModeMulti(self)
  return function()
    self.answerMode:setKeyboardEnabled(false)
    DeviceKeyboard.setEnabled(false)

    local promptLabel = self.answerMode:getPromptLabel()
    local promptLabelClone = promptLabel:createClone(promptLabel:getText())
    promptLabelClone:setLoc(self:transformObjectLoc(promptLabel))
    self:insert(promptLabelClone, Z_INDICES.TRANSITION_EFFECTS)

    local inputField = self.answerMode:getInputField()
    local inputFieldLabelClone = inputField:createClone(inputField:getText())
    inputFieldLabelClone:setLoc(self:transformObjectLoc(inputField))
    inputFieldLabelClone:setScl(inputField:getScl())
    self:insert(inputFieldLabelClone, Z_INDICES.TRANSITION_EFFECTS)

    -- preload the next mode so that we may access its ui element positions as animation destination positions
    preloadModeMulti(self)
    local destinationPromptLabel = self.preloadedAnswerMode:getPromptLabel()
    local _, destinationPromptLabelY = self:transformObjectLoc(destinationPromptLabel)
    local destinationInputField = self.preloadedAnswerMode:getInputField()
    local _, destinationInputFieldY = self:transformObjectLoc(destinationInputField)

    local slideDuration = 0.3
    local colorDuration = 0.15
    local ease = AHEaseSimpleInOut.new()
    local combinedAction = AHSequence.new(
                             function()
                               promptLabel:hide()
                               inputField:hide()
                             end,
                             AHSequence.new(
                               AHSpawn.new(
                                 AHLocationYTo.new(promptLabelClone, slideDuration, ease, destinationPromptLabelY),
                                 AHColorTo.new(promptLabelClone, colorDuration, nil, SynonymsCommon.COLORS.WHITE),
                                 AHLocationYTo.new(inputFieldLabelClone, slideDuration, ease, destinationInputFieldY)
                               ),
                               AHSpawn.new(
                                 AHRemove.new(promptLabelClone),
                                 AHRemove.new(inputFieldLabelClone)
                               )
                             )
                           )

    local delayToEnsureClonesVisibleBeforeRunningActionInFrames = 1
    waitForFrames(delayToEnsureClonesVisibleBeforeRunningActionInFrames)

    combinedAction:start()
  end
end

local function setModeMulti(self)
  local setModeFunction = function()
    replaceAnswerMode(self, self.preloadedAnswerMode)

    local preexistingText = DeviceKeyboard.getText()
    if ((preexistingText ~= nil) and (preexistingText ~= "")) then
      self.answerMode.inputField:setText(string.lower(preexistingText))
    end
  end

  setModeInternal(self, setModeFunction)
end

local function showScoreCard(self, timeUpScreen, answers, shouldShowTutorial, onFinished)
  return runInCoroutine(function()
    local yDistance = self:getHeight()
    local card = SynonymsScoreCard.new(answers, onFinished)
    card:setAnchorPointBottom()
    card:setLoc(self:getWidthPart(0.5), yDistance)
    self:insert(card, Z_INDICES.SCORE_CARD)
    local ease = AHEaseSineInOut.new()
    local duration = 0.5
    wait(AHSpawn.new(
      AHLocationYBy.new(timeUpScreen, duration, ease, -yDistance),
      AHLocationYBy.new(card, duration, ease, -yDistance)
    ):start())
    if shouldShowTutorial then
      card:showTutorial()
    else
      card:showCard()
    end

    self:setTouchEnabled(true)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsUI:stop()
  self.running = false
  self.answerMode:setTouchEnabled(false)
  self.setModeInterrupt:stop()
  slideKeyboardOut(self)
end

function SynonymsUI:setTouchEnabled(isEnabled)
  self.helperButtons:setTouchEnabled(isEnabled)
  if self.answerMode then
    self.answerMode:setTouchEnabled(isEnabled)
  end
end

function SynonymsUI:setConcept(concept)
  if self.running then
    DeviceKeyboard.clear()
    self.concept = concept
    self.setModeCoroutine = setModeNormal(self)

    self:setTouchEnabled(true)
  end
end

function SynonymsUI:runEndSequence(answers, shouldShowTutorial, onFinished)
  return runInCoroutine(function()
    local timeUpSequence = SynonymsTimeUpSequence.new()
    self:insertAtCenter(timeUpSequence, Z_INDICES.TIMEOUT)
    timeUpSequence:setOpacity(0)

    local fade = AHOpacityTo.new(timeUpSequence, 0.3, nil, 1)
    fade:start()

    local yDistance = getKeyboardTopY(self) + self.helperButtons:getHeight()
    local ease = AHEaseSineOut.new()
    local duration = 0.3
    local slide = AHLocationYBy.new(self.helperButtons, duration, ease, -yDistance)
    slide:start()

    wait(timeUpSequence:start())

    if #answers > 0 then
      showScoreCard(self, timeUpSequence, answers, shouldShowTutorial, onFinished)
    else
      onFinished()
    end
  end)
end

function SynonymsUI:animateSkip()
  return self.answerMode:animateSkip()
end

function SynonymsUI:animateAutocorrect(answer)
  self.answerMode:animateAutocorrect(answer)
end

function SynonymsUI:showAndWaitForRationale(answer, isNoun, isVerb)
  self.helperButtons:setTouchEnabled(false)
  local isPaused = true
  local fadeDuration = 0.5
  self.answerMode.inputField:toggleBlink(isPaused)
  DeviceKeyboard.clear()
  DeviceKeyboard.hide()
  self.answerMode:setTouchEnabled(false)

  local rationale = SynonymsRationale.new(answer, isNoun, isVerb)
  self:insertAtCenter(rationale, Z_INDICES.RATIONALE)

  local fadeInAction = AHOpacityTo.new(rationale, fadeDuration, nil, 1.0)
  wait(fadeInAction:start())

  local tapToContinue = TapToContinue.new()
  tapToContinue:waitForTrigger()

  tapToContinue:removeSelf()
  DeviceKeyboard.show()
  local fadeOutAction = AHOpacityTo.new(rationale, fadeDuration, nil, 0.0)
  wait(fadeOutAction:start())
  rationale:removeSelf()
  isPaused = false
  self.answerMode.inputField:toggleBlink(isPaused)
  self.helperButtons:setTouchEnabled(true)
  self.answerMode:setTouchEnabled(true)
end

function SynonymsUI:getActiveMode()
  return self.answerMode:getMode()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newHelperBankFunction(self, shouldRemoveKeyboard, extraDelay, transitionStartCallback, transitionCompleteCallback)
  return function(button, color)
    if (transitionStartCallback ~= nil) then
      transitionStartCallback()
    end

    showCircleOverlay(self, button, color)

    if shouldRemoveKeyboard then
      AHDelayed.new(0.1, function() slideKeyboardOut(self) end):start()
    end
    AHDelayed.new(0.1 + extraDelay, transitionCompleteCallback):start()
  end
end

-- Constructor ------------------------

function SynonymsUI.new(timer, onAnswerSubmittedCallback, onRoundCompleteCallback, onSkipCallback, isRepeatAnswerCallback, isCorrectAnswerCallback)
  local self = FilledRectangle.newFullScreen(SynonymsCommon.COLORS.WHITE)
  Mixin.mix(self, SynonymsUI)

  self.onAnswerSubmittedCallback = onAnswerSubmittedCallback
  self.onRoundCompleteCallback = onRoundCompleteCallback
  self.onSkipCallback = onSkipCallback
  self.isRepeatAnswerCallback = isRepeatAnswerCallback
  self.isCorrectAnswerCallback = isCorrectAnswerCallback

  local baseValidateCallback = DeviceKeyboard.createCharacterBasedValidationCallback(VALID_KEYS)

  DeviceKeyboard.setValidateCallback(function(replacementText, resultingText)
    local enteredInvalidCharacter = not baseValidateCallback(replacementText)
    local willExceedMaxInputLength = (#resultingText > MAX_INPUT_LENGTH)
    local enteredLeadingWhitespace = (resultingText == " ")
    local enteredExcessiveTrailingWhitespace = containsConsecutiveWhitespace(resultingText)
    local enteredSecondWord = containsPrintableCharacterAfterWhitespace(resultingText)

    if (enteredInvalidCharacter or
        willExceedMaxInputLength or
        enteredLeadingWhitespace or
        enteredExcessiveTrailingWhitespace or
        enteredSecondWord) then
      return false
    end

    return true
  end)

  DeviceKeyboard.setWillAppearWithDimensionsCallback(function(originXProportion, originYProportion, widthProportion, heightProportion)
    local oldTopY = self.keyboardTopY or 0
    self.keyboardTopY = self:getHeightPart(originYProportion) + self:getHeightPart(heightProportion)
    if self.keyboardTopY > oldTopY then
      repositionHelperButtons(self)
    end
  end)

  local showSingleWordWarning = function()
    timer:pause()
    DeviceKeyboard.hide()
    KeyboardWarning.newSingleWordWarning():show(function()
      DeviceKeyboard.show()
      timer:resume()
    end)
  end

  DeviceKeyboard.setInputRejectedCallback(function(replacementText, resultingText)
    if containsPrintableCharacterAfterWhitespace(resultingText) then
      showSingleWordWarning()
    end
  end)

  self.helperButtons = SynonymsHelperBank.new(
    newHelperBankFunction(self, false, 0, getTransitionEffectToModeMulti(self), function() setModeMulti(self) end),
    newHelperBankFunction(self, false, 0.07, nil, function() setModeHint(self) end),
    newHelperBankFunction(self, true, 0.01, nil, function() setModePick(self) end),
    function() self.helperButtons:retakeSkipButton() onSkip(self) end
  )
  self.helperButtons:setAnchorPointBottomLeft()
  self.helperButtons:setLoc(Screen.safeLeft, -self.helperButtons:getHeight())
  self:insert(self.helperButtons, Z_INDICES.HELPER_BUTTONS)

  self.running = true
  self.setModeInterrupt = ActionInterrupt.new()

  self.keyboardActionQueue = ActionQueue.new()

  self.timer = timer

  return self
end
