--- SynonymsValidator
-- @author Created by karl on January 05, 2015.
-- ConceptValidator for Synonyms.

class("SynonymsValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "prompt",
  "basic",
  "medium",
  "advanced",
  "suggestions",
  "hints",
  "distractors",
  "correct",
  "difficulty"
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validatePromptIsNotInOtherSets(self, prompt, basic, medium, advanced, suggestions, hints, correct, distractors, nouns, verbs)
  self:expectValueIsNotInSet(prompt, basic, "prompt", "basic")
  self:expectValueIsNotInSet(prompt, medium, "prompt", "medium")
  self:expectValueIsNotInSet(prompt, advanced, "prompt", "advanced")
  self:expectValueIsNotInSet(prompt, suggestions, "prompt", "suggestions")
  self:expectValueIsNotInSet(prompt, hints, "prompt", "hints")
  self:expectValueIsNotInSet(prompt, correct, "prompt", "correct")
  self:expectValueIsNotInSet(prompt, distractors, "prompt", "distractors")
  self:expectValueIsNotInSet(prompt, nouns, "prompt", "nouns")
  self:expectValueIsNotInSet(prompt, verbs, "prompt", "verbs")
end

local function validateDistractorsAreNotInOtherSets(self, distractors, basic, medium, advanced, suggestions, hints, correct, nouns, verbs)
  self:expectValuesAreNotInSet(distractors, basic, "distractors", "basic")
  self:expectValuesAreNotInSet(distractors, medium, "distractors", "medium")
  self:expectValuesAreNotInSet(distractors, advanced, "distractors", "advanced")
  self:expectValuesAreNotInSet(distractors, suggestions, "distractors", "suggestions")
  self:expectValuesAreNotInSet(distractors, hints, "distractors", "hints")
  self:expectValuesAreNotInSet(distractors, correct, "distractors", "correct")
  self:expectValuesAreNotInSet(distractors, nouns, "distractors", "nouns")
  self:expectValuesAreNotInSet(distractors, verbs, "distractors", "verbs")
end

local function validateSuggestionsHintsCorrectAreInRightSets(self, basic, medium, advanced, suggestions, hints, correct)
  local basicOrMedium = TableUtils.unionSet(basic, medium)
  local mediumOrAdvanced = TableUtils.unionSet(medium, advanced)
  self:expectValuesAreInSet(suggestions, mediumOrAdvanced, "suggestions",  "medium or advanced")
  self:expectValuesAreInSet(hints, basicOrMedium, "hints",  "basic or medium")
  self:expectValuesAreInSet(correct, mediumOrAdvanced, "correct",  "medium or advanced")
end

local function validateBasicMediumAdvancedNounsVerbsAreDisjoint(self, basic, medium, advanced, nouns, verbs)
  self:expectValuesAreNotInSet(basic, medium, "basic", "medium")
  self:expectValuesAreNotInSet(basic, advanced, "basic", "advanced")
  self:expectValuesAreNotInSet(medium, advanced, "medium", "advanced")

  self:expectValuesAreNotInSet(nouns, verbs, "nouns", "verbs")
  self:expectValuesAreNotInSet(nouns, basic, "nouns", "basic")
  self:expectValuesAreNotInSet(nouns, medium, "nouns", "medium")
  self:expectValuesAreNotInSet(nouns, advanced, "nouns", "advanced")
  self:expectValuesAreNotInSet(verbs, basic, "verbs", "basic")
  self:expectValuesAreNotInSet(verbs, medium, "verbs", "medium")
  self:expectValuesAreNotInSet(verbs, advanced, "verbs", "advanced")
end

local function validateAllArraysHaveNoDuplicates(self, basic, medium, advanced, suggestions, hints, correct, distractors, nouns, verbs)
  self:expectNoDuplicates(basic, "basic")
  self:expectNoDuplicates(medium, "medium")
  self:expectNoDuplicates(advanced, "advanced")
  self:expectNoDuplicates(suggestions, "suggestions")
  self:expectNoDuplicates(hints, "hints")
  self:expectNoDuplicates(correct, "correct")
  self:expectNoDuplicates(distractors, "distractors")
  self:expectNoDuplicates(nouns, "nouns")
  self:expectNoDuplicates(verbs, "verbs")
end

local function validateValuesHaveLowercaseLettersOnly(self, valuesSet, valuesName)
  for value, _ in pairs(valuesSet) do
    local ignoredCharacters = "-"
    local fieldName = string.format("%s (word: '%s')", valuesName, value)
    self:expectStringIsLettersOnly(value, fieldName, ignoredCharacters)

    local lowercase = string.lower(value)
    self:expectEqual(value, lowercase, "Word is not lowercase")
  end
end

local function validateAllSetsHaveLowercaseLettersOnly(self, prompt, basic, medium, advanced, suggestions, hints, correct, distractors, nouns, verbs)
  self:expectStringIsLettersOnly(prompt, "prompt")
  validateValuesHaveLowercaseLettersOnly(self, basic, "basic")
  validateValuesHaveLowercaseLettersOnly(self, medium, "medium")
  validateValuesHaveLowercaseLettersOnly(self, advanced, "advanced")
  validateValuesHaveLowercaseLettersOnly(self, suggestions, "suggestions")
  validateValuesHaveLowercaseLettersOnly(self, hints, "hints")
  validateValuesHaveLowercaseLettersOnly(self, correct, "correct")
  validateValuesHaveLowercaseLettersOnly(self, distractors, "distractors")
  validateValuesHaveLowercaseLettersOnly(self, nouns, "nouns")
  validateValuesHaveLowercaseLettersOnly(self, verbs, "verbs")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function SynonymsValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local prompt = concept.prompt

  local basicArray = TableUtils.csvToArray(concept.basic)
  local mediumArray = TableUtils.csvToArray(concept.medium)
  local advancedArray = TableUtils.csvToArray(concept.advanced)
  local suggestionsArray = TableUtils.csvToArray(concept.suggestions)
  local hintsArray = TableUtils.csvToArray(concept.hints)
  local correctArray = TableUtils.csvToArray(concept.correct)
  local distractorsArray = TableUtils.csvToArray(concept.distractors)
  local nounsArray = TableUtils.csvToArray(concept.nouns)
  local verbsArray = TableUtils.csvToArray(concept.verbs)
  validateAllArraysHaveNoDuplicates(self, basicArray, mediumArray, advancedArray, suggestionsArray, hintsArray, correctArray, distractorsArray, nounsArray, verbsArray)

  local basic = TableUtils.createSetFromArray(basicArray)
  local medium = TableUtils.createSetFromArray(mediumArray)
  local advanced = TableUtils.createSetFromArray(advancedArray)
  local suggestions = TableUtils.createSetFromArray(suggestionsArray)
  local hints = TableUtils.createSetFromArray(hintsArray)
  local correct = TableUtils.createSetFromArray(correctArray)
  local distractors = TableUtils.createSetFromArray(distractorsArray)
  local nouns = TableUtils.createSetFromArray(nounsArray)
  local verbs = TableUtils.createSetFromArray(verbsArray)

  validatePromptIsNotInOtherSets(self, prompt, basic, medium, advanced, suggestions, hints, correct, distractors, nouns, verbs)
  validateDistractorsAreNotInOtherSets(self, distractors, basic, medium, advanced, suggestions, hints, correct, nouns, verbs)
  validateSuggestionsHintsCorrectAreInRightSets(self, basic, medium, advanced, suggestions, hints, correct)
  validateBasicMediumAdvancedNounsVerbsAreDisjoint(self, basic, medium, advanced, nouns, verbs)
  validateAllSetsHaveLowercaseLettersOnly(self, prompt, basic, medium, advanced, suggestions, hints, correct, distractors, nouns, verbs)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function SynonymsValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, SynonymsValidator)

  return self
end
