--- HintOverlay
-- @author Created by karl on June 27, 2014.

class("HintOverlay")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_BG = {52/255, 174/255, 157/255}
local COLOR_TEXT = {0, 0, 0, 0.8}

local DURATION_ANIMATE_IN = 0.2
local DURATION_DELAY = 0.05
local DURATION_ANIMATE_OUT = 0.15

local SPACING_ADJUST_NUMERATOR = -3
local SPACING_ADJUST_DENOMINATOR = 6
local SPACING_ADJUST_OPERATOR = 4
local SPACING_ROW = 6


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getSmallFontSize(width, height)
  return math.floor(math.min(width, height) * 0.22)
end

local function getLargeFontSize(width, height)
  return math.floor(math.min(width, height) * 0.28)
end

local function newLabel(text, fontSize)
  local color = COLOR_TEXT
  local font = TotemCommon.FONT_TOTEM_FACE
  local label = DeprecatedLabel.newSingleLine(text, font, fontSize)
  label:setColor(unpack(color))
  return label
end

class("HintBox")

function HintBox:setConcept(concept)
  error("Subclasses must implement this method")
end

function HintBox.new(width, height)
  local boxCorner = width * 0.15
  local self = FilledRoundedRectangle.new(width, height, boxCorner, COLOR_BG)
  Mixin.mix(self, HintBox)

  local triangleW, triangleH = math.floor(width * 0.2), math.floor(width * 0.2)
  local tri = FilledTriangle.new(triangleW, 0, triangleW,triangleH, 0, triangleH/2, COLOR_BG)
  tri:setAnchorPointRight()
  tri:setLoc(0, height/2)
  self:insert(tri)

  return self
end


class("Fraction")

function Fraction:setRatio(numerator, denominator)
  self.numeratorLabel:setText(tostring(numerator))
  self.denominatorLabel:setText(tostring(denominator))
  local maxW = math.max(self.numeratorLabel:getWidth(), self.denominatorLabel:getWidth())
  local separatorH = self.separator:getHeight()
  self.separator:setDims(math.floor(maxW + 4), separatorH)

  self:alignChildren()
  local numX, numY = self.numeratorLabel:getLoc()
  self.numeratorLabel:setLoc(numX, numY + SPACING_ADJUST_NUMERATOR)
  local denX, denY = self.denominatorLabel:getLoc()
  self.denominatorLabel:setLoc(denX, denY + SPACING_ADJUST_DENOMINATOR)
end

function Fraction.new(fontSize)
  local self = Column.new(0)
  Mixin.mix(self, Fraction)

  local numeratorLabel = newLabel("", fontSize)
  local denominatorLabel = newLabel("", fontSize)

  local spacing = math.max(math.floor(fontSize * 0.05), 1)
  local separator = FilledRectangle.new(0, spacing, COLOR_TEXT)

  self:insert(numeratorLabel)
  self:insert(separator)
  self:insert(denominatorLabel)

  self.numeratorLabel = numeratorLabel
  self.denominatorLabel = denominatorLabel
  self.separator = separator

  return self
end


class("FractionEquality")

function FractionEquality:setRatios(numerator1, denominator1, numerator2, denominator2, isEqual)
  self.fraction1:setRatio(numerator1, denominator1)
  self.fraction2:setRatio(numerator2, denominator2)

  if isEqual then
    self.equals:setText("=")
  else
    self.equals:setText("≈")
  end

  self:alignChildren()

  local equalsX, equalsY = self.equals:getLoc()
  self.equals:setLoc(equalsX, equalsY+SPACING_ADJUST_OPERATOR)
end

function FractionEquality.new(fontSize)
  local self = Row.new(SPACING_ROW)
  Mixin.mix(self, FractionEquality)

  self.fraction1 = Fraction.new(fontSize)
  self.fraction2 = Fraction.new(fontSize)

  local color = COLOR_TEXT
  local font = TotemCommon.FONT_TOTEM_FACE
  self.equals = newLabel("", fontSize)

  self:insert(self.fraction1)
  self:insert(self.equals)
  self:insert(self.fraction2)

  return self
end


class("RatioBox")

function RatioBox:setConcept(concept)
  self.fraction:setRatio(MathUtils.fractionLowestTerms(concept.numerator, concept.denominator))
end

function RatioBox.new(width, height)
  local self = HintBox.new(width, height)
  Mixin.mix(self, RatioBox)

  local fontSize = getLargeFontSize(width, height)
  self.fraction = Fraction.new(fontSize)
  self:insertAtCenter(self.fraction)

  return self
end

class("DividedFractionBox")

function DividedFractionBox:setConcept(concept)
  local numerator, denominator = concept[self.numeratorName], concept[self.denominatorName]
  local numerator2, denominator2 = MathUtils.fractionLowestTerms(numerator, denominator)
  local isEqual = denominator == denominator2

  if isEqual then
    self.fraction:setRatio(numerator, denominator)
    self.fraction:setOpacity(1)
    self.equalFractionRow:setOpacity(0)
  else
    self.fraction:setOpacity(0)
    self.equalFractionRow:setOpacity(1)
    local divideNum, divideDen = numerator / numerator2, numerator / numerator2
    local isEqual = true
    self.equalFraction:setRatios(divideNum, divideDen, numerator2, denominator2, isEqual)
    self.equalFractionRow:alignChildren()
  end

  local divideX, divideY = self.divide:getLoc()
  self.divide:setLoc(divideX, divideY+SPACING_ADJUST_OPERATOR)
end

function DividedFractionBox.new(width, height, numeratorName, denominatorName)
  local self = HintBox.new(width, height)
  Mixin.mix(self, DividedFractionBox)

  self.numeratorName = numeratorName or "numerator"
  self.denominatorName = denominatorName or "denominator"

  local fontSize = getLargeFontSize(width, height)
  self.equalFraction = FractionEquality.new(fontSize)
  self.divide = newLabel("÷", fontSize)

  self.equalFractionRow = Row.new(SPACING_ROW)
  self:insertAtCenter(self.equalFractionRow)
  self.equalFractionRow:insert(self.divide)
  self.equalFractionRow:insert(self.equalFraction, 10)
  self.equalFractionRow:linkColorAttrToChildren(true)

  self.fraction = Fraction.new(fontSize)
  self:insertAtCenter(self.fraction)
  self.fraction:linkColorAttrToChildren(true)

  self.fraction:setOpacity(0)
  self.equalFractionRow:setOpacity(0)

  return self
end

class("MoneyFractionBox")

function MoneyFractionBox.new(width, height)
  return DividedFractionBox.new(width, height, "moneyNumerator", "moneyDenominator")
end

class("PercentageBox")

function PercentageBox:setConcept(concept)
  local numerator1, denominator1 = concept.numerator / concept.denominator * 100, 100
  local numerator2, denominator2 = MathUtils.fractionLowestTerms(concept.numerator, concept.denominator)
  local isEqual = (math.floor(numerator1 * 100) / 100) == numerator1
  local formatString = "%.2f"

  if numerator1 % 1 == 0 then
    formatString = "%g"
  end
  numerator1 = string.format(formatString, numerator1)

  self.fraction:setRatios(numerator1, denominator1, numerator2, denominator2, isEqual)
end

function PercentageBox.new(width, height)
  local self = HintBox.new(width, height)
  Mixin.mix(self, PercentageBox)

  local fontSize = getSmallFontSize(width, height)
  self.fraction = FractionEquality.new(fontSize)

  self:insertAtCenter(self.fraction)

  return self
end


local function makeHintBox(self, class)
  local selfW, selfH = self:getDims()

  local boxW = math.floor(selfW * 0.3)
  local boxH = math.floor(selfH * 0.2)

  return class.new(boxW, boxH)
end

local function makeZoomAction(helpBox, duration)
  return AHSequence.new(
    AHScaleTo.new(helpBox, duration, AHEaseSineOut.new(), 1.2),
    AHScaleTo.new(helpBox, 0.1, AHEaseSineInOut.new(), 1)
  )
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function HintOverlay:animateIn()
  self.isShowing = true
  return runInCoroutine(function()
    for _, helpBox in ipairs(self.helpBoxes) do
      waitFor(DURATION_DELAY)
      makeZoomAction(helpBox, DURATION_ANIMATE_IN):start()
    end
  end)
end

function HintOverlay:animateOut()
  self.isShowing = false
  return runInCoroutine(function()
    for _, helpBox in ipairs(self.helpBoxes) do
      AHScaleTo.new(helpBox, DURATION_ANIMATE_OUT, AHEaseSineOut.new(), 0):start()
    end
  end)
end

function HintOverlay:setConcept(index, concept)
    local hintBox = self.helpBoxes[index]
    hintBox:setConcept(concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function HintOverlay.new(config, onDismiss)
  local self = Group.newFullScreen()
  Mixin.mix(self, HintOverlay)
  local selfW, selfH = self:getDims()

  local helpBoxX = math.floor(selfW * 0.68)
  local helpBoxYValues = {
    math.floor(selfH * 0.89),
    math.floor(selfH * 0.62),
    math.floor(selfH * 0.347),
    math.floor(selfH * 0.1)
  }

  local classes = {
    percent = PercentageBox,
    fraction = DividedFractionBox,
    ratio = RatioBox,
    money = MoneyFractionBox
  }

  self.helpBoxes = {}
  for i, helpBoxY in ipairs(helpBoxYValues) do
    local conceptType = config[i].type
    local helpBox = makeHintBox(self, classes[conceptType])
    helpBox:setAnchorPointLeft()
    helpBox:setLoc(helpBoxX, helpBoxY)
    helpBox:setScl(0)
    self:insert(helpBox)
    table.insert(self.helpBoxes, helpBox)
  end

  return self
end
