--- Totem In-game UI

class("InGameUI")
require("TotemCommon")
require("InstructionsOverlay")
require("HintOverlay")
require("HelpOverlay")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_OVERLAY = 90
local Z_NOTIFICATION = 100

local NOTIFICATION_BUMP_DISTANCE = 3
local NOTIFICATION_BOX_WIDTH = Screen.width
local NOTIFICATION_BOX_HEIGHT = math.floor(Screen.height * 0.08)
local NOTIFICATION_BG_COLOR = {5/255, 37/255, 33/255, 0.7}
local NOTIFICATION_TEXT_COLOR = {1, 1, 1, 0.7}

local SHOW_NOTIFICATION_KEY = "SHOW_NOTIFICATION_KEY"


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function onTouch(self)
  local actions = self.touchActions
  self.touchActions = AHSequence.new()
  actions:start()
end

local function addTouchAction(self, action)
  self.touchActions:addAction(action)
end

local function moveNotificationBox(self, isMoveIn)
  return runExclusivelyInCoroutine(self, SHOW_NOTIFICATION_KEY, true, function()
    local selfW, selfH = self:getDims()
    local box = self.notificationBox
    local boxW, boxH = box:getDims()

    local upY = NOTIFICATION_BOX_HEIGHT + NOTIFICATION_BUMP_DISTANCE
    local downY = NOTIFICATION_BOX_HEIGHT
    local upDuration = 0.29
    local downDuration = 0.13

    if not isMoveIn then
      downY = 0
      upDuration = 0.13
      downDuration = 0.29
    end

    wait(AHSequence.new(
                        AHLocationYTo.new(box, upDuration, AHEaseSineOut.new(), upY),
                        AHLocationYTo.new(box, downDuration, AHEaseSineInOut.new(), downY)
                        ):start())
  end)
end

local function hideBottomNotificationOnTouch(self)
  addTouchAction(self, function()
    self:hideBottomNotification()
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function InGameUI:showBottomNotification(text)
  return runInCoroutine(function()
    if self._isBottomNotificationShowing then
      wait(self:hideBottomNotification())
    end
    self._isBottomNotificationShowing = true
    hideBottomNotificationOnTouch(self)
    self.notificationBox.label:setText(text)
    wait(moveNotificationBox(self, true))
  end)
end

function InGameUI:hideBottomNotification()
  return runInCoroutine(function()
    if self._isBottomNotificationShowing then
      self._isBottomNotificationShowing = false
      wait(moveNotificationBox(self, false))
    end
  end)
end

function InGameUI:isBottomNotificationShowing()
  return self._isBottomNotificationShowing
end

function InGameUI:showInstructions()
  self.instructionsOverlay:animateIn()
end

function InGameUI:showHint(cubes)
  for index, cube in ipairs(cubes) do
    local concept = cube:getFrontFaceConcept()
    self.hintOverlay:setConcept(index, concept)
  end
  return self.hintOverlay:animateIn()
end

function InGameUI:removeHint()
  if self.hintOverlay.isShowing then
    self.hintOverlay:animateOut()
  end
end

function InGameUI:isHintShowing()
  return self.hintOverlay.isShowing
end

function InGameUI:updateHint(cube)
  if self:isHintShowing() then
    self.hintOverlay:setConcept(cube.cubeIndex, cube:getFrontFaceConcept())
  end
end

function InGameUI:showHelp()
  self.helpOverlay:animateIn()

  addTouchAction(self, function()
    self.helpOverlay:animateOut()
  end)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makeNotificationBox()
  local rectW, rectH = NOTIFICATION_BOX_WIDTH, NOTIFICATION_BOX_HEIGHT + NOTIFICATION_BUMP_DISTANCE
  local rect = FilledRectangle.new(rectW, rectH, NOTIFICATION_BG_COLOR)
  local fontSize = math.floor(rectH * 0.5)
  fontSize = fontSize - fontSize % 2
  rect.label = DeprecatedLabel.newSingleLine("", TotemCommon.FONT_NORMAL, fontSize)
  rect.label:setColor(unpack(NOTIFICATION_TEXT_COLOR))
  rect.label:setLoc(rectW/2, rectH/2 + NOTIFICATION_BUMP_DISTANCE)
  rect:insert(rect.label)
  return rect
end

-- Constructor ------------------------

function InGameUI.new(iconBar, totemContentTemplate, instructionFields, onInstructionsDismissed)
  local self = Group.newFullScreen()
  Mixin.mix(self, InGameUI)
  local selfW, selfH = self:getDims()

  self.iconBar = iconBar

  self.instructionsOverlay = InstructionsOverlay.new(instructionFields, onInstructionsDismissed)
  self.instructionsOverlay:setLoc(selfW/2, selfH/2)
  self:insert(self.instructionsOverlay, Z_OVERLAY)

  self.hintOverlay = HintOverlay.new(totemContentTemplate)
  self.hintOverlay:setLoc(selfW/2, selfH/2)
  self:insert(self.hintOverlay, Z_OVERLAY)

  self.helpOverlay = HelpOverlay.new()
  self:insertAtCenter(self.helpOverlay, Z_OVERLAY)

  self.notificationBox = makeNotificationBox()
  local boxW, boxH = self.notificationBox:getDims()
  self.notificationBox:setAnchorPointTop()
  self.notificationBox:setLoc(selfW/2, 0)
  self:insert(self.notificationBox, Z_NOTIFICATION)

  self.touchActions = AHSequence.new()
  TouchManager.addTouchHandler(self, function(touchEvent)
    if touchEvent.phase == TouchManager.TOUCH_BEGAN then
      onTouch(self)
    end
  end, TouchHandler.BEHAVIOR_ALWAYS)

  return self
end
