--- Instructions Overlay

class("InstructionsOverlay")
require("TotemCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLOR_BG = {52/255, 174/255, 157/255}
local COLOR_TEXT = {1, 1, 1}

local DURATION_ANIMATE_IN = 0.2
local DURATION_DELAY = 0.05
local DURATION_ANIMATE_OUT = 0.15


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeInstructionHelpBox(text)
  local helpBoxH = 20
  local fontSize = helpBoxH - 2

  local label = DeprecatedLabel.newSingleLine(text, TotemCommon.FONT_NORMAL, 20)
  local labelW, labelH = label:getDims()
  local helpBox = FilledRoundedRectangle.new(labelW+8, labelH, 3, COLOR_BG)
  local helpBoxW, helpBoxH = helpBox:getDims()
  label:setLoc(helpBoxW/2, helpBoxH/2)
  helpBox:insert(label)
  helpBox.label = label

  local triangleW, triangleH = math.floor(helpBoxH * 0.6), math.floor(helpBoxH * 0.4)
  local tri = FilledTriangle.new(triangleW, 0, triangleW,triangleH, 0, triangleH/2, COLOR_BG)
  tri:setAnchorPointRight()
  tri:setLoc(0, helpBoxH/2)
  helpBox:insert(tri)

  return helpBox
end

local function makeZoomAction(helpBox, duration)
  return AHSequence.new(
                        AHScaleTo.new(helpBox, duration, AHEaseSineOut.new(), 1.2),
                        AHScaleTo.new(helpBox, 0.1, AHEaseSineInOut.new(), 1)
                        )
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function InstructionsOverlay:animateIn()
  return runInCoroutine(function()
    local topBarW, topBarH = self.topBar:getDims()
    AHLocationYBy.new(self.topBar, DURATION_ANIMATE_IN, AHEaseSineOut.new(), -topBarH):start()
    for _, helpBox in ipairs(self.helpBoxes) do
      waitFor(DURATION_DELAY)
      makeZoomAction(helpBox, DURATION_ANIMATE_IN):start()
    end
  end)
end

function InstructionsOverlay:animateOut()
  return runInCoroutine(function()
    for _, helpBox in ipairs(self.helpBoxes) do
      AHScaleTo.new(helpBox, DURATION_ANIMATE_OUT, AHEaseSineOut.new(), 0):start()
    end
    local topBarW, topBarH = self.topBar:getDims()
    wait(AHLocationYBy.new(self.topBar, DURATION_ANIMATE_OUT, AHEaseSineIn.new(), topBarH):start())
  end)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function InstructionsOverlay.new(names, onDismiss)
  local self = Group.new(Screen.width, Screen.height)
  Mixin.mix(self, InstructionsOverlay)
  local selfW, selfH = self:getDims()
  local safeAreaHeight = Screen.height - Screen.safeTop
  local labelAreaHeight = 30
  local topBarH = labelAreaHeight + safeAreaHeight
  local topBarFontSize = labelAreaHeight - 10

  self.topBar = FilledRectangle.new(Screen.width, topBarH, COLOR_BG)
  self.topBar:setAnchorPointBottom()
  self.topBar:setLoc(selfW/2, selfH)
  self:insert(self.topBar)
  self.topBar.label = DeprecatedLabel.newSingleLine("SPIN AND MATCH CUBES", TotemCommon.FONT_NORMAL, topBarFontSize)
  self.topBar.label:setLoc(selfW/2, labelAreaHeight/2)
  self.topBar:insert(self.topBar.label)

  local helpBoxX = math.floor(selfW * 0.68)
  local helpBoxYValues = {
    math.floor(selfH * 0.89),
    math.floor(selfH * 0.62),
    math.floor(selfH * 0.347),
    math.floor(selfH * 0.075)
  }

  self.helpBoxes = {}
  for i, helpBoxY in ipairs(helpBoxYValues) do
    local helpBox = makeInstructionHelpBox(names[i])
    helpBox:setAnchorPointLeft()
    helpBox:setLoc(helpBoxX, helpBoxY)
    helpBox:setScl(0)
    self:insert(helpBox)
    table.insert(self.helpBoxes, helpBox)
  end

  TouchManager.addTouchHandler(self, function(touchEvent)
    if touchEvent.phase == TouchManager.TOUCH_BEGAN then
      TouchManager.removeTouchHandler(self)
      runInCoroutine(function()
        wait(self:animateOut())
        self:removeSelf()
        self.topBar.label:releaseTexture()
        for i, helpBox in ipairs(self.helpBoxes) do
          helpBox.label:releaseTexture()
        end
        if onDismiss then
          onDismiss()
        end
      end)
    end
  end, TouchHandler.BEHAVIOR_ALWAYS)

  return self
end
