--- Totem
-- @author Created by a Adam Villalobos on Tuesday, August 27, 2013.

class("Totem")
require("TotemCylinder")
require("TotemCube")
require("TotemCommon")
require("Particles")
require("TotemFraction")
require("TotemRatio")
require("MoneyFraction")
require("TotemPercentage")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- General Configuration --------------

-- Nuber of cubes on a single totem
local CUBES_PER_TOTEM = 4

-- Number of sides that must be aligned for the totem to be complete
local REQUIRED_COMPLETED_SIDES = 3

-- Min distance to swipe for it to register as a gesture
local MIN_SWIPE_DISTANCE = Screen.width * 0.05

-- Number of rays to show during completion sequence
local RAY_COUNT = 8


-- Sizing, spacing --------------------

local CUBE_WIDTH = math.floor(Screen.width * 0.3)
local CUBE_HEIGHT = math.floor(Screen.height * 0.23)
local CUBE_SPACING = math.floor(Screen.height * 0.02)
local CYLINDER_RADIUS = math.floor(CUBE_WIDTH * 0.3)

-- Text & Fonts -----------------------

local FONT_SIZE_MIN = 10
local FONT_SIZE_MAX = 30
local FONT_SIZE_GRANULARITY = 5
local STRING_TEXTURE_SIZE = 128 -- Must be power-of-2

-- Z Order ----------------------------

local Z_DARK_OVERLAY = 10
local Z_RAYS = 20
local Z_BACK_BEAM = 30
local Z_FRONT_BEAM = 40
local Z_TOTEM_PILLAR = 50
local Z_TOTEM_OUTER_BLOCKS = 51
local Z_TOTEM_INNER_BLOCKS = 52
local Z_TOTEMS = 50
local Z_SMOKE = 60
local Z_PARTICLES = 70

-- Colors -----------------------------

local COLOR_BACK_BEAM = {1, 1, 0.8, 0.6}
local COLOR_FRONT_BEAM = {1, 1, 1, 0.7}
local DARK_OVERLAY_COLOR = {54/255, 32/255, 68/255, 0.5}

-- Other ------------------------------

local ANIM_FRAMERATE = 30

local LARGE_SMOKE_SCALE = 1.2
local SMALL_SMOKE_SCALE = 0.8

local RENDER_ENGINE_COUNT = 16

local HIGHEST_TERM_DIFFICULTIES = {
  medium = true,
  hard = true
}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getSemiRandomStartingFaceOrder()
  local sampleCount = 4
  -- 2 of each index insures that at most two faces match when cubes are shown
  local possibleValues = {1,1, 2,2, 3,3, 4,4}
  return TableUtils.createRandomSampleOfArray(possibleValues, sampleCount)
end

local function areCubesUnhighlightedAndAligned(self)
  local concept = self.cubes[1]:getFrontFaceConcept()

  for _, cube in pairs(self.cubes) do
    if cube:isRotating() then
      return false
    end
    if cube:getFrontFaceConcept() ~= concept then
      return false
    end
    if cube:isCurrentHighlighted() then
      return false
    end
  end

  return true
end

local function getFacingConcept(self)
  return self.cubes[1]:getFrontFaceConcept()
end

local function getSemiRandomCubeFaceOrder(firstFaceIndex)
  local conceptOrder = {1, 2, 3, 4}
  TableUtils.removeValueFromArray(conceptOrder, firstFaceIndex)
  TableUtils.shuffleArray(conceptOrder)
  table.insert(conceptOrder, 1, firstFaceIndex)
  return conceptOrder
end

local function prepareCubeFaceTextures(self, cube, createFunction, firstFaceIndex)
  local conceptOrder = getSemiRandomCubeFaceOrder(firstFaceIndex)

  for faceIdx, conceptIdx in ipairs(conceptOrder) do
    local concept = self.concepts[conceptIdx]
    local group = createFunction(concept)
    self.showCubeFacesAction:addAction(function()
      cube:setContentOnFace(faceIdx, group, self.concepts[conceptIdx])
    end)
  end
end

local function onSwipeGesture(self, deltaX, locY)
  --determine column of swipe
  local cubeIdx
  if locY < Screen.height*0.25 then
    cubeIdx = 4
  elseif locY < Screen.height*0.5 then
    cubeIdx = 3
  elseif locY < Screen.height*0.75 then
    cubeIdx = 2
  else
    cubeIdx = 1
  end

  local cube = self.cubes[cubeIdx]

  runInCoroutine(function()
    if deltaX > 0 then
      wait(cube:rotateLeft())
    else
      wait(cube:rotateRight())
    end
  end)
end

local function newRay(self)
  local selfW, selfH = self:getDims()
  local rayW, rayH = math.floor(selfW * 0.4), math.floor(selfH * 1.1)

  return FilledTriangle.new(rayW/2, 0, rayW, rayH, 0, rayH, {1, 1, 1})
end

local function buildRays(self)
  local selfW, selfH = self:getDims()
  local rotateAmount = 360 / RAY_COUNT
  local currentRotation = 0

  local rays = Group.new(selfW, selfH)
  for i=1,RAY_COUNT do
    local ray = newRay(self)
    ray:setAnchorPointBottom()
    ray:setRot(rotateAmount * i)
    ray:setLoc(selfW/2, selfH/2)
    rays:insert(ray)
  end
  rays:linkAttrToChildren(MOAIColor.INHERIT_COLOR, MOAIColor.COLOR_TRAIT)

  return rays
end

local function buildBeam(self, filename, width, color, zOrder)
  local selfW, selfH = self:getDims()
  local beam = Sprite.newWithName(filename)
  beam:setEffectiveDims(width, selfH)
  beam:setColor(unpack(color))
  beam:setLoc(selfW/2, -selfH/2)
  return beam
end

local function runBeamAction(self)
  return runInCoroutine(function()
    local selfW, selfH = self:getDims()

    local backBeam = buildBeam(self, "totem_beamback.png", CUBE_WIDTH * 1.4, COLOR_BACK_BEAM)
    self.backgroundGroup:insert(backBeam, Z_BACK_BEAM)
    local frontBeam = buildBeam(self, "totem_beam.png", CUBE_WIDTH * 1.2, COLOR_FRONT_BEAM)
    self.backgroundGroup:insert(frontBeam, Z_FRONT_BEAM)

    local backScaleX, backScaleY = backBeam:getScl()
    local frontScaleX, frontScaleY = frontBeam:getScl()

    local scaleAction = AHLoop.new(
                                   AHSequence.new(
                                                  AHSpawn.new(
                                                              AHScaleTo.new(backBeam, 0.1, nil, backScaleX * 1.02, backScaleY),
                                                              AHScaleTo.new(frontBeam, 0.1, nil, frontScaleX * 1.02, frontScaleY)
                                                              ),
                                                  AHSpawn.new(
                                                              AHScaleTo.new(backBeam, 0.1, nil, backScaleX, backScaleY),
                                                              AHScaleTo.new(frontBeam, 0.1, nil, frontScaleX, frontScaleY)
                                                              )
                                                  )
                                   )
    scaleAction:start()

    wait(
         AHSequence.new(
                        AHSpawn.new(
                                    AHLocationYTo.new(backBeam, 0.21, AHEaseSineIn.new(), selfH / 2),
                                    AHLocationYTo.new(frontBeam, 0.21, AHEaseSineIn.new(), selfH / 2)
                                    ),
                        AHDelay.new(1.5),
                        AHSpawn.new(
                                    AHOpacityTo.new(backBeam, 0.38, nil, 0),
                                    AHOpacityTo.new(frontBeam, 0.38, nil, 0)
                                    )
                        ):start()
         )

    scaleAction:stop()
    backBeam:removeSelf()
    frontBeam:removeSelf()
  end)
end

local function runRaysAction(self)
  return runInCoroutine(function()
    local selfW, selfH = self:getDims()
    local rotateAmount = -180
    local duration = 1.83
    local opacityInDuration = 0.17
    local opacityOutDuration = 0.33

    local rays = buildRays(self)
    rays:setLoc(selfW/2, -selfH * 0.05)
    rays:setOpacity(0)
    self.backgroundGroup:insert(rays, Z_RAYS)

    wait(
         AHSequence.new(
                        AHSpawn.new(
                                    AHRotationBy.new(rays, duration, nil, rotateAmount),
                                    AHSequence.new(
                                                   AHOpacityTo.new(rays, opacityInDuration, AHEaseSineIn.new(), 0.2),
                                                   AHDelay.new(duration - opacityInDuration - opacityOutDuration),
                                                   AHOpacityTo.new(rays, opacityOutDuration, AHEaseSineIn.new(), 0)
                                                   )
                                    ),
                        AHCallLogic.new(function()
                                            rays:removeSelf()
                                          end)
                        ):start()
         )
  end)
end

local function runDarkenAction(self, duration)
  return runInCoroutine(function()
    local selfW, selfH = self:getDims()
    local boxColor = DARK_OVERLAY_COLOR
    local opacity = boxColor[4]
    boxColor[4] = 0
    local box = FilledRectangle.new(selfW, selfH*4, boxColor)
    box:setLoc(selfW/2, selfH/2)
    self.backgroundGroup:insert(box, Z_DARK_OVERLAY)

    wait(
         AHSequence.new(
                        AHOpacityTo.new(box, 0.21, nil, opacity),
                        AHDelay.new(duration),
                        AHOpacityTo.new(box, 0.42, nil, 0)
                        ):start()
         )
    box:removeSelf()
  end)
end

local function runHighlightAction(self, concept)
  runInCoroutine(function()
    local delayAmount = 0.05
    for _, cube in pairs(self.cubes) do
      cube:highlight(concept)
      waitFor(delayAmount)
    end
  end)
end

local function getFontTexture(self, text)
  local fontRenderEngine = self.fontRenderEngines[self.fontRenderEngineIdx]
  self.fontRenderEngineIdx = self.fontRenderEngineIdx % RENDER_ENGINE_COUNT + 1
  wait(fontRenderEngine:calculateOptimalFontSizeForText(text))
  return fontRenderEngine:renderTexture(text)
end

local function getCubeContentDims(self)
  local cube = self.cubes[1]
  return MathUtils.incrementUntilPowerOf2(cube.contentW), MathUtils.incrementUntilPowerOf2(cube.contentH)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function Totem:rotateCube(cubeIndex, isRight)
  local cube = self.cubes[cubeIndex]

  if isRight then
    return cube:rotateRight()
  else
    return cube:rotateLeft()
  end
end

function Totem:enableTouch()
  local handler = function(touchEvent)
    if touchEvent.phase == TouchManager.TOUCH_BEGAN then
      self.touchDownEvent = touchEvent
    elseif self.touchDownEvent then
      local deltaX = self.touchDownEvent.x - touchEvent.x
      if math.abs(deltaX) >= MIN_SWIPE_DISTANCE then
        onSwipeGesture(self, deltaX, self.touchDownEvent.y)
        self.touchDownEvent = nil
      end
    end
  end
  local handlerBehaviors = nil
  local priority = -1 -- Allow any buttons tapped to take priority before checking for a swipe
  TouchManager.addTouchHandler(self, handler, handlerBehaviors, priority)
end

function Totem:disableTouch()
  TouchManager.removeTouchHandler(self)
end

function Totem:setConcepts(concepts)
  return runInCoroutine(function()
    self.concepts = concepts
    self.sideCompletionCount = 0
    self.showCubeFacesAction = AHSpawn.new()
    self.expectedFaces = CUBES_PER_TOTEM * 4
    local startingFaceOrder = getSemiRandomStartingFaceOrder()
    local contentW, contentH = getCubeContentDims(self)

    for i=1,CUBES_PER_TOTEM do
      local config = self.config[i]
      local createFunction
      local firstField = config.fields[1]
      if config.type == "percent" then
        createFunction = function(concept)
          return TotemPercentage.new(contentW, contentH, concept.percent)
        end
      elseif config.type == "fraction" then
        createFunction = function(concept)
          return TotemFraction.new(contentW, contentH, concept.numerator, concept.denominator)
        end
      elseif config.type == "ratio" then
        createFunction = function(concept)
          local difficulty = string.lower(concept.difficulty)
          local isHighestTerms = HIGHEST_TERM_DIFFICULTIES[difficulty]
          return TotemRatio.new(contentW, contentH, concept.numerator, concept.denominator, isHighestTerms)
        end
      elseif config.type == "money" then
        createFunction = function(concept)
          return MoneyFraction.new(contentW, contentH, concept.moneyNumerator, concept.moneyDenominator)
        end
      else
        assert(false, tostring(config.type) .. ": Unknown totem content type")
      end
      self.cubes[i]:reset()
      prepareCubeFaceTextures(self, self.cubes[i], createFunction, startingFaceOrder[i])
    end
  end)
end

function Totem:showConcepts()
  return runInCoroutine(function()
    while #self.showCubeFacesAction:getChildActions() < self.expectedFaces do
      coroutine.yield()
    end
    self.showCubeFacesAction:start()
  end)
end

function Totem:markCurrentSideComplete()
  self.sideCompletionCount = self.sideCompletionCount + 1
end

function Totem:incompleteSidesRemaining()
  local result = REQUIRED_COMPLETED_SIDES - self.sideCompletionCount
  if result < 0 then
    result = 0
  end
  return result
end

function Totem:isComplete()
  return self:incompleteSidesRemaining() == 0
end

function Totem:getCubes()
  return self.cubes
end

function Totem:playSideCompleteSequence(concept, duration, soundName)
  duration = duration or 1.25
  soundName = soundName or "correct"
  return runInCoroutine(function()
    Sound.playSoundEffect(soundName)

    local emitDuration = 0.7
    self.particles:emitFor(emitDuration)

    local allActions = {
      runHighlightAction(self, concept),
      runRaysAction(self),
      runDarkenAction(self, duration),
      runBeamAction(self)
    }
    for _, action in pairs(allActions) do
      wait(action)
    end

  end)
end

function Totem:playTotemCompleteSequence(concept)
  return self:playSideCompleteSequence(concept, 1.9, "completeset")
end

function Totem:spin()
  return runInCoroutine(function()
    local isLeft = false
    local duration = 0.5
    local delay = 0.2
    local lastAction = nil
    for _, cube in ipairs(self.cubes) do
      lastAction = cube:spin(duration, isLeft)
      duration = duration + delay
      isLeft = not isLeft
    end
    wait(lastAction)
  end)
end

function Totem:release()
  for _, cube in ipairs(self.cubes) do
    cube:release()
  end
end

function Totem:getIdentifier()
  local concept = self.concepts[1]
  return concept.identifier
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makefontRenderEngine()
  local fontRenderEngine = FontRenderEngine.new()
  fontRenderEngine:setFont(TotemCommon.FONT_NORMAL)
  fontRenderEngine:setGranularity(FONT_SIZE_GRANULARITY)
  fontRenderEngine:setWidth(STRING_TEXTURE_SIZE)
  fontRenderEngine:setHeight(STRING_TEXTURE_SIZE)
  fontRenderEngine:setMinFontSize(FONT_SIZE_MIN)
  fontRenderEngine:setMaxFontSize(FONT_SIZE_MAX)
  fontRenderEngine:setAlignment(Alignment.CENTER, Alignment.CENTER)

  return fontRenderEngine
end

-- Constructor ------------------------

function Totem.new(config, onFaceChanged, onSideComplete, backgroundGroup, foregroundGroup)
  local self = Group.newFullScreen()
  Mixin.mix(self, Totem)

  local selfW, selfH = self:getDims()

  self.config = config
  self.backgroundGroup = backgroundGroup
  self.foregroundGroup = foregroundGroup

  local cylinder = TotemCylinder.new(CYLINDER_RADIUS, selfH)
  cylinder:setLoc(selfW/2, selfH/2, 0)
  self:insert(cylinder, Z_TOTEM_PILLAR)

  local yPos = math.floor(CUBE_HEIGHT * (CUBES_PER_TOTEM - 0.5) + CUBE_SPACING * (CUBES_PER_TOTEM-1))

  self.cubes = {}
  for i=1,CUBES_PER_TOTEM do
    local textureName = "totem_block_" .. i .. ".png"
    local cube = TotemCube.new(CUBE_WIDTH, CUBE_HEIGHT, textureName, i)
    local zIndex = Z_TOTEM_INNER_BLOCKS
    if ((i == 1) or (i == CUBES_PER_TOTEM)) then
      zIndex = Z_TOTEM_OUTER_BLOCKS
    end
    self:insert(cube, zIndex)
    cube:setLoc(selfW/2, yPos, 0)
    cube:setRotateCallback(function()
      local didCompleteSide = areCubesUnhighlightedAndAligned(self)

      onFaceChanged(cube, didCompleteSide)

      if didCompleteSide then
        onSideComplete(getFacingConcept(self))
      end
    end)
    table.insert(self.cubes, cube)
    yPos = yPos - CUBE_HEIGHT - CUBE_SPACING
  end

  self.particles = Particles.new("starburst.pex")
  self.particles:setLoc(Screen.width/2, 0)
  foregroundGroup:insert(self.particles, Z_PARTICLES)
  self.particles:start()

  self.fontRenderEngines = {}
  for i=1,RENDER_ENGINE_COUNT do
    table.insert(self.fontRenderEngines, makefontRenderEngine())
  end
  self.fontRenderEngineIdx = 1

  return self
end
