--- TotemCube
-- @author Created by a Adam Villalobos on Tuesday, August 27, 2013.

class("TotemCube")
require("Totem3D")
require("graphics.OffscreenRenderedTexture")


TotemCube.FACE_FRONT = 1
TotemCube.FACE_RIGHT = 2
TotemCube.FACE_BACK = 3
TotemCube.FACE_LEFT = 4
TotemCube.FACE_BOTTOM = 5
TotemCube.FACE_TOP = 6


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Colors -----------------------------

local COLOR_WHITE = {1, 1, 1}

-- Other ------------------------------

local HIGHLIGHT_DURATION = 0.5
local ROTATE_DURATION = 0.4


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function rotateFaceData(self, isLeft)
  if isLeft then
    if self.currentFace >= TotemCube.FACE_LEFT then
      self.currentFace = TotemCube.FACE_FRONT
    else
      self.currentFace = self.currentFace + 1
    end
  else
    if self.currentFace <= TotemCube.FACE_FRONT then
      self.currentFace = TotemCube.FACE_LEFT
    else
      self.currentFace = self.currentFace - 1
    end
  end
end

local function getFaceForConcept(self, concept)
  for _, face in ipairs(self.faces) do
    if face.concept == concept then
      return face
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TotemCube:reset()
  for _, face in pairs(self.faces) do
    face.highlightProp:setOpacity(0)
    face.isHighlighted = false
  end
end

function TotemCube:highlight(concept)
  local face = getFaceForConcept(self, concept)
  local lightProp = face.highlightProp

  face.isHighlighted = true
  AHOpacityTo.new(lightProp, HIGHLIGHT_DURATION, nil, 1):start()
  if face.group.setTextColor then
    face.group:setTextColor({1, 1, 1})
  end
end

function TotemCube:isCurrentHighlighted()
  return self.faces[self.currentFace].isHighlighted
end

function TotemCube:getFrontFaceConcept()
  return self.faces[self.currentFace].concept
end

function TotemCube:getCubeIndex()
  return self.cubeIndex
end

function TotemCube:setRotateCallback(callback)
  self.rotateCallback = callback
end

function TotemCube:rotateLeft()
  return self:rotate(true)
end

function TotemCube:rotateRight()
  return self:rotate(false)
end

function TotemCube:rotate(isLeft)
  if not self.rotateAction then
    local rotateAmount = 90
    if isLeft then
      rotateAmount = -rotateAmount
    end

    local attributeID = MOAITransform.ATTR_Y_ROT
    local yRotationAction = AHAttributeBy.new(self, ROTATE_DURATION, attributeID, ease, rotateAmount)

    self.rotateAction = AHSequence.new(
      yRotationAction,
      function()
        self.rotateAction = nil
        rotateFaceData(self, isLeft)
        self.rotateCallback()
        Sound.playSoundEffect("slide_click")
      end
    )
    return self.rotateAction:start()
  end
end

function TotemCube:isRotating()
  if self.rotateAction then
    return self.rotateAction:isBusy()
  end
  return false
end

function TotemCube:spin(duration, isLeft)
  local attributeID = MOAITransform.ATTR_Y_ROT
  local ease = AHEaseExponentialOut.new()
  local rotateAmount = 360
  if isLeft then
    rotateAmount = -rotateAmount
  end

  return AHAttributeBy.new(self, duration, attributeID, ease, rotateAmount):start()
end

function TotemCube:setContentOnFace(faceIdx, group, concept)
  local face = self.faces[faceIdx]
  face.contentMesh:setTexture(OffscreenRenderedTexture.new(group))
  face.group = group
  face.concept = concept
end

function TotemCube:release()
  for _, face in ipairs(self.faces) do
    face.contentMesh:setTexture(face.blankTexture)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makeCube(cubeW, cubeH, cubeD, textureName)
  local prop = Group.new(cubeW, cubeH)
  local extentX, extentY, extentZ = math.floor(cubeW / 2), math.floor(cubeH / 2), math.floor(cubeD / 2)

  local mesh = Totem3D.makeBoxMesh(-extentX, -extentY, -extentZ, extentX, extentY, extentZ, COLOR_WHITE)
  local texture = TextureLoader.loadTexture(textureName)
  mesh:setTexture(texture)

  prop:setDeck(mesh)
  prop:setCullMode(MOAIProp.CULL_BACK)
  prop:setDepthTest(MOAIProp.DEPTH_TEST_LESS)
  prop:setPiv(0, cubeH/2, cubeD/2)
  prop:markAs3DObject()

  return prop
end

local function makeCubeWall(self, mesh, distance)
  local selfW, selfH = self:getDims()
  local prop = MOAIProp.new()
  prop:setDeck(mesh)
  prop:setCullMode(MOAIProp.CULL_BACK)

  -- Voodoo fix: Turned off depth test because it causes the cube wall
  -- to not be shown on OpenGL ES (works fine on regular OpenGL...)
  -- prop:setDepthTest(MOAIProp.DEPTH_TEST_LESS)

  -- TODO: Setting full Y pos instead of half...
  prop:setLoc(selfW/2, selfH, 0)
  prop:setPiv(0, selfH/2, -distance)
  Group.markAs3DObject(prop)
  return prop
end

-- Constructor ------------------------

function TotemCube.new(selfW, selfH, textureName, cubeIndex)
  local self = Group.new(selfW, selfH)
  Mixin.mix(self, TotemCube)

  self.faces = {}
  self.currentFace = TotemCube.FACE_FRONT
  self.cubeIndex = cubeIndex

  local cube = makeCube(selfW, selfH, selfW, textureName)
  -- TODO: Totem3D's cube implementation seems weird.
  -- Requires us to set the Y loc to the full height.
  cube:setLoc(selfW/2, selfH, selfW/2)
  self:insert(cube)

  local blankTexture = TextureLoader.loadTexture("blank.png")

  local highlightExtentW, highlightExtentH = math.floor(selfW * 0.5), math.floor(selfH * 0.5)
  local highlightMesh = Totem3D.makeQuadMesh(highlightExtentW, highlightExtentH, -highlightExtentW, -highlightExtentH, COLOR_WHITE)
  highlightMesh:setTexture(TextureLoader.loadTexture("totem_correctside.png"))

  -- Just make the content a square.
  self.contentW, self.contentH = math.floor(selfW * 0.38) * 2, math.floor(selfW * 0.38) * 2

  for i=1,4 do
    local rotation = (i-1)*90
    local face = {}
    table.insert(self.faces, face)

    face.highlightProp = makeCubeWall(self, highlightMesh, selfW/2 + 1)
    face.highlightProp:setOpacity(0)
    face.highlightProp:setRot(0, rotation, 0)
    self:insert(face.highlightProp)

    face.contentMesh = Totem3D.makeQuadMesh(self.contentW/2, self.contentH/2, -self.contentW/2, -self.contentH/2, COLOR_WHITE)
    face.contentMesh:setTexture(blankTexture)
    face.blankTexture = blankTexture
    local prop = makeCubeWall(self, face.contentMesh, selfW/2 + 2)
    prop:setRot(0, rotation, 0)
    self:insert(prop)
  end

  self:markAs3DObject()

  return self
end
