--- TotemGame
-- @author Created by a Adam Villalobos on Monday, August 19, 2013.

class("TotemGame")
require("StandardUIBuilder")

require("TotemConfiguration")
require("Totem")
require("Background")
require("InGameUI")
require("widgets.progress.NumericProgressIndicator")
require("TotemRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Game Settings ----------------------

local MOVES_TILL_HINT = 10

local TIMER_HELP = 5

local SIDES_TO_COMPLETE_TOTEM = 3

-- Z Order ----------------------------

local Z_BACKGROUND = 0
local Z_GRADIENT = 10
local Z_TOTEM = 20

-- Colors -----------------------------

local COLOR_UI = {1, 1, 1, 0.75}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local pointsPerAnswer = TotemConfiguration.POINTS_PER_CORRECT_RESPONSE
  local passDuration = TotemConfiguration.GAME_DURATION_IN_SECONDS
  local pointsPerSecond = TotemConfiguration.POINTS_PER_EXTRA_SECOND_REMAINING

  local passScore = TotemConfiguration.getPassScore()
  local rank2Score = TotemConfiguration.getRankTwoScore()
  local rank3Score = TotemConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rank2Score, rank3Score)

  local difficultyKeyFrames = TotemConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  self.totemCount = TotemConfiguration.TOTEMS_TO_COMPLETE_COUNT
  self.timerDuration = passDuration
  self.pointsPerSecond = pointsPerSecond

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.TIME_BONUS)
end

local getTotemSideCompletionHandler

local function swapTotems(self)
  local nextTotem = self.currentTotem
  self.currentTotem = self.nextTotem
  self.nextTotem = nextTotem
end

local function prepareNextTotem(self)
  return runInCoroutine(function()
    local selfW, selfH = self:getDims()
    self.nextTotem:setLoc(selfW/2, selfH * 1.5, 0)
    wait(self.nextTotem:setConcepts(self.roundGenerator:next()))
    wait(self.nextTotem:showConcepts())
  end)
end

local function showNextTotem(self)
  return runInCoroutine(function()
    local selfW, selfH = self:getDims()

    local moveDuration = 0.7

    Sound.playSoundEffect("moveuptotem")
    wait(AHSpawn.new(
                     AHLocationYBy.new(self.currentTotem, moveDuration, AHEaseSineInOut.new(), -selfH),
                     AHLocationYBy.new(self.nextTotem, moveDuration, AHEaseSineInOut.new(), -selfH)
                     ):start())
    local oldTotem = self.currentTotem
    swapTotems(self)
    oldTotem:release()
    self.currentTotem:enableTouch()
  end)
end

local function onGameWin(self)
  self.timer:stop()
  local timeLeft = self.timer:getTimeRemaining()
  ScoringUtils.applyTimeBonus(timeLeft, TotemConfiguration.POINTS_PER_EXTRA_SECOND_REMAINING, self.bonusTray)
  self.scoreLabel:setScore(GameSession.getScore())

  local roundedTimeLeft = MathUtils.round(timeLeft)
  GameSession.bonusReporter:setBonusCount(BonusTypes.TYPES.TIME_BONUS, roundedTimeLeft)

  return runInCoroutine(function()
    self.currentTotem:disableTouch()
    waitFor(1.7)
    GameSession.notifyGameWin()
  end)
end

local function onGameLoss(self)
  return runInCoroutine(function()
    self.timer:stop()
    self.currentTotem:disableTouch()

    GameSession.contentTracker:notifyTimeout(self.currentTotem:getIdentifier())

    waitFor(1.7)

    GameSession.notifyGameLoss()
  end)
end

local function showHelp(self)
  self.inGameUI:showBottomNotification("Rotate the cubes!")
  self.inGameUI:showHelp()
end

local function stopHelpTimer(self)
  if self.helpTimer then
    self.helpTimer:stop()
  end
end

local function startHelpTimer(self)
  stopHelpTimer(self)
  self.helpTimer = AHSequence.new(
    AHDelay.new(TIMER_HELP),
    function() showHelp(self) end
  ):start()
end

local function showHintNotification(self)
  self.inGameUI:showBottomNotification("HINT")
end

local function showBottomNotification(self)
  local sidesRemaining = self.currentTotem:incompleteSidesRemaining()
  local suffix = " SETS TO GO"
  if sidesRemaining == 1 then
    suffix = " SET TO GO"
  end
  self.inGameUI:showBottomNotification(sidesRemaining .. suffix)
end

local function showHint(self)
  self.inGameUI:showHint(self.currentTotem:getCubes())
  showHintNotification(self)
end

local function removeHint(self)
  self.inGameUI:removeHint()
end

local function onSideComplete(self, concept)
  return runInCoroutine(function()
    runInCoroutine(function()
      self.currentTotem:disableTouch()
      waitFor(0.5)
      self.currentTotem:enableTouch()
    end)
    self.progressIndicator:incrementProgress()
    self.currentTotem:playSideCompleteSequence(concept)
    showBottomNotification(self)
    waitFor(0.5)
    startHelpTimer(self)
    self.movesSinceSideCompleted = 0
  end)
end

local function onTotemComplete(self, concept)
  return runInCoroutine(function()
    self.currentTotem:disableTouch()
    self.progressIndicator:incrementProgress()
    self.completedTotems = self.completedTotems + 1
    local hasWon = self.completedTotems >= self.totemCount
    if hasWon then
      self.timer:stop()
    end
    self.currentTotem:playTotemCompleteSequence(concept)
    waitFor(2)
    if hasWon then
      onGameWin(self)
    else
      self.movesSinceSideCompleted = 0
      removeHint(self)
      GameSession.contentTracker:notifyCorrectAnswer(self.currentTotem:getIdentifier())
      wait(showNextTotem(self))
      GameSession.contentTracker:notifyPresentedConcept(self.currentTotem:getIdentifier())
      -- We don't prepare next totem anymore because there are only two.
      -- prepareNextTotem(self)
    end
  end)
end

getTotemSideCompletionHandler = function(self)
  return function(concept)
    self.hasCompletedSide = true
    runInCoroutine(function()
      GameSession.notifyCorrectAnswer(concept.identifier)
      GameSession.increaseScore(TotemConfiguration.POINTS_PER_CORRECT_RESPONSE)

      self.scoreLabel:setScore(GameSession.getScore())
      self.currentTotem:markCurrentSideComplete()
      stopHelpTimer(self)
      if self.currentTotem:isComplete() then
        onTotemComplete(self, concept)
      else
        onSideComplete(self, concept)
      end
    end)
  end
end

local function getTotemFaceChangedHandler(self)
  return function(cube, didCompleteSide)
    stopHelpTimer(self)
    self.movesSinceSideCompleted = self.movesSinceSideCompleted + 1
    local shouldRemoveHint = self.movesSinceSideCompleted < MOVES_TILL_HINT
    local shouldShowHint = (self.movesSinceSideCompleted == MOVES_TILL_HINT) and not didCompleteSide
    local isHintShowing = self.inGameUI:isHintShowing()

    if shouldRemoveHint then
      removeHint(self)
    elseif shouldShowHint then
      showHint(self)
    elseif isHintShowing then
      self.inGameUI:updateHint(cube)
    end
  end
end

local function getTimerCompletionHandler(self)
  return function()
    onGameLoss(self)
  end
end

local function getInstructionFields(self)
  local fields = {}
  for _, entry in ipairs(self.totemContentTemplate) do
    table.insert(fields, entry.help)
  end
  return fields
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TotemGame:gameFlow()
  runInCoroutine(function()
    wait(self.currentTotem:setConcepts(self.roundGenerator:next()))
    GameSession.contentTracker:notifyPresentedConcept(self.currentTotem:getIdentifier())
    prepareNextTotem(self)
  end)

  self.currentTotem:spin()
  waitFor(0.2)
  self.currentTotem:showConcepts()
  self.completedTotems = 0

  self.background:animate()
  self.inGameUI:showInstructions()
  self.currentTotem:enableTouch()
  startHelpTimer(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local selfW, selfH = self:getDims()
  self.background = Background.new()
  self.background:setLoc(selfW/2, selfH/2)
  self:insert(self.background, Z_BACKGROUND)

  local onDismissed = function()
    self.timer:start()
  end

  self.inGameUI = InGameUI.new(self.iconBar, self.totemContentTemplate, getInstructionFields(self), onDismissed)

  local backgroundGroup = self
  local foregroundGroup = self.inGameUI

  self.currentTotem = Totem.new(self.totemContentTemplate, getTotemFaceChangedHandler(self), getTotemSideCompletionHandler(self), backgroundGroup, foregroundGroup)
  self.currentTotem:setLoc(selfW/2, selfH/2)
  self.totemRootGroup:insert(self.currentTotem, Z_TOTEM)

  self.nextTotem = Totem.new(self.totemContentTemplate, getTotemFaceChangedHandler(self), getTotemSideCompletionHandler(self), backgroundGroup, foregroundGroup)
  self.nextTotem:setLoc(selfW/2, selfH * 1.5, 0)
  self.totemRootGroup:insert(self.nextTotem, Z_TOTEM)

  local gradientColorTop = {111/255, 144/255, 160/255, 0}
  local gradientColorBottom = {111/255, 144/255, 160/255, 0.7}
  self.gradient = RectangleGradient.new(selfW, selfH/2, gradientColorTop, gradientColorBottom)
  self.gradient:setAnchorPointBottomLeft()
  self:insert(self.gradient, Z_GRADIENT)
end

local function createUI(self)
  local selfW, selfH = self:getDims()
  local builder = StandardUIBuilder.new(self:getUIOverlay(), COLOR_UI)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.timer = builder:addTimer(self.timerDuration, getTimerCompletionHandler(self), UIOverlay.NORTHEAST, -2)
  self.bonusTray = builder:addBonusTray()

  local totalSides = SIDES_TO_COMPLETE_TOTEM * TotemConfiguration.TOTEMS_TO_COMPLETE_COUNT
  self.progressIndicator = NumericProgressIndicator.new(FontFaces.DIN_OT_MEDIUM, 20, 0, totalSides)
  builder:addWidget(self.progressIndicator, UIOverlay.NORTHEAST, -3, -24)
end

local function configure3DLayer(self)
  local viewport = MOAIViewport.new()
  viewport:setSize(MOAIEnvironment.horizontalResolution, MOAIEnvironment.verticalResolution)
  viewport:setScale(Screen.width, Screen.height)

  local layer = MOAILayer.new()
  layer:setViewport(viewport)
  layer:setSortMode(MOAILayer.SORT_Z_ASCENDING)

  local camera = MOAICamera.new()
  local focalLength = Screen.width * Host.contentScale
  camera:setNearPlane(0.5)
  camera:setFarPlane(focalLength)
  camera:setFieldOfView(60)
  camera:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5), camera:getFocalLength(focalLength))
  layer:setCamera(camera)

  local rootGroup = Group.newFullScreen()
  rootGroup:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))
  layer:insertProp(rootGroup)
  rootGroup.parentLayer = layer
  layer.rootGroup = rootGroup
  self.totemRootGroup = rootGroup

  Screen.addManagedLayer(layer)
  MOAIRenderMgr.addToRenderTable(layer)

  MOAIGfxDevice.getFrameBuffer():setClearDepth(true)
end

local function configureInGameUILayer(self)
  local viewport = Screen.getViewport()

  local layer = MOAILayer2D.new()
  layer:setViewport(viewport)
  layer:setSortMode(MOAILayer2D.SORT_PRIORITY_ASCENDING)

  local rootGroup = Group.newFullScreen()
  rootGroup:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))
  layer:insertProp(rootGroup)
  rootGroup.parentLayer = layer
  layer.rootGroup = rootGroup

  local inGameUI = self.inGameUI
  rootGroup:insertAtCenter(inGameUI)

  Screen.addManagedLayer(layer)
  MOAIRenderMgr.addToRenderTable(layer)
end

-- Constructor ------------------------

function TotemGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, TotemGame)

  configure3DLayer(self)

  configureGameDifficulty(self)
  self.roundGenerator = TotemRoundGenerator.new()
  self.totemContentTemplate = TotemConfiguration.getTotemContentTemplate()

  createUI(self)
  createGameElements(self)
  configureInGameUILayer(self)

  self.movesSinceSideCompleted = 0

  return self
end
