--- TransitionsAboutScreen
-- @author Created by Steven Mah on May 12, 2016

class("TransitionsAboutScreen")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local HEADER_FONT = Font.new(FontFaces.DIN_OT_LIGHT, 42)
local BODY_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 16)
local BUTTON_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 15)
local EXAMPLE_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 20)

local FIRST_PAGE_FIRST_PASSAGE = "Transitions draw logical connections between sentences and make writing easier to follow."
local FIRST_PAGE_SECOND_PASSAGE = "Context clues help determine the best transitions to use."
local EXAMPLE_TEXT = "The Louvre in Paris holds many priceless artworks. <blank/>, Leonardo da Vinci's famous painting <i>Mona Lisa</i> has been displayed there since 1797."
local EXPLANATION_FIRST_PASSAGE = "The <strong>famous painting</strong> <i>Mona Lisa</i> is an example of the <strong>many priceless artworks</strong> in the museum."
local EXPLANATION_SECOND_PASSAGE = "Therefore, <strong>\"for example\"</strong> is a good transition to use."

local WIDTH = Screen.getWidthPart(0.8)
local SPACING = Screen.getHeightPart(0.025)

local COLORS = {
  HEADER = {1.0, 0.98, 0.93},
  BODY = {1.0, 1.0, 1.0},
  BUTTON = {0.75, 0.89, 0.94, 0.0}
}
local BUTTON_VISIBLY_OPACITY = 0.5

local SLIDE_DURATION = 1.0
local SLIDE_EASE = AHEaseExponentialOut.new()

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function moveLeft(self)
  self.actionInterrupt:insert(
    AHSpawn.new(
      AHLocationXTo.new(self.firstPage, SLIDE_DURATION, SLIDE_EASE, Screen.getWidthPart(-0.5)),
      AHLocationXTo.new(self.secondPage, SLIDE_DURATION, SLIDE_EASE, Screen.getWidthPart(0.5))
    )
  )
end

local function moveRight(self)
  self.actionInterrupt:insert(
    AHSpawn.new(
      AHLocationXTo.new(self.firstPage, SLIDE_DURATION, SLIDE_EASE, Screen.getWidthPart(0.5)),
      AHLocationXTo.new(self.secondPage, SLIDE_DURATION, SLIDE_EASE, Screen.getWidthPart(1.5))
    )
  )
end

local function addVirtualButtonToLabel(label, callback, width, height)
  local labelWidth, labelHeight = label:getDims()
  local buttonPadding = labelWidth
  width = width or labelWidth + buttonPadding
  height = height or labelHeight + buttonPadding

  local button = Button.addVirtualButton(label, width, height, callback)

  return button
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsAboutScreen:getRevealAction(duration, ease)
  return AHSpawn.new(
    AHOpacityTo.new(self, duration, ease, 1.0),
    AHOpacityTo.new(self.tapToContinueButton, duration, ease, BUTTON_VISIBLY_OPACITY),
    AHOpacityTo.new(self.exampleButton, duration, ease, BUTTON_VISIBLY_OPACITY),
    AHOpacityTo.new(self.returnButton, duration, ease, BUTTON_VISIBLY_OPACITY)
  )
end

function TransitionsAboutScreen:getHideAndRemoveAction(duration, ease)
  local hideAction = AHSpawn.new(
    AHOpacityTo.new(self, duration, ease, 0.0),
    AHOpacityTo.new(self.tapToContinueButton, duration, ease, 0.0),
    AHOpacityTo.new(self.exampleButton, duration, ease, 0.0),
    AHOpacityTo.new(self.returnButton, duration, ease, 0.0)
  )

  local removeAction = AHSpawn.new(
    AHRemove.new(self),
    AHRemove.new(self.firstPage),
    AHRemove.new(self.secondPage),
    AHRemove.new(self.tapToContinueButton)
  )

  return AHSequence.new(
    hideAction, removeAction
  )
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function makeButtonWithArrow(label, callback, isLeft)
  local buttonGroup = Row.new(Screen.getWidthPart(0.015))
  local arrow = Sprite.newWithName("arrow.png")
  arrow:inheritColor(label)

  if (isLeft) then
    arrow:setRot(180)
    buttonGroup:insert(arrow)
    buttonGroup:insert(label)
  else
    buttonGroup:insert(label)
    buttonGroup:insert(arrow)
  end

  buttonGroup:alignChildren()
  addVirtualButtonToLabel(buttonGroup, callback)

  return buttonGroup
end

local function createLayout(self, tapToContinueCallback)
  local lineSpacing = Screen.getHeightPart(0.01)
  local firstPage = Group.newFullScreen()
  local aboutHeader = DynamicTextArea.newSimple("ABOUT<br/>THIS GAME", WIDTH, HEADER_FONT)
  aboutHeader:setColor(unpack(COLORS.HEADER))
  local firstPageFirstBody = DynamicTextArea.newSimple(FIRST_PAGE_FIRST_PASSAGE, WIDTH, BODY_FONT, nil, nil, lineSpacing)
  firstPageFirstBody:setColor(unpack(COLORS.BODY))
  local firstPageSecondBody = DynamicTextArea.newSimple(FIRST_PAGE_SECOND_PASSAGE, WIDTH, BODY_FONT, nil, nil, lineSpacing)
  firstPageSecondBody:setColor(unpack(COLORS.BODY))
  local exampleButtonLabel = Label.new("EXAMPLE", BUTTON_FONT)
  exampleButtonLabel:setColor(unpack(COLORS.BUTTON))
  local exampleCallback = function()
    Sound.playSoundEffect("other_buttons")
    moveLeft(self)
  end
  local exampleButton = makeButtonWithArrow(exampleButtonLabel, exampleCallback, false)

  local firstContentColumn = Column.new(nil, Alignment.LEFT)
  firstContentColumn:insert(aboutHeader)
  firstContentColumn:insertSpacer(Screen.getHeightPart(0.04))
  firstContentColumn:insert(firstPageFirstBody)
  firstContentColumn:insertSpacer(Screen.getHeightPart(0.04))
  firstContentColumn:insert(firstPageSecondBody)
  firstContentColumn:insertSpacer(Screen.getHeightPart(0.02))
  firstContentColumn:insert(exampleButton)
  firstContentColumn:alignChildren()
  firstPage:insertAtCenter(firstContentColumn)
  firstPage:linkOpacityAttrToChildren(true)
  firstPage:inheritOpacity(self)
  exampleButton:clearAttrLink(MOAIColor.ATTR_OPACITY)
  exampleButton:setOpacity(0.0)
  self.textLayer:insertAtLocationFractional(firstPage, 0.5, 0.53)
  self.firstPage = firstPage
  self.exampleButton = exampleButton

  local secondPage = Group.newFullScreen()
  local returnButtonLabel = Label.new("RETURN", BUTTON_FONT)
  returnButtonLabel:setColor(unpack(COLORS.BUTTON))
  local returnCallback = function()
    Sound.playSoundEffect("other_buttons")
    moveRight(self)
  end
  local returnButton = makeButtonWithArrow(returnButtonLabel, returnCallback, true)

  local exampleHeader = Label.new("EXAMPLE", EXAMPLE_FONT)
  exampleHeader:setColor(unpack(COLORS.HEADER))
  local white = {1.0, 1.0, 1.0}
  local gray = {0.73, 0.73, 0.75}
  local exampleTextArea = TransitionsPromptTextView.new(EXAMPLE_TEXT, EXAMPLE_FONT, white)
  local baseAttributes = {
    font = BODY_FONT,
    color = gray
  }
  local attributesForLabels = {
    strong = {
      color = white,
      font = Font.new(FontFaces.DIN_OT_BOLD, 16)
    },
    i = {
      font = Font.new(FontFaces.DIN_OT_MEDIUM_ITALIC, 16)
    }
  }
  local exampleFirstExplanation = DynamicTextArea.new(EXPLANATION_FIRST_PASSAGE, WIDTH, baseAttributes, attributesForLabels)
  local exampleSecondExplanation = DynamicTextArea.new(EXPLANATION_SECOND_PASSAGE, WIDTH, baseAttributes, attributesForLabels)

  local secondContentColumn = Column.new(SPACING, Alignment.LEFT)
  secondContentColumn:insert(returnButton)
  secondContentColumn:insertSpacer(Screen.getHeightPart(0.01))
  secondContentColumn:insert(exampleHeader)
  secondContentColumn:insert(exampleTextArea)
  secondContentColumn:insertSpacer(Screen.getHeightPart(0.01))
  secondContentColumn:insert(exampleFirstExplanation)
  secondContentColumn:insert(exampleSecondExplanation)
  secondContentColumn:alignChildren()
  secondPage:insertAtCenter(secondContentColumn)
  secondPage:linkOpacityAttrToChildren(true)
  secondPage:inheritOpacity(self)
  returnButton:clearAttrLink(MOAIColor.ATTR_OPACITY)
  self.textLayer:insertAtLocationFractional(secondPage, 1.5, 0.55)
  self.secondPage = secondPage
  self.returnButton = returnButton

  local tapToContinueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), BUTTON_FONT)
  tapToContinueLabel:setColor(unpack(COLORS.BUTTON))
  addVirtualButtonToLabel(tapToContinueLabel, tapToContinueCallback, Screen.width, Screen.getHeightPart(0.2))
  self.textLayer:insertAtLocationFractional(tapToContinueLabel, 0.5, 0.1)
  self.tapToContinueButton = tapToContinueLabel
end

-- Constructor ------------------------

function TransitionsAboutScreen.new(tapToContinueCallback, textLayer)
  local self = Group.newFullScreen()
  Mixin.mix(self, TransitionsAboutScreen)

  self.actionInterrupt = ActionInterrupt.new()
  self.textLayer = textLayer

  createLayout(self, tapToContinueCallback)
  local isRecursive = true
  self:linkOpacityAttrToChildren(isRecursive)

  return self
end
