--- TransitionsBlendCanvas
-- @author Created by Steven Mah on February 23, 2017

class("TransitionsBlendCanvas")

require("graphics.OffscreenRenderedTexture")
require("utils.ShaderUtils")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SHADER_UNIFORM_INDEX = {
  SAMPLER = 1,
  NORMALIZATION_VALUE_HORIZONTAL = 2,
  NORMALIZATION_VALUE_VERTICAL = 3,
  COLOR = 4,
  LOWER_LEFT_COLOR = 5,
  LOWER_RIGHT_COLOR = 6,
  UPPER_RIGHT_COLOR = 7,
  UPPER_LEFT_COLOR = 8,
  THREE_OPTION = 9,
  LOWER_LEFT_COLOR_THREE_OPTION = 10,
  LOWER_RIGHT_COLOR_THREE_OPTION = 11,
  UPPER_RIGHT_COLOR_THREE_OPTION = 12,
  UPPER_LEFT_COLOR_THREE_OPTION = 13
}

local BLEND_NEUTRAL_TINT = {1.0, 1.0, 1.0, 0.0}
local BLEND_COLOR_PLACEMENTS = {
  LOWER_LEFT_COLOR = "lowerLeftColor",
  LOWER_RIGHT_COLOR = "lowerRightColor",
  UPPER_RIGHT_COLOR = "upperRightColor",
  UPPER_LEFT_COLOR = "upperLeftColor"
}

local THREE_OPTION_COLOR_PLACEMENTS = {
  LOWER_LEFT_COLOR_THREE_OPTION = "threeOptionLowerLeftColor",
  LOWER_RIGHT_COLOR_THREE_OPTION =  "threeOptionLowerRightColor",
  UPPER_RIGHT_COLOR_THREE_OPTION = "threeOptionUpperRightColor",
  UPPER_LEFT_COLOR_THREE_OPTION = "threeOptionUpperLeftColor"
}

local THREE_OPTION_COLORS = {
  LOWER_LEFT_COLOR_THREE_OPTION = {0.65, 0.65, 0.83, 0.9},
  LOWER_RIGHT_COLOR_THREE_OPTION = {0.47, 0.74, 0.83, 0.9},
  UPPER_RIGHT_COLOR_THREE_OPTION = {0.47, 0.74, 0.83, 0.9},
  UPPER_LEFT_COLOR_THREE_OPTION = {0.65, 0.65, 0.83, 0.9}
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function declareColorUniforms(self, shader)
  local defaultTintColor = BLEND_NEUTRAL_TINT

  local blendColors = {}
  for locationKey, locationString in pairs(BLEND_COLOR_PLACEMENTS) do
    local colorUniformKey = locationKey
    local colorUniformIndex = SHADER_UNIFORM_INDEX[colorUniformKey]
    local uniformName = locationString

    shader:declareUniform(colorUniformIndex, uniformName, MOAIShader.UNIFORM_COLOR)

    local colorObject = MOAIColor.new()
    colorObject:setColor(unpack(defaultTintColor))
    shader:setAttrLink(colorUniformIndex, colorObject, MOAIColor.COLOR_TRAIT)
    blendColors[locationKey] = colorObject
  end
  self.blendColors = blendColors

  local threeOptionColors = {}
  for locationKey, locationString in pairs(THREE_OPTION_COLOR_PLACEMENTS) do
    local colorUniformKey = locationKey
    local colorUniformIndex = SHADER_UNIFORM_INDEX[colorUniformKey]
    local uniformName = locationString

    shader:declareUniform(colorUniformIndex, uniformName, MOAIShader.UNIFORM_COLOR)

    local colorObject = MOAIColor.new()
    colorObject:setColor(unpack(defaultTintColor))
    shader:setAttrLink(colorUniformIndex, colorObject, MOAIColor.COLOR_TRAIT)
    threeOptionColors[locationKey] = colorObject
  end
  self.threeOptionColors = threeOptionColors
end

local function getShader(self)
  local shader = MOAIShader.new()
  shader:reserveUniforms(TableUtils.getSize(SHADER_UNIFORM_INDEX))

  shader:declareUniformSampler(SHADER_UNIFORM_INDEX.SAMPLER, "sampler", 1)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_HORIZONTAL, "normalizationValueHorizontal", 0.0)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_VERTICAL, "normalizationValueVertical", 0.0)

  -- NOTE: Using floats because can't seem to pass bools to shader and int comparisons are incorrect
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.COLOR, "doColor", 0.0)
  shader:declareUniformFloat(SHADER_UNIFORM_INDEX.THREE_OPTION, "doThreeOptionColor", 0.0)
  declareColorUniforms(self, shader)

  local normalizationValueHorizontal = (self.group:getWidth() * Host.contentScale / self.textureWidth)
  shader:setAttr(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_HORIZONTAL, normalizationValueHorizontal)
  local normalizationValueVertical = 1.0 - (Screen.height * Host.contentScale / self.textureHeight)
  shader:setAttr(SHADER_UNIFORM_INDEX.NORMALIZATION_VALUE_VERTICAL, normalizationValueVertical)

  shader:setVertexAttribute(1, "position")
  shader:setVertexAttribute(2, "uv")

  local vsh = ShaderUtils.loadShader("TransitionsBlendShader.vsh")
  local fsh = ShaderUtils.loadShader("TransitionsBlendShader.fsh")
  shader:load(vsh, fsh)

  return shader
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsBlendCanvas:getShaderColorsToAction(colors, duration, ease)
  duration = duration or 0.5
  local fullColorAction = AHSpawn.new()
  local colorsTable = self.blendColors
  for colorLocationKey, color in pairs(colors) do
    if (color ~= nil and color ~= {}) then
      local moaiColor = colorsTable[colorLocationKey]
      local colorAction = AHColorTo.new(moaiColor, duration, ease, color)
      fullColorAction:addAction(colorAction)
    end
  end

  return fullColorAction
end

function TransitionsBlendCanvas:getResetShaderColorsAction(duration, ease)
  local colorTable = self.blendColors
  local moaiColor = colorTable["UPPER_LEFT_COLOR"]
  return AHOpacityTo.new(moaiColor, duration, ease, 0.0)
end

function TransitionsBlendCanvas:setBlendEnabled(isEnabled, isThreeOption)
  local toggleValue = 0.0
  if (isEnabled) then
    toggleValue = 1.0
  end
  local blendIndex
  if (isThreeOption) then
    blendIndex = SHADER_UNIFORM_INDEX["THREE_OPTION"]
  else
    blendIndex = SHADER_UNIFORM_INDEX["COLOR"]
  end
  self.shader:setAttr(blendIndex, toggleValue)
end

function TransitionsBlendCanvas:getThreeOptionShaderAction(duration, ease)
  duration = duration or 0.5
  local fullColorAction = AHSpawn.new()
  local colorsTable = self.threeOptionColors
  for colorLocationKey, color in pairs(THREE_OPTION_COLORS) do
    if (color ~= nil and color ~= {}) then
      local moaiColor = colorsTable[colorLocationKey]
      local colorAction = AHColorTo.new(moaiColor, duration, ease, color)
      fullColorAction:addAction(colorAction)
    end
  end

  return fullColorAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createTexture(group)
  local groupWidth = group:getWidth()
  local groupHeight = group:getHeight()
  local contentScale = Host.contentScale
  local textureWidth = MathUtils.incrementUntilPowerOf2(groupWidth * contentScale)
  local textureHeight = MathUtils.incrementUntilPowerOf2(groupHeight * contentScale)
  local texture = OffscreenRenderedTexture.new(group, textureWidth, textureHeight)
  return texture
end

-- Constructor ------------------------

function TransitionsBlendCanvas.new(group)
  local texture = createTexture(group)
  local self = Sprite.newWithTexture(texture)
  Mixin.mix(self, TransitionsBlendCanvas)

  self.group = group
  self.blendColors = {}
  self.threeOptionColors = {}

  self.textureWidth, self.textureHeight = texture:getSize()
  self.shader = getShader(self)
  self:setShader(self.shader)

  self:setBlendEnabled(true)
  self:setBlendEnabled(true, true)

  return self
end
