--- TransitionsButton
-- @author Created by Steven Mah on February 21, 2017

class("TransitionsButton")

require("TransitionsCross")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAX_BACKGROUND_OPACITY = TransitionsCommon.BUTTON_MAX_OPACITY
local BUTTON_COLOR = {1.0, 0.85, 0.8, MAX_BACKGROUND_OPACITY}
local INVISIBLE_BUTTON_COLOR = {1.0, 0.85, 0.8, 0.0}
local OUTLINE_COLOR = {1.0, 1.0, 1.0}
local LABEL_COLOR = {0.22, 0.39, 0.41}
local THREE_OPTION_LABEL_COLOR = {0.2, 0.29, 0.41}
local INCORRECT_LABEL_COLOR = {0.41, 0.22, 0.22}

local WIDTH = Screen.getWidthPart(0.74)
local HEIGHT = Screen.getHeightPart(0.09)
local OUTLINE_PADDING = HEIGHT * 0.5

local Z_INDICES = {
  BACKGROUND = 0,
  OUTLINE = 10,
  TEXT = 20
}

local CROSS_STARTING_SCALE = 1.25

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function inheritTransform(self, object)
  local width, height = self:getDims()
  object:setAttrLink(MOAIProp2D.INHERIT_TRANSFORM, self, MOAIProp2D.TRANSFORM_TRAIT)
  object:setLoc(width / 2, height / 2)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsButton:setData(text, isCorrect)
  self.label:setText(text)
  self.isCorrect = isCorrect
end

function TransitionsButton:getIsCorrect()
  return self.isCorrect
end

function TransitionsButton:getText()
  return self.label:getText()
end

function TransitionsButton:getCorrectAction(duration)
  local background = self.background
  local outline = self.outline
  local color = {1.0, 1.0, 1.0, 0.75}
  local targetWidth, targetHeight = OUTLINE_PADDING + 1.05 * WIDTH, OUTLINE_PADDING + 1.05 * HEIGHT
  local outlineXScale, outlineYScale = outline:scaleForEffectiveDims(targetWidth, targetHeight)
  local scaleEase = AHEaseExponentialOut.new()
  local scaleDownEase = AHEaseExponentialInOut.new()

  local scaleAction = AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(background, duration * 0.75, scaleEase, 1.05),
      AHScaleTo.new(self.label, duration * 0.75, scaleEase, 1.05)
    ),
    AHSpawn.new(
      AHScaleTo.new(background, duration, scaleDownEase, 1.0),
      AHScaleTo.new(self.label, duration, scaleDownEase, 1.0),
      AHColorTo.new(self.label, duration, ease, self.defaultLabelColor)
    )
  )

  local outlineAction = AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(outline, duration, scaleEase, outlineXScale, outlineYScale),
      AHOpacityTo.new(outline, duration + 0.5, nil, 0.0)
    ),
    function()
      outline:setScl(1.0)
      outline:setOpacity(1.0)
    end
  )

  local colorAction = AHSequence.new(
    function()
      background:setColor(unpack(color))
    end,
    AHColorTo.new(background, 2.0, nil, BUTTON_COLOR)
  )

  local correctAction = AHSpawn.new(
    outlineAction,
    colorAction,
    scaleAction
  )

  return correctAction
end

function TransitionsButton:getNotSelectedAction(skipRevert, isIncorrect)
  local triggerDuration = 0.25
  local delayDuration = 0.5
  local ease = nil
  local revertDuration = 2.0 - (triggerDuration + delayDuration)

  local firstPhaseAction = AHSpawn.new(
    AHOpacityTo.new(self.background, triggerDuration, ease, 0.1),
    AHOpacityTo.new(self.label, triggerDuration, ease, 0.25),
    AHOpacityTo.new(self.outline, triggerDuration, ease, 0.25)
  )

  if (isIncorrect) then
    firstPhaseAction:addAction(
      AHColorTo.new(self.label, triggerDuration, ease, INCORRECT_LABEL_COLOR)
    )
  end

  local fullAction = AHSequence.new(
    firstPhaseAction,
    AHDelay.new(delayDuration)
  )

  if (not skipRevert) then
    fullAction:addAction(
      AHSpawn.new(
        AHColorTo.new(self.background, revertDuration, ease, BUTTON_COLOR),
        AHColorTo.new(self.label, revertDuration, ease, self.defaultLabelColor),
        AHOpacityTo.new(self.outline, revertDuration, ease, 1.0)
      )
    )
  end

  return fullAction
end

function TransitionsButton:getIncorrectAction(duration)
  local color = {0.8, 0.2, 0.1, 1.0}
  local cross = self.cross
  duration = duration or 0.5
  local ease = AHEaseSineInOut.new()
  local reducedDuration = duration / 4

  local changeToColor = AHSpawn.new(
    AHColorTo.new(self.background, duration, nil, color),
    AHColorTo.new(self.label, duration, nil, INCORRECT_LABEL_COLOR),
    AHOpacityTo.new(self.outline, duration, nil, 0.15)
  )
  local scaleAction = AHSequence.new(
    AHScaleTo.new(self, reducedDuration, ease, 0.975),
    AHScaleTo.new(self, reducedDuration, ease, 1.0)
  )
  local shakeAction = AHSequence.new(
    AHRotationTo.new(self, reducedDuration, ease, 3),
    AHRotationTo.new(self, reducedDuration, ease, -2),
    AHRotationTo.new(self, reducedDuration, ease, 1),
    AHRotationTo.new(self, reducedDuration, ease, 0)
  )
  local crossAction = cross:getAppearAnimation(duration / 2)

  return AHSpawn.new(
    changeToColor,
    shakeAction,
    scaleAction,
    crossAction
  )
end

function TransitionsButton:getLabel()
  return self.label
end

function TransitionsButton:getFullFadeAction(opacity, duration, ease)
  duration = duration or 0.5
  local backgroundOpacity = MathUtils.clamp(opacity, 0.0, MAX_BACKGROUND_OPACITY)
  local fadeAction = AHSpawn.new(
    AHOpacityTo.new(self.background, duration, ease, backgroundOpacity),
    AHOpacityTo.new(self.outline, duration, ease, opacity),
    AHOpacityTo.new(self.label, duration, ease, opacity),
    AHColorTo.new(self.label, duration, ease, self.defaultLabelColor),
    self.cross:getHideAnimation(duration)
  )

  return fadeAction
end

function TransitionsButton:getResetAction(duration, ease)
  duration = duration or 0.5
  local fadeAction = AHSpawn.new(
    AHColorTo.new(self.background, duration, ease, BUTTON_COLOR),
    AHColorTo.new(self.label, duration, ease, self.defaultLabelColor),
    AHOpacityTo.new(self.outline, duration, ease, 1.0),
    function()
      self.cross:setScl(CROSS_STARTING_SCALE)
    end
  )

  return fadeAction
end

function TransitionsButton:getResizeBackingHeight(targetHeight, duration, ease)
  return AHSpawn.new(
    AHDimYTo.new(self.background, duration, ease, targetHeight),
    AHDimYTo.new(self.outline, duration, ease, targetHeight)
  )
end

function TransitionsButton:getFadeOutAndResetAction(duration, ease)
  return AHSpawn.new(
    AHColorTo.new(self.label, duration, ease, self.defaultLabelColor),
    AHColorTo.new(self.background, duration, ease, INVISIBLE_BUTTON_COLOR),
    AHOpacityTo.new(self.outline, duration, ease, 0.0)
  )
end

function TransitionsButton:switchToThreeOptionLabelColor()
  self.defaultLabelColor = THREE_OPTION_LABEL_COLOR
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createCross()
  local length = WIDTH * 0.33
  local crossArmColor = {1.0, 1.0, 1.0}
  local cross = TransitionsCross.new(length, crossArmColor)
  cross:setRot(45)

  return cross
end

local function createLayout(self)
  local width, height = self:getDims()
  local background = FilledRectangle.new(width, height, BUTTON_COLOR)
  background:setBlendMode(MOAIProp2D.BLEND_ADD)
  self:insertAtCenter(background, Z_INDICES.BACKGROUND)
  self.background = background

  local outline = HollowRectangle.new(width, height, 1.0, nil, OUTLINE_COLOR)
  outline:setBlendMode(MOAIProp2D.BLEND_ADD)
  self:insertAtCenter(outline, Z_INDICES.OUTLINE)
  self.outline = outline

  local font = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 21.0)
  local wordLabel = Label.new("", font)
  wordLabel:setColor(unpack(self.defaultLabelColor))
  self.textLayer:insert(wordLabel, Z_INDICES.TEXT)
  inheritTransform(self, wordLabel)
  wordLabel:setOpacity(0.0)
  self.label = wordLabel

  local cross = createCross()
  cross:setOpacity(0.0)
  cross:setScl(CROSS_STARTING_SCALE)
  self.textLayer:insert(cross, Z_INDICES.TEXT)
  inheritTransform(self, cross)
  self.cross = cross
end

-- Constructor ------------------------

function TransitionsButton.new(textLayer)
  local self = Group.new(WIDTH, HEIGHT)
  Mixin.mix(self, TransitionsButton)

  self.textLayer = textLayer
  self.defaultLabelColor = LABEL_COLOR

  createLayout(self)

  self.isCorrect = false

  return self
end
