--- TransitionsCubeVertices
-- @author Created by Steven Mah on May 8, 2017

class("TransitionsCubeVertices")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH, HEIGHT = TransitionsCommon.CUBE_WIDTH, TransitionsCommon.CUBE_HEIGHT
local BACK_FACE_PADDING = WIDTH * 0.1
local TOP_BACK_FACE_PADDING = BACK_FACE_PADDING * 3

local LEFT_X = 0
local RIGHT_X = WIDTH
local TOP_Y = HEIGHT
local BOTTOM_Y = 0

local BACK_LEFT_X = 0 + BACK_FACE_PADDING
local BACK_RIGHT_X = WIDTH - BACK_FACE_PADDING
local BACK_TOP_Y = HEIGHT - TOP_BACK_FACE_PADDING
local BACK_BOTTOM_Y = 0 + BACK_FACE_PADDING

local MIDDLE_X = WIDTH / 2

local VERTICES_BY_KEYFRAME = {
  {
    { x = LEFT_X, y = BOTTOM_Y },
    { x = LEFT_X, y = TOP_Y },
    { x = RIGHT_X, y = TOP_Y },
    { x = RIGHT_X, y = BOTTOM_Y },
    { x = BACK_LEFT_X, y = BACK_BOTTOM_Y},
    { x = BACK_LEFT_X, y = BACK_TOP_Y },
    { x = BACK_RIGHT_X, y = BACK_TOP_Y },
    { x = BACK_RIGHT_X, y = BACK_BOTTOM_Y }
  },
  {
    { x = LEFT_X, y = (BACK_BOTTOM_Y + BOTTOM_Y) / 2 },
    { x = LEFT_X, y = (BACK_TOP_Y + TOP_Y) / 2 },
    { x = MIDDLE_X, y = TOP_Y + BACK_FACE_PADDING },
    { x = MIDDLE_X, y = BOTTOM_Y - BACK_FACE_PADDING / 2 },
    { x = MIDDLE_X, y = BACK_BOTTOM_Y + BACK_FACE_PADDING / 4 },
    { x = MIDDLE_X, y = BACK_TOP_Y - BACK_FACE_PADDING / 4 },
    { x = RIGHT_X, y = (BACK_TOP_Y + TOP_Y) / 2 },
    { x = RIGHT_X, y = (BACK_BOTTOM_Y + BOTTOM_Y) / 2 }
  },
  {
    { x = BACK_LEFT_X, y = BACK_BOTTOM_Y},
    { x = BACK_LEFT_X, y = BACK_TOP_Y },
    { x = LEFT_X, y = TOP_Y },
    { x = LEFT_X, y = BOTTOM_Y },
    { x = BACK_RIGHT_X, y = BACK_BOTTOM_Y },
    { x = BACK_RIGHT_X, y = BACK_TOP_Y },
    { x = RIGHT_X, y = TOP_Y },
    { x = RIGHT_X, y = BOTTOM_Y }
  }
}

TransitionsCubeVertices.DISTANCES_BETWEEN_KEYFRAMES = {
  BACK_FACE_PADDING,
  WIDTH / 2,
  WIDTH / 2,
  BACK_FACE_PADDING
}

TransitionsCubeVertices.PRECALCULATED_VERTICES = {}

local DEFAULT_SPIN_DURATION = 1.0
local FPS = 60

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function compileVerticesForFaces()
  -- NOTE: Face names are by starting positions
  local compiledVertices = {
    FRONT = {},
    BACK = {},
    LEFT = {},
    RIGHT = {},
    TOP = {},
    BOTTOM = {}
  }

  for _, vertices in ipairs(VERTICES_BY_KEYFRAME) do
    local front = {
      vertices[1],
      vertices[2],
      vertices[3],
      vertices[4]
    }
    table.insert(compiledVertices.FRONT, front)

    local back = {
      vertices[5],
      vertices[6],
      vertices[7],
      vertices[8]
    }
    table.insert(compiledVertices.BACK, back)

    local left = {
      vertices[1],
      vertices[2],
      vertices[6],
      vertices[5]
    }
    table.insert(compiledVertices.LEFT, left)

    local right = {
      vertices[3],
      vertices[4],
      vertices[8],
      vertices[7]
    }
    table.insert(compiledVertices.RIGHT, right)

    local top = {
      vertices[2],
      vertices[3],
      vertices[7],
      vertices[6]
    }
    table.insert(compiledVertices.TOP, top)

    local bottom = {
      vertices[1],
      vertices[4],
      vertices[8],
      vertices[5]
    }
    table.insert(compiledVertices.BOTTOM, bottom)
  end

  return compiledVertices
end

local function preinterpolateVertices(vertexKeyFrames)
  local vertices = {}
  local frameCount = FPS * DEFAULT_SPIN_DURATION

  for keyframeIndex = 1, #vertexKeyFrames - 1 do
    local currentKeyframe = vertexKeyFrames[keyframeIndex]
    local nextKeyframe = vertexKeyFrames[keyframeIndex + 1]

    local framesPerKeyframePair = frameCount / (#vertexKeyFrames - 1)
    for frame = 1, framesPerKeyframePair do
      local frameVertices = {}
      local interpolationValue = (frame - 1) / framesPerKeyframePair
      for vertexIndex, currentVertex in ipairs(currentKeyframe) do
        local nextKeyframeVertex = nextKeyframe[vertexIndex]
        local blendedX = MathUtils.interpolateLinearFromNormalized(interpolationValue, currentVertex.x, nextKeyframeVertex.x)
        local blendedY = MathUtils.interpolateLinearFromNormalized(interpolationValue, currentVertex.y, nextKeyframeVertex.y)
        table.insert(frameVertices, blendedX)
        table.insert(frameVertices, blendedY)
      end
      table.insert(vertices, frameVertices)
    end
  end

  return vertices
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsCubeVertices.getBackFacePadding()
  return BACK_FACE_PADDING
end

function TransitionsCubeVertices.getApexOffset()
  return APEX_OFFSET
end

function TransitionsCubeVertices.getPrecalculatedFrameVertices()
  if (#TransitionsCubeVertices.PRECALCULATED_VERTICES == 0) then
    local verticesByFace = compileVerticesForFaces()
    local finalComputedVertices = {}
    for cubeFaceName, vertexKeyFrames in pairs(verticesByFace) do
      local vertexFrames = preinterpolateVertices(vertexKeyFrames)
      finalComputedVertices[cubeFaceName] = vertexFrames
    end
    TransitionsCubeVertices.PRECALCULATED_VERTICES = finalComputedVertices
  end

  assert(TableUtils.getSize(TransitionsCubeVertices.PRECALCULATED_VERTICES) > 0, "Cube vertices are empty.")
  return TransitionsCubeVertices.PRECALCULATED_VERTICES
end
