--- TransitionsDiamondTexture
-- @author Created by Steven Mah on May 8, 2017.

class("TransitionsDiamondTexture")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function drawLine(imageDataSource, bounds, color, doTaper)
  local startRow = bounds.startRow
  local endRow = bounds.endRow
  local startColumn = bounds.startColumn
  local endColumn = bounds.endColumn

  local isHorizontal = (endColumn - startColumn > endRow - startRow)
  local middleRow = (endRow + startRow) / 2
  local middleColumn = (endColumn + startColumn) / 2

  for row = startRow, endRow do
    for column = startColumn, endColumn do
      local alpha = 1.0
      local newColor = TableUtils.deepCopy(color)

      if (doTaper) then
        if (isHorizontal) then
          alpha = 1.0 - math.abs(column - middleColumn) / (endColumn - middleColumn)
        else
          alpha = 1.0 - math.abs(row - middleRow) / (endRow - middleRow)
        end

        local r, g, b, a = imageDataSource:getRGBA(column, row)
        local currentColor = {r, g, b, a}

        for i = 1, #newColor do
          newColor[i] = math.max(newColor[i] * alpha, currentColor[i])
        end
      end

      imageDataSource:setRGBA(column, row, unpack(newColor))
    end
  end
end

local function drawGlow(imageDataSource, bounds, glowColor, glowMarginRowCount, glowMarginColumnCount)
  local startRow = bounds.startRow
  local endRow = bounds.endRow
  local startColumn = bounds.startColumn
  local endColumn = bounds.endColumn

  local isHorizontal = (endColumn - startColumn > endRow - startRow)
  local middleRow = (endRow + startRow) / 2
  local middleColumn = (endColumn + startColumn) / 2
  local doTaper = true

  if (isHorizontal) then
    for column = startColumn, endColumn do
      local lineBounds = {
        startRow = middleRow - glowMarginColumnCount,
        endRow = middleRow + glowMarginColumnCount,
        startColumn = column,
        endColumn = column
      }
      drawLine(imageDataSource, lineBounds, glowColor, doTaper)
    end
  else
    for row = startRow, endRow do
      local lineBounds = {
        startRow = row,
        endRow = row,
        startColumn = middleColumn - glowMarginRowCount,
        endColumn = middleColumn + glowMarginRowCount
      }
      drawLine(imageDataSource, lineBounds, glowColor, doTaper)
    end
  end
end

local function drawCornerGlow(imageDataSource, cornerCenter, glowColor, glowLength)
  local centerRow = cornerCenter.row
  local centerColumn = cornerCenter.column

  local startRow = centerRow - glowLength * 0.5
  local endRow = startRow + glowLength
  local startColumn = centerColumn - glowLength * 0.5
  local endColumn = startColumn + glowLength

  for row = startRow, endRow do
    for column = startColumn, endColumn do
      local alpha = math.max(0.0, 1.0 - MathUtils.distance(column, row, centerColumn, centerRow) / (0.5 * glowLength))
      local newColor = TableUtils.deepCopy(glowColor)

      local r, g, b, a = imageDataSource:getRGBA(column, row)
      local currentColor = {r, g, b, a}

      for i = 1, #newColor do
        newColor[i] = math.max(newColor[i] * alpha, currentColor[i])
      end

      imageDataSource:setRGBA(column, row, unpack(newColor))
    end
  end
end

local function populateImage(imageDataSource, parameters)
  local width, height = imageDataSource:getSize()

  local glowMarginFraction = parameters.glowMarginFraction or 0.0
  local glowMarginRowCount = math.floor(height * glowMarginFraction)
  local glowMarginColumnCount = math.floor(width * glowMarginFraction)

  local lineWidthFraction = parameters.lineWidthFraction or 0.1
  local lineWidthRowCount = math.floor(height * lineWidthFraction)
  local lineWidthColumnCount = math.floor(width * lineWidthFraction)

  local lineColor = parameters.lineColor or {1.0, 1.0, 1.0, 1.0}
  local glowColor = parameters.glowColor or {0.0, 1.0, 0.0, 1.0}
  local lineMinimumX = glowMarginColumnCount
  local lineMaximumX = height - glowMarginColumnCount - 1
  local lineMinimumY = glowMarginRowCount
  local lineMaximumY = width - glowMarginRowCount - 1

  -- row/column is (0, 0) at upper-left

  if (parameters.showGlow) then
    local cornerCenters = {
      -- upper-left
      {
        row = lineMinimumY + lineWidthRowCount * 0.5,
        column = lineMinimumX + lineWidthColumnCount * 0.5
      },
      -- upper-right
      {
        row = lineMinimumY + lineWidthRowCount * 0.5,
        column = lineMaximumX - lineWidthColumnCount * 0.5
      },
      -- lower-right
      {
        row = lineMaximumY - lineWidthRowCount * 0.5,
        column = lineMaximumX - lineWidthColumnCount * 0.5
      },
      -- lower-left
      {
        row = lineMaximumY - lineWidthRowCount * 0.5,
        column = lineMinimumX + lineWidthColumnCount * 0.5
      }
    }

    for _, cornerCenter in ipairs(cornerCenters) do
      local glowLength = 2.0 * math.max(glowMarginRowCount, glowMarginColumnCount)
      drawCornerGlow(imageDataSource, cornerCenter, glowColor, glowLength)
    end
  end

  local lineBounds = {
    -- top piece
    {
      startRow = lineMinimumY,
      endRow = lineMinimumY + lineWidthRowCount,
      startColumn = lineMinimumX,
      endColumn = lineMaximumX
    },
    -- bottom piece
    {
      startRow = lineMaximumY - lineWidthRowCount,
      endRow = lineMaximumY,
      startColumn = lineMinimumX,
      endColumn = lineMaximumX
    },
    -- left piece
    {
      startRow = lineMinimumY,
      endRow = lineMaximumY,
      startColumn = lineMinimumX,
      endColumn = lineMinimumX + lineWidthColumnCount
    },
    -- right piece
    {
      startRow = lineMinimumY,
      endRow = lineMaximumY,
      startColumn = lineMaximumX - lineWidthColumnCount,
      endColumn = lineMaximumX
    }
  }

  if (parameters.showGlow) then
    for _, bounds in ipairs(lineBounds) do
      drawGlow(imageDataSource, bounds, glowColor, glowMarginRowCount, glowMarginColumnCount)
    end
  end

  -- optional inner fill
  if (parameters.fillCenter) then
    local centerRow = height / 2
    local centerColumn = width / 2
    local maximumDistance = 1.41 * (lineMaximumY - lineMinimumY) * 3.0
    for row = lineMinimumY + lineWidthRowCount, lineMaximumY - lineWidthRowCount do
      for column = lineMinimumX + lineWidthColumnCount, lineMaximumX - lineWidthColumnCount do
        local color = TableUtils.deepCopy(parameters.fillColor)
        local alpha = 0.9 - MathUtils.distance(column, row, centerColumn, centerRow) / maximumDistance
        for i = 1, #color do
          color[i] = color[i] * alpha
        end
        imageDataSource:setRGBA(column, row, unpack(color))
      end
    end
  end

  -- square line
  local doTaper = false
  for _, bounds in ipairs(lineBounds) do
    drawLine(imageDataSource, bounds, lineColor, doTaper)
  end
end

local function getDataSource(parameters)
  local rowCount = parameters.rowCount or 32
  local columnCount = parameters.columnCount or 32

  local dataSource = MOAIImage.new()
  dataSource:init(columnCount, rowCount)
  populateImage(dataSource, parameters)
  return dataSource
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TransitionsDiamondTexture.new(parameters)
  local self = MOAITexture.new()
  Mixin.mix(self, TransitionsDiamondTexture)

  local dataSource = getDataSource(parameters)
  self:load(dataSource)

  return self
end
