--- TransitionsGame
-- @author Created by Steven Mah on October 20, 2016.

class("TransitionsGame")

-- NOTE: For iPad scaling
FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("TransitionsCommon")
require("TransitionsConfiguration")
require("TransitionsScene")
require("TransitionsRoundGenerator")
require("bonuses.ResponseSpeedTracker")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {1, 1, 1}

local TIME_PENALTY = TransitionsConfiguration.getTimePenalty()
local POINTS_PER_CORRECT_ANSWER = TransitionsConfiguration.getPointsPerAnswer()

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function showAndWaitForInterstitial(self, class)
  self.canStart = false
  local waitCallback = function()
    return self.canStart
  end
  local tapToContinueCallback = function()
    Sound.playSoundEffect("other_buttons")
    self.canStart = true
  end

  wait(self.scene:showAndWaitForInterstitial(class, waitCallback, tapToContinueCallback):start())
end

local function startNewRound(self)
  self.submittedAnswer = nil
  local switchedToThree = self.roundGenerator:switchToThreeOptionsIfNeeded(self.numberCorrect)
  self.round = self.roundGenerator:getNextRound()

  local presentationAnimation
  if (self.previousRoundCorrect == nil) then
    presentationAnimation = AHStaggeredSpawn.new(
      0.1,
      function()
        Sound.playSoundEffect("return_to_game")
      end,
      self.scene:getIntroAction(self.round),
      self.uiOverlay:createFadeInAction()
    )
  elseif (self.previousRoundCorrect) then
    presentationAnimation = self.scene:getRevealRoundAfterCorrectResponse(self.round, switchedToThree)
  else
    presentationAnimation = self.scene:getRevealRoundAfterIncorrectResponse(self.round)
    Sound.playSoundEffect("return_to_game")
  end

  wait(presentationAnimation:start())
  self.scene:setTouchEnabled(true)
  self.timer:resumeAndHideIcon()
  self.responseSpeedTracker:notifyStart()
end

local function waitForResponseOrTimeout(self)
  local responseSubmittedOrTimeOut = function()
    return (self.submittedAnswer or self.isOutOfTime)
  end
  waitUntil(responseSubmittedOrTimeOut)
  self.scene:setTouchEnabled(false)
end

local function processResponse(self)
  local correctAnswer = self.round.correct
  local userAnsweredCorrectly = (self.submittedAnswer == correctAnswer)
  self.previousRoundCorrect = userAnsweredCorrectly

  if (self.timer:getTimeRemaining() <= 2.0) then
    self.isOutOfTime = true
  end

  local responseAction
  if (userAnsweredCorrectly) then
    Sound.playSoundEffect("right_answer", 0.4)
    self.numberCorrect = self.numberCorrect + 1
    local willSwitchToThree = (self.numberCorrect == TransitionsConfiguration.getCorrectConceptsBeforeThreeOption()) and (TransitionsConfiguration.getIsHighLevel())
    GameSession.increaseScore(POINTS_PER_CORRECT_ANSWER)
    self.responseSpeedTracker:notifyCorrectResponse()
    self.scoreLabel:setScore(GameSession.getScore())
    GameSession.contentTracker:notifyCorrectAnswer(self.round.identifier)
    GameSession.incrementCorrectAnswerCount()
    responseAction = self.scene:getCorrectResponseAction(correctAnswer, willSwitchToThree, self.isOutOfTime)
  else
    Sound.playSoundEffect("error", 0.3)
    self.responseSpeedTracker:notifyIncorrectResponse()
    self.timer:addTime(TIME_PENALTY)
    self.timer:pauseAndShowIcon(nil)
    if (self.timer:getTimeRemaining() <= 1.0) then
      self.isOutOfTime = true
    end
    GameSession.contentTracker:notifyIncorrectAnswer(self.round.identifier, self.submittedAnswer)
    GameSession.incrementIncorrectAnswerCount()
    responseAction = self.scene:getIncorrectReponseAction(correctAnswer, self.submittedAnswer)
  end

  wait(responseAction:start())
end

local function doOutro(self)
  if (GameSession.getScore() > 0) then
    local outroAction = self.scene:getOutroAnimation(self.previousRoundCorrect, self.submittedAnswer)
    self.timer:resumeAndHideIcon()
    self.uiOverlay:fadeOut()
    outroAction:start()
    waitFor(4.0) -- NOTE: animation loops indefinitely; give it some time before ending game
    outroAction:stop()
    GameSession.notifyGameWin()
  else
    local duration = 0.25
    local clearScreenAction = AHSpawn.new(
      AHOpacityTo.new(self.timer:getPausedLabel(), duration, nil, 0.0),
      self.uiOverlay:createFadeOutAction(duration),
      self.scene:getClearSceneAnimation(duration)
    )

    wait(clearScreenAction:start())
    waitFor(0.25)
    GameSession.notifyGameLoss()
  end
end

local function showAndWaitforPregameOverlays(self)
  local hasNeverWon = GameSession.hasNeverWon()
  local shouldShowHighDifficultyInterstitial = TransitionsConfiguration.shouldShowHighDifficultyInterstitial()
  if (hasNeverWon or shouldShowHighDifficultyInterstitial) then
    wait(self.scene:getShowBlueTintAction(0.0):start())
  end
  self.scene:getFadeOutOverlayAction(0.15):start()

  if (hasNeverWon) then
    showAndWaitForInterstitial(self, TransitionsAboutScreen)
  end
  if (shouldShowHighDifficultyInterstitial) then
    showAndWaitForInterstitial(self, TransitionsHighLevelScreen)
  end

  self.scene:getResetShaderColorsAction(0.5):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsGame:isUIAutoFadeInEnabled()
  return false
end

function TransitionsGame:gameFlow()
  showAndWaitforPregameOverlays(self)

  while (not self.isOutOfTime) do
    startNewRound(self)
    waitForResponseOrTimeout(self)

    if (self.submittedAnswer) then
      processResponse(self)
    end
  end

  doOutro(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local rankTwoScore = TransitionsConfiguration.getRankTwoScore()
  local rankThreeScore = TransitionsConfiguration.getRankThreeScore()
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = TransitionsConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  local roundGenerator = TransitionsRoundGenerator.new(GameSession.conceptGenerator, difficultyNameChooser)
  self.roundGenerator = roundGenerator

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end

local function createGameElements(self)
  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 TransitionsConfiguration.getPointsPerFullSpeedBonus(),
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(TransitionsConfiguration.getSpeedBonusWindow(),
                                                       quickResponseWrapper,
                                                       TransitionsConfiguration.getSpeedBonusDecayStart())

  local onResponseCallback = function(transitionWord)
    self.submittedAnswer = transitionWord
  end

  self.scene = TransitionsScene.new(onResponseCallback, self.round)
  self:insertAtCenter(self.scene)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)
  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local gameDuration = TransitionsConfiguration.getGameDuration()
  local onTimeOut = function()
    self.isOutOfTime = true
  end

  self.timer = builder:addDigitTimer(gameDuration, onTimeOut)
  self.timer:initialize()
end

-- Constructor ------------------------

function TransitionsGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, TransitionsGame)

  self.previousRoundCorrect = nil
  self.numberCorrect = 0

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
