--- TransitionsGlitchyLine
-- @author Created by Steven Mah on May 10, 2017.

class("TransitionsGlitchyLine")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

TransitionsGlitchyLine.DIRECTIONS = {
  UP = 1,
  RIGHT = 2,
  LEFT = 3
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getLineSegment(self, length, direction)
  local width = self:getWidthPart(0.01)
  local randomOpacity = MathUtils.getRandomFloat(0.1, 0.75)
  local color = {1.0, 1.0, 1.0, randomOpacity}
  local line = FilledRectangle.new(width, length, color)
  line:setAnchorPointBottom()
  line:setScl(1.0, 0.0)

  if (direction == TransitionsGlitchyLine.DIRECTIONS.RIGHT) then
    line:setRot(-90.0)
  elseif (direction == TransitionsGlitchyLine.DIRECTIONS.LEFT) then
    line:setRot(90.0)
  end

  return line
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsGlitchyLine:startAnimation()
  self.animation:start()
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, turnDirection)
  local segmentCount = 2
  local verticalLength = self:getHeightPart(0.8)
  local horizontalLength = 0.35 * verticalLength
  local position = {
    x = self:getWidthPart(0.5),
    y = 0.0
  }

  local direction = TransitionsGlitchyLine.DIRECTIONS.UP
  local lines = {}

  for i = 1, segmentCount do
    local length = horizontalLength
    if (direction == TransitionsGlitchyLine.DIRECTIONS.UP) then
      length = verticalLength
    end

    local line = getLineSegment(self, length, direction)
    line:setLoc(position.x, position.y)
    self:insert(line)
    table.insert(lines, line)

    local goingRight = (direction == TransitionsGlitchyLine.DIRECTIONS.RIGHT)
    local goingLeft = (direction == TransitionsGlitchyLine.DIRECTIONS.LEFT)

    -- determine next line starting point
    if (goingRight) then
      position.x = position.x + length
    elseif (goingLeft) then
      position.x = position.x - length
    else
      position.y = position.y + length
    end

    -- determine next line direction
    if (goingRight or goingLeft) then
      direction = TransitionsGlitchyLine.DIRECTIONS.UP
    else
      direction = turnDirection
    end

    verticalLength = verticalLength * 0.75
    horizontalLength = horizontalLength * 0.75
  end

  local duration = 0.025
  local delay = 0.0
  self.animation = AHSpawn.new()

  for _, line in ipairs(lines) do
    local scale = AHScaleYTo.new(line, duration, nil, 1.0)
    local delayedAnimation = AHDelayed.new(delay, scale)
    self.animation:addAction(delayedAnimation)

    delay = delay + duration
    duration = duration * 0.75
  end

  local scaleDuration = 20.0 * duration * segmentCount
  local scaleEase = AHEaseBounceOut.new()
  local destinationScale = 1.5
  local scale = AHScaleYTo.new(self, scaleDuration, scaleEase, destinationScale)

  local maximumScale = destinationScale + MathUtils.getRandomFloat(0.00625, 0.0125)
  local expand = AHScaleYTo.new(self, duration, nil, maximumScale)
  local collapse = AHScaleYTo.new(self, duration, nil, destinationScale)
  local glitchOnce = AHSequence.new(expand, collapse)
  local glitchCount = math.random(2, 5)
  local glitch = AHLoop.new(glitchOnce, glitchCount)
  local delay = MathUtils.getRandomFloat(1.0, 3.0)
  local wait = AHDelay.new(delay)
  local glitchAndWait = AHSequence.new(glitch, wait)
  local glitchLoop = AHLoop.new(glitchAndWait)

  local scaleThenGlitch = AHSequence.new(scale, AHDelay.new(delay), glitchLoop)
  self.animation:addAction(scaleThenGlitch)

  self:setAnchorPointBottom()
end

-- Constructor ------------------------

function TransitionsGlitchyLine.new(width, turnDirection)
  local self = Group.new(width, Screen.getWidthPart(0.5))
  Mixin.mix(self, TransitionsGlitchyLine)

  createLayout(self, turnDirection)

  return self
end
