--- TransitionsMenu
-- @author Created by Steven Mah on October 21, 2016.

class("TransitionsMenu")

require("TransitionsButton")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SPACING = Screen.getHeightPart(0.035)

local MAX_BACKGROUND_OPACITY = TransitionsCommon.BUTTON_MAX_OPACITY

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTouchCallback(self, button)
  local onTouchDown = function()
    self:setTouchEnabled(false)
    self.selectedButton = button
    self.onResponseCallback(button:getText())
  end

  return onTouchDown
end

local function makeButtons(self, buttonCount)
  self.buttons = {}
  local spacing = SPACING
  local x = Screen.getWidthPart(0.5)
  local y = self:getHeightPart(0.5) - Screen.getHeightPart(0.05) - spacing / 2
  for buttonIndex = 1, buttonCount do
    local button = TransitionsButton.new(self.textLayer)
    button:setLoc(x, y)
    y = y + Screen.getHeightPart(0.09) + spacing / 2
    self:insert(button)
    local onTouchDown = getTouchCallback(self, button)
    Button.makeButton(button, onTouchDown)
    table.insert(self.buttons, button)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsMenu:getCorrectAction(triggerDuration, revertDuration)
  triggerDuration = triggerDuration or 0.8
  revertDuration = revertDuration or 0.5

  local correctAction = AHSpawn.new(
    self.selectedButton:getCorrectAction(triggerDuration, revertDuration)
  )

  for _, button in ipairs(self.buttons) do
    if (button ~= self.selectedButton) then
      correctAction:addAction(
        button:getNotSelectedAction()
      )
    end
  end

  return correctAction
end

function TransitionsMenu:getIncorrectAction(duration)
  local selectedButtonAction = self.selectedButton:getIncorrectAction(duration)
  local skipRevert = true
  local isIncorrect = true
  local otherAction = AHSpawn.new()
  for _, button in ipairs(self.buttons) do
    if (button ~= self.selectedButton) then
      otherAction:addAction(
        button:getNotSelectedAction(skipRevert, isIncorrect)
      )
    end
  end

  return AHSpawn.new(
    selectedButtonAction,
    otherAction
  )
end

function TransitionsMenu:getRevealThirdButtonAction(duration, ease)
  local bottomButton = TransitionsButton.new(self.textLayer)
  bottomButton:switchToThreeOptionLabelColor()
  local middleButton = self.buttons[1]
  local topButton = self.buttons[2]
  bottomButton:setLoc(middleButton:getX(), middleButton:getY())
  wait(bottomButton:getFullFadeAction(0.0, 0.0):start()) -- Make sure it's invisble before it gets added
  self:insert(bottomButton)

  local onTouchDown = getTouchCallback(self, bottomButton)
  Button.makeButton(bottomButton, onTouchDown)
  bottomButton:setTouchEnabled(false)
  table.insert(self.buttons, bottomButton)

  local scale = 0.85
  local newHeight = topButton:getHeightPart(scale)
  local buttonResizeAction = AHSpawn.new(
    topButton:getResizeBackingHeight(newHeight, duration, ease),
    middleButton:getResizeBackingHeight(newHeight, duration, ease),
    bottomButton:getResizeBackingHeight(newHeight, duration, ease)
  )

  local bottomYProportion = 0.05
  local fullRevealAction = AHSpawn.new(
    buttonResizeAction,
    AHLocationYTo.new(self.buttons[2], duration, ease, Screen.getHeightPart(bottomYProportion + 0.2) * scale + SPACING / 3),
    AHLocationYTo.new(middleButton, duration, ease, Screen.getHeightPart(bottomYProportion + 0.1) * scale + SPACING / 6),
    AHLocationYTo.new(bottomButton, duration, ease, Screen.getHeightPart(bottomYProportion) * scale),
    AHOpacityTo.new(bottomButton.background, duration, ease, MAX_BACKGROUND_OPACITY),
    AHOpacityTo.new(bottomButton.outline, duration, ease, 1.0)
  )

  return fullRevealAction
end

function TransitionsMenu:swapWords(round)
  local replacementOptions = TableUtils.createShuffledArray(round.replacements)
  local correctReplacement = StringUtils.capitalizeFirstLetter(round.correct)

  -- Edge case when testing concepts
  while (#replacementOptions > #self.buttons) do
    TableUtils.removeValueFromArray(replacementOptions, correctReplacement)
    table.remove(replacementOptions)
    table.insert(replacementOptions, correctReplacement)
    TableUtils.shuffleArray(replacementOptions)
  end

  local replaceWordsAction = AHSpawn.new()
  for index, word in ipairs(replacementOptions) do
    local button = self.buttons[index]
    local capitalizedWord = StringUtils.capitalizeFirstLetter(word)
    button:setData(capitalizedWord, (word == correctReplacement))
  end
end

function TransitionsMenu:setTouchEnabled(isEnabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(isEnabled)
  end
end

function TransitionsMenu:getFadeButtonLabelsAction(opacity, duration, ease)
  duration = duration or 0.5
  local fadeAction = AHSpawn.new()
  for _, button in ipairs(self.buttons) do
    fadeAction:addAction(
      AHOpacityTo.new(button:getLabel(), duration, ease, opacity)
    )
  end

  return fadeAction
end

function TransitionsMenu:getFullFadeAction(opacity, duration, ease)
  duration = duration or 0.5
  local fadeAction = AHSpawn.new()
  for _, button in ipairs(self.buttons) do
    fadeAction:addAction(
      button:getFullFadeAction(opacity, duration, ease)
    )
  end

  return fadeAction
end

function TransitionsMenu:getResetAction(duration, ease)
  local fullResetAction = AHSpawn.new()
  for _, button in ipairs(self.buttons) do
    fullResetAction:addAction(
      button:getResetAction(duration, ease)
    )
  end

  return fullResetAction
end

function TransitionsMenu:getFadeOutAndResetAction(duration)
  local fadeAction = AHSpawn.new()
  for _, button in ipairs(self.buttons) do
    fadeAction:addAction(
      button:getFadeOutAndResetAction(duration)
    )
  end

  return fadeAction
end

function TransitionsMenu:switchToThreeOptionLabelColor()
  for _, button in ipairs(self.buttons) do
    button:switchToThreeOptionLabelColor()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local buttonCount = 2
  makeButtons(self, buttonCount)
end

-- Constructor ------------------------

function TransitionsMenu.new(onResponseCallback, textLayer)
  local width, height = Screen.width, Screen.getHeightPart(0.35)
  local self = Group.new(width, height)
  Mixin.mix(self, TransitionsMenu)

  self.onResponseCallback = onResponseCallback
  self.buttons = {}
  self.selectedButton = nil
  self.textLayer = textLayer

  createLayout(self)

  self:setTouchEnabled(false)

  return self
end
