--- TransitionsOutroAnimation
-- @author Created by Steven Mah on May 9, 2017.

class("TransitionsOutroAnimation")

require("TransitionsDiamond")
require("TransitionsGlitchyLine")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local DURATION = 0.2

local Z_INDEX = {
  LINES = 10,
  DIAMONDS = 20
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getIdleDiamondAnimation(self, animationParameters, textureParameters)
  local diamond = TransitionsDiamond.new(textureParameters)
  diamond:setBlendMode(MOAIProp2D.BLEND_ADD)
  diamond:setLoc(animationParameters.endPosition.x, animationParameters.endPosition.y)
  diamond:setOpacity(0.0)
  self:insert(diamond, Z_INDEX.DIAMONDS)

  local deltaY = textureParameters.sideLength * MathUtils.getRandomFloat(0.05, 0.125)
  local fadeDuration = 0.05
  local fade = AHOpacityTo.new(diamond, fadeDuration, nil, 1.0)

  local bounceDeltaY = deltaY * 12.0
  local minimumBounceDeltaY = deltaY * 0.35
  local theta = MathUtils.getRandomFloat(0.0, 0.125 * math.pi)
  local deltaTheta = 0.1
  local minimumDeltaTheta = 0.045
  local defaultY = diamond:getY() + deltaY

  local updateBounce = function()
    local offset = bounceDeltaY * math.sin(theta)
    diamond:setY(defaultY + offset)

    bounceDeltaY = math.max(minimumBounceDeltaY, bounceDeltaY * 0.98)
    deltaTheta = math.max(minimumDeltaTheta, deltaTheta * 0.98)

    theta = theta + deltaTheta
    if (theta > 2.0 * math.pi) then
      theta = 0.0
    end
  end
  local bounce = AHLoop.new(updateBounce)

  local flickerDuration = 0.0125
  local minimumOpacity = MathUtils.getRandomFloat(0.7, 0.9)
  local flickerDown = AHOpacityTo.new(diamond, flickerDuration, nil, minimumOpacity)
  local flickerUp = AHOpacityTo.new(diamond, flickerDuration, nil, 1.0)
  local flickerOnce = AHSequence.new(flickerDown, flickerUp)
  local flickerCount = math.random(2, 5)
  local flicker = AHLoop.new(flickerOnce, flickerCount)
  local delay = MathUtils.getRandomFloat(3.0, 7.0)
  local wait = AHDelay.new(delay)
  local flickerSequence = AHSequence.new(wait, flicker)
  local flickerLoop = AHLoop.new(flickerSequence)

  local showAndBounce = AHSpawn.new(fade, bounce, flickerLoop)
  return showAndBounce
end

local function getPopAnimation(diamond)
  local popScaleDuration = 0.2
  local popScale = AHScaleTo.new(diamond, popScaleDuration, nil, 40.0)
  local popFade = AHOpacityTo.new(diamond, popScaleDuration * 0.5, nil, 0.0)
  local pop = AHSpawn.new(popScale, popFade)
  return pop
end

local function getShakeAnimation(diamond, sideLength)
  local shakeCount = 4
  local shakeAmplitude = sideLength * MathUtils.getRandomFloat(0.05, 0.1)
  local shakeDuration = 0.05
  local scale = diamond:getScl()
  local shake = AHSequence.new()

  for i = 1, shakeCount do
    local shakeUp = AHLocationYBy.new(diamond, shakeDuration, nil, shakeAmplitude)
    local shakeDown = AHLocationYBy.new(diamond, shakeDuration, nil, -shakeAmplitude)
    local shakeOnce = AHSequence.new(shakeUp, shakeDown)
    local scaleDown = AHScaleTo.new(diamond, 2.0 * shakeDuration, nil, scale)
    local shakeAndScale = AHSpawn.new(shakeOnce, scaleDown)
    shake:addAction(shakeAndScale)

    shakeDuration = shakeDuration * 0.6
    shakeAmplitude = shakeAmplitude * 1.025
    scale = scale * 1.0075
  end

  return shake
end

local function getshowLinesAnimation(self, animationParameters, sideLength)
  local showLinesAnimation = AHSpawn.new()

  for _, parameters in ipairs(animationParameters.lineParameters) do
    local lines = TransitionsGlitchyLine.new(sideLength, parameters.turnDirection)
    lines:setBlendMode(MOAIProp2D.BLEND_ADD)
    lines:setLoc(parameters.position.x, parameters.position.y)
    lines:setRot(parameters.angle)
    self:insert(lines, Z_INDEX.LINES)

    local showLines = function()
      lines:startAnimation()
    end
    showLinesAnimation:addAction(showLines)
  end

  return showLinesAnimation
end

local function getMoveAnimation(diamond, duration, animationParameters)
  local moveEase = AHEaseElasticOut.new(1.0)
  local move = AHLocationTo.new(diamond, duration * 4.0, moveEase, animationParameters.endPosition.x, animationParameters.endPosition.y)
  local fade = AHOpacityTo.new(diamond, duration, nil, 1.0)
  local moveAnimation = AHSpawn.new(move, fade)
  return moveAnimation
end

local function getDiamondsAnimation(self, sideLength, diamondParameters, diamondParametersIdle)
  local textureParameters = {
    rowCount = 64,
    columnCount = 64,
    lineColor = {1.0, 1.0, 1.0, 1.0},
    glowColor = {0.5, 0.5, 0.5, 1.0},
    fillColor = {0.4, 0.4, 0.4, 1.0},
    glowMarginFraction = 0.2,
    lineWidthFraction = 0.005,
    sideLength = sideLength,
    showGlow = false,
    fillCenter = false
  }

  local duration = DURATION
  local shakeDelay = 3.5 * duration + MathUtils.getRandomFloat(0.0, 0.05)
  local animation = AHSpawn.new()

  for i = 1, #diamondParameters do
    local animationParameters = diamondParameters[i]

    local diamond = TransitionsDiamond.new(textureParameters)
    diamond:setLoc(animationParameters.startPosition.x, animationParameters.startPosition.y)
    diamond:setOpacity(0.0)
    diamond:setBlendMode(MOAIProp2D.BLEND_ADD)
    self:insert(diamond, Z_INDEX.DIAMONDS)

    local move = getMoveAnimation(diamond, duration, animationParameters)
    local shake = getShakeAnimation(diamond, sideLength)
    local lines = getshowLinesAnimation(self, animationParameters, sideLength)
    local pop = getPopAnimation(diamond)

    if (diamondParametersIdle ~= nil) then
      local animationParametersIdle = diamondParametersIdle[i]
      local textureParametersIdle = TableUtils.deepCopy(textureParameters)
      textureParametersIdle.fillCenter = true
      textureParametersIdle.showGlow = true
      local idleDiamondAnimation = getIdleDiamondAnimation(self, animationParametersIdle, textureParametersIdle)
      pop:addAction(idleDiamondAnimation)
    end

    animation:addAction(move)
    pop:addAction(lines)
    shake:addAction(pop)

    local delayedShake = AHDelayed.new(shakeDelay, shake)
    animation:addAction(delayedShake)

    duration = duration * 1.25
  end

  return animation
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsOutroAnimation:getAnimation()
  return self.animation
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local sideLength = self:getWidthPart(0.25)
  local spacingFraction = 0.425
  local defaultStartY = self:getHeightPart(1.2)

  -- hollow moving diamonds
  local leftMostEndPosition = {
    x = self:getWidthPart(0.34),
    y = self:getHeightPart(0.55)
  }

  local diamondParametersTop = {
    {
      startPosition = {
        x = leftMostEndPosition.x,
        y = defaultStartY
      },
      endPosition = {
        x = leftMostEndPosition.x,
        y = leftMostEndPosition.y
      }
    },
    {
      startPosition = {
        x = leftMostEndPosition.x + sideLength * spacingFraction,
        y = defaultStartY
      },
      endPosition = {
        x = leftMostEndPosition.x + sideLength * spacingFraction,
        y = leftMostEndPosition.y - sideLength * spacingFraction
      }
    },
    {
      startPosition = {
        x = leftMostEndPosition.x + sideLength * 2.0 * spacingFraction,
        y = defaultStartY
      },
      endPosition = {
        x = leftMostEndPosition.x + sideLength * 2.0 * spacingFraction,
        y = leftMostEndPosition.y
      }
    },
    {
      startPosition = {
        x = leftMostEndPosition.x + sideLength * 3.0 * spacingFraction,
        y = defaultStartY
      },
      endPosition = {
        x = leftMostEndPosition.x + sideLength * 3.0 * spacingFraction,
        y = leftMostEndPosition.y + sideLength * spacingFraction
      }
    }
  }

  local diamondParametersBottom = TableUtils.deepCopy(diamondParametersTop)
  for _, parameters in ipairs(diamondParametersBottom) do
    parameters.startPosition.y = 0.0
    parameters.lineParameters = {}
  end

  -- filled idle diamonds
  local diamondParametersIdle = {}
  for _, parameters in ipairs(diamondParametersBottom) do
    local idleParameters = {
      endPosition = TableUtils.deepCopy(parameters.endPosition)
    }
    table.insert(diamondParametersIdle, idleParameters)
  end

  local topLineParameters = {
    {
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = -sideLength * spacingFraction,
          y = sideLength * spacingFraction * 1.2
        },
        angle = 180.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.RIGHT,
        position = {
          x = sideLength * spacingFraction,
          y = sideLength * spacingFraction * 0.5
        },
        angle = 180.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = -sideLength * spacingFraction,
          y = sideLength * spacingFraction * 1.2
        },
        angle = -90.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.RIGHT,
        position = {
          x = sideLength * spacingFraction,
          y = sideLength * spacingFraction * 0.5
        },
        angle = 90.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.RIGHT,
        position = {
          x = -sideLength * spacingFraction * 1.65,
          y = sideLength * spacingFraction * 0.075
        },
        angle = -90.0
      }
    },
    {

    },
    {

    },
    {
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = sideLength * spacingFraction,
          y = -sideLength * spacingFraction * 1.2
        },
        angle = 0.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.RIGHT,
        position = {
          x = -sideLength * spacingFraction,
          y = -sideLength * spacingFraction * 0.65
        },
        angle = 0.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = -sideLength * spacingFraction,
          y = -sideLength * spacingFraction * 0.65
        },
        angle = -90.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = sideLength * spacingFraction,
          y = -sideLength * spacingFraction * 1.2
        },
        angle = 90.0
      },
      {
        turnDirection = TransitionsGlitchyLine.DIRECTIONS.LEFT,
        position = {
          x = sideLength * spacingFraction * 1.65,
          y = -sideLength * spacingFraction * 0.3
        },
        angle = 90.0
      }
    }
  }

  for i = 1, #diamondParametersTop do
    local diamondParameters = diamondParametersTop[i]
    local lineParameters = topLineParameters[i]
    for _, parameters in ipairs(lineParameters) do
      parameters.position.x = parameters.position.x + diamondParameters.endPosition.x
      parameters.position.y = parameters.position.y + diamondParameters.endPosition.y
    end
    diamondParameters.lineParameters = lineParameters
  end

  local animationDiamondsTop = getDiamondsAnimation(self, sideLength, diamondParametersTop, diamondParametersIdle)
  local animationDiamondsBottom = getDiamondsAnimation(self, sideLength, diamondParametersBottom)
  self.animation = AHSpawn.new(animationDiamondsTop, animationDiamondsBottom)
end

-- Constructor ------------------------

function TransitionsOutroAnimation.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, TransitionsOutroAnimation)

  createLayout(self)

  return self
end
