--- TransitionsPromptTextView
-- @author Created by Steven Mah on October 20, 2016.

class("TransitionsPromptTextView")

require("TransitionsTextAnimationBuilder")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ITEM_TYPE_BLANK = "blank"
local TEXT_COLOR = {1.0, 1.0, 0.95}
local INCORRECT_REPLACEMENT_COLOR = TransitionsCommon.CYAN_HIGHLIGHT
local CORRECT_REPLACEMENT_COLOR = {0.71, 0.84, 0.45}

local WIDTH = math.min(Screen.getWidthPart(0.85), Screen.safeWidth)
local LINE_SPACING = Screen.getWidthPart(0.0225)

local SMALL_FONT_SIZE = 18
local MEDIUM_FONT_SIZE = 20
local BIG_FONT_SIZE = 22

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function switchFontSizeIfNeeded(self, text)
  local characterCount = #text
  local textInterpreter = self.textInterpreter
  if (characterCount >= 140) then
    textInterpreter.baseAttributes.font = self.mediumFont
    textInterpreter.attributesForLabels.i.font = self.mediumItalicFont
  else
    textInterpreter.baseAttributes.font = self.bigFont
    textInterpreter.attributesForLabels.i.font = self.bigItalicFont
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsPromptTextView:createItemView(item)
  if (item:getItemType() == ITEM_TYPE_BLANK) then
    local underlineWidth = Screen.getWidthPart(0.25)
    local underlineHeight = Screen.getHeightPart(0.0025)
    local color = item:getAttribute("color")
    local underline = FilledRectangle.new(underlineWidth, underlineHeight, color)
    underline:setScl(0.95, 1.0)

    return underline
  else
    return self.super.createItemView(self, item)
  end
end

function TransitionsPromptTextView:setText(promptText)
  switchFontSizeIfNeeded(self, promptText)
  local textInterpreter = self.textInterpreter
  local textModel = self.textModel
  local items = textInterpreter:interpretXmlText(promptText)
  textModel:resetItems(items)

  self.text = promptText
end

function TransitionsPromptTextView:getReplaceBlankAction(correctReplacement, isCorrect)
  local words = StringUtils.split(correctReplacement, " ")
  local attributes = TableUtils.shallowCopy(self.baseAttributes)
  local newItems = {}
  for index, word in ipairs(words) do
    if index == #words then
      attributes.skipNextSpacing = true
      attributes.allowBreakLineAfter = false
    end
    if (isCorrect) then
      attributes.color = CORRECT_REPLACEMENT_COLOR
    else
      attributes.color = INCORRECT_REPLACEMENT_COLOR
    end
    local textItem = TextItem.newTextItem(word, attributes)
    table.insert(newItems, textItem)
  end
  local items = self.textModel:getItems()
  local blankIndex = 0
  for index, item in ipairs(items) do
    if (item:getItemType() == ITEM_TYPE_BLANK) then
      blankIndex = index
      break
    end
  end

  local replaceAction = function()
    self.textModel:replaceItems(blankIndex, 1, newItems)
    wait(self.animation)
  end

  return replaceAction
 end

 function TransitionsPromptTextView:getText()
   return self.text
 end

 function TransitionsPromptTextView:getHideAndRemoveAction(duration, ease)
   local hideAndRemoveAction = AHSequence.new(
     function()
       self:linkOpacityAttrToChildren(true)
     end,
     AHSpawn.new(
       AHOpacityTo.new(self, duration, ease, 0.0)
     ),
     function()
       self:removeSelf()
     end
   )

   return hideAndRemoveAction
 end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TransitionsPromptTextView.new(text)
  local self = TextView.new(WIDTH, nil, nil, LINE_SPACING)
  Mixin.mix(self, TransitionsPromptTextView)

  self.smallFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, SMALL_FONT_SIZE)
  self.mediumFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, MEDIUM_FONT_SIZE)
  self.bigFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, BIG_FONT_SIZE)
  self.smallItalicFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM_ITALIC, SMALL_FONT_SIZE)
  self.mediumItalicFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM_ITALIC, MEDIUM_FONT_SIZE)
  self.bigItalicFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM_ITALIC, BIG_FONT_SIZE)

  self.baseAttributes = {
    font = self.bigFont,
    color = TEXT_COLOR
  }

  self.blankAttributes = {
    blank = {
      itemType = ITEM_TYPE_BLANK,
      skipNextSpacing = true,
      allowBreakLineAfter = false,
      yOffset = Screen.getHeightPart(0.0075) -- NOTE: Needed to line up underline with comma more
    },
    i = {
      font = self.bigItalicFont
    }
  }

  local textInterpreter = TextInterpreter.new(self.baseAttributes, self.blankAttributes)
  local textModel = TextModel.new(textInterpreter:interpretXmlText(""))
  local textAnimationBuilder = TransitionsTextAnimationBuilder.new()
  self:setModel(textModel)

  self.textInterpreter = textInterpreter
  self.textModel = textModel
  self.animationBuilder = textAnimationBuilder

  if (text) then
    self:setText(text)
  end

  return self
end
