--- TransitionsRationale
-- @author Created by Steven Mah on October 20, 2016.

class("TransitionsRationale")

require("text.SizeConstrainedTextArea")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BASE_COLOR = {1.0, 1.0, 1.0}
local BUTTON_COLOR = {0.75, 0.89, 0.94}
local BUTTON_MAX_OPACITY = 0.5

local LINE_SPACING = Screen.getHeightPart(0.005)
local MARGIN_Y = Screen.getHeightPart(0.03)

local MAIN_RATIONALE_STRONG_ATTRIBUTE_TAG = "strongMain"
local GENERIC_RATIONALE_STRONG_ATTRIBUTE_TAG = "strongGeneric"

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createOrReplaceTextView(self)
  assert(self.returnToGameLabel ~= nil, "returnToGameLabel must be created first to determine text view dimensions")

  if (self.textView ~= nil) then
    self.textView:removeSelf()
  end

  local fontSize = 20
  local baseFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, fontSize)
  local italicFont = PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM_ITALIC, fontSize)

  local basicAttributes = {
    font = baseFont
  }

  local customAttributes = {
    [MAIN_RATIONALE_STRONG_ATTRIBUTE_TAG] = {
      color = TransitionsCommon.CYAN_HIGHLIGHT
    },
    [GENERIC_RATIONALE_STRONG_ATTRIBUTE_TAG] = {
      color = {1.0, 0.18, 0.38}
    },
    i = {
      font = italicFont
    }
  }

  local maxHeight = self:getHeight() - self.returnToGameLabel:getYTop() - (2 * MARGIN_Y)
  local minimumFontSize = 10
  self.textView = SizeConstrainedTextArea.new("--unset--", self:getWidth(), maxHeight, basicAttributes, customAttributes, minimumFontSize,
    nil, nil, LINE_SPACING)
  self.textView:setColor(unpack(BASE_COLOR))
  self.textView:setAnchorPointTop()
  self:insertAtLocation(self.textView, self:getWidthPart(0.5), self:getHeight() - MARGIN_Y)

  self:linkOpacityAttrToChildren(true)
  self.returnToGameLabel:clearAttrLink(MOAIColor.ATTR_OPACITY)
end

local function createCombinedMainAndGenericRationaleString(self, selectedAnswer)
  local replacementOpeningTag = string.format("<%s>", MAIN_RATIONALE_STRONG_ATTRIBUTE_TAG)
  local replacementClosingTag = string.format("</%s>", MAIN_RATIONALE_STRONG_ATTRIBUTE_TAG)
  local tagReplacedMainRationaleText = self.mainRationaleText:gsub("<strong>", replacementOpeningTag)
  tagReplacedMainRationaleText = tagReplacedMainRationaleText:gsub("</strong>", replacementClosingTag)

  replacementOpeningTag = string.format("<%s>", GENERIC_RATIONALE_STRONG_ATTRIBUTE_TAG)
  replacementClosingTag = string.format("</%s>", GENERIC_RATIONALE_STRONG_ATTRIBUTE_TAG)

  local genericRationaleText = self.incorrectWordsForRationales[selectedAnswer]
  local tagReplacedGenericRationaleText = genericRationaleText:gsub("<strong>", replacementOpeningTag)
  tagReplacedGenericRationaleText = tagReplacedGenericRationaleText:gsub("</strong>", replacementClosingTag)

  return string.format("%s<br/><br/>%s", tagReplacedMainRationaleText, tagReplacedGenericRationaleText)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsRationale:setRoundData(round)
  self.mainRationaleText = round.customRationale
  self.incorrectWordsForRationales = {}
  for i = 1, round.wordCount - 1 do
    local incorrectWord = round["incorrect"..i]
    local rationaleForWord = round["genericRationale"..i]
    self.incorrectWordsForRationales[incorrectWord] = rationaleForWord
  end
end

function TransitionsRationale:setText(selectedAnswer)
  local combinedRationaleText = createCombinedMainAndGenericRationaleString(self, selectedAnswer)
  self.textView:setText(combinedRationaleText)
  self.textView:rerender()
end

function TransitionsRationale:waitForDismissal()
  self.tapToContinue:rearm()
  self.tapToContinue:waitForTrigger()
end

function TransitionsRationale:getReturnLabel()
  return self.returnToGameLabel
end

function TransitionsRationale:getFadeAction(opacity, duration, ease)
  return AHSpawn.new(
    AHOpacityTo.new(self, duration, ease, opacity),
    AHOpacityTo.new(self.returnToGameLabel, duration, ease, MathUtils.clamp(opacity, 0.0, BUTTON_MAX_OPACITY))
  )
end

function TransitionsRationale:setHeight(height)
  self.super.setHeight(self, height)

  -- Necessary since if the rationale's dimensions change, it modifies the height
  -- of available space to constrain text
  createOrReplaceTextView(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TransitionsRationale.new(height)
  local width = math.min(Screen.getWidthPart(0.85), Screen.safeWidth)
  local self = Group.new(width, height)
  Mixin.mix(self, TransitionsRationale)

  local returnToGameLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  returnToGameLabel:setColor(unpack(BUTTON_COLOR))
  returnToGameLabel:setOpacity(0.0)
  local labelY = self:getHeightPart(0.15)
  local minimumY = Screen.safeBottom + (returnToGameLabel:getHeight() * 1.5)
  self:insertAtLocation(returnToGameLabel, self:getWidthPart(0.5), math.max(labelY, minimumY))
  self.returnToGameLabel = returnToGameLabel

  createOrReplaceTextView(self)

  self.tapToContinue = TapToContinue.new()
  self.tapToContinue:disarm()

  return self
end
