--- TransitionsValidator
-- @author Created by joepoka on October 20, 2016.
-- ConceptValidator for Transitions.

class("TransitionsValidator")

require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "difficulty",
  "types",
  "wordCount",
  "prompt",
  "correct",
  "incorrect1",
  "customRationale",
  "genericRationale1"
}

local SEMI_OPTIONAL_KEYS = {
  "incorrect2",
  "genericRationale2"
}

local PROMPT_SINGLE_TAGS = {"blank"}
local RATIONALE_PAIRED_TAGS = {"strong"}
local TERTIARY_TAGS = nil
local OPTIONAL_PAIRED_TAGS = {"i"}

local DUPLICATE_WORD_MESSAGE = "Concept has duplicated word among answer choices."

local GENERIC_STRONG_FORMAT = "<strong>%s</strong> "
local CUSTOM_STRONG_FORMAT = ", so <strong>%s</strong> "
local CUSTOM_STRONG_QUOTE_FORMAT = ",\" so <strong>%s</strong> "

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isNotEmptyOrNil(string)
  return (string ~= nil) and (string ~= "")
end

local function getStrongWrappedTransition(transition, template)
  return string.format(template, transition)
end

local function quotePrecedesTransition(concept)
  local text = concept.customRationale
  local plain = true
  return (text:find(",\" so", 1, plain) ~= nil)
end

local function validateThirdOption(self, concept)
  if ((isNotEmptyOrNil(concept.incorrect2)) or (isNotEmptyOrNil(concept.genericRationale2))) then
    local formattedStronTransition = getStrongWrappedTransition(concept.incorrect2, GENERIC_STRONG_FORMAT)

    self:expectEqual(concept.wordCount, 3, "filter_part value does not match number of word options.")
    self:validateRequiredKeys(concept, SEMI_OPTIONAL_KEYS)
    self:expectNotEqual(concept.correct, concept.incorrect2, DUPLICATE_WORD_MESSAGE)
    self:expectNotEqual(concept.incorrect1, concept.incorrect2, DUPLICATE_WORD_MESSAGE)
    self:validateTags(concept.genericRationale2, nil, RATIONALE_PAIRED_TAGS, nil, OPTIONAL_PAIRED_TAGS)
    self:expectStringContains(concept.genericRationale2, formattedStronTransition, "generic_rationale2")
    self:expectNoLeadingOrTrailingWhitespace(concept.genericRationale2, "generic_rationale2")

    local capitalizedDistractor = StringUtils.capitalizeFirstLetter(concept.incorrect2)
    self:expectEqual(concept.incorrect2, capitalizedDistractor, "incorrect_replace2 is not capitalized.")
  else
    self:expectEqual(concept.wordCount, 2, "filter_part value does not match number of word options.")
    self:expectEmpty(concept.incorrect2, "incorrect_replace2")
    self:expectEmpty(concept.genericRationale2, "generic_rationale2")
  end
end

local function validateWordsInRationales(self, concept)
  local lowerCaseCorrect = string.lower(concept.correct)
  local formattedStrongCorrect
  if (quotePrecedesTransition(concept)) then
    formattedStrongCorrect = getStrongWrappedTransition(lowerCaseCorrect, CUSTOM_STRONG_QUOTE_FORMAT)
  else
    formattedStrongCorrect = getStrongWrappedTransition(lowerCaseCorrect, CUSTOM_STRONG_FORMAT)
  end
  local formattedStrongIncorrect = getStrongWrappedTransition(concept.incorrect1, GENERIC_STRONG_FORMAT)
  self:expectStringContains(concept.customRationale, formattedStrongCorrect, "custom_rationale")
  self:expectStringContains(concept.genericRationale1, formattedStrongIncorrect, "generic_rationale1")
end

local function validateNoWordRepeats(self, concept)
  self:expectNotEqual(concept.correct, concept.incorrect1, DUPLICATE_WORD_MESSAGE)
end

local function validateWhiteSpace(self, concept)
  self:expectNoLeadingOrTrailingWhitespace(concept.prompt, "prompt")
  self:expectNoLeadingOrTrailingWhitespace(concept.customRationale, "custom_rationale")
  self:expectNoLeadingOrTrailingWhitespace(concept.genericRationale1, "generic_rationale1")
end

local function validateCapitalizedOptions(self, concept)
  local capitalizedAnswer = StringUtils.capitalizeFirstLetter(concept.correct)
  local capitalizedDistractor = StringUtils.capitalizeFirstLetter(concept.incorrect1)
  self:expectEqual(concept.correct, capitalizedAnswer, "correct_replace is not capitalized.")
  self:expectEqual(concept.incorrect1, capitalizedDistractor, "incorrect_replace1 is not capitalized.")
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TransitionsValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  self:validateTags(concept.prompt, PROMPT_SINGLE_TAGS, nil, nil, OPTIONAL_PAIRED_TAGS)
  self:validateTags(concept.customRationale, nil, RATIONALE_PAIRED_TAGS, nil, OPTIONAL_PAIRED_TAGS)
  self:validateTags(concept.genericRationale1, nil, RATIONALE_PAIRED_TAGS, nil, OPTIONAL_PAIRED_TAGS)
  self:expectStringContains(concept.prompt, " <blank/>, ", "prompt")

  validateCapitalizedOptions(self, concept)
  validateWhiteSpace(self, concept)
  validateNoWordRepeats(self, concept)
  validateWordsInRationales(self, concept)
  validateThirdOption(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TransitionsValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, TransitionsValidator)

  return self
end
