--- TypoAnimationHelper
-- @author Created by anton on August 12, 2015.

class("TypoAnimationHelper")

require("TypoBubbleSwarm")
require("TypoParticles")
require("TypoTimeoutPanel")
require("TypoGameIntro")
require("TypoCheckmark")
require("TypoBubbleLayouts")
require("TypoSound")
require("TypoNoErrorLabel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND = { 1, 1, 1 }

local Z_BACKGROUND = 10
local Z_BOUNCING_GROUP = 20
local Z_GAME_INTRO = 30
local Z_ROUND_INTRO = 40
local Z_WIN_BACKGROUND = 50
local Z_BUBBLE_SWARM = 60
local Z_NO_ERROR_LABEL = 70
local Z_CHECKMARK = 80
local Z_TRACKING_BLOCKS = 90
local Z_PARTICLES = 100

-- Tracking block positions -----------

local BLOCK_POSITION_TOP = "BLOCK_POSITION_TOP"
local BLOCK_POSITION_LEFT = "BLOCK_POSITION_LEFT"
local BLOCK_POSITION_RIGHT = "BLOCK_POSITION_RIGHT"

-- Animation parameters ---------------

local SLIDE_DURATION = 1.3

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Slide helpers ----------------------

local function prepareToSlideDown(group)
  local y = group:getY()
  if y <= Screen.height then
    local hiddenY = group:getY() + group:getHeight()
    group:setY(hiddenY)
  end
end

local function getSlideDownAnimation(group)
  local distanceY = -group:getHeight()
  return AHLocationYBy.new(group, SLIDE_DURATION, AHEaseQuadraticOut.new(), distanceY)
end

local function getSlideUpAnimation(group)
  local distanceY = group:getHeight()
  return AHLocationYBy.new(group, SLIDE_DURATION, AHEaseQuadraticIn.new(), distanceY)
end

-- Fake background pieces -------------

local function updateTrackingBlockPosition(block, trackedObject, blockPosition)
  if blockPosition == BLOCK_POSITION_LEFT then
    block:setX(trackedObject:getXLeft())
  elseif blockPosition == BLOCK_POSITION_RIGHT then
    block:setX(trackedObject:getXRight())
  elseif blockPosition == BLOCK_POSITION_TOP then
    block:setY(trackedObject:getYTop())
  end
end

local function startTracking(block, trackedObject, blockPosition)
  local scriptNode = MOAIScriptNode.new()
  scriptNode:reserveAttrs(1)
  scriptNode:setCallback(function(node)
    updateTrackingBlockPosition(block, trackedObject, blockPosition)
  end)

  local attributeId = 0
  scriptNode:setAttrLink(attributeId, trackedObject, MOAITransform2D.ATTR_X_SCL)
  block.scriptNode = scriptNode
end

local function createTrackingBlock(trackedObject, blockPosition)
  assert(TableUtils.contains({BLOCK_POSITION_LEFT, BLOCK_POSITION_RIGHT, BLOCK_POSITION_TOP}, blockPosition),
         string.format("Unexpected block position: %s", tostring(blockPosition)))

  -- We don't pass any color here because it will be set as needed.
  local block = FilledRectangle.newFullScreen()

  if blockPosition == BLOCK_POSITION_TOP then
    block:setX(trackedObject:getWidthPart(0.5))
  else
    block:setY(trackedObject:getHeightPart(0.5))
  end

  updateTrackingBlockPosition(block, trackedObject, blockPosition)
  startTracking(block, trackedObject, blockPosition)

  return block
end

local function createTopTrackingBlock(trackedObject)
  local block = createTrackingBlock(trackedObject, BLOCK_POSITION_TOP)
  block:setAnchorPointBottom()
  block:setX(trackedObject:getWidthPart(0.5))
  return block
end

local function createLeftTrackingBlock(trackedObject)
  local block = createTrackingBlock(trackedObject, BLOCK_POSITION_LEFT)
  block:setAnchorPointRight()
  block:setY(trackedObject:getHeightPart(0.5))
  return block
end

local function createRightTrackingBlock(trackedObject)
  local block = createTrackingBlock(trackedObject, BLOCK_POSITION_RIGHT)
  block:setAnchorPointLeft()
  block:setY(trackedObject:getHeightPart(0.5))
  return block
end

local function createBounceAnimation(self, tapX, tapY)
  local bouncingGroup = self.bouncingGroup
  local anchorX = tapX / bouncingGroup:getWidth()
  local anchorY = tapY / bouncingGroup:getHeight()
  bouncingGroup:transformAnchorPoint(anchorX, anchorY)

  return AHSequence.new(
    AHScaleTo.new(bouncingGroup, 0.17, AHEaseSineInOut.new(), 0.90),
    AHScaleTo.new(bouncingGroup, 0.15, AHEaseSineInOut.new(), 1.03),
    AHScaleTo.new(bouncingGroup, 0.17, AHEaseSineInOut.new(), 0.98),
    AHScaleTo.new(bouncingGroup, 0.28, AHEaseSineInOut.new(), 1)
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Transitions ------------------------

function TypoAnimationHelper:playRoundIntroTransition(round, roundIntro)
  local bubbleSwarm = self.bubbleSwarm

  if not self.firstTransitionDone then
    local gameIntro = self.gameIntro -- Will be nil if GameSession.hasNeverWon() is false

    bubbleSwarm:setBubbleLayouts(round:getBubbleEnterLayouts(), round:getColor())
    bubbleSwarm:attachToBottomOffscreen(roundIntro, true)
    roundIntro:prepareToShow()

    if gameIntro then
      local showGameIntro = gameIntro:getShowAnimation()
      wait(showGameIntro:start())
      gameIntro:waitForTap()
    end

    self:insertAtCenter(roundIntro, Z_ROUND_INTRO)

    local swirlUpSwarm = bubbleSwarm:getSwirlUpAnimation(round:getColor())
    local showRoundIntro = roundIntro:getShowAnimation()

    local hideGameIntro = AHDoNothing.new()
    if gameIntro then
      hideGameIntro = gameIntro:getHideAnimation()
    end

    local animation = AHSpawn.new(
      swirlUpSwarm,
      AHDelayed.new(0.1, hideGameIntro),
      AHDelayed.new(0.1, showRoundIntro)
    )

    TypoSound.playBubbleWoosh()
    wait(animation:start())
    self.firstTransitionDone = true
  else
    local textPanel = self.textPanel
    local header = self.header
    local scrollDownTextPanel = textPanel:newScrollToBottomAction(SLIDE_DURATION)
    local slideUpTextPanel = getSlideUpAnimation(textPanel)
    local slideUpHeader = getSlideUpAnimation(header)
    local showRoundIntro = roundIntro:getShowAnimation()

    local swirlUpSwarm = nil
    if round:isBonusRound() then
      swirlUpSwarm = bubbleSwarm:getSwirlUpBonusAnimation()
    else
      swirlUpSwarm = bubbleSwarm:getSwirlUpAnimation(round:getColor())
    end

    self:insertAtCenter(roundIntro, Z_ROUND_INTRO)
    roundIntro:prepareToShow()
    bubbleSwarm:attachToTop(roundIntro)

    local animation = AHSpawn.new(
      swirlUpSwarm,
      AHSpawn.new(slideUpTextPanel, scrollDownTextPanel),
      slideUpHeader,
      AHDelayed.new(0.4, showRoundIntro)
    )

    TypoSound.playBubbleWoosh()
    wait(animation:start())
  end
end

function TypoAnimationHelper:playRoundStartTransition(round, roundIntro)
  local textPanel = self.textPanel
  local header = self.header
  local bubbleSwarm = self.bubbleSwarm
  local hideRoundIntro = roundIntro:getHideAnimation()
  local slideDownHeader = getSlideDownAnimation(header)
  local slideDownTextPanel = getSlideDownAnimation(textPanel)
  local scrollUpTextPanel = textPanel:newScrollToTopAction(1)

  local swirlDownSwarm = nil
  if round:isBonusRound() then
    swirlDownSwarm = bubbleSwarm:getSwirlDownBonusAnimation()
  else
    swirlDownSwarm = bubbleSwarm:getSwirlDownAnimation()
  end

  prepareToSlideDown(header)
  prepareToSlideDown(textPanel)
  textPanel:scrollToBottom()

  local animation = AHSpawn.new(
    swirlDownSwarm,
    hideRoundIntro,
    AHDelayed.new(0.4, slideDownHeader),
    AHDelayed.new(0.4, AHSpawn.new(slideDownTextPanel, scrollUpTextPanel))
  )

  TypoSound.playBubbleWoosh()
  wait(animation:start())


  if round:isBonusRound() then
    self.bouncingGroup:takeChild(bubbleSwarm)
  else
    bubbleSwarm:setBubbleLayouts(round:getBubbleExitLayouts(), round:getColor())
    bubbleSwarm:attachToBottom(textPanel:getBottomArea())
  end

  roundIntro:removeSelf()
end

-- Answers ----------------------------

function TypoAnimationHelper:playCorrectAnswerAnimation(round, errorId, tapX, tapY)
  local textController = self.textController
  local textView = self.textView
  local header = self.header
  local particles = self.particles
  local trackingBlocks = self.trackingBlocks
  local roundIsBonus = round:isBonusRound()

  local bounceAnimation = createBounceAnimation(self, tapX, tapY)

  for _, block in ipairs(trackingBlocks) do
    if roundIsBonus then
      TypoCommon.startBonusColorAnimation(block)
    else
      block:setColor(unpack(round:getColor()))
    end
  end

  particles:setLoc(tapX, tapY)
  if roundIsBonus then
    particles:setIsBonus(true)
  else
    particles:setMainColor(round:getColor())
  end


  local correctAnimation = AHSpawn.new(
    bounceAnimation,
    function()
      textController:correctError(errorId)
      particles:show()
      TypoSound.playCorrectAnswer()
      header:incrementErrorsFound()
      textView:getCorrectAnswerColorAnimation(errorId, roundIsBonus):start()
      wait(textView:getAnimation())
    end
  )

  wait(correctAnimation:start())
end

function TypoAnimationHelper:playIncorrectAnswerAnimation(round, tappedItem)
  local textView = self.textView
  local header = self.header
  local bubbleSwarm = self.bubbleSwarm
  local noErrorLabel = self.noErrorLabel
  local tappedItemView = textView.viewsForItems[tappedItem]

  local colorChangeDuration = 0.2
  local headerAnimation = header:getIncorrectAnswerAnimation(colorChangeDuration)
  local textViewAnimation = textView:getIncorrectAnswerAnimation(tappedItem, colorChangeDuration)
  local noErrorLabelAnimation = noErrorLabel:getShowAnimation(tappedItemView, colorChangeDuration)
  local dimBubbleSwarm = bubbleSwarm:getDimDownAnimation(colorChangeDuration)
  local preAnimation = AHSpawn.new(headerAnimation, textViewAnimation, noErrorLabelAnimation, dimBubbleSwarm)

  self.isShowingIncorrectAnimation = true

  TypoSound.playIncorrectAnswer()
  wait(preAnimation:start())
  waitFor(0.5)

  local revertHeaderAnimation = header:getRevertFromIncorrectAnswerAnimation(colorChangeDuration)
  local revertTextViewAnimation = textView:getRevertFromIncorrectAnswerAnimation(colorChangeDuration)
  local revertNoErrorLabelAnimation = noErrorLabel:getHideAnimation(colorChangeDuration)
  local reverttDimBubbleSwarm = bubbleSwarm:getRevertDimDownAnimation(colorChangeDuration)
  local revertAnimation = AHSpawn.new(revertHeaderAnimation, revertTextViewAnimation, revertNoErrorLabelAnimation, reverttDimBubbleSwarm)

  wait(revertAnimation:start())

  self.isShowingIncorrectAnimation = false
end

function TypoAnimationHelper:playFixedAnswerAnimation(round, errorId, tapX, tapY)
  local textController = self.textController
  local textView = self.textView
  local header = self.header
  local forceReplace = true

  local bounceAnimation = createBounceAnimation(self, tapX, tapY)

  local correctAnimation = AHSpawn.new(
    bounceAnimation,
    function()
      textController:correctError(errorId, forceReplace)
      TypoSound.playCorrectAnswer()
      header:incrementErrorsFound()
      textView:getCorrectAnswerColorAnimation(errorId, roundIsBonus):start()
      wait(textView:getAnimation())
    end
  )

  wait(correctAnimation:start())
end

-- Timeout ----------------------------

function TypoAnimationHelper:playTimeoutAnimation(round, errorsRemainingCount)
  local textView = self.textView
  local header = self.header
  local bubbleSwarm = self.bubbleSwarm

  waitWhile(function() return self.isShowingIncorrectAnimation end)

  local colorChangeDuration = 0.2
  local headerAnimation = header:getTimeoutAnimation(colorChangeDuration)
  local textViewAnimation = textView:getTimeoutAnimation(colorChangeDuration)
  local dimBubbleSwarm = bubbleSwarm:getDimDownAnimation(colorChangeDuration)
  local preAnimation = AHSpawn.new(headerAnimation, textViewAnimation, dimBubbleSwarm)
  wait(preAnimation:start())

  local timeoutPanel = TypoTimeoutPanel.new(errorsRemainingCount)
  self:insertAtCenter(timeoutPanel)
  self.timeoutPanel = timeoutPanel
  timeoutPanel:bringToFront()
  timeoutPanel:waitForTap(colorChangeDuration)
  timeoutPanel:removeSelf()

  local revertHeaderAnimation = header:getRevertTimeoutAnimation(colorChangeDuration)
  local revertTextViewAnimation = textView:getRevertTimeoutAnimation(colorChangeDuration)
  local reverttDimBubbleSwarm = bubbleSwarm:getRevertDimDownAnimation(colorChangeDuration)
  local revertAnimation = AHSpawn.new(revertHeaderAnimation, revertTextViewAnimation, reverttDimBubbleSwarm)

  wait(revertAnimation:start())
end

function TypoAnimationHelper:allErrorsCorrectedAfterTimeout()
  if (self.timeoutPanel ~= nil) then
    self.timeoutPanel:allErrorsCorrected()
  end
end

-- Win --------------------------------

function TypoAnimationHelper:playWinAnimation()
  local textPanel = self.textPanel
  local header = self.header
  local uiOverlay = self.uiOverlay
  local checkmark = self.checkmark
  local bubbleSwarm = self.bubbleSwarm
  local winBackground = self.winBackground

  local scrollDownTextPanel = textPanel:newScrollToBottomAction(1)
  local slideUpTextPanel = getSlideUpAnimation(textPanel)
  local slideUpHeader = getSlideUpAnimation(header)
  local slideUpUi = getSlideUpAnimation(uiOverlay)
  local showCheckmark = checkmark:getShowAnimation(1.5)
  local showWinBackground = AHOpacityTo.new(winBackground, 0.5, AHEaseQuadraticInOut.new(), 1)

  bubbleSwarm:addBubbleLayouts(TypoBubbleLayouts.LAYOUTS_WIN)
  local swirlUpSwarm = bubbleSwarm:getSwirlUpWinAnimation()

  self:takeChild(bubbleSwarm, Z_BUBBLE_SWARM)

  local showAnimation = AHSpawn.new(
    swirlUpSwarm,
    AHDelayed.new(0.2, AHSpawn.new(slideUpTextPanel, scrollDownTextPanel)),
    AHDelayed.new(0.4, showWinBackground),
    AHDelayed.new(0.4, AHSpawn.new(slideUpHeader, slideUpUi)),
    AHDelayed.new(0.8, showCheckmark)
  )

  TypoSound.playGameWin()
  wait(showAnimation:start())
  waitFor(1)

  local hideSwarm = bubbleSwarm:getHideAnimation()
  local hideCheckmark = checkmark:getHideAnimation(0.5)

  local hideAnimation = AHSpawn.new(
    hideSwarm,
    hideCheckmark
  )

  wait(hideAnimation:start())
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createElements(self, textPanel, header, uiOverlay)
  local background = FilledRectangle.newFullScreen(BACKGROUND)
  self:insertAtCenter(background, Z_BACKGROUND)

  local bouncingGroup = Group.newFullScreen()
  self:insertAtCenter(bouncingGroup, Z_BOUNCING_GROUP)
  bouncingGroup:insert(textPanel)
  bouncingGroup:insert(header)
  bouncingGroup:takeChild(uiOverlay)
  self.bouncingGroup = bouncingGroup

  local bubbleSwarm = TypoBubbleSwarm.new()
  self:insertAtCenter(bubbleSwarm, Z_BUBBLE_SWARM)
  self.bubbleSwarm = bubbleSwarm

  local topTrackingBlock = createTopTrackingBlock(bouncingGroup)
  local leftTrackingBlock = createLeftTrackingBlock(bouncingGroup)
  local rightTrackingBlock = createRightTrackingBlock(bouncingGroup)
  self:insert(topTrackingBlock, Z_TRACKING_BLOCKS)
  self:insert(leftTrackingBlock, Z_TRACKING_BLOCKS)
  self:insert(rightTrackingBlock, Z_TRACKING_BLOCKS)
  self.trackingBlocks = { topTrackingBlock, leftTrackingBlock, rightTrackingBlock }

  local particles = TypoParticles.new()
  self:insert(particles, Z_PARTICLES)
  self.particles = particles

  local shouldShowGameIntro = GameSession.hasNeverWon()
  if shouldShowGameIntro then
    gameIntro = TypoGameIntro.new()
    gameIntro:prepareToShow()
    self:insertAtCenter(gameIntro, Z_GAME_INTRO)
    self.gameIntro = gameIntro
  end

  local checkmark = TypoCheckmark.new()
  self:insertAtCenter(checkmark, Z_CHECKMARK)
  checkmark:prepareToShow()
  self.checkmark = checkmark

  local noErrorLabel = TypoNoErrorLabel.new()
  self:insert(noErrorLabel, Z_NO_ERROR_LABEL)
  self.noErrorLabel = noErrorLabel

  local winBackground = FilledRectangle.newFullScreen(WIN_BACKGROUND_COLOR)
  winBackground:setOpacity(0)
  self:insertAtCenter(winBackground, Z_WIN_BACKGROUND)
  self.winBackground = winBackground
end

-- Constructor ------------------------

function TypoAnimationHelper.new(textController, textPanel, header, uiOverlay)
  local self = Group.newFullScreen()
  Mixin.mix(self, TypoAnimationHelper)

  createElements(self, textPanel, header, uiOverlay)

  self.textPanel = textPanel
  self.header = header
  self.uiOverlay = uiOverlay
  self.textView = textPanel:getTextView()
  self.textController = textController
  self.firstTransitionDone = false
  self.isShowingIncorrectAnimation = false

  return self
end

