--- TypoBubble
-- @author Created by anton on August 10, 2015.

class("TypoBubble")

require("actions.advanced.PathBezier")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WOBBLE_MAX_RADIUS = 3
local WOBBLE_MAX_SPEED = 3
-- These provide a smooth ramp-up to the wobble effect
local WOBBLE_INITIAL_SPEED = 0.5
local WOBBLE_SPEED_STEP = 0.5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function startWobbleAnimation(self)
  local originX, originY = self:getLoc()
  local speed = WOBBLE_INITIAL_SPEED

  local wobbleAnimation = AHLoop.new(function()
    local distanceFromCenterToTarget = MathUtils.getRandomFloat(0, WOBBLE_MAX_RADIUS)
    local direction = MathUtils.getRandomFloat(0, 2 * math.pi)
    local targetX = originX + distanceFromCenterToTarget * math.cos(direction)
    local targetY = originY + distanceFromCenterToTarget * math.sin(direction)

    local distanceFromCurrentLocationToTarget = MathUtils.distance(targetX, targetY, self:getLoc())
    speed = math.min(speed + WOBBLE_SPEED_STEP, WOBBLE_MAX_SPEED)
    local duration = distanceFromCurrentLocationToTarget / speed

    local wobbleIteration = AHLocationTo.new(self, duration, AHEaseSineInOut.new(), targetX, targetY)
    wait(wobbleIteration:start())
  end)

  self.wobbleAnimation = wobbleAnimation
  wobbleAnimation:start()
end

local function stopWobbleAnimation(self)
  if self.wobbleAnimation then
    self.wobbleAnimation:stop()
    self.wobbleAnimation = nil
  end
end

local function wrapActionInWobblingPause(self, action)
  return AHSequence.new(
    function() stopWobbleAnimation(self) end,
    action,
    function() startWobbleAnimation(self) end
  )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoBubble:getSinkAnimation(endPoint, duration)
  local fallAnimation = AHLocationTo.new(self, duration, AHEaseQuadraticIn.new(), endPoint.x, endPoint.y)

  return wrapActionInWobblingPause(self, fallAnimation)
end

function TypoBubble:getSinkBonusAnimation(endPoint, duration)
  local fallAnimation = AHLocationTo.new(self, duration, AHEaseQuadraticInOut.new(), endPoint.x, endPoint.y)

  return wrapActionInWobblingPause(self, fallAnimation)
end

function TypoBubble:getFloatAnimation(endPoint, duration, color)
  local floatAnimation = AHSpawn.new(
    AHLocationTo.new(self, duration, AHEaseQuadraticInOut.new(), endPoint.x, endPoint.y)
  )

  if color then
    floatAnimation:addAction(AHColorTo.new(self, duration, AHEaseQuadraticInOut.new(), color))
  end

  return wrapActionInWobblingPause(self, floatAnimation)
end

function TypoBubble:getDimDownAnimation(dimmedDownColor, colorChangeDuration)
  self.originalColor = table.pack(self:getColor())

  local colorAnimation = AHColorTo.new(self, colorChangeDuration, nil, dimmedDownColor)
  local stopWobbling = function() stopWobbleAnimation(self) end

  return AHSpawn.new(colorAnimation, stopWobbling)
end

function TypoBubble:getRevertDimDownAnimation(colorChangeDuration)
  local originalColor = self.originalColor

  if not originalColor then
    return AHDoNothing.new()
  end

  local colorAnimation = AHSpawn.new(
    AHColorTo.new(self, colorChangeDuration, nil, originalColor),
    function()
      self.originalColor = nil
    end
  )

  local resumeWobbling = function() startWobbleAnimation(self) end

  return AHSpawn.new(colorAnimation, resumeWobbling)
end

function TypoBubble:setLoc(x, y)
  stopWobbleAnimation(self)
  self.super.setLoc(self, x, y)
  startWobbleAnimation(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TypoBubble.new(radius, color)
  local self = FilledCircle.new(radius, color)
  Mixin.mix(self, TypoBubble)

  self.wobbleAnimation = nil

  return self
end
