--- TypoBubbleSwarm
-- @author Created by anton on August 12, 2015.

class("TypoBubbleSwarm")

require("TypoBubble")
require("TypoBubbleLayouts")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local MAX_Z_INDEX = 10000
local MAXIMUM_BUBBLES_COUNT = 110

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Point math -------------------------

local function makePoint(x, y)
  return { x = x, y = y }
end

local function getOffscreenEndPoint(startPoint)
  local x = startPoint.x
  local offscreenY = startPoint.y - Screen.getHeightPart(1.5)
  return makePoint(x, offscreenY)
end

local function getFlippedEndPoint(startPoint)
  local x = startPoint.x
  local flippedY = Screen.height - startPoint.y
  -- Offset a little bit. We don't want the swarm to take as much space when placed at the bottom.
  local offsetY = flippedY - Screen.getHeightPart(0.12)
  return makePoint(x, offsetY)
end

-- Bubble activation ------------------

local function deactivateBubbles(self)
  local activeBubbles = self.activeBubbles
  local inactiveBubbles = self.inactiveBubbles

  while #activeBubbles > 0 do
    local bubble = table.remove(activeBubbles)
    bubble:setVisibleRecursive(false)
    table.insert(inactiveBubbles, bubble)
  end
end

local function activateBubbles(self, bubbleLayouts, getColorFunction)
  local activeBubbles = self.activeBubbles
  local inactiveBubbles = self.inactiveBubbles

  for index, layout in ipairs(bubbleLayouts) do
    local bubble = table.remove(inactiveBubbles)
    local startPoint = layout.startPoint
    local color = getColorFunction(layout)

    bubble:setDims(2 * layout.radius, 2 * layout.radius)
    bubble:setColor(unpack(color))
    bubble:setLoc(startPoint.x, startPoint.y)
    bubble:setVisibleRecursive(true)
    table.insert(activeBubbles, bubble)

    local zIndex = MAX_Z_INDEX - layout.radius * 100
    bubble:setZIndex(zIndex)
  end
end

-- Animations -------------------------

local function createSwirlDownAnimation(self)
  local animation = AHSpawn.new()

  local minY = math.min(unpack(TableUtils.map(self.bubbleLayouts, function(layout)
    return layout.startPoint.y
  end)))

  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local startPoint = layout.startPoint
    local endPoint = getOffscreenEndPoint(startPoint)

    local duration = 0.4 + (startPoint.y - minY)/self:getHeight() * 3
    if duration >= 3.2 then
      duration = duration * 1.5
    end

    animation:addAction(bubble:getSinkAnimation(endPoint, duration))
  end

  return animation
end

local function createSwirlDownBonusAnimation(self)
  local animation = AHSpawn.new()

  local maxX = math.max(unpack(TableUtils.map(self.bubbleLayouts, function(layout)
    return layout.startPoint.x
  end)))

  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local startPoint = layout.startPoint
    local endPoint = getFlippedEndPoint(startPoint)

    local duration = (maxX - startPoint.x)/self:getWidth() * 1.3 + math.random() * 0.3

    animation:addAction(bubble:getSinkBonusAnimation(endPoint, duration))
  end

  return animation
end

local function createSwirlUpAnimation(self, getColorFunction)
  local animation = AHSpawn.new()

  local maxY = math.max(unpack(TableUtils.map(self.bubbleLayouts, function(layout)
    return layout.startPoint.y
  end)))

  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local startPoint = layout.startPoint
    local color = getColorFunction(layout)

    local duration = 0.4 + (maxY - startPoint.y)/self:getHeight() * 3
    if duration >= 3.2 then
      duration = duration * 1.5
    end

    animation:addAction(bubble:getFloatAnimation(startPoint, duration, color))
  end

  return animation
end

local function getSwirlUpAnimationNoFlip(self, offsetY)
  local animation = AHSpawn.new()

  local maxY = math.max(unpack(TableUtils.map(self.bubbleLayouts, function(layout)
    return layout.startPoint.y
  end)))

  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local bubbleX, bubbleY = bubble:getLoc()
    local newY = bubbleY + offsetY
    local endPoint = makePoint(bubbleX, newY)

    local duration = 0.4 + (maxY - endPoint.y)/self:getHeight() * 2 + math.random() * 0.3

    animation:addAction(bubble:getFloatAnimation(endPoint, duration))
  end

  return animation
end

local function getSwirlUpAfterWinAnimation(self, offsetY)
  local animation = AHSpawn.new()

  -- Note: Other animations use layout info to determine maxY, which is safer. Unfortunately, when
  -- this animation is created we have combined two different layout sets (LAYOUTS_BONUS and
  -- LAYOUTS_WIN). Part of the bubbles are flipped when this is called, so we must use the current
  -- location.
  local maxY = math.max(unpack(TableUtils.map(self.activeBubbles, function(bubble)
    return bubble:getY()
  end)))

  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local bubbleX, bubbleY = bubble:getLoc()
    local newY = bubbleY + offsetY
    local endPoint = makePoint(bubbleX, newY)

    local duration = 0.4 + (maxY - bubble:getY())/self:getHeight() * 0.7 + math.random() * 0.5

    animation:addAction(bubble:getFloatAnimation(endPoint, duration))
  end

  return animation
end

-- getBubbleColorFunctions ------------

local function getBubbleColorFunctionForRound(roundColor)
  return function(layout)
    return TypoBubbleLayouts.getColor(layout.colorKey, roundColor)
  end
end

local function getBubbleColorFunctionForBonusRound()
  return function(layout)
    return TypoBubbleLayouts.getColorBonus(layout)
  end
end

-- Misc helpers -----------------------

local function getDistanceToSwirlForWinAnimation(self)
  -- This depends on the actual bubble layouts looking okay when flipped and moved around the
  -- screen. In particular we are using the bonus layout, flipping it and moving it down to the
  -- bottom of the screen, then moving it up to the top without un-flipping. Since we don't
  -- un-flip, the distance for this translation is less than the whole screen height.
  return Screen.getHeightPart(0.9)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Set layouts ------------------------

function TypoBubbleSwarm:setBubbleLayouts(bubbleLayouts, roundColor)
  assert(#bubbleLayouts <= MAXIMUM_BUBBLES_COUNT, string.format("Too many bubble layouts: %s, maximum: %s", #bubbleLayouts, MAXIMUM_BUBBLES_COUNT))
  deactivateBubbles(self)
  activateBubbles(self, bubbleLayouts, getBubbleColorFunctionForRound(roundColor))

  self.bubbleLayouts = bubbleLayouts
end

function TypoBubbleSwarm:addBubbleLayouts(bubbleLayouts, roundColor)
  assert(#bubbleLayouts <= #self.inactiveBubbles, string.format("Too many bubble layouts: %s, currently inactive: %s", #bubbleLayouts, #self.inactiveBubbles))

  local activeBubbles = self.activeBubbles
  local firstNewBubbleIndex = #self.activeBubbles + 1

  activateBubbles(self, bubbleLayouts, getBubbleColorFunctionForBonusRound())

  for i = firstNewBubbleIndex, #activeBubbles do
    local bubble = activeBubbles[i]
    local offscreenY = bubble:getY() - getDistanceToSwirlForWinAnimation(self)
    bubble:setY(offscreenY)
  end

  TableUtils.concatenateArrays(self.bubbleLayouts, bubbleLayouts)
end

-- Animations -------------------------

function TypoBubbleSwarm:getSwirlDownAnimation()
  return createSwirlDownAnimation(self, getOffscreenEndPoint)
end

function TypoBubbleSwarm:getSwirlDownBonusAnimation()
  return createSwirlDownBonusAnimation(self, getFlippedEndPoint)
end

function TypoBubbleSwarm:getSwirlUpAnimation(roundColor)
  return createSwirlUpAnimation(self, getBubbleColorFunctionForRound(roundColor))
end

function TypoBubbleSwarm:getSwirlUpBonusAnimation()
  return createSwirlUpAnimation(self, getBubbleColorFunctionForBonusRound())
end

function TypoBubbleSwarm:getSwirlUpWinAnimation()
  local offsetY = getDistanceToSwirlForWinAnimation(self)
  return getSwirlUpAnimationNoFlip(self, offsetY)
end

function TypoBubbleSwarm:getHideAnimation()
  local offsetY = Screen.getHeightPart(1.2)
  return getSwirlUpAfterWinAnimation(self, offsetY)
end

function TypoBubbleSwarm:getDimDownAnimation(colorChangeDuration)
  local activeBubbles = self.activeBubbles
  local bubbleLayouts = self.bubbleLayouts
  local animation = AHSpawn.new()

  for i, bubble in ipairs(activeBubbles) do
    local layout = bubbleLayouts[i]
    local dimmedColor = TypoBubbleLayouts.getDimmedColor(layout.colorKey)
    animation:addAction(bubble:getDimDownAnimation(dimmedColor, colorChangeDuration))
  end

  return animation
end

function TypoBubbleSwarm:getRevertDimDownAnimation(colorChangeDuration)
  local activeBubbles = self.activeBubbles
  local animation = AHSpawn.new()

  for _, bubble in ipairs(activeBubbles) do
    animation:addAction(bubble:getRevertDimDownAnimation(colorChangeDuration))
  end

  return animation
end

-- Re-attachment ----------------------

function TypoBubbleSwarm:attachToBottom(newParent)
  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local startPoint = layout.startPoint
    local endPoint = getFlippedEndPoint(startPoint)

    bubble:setLoc(endPoint.x, endPoint.y)
  end

  self:setAnchorPointBottom()
  newParent:takeChild(self)
  self:setLoc(newParent:getWidthPart(0.5), 0)
end

function TypoBubbleSwarm:attachToBottomOffscreen(newParent)
  for index, layout in ipairs(self.bubbleLayouts) do
    local bubble = self.activeBubbles[index]
    local startPoint = layout.startPoint
    local endPoint = getFlippedEndPoint(startPoint)
    endPoint.y = endPoint.y - Screen.getHeightPart(0.5)

    bubble:setLoc(endPoint.x, endPoint.y)
  end

  self:setAnchorPointBottom()
  newParent:takeChild(self)
  self:setLoc(newParent:getWidthPart(0.5), 0)
end

function TypoBubbleSwarm:attachToTop(newParent)
  for _, bubble in ipairs(self.activeBubbles) do
    local newX, newY = newParent:transformObjectLoc(bubble)
    bubble:setLoc(newX, newY)
  end

  self:setAnchorPointTop()
  newParent:takeChild(self)
  self:setLoc(newParent:getWidthPart(0.5), newParent:getHeight())
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createBubbles(self)
  -- This is an optimization. Create and insert bubbles during creation, so that we do not need to
  -- do insertions and removals during animations.
  local inactiveBubbles = {}
  local defaultColor = { 1, 1, 1 }
  local defaultRadius = 1

  for i = 1, MAXIMUM_BUBBLES_COUNT do
    local bubble = TypoBubble.new(defaultRadius, defaultColor)
    bubble:setVisibleRecursive(false)
    self:insert(bubble)
    table.insert(inactiveBubbles, bubble)
  end

  self.inactiveBubbles = inactiveBubbles
  self.activeBubbles = {}
end

-- Constructor ------------------------

function TypoBubbleSwarm.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, TypoBubbleSwarm)

  createBubbles(self, bubbleLayouts, roundColor)

  self.bubbleLayouts = nil

  return self
end
