--- TypoConfiguration
-- @author Created by amorant on June 29, 2015.

class("TypoConfiguration")

require("data.keyframes.KeyframeSet")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Game parameters --------------------

local LIVES_COUNT = 5
local ROUND_DURATION_THRESHOLDS = KeyframeSet.newThreshold(
  Keyframe.new(0.0, { 75, 65, 55 }),
  Keyframe.new(0.15, { 65, 55, 45 })
)
local BONUS_ROUND_DURATION = 12
local ERRORS_PER_ROUND = 4

local SECONDS_TO_TAKE_ON_INCORRECT_THRESHOLDS = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {seconds = 2}),
  Keyframe.new(0.2, {seconds = 3}),
  Keyframe.new(0.6, {seconds = 4}),
  Keyframe.new(0.8, {seconds = 5})
)

-- Difficulty -------------------------

local DIFFICULTY_LOWER_BOUND_FOR_HARD_CONCEPT = 0.3
local DIFFICULTY_UPPER_BOUND_FOR_EASY_CONCEPT = 0.55

local TOTAL_ERRORS_COUNT = 3 * ERRORS_PER_ROUND
local MINIMUM_TOTAL_HARD_ERRORS = 0
local MAXIMUM_TOTAL_HARD_ERRORS = TOTAL_ERRORS_COUNT - 2

local HARD_ERRORS_FOR_BASIC_CONCEPTS_KEYFRAMES = KeyframeSet.newThreshold(
  Keyframe.new(0.0, {hardErrorsCount = 0})
)

local HARD_ERRORS_FOR_EASY_CONCEPTS_KEYFRAMES = KeyframeSet.newInterpolated(
  Keyframe.new(0.0, {hardErrorsCount = MINIMUM_TOTAL_HARD_ERRORS}),
  Keyframe.new(0.1, {hardErrorsCount = MINIMUM_TOTAL_HARD_ERRORS}),
  Keyframe.new(DIFFICULTY_UPPER_BOUND_FOR_EASY_CONCEPT, {hardErrorsCount = MAXIMUM_TOTAL_HARD_ERRORS}),
  Keyframe.new(1.0, {hardErrorsCount = MAXIMUM_TOTAL_HARD_ERRORS})
)

local HARD_ERRORS_FOR_HARD_CONCEPTS_KEYFRAMES = KeyframeSet.newInterpolated(
  Keyframe.new(0.0, {hardErrorsCount = MINIMUM_TOTAL_HARD_ERRORS}),
  Keyframe.new(DIFFICULTY_LOWER_BOUND_FOR_HARD_CONCEPT, {hardErrorsCount = MINIMUM_TOTAL_HARD_ERRORS}),
  Keyframe.new(1.0, {hardErrorsCount = MAXIMUM_TOTAL_HARD_ERRORS})
)

-- Scoring ----------------------------

local SCORE_PER_CORRECT_ANSWER = 700
local BONUS_SCORE_PER_EXTRA_LIFE = 625
-- Bonus round makes up for 1 mistake:
local SCORE_FOR_BONUS_ROUND = SCORE_PER_CORRECT_ANSWER + BONUS_SCORE_PER_EXTRA_LIFE
-- Add some sugar on top for score variability
local MINIMUM_BONUS_SCORE_FOR_BONUS_ROUND_SPEED = 100
local MAXIMUM_BONUS_SCORE_FOR_BONUS_ROUND_SPEED = 200

local MAXIMUM_INCORRECT_ANSWERS_RANK_TWO = 2
local MAXIMUM_INCORRECT_ANSWERS_RANK_THREE = 0
local EXTRA_LIVES_RANK_TWO = LIVES_COUNT - MAXIMUM_INCORRECT_ANSWERS_RANK_TWO - 1
local EXTRA_LIVES_RANK_THREE = LIVES_COUNT - MAXIMUM_INCORRECT_ANSWERS_RANK_THREE - 1


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getKeyframeSetForConceptDifficulty(conceptDifficulty)
  if (conceptDifficulty == "Basic") then
    return HARD_ERRORS_FOR_BASIC_CONCEPTS_KEYFRAMES
  elseif (conceptDifficulty == "Easy") then
    return HARD_ERRORS_FOR_EASY_CONCEPTS_KEYFRAMES
  elseif (conceptDifficulty == "Hard") then
    return HARD_ERRORS_FOR_HARD_CONCEPTS_KEYFRAMES
  else
    error(string.format("Unexpected concept difficulty %s", tostring(conceptDifficulty)))
  end
end

local function getKeyframeForDifficulty(conceptDifficulty, gameDifficulty)
  local keyframeSet = getKeyframeSetForConceptDifficulty(conceptDifficulty)
  return keyframeSet:getFrameAt(gameDifficulty)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Scoring ----------------------------

function TypoConfiguration.getScorePerCorrectAnswer()
  return SCORE_PER_CORRECT_ANSWER
end

function TypoConfiguration.getBonusScorePerExtraLife()
  return BONUS_SCORE_PER_EXTRA_LIFE
end

function TypoConfiguration.getScoreForBonusRound()
  return SCORE_FOR_BONUS_ROUND
end

function TypoConfiguration.getBonusScoreForBonusRoundSpeed(remainingSeconds)
  return MathUtils.interpolateLinear(remainingSeconds,
                                     0,
                                     BONUS_ROUND_DURATION,
                                     MINIMUM_BONUS_SCORE_FOR_BONUS_ROUND_SPEED,
                                     MAXIMUM_BONUS_SCORE_FOR_BONUS_ROUND_SPEED)
end

function TypoConfiguration.getPassScore()
  -- Pass rank is attained by finising the game without losing all hearts.
  local minimumCorrectAnswers = TOTAL_ERRORS_COUNT - LIVES_COUNT + 1
  local baseScore = minimumCorrectAnswers * SCORE_PER_CORRECT_ANSWER
  return baseScore
end

function TypoConfiguration.getRankTwoScore()
  -- Rank 2 is attained by losing 1 or 2 hearts, or losing 3 hearts and getting right the bonus round.
  local minimumCorrectAnswers = TOTAL_ERRORS_COUNT - MAXIMUM_INCORRECT_ANSWERS_RANK_TWO
  local baseScore = minimumCorrectAnswers * SCORE_PER_CORRECT_ANSWER
  local livesBonusScore = EXTRA_LIVES_RANK_TWO * BONUS_SCORE_PER_EXTRA_LIFE
  return baseScore + livesBonusScore
end

function TypoConfiguration.getRankThreeScore()
  -- Rank 3 is attained by losing no hearts, or losing 1 heart and getting right the bonus round.
  local minimumCorrectAnswers = TOTAL_ERRORS_COUNT - MAXIMUM_INCORRECT_ANSWERS_RANK_THREE
  local baseScore = minimumCorrectAnswers * SCORE_PER_CORRECT_ANSWER
  local livesBonusScore = EXTRA_LIVES_RANK_THREE * BONUS_SCORE_PER_EXTRA_LIFE
  return baseScore + livesBonusScore
end

-- Game parameters --------------------

function TypoConfiguration.getLivesCount()
  return LIVES_COUNT
end

function TypoConfiguration.getRoundDuration(roundIndex)
  local difficulty = GameSession.getDifficulty()
  local roundDurations = ROUND_DURATION_THRESHOLDS:getFrameAt(difficulty)

  assert(roundIndex >= 1 and roundIndex <= #roundDurations, string.format("roundIndex is out of bounds: %s", tostring(roundIndex)))

  return roundDurations[roundIndex]
end

function TypoConfiguration.getBonusRoundDuration()
  return BONUS_ROUND_DURATION
end

function TypoConfiguration.getSecondsToTakeOnIncorrect()
  local difficulty = GameSession.getDifficulty()
  local thresholdForDifficulty = SECONDS_TO_TAKE_ON_INCORRECT_THRESHOLDS:getFrameAt(difficulty)
  return thresholdForDifficulty.seconds
end

function TypoConfiguration.getErrorsPerRound()
  return ERRORS_PER_ROUND
end

-- Difficulty -------------------------

function TypoConfiguration.getEasyAndHardErrorsCounts(conceptDifficulty)
  local gameDifficulty = GameSession.getDifficulty()

  local keyframeForDifficulty = getKeyframeForDifficulty(conceptDifficulty, gameDifficulty)

  local hardErrorsCount = MathUtils.round(keyframeForDifficulty.hardErrorsCount)
  local easyErrorsCount = TOTAL_ERRORS_COUNT - hardErrorsCount

  return easyErrorsCount, hardErrorsCount
end

function TypoConfiguration.getDifficultyKeyFrames()
  return {
    { key = 0.0,  valueProbabilities = { Basic = 1.0 } },
    { key = 0.15,  valueProbabilities = { Easy = 1.0 } },
    { key = DIFFICULTY_LOWER_BOUND_FOR_HARD_CONCEPT, valueProbabilities = { Easy = 1.0 } },
    { key = DIFFICULTY_UPPER_BOUND_FOR_EASY_CONCEPT, valueProbabilities = { Hard = 1.0 } },
    { key = 1.0,  valueProbabilities = { Hard = 1.0 } }
  }
end
