--- TypoGame
-- @author Created by amorant on June 29, 2015.

class("TypoGame")

-- Font setup
FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("TypoConfiguration")
require("TypoConceptGenerator")
require("TypoTextController")
require("TypoTextPanel")
require("TypoRoundIntro")
require("widgets.progress.TimeTimer")
require("TypoRoundConfiguration")
require("TypoHeader")
require("TypoAnimationHelper")
require("TypoHowToPlayPanel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Colors -----------------------------

local BACKGROUND_COLOR = { 1, 1, 1 }
local UI_COMPONENTS_COLOR = TypoCommon.COLOR_TEXT_GRAY_LIGHT

-- Dimensions -------------------------

local HEADER_HEIGHT = math.max(Screen.getHeightPart(0.23), Screen.getHeightPart(0.18) + Host.safeAreaTopInset)
local MARGIN_X = Screen.getWidthPart(0.07)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function scrollToNextError(self)
  self.textPanel:scrollToNextError(self.remainingErrorsInRound)
end

local function isRoundComplete(self)
  return (#self.remainingErrorsInRound == 0)
end

-- Wait lock --------------------------

local function waitForRoundCompleteOrTimeout(self)
  local waitLock = AHDelay.new(math.huge)
  self.waitLock = waitLock
  wait(waitLock:start())
end

local function unlockWait(self)
  if self.waitLock then
    self.waitLock:stop()
    self.waitLock = nil
  end
end

-- Handlers ---------------------------

local function handleBonusRoundSpeedyAnswer(self)
  local bonusRoundRemainingSeconds = self.timer:getTimeRemaining()
  local bonusRoundSpeedBonusScore = TypoConfiguration.getBonusScoreForBonusRoundSpeed(bonusRoundRemainingSeconds)
  GameSession.increaseScore(bonusRoundSpeedBonusScore)

  local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.BONUS_ROUND_SPEED_BONUS)
  self.bonusTray:indicateSimpleTextBonus(notificationString, bonusRoundSpeedBonusScore)
end

local function handleFixedAnswerAnimation(self, errorId, tapX, tapY)
  self.animationHelper:playFixedAnswerAnimation(self.currentRound, errorId, tapX, tapY)
end

local function handleCorrectAnswerScoreAndAnimation(self, errorId, tapX, tapY)
  GameSession.contentTracker:notifyCorrectAnswer(self.conceptIdentifier, errorId)
  GameSession.incrementCorrectAnswerCount()

  local answerScore = TypoConfiguration.getScorePerCorrectAnswer()
  if self.currentRound:isBonusRound() then
    answerScore = TypoConfiguration.getScoreForBonusRound()
    handleBonusRoundSpeedyAnswer(self)
  end

  GameSession.increaseScore(answerScore)
  self.scoreLabel:setScore(GameSession.getScore())

  self.animationHelper:playCorrectAnswerAnimation(self.currentRound, errorId, tapX, tapY)
end

local function handleCorrectAnswer(self, errorId, tapX, tapY)
  if self.fixAnswerModeEnabled then
    handleFixedAnswerAnimation(self, errorId, tapX, tapY)
  else
    handleCorrectAnswerScoreAndAnimation(self, errorId, tapX, tapY)
  end

  local roundIsComplete = isRoundComplete(self)

  wait(self.textView:getAnimation())
  if (self.fixAnswerModeEnabled and (not roundIsComplete)) then
    scrollToNextError(self)
  end

  if roundIsComplete then
    if self.fixAnswerModeEnabled then
      self.animationHelper:allErrorsCorrectedAfterTimeout()
    else
      unlockWait(self)
    end
  end
end

local function handleIncorrectAnswer(self, item)
  if not self.fixAnswerModeEnabled then
    local roundErrorType = self.currentRound:getErrorType()
    local tappedWord = item:getText()
    GameSession.contentTracker:notifyIncorrectAnswer(self.conceptIdentifier, roundErrorType, tappedWord)
    GameSession.incrementIncorrectAnswerCount()
    self.animationHelper:playIncorrectAnswerAnimation(self.currentRound, item)
  end
end

local function getAnswerHandler(self)
  return function(isIncorrectWord, errorId, item, tapX, tapY)
    self.textView:setTouchEnabled(false)
    if (isIncorrectWord) then
      TableUtils.removeValueFromArray(self.remainingErrorsInRound, errorId)
    end
    runInCoroutine(function()
      local isTapBelowHeader = (tapY < (self:getHeight() - HEADER_HEIGHT))

      if (isTapBelowHeader) then
        if (isIncorrectWord) then
          handleCorrectAnswer(self, errorId, tapX, tapY)
        else
          handleIncorrectAnswer(self, item)
        end
      end

      if not isRoundComplete(self) then
        if ((self.timer:getTimeRemaining() > 0) or self.fixAnswerModeEnabled) then
          self.textView:setTouchEnabled(true)
        end
      end
    end)
  end
end

local function getTimeoutHandler(self)
  return function()
    local conceptId = self.conceptIdentifier
    self.textView:setTouchEnabled(false)

    if not self.currentRound:isBonusRound() then
      for _, errorId in ipairs(self.remainingErrorsInRound) do
        GameSession.contentTracker:notifyTimeout(conceptId, errorId)
        self.livesCounter:decrement()
      end
    end

    -- We might get a timeout while the correct animation is happening, so we must check if the
    -- round is complete before triggering timeout animation.
    local errorsRemainingCount = #self.remainingErrorsInRound
    local roundIsComplete = (errorsRemainingCount == 0)
    if not roundIsComplete then
      self.fixAnswerModeEnabled = true
      self.textView:setTouchEnabled(true)
      scrollToNextError(self)
      self.animationHelper:playTimeoutAnimation(self.currentRound, errorsRemainingCount)
      self.fixAnswerModeEnabled = false
    end

    unlockWait(self)
  end
end

-- Game flow --------------------------

local function startRound(self, round)
  self.currentRound = round

  local roundLabel = self.roundLabel
  local timer = self.timer
  local textView = self.textView
  local textController = self.textController
  local roundIsBonus = round:isBonusRound()
  local errorsIds = round:getErrorIds()
  self.remainingErrorsInRound = TableUtils.shallowCopy(errorsIds)

  local animationHelper = self.animationHelper
  local roundIntro = nil
  if roundIsBonus then
    roundIntro = TypoRoundIntro.newBonus(round)
  else
    roundIntro = TypoRoundIntro.new(round, self.roundConfiguration:getRoundsCount())
  end

  animationHelper:playRoundIntroTransition(round, roundIntro, self.header, self.textPanel)

  if not roundIsBonus then
    textView:setCorrectTextColor(round:getColor())
  end
  textController:startRound(round)
  textView:updateZIndices()
  self.header:setRound(round)
  timer:setDuration(round:getDuration())

  roundIntro:waitForTap()

  animationHelper:playRoundStartTransition(round, roundIntro, self.header, self.textPanel)

  local shouldShowHowToPlay = (GameSession.hasNeverWon() and not roundIsBonus and round:getRoundIndex() == 1)
  if (shouldShowHowToPlay) then
    local howToPlayPanel = TypoHowToPlayPanel.new()
    self:insertAtCenter(howToPlayPanel)
    howToPlayPanel:showAndWaitForTap()
  end

  timer:start()
  textView:setTouchEnabled(true)
end

local function endRound(self)
  self.textView:setTouchEnabled(false)
  self.timer:stop()
end

local function doRound(self, round)
  startRound(self, round)
  waitForRoundCompleteOrTimeout(self)
  endRound(self)
end

local function isGameLoss(self)
  return self.livesCounter:getLives() <= 0
end

local function endGame(self)
  if isGameLoss(self) then
    GameSession.notifyGameLoss()
  else
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), TypoConfiguration.getBonusScorePerExtraLife(), self.bonusTray)
    self.animationHelper:playWinAnimation()
    GameSession.notifyGameWin()
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoGame:gameFlow()
  local rounds = self.roundConfiguration:getRounds()

  for _, round in ipairs(rounds) do
    doRound(self, round)
    if isGameLoss(self) then
      break
    end
  end

  if not isGameLoss(self) then
    local bonusRound = self.roundConfiguration:getBonusRound()
    doRound(self, bonusRound)
  end

  endGame(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  local passScore = TypoConfiguration.getPassScore()
  local rankTwoScore = TypoConfiguration.getRankTwoScore()
  local rankThreeScore = TypoConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  local difficultyKeyFrames = TypoConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, GameSession.getDifficulty())
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  GameSession.setInitialLives(TypoConfiguration.getLivesCount())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

local function createGameElements(self)
  local concept = TypoConceptGenerator.new():generateNextValue()
  GameSession.contentTracker:notifyPresentedConcept(concept.identifier)
  local roundConfiguration = TypoRoundConfiguration.new(concept)

  local textPanelHeight = self:getHeight() - HEADER_HEIGHT
  local textPanel = TypoTextPanel.new(self:getWidth() - 2 * MARGIN_X, textPanelHeight, getAnswerHandler(self))
  local textView = textPanel:getTextView()
  local textController = TypoTextController.new(concept.correctionsByErrorId, textView)
  local textModel = textController:getTextModel()
  textView:setModel(textModel)
  textPanel:setAnchorPointTop()
  local negativeMargin = self:getHeightPart(0.03)
  textPanel:setLoc(self:getWidthPart(0.5), self:getHeight() - HEADER_HEIGHT + negativeMargin)

  local header = TypoHeader.new(self:getWidth(), HEADER_HEIGHT, MARGIN_X, getTimeoutHandler(self))
  header:setAnchorPointTop()
  header:setLoc(self:getWidthPart(0.5), self:getHeight() + HEADER_HEIGHT)

  local animationHelper = TypoAnimationHelper.new(textController, textPanel, header, self.uiOverlay)
  self:insertAtCenter(animationHelper)

  self.conceptIdentifier = concept.identifier
  self.roundConfiguration = roundConfiguration
  self.background = background
  self.textPanel = textPanel
  self.textView = textView
  self.textController = textController
  self.header = header
  self.timer = header:getTimer()
  self.animationHelper = animationHelper
  self.fixAnswerModeEnabled = false
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.livesCounter = builder:addLivesDisplay(TypoConfiguration.getLivesCount())
end

-- Constructor ------------------------

function TypoGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, TypoGame)

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  return self
end
