--- TypoHeader
-- @author Created by anton on August 12, 2015.

class("TypoHeader")

require("TypoCommon")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = { 1, 1, 1 }
local SEPARATOR_COLOR = { 0, 0, 0, 0.3 }
local ROUND_TEXT_COLOR = TypoCommon.COLOR_TEXT_GRAY_LIGHT

local ROUND_LABEL_TEXT_SIZE = 16
local ERRORS_LABEL_TEXT_SIZE = 14
local TIMER_TEXT_SIZE = 25

local TOP_MARGIN_PADDING = 48.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Update helpers ---------------------

local function alignLeftLabels(self)
  local topMarginY = self:getHeight() - (Host.safeAreaTopInset + TOP_MARGIN_PADDING)

  local roundLabelY = self:getHeightPart(0.35)
  local errorsFoundLabelY = self:getHeightPart(0.22)

  local roundLabelTopY = roundLabelY + self.roundLabel:getHeight()
  local realignmentNeeded = roundLabelTopY > topMarginY
  if (realignmentNeeded) then
    local roundLabelOffset = roundLabelTopY - topMarginY

    local scaleFactor = 0.75
    local errorsFoundLabelOffset = roundLabelOffset * scaleFactor

    roundLabelY = roundLabelY - roundLabelOffset
    errorsFoundLabelY = errorsFoundLabelY - errorsFoundLabelOffset
  end

  self.roundLabel:setY(roundLabelY)
  self.errorsFoundLabel:setY(errorsFoundLabelY)
end

local function updateRoundLabel(self, round)
  local roundLabel = self.roundLabel
  local roundDescription = round:getDescription()
  local text = roundDescription:upper()
  roundLabel:setText(text)
end

local function updateColors(self, round)
  local errorsFoundLabel = self.errorsFoundLabel
  local timer = self.timer

  if not round:isBonusRound() then
    local roundColor = round:getColor()
    errorsFoundLabel:setColor(unpack(roundColor))
    timer:setColor(unpack(roundColor))
  else
    TypoCommon.startBonusColorAnimation(errorsFoundLabel)
    TypoCommon.startBonusColorAnimation(timer)
  end
end

local function updateErrorsFoundLabel(self)
  local errorsFoundLabel = self.errorsFoundLabel
  local errorsFoundCount = self.errorsFoundCount
  local errorsTotalCount = self.errorsTotalCount
  errorsFoundLabel:setText(string.format("%s of %s found", errorsFoundCount, errorsTotalCount))
end

-- Color animations -------------------

local function createColorChangeAnimation(self, duration, errorsFoundLabelColor, timerColor)
  local errorsFoundLabel = self.errorsFoundLabel
  local timer = self.timer
  local colorErrorsFoundLabel = AHColorTo.new(errorsFoundLabel, duration, nil, errorsFoundLabelColor)
  local colorTimer = AHColorTo.new(timer, duration, nil, timerColor)
  local animation = AHSpawn.new(colorErrorsFoundLabel, colorTimer)

  return AHCallLogic.new(function()
    local round = self.round
    if round:isBonusRound() then
      TypoCommon.stopBonusColorAnimation(errorsFoundLabel)
      TypoCommon.stopBonusColorAnimation(timer)
    end
    self.colorInterrupt:insert(animation)
  end)
end

local function createIncorrectDimTextAnimation(self, duration)
  return createColorChangeAnimation(self, duration, ROUND_TEXT_COLOR, TypoCommon.COLOR_WRONG_ANSWER)
end

local function createTimeoutDimTextAnimation(self, duration)
  return createColorChangeAnimation(self, duration, ROUND_TEXT_COLOR, ROUND_TEXT_COLOR)
end

local function createRevertDimTextAnimation(self, duration)
  local round = self.round
  if round:isBonusRound() then
    return AHCallLogic.new(function()
      updateColors(self, round)
    end)
  else
    local roundColor = round:getColor()
    return createColorChangeAnimation(self, duration, roundColor, roundColor)
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoHeader:getTimer()
  return self.timer
end

function TypoHeader:setRound(round)
  self.round = round
  self.errorsFoundCount = 0
  self.errorsTotalCount = #round:getErrorIds()

  updateRoundLabel(self, round)
  updateColors(self, round)
  updateErrorsFoundLabel(self)
  alignLeftLabels(self)

  self:setVisibleRecursive(true)
end

function TypoHeader:incrementErrorsFound()
  self.errorsFoundCount = self.errorsFoundCount + 1
  updateErrorsFoundLabel(self)
end

function TypoHeader:getIncorrectAnswerAnimation(colorChangeDuration)
  local timer = self.timer
  local colorAnimation = createIncorrectDimTextAnimation(self, colorChangeDuration)
  local secondsToTake = TypoConfiguration.getSecondsToTakeOnIncorrect()

  return AHSpawn.new(
    colorAnimation,
    function()
      timer:pause()
      timer:addTimeRemaining(-secondsToTake)
    end
  )
end

function TypoHeader:getRevertFromIncorrectAnswerAnimation(colorChangeDuration)
  local timer = self.timer

  local colorAnimation = createRevertDimTextAnimation(self, colorChangeDuration)

  return AHSpawn.new(
    colorAnimation,
    function() timer:resume() end)
end

function TypoHeader:getTimeoutAnimation(colorChangeDuration)
  return createTimeoutDimTextAnimation(self, colorChangeDuration)
end

function TypoHeader:getRevertTimeoutAnimation(colorChangeDuration)
  return createRevertDimTextAnimation(self, colorChangeDuration)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createElements(self, marginX, timeoutHandler)
  local background = FilledRectangle.new(self:getWidth(), self:getHeight(), BACKGROUND_COLOR)
  self:insertAtCenter(background)

  local separator = FilledRectangle.new(self:getWidth() - marginX, 1, SEPARATOR_COLOR)
  separator:setAnchorPointLeft()
  separator:setLoc(marginX, 0)
  self:insert(separator)

  local roundLabelWidth = self:getWidthPart(0.86) - 2 * marginX
  local roundLabel = TextArea.new("", Font.new(FontFaces.DIN_OT_MEDIUM, ROUND_LABEL_TEXT_SIZE), roundLabelWidth)
  roundLabel:setColor(unpack(ROUND_TEXT_COLOR))
  roundLabel:setAnchorPointBottomLeft()
  roundLabel:setX(marginX)
  self:insert(roundLabel)
  self.roundLabel = roundLabel

  local errorsFoundLabel = Label.new("", Font.new(FontFaces.DIN_OT_MEDIUM, ERRORS_LABEL_TEXT_SIZE))
  errorsFoundLabel:setAnchorPointLeft()
  errorsFoundLabel:setX(marginX)
  self:insert(errorsFoundLabel)
  self.errorsFoundLabel = errorsFoundLabel

  alignLeftLabels(self)

  local adjustedTimerTextSize = FontScaling.getScaledSize(TIMER_TEXT_SIZE)
  local secondsToPreserveMinutePlace = 61
  local timer =  TimeTimer.new(FontFaces.DIN_OT_BOLD, adjustedTimerTextSize, secondsToPreserveMinutePlace)
  timer:setTimeoutHandler(timeoutHandler)
  timer:addStandardWarningSequence()
  timer:setAnchorPointRight()
  timer:setLoc(self:getWidth() - marginX, self:getHeightPart(0.3))
  self:insert(timer)
  self.timer = timer
end

-- Constructor ------------------------

function TypoHeader.new(width, height, marginX, timeoutHandler)
  local self = Group.new(width, height)
  Mixin.mix(self, TypoHeader)

  createElements(self, marginX, timeoutHandler)

  -- This is the only visible element on frame 1. By making it invisible, we start off with a white
  -- clear frame.
  self:setVisibleRecursive(false)

  self.errorsFoundCount = 0
  self.round = nil
  self.colorInterrupt = ActionInterrupt.new()

  return self
end
