--- TypoRoundConfiguration
-- @author Created by anton on July 24, 2015.

class("TypoRoundConfiguration")

require("TypoRound")
require("TypoCommon")
require("TypoBubbleLayouts")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ROUND_TYPES = {
  TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION,
  TypoCommon.ERROR_SPELLING,
  TypoCommon.ERROR_EXTRA_MISSING
}

local ROUND_COLORS = {
  TypoCommon.COLOR_PURPLE,
  TypoCommon.COLOR_PINK,
  TypoCommon.COLOR_ORANGE
}

local ROUND_ENTER_LAYOUTS = {
  TypoBubbleLayouts.LAYOUTS_ONE,
  TypoBubbleLayouts.LAYOUTS_TWO,
  TypoBubbleLayouts.LAYOUTS_THREE
}

local ROUND_EXIT_LAYOUTS = {
  TypoBubbleLayouts.LAYOUTS_TWO,
  TypoBubbleLayouts.LAYOUTS_THREE,
  TypoBubbleLayouts.LAYOUTS_BONUS
}

local BONUS_ROUND_TYPE = TypoCommon.ERROR_BONUS
local BONUS_ROUND_ENTER_LAYOUTS = TypoBubbleLayouts.LAYOUTS_BONUS
local BONUS_ROUND_EXIT_LAYOUTS = TypoBubbleLayouts.LAYOUTS_BONUS

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getClassifiedErrorIds(correctionsByErrorId)
  local classifiedErrorIds = {}

  for roundIndex, errorType in ipairs(ROUND_TYPES) do
    classifiedErrorIds[roundIndex] = {
      ["Easy"] = {},
      ["Hard"] = {}
    }
    for _, correction in pairs(correctionsByErrorId) do
      if correction:getErrorType() == errorType then
        table.insert(classifiedErrorIds[roundIndex][correction:getDifficulty()], correction:getErrorId())
      end
    end
  end

  return classifiedErrorIds
end

local function makeDifficultiesArray(easyErrorsCount, hardErrorsCount)
  local difficulties = {}

  for i = 1, easyErrorsCount do
    table.insert(difficulties, "Easy")
  end
  for i = 1, hardErrorsCount do
    table.insert(difficulties, "Hard")
  end

  TableUtils.shuffleArray(difficulties)
  return difficulties
end

local function makeRounds(passagesByErrorType, classifiedErrorIds, conceptDifficulty)
  local easyErrorsCount, hardErrorsCount = TypoConfiguration.getEasyAndHardErrorsCounts(conceptDifficulty)
  local difficulties = makeDifficultiesArray(easyErrorsCount, hardErrorsCount)

  local rounds = {}

  for roundIndex, roundType in pairs(ROUND_TYPES) do
    local selectedErrors = {}

    for i = 1, TypoConfiguration.getErrorsPerRound() do
      local nextDifficulty = table.remove(difficulties, 1)
      local errorsToSelectFrom = classifiedErrorIds[roundIndex][nextDifficulty]

      if #errorsToSelectFrom == 0 then
        table.insert(difficulties, nextDifficulty)
        if nextDifficulty == "Easy" then
          nextDifficulty = "Hard"
        else
          nextDifficulty = "Easy"
        end
        TableUtils.removeValueFromArray(difficulties, nextDifficulty)
        errorsToSelectFrom = classifiedErrorIds[roundIndex][nextDifficulty]
      end

      assert(#errorsToSelectFrom > 0, string.format("No more errors to select from for round %s", roundIndex))

      local doRemove = true
      local errorId = TableUtils.getRandomElement(errorsToSelectFrom, doRemove)
      table.insert(selectedErrors, errorId)
    end

    assert(#selectedErrors == TypoConfiguration.getErrorsPerRound(),
           string.format("Selected the wrong number of errors for round %s: %s", roundIndex, #selectedErrors))

    local passageText = passagesByErrorType[roundType]
    local roundColor = ROUND_COLORS[roundIndex]
    local roundEnterLayouts = ROUND_ENTER_LAYOUTS[roundIndex]
    local roundExitLayouts = ROUND_EXIT_LAYOUTS[roundIndex]
    table.insert(rounds, TypoRound.new(passageText, roundIndex, roundType, selectedErrors, roundColor, roundEnterLayouts, roundExitLayouts))
  end

  assert(#difficulties == 0, string.format("The remaining number of difficulties to pick should be 0, it was: %s", #difficulties))

  return rounds
end

local function makeBonusRound(bonusText, correctionsByErrorId)
  local filteredErrorIds = {}
  for errorId, correction in pairs(correctionsByErrorId) do
    if correction:getErrorType() == BONUS_ROUND_TYPE then
      table.insert(filteredErrorIds, errorId)
    end
  end

  local selectedErrorIds = { TableUtils.getRandomElement(filteredErrorIds) }
  return TypoRound.newBonusRound(bonusText, selectedErrorIds, BONUS_ROUND_ENTER_LAYOUTS, BONUS_ROUND_EXIT_LAYOUTS)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoRoundConfiguration:getRounds()
  return self.rounds
end

function TypoRoundConfiguration:getRoundsCount()
  return #self.rounds
end

function TypoRoundConfiguration:getBonusRound()
  return self.bonusRound
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TypoRoundConfiguration.new(concept)
  local self = {}
  Mixin.mix(self, TypoRoundConfiguration)

  local passagesByErrorType = concept.passagesByErrorType
  local bonusText = concept.bonusText
  local correctionsByErrorId = concept.correctionsByErrorId
  local conceptDifficulty = concept.difficulty

  local classifiedErrorIds = getClassifiedErrorIds(correctionsByErrorId)
  local rounds = makeRounds(passagesByErrorType, classifiedErrorIds, conceptDifficulty)
  local bonusRound = makeBonusRound(bonusText, correctionsByErrorId)

  self.rounds = rounds
  self.bonusRound = bonusRound

  return self
end
