--- TypoRoundIntro
-- @author Created by anton on July 17, 2015.

class("TypoRoundIntro")

require("TypoCommon")
require("TypoMoreInfoPanel")
require("TypoBonusLabel")
require("TypoSound")
require("TypoCorrection")
require("TypoTextController")
require("TypoExampleTextController")
require("TypoExampleTextView")
require("widgets.TapToContinue")
require("graphics.primitives.Arrow")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = { 1, 1, 1 }
local TEXT_COLOR = TypoCommon.COLOR_TEXT_GRAY_DARK
local TITLE_TEXT_COLOR = TypoCommon.COLOR_TEXT_GRAY_LIGHT

local TITLE_TEXT_SIZE = 16
local DESCRIPTION_TEXT_SIZE = 31
local EXAMPLE_TEXT_SIZE = 18
local CONTINUE_TEXT_SIZE = 14
local MORE_INFO_TEXT_SIZE = 16
local BONUS_TITLE_TEXT_SIZE = 68
local BONUS_TEXT_SIZE = 21

local EXAMPLES = {
  [TypoCommon.ERROR_SPELLING] = {
    text = string.format("In addition to his <%s errorId=\"error_1\">pionering</%s> work on photosynthesis, Ingenhousz helped promote inoculation against smallpox.",
                         TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL),
    corrections = {
      TypoCorrection.newExampleCorrection(1, "pioneering", "pionering")
    }
  },

  [TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION] = {
    text = string.format("In addition to his pioneering work on photosynthesis, Ingenhousz helped promote <%s errorId=\"error_1\">Inoculation</%s> against <%s errorId=\"error_2\">smallpox</%s>",
                         TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL),
    corrections = {
      TypoCorrection.newExampleCorrection(1, "inoculation", "Inoculation"),
      TypoCorrection.newExampleCorrection(2, "smallpox.", "smallpox")
    }
  },

  [TypoCommon.ERROR_EXTRA_MISSING] = {
    text = string.format("In addition <%s errorId=\"error_1\">his</%s> pioneering work <%s errorId=\"error_2\">on on</%s> photosynthesis, Ingenhousz helped promote inoculation against smallpox.",
                         TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL, TypoCommon.ERROR_TAG_LABEL),
    corrections = {
      TypoCorrection.newExampleCorrection(1, "to his", "his"),
      TypoCorrection.newExampleCorrection(2, "on", "on on")
    }
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- Animations -------------------------

local function getReplacementAnimation(self)
  if self.isBonus then
    return AHDoNothing.new()
  end

  local textController = self.textController
  local corrections = self.example.corrections
  local textModel = textController:getTextModel()
  local exampleTextView = self.exampleTextView

  local function showCorrect()
    for _, correction in ipairs(corrections) do
      local errorId = correction:getErrorId()
      textController:correctError(errorId)
      exampleTextView:getCorrectAnswerColorAnimation(errorId):start()
      waitFor(1.5)
    end
  end

  local function showIncorrect()
    -- We use a transaction to revert corrections all at once
    textModel:beginTransaction()
    for _, correction in ipairs(corrections) do
      local errorId = correction:getErrorId()
      textController:revertError(errorId)
      exampleTextView:getRevertCorrectAnimation(errorId):start()
    end
    textModel:endTransaction()
  end

  return AHLoop.new(function()
    waitFor(1)
    showCorrect()
    waitFor(1.5)
    showIncorrect()
    waitFor(2)
  end)
end

local function getShowAnimation(self)
  local replacementAnimation = getReplacementAnimation(self)
  self.replacementAnimation = replacementAnimation

  local translationDuration = 1

  local showAnimation = AHSpawn.new(
    AHLocationYBy.new(self.continueLabel, translationDuration, AHEaseQuadraticOut.new(), self:getHeight()),
    AHOpacityTo.new(self, translationDuration * 0.5, AHEaseQuadraticOut.new(), 1),
    function()
      replacementAnimation:start()
    end
  )

  if self.isBonus then
    showAnimation:addAction(self.bonusLabel:getShowAnimation(2))
    showAnimation:addAction(AHDelayed.new(0.2, AHLocationYBy.new(self.roundLabel, 1.2, AHEaseQuadraticOut.new(), self:getHeight())))
    showAnimation:addAction(AHDelayed.new(0.4, AHLocationYBy.new(self.bonusExplanationLabel, 1.2, AHEaseQuadraticOut.new(), self:getHeight())))
  else
    showAnimation:addAction(AHLocationYBy.new(self.column, translationDuration, AHEaseQuadraticOut.new(), self:getHeight()))
  end

  return showAnimation
end

local function getHideAnimation(self)
  local translationDuration = 0.8

  local hideAnimation = AHSpawn.new(
    AHLocationYBy.new(self.continueLabel, translationDuration, AHEaseQuadraticIn.new(), -self:getHeight()),
    AHDelayed.new(0.3, AHOpacityTo.new(self, 0.3, AHEaseQuadraticIn.new(), 0)),
    function()
      self.replacementAnimation:stop()
    end
  )

  if self.isBonus then
    hideAnimation:addAction(AHLocationYBy.new(self.bonusExplanationLabel, 0.5, AHEaseQuadraticIn.new(), -self:getHeight()))
    hideAnimation:addAction(AHDelayed.new(0.05, AHLocationYBy.new(self.roundLabel, 0.5, AHEaseQuadraticIn.new(), -self:getHeight())))
    hideAnimation:addAction(AHDelayed.new(0.1, AHLocationYBy.new(self.bonusLabel, 0.5, AHEaseQuadraticIn.new(), -self:getHeight())))
  else
    hideAnimation:addAction(AHLocationYBy.new(self.column, translationDuration, AHEaseQuadraticIn.new(), -self:getHeight()))
    hideAnimation:addAction(function()
      self.moreInfoButton:setTouchEnabled(false)
    end)
  end

  return hideAnimation
end

-- More info panel --------------------

local function prepareToShowMoreInfoPanel(self)
  local moreInfoPanel = self.moreInfoPanel
  local offscreenX = moreInfoPanel:getX() + Screen.width
  moreInfoPanel:setX(offscreenX)
end

local function playShowMoreInfoPanel(self)
  local moreInfoPanel = self.moreInfoPanel
  local column = self.column
  local distance = -Screen.width
  local animation = AHSpawn.new(
    AHLocationXBy.new(moreInfoPanel, 0.5, AHEaseQuadraticInOut.new(), distance),
    AHLocationXBy.new(column, 0.5, AHEaseQuadraticInOut.new(), distance)
  )
  wait(animation:start())
end

local function playHideMoreInfoPanel(self)
  local moreInfoPanel = self.moreInfoPanel
  local column = self.column
  local distance = Screen.width
  local animation = AHSpawn.new(
    AHLocationXBy.new(moreInfoPanel, 0.5, AHEaseQuadraticInOut.new(), distance),
    AHLocationXBy.new(column, 0.5, AHEaseQuadraticInOut.new(), distance)
  )
  wait(animation:start())
end

-- Touch handlers ---------------------

local function getMoreInfoTouchHandler(self)
  return function()
    assert(not self.isShowingMoreInfo, "self.isShowingMoreInfo should be false")
    self.isShowingMoreInfo = true
    TypoSound.playButtonTap()
    playShowMoreInfoPanel(self)
    self.moreInfoPanel:setTouchEnabled(true)
  end
end

local function getBackTouchHandler(self)
  return function()
    assert(self.isShowingMoreInfo, "self.isShowingMoreInfo should be true")
    self.isShowingMoreInfo = false
    self.moreInfoPanel:setTouchEnabled(false)
    TypoSound.playButtonTap()
    playHideMoreInfoPanel(self)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoRoundIntro:getShowAnimation()
  return getShowAnimation(self)
end

function TypoRoundIntro:getHideAnimation()
  return getHideAnimation(self)
end

function TypoRoundIntro:prepareToShow()
  self:setOpacity(0)
  self.continueLabel:setY(self.continueLabel:getY() - self:getHeight())

  if self.isBonus then
    self.bonusLabel:prepareToShow()
    self.roundLabel:setY(self.roundLabel:getY() - self:getHeight())
    self.bonusExplanationLabel:setY(self.bonusExplanationLabel:getY() - self:getHeight())
  else
    self.column:setY(self.column:getY() - self:getHeight())
  end
end

function TypoRoundIntro:waitForTap()
  local tapToContinue = self.tapToContinue
  waitFor(0.25)
  tapToContinue:rearm()

  tapToContinue:waitForTrigger()
  -- Note: This disallows continuing from the more info panel, to avoid doing a conditional
  -- animation to hide this object if showing more info. Ideally we'd allow to continue from
  -- more info panel.
  while self.isShowingMoreInfo do
    getBackTouchHandler(self)()
    tapToContinue:rearm()
    tapToContinue:waitForTrigger()
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createCommonElements(self)
  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)
  background:inheritOpacity(self)
  self:insertAtCenter(background)
  self.background = background

  local safeCanvas = Group.new(Screen.safeWidth, Screen.safeHeight)
  safeCanvas:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  self:insert(safeCanvas)
  self.safeCanvas = safeCanvas

  local continueLabel = Label.new("CONTINUE", Font.new(FontFaces.DIN_OT_MEDIUM, CONTINUE_TEXT_SIZE), nil, Alignment.CENTER)
  continueLabel:setColor(unpack(TITLE_TEXT_COLOR))
  continueLabel:setLoc(self.safeCanvas:getWidthPart(0.5), self.safeCanvas:getHeightPart(0.07))
  continueLabel:inheritOpacity(self)
  self.safeCanvas:insert(continueLabel)
  self.continueLabel = continueLabel

  local underlineThickness = TypoCommon.UNDERLINE_THICKNESS
  local continueUnderline = FilledRectangle.new(continueLabel:getWidth(), underlineThickness)
  local underlineYOffset = continueLabel:getHeightPart(-0.03)
  continueUnderline:setLoc(continueLabel:getWidthPart(0.5), underlineYOffset)
  continueLabel:insert(continueUnderline)
  continueLabel:linkColorAttrToChildren()

  local tapToContinue = TapToContinue.newInGroup(continueLabel)
  tapToContinue:setDims(Screen.width, Screen.getSafeY(0.2))
  tapToContinue:bringToBack()
  tapToContinue:disarm()
  self.tapToContinue = tapToContinue
end

local function createElements(self, round, roundsCount)
  createCommonElements(self)

  local roundColor = round:getColor()

  local column = Column.new(self.safeCanvas:getHeightPart(0.01), Alignment.LEFT)
  column:insertSpacer(self.safeCanvas:getHeightPart(0.32))
  self.column = column

  local indexText = string.format("ROUND %d OF %d", round:getRoundIndex(), roundsCount)
  local indexLabel = Label.new(indexText, Font.new(FontFaces.DIN_OT_MEDIUM, TITLE_TEXT_SIZE))
  indexLabel:setColor(unpack(TITLE_TEXT_COLOR))
  column:insert(indexLabel)

  local descriptionText = round:getDescription():upper()
  local descriptionLabel = TextArea.new(descriptionText, Font.new(FontFaces.DIN_OT_BOLD, DESCRIPTION_TEXT_SIZE), self:getWidthPart(0.86))
  descriptionLabel:setColor(unpack(roundColor))
  column:insertSpacer(self.safeCanvas:getHeightPart(0.0002))
  column:insert(descriptionLabel)

  local moreInfoLabel = Label.new("MORE INFO", Font.new(FontFaces.DIN_OT_MEDIUM, MORE_INFO_TEXT_SIZE))
  local moreInfoArrow = Sprite.newWithName("arrow.png")
  local moreInfoLabelAndArrow = Row.new(5)
  moreInfoLabel:setColor(unpack(roundColor))
  moreInfoArrow:setColor(unpack(roundColor))
  moreInfoLabelAndArrow:insert(moreInfoLabel)
  moreInfoLabelAndArrow:insert(moreInfoArrow)
  moreInfoLabelAndArrow:alignChildren()
  local buttonWidth = moreInfoLabelAndArrow:getWidthPart(1.8)
  local buttonHeight = moreInfoLabelAndArrow:getHeightPart(4)
  Button.addVirtualButton(moreInfoLabelAndArrow, buttonWidth, buttonHeight, getMoreInfoTouchHandler(self))
  column:insert(moreInfoLabelAndArrow)
  self.moreInfoButton = moreInfoLabelAndArrow

  local exampleTitleLabel = Label.new("EXAMPLE", Font.new(FontFaces.DIN_OT_MEDIUM, TITLE_TEXT_SIZE))
  exampleTitleLabel:setColor(unpack(TITLE_TEXT_COLOR))
  column:insertSpacer(self.safeCanvas:getHeightPart(0.035))
  column:insert(exampleTitleLabel)
  column:insertSpacer(self.safeCanvas:getHeightPart(0.0002))

  local example = EXAMPLES[round:getErrorType()]
  local exampleTextView = TypoExampleTextView.new(self.safeCanvas:getWidthPart(0.86))
  local textController = TypoExampleTextController.new(example.corrections, exampleTextView, EXAMPLE_TEXT_SIZE)
  local textModel = textController:getTextModel()
  local textItems = textController.textInterpreter:interpretXmlText(example.text)
  textModel:resetItems(textItems)
  exampleTextView:setTextColor(TEXT_COLOR)
  exampleTextView:setCorrectTextColor(roundColor)
  exampleTextView:setModel(textModel)
  column:insert(exampleTextView)
  self.textController = textController
  self.exampleTextView = exampleTextView
  self.example = example

  column:alignChildren()
  column:linkOpacityAttrToChildren(true)
  column:inheritOpacity(self)
  column:setLoc(self.safeCanvas:getWidthPart(0.5), self.safeCanvas:getHeightPart(0.55))
  self.safeCanvas:insert(column)

  local moreInfoPanel = TypoMoreInfoPanel.new(round, getBackTouchHandler(self))
  self:insertAtCenter(moreInfoPanel)
  self.moreInfoPanel = moreInfoPanel
  prepareToShowMoreInfoPanel(self)
end

local function createBonusElements(self, round)
  createCommonElements(self)

  local bonusLabel = TypoBonusLabel.new("BONUS", FontFaces.DIN_OT_BOLD, BONUS_TITLE_TEXT_SIZE)
  bonusLabel:setLoc(self.safeCanvas:getWidthPart(0.5), self.safeCanvas:getHeightPart(0.53))
  self.safeCanvas:insert(bonusLabel)
  self.bonusLabel = bonusLabel

  local roundLabel = Label.new("ROUND", Font.new(FontFaces.DIN_OT_BOLD, BONUS_TITLE_TEXT_SIZE))
  roundLabel:setColor(unpack(TEXT_COLOR))
  roundLabel:setLoc(self.safeCanvas:getWidthPart(0.5), self.safeCanvas:getHeightPart(0.4))
  self.safeCanvas:insert(roundLabel)
  self.roundLabel = roundLabel

  local bonusExplanationText = "For extra points, find the<br/>single error in the next selection."
  local bonusExplanationLabel = DynamicTextArea.newSimple(bonusExplanationText, self.safeCanvas:getWidthPart(0.86), Font.new(FontFaces.DIN_OT_MEDIUM, BONUS_TEXT_SIZE), Alignment.CENTER)
  bonusExplanationLabel:setColor(unpack(TEXT_COLOR))
  bonusExplanationLabel:setLoc(self.safeCanvas:getWidthPart(0.5), self.safeCanvas:getHeightPart(0.23))
  local indicesOfWordsToColor = { 7, 8 } -- "single error"
  for _, indexToColor in ipairs(indicesOfWordsToColor) do
    TypoCommon.startBonusColorAnimation(bonusExplanationLabel.itemViews[indexToColor])
  end
  self.safeCanvas:insert(bonusExplanationLabel)
  self.bonusExplanationLabel = bonusExplanationLabel
end

-- Constructor ------------------------

function TypoRoundIntro.new(round, roundsCount)
  local self = Group.newFullScreen()
  Mixin.mix(self, TypoRoundIntro)

  createElements(self, round, roundsCount)

  self.isBonus = false
  self.isShowingMoreInfo = false

  return self
end

function TypoRoundIntro.newBonus(round)
  local self = Group.newFullScreen()
  Mixin.mix(self, TypoRoundIntro)

  createBonusElements(self, round)

  self.isBonus = true
  self.isShowingMoreInfo = false

  return self
end
