--- TypoTextView
-- @author Created by amorant on June 29, 2015.

class("TypoTextView")

require("text.dynamic.TextView")
require("widgets.Button")
require("TypoCommon")
require("TypoTextAnimationBuilder")
require("TypoConfiguration")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local TAPPABLE_AREA_EXTRA_WIDTH = ReferenceDevice.getScaledWidth(50)
local TAPPABLE_AREA_EXTRA_HEIGHT = ReferenceDevice.getScaledHeight(12)

local UNDERLINE_OFFSET_Y = FontScaling.getScaledSize(3)
local UNDERLINE_HEIGHT = FontScaling.getScaledSize(1.5)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createUnderlineErrorsAnimation(self, color, duration)
  local animation = AHSpawn.new()

  for item, itemView in pairs(self.viewsForItems) do
    if item:getAttribute(TypoCommon.IS_INCORRECT_ATTRIBUTE) then
      local underlineWidth = itemView:getWidth()
      local underline = FilledRectangle.new(underlineWidth, UNDERLINE_HEIGHT, color)
      underline:setOpacity(0)
      itemView:insertAtBottom(underline)
      underline:offsetY(UNDERLINE_OFFSET_Y)
      animation:addAction(AHOpacityTo.new(underline, duration, nil, 1))
    end
  end

  return animation
end

local function getTouchUpHandler(self, item)
  return function(label, event)
    if not self.isTouchEnabled then
      return
    end

    local isIncorrectWord = item:getAttribute(TypoCommon.IS_INCORRECT_ATTRIBUTE)
    local errorId = item:getAttribute(TypoCommon.ERROR_ID_ATTRIBUTE)
    self.wordTouchedHandler(isIncorrectWord, errorId, item, event.x, event.y)
  end
end

-- Incorrect answer animation ---------

local function createShakeAnimation(itemView)
  local shakeDuration = 0.5
  local shakeAmplitude = 7
  return AHSequence.new(
    AHLocationXBy.new(itemView, 0.2 * shakeDuration, AHEaseSineOut.new(), -1 * shakeAmplitude),
    AHLocationXBy.new(itemView, 0.2 * shakeDuration, AHEaseSineInOut.new(), 1.7 * shakeAmplitude),
    AHLocationXBy.new(itemView, 0.2 * shakeDuration, AHEaseSineInOut.new(), -1.2 * shakeAmplitude),
    AHLocationXBy.new(itemView, 0.2 * shakeDuration, AHEaseSineInOut.new(), 0.8 * shakeAmplitude),
    AHLocationXBy.new(itemView, 0.2 * shakeDuration, AHEaseSineInOut.new(), -0.3 * shakeAmplitude)
  )
end

-- Color change animations ------------

local function createColorChangeAnimation(self, duration, correctColor, errorColor, defaultColor, excludedItem)
  local animation = AHSpawn.new()

  for item, itemView in pairs(self.viewsForItems) do
    if item:getAttribute(TypoCommon.IS_CORRECT_ATTRIBUTE) then
      animation:addAction(AHColorTo.new(itemView, duration, nil, correctColor))
    elseif item:getAttribute(TypoCommon.IS_INCORRECT_ATTRIBUTE) then
      animation:addAction(AHColorTo.new(itemView, duration, nil, errorColor))
    elseif item ~= excludedItem then
      animation:addAction(AHColorTo.new(itemView, duration, nil, defaultColor))
    end
  end

  return animation
end

local function createIncorrectDimTextAnimation(self, duration, tappedItem)
  local excludedItem = tappedItem
  return createColorChangeAnimation(self,
                                    duration,
                                    TypoCommon.COLOR_DIM_GRAY_LIGHT,
                                    TypoCommon.COLOR_DIM_GRAY_LIGHT,
                                    TypoCommon.COLOR_DIM_GRAY_LIGHT,
                                    excludedItem)
end

local function createTimeoutDimTextAnimation(self, duration)
  return createColorChangeAnimation(self,
                                    duration,
                                    TypoCommon.COLOR_DIM_GRAY_LIGHT,
                                    TypoCommon.COLOR_TEXT_ERROR_RED,
                                    TypoCommon.COLOR_DIM_GRAY_LIGHT)
end

local function createRevertDimTextAnimation(self, duration)
  return createColorChangeAnimation(self,
                                    duration,
                                    self.correctTextColor,
                                    self.textColor,
                                    self.textColor)
end

-- Label caching ----------------------

local function getCachedItemView(self, item)
  local cachedViews = self.cachedViews
  local errorId = item:getAttribute(TypoCommon.ERROR_ID_ATTRIBUTE)
  local cachedViewsForErrorId = cachedViews[errorId]

  if cachedViewsForErrorId then
    for cachedItem, cachedItemView in pairs(cachedViewsForErrorId) do
      if cachedItem:equals(item) then
        return cachedItemView
      end
    end
  end

  -- No cached view found is acceptable
  return nil
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- Item operations --------------------

function TypoTextView:createViewForTextItem(item)
  local cachedItemView = getCachedItemView(self, item)
  if cachedItemView then
    return cachedItemView
  end

  local label = self.super.createViewForTextItem(self, item)
  label:setColor(unpack(self.textColor))

  if item:getAttribute(TypoCommon.IS_CORRECT_ATTRIBUTE) then
    label:setColor(unpack(self.correctTextColor))
  elseif item:getAttribute(TypoCommon.IS_INCORRECT_ATTRIBUTE) then
    local tappableAreaWidth = label:getWidth() + TAPPABLE_AREA_EXTRA_WIDTH
    local tappableAreaHeight = label:getHeight() + TAPPABLE_AREA_EXTRA_HEIGHT
    Button.addVirtualButton(label, tappableAreaWidth, tappableAreaHeight, nil, getTouchUpHandler(self, item))
  else
    Button.makeButton(label, nil, getTouchUpHandler(self, item))
  end

  return label
end

function TypoTextView:prerenderViewsForTextItems(errorId, items)
  local cachedViews = self.cachedViews
  assert(not cachedViews[errorId], string.format("Already cached views for errorId: %s", tostring(errorId)))

  itemViews = {}

  for _, item in ipairs(items) do
    local itemView = self:createViewForTextItem(item)
    itemViews[item] = itemView
    -- NB: Normally, the animation built by the TextAnimationBuilder does the insertion of the new
    --     item view. This is an optimization, inserting it early to make the animation smoother.
    self:insert(itemView)
    itemView:setVisibleRecursive(false)
  end

  cachedViews[errorId] = itemViews
end

-- Words belonging to errors have extended tap areas. This makes sure they have priority to handle
-- touches over any overlapping neighbours.
function TypoTextView:updateZIndices()
  for item, itemView in pairs(self.viewsForItems) do
    if item:getAttribute(TypoCommon.IS_INCORRECT_ATTRIBUTE) then
      itemView:bringToFront()
    end
  end
end

-- State ------------------------------

function TypoTextView:setTouchEnabled(enabled)
  self.isTouchEnabled = enabled
end

function TypoTextView:setTextColor(textColor)
  self.textColor = textColor
end

function TypoTextView:setCorrectTextColor(correctTextColor)
  self.correctTextColor = correctTextColor
end

-- Animations -------------------------

function TypoTextView:getViewsForError(errorId)
  return TableUtils.getFilteredTable(self.viewsForItems, function(itemView)
    local item = self.itemsForViews[itemView]
    local itemErrorId = item:getAttribute(TypoCommon.ERROR_ID_ATTRIBUTE)
    return itemErrorId == errorId
  end)
end

function TypoTextView:getCorrectAnswerColorAnimation(errorId, roundIsBonus)
  local itemViewsToAnimate = self:getViewsForError(errorId)

  local animation = AHSpawn.new()
  for item, itemView in pairs(itemViewsToAnimate) do
    if roundIsBonus then
      animation:addAction(function()
          TypoCommon.startBonusColorAnimation(itemView)
      end)
    else
      animation:addAction(AHColorTo.new(itemView, 0.5, nil, self.correctTextColor))
    end

    animation:addAction(function()
      -- Hack: TextItem attributes are supposed to be read-only.
      -- The game was designed to swap all incorrect items, but for animation reasons now we leave
      -- some of them untouched (instead of replacing them with new items with equal text).
      item.attributes[TypoCommon.IS_INCORRECT_ATTRIBUTE] = false
      item.attributes[TypoCommon.IS_CORRECT_ATTRIBUTE] = true
      if itemView.setTouchEnabled then
        itemView:setTouchEnabled(false)
      end
    end)
  end

  return animation
end

function TypoTextView:getIncorrectAnswerAnimation(tappedItem, colorChangeDuration)
  local itemView = self.viewsForItems[tappedItem]

  local shakeAnimation = createShakeAnimation(itemView)
  local dimAnimation = createIncorrectDimTextAnimation(self, colorChangeDuration, tappedItem)

  return AHSpawn.new(shakeAnimation, dimAnimation)
end

function TypoTextView:getRevertFromIncorrectAnswerAnimation(colorChangeDuration)
  return createRevertDimTextAnimation(self, colorChangeDuration)
end

function TypoTextView:getTimeoutAnimation(colorChangeDuration)
  return AHSpawn.new(
    createTimeoutDimTextAnimation(self, colorChangeDuration),
    createUnderlineErrorsAnimation(self, TypoCommon.COLOR_TEXT_ERROR_RED, colorChangeDuration)
  )
end

function TypoTextView:getRevertTimeoutAnimation(colorChangeDuration)
  return createRevertDimTextAnimation(self, colorChangeDuration)
end

function TypoTextView:getAnimation()
  return self.animation
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TypoTextView.new(textWidth, wordTouchedHandler)
  local self = TextView.new(textWidth)
  Mixin.mix(self, TypoTextView)

  self.wordTouchedHandler = wordTouchedHandler
  self.isTouchEnabled = false
  self.textColor = nil
  self.correctTextColor = nil
  self.cachedViews = {}
  self.animationBuilder = TypoTextAnimationBuilder.new()

  return self
end
