--- TypoTimeoutPanel
-- @author Created by anton on August 17, 2015.

class("TypoTimeoutPanel")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONTINUE_TEXT = "CONTINUE"
local TAP_ERRORS_PLURAL_TEXT = "TAP THE REMAINING<br/>ERRORS TO CORRECT THEM"
local TAP_ERRORS_SINGULAR_TEXT = "TAP THE REMAINING<br/>ERROR TO CORRECT IT"

local BUTTON_BACKGROUND_COLOR = ColorUtils.colorRGBToStd(57, 57, 57)

local TEXT_SIZE = 15

local LABEL_HEIGHT = FontScaling.getScaledSize(TEXT_SIZE) * 5

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getInitialText(errorsRemainingCount)
  if (errorsRemainingCount == 1) then
    return TAP_ERRORS_SINGULAR_TEXT
  elseif (errorsRemainingCount > 1) then
    return TAP_ERRORS_PLURAL_TEXT
  end
  error("getInitialText should not be called if there are no errors to fix.")
end

local function getTouchDownHandler(self)
  return function()
    self.buttonWasTapped = true
  end
end

local function createLabelBackground(self)
  local width = self:getWidth()
  local height = LABEL_HEIGHT
  local safeAreaOffset = Screen.safeBottom
  local backgroundHeight = LABEL_HEIGHT + safeAreaOffset

  local background = FilledRectangle.new(width, backgroundHeight, BUTTON_BACKGROUND_COLOR)
  background:setAnchorPointBottom()
  background:setLoc(self:getWidthPart(0.5), 0)

  return background
end

local function createContinueButton(self)
  local button = createLabelBackground(self)

  local label = Label.new(CONTINUE_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, TEXT_SIZE))

  local underlineThickness = TypoCommon.UNDERLINE_THICKNESS
  local continueUnderline = FilledRectangle.new(label:getWidth(), underlineThickness)
  local underlineYOffset = label:getHeightPart(-0.03)
  continueUnderline:setLoc(label:getWidthPart(0.5), underlineYOffset)
  label:insert(continueUnderline)
  label:linkColorAttrToChildren()

  label:setColor(TypoCommon.COLOR_TEXT_CONTINUE)

  button:insertAtCenter(label)

  local safeAreaOffset = Screen.safeBottom
  label:offsetY(safeAreaOffset)

  Button.makeButton(button, getTouchDownHandler(self))
  self:insert(button)
  return button
end

local function createFixErrorsLabel(self, errorsRemainingCount)
  local width = self:getWidth()
  local font = Font.new(FontFaces.DIN_OT_MEDIUM, TEXT_SIZE)

  local initialText = getInitialText(errorsRemainingCount)
  local label = DynamicTextArea.newSimple(initialText, width, font, Alignment.CENTER, Alignment.CENTER)

  local background = createLabelBackground(self)
  background:insertAtCenter(label)

  local safeAreaOffset = Screen.safeBottom
  label:offsetY(safeAreaOffset)

  self:insert(background)
  return background
end

local function playShowActiveLabel(self, duration)
  local label = self.activeLabel
  local labelHeight = label:getHeight()
  label:setY(label:getY() - labelHeight)
  return AHLocationYBy.new(label, duration, AHEaseExponentialOut.new(), labelHeight):start()
end

local function playHideActiveLabel(self, duration)
  local label = self.activeLabel
  local labelHeight = label:getHeight()
  return AHLocationYBy.new(label, duration, AHEaseExponentialOut.new(), -labelHeight):start()
end

local function swapToContinueLabel(self)
  local duration = 0.6
  wait(playHideActiveLabel(self, duration))
  self.activeLabel = createContinueButton(self)
  playShowActiveLabel(self, duration)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoTimeoutPanel:waitForTap(buttonSlideDuration)
  wait(playShowActiveLabel(self, buttonSlideDuration))
  waitUntil(function() return self.buttonWasTapped end)
  wait(playHideActiveLabel(self, buttonSlideDuration))
end

function TypoTimeoutPanel:allErrorsCorrected()
  swapToContinueLabel(self)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TypoTimeoutPanel.new(errorsRemainingCount)
  local self = Group.newFullScreen()
  Mixin.mix(self, TypoTimeoutPanel)

  if errorsRemainingCount == 0 then
    self.activeLabel = createContinueButton(self)
  else
    self.activeLabel = createFixErrorsLabel(self, errorsRemainingCount)
  end

  self.buttonWasTapped = false

  return self
end
