--- TypoValidator
-- @author Created by amorant on June 29, 2015.
-- ConceptValidator for Typo.

class("TypoValidator")

-- NOTE: TypoCommon contains a reference to Host.contentScale, but because
-- TypoCommon is required here in the validator where Host is undefined, it was
-- causing validators to fail. A dummy value has been inserted here to prevent
-- crashing, but beyond that, these two lines won't affect anything.
Host = {}
Host.contentScale = 4

require("validators.ConceptValidator")
require("validators.TagValidator")
require("utils.XmlParser")
require("TypoCommon")
require("TypoConceptGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "passageSpelling",
  "passagePunctuationCapitalization",
  "passageMissingExtra",
  "bonusText",
  "errors",
  "difficulty"
}

local ROUND_TYPES = {
  TypoCommon.ERROR_SPELLING,
  TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION,
  TypoCommon.ERROR_EXTRA_MISSING,
  TypoCommon.ERROR_BONUS
}

local CONCEPT_DIFFICULTIES = {
  "Basic",
  "Easy",
  "Hard"
}

local ERROR_DIFFICULTIES = {
  "Easy",
  "Hard"
}

local OPTIONAL_SINGLE_TAGS = { "br" }
local OPTIONAL_PAIRED_TAGS = { "i" }

local EXPECTED_ERROR_COUNTS = {
  [TypoCommon.ERROR_SPELLING] = { Easy = 4, Hard = 4 },
  [TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION] = { Easy = 4, Hard = 4 },
  [TypoCommon.ERROR_EXTRA_MISSING] = { Easy = 4, Hard = 4 },
  [TypoCommon.ERROR_BONUS] = { Easy = 0, Hard = 6 }
}

local EXPECTED_ERROR_COUNTS_FOR_BASIC_CONCEPTS = {
  [TypoCommon.ERROR_SPELLING] = { Easy = 4, Hard = 0 },
  [TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION] = { Easy = 4, Hard = 0 },
  [TypoCommon.ERROR_EXTRA_MISSING] = { Easy = 4, Hard = 0 },
  [TypoCommon.ERROR_BONUS] = { Easy = 6, Hard = 0 }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function isBasicDifficulty(difficulty)
  return (difficulty == "Basic")
end

local function getErrorNumbers(firstErrorNumber, secondErrorNumber)
  local errorNumbers = TableUtils.getNumberSequence(firstErrorNumber, secondErrorNumber)
  local errorNumbersAsStrings = TableUtils.map(errorNumbers, function(n) return tostring(n) end)
  return errorNumbersAsStrings
end

local function validateRawText(self, text, requiredTags)
  self:expectNoError(function()
    XmlParser.parse(text)
  end)
  self:validateTags(text, nil, requiredTags, nil, OPTIONAL_PAIRED_TAGS, OPTIONAL_SINGLE_TAGS)

  local doesTextStartWithTag = text:find("^<") ~= nil
  self:expectFalse(doesTextStartWithTag, "Text should not start with a tag so that we can do dropcaps")
end

local function validateErrorsForEasyOrHardConcept(self, errors)
  local errorsByNumber = {}

  local errorTypeCounts = {}
  for _, errorType in ipairs(ROUND_TYPES) do
    errorTypeCounts[errorType] = { Easy = 0, Hard = 0 }
  end

  for _, errorData in ipairs(errors) do
    local errorNumberAsString = errorData.error_number
    local errorNumber = tonumber(errorNumberAsString)
    self:expectFalse(errorsByNumber[errorNumberAsString], string.format("Two entries for error %s found", errorNumberAsString))
    errorsByNumber[errorNumberAsString] = errorData

    local difficulty = errorData.difficulty
    self:expectIsContained(difficulty, ERROR_DIFFICULTIES, "error difficulty")

    local errorType = errorData.round
    errorTypeCounts[errorType][difficulty] = errorTypeCounts[errorType][difficulty] + 1

    if errorNumber <= 8 then
      self:expectEqual(errorType, TypoCommon.ERROR_SPELLING)
    elseif errorNumber <= 16 then
      self:expectEqual(errorType, TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION)
    elseif errorNumber <= 24 then
      self:expectEqual(errorType, TypoCommon.ERROR_EXTRA_MISSING)
    elseif errorNumber <= 30 then
      self:expectEqual(errorType, TypoCommon.ERROR_BONUS)
    end
  end

  for errorType, errorCountsByDifficulty in pairs(errorTypeCounts) do
    for difficulty, errorCount in pairs(errorCountsByDifficulty) do
      local expectedErrorCount = EXPECTED_ERROR_COUNTS[errorType][difficulty]
      local errorMessage = string.format("Incorrect number of errors of type %s, difficulty %s", errorType, difficulty)
      self:expectEqual(errorCount, expectedErrorCount, errorMessage)
    end
  end
end

local function validateErrorsForBasicConcept(self, errors)
  local errorsByNumber = {}

  local errorTypeCounts = {}
  for _, errorType in ipairs(ROUND_TYPES) do
    errorTypeCounts[errorType] = { Easy = 0, Hard = 0 }
  end

  for _, errorData in ipairs(errors) do
    local errorNumberAsString = errorData.error_number
    local errorNumber = tonumber(errorNumberAsString)
    self:expectFalse(errorsByNumber[errorNumberAsString], string.format("Two entries for error %s found", errorNumberAsString))
    errorsByNumber[errorNumberAsString] = errorData

    local difficulty = errorData.difficulty
    self:expectIsContained(difficulty, ERROR_DIFFICULTIES, "error difficulty")

    local errorType = errorData.round
    errorTypeCounts[errorType][difficulty] = errorTypeCounts[errorType][difficulty] + 1

    if errorNumber <= 4 then
      self:expectEqual(errorType, TypoCommon.ERROR_SPELLING)
    elseif errorNumber <= 8 then
      self:expectEqual(errorType, TypoCommon.ERROR_PUNCTUATION_CAPITALIZATION)
    elseif errorNumber <= 12 then
      self:expectEqual(errorType, TypoCommon.ERROR_EXTRA_MISSING)
    elseif errorNumber <= 18 then
      self:expectEqual(errorType, TypoCommon.ERROR_BONUS)
    end
  end

  for errorType, errorCountsByDifficulty in pairs(errorTypeCounts) do
    for difficulty, errorCount in pairs(errorCountsByDifficulty) do
      local expectedErrorCount = EXPECTED_ERROR_COUNTS_FOR_BASIC_CONCEPTS[errorType][difficulty]
      local errorMessage = string.format("Incorrect number of errors of type %s, difficulty %s", errorType, difficulty)
      self:expectEqual(errorCount, expectedErrorCount, errorMessage)
    end
  end
end

local function validateErrors(self, difficulty, errors)
  if isBasicDifficulty(difficulty) then
    validateErrorsForBasicConcept(self, errors)
  else
    validateErrorsForEasyOrHardConcept(self, errors)
  end
end

local function validateConceptDifficulty(self, difficulty)
  self:expectIsContained(difficulty, CONCEPT_DIFFICULTIES, "difficulty")
end

local function validateConceptCanBeProcessed(self, concept)
  -- This method throws an error if something is amiss. In particular, it processes the passages
  -- and verifies that each error has different correct/incorrect text.
  self:expectNoError(function()
    TypoConceptGenerator.processConcept(concept)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function TypoValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)

  local passageSpelling = concept.passageSpelling
  local passagePunctuationCapitalization = concept.passagePunctuationCapitalization
  local passageMissingExtra = concept.passageMissingExtra
  local bonusText = concept.bonusText
  local errors = concept.errors
  local difficulty = concept.difficulty

  validateConceptDifficulty(self, difficulty)
  if isBasicDifficulty(difficulty) then
    validateRawText(self, passageSpelling, getErrorNumbers(1, 4))
    validateRawText(self, passagePunctuationCapitalization, getErrorNumbers(5, 8))
    validateRawText(self, passageMissingExtra, getErrorNumbers(9, 12))
    validateRawText(self, bonusText, getErrorNumbers(13, 18))
  else
    validateRawText(self, passageSpelling, getErrorNumbers(1, 8))
    validateRawText(self, passagePunctuationCapitalization, getErrorNumbers(9, 16))
    validateRawText(self, passageMissingExtra, getErrorNumbers(17, 24))
    validateRawText(self, bonusText, getErrorNumbers(25, 30))
  end
  validateErrors(self, difficulty, errors)
  validateConceptCanBeProcessed(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function TypoValidator.new()
  local self = ConceptValidator.new()
  TagValidator.make(self)
  Mixin.mix(self, TypoValidator)

  return self
end
