--- UpperAboutGameInterstitial
-- @author Created by Steven Mah on February 2, 2016.

class("UpperAboutGameInterstitial")

require("UpperCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local SECOND_HEADER_FONT = Font.new(FontFaces.DIN_OT_BOLD, 17)
local BODY_FONT = Font.new(FontFaces.DIN_OT_MEDIUM, 16)

local HEADER_COLOR = ColorUtils.colorHex24ToStd(0x37EDFD)
local BACKGROUND_START_COLOR = ColorUtils.colorHex32ToStd(0x104E7F00)
local BACKGROUND_END_COLOR = ColorUtils.colorHex32ToStd(0x17798B00)
local BODY_TEXT_COLOR = {1.0, 1.0, 1.0}

local HEADER_TEXT = Localizer.get("ABOUT_TITLE")
local SECOND_HEADER_TEXT = Localizer.get("SUBVOCALIZATION")

local BODY_WIDTH = Screen.getSafeWidthPart(0.85)
local OKAY_BUTTON_WIDTH = Screen.getSafeWidthPart(0.8)
local OKAY_BUTTON_HEIGHT = Screen.getSafeHeightPart(0.075)
local COLUMN_SPACING = Screen.getSafeHeightPart(0.025)

local FADE_IN_DURATION = 1.5
local PAGE_SLIDE_DURATION = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeButton(group, callback)
  local onPressReturn = function()
    group:setOpacity(0.2)
  end

  local onCancelReturn = function()
    AHOpacityTo.new(group, 0.2, nil, 1.0):start()
  end

  local onReleaseReturn = function()
    AHOpacityTo.new(group, 0.2, nil, 1.0):start()
    Sound.playSoundEffect("button_click")
    group:setTouchEnabled(false)
    callback()
  end

  Button.addVirtualButton(group, group:getWidthPart(2.0), group:getHeightPart(2.0), onPressReturn, onReleaseReturn, onCancelReturn)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function UpperAboutGameInterstitial:enableButtons()
  self.okayButton:setTouchEnabled(true)
  self.moreInfoButton:setTouchEnabled(true)
end

function UpperAboutGameInterstitial:createFadeInAction()
  local pageOne = self.pageOne
  local okayButton = self.okayButton
  local background = self.background

  local okayButtonY = self:getHeightPart(0.08)

  local bodyAndButtonAction = AHSpawn.new(
    self.header:createAnimateInAction(FADE_IN_DURATION),
    AHScaleTo.new(pageOne, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0),
    AHOpacityTo.new(self.header, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0),
    AHOpacityTo.new(self.body, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0),
    AHOpacityTo.new(self.moreInfoButton, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0),
    AHLocationYTo.new(okayButton, FADE_IN_DURATION, AHEaseExponentialOut.new(), okayButtonY),
    AHScaleTo.new(okayButton, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0)
  )

  local staggeredRevealAction = AHStaggeredSpawn.new(
    0.25,
    background:newOpacityToAction(0.5, AHEaseExponentialOut.new(), 1.0, 1.0),
    AHSpawn.new(
      AHScaleTo.new(self.bubbleLayer, FADE_IN_DURATION, AHEaseExponentialOut.new(), 1.0),
      self.bubbleLayer:createShowBubblesAction(0.05, 0.2, 1.0)
    ),
    bodyAndButtonAction,
    AHDelayed.new(0.25, function() self:enableButtons() end)
  )

  return staggeredRevealAction
end

function UpperAboutGameInterstitial:createFadeOutAction()
  local fadeOutAction = AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(self.circleOverlay, 0.4, AHEaseSineOut.new(), 1.0),
      AHOpacityTo.new(self.okayButton, 0.2, nil, 0.0)
    ),
    function() wait(self.gameBeginScreen:resetInterstitial()) end
  )
  return fadeOutAction
end

-- ----------------------------------------------------------------------------
-- Object creation
-- ----------------------------------------------------------------------------

local function createLayout(self, loadWithBackgroundVisible)
  local backgroundStartColor = BACKGROUND_START_COLOR
  local backgroundEndColor = BACKGROUND_END_COLOR
  if loadWithBackgroundVisible then
    backgroundStartColor = ColorUtils.colorWithOpacity(backgroundStartColor, 1.0)
    backgroundEndColor = ColorUtils.colorWithOpacity(backgroundEndColor, 1.0)
  end
  local background = RectangleGradient.new(Screen.fullWidth, Screen.fullHeight, backgroundStartColor, backgroundEndColor)
  self.background = background
  self:insertAtCenter(background)

  local isSimple = true
  local bubbleLayer = UpperBubbleLayer.new(isSimple, 12, Screen.getHeightPart(0.05), Screen.getHeightPart(0.1))
  bubbleLayer:hideBubbles()
  bubbleLayer:setScrollSpeed(0.3, 0.1)
  bubbleLayer:setBubblesToWander()
  bubbleLayer:setScl(2.0)
  self:insertAtCenter(bubbleLayer)
  self.bubbleLayer = bubbleLayer

  local pageOne = Column.new(COLUMN_SPACING, Alignment.LEFT)
  local pageTwo = Column.new(COLUMN_SPACING, Alignment.LEFT)

  local firstHeaderFont = PrerenderedFont.new(FontFaces.DIN_OT_LIGHT, 40)
  local firstHeaderTextArea = UpperDynamicTextArea.newSimple(HEADER_TEXT, BODY_WIDTH * 0.65, firstHeaderFont, Alignment.LEFT, HEADER_COLOR)
  firstHeaderTextArea:setColor(unpack(HEADER_COLOR))
  firstHeaderTextArea:setOpacity(0.0)

  self.header = firstHeaderTextArea
  pageOne:insertAtCenter(firstHeaderTextArea)

  local thisGameText = string.format("%s %s<br/><br/>%s",
                                     Localizer.get("THIS_GAME_1"),
                                     Localizer.get("THIS_GAME_2"),
                                     Localizer.get("THIS_GAME_3"))
  local firstBodyText = DynamicTextArea.newSimple(thisGameText, BODY_WIDTH, BODY_FONT, Alignment.LEFT)
  firstBodyText:setOpacity(0.0)

  self.body = firstBodyText
  pageOne:insertAtCenter(firstBodyText)

  local moreInfoLabel = Label.new(Localizer.get("MORE_INFO"), Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  moreInfoLabel:setColor(unpack(HEADER_COLOR))
  moreInfoLabel:setAnchorPointLeft()

  local moreArrow = Sprite.newWithName("arrow.png")
  moreArrow:setAnchorPointRight()
  moreArrow:setColor(unpack(HEADER_COLOR))

  local moreInfoGroup = Group.new(moreInfoLabel:getWidthPart(1.1) + moreArrow:getWidth(), moreInfoLabel:getHeightPart(1.5))
  moreInfoLabel:setLoc(0.0, moreInfoGroup:getHeightPart(0.5))
  moreInfoGroup:insert(moreInfoLabel)
  moreArrow:setLoc(moreInfoGroup:getWidthPart(1.0), moreInfoGroup:getHeightPart(0.5))
  moreInfoGroup:insert(moreArrow)

  local isRecursive = true
  moreInfoGroup:setOpacity(0.0)
  moreInfoGroup:linkOpacityAttrToChildren(isRecursive)

  local moreInfoCallBack = function()
    local moreInfoAction = AHSequence.new(
      AHSpawn.new(
        AHLocationXTo.new(self.pageOne, PAGE_SLIDE_DURATION, AHEaseExponentialOut.new(), -Screen.getWidthPart(0.5)),
        AHLocationXTo.new(self.pageTwo, PAGE_SLIDE_DURATION, AHEaseExponentialOut.new(), Screen.getSafeX(0.5))
      ),
      function() self.returnButton:setTouchEnabled(true) end
    )

    moreInfoAction:start()
  end

  makeButton(moreInfoGroup, moreInfoCallBack)
  moreInfoGroup:setTouchEnabled(false)
  self.moreInfoButton = moreInfoGroup
  pageOne:insertAtCenter(moreInfoGroup)
  pageOne:alignChildren()
  firstHeaderTextArea:scrambleAndHideText()

  pageOne:setScl(2.0)
  self.pageOne = pageOne
  self:insert(pageOne)
  pageOne:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.55))

  local returnLabel = Label.new(Localizer.get("RETURN"), Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  returnLabel:setColor(unpack(HEADER_COLOR))
  returnLabel:setAnchorPointRight()

  local returnArrow = Sprite.newWithName("arrow.png")
  returnArrow:setAnchorPointRight()
  returnArrow:setColor(unpack(HEADER_COLOR))
  returnArrow:setRot(180)

  local returnGroup = Group.new(returnLabel:getWidthPart(1.1) + returnArrow:getWidth(), returnLabel:getHeightPart(1.5))
  returnLabel:setLoc(returnGroup:getWidthPart(1.0), returnGroup:getHeightPart(0.5))
  returnGroup:insert(returnLabel)
  returnArrow:setLoc(0.0, returnGroup:getHeightPart(0.5))
  returnGroup:insert(returnArrow)

  returnGroup:linkOpacityAttrToChildren(isRecursive)

  local returnCallback = function()
    local returnAction = AHSequence.new(
      AHSpawn.new(
        AHLocationXBy.new(self.pageTwo, PAGE_SLIDE_DURATION, AHEaseExponentialOut.new(), Screen.fullWidth),
        AHLocationXBy.new(self.pageOne, PAGE_SLIDE_DURATION, AHEaseExponentialOut.new(), Screen.fullWidth)
      ),
      function() self.moreInfoButton:setTouchEnabled(true) end
    )

    returnAction:start()
  end

  makeButton(returnGroup, returnCallback)
  returnGroup:setTouchEnabled(false)
  self.returnButton = returnGroup
  pageTwo:insertAtCenter(returnGroup)

  local secondHeaderTextArea = Label.new(SECOND_HEADER_TEXT, SECOND_HEADER_FONT, BODY_WIDTH, Alignment.LEFT)
  secondHeaderTextArea:setColor(unpack(HEADER_COLOR))
  pageTwo:insertAtCenter(secondHeaderTextArea)

  local subvocalizationText = string.format("%s<br/><br/>%s<br/><br/>%s<br/><br/>%s",
                                            Localizer.get("SUBVOCALIZATION_SUMMARY_1"),
                                            Localizer.get("SUBVOCALIZATION_SUMMARY_2"),
                                            Localizer.get("SUBVOCALIZATION_SUMMARY_3"),
                                            Localizer.get("SUBVOCALIZATION_SUMMARY_4"))
  local secondBodyText = DynamicTextArea.newSimple(subvocalizationText, BODY_WIDTH, BODY_FONT, Alignment.LEFT)
  pageTwo:insertAtCenter(secondBodyText)

  pageTwo:alignChildren()

  pageTwo:setLoc(Screen.getWidthPart(1.5), Screen.getSafeY(0.575))
  self:insert(pageTwo)
  self.pageTwo = pageTwo

  local circleOverlay = FilledCircle.new(Screen.safeHeight, {1.0, 1.0, 1.0})
  circleOverlay:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.1))
  circleOverlay:setScl(0.0)
  self.circleOverlay = circleOverlay
  self:insert(circleOverlay)

  local okayLabel = Label.new(Localizer.get("OKAY"), Font.new(FontFaces.DIN_OT_MEDIUM, 18))
  okayLabel:setColor(unpack(BACKGROUND_START_COLOR))
  local okayBox = FilledRoundedRectangle.new(OKAY_BUTTON_WIDTH, OKAY_BUTTON_HEIGHT, 3)
  okayBox:insertAtCenter(okayLabel)
  okayBox:linkOpacityAttrToChildren(isRecursive)

  local onPressContinue = function()
    Sound.playSoundEffect("button_click")
    self.okayButton:setTouchEnabled(false)
    wait(self:createFadeOutAction():start())
    self:removeSelf()
    self.gameBeginScreen:createFadeInAction():start()
  end

  local okayButton = Button.makeButton(okayBox, onPressContinue)
  okayButton:setTouchEnabled(false)
  okayButton:setLoc(self:getWidthPart(0.5), -self:getHeightPart(0.1))
  okayButton:setScl(1.25)
  self:insert(okayButton)
  self.okayButton = okayButton
end

-- Constructor ------------------------

function UpperAboutGameInterstitial.new(gameBeginScreen, loadWithBackgroundVisible)
  local self = Group.newFullScreen()
  Mixin.mix(self, UpperAboutGameInterstitial)

  self.gameBeginScreen = gameBeginScreen
  createLayout(self, loadWithBackgroundVisible)

  return self
end
