--- UpperBubbleLayer
-- @author Created by steven on February 17, 2016.

class("UpperBubbleLayer")

require("UpperBubble")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local OFF_SCREEN_THRESHOLD_Y = Screen.getHeightPart(1.3)
local OFF_SCREEN_THRESHOLD_X = Screen.getWidthPart(1.3)

local MIN_X_VELOCITY = -Screen.getWidthPart(0.001)
local MAX_X_VELOCITY = Screen.getWidthPart(0.001)
local MIN_Y_VELOCITY = -Screen.getHeightPart(0.0005)
local MAX_Y_VELOCITY = -Screen.getHeightPart(0.0015)

local COLORS = {
  [UpperCommon.COLOR_STATES.NEUTRAL] = {
    {0.972, 0.792, 0.941},
    {0.996, 0.756, 0.792}
  },
  [UpperCommon.COLOR_STATES.CORRECT] = {
    {0.0, 0.698, 0.913},
    {0.960, 0.827, 0.976}
  },
  [UpperCommon.COLOR_STATES.OUTRO] = {
    {0.0, 0.698, 0.913},
    {0.776, 0.819, 0.992},
    {0.972, 0.792, 0.941}
  }
}

local SCROLL_SPEED_CHANGE_GRANULARITY = 0.1 -- update speed once every 0.1 seconds

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function makeRandomVelocity(self)
  local randomXVelocity = MathUtils.getRandomFloat(MIN_X_VELOCITY, MAX_X_VELOCITY)
  local randomYVelocity = MathUtils.getRandomFloat(MIN_Y_VELOCITY, MAX_Y_VELOCITY)
  local velocity = {x = randomXVelocity, y = randomYVelocity}
  return velocity
end

local function makeRandomBubble(self, isSimple)
  local randomX = MathUtils.getRandomFloat(0, Screen.fullWidth)
  local randomY = MathUtils.getRandomFloat(0, Screen.fullHeight)
  local randomRadius = MathUtils.getRandomFloat(self.bubbleMinRadius, self.bubbleMaxRadius)

  local isBlurred = MathUtils.getRandomBoolean()
  local bubble = nil
  if (isSimple == true) then
    bubble = UpperBubble.newSimple(randomRadius, isBlurred)
  else
    bubble = UpperBubble.newComplex(randomRadius)
  end

  local velocity = makeRandomVelocity(self)
  bubble.velocity = velocity
  bubble:setLoc(randomX, randomY)

  return bubble
end

local function isBubbleOffScreen(bubble)
  local x, y = bubble:getLoc()
  local width, height = bubble:getDims()
  local offScreenX = Screen.fullWidth + width
  local offScreenY = Screen.fullHeight + height
  local outOfBoundsY = ((y < -width) or (y > offScreenY))
  local outOfBoundsX = ((x < -height) or (x > offScreenX))

  return (outOfBoundsY or outOfBoundsX)
end

local function generateBubbles(self, isSimple, numberOfBubbles)
  local numberOfBubbles = numberOfBubbles or self.maxBubbles
  for bubbleIndex = 1, self.maxBubbles do
    local bubble = makeRandomBubble(self, isSimple)
    self:insert(bubble)
    table.insert(self.bubbles, bubble)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function UpperBubbleLayer:setScrollSpeed(speedX, speedY)
  local speedY = (speedY or speedX)
  self.scrollSpeedX = speedX
  self.scrollSpeedY = speedY
end

function UpperBubbleLayer:createShowBubblesAction(minOpacity, maxOpacity, duration)
  local duration = duration or 0.1
  local fullShowAction = AHSpawn.new()
  for _, bubble in ipairs(self.bubbles) do
    local minOpacity = (minOpacity or 0.1)
    local maxOpacity = (maxOpacity or 0.2)
    local randomOpacity = MathUtils.getRandomFloat(minOpacity, maxOpacity)
    local bubbleOpacityAction = AHOpacityTo.new(bubble.colorLayer, duration, nil, randomOpacity)
    if (bubble.transparentLayer ~= nil) then
      fullShowAction:addAction(
        AHOpacityTo.new(bubble.transparentLayer, duration, nil, 1.0)
      )
    end
    fullShowAction:addAction(bubbleOpacityAction)
  end
  return fullShowAction
end

function UpperBubbleLayer:createHideBubblesAction()
  local fullHideAction = AHSpawn.new()
  for _, bubble in ipairs(self.bubbles) do
    local bubbleOpacityAction = AHOpacityTo.new(bubble.colorLayer, 0.5, nil, 0.0)
    if (bubble.transparentLayer ~= nil) then
      fullHideAction:addAction(
        AHOpacityTo.new(bubble.transparentLayer, 0.1, nil, 0.0)
      )
    end
    fullHideAction:addAction(bubbleOpacityAction)
  end
  return fullHideAction
end

function UpperBubbleLayer:hideBubbles()
  local fullShowAction = AHSpawn.new()
  for _, bubble in ipairs(self.bubbles) do
    bubble.colorLayer:setOpacity(0.0)
  end
end

function UpperBubbleLayer:setBubblesToWander()
  local bubbles = self.bubbles
  for _, bubble in ipairs(self.bubbles) do
    local oldVelocity = bubble.velocity
    local newVelocity = {x = MathUtils.randomSign() * oldVelocity.x, y = MathUtils.randomSign() * oldVelocity.y}
    bubble.velocity = newVelocity
  end
end

function UpperBubbleLayer:createChangeColorsForStateAction(state, duration)
  local duration = duration or 0.1
  local newColors = COLORS[state]
  local colorIndex = 1
  local allBubblesColorAction = AHSpawn.new()

  for _, bubble in ipairs(self.bubbles) do
    local newColor = newColors[colorIndex]
    colorIndex = (colorIndex % #newColors) + 1

    allBubblesColorAction:addAction(
      AHColorTo.new(bubble.colorLayer, duration, nil, newColor)
    )
  end

  self.currentMode = mode
  return allBubblesColorAction
end

function UpperBubbleLayer:changeColorsForState(state, duration)
  local allBubblesColorAction = self:createChangeColorsForStateAction(state, duration)

  allBubblesColorAction:start()
end

function UpperBubbleLayer:disableBubbleReset()
  self.shouldResetBubbles = false
end

function UpperBubbleLayer:reverseYVelocity()
  for _, bubble in ipairs(self.bubbles) do
    bubble.velocity.y = -bubble.velocity.y
  end
end

function UpperBubbleLayer:graduallySetScrollSpeedY(duration, targetSpeed)
  local numberOfTimeSteps = duration / SCROLL_SPEED_CHANGE_GRANULARITY
  local speedDeltaPerTimeStep = (targetSpeed - self.scrollSpeedY) / numberOfTimeSteps

  local scrollSpeedTo = function()
    for timeStepIndex = 1, numberOfTimeSteps do
      waitFor(SCROLL_SPEED_CHANGE_GRANULARITY)
      self.scrollSpeedY = self.scrollSpeedY + speedDeltaPerTimeStep
    end
  end

  runInCoroutine(scrollSpeedTo)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

function UpperBubbleLayer.new(isSimple, numberOfBubbles, minRadius, maxRadius)
  local self = Group.newFullScreen()
  Mixin.mix(self, UpperBubbleLayer)

  self.maxBubbles = numberOfBubbles
  self.bubbleMinRadius = minRadius
  self.bubbleMaxRadius = maxRadius

  self.bubbles = {}
  self.scrollSpeedX = 0.1
  self.scrollSpeedY = 0.1

  self.shouldResetBubbles = true

  generateBubbles(self, isSimple)
  self:createHideBubblesAction():start()

  local updateFunction = function()
    while true do
      for _, bubble in ipairs(self.bubbles) do
        local oldX, oldY = bubble:getLoc()
        local deltaX = bubble.velocity.x * self.scrollSpeedX
        local deltaY = bubble.velocity.y * self.scrollSpeedY
        local newX = oldX + deltaX
        local newY = oldY + deltaY
        bubble:setLoc(newX, newY)
        if (isBubbleOffScreen(bubble) and self.shouldResetBubbles) then
          local bubbleHeight = bubble:getHeight()
          local randomX = MathUtils.getRandomFloat(0, Screen.fullWidth)
          newY = Screen.fullHeight + bubbleHeight
          if deltaY > 0 then
            newY = -bubbleHeight
          end
          bubble:setLoc(randomX, newY)
        end
      end
      waitForFrames(1)
    end
  end

  self.updateCoroutine = runInCoroutine(updateFunction)

  return self
end
