--- UpperGame
-- @author Created by steven on February 01, 2016.

class("UpperGame")

require("StandardUIBuilder")
require("UpperConfiguration")
require("UpperCommon")
require("UpperBrick")
require("UpperWordBrick")
require("UpperResponseBrick")
require("UpperTower")
require("UpperGameBeginInterstitial")
require("UpperRoundStartInterstitial")
require("UpperRoundGenerator")
require("UpperBubbleLayer")
require("UpperRationale")
require("UpperProgressIndicator")
require("bonuses.ResponseSpeedTracker")
require("UpperBackground")
require("UpperOutroOverlay")
require("UpperSpecksLayer")
require("UpperGradientBrick")
require("UpperPieTimer")
require("graphics.MultiGradient")
require("tutorials.TutorialHighlighter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = UpperCommon.NEUTRAL_PURPLE
local UI_CORRECT_COLOR = {0.31, 0.68, 0.82}
local UI_INCORRECT_COLOR = {0.92, 0.25, 0.58}
local UI_VERTICAL_PADDING = 10

local QUESTION_AREA_COLOR = UpperCommon.NEUTRAL_PURPLE

local QUESTION_WIDTH = UpperCommon.RESPONSE_BRICK_WIDTH
local QUESTION_VERTICAL_PADDING = Screen.getSafeHeightPart(0.04)
local QUESTION_SHADOW_PADDING = 75

local Z_BACKGROUND = 0
local Z_BACK_BUBBLE_LAYER = 100
local Z_TOWER = 200
local Z_FORE_BUBBLE_LAYER = 300
local Z_SHADOW_FOR_QUESTION = 350
local Z_ABOVE_BUBBLES = 400
local Z_WORD_BRICK = 500
local Z_WIN_ANIMATION_OVERLAY = 600
local Z_COUNTDOWN_INTERSTITIAL = 800
local Z_RATIONALE = 900

local POST_WIN_ANIMATION_DELAY = 1.0

local TUTORIAL_FONT_SIZE = 20

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function createHideUiAction(self, duration)
  local duration = duration or 1.0
  local hideUiAction = AHSpawn.new(
    AHOpacityTo.new(self.uiOverlay, duration, nil, 0.0),
    AHOpacityTo.new(self.progressIndicator, duration, nil, 0.0)
    )
  return hideUiAction
end

local function winGame(self)
  waitFor(POST_WIN_ANIMATION_DELAY)
  GameSession.notifyGameWin()
end

local function waitForWinGameAnimation(self)
  local tower = self.tower
  local wordBrick = tower.wordBrick

  local fadeOutElementsAction = AHSpawn.new(
    AHOpacityTo.new(self.questionTextArea, 0.1, nil, 0.0),
    AHOpacityTo.new(self.correctBrick:getLabel(), 0.1, nil, 0.0)
  )

  local checkmark = Sprite.newWithName("checkmark.png")
  checkmark:setLoc(Screen.getSafeX(0.5), Screen.getHeightPart(1.5))
  self:insert(checkmark, Z_ABOVE_BUBBLES)

  local whiteOverlay = UpperOutroOverlay.new()
  self:insertAtCenter(whiteOverlay, Z_WIN_ANIMATION_OVERLAY)
  local overlayPulse = whiteOverlay:createFlashAction(UpperTower.OUTRO_PAN_DURATION)

  local createTowerColorAction = function()
    local towerColorAction = tower:createOutroColorAction(self.background)
    towerColorAction:start()
  end

  local winAction = AHSpawn.new(
    fadeOutElementsAction,
    AHSequence.new(
      tower:createGameWinPanAction(),
      function() self.foreBubbleLayer:setScrollSpeed(1.0) end,
      createTowerColorAction,
      AHLocationYTo.new(checkmark, 1.5, AHEaseExponentialOut.new(), Screen.getSafeY(0.75)),
      self.foreBubbleLayer:disableBubbleReset(),
      self.backBubbleLayer:disableBubbleReset()
    ),
    AHDelayed.new(1.0, createHideUiAction(self, 0.3)),
    self.background:createToOutroAction(0.1),
    self.foreBubbleLayer:changeColorsForState(UpperCommon.COLOR_STATES.OUTRO)
  )

  self.foreBubbleLayer:setZIndex(Z_FORE_BUBBLE_LAYER)
  self.backBubbleLayer:createShowBubblesAction():start()
  self.foreBubbleLayer:createShowBubblesAction(1.0, 1.0):start()
  self.backBubbleLayer:setScrollSpeed(0.5)
  self.foreBubbleLayer:setScrollSpeed(3.0)

  self:insertAtCenter(self.specksLayer)
  self.specksLayer:fadeInSpecks()

  overlayPulse:start()
  wait(winAction:start())
end

local function registerCorrectResponse(self)
  GameSession.incrementCorrectAnswerCount()
  self.responseSpeedTracker:notifyCorrectResponse()

  self.currentWordsPerMinute = self.currentWordsPerMinute + UpperConfiguration.WORDS_PER_MINUTE_INCREMENT
  self.currentCountdownDuration = self.currentCountdownDuration + self.countdownDecrement
  GameSession.increaseScore(UpperConfiguration.getCorrectAnswerScore(GameSession.getCorrectAnswerCount()))
  self.scoreLabel:setScore(GameSession.getScore())
end

local function checkAndWaitForGameEnd(self)
  if (self.progressIndicator:getProgress() >= 1 and GameSession.getScore() >= GameSession.getPassScore()) then
    ScoringUtils.applyLivesBonus(self.livesCounter:getLives(), UpperConfiguration.BONUS_SCORE_PER_EXTRA_LIFE, self.bonusTray)
    self.scoreLabel:setScore(GameSession.getScore())

    waitForWinGameAnimation(self)
    winGame(self)
  elseif self.livesCounter:getLives() <= 0 then
    GameSession.notifyGameLoss()
  end
end

local function registerIncorrectResponse(self)
  GameSession.incrementIncorrectAnswerCount()
  self.responseSpeedTracker:notifyIncorrectResponse()
  Sound.playSoundEffect("incorrect_answer")
  self.livesCounter:decrement()
  local progressIndicator = self.progressIndicator

  progressIndicator:displayIncorrect()
end

local function calculateDelayForWpm(self)
  local currentWpm = self.currentWordsPerMinute
  local secondsPerMinute = 60
  local actualDelay = secondsPerMinute/currentWpm
  return actualDelay
end

local function showUI(self)
  local showUiAction = AHSpawn.new(
    AHOpacityTo.new(self.uiOverlay, 1.0, nil, 1.0),
    AHOpacityTo.new(self.progressIndicator, 1.0, nil, 1.0)
    )
  showUiAction:start()
end

local function createUiColorAction(self, color)
  local pauseButton = self.pauseButton
  local scoreLabel = self.scoreLabel
  local livesCounter = self.livesCounter
  local bonusTray = self.bonusTray
  local fullColorAction = AHSpawn.new(
    AHColorTo.new(pauseButton, 0.1, nil, color),
    AHColorTo.new(scoreLabel, 0.1, nil, color),
    AHColorTo.new(livesCounter, 0.1, nil, color),
    AHColorTo.new(bonusTray, 0.1, nil, color)
  )

  return fullColorAction
end

local function uiColorReset(self)
  createUiColorAction(self, UI_COMPONENTS_COLOR):start()
end

local function getResetBackgroundColorAction(self)
  local background = self.background
  local backgroundReset = AHSpawn.new(
    background:createToNeutralAction(0.5),
    function() uiColorReset(self) end
    )
  return backgroundReset
end

local function buildWordLabelsForRound(self, round)
  local wordBrick = self.wordBrick

  wait(wordBrick:removeAllLabels())

  local words = StringUtils.split(round.snippet, " ")
  local labelsForRound = {}

  for wordIndex, word in ipairs(words) do
    local label = Label.new(word, UpperCommon.STANDARD_BRICK_FONT, nil, Alignment.CENTER)
    label:setColor(unpack(UpperCommon.BRICK_TEXT_COLOR))
    label:setOpacity(0.0)
    table.insert(labelsForRound, label)

    wordBrick:insertAtCenter(label, 20)
  end

  wordBrick:setWordLabelsForRound(labelsForRound)
end

local function buildQuestionAndResponsesForRound(self, round)
  local questionTextArea = DynamicTextArea.newSimple(round.question,
                                                      QUESTION_WIDTH,
                                                      UpperCommon.QUESTION_FONT,
                                                      Alignment.CENTER)
  questionTextArea:setColor(unpack(QUESTION_AREA_COLOR))
  self.questionTextArea = questionTextArea

  local correctBrick = UpperResponseBrick.new(round.correctResponse, true)
  correctBrick:setVisibleRecursive(false)
  local incorrectBrick = UpperResponseBrick.new(round.incorrectResponse, false)
  incorrectBrick:setVisibleRecursive(false)

  local function correctResponseBrickSelected()
    self.isSelectedBrickCorrect = correctBrick.isCorrect
    self.isAnswerSubmitted = true
  end

  local function incorrectResponseBrickSelected()
    self.isSelectedBrickCorrect = incorrectBrick.isCorrect
    self.isAnswerSubmitted = true
  end

  local correctButton = Button.makeButton(correctBrick, correctResponseBrickSelected)
  correctButton:setTouchEnabled(false)
  local incorrectButton = Button.makeButton(incorrectBrick, incorrectResponseBrickSelected)
  incorrectButton:setTouchEnabled(false)

  self.correctButtonCallback = correctResponseBrickSelected

  --- TODO: Update "correctSide" logic from left-right to top-bottom once content
  -- is ready for it.
  local randomSide = MathUtils.randomSign()
  local randomChooseTop = ((round.correctSide == "random") and (randomSide == -1))
  local randomChooseBottom = ((round.correctSide == "random") and (randomSide == 1))

  local topBrick = nil
  local bottomBrick = nil

  if ((round.correctSide == "left") or (randomChooseTop)) then
    topBrick = correctButton
    bottomBrick = incorrectButton
  elseif ((round.correctSide == "right") or (randomChooseBottom)) then
    topBrick = incorrectButton
    bottomBrick = correctButton
  end

  self.correctBrick = correctButton

  bottomBrick:setLoc(Screen.getSafeX(0.5), Screen.getHeightPart(1.1))
  topBrick:setLoc(bottomBrick:getX(), bottomBrick:getY() + UpperCommon.RESPONSE_BRICK_HEIGHT + UpperCommon.BRICK_PADDING)
  local questionTextAreaSpacing = ((questionTextArea:getHeight() / 2) + QUESTION_VERTICAL_PADDING)
  questionTextArea:setLoc(topBrick:getX(), topBrick:getYTop() + questionTextAreaSpacing + UpperCommon.BRICK_PADDING)

  self.topBrick = topBrick
  self.bottomBrick = bottomBrick
  self.questionTextArea = questionTextArea

  self.zoomGroup:insert(bottomBrick, Z_WORD_BRICK)
  self.zoomGroup:insert(topBrick, Z_WORD_BRICK)
  self.zoomGroup:insert(questionTextArea, Z_WORD_BRICK)
end

local function buildRationaleForRound(self, round)
  local rationale = UpperRationale.newHidden(round)
  rationale:setTouchEnabled(false)
  self:insertAtCenter(rationale, Z_RATIONALE)

  self.rationaleForRound = rationale
end

local function loadBeginGameInterstitial(self)
  local beginGameInterstitial = UpperGameBeginInterstitial.new(self.currentWordsPerMinute)
  self:insertAtCenter(beginGameInterstitial, Z_COUNTDOWN_INTERSTITIAL)
  beginGameInterstitial:preloadFirstAboutScreenIfNeeded()
  self.beginGameInterstitial = beginGameInterstitial
end

local function waitForBeginGameInterstitial(self, round)
  local beginGameInterstitial = self.beginGameInterstitial
  beginGameInterstitial.gameCanBegin = false

  self.isAnswerSubmitted = false
  self.isSelectedBrickCorrect = nil
  local interstitialFadeIn = beginGameInterstitial:createFadeInAction()
  buildWordLabelsForRound(self, round)
  buildQuestionAndResponsesForRound(self, round)
  interstitialFadeIn:start()

  self.foreBubbleLayer:changeColorsForState(UpperCommon.COLOR_STATES.NEUTRAL)

  self.wordBrick:fadeInWordBrick()
  self.backBubbleLayer:createShowBubblesAction():start()
  self.foreBubbleLayer:createShowBubblesAction():start()

  getResetBackgroundColorAction(self):start()

  waitUntil(function() return beginGameInterstitial.gameCanBegin end)

  self.tower:setVisibility(true)

  local interstitialFadeOut = AHStaggeredSpawn.new(
    0.75,
    AHSpawn.new(
      beginGameInterstitial:createFadeOutToGameAction(),
      AHScaleTo.new(self.zoomGroup, 2.0, AHEaseExponentialOut.new(), 1.0)
    ),
    function() Sound.playSoundEffect("countdown") end
  )

  wait(interstitialFadeOut:start())
end

local function waitForRoundStart(self, roundIndex, round)
  local correctBrick = self.correctBrick
  local zoomGroup = self.zoomGroup

  local answeredCorrectly = (self.isAnswerSubmitted and self.isSelectedBrickCorrect)

  local roundStartIntersitial = UpperRoundStartInterstitial.new(roundIndex, self.currentWordsPerMinute, answeredCorrectly)
  zoomGroup:insertAtCenter(roundStartIntersitial, Z_COUNTDOWN_INTERSTITIAL)

  local zoomOutAction = AHDoNothing.new()
  local actionStaggerDuration = 0.05

  if answeredCorrectly then
    actionStaggerDuration = 0.25

    zoomOutAction = AHStaggeredSpawn.new(
      0.1,
      AHScaleTo.new(zoomGroup, 1.0,  AHEaseExponentialOut.new(), 0.75),
      AHSequence.new(
        AHOpacityTo.new(correctBrick:getLabel(), 0.1, nil, 0.0),
        AHOpacityTo.new(self.questionTextArea, 0.1, nil, 0.0),
        AHRemove.new(self.questionTextArea)
      )
    )
  end

  local roundEndAction = AHSpawn.new(
    zoomOutAction,
    AHDelayed.new(
      actionStaggerDuration,
      function() Sound.playSoundEffect("round_transition") end,
      roundStartIntersitial:createFadeInAction(),
      AHOpacityTo.new(correctBrick:getLabel(), 0.1, nil, 0.0),
      createHideUiAction(self)
    )
  )

  wait(roundEndAction:start())

  self.rationaleForRound:removeSelf()

  buildWordLabelsForRound(self, round)
  buildQuestionAndResponsesForRound(self, round)

  if (roundIndex == UpperConfiguration.NUMBER_OF_STAGES_TO_WIN) then
    local specksLayer = UpperSpecksLayer.new()
    self.specksLayer = specksLayer
    self.tower:addWinBricks()
  end

  local delayUntilFadeOut = 0.5

  if (not answeredCorrectly) then
    delayUntilFadeOut = 1.0
  end

  waitFor(delayUntilFadeOut)

  if ((self.isAnswerSubmitted == true) and (self.isSelectedBrickCorrect == true)) then
    local wordBrick = self.tower.wordBrick
    wordBrick:setLoc(self.tower.correctBrick:getLoc())
    wordBrick:resizeWordBrickToResponseBrick()
    self.tower:addEmptyBrick(UpperCommon.RESPONSE_BRICK_WIDTH, nil, UpperCommon.RESPONSE_BRICK_HEIGHT)
    correctBrick:removeSelf()
  end

  local interstitialFadeOut = AHSequence.new(
    roundStartIntersitial:createFadeOutToGameAction(),
    AHRemove.new(roundStartIntersitial)
  )

  local unzoomAction = AHScaleTo.new(zoomGroup, 1.0, AHEaseExponentialInOut.new(), 1.0)
  local resetBrickColorAction = self.tower:createResetBrickColorAction()
  local showBubblesAction = AHSpawn.new(
    self.backBubbleLayer:createShowBubblesAction(),
    self.foreBubbleLayer:createShowBubblesAction()
  )
  local backgroundColorAction = getResetBackgroundColorAction(self)

  local shiftDistance = self.wordBrick:getY() - Screen.getHeightPart(0.5)
  local shiftDuration = 0.5

  local shiftTowerDownAction = AHDoNothing.new()
  if (self.isSelectedBrickCorrect == true) then
    shiftTowerDownAction = self.tower:createShiftBricksDownAction(shiftDistance, shiftDuration, AHEaseExponentialIn.new())
  end

  local shiftResponseBrickDown = AHLocationYBy.new(
    self.wordBrick,
    shiftDuration,
    AHEaseExponentialIn.new(),
    -shiftDistance
  )

  local returnToGameAction = AHSpawn.new(
    self.foreBubbleLayer:createChangeColorsForStateAction(UpperCommon.COLOR_STATES.NEUTRAL, 0.5),
    interstitialFadeOut,
    resetBrickColorAction,
    unzoomAction,
    backgroundColorAction,
    showBubblesAction,
    shiftResponseBrickDown,
    shiftTowerDownAction
  )

  returnToGameAction:start()
end

local function waitForBricks(self, roundIndex, round)
  local delayForBricks = calculateDelayForWpm(self)
  local wordBrick = self.wordBrick
  local wordsInRound = wordBrick.wordsInCurrentRound
  local tower = self.tower
  local isFirstRound = (roundIndex == 1)


  if (not isFirstRound) then
    waitFor(0.4)
  end
  showUI(self)

  wait(self.tower:startCountdown(isFirstRound))

  self.backBubbleLayer:setScrollSpeed(1.0)
  self.foreBubbleLayer:setScrollSpeed(1.0)

  self.isAnswerSubmitted = false
  self.isSelectedBrickCorrect = nil

  local words = StringUtils.split(round.snippet, " ")
  local progressIndicator = self.progressIndicator
  local currentStep = progressIndicator:getProgressStep()
  local updatedStep = currentStep + 1
  local fillDuration = delayForBricks * #words

  local shiftDistance = UpperCommon.STANDARD_BRICK_HEIGHT + UpperCommon.BRICK_PADDING

  local fullScreenBounce = AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(self.background, 0.5, AHEaseExponentialIn.new(), 1.5),
      AHLocationYBy.new(self, 0.5, AHEaseExponentialIn.new(), -Screen.getHeightPart(0.05))
    ),
    AHSpawn.new(
      AHScaleTo.new(self.background, 0.5, AHEaseSineInOut.new(), 1.0),
      AHLocationYBy.new(self, 0.5, AHEaseSineInOut.new(), Screen.getHeightPart(0.05))
    )
  )

  for labelIndex, label in ipairs(wordsInRound) do
    progressIndicator:setProgressStepWithDuration(updatedStep, fillDuration)
    tower:addEmptyBrick(wordBrick:getAssignedWidth())
    tower:shiftBricksDown(shiftDistance)
    wordBrick:showNextWord()

    if (labelIndex == wordsInRound) then
      fullScreenBounce:start()
    end

    waitFor(delayForBricks)
  end

end

local function createShadowForTower(self)
  local shadow = Sprite.newWithName("shadow.png")
  local bottomBrickX, bottomBrickY = self.bottomBrick:getLoc()

  shadow:setEffectiveDims(UpperCommon.RESPONSE_BRICK_WIDTH + QUESTION_SHADOW_PADDING, 2 * UpperCommon.RESPONSE_BRICK_HEIGHT + UpperCommon.BRICK_PADDING + QUESTION_SHADOW_PADDING)
  shadow:setLoc(bottomBrickX, bottomBrickY + (UpperCommon.RESPONSE_BRICK_HEIGHT + UpperCommon.BRICK_PADDING) / 2)
  shadow:setOpacity(0.0)
  self.zoomGroup:insert(shadow, Z_SHADOW_FOR_QUESTION)

  return shadow
end

local function setAnswerBricksTouchEnabled(self, isEnabled)
  self.topBrick:setTouchEnabled(isEnabled)
  self.bottomBrick:setTouchEnabled(isEnabled)
end

local function createQuestionOverlay(self)
  local questionOverlay = UpperBackground.newQuestionTopLayer()
  questionOverlay:setLoc(0.0, -Screen.getHeightPart(0.45))
  self:insert(questionOverlay, Z_ABOVE_BUBBLES)
  self.questionOverlay = questionOverlay

  return questionOverlay
end

local function registerSubmittedAnswer(self, round)
  local tower = self.tower
  local timer = self.questionTimer

  timer:pause()

  local answerSubmittedAction = AHSpawn.new(
    AHSequence.new(
      self.questionOverlay:newOpacityToAction(0.5, nil, 0.0, 0.0),
      AHRemove.new(self.questionOverlay)
    ),
    AHSequence.new(
      AHScaleTo.new(timer, 0.1, nil, 0.0)
    )
  )

  answerSubmittedAction:start()

  tower:setAnswerHasBeenSubmitted(self.isAnswerSubmitted, self.isSelectedBrickCorrect)
  if self.isSelectedBrickCorrect then
    Sound.playSoundEffect("correct_answer")
    registerCorrectResponse(self)

    local towerAction = tower:createCorrectAnswerAction()
    local uiColorAction = createUiColorAction(self, UI_CORRECT_COLOR)
    local correctResponseAction = AHSpawn.new(towerAction, uiColorAction)

    GameSession.contentTracker:notifyCorrectAnswer(self.roundGenerator.conceptIdentifier, round.questionIdentifier)

    waitFor(0.1)
    correctResponseAction:start()
    checkAndWaitForGameEnd(self)
    local screenAndBubblesAction = AHSpawn.new(
      self.foreBubbleLayer:createChangeColorsForStateAction(UpperCommon.COLOR_STATES.CORRECT),
      self.background:createToCorrectAction(0.1),
      AHSequence.new(
        AHScaleTo.new(self, 0.1, nil, 0.99),
        AHScaleTo.new(self, 0.3, nil, 1.01),
        AHScaleTo.new(self, 0.3, nil, 0.995),
        AHScaleTo.new(self, 0.3, nil, 1.005),
        AHScaleTo.new(self, 0.8, AHEaseSineOut.new(), 1.0)
      )
    )
    screenAndBubblesAction:start()
    waitFor(1.25)
  else
    registerIncorrectResponse(self)
    local incorrectOverlay = UpperBackground.newIncorrectTopLayer()
    local screenColorAction = self.background:createToIncorrectAction(0.25)
    local hideBubblesAction = self.foreBubbleLayer:createHideBubblesAction()

    local isTimeOut = ((self.isAnswerSubmitted == false) and (self.isSelectedBrickCorrect == false))
    local towerAction = self.tower:createIncorrectAnswerAction(isTimeOut)
    local uiColorAction = createUiColorAction(self, UI_INCORRECT_COLOR)
    local incorrectResponseAction = AHSpawn.new(hideBubblesAction, screenColorAction, towerAction, uiColorAction)

    if isTimeOut then
      GameSession.contentTracker:notifyTimeout(self.roundGenerator.conceptIdentifier, round.questionIdentifier)
    else
      GameSession.contentTracker:notifyIncorrectAnswer(self.roundGenerator.conceptIdentifier, round.questionIdentifier)
    end

    waitFor(0.1)
    self:insert(incorrectOverlay, Z_ABOVE_BUBBLES)
    self.incorrectOverlay = incorrectOverlay
    incorrectResponseAction:start()
  end
end

local function isTutorialEnabled(roundIndex)
  return (roundIndex == 1)
end

local function displaySceneForQuestionAndAnswer(self, roundIndex)
  local topBrick = self.topBrick
  local bottomBrick = self.bottomBrick
  local tower = self.tower
  local timer = self.questionTimer

  if (roundIndex >= UpperConfiguration.NUMBER_OF_STAGES_TO_WIN) then
    tower:repositionVictoryBricks()
  end

  topBrick:setVisibleRecursive(true)
  bottomBrick:setVisibleRecursive(true)

  self.backBubbleLayer:setScrollSpeed(0.1)

  local shadow = createShadowForTower(self)
  tower:addAnswerBricksAndQuestion(bottomBrick, topBrick, self.questionTextArea, shadow)

  self.questionOverlay = createQuestionOverlay(self)

  local timerDuration = self.currentCountdownDuration
  if isTutorialEnabled(roundIndex) then
    timerDuration = UpperConfiguration.getTutorialCountdownDuration()
  end

  timer:setDuration(timerDuration)
  timer:setTimeRemaining(timer:getDuration())

  -- visibily iniitialize the timer by starting and immediately pausing
  timer:start()
  timer:pause()
end

local function createAnswerHole(self)
  local answerRect = FilledRectangle.new(self.correctBrick:getWidth(), self.correctBrick:getHeight(), {1.0, 1.0, 1.0, 1.0})
  local answerX, answerY = self.correctBrick:getLoc()
  answerRect:setLoc(answerX, answerY)
  return answerRect
end

local function showHelpAndWaitForTap(self)
  self.questionTimer:pause()

  local answerHoleFactory = function() return createAnswerHole(self) end

  local helpLayer = TutorialHighlighter.newWithMasks({answerHoleFactory})
  helpLayer:addInstructionLabel(Localizer.get("TUTORIAL"), TUTORIAL_FONT_SIZE)
  self:insertAtCenter(helpLayer)

  helpLayer:createFadeInAnimation():start()

  helpLayer:waitForTargetTap()

  wait(helpLayer:createFadeOutAnimation():start())
  helpLayer:removeSelf()
end

local function waitForQuestionAndAnswer(self, round, roundIndex)
  local timer = self.questionTimer

  buildRationaleForRound(self, round)
  displaySceneForQuestionAndAnswer(self, roundIndex)

  local backgroundToWhiteAction = self.background:createNeutralToQuestionAction(1.0)
  local fadeInQuestionOverlayAction = self.questionOverlay:newOpacityToAction(1.0, nil, 0.0, 1.0)
  local hideBackBubblesAction = self.backBubbleLayer:createHideBubblesAction()
  local hideProgressBarAction = AHOpacityTo.new(self.progressIndicator, 0.5, nil, 0.0)
  local slowForeBubbles = AHDelayed.new(0.5, function() self.foreBubbleLayer:setScrollSpeed(0.05) end)

  local fullSceneAction = AHSequence.new(
    AHSpawn.new(
      hideBackBubblesAction,
      backgroundToWhiteAction,
      hideProgressBarAction,
      fadeInQuestionOverlayAction,
      slowForeBubbles
    ),
    function()
      timer:setLocUsingAnchorPoint(Screen.getSafeX(0.5),
                                   self.questionTextArea:getYTop() + QUESTION_VERTICAL_PADDING,
                                   nil,
                                   0.0,
                                   -0.25)
    end,
    AHScaleTo.new(timer, 0.5, AHEaseExponentialOut.new(), 1.0),
    function() timer:start() end
  )

  local answerSubmittedOrTimeOut = function()
    if isTutorialEnabled(roundIndex) then
      return (self.isAnswerSubmitted or timer:getTime() > UpperConfiguration.TUTORIAL_DELAY)
    end

    return (self.isAnswerSubmitted or (timer:getTimeRemaining() <= 0.001))
  end

  -- Start response speed tracking now since input will be immediately allowed.
  -- Question timer is started near the end of the fullSceneAction
  self.responseSpeedTracker:notifyStart()

  -- Start scene animations and enable touch simultaneously...
  fullSceneAction:start()
  setAnswerBricksTouchEnabled(self, true)

  waitUntil(answerSubmittedOrTimeOut)

  if isTutorialEnabled(roundIndex) and not self.isAnswerSubmitted then
    showHelpAndWaitForTap(self)
    self.correctButtonCallback()
  end

  -- ...and stop scene animations and disable touch after waiting.
  fullSceneAction:stop()
  setAnswerBricksTouchEnabled(self, false)

  registerSubmittedAnswer(self, round)
end

local function waitForRationale(self, round, roundIndex)
  local rationale = self.rationaleForRound
  local isTimeOut = ((self.isAnswerSubmitted == false) and (self.isSelectedBrickCorrect == false))

  if (not isTimeOut) then
    local incorrectResponse = round.incorrectResponse
    rationale:addPlayerResponse(incorrectResponse)
  end

  local showRationaleAction = AHStaggeredSpawn.new(
    1.0,
    createHideUiAction(self),
    rationale:createFadeInAction()
  )
  local delayForAnswerAnimation = 0.5
  waitFor(delayForAnswerAnimation)
  wait(showRationaleAction:start())

  rationale:setTouchEnabled(true)

  local prepColorAction = AHSpawn.new(
    self.backBubbleLayer:createShowBubblesAction(),
    self.foreBubbleLayer:createShowBubblesAction(),
    self.foreBubbleLayer:createChangeColorsForStateAction(UpperCommon.COLOR_STATES.NEUTRAL, 0.0),
    self.background:createToNeutralAction(0.0),
    AHRemove.new(self.incorrectOverlay)
  )

  wait(prepColorAction:start())

  waitUntil(function() return rationale.canContinue end)
  rationale:setTouchEnabled(false)

  local isFirstRationale = (roundIndex == 1)

  local reverseBubbles = function()
    self.backBubbleLayer:graduallySetScrollSpeedY(0.5, -0.5)
    self.foreBubbleLayer:graduallySetScrollSpeedY(0.5, -5.0)
  end

  local settleBubbles = function()
    self.backBubbleLayer:graduallySetScrollSpeedY(1.3, 0.01)
    self.foreBubbleLayer:graduallySetScrollSpeedY(1.3, 0.05)
  end

  local backToGameAction = AHSpawn.new(
    rationale:createFadeOutAction(),
    AHDelayed.new(1.0, self.tower:createPanUpAction(isFirstRationale)),
    AHDelayed.new(1.8, settleBubbles),
    reverseBubbles
  )

  wait(backToGameAction:start())
  checkAndWaitForGameEnd(self)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function UpperGame:gameFlow()
  GameSession.contentTracker:notifyPresentedConcept(self.roundGenerator:getConceptIdentifier(), self.roundGenerator:getConceptDifficulty())

  for roundIndex = 1, #self.roundGenerator.rounds do
    local round = self.roundGenerator:next()
    local isFirstRound = (roundIndex == 1)

    if isFirstRound then
      waitForBeginGameInterstitial(self, round)
    else
      waitForRoundStart(self, roundIndex, round)
    end

    waitForBricks(self, roundIndex, round)
    waitForQuestionAndAnswer(self, round, roundIndex)
    if not self.isSelectedBrickCorrect then
      waitForRationale(self, round, roundIndex)
    end
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------
local function configureGameDifficulty(self)
  self.difficulty = GameSession.getDifficulty()
  self.setsToComplete = UpperConfiguration.NUMBER_OF_STAGES_TO_WIN
  self.currentWordsPerMinute = UpperConfiguration.getInitialWordsPerMinuteForDifficulty(self.difficulty)

  local difficultyKeyFrames = UpperConfiguration.getDifficultyKeyFrames()
  local difficultyNameChooser = KeyFrameInterpolator.new(difficultyKeyFrames, self.difficulty)
  GameSession.conceptGenerator:setDifficultyNameChooser(difficultyNameChooser)

  local passScore = UpperConfiguration.getPassScore()
  local rankTwoScore = UpperConfiguration.getRankTwoScore()
  local rankThreeScore = UpperConfiguration.getRankThreeScore()

  GameSession.setPassScore(passScore)
  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.setInitialLives(UpperConfiguration.LIVES_COUNT)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.LIVES_BONUS)
end

local function createGameElements(self)
  local background = UpperBackground.new()
  self:insertAtCenter(background, Z_BACKGROUND)
  self.background = background

  -- This is for zooming and unzooming during round transition
  local zoomGroup = Group.newFullScreen()
  self.zoomGroup = zoomGroup
  self:insertAtCenter(zoomGroup)

  local wordBrick = UpperWordBrick.new(Localizer.get("READY"))
  wordBrick:createFadeOutWordBrickAction(0.0):start()
  zoomGroup:insertAtCenter(wordBrick, Z_WORD_BRICK)
  self.wordBrick = wordBrick

  local tower = UpperTower.new(wordBrick)
  zoomGroup:insertAtCenter(tower, Z_TOWER)
  self.tower = tower
  self.isAnswerSubmitted = false
  self.isSelectedBrickCorrect = nil

  local backBubbleLayer = UpperBubbleLayer.new(true, 4, Screen.getHeightPart(0.08), Screen.getHeightPart(0.08))
  zoomGroup:insertAtCenter(backBubbleLayer, Z_BACK_BUBBLE_LAYER)
  self.backBubbleLayer = backBubbleLayer

  local foreBubbleLayer = UpperBubbleLayer.new(false, 2, Screen.getHeightPart(0.14), Screen.getHeightPart(0.26))
  zoomGroup:insertAtCenter(foreBubbleLayer, Z_BACK_BUBBLE_LAYER)
  self.foreBubbleLayer = foreBubbleLayer

  loadBeginGameInterstitial(self)

  zoomGroup:setScl(5.0)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  local numberOfCountdownDecrements = UpperConfiguration.NUMBER_OF_STAGES_TO_WIN - 1
  local countdownDecrementRange = UpperConfiguration.getMinimumCountdownDuration() - UpperConfiguration.getInitialCountdownDuration()
  self.currentCountdownDuration = UpperConfiguration.getInitialCountdownDuration()
  self.countdownDecrement = MathUtils.round(countdownDecrementRange / numberOfCountdownDecrements, 2)

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.bonusTray:setColor(unpack(UI_CORRECT_COLOR))
  self.livesCounter = builder:addLivesDisplay(UpperConfiguration.LIVES_COUNT)

  local onTimeOut = function()
    self.isAnswerSubmitted = false
    self.isSelectedBrickCorrect = false
  end

  self.questionTimer = UpperPieTimer.new(UpperConfiguration.getInitialCountdownDuration(), onTimeOut)
  self.questionTimer:addStandardWarningSequence()
  self:insert(self.questionTimer)
  self.questionTimer:setScl(0.0)

  local progressIndicator = UpperProgressIndicator.newWithStagesCount(UpperConfiguration.NUMBER_OF_STAGES_TO_WIN)
  local progressIndicatorWidth, progressIndicatorHeight = progressIndicator:getDims()
  local xOffset = 6.0
  local yOffset = -self.livesCounter:getHeightPart(0.5) - progressIndicatorHeight * 0.5 - UI_VERTICAL_PADDING

  builder:addWidget(progressIndicator, UIOverlay.NORTHEAST, xOffset, yOffset)
  self.progressIndicator = progressIndicator

  self.uiOverlay:linkColorAttrToChildren(true)
  self.uiOverlay:setOpacity(0.0)

  local quickResponseWrapper = function(earnedBonusProportion, bonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 bonusProportionsSum,
                                 UpperConfiguration.MAX_SPEED_BONUS_PER_RESPONSE,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(UpperConfiguration.SPEED_BONUS_DURATION,
                                                       quickResponseWrapper,
                                                       UpperConfiguration.SPEED_BONUS_DECAY_START)
end

-- Constructor ------------------------

function UpperGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, UpperGame)

  UpperCommon.initialize()

  configureGameDifficulty(self)
  createGameElements(self)
  createUI(self)

  self.roundGenerator = UpperRoundGenerator.new()

  return self
end
