--- UpperProgressIndicator
-- @author Created by steven on February 23, 2016.

class("UpperProgressIndicator")

require("widgets.progress.ProgressIndicator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local WIDTH = 8.0
local HEIGHT = Screen.safeHeight * 0.3

local CENTER_X = WIDTH * 0.5

local MAX_BRICKS_COUNT = 25
local BRICK_SPACING = 1.0

local BRICK_WIDTHS = {
  WIDTH * 1.00,
  WIDTH * 0.72,
  WIDTH * 0.62
}
local BRICK_WIDTHS_COUNT = #BRICK_WIDTHS

local BRICK_COLOR_UNLIT = UpperCommon.NEUTRAL_PURPLE
local BRICK_COLOR_LIT = UpperCommon.WHITE
local BRICK_COLOR_WRONG = UpperCommon.INCORRECT_RED
local BRICK_OPACITY_UNLIT = 0.3
local BRICK_OPACITY_LIT = 1.0

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function UpperProgressIndicator:emptyQuickly()
  self.super.setProgressStep(self, 0)

  local allBricks = self.allBricks
  local bricksCount = #allBricks

  local drainDurationPerBrick = 0.008
  local brickDrainActions = {}

  for i = bricksCount, 1, -1 do
    local brick = allBricks[i]

    local brickDrainAction = AHOpacityTo.new(brick, drainDurationPerBrick, nil, BRICK_OPACITY_UNLIT)
    table.insert(brickDrainActions, brickDrainAction)
  end

  local emptyAction = AHSequence.new(unpack(brickDrainActions))
  emptyAction:start()
end

function UpperProgressIndicator:setProgressStepWithDuration(step, duration)
  self.super.setProgressStep(self, step)
  step = self:getProgressStep()

  local previousStepValue = self.previousStepValue
  self.previousStepValue = step

  local isIncreasing = (step > previousStepValue)
  local isDecreasing = (step < previousStepValue)

  local stageNumber = nil
  if (isIncreasing) then
    stageNumber = step
  elseif (isDecreasing) then
    stageNumber = previousStepValue
  else
    return
  end

  local bricksForStage = self.bricksForStage
  local bricks = bricksForStage[stageNumber]

  local bricksCount = #bricks
  local lightDurationPerBrick = duration / bricksCount

  local firstBrickIndex = nil
  local lastBrickIndex = nil
  local direction = nil
  local destinationOpacity = nil
  local destinationColor = nil
  if (isIncreasing) then
    firstBrickIndex = 1
    lastBrickIndex = bricksCount
    direction = 1
    destinationOpacity = BRICK_OPACITY_LIT
    destinationColor = BRICK_COLOR_LIT
  elseif (isDecreasing) then
    firstBrickIndex = bricksCount
    lastBrickIndex = 1
    direction = -1
    destinationOpacity = BRICK_OPACITY_UNLIT
    destinationColor = BRICK_COLOR_UNLIT
  end

  local brickActions = {}
  for i = firstBrickIndex, lastBrickIndex, direction do
    local brick = bricks[i]
    local brickAction = AHSpawn.new(
      AHColorTo.new(brick, lightDurationPerBrick, nil, destinationColor),
      AHOpacityTo.new(brick, lightDurationPerBrick, nil, destinationOpacity)
    )
    table.insert(brickActions, brickAction)
  end

  local lightAction = AHSequence.new(unpack(brickActions))
  lightAction:start()
end

function UpperProgressIndicator:setProgressStep(step)
  local defaultDuration = 0.3
  self:setProgressStepWithDuration(step, defaultDuration)
end

function UpperProgressIndicator:displayIncorrect()
  local stageNumber = self.previousStepValue
  local bricksForStage = self.bricksForStage
  local bricks = bricksForStage[stageNumber]
  local durationPerBrick = 0.3 / #bricks

  local brickActions = {}
  for _, brick in ipairs(bricks) do
    local brickAction = AHSpawn.new(
      AHColorTo.new(brick, durationPerBrick, nil, UpperCommon.INCORRECT_RED),
      AHOpacityTo.new(brick, durationPerBrick, nil, 0.25)
    )
    table.insert(brickActions, brickAction)
  end

  AHSequence.new(unpack(brickActions)):start()
end

-- UIOverlay Widget -------------------

function UpperProgressIndicator:prepareForGameIntro()
  local allBricks = self.allBricks
  for i, brick in ipairs(allBricks) do
    brick:setOpacity(0.0)
  end
end

function UpperProgressIndicator:createFadeInAction(duration)
  local brickFadeInActions = {}

  local allBricks = self.allBricks
  for i, brick in ipairs(allBricks) do
    local brickCurrentOpacity = brick.currentOpacity
    local brickFadeInAction = AHOpacityTo.new(brick, duration, nil, brickCurrentOpacity)
    table.insert(brickFadeInActions, brickFadeInAction)
  end

  local fadeInAction = AHSpawn.new(unpack(brickFadeInActions))
  return fadeInAction
end

function UpperProgressIndicator:createFadeOutAction(duration)
  local brickFadeOutActions = {}

  local allBricks = self.allBricks
  for i, brick in ipairs(allBricks) do
    brick.currentOpacity = brick:getOpacity()
    local brickFadeOutAction = AHOpacityTo.new(brick, duration, nil, 0.0)
    table.insert(brickFadeOutActions, brickFadeOutAction)
  end

  local fadeOutAction = AHSpawn.new(unpack(brickFadeOutActions))
  return fadeOutAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function addBricks(self, stagesCount)
  local bricksPerStage = math.floor(MAX_BRICKS_COUNT / stagesCount)
  local bricksCount = bricksPerStage * stagesCount

  local segmentHeight = HEIGHT / bricksCount
  local brickHeight = segmentHeight - BRICK_SPACING

  local allBricks = {}
  local bricksForStage = {}

  local brickIndex = 0
  local firstBrickY = brickHeight * 0.5
  for stageNumber = 1, stagesCount do
    local bricks = {}

    for i = 1, bricksPerStage do
      local brickWidthIndex = (brickIndex % BRICK_WIDTHS_COUNT) + 1
      local brickWidth = BRICK_WIDTHS[brickWidthIndex]
      local brick = FilledRectangle.new(brickWidth, brickHeight, BRICK_COLOR_UNLIT)
      brick:setOpacity(BRICK_OPACITY_UNLIT)
      brick.currentOpacity = BRICK_OPACITY_UNLIT

      local y = firstBrickY + brickIndex * segmentHeight
      brick:setLoc(CENTER_X, y)
      self:insert(brick)
      table.insert(allBricks, brick)
      table.insert(bricks, brick)

      brickIndex = brickIndex + 1
    end

    table.insert(bricksForStage, bricks)
  end

  self.allBricks = allBricks
  self.bricksForStage = bricksForStage
end

-- Constructor ------------------------

function UpperProgressIndicator.newWithStagesCount(stagesCount)
  local self = Group.new(WIDTH, HEIGHT)
  ProgressIndicator.makeProgressIndicator(self, stagesCount)
  Mixin.mix(self, UpperProgressIndicator)

  addBricks(self, stagesCount)

  self.previousStepValue = 0

  return self
end
