-- VendettaCircuitLayer
-- Created by Steven on August 22, 2016

class("VendettaCircuitLayer")

require("actions.advanced.PathBezier")
require("VendettaOrb")
require("VendettaAnswerBeam")
require("VendettaAnswerLine")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ORB_RADIUS = 6

local COLORS = {
  {0.76, 0.3, 0.57, 1.0},
  {0.77, 0.3, 0.45, 1.0},
  {0.80, 0.29, 0.38, 1.0},
  {0.77, 0.35, 0.3, 1.0}
}

local WHITE = {1.0, 1.0, 1.0, 1.0}

local Z_INDICES = {
  LINE = 0,
  BEAM = 1,
  ORB = 5
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getBezierPoints(startPoint, midPoint, endPoint)
  local curveOffset = Screen.getWidthPart(0.05)
  local offshootPoint, bezierStartPoint, bezierEndPoint
  local isInlineWithEnd = (midPoint.x == endPoint.x)

  if (isInlineWithEnd) then
    offshootPoint = startPoint
  else
    offshootPoint = endPoint
  end

  local angleBetweenOffshootAndMid = MathUtils.angleBetweenVectors(midPoint.x, midPoint.y, 0.0, midPoint.y, offshootPoint.x, offshootPoint.y)
  local bezierXOffset = curveOffset * -math.cos(angleBetweenOffshootAndMid)
  local bezierYOffset = curveOffset * -math.sin(angleBetweenOffshootAndMid)

  if (isInlineWithEnd) then
    curveOffset = -curveOffset
    bezierEndPoint = {x = midPoint.x, y = midPoint.y + curveOffset}
    bezierStartPoint = {x = midPoint.x + bezierXOffset, y = midPoint.y + bezierYOffset}
  else
    bezierStartPoint = {x = midPoint.x, y = midPoint.y + curveOffset}
    bezierEndPoint = {x = midPoint.x + bezierXOffset, y = midPoint.y + bezierYOffset}
  end

  return bezierStartPoint, bezierEndPoint
end

local function makeButton(self, prompt)
  local index = self.promptsForIndex[prompt]
  local orb = self.orbs[index]
  local orbTopY = orb:getYTop()
  local promptBottomY = prompt:getYBottom()

  local buttonHeight = VendettaCommon.PROMPT_HEIGHT
  local buttonWidth = VendettaCommon.PROMPT_WIDTH
  local y = (orbTopY - promptBottomY) / 2 + promptBottomY

  -- callback sets the selected prompt in VendettaScene to advance game
  local onTapWrapper = function()
    self:setTouchEnabled(false)
    self.onTapCallback(prompt)
  end

  local button = Button.new(buttonWidth, buttonHeight, onTapWrapper)
  button:setLoc(orb:getX(), y)
  button.touchHandler = onTapWrapper
  self:insert(button, Z_INDICES.ORB)

  self.buttons[prompt] = button
end

local function getLineSplitPoints(orbPoint, numberOfPrompts, isMiddleLine)
  local orbX, orbY = orbPoint.x, orbPoint.y
  local synonymX, synonymY = Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y

  local endXOffset = ORB_RADIUS
  local midYOffset = Screen.getHeightPart(0.05)
  local endYOffset = 0

  if (orbX < synonymX) then
    endXOffset = endXOffset * -1
  end

  if (isMiddleLine and (numberOfPrompts == 4)) then
    endXOffset = endXOffset * 1.5
    endYOffset = ORB_RADIUS * -1.5
    midYOffset = midYOffset * 1.5
  elseif (isMiddleLine and (numberOfPrompts < 4)) then
    endXOffset = 0
    endYOffset = ORB_RADIUS * -2.0
  else
    endXOffset = endXOffset * 2.0
  end

  local points = {
    {x = synonymX + endXOffset, y = synonymY + endYOffset},
    {x = orbX, y = synonymY - midYOffset},
    {x = orbX, y = orbY}
  }

  local curveStartPoint, curveEndPoint = getBezierPoints(points[1], points[2], points[3])

  local pathPoints = {
    {points[1], points[1], curveStartPoint, curveStartPoint},
    {points[2], curveStartPoint, curveEndPoint, points[2]},
    {curveEndPoint, curveEndPoint, points[3], points[3]}
  }

  return pathPoints
end

local function connectN(pointsCount, ...)
  local result = {}
  local maxPointIndexPerCurve = (pointsCount - 1) / #{...}

  for _, bezierData in ipairs({...}) do
    local bezierPath = PathBezier.new(nil, bezierData[1], bezierData[2], bezierData[3], bezierData[4])

    for i = 0, maxPointIndexPerCurve do
      local t = i/maxPointIndexPerCurve
      local x, y = bezierPath:getLocationAtProgress(t)

      table.insert(result, {x, y})
    end
  end

  return result
end

local function getLineBounceAction(line, bounceDuration, bounceEase, yDelta)
  local lineBounce = AHSequence.new(
    AHSpawn.new(
      AHLocationYBy.new(line, bounceDuration, bounceEase, yDelta),
      AHScaleYTo.new(line, bounceDuration, bounceEase, 1.1)
    ),
    AHSpawn.new(
      AHLocationYBy.new(line, bounceDuration, bounceEase, -yDelta * 1.5),
      AHScaleYTo.new(line, bounceDuration, bounceEase, 0.95)
    ),
    AHSpawn.new(
      AHLocationYBy.new(line, bounceDuration, bounceEase, yDelta * 0.75),
      AHScaleYTo.new(line, bounceDuration, bounceEase, 1.025)
    ),
    AHSpawn.new(
      AHLocationYBy.new(line, bounceDuration, bounceEase, -yDelta * 0.25),
      AHScaleYTo.new(line, bounceDuration, bounceEase, 1.0)
    )
  )

  return lineBounce
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaCircuitLayer:createCorrectAction(prompt)
  local index = self.promptsForIndex[prompt]
  local orb = self.orbs[index]
  local line = self.lines[index]
  local beam = self.linesToBeams[line]

  local yDelta = -orb:getHeightPart(1.1)
  local bounceEase = AHEaseSineInOut.new()
  local bounceDuration = 0.1

  local lineBounce = getLineBounceAction(line, bounceDuration, bounceEase, yDelta)
  local beamBounce = getLineBounceAction(beam, bounceDuration, bounceEase, yDelta)

  local orbBounce = AHSequence.new(
    AHLocationYBy.new(orb, bounceDuration, bounceEase, yDelta * 2),
    AHLocationYBy.new(orb, bounceDuration, bounceEase, -yDelta * 3),
    AHLocationYBy.new(orb, bounceDuration, bounceEase, yDelta * 1.5),
    AHLocationYBy.new(orb, bounceDuration, bounceEase, -yDelta * 0.5)
  )

  local promptBounce = AHSequence.new(
    AHDelay.new(0.005),
    AHLocationYBy.new(prompt, bounceDuration, bounceEase, yDelta * 2),
    AHLocationYBy.new(prompt, bounceDuration, bounceEase, -yDelta * 3),
    AHLocationYBy.new(prompt, bounceDuration, bounceEase, yDelta * 1.5),
    AHLocationYBy.new(prompt, bounceDuration, bounceEase, -yDelta * 0.5)
  )

  local bounceAction = AHSpawn.new(
    lineBounce,
    beamBounce,
    orbBounce,
    promptBounce
  )

  index = self.promptsForIndex[prompt]
  local beamAction = self.beamLineActions[index]
  beam:setOpacity(1.0)

  local correctAction = AHSpawn.new(
    function() orb:playCorrectAction(bounceDuration * 3.0) end,
    bounceAction
  )

  return correctAction, beamAction
end

function VendettaCircuitLayer:createIncorrectAction(prompt)
  local index = self.promptsForIndex[prompt]
  local line = self.lines[index]
  local beam = self.linesToBeams[line]
  local orb = self.orbs[index]
  self.invalidPrompts[prompt] = true

  local dimUnselectedPrompts = AHSpawn.new()
  local revertUnselectedPrompts = AHSpawn.new()

  local duration = 0.5

  for i = 1, self.numberOfPrompts do
    local unselectedPrompt = self.prompts[i]

    if (i ~= index and (not self.invalidPrompts[unselectedPrompt])) then
      local unselectedLine = self.lines[i]
      local unselectedOrb = self.orbs[i]

      dimUnselectedPrompts:addAction(
        AHSpawn.new(
          function()
            unselectedOrb:playFadeAction(true, duration / 2)
            unselectedOrb:playRevertSizeAction(duration / 2)
            unselectedLine:changeColorTo(nil, 0.25, duration / 2)
            self.linesToBeams[unselectedLine]:hide(duration / 2)
          end,
          AHOpacityTo.new(unselectedPrompt, duration / 2, nil, 0.25)
        )
      )

      revertUnselectedPrompts:addAction(
        AHSpawn.new(
          AHSpawn.new(
            function()
              unselectedOrb:playFadeAction(false, duration)
              unselectedLine:changeColorTo(nil, 1.0, duration)
            end,
            AHOpacityTo.new(unselectedPrompt, duration, nil, 1.0)
          )
        )
      )
    end
  end

  local promptShake = VendettaCommon.createShakeAction(prompt)
  local orbShake = VendettaCommon.createShakeAction(orb)
  local lineShake = VendettaCommon.createShakeAction(line)
  local beamShake = VendettaCommon.createShakeAction(beam)

  local ease = nil
  local incorrectAction = AHSpawn.new(
    promptShake,
    orbShake,
    lineShake,
    beamShake,
    function()
      orb:playIncorrectAction(0.25, ease)
      self.linesToBeams[line]:hide(0.25)
      line:changeColorTo({1.0, 1.0, 1.0}, 1.0)
    end,
    dimUnselectedPrompts
  )

  return incorrectAction, revertUnselectedPrompts
end

function VendettaCircuitLayer:getReshowInvalidPromptsAction(duration, isRoundComplete)
  local reenableColorAction = AHSpawn.new()
  duration = duration or 0.5
  for prompt, _ in pairs(self.invalidPrompts) do
    local index = self.promptsForIndex[prompt]
    local line = self.lines[index]
    local orb = self.orbs[index]
    local colorAction = AHSpawn.new(
      function() orb:playFullRevertAction(duration) end,
      AHColorTo.new(prompt, duration, nil, WHITE)
    )
    if (not isRoundComplete) then
      colorAction:addAction(
        function() line:changeColorTo(nil, 1.0, duration) end
      )
    end
    reenableColorAction:addAction(colorAction)
  end

  self.invalidPrompts = {}

  return reenableColorAction
end

-- For use when all buttons need to be disabled or enabled at the same time
function VendettaCircuitLayer:setTouchEnabled(isEnabled)
  for _, button in pairs(self.buttons) do
    button:setTouchEnabled(isEnabled)
  end
end

-- For use when disabling an incorrectly selected button for the first time
function VendettaCircuitLayer:getDisableIncorrectButtonAction(prompt)
  local button = self.buttons[prompt]
  button:setTouchEnabled(false)

  local shouldHide = true
  local index = self.promptsForIndex[prompt]
  local line = self.lines[index]
  local orb = self.orbs[index]
  local duration = 0.25
  local disabledColorAction = AHSpawn.new(
    function()
      orb:playFadeAction(shouldHide, duration)
      line:changeColorTo(nil, 0.25, duration)
    end,
    AHOpacityTo.new(prompt, duration, nil, 0.1)
  )

  return disabledColorAction
end

-- For use after the rationale screen to enable buttons that weren't incorrectly
-- selected yet
function VendettaCircuitLayer:enableValidButtons()
  for prompt, button in pairs(self.buttons) do
    local isPromptInvalid = self.invalidPrompts[prompt]

    if (not isPromptInvalid) then
      button:setTouchEnabled(true)
    end
  end
end

function VendettaCircuitLayer:setPrompts(prompts)
  for index, prompt in ipairs(prompts) do
    self.promptsForIndex[prompt] = index
  end

  self.prompts = prompts
end

function VendettaCircuitLayer:getHideForReviewAction(duration)
  duration = duration or 0.1
  local ease = nil
  local hideAction = AHSpawn.new(
    AHOpacityTo.new(self.synonymOrb, duration, ease, 0.0)
  )

  for index, line in pairs(self.lines) do
    local orb = self.orbs[index]
    hideAction:addAction(
      function()
        line:changeColorTo(nil, 0.0, duration, ease)
        orb:playFullRevertAction(duration)
      end
    )
  end

  return hideAction
end

function VendettaCircuitLayer:getOrb(prompt)
  local index = self.promptsForIndex[prompt]
  local orb = self.orbs[index]

  return orb
end

function VendettaCircuitLayer:getOrbs()
  return self.orbs
end

function VendettaCircuitLayer:createRoundIntro(isFirstRound)
  local duration = 0.5

  local revealPromptsAction = AHSpawn.new()
  local cascadePulseAction = AHStaggeredSpawn.new(0.1)
  local pulseUpEase = AHEaseSimpleInOut.new(5)
  local pulseDownEase = AHEaseSimpleInOut.new(5)
  local pulseMaxScale = 2.0
  local pulseDuration = 1.0

  for index, prompt in ipairs(self.prompts) do
    revealPromptsAction:addAction(
      AHSequence.new(
        AHOpacityTo.new(prompt, duration, nil, 1.0),
        function() makeButton(self, prompt) end
      )
    )

    if (isFirstRound) then
      local orb = self.orbs[index]
      cascadePulseAction:addAction(
        orb:getPulse(pulseDuration, pulseUpEase, pulseDownEase, pulseMaxScale)
      )
    end
  end

  local fullAction = AHSpawn.new(
    self.lineSplitAction,
    AHDelayed.new(0.85, cascadePulseAction),
    AHDelayed.new(0.95, revealPromptsAction)
  )

  return fullAction
end

function VendettaCircuitLayer:createRoundOutro()
  local fullLineAction = AHSpawn.new()
  local shiftAction = AHSpawn.new()

  local nextSynonymOrbX, nextSynonymOrbY = Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y
  nextSynonymOrbY = nextSynonymOrbY + Screen.getHeightPart(-2.0)
  local lineOffset = Screen.getWidthPart(0.025)

  local numberOfLines = #self.orbs

  if (numberOfLines == 0) then
    numberOfLines = 3
  end

  local hasFourLines = (numberOfLines == 4)
  local hasEvenLines = (numberOfLines % 2) == 0

  if (hasEvenLines) then
    lineOffset = lineOffset / 2
  end

  local leftFirstTurnY, leftSecondTurnY, leftMiddleY
  local rightFirstTurnY, rightSecondTurnY, rightMiddleY

  local transitionLines = {}

  for index = 1, numberOfLines do
    local orbX, orbY
    local orb = nil
    local hasOrbs = #self.orbs > 0
    if (hasOrbs) then
      orb = self.orbs[index]
      orbX, orbY = orb:getLoc()
    else
      -- Fake locations for intro lines since there aren't any orbs
      local point = VendettaCommon.REVIEW_LAYOUTS[2][index]
      orbX = point.x
      orbY = Screen.getHeightPart(0.95)
    end
    local destinationX = nextSynonymOrbX
    local isLowerPrompt = ((index == 2) or (index == 3))
    local isOnLeft = orbX < Screen.getWidthPart(0.5)
    local isOnRight = orbX > Screen.getWidthPart(0.5)
    local xShift = 0.0

    local firstTurnY, secondTurnY, middleY = 0, 0, 0
    local pathPoints

    if (not (isOnRight or isOnLeft)) then
      pathPoints = {
        {
          {x = orbX, y = orbY},
          {x = orbX, y = orbY},
          {x = nextSynonymOrbX, y = nextSynonymOrbY},
          {x = nextSynonymOrbX, y = nextSynonymOrbY}
        }
      }
    elseif (isOnLeft) then
      destinationX = destinationX - lineOffset
      if (leftFirstTurnY == nil) then
        leftFirstTurnY = MathUtils.getRandomFloat(orbY - VendettaCommon.SYNONYM_ORB_Y - Screen.getHeightPart(0.3), orbY - Screen.getHeightPart(1.1))
        leftSecondTurnY = MathUtils.getRandomFloat(leftFirstTurnY - Screen.getHeightPart(0.3), leftFirstTurnY - Screen.getHeightPart(0.4))
        leftMiddleY = (leftFirstTurnY + leftSecondTurnY) / 2
      end
      firstTurnY, secondTurnY, middleY = leftFirstTurnY, leftSecondTurnY, leftMiddleY
      if (hasFourLines and isLowerPrompt) then
        xShift = lineOffset * 2.0
      elseif (hasFourLines) then
        destinationX = destinationX - 2 * lineOffset
      end
    elseif (isOnRight) then
      destinationX = destinationX + lineOffset
      if (rightFirstTurnY == nil) then
        rightFirstTurnY = MathUtils.getRandomFloat(orbY - VendettaCommon.SYNONYM_ORB_Y - Screen.getHeightPart(0.3), orbY - Screen.getHeightPart(1.1))
        rightSecondTurnY = MathUtils.getRandomFloat(rightFirstTurnY - Screen.getHeightPart(0.3), rightFirstTurnY - Screen.getHeightPart(0.4))
        rightMiddleY = (rightFirstTurnY + rightSecondTurnY) / 2
      end
      firstTurnY, secondTurnY, middleY = rightFirstTurnY, rightSecondTurnY, rightMiddleY

      if (hasFourLines and isLowerPrompt) then
        xShift = -lineOffset * 2.0
      elseif (hasFourLines) then
        destinationX = destinationX + 2 * lineOffset
      end
    end

    if (isOnRight or isOnLeft) then
      local middleX = (orbX + destinationX) / 2

      local keyPoints = {
        {x = orbX + xShift, y = orbY}, -- start at old (self) promptOrbs
        {x = orbX + xShift, y = firstTurnY}, -- first corner to turn inward
        {x = middleX, y = middleY}, -- midPoint between two corners
        {x = destinationX, y = secondTurnY}, -- second corner where they meet
        {x = destinationX, y = nextSynonymOrbY} -- ends at new synonymOrb
      }

      local firstCurveStartPoint, firstCurveEndPoint = getBezierPoints(keyPoints[1], keyPoints[2], keyPoints[3])
      local secondCurveStartPoint, secondCurveEndPoint = getBezierPoints(keyPoints[3], keyPoints[4], keyPoints[5])

      pathPoints = {
        {keyPoints[1], keyPoints[1], firstCurveStartPoint, firstCurveStartPoint}, -- first straight segment
        {keyPoints[2], firstCurveStartPoint, firstCurveEndPoint, keyPoints[2]}, -- first turn
        {firstCurveEndPoint, firstCurveEndPoint, secondCurveStartPoint, secondCurveStartPoint}, -- middle straight segment
        {keyPoints[4], secondCurveStartPoint, secondCurveEndPoint, keyPoints[4]}, -- second turn
        {secondCurveEndPoint, secondCurveEndPoint, keyPoints[5], keyPoints[5]} -- last straight segment
      }
    end

    local pointsCount = 80
    local flatPoints = connectN(pointsCount, unpack(pathPoints))
    local sourcePoints = {}
    for _, p in ipairs(flatPoints) do
      table.insert(sourcePoints, {x = p[1], y = p[2]})
    end

    local lineWidth = Screen.getWidthPart(0.01)
    local transitionLine = VendettaRoundTransitionLine.new(sourcePoints, lineWidth, COLORS[index])
    table.insert(transitionLines, transitionLine)

    if (hasFourLines and isLowerPrompt and hasOrbs) then
      shiftAction:addAction(
        AHLocationXBy.new(orb, 0.1, nil, xShift)
      )
    end
  end

  local fullTransitionAction = AHSpawn.new(
    shiftAction,
    fullLineAction
  )

  return fullTransitionAction, transitionLines
end

function VendettaCircuitLayer:addExplosion(explosion)
  local y = self.synonymOrb:getY()
  explosion:setLoc(Screen.getWidthPart(0.5), y)
  self:insert(explosion, Z_INDICES.ORB)
end

function VendettaCircuitLayer:setTapCallback(callback)
  self.onTapCallback = callback
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local synonymX, synonymY = Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y
  local synonymOrb = FilledCircle.new(ORB_RADIUS, WHITE)
  synonymOrb:setLoc(synonymX, synonymY)
  self:insert(synonymOrb, Z_INDICES.ORB)
  self.synonymOrb = synonymOrb

  local numberOfPrompts = self.numberOfPrompts

  self.lineSplitAction = AHSpawn.new()

  local duration = 2.0
  local ease = nil

  local layoutIndex = numberOfPrompts - 1
  local layoutPoints = VendettaCommon.ROUND_LAYOUTS[layoutIndex]

  for index = 1, numberOfPrompts do
    local orbPoint = layoutPoints[index]
    local isMiddleLine = (index ~= 1) and (index ~= numberOfPrompts)
    local pathPoints = getLineSplitPoints(orbPoint, numberOfPrompts, isMiddleLine)
    local color = COLORS[index]

    local promptOrbRadius = ORB_RADIUS * 1.1
    local promptOrb = VendettaOrb.newHidden(promptOrbRadius, color)

    local pointsCount = 30
    local flatPoints = connectN(pointsCount, unpack(pathPoints))
    local sourcePoints = {}
    for _, p in ipairs(flatPoints) do
      table.insert(sourcePoints, {x = p[1], y = p[2]})
    end

    local lineWidth = Screen.getWidthPart(0.0038)
    local line = VendettaAnswerLine.new(sourcePoints, lineWidth, color)
    line:setTipOrb(promptOrb)
    self:insertAtCenter(line, Z_INDICES.LINE)
    table.insert(self.lines, line)

    local lineDuration = 1.035
    local lineAppearAction = AHSpawn.new(
      line:getAppearAction(lineDuration, ease),
      AHDelayed.new(duration * 0.25, promptOrb:getAppearAction(duration * 0.25, ease))
    )

    self.lineSplitAction:addAction(lineAppearAction)

    local beamLineWidth = 1.9 * lineWidth
    local visibleSegmentCount = 64
    local answerBeam = VendettaAnswerBeam.new(sourcePoints, beamLineWidth, color, visibleSegmentCount)
    self:insertAtCenter(answerBeam, Z_INDICES.BEAM)
    self.linesToBeams[line] = answerBeam

    local beamAction = answerBeam:getAppearAction(0.9, AHEaseSineInOut.new())
    table.insert(self.beamLineActions, beamAction)

    self:insert(promptOrb, Z_INDICES.ORB)
    table.insert(self.orbs, promptOrb)
  end
end

-- Constructor ------------------------

function VendettaCircuitLayer.new(numberOfPrompts)
  local self = Group.newFullScreen()
  Mixin.mix(self, VendettaCircuitLayer)

  self.numberOfPrompts = numberOfPrompts
  self.orbs = {}

  if (numberOfPrompts > 0) then
    self.promptsForIndex = {}
    self.lines = {}
    self.beamLineActions = {}
    self.buttons = {}
    self.invalidPrompts = {}
    self.linesToBeams = {}

    createLayout(self)
  end

  return self
end
