--- VendettaExplosion
-- @author Created by Joe Poka on September 7, 2016.

class("VendettaExplosion")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  rectangle = 2,
  disk = 3,
  halo = 4
}

local DISK_DURATION = {
  diskScaleUp = 0.6,
  diskScaleDownDelay = 0.4,

  diskScaleDown = 0.6,

  haloScaleUp = 1.0,
  haloScaleDownDelay = 0.0,

  haloScaleDown = 0.7,

  arcRotation = 2.0,
  arcScaleUp = 1.2,
  arcRemovalDelay = 0.3,
  arcShrink = 1.0,
  extraFinishingTime = 1.0
}

local RECTANGLE_DURATION = {
  -- top row
  {
    scaleDuration = 0.6,
    translationDelay = 0.15,
    translationDuration = 0.75
  },
  {
    scaleDuration = 0.6,
    translationDelay = 0.15,
    translationDuration = 0.75
  },
  -- center row
  {
    scaleDuration = 0.6,
    translationDelay = 0.15,
    translationDuration = 0.75
  },
  {
    scaleDuration = 0.6,
    translationDelay = 0.1,
    translationDuration = 0.75
  },
  -- bottom row
  {
    scaleDuration = 0.6,
    translationDelay = 0.2,
    translationDuration = 0.75
  },
  {
    scaleDuration = 0.6,
    translationDelay = 0.2,
    translationDuration = 0.75
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addDiskAndHalos(self)
  -- add disk
  local minimumDiskRadius = Screen.getWidthPart(0.02)
  local maximumDiskRadius = Screen.getWidthPart(0.1)
  local initialDiskScale = minimumDiskRadius / maximumDiskRadius
  local diskColor = {1.0, 1.0, 1.0, 1.0}
  local disk = FilledCircle.new(maximumDiskRadius, diskColor)
  disk:setScl(initialDiskScale)
  self:insertAtCenter(disk, Z_INDEX.disk)

  -- disk scales up, waits, and scales down
  local finalDiskScale = 1.0
  local diskScaleUp = AHScaleTo.new(disk, DISK_DURATION.diskScaleUp, AHEaseExponentialOut.new(), finalDiskScale)
  local diskScaleDown = AHScaleTo.new(disk, DISK_DURATION.diskScaleDown, AHEaseSimpleInOut.new(5.0), 0.0)
  local delayedDiskScaleDown = AHDelayed.new(DISK_DURATION.diskScaleDownDelay, diskScaleDown)
  local diskSequence = AHSequence.new(diskScaleUp, delayedDiskScaleDown)
  table.insert(self.actions, diskSequence)

  -- add halo
  local minimumHaloRadius = minimumDiskRadius * 1.2
  local maximumHaloRadius = maximumDiskRadius * 1.3
  local initialHaloScale = minimumHaloRadius / maximumHaloRadius
  local haloColor = {1.0, 1.0, 1.0, 1.0}
  local haloLineWidth = Screen.getWidthPart(0.004)
  local haloBlurMargin = 0.0
  local halo = HollowCircle.new(maximumHaloRadius, haloColor, haloLineWidth, haloBlurMargin)
  halo:setScl(initialHaloScale)
  self:insertAtCenter(halo, Z_INDEX.halo)

  -- halo scales up, waits, and scales down
  local finalHaloScale = 1.0
  local haloScaleUp = AHScaleTo.new(halo, DISK_DURATION.haloScaleUp, AHEaseExponentialOut.new(), finalHaloScale)
  local haloScaleDown = AHScaleTo.new(halo, DISK_DURATION.haloScaleDown, AHEaseSimpleInOut.new(5.0), 0.0)
  local delayedHaloScaleDown = AHDelayed.new(DISK_DURATION.haloScaleDownDelay, haloScaleDown)
  local haloSequence = AHSequence.new(haloScaleUp, delayedHaloScaleDown)
  table.insert(self.actions, haloSequence)

  -- add rotating halo fragments/arcs
  local minimumArcRadius = minimumHaloRadius
  local maximumArcRadius = maximumHaloRadius * 1.3
  local initalArcScale = minimumArcRadius / maximumArcRadius
  local arcColor = {1.0, 1.0, 1.0, 1.0}
  local arcLineWidth = haloLineWidth
  local arcBlurMargin = 0.0
  local arcHasCenterPoint = false

  local arcData = {
    {
      startAngleRadians = 0.0,
      endAngleRadians = 0.8
    },
    {
      startAngleRadians = 2.4,
      endAngleRadians = 3.0
    },
    {
      startAngleRadians = 4.6,
      endAngleRadians = 5.2
    }
  }

  -- arcs rotate then shrink and fade out
  for _, data in ipairs(arcData) do
    local startAngleRadians = data.startAngleRadians
    local endAngleRadians = data.endAngleRadians

    local arc = HollowArc.new(maximumArcRadius, startAngleRadians, endAngleRadians, arcColor, arcLineWidth, arcBlurMargin, arcHasCenterPoint)
    arc:setScl(initalArcScale)
    self:insertAtCenter(arc, Z_INDEX.halo)

    local finalArcScale = 1.0
    local arcScaleUp = AHScaleTo.new(arc, DISK_DURATION.arcScaleUp, AHEaseExponentialOut.new(), finalArcScale)
    table.insert(self.actions, arcScaleUp)

    local rotate = AHRotationBy.new(arc, DISK_DURATION.arcRotation, nil, 360.0)
    local rotateLoop = AHLoop.new(rotate)
    table.insert(self.actions, rotateLoop)

    local removeArc = function()
      -- note: moving startAngle too close to endAngle causes arc to abruptly disappear
      local destinationShrinkAngle = startAngleRadians + 0.65 * (endAngleRadians - startAngleRadians)
      local shrink = arc:newStartAngleToAction(DISK_DURATION.arcShrink, nil, destinationShrinkAngle)
      shrink:start()

      local delayBeforeFadeOut = 0.4 * DISK_DURATION.arcShrink
      local fadeDuration = DISK_DURATION.arcShrink - delayBeforeFadeOut
      local fadeOut = AHOpacityTo.new(arc, fadeDuration, nil, 0.0)
      local delayedFadeOut = AHDelayed.new(delayBeforeFadeOut, fadeOut)
      delayedFadeOut:start()
    end

    local delayedArcRemoval = AHDelayed.new(DISK_DURATION.arcRemovalDelay, removeArc)
    table.insert(self.actions, delayedArcRemoval)
  end

  -- final clean-up after arcs are gone
  local cleanUpDelay = DISK_DURATION.arcRemovalDelay + DISK_DURATION.arcShrink + DISK_DURATION.extraFinishingTime
  local delayedCleanUp = AHDelayed.new(cleanUpDelay, function()
    self:removeSelf()
  end)
  table.insert(self.actions, delayedCleanUp)
end

local function addRectangles(self)
  local minimumRectangleWidth = Screen.getWidthPart(0.06)
  local rectangleHeight = Screen.getHeightPart(0.03)

  local rectangleData = {
    -- top row
    {
      width = minimumRectangleWidth,
      xOffset = -0.5 * minimumRectangleWidth,
      yOffset = rectangleHeight * 1.5,
      destinationScaleX = 12.0,
      translationDirection = 1.0,
      durations = RECTANGLE_DURATION[1]
    },
    {
      width = minimumRectangleWidth,
      xOffset = 0.5 * minimumRectangleWidth,
      yOffset = rectangleHeight * 1.5,
      destinationScaleX = 8.0,
      translationDirection = 1.0,
      durations = RECTANGLE_DURATION[2]
    },
    -- center row
    {
      width = minimumRectangleWidth * 1.5,
      xOffset = 0.5 * minimumRectangleWidth,
      yOffset = rectangleHeight * 0.1,
      destinationScaleX = 8.0,
      translationDirection = -1.0,
      durations = RECTANGLE_DURATION[3]
    },
    {
      width = minimumRectangleWidth * 1.5,
      xOffset = 2.0 * minimumRectangleWidth,
      yOffset = rectangleHeight * 0.1,
      destinationScaleX = 5.0,
      translationDirection = -1.0,
      durations = RECTANGLE_DURATION[4]
    },
    -- bottom row
    {
      width = minimumRectangleWidth,
      xOffset = 0.5 * minimumRectangleWidth,
      yOffset = -rectangleHeight * 1.4,
      destinationScaleX = 11.0,
      translationDirection = 1.0,
      durations = RECTANGLE_DURATION[5]
    },
    {
      width = minimumRectangleWidth,
      xOffset = 1.5 * minimumRectangleWidth,
      yOffset = -rectangleHeight * 1.4,
      destinationScaleX = 5.5,
      translationDirection = 1.0,
      durations = RECTANGLE_DURATION[6]
    }
  }

  local topBeginColor = {0.98, 0.35, 0.57, 1.0}
  local topEndColor = {0.99, 0.25, 0.98, 1.0}
  local bottomBeginColor = {0.98, 0.27, 0.33, 1.0}
  local bottomEndColor = {0.98, 0.14, 0.53, 1.0}

  for i = 1, #rectangleData do
    local data = rectangleData[i]

    local isTop = (i % 2 == 0)
    local beginColor = (isTop and topBeginColor or bottomBeginColor)
    local endColor = (isTop and topEndColor or bottomEndColor)

    -- add rectangle
    local rectangle = RectangleGradient.new(data.width, rectangleHeight, beginColor, endColor, true)
    self:insertAtCenter(rectangle, Z_INDEX.rectangle)
    rectangle:setLoc(rectangle:getX() + data.xOffset, rectangle:getY() + data.yOffset)

    -- rectangle starts to scale up, then translates horizontally after a short delay
    local scaleUpRectangle = AHScaleXTo.new(rectangle, data.durations.scaleDuration, AHEaseExponentialOut.new(), data.destinationScaleX)
    table.insert(self.actions, scaleUpRectangle)

    local deltaX = rectangle:getWidth() * data.destinationScaleX * 2.0 * data.translationDirection
    local translateRectangle = AHLocationXBy.new(rectangle, data.durations.translationDuration, AHEaseExponentialIn.new(), deltaX)
    local delayedTranslateRectangle = AHDelayed.new(data.durations.translationDelay, translateRectangle)
    table.insert(self.actions, delayedTranslateRectangle)
  end
end

local function adjustTiming(timeFactor)
  for key, duration in pairs(DISK_DURATION) do
    DISK_DURATION[key] = duration * timeFactor
  end

  for _, durations in ipairs(RECTANGLE_DURATION) do
    for key, duration in pairs(durations) do
      durations[key] = duration * timeFactor
    end
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaExplosion:getAnimation()
  return AHSpawn.new(self.actions)
end

function VendettaExplosion:start()
  local animation = self:getAnimation()
  animation:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, hideRectangles)
  self.actions = {}

  if (not hideRectangles) then
    addRectangles(self)
  end

  addDiskAndHalos(self)
end

-- Constructor ------------------------

function VendettaExplosion.new(timeFactor, hideRectangles)
  timeFactor = timeFactor or 1.0
  local self = Group.newFullScreen()
  Mixin.mix(self, VendettaExplosion)

  adjustTiming(timeFactor)
  createLayout(self, hideRectangles)

  -- this is a quick fix for consistent timing
  adjustTiming(1.0 / timeFactor)

  return self
end
