--- VendettaGame
-- @author Created by steven on July 07, 2016.

class("VendettaGame")

FontScaling.setScalingMethod(FontScaling.SCALING_METHOD_HEIGHT)

require("StandardUIBuilder")
require("VendettaCommon")
require("VendettaRoundGenerator")
require("VendettaSynonymGenerator")
require("VendettaConfiguration")
require("VendettaScene")
require("widgets.TapToContinue")
require("bonuses.ResponseSpeedTracker")
require("tutorials.TutorialHighlighter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local UI_COMPONENTS_COLOR = {0.80, 0.29, 0.38}
local UI_INCORRECT_COLOR = {0.68, 0.35, 0.52}
local OVERLAY_COLOR = {0.039, 0.055, 0.082, 0.7}
local MASK_COLOR = {1.0, 1.0, 1.0, 1.0}
local GAME_DURATION = VendettaConfiguration.GAME_DURATION

local function configureGameDifficulty()
  local rankTwoScore = VendettaConfiguration.getRankTwoScore()
  local rankThreeScore = VendettaConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
  GameSession.contentTracker:setSquashData(true)
end

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function updateScoreOnCorrectResponse(self)
  GameSession.increaseScore(VendettaConfiguration.POINTS_PER_CORRECT_ANSWER)
  self.scoreLabel:setScore(GameSession.getScore())
end

local function endGame(self)
  if (GameSession:getScore() > 0) then
    self.scoreLabel:setScore(GameSession:getScore())
    self.scene:waitForGameWinOutro()
    GameSession.notifyGameWin()
  else
    self.scene:waitForGameLoseOutro()
    GameSession.notifyGameLoss()
  end
end

local function uiColorTo(self, color, duration)
  local colorAction = AHSpawn.new()
  for _, widget in ipairs(self.uiOverlay.widgets) do
    if (widget.isTimer == nil) then
      colorAction:addAction(
        AHColorTo.new(widget, duration, nil, color)
      )
    end
  end

  colorAction:start()
end

local function processCorrectResponse(self, prompt, synonym)
  Sound.playSoundEffect("correct", 0.25)
  self.responseSpeedTracker:notifyCorrectResponse()
  updateScoreOnCorrectResponse(self)

  if (self.hasAnsweredIncorrectlyForCurrentWord == false) then
    GameSession.contentTracker:notifyCorrectAnswer(synonym:getPromptIdentifier(), synonym:getWord())
    GameSession.incrementCorrectAnswerCount()
  end
  self.hasAnsweredIncorrectlyForCurrentWord = false

  self.collectedWords = self.collectedWords + 1

  self.activeSynonym = self.synonymGenerator:next()

  local notEnoughTimeForNextWord = (self.timer:getTimeRemaining() <= VendettaConfiguration.MINIMUM_TIME_NEEDED_FOR_NEW_WORD)
  local hasCollectedAllWords = (self.collectedWords == self.totalWords)
  local isRoundComplete = (hasCollectedAllWords or notEnoughTimeForNextWord)
  local correctAnimation = self.scene:getCorrectResponseAction(isRoundComplete, prompt, self.activeSynonym)
  wait(correctAnimation:start())
end

local function processIncorrectResponse(self, prompt, synonym)
  Sound.playSoundEffect("incorrect", 0.25)
  self.responseSpeedTracker:notifyIncorrectResponse()
  GameSession.contentTracker:notifyIncorrectAnswer(synonym:getPromptIdentifier(), synonym:getWord(), prompt:getIdentifier())
  GameSession.incrementIncorrectAnswerCount()

  self.hasAnsweredIncorrectlyForCurrentWord = true

  self.synonymGenerator:notifyMistake(synonym:getWord())

  self.timer:addTime(VendettaConfiguration.INCORRECT_RESPONSE_TIME_PENALTY)

  local revealRationaleAction, hideRationaleAction = self.scene:getIncorrectResponseActions(prompt)

  local fadeDuration = 0.3
  local fadedOpacity = 0.0
  local fullRevealAction = AHSpawn.new(
    AHOpacityTo.new(self.pauseButton, fadeDuration, nil, fadedOpacity),
    AHOpacityTo.new(self.scoreLabel, fadeDuration, nil, fadedOpacity),
    revealRationaleAction,
    function()
      self.timer:pauseAndShowIcon(fadeDuration)
    end
  )
  uiColorTo(self, UI_INCORRECT_COLOR, 0.5)
  wait(fullRevealAction:start())

  local tapToContinue = TapToContinue.new()
  tapToContinue:waitForTrigger()

  uiColorTo(self, UI_COMPONENTS_COLOR, 0.5)
  self.timer:resumeAndHideIcon()
  wait(hideRationaleAction:start())
end

local function createMask(self, element, width, height)
  if width == nil then
    width = element:getWidth()
    height = element:getHeight()
  end

  local paddedWidth = width + 10
  local mask = FilledRectangle.new(paddedWidth, height, MASK_COLOR)
  local x, y = self:transformLoc(element)
  x = x + width / 2
  y = y + height / 2
  mask:setLoc(x, y)

  return mask
end

local function showHelpAndWaitForTap(self, correctPrompt)
  self.timer:pauseAndShowIcon()

  local promptWidth, promptHeight = correctPrompt:getEffectiveDims()
  local getButtonMask = function() return createMask(self, correctPrompt, promptWidth, promptHeight) end

  local helpLayer = TutorialHighlighter.newWithMasks({getButtonMask})
  self.scene:insertTutorial(helpLayer)

  helpLayer:createFadeInAnimation():start()

  helpLayer:waitForTargetTap()

  wait(helpLayer:createFadeOutAnimation():start())
  helpLayer:removeSelf()

  self.timer:resumeAndHideIcon()
end

local function processResponse(self, prompt, synonym)
  self.scene:setHasUserResponded(true)

  self.numberOfResponseAnimationsInProgress = self.numberOfResponseAnimationsInProgress + 1

  local promptWord = prompt:getWord()
  local wasCorrect = synonym:isWordSynonym(promptWord)

  if (self.isShowingHelper == true) then
    self.scene:hideHelper()
    self.isShowingHelper = false
  end

  if (wasCorrect) then
    processCorrectResponse(self, prompt, synonym)
  else
    processIncorrectResponse(self, prompt, synonym)
  end

  self.numberOfResponseAnimationsInProgress = self.numberOfResponseAnimationsInProgress - 1
end

local function prepareNewRound(self)
  GameSession.contentTracker:clearSeenConcepts()

  self.round = self.roundGenerator:getNextRound()

  self.collectedWords = 0
  self.totalWords = #self.round.synonyms

  self.conceptsDataForRound = TableUtils.deepCopy(self.round.concepts)

  self.synonymGenerator = VendettaSynonymGenerator.new(self.round.synonyms)
end

local function showNewRound(self)
  local isFirstRound = true
  local firstSynonymData = self.synonymGenerator:next()
  local revealNewRoundAction = self.scene:createNewRound(firstSynonymData, self.conceptsDataForRound, isFirstRound)
  self.activeSynonymData = firstSynonymData

  local startRoundFunction = function()
    self.uiOverlay:fadeIn(0.5)
    self.timer:resume()
    self.responseSpeedTracker:notifyStart()
  end

  local fullAction = AHStaggeredSpawn.new(
    2.25,
    revealNewRoundAction,
    startRoundFunction
  )

  wait(fullAction:start())
end

local function waitForRoundEnded(self)
  waitUntil(function()
    local allWordsCollected = (self.collectedWords == self.totalWords)
    local allAnimationsDone = (self.numberOfResponseAnimationsInProgress == 0)

    return ((allWordsCollected or self.isOutOfTime) and allAnimationsDone)
  end)

  local fadeDuration = 0.3
  local fadedOpacity = 0.0
  local fadeOutUiAction = AHDelayed.new(
    0.75,
    AHOpacityTo.new(self.pauseButton, fadeDuration, nil, fadedOpacity),
    AHOpacityTo.new(self.scoreLabel, fadeDuration, nil, fadedOpacity),
    function()
      self.timer:pauseAndShowIcon(fadeDuration)
    end
  )

  fadeOutUiAction:start()
  local timeRemaining = self.timer:getTimeRemaining()
  if (timeRemaining < VendettaConfiguration.MINIMUM_TIME_NEEDED_FOR_ROUND) then
    self.isOutOfTime = true
  end

  local hideCircuitAction = self.scene:getHideCircuitAction()
  wait(hideCircuitAction:start())

  if (self.collectedWords > 0) then
    local reviewAction = self.scene:getReviewAction()
    wait(reviewAction:start())

    local tapToContinue = TapToContinue.new()
    tapToContinue:waitForTrigger()
    tapToContinue:removeSelf()
  end

  local hiddenOpacity = 0.0
  local hideUiAction = AHSpawn.new(
    AHOpacityTo.new(self.timer:getPausedLabel(), fadeDuration, nil, hiddenOpacity),
    AHOpacityTo.new(self.timer, fadeDuration, nil, hiddenOpacity)
  )

  wait(hideUiAction:start())

  self.scene:setRoundOutroAction(self.isOutOfTime)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaGame:isUIAutoFadeInEnabled()
  return false
end

function VendettaGame:gameFlow()
  self.timer:initialize()

  self.scene:startDelayedPresentHelper()
  self.isShowingHelper = true

  while (not self.isOutOfTime) do
    prepareNewRound(self)
    showNewRound(self)
    waitForRoundEnded(self)
  end

  endGame(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  local roundGenerator = VendettaRoundGenerator.new()
  self.roundGenerator = roundGenerator


  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 VendettaConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(VendettaConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       VendettaConfiguration.SPEED_BONUS_DECAY_START)

  local scene = VendettaScene.new()

  scene:setConceptPresentedCallback(function()
    self.responseSpeedTracker:notifyStart()
  end)

  scene:setOnResponseCallback(function(prompt, synonym)
    processResponse(self, prompt, synonym)
  end)

  scene:setTutorialCallback(function(correctPrompt)
      showHelpAndWaitForTap(self, correctPrompt)
    end,
    self.timer)

  self:insertAtCenter(scene, 0)
  self.scene = scene
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  self.pauseButton = builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()

  local onCompletion = function()
    self.isOutOfTime = true
  end

  self.timer = builder:addDigitTimer(GAME_DURATION, onCompletion)
end

-- Constructor ------------------------

function VendettaGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, VendettaGame)

  VendettaCommon.initialize()

  self.collectedWords = 0
  self.totalWords = 0
  self.hasAnsweredIncorrectlyForCurrentWord = false
  self.isOutOfTime = false
  self.numberOfResponseAnimationsInProgress = 0
  self.conceptsDataForRound = nil

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)

  return self
end
