-- VendettaCircuitLayer
-- Created by Steven on September 9, 2016

class("VendettaOrb")

require("graphics.CircleGradient")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local GLOW_CENTER_OPACITY = 1.0
local GLOW_EDGE_OPACITY = 0.5
local OUTER_CIRCLE_OPACITY = 0.5

local DEFAULT_COLOR = {1.0, 1.0, 1.0}

local Z_INDICES = {
  OUTER = 0,
  INNER = 5,
  GLOW = 10,
  CROSS = 15
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaOrb:getAppearAction(duration, ease)
  local appearAction = AHSpawn.new(
    AHColorTo.new(self.innerCircle, duration, ease, self.baseColor),
    AHColorTo.new(self.outerCircle, duration, ease, self.outerBaseColor)
  )

  return appearAction
end

function VendettaOrb:playCorrectAction(duration, ease)
  local scaleDownEase = AHEaseSimpleInOut.new(5)
  local revertDuration = 0.75

  local scaleAction = AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(self.outerCircle, duration, ease, 2.0),
      AHScaleTo.new(self.glow, duration, ease, 2.0),
      AHDelayed.new(duration * 5, AHScaleTo.new(self.innerCircle, duration / 2, ease, 2.0))
    ),
    AHDelay.new(0.25),
    AHSpawn.new(
      AHScaleTo.new(self.innerCircle, revertDuration * 1.1, scaleDownEase, 1.0),
      AHScaleTo.new(self.outerCircle, revertDuration * 1.5, scaleDownEase, 1.0),
      AHScaleTo.new(self.glow, revertDuration, scaleDownEase, 1.0)
    )
  )

  local colorAction = AHSequence.new(
    AHSpawn.new(
      AHColorTo.new(self.innerCircle, duration, ease, VendettaCommon.CORRECT_TEXT_COLOR),
      self.glow:newOpacityToAction(duration, ease, GLOW_CENTER_OPACITY, GLOW_EDGE_OPACITY)
    ),
    AHDelay.new(2.0),
    AHSpawn.new(
      AHColorTo.new(self.innerCircle, revertDuration / 2, ease, self.baseColor),
      AHColorTo.new(self.outerCircle, revertDuration / 2, ease, self.outerBaseColor),
      self.glow:newOpacityToAction(revertDuration, ease, 0.0, 0.0)
    )
  )

  self.scaleInterrupt:insert(scaleAction)
  self.colorAndOpacityInterrupt:insert(colorAction)
end

function VendettaOrb:playRevertSizeAction(duration, ease)
  ease = ease or AHEaseSimpleInOut.new(5)
  local scaleAction = AHSpawn.new(
    AHScaleTo.new(self.innerCircle, duration * 1.1, ease, 1.0),
    AHScaleTo.new(self.outerCircle, duration * 1.5, ease, 1.0),
    AHScaleTo.new(self.glow, duration, ease, 1.0)
  )

  self.scaleInterrupt:insert(scaleAction)
end

function VendettaOrb:playRevertColorAction(duration, ease)
  local colorAction = AHSpawn.new(
    AHColorTo.new(self.innerCircle, duration, ease, self.baseColor),
    AHColorTo.new(self.outerCircle, duration, ease, self.outerBaseColor),
    self.glow:newOpacityToAction(duration, ease, 0.0, 0.0),
    AHOpacityTo.new(self.cross, duration, ease, 0.0)
  )

  self.colorAndOpacityInterrupt:insert(colorAction)
end

function VendettaOrb:playFullRevertAction(duration, ease)
  self:playRevertColorAction(duration, ease)
  self:playRevertSizeAction(duration)
end

function VendettaOrb:playIncorrectAction(duration, ease)
  local colorAction = AHSpawn.new(
    AHOpacityTo.new(self.innerCircle, duration, ease, 0.0),
    AHOpacityTo.new(self.outerCircle, duration, ease, 0.0),
    AHOpacityTo.new(self.cross, duration, ease, 1.0),
    self.glow:newOpacityToAction(duration, ease, 0.0, 0.0)
  )

  self.colorAndOpacityInterrupt:insert(colorAction)
end

function VendettaOrb:playFadeAction(shouldHide, duration, ease)
  local innerOpacity = 1.0
  local outerOpacity = OUTER_CIRCLE_OPACITY

  if (shouldHide) then
    innerOpacity = 0.25
    outerOpacity = OUTER_CIRCLE_OPACITY / 4
  end

  local colorAction = AHSpawn.new(
    AHOpacityTo.new(self.outerCircle, duration, ease, outerOpacity),
    AHColorTo.new(self.innerCircle, duration, ease, ColorUtils.colorWithOpacity(self.baseColor, innerOpacity)),
    self.glow:newOpacityToAction(duration, ease, 0.0, 0.0),
    AHOpacityTo.new(self.cross, duration, ease, 0.0)
  )

  self.colorAndOpacityInterrupt:insert(colorAction)
end

function VendettaOrb:getPulse(pulseDuration, scaleUpEase, scaleDownEase, maxScale)
  local pulseAction = AHSpawn.new(
    AHScaleTo.new(self, pulseDuration, scaleUpEase, maxScale),
    AHDelayed.new(pulseDuration * 0.5, AHScaleTo.new(self, pulseDuration, scaleDownEase, 0.6))
  )

  return pulseAction
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self, radius)
  local innerCircle = FilledCircle.new(radius, DEFAULT_COLOR)
  self:insertAtCenter(innerCircle, Z_INDICES.INNER)
  self.innerCircle = innerCircle

  local outerRadius = radius * 1.5
  local outerColor = ColorUtils.colorWithOpacity(DEFAULT_COLOR, OUTER_CIRCLE_OPACITY)
  local outerCircle = FilledCircle.new(outerRadius, outerColor)
  self:insertAtCenter(outerCircle, Z_INDICES.OUTER)
  self.outerCircle = outerCircle

  local glowRadius = radius * 2.0
  local blurMargin = glowRadius / 2
  local glowCenterColor = ColorUtils.colorWithOpacity(self.baseColor, 0.0)
  local glowEdgeColor = ColorUtils.colorWithOpacity(self.baseColor, 0.0)
  local glow = CircleGradient.new(glowRadius, glowCenterColor, glowEdgeColor, blurMargin)
  glow:setBlendMode(MOAIProp2D.BLEND_ADD)
  self:insertAtCenter(glow, Z_INDICES.GLOW)
  self.glow = glow

  local cross = Sprite.newWithName("cross.png")
  cross:setOpacity(0.0)
  self:insertAtCenter(cross, Z_INDICES.CROSS)
  self.cross = cross
end

-- Constructor ------------------------

function VendettaOrb.new(radius, color)
  local self = Group.new(radius * 2, radius * 2)
  Mixin.mix(self, VendettaOrb)

  self.baseColor = color
  self.outerBaseColor = ColorUtils.colorWithOpacity(self.baseColor, OUTER_CIRCLE_OPACITY)
  self.scaleInterrupt = ActionInterrupt.new()
  self.colorAndOpacityInterrupt = ActionInterrupt.new()

  createLayout(self, radius)

  return self
end

function VendettaOrb.newHidden(radius, color)
  local orb = VendettaOrb.new(radius, color)
  orb:setScl(0.0)

  return orb
end
