--- VendettaRoundGenerator
-- @author Created by steven on July 14, 2016.

class("VendettaRoundGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function markRoundConceptsAsSeen(round)
  local conceptGenerator = GameSession.conceptGenerator
  for _, concept in ipairs(round.concepts) do
    local conceptIdentifier = concept.identifier
    conceptGenerator:markConceptIdentifierAsSeen(conceptIdentifier)
  end
end

-- Needed because values of TableUtils.deepCopy
local function removeConceptByPrompt(concepts, prompt)
  for index, concept in ipairs(concepts) do
    if (concept.prompt == prompt) then
      table.remove(concepts, index)
      break
    end
  end
end

local function refreshAndShufflePool(self)
  local refreshedPool = TableUtils.deepCopy(self.usedConceptBuffer)
  -- if #self.conceptPool is greater than zero, it means there were unused concepts left;
  -- want to make sure they're included in future passes
  if (#self.conceptPool > 0) then
    local remainingConcepts = TableUtils.deepCopy(self.conceptPool)
    refreshedPool = TableUtils.concatenateArrays(refreshedPool, remainingConcepts)
  end

  self.conceptPool = TableUtils.shuffleArray(refreshedPool)
  self.usedConceptBuffer = {}
end

local function setDifficultyRatios(self)
  local difficulty = GameSession.getDifficulty()
  local interpolator = KeyFrameInterpolator.new(VendettaConfiguration.getDifficultyKeyFrames(), difficulty)

  local logString = "Concept Difficulty Weights for " .. tostring(difficulty)

  local difficultyRatio = {}
  for _, difficultyBucketName in ipairs(VendettaConfiguration.DIFFICULTY_BUCKETS) do
    local weightForDifficulty = interpolator:getWeightForValue(difficultyBucketName)
    difficultyRatio[difficultyBucketName] = weightForDifficulty
    logString = logString .. "\n\t" .. tostring(difficultyBucketName) .. ": " .. tostring(weightForDifficulty)
  end

  Log.debug(logString)

  self.difficultyRatio = difficultyRatio
end

local function formatSynonyms(self, concept)
  local formattedSynonymsByDifficulty = {}

  for _, synonymData in pairs(concept.synonyms) do
    synonymData.prompt = concept.prompt
    synonymData.promptIdentifier = concept.identifier

    if (formattedSynonymsByDifficulty[synonymData.difficulty] ~= nil) then
      table.insert(formattedSynonymsByDifficulty[synonymData.difficulty], synonymData)
    else
      formattedSynonymsByDifficulty[synonymData.difficulty] = {}
    end
  end

  self.synonymsByPromptWord[concept.prompt] = formattedSynonymsByDifficulty

  concept.synonymsByDifficulty = TableUtils.deepCopy(formattedSynonymsByDifficulty)
end

local function createExclusionsSet(concept)
  local exclusionsSet = {}

  local exclusionsList = concept.exclusionsList
  if ((exclusionsList ~= nil) and (exclusionsList ~= "")) then
    local seperatorString = ", "
    local exclusionsArray = StringUtils.split(exclusionsList, seperatorString)
    exclusionsSet = TableUtils.createSetFromArray(exclusionsArray)
  end

  concept.exclusionsSet = exclusionsSet
end

local function formatSynonymsAndExclusions(self)
  for _, concept in ipairs(self.conceptPool) do
    formatSynonyms(self, concept)
    createExclusionsSet(concept)
  end
end

local function checkForDuplicateOrExclusion(self, concept)
  local conceptPrompt = concept.prompt
  local conceptId = concept.identifier

  for _, activeConcept in ipairs(self.activeConcepts) do
    -- isDuplicate can only be true after the conceptPool refreshes in the middle of selecting concepts for a round
    local isDuplicate = (activeConcept.prompt == conceptPrompt)
    local isExclusion = (activeConcept.exclusionsSet[conceptId])
    if (isDuplicate or isExclusion) then
      return true
    end
  end

  return false
end

local function grabConceptsLinearly(self, numberOfPromptsNeeded)
  local conceptPoolCount = #self.conceptPool

  -- Iterate in reverse so that we may perform removals from self.conceptPool
  -- without affecting indexes
  for conceptIndex = conceptPoolCount, 1, -1 do
    local concept = self.conceptPool[conceptIndex]

    local isDuplicateOrExclusion = checkForDuplicateOrExclusion(self, concept)

    if (not isDuplicateOrExclusion) then
      table.insert(self.activeConcepts, concept)
      table.insert(self.usedConceptBuffer, concept)
      removeConceptByPrompt(self.conceptPool, concept.prompt)

      if (#self.activeConcepts == numberOfPromptsNeeded) then
        break
      end
    end
  end
end

local function getRoundConcepts(self, numberOfPromptsNeeded)
  self.activeConcepts = {}

  grabConceptsLinearly(self, numberOfPromptsNeeded)

  if (#self.activeConcepts < numberOfPromptsNeeded) then
    refreshAndShufflePool(self)
    grabConceptsLinearly(self, numberOfPromptsNeeded)
  end

  return self.activeConcepts
end

local function getPrioritizedDifficultyProportions(self)
  local prioritizedDifficultyProportion = {}
  for difficultyName, proportion in pairs(self.difficultyRatio) do
    local difficultyProprotion = {
      difficultyName = difficultyName,
      proportion = proportion
    }

    table.insert(prioritizedDifficultyProportion, difficultyProprotion)
  end
  table.sort(prioritizedDifficultyProportion, function(a, b)
    return a.proportion > b.proportion
  end)

  return prioritizedDifficultyProportion
end

local function getTargetCountsForBuckets(self, totalNeeded)
  local targetCounts = {}
  local prioritizedDifficultyProportion = getPrioritizedDifficultyProportions(self)
  local currentCount = 0

  -- Get proportions of all but the last bucket
  for proportionIndex = 1, #prioritizedDifficultyProportion - 1 do
    local difficultyProprotion = prioritizedDifficultyProportion[proportionIndex]
    local targetCount = MathUtils.round(difficultyProprotion.proportion * totalNeeded)
    targetCounts[difficultyProprotion.difficultyName] = targetCount
    currentCount = currentCount + targetCount
  end

  -- Fill the remainder with the last bucket if there is still a shortage
  local countOfLowestPriorityBucket = (totalNeeded - currentCount)
  local nameOfLowestPriorityBucket = prioritizedDifficultyProportion[#prioritizedDifficultyProportion].difficultyName
  targetCounts[nameOfLowestPriorityBucket] = countOfLowestPriorityBucket

  return targetCounts
end

local function getSynonymsFromConceptForRound(self, concept, synonymsNeededFromConcept)
  local synonymsByDifficulty = concept.synonymsByDifficulty
  local synonymsForConcept = {}
  local remainingSynonymsNeededFromConcept = synonymsNeededFromConcept
  local targetCounts = getTargetCountsForBuckets(self, synonymsNeededFromConcept)

  for difficultyName, synonymsInDiffultyBucket in pairs(synonymsByDifficulty) do
    --- if encountering this concept more than once and it has no synonyms
    -- at a certain difficulty, refresh that table
    if (#synonymsInDiffultyBucket == 0) then
      local currentSynonyms = self.synonymsByPromptWord[concept.prompt]
      synonymsInDiffultyBucket = TableUtils.deepCopy(currentSynonyms[difficultyName])
    end

    local targetCount = targetCounts[difficultyName]
    local achievableCount = math.min(targetCount, #synonymsInDiffultyBucket, remainingSynonymsNeededFromConcept)

    for synonymIndex = 1, achievableCount do
      local doRemove = true
      local randomSynonym = TableUtils.getRandomElement(synonymsInDiffultyBucket, doRemove)
      table.insert(synonymsForConcept, randomSynonym)
    end

    remainingSynonymsNeededFromConcept = remainingSynonymsNeededFromConcept - achievableCount
  end

  --- If not enough synonyms in respective difficulty buckets, grab words from
  -- other buckets
  while (synonymsNeededFromConcept > #synonymsForConcept) do
    local tableForSelection = nil

    local difficultyPreferencesOrder = {
      VendettaConfiguration.DIFFICULTY_BUCKETS[2],
      VendettaConfiguration.DIFFICULTY_BUCKETS[1],
      VendettaConfiguration.DIFFICULTY_BUCKETS[3]
    }

    for _, difficultyBucketName in ipairs(difficultyPreferencesOrder) do
      if (TableUtils.getSize(synonymsByDifficulty[difficultyBucketName]) ~= 0) then
        tableForSelection = synonymsByDifficulty[difficultyBucketName]
        break
      end
    end
    assert((tableForSelection ~= nil), "Not enough synonyms in concept")

    local doRemove = true
    local randomSynonym = TableUtils.getRandomElement(tableForSelection, doRemove)
    table.insert(synonymsForConcept, randomSynonym)
  end

  return synonymsForConcept
end

local function getSynonymCountsForRound(conceptCount)
  local synonymCountsPerConcept = {}
  for i = 1, conceptCount do
    synonymCountsPerConcept[i] = VendettaConfiguration.SYNONYMS_TO_PROMPT_RATIO
  end
  local totalPromptCountModifier = VendettaConfiguration.getRoundPromptCountModifierForBucketCount(conceptCount)
  for i = 1, math.abs(totalPromptCountModifier) do
    local constrainedIndex = i
    while (constrainedIndex > conceptCount) do
      constrainedIndex = constrainedIndex - conceptCount
    end
    if (totalPromptCountModifier > 0) then
      synonymCountsPerConcept[constrainedIndex] = synonymCountsPerConcept[constrainedIndex] + 1
    else
      synonymCountsPerConcept[constrainedIndex] = synonymCountsPerConcept[constrainedIndex] - 1
    end
  end
  TableUtils.shuffleArray(synonymCountsPerConcept)
  return synonymCountsPerConcept
end

-- Order the synonyms in a way that feels more random than a true shuffle:
-- There are between 2 and 4 synonyms per concept. By pulling 1 or 2 for each concept,
-- and shuffling the order, we make sure all the synonyms for one concept aren't left to the end.
local function curatedShuffle(concepts, synonyms)
  local orderedSynonyms = {}
  for _, concept in ipairs(concepts) do
    local identifier = concept.identifier
    local countToPull = math.random(1, 2)
    for count = 1, countToPull do
      for i, synonym in ipairs(synonyms) do
        if (synonym.promptIdentifier == identifier) then
          table.insert(orderedSynonyms, synonym)
          table.remove(synonyms, i)
          break
        end
      end
    end
  end
  TableUtils.shuffleArray(orderedSynonyms)
  TableUtils.shuffleArray(synonyms)

  -- If these two synonyms share a prompt, we reverse the first array to guarentee that none
  -- of our prompts will have all of their synonyms back-to-back
  if (orderedSynonyms[#orderedSynonyms].promptIdentifier == synonyms[1].promptIdentifier) then
    orderedSynonyms = TableUtils.getReversedArray(orderedSynonyms)
  end
  TableUtils.concatenateArrays(orderedSynonyms, synonyms)

  -- Synonyms are presented to the user from the end of the array first, so we reverse our order.
  orderedSynonyms = TableUtils.getReversedArray(orderedSynonyms)
  return orderedSynonyms
end

local function getSynonymsForRound(self, concepts)
  local synonymCounts = getSynonymCountsForRound(#concepts)

  local synonymsForRound = {}

  for index, concept in ipairs(concepts) do
    local synonymsForConcept = getSynonymsFromConceptForRound(self, concept, synonymCounts[index])
    TableUtils.concatenateArrays(synonymsForRound, synonymsForConcept)
  end

  synonymsForRound = curatedShuffle(concepts, synonymsForRound)
  return synonymsForRound
end

local function generateRound(self, bucketCount)
  local round = {}
  local concepts = getRoundConcepts(self, bucketCount)
  local synonyms = getSynonymsForRound(self, concepts)

  round.concepts = concepts
  round.synonyms = synonyms

  return round
end

local function prebuildRounds(self)
  local rounds = {}

  local currentBucketCount = VendettaConfiguration.getInitialBucketCount()
  local maxBucketCount = VendettaConfiguration.getMaxBucketCount()

  while (currentBucketCount <= maxBucketCount) do
    local round = generateRound(self, currentBucketCount)
    table.insert(rounds, round)

    currentBucketCount = currentBucketCount + 1
  end

  self.rounds = rounds
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- A round consists of three prompts, their synonyms, and those synonyms' definitions
function VendettaRoundGenerator:getNextRound()
  local round = self.rounds[self.currentRoundIndex]

  local isOutOfPrebuiltRounds = (round == nil)
  if (isOutOfPrebuiltRounds) then
    round = generateRound(self, VendettaConfiguration.getMaxBucketCount())
  end

  self.currentRoundIndex = self.currentRoundIndex + 1

  markRoundConceptsAsSeen(round)

  return round
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function VendettaRoundGenerator.new()
  local self = {}
  Mixin.mix(self, VendettaRoundGenerator)

  local conceptGenerator = GameSession.conceptGenerator
  conceptGenerator:setMarkGeneratedConceptsAsSeen(false)

  self.currentRoundIndex = 1
  self.conceptPool = conceptGenerator:nextAsArray(VendettaConfiguration.CONCEPTS_COUNT_PER_GAME)
  self.usedConceptBuffer = {}
  self.synonymsByPromptWord = {}

  setDifficultyRatios(self)
  formatSynonymsAndExclusions(self)
  prebuildRounds(self)

  return self
end
