--- VendettaScene
-- @author Created by steven on July 11, 2016.

class("VendettaScene")

require("VendettaPrompt")
require("VendettaSynonym")
require("VendettaRationale")
require("VendettaCircuitLayer")
require("VendettaExplosion")
require("VendettaReviewOverlay")
require("VendettaRoundTransitionLine")
require("graphics.primitives.Checkmark")
require("tutorials.TutorialHighlighter")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_TOP_COLOR = {0.25, 0.10, 0.25}
local BACKGROUND_BOTTOM_COLOR = {0.25, 0.10, 0.25}
local BACKGROUND_INCORRECT_TOP_COLOR = {0.3, 0.0, 0.15}
local BACKGROUND_INCORRECT_BOTTOM_COLOR = {0.22, 0.0, 0.11}

local GLOW_CENTER_COLOR = {0.68, 0.25, 0.27, 0.5}
local GLOW_EDGE_COLOR = {0.25, 0.10, 0.25, 0.0}
local GLOW_RATIONALE_CENTER_COLOR = {0.4, 0.06, 0.25}
local GLOW_RATIONALE_EDGE_COLOR = {0.5, 0.5, 0.5, 0.0}

local Z_INDICES = {
  CIRCUIT = 5,
  SHADOW = 10,
  TUTORIAL = 20,
  UI = 30
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function scalePromptsIfNeeded(self)
  local minimumScale = 1.0
  for index, prompt in ipairs(self.prompts) do
    local promptScale = prompt:getScl()
    minimumScale = math.min(promptScale, minimumScale)
  end

  if (minimumScale < 1.0) then
    local layoutIndex = #self.prompts - 1
    local promptLocations = VendettaCommon.ROUND_LAYOUTS[layoutIndex]
    for index, prompt in ipairs(self.prompts) do
      prompt:setScl(minimumScale)
      local point = promptLocations[index]
      local offsetYAfterScaling = point.y - prompt:getEffectiveHeightPart(1.5)
      prompt:setY(offsetYAfterScaling)
    end
  end
end

local function createPrompt(self, promptData)
  local prompt = VendettaPrompt.new(promptData)
  prompt:setOpacity(0.0)
  table.insert(self.prompts, prompt)

  return prompt
end

local function createPrompts(self, promptsData)
  local layoutIndex = #promptsData - 1
  local promptLocations = VendettaCommon.ROUND_LAYOUTS[layoutIndex]

  for index, promptData in ipairs(promptsData) do
    local prompt = createPrompt(self, promptData)
    local point = promptLocations[index]
    prompt:setLoc(point.x, point.y - prompt:getHeightPart(1.5))
    self:insert(prompt, Z_INDICES.CIRCUIT)
  end
end

local function createSynonym(self, synonymData, isFirstRound)
  if (isFirstRound == nil) then
    isFirstRound = false
  end

  local shouldStackLetters = (not isFirstRound)
  local synonym = VendettaSynonym.new(synonymData, shouldStackLetters)
  synonym:setOpacity(0.0)
  synonym:setLoc(Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y)
  self:insert(synonym, Z_INDICES.UI)
  table.insert(self.synonyms, synonym)

  return synonym
end

local function createGameOutroAction(self)
  local destinationX = Screen.getWidthPart(0.5)
  local destinationY = VendettaCommon.SYNONYM_ORB_Y

  local orbActions = AHSpawn.new()
  local halfWidth = Screen.getWidthPart(0.5)
  local durationOut = 0.7
  local durationBack = 0.6
  local animateBackDelay = durationOut * 0.5
  local randomFunction = math.random

  local orbs = self.circuitLayer:getOrbs()
  for _, orb in ipairs(orbs) do
    -- move out and back
    local x = orb:getX()
    local y = orb:getY()
    local moveLeft = (x < halfWidth)
    local moveDown = (y < destinationY)
    local randomX = randomFunction(4, 15) / 100.0
    local randomY = randomFunction(4, 15) / 100.0
    local dx = Screen.getWidthPart(randomX) * (moveLeft and -1.0 or 1.0)
    local dy = Screen.getHeightPart(randomY) * (moveDown and -1.0 or 1.0)
    local moveOut = AHLocationBy.new(orb, durationOut, AHEaseExponentialOut.new(), dx, dy)
    local moveBack = AHLocationTo.new(orb, durationBack, AHEaseExponentialIn.new(), destinationX, destinationY)

    -- scale up/down on the way out, then down on the way back
    local minScale = randomFunction(6, 8) / 10.0
    local maxScale = randomFunction(11, 14) / 10.0
    local destinationScale = (randomFunction(1, 2) == 1 and minScale or maxScale)
    local scaleOut = AHScaleTo.new(orb, durationOut, AHEaseExponentialOut.new(), destinationScale)
    local scaleBack = AHScaleTo.new(orb, durationBack, AHEaseExponentialIn.new(), 0.3)

    local animateOut = AHSpawn.new(moveOut, scaleOut)
    local animateBack = AHSpawn.new(moveBack, scaleBack)
    local disappear = function()
      orb:removeSelf()
    end

    local delayedAnimateBack = AHDelayed.new(animateBackDelay, animateBack)
    local motionSpawn = AHSpawn.new(animateOut, delayedAnimateBack)

    local initialDelay = randomFunction(1, 10) / 100.0
    local stagger = AHDelay.new(initialDelay)

    local sequence = AHSequence.new(stagger,motionSpawn, disappear)
    orbActions:addAction(sequence)
  end

  -- delayed explosion
  local explosion = VendettaExplosion.new()
  local explosionDelay = (durationOut + durationBack - animateBackDelay) * 1.025

  local shakeDelta = Screen.getWidthPart(0.01)
  local screenShakeAction = AHSequence.new(
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), -shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), -shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), shakeDelta)
  )

  local delayedExplosion = AHDelayed.new(
    explosionDelay,
    function()
      local explosionZIndex = self.circuitLayer:getTopZIndex() + 1
      explosion:setLoc(Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y)
      self:insert(explosion, explosionZIndex)
      explosion:start()
    end,
    screenShakeAction
  )

  -- delayed checkmark
  local width = Screen.getWidthPart(0.4)
  local shortSegmentProportion = 0.425
  local thicknessProportion = 0.25
  local checkmarkInitalColor = {1.0, 1.0, 1.0, 1.0}
  local checkmarkFinalColor = {0.78, 0.29, 0.45, 1.0}
  local lineWidth = Screen.getWidthPart(0.002)
  local blurMargin = Screen.getWidthPart(0.003)
  local haloBlurMargin = blurMargin * 1.8
  local checkmark = HollowCheckmark.new(width, shortSegmentProportion, thicknessProportion, checkmarkInitalColor, lineWidth, blurMargin)
  local checkmarkHalo = HollowCheckmark.new(width, shortSegmentProportion, thicknessProportion, checkmarkFinalColor, lineWidth, haloBlurMargin)

  local initialCheckmarkScale = 0.05
  checkmark:setScl(initialCheckmarkScale)

  local scaleUpDuration = 1.2
  local checkmarkScaleUp = AHScaleTo.new(checkmark, scaleUpDuration, AHEaseBounceOut.new(), 1.0)

  local haloDestinationAngle = 16.0
  checkmarkHalo:setOpacity(0.2)
  checkmarkHalo:setScl(1.05)

  -- halo scales up, rotates, and fades out
  local haloFadeOutDuration = scaleUpDuration * 2.6
  local checkmarkHaloRotate = AHRotationBy.new(checkmarkHalo, haloFadeOutDuration, AHEaseExponentialOut.new(), haloDestinationAngle)
  local checkmarkHaloFadeOut = AHOpacityTo.new(checkmarkHalo, haloFadeOutDuration, AHEaseExponentialOut.new(), 0.0)
  local checkmarkHaloScaleUp = AHScaleTo.new(checkmarkHalo, haloFadeOutDuration, AHEaseExponentialOut.new(), 1.3)

  local checkmarkAngle = 180.0
  local haloInitialAngle = 13.0
  checkmark:setRot(-checkmarkAngle)
  checkmarkHalo:setRot(-haloInitialAngle)

  local checkmarkRotate = AHRotationBy.new(checkmark, scaleUpDuration, AHEaseExponentialOut.new(), checkmarkAngle)

  local colorChangeDelay = scaleUpDuration * 0.5
  local colorChangeDuration = scaleUpDuration - colorChangeDelay
  local checkmarkColorChange = AHColorTo.new(checkmark, colorChangeDuration, nil, checkmarkFinalColor)
  local delayedColorChange = AHDelayed.new(colorChangeDelay, checkmarkColorChange)

  local checkmarkDelay = 2.0
  local delayedCheckmark = AHDelayed.new(checkmarkDelay, function()
    checkmark:setLoc(destinationX, destinationY)
    self:insert(checkmark)

    checkmarkScaleUp:start()
    checkmarkRotate:start()
    delayedColorChange:start()

    local haloDelay = 0.425
    local delayedHalo = AHDelayed.new(haloDelay, function()
      checkmarkHalo:setLoc(destinationX, destinationY)
      self:insert(checkmarkHalo)
      checkmarkHaloRotate:start()
      checkmarkHaloFadeOut:start()
      checkmarkHaloScaleUp:start()
    end)

    delayedHalo:start()
  end)


  local outroAnimations = AHSpawn.new(orbActions, delayedExplosion, delayedCheckmark)
  local letAnimationFinish = AHDelay.new(3.0)

  return AHSequence.new(outroAnimations, letAnimationFinish)
end

local function getClearReviewAction(self)
  local hideTextAction = AHSpawn.new()

  for _, prompt in ipairs(self.prompts) do
    hideTextAction:addAction(
      AHSequence.new(
        prompt:createHideAction(0.1),
        function() prompt:removeSelf() end
      )
    )
  end

  return hideTextAction
end

local function getRevealReviewAction(self)
  local topHeight = Screen.getHeightPart(0.625)
  local largestColumnHeight = 0
  local shiftPromptsAction = AHSpawn.new()
  local revealSynonymsAction = AHSpawn.new()

  if (self.numberOfPrompts > 2) then
    local firstPrompt = self.prompts[1]
    local lastPrompt = self.prompts[self.numberOfPrompts]
    largestColumnHeight = math.max(firstPrompt:getSynonymColumnHeight(), lastPrompt:getSynonymColumnHeight())
  end

  local layoutIndex = self.numberOfPrompts - 1
  local promptLocations = VendettaCommon.REVIEW_LAYOUTS[layoutIndex]

  for index, prompt in ipairs(self.prompts) do
    local point = promptLocations[index]
    local orb = self.circuitLayer:getOrb(prompt)

    local destinationY = 0.0
    if (point.y == 0) then
      destinationY = topHeight - largestColumnHeight - prompt:getHeightPart(4.0)
    else
      destinationY = point.y
    end

    local shiftEase = AHEaseSimpleInOut.new(5)
    local shiftDuration = 0.75
    local individualShiftAction = AHSpawn.new(
      AHLocationTo.new(prompt, shiftDuration, shiftEase, point.x, destinationY),
      AHLocationTo.new(orb, shiftDuration, shiftEase, point.x, destinationY + prompt:getHeightPart(1.5))
    )

    shiftPromptsAction:addAction(individualShiftAction)
    revealSynonymsAction:addAction(
      prompt:createRevealSynonymsAction(0.5)
    )
  end

  local reviewOverlay = VendettaReviewOverlay.new(self.numberOfPrompts)
  self:insertAtCenter(reviewOverlay)
  self.reviewOverlay = reviewOverlay

  local reviewAction = AHSpawn.new(
    shiftPromptsAction,
    revealSynonymsAction,
    reviewOverlay:createFadeAction(1.0, 0.5)
  )

  return reviewAction
end

local function createCircuits(self, numberOfPrompts)
  local circuitLayer = VendettaCircuitLayer.new(numberOfPrompts)
  self:insertAtCenter(circuitLayer, Z_INDICES.CIRCUIT)

  local onTap = function(prompt)
    runInCoroutine(function()
      self.onResponseCallback(prompt, self.activeSynonym)
    end)
  end

  circuitLayer:setTapCallback(onTap)

  return circuitLayer
end

local function createRevealSynonymAction(self)
  local synonym = self.activeSynonym
  local duration = 0.5

  return AHDelayed.new(
           0.2,
           AHStaggeredSpawn.new(
             0.2,
             AHLocationYBy.new(synonym, duration, AHEaseSineOut.new(), synonym:getHeight()),
             AHOpacityTo.new(synonym, duration, AHEaseSineOut.new(), 1.0)
           )
         )
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaScene:createNewRound(firstSynonymData, promptsData, isFirstRound)
  local numberOfPrompts = #promptsData
  local oldCircuitLayer = self.circuitLayer

  local nextCircuit = createCircuits(self, numberOfPrompts)
  nextCircuit:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(-1.5))
  self.circuitLayer = nextCircuit

  self.prompts = {}
  createPrompts(self, promptsData)
  scalePromptsIfNeeded(self)

  self.activeSynonym = createSynonym(self, firstSynonymData, isFirstRound)

  nextCircuit:setPrompts(self.prompts)

  local shiftDistance = Screen.getHeightPart(2.0)

  local panDuration = 1.5
  local panEase = AHEaseSineInOut.new()
  local roundStartStagger = 2.75

  local shakeDelta = Screen.getWidthPart(0.01)
  local screenShakeAction = AHSequence.new(
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), -shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), -shakeDelta),
    AHLocationXBy.new(self, 0.1, AHEaseSineInOut.new(), shakeDelta)
  )

  local panAction = AHSpawn.new(
    self.roundOutroAction,
    AHLocationYBy.new(oldCircuitLayer, panDuration, panEase, shiftDistance),
    AHDelayed.new(
      1.3,
      function()
        local explosion = VendettaExplosion.new()
        nextCircuit:addExplosion(explosion)
        explosion:start()
      end,
      screenShakeAction
    ),
    AHLocationYBy.new(nextCircuit, panDuration, panEase, shiftDistance)
  )

  local transitionAction = AHStaggeredSpawn.new(
    roundStartStagger,
    panAction,
    AHSpawn.new(nextCircuit:createRoundIntro(isFirstRound), createRevealSynonymAction(self)),
    function()
      self.numberOfPrompts = numberOfPrompts

      if (oldCircuitLayer ~= nil) then
        oldCircuitLayer:removeSelf()
      end
    end
  )

  local shouldShowTutorial = isFirstRound

  if shouldShowTutorial and (not self.hasUserResponded) then
    runInCoroutine(function()
      waitUntil(function()
        return (self.timer:getTime() > VendettaConfiguration.TUTORIAL_DELAY)
      end)
      if shouldShowTutorial and (not self.hasUserResponded) then
        self.circuitLayer:setTouchEnabled(false)

        local correctPrompt = nil
        for _, prompt in ipairs(self.prompts) do
          local promptWord = prompt:getWord()
          local isCorrect = self.activeSynonym:isWordSynonym(promptWord)
          if isCorrect then
            correctPrompt = prompt
            break
          end
        end

        self.tutorialCallback(correctPrompt)

        self.circuitLayer:setTouchEnabled(true)

        local holeButton = self.circuitLayer.buttons[correctPrompt]
        holeButton.touchHandler()
      end
    end)
  end

  return transitionAction
end

function VendettaScene:getHideCircuitAction()
  local duration = 0.5
  self.circuitLayer:setTouchEnabled(false)

  local hideElementsAction = AHSpawn.new(
    self.circuitLayer:getHideForReviewAction(duration),
    self.circuitLayer:getReshowInvalidPromptsAction(duration, true)
  )

  if (self.activeSynonym ~= nil) then
    hideElementsAction:addAction(
      AHOpacityTo.new(self.activeSynonym, duration, nil, 0.0)
    )
  end

  return hideElementsAction
end

function VendettaScene:getReviewAction()
  local reviewAction = getRevealReviewAction(self)

  return reviewAction
end

function VendettaScene:setRoundOutroAction(isOutOfTime)
  -- NOTE: Hide computation time behind review screen
  local transitionAction = AHSpawn.new()

  if (self.prompts ~= nil) then
    transitionAction:addAction(getClearReviewAction(self))
  end

  if (not isOutOfTime) then
    local roundOutro, transitionLines = self.circuitLayer:createRoundOutro()
    local transitionLinesAction = AHSpawn.new(
      roundOutro
    )

    for _, transitionLine in ipairs(transitionLines) do
      transitionLine:setCircuitLayer(self.circuitLayer)
      self:insertAtCenter(transitionLine)

      local lineAction = AHSequence.new(
        transitionLine:getAppearAction(2.5, AHEaseSineInOut.new()),
        AHRemove.new(transitionLine)
      )

      transitionLinesAction:addAction(lineAction)
    end

    transitionAction:addAction(transitionLinesAction)
    transitionAction:addAction(
      AHDelayed.new(0.25, function() Sound.playSoundEffect("explosion", 0.5) end)
    )
  else
    transitionAction:addAction(createGameOutroAction(self))
    transitionAction:addAction(function() Sound.playSoundEffect("outro", 0.5) end)
  end

  if (self.reviewOverlay ~= nil) then
    self.reviewOverlay:createFadeAction(0.0, 0.1):start()
  end

  self.roundOutroAction = transitionAction
end

function VendettaScene:waitForGameWinOutro()
  wait(self.roundOutroAction:start())
end

function VendettaScene:waitForGameLoseOutro()
  local outroLoseDuration = 0.5
  local outroLoseAction = AHSpawn.new()

  if (self.interactionHelper) then
    local hideAction = AHSequence.new(
      AHOpacityTo.new(self.interactionHelper, outroLoseDuration, nil, 0.0),
      AHRemove.new(self.interactionHelper)
    )
    outroLoseAction:addAction(hideAction)
  end

  local orbs = self.circuitLayer:getOrbs()
  for _, orb in ipairs(orbs) do
    local hideOrb = AHScaleTo.new(orb, outroLoseDuration, nil, 0.0)
    outroLoseAction:addAction(hideOrb)
  end

  for _, prompt in ipairs(self.prompts) do
    local hidePrompt = AHScaleTo.new(prompt, outroLoseDuration, nil, 0.0)
    outroLoseAction:addAction(hidePrompt)
  end

  wait(outroLoseAction:start())
end

function VendettaScene:setConceptPresentedCallback(callback)
  self.conceptPresentedCallback = callback
end

function VendettaScene:startDelayedPresentHelper()
  local presentAction = AHDelayed.new(3.1, AHOpacityTo.new(self.interactionHelper, 1.0, nil, 1.0))

  return presentAction:start()
end

function VendettaScene:hideHelper()
  local hideAction = AHSequence.new(
    AHOpacityTo.new(self.interactionHelper, 0.1, nil, 0.0),
    AHRemove.new(self.interactionHelper)
  )

  return hideAction:start()
end

function VendettaScene:getCorrectResponseAction(isRoundComplete, prompt, nextSynonymData)
  self.shouldShowTutorial = false
  local oldSynonym = self.activeSynonym

  if (nextSynonymData ~= nil) then
    self.activeSynonym = createSynonym(self, nextSynonymData)
  else
    self.activeSynonym = nil
  end

  local lineBounce, beamAction = self.circuitLayer:createCorrectAction(prompt)

  local duration = 0.25
  local showNextSynonym = AHDoNothing.new()
  if ((self.activeSynonym ~= nil) and (not isRoundComplete)) then
    showNextSynonym = AHStaggeredSpawn.new(
      duration / 4,
      AHOpacityTo.new(self.activeSynonym, duration, nil, 1.0),
      AHSpawn.new(
        AHLocationYBy.new(self.activeSynonym, duration * 6.0, AHEaseElasticOut.new(0.7), self.activeSynonym:getHeight()),
        self.activeSynonym:getAppearAction(duration * 3.0, AHEaseSineOut.new()),
        function()
          Sound.playSoundEffect("word_intro", 0.1)
        end
      ),
      function()
        self.conceptPresentedCallback()
        self.circuitLayer:setTouchEnabled(true)
      end
    )
  end

  local hideSynonymAction = AHSequence.new(
    oldSynonym:getHideAction(duration * 4, AHEaseExponentialIn.new()),
    function()
      prompt:takeSynonymIfUnique(oldSynonym)
    end
  )

  local explosionAction = AHSequence.new(
    AHDelay.new(0.1),
    function()
      local explosion = VendettaExplosion.new(0.95, true)
      explosion:setScl(0.7)
      self.circuitLayer:addExplosion(explosion)
      explosion:start()
      waitFor(1.1)
      beamAction:start()
      Sound.playSoundEffect("whoosh_animation", 0.25)
    end
  )

  local fullResponseAction = AHSpawn.new(
    explosionAction,
    hideSynonymAction,
    AHDelayed.new(
      1.2,
      showNextSynonym,
      self.circuitLayer:getReshowInvalidPromptsAction(duration, isRoundComplete)
    ),
    lineBounce
  )

  return fullResponseAction
end

function VendettaScene:setHasUserResponded(hasUserResponded)
  self.hasUserResponded = hasUserResponded
end

function VendettaScene:getIncorrectResponseActions(prompt)
  local ease = nil
  local rationale = VendettaRationale.new(self.activeSynonym)
  self:insertAtCenter(rationale)

  local synonymShakeAction = VendettaCommon.createShakeAction(self.activeSynonym)
  local dimUnselectedPrompts, revertUnselectedPrompts = self.circuitLayer:createIncorrectAction(prompt)

  local duration = 0.5
  local revealRationaleAction = AHSpawn.new(
    synonymShakeAction,
    self.glow:newGradientToAction(duration, ease, GLOW_RATIONALE_CENTER_COLOR, GLOW_RATIONALE_EDGE_COLOR),
    dimUnselectedPrompts,
    rationale:createRevealAction(duration),
    self.background:newGradientToAction(duration, ease, BACKGROUND_INCORRECT_TOP_COLOR, BACKGROUND_INCORRECT_BOTTOM_COLOR)
  )

  local hideRationaleAction = AHSpawn.new(
    self.glow:newGradientToAction(duration, ease, GLOW_CENTER_COLOR, GLOW_EDGE_COLOR),
    rationale:createHideAction(duration),
    self.background:newGradientToAction(duration, ease, BACKGROUND_TOP_COLOR, BACKGROUND_BOTTOM_COLOR),
    revertUnselectedPrompts,
    self.circuitLayer:getDisableIncorrectButtonAction(prompt),
    function()
      waitFor(duration)
      self.circuitLayer:enableValidButtons()
      rationale:removeSelf()
    end
  )

  return revealRationaleAction, hideRationaleAction
end

function VendettaScene:setOnResponseCallback(callback)
  self.onResponseCallback = callback
end

function VendettaScene:setTutorialCallback(callback, timer)
  self.tutorialCallback = callback
  self.timer = timer
end

function VendettaScene:insertTutorial(tutorial)
  self:insertAtCenter(tutorial, Z_INDICES.TUTORIAL)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createLayout(self)
  local width = Screen.width
  local background = RectangleGradient.new(width, Screen.height, BACKGROUND_TOP_COLOR, BACKGROUND_BOTTOM_COLOR)
  self:insertAtCenter(background)
  self.background = background

  local glowRadius = Screen.getWidthPart(0.5)
  local blurMargin = Screen.getWidthPart(0.1)
  local glow = CircleGradient.new(glowRadius, GLOW_CENTER_COLOR, GLOW_EDGE_COLOR, blurMargin)
  glow:setLoc(Screen.getWidthPart(0.5), VendettaCommon.SYNONYM_ORB_Y)
  self:insert(glow)
  self.glow = glow

  local topGradientColor = {0.152, 0.062, 0.156, 0.75}
  local bottomGradientColor = {0.152, 0.062, 0.156, 0.0}
  local gradientHeight = Screen.getHeightPart(0.2)
  local topShadow = RectangleGradient.new(width, gradientHeight, topGradientColor, bottomGradientColor)
  topShadow:setAnchorPointTop()
  topShadow:setLoc(Screen.getWidthPart(0.5), Screen.height)
  self:insert(topShadow, Z_INDICES.SHADOW)

  local bottomShadow = RectangleGradient.new(width, gradientHeight, bottomGradientColor, topGradientColor)
  bottomShadow:setAnchorPointBottom()
  bottomShadow:setX(Screen.getWidthPart(0.5))
  self:insert(bottomShadow, Z_INDICES.SHADOW)

  local interactionHelper = Label.new("TAP THE MATCHING VERB", Font.new(FontFaces.DIN_OT_MEDIUM, 14))
  interactionHelper:setColor(unpack({0.56, 0.36, 0.49}))
  interactionHelper:setLoc(self:getWidthPart(0.5), Screen.getSafeY(0.15))
  interactionHelper:setOpacity(0)
  self:insert(interactionHelper, Z_INDICES.UI)
  self.interactionHelper = interactionHelper

  local introCircuits = VendettaCircuitLayer.new(0)
  self:insertAtCenter(introCircuits, Z_INDICES.CIRCUIT)
  self.circuitLayer = introCircuits
  self:setRoundOutroAction()
end

-- Constructor ------------------------

function VendettaScene.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, VendettaScene)

  self.hasUserResponded = false
  self.synonyms = {}

  createLayout(self)

  return self
end
