--- VendettaValidator
-- @author Created by steven on July 07, 2016.
-- ConceptValidator for Vendetta.

class("VendettaValidator")

require("validators.ConceptValidator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local CONCEPT_REQUIRED_KEYS = {
  'prompt',
  'synonyms'
}

local CONCEPT_OPTIONAL_KEYS = {
  'exclusionsList'
}

local SYNONYM_REQUIRED_KEYS = {
  "difficulty",
  "word",
  "rationale"
}

local MINIMUM_SYNONYM_COUNT_PER_DIFFICULTY = 4
local MAXIMUM_PROMPT_LENGTH = 10
local MAXIMUM_SYNONYM_LENGTH = 14
local MAXIMUM_RATIONALE_LENGTH = 40

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateSynonymRequiredKeys(self, synonym)
  local isOptional = false
  self:validateRequiredKeys(synonym, SYNONYM_REQUIRED_KEYS, isOptional)
end

local function validateNoDuplicateSynonyms(self, synonyms)
  local duplicateItems = TableUtils.getDuplicatesInArray(synonyms)

  self:expectEqual(#duplicateItems, 0, "Duplicate synonyms found for concept.")
end

local function validateMinimumSynonymsCount(self, synonyms)
  local easyCount = 0
  local mediumCount = 0
  local hardCount = 0

  for _, synonym in ipairs(synonyms) do
    if (synonym.difficulty == "Easy") then
      easyCount = easyCount + 1
    elseif (synonym.difficulty == "Medium") then
      mediumCount = mediumCount + 1
    elseif (synonym.difficulty == "Hard") then
      hardCount = hardCount + 1
    else
      self:expectTrue(false, string.format("Synonym %s has a misspelled difficulty tag.", synonym.word))
    end
  end

  self:expectGreaterOrEqual(easyCount, MINIMUM_SYNONYM_COUNT_PER_DIFFICULTY, "Concept does not have enough easy synonyms.")
  self:expectGreaterOrEqual(mediumCount, MINIMUM_SYNONYM_COUNT_PER_DIFFICULTY, "Concept does not have enough medium synonyms.")
  self:expectGreaterOrEqual(hardCount, MINIMUM_SYNONYM_COUNT_PER_DIFFICULTY, "Concept does not have enough hard synonyms.")
end

local function validatePromptLength(self, concept)
  local prompt = concept.prompt
  local promptLength = string.len(prompt)

  self:expectLessOrEqual(promptLength, MAXIMUM_PROMPT_LENGTH, string.format("Prompt %s has more than 10 characters.", prompt))
end

local function validateSynonymLength(self, synonymWord)
  local synonymLength = string.len(synonymWord)

  self:expectLessOrEqual(synonymLength, MAXIMUM_SYNONYM_LENGTH, string.format("Synonym %s has more than 14 characters.", synonymWord))
end

local function validateRationale(self, synonymData)
  local rationale = synonymData.rationale
  self:expectTrue(((rationale ~= nil) and (rationale ~= "")), string.format("Synonym $s does not have a rationale.", synonymData.word))

  if (rationale ~= nil) then
    local rationaleLength = string.len(rationale)
    self:expectLessOrEqual(rationaleLength, MAXIMUM_RATIONALE_LENGTH, string.format("Rationale for %s has more than 40 characters.", synonymData.word))
  end
end


-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function VendettaValidator:validateConcept(concept)
  local synonyms = concept.synonyms

  local isOptional = false
  self:validateRequiredKeys(concept, CONCEPT_REQUIRED_KEYS, isOptional)

  isOptional = true
  self:validateRequiredKeys(concept, CONCEPT_OPTIONAL_KEYS, isOptional)

  validatePromptLength(self, concept)
  validateNoDuplicateSynonyms(self, synonyms)
  validateMinimumSynonymsCount(self, synonyms)

  for _, synonymData in ipairs(synonyms) do
    local synonymWord = synonymData.word

    validateSynonymRequiredKeys(self, synonymData)
    validateSynonymLength(self, synonymWord)
    validateRationale(self, synonymData)
  end
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function VendettaValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, VendettaValidator)

  return self
end
