--- WitGame
-- @author Created by alexeychernikov on July 15, 2014.


class("WitGame")

-- NOTE: Needed to fix sizes for iPad
FontScaling.setAutoScaling(Screen.fullWidth, Screen.fullHeight)

require("StandardUIBuilder")
require("WitConfiguration")
require("scene.WitScene")

require("data.WitRoundGenerator")
require("data.WitDictionary")
require("data.DictionaryWordsMatcher")
require("scene.widgets.WitCountdownTimer")
require("WitConstants")

require("WitTutorial")
require("WitAnalytics")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local BACKGROUND_COLOR = {0.3, 0, 0.6}
local UI_COMPONENTS_COLOR = {1, 1, 1}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getTimeoutHandler(self)
  return function()
    self.roundLock:stop()
  end
end

local function incrementCounter(self)
  self.currentWordCount = self.currentWordCount + 1
  local newCountString = string.format("%d/%d", self.currentWordCount, WitConfiguration.getCorrectResponsesNeededToPass())
  self.counterLabel:setText(newCountString)
end

local function hideUiOverlay(self, duration)
  local duration = duration or 0.5
  local isRecursive = true
  self.counterLabel:linkOpacityAttrToChildren(isRecursive)
  self.uiOverlay:createFadeOutAction(duration):start()
end

local function getCorrectAnswerHandler(self)
  return function(word, originalString, isAdvanced)
    incrementCounter(self)
    if(isAdvanced) then
      Sound.playSoundEffect("wit_advancedword")
      GameSession.increaseScore(WitConfiguration.POINTS_PER_ADVANCED_RESPONSE)

      local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.ADVANCED_RESPONSE_BONUS)
      self.bonusTray:indicateSimpleTextBonus(notificationString, WitConfiguration.POINTS_PER_ADVANCED_RESPONSE)
      GameSession.bonusReporter:incrementBonusCount(BonusTypes.TYPES.ADVANCED_RESPONSE_BONUS)
    else
      Sound.playSoundEffect("wit_correct")
    end

    self.correctAnswers = self.correctAnswers + 1
    self.correctAnswersForCurrentRound = self.correctAnswersForCurrentRound + 1

    if(self.correctAnswers <= WitConfiguration.getCorrectResponsesNeededToPass()) then
      GameSession.increaseScore(WitConfiguration.POINTS_PER_CORRECT_RESPONSE)
    else
      GameSession.increaseScore(WitConfiguration.POINTS_PER_EXTRA_CORRECT_RESPONSE)

      local notificationString = BonusTypes.getNotificationString(BonusTypes.TYPES.EXTRA_COLLECTED_BONUS)
      self.bonusTray:indicateSimpleTextBonus(notificationString, WitConfiguration.POINTS_PER_EXTRA_CORRECT_RESPONSE)
    end

    self.scoreLabel:setScore(GameSession.getScore())

    GameSession.contentTracker:notifyCorrectAnswer(self.round.identifier, nil, originalString)
    GameSession.incrementCorrectAnswerCount()
  end
end

local function getIncorrectAnswerHandler(self)
  return function(word)
    GameSession.contentTracker:notifyIncorrectAnswer(self.round.identifier, nil, word)
    GameSession.incrementIncorrectAnswerCount()
    Sound.playSoundEffect("wit_incorrect")
  end
end

local function getPassRoundCallback(self)
  return function()
    WitAnalytics.roundEndedEarly(self.correctAnswersForCurrentRound)
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

local function doRound(self)
  Sound.playSoundEffect("wit_roundstart")
  self.correctAnswersForCurrentRound = 0
  self.round = self.roundGenerator:generateNextValue(self.roundNumber)
  GameSession.reportSeenConcept(self.round.identifier)
  GameSession.contentTracker:notifyPresentedConcept(self.round.identifier)
  self.scene:setRound(self.round, self.roundNumber)
  -- Timer cancels the round lock
  wait(self.roundLock:start())
  self.scene:checkForCorrectAnswerOnRoundTimeout()
  WitAnalytics.roundTimedOut(self.correctAnswersForCurrentRound)
  Sound.playSoundEffect("wit_roundend")
end

local function reportBonuses(self)
  -- TODO
end

local function waitForRoundInterstitial(self, roundStats)
  self.scene:stopAnimations()
  wait(self.scene:createRoundInterstitialAction(roundStats):start())
end

local function buildWitScene(self)
  local scene = WitScene.new(self.timer, getPassRoundCallback(self))
  self:insertAtCenter(scene)

  self.scene = scene
  self.correctAnswers = 0
  scene:setCorrectAnswerHandler(getCorrectAnswerHandler(self))
  scene:setIncorrectAnswerHandler(getIncorrectAnswerHandler(self))

  self.intro = self.scene:doIntro()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitGame:isUIAutoFadeInEnabled()
  return (self.tutorial == nil)
end

function WitGame:gameFlow()
  if (self.tutorial ~= nil) then
    wait(self.tutorial:playTutorial())

    buildWitScene(self)

    local uiOverlay = self:getUIOverlay()
    uiOverlay:fadeIn()

    --Wait a frame to prevent the screen from flickering when the background is removed.
    waitForFrames(1)

    self.tutorial:removeSelf()
    self.tutorial = nil
  end

  wait(AHDelay.new(self.intro:getDuration()))

  local roundsLeft = self.setsToComplete
  self.roundLock = AHDelay.new(math.huge)
  self.roundNumber = 1
  repeat
    doRound(self)
    roundsLeft = roundsLeft - 1

    local roundStats =
    {
      passedRoundNumber = self.roundNumber,
      wordsForPassedRound = self.correctAnswersForCurrentRound,
      wordsForAllRounds = self.correctAnswers,
      goalWordCount = WitConfiguration.getCorrectResponsesNeededToPass()
    }

    waitForRoundInterstitial(self, roundStats)

    self.roundNumber = self.roundNumber + 1
  until(roundsLeft <= 1)

  -- Do an outro on the last round
  doRound(self)
  local didWin = (self.currentWordCount >= WitConfiguration.getCorrectResponsesNeededToPass())
  WitAnalytics.gameEnded(didWin)
  if (didWin) then
    Sound.playSoundEffect("wit_victory")
    hideUiOverlay(self)
    wait(self.scene:createWinOutroAction():start())
    reportBonuses(self)

    GameSession.notifyGameWin()
  else
    hideUiOverlay(self)
    wait(self.scene:createLoseOutroAction():start())
    GameSession.notifyGameLoss()
  end
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function configureGameDifficulty(self)
  self.setsToComplete = WitConfiguration.ROUNDS_PER_GAME_COUNT

  local rankTwoScore = WitConfiguration.getRankTwoScore()
  local rankThreeScore = WitConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rankTwoScore, rankThreeScore)

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.ADVANCED_RESPONSE_BONUS)
end

local function createRoundGenerator(self)
  self.roundGenerator = WitRoundGenerator.new()
end

local function createGameElements(self)
  if (GameSession.hasNeverWon()) then
    local tutorial = WitTutorial.new()
    self:insertAtCenter(tutorial)
    self.tutorial = tutorial
  else
    buildWitScene(self)
  end
end

local function createTimer(self)
  local timeoutHandler = getTimeoutHandler(self)
  local timer = WitCountdownTimer.new(timeoutHandler)
  return timer
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self.uiOverlay, UI_COMPONENTS_COLOR)

  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.timer = createTimer(self) -- Overrides default view

  self.currentWordCount = 0
  local initialCountString = string.format("%d/%d", self.currentWordCount, WitConfiguration.getCorrectResponsesNeededToPass())
  self.counterLabel = Label.new(initialCountString, Font.new(FontFaces.DIN_OT_MEDIUM, 16))
  self.counterLabel:setAnchorPointRight()
  builder:addWidget(self.counterLabel, UIOverlay.NORTHEAST)
end

-- Constructor ------------------------

function WitGame.new()
  local self = BaseGame.new()
  Mixin.mix(self, WitGame)

  configureGameDifficulty(self)
  createRoundGenerator(self)
  createUI(self)
  createGameElements(self)

  WitAnalytics.instantiate()

  return self
end
