--- WitInputField
-- Created by Alexey Chernikov on 02/25/2013

class("WitInputField")

require("actions.advanced.AHSmoothShake")

-- ---------------------------------------------------------------------------
-- Configuration
-- ---------------------------------------------------------------------------

local TEXT_COLOR = {0, 0, 0, 0.8}
local INPUT_FIELD_BOX_COLOR = {1.0, 1.0, 1.0}
local MESSAGE_COLOR = {0, 0, 0.3, 0.5}
local DEFAULT_INCORRECT_BACKGROUND_COLOR = {1, 0, 0}
local DISABLED_COLOR = {1, 1, 1, 0}
local ENABLED_COLOR = {1, 1, 1, 1}

local CURSOR_WIDTH = 2.0
local CURSOR_HEIGHT = 20.0
local CURSOR_COLOR = {0.6, 0.6, 0.6, 0.5}
local PROMPT_COLOR = {0.5, 0.5, 0.5}

local CURSOR_BLINK_DELAY_DURATION = 0.3
local INPUT_PROMPT_DELAY_DURATION = 7.0

local INPUT_PROMPT_TEXT_INITIAL = "ENTER A MORE %s<br/>WORD TO REPLACE \"%s\""
local INPUT_PROMPT_TEXT_MORE = "ENTER MORE WORDS OR<br/>TAP THE END ROUND BUTTON"

-- ---------------------------------------------------------------------------
-- Utility
-- ---------------------------------------------------------------------------

local function startCursorBlink(self)
  if self.cursorBlinkCoroutine and self.cursorBlinkCoroutine:isBusy() then
    return
  end

  self.cursorBlinkCoroutine = runInCoroutine(function()
    local blinkDelay = AHDelay.new(CURSOR_BLINK_DELAY_DURATION)
    while (true) do
      self.cursor:setVisible(false)
      wait(blinkDelay:start())
      self.cursor:setVisible(true)
      wait(blinkDelay:start())
    end
  end)
end

local function stopCursorBlink(self)
  if self.cursorBlinkCoroutine and self.cursorBlinkCoroutine:isBusy() then
    self.cursorBlinkCoroutine:stop()
  end

  self.cursor:setVisible(false)
end

local function updateCursorLocation(self)
  local label = self.label
  local labelX, _ = label:getLoc()
  local labelWidth = label:getWidth()

  local cursorX = labelX + labelWidth * 0.5
  self.cursor:setX(cursorX)
end

local function showInputPrompt(self)
  if self.inputPrompt:getOpacity() < 1 and self:getOpacity() > 0 then
    wait(AHOpacityTo.new(self.inputPrompt, 0.1, nil, 1.0):start())
  end
end

local function hideInputPrompt(self)
  if self.inputPrompt:getOpacity() > 0 then
    wait(AHOpacityTo.new(self.inputPrompt, 0.1, nil, 0.0):start())
  end
end

local function fade(self, color, animTime)
  if(self.fadeAction) then
    self.fadeAction:stop()
  end
  local panelFade = AHColorTo.new(self, animTime, nil, color)
  if color[4] > 0 then
    self.fadeAction = AHSequence.new(
      panelFade,
      function() showInputPrompt(self) end
    )
  else
    self.fadeAction = AHSpawn.new(
      panelFade,
      function() hideInputPrompt(self) end
    )
  end

  self.fadeAction:start()
end

local function showMessage(self, message, displaySeconds, color)
  color = color or MESSAGE_COLOR
  local fadeInSeconds = displaySeconds * 0.2
  local fadeOutSeconds = displaySeconds * 0.2
  local showSeconds = displaySeconds - fadeInSeconds - fadeOutSeconds

  local messageView = Label.new(message, Font.new(FontFaces.DIN_OT_LIGHT, self.fontSize))
  messageView:setColor(unpack(color))
  self:insert(messageView)

  local totalWidth, totalHeight = self:getDims()
  local messageWidth = messageView:getWidth()
  local x = totalWidth - messageWidth/2 - 10
  local y = totalHeight/2
  messageView:setLoc(x, y)

  messageView:setOpacity(0)
  local showMessage = AHOpacityTo.new(messageView, fadeInSeconds, nil, 1)
  local presentMessage = AHDelay.new(showSeconds)
  local hideMessage = AHOpacityTo.new(messageView, fadeOutSeconds, nil, 0)
  local removeMessage = function() messageView:removeSelf() end


  local showMessageAction = AHSequence.new(showMessage, presentMessage, hideMessage, removeMessage)
  showMessageAction:start()
end

-- ---------------------------------------------------------------------------
-- API
-- ---------------------------------------------------------------------------

function WitInputField:setTouchEnabled(isEnabled)
  self.isCurrentlyInRound = isEnabled
  self.skipButton:setTouchEnabled(isEnabled)
end

function WitInputField:setTouchEnabledVisual(isEnabled)
  self.isCurrentlyInRound = isEnabled
  self.skipButton:setTouchEnabledVisual(isEnabled)
end

function WitInputField:append(symbol)
  symbol = string.lower(symbol)
  local labelText = self.label:getText()
  if (self.maxLength and string.len(labelText) >= self.maxLength) then
    return
  end
  self.label:appendText(symbol)
  updateCursorLocation(self)
end

function WitInputField:setText(text)
  self.label:setText(string.lower(text))
  if self.isRoundOver then
    hideInputPrompt(self)
    stopCursorBlink(self)
  elseif text:len() > 0 then
    hideInputPrompt(self)
    startCursorBlink(self)
  else
    showInputPrompt(self)
    stopCursorBlink(self)
  end
  updateCursorLocation(self)
end

function WitInputField:backspace()
  self.label:eraseLastCharacter()
  updateCursorLocation(self)
end

function WitInputField:erase()
  self.label:setOpacity(TEXT_COLOR[4])
  self.label:setText("")
  updateCursorLocation(self)
end

function WitInputField:getText()
  return self.label:getText()
end

function WitInputField:setMaxLength(maxLength)
  self.maxLength = maxLength
end

function WitInputField:createIncorrectAnswerAction(backgroundColor)
  local animationSeconds = 0.3
  backgroundColor = backgroundColor or DEFAULT_INCORRECT_BACKGROUND_COLOR
  local totalAnimationTime = animationSeconds

  local xShakeAmp = 20
  local yShakeAmp = 0
  local shakesPerSecond = 3
  local shakeSeconds = totalAnimationTime
  local colorChangeSeconds = totalAnimationTime/4
  local colorAndShakeOverlapSeconds = totalAnimationTime/3
  local colorKeepSeconds = totalAnimationTime - colorChangeSeconds * 2 - colorAndShakeOverlapSeconds

  local shakeText = AHSmoothShake.new(self.label, xShakeAmp, yShakeAmp, shakesPerSecond, shakeSeconds)
  local highlightBackground = AHColorTo.new(self.background, colorChangeSeconds, nil, backgroundColor)
  local keepColor = AHDelay.new(colorKeepSeconds)
  local removeHighlight = AHColorTo.new(self.background, colorChangeSeconds, nil, INPUT_FIELD_BOX_COLOR)
  local backgroundAnimation = AHSequence.new(highlightBackground, keepColor, removeHighlight)

  local incorrectAnimation = AHSpawn.new(shakeText, backgroundAnimation)

  return incorrectAnimation
end

function WitInputField:createCorrectAnswerAction(backgroundColor)
  local totalAnimationTime = 0.1
  local colorChangeSeconds = totalAnimationTime * 0.2
  local colorResumeSeconds = totalAnimationTime/2
  local colorKeepSeconds = totalAnimationTime - colorChangeSeconds - colorResumeSeconds

  local highlightBackground = AHColorTo.new(self.background, colorChangeSeconds, nil, backgroundColor)
  local keepColor = AHDelay.new(colorKeepSeconds)
  local removeHighlight = AHColorTo.new(self.background, colorResumeSeconds, nil, INPUT_FIELD_BOX_COLOR)
  local backgroundAnimation = AHSequence.new(highlightBackground, keepColor, removeHighlight)

  self.inputPrompt:setText(INPUT_PROMPT_TEXT_MORE)

  return backgroundAnimation
end

function WitInputField:createFadeInputAction(animationSeconds)
  local action = AHOpacityTo.new(self.label, animationSeconds, nil, 0)
  return action
end

function WitInputField:createDisableAction(animSeconds)
  return AHCallLogic.new(function()
    fade(self, DISABLED_COLOR, animSeconds)
  end)
end

function WitInputField:createEnableAction(animSeconds)
  return AHCallLogic.new(function ()
    fade(self, ENABLED_COLOR, animSeconds)
  end)
end

function WitInputField:showAutocorrectMessage(color)
  showMessage(self, "autocorrect", 0.9, color)
end

function WitInputField:showDuplicateCorrectAnswerMessage(color)
  showMessage(self, "Already entered", 0.9, color)
end

function WitInputField:addLeftWidget(widget)
  widget:setLoc(widget:getWidthPart(0.5) + 5, self.background:getHeightPart(0.5))
  self:insert(widget)
  widget:inheritColor(self)
end

function WitInputField:addDigitTimer(timer)
  local timerLocationX = timer:getWidthPart(1)
  local timerLocationY = self.skipButton:getY()
  timer:setLoc(timerLocationX, timerLocationY)
  self:insert(timer)
end

function WitInputField:startNewRound(word, descriptor)
  self.isRoundOver = false
  local inputText = string.format(INPUT_PROMPT_TEXT_INITIAL, string.upper(descriptor), string.upper(word))
  self.inputPrompt:setText(inputText)
  showInputPrompt(self)
end

function WitInputField:notifyRoundEnded()
  self.isRoundOver = true
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ---------------------------------------------------------------------------

local function createSkipButton(self, onPressed)
  local downOpacity = 0.8
  local upOpacity = 0.3
  local skipBackground = Sprite.newWithName("end_round.png")
  local skipButton = Group.new(skipBackground:getDims())
  skipButton:insertAtCenter(skipBackground)
  local skipLabel = Label.new("END ROUND", Font.new(FontFaces.DIN_OT_MEDIUM, 13))
  skipLabel:setLoc(skipButton:getWidthPart(0.6), skipButton:getHeightPart(0.5))
  skipButton:insert(skipLabel)
  skipButton:linkColorAttrToChildren(true)
  skipBackground:setOpacity(upOpacity)
  local onTouchDown = function()
    onPressed()
    hideInputPrompt(self)
    skipBackground:setOpacity(downOpacity)
    AHDelayed.new(0.2, AHOpacityTo.new(skipBackground, 0.2, nil, upOpacity)):start()
  end

  Button.makeButton(skipButton, onTouchDown)
  skipButton.skipButtonInterupt = ActionInterrupt.new()
  skipButton.setTouchEnabledVisual = function(self, isEnabled)
    self:setTouchEnabled(isEnabled)
    local opacity = 0
    if isEnabled then
      opacity = 1
    end
    self.skipButtonInterupt:insert(AHOpacityTo.new(skipLabel, 0.3, nil, opacity))
  end
  skipButton:setTouchEnabled(false)
  skipLabel:setOpacity(0)
  return skipButton
end

local function createLayout(self, onSkip)
  local maxLength = 25
  local fontSize = 18
  local background = FilledRectangle.new(Screen.fullWidth, Screen.getHeightPart(0.09), INPUT_FIELD_BOX_COLOR)
  local cursor = FilledRectangle.new(CURSOR_WIDTH, CURSOR_HEIGHT, CURSOR_COLOR)
  cursor:setAnchorPointLeft()

  local initialText = ""
  local label = Label.new(initialText, PrerenderedFont.new(FontFaces.DIN_OT_MEDIUM, 18))
  background:insertAtCenter(label)
  background:insertAtCenter(cursor)

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 14)
  local inputPrompt = DynamicTextArea.newSimple(
    "",
    Screen.getSafeWidthPart(0.6),
    font,
    Alignment.CENTER,
    Alignment.CENTER
  )
  inputPrompt:setColor(unpack(PROMPT_COLOR))
  inputPrompt:setOpacity(0.0)
  self.inputPrompt = inputPrompt
  background:insertAtCenter(inputPrompt)

  local skipButton = createSkipButton(self, onSkip)
  self:insert(skipButton)
  self:insert(background)

  self:alignChildren()

  background:inheritColor(self)
  cursor:inheritColor(self)
  label:inheritColor(self)
  skipButton:inheritColor(self)

  self.label = label
  self.cursor = cursor
  self.skipButton = skipButton
  self.background = background
  self.fontSize = fontSize

  stopCursorBlink(self)

  self:setMaxLength(maxLength)

  label:setColor(unpack(TEXT_COLOR))
end

-- Constructor ------------------------

function WitInputField.new(onSkip)
  local self = Column.new(0, Alignment.RIGHT)
  Mixin.mix(self, WitInputField)
  createLayout(self, onSkip)
  self:setTouchEnabled(false)
  self.hasPromptBeenShown = false
  self.isCurrentlyInRound = false
  self.isRoundOver = false

  return self
end
