--- WitTutorial
-- @author Created by Aaron Barrett on July 29, 2015.

class("WitTutorial")

require("WitBackground")
require("WitTutorialBox")
require("WitTutorialCommon")
require("widgets.progress.NumericProgressIndicator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INITIAL_BLURB = "Eloquence is a game to help you change the meaning of a sentence."

local PROMPT_EXPLANATION_TEXT = "Replace the word <b>tiny</b> with a new word that retains the meaning <b>small</b> but also makes the sentence more complimentary."

WitTutorial.WORD_DESCRIPTORS = {
  {text = "charming", correct = true, blurbText = "<b>Charming</b> makes the sentence more complimentary and still suggests <b>tiny</b>."},
  {text = "intimate", correct = true, blurbText = "<b>Intimate</b> makes the sentence more complimentary and still suggests <b>tiny</b>."},
  {text = "nice", correct = false, blurbText = "<b>Nice</b> makes the sentence more complimentary, but it doesn’t suggest the meaning <b>tiny</b>."},
  {text = "little", correct = false, blurbText = "<b>Little</b> suggests the meaning <b>tiny</b>, but it doesn’t make the sentence more complimentary."}
}

local Z_BACKGROUND = 10
local Z_BOX = 20
local Z_BLURB_LABEL = 30
local Z_TAP_TO_CONTINUE_LABEL = 40
local Z_STEP_COUNTER = 50

local STEPS_COUNT = 6

local DROP_DISTANCE = Screen.getHeightPart(0.04)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function waitForTap(self)
  local tapToContinueLabel = self.tapToContinueLabel
  local tapToContinueInterrupt = self.tapToContinueInterrupt

  local tapToContinueAppearAction = AHOpacityTo.new(tapToContinueLabel, 0.3, nil, 1.0)
  tapToContinueInterrupt:insert(tapToContinueAppearAction)

  local tapToContinue = TapToContinue.new()
  tapToContinue:waitForTrigger()

  local tapToContinueDisappearAction = AHOpacityTo.new(tapToContinueLabel, 0.3, nil, 0.0)
  tapToContinueInterrupt:insert(tapToContinueDisappearAction)
end

local function playFadeIn(group, destinationOpacity)
  destinationOpacity = destinationOpacity or 1.0

  group:setY(group:getY() + DROP_DISTANCE)

  local dropAction = AHSpawn.new(
                                 AHLocationYBy.new(group, 0.25, AHEaseSimpleOut.new(3.0), -DROP_DISTANCE),
                                 AHOpacityTo.new(group, 0.25, nil, destinationOpacity)
                                 )
  return dropAction:start()
end

local function playFadeOut(group)
  local dropAction = AHSequence.new(
                                    AHSpawn.new(
                                                AHLocationYBy.new(group, 0.25, AHEaseSimpleIn.new(3.0), -DROP_DISTANCE),
                                                AHOpacityTo.new(group, 0.25, nil, 0.0)
                                                ),
                                    AHRemove.new(group)
                                    )
  return dropAction:start()
end

local function createWordBlurbLabel(self, text)
  local width = Screen.getSafeWidthPart(0.85)
  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_MEDIUM, 20),
    colored = false
  }
  local horizontalAlignment = Alignment.LEFT
  local blurbLabel = DynamicTextArea.new(text, width, baseAttributes, WitTutorialCommon.ATTRIBUTES_FOR_LABELS, horizontalAlignment)
  WitTutorialCommon.colorWordsInLabel(blurbLabel)
  blurbLabel:setOpacity(0.0)
  return blurbLabel
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitTutorial:playTutorial()
  return runInCoroutine(function()
    local stepCounter = self.stepCounter
    local box = self.box
    local blurbLabel = self.blurbLabel
    self.blurbLabel = nil

    stepCounter:incrementProgress()
    AHOpacityTo.new(stepCounter, 0.25, nil, 1.0):start()

    wait(playFadeIn(blurbLabel))
    waitForTap(self)
    stepCounter:incrementProgress()
    wait(playFadeOut(blurbLabel))

    blurbLabel = createWordBlurbLabel(self, PROMPT_EXPLANATION_TEXT)
    blurbLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.35))
    self:insert(blurbLabel, Z_BLURB_LABEL)

    box:playFadeInLabels()
    playFadeIn(blurbLabel)
    local destinationOpacity = 0.25
    wait(playFadeIn(box, destinationOpacity))

    waitForTap(self)
    stepCounter:incrementProgress()

    local wordsCount = #(WitTutorial.WORD_DESCRIPTORS)

    for wordIndex, wordDesctiptor in ipairs(WitTutorial.WORD_DESCRIPTORS) do
      box:playShowWord(wordIndex)
      wait(playFadeOut(blurbLabel))

      local text = wordDesctiptor.blurbText
      blurbLabel = createWordBlurbLabel(self, text)
      blurbLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.35))
      self:insert(blurbLabel, Z_BLURB_LABEL)

      wait(playFadeIn(blurbLabel))
      waitForTap(self)
      if (wordIndex < wordsCount) then
        stepCounter:incrementProgress()
      end
    end

    playFadeOut(blurbLabel)

    box:playFadeOutLabels()
    playFadeOut(box)

    wait(AHOpacityTo.new(stepCounter, 0.25, nil, 0.0):start())
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WitTutorial.new()
  local self = Group.newFullScreen()
  Mixin.mix(self, WitTutorial)

  local background = WitBackground.new()
  self:insertAtCenter(background, Z_BACKGROUND)

  local width = Screen.getSafeWidthPart(0.82)
  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_LIGHT, 24),
    colored = false
  }
  local horizontalAlignment = Alignment.LEFT
  local blurbLabel = DynamicTextArea.new(INITIAL_BLURB, width, baseAttributes, WitTutorialCommon.ATTRIBUTES_FOR_LABELS, horizontalAlignment)
  blurbLabel:setOpacity(0.0)
  blurbLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.5))
  self:insert(blurbLabel, Z_BLURB_LABEL)
  self.blurbLabel = blurbLabel

  local tapToContinueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 16))
  tapToContinueLabel:setOpacity(0.0)
  tapToContinueLabel:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.1))
  self:insert(tapToContinueLabel, Z_TAP_TO_CONTINUE_LABEL)
  self.tapToContinueLabel = tapToContinueLabel
  self.tapToContinueInterrupt = ActionInterrupt.new()

  local minStepIndex = 0
  local maxStepIndex = STEPS_COUNT
  local shouldShowMaxValue = true
  local separatorText = " of "
  local stepCounter = NumericProgressIndicator.new(FontFaces.DIN_OT_MEDIUM,
                                                   16,
                                                   minStepIndex,
                                                   maxStepIndex,
                                                   shouldShowMaxValue,
                                                   separatorText)
  stepCounter:setColor(unpack(WitTutorialCommon.COLOR_HIGHLIGHT))
  stepCounter:setOpacity(0.0)
  stepCounter:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.9))
  self:insert(stepCounter, Z_STEP_COUNTER)
  self.stepCounter = stepCounter

  local box = WitTutorialBox.new()
  box:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.64))
  self:insert(box, Z_BOX)
  self.box = box

  return self
end
