--- WitTutorialBox
-- @author Created by Aaron Barrett on July 30, 2015.

class("WitTutorialBox")

require("WitTutorialCommon")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local PROMPT_TEXT = "BE MORE COMPLIMENTARY"
local SENTENCE_TEXT = "Sam's new restaurant is <b>tiny</b>."

local Z_BACKING = 10
local Z_SEPARATOR = 20
local Z_TOP_LABEL_HOLDER = 30
local Z_BOTTOM_LABEL_HOLDER = 40

local COLOR_CHECKMARK = {0.31, 0.98, 0.83}
local COLOR_X = {0.97, 0.37, 0.37}
local COLOR_SEPARATOR = {0.15, 0.39, 0.50}

local HORIZONTAL_MARGIN = Screen.getWidthPart(0.15)
local VERTICAL_MARGIN = Screen.getHeightPart(0.04)
local SEPARATOR_MARGIN = Screen.getHeightPart(0.025)

local WORD_DROP_DISTANCE = Screen.getHeightPart(0.02)

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitTutorialBox:playFadeInLabels()
  local topLabelHolder = self.topLabelHolder
  local fadeAction = AHOpacityTo.new(topLabelHolder, 0.25, nil, 1.0)
  return fadeAction:start()
end

function WitTutorialBox:playFadeOutLabels()
  local elementsToFade = {
    self.topLabelHolder,
    self.separator,
    unpack(self.wordHolders)
  }

  for _, element in ipairs(elementsToFade) do
    local fadeAction = AHOpacityTo.new(element, 0.25, nil, 0.0)
    fadeAction:start()
  end
end

function WitTutorialBox:getBackingHeight()
  return self.backingHeight
end

function WitTutorialBox:setBackingHeight(height)
  self.backingHeight = height

  local backing = self.backing
  local topLabelHolder = self.topLabelHolder
  local separator = self.separator
  local maxBackingHeight = self.maxBackingHeight

  local backingScale = height / maxBackingHeight
  backing:setScl(1.0, backingScale)

  local topLabelHolderHeight = topLabelHolder:getHeight()
  local topOffset = maxBackingHeight + height - VERTICAL_MARGIN

  local topLabelHolderY = 0.5 * (topOffset - topLabelHolderHeight)
  topLabelHolder:setY(topLabelHolderY)
  local separatorY = 0.5 * (topOffset - SEPARATOR_MARGIN) - topLabelHolderHeight
  separator:setY(separatorY)
end

function WitTutorialBox:createBackingHeightToAction(duration, ease, destinationHeight)
  return AHNonAttributeTo.new(self, duration, self.getBackingHeight, self.setBackingHeight, ease, destinationHeight)
end

function WitTutorialBox:playShowWord(wordIndex)
  local topLabelHolder = self.topLabelHolder
  local rowsByWordHolder = self.rowsByWordHolder
  local bottomLabelHolder = self.bottomLabelHolder
  local separator = self.separator
  local maxBackingHeight = self.maxBackingHeight

  local wordHolders = self.wordHolders
  local wordHolder = wordHolders[wordIndex]

  local row = rowsByWordHolder[wordHolder]

  local originalY = wordHolder:getY()
  wordHolder:setY(originalY + WORD_DROP_DISTANCE)

  local destinationHeight = topLabelHolder:getHeight() + VERTICAL_MARGIN + bottomLabelHolder:getHeight() - row:getY() + row:getHeightPart(0.5) + SEPARATOR_MARGIN
  local bottomLabelHolderDestinationY = 0.5 * (maxBackingHeight - destinationHeight + VERTICAL_MARGIN + bottomLabelHolder:getHeight() + row:getHeight()) - row:getY()

  local showAction = AHSpawn.new(
                                 AHDelayed.new(0.15, AHOpacityTo.new(wordHolder, 0.35, AHEaseSineIn.new(), 1.0)),
                                 AHOpacityTo.new(separator, 0.5, AHEaseSimpleIn.new(3.0), 1.0),
                                 AHDelayed.new(0.15, AHLocationYTo.new(wordHolder, 0.35, AHEaseSineOut.new(), originalY)),
                                 AHLocationYTo.new(bottomLabelHolder, 0.5, AHEaseSineInOut.new(2.0), bottomLabelHolderDestinationY),
                                 self:createBackingHeightToAction(0.5, AHEaseSineInOut.new(4.0), destinationHeight)
                                 )
  return showAction:start()
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createPromptLabel()
  local promptLabel = Label.new(PROMPT_TEXT, Font.new(FontFaces.DIN_OT_MEDIUM, 12))
  promptLabel:setColor(unpack(WitTutorialCommon.COLOR_HIGHLIGHT))
  return promptLabel
end

local function createSentenceLabel(width)
  local baseAttributes = {
    font = Font.new(FontFaces.DIN_OT_LIGHT, 22),
    colored = false
  }
  local sentenceLabel = DynamicTextArea.new(SENTENCE_TEXT, width, baseAttributes, WitTutorialCommon.ATTRIBUTES_FOR_LABELS)
  WitTutorialCommon.colorWordsInLabel(sentenceLabel)
  return sentenceLabel
end

local function createTopLabelHolder(promptLabel, sentenceLabel)
  local spacing = Screen.getHeightPart(0.01)
  local alignment = Alignment.LEFT
  local holder = Column.new(spacing, alignment)

  promptLabel:inheritOpacity(holder)
  holder:insert(promptLabel)

  sentenceLabel:inheritOpacity(holder)
  holder:insert(sentenceLabel)

  holder:alignChildren()

  holder:setOpacity(0.0)

  return holder
end

local function createWordHolder(wordDescriptor)
  local spacing = Screen.getWidthPart(0.01)
  local holder = Row.new(spacing)

  local wordText = wordDescriptor.text
  local wordLabel = Label.new(wordText, Font.new(FontFaces.DIN_OT_MEDIUM, 18))
  wordLabel:setColor(unpack(WitTutorialCommon.COLOR_HIGHLIGHT))
  wordLabel:inheritOpacity(holder)
  holder:insert(wordLabel)

  local markFilename = nil
  local markColor = nil
  if (wordDescriptor.correct) then
    markFilename = "witcheck.png"
    markColor = COLOR_CHECKMARK
  else
    markFilename = "witx.png"
    markColor = COLOR_X
  end

  local scaleFactor = 0.5
  local verticalMarkAlignmentProportion = 0.4

  local mark = Sprite.newWithName(markFilename)
  mark:setScl(scaleFactor)
  mark:setColor(unpack(markColor))
  mark:inheritOpacity(holder)

  local alignerWidth = mark:getWidthPart(scaleFactor)
  local alignerHeight = wordLabel:getHeight()
  local markAligner = Group.new(alignerWidth, alignerHeight)
  mark:setX(alignerWidth * 0.5)
  mark:setY(alignerHeight * verticalMarkAlignmentProportion)
  markAligner:insert(mark)

  holder:insert(markAligner)

  holder:alignChildren()

  holder:setOpacity(0.0)

  return holder
end

local function createWordHolders()
  local wordHolders = {}
  for _, wordDescriptor in ipairs(WitTutorial.WORD_DESCRIPTORS) do
    local wordHolder = createWordHolder(wordDescriptor)
    table.insert(wordHolders, wordHolder)
  end
  return wordHolders
end

local function createBottomLabelHolder(wordHolders)
  local columnsCount = 3
  local verticalSpacing = Screen.getHeightPart(0.01)
  local bottomLabelHolder = Column.new(verticalSpacing, Alignment.LEFT)

  local horizontalSpacing = Screen.getWidthPart(0.04)

  local rowsByWordHolder = {}

  local columnIndex = 1
  local currentRow = Row.new(horizontalSpacing)

  for _, wordHolder in ipairs(wordHolders) do
    if (columnIndex > columnsCount) then
      currentRow:alignChildren()
      bottomLabelHolder:insert(currentRow)

      currentRow = Row.new(horizontalSpacing)
      columnIndex = 1
    end

    currentRow:insert(wordHolder)
    rowsByWordHolder[wordHolder] = currentRow
    columnIndex = columnIndex + 1
  end

  currentRow:alignChildren()
  bottomLabelHolder:insert(currentRow)

  bottomLabelHolder:alignChildren()

  bottomLabelHolder:setOpacity(0.0)

  return bottomLabelHolder, rowsByWordHolder
end

local function createBacking(backingWidth, maxBackingHeight)
  local cornerRadius = Screen.getWidthPart(0.01)
  local color = {0.0, 0.0, 0.0}
  local backing = FilledRoundedRectangle.new(backingWidth, maxBackingHeight, cornerRadius, color)
  backing:setOpacity(0.25)
  return backing
end

local function createSeparator(backingWidth)
  local height = Screen.getHeightPart(0.0025)
  local separator = FilledRectangle.new(backingWidth, height, COLOR_SEPARATOR)
  separator:setOpacity(0.0)
  return separator
end

-- Constructor ------------------------

function WitTutorialBox.new()
  local wordHolders = createWordHolders()
  local bottomLabelHolder, rowsByWordHolder = createBottomLabelHolder(wordHolders)

  local promptLabel = createPromptLabel()
  local sentenceLabel = createSentenceLabel(bottomLabelHolder:getWidth())
  local topLabelHolder = createTopLabelHolder(promptLabel, sentenceLabel)

  local backingWidth = math.max(topLabelHolder:getWidth(), bottomLabelHolder:getWidth()) + HORIZONTAL_MARGIN
  local maxBackingHeight = topLabelHolder:getHeight() + bottomLabelHolder:getHeight() + VERTICAL_MARGIN

  local backing = createBacking(backingWidth, maxBackingHeight)

  local separator = createSeparator(backingWidth)

  local self = Group.new(backingWidth, maxBackingHeight)
  Mixin.mix(self, WitTutorialBox)

  backing:inheritOpacity(self)
  self:insertAtCenter(backing, Z_BACKING)
  self.backing = backing

  topLabelHolder:setX(backingWidth * 0.5)
  self:insert(topLabelHolder, Z_TOP_LABEL_HOLDER)
  self.topLabelHolder = topLabelHolder

  self.wordHolders = wordHolders
  self.rowsByWordHolder = rowsByWordHolder
  self:insert(bottomLabelHolder, Z_BOTTOM_LABEL_HOLDER)
  self.bottomLabelHolder = bottomLabelHolder
  self.wordHolders = wordHolders

  separator:setX(backingWidth * 0.5)
  self:insert(separator, Z_SEPARATOR)
  self.separator = separator

  self.maxBackingHeight = maxBackingHeight

  bottomLabelHolder:setLoc(backingWidth * 0.5, 0.5 * VERTICAL_MARGIN)

  local initialBackingHeight = topLabelHolder:getHeight() + VERTICAL_MARGIN
  self:setBackingHeight(initialBackingHeight)

  self:setOpacity(0.0)

  return self
end
