--- WitHintView

class("WitHintView")

require("WitSolutionExamples")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------
local X_MARGIN = Screen.getWidthPart(0.075)
local WIDTH = Screen.safeWidth - X_MARGIN * 2
local BACKGROUND_COLOR = {0, 0, 0, 0.8}
local TEXT_COLOR = {1, 1, 1}
local CHECKMARK_COLOR = {0.24, 0.85, 0.27}

local HIGHLIGHT_ATTRIBUTES = {
  keyword = {
    blueColored = true
  },
  replace = {
    blueColored = true
  },
  wrong = {
    redColored = true
  }
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function setColorForTags(textView, attribute, color)
  local views = textView:getViewsWithAttribute(attribute, true)
  for _, view in ipairs(views) do
    view:clearAttrLink(MOAIColor.INHERIT_COLOR)
    view:setColor(unpack(color))
  end
end

function WitHintView:createAppearAction(duration)
  duration = duration or 0.3
  self:setOpacity(0)
  return AHOpacityTo.new(self, duration, nil, 1)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitHintView:createDisappearAction(duration)
  duration = duration or 0.3
  return AHOpacityTo.new(self, duration, nil, 0)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createTextWithHighlight(self, text, tag, fontSize, fontFace)
  local basicAttributes = {
    font = Font.new(fontFace, fontSize)
  }
  local alignment = nil
  local lineSpacing = Screen.getHeightPart(0.005)
  local body = DynamicTextArea.new(text, WIDTH, basicAttributes, HIGHLIGHT_ATTRIBUTES, alignment, alignment, lineSpacing)

  setColorForTags(body, "blueColored", self.highlightColor)
  setColorForTags(body, "redColored", self.wrongColor)

  return body
end

local function createHeader(self, headerText, bodyText)
  local view = Group.newDeprecated()

  local headerFont = FontFaces.DIN_OT_BOLD
  local bodyFont = FontFaces.DIN_OT_REGULAR
  local headerSize = 24
  local bodyFontSize = 20
  local header = Label.new(headerText, Font.new(headerFont, headerSize))

  local body = createTextWithHighlight(self, bodyText, "keyword", bodyFontSize, bodyFont)

  local headerW, headerH = header:getDims()
  local bodyW, bodyH = body:getDims()
  local dy = Screen.getHeightPart(1.0/113)
  local totalH = headerH + bodyH + dy
  local totalW = bodyW

  header:setLoc(-totalW/2 + headerW/2, -headerH/2)
  body:setLoc(0, -headerH - bodyH/2 - dy)

  view:insert(header)
  view:insert(body)

  view:setBounds(0, 0, 0, totalW, totalH, 0)
  view.body = body

  return view
end

local function createTipView(self, taskKeyword, specifiedWord)
  local bodyText = "The word <wrong>"..specifiedWord.."</wrong> fits the sentence, but it's too unrelated to the word <keyword>"..taskKeyword.."</keyword>. Try submitting words that are more related to the original sentence."
  local headerText = "Why is my answer incorrect?"

  local tipView = createHeader(self, headerText, bodyText)

  return tipView
end

local function createExampleAnswer(self, word, width, isCorrect)
  local fontSize = WitConstants.LARGE_FONT_SIZE
  local w, h = width, fontSize * 1.5
  local view = Group.new(w, h)
  local wordView = Label.new(word, Font.new(FontFaces.DIN_OT_LIGHT, fontSize))

  local iconName = "witx.png"
  local color = self.wrongColor
  if(isCorrect) then
    color = CHECKMARK_COLOR
    iconName = "witcheck.png"
  end

  local icon = Sprite.newWithName(iconName)
  local wordW, wordH = wordView:getDims()

  wordView:setLoc(-w/2, 0)
  wordView:setAnchorPointLeft()
  icon:setAnchorPointLeft()
  icon:setLoc(w/2, 0)

  icon:setColor(unpack(color))

  view:insert(wordView)
  view:insert(icon)

  return view
end

local function createExampleAnswersFromList(self, list, isCorrect)
  local result = Group.newDeprecated()
  local y = 0
  local dy = 3
  local width = Screen.getSafeWidthPart(0.5)
  for _, answer in ipairs(list) do
    local answerView = createExampleAnswer(self, answer, width, isCorrect)
    answerView:setLoc(0, y)
    local w, h = answerView:getDims()
    y = y - h - dy
    result:insert(answerView)
    answerView:inheritColor(result)
  end
  local height = math.abs(y)
  result:setBounds(0, 0, 0, width, height, 0)
  return result
end

local function createExampleAnswersView(self, exampleInfo)
  local result = Group.newDeprecated()
  local corrects = createExampleAnswersFromList(self, exampleInfo.acceptable, true)
  local wrongs = createExampleAnswersFromList(self, exampleInfo.wrong, false)

  local correctsW, correctsH = corrects:getDims()
  wrongs:setLoc(0, -correctsH)

  result:insert(corrects)
  result:insert(wrongs)

  return result
end

local function createExampleView(self, exampleInfo)
  local view = Group.newDeprecated()

  local headerText = "Example:"
  local taskText = "Make the sentence more <keyword>"..exampleInfo.task.."</keyword>."
  local font = FontFaces.DIN_OT_LIGHT
  local sentenceFontSize = WitConstants.SMALL_FONT_SIZE
  local taskView = createHeader(self, headerText, taskText)
  local sentenceView = createTextWithHighlight(self, exampleInfo.prompt, "replace", sentenceFontSize, font)
  local exampleAnswersView = createExampleAnswersView(self, exampleInfo)

  local taskW, taskH = taskView:getDims()
  local sentenceW, sentenceH = sentenceView:getDims()
  local dy = Screen.getHeightPart(0.02)

  local sentenceY = -taskH - sentenceH/2 - dy
  local exampleAnswersY = sentenceY - sentenceH/2
  sentenceView:setLoc(0, sentenceY)
  exampleAnswersView:setLoc(X_MARGIN, exampleAnswersY - dy)

  view:insert(taskView)
  view:insert(sentenceView)
  view:insert(exampleAnswersView)

  return view
end

local function createLayout(self, round, specifiedWord)
  local exampleInfo = WitSolutionExamples.getExampleForRound(round)

  local background = FilledRectangle.newFullScreen(BACKGROUND_COLOR)

  local tipView = createTipView(self, exampleInfo.keyword, specifiedWord)
  local exampleView = createExampleView(self, exampleInfo)
  local tapToContinueLabel = Label.new(Localizer.get("Misc.TAP_TO_CONTINUE"), Font.new(FontFaces.DIN_OT_MEDIUM, 15))

  local tipW, tipH = tipView:getDims()
  local tipY = Screen.getSafeY(0.9)
  local dy = Screen.getHeightPart(0.03)
  local exampleY = tipY - tipH - dy

  local x = Screen.getSafeX(0.5)
  tipView:setLoc(x, tipY)
  exampleView:setLoc(x, exampleY)
  local minimumY = Screen.safeBottom + tapToContinueLabel:getHeightPart(0.5)
  local tapToContinueY = math.max(minimumY, tapToContinueLabel:getHeightPart(1.5))
  tapToContinueLabel:setLoc(x, tapToContinueY)
  background:setLoc(Screen.getWidthPart(0.5), Screen.getHeightPart(0.5))

  self:insert(background)
  self:insert(tipView)
  self:insert(exampleView)
  self:insert(tapToContinueLabel)

  local linkColorRecursively = true
  self:linkColorAttrToChildren(linkColorRecursively)
end

--- Creates a popup explaining how the game decides if an answer is correct or not.
-- @param colorScheme = {highlight = {r, g, b, a}, wrong = {r, g, b, a}}
function WitHintView.new(colorScheme, round, specifiedWord)
  local self = Group.newDeprecated()
  Mixin.mix(self, WitHintView)

  self.highlightColor = colorScheme.highlight
  self.wrongColor = colorScheme.wrong

  createLayout(self, round, specifiedWord)
  self:setOpacity(0)

  return self
end
