--- WitScene

class("WitScene")

require("WitInputField")
require("WitCollectedWordsView")
require("WitImprovableSentenceView")
require("WitAdvancedWordNotification")
require("WitUnusedWordsView")
require("scene.widgets.DotProgressBar")
require("scene.WitBackground")
require("scene.WitTaskView")
require("WitColors")
require("data.WitDictionary")
require("data.DictionaryWordsMatcher")
require("scene.InterstitialView")
require("scene.WitHintView")
require("widgets.TapToContinue")
require("DeviceKeyboard")
require("widgets.KeyboardWarning")
require("ExperimentsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Y_DISTANCE_BETWEEN_SENTENCE_AND_ANSWERS = Screen.getHeightPart(0.023)
local Y_DISTANCE_BETWEEN_PROGRESS_BAR_TO_TASK = 4 * Y_DISTANCE_BETWEEN_SENTENCE_AND_ANSWERS
local Y_DISTANCE_BETWEEN_TASK_AND_SENTENCE = 1.5 * Y_DISTANCE_BETWEEN_SENTENCE_AND_ANSWERS
local Y_DISTANCE_FROM_TOP_TO_PROGRESS_BAR = Y_DISTANCE_BETWEEN_SENTENCE_AND_ANSWERS
local X_MARGIN = 25
local LOSE_OUTRO_COLOR_SCHEME_INDEX = 5
local WIN_OUTRO_COLOR_SCHEME_INDEX = 4
local MAX_INPUT_LENGTH = 30
local VALID_KEYS = "abcdefghijklmnopqrstuvwxyz- "


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function clearInput(self)
  self.input:erase()
  DeviceKeyboard.clear()
end

local function ensureInputFieldShows(self)
  runInCoroutine(function()
    waitFor(0.5)
    if self.input:getY() <= 0 then
      self.input:setY(Screen.getSafeY(0.4))
    end
  end)
end

local function setInputEnabled(self, isEnabled)
  DeviceKeyboard.setEnabled(isEnabled)
end

local function showInputEnabled(self, isEnabled)
  self.input:setTouchEnabledVisual(isEnabled)
  if isEnabled then
    DeviceKeyboard.show()
    ensureInputFieldShows(self)
  else
    DeviceKeyboard.hide()
  end
end

local function createShowDisableInputAction(self, fadeTime)
  return AHSpawn.new(function() showInputEnabled(self, false) end, self.input:createDisableAction(fadeTime))
end

local function createEnableInputAction(self, fadeTime)
  local enableInputAction = AHSpawn.new(
    self.input:createEnableAction(fadeTime),
    function()
      showInputEnabled(self, true)
      setInputEnabled(self, true)
    end
  )

  return enableInputAction
end

local function createBlockInputAnimation(self, fadeSeconds, showReset)
  local blockInput = function()
    DeviceKeyboard.setEnabled(false)
  end

  local fadeInput = self.input:createFadeInputAction(fadeSeconds)

  if showReset then
    local resetInput = function()
      showInputEnabled(self, true)
      clearInput(self)
      setInputEnabled(self, true)
    end

    local blockInputAnimation = AHSequence.new(blockInput, fadeInput, resetInput)
    return blockInputAnimation
  else
    return fadeInput
  end
end

local function processCorrectAnswer(self, answer, originalString, isRoundOver)
  -- Don't accept words twice
  if(self.dictionary:isUsed(answer)) then
    local bgColors = WitColors.getBackgroundGradientByColorSchemeIndex(self.roundNumber)
    local color = bgColors[1]
    self.input:showDuplicateCorrectAnswerMessage(color)
    clearInput(self)
    return
  end

  self.dictionary:use(answer)
  self.collectedWords:addWord(answer)
  local isAdvanced = self.dictionary:isAdvanced(answer)
  if(isAdvanced) then
    self.advancedWordNotification:show()
  end

  if(self.correctAnswerHandler) then
    self.correctAnswerHandler(answer, originalString, isAdvanced)
  end

  local inputAction = self.input:createCorrectAnswerAction(WitColors.getCorrectAnswerColor())
  inputAction:start()

  local showResetAfterBlock = (not isRoundOver)
  local blockInputAnimation = createBlockInputAnimation(self, 0.0, showResetAfterBlock)
  blockInputAnimation:start()

  self.submitAnimationInterrupt:insert(blockInputAnimation)
end

local function showHint(self, providedAnswer)
  local colorScheme = WitColors.getHintColorSchemeForRound(self.roundNumber)
  local hintView = WitHintView.new(colorScheme, self.round, providedAnswer)
  self:insert(hintView)
  self.roundTimer:pause()

  clearInput(self)

  local fadeTime = 0.3
  local fadeOpacity = 0.3

  local disableInput = function() setInputEnabled(self, false) end
  local showDisableInputAction = createShowDisableInputAction(self, fadeTime)
  local hideSentenceAction = self.sentence:createFadeAction(fadeTime, fadeOpacity)
  local hideTaskAction = self.taskView:createFadeAction(fadeTime, fadeOpacity)
  local hintAppearAction = hintView:createAppearAction(fadeTime)
  local showHintAnimation = AHSpawn.new(disableInput, hintAppearAction, hideTaskAction, showDisableInputAction, hideSentenceAction)

  local enableInputAction = createEnableInputAction(self, fadeTime)
  local showSentenceAction = self.sentence:createFadeAction(fadeTime, 1)
  local showTaskAction = self.taskView:createFadeAction(fadeTime, 1)
  local hintDisappearAction = hintView:createDisappearAction(fadeTime)
  local hideHint = AHSpawn.new(hintDisappearAction, showTaskAction, enableInputAction, showSentenceAction)

  local unpauseTimerAction = function()
    local tap = TapToContinue.new()
    waitWhile(function() return tap.isArmed end)
  end

  local finishHint = function ()
    hintView:removeSelf()
    self.roundTimer:resume()
    showInputEnabled(self, true)
    setInputEnabled(self, true)
  end

  AHSequence.new(showHintAnimation, unpauseTimerAction, hideHint, finishHint):start()
end

local function processIncorrectAnswer(self, providedAnswer, isRoundOver)
  if(self.incorrectAnswerHandler) then
    self.incorrectAnswerHandler(providedAnswer)
  end
  if(not self.hintShown) then
     local bestExclusionMatch, _ = self.dictionaryMatcher:getBestExclusionMatch(providedAnswer)
     if(bestExclusionMatch) then
      showHint(self, bestExclusionMatch)
      self.hintShown = true
      return
     end
  end
  local inputAction = self.input:createIncorrectAnswerAction(WitColors.getIncorrectAnswerColor())
  inputAction:start()

  local showResetAfterBlock = (not isRoundOver)
  local blockInputAnimation = createBlockInputAnimation(self, 0.5, showResetAfterBlock)
  blockInputAnimation:start()

  self.submitAnimationInterrupt:insert(blockInputAnimation)
end

local function containsConsecutiveWhitespace(text)
  return (text:find("%s%s") ~= nil)
end

local function containsPrintableCharacterAfterWhitespace(text)
  return (text:find("%s%g") ~= nil)
end

local function getInputConfirmHandler(self, isRoundOver)
  return function()
    if isRoundOver then
      setInputEnabled(self, false)
    end

    local currentText = self.input:getText()
    local inputString = StringUtils.trim(currentText)
    local bestMatch, isExactMatch = self.dictionaryMatcher:getBestMatch(inputString)
    if(bestMatch) then
      processCorrectAnswer(self, bestMatch, inputString, isRoundOver)
      if(not isExactMatch) then
        local bgColors = WitColors.getBackgroundGradientByColorSchemeIndex(self.roundNumber)
        self.input:showAutocorrectMessage(bgColors[1])
      end
    else
      processIncorrectAnswer(self, inputString, isRoundOver)
    end
  end
end

local function handleRoundTimeoutInputIfNeeded(self, inputConfirmHandler)
  if (#self.input:getText() > 0) then
    inputConfirmHandler()
    waitFor(0.667)
  end
end

local function getPassHandler(self, passRoundCallback)
  return function()
    self.input:setTouchEnabled(false)
    self.roundTimer:setTimeRemaining(0.0)
    if (passRoundCallback ~= nil) then
      passRoundCallback()
    end
  end
end

local function createLayout(self, roundTimer, passRoundCallback)
  local progress = DotProgressBar.new(3, WitColors.getProgressBarColorsByRound())
  local input = WitInputField.new(getPassHandler(self, passRoundCallback))
  input:setAnchorPointBottom()

  local baseH = Screen.getSafeHeightPart(0.5)
  local advancedWordNotificationHeight = 24
  local dy = 5
  -- TODO: switch to everything having (0, 0) anchor point, that will severely simplify the layout logic
  -- (requires removing fake backgrounds from all the independent components)
  local topMargin = 60 + (Screen.fullHeight - Screen.safeTop)
  local leftMargin = X_MARGIN + Screen.safeLeft
  local heightLeft = Screen.fullHeight - baseH - 2 * dy - topMargin - advancedWordNotificationHeight
  local collectedWordsHeight = heightLeft/2
  local sentenceHeight = heightLeft - collectedWordsHeight
  local textAreaWidth = Screen.safeWidth - leftMargin * 2

  local background = WitBackground.new()

  local collectedWords = WitCollectedWordsView.new(textAreaWidth, collectedWordsHeight)
  local sentence = WitImprovableSentenceView.new(textAreaWidth)
  local advancedWordNotification = WitAdvancedWordNotification.new(input:getHeight())

  sentence:setAnchorPointLeft()
  collectedWords:setAnchorPointLeft()

  local progressY = Screen.getSafeY(1) - Y_DISTANCE_FROM_TOP_TO_PROGRESS_BAR
  local sentenceX, sentenceY = leftMargin, Screen.getSafeY(0.5)
  local collectedWordsX, collectedWordsY = leftMargin, Screen.getSafeY(0.5)

  advancedWordNotification:setX(input:getWidthPart(0.5))
  input:setLoc(Screen.getSafeX(0.5), input:getHeightPart(-1.0))
  sentence:setLoc(sentenceX, sentenceY)
  collectedWords:setLoc(collectedWordsX, collectedWordsY)
  progress:setLoc(Screen.getSafeX(0.5), progressY)

  input:addDigitTimer(roundTimer)
  self:insertAtCenter(background)
  self:insert(progress)
  self:insert(input)
  self:insert(collectedWords)
  self:insert(sentence)
  input:insert(advancedWordNotification)

  advancedWordNotification:hide()

  self.collectedWords = collectedWords
  self.sentence = sentence
  self.advancedWordNotification = advancedWordNotification
  self.input = input
  self.background = background
  self.progress = progress
  self.leftMargin = leftMargin
end


local function clearRound(self)
  if(self.taskView) then
    self.taskView:removeSelf()
    self.taskView = nil
  end
  self.collectedWords:clear()
  self.collectedWords:resetSizing()
end

local function createRoundLabel(self, roundNumber)
  local labelFont = Font.new(FontFaces.DIN_OT_MEDIUM, 36)
  local roundLabel = Label.new("Round "..roundNumber, labelFont)
  roundLabel:setOpacity(0.0)
  roundLabel:setScl(2.0)
  roundLabel:setLoc(self.leftMargin + roundLabel:getWidthPart(0.5), Screen.getHeightPart(1.5))
  self:insert(roundLabel)

  self.roundLabel = roundLabel
end

local function createTaskView(self, round)
  local taskColor = WitColors.getHeaderColorForRound(self.roundNumber)
  local taskView = WitTaskView.new(round.task, taskColor)
  taskView:setScl(1.5)
  taskView:setLoc(self.leftMargin, Screen.getSafeY(0.5))
  self:insert(taskView)
  self.taskView = taskView
end

local function updateColorScheme(self, roundNumber)
  local highlightColor = WitColors.getHighlightColorForRound(roundNumber)
  local collectedWordsColor = WitColors.getCollectedWordsColorForRound(roundNumber)

  self.background:setColorScheme(roundNumber)
  self.sentence:setActiveColor(highlightColor)
  self.collectedWords:setWordColor(collectedWordsColor)
end

local function getFinalElementLocations(self)
  local progressY = self.progress:getY()
  local taskOffset = self.taskView:getHeightPart(0.5)
  local sentenceOffset = self.sentence:getHeightPart(0.5)
  local wordsOffset = self.collectedWords:getHeightPart(0.5)

  local taskY = progressY - taskOffset - Y_DISTANCE_BETWEEN_PROGRESS_BAR_TO_TASK
  local sentenceY = taskY - sentenceOffset - Y_DISTANCE_BETWEEN_TASK_AND_SENTENCE
  local wordsY = sentenceY - wordsOffset - sentenceOffset - Y_DISTANCE_BETWEEN_SENTENCE_AND_ANSWERS

  return taskY, sentenceY, wordsY
end

local function createShowRoundAction(self, round, roundNumber)
  self.sentence:setOpacity(0.0)

  self.round = round
  self.dictionary = WitDictionary.new(round)
  self.dictionaryMatcher = DictionaryWordsMatcher.new(self.dictionary)

  createRoundLabel(self, roundNumber)
  createTaskView(self, round)
  updateColorScheme(self, roundNumber)
  self.sentence:setSentence(round.prompt)

  self.collectedWords:setOpacity(1.0)
  -- Need to duplicate this call here, because otherwise the highlighted word is visible
  -- (If we only use the call here, there is a first frame bug)
  self.sentence:setOpacity(0.0)
  self.sentence:setY(Screen.getSafeY(0.4))

  local roundStartTime = 0.2

  local roundLabel = self.roundLabel
  local taskView = self.taskView
  local animationBaseTime = 1.0

  local taskY, sentenceY, wordsY = getFinalElementLocations(self)
  self.collectedWords:setLoc(self.leftMargin, wordsY)

  local replaceWord = StringUtils.findSubstringsInTags(round.prompt, "replace")[1]
  self.input:startNewRound(replaceWord, round.task)

  local roundLabelAction = AHDelayed.new(
    0.1,
    AHScaleTo.new(roundLabel, animationBaseTime, AHEaseElasticOut.new(1.0), 1.0),
    AHLocationYTo.new(roundLabel, animationBaseTime, AHEaseElasticOut.new(1.0), Screen.getSafeY(0.57)),
    AHDelayed.new(0.1, AHOpacityTo.new(roundLabel, animationBaseTime, nil, 1.0))
  )

  local animateInAction = AHSequence.new(
    roundLabelAction,
    taskView:createAnimateInAction()
  )

  local easeOnOut = AHEaseExponentialInOut.new()
  local roundLabelOffset = (roundLabel:getY() - taskView:getY()) / 10
  local animateOutAction = AHSpawn.new(
    AHOpacityTo.new(roundLabel, animationBaseTime * 0.75, nil, 0.0),
    AHLocationYTo.new(roundLabel, animationBaseTime * 1.5, easeOnOut, taskY + roundLabelOffset),
    AHLocationYTo.new(taskView, animationBaseTime * 1.5, easeOnOut, taskY),
    AHScaleTo.new(taskView, animationBaseTime * 1.5, easeOnOut, 1.0),
    AHLocationYTo.new(self.sentence, animationBaseTime * 1.5, easeOnOut, sentenceY),
    AHDelayed.new(
      animationBaseTime * 0.65,
      self.progress:createAnnounceRoundAction(roundNumber),
      self.sentence:createFadeAction(roundStartTime, 1.0)
    )
  )

  local textAction = AHSequence.new(
    animateInAction,
    AHDelay.new(0.25),
    animateOutAction,
    function() roundLabel:removeSelf() end
  )

  local showAndStartTimer = function()
    self.roundTimer:setOpacity(1.0)
    self.roundTimer:start()
  end

  local showRoundAction = AHSequence.new(
    textAction,
    showAndStartTimer,
    createEnableInputAction(self, roundStartTime)
  )
  return showRoundAction
end

local function createHideRoundViewsAnimation(self)
  local hideTaskSeconds = 0.3
  local hideCollectedWordsSeconds = 0.3
  local hideSentenceSeconds = 0.3
  local hideTimerSeconds = 0.3

  local hideCollectedWords = AHOpacityTo.new(self.collectedWords, hideCollectedWordsSeconds, nil, 0)
  local hideSentence = self.sentence:createFadeAction(hideSentenceSeconds)
  local hideTimer = AHOpacityTo.new(self.roundTimer, hideTimerSeconds, nil, 0)
  local disableInput = function()
    setInputEnabled(self, false)
    clearInput(self)
    createShowDisableInputAction(self, 0.6):start()
  end
  local hideTask = self.taskView:createFadeAction(hideTaskSeconds, 0.0)
  local animation = AHSpawn.new(hideCollectedWords, hideSentence, hideTimer, disableInput, hideTask)

  return animation
end

local function initUnusedWordsView(self, colorSchemeIndex)
  colorSchemeIndex = colorSchemeIndex or self.roundNumber
  local suggestionColor = WitColors.getWordSuggestionsColor(colorSchemeIndex)
  local unusedWords = WitUnusedWordsView.new(self.dictionary, self.round.suggestions, nil, nil, suggestionColor)
  unusedWords:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.625))
  self:insert(unusedWords)
  return unusedWords
end

local function createEmailText(self, colorSchemeIndex)
  local topText = "Has a word been scored incorrectly?"
  local bottomText = "Email eloquence@elevateapp.com"

  local font = Font.new(FontFaces.DIN_OT_MEDIUM, 13)
  local topLabel = Label.new(topText, font)

  local bottomLabel = Label.new(bottomText, font)
  local highlightColor = WitColors.getWordSuggestionsColor(colorSchemeIndex)
  bottomLabel:setColor(unpack(highlightColor))

  local textColumn = Column.new(Screen.getHeightPart(0.01))
  textColumn:insert(topLabel)
  textColumn:insert(bottomLabel)
  textColumn:alignChildren()
  local isRecursive = true
  textColumn:linkOpacityAttrToChildren(isRecursive)
  textColumn:setOpacity(0.0)

  textColumn:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.075))
  self:insert(textColumn)

  return textColumn
end

local function setupKeyboard(self)
  local baseValidateCallback = DeviceKeyboard.createCharacterBasedValidationCallback(VALID_KEYS)

  DeviceKeyboard.setValidateCallback(function(replacementText, resultingText)
    local enteredInvalidCharacter = not baseValidateCallback(replacementText)
    local willExceedMaxInputLength = (#resultingText > MAX_INPUT_LENGTH)
    local enteredLeadingWhitespace = (resultingText == " ")
    local enteredExcessiveTrailingWhitespace = containsConsecutiveWhitespace(resultingText)
    local enteredSecondWord = containsPrintableCharacterAfterWhitespace(resultingText)

    if (enteredInvalidCharacter or
        willExceedMaxInputLength or
        enteredLeadingWhitespace or
        enteredExcessiveTrailingWhitespace or
        enteredSecondWord) then
      return false
    end

    return true
  end)

  DeviceKeyboard.setWillAppearWithDimensionsCallback(function(originXProportion, originYProportion, widthProportion, heightProportion)
    local yPosition = Screen.getHeightPart(originYProportion + heightProportion)
    self.relocateInputQueue:insert(AHSequence.new(
      AHDelay.new(0.2),
      AHLocationYTo.new(self.input, 0.3, AHEaseExponentialOut.new(), yPosition)
    ))
  end)

  DeviceKeyboard.setInputCallback(function(updatedText)
    if not DeviceKeyboard.isEnabled() and #updatedText > 0 then
      DeviceKeyboard.clear()
    else
      self.input:setText(updatedText)
    end
  end)

  local onReturn = getInputConfirmHandler(self)
  DeviceKeyboard.setReturnCallback(function(answerText)
    onReturn()
  end)

  local showSingleWordWarning = function()
    self.roundTimer:pause()
    DeviceKeyboard.hide()
    KeyboardWarning.newSingleWordWarning():show(function()
      DeviceKeyboard.show()
      self.roundTimer:resume()
    end)
  end

  DeviceKeyboard.setInputRejectedCallback(function(replacementText, resultingText)
    if containsPrintableCharacterAfterWhitespace(resultingText) then
      showSingleWordWarning()
    end
  end)

  self.relocateInputQueue = ActionInterrupt.new()
  DeviceKeyboard.setEnabled(false)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitScene:checkForCorrectAnswerOnRoundTimeout()
  self.input:notifyRoundEnded()
  local isRoundOver = true
  local inputConfirmHandler = getInputConfirmHandler(self, isRoundOver)
  handleRoundTimeoutInputIfNeeded(self, inputConfirmHandler)
end

function WitScene:setRound(round, roundNumber)
  clearRound(self)
  self.roundNumber = roundNumber

  local startRoundAction = createShowRoundAction(self, round, roundNumber)

  startRoundAction:start()
end

function WitScene:stopAnimations()
  self.submitAnimationInterrupt:stop()
end

--- Shows a popup giving stats about the passed round, in anticipation of the next round
-- @param roundStats =
-- {
--    passedRoundNumber = indexOfRoundThatJustEnded,
--    wordsForPassedRound = howManyAnswersWereAcceptedThisRound,
--    wordsForAllRounds = howManyAnswersWereAcceptedAcrossAllRounds,
--    goalWordCount = howManyWordsNeededToPassGame,
-- }
function WitScene:createRoundInterstitialAction(roundStats)
  local suggestionInfo =
  {
    dictionary = self.dictionary,
    suggestions = self.round.suggestions,
    colorSchemeIndex = self.roundNumber
  }
  local interstitialView = InterstitialView.new(roundStats, suggestionInfo)
  self:insert(interstitialView)
  local hideRoundViews = createHideRoundViewsAnimation(self)
  local showInterstitial = interstitialView:initShowAction()
  local roundComplete = self.progress:createCompleteAction(self.roundNumber)
  local waitAfterRoundComplete = AHDelay.new(0.2)

  local interstitialAction = AHSpawn.new(showInterstitial, roundComplete)
  local action = AHSequence.new(hideRoundViews, interstitialAction, waitAfterRoundComplete)

  return action
end


function WitScene:setCorrectAnswerHandler(handler)
  self.correctAnswerHandler = handler
end

function WitScene:setIncorrectAnswerHandler(handler)
  self.incorrectAnswerHandler = handler
end

function WitScene:doIntro()
  self.input:createDisableAction(0.0):start()

  local keyboardSlideSeconds = 0.4
  local waitBeforeSlideInTime = 0.2

  local intro = AHDelay.new(waitBeforeSlideInTime)
  intro:start()
  intro.getDuration = function()
    return (keyboardSlideSeconds + waitBeforeSlideInTime) * 0.9
  end

  return intro
end


function WitScene:createWinOutroAction()
  local colorSchemeIndex = WIN_OUTRO_COLOR_SCHEME_INDEX
  local unusedWords = initUnusedWordsView(self, colorSchemeIndex)

  local tapToContinue = BlinkingSign.newTapToContinue()
  tapToContinue:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.35))
  self:insert(tapToContinue)

  local emailText = createEmailText(self, colorSchemeIndex)

  local offscreenY = -self.input:getHeight()

  local keyboardSlideSeconds = 0.4
  local backgroundColorChanger = function() self.background:setColorScheme(colorSchemeIndex) end
  -- The progress bar is stuck in a state where one of the dots is scaled up
  local restoreProgressBar = self.progress:createNeutralizingAction()
  local hideRoundViews = createHideRoundViewsAnimation(self)
  local slideKeyboard = AHLocationYTo.new(self.input, keyboardSlideSeconds, nil, offscreenY)
  local showCheckmarkAction = self.progress:createUltimateCompleteAction(backgroundColorChanger)
  local showUnusedWords = unusedWords:createShowAction()
  local showEmailText = AHOpacityTo.new(emailText, 0.3, nil, 1.0)
  local staggerUnusedWords = AHDelay.new(0.5)
  local presentUnusedWords = AHSequence.new(staggerUnusedWords, showUnusedWords, showEmailText)

  local endOutroAction = function()
    local tap = TapToContinue.new()
    waitWhile(function() return tap.isArmed end)
  end
  local finishOutro = AHSequence.new(presentUnusedWords, endOutroAction)
  local outroAction = AHSpawn.new(restoreProgressBar, hideRoundViews, slideKeyboard, showCheckmarkAction, finishOutro)

  return outroAction
end

function WitScene:createLoseOutroAction()
  local failedRound = self.roundNumber
  local colorSchemeIndex = LOSE_OUTRO_COLOR_SCHEME_INDEX
  local unusedWords = initUnusedWordsView(self, colorSchemeIndex)

  local tapToContinue = BlinkingSign.newTapToContinue()
  tapToContinue:setLoc(Screen.getSafeX(0.5), Screen.getSafeY(0.4))
  self:insert(tapToContinue)

  local emailText = createEmailText(self, colorSchemeIndex)

  local offscreenY = -self.input:getHeight()

  local keyboardSlideSeconds = 0.4
  local backgroundColorChanger = function() self.background:setColorScheme(colorSchemeIndex) end
  -- The progress bar is stuck in a state where one of the dots is scaled up
  local restoreProgressBar = self.progress:createNeutralizingAction()
  local hideRoundViews = createHideRoundViewsAnimation(self)
  local slideKeyboard = AHLocationYTo.new(self.input, keyboardSlideSeconds, nil, offscreenY)
  local failProgressBarAction = self.progress:createFailAction(failedRound, backgroundColorChanger)
  local fadeOutProgressBar = AHOpacityTo.new(self.progress, 0.3, nil, 0.5)
  local showEmailText = AHOpacityTo.new(emailText, 0.3, nil, 1.0)
  local showUnusedWords = unusedWords:createShowAction()
  local staggerUnusedWords = AHDelay.new(0.5)
  local presentUnusedWords = AHSequence.new(staggerUnusedWords, showUnusedWords, showEmailText)


  local endOutroAction = function()
    local tap = TapToContinue.new()
    waitWhile(function() return tap.isArmed end)
  end
  local finishOutro = AHSequence.new(presentUnusedWords, endOutroAction)
  local outroAction = AHSpawn.new(restoreProgressBar, hideRoundViews, slideKeyboard, fadeOutProgressBar, failProgressBarAction, finishOutro)

  return outroAction
end


-- ----------------------------------------------------------------------------
-- Object construction
-- ----------------------------------------------------------------------------

function WitScene.new(roundTimer, passRoundCallback)
  local self = Group.newFullScreen()
  Mixin.mix(self, WitScene)
  createLayout(self, roundTimer, passRoundCallback)
  self.roundTimer = roundTimer
  self.submitAnimationInterrupt = ActionInterrupt.new()

  setupKeyboard(self)

  return self
end
