module ("BlockPicture", package.seeall)

-------------------------------------------------
----------------- CONFIGURATION -----------------
-------------------------------------------------

BlockPicture.BlockTypes =
{
  Rhombus = 1,
  Square = 2
}

local colors =
{
  {0.8, 0.1, 0},
  {0.1, 0.9, 0.2},
  {0.3, 0.9, 0},
}

local DIAG_MULTIPLIER = math.sqrt(2)


-------------------------------------------------
----------------- UTILITY -----------------------
-------------------------------------------------

local function getColor(self, type)
  return self.color
end

local function getDiagSize(self, type)
  if(self.knownDiagSizes[type]) then
    return self.knownDiagSizes[type]
  end
  if(type == BlockPicture.BlockTypes.Rhombus) then
    -- For we rotate it by 45 degrees
    self.knownDiagSizes[type] = math.sqrt(self.blockW * self.blockW + self.blockH * self.blockH)
  end
  return self.knownDiagSizes[type]
end

local function createBlock(self, type)
  local color = getColor(self, type)
  if(type == BlockPicture.BlockTypes.Rhombus) then
    local diagSize = getDiagSize(self, type)
    local view = FilledRectangle.new(diagSize, diagSize, color)
    view:setRot(45)
    return view
  elseif(type == BlockPicture.BlockTypes.Square) then
    local view = FilledRectangle.new(self.blockW, self.blockH, color)
    return view
  end
end

local function getBlockXY(self, type, gridX, gridY)
  local blockW, blockH = self.blockW, self.blockH
  local gridH = self.gridH
  local diagSize = getDiagSize(self, type)
  local x = blockW * gridX
  local y = blockH * (gridH - gridY)

  if(type == BlockPicture.BlockTypes.Rhombus) then
    y = y + 3 * DIAG_MULTIPLIER * blockH/4
  elseif(type == BlockPicture.BlockTypes.Square) then
    y = y + blockH/2
  end

  return x, y
end

local function updateBounds(minX, maxX, minY, maxY, x, y, blockType, block)
  local w, h = block:getDims()
  local minXCandidate = x
  local maxXCandidate = x
  local minYCandidate = y
  local maxYCandidate = y
  if(blockType == BlockPicture.BlockTypes.Rhombus) then
    -- This calculates the honest bounds, accounting for the edges of the rhombuses that stick out
    minXCandidate = x - 4 * math.sqrt(2) * w/8
    maxXCandidate = x + 4 * math.sqrt(2) * w/8
    minYCandidate = y - 4 * math.sqrt(2) * h/8
    maxYCandidate = y + 4 * math.sqrt(2) * h/8
  elseif(blockType == BlockPicture.BlockTypes.Square) then
    maxXCandidate = x + w
    maxYCandidate = y + h/2
    minYCandidate = y - h/2
  end

  if(minXCandidate < minX) then minX = minXCandidate end
  if(maxXCandidate > maxX) then maxX = maxXCandidate end
  if(minYCandidate < minY) then minY = minYCandidate end
  if(maxYCandidate > maxY) then maxY = maxYCandidate end

  return minX, maxX, minY, maxY
end

local function createLayout(self, shape, blockW, blockH, color)
  self.gridH = #shape
  self.gridW = #shape[1]
  local minX = math.huge
  local maxX = -math.huge
  local minY = math.huge
  local maxY = -math.huge
  local views = {}
  local picture = Group.newDeprecated()

  for gridY = 1, self.gridH do
    local row = shape[gridY]
    local viewRow = {}
    views[#views + 1] = viewRow
    for gridX = 1, self.gridW do
      local blockType = row[gridX]
      local block = createBlock(self, blockType)
      if(block) then
        local x, y = getBlockXY(self, blockType, gridX, gridY)
        minX, maxX, minY, maxY = updateBounds(minX, maxX, minY, maxY, x, y, blockType, block)
        block:setLoc(x, y)
        picture:insert(block)
        viewRow[gridX] = block
      end
    end
  end

  self:insert(picture)
  picture:setLoc(-minX/2, -minY/2)

  local w, h = maxX - minX, maxY - minY
  self:setBounds(0, 0, 0, w, h, 0)
  self:setAnchorPointCenter()

  self.views = views
end

-------------------------------------------------
--------------------- API -----------------------
-------------------------------------------------
function BlockPicture:iterateViews(iterator, rowCallback)
  for gridY, row in ipairs(self.views) do
    rowCallback(row, gridY)
    for gridX, view in pairs(row) do
      iterator(view, gridX, gridY)
    end
  end
end

function BlockPicture:getRowCount()
  return #self.views
end

function BlockPicture:getView(x, y)
  -- Stored row first
  return self.views[y][x]
end

function BlockPicture:getGridWidth()
  return self.gridW
end

function BlockPicture:getGridHeight()
  return self.gridH
end

-------------------------------------------------
----------------- CONSTRUCTOR -------------------
-------------------------------------------------
function BlockPicture.new(shape, blockW, blockH, color)
  local self = Group.newDeprecated()
  Mixin.mix(self, BlockPicture)
  self.knownDiagSizes = {}
  self.blockW = blockW
  self.blockH = blockH
  self.color = color
  createLayout(self, shape, blockW, blockH, color)
  return self
end
