module ("DotProgressBar", package.seeall)
require("scene.effects.BlockCheckmark")
require("scene.effects.BlockPictureFadeInAction")
require("scene.effects.BlockPictureFlowUpAction")

-------------------------------------------------
----------------- CONFIGURATION -----------------
-------------------------------------------------

local DEFAULT_COLORS =
{
  initial =
  {{1, 0, 0}, {0.5, 0.5, 0.5}, {0, 0, 1}},
  complete =
  {{0, 1, 0}, {0, 1, 0}, {0, 1, 0}},
}

local DOT_SCALE_AMPLITUDE_ANNOUNCE = 1.5
local DOT_SCALE_AMPLITUDE_IMPACT = 1.6
local CHECKMARK_BLOCK_SIZE = 12


-------------------------------------------------
----------------- UTILITY -----------------------
-------------------------------------------------

local function createLayout(self, n, r, dx, colors)
  local x = 0
  local dots = {}
  local colorIndex = 1
  local totalColors = #colors.initial
  for i = 1, n do
    local dot = FilledCircle.new(r, colors.initial[colorIndex])
    colorIndex = colorIndex % totalColors + 1
    self:insert(dot)
    dot:setLoc(x, r)

    x = x + r + dx
    dots[i] = dot
    dot:inheritColor(self)
  end
  local w = x - dx - r
  local h = 2 * r
  self:setBounds(0, 0, 0, w, h, 0)
  self:setAnchorPointCenter()

  local checkmark = BlockCheckmark.new(CHECKMARK_BLOCK_SIZE, CHECKMARK_BLOCK_SIZE, WitColors.getCorrectAnswerColor())
  local checkmarkW, checkmarkH = checkmark:getDims()
  local checkmarkX, checkmarkY = w/2, -checkmarkH/2 - h - 16
  self:insert(checkmark)

  checkmark:setVisibleRecursive(false)
  local showCheckmark = function()
    checkmark:setScl(1, 1)
    checkmark:setLoc(checkmarkX, checkmarkY)
    checkmark:setVisibleRecursive(true)
  end
  local checkmarkFadeInDuration = 0.8
  local checkmarkFadeIn = BlockPictureFadeInAction.new(checkmark, checkmarkFadeInDuration)
  local checkmarkSlideIn = BlockPictureFlowUpAction.new(checkmark)

  self.checkmarkAnimation = AHSpawn.new(checkmarkFadeIn, showCheckmark)
  self.finalCheckmarkAnimation = AHSpawn.new(checkmarkSlideIn, showCheckmark)

  self.checkmarkAnimation.getDuration = function()
    return checkmarkFadeInDuration
  end

  self.checkmark = checkmark
  self.dots = dots
  self.colors = colors
end

local function createRestoreScaleActionForOtherDots(self, excludedDotNumber, duration, ease)
  local result = AHSpawn.new()
  for i = 1, #self.dots do
    if(i ~= excludedDotNumber) then
      local restoreScale = AHScaleTo.new(self.dots[i], duration, ease, 1, 1)
      result:addAction(restoreScale)
    end
  end
  return result
end

-------------------------------------------------
--------------------- API -----------------------
-------------------------------------------------
function DotProgressBar:createAnnounceRoundAction(roundNumber)
  local dot = self.dots[roundNumber]

  local dotScaleUpTime = 0.2
  local dotShowTime = 0.3
  local dotScaleDownTime = 0.2
  local scaleAmplitude = DOT_SCALE_AMPLITUDE_ANNOUNCE

  local scaleUpEase = AHEaseSimpleOut.new()
  local scaleDownEase = AHEaseSimpleIn.new()

  local restoreScaleForOtherDots = createRestoreScaleActionForOtherDots(self, roundNumber, dotScaleUpTime, scaleUpEase)
  local dotScaleUp = AHScaleTo.new(dot, dotScaleUpTime, scaleUpEase, scaleAmplitude, scaleAmplitude)
  local dotScaleDown = AHScaleTo.new(dot, dotScaleDownTime, scaleDownEase, 1, 1)

  local presentCurrentDot = AHSpawn.new(dotScaleUp, restoreScaleForOtherDots)
  local finishDotAnnouncement = dotScaleDown
  local dotAnimation = AHSequence.new(presentCurrentDot, finishDotAnnouncement)

  return dotAnimation
end

function DotProgressBar:createShowCheckmarkAction()
  return self.checkmarkAnimation
end

function DotProgressBar:createNeutralizingAction()
  local duration = 0.2
  local ease = nil
  return createRestoreScaleActionForOtherDots(self, nil, duration, ease)
end


function DotProgressBar:createUltimateCompleteAction(dotColorChangeHandler)
  local n = #self.dots
  local dot = self.dots[n]
  local animationTime = 0.3
  local dotScaleUpTime = animationTime/3
  local dotScaleDownTime = animationTime/3
  local dotColorTime = dotScaleDownTime
  local dotStaggerTime = animationTime * 0.95

  local dotScaleAmp = DOT_SCALE_AMPLITUDE_IMPACT

  local scaleDotUp = AHScaleTo.new(dot, dotScaleUpTime, nil, dotScaleAmp, dotScaleAmp)
  local scaleDotDown = AHScaleTo.new(dot, dotScaleDownTime, nil, 1, 1)
  local staggerDot = AHDelay.new(dotStaggerTime)
  local colorDot = AHColorTo.new(dot, dotColorTime, nil, self.colors.complete[n])
  local arouseDot = AHSpawn.new(colorDot, scaleDotUp, dotColorChangeHandler)
  local dotAnimation = AHSequence.new(arouseDot, scaleDotDown)

  local ultimateCompleteAction = AHSpawn.new(self.finalCheckmarkAnimation, dotAnimation)
  return ultimateCompleteAction
end


local function createDotScaleUpDownColorAction(self, dotNumber, targetColor, dotColorChangeHandler)
  local dot = self.dots[dotNumber]
  local animationTime = 0.3
  local dotScaleUpTime = animationTime/2
  local dotScaleDownTime = animationTime/2
  local dotColorTime = dotScaleDownTime

  local dotScaleAmp = DOT_SCALE_AMPLITUDE_IMPACT

  local scaleDotUp = AHScaleTo.new(dot, dotScaleUpTime, nil, dotScaleAmp, dotScaleAmp)
  local scaleDotDown = AHScaleTo.new(dot, dotScaleDownTime, nil, 1, 1)
  local colorDot = AHColorTo.new(dot, dotColorTime, nil, targetColor)
  local arouseDot = AHSpawn.new(colorDot, scaleDotUp, dotColorChangeHandler)
  local dotAnimation = AHSequence.new(arouseDot, scaleDotDown)

  return dotAnimation
end
function DotProgressBar:createFailAction(dotIndex, dotColorChangeHandler)
  return createDotScaleUpDownColorAction(self, dotIndex, self.colors.failed[dotIndex], dotColorChangeHandler)
end

function DotProgressBar:createCompleteAction(dotIndex)
  return createDotScaleUpDownColorAction(self, dotIndex, self.colors.complete[dotIndex])
end

-------------------------------------------------
----------------- CONSTRUCTOR -------------------
-------------------------------------------------
function DotProgressBar.new(numberOfDots, colors, r, dx)
  numberOfDots = numberOfDots or 3
  r = r or 4
  dx = dx or 24
  colors = colors or DEFAULT_COLORS
  local self = Group.newDeprecated()
  Mixin.mix(self, DotProgressBar)
  createLayout(self, numberOfDots, r, dx, colors)
  return self
end
