--- WitValidator
-- @author Created by alexeychernikov on July 15, 2014.
-- ConceptValidator for Wit.

class("WitValidator")

require("validators.ConceptValidator")
require("WitSolutionExamples")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  "prompt",
  "common",
  "advanced",
  "task",
  "difficulty",
  "replaceWord",
  "suggestList"
}

local REQUIRED_EMPTY_ALLOWED_KEYS = {
  "noAccept"
}

local MINIMUM_COMMON_ITEMS = 6
local MINIMUM_ADVANCED_ITEMS = 2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateItemCount(self, commonItems, advancedItems)
  self:expectGreaterOrEqual(#commonItems, MINIMUM_COMMON_ITEMS, "Not enough common items found")
  self:expectGreaterOrEqual(#advancedItems, MINIMUM_ADVANCED_ITEMS, "Not enough advanced items found")
end

local function validateUniqueAndLowercaseItems(self, commonItems, advancedItems)
  local commonCopy = TableUtils.shallowCopy(commonItems)
  local advancedCopy = TableUtils.shallowCopy(advancedItems)
  table.sort(commonCopy)
  table.sort(advancedCopy)

  for i = 1, #commonCopy - 1 do
    local word = commonCopy[i]
    local lowerCase = string.lower(word)
    self:expectEqual(word, lowerCase, "Word is not lowercase in common items")
    self:expectNotEqual(word, commonCopy[i+1], "Duplicate item found in common items")
  end

  local lastCommonWord = commonCopy[#commonCopy]
  local lowerCaseCommonWord = string.lower(lastCommonWord)
  self:expectEqual(lastCommonWord, lowerCaseCommonWord, "Word is not lowercase in common items")

  for i = 1, #advancedCopy - 1 do
    local word = advancedCopy[i]
    local lowerCase = string.lower(word)
    self:expectEqual(word, lowerCase, "Word is not lowercase in advanced items")
    self:expectNotEqual(word, advancedCopy[i+1], "Duplicate item found in advanced items")
  end

  local lastAdvancedWord = advancedCopy[#advancedCopy]
  local lowerCaseAdvancedWord = string.lower(lastAdvancedWord)
  self:expectEqual(lastAdvancedWord, lowerCaseAdvancedWord, "Word is not lowercase in advanced items")

  local commonSet = TableUtils.createSetFromArray(commonItems)
  local advancedSet = TableUtils.createSetFromArray(advancedItems)
  local repeatedSet = TableUtils.intersectionSet(commonSet, advancedSet)
  for word, _ in pairs(repeatedSet) do
    self:expectTrue(false, string.format("Duplicate item found in common and advanced items (%s)", word))
  end
end

local function validateSuggestedItems(self, suggestedItems, commonItems, advancedItems)
  local suggestSet = TableUtils.createSetFromArray(suggestedItems)
  local commonSet = TableUtils.createSetFromArray(commonItems)
  local advancedSet = TableUtils.createSetFromArray(advancedItems)
  local allItems = TableUtils.unionSet(commonSet, advancedSet)

  local unexpectedSet = TableUtils.differenceSet(suggestSet, allItems)
  local unexpectedItems = TableUtils.createArrayFromSet(unexpectedSet)
  local errorMessage = string.format("Items in suggested list should belong to either common or advanced. Unexpected items found: %s", StringUtils.tostring(unexpectedItems))
  self:expectTrue(#unexpectedItems == 0, errorMessage)
end

local function validateNoAcceptItems(self, noAcceptItems, commonItems, advancedItems)
  local noAcceptSet = TableUtils.createSetFromArray(noAcceptItems)
  local commonSet = TableUtils.createSetFromArray(commonItems)
  local advancedSet = TableUtils.createSetFromArray(advancedItems)
  local allItems = TableUtils.unionSet(commonSet, advancedSet)

  local unexpectedSet = TableUtils.intersectionSet(allItems, noAcceptSet)
  local unexpectedItems = TableUtils.createArrayFromSet(unexpectedSet)
  local errorMessage = string.format("Items in noAccept should not belong to common nor advanced. Unexpected items found: %s", StringUtils.tostring(unexpectedItems))
  self:expectTrue(#unexpectedItems == 0, errorMessage)
end

local function validateHasExample(self, concept, noAcceptItems)
  if(#noAcceptItems < 1) then return end

  local task = concept.task
  local hasExample = WitSolutionExamples.hasExampleForTask(task)
  self:expectTrue(hasExample, "No example for task type: "..task)
end

local function validateReplaceWordIsNotInOtherSets(self, replaceWord, commonItems, advancedItems, suggestedItems)
  local commonSet = TableUtils.createSetFromArray(commonItems)
  local advancedSet = TableUtils.createSetFromArray(advancedItems)
  local suggestSet = TableUtils.createSetFromArray(suggestedItems)

  self:expectTrue((commonSet[replaceWord] == nil), "Prompt was found in the common set")
  self:expectTrue((advancedSet[replaceWord] == nil), "Prompt was found in the advanced set")
  self:expectTrue((suggestSet[replaceWord] == nil), "Prompt was found in the suggested words set")
end

local function validatePromptContainsReplaceWord(self, prompt, replaceWord)
  local replaceTagStart = "<replace>"
  local replaceTagEnd = "</replace>"

  local indexToBeginSearch = 1
  local willUsePlainTextSearch = true
  local _, startTagEndIndex = prompt:find(replaceTagStart, indexToBeginSearch, willUsePlainTextSearch)
  local endTagStartIndex = prompt:find(replaceTagEnd, startTagEndIndex, willUsePlainTextSearch)

  local wordBetweenTagsStartIndex = startTagEndIndex + 1
  local wordBetweenTagsEndIndex = endTagStartIndex - 1
  local wordBetweenTags = prompt:sub(wordBetweenTagsStartIndex, wordBetweenTagsEndIndex)

  local errorString = string.format("Word between replace tags (%s) in the prompt didn't match the expected replace word (%s)", wordBetweenTags, replaceWord)
  self:expectTrue(wordBetweenTags == replaceWord, errorString)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WitValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  self:validateRequiredKeys(concept, REQUIRED_EMPTY_ALLOWED_KEYS, true)

  local replaceWord = concept.replaceWord
  local commonItems = TableUtils.csvToArray(concept.common)
  local advancedItems = TableUtils.csvToArray(concept.advanced)
  local suggestedItems = TableUtils.csvToArray(concept.suggestList)
  local noAcceptItems = TableUtils.csvToArray(concept.noAccept)
  local prompt = concept.prompt

  validateItemCount(self, commonItems, advancedItems)
  validateUniqueAndLowercaseItems(self, commonItems, advancedItems)
  validateSuggestedItems(self, suggestedItems, commonItems, advancedItems)
  validateNoAcceptItems(self, noAcceptItems, commonItems, advancedItems)
  validateHasExample(self, concept, noAcceptItems)
  validateReplaceWordIsNotInOtherSets(self, replaceWord, commonItems, advancedItems, suggestedItems)
  validatePromptContainsReplaceWord(self, prompt, replaceWord)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WitValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, WitValidator)

  return self
end
