--- WordBirdsBackground.lua
-- @author created by Jay Austria on April 26, 2016

class("WordBirdsBackground")
require("WordBirdsBird")
require("WordBirdsBirdData")
require("widgets.Scroller")
require("WordBirdsFallingStarGenerator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

-- Data values used in BirdData -------
local FACING_LEFT = 1
local FACING_RIGHT = 2
local BIRD_TYPE_PEEK = 3
local BIRD_TYPE_PERCH = 4
local BIRD_TYPE_HOP = 5

-- Z Order ----------------------------

local Z_SKY = 10
local Z_CLOUDS = 20
local Z_STARS = 30
local Z_TREES = 40
local Z_BIRDS_LOW = 50
local Z_WINDOWS = 60
local Z_HOUSES = 70
local Z_BIRDS_HIGH = 80

local Z_BIRDS = {
  [BIRD_TYPE_PEEK] = Z_BIRDS_LOW,
  [BIRD_TYPE_PERCH] = Z_BIRDS_HIGH,
  [BIRD_TYPE_HOP] = Z_BIRDS_HIGH
}

-- Other ------------------------------

local BIRD_CLASSES = {
  [BIRD_TYPE_PEEK] = WordBirdsBirdTypePeek,
  [BIRD_TYPE_PERCH] = WordBirdsBirdTypePerch,
  [BIRD_TYPE_HOP] = WordBirdsBirdTypeHop
}

local BIRD_X_LEFT = -Screen.getWidthPart(0.2)
local BIRD_X_RIGHT = Screen.getWidthPart(1.4)

local SCENE_COUNT = 3

local SCENE_TRANSITION_DURATION = 10

local SCROLL_SPEED = 8

local SKY_OPACITY = 0.15

local SKY_TINTS = {
  {ColorUtils.colorHex24ToStd(0x1fa3bc),  ColorUtils.colorHex24ToStd(0x72c3d9)},
  {ColorUtils.colorHex24ToStd(0x528aa6),  ColorUtils.colorHex24ToStd(0x9b7c8e)},
  {ColorUtils.colorHex24ToStd(0x062236),  ColorUtils.colorHex24ToStd(0x006374)}
}

local STAR_TINTS = {
  {1, 1, 1, 0},
  {1, 1, 1, 0},
  {1, 1, 1, 1}
}

local CLOUD_TINTS = {
  ColorUtils.colorHex32ToStd(0xcaedf450),
  ColorUtils.colorHex32ToStd(0x54396c44),
  ColorUtils.colorHex32ToStd(0x00193628)
}

local TREE_TINTS = {
  ColorUtils.colorHex24ToStd(0xcaedf4),
  ColorUtils.colorHex24ToStd(0xddc1cc),
  ColorUtils.colorHex24ToStd(0x25394d)
}

local HOUSE_TINTS = {
  ColorUtils.colorHex24ToStd(0xffffff),
  ColorUtils.colorHex24ToStd(0xcccccc),
  ColorUtils.colorHex24ToStd(0x889099)
}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function proportionalToAbsolute(value, proportionOf)
  if value > -1 and value < 1 then
    return value * proportionOf
  end
  return value
end

local function toHorizontalValue(locX)
  return Screen.getWidthPart(0.5) + proportionalToAbsolute(locX, Screen.fullWidth)
end

local function toVerticalValue(locX)
  return proportionalToAbsolute(locX, Screen.fullHeight)
end

local function processBirdData(data)
  for _, birdData in ipairs(data) do
    if birdData.facing == FACING_RIGHT then
      birdData.fly_in_from_x = BIRD_X_LEFT
      birdData.fly_out_to_x = BIRD_X_RIGHT
    else
      birdData.fly_in_from_x = BIRD_X_RIGHT
      birdData.fly_out_to_x = BIRD_X_LEFT
    end

    birdData.fly_in_from_y = toVerticalValue(birdData.fly_in_from_y)
    birdData.fly_in_to_x = toHorizontalValue(birdData.fly_in_to_x)
    birdData.fly_in_to_y = toVerticalValue(birdData.fly_in_to_y)
    birdData.fly_out_to_y = toVerticalValue(birdData.fly_out_to_y)
  end

  return data
end

local function insertSky(self)
  local selfW, selfH = self:getDims()

  self.skyGradient = RectangleGradient.new(selfW, selfH, SKY_TINTS[1][1], SKY_TINTS[1][2])
  self:insertAtCenter(self.skyGradient, Z_SKY)

  self.stars = Sprite.newWithName("stars.png")
  self.stars:setEffectiveDims(selfW, selfH*0.66)
  self.stars:setAnchorPointTop()
  self.stars:setLoc(selfW/2, selfH)
  self:insert(self.stars, Z_SKY)

  self.sky = Sprite.newWithName("wordbirds_sky.png")
  self.sky:setEffectiveDims(selfW, selfH)
  self.sky:setOpacity(SKY_OPACITY)
  self:insertAtCenter(self.sky, Z_SKY)
end

local function insertTrees(self)
  self.trees = Sprite.newWithName("wordbirds_trees.png")
  self.trees:setAnchorPointBottom()
  self.trees:setX(self:getWidthPart(0.5))
  self:insert(self.trees, Z_TREES)
end

local function insertHouses(self)
  self.houses = Sprite.newWithName("wordbirds_houses.png")
  self.houses:setAnchorPointBottom()

  local housesWidth = self.houses:getWidth()
  local scaleToFitScreen = math.max(1.0, Screen.width / housesWidth)
  self.houses:setScl(scaleToFitScreen)

  local housesYOffset = -1
  self.houses:setLoc(self:getWidthPart(0.5), housesYOffset)

  self:insert(self.houses, Z_HOUSES)

  self.bottomMargin = self.houses:getHeight()
end

local function insertWindows(self, index)
  local windowSpriteName = "wordbirds_windows_unlit.png"
  if index > 1 then
    windowSpriteName = "wordbirds_windows_lit.png"
  end

  self.windows = Sprite.newWithName(windowSpriteName)

  local housesScale = self.houses:getScl()
  self.windows:setScl(housesScale)

  local windowsXOffset = 38 * housesScale
  local windowsYOffset = 31 * housesScale
  self.windows:setLoc(self:getWidthPart(0.5) - windowsXOffset, windowsYOffset)

  self:insert(self.windows, Z_WINDOWS)
end


local function makeSmallCloudScroller()
  local assets = {
    Sprite.newWithName("clouds/clouds_1.png"),
    Sprite.newWithName("clouds/clouds_1.png")
  }
  for _, asset in ipairs(assets) do
    asset:setScl(4)
  end
  local scroller = FixedScroller.new(assets, true)
  local positionedScroller = PositionedScroller.new(scroller, Screen.fullWidth, Screen.fullHeight, Screen.getWidthPart(0.5), Screen.getHeightPart(0.52))
  return positionedScroller
end

local function makeBigCloudScroller(type, x, y)
  local assets = {}
  assets[Sprite.newWithName(string.format("clouds/clouds_%d.png", type))] = {x=math.floor(x), y=math.floor(y)}
  for asset, _ in pairs(assets) do
    asset:setScl(4)
  end

  local scroller = ItemScroller.new(assets, Screen.fullWidth*2, Screen.fullWidth*2, Screen.fullHeight, true)
  return scroller
end

local function insertClouds(self)
  local width = Screen.fullWidth * 2
  local height = Screen.fullHeight
  local scrollers = {
    makeSmallCloudScroller(),
    makeBigCloudScroller(2, width * 0.00, height * 0.25),
    makeBigCloudScroller(3, width * 0.25, height * 0.1),
    makeBigCloudScroller(4, width * 0.5, height * 0.37),
    makeBigCloudScroller(5, width * 0, height * 0.25)
  }

  local scrollerProportions = {
    1.1,
    1.2,
    1.4,
    1.8,
    2.0
  }

  self.clouds = ParallaxScroller.new(scrollers, scrollerProportions)
  self.clouds:linkColorAttrToChildren(true)
  self:insertAtCenter(self.clouds, Z_CLOUDS)
end

local function newTintAction(self, index, duration)
  local treeColor = TREE_TINTS[index]
  local skyStartColor = SKY_TINTS[index][1]
  local skyEndColor = SKY_TINTS[index][2]
  local starsColor = STAR_TINTS[index]
  local cloudColor = CLOUD_TINTS[index]
  local houseColor = HOUSE_TINTS[index]

  return AHSpawn.new(
    AHColorTo.new(self.trees, duration, nil, treeColor),
    AHColorTo.new(self.clouds, duration, nil, cloudColor),
    AHColorTo.new(self.houses, duration, nil, houseColor),
    AHColorTo.new(self.stars, duration, nil, starsColor),
    self.skyGradient:newGradientToAction(duration, nil, skyStartColor, skyEndColor)
  )
end

local function changeScene(self, sceneIndex)
  self.sceneChangeCoroutine = runInCoroutine(function()
    local oldWindows = self.windows

    insertWindows(self, sceneIndex)
    self.windows:setOpacity(0)
    waitFor(0.01) -- Workaround for attribute lag glitch

    local action = newTintAction(self, sceneIndex, SCENE_TRANSITION_DURATION)
    wait(action:start())
    self.windows:setOpacity(1)
    oldWindows:removeSelf()

    if (sceneIndex == 3) then
      self.starGenerator:start()
    end
  end)

  return self.sceneChangeCoroutine
end

local function startNextScene(self)
  self.sceneIndex = self.sceneIndex + 1

  if (self.sceneIndex > SCENE_COUNT) then
    self.sceneIndex = 1
  end

  changeScene(self, self.sceneIndex)
end

local function makeBird(birdData)
  if birdData then
    local birdClass = BIRD_CLASSES[birdData.bird_type]
    local bird = birdClass.new(birdData.facing == FACING_RIGHT)
    bird.birdData = birdData
    bird:setScl(birdData.scale)

    return bird
  end
  return nil
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WordBirdsBackground:showOutro()
  return runInCoroutine(function()
    local lastAction = nil
    for _, bird in ipairs(self.birds) do
      lastAction = bird:flyOut(bird.birdData.fly_out_duration, bird.birdData.fly_out_to_x, bird.birdData.fly_out_to_y)
      waitFor(0.1)
    end
    wait(lastAction)
  end)
end

function WordBirdsBackground:addBird()
  return runInCoroutine(function()
    if #self.availableBirdData > 0 then
      local birdData = table.remove(self.availableBirdData,1)
      local bird = makeBird(birdData)
      table.insert(self.birds, bird)
      bird:setLoc(bird.birdData.fly_in_from_x, bird.birdData.fly_in_from_y)
      self:insert(bird, Z_BIRDS[bird.birdData.bird_type])
      runInCoroutine(function()
        wait(bird:flyIn(bird.birdData.fly_in_duration, bird.birdData.fly_in_to_x, bird.birdData.fly_in_to_y))
        bird:beginIdling()
      end)
    else
      -- todo, is this safe enough?
      print("No more birds")
    end
  end)
end

function WordBirdsBackground:addBirds(count)
  for i=1, count do
    self:addBird()
  end
end

local function removeOneIfAvailableAndAddToList(destinationList, sourceList)
  local item = table.remove(sourceList, 1)
  if (item) then
    table.insert(destinationList, item)
  end
end

function WordBirdsBackground:resetBirds()
  self.birds = {}
  self.availableBirdData = {}

  -- generate availableBirdData in a balanced way, but ranomized within each roof
  -- for as even as possible a distribution of birds
  local processedBirdDescriptors =
  {
    processBirdData(WordBirdsBirdData.BIRD_DATA_ROOF_1),
    processBirdData(WordBirdsBirdData.BIRD_DATA_ROOF_2),
    processBirdData(WordBirdsBirdData.BIRD_DATA_ROOF_3)
  }

  for _, birdDescriptorList in ipairs(processedBirdDescriptors) do
    TableUtils.shuffleArray(birdDescriptorList)
  end

  repeat
    for _, birdDescriptorList in ipairs(processedBirdDescriptors) do
      removeOneIfAvailableAndAddToList(self.availableBirdData, birdDescriptorList)
    end
  until (TableUtils.areEmpty(unpack(processedBirdDescriptors)))
end

function WordBirdsBackground:start(sceneDuration)
  self.backgroundCoroutine = runInCoroutine(function()
    waitFor(sceneDuration)
    startNextScene(self)
    waitFor(sceneDuration)
    startNextScene(self)
  end)

  return self.backgroundCoroutine
end

function WordBirdsBackground:stop()
  self.starGenerator:stop()
end

function WordBirdsBackground:getBottomMargin()
  return self.bottomMargin
end

function WordBirdsBackground:pauseTransition()
  if self.backgroundCoroutine and self.backgroundCoroutine:isBusy() then
    self.backgroundCoroutine:pause(true)
  end

  if self.sceneChangeCoroutine and self.sceneChangeCoroutine:isBusy() then
    self.sceneChangeCoroutine:pause(true)
  end
end

function WordBirdsBackground:resumeTransition()
  if self.sceneChangeCoroutine and self.sceneChangeCoroutine:isBusy() then
    self.sceneChangeCoroutine:pause(false)
  end

  if self.backgroundCoroutine and self.backgroundCoroutine:isBusy() then
    self.backgroundCoroutine:pause(false)
  end
end

-- ---------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WordBirdsBackground.new()
  local self = Group.new(Screen.fullWidth, Screen.fullHeight)
  Mixin.mix(self, WordBirdsBackground)

  local sceneIndex = 1
  insertSky(self)
  insertTrees(self)
  insertHouses(self)
  insertWindows(self, sceneIndex)
  insertClouds(self)
  self.clouds:setScrollSpeed(SCROLL_SPEED)
  newTintAction(self, sceneIndex, 0):start()

  self.starGenerator = WordBirdsFallingStarGenerator.new()
  self:insertAtCenter(self.starGenerator, Z_STARS)

  self.sceneIndex = sceneIndex

  self.birds = {}
  self:resetBirds()

  self.sceneChangeCoroutine = nil
  self.backgroundCoroutine = nil

  return self
end
