--- WordBirdsButton
-- @author Created by karl on October 13, 2014.

class("WordBirdsButton")
require("WordBirdsCommon")
require("WordBirdsConfiguration")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local COLORS_LETTERS = {
  colorBgNormal = {0.0, 0.0, 0.1, 0.5},
  colorTextNormal = {1.0, 1.0, 1.0, 1.0},
  colorBgTouched = {0.0, 0.0, 0.1, 0.5},
  colorTextTouched = {1.0, 1.0, 1.0, 1.0},
  colorBgDisabled = {0.0, 0.0, 0.1, 0.05},
  colorTextDisabled = {1.0, 1.0, 1.0, 0.05}
}

local COLORS_CONTROLS = {
  colorBgNormal = ColorUtils.colorHex32ToStd(0xffffffa0),
  colorTextNormal = ColorUtils.colorHex32ToStd(0x335e74ff),
  colorBgTouched = ColorUtils.colorHex32ToStd(0x325e7482),
  colorTextTouched = ColorUtils.colorHex32ToStd(0xffffffff),
  colorBgDisabled = ColorUtils.colorHex32ToStd(0xffffff33),
  colorTextDisabled = ColorUtils.colorHex32ToStd(0x325e7433)
}

local FONT = WordBirdsConfiguration.NORMAL_FONT

local LETTER_FONT_SIZE = WordBirdsConfiguration.SQUARE_BUTTON_SIZE * 0.6
local WORD_FONT_SIZE = WordBirdsConfiguration.SQUARE_BUTTON_SIZE * 0.4

local LETTER_ADJUST_Y = WordBirdsConfiguration.SQUARE_BUTTON_SIZE * 0.04
local WORD_ADJUST_Y = WordBirdsConfiguration.SQUARE_BUTTON_SIZE * 0.01

local LABEL_CACHE = {}

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function addIcon(text, iconName)
  local icon = Sprite.newWithName(iconName)
  icon:setEffectiveDims(WordBirdsConfiguration.SQUARE_BUTTON_SIZE * 0.55)
  icon.adjustY = 0
  LABEL_CACHE[text] = icon
end

local function addToCache(words, fontSize, adjustY)
  for _, word in ipairs(words) do
    local label = Label.new(word, Font.newWithActualSize(FONT, fontSize))
    label.adjustY = adjustY
    LABEL_CACHE[word] = label
  end
end

local function getCachedLabel(text)
  local label = LABEL_CACHE[text]
  if label then
    if label.clone then
      local clone = label:clone()
      clone.adjustY = label.adjustY
      return clone
    end
    return label
  end
end

local function newRoundedRectangle(self)
  local rect = FilledRoundedRectangle.new(self:getWidth(), self:getHeight(), self.cornerRadius, self.colors.colorBgNormal)
  rect:setLoc(self:getCenter())
  return rect
end

local function newLabel(self, text)
  local label = getCachedLabel(text)
  label:setColor(unpack(self.colors.colorTextNormal))
  label:setLoc(self:getWidthPart(0.5), self:getHeightPart(0.5) + label.adjustY)
  return label
end

local function newTextRectangle(self, text)
  local rect = newRoundedRectangle(self)
  local label = newLabel(self, text)
  local group = Group.new(rect:getDims())
  group:insertAtCenter(rect)
  group:insert(label)
  group.rect = rect
  group.label = label
  group:linkColorAttrToChildren()
  return group
end

local function getColorsForTouchEnabled(self, isEnabled)
  local bgColor = self.colors.colorBgNormal
  local textColor = self.colors.colorTextNormal
  if not isEnabled then
    bgColor = self.colors.colorBgDisabled
    textColor = self.colors.colorTextDisabled
  end

  return bgColor, textColor
end

local function animateTouch(self)
  return runInCoroutine(function()
    local textRect = self.textRectangle
    local rect = textRect.rect
    local label = textRect.label
    local scaleDownDuration = 0.04
    local scaleUpDuration = 0.05

    self:setTouchEnabled(false)

    local silhouetteEase = AHEaseSineOut.new()
    wait(AHSpawn.new(
       AHScaleTo.new(textRect, scaleDownDuration, silhouetteEase, 0.9),
       AHColorTo.new(rect, scaleDownDuration, silhouetteEase, self.colors.colorBgTouched),
       AHColorTo.new(label, scaleDownDuration, silhouetteEase, self.colors.colorTextTouched)
    ):start())

    local bgColor, textColor = getColorsForTouchEnabled(self, self.shouldRearm)

    wait(AHSpawn.new(
      AHScaleTo.new(textRect, scaleUpDuration, nil, 1.0),
       AHColorTo.new(rect, scaleUpDuration, silhouetteEase, bgColor),
       AHColorTo.new(label, scaleUpDuration, silhouetteEase, textColor)
    ):start())
    self:setTouchEnabled(self.shouldRearm)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WordBirdsButton:setTouchEnabled(isEnabled)
  self.virtualButton:setTouchEnabled(isEnabled)
end

function WordBirdsButton:setTouchEnabledVisible(isEnabled)
  self:setTouchEnabled(isEnabled)
  local bgColor, textColor = getColorsForTouchEnabled(self, isEnabled)
  self.textRectangle.rect:setColor(unpack(bgColor))
  self.textRectangle.label:setColor(unpack(textColor))
end

function WordBirdsButton:isTouchEnabled()
  return self.virtualButton:isTouchEnabled()
end

function WordBirdsButton:animateTouchEnabled(isEnabled)
  return runInCoroutine(function()
    if not isEnabled then self:setTouchEnabled(isEnabled) end
    local bgColor, textColor = getColorsForTouchEnabled(self, isEnabled)

    local textRect = self.textRectangle
    local rect = textRect.rect
    local label = textRect.label
    local duration = 0.1

    wait(AHSpawn.new(
       AHColorTo.new(rect, duration, silhouetteEase, bgColor),
       AHColorTo.new(label, duration, silhouetteEase, textColor)
    ):start())
    if isEnabled then self:setTouchEnabled(isEnabled) end
  end)
end

function WordBirdsButton.initLabelCache()
  addToCache(TableUtils.createArrayFromString(Localizer.get("ALL_CHARACTERS")), LETTER_FONT_SIZE, LETTER_ADJUST_Y)
  addToCache({Localizer.get("SUBMIT")}, WORD_FONT_SIZE, WORD_ADJUST_Y)
  addToCache({Localizer.get("SKIP")}, WORD_FONT_SIZE * 0.8, WORD_ADJUST_Y)
  addIcon("<<", "back-normal.png")
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WordBirdsButton.new(widthInSquares, text, shouldRearm, onTouched)
  local selfW = WordBirdsConfiguration.SQUARE_BUTTON_SIZE * widthInSquares + WordBirdsConfiguration.BUTTON_SPACING * (widthInSquares-1)
  local selfH = WordBirdsConfiguration.SQUARE_BUTTON_SIZE
  local self = Group.new(selfW, selfH)
  Mixin.mix(self, WordBirdsButton)

  if #text == 1 then
    self.colors = COLORS_LETTERS
  else
    self.colors = COLORS_CONTROLS
  end
  self.text = text
  self.cornerRadius = selfH / 10

  self.shouldRearm = shouldRearm
  self.textRectangle = newTextRectangle(self, text)
  self:insertAtCenter(self.textRectangle)

  self.touchHandler = function()
    self:setTouchEnabled(false)
    animateTouch(self)
    onTouched(self)
  end
  self.virtualButton = Button.addVirtualButton(self, self:getWidthPart(1.25), self:getHeightPart(1.25), self.touchHandler)
  return self
end
