--- WordBirdsButtonSet
-- @author Created by karl on October 23, 2014.

class("WordBirdsButtonSet")
require("WordBirdsButton")
require("data.generators.ValueGenerators")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local ALL_LETTERS = TableUtils.createSetFromString(Localizer.get("ALL_CHARACTERS"))
local ANIMATE_IN_DELAY = 0.02
local ANIMATE_OUT_DELAY = 0.02

local MAX_HEIGHT_RATIO = 0.2

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function newAnimateButtonInAction(button)
  return AHSequence.new(
    AHSpawn.new(
      AHOpacityTo.new(button, 0.30, AHEaseSineInOut.new(), 1),
      AHScaleTo.new(button, 0.30, AHEaseSineInOut.new(), 1, 1.1)
    ),
    AHScaleTo.new(button, 0.15, AHEaseSineInOut.new(), 1, 1)
  )
end

local function newAnimateButtonOutAction(button)
  local duration = 0.19
  return AHSpawn.new(
    AHOpacityTo.new(button, duration, AHEaseSineIn.new(), 0),
    AHScaleTo.new(button, duration, AHEaseSineIn.new(), 0, 0)
  )
end

local function handleLetterButtonPress(self, letter, button)
  table.insert(self.buttonPressOrder, button)
  self.onLetterPressed(letter)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WordBirdsButtonSet:setEnabled(isEnabled)
  for _, button in ipairs(self.buttons) do
    button:setTouchEnabled(isEnabled)
  end
end

function WordBirdsButtonSet:animateIn()
  return runInCoroutine(function()
    local action = AHStaggeredSpawn.new(ANIMATE_IN_DELAY)
    for _, button in ipairs(self.buttons) do
      action:addAction(newAnimateButtonInAction(button))
    end
    self.inOutActions:insert(action)
    self.inOutActions:wait()
  end)
end

function WordBirdsButtonSet:animateOut()
  return runInCoroutine(function()
    local action = AHStaggeredSpawn.new(ANIMATE_OUT_DELAY)
    for _, button in ipairs(self.buttons) do
      action:addAction(newAnimateButtonOutAction(button))
    end
    self.inOutActions:insert(action)
    self.inOutActions:wait()
  end)
end

function WordBirdsButtonSet:markIncorrect()
  for _, button in ipairs(self.buttons) do
    button:animateTouchEnabled(false)
  end
end

function WordBirdsButtonSet:reactivateLastLetterButton()
  local button = table.remove(self.buttonPressOrder)
  if button then
    button:animateTouchEnabled(true)
  end
end

function WordBirdsButtonSet:setBackspaceEnabled(isEnabled)
  self.backspaceButton:setTouchEnabledVisible(isEnabled)
end

function WordBirdsButtonSet:setBackspaceRearm(shouldRearm)
  self.backspaceButton.shouldRearm = shouldRearm
end

function WordBirdsButtonSet:findLetterButton(letter)
  return TableUtils.find(self.buttons, function(button)
    return button.text == letter and not TableUtils.contains(self.buttonPressOrder, button)
  end)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function newCommandRow(self, onSkip, onSubmit, onBackspace)
  local skipButton = WordBirdsButton.new(1, Localizer.get("SKIP"), false, onSkip)
  local submitButton = WordBirdsButton.new(3, Localizer.get("SUBMIT"), false, onSubmit)
  local backspaceButton = WordBirdsButton.new(1, "<<", true, onBackspace)

  local commandRow = Row.new(WordBirdsConfiguration.BUTTON_SPACING)
  commandRow:insert(skipButton)
  commandRow:insert(submitButton)
  commandRow:insert(backspaceButton)
  commandRow:alignChildren()

  for _, child in ipairs(commandRow.children) do
    table.insert(self.buttons, child)
  end

  return commandRow, backspaceButton, submitButton, skipButton
end

local function newLetterGenerator(unrevealedWordString, targetLetterCount, disallowedLetters)
  local letters = TableUtils.createArrayFromString(unrevealedWordString)
  local distractorLetters = TableUtils.symmetricDifferenceSet(ALL_LETTERS, TableUtils.createSetFromString(disallowedLetters))
  local generator = RandomlySelectedValueGenerator.new(distractorLetters, true)
  local extraLetterCount = targetLetterCount - #letters
  TableUtils.concatenateArrays(letters, generator:nextAsArray(extraLetterCount))
  return RandomlySelectedValueGenerator.new(letters)
end

-- Constructor ------------------------

function WordBirdsButtonSet.new(unrevealedWordString, disallowedLetters, onLetter, onBackspace, onSubmit, onSkip)
  local self = Column.new(WordBirdsConfiguration.BUTTON_SPACING)
  Mixin.mix(self, WordBirdsButtonSet)

  local rowCount = 2
  local buttonsPerRow = WordBirdsConfiguration.MAX_SQUARE_BUTTONS_PER_ROW
  local letterCount = buttonsPerRow * rowCount
  assert(#unrevealedWordString <= letterCount, "[" .. unrevealedWordString .. "] is too long. Max unrevealed word length is " .. tostring(letterCount))
  local generator = newLetterGenerator(unrevealedWordString, letterCount, disallowedLetters)

  local widthInSquares = 1
  local shouldRearm = false

  self.buttons = {}
  local letterRows = {}
  for i = 1, rowCount do
    local row = Row.new(WordBirdsConfiguration.BUTTON_SPACING)
    table.insert(letterRows, row)
    self:insert(row)
    for j = 1, buttonsPerRow do
      local letter = generator:next()
      local button
      button = WordBirdsButton.new(widthInSquares, letter, shouldRearm, function() handleLetterButtonPress(self, letter, button) end)
      row:insert(button)
      table.insert(self.buttons, button)
    end
    row:alignChildren()
  end

  self:alignChildren()

  if self:getHeight() / Screen.safeHeight < MAX_HEIGHT_RATIO then
    self:insertSpacer(0)
  end

  local commandRow, backspaceButton, submitButton, skipButton = newCommandRow(self, onSkip, onSubmit, onBackspace)
  self:insert(commandRow)
  self:alignChildren()

  self.backspaceButton = backspaceButton
  self.submitButton = submitButton
  self.skipButton = skipButton

  for _, button in ipairs(self.buttons) do
    button:setScl(1, 0)
    button:setOpacity(0)
  end

  self.onLetterPressed = onLetter
  self.buttonPressOrder = {}

  self.inOutActions = ActionInterrupt.new()

  return self
end
