--- WordBirdsConceptGenerator.lua
-- @author created by Jay Austria on April 26, 2016

class("WordBirdsConceptGenerator")
require("WordBirdsConfiguration")
require("data.generators.ValueGenerators")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INITIAL_POOL_SIZE = 6
local DIFFICULT_POOL_SIZE = 25
local MAXIMUM_CONCEPTS_PER_PACK = 100

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getConceptsByDifficulty(allConcepts)
  local conceptsByDifficulty = {}

  for _, contentDifficulty in ipairs(WordBirdsConfiguration.DIFFICULTY_BUCKETS) do
    conceptsByDifficulty[contentDifficulty] = {}
  end

  for _, concept in ipairs(allConcepts) do
    local contentDifficulty = concept.contentDifficulty
    table.insert(conceptsByDifficulty[contentDifficulty], concept)
  end

  return conceptsByDifficulty
end

local function getContentDifficultyCounts(poolSize, gameDifficulty)
  local keyFrames = WordBirdsConfiguration.getDifficultyKeyFrames()
  local contentDifficultyChooser = KeyFrameInterpolator.new(keyFrames, gameDifficulty)
  local difficulties = TableUtils.shallowCopy(contentDifficultyChooser:getValues())
  local weights = contentDifficultyChooser:getWeights()

  local difficultyCounts = {}
  local currentPoolSize = 0
  for i, difficulty in ipairs(difficulties) do
    local weight = weights[i]
    local count = MathUtils.round(weight * poolSize)
    difficultyCounts[difficulty] = count
    currentPoolSize = currentPoolSize + count
  end

  local indexToAdjust = 1
  table.sort(difficulties)
  while (currentPoolSize > poolSize) do
    local difficultyToAdjust = difficulties[indexToAdjust]
    difficultyCounts[difficultyToAdjust] = math.max(0, difficultyCounts[difficultyToAdjust] - 1)
    indexToAdjust = indexToAdjust + 1
    currentPoolSize = currentPoolSize - 1
  end
  while (currentPoolSize < poolSize) do
    local difficultyToAdjust = difficulties[indexToAdjust]
    difficultyCounts[difficultyToAdjust] = difficultyCounts[difficultyToAdjust] + 1
    indexToAdjust = indexToAdjust + 1
    currentPoolSize = currentPoolSize + 1
  end

  local logString = "Selecting content difficulty counts for " .. tostring(gameDifficulty)
  for difficulty, count in pairs(difficultyCounts) do
    logString = logString .. "\n\t" .. tostring(difficulty) .. ":\t\t" .. tostring(count)
  end
  Log.debug(logString)

  return difficultyCounts
end

local function selectConcepts(conceptsByDifficulty, poolSize, gameDifficulty)
  local selectedConcepts = {}
  local difficultyCounts = getContentDifficultyCounts(poolSize, gameDifficulty)

  for contentDifficulty, concepts in pairs(conceptsByDifficulty) do
    table.sort(concepts, function(conceptA, conceptB)
      return conceptA.identifier < conceptB.identifier
    end)

    local scaleHashFactor = 500
    local startIndex = math.floor(gameDifficulty * scaleHashFactor) % #concepts
    local count = difficultyCounts[contentDifficulty]

    -- Note: If there are not enough concepts for this difficulty, the following will just select
    -- some repeated concepts. However, the Content Tests check that all packs have the required
    -- number of concepts per difficulty.
    for i = 0, count - 1 do
      local index = (startIndex + i - 1) % #concepts + 1
      table.insert(selectedConcepts, concepts[index])
    end
  end

  TableUtils.shuffleArray(selectedConcepts)

  return selectedConcepts
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WordBirdsConceptGenerator.new()
  local conceptGenerator = GameSession.conceptGenerator
  conceptGenerator:setAllowFewerResults(true)

  local allConcepts = conceptGenerator:nextAsArray(MAXIMUM_CONCEPTS_PER_PACK)
  local conceptsByDifficulty = getConceptsByDifficulty(allConcepts)

  local initialPoolDifficulty = GameSession.getDifficulty()
  local difficultPoolDifficulty = math.min(1.0, initialPoolDifficulty + 0.3)

  local initialConceptPool = selectConcepts(conceptsByDifficulty, INITIAL_POOL_SIZE, initialPoolDifficulty)
  local difficultConceptPool = selectConcepts(conceptsByDifficulty, DIFFICULT_POOL_SIZE, difficultPoolDifficulty)

  local combinedConceptPool = TableUtils.createConcatenatedArray(initialConceptPool, difficultConceptPool)
  local uniquesConceptPool = TableUtils.getUniquesArray(combinedConceptPool)

  -- todo - remove...but its useful for testing
  -- for _, concept in ipairs(uniquesConceptPool) do
  --   print(concept.identifier, concept.contentDifficulty)
  -- end

  local self = SequentialValueGenerator.new(uniquesConceptPool)
  Mixin.mix(self, WordBirdsConceptGenerator)

  return self
end
