--- WordBirdsGame.lua
-- @author created by Jay Austria on April 26, 2016

class("WordBirdsGame")
require("bonuses.ResponseSpeedTracker")
require("WordBirdsBackground")
require("WordBirdsInGameUi")
require("WordBirdsCommon")
require("WordBirdsConfiguration")
require("StandardUIBuilder")
require("WordBirdsConceptGenerator")
require("WordBirdsBirdCounter")
require("WordBirdsAutocorrect")
require("WordBirdsErrorInterstitial")
require("widgets.TapToContinue")
require("tutorials.TutorialHighlighter")
require("data.generators.LearningFocusedRoundGenerator")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local Z_INDEX = {
  BACKGROUND = 0,
  UI = 10,
  INTERSTITIAL = 20,
  TUTORIAL = 30
}

local LOST_SECONDS_ON_SKIP = 1.5

local OVERLAY_COLOR = {0.039, 0.055, 0.082, 0.7}
local MASK_COLOR = {1.0, 1.0, 1.0, 1.0}

local TUTORIAL_FONT_SIZE = 20
local TUTORIAL_FONT_SIZE_TABLET = 15
local TUTORIAL_LABEL_PADDING_Y = 85
local TUTORIAL_LABEL_MINIMUM_Y_PROPORTION = 0.5

local function getDifficultyUpdateHandler()
  return function(ramp)
    if (ramp.correctAnswerStreak >= 2) then
      ramp:addDifficulty(1)
      ramp:resetCorrectAnswerStreak()
    elseif (ramp.incorrectAnswerStreak >= 2) then
      ramp:addDifficulty(-1)
      ramp:resetIncorrectAnswerStreak()
    end
  end
end

local function configureGameDifficulty(self)
  local passDuration = WordBirdsConfiguration.GAME_DURATION_IN_SECONDS

  local rank2Score = WordBirdsConfiguration.getRankTwoScore()
  local rank3Score = WordBirdsConfiguration.getRankThreeScore()

  GameSession.setScoreThresholds(rank2Score, rank3Score)

  self.questionCount = WordBirdsConfiguration.CORRECT_NEEDED_COUNT
  self.timerDuration = passDuration

  self.difficultyRamp = DifficultyRamp.new(WordBirdsConfiguration.INITIAL_DIFFICULTY, WordBirdsCommon.MAX_DIFFICULTY, getDifficultyUpdateHandler())

  GameSession.bonusReporter:initializeBonusCount(BonusTypes.TYPES.SPEED_BONUS)
end


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function getClosestWordInSet(word, set)
  local closestWord = nil
  local closestDistance = WordBirdsCommon.MAX_WORD_MATCH_DISTANCE + 1
  for setWord in pairs(set) do
    local distance = WordBirdsAutocorrect.getDistance(word, setWord)
    if (distance <= WordBirdsCommon.MAX_WORD_MATCH_DISTANCE and closestDistance > distance) then
      closestWord = setWord
      closestDistance = distance
    end
  end

  return closestWord
end

local function isResponseSimilarToAlternateForm(self, answer)
  local closeWord = getClosestWordInSet(answer, self.concept.alternateFormsSet)
  return closeWord ~= nil
end

local function getPrefillCharacterCount(self)
  local difficulty = self.difficultyRamp:getDifficulty()
  return WordBirdsCommon.MAX_CHAR_COUNT - difficulty
end

local function isStartOfWordCorrect(self, startOfWord)
  for word in pairs(self.concept.answerSet) do
    if StringUtils.startsWith(word, startOfWord) then
      return true
    end
  end
  return false
end

local function markSpeedyAnswerTimeIfValid(self, existingWord, newLetter)
  if ((existingWord ~= nil) and (newLetter ~= nil)) then
    if (self.firstCorrectLetterTime == nil) then
      if isStartOfWordCorrect(self, existingWord .. newLetter) then
        self.firstCorrectLetterTime = self.responseSpeedTracker:getCurrentTime()
      end
    end
  end
end

local function setMaskPos(self, mask, element)
  local x, y = self:transformLoc(element)
  x = x + element:getWidthPart(0.5)
  y = y + element:getHeightPart(0.5)
  mask:setLoc(x, y)
end

local function showHelpAndWaitForTap(self)
  self.timer:pause()
  self.background:pauseTransition()

  local answerArray = TableUtils.getKeys(self.concept.answerSet)
  local answer = answerArray[1]
  local currentAnswer = self.inGameUi.editor:getWord()
  local charIndex = #currentAnswer + 1
  local letterButton = self.inGameUi.buttonSet:findLetterButton(answer:sub(charIndex, charIndex))
  local tutorialText = Localizer.get("TUTORIAL_SPELL")
  local tutorialLabelWidth = Screen.getSafeWidthPart(0.6)
  local tutorialFontSize = TUTORIAL_FONT_SIZE
  local isUnsubmittedFullAnswer = (answer == currentAnswer)

  if isUnsubmittedFullAnswer then
    letterButton = self.inGameUi.buttonSet.submitButton
    tutorialText = Localizer.get("TUTORIAL_SUBMIT")
    tutorialLabelWidth = Screen.getSafeWidthPart(0.8)
  end

  if Screen.isTabletAspectRatio() then
    tutorialFontSize = TUTORIAL_FONT_SIZE_TABLET

    if isUnsubmittedFullAnswer then
      tutorialLabelWidth = Screen.getSafeWidthPart(0.6)
    else
      tutorialLabelWidth = Screen.getSafeWidthPart(0.5)
    end
  end

  local getLetterHole = function()
    local letterRect = FilledRoundedRectangle.new(letterButton:getWidth(), letterButton:getHeight(), letterButton.cornerRadius, MASK_COLOR)
    setMaskPos(self, letterRect, letterButton)
    return letterRect
  end

  local answerLabel = self.inGameUi.answerLabel
  local answerRect = FilledRoundedRectangle.new(answerLabel:getWidth(), answerLabel:getHeight(), answerLabel.cornerRadius, MASK_COLOR)
  setMaskPos(self, answerRect, answerLabel)

  local questionLabel = self.inGameUi.questionLabel
  local questionRect = FilledRectangle.new(questionLabel:getWidth(), questionLabel:getHeight(), MASK_COLOR)
  setMaskPos(self, questionRect, questionLabel)

  local secondaryMasks = {answerRect, questionRect}
  local helpLayer = TutorialHighlighter.newWithMasks({getLetterHole},
                                                     secondaryMasks)

  local _, submitPositionYBottom = self:transformLoc(self.inGameUi.buttonSet.submitButton)
  local positionYUnderSubmit = submitPositionYBottom - TUTORIAL_LABEL_PADDING_Y
  local positionYMinimum = submitPositionYBottom * TUTORIAL_LABEL_MINIMUM_Y_PROPORTION
  local instructionLocationY = math.max(positionYUnderSubmit, positionYMinimum)

  local shouldShowArrows = false
  helpLayer:addInstructionLabel(tutorialText, tutorialFontSize, instructionLocationY, shouldShowArrows, tutorialLabelWidth)
  self:insertAtCenter(helpLayer, Z_INDEX.TUTORIAL)

  helpLayer:createFadeInAnimation():start()

  helpLayer:waitForTargetTap()

  wait(helpLayer:createFadeOutAnimation():start())
  helpLayer:removeSelf()

  self.background:resumeTransition()
  self.timer:resume()
  return letterButton
end

local function initializeConcept(self, concept, prefillCharacterCount)
  self.concept = concept
  if (prefillCharacterCount == nil) then
    prefillCharacterCount = getPrefillCharacterCount(self)
  end
  self.concept.prefillCount = prefillCharacterCount

  self.concept.alternateFormsSet = {}
  if ((self.concept.alternateForms ~= nil) and (self.concept.alternateForms ~= "")) then
    self.concept.alternateFormsSet = TableUtils.createSetFromArray(TableUtils.csvToArray(self.concept.alternateForms))
  end

  self.concept.acceptedMisspellingsSet = {}
  if ((self.concept.acceptedMisspellings ~= nil) and (self.concept.acceptedMisspellings ~= "")) then
    self.concept.acceptedMisspellingsSet = TableUtils.createSetFromArray(TableUtils.csvToArray(self.concept.acceptedMisspellings))
  end

  self.concept.answerSet = {}
  if ((self.concept.answer ~= nil) and (self.concept.answer ~= "")) then
    self.concept.answerSet = TableUtils.createSetFromArray(TableUtils.csvToArray(self.concept.answer))
  end
end

local function startNextQuestion(self)
  if self.questionCoroutine then
    self.questionCoroutine:stop()
  end

  self.questionCoroutine = runInCoroutine(function()
    if (self.isGameRunning) then
      if (self.timer:getTimeRemaining() < WordBirdsConfiguration.TIME_REMAINING_TO_START_RESURFACING_MISTAKES) then
        self.conceptGenerator:allowMistakesToResurface()
      end
      initializeConcept(self, self.conceptGenerator:next())
      GameSession.contentTracker:notifyPresentedConcept(self.concept.identifier)

      wait(self.inGameUi:showQuestion(self.concept))
      self.responseSpeedTracker:notifyStart()
      self.firstCorrectLetterTime = nil

      while self.questionsCompleted == 0 and self.tutorialEnabled do
        waitUntil(function() return self.timer:getTime() - self.lastActionTime > WordBirdsConfiguration.TUTORIAL_DELAY end)
        if self.questionsCompleted == 0 and self.tutorialEnabled then
          self.inGameUi:setEnabled(false)
          self.showingTutorial = true

          local letterButton = showHelpAndWaitForTap(self)
          letterButton.touchHandler()

          self.showingTutorial = false
          self.inGameUi:setEnabled(true)
        end
      end
    end
  end)

  return self.questionCoroutine
end

local function stopGame(self)
  self.isGameRunning = false
  self.inGameUi:setEnabled(false)
end

-- High level event handlers ----------

local function onGameWin(self)
  stopGame(self)
  GameSession.notifyGameTimeEnd()
  self.scoreLabel:setScore(GameSession.getScore())
  self.inGameUi:removeAll()
  Sound.playSoundEffect("game_win")
  self:showDefaultGameWinOutro()
  wait(self.background:showOutro())
  self.background:stop()

  GameSession.notifyGameWin()
end

local function onGameLoss(self)
  stopGame(self)
  self:showDefaultGameLossOutro()
  wait(self.inGameUi:removeAll())
  self.background:stop()

  GameSession.notifyGameLoss()
end

local function onCorrectAnswer(self, answer, matchingAnswer, isExactMatch)
  if (self.firstCorrectLetterTime == nil) then
    self.firstCorrectLetterTime = self.responseSpeedTracker:getCurrentTime()
  end
  self.responseSpeedTracker:notifyCorrectResponseWithValue(self.firstCorrectLetterTime)

  self.birdCounter:incrementProgress()

  GameSession.contentTracker:notifyCorrectAnswer(self.concept.identifier, nil, answer)
  GameSession.notifyCorrectAnswer(self.concept.identifier)
  GameSession.incrementCorrectAnswerCount()
  GameSession.increaseScore(WordBirdsConfiguration.POINTS_PER_CORRECT_RESPONSE)

  self.difficultyRamp:notifyCorrectAnswer()
  self.scoreLabel:setScore(GameSession.getScore())
  Sound.playSoundEffect("wordbirds_correct")
  self.background:addBird()
  Sound.playSoundEffect("wordbirds_flap", 0.2)
  self.questionsCompleted = self.questionsCompleted + 1

  wait(self.inGameUi:runCorrectSequence(matchingAnswer, isExactMatch))
end

local function onIncorrectAnswer(self, answer, willShowInterstitial)
  GameSession.contentTracker:notifyIncorrectAnswer(self.concept.identifier, nil, answer)
  self.responseSpeedTracker:notifyIncorrectResponse()
  GameSession.scoredIncorrectAnswer(self.concept.identifier)
  GameSession.incrementIncorrectAnswerCount()
  self.difficultyRamp:notifyIncorrectAnswer()
  self.conceptGenerator:notifyMistake(self.concept.identifier)
  Sound.playSoundEffect("wordbirds_incorrect")

  if (willShowInterstitial) then
    self.timer:pauseAndShowIcon()

    local interstitial = WordBirdsErrorInterstitial.newHidden(self.concept)
    self:insertAtCenter(interstitial, Z_INDEX.INTERSTITIAL)

    wait(interstitial:playShowAction())

    local tapToContinue = TapToContinue.new()
    tapToContinue:waitForTrigger()
    tapToContinue:removeSelf()

    wait(interstitial:playHideAction())

    interstitial:removeSelf()

    self.timer:resumeAndHideIcon()
  end

  wait(self.inGameUi:runIncorrectSequence())
end

local function onConceptSkipped(self)
  GameSession.contentTracker:notifySkippedAnswer(self.concept.identifier)
  self.responseSpeedTracker:notifyIncorrectResponse()
  GameSession.scoredIncorrectAnswer(self.concept.identifier)
  self.conceptGenerator:notifyMistake(self.concept.identifier)
  self.timer:addTime(LOST_SECONDS_ON_SKIP)
  self.tutorialEnabled = false
  Sound.playSoundEffect("wordbirds_skip")
  wait(self.inGameUi:runSkipSequence())
end

local function tryToTimeout(self)
  if (self.isGameRunning) then
    local gameHasTimedOutNaturally = (self.isTimeout and not self.isSubmittingAnswer)
    local timeRemainingIsLessThanMinimum = (self.timer:getTimeRemaining() < WordBirdsConfiguration.MINIMUM_TIME_TO_START_NEXT_ROUND)
    if (gameHasTimedOutNaturally or timeRemainingIsLessThanMinimum) then
      stopGame(self)
      if gameHasTimedOutNaturally then
        wait(self.inGameUi:runTimeoutSequence())
      end

      local didWin = (self.questionsCompleted > 0)
      if (didWin) then
        onGameWin(self)
      else
        onGameLoss(self)
      end
    end
  end
end

-- Low level event handlers -----------

local function onSubmitAnswer(self, answer)
  self.inGameUi:setEnabled(false)
  return runInCoroutine(function()
    if (not self.isGameRunning) then return end
    self.isSubmittingAnswer = true

    local matchingAnswer = getClosestWordInSet(answer, self.concept.answerSet)
    if (not matchingAnswer) then
      local misspelling = getClosestWordInSet(answer, self.concept.acceptedMisspellingsSet)
      if (misspelling) then
        local answerArray = TableUtils.getKeys(self.concept.answerSet)
        matchingAnswer = answerArray[1]
      end
    end

    local matchingAlternateForm = getClosestWordInSet(answer, self.concept.alternateFormsSet)
    local isExactMatchToAlternateForm = (answer == matchingAlternateForm)
    if (matchingAnswer and (not isExactMatchToAlternateForm)) then
      local isExactMatch = (answer == matchingAnswer)
      onCorrectAnswer(self, answer, matchingAnswer, isExactMatch)
    else
      local willShowInterstitial = isResponseSimilarToAlternateForm(self, answer)
      onIncorrectAnswer(self, answer, willShowInterstitial)
    end
    self.isSubmittingAnswer = false
    tryToTimeout(self)
    if (self.isGameRunning) then
      startNextQuestion(self)
    end
  end)
end

local function onSkipButtonPressed(self)
  self.inGameUi:setEnabled(false)
  return runInCoroutine(function()
    if (not self.isGameRunning) then return end
    onConceptSkipped(self)
    startNextQuestion(self)
  end)
end

local function onTimeout(self)
  self.inGameUi:setEnabled(false)
  return runInCoroutine(function()
    if (not self.isGameRunning) then return end
    self.isTimeout = true
    tryToTimeout(self)
  end)
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

-- gameFlow ---------------------------

function WordBirdsGame:gameFlow()
  self.questionsCompleted = 0
  local sceneDuration = WordBirdsConfiguration.GAME_DURATION_IN_SECONDS / 3
  self.background:start(sceneDuration)

  self.timer:start()
  waitFor(0.3)
  startNextQuestion(self)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

local function createGameElements(self)
  self.background = WordBirdsBackground.new()
  self:insertAtCenter(self.background, Z_INDEX.BACKGROUND)

  local topMarginOffset = 5
  local topMargin = self:getHeight() - self.timer:getHeight() - topMarginOffset
  local bottomMargin = self.background:getBottomMargin()

  local tapHandler = function(existingWord, newLetter)
    markSpeedyAnswerTimeIfValid(self, existingWord, newLetter)
    if self.showingTutorial then
      self.lastActionTime = self.timer:getTime()
    else
      self.tutorialEnabled = false
    end
  end

  self.inGameUi = WordBirdsInGameUi.new(topMargin,
                                        bottomMargin,
                                        function(answer) onSubmitAnswer(self, answer) end,
                                        function() onSkipButtonPressed(self) end,
                                        tapHandler)
  self:insertAtCenter(self.inGameUi, Z_INDEX.UI)

  local function quickResponseWrapper(earnedBonusProportion, earnedBonusProportionsSum)
    ScoringUtils.applySpeedBonus(earnedBonusProportion,
                                 earnedBonusProportionsSum,
                                 WordBirdsConfiguration.POINTS_PER_FULL_SPEED_BONUS,
                                 self.bonusTray)
  end

  self.responseSpeedTracker = ResponseSpeedTracker.new(WordBirdsConfiguration.SPEED_BONUS_WINDOW_IN_SECONDS,
                                                       quickResponseWrapper,
                                                       WordBirdsConfiguration.SPEED_BONUS_DECAY_START)
end

local function createUI(self)
  local builder = StandardUIBuilder.new(self:getUIOverlay(), WordBirdsConfiguration.UI_COMPONENTS_COLOR)
  builder:addPauseButton()
  self.scoreLabel = builder:addScoreLabel()
  self.bonusTray = builder:addBonusTray()
  self.timer = builder:addDigitTimer(self.timerDuration, function() onTimeout(self) end)
  self.timer:setTimeLostOpacityOverride(1.0)

  self.lastActionTime = 0
  self.tutorialEnabled = true
  self.showingTutorial = false

  local birdCounter = WordBirdsBirdCounter.new()
  builder:addWidget(birdCounter, UIOverlay.NORTH)
  self.birdCounter = birdCounter
end

-- Constructor ------------------------

function WordBirdsGame.new()
  WordBirdsButton.initLabelCache()

  local self = BaseGame.new()
  Mixin.mix(self, WordBirdsGame)

  configureGameDifficulty(self)
  createUI(self)
  createGameElements(self)

  self.conceptGenerator = LearningFocusedRoundGenerator.new(WordBirdsConceptGenerator.new())

  return self
end
