--- Word Birds In-game UI

class("WordBirdsInGameUi")
require("WordBirdsButtonSet")
require("WordBirdsQuestionLabel")
require("WordBirdsAnswerLabel")
require("WordBirdsInterstitial")
require("WordBirdsEditor")
require("WordBirdsNotePlayer")


-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local INCORRECT_SEQUENCE_DELAY = 1.7
local SKIP_SEQUENCE_DELAY = 1.0
local CORRECT_SEQUENCE_DELAY = 0.01

local COLOR_AUTOCORRECT = {1, 1, 1, 0.7}


-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function newScaleBoxInAction(box)
  local scaleUpDuration = 0.25
  local scaleDownDuration = 0.21
  return AHSequence.new(
    AHSpawn.new(
      AHScaleTo.new(box, scaleUpDuration, AHEaseSineOut.new(), 1.02),
      AHOpacityTo.new(box, scaleUpDuration, AHEaseSineOut.new(), 1)
    ),
    AHScaleTo.new(box, scaleDownDuration, AHEaseSineInOut.new(), 1)
  )
end

local function animateBoxesIn(self)
  return runInCoroutine(function()
    self.inOutActions:insert(AHStaggeredSpawn.new(0.08,
      newScaleBoxInAction(self.questionLabel),
      newScaleBoxInAction(self.answerLabel)
    ))
    self.inOutActions:wait()
  end)
end

local function newMoveBoxOutAction(self, box)
  local answerY = box:getY()
  local answerHeight = box:getHeight()
  local translateDown = -10
  local translateUp = self:getHeight() - answerY + translateDown + answerHeight

  return AHSequence.new(
    AHLocationYBy.new(box, 0.10, AHEaseSineInOut.new(), translateDown),
    AHLocationYBy.new(box, 0.26, AHEaseSineIn.new(), translateUp)
  )
end

local function animateBoxesOut(self)
  return runInCoroutine(function()
    self.inOutActions:insert(AHStaggeredSpawn.new(0.06,
      newMoveBoxOutAction(self, self.questionLabel),
      newMoveBoxOutAction(self, self.answerLabel)
    ))
    self.inOutActions:wait()
  end)
end

local function animateQuestionIn(self)
  return runInCoroutine(function()
    wait(
      self.buttonSet:animateIn(),
      animateBoxesIn(self)
    )
  end)
end

local function animateQuestionOut(self)
  return runInCoroutine(function()
    wait(
      self.buttonSet:animateOut(),
      animateBoxesOut(self)
    )
  end)
end

local function showInterstitial(self, message, showDuration, shouldDarken)
  return runInCoroutine(function()
    local interstitial = WordBirdsInterstitial.new(message)
    self:insertAtCenter(interstitial, self:getTopZIndex() + 1)

    wait(interstitial:animateIn(shouldDarken))
    waitFor(showDuration)
    wait(interstitial:animateOut())

    interstitial:removeSelf()
  end)
end

local function onAddLetter(self, letter)
  self.onTapAction(self.editor:getWord(), letter)
  self.editor:appendLetter(letter)
  self.answerLabel:setText(self.editor:getWord())
  self.buttonSet:setBackspaceEnabled(true)
  self.notePlayer:playNextNote()
end

local function onBackspace(self)
  if self.editor:canBackspace() then
    self.onTapAction()
    self.editor:backspace()
    self.answerLabel:setText(self.editor:getWord())
    self.buttonSet:reactivateLastLetterButton()
    self.buttonSet:setBackspaceRearm(self.editor:getBackspacesRemaining() > 0)
    self.notePlayer:playPreviousNote()
  end
end

local function onSubmit(self, submitCallback)
  self.onTapAction()
  submitCallback(self.editor:getWord())
end

local function onSkip(self, skipCallback)
  skipCallback()
end

local function newButtonSet(self, concept, submitCallback, skipCallback)
  local indexOfFirstUnrevealedCharacter = concept.prefillCount + 1

  local keyChars = {}
  for answer in pairs(concept.answerSet) do
    local unrevealedChars = TableUtils.createArrayFromString(answer:sub(indexOfFirstUnrevealedCharacter))
    for _, char in ipairs(unrevealedChars) do
      local count = TableUtils.tableCount(unrevealedChars, char)
      if keyChars[char] then
        keyChars[char] = math.max(keyChars[char], count)
      else
        keyChars[char] = count
      end
    end
  end

  local unrevealedWordString = ""
  for char, count in pairs(keyChars) do
    for _ = 1, count do
      unrevealedWordString = unrevealedWordString .. char
    end
  end

  local disallowedLetters = concept.disallowedLetters
  local submitHandler = function() onSubmit(self, submitCallback) end
  local skipHandler = function() onSkip(self, skipCallback) end
  local backspaceHandler = function() onBackspace(self) end
  local letterHandler = function(letter) onAddLetter(self, letter) end
  return WordBirdsButtonSet.new(unrevealedWordString, disallowedLetters, letterHandler, backspaceHandler, submitHandler, skipHandler)
end

local function animateAutocorrect(self)
  local label = self.autocorrectLabel
  local fullOpacity = COLOR_AUTOCORRECT[4]
  return AHSequence.new(
    AHOpacityTo.new(label, 0, nil, fullOpacity),
    AHDelay.new(0.4),
    AHOpacityTo.new(label, 0.3, nil, 0)
  ):start()
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WordBirdsInGameUi:setEnabled(isEnabled)
  self.buttonSet:setEnabled(isEnabled)
end

function WordBirdsInGameUi:showQuestion(concept)
  return runInCoroutine(function()
    local startScale = 0.4
    local columnSpacing = self:getHeightPart(0.05)

    local answerArray = TableUtils.getKeys(concept.answerSet)
    local answer = answerArray[1]
    self.editor:reset(string.sub(answer, 1, concept.prefillCount))
    self.notePlayer:reset()

    if self.questionView then
      self.questionView:removeSelf()
    end
    self.questionView = Column.new()
    self.workArea:insertAtCenter(self.questionView)

    self.questionLabel = WordBirdsQuestionLabel.new(concept.prompt)
    self.questionLabel:setOpacity(0)
    self.questionView:insert(self.questionLabel)

    local autocorrectText = Localizer.get("Misc.AUTOCORRECT"):lower()
    self.autocorrectLabel = Label.new(autocorrectText, Font.newWithActualSize(FontFaces.DIN_OT_BOLD, columnSpacing * 0.6))
    self.autocorrectLabel:setColor(unpack(COLOR_AUTOCORRECT))
    self.autocorrectLabel:setOpacity(0)

    local autocorrectPadding = (columnSpacing - self.autocorrectLabel:getHeight()) / 2

    self.questionView:insertSpacer(autocorrectPadding)
    self.questionView:insert(self.autocorrectLabel)
    self.questionView:insertSpacer(autocorrectPadding)


    self.answerLabel = WordBirdsAnswerLabel.new(answer)
    self.answerLabel:setText(self.editor:getWord())
    self.answerLabel:setOpacity(0)
    self.questionView:insert(self.answerLabel)
    self.questionView:insertSpacer(columnSpacing)

    self.buttonSet = newButtonSet(self, concept, self.onSubmit, self.onSkip)
    self.buttonSet:setBackspaceEnabled(false)
    self.questionView:insert(self.buttonSet)

    self.questionView:alignChildren()

    self.questionLabel:setScl(startScale)
    self.answerLabel:setScl(startScale)

    waitFor(0.3) -- Give labels time to render.
    animateQuestionIn(self)
  end)
end

function WordBirdsInGameUi:runCorrectSequence(answer, isExactMatch)
  return runInCoroutine(function()
    if not isExactMatch then
      animateAutocorrect(self)
      waitFor(0.3)
    end
    wait(self.answerLabel:markCorrect(answer))
    waitFor(CORRECT_SEQUENCE_DELAY)
    wait(animateQuestionOut(self))
  end)
end

function WordBirdsInGameUi:runIncorrectSequence()
  return runInCoroutine(function()
    self.buttonSet:markIncorrect()
    wait(self.answerLabel:markIncorrect())
    waitFor(INCORRECT_SEQUENCE_DELAY)
    wait(animateQuestionOut(self))
  end)
end

function WordBirdsInGameUi:runSkipSequence()
  return runInCoroutine(function()
    self.buttonSet:markIncorrect()
    wait(self.answerLabel:markSkipped())
    waitFor(SKIP_SEQUENCE_DELAY)
    wait(animateQuestionOut(self))
  end)
end

function WordBirdsInGameUi:runTimeoutSequence()
  return runInCoroutine(function()
    self.buttonSet:markIncorrect()
    wait(self.answerLabel:markTimeout())
    waitFor(INCORRECT_SEQUENCE_DELAY)
    wait(animateQuestionOut(self))
  end)
end

function WordBirdsInGameUi:runLongWordSequence()
  Sound.playSoundEffect("wordbirds_longword")
end

function WordBirdsInGameUi:removeAll()
  return runInCoroutine(function()
    animateBoxesOut(self)
    wait(self.buttonSet:animateOut())
  end)
end

function WordBirdsInGameUi:showBirdsRemaining(birdCount)
  return showInterstitial(self, birdCount .. " birds to go!", 1.5, true)
end

function WordBirdsInGameUi:showInstructions()
  return showInterstitial(self, "Clear all the birds", 0.8, false)
end


-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WordBirdsInGameUi.new(topMargin, bottomMargin, onSubmit, onSkip, onTapAction)
  local self = Group.newFullScreen()
  Mixin.mix(self, WordBirdsInGameUi)
  local selfWidth = self:getWidth()

  local uiW, uiH = selfWidth, topMargin - bottomMargin
  self.workArea = Group.new(uiW, uiH)
  self.workArea:setLoc(selfWidth/2, uiH/2 + bottomMargin)
  self:insert(self.workArea)

  self.notePlayer = WordBirdsNotePlayer.new()

  self.onSubmit = onSubmit
  self.onSkip = onSkip
  self.onTapAction = onTapAction
  self.editor = WordBirdsEditor.new()
  self.inOutActions = ActionInterrupt.new()

  return self
end
