--- WordBirdsValidator
-- @author Created by Anton Morant on April 17, 2014.
-- ConceptValidator for WordBirds.

class("WordBirdsValidator")

require("WordBirdsCommon")
require("validators.ConceptValidator")
require("validators.TagValidator")

-- ----------------------------------------------------------------------------
-- Configuration
-- ----------------------------------------------------------------------------

local REQUIRED_KEYS = {
  'prompt',
  'answer',
  'contentDifficulty'
}

local REQUIRED_EMPTY_ALLOWED_KEYS = {
  'disallowedLetters',
  'alternateForms',
  'rationale',
  'acceptedMisspellings'
}

local RATIONALE_OPTINAL_PAIRED_TAGS = {"highlight1", "highlight2"}
local RATIONALE_OPTIONAL_SINGLE_TAGS = {"br"}

local MINIMUM_WORD_LENGTH = 7
local MAXIMUM_WORD_LENGTH = 10
local MAXIMUM_PROMPT_LENGTH = 50
local MAXIMUM_LETTER_BUTTONS = 10

-- ----------------------------------------------------------------------------
-- Utility
-- ----------------------------------------------------------------------------

local function validateLengths(self, concept, answerArray)
  local prompt = concept.prompt

  for _, answer in ipairs(answerArray) do
    self:expectNumberInRange(#answer, MINIMUM_WORD_LENGTH, MAXIMUM_WORD_LENGTH, string.format("length of answer (%s)", answer))
  end

  self:expectLessOrEqual(#prompt, MAXIMUM_PROMPT_LENGTH, string.format("length of prompt (%s)", prompt))
end

local function validatePrefixes(self, answerArray)
  local prefix = answerArray[1]:sub(1, WordBirdsCommon.MAX_CHAR_COUNT)

  for _, answer in ipairs(answerArray) do
    self:expectEqual(answer:sub(1, WordBirdsCommon.MAX_CHAR_COUNT), prefix, string.format("First %d characters of '%s' do not match prefix '%s'", WordBirdsCommon.MAX_CHAR_COUNT, answer, prefix))
  end
end

local function validateTotalCharCount(self, answerArray)
  local minPrefixCount = WordBirdsCommon.MAX_CHAR_COUNT - WordBirdsCommon.MAX_DIFFICULTY
  local charButtons = {}

  for _, answer in ipairs(answerArray) do
    local answerChars = TableUtils.createArrayFromString(answer:sub(minPrefixCount + 1))
    for _, char in ipairs(answerChars) do
      local count = TableUtils.tableCount(answerChars, char)
      if (charButtons[char]) then
        charButtons[char] = math.max(charButtons[char], count)
      else
        charButtons[char] = count
      end
    end
  end

  local totalCount = TableUtils.sum(TableUtils.getValues(charButtons))
  self:expectLessOrEqual(totalCount, MAXIMUM_LETTER_BUTTONS, string.format("Answers have %d letters that need buttons, but only %d buttons available", totalCount, MAXIMUM_LETTER_BUTTONS))
end

local function validateRationales(self, concept)
  local alternateForms = concept.alternateForms
  local rationale = concept.rationale

  local hasAlternateForms = ((alternateForms ~= nil) and (alternateForms ~= ""))
  local hasRationale = ((rationale ~= nil) and (rationale ~= ""))

  if (hasAlternateForms) then
    self:assertTrue(hasRationale, "rationale must be defined for concepts with alternateForms")
    self:validateTags(rationale, nil, RATIONALE_OPTINAL_PAIRED_TAGS, nil, nil, RATIONALE_OPTIONAL_SINGLE_TAGS)
    self:expectStringIsPrintable(rationale, "rationale")
    self:expectNoDoubleSpaces(rationale, "rationale")
  end
end

-- ----------------------------------------------------------------------------
-- API
-- ----------------------------------------------------------------------------

function WordBirdsValidator:validateConcept(concept)
  self:validateRequiredKeys(concept, REQUIRED_KEYS)
  self:validateRequiredKeys(concept, REQUIRED_EMPTY_ALLOWED_KEYS, true)

  local answerArray = {}
  if ((concept.answer ~= nil) and (concept.answer ~= "")) then
    answerArray = TableUtils.csvToArray(concept.answer)
  end

  self:expectGreaterOrEqual(#answerArray, 1, "Must provide at least one answer in vocab_word")
  self:expectNoDuplicates(answerArray, "vocab_word")
  validateLengths(self, concept, answerArray)
  validatePrefixes(self, answerArray)
  validateTotalCharCount(self, answerArray)
  validateRationales(self, concept)
end

-- ----------------------------------------------------------------------------
-- Object Creation
-- ----------------------------------------------------------------------------

-- Constructor ------------------------

function WordBirdsValidator.new()
  local self = ConceptValidator.new()
  Mixin.mix(self, WordBirdsValidator)
  TagValidator.make(self)

  return self
end
